/*
MIT License

Copyright (C) 2022 Ryan L. Guy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include "fluidsimulation.h"

#include <cstring>
#include <iomanip>

#include "threadutils.h"
#include "stopwatch.h"
#include "openclutils.h"
#include "viscositysolver.h"
#include "particlemesher.h"
#include "polygonizer3d.h"
#include "diffuseparticle.h"
#include "particlemaskgrid.h"
#include "triangle.h"
#include "scalarfield.h"
#include "particlesheeter.h"
#include "versionutils.h"
#include "interpolation.h"
#include "gridutils.h"
#include "attributetogridtransfer.h"
#include "mixbox/mixbox.h"


FluidSimulation::FluidSimulation() {
}

FluidSimulation::FluidSimulation(int isize, int jsize, int ksize, double dx) :
                                _isize(isize), _jsize(jsize), _ksize(ksize), _dx(dx) {
    _logGreeting();
}

FluidSimulation::~FluidSimulation() {
}

/*******************************************************************************
    PUBLIC
********************************************************************************/

void FluidSimulation::getVersion(int *major, int *minor, int *revision) { 
    VersionUtils::getVersion(major, minor, revision);
}

void FluidSimulation::upscaleOnInitialization(int isizePrev, int jsizePrev, int ksizePrev, double dxPrev) { 
    if (isizePrev <= 0 || jsizePrev <= 0 || ksizePrev <= 0 || dxPrev <= 0.0) {
        std::string msg = "Error: dimensions and cell size must be greater than 0.\n";
        msg += "grid: " + _toString(isizePrev) + _toString(jsizePrev) + _toString(ksizePrev) + " " + _toString(dxPrev) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " upscaleOnInitialization: " << isizePrev << " " << jsizePrev << " " << ksizePrev << " " << dxPrev << std::endl);

    _isUpscalingOnInitializationEnabled = true;
    _upscalingPreviousIsize = isizePrev;
    _upscalingPreviousJsize = jsizePrev;
    _upscalingPreviousKsize = ksizePrev;
    _upscalingPreviousCellSize = dxPrev;
}

void FluidSimulation::initialize() {
    if (!_isSimulationInitialized) {
        _logfile.log(std::ostringstream().flush() << 
                     _logfile.getTime() << " initialize" << std::endl);
        _initializeSimulation();
    }
}

bool FluidSimulation::isInitialized() {
    return _isSimulationInitialized;
}

void FluidSimulation::initializeMixbox(FluidSimulationMixboxLutData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " initializeMixbox: " << data.size << std::endl);

    mixbox_initialize(data.data, data.size);

    // Output a simple test to the logfile
    unsigned char r,g,b;
    mixbox_lerp_srgb8(252, 211, 0, 0, 0, 96, 0.5f, &r,&g,&b);
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " initializeMixboxTest: " << (int)r << " " << (int)g << " " << (int)b << std::endl);
}

int FluidSimulation::getCurrentFrame() {
    return _currentFrame;
}

void FluidSimulation::setCurrentFrame(int frameno) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setCurrentFrame: " << frameno << std::endl);

    _currentFrame = frameno;
}

void FluidSimulation::setTimelineFrameStart(int frameno) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setTimelineFrameStart: " << frameno << std::endl);

    _timelineFrameStart = frameno;
}

void FluidSimulation::setTimelineFrameEnd(int frameno) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setTimelineFrameEnd: " << frameno << std::endl);

    _timelineFrameEnd = frameno;
}

bool FluidSimulation::isCurrentFrameFinished() { 
    return _isCurrentFrameFinished; 
}

double FluidSimulation::getCellSize() { 
    return _dx; 
}

void FluidSimulation::getGridDimensions(int *i, int *j, int *k) { 
    *i = _isize; *j = _jsize; *k = _ksize; 
}

int FluidSimulation::getGridWidth() {
    return _isize;
}

int FluidSimulation::getGridHeight() {
    return _jsize;
}

int FluidSimulation::getGridDepth() {
    return _ksize;
}

void FluidSimulation::getSimulationDimensions(double *w, double *h, double *d) { 
    *w = (double)_isize*_dx;
    *h = (double)_jsize*_dx;
    *d = (double)_ksize*_dx;
}

double FluidSimulation::getSimulationWidth() {  
    return (double)_isize*_dx; 
}

double FluidSimulation::getSimulationHeight() { 
    return (double)_jsize*_dx; 
}

double FluidSimulation::getSimulationDepth() {  
    return (double)_ksize*_dx; 
}

double FluidSimulation::getDensity() { 
    return _density; 
}

void FluidSimulation::setDensity(double p) { 
    if (p <= 0.0) {
        std::string msg = "Error: density must be greater than 0.\n";
        msg += "density: " + _toString(p) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDensity: " << p << std::endl);

    _density = p; 
}

double FluidSimulation::getMarkerParticleScale() {
    return _markerParticleScale;
}

void FluidSimulation::setMarkerParticleScale(double s) { 
    if (s < 0.0) {
        std::string msg = "Error: marker particle scale must be greater than or equal to 0.\n";
        msg += "scale: " + _toString(s) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMarkerParticleScale: " << s << std::endl);

    _markerParticleScale = s; 
}

double FluidSimulation::getMarkerParticleJitterFactor() {
    return _markerParticleJitterFactor;
}

void FluidSimulation::setMarkerParticleJitterFactor(double jit) { 
    if (jit < 0.0) {
        std::string msg = "Error: marker particle jitter must be greater than or equal to 0.\n";
        msg += "jitter: " + _toString(jit) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMarkerParticleJitterFactor: " << jit << std::endl);

    _markerParticleJitterFactor = jit; 
}

void FluidSimulation::enableJitterSurfaceMarkerParticles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableJitterSurfaceMarkerParticles" << std::endl);

    _isJitterSurfaceMarkerParticlesEnabled = true;
}

void FluidSimulation::disableJitterSurfaceMarkerParticles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableJitterSurfaceMarkerParticles" << std::endl);

    _isJitterSurfaceMarkerParticlesEnabled = false;
}

bool FluidSimulation::isJitterSurfaceMarkerParticlesEnabled() {
    return _isJitterSurfaceMarkerParticlesEnabled;
}

int FluidSimulation::getPressureSolverMaxIterations() {
    return _maxPressureSolveIterations;
}

void FluidSimulation::setPressureSolverMaxIterations(int n) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setPressureSolverMaxIterations: " << n << std::endl);

    _maxPressureSolveIterations = n;
}

int FluidSimulation::getViscositySolverMaxIterations() {
    return _maxViscositySolveIterations;
}

void FluidSimulation::setViscositySolverMaxIterations(int n) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setViscositySolverMaxIterations: " << n << std::endl);

    _maxViscositySolveIterations = n;
}

int FluidSimulation::getSurfaceSubdivisionLevel() {
    return _outputFluidSurfaceSubdivisionLevel;
}

void FluidSimulation::setSurfaceSubdivisionLevel(int n) {
    if (n < 1) {
        std::string msg = "Error: subdivision level must be greater than or equal to 1.\n";
        msg += "subdivision level: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSurfaceSubdivisionLevel: " << n << std::endl);

    _outputFluidSurfaceSubdivisionLevel = n;
}

int FluidSimulation::getNumPolygonizerSlices() {
    return _numSurfaceReconstructionPolygonizerSlices;
}

void FluidSimulation::setNumPolygonizerSlices(int n) {
    if (n < 1) {
        std::string msg = "Error: number of polygonizer slices must be greater than or equal to 1.\n";
        msg += "polygonizer slices: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setNumPolygonizerSlices: " << n << std::endl);

    _numSurfaceReconstructionPolygonizerSlices = n;
}

double FluidSimulation::getSurfaceSmoothingValue() {
    return _surfaceReconstructionSmoothingValue;
}

void FluidSimulation::setSurfaceSmoothingValue(double s) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSurfaceSmoothingValue: " << s << std::endl);

    _surfaceReconstructionSmoothingValue = s;
}

int FluidSimulation::getSurfaceSmoothingIterations() {
    return _surfaceReconstructionSmoothingIterations;
}

void FluidSimulation::setSurfaceSmoothingIterations(int n) {
    if (n < 0) {
        std::string msg = "Error: number of smoothing iterations must be positive.\n";
        msg += "smoothing iterations: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSurfaceSmoothingIterations: " << n << std::endl);

    _surfaceReconstructionSmoothingIterations = n;
}

void FluidSimulation::setMeshingVolume(MeshObject *volumeObject) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMeshingVolume: " << volumeObject << std::endl);

    _meshingVolume = volumeObject;
    _isMeshingVolumeSet = true;
}

int FluidSimulation::getMinPolyhedronTriangleCount() {
    return _minimumSurfacePolyhedronTriangleCount;
}

void FluidSimulation::setMinPolyhedronTriangleCount(int n) {
    if (n < 0) {
        std::string msg = "Error: minimum polyhedron triangle count must be greater than or equal to 0.\n";
        msg += "triangle count: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinPolyhedronTriangleCount: " << n << std::endl);

    _minimumSurfacePolyhedronTriangleCount = n;
}

vmath::vec3 FluidSimulation::getDomainOffset() {
    return _domainOffset;
}

void FluidSimulation::setDomainOffset(double x, double y, double z) {
    setDomainOffset(vmath::vec3(x, y, z));
}

void FluidSimulation::setDomainOffset(vmath::vec3 offset) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDomainOffset: " << 
                 offset.x << " " << offset.y << " " << offset.z << std::endl);

    _domainOffset = offset;
    _diffuseMaterial.setDomainOffset(offset);
}

double FluidSimulation::getDomainScale() {
    return _domainScale;
}

void FluidSimulation::setDomainScale(double scale) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDomainScale: " << scale << std::endl);

    _domainScale = scale;
    _diffuseMaterial.setDomainScale(scale);
}

void FluidSimulation::setMeshOutputFormatAsPLY() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMeshOutputFormatAsPLY" << std::endl);

    _meshOutputFormat = TriangleMeshFormat::ply;
}

void FluidSimulation::setMeshOutputFormatAsBOBJ() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMeshOutputFormatAsBOBJ" << std::endl);

    _meshOutputFormat = TriangleMeshFormat::bobj;
}

void FluidSimulation::enableConsoleOutput() {
    _logfile.enableConsole();

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableConsoleOutput" << std::endl);
}

void FluidSimulation::disableConsoleOutput() {
    _logfile.disableConsole();

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableConsoleOutput" << std::endl);
}

bool FluidSimulation::isConsoleOutputEnabled() {
    return _logfile.isConsoleEnabled();
}

void FluidSimulation::enableSurfaceReconstruction() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceReconstruction" << std::endl);

    _isSurfaceMeshReconstructionEnabled = true;
}

void FluidSimulation::disableSurfaceReconstruction() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceReconstruction" << std::endl);

    _isSurfaceMeshReconstructionEnabled = false;
}

bool FluidSimulation::isSurfaceReconstructionEnabled() {
    return _isSurfaceMeshReconstructionEnabled;
}

void FluidSimulation::enableAsynchronousMeshing() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableAsynchronousMeshing" << std::endl);

    _isAsynchronousMeshingEnabled = true;
}

void FluidSimulation::disableAsynchronousMeshing() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableAsynchronousMeshing" << std::endl);

    _isAsynchronousMeshingEnabled = false;
}

bool FluidSimulation::isAsynchronousMeshingEnabled() {
    return _isAsynchronousMeshingEnabled;
}

void FluidSimulation::enablePreviewMeshOutput(double cellsize) {
    if (cellsize <= 0.0) {
        std::string msg = "Error: cell size must be greater than 0.0.\n";
        msg += "cellsize: " + _toString(cellsize) + "\n";
        throw std::domain_error(msg);
    }

     _logfile.log(std::ostringstream().flush() << 
             _logfile.getTime() << " enablePreviewMeshOutput: " << cellsize << std::endl);

    _isPreviewSurfaceMeshEnabled = true;
    _previewdx = cellsize;
}

void FluidSimulation::disablePreviewMeshOutput() {
     _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disablePreviewMeshOutput" << std::endl);

    _isPreviewSurfaceMeshEnabled = false;
}

bool FluidSimulation::isPreviewMeshOutputEnabled() {
    return _isPreviewSurfaceMeshEnabled;
}

void FluidSimulation::enableObstacleMeshingOffset() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableObstacleMeshingOffset" << std::endl);

    _isObstacleMeshingOffsetEnabled = true;
}

void FluidSimulation::disableObstacleMeshingOffset() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableObstacleMeshingOffset" << std::endl);

    _isObstacleMeshingOffsetEnabled = false;
}

bool FluidSimulation::isObstacleMeshingOffsetEnabled() {
    return _isObstacleMeshingOffsetEnabled;
}

double FluidSimulation::getObstacleMeshingOffset() {
    return _obstacleMeshingOffset;
}

void FluidSimulation::setObstacleMeshingOffset(double s) { 
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setObstacleMeshingOffset: " << s << std::endl);

    _obstacleMeshingOffset = s; 
}

void FluidSimulation::enableInvertedContactNormals() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableInvertedContactNormals" << std::endl);

    _isInvertedContactNormalsEnabled = true;
}

void FluidSimulation::disableInvertedContactNormals() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableInvertedContactNormals" << std::endl);

    _isInvertedContactNormalsEnabled = false;
}

bool FluidSimulation::isInvertedContactNormalsEnabled() {
    return _isInvertedContactNormalsEnabled;
}

void FluidSimulation::enableSurfaceMotionBlur() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceMotionBlur" << std::endl);

    _isSurfaceMotionBlurEnabled = true;
}

void FluidSimulation::disableSurfaceMotionBlur() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceMotionBlur" << std::endl);

    _isSurfaceMotionBlurEnabled = false;
}

bool FluidSimulation::isSurfaceMotionBlurEnabled() {
    return _isSurfaceMotionBlurEnabled;
}

void FluidSimulation::enableWhitewaterMotionBlur() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableWhitewaterMotionBlur" << std::endl);

    _isWhitewaterMotionBlurEnabled = true;
}

void FluidSimulation::disableWhitewaterMotionBlur() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableWhitewaterMotionBlur" << std::endl);

    _isWhitewaterMotionBlurEnabled = false;
}

bool FluidSimulation::isWhitewaterMotionBlurEnabled() {
    return _isWhitewaterMotionBlurEnabled;
}

void FluidSimulation::enableSurfaceVelocityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceVelocityAttribute" << std::endl);

    _isSurfaceVelocityAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceVelocityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceVelocityAttribute" << std::endl);

    _isSurfaceVelocityAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceVelocityAttributeEnabled() {
    return _isSurfaceVelocityAttributeEnabled;
}

void FluidSimulation::enableSurfaceVelocityAttributeAgainstObstacles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceVelocityAttributeAgainstObstacles" << std::endl);

    _isSurfaceVelocityAttributeAgainstObstaclesEnabled = true;
}

void FluidSimulation::disableSurfaceVelocityAttributeAgainstObstacles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceVelocityAttributeAgainstObstacles" << std::endl);

    _isSurfaceVelocityAttributeAgainstObstaclesEnabled = false;
}

bool FluidSimulation::isSurfaceVelocityAttributeAgainstObstaclesEnabled() {
    return _isSurfaceVelocityAttributeAgainstObstaclesEnabled;
}

void FluidSimulation::enableWhitewaterVelocityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableWhitewaterVelocityAttribute" << std::endl);

    _isWhitewaterVelocityAttributeEnabled = true;
}

void FluidSimulation::disableWhitewaterVelocityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableWhitewaterVelocityAttribute" << std::endl);

    _isWhitewaterVelocityAttributeEnabled = false;
}

bool FluidSimulation::isWhitewaterVelocityAttributeEnabled() {
    return _isWhitewaterVelocityAttributeEnabled;
}

void FluidSimulation::enableSurfaceVorticityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceVorticityAttribute" << std::endl);

    _isSurfaceVorticityAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceVorticityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceVorticityAttribute" << std::endl);

    _isSurfaceVorticityAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceVorticityAttributeEnabled() {
    return _isSurfaceVorticityAttributeEnabled;
}

void FluidSimulation::enableSurfaceSpeedAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceSpeedAttribute" << std::endl);

    _isSurfaceSpeedAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceSpeedAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceSpeedAttribute" << std::endl);

    _isSurfaceSpeedAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceSpeedAttributeEnabled() {
    return _isSurfaceSpeedAttributeEnabled;
}

void FluidSimulation::enableSurfaceAgeAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceAgeAttribute" << std::endl);

    _isSurfaceAgeAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceAgeAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceAgeAttribute" << std::endl);

    _isSurfaceAgeAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceAgeAttributeEnabled() {
    return _isSurfaceAgeAttributeEnabled;
}

double FluidSimulation::getSurfaceAgeAttributeRadius() {
    return _ageAttributeRadius;
}

void FluidSimulation::setSurfaceAgeAttributeRadius(double r) {
    if (r <= 0.0) {
        std::string msg = "Error: Radius must be greater than or equal to 0.0.\n";
        msg += "radius: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setSurfaceAgeAttributeRadius: " << r << std::endl);

    _ageAttributeRadius = r;
}

void FluidSimulation::enableSurfaceColorAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceColorAttribute" << std::endl);

    _isSurfaceSourceColorAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceColorAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceColorAttribute" << std::endl);

    _isSurfaceSourceColorAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceColorAttributeEnabled() {
    return _isSurfaceSourceColorAttributeEnabled;
}

double FluidSimulation::getSurfaceColorAttributeRadius() {
    return _colorAttributeRadius;
}

void FluidSimulation::setSurfaceColorAttributeRadius(double r) {
    if (r <= 0.0) {
        std::string msg = "Error: Radius must be greater than or equal to 0.0.\n";
        msg += "radius: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setSurfaceColorAttributeRadius: " << r << std::endl);

    _colorAttributeRadius = r;
}

void FluidSimulation::enableSurfaceColorAttributeMixing() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceColorAttributeMixing" << std::endl);

    _isSurfaceSourceColorAttributeMixingEnabled = true;
}

void FluidSimulation::disableSurfaceColorAttributeMixing() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceColorAttributeMixing" << std::endl);

    _isSurfaceSourceColorAttributeMixingEnabled = false;
}

bool FluidSimulation::isSurfaceColorAttributeMixingEnabled() {
    return _isSurfaceSourceColorAttributeMixingEnabled;
}

double FluidSimulation::getSurfaceColorAttributeMixingRadius() {
    return _colorAttributeMixingRadius;
}

void FluidSimulation::setSurfaceColorAttributeMixingRadius(double r) {
    if (r <= 0.0) {
        std::string msg = "Error: Radius must be greater than or equal to 0.0.\n";
        msg += "radius: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setSurfaceColorAttributeMixingRadius: " << r << std::endl);

    _colorAttributeMixingRadius = r;
}

double FluidSimulation::getSurfaceColorAttributeMixingRate() {
    return _colorAttributeMixingRate;
}

void FluidSimulation::setSurfaceColorAttributeMixingRate(double r) {
    if (r <= 0.0) {
        std::string msg = "Error: Rate must be greater than or equal to 0.0.\n";
        msg += "rate: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setSurfaceColorAttributeMixingRate: " << r << std::endl);

    _colorAttributeMixingRate = r;
}

void FluidSimulation::enableMixbox() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableMixbox" << std::endl);

    _isMixboxEnabled = true;
}

void FluidSimulation::disableMixbox() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableMixbox" << std::endl);

    _isMixboxEnabled = false;
}

bool FluidSimulation::isMixboxEnabled() {
    return _isMixboxEnabled;
}

void FluidSimulation::enableSurfaceSourceIDAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceSourceIDAttribute" << std::endl);

    _isSurfaceSourceIDAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceSourceIDAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceSourceIDAttribute" << std::endl);

    _isSurfaceSourceIDAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceSourceIDAttributeEnabled() {
    return _isSurfaceSourceIDAttributeEnabled;
}

void FluidSimulation::enableSurfaceViscosityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSurfaceViscosityAttribute" << std::endl);

    _isSurfaceSourceViscosityAttributeEnabled = true;
}

void FluidSimulation::disableSurfaceViscosityAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSurfaceViscosityAttribute" << std::endl);

    _isSurfaceSourceViscosityAttributeEnabled = false;
}

bool FluidSimulation::isSurfaceViscosityAttributeEnabled() {
    return _isSurfaceSourceViscosityAttributeEnabled;
}

void FluidSimulation::enableWhitewaterIDAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableWhitewaterIDAttribute" << std::endl);

    _isWhitewaterIDAttributeEnabled = true;
}

void FluidSimulation::disableWhitewaterIDAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableWhitewaterIDAttribute" << std::endl);

    _isWhitewaterIDAttributeEnabled = false;
}

bool FluidSimulation::isWhitewaterIDAttributeEnabled() {
    return _isWhitewaterIDAttributeEnabled;
}

void FluidSimulation::enableWhitewaterLifetimeAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableWhitewaterLifetimeAttribute" << std::endl);

    _isWhitewaterLifetimeAttributeEnabled = true;
}

void FluidSimulation::disableWhitewaterLifetimeAttribute() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableWhitewaterLifetimeAttribute" << std::endl);

    _isWhitewaterLifetimeAttributeEnabled = false;
}

bool FluidSimulation::isWhitewaterLifetimeAttributeEnabled() {
    return _isWhitewaterLifetimeAttributeEnabled;
}

void FluidSimulation::enableRemoveSurfaceNearDomain() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableRemoveSurfaceNearDomain" << std::endl);

    _isRemoveSurfaceNearDomainEnabled = true;
}

void FluidSimulation::disableRemoveSurfaceNearDomain() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableRemoveSurfaceNearDomain" << std::endl);

    _isRemoveSurfaceNearDomainEnabled = false;
}

bool FluidSimulation::isRemoveSurfaceNearDomainEnabled() {
    return _isRemoveSurfaceNearDomainEnabled;
}

int FluidSimulation::getRemoveSurfaceNearDomainDistance() {
    return _removeSurfaceNearDomainDistance;
}

void FluidSimulation::setRemoveSurfaceNearDomainDistance(int n) {
    if (n < 0) {
        std::string msg = "Error: distance must be greater than or equal to zero.\n";
        msg += "distance: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setRemoveSurfaceNearDomainDistance: " << n << std::endl);

    _removeSurfaceNearDomainDistance = n;
}

void FluidSimulation::enableFluidParticleOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableFluidParticleOutput" << std::endl);

    _isFluidParticleOutputEnabled = true;
}

void FluidSimulation::disableFluidParticleOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableFluidParticleOutput" << std::endl);

    _isFluidParticleOutputEnabled = false;
}

bool FluidSimulation::isFluidParticleOutputEnabled() {
    return _isFluidParticleOutputEnabled;
}

void FluidSimulation::enableInternalObstacleMeshOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableInternalObstacleMeshOutput" << std::endl);

    _isInternalObstacleMeshOutputEnabled = true;
}

void FluidSimulation::disableInternalObstacleMeshOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableInternalObstacleMeshOutput" << std::endl);

    _isInternalObstacleMeshOutputEnabled = false;
}

bool FluidSimulation::isInternalObstacleMeshOutputEnabled() {
    return _isInternalObstacleMeshOutputEnabled;
}

void FluidSimulation::enableForceFieldDebugOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableForceFieldDebugOutput" << std::endl);

    _isForceFieldDebugOutputEnabled = true;
}

void FluidSimulation::disableForceFieldDebugOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableForceFieldDebugOutput" << std::endl);

    _isForceFieldDebugOutputEnabled = false;
}

bool FluidSimulation::isForceFieldDebugOutputEnabled() {
    return _isForceFieldDebugOutputEnabled;
}

void FluidSimulation::enableDiffuseMaterialOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseMaterialOutput" << std::endl);

    _isDiffuseMaterialOutputEnabled = true;
}

void FluidSimulation::disableDiffuseMaterialOutput() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseMaterialOutput" << std::endl);

    _isDiffuseMaterialOutputEnabled = false;
}

bool FluidSimulation::isDiffuseMaterialOutputEnabled() {
    return _isDiffuseMaterialOutputEnabled;
}

void FluidSimulation::enableDiffuseParticleEmission() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseParticleEmission" << std::endl);

    _diffuseMaterial.enableDiffuseParticleEmission();
}

void FluidSimulation::disableDiffuseParticleEmission() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseParticleEmission" << std::endl);

    _diffuseMaterial.disableDiffuseParticleEmission();
}

bool FluidSimulation::isDiffuseParticleEmissionEnabled() {
    return _diffuseMaterial.isDiffuseParticleEmissionEnabled();
}

void FluidSimulation::enableDiffuseFoam() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseFoam" << std::endl);

    _diffuseMaterial.enableFoam();
}

void FluidSimulation::disableDiffuseFoam() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseFoam" << std::endl);

    _diffuseMaterial.disableFoam();
}

bool FluidSimulation::isDiffuseFoamEnabled() {
    return _diffuseMaterial.isFoamEnabled();
}

void FluidSimulation::enableDiffuseBubbles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseBubbles" << std::endl);

    _diffuseMaterial.enableBubbles();
}

void FluidSimulation::disableDiffuseBubbles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseBubbles" << std::endl);

    _diffuseMaterial.disableBubbles();
}

bool FluidSimulation::isDiffuseBubblesEnabled() {
    return _diffuseMaterial.isBubblesEnabled();
}

void FluidSimulation::enableDiffuseSpray() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseSpray" << std::endl);

    _diffuseMaterial.enableSpray();
}

void FluidSimulation::disableDiffuseSpray() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseSpray" << std::endl);

    _diffuseMaterial.disableSpray();
}

bool FluidSimulation::isDiffuseSprayEnabled() {
    return _diffuseMaterial.isSprayEnabled();
}


void FluidSimulation::enableDiffuseDust() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffuseDust" << std::endl);

    _diffuseMaterial.enableDust();
}

void FluidSimulation::disableDiffuseDust() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffuseDust" << std::endl);

    _diffuseMaterial.disableDust();
}

bool FluidSimulation::isDiffuseDustEnabled() {
    return _diffuseMaterial.isDustEnabled();
}

void FluidSimulation::enableBoundaryDiffuseDustEmission() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableBoundaryDiffuseDustEmission" << std::endl);

    _diffuseMaterial.enableBoundaryDustEmission();
}

void FluidSimulation::disableBoundaryDiffuseDustEmission() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableBoundaryDiffuseDustEmission" << std::endl);

    _diffuseMaterial.disableBoundaryDustEmission();
}

bool FluidSimulation::isBoundaryDustDiffuseEmissionEnabled() {
    return _diffuseMaterial.isBoundaryDustEmissionEnabled();
}

void FluidSimulation::enableBubbleDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableBubbleDiffuseMaterial" << std::endl);

    _isBubbleDiffuseMaterialEnabled = true;
}

void FluidSimulation::enableSprayDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSprayDiffuseMaterial" << std::endl);

    _isSprayDiffuseMaterialEnabled = true;
}

void FluidSimulation::enableFoamDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableFoamDiffuseMaterial" << std::endl);

    _isFoamDiffuseMaterialEnabled = true;
}

void FluidSimulation::disableBubbleDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableBubbleDiffuseMaterial" << std::endl);

    _isBubbleDiffuseMaterialEnabled = false;
}

void FluidSimulation::disableSprayDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSprayDiffuseMaterial" << std::endl);

    _isSprayDiffuseMaterialEnabled = false;
}

void FluidSimulation::disableFoamDiffuseMaterial() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableFoamDiffuseMaterial" << std::endl);

    _isFoamDiffuseMaterialEnabled = false;
}

bool FluidSimulation::isBubbleDiffuseMaterialEnabled() {
    return _isBubbleDiffuseMaterialEnabled;
}

bool FluidSimulation::isSprayDiffuseMaterialEnabled() {
    return _isSprayDiffuseMaterialEnabled;
}

bool FluidSimulation::isFoamDiffuseMaterialEnabled() {
    return _isFoamDiffuseMaterialEnabled;
}

void FluidSimulation::outputDiffuseMaterialAsSeparateFiles() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " outputDiffuseMaterialAsSeparateFiles" << std::endl);

    _isDiffuseMaterialFilesSeparated = true;
}

void FluidSimulation::outputDiffuseMaterialAsSingleFile() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " outputDiffuseMaterialAsSingleFile" << std::endl);

    _isDiffuseMaterialFilesSeparated = false;
}

bool FluidSimulation::isDiffuseMaterialOutputAsSeparateFiles() {
    return _isDiffuseMaterialFilesSeparated;
}

double FluidSimulation::getDiffuseEmitterGenerationRate() {
    return _diffuseMaterial.getEmitterGenerationRate();
}

void FluidSimulation::setDiffuseEmitterGenerationRate(double rate) {
    if (rate < 0.0 || rate > 1.0) {
        std::string msg = "Error: emitter generation rate must be in range [0.0, 1.0].\n";
        msg += "rate: " + _toString(rate) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseEmitterGenerationRate: " << rate << std::endl);

    _diffuseMaterial.setEmitterGenerationRate(rate);
}

double FluidSimulation::getMinDiffuseEmitterEnergy() {
    return _diffuseMaterial.getMinEmitterEnergy();
}

void FluidSimulation::setMinDiffuseEmitterEnergy(double e) {
    if (e < 0) {
        std::string msg = "Error: min diffuse emitter energy must be greater than or equal to 0.\n";
        msg += "energy: " + _toString(e) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinDiffuseEmitterEnergy: " << e << std::endl);

    _diffuseMaterial.setMinEmitterEnergy(e);
}

double FluidSimulation::getMaxDiffuseEmitterEnergy() {
    return _diffuseMaterial.getMaxEmitterEnergy();
}

void FluidSimulation::setMaxDiffuseEmitterEnergy(double e) {
    if (e < 0) {
        std::string msg = "Error: max diffuse emitter energy must be greater than or equal to 0.\n";
        msg += "energy: " + _toString(e) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxDiffuseEmitterEnergy: " << e << std::endl);

    _diffuseMaterial.setMaxEmitterEnergy(e);
}

double FluidSimulation::getMinDiffuseWavecrestCurvature() {
    return _diffuseMaterial.getMinWavecrestCurvature();
}

void FluidSimulation::setMinDiffuseWavecrestCurvature(double k) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinDiffuseWavecrestCurvature: " << k << std::endl);

    _diffuseMaterial.setMinWavecrestCurvature(k);
}

double FluidSimulation::getMaxDiffuseWavecrestCurvature() {
    return _diffuseMaterial.getMaxWavecrestCurvature();
}

void FluidSimulation::setMaxDiffuseWavecrestCurvature(double k) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxDiffuseWavecrestCurvature: " << k << std::endl);

    _diffuseMaterial.setMaxWavecrestCurvature(k);
}

double FluidSimulation::getMinDiffuseTurbulence() {
    return _diffuseMaterial.getMinTurbulence();
}

void FluidSimulation::setMinDiffuseTurbulence(double t) {
    if (t < 0) {
        std::string msg = "Error: min diffuse turbulence must be greater than or equal to 0.\n";
        msg += "turbulence: " + _toString(t) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinDiffuseTurbulence: " << t << std::endl);

    _diffuseMaterial.setMinTurbulence(t);
}

double FluidSimulation::getMaxDiffuseTurbulence() {
    return _diffuseMaterial.getMaxTurbulence();
}

void FluidSimulation::setMaxDiffuseTurbulence(double t) {
    if (t < 0) {
        std::string msg = "Error: max diffuse turbulence must be greater than or equal to 0.\n";
        msg += "turbulence: " + _toString(t) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxDiffuseTurbulence: " << t << std::endl);

    _diffuseMaterial.setMaxTurbulence(t);
}

int FluidSimulation::getMaxNumDiffuseParticles() {
    return _diffuseMaterial.getMaxNumDiffuseParticles();
}

void FluidSimulation::setMaxNumDiffuseParticles(int n) {
    if (n < 0) {
        std::string msg = "Error: maxNumDiffuseParticles must be greater than or equal to 0.\n";
        msg += "n: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxNumDiffuseParticles: " << n << std::endl);

    _diffuseMaterial.setMaxNumDiffuseParticles(n);
}

AABB FluidSimulation::getDiffuseEmitterGenerationBounds() {
    return _diffuseMaterial.getEmitterGenerationBounds();
}

void FluidSimulation::setDiffuseEmitterGenerationBounds(AABB bbox) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseEmitterGenerationBounds: " << 
                 bbox.position.x << " " << bbox.position.y << " " << bbox.position.z << " " <<
                 bbox.width << " " << bbox.height << " " << bbox.depth << std::endl);

    _diffuseMaterial.setEmitterGenerationBounds(bbox);
}

double FluidSimulation::getMinDiffuseParticleLifetime() {
    return _diffuseMaterial.getMinDiffuseParticleLifetime();
}   

void FluidSimulation::setMinDiffuseParticleLifetime(double lifetime) {
    if (lifetime < 0) {
        std::string msg = "Error: min lifetime must be greater than or equal to 0.\n";
        msg += "lifetime: " + _toString(lifetime) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinDiffuseParticleLifetime: " << lifetime << std::endl);

    _diffuseMaterial.setMinDiffuseParticleLifetime(lifetime);
}

double FluidSimulation::getMaxDiffuseParticleLifetime() {
    return _diffuseMaterial.getMaxDiffuseParticleLifetime();
}   

void FluidSimulation::setMaxDiffuseParticleLifetime(double lifetime) {
    if (lifetime < 0) {
        std::string msg = "Error: max lifetime must be greater than or equal to 0.\n";
        msg += "lifetime: " + _toString(lifetime) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxDiffuseParticleLifetime: " << lifetime << std::endl);

    _diffuseMaterial.setMaxDiffuseParticleLifetime(lifetime);
}

double FluidSimulation::getDiffuseParticleLifetimeVariance() {
    return _diffuseMaterial.getDiffuseParticleLifetimeVariance();
}

void FluidSimulation::setDiffuseParticleLifetimeVariance(double variance) {
    if (variance < 0) {
        std::string msg = "Error: lifetime variance must be greater than or equal to 0.\n";
        msg += "variance: " + _toString(variance) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseParticleLifetimeVariance: " << variance << std::endl);

    _diffuseMaterial.setDiffuseParticleLifetimeVariance(variance);
}

double FluidSimulation::getFoamParticleLifetimeModifier() {
    return _diffuseMaterial.getFoamParticleLifetimeModifier();
}

void FluidSimulation::setFoamParticleLifetimeModifier(double modifier) {
    if (modifier < 0) {
        std::string msg = "Error: foam lifetime modifier must be greater than or equal to 0.\n";
        msg += "modifier: " + _toString(modifier) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setFoamParticleLifetimeModifier: " << modifier << std::endl);

    _diffuseMaterial.setFoamParticleLifetimeModifier(modifier);
}

double FluidSimulation::getBubbleParticleLifetimeModifier() {
    return _diffuseMaterial.getBubbleParticleLifetimeModifier();
}

void FluidSimulation::setBubbleParticleLifetimeModifier(double modifier) {
    if (modifier < 0) {
        std::string msg = "Error: bubble lifetime modifier must be greater than or equal to 0.\n";
        msg += "modifier: " + _toString(modifier) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setBubbleParticleLifetimeModifier: " << modifier << std::endl);

    _diffuseMaterial.setBubbleParticleLifetimeModifier(modifier);
}

double FluidSimulation::getSprayParticleLifetimeModifier() {
    return _diffuseMaterial.getSprayParticleLifetimeModifier();
}

void FluidSimulation::setSprayParticleLifetimeModifier(double modifier) {
    if (modifier < 0) {
        std::string msg = "Error: spray lifetime modifier must be greater than or equal to 0.\n";
        msg += "modifier: " + _toString(modifier) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSprayParticleLifetimeModifier: " << modifier << std::endl);

    _diffuseMaterial.setSprayParticleLifetimeModifier(modifier);
}

double FluidSimulation::getDustParticleLifetimeModifier() {
    return _diffuseMaterial.getDustParticleLifetimeModifier();
}

void FluidSimulation::setDustParticleLifetimeModifier(double modifier) {
    if (modifier < 0) {
        std::string msg = "Error: dust lifetime modifier must be greater than or equal to 0.\n";
        msg += "modifier: " + _toString(modifier) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDustParticleLifetimeModifier: " << modifier << std::endl);

    _diffuseMaterial.setDustParticleLifetimeModifier(modifier);
}

double FluidSimulation::getDiffuseParticleWavecrestEmissionRate() {
    return _diffuseMaterial.getDiffuseParticleWavecrestEmissionRate();
}

void FluidSimulation::setDiffuseParticleWavecrestEmissionRate(double r) {
    if (r < 0) {
        std::string msg = "Error: wavecrest emission rate must be greater than or equal to 0.\n";
        msg += "rate: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseParticleWavecrestEmissionRate: " << r << std::endl);

    _diffuseMaterial.setDiffuseParticleWavecrestEmissionRate(r);
}

double FluidSimulation::getDiffuseParticleTurbulenceEmissionRate() {
    return _diffuseMaterial.getDiffuseParticleTurbulenceEmissionRate();
}

void FluidSimulation::setDiffuseParticleTurbulenceEmissionRate(double r) {
    if (r < 0) {
        std::string msg = "Error: turbulence emission rate must be greater than or equal to 0.\n";
        msg += "rate: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseParticleTurbulenceEmissionRate: " << r << std::endl);

    _diffuseMaterial.setDiffuseParticleTurbulenceEmissionRate(r);
}

double FluidSimulation::getDiffuseParticleDustEmissionRate() {
    return _diffuseMaterial.getDiffuseParticleDustEmissionRate();
}

void FluidSimulation::setDiffuseParticleDustEmissionRate(double r) {
    if (r < 0) {
        std::string msg = "Error: dust emission rate must be greater than or equal to 0.\n";
        msg += "rate: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseParticleDustEmissionRate: " << r << std::endl);

    _diffuseMaterial.setDiffuseParticleDustEmissionRate(r);
}

double FluidSimulation::getDiffuseFoamAdvectionStrength() {
    return _diffuseMaterial.getFoamAdvectionStrength();
}

void FluidSimulation::setDiffuseFoamAdvectionStrength(double s) {
    if (s < 0.0 || s > 1.0) {
        std::string msg = "Error: advection strength must be in range [0.0, 1.0].\n";
        msg += "strendth: " + _toString(s) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamAdvectionStrength: " << s << std::endl);

    _diffuseMaterial.setFoamAdvectionStrength(s);
}

double FluidSimulation::getDiffuseFoamLayerDepth() {
    return _diffuseMaterial.getFoamLayerDepth();
}

void FluidSimulation::setDiffuseFoamLayerDepth(double depth) {
    if (depth < 0.0 || depth > 1.0) {
        std::string msg = "Error: foam layer depth must be in range [0.0, 1.0].\n";
        msg += "depth: " + _toString(depth) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamLayerDepth: " << depth << std::endl);

    _diffuseMaterial.setFoamLayerDepth(depth);
}

double FluidSimulation::getDiffuseFoamLayerOffset() {
    return _diffuseMaterial.getFoamLayerOffset();
}

void FluidSimulation::setDiffuseFoamLayerOffset(double offset) {
    if (offset < -1.0 || offset > 1.0) {
        std::string msg = "Error: foam layer offset must be in range [-1.0, 1.0].\n";
        msg += "offset: " + _toString(offset) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamLayerOffset: " << offset << std::endl);

    _diffuseMaterial.setFoamLayerOffset(offset);

}

void FluidSimulation::enableDiffusePreserveFoam() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableDiffusePreserveFoam" << std::endl);

    _diffuseMaterial.enablePreserveFoam();
}

void FluidSimulation::disableDiffusePreserveFoam() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableDiffusePreserveFoam" << std::endl);

    _diffuseMaterial.disablePreserveFoam();
}

bool FluidSimulation::isDiffusePreserveFoamEnabled() {
    return _diffuseMaterial.isPreserveFoamEnabled();
}

double FluidSimulation::getDiffuseFoamPreservationRate() {
    return _diffuseMaterial.getFoamPreservationRate();
}

void FluidSimulation::setDiffuseFoamPreservationRate(double rate) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamPreservationRate: " << rate << std::endl);

    _diffuseMaterial.setFoamPreservationRate(rate);
}

double FluidSimulation::getMinDiffuseFoamDensity() {
    return _diffuseMaterial.getMinFoamDensity();
}

void FluidSimulation::setMinDiffuseFoamDensity(double d) {
    if (d < 0) {
        std::string msg = "Error: min density must be greater than or equal to 0.\n";
        msg += "density: " + _toString(d) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMinDiffuseFoamDensity: " << d << std::endl);

    _diffuseMaterial.setMinFoamDensity(d);
}

double FluidSimulation::getMaxDiffuseFoamDensity() {
    return _diffuseMaterial.getMaxFoamDensity();
}

void FluidSimulation::setMaxDiffuseFoamDensity(double d) {
    if (d < 0) {
        std::string msg = "Error: max density must be greater than or equal to 0.\n";
        msg += "density: " + _toString(d) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setMaxDiffuseFoamDensity: " << d << std::endl);

    _diffuseMaterial.setMaxFoamDensity(d);
}

double FluidSimulation::getDiffuseBubbleDragCoefficient() {
    return _diffuseMaterial.getBubbleDragCoefficient();
}

void FluidSimulation::setDiffuseBubbleDragCoefficient(double d) {
    if (d < 0.0 || d > 1.0) {
        std::string msg = "Error: drag coefficient must be in range [0.0, 1.0].\n";
        msg += "coefficient: " + _toString(d) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseBubbleDragCoefficient: " << d << std::endl);

    _diffuseMaterial.setBubbleDragCoefficient(d);
}

double FluidSimulation::getDiffuseDustDragCoefficient() {
    return _diffuseMaterial.getDustDragCoefficient();
}

void FluidSimulation::setDiffuseDustDragCoefficient(double d) {
    if (d < 0.0 || d > 1.0) {
        std::string msg = "Error: drag coefficient must be in range [0.0, 1.0].\n";
        msg += "coefficient: " + _toString(d) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseDustDragCoefficient: " << d << std::endl);

    _diffuseMaterial.setDustDragCoefficient(d);
}

double FluidSimulation::getDiffuseBubbleBouyancyCoefficient() {
    return _diffuseMaterial.getBubbleBouyancyCoefficient();
}

void FluidSimulation::setDiffuseBubbleBouyancyCoefficient(double b) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseBubbleBouyancyCoefficient: " << b << std::endl);

    _diffuseMaterial.setBubbleBouyancyCoefficient(b);
}

double FluidSimulation::getDiffuseDustBouyancyCoefficient() {
    return _diffuseMaterial.getDustBouyancyCoefficient();
}

void FluidSimulation::setDiffuseDustBouyancyCoefficient(double b) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseDustBouyancyCoefficient: " << b << std::endl);

    _diffuseMaterial.setDustBouyancyCoefficient(b);
}

double FluidSimulation::getDiffuseSprayDragCoefficient() {
    return _diffuseMaterial.getSprayDragCoefficient();
}

void FluidSimulation::setDiffuseSprayDragCoefficient(double d) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseSprayDragCoefficient: " << d << std::endl);

    _diffuseMaterial.setSprayDragCoefficient(d);
}

double FluidSimulation::getDiffuseSprayEmissionSpeed() {
    return _diffuseMaterial.getSprayEmissionSpeed();
}

void FluidSimulation::setDiffuseSprayEmissionSpeed(double d) {
    if (d < 0.0) {
        std::string msg = "Error: spray emission speed must be greater than or equal to 0.0.\n";
        msg += "speed: " + _toString(d) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseSprayEmissionSpeed: " << d << std::endl);

    _diffuseMaterial.setSprayEmissionSpeed(d);
}

LimitBehaviour FluidSimulation::getDiffuseFoamLimitBehaviour() {
    return _diffuseMaterial.getFoamLimitBehaviour();
}

void FluidSimulation::setDiffuseFoamLimitBehaviour(LimitBehaviour b) {
    std::string typestr;
    if (b == LimitBehaviour::collide) {
        typestr = "collide";
    } else if (b == LimitBehaviour::ballistic) {
        typestr = "ballistic";
    } else if (b == LimitBehaviour::kill) {
        typestr = "kill";
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamLimitBehavour: " << typestr << std::endl);

    _diffuseMaterial.setFoamLimitBehavour(b);
}

LimitBehaviour FluidSimulation::getDiffuseBubbleLimitBehaviour() {
    return _diffuseMaterial.getBubbleLimitBehaviour();
}

void FluidSimulation::setDiffuseBubbleLimitBehaviour(LimitBehaviour b) {
    std::string typestr;
    if (b == LimitBehaviour::collide) {
        typestr = "collide";
    } else if (b == LimitBehaviour::ballistic) {
        typestr = "ballistic";
    } else if (b == LimitBehaviour::kill) {
        typestr = "kill";
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseBubbleLimitBehavour: " << typestr << std::endl);

    _diffuseMaterial.setBubbleLimitBehavour(b);
}

LimitBehaviour FluidSimulation::getDiffuseSprayLimitBehaviour() {
    return _diffuseMaterial.getSprayLimitBehaviour();
}

void FluidSimulation::setDiffuseSprayLimitBehaviour(LimitBehaviour b) {
    std::string typestr;
    if (b == LimitBehaviour::collide) {
        typestr = "collide";
    } else if (b == LimitBehaviour::ballistic) {
        typestr = "ballistic";
    } else if (b == LimitBehaviour::kill) {
        typestr = "kill";
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseSprayLimitBehavour: " << typestr << std::endl);

    _diffuseMaterial.setSprayLimitBehavour(b);
}

LimitBehaviour FluidSimulation::getDiffuseDustLimitBehaviour() {
    return _diffuseMaterial.getDustLimitBehaviour();
}

void FluidSimulation::setDiffuseDustLimitBehaviour(LimitBehaviour b) {
    std::string typestr;
    if (b == LimitBehaviour::collide) {
        typestr = "collide";
    } else if (b == LimitBehaviour::ballistic) {
        typestr = "ballistic";
    } else if (b == LimitBehaviour::kill) {
        typestr = "kill";
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseDustLimitBehaviour: " << typestr << std::endl);

    _diffuseMaterial.setDustLimitBehavour(b);
}

std::vector<bool> FluidSimulation::getDiffuseFoamActiveBoundarySides() {
    return _diffuseMaterial.getFoamActiveBoundarySides();
}

void FluidSimulation::setDiffuseFoamActiveBoundarySides(std::vector<bool> active) {
    if (active.size() != 6) {
        std::string msg = "Error: foam active boundary vector must be of length 6.\n";
        msg += "length: " + _toString(active.size()) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseFoamActiveBoundarySides: " << 
                 active[0] << " " << active[1] << " " << active[2] << " " << 
                 active[3] << " " << active[4] << " " << active[5] << std::endl);

    _diffuseMaterial.setFoamActiveBoundarySides(active);
}

std::vector<bool> FluidSimulation::getDiffuseBubbleActiveBoundarySides() {
    return _diffuseMaterial.getBubbleActiveBoundarySides();
}

void FluidSimulation::setDiffuseBubbleActiveBoundarySides(std::vector<bool> active) {
    if (active.size() != 6) {
        std::string msg = "Error: bubble active boundary vector must be of length 6.\n";
        msg += "length: " + _toString(active.size()) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseBubbleActiveBoundarySides: " << 
                 active[0] << " " << active[1] << " " << active[2] << " " << 
                 active[3] << " " << active[4] << " " << active[5] << std::endl);

    _diffuseMaterial.setBubbleActiveBoundarySides(active);
}

std::vector<bool> FluidSimulation::getDiffuseSprayActiveBoundarySides() {
    return _diffuseMaterial.getSprayActiveBoundarySides();
}

void FluidSimulation::setDiffuseSprayActiveBoundarySides(std::vector<bool> active) {
    if (active.size() != 6) {
        std::string msg = "Error: spray active boundary vector must be of length 6.\n";
        msg += "length: " + _toString(active.size()) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseSprayActiveBoundarySides: " << 
                 active[0] << " " << active[1] << " " << active[2] << " " << 
                 active[3] << " " << active[4] << " " << active[5] << std::endl);

    _diffuseMaterial.setSprayActiveBoundarySides(active);
}

std::vector<bool> FluidSimulation::getDiffuseDustActiveBoundarySides() {
    return _diffuseMaterial.getDustActiveBoundarySides();
}

void FluidSimulation::setDiffuseDustActiveBoundarySides(std::vector<bool> active) {
    if (active.size() != 6) {
        std::string msg = "Error: dust active boundary vector must be of length 6.\n";
        msg += "length: " + _toString(active.size()) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseDustActiveBoundarySides: " << 
                 active[0] << " " << active[1] << " " << active[2] << " " << 
                 active[3] << " " << active[4] << " " << active[5] << std::endl);

    _diffuseMaterial.setDustActiveBoundarySides(active);
}

double FluidSimulation::getDiffuseObstacleInfluenceBaseLevel() {
    return _diffuseObstacleInfluenceBaseLevel;
}

void FluidSimulation::setDiffuseObstacleInfluenceBaseLevel(double level) {
    if (level < 0.0) {
        std::string msg = "Error: base level must be greater than or equal to 0.0.\n";
        msg += "base level: " + _toString(level) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseObstacleInfluenceBaseLevel: " << level << std::endl);

    _diffuseObstacleInfluenceBaseLevel = level;
}

double FluidSimulation::getDiffuseObstacleInfluenceDecayRate() {
    return _diffuseObstacleInfluenceDecayRate;
}

void FluidSimulation::setDiffuseObstacleInfluenceDecayRate(double decay) {
    if (decay < 0.0) {
        std::string msg = "Error: decay rate must be greater than or equal to 0.0.\n";
        msg += "decay rate: " + _toString(decay) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setDiffuseObstacleInfluenceDecayRate: " << decay << std::endl);

    _diffuseObstacleInfluenceDecayRate = decay;
}

void FluidSimulation::enableOpenCLParticleAdvection() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableOpenCLParticleAdvection" << std::endl);

    _particleAdvector.enableOpenCL();
}

void FluidSimulation::disableOpenCLParticleAdvection() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableOpenCLParticleAdvection" << std::endl);

    _particleAdvector.disableOpenCL();
}

bool FluidSimulation::isOpenCLParticleAdvectionEnabled() {
    return _particleAdvector.isOpenCLEnabled();
}

void FluidSimulation::enableOpenCLScalarField() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableOpenCLScalarField" << std::endl);

    _mesherScalarFieldAccelerator.enableOpenCL();
}

void FluidSimulation::disableOpenCLScalarField() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableOpenCLScalarField" << std::endl);

    _mesherScalarFieldAccelerator.disableOpenCL();
}

bool FluidSimulation::isOpenCLScalarFieldEnabled() {
    return _mesherScalarFieldAccelerator.isOpenCLEnabled();
}

int FluidSimulation::getParticleAdvectionKernelWorkLoadSize() {
    return _particleAdvector.getKernelWorkLoadSize();
}

void FluidSimulation::setParticleAdvectionKernelWorkLoadSize(int n) {
    if (n < 1) {
        std::string msg = "Error: work load size must be greater than or equal to 1.\n";
        msg += "size: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setParticleAdvectionKernelWorkLoadSize: " << n << std::endl);

    _particleAdvector.setKernelWorkLoadSize(n);
}

int FluidSimulation::getScalarFieldKernelWorkLoadSize() {
    return _mesherScalarFieldAccelerator.getKernelWorkLoadSize();
}

void FluidSimulation::setScalarFieldKernelWorkLoadSize(int n) {
    if (n < 1) {
        std::string msg = "Error: work load size must be greater than or equal to 1.\n";
        msg += "size: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setScalarFieldKernelWorkLoadSize: " << n << std::endl);

    _mesherScalarFieldAccelerator.setKernelWorkLoadSize(n);
}

int FluidSimulation::getMaxThreadCount() {
    return ThreadUtils::getMaxThreadCount();
}

void FluidSimulation::setMaxThreadCount(int n) {
    if (n < 1) {
        std::string msg = "Error: thread count must be greater than or equal to 1.\n";
        msg += "thread count: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setMaxThreadCount: " << n << std::endl);

    ThreadUtils::setMaxThreadCount(n);
}

void FluidSimulation::addBodyForce(double fx, double fy, double fz) { 
    addBodyForce(vmath::vec3(fx, fy, fz)); 
}

void FluidSimulation::addBodyForce(vmath::vec3 f) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " addBodyForce: " <<
                 f.x << " " << f.y << " " << f.z << std::endl);

    _constantBodyForces.push_back(f);
}

vmath::vec3 FluidSimulation::getConstantBodyForce() {
    return _getConstantBodyForce();
}

void FluidSimulation::resetBodyForce() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " resetBodyForce" << std::endl);

    _constantBodyForces.clear();
}

void FluidSimulation::enableForceFields() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableForceFields" << std::endl);

    _isForceFieldsEnabled = true;
}

void FluidSimulation::disableForceFields() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableForceFields" << std::endl);

    _isForceFieldsEnabled = false;
}

bool FluidSimulation::isForceFieldsEnabled() {
    return _isForceFieldsEnabled;
}

int FluidSimulation::getForceFieldReductionLevel() {
    return _forceFieldReductionLevel;
}

void FluidSimulation::setForceFieldReductionLevel(int level) {
    if (level < 1) {
        std::string msg = "Error: reduction level must be greater than or equal to 1.\n";
        msg += "reduction level: " + _toString(level) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setForceFieldReductionLevel: " << level << std::endl);

    _forceFieldReductionLevel = level;
}

ForceFieldGrid* FluidSimulation::getForceFieldGrid() {
    if (!_isForceFieldsEnabled) {
        std::string msg = "Error: force fields must be enabled before using this method.\n";
        msg += "is force fields enabled: " + _toString(_isForceFieldsEnabled) + "\n";
        throw std::domain_error(msg);
    }

    return &_forceFieldGrid;
}

double FluidSimulation::getViscosity() {
    return _constantViscosityValue;
}

void FluidSimulation::setViscosity(double v) {
    if (v < 0.0) {
        std::string msg = "Error: viscosity must be greater than or equal to 0.\n";
        msg += "viscosity: " + _toString(v) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setViscosity: " << v << std::endl);

    if (!_isViscosityEnabled) {
        _viscosity = Array3d<float>(_isize + 1, _jsize + 1, _ksize + 1);
        _isViscosityEnabled = true;
    }
    _viscosity.fill(v);
    _constantViscosityValue = v;
}

double FluidSimulation::getViscositySolverErrorTolerance() {
    return _viscositySolverErrorTolerance;
}

void FluidSimulation::setViscositySolverErrorTolerance(double tol) {
    if (tol < 0.0) {
        std::string msg = "Error: viscosity solver error tolerance must be greater than or equal to 0.\n";
        msg += "error tolerance: " + _toString(tol) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setViscositySolverErrorTolerance: " << tol << std::endl);

    _viscositySolverErrorTolerance = tol;
}

double FluidSimulation::getSurfaceTension() {
    return _surfaceTensionConstant;
}

void FluidSimulation::setSurfaceTension(double k) {
    if (k < 0.0) {
        std::string msg = "Error: surface tension must be greater than or equal to 0.\n";
        msg += "surface tension: " + _toString(k) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSurfaceTension: " << k << std::endl);

    double eps = 1e-6;
    _isSurfaceTensionEnabled = k > eps;
    _surfaceTensionConstant = k;
}

void FluidSimulation::enableSheetSeeding() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSheetSeeding" << std::endl);

    _isSheetSeedingEnabled = true;
}

void FluidSimulation::disableSheetSeeding() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSheetSeeding" << std::endl);

    _isSheetSeedingEnabled = false;
}

bool FluidSimulation::isSheetSeedingEnabled() {
    return _isSheetSeedingEnabled;
}

double FluidSimulation::getSheetFillThreshold() {
    return _sheetFillThreshold;
}

void FluidSimulation::setSheetFillThreshold(double f) {
    if (f < -1.0 || f > 0.0) {
        std::string msg = "Error: sheet fill threshold must be in range [-1.0, 0.0].\n";
        msg += "threshold: " + _toString(f) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSheetFillThreshold: " << f << std::endl);

    _sheetFillThreshold = f;
}

double FluidSimulation::getSheetFillRate() {
    return _sheetFillRate;
}

void FluidSimulation::setSheetFillRate(double r) {
    if (r < 0.0 || r > 1.0) {
        std::string msg = "Error: sheet fill rate must be in range [0.0, 1.0].\n";
        msg += "threshold: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setSheetFillRate: " << r << std::endl);

    _sheetFillRate = r;
}

double FluidSimulation::getBoundaryFriction() {
    return _domainBoundaryFriction;
}

void FluidSimulation::setBoundaryFriction(double f) {
    if (f < 0.0 || f > 1.0) {
        std::string msg = "Error: boundary friction must be in range [0.0, 1.0].\n";
        msg += "friction: " + _toString(f) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setBoundaryFriction: " << f << std::endl);

    _domainMeshObject.setFriction(f);
    _domainBoundaryFriction = f;
}

int FluidSimulation::getCFLConditionNumber() {
    return _CFLConditionNumber;
}

void FluidSimulation::setCFLConditionNumber(int n) {
    if (n < 1) {
        std::string msg = "Error: CFL must be greater than or equal to 1.\n";
        msg += "CFL: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setCFLConditionNumber: " << n << std::endl);

    _CFLConditionNumber = n;
}

double FluidSimulation::getSurfaceTensionConditionNumber() {
    return _surfaceTensionConditionNumber;
}

void FluidSimulation::setSurfaceTensionConditionNumber(double n) {
    if (n <= 0.0) {
        std::string msg = "Error: condition number must be greater than 0.0.\n";
        msg += "number: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setSurfaceTensionConditionNumber: " << n << std::endl);

    _surfaceTensionConditionNumber = n;
}

bool FluidSimulation::isSmoothSurfaceTensionKernelEnabled() {
    return _isSmoothSurfaceTensionKernelEnabled;
}

void FluidSimulation::enableSmoothSurfaceTensionKernel() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableSmoothSurfaceTensionKernel" << std::endl);

    _isSmoothSurfaceTensionKernelEnabled = true;
}

void FluidSimulation::disableSmoothSurfaceTensionKernel() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableSmoothSurfaceTensionKernel" << std::endl);

    _isSmoothSurfaceTensionKernelEnabled = false;
}

int FluidSimulation::getMinTimeStepsPerFrame() {
    return _minFrameTimeSteps;
}

void FluidSimulation::setMinTimeStepsPerFrame(int n) {
    if (n < 1) {
        std::string msg = "Error: min step count must be greater than or equal to 1.\n";
        msg += "Step count: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setMinTimeStepsPerFrame: " << n << std::endl);

    _minFrameTimeSteps = n;
}

int FluidSimulation::getMaxTimeStepsPerFrame() {
    return _maxFrameTimeSteps;
}

void FluidSimulation::setMaxTimeStepsPerFrame(int n) {
    if (n < 1) {
        std::string msg = "Error: max step count must be greater than or equal to 1.\n";
        msg += "Step count: " + _toString(n) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setMaxTimeStepsPerFrame: " << n << std::endl);

    _maxFrameTimeSteps = n;
}

void FluidSimulation::enableAdaptiveObstacleTimeStepping() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableAdaptiveObstacleTimeStepping" << std::endl);

    _isAdaptiveObstacleTimeSteppingEnabled = true;
}

void FluidSimulation::disableAdaptiveObstacleTimeStepping() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableAdaptiveObstacleTimeStepping" << std::endl);

    _isAdaptiveObstacleTimeSteppingEnabled = false;
}

bool FluidSimulation::isAdaptiveObstacleTimeSteppingEnabled() {
    return _isAdaptiveObstacleTimeSteppingEnabled;
}

void FluidSimulation::enableAdaptiveForceFieldTimeStepping() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableAdaptiveForceFieldTimeStepping" << std::endl);

    _isAdaptiveForceFieldTimeSteppingEnabled = true;
}

void FluidSimulation::disableAdaptiveForceFieldTimeStepping() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableAdaptiveForceFieldTimeStepping" << std::endl);

    _isAdaptiveForceFieldTimeSteppingEnabled = false;
}

bool FluidSimulation::isAdaptiveForceFieldTimeSteppingEnabled() {
    return _isAdaptiveForceFieldTimeSteppingEnabled;
}

void FluidSimulation::enableExtremeVelocityRemoval() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableExtremeVelocityRemoval" << std::endl);

    _isExtremeVelocityRemovalEnabled = true;
}

void FluidSimulation::disableExtremeVelocityRemoval() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableExtremeVelocityRemoval" << std::endl);

    _isExtremeVelocityRemovalEnabled = false;
}

bool FluidSimulation::isExtremeVelocityRemovalEnabled() {
    return _isExtremeVelocityRemovalEnabled;
}

void FluidSimulation::setVelocityTransferMethodFLIP() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setVelocityTransferMethodFLIP" << std::endl);

    _velocityTransferMethod = VelocityTransferMethod::FLIP;
}

void FluidSimulation::setVelocityTransferMethodAPIC() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " setVelocityTransferMethodAPIC" << std::endl);

    _velocityTransferMethod = VelocityTransferMethod::APIC;
}

bool FluidSimulation::isVelocityTransferMethodFLIP() {
    return _velocityTransferMethod == VelocityTransferMethod::FLIP;
}

bool FluidSimulation::isVelocityTransferMethodAPIC() {
    return _velocityTransferMethod == VelocityTransferMethod::APIC;
}

double FluidSimulation::getPICFLIPRatio() {
    return _ratioPICFLIP;
}

void FluidSimulation::setPICFLIPRatio(double r) {
    if (r < 0.0 || r > 1.0) {
        std::string msg = "Error: PICFLIP ratio must be in range [0.0, 1.0].\n";
        msg += "ratio: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setPICFLIPRatio: " << r << std::endl);

    _ratioPICFLIP = r;
}

double FluidSimulation::getPICAPICRatio() {
    return _ratioPICAPIC;
}

void FluidSimulation::setPICAPICRatio(double r) {
    if (r < 0.0 || r > 1.0) {
        std::string msg = "Error: PICAPIC ratio must be in range [0.0, 1.0].\n";
        msg += "ratio: " + _toString(r) + "\n";
        throw std::domain_error(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setPICAPICRatio: " << r << std::endl);

    _ratioPICAPIC = r;
}

void FluidSimulation::setPreferredGPUDevice(std::string deviceName) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << 
                 " setPreferredGPUDevice: " << deviceName << std::endl);

    OpenCLUtils::setPreferredGPUDevice(deviceName);
}

std::string FluidSimulation::getPreferredGPUDevice() {
    return OpenCLUtils::getPreferredGPUDevice();
}

void FluidSimulation::enableStaticSolidLevelSetPrecomputation() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableStaticSolidLevelSetPrecomputation" << std::endl);

    _isStaticSolidLevelSetPrecomputed = true;
}

void FluidSimulation::disableStaticSolidLevelSetPrecomputation() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableStaticSolidLevelSetPrecomputation" << std::endl);

    _isStaticSolidLevelSetPrecomputed = false;
}

bool FluidSimulation::isStaticSolidLevelSetPrecomputationEnabled() {
    return _isStaticSolidLevelSetPrecomputed;
}

void FluidSimulation::enableTemporaryMeshLevelSet() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " enableTemporaryMeshLevelSet" << std::endl);

    _isTempSolidLevelSetEnabled = true;
}

void FluidSimulation::disableTemporaryMeshLevelSet() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " disableTemporaryMeshLevelSet" << std::endl);

    _isTempSolidLevelSetEnabled = false;
}

bool FluidSimulation::isTemporaryMeshLevelSetEnabled() {
    return _isTempSolidLevelSetEnabled;
}

void FluidSimulation::addMeshFluidSource(MeshFluidSource *source) {
    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (source->getID() == _meshFluidSources[i]->getID()) {
            std::string msg = "Error: Mesh fluid source has already been added.\n";
            throw std::runtime_error(msg);
        }
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " addMeshFluidSource: " << source << std::endl);

    _meshFluidSources.push_back(source);
}

void FluidSimulation::removeMeshFluidSource(MeshFluidSource *source) {
    bool isFound = false;
    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (source->getID() == _meshFluidSources[i]->getID()) {
            _meshFluidSources.erase(_meshFluidSources.begin() + i);
            isFound = true;
            break;
        }
    }

    if (!isFound) {
        std::string msg = "Error: could not find mesh fluid source to remove.\n";
        msg += "mesh fluid source: " + _toString(source) + "\n";
        throw std::invalid_argument(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " removeMeshFluidSource: " << source << std::endl);
}

void FluidSimulation::removeMeshFluidSources() {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " removeMeshFluidSources" << std::endl);

    _meshFluidSources.clear();
}

void FluidSimulation::addMeshObstacle(MeshObject *obstacle) {
    for (unsigned int i = 0; i < _obstacles.size(); i++) {
        if (obstacle == _obstacles[i]) {
            std::string msg = "Error: mesh obstacle has already been added.\n";
            throw std::runtime_error(msg);
        }
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " addMeshObstacle: " << obstacle << std::endl);

    _obstacles.push_back(obstacle);

    _isSolidLevelSetUpToDate = false;
}

void FluidSimulation::removeMeshObstacle(MeshObject *obstacle) {
    bool isFound = false;
    for (unsigned int i = 0; i < _obstacles.size(); i++) {
        if (obstacle == _obstacles[i]) {
            _obstacles.erase(_obstacles.begin() + i);
            isFound = true;
            break;
        }
    }

    if (!isFound) {
        std::string msg = "Error: could not find mesh obstacle to remove.\n";
        msg += "mesh obstacle: " + _toString(obstacle) + "\n";
        throw std::invalid_argument(msg);
    }

    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " removeMeshObstacle: " << obstacle << std::endl);

    _isSolidLevelSetUpToDate = false;
}

void FluidSimulation::removeMeshObstacles() {
    _obstacles.clear();

    _isSolidLevelSetUpToDate = false;
}

void FluidSimulation::addMeshFluid(MeshObject fluid) {
    addMeshFluid(fluid, vmath::vec3(0.0, 0.0, 0.0));
}

void FluidSimulation::addMeshFluid(MeshObject fluid, vmath::vec3 velocity) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " addMeshFluid: " << velocity << std::endl);

    _addedFluidMeshObjectQueue.push_back(FluidMeshObject(fluid, velocity));
}

unsigned int FluidSimulation::getNumMarkerParticles() {
    return _markerParticles.size();
}

std::vector<MarkerParticle> FluidSimulation::getMarkerParticles() {
    return getMarkerParticles(0, _markerParticles.size());
}

std::vector<MarkerParticle> FluidSimulation::getMarkerParticles(int startidx, int endidx) {
    int size = _markerParticles.size();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<MarkerParticle> particles;
    particles.reserve(endidx - startidx);

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    for (int i = startidx; i < endidx; i++) {
        MarkerParticle mp(positions->at(i), velocities->at(i));
        particles.push_back(mp);
    }

    return particles;
}

std::vector<vmath::vec3> FluidSimulation::getMarkerParticlePositions() {
    return getMarkerParticlePositions(0, _markerParticles.size());
}

std::vector<vmath::vec3> FluidSimulation::getMarkerParticlePositions(int startidx, int endidx) {
    int size = _markerParticles.size();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<vmath::vec3> particles;
    particles.reserve(endidx - startidx);

    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    for (int i = startidx; i < endidx; i++) {
        particles.push_back(positions->at(i));
    }

    return particles;
}

std::vector<vmath::vec3> FluidSimulation::getMarkerParticleVelocities() {
    return getMarkerParticleVelocities(0, _markerParticles.size());
}

std::vector<vmath::vec3> FluidSimulation::getMarkerParticleVelocities(int startidx, int endidx) {
    int size = _markerParticles.size();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<vmath::vec3> velocities;
    velocities.reserve(endidx - startidx);

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("VELOCITY", values);

    for (int i = startidx; i < endidx; i++) {
        velocities.push_back(values->at(i));
    }

    return velocities;
}

unsigned int FluidSimulation::getNumDiffuseParticles() {
    return _diffuseMaterial.getNumDiffuseParticles();
}

std::vector<vmath::vec3> FluidSimulation::getDiffuseParticlePositions() {
    int size = getNumDiffuseParticles();
    return getDiffuseParticlePositions(0, size);
}

std::vector<vmath::vec3> FluidSimulation::getDiffuseParticlePositions(int startidx, int endidx) {
    int size = getNumDiffuseParticles();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<vmath::vec3> *positions;
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("POSITION", positions);

    return *positions;
}

std::vector<vmath::vec3> FluidSimulation::getDiffuseParticleVelocities() {
    int size = getNumDiffuseParticles();
    return getDiffuseParticleVelocities(0, size);
}

std::vector<vmath::vec3> FluidSimulation::getDiffuseParticleVelocities(int startidx, int endidx) {
    int size = getNumDiffuseParticles();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<vmath::vec3> *velocities;
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("VELOCITY", velocities);
    
    return *velocities;
}

std::vector<float> FluidSimulation::getDiffuseParticleLifetimes() {
    int size = getNumDiffuseParticles();
    return getDiffuseParticleLifetimes(0, size);
}

std::vector<float> FluidSimulation::getDiffuseParticleLifetimes(int startidx, int endidx) {
    int size = getNumDiffuseParticles();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<float> *lifetimes;
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("LIFETIME", lifetimes);
    
    return *lifetimes;
}

std::vector<char> FluidSimulation::getDiffuseParticleTypes() {
    int size = getNumDiffuseParticles();
    return getDiffuseParticleTypes(0, size);
}

std::vector<char> FluidSimulation::getDiffuseParticleTypes(int startidx, int endidx) {
    int size = getNumDiffuseParticles();
    if (!(startidx >= 0 && startidx <= size) || !(endidx >= 0 && endidx <= size)) {
        std::string msg = "Error: invalid index range.\n";
        msg += "start index: " + _toString(startidx) + " end index: " + _toString(endidx) + "\n";
        throw std::out_of_range(msg);
    }

    std::vector<char> *types;
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("TYPE", types);
    
    return *types;
}

MACVelocityField* FluidSimulation::getVelocityField() { 
    return &_MACVelocity; 
}

std::vector<char>* FluidSimulation::getSurfaceData() {
    return &_outputData.surfaceData;
}

std::vector<char>* FluidSimulation::getSurfaceBlurData() {
    return &_outputData.surfaceBlurData;
}

std::vector<char>* FluidSimulation::getSurfaceVelocityAttributeData() {
    return &_outputData.surfaceVelocityAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceVorticityAttributeData() {
    return &_outputData.surfaceVorticityAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceSpeedAttributeData() {
    return &_outputData.surfaceSpeedAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceAgeAttributeData() {
    return &_outputData.surfaceAgeAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceColorAttributeData() {
    return &_outputData.surfaceColorAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceSourceIDAttributeData() {
    return &_outputData.surfaceSourceIDAttributeData;
}

std::vector<char>* FluidSimulation::getSurfaceViscosityAttributeData() {
    return &_outputData.surfaceViscosityAttributeData;
}

std::vector<char>* FluidSimulation::getSurfacePreviewData() {
    return &_outputData.surfacePreviewData;
}

std::vector<char>* FluidSimulation::getDiffuseData() {
    return &_outputData.diffuseData;
}

std::vector<char>* FluidSimulation::getDiffuseFoamData() {
    return &_outputData.diffuseFoamData;
}

std::vector<char>* FluidSimulation::getDiffuseBubbleData() {
    return &_outputData.diffuseBubbleData;
}

std::vector<char>* FluidSimulation::getDiffuseSprayData() {
    return &_outputData.diffuseSprayData;
}

std::vector<char>* FluidSimulation::getDiffuseDustData() {
    return &_outputData.diffuseDustData;
}

std::vector<char>* FluidSimulation::getDiffuseFoamBlurData() {
    return &_outputData.diffuseFoamBlurData;
}

std::vector<char>* FluidSimulation::getDiffuseBubbleBlurData() {
    return &_outputData.diffuseBubbleBlurData;
}

std::vector<char>* FluidSimulation::getDiffuseSprayBlurData() {
    return &_outputData.diffuseSprayBlurData;
}

std::vector<char>* FluidSimulation::getDiffuseDustBlurData() {
    return &_outputData.diffuseDustBlurData;
}

std::vector<char>* FluidSimulation::getWhitewaterFoamVelocityAttributeData() {
    return &_outputData.whitewaterFoamVelocityAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterBubbleVelocityAttributeData() {
    return &_outputData.whitewaterBubbleVelocityAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterSprayVelocityAttributeData() {
    return &_outputData.whitewaterSprayVelocityAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterDustVelocityAttributeData() {
    return &_outputData.whitewaterDustVelocityAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterFoamIDAttributeData() {
    return &_outputData.whitewaterFoamIDAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterBubbleIDAttributeData() {
    return &_outputData.whitewaterBubbleIDAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterSprayIDAttributeData() {
    return &_outputData.whitewaterSprayIDAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterDustIDAttributeData() {
    return &_outputData.whitewaterDustIDAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterFoamLifetimeAttributeData() {
    return &_outputData.whitewaterFoamLifetimeAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterBubbleLifetimeAttributeData() {
    return &_outputData.whitewaterBubbleLifetimeAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterSprayLifetimeAttributeData() {
    return &_outputData.whitewaterSprayLifetimeAttributeData;
}

std::vector<char>* FluidSimulation::getWhitewaterDustLifetimeAttributeData() {
    return &_outputData.whitewaterDustLifetimeAttributeData;
}

std::vector<char>* FluidSimulation::getFluidParticleData() {
    return &_outputData.fluidParticleData;
}

std::vector<char>* FluidSimulation::getInternalObstacleMeshData() {
    return &_outputData.internalObstacleMeshData;
}

std::vector<char>* FluidSimulation::getForceFieldDebugData() {
    return &_outputData.forceFieldDebugData;
}

std::vector<char>* FluidSimulation::getLogFileData() {
    return &_outputData.logfileData;
}

FluidSimulationFrameStats FluidSimulation::getFrameStatsData() {
    return _outputData.frameData;
}

void FluidSimulation::getMarkerParticlePositionDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("POSITION", values);

    vmath::vec3 *positions = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        positions[i - start_idx] = values->at(i) * _domainScale + _domainOffset;
    }
}

void FluidSimulation::getMarkerParticleVelocityDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("VELOCITY", values);

    vmath::vec3 *velocities = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        velocities[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleAffineXDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("AFFINEX", values);

    vmath::vec3 *dataValues = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleAffineYDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("AFFINEY", values);

    vmath::vec3 *dataValues = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleAffineZDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("AFFINEZ", values);

    vmath::vec3 *dataValues = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleAgeDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<float> *values;
    _markerParticles.getAttributeValues("AGE", values);

    float *dataValues = (float*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleColorDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("COLOR", values);

    vmath::vec3 *dataValues = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleSourceIDDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<int> *values;
    _markerParticles.getAttributeValues("SOURCEID", values);

    int *dataValues = (int*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getMarkerParticleViscosityDataRange(int start_idx, int end_idx, char *data) {
    if (start_idx < 0 || end_idx > (int)_markerParticles.size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<float> *values;
    _markerParticles.getAttributeValues("VISCOSITY", values);

    float *dataValues = (float*)data;
    for (int i = start_idx; i < end_idx; i++) {
        dataValues[i - start_idx] = values->at(i);
    }
}

void FluidSimulation::getDiffuseParticlePositionDataRange(int start_idx, int end_idx, char *data) {
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    if (start_idx < 0 || end_idx > (int)dps->size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *particlePositions;
    dps->getAttributeValues("POSITION", particlePositions);

    vmath::vec3 *positions = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        positions[i - start_idx] = particlePositions->at(i) * _domainScale + _domainOffset;
    }
}

void FluidSimulation::getDiffuseParticleVelocityDataRange(int start_idx, int end_idx, char *data) {
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    if (start_idx < 0 || end_idx > (int)dps->size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<vmath::vec3> *particleVelocities;
    dps->getAttributeValues("VELOCITY", particleVelocities);

    vmath::vec3 *velocities = (vmath::vec3*)data;
    for (int i = start_idx; i < end_idx; i++) {
        velocities[i - start_idx] = particleVelocities->at(i);
    }
}

void FluidSimulation::getDiffuseParticleLifetimeDataRange(int start_idx, int end_idx, char *data) {
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    if (start_idx < 0 || end_idx > (int)dps->size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<float> *particleLifetimes;
    dps->getAttributeValues("LIFETIME", particleLifetimes);

    float *lifetimes = (float*)data;
    for (int i = start_idx; i < end_idx; i++) {
        lifetimes[i - start_idx] = particleLifetimes->at(i);
    }
}

void FluidSimulation::getDiffuseParticleTypeDataRange(int start_idx, int end_idx, char *data) {
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    if (start_idx < 0 || end_idx > (int)dps->size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<char> *particleTypes;
    dps->getAttributeValues("TYPE", particleTypes);

    for (int i = start_idx; i < end_idx; i++) {
        data[i - start_idx] = particleTypes->at(i);
    }
}

void FluidSimulation::getDiffuseParticleIdDataRange(int start_idx, int end_idx, char *data) {
    ParticleSystem *dps = _diffuseMaterial.getDiffuseParticles();
    if (start_idx < 0 || end_idx > (int)dps->size() || start_idx > end_idx) {
        std::string msg = "Error: invalid range.\n";
        msg += "range: [" + _toString(start_idx) + ", " + _toString(end_idx) + "]\n";
        throw std::domain_error(msg);
    }

    std::vector<unsigned char> *particleIds;
    dps->getAttributeValues("ID", particleIds);

    for (int i = start_idx; i < end_idx; i++) {
        data[i - start_idx] = (unsigned char)(particleIds->at(i));
    }
}

void FluidSimulation::getMarkerParticlePositionData(char *data) {
    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("POSITION", values);

    vmath::vec3 *positions = (vmath::vec3*)data;
    for (size_t i = 0; i < _markerParticles.size(); i++) {
        positions[i] = values->at(i) * _domainScale + _domainOffset;
    }
}

void FluidSimulation::getMarkerParticleVelocityData(char *data) {
    std::vector<vmath::vec3> *values;
    _markerParticles.getAttributeValues("POSITION", values);

    vmath::vec3 *velocities = (vmath::vec3*)data;
    for (size_t i = 0; i < _markerParticles.size(); i++) {
        velocities[i] = values->at(i);
    }
}

void FluidSimulation::getDiffuseParticlePositionData(char *data) {
    std::vector<vmath::vec3> *particlePositions;
    ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("POSITION", particlePositions);

    vmath::vec3 *positions = (vmath::vec3*)data;
    for (size_t i = 0; i < dps->size(); i++) {
        positions[i] = particlePositions->at(i) * _domainScale + _domainOffset;
    }
}

void FluidSimulation::getDiffuseParticleVelocityData(char *data) {
    std::vector<vmath::vec3> *particleVelocities;
    ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("VELOCITY", particleVelocities);

    vmath::vec3 *velocities = (vmath::vec3*)data;
    for (size_t i = 0; i < dps->size(); i++) {
        velocities[i] = particleVelocities->at(i);
    }
}

void FluidSimulation::getDiffuseParticleLifetimeData(char *data) {
    std::vector<float> *particleLifetimes;
    ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("LIFETIME", particleLifetimes);

    float *lifetimes = (float*)data;
    for (size_t i = 0; i < dps->size(); i++) {
        lifetimes[i] = particleLifetimes->at(i);
    }
}

void FluidSimulation::getDiffuseParticleTypeData(char *data) {
    std::vector<char> *particleTypes;
    ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("TYPE", particleTypes);

    for (size_t i = 0; i < dps->size(); i++) {
        data[i] = particleTypes->at(i);
    }
}

void FluidSimulation::getDiffuseParticleIdData(char *data) {
    std::vector<unsigned char> *particleIds;
    ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
    dps->getAttributeValues("ID", particleIds);

    for (size_t i = 0; i < dps->size(); i++) {
        data[i] = (unsigned char)(particleIds->at(i));
    }
}

unsigned int FluidSimulation::getMarkerParticlePositionDataSize() {
    return (unsigned int)(getNumMarkerParticles() * sizeof(vmath::vec3));
}

unsigned int FluidSimulation::getMarkerParticleVelocityDataSize() {
    return getMarkerParticlePositionDataSize();
}

unsigned int FluidSimulation::getDiffuseParticlePositionDataSize() {
    return (unsigned int)(getNumDiffuseParticles() * sizeof(vmath::vec3));
}

unsigned int FluidSimulation::getDiffuseParticleVelocityDataSize() {
    return getDiffuseParticlePositionDataSize();
}

unsigned int FluidSimulation::getDiffuseParticleLifetimeDataSize() {
    return (unsigned int)(getNumDiffuseParticles() * sizeof(float));
}

unsigned int FluidSimulation::getDiffuseParticleTypeDataSize() {
    return (unsigned int)(getNumDiffuseParticles() * sizeof(char));
}

unsigned int FluidSimulation::getDiffuseParticleIdDataSize() {
    return (unsigned int)(getNumDiffuseParticles() * sizeof(char));
}

void FluidSimulation::loadMarkerParticleData(FluidSimulationMarkerParticleData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    vmath::vec3 *positions = (vmath::vec3*)(data.positions);
    vmath::vec3 *velocities = (vmath::vec3*)(data.velocities);

    MarkerParticleLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticle(positions[i], velocities[i]));
    }

    _markerParticleLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadMarkerParticleAffineData(FluidSimulationMarkerParticleAffineData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleAffineData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    vmath::vec3 *affineX = (vmath::vec3*)(data.affineX);
    vmath::vec3 *affineY = (vmath::vec3*)(data.affineY);
    vmath::vec3 *affineZ = (vmath::vec3*)(data.affineZ);

    MarkerParticleAffineLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticleAffine(affineX[i], affineY[i], affineZ[i]));
    }

    _markerParticleAffineLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadMarkerParticleAgeData(FluidSimulationMarkerParticleAgeData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleAgeData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    float *age = (float*)(data.age);

    MarkerParticleAgeLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticleAge(age[i]));
    }

    _markerParticleAgeLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadMarkerParticleColorData(FluidSimulationMarkerParticleColorData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleColorData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    vmath::vec3 *colors = (vmath::vec3*)(data.color);

    MarkerParticleColorLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticleColor(colors[i]));
    }

    _markerParticleColorLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadMarkerParticleSourceIDData(FluidSimulationMarkerParticleSourceIDData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleSourceIDData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    int *sourceid = (int*)(data.sourceid);

    MarkerParticleSourceIDLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticleSourceID(sourceid[i]));
    }

    _markerParticleSourceIDLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadMarkerParticleViscosityData(FluidSimulationMarkerParticleViscosityData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadMarkerParticleViscosityData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    float *viscosity = (float*)(data.viscosity);

    MarkerParticleViscosityLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        loadData.particles.push_back(MarkerParticleViscosity(viscosity[i]));
    }

    _markerParticleViscosityLoadQueue.push_back(loadData);

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::loadDiffuseParticleData(FluidSimulationDiffuseParticleData data) {
    _logfile.log(std::ostringstream().flush() << 
                 _logfile.getTime() << " loadDiffuseParticleData: " << data.size << std::endl);

    if (data.size == 0) {
        return;
    }

    vmath::vec3 *positions = (vmath::vec3*)(data.positions);
    vmath::vec3 *velocities = (vmath::vec3*)(data.velocities);
    float *lifetimes = (float*)(data.lifetimes);
    DiffuseParticleType *types = (DiffuseParticleType*)(data.types);
    unsigned char *ids = (unsigned char*)(data.ids);

    DiffuseParticleLoadData loadData;
    loadData.particles.reserve(data.size);

    for (unsigned int i = 0; i < (unsigned int)data.size; i++) {
        DiffuseParticle dp;
        dp.position = positions[i];
        dp.velocity = velocities[i];
        dp.lifetime = lifetimes[i];
        dp.type = types[i];
        dp.id = ids[i];
        loadData.particles.push_back(dp);
    }

    _diffuseParticleLoadQueue.push_back(loadData);

    _isDiffuseParticleLoadPending = true;
}


/********************************************************************************
    Initializing the Fluid Simulator
********************************************************************************/

void FluidSimulation::_initializeSimulationGrids(int isize, int jsize, int ksize, double dx) {
    _logfile.separator();
    _logfile.timestamp();
    _logfile.newline();
    _logfile.log(std::ostringstream().flush() << 
                 "Initializing Simulation Grids:" << std::endl <<
                 "\tGrid Dimensions: " << isize << " x " << 
                                        jsize << " x " << 
                                        ksize << std::endl <<
                 "\tCell Size:       " << dx << std::endl);

    StopWatch t;
    t.start();
    _MACVelocity = MACVelocityField(isize, jsize, ksize, dx);
    _validVelocities = ValidVelocityComponentGrid(_isize, _jsize, _ksize);
    t.stop();

    _logfile.log("Constructing MACVelocityField: \t", t.getTime(), 4, 1);

    t.reset();
    t.start();
    _solidSDF = MeshLevelSet(isize, jsize, ksize, dx);
    if (_isStaticSolidLevelSetPrecomputed) {
        _staticSolidSDF = MeshLevelSet(isize, jsize, ksize, dx);
    }
    if (_isTempSolidLevelSetEnabled) {
        _tempSolidSDF = MeshLevelSet(isize, jsize, ksize, dx);
    }
    if (_isMeshingVolumeSet) {
        _meshingVolumeSDF = MeshLevelSet(isize, jsize, ksize, dx);
    }
    _liquidSDF = ParticleLevelSet(isize, jsize, ksize, dx);

    TriangleMesh domainBoundaryMesh = _getBoundaryTriangleMesh();
    _domainMeshObject = MeshObject(isize, jsize, ksize, dx);
    _domainMeshObject.updateMeshStatic(domainBoundaryMesh);
    _domainMeshObject.setFriction(_domainBoundaryFriction);
    _domainMeshObject.setWhitewaterInfluence(1.0);
    _domainMeshObject.setDustEmissionStrength(1.0);
    _domainMeshObject.setAsDomainObject();
    t.stop();

    _logfile.log("Constructing Level Sets:       \t", t.getTime(), 4, 1);

    t.reset();
    t.start();
    _weightGrid = WeightGrid(_isize, _jsize, _ksize);
    t.stop();

    _logfile.log("Constructing Weight Grid:      \t", t.getTime(), 4, 1);

    if (_isForceFieldsEnabled) {
        t.reset();
        t.start();
        _initializeForceFieldGrid(isize, jsize, ksize, dx);
        t.stop();

        _logfile.log("Constructing Force Field Grid: \t", t.getTime(), 4, 1);
    }

    _vertx = std::vector<float>({-0.4928, -0.0, -0.0364, -0.4928, 0.0017, -0.0364, -0.4928, 0.0039, -0.0364, -0.4868, -0.0, -0.0456, -0.485, -0.0, -0.0481, -0.486, -0.0, -0.0467, -0.4868, 0.0017, -0.0457, -0.4868, 0.0039, -0.0457, -0.4849, 0.0017, -0.0482, -0.486, 0.0017, -0.0468, -0.4849, 0.0039, -0.0482, -0.486, 0.0039, -0.0468, -0.4886, -0.0, -0.043, -0.4893, -0.0, -0.042, -0.4912, -0.0, -0.0391, -0.4922, -0.0, -0.0375, -0.4905, -0.0, -0.0402, -0.4896, -0.0, -0.0374, -0.4886, 0.0017, -0.043, -0.4895, 0.0017, -0.0418, -0.4886, 0.0039, -0.043, -0.4895, 0.0039, -0.0418, -0.4911, 0.0017, -0.0393, -0.4922, 0.0017, -0.0375, -0.4911, 0.0039, -0.0393, -0.4922, 0.0039, -0.0375, -0.4904, 0.0017, -0.0403, -0.4904, 0.0039, -0.0403, -0.4879, -0.0, -0.0441, -0.4874, -0.0, -0.0418,
-0.4852, -0.0, -0.0439, -0.4852, -0.0, -0.0418, -0.4874, -0.0, -0.0396, -0.4869, -0.0, -0.0368, -0.4845, -0.0, -0.0386, -0.4856, -0.0, -0.0378, -0.4879, 0.0017, -0.0441, -0.4879, 0.0039, -0.0441, -0.4868, 0.0017, -0.0368, -0.4868, 0.0039, -0.0368, -0.4844, 0.0017, -0.0387, -0.4855, 0.0017, -0.0379, -0.4844, 0.0039, -0.0387, -0.4855, 0.0039, -0.0379, -0.4951, -0.0, -0.0321, -0.4934, -0.0, -0.0354, -0.4945, -0.0, -0.0332, -0.4961, -0.0, -0.0297, -0.4935, -0.0, -0.0308, -0.4951, 0.0017, -0.0321, -0.4951, 0.0039, -0.0321, -0.4935, 0.0017, -0.0353, -0.4945, 0.0017, -0.0332, -0.4935, 0.0039, -0.0353, -0.4945, 0.0039, -0.0332, -0.4961, 0.0017, -0.0297, -0.4961, 0.0039, -0.0297, -0.4938, 0.0017, -0.0306, -0.4938, 0.0039, -0.0306, -0.4934, -0.0, -0.024,
-0.4953, -0.0, -0.0215, -0.496, -0.0, -0.0203, -0.4945, -0.0, -0.0226, -0.4935, -0.0, -0.0195, -0.4934, 0.0017, -0.0241, -0.4934, 0.0039, -0.0241, -0.4952, 0.0017, -0.0215, -0.4961, 0.0017, -0.0201, -0.4952, 0.0039, -0.0215, -0.4961, 0.0039, -0.0201, -0.4945, 0.0017, -0.0226, -0.4935, 0.0017, -0.0195, -0.4945, 0.0039, -0.0226, -0.4935, 0.0039, -0.0195, -0.4917, -0.0, -0.0353, -0.4914, -0.0, -0.0326, -0.4891, -0.0, -0.0347, -0.4902, -0.0, -0.0337, -0.4925, -0.0, -0.0317, -0.4894, -0.0, -0.0286, -0.4914, 0.0017, -0.0326, -0.4914, 0.0039, -0.0326, -0.4891, 0.0017, -0.0347, -0.4902, 0.0017, -0.0337, -0.4891, 0.0039, -0.0347, -0.4902, 0.0039, -0.0337, -0.4925, 0.0017, -0.0317, -0.4925, 0.0039, -0.0317, -0.4894, 0.0017, -0.0286, -0.4894, 0.0039, -0.0286,
-0.4879, -0.0, -0.0358, -0.4849, -0.0, -0.0327, -0.4872, -0.0, -0.0307, -0.4883, -0.0, -0.0297, -0.486, -0.0, -0.0318, -0.4852, -0.0, -0.0288, -0.4879, 0.0017, -0.0358, -0.4879, 0.0039, -0.0358, -0.4849, 0.0017, -0.0326, -0.4849, 0.0039, -0.0326, -0.4872, 0.0017, -0.0307, -0.4883, 0.0017, -0.0296, -0.4872, 0.0039, -0.0307, -0.4883, 0.0039, -0.0296, -0.486, 0.0017, -0.0317, -0.486, 0.0039, -0.0317, -0.4915, -0.0, -0.0263, -0.4926, -0.0, -0.0251, -0.4905, -0.0, -0.0274, -0.4896, -0.0, -0.0244, -0.4911, -0.0, -0.0214, -0.4922, -0.0, -0.0206, -0.4894, -0.0, -0.0227, -0.4915, 0.0017, -0.0264, -0.4925, 0.0017, -0.0252, -0.4915, 0.0039, -0.0264, -0.4925, 0.0039, -0.0252, -0.4905, 0.0017, -0.0274, -0.4905, 0.0039, -0.0274, -0.4911, 0.0017, -0.0215,
-0.4922, 0.0017, -0.0206, -0.4911, 0.0039, -0.0215, -0.4922, 0.0039, -0.0206, -0.4896, 0.0017, -0.0225, -0.4896, 0.0039, -0.0225, -0.4874, -0.0, -0.0266, -0.4873, -0.0, -0.0239, -0.4852, -0.0, -0.0266, -0.4851, -0.0, -0.025, -0.4884, -0.0, -0.0233, -0.4873, 0.0017, -0.0239, -0.4873, 0.0039, -0.0239, -0.4853, 0.0017, -0.025, -0.4853, 0.0039, -0.025, -0.4884, 0.0017, -0.0233, -0.4884, 0.0039, -0.0233, -0.4975, -0.0, -0.0173, -0.498, -0.0, -0.016, -0.4975, 0.0017, -0.0175, -0.498, 0.0017, -0.016, -0.4975, 0.0039, -0.0175, -0.498, 0.0039, -0.016, -0.4983, -0.0, -0.0157, -0.497, -0.0, -0.0167, -0.4946, -0.0, -0.0186, -0.4983, 0.0017, -0.0157, -0.497, 0.0017, -0.0167, -0.4983, 0.0039, -0.0157, -0.497, 0.0039, -0.0167, -0.4945, 0.0017, -0.0186,
-0.4945, 0.0039, -0.0186, -0.4561, -0.0, -0.0713, -0.4541, -0.0, -0.0736, -0.4552, -0.0, -0.0724, -0.4561, 0.0017, -0.0713, -0.4561, 0.0039, -0.0713, -0.4541, 0.0017, -0.0736, -0.4552, 0.0017, -0.0724, -0.4541, 0.0039, -0.0736, -0.4552, 0.0039, -0.0724, -0.452, -0.0, -0.0759, -0.4499, -0.0, -0.0781, -0.451, -0.0, -0.077, -0.4531, -0.0, -0.0747, -0.4527, -0.0, -0.0722, -0.4505, -0.0, -0.0743, -0.4505, -0.0, -0.0722, -0.452, 0.0017, -0.0759, -0.452, 0.0039, -0.0759, -0.4499, 0.0017, -0.0781, -0.451, 0.0017, -0.077, -0.4499, 0.0039, -0.0781, -0.451, 0.0039, -0.077, -0.4531, 0.0017, -0.0747, -0.4531, 0.0039, -0.0747, -0.4371, -0.0, -0.0891, -0.4372, 0.0017, -0.089, -0.4372, 0.0039, -0.089, -0.4345, -0.0, -0.0908, -0.4356, -0.0, -0.0901,
-0.4334, -0.0, -0.0916, -0.4332, -0.0, -0.0895, -0.4345, 0.0017, -0.0908, -0.4356, 0.0017, -0.0901, -0.4345, 0.0039, -0.0908, -0.4356, 0.0039, -0.0901, -0.4332, 0.0017, -0.0917, -0.4332, 0.0039, -0.0917, -0.4478, -0.0, -0.0802, -0.4455, -0.0, -0.0823, -0.4466, -0.0, -0.0813, -0.4478, 0.0017, -0.0802, -0.4478, 0.0039, -0.0802, -0.4455, 0.0017, -0.0823, -0.4466, 0.0017, -0.0813, -0.4455, 0.0039, -0.0823, -0.4466, 0.0039, -0.0813, -0.4432, -0.0, -0.0844, -0.4408, -0.0, -0.0863, -0.442, -0.0, -0.0855, -0.4443, -0.0, -0.0834, -0.444, -0.0, -0.0808, -0.4418, -0.0, -0.083, -0.4418, -0.0, -0.0808, -0.4432, 0.0017, -0.0844, -0.4432, 0.0039, -0.0844, -0.4408, 0.0017, -0.0863, -0.4419, 0.0017, -0.0855, -0.4408, 0.0039, -0.0863, -0.4419, 0.0039, -0.0855,
-0.4443, 0.0017, -0.0834, -0.4443, 0.0039, -0.0834, -0.4488, -0.0, -0.0792, -0.4483, -0.0, -0.0765, -0.4462, -0.0, -0.0787, -0.4462, -0.0, -0.0765, -0.4483, -0.0, -0.0743, -0.4483, -0.0, -0.0722, -0.4462, -0.0, -0.0743, -0.4462, -0.0, -0.0722, -0.4488, 0.0017, -0.0792, -0.4488, 0.0039, -0.0792, -0.444, -0.0, -0.0787, -0.444, -0.0, -0.0765, -0.4418, -0.0, -0.0787, -0.4418, -0.0, -0.0765, -0.444, -0.0, -0.0743, -0.444, -0.0, -0.0722, -0.4418, -0.0, -0.0743, -0.4418, -0.0, -0.0722, -0.4399, -0.0, -0.087, -0.4397, -0.0, -0.0852, -0.4382, -0.0, -0.0883, -0.4375, -0.0, -0.0852, -0.4397, -0.0, -0.083, -0.4397, -0.0, -0.0808, -0.4375, -0.0, -0.083, -0.4375, -0.0, -0.0808, -0.4397, 0.0017, -0.0872, -0.4397, 0.0039, -0.0872,
-0.4383, 0.0017, -0.0882, -0.4383, 0.0039, -0.0882, -0.4353, -0.0, -0.0874, -0.4353, -0.0, -0.0852, -0.4332, -0.0, -0.0874, -0.4332, -0.0, -0.0852, -0.4353, -0.0, -0.083, -0.4353, -0.0, -0.0808, -0.4332, -0.0, -0.083, -0.4332, -0.0, -0.0808, -0.4397, -0.0, -0.0787, -0.4397, -0.0, -0.0765, -0.4375, -0.0, -0.0787, -0.4375, -0.0, -0.0765, -0.4397, -0.0, -0.0743, -0.4397, -0.0, -0.0722, -0.4375, -0.0, -0.0743, -0.4375, -0.0, -0.0722, -0.4353, -0.0, -0.0787, -0.4353, -0.0, -0.0765, -0.4332, -0.0, -0.0787, -0.4332, -0.0, -0.0765, -0.4353, -0.0, -0.0743, -0.4353, -0.0, -0.0722, -0.4332, -0.0, -0.0743, -0.4332, -0.0, -0.0722, -0.4234, -0.0, -0.0972, -0.4235, 0.0017, -0.0972, -0.4235, 0.0039, -0.0972, -0.4306, -0.0, -0.0933,
-0.4277, -0.0, -0.0949, -0.4289, -0.0, -0.0943, -0.4316, -0.0, -0.0927, -0.431, -0.0, -0.0895, -0.4288, -0.0, -0.0917, -0.4288, -0.0, -0.0895, -0.4307, 0.0017, -0.0933, -0.4307, 0.0039, -0.0933, -0.4277, 0.0017, -0.0949, -0.4288, 0.0017, -0.0943, -0.4277, 0.0039, -0.0949, -0.4288, 0.0039, -0.0943, -0.4318, 0.0017, -0.0926, -0.4318, 0.0039, -0.0926, -0.4268, -0.0, -0.0955, -0.4266, -0.0, -0.0939, -0.4246, -0.0, -0.0966, -0.4245, -0.0, -0.0939, -0.4266, -0.0, -0.0917, -0.4266, -0.0, -0.0895, -0.4245, -0.0, -0.0917, -0.4245, -0.0, -0.0895, -0.4266, 0.0017, -0.0955, -0.4266, 0.0039, -0.0955, -0.4245, 0.0017, -0.0966, -0.4245, 0.0039, -0.0966, -0.4224, -0.0, -0.0977, -0.4205, -0.0, -0.0986, -0.4223, 0.0017, -0.0977, -0.4223, 0.0039, -0.0977,
-0.4201, 0.0017, -0.0988, -0.4201, 0.0039, -0.0988, -0.4177, -0.0, -0.0998, -0.4186, -0.0, -0.0995, -0.416, -0.0, -0.1005, -0.4158, -0.0, -0.0982, -0.4179, 0.0017, -0.0997, -0.419, 0.0017, -0.0993, -0.4179, 0.0039, -0.0997, -0.419, 0.0039, -0.0993, -0.4158, 0.0017, -0.1006, -0.4158, 0.0039, -0.1006, -0.4223, -0.0, -0.096, -0.4223, -0.0, -0.0939, -0.4201, -0.0, -0.096, -0.4201, -0.0, -0.0939, -0.4223, -0.0, -0.0917, -0.4223, -0.0, -0.0895, -0.4201, -0.0, -0.0917, -0.4201, -0.0, -0.0895, -0.418, -0.0, -0.096, -0.418, -0.0, -0.0939, -0.4158, -0.0, -0.096, -0.4158, -0.0, -0.0939, -0.418, -0.0, -0.0917, -0.418, -0.0, -0.0895, -0.4158, -0.0, -0.0917, -0.4158, -0.0, -0.0895, -0.431, -0.0, -0.0874, -0.431, -0.0, -0.0852,
-0.4288, -0.0, -0.0874, -0.4288, -0.0, -0.0852, -0.431, -0.0, -0.083, -0.431, -0.0, -0.0808, -0.4288, -0.0, -0.083, -0.4288, -0.0, -0.0808, -0.4266, -0.0, -0.0874, -0.4266, -0.0, -0.0852, -0.4245, -0.0, -0.0874, -0.4245, -0.0, -0.0852, -0.4266, -0.0, -0.083, -0.4266, -0.0, -0.0808, -0.4245, -0.0, -0.083, -0.4245, -0.0, -0.0808, -0.431, -0.0, -0.0787, -0.431, -0.0, -0.0765, -0.4288, -0.0, -0.0787, -0.4288, -0.0, -0.0765, -0.431, -0.0, -0.0743, -0.4307, -0.0, -0.072, -0.4287, -0.0, -0.0741, -0.4297, -0.0, -0.0731, -0.4307, 0.0017, -0.072, -0.4307, 0.0039, -0.072, -0.4287, 0.0017, -0.0741, -0.4297, 0.0017, -0.0731, -0.4287, 0.0039, -0.0741, -0.4297, 0.0039, -0.0731, -0.4266, -0.0, -0.0787, -0.4263, -0.0, -0.076,
-0.4239, -0.0, -0.0778, -0.4249, -0.0, -0.077, -0.4274, -0.0, -0.0751, -0.4263, 0.0017, -0.076, -0.4263, 0.0039, -0.076, -0.4238, 0.0017, -0.0779, -0.4249, 0.0017, -0.0771, -0.4238, 0.0039, -0.0779, -0.4249, 0.0039, -0.0771, -0.4274, 0.0017, -0.0752, -0.4274, 0.0039, -0.0752, -0.4223, -0.0, -0.0874, -0.4223, -0.0, -0.0852, -0.4201, -0.0, -0.0874, -0.4201, -0.0, -0.0852, -0.4223, -0.0, -0.083, -0.4223, -0.0, -0.0808, -0.4201, -0.0, -0.083, -0.42, -0.0, -0.0804, -0.42, 0.0017, -0.0804, -0.42, 0.0039, -0.0804, -0.418, -0.0, -0.0874, -0.418, -0.0, -0.0852, -0.4158, -0.0, -0.0874, -0.4158, -0.0, -0.0852, -0.4174, -0.0, -0.082, -0.4184, -0.0, -0.0814, -0.4155, -0.0, -0.083, -0.4172, 0.0017, -0.0821, -0.4183, 0.0017, -0.0815,
-0.4172, 0.0039, -0.0821, -0.4183, 0.0039, -0.0815, -0.4158, 0.0017, -0.0828, -0.4158, 0.0039, -0.0828, -0.4221, -0.0, -0.079, -0.4211, -0.0, -0.0797, -0.4223, 0.0017, -0.0789, -0.4223, 0.0039, -0.0789, -0.4211, 0.0017, -0.0797, -0.4211, 0.0039, -0.0797, -0.4799, -0.0, -0.0538, -0.4799, 0.0017, -0.0538, -0.4799, 0.0039, -0.0538, -0.4777, -0.0, -0.0559, -0.4788, -0.0, -0.0549, -0.4768, -0.0, -0.0567, -0.4766, -0.0, -0.0548, -0.4777, 0.0017, -0.0559, -0.4788, 0.0017, -0.0549, -0.4777, 0.0039, -0.0559, -0.4788, 0.0039, -0.0549, -0.4766, 0.0017, -0.0569, -0.4766, 0.0039, -0.0569, -0.469, -0.0, -0.0624, -0.468, -0.0, -0.063, -0.469, 0.0017, -0.0624, -0.469, 0.0039, -0.0624, -0.4679, 0.0017, -0.0631, -0.4679, 0.0039, -0.0631, -0.4742, -0.0, -0.0589,
-0.4717, -0.0, -0.0607, -0.4728, -0.0, -0.0599, -0.4753, -0.0, -0.058, -0.4744, -0.0, -0.0548, -0.4722, -0.0, -0.057, -0.4722, -0.0, -0.0548, -0.4743, 0.0017, -0.0588, -0.4743, 0.0039, -0.0588, -0.4717, 0.0017, -0.0606, -0.4728, 0.0017, -0.0599, -0.4717, 0.0039, -0.0606, -0.4728, 0.0039, -0.0599, -0.4753, 0.0017, -0.058, -0.4753, 0.0039, -0.058, -0.4701, -0.0, -0.0617, -0.47, -0.0, -0.0591, -0.4679, -0.0, -0.0613, -0.4679, -0.0, -0.0591, -0.47, -0.0, -0.057, -0.47, -0.0, -0.0548, -0.4679, -0.0, -0.057, -0.4679, -0.0, -0.0548, -0.4701, 0.0017, -0.0617, -0.4701, 0.0039, -0.0617, -0.482, -0.0, -0.0516, -0.4828, -0.0, -0.0507, -0.481, -0.0, -0.0527, -0.4809, -0.0, -0.0505, -0.4841, -0.0, -0.0492, -0.4831, -0.0, -0.0461,
-0.4809, -0.0, -0.0483, -0.4809, -0.0, -0.0461, -0.482, 0.0017, -0.0516, -0.483, 0.0017, -0.0505, -0.482, 0.0039, -0.0516, -0.483, 0.0039, -0.0505, -0.481, 0.0017, -0.0527, -0.481, 0.0039, -0.0527, -0.4841, 0.0017, -0.0493, -0.4841, 0.0039, -0.0493, -0.4787, -0.0, -0.0526, -0.4787, -0.0, -0.0505, -0.4766, -0.0, -0.0526, -0.4766, -0.0, -0.0505, -0.4787, -0.0, -0.0483, -0.4787, -0.0, -0.0461, -0.4766, -0.0, -0.0483, -0.4766, -0.0, -0.0461, -0.4831, -0.0, -0.0439, -0.4831, -0.0, -0.0418, -0.4809, -0.0, -0.0439, -0.4807, -0.0, -0.0413, -0.4828, -0.0, -0.0398, -0.4818, -0.0, -0.0405, -0.4807, 0.0017, -0.0413, -0.4807, 0.0039, -0.0413, -0.4831, 0.0017, -0.0397, -0.4831, 0.0039, -0.0397, -0.4818, 0.0017, -0.0406, -0.4818, 0.0039, -0.0406,
-0.4781, -0.0, -0.043, -0.4791, -0.0, -0.0423, -0.4763, -0.0, -0.044, -0.4783, -0.0, -0.0369, -0.4768, -0.0, -0.0376, -0.4779, 0.0017, -0.0431, -0.479, 0.0017, -0.0424, -0.4779, 0.0039, -0.0431, -0.479, 0.0039, -0.0424, -0.4766, 0.0017, -0.0439, -0.4766, 0.0039, -0.0439, -0.4785, 0.0017, -0.0368, -0.4785, 0.0039, -0.0368, -0.4766, 0.0017, -0.0377, -0.4766, 0.0039, -0.0377, -0.4744, -0.0, -0.0526, -0.4744, -0.0, -0.0505, -0.4722, -0.0, -0.0526, -0.4722, -0.0, -0.0505, -0.4744, -0.0, -0.0483, -0.474, -0.0, -0.0452, -0.4722, -0.0, -0.0483, -0.472, -0.0, -0.046, -0.4738, 0.0017, -0.0453, -0.4738, 0.0039, -0.0453, -0.4722, 0.0017, -0.0459, -0.4722, 0.0039, -0.0459, -0.47, -0.0, -0.0526, -0.47, -0.0, -0.0505, -0.4679, -0.0, -0.0526,
-0.4679, -0.0, -0.0505, -0.47, -0.0, -0.0483, -0.47, -0.0, -0.0468, -0.4679, -0.0, -0.0476, -0.4689, -0.0, -0.0472, -0.4701, 0.0017, -0.0468, -0.4701, 0.0039, -0.0468, -0.4678, 0.0017, -0.0476, -0.4689, 0.0017, -0.0472, -0.4678, 0.0039, -0.0476, -0.4689, 0.0039, -0.0472, -0.475, -0.0, -0.0447, -0.4737, -0.0, -0.0388, -0.4747, -0.0, -0.0384, -0.4723, -0.0, -0.0393, -0.4722, -0.0, -0.0374, -0.4749, 0.0017, -0.0448, -0.4749, 0.0039, -0.0448, -0.4739, 0.0017, -0.0387, -0.475, 0.0017, -0.0383, -0.4739, 0.0039, -0.0387, -0.475, 0.0039, -0.0383, -0.4722, 0.0017, -0.0393, -0.4722, 0.0039, -0.0393, -0.467, -0.0, -0.0408, -0.4702, -0.0, -0.0399, -0.47, -0.0, -0.0374, -0.468, -0.0, -0.0405, -0.4679, -0.0, -0.0374, -0.4671, 0.0017, -0.0408,
-0.4671, 0.0039, -0.0408, -0.47, 0.0017, -0.04, -0.47, 0.0039, -0.04, -0.4682, 0.0017, -0.0405, -0.4682, 0.0039, -0.0405, -0.4648, -0.0, -0.0647, -0.4659, -0.0, -0.0641, -0.4637, -0.0, -0.0653, -0.4635, -0.0, -0.0635, -0.4649, 0.0017, -0.0647, -0.466, 0.0017, -0.0641, -0.4649, 0.0039, -0.0647, -0.466, 0.0039, -0.0641, -0.4635, 0.0017, -0.0654, -0.4635, 0.0039, -0.0654, -0.4606, -0.0, -0.067, -0.4582, -0.0, -0.069, -0.4593, -0.0, -0.068, -0.4617, -0.0, -0.0664, -0.4614, -0.0, -0.0635, -0.4592, -0.0, -0.0656, -0.4592, -0.0, -0.0635, -0.4607, 0.0017, -0.067, -0.4607, 0.0039, -0.067, -0.4582, 0.0017, -0.069, -0.4593, 0.0017, -0.068, -0.4582, 0.0039, -0.069, -0.4593, 0.0039, -0.068, -0.4618, 0.0017, -0.0663, -0.4618, 0.0039, -0.0663,
-0.4657, -0.0, -0.0613, -0.4657, -0.0, -0.0591, -0.4635, -0.0, -0.0613, -0.4635, -0.0, -0.0591, -0.4657, -0.0, -0.057, -0.4657, -0.0, -0.0548, -0.4635, -0.0, -0.057, -0.4635, -0.0, -0.0548, -0.4614, -0.0, -0.0613, -0.4614, -0.0, -0.0591, -0.4592, -0.0, -0.0613, -0.4592, -0.0, -0.0591, -0.4614, -0.0, -0.057, -0.4614, -0.0, -0.0548, -0.4592, -0.0, -0.057, -0.4592, -0.0, -0.0548, -0.4572, -0.0, -0.0701, -0.457, -0.0, -0.0678, -0.4549, -0.0, -0.07, -0.4549, -0.0, -0.0678, -0.457, -0.0, -0.0656, -0.457, -0.0, -0.0635, -0.4549, -0.0, -0.0656, -0.4549, -0.0, -0.0635, -0.4572, 0.0017, -0.0701, -0.4572, 0.0039, -0.0701, -0.4527, -0.0, -0.07, -0.4527, -0.0, -0.0678, -0.4505, -0.0, -0.07, -0.4505, -0.0, -0.0678,
-0.4527, -0.0, -0.0656, -0.4527, -0.0, -0.0635, -0.4505, -0.0, -0.0656, -0.4505, -0.0, -0.0635, -0.457, -0.0, -0.0613, -0.457, -0.0, -0.0591, -0.4549, -0.0, -0.0613, -0.4549, -0.0, -0.0591, -0.457, -0.0, -0.057, -0.457, -0.0, -0.0548, -0.4549, -0.0, -0.057, -0.4549, -0.0, -0.0548, -0.4527, -0.0, -0.0613, -0.4527, -0.0, -0.0591, -0.4505, -0.0, -0.0613, -0.4505, -0.0, -0.0591, -0.4527, -0.0, -0.057, -0.4527, -0.0, -0.0548, -0.4505, -0.0, -0.057, -0.4505, -0.0, -0.0548, -0.4657, -0.0, -0.0526, -0.4657, -0.0, -0.0505, -0.4635, -0.0, -0.0526, -0.4635, -0.0, -0.0505, -0.4656, -0.0, -0.0483, -0.4635, -0.0, -0.0488, -0.4657, 0.0017, -0.0482, -0.4657, 0.0039, -0.0482, -0.4635, 0.0017, -0.0487, -0.4635, 0.0039, -0.0487,
-0.4614, -0.0, -0.0526, -0.461, -0.0, -0.0493, -0.4592, -0.0, -0.0526, -0.459, -0.0, -0.0497, -0.4618, -0.0, -0.0491, -0.4604, 0.0017, -0.0494, -0.4604, 0.0039, -0.0494, -0.4592, 0.0017, -0.0497, -0.4592, 0.0039, -0.0497, -0.4615, 0.0017, -0.0492, -0.4615, 0.0039, -0.0492, -0.4658, -0.0, -0.0411, -0.4636, -0.0, -0.0416, -0.4657, -0.0, -0.0396, -0.4657, -0.0, -0.0374, -0.4635, -0.0, -0.0396, -0.4635, -0.0, -0.0374, -0.4657, 0.0017, -0.0411, -0.4657, 0.0039, -0.0411, -0.4635, 0.0017, -0.0416, -0.4635, 0.0039, -0.0416, -0.4615, -0.0, -0.0421, -0.4593, -0.0, -0.0424, -0.4614, -0.0, -0.0396, -0.4614, -0.0, -0.0374, -0.4592, -0.0, -0.0396, -0.4592, -0.0, -0.0374, -0.4614, 0.0017, -0.0421, -0.4614, 0.0039, -0.0421, -0.4592, 0.0017, -0.0424,
-0.4592, 0.0039, -0.0424, -0.457, -0.0, -0.0526, -0.4569, -0.0, -0.05, -0.4549, -0.0, -0.0526, -0.4548, -0.0, -0.0502, -0.457, 0.0017, -0.05, -0.457, 0.0039, -0.05, -0.4549, 0.0017, -0.0502, -0.4549, 0.0039, -0.0502, -0.4527, -0.0, -0.0526, -0.4527, -0.0, -0.0503, -0.4505, -0.0, -0.0526, -0.4505, -0.0, -0.0503, -0.4527, 0.0017, -0.0503, -0.4527, 0.0039, -0.0503, -0.4505, 0.0017, -0.0503, -0.4505, 0.0039, -0.0503, -0.457, -0.0, -0.0425, -0.4549, -0.0, -0.0425, -0.457, -0.0, -0.0396, -0.457, -0.0, -0.0374, -0.4549, -0.0, -0.0396, -0.4549, -0.0, -0.0374, -0.457, 0.0017, -0.0425, -0.457, 0.0039, -0.0425, -0.4549, 0.0017, -0.0425, -0.4549, 0.0039, -0.0425, -0.4527, -0.0, -0.0425, -0.4505, -0.0, -0.0425, -0.4527, -0.0, -0.0396,
-0.4527, -0.0, -0.0374, -0.4505, -0.0, -0.0396, -0.4505, -0.0, -0.0374, -0.4527, 0.0017, -0.0425, -0.4527, 0.0039, -0.0425, -0.4505, 0.0017, -0.0425, -0.4505, 0.0039, -0.0425, -0.4824, -0.0, -0.0345, -0.4835, -0.0, -0.0338, -0.4812, -0.0, -0.0353, -0.4809, -0.0, -0.0331, -0.4831, -0.0, -0.0309, -0.4831, -0.0, -0.0288, -0.4809, -0.0, -0.0309, -0.4809, -0.0, -0.0288, -0.4824, 0.0017, -0.0345, -0.4835, 0.0017, -0.0337, -0.4824, 0.0039, -0.0345, -0.4835, 0.0039, -0.0337, -0.4809, 0.0017, -0.0355, -0.4809, 0.0039, -0.0355, -0.4793, -0.0, -0.0363, -0.4787, -0.0, -0.0331, -0.4766, -0.0, -0.0353, -0.4766, -0.0, -0.0331, -0.4787, -0.0, -0.0309, -0.4787, -0.0, -0.0288, -0.4766, -0.0, -0.0309, -0.4765, -0.0, -0.0279, -0.4795, 0.0017, -0.0362,
-0.4795, 0.0039, -0.0362, -0.4764, 0.0017, -0.0279, -0.4764, 0.0039, -0.0279, -0.483, -0.0, -0.0259, -0.4841, -0.0, -0.0255, -0.4807, -0.0, -0.0267, -0.483, 0.0017, -0.0259, -0.4841, 0.0017, -0.0255, -0.483, 0.0039, -0.0259, -0.4841, 0.0039, -0.0255, -0.4809, 0.0017, -0.0266, -0.4809, 0.0039, -0.0266, -0.4787, -0.0, -0.0273, -0.4775, -0.0, -0.0276, -0.4787, 0.0017, -0.0273, -0.4787, 0.0039, -0.0273, -0.4775, 0.0017, -0.0276, -0.4775, 0.0039, -0.0276, -0.4744, -0.0, -0.0353, -0.4744, -0.0, -0.0331, -0.4722, -0.0, -0.0353, -0.4722, -0.0, -0.0331, -0.4744, -0.0, -0.0309, -0.4743, -0.0, -0.0284, -0.4722, -0.0, -0.0309, -0.4722, -0.0, -0.0286, -0.4744, 0.0017, -0.0283, -0.4744, 0.0039, -0.0283, -0.4722, 0.0017, -0.0286, -0.4722, 0.0039, -0.0286,
-0.47, -0.0, -0.0353, -0.47, -0.0, -0.0331, -0.4679, -0.0, -0.0353, -0.4679, -0.0, -0.0331, -0.47, -0.0, -0.0309, -0.47, -0.0, -0.0287, -0.4679, -0.0, -0.0309, -0.4679, -0.0, -0.0287, -0.47, 0.0017, -0.0287, -0.47, 0.0039, -0.0287, -0.4679, 0.0017, -0.0287, -0.4679, 0.0039, -0.0287, -0.4801, -0.0, -0.0148, -0.4824, -0.0, -0.014, -0.4823, -0.0, -0.0114, -0.4811, -0.0, -0.0145, -0.4809, -0.0, -0.0114, -0.4801, 0.0017, -0.0148, -0.4801, 0.0039, -0.0148, -0.4824, 0.0017, -0.014, -0.4823, 0.0017, -0.0114, -0.4824, 0.0039, -0.014, -0.4823, 0.0039, -0.0114, -0.4812, 0.0017, -0.0144, -0.4812, 0.0039, -0.0144, -0.4788, -0.0, -0.0152, -0.4767, -0.0, -0.0158, -0.4787, -0.0, -0.0136, -0.4787, -0.0, -0.0114, -0.4766, -0.0, -0.0136,
-0.4766, -0.0, -0.0114, -0.4787, 0.0017, -0.0152, -0.4787, 0.0039, -0.0152, -0.4766, 0.0017, -0.0159, -0.4766, 0.0039, -0.0159, -0.4821, -0.0, -0.0097, -0.482, -0.0, -0.0089, -0.4817, -0.0, -0.0069, -0.4814, -0.0, -0.0048, -0.4811, -0.0, -0.0026, -0.4821, 0.0017, -0.0094, -0.4821, 0.0039, -0.0094, -0.482, 0.0017, -0.0083, -0.4818, 0.0017, -0.0071, -0.482, 0.0039, -0.0083, -0.4818, 0.0039, -0.0071, -0.4815, 0.0017, -0.0049, -0.4812, 0.0017, -0.0027, -0.4815, 0.0039, -0.0049, -0.4812, 0.0039, -0.0027, -0.4787, -0.0, -0.0092, -0.4787, -0.0, -0.0071, -0.4766, -0.0, -0.0092, -0.4766, -0.0, -0.0071, -0.4787, -0.0, -0.0049, -0.4787, -0.0, -0.0027, -0.4766, -0.0, -0.0049, -0.4766, -0.0, -0.0027, -0.4746, -0.0, -0.0164, -0.4713, -0.0, -0.0168,
-0.4722, -0.0, -0.0168, -0.4744, -0.0, -0.0136, -0.4744, -0.0, -0.0114, -0.4722, -0.0, -0.0136, -0.4722, -0.0, -0.0114, -0.4744, 0.0017, -0.0164, -0.4744, 0.0039, -0.0164, -0.4713, 0.0017, -0.0168, -0.4724, 0.0017, -0.0167, -0.4713, 0.0039, -0.0168, -0.4724, 0.0039, -0.0167, -0.47, -0.0, -0.0169, -0.47, -0.0, -0.0157, -0.4679, -0.0, -0.017, -0.4679, -0.0, -0.0157, -0.47, -0.0, -0.0136, -0.47, -0.0, -0.0114, -0.4679, -0.0, -0.0136, -0.4679, -0.0, -0.0114, -0.47, 0.0017, -0.0169, -0.47, 0.0039, -0.0169, -0.4679, 0.0017, -0.017, -0.4679, 0.0039, -0.017, -0.4744, -0.0, -0.0092, -0.4744, -0.0, -0.0071, -0.4722, -0.0, -0.0092, -0.4722, -0.0, -0.0071, -0.4744, -0.0, -0.0049, -0.4744, -0.0, -0.0027, -0.4722, -0.0, -0.0049,
-0.4722, -0.0, -0.0027, -0.47, -0.0, -0.0092, -0.47, -0.0, -0.0071, -0.4679, -0.0, -0.0092, -0.4679, -0.0, -0.0071, -0.47, -0.0, -0.0049, -0.47, -0.0, -0.0027, -0.4679, -0.0, -0.0049, -0.4679, -0.0, -0.0027, -0.4657, -0.0, -0.0353, -0.4657, -0.0, -0.0331, -0.4635, -0.0, -0.0353, -0.4635, -0.0, -0.0331, -0.4657, -0.0, -0.0309, -0.4657, -0.0, -0.0287, -0.4635, -0.0, -0.0309, -0.4636, -0.0, -0.0287, -0.4657, 0.0017, -0.0287, -0.4657, 0.0039, -0.0287, -0.4635, 0.0017, -0.0287, -0.4635, 0.0039, -0.0287, -0.4614, -0.0, -0.0353, -0.4614, -0.0, -0.0331, -0.4592, -0.0, -0.0353, -0.4592, -0.0, -0.0331, -0.4614, -0.0, -0.0309, -0.4615, -0.0, -0.0284, -0.4592, -0.0, -0.0309, -0.4594, -0.0, -0.028, -0.4614, 0.0017, -0.0284,
-0.4614, 0.0039, -0.0284, -0.4592, 0.0017, -0.028, -0.4592, 0.0039, -0.028, -0.457, -0.0, -0.0353, -0.457, -0.0, -0.0331, -0.4549, -0.0, -0.0353, -0.4549, -0.0, -0.0331, -0.457, -0.0, -0.0309, -0.4574, -0.0, -0.0276, -0.4549, -0.0, -0.0309, -0.4549, -0.0, -0.0288, -0.458, 0.0017, -0.0277, -0.458, 0.0039, -0.0277, -0.4527, -0.0, -0.0353, -0.4527, -0.0, -0.0331, -0.4505, -0.0, -0.0353, -0.4505, -0.0, -0.0331, -0.4527, -0.0, -0.0309, -0.4527, -0.0, -0.0288, -0.4505, -0.0, -0.0309, -0.4505, -0.0, -0.0288, -0.4566, -0.0, -0.0274, -0.4549, -0.0, -0.027, -0.4569, 0.0017, -0.0274, -0.4569, 0.0039, -0.0274, -0.4549, 0.0017, -0.027, -0.4549, 0.0039, -0.027, -0.4528, -0.0, -0.0265, -0.4507, -0.0, -0.0261, -0.4527, 0.0017, -0.0265,
-0.4527, 0.0039, -0.0265, -0.4505, 0.0017, -0.026, -0.4505, 0.0039, -0.026, -0.4657, -0.0, -0.017, -0.4657, -0.0, -0.0158, -0.4655, -0.0, -0.0136, -0.4652, -0.0, -0.0115, -0.4657, 0.0017, -0.017, -0.4657, 0.0017, -0.0157, -0.4657, 0.0039, -0.017, -0.4657, 0.0039, -0.0157, -0.4655, 0.0017, -0.0136, -0.4651, 0.0017, -0.0114, -0.4655, 0.0039, -0.0136, -0.4651, 0.0039, -0.0114, -0.4647, -0.0, -0.0093, -0.4657, -0.0, -0.0071, -0.4645, -0.0, -0.0083, -0.4643, -0.0, -0.0071, -0.4657, -0.0, -0.0049, -0.4657, -0.0, -0.0027, -0.4638, -0.0, -0.005, -0.4633, -0.0, -0.0029, -0.4648, 0.0017, -0.0095, -0.4648, 0.0039, -0.0095, -0.4646, 0.0017, -0.0084, -0.4643, 0.0017, -0.0071, -0.4646, 0.0039, -0.0084, -0.4643, 0.0039, -0.0071, -0.4638, 0.0017, -0.0049,
-0.4633, 0.0017, -0.0027, -0.4638, 0.0039, -0.0049, -0.4633, 0.0039, -0.0027, -0.4483, -0.0, -0.07, -0.4483, -0.0, -0.0678, -0.4462, -0.0, -0.07, -0.4462, -0.0, -0.0678, -0.4483, -0.0, -0.0656, -0.4483, -0.0, -0.0635, -0.4462, -0.0, -0.0656, -0.4462, -0.0, -0.0635, -0.444, -0.0, -0.07, -0.444, -0.0, -0.0678, -0.4418, -0.0, -0.07, -0.4418, -0.0, -0.0678, -0.444, -0.0, -0.0656, -0.444, -0.0, -0.0635, -0.4418, -0.0, -0.0656, -0.4418, -0.0, -0.0635, -0.4483, -0.0, -0.0613, -0.4483, -0.0, -0.0591, -0.4462, -0.0, -0.0613, -0.4462, -0.0, -0.0591, -0.4483, -0.0, -0.057, -0.4483, -0.0, -0.0548, -0.4462, -0.0, -0.057, -0.4462, -0.0, -0.0548, -0.444, -0.0, -0.0613, -0.444, -0.0, -0.0591, -0.4418, -0.0, -0.0613,
-0.4418, -0.0, -0.0591, -0.444, -0.0, -0.057, -0.444, -0.0, -0.0548, -0.4418, -0.0, -0.057, -0.4418, -0.0, -0.0548, -0.4397, -0.0, -0.07, -0.4397, -0.0, -0.0678, -0.4375, -0.0, -0.07, -0.4375, -0.0, -0.0678, -0.4397, -0.0, -0.0656, -0.4397, -0.0, -0.0635, -0.4375, -0.0, -0.0656, -0.4375, -0.0, -0.0635, -0.4353, -0.0, -0.07, -0.4353, -0.0, -0.0678, -0.433, -0.0, -0.069, -0.4332, -0.0, -0.0678, -0.4353, -0.0, -0.0656, -0.4353, -0.0, -0.0635, -0.4332, -0.0, -0.0656, -0.4332, -0.0, -0.0635, -0.4397, -0.0, -0.0613, -0.4397, -0.0, -0.0591, -0.4375, -0.0, -0.0613, -0.4375, -0.0, -0.0591, -0.4397, -0.0, -0.057, -0.4397, -0.0, -0.0548, -0.4375, -0.0, -0.057, -0.4375, -0.0, -0.0548, -0.4353, -0.0, -0.0613,
-0.4353, -0.0, -0.0591, -0.4332, -0.0, -0.0613, -0.4332, -0.0, -0.0591, -0.4353, -0.0, -0.057, -0.4353, -0.0, -0.0548, -0.4332, -0.0, -0.057, -0.4332, -0.0, -0.0548, -0.4483, -0.0, -0.0526, -0.4484, -0.0, -0.0503, -0.4462, -0.0, -0.0526, -0.4463, -0.0, -0.0501, -0.4483, 0.0017, -0.0503, -0.4483, 0.0039, -0.0503, -0.4462, 0.0017, -0.0501, -0.4462, 0.0039, -0.0501, -0.444, -0.0, -0.0526, -0.4441, -0.0, -0.0498, -0.4418, -0.0, -0.0526, -0.442, -0.0, -0.0494, -0.4411, -0.0, -0.0492, -0.444, 0.0017, -0.0497, -0.444, 0.0039, -0.0497, -0.4424, 0.0017, -0.0495, -0.4424, 0.0039, -0.0495, -0.4413, 0.0017, -0.0493, -0.4413, 0.0039, -0.0493, -0.4483, -0.0, -0.0425, -0.4461, -0.0, -0.0423, -0.4483, -0.0, -0.0396, -0.4483, -0.0, -0.0374,
-0.4462, -0.0, -0.0396, -0.4462, -0.0, -0.0374, -0.4483, 0.0017, -0.0425, -0.4483, 0.0039, -0.0425, -0.4462, 0.0017, -0.0423, -0.4462, 0.0039, -0.0423, -0.4439, -0.0, -0.042, -0.4417, -0.0, -0.0416, -0.444, -0.0, -0.0396, -0.444, -0.0, -0.0374, -0.4418, -0.0, -0.0396, -0.4418, -0.0, -0.0374, -0.444, 0.0017, -0.042, -0.444, 0.0039, -0.042, -0.4418, 0.0017, -0.0416, -0.4418, 0.0039, -0.0416, -0.4397, -0.0, -0.0526, -0.4397, -0.0, -0.0505, -0.4375, -0.0, -0.0526, -0.4375, -0.0, -0.0505, -0.4397, -0.0, -0.049, -0.4376, -0.0, -0.0486, -0.4397, 0.0017, -0.049, -0.4397, 0.0039, -0.049, -0.4375, 0.0017, -0.0486, -0.4375, 0.0039, -0.0486, -0.4353, -0.0, -0.0526, -0.4353, -0.0, -0.0505, -0.4332, -0.0, -0.0526, -0.4332, -0.0, -0.0505,
-0.4355, -0.0, -0.0482, -0.4334, -0.0, -0.0476, -0.4353, 0.0017, -0.0481, -0.4353, 0.0039, -0.0481, -0.4331, 0.0017, -0.0476, -0.4331, 0.0039, -0.0476, -0.4396, -0.0, -0.0412, -0.4382, -0.0, -0.0409, -0.4397, -0.0, -0.0396, -0.4397, -0.0, -0.0374, -0.4373, -0.0, -0.0406, -0.4375, -0.0, -0.0374, -0.4397, 0.0017, -0.0412, -0.4397, 0.0039, -0.0412, -0.4381, 0.0017, -0.0408, -0.4381, 0.0039, -0.0408, -0.437, 0.0017, -0.0405, -0.437, 0.0039, -0.0405, -0.4351, -0.0, -0.0401, -0.4353, -0.0, -0.0374, -0.433, -0.0, -0.0396, -0.4332, -0.0, -0.0374, -0.4353, 0.0017, -0.0401, -0.4353, 0.0039, -0.0401, -0.4331, 0.0017, -0.0396, -0.4331, 0.0039, -0.0396, -0.4328, -0.0, -0.069, -0.4311, -0.0, -0.0689, -0.4286, -0.0, -0.0686, -0.431, -0.0, -0.0656,
-0.431, -0.0, -0.0635, -0.4288, -0.0, -0.0656, -0.4288, -0.0, -0.0635, -0.4328, 0.0017, -0.069, -0.4308, 0.0017, -0.0689, -0.4328, 0.0039, -0.069, -0.4308, 0.0039, -0.0689, -0.4288, 0.0017, -0.0686, -0.4288, 0.0039, -0.0686, -0.4265, -0.0, -0.0681, -0.4244, -0.0, -0.0675, -0.4266, -0.0, -0.0656, -0.4266, -0.0, -0.0635, -0.4245, -0.0, -0.0656, -0.4245, -0.0, -0.0635, -0.4266, 0.0017, -0.0681, -0.4266, 0.0039, -0.0681, -0.4245, 0.0017, -0.0675, -0.4245, 0.0039, -0.0675, -0.431, -0.0, -0.0613, -0.431, -0.0, -0.0591, -0.4288, -0.0, -0.0613, -0.4288, -0.0, -0.0591, -0.431, -0.0, -0.057, -0.431, -0.0, -0.0548, -0.4288, -0.0, -0.057, -0.4288, -0.0, -0.0548, -0.4266, -0.0, -0.0613, -0.4266, -0.0, -0.0591, -0.4245, -0.0, -0.0613,
-0.4245, -0.0, -0.0591, -0.4266, -0.0, -0.057, -0.4266, -0.0, -0.0548, -0.4245, -0.0, -0.057, -0.4245, -0.0, -0.0548, -0.4229, -0.0, -0.067, -0.4219, -0.0, -0.0667, -0.4223, -0.0, -0.0635, -0.4199, -0.0, -0.0659, -0.4201, -0.0, -0.0635, -0.4227, 0.0017, -0.067, -0.4227, 0.0039, -0.067, -0.4216, 0.0017, -0.0666, -0.4216, 0.0039, -0.0666, -0.4201, 0.0017, -0.066, -0.4201, 0.0039, -0.066, -0.4179, -0.0, -0.0651, -0.418, -0.0, -0.0635, -0.4167, -0.0, -0.0647, -0.4157, -0.0, -0.0642, -0.418, 0.0017, -0.0651, -0.418, 0.0039, -0.0651, -0.4167, 0.0017, -0.0646, -0.4156, 0.0017, -0.0642, -0.4167, 0.0039, -0.0646, -0.4156, 0.0039, -0.0642, -0.4223, -0.0, -0.0613, -0.4223, -0.0, -0.0591, -0.4201, -0.0, -0.0613, -0.4201, -0.0, -0.0591,
-0.4223, -0.0, -0.057, -0.4223, -0.0, -0.0548, -0.4201, -0.0, -0.057, -0.4201, -0.0, -0.0548, -0.418, -0.0, -0.0613, -0.418, -0.0, -0.0591, -0.4158, -0.0, -0.0613, -0.4158, -0.0, -0.0591, -0.418, -0.0, -0.057, -0.418, -0.0, -0.0548, -0.4158, -0.0, -0.057, -0.4158, -0.0, -0.0548, -0.431, -0.0, -0.0526, -0.431, -0.0, -0.0505, -0.4288, -0.0, -0.0526, -0.4288, -0.0, -0.0505, -0.4315, -0.0, -0.0471, -0.4306, -0.0, -0.0468, -0.4288, -0.0, -0.0483, -0.429, -0.0, -0.0463, -0.432, 0.0017, -0.0472, -0.4309, 0.0017, -0.0469, -0.432, 0.0039, -0.0472, -0.4309, 0.0039, -0.0469, -0.4288, 0.0017, -0.0462, -0.4288, 0.0039, -0.0462, -0.4266, -0.0, -0.0526, -0.4266, -0.0, -0.0505, -0.4245, -0.0, -0.0526, -0.4245, -0.0, -0.0505,
-0.4266, -0.0, -0.0483, -0.4267, -0.0, -0.0454, -0.4245, -0.0, -0.0483, -0.4245, -0.0, -0.0461, -0.4267, 0.0017, -0.0454, -0.4267, 0.0039, -0.0454, -0.4309, -0.0, -0.039, -0.431, -0.0, -0.0374, -0.4298, -0.0, -0.0386, -0.4287, -0.0, -0.0382, -0.431, 0.0017, -0.039, -0.431, 0.0039, -0.039, -0.4297, 0.0017, -0.0386, -0.4286, 0.0017, -0.0382, -0.4297, 0.0039, -0.0386, -0.4286, 0.0039, -0.0382, -0.4256, -0.0, -0.045, -0.4245, -0.0, -0.0446, -0.4265, -0.0, -0.0373, -0.4251, -0.0, -0.0367, -0.4256, 0.0017, -0.045, -0.4256, 0.0039, -0.045, -0.4245, 0.0017, -0.0446, -0.4245, 0.0039, -0.0446, -0.4266, 0.0017, -0.0373, -0.4266, 0.0039, -0.0373, -0.425, 0.0017, -0.0366, -0.425, 0.0039, -0.0366, -0.4223, -0.0, -0.0526, -0.4223, -0.0, -0.0505,
-0.4201, -0.0, -0.0526, -0.4201, -0.0, -0.0505, -0.4223, -0.0, -0.0483, -0.4223, -0.0, -0.0461, -0.4201, -0.0, -0.0483, -0.4201, -0.0, -0.0461, -0.418, -0.0, -0.0526, -0.418, -0.0, -0.0505, -0.4158, -0.0, -0.0526, -0.4158, -0.0, -0.0505, -0.418, -0.0, -0.0483, -0.418, -0.0, -0.0461, -0.4158, -0.0, -0.0483, -0.4158, -0.0, -0.0461, -0.4226, -0.0, -0.0437, -0.4206, -0.0, -0.0429, -0.4196, -0.0, -0.0425, -0.4223, 0.0017, -0.0436, -0.4223, 0.0039, -0.0436, -0.4209, 0.0017, -0.043, -0.4198, 0.0017, -0.0425, -0.4209, 0.0039, -0.043, -0.4198, 0.0039, -0.0425, -0.418, -0.0, -0.0439, -0.4182, -0.0, -0.0419, -0.4158, -0.0, -0.0439, -0.4161, -0.0, -0.0409, -0.4151, -0.0, -0.0404, -0.418, 0.0017, -0.0418, -0.418, 0.0039, -0.0418,
-0.4162, 0.0017, -0.041, -0.4162, 0.0039, -0.041, -0.4152, 0.0017, -0.0405, -0.4152, 0.0039, -0.0405, -0.4483, -0.0, -0.0353, -0.4483, -0.0, -0.0331, -0.4462, -0.0, -0.0353, -0.4462, -0.0, -0.0331, -0.4483, -0.0, -0.0309, -0.4483, -0.0, -0.0288, -0.4462, -0.0, -0.0309, -0.4462, -0.0, -0.0288, -0.444, -0.0, -0.0353, -0.444, -0.0, -0.0331, -0.4418, -0.0, -0.0353, -0.4418, -0.0, -0.0331, -0.444, -0.0, -0.0309, -0.444, -0.0, -0.0288, -0.4418, -0.0, -0.0309, -0.4418, -0.0, -0.0288, -0.4486, -0.0, -0.0255, -0.4477, -0.0, -0.0253, -0.4462, -0.0, -0.0266, -0.4463, -0.0, -0.0249, -0.4489, 0.0017, -0.0256, -0.4479, 0.0017, -0.0253, -0.4489, 0.0039, -0.0256, -0.4479, 0.0039, -0.0253, -0.4462, 0.0017, -0.0248, -0.4462, 0.0039, -0.0248,
-0.444, -0.0, -0.0266, -0.4442, -0.0, -0.0242, -0.4418, -0.0, -0.0266, -0.4422, -0.0, -0.0234, -0.4412, -0.0, -0.0231, -0.444, 0.0017, -0.0241, -0.444, 0.0039, -0.0241, -0.4424, 0.0017, -0.0235, -0.4424, 0.0039, -0.0235, -0.4413, 0.0017, -0.0231, -0.4413, 0.0039, -0.0231, -0.4397, -0.0, -0.0353, -0.4397, -0.0, -0.0331, -0.4375, -0.0, -0.0353, -0.4375, -0.0, -0.0331, -0.4397, -0.0, -0.0309, -0.4397, -0.0, -0.0288, -0.4375, -0.0, -0.0309, -0.4375, -0.0, -0.0288, -0.4353, -0.0, -0.0353, -0.4353, -0.0, -0.0331, -0.4332, -0.0, -0.0353, -0.4332, -0.0, -0.0331, -0.4353, -0.0, -0.0309, -0.4353, -0.0, -0.0288, -0.4332, -0.0, -0.0309, -0.4332, -0.0, -0.0288, -0.4397, -0.0, -0.0266, -0.4397, -0.0, -0.0244, -0.4375, -0.0, -0.0266,
-0.4375, -0.0, -0.0244, -0.4398, -0.0, -0.0225, -0.4375, -0.0, -0.0216, -0.4364, -0.0, -0.0212, -0.4397, 0.0017, -0.0225, -0.4397, 0.0039, -0.0225, -0.4375, 0.0017, -0.0216, -0.4364, 0.0017, -0.0212, -0.4375, 0.0039, -0.0216, -0.4364, 0.0039, -0.0212, -0.4353, -0.0, -0.0266, -0.4353, -0.0, -0.0244, -0.4332, -0.0, -0.0266, -0.4332, -0.0, -0.0244, -0.4353, -0.0, -0.0222, -0.4354, -0.0, -0.0207, -0.4332, -0.0, -0.0222, -0.4334, -0.0, -0.0199, -0.4353, 0.0017, -0.0207, -0.4353, 0.0039, -0.0207, -0.4332, 0.0017, -0.0197, -0.4332, 0.0039, -0.0197, -0.431, -0.0, -0.0353, -0.431, -0.0, -0.0331, -0.4288, -0.0, -0.0353, -0.4288, -0.0, -0.0331, -0.431, -0.0, -0.0309, -0.431, -0.0, -0.0288, -0.4288, -0.0, -0.0309, -0.4288, -0.0, -0.0288,
-0.4266, -0.0, -0.0353, -0.4266, -0.0, -0.0331, -0.4241, -0.0, -0.0362, -0.4245, -0.0, -0.0331, -0.4266, -0.0, -0.0309, -0.4266, -0.0, -0.0288, -0.4245, -0.0, -0.0309, -0.4245, -0.0, -0.0288, -0.4239, 0.0017, -0.0361, -0.4239, 0.0039, -0.0361, -0.431, -0.0, -0.0266, -0.431, -0.0, -0.0244, -0.4288, -0.0, -0.0266, -0.4288, -0.0, -0.0244, -0.431, -0.0, -0.0222, -0.4315, -0.0, -0.019, -0.4288, -0.0, -0.0222, -0.4288, -0.0, -0.0201, -0.4318, 0.0017, -0.0191, -0.4318, 0.0039, -0.0191, -0.4266, -0.0, -0.0266, -0.4266, -0.0, -0.0244, -0.4245, -0.0, -0.0266, -0.4245, -0.0, -0.0244, -0.4266, -0.0, -0.0222, -0.4266, -0.0, -0.0201, -0.4245, -0.0, -0.0222, -0.4245, -0.0, -0.0201, -0.4221, -0.0, -0.0353, -0.4223, -0.0, -0.0331,
-0.4206, -0.0, -0.0346, -0.4196, -0.0, -0.0342, -0.4223, -0.0, -0.0309, -0.4223, -0.0, -0.0288, -0.4201, -0.0, -0.0309, -0.4201, -0.0, -0.0288, -0.4223, 0.0017, -0.0354, -0.4223, 0.0039, -0.0354, -0.4204, 0.0017, -0.0345, -0.4194, 0.0017, -0.034, -0.4204, 0.0039, -0.0345, -0.4194, 0.0039, -0.034, -0.4177, -0.0, -0.0332, -0.4163, -0.0, -0.0325, -0.418, -0.0, -0.0309, -0.418, -0.0, -0.0288, -0.4153, -0.0, -0.0319, -0.4158, -0.0, -0.0288, -0.4179, 0.0017, -0.0333, -0.4179, 0.0039, -0.0333, -0.4162, 0.0017, -0.0324, -0.4162, 0.0039, -0.0324, -0.4151, 0.0017, -0.0318, -0.4151, 0.0039, -0.0318, -0.4223, -0.0, -0.0266, -0.4223, -0.0, -0.0244, -0.4201, -0.0, -0.0266, -0.4201, -0.0, -0.0244, -0.4223, -0.0, -0.0222, -0.4223, -0.0, -0.0201,
-0.4201, -0.0, -0.0222, -0.4201, -0.0, -0.0201, -0.418, -0.0, -0.0266, -0.418, -0.0, -0.0244, -0.4158, -0.0, -0.0266, -0.4158, -0.0, -0.0244, -0.418, -0.0, -0.0222, -0.418, -0.0, -0.0201, -0.4158, -0.0, -0.0222, -0.4158, -0.0, -0.0201, -0.4305, -0.0, -0.0185, -0.429, -0.0, -0.0179, -0.4307, 0.0017, -0.0186, -0.4307, 0.0039, -0.0186, -0.4288, 0.0017, -0.0178, -0.4288, 0.0039, -0.0178, -0.427, -0.0, -0.017, -0.426, -0.0, -0.0165, -0.4245, -0.0, -0.0179, -0.4247, -0.0, -0.0158, -0.4272, 0.0017, -0.017, -0.4262, 0.0017, -0.0165, -0.4272, 0.0039, -0.017, -0.4262, 0.0039, -0.0165, -0.4245, 0.0017, -0.0157, -0.4245, 0.0039, -0.0157, -0.4257, -0.0, -0.0018, -0.4246, -0.0, -0.0023, -0.4258, 0.0017, -0.0017, -0.4258, 0.0039, -0.0017,
-0.4245, 0.0017, -0.0023, -0.4245, 0.0039, -0.0023, -0.4223, -0.0, -0.0179, -0.4228, -0.0, -0.0148, -0.4201, -0.0, -0.0179, -0.4201, -0.0, -0.0157, -0.4217, -0.0, -0.0142, -0.4204, -0.0, -0.0135, -0.423, 0.0017, -0.0149, -0.423, 0.0039, -0.0149, -0.4219, 0.0017, -0.0143, -0.4219, 0.0039, -0.0143, -0.4201, 0.0017, -0.0134, -0.4201, 0.0039, -0.0134, -0.418, -0.0, -0.0179, -0.418, -0.0, -0.0157, -0.4158, -0.0, -0.0179, -0.4158, -0.0, -0.0157, -0.4185, -0.0, -0.0126, -0.4175, -0.0, -0.012, -0.4158, -0.0, -0.0136, -0.4161, -0.0, -0.0113, -0.4187, 0.0017, -0.0127, -0.4176, 0.0017, -0.0121, -0.4187, 0.0039, -0.0127, -0.4176, 0.0039, -0.0121, -0.4158, 0.0017, -0.0111, -0.4158, 0.0039, -0.0111, -0.4224, -0.0, -0.0029, -0.4201, -0.0, -0.0032,
-0.4223, 0.0017, -0.003, -0.4223, 0.0039, -0.003, -0.4201, 0.0017, -0.0032, -0.4201, 0.0039, -0.0032, -0.4179, -0.0, -0.003, -0.4156, -0.0, -0.0025, -0.4179, 0.0017, -0.003, -0.4179, 0.0039, -0.003, -0.4158, 0.0017, -0.0026, -0.4158, 0.0039, -0.0026, -0.3976, -0.0, -0.1059, -0.3977, 0.0017, -0.1059, -0.3977, 0.0039, -0.1059, -0.3963, -0.0, -0.1061, -0.3942, -0.0, -0.1066, -0.3963, 0.0017, -0.1061, -0.3963, 0.0039, -0.1061, -0.3941, 0.0017, -0.1066, -0.3941, 0.0039, -0.1066, -0.392, -0.0, -0.1069, -0.3898, -0.0, -0.1071, -0.3919, 0.0017, -0.1069, -0.3919, 0.0039, -0.1069, -0.3898, 0.0017, -0.1071, -0.3898, 0.0039, -0.1071, -0.3876, -0.0, -0.1072, -0.3854, -0.0, -0.1072, -0.3876, 0.0017, -0.1072, -0.3876, 0.0039, -0.1072, -0.3854, 0.0017, -0.1072,
-0.3854, 0.0039, -0.1072, -0.3832, -0.0, -0.1072, -0.3811, -0.0, -0.1072, -0.3832, 0.0017, -0.1072, -0.3832, 0.0039, -0.1072, -0.3811, 0.0017, -0.1072, -0.3811, 0.0039, -0.1072, -0.4129, -0.0, -0.1017, -0.4116, -0.0, -0.1022, -0.4139, -0.0, -0.1013, -0.4136, -0.0, -0.0982, -0.4115, -0.0, -0.1004, -0.4115, -0.0, -0.0982, -0.413, 0.0017, -0.1016, -0.413, 0.0039, -0.1016, -0.4115, 0.0017, -0.1022, -0.4115, 0.0039, -0.1022, -0.4141, 0.0017, -0.1012, -0.4141, 0.0039, -0.1012, -0.4095, -0.0, -0.1029, -0.4065, -0.0, -0.1039, -0.4075, -0.0, -0.1036, -0.4093, -0.0, -0.1004, -0.4093, -0.0, -0.0982, -0.4071, -0.0, -0.1004, -0.4071, -0.0, -0.0982, -0.4093, 0.0017, -0.103, -0.4093, 0.0039, -0.103, -0.4067, 0.0017, -0.1038, -0.4078, 0.0017, -0.1035,
-0.4067, 0.0039, -0.1038, -0.4078, 0.0039, -0.1035, -0.4136, -0.0, -0.096, -0.4136, -0.0, -0.0939, -0.4115, -0.0, -0.096, -0.4115, -0.0, -0.0939, -0.4136, -0.0, -0.0917, -0.4136, -0.0, -0.0895, -0.4115, -0.0, -0.0917, -0.4115, -0.0, -0.0895, -0.4093, -0.0, -0.096, -0.4093, -0.0, -0.0939, -0.4071, -0.0, -0.096, -0.4071, -0.0, -0.0939, -0.4093, -0.0, -0.0917, -0.4093, -0.0, -0.0895, -0.4071, -0.0, -0.0917, -0.4071, -0.0, -0.0895, -0.405, -0.0, -0.1043, -0.4049, -0.0, -0.1025, -0.4029, -0.0, -0.1048, -0.4028, -0.0, -0.1025, -0.4049, -0.0, -0.1004, -0.4049, -0.0, -0.0982, -0.4028, -0.0, -0.1004, -0.4028, -0.0, -0.0982, -0.4049, 0.0017, -0.1043, -0.4049, 0.0039, -0.1043, -0.4028, 0.0017, -0.1048, -0.4028, 0.0039, -0.1048,
-0.4008, -0.0, -0.1053, -0.4006, -0.0, -0.1025, -0.3985, -0.0, -0.1057, -0.3984, -0.0, -0.1025, -0.4006, -0.0, -0.1004, -0.4006, -0.0, -0.0982, -0.3984, -0.0, -0.1004, -0.3984, -0.0, -0.0982, -0.4006, 0.0017, -0.1053, -0.4006, 0.0039, -0.1053, -0.3988, 0.0017, -0.1056, -0.3988, 0.0039, -0.1056, -0.4049, -0.0, -0.096, -0.4049, -0.0, -0.0939, -0.4028, -0.0, -0.096, -0.4028, -0.0, -0.0939, -0.4049, -0.0, -0.0917, -0.4049, -0.0, -0.0895, -0.4028, -0.0, -0.0917, -0.4026, -0.0, -0.0886, -0.4024, 0.0017, -0.0887, -0.4024, 0.0039, -0.0887, -0.4006, -0.0, -0.096, -0.4006, -0.0, -0.0939, -0.3984, -0.0, -0.096, -0.3984, -0.0, -0.0939, -0.4006, -0.0, -0.0917, -0.4004, -0.0, -0.0894, -0.3984, -0.0, -0.0917, -0.3984, -0.0, -0.09,
-0.4006, 0.0017, -0.0893, -0.4006, 0.0039, -0.0893, -0.3984, 0.0017, -0.0899, -0.3984, 0.0039, -0.0899, -0.4136, -0.0, -0.0874, -0.4131, -0.0, -0.0842, -0.4115, -0.0, -0.0874, -0.4112, -0.0, -0.0851, -0.4141, -0.0, -0.0837, -0.4129, 0.0017, -0.0843, -0.4129, 0.0039, -0.0843, -0.4115, 0.0017, -0.085, -0.4115, 0.0039, -0.085, -0.414, 0.0017, -0.0838, -0.414, 0.0039, -0.0838, -0.4087, -0.0, -0.0862, -0.4097, -0.0, -0.0858, -0.4068, -0.0, -0.087, -0.4084, 0.0017, -0.0863, -0.4095, 0.0017, -0.0859, -0.4084, 0.0039, -0.0863, -0.4095, 0.0039, -0.0859, -0.4071, 0.0017, -0.0869, -0.4071, 0.0039, -0.0869, -0.4048, -0.0, -0.0878, -0.4036, -0.0, -0.0883, -0.4049, 0.0017, -0.0877, -0.4049, 0.0039, -0.0877, -0.4035, 0.0017, -0.0883, -0.4035, 0.0039, -0.0883,
-0.3963, -0.0, -0.1047, -0.3963, -0.0, -0.1025, -0.3941, -0.0, -0.1047, -0.3941, -0.0, -0.1025, -0.3963, -0.0, -0.1004, -0.3963, -0.0, -0.0982, -0.3941, -0.0, -0.1004, -0.3941, -0.0, -0.0982, -0.3919, -0.0, -0.1047, -0.3919, -0.0, -0.1025, -0.3897, -0.0, -0.1047, -0.3897, -0.0, -0.1025, -0.3919, -0.0, -0.1004, -0.3919, -0.0, -0.0982, -0.3897, -0.0, -0.1004, -0.3897, -0.0, -0.0982, -0.3963, -0.0, -0.096, -0.3963, -0.0, -0.0939, -0.3941, -0.0, -0.096, -0.3941, -0.0, -0.0939, -0.3959, -0.0, -0.0905, -0.3967, -0.0, -0.0903, -0.3939, -0.0, -0.091, -0.3954, 0.0017, -0.0907, -0.3965, 0.0017, -0.0904, -0.3954, 0.0039, -0.0907, -0.3965, 0.0039, -0.0904, -0.3941, 0.0017, -0.091, -0.3941, 0.0039, -0.091, -0.3919, -0.0, -0.096,
-0.3919, -0.0, -0.0939, -0.3897, -0.0, -0.096, -0.3897, -0.0, -0.0939, -0.3918, -0.0, -0.0916, -0.3897, -0.0, -0.0921, -0.3919, 0.0017, -0.0915, -0.3919, 0.0039, -0.0915, -0.3898, 0.0017, -0.0921, -0.3898, 0.0039, -0.0921, -0.3876, -0.0, -0.1047, -0.3876, -0.0, -0.1025, -0.3854, -0.0, -0.1047, -0.3854, -0.0, -0.1025, -0.3876, -0.0, -0.1004, -0.3876, -0.0, -0.0982, -0.3854, -0.0, -0.1004, -0.3854, -0.0, -0.0982, -0.3832, -0.0, -0.1047, -0.3832, -0.0, -0.1025, -0.3811, -0.0, -0.1047, -0.3811, -0.0, -0.1025, -0.3832, -0.0, -0.1004, -0.3832, -0.0, -0.0982, -0.3811, -0.0, -0.1004, -0.3811, -0.0, -0.0982, -0.3876, -0.0, -0.096, -0.3876, -0.0, -0.0939, -0.3854, -0.0, -0.096, -0.3854, -0.0, -0.0939, -0.3876, -0.0, -0.0924,
-0.3854, -0.0, -0.0925, -0.3876, 0.0017, -0.0924, -0.3876, 0.0039, -0.0924, -0.3854, 0.0017, -0.0925, -0.3854, 0.0039, -0.0925, -0.3832, -0.0, -0.096, -0.3831, -0.0, -0.0928, -0.3811, -0.0, -0.096, -0.381, -0.0, -0.0931, -0.3839, -0.0, -0.0927, -0.3827, 0.0017, -0.0928, -0.3827, 0.0039, -0.0928, -0.3811, 0.0017, -0.0931, -0.3811, 0.0039, -0.0931, -0.3838, 0.0017, -0.0927, -0.3838, 0.0039, -0.0927, -0.3789, -0.0, -0.1072, -0.3767, -0.0, -0.1072, -0.3789, 0.0017, -0.1072, -0.3789, 0.0039, -0.1072, -0.3767, 0.0017, -0.1072, -0.3767, 0.0039, -0.1072, -0.3746, -0.0, -0.1072, -0.3724, -0.0, -0.1072, -0.3746, 0.0017, -0.1072, -0.3746, 0.0039, -0.1072, -0.3724, 0.0017, -0.1072, -0.3724, 0.0039, -0.1072, -0.3702, -0.0, -0.1072, -0.368, -0.0, -0.1071,
-0.3702, 0.0017, -0.1072, -0.3702, 0.0039, -0.1072, -0.368, 0.0017, -0.1072, -0.368, 0.0039, -0.1072, -0.3658, -0.0, -0.107, -0.3636, -0.0, -0.1067, -0.3659, 0.0017, -0.107, -0.3659, 0.0039, -0.107, -0.3637, 0.0017, -0.1067, -0.3637, 0.0039, -0.1067, -0.3615, -0.0, -0.1062, -0.3601, -0.0, -0.1059, -0.3615, 0.0017, -0.1062, -0.3615, 0.0039, -0.1062, -0.36, 0.0017, -0.1059, -0.36, 0.0039, -0.1059, -0.3789, -0.0, -0.1047, -0.3789, -0.0, -0.1025, -0.3767, -0.0, -0.1047, -0.3767, -0.0, -0.1025, -0.3789, -0.0, -0.1004, -0.3789, -0.0, -0.0982, -0.3767, -0.0, -0.1004, -0.3767, -0.0, -0.0982, -0.3746, -0.0, -0.1047, -0.3746, -0.0, -0.1025, -0.3724, -0.0, -0.1047, -0.3724, -0.0, -0.1025, -0.3746, -0.0, -0.1004, -0.3746, -0.0, -0.0982,
-0.3724, -0.0, -0.1004, -0.3724, -0.0, -0.0982, -0.3789, -0.0, -0.096, -0.3788, -0.0, -0.0933, -0.3767, -0.0, -0.096, -0.3767, -0.0, -0.0934, -0.3789, 0.0017, -0.0933, -0.3789, 0.0039, -0.0933, -0.3767, 0.0017, -0.0934, -0.3767, 0.0039, -0.0934, -0.3746, -0.0, -0.096, -0.3746, -0.0, -0.0934, -0.3724, -0.0, -0.096, -0.3724, -0.0, -0.0934, -0.3746, 0.0017, -0.0934, -0.3746, 0.0039, -0.0934, -0.3724, 0.0017, -0.0934, -0.3724, 0.0039, -0.0934, -0.3702, -0.0, -0.1047, -0.3702, -0.0, -0.1025, -0.368, -0.0, -0.1047, -0.368, -0.0, -0.1025, -0.3702, -0.0, -0.1004, -0.3702, -0.0, -0.0982, -0.368, -0.0, -0.1004, -0.368, -0.0, -0.0982, -0.3659, -0.0, -0.1047, -0.3659, -0.0, -0.1025, -0.3637, -0.0, -0.1047, -0.3637, -0.0, -0.1025,
-0.3659, -0.0, -0.1004, -0.3659, -0.0, -0.0982, -0.3637, -0.0, -0.1004, -0.3637, -0.0, -0.0982, -0.3702, -0.0, -0.096, -0.3703, -0.0, -0.0934, -0.368, -0.0, -0.096, -0.3682, -0.0, -0.0931, -0.3702, 0.0017, -0.0934, -0.3702, 0.0039, -0.0934, -0.368, 0.0017, -0.0931, -0.368, 0.0039, -0.0931, -0.3659, -0.0, -0.096, -0.366, -0.0, -0.0928, -0.3637, -0.0, -0.096, -0.3637, -0.0, -0.0939, -0.3651, -0.0, -0.0927, -0.3637, -0.0, -0.0925, -0.3664, 0.0017, -0.0929, -0.3664, 0.0039, -0.0929, -0.3653, 0.0017, -0.0927, -0.3653, 0.0039, -0.0927, -0.3637, 0.0017, -0.0925, -0.3637, 0.0039, -0.0925, -0.3615, -0.0, -0.1047, -0.3615, -0.0, -0.1025, -0.3592, -0.0, -0.1057, -0.3594, -0.0, -0.1025, -0.3615, -0.0, -0.1004, -0.3615, -0.0, -0.0982,
-0.3594, -0.0, -0.1004, -0.3594, -0.0, -0.0982, -0.3589, 0.0017, -0.1057, -0.3589, 0.0039, -0.1057, -0.357, -0.0, -0.1052, -0.3572, -0.0, -0.1025, -0.3549, -0.0, -0.1048, -0.355, -0.0, -0.1025, -0.3572, -0.0, -0.1004, -0.3572, -0.0, -0.0982, -0.355, -0.0, -0.1004, -0.355, -0.0, -0.0982, -0.3572, 0.0017, -0.1053, -0.3572, 0.0039, -0.1053, -0.355, 0.0017, -0.1048, -0.355, 0.0039, -0.1048, -0.3615, -0.0, -0.096, -0.3615, -0.0, -0.0939, -0.3594, -0.0, -0.096, -0.3594, -0.0, -0.0939, -0.3616, -0.0, -0.0924, -0.3594, -0.0, -0.0922, -0.3615, 0.0017, -0.0924, -0.3615, 0.0039, -0.0924, -0.3594, 0.0017, -0.0922, -0.3594, 0.0039, -0.0922, -0.3572, -0.0, -0.096, -0.3572, -0.0, -0.0939, -0.355, -0.0, -0.096, -0.355, -0.0, -0.0939,
-0.3573, -0.0, -0.0918, -0.3552, -0.0, -0.0913, -0.3572, 0.0017, -0.0918, -0.3572, 0.0039, -0.0918, -0.355, 0.0017, -0.0912, -0.355, 0.0039, -0.0912, -0.3528, -0.0, -0.1043, -0.3529, -0.0, -0.1025, -0.3512, -0.0, -0.1039, -0.3503, -0.0, -0.1037, -0.3529, -0.0, -0.1004, -0.3529, -0.0, -0.0982, -0.3507, -0.0, -0.1004, -0.3507, -0.0, -0.0982, -0.3529, 0.0017, -0.1043, -0.3529, 0.0039, -0.1043, -0.3509, 0.0017, -0.1039, -0.3498, 0.0017, -0.1036, -0.3509, 0.0039, -0.1039, -0.3498, 0.0039, -0.1036, -0.3483, -0.0, -0.1031, -0.3462, -0.0, -0.1024, -0.3485, -0.0, -0.1004, -0.3485, -0.0, -0.0982, -0.3463, -0.0, -0.1004, -0.3463, -0.0, -0.0982, -0.3485, 0.0017, -0.1032, -0.3485, 0.0039, -0.1032, -0.3463, 0.0017, -0.1025, -0.3463, 0.0039, -0.1025,
-0.3529, -0.0, -0.096, -0.3529, -0.0, -0.0939, -0.3507, -0.0, -0.096, -0.3507, -0.0, -0.0939, -0.3531, -0.0, -0.0907, -0.3521, -0.0, -0.0904, -0.3507, -0.0, -0.0917, -0.3508, -0.0, -0.0901, -0.3534, 0.0017, -0.0908, -0.3523, 0.0017, -0.0905, -0.3534, 0.0039, -0.0908, -0.3523, 0.0039, -0.0905, -0.3507, 0.0017, -0.0901, -0.3507, 0.0039, -0.0901, -0.3485, -0.0, -0.096, -0.3485, -0.0, -0.0939, -0.3463, -0.0, -0.096, -0.3463, -0.0, -0.0939, -0.3485, -0.0, -0.0917, -0.3487, -0.0, -0.0895, -0.3463, -0.0, -0.0917, -0.3464, -0.0, -0.0888, -0.3485, 0.0017, -0.0895, -0.3485, 0.0039, -0.0895, -0.3465, 0.0017, -0.0888, -0.3465, 0.0039, -0.0888, -0.3454, -0.0, -0.0884, -0.3454, 0.0017, -0.0884, -0.3454, 0.0039, -0.0884, -0.4134, -0.0, -0.0633,
-0.412, -0.0, -0.0627, -0.4136, 0.0017, -0.0634, -0.4136, 0.0039, -0.0634, -0.4119, 0.0017, -0.0627, -0.4119, 0.0039, -0.0627, -0.4136, -0.0, -0.0613, -0.4136, -0.0, -0.0591, -0.411, -0.0, -0.0623, -0.4115, -0.0, -0.0591, -0.4136, -0.0, -0.057, -0.4136, -0.0, -0.0548, -0.4115, -0.0, -0.057, -0.4115, -0.0, -0.0548, -0.4108, 0.0017, -0.0622, -0.4108, 0.0039, -0.0622, -0.409, -0.0, -0.0614, -0.4093, -0.0, -0.0591, -0.4076, -0.0, -0.0607, -0.4065, -0.0, -0.0602, -0.4093, -0.0, -0.057, -0.4093, -0.0, -0.0548, -0.4071, -0.0, -0.057, -0.4071, -0.0, -0.0548, -0.4093, 0.0017, -0.0615, -0.4093, 0.0039, -0.0615, -0.4074, 0.0017, -0.0606, -0.4063, 0.0017, -0.0601, -0.4074, 0.0039, -0.0606, -0.4063, 0.0039, -0.0601, -0.4047, -0.0, -0.0592,
-0.4033, -0.0, -0.0585, -0.4049, -0.0, -0.057, -0.4049, -0.0, -0.0548, -0.4023, -0.0, -0.0579, -0.4028, -0.0, -0.0548, -0.4049, 0.0017, -0.0594, -0.4049, 0.0039, -0.0594, -0.4032, 0.0017, -0.0584, -0.4032, 0.0039, -0.0584, -0.4021, 0.0017, -0.0578, -0.4021, 0.0039, -0.0578, -0.4004, -0.0, -0.0568, -0.4006, -0.0, -0.0548, -0.3992, -0.0, -0.0561, -0.3981, -0.0, -0.0555, -0.4006, 0.0017, -0.0569, -0.4006, 0.0039, -0.0569, -0.3991, 0.0017, -0.0561, -0.398, 0.0017, -0.0555, -0.3991, 0.0039, -0.0561, -0.398, 0.0039, -0.0555, -0.4136, -0.0, -0.0526, -0.4136, -0.0, -0.0505, -0.4115, -0.0, -0.0526, -0.4115, -0.0, -0.0505, -0.4136, -0.0, -0.0483, -0.4136, -0.0, -0.0461, -0.4115, -0.0, -0.0483, -0.4115, -0.0, -0.0461, -0.4093, -0.0, -0.0526,
-0.4093, -0.0, -0.0505, -0.4071, -0.0, -0.0526, -0.4071, -0.0, -0.0505, -0.4093, -0.0, -0.0483, -0.4093, -0.0, -0.0461, -0.4071, -0.0, -0.0483, -0.4071, -0.0, -0.0461, -0.4136, -0.0, -0.0439, -0.4136, -0.0, -0.0418, -0.4115, -0.0, -0.0439, -0.4115, -0.0, -0.0418, -0.4138, -0.0, -0.0398, -0.4118, -0.0, -0.0388, -0.4107, -0.0, -0.0383, -0.4136, 0.0017, -0.0397, -0.4136, 0.0039, -0.0397, -0.4119, 0.0017, -0.0389, -0.4108, 0.0017, -0.0383, -0.4119, 0.0039, -0.0389, -0.4108, 0.0039, -0.0383, -0.4093, -0.0, -0.0439, -0.4093, -0.0, -0.0418, -0.4071, -0.0, -0.0439, -0.4071, -0.0, -0.0418, -0.4093, -0.0, -0.0396, -0.4095, -0.0, -0.0376, -0.4071, -0.0, -0.0396, -0.4076, -0.0, -0.0366, -0.4093, 0.0017, -0.0375, -0.4093, 0.0039, -0.0375,
-0.4077, 0.0017, -0.0366, -0.4077, 0.0039, -0.0366, -0.4049, -0.0, -0.0526, -0.4049, -0.0, -0.0505, -0.4028, -0.0, -0.0526, -0.4028, -0.0, -0.0505, -0.4049, -0.0, -0.0483, -0.4049, -0.0, -0.0461, -0.4028, -0.0, -0.0483, -0.4028, -0.0, -0.0461, -0.4006, -0.0, -0.0526, -0.4006, -0.0, -0.0505, -0.3984, -0.0, -0.0526, -0.3984, -0.0, -0.0505, -0.4006, -0.0, -0.0483, -0.4006, -0.0, -0.0461, -0.3984, -0.0, -0.0483, -0.3984, -0.0, -0.0461, -0.4049, -0.0, -0.0439, -0.4049, -0.0, -0.0418, -0.4028, -0.0, -0.0439, -0.4028, -0.0, -0.0418, -0.4049, -0.0, -0.0396, -0.4049, -0.0, -0.0374, -0.4028, -0.0, -0.0396, -0.4028, -0.0, -0.0374, -0.4006, -0.0, -0.0439, -0.4006, -0.0, -0.0418, -0.3984, -0.0, -0.0439, -0.3984, -0.0, -0.0418,
-0.4006, -0.0, -0.0396, -0.4006, -0.0, -0.0374, -0.3984, -0.0, -0.0396, -0.3984, -0.0, -0.0374, -0.3961, -0.0, -0.0544, -0.3951, -0.0, -0.0538, -0.3963, 0.0017, -0.0545, -0.3963, 0.0039, -0.0545, -0.395, 0.0017, -0.0538, -0.395, 0.0039, -0.0538, -0.3963, -0.0, -0.0526, -0.3963, -0.0, -0.0505, -0.3939, -0.0, -0.0531, -0.3941, -0.0, -0.0505, -0.3963, -0.0, -0.0483, -0.3963, -0.0, -0.0461, -0.3941, -0.0, -0.0483, -0.3941, -0.0, -0.0461, -0.394, 0.0017, -0.0532, -0.394, 0.0039, -0.0532, -0.3923, -0.0, -0.0521, -0.3912, -0.0, -0.0515, -0.3895, -0.0, -0.0504, -0.3919, -0.0, -0.0483, -0.3919, -0.0, -0.0461, -0.3897, -0.0, -0.0483, -0.3897, -0.0, -0.0461, -0.3921, 0.0017, -0.0521, -0.3911, 0.0017, -0.0514, -0.3921, 0.0039, -0.0521,
-0.3911, 0.0039, -0.0514, -0.3897, 0.0017, -0.0506, -0.3897, 0.0039, -0.0506, -0.3963, -0.0, -0.0439, -0.3963, -0.0, -0.0418, -0.3941, -0.0, -0.0439, -0.3941, -0.0, -0.0418, -0.3963, -0.0, -0.0396, -0.3963, -0.0, -0.0374, -0.3941, -0.0, -0.0396, -0.3941, -0.0, -0.0374, -0.3919, -0.0, -0.0439, -0.3919, -0.0, -0.0418, -0.3897, -0.0, -0.0439, -0.3897, -0.0, -0.0418, -0.3919, -0.0, -0.0396, -0.3919, -0.0, -0.0374, -0.3897, -0.0, -0.0396, -0.3897, -0.0, -0.0374, -0.3883, -0.0, -0.0497, -0.3872, -0.0, -0.049, -0.3876, -0.0, -0.0461, -0.3857, -0.0, -0.0479, -0.3846, -0.0, -0.0471, -0.3883, 0.0017, -0.0496, -0.3883, 0.0039, -0.0496, -0.3872, 0.0017, -0.0489, -0.3872, 0.0039, -0.0489, -0.3856, 0.0017, -0.0478, -0.3845, 0.0017, -0.0471,
-0.3856, 0.0039, -0.0478, -0.3845, 0.0039, -0.0471, -0.383, -0.0, -0.0459, -0.382, -0.0, -0.0451, -0.3832, 0.0017, -0.0461, -0.3832, 0.0039, -0.0461, -0.382, 0.0017, -0.0451, -0.382, 0.0039, -0.0451, -0.3876, -0.0, -0.0439, -0.3876, -0.0, -0.0418, -0.3854, -0.0, -0.0439, -0.3854, -0.0, -0.0418, -0.3876, -0.0, -0.0396, -0.3876, -0.0, -0.0374, -0.3854, -0.0, -0.0396, -0.3854, -0.0, -0.0374, -0.3832, -0.0, -0.0439, -0.3832, -0.0, -0.0418, -0.3809, -0.0, -0.0443, -0.3811, -0.0, -0.0418, -0.3832, -0.0, -0.0396, -0.3832, -0.0, -0.0374, -0.3811, -0.0, -0.0396, -0.3811, -0.0, -0.0374, -0.3809, 0.0017, -0.0443, -0.3809, 0.0039, -0.0443, -0.4134, -0.0, -0.0308, -0.4136, -0.0, -0.0288, -0.4122, -0.0, -0.0301, -0.4111, -0.0, -0.0295,
-0.4136, 0.0017, -0.0309, -0.4136, 0.0039, -0.0309, -0.4121, 0.0017, -0.0301, -0.411, 0.0017, -0.0295, -0.4121, 0.0039, -0.0301, -0.411, 0.0039, -0.0295, -0.4065, -0.0, -0.036, -0.4091, -0.0, -0.0283, -0.4081, -0.0, -0.0277, -0.4066, 0.0017, -0.036, -0.4066, 0.0039, -0.036, -0.4093, 0.0017, -0.0284, -0.4093, 0.0039, -0.0284, -0.4081, 0.0017, -0.0277, -0.4081, 0.0039, -0.0277, -0.4136, -0.0, -0.0266, -0.4136, -0.0, -0.0244, -0.4115, -0.0, -0.0266, -0.4115, -0.0, -0.0244, -0.4136, -0.0, -0.0222, -0.4136, -0.0, -0.0201, -0.4115, -0.0, -0.0222, -0.4115, -0.0, -0.0201, -0.4093, -0.0, -0.0266, -0.4093, -0.0, -0.0244, -0.407, -0.0, -0.0271, -0.4071, -0.0, -0.0244, -0.4093, -0.0, -0.0222, -0.4093, -0.0, -0.0201, -0.4071, -0.0, -0.0222,
-0.4071, -0.0, -0.0201, -0.407, 0.0017, -0.0271, -0.407, 0.0039, -0.0271, -0.4052, -0.0, -0.0352, -0.4035, -0.0, -0.0342, -0.4025, -0.0, -0.0335, -0.4049, 0.0017, -0.0351, -0.4049, 0.0039, -0.0351, -0.4037, 0.0017, -0.0343, -0.4026, 0.0017, -0.0336, -0.4037, 0.0039, -0.0343, -0.4026, 0.0039, -0.0336, -0.4006, -0.0, -0.0353, -0.4008, -0.0, -0.0325, -0.3984, -0.0, -0.0353, -0.3984, -0.0, -0.0331, -0.3997, -0.0, -0.0318, -0.3987, -0.0, -0.0312, -0.4009, 0.0017, -0.0325, -0.4009, 0.0039, -0.0325, -0.3998, 0.0017, -0.0319, -0.3998, 0.0039, -0.0319, -0.3984, 0.0017, -0.031, -0.3984, 0.0039, -0.031, -0.4054, -0.0, -0.0261, -0.4043, -0.0, -0.0254, -0.4025, -0.0, -0.0242, -0.4049, -0.0, -0.0222, -0.4049, -0.0, -0.0201, -0.4028, -0.0, -0.0222,
-0.4028, -0.0, -0.0201, -0.4053, 0.0017, -0.026, -0.4042, 0.0017, -0.0253, -0.4053, 0.0039, -0.026, -0.4042, 0.0039, -0.0253, -0.4028, 0.0017, -0.0244, -0.4028, 0.0039, -0.0244, -0.4015, -0.0, -0.0235, -0.4004, -0.0, -0.0228, -0.4006, -0.0, -0.0201, -0.3988, -0.0, -0.0218, -0.3978, -0.0, -0.0211, -0.4014, 0.0017, -0.0235, -0.4014, 0.0039, -0.0235, -0.4003, 0.0017, -0.0228, -0.4003, 0.0039, -0.0228, -0.3987, 0.0017, -0.0217, -0.3976, 0.0017, -0.021, -0.3987, 0.0039, -0.0217, -0.3976, 0.0039, -0.021, -0.4136, -0.0, -0.0179, -0.4136, -0.0, -0.0157, -0.4115, -0.0, -0.0179, -0.4115, -0.0, -0.0157, -0.4136, -0.0, -0.0136, -0.4143, -0.0, -0.0103, -0.4115, -0.0, -0.0136, -0.4115, -0.0, -0.0114, -0.4145, 0.0017, -0.0104, -0.4145, 0.0039, -0.0104,
-0.4093, -0.0, -0.0179, -0.4093, -0.0, -0.0157, -0.4071, -0.0, -0.0179, -0.4071, -0.0, -0.0157, -0.4093, -0.0, -0.0136, -0.4093, -0.0, -0.0114, -0.4071, -0.0, -0.0136, -0.4071, -0.0, -0.0114, -0.4133, -0.0, -0.0097, -0.4118, -0.0, -0.0089, -0.4142, -0.0, -0.002, -0.4134, 0.0017, -0.0098, -0.4134, 0.0039, -0.0098, -0.4115, 0.0017, -0.0087, -0.4115, 0.0039, -0.0087, -0.4141, 0.0017, -0.0019, -0.4141, 0.0039, -0.0019, -0.41, -0.0, -0.008, -0.4091, -0.0, -0.0075, -0.4071, -0.0, -0.0092, -0.4073, -0.0, -0.0065, -0.4062, -0.0, -0.0058, -0.4103, 0.0017, -0.0081, -0.4093, 0.0017, -0.0076, -0.4103, 0.0039, -0.0081, -0.4093, 0.0039, -0.0076, -0.4073, 0.0017, -0.0065, -0.4073, 0.0039, -0.0065, -0.4062, 0.0017, -0.0058, -0.4062, 0.0039, -0.0058,
-0.4049, -0.0, -0.0179, -0.4049, -0.0, -0.0157, -0.4028, -0.0, -0.0179, -0.4028, -0.0, -0.0157, -0.4049, -0.0, -0.0136, -0.4049, -0.0, -0.0114, -0.4028, -0.0, -0.0136, -0.4028, -0.0, -0.0114, -0.4006, -0.0, -0.0179, -0.4006, -0.0, -0.0157, -0.3984, -0.0, -0.0179, -0.3984, -0.0, -0.0157, -0.4006, -0.0, -0.0136, -0.4006, -0.0, -0.0114, -0.3984, -0.0, -0.0136, -0.3984, -0.0, -0.0114, -0.4049, -0.0, -0.0092, -0.4049, -0.0, -0.0071, -0.4028, -0.0, -0.0092, -0.4028, -0.0, -0.0071, -0.4051, -0.0, -0.0052, -0.4032, -0.0, -0.0041, -0.4021, -0.0, -0.0035, -0.4049, 0.0017, -0.0051, -0.4049, 0.0039, -0.0051, -0.4032, 0.0017, -0.0041, -0.4022, 0.0017, -0.0035, -0.4032, 0.0039, -0.0041, -0.4022, 0.0039, -0.0035, -0.4006, -0.0, -0.0092,
-0.4006, -0.0, -0.0071, -0.3984, -0.0, -0.0092, -0.3984, -0.0, -0.0071, -0.4006, -0.0, -0.0049, -0.4009, -0.0, -0.0028, -0.3984, -0.0, -0.0049, -0.399, -0.0, -0.0017, -0.4006, 0.0017, -0.0026, -0.4006, 0.0039, -0.0026, -0.3992, 0.0017, -0.0018, -0.3992, 0.0039, -0.0018, -0.3963, -0.0, -0.0353, -0.3963, -0.0, -0.0331, -0.3941, -0.0, -0.0353, -0.3941, -0.0, -0.0331, -0.3968, -0.0, -0.03, -0.3958, -0.0, -0.0293, -0.3941, -0.0, -0.0309, -0.3942, -0.0, -0.0284, -0.397, 0.0017, -0.0301, -0.3959, 0.0017, -0.0294, -0.397, 0.0039, -0.0301, -0.3959, 0.0039, -0.0294, -0.3941, 0.0017, -0.0283, -0.3941, 0.0039, -0.0283, -0.3919, -0.0, -0.0353, -0.3919, -0.0, -0.0331, -0.3897, -0.0, -0.0353, -0.3897, -0.0, -0.0331, -0.3919, -0.0, -0.0309,
-0.3919, -0.0, -0.0288, -0.3897, -0.0, -0.0309, -0.3897, -0.0, -0.0288, -0.393, -0.0, -0.0276, -0.396, -0.0, -0.0199, -0.395, -0.0, -0.0192, -0.393, 0.0017, -0.0277, -0.393, 0.0039, -0.0277, -0.3963, 0.0017, -0.0201, -0.3963, 0.0039, -0.0201, -0.3949, 0.0017, -0.0192, -0.3949, 0.0039, -0.0192, -0.3921, -0.0, -0.027, -0.3902, -0.0, -0.0258, -0.3891, -0.0, -0.0251, -0.3919, 0.0017, -0.0269, -0.3919, 0.0039, -0.0269, -0.3903, 0.0017, -0.0259, -0.3892, 0.0017, -0.0251, -0.3903, 0.0039, -0.0259, -0.3892, 0.0039, -0.0251, -0.3876, -0.0, -0.0353, -0.3876, -0.0, -0.0331, -0.3854, -0.0, -0.0353, -0.3854, -0.0, -0.0331, -0.3876, -0.0, -0.0309, -0.3876, -0.0, -0.0288, -0.3854, -0.0, -0.0309, -0.3854, -0.0, -0.0288, -0.3832, -0.0, -0.0353,
-0.3832, -0.0, -0.0331, -0.3811, -0.0, -0.0353, -0.3811, -0.0, -0.0331, -0.3832, -0.0, -0.0309, -0.3832, -0.0, -0.0288, -0.3811, -0.0, -0.0309, -0.3811, -0.0, -0.0288, -0.3876, -0.0, -0.0266, -0.3879, -0.0, -0.0243, -0.3854, -0.0, -0.0266, -0.3863, -0.0, -0.0232, -0.3853, -0.0, -0.0225, -0.3876, 0.0017, -0.0241, -0.3876, 0.0039, -0.0241, -0.3865, 0.0017, -0.0233, -0.3865, 0.0039, -0.0233, -0.3854, 0.0017, -0.0226, -0.3854, 0.0039, -0.0226, -0.3832, -0.0, -0.0266, -0.3832, -0.0, -0.0244, -0.3811, -0.0, -0.0266, -0.3811, -0.0, -0.0244, -0.3837, -0.0, -0.0215, -0.3827, -0.0, -0.0207, -0.3811, -0.0, -0.0222, -0.3812, -0.0, -0.0197, -0.3838, 0.0017, -0.0215, -0.3827, 0.0017, -0.0207, -0.3838, 0.0039, -0.0215, -0.3827, 0.0039, -0.0207,
-0.3812, 0.0017, -0.0197, -0.3812, 0.0039, -0.0197, -0.3963, -0.0, -0.0179, -0.3963, -0.0, -0.0157, -0.3938, -0.0, -0.0184, -0.3941, -0.0, -0.0157, -0.3963, -0.0, -0.0136, -0.3963, -0.0, -0.0114, -0.3941, -0.0, -0.0136, -0.3941, -0.0, -0.0114, -0.3938, 0.0017, -0.0184, -0.3938, 0.0039, -0.0184, -0.3923, -0.0, -0.0174, -0.3913, -0.0, -0.0167, -0.3895, -0.0, -0.0156, -0.3919, -0.0, -0.0136, -0.3919, -0.0, -0.0114, -0.3897, -0.0, -0.0136, -0.3897, -0.0, -0.0114, -0.3922, 0.0017, -0.0174, -0.3911, 0.0017, -0.0166, -0.3922, 0.0039, -0.0174, -0.3911, 0.0039, -0.0166, -0.3897, 0.0017, -0.0157, -0.3897, 0.0039, -0.0157, -0.3963, -0.0, -0.0092, -0.3963, -0.0, -0.0071, -0.3941, -0.0, -0.0092, -0.3941, -0.0, -0.0071, -0.3963, -0.0, -0.0049,
-0.3963, -0.0, -0.0027, -0.3941, -0.0, -0.0049, -0.3941, -0.0, -0.0027, -0.3919, -0.0, -0.0092, -0.3919, -0.0, -0.0071, -0.3897, -0.0, -0.0092, -0.3897, -0.0, -0.0071, -0.3919, -0.0, -0.0049, -0.3919, -0.0, -0.0027, -0.3897, -0.0, -0.0049, -0.3897, -0.0, -0.0027, -0.3884, -0.0, -0.0148, -0.3873, -0.0, -0.0141, -0.3876, -0.0, -0.0114, -0.3859, -0.0, -0.013, -0.3848, -0.0, -0.0123, -0.3884, 0.0017, -0.0148, -0.3884, 0.0039, -0.0148, -0.3873, 0.0017, -0.0141, -0.3873, 0.0039, -0.0141, -0.3858, 0.0017, -0.013, -0.3847, 0.0017, -0.0122, -0.3858, 0.0039, -0.013, -0.3847, 0.0039, -0.0122, -0.3801, -0.0, -0.0189, -0.383, -0.0, -0.011, -0.3821, -0.0, -0.0104, -0.3801, 0.0017, -0.0189, -0.3801, 0.0039, -0.0189, -0.3832, 0.0017, -0.0112,
-0.3832, 0.0039, -0.0112, -0.3821, 0.0017, -0.0104, -0.3821, 0.0039, -0.0104, -0.3876, -0.0, -0.0092, -0.3876, -0.0, -0.0071, -0.3854, -0.0, -0.0092, -0.3854, -0.0, -0.0071, -0.3876, -0.0, -0.0049, -0.3876, -0.0, -0.0027, -0.3854, -0.0, -0.0049, -0.3854, -0.0, -0.0027, -0.3832, -0.0, -0.0092, -0.3832, -0.0, -0.0071, -0.381, -0.0, -0.0096, -0.3811, -0.0, -0.0071, -0.3832, -0.0, -0.0049, -0.3832, -0.0, -0.0027, -0.3811, -0.0, -0.0049, -0.3811, -0.0, -0.0027, -0.381, 0.0017, -0.0096, -0.381, 0.0039, -0.0096, -0.3796, -0.0, -0.0432, -0.3785, -0.0, -0.0423, -0.3772, -0.0, -0.0413, -0.3789, -0.0, -0.0396, -0.3789, -0.0, -0.0374, -0.3761, -0.0, -0.0404, -0.3767, -0.0, -0.0374, -0.3796, 0.0017, -0.0432, -0.3785, 0.0017, -0.0423,
-0.3796, 0.0039, -0.0432, -0.3785, 0.0039, -0.0423, -0.3771, 0.0017, -0.0412, -0.3771, 0.0039, -0.0412, -0.376, 0.0017, -0.0404, -0.376, 0.0039, -0.0404, -0.3747, -0.0, -0.0394, -0.3737, -0.0, -0.0385, -0.3721, -0.0, -0.0373, -0.3746, 0.0017, -0.0393, -0.3736, 0.0017, -0.0385, -0.3746, 0.0039, -0.0393, -0.3736, 0.0039, -0.0385, -0.3724, 0.0017, -0.0376, -0.3724, 0.0039, -0.0376, -0.3711, -0.0, -0.0365, -0.3711, 0.0017, -0.0365, -0.3711, 0.0039, -0.0365, -0.3789, -0.0, -0.0353, -0.3789, -0.0, -0.0331, -0.3767, -0.0, -0.0353, -0.3767, -0.0, -0.0331, -0.3789, -0.0, -0.0309, -0.3789, -0.0, -0.0288, -0.3767, -0.0, -0.0309, -0.3767, -0.0, -0.0288, -0.3746, -0.0, -0.0353, -0.3746, -0.0, -0.0331, -0.3724, -0.0, -0.0353, -0.3724, -0.0, -0.0331,
-0.3746, -0.0, -0.0309, -0.3746, -0.0, -0.0288, -0.3724, -0.0, -0.0309, -0.3724, -0.0, -0.0288, -0.3789, -0.0, -0.0266, -0.3789, -0.0, -0.0244, -0.3767, -0.0, -0.0266, -0.3767, -0.0, -0.0244, -0.3789, -0.0, -0.0222, -0.3789, -0.0, -0.0201, -0.3767, -0.0, -0.0222, -0.3767, -0.0, -0.0201, -0.3746, -0.0, -0.0266, -0.3746, -0.0, -0.0244, -0.3724, -0.0, -0.0266, -0.3724, -0.0, -0.0244, -0.3746, -0.0, -0.0222, -0.3746, -0.0, -0.0201, -0.3724, -0.0, -0.0222, -0.3724, -0.0, -0.0201, -0.37, -0.0, -0.0357, -0.3702, -0.0, -0.0331, -0.3686, -0.0, -0.0346, -0.3676, -0.0, -0.0338, -0.3702, -0.0, -0.0309, -0.3702, -0.0, -0.0288, -0.368, -0.0, -0.0309, -0.368, -0.0, -0.0288, -0.37, 0.0017, -0.0357, -0.37, 0.0039, -0.0357,
-0.3686, 0.0017, -0.0346, -0.3675, 0.0017, -0.0337, -0.3686, 0.0039, -0.0346, -0.3675, 0.0039, -0.0337, -0.3662, -0.0, -0.0327, -0.3651, -0.0, -0.0318, -0.3659, -0.0, -0.0288, -0.3635, -0.0, -0.0305, -0.3637, -0.0, -0.0288, -0.3662, 0.0017, -0.0327, -0.3662, 0.0039, -0.0327, -0.3651, 0.0017, -0.0318, -0.3651, 0.0039, -0.0318, -0.3637, 0.0017, -0.0307, -0.3637, 0.0039, -0.0307, -0.3702, -0.0, -0.0266, -0.3702, -0.0, -0.0244, -0.368, -0.0, -0.0266, -0.368, -0.0, -0.0244, -0.3702, -0.0, -0.0222, -0.3702, -0.0, -0.0201, -0.368, -0.0, -0.0222, -0.368, -0.0, -0.0201, -0.3659, -0.0, -0.0266, -0.3659, -0.0, -0.0244, -0.3637, -0.0, -0.0266, -0.3637, -0.0, -0.0244, -0.3659, -0.0, -0.0222, -0.3659, -0.0, -0.0201, -0.3637, -0.0, -0.0222,
-0.3637, -0.0, -0.0201, -0.3791, -0.0, -0.0182, -0.3774, -0.0, -0.017, -0.3763, -0.0, -0.0162, -0.3789, 0.0017, -0.018, -0.3789, 0.0039, -0.018, -0.3775, 0.0017, -0.017, -0.3764, 0.0017, -0.0163, -0.3775, 0.0039, -0.017, -0.3764, 0.0039, -0.0163, -0.3746, -0.0, -0.0179, -0.3749, -0.0, -0.0152, -0.3724, -0.0, -0.0179, -0.3724, -0.0, -0.0157, -0.3738, -0.0, -0.0144, -0.3727, -0.0, -0.0136, -0.3749, 0.0017, -0.0152, -0.3749, 0.0039, -0.0152, -0.3738, 0.0017, -0.0144, -0.3738, 0.0039, -0.0144, -0.3724, 0.0017, -0.0134, -0.3724, 0.0039, -0.0134, -0.3795, -0.0, -0.0086, -0.3784, -0.0, -0.0078, -0.3769, -0.0, -0.0068, -0.3789, -0.0, -0.0049, -0.3789, -0.0, -0.0027, -0.3759, -0.0, -0.006, -0.3767, -0.0, -0.0027, -0.3794, 0.0017, -0.0085,
-0.3783, 0.0017, -0.0078, -0.3794, 0.0039, -0.0085, -0.3783, 0.0039, -0.0078, -0.3768, 0.0017, -0.0067, -0.3768, 0.0039, -0.0067, -0.3757, 0.0017, -0.0059, -0.3757, 0.0039, -0.0059, -0.3742, -0.0, -0.0049, -0.3746, -0.0, -0.0027, -0.3731, -0.0, -0.0041, -0.372, -0.0, -0.0033, -0.3746, 0.0017, -0.0051, -0.3746, 0.0039, -0.0051, -0.3731, 0.0017, -0.0041, -0.372, 0.0017, -0.0033, -0.3731, 0.0039, -0.0041, -0.372, 0.0039, -0.0033, -0.3702, -0.0, -0.0179, -0.3702, -0.0, -0.0157, -0.368, -0.0, -0.0179, -0.368, -0.0, -0.0157, -0.3711, -0.0, -0.0124, -0.37, -0.0, -0.0117, -0.368, -0.0, -0.0136, -0.3685, -0.0, -0.0106, -0.3712, 0.0017, -0.0125, -0.3701, 0.0017, -0.0118, -0.3712, 0.0039, -0.0125, -0.3701, 0.0039, -0.0118, -0.3686, 0.0017, -0.0107,
-0.3686, 0.0039, -0.0107, -0.3659, -0.0, -0.0179, -0.3659, -0.0, -0.0157, -0.3637, -0.0, -0.0179, -0.3637, -0.0, -0.0157, -0.3659, -0.0, -0.0136, -0.3659, -0.0, -0.0114, -0.3637, -0.0, -0.0136, -0.3637, -0.0, -0.0114, -0.3675, -0.0, -0.0099, -0.3706, -0.0, -0.0023, -0.3675, 0.0017, -0.0099, -0.3675, 0.0039, -0.0099, -0.3704, 0.0017, -0.0022, -0.3704, 0.0039, -0.0022, -0.366, -0.0, -0.0088, -0.3649, -0.0, -0.008, -0.3637, -0.0, -0.0092, -0.3639, -0.0, -0.0074, -0.366, 0.0017, -0.0088, -0.3649, 0.0017, -0.0081, -0.366, 0.0039, -0.0088, -0.3649, 0.0039, -0.0081, -0.3637, 0.0017, -0.0072, -0.3637, 0.0039, -0.0072, -0.3626, -0.0, -0.0298, -0.3615, -0.0, -0.029, -0.3602, -0.0, -0.0279, -0.3626, 0.0017, -0.0298, -0.3615, 0.0017, -0.029,
-0.3626, 0.0039, -0.0298, -0.3615, 0.0039, -0.029, -0.3601, 0.0017, -0.0279, -0.3601, 0.0039, -0.0279, -0.3615, -0.0, -0.0266, -0.3615, -0.0, -0.0244, -0.3591, -0.0, -0.0271, -0.3594, -0.0, -0.0244, -0.3615, -0.0, -0.0222, -0.3615, -0.0, -0.0201, -0.3594, -0.0, -0.0222, -0.3594, -0.0, -0.0201, -0.3591, 0.0017, -0.0271, -0.3591, 0.0039, -0.0271, -0.3577, -0.0, -0.026, -0.3566, -0.0, -0.0252, -0.3553, -0.0, -0.0241, -0.3572, -0.0, -0.0222, -0.3572, -0.0, -0.0201, -0.3543, -0.0, -0.0231, -0.355, -0.0, -0.0201, -0.3577, 0.0017, -0.026, -0.3566, 0.0017, -0.0251, -0.3577, 0.0039, -0.026, -0.3566, 0.0039, -0.0251, -0.3553, 0.0017, -0.024, -0.3553, 0.0039, -0.024, -0.3542, 0.0017, -0.0231, -0.3542, 0.0039, -0.0231, -0.353, -0.0, -0.0221,
-0.3519, -0.0, -0.0212, -0.3504, -0.0, -0.0199, -0.3529, 0.0017, -0.022, -0.3519, 0.0017, -0.0211, -0.3529, 0.0039, -0.022, -0.3519, 0.0039, -0.0211, -0.3507, 0.0017, -0.0201, -0.3507, 0.0039, -0.0201, -0.3495, -0.0, -0.0191, -0.3494, 0.0017, -0.0191, -0.3494, 0.0039, -0.0191, -0.3615, -0.0, -0.0179, -0.3615, -0.0, -0.0157, -0.3594, -0.0, -0.0179, -0.3594, -0.0, -0.0157, -0.3615, -0.0, -0.0136, -0.3615, -0.0, -0.0114, -0.3594, -0.0, -0.0136, -0.3594, -0.0, -0.0114, -0.3572, -0.0, -0.0179, -0.3572, -0.0, -0.0157, -0.355, -0.0, -0.0179, -0.355, -0.0, -0.0157, -0.3572, -0.0, -0.0136, -0.3572, -0.0, -0.0114, -0.355, -0.0, -0.0136, -0.355, -0.0, -0.0114, -0.3615, -0.0, -0.0092, -0.3622, -0.0, -0.0061, -0.3594, -0.0, -0.0092,
-0.3594, -0.0, -0.0071, -0.3611, -0.0, -0.0054, -0.3596, -0.0, -0.0044, -0.3585, -0.0, -0.0036, -0.3623, 0.0017, -0.0062, -0.3623, 0.0039, -0.0062, -0.3612, 0.0017, -0.0055, -0.3612, 0.0039, -0.0055, -0.3596, 0.0017, -0.0044, -0.3585, 0.0017, -0.0036, -0.3596, 0.0039, -0.0044, -0.3585, 0.0039, -0.0036, -0.3572, -0.0, -0.0092, -0.3572, -0.0, -0.0071, -0.355, -0.0, -0.0092, -0.355, -0.0, -0.0071, -0.3572, -0.0, -0.0049, -0.3574, -0.0, -0.0029, -0.355, -0.0, -0.0049, -0.3557, -0.0, -0.0017, -0.3572, 0.0017, -0.0027, -0.3572, 0.0039, -0.0027, -0.3559, 0.0017, -0.0018, -0.3559, 0.0039, -0.0018, -0.3529, -0.0, -0.0179, -0.3529, -0.0, -0.0157, -0.3507, -0.0, -0.0179, -0.3507, -0.0, -0.0157, -0.3529, -0.0, -0.0136, -0.3529, -0.0, -0.0114,
-0.3507, -0.0, -0.0136, -0.3507, -0.0, -0.0114, -0.3483, -0.0, -0.0182, -0.3485, -0.0, -0.0157, -0.347, -0.0, -0.0172, -0.3459, -0.0, -0.0163, -0.3485, -0.0, -0.0136, -0.3485, -0.0, -0.0114, -0.3463, -0.0, -0.0136, -0.3463, -0.0, -0.0114, -0.3484, 0.0017, -0.0182, -0.3484, 0.0039, -0.0182, -0.347, 0.0017, -0.0172, -0.3459, 0.0017, -0.0163, -0.347, 0.0039, -0.0172, -0.3459, 0.0039, -0.0163, -0.3529, -0.0, -0.0092, -0.3529, -0.0, -0.0071, -0.3507, -0.0, -0.0092, -0.3507, -0.0, -0.0071, -0.3529, -0.0, -0.0049, -0.3529, -0.0, -0.0027, -0.3507, -0.0, -0.0049, -0.3507, -0.0, -0.0027, -0.3485, -0.0, -0.0092, -0.3485, -0.0, -0.0071, -0.3463, -0.0, -0.0092, -0.3463, -0.0, -0.0071, -0.3485, -0.0, -0.0049, -0.3485, -0.0, -0.0027,
-0.3463, -0.0, -0.0049, -0.3463, -0.0, -0.0027, -0.3446, -0.0, -0.1019, -0.3437, -0.0, -0.1015, -0.3442, -0.0, -0.0982, -0.3417, -0.0, -0.1007, -0.342, -0.0, -0.0982, -0.3444, 0.0017, -0.1018, -0.3444, 0.0039, -0.1018, -0.3433, 0.0017, -0.1014, -0.3433, 0.0039, -0.1014, -0.342, 0.0017, -0.1008, -0.342, 0.0039, -0.1008, -0.3397, -0.0, -0.0999, -0.3398, -0.0, -0.0982, -0.3385, -0.0, -0.0994, -0.3375, -0.0, -0.099, -0.3398, 0.0017, -0.1, -0.3398, 0.0039, -0.1, -0.3385, 0.0017, -0.0994, -0.3374, 0.0017, -0.099, -0.3385, 0.0039, -0.0994, -0.3374, 0.0039, -0.099, -0.3442, -0.0, -0.096, -0.3442, -0.0, -0.0939, -0.342, -0.0, -0.096, -0.342, -0.0, -0.0939, -0.3442, -0.0, -0.0917, -0.3442, -0.0, -0.0895, -0.342, -0.0, -0.0917,
-0.342, -0.0, -0.0895, -0.3398, -0.0, -0.096, -0.3398, -0.0, -0.0939, -0.3377, -0.0, -0.096, -0.3377, -0.0, -0.0939, -0.3398, -0.0, -0.0917, -0.3398, -0.0, -0.0895, -0.3377, -0.0, -0.0917, -0.3377, -0.0, -0.0895, -0.3353, -0.0, -0.0981, -0.334, -0.0, -0.0974, -0.3355, 0.0017, -0.0981, -0.3355, 0.0039, -0.0981, -0.3339, 0.0017, -0.0974, -0.3339, 0.0039, -0.0974, -0.3355, -0.0, -0.096, -0.3355, -0.0, -0.0939, -0.3329, -0.0, -0.0969, -0.3333, -0.0, -0.0939, -0.3355, -0.0, -0.0917, -0.3355, -0.0, -0.0895, -0.3333, -0.0, -0.0917, -0.3333, -0.0, -0.0895, -0.3328, 0.0017, -0.0968, -0.3328, 0.0039, -0.0968, -0.3309, -0.0, -0.0958, -0.3312, -0.0, -0.0939, -0.3298, -0.0, -0.0952, -0.3287, -0.0, -0.0946, -0.3312, -0.0, -0.0917,
-0.3312, -0.0, -0.0895, -0.329, -0.0, -0.0917, -0.329, -0.0, -0.0895, -0.3311, 0.0017, -0.096, -0.3311, 0.0039, -0.096, -0.3297, 0.0017, -0.0952, -0.3286, 0.0017, -0.0945, -0.3297, 0.0039, -0.0952, -0.3286, 0.0039, -0.0945, -0.3443, -0.0, -0.0879, -0.3423, -0.0, -0.087, -0.3442, 0.0017, -0.0879, -0.3442, 0.0039, -0.0879, -0.342, 0.0017, -0.0869, -0.342, 0.0039, -0.0869, -0.3404, -0.0, -0.0862, -0.3394, -0.0, -0.0858, -0.3377, -0.0, -0.0874, -0.3379, -0.0, -0.0851, -0.3407, 0.0017, -0.0864, -0.3396, 0.0017, -0.0859, -0.3407, 0.0039, -0.0864, -0.3396, 0.0039, -0.0859, -0.3377, 0.0017, -0.085, -0.3377, 0.0039, -0.085, -0.3355, -0.0, -0.0874, -0.336, -0.0, -0.0842, -0.3333, -0.0, -0.0874, -0.3333, -0.0, -0.0852, -0.3349, -0.0, -0.0837,
-0.3336, -0.0, -0.083, -0.3362, 0.0017, -0.0843, -0.3362, 0.0039, -0.0843, -0.3351, 0.0017, -0.0838, -0.3351, 0.0039, -0.0838, -0.3333, 0.0017, -0.0829, -0.3333, 0.0039, -0.0829, -0.3312, -0.0, -0.0874, -0.3312, -0.0, -0.0852, -0.329, -0.0, -0.0874, -0.329, -0.0, -0.0852, -0.3317, -0.0, -0.082, -0.3307, -0.0, -0.0814, -0.329, -0.0, -0.083, -0.3291, -0.0, -0.0804, -0.3319, 0.0017, -0.0821, -0.3308, 0.0017, -0.0815, -0.3319, 0.0039, -0.0821, -0.3308, 0.0039, -0.0815, -0.329, 0.0017, -0.0804, -0.329, 0.0039, -0.0804, -0.328, -0.0, -0.0797, -0.328, 0.0017, -0.0797, -0.328, 0.0039, -0.0797, -0.3267, -0.0, -0.0934, -0.3257, -0.0, -0.0928, -0.3268, -0.0, -0.0917, -0.3268, -0.0, -0.0895, -0.3246, -0.0, -0.0921, -0.3246, -0.0, -0.0895,
-0.3268, 0.0017, -0.0935, -0.3268, 0.0039, -0.0935, -0.3257, 0.0017, -0.0928, -0.3257, 0.0039, -0.0928, -0.3246, 0.0017, -0.0921, -0.3246, 0.0039, -0.0921, -0.323, -0.0, -0.0911, -0.3219, -0.0, -0.0903, -0.3205, -0.0, -0.0893, -0.3229, 0.0017, -0.091, -0.3219, 0.0017, -0.0903, -0.3229, 0.0039, -0.091, -0.3219, 0.0039, -0.0903, -0.3203, 0.0017, -0.0892, -0.3203, 0.0039, -0.0892, -0.3268, -0.0, -0.0874, -0.3268, -0.0, -0.0852, -0.3246, -0.0, -0.0874, -0.3246, -0.0, -0.0852, -0.3268, -0.0, -0.083, -0.3268, -0.0, -0.0808, -0.3246, -0.0, -0.083, -0.3246, -0.0, -0.0808, -0.3225, -0.0, -0.0874, -0.3225, -0.0, -0.0852, -0.3194, -0.0, -0.0885, -0.3203, -0.0, -0.0852, -0.3225, -0.0, -0.083, -0.3225, -0.0, -0.0808, -0.3203, -0.0, -0.083,
-0.3203, -0.0, -0.0808, -0.3193, 0.0017, -0.0884, -0.3193, 0.0039, -0.0884, -0.327, -0.0, -0.0791, -0.3252, -0.0, -0.0779, -0.3241, -0.0, -0.0771, -0.3268, 0.0017, -0.079, -0.3268, 0.0039, -0.079, -0.3253, 0.0017, -0.0779, -0.3242, 0.0017, -0.0772, -0.3253, 0.0039, -0.0779, -0.3242, 0.0039, -0.0772, -0.3225, -0.0, -0.0787, -0.3226, -0.0, -0.0761, -0.3203, -0.0, -0.0787, -0.3203, -0.0, -0.0765, -0.3215, -0.0, -0.0753, -0.3206, -0.0, -0.0745, -0.3226, 0.0017, -0.0761, -0.3226, 0.0039, -0.0761, -0.3216, 0.0017, -0.0753, -0.3216, 0.0039, -0.0753, -0.3203, 0.0017, -0.0743, -0.3203, 0.0039, -0.0743, -0.3178, -0.0, -0.0873, -0.3181, -0.0, -0.0852, -0.3168, -0.0, -0.0865, -0.3157, -0.0, -0.0856, -0.3181, -0.0, -0.083, -0.3181, -0.0, -0.0808,
-0.316, -0.0, -0.083, -0.316, -0.0, -0.0808, -0.3181, 0.0017, -0.0875, -0.3181, 0.0039, -0.0875, -0.3168, 0.0017, -0.0865, -0.3157, 0.0017, -0.0856, -0.3168, 0.0039, -0.0865, -0.3157, 0.0039, -0.0856, -0.3144, -0.0, -0.0846, -0.3133, -0.0, -0.0837, -0.3138, -0.0, -0.0808, -0.312, -0.0, -0.0826, -0.3109, -0.0, -0.0816, -0.3143, 0.0017, -0.0845, -0.3143, 0.0039, -0.0845, -0.3133, 0.0017, -0.0836, -0.3133, 0.0039, -0.0836, -0.312, 0.0017, -0.0826, -0.3109, 0.0017, -0.0816, -0.312, 0.0039, -0.0826, -0.3109, 0.0039, -0.0816, -0.3181, -0.0, -0.0787, -0.3181, -0.0, -0.0765, -0.316, -0.0, -0.0787, -0.316, -0.0, -0.0765, -0.319, -0.0, -0.0733, -0.318, -0.0, -0.0724, -0.316, -0.0, -0.0743, -0.3167, -0.0, -0.0713, -0.3191, 0.0017, -0.0733,
-0.318, 0.0017, -0.0724, -0.3191, 0.0039, -0.0733, -0.318, 0.0039, -0.0724, -0.3167, 0.0017, -0.0713, -0.3167, 0.0039, -0.0713, -0.3138, -0.0, -0.0787, -0.3138, -0.0, -0.0765, -0.3116, -0.0, -0.0787, -0.3116, -0.0, -0.0765, -0.3138, -0.0, -0.0743, -0.3138, -0.0, -0.0722, -0.3116, -0.0, -0.0743, -0.3116, -0.0, -0.0722, -0.3097, -0.0, -0.0806, -0.3097, 0.0017, -0.0805, -0.3097, 0.0039, -0.0805, -0.3086, -0.0, -0.0795, -0.3095, -0.0, -0.0765, -0.3075, -0.0, -0.0784, -0.3065, -0.0, -0.0773, -0.3095, -0.0, -0.0743, -0.3095, -0.0, -0.0722, -0.3073, -0.0, -0.0743, -0.3073, -0.0, -0.0722, -0.3086, 0.0017, -0.0795, -0.3086, 0.0039, -0.0795, -0.3075, 0.0017, -0.0784, -0.3065, 0.0017, -0.0773, -0.3075, 0.0039, -0.0784, -0.3065, 0.0039, -0.0773,
-0.3054, -0.0, -0.0762, -0.3043, -0.0, -0.075, -0.3051, -0.0, -0.0722, -0.3034, -0.0, -0.0739, -0.3023, -0.0, -0.0726, -0.3054, 0.0017, -0.0762, -0.3054, 0.0039, -0.0762, -0.3044, 0.0017, -0.075, -0.3044, 0.0039, -0.075, -0.3034, 0.0017, -0.0739, -0.3024, 0.0017, -0.0726, -0.3034, 0.0039, -0.0739, -0.3024, 0.0039, -0.0726, -0.3015, -0.0, -0.0715, -0.3015, 0.0017, -0.0715, -0.3015, 0.0039, -0.0715, -0.3156, -0.0, -0.0704, -0.3156, 0.0017, -0.0704, -0.3156, 0.0039, -0.0704, -0.3144, -0.0, -0.0693, -0.3133, -0.0, -0.0683, -0.3116, -0.0, -0.07, -0.3122, -0.0, -0.0673, -0.3112, -0.0, -0.0661, -0.3144, 0.0017, -0.0693, -0.3133, 0.0017, -0.0683, -0.3144, 0.0039, -0.0693, -0.3133, 0.0039, -0.0683, -0.3122, 0.0017, -0.0672, -0.3122, 0.0039, -0.0672,
-0.3111, 0.0017, -0.0661, -0.3111, 0.0039, -0.0661, -0.3446, -0.0, -0.0152, -0.3435, -0.0, -0.0144, -0.3442, -0.0, -0.0114, -0.3422, -0.0, -0.0133, -0.3411, -0.0, -0.0125, -0.3446, 0.0017, -0.0152, -0.3446, 0.0039, -0.0152, -0.3435, 0.0017, -0.0143, -0.3435, 0.0039, -0.0143, -0.3421, 0.0017, -0.0132, -0.341, 0.0017, -0.0124, -0.3421, 0.0039, -0.0132, -0.341, 0.0039, -0.0124, -0.3396, -0.0, -0.0113, -0.3385, -0.0, -0.0105, -0.3398, 0.0017, -0.0115, -0.3398, 0.0039, -0.0115, -0.3385, 0.0017, -0.0105, -0.3385, 0.0039, -0.0105, -0.3442, -0.0, -0.0092, -0.3442, -0.0, -0.0071, -0.342, -0.0, -0.0092, -0.342, -0.0, -0.0071, -0.3442, -0.0, -0.0049, -0.3442, -0.0, -0.0027, -0.342, -0.0, -0.0049, -0.342, -0.0, -0.0027, -0.3398, -0.0, -0.0092,
-0.3398, -0.0, -0.0071, -0.3374, -0.0, -0.0097, -0.3377, -0.0, -0.0071, -0.3398, -0.0, -0.0049, -0.3398, -0.0, -0.0027, -0.3377, -0.0, -0.0049, -0.3377, -0.0, -0.0027, -0.3374, 0.0017, -0.0097, -0.3374, 0.0039, -0.0097, -0.3361, -0.0, -0.0086, -0.335, -0.0, -0.0078, -0.3336, -0.0, -0.0067, -0.3355, -0.0, -0.0049, -0.3355, -0.0, -0.0027, -0.3325, -0.0, -0.0059, -0.3333, -0.0, -0.0027, -0.336, 0.0017, -0.0086, -0.3349, 0.0017, -0.0078, -0.336, 0.0039, -0.0086, -0.3349, 0.0039, -0.0078, -0.3335, 0.0017, -0.0067, -0.3335, 0.0039, -0.0067, -0.3324, 0.0017, -0.0059, -0.3324, 0.0039, -0.0059, -0.3309, -0.0, -0.0047, -0.3312, -0.0, -0.0027, -0.3298, -0.0, -0.004, -0.3287, -0.0, -0.0032, -0.3312, 0.0017, -0.0049, -0.3312, 0.0039, -0.0049,
-0.3298, 0.0017, -0.004, -0.3287, 0.0017, -0.0032, -0.3298, 0.0039, -0.004, -0.3287, 0.0039, -0.0032, -0.3273, -0.0, -0.0022, -0.3272, 0.0017, -0.0021, -0.3272, 0.0039, -0.0021, -0.3095, -0.0, -0.07, -0.3095, -0.0, -0.0678, -0.3073, -0.0, -0.07, -0.3073, -0.0, -0.0678, -0.3102, -0.0, -0.065, -0.3092, -0.0, -0.0637, -0.3073, -0.0, -0.0656, -0.3083, -0.0, -0.0626, -0.3102, 0.0017, -0.065, -0.3091, 0.0017, -0.0636, -0.3102, 0.0039, -0.065, -0.3091, 0.0039, -0.0636, -0.3082, 0.0017, -0.0626, -0.3082, 0.0039, -0.0626, -0.3051, -0.0, -0.07, -0.3051, -0.0, -0.0678, -0.3029, -0.0, -0.07, -0.3029, -0.0, -0.0678, -0.3051, -0.0, -0.0656, -0.3051, -0.0, -0.0635, -0.3029, -0.0, -0.0656, -0.3029, -0.0, -0.0635, -0.307, -0.0, -0.0611,
-0.3063, -0.0, -0.0601, -0.3072, 0.0017, -0.0613, -0.3063, 0.0017, -0.0601, -0.3072, 0.0039, -0.0613, -0.3063, 0.0039, -0.0601, -0.3051, -0.0, -0.0613, -0.3054, -0.0, -0.059, -0.3029, -0.0, -0.0613, -0.3029, -0.0, -0.0591, -0.3044, -0.0, -0.0576, -0.3036, -0.0, -0.0565, -0.3026, -0.0, -0.0551, -0.3055, 0.0017, -0.059, -0.3055, 0.0039, -0.059, -0.3044, 0.0017, -0.0576, -0.3044, 0.0039, -0.0576, -0.3036, 0.0017, -0.0565, -0.3025, 0.0017, -0.055, -0.3036, 0.0039, -0.0565, -0.3025, 0.0039, -0.055, -0.3004, -0.0, -0.0702, -0.3008, -0.0, -0.0678, -0.2996, -0.0, -0.069, -0.2988, -0.0, -0.0681, -0.3008, -0.0, -0.0656, -0.3008, -0.0, -0.0635, -0.2976, -0.0, -0.0663, -0.2986, -0.0, -0.0635, -0.3004, 0.0017, -0.0701, -0.3004, 0.0039, -0.0701,
-0.2996, 0.0017, -0.069, -0.2987, 0.0017, -0.0678, -0.2996, 0.0039, -0.069, -0.2987, 0.0039, -0.0678, -0.2977, 0.0017, -0.0665, -0.2977, 0.0039, -0.0665, -0.2969, -0.0, -0.0653, -0.2959, -0.0, -0.0637, -0.2952, -0.0, -0.0626, -0.297, 0.0017, -0.0654, -0.2959, 0.0017, -0.0637, -0.297, 0.0039, -0.0654, -0.2959, 0.0039, -0.0637, -0.2952, 0.0017, -0.0627, -0.2952, 0.0039, -0.0627, -0.3008, -0.0, -0.0613, -0.3008, -0.0, -0.0591, -0.2986, -0.0, -0.0613, -0.2986, -0.0, -0.0591, -0.3008, -0.0, -0.057, -0.3019, -0.0, -0.054, -0.2986, -0.0, -0.057, -0.2986, -0.0, -0.0548, -0.3017, 0.0017, -0.0539, -0.3017, 0.0039, -0.0539, -0.2964, -0.0, -0.0613, -0.2964, -0.0, -0.0591, -0.2945, -0.0, -0.0615, -0.2933, -0.0, -0.0597, -0.2964, -0.0, -0.057,
-0.2964, -0.0, -0.0548, -0.2943, -0.0, -0.057, -0.2943, -0.0, -0.0548, -0.2943, 0.0017, -0.0613, -0.2934, 0.0017, -0.0599, -0.2943, 0.0039, -0.0613, -0.2934, 0.0039, -0.0599, -0.3007, -0.0, -0.0524, -0.3, -0.0, -0.0512, -0.2986, -0.0, -0.0526, -0.2993, -0.0, -0.0501, -0.2982, -0.0, -0.0481, -0.2976, -0.0, -0.047, -0.3009, 0.0017, -0.0526, -0.3, 0.0017, -0.0511, -0.3009, 0.0039, -0.0526, -0.3, 0.0039, -0.0511, -0.2993, 0.0017, -0.05, -0.2993, 0.0039, -0.05, -0.2983, 0.0017, -0.0483, -0.2976, 0.0017, -0.047, -0.2983, 0.0039, -0.0483, -0.2976, 0.0039, -0.047, -0.2964, -0.0, -0.0526, -0.2964, -0.0, -0.0505, -0.2943, -0.0, -0.0526, -0.2943, -0.0, -0.0505, -0.2964, -0.0, -0.0483, -0.2971, -0.0, -0.0459, -0.2943, -0.0, -0.0483,
-0.2943, -0.0, -0.0461, -0.2971, 0.0017, -0.0459, -0.2971, 0.0039, -0.0459, -0.2961, -0.0, -0.0438, -0.2956, -0.0, -0.0426, -0.2943, -0.0, -0.0439, -0.2951, -0.0, -0.0415, -0.2943, -0.0, -0.0394, -0.2936, -0.0, -0.0378, -0.2962, 0.0017, -0.044, -0.2955, 0.0017, -0.0425, -0.2962, 0.0039, -0.044, -0.2955, 0.0039, -0.0425, -0.2951, 0.0017, -0.0414, -0.2951, 0.0039, -0.0414, -0.2943, 0.0017, -0.0396, -0.2935, 0.0017, -0.0375, -0.2943, 0.0039, -0.0396, -0.2935, 0.0039, -0.0375, -0.2927, -0.0, -0.0587, -0.2917, -0.0, -0.057, -0.2921, -0.0, -0.0548, -0.291, -0.0, -0.0559, -0.2904, -0.0, -0.0549, -0.2927, 0.0017, -0.0588, -0.2927, 0.0039, -0.0588, -0.2916, 0.0017, -0.057, -0.2916, 0.0039, -0.057, -0.291, 0.0017, -0.0559, -0.2904, 0.0017, -0.0548,
-0.291, 0.0039, -0.0559, -0.2904, 0.0039, -0.0548, -0.2921, -0.0, -0.0526, -0.2921, -0.0, -0.0505, -0.2893, -0.0, -0.0527, -0.2899, -0.0, -0.0505, -0.2921, -0.0, -0.0483, -0.2921, -0.0, -0.0461, -0.2899, -0.0, -0.0483, -0.2899, -0.0, -0.0461, -0.2893, 0.0017, -0.0527, -0.2893, 0.0039, -0.0527, -0.2888, -0.0, -0.0516, -0.2884, -0.0, -0.0506, -0.2875, -0.0, -0.0486, -0.2867, -0.0, -0.0466, -0.2862, -0.0, -0.0457, -0.2888, 0.0017, -0.0516, -0.2883, 0.0017, -0.0505, -0.2888, 0.0039, -0.0516, -0.2883, 0.0039, -0.0505, -0.2874, 0.0017, -0.0483, -0.2868, 0.0017, -0.0469, -0.2874, 0.0039, -0.0483, -0.2868, 0.0039, -0.0469, -0.2863, 0.0017, -0.0458, -0.2863, 0.0039, -0.0458, -0.2921, -0.0, -0.0439, -0.2921, -0.0, -0.0418, -0.2899, -0.0, -0.0439,
-0.2899, -0.0, -0.0418, -0.2921, -0.0, -0.0396, -0.2933, -0.0, -0.0369, -0.2899, -0.0, -0.0396, -0.2899, -0.0, -0.0374, -0.2931, 0.0017, -0.0364, -0.2931, 0.0039, -0.0364, -0.2878, -0.0, -0.0439, -0.2878, -0.0, -0.0418, -0.2856, -0.0, -0.0441, -0.2848, -0.0, -0.0419, -0.2878, -0.0, -0.0396, -0.2878, -0.0, -0.0374, -0.2856, -0.0, -0.0396, -0.2856, -0.0, -0.0374, -0.2855, 0.0017, -0.044, -0.2848, 0.0017, -0.0419, -0.2855, 0.0039, -0.044, -0.2848, 0.0039, -0.0419, -0.2844, -0.0, -0.0408, -0.2841, -0.0, -0.0397, -0.2835, -0.0, -0.0376, -0.2845, 0.0017, -0.0408, -0.2845, 0.0039, -0.0408, -0.2841, 0.0017, -0.0396, -0.2834, 0.0017, -0.0374, -0.2841, 0.0039, -0.0396, -0.2834, 0.0039, -0.0374, -0.2926, -0.0, -0.035, -0.292, -0.0, -0.033,
-0.2899, -0.0, -0.0353, -0.2899, -0.0, -0.0331, -0.2914, -0.0, -0.0309, -0.2911, -0.0, -0.0297, -0.2899, -0.0, -0.0309, -0.2908, -0.0, -0.0287, -0.2927, 0.0017, -0.0353, -0.292, 0.0017, -0.0331, -0.2927, 0.0039, -0.0353, -0.292, 0.0039, -0.0331, -0.2914, 0.0017, -0.0309, -0.2911, 0.0017, -0.0297, -0.2914, 0.0039, -0.0309, -0.2911, 0.0039, -0.0297, -0.2908, 0.0017, -0.0286, -0.2908, 0.0039, -0.0286, -0.2878, -0.0, -0.0353, -0.2878, -0.0, -0.0331, -0.2856, -0.0, -0.0353, -0.2856, -0.0, -0.0331, -0.2878, -0.0, -0.0309, -0.2878, -0.0, -0.0288, -0.2856, -0.0, -0.0309, -0.2856, -0.0, -0.0288, -0.2903, -0.0, -0.0264, -0.2899, -0.0, -0.0243, -0.2894, -0.0, -0.0222, -0.2891, -0.0, -0.0206, -0.2904, 0.0017, -0.0266, -0.2899, 0.0017, -0.0244,
-0.2904, 0.0039, -0.0266, -0.2899, 0.0039, -0.0244, -0.2894, 0.0017, -0.0222, -0.289, 0.0017, -0.0204, -0.2894, 0.0039, -0.0222, -0.289, 0.0039, -0.0204, -0.2878, -0.0, -0.0266, -0.2878, -0.0, -0.0244, -0.2856, -0.0, -0.0266, -0.2856, -0.0, -0.0244, -0.2878, -0.0, -0.0222, -0.2889, -0.0, -0.0198, -0.2856, -0.0, -0.0222, -0.2856, -0.0, -0.0201, -0.2888, 0.0017, -0.0193, -0.2888, 0.0039, -0.0193, -0.2829, -0.0, -0.0355, -0.2823, -0.0, -0.0334, -0.2821, -0.0, -0.0325, -0.2834, -0.0, -0.0309, -0.2834, -0.0, -0.0288, -0.2817, -0.0, -0.031, -0.2812, -0.0, -0.0289, -0.2828, 0.0017, -0.0353, -0.2824, 0.0017, -0.0338, -0.2828, 0.0039, -0.0353, -0.2824, 0.0039, -0.0338, -0.2821, 0.0017, -0.0328, -0.2821, 0.0039, -0.0328, -0.2816, 0.0017, -0.0309,
-0.2812, 0.0017, -0.0288, -0.2816, 0.0039, -0.0309, -0.2812, 0.0039, -0.0288, -0.2834, -0.0, -0.0266, -0.2834, -0.0, -0.0244, -0.2808, -0.0, -0.0267, -0.2804, -0.0, -0.0246, -0.2834, -0.0, -0.0222, -0.2834, -0.0, -0.0201, -0.2801, -0.0, -0.0225, -0.2812, -0.0, -0.0201, -0.2808, 0.0017, -0.0266, -0.2804, 0.0017, -0.0244, -0.2808, 0.0039, -0.0266, -0.2804, 0.0039, -0.0244, -0.2802, 0.0017, -0.0231, -0.2802, 0.0039, -0.0231, -0.28, -0.0, -0.0218, -0.2797, -0.0, -0.0201, -0.28, 0.0017, -0.022, -0.2797, 0.0017, -0.0201, -0.28, 0.0039, -0.022, -0.2797, 0.0039, -0.0201, -0.2885, -0.0, -0.0178, -0.2883, -0.0, -0.0157, -0.2856, -0.0, -0.0179, -0.2856, -0.0, -0.0157, -0.2882, -0.0, -0.0136, -0.2882, -0.0, -0.0114, -0.2856, -0.0, -0.0136,
-0.2856, -0.0, -0.0114, -0.2885, 0.0017, -0.0179, -0.2883, 0.0017, -0.0157, -0.2885, 0.0039, -0.0179, -0.2883, 0.0039, -0.0157, -0.2882, 0.0017, -0.0136, -0.2882, 0.0017, -0.0114, -0.2882, 0.0039, -0.0136, -0.2882, 0.0039, -0.0114, -0.2882, -0.0, -0.0092, -0.2882, -0.0, -0.0071, -0.2856, -0.0, -0.0092, -0.2856, -0.0, -0.0071, -0.2882, -0.0, -0.0049, -0.2882, -0.0, -0.0027, -0.2856, -0.0, -0.0049, -0.2856, -0.0, -0.0027, -0.2882, 0.0017, -0.0092, -0.2882, 0.0017, -0.0071, -0.2882, 0.0039, -0.0092, -0.2882, 0.0039, -0.0071, -0.2882, 0.0017, -0.0049, -0.2882, 0.0017, -0.0027, -0.2882, 0.0039, -0.0049, -0.2882, 0.0039, -0.0027, -0.2834, -0.0, -0.0179, -0.2834, -0.0, -0.0157, -0.2812, -0.0, -0.0179, -0.2812, -0.0, -0.0157, -0.2834, -0.0, -0.0136,
-0.2834, -0.0, -0.0114, -0.2812, -0.0, -0.0136, -0.2812, -0.0, -0.0114, -0.2795, -0.0, -0.0179, -0.2794, -0.0, -0.0157, -0.2794, -0.0, -0.0136, -0.2794, -0.0, -0.0114, -0.2795, 0.0017, -0.0179, -0.2794, 0.0017, -0.0157, -0.2795, 0.0039, -0.0179, -0.2794, 0.0039, -0.0157, -0.2794, 0.0017, -0.0136, -0.2794, 0.0017, -0.0114, -0.2794, 0.0039, -0.0136, -0.2794, 0.0039, -0.0114, -0.2834, -0.0, -0.0092, -0.2834, -0.0, -0.0071, -0.2812, -0.0, -0.0092, -0.2812, -0.0, -0.0071, -0.2834, -0.0, -0.0049, -0.2834, -0.0, -0.0027, -0.2812, -0.0, -0.0049, -0.2812, -0.0, -0.0027, -0.2794, -0.0, -0.0092, -0.2794, -0.0, -0.0071, -0.2794, -0.0, -0.0049, -0.2794, -0.0, -0.0027, -0.2794, 0.0017, -0.0092, -0.2794, 0.0017, -0.0071, -0.2794, 0.0039, -0.0092,
-0.2794, 0.0039, -0.0071, -0.2794, 0.0017, -0.0049, -0.2794, 0.0017, -0.0027, -0.2794, 0.0039, -0.0049, -0.2794, 0.0039, -0.0027, -0.2196, -0.0, -0.067, -0.2195, -0.0, -0.0666, -0.2196, 0.0017, -0.067, -0.2196, 0.0039, -0.067, -0.2195, 0.0017, -0.0666, -0.2195, 0.0039, -0.0666, -0.2183, -0.0, -0.067, -0.2172, -0.0, -0.0667, -0.219, -0.0, -0.0655, -0.218, -0.0, -0.0633, -0.216, -0.0, -0.0665, -0.2161, -0.0, -0.0635, -0.2183, 0.0017, -0.067, -0.2183, 0.0039, -0.067, -0.2172, 0.0017, -0.0667, -0.2172, 0.0039, -0.0667, -0.219, 0.0017, -0.0655, -0.2181, 0.0017, -0.0635, -0.219, 0.0039, -0.0655, -0.2181, 0.0039, -0.0635, -0.216, 0.0017, -0.0665, -0.216, 0.0039, -0.0665, -0.2174, -0.0, -0.0622, -0.2168, -0.0, -0.0611, -0.2158, -0.0, -0.0593,
-0.2174, 0.0017, -0.0621, -0.2174, 0.0039, -0.0621, -0.2168, 0.0017, -0.061, -0.2157, 0.0017, -0.0592, -0.2168, 0.0039, -0.061, -0.2157, 0.0039, -0.0592, -0.2142, -0.0, -0.0654, -0.2131, -0.0, -0.0641, -0.2124, -0.0, -0.063, -0.2141, 0.0017, -0.0652, -0.2132, 0.0017, -0.0642, -0.2141, 0.0039, -0.0652, -0.2132, 0.0039, -0.0642, -0.2124, 0.0017, -0.0631, -0.2124, 0.0039, -0.0631, -0.214, -0.0, -0.0613, -0.2152, -0.0, -0.0583, -0.2114, -0.0, -0.0614, -0.2118, -0.0, -0.0591, -0.2142, -0.0, -0.0566, -0.2134, -0.0, -0.0553, -0.2118, -0.0, -0.057, -0.2127, -0.0, -0.0542, -0.215, 0.0017, -0.0581, -0.215, 0.0039, -0.0581, -0.2113, 0.0017, -0.0614, -0.2113, 0.0039, -0.0614, -0.2144, 0.0017, -0.057, -0.2133, 0.0017, -0.0552, -0.2144, 0.0039, -0.057,
-0.2133, 0.0039, -0.0552, -0.2126, 0.0017, -0.0541, -0.2126, 0.0039, -0.0541, -0.2107, -0.0, -0.0603, -0.2101, -0.0, -0.0593, -0.2089, -0.0, -0.0573, -0.2096, -0.0, -0.0548, -0.2083, -0.0, -0.0562, -0.2075, -0.0, -0.0551, -0.2107, 0.0017, -0.0603, -0.21, 0.0017, -0.0591, -0.2107, 0.0039, -0.0603, -0.21, 0.0039, -0.0591, -0.2089, 0.0017, -0.0574, -0.2089, 0.0039, -0.0574, -0.2083, 0.0017, -0.0563, -0.2074, 0.0017, -0.0548, -0.2083, 0.0039, -0.0563, -0.2074, 0.0039, -0.0548, -0.2115, -0.0, -0.0524, -0.2109, -0.0, -0.0513, -0.2117, 0.0017, -0.0526, -0.2109, 0.0017, -0.0513, -0.2117, 0.0039, -0.0526, -0.2109, 0.0039, -0.0513, -0.2096, -0.0, -0.0526, -0.2102, -0.0, -0.0502, -0.2065, -0.0, -0.0532, -0.2075, -0.0, -0.0505, -0.2092, -0.0, -0.0486,
-0.2086, -0.0, -0.0475, -0.2074, -0.0, -0.0458, -0.2102, 0.0017, -0.0502, -0.2102, 0.0039, -0.0502, -0.2066, 0.0017, -0.0534, -0.2066, 0.0039, -0.0534, -0.2091, 0.0017, -0.0484, -0.2091, 0.0039, -0.0484, -0.2085, 0.0017, -0.0473, -0.2076, 0.0017, -0.0461, -0.2085, 0.0039, -0.0473, -0.2076, 0.0039, -0.0461, -0.2066, -0.0, -0.0448, -0.2066, 0.0017, -0.0448, -0.2066, 0.0039, -0.0448, -0.2196, -0.0, -0.0219, -0.2196, -0.0, -0.0201, -0.2196, 0.0017, -0.0219, -0.2196, 0.0017, -0.0201, -0.2196, 0.0039, -0.0219, -0.2196, 0.0039, -0.0201, -0.2183, -0.0, -0.0219, -0.2183, -0.0, -0.0201, -0.2161, -0.0, -0.0219, -0.2161, -0.0, -0.0201, -0.2183, 0.0017, -0.0219, -0.2183, 0.0039, -0.0219, -0.2161, 0.0017, -0.0219, -0.2161, 0.0039, -0.0219, -0.214, -0.0, -0.0219,
-0.214, -0.0, -0.0201, -0.2118, -0.0, -0.0219, -0.2118, -0.0, -0.0201, -0.214, 0.0017, -0.0219, -0.214, 0.0039, -0.0219, -0.2118, 0.0017, -0.0219, -0.2118, 0.0039, -0.0219, -0.2088, -0.0, -0.0219, -0.2088, -0.0, -0.0201, -0.2088, 0.0017, -0.0219, -0.2088, 0.0017, -0.0201, -0.2088, 0.0039, -0.0219, -0.2088, 0.0039, -0.0201, -0.2196, -0.0, -0.0179, -0.2196, -0.0, -0.0157, -0.2196, -0.0, -0.0136, -0.2196, -0.0, -0.0114, -0.2196, 0.0017, -0.0179, -0.2196, 0.0017, -0.0157, -0.2196, 0.0039, -0.0179, -0.2196, 0.0039, -0.0157, -0.2196, 0.0017, -0.0136, -0.2196, 0.0017, -0.0114, -0.2196, 0.0039, -0.0136, -0.2196, 0.0039, -0.0114, -0.2183, -0.0, -0.0179, -0.2183, -0.0, -0.0157, -0.2161, -0.0, -0.0179, -0.2161, -0.0, -0.0157, -0.2183, -0.0, -0.0136,
-0.2183, -0.0, -0.0114, -0.2161, -0.0, -0.0136, -0.2161, -0.0, -0.0114, -0.2196, -0.0, -0.0092, -0.2196, -0.0, -0.0071, -0.2196, -0.0, -0.0049, -0.2196, -0.0, -0.0027, -0.2196, 0.0017, -0.0092, -0.2196, 0.0017, -0.0071, -0.2196, 0.0039, -0.0092, -0.2196, 0.0039, -0.0071, -0.2196, 0.0017, -0.0049, -0.2196, 0.0017, -0.0027, -0.2196, 0.0039, -0.0049, -0.2196, 0.0039, -0.0027, -0.2183, -0.0, -0.0092, -0.2183, -0.0, -0.0071, -0.2161, -0.0, -0.0092, -0.2161, -0.0, -0.0071, -0.2183, -0.0, -0.0049, -0.2183, -0.0, -0.0027, -0.2161, -0.0, -0.0049, -0.2161, -0.0, -0.0027, -0.214, -0.0, -0.0179, -0.214, -0.0, -0.0157, -0.2118, -0.0, -0.0179, -0.2118, -0.0, -0.0157, -0.214, -0.0, -0.0136, -0.214, -0.0, -0.0114, -0.2118, -0.0, -0.0136,
-0.2118, -0.0, -0.0114, -0.2088, -0.0, -0.0179, -0.2088, -0.0, -0.0157, -0.2088, -0.0, -0.0136, -0.2088, -0.0, -0.0114, -0.2088, 0.0017, -0.0179, -0.2088, 0.0017, -0.0157, -0.2088, 0.0039, -0.0179, -0.2088, 0.0039, -0.0157, -0.2088, 0.0017, -0.0136, -0.2088, 0.0017, -0.0114, -0.2088, 0.0039, -0.0136, -0.2088, 0.0039, -0.0114, -0.214, -0.0, -0.0092, -0.214, -0.0, -0.0071, -0.2118, -0.0, -0.0092, -0.2118, -0.0, -0.0071, -0.214, -0.0, -0.0049, -0.214, -0.0, -0.0027, -0.2118, -0.0, -0.0049, -0.2118, -0.0, -0.0027, -0.2088, -0.0, -0.0092, -0.2088, -0.0, -0.0071, -0.2088, -0.0, -0.0049, -0.2088, -0.0, -0.0027, -0.2088, 0.0017, -0.0092, -0.2088, 0.0017, -0.0071, -0.2088, 0.0039, -0.0092, -0.2088, 0.0039, -0.0071, -0.2088, 0.0017, -0.0049,
-0.2088, 0.0017, -0.0027, -0.2088, 0.0039, -0.0049, -0.2088, 0.0039, -0.0027, -0.2069, -0.0, -0.0591, -0.2031, -0.0, -0.0591, -0.206, -0.0, -0.0573, -0.2033, -0.0, -0.0563, -0.2069, 0.0017, -0.0591, -0.2069, 0.0039, -0.0591, -0.2031, 0.0017, -0.0591, -0.2031, 0.0039, -0.0591, -0.206, 0.0017, -0.0573, -0.206, 0.0039, -0.0573, -0.2032, 0.0017, -0.0563, -0.2032, 0.0039, -0.0563, -0.2009, -0.0, -0.0591, -0.1988, -0.0, -0.0591, -0.201, -0.0, -0.0562, -0.2001, -0.0, -0.054, -0.199, -0.0, -0.0562, -0.1995, -0.0, -0.055, -0.2009, 0.0017, -0.0591, -0.2009, 0.0039, -0.0591, -0.1988, 0.0017, -0.0591, -0.1988, 0.0039, -0.0591, -0.2009, 0.0017, -0.0562, -0.2, 0.0017, -0.054, -0.2009, 0.0039, -0.0562, -0.2, 0.0039, -0.054, -0.199, 0.0017, -0.0562,
-0.1995, 0.0017, -0.0551, -0.199, 0.0039, -0.0562, -0.1995, 0.0039, -0.0551, -0.1942, -0.0, -0.0634, -0.1941, 0.0017, -0.0635, -0.1941, 0.0039, -0.0635, -0.1893, -0.0, -0.0668, -0.1923, -0.0, -0.0656, -0.1934, -0.0, -0.0645, -0.1904, -0.0, -0.0667, -0.1891, -0.0, -0.063, -0.1894, 0.0017, -0.0668, -0.1894, 0.0039, -0.0668, -0.1924, 0.0017, -0.0656, -0.1933, 0.0017, -0.0645, -0.1924, 0.0039, -0.0656, -0.1933, 0.0039, -0.0645, -0.1905, 0.0017, -0.0666, -0.1892, 0.0017, -0.0628, -0.1905, 0.0039, -0.0666, -0.1892, 0.0039, -0.0628, -0.1957, -0.0, -0.0605, -0.1961, -0.0, -0.0591, -0.1952, -0.0, -0.0616, -0.1944, -0.0, -0.0591, -0.1966, -0.0, -0.057, -0.1966, -0.0, -0.0548, -0.1944, -0.0, -0.057, -0.1939, -0.0, -0.0546, -0.1956, 0.0017, -0.0605,
-0.1961, 0.0017, -0.0591, -0.1956, 0.0039, -0.0605, -0.1961, 0.0039, -0.0591, -0.1952, 0.0017, -0.0616, -0.1952, 0.0039, -0.0616, -0.1939, 0.0017, -0.0546, -0.1939, 0.0039, -0.0546, -0.1923, -0.0, -0.0613, -0.1915, -0.0, -0.0588, -0.1902, -0.0, -0.0611, -0.1909, -0.0, -0.0599, -0.1927, -0.0, -0.0568, -0.1933, -0.0, -0.0558, -0.1916, 0.0017, -0.0587, -0.1916, 0.0039, -0.0587, -0.1901, 0.0017, -0.0613, -0.1909, 0.0017, -0.0598, -0.1901, 0.0039, -0.0613, -0.1909, 0.0039, -0.0598, -0.1926, 0.0017, -0.057, -0.1933, 0.0017, -0.0557, -0.1926, 0.0039, -0.057, -0.1933, 0.0039, -0.0557, -0.2059, -0.0, -0.0522, -0.205, -0.0, -0.0508, -0.2026, -0.0, -0.05, -0.2041, -0.0, -0.0489, -0.2053, -0.0, -0.0461, -0.2039, -0.0, -0.0484, -0.2031, -0.0, -0.0461,
-0.2059, 0.0017, -0.0523, -0.2049, 0.0017, -0.0505, -0.2059, 0.0039, -0.0523, -0.2049, 0.0039, -0.0505, -0.2026, 0.0017, -0.0501, -0.2026, 0.0039, -0.0501, -0.2043, 0.0017, -0.0492, -0.2043, 0.0039, -0.0492, -0.2039, 0.0017, -0.0484, -0.2039, 0.0039, -0.0484, -0.2007, -0.0, -0.0528, -0.2019, -0.0, -0.0511, -0.1988, -0.0, -0.0526, -0.1988, -0.0, -0.0505, -0.2009, -0.0, -0.0483, -0.2009, -0.0, -0.0461, -0.1978, -0.0, -0.0477, -0.199, -0.0, -0.0458, -0.2008, 0.0017, -0.0526, -0.2018, 0.0017, -0.0512, -0.2008, 0.0039, -0.0526, -0.2018, 0.0039, -0.0512, -0.1979, 0.0017, -0.0476, -0.1989, 0.0017, -0.046, -0.1979, 0.0039, -0.0476, -0.1989, 0.0039, -0.046, -0.2052, -0.0, -0.0436, -0.2026, -0.0, -0.043, -0.2052, 0.0017, -0.0436, -0.2052, 0.0039, -0.0436,
-0.2026, 0.0017, -0.043, -0.2026, 0.0039, -0.043, -0.2006, -0.0, -0.0441, -0.1998, -0.0, -0.045, -0.2006, 0.0017, -0.0441, -0.2006, 0.0039, -0.0441, -0.1998, 0.0017, -0.045, -0.1998, 0.0039, -0.045, -0.1955, -0.0, -0.0519, -0.1965, -0.0, -0.0502, -0.1949, -0.0, -0.0529, -0.1972, -0.0, -0.0488, -0.1956, 0.0017, -0.0517, -0.1963, 0.0017, -0.0505, -0.1956, 0.0039, -0.0517, -0.1963, 0.0039, -0.0505, -0.195, 0.0017, -0.0528, -0.195, 0.0039, -0.0528, -0.1973, 0.0017, -0.0487, -0.1973, 0.0039, -0.0487, -0.1873, -0.0, -0.067, -0.1879, -0.0, -0.0654, -0.1886, -0.0, -0.064, -0.1873, 0.0017, -0.067, -0.1873, 0.0039, -0.067, -0.1878, 0.0017, -0.0656, -0.1886, 0.0017, -0.0639, -0.1878, 0.0039, -0.0656, -0.1886, 0.0039, -0.0639, -0.1686, -0.0, -0.067,
-0.1686, -0.0, -0.0656, -0.1686, -0.0, -0.0635, -0.1686, 0.0017, -0.067, -0.1686, 0.0039, -0.067, -0.1686, 0.0017, -0.0656, -0.1686, 0.0017, -0.0635, -0.1686, 0.0039, -0.0656, -0.1686, 0.0039, -0.0635, -0.1662, -0.0, -0.067, -0.1641, -0.0, -0.067, -0.1662, -0.0, -0.0656, -0.1662, -0.0, -0.0635, -0.1641, -0.0, -0.0656, -0.1647, -0.0, -0.0631, -0.1662, 0.0017, -0.067, -0.1662, 0.0039, -0.067, -0.1641, 0.0017, -0.067, -0.1641, 0.0039, -0.067, -0.1647, 0.0017, -0.0631, -0.1647, 0.0039, -0.0631, -0.1686, -0.0, -0.0613, -0.1686, -0.0, -0.0591, -0.1686, -0.0, -0.057, -0.1686, -0.0, -0.0548, -0.1686, 0.0017, -0.0613, -0.1686, 0.0017, -0.0591, -0.1686, 0.0039, -0.0613, -0.1686, 0.0039, -0.0591, -0.1686, 0.0017, -0.057, -0.1686, 0.0017, -0.0548,
-0.1686, 0.0039, -0.057, -0.1686, 0.0039, -0.0548, -0.1662, -0.0, -0.0613, -0.1662, -0.0, -0.0591, -0.1647, -0.0, -0.0613, -0.1647, -0.0, -0.0591, -0.1662, -0.0, -0.057, -0.1662, -0.0, -0.0548, -0.1647, -0.0, -0.057, -0.1647, -0.0, -0.0548, -0.1647, 0.0017, -0.0613, -0.1647, 0.0017, -0.0591, -0.1647, 0.0039, -0.0613, -0.1647, 0.0039, -0.0591, -0.1647, 0.0017, -0.057, -0.1647, 0.0017, -0.0548, -0.1647, 0.0039, -0.057, -0.1647, 0.0039, -0.0548, -0.1619, -0.0, -0.067, -0.1597, -0.0, -0.067, -0.1619, -0.0, -0.0656, -0.1619, -0.0, -0.0631, -0.1597, -0.0, -0.0656, -0.1597, -0.0, -0.0631, -0.1619, 0.0017, -0.067, -0.1619, 0.0039, -0.067, -0.1597, 0.0017, -0.067, -0.1597, 0.0039, -0.067, -0.1619, 0.0017, -0.0631, -0.1619, 0.0039, -0.0631,
-0.1597, 0.0017, -0.0631, -0.1597, 0.0039, -0.0631, -0.1575, -0.0, -0.067, -0.1554, -0.0, -0.067, -0.1575, -0.0, -0.0656, -0.1576, -0.0, -0.0631, -0.1554, -0.0, -0.0656, -0.1554, -0.0, -0.063, -0.1575, 0.0017, -0.067, -0.1575, 0.0039, -0.067, -0.1554, 0.0017, -0.067, -0.1554, 0.0039, -0.067, -0.1575, 0.0017, -0.0631, -0.1575, 0.0039, -0.0631, -0.1554, 0.0017, -0.063, -0.1554, 0.0039, -0.063, -0.1686, -0.0, -0.0526, -0.1686, -0.0, -0.0505, -0.1686, -0.0, -0.0483, -0.1686, -0.0, -0.0461, -0.1686, 0.0017, -0.0526, -0.1686, 0.0017, -0.0505, -0.1686, 0.0039, -0.0526, -0.1686, 0.0039, -0.0505, -0.1686, 0.0017, -0.0483, -0.1686, 0.0017, -0.0461, -0.1686, 0.0039, -0.0483, -0.1686, 0.0039, -0.0461, -0.1662, -0.0, -0.0526, -0.1662, -0.0, -0.0505,
-0.1647, -0.0, -0.0526, -0.1647, -0.0, -0.0505, -0.1662, -0.0, -0.0483, -0.1662, -0.0, -0.0461, -0.1647, -0.0, -0.0483, -0.1647, -0.0, -0.0454, -0.1647, 0.0017, -0.0526, -0.1647, 0.0017, -0.0505, -0.1647, 0.0039, -0.0526, -0.1647, 0.0039, -0.0505, -0.1647, 0.0017, -0.0483, -0.1647, 0.0017, -0.0454, -0.1647, 0.0039, -0.0483, -0.1647, 0.0039, -0.0454, -0.1686, -0.0, -0.0439, -0.1686, -0.0, -0.0425, -0.1686, 0.0017, -0.0439, -0.1686, 0.0017, -0.0425, -0.1686, 0.0039, -0.0439, -0.1686, 0.0039, -0.0425, -0.1662, -0.0, -0.0439, -0.1662, -0.0, -0.0425, -0.1641, -0.0, -0.0439, -0.1641, -0.0, -0.0425, -0.1662, 0.0017, -0.0425, -0.1662, 0.0039, -0.0425, -0.1641, 0.0017, -0.0425, -0.1641, 0.0039, -0.0425, -0.1619, -0.0, -0.0454, -0.1597, -0.0, -0.0454,
-0.1619, 0.0017, -0.0454, -0.1619, 0.0039, -0.0454, -0.1597, 0.0017, -0.0454, -0.1597, 0.0039, -0.0454, -0.1576, -0.0, -0.0455, -0.1554, -0.0, -0.0456, -0.1575, 0.0017, -0.0455, -0.1575, 0.0039, -0.0455, -0.1554, 0.0017, -0.0457, -0.1554, 0.0039, -0.0457, -0.1619, -0.0, -0.0439, -0.1619, -0.0, -0.0425, -0.1597, -0.0, -0.0439, -0.1597, -0.0, -0.0425, -0.1619, 0.0017, -0.0425, -0.1619, 0.0039, -0.0425, -0.1597, 0.0017, -0.0425, -0.1597, 0.0039, -0.0425, -0.1575, -0.0, -0.0439, -0.1575, -0.0, -0.0425, -0.1554, -0.0, -0.0439, -0.1554, -0.0, -0.0425, -0.1575, 0.0017, -0.0425, -0.1575, 0.0039, -0.0425, -0.1554, 0.0017, -0.0425, -0.1554, 0.0039, -0.0425, -0.1537, -0.0, -0.0669, -0.1529, -0.0, -0.0668, -0.1533, -0.0, -0.0629, -0.1507, -0.0, -0.0664,
-0.1513, -0.0, -0.0624, -0.1535, 0.0017, -0.0669, -0.1535, 0.0039, -0.0669, -0.1524, 0.0017, -0.0667, -0.1532, 0.0017, -0.0629, -0.1524, 0.0039, -0.0667, -0.1532, 0.0039, -0.0629, -0.151, 0.0017, -0.0665, -0.1516, 0.0017, -0.0625, -0.151, 0.0039, -0.0665, -0.1516, 0.0039, -0.0625, -0.1486, -0.0, -0.0655, -0.1489, -0.0, -0.0635, -0.1475, -0.0, -0.0648, -0.1463, -0.0, -0.064, -0.1488, 0.0017, -0.0656, -0.1488, 0.0039, -0.0656, -0.1475, 0.0017, -0.0648, -0.1463, 0.0017, -0.064, -0.1475, 0.0039, -0.0648, -0.1463, 0.0039, -0.064, -0.1504, -0.0, -0.0621, -0.1504, 0.0017, -0.0621, -0.1504, 0.0039, -0.0621, -0.149, -0.0, -0.0612, -0.148, -0.0, -0.06, -0.1467, -0.0, -0.0613, -0.1472, -0.0, -0.059, -0.1464, -0.0, -0.0569, -0.1462, -0.0, -0.0548,
-0.1489, 0.0017, -0.0611, -0.148, 0.0017, -0.06, -0.1489, 0.0039, -0.0611, -0.148, 0.0039, -0.06, -0.1472, 0.0017, -0.059, -0.1472, 0.0039, -0.059, -0.1465, 0.0017, -0.057, -0.1462, 0.0017, -0.0548, -0.1465, 0.0039, -0.057, -0.1462, 0.0039, -0.0548, -0.1452, -0.0, -0.0628, -0.1452, 0.0017, -0.0628, -0.1452, 0.0039, -0.0628, -0.1441, -0.0, -0.0615, -0.1445, -0.0, -0.0591, -0.1434, -0.0, -0.0603, -0.1429, -0.0, -0.0593, -0.1445, -0.0, -0.057, -0.1445, -0.0, -0.0548, -0.1423, -0.0, -0.0571, -0.1422, -0.0, -0.0548, -0.144, 0.0017, -0.0614, -0.144, 0.0039, -0.0614, -0.1434, 0.0017, -0.0603, -0.1428, 0.0017, -0.0592, -0.1434, 0.0039, -0.0603, -0.1428, 0.0039, -0.0592, -0.1423, 0.0017, -0.057, -0.1422, 0.0017, -0.0548, -0.1423, 0.0039, -0.057,
-0.1422, 0.0039, -0.0548, -0.1533, -0.0, -0.0461, -0.1502, -0.0, -0.0474, -0.1512, -0.0, -0.0469, -0.1532, 0.0017, -0.0461, -0.1532, 0.0039, -0.0461, -0.1503, 0.0017, -0.0474, -0.1514, 0.0017, -0.0468, -0.1503, 0.0039, -0.0474, -0.1514, 0.0039, -0.0468, -0.1479, -0.0, -0.0495, -0.1464, -0.0, -0.0527, -0.1471, -0.0, -0.0505, -0.1489, -0.0, -0.0484, -0.1489, -0.0, -0.0461, -0.1467, -0.0, -0.0483, -0.1458, -0.0, -0.0451, -0.1479, 0.0017, -0.0495, -0.1479, 0.0039, -0.0495, -0.1464, 0.0017, -0.0526, -0.1469, 0.0017, -0.0509, -0.1464, 0.0039, -0.0526, -0.1469, 0.0039, -0.0509, -0.1489, 0.0017, -0.0484, -0.1489, 0.0039, -0.0484, -0.1458, 0.0017, -0.0452, -0.1458, 0.0039, -0.0452, -0.1532, -0.0, -0.0439, -0.1532, -0.0, -0.0425, -0.1506, -0.0, -0.0428,
-0.1515, -0.0, -0.0426, -0.1532, 0.0017, -0.0425, -0.1532, 0.0039, -0.0425, -0.1501, 0.0017, -0.0429, -0.1511, 0.0017, -0.0426, -0.1501, 0.0039, -0.0429, -0.1511, 0.0039, -0.0426, -0.1485, -0.0, -0.0433, -0.1469, -0.0, -0.0443, -0.1488, 0.0017, -0.0432, -0.1488, 0.0039, -0.0432, -0.1468, 0.0017, -0.0443, -0.1468, 0.0039, -0.0443, -0.1445, -0.0, -0.0526, -0.1445, -0.0, -0.0505, -0.1422, -0.0, -0.0526, -0.1425, -0.0, -0.0503, -0.1436, -0.0, -0.0478, -0.1447, -0.0, -0.0462, -0.143, -0.0, -0.0489, -0.1422, 0.0017, -0.0526, -0.1425, 0.0017, -0.0504, -0.1422, 0.0039, -0.0526, -0.1425, 0.0039, -0.0504, -0.1437, 0.0017, -0.0477, -0.1447, 0.0017, -0.0462, -0.1437, 0.0039, -0.0477, -0.1447, 0.0039, -0.0462, -0.1431, 0.0017, -0.0487, -0.1431, 0.0039, -0.0487,
-0.151, -0.0, -0.0219, -0.151, -0.0, -0.0201, -0.151, 0.0017, -0.0219, -0.151, 0.0017, -0.0201, -0.151, 0.0039, -0.0219, -0.151, 0.0039, -0.0201, -0.1489, -0.0, -0.0219, -0.1489, -0.0, -0.0201, -0.1467, -0.0, -0.0219, -0.1467, -0.0, -0.0201, -0.1489, 0.0017, -0.0219, -0.1489, 0.0039, -0.0219, -0.1467, 0.0017, -0.0219, -0.1467, 0.0039, -0.0219, -0.1445, -0.0, -0.0219, -0.1445, -0.0, -0.0201, -0.1424, -0.0, -0.0219, -0.1424, -0.0, -0.0201, -0.1445, 0.0017, -0.0219, -0.1445, 0.0039, -0.0219, -0.1424, 0.0017, -0.0219, -0.1424, 0.0039, -0.0219, -0.1402, -0.0, -0.0219, -0.1402, -0.0, -0.0201, -0.138, -0.0, -0.0219, -0.138, -0.0, -0.0201, -0.1402, 0.0017, -0.0219, -0.1402, 0.0039, -0.0219, -0.138, 0.0017, -0.0219, -0.138, 0.0039, -0.0219,
-0.151, -0.0, -0.0179, -0.151, -0.0, -0.0157, -0.151, -0.0, -0.0136, -0.151, -0.0, -0.0114, -0.151, 0.0017, -0.0179, -0.151, 0.0017, -0.0157, -0.151, 0.0039, -0.0179, -0.151, 0.0039, -0.0157, -0.151, 0.0017, -0.0136, -0.151, 0.0017, -0.0114, -0.151, 0.0039, -0.0136, -0.151, 0.0039, -0.0114, -0.1489, -0.0, -0.0179, -0.1489, -0.0, -0.0157, -0.1467, -0.0, -0.0179, -0.1467, -0.0, -0.0157, -0.1489, -0.0, -0.0136, -0.1489, -0.0, -0.0114, -0.1467, -0.0, -0.0136, -0.1467, -0.0, -0.0114, -0.151, -0.0, -0.0092, -0.151, -0.0, -0.0071, -0.151, -0.0, -0.0049, -0.151, -0.0, -0.0027, -0.151, 0.0017, -0.0092, -0.151, 0.0017, -0.0071, -0.151, 0.0039, -0.0092, -0.151, 0.0039, -0.0071, -0.151, 0.0017, -0.0049, -0.151, 0.0017, -0.0027,
-0.151, 0.0039, -0.0049, -0.151, 0.0039, -0.0027, -0.1489, -0.0, -0.0092, -0.1489, -0.0, -0.0071, -0.1467, -0.0, -0.0092, -0.1467, -0.0, -0.0071, -0.1489, -0.0, -0.0049, -0.1489, -0.0, -0.0027, -0.1467, -0.0, -0.0049, -0.1467, -0.0, -0.0027, -0.1445, -0.0, -0.0179, -0.1445, -0.0, -0.0157, -0.1424, -0.0, -0.0179, -0.1424, -0.0, -0.0157, -0.1445, -0.0, -0.0136, -0.1445, -0.0, -0.0114, -0.1424, -0.0, -0.0136, -0.1424, -0.0, -0.0114, -0.1402, -0.0, -0.0179, -0.1402, -0.0, -0.0157, -0.138, -0.0, -0.0179, -0.138, -0.0, -0.0157, -0.1402, -0.0, -0.0136, -0.1402, -0.0, -0.0114, -0.138, -0.0, -0.0136, -0.138, -0.0, -0.0114, -0.1445, -0.0, -0.0092, -0.1445, -0.0, -0.0071, -0.1424, -0.0, -0.0092, -0.1424, -0.0, -0.0071,
-0.1445, -0.0, -0.0049, -0.1445, -0.0, -0.0027, -0.1424, -0.0, -0.0049, -0.1424, -0.0, -0.0027, -0.1402, -0.0, -0.0092, -0.1402, -0.0, -0.0071, -0.1375, -0.0, -0.0087, -0.1387, -0.0, -0.0072, -0.1391, -0.0, -0.0048, -0.1392, -0.0, -0.0027, -0.139, -0.0, -0.0056, -0.1376, 0.0017, -0.0086, -0.1387, 0.0017, -0.0072, -0.1376, 0.0039, -0.0086, -0.1387, 0.0039, -0.0072, -0.1391, 0.0017, -0.0043, -0.1392, 0.0017, -0.0027, -0.1391, 0.0039, -0.0043, -0.1392, 0.0039, -0.0027, -0.139, 0.0017, -0.0054, -0.139, 0.0039, -0.0054, -0.1225, -0.0, -0.067, -0.1207, -0.0, -0.067, -0.1225, -0.0, -0.0656, -0.1225, -0.0, -0.0635, -0.1207, -0.0, -0.0656, -0.1207, -0.0, -0.0635, -0.1225, 0.0017, -0.067, -0.1225, 0.0039, -0.067, -0.1207, 0.0017, -0.067,
-0.1207, 0.0039, -0.067, -0.1225, 0.0017, -0.0656, -0.1225, 0.0017, -0.0635, -0.1225, 0.0039, -0.0656, -0.1225, 0.0039, -0.0635, -0.1225, -0.0, -0.0613, -0.1225, -0.0, -0.0591, -0.1207, -0.0, -0.0613, -0.1207, -0.0, -0.0591, -0.1225, -0.0, -0.057, -0.1225, -0.0, -0.0548, -0.1207, -0.0, -0.057, -0.1207, -0.0, -0.0548, -0.1225, 0.0017, -0.0613, -0.1225, 0.0017, -0.0591, -0.1225, 0.0039, -0.0613, -0.1225, 0.0039, -0.0591, -0.1225, 0.0017, -0.057, -0.1225, 0.0017, -0.0548, -0.1225, 0.0039, -0.057, -0.1225, 0.0039, -0.0548, -0.1225, -0.0, -0.0526, -0.1225, -0.0, -0.0505, -0.1207, -0.0, -0.0526, -0.1207, -0.0, -0.0505, -0.1225, -0.0, -0.0483, -0.1225, -0.0, -0.0461, -0.1207, -0.0, -0.0483, -0.1207, -0.0, -0.0461, -0.1225, 0.0017, -0.0526,
-0.1225, 0.0017, -0.0505, -0.1225, 0.0039, -0.0526, -0.1225, 0.0039, -0.0505, -0.1225, 0.0017, -0.0483, -0.1225, 0.0017, -0.0461, -0.1225, 0.0039, -0.0483, -0.1225, 0.0039, -0.0461, -0.1225, -0.0, -0.0439, -0.1225, -0.0, -0.0425, -0.1207, -0.0, -0.0439, -0.1207, -0.0, -0.0425, -0.1225, 0.0017, -0.0439, -0.1225, 0.0017, -0.0425, -0.1225, 0.0039, -0.0439, -0.1225, 0.0039, -0.0425, -0.1207, 0.0017, -0.0425, -0.1207, 0.0039, -0.0425, -0.1185, -0.0, -0.067, -0.1163, -0.0, -0.067, -0.1185, -0.0, -0.0656, -0.1186, -0.0, -0.0631, -0.1163, -0.0, -0.0656, -0.1163, -0.0, -0.0631, -0.1185, 0.0017, -0.067, -0.1185, 0.0039, -0.067, -0.1163, 0.0017, -0.067, -0.1163, 0.0039, -0.067, -0.1186, 0.0017, -0.0631, -0.1186, 0.0039, -0.0631, -0.1163, 0.0017, -0.0631,
-0.1163, 0.0039, -0.0631, -0.1141, -0.0, -0.067, -0.112, -0.0, -0.067, -0.1141, -0.0, -0.0656, -0.1141, -0.0, -0.0631, -0.112, -0.0, -0.0656, -0.112, -0.0, -0.063, -0.1141, 0.0017, -0.067, -0.1141, 0.0039, -0.067, -0.112, 0.0017, -0.067, -0.112, 0.0039, -0.067, -0.1141, 0.0017, -0.0631, -0.1141, 0.0039, -0.0631, -0.112, 0.0017, -0.063, -0.112, 0.0039, -0.063, -0.1186, -0.0, -0.0613, -0.1186, -0.0, -0.0591, -0.1186, -0.0, -0.057, -0.1186, -0.0, -0.0548, -0.1186, 0.0017, -0.0613, -0.1186, 0.0017, -0.0591, -0.1186, 0.0039, -0.0613, -0.1186, 0.0039, -0.0591, -0.1186, 0.0017, -0.057, -0.1186, 0.0017, -0.0548, -0.1186, 0.0039, -0.057, -0.1186, 0.0039, -0.0548, -0.1098, -0.0, -0.067, -0.108, -0.0, -0.0669, -0.1098, -0.0, -0.0656,
-0.1098, -0.0, -0.063, -0.1072, -0.0, -0.0669, -0.1078, -0.0, -0.0629, -0.1098, 0.0017, -0.067, -0.1098, 0.0039, -0.067, -0.1076, 0.0017, -0.0669, -0.1076, 0.0039, -0.0669, -0.1098, 0.0017, -0.063, -0.1098, 0.0039, -0.063, -0.1066, 0.0017, -0.0667, -0.1076, 0.0017, -0.0629, -0.1066, 0.0039, -0.0667, -0.1076, 0.0039, -0.0629, -0.1052, -0.0, -0.0665, -0.1058, -0.0, -0.0624, -0.1031, -0.0, -0.0658, -0.1033, -0.0, -0.0635, -0.1054, 0.0017, -0.0666, -0.1061, 0.0017, -0.0625, -0.1054, 0.0039, -0.0666, -0.1061, 0.0039, -0.0625, -0.1033, 0.0017, -0.0658, -0.1033, 0.0039, -0.0658, -0.1048, -0.0, -0.0622, -0.1034, -0.0, -0.0616, -0.105, 0.0017, -0.0622, -0.105, 0.0039, -0.0622, -0.103, 0.0017, -0.0613, -0.103, 0.0039, -0.0613, -0.1186, -0.0, -0.0526,
-0.1186, -0.0, -0.0505, -0.1186, -0.0, -0.0483, -0.1186, -0.0, -0.0454, -0.1163, -0.0, -0.0454, -0.1186, 0.0017, -0.0526, -0.1186, 0.0017, -0.0505, -0.1186, 0.0039, -0.0526, -0.1186, 0.0039, -0.0505, -0.1186, 0.0017, -0.0483, -0.1186, 0.0017, -0.0454, -0.1186, 0.0039, -0.0483, -0.1186, 0.0039, -0.0454, -0.1163, 0.0017, -0.0454, -0.1163, 0.0039, -0.0454, -0.1141, -0.0, -0.0454, -0.112, -0.0, -0.0454, -0.1141, 0.0017, -0.0454, -0.1141, 0.0039, -0.0454, -0.112, 0.0017, -0.0454, -0.112, 0.0039, -0.0454, -0.1185, -0.0, -0.0439, -0.1185, -0.0, -0.0425, -0.1163, -0.0, -0.0439, -0.1163, -0.0, -0.0425, -0.1185, 0.0017, -0.0425, -0.1185, 0.0039, -0.0425, -0.1163, 0.0017, -0.0425, -0.1163, 0.0039, -0.0425, -0.1141, -0.0, -0.0439, -0.1141, -0.0, -0.0425,
-0.112, -0.0, -0.0439, -0.112, -0.0, -0.0425, -0.1141, 0.0017, -0.0425, -0.1141, 0.0039, -0.0425, -0.112, 0.0017, -0.0425, -0.112, 0.0039, -0.0425, -0.1098, -0.0, -0.0455, -0.1077, -0.0, -0.0459, -0.1098, 0.0017, -0.0455, -0.1098, 0.0039, -0.0455, -0.1076, 0.0017, -0.0459, -0.1076, 0.0039, -0.0459, -0.1057, -0.0, -0.0464, -0.103, -0.0, -0.0479, -0.104, -0.0, -0.0473, -0.1054, 0.0017, -0.0465, -0.1054, 0.0039, -0.0465, -0.1031, 0.0017, -0.0478, -0.1043, 0.0017, -0.0471, -0.1031, 0.0039, -0.0478, -0.1043, 0.0039, -0.0471, -0.1098, -0.0, -0.0439, -0.1098, -0.0, -0.0425, -0.1076, -0.0, -0.0439, -0.1076, -0.0, -0.0425, -0.1098, 0.0017, -0.0425, -0.1098, 0.0039, -0.0425, -0.1076, 0.0017, -0.0425, -0.1076, 0.0039, -0.0425, -0.1055, -0.0, -0.0439,
-0.1054, -0.0, -0.0426, -0.1032, -0.0, -0.0431, -0.1043, -0.0, -0.0428, -0.1054, 0.0017, -0.0426, -0.1054, 0.0039, -0.0426, -0.1031, 0.0017, -0.0431, -0.1042, 0.0017, -0.0428, -0.1031, 0.0039, -0.0431, -0.1042, 0.0039, -0.0428, -0.1358, -0.0, -0.0219, -0.1358, -0.0, -0.0201, -0.1337, -0.0, -0.0219, -0.1337, -0.0, -0.0201, -0.1358, 0.0017, -0.0219, -0.1358, 0.0039, -0.0219, -0.1337, 0.0017, -0.0219, -0.1337, 0.0039, -0.0219, -0.1315, -0.0, -0.0219, -0.1315, -0.0, -0.0201, -0.1293, -0.0, -0.0219, -0.1293, -0.0, -0.0201, -0.1315, 0.0017, -0.0219, -0.1315, 0.0039, -0.0219, -0.1293, 0.0017, -0.0219, -0.1293, 0.0039, -0.0219, -0.1272, -0.0, -0.0219, -0.1272, -0.0, -0.0201, -0.125, -0.0, -0.0219, -0.125, -0.0, -0.0201, -0.1272, 0.0017, -0.0219,
-0.1272, 0.0039, -0.0219, -0.125, 0.0017, -0.0219, -0.125, 0.0039, -0.0219, -0.1228, -0.0, -0.0219, -0.1228, -0.0, -0.0201, -0.1207, -0.0, -0.0219, -0.1207, -0.0, -0.0201, -0.1228, 0.0017, -0.0219, -0.1228, 0.0039, -0.0219, -0.1207, 0.0017, -0.0219, -0.1207, 0.0039, -0.0219, -0.1358, -0.0, -0.0179, -0.1358, -0.0, -0.0157, -0.1337, -0.0, -0.0179, -0.1337, -0.0, -0.0157, -0.1358, -0.0, -0.0136, -0.1358, -0.0, -0.0114, -0.1337, -0.0, -0.0136, -0.1337, -0.0, -0.0114, -0.1315, -0.0, -0.0179, -0.1315, -0.0, -0.0157, -0.1293, -0.0, -0.0179, -0.1293, -0.0, -0.0157, -0.1315, -0.0, -0.0136, -0.1315, -0.0, -0.0114, -0.1293, -0.0, -0.0136, -0.1293, -0.0, -0.0114, -0.1358, -0.0, -0.0097, -0.1337, -0.0, -0.01, -0.1358, 0.0017, -0.0097,
-0.1358, 0.0039, -0.0097, -0.1337, 0.0017, -0.01, -0.1337, 0.0039, -0.01, -0.1315, -0.0, -0.0101, -0.1293, -0.0, -0.0101, -0.1315, 0.0017, -0.0101, -0.1315, 0.0039, -0.0101, -0.1293, 0.0017, -0.0101, -0.1293, 0.0039, -0.0101, -0.1272, -0.0, -0.0179, -0.1272, -0.0, -0.0157, -0.125, -0.0, -0.0179, -0.125, -0.0, -0.0157, -0.1272, -0.0, -0.0136, -0.1272, -0.0, -0.0114, -0.125, -0.0, -0.0136, -0.125, -0.0, -0.0114, -0.1228, -0.0, -0.0179, -0.1228, -0.0, -0.0157, -0.1207, -0.0, -0.0179, -0.1207, -0.0, -0.0157, -0.1228, -0.0, -0.0136, -0.1228, -0.0, -0.0114, -0.1207, -0.0, -0.0136, -0.1207, -0.0, -0.0114, -0.1272, -0.0, -0.0101, -0.125, -0.0, -0.0101, -0.1272, 0.0017, -0.0101, -0.1272, 0.0039, -0.0101, -0.125, 0.0017, -0.0101,
-0.125, 0.0039, -0.0101, -0.1228, -0.0, -0.0101, -0.1207, -0.0, -0.0101, -0.1228, 0.0017, -0.0101, -0.1228, 0.0039, -0.0101, -0.1207, 0.0017, -0.0101, -0.1207, 0.0039, -0.0101, -0.1185, -0.0, -0.0219, -0.1185, -0.0, -0.0201, -0.1163, -0.0, -0.0219, -0.1163, -0.0, -0.0201, -0.1185, 0.0017, -0.0219, -0.1185, 0.0039, -0.0219, -0.1163, 0.0017, -0.0219, -0.1163, 0.0039, -0.0219, -0.1141, -0.0, -0.0219, -0.1141, -0.0, -0.0201, -0.112, -0.0, -0.0219, -0.112, -0.0, -0.0201, -0.1141, 0.0017, -0.0219, -0.1141, 0.0039, -0.0219, -0.112, 0.0017, -0.0219, -0.112, 0.0039, -0.0219, -0.1098, -0.0, -0.0219, -0.1098, -0.0, -0.0201, -0.1076, -0.0, -0.0219, -0.1076, -0.0, -0.0201, -0.1098, 0.0017, -0.0219, -0.1098, 0.0039, -0.0219, -0.1076, 0.0017, -0.0219,
-0.1076, 0.0039, -0.0219, -0.1055, -0.0, -0.0219, -0.1055, -0.0, -0.0201, -0.1033, -0.0, -0.0219, -0.1033, -0.0, -0.0201, -0.1055, 0.0017, -0.0219, -0.1055, 0.0039, -0.0219, -0.1033, 0.0017, -0.0219, -0.1033, 0.0039, -0.0219, -0.1185, -0.0, -0.0179, -0.1185, -0.0, -0.0157, -0.1163, -0.0, -0.0179, -0.1163, -0.0, -0.0157, -0.1185, -0.0, -0.0136, -0.1185, -0.0, -0.0114, -0.1163, -0.0, -0.0136, -0.1163, -0.0, -0.0114, -0.1141, -0.0, -0.0179, -0.1141, -0.0, -0.0157, -0.112, -0.0, -0.0179, -0.112, -0.0, -0.0157, -0.1141, -0.0, -0.0136, -0.1141, -0.0, -0.0114, -0.112, -0.0, -0.0136, -0.112, -0.0, -0.0114, -0.1185, -0.0, -0.0101, -0.1163, -0.0, -0.0101, -0.1185, 0.0017, -0.0101, -0.1185, 0.0039, -0.0101, -0.1163, 0.0017, -0.0101,
-0.1163, 0.0039, -0.0101, -0.1141, -0.0, -0.0101, -0.112, -0.0, -0.0101, -0.1141, 0.0017, -0.0101, -0.1141, 0.0039, -0.0101, -0.112, 0.0017, -0.0101, -0.112, 0.0039, -0.0101, -0.1098, -0.0, -0.0179, -0.1098, -0.0, -0.0157, -0.1076, -0.0, -0.0179, -0.1076, -0.0, -0.0157, -0.1098, -0.0, -0.0136, -0.1098, -0.0, -0.0114, -0.1076, -0.0, -0.0136, -0.1076, -0.0, -0.0114, -0.1055, -0.0, -0.0179, -0.1055, -0.0, -0.0157, -0.1033, -0.0, -0.0179, -0.1033, -0.0, -0.0157, -0.1055, -0.0, -0.0136, -0.1055, -0.0, -0.0114, -0.1033, -0.0, -0.0136, -0.1033, -0.0, -0.0114, -0.1098, -0.0, -0.0101, -0.1076, -0.0, -0.0101, -0.1098, 0.0017, -0.0101, -0.1098, 0.0039, -0.0101, -0.1076, 0.0017, -0.0101, -0.1076, 0.0039, -0.0101, -0.1055, -0.0, -0.0101,
-0.1033, -0.0, -0.0101, -0.1055, 0.0017, -0.0101, -0.1055, 0.0039, -0.0101, -0.1033, 0.0017, -0.0101, -0.1033, 0.0039, -0.0101, -0.1017, -0.0, -0.065, -0.1006, -0.0, -0.0642, -0.0994, -0.0, -0.0631, -0.1016, 0.0017, -0.065, -0.1005, 0.0017, -0.0642, -0.1016, 0.0039, -0.065, -0.1005, 0.0039, -0.0642, -0.0994, 0.0017, -0.0631, -0.0994, 0.0039, -0.0631, -0.102, -0.0, -0.0604, -0.1007, -0.0, -0.059, -0.0983, -0.0, -0.0618, -0.099, -0.0, -0.0591, -0.1002, -0.0, -0.0578, -0.0998, -0.0, -0.0569, -0.0995, -0.0, -0.0548, -0.102, 0.0017, -0.0604, -0.1007, 0.0017, -0.0591, -0.102, 0.0039, -0.0604, -0.1007, 0.0039, -0.0591, -0.0983, 0.0017, -0.0618, -0.0983, 0.0039, -0.0618, -0.1002, 0.0017, -0.0577, -0.1002, 0.0039, -0.0577, -0.0997, 0.0017, -0.0566,
-0.0995, 0.0017, -0.0548, -0.0997, 0.0039, -0.0566, -0.0995, 0.0039, -0.0548, -0.0975, -0.0, -0.0606, -0.0968, -0.0, -0.0594, -0.0962, -0.0, -0.0571, -0.0961, -0.0, -0.0548, -0.0976, 0.0017, -0.0607, -0.0968, 0.0017, -0.0592, -0.0976, 0.0039, -0.0607, -0.0968, 0.0039, -0.0592, -0.0962, 0.0017, -0.057, -0.0961, 0.0017, -0.0548, -0.0962, 0.0039, -0.057, -0.0961, 0.0039, -0.0548, -0.1002, -0.0, -0.0518, -0.1006, -0.0, -0.0506, -0.0998, -0.0, -0.0527, -0.099, -0.0, -0.0505, -0.1019, -0.0, -0.049, -0.1011, -0.0, -0.0461, -0.0978, -0.0, -0.0474, -0.099, -0.0, -0.0457, -0.1002, 0.0017, -0.0518, -0.1008, 0.0017, -0.0504, -0.1002, 0.0039, -0.0518, -0.1008, 0.0039, -0.0504, -0.0997, 0.0017, -0.0529, -0.0997, 0.0039, -0.0529, -0.1019, 0.0017, -0.049,
-0.1019, 0.0039, -0.049, -0.0979, 0.0017, -0.0473, -0.0988, 0.0017, -0.046, -0.0979, 0.0039, -0.0473, -0.0988, 0.0039, -0.046, -0.0961, -0.0, -0.0526, -0.0965, -0.0, -0.0503, -0.0971, -0.0, -0.0485, -0.0961, 0.0017, -0.0526, -0.0964, 0.0017, -0.0504, -0.0961, 0.0039, -0.0526, -0.0964, 0.0039, -0.0504, -0.0972, 0.0017, -0.0483, -0.0972, 0.0039, -0.0483, -0.1009, -0.0, -0.0442, -0.1, -0.0, -0.045, -0.101, 0.0017, -0.0441, -0.101, 0.0039, -0.0441, -0.1, 0.0017, -0.045, -0.1, 0.0039, -0.045, -0.0766, -0.0, -0.0587, -0.0771, -0.0, -0.057, -0.0774, -0.0, -0.0548, -0.0766, 0.0017, -0.0589, -0.0766, 0.0039, -0.0589, -0.0771, 0.0017, -0.057, -0.0774, 0.0017, -0.0548, -0.0771, 0.0039, -0.057, -0.0774, 0.0039, -0.0548, -0.0742, -0.0, -0.0626,
-0.072, -0.0, -0.0647, -0.0732, -0.0, -0.0638, -0.0742, 0.0017, -0.0627, -0.0742, 0.0039, -0.0627, -0.072, 0.0017, -0.0647, -0.0731, 0.0017, -0.0638, -0.072, 0.0039, -0.0647, -0.0731, 0.0039, -0.0638, -0.071, -0.0, -0.0654, -0.0707, -0.0, -0.0635, -0.0688, -0.0, -0.0663, -0.0681, -0.0, -0.0623, -0.0708, 0.0017, -0.0655, -0.0708, 0.0039, -0.0655, -0.0686, 0.0017, -0.0664, -0.0676, 0.0017, -0.0624, -0.0686, 0.0039, -0.0664, -0.0676, 0.0039, -0.0624, -0.0751, -0.0, -0.0617, -0.0762, -0.0, -0.0597, -0.0729, -0.0, -0.0613, -0.0722, -0.0, -0.0588, -0.0751, -0.0, -0.057, -0.0751, -0.0, -0.0548, -0.073, -0.0, -0.0568, -0.0733, -0.0, -0.0548, -0.0753, 0.0017, -0.0614, -0.0761, 0.0017, -0.06, -0.0753, 0.0039, -0.0614, -0.0761, 0.0039, -0.06,
-0.0723, 0.0017, -0.0587, -0.0723, 0.0039, -0.0587, -0.0729, 0.0017, -0.057, -0.0733, 0.0017, -0.0548, -0.0729, 0.0039, -0.057, -0.0733, 0.0039, -0.0548, -0.0705, -0.0, -0.061, -0.0714, -0.0, -0.0598, -0.0689, -0.0, -0.0619, -0.0707, 0.0017, -0.0608, -0.0715, 0.0017, -0.0598, -0.0707, 0.0039, -0.0608, -0.0715, 0.0039, -0.0598, -0.0687, 0.0017, -0.062, -0.0687, 0.0039, -0.062, -0.0773, -0.0, -0.0526, -0.0769, -0.0, -0.0503, -0.0764, -0.0, -0.0491, -0.0773, 0.0017, -0.0526, -0.0769, 0.0017, -0.0504, -0.0773, 0.0039, -0.0526, -0.0769, 0.0039, -0.0504, -0.0763, 0.0017, -0.049, -0.0763, 0.0039, -0.049, -0.0751, -0.0, -0.0526, -0.0751, -0.0, -0.0505, -0.0732, -0.0, -0.0527, -0.0725, -0.0, -0.0505, -0.0758, -0.0, -0.048, -0.0748, -0.0, -0.0464,
-0.0729, -0.0, -0.0483, -0.0736, -0.0, -0.0453, -0.0732, 0.0017, -0.0526, -0.0727, 0.0017, -0.0508, -0.0732, 0.0039, -0.0526, -0.0727, 0.0039, -0.0508, -0.0758, 0.0017, -0.0479, -0.0747, 0.0017, -0.0464, -0.0758, 0.0039, -0.0479, -0.0747, 0.0039, -0.0464, -0.0736, 0.0017, -0.0453, -0.0736, 0.0039, -0.0453, -0.0717, -0.0, -0.0495, -0.0708, -0.0, -0.0483, -0.0707, -0.0, -0.0461, -0.0694, -0.0, -0.0474, -0.0684, -0.0, -0.0468, -0.0717, 0.0017, -0.0495, -0.0717, 0.0039, -0.0495, -0.0707, 0.0017, -0.0482, -0.0707, 0.0039, -0.0482, -0.0694, 0.0017, -0.0474, -0.0683, 0.0017, -0.0468, -0.0694, 0.0039, -0.0474, -0.0683, 0.0039, -0.0468, -0.0725, -0.0, -0.0445, -0.0726, 0.0017, -0.0445, -0.0726, 0.0039, -0.0445, -0.0711, -0.0, -0.0437, -0.0689, -0.0, -0.0429,
-0.0679, -0.0, -0.0427, -0.0708, 0.0017, -0.0435, -0.0708, 0.0039, -0.0435, -0.0692, 0.0017, -0.043, -0.0681, 0.0017, -0.0427, -0.0692, 0.0039, -0.043, -0.0681, 0.0039, -0.0427, -0.1011, -0.0, -0.0219, -0.1011, -0.0, -0.0201, -0.099, -0.0, -0.0219, -0.099, -0.0, -0.0201, -0.1011, 0.0017, -0.0219, -0.1011, 0.0039, -0.0219, -0.099, 0.0017, -0.0219, -0.099, 0.0039, -0.0219, -0.0968, -0.0, -0.0219, -0.0968, -0.0, -0.0201, -0.0946, -0.0, -0.0219, -0.0946, -0.0, -0.0201, -0.0968, 0.0017, -0.0219, -0.0968, 0.0039, -0.0219, -0.0946, 0.0017, -0.0219, -0.0946, 0.0039, -0.0219, -0.0924, -0.0, -0.0219, -0.0924, -0.0, -0.0201, -0.0902, -0.0, -0.0219, -0.0902, -0.0, -0.0201, -0.0924, 0.0017, -0.0219, -0.0924, 0.0039, -0.0219, -0.0902, 0.0017, -0.0219,
-0.0902, 0.0017, -0.0201, -0.0902, 0.0039, -0.0219, -0.0902, 0.0039, -0.0201, -0.1011, -0.0, -0.0179, -0.1011, -0.0, -0.0157, -0.099, -0.0, -0.0179, -0.099, -0.0, -0.0157, -0.1011, -0.0, -0.0136, -0.1011, -0.0, -0.0114, -0.099, -0.0, -0.0136, -0.099, -0.0, -0.0114, -0.0968, -0.0, -0.0179, -0.0968, -0.0, -0.0157, -0.0946, -0.0, -0.0179, -0.0946, -0.0, -0.0157, -0.0968, -0.0, -0.0136, -0.0968, -0.0, -0.0114, -0.0946, -0.0, -0.0136, -0.0946, -0.0, -0.0114, -0.1011, -0.0, -0.0101, -0.099, -0.0, -0.0101, -0.1011, 0.0017, -0.0101, -0.1011, 0.0039, -0.0101, -0.099, 0.0017, -0.0101, -0.099, 0.0039, -0.0101, -0.0968, -0.0, -0.0101, -0.0946, -0.0, -0.0101, -0.0968, 0.0017, -0.0101, -0.0968, 0.0039, -0.0101, -0.0946, 0.0017, -0.0101,
-0.0946, 0.0039, -0.0101, -0.0924, -0.0, -0.0179, -0.0924, -0.0, -0.0157, -0.0902, -0.0, -0.0179, -0.0902, -0.0, -0.0157, -0.0924, -0.0, -0.0136, -0.0924, -0.0, -0.0114, -0.0902, -0.0, -0.0136, -0.0902, -0.0, -0.0114, -0.0902, 0.0017, -0.0179, -0.0902, 0.0017, -0.0157, -0.0902, 0.0039, -0.0179, -0.0902, 0.0039, -0.0157, -0.0902, 0.0017, -0.0136, -0.0902, 0.0017, -0.0114, -0.0902, 0.0039, -0.0136, -0.0902, 0.0039, -0.0114, -0.0924, -0.0, -0.0101, -0.0902, -0.0, -0.0101, -0.0924, 0.0017, -0.0101, -0.0924, 0.0039, -0.0101, -0.0902, 0.0017, -0.0101, -0.0902, 0.0039, -0.0101, -0.0775, -0.0, -0.0219, -0.0775, -0.0, -0.0201, -0.0775, 0.0017, -0.0219, -0.0775, 0.0017, -0.0201, -0.0775, 0.0039, -0.0219, -0.0775, 0.0039, -0.0201, -0.0751, -0.0, -0.0219,
-0.0751, -0.0, -0.0201, -0.0729, -0.0, -0.0219, -0.0729, -0.0, -0.0201, -0.0751, 0.0017, -0.0219, -0.0751, 0.0039, -0.0219, -0.0729, 0.0017, -0.0219, -0.0729, 0.0039, -0.0219, -0.0707, -0.0, -0.0219, -0.0707, -0.0, -0.0201, -0.0686, -0.0, -0.0219, -0.0686, -0.0, -0.0201, -0.0707, 0.0017, -0.0219, -0.0707, 0.0039, -0.0219, -0.0686, 0.0017, -0.0219, -0.0686, 0.0039, -0.0219, -0.0775, -0.0, -0.0179, -0.0775, -0.0, -0.0157, -0.0775, -0.0, -0.0136, -0.0775, -0.0, -0.0114, -0.0775, 0.0017, -0.0179, -0.0775, 0.0017, -0.0157, -0.0775, 0.0039, -0.0179, -0.0775, 0.0039, -0.0157, -0.0775, 0.0017, -0.0136, -0.0775, 0.0017, -0.0114, -0.0775, 0.0039, -0.0136, -0.0775, 0.0039, -0.0114, -0.0775, -0.0, -0.0092, -0.0775, -0.0, -0.0071, -0.0775, -0.0, -0.0049,
-0.0775, -0.0, -0.0027, -0.0775, 0.0017, -0.0092, -0.0775, 0.0017, -0.0071, -0.0775, 0.0039, -0.0092, -0.0775, 0.0039, -0.0071, -0.0775, 0.0017, -0.0049, -0.0775, 0.0017, -0.0027, -0.0775, 0.0039, -0.0049, -0.0775, 0.0039, -0.0027, -0.0751, -0.0, -0.0179, -0.0751, -0.0, -0.0157, -0.0729, -0.0, -0.0179, -0.0729, -0.0, -0.0157, -0.0751, -0.0, -0.0136, -0.0751, -0.0, -0.0114, -0.0729, -0.0, -0.0136, -0.0729, -0.0, -0.0114, -0.0707, -0.0, -0.0179, -0.0707, -0.0, -0.0157, -0.0686, -0.0, -0.0179, -0.0686, -0.0, -0.0157, -0.0707, -0.0, -0.0136, -0.0707, -0.0, -0.0114, -0.0686, -0.0, -0.0136, -0.0686, -0.0, -0.0114, -0.0751, -0.0, -0.0092, -0.0751, -0.0, -0.0071, -0.0729, -0.0, -0.0092, -0.0729, -0.0, -0.0071, -0.0751, -0.0, -0.0049,
-0.0751, -0.0, -0.0027, -0.0729, -0.0, -0.0049, -0.0729, -0.0, -0.0027, -0.0707, -0.0, -0.0092, -0.0707, -0.0, -0.0071, -0.0686, -0.0, -0.0092, -0.0686, -0.0, -0.0071, -0.0707, -0.0, -0.0049, -0.0707, -0.0, -0.0027, -0.0686, -0.0, -0.0049, -0.0686, -0.0, -0.0027, -0.0659, -0.0, -0.0669, -0.0642, -0.0, -0.067, -0.0667, -0.0, -0.0668, -0.0663, -0.0, -0.0628, -0.0642, -0.0, -0.0656, -0.0642, -0.0, -0.063, -0.0661, 0.0017, -0.0669, -0.0661, 0.0039, -0.0669, -0.0642, 0.0017, -0.067, -0.0642, 0.0039, -0.067, -0.0672, 0.0017, -0.0667, -0.0664, 0.0017, -0.0627, -0.0672, 0.0039, -0.0667, -0.0664, 0.0039, -0.0627, -0.0642, 0.0017, -0.063, -0.0642, 0.0039, -0.063, -0.0621, -0.0, -0.067, -0.0599, -0.0, -0.067, -0.0621, -0.0, -0.0656,
-0.062, -0.0, -0.063, -0.0599, -0.0, -0.0656, -0.0599, -0.0, -0.0631, -0.0621, 0.0017, -0.067, -0.0621, 0.0039, -0.067, -0.0599, 0.0017, -0.067, -0.0599, 0.0039, -0.067, -0.0621, 0.0017, -0.063, -0.0621, 0.0039, -0.063, -0.0599, 0.0017, -0.0631, -0.0599, 0.0039, -0.0631, -0.0577, -0.0, -0.0669, -0.0566, -0.0, -0.0668, -0.0577, -0.0, -0.0656, -0.0578, -0.0, -0.063, -0.0555, -0.0, -0.0666, -0.0557, -0.0, -0.0625, -0.0577, 0.0017, -0.0669, -0.0577, 0.0039, -0.0669, -0.0565, 0.0017, -0.0667, -0.0565, 0.0039, -0.0667, -0.0577, 0.0017, -0.063, -0.0577, 0.0039, -0.063, -0.0554, 0.0017, -0.0666, -0.0559, 0.0017, -0.0626, -0.0554, 0.0039, -0.0666, -0.0559, 0.0039, -0.0626, -0.0531, -0.0, -0.0659, -0.0534, -0.0, -0.0635, -0.0515, -0.0, -0.0652,
-0.0505, -0.0, -0.0646, -0.0534, 0.0017, -0.066, -0.0534, 0.0039, -0.066, -0.0514, 0.0017, -0.0651, -0.0503, 0.0017, -0.0645, -0.0514, 0.0039, -0.0651, -0.0503, 0.0039, -0.0645, -0.0547, -0.0, -0.0623, -0.0548, 0.0017, -0.0623, -0.0548, 0.0039, -0.0623, -0.0535, -0.0, -0.0617, -0.052, -0.0, -0.0605, -0.0507, -0.0, -0.059, -0.0502, -0.0, -0.058, -0.0533, 0.0017, -0.0615, -0.0533, 0.0039, -0.0615, -0.052, 0.0017, -0.0605, -0.0509, 0.0017, -0.0592, -0.052, 0.0039, -0.0605, -0.0509, 0.0039, -0.0592, -0.0502, 0.0017, -0.058, -0.0502, 0.0039, -0.058, -0.0663, -0.0, -0.0461, -0.0642, -0.0, -0.0456, -0.0664, 0.0017, -0.0461, -0.0664, 0.0039, -0.0461, -0.0642, 0.0017, -0.0456, -0.0642, 0.0039, -0.0456, -0.0621, -0.0, -0.0454, -0.0599, -0.0, -0.0455,
-0.0621, 0.0017, -0.0454, -0.0621, 0.0039, -0.0454, -0.0599, 0.0017, -0.0455, -0.0599, 0.0039, -0.0455, -0.0664, -0.0, -0.0439, -0.0664, -0.0, -0.0425, -0.0642, -0.0, -0.0439, -0.0642, -0.0, -0.0425, -0.0664, 0.0017, -0.0425, -0.0664, 0.0039, -0.0425, -0.0642, 0.0017, -0.0425, -0.0642, 0.0039, -0.0425, -0.0621, -0.0, -0.0439, -0.0621, -0.0, -0.0425, -0.0599, -0.0, -0.0439, -0.0599, -0.0, -0.0425, -0.0621, 0.0017, -0.0425, -0.0621, 0.0039, -0.0425, -0.0599, 0.0017, -0.0425, -0.0599, 0.0039, -0.0425, -0.0578, -0.0, -0.0457, -0.0557, -0.0, -0.0463, -0.0577, 0.0017, -0.0458, -0.0577, 0.0039, -0.0458, -0.0555, 0.0017, -0.0463, -0.0555, 0.0039, -0.0463, -0.0507, -0.0, -0.0501, -0.0528, -0.0, -0.0477, -0.0538, -0.0, -0.0471, -0.0517, -0.0, -0.0487,
-0.0512, -0.0, -0.0461, -0.0506, 0.0017, -0.0502, -0.0506, 0.0039, -0.0502, -0.0529, 0.0017, -0.0476, -0.054, 0.0017, -0.047, -0.0529, 0.0039, -0.0476, -0.054, 0.0039, -0.047, -0.0517, 0.0017, -0.0487, -0.0517, 0.0039, -0.0487, -0.0577, -0.0, -0.0439, -0.0577, -0.0, -0.0425, -0.0555, -0.0, -0.0439, -0.0555, -0.0, -0.0426, -0.0577, 0.0017, -0.0425, -0.0577, 0.0039, -0.0425, -0.0555, 0.0017, -0.0426, -0.0555, 0.0039, -0.0426, -0.0533, -0.0, -0.0431, -0.0544, -0.0, -0.0428, -0.0509, -0.0, -0.0442, -0.0532, 0.0017, -0.0431, -0.0544, 0.0017, -0.0428, -0.0532, 0.0039, -0.0431, -0.0544, 0.0039, -0.0428, -0.0511, 0.0017, -0.044, -0.0511, 0.0039, -0.044, -0.049, -0.0, -0.0636, -0.049, 0.0017, -0.0635, -0.049, 0.0039, -0.0635, -0.048, -0.0, -0.0622,
-0.049, -0.0, -0.0591, -0.0472, -0.0, -0.0611, -0.0465, -0.0, -0.0594, -0.0497, -0.0, -0.0569, -0.0492, -0.0, -0.0548, -0.0461, -0.0, -0.0571, -0.0461, -0.0, -0.0548, -0.0481, 0.0017, -0.0623, -0.0481, 0.0039, -0.0623, -0.0473, 0.0017, -0.0613, -0.0465, 0.0017, -0.0596, -0.0473, 0.0039, -0.0613, -0.0465, 0.0039, -0.0596, -0.0497, 0.0017, -0.0569, -0.0492, 0.0017, -0.0547, -0.0497, 0.0039, -0.0569, -0.0492, 0.0039, -0.0547, -0.0461, 0.0017, -0.057, -0.0461, 0.0017, -0.0548, -0.0461, 0.0039, -0.057, -0.0461, 0.0039, -0.0548, -0.0495, -0.0, -0.0528, -0.0502, -0.0, -0.051, -0.0461, -0.0, -0.0526, -0.0464, -0.0, -0.0503, -0.049, -0.0, -0.0483, -0.0489, -0.0, -0.0459, -0.0471, -0.0, -0.048, -0.0479, -0.0, -0.0471, -0.0496, 0.0017, -0.0526,
-0.05, 0.0017, -0.0513, -0.0496, 0.0039, -0.0526, -0.05, 0.0039, -0.0513, -0.0461, 0.0017, -0.0526, -0.0463, 0.0017, -0.0504, -0.0461, 0.0039, -0.0526, -0.0463, 0.0039, -0.0504, -0.0488, 0.0017, -0.0459, -0.0488, 0.0039, -0.0459, -0.047, 0.0017, -0.0482, -0.0479, 0.0017, -0.0471, -0.047, 0.0039, -0.0482, -0.0479, 0.0039, -0.0471, -0.05, -0.0, -0.0449, -0.05, 0.0017, -0.045, -0.05, 0.0039, -0.045, -0.0667, -0.0, -0.0219, -0.0667, -0.0, -0.0201, -0.0667, 0.0017, -0.0219, -0.0667, 0.0017, -0.0201, -0.0667, 0.0039, -0.0219, -0.0667, 0.0039, -0.0201, -0.0667, -0.0, -0.0179, -0.0667, -0.0, -0.0157, -0.0667, -0.0, -0.0136, -0.0667, -0.0, -0.0114, -0.0667, 0.0017, -0.0179, -0.0667, 0.0017, -0.0157, -0.0667, 0.0039, -0.0179, -0.0667, 0.0039, -0.0157,
-0.0667, 0.0017, -0.0136, -0.0667, 0.0017, -0.0114, -0.0667, 0.0039, -0.0136, -0.0667, 0.0039, -0.0114, -0.0667, -0.0, -0.0092, -0.0667, -0.0, -0.0071, -0.0667, -0.0, -0.0049, -0.0667, -0.0, -0.0027, -0.0667, 0.0017, -0.0092, -0.0667, 0.0017, -0.0071, -0.0667, 0.0039, -0.0092, -0.0667, 0.0039, -0.0071, -0.0667, 0.0017, -0.0049, -0.0667, 0.0017, -0.0027, -0.0667, 0.0039, -0.0049, -0.0667, 0.0039, -0.0027, -0.05, -0.0, -0.0219, -0.05, -0.0, -0.0201, -0.0469, -0.0, -0.0219, -0.0469, -0.0, -0.0201, -0.05, 0.0017, -0.0219, -0.05, 0.0017, -0.0201, -0.05, 0.0039, -0.0219, -0.05, 0.0039, -0.0201, -0.0469, 0.0017, -0.0219, -0.0469, 0.0039, -0.0219, -0.0447, -0.0, -0.0219, -0.0447, -0.0, -0.0201, -0.0425, -0.0, -0.0219, -0.0425, -0.0, -0.0201,
-0.0447, 0.0017, -0.0219, -0.0447, 0.0039, -0.0219, -0.0425, 0.0017, -0.0219, -0.0425, 0.0039, -0.0219, -0.0404, -0.0, -0.0219, -0.0404, -0.0, -0.0201, -0.0382, -0.0, -0.0219, -0.0382, -0.0, -0.0201, -0.0404, 0.0017, -0.0219, -0.0404, 0.0039, -0.0219, -0.0382, 0.0017, -0.0219, -0.0382, 0.0017, -0.0201, -0.0382, 0.0039, -0.0219, -0.0382, 0.0039, -0.0201, -0.05, -0.0, -0.0179, -0.05, -0.0, -0.0157, -0.0469, -0.0, -0.0179, -0.0469, -0.0, -0.0157, -0.05, -0.0, -0.0136, -0.05, -0.0, -0.0114, -0.0469, -0.0, -0.0136, -0.0469, -0.0, -0.0114, -0.05, 0.0017, -0.0179, -0.05, 0.0017, -0.0157, -0.05, 0.0039, -0.0179, -0.05, 0.0039, -0.0157, -0.05, 0.0017, -0.0136, -0.05, 0.0017, -0.0114, -0.05, 0.0039, -0.0136, -0.05, 0.0039, -0.0114,
-0.0447, -0.0, -0.0179, -0.0447, -0.0, -0.0157, -0.0425, -0.0, -0.0179, -0.0425, -0.0, -0.0157, -0.0447, -0.0, -0.0136, -0.0447, -0.0, -0.0114, -0.0425, -0.0, -0.0136, -0.0425, -0.0, -0.0114, -0.05, -0.0, -0.0092, -0.05, -0.0, -0.0071, -0.0469, -0.0, -0.0092, -0.0469, -0.0, -0.0071, -0.05, -0.0, -0.0049, -0.05, -0.0, -0.0027, -0.0469, -0.0, -0.0049, -0.0469, -0.0, -0.0027, -0.05, 0.0017, -0.0092, -0.05, 0.0017, -0.0071, -0.05, 0.0039, -0.0092, -0.05, 0.0039, -0.0071, -0.05, 0.0017, -0.0049, -0.05, 0.0017, -0.0027, -0.05, 0.0039, -0.0049, -0.05, 0.0039, -0.0027, -0.0447, -0.0, -0.0092, -0.0447, -0.0, -0.0071, -0.0425, -0.0, -0.0092, -0.0425, -0.0, -0.0071, -0.0447, -0.0, -0.0049, -0.0447, -0.0, -0.0027,
-0.0425, -0.0, -0.0049, -0.0425, -0.0, -0.0027, -0.0404, -0.0, -0.0179, -0.0404, -0.0, -0.0157, -0.0382, -0.0, -0.0179, -0.0382, -0.0, -0.0157, -0.0404, -0.0, -0.0136, -0.0404, -0.0, -0.0114, -0.0382, -0.0, -0.0136, -0.0382, -0.0, -0.0114, -0.0382, 0.0017, -0.0179, -0.0382, 0.0017, -0.0157, -0.0382, 0.0039, -0.0179, -0.0382, 0.0039, -0.0157, -0.0382, 0.0017, -0.0136, -0.0382, 0.0017, -0.0114, -0.0382, 0.0039, -0.0136, -0.0382, 0.0039, -0.0114, -0.0404, -0.0, -0.0092, -0.0404, -0.0, -0.0071, -0.0382, -0.0, -0.0092, -0.0382, -0.0, -0.0071, -0.0404, -0.0, -0.0049, -0.0404, -0.0, -0.0027, -0.0382, -0.0, -0.0049, -0.0382, -0.0, -0.0027, -0.0382, 0.0017, -0.0092, -0.0382, 0.0017, -0.0071, -0.0382, 0.0039, -0.0092, -0.0382, 0.0039, -0.0071,
-0.0382, 0.0017, -0.0049, -0.0382, 0.0017, -0.0027, -0.0382, 0.0039, -0.0049, -0.0382, 0.0039, -0.0027, -0.0265, -0.0, -0.067, -0.0252, -0.0, -0.067, -0.0265, -0.0, -0.0656, -0.0265, -0.0, -0.0635, -0.0252, -0.0, -0.0656, -0.0252, -0.0, -0.0635, -0.0265, 0.0017, -0.067, -0.0265, 0.0039, -0.067, -0.0252, 0.0017, -0.067, -0.0252, 0.0039, -0.067, -0.0265, 0.0017, -0.0656, -0.0265, 0.0017, -0.0635, -0.0265, 0.0039, -0.0656, -0.0265, 0.0039, -0.0635, -0.0265, -0.0, -0.0613, -0.0265, -0.0, -0.0591, -0.0252, -0.0, -0.0613, -0.0252, -0.0, -0.0591, -0.0265, -0.0, -0.057, -0.0265, -0.0, -0.0548, -0.0252, -0.0, -0.057, -0.0252, -0.0, -0.0548, -0.0265, 0.0017, -0.0613, -0.0265, 0.0017, -0.0591, -0.0265, 0.0039, -0.0613, -0.0265, 0.0039, -0.0591,
-0.0265, 0.0017, -0.057, -0.0265, 0.0017, -0.0548, -0.0265, 0.0039, -0.057, -0.0265, 0.0039, -0.0548, -0.0225, -0.0, -0.067, -0.0225, -0.0, -0.0656, -0.0225, -0.0, -0.0635, -0.0225, 0.0017, -0.067, -0.0225, 0.0039, -0.067, -0.0225, 0.0017, -0.0656, -0.0225, 0.0017, -0.0635, -0.0225, 0.0039, -0.0656, -0.0225, 0.0039, -0.0635, -0.0225, -0.0, -0.0613, -0.0225, -0.0, -0.0591, -0.0225, -0.0, -0.057, -0.0225, -0.0, -0.0548, -0.0225, 0.0017, -0.0613, -0.0225, 0.0017, -0.0591, -0.0225, 0.0039, -0.0613, -0.0225, 0.0039, -0.0591, -0.0225, 0.0017, -0.057, -0.0225, 0.0017, -0.0548, -0.0225, 0.0039, -0.057, -0.0225, 0.0039, -0.0548, -0.0179, -0.0, -0.054, -0.0157, -0.0, -0.0562, -0.0168, -0.0, -0.0551, -0.0179, 0.0017, -0.054, -0.0179, 0.0039, -0.054,
-0.0157, 0.0017, -0.0562, -0.0168, 0.0017, -0.0551, -0.0157, 0.0039, -0.0562, -0.0168, 0.0039, -0.0551, -0.0265, -0.0, -0.0526, -0.0265, -0.0, -0.0505, -0.0252, -0.0, -0.0526, -0.0252, -0.0, -0.0505, -0.0265, -0.0, -0.0483, -0.0264, -0.0, -0.0468, -0.0252, -0.0, -0.0483, -0.0262, -0.0, -0.0459, -0.0265, 0.0017, -0.0526, -0.0265, 0.0017, -0.0505, -0.0265, 0.0039, -0.0526, -0.0265, 0.0039, -0.0505, -0.0265, 0.0017, -0.0483, -0.0264, 0.0017, -0.0466, -0.0265, 0.0039, -0.0483, -0.0264, 0.0039, -0.0466, -0.0262, 0.0017, -0.0455, -0.0262, 0.0039, -0.0455, -0.0255, -0.0, -0.0437, -0.0255, 0.0017, -0.0437, -0.0255, 0.0039, -0.0437, -0.0225, -0.0, -0.0526, -0.0225, -0.0, -0.0503, -0.0202, -0.0, -0.052, -0.0213, -0.0, -0.0511, -0.023, -0.0, -0.0483,
-0.023, -0.0, -0.0461, -0.0208, -0.0, -0.0483, -0.0201, -0.0, -0.0452, -0.0225, 0.0017, -0.0526, -0.0225, 0.0017, -0.0503, -0.0225, 0.0039, -0.0526, -0.0225, 0.0039, -0.0503, -0.0202, 0.0017, -0.052, -0.0213, 0.0017, -0.0511, -0.0202, 0.0039, -0.052, -0.0213, 0.0039, -0.0511, -0.02, 0.0017, -0.0452, -0.02, 0.0039, -0.0452, -0.019, -0.0, -0.053, -0.0187, -0.0, -0.0505, -0.0165, -0.0, -0.0526, -0.0155, -0.0, -0.0495, -0.0177, -0.0, -0.0473, -0.0188, -0.0, -0.0463, -0.0166, -0.0, -0.0484, -0.019, 0.0017, -0.053, -0.019, 0.0039, -0.053, -0.0155, 0.0017, -0.0495, -0.0155, 0.0039, -0.0495, -0.0177, 0.0017, -0.0473, -0.0188, 0.0017, -0.0463, -0.0177, 0.0039, -0.0473, -0.0188, 0.0039, -0.0463, -0.0166, 0.0017, -0.0484, -0.0166, 0.0039, -0.0484,
-0.0231, -0.0, -0.0433, -0.0211, -0.0, -0.0443, -0.0231, 0.0017, -0.0433, -0.0231, 0.0039, -0.0433, -0.021, 0.0017, -0.0444, -0.021, 0.0039, -0.0444, -0.0092, -0.0, -0.0627, -0.0069, -0.0, -0.0647, -0.008, -0.0, -0.0638, -0.0092, 0.0017, -0.0627, -0.0092, 0.0039, -0.0627, -0.0069, 0.0017, -0.0647, -0.008, 0.0017, -0.0638, -0.0069, 0.0039, -0.0647, -0.008, 0.0039, -0.0638, -0.0135, -0.0, -0.0584, -0.0114, -0.0, -0.0605, -0.0124, -0.0, -0.0594, -0.0146, -0.0, -0.0573, -0.0143, -0.0, -0.0548, -0.0121, -0.0, -0.057, -0.0112, -0.0, -0.0538, -0.0135, 0.0017, -0.0584, -0.0135, 0.0039, -0.0584, -0.0114, 0.0017, -0.0605, -0.0124, 0.0017, -0.0594, -0.0114, 0.0039, -0.0605, -0.0124, 0.0039, -0.0594, -0.0146, 0.0017, -0.0573, -0.0146, 0.0039, -0.0573,
-0.0112, 0.0017, -0.0538, -0.0112, 0.0039, -0.0538, -0.0103, -0.0, -0.0616, -0.01, -0.0, -0.0591, -0.0078, -0.0, -0.0613, -0.0068, -0.0, -0.0582, -0.009, -0.0, -0.056, -0.0101, -0.0, -0.0549, -0.0079, -0.0, -0.0571, -0.0103, 0.0017, -0.0616, -0.0103, 0.0039, -0.0616, -0.0068, 0.0017, -0.0582, -0.0068, 0.0039, -0.0582, -0.009, 0.0017, -0.056, -0.0101, 0.0017, -0.0549, -0.009, 0.0039, -0.056, -0.0101, 0.0039, -0.0549, -0.0079, 0.0017, -0.0571, -0.0079, 0.0039, -0.0571, -0.0026, -0.0, -0.0668, -0.0059, -0.0, -0.0655, -0.0056, -0.0, -0.0635, -0.0037, -0.0, -0.0666, -0.0035, -0.0, -0.0635, -0.0028, 0.0017, -0.0668, -0.0028, 0.0039, -0.0668, -0.0057, 0.0017, -0.0656, -0.0057, 0.0039, -0.0656, -0.0039, 0.0017, -0.0666, -0.0039, 0.0039, -0.0666,
-0.002, -0.0, -0.067, -0.002, -0.0, -0.0656, -0.002, -0.0, -0.0635, -0.002, 0.0017, -0.067, -0.002, 0.0039, -0.067, -0.002, 0.0017, -0.0656, -0.002, 0.0017, -0.0635, -0.002, 0.0039, -0.0656, -0.002, 0.0039, -0.0635, -0.0056, -0.0, -0.0613, -0.0059, -0.0, -0.0591, -0.0035, -0.0, -0.0613, -0.0035, -0.0, -0.0591, -0.0059, -0.0, -0.057, -0.0059, -0.0, -0.0548, -0.0035, -0.0, -0.057, -0.0035, -0.0, -0.0548, -0.0059, 0.0017, -0.0591, -0.0059, 0.0039, -0.0591, -0.0059, 0.0017, -0.057, -0.0059, 0.0017, -0.0548, -0.0059, 0.0039, -0.057, -0.0059, 0.0039, -0.0548, -0.002, -0.0, -0.0613, -0.002, -0.0, -0.0591, -0.002, -0.0, -0.057, -0.002, -0.0, -0.0548, -0.002, 0.0017, -0.0613, -0.002, 0.0017, -0.0591, -0.002, 0.0039, -0.0613,
-0.002, 0.0039, -0.0591, -0.002, 0.0017, -0.057, -0.002, 0.0017, -0.0548, -0.002, 0.0039, -0.057, -0.002, 0.0039, -0.0548, -0.0134, -0.0, -0.0517, -0.0144, -0.0, -0.0506, -0.0123, -0.0, -0.0528, -0.0134, 0.0017, -0.0517, -0.0144, 0.0017, -0.0506, -0.0134, 0.0039, -0.0517, -0.0144, 0.0039, -0.0506, -0.0123, 0.0017, -0.0528, -0.0123, 0.0039, -0.0528, -0.0059, -0.0, -0.0526, -0.0059, -0.0, -0.0505, -0.0035, -0.0, -0.0526, -0.0035, -0.0, -0.0505, -0.0059, -0.0, -0.0483, -0.0059, -0.0, -0.0461, -0.0035, -0.0, -0.0483, -0.0035, -0.0, -0.0461, -0.0059, 0.0017, -0.0526, -0.0059, 0.0017, -0.0505, -0.0059, 0.0039, -0.0526, -0.0059, 0.0039, -0.0505, -0.0059, 0.0017, -0.0483, -0.0059, 0.0017, -0.0461, -0.0059, 0.0039, -0.0483, -0.0059, 0.0039, -0.0461,
-0.002, -0.0, -0.0526, -0.002, -0.0, -0.0505, -0.002, -0.0, -0.0483, -0.002, -0.0, -0.0461, -0.002, 0.0017, -0.0526, -0.002, 0.0017, -0.0505, -0.002, 0.0039, -0.0526, -0.002, 0.0039, -0.0505, -0.002, 0.0017, -0.0483, -0.002, 0.0017, -0.0461, -0.002, 0.0039, -0.0483, -0.002, 0.0039, -0.0461, -0.0059, -0.0, -0.0439, -0.0059, -0.0, -0.0425, -0.0035, -0.0, -0.0439, -0.0035, -0.0, -0.0425, -0.0059, 0.0017, -0.0439, -0.0059, 0.0017, -0.0425, -0.0059, 0.0039, -0.0439, -0.0059, 0.0039, -0.0425, -0.0035, 0.0017, -0.0425, -0.0035, 0.0039, -0.0425, -0.002, -0.0, -0.0439, -0.002, -0.0, -0.0425, -0.002, 0.0017, -0.0439, -0.002, 0.0017, -0.0425, -0.002, 0.0039, -0.0439, -0.002, 0.0039, -0.0425, -0.4807, -0.0, -0.0004, -0.4802, -0.0, 0.0013,
-0.4807, 0.0017, -0.0005, -0.4801, 0.0017, 0.0016, -0.4807, 0.0039, -0.0005, -0.4801, 0.0039, 0.0016, -0.4787, -0.0, -0.0005, -0.48, -0.0, 0.0021, -0.4766, -0.0, -0.0005, -0.4766, -0.0, 0.0016, -0.4795, -0.0, 0.004, -0.4789, -0.0, 0.0061, -0.4766, -0.0, 0.0038, -0.4766, -0.0, 0.006, -0.4798, 0.0017, 0.0027, -0.4798, 0.0039, 0.0027, -0.4795, 0.0017, 0.0038, -0.4789, 0.0017, 0.006, -0.4795, 0.0039, 0.0038, -0.4789, 0.0039, 0.006, -0.4784, -0.0, 0.0082, -0.478, -0.0, 0.0099, -0.4766, -0.0, 0.0081, -0.4777, -0.0, 0.0107, -0.4772, -0.0, 0.0127, -0.4765, -0.0, 0.0148, -0.4784, 0.0017, 0.0081, -0.4779, 0.0017, 0.0102, -0.4784, 0.0039, 0.0081, -0.4779, 0.0039, 0.0102, -0.4776, 0.0017, 0.0112, -0.4776, 0.0039, 0.0112,
-0.4772, 0.0017, 0.0125, -0.4765, 0.0017, 0.0147, -0.4772, 0.0039, 0.0125, -0.4765, 0.0039, 0.0147, -0.4744, -0.0, -0.0005, -0.4744, -0.0, 0.0016, -0.4722, -0.0, -0.0005, -0.4722, -0.0, 0.0016, -0.4744, -0.0, 0.0038, -0.4744, -0.0, 0.006, -0.4722, -0.0, 0.0038, -0.4722, -0.0, 0.006, -0.47, -0.0, -0.0005, -0.47, -0.0, 0.0016, -0.4679, -0.0, -0.0005, -0.4679, -0.0, 0.0016, -0.47, -0.0, 0.0038, -0.47, -0.0, 0.006, -0.4679, -0.0, 0.0038, -0.4679, -0.0, 0.006, -0.4744, -0.0, 0.0081, -0.4744, -0.0, 0.0103, -0.4722, -0.0, 0.0081, -0.4722, -0.0, 0.0103, -0.4744, -0.0, 0.0125, -0.4744, -0.0, 0.0146, -0.4722, -0.0, 0.0125, -0.4722, -0.0, 0.0146, -0.47, -0.0, 0.0081, -0.47, -0.0, 0.0103,
-0.4679, -0.0, 0.0081, -0.4679, -0.0, 0.0103, -0.47, -0.0, 0.0125, -0.47, -0.0, 0.0146, -0.4679, -0.0, 0.0125, -0.4679, -0.0, 0.0146, -0.4759, -0.0, 0.0164, -0.4758, 0.0017, 0.0166, -0.4758, 0.0039, 0.0166, -0.4755, -0.0, 0.0173, -0.4749, -0.0, 0.0192, -0.4722, -0.0, 0.0168, -0.4722, -0.0, 0.019, -0.4742, -0.0, 0.0213, -0.4736, -0.0, 0.0228, -0.4722, -0.0, 0.0212, -0.4733, -0.0, 0.0237, -0.4754, 0.0017, 0.0177, -0.4749, 0.0017, 0.019, -0.4754, 0.0039, 0.0177, -0.4749, 0.0039, 0.019, -0.4742, 0.0017, 0.0212, -0.4736, 0.0017, 0.0229, -0.4742, 0.0039, 0.0212, -0.4736, 0.0039, 0.0229, -0.4732, 0.0017, 0.024, -0.4732, 0.0039, 0.024, -0.47, -0.0, 0.0168, -0.47, -0.0, 0.019, -0.4679, -0.0, 0.0168,
-0.4679, -0.0, 0.019, -0.47, -0.0, 0.0212, -0.47, -0.0, 0.0233, -0.4679, -0.0, 0.0212, -0.4679, -0.0, 0.0233, -0.4724, -0.0, 0.0258, -0.4716, -0.0, 0.0278, -0.4711, -0.0, 0.0288, -0.4726, 0.0017, 0.0255, -0.4716, 0.0017, 0.0277, -0.4726, 0.0039, 0.0255, -0.4716, 0.0039, 0.0277, -0.4711, 0.0017, 0.0288, -0.4711, 0.0039, 0.0288, -0.47, -0.0, 0.0255, -0.47, -0.0, 0.0277, -0.4679, -0.0, 0.0255, -0.4679, -0.0, 0.0277, -0.4707, -0.0, 0.0299, -0.4697, -0.0, 0.0322, -0.4679, -0.0, 0.0298, -0.4679, -0.0, 0.032, -0.4707, 0.0017, 0.0299, -0.4698, 0.0017, 0.032, -0.4707, 0.0039, 0.0299, -0.4698, 0.0039, 0.032, -0.4657, -0.0, -0.0005, -0.4657, -0.0, 0.0016, -0.4628, -0.0, -0.0008, -0.4623, -0.0, 0.0012,
-0.4657, -0.0, 0.0038, -0.4657, -0.0, 0.006, -0.4635, -0.0, 0.0038, -0.4635, -0.0, 0.006, -0.4628, 0.0017, -0.0005, -0.4625, 0.0017, 0.0006, -0.4628, 0.0039, -0.0005, -0.4625, 0.0039, 0.0006, -0.4621, -0.0, 0.002, -0.4616, -0.0, 0.0037, -0.4609, -0.0, 0.0057, -0.4622, 0.0017, 0.0017, -0.4622, 0.0039, 0.0017, -0.4615, 0.0017, 0.0038, -0.4608, 0.0017, 0.006, -0.4615, 0.0039, 0.0038, -0.4608, 0.0039, 0.006, -0.4657, -0.0, 0.0081, -0.4657, -0.0, 0.0103, -0.4635, -0.0, 0.0081, -0.4635, -0.0, 0.0103, -0.4657, -0.0, 0.0125, -0.4657, -0.0, 0.0146, -0.4635, -0.0, 0.0125, -0.4635, -0.0, 0.0146, -0.4603, -0.0, 0.0078, -0.4614, -0.0, 0.0103, -0.46, -0.0, 0.0087, -0.4595, -0.0, 0.0102, -0.4614, -0.0, 0.0125,
-0.4614, -0.0, 0.0146, -0.4588, -0.0, 0.0122, -0.4581, -0.0, 0.0142, -0.4604, 0.0017, 0.0074, -0.4604, 0.0039, 0.0074, -0.4601, 0.0017, 0.0085, -0.4595, 0.0017, 0.0103, -0.4601, 0.0039, 0.0085, -0.4595, 0.0039, 0.0103, -0.4587, 0.0017, 0.0125, -0.4582, 0.0017, 0.0138, -0.4587, 0.0039, 0.0125, -0.4582, 0.0039, 0.0138, -0.4577, -0.0, 0.0151, -0.4578, 0.0017, 0.0149, -0.4578, 0.0039, 0.0149, -0.4657, -0.0, 0.0168, -0.4657, -0.0, 0.019, -0.4635, -0.0, 0.0168, -0.4635, -0.0, 0.019, -0.4657, -0.0, 0.0212, -0.4657, -0.0, 0.0233, -0.4635, -0.0, 0.0212, -0.4635, -0.0, 0.0233, -0.4614, -0.0, 0.0168, -0.4614, -0.0, 0.019, -0.4592, -0.0, 0.0168, -0.4592, -0.0, 0.019, -0.4614, -0.0, 0.0212, -0.4614, -0.0, 0.0233,
-0.4592, -0.0, 0.0212, -0.4592, -0.0, 0.0233, -0.4657, -0.0, 0.0255, -0.4657, -0.0, 0.0277, -0.4635, -0.0, 0.0255, -0.4635, -0.0, 0.0277, -0.4657, -0.0, 0.0298, -0.4657, -0.0, 0.032, -0.4635, -0.0, 0.0298, -0.4635, -0.0, 0.032, -0.4614, -0.0, 0.0255, -0.4614, -0.0, 0.0277, -0.4592, -0.0, 0.0255, -0.4592, -0.0, 0.0277, -0.4614, -0.0, 0.0298, -0.4614, -0.0, 0.032, -0.4592, -0.0, 0.0298, -0.4592, -0.0, 0.032, -0.4572, -0.0, 0.0167, -0.4564, -0.0, 0.0187, -0.4558, -0.0, 0.0206, -0.457, -0.0, 0.0233, -0.4555, -0.0, 0.0214, -0.4549, -0.0, 0.0231, -0.4571, 0.0017, 0.0168, -0.4563, 0.0017, 0.019, -0.4571, 0.0039, 0.0168, -0.4563, 0.0039, 0.019, -0.4559, 0.0017, 0.0201, -0.4559, 0.0039, 0.0201,
-0.4556, 0.0017, 0.0212, -0.4548, 0.0017, 0.0233, -0.4556, 0.0039, 0.0212, -0.4548, 0.0039, 0.0233, -0.457, -0.0, 0.0255, -0.457, -0.0, 0.0277, -0.454, -0.0, 0.0252, -0.4549, -0.0, 0.0277, -0.457, -0.0, 0.0298, -0.457, -0.0, 0.032, -0.4549, -0.0, 0.0298, -0.4549, -0.0, 0.032, -0.4541, 0.0017, 0.0251, -0.4541, 0.0039, 0.0251, -0.4536, -0.0, 0.0263, -0.453, -0.0, 0.0275, -0.4519, -0.0, 0.0296, -0.4527, -0.0, 0.032, -0.4514, -0.0, 0.0307, -0.4508, -0.0, 0.0318, -0.4536, 0.0017, 0.0262, -0.4529, 0.0017, 0.0277, -0.4536, 0.0039, 0.0262, -0.4529, 0.0039, 0.0277, -0.452, 0.0017, 0.0295, -0.452, 0.0039, 0.0295, -0.4514, 0.0017, 0.0306, -0.4507, 0.0017, 0.032, -0.4514, 0.0039, 0.0306, -0.4507, 0.0039, 0.032,
-0.4691, -0.0, 0.0333, -0.4686, -0.0, 0.0344, -0.4676, -0.0, 0.0365, -0.467, -0.0, 0.0377, -0.4691, 0.0017, 0.0334, -0.4691, 0.0039, 0.0334, -0.4686, 0.0017, 0.0345, -0.4677, 0.0017, 0.0364, -0.4686, 0.0039, 0.0345, -0.4677, 0.0039, 0.0364, -0.467, 0.0017, 0.0378, -0.467, 0.0039, 0.0378, -0.4657, -0.0, 0.0342, -0.4657, -0.0, 0.0363, -0.4635, -0.0, 0.0342, -0.4635, -0.0, 0.0363, -0.4664, -0.0, 0.0388, -0.4654, -0.0, 0.0409, -0.4635, -0.0, 0.0385, -0.4635, -0.0, 0.0407, -0.4664, 0.0017, 0.0389, -0.4655, 0.0017, 0.0407, -0.4664, 0.0039, 0.0389, -0.4655, 0.0039, 0.0407, -0.4614, -0.0, 0.0342, -0.4614, -0.0, 0.0363, -0.4592, -0.0, 0.0342, -0.4592, -0.0, 0.0363, -0.4614, -0.0, 0.0385, -0.4614, -0.0, 0.0407,
-0.4592, -0.0, 0.0385, -0.4592, -0.0, 0.0407, -0.4648, -0.0, 0.042, -0.4641, -0.0, 0.0431, -0.4631, -0.0, 0.0448, -0.4647, 0.0017, 0.042, -0.4647, 0.0039, 0.042, -0.4641, 0.0017, 0.0431, -0.463, 0.0017, 0.0449, -0.4641, 0.0039, 0.0431, -0.463, 0.0039, 0.0449, -0.4614, -0.0, 0.0429, -0.4625, -0.0, 0.0458, -0.4592, -0.0, 0.0429, -0.4592, -0.0, 0.045, -0.4615, -0.0, 0.0475, -0.4607, -0.0, 0.0488, -0.4592, -0.0, 0.0472, -0.46, -0.0, 0.0498, -0.4624, 0.0017, 0.046, -0.4624, 0.0039, 0.046, -0.4617, 0.0017, 0.0472, -0.4606, 0.0017, 0.0488, -0.4617, 0.0039, 0.0472, -0.4606, 0.0039, 0.0488, -0.4599, 0.0017, 0.0499, -0.4599, 0.0039, 0.0499, -0.457, -0.0, 0.0342, -0.457, -0.0, 0.0363, -0.4549, -0.0, 0.0342,
-0.4549, -0.0, 0.0363, -0.457, -0.0, 0.0385, -0.457, -0.0, 0.0407, -0.4549, -0.0, 0.0385, -0.4549, -0.0, 0.0407, -0.4527, -0.0, 0.0342, -0.4527, -0.0, 0.0363, -0.4498, -0.0, 0.0339, -0.4505, -0.0, 0.0363, -0.4527, -0.0, 0.0385, -0.4515, -0.0, 0.0415, -0.4499, -0.0, 0.0393, -0.4505, -0.0, 0.0408, -0.4498, 0.0017, 0.0339, -0.4498, 0.0039, 0.0339, -0.4517, 0.0017, 0.0417, -0.4517, 0.0039, 0.0417, -0.4498, 0.0017, 0.0392, -0.4505, 0.0017, 0.0408, -0.4498, 0.0039, 0.0392, -0.4505, 0.0039, 0.0408, -0.457, -0.0, 0.0429, -0.457, -0.0, 0.045, -0.4549, -0.0, 0.0429, -0.4549, -0.0, 0.045, -0.457, -0.0, 0.0472, -0.457, -0.0, 0.0494, -0.4549, -0.0, 0.0472, -0.4549, -0.0, 0.0494, -0.4525, -0.0, 0.0431,
-0.452, -0.0, 0.0453, -0.451, -0.0, 0.0425, -0.4515, -0.0, 0.045, -0.452, -0.0, 0.0468, -0.4523, -0.0, 0.0492, -0.4516, -0.0, 0.0472, -0.4508, -0.0, 0.0496, -0.4523, 0.0017, 0.0426, -0.452, 0.0017, 0.0452, -0.4523, 0.0039, 0.0426, -0.452, 0.0039, 0.0452, -0.4512, 0.0017, 0.0429, -0.4515, 0.0017, 0.0447, -0.4512, 0.0039, 0.0429, -0.4515, 0.0039, 0.0447, -0.4519, 0.0017, 0.0469, -0.4522, 0.0017, 0.0494, -0.4519, 0.0039, 0.0469, -0.4522, 0.0039, 0.0494, -0.4516, 0.0017, 0.0473, -0.4509, 0.0017, 0.0494, -0.4516, 0.0039, 0.0473, -0.4509, 0.0039, 0.0494, -0.459, -0.0, 0.0514, -0.4589, 0.0017, 0.0515, -0.4589, 0.0039, 0.0515, -0.4582, -0.0, 0.0524, -0.457, -0.0, 0.054, -0.4549, -0.0, 0.0515, -0.4549, -0.0, 0.0537,
-0.4562, -0.0, 0.0551, -0.4553, -0.0, 0.0562, -0.4543, -0.0, 0.0575, -0.4581, 0.0017, 0.0526, -0.4573, 0.0017, 0.0537, -0.4581, 0.0039, 0.0526, -0.4573, 0.0039, 0.0537, -0.4562, 0.0017, 0.0551, -0.4562, 0.0039, 0.0551, -0.4553, 0.0017, 0.0562, -0.4543, 0.0017, 0.0576, -0.4553, 0.0039, 0.0562, -0.4543, 0.0039, 0.0576, -0.4515, -0.0, 0.0508, -0.4527, -0.0, 0.0537, -0.4504, -0.0, 0.0514, -0.4498, -0.0, 0.0529, -0.4527, -0.0, 0.0559, -0.4534, -0.0, 0.0586, -0.4505, -0.0, 0.0559, -0.4505, -0.0, 0.058, -0.4517, 0.0017, 0.0505, -0.4517, 0.0039, 0.0505, -0.4499, 0.0017, 0.0521, -0.4497, 0.0017, 0.0529, -0.4499, 0.0039, 0.0521, -0.4497, 0.0039, 0.0529, -0.4534, 0.0017, 0.0586, -0.4534, 0.0039, 0.0586, -0.4523, -0.0, 0.0599,
-0.4513, -0.0, 0.061, -0.4502, -0.0, 0.0621, -0.4523, 0.0017, 0.0599, -0.4523, 0.0039, 0.0599, -0.4513, 0.0017, 0.061, -0.4502, 0.0017, 0.0621, -0.4513, 0.0039, 0.061, -0.4502, 0.0039, 0.0621, -0.4323, -0.0, 0.0046, -0.4328, -0.0, 0.0058, -0.4322, 0.0017, 0.0045, -0.4328, 0.0017, 0.0059, -0.4322, 0.0039, 0.0045, -0.4328, 0.0039, 0.0059, -0.4332, -0.0, 0.0081, -0.4333, -0.0, 0.0103, -0.4333, -0.0, 0.0125, -0.4331, -0.0, 0.0147, -0.4332, 0.0017, 0.0081, -0.4333, 0.0017, 0.0103, -0.4332, 0.0039, 0.0081, -0.4333, 0.0039, 0.0103, -0.4333, 0.0017, 0.0125, -0.4331, 0.0017, 0.0147, -0.4333, 0.0039, 0.0125, -0.4331, 0.0039, 0.0147, -0.4326, -0.0, 0.0165, -0.4325, 0.0017, 0.0168, -0.4325, 0.0039, 0.0168, -0.4306, -0.0, 0.002,
-0.4284, -0.0, -0.0001, -0.4295, -0.0, 0.0008, -0.4317, -0.0, 0.0035, -0.431, -0.0, 0.006, -0.4288, -0.0, 0.0038, -0.4288, -0.0, 0.006, -0.4306, 0.0017, 0.0019, -0.4306, 0.0039, 0.0019, -0.4285, 0.0017, -0.0001, -0.4296, 0.0017, 0.0009, -0.4285, 0.0039, -0.0001, -0.4296, 0.0039, 0.0009, -0.4317, 0.0017, 0.0034, -0.4317, 0.0039, 0.0034, -0.4269, -0.0, -0.0012, -0.4266, -0.0, 0.0016, -0.4245, -0.0, -0.0005, -0.4245, -0.0, 0.0016, -0.4266, -0.0, 0.0038, -0.4266, -0.0, 0.006, -0.4245, -0.0, 0.0038, -0.4245, -0.0, 0.006, -0.4269, 0.0017, -0.0012, -0.4269, 0.0039, -0.0012, -0.431, -0.0, 0.0081, -0.431, -0.0, 0.0103, -0.4288, -0.0, 0.0081, -0.4288, -0.0, 0.0103, -0.431, -0.0, 0.0125, -0.431, -0.0, 0.0146,
-0.4288, -0.0, 0.0125, -0.4288, -0.0, 0.0146, -0.4266, -0.0, 0.0081, -0.4266, -0.0, 0.0103, -0.4245, -0.0, 0.0081, -0.4245, -0.0, 0.0103, -0.4266, -0.0, 0.0125, -0.4266, -0.0, 0.0146, -0.4245, -0.0, 0.0125, -0.4245, -0.0, 0.0146, -0.4223, -0.0, -0.0005, -0.4223, -0.0, 0.0016, -0.4201, -0.0, -0.0005, -0.4201, -0.0, 0.0016, -0.4223, -0.0, 0.0038, -0.4223, -0.0, 0.006, -0.4201, -0.0, 0.0038, -0.4201, -0.0, 0.006, -0.418, -0.0, -0.0005, -0.418, -0.0, 0.0016, -0.4158, -0.0, -0.0005, -0.4158, -0.0, 0.0016, -0.418, -0.0, 0.0038, -0.418, -0.0, 0.006, -0.4158, -0.0, 0.0038, -0.4158, -0.0, 0.006, -0.4223, -0.0, 0.0081, -0.4223, -0.0, 0.0103, -0.4201, -0.0, 0.0081, -0.4201, -0.0, 0.0103,
-0.4223, -0.0, 0.0125, -0.4223, -0.0, 0.0146, -0.4201, -0.0, 0.0125, -0.4201, -0.0, 0.0146, -0.418, -0.0, 0.0081, -0.418, -0.0, 0.0103, -0.4158, -0.0, 0.0081, -0.4158, -0.0, 0.0103, -0.418, -0.0, 0.0125, -0.418, -0.0, 0.0146, -0.4158, -0.0, 0.0125, -0.4158, -0.0, 0.0146, -0.4322, -0.0, 0.0175, -0.4314, -0.0, 0.0194, -0.4288, -0.0, 0.0168, -0.4288, -0.0, 0.019, -0.4304, -0.0, 0.0206, -0.4293, -0.0, 0.0217, -0.4281, -0.0, 0.0226, -0.4321, 0.0017, 0.0179, -0.4315, 0.0017, 0.0191, -0.4321, 0.0039, 0.0179, -0.4315, 0.0039, 0.0191, -0.4304, 0.0017, 0.0206, -0.4304, 0.0039, 0.0206, -0.4293, 0.0017, 0.0217, -0.4282, 0.0017, 0.0226, -0.4293, 0.0039, 0.0217, -0.4282, 0.0039, 0.0226, -0.4266, -0.0, 0.0168,
-0.4266, -0.0, 0.019, -0.4245, -0.0, 0.0168, -0.4245, -0.0, 0.019, -0.4266, -0.0, 0.0212, -0.427, -0.0, 0.0234, -0.4245, -0.0, 0.0212, -0.4247, -0.0, 0.0243, -0.4267, 0.0017, 0.0236, -0.4267, 0.0039, 0.0236, -0.4249, 0.0017, 0.0242, -0.4249, 0.0039, 0.0242, -0.4237, -0.0, 0.0246, -0.4238, 0.0017, 0.0246, -0.4238, 0.0039, 0.0246, -0.4223, -0.0, 0.0168, -0.4223, -0.0, 0.019, -0.4201, -0.0, 0.0168, -0.4201, -0.0, 0.019, -0.4223, -0.0, 0.0212, -0.4223, -0.0, 0.0233, -0.4201, -0.0, 0.0212, -0.4201, -0.0, 0.0233, -0.418, -0.0, 0.0168, -0.418, -0.0, 0.019, -0.4158, -0.0, 0.0168, -0.4158, -0.0, 0.019, -0.418, -0.0, 0.0212, -0.418, -0.0, 0.0233, -0.4158, -0.0, 0.0212, -0.4153, -0.0, 0.0246,
-0.4148, 0.0017, 0.0244, -0.4148, 0.0039, 0.0244, -0.4223, -0.0, 0.0249, -0.4201, -0.0, 0.0251, -0.4223, 0.0017, 0.0249, -0.4223, 0.0039, 0.0249, -0.4201, 0.0017, 0.0251, -0.4201, 0.0039, 0.0251, -0.4179, -0.0, 0.025, -0.4162, -0.0, 0.0247, -0.418, 0.0017, 0.025, -0.418, 0.0039, 0.025, -0.4159, 0.0017, 0.0247, -0.4159, 0.0039, 0.0247, -0.4493, -0.0, 0.035, -0.4487, -0.0, 0.0362, -0.4478, -0.0, 0.0387, -0.4483, -0.0, 0.0407, -0.4463, -0.0, 0.0385, -0.4464, -0.0, 0.0407, -0.4493, 0.0017, 0.035, -0.4487, 0.0017, 0.0363, -0.4493, 0.0039, 0.035, -0.4487, 0.0039, 0.0363, -0.4478, 0.0017, 0.0387, -0.4483, 0.0017, 0.0407, -0.4478, 0.0039, 0.0387, -0.4483, 0.0039, 0.0407, -0.4462, 0.0017, 0.0385, -0.4471, 0.0017, 0.0414,
-0.4462, 0.0039, 0.0385, -0.4471, 0.0039, 0.0414, -0.444, -0.0, 0.0384, -0.444, -0.0, 0.0407, -0.4417, -0.0, 0.0388, -0.4418, -0.0, 0.0407, -0.444, 0.0017, 0.0384, -0.444, 0.0039, 0.0384, -0.4418, 0.0017, 0.0388, -0.4418, 0.0039, 0.0388, -0.4483, -0.0, 0.0426, -0.4483, -0.0, 0.045, -0.4471, -0.0, 0.0419, -0.4462, -0.0, 0.045, -0.4483, -0.0, 0.0472, -0.4483, -0.0, 0.0494, -0.4462, -0.0, 0.0472, -0.4462, -0.0, 0.0494, -0.4476, 0.0017, 0.0419, -0.4476, 0.0039, 0.0419, -0.4471, 0.0017, 0.0419, -0.4471, 0.0039, 0.0419, -0.444, -0.0, 0.0429, -0.444, -0.0, 0.045, -0.4418, -0.0, 0.0429, -0.4418, -0.0, 0.045, -0.444, -0.0, 0.0472, -0.444, -0.0, 0.0494, -0.4418, -0.0, 0.0472, -0.4418, -0.0, 0.0494,
-0.4402, -0.0, 0.0392, -0.4392, -0.0, 0.0397, -0.4375, -0.0, 0.0406, -0.44, 0.0017, 0.0393, -0.439, 0.0017, 0.0398, -0.44, 0.0039, 0.0393, -0.439, 0.0039, 0.0398, -0.438, 0.0017, 0.0402, -0.438, 0.0039, 0.0402, -0.4397, -0.0, 0.0429, -0.4397, -0.0, 0.045, -0.4364, -0.0, 0.0421, -0.4375, -0.0, 0.045, -0.4397, -0.0, 0.0472, -0.4397, -0.0, 0.0494, -0.4375, -0.0, 0.0472, -0.4363, -0.0, 0.0501, -0.4365, 0.0017, 0.0419, -0.4365, 0.0039, 0.0419, -0.4364, 0.0017, 0.0504, -0.4364, 0.0039, 0.0504, -0.4359, -0.0, 0.0433, -0.4355, -0.0, 0.045, -0.4354, -0.0, 0.0473, -0.4358, -0.0, 0.0491, -0.4359, 0.0017, 0.043, -0.4355, 0.0017, 0.045, -0.4359, 0.0039, 0.043, -0.4355, 0.0039, 0.045, -0.4354, 0.0017, 0.0472,
-0.4359, 0.0017, 0.0493, -0.4354, 0.0039, 0.0472, -0.4359, 0.0039, 0.0493, -0.4488, -0.0, 0.0521, -0.4481, -0.0, 0.0535, -0.4462, -0.0, 0.0515, -0.4464, -0.0, 0.0538, -0.4475, -0.0, 0.0547, -0.4483, -0.0, 0.058, -0.446, -0.0, 0.0556, -0.4462, -0.0, 0.058, -0.4489, 0.0017, 0.0521, -0.4481, 0.0017, 0.0535, -0.4489, 0.0039, 0.0521, -0.4481, 0.0039, 0.0535, -0.4462, 0.0017, 0.0538, -0.4462, 0.0039, 0.0538, -0.4473, 0.0017, 0.0548, -0.4473, 0.0039, 0.0548, -0.4461, 0.0017, 0.0556, -0.4461, 0.0039, 0.0556, -0.444, -0.0, 0.0515, -0.444, -0.0, 0.0543, -0.4418, -0.0, 0.0515, -0.4417, -0.0, 0.0541, -0.444, -0.0, 0.0559, -0.444, -0.0, 0.058, -0.4419, -0.0, 0.0558, -0.4418, -0.0, 0.058, -0.444, 0.0017, 0.0543,
-0.444, 0.0039, 0.0543, -0.4418, 0.0017, 0.0541, -0.4418, 0.0039, 0.0541, -0.444, 0.0017, 0.0559, -0.444, 0.0039, 0.0559, -0.4418, 0.0017, 0.0557, -0.4418, 0.0039, 0.0557, -0.4483, -0.0, 0.0602, -0.4491, -0.0, 0.0632, -0.4462, -0.0, 0.0602, -0.4462, -0.0, 0.0624, -0.448, -0.0, 0.0642, -0.4468, -0.0, 0.0653, -0.4457, -0.0, 0.0662, -0.4491, 0.0017, 0.0632, -0.4491, 0.0039, 0.0632, -0.4481, 0.0017, 0.0642, -0.4481, 0.0039, 0.0642, -0.4468, 0.0017, 0.0653, -0.4457, 0.0017, 0.0662, -0.4468, 0.0039, 0.0653, -0.4457, 0.0039, 0.0662, -0.444, -0.0, 0.0602, -0.444, -0.0, 0.0624, -0.4418, -0.0, 0.0602, -0.4418, -0.0, 0.0624, -0.444, -0.0, 0.0646, -0.4444, -0.0, 0.0673, -0.4418, -0.0, 0.0646, -0.4418, -0.0, 0.0667,
-0.4444, 0.0017, 0.0673, -0.4444, 0.0039, 0.0673, -0.4397, -0.0, 0.0515, -0.4393, -0.0, 0.0543, -0.4375, -0.0, 0.052, -0.4378, 0.0003, 0.054, -0.4398, -0.0, 0.055, -0.4397, -0.0, 0.058, -0.4375, -0.0, 0.0559, -0.4375, -0.0, 0.058, -0.4395, 0.0017, 0.0544, -0.4395, 0.0039, 0.0544, -0.4373, 0.0017, 0.0518, -0.4378, 0.0017, 0.0539, -0.4373, 0.0039, 0.0518, -0.4378, 0.0039, 0.0539, -0.44, 0.0017, 0.0551, -0.44, 0.0039, 0.0551, -0.436, -0.0, 0.0545, -0.4349, -0.0, 0.0548, -0.4353, -0.0, 0.058, -0.4329, -0.0, 0.056, -0.4332, -0.0, 0.058, -0.4358, 0.0017, 0.0545, -0.4358, 0.0039, 0.0545, -0.4347, 0.0017, 0.0549, -0.4347, 0.0039, 0.0549, -0.4331, 0.0017, 0.0558, -0.4331, 0.0039, 0.0558, -0.4397, -0.0, 0.0602,
-0.4397, -0.0, 0.0624, -0.4375, -0.0, 0.0602, -0.4375, -0.0, 0.0624, -0.4397, -0.0, 0.0646, -0.4397, -0.0, 0.0667, -0.4375, -0.0, 0.0646, -0.4375, -0.0, 0.0667, -0.4353, -0.0, 0.0602, -0.4353, -0.0, 0.0624, -0.4332, -0.0, 0.0602, -0.4332, -0.0, 0.0624, -0.4353, -0.0, 0.0646, -0.4353, -0.0, 0.0667, -0.4332, -0.0, 0.0646, -0.4332, -0.0, 0.0667, -0.4319, -0.0, 0.0568, -0.4307, -0.0, 0.0576, -0.4294, -0.0, 0.0587, -0.4318, 0.0017, 0.0568, -0.4307, 0.0017, 0.0576, -0.4318, 0.0039, 0.0568, -0.4307, 0.0039, 0.0576, -0.4294, 0.0017, 0.0587, -0.4294, 0.0039, 0.0587, -0.431, -0.0, 0.0602, -0.431, -0.0, 0.0624, -0.4283, -0.0, 0.0596, -0.4288, -0.0, 0.0624, -0.431, -0.0, 0.0646, -0.431, -0.0, 0.0667,
-0.4288, -0.0, 0.0646, -0.4288, -0.0, 0.0667, -0.4283, 0.0017, 0.0596, -0.4283, 0.0039, 0.0596, -0.4271, -0.0, 0.0607, -0.426, -0.0, 0.0615, -0.4243, -0.0, 0.0628, -0.4266, -0.0, 0.0646, -0.4266, -0.0, 0.0667, -0.4245, -0.0, 0.0646, -0.4245, -0.0, 0.0667, -0.427, 0.0017, 0.0607, -0.4259, 0.0017, 0.0616, -0.427, 0.0039, 0.0607, -0.4259, 0.0039, 0.0616, -0.4245, 0.0017, 0.0627, -0.4245, 0.0039, 0.0627, -0.4234, -0.0, 0.0635, -0.4222, -0.0, 0.0642, -0.4223, -0.0, 0.0667, -0.4207, -0.0, 0.0652, -0.4197, -0.0, 0.0659, -0.4234, 0.0017, 0.0635, -0.4234, 0.0039, 0.0635, -0.4223, 0.0017, 0.0642, -0.4223, 0.0039, 0.0642, -0.4206, 0.0017, 0.0653, -0.4195, 0.0017, 0.066, -0.4206, 0.0039, 0.0653, -0.4195, 0.0039, 0.066,
-0.4178, -0.0, 0.067, -0.4167, -0.0, 0.0677, -0.418, 0.0017, 0.0669, -0.418, 0.0039, 0.0669, -0.4167, 0.0017, 0.0677, -0.4167, 0.0039, 0.0677, -0.4433, -0.0, 0.0682, -0.4421, -0.0, 0.0693, -0.4409, -0.0, 0.0702, -0.4434, 0.0017, 0.0682, -0.4434, 0.0039, 0.0682, -0.442, 0.0017, 0.0693, -0.441, 0.0017, 0.0701, -0.442, 0.0039, 0.0693, -0.441, 0.0039, 0.0701, -0.4397, -0.0, 0.0689, -0.4399, -0.0, 0.0709, -0.4375, -0.0, 0.0689, -0.4383, -0.0, 0.0721, -0.4373, -0.0, 0.0729, -0.4397, 0.0017, 0.0711, -0.4397, 0.0039, 0.0711, -0.4384, 0.0017, 0.0721, -0.4384, 0.0039, 0.0721, -0.4374, 0.0017, 0.0728, -0.4374, 0.0039, 0.0728, -0.4353, -0.0, 0.0689, -0.4353, -0.0, 0.0711, -0.4332, -0.0, 0.0689, -0.4332, -0.0, 0.0711,
-0.4357, -0.0, 0.0739, -0.4346, -0.0, 0.0747, -0.4332, -0.0, 0.0732, -0.4334, -0.0, 0.0754, -0.4358, 0.0017, 0.0739, -0.4347, 0.0017, 0.0746, -0.4358, 0.0039, 0.0739, -0.4347, 0.0039, 0.0746, -0.4332, 0.0017, 0.0755, -0.4332, 0.0039, 0.0755, -0.4321, -0.0, 0.087, -0.4322, -0.0, 0.0879, -0.4321, 0.0017, 0.0868, -0.4322, 0.0017, 0.0882, -0.4321, 0.0039, 0.0868, -0.4322, 0.0039, 0.0882, -0.431, -0.0, 0.0689, -0.431, -0.0, 0.0711, -0.4288, -0.0, 0.0689, -0.4288, -0.0, 0.0711, -0.431, -0.0, 0.0732, -0.4309, -0.0, 0.0763, -0.4288, -0.0, 0.0732, -0.4286, -0.0, 0.0757, -0.4309, 0.0017, 0.0763, -0.4309, 0.0039, 0.0763, -0.4288, 0.0017, 0.0758, -0.4288, 0.0039, 0.0758, -0.4266, -0.0, 0.0689, -0.4266, -0.0, 0.0711,
-0.4245, -0.0, 0.0689, -0.4245, -0.0, 0.0711, -0.426, -0.0, 0.0744, -0.4266, -0.0, 0.0754, -0.4239, -0.0, 0.0741, -0.4246, -0.0, 0.0748, -0.4257, 0.0017, 0.0742, -0.4266, 0.0017, 0.0753, -0.4257, 0.0039, 0.0742, -0.4266, 0.0039, 0.0753, -0.4242, 0.0017, 0.0742, -0.4246, 0.0017, 0.0748, -0.4242, 0.0039, 0.0742, -0.4246, 0.0039, 0.0748, -0.4304, -0.0, 0.0802, -0.4286, -0.0, 0.0777, -0.4297, -0.0, 0.0791, -0.4311, -0.0, 0.0817, -0.4317, -0.0, 0.0839, -0.4278, -0.0, 0.0829, -0.4276, -0.0, 0.0836, -0.4303, 0.0017, 0.0801, -0.4303, 0.0039, 0.0801, -0.4286, 0.0017, 0.0777, -0.4296, 0.0017, 0.079, -0.4286, 0.0039, 0.0777, -0.4296, 0.0039, 0.079, -0.4311, 0.0017, 0.0819, -0.4318, 0.0017, 0.0841, -0.4311, 0.0039, 0.0819,
-0.4318, 0.0039, 0.0841, -0.4278, 0.0017, 0.0829, -0.4277, 0.0017, 0.083, -0.4278, 0.0039, 0.0829, -0.4277, 0.0039, 0.083, -0.4275, -0.0, 0.0766, -0.4264, -0.0, 0.0798, -0.4246, -0.0, 0.0778, -0.4248, -0.0, 0.0794, -0.4285, -0.0, 0.0847, -0.4275, -0.0, 0.084, -0.4253, -0.0, 0.0811, -0.4245, -0.0, 0.0841, -0.4275, 0.0017, 0.0767, -0.4264, 0.0017, 0.0798, -0.4275, 0.0039, 0.0767, -0.4264, 0.0039, 0.0798, -0.4246, 0.0017, 0.0778, -0.4248, 0.0017, 0.0794, -0.4246, 0.0039, 0.0778, -0.4248, 0.0039, 0.0794, -0.4285, 0.0017, 0.0847, -0.4275, 0.0017, 0.084, -0.4285, 0.0039, 0.0847, -0.4275, 0.0039, 0.084, -0.4253, 0.0017, 0.0811, -0.4253, 0.0039, 0.0811, -0.4223, -0.0, 0.0689, -0.4223, -0.0, 0.0711, -0.4201, -0.0, 0.0689,
-0.4201, -0.0, 0.0711, -0.4222, -0.0, 0.0738, -0.4223, -0.0, 0.0754, -0.4201, -0.0, 0.0734, -0.4201, -0.0, 0.0754, -0.418, -0.0, 0.0689, -0.418, -0.0, 0.0711, -0.4156, -0.0, 0.0683, -0.4158, -0.0, 0.0711, -0.418, -0.0, 0.0733, -0.418, -0.0, 0.0754, -0.4158, -0.0, 0.0736, -0.4158, -0.0, 0.0754, -0.4156, 0.0017, 0.0683, -0.4156, 0.0039, 0.0683, -0.422, -0.0, 0.0778, -0.4229, -0.0, 0.0789, -0.4204, -0.0, 0.0775, -0.4201, -0.0, 0.0798, -0.4223, -0.0, 0.0819, -0.4223, -0.0, 0.0841, -0.4201, -0.0, 0.0819, -0.4201, -0.0, 0.0841, -0.4224, 0.0017, 0.078, -0.4229, 0.0017, 0.079, -0.4224, 0.0039, 0.078, -0.4229, 0.0039, 0.079, -0.418, -0.0, 0.0776, -0.418, -0.0, 0.0798, -0.4158, -0.0, 0.0776,
-0.4158, -0.0, 0.0798, -0.418, -0.0, 0.0819, -0.418, -0.0, 0.0841, -0.4158, -0.0, 0.0819, -0.4158, -0.0, 0.0841, -0.432, -0.0, 0.0862, -0.4321, -0.0, 0.0887, -0.4288, -0.0, 0.0863, -0.4288, -0.0, 0.0884, -0.4317, -0.0, 0.0909, -0.4309, -0.0, 0.093, -0.4288, -0.0, 0.0906, -0.4288, -0.0, 0.0928, -0.432, 0.0017, 0.0858, -0.432, 0.0017, 0.0893, -0.432, 0.0039, 0.0858, -0.432, 0.0039, 0.0893, -0.4318, 0.0017, 0.0906, -0.431, 0.0017, 0.0928, -0.4318, 0.0039, 0.0906, -0.431, 0.0039, 0.0928, -0.4266, -0.0, 0.0863, -0.4266, -0.0, 0.0884, -0.4245, -0.0, 0.0863, -0.4245, -0.0, 0.0884, -0.4266, -0.0, 0.0906, -0.4266, -0.0, 0.0928, -0.4245, -0.0, 0.0906, -0.4245, -0.0, 0.0928, -0.4301, -0.0, 0.0941,
-0.4293, -0.0, 0.0953, -0.4282, -0.0, 0.0965, -0.4301, 0.0017, 0.0942, -0.4301, 0.0039, 0.0942, -0.4293, 0.0017, 0.0953, -0.4282, 0.0017, 0.0965, -0.4293, 0.0039, 0.0953, -0.4282, 0.0039, 0.0965, -0.4266, -0.0, 0.0949, -0.427, -0.0, 0.0976, -0.4245, -0.0, 0.0949, -0.4245, -0.0, 0.0971, -0.4258, -0.0, 0.0984, -0.4247, -0.0, 0.0992, -0.427, 0.0017, 0.0976, -0.427, 0.0039, 0.0976, -0.4259, 0.0017, 0.0984, -0.4259, 0.0039, 0.0984, -0.4245, 0.0017, 0.0993, -0.4245, 0.0039, 0.0993, -0.4223, -0.0, 0.0863, -0.4223, -0.0, 0.0884, -0.4201, -0.0, 0.0863, -0.4201, -0.0, 0.0884, -0.4223, -0.0, 0.0906, -0.4223, -0.0, 0.0928, -0.4201, -0.0, 0.0906, -0.4201, -0.0, 0.0928, -0.418, -0.0, 0.0863, -0.418, -0.0, 0.0884,
-0.4158, -0.0, 0.0863, -0.4158, -0.0, 0.0884, -0.418, -0.0, 0.0906, -0.418, -0.0, 0.0928, -0.4158, -0.0, 0.0906, -0.4158, -0.0, 0.0928, -0.4223, -0.0, 0.0949, -0.4223, -0.0, 0.0971, -0.4201, -0.0, 0.0949, -0.4201, -0.0, 0.0971, -0.4226, -0.0, 0.1, -0.4205, -0.0, 0.1005, -0.4197, -0.0, 0.1006, -0.4223, 0.0017, 0.1001, -0.4223, 0.0039, 0.1001, -0.4211, 0.0017, 0.1003, -0.42, 0.0017, 0.1005, -0.4211, 0.0039, 0.1003, -0.42, 0.0039, 0.1005, -0.418, -0.0, 0.0949, -0.418, -0.0, 0.0971, -0.4158, -0.0, 0.0949, -0.4158, -0.0, 0.0971, -0.418, -0.0, 0.0993, -0.418, -0.0, 0.1006, -0.4157, -0.0, 0.1003, -0.4166, -0.0, 0.1004, -0.418, 0.0017, 0.1006, -0.418, 0.0039, 0.1006, -0.4154, 0.0017, 0.1003,
-0.4165, 0.0017, 0.1004, -0.4154, 0.0039, 0.1003, -0.4165, 0.0039, 0.1004, -0.4131, -0.0, -0.0014, -0.4136, -0.0, 0.0016, -0.4112, -0.0, -0.0002, -0.4115, -0.0, 0.0016, -0.4136, -0.0, 0.0038, -0.4136, -0.0, 0.006, -0.4115, -0.0, 0.0038, -0.4115, -0.0, 0.006, -0.413, 0.0017, -0.0014, -0.413, 0.0039, -0.0014, -0.4114, 0.0017, -0.0004, -0.4114, 0.0039, -0.0004, -0.4104, -0.0, 0.0005, -0.4094, -0.0, 0.0012, -0.4081, -0.0, 0.003, -0.4093, -0.0, 0.006, -0.4074, -0.0, 0.004, -0.4065, -0.0, 0.0057, -0.4104, 0.0017, 0.0005, -0.4091, 0.0017, 0.0015, -0.4104, 0.0039, 0.0005, -0.4091, 0.0039, 0.0015, -0.4082, 0.0017, 0.0027, -0.4082, 0.0039, 0.0027, -0.4075, 0.0017, 0.0038, -0.4064, 0.0017, 0.0059, -0.4075, 0.0039, 0.0038,
-0.4064, 0.0039, 0.0059, -0.4136, -0.0, 0.0081, -0.4136, -0.0, 0.0103, -0.4115, -0.0, 0.0081, -0.4115, -0.0, 0.0103, -0.4136, -0.0, 0.0125, -0.4136, -0.0, 0.0146, -0.4115, -0.0, 0.0125, -0.4115, -0.0, 0.0146, -0.4093, -0.0, 0.0081, -0.4093, -0.0, 0.0103, -0.406, -0.0, 0.008, -0.4071, -0.0, 0.0103, -0.4093, -0.0, 0.0125, -0.4093, -0.0, 0.0146, -0.4071, -0.0, 0.0125, -0.4061, -0.0, 0.0146, -0.4061, 0.0017, 0.0076, -0.4061, 0.0039, 0.0076, -0.4061, 0.0017, 0.0147, -0.4061, 0.0039, 0.0147, -0.398, -0.0, -0.0011, -0.3981, 0.0017, -0.0012, -0.3981, 0.0039, -0.0012, -0.4059, -0.0, 0.0088, -0.4059, -0.0, 0.0103, -0.4059, -0.0, 0.0125, -0.406, -0.0, 0.0136, -0.406, 0.0017, 0.0087, -0.4059, 0.0017, 0.0103,
-0.406, 0.0039, 0.0087, -0.4059, 0.0039, 0.0103, -0.4059, 0.0017, 0.0125, -0.406, 0.0017, 0.0136, -0.4059, 0.0039, 0.0125, -0.406, 0.0039, 0.0136, -0.4136, -0.0, 0.0168, -0.4136, -0.0, 0.019, -0.4115, -0.0, 0.0168, -0.4115, -0.0, 0.019, -0.4136, -0.0, 0.0212, -0.4133, -0.0, 0.0239, -0.4105, -0.0, 0.0222, -0.4116, -0.0, 0.0231, -0.4136, 0.0017, 0.024, -0.4136, 0.0039, 0.024, -0.4105, 0.0017, 0.0221, -0.4115, 0.0017, 0.023, -0.4105, 0.0039, 0.0221, -0.4115, 0.0039, 0.023, -0.4093, -0.0, 0.0168, -0.4083, -0.0, 0.0198, -0.4068, -0.0, 0.0171, -0.4075, -0.0, 0.0187, -0.4093, -0.0, 0.0211, -0.4083, 0.0017, 0.0199, -0.4083, 0.0039, 0.0199, -0.4067, 0.0017, 0.0168, -0.4076, 0.0017, 0.0189, -0.4067, 0.0039, 0.0168,
-0.4076, 0.0039, 0.0189, -0.4093, 0.0017, 0.0211, -0.4093, 0.0039, 0.0211, -0.3966, -0.0, -0.0003, -0.3948, -0.0, 0.0007, -0.3939, -0.0, 0.0012, -0.3963, 0.0017, -0.0001, -0.3963, 0.0039, -0.0001, -0.3951, 0.0017, 0.0005, -0.394, 0.0017, 0.0011, -0.3951, 0.0039, 0.0005, -0.394, 0.0039, 0.0011, -0.3919, -0.0, -0.0005, -0.3921, -0.0, 0.0022, -0.3897, -0.0, -0.0005, -0.3897, -0.0, 0.0016, -0.391, -0.0, 0.0029, -0.3899, -0.0, 0.0035, -0.3921, 0.0017, 0.0022, -0.3921, 0.0039, 0.0022, -0.3911, 0.0017, 0.0028, -0.3911, 0.0039, 0.0028, -0.3897, 0.0017, 0.0036, -0.3897, 0.0039, 0.0036, -0.3876, -0.0, -0.0005, -0.3876, -0.0, 0.0016, -0.3854, -0.0, -0.0005, -0.3854, -0.0, 0.0016, -0.388, -0.0, 0.0046, -0.3869, -0.0, 0.0052,
-0.3854, -0.0, 0.0038, -0.3857, -0.0, 0.0059, -0.3881, 0.0017, 0.0045, -0.387, 0.0017, 0.0052, -0.3881, 0.0039, 0.0045, -0.387, 0.0039, 0.0052, -0.3854, 0.0017, 0.0061, -0.3854, 0.0039, 0.0061, -0.3832, -0.0, -0.0005, -0.3832, -0.0, 0.0016, -0.3811, -0.0, -0.0005, -0.3811, -0.0, 0.0016, -0.3832, -0.0, 0.0038, -0.3839, -0.0, 0.007, -0.3811, -0.0, 0.0038, -0.3811, -0.0, 0.006, -0.3841, 0.0017, 0.0069, -0.3841, 0.0039, 0.0069, -0.3828, -0.0, 0.0076, -0.3812, -0.0, 0.0086, -0.38, -0.0, 0.0093, -0.383, 0.0017, 0.0075, -0.383, 0.0039, 0.0075, -0.3811, 0.0017, 0.0086, -0.38, 0.0017, 0.0093, -0.3811, 0.0039, 0.0086, -0.38, 0.0039, 0.0093, -0.4024, -0.0, 0.0452, -0.4026, -0.0, 0.0472, -0.4021, -0.0, 0.0488,
-0.4023, 0.0017, 0.0451, -0.4023, 0.0039, 0.0451, -0.4026, 0.0017, 0.0472, -0.402, 0.0017, 0.0489, -0.4026, 0.0039, 0.0472, -0.402, 0.0039, 0.0489, -0.4011, -0.0, 0.0429, -0.4018, -0.0, 0.0442, -0.3986, -0.0, 0.042, -0.3984, -0.0, 0.045, -0.4006, -0.0, 0.0472, -0.4013, -0.0, 0.05, -0.3984, -0.0, 0.0472, -0.3984, -0.0, 0.0494, -0.4011, 0.0017, 0.0429, -0.4017, 0.0017, 0.044, -0.4011, 0.0039, 0.0429, -0.4017, 0.0039, 0.044, -0.3987, 0.0017, 0.042, -0.3987, 0.0039, 0.042, -0.4014, 0.0017, 0.05, -0.4014, 0.0039, 0.05, -0.4001, -0.0, 0.0509, -0.3986, -0.0, 0.0514, -0.4002, 0.0017, 0.0508, -0.4002, 0.0039, 0.0508, -0.3986, 0.0017, 0.0514, -0.3986, 0.0039, 0.0514, -0.396, -0.0, 0.0423, -0.3963, -0.0, 0.045,
-0.3946, -0.0, 0.0434, -0.3935, -0.0, 0.0444, -0.3963, -0.0, 0.0472, -0.3963, -0.0, 0.0494, -0.3928, -0.0, 0.0472, -0.3941, -0.0, 0.0498, -0.396, 0.0017, 0.0423, -0.396, 0.0039, 0.0423, -0.3946, 0.0017, 0.0435, -0.3935, 0.0017, 0.0444, -0.3946, 0.0039, 0.0435, -0.3935, 0.0039, 0.0444, -0.3928, 0.0017, 0.0472, -0.3938, 0.0017, 0.0494, -0.3928, 0.0039, 0.0472, -0.3938, 0.0039, 0.0494, -0.3959, -0.0, 0.0512, -0.3951, -0.0, 0.0505, -0.3959, 0.0017, 0.0512, -0.3959, 0.0039, 0.0512, -0.3951, 0.0017, 0.0505, -0.3951, 0.0039, 0.0505, -0.3789, -0.0, -0.0005, -0.3789, -0.0, 0.0016, -0.3767, -0.0, -0.0005, -0.3767, -0.0, 0.0016, -0.3789, -0.0, 0.0038, -0.3789, -0.0, 0.006, -0.3767, -0.0, 0.0038, -0.3767, -0.0, 0.006,
-0.3746, -0.0, -0.0005, -0.3746, -0.0, 0.0016, -0.3724, -0.0, -0.0005, -0.3724, -0.0, 0.0016, -0.3746, -0.0, 0.0038, -0.3746, -0.0, 0.006, -0.3724, -0.0, 0.0038, -0.3724, -0.0, 0.006, -0.3789, -0.0, 0.0081, -0.379, -0.0, 0.0098, -0.3767, -0.0, 0.0081, -0.3769, -0.0, 0.0109, -0.3758, -0.0, 0.0115, -0.3789, 0.0017, 0.0099, -0.3789, 0.0039, 0.0099, -0.377, 0.0017, 0.0109, -0.377, 0.0039, 0.0109, -0.3759, 0.0017, 0.0115, -0.3759, 0.0039, 0.0115, -0.3746, -0.0, 0.0081, -0.3746, -0.0, 0.0103, -0.3724, -0.0, 0.0081, -0.3724, -0.0, 0.0103, -0.3747, -0.0, 0.0122, -0.3728, -0.0, 0.0133, -0.3717, -0.0, 0.0139, -0.3746, 0.0017, 0.0123, -0.3746, 0.0039, 0.0123, -0.3729, 0.0017, 0.0132, -0.3718, 0.0017, 0.0138,
-0.3729, 0.0039, 0.0132, -0.3718, 0.0039, 0.0138, -0.3695, -0.0, -0.0016, -0.3702, -0.0, 0.0016, -0.3678, -0.0, -0.0004, -0.368, -0.0, 0.0016, -0.3702, -0.0, 0.0038, -0.3702, -0.0, 0.006, -0.368, -0.0, 0.0038, -0.368, -0.0, 0.006, -0.3694, 0.0017, -0.0015, -0.3694, 0.0039, -0.0015, -0.3681, 0.0017, -0.0006, -0.3681, 0.0039, -0.0006, -0.3667, -0.0, 0.0003, -0.3656, -0.0, 0.001, -0.364, -0.0, 0.0021, -0.3659, -0.0, 0.0038, -0.3659, -0.0, 0.006, -0.363, -0.0, 0.0027, -0.3637, -0.0, 0.006, -0.3667, 0.0017, 0.0003, -0.3656, 0.0017, 0.0011, -0.3667, 0.0039, 0.0003, -0.3656, 0.0039, 0.0011, -0.3639, 0.0017, 0.0021, -0.3639, 0.0039, 0.0021, -0.3628, 0.0017, 0.0029, -0.3628, 0.0039, 0.0029, -0.3702, -0.0, 0.0081,
-0.3702, -0.0, 0.0103, -0.368, -0.0, 0.0081, -0.368, -0.0, 0.0103, -0.3702, -0.0, 0.0125, -0.3705, -0.0, 0.0146, -0.368, -0.0, 0.0125, -0.3686, -0.0, 0.0156, -0.3702, 0.0017, 0.0147, -0.3702, 0.0039, 0.0147, -0.3687, 0.0017, 0.0155, -0.3687, 0.0039, 0.0155, -0.3659, -0.0, 0.0081, -0.3659, -0.0, 0.0103, -0.3637, -0.0, 0.0081, -0.3637, -0.0, 0.0103, -0.3659, -0.0, 0.0125, -0.3659, -0.0, 0.0146, -0.3637, -0.0, 0.0125, -0.3637, -0.0, 0.0146, -0.3675, -0.0, 0.0162, -0.3677, 0.0017, 0.0161, -0.3677, 0.0039, 0.0161, -0.3662, -0.0, 0.0169, -0.3643, -0.0, 0.0179, -0.3633, -0.0, 0.0184, -0.3659, 0.0017, 0.0171, -0.3659, 0.0039, 0.0171, -0.3646, 0.0017, 0.0178, -0.3635, 0.0017, 0.0184, -0.3646, 0.0039, 0.0178,
-0.3635, 0.0039, 0.0184, -0.3613, -0.0, 0.0039, -0.3615, -0.0, 0.006, -0.3601, -0.0, 0.0046, -0.359, -0.0, 0.0053, -0.3615, 0.0017, 0.0037, -0.3615, 0.0039, 0.0037, -0.36, 0.0017, 0.0046, -0.3589, 0.0017, 0.0053, -0.36, 0.0039, 0.0046, -0.3589, 0.0039, 0.0053, -0.3547, -0.0, -0.001, -0.3574, -0.0, 0.0063, -0.3548, 0.0017, -0.0011, -0.3548, 0.0039, -0.0011, -0.3572, 0.0017, 0.0064, -0.3572, 0.0039, 0.0064, -0.3615, -0.0, 0.0081, -0.3615, -0.0, 0.0103, -0.3594, -0.0, 0.0081, -0.3594, -0.0, 0.0103, -0.3615, -0.0, 0.0125, -0.3615, -0.0, 0.0146, -0.3594, -0.0, 0.0125, -0.3594, -0.0, 0.0146, -0.3563, -0.0, 0.0069, -0.3572, -0.0, 0.0103, -0.3547, -0.0, 0.0079, -0.355, -0.0, 0.0103, -0.3572, -0.0, 0.0125,
-0.3572, -0.0, 0.0146, -0.355, -0.0, 0.0125, -0.355, -0.0, 0.0146, -0.3561, 0.0017, 0.0071, -0.3561, 0.0039, 0.0071, -0.355, 0.0017, 0.0077, -0.355, 0.0039, 0.0077, -0.3531, -0.0, 0.0, -0.352, -0.0, 0.0007, -0.3507, -0.0, -0.0005, -0.3509, -0.0, 0.0014, -0.3531, 0.0017, -0.0, -0.352, 0.0017, 0.0007, -0.3531, 0.0039, -0.0, -0.352, 0.0039, 0.0007, -0.3507, 0.0017, 0.0015, -0.3507, 0.0039, 0.0015, -0.3485, -0.0, -0.0005, -0.3491, -0.0, 0.0025, -0.3463, -0.0, -0.0005, -0.3463, -0.0, 0.0016, -0.348, -0.0, 0.0032, -0.3467, -0.0, 0.004, -0.3492, 0.0017, 0.0025, -0.3492, 0.0039, 0.0025, -0.3481, 0.0017, 0.0031, -0.3481, 0.0039, 0.0031, -0.3463, 0.0017, 0.0042, -0.3463, 0.0039, 0.0042, -0.3534, -0.0, 0.0088,
-0.3523, -0.0, 0.0094, -0.3505, -0.0, 0.0105, -0.3529, -0.0, 0.0125, -0.3529, -0.0, 0.0146, -0.3507, -0.0, 0.0125, -0.3507, -0.0, 0.0146, -0.3533, 0.0017, 0.0088, -0.3522, 0.0017, 0.0095, -0.3533, 0.0039, 0.0088, -0.3522, 0.0039, 0.0095, -0.3507, 0.0017, 0.0104, -0.3507, 0.0039, 0.0104, -0.3494, -0.0, 0.0112, -0.3483, -0.0, 0.0119, -0.3485, -0.0, 0.0146, -0.3465, -0.0, 0.0129, -0.3456, -0.0, 0.0134, -0.3493, 0.0017, 0.0112, -0.3493, 0.0039, 0.0112, -0.3483, 0.0017, 0.0119, -0.3483, 0.0039, 0.0119, -0.3464, 0.0017, 0.013, -0.3453, 0.0017, 0.0136, -0.3464, 0.0039, 0.013, -0.3453, 0.0039, 0.0136, -0.3615, -0.0, 0.0168, -0.3619, -0.0, 0.0192, -0.3594, -0.0, 0.0168, -0.3601, -0.0, 0.0202, -0.3591, -0.0, 0.0207,
-0.3615, 0.0017, 0.0194, -0.3615, 0.0039, 0.0194, -0.3604, 0.0017, 0.02, -0.3604, 0.0039, 0.02, -0.3593, 0.0017, 0.0206, -0.3593, 0.0039, 0.0206, -0.3572, -0.0, 0.0168, -0.3572, -0.0, 0.019, -0.355, -0.0, 0.0168, -0.355, -0.0, 0.019, -0.3573, -0.0, 0.0217, -0.3562, -0.0, 0.0223, -0.355, -0.0, 0.0212, -0.3551, -0.0, 0.0228, -0.3573, 0.0017, 0.0217, -0.3562, 0.0017, 0.0223, -0.3573, 0.0039, 0.0217, -0.3562, 0.0039, 0.0223, -0.355, 0.0017, 0.0228, -0.355, 0.0039, 0.0228, -0.3529, -0.0, 0.0168, -0.3529, -0.0, 0.019, -0.3507, -0.0, 0.0168, -0.3507, -0.0, 0.019, -0.3529, -0.0, 0.0212, -0.3531, -0.0, 0.0237, -0.3507, -0.0, 0.0212, -0.3513, -0.0, 0.0245, -0.3529, 0.0017, 0.0238, -0.3529, 0.0039, 0.0238,
-0.3516, 0.0017, 0.0243, -0.3516, 0.0039, 0.0243, -0.3485, -0.0, 0.0168, -0.3485, -0.0, 0.019, -0.3463, -0.0, 0.0168, -0.3463, -0.0, 0.019, -0.3485, -0.0, 0.0212, -0.3485, -0.0, 0.0233, -0.3463, -0.0, 0.0212, -0.3463, -0.0, 0.0233, -0.3503, -0.0, 0.0249, -0.3505, 0.0017, 0.0248, -0.3505, 0.0039, 0.0248, -0.3488, -0.0, 0.0256, -0.3468, -0.0, 0.0265, -0.3458, -0.0, 0.0269, -0.3485, 0.0017, 0.0257, -0.3485, 0.0039, 0.0257, -0.347, 0.0017, 0.0264, -0.3459, 0.0017, 0.0269, -0.347, 0.0039, 0.0264, -0.3459, 0.0039, 0.0269, -0.3715, -0.0, 0.0633, -0.3738, -0.0, 0.0653, -0.3747, -0.0, 0.0664, -0.3727, -0.0, 0.0641, -0.3724, -0.0, 0.0667, -0.3715, 0.0017, 0.0633, -0.3715, 0.0039, 0.0633, -0.3738, 0.0017, 0.0653,
-0.3749, 0.0017, 0.0667, -0.3738, 0.0039, 0.0653, -0.3749, 0.0039, 0.0667, -0.3727, 0.0017, 0.0641, -0.3727, 0.0039, 0.0641, -0.3704, -0.0, 0.0626, -0.3682, -0.0, 0.0619, -0.3702, -0.0, 0.0646, -0.3702, -0.0, 0.0667, -0.368, -0.0, 0.0646, -0.368, -0.0, 0.0667, -0.3703, 0.0017, 0.0625, -0.3703, 0.0039, 0.0625, -0.3681, 0.0017, 0.0619, -0.3681, 0.0039, 0.0619, -0.3659, -0.0, 0.0617, -0.3635, -0.0, 0.062, -0.3659, -0.0, 0.0646, -0.3659, -0.0, 0.0667, -0.3637, -0.0, 0.0646, -0.3637, -0.0, 0.0667, -0.3659, 0.0017, 0.0617, -0.3659, 0.0039, 0.0617, -0.3636, 0.0017, 0.062, -0.3636, 0.0039, 0.062, -0.3549, -0.0, 0.0408, -0.3548, 0.0017, 0.0406, -0.3548, 0.0039, 0.0406, -0.3561, -0.0, 0.0439, -0.3563, -0.0, 0.045,
-0.3558, -0.0, 0.0428, -0.355, -0.0, 0.045, -0.3564, -0.0, 0.0468, -0.3562, -0.0, 0.0476, -0.3558, -0.0, 0.0497, -0.3561, 0.0017, 0.0439, -0.3563, 0.0017, 0.045, -0.3561, 0.0039, 0.0439, -0.3563, 0.0039, 0.045, -0.3558, 0.0017, 0.0427, -0.3558, 0.0039, 0.0427, -0.3563, 0.0017, 0.0472, -0.3563, 0.0039, 0.0472, -0.3561, 0.0017, 0.0483, -0.3559, 0.0017, 0.0494, -0.3561, 0.0039, 0.0483, -0.3559, 0.0039, 0.0494, -0.3501, -0.0, 0.0372, -0.3529, -0.0, 0.0386, -0.354, -0.0, 0.0397, -0.3511, -0.0, 0.0376, -0.3507, -0.0, 0.0407, -0.3502, 0.0017, 0.0372, -0.3502, 0.0039, 0.0372, -0.3533, 0.0017, 0.0389, -0.3539, 0.0017, 0.0396, -0.3533, 0.0039, 0.0389, -0.3539, 0.0039, 0.0396, -0.3513, 0.0017, 0.0377, -0.3513, 0.0039, 0.0377,
-0.3486, -0.0, 0.0367, -0.3464, -0.0, 0.0364, -0.3485, -0.0, 0.0385, -0.3485, -0.0, 0.0407, -0.3463, -0.0, 0.0385, -0.3463, -0.0, 0.0407, -0.3485, 0.0017, 0.0366, -0.3485, 0.0039, 0.0366, -0.3464, 0.0017, 0.0364, -0.3464, 0.0039, 0.0364, -0.3529, -0.0, 0.0429, -0.3529, -0.0, 0.045, -0.3507, -0.0, 0.0429, -0.3507, -0.0, 0.045, -0.3529, -0.0, 0.0472, -0.3529, -0.0, 0.0494, -0.3507, -0.0, 0.0472, -0.3507, -0.0, 0.0494, -0.3485, -0.0, 0.0429, -0.3485, -0.0, 0.045, -0.3463, -0.0, 0.0429, -0.3463, -0.0, 0.045, -0.3485, -0.0, 0.0472, -0.3485, -0.0, 0.0494, -0.3463, -0.0, 0.0472, -0.3463, -0.0, 0.0494, -0.3549, -0.0, 0.0515, -0.3548, 0.0017, 0.0516, -0.3548, 0.0039, 0.0516, -0.3618, -0.0, 0.0628,
-0.3625, -0.0, 0.0622, -0.3607, -0.0, 0.0661, -0.3593, -0.0, 0.0645, -0.3591, -0.0, 0.0664, -0.3617, 0.0017, 0.0629, -0.3617, 0.0039, 0.0629, -0.3625, 0.0017, 0.0622, -0.3605, 0.0017, 0.0665, -0.3625, 0.0039, 0.0622, -0.3605, 0.0039, 0.0665, -0.3591, 0.0017, 0.0647, -0.3592, 0.0017, 0.0663, -0.3591, 0.0039, 0.0647, -0.3592, 0.0039, 0.0663, -0.3577, -0.0, 0.0671, -0.3578, 0.0017, 0.0669, -0.3578, 0.0039, 0.0669, -0.354, -0.0, 0.0525, -0.353, -0.0, 0.0538, -0.3507, -0.0, 0.0515, -0.3512, -0.0, 0.0547, -0.3501, -0.0, 0.0551, -0.3539, 0.0017, 0.0526, -0.3532, 0.0017, 0.0536, -0.3539, 0.0039, 0.0526, -0.3532, 0.0039, 0.0536, -0.3514, 0.0017, 0.0546, -0.3514, 0.0039, 0.0546, -0.3503, 0.0017, 0.0551, -0.3503, 0.0039, 0.0551,
-0.3485, -0.0, 0.0515, -0.3485, -0.0, 0.0537, -0.3463, -0.0, 0.0515, -0.3463, -0.0, 0.0537, -0.3486, -0.0, 0.0554, -0.3463, -0.0, 0.0555, -0.3485, 0.0017, 0.0554, -0.3485, 0.0039, 0.0554, -0.3463, 0.0017, 0.0555, -0.3463, 0.0039, 0.0555, -0.4139, -0.0, 0.0693, -0.4129, -0.0, 0.0699, -0.4111, -0.0, 0.0708, -0.414, -0.0, 0.0738, -0.4133, -0.0, 0.0745, -0.4118, -0.0, 0.074, -0.411, -0.0, 0.0749, -0.4137, 0.0017, 0.0694, -0.4126, 0.0017, 0.07, -0.4137, 0.0039, 0.0694, -0.4126, 0.0039, 0.07, -0.4115, 0.0017, 0.0706, -0.4115, 0.0039, 0.0706, -0.4136, 0.0017, 0.0739, -0.4131, 0.0017, 0.0746, -0.4136, 0.0039, 0.0739, -0.4131, 0.0039, 0.0746, -0.4119, 0.0017, 0.074, -0.4114, 0.0017, 0.0747, -0.4119, 0.0039, 0.074,
-0.4114, 0.0039, 0.0747, -0.4095, -0.0, 0.0716, -0.4086, -0.0, 0.072, -0.4095, -0.0, 0.0757, -0.4068, -0.0, 0.0728, -0.4071, -0.0, 0.0754, -0.4093, 0.0017, 0.0717, -0.4093, 0.0039, 0.0717, -0.4082, 0.0017, 0.0722, -0.4082, 0.0039, 0.0722, -0.4071, 0.0017, 0.0727, -0.4071, 0.0039, 0.0727, -0.4136, -0.0, 0.0776, -0.4136, -0.0, 0.0798, -0.4115, -0.0, 0.0776, -0.4115, -0.0, 0.0798, -0.4136, -0.0, 0.0819, -0.4136, -0.0, 0.0841, -0.4115, -0.0, 0.0819, -0.4115, -0.0, 0.0841, -0.4086, -0.0, 0.0772, -0.4093, -0.0, 0.0798, -0.4081, -0.0, 0.0784, -0.4069, -0.0, 0.0794, -0.4093, -0.0, 0.0819, -0.4093, -0.0, 0.0841, -0.4061, -0.0, 0.0816, -0.4071, -0.0, 0.0841, -0.4068, 0.0017, 0.0797, -0.4068, 0.0039, 0.0797,
-0.4062, 0.0017, 0.0814, -0.4062, 0.0039, 0.0814, -0.4048, -0.0, 0.0735, -0.4049, -0.0, 0.0754, -0.4033, -0.0, 0.074, -0.4024, -0.0, 0.0743, -0.4049, 0.0017, 0.0735, -0.4049, 0.0039, 0.0735, -0.4032, 0.0017, 0.0741, -0.4021, 0.0017, 0.0745, -0.4032, 0.0039, 0.0741, -0.4021, 0.0039, 0.0745, -0.4004, -0.0, 0.075, -0.3983, -0.0, 0.0757, -0.4006, 0.0017, 0.0749, -0.4006, 0.0039, 0.0749, -0.3984, 0.0017, 0.0757, -0.3984, 0.0039, 0.0757, -0.4049, -0.0, 0.0776, -0.4057, -0.0, 0.0802, -0.4028, -0.0, 0.0776, -0.4028, -0.0, 0.0798, -0.4055, -0.0, 0.0823, -0.4054, -0.0, 0.0842, -0.4028, -0.0, 0.0819, -0.4028, -0.0, 0.0841, -0.4057, 0.0017, 0.0803, -0.4057, 0.0039, 0.0803, -0.4056, 0.0017, 0.0819, -0.4056, 0.0039, 0.0819,
-0.4006, -0.0, 0.0776, -0.4006, -0.0, 0.0798, -0.3984, -0.0, 0.0776, -0.3984, -0.0, 0.0798, -0.4006, -0.0, 0.0819, -0.4006, -0.0, 0.0841, -0.3984, -0.0, 0.0819, -0.3984, -0.0, 0.0841, -0.4136, -0.0, 0.0863, -0.4136, -0.0, 0.0884, -0.4115, -0.0, 0.0863, -0.4115, -0.0, 0.0884, -0.4136, -0.0, 0.0906, -0.4136, -0.0, 0.0928, -0.4115, -0.0, 0.0906, -0.4115, -0.0, 0.0928, -0.4093, -0.0, 0.0863, -0.4093, -0.0, 0.0884, -0.4071, -0.0, 0.0863, -0.4071, -0.0, 0.0884, -0.4093, -0.0, 0.0906, -0.4093, -0.0, 0.0928, -0.4071, -0.0, 0.0906, -0.4064, -0.0, 0.0929, -0.4064, 0.0017, 0.0928, -0.4064, 0.0039, 0.0928, -0.4136, -0.0, 0.0949, -0.4136, -0.0, 0.0971, -0.4115, -0.0, 0.0949, -0.4107, -0.0, 0.0981,
-0.4133, -0.0, 0.0997, -0.4118, -0.0, 0.0988, -0.4106, 0.0017, 0.098, -0.4106, 0.0039, 0.098, -0.4136, 0.0017, 0.0998, -0.4136, 0.0039, 0.0998, -0.4117, 0.0017, 0.0987, -0.4117, 0.0039, 0.0987, -0.4083, -0.0, 0.0957, -0.4094, -0.0, 0.097, -0.4075, -0.0, 0.0946, -0.4084, 0.0017, 0.0958, -0.4094, 0.0017, 0.097, -0.4084, 0.0039, 0.0958, -0.4094, 0.0039, 0.097, -0.4075, 0.0017, 0.0947, -0.4075, 0.0039, 0.0947, -0.4045, -0.0, 0.0862, -0.4045, -0.0, 0.0887, -0.4028, -0.0, 0.0863, -0.4029, -0.0, 0.089, -0.4056, -0.0, 0.0907, -0.406, -0.0, 0.0917, -0.4045, 0.0017, 0.0887, -0.4045, 0.0039, 0.0887, -0.4028, 0.0017, 0.089, -0.4028, 0.0039, 0.089, -0.4056, 0.0017, 0.0906, -0.406, 0.0017, 0.0917, -0.4056, 0.0039, 0.0906,
-0.406, 0.0039, 0.0917, -0.4006, -0.0, 0.0863, -0.4005, -0.0, 0.0893, -0.3984, -0.0, 0.0863, -0.3984, -0.0, 0.0884, -0.3995, -0.0, 0.0895, -0.3984, -0.0, 0.0897, -0.4006, 0.0017, 0.0893, -0.4006, 0.0039, 0.0893, -0.3995, 0.0017, 0.0895, -0.3995, 0.0039, 0.0895, -0.3984, 0.0017, 0.0897, -0.3984, 0.0039, 0.0897, -0.3968, -0.0, 0.0762, -0.3966, 0.0017, 0.0763, -0.3966, 0.0039, 0.0763, -0.3959, -0.0, 0.0765, -0.3963, -0.0, 0.0798, -0.3939, -0.0, 0.077, -0.3941, -0.0, 0.0798, -0.3963, -0.0, 0.0819, -0.3963, -0.0, 0.0841, -0.3941, -0.0, 0.0819, -0.3941, -0.0, 0.0841, -0.3955, 0.0017, 0.0766, -0.3955, 0.0039, 0.0766, -0.3941, 0.0017, 0.077, -0.3941, 0.0039, 0.077, -0.3918, -0.0, 0.0775, -0.3919, -0.0, 0.0798,
-0.3897, -0.0, 0.0779, -0.3897, -0.0, 0.0798, -0.3919, -0.0, 0.0819, -0.3919, -0.0, 0.0841, -0.3897, -0.0, 0.0819, -0.3897, -0.0, 0.0841, -0.3919, 0.0017, 0.0775, -0.3919, 0.0039, 0.0775, -0.3898, 0.0017, 0.0779, -0.3898, 0.0039, 0.0779, -0.3875, -0.0, 0.0783, -0.3876, -0.0, 0.0798, -0.3861, -0.0, 0.0785, -0.3852, -0.0, 0.0787, -0.3876, -0.0, 0.0819, -0.3876, -0.0, 0.0841, -0.3854, -0.0, 0.0819, -0.3854, -0.0, 0.0841, -0.3876, 0.0017, 0.0783, -0.3876, 0.0039, 0.0783, -0.3858, 0.0017, 0.0786, -0.3848, 0.0017, 0.0787, -0.3858, 0.0039, 0.0786, -0.3848, 0.0039, 0.0787, -0.3832, -0.0, 0.0789, -0.381, -0.0, 0.079, -0.3832, -0.0, 0.0819, -0.3832, -0.0, 0.0841, -0.3811, -0.0, 0.0819, -0.3811, -0.0, 0.0841,
-0.3832, 0.0017, 0.0789, -0.3832, 0.0039, 0.0789, -0.3811, 0.0017, 0.079, -0.3811, 0.0039, 0.079, -0.3963, -0.0, 0.0863, -0.3963, -0.0, 0.0884, -0.3941, -0.0, 0.0863, -0.3941, -0.0, 0.0884, -0.3963, -0.0, 0.0901, -0.3942, -0.0, 0.0904, -0.3963, 0.0017, 0.0901, -0.3963, 0.0039, 0.0901, -0.3941, 0.0017, 0.0904, -0.3941, 0.0039, 0.0904, -0.3919, -0.0, 0.0863, -0.3919, -0.0, 0.0884, -0.3897, -0.0, 0.0863, -0.3897, -0.0, 0.0884, -0.392, -0.0, 0.0908, -0.3899, -0.0, 0.0911, -0.3919, 0.0017, 0.0908, -0.3919, 0.0039, 0.0908, -0.3898, 0.0017, 0.0912, -0.3898, 0.0039, 0.0912, -0.3876, -0.0, 0.0863, -0.3876, -0.0, 0.0884, -0.3854, -0.0, 0.0863, -0.3854, -0.0, 0.0884, -0.3877, -0.0, 0.0915, -0.3856, -0.0, 0.0917,
-0.3848, -0.0, 0.0918, -0.3876, 0.0017, 0.0915, -0.3876, 0.0039, 0.0915, -0.3861, 0.0017, 0.0917, -0.3851, 0.0017, 0.0917, -0.3861, 0.0039, 0.0917, -0.3851, 0.0039, 0.0917, -0.3832, -0.0, 0.0863, -0.3832, -0.0, 0.0884, -0.3811, -0.0, 0.0863, -0.3811, -0.0, 0.0884, -0.3832, -0.0, 0.0906, -0.3832, -0.0, 0.0918, -0.3811, -0.0, 0.0906, -0.3811, -0.0, 0.0918, -0.3832, 0.0017, 0.0918, -0.3832, 0.0039, 0.0918, -0.3811, 0.0017, 0.0918, -0.3811, 0.0039, 0.0918, -0.3849, -0.0, 0.0954, -0.386, -0.0, 0.0967, -0.3867, -0.0, 0.0991, -0.3862, -0.0, 0.1015, -0.3849, 0.0017, 0.0954, -0.3861, 0.0017, 0.0969, -0.3849, 0.0039, 0.0954, -0.3861, 0.0039, 0.0969, -0.3867, 0.0017, 0.0991, -0.3862, 0.0017, 0.1015, -0.3867, 0.0039, 0.0991,
-0.3862, 0.0039, 0.1015, -0.3832, -0.0, 0.0941, -0.3832, -0.0, 0.0971, -0.3808, -0.0, 0.0943, -0.3811, -0.0, 0.0971, -0.3832, -0.0, 0.0993, -0.3832, -0.0, 0.1015, -0.3811, -0.0, 0.0993, -0.3811, -0.0, 0.1015, -0.3832, 0.0017, 0.0941, -0.3832, 0.0039, 0.0941, -0.3804, 0.0017, 0.0944, -0.3804, 0.0039, 0.0944, -0.3847, -0.0, 0.1029, -0.3848, 0.0017, 0.1029, -0.3848, 0.0039, 0.1029, -0.3834, -0.0, 0.1037, -0.381, -0.0, 0.1039, -0.3832, 0.0017, 0.1037, -0.3832, 0.0039, 0.1037, -0.381, 0.0017, 0.1039, -0.381, 0.0039, 0.1039, -0.3759, -0.0, 0.0695, -0.3761, -0.0, 0.071, -0.376, -0.0, 0.0728, -0.3758, 0.0017, 0.0693, -0.3761, 0.0017, 0.0711, -0.3758, 0.0039, 0.0693, -0.3761, 0.0039, 0.0711, -0.376, 0.0017, 0.0731,
-0.376, 0.0039, 0.0731, -0.3757, -0.0, 0.0685, -0.3746, -0.0, 0.0711, -0.3724, -0.0, 0.0689, -0.3724, -0.0, 0.0711, -0.3758, -0.0, 0.0738, -0.3749, -0.0, 0.0758, -0.3724, -0.0, 0.0732, -0.3724, -0.0, 0.0754, -0.3755, 0.0017, 0.0682, -0.3755, 0.0039, 0.0682, -0.3756, 0.0017, 0.0742, -0.3751, 0.0017, 0.0755, -0.3756, 0.0039, 0.0742, -0.3751, 0.0039, 0.0755, -0.3788, -0.0, 0.0791, -0.3766, -0.0, 0.0793, -0.3789, -0.0, 0.0819, -0.3789, -0.0, 0.0841, -0.3767, -0.0, 0.0819, -0.3767, -0.0, 0.0841, -0.3789, 0.0017, 0.0791, -0.3789, 0.0039, 0.0791, -0.3767, 0.0017, 0.0792, -0.3767, 0.0039, 0.0792, -0.3741, -0.0, 0.0772, -0.3744, -0.0, 0.0799, -0.3731, -0.0, 0.0784, -0.3725, -0.0, 0.0799, -0.3746, -0.0, 0.0819,
-0.3746, -0.0, 0.0841, -0.372, -0.0, 0.0809, -0.3724, -0.0, 0.0841, -0.374, 0.0017, 0.0773, -0.3745, 0.0017, 0.0798, -0.374, 0.0039, 0.0773, -0.3745, 0.0039, 0.0798, -0.3732, 0.0017, 0.0783, -0.3724, 0.0017, 0.0799, -0.3732, 0.0039, 0.0783, -0.3724, 0.0039, 0.0799, -0.3717, 0.0017, 0.081, -0.3717, 0.0039, 0.081, -0.3702, -0.0, 0.0689, -0.3702, -0.0, 0.0711, -0.368, -0.0, 0.0689, -0.368, -0.0, 0.0711, -0.3702, -0.0, 0.0732, -0.3702, -0.0, 0.0754, -0.368, -0.0, 0.0732, -0.368, -0.0, 0.0754, -0.3659, -0.0, 0.0689, -0.3659, -0.0, 0.0711, -0.3637, -0.0, 0.0689, -0.3637, -0.0, 0.0711, -0.3659, -0.0, 0.0732, -0.3659, -0.0, 0.0754, -0.3637, -0.0, 0.0732, -0.3637, -0.0, 0.0754, -0.3702, -0.0, 0.0776,
-0.3705, -0.0, 0.0802, -0.368, -0.0, 0.0776, -0.3682, -0.0, 0.0808, -0.37, -0.0, 0.0817, -0.3702, -0.0, 0.0841, -0.3678, -0.0, 0.0819, -0.368, -0.0, 0.0841, -0.3703, 0.0017, 0.0803, -0.3703, 0.0039, 0.0803, -0.3686, 0.0017, 0.0808, -0.3686, 0.0039, 0.0808, -0.3702, 0.0017, 0.0816, -0.3702, 0.0039, 0.0816, -0.368, 0.0017, 0.0818, -0.368, 0.0039, 0.0818, -0.3659, -0.0, 0.0776, -0.3657, -0.0, 0.0809, -0.3637, -0.0, 0.0776, -0.3634, -0.0, 0.0807, -0.366, -0.0, 0.082, -0.3659, -0.0, 0.0841, -0.3638, -0.0, 0.0821, -0.3637, -0.0, 0.0841, -0.365, 0.0017, 0.0808, -0.365, 0.0039, 0.0808, -0.3637, 0.0017, 0.0807, -0.3637, 0.0039, 0.0807, -0.367, 0.0017, 0.082, -0.367, 0.0039, 0.082, -0.3637, 0.0017, 0.0821,
-0.3637, 0.0039, 0.0821, -0.3789, -0.0, 0.0863, -0.3789, -0.0, 0.0884, -0.3767, -0.0, 0.0863, -0.3767, -0.0, 0.0884, -0.3789, -0.0, 0.0906, -0.3789, -0.0, 0.0918, -0.3767, -0.0, 0.0906, -0.3767, -0.0, 0.0918, -0.3789, 0.0017, 0.0918, -0.3789, 0.0039, 0.0918, -0.3767, 0.0017, 0.0918, -0.3767, 0.0039, 0.0918, -0.3746, -0.0, 0.0863, -0.3746, -0.0, 0.0884, -0.3724, -0.0, 0.0863, -0.3724, -0.0, 0.0884, -0.3746, -0.0, 0.0906, -0.3746, -0.0, 0.0918, -0.3724, -0.0, 0.0906, -0.3724, -0.0, 0.0918, -0.3746, 0.0017, 0.0918, -0.3746, 0.0039, 0.0918, -0.3724, 0.0017, 0.0918, -0.3724, 0.0039, 0.0918, -0.379, -0.0, 0.0951, -0.3779, -0.0, 0.0963, -0.3773, -0.0, 0.0974, -0.3789, -0.0, 0.0993, -0.3782, -0.0, 0.102,
-0.3769, -0.0, 0.0991, -0.3774, -0.0, 0.1009, -0.3789, 0.0017, 0.0952, -0.3779, 0.0017, 0.0962, -0.3789, 0.0039, 0.0952, -0.3779, 0.0039, 0.0962, -0.3773, 0.0017, 0.0972, -0.3773, 0.0039, 0.0972, -0.3782, 0.0017, 0.1021, -0.3782, 0.0039, 0.1021, -0.3769, 0.0017, 0.0991, -0.3776, 0.0017, 0.1014, -0.3769, 0.0039, 0.0991, -0.3776, 0.0039, 0.1014, -0.3702, -0.0, 0.0863, -0.3702, -0.0, 0.0884, -0.368, -0.0, 0.0863, -0.368, -0.0, 0.0884, -0.3703, -0.0, 0.0916, -0.3712, -0.0, 0.0917, -0.3679, -0.0, 0.0913, -0.3701, 0.0017, 0.0916, -0.3712, 0.0017, 0.0917, -0.3701, 0.0039, 0.0916, -0.3712, 0.0039, 0.0917, -0.368, 0.0017, 0.0913, -0.368, 0.0039, 0.0913, -0.3659, -0.0, 0.0863, -0.3659, -0.0, 0.0884, -0.3637, -0.0, 0.0863,
-0.3637, -0.0, 0.0884, -0.3658, -0.0, 0.0908, -0.3636, -0.0, 0.0904, -0.3659, 0.0017, 0.0909, -0.3659, 0.0039, 0.0909, -0.3637, 0.0017, 0.0905, -0.3637, 0.0039, 0.0905, -0.3603, -0.0, 0.0682, -0.3615, -0.0, 0.0711, -0.3596, -0.0, 0.0691, -0.3587, -0.0, 0.0703, -0.3615, -0.0, 0.0732, -0.3615, -0.0, 0.0754, -0.3594, -0.0, 0.0732, -0.3582, -0.0, 0.0763, -0.3605, 0.0017, 0.0679, -0.3605, 0.0039, 0.0679, -0.3588, 0.0017, 0.0704, -0.3588, 0.0039, 0.0704, -0.3583, 0.0017, 0.0765, -0.3583, 0.0039, 0.0765, -0.3571, -0.0, 0.0687, -0.3569, -0.0, 0.071, -0.3571, -0.0, 0.0733, -0.3575, -0.0, 0.0752, -0.3571, 0.0017, 0.0689, -0.3569, 0.0017, 0.0711, -0.3571, 0.0039, 0.0689, -0.3569, 0.0039, 0.0711, -0.357, 0.0017, 0.0733,
-0.3577, 0.0017, 0.0757, -0.357, 0.0039, 0.0733, -0.3577, 0.0039, 0.0757, -0.3615, -0.0, 0.0776, -0.3611, -0.0, 0.0803, -0.3594, -0.0, 0.078, -0.3599, -0.0, 0.0795, -0.3618, -0.0, 0.0812, -0.3615, -0.0, 0.0841, -0.3594, -0.0, 0.0819, -0.3594, -0.0, 0.0841, -0.3614, 0.0017, 0.0805, -0.3614, 0.0039, 0.0805, -0.3591, 0.0017, 0.0777, -0.3596, 0.0017, 0.0793, -0.3591, 0.0039, 0.0777, -0.3596, 0.0039, 0.0793, -0.3619, 0.0017, 0.0812, -0.3619, 0.0039, 0.0812, -0.3569, -0.0, 0.0781, -0.3579, -0.0, 0.0787, -0.3553, -0.0, 0.0773, -0.355, -0.0, 0.0798, -0.3572, -0.0, 0.0819, -0.3572, -0.0, 0.0841, -0.355, -0.0, 0.0819, -0.355, -0.0, 0.0841, -0.357, 0.0017, 0.0781, -0.3581, 0.0017, 0.0788, -0.357, 0.0039, 0.0781,
-0.3581, 0.0039, 0.0788, -0.3551, 0.0017, 0.0772, -0.3551, 0.0039, 0.0772, -0.352, -0.0, 0.0764, -0.3508, -0.0, 0.076, -0.3521, 0.0017, 0.0764, -0.3521, 0.0039, 0.0764, -0.3507, 0.0017, 0.076, -0.3507, 0.0039, 0.076, -0.3487, -0.0, 0.0754, -0.3454, -0.0, 0.0742, -0.3465, -0.0, 0.0746, -0.3485, 0.0017, 0.0753, -0.3485, 0.0039, 0.0753, -0.3455, 0.0017, 0.0742, -0.3466, 0.0017, 0.0746, -0.3455, 0.0039, 0.0742, -0.3466, 0.0039, 0.0746, -0.353, -0.0, 0.0767, -0.3529, -0.0, 0.0798, -0.3507, -0.0, 0.0776, -0.3507, -0.0, 0.0798, -0.3529, -0.0, 0.0819, -0.3529, -0.0, 0.0841, -0.3507, -0.0, 0.0819, -0.3507, -0.0, 0.0841, -0.3532, 0.0017, 0.0767, -0.3532, 0.0039, 0.0767, -0.3485, -0.0, 0.0776, -0.3485, -0.0, 0.0798,
-0.3463, -0.0, 0.0776, -0.3463, -0.0, 0.0798, -0.3485, -0.0, 0.0819, -0.3485, -0.0, 0.0841, -0.3463, -0.0, 0.0819, -0.3463, -0.0, 0.0841, -0.3615, -0.0, 0.0863, -0.3615, -0.0, 0.0884, -0.3594, -0.0, 0.0863, -0.359, -0.0, 0.0896, -0.3615, -0.0, 0.09, -0.3597, -0.0, 0.0898, -0.3583, 0.0017, 0.0895, -0.3583, 0.0039, 0.0895, -0.3615, 0.0017, 0.0901, -0.3615, 0.0039, 0.0901, -0.3594, 0.0017, 0.0897, -0.3594, 0.0039, 0.0897, -0.3572, -0.0, 0.0863, -0.357, -0.0, 0.0893, -0.355, -0.0, 0.0863, -0.3549, -0.0, 0.0889, -0.3572, 0.0017, 0.0893, -0.3572, 0.0039, 0.0893, -0.355, 0.0017, 0.0889, -0.355, 0.0039, 0.0889, -0.3529, -0.0, 0.0863, -0.3527, -0.0, 0.0884, -0.3507, -0.0, 0.0863, -0.3506, -0.0, 0.0879,
-0.3529, 0.0017, 0.0885, -0.3529, 0.0039, 0.0885, -0.3507, 0.0017, 0.0879, -0.3507, 0.0039, 0.0879, -0.3483, -0.0, 0.0871, -0.3494, -0.0, 0.0875, -0.3461, -0.0, 0.0864, -0.3482, 0.0017, 0.0871, -0.3493, 0.0017, 0.0875, -0.3482, 0.0039, 0.0871, -0.3493, 0.0039, 0.0875, -0.3463, 0.0017, 0.0864, -0.3463, 0.0039, 0.0864, -0.3791, -0.0, 0.1031, -0.3791, 0.0017, 0.1031, -0.3791, 0.0039, 0.1031, -0.3442, -0.0, -0.0005, -0.3442, -0.0, 0.0016, -0.342, -0.0, -0.0005, -0.342, -0.0, 0.0016, -0.345, -0.0, 0.005, -0.344, -0.0, 0.0056, -0.342, -0.0, 0.0038, -0.3423, -0.0, 0.0066, -0.3452, 0.0017, 0.0049, -0.3442, 0.0017, 0.0055, -0.3452, 0.0039, 0.0049, -0.3442, 0.0039, 0.0055, -0.3423, 0.0017, 0.0066, -0.3423, 0.0039, 0.0066,
-0.3398, -0.0, -0.0005, -0.3398, -0.0, 0.0016, -0.3377, -0.0, -0.0005, -0.3377, -0.0, 0.0016, -0.3398, -0.0, 0.0038, -0.3398, -0.0, 0.006, -0.3377, -0.0, 0.0038, -0.3377, -0.0, 0.006, -0.3412, -0.0, 0.0072, -0.3438, -0.0, 0.0144, -0.3424, -0.0, 0.0152, -0.3412, 0.0017, 0.0072, -0.3412, 0.0039, 0.0072, -0.3442, 0.0017, 0.0142, -0.3442, 0.0039, 0.0142, -0.3423, 0.0017, 0.0153, -0.3423, 0.0039, 0.0153, -0.34, -0.0, 0.0079, -0.338, -0.0, 0.009, -0.337, -0.0, 0.0095, -0.3398, 0.0017, 0.008, -0.3398, 0.0039, 0.008, -0.3381, 0.0017, 0.0089, -0.3371, 0.0017, 0.0095, -0.3381, 0.0039, 0.0089, -0.3371, 0.0039, 0.0095, -0.3355, -0.0, -0.0005, -0.3355, -0.0, 0.0016, -0.3333, -0.0, -0.0005, -0.3333, -0.0, 0.0016,
-0.3355, -0.0, 0.0038, -0.3355, -0.0, 0.006, -0.3333, -0.0, 0.0038, -0.3333, -0.0, 0.006, -0.3312, -0.0, -0.0005, -0.3312, -0.0, 0.0016, -0.329, -0.0, -0.0005, -0.329, -0.0, 0.0016, -0.3312, -0.0, 0.0038, -0.3312, -0.0, 0.006, -0.329, -0.0, 0.0038, -0.329, -0.0, 0.006, -0.3355, -0.0, 0.0081, -0.3357, -0.0, 0.0101, -0.3333, -0.0, 0.0081, -0.3336, -0.0, 0.0111, -0.3326, -0.0, 0.0116, -0.3355, 0.0017, 0.0102, -0.3355, 0.0039, 0.0102, -0.3337, 0.0017, 0.0111, -0.3337, 0.0039, 0.0111, -0.3327, 0.0017, 0.0116, -0.3327, 0.0039, 0.0116, -0.3312, -0.0, 0.0081, -0.3312, -0.0, 0.0103, -0.329, -0.0, 0.0081, -0.329, -0.0, 0.0103, -0.3313, -0.0, 0.0122, -0.3292, -0.0, 0.0132, -0.3282, -0.0, 0.0137,
-0.3312, 0.0017, 0.0123, -0.3312, 0.0039, 0.0123, -0.3293, 0.0017, 0.0132, -0.3282, 0.0017, 0.0137, -0.3293, 0.0039, 0.0132, -0.3282, 0.0039, 0.0137, -0.3442, -0.0, 0.0168, -0.3442, -0.0, 0.019, -0.3414, -0.0, 0.0158, -0.342, -0.0, 0.019, -0.3442, -0.0, 0.0212, -0.3442, -0.0, 0.0233, -0.342, -0.0, 0.0212, -0.342, -0.0, 0.0233, -0.3412, 0.0017, 0.0159, -0.3412, 0.0039, 0.0159, -0.3396, -0.0, 0.0168, -0.3398, -0.0, 0.019, -0.3383, -0.0, 0.0175, -0.3372, -0.0, 0.0181, -0.3398, -0.0, 0.0212, -0.3398, -0.0, 0.0233, -0.3377, -0.0, 0.0212, -0.3377, -0.0, 0.0233, -0.3398, 0.0017, 0.0167, -0.3398, 0.0039, 0.0167, -0.3382, 0.0017, 0.0176, -0.3371, 0.0017, 0.0181, -0.3382, 0.0039, 0.0176, -0.3371, 0.0039, 0.0181,
-0.3442, -0.0, 0.0255, -0.3444, -0.0, 0.0276, -0.342, -0.0, 0.0255, -0.3422, -0.0, 0.0285, -0.3411, -0.0, 0.0289, -0.3442, 0.0017, 0.0277, -0.3442, 0.0039, 0.0277, -0.3423, 0.0017, 0.0284, -0.3423, 0.0039, 0.0284, -0.3412, 0.0017, 0.0289, -0.3412, 0.0039, 0.0289, -0.3398, -0.0, 0.0255, -0.3398, -0.0, 0.0277, -0.3377, -0.0, 0.0255, -0.3377, -0.0, 0.0277, -0.34, -0.0, 0.0294, -0.3379, -0.0, 0.0302, -0.3398, 0.0017, 0.0294, -0.3398, 0.0039, 0.0294, -0.3377, 0.0017, 0.0303, -0.3377, 0.0039, 0.0303, -0.3353, -0.0, 0.019, -0.3339, -0.0, 0.0197, -0.3355, -0.0, 0.0212, -0.3355, -0.0, 0.0233, -0.3329, -0.0, 0.0202, -0.3333, -0.0, 0.0233, -0.3355, 0.0017, 0.0189, -0.3355, 0.0039, 0.0189, -0.3337, 0.0017, 0.0198,
-0.3337, 0.0039, 0.0198, -0.3327, 0.0017, 0.0203, -0.3327, 0.0039, 0.0203, -0.3309, -0.0, 0.021, -0.3312, -0.0, 0.0233, -0.3289, -0.0, 0.0218, -0.329, -0.0, 0.0233, -0.3312, 0.0017, 0.0209, -0.3312, 0.0039, 0.0209, -0.329, 0.0017, 0.0218, -0.329, 0.0039, 0.0218, -0.3355, -0.0, 0.0255, -0.3355, -0.0, 0.0277, -0.3333, -0.0, 0.0255, -0.3333, -0.0, 0.0277, -0.3359, -0.0, 0.0309, -0.335, -0.0, 0.0313, -0.3333, -0.0, 0.0298, -0.3335, -0.0, 0.0318, -0.3363, 0.0017, 0.0308, -0.3352, 0.0017, 0.0312, -0.3363, 0.0039, 0.0308, -0.3352, 0.0039, 0.0312, -0.3333, 0.0017, 0.0318, -0.3333, 0.0039, 0.0318, -0.3312, -0.0, 0.0255, -0.3312, -0.0, 0.0277, -0.329, -0.0, 0.0255, -0.329, -0.0, 0.0277, -0.3312, -0.0, 0.0298,
-0.3314, -0.0, 0.0324, -0.329, -0.0, 0.0298, -0.3293, -0.0, 0.0331, -0.3312, 0.0017, 0.0325, -0.3312, 0.0039, 0.0325, -0.3296, 0.0017, 0.033, -0.3296, 0.0039, 0.033, -0.3262, -0.0, -0.0015, -0.3268, -0.0, 0.0016, -0.3244, -0.0, -0.0003, -0.3246, -0.0, 0.0016, -0.3268, -0.0, 0.0038, -0.3268, -0.0, 0.006, -0.3246, -0.0, 0.0038, -0.3246, -0.0, 0.006, -0.3261, 0.0017, -0.0014, -0.3261, 0.0039, -0.0014, -0.3246, 0.0017, -0.0004, -0.3246, 0.0039, -0.0004, -0.3234, -0.0, 0.0004, -0.3223, -0.0, 0.0011, -0.3207, -0.0, 0.0021, -0.3225, -0.0, 0.0038, -0.3225, -0.0, 0.006, -0.3197, -0.0, 0.0027, -0.3203, -0.0, 0.006, -0.3234, 0.0017, 0.0004, -0.3223, 0.0017, 0.0011, -0.3234, 0.0039, 0.0004, -0.3223, 0.0039, 0.0011,
-0.3205, 0.0017, 0.0022, -0.3205, 0.0039, 0.0022, -0.3194, 0.0017, 0.0028, -0.3194, 0.0039, 0.0028, -0.3268, -0.0, 0.0081, -0.3268, -0.0, 0.0103, -0.3246, -0.0, 0.0081, -0.3246, -0.0, 0.0103, -0.3268, -0.0, 0.0125, -0.327, -0.0, 0.0143, -0.3246, -0.0, 0.0125, -0.3249, -0.0, 0.0151, -0.3268, 0.0017, 0.0144, -0.3268, 0.0039, 0.0144, -0.3247, 0.0017, 0.0153, -0.3247, 0.0039, 0.0153, -0.3225, -0.0, 0.0081, -0.3225, -0.0, 0.0103, -0.3203, -0.0, 0.0081, -0.3203, -0.0, 0.0103, -0.3225, -0.0, 0.0125, -0.323, -0.0, 0.0159, -0.3203, -0.0, 0.0125, -0.3203, -0.0, 0.0146, -0.3235, 0.0017, 0.0157, -0.3235, 0.0039, 0.0157, -0.3179, -0.0, 0.0037, -0.3181, -0.0, 0.006, -0.3165, -0.0, 0.0044, -0.3154, -0.0, 0.005,
-0.3181, 0.0017, 0.0036, -0.3181, 0.0039, 0.0036, -0.3164, 0.0017, 0.0045, -0.3153, 0.0017, 0.0051, -0.3164, 0.0039, 0.0045, -0.3153, 0.0039, 0.0051, -0.3136, -0.0, 0.0061, -0.3123, -0.0, 0.0067, -0.3138, 0.0017, 0.0059, -0.3138, 0.0039, 0.0059, -0.3122, 0.0017, 0.0068, -0.3122, 0.0039, 0.0068, -0.3181, -0.0, 0.0081, -0.3181, -0.0, 0.0103, -0.316, -0.0, 0.0081, -0.316, -0.0, 0.0103, -0.3181, -0.0, 0.0125, -0.3181, -0.0, 0.0146, -0.316, -0.0, 0.0125, -0.316, -0.0, 0.0146, -0.3138, -0.0, 0.0081, -0.3138, -0.0, 0.0103, -0.3113, -0.0, 0.0073, -0.3116, -0.0, 0.0103, -0.3138, -0.0, 0.0125, -0.3138, -0.0, 0.0146, -0.3116, -0.0, 0.0125, -0.3116, -0.0, 0.0146, -0.3111, 0.0017, 0.0074, -0.3111, 0.0039, 0.0074,
-0.3279, -0.0, 0.0222, -0.3268, -0.0, 0.0227, -0.3245, -0.0, 0.0236, -0.3279, 0.0017, 0.0222, -0.3268, 0.0017, 0.0227, -0.3279, 0.0039, 0.0222, -0.3268, 0.0039, 0.0227, -0.3246, 0.0017, 0.0235, -0.3246, 0.0039, 0.0235, -0.3221, -0.0, 0.0162, -0.3205, -0.0, 0.0168, -0.3231, -0.0, 0.0241, -0.3224, 0.0017, 0.0161, -0.3224, 0.0039, 0.0161, -0.3203, 0.0017, 0.0169, -0.3203, 0.0039, 0.0169, -0.3229, 0.0017, 0.0242, -0.3229, 0.0039, 0.0242, -0.3268, -0.0, 0.0255, -0.3268, -0.0, 0.0277, -0.3246, -0.0, 0.0255, -0.3246, -0.0, 0.0277, -0.3268, -0.0, 0.0298, -0.3268, -0.0, 0.032, -0.3246, -0.0, 0.0298, -0.3246, -0.0, 0.032, -0.3221, -0.0, 0.0245, -0.3225, -0.0, 0.0277, -0.3201, -0.0, 0.0251, -0.3203, -0.0, 0.0277,
-0.3225, -0.0, 0.0298, -0.3225, -0.0, 0.032, -0.3203, -0.0, 0.0298, -0.3203, -0.0, 0.032, -0.3218, 0.0017, 0.0246, -0.3218, 0.0039, 0.0246, -0.3203, 0.0017, 0.0251, -0.3203, 0.0039, 0.0251, -0.3183, -0.0, 0.0176, -0.3173, -0.0, 0.018, -0.316, -0.0, 0.0168, -0.3161, -0.0, 0.0185, -0.3184, 0.0017, 0.0176, -0.3173, 0.0017, 0.018, -0.3184, 0.0039, 0.0176, -0.3173, 0.0039, 0.018, -0.316, 0.0017, 0.0185, -0.316, 0.0039, 0.0185, -0.3138, -0.0, 0.0168, -0.314, -0.0, 0.0192, -0.3116, -0.0, 0.0168, -0.3117, -0.0, 0.0199, -0.3107, -0.0, 0.0201, -0.3138, 0.0017, 0.0192, -0.3138, 0.0039, 0.0192, -0.3118, 0.0017, 0.0198, -0.3118, 0.0039, 0.0198, -0.3107, 0.0017, 0.0201, -0.3107, 0.0039, 0.0201, -0.318, -0.0, 0.0257,
-0.3181, -0.0, 0.0277, -0.3164, -0.0, 0.0262, -0.3155, -0.0, 0.0265, -0.3181, -0.0, 0.0298, -0.3181, -0.0, 0.032, -0.316, -0.0, 0.0298, -0.316, -0.0, 0.032, -0.3181, 0.0017, 0.0257, -0.3181, 0.0039, 0.0257, -0.3162, 0.0017, 0.0263, -0.3151, 0.0017, 0.0267, -0.3162, 0.0039, 0.0263, -0.3151, 0.0039, 0.0267, -0.3136, -0.0, 0.0271, -0.3115, -0.0, 0.0277, -0.3138, -0.0, 0.0298, -0.3138, -0.0, 0.032, -0.3116, -0.0, 0.0298, -0.3116, -0.0, 0.032, -0.3138, 0.0017, 0.0271, -0.3138, 0.0039, 0.0271, -0.3116, 0.0017, 0.0277, -0.3116, 0.0039, 0.0277, -0.3441, -0.0, 0.0367, -0.3428, -0.0, 0.0372, -0.3442, -0.0, 0.0385, -0.3442, -0.0, 0.0407, -0.3417, -0.0, 0.0377, -0.342, -0.0, 0.0407, -0.3441, 0.0017, 0.0367,
-0.3441, 0.0039, 0.0367, -0.3427, 0.0017, 0.0373, -0.3427, 0.0039, 0.0373, -0.3416, 0.0017, 0.0378, -0.3416, 0.0039, 0.0378, -0.3401, -0.0, 0.0388, -0.3389, -0.0, 0.0401, -0.3383, -0.0, 0.0412, -0.3401, 0.0017, 0.0388, -0.339, 0.0017, 0.04, -0.3401, 0.0039, 0.0388, -0.339, 0.0039, 0.04, -0.3384, 0.0017, 0.0411, -0.3384, 0.0039, 0.0411, -0.3442, -0.0, 0.0429, -0.3442, -0.0, 0.045, -0.342, -0.0, 0.0429, -0.342, -0.0, 0.045, -0.3442, -0.0, 0.0472, -0.3442, -0.0, 0.0494, -0.342, -0.0, 0.0472, -0.342, -0.0, 0.0494, -0.3398, -0.0, 0.0429, -0.3398, -0.0, 0.045, -0.3376, -0.0, 0.0426, -0.3373, -0.0, 0.045, -0.3398, -0.0, 0.0472, -0.3398, -0.0, 0.0494, -0.3373, -0.0, 0.0473, -0.3378, -0.0, 0.0496,
-0.3378, 0.0017, 0.0422, -0.3373, 0.0017, 0.045, -0.3378, 0.0039, 0.0422, -0.3373, 0.0039, 0.045, -0.3373, 0.0017, 0.0472, -0.3377, 0.0017, 0.0494, -0.3373, 0.0039, 0.0472, -0.3377, 0.0039, 0.0494, -0.3284, -0.0, 0.0334, -0.3286, 0.0017, 0.0333, -0.3286, 0.0039, 0.0333, -0.3442, -0.0, 0.0515, -0.3442, -0.0, 0.0537, -0.342, -0.0, 0.0515, -0.3418, -0.0, 0.0544, -0.3441, -0.0, 0.0553, -0.343, -0.0, 0.0548, -0.3418, 0.0017, 0.0544, -0.3418, 0.0039, 0.0544, -0.3441, 0.0017, 0.0553, -0.3441, 0.0039, 0.0553, -0.343, 0.0017, 0.0548, -0.343, 0.0039, 0.0548, -0.3391, -0.0, 0.052, -0.3403, -0.0, 0.0532, -0.3385, -0.0, 0.0508, -0.3392, 0.0017, 0.0521, -0.3403, 0.0017, 0.0532, -0.3392, 0.0039, 0.0521, -0.3403, 0.0039, 0.0532,
-0.3385, 0.0017, 0.0509, -0.3385, 0.0039, 0.0509, -0.3309, -0.0, 0.0672, -0.3281, -0.0, 0.0655, -0.3292, -0.0, 0.0662, -0.3311, 0.0017, 0.0673, -0.3311, 0.0039, 0.0673, -0.3282, 0.0017, 0.0655, -0.3292, 0.0017, 0.0662, -0.3282, 0.0039, 0.0655, -0.3292, 0.0039, 0.0662, -0.3269, -0.0, 0.0338, -0.3248, -0.0, 0.0344, -0.3268, 0.0017, 0.0339, -0.3268, 0.0039, 0.0339, -0.3247, 0.0017, 0.0345, -0.3247, 0.0039, 0.0345, -0.3227, -0.0, 0.0349, -0.3206, -0.0, 0.0353, -0.3198, -0.0, 0.0355, -0.3225, 0.0017, 0.035, -0.3225, 0.0039, 0.035, -0.3212, 0.0017, 0.0352, -0.3201, 0.0017, 0.0354, -0.3212, 0.0039, 0.0352, -0.3201, 0.0039, 0.0354, -0.3181, -0.0, 0.0342, -0.3182, -0.0, 0.0358, -0.316, -0.0, 0.0342, -0.3161, -0.0, 0.0363,
-0.3181, 0.0017, 0.0358, -0.3181, 0.0039, 0.0358, -0.316, 0.0017, 0.0363, -0.316, 0.0039, 0.0363, -0.3138, -0.0, 0.0342, -0.3139, -0.0, 0.0366, -0.3116, -0.0, 0.0342, -0.3117, -0.0, 0.0368, -0.3138, 0.0017, 0.0366, -0.3138, 0.0039, 0.0366, -0.3116, 0.0017, 0.0369, -0.3116, 0.0039, 0.0369, -0.3106, -0.0, 0.0504, -0.3106, 0.0017, 0.0504, -0.3106, 0.0039, 0.0504, -0.3193, -0.0, 0.059, -0.3193, 0.0017, 0.059, -0.3193, 0.0039, 0.059, -0.3243, -0.0, 0.0629, -0.3271, -0.0, 0.0648, -0.3268, -0.0, 0.0667, -0.3253, -0.0, 0.0636, -0.3246, -0.0, 0.0667, -0.3243, 0.0017, 0.0629, -0.3243, 0.0039, 0.0629, -0.3268, 0.0017, 0.0646, -0.3268, 0.0039, 0.0646, -0.3254, 0.0017, 0.0637, -0.3254, 0.0039, 0.0637, -0.3217, -0.0, 0.061,
-0.3228, -0.0, 0.0619, -0.3204, -0.0, 0.06, -0.3203, -0.0, 0.0624, -0.3225, -0.0, 0.0646, -0.3225, -0.0, 0.0667, -0.3203, -0.0, 0.0646, -0.3203, -0.0, 0.0667, -0.3217, 0.0017, 0.061, -0.3228, 0.0017, 0.0619, -0.3217, 0.0039, 0.061, -0.3228, 0.0039, 0.0619, -0.3204, 0.0017, 0.06, -0.3204, 0.0039, 0.06, -0.3149, -0.0, 0.0548, -0.3171, -0.0, 0.0569, -0.3182, -0.0, 0.058, -0.316, -0.0, 0.0558, -0.316, -0.0, 0.058, -0.3149, 0.0017, 0.0548, -0.3149, 0.0039, 0.0548, -0.3171, 0.0017, 0.0569, -0.3182, 0.0017, 0.058, -0.3171, 0.0039, 0.0569, -0.3182, 0.0039, 0.058, -0.316, 0.0017, 0.0558, -0.316, 0.0039, 0.0558, -0.3127, -0.0, 0.0526, -0.3138, -0.0, 0.0537, -0.3117, -0.0, 0.0515, -0.3116, -0.0, 0.0537,
-0.3138, -0.0, 0.0559, -0.3138, -0.0, 0.058, -0.3116, -0.0, 0.0559, -0.3116, -0.0, 0.058, -0.3127, 0.0017, 0.0526, -0.3138, 0.0017, 0.0537, -0.3127, 0.0039, 0.0526, -0.3138, 0.0039, 0.0537, -0.3117, 0.0017, 0.0515, -0.3117, 0.0039, 0.0515, -0.3181, -0.0, 0.0602, -0.3181, -0.0, 0.0624, -0.316, -0.0, 0.0602, -0.316, -0.0, 0.0624, -0.3181, -0.0, 0.0646, -0.3181, -0.0, 0.0667, -0.316, -0.0, 0.0646, -0.316, -0.0, 0.0667, -0.3138, -0.0, 0.0602, -0.3138, -0.0, 0.0624, -0.3116, -0.0, 0.0602, -0.3116, -0.0, 0.0624, -0.3138, -0.0, 0.0646, -0.313, -0.0, 0.0676, -0.3107, -0.0, 0.0655, -0.3118, -0.0, 0.0665, -0.313, 0.0017, 0.0676, -0.313, 0.0039, 0.0676, -0.3107, 0.0017, 0.0655, -0.3118, 0.0017, 0.0665,
-0.3107, 0.0039, 0.0655, -0.3118, 0.0039, 0.0665, -0.3093, -0.0, 0.0083, -0.3095, -0.0, 0.0103, -0.3078, -0.0, 0.0089, -0.3069, -0.0, 0.0093, -0.3095, -0.0, 0.0125, -0.3095, -0.0, 0.0146, -0.3073, -0.0, 0.0125, -0.3073, -0.0, 0.0146, -0.3095, 0.0017, 0.0082, -0.3095, 0.0039, 0.0082, -0.3077, 0.0017, 0.0089, -0.3066, 0.0017, 0.0094, -0.3077, 0.0039, 0.0089, -0.3066, 0.0039, 0.0094, -0.3049, -0.0, 0.0101, -0.3029, -0.0, 0.0109, -0.3051, -0.0, 0.0125, -0.3051, -0.0, 0.0146, -0.3029, -0.0, 0.0125, -0.3029, -0.0, 0.0146, -0.3051, 0.0017, 0.01, -0.3051, 0.0039, 0.01, -0.3029, 0.0017, 0.0108, -0.3029, 0.0039, 0.0108, -0.3017, -0.0, 0.0113, -0.3007, -0.0, 0.0117, -0.3008, -0.0, 0.0146, -0.2985, -0.0, 0.0125,
-0.2986, -0.0, 0.0146, -0.3017, 0.0017, 0.0113, -0.3017, 0.0039, 0.0113, -0.3006, 0.0017, 0.0118, -0.3006, 0.0039, 0.0118, -0.2986, 0.0017, 0.0125, -0.2986, 0.0039, 0.0125, -0.2964, -0.0, 0.0131, -0.2964, -0.0, 0.0146, -0.295, -0.0, 0.0134, -0.2941, -0.0, 0.0136, -0.2964, 0.0017, 0.0131, -0.2964, 0.0039, 0.0131, -0.2948, 0.0017, 0.0134, -0.2937, 0.0017, 0.0137, -0.2948, 0.0039, 0.0134, -0.2937, 0.0039, 0.0137, -0.3095, -0.0, 0.0168, -0.3095, -0.0, 0.019, -0.3073, -0.0, 0.0168, -0.3073, -0.0, 0.019, -0.3095, -0.0, 0.0205, -0.3074, -0.0, 0.021, -0.3095, 0.0017, 0.0205, -0.3095, 0.0039, 0.0205, -0.3073, 0.0017, 0.021, -0.3073, 0.0039, 0.021, -0.3051, -0.0, 0.0168, -0.3051, -0.0, 0.019, -0.3029, -0.0, 0.0168,
-0.3029, -0.0, 0.019, -0.3053, -0.0, 0.0215, -0.3031, -0.0, 0.0219, -0.3051, 0.0017, 0.0215, -0.3051, 0.0039, 0.0215, -0.3029, 0.0017, 0.022, -0.3029, 0.0039, 0.022, -0.3094, -0.0, 0.0282, -0.3077, -0.0, 0.0285, -0.3095, -0.0, 0.0298, -0.3095, -0.0, 0.032, -0.3069, -0.0, 0.0287, -0.3073, -0.0, 0.032, -0.3095, 0.0017, 0.0282, -0.3095, 0.0039, 0.0282, -0.3074, 0.0017, 0.0286, -0.3074, 0.0039, 0.0286, -0.3063, 0.0017, 0.0288, -0.3063, 0.0039, 0.0288, -0.305, -0.0, 0.029, -0.3051, -0.0, 0.032, -0.3028, -0.0, 0.0293, -0.3029, -0.0, 0.032, -0.3051, 0.0017, 0.029, -0.3051, 0.0039, 0.029, -0.3029, 0.0017, 0.0293, -0.3029, 0.0039, 0.0293, -0.3008, -0.0, 0.0168, -0.3008, -0.0, 0.019, -0.2986, -0.0, 0.0168,
-0.2986, -0.0, 0.019, -0.301, -0.0, 0.0223, -0.3003, -0.0, 0.0224, -0.2986, -0.0, 0.0212, -0.2986, -0.0, 0.0227, -0.3016, 0.0017, 0.0222, -0.3005, 0.0017, 0.0224, -0.3016, 0.0039, 0.0222, -0.3005, 0.0039, 0.0224, -0.2986, 0.0017, 0.0227, -0.2986, 0.0039, 0.0227, -0.2964, -0.0, 0.0168, -0.2964, -0.0, 0.019, -0.2943, -0.0, 0.0168, -0.2943, -0.0, 0.019, -0.2964, -0.0, 0.0212, -0.2965, -0.0, 0.023, -0.2943, -0.0, 0.0212, -0.2943, -0.0, 0.0232, -0.2964, 0.0017, 0.023, -0.2964, 0.0039, 0.023, -0.2943, 0.0017, 0.0232, -0.2943, 0.0039, 0.0232, -0.3007, -0.0, 0.0297, -0.3008, -0.0, 0.032, -0.2986, -0.0, 0.03, -0.2986, -0.0, 0.032, -0.3008, 0.0017, 0.0297, -0.3008, 0.0039, 0.0297, -0.2986, 0.0017, 0.0299,
-0.2986, 0.0039, 0.0299, -0.2964, -0.0, 0.03, -0.2964, -0.0, 0.032, -0.2943, -0.0, 0.0301, -0.2943, -0.0, 0.032, -0.2964, 0.0017, 0.03, -0.2964, 0.0039, 0.03, -0.2943, 0.0017, 0.0301, -0.2943, 0.0039, 0.0301, -0.289, -0.0, 0.0066, -0.2889, 0.0017, 0.0064, -0.2889, 0.0039, 0.0064, -0.2882, -0.0, -0.0006, -0.2883, -0.0, 0.0016, -0.2856, -0.0, -0.0005, -0.2856, -0.0, 0.0016, -0.2885, -0.0, 0.0037, -0.2888, -0.0, 0.0058, -0.2856, -0.0, 0.0038, -0.2856, -0.0, 0.006, -0.2882, 0.0017, -0.0005, -0.2883, 0.0017, 0.0016, -0.2882, 0.0039, -0.0005, -0.2883, 0.0039, 0.0016, -0.2885, 0.0017, 0.0038, -0.2888, 0.0017, 0.0053, -0.2885, 0.0039, 0.0038, -0.2888, 0.0039, 0.0053, -0.2892, -0.0, 0.0081, -0.2896, -0.0, 0.0102,
-0.2919, -0.0, 0.014, -0.29, -0.0, 0.0124, -0.2902, -0.0, 0.0144, -0.2892, 0.0017, 0.0081, -0.2896, 0.0017, 0.0103, -0.2892, 0.0039, 0.0081, -0.2896, 0.0039, 0.0103, -0.2921, 0.0017, 0.014, -0.2921, 0.0039, 0.014, -0.29, 0.0017, 0.0125, -0.2902, 0.0017, 0.0144, -0.29, 0.0039, 0.0125, -0.2902, 0.0039, 0.0144, -0.2878, -0.0, 0.0081, -0.2878, -0.0, 0.0103, -0.2856, -0.0, 0.0081, -0.2856, -0.0, 0.0103, -0.2878, -0.0, 0.0125, -0.2878, -0.0, 0.0146, -0.2856, -0.0, 0.0125, -0.2856, -0.0, 0.0146, -0.2834, -0.0, -0.0005, -0.2834, -0.0, 0.0016, -0.2812, -0.0, -0.0005, -0.2812, -0.0, 0.0016, -0.2834, -0.0, 0.0038, -0.2834, -0.0, 0.006, -0.2812, -0.0, 0.0038, -0.2801, -0.0, 0.0063, -0.2802, 0.0017, 0.0069,
-0.2802, 0.0039, 0.0069, -0.2794, -0.0, -0.0005, -0.2795, -0.0, 0.0017, -0.2797, -0.0, 0.0038, -0.28, -0.0, 0.0055, -0.2794, 0.0017, -0.0005, -0.2795, 0.0017, 0.0016, -0.2794, 0.0039, -0.0005, -0.2795, 0.0039, 0.0016, -0.2797, 0.0017, 0.0038, -0.28, 0.0017, 0.0058, -0.2797, 0.0039, 0.0038, -0.28, 0.0039, 0.0058, -0.2834, -0.0, 0.0081, -0.2834, -0.0, 0.0103, -0.2804, -0.0, 0.0083, -0.2807, -0.0, 0.0104, -0.2834, -0.0, 0.0125, -0.2834, -0.0, 0.0146, -0.281, -0.0, 0.0126, -0.2813, -0.0, 0.0147, -0.2803, 0.0017, 0.0081, -0.2806, 0.0017, 0.0103, -0.2803, 0.0039, 0.0081, -0.2806, 0.0039, 0.0103, -0.281, 0.0017, 0.0125, -0.2813, 0.0017, 0.0146, -0.281, 0.0039, 0.0125, -0.2813, 0.0039, 0.0146, -0.2921, -0.0, 0.0168,
-0.2921, -0.0, 0.019, -0.2899, -0.0, 0.0168, -0.2899, -0.0, 0.019, -0.2921, -0.0, 0.0212, -0.2921, -0.0, 0.0232, -0.2899, -0.0, 0.0212, -0.2899, -0.0, 0.0232, -0.2921, 0.0017, 0.0232, -0.2921, 0.0039, 0.0232, -0.2899, 0.0017, 0.0232, -0.2899, 0.0039, 0.0232, -0.2878, -0.0, 0.0168, -0.2878, -0.0, 0.019, -0.2856, -0.0, 0.0168, -0.2856, -0.0, 0.019, -0.2878, -0.0, 0.0212, -0.2878, -0.0, 0.0232, -0.2856, -0.0, 0.0212, -0.2856, -0.0, 0.0232, -0.2878, 0.0017, 0.0232, -0.2878, 0.0039, 0.0232, -0.2856, 0.0017, 0.0232, -0.2856, 0.0039, 0.0232, -0.2921, -0.0, 0.0301, -0.2921, -0.0, 0.032, -0.2899, -0.0, 0.0301, -0.2899, -0.0, 0.032, -0.2921, 0.0017, 0.0301, -0.2921, 0.0039, 0.0301, -0.2899, 0.0017, 0.0301,
-0.2899, 0.0039, 0.0301, -0.2878, -0.0, 0.0301, -0.2878, -0.0, 0.032, -0.2856, -0.0, 0.03, -0.2856, -0.0, 0.032, -0.2878, 0.0017, 0.0301, -0.2878, 0.0039, 0.0301, -0.2856, 0.0017, 0.03, -0.2856, 0.0039, 0.03, -0.2834, -0.0, 0.0168, -0.2834, -0.0, 0.019, -0.2814, -0.0, 0.0163, -0.2812, -0.0, 0.019, -0.2834, -0.0, 0.0212, -0.2834, -0.0, 0.0232, -0.2812, -0.0, 0.0212, -0.2812, -0.0, 0.0232, -0.2814, 0.0017, 0.0163, -0.2814, 0.0039, 0.0163, -0.2834, 0.0017, 0.0232, -0.2834, 0.0039, 0.0232, -0.2812, 0.0017, 0.0232, -0.2812, 0.0039, 0.0232, -0.2791, -0.0, 0.0163, -0.2791, -0.0, 0.019, -0.2769, -0.0, 0.0163, -0.2769, -0.0, 0.019, -0.2791, -0.0, 0.0212, -0.279, -0.0, 0.0232, -0.2769, -0.0, 0.0212,
-0.2769, -0.0, 0.023, -0.2791, 0.0017, 0.0163, -0.2791, 0.0039, 0.0163, -0.2769, 0.0017, 0.0163, -0.2769, 0.0039, 0.0163, -0.2791, 0.0017, 0.0232, -0.2791, 0.0039, 0.0232, -0.2769, 0.0017, 0.023, -0.2769, 0.0039, 0.023, -0.2835, -0.0, 0.0299, -0.2834, -0.0, 0.032, -0.2813, -0.0, 0.0296, -0.2812, -0.0, 0.032, -0.2834, 0.0017, 0.0299, -0.2834, 0.0039, 0.0299, -0.2812, 0.0017, 0.0296, -0.2812, 0.0039, 0.0296, -0.2762, -0.0, 0.0286, -0.2792, -0.0, 0.0292, -0.2791, -0.0, 0.032, -0.2771, -0.0, 0.0288, -0.2769, -0.0, 0.032, -0.2764, 0.0017, 0.0286, -0.2764, 0.0039, 0.0286, -0.2791, 0.0017, 0.0292, -0.2791, 0.0039, 0.0292, -0.2774, 0.0017, 0.0289, -0.2774, 0.0039, 0.0289, -0.3095, -0.0, 0.0342, -0.3095, -0.0, 0.037,
-0.3073, -0.0, 0.0342, -0.3074, -0.0, 0.0371, -0.3095, 0.0017, 0.037, -0.3095, 0.0039, 0.037, -0.3073, 0.0017, 0.0371, -0.3073, 0.0039, 0.0371, -0.3051, -0.0, 0.0342, -0.3053, -0.0, 0.0374, -0.3029, -0.0, 0.0342, -0.3029, -0.0, 0.0363, -0.3045, -0.0, 0.0376, -0.301, -0.0, 0.038, -0.3027, -0.0, 0.0405, -0.3058, 0.0017, 0.0374, -0.3058, 0.0039, 0.0374, -0.3047, 0.0017, 0.0375, -0.3047, 0.0039, 0.0375, -0.301, 0.0017, 0.038, -0.3029, 0.0017, 0.0407, -0.301, 0.0039, 0.038, -0.3029, 0.0039, 0.0407, -0.3065, -0.0, 0.0457, -0.3085, -0.0, 0.0481, -0.3096, -0.0, 0.0493, -0.3076, -0.0, 0.047, -0.3073, -0.0, 0.0494, -0.3065, 0.0017, 0.0457, -0.3065, 0.0039, 0.0457, -0.3085, 0.0017, 0.0481, -0.3096, 0.0017, 0.0493,
-0.3085, 0.0039, 0.0481, -0.3096, 0.0039, 0.0493, -0.3076, 0.0017, 0.0471, -0.3076, 0.0039, 0.0471, -0.3047, -0.0, 0.0432, -0.3057, -0.0, 0.0447, -0.3039, -0.0, 0.0422, -0.3029, -0.0, 0.045, -0.3051, -0.0, 0.0472, -0.3051, -0.0, 0.0494, -0.3029, -0.0, 0.0472, -0.3029, -0.0, 0.0494, -0.3046, 0.0017, 0.0432, -0.3057, 0.0017, 0.0446, -0.3046, 0.0039, 0.0432, -0.3057, 0.0039, 0.0446, -0.3038, 0.0017, 0.0421, -0.3038, 0.0039, 0.0421, -0.3008, -0.0, 0.0342, -0.3008, -0.0, 0.0363, -0.2986, -0.0, 0.0342, -0.2986, -0.0, 0.0363, -0.3013, -0.0, 0.0384, -0.3008, -0.0, 0.0407, -0.2986, -0.0, 0.0385, -0.2986, -0.0, 0.0407, -0.2964, -0.0, 0.0342, -0.2964, -0.0, 0.0363, -0.2943, -0.0, 0.0342, -0.2943, -0.0, 0.0363,
-0.2964, -0.0, 0.0385, -0.2964, -0.0, 0.0407, -0.2943, -0.0, 0.0385, -0.2943, -0.0, 0.0407, -0.3008, -0.0, 0.0429, -0.3008, -0.0, 0.045, -0.2986, -0.0, 0.0429, -0.2986, -0.0, 0.045, -0.3008, -0.0, 0.0472, -0.3008, -0.0, 0.0494, -0.2986, -0.0, 0.0472, -0.2986, -0.0, 0.0494, -0.2964, -0.0, 0.0429, -0.2964, -0.0, 0.045, -0.2931, -0.0, 0.0436, -0.2942, -0.0, 0.0453, -0.2956, -0.0, 0.0477, -0.2969, -0.0, 0.0496, -0.295, -0.0, 0.0466, -0.2932, 0.0017, 0.0438, -0.294, 0.0017, 0.045, -0.2932, 0.0039, 0.0438, -0.294, 0.0039, 0.045, -0.2957, 0.0017, 0.0477, -0.2968, 0.0017, 0.0494, -0.2957, 0.0039, 0.0477, -0.2968, 0.0039, 0.0494, -0.295, 0.0017, 0.0467, -0.295, 0.0039, 0.0467, -0.3095, -0.0, 0.0515,
-0.3095, -0.0, 0.0537, -0.3073, -0.0, 0.0515, -0.3073, -0.0, 0.0537, -0.3095, -0.0, 0.0559, -0.3095, -0.0, 0.058, -0.3073, -0.0, 0.0559, -0.3073, -0.0, 0.058, -0.3051, -0.0, 0.0515, -0.3051, -0.0, 0.0537, -0.3029, -0.0, 0.0515, -0.3029, -0.0, 0.0537, -0.3051, -0.0, 0.0559, -0.3042, -0.0, 0.0588, -0.3022, -0.0, 0.0565, -0.3033, -0.0, 0.0577, -0.3043, 0.0017, 0.0589, -0.3043, 0.0039, 0.0589, -0.3022, 0.0017, 0.0565, -0.3033, 0.0017, 0.0578, -0.3022, 0.0039, 0.0565, -0.3033, 0.0039, 0.0578, -0.3095, -0.0, 0.0602, -0.3084, -0.0, 0.0634, -0.3063, -0.0, 0.0611, -0.3073, -0.0, 0.0623, -0.3095, -0.0, 0.0645, -0.3084, 0.0017, 0.0634, -0.3084, 0.0039, 0.0634, -0.3063, 0.0017, 0.0612, -0.3074, 0.0017, 0.0624,
-0.3063, 0.0039, 0.0612, -0.3074, 0.0039, 0.0624, -0.3095, 0.0017, 0.0644, -0.3095, 0.0039, 0.0644, -0.3053, -0.0, 0.0601, -0.3053, 0.0017, 0.0601, -0.3053, 0.0039, 0.0601, -0.3008, -0.0, 0.0515, -0.3002, -0.0, 0.0541, -0.2983, -0.0, 0.0516, -0.2994, -0.0, 0.053, -0.3013, -0.0, 0.0554, -0.3002, 0.0017, 0.0541, -0.3002, 0.0039, 0.0541, -0.2983, 0.0017, 0.0516, -0.2994, 0.0017, 0.053, -0.2983, 0.0039, 0.0516, -0.2994, 0.0039, 0.053, -0.3013, 0.0017, 0.0554, -0.3013, 0.0039, 0.0554, -0.2975, -0.0, 0.0505, -0.2975, 0.0017, 0.0505, -0.2975, 0.0039, 0.0505, -0.2921, -0.0, 0.0342, -0.2921, -0.0, 0.0363, -0.2899, -0.0, 0.0342, -0.2902, -0.0, 0.0369, -0.2908, -0.0, 0.0391, -0.2916, -0.0, 0.041, -0.2905, -0.0, 0.0382,
-0.2902, 0.0017, 0.0369, -0.2902, 0.0039, 0.0369, -0.291, 0.0017, 0.0396, -0.2915, 0.0017, 0.0407, -0.291, 0.0039, 0.0396, -0.2915, 0.0039, 0.0407, -0.2906, 0.0017, 0.0385, -0.2906, 0.0039, 0.0385, -0.2878, -0.0, 0.0342, -0.2877, -0.0, 0.0369, -0.2856, -0.0, 0.0342, -0.2855, -0.0, 0.0367, -0.2878, 0.0017, 0.0369, -0.2878, 0.0039, 0.0369, -0.2856, 0.0017, 0.0367, -0.2856, 0.0039, 0.0367, -0.2925, -0.0, 0.0426, -0.2926, 0.0017, 0.0428, -0.2926, 0.0039, 0.0428, -0.2834, -0.0, 0.0342, -0.2833, -0.0, 0.0363, -0.2812, -0.0, 0.0342, -0.2812, -0.0, 0.0359, -0.2834, 0.0017, 0.0363, -0.2834, 0.0039, 0.0363, -0.2812, 0.0017, 0.0359, -0.2812, 0.0039, 0.0359, -0.2787, -0.0, 0.0353, -0.2796, -0.0, 0.0355, -0.2767, -0.0, 0.0347,
-0.2783, 0.0017, 0.0352, -0.2794, 0.0017, 0.0355, -0.2783, 0.0039, 0.0352, -0.2794, 0.0039, 0.0355, -0.2769, 0.0017, 0.0348, -0.2769, 0.0039, 0.0348, -0.3443, -0.0, 0.0738, -0.3442, -0.0, 0.0754, -0.3423, -0.0, 0.0729, -0.342, -0.0, 0.0754, -0.3442, 0.0017, 0.0737, -0.3442, 0.0039, 0.0737, -0.342, 0.0017, 0.0728, -0.342, 0.0039, 0.0728, -0.3395, -0.0, 0.0716, -0.338, -0.0, 0.0708, -0.3405, -0.0, 0.0721, -0.3398, -0.0, 0.0754, -0.3377, -0.0, 0.0732, -0.3377, -0.0, 0.0754, -0.3397, 0.0017, 0.0717, -0.3397, 0.0039, 0.0717, -0.3377, 0.0017, 0.0707, -0.3377, 0.0039, 0.0707, -0.3408, 0.0017, 0.0722, -0.3408, 0.0039, 0.0722, -0.3442, -0.0, 0.0776, -0.3442, -0.0, 0.0798, -0.342, -0.0, 0.0776, -0.342, -0.0, 0.0798,
-0.3442, -0.0, 0.0819, -0.3442, -0.0, 0.0841, -0.342, -0.0, 0.0819, -0.3419, -0.0, 0.0849, -0.3419, 0.0017, 0.0848, -0.3419, 0.0039, 0.0848, -0.3398, -0.0, 0.0776, -0.3398, -0.0, 0.0798, -0.3377, -0.0, 0.0776, -0.3377, -0.0, 0.0798, -0.3398, -0.0, 0.0819, -0.3396, -0.0, 0.0841, -0.3371, -0.0, 0.0831, -0.3381, -0.0, 0.0835, -0.3398, 0.0017, 0.0841, -0.3398, 0.0039, 0.0841, -0.3368, 0.0017, 0.0829, -0.3379, 0.0017, 0.0834, -0.3368, 0.0039, 0.0829, -0.3379, 0.0039, 0.0834, -0.3351, -0.0, 0.0694, -0.3361, -0.0, 0.0699, -0.3336, -0.0, 0.0687, -0.3333, -0.0, 0.0711, -0.3355, -0.0, 0.0732, -0.3355, -0.0, 0.0754, -0.3333, -0.0, 0.0732, -0.3333, -0.0, 0.0754, -0.3353, 0.0017, 0.0695, -0.3364, 0.0017, 0.0701,
-0.3353, 0.0039, 0.0695, -0.3364, 0.0039, 0.0701, -0.3333, 0.0017, 0.0685, -0.3333, 0.0039, 0.0685, -0.3319, -0.0, 0.0677, -0.3312, -0.0, 0.0711, -0.329, -0.0, 0.0689, -0.329, -0.0, 0.0711, -0.3312, -0.0, 0.0732, -0.3312, -0.0, 0.0754, -0.329, -0.0, 0.0732, -0.329, -0.0, 0.0754, -0.3321, 0.0017, 0.0679, -0.3321, 0.0039, 0.0679, -0.3355, -0.0, 0.0776, -0.3355, -0.0, 0.0798, -0.3333, -0.0, 0.0776, -0.3327, -0.0, 0.0809, -0.3352, -0.0, 0.0822, -0.3336, -0.0, 0.0814, -0.3324, 0.0017, 0.0808, -0.3324, 0.0039, 0.0808, -0.3355, 0.0017, 0.0823, -0.3355, 0.0039, 0.0823, -0.3335, 0.0017, 0.0813, -0.3335, 0.0039, 0.0813, -0.3312, -0.0, 0.0776, -0.3308, -0.0, 0.08, -0.3284, -0.0, 0.0786, -0.3294, -0.0, 0.0792,
-0.3311, 0.0017, 0.0802, -0.3311, 0.0039, 0.0802, -0.3282, 0.0017, 0.0785, -0.3293, 0.0017, 0.0791, -0.3282, 0.0039, 0.0785, -0.3293, 0.0039, 0.0791, -0.3441, -0.0, 0.0856, -0.343, -0.0, 0.0852, -0.3442, 0.0017, 0.0856, -0.3442, 0.0039, 0.0856, -0.343, 0.0017, 0.0852, -0.343, 0.0039, 0.0852, -0.3268, -0.0, 0.0689, -0.3268, -0.0, 0.0711, -0.3246, -0.0, 0.0689, -0.3246, -0.0, 0.0711, -0.3268, -0.0, 0.0732, -0.3268, -0.0, 0.0754, -0.3246, -0.0, 0.0732, -0.3243, -0.0, 0.0761, -0.3242, 0.0017, 0.0761, -0.3242, 0.0039, 0.0761, -0.3225, -0.0, 0.0689, -0.3225, -0.0, 0.0711, -0.3203, -0.0, 0.0689, -0.3203, -0.0, 0.0711, -0.3217, -0.0, 0.0743, -0.3227, -0.0, 0.0751, -0.32, -0.0, 0.0732, -0.3215, 0.0017, 0.0742,
-0.3226, 0.0017, 0.075, -0.3215, 0.0039, 0.0742, -0.3226, 0.0039, 0.075, -0.3203, 0.0017, 0.0734, -0.3203, 0.0039, 0.0734, -0.3265, -0.0, 0.0775, -0.3254, -0.0, 0.0768, -0.3268, 0.0017, 0.0777, -0.3268, 0.0039, 0.0777, -0.3253, 0.0017, 0.0768, -0.3253, 0.0039, 0.0768, -0.3181, -0.0, 0.0689, -0.3178, -0.0, 0.0716, -0.3153, -0.0, 0.0696, -0.3164, -0.0, 0.0706, -0.3189, -0.0, 0.0724, -0.3178, 0.0017, 0.0716, -0.3178, 0.0039, 0.0716, -0.3153, 0.0017, 0.0696, -0.3164, 0.0017, 0.0705, -0.3153, 0.0039, 0.0696, -0.3164, 0.0039, 0.0705, -0.3189, 0.0017, 0.0724, -0.3189, 0.0039, 0.0724, -0.3141, -0.0, 0.0686, -0.3141, 0.0017, 0.0685, -0.3141, 0.0039, 0.0685, -0.2652, -0.0, 0.0157, -0.2639, -0.0, 0.0154, -0.2653, 0.0017, 0.0157,
-0.2653, 0.0039, 0.0157, -0.2639, 0.0017, 0.0154, -0.2639, 0.0039, 0.0154, -0.2618, -0.0, 0.015, -0.2596, -0.0, 0.0146, -0.2617, 0.0017, 0.015, -0.2617, 0.0039, 0.015, -0.2595, 0.0017, 0.0146, -0.2595, 0.0039, 0.0146, -0.2747, -0.0, 0.0163, -0.2747, -0.0, 0.019, -0.2726, -0.0, 0.0163, -0.2726, -0.0, 0.019, -0.2747, -0.0, 0.0212, -0.2747, -0.0, 0.0226, -0.2725, -0.0, 0.0221, -0.2735, -0.0, 0.0223, -0.2747, 0.0017, 0.0163, -0.2747, 0.0039, 0.0163, -0.2726, 0.0017, 0.0163, -0.2726, 0.0039, 0.0163, -0.2747, 0.0017, 0.0226, -0.2747, 0.0039, 0.0226, -0.2723, 0.0017, 0.022, -0.2734, 0.0017, 0.0223, -0.2723, 0.0039, 0.022, -0.2734, 0.0039, 0.0223, -0.2704, -0.0, 0.0163, -0.2704, -0.0, 0.019, -0.2683, -0.0, 0.0161,
-0.2682, -0.0, 0.019, -0.2702, -0.0, 0.0216, -0.2681, -0.0, 0.0211, -0.2704, 0.0017, 0.0163, -0.2704, 0.0039, 0.0163, -0.2682, 0.0017, 0.0161, -0.2682, 0.0039, 0.0161, -0.2704, 0.0017, 0.0216, -0.2704, 0.0039, 0.0216, -0.2682, 0.0017, 0.0211, -0.2682, 0.0039, 0.0211, -0.2748, -0.0, 0.0283, -0.2727, -0.0, 0.0277, -0.2747, -0.0, 0.0298, -0.2747, -0.0, 0.032, -0.2726, -0.0, 0.0298, -0.2721, -0.0, 0.0331, -0.2747, 0.0017, 0.0283, -0.2747, 0.0039, 0.0283, -0.2726, 0.0017, 0.0277, -0.2726, 0.0039, 0.0277, -0.2717, 0.0017, 0.033, -0.2717, 0.0039, 0.033, -0.2706, -0.0, 0.0271, -0.2678, -0.0, 0.0262, -0.2687, -0.0, 0.0265, -0.2704, -0.0, 0.0298, -0.2701, -0.0, 0.0324, -0.2682, -0.0, 0.0298, -0.2681, -0.0, 0.0316,
-0.2704, 0.0017, 0.027, -0.2704, 0.0039, 0.027, -0.2681, 0.0017, 0.0263, -0.2692, 0.0017, 0.0266, -0.2681, 0.0039, 0.0263, -0.2692, 0.0039, 0.0266, -0.2704, 0.0017, 0.0325, -0.2704, 0.0039, 0.0325, -0.2682, 0.0017, 0.0317, -0.2682, 0.0039, 0.0317, -0.2661, -0.0, 0.0158, -0.2661, -0.0, 0.019, -0.2639, -0.0, 0.0168, -0.2637, -0.0, 0.0199, -0.266, -0.0, 0.0206, -0.2647, -0.0, 0.0202, -0.2664, 0.0017, 0.0159, -0.2664, 0.0039, 0.0159, -0.2635, 0.0017, 0.0198, -0.2635, 0.0039, 0.0198, -0.266, 0.0017, 0.0206, -0.266, 0.0039, 0.0206, -0.2646, 0.0017, 0.0202, -0.2646, 0.0039, 0.0202, -0.2617, -0.0, 0.0168, -0.2615, -0.0, 0.0192, -0.2595, -0.0, 0.0168, -0.2594, -0.0, 0.0185, -0.2614, -0.0, 0.024, -0.2598, -0.0, 0.0233,
-0.2617, 0.0017, 0.0193, -0.2617, 0.0039, 0.0193, -0.2595, 0.0017, 0.0185, -0.2595, 0.0039, 0.0185, -0.2616, 0.0017, 0.0241, -0.2616, 0.0039, 0.0241, -0.2595, 0.0017, 0.0232, -0.2595, 0.0039, 0.0232, -0.2662, -0.0, 0.0256, -0.2661, -0.0, 0.0277, -0.2641, -0.0, 0.025, -0.2639, -0.0, 0.0277, -0.2658, -0.0, 0.0307, -0.2669, -0.0, 0.0311, -0.2637, -0.0, 0.0297, -0.2661, 0.0017, 0.0256, -0.2661, 0.0039, 0.0256, -0.2639, 0.0017, 0.0249, -0.2639, 0.0039, 0.0249, -0.2657, 0.0017, 0.0306, -0.2668, 0.0017, 0.0311, -0.2657, 0.0039, 0.0306, -0.2668, 0.0039, 0.0311, -0.2639, 0.0017, 0.0298, -0.2639, 0.0039, 0.0298, -0.2622, -0.0, 0.0243, -0.2613, -0.0, 0.0286, -0.2595, -0.0, 0.0255, -0.2593, -0.0, 0.0275, -0.2623, -0.0, 0.0291,
-0.2627, 0.0017, 0.0245, -0.2611, 0.0017, 0.0285, -0.2627, 0.0039, 0.0245, -0.2611, 0.0039, 0.0285, -0.2595, 0.0017, 0.0276, -0.2595, 0.0039, 0.0276, -0.2622, 0.0017, 0.029, -0.2622, 0.0039, 0.029, -0.2575, -0.0, 0.0142, -0.2541, -0.0, 0.0135, -0.2551, -0.0, 0.0138, -0.2574, 0.0017, 0.0142, -0.2574, 0.0039, 0.0142, -0.2542, 0.0017, 0.0136, -0.2552, 0.0017, 0.0138, -0.2542, 0.0039, 0.0136, -0.2552, 0.0039, 0.0138, -0.2531, -0.0, 0.0133, -0.2527, -0.0, 0.0155, -0.2509, -0.0, 0.0128, -0.2506, -0.0, 0.0144, -0.253, 0.0017, 0.0133, -0.2526, 0.0017, 0.0154, -0.253, 0.0039, 0.0133, -0.2526, 0.0039, 0.0154, -0.2509, 0.0017, 0.0128, -0.2508, 0.0017, 0.0145, -0.2509, 0.0039, 0.0128, -0.2508, 0.0039, 0.0145, -0.2497, -0.0, 0.0131,
-0.2496, -0.0, 0.0137, -0.2467, 0.0017, 0.0121, -0.2496, 0.0017, 0.0137, -0.2467, 0.0039, 0.0121, -0.2496, 0.0039, 0.0137, -0.2572, -0.0, 0.0176, -0.2583, -0.0, 0.018, -0.255, -0.0, 0.0166, -0.2568, -0.0, 0.0219, -0.2578, -0.0, 0.0224, -0.2554, -0.0, 0.0212, -0.2544, -0.0, 0.0245, -0.2571, 0.0017, 0.0176, -0.2582, 0.0017, 0.018, -0.2571, 0.0039, 0.0176, -0.2582, 0.0039, 0.018, -0.2552, 0.0017, 0.0167, -0.2552, 0.0039, 0.0167, -0.2569, 0.0017, 0.022, -0.258, 0.0017, 0.0225, -0.2569, 0.0039, 0.022, -0.258, 0.0039, 0.0225, -0.2552, 0.0017, 0.0211, -0.2542, 0.0017, 0.0244, -0.2552, 0.0039, 0.0211, -0.2542, 0.0039, 0.0244, -0.2538, -0.0, 0.016, -0.2524, -0.0, 0.0197, -0.2511, -0.0, 0.0191, -0.2534, -0.0, 0.0203,
-0.2527, -0.0, 0.0233, -0.2505, -0.0, 0.0217, -0.2516, -0.0, 0.0225, -0.2537, 0.0017, 0.016, -0.2525, 0.0017, 0.0198, -0.2537, 0.0039, 0.016, -0.2525, 0.0039, 0.0198, -0.2509, 0.0017, 0.019, -0.2509, 0.0039, 0.019, -0.2536, 0.0017, 0.0203, -0.253, 0.0017, 0.0235, -0.2536, 0.0039, 0.0203, -0.253, 0.0039, 0.0235, -0.2505, 0.0017, 0.0217, -0.2516, 0.0017, 0.0225, -0.2505, 0.0039, 0.0217, -0.2516, 0.0039, 0.0225, -0.257, -0.0, 0.0262, -0.2581, -0.0, 0.0268, -0.2554, -0.0, 0.0252, -0.257, 0.0017, 0.0262, -0.2581, 0.0017, 0.0268, -0.257, 0.0039, 0.0262, -0.2581, 0.0039, 0.0268, -0.2553, 0.0017, 0.0251, -0.2553, 0.0039, 0.0251, -0.248, -0.0, 0.0176, -0.2486, -0.0, 0.019, -0.2449, -0.0, 0.0163, -0.2469, -0.0, 0.0186,
-0.2491, -0.0, 0.0207, -0.2481, 0.0017, 0.0176, -0.2486, 0.0017, 0.0189, -0.2481, 0.0039, 0.0176, -0.2486, 0.0039, 0.0189, -0.2449, 0.0017, 0.0163, -0.2469, 0.0017, 0.0186, -0.2449, 0.0039, 0.0163, -0.2469, 0.0039, 0.0186, -0.2491, 0.0017, 0.0206, -0.2491, 0.0039, 0.0206, -0.2746, -0.0, 0.034, -0.273, -0.0, 0.0335, -0.2747, 0.0017, 0.0341, -0.2747, 0.0039, 0.0341, -0.2728, 0.0017, 0.0334, -0.2728, 0.0039, 0.0334, -0.2196, -0.0, -0.0005, -0.2196, -0.0, 0.0016, -0.2196, -0.0, 0.0038, -0.2196, -0.0, 0.006, -0.2196, 0.0017, -0.0005, -0.2196, 0.0017, 0.0016, -0.2196, 0.0039, -0.0005, -0.2196, 0.0039, 0.0016, -0.2196, 0.0017, 0.0038, -0.2196, 0.0017, 0.006, -0.2196, 0.0039, 0.0038, -0.2196, 0.0039, 0.006, -0.2183, -0.0, -0.0005,
-0.2183, -0.0, 0.0016, -0.2161, -0.0, -0.0005, -0.2161, -0.0, 0.0016, -0.2183, -0.0, 0.0038, -0.2183, -0.0, 0.006, -0.2161, -0.0, 0.0038, -0.2161, -0.0, 0.006, -0.2196, -0.0, 0.0081, -0.2196, -0.0, 0.0103, -0.2196, -0.0, 0.0125, -0.2196, -0.0, 0.0146, -0.2196, 0.0017, 0.0081, -0.2196, 0.0017, 0.0103, -0.2196, 0.0039, 0.0081, -0.2196, 0.0039, 0.0103, -0.2196, 0.0017, 0.0125, -0.2196, 0.0017, 0.0146, -0.2196, 0.0039, 0.0125, -0.2196, 0.0039, 0.0146, -0.2183, -0.0, 0.0081, -0.2183, -0.0, 0.0103, -0.2161, -0.0, 0.0081, -0.2161, -0.0, 0.0103, -0.2183, -0.0, 0.0125, -0.2183, -0.0, 0.0146, -0.2161, -0.0, 0.0125, -0.2161, -0.0, 0.0146, -0.214, -0.0, -0.0005, -0.214, -0.0, 0.0016, -0.2118, -0.0, -0.0005,
-0.2118, -0.0, 0.0016, -0.214, -0.0, 0.0038, -0.214, -0.0, 0.006, -0.2118, -0.0, 0.0038, -0.2118, -0.0, 0.006, -0.2088, -0.0, -0.0005, -0.2088, -0.0, 0.0016, -0.2088, -0.0, 0.0038, -0.2088, -0.0, 0.006, -0.2088, 0.0017, -0.0005, -0.2088, 0.0017, 0.0016, -0.2088, 0.0039, -0.0005, -0.2088, 0.0039, 0.0016, -0.2088, 0.0017, 0.0038, -0.2088, 0.0017, 0.006, -0.2088, 0.0039, 0.0038, -0.2088, 0.0039, 0.006, -0.214, -0.0, 0.0081, -0.214, -0.0, 0.0103, -0.2118, -0.0, 0.0081, -0.2118, -0.0, 0.0103, -0.214, -0.0, 0.0125, -0.214, -0.0, 0.0146, -0.2118, -0.0, 0.0125, -0.2118, -0.0, 0.0146, -0.2088, -0.0, 0.0085, -0.2096, -0.0, 0.0103, -0.2075, -0.0, 0.0085, -0.2075, -0.0, 0.0103, -0.2096, -0.0, 0.0125,
-0.2096, -0.0, 0.0146, -0.2075, -0.0, 0.0125, -0.2075, -0.0, 0.0146, -0.2088, 0.0017, 0.0085, -0.2088, 0.0039, 0.0085, -0.2075, 0.0017, 0.0085, -0.2075, 0.0039, 0.0085, -0.2196, -0.0, 0.0168, -0.2196, -0.0, 0.019, -0.2196, -0.0, 0.0212, -0.2196, -0.0, 0.0233, -0.2196, 0.0017, 0.0168, -0.2196, 0.0017, 0.019, -0.2196, 0.0039, 0.0168, -0.2196, 0.0039, 0.019, -0.2196, 0.0017, 0.0212, -0.2196, 0.0017, 0.0233, -0.2196, 0.0039, 0.0212, -0.2196, 0.0039, 0.0233, -0.2183, -0.0, 0.0168, -0.2183, -0.0, 0.019, -0.2161, -0.0, 0.0168, -0.2161, -0.0, 0.019, -0.2183, -0.0, 0.0212, -0.2183, -0.0, 0.0233, -0.2161, -0.0, 0.0212, -0.2161, -0.0, 0.0233, -0.2196, -0.0, 0.0255, -0.2196, -0.0, 0.0277, -0.2196, -0.0, 0.0298,
-0.2196, -0.0, 0.032, -0.2196, 0.0017, 0.0255, -0.2196, 0.0017, 0.0277, -0.2196, 0.0039, 0.0255, -0.2196, 0.0039, 0.0277, -0.2196, 0.0017, 0.0298, -0.2196, 0.0017, 0.032, -0.2196, 0.0039, 0.0298, -0.2196, 0.0039, 0.032, -0.2183, -0.0, 0.0255, -0.2183, -0.0, 0.0277, -0.2161, -0.0, 0.0255, -0.2161, -0.0, 0.0277, -0.2183, -0.0, 0.0298, -0.2183, -0.0, 0.032, -0.2161, -0.0, 0.0298, -0.2161, -0.0, 0.032, -0.214, -0.0, 0.0168, -0.214, -0.0, 0.019, -0.2118, -0.0, 0.0168, -0.2118, -0.0, 0.019, -0.214, -0.0, 0.0212, -0.214, -0.0, 0.0233, -0.2118, -0.0, 0.0212, -0.2118, -0.0, 0.0233, -0.2096, -0.0, 0.0168, -0.2096, -0.0, 0.019, -0.2075, -0.0, 0.0168, -0.2075, -0.0, 0.019, -0.2088, -0.0, 0.0203,
-0.2088, -0.0, 0.0233, -0.2075, -0.0, 0.0203, -0.2088, 0.0017, 0.0203, -0.2088, 0.0017, 0.0233, -0.2088, 0.0039, 0.0203, -0.2088, 0.0039, 0.0233, -0.2075, 0.0017, 0.0203, -0.2075, 0.0039, 0.0203, -0.214, -0.0, 0.0255, -0.214, -0.0, 0.0277, -0.2118, -0.0, 0.0255, -0.2118, -0.0, 0.0277, -0.214, -0.0, 0.0298, -0.214, -0.0, 0.032, -0.2118, -0.0, 0.0298, -0.2118, -0.0, 0.032, -0.2088, -0.0, 0.0255, -0.2088, -0.0, 0.0277, -0.2088, -0.0, 0.0298, -0.2088, -0.0, 0.0319, -0.2088, 0.0017, 0.0255, -0.2088, 0.0017, 0.0277, -0.2088, 0.0039, 0.0255, -0.2088, 0.0039, 0.0277, -0.2088, 0.0017, 0.0298, -0.2088, 0.0017, 0.032, -0.2088, 0.0039, 0.0298, -0.2088, 0.0039, 0.032, -0.2196, -0.0, 0.0342, -0.2196, -0.0, 0.0363,
-0.2196, -0.0, 0.0385, -0.2196, -0.0, 0.0407, -0.2196, 0.0017, 0.0342, -0.2196, 0.0017, 0.0363, -0.2196, 0.0039, 0.0342, -0.2196, 0.0039, 0.0363, -0.2196, 0.0017, 0.0385, -0.2196, 0.0017, 0.0407, -0.2196, 0.0039, 0.0385, -0.2196, 0.0039, 0.0407, -0.2183, -0.0, 0.0342, -0.2183, -0.0, 0.0363, -0.2161, -0.0, 0.0342, -0.2161, -0.0, 0.0363, -0.2183, -0.0, 0.0385, -0.2183, -0.0, 0.0407, -0.2161, -0.0, 0.0385, -0.2161, -0.0, 0.0407, -0.2196, -0.0, 0.0429, -0.2196, -0.0, 0.045, -0.2196, -0.0, 0.0472, -0.2196, -0.0, 0.0487, -0.2196, 0.0017, 0.0429, -0.2196, 0.0017, 0.045, -0.2196, 0.0039, 0.0429, -0.2196, 0.0039, 0.045, -0.2196, 0.0017, 0.0472, -0.2196, 0.0017, 0.0487, -0.2196, 0.0039, 0.0472, -0.2196, 0.0039, 0.0487,
-0.2183, -0.0, 0.0429, -0.2183, -0.0, 0.045, -0.2161, -0.0, 0.0429, -0.2161, -0.0, 0.045, -0.2183, -0.0, 0.0472, -0.2183, -0.0, 0.0487, -0.2161, -0.0, 0.0472, -0.2161, -0.0, 0.0487, -0.2183, 0.0017, 0.0487, -0.2183, 0.0039, 0.0487, -0.2161, 0.0017, 0.0487, -0.2161, 0.0039, 0.0487, -0.214, -0.0, 0.0342, -0.214, -0.0, 0.0363, -0.2118, -0.0, 0.0342, -0.2118, -0.0, 0.0363, -0.214, -0.0, 0.0385, -0.214, -0.0, 0.0407, -0.2118, -0.0, 0.0385, -0.2118, -0.0, 0.0407, -0.2085, -0.0, 0.0339, -0.2096, -0.0, 0.0363, -0.2083, -0.0, 0.0347, -0.2076, -0.0, 0.0361, -0.2096, -0.0, 0.0385, -0.2096, -0.0, 0.0407, -0.2075, -0.0, 0.0385, -0.2075, -0.0, 0.0407, -0.2086, 0.0017, 0.0334, -0.2086, 0.0039, 0.0334,
-0.2083, 0.0017, 0.0345, -0.2076, 0.0017, 0.0361, -0.2083, 0.0039, 0.0345, -0.2076, 0.0039, 0.0361, -0.214, -0.0, 0.0429, -0.214, -0.0, 0.045, -0.2118, -0.0, 0.0429, -0.2118, -0.0, 0.045, -0.214, -0.0, 0.0472, -0.214, -0.0, 0.0487, -0.2118, -0.0, 0.0472, -0.2118, -0.0, 0.0487, -0.214, 0.0017, 0.0487, -0.214, 0.0039, 0.0487, -0.2118, 0.0017, 0.0487, -0.2118, 0.0039, 0.0487, -0.2096, -0.0, 0.0429, -0.2096, -0.0, 0.045, -0.2075, -0.0, 0.0429, -0.2075, -0.0, 0.045, -0.2096, -0.0, 0.0472, -0.2096, -0.0, 0.0487, -0.2075, -0.0, 0.0472, -0.2075, -0.0, 0.0487, -0.2096, 0.0017, 0.0487, -0.2096, 0.0039, 0.0487, -0.2075, 0.0017, 0.0487, -0.2075, 0.0039, 0.0487, -0.2053, -0.0, 0.0085, -0.2053, -0.0, 0.0103,
-0.2031, -0.0, 0.0085, -0.2031, -0.0, 0.0103, -0.2053, -0.0, 0.0125, -0.2053, -0.0, 0.0146, -0.2031, -0.0, 0.0125, -0.2031, -0.0, 0.0146, -0.2053, 0.0017, 0.0085, -0.2053, 0.0039, 0.0085, -0.2031, 0.0017, 0.0085, -0.2031, 0.0039, 0.0085, -0.2009, -0.0, 0.0085, -0.2009, -0.0, 0.0103, -0.1988, -0.0, 0.0085, -0.1988, -0.0, 0.0103, -0.2009, -0.0, 0.0125, -0.2009, -0.0, 0.0146, -0.1988, -0.0, 0.0125, -0.1988, -0.0, 0.0146, -0.2009, 0.0017, 0.0085, -0.2009, 0.0039, 0.0085, -0.1988, 0.0017, 0.0085, -0.1988, 0.0039, 0.0085, -0.1966, -0.0, 0.0085, -0.1966, -0.0, 0.0103, -0.1944, -0.0, 0.0085, -0.1944, -0.0, 0.0103, -0.1966, -0.0, 0.0125, -0.1966, -0.0, 0.0146, -0.1944, -0.0, 0.0125, -0.1944, -0.0, 0.0146,
-0.1966, 0.0017, 0.0085, -0.1966, 0.0039, 0.0085, -0.1944, 0.0017, 0.0085, -0.1944, 0.0039, 0.0085, -0.1923, -0.0, 0.0085, -0.1923, -0.0, 0.0103, -0.1901, -0.0, 0.0085, -0.1901, -0.0, 0.0103, -0.1923, -0.0, 0.0125, -0.1923, -0.0, 0.0146, -0.1901, -0.0, 0.0125, -0.1901, -0.0, 0.0146, -0.1923, 0.0017, 0.0085, -0.1923, 0.0039, 0.0085, -0.1901, 0.0017, 0.0085, -0.1901, 0.0039, 0.0085, -0.2053, -0.0, 0.0168, -0.2053, -0.0, 0.019, -0.2031, -0.0, 0.0168, -0.2031, -0.0, 0.019, -0.2053, -0.0, 0.0203, -0.2031, -0.0, 0.0203, -0.2053, 0.0017, 0.0203, -0.2053, 0.0039, 0.0203, -0.2031, 0.0017, 0.0203, -0.2031, 0.0039, 0.0203, -0.2009, -0.0, 0.0168, -0.2009, -0.0, 0.019, -0.1988, -0.0, 0.0168, -0.1988, -0.0, 0.019,
-0.2009, -0.0, 0.0203, -0.1988, -0.0, 0.0203, -0.2009, 0.0017, 0.0203, -0.2009, 0.0039, 0.0203, -0.1988, 0.0017, 0.0203, -0.1988, 0.0039, 0.0203, -0.1966, -0.0, 0.0168, -0.1966, -0.0, 0.019, -0.1944, -0.0, 0.0168, -0.1944, -0.0, 0.019, -0.1966, -0.0, 0.0203, -0.1944, -0.0, 0.0203, -0.1966, 0.0017, 0.0203, -0.1966, 0.0039, 0.0203, -0.1944, 0.0017, 0.0203, -0.1944, 0.0039, 0.0203, -0.1923, -0.0, 0.0168, -0.1923, -0.0, 0.019, -0.1901, -0.0, 0.0168, -0.1901, -0.0, 0.019, -0.1923, -0.0, 0.0203, -0.1901, -0.0, 0.0203, -0.1923, 0.0017, 0.0203, -0.1923, 0.0039, 0.0203, -0.1901, 0.0017, 0.0203, -0.1901, 0.0039, 0.0203, -0.1879, -0.0, 0.0085, -0.1879, -0.0, 0.0103, -0.1858, -0.0, 0.0085, -0.1858, -0.0, 0.0103,
-0.1879, -0.0, 0.0125, -0.1879, -0.0, 0.0146, -0.1858, -0.0, 0.0125, -0.1858, -0.0, 0.0146, -0.1879, 0.0017, 0.0085, -0.1879, 0.0039, 0.0085, -0.1858, 0.0017, 0.0085, -0.1858, 0.0039, 0.0085, -0.1836, -0.0, 0.0085, -0.1836, -0.0, 0.0103, -0.1814, -0.0, 0.0085, -0.1814, -0.0, 0.0103, -0.1836, -0.0, 0.0125, -0.1836, -0.0, 0.0146, -0.1814, -0.0, 0.0125, -0.1814, -0.0, 0.0146, -0.1836, 0.0017, 0.0085, -0.1836, 0.0039, 0.0085, -0.1814, 0.0017, 0.0085, -0.1814, 0.0039, 0.0085, -0.1792, -0.0, 0.0085, -0.1792, -0.0, 0.0103, -0.1771, -0.0, 0.0085, -0.1771, -0.0, 0.0103, -0.1792, -0.0, 0.0125, -0.1792, -0.0, 0.0146, -0.1771, -0.0, 0.0125, -0.1771, -0.0, 0.0146, -0.1792, 0.0017, 0.0085, -0.1792, 0.0039, 0.0085,
-0.1771, 0.0017, 0.0085, -0.1771, 0.0039, 0.0085, -0.1749, -0.0, 0.0085, -0.1749, -0.0, 0.0103, -0.1727, -0.0, 0.0085, -0.1727, -0.0, 0.0103, -0.1749, -0.0, 0.0125, -0.1749, -0.0, 0.0146, -0.1727, -0.0, 0.0125, -0.1727, -0.0, 0.0146, -0.1749, 0.0017, 0.0085, -0.1749, 0.0039, 0.0085, -0.1727, 0.0017, 0.0085, -0.1727, 0.0039, 0.0085, -0.1879, -0.0, 0.0168, -0.1879, -0.0, 0.019, -0.1858, -0.0, 0.0168, -0.1858, -0.0, 0.019, -0.1879, -0.0, 0.0203, -0.1858, -0.0, 0.0203, -0.1879, 0.0017, 0.0203, -0.1879, 0.0039, 0.0203, -0.1858, 0.0017, 0.0203, -0.1858, 0.0039, 0.0203, -0.1836, -0.0, 0.0168, -0.1836, -0.0, 0.019, -0.1814, -0.0, 0.0168, -0.1814, -0.0, 0.019, -0.1836, -0.0, 0.0203, -0.1814, -0.0, 0.0203,
-0.1836, 0.0017, 0.0203, -0.1836, 0.0039, 0.0203, -0.1814, 0.0017, 0.0203, -0.1814, 0.0039, 0.0203, -0.1792, -0.0, 0.0168, -0.1792, -0.0, 0.019, -0.1771, -0.0, 0.0168, -0.1771, -0.0, 0.019, -0.1792, -0.0, 0.0203, -0.1771, -0.0, 0.0203, -0.1792, 0.0017, 0.0203, -0.1792, 0.0039, 0.0203, -0.1771, 0.0017, 0.0203, -0.1771, 0.0039, 0.0203, -0.1749, -0.0, 0.0168, -0.1749, -0.0, 0.019, -0.1727, -0.0, 0.0168, -0.1727, -0.0, 0.019, -0.1749, -0.0, 0.0203, -0.1727, -0.0, 0.0203, -0.1749, 0.0017, 0.0203, -0.1749, 0.0039, 0.0203, -0.1727, 0.0017, 0.0203, -0.1727, 0.0039, 0.0203, -0.206, -0.0, 0.0372, -0.2051, -0.0, 0.0375, -0.2053, -0.0, 0.0407, -0.203, -0.0, 0.0378, -0.2031, -0.0, 0.0407, -0.2059, 0.0017, 0.0372,
-0.2059, 0.0039, 0.0372, -0.2048, 0.0017, 0.0376, -0.2048, 0.0039, 0.0376, -0.2031, 0.0017, 0.0378, -0.2031, 0.0039, 0.0378, -0.2009, -0.0, 0.0379, -0.2009, -0.0, 0.0407, -0.1988, -0.0, 0.0379, -0.1988, -0.0, 0.0407, -0.2009, 0.0017, 0.0379, -0.2009, 0.0039, 0.0379, -0.1988, 0.0017, 0.0379, -0.1988, 0.0039, 0.0379, -0.2053, -0.0, 0.0429, -0.2053, -0.0, 0.045, -0.2031, -0.0, 0.0429, -0.2031, -0.0, 0.045, -0.2053, -0.0, 0.0472, -0.2053, -0.0, 0.0487, -0.2031, -0.0, 0.0472, -0.2031, -0.0, 0.0487, -0.2053, 0.0017, 0.0487, -0.2053, 0.0039, 0.0487, -0.2031, 0.0017, 0.0487, -0.2031, 0.0039, 0.0487, -0.2009, -0.0, 0.0429, -0.2009, -0.0, 0.045, -0.1988, -0.0, 0.0429, -0.1988, -0.0, 0.045, -0.2009, -0.0, 0.0472,
-0.2009, -0.0, 0.0487, -0.1988, -0.0, 0.0472, -0.1988, -0.0, 0.0487, -0.2009, 0.0017, 0.0487, -0.2009, 0.0039, 0.0487, -0.1988, 0.0017, 0.0487, -0.1988, 0.0039, 0.0487, -0.1966, -0.0, 0.0379, -0.1966, -0.0, 0.0407, -0.1944, -0.0, 0.0379, -0.1944, -0.0, 0.0407, -0.1966, 0.0017, 0.0379, -0.1966, 0.0039, 0.0379, -0.1944, 0.0017, 0.0379, -0.1944, 0.0039, 0.0379, -0.1923, -0.0, 0.0379, -0.1923, -0.0, 0.0407, -0.1901, -0.0, 0.0379, -0.1901, -0.0, 0.0407, -0.1923, 0.0017, 0.0379, -0.1923, 0.0039, 0.0379, -0.1901, 0.0017, 0.0379, -0.1901, 0.0039, 0.0379, -0.1966, -0.0, 0.0429, -0.1966, -0.0, 0.045, -0.1944, -0.0, 0.0429, -0.1944, -0.0, 0.045, -0.1966, -0.0, 0.0472, -0.1966, -0.0, 0.0487, -0.1944, -0.0, 0.0472,
-0.1944, -0.0, 0.0487, -0.1966, 0.0017, 0.0487, -0.1966, 0.0039, 0.0487, -0.1944, 0.0017, 0.0487, -0.1944, 0.0039, 0.0487, -0.1923, -0.0, 0.0429, -0.1923, -0.0, 0.045, -0.1901, -0.0, 0.0429, -0.1901, -0.0, 0.045, -0.1923, -0.0, 0.0472, -0.1923, -0.0, 0.0487, -0.1901, -0.0, 0.0472, -0.1901, -0.0, 0.0487, -0.1923, 0.0017, 0.0487, -0.1923, 0.0039, 0.0487, -0.1901, 0.0017, 0.0487, -0.1901, 0.0039, 0.0487, -0.1879, -0.0, 0.0379, -0.1879, -0.0, 0.0407, -0.1858, -0.0, 0.0379, -0.1858, -0.0, 0.0407, -0.1879, 0.0017, 0.0379, -0.1879, 0.0039, 0.0379, -0.1858, 0.0017, 0.0379, -0.1858, 0.0039, 0.0379, -0.1836, -0.0, 0.0379, -0.1836, -0.0, 0.0407, -0.1814, -0.0, 0.0379, -0.1814, -0.0, 0.0407, -0.1836, 0.0017, 0.0379,
-0.1836, 0.0039, 0.0379, -0.1814, 0.0017, 0.0379, -0.1814, 0.0039, 0.0379, -0.1879, -0.0, 0.0429, -0.1879, -0.0, 0.045, -0.1858, -0.0, 0.0429, -0.1858, -0.0, 0.045, -0.1879, -0.0, 0.0472, -0.1879, -0.0, 0.0487, -0.1858, -0.0, 0.0472, -0.1858, -0.0, 0.0487, -0.1879, 0.0017, 0.0487, -0.1879, 0.0039, 0.0487, -0.1858, 0.0017, 0.0487, -0.1858, 0.0039, 0.0487, -0.1836, -0.0, 0.0429, -0.1836, -0.0, 0.045, -0.1814, -0.0, 0.0429, -0.1814, -0.0, 0.045, -0.1836, -0.0, 0.0472, -0.1836, -0.0, 0.0487, -0.1814, -0.0, 0.0472, -0.1814, -0.0, 0.0487, -0.1836, 0.0017, 0.0487, -0.1836, 0.0039, 0.0487, -0.1814, 0.0017, 0.0487, -0.1814, 0.0039, 0.0487, -0.1792, -0.0, 0.0379, -0.1792, -0.0, 0.0407, -0.1771, -0.0, 0.0379,
-0.1771, -0.0, 0.0407, -0.1792, 0.0017, 0.0379, -0.1792, 0.0039, 0.0379, -0.1771, 0.0017, 0.0379, -0.1771, 0.0039, 0.0379, -0.1749, -0.0, 0.0379, -0.1749, -0.0, 0.0407, -0.1727, -0.0, 0.0379, -0.1727, -0.0, 0.0407, -0.1749, 0.0017, 0.0379, -0.1749, 0.0039, 0.0379, -0.1727, 0.0017, 0.0379, -0.1727, 0.0039, 0.0379, -0.1792, -0.0, 0.0429, -0.1792, -0.0, 0.045, -0.1771, -0.0, 0.0429, -0.1771, -0.0, 0.045, -0.1792, -0.0, 0.0472, -0.1792, -0.0, 0.0487, -0.1771, -0.0, 0.0472, -0.1771, -0.0, 0.0487, -0.1792, 0.0017, 0.0487, -0.1792, 0.0039, 0.0487, -0.1771, 0.0017, 0.0487, -0.1771, 0.0039, 0.0487, -0.1749, -0.0, 0.0429, -0.1749, -0.0, 0.045, -0.1727, -0.0, 0.0429, -0.1727, -0.0, 0.045, -0.1749, -0.0, 0.0472,
-0.1749, -0.0, 0.0487, -0.1727, -0.0, 0.0472, -0.1727, -0.0, 0.0487, -0.1749, 0.0017, 0.0487, -0.1749, 0.0039, 0.0487, -0.1727, 0.0017, 0.0487, -0.1727, 0.0039, 0.0487, -0.1706, -0.0, 0.0085, -0.1706, -0.0, 0.0103, -0.1684, -0.0, 0.0085, -0.1684, -0.0, 0.0103, -0.1706, -0.0, 0.0125, -0.1706, -0.0, 0.0146, -0.1684, -0.0, 0.0125, -0.1684, -0.0, 0.0146, -0.1706, 0.0017, 0.0085, -0.1706, 0.0039, 0.0085, -0.1684, 0.0017, 0.0085, -0.1684, 0.0039, 0.0085, -0.1662, -0.0, 0.0085, -0.1662, -0.0, 0.0103, -0.1641, -0.0, 0.0085, -0.1641, -0.0, 0.0103, -0.1662, -0.0, 0.0125, -0.1662, -0.0, 0.0146, -0.1641, -0.0, 0.0125, -0.1641, -0.0, 0.0146, -0.1662, 0.0017, 0.0085, -0.1662, 0.0039, 0.0085, -0.1641, 0.0017, 0.0085,
-0.1641, 0.0039, 0.0085, -0.1627, -0.0, 0.0085, -0.1627, -0.0, 0.0103, -0.1627, -0.0, 0.0125, -0.1627, -0.0, 0.0146, -0.1627, 0.0017, 0.0085, -0.1627, 0.0017, 0.0103, -0.1627, 0.0039, 0.0085, -0.1627, 0.0039, 0.0103, -0.1627, 0.0017, 0.0125, -0.1627, 0.0017, 0.0146, -0.1627, 0.0039, 0.0125, -0.1627, 0.0039, 0.0146, -0.1706, -0.0, 0.0168, -0.1706, -0.0, 0.019, -0.1684, -0.0, 0.0168, -0.1684, -0.0, 0.019, -0.1706, -0.0, 0.0203, -0.1684, -0.0, 0.0203, -0.1706, 0.0017, 0.0203, -0.1706, 0.0039, 0.0203, -0.1684, 0.0017, 0.0203, -0.1684, 0.0039, 0.0203, -0.1662, -0.0, 0.0168, -0.1662, -0.0, 0.019, -0.1641, -0.0, 0.0168, -0.1641, -0.0, 0.019, -0.1662, -0.0, 0.0203, -0.1641, -0.0, 0.0203, -0.1662, 0.0017, 0.0203,
-0.1662, 0.0039, 0.0203, -0.1641, 0.0017, 0.0203, -0.1641, 0.0039, 0.0203, -0.1627, -0.0, 0.0168, -0.1627, -0.0, 0.019, -0.1627, -0.0, 0.0203, -0.1627, 0.0017, 0.0168, -0.1627, 0.0017, 0.019, -0.1627, 0.0039, 0.0168, -0.1627, 0.0039, 0.019, -0.1627, 0.0017, 0.0203, -0.1627, 0.0039, 0.0203, -0.151, -0.0, -0.0005, -0.151, -0.0, 0.0016, -0.151, -0.0, 0.0038, -0.151, -0.0, 0.006, -0.151, 0.0017, -0.0005, -0.151, 0.0017, 0.0016, -0.151, 0.0039, -0.0005, -0.151, 0.0039, 0.0016, -0.151, 0.0017, 0.0038, -0.151, 0.0017, 0.006, -0.151, 0.0039, 0.0038, -0.151, 0.0039, 0.006, -0.1489, -0.0, -0.0005, -0.1489, -0.0, 0.0016, -0.1467, -0.0, -0.0005, -0.1467, -0.0, 0.0016, -0.1489, -0.0, 0.0038, -0.1489, -0.0, 0.006,
-0.1467, -0.0, 0.0038, -0.1467, -0.0, 0.006, -0.151, -0.0, 0.0081, -0.151, -0.0, 0.0103, -0.151, -0.0, 0.0125, -0.151, -0.0, 0.0146, -0.151, 0.0017, 0.0081, -0.151, 0.0017, 0.0103, -0.151, 0.0039, 0.0081, -0.151, 0.0039, 0.0103, -0.151, 0.0017, 0.0125, -0.151, 0.0017, 0.0146, -0.151, 0.0039, 0.0125, -0.151, 0.0039, 0.0146, -0.1489, -0.0, 0.0081, -0.1489, -0.0, 0.0103, -0.1467, -0.0, 0.0081, -0.1467, -0.0, 0.0103, -0.1489, -0.0, 0.0125, -0.1489, -0.0, 0.0146, -0.1467, -0.0, 0.0125, -0.1467, -0.0, 0.0146, -0.1445, -0.0, -0.0005, -0.1445, -0.0, 0.0016, -0.1424, -0.0, -0.0005, -0.1424, -0.0, 0.0016, -0.1445, -0.0, 0.0038, -0.1445, -0.0, 0.006, -0.1424, -0.0, 0.0038, -0.1424, -0.0, 0.006,
-0.1392, -0.0, -0.0005, -0.1392, -0.0, 0.0016, -0.1392, -0.0, 0.0038, -0.1392, -0.0, 0.006, -0.1392, 0.0017, -0.0005, -0.1392, 0.0017, 0.0016, -0.1392, 0.0039, -0.0005, -0.1392, 0.0039, 0.0016, -0.1392, 0.0017, 0.0038, -0.1392, 0.0017, 0.006, -0.1392, 0.0039, 0.0038, -0.1392, 0.0039, 0.006, -0.1445, -0.0, 0.0081, -0.1445, -0.0, 0.0103, -0.1424, -0.0, 0.0081, -0.1424, -0.0, 0.0103, -0.1445, -0.0, 0.0125, -0.1445, -0.0, 0.0146, -0.1424, -0.0, 0.0125, -0.1424, -0.0, 0.0146, -0.1392, -0.0, 0.0081, -0.1392, -0.0, 0.0103, -0.1392, -0.0, 0.0125, -0.1392, -0.0, 0.0146, -0.1392, 0.0017, 0.0081, -0.1392, 0.0017, 0.0103, -0.1392, 0.0039, 0.0081, -0.1392, 0.0039, 0.0103, -0.1392, 0.0017, 0.0125, -0.1392, 0.0017, 0.0146,
-0.1392, 0.0039, 0.0125, -0.1392, 0.0039, 0.0146, -0.151, -0.0, 0.0168, -0.151, -0.0, 0.019, -0.151, -0.0, 0.0212, -0.151, -0.0, 0.0233, -0.151, 0.0017, 0.0168, -0.151, 0.0017, 0.019, -0.151, 0.0039, 0.0168, -0.151, 0.0039, 0.019, -0.151, 0.0017, 0.0212, -0.151, 0.0017, 0.0233, -0.151, 0.0039, 0.0212, -0.151, 0.0039, 0.0233, -0.1489, -0.0, 0.0168, -0.1489, -0.0, 0.019, -0.1467, -0.0, 0.0168, -0.1467, -0.0, 0.019, -0.1489, -0.0, 0.0212, -0.1489, -0.0, 0.0233, -0.1467, -0.0, 0.0212, -0.1467, -0.0, 0.0233, -0.151, -0.0, 0.0255, -0.151, -0.0, 0.0277, -0.151, -0.0, 0.0298, -0.151, -0.0, 0.032, -0.151, 0.0017, 0.0255, -0.151, 0.0017, 0.0277, -0.151, 0.0039, 0.0255, -0.151, 0.0039, 0.0277,
-0.151, 0.0017, 0.0298, -0.151, 0.0017, 0.032, -0.151, 0.0039, 0.0298, -0.151, 0.0039, 0.032, -0.1489, -0.0, 0.0255, -0.1489, -0.0, 0.0277, -0.1467, -0.0, 0.0255, -0.1467, -0.0, 0.0277, -0.1489, -0.0, 0.0298, -0.1489, -0.0, 0.032, -0.1467, -0.0, 0.0298, -0.1467, -0.0, 0.032, -0.1445, -0.0, 0.0168, -0.1445, -0.0, 0.019, -0.1424, -0.0, 0.0168, -0.1424, -0.0, 0.019, -0.1445, -0.0, 0.0212, -0.1445, -0.0, 0.0233, -0.1424, -0.0, 0.0212, -0.1424, -0.0, 0.0233, -0.1392, -0.0, 0.0168, -0.1392, -0.0, 0.019, -0.1392, -0.0, 0.0212, -0.1392, -0.0, 0.0233, -0.1392, 0.0017, 0.0168, -0.1392, 0.0017, 0.019, -0.1392, 0.0039, 0.0168, -0.1392, 0.0039, 0.019, -0.1392, 0.0017, 0.0212, -0.1392, 0.0017, 0.0233,
-0.1392, 0.0039, 0.0212, -0.1392, 0.0039, 0.0233, -0.1445, -0.0, 0.0255, -0.1445, -0.0, 0.0277, -0.1424, -0.0, 0.0255, -0.1424, -0.0, 0.0277, -0.1445, -0.0, 0.0298, -0.1445, -0.0, 0.032, -0.1424, -0.0, 0.0298, -0.1424, -0.0, 0.032, -0.1392, -0.0, 0.0255, -0.1392, -0.0, 0.0277, -0.1392, -0.0, 0.0298, -0.1392, -0.0, 0.032, -0.1392, 0.0017, 0.0255, -0.1392, 0.0017, 0.0277, -0.1392, 0.0039, 0.0255, -0.1392, 0.0039, 0.0277, -0.1392, 0.0017, 0.0298, -0.1392, 0.0017, 0.032, -0.1392, 0.0039, 0.0298, -0.1392, 0.0039, 0.032, -0.1706, -0.0, 0.0379, -0.1706, -0.0, 0.0407, -0.1684, -0.0, 0.0379, -0.1684, -0.0, 0.0407, -0.1706, 0.0017, 0.0379, -0.1706, 0.0039, 0.0379, -0.1684, 0.0017, 0.0379, -0.1684, 0.0039, 0.0379,
-0.1662, -0.0, 0.0379, -0.1662, -0.0, 0.0407, -0.1641, -0.0, 0.0379, -0.1641, -0.0, 0.0407, -0.1662, 0.0017, 0.0379, -0.1662, 0.0039, 0.0379, -0.1641, 0.0017, 0.0379, -0.1641, 0.0039, 0.0379, -0.1706, -0.0, 0.0429, -0.1706, -0.0, 0.045, -0.1684, -0.0, 0.0429, -0.1684, -0.0, 0.045, -0.1706, -0.0, 0.0472, -0.1706, -0.0, 0.0487, -0.1684, -0.0, 0.0472, -0.1684, -0.0, 0.0487, -0.1706, 0.0017, 0.0487, -0.1706, 0.0039, 0.0487, -0.1684, 0.0017, 0.0487, -0.1684, 0.0039, 0.0487, -0.1662, -0.0, 0.0429, -0.1662, -0.0, 0.045, -0.1641, -0.0, 0.0429, -0.1641, -0.0, 0.045, -0.1662, -0.0, 0.0472, -0.1662, -0.0, 0.0487, -0.1641, -0.0, 0.0472, -0.1641, -0.0, 0.0487, -0.1662, 0.0017, 0.0487, -0.1662, 0.0039, 0.0487,
-0.1641, 0.0017, 0.0487, -0.1641, 0.0039, 0.0487, -0.1619, -0.0, 0.0379, -0.1619, -0.0, 0.0407, -0.1588, -0.0, 0.0379, -0.1588, -0.0, 0.0407, -0.1619, 0.0017, 0.0379, -0.1619, 0.0039, 0.0379, -0.1588, 0.0017, 0.0379, -0.1588, 0.0017, 0.0407, -0.1588, 0.0039, 0.0379, -0.1588, 0.0039, 0.0407, -0.1619, -0.0, 0.0429, -0.1619, -0.0, 0.045, -0.1588, -0.0, 0.0429, -0.1588, -0.0, 0.045, -0.1619, -0.0, 0.0472, -0.1619, -0.0, 0.0487, -0.1588, -0.0, 0.0472, -0.1588, -0.0, 0.0487, -0.1588, 0.0017, 0.0429, -0.1588, 0.0017, 0.045, -0.1588, 0.0039, 0.0429, -0.1588, 0.0039, 0.045, -0.1619, 0.0017, 0.0487, -0.1619, 0.0039, 0.0487, -0.1588, 0.0017, 0.0472, -0.1588, 0.0017, 0.0487, -0.1588, 0.0039, 0.0472, -0.1588, 0.0039, 0.0487,
-0.151, -0.0, 0.0342, -0.151, -0.0, 0.0363, -0.151, -0.0, 0.0385, -0.151, -0.0, 0.0407, -0.151, 0.0017, 0.0342, -0.151, 0.0017, 0.0363, -0.151, 0.0039, 0.0342, -0.151, 0.0039, 0.0363, -0.151, 0.0017, 0.0385, -0.151, 0.0017, 0.0407, -0.151, 0.0039, 0.0385, -0.151, 0.0039, 0.0407, -0.1489, -0.0, 0.0342, -0.1489, -0.0, 0.0363, -0.1467, -0.0, 0.0342, -0.1467, -0.0, 0.0363, -0.1489, -0.0, 0.0385, -0.1489, -0.0, 0.0407, -0.1467, -0.0, 0.0385, -0.1467, -0.0, 0.0407, -0.151, -0.0, 0.0429, -0.151, -0.0, 0.045, -0.151, -0.0, 0.0472, -0.151, -0.0, 0.0487, -0.151, 0.0017, 0.0429, -0.151, 0.0017, 0.045, -0.151, 0.0039, 0.0429, -0.151, 0.0039, 0.045, -0.151, 0.0017, 0.0472, -0.151, 0.0017, 0.0487,
-0.151, 0.0039, 0.0472, -0.151, 0.0039, 0.0487, -0.1489, -0.0, 0.0429, -0.1489, -0.0, 0.045, -0.1467, -0.0, 0.0429, -0.1467, -0.0, 0.045, -0.1489, -0.0, 0.0472, -0.1489, -0.0, 0.0487, -0.1467, -0.0, 0.0472, -0.1467, -0.0, 0.0487, -0.1489, 0.0017, 0.0487, -0.1489, 0.0039, 0.0487, -0.1467, 0.0017, 0.0487, -0.1467, 0.0039, 0.0487, -0.1445, -0.0, 0.0342, -0.1445, -0.0, 0.0363, -0.1424, -0.0, 0.0342, -0.1424, -0.0, 0.0363, -0.1445, -0.0, 0.0385, -0.1445, -0.0, 0.0407, -0.1424, -0.0, 0.0385, -0.1424, -0.0, 0.0407, -0.1392, -0.0, 0.0342, -0.1392, -0.0, 0.0363, -0.1392, -0.0, 0.0385, -0.1392, -0.0, 0.0407, -0.1392, 0.0017, 0.0342, -0.1392, 0.0017, 0.0363, -0.1392, 0.0039, 0.0342, -0.1392, 0.0039, 0.0363,
-0.1392, 0.0017, 0.0385, -0.1392, 0.0017, 0.0407, -0.1392, 0.0039, 0.0385, -0.1392, 0.0039, 0.0407, -0.1445, -0.0, 0.0429, -0.1445, -0.0, 0.045, -0.1424, -0.0, 0.0429, -0.1424, -0.0, 0.045, -0.1445, -0.0, 0.0472, -0.1445, -0.0, 0.0487, -0.1424, -0.0, 0.0472, -0.1424, -0.0, 0.0487, -0.1445, 0.0017, 0.0487, -0.1445, 0.0039, 0.0487, -0.1424, 0.0017, 0.0487, -0.1424, 0.0039, 0.0487, -0.1392, -0.0, 0.0429, -0.1392, -0.0, 0.045, -0.1392, -0.0, 0.0472, -0.1392, -0.0, 0.0487, -0.1392, 0.0017, 0.0429, -0.1392, 0.0017, 0.045, -0.1392, 0.0039, 0.0429, -0.1392, 0.0039, 0.045, -0.1392, 0.0017, 0.0472, -0.1392, 0.0017, 0.0487, -0.1392, 0.0039, 0.0472, -0.1392, 0.0039, 0.0487, -0.0775, -0.0, -0.0005, -0.0775, -0.0, 0.0016,
-0.0775, -0.0, 0.0038, -0.0775, -0.0, 0.006, -0.0775, 0.0017, -0.0005, -0.0775, 0.0017, 0.0016, -0.0775, 0.0039, -0.0005, -0.0775, 0.0039, 0.0016, -0.0775, 0.0017, 0.0038, -0.0775, 0.0017, 0.006, -0.0775, 0.0039, 0.0038, -0.0775, 0.0039, 0.006, -0.0775, -0.0, 0.0081, -0.0775, -0.0, 0.0103, -0.0775, -0.0, 0.0125, -0.0775, -0.0, 0.0146, -0.0775, 0.0017, 0.0081, -0.0775, 0.0017, 0.0103, -0.0775, 0.0039, 0.0081, -0.0775, 0.0039, 0.0103, -0.0775, 0.0017, 0.0125, -0.0775, 0.0017, 0.0146, -0.0775, 0.0039, 0.0125, -0.0775, 0.0039, 0.0146, -0.0751, -0.0, -0.0005, -0.0751, -0.0, 0.0016, -0.0729, -0.0, -0.0005, -0.0729, -0.0, 0.0016, -0.0751, -0.0, 0.0038, -0.0751, -0.0, 0.006, -0.0729, -0.0, 0.0038, -0.0729, -0.0, 0.006,
-0.0707, -0.0, -0.0005, -0.0707, -0.0, 0.0016, -0.0686, -0.0, -0.0005, -0.0686, -0.0, 0.0016, -0.0707, -0.0, 0.0038, -0.0707, -0.0, 0.006, -0.0686, -0.0, 0.0038, -0.0686, -0.0, 0.006, -0.0751, -0.0, 0.0081, -0.0751, -0.0, 0.0103, -0.0729, -0.0, 0.0081, -0.0729, -0.0, 0.0103, -0.0751, -0.0, 0.0125, -0.0751, -0.0, 0.0146, -0.0729, -0.0, 0.0125, -0.0729, -0.0, 0.0146, -0.0707, -0.0, 0.0081, -0.0707, -0.0, 0.0103, -0.0686, -0.0, 0.0081, -0.0686, -0.0, 0.0103, -0.0707, -0.0, 0.0125, -0.0707, -0.0, 0.0146, -0.0686, -0.0, 0.0125, -0.0686, -0.0, 0.0146, -0.0775, -0.0, 0.0168, -0.0775, -0.0, 0.019, -0.0775, -0.0, 0.0212, -0.0775, -0.0, 0.0233, -0.0775, 0.0017, 0.0168, -0.0775, 0.0017, 0.019,
-0.0775, 0.0039, 0.0168, -0.0775, 0.0039, 0.019, -0.0775, 0.0017, 0.0212, -0.0775, 0.0017, 0.0233, -0.0775, 0.0039, 0.0212, -0.0775, 0.0039, 0.0233, -0.0775, -0.0, 0.0255, -0.0775, -0.0, 0.0277, -0.0775, -0.0, 0.0298, -0.0775, -0.0, 0.032, -0.0775, 0.0017, 0.0255, -0.0775, 0.0017, 0.0277, -0.0775, 0.0039, 0.0255, -0.0775, 0.0039, 0.0277, -0.0775, 0.0017, 0.0298, -0.0775, 0.0017, 0.032, -0.0775, 0.0039, 0.0298, -0.0775, 0.0039, 0.032, -0.0751, -0.0, 0.0168, -0.0751, -0.0, 0.019, -0.0729, -0.0, 0.0168, -0.0729, -0.0, 0.019, -0.0751, -0.0, 0.0212, -0.0751, -0.0, 0.0233, -0.0729, -0.0, 0.0212, -0.0729, -0.0, 0.0233, -0.0707, -0.0, 0.0168, -0.0707, -0.0, 0.019, -0.0686, -0.0, 0.0168, -0.0686, -0.0, 0.019,
-0.0707, -0.0, 0.0212, -0.0707, -0.0, 0.0233, -0.0686, -0.0, 0.0212, -0.0686, -0.0, 0.0233, -0.0751, -0.0, 0.0255, -0.0751, -0.0, 0.0277, -0.0729, -0.0, 0.0255, -0.0729, -0.0, 0.0277, -0.0751, -0.0, 0.0298, -0.0751, -0.0, 0.032, -0.0729, -0.0, 0.0298, -0.0729, -0.0, 0.032, -0.0707, -0.0, 0.0255, -0.0707, -0.0, 0.0277, -0.0686, -0.0, 0.0255, -0.0686, -0.0, 0.0277, -0.0707, -0.0, 0.0298, -0.0707, -0.0, 0.032, -0.0686, -0.0, 0.0298, -0.0686, -0.0, 0.032, -0.0775, -0.0, 0.0342, -0.0775, -0.0, 0.0363, -0.0775, -0.0, 0.0385, -0.0775, -0.0, 0.0407, -0.0775, 0.0017, 0.0342, -0.0775, 0.0017, 0.0363, -0.0775, 0.0039, 0.0342, -0.0775, 0.0039, 0.0363, -0.0775, 0.0017, 0.0385, -0.0775, 0.0017, 0.0407,
-0.0775, 0.0039, 0.0385, -0.0775, 0.0039, 0.0407, -0.0775, -0.0, 0.0429, -0.0775, -0.0, 0.045, -0.0775, -0.0, 0.0472, -0.0775, -0.0, 0.0487, -0.0775, 0.0017, 0.0429, -0.0775, 0.0017, 0.045, -0.0775, 0.0039, 0.0429, -0.0775, 0.0039, 0.045, -0.0775, 0.0017, 0.0472, -0.0775, 0.0017, 0.0487, -0.0775, 0.0039, 0.0472, -0.0775, 0.0039, 0.0487, -0.0751, -0.0, 0.0342, -0.0751, -0.0, 0.0363, -0.0729, -0.0, 0.0342, -0.0729, -0.0, 0.0363, -0.0751, -0.0, 0.0385, -0.0751, -0.0, 0.0407, -0.0729, -0.0, 0.0385, -0.0729, -0.0, 0.0407, -0.0707, -0.0, 0.0342, -0.0707, -0.0, 0.0363, -0.0686, -0.0, 0.0342, -0.0686, -0.0, 0.0363, -0.0707, -0.0, 0.0385, -0.0707, -0.0, 0.0407, -0.0686, -0.0, 0.0385, -0.0686, -0.0, 0.0407,
-0.0751, -0.0, 0.0429, -0.0751, -0.0, 0.045, -0.0729, -0.0, 0.0429, -0.0729, -0.0, 0.045, -0.0751, -0.0, 0.0472, -0.0751, -0.0, 0.0487, -0.0729, -0.0, 0.0472, -0.0729, -0.0, 0.0487, -0.0751, 0.0017, 0.0487, -0.0751, 0.0039, 0.0487, -0.0729, 0.0017, 0.0487, -0.0729, 0.0039, 0.0487, -0.0707, -0.0, 0.0429, -0.0707, -0.0, 0.045, -0.0686, -0.0, 0.0429, -0.0686, -0.0, 0.045, -0.0707, -0.0, 0.0472, -0.0707, -0.0, 0.0487, -0.0686, -0.0, 0.0472, -0.0686, -0.0, 0.0487, -0.0707, 0.0017, 0.0487, -0.0707, 0.0039, 0.0487, -0.0686, 0.0017, 0.0487, -0.0686, 0.0039, 0.0487, -0.0667, -0.0, -0.0005, -0.0667, -0.0, 0.0016, -0.0667, -0.0, 0.0038, -0.0667, -0.0, 0.006, -0.0667, 0.0017, -0.0005, -0.0667, 0.0017, 0.0016,
-0.0667, 0.0039, -0.0005, -0.0667, 0.0039, 0.0016, -0.0667, 0.0017, 0.0038, -0.0667, 0.0017, 0.006, -0.0667, 0.0039, 0.0038, -0.0667, 0.0039, 0.006, -0.0667, -0.0, 0.0081, -0.0667, -0.0, 0.0103, -0.0667, -0.0, 0.0125, -0.0667, -0.0, 0.0146, -0.0667, 0.0017, 0.0081, -0.0667, 0.0017, 0.0103, -0.0667, 0.0039, 0.0081, -0.0667, 0.0039, 0.0103, -0.0667, 0.0017, 0.0125, -0.0667, 0.0017, 0.0146, -0.0667, 0.0039, 0.0125, -0.0667, 0.0039, 0.0146, -0.0667, -0.0, 0.0168, -0.0667, -0.0, 0.019, -0.0667, -0.0, 0.0212, -0.0667, -0.0, 0.0233, -0.0667, 0.0017, 0.0168, -0.0667, 0.0017, 0.019, -0.0667, 0.0039, 0.0168, -0.0667, 0.0039, 0.019, -0.0667, 0.0017, 0.0212, -0.0667, 0.0017, 0.0233, -0.0667, 0.0039, 0.0212, -0.0667, 0.0039, 0.0233,
-0.0667, -0.0, 0.0255, -0.0667, -0.0, 0.0277, -0.0667, -0.0, 0.0298, -0.0667, -0.0, 0.032, -0.0667, 0.0017, 0.0255, -0.0667, 0.0017, 0.0277, -0.0667, 0.0039, 0.0255, -0.0667, 0.0039, 0.0277, -0.0667, 0.0017, 0.0298, -0.0667, 0.0017, 0.032, -0.0667, 0.0039, 0.0298, -0.0667, 0.0039, 0.032, -0.05, -0.0, -0.0005, -0.05, -0.0, 0.0016, -0.0469, -0.0, -0.0005, -0.0469, -0.0, 0.0016, -0.05, -0.0, 0.0038, -0.05, -0.0, 0.006, -0.0469, -0.0, 0.0038, -0.0469, -0.0, 0.006, -0.05, 0.0017, -0.0005, -0.05, 0.0017, 0.0016, -0.05, 0.0039, -0.0005, -0.05, 0.0039, 0.0016, -0.05, 0.0017, 0.0038, -0.05, 0.0017, 0.006, -0.05, 0.0039, 0.0038, -0.05, 0.0039, 0.006, -0.0447, -0.0, -0.0005, -0.0447, -0.0, 0.0016,
-0.0425, -0.0, -0.0005, -0.0425, -0.0, 0.0016, -0.0447, -0.0, 0.0038, -0.0447, -0.0, 0.006, -0.0425, -0.0, 0.0038, -0.0425, -0.0, 0.006, -0.05, -0.0, 0.0081, -0.05, -0.0, 0.0103, -0.0469, -0.0, 0.0081, -0.0469, -0.0, 0.0103, -0.05, -0.0, 0.0125, -0.05, -0.0, 0.0146, -0.0469, -0.0, 0.0125, -0.0469, -0.0, 0.0146, -0.05, 0.0017, 0.0081, -0.05, 0.0017, 0.0103, -0.05, 0.0039, 0.0081, -0.05, 0.0039, 0.0103, -0.05, 0.0017, 0.0125, -0.05, 0.0017, 0.0146, -0.05, 0.0039, 0.0125, -0.05, 0.0039, 0.0146, -0.0447, -0.0, 0.0081, -0.0447, -0.0, 0.0103, -0.0425, -0.0, 0.0081, -0.0425, -0.0, 0.0103, -0.0447, -0.0, 0.0125, -0.0447, -0.0, 0.0146, -0.0425, -0.0, 0.0125, -0.0425, -0.0, 0.0146,
-0.0404, -0.0, -0.0005, -0.0404, -0.0, 0.0016, -0.0382, -0.0, -0.0013, -0.0382, -0.0, 0.0016, -0.0404, -0.0, 0.0038, -0.0404, -0.0, 0.006, -0.0382, -0.0, 0.0038, -0.0382, -0.0, 0.006, -0.0382, 0.0017, -0.0013, -0.0382, 0.0039, -0.0013, -0.036, -0.0, -0.0013, -0.036, -0.0, 0.0016, -0.0338, -0.0, -0.0013, -0.0338, -0.0, 0.0016, -0.036, -0.0, 0.0038, -0.036, -0.0, 0.006, -0.0338, -0.0, 0.0038, -0.0338, -0.0, 0.006, -0.036, 0.0017, -0.0013, -0.036, 0.0039, -0.0013, -0.0338, 0.0017, -0.0013, -0.0338, 0.0039, -0.0013, -0.0404, -0.0, 0.0081, -0.0404, -0.0, 0.0103, -0.0382, -0.0, 0.0081, -0.0382, -0.0, 0.0105, -0.0404, -0.0, 0.0125, -0.0404, -0.0, 0.0146, -0.0382, -0.0, 0.0125, -0.0382, -0.0, 0.0146,
-0.0382, 0.0017, 0.0105, -0.0382, 0.0039, 0.0105, -0.0382, 0.0017, 0.0125, -0.0382, 0.0017, 0.0146, -0.0382, 0.0039, 0.0125, -0.0382, 0.0039, 0.0146, -0.036, -0.0, 0.0081, -0.036, -0.0, 0.0105, -0.0338, -0.0, 0.0081, -0.0338, -0.0, 0.0105, -0.036, 0.0017, 0.0105, -0.036, 0.0039, 0.0105, -0.0338, 0.0017, 0.0105, -0.0338, 0.0039, 0.0105, -0.05, -0.0, 0.0168, -0.05, -0.0, 0.019, -0.0469, -0.0, 0.0168, -0.0469, -0.0, 0.019, -0.05, -0.0, 0.0212, -0.05, -0.0, 0.0233, -0.0469, -0.0, 0.0212, -0.0469, -0.0, 0.0233, -0.05, 0.0017, 0.0168, -0.05, 0.0017, 0.019, -0.05, 0.0039, 0.0168, -0.05, 0.0039, 0.019, -0.05, 0.0017, 0.0212, -0.05, 0.0017, 0.0233, -0.05, 0.0039, 0.0212, -0.05, 0.0039, 0.0233,
-0.0447, -0.0, 0.0168, -0.0447, -0.0, 0.019, -0.0425, -0.0, 0.0168, -0.0425, -0.0, 0.019, -0.0447, -0.0, 0.0212, -0.0447, -0.0, 0.0233, -0.0425, -0.0, 0.0212, -0.0425, -0.0, 0.0233, -0.05, -0.0, 0.0255, -0.05, -0.0, 0.0277, -0.0469, -0.0, 0.0255, -0.0469, -0.0, 0.0277, -0.05, -0.0, 0.0298, -0.05, -0.0, 0.032, -0.0469, -0.0, 0.0298, -0.0469, -0.0, 0.032, -0.05, 0.0017, 0.0255, -0.05, 0.0017, 0.0277, -0.05, 0.0039, 0.0255, -0.05, 0.0039, 0.0277, -0.05, 0.0017, 0.0298, -0.05, 0.0017, 0.032, -0.05, 0.0039, 0.0298, -0.05, 0.0039, 0.032, -0.0447, -0.0, 0.0255, -0.0447, -0.0, 0.0277, -0.0425, -0.0, 0.0255, -0.0425, -0.0, 0.0277, -0.0447, -0.0, 0.0298, -0.0447, -0.0, 0.032,
-0.0425, -0.0, 0.0298, -0.0425, -0.0, 0.032, -0.0404, -0.0, 0.0168, -0.0404, -0.0, 0.019, -0.0382, -0.0, 0.0168, -0.0382, -0.0, 0.019, -0.0404, -0.0, 0.0212, -0.0404, -0.0, 0.0233, -0.0382, -0.0, 0.0212, -0.0382, -0.0, 0.0233, -0.0382, 0.0017, 0.0168, -0.0382, 0.0017, 0.019, -0.0382, 0.0039, 0.0168, -0.0382, 0.0039, 0.019, -0.0382, 0.0017, 0.0212, -0.0382, 0.0017, 0.0233, -0.0382, 0.0039, 0.0212, -0.0382, 0.0039, 0.0233, -0.0404, -0.0, 0.0255, -0.0404, -0.0, 0.0277, -0.0382, -0.0, 0.0255, -0.0382, -0.0, 0.0277, -0.0404, -0.0, 0.0298, -0.0404, -0.0, 0.032, -0.0382, -0.0, 0.0298, -0.0382, -0.0, 0.032, -0.0382, 0.0017, 0.0255, -0.0382, 0.0017, 0.0277, -0.0382, 0.0039, 0.0255, -0.0382, 0.0039, 0.0277,
-0.0382, 0.0017, 0.0298, -0.0382, 0.0017, 0.032, -0.0382, 0.0039, 0.0298, -0.0382, 0.0039, 0.032, -0.0667, -0.0, 0.0342, -0.0667, -0.0, 0.0363, -0.0667, -0.0, 0.0385, -0.0667, -0.0, 0.0407, -0.0667, 0.0017, 0.0342, -0.0667, 0.0017, 0.0363, -0.0667, 0.0039, 0.0342, -0.0667, 0.0039, 0.0363, -0.0667, 0.0017, 0.0385, -0.0667, 0.0017, 0.0407, -0.0667, 0.0039, 0.0385, -0.0667, 0.0039, 0.0407, -0.0667, -0.0, 0.0429, -0.0667, -0.0, 0.045, -0.0667, -0.0, 0.0472, -0.0667, -0.0, 0.0487, -0.0667, 0.0017, 0.0429, -0.0667, 0.0017, 0.045, -0.0667, 0.0039, 0.0429, -0.0667, 0.0039, 0.045, -0.0667, 0.0017, 0.0472, -0.0667, 0.0017, 0.0487, -0.0667, 0.0039, 0.0472, -0.0667, 0.0039, 0.0487, -0.05, -0.0, 0.0342, -0.05, -0.0, 0.0363,
-0.0469, -0.0, 0.0342, -0.0469, -0.0, 0.0363, -0.05, -0.0, 0.0385, -0.05, -0.0, 0.0407, -0.0469, -0.0, 0.0385, -0.0469, -0.0, 0.0407, -0.05, 0.0017, 0.0342, -0.05, 0.0017, 0.0363, -0.05, 0.0039, 0.0342, -0.05, 0.0039, 0.0363, -0.05, 0.0017, 0.0385, -0.05, 0.0017, 0.0407, -0.05, 0.0039, 0.0385, -0.05, 0.0039, 0.0407, -0.0447, -0.0, 0.0342, -0.0447, -0.0, 0.0363, -0.0425, -0.0, 0.0342, -0.0425, -0.0, 0.0363, -0.0447, -0.0, 0.0385, -0.0447, -0.0, 0.0407, -0.0425, -0.0, 0.0385, -0.0425, -0.0, 0.0407, -0.05, -0.0, 0.0429, -0.05, -0.0, 0.045, -0.0469, -0.0, 0.0429, -0.0469, -0.0, 0.045, -0.05, -0.0, 0.0472, -0.05, -0.0, 0.0487, -0.0469, -0.0, 0.0472, -0.0469, -0.0, 0.0487,
-0.05, 0.0017, 0.0429, -0.05, 0.0017, 0.045, -0.05, 0.0039, 0.0429, -0.05, 0.0039, 0.045, -0.05, 0.0017, 0.0472, -0.05, 0.0017, 0.0487, -0.05, 0.0039, 0.0472, -0.05, 0.0039, 0.0487, -0.0469, 0.0017, 0.0487, -0.0469, 0.0039, 0.0487, -0.0447, -0.0, 0.0429, -0.0447, -0.0, 0.045, -0.0425, -0.0, 0.0429, -0.0425, -0.0, 0.045, -0.0447, -0.0, 0.0472, -0.0447, -0.0, 0.0487, -0.0425, -0.0, 0.0472, -0.0425, -0.0, 0.0487, -0.0447, 0.0017, 0.0487, -0.0447, 0.0039, 0.0487, -0.0425, 0.0017, 0.0487, -0.0425, 0.0039, 0.0487, -0.0404, -0.0, 0.0342, -0.0404, -0.0, 0.0363, -0.0382, -0.0, 0.0342, -0.0382, -0.0, 0.0363, -0.0404, -0.0, 0.0385, -0.0404, -0.0, 0.0407, -0.0382, -0.0, 0.0379, -0.0382, -0.0, 0.0407,
-0.0382, 0.0017, 0.0342, -0.0382, 0.0017, 0.0363, -0.0382, 0.0039, 0.0342, -0.0382, 0.0039, 0.0363, -0.0382, 0.0017, 0.0379, -0.0382, 0.0039, 0.0379, -0.036, -0.0, 0.0379, -0.036, -0.0, 0.0407, -0.0338, -0.0, 0.0379, -0.0338, -0.0, 0.0407, -0.036, 0.0017, 0.0379, -0.036, 0.0039, 0.0379, -0.0338, 0.0017, 0.0379, -0.0338, 0.0039, 0.0379, -0.0404, -0.0, 0.0429, -0.0404, -0.0, 0.045, -0.0382, -0.0, 0.0429, -0.0382, -0.0, 0.045, -0.0404, -0.0, 0.0472, -0.0404, -0.0, 0.0487, -0.0382, -0.0, 0.0472, -0.0382, -0.0, 0.0487, -0.0404, 0.0017, 0.0487, -0.0404, 0.0039, 0.0487, -0.0382, 0.0017, 0.0487, -0.0382, 0.0039, 0.0487, -0.036, -0.0, 0.0429, -0.036, -0.0, 0.045, -0.0338, -0.0, 0.0429, -0.0338, -0.0, 0.045,
-0.036, -0.0, 0.0472, -0.036, -0.0, 0.0487, -0.0338, -0.0, 0.0472, -0.0338, -0.0, 0.0487, -0.036, 0.0017, 0.0487, -0.036, 0.0039, 0.0487, -0.0338, 0.0017, 0.0487, -0.0338, 0.0039, 0.0487, -0.0317, -0.0, -0.0013, -0.0317, -0.0, 0.0016, -0.0295, -0.0, -0.0013, -0.0295, -0.0, 0.0016, -0.0317, -0.0, 0.0038, -0.0317, -0.0, 0.006, -0.0295, -0.0, 0.0038, -0.0295, -0.0, 0.006, -0.0317, 0.0017, -0.0013, -0.0317, 0.0039, -0.0013, -0.0295, 0.0017, -0.0013, -0.0295, 0.0039, -0.0013, -0.0273, -0.0, -0.0013, -0.0273, -0.0, 0.0016, -0.0252, -0.0, -0.0013, -0.0252, -0.0, 0.0016, -0.0273, -0.0, 0.0038, -0.0273, -0.0, 0.006, -0.0252, -0.0, 0.0038, -0.0252, -0.0, 0.006, -0.0273, 0.0017, -0.0013, -0.0273, 0.0039, -0.0013,
-0.0252, 0.0017, -0.0013, -0.0252, 0.0039, -0.0013, -0.0317, -0.0, 0.0081, -0.0317, -0.0, 0.0105, -0.0295, -0.0, 0.0081, -0.0295, -0.0, 0.0105, -0.0317, 0.0017, 0.0105, -0.0317, 0.0039, 0.0105, -0.0295, 0.0017, 0.0105, -0.0295, 0.0039, 0.0105, -0.0273, -0.0, 0.0081, -0.0273, -0.0, 0.0105, -0.0252, -0.0, 0.0081, -0.0252, -0.0, 0.0105, -0.0273, 0.0017, 0.0105, -0.0273, 0.0039, 0.0105, -0.0252, 0.0017, 0.0105, -0.0252, 0.0039, 0.0105, -0.023, -0.0, -0.0013, -0.023, -0.0, 0.0016, -0.0208, -0.0, -0.0013, -0.0208, -0.0, 0.0016, -0.023, -0.0, 0.0038, -0.023, -0.0, 0.006, -0.0208, -0.0, 0.0038, -0.0208, -0.0, 0.006, -0.023, 0.0017, -0.0013, -0.023, 0.0039, -0.0013, -0.0208, 0.0017, -0.0013, -0.0208, 0.0039, -0.0013,
-0.0187, -0.0, -0.0013, -0.0187, -0.0, 0.0016, -0.0165, -0.0, -0.0013, -0.0165, -0.0, 0.0016, -0.0187, -0.0, 0.0038, -0.0187, -0.0, 0.006, -0.0165, -0.0, 0.0038, -0.0165, -0.0, 0.006, -0.0187, 0.0017, -0.0013, -0.0187, 0.0039, -0.0013, -0.0165, 0.0017, -0.0013, -0.0165, 0.0039, -0.0013, -0.023, -0.0, 0.0081, -0.023, -0.0, 0.0105, -0.0208, -0.0, 0.0081, -0.0208, -0.0, 0.0105, -0.023, 0.0017, 0.0105, -0.023, 0.0039, 0.0105, -0.0208, 0.0017, 0.0105, -0.0208, 0.0039, 0.0105, -0.0187, -0.0, 0.0081, -0.0187, -0.0, 0.0105, -0.0165, -0.0, 0.0081, -0.0165, -0.0, 0.0105, -0.0187, 0.0017, 0.0105, -0.0187, 0.0039, 0.0105, -0.0165, 0.0017, 0.0105, -0.0165, 0.0039, 0.0105, -0.0143, -0.0, -0.0013, -0.0143, -0.0, 0.0016,
-0.0121, -0.0, -0.0013, -0.0121, -0.0, 0.0016, -0.0143, -0.0, 0.0038, -0.0143, -0.0, 0.006, -0.0121, -0.0, 0.0038, -0.0121, -0.0, 0.006, -0.0143, 0.0017, -0.0013, -0.0143, 0.0039, -0.0013, -0.0121, 0.0017, -0.0013, -0.0121, 0.0039, -0.0013, -0.01, -0.0, -0.0013, -0.01, -0.0, 0.0016, -0.0078, -0.0, -0.0013, -0.0078, -0.0, 0.0016, -0.01, -0.0, 0.0038, -0.01, -0.0, 0.006, -0.0078, -0.0, 0.0038, -0.0078, -0.0, 0.006, -0.01, 0.0017, -0.0013, -0.01, 0.0039, -0.0013, -0.0078, 0.0017, -0.0013, -0.0078, 0.0039, -0.0013, -0.0143, -0.0, 0.0081, -0.0143, -0.0, 0.0105, -0.0121, -0.0, 0.0081, -0.0121, -0.0, 0.0105, -0.0143, 0.0017, 0.0105, -0.0143, 0.0039, 0.0105, -0.0121, 0.0017, 0.0105, -0.0121, 0.0039, 0.0105,
-0.01, -0.0, 0.0081, -0.01, -0.0, 0.0105, -0.0078, -0.0, 0.0081, -0.0078, -0.0, 0.0105, -0.01, 0.0017, 0.0105, -0.01, 0.0039, 0.0105, -0.0078, 0.0017, 0.0105, -0.0078, 0.0039, 0.0105, -0.0056, -0.0, -0.0013, -0.0056, -0.0, 0.0016, -0.0035, -0.0, -0.0013, -0.0035, -0.0, 0.0016, -0.0056, -0.0, 0.0038, -0.0056, -0.0, 0.006, -0.0035, -0.0, 0.0038, -0.0035, -0.0, 0.006, -0.0056, 0.0017, -0.0013, -0.0056, 0.0039, -0.0013, -0.0035, 0.0017, -0.0013, -0.0035, 0.0039, -0.0013, -0.0013, -0.0, -0.0013, -0.0013, -0.0, 0.0016, 0.0009, -0.0, -0.0013, 0.0009, -0.0, 0.0016, -0.0013, -0.0, 0.0038, -0.0013, -0.0, 0.006, 0.0009, -0.0, 0.0038, 0.0009, -0.0, 0.006, -0.0013, 0.0017, -0.0013, -0.0013, 0.0039, -0.0013,
0.0009, 0.0017, -0.0013, 0.0009, 0.0039, -0.0013, -0.0056, -0.0, 0.0081, -0.0056, -0.0, 0.0105, -0.0035, -0.0, 0.0081, -0.0035, -0.0, 0.0105, -0.0056, 0.0017, 0.0105, -0.0056, 0.0039, 0.0105, -0.0035, 0.0017, 0.0105, -0.0035, 0.0039, 0.0105, -0.0013, -0.0, 0.0081, -0.0013, -0.0, 0.0105, 0.0009, -0.0, 0.0081, 0.0009, -0.0, 0.0105, -0.0013, 0.0017, 0.0105, -0.0013, 0.0039, 0.0105, 0.0009, 0.0017, 0.0105, 0.0009, 0.0039, 0.0105, -0.0317, -0.0, 0.0379, -0.0317, -0.0, 0.0407, -0.0295, -0.0, 0.0379, -0.0295, -0.0, 0.0407, -0.0317, 0.0017, 0.0379, -0.0317, 0.0039, 0.0379, -0.0295, 0.0017, 0.0379, -0.0295, 0.0039, 0.0379, -0.0273, -0.0, 0.0379, -0.0273, -0.0, 0.0407, -0.0252, -0.0, 0.0379, -0.0252, -0.0, 0.0407,
-0.0273, 0.0017, 0.0379, -0.0273, 0.0039, 0.0379, -0.0252, 0.0017, 0.0379, -0.0252, 0.0039, 0.0379, -0.0317, -0.0, 0.0429, -0.0317, -0.0, 0.045, -0.0295, -0.0, 0.0429, -0.0295, -0.0, 0.045, -0.0317, -0.0, 0.0472, -0.0317, -0.0, 0.0487, -0.0295, -0.0, 0.0472, -0.0295, -0.0, 0.0487, -0.0317, 0.0017, 0.0487, -0.0317, 0.0039, 0.0487, -0.0295, 0.0017, 0.0487, -0.0295, 0.0039, 0.0487, -0.0273, -0.0, 0.0429, -0.0273, -0.0, 0.045, -0.0252, -0.0, 0.0429, -0.0252, -0.0, 0.045, -0.0273, -0.0, 0.0472, -0.0273, -0.0, 0.0487, -0.0252, -0.0, 0.0472, -0.0252, -0.0, 0.0487, -0.0273, 0.0017, 0.0487, -0.0273, 0.0039, 0.0487, -0.0252, 0.0017, 0.0487, -0.0252, 0.0039, 0.0487, -0.023, -0.0, 0.0379, -0.023, -0.0, 0.0407,
-0.0208, -0.0, 0.0379, -0.0208, -0.0, 0.0407, -0.023, 0.0017, 0.0379, -0.023, 0.0039, 0.0379, -0.0208, 0.0017, 0.0379, -0.0208, 0.0039, 0.0379, -0.0187, -0.0, 0.0379, -0.0187, -0.0, 0.0407, -0.0165, -0.0, 0.0379, -0.0165, -0.0, 0.0407, -0.0187, 0.0017, 0.0379, -0.0187, 0.0039, 0.0379, -0.0165, 0.0017, 0.0379, -0.0165, 0.0039, 0.0379, -0.023, -0.0, 0.0429, -0.023, -0.0, 0.045, -0.0208, -0.0, 0.0429, -0.0208, -0.0, 0.045, -0.023, -0.0, 0.0472, -0.023, -0.0, 0.0487, -0.0208, -0.0, 0.0472, -0.0208, -0.0, 0.0487, -0.023, 0.0017, 0.0487, -0.023, 0.0039, 0.0487, -0.0208, 0.0017, 0.0487, -0.0208, 0.0039, 0.0487, -0.0187, -0.0, 0.0429, -0.0187, -0.0, 0.045, -0.0165, -0.0, 0.0429, -0.0165, -0.0, 0.045,
-0.0187, -0.0, 0.0472, -0.0187, -0.0, 0.0487, -0.0165, -0.0, 0.0472, -0.0165, -0.0, 0.0487, -0.0187, 0.0017, 0.0487, -0.0187, 0.0039, 0.0487, -0.0165, 0.0017, 0.0487, -0.0165, 0.0039, 0.0487, -0.0143, -0.0, 0.0379, -0.0143, -0.0, 0.0407, -0.0121, -0.0, 0.0379, -0.0121, -0.0, 0.0407, -0.0143, 0.0017, 0.0379, -0.0143, 0.0039, 0.0379, -0.0121, 0.0017, 0.0379, -0.0121, 0.0039, 0.0379, -0.01, -0.0, 0.0379, -0.01, -0.0, 0.0407, -0.0078, -0.0, 0.0379, -0.0078, -0.0, 0.0407, -0.01, 0.0017, 0.0379, -0.01, 0.0039, 0.0379, -0.0078, 0.0017, 0.0379, -0.0078, 0.0039, 0.0379, -0.0143, -0.0, 0.0429, -0.0143, -0.0, 0.045, -0.0121, -0.0, 0.0429, -0.0121, -0.0, 0.045, -0.0143, -0.0, 0.0472, -0.0143, -0.0, 0.0487,
-0.0121, -0.0, 0.0472, -0.0121, -0.0, 0.0487, -0.0143, 0.0017, 0.0487, -0.0143, 0.0039, 0.0487, -0.0121, 0.0017, 0.0487, -0.0121, 0.0039, 0.0487, -0.01, -0.0, 0.0429, -0.01, -0.0, 0.045, -0.0078, -0.0, 0.0429, -0.0078, -0.0, 0.045, -0.01, -0.0, 0.0472, -0.01, -0.0, 0.0487, -0.0078, -0.0, 0.0472, -0.0078, -0.0, 0.0487, -0.01, 0.0017, 0.0487, -0.01, 0.0039, 0.0487, -0.0078, 0.0017, 0.0487, -0.0078, 0.0039, 0.0487, -0.0056, -0.0, 0.0379, -0.0056, -0.0, 0.0407, -0.0035, -0.0, 0.0379, -0.0035, -0.0, 0.0407, -0.0056, 0.0017, 0.0379, -0.0056, 0.0039, 0.0379, -0.0035, 0.0017, 0.0379, -0.0035, 0.0039, 0.0379, -0.0013, -0.0, 0.0379, -0.0013, -0.0, 0.0407, 0.0008, -0.0, 0.0379, 0.0009, -0.0, 0.0407,
-0.0013, 0.0017, 0.0379, -0.0013, 0.0039, 0.0379, 0.0009, 0.0017, 0.0379, 0.0009, 0.0039, 0.0379, -0.0056, -0.0, 0.0429, -0.0056, -0.0, 0.045, -0.0035, -0.0, 0.0429, -0.0035, -0.0, 0.045, -0.0056, -0.0, 0.0472, -0.0056, -0.0, 0.0487, -0.0035, -0.0, 0.0472, -0.0035, -0.0, 0.0487, -0.0056, 0.0017, 0.0487, -0.0056, 0.0039, 0.0487, -0.0035, 0.0017, 0.0487, -0.0035, 0.0039, 0.0487, -0.0013, -0.0, 0.0429, -0.0013, -0.0, 0.045, 0.0009, -0.0, 0.0429, 0.0009, -0.0, 0.045, -0.0013, -0.0, 0.0472, -0.0013, -0.0, 0.0487, 0.0009, -0.0, 0.0472, 0.0009, -0.0, 0.0487, -0.0013, 0.0017, 0.0487, -0.0013, 0.0039, 0.0487, 0.0009, 0.0017, 0.0487, 0.0009, 0.0039, 0.0487, -0.4928, 0.0061, -0.0364, -0.4928, 0.0083, -0.0364,
-0.4928, 0.01, -0.0364, -0.4868, 0.0061, -0.0457, -0.4868, 0.0083, -0.0457, -0.4849, 0.0061, -0.0482, -0.486, 0.0061, -0.0468, -0.4849, 0.0083, -0.0482, -0.486, 0.0083, -0.0468, -0.4868, 0.01, -0.0456, -0.485, 0.01, -0.0481, -0.486, 0.01, -0.0467, -0.4886, 0.0061, -0.043, -0.4895, 0.0061, -0.0418, -0.4886, 0.0083, -0.043, -0.4895, 0.0083, -0.0418, -0.4911, 0.0061, -0.0393, -0.4922, 0.0061, -0.0375, -0.4911, 0.0083, -0.0393, -0.4922, 0.0083, -0.0375, -0.4904, 0.0061, -0.0403, -0.4904, 0.0083, -0.0403, -0.4886, 0.01, -0.043, -0.4893, 0.01, -0.042, -0.4912, 0.01, -0.0391, -0.4922, 0.01, -0.0375, -0.4905, 0.01, -0.0402, -0.4896, 0.01, -0.0374, -0.4879, 0.0061, -0.0441, -0.4879, 0.0083, -0.0441, -0.4868, 0.0061, -0.0368, -0.4868, 0.0083, -0.0368,
-0.4844, 0.0061, -0.0387, -0.4855, 0.0061, -0.0379, -0.4844, 0.0083, -0.0387, -0.4855, 0.0083, -0.0379, -0.4879, 0.01, -0.0441, -0.4874, 0.01, -0.0418, -0.4852, 0.01, -0.0439, -0.4852, 0.01, -0.0418, -0.4874, 0.01, -0.0396, -0.4869, 0.01, -0.0368, -0.4845, 0.01, -0.0386, -0.4856, 0.01, -0.0378, -0.4951, 0.0061, -0.0321, -0.4951, 0.0083, -0.0321, -0.4935, 0.0061, -0.0353, -0.4945, 0.0061, -0.0332, -0.4935, 0.0083, -0.0353, -0.4945, 0.0083, -0.0332, -0.4961, 0.0061, -0.0297, -0.4961, 0.0083, -0.0297, -0.4938, 0.0061, -0.0306, -0.4938, 0.0083, -0.0306, -0.4951, 0.01, -0.0321, -0.4934, 0.01, -0.0354, -0.4945, 0.01, -0.0332, -0.4961, 0.01, -0.0297, -0.4935, 0.01, -0.0308, -0.4934, 0.0061, -0.0241, -0.4934, 0.0083, -0.0241, -0.4952, 0.0061, -0.0215,
-0.4961, 0.0061, -0.0201, -0.4952, 0.0083, -0.0215, -0.4961, 0.0083, -0.0201, -0.4945, 0.0061, -0.0226, -0.4935, 0.0061, -0.0195, -0.4945, 0.0083, -0.0226, -0.4935, 0.0083, -0.0195, -0.4934, 0.01, -0.024, -0.4953, 0.01, -0.0215, -0.496, 0.01, -0.0203, -0.4945, 0.01, -0.0226, -0.4935, 0.01, -0.0195, -0.4914, 0.0061, -0.0326, -0.4914, 0.0083, -0.0326, -0.4891, 0.0061, -0.0347, -0.4902, 0.0061, -0.0337, -0.4891, 0.0083, -0.0347, -0.4902, 0.0083, -0.0337, -0.4925, 0.0061, -0.0317, -0.4925, 0.0083, -0.0317, -0.4894, 0.0061, -0.0286, -0.4894, 0.0083, -0.0286, -0.4917, 0.01, -0.0353, -0.4914, 0.01, -0.0326, -0.4891, 0.01, -0.0347, -0.4902, 0.01, -0.0337, -0.4925, 0.01, -0.0317, -0.4894, 0.01, -0.0286, -0.4879, 0.0061, -0.0358, -0.4879, 0.0083, -0.0358,
-0.4849, 0.0061, -0.0326, -0.4849, 0.0083, -0.0326, -0.4872, 0.0061, -0.0307, -0.4883, 0.0061, -0.0296, -0.4872, 0.0083, -0.0307, -0.4883, 0.0083, -0.0296, -0.486, 0.0061, -0.0317, -0.486, 0.0083, -0.0317, -0.4879, 0.01, -0.0358, -0.4849, 0.01, -0.0327, -0.4872, 0.01, -0.0307, -0.4883, 0.01, -0.0297, -0.486, 0.01, -0.0318, -0.4852, 0.01, -0.0288, -0.4915, 0.0061, -0.0264, -0.4925, 0.0061, -0.0252, -0.4915, 0.0083, -0.0264, -0.4925, 0.0083, -0.0252, -0.4905, 0.0061, -0.0274, -0.4905, 0.0083, -0.0274, -0.4911, 0.0061, -0.0215, -0.4922, 0.0061, -0.0206, -0.4911, 0.0083, -0.0215, -0.4922, 0.0083, -0.0206, -0.4896, 0.0061, -0.0225, -0.4896, 0.0083, -0.0225, -0.4915, 0.01, -0.0263, -0.4926, 0.01, -0.0251, -0.4905, 0.01, -0.0274, -0.4896, 0.01, -0.0244,
-0.4911, 0.01, -0.0214, -0.4922, 0.01, -0.0206, -0.4894, 0.01, -0.0227, -0.4873, 0.0061, -0.0239, -0.4873, 0.0083, -0.0239, -0.4853, 0.0061, -0.025, -0.4853, 0.0083, -0.025, -0.4884, 0.0061, -0.0233, -0.4884, 0.0083, -0.0233, -0.4874, 0.01, -0.0266, -0.4873, 0.01, -0.0239, -0.4852, 0.01, -0.0266, -0.4851, 0.01, -0.025, -0.4884, 0.01, -0.0233, -0.4975, 0.0061, -0.0175, -0.498, 0.0061, -0.016, -0.4975, 0.0083, -0.0175, -0.498, 0.0083, -0.016, -0.4975, 0.01, -0.0173, -0.498, 0.01, -0.016, -0.4983, 0.0061, -0.0157, -0.497, 0.0061, -0.0167, -0.4983, 0.0083, -0.0157, -0.497, 0.0083, -0.0167, -0.4945, 0.0061, -0.0186, -0.4945, 0.0083, -0.0186, -0.4983, 0.01, -0.0157, -0.497, 0.01, -0.0167, -0.4946, 0.01, -0.0186, -0.4561, 0.0061, -0.0713,
-0.4561, 0.0083, -0.0713, -0.4541, 0.0061, -0.0736, -0.4552, 0.0061, -0.0724, -0.4541, 0.0083, -0.0736, -0.4552, 0.0083, -0.0724, -0.4561, 0.01, -0.0713, -0.4541, 0.01, -0.0736, -0.4552, 0.01, -0.0724, -0.452, 0.0061, -0.0759, -0.452, 0.0083, -0.0759, -0.4499, 0.0061, -0.0781, -0.451, 0.0061, -0.077, -0.4499, 0.0083, -0.0781, -0.451, 0.0083, -0.077, -0.4531, 0.0061, -0.0747, -0.4531, 0.0083, -0.0747, -0.452, 0.01, -0.0759, -0.4499, 0.01, -0.0781, -0.451, 0.01, -0.077, -0.4531, 0.01, -0.0747, -0.4527, 0.01, -0.0722, -0.4505, 0.01, -0.0743, -0.4505, 0.01, -0.0722, -0.4372, 0.0061, -0.089, -0.4372, 0.0083, -0.089, -0.4371, 0.01, -0.0891, -0.4345, 0.0061, -0.0908, -0.4356, 0.0061, -0.0901, -0.4345, 0.0083, -0.0908, -0.4356, 0.0083, -0.0901,
-0.4332, 0.0061, -0.0917, -0.4332, 0.0083, -0.0917, -0.4345, 0.01, -0.0908, -0.4356, 0.01, -0.0901, -0.4334, 0.01, -0.0916, -0.4332, 0.01, -0.0895, -0.4478, 0.0061, -0.0802, -0.4478, 0.0083, -0.0802, -0.4455, 0.0061, -0.0823, -0.4466, 0.0061, -0.0813, -0.4455, 0.0083, -0.0823, -0.4466, 0.0083, -0.0813, -0.4478, 0.01, -0.0802, -0.4455, 0.01, -0.0823, -0.4466, 0.01, -0.0813, -0.4432, 0.0061, -0.0844, -0.4432, 0.0083, -0.0844, -0.4408, 0.0061, -0.0863, -0.4419, 0.0061, -0.0855, -0.4408, 0.0083, -0.0863, -0.4419, 0.0083, -0.0855, -0.4443, 0.0061, -0.0834, -0.4443, 0.0083, -0.0834, -0.4432, 0.01, -0.0844, -0.4408, 0.01, -0.0863, -0.442, 0.01, -0.0855, -0.4443, 0.01, -0.0834, -0.444, 0.01, -0.0808, -0.4418, 0.01, -0.083, -0.4418, 0.01, -0.0808,
-0.4488, 0.0061, -0.0792, -0.4488, 0.0083, -0.0792, -0.4488, 0.01, -0.0792, -0.4483, 0.01, -0.0765, -0.4462, 0.01, -0.0787, -0.4462, 0.01, -0.0765, -0.4483, 0.01, -0.0743, -0.4483, 0.01, -0.0722, -0.4462, 0.01, -0.0743, -0.4462, 0.01, -0.0722, -0.444, 0.01, -0.0787, -0.444, 0.01, -0.0765, -0.4418, 0.01, -0.0787, -0.4418, 0.01, -0.0765, -0.444, 0.01, -0.0743, -0.444, 0.01, -0.0722, -0.4418, 0.01, -0.0743, -0.4418, 0.01, -0.0722, -0.4397, 0.0061, -0.0872, -0.4397, 0.0083, -0.0872, -0.4383, 0.0061, -0.0882, -0.4383, 0.0083, -0.0882, -0.4399, 0.01, -0.087, -0.4397, 0.01, -0.0852, -0.4382, 0.01, -0.0883, -0.4375, 0.01, -0.0852, -0.4397, 0.01, -0.083, -0.4397, 0.01, -0.0808, -0.4375, 0.01, -0.083, -0.4375, 0.01, -0.0808,
-0.4353, 0.01, -0.0874, -0.4353, 0.01, -0.0852, -0.4332, 0.01, -0.0874, -0.4332, 0.01, -0.0852, -0.4353, 0.01, -0.083, -0.4353, 0.01, -0.0808, -0.4332, 0.01, -0.083, -0.4332, 0.01, -0.0808, -0.4397, 0.01, -0.0787, -0.4397, 0.01, -0.0765, -0.4375, 0.01, -0.0787, -0.4375, 0.01, -0.0765, -0.4397, 0.01, -0.0743, -0.4397, 0.01, -0.0722, -0.4375, 0.01, -0.0743, -0.4375, 0.01, -0.0722, -0.4353, 0.01, -0.0787, -0.4353, 0.01, -0.0765, -0.4332, 0.01, -0.0787, -0.4332, 0.01, -0.0765, -0.4353, 0.01, -0.0743, -0.4353, 0.01, -0.0722, -0.4332, 0.01, -0.0743, -0.4332, 0.01, -0.0722, -0.4235, 0.0061, -0.0972, -0.4235, 0.0083, -0.0972, -0.4234, 0.01, -0.0972, -0.4307, 0.0061, -0.0933, -0.4307, 0.0083, -0.0933, -0.4277, 0.0061, -0.0949,
-0.4288, 0.0061, -0.0943, -0.4277, 0.0083, -0.0949, -0.4288, 0.0083, -0.0943, -0.4318, 0.0061, -0.0926, -0.4318, 0.0083, -0.0926, -0.4306, 0.01, -0.0933, -0.4277, 0.01, -0.0949, -0.4289, 0.01, -0.0943, -0.4316, 0.01, -0.0927, -0.431, 0.01, -0.0895, -0.4288, 0.01, -0.0917, -0.4288, 0.01, -0.0895, -0.4266, 0.0061, -0.0955, -0.4266, 0.0083, -0.0955, -0.4245, 0.0061, -0.0966, -0.4245, 0.0083, -0.0966, -0.4268, 0.01, -0.0955, -0.4266, 0.01, -0.0939, -0.4246, 0.01, -0.0966, -0.4245, 0.01, -0.0939, -0.4266, 0.01, -0.0917, -0.4266, 0.01, -0.0895, -0.4245, 0.01, -0.0917, -0.4245, 0.01, -0.0895, -0.4223, 0.0061, -0.0977, -0.4223, 0.0083, -0.0977, -0.4201, 0.0061, -0.0988, -0.4201, 0.0083, -0.0988, -0.4224, 0.01, -0.0977, -0.4205, 0.01, -0.0986,
-0.4179, 0.0061, -0.0997, -0.419, 0.0061, -0.0993, -0.4179, 0.0083, -0.0997, -0.419, 0.0083, -0.0993, -0.4158, 0.0061, -0.1006, -0.4158, 0.0083, -0.1006, -0.4177, 0.01, -0.0998, -0.4186, 0.01, -0.0995, -0.416, 0.01, -0.1005, -0.4158, 0.01, -0.0982, -0.4223, 0.01, -0.096, -0.4223, 0.01, -0.0939, -0.4201, 0.01, -0.096, -0.4201, 0.01, -0.0939, -0.4223, 0.01, -0.0917, -0.4223, 0.01, -0.0895, -0.4201, 0.01, -0.0917, -0.4201, 0.01, -0.0895, -0.418, 0.01, -0.096, -0.418, 0.01, -0.0939, -0.4158, 0.01, -0.096, -0.4158, 0.01, -0.0939, -0.418, 0.01, -0.0917, -0.418, 0.01, -0.0895, -0.4158, 0.01, -0.0917, -0.4158, 0.01, -0.0895, -0.431, 0.01, -0.0874, -0.431, 0.01, -0.0852, -0.4288, 0.01, -0.0874, -0.4288, 0.01, -0.0852,
-0.431, 0.01, -0.083, -0.431, 0.01, -0.0808, -0.4288, 0.01, -0.083, -0.4288, 0.01, -0.0808, -0.4266, 0.01, -0.0874, -0.4266, 0.01, -0.0852, -0.4245, 0.01, -0.0874, -0.4245, 0.01, -0.0852, -0.4266, 0.01, -0.083, -0.4266, 0.01, -0.0808, -0.4245, 0.01, -0.083, -0.4245, 0.01, -0.0808, -0.4307, 0.0061, -0.072, -0.4307, 0.0083, -0.072, -0.4287, 0.0061, -0.0741, -0.4297, 0.0061, -0.0731, -0.4287, 0.0083, -0.0741, -0.4297, 0.0083, -0.0731, -0.431, 0.01, -0.0787, -0.431, 0.01, -0.0765, -0.4288, 0.01, -0.0787, -0.4288, 0.01, -0.0765, -0.431, 0.01, -0.0743, -0.4307, 0.01, -0.072, -0.4287, 0.01, -0.0741, -0.4297, 0.01, -0.0731, -0.4263, 0.0061, -0.076, -0.4263, 0.0083, -0.076, -0.4238, 0.0061, -0.0779, -0.4249, 0.0061, -0.0771,
-0.4238, 0.0083, -0.0779, -0.4249, 0.0083, -0.0771, -0.4274, 0.0061, -0.0752, -0.4274, 0.0083, -0.0752, -0.4266, 0.01, -0.0787, -0.4263, 0.01, -0.076, -0.4239, 0.01, -0.0778, -0.4249, 0.01, -0.077, -0.4274, 0.01, -0.0751, -0.42, 0.0061, -0.0804, -0.42, 0.0083, -0.0804, -0.4223, 0.01, -0.0874, -0.4223, 0.01, -0.0852, -0.4201, 0.01, -0.0874, -0.4201, 0.01, -0.0852, -0.4223, 0.01, -0.083, -0.4223, 0.01, -0.0808, -0.4201, 0.01, -0.083, -0.42, 0.01, -0.0804, -0.4172, 0.0061, -0.0821, -0.4183, 0.0061, -0.0815, -0.4172, 0.0083, -0.0821, -0.4183, 0.0083, -0.0815, -0.4158, 0.0061, -0.0828, -0.4158, 0.0083, -0.0828, -0.418, 0.01, -0.0874, -0.418, 0.01, -0.0852, -0.4158, 0.01, -0.0874, -0.4158, 0.01, -0.0852, -0.4174, 0.01, -0.082,
-0.4184, 0.01, -0.0814, -0.4155, 0.01, -0.083, -0.4223, 0.0061, -0.0789, -0.4223, 0.0083, -0.0789, -0.4211, 0.0061, -0.0797, -0.4211, 0.0083, -0.0797, -0.4221, 0.01, -0.079, -0.4211, 0.01, -0.0797, -0.4799, 0.0061, -0.0538, -0.4799, 0.0083, -0.0538, -0.4799, 0.01, -0.0538, -0.4777, 0.0061, -0.0559, -0.4788, 0.0061, -0.0549, -0.4777, 0.0083, -0.0559, -0.4788, 0.0083, -0.0549, -0.4766, 0.0061, -0.0569, -0.4766, 0.0083, -0.0569, -0.4777, 0.01, -0.0559, -0.4788, 0.01, -0.0549, -0.4768, 0.01, -0.0567, -0.4766, 0.01, -0.0548, -0.469, 0.0061, -0.0624, -0.469, 0.0083, -0.0624, -0.4679, 0.0061, -0.0631, -0.4679, 0.0083, -0.0631, -0.469, 0.01, -0.0624, -0.468, 0.01, -0.063, -0.4743, 0.0061, -0.0588, -0.4743, 0.0083, -0.0588, -0.4717, 0.0061, -0.0606,
-0.4728, 0.0061, -0.0599, -0.4717, 0.0083, -0.0606, -0.4728, 0.0083, -0.0599, -0.4753, 0.0061, -0.058, -0.4753, 0.0083, -0.058, -0.4742, 0.01, -0.0589, -0.4717, 0.01, -0.0607, -0.4728, 0.01, -0.0599, -0.4753, 0.01, -0.058, -0.4744, 0.01, -0.0548, -0.4722, 0.01, -0.057, -0.4722, 0.01, -0.0548, -0.4701, 0.0061, -0.0617, -0.4701, 0.0083, -0.0617, -0.4701, 0.01, -0.0617, -0.47, 0.01, -0.0591, -0.4679, 0.01, -0.0613, -0.4679, 0.01, -0.0591, -0.47, 0.01, -0.057, -0.47, 0.01, -0.0548, -0.4679, 0.01, -0.057, -0.4679, 0.01, -0.0548, -0.482, 0.0061, -0.0516, -0.483, 0.0061, -0.0505, -0.482, 0.0083, -0.0516, -0.483, 0.0083, -0.0505, -0.481, 0.0061, -0.0527, -0.481, 0.0083, -0.0527, -0.4841, 0.0061, -0.0493, -0.4841, 0.0083, -0.0493,
-0.482, 0.01, -0.0516, -0.4828, 0.01, -0.0507, -0.481, 0.01, -0.0527, -0.4809, 0.01, -0.0505, -0.4841, 0.01, -0.0492, -0.4831, 0.01, -0.0461, -0.4809, 0.01, -0.0483, -0.4809, 0.01, -0.0461, -0.4787, 0.01, -0.0526, -0.4787, 0.01, -0.0505, -0.4766, 0.01, -0.0526, -0.4766, 0.01, -0.0505, -0.4787, 0.01, -0.0483, -0.4787, 0.01, -0.0461, -0.4766, 0.01, -0.0483, -0.4766, 0.01, -0.0461, -0.4807, 0.0061, -0.0413, -0.4807, 0.0083, -0.0413, -0.4831, 0.0061, -0.0397, -0.4831, 0.0083, -0.0397, -0.4818, 0.0061, -0.0406, -0.4818, 0.0083, -0.0406, -0.4831, 0.01, -0.0439, -0.4831, 0.01, -0.0418, -0.4809, 0.01, -0.0439, -0.4807, 0.01, -0.0413, -0.4828, 0.01, -0.0398, -0.4818, 0.01, -0.0405, -0.4779, 0.0061, -0.0431, -0.479, 0.0061, -0.0424,
-0.4779, 0.0083, -0.0431, -0.479, 0.0083, -0.0424, -0.4766, 0.0061, -0.0439, -0.4766, 0.0083, -0.0439, -0.4785, 0.0061, -0.0368, -0.4785, 0.0083, -0.0368, -0.4766, 0.0061, -0.0377, -0.4766, 0.0083, -0.0377, -0.4781, 0.01, -0.043, -0.4791, 0.01, -0.0423, -0.4763, 0.01, -0.044, -0.4783, 0.01, -0.0369, -0.4768, 0.01, -0.0376, -0.4738, 0.0061, -0.0453, -0.4738, 0.0083, -0.0453, -0.4722, 0.0061, -0.0459, -0.4722, 0.0083, -0.0459, -0.4744, 0.01, -0.0526, -0.4744, 0.01, -0.0505, -0.4722, 0.01, -0.0526, -0.4722, 0.01, -0.0505, -0.4744, 0.01, -0.0483, -0.474, 0.01, -0.0452, -0.4722, 0.01, -0.0483, -0.472, 0.01, -0.046, -0.4701, 0.0061, -0.0468, -0.4701, 0.0083, -0.0468, -0.4678, 0.0061, -0.0476, -0.4689, 0.0061, -0.0472, -0.4678, 0.0083, -0.0476,
-0.4689, 0.0083, -0.0472, -0.47, 0.01, -0.0526, -0.47, 0.01, -0.0505, -0.4679, 0.01, -0.0526, -0.4679, 0.01, -0.0505, -0.47, 0.01, -0.0483, -0.47, 0.01, -0.0468, -0.4679, 0.01, -0.0476, -0.4689, 0.01, -0.0472, -0.4749, 0.0061, -0.0448, -0.4749, 0.0083, -0.0448, -0.4739, 0.0061, -0.0387, -0.475, 0.0061, -0.0383, -0.4739, 0.0083, -0.0387, -0.475, 0.0083, -0.0383, -0.4722, 0.0061, -0.0393, -0.4722, 0.0083, -0.0393, -0.475, 0.01, -0.0447, -0.4737, 0.01, -0.0388, -0.4747, 0.01, -0.0384, -0.4723, 0.01, -0.0393, -0.4722, 0.01, -0.0374, -0.4671, 0.0061, -0.0408, -0.4671, 0.0083, -0.0408, -0.47, 0.0061, -0.04, -0.47, 0.0083, -0.04, -0.4682, 0.0061, -0.0405, -0.4682, 0.0083, -0.0405, -0.467, 0.01, -0.0408, -0.4702, 0.01, -0.0399,
-0.47, 0.01, -0.0374, -0.468, 0.01, -0.0405, -0.4679, 0.01, -0.0374, -0.4649, 0.0061, -0.0647, -0.466, 0.0061, -0.0641, -0.4649, 0.0083, -0.0647, -0.466, 0.0083, -0.0641, -0.4635, 0.0061, -0.0654, -0.4635, 0.0083, -0.0654, -0.4648, 0.01, -0.0647, -0.4659, 0.01, -0.0641, -0.4637, 0.01, -0.0653, -0.4635, 0.01, -0.0635, -0.4607, 0.0061, -0.067, -0.4607, 0.0083, -0.067, -0.4582, 0.0061, -0.069, -0.4593, 0.0061, -0.068, -0.4582, 0.0083, -0.069, -0.4593, 0.0083, -0.068, -0.4618, 0.0061, -0.0663, -0.4618, 0.0083, -0.0663, -0.4606, 0.01, -0.067, -0.4582, 0.01, -0.069, -0.4593, 0.01, -0.068, -0.4617, 0.01, -0.0664, -0.4614, 0.01, -0.0635, -0.4592, 0.01, -0.0656, -0.4592, 0.01, -0.0635, -0.4657, 0.01, -0.0613, -0.4657, 0.01, -0.0591,
-0.4635, 0.01, -0.0613, -0.4635, 0.01, -0.0591, -0.4657, 0.01, -0.057, -0.4657, 0.01, -0.0548, -0.4635, 0.01, -0.057, -0.4635, 0.01, -0.0548, -0.4614, 0.01, -0.0613, -0.4614, 0.01, -0.0591, -0.4592, 0.01, -0.0613, -0.4592, 0.01, -0.0591, -0.4614, 0.01, -0.057, -0.4614, 0.01, -0.0548, -0.4592, 0.01, -0.057, -0.4592, 0.01, -0.0548, -0.4572, 0.0061, -0.0701, -0.4572, 0.0083, -0.0701, -0.4572, 0.01, -0.0701, -0.457, 0.01, -0.0678, -0.4549, 0.01, -0.07, -0.4549, 0.01, -0.0678, -0.457, 0.01, -0.0656, -0.457, 0.01, -0.0635, -0.4549, 0.01, -0.0656, -0.4549, 0.01, -0.0635, -0.4527, 0.01, -0.07, -0.4527, 0.01, -0.0678, -0.4505, 0.01, -0.07, -0.4505, 0.01, -0.0678, -0.4527, 0.01, -0.0656, -0.4527, 0.01, -0.0635,
-0.4505, 0.01, -0.0656, -0.4505, 0.01, -0.0635, -0.457, 0.01, -0.0613, -0.457, 0.01, -0.0591, -0.4549, 0.01, -0.0613, -0.4549, 0.01, -0.0591, -0.457, 0.01, -0.057, -0.457, 0.01, -0.0548, -0.4549, 0.01, -0.057, -0.4549, 0.01, -0.0548, -0.4527, 0.01, -0.0613, -0.4527, 0.01, -0.0591, -0.4505, 0.01, -0.0613, -0.4505, 0.01, -0.0591, -0.4527, 0.01, -0.057, -0.4527, 0.01, -0.0548, -0.4505, 0.01, -0.057, -0.4505, 0.01, -0.0548, -0.4657, 0.0061, -0.0482, -0.4657, 0.0083, -0.0482, -0.4635, 0.0061, -0.0487, -0.4635, 0.0083, -0.0487, -0.4657, 0.01, -0.0526, -0.4657, 0.01, -0.0505, -0.4635, 0.01, -0.0526, -0.4635, 0.01, -0.0505, -0.4656, 0.01, -0.0483, -0.4635, 0.01, -0.0488, -0.4604, 0.0061, -0.0494, -0.4604, 0.0083, -0.0494,
-0.4592, 0.0061, -0.0497, -0.4592, 0.0083, -0.0497, -0.4615, 0.0061, -0.0492, -0.4615, 0.0083, -0.0492, -0.4614, 0.01, -0.0526, -0.461, 0.01, -0.0493, -0.4592, 0.01, -0.0526, -0.459, 0.01, -0.0497, -0.4618, 0.01, -0.0491, -0.4657, 0.0061, -0.0411, -0.4657, 0.0083, -0.0411, -0.4635, 0.0061, -0.0416, -0.4635, 0.0083, -0.0416, -0.4658, 0.01, -0.0411, -0.4636, 0.01, -0.0416, -0.4657, 0.01, -0.0396, -0.4657, 0.01, -0.0374, -0.4635, 0.01, -0.0396, -0.4635, 0.01, -0.0374, -0.4614, 0.0061, -0.0421, -0.4614, 0.0083, -0.0421, -0.4592, 0.0061, -0.0424, -0.4592, 0.0083, -0.0424, -0.4615, 0.01, -0.0421, -0.4593, 0.01, -0.0424, -0.4614, 0.01, -0.0396, -0.4614, 0.01, -0.0374, -0.4592, 0.01, -0.0396, -0.4592, 0.01, -0.0374, -0.457, 0.0061, -0.05,
-0.457, 0.0083, -0.05, -0.4549, 0.0061, -0.0502, -0.4549, 0.0083, -0.0502, -0.457, 0.01, -0.0526, -0.4569, 0.01, -0.05, -0.4549, 0.01, -0.0526, -0.4548, 0.01, -0.0502, -0.4527, 0.0061, -0.0503, -0.4527, 0.0083, -0.0503, -0.4505, 0.0061, -0.0503, -0.4505, 0.0083, -0.0503, -0.4527, 0.01, -0.0526, -0.4527, 0.01, -0.0503, -0.4505, 0.01, -0.0526, -0.4505, 0.01, -0.0503, -0.457, 0.0061, -0.0425, -0.457, 0.0083, -0.0425, -0.4549, 0.0061, -0.0425, -0.4549, 0.0083, -0.0425, -0.457, 0.01, -0.0425, -0.4549, 0.01, -0.0425, -0.457, 0.01, -0.0396, -0.457, 0.01, -0.0374, -0.4549, 0.01, -0.0396, -0.4549, 0.01, -0.0374, -0.4527, 0.0061, -0.0425, -0.4527, 0.0083, -0.0425, -0.4505, 0.0061, -0.0425, -0.4505, 0.0083, -0.0425, -0.4527, 0.01, -0.0425,
-0.4505, 0.01, -0.0425, -0.4527, 0.01, -0.0396, -0.4527, 0.01, -0.0374, -0.4505, 0.01, -0.0396, -0.4505, 0.01, -0.0374, -0.4824, 0.0061, -0.0345, -0.4835, 0.0061, -0.0337, -0.4824, 0.0083, -0.0345, -0.4835, 0.0083, -0.0337, -0.4809, 0.0061, -0.0355, -0.4809, 0.0083, -0.0355, -0.4824, 0.01, -0.0345, -0.4835, 0.01, -0.0338, -0.4812, 0.01, -0.0353, -0.4809, 0.01, -0.0331, -0.4831, 0.01, -0.0309, -0.4831, 0.01, -0.0288, -0.4809, 0.01, -0.0309, -0.4809, 0.01, -0.0288, -0.4795, 0.0061, -0.0362, -0.4795, 0.0083, -0.0362, -0.4764, 0.0061, -0.0279, -0.4764, 0.0083, -0.0279, -0.4793, 0.01, -0.0363, -0.4787, 0.01, -0.0331, -0.4766, 0.01, -0.0353, -0.4766, 0.01, -0.0331, -0.4787, 0.01, -0.0309, -0.4787, 0.01, -0.0288, -0.4766, 0.01, -0.0309,
-0.4765, 0.01, -0.0279, -0.483, 0.0061, -0.0259, -0.4841, 0.0061, -0.0255, -0.483, 0.0083, -0.0259, -0.4841, 0.0083, -0.0255, -0.4809, 0.0061, -0.0266, -0.4809, 0.0083, -0.0266, -0.483, 0.01, -0.0259, -0.4841, 0.01, -0.0255, -0.4807, 0.01, -0.0267, -0.4787, 0.0061, -0.0273, -0.4787, 0.0083, -0.0273, -0.4775, 0.0061, -0.0276, -0.4775, 0.0083, -0.0276, -0.4787, 0.01, -0.0273, -0.4775, 0.01, -0.0276, -0.4744, 0.0061, -0.0283, -0.4744, 0.0083, -0.0283, -0.4722, 0.0061, -0.0286, -0.4722, 0.0083, -0.0286, -0.4744, 0.01, -0.0353, -0.4744, 0.01, -0.0331, -0.4722, 0.01, -0.0353, -0.4722, 0.01, -0.0331, -0.4744, 0.01, -0.0309, -0.4743, 0.01, -0.0284, -0.4722, 0.01, -0.0309, -0.4722, 0.01, -0.0286, -0.47, 0.0061, -0.0287, -0.47, 0.0083, -0.0287,
-0.4679, 0.0061, -0.0287, -0.4679, 0.0083, -0.0287, -0.47, 0.01, -0.0353, -0.47, 0.01, -0.0331, -0.4679, 0.01, -0.0353, -0.4679, 0.01, -0.0331, -0.47, 0.01, -0.0309, -0.47, 0.01, -0.0287, -0.4679, 0.01, -0.0309, -0.4679, 0.01, -0.0287, -0.4801, 0.0061, -0.0148, -0.4801, 0.0083, -0.0148, -0.4824, 0.0061, -0.014, -0.4823, 0.0061, -0.0114, -0.4824, 0.0083, -0.014, -0.4823, 0.0083, -0.0114, -0.4812, 0.0061, -0.0144, -0.4812, 0.0083, -0.0144, -0.4801, 0.01, -0.0148, -0.4824, 0.01, -0.014, -0.4823, 0.01, -0.0114, -0.4811, 0.01, -0.0145, -0.4809, 0.01, -0.0114, -0.4787, 0.0061, -0.0152, -0.4787, 0.0083, -0.0152, -0.4766, 0.0061, -0.0159, -0.4766, 0.0083, -0.0159, -0.4788, 0.01, -0.0152, -0.4767, 0.01, -0.0158, -0.4787, 0.01, -0.0136,
-0.4787, 0.01, -0.0114, -0.4766, 0.01, -0.0136, -0.4766, 0.01, -0.0114, -0.4821, 0.0061, -0.0094, -0.4821, 0.0083, -0.0094, -0.482, 0.0061, -0.0083, -0.4818, 0.0061, -0.0071, -0.482, 0.0083, -0.0083, -0.4818, 0.0083, -0.0071, -0.4815, 0.0061, -0.0049, -0.4812, 0.0061, -0.0027, -0.4815, 0.0083, -0.0049, -0.4812, 0.0083, -0.0027, -0.4821, 0.01, -0.0097, -0.482, 0.01, -0.0089, -0.4817, 0.01, -0.0069, -0.4814, 0.01, -0.0048, -0.4811, 0.01, -0.0026, -0.4787, 0.01, -0.0092, -0.4787, 0.01, -0.0071, -0.4766, 0.01, -0.0092, -0.4766, 0.01, -0.0071, -0.4787, 0.01, -0.0049, -0.4787, 0.01, -0.0027, -0.4766, 0.01, -0.0049, -0.4766, 0.01, -0.0027, -0.4744, 0.0061, -0.0164, -0.4744, 0.0083, -0.0164, -0.4713, 0.0061, -0.0168, -0.4724, 0.0061, -0.0167,
-0.4713, 0.0083, -0.0168, -0.4724, 0.0083, -0.0167, -0.4746, 0.01, -0.0164, -0.4713, 0.01, -0.0168, -0.4722, 0.01, -0.0168, -0.4744, 0.01, -0.0136, -0.4744, 0.01, -0.0114, -0.4722, 0.01, -0.0136, -0.4722, 0.01, -0.0114, -0.47, 0.0061, -0.0169, -0.47, 0.0083, -0.0169, -0.4679, 0.0061, -0.017, -0.4679, 0.0083, -0.017, -0.47, 0.01, -0.0169, -0.47, 0.01, -0.0157, -0.4679, 0.01, -0.017, -0.4679, 0.01, -0.0157, -0.47, 0.01, -0.0136, -0.47, 0.01, -0.0114, -0.4679, 0.01, -0.0136, -0.4679, 0.01, -0.0114, -0.4744, 0.01, -0.0092, -0.4744, 0.01, -0.0071, -0.4722, 0.01, -0.0092, -0.4722, 0.01, -0.0071, -0.4744, 0.01, -0.0049, -0.4744, 0.01, -0.0027, -0.4722, 0.01, -0.0049, -0.4722, 0.01, -0.0027, -0.47, 0.01, -0.0092,
-0.47, 0.01, -0.0071, -0.4679, 0.01, -0.0092, -0.4679, 0.01, -0.0071, -0.47, 0.01, -0.0049, -0.47, 0.01, -0.0027, -0.4679, 0.01, -0.0049, -0.4679, 0.01, -0.0027, -0.4657, 0.0061, -0.0287, -0.4657, 0.0083, -0.0287, -0.4635, 0.0061, -0.0287, -0.4635, 0.0083, -0.0287, -0.4657, 0.01, -0.0353, -0.4657, 0.01, -0.0331, -0.4635, 0.01, -0.0353, -0.4635, 0.01, -0.0331, -0.4657, 0.01, -0.0309, -0.4657, 0.01, -0.0287, -0.4635, 0.01, -0.0309, -0.4636, 0.01, -0.0287, -0.4614, 0.0061, -0.0284, -0.4614, 0.0083, -0.0284, -0.4592, 0.0061, -0.028, -0.4592, 0.0083, -0.028, -0.4614, 0.01, -0.0353, -0.4614, 0.01, -0.0331, -0.4592, 0.01, -0.0353, -0.4592, 0.01, -0.0331, -0.4614, 0.01, -0.0309, -0.4615, 0.01, -0.0284, -0.4592, 0.01, -0.0309,
-0.4594, 0.01, -0.028, -0.458, 0.0061, -0.0277, -0.458, 0.0083, -0.0277, -0.457, 0.01, -0.0353, -0.457, 0.01, -0.0331, -0.4549, 0.01, -0.0353, -0.4549, 0.01, -0.0331, -0.457, 0.01, -0.0309, -0.4574, 0.01, -0.0276, -0.4549, 0.01, -0.0309, -0.4549, 0.01, -0.0288, -0.4527, 0.01, -0.0353, -0.4527, 0.01, -0.0331, -0.4505, 0.01, -0.0353, -0.4505, 0.01, -0.0331, -0.4527, 0.01, -0.0309, -0.4527, 0.01, -0.0288, -0.4505, 0.01, -0.0309, -0.4505, 0.01, -0.0288, -0.4569, 0.0061, -0.0274, -0.4569, 0.0083, -0.0274, -0.4549, 0.0061, -0.027, -0.4549, 0.0083, -0.027, -0.4566, 0.01, -0.0274, -0.4549, 0.01, -0.027, -0.4527, 0.0061, -0.0265, -0.4527, 0.0083, -0.0265, -0.4505, 0.0061, -0.026, -0.4505, 0.0083, -0.026, -0.4528, 0.01, -0.0265,
-0.4507, 0.01, -0.0261, -0.4657, 0.0061, -0.017, -0.4657, 0.0061, -0.0157, -0.4657, 0.0083, -0.017, -0.4657, 0.0083, -0.0157, -0.4655, 0.0061, -0.0136, -0.4651, 0.0061, -0.0114, -0.4655, 0.0083, -0.0136, -0.4651, 0.0083, -0.0114, -0.4657, 0.01, -0.017, -0.4657, 0.01, -0.0158, -0.4655, 0.01, -0.0136, -0.4652, 0.01, -0.0115, -0.4648, 0.0061, -0.0095, -0.4648, 0.0083, -0.0095, -0.4646, 0.0061, -0.0084, -0.4643, 0.0061, -0.0071, -0.4646, 0.0083, -0.0084, -0.4643, 0.0083, -0.0071, -0.4638, 0.0061, -0.0049, -0.4633, 0.0061, -0.0027, -0.4638, 0.0083, -0.0049, -0.4633, 0.0083, -0.0027, -0.4647, 0.01, -0.0093, -0.4657, 0.01, -0.0071, -0.4645, 0.01, -0.0083, -0.4643, 0.01, -0.0071, -0.4657, 0.01, -0.0049, -0.4657, 0.01, -0.0027, -0.4638, 0.01, -0.005,
-0.4633, 0.01, -0.0029, -0.4483, 0.01, -0.07, -0.4483, 0.01, -0.0678, -0.4462, 0.01, -0.07, -0.4462, 0.01, -0.0678, -0.4483, 0.01, -0.0656, -0.4483, 0.01, -0.0635, -0.4462, 0.01, -0.0656, -0.4462, 0.01, -0.0635, -0.444, 0.01, -0.07, -0.444, 0.01, -0.0678, -0.4418, 0.01, -0.07, -0.4418, 0.01, -0.0678, -0.444, 0.01, -0.0656, -0.444, 0.01, -0.0635, -0.4418, 0.01, -0.0656, -0.4418, 0.01, -0.0635, -0.4483, 0.01, -0.0613, -0.4483, 0.01, -0.0591, -0.4462, 0.01, -0.0613, -0.4462, 0.01, -0.0591, -0.4483, 0.01, -0.057, -0.4483, 0.01, -0.0548, -0.4462, 0.01, -0.057, -0.4462, 0.01, -0.0548, -0.444, 0.01, -0.0613, -0.444, 0.01, -0.0591, -0.4418, 0.01, -0.0613, -0.4418, 0.01, -0.0591, -0.444, 0.01, -0.057,
-0.444, 0.01, -0.0548, -0.4418, 0.01, -0.057, -0.4418, 0.01, -0.0548, -0.4397, 0.01, -0.07, -0.4397, 0.01, -0.0678, -0.4375, 0.01, -0.07, -0.4375, 0.01, -0.0678, -0.4397, 0.01, -0.0656, -0.4397, 0.01, -0.0635, -0.4375, 0.01, -0.0656, -0.4375, 0.01, -0.0635, -0.4353, 0.01, -0.07, -0.4353, 0.01, -0.0678, -0.433, 0.01, -0.069, -0.4332, 0.01, -0.0678, -0.4353, 0.01, -0.0656, -0.4353, 0.01, -0.0635, -0.4332, 0.01, -0.0656, -0.4332, 0.01, -0.0635, -0.4397, 0.01, -0.0613, -0.4397, 0.01, -0.0591, -0.4375, 0.01, -0.0613, -0.4375, 0.01, -0.0591, -0.4397, 0.01, -0.057, -0.4397, 0.01, -0.0548, -0.4375, 0.01, -0.057, -0.4375, 0.01, -0.0548, -0.4353, 0.01, -0.0613, -0.4353, 0.01, -0.0591, -0.4332, 0.01, -0.0613,
-0.4332, 0.01, -0.0591, -0.4353, 0.01, -0.057, -0.4353, 0.01, -0.0548, -0.4332, 0.01, -0.057, -0.4332, 0.01, -0.0548, -0.4483, 0.0061, -0.0503, -0.4483, 0.0083, -0.0503, -0.4462, 0.0061, -0.0501, -0.4462, 0.0083, -0.0501, -0.4483, 0.01, -0.0526, -0.4484, 0.01, -0.0503, -0.4462, 0.01, -0.0526, -0.4463, 0.01, -0.0501, -0.444, 0.0061, -0.0497, -0.444, 0.0083, -0.0497, -0.4424, 0.0061, -0.0495, -0.4424, 0.0083, -0.0495, -0.4413, 0.0061, -0.0493, -0.4413, 0.0083, -0.0493, -0.444, 0.01, -0.0526, -0.4441, 0.01, -0.0498, -0.4418, 0.01, -0.0526, -0.442, 0.01, -0.0494, -0.4411, 0.01, -0.0492, -0.4483, 0.0061, -0.0425, -0.4483, 0.0083, -0.0425, -0.4462, 0.0061, -0.0423, -0.4462, 0.0083, -0.0423, -0.4483, 0.01, -0.0425, -0.4461, 0.01, -0.0423,
-0.4483, 0.01, -0.0396, -0.4483, 0.01, -0.0374, -0.4462, 0.01, -0.0396, -0.4462, 0.01, -0.0374, -0.444, 0.0061, -0.042, -0.444, 0.0083, -0.042, -0.4418, 0.0061, -0.0416, -0.4418, 0.0083, -0.0416, -0.4439, 0.01, -0.042, -0.4417, 0.01, -0.0416, -0.444, 0.01, -0.0396, -0.444, 0.01, -0.0374, -0.4418, 0.01, -0.0396, -0.4418, 0.01, -0.0374, -0.4397, 0.0061, -0.049, -0.4397, 0.0083, -0.049, -0.4375, 0.0061, -0.0486, -0.4375, 0.0083, -0.0486, -0.4397, 0.01, -0.0526, -0.4397, 0.01, -0.0505, -0.4375, 0.01, -0.0526, -0.4375, 0.01, -0.0505, -0.4397, 0.01, -0.049, -0.4376, 0.01, -0.0486, -0.4353, 0.0061, -0.0481, -0.4353, 0.0083, -0.0481, -0.4331, 0.0061, -0.0476, -0.4331, 0.0083, -0.0476, -0.4353, 0.01, -0.0526, -0.4353, 0.01, -0.0505,
-0.4332, 0.01, -0.0526, -0.4332, 0.01, -0.0505, -0.4355, 0.01, -0.0482, -0.4334, 0.01, -0.0476, -0.4397, 0.0061, -0.0412, -0.4397, 0.0083, -0.0412, -0.4381, 0.0061, -0.0408, -0.4381, 0.0083, -0.0408, -0.437, 0.0061, -0.0405, -0.437, 0.0083, -0.0405, -0.4396, 0.01, -0.0412, -0.4382, 0.01, -0.0409, -0.4397, 0.01, -0.0396, -0.4397, 0.01, -0.0374, -0.4373, 0.01, -0.0406, -0.4375, 0.01, -0.0374, -0.4353, 0.0061, -0.0401, -0.4353, 0.0083, -0.0401, -0.4331, 0.0061, -0.0396, -0.4331, 0.0083, -0.0396, -0.4351, 0.01, -0.0401, -0.4353, 0.01, -0.0374, -0.433, 0.01, -0.0396, -0.4332, 0.01, -0.0374, -0.4328, 0.0061, -0.069, -0.4308, 0.0061, -0.0689, -0.4328, 0.0083, -0.069, -0.4308, 0.0083, -0.0689, -0.4288, 0.0061, -0.0686, -0.4288, 0.0083, -0.0686,
-0.4328, 0.01, -0.069, -0.4311, 0.01, -0.0689, -0.4286, 0.01, -0.0686, -0.431, 0.01, -0.0656, -0.431, 0.01, -0.0635, -0.4288, 0.01, -0.0656, -0.4288, 0.01, -0.0635, -0.4266, 0.0061, -0.0681, -0.4266, 0.0083, -0.0681, -0.4245, 0.0061, -0.0675, -0.4245, 0.0083, -0.0675, -0.4265, 0.01, -0.0681, -0.4244, 0.01, -0.0675, -0.4266, 0.01, -0.0656, -0.4266, 0.01, -0.0635, -0.4245, 0.01, -0.0656, -0.4245, 0.01, -0.0635, -0.431, 0.01, -0.0613, -0.431, 0.01, -0.0591, -0.4288, 0.01, -0.0613, -0.4288, 0.01, -0.0591, -0.431, 0.01, -0.057, -0.431, 0.01, -0.0548, -0.4288, 0.01, -0.057, -0.4288, 0.01, -0.0548, -0.4266, 0.01, -0.0613, -0.4266, 0.01, -0.0591, -0.4245, 0.01, -0.0613, -0.4245, 0.01, -0.0591, -0.4266, 0.01, -0.057,
-0.4266, 0.01, -0.0548, -0.4245, 0.01, -0.057, -0.4245, 0.01, -0.0548, -0.4227, 0.0061, -0.067, -0.4227, 0.0083, -0.067, -0.4216, 0.0061, -0.0666, -0.4216, 0.0083, -0.0666, -0.4201, 0.0061, -0.066, -0.4201, 0.0083, -0.066, -0.4229, 0.01, -0.067, -0.4219, 0.01, -0.0667, -0.4223, 0.01, -0.0635, -0.4199, 0.01, -0.0659, -0.4201, 0.01, -0.0635, -0.418, 0.0061, -0.0651, -0.418, 0.0083, -0.0651, -0.4167, 0.0061, -0.0646, -0.4156, 0.0061, -0.0642, -0.4167, 0.0083, -0.0646, -0.4156, 0.0083, -0.0642, -0.4179, 0.01, -0.0651, -0.418, 0.01, -0.0635, -0.4167, 0.01, -0.0647, -0.4157, 0.01, -0.0642, -0.4223, 0.01, -0.0613, -0.4223, 0.01, -0.0591, -0.4201, 0.01, -0.0613, -0.4201, 0.01, -0.0591, -0.4223, 0.01, -0.057, -0.4223, 0.01, -0.0548,
-0.4201, 0.01, -0.057, -0.4201, 0.01, -0.0548, -0.418, 0.01, -0.0613, -0.418, 0.01, -0.0591, -0.4158, 0.01, -0.0613, -0.4158, 0.01, -0.0591, -0.418, 0.01, -0.057, -0.418, 0.01, -0.0548, -0.4158, 0.01, -0.057, -0.4158, 0.01, -0.0548, -0.432, 0.0061, -0.0472, -0.4309, 0.0061, -0.0469, -0.432, 0.0083, -0.0472, -0.4309, 0.0083, -0.0469, -0.4288, 0.0061, -0.0462, -0.4288, 0.0083, -0.0462, -0.431, 0.01, -0.0526, -0.431, 0.01, -0.0505, -0.4288, 0.01, -0.0526, -0.4288, 0.01, -0.0505, -0.4315, 0.01, -0.0471, -0.4306, 0.01, -0.0468, -0.4288, 0.01, -0.0483, -0.429, 0.01, -0.0463, -0.4267, 0.0061, -0.0454, -0.4267, 0.0083, -0.0454, -0.4266, 0.01, -0.0526, -0.4266, 0.01, -0.0505, -0.4245, 0.01, -0.0526, -0.4245, 0.01, -0.0505,
-0.4266, 0.01, -0.0483, -0.4267, 0.01, -0.0454, -0.4245, 0.01, -0.0483, -0.4245, 0.01, -0.0461, -0.431, 0.0061, -0.039, -0.431, 0.0083, -0.039, -0.4297, 0.0061, -0.0386, -0.4286, 0.0061, -0.0382, -0.4297, 0.0083, -0.0386, -0.4286, 0.0083, -0.0382, -0.4309, 0.01, -0.039, -0.431, 0.01, -0.0374, -0.4298, 0.01, -0.0386, -0.4287, 0.01, -0.0382, -0.4256, 0.0061, -0.045, -0.4256, 0.0083, -0.045, -0.4245, 0.0061, -0.0446, -0.4245, 0.0083, -0.0446, -0.4266, 0.0061, -0.0373, -0.4266, 0.0083, -0.0373, -0.425, 0.0061, -0.0366, -0.425, 0.0083, -0.0366, -0.4256, 0.01, -0.045, -0.4245, 0.01, -0.0446, -0.4265, 0.01, -0.0373, -0.4251, 0.01, -0.0367, -0.4223, 0.01, -0.0526, -0.4223, 0.01, -0.0505, -0.4201, 0.01, -0.0526, -0.4201, 0.01, -0.0505,
-0.4223, 0.01, -0.0483, -0.4223, 0.01, -0.0461, -0.4201, 0.01, -0.0483, -0.4201, 0.01, -0.0461, -0.418, 0.01, -0.0526, -0.418, 0.01, -0.0505, -0.4158, 0.01, -0.0526, -0.4158, 0.01, -0.0505, -0.418, 0.01, -0.0483, -0.418, 0.01, -0.0461, -0.4158, 0.01, -0.0483, -0.4158, 0.01, -0.0461, -0.4223, 0.0061, -0.0436, -0.4223, 0.0083, -0.0436, -0.4209, 0.0061, -0.043, -0.4198, 0.0061, -0.0425, -0.4209, 0.0083, -0.043, -0.4198, 0.0083, -0.0425, -0.4226, 0.01, -0.0437, -0.4206, 0.01, -0.0429, -0.4196, 0.01, -0.0425, -0.418, 0.0061, -0.0418, -0.418, 0.0083, -0.0418, -0.4162, 0.0061, -0.041, -0.4162, 0.0083, -0.041, -0.4152, 0.0061, -0.0405, -0.4152, 0.0083, -0.0405, -0.418, 0.01, -0.0439, -0.4182, 0.01, -0.0419, -0.4158, 0.01, -0.0439,
-0.4161, 0.01, -0.0409, -0.4151, 0.01, -0.0404, -0.4483, 0.01, -0.0353, -0.4483, 0.01, -0.0331, -0.4462, 0.01, -0.0353, -0.4462, 0.01, -0.0331, -0.4483, 0.01, -0.0309, -0.4483, 0.01, -0.0288, -0.4462, 0.01, -0.0309, -0.4462, 0.01, -0.0288, -0.444, 0.01, -0.0353, -0.444, 0.01, -0.0331, -0.4418, 0.01, -0.0353, -0.4418, 0.01, -0.0331, -0.444, 0.01, -0.0309, -0.444, 0.01, -0.0288, -0.4418, 0.01, -0.0309, -0.4418, 0.01, -0.0288, -0.4489, 0.0061, -0.0256, -0.4479, 0.0061, -0.0253, -0.4489, 0.0083, -0.0256, -0.4479, 0.0083, -0.0253, -0.4462, 0.0061, -0.0248, -0.4462, 0.0083, -0.0248, -0.4486, 0.01, -0.0255, -0.4477, 0.01, -0.0253, -0.4462, 0.01, -0.0266, -0.4463, 0.01, -0.0249, -0.444, 0.0061, -0.0241, -0.444, 0.0083, -0.0241,
-0.4424, 0.0061, -0.0235, -0.4424, 0.0083, -0.0235, -0.4413, 0.0061, -0.0231, -0.4413, 0.0083, -0.0231, -0.444, 0.01, -0.0266, -0.4442, 0.01, -0.0242, -0.4418, 0.01, -0.0266, -0.4422, 0.01, -0.0234, -0.4412, 0.01, -0.0231, -0.4397, 0.01, -0.0353, -0.4397, 0.01, -0.0331, -0.4375, 0.01, -0.0353, -0.4375, 0.01, -0.0331, -0.4397, 0.01, -0.0309, -0.4397, 0.01, -0.0288, -0.4375, 0.01, -0.0309, -0.4375, 0.01, -0.0288, -0.4353, 0.01, -0.0353, -0.4353, 0.01, -0.0331, -0.4332, 0.01, -0.0353, -0.4332, 0.01, -0.0331, -0.4353, 0.01, -0.0309, -0.4353, 0.01, -0.0288, -0.4332, 0.01, -0.0309, -0.4332, 0.01, -0.0288, -0.4397, 0.0061, -0.0225, -0.4397, 0.0083, -0.0225, -0.4375, 0.0061, -0.0216, -0.4364, 0.0061, -0.0212, -0.4375, 0.0083, -0.0216,
-0.4364, 0.0083, -0.0212, -0.4397, 0.01, -0.0266, -0.4397, 0.01, -0.0244, -0.4375, 0.01, -0.0266, -0.4375, 0.01, -0.0244, -0.4398, 0.01, -0.0225, -0.4375, 0.01, -0.0216, -0.4364, 0.01, -0.0212, -0.4353, 0.0061, -0.0207, -0.4353, 0.0083, -0.0207, -0.4332, 0.0061, -0.0197, -0.4332, 0.0083, -0.0197, -0.4353, 0.01, -0.0266, -0.4353, 0.01, -0.0244, -0.4332, 0.01, -0.0266, -0.4332, 0.01, -0.0244, -0.4353, 0.01, -0.0222, -0.4354, 0.01, -0.0207, -0.4332, 0.01, -0.0222, -0.4334, 0.01, -0.0199, -0.431, 0.01, -0.0353, -0.431, 0.01, -0.0331, -0.4288, 0.01, -0.0353, -0.4288, 0.01, -0.0331, -0.431, 0.01, -0.0309, -0.431, 0.01, -0.0288, -0.4288, 0.01, -0.0309, -0.4288, 0.01, -0.0288, -0.4239, 0.0061, -0.0361, -0.4239, 0.0083, -0.0361,
-0.4266, 0.01, -0.0353, -0.4266, 0.01, -0.0331, -0.4241, 0.01, -0.0362, -0.4245, 0.01, -0.0331, -0.4266, 0.01, -0.0309, -0.4266, 0.01, -0.0288, -0.4245, 0.01, -0.0309, -0.4245, 0.01, -0.0288, -0.4318, 0.0061, -0.0191, -0.4318, 0.0083, -0.0191, -0.431, 0.01, -0.0266, -0.431, 0.01, -0.0244, -0.4288, 0.01, -0.0266, -0.4288, 0.01, -0.0244, -0.431, 0.01, -0.0222, -0.4315, 0.01, -0.019, -0.4288, 0.01, -0.0222, -0.4288, 0.01, -0.0201, -0.4266, 0.01, -0.0266, -0.4266, 0.01, -0.0244, -0.4245, 0.01, -0.0266, -0.4245, 0.01, -0.0244, -0.4266, 0.01, -0.0222, -0.4266, 0.01, -0.0201, -0.4245, 0.01, -0.0222, -0.4245, 0.01, -0.0201, -0.4223, 0.0061, -0.0354, -0.4223, 0.0083, -0.0354, -0.4204, 0.0061, -0.0345, -0.4194, 0.0061, -0.034,
-0.4204, 0.0083, -0.0345, -0.4194, 0.0083, -0.034, -0.4221, 0.01, -0.0353, -0.4223, 0.01, -0.0331, -0.4206, 0.01, -0.0346, -0.4196, 0.01, -0.0342, -0.4223, 0.01, -0.0309, -0.4223, 0.01, -0.0288, -0.4201, 0.01, -0.0309, -0.4201, 0.01, -0.0288, -0.4179, 0.0061, -0.0333, -0.4179, 0.0083, -0.0333, -0.4162, 0.0061, -0.0324, -0.4162, 0.0083, -0.0324, -0.4151, 0.0061, -0.0318, -0.4151, 0.0083, -0.0318, -0.4177, 0.01, -0.0332, -0.4163, 0.01, -0.0325, -0.418, 0.01, -0.0309, -0.418, 0.01, -0.0288, -0.4153, 0.01, -0.0319, -0.4158, 0.01, -0.0288, -0.4223, 0.01, -0.0266, -0.4223, 0.01, -0.0244, -0.4201, 0.01, -0.0266, -0.4201, 0.01, -0.0244, -0.4223, 0.01, -0.0222, -0.4223, 0.01, -0.0201, -0.4201, 0.01, -0.0222, -0.4201, 0.01, -0.0201,
-0.418, 0.01, -0.0266, -0.418, 0.01, -0.0244, -0.4158, 0.01, -0.0266, -0.4158, 0.01, -0.0244, -0.418, 0.01, -0.0222, -0.418, 0.01, -0.0201, -0.4158, 0.01, -0.0222, -0.4158, 0.01, -0.0201, -0.4307, 0.0061, -0.0186, -0.4307, 0.0083, -0.0186, -0.4288, 0.0061, -0.0178, -0.4288, 0.0083, -0.0178, -0.4305, 0.01, -0.0185, -0.429, 0.01, -0.0179, -0.4272, 0.0061, -0.017, -0.4262, 0.0061, -0.0165, -0.4272, 0.0083, -0.017, -0.4262, 0.0083, -0.0165, -0.4245, 0.0061, -0.0157, -0.4245, 0.0083, -0.0157, -0.427, 0.01, -0.017, -0.426, 0.01, -0.0165, -0.4245, 0.01, -0.0179, -0.4247, 0.01, -0.0158, -0.4258, 0.0061, -0.0017, -0.4258, 0.0083, -0.0017, -0.4245, 0.0061, -0.0023, -0.4245, 0.0083, -0.0023, -0.4257, 0.01, -0.0018, -0.4246, 0.01, -0.0023,
-0.423, 0.0061, -0.0149, -0.423, 0.0083, -0.0149, -0.4219, 0.0061, -0.0143, -0.4219, 0.0083, -0.0143, -0.4201, 0.0061, -0.0134, -0.4201, 0.0083, -0.0134, -0.4223, 0.01, -0.0179, -0.4228, 0.01, -0.0148, -0.4201, 0.01, -0.0179, -0.4201, 0.01, -0.0157, -0.4217, 0.01, -0.0142, -0.4204, 0.01, -0.0135, -0.4187, 0.0061, -0.0127, -0.4176, 0.0061, -0.0121, -0.4187, 0.0083, -0.0127, -0.4176, 0.0083, -0.0121, -0.4158, 0.0061, -0.0111, -0.4158, 0.0083, -0.0111, -0.418, 0.01, -0.0179, -0.418, 0.01, -0.0157, -0.4158, 0.01, -0.0179, -0.4158, 0.01, -0.0157, -0.4185, 0.01, -0.0126, -0.4175, 0.01, -0.012, -0.4158, 0.01, -0.0136, -0.4161, 0.01, -0.0113, -0.4223, 0.0061, -0.003, -0.4223, 0.0083, -0.003, -0.4201, 0.0061, -0.0032, -0.4201, 0.0083, -0.0032,
-0.4224, 0.01, -0.0029, -0.4201, 0.01, -0.0032, -0.4179, 0.0061, -0.003, -0.4179, 0.0083, -0.003, -0.4158, 0.0061, -0.0026, -0.4158, 0.0083, -0.0026, -0.4179, 0.01, -0.003, -0.4156, 0.01, -0.0025, -0.3977, 0.0061, -0.1059, -0.3977, 0.0083, -0.1059, -0.3976, 0.01, -0.1059, -0.3963, 0.0061, -0.1061, -0.3963, 0.0083, -0.1061, -0.3941, 0.0061, -0.1066, -0.3941, 0.0083, -0.1066, -0.3963, 0.01, -0.1061, -0.3942, 0.01, -0.1066, -0.3919, 0.0061, -0.1069, -0.3919, 0.0083, -0.1069, -0.3898, 0.0061, -0.1071, -0.3898, 0.0083, -0.1071, -0.392, 0.01, -0.1069, -0.3898, 0.01, -0.1071, -0.3876, 0.0061, -0.1072, -0.3876, 0.0083, -0.1072, -0.3854, 0.0061, -0.1072, -0.3854, 0.0083, -0.1072, -0.3876, 0.01, -0.1072, -0.3854, 0.01, -0.1072, -0.3832, 0.0061, -0.1072,
-0.3832, 0.0083, -0.1072, -0.3811, 0.0061, -0.1072, -0.3811, 0.0083, -0.1072, -0.3832, 0.01, -0.1072, -0.3811, 0.01, -0.1072, -0.413, 0.0061, -0.1016, -0.413, 0.0083, -0.1016, -0.4115, 0.0061, -0.1022, -0.4115, 0.0083, -0.1022, -0.4141, 0.0061, -0.1012, -0.4141, 0.0083, -0.1012, -0.4129, 0.01, -0.1017, -0.4116, 0.01, -0.1022, -0.4139, 0.01, -0.1013, -0.4136, 0.01, -0.0982, -0.4115, 0.01, -0.1004, -0.4115, 0.01, -0.0982, -0.4093, 0.0061, -0.103, -0.4093, 0.0083, -0.103, -0.4067, 0.0061, -0.1038, -0.4078, 0.0061, -0.1035, -0.4067, 0.0083, -0.1038, -0.4078, 0.0083, -0.1035, -0.4095, 0.01, -0.1029, -0.4065, 0.01, -0.1039, -0.4075, 0.01, -0.1036, -0.4093, 0.01, -0.1004, -0.4093, 0.01, -0.0982, -0.4071, 0.01, -0.1004, -0.4071, 0.01, -0.0982,
-0.4136, 0.01, -0.096, -0.4136, 0.01, -0.0939, -0.4115, 0.01, -0.096, -0.4115, 0.01, -0.0939, -0.4136, 0.01, -0.0917, -0.4136, 0.01, -0.0895, -0.4115, 0.01, -0.0917, -0.4115, 0.01, -0.0895, -0.4093, 0.01, -0.096, -0.4093, 0.01, -0.0939, -0.4071, 0.01, -0.096, -0.4071, 0.01, -0.0939, -0.4093, 0.01, -0.0917, -0.4093, 0.01, -0.0895, -0.4071, 0.01, -0.0917, -0.4071, 0.01, -0.0895, -0.4049, 0.0061, -0.1043, -0.4049, 0.0083, -0.1043, -0.4028, 0.0061, -0.1048, -0.4028, 0.0083, -0.1048, -0.405, 0.01, -0.1043, -0.4049, 0.01, -0.1025, -0.4029, 0.01, -0.1048, -0.4028, 0.01, -0.1025, -0.4049, 0.01, -0.1004, -0.4049, 0.01, -0.0982, -0.4028, 0.01, -0.1004, -0.4028, 0.01, -0.0982, -0.4006, 0.0061, -0.1053, -0.4006, 0.0083, -0.1053,
-0.3988, 0.0061, -0.1056, -0.3988, 0.0083, -0.1056, -0.4008, 0.01, -0.1053, -0.4006, 0.01, -0.1025, -0.3985, 0.01, -0.1057, -0.3984, 0.01, -0.1025, -0.4006, 0.01, -0.1004, -0.4006, 0.01, -0.0982, -0.3984, 0.01, -0.1004, -0.3984, 0.01, -0.0982, -0.4024, 0.0061, -0.0887, -0.4024, 0.0083, -0.0887, -0.4049, 0.01, -0.096, -0.4049, 0.01, -0.0939, -0.4028, 0.01, -0.096, -0.4028, 0.01, -0.0939, -0.4049, 0.01, -0.0917, -0.4049, 0.01, -0.0895, -0.4028, 0.01, -0.0917, -0.4026, 0.01, -0.0886, -0.4006, 0.0061, -0.0893, -0.4006, 0.0083, -0.0893, -0.3984, 0.0061, -0.0899, -0.3984, 0.0083, -0.0899, -0.4006, 0.01, -0.096, -0.4006, 0.01, -0.0939, -0.3984, 0.01, -0.096, -0.3984, 0.01, -0.0939, -0.4006, 0.01, -0.0917, -0.4004, 0.01, -0.0894,
-0.3984, 0.01, -0.0917, -0.3984, 0.01, -0.09, -0.4129, 0.0061, -0.0843, -0.4129, 0.0083, -0.0843, -0.4115, 0.0061, -0.085, -0.4115, 0.0083, -0.085, -0.414, 0.0061, -0.0838, -0.414, 0.0083, -0.0838, -0.4136, 0.01, -0.0874, -0.4131, 0.01, -0.0842, -0.4115, 0.01, -0.0874, -0.4112, 0.01, -0.0851, -0.4141, 0.01, -0.0837, -0.4084, 0.0061, -0.0863, -0.4095, 0.0061, -0.0859, -0.4084, 0.0083, -0.0863, -0.4095, 0.0083, -0.0859, -0.4071, 0.0061, -0.0869, -0.4071, 0.0083, -0.0869, -0.4087, 0.01, -0.0862, -0.4097, 0.01, -0.0858, -0.4068, 0.01, -0.087, -0.4049, 0.0061, -0.0877, -0.4049, 0.0083, -0.0877, -0.4035, 0.0061, -0.0883, -0.4035, 0.0083, -0.0883, -0.4048, 0.01, -0.0878, -0.4036, 0.01, -0.0883, -0.3963, 0.01, -0.1047, -0.3963, 0.01, -0.1025,
-0.3941, 0.01, -0.1047, -0.3941, 0.01, -0.1025, -0.3963, 0.01, -0.1004, -0.3963, 0.01, -0.0982, -0.3941, 0.01, -0.1004, -0.3941, 0.01, -0.0982, -0.3919, 0.01, -0.1047, -0.3919, 0.01, -0.1025, -0.3897, 0.01, -0.1047, -0.3897, 0.01, -0.1025, -0.3919, 0.01, -0.1004, -0.3919, 0.01, -0.0982, -0.3897, 0.01, -0.1004, -0.3897, 0.01, -0.0982, -0.3954, 0.0061, -0.0907, -0.3965, 0.0061, -0.0904, -0.3954, 0.0083, -0.0907, -0.3965, 0.0083, -0.0904, -0.3941, 0.0061, -0.091, -0.3941, 0.0083, -0.091, -0.3963, 0.01, -0.096, -0.3963, 0.01, -0.0939, -0.3941, 0.01, -0.096, -0.3941, 0.01, -0.0939, -0.3959, 0.01, -0.0905, -0.3967, 0.01, -0.0903, -0.3939, 0.01, -0.091, -0.3919, 0.0061, -0.0915, -0.3919, 0.0083, -0.0915, -0.3898, 0.0061, -0.0921,
-0.3898, 0.0083, -0.0921, -0.3919, 0.01, -0.096, -0.3919, 0.01, -0.0939, -0.3897, 0.01, -0.096, -0.3897, 0.01, -0.0939, -0.3918, 0.01, -0.0916, -0.3897, 0.01, -0.0921, -0.3876, 0.01, -0.1047, -0.3876, 0.01, -0.1025, -0.3854, 0.01, -0.1047, -0.3854, 0.01, -0.1025, -0.3876, 0.01, -0.1004, -0.3876, 0.01, -0.0982, -0.3854, 0.01, -0.1004, -0.3854, 0.01, -0.0982, -0.3832, 0.01, -0.1047, -0.3832, 0.01, -0.1025, -0.3811, 0.01, -0.1047, -0.3811, 0.01, -0.1025, -0.3832, 0.01, -0.1004, -0.3832, 0.01, -0.0982, -0.3811, 0.01, -0.1004, -0.3811, 0.01, -0.0982, -0.3876, 0.0061, -0.0924, -0.3876, 0.0083, -0.0924, -0.3854, 0.0061, -0.0925, -0.3854, 0.0083, -0.0925, -0.3876, 0.01, -0.096, -0.3876, 0.01, -0.0939, -0.3854, 0.01, -0.096,
-0.3854, 0.01, -0.0939, -0.3876, 0.01, -0.0924, -0.3854, 0.01, -0.0925, -0.3827, 0.0061, -0.0928, -0.3827, 0.0083, -0.0928, -0.3811, 0.0061, -0.0931, -0.3811, 0.0083, -0.0931, -0.3838, 0.0061, -0.0927, -0.3838, 0.0083, -0.0927, -0.3832, 0.01, -0.096, -0.3831, 0.01, -0.0928, -0.3811, 0.01, -0.096, -0.381, 0.01, -0.0931, -0.3839, 0.01, -0.0927, -0.3789, 0.0061, -0.1072, -0.3789, 0.0083, -0.1072, -0.3767, 0.0061, -0.1072, -0.3767, 0.0083, -0.1072, -0.3789, 0.01, -0.1072, -0.3767, 0.01, -0.1072, -0.3746, 0.0061, -0.1072, -0.3746, 0.0083, -0.1072, -0.3724, 0.0061, -0.1072, -0.3724, 0.0083, -0.1072, -0.3746, 0.01, -0.1072, -0.3724, 0.01, -0.1072, -0.3702, 0.0061, -0.1072, -0.3702, 0.0083, -0.1072, -0.368, 0.0061, -0.1072, -0.368, 0.0083, -0.1072,
-0.3702, 0.01, -0.1072, -0.368, 0.01, -0.1071, -0.3659, 0.0061, -0.107, -0.3659, 0.0083, -0.107, -0.3637, 0.0061, -0.1067, -0.3637, 0.0083, -0.1067, -0.3658, 0.01, -0.107, -0.3636, 0.01, -0.1067, -0.3615, 0.0061, -0.1062, -0.3615, 0.0083, -0.1062, -0.36, 0.0061, -0.1059, -0.36, 0.0083, -0.1059, -0.3615, 0.01, -0.1062, -0.3601, 0.01, -0.1059, -0.3789, 0.01, -0.1047, -0.3789, 0.01, -0.1025, -0.3767, 0.01, -0.1047, -0.3767, 0.01, -0.1025, -0.3789, 0.01, -0.1004, -0.3789, 0.01, -0.0982, -0.3767, 0.01, -0.1004, -0.3767, 0.01, -0.0982, -0.3746, 0.01, -0.1047, -0.3746, 0.01, -0.1025, -0.3724, 0.01, -0.1047, -0.3724, 0.01, -0.1025, -0.3746, 0.01, -0.1004, -0.3746, 0.01, -0.0982, -0.3724, 0.01, -0.1004, -0.3724, 0.01, -0.0982,
-0.3789, 0.0061, -0.0933, -0.3789, 0.0083, -0.0933, -0.3767, 0.0061, -0.0934, -0.3767, 0.0083, -0.0934, -0.3789, 0.01, -0.096, -0.3788, 0.01, -0.0933, -0.3767, 0.01, -0.096, -0.3767, 0.01, -0.0934, -0.3746, 0.0061, -0.0934, -0.3746, 0.0083, -0.0934, -0.3724, 0.0061, -0.0934, -0.3724, 0.0083, -0.0934, -0.3746, 0.01, -0.096, -0.3746, 0.01, -0.0934, -0.3724, 0.01, -0.096, -0.3724, 0.01, -0.0934, -0.3702, 0.01, -0.1047, -0.3702, 0.01, -0.1025, -0.368, 0.01, -0.1047, -0.368, 0.01, -0.1025, -0.3702, 0.01, -0.1004, -0.3702, 0.01, -0.0982, -0.368, 0.01, -0.1004, -0.368, 0.01, -0.0982, -0.3659, 0.01, -0.1047, -0.3659, 0.01, -0.1025, -0.3637, 0.01, -0.1047, -0.3637, 0.01, -0.1025, -0.3659, 0.01, -0.1004, -0.3659, 0.01, -0.0982,
-0.3637, 0.01, -0.1004, -0.3637, 0.01, -0.0982, -0.3702, 0.0061, -0.0934, -0.3702, 0.0083, -0.0934, -0.368, 0.0061, -0.0931, -0.368, 0.0083, -0.0931, -0.3702, 0.01, -0.096, -0.3703, 0.01, -0.0934, -0.368, 0.01, -0.096, -0.3682, 0.01, -0.0931, -0.3664, 0.0061, -0.0929, -0.3664, 0.0083, -0.0929, -0.3653, 0.0061, -0.0927, -0.3653, 0.0083, -0.0927, -0.3637, 0.0061, -0.0925, -0.3637, 0.0083, -0.0925, -0.3659, 0.01, -0.096, -0.366, 0.01, -0.0928, -0.3637, 0.01, -0.096, -0.3637, 0.01, -0.0939, -0.3651, 0.01, -0.0927, -0.3637, 0.01, -0.0925, -0.3589, 0.0061, -0.1057, -0.3589, 0.0083, -0.1057, -0.3615, 0.01, -0.1047, -0.3615, 0.01, -0.1025, -0.3592, 0.01, -0.1057, -0.3594, 0.01, -0.1025, -0.3615, 0.01, -0.1004, -0.3615, 0.01, -0.0982,
-0.3594, 0.01, -0.1004, -0.3594, 0.01, -0.0982, -0.3572, 0.0061, -0.1053, -0.3572, 0.0083, -0.1053, -0.355, 0.0061, -0.1048, -0.355, 0.0083, -0.1048, -0.357, 0.01, -0.1052, -0.3572, 0.01, -0.1025, -0.3549, 0.01, -0.1048, -0.355, 0.01, -0.1025, -0.3572, 0.01, -0.1004, -0.3572, 0.01, -0.0982, -0.355, 0.01, -0.1004, -0.355, 0.01, -0.0982, -0.3615, 0.0061, -0.0924, -0.3615, 0.0083, -0.0924, -0.3594, 0.0061, -0.0922, -0.3594, 0.0083, -0.0922, -0.3615, 0.01, -0.096, -0.3615, 0.01, -0.0939, -0.3594, 0.01, -0.096, -0.3594, 0.01, -0.0939, -0.3616, 0.01, -0.0924, -0.3594, 0.01, -0.0922, -0.3572, 0.0061, -0.0918, -0.3572, 0.0083, -0.0918, -0.355, 0.0061, -0.0912, -0.355, 0.0083, -0.0912, -0.3572, 0.01, -0.096, -0.3572, 0.01, -0.0939,
-0.355, 0.01, -0.096, -0.355, 0.01, -0.0939, -0.3573, 0.01, -0.0918, -0.3552, 0.01, -0.0913, -0.3529, 0.0061, -0.1043, -0.3529, 0.0083, -0.1043, -0.3509, 0.0061, -0.1039, -0.3498, 0.0061, -0.1036, -0.3509, 0.0083, -0.1039, -0.3498, 0.0083, -0.1036, -0.3528, 0.01, -0.1043, -0.3529, 0.01, -0.1025, -0.3512, 0.01, -0.1039, -0.3503, 0.01, -0.1037, -0.3529, 0.01, -0.1004, -0.3529, 0.01, -0.0982, -0.3507, 0.01, -0.1004, -0.3507, 0.01, -0.0982, -0.3485, 0.0061, -0.1032, -0.3485, 0.0083, -0.1032, -0.3463, 0.0061, -0.1025, -0.3463, 0.0083, -0.1025, -0.3483, 0.01, -0.1031, -0.3462, 0.01, -0.1024, -0.3485, 0.01, -0.1004, -0.3485, 0.01, -0.0982, -0.3463, 0.01, -0.1004, -0.3463, 0.01, -0.0982, -0.3534, 0.0061, -0.0908, -0.3523, 0.0061, -0.0905,
-0.3534, 0.0083, -0.0908, -0.3523, 0.0083, -0.0905, -0.3507, 0.0061, -0.0901, -0.3507, 0.0083, -0.0901, -0.3529, 0.01, -0.096, -0.3529, 0.01, -0.0939, -0.3507, 0.01, -0.096, -0.3507, 0.01, -0.0939, -0.3531, 0.01, -0.0907, -0.3521, 0.01, -0.0904, -0.3507, 0.01, -0.0917, -0.3508, 0.01, -0.0901, -0.3485, 0.0061, -0.0895, -0.3485, 0.0083, -0.0895, -0.3465, 0.0061, -0.0888, -0.3465, 0.0083, -0.0888, -0.3485, 0.01, -0.096, -0.3485, 0.01, -0.0939, -0.3463, 0.01, -0.096, -0.3463, 0.01, -0.0939, -0.3485, 0.01, -0.0917, -0.3487, 0.01, -0.0895, -0.3463, 0.01, -0.0917, -0.3464, 0.01, -0.0888, -0.3454, 0.0061, -0.0884, -0.3454, 0.0083, -0.0884, -0.3454, 0.01, -0.0884, -0.4136, 0.0061, -0.0634, -0.4136, 0.0083, -0.0634, -0.4119, 0.0061, -0.0627,
-0.4119, 0.0083, -0.0627, -0.4134, 0.01, -0.0633, -0.412, 0.01, -0.0627, -0.4108, 0.0061, -0.0622, -0.4108, 0.0083, -0.0622, -0.4136, 0.01, -0.0613, -0.4136, 0.01, -0.0591, -0.411, 0.01, -0.0623, -0.4115, 0.01, -0.0591, -0.4136, 0.01, -0.057, -0.4136, 0.01, -0.0548, -0.4115, 0.01, -0.057, -0.4115, 0.01, -0.0548, -0.4093, 0.0061, -0.0615, -0.4093, 0.0083, -0.0615, -0.4074, 0.0061, -0.0606, -0.4063, 0.0061, -0.0601, -0.4074, 0.0083, -0.0606, -0.4063, 0.0083, -0.0601, -0.409, 0.01, -0.0614, -0.4093, 0.01, -0.0591, -0.4076, 0.01, -0.0607, -0.4065, 0.01, -0.0602, -0.4093, 0.01, -0.057, -0.4093, 0.01, -0.0548, -0.4071, 0.01, -0.057, -0.4071, 0.01, -0.0548, -0.4049, 0.0061, -0.0594, -0.4049, 0.0083, -0.0594, -0.4032, 0.0061, -0.0584,
-0.4032, 0.0083, -0.0584, -0.4021, 0.0061, -0.0578, -0.4021, 0.0083, -0.0578, -0.4047, 0.01, -0.0592, -0.4033, 0.01, -0.0585, -0.4049, 0.01, -0.057, -0.4049, 0.01, -0.0548, -0.4023, 0.01, -0.0579, -0.4028, 0.01, -0.0548, -0.4006, 0.0061, -0.0569, -0.4006, 0.0083, -0.0569, -0.3991, 0.0061, -0.0561, -0.398, 0.0061, -0.0555, -0.3991, 0.0083, -0.0561, -0.398, 0.0083, -0.0555, -0.4004, 0.01, -0.0568, -0.4006, 0.01, -0.0548, -0.3992, 0.01, -0.0561, -0.3981, 0.01, -0.0555, -0.4136, 0.01, -0.0526, -0.4136, 0.01, -0.0505, -0.4115, 0.01, -0.0526, -0.4115, 0.01, -0.0505, -0.4136, 0.01, -0.0483, -0.4136, 0.01, -0.0461, -0.4115, 0.01, -0.0483, -0.4115, 0.01, -0.0461, -0.4093, 0.01, -0.0526, -0.4093, 0.01, -0.0505, -0.4071, 0.01, -0.0526,
-0.4071, 0.01, -0.0505, -0.4093, 0.01, -0.0483, -0.4093, 0.01, -0.0461, -0.4071, 0.01, -0.0483, -0.4071, 0.01, -0.0461, -0.4136, 0.0061, -0.0397, -0.4136, 0.0083, -0.0397, -0.4119, 0.0061, -0.0389, -0.4108, 0.0061, -0.0383, -0.4119, 0.0083, -0.0389, -0.4108, 0.0083, -0.0383, -0.4136, 0.01, -0.0439, -0.4136, 0.01, -0.0418, -0.4115, 0.01, -0.0439, -0.4115, 0.01, -0.0418, -0.4138, 0.01, -0.0398, -0.4118, 0.01, -0.0388, -0.4107, 0.01, -0.0383, -0.4093, 0.0061, -0.0375, -0.4093, 0.0083, -0.0375, -0.4077, 0.0061, -0.0366, -0.4077, 0.0083, -0.0366, -0.4093, 0.01, -0.0439, -0.4093, 0.01, -0.0418, -0.4071, 0.01, -0.0439, -0.4071, 0.01, -0.0418, -0.4093, 0.01, -0.0396, -0.4095, 0.01, -0.0376, -0.4071, 0.01, -0.0396, -0.4076, 0.01, -0.0366,
-0.4049, 0.01, -0.0526, -0.4049, 0.01, -0.0505, -0.4028, 0.01, -0.0526, -0.4028, 0.01, -0.0505, -0.4049, 0.01, -0.0483, -0.4049, 0.01, -0.0461, -0.4028, 0.01, -0.0483, -0.4028, 0.01, -0.0461, -0.4006, 0.01, -0.0526, -0.4006, 0.01, -0.0505, -0.3984, 0.01, -0.0526, -0.3984, 0.01, -0.0505, -0.4006, 0.01, -0.0483, -0.4006, 0.01, -0.0461, -0.3984, 0.01, -0.0483, -0.3984, 0.01, -0.0461, -0.4049, 0.01, -0.0439, -0.4049, 0.01, -0.0418, -0.4028, 0.01, -0.0439, -0.4028, 0.01, -0.0418, -0.4049, 0.01, -0.0396, -0.4049, 0.01, -0.0374, -0.4028, 0.01, -0.0396, -0.4028, 0.01, -0.0374, -0.4006, 0.01, -0.0439, -0.4006, 0.01, -0.0418, -0.3984, 0.01, -0.0439, -0.3984, 0.01, -0.0418, -0.4006, 0.01, -0.0396, -0.4006, 0.01, -0.0374,
-0.3984, 0.01, -0.0396, -0.3984, 0.01, -0.0374, -0.3963, 0.0061, -0.0545, -0.3963, 0.0083, -0.0545, -0.395, 0.0061, -0.0538, -0.395, 0.0083, -0.0538, -0.3961, 0.01, -0.0544, -0.3951, 0.01, -0.0538, -0.394, 0.0061, -0.0532, -0.394, 0.0083, -0.0532, -0.3963, 0.01, -0.0526, -0.3963, 0.01, -0.0505, -0.3939, 0.01, -0.0531, -0.3941, 0.01, -0.0505, -0.3963, 0.01, -0.0483, -0.3963, 0.01, -0.0461, -0.3941, 0.01, -0.0483, -0.3941, 0.01, -0.0461, -0.3921, 0.0061, -0.0521, -0.3911, 0.0061, -0.0514, -0.3921, 0.0083, -0.0521, -0.3911, 0.0083, -0.0514, -0.3897, 0.0061, -0.0506, -0.3897, 0.0083, -0.0506, -0.3923, 0.01, -0.0521, -0.3912, 0.01, -0.0515, -0.3895, 0.01, -0.0504, -0.3919, 0.01, -0.0483, -0.3919, 0.01, -0.0461, -0.3897, 0.01, -0.0483,
-0.3897, 0.01, -0.0461, -0.3963, 0.01, -0.0439, -0.3963, 0.01, -0.0418, -0.3941, 0.01, -0.0439, -0.3941, 0.01, -0.0418, -0.3963, 0.01, -0.0396, -0.3963, 0.01, -0.0374, -0.3941, 0.01, -0.0396, -0.3941, 0.01, -0.0374, -0.3919, 0.01, -0.0439, -0.3919, 0.01, -0.0418, -0.3897, 0.01, -0.0439, -0.3897, 0.01, -0.0418, -0.3919, 0.01, -0.0396, -0.3919, 0.01, -0.0374, -0.3897, 0.01, -0.0396, -0.3897, 0.01, -0.0374, -0.3883, 0.0061, -0.0496, -0.3883, 0.0083, -0.0496, -0.3872, 0.0061, -0.0489, -0.3872, 0.0083, -0.0489, -0.3856, 0.0061, -0.0478, -0.3845, 0.0061, -0.0471, -0.3856, 0.0083, -0.0478, -0.3845, 0.0083, -0.0471, -0.3883, 0.01, -0.0497, -0.3872, 0.01, -0.049, -0.3876, 0.01, -0.0461, -0.3857, 0.01, -0.0479, -0.3846, 0.01, -0.0471,
-0.3832, 0.0061, -0.0461, -0.3832, 0.0083, -0.0461, -0.382, 0.0061, -0.0451, -0.382, 0.0083, -0.0451, -0.383, 0.01, -0.0459, -0.382, 0.01, -0.0451, -0.3876, 0.01, -0.0439, -0.3876, 0.01, -0.0418, -0.3854, 0.01, -0.0439, -0.3854, 0.01, -0.0418, -0.3876, 0.01, -0.0396, -0.3876, 0.01, -0.0374, -0.3854, 0.01, -0.0396, -0.3854, 0.01, -0.0374, -0.3809, 0.0061, -0.0443, -0.3809, 0.0083, -0.0443, -0.3832, 0.01, -0.0439, -0.3832, 0.01, -0.0418, -0.3809, 0.01, -0.0443, -0.3811, 0.01, -0.0418, -0.3832, 0.01, -0.0396, -0.3832, 0.01, -0.0374, -0.3811, 0.01, -0.0396, -0.3811, 0.01, -0.0374, -0.4136, 0.0061, -0.0309, -0.4136, 0.0083, -0.0309, -0.4121, 0.0061, -0.0301, -0.411, 0.0061, -0.0295, -0.4121, 0.0083, -0.0301, -0.411, 0.0083, -0.0295,
-0.4134, 0.01, -0.0308, -0.4136, 0.01, -0.0288, -0.4122, 0.01, -0.0301, -0.4111, 0.01, -0.0295, -0.4066, 0.0061, -0.036, -0.4066, 0.0083, -0.036, -0.4093, 0.0061, -0.0284, -0.4093, 0.0083, -0.0284, -0.4081, 0.0061, -0.0277, -0.4081, 0.0083, -0.0277, -0.4065, 0.01, -0.036, -0.4091, 0.01, -0.0283, -0.4081, 0.01, -0.0277, -0.4136, 0.01, -0.0266, -0.4136, 0.01, -0.0244, -0.4115, 0.01, -0.0266, -0.4115, 0.01, -0.0244, -0.4136, 0.01, -0.0222, -0.4136, 0.01, -0.0201, -0.4115, 0.01, -0.0222, -0.4115, 0.01, -0.0201, -0.407, 0.0061, -0.0271, -0.407, 0.0083, -0.0271, -0.4093, 0.01, -0.0266, -0.4093, 0.01, -0.0244, -0.407, 0.01, -0.0271, -0.4071, 0.01, -0.0244, -0.4093, 0.01, -0.0222, -0.4093, 0.01, -0.0201, -0.4071, 0.01, -0.0222,
-0.4071, 0.01, -0.0201, -0.4049, 0.0061, -0.0351, -0.4049, 0.0083, -0.0351, -0.4037, 0.0061, -0.0343, -0.4026, 0.0061, -0.0336, -0.4037, 0.0083, -0.0343, -0.4026, 0.0083, -0.0336, -0.4052, 0.01, -0.0352, -0.4035, 0.01, -0.0342, -0.4025, 0.01, -0.0335, -0.4009, 0.0061, -0.0325, -0.4009, 0.0083, -0.0325, -0.3998, 0.0061, -0.0319, -0.3998, 0.0083, -0.0319, -0.3984, 0.0061, -0.031, -0.3984, 0.0083, -0.031, -0.4006, 0.01, -0.0353, -0.4008, 0.01, -0.0325, -0.3984, 0.01, -0.0353, -0.3984, 0.01, -0.0331, -0.3997, 0.01, -0.0318, -0.3987, 0.01, -0.0312, -0.4053, 0.0061, -0.026, -0.4042, 0.0061, -0.0253, -0.4053, 0.0083, -0.026, -0.4042, 0.0083, -0.0253, -0.4028, 0.0061, -0.0244, -0.4028, 0.0083, -0.0244, -0.4054, 0.01, -0.0261, -0.4043, 0.01, -0.0254,
-0.4025, 0.01, -0.0242, -0.4049, 0.01, -0.0222, -0.4049, 0.01, -0.0201, -0.4028, 0.01, -0.0222, -0.4028, 0.01, -0.0201, -0.4014, 0.0061, -0.0235, -0.4014, 0.0083, -0.0235, -0.4003, 0.0061, -0.0228, -0.4003, 0.0083, -0.0228, -0.3987, 0.0061, -0.0217, -0.3976, 0.0061, -0.021, -0.3987, 0.0083, -0.0217, -0.3976, 0.0083, -0.021, -0.4015, 0.01, -0.0235, -0.4004, 0.01, -0.0228, -0.4006, 0.01, -0.0201, -0.3988, 0.01, -0.0218, -0.3978, 0.01, -0.0211, -0.4145, 0.0061, -0.0104, -0.4145, 0.0083, -0.0104, -0.4136, 0.01, -0.0179, -0.4136, 0.01, -0.0157, -0.4115, 0.01, -0.0179, -0.4115, 0.01, -0.0157, -0.4136, 0.01, -0.0136, -0.4143, 0.01, -0.0103, -0.4115, 0.01, -0.0136, -0.4115, 0.01, -0.0114, -0.4093, 0.01, -0.0179, -0.4093, 0.01, -0.0157,
-0.4071, 0.01, -0.0179, -0.4071, 0.01, -0.0157, -0.4093, 0.01, -0.0136, -0.4093, 0.01, -0.0114, -0.4071, 0.01, -0.0136, -0.4071, 0.01, -0.0114, -0.4134, 0.0061, -0.0098, -0.4134, 0.0083, -0.0098, -0.4115, 0.0061, -0.0087, -0.4115, 0.0083, -0.0087, -0.4141, 0.0061, -0.0019, -0.4141, 0.0083, -0.0019, -0.4133, 0.01, -0.0097, -0.4118, 0.01, -0.0089, -0.4142, 0.01, -0.002, -0.4103, 0.0061, -0.0081, -0.4093, 0.0061, -0.0076, -0.4103, 0.0083, -0.0081, -0.4093, 0.0083, -0.0076, -0.4073, 0.0061, -0.0065, -0.4073, 0.0083, -0.0065, -0.4062, 0.0061, -0.0058, -0.4062, 0.0083, -0.0058, -0.41, 0.01, -0.008, -0.4091, 0.01, -0.0075, -0.4071, 0.01, -0.0092, -0.4073, 0.01, -0.0065, -0.4062, 0.01, -0.0058, -0.4049, 0.01, -0.0179, -0.4049, 0.01, -0.0157,
-0.4028, 0.01, -0.0179, -0.4028, 0.01, -0.0157, -0.4049, 0.01, -0.0136, -0.4049, 0.01, -0.0114, -0.4028, 0.01, -0.0136, -0.4028, 0.01, -0.0114, -0.4006, 0.01, -0.0179, -0.4006, 0.01, -0.0157, -0.3984, 0.01, -0.0179, -0.3984, 0.01, -0.0157, -0.4006, 0.01, -0.0136, -0.4006, 0.01, -0.0114, -0.3984, 0.01, -0.0136, -0.3984, 0.01, -0.0114, -0.4049, 0.0061, -0.0051, -0.4049, 0.0083, -0.0051, -0.4032, 0.0061, -0.0041, -0.4022, 0.0061, -0.0035, -0.4032, 0.0083, -0.0041, -0.4022, 0.0083, -0.0035, -0.4049, 0.01, -0.0092, -0.4049, 0.01, -0.0071, -0.4028, 0.01, -0.0092, -0.4028, 0.01, -0.0071, -0.4051, 0.01, -0.0052, -0.4032, 0.01, -0.0041, -0.4021, 0.01, -0.0035, -0.4006, 0.0061, -0.0026, -0.4006, 0.0083, -0.0026, -0.3992, 0.0061, -0.0018,
-0.3992, 0.0083, -0.0018, -0.4006, 0.01, -0.0092, -0.4006, 0.01, -0.0071, -0.3984, 0.01, -0.0092, -0.3984, 0.01, -0.0071, -0.4006, 0.01, -0.0049, -0.4009, 0.01, -0.0028, -0.3984, 0.01, -0.0049, -0.399, 0.01, -0.0017, -0.397, 0.0061, -0.0301, -0.3959, 0.0061, -0.0294, -0.397, 0.0083, -0.0301, -0.3959, 0.0083, -0.0294, -0.3941, 0.0061, -0.0283, -0.3941, 0.0083, -0.0283, -0.3963, 0.01, -0.0353, -0.3963, 0.01, -0.0331, -0.3941, 0.01, -0.0353, -0.3941, 0.01, -0.0331, -0.3968, 0.01, -0.03, -0.3958, 0.01, -0.0293, -0.3941, 0.01, -0.0309, -0.3942, 0.01, -0.0284, -0.3919, 0.01, -0.0353, -0.3919, 0.01, -0.0331, -0.3897, 0.01, -0.0353, -0.3897, 0.01, -0.0331, -0.3919, 0.01, -0.0309, -0.3919, 0.01, -0.0288, -0.3897, 0.01, -0.0309,
-0.3897, 0.01, -0.0288, -0.393, 0.0061, -0.0277, -0.393, 0.0083, -0.0277, -0.3963, 0.0061, -0.0201, -0.3963, 0.0083, -0.0201, -0.3949, 0.0061, -0.0192, -0.3949, 0.0083, -0.0192, -0.393, 0.01, -0.0276, -0.396, 0.01, -0.0199, -0.395, 0.01, -0.0192, -0.3919, 0.0061, -0.0269, -0.3919, 0.0083, -0.0269, -0.3903, 0.0061, -0.0259, -0.3892, 0.0061, -0.0251, -0.3903, 0.0083, -0.0259, -0.3892, 0.0083, -0.0251, -0.3921, 0.01, -0.027, -0.3902, 0.01, -0.0258, -0.3891, 0.01, -0.0251, -0.3876, 0.01, -0.0353, -0.3876, 0.01, -0.0331, -0.3854, 0.01, -0.0353, -0.3854, 0.01, -0.0331, -0.3876, 0.01, -0.0309, -0.3876, 0.01, -0.0288, -0.3854, 0.01, -0.0309, -0.3854, 0.01, -0.0288, -0.3832, 0.01, -0.0353, -0.3832, 0.01, -0.0331, -0.3811, 0.01, -0.0353,
-0.3811, 0.01, -0.0331, -0.3832, 0.01, -0.0309, -0.3832, 0.01, -0.0288, -0.3811, 0.01, -0.0309, -0.3811, 0.01, -0.0288, -0.3876, 0.0061, -0.0241, -0.3876, 0.0083, -0.0241, -0.3865, 0.0061, -0.0233, -0.3865, 0.0083, -0.0233, -0.3854, 0.0061, -0.0226, -0.3854, 0.0083, -0.0226, -0.3876, 0.01, -0.0266, -0.3879, 0.01, -0.0243, -0.3854, 0.01, -0.0266, -0.3863, 0.01, -0.0232, -0.3853, 0.01, -0.0225, -0.3838, 0.0061, -0.0215, -0.3827, 0.0061, -0.0207, -0.3838, 0.0083, -0.0215, -0.3827, 0.0083, -0.0207, -0.3812, 0.0061, -0.0197, -0.3812, 0.0083, -0.0197, -0.3832, 0.01, -0.0266, -0.3832, 0.01, -0.0244, -0.3811, 0.01, -0.0266, -0.3811, 0.01, -0.0244, -0.3837, 0.01, -0.0215, -0.3827, 0.01, -0.0207, -0.3811, 0.01, -0.0222, -0.3812, 0.01, -0.0197,
-0.3938, 0.0061, -0.0184, -0.3938, 0.0083, -0.0184, -0.3963, 0.01, -0.0179, -0.3963, 0.01, -0.0157, -0.3938, 0.01, -0.0184, -0.3941, 0.01, -0.0157, -0.3963, 0.01, -0.0136, -0.3963, 0.01, -0.0114, -0.3941, 0.01, -0.0136, -0.3941, 0.01, -0.0114, -0.3922, 0.0061, -0.0174, -0.3911, 0.0061, -0.0166, -0.3922, 0.0083, -0.0174, -0.3911, 0.0083, -0.0166, -0.3897, 0.0061, -0.0157, -0.3897, 0.0083, -0.0157, -0.3923, 0.01, -0.0174, -0.3913, 0.01, -0.0167, -0.3895, 0.01, -0.0156, -0.3919, 0.01, -0.0136, -0.3919, 0.01, -0.0114, -0.3897, 0.01, -0.0136, -0.3897, 0.01, -0.0114, -0.3963, 0.01, -0.0092, -0.3963, 0.01, -0.0071, -0.3941, 0.01, -0.0092, -0.3941, 0.01, -0.0071, -0.3963, 0.01, -0.0049, -0.3963, 0.01, -0.0027, -0.3941, 0.01, -0.0049,
-0.3941, 0.01, -0.0027, -0.3919, 0.01, -0.0092, -0.3919, 0.01, -0.0071, -0.3897, 0.01, -0.0092, -0.3897, 0.01, -0.0071, -0.3919, 0.01, -0.0049, -0.3919, 0.01, -0.0027, -0.3897, 0.01, -0.0049, -0.3897, 0.01, -0.0027, -0.3884, 0.0061, -0.0148, -0.3884, 0.0083, -0.0148, -0.3873, 0.0061, -0.0141, -0.3873, 0.0083, -0.0141, -0.3858, 0.0061, -0.013, -0.3847, 0.0061, -0.0122, -0.3858, 0.0083, -0.013, -0.3847, 0.0083, -0.0122, -0.3884, 0.01, -0.0148, -0.3873, 0.01, -0.0141, -0.3876, 0.01, -0.0114, -0.3859, 0.01, -0.013, -0.3848, 0.01, -0.0123, -0.3801, 0.0061, -0.0189, -0.3801, 0.0083, -0.0189, -0.3832, 0.0061, -0.0112, -0.3832, 0.0083, -0.0112, -0.3821, 0.0061, -0.0104, -0.3821, 0.0083, -0.0104, -0.3801, 0.01, -0.0189, -0.383, 0.01, -0.011,
-0.3821, 0.01, -0.0104, -0.3876, 0.01, -0.0092, -0.3876, 0.01, -0.0071, -0.3854, 0.01, -0.0092, -0.3854, 0.01, -0.0071, -0.3876, 0.01, -0.0049, -0.3876, 0.01, -0.0027, -0.3854, 0.01, -0.0049, -0.3854, 0.01, -0.0027, -0.381, 0.0061, -0.0096, -0.381, 0.0083, -0.0096, -0.3832, 0.01, -0.0092, -0.3832, 0.01, -0.0071, -0.381, 0.01, -0.0096, -0.3811, 0.01, -0.0071, -0.3832, 0.01, -0.0049, -0.3832, 0.01, -0.0027, -0.3811, 0.01, -0.0049, -0.3811, 0.01, -0.0027, -0.3796, 0.0061, -0.0432, -0.3785, 0.0061, -0.0423, -0.3796, 0.0083, -0.0432, -0.3785, 0.0083, -0.0423, -0.3771, 0.0061, -0.0412, -0.3771, 0.0083, -0.0412, -0.376, 0.0061, -0.0404, -0.376, 0.0083, -0.0404, -0.3796, 0.01, -0.0432, -0.3785, 0.01, -0.0423, -0.3772, 0.01, -0.0413,
-0.3789, 0.01, -0.0396, -0.3789, 0.01, -0.0374, -0.3761, 0.01, -0.0404, -0.3767, 0.01, -0.0374, -0.3746, 0.0061, -0.0393, -0.3736, 0.0061, -0.0385, -0.3746, 0.0083, -0.0393, -0.3736, 0.0083, -0.0385, -0.3724, 0.0061, -0.0376, -0.3724, 0.0083, -0.0376, -0.3747, 0.01, -0.0394, -0.3737, 0.01, -0.0385, -0.3721, 0.01, -0.0373, -0.3711, 0.0061, -0.0365, -0.3711, 0.0083, -0.0365, -0.3711, 0.01, -0.0365, -0.3789, 0.01, -0.0353, -0.3789, 0.01, -0.0331, -0.3767, 0.01, -0.0353, -0.3767, 0.01, -0.0331, -0.3789, 0.01, -0.0309, -0.3789, 0.01, -0.0288, -0.3767, 0.01, -0.0309, -0.3767, 0.01, -0.0288, -0.3746, 0.01, -0.0353, -0.3746, 0.01, -0.0331, -0.3724, 0.01, -0.0353, -0.3724, 0.01, -0.0331, -0.3746, 0.01, -0.0309, -0.3746, 0.01, -0.0288,
-0.3724, 0.01, -0.0309, -0.3724, 0.01, -0.0288, -0.3789, 0.01, -0.0266, -0.3789, 0.01, -0.0244, -0.3767, 0.01, -0.0266, -0.3767, 0.01, -0.0244, -0.3789, 0.01, -0.0222, -0.3789, 0.01, -0.0201, -0.3767, 0.01, -0.0222, -0.3767, 0.01, -0.0201, -0.3746, 0.01, -0.0266, -0.3746, 0.01, -0.0244, -0.3724, 0.01, -0.0266, -0.3724, 0.01, -0.0244, -0.3746, 0.01, -0.0222, -0.3746, 0.01, -0.0201, -0.3724, 0.01, -0.0222, -0.3724, 0.01, -0.0201, -0.37, 0.0061, -0.0357, -0.37, 0.0083, -0.0357, -0.3686, 0.0061, -0.0346, -0.3675, 0.0061, -0.0337, -0.3686, 0.0083, -0.0346, -0.3675, 0.0083, -0.0337, -0.37, 0.01, -0.0357, -0.3702, 0.01, -0.0331, -0.3686, 0.01, -0.0346, -0.3676, 0.01, -0.0338, -0.3702, 0.01, -0.0309, -0.3702, 0.01, -0.0288,
-0.368, 0.01, -0.0309, -0.368, 0.01, -0.0288, -0.3662, 0.0061, -0.0327, -0.3662, 0.0083, -0.0327, -0.3651, 0.0061, -0.0318, -0.3651, 0.0083, -0.0318, -0.3637, 0.0061, -0.0307, -0.3637, 0.0083, -0.0307, -0.3662, 0.01, -0.0327, -0.3651, 0.01, -0.0318, -0.3659, 0.01, -0.0288, -0.3635, 0.01, -0.0305, -0.3637, 0.01, -0.0288, -0.3702, 0.01, -0.0266, -0.3702, 0.01, -0.0244, -0.368, 0.01, -0.0266, -0.368, 0.01, -0.0244, -0.3702, 0.01, -0.0222, -0.3702, 0.01, -0.0201, -0.368, 0.01, -0.0222, -0.368, 0.01, -0.0201, -0.3659, 0.01, -0.0266, -0.3659, 0.01, -0.0244, -0.3637, 0.01, -0.0266, -0.3637, 0.01, -0.0244, -0.3659, 0.01, -0.0222, -0.3659, 0.01, -0.0201, -0.3637, 0.01, -0.0222, -0.3637, 0.01, -0.0201, -0.3789, 0.0061, -0.018,
-0.3789, 0.0083, -0.018, -0.3775, 0.0061, -0.017, -0.3764, 0.0061, -0.0163, -0.3775, 0.0083, -0.017, -0.3764, 0.0083, -0.0163, -0.3791, 0.01, -0.0182, -0.3774, 0.01, -0.017, -0.3763, 0.01, -0.0162, -0.3749, 0.0061, -0.0152, -0.3749, 0.0083, -0.0152, -0.3738, 0.0061, -0.0144, -0.3738, 0.0083, -0.0144, -0.3724, 0.0061, -0.0134, -0.3724, 0.0083, -0.0134, -0.3746, 0.01, -0.0179, -0.3749, 0.01, -0.0152, -0.3724, 0.01, -0.0179, -0.3724, 0.01, -0.0157, -0.3738, 0.01, -0.0144, -0.3727, 0.01, -0.0136, -0.3794, 0.0061, -0.0085, -0.3783, 0.0061, -0.0078, -0.3794, 0.0083, -0.0085, -0.3783, 0.0083, -0.0078, -0.3768, 0.0061, -0.0067, -0.3768, 0.0083, -0.0067, -0.3757, 0.0061, -0.0059, -0.3757, 0.0083, -0.0059, -0.3795, 0.01, -0.0086, -0.3784, 0.01, -0.0078,
-0.3769, 0.01, -0.0068, -0.3789, 0.01, -0.0049, -0.3789, 0.01, -0.0027, -0.3759, 0.01, -0.006, -0.3767, 0.01, -0.0027, -0.3746, 0.0061, -0.0051, -0.3746, 0.0083, -0.0051, -0.3731, 0.0061, -0.0041, -0.372, 0.0061, -0.0033, -0.3731, 0.0083, -0.0041, -0.372, 0.0083, -0.0033, -0.3742, 0.01, -0.0049, -0.3746, 0.01, -0.0027, -0.3731, 0.01, -0.0041, -0.372, 0.01, -0.0033, -0.3712, 0.0061, -0.0125, -0.3701, 0.0061, -0.0118, -0.3712, 0.0083, -0.0125, -0.3701, 0.0083, -0.0118, -0.3686, 0.0061, -0.0107, -0.3686, 0.0083, -0.0107, -0.3702, 0.01, -0.0179, -0.3702, 0.01, -0.0157, -0.368, 0.01, -0.0179, -0.368, 0.01, -0.0157, -0.3711, 0.01, -0.0124, -0.37, 0.01, -0.0117, -0.368, 0.01, -0.0136, -0.3685, 0.01, -0.0106, -0.3659, 0.01, -0.0179,
-0.3659, 0.01, -0.0157, -0.3637, 0.01, -0.0179, -0.3637, 0.01, -0.0157, -0.3659, 0.01, -0.0136, -0.3659, 0.01, -0.0114, -0.3637, 0.01, -0.0136, -0.3637, 0.01, -0.0114, -0.3675, 0.0061, -0.0099, -0.3675, 0.0083, -0.0099, -0.3704, 0.0061, -0.0022, -0.3704, 0.0083, -0.0022, -0.3675, 0.01, -0.0099, -0.3706, 0.01, -0.0023, -0.366, 0.0061, -0.0088, -0.3649, 0.0061, -0.0081, -0.366, 0.0083, -0.0088, -0.3649, 0.0083, -0.0081, -0.3637, 0.0061, -0.0072, -0.3637, 0.0083, -0.0072, -0.366, 0.01, -0.0088, -0.3649, 0.01, -0.008, -0.3637, 0.01, -0.0092, -0.3639, 0.01, -0.0074, -0.3626, 0.0061, -0.0298, -0.3615, 0.0061, -0.029, -0.3626, 0.0083, -0.0298, -0.3615, 0.0083, -0.029, -0.3601, 0.0061, -0.0279, -0.3601, 0.0083, -0.0279, -0.3626, 0.01, -0.0298,
-0.3615, 0.01, -0.029, -0.3602, 0.01, -0.0279, -0.3591, 0.0061, -0.0271, -0.3591, 0.0083, -0.0271, -0.3615, 0.01, -0.0266, -0.3615, 0.01, -0.0244, -0.3591, 0.01, -0.0271, -0.3594, 0.01, -0.0244, -0.3615, 0.01, -0.0222, -0.3615, 0.01, -0.0201, -0.3594, 0.01, -0.0222, -0.3594, 0.01, -0.0201, -0.3577, 0.0061, -0.026, -0.3566, 0.0061, -0.0251, -0.3577, 0.0083, -0.026, -0.3566, 0.0083, -0.0251, -0.3553, 0.0061, -0.024, -0.3553, 0.0083, -0.024, -0.3542, 0.0061, -0.0231, -0.3542, 0.0083, -0.0231, -0.3577, 0.01, -0.026, -0.3566, 0.01, -0.0252, -0.3553, 0.01, -0.0241, -0.3572, 0.01, -0.0222, -0.3572, 0.01, -0.0201, -0.3543, 0.01, -0.0231, -0.355, 0.01, -0.0201, -0.3529, 0.0061, -0.022, -0.3519, 0.0061, -0.0211, -0.3529, 0.0083, -0.022,
-0.3519, 0.0083, -0.0211, -0.3507, 0.0061, -0.0201, -0.3507, 0.0083, -0.0201, -0.353, 0.01, -0.0221, -0.3519, 0.01, -0.0212, -0.3504, 0.01, -0.0199, -0.3494, 0.0061, -0.0191, -0.3494, 0.0083, -0.0191, -0.3495, 0.01, -0.0191, -0.3615, 0.01, -0.0179, -0.3615, 0.01, -0.0157, -0.3594, 0.01, -0.0179, -0.3594, 0.01, -0.0157, -0.3615, 0.01, -0.0136, -0.3615, 0.01, -0.0114, -0.3594, 0.01, -0.0136, -0.3594, 0.01, -0.0114, -0.3572, 0.01, -0.0179, -0.3572, 0.01, -0.0157, -0.355, 0.01, -0.0179, -0.355, 0.01, -0.0157, -0.3572, 0.01, -0.0136, -0.3572, 0.01, -0.0114, -0.355, 0.01, -0.0136, -0.355, 0.01, -0.0114, -0.3623, 0.0061, -0.0062, -0.3623, 0.0083, -0.0062, -0.3612, 0.0061, -0.0055, -0.3612, 0.0083, -0.0055, -0.3596, 0.0061, -0.0044,
-0.3585, 0.0061, -0.0036, -0.3596, 0.0083, -0.0044, -0.3585, 0.0083, -0.0036, -0.3615, 0.01, -0.0092, -0.3622, 0.01, -0.0061, -0.3594, 0.01, -0.0092, -0.3594, 0.01, -0.0071, -0.3611, 0.01, -0.0054, -0.3596, 0.01, -0.0044, -0.3585, 0.01, -0.0036, -0.3572, 0.0061, -0.0027, -0.3572, 0.0083, -0.0027, -0.3559, 0.0061, -0.0018, -0.3559, 0.0083, -0.0018, -0.3572, 0.01, -0.0092, -0.3572, 0.01, -0.0071, -0.355, 0.01, -0.0092, -0.355, 0.01, -0.0071, -0.3572, 0.01, -0.0049, -0.3574, 0.01, -0.0029, -0.355, 0.01, -0.0049, -0.3557, 0.01, -0.0017, -0.3529, 0.01, -0.0179, -0.3529, 0.01, -0.0157, -0.3507, 0.01, -0.0179, -0.3507, 0.01, -0.0157, -0.3529, 0.01, -0.0136, -0.3529, 0.01, -0.0114, -0.3507, 0.01, -0.0136, -0.3507, 0.01, -0.0114,
-0.3484, 0.0061, -0.0182, -0.3484, 0.0083, -0.0182, -0.347, 0.0061, -0.0172, -0.3459, 0.0061, -0.0163, -0.347, 0.0083, -0.0172, -0.3459, 0.0083, -0.0163, -0.3483, 0.01, -0.0182, -0.3485, 0.01, -0.0157, -0.347, 0.01, -0.0172, -0.3459, 0.01, -0.0163, -0.3485, 0.01, -0.0136, -0.3485, 0.01, -0.0114, -0.3463, 0.01, -0.0136, -0.3463, 0.01, -0.0114, -0.3529, 0.01, -0.0092, -0.3529, 0.01, -0.0071, -0.3507, 0.01, -0.0092, -0.3507, 0.01, -0.0071, -0.3529, 0.01, -0.0049, -0.3529, 0.01, -0.0027, -0.3507, 0.01, -0.0049, -0.3507, 0.01, -0.0027, -0.3485, 0.01, -0.0092, -0.3485, 0.01, -0.0071, -0.3463, 0.01, -0.0092, -0.3463, 0.01, -0.0071, -0.3485, 0.01, -0.0049, -0.3485, 0.01, -0.0027, -0.3463, 0.01, -0.0049, -0.3463, 0.01, -0.0027,
-0.3444, 0.0061, -0.1018, -0.3444, 0.0083, -0.1018, -0.3433, 0.0061, -0.1014, -0.3433, 0.0083, -0.1014, -0.342, 0.0061, -0.1008, -0.342, 0.0083, -0.1008, -0.3446, 0.01, -0.1019, -0.3437, 0.01, -0.1015, -0.3442, 0.01, -0.0982, -0.3417, 0.01, -0.1007, -0.342, 0.01, -0.0982, -0.3398, 0.0061, -0.1, -0.3398, 0.0083, -0.1, -0.3385, 0.0061, -0.0994, -0.3374, 0.0061, -0.099, -0.3385, 0.0083, -0.0994, -0.3374, 0.0083, -0.099, -0.3397, 0.01, -0.0999, -0.3398, 0.01, -0.0982, -0.3385, 0.01, -0.0994, -0.3375, 0.01, -0.099, -0.3442, 0.01, -0.096, -0.3442, 0.01, -0.0939, -0.342, 0.01, -0.096, -0.342, 0.01, -0.0939, -0.3442, 0.01, -0.0917, -0.3442, 0.01, -0.0895, -0.342, 0.01, -0.0917, -0.342, 0.01, -0.0895, -0.3398, 0.01, -0.096,
-0.3398, 0.01, -0.0939, -0.3377, 0.01, -0.096, -0.3377, 0.01, -0.0939, -0.3398, 0.01, -0.0917, -0.3398, 0.01, -0.0895, -0.3377, 0.01, -0.0917, -0.3377, 0.01, -0.0895, -0.3355, 0.0061, -0.0981, -0.3355, 0.0083, -0.0981, -0.3339, 0.0061, -0.0974, -0.3339, 0.0083, -0.0974, -0.3353, 0.01, -0.0981, -0.334, 0.01, -0.0974, -0.3328, 0.0061, -0.0968, -0.3328, 0.0083, -0.0968, -0.3355, 0.01, -0.096, -0.3355, 0.01, -0.0939, -0.3329, 0.01, -0.0969, -0.3333, 0.01, -0.0939, -0.3355, 0.01, -0.0917, -0.3355, 0.01, -0.0895, -0.3333, 0.01, -0.0917, -0.3333, 0.01, -0.0895, -0.3311, 0.0061, -0.096, -0.3311, 0.0083, -0.096, -0.3297, 0.0061, -0.0952, -0.3286, 0.0061, -0.0945, -0.3297, 0.0083, -0.0952, -0.3286, 0.0083, -0.0945, -0.3309, 0.01, -0.0958,
-0.3312, 0.01, -0.0939, -0.3298, 0.01, -0.0952, -0.3287, 0.01, -0.0946, -0.3312, 0.01, -0.0917, -0.3312, 0.01, -0.0895, -0.329, 0.01, -0.0917, -0.329, 0.01, -0.0895, -0.3442, 0.0061, -0.0879, -0.3442, 0.0083, -0.0879, -0.342, 0.0061, -0.0869, -0.342, 0.0083, -0.0869, -0.3443, 0.01, -0.0879, -0.3423, 0.01, -0.087, -0.3407, 0.0061, -0.0864, -0.3396, 0.0061, -0.0859, -0.3407, 0.0083, -0.0864, -0.3396, 0.0083, -0.0859, -0.3377, 0.0061, -0.085, -0.3377, 0.0083, -0.085, -0.3404, 0.01, -0.0862, -0.3394, 0.01, -0.0858, -0.3377, 0.01, -0.0874, -0.3379, 0.01, -0.0851, -0.3362, 0.0061, -0.0843, -0.3362, 0.0083, -0.0843, -0.3351, 0.0061, -0.0838, -0.3351, 0.0083, -0.0838, -0.3333, 0.0061, -0.0829, -0.3333, 0.0083, -0.0829, -0.3355, 0.01, -0.0874,
-0.336, 0.01, -0.0842, -0.3333, 0.01, -0.0874, -0.3333, 0.01, -0.0852, -0.3349, 0.01, -0.0837, -0.3336, 0.01, -0.083, -0.3319, 0.0061, -0.0821, -0.3308, 0.0061, -0.0815, -0.3319, 0.0083, -0.0821, -0.3308, 0.0083, -0.0815, -0.329, 0.0061, -0.0804, -0.329, 0.0083, -0.0804, -0.3312, 0.01, -0.0874, -0.3312, 0.01, -0.0852, -0.329, 0.01, -0.0874, -0.329, 0.01, -0.0852, -0.3317, 0.01, -0.082, -0.3307, 0.01, -0.0814, -0.329, 0.01, -0.083, -0.3291, 0.01, -0.0804, -0.328, 0.0061, -0.0797, -0.328, 0.0083, -0.0797, -0.328, 0.01, -0.0797, -0.3268, 0.0061, -0.0935, -0.3268, 0.0083, -0.0935, -0.3257, 0.0061, -0.0928, -0.3257, 0.0083, -0.0928, -0.3246, 0.0061, -0.0921, -0.3246, 0.0083, -0.0921, -0.3267, 0.01, -0.0934, -0.3257, 0.01, -0.0928,
-0.3268, 0.01, -0.0917, -0.3268, 0.01, -0.0895, -0.3246, 0.01, -0.0921, -0.3246, 0.01, -0.0895, -0.3229, 0.0061, -0.091, -0.3219, 0.0061, -0.0903, -0.3229, 0.0083, -0.091, -0.3219, 0.0083, -0.0903, -0.3203, 0.0061, -0.0892, -0.3203, 0.0083, -0.0892, -0.323, 0.01, -0.0911, -0.3219, 0.01, -0.0903, -0.3205, 0.01, -0.0893, -0.3268, 0.01, -0.0874, -0.3268, 0.01, -0.0852, -0.3246, 0.01, -0.0874, -0.3246, 0.01, -0.0852, -0.3268, 0.01, -0.083, -0.3268, 0.01, -0.0808, -0.3246, 0.01, -0.083, -0.3246, 0.01, -0.0808, -0.3193, 0.0061, -0.0884, -0.3193, 0.0083, -0.0884, -0.3225, 0.01, -0.0874, -0.3225, 0.01, -0.0852, -0.3194, 0.01, -0.0885, -0.3203, 0.01, -0.0852, -0.3225, 0.01, -0.083, -0.3225, 0.01, -0.0808, -0.3203, 0.01, -0.083,
-0.3203, 0.01, -0.0808, -0.3268, 0.0061, -0.079, -0.3268, 0.0083, -0.079, -0.3253, 0.0061, -0.0779, -0.3242, 0.0061, -0.0772, -0.3253, 0.0083, -0.0779, -0.3242, 0.0083, -0.0772, -0.327, 0.01, -0.0791, -0.3252, 0.01, -0.0779, -0.3241, 0.01, -0.0771, -0.3226, 0.0061, -0.0761, -0.3226, 0.0083, -0.0761, -0.3216, 0.0061, -0.0753, -0.3216, 0.0083, -0.0753, -0.3203, 0.0061, -0.0743, -0.3203, 0.0083, -0.0743, -0.3225, 0.01, -0.0787, -0.3226, 0.01, -0.0761, -0.3203, 0.01, -0.0787, -0.3203, 0.01, -0.0765, -0.3215, 0.01, -0.0753, -0.3206, 0.01, -0.0745, -0.3181, 0.0061, -0.0875, -0.3181, 0.0083, -0.0875, -0.3168, 0.0061, -0.0865, -0.3157, 0.0061, -0.0856, -0.3168, 0.0083, -0.0865, -0.3157, 0.0083, -0.0856, -0.3178, 0.01, -0.0873, -0.3181, 0.01, -0.0852,
-0.3168, 0.01, -0.0865, -0.3157, 0.01, -0.0856, -0.3181, 0.01, -0.083, -0.3181, 0.01, -0.0808, -0.316, 0.01, -0.083, -0.316, 0.01, -0.0808, -0.3143, 0.0061, -0.0845, -0.3143, 0.0083, -0.0845, -0.3133, 0.0061, -0.0836, -0.3133, 0.0083, -0.0836, -0.312, 0.0061, -0.0826, -0.3109, 0.0061, -0.0816, -0.312, 0.0083, -0.0826, -0.3109, 0.0083, -0.0816, -0.3144, 0.01, -0.0846, -0.3133, 0.01, -0.0837, -0.3138, 0.01, -0.0808, -0.312, 0.01, -0.0826, -0.3109, 0.01, -0.0816, -0.3191, 0.0061, -0.0733, -0.318, 0.0061, -0.0724, -0.3191, 0.0083, -0.0733, -0.318, 0.0083, -0.0724, -0.3167, 0.0061, -0.0713, -0.3167, 0.0083, -0.0713, -0.3181, 0.01, -0.0787, -0.3181, 0.01, -0.0765, -0.316, 0.01, -0.0787, -0.316, 0.01, -0.0765, -0.319, 0.01, -0.0733,
-0.318, 0.01, -0.0724, -0.316, 0.01, -0.0743, -0.3167, 0.01, -0.0713, -0.3138, 0.01, -0.0787, -0.3138, 0.01, -0.0765, -0.3116, 0.01, -0.0787, -0.3116, 0.01, -0.0765, -0.3138, 0.01, -0.0743, -0.3138, 0.01, -0.0722, -0.3116, 0.01, -0.0743, -0.3116, 0.01, -0.0722, -0.3097, 0.0061, -0.0805, -0.3097, 0.0083, -0.0805, -0.3097, 0.01, -0.0806, -0.3086, 0.0061, -0.0795, -0.3086, 0.0083, -0.0795, -0.3075, 0.0061, -0.0784, -0.3065, 0.0061, -0.0773, -0.3075, 0.0083, -0.0784, -0.3065, 0.0083, -0.0773, -0.3086, 0.01, -0.0795, -0.3095, 0.01, -0.0765, -0.3075, 0.01, -0.0784, -0.3065, 0.01, -0.0773, -0.3095, 0.01, -0.0743, -0.3095, 0.01, -0.0722, -0.3073, 0.01, -0.0743, -0.3073, 0.01, -0.0722, -0.3054, 0.0061, -0.0762, -0.3054, 0.0083, -0.0762,
-0.3044, 0.0061, -0.075, -0.3044, 0.0083, -0.075, -0.3034, 0.0061, -0.0739, -0.3024, 0.0061, -0.0726, -0.3034, 0.0083, -0.0739, -0.3024, 0.0083, -0.0726, -0.3054, 0.01, -0.0762, -0.3043, 0.01, -0.075, -0.3051, 0.01, -0.0722, -0.3034, 0.01, -0.0739, -0.3023, 0.01, -0.0726, -0.3015, 0.0061, -0.0715, -0.3015, 0.0083, -0.0715, -0.3015, 0.01, -0.0715, -0.3156, 0.0061, -0.0704, -0.3156, 0.0083, -0.0704, -0.3156, 0.01, -0.0704, -0.3144, 0.0061, -0.0693, -0.3133, 0.0061, -0.0683, -0.3144, 0.0083, -0.0693, -0.3133, 0.0083, -0.0683, -0.3122, 0.0061, -0.0672, -0.3122, 0.0083, -0.0672, -0.3111, 0.0061, -0.0661, -0.3111, 0.0083, -0.0661, -0.3144, 0.01, -0.0693, -0.3133, 0.01, -0.0683, -0.3116, 0.01, -0.07, -0.3122, 0.01, -0.0673, -0.3112, 0.01, -0.0661,
-0.3446, 0.0061, -0.0152, -0.3446, 0.0083, -0.0152, -0.3435, 0.0061, -0.0143, -0.3435, 0.0083, -0.0143, -0.3421, 0.0061, -0.0132, -0.341, 0.0061, -0.0124, -0.3421, 0.0083, -0.0132, -0.341, 0.0083, -0.0124, -0.3446, 0.01, -0.0152, -0.3435, 0.01, -0.0144, -0.3442, 0.01, -0.0114, -0.3422, 0.01, -0.0133, -0.3411, 0.01, -0.0125, -0.3398, 0.0061, -0.0115, -0.3398, 0.0083, -0.0115, -0.3385, 0.0061, -0.0105, -0.3385, 0.0083, -0.0105, -0.3396, 0.01, -0.0113, -0.3385, 0.01, -0.0105, -0.3442, 0.01, -0.0092, -0.3442, 0.01, -0.0071, -0.342, 0.01, -0.0092, -0.342, 0.01, -0.0071, -0.3442, 0.01, -0.0049, -0.3442, 0.01, -0.0027, -0.342, 0.01, -0.0049, -0.342, 0.01, -0.0027, -0.3374, 0.0061, -0.0097, -0.3374, 0.0083, -0.0097, -0.3398, 0.01, -0.0092,
-0.3398, 0.01, -0.0071, -0.3374, 0.01, -0.0097, -0.3377, 0.01, -0.0071, -0.3398, 0.01, -0.0049, -0.3398, 0.01, -0.0027, -0.3377, 0.01, -0.0049, -0.3377, 0.01, -0.0027, -0.336, 0.0061, -0.0086, -0.3349, 0.0061, -0.0078, -0.336, 0.0083, -0.0086, -0.3349, 0.0083, -0.0078, -0.3335, 0.0061, -0.0067, -0.3335, 0.0083, -0.0067, -0.3324, 0.0061, -0.0059, -0.3324, 0.0083, -0.0059, -0.3361, 0.01, -0.0086, -0.335, 0.01, -0.0078, -0.3336, 0.01, -0.0067, -0.3355, 0.01, -0.0049, -0.3355, 0.01, -0.0027, -0.3325, 0.01, -0.0059, -0.3333, 0.01, -0.0027, -0.3312, 0.0061, -0.0049, -0.3312, 0.0083, -0.0049, -0.3298, 0.0061, -0.004, -0.3287, 0.0061, -0.0032, -0.3298, 0.0083, -0.004, -0.3287, 0.0083, -0.0032, -0.3309, 0.01, -0.0047, -0.3312, 0.01, -0.0027,
-0.3298, 0.01, -0.004, -0.3287, 0.01, -0.0032, -0.3272, 0.0061, -0.0021, -0.3272, 0.0083, -0.0021, -0.3273, 0.01, -0.0022, -0.3102, 0.0061, -0.065, -0.3091, 0.0061, -0.0636, -0.3102, 0.0083, -0.065, -0.3091, 0.0083, -0.0636, -0.3082, 0.0061, -0.0626, -0.3082, 0.0083, -0.0626, -0.3095, 0.01, -0.07, -0.3095, 0.01, -0.0678, -0.3073, 0.01, -0.07, -0.3073, 0.01, -0.0678, -0.3102, 0.01, -0.065, -0.3092, 0.01, -0.0637, -0.3073, 0.01, -0.0656, -0.3083, 0.01, -0.0626, -0.3051, 0.01, -0.07, -0.3051, 0.01, -0.0678, -0.3029, 0.01, -0.07, -0.3029, 0.01, -0.0678, -0.3051, 0.01, -0.0656, -0.3051, 0.01, -0.0635, -0.3029, 0.01, -0.0656, -0.3029, 0.01, -0.0635, -0.3072, 0.0061, -0.0613, -0.3063, 0.0061, -0.0601, -0.3072, 0.0083, -0.0613,
-0.3063, 0.0083, -0.0601, -0.307, 0.01, -0.0611, -0.3063, 0.01, -0.0601, -0.3055, 0.0061, -0.059, -0.3055, 0.0083, -0.059, -0.3044, 0.0061, -0.0576, -0.3044, 0.0083, -0.0576, -0.3036, 0.0061, -0.0565, -0.3025, 0.0061, -0.055, -0.3036, 0.0083, -0.0565, -0.3025, 0.0083, -0.055, -0.3051, 0.01, -0.0613, -0.3054, 0.01, -0.059, -0.3029, 0.01, -0.0613, -0.3029, 0.01, -0.0591, -0.3044, 0.01, -0.0576, -0.3036, 0.01, -0.0565, -0.3026, 0.01, -0.0551, -0.3004, 0.0061, -0.0701, -0.3004, 0.0083, -0.0701, -0.2996, 0.0061, -0.069, -0.2987, 0.0061, -0.0678, -0.2996, 0.0083, -0.069, -0.2987, 0.0083, -0.0678, -0.2977, 0.0061, -0.0665, -0.2977, 0.0083, -0.0665, -0.3004, 0.01, -0.0702, -0.3008, 0.01, -0.0678, -0.2996, 0.01, -0.069, -0.2988, 0.01, -0.0681,
-0.3008, 0.01, -0.0656, -0.3008, 0.01, -0.0635, -0.2976, 0.01, -0.0663, -0.2986, 0.01, -0.0635, -0.297, 0.0061, -0.0654, -0.2959, 0.0061, -0.0637, -0.297, 0.0083, -0.0654, -0.2959, 0.0083, -0.0637, -0.2952, 0.0061, -0.0627, -0.2952, 0.0083, -0.0627, -0.2969, 0.01, -0.0653, -0.2959, 0.01, -0.0637, -0.2952, 0.01, -0.0626, -0.3017, 0.0061, -0.0539, -0.3017, 0.0083, -0.0539, -0.3008, 0.01, -0.0613, -0.3008, 0.01, -0.0591, -0.2986, 0.01, -0.0613, -0.2986, 0.01, -0.0591, -0.3008, 0.01, -0.057, -0.3019, 0.01, -0.054, -0.2986, 0.01, -0.057, -0.2986, 0.01, -0.0548, -0.2943, 0.0061, -0.0613, -0.2934, 0.0061, -0.0599, -0.2943, 0.0083, -0.0613, -0.2934, 0.0083, -0.0599, -0.2964, 0.01, -0.0613, -0.2964, 0.01, -0.0591, -0.2945, 0.01, -0.0615,
-0.2933, 0.01, -0.0597, -0.2964, 0.01, -0.057, -0.2964, 0.01, -0.0548, -0.2943, 0.01, -0.057, -0.2943, 0.01, -0.0548, -0.3009, 0.0061, -0.0526, -0.3, 0.0061, -0.0511, -0.3009, 0.0083, -0.0526, -0.3, 0.0083, -0.0511, -0.2993, 0.0061, -0.05, -0.2993, 0.0083, -0.05, -0.2983, 0.0061, -0.0483, -0.2976, 0.0061, -0.047, -0.2983, 0.0083, -0.0483, -0.2976, 0.0083, -0.047, -0.3007, 0.01, -0.0524, -0.3, 0.01, -0.0512, -0.2986, 0.01, -0.0526, -0.2993, 0.01, -0.0501, -0.2982, 0.01, -0.0481, -0.2976, 0.01, -0.047, -0.2971, 0.0061, -0.0459, -0.2971, 0.0083, -0.0459, -0.2964, 0.01, -0.0526, -0.2964, 0.01, -0.0505, -0.2943, 0.01, -0.0526, -0.2943, 0.01, -0.0505, -0.2964, 0.01, -0.0483, -0.2971, 0.01, -0.0459, -0.2943, 0.01, -0.0483,
-0.2943, 0.01, -0.0461, -0.2962, 0.0061, -0.044, -0.2955, 0.0061, -0.0425, -0.2962, 0.0083, -0.044, -0.2955, 0.0083, -0.0425, -0.2951, 0.0061, -0.0414, -0.2951, 0.0083, -0.0414, -0.2943, 0.0061, -0.0396, -0.2935, 0.0061, -0.0375, -0.2943, 0.0083, -0.0396, -0.2935, 0.0083, -0.0375, -0.2961, 0.01, -0.0438, -0.2956, 0.01, -0.0426, -0.2943, 0.01, -0.0439, -0.2951, 0.01, -0.0415, -0.2943, 0.01, -0.0394, -0.2936, 0.01, -0.0378, -0.2927, 0.0061, -0.0588, -0.2927, 0.0083, -0.0588, -0.2916, 0.0061, -0.057, -0.2916, 0.0083, -0.057, -0.291, 0.0061, -0.0559, -0.2904, 0.0061, -0.0548, -0.291, 0.0083, -0.0559, -0.2904, 0.0083, -0.0548, -0.2927, 0.01, -0.0587, -0.2917, 0.01, -0.057, -0.2921, 0.01, -0.0548, -0.291, 0.01, -0.0559, -0.2904, 0.01, -0.0549,
-0.2893, 0.0061, -0.0527, -0.2893, 0.0083, -0.0527, -0.2921, 0.01, -0.0526, -0.2921, 0.01, -0.0505, -0.2893, 0.01, -0.0527, -0.2899, 0.01, -0.0505, -0.2921, 0.01, -0.0483, -0.2921, 0.01, -0.0461, -0.2899, 0.01, -0.0483, -0.2899, 0.01, -0.0461, -0.2888, 0.0061, -0.0516, -0.2883, 0.0061, -0.0505, -0.2888, 0.0083, -0.0516, -0.2883, 0.0083, -0.0505, -0.2874, 0.0061, -0.0483, -0.2868, 0.0061, -0.0469, -0.2874, 0.0083, -0.0483, -0.2868, 0.0083, -0.0469, -0.2863, 0.0061, -0.0458, -0.2863, 0.0083, -0.0458, -0.2888, 0.01, -0.0516, -0.2884, 0.01, -0.0506, -0.2875, 0.01, -0.0486, -0.2867, 0.01, -0.0466, -0.2862, 0.01, -0.0457, -0.2931, 0.0061, -0.0364, -0.2931, 0.0083, -0.0364, -0.2921, 0.01, -0.0439, -0.2921, 0.01, -0.0418, -0.2899, 0.01, -0.0439,
-0.2899, 0.01, -0.0418, -0.2921, 0.01, -0.0396, -0.2933, 0.01, -0.0369, -0.2899, 0.01, -0.0396, -0.2899, 0.01, -0.0374, -0.2855, 0.0061, -0.044, -0.2848, 0.0061, -0.0419, -0.2855, 0.0083, -0.044, -0.2848, 0.0083, -0.0419, -0.2878, 0.01, -0.0439, -0.2878, 0.01, -0.0418, -0.2856, 0.01, -0.0441, -0.2848, 0.01, -0.0419, -0.2878, 0.01, -0.0396, -0.2878, 0.01, -0.0374, -0.2856, 0.01, -0.0396, -0.2856, 0.01, -0.0374, -0.2845, 0.0061, -0.0408, -0.2845, 0.0083, -0.0408, -0.2841, 0.0061, -0.0396, -0.2834, 0.0061, -0.0374, -0.2841, 0.0083, -0.0396, -0.2834, 0.0083, -0.0374, -0.2844, 0.01, -0.0408, -0.2841, 0.01, -0.0397, -0.2835, 0.01, -0.0376, -0.2927, 0.0061, -0.0353, -0.292, 0.0061, -0.0331, -0.2927, 0.0083, -0.0353, -0.292, 0.0083, -0.0331,
-0.2914, 0.0061, -0.0309, -0.2911, 0.0061, -0.0297, -0.2914, 0.0083, -0.0309, -0.2911, 0.0083, -0.0297, -0.2908, 0.0061, -0.0286, -0.2908, 0.0083, -0.0286, -0.2926, 0.01, -0.035, -0.292, 0.01, -0.033, -0.2899, 0.01, -0.0353, -0.2899, 0.01, -0.0331, -0.2914, 0.01, -0.0309, -0.2911, 0.01, -0.0297, -0.2899, 0.01, -0.0309, -0.2908, 0.01, -0.0287, -0.2878, 0.01, -0.0353, -0.2878, 0.01, -0.0331, -0.2856, 0.01, -0.0353, -0.2856, 0.01, -0.0331, -0.2878, 0.01, -0.0309, -0.2878, 0.01, -0.0288, -0.2856, 0.01, -0.0309, -0.2856, 0.01, -0.0288, -0.2904, 0.0061, -0.0266, -0.2899, 0.0061, -0.0244, -0.2904, 0.0083, -0.0266, -0.2899, 0.0083, -0.0244, -0.2894, 0.0061, -0.0222, -0.289, 0.0061, -0.0204, -0.2894, 0.0083, -0.0222, -0.289, 0.0083, -0.0204,
-0.2903, 0.01, -0.0264, -0.2899, 0.01, -0.0243, -0.2894, 0.01, -0.0222, -0.2891, 0.01, -0.0206, -0.2888, 0.0061, -0.0193, -0.2888, 0.0083, -0.0193, -0.2878, 0.01, -0.0266, -0.2878, 0.01, -0.0244, -0.2856, 0.01, -0.0266, -0.2856, 0.01, -0.0244, -0.2878, 0.01, -0.0222, -0.2889, 0.01, -0.0198, -0.2856, 0.01, -0.0222, -0.2856, 0.01, -0.0201, -0.2828, 0.0061, -0.0353, -0.2824, 0.0061, -0.0338, -0.2828, 0.0083, -0.0353, -0.2824, 0.0083, -0.0338, -0.2821, 0.0061, -0.0328, -0.2821, 0.0083, -0.0328, -0.2816, 0.0061, -0.0309, -0.2812, 0.0061, -0.0288, -0.2816, 0.0083, -0.0309, -0.2812, 0.0083, -0.0288, -0.2829, 0.01, -0.0355, -0.2823, 0.01, -0.0334, -0.2821, 0.01, -0.0325, -0.2834, 0.01, -0.0309, -0.2834, 0.01, -0.0288, -0.2817, 0.01, -0.031,
-0.2812, 0.01, -0.0289, -0.2808, 0.0061, -0.0266, -0.2804, 0.0061, -0.0244, -0.2808, 0.0083, -0.0266, -0.2804, 0.0083, -0.0244, -0.2802, 0.0061, -0.0231, -0.2802, 0.0083, -0.0231, -0.2834, 0.01, -0.0266, -0.2834, 0.01, -0.0244, -0.2808, 0.01, -0.0267, -0.2804, 0.01, -0.0246, -0.2834, 0.01, -0.0222, -0.2834, 0.01, -0.0201, -0.2801, 0.01, -0.0225, -0.2812, 0.01, -0.0201, -0.28, 0.0061, -0.022, -0.2797, 0.0061, -0.0201, -0.28, 0.0083, -0.022, -0.2797, 0.0083, -0.0201, -0.28, 0.01, -0.0218, -0.2797, 0.01, -0.0201, -0.2885, 0.0061, -0.0179, -0.2883, 0.0061, -0.0157, -0.2885, 0.0083, -0.0179, -0.2883, 0.0083, -0.0157, -0.2882, 0.0061, -0.0136, -0.2882, 0.0061, -0.0114, -0.2882, 0.0083, -0.0136, -0.2882, 0.0083, -0.0114, -0.2885, 0.01, -0.0178,
-0.2883, 0.01, -0.0157, -0.2856, 0.01, -0.0179, -0.2856, 0.01, -0.0157, -0.2882, 0.01, -0.0136, -0.2882, 0.01, -0.0114, -0.2856, 0.01, -0.0136, -0.2856, 0.01, -0.0114, -0.2882, 0.0061, -0.0092, -0.2882, 0.0061, -0.0071, -0.2882, 0.0083, -0.0092, -0.2882, 0.0083, -0.0071, -0.2882, 0.0061, -0.0049, -0.2882, 0.0061, -0.0027, -0.2882, 0.0083, -0.0049, -0.2882, 0.0083, -0.0027, -0.2882, 0.01, -0.0092, -0.2882, 0.01, -0.0071, -0.2856, 0.01, -0.0092, -0.2856, 0.01, -0.0071, -0.2882, 0.01, -0.0049, -0.2882, 0.01, -0.0027, -0.2856, 0.01, -0.0049, -0.2856, 0.01, -0.0027, -0.2834, 0.01, -0.0179, -0.2834, 0.01, -0.0157, -0.2812, 0.01, -0.0179, -0.2812, 0.01, -0.0157, -0.2834, 0.01, -0.0136, -0.2834, 0.01, -0.0114, -0.2812, 0.01, -0.0136,
-0.2812, 0.01, -0.0114, -0.2795, 0.0061, -0.0179, -0.2794, 0.0061, -0.0157, -0.2795, 0.0083, -0.0179, -0.2794, 0.0083, -0.0157, -0.2794, 0.0061, -0.0136, -0.2794, 0.0061, -0.0114, -0.2794, 0.0083, -0.0136, -0.2794, 0.0083, -0.0114, -0.2795, 0.01, -0.0179, -0.2794, 0.01, -0.0157, -0.2794, 0.01, -0.0136, -0.2794, 0.01, -0.0114, -0.2834, 0.01, -0.0092, -0.2834, 0.01, -0.0071, -0.2812, 0.01, -0.0092, -0.2812, 0.01, -0.0071, -0.2834, 0.01, -0.0049, -0.2834, 0.01, -0.0027, -0.2812, 0.01, -0.0049, -0.2812, 0.01, -0.0027, -0.2794, 0.0061, -0.0092, -0.2794, 0.0061, -0.0071, -0.2794, 0.0083, -0.0092, -0.2794, 0.0083, -0.0071, -0.2794, 0.0061, -0.0049, -0.2794, 0.0061, -0.0027, -0.2794, 0.0083, -0.0049, -0.2794, 0.0083, -0.0027, -0.2794, 0.01, -0.0092,
-0.2794, 0.01, -0.0071, -0.2794, 0.01, -0.0049, -0.2794, 0.01, -0.0027, -0.2196, 0.0061, -0.067, -0.2196, 0.0083, -0.067, -0.2195, 0.0061, -0.0666, -0.2195, 0.0083, -0.0666, -0.2196, 0.01, -0.067, -0.2195, 0.01, -0.0666, -0.2183, 0.0061, -0.067, -0.2183, 0.0083, -0.067, -0.2172, 0.0061, -0.0667, -0.2172, 0.0083, -0.0667, -0.219, 0.0061, -0.0655, -0.2181, 0.0061, -0.0635, -0.219, 0.0083, -0.0655, -0.2181, 0.0083, -0.0635, -0.216, 0.0061, -0.0665, -0.216, 0.0083, -0.0665, -0.2183, 0.01, -0.067, -0.2172, 0.01, -0.0667, -0.219, 0.01, -0.0655, -0.218, 0.01, -0.0633, -0.216, 0.01, -0.0665, -0.2161, 0.01, -0.0635, -0.2174, 0.0061, -0.0621, -0.2174, 0.0083, -0.0621, -0.2168, 0.0061, -0.061, -0.2157, 0.0061, -0.0592, -0.2168, 0.0083, -0.061,
-0.2157, 0.0083, -0.0592, -0.2174, 0.01, -0.0622, -0.2168, 0.01, -0.0611, -0.2158, 0.01, -0.0593, -0.2141, 0.0061, -0.0652, -0.2132, 0.0061, -0.0642, -0.2141, 0.0083, -0.0652, -0.2132, 0.0083, -0.0642, -0.2124, 0.0061, -0.0631, -0.2124, 0.0083, -0.0631, -0.2142, 0.01, -0.0654, -0.2131, 0.01, -0.0641, -0.2124, 0.01, -0.063, -0.215, 0.0061, -0.0581, -0.215, 0.0083, -0.0581, -0.2113, 0.0061, -0.0614, -0.2113, 0.0083, -0.0614, -0.2144, 0.0061, -0.057, -0.2133, 0.0061, -0.0552, -0.2144, 0.0083, -0.057, -0.2133, 0.0083, -0.0552, -0.2126, 0.0061, -0.0541, -0.2126, 0.0083, -0.0541, -0.214, 0.01, -0.0613, -0.2152, 0.01, -0.0583, -0.2114, 0.01, -0.0614, -0.2118, 0.01, -0.0591, -0.2142, 0.01, -0.0566, -0.2134, 0.01, -0.0553, -0.2118, 0.01, -0.057,
-0.2127, 0.01, -0.0542, -0.2107, 0.0061, -0.0603, -0.21, 0.0061, -0.0591, -0.2107, 0.0083, -0.0603, -0.21, 0.0083, -0.0591, -0.2089, 0.0061, -0.0574, -0.2089, 0.0083, -0.0574, -0.2083, 0.0061, -0.0563, -0.2074, 0.0061, -0.0548, -0.2083, 0.0083, -0.0563, -0.2074, 0.0083, -0.0548, -0.2107, 0.01, -0.0603, -0.2101, 0.01, -0.0593, -0.2089, 0.01, -0.0573, -0.2096, 0.01, -0.0548, -0.2083, 0.01, -0.0562, -0.2075, 0.01, -0.0551, -0.2117, 0.0061, -0.0526, -0.2109, 0.0061, -0.0513, -0.2117, 0.0083, -0.0526, -0.2109, 0.0083, -0.0513, -0.2115, 0.01, -0.0524, -0.2109, 0.01, -0.0513, -0.2102, 0.0061, -0.0502, -0.2102, 0.0083, -0.0502, -0.2066, 0.0061, -0.0534, -0.2066, 0.0083, -0.0534, -0.2091, 0.0061, -0.0484, -0.2091, 0.0083, -0.0484, -0.2085, 0.0061, -0.0473,
-0.2076, 0.0061, -0.0461, -0.2085, 0.0083, -0.0473, -0.2076, 0.0083, -0.0461, -0.2096, 0.01, -0.0526, -0.2102, 0.01, -0.0502, -0.2065, 0.01, -0.0532, -0.2075, 0.01, -0.0505, -0.2092, 0.01, -0.0486, -0.2086, 0.01, -0.0475, -0.2074, 0.01, -0.0458, -0.2066, 0.0061, -0.0448, -0.2066, 0.0083, -0.0448, -0.2066, 0.01, -0.0448, -0.2196, 0.0061, -0.0219, -0.2196, 0.0061, -0.0201, -0.2196, 0.0083, -0.0219, -0.2196, 0.0083, -0.0201, -0.2196, 0.01, -0.0219, -0.2196, 0.01, -0.0201, -0.2183, 0.0061, -0.0219, -0.2183, 0.0083, -0.0219, -0.2161, 0.0061, -0.0219, -0.2161, 0.0083, -0.0219, -0.2183, 0.01, -0.0219, -0.2183, 0.01, -0.0201, -0.2161, 0.01, -0.0219, -0.2161, 0.01, -0.0201, -0.214, 0.0061, -0.0219, -0.214, 0.0083, -0.0219, -0.2118, 0.0061, -0.0219,
-0.2118, 0.0083, -0.0219, -0.214, 0.01, -0.0219, -0.214, 0.01, -0.0201, -0.2118, 0.01, -0.0219, -0.2118, 0.01, -0.0201, -0.2088, 0.0061, -0.0219, -0.2088, 0.0061, -0.0201, -0.2088, 0.0083, -0.0219, -0.2088, 0.0083, -0.0201, -0.2088, 0.01, -0.0219, -0.2088, 0.01, -0.0201, -0.2196, 0.0061, -0.0179, -0.2196, 0.0061, -0.0157, -0.2196, 0.0083, -0.0179, -0.2196, 0.0083, -0.0157, -0.2196, 0.0061, -0.0136, -0.2196, 0.0061, -0.0114, -0.2196, 0.0083, -0.0136, -0.2196, 0.0083, -0.0114, -0.2196, 0.01, -0.0179, -0.2196, 0.01, -0.0157, -0.2196, 0.01, -0.0136, -0.2196, 0.01, -0.0114, -0.2183, 0.01, -0.0179, -0.2183, 0.01, -0.0157, -0.2161, 0.01, -0.0179, -0.2161, 0.01, -0.0157, -0.2183, 0.01, -0.0136, -0.2183, 0.01, -0.0114, -0.2161, 0.01, -0.0136,
-0.2161, 0.01, -0.0114, -0.2196, 0.0061, -0.0092, -0.2196, 0.0061, -0.0071, -0.2196, 0.0083, -0.0092, -0.2196, 0.0083, -0.0071, -0.2196, 0.0061, -0.0049, -0.2196, 0.0061, -0.0027, -0.2196, 0.0083, -0.0049, -0.2196, 0.0083, -0.0027, -0.2196, 0.01, -0.0092, -0.2196, 0.01, -0.0071, -0.2196, 0.01, -0.0049, -0.2196, 0.01, -0.0027, -0.2183, 0.01, -0.0092, -0.2183, 0.01, -0.0071, -0.2161, 0.01, -0.0092, -0.2161, 0.01, -0.0071, -0.2183, 0.01, -0.0049, -0.2183, 0.01, -0.0027, -0.2161, 0.01, -0.0049, -0.2161, 0.01, -0.0027, -0.214, 0.01, -0.0179, -0.214, 0.01, -0.0157, -0.2118, 0.01, -0.0179, -0.2118, 0.01, -0.0157, -0.214, 0.01, -0.0136, -0.214, 0.01, -0.0114, -0.2118, 0.01, -0.0136, -0.2118, 0.01, -0.0114, -0.2088, 0.0061, -0.0179,
-0.2088, 0.0061, -0.0157, -0.2088, 0.0083, -0.0179, -0.2088, 0.0083, -0.0157, -0.2088, 0.0061, -0.0136, -0.2088, 0.0061, -0.0114, -0.2088, 0.0083, -0.0136, -0.2088, 0.0083, -0.0114, -0.2088, 0.01, -0.0179, -0.2088, 0.01, -0.0157, -0.2088, 0.01, -0.0136, -0.2088, 0.01, -0.0114, -0.214, 0.01, -0.0092, -0.214, 0.01, -0.0071, -0.2118, 0.01, -0.0092, -0.2118, 0.01, -0.0071, -0.214, 0.01, -0.0049, -0.214, 0.01, -0.0027, -0.2118, 0.01, -0.0049, -0.2118, 0.01, -0.0027, -0.2088, 0.0061, -0.0092, -0.2088, 0.0061, -0.0071, -0.2088, 0.0083, -0.0092, -0.2088, 0.0083, -0.0071, -0.2088, 0.0061, -0.0049, -0.2088, 0.0061, -0.0027, -0.2088, 0.0083, -0.0049, -0.2088, 0.0083, -0.0027, -0.2088, 0.01, -0.0092, -0.2088, 0.01, -0.0071, -0.2088, 0.01, -0.0049,
-0.2088, 0.01, -0.0027, -0.2069, 0.0061, -0.0591, -0.2069, 0.0083, -0.0591, -0.2031, 0.0061, -0.0591, -0.2031, 0.0083, -0.0591, -0.206, 0.0061, -0.0573, -0.206, 0.0083, -0.0573, -0.2032, 0.0061, -0.0563, -0.2032, 0.0083, -0.0563, -0.2069, 0.01, -0.0591, -0.2031, 0.01, -0.0591, -0.206, 0.01, -0.0573, -0.2033, 0.01, -0.0563, -0.2009, 0.0061, -0.0591, -0.2009, 0.0083, -0.0591, -0.1988, 0.0061, -0.0591, -0.1988, 0.0083, -0.0591, -0.2009, 0.0061, -0.0562, -0.2, 0.0061, -0.054, -0.2009, 0.0083, -0.0562, -0.2, 0.0083, -0.054, -0.199, 0.0061, -0.0562, -0.1995, 0.0061, -0.0551, -0.199, 0.0083, -0.0562, -0.1995, 0.0083, -0.0551, -0.2009, 0.01, -0.0591, -0.1988, 0.01, -0.0591, -0.201, 0.01, -0.0562, -0.2001, 0.01, -0.054, -0.199, 0.01, -0.0562,
-0.1995, 0.01, -0.055, -0.1941, 0.0061, -0.0635, -0.1941, 0.0083, -0.0635, -0.1942, 0.01, -0.0634, -0.1894, 0.0061, -0.0668, -0.1894, 0.0083, -0.0668, -0.1924, 0.0061, -0.0656, -0.1933, 0.0061, -0.0645, -0.1924, 0.0083, -0.0656, -0.1933, 0.0083, -0.0645, -0.1905, 0.0061, -0.0666, -0.1892, 0.0061, -0.0628, -0.1905, 0.0083, -0.0666, -0.1892, 0.0083, -0.0628, -0.1893, 0.01, -0.0668, -0.1923, 0.01, -0.0656, -0.1934, 0.01, -0.0645, -0.1904, 0.01, -0.0667, -0.1891, 0.01, -0.063, -0.1956, 0.0061, -0.0605, -0.1961, 0.0061, -0.0591, -0.1956, 0.0083, -0.0605, -0.1961, 0.0083, -0.0591, -0.1952, 0.0061, -0.0616, -0.1952, 0.0083, -0.0616, -0.1939, 0.0061, -0.0546, -0.1939, 0.0083, -0.0546, -0.1957, 0.01, -0.0605, -0.1961, 0.01, -0.0591, -0.1952, 0.01, -0.0616,
-0.1944, 0.01, -0.0591, -0.1966, 0.01, -0.057, -0.1966, 0.01, -0.0548, -0.1944, 0.01, -0.057, -0.1939, 0.01, -0.0546, -0.1916, 0.0061, -0.0587, -0.1916, 0.0083, -0.0587, -0.1901, 0.0061, -0.0613, -0.1909, 0.0061, -0.0598, -0.1901, 0.0083, -0.0613, -0.1909, 0.0083, -0.0598, -0.1926, 0.0061, -0.057, -0.1933, 0.0061, -0.0557, -0.1926, 0.0083, -0.057, -0.1933, 0.0083, -0.0557, -0.1923, 0.01, -0.0613, -0.1915, 0.01, -0.0588, -0.1902, 0.01, -0.0611, -0.1909, 0.01, -0.0599, -0.1927, 0.01, -0.0568, -0.1933, 0.01, -0.0558, -0.2059, 0.0061, -0.0523, -0.2049, 0.0061, -0.0505, -0.2059, 0.0083, -0.0523, -0.2049, 0.0083, -0.0505, -0.2026, 0.0061, -0.0501, -0.2026, 0.0083, -0.0501, -0.2043, 0.0061, -0.0492, -0.2043, 0.0083, -0.0492, -0.2039, 0.0061, -0.0484,
-0.2039, 0.0083, -0.0484, -0.2059, 0.01, -0.0522, -0.205, 0.01, -0.0508, -0.2026, 0.01, -0.05, -0.2041, 0.01, -0.0489, -0.2053, 0.01, -0.0461, -0.2039, 0.01, -0.0484, -0.2031, 0.01, -0.0461, -0.2008, 0.0061, -0.0526, -0.2018, 0.0061, -0.0512, -0.2008, 0.0083, -0.0526, -0.2018, 0.0083, -0.0512, -0.1979, 0.0061, -0.0476, -0.1989, 0.0061, -0.046, -0.1979, 0.0083, -0.0476, -0.1989, 0.0083, -0.046, -0.2007, 0.01, -0.0528, -0.2019, 0.01, -0.0511, -0.1988, 0.01, -0.0526, -0.1988, 0.01, -0.0505, -0.2009, 0.01, -0.0483, -0.2009, 0.01, -0.0461, -0.1978, 0.01, -0.0477, -0.199, 0.01, -0.0458, -0.2052, 0.0061, -0.0436, -0.2052, 0.0083, -0.0436, -0.2026, 0.0061, -0.043, -0.2026, 0.0083, -0.043, -0.2052, 0.01, -0.0436, -0.2026, 0.01, -0.043,
-0.2006, 0.0061, -0.0441, -0.2006, 0.0083, -0.0441, -0.1998, 0.0061, -0.045, -0.1998, 0.0083, -0.045, -0.2006, 0.01, -0.0441, -0.1998, 0.01, -0.045, -0.1956, 0.0061, -0.0517, -0.1963, 0.0061, -0.0505, -0.1956, 0.0083, -0.0517, -0.1963, 0.0083, -0.0505, -0.195, 0.0061, -0.0528, -0.195, 0.0083, -0.0528, -0.1973, 0.0061, -0.0487, -0.1973, 0.0083, -0.0487, -0.1955, 0.01, -0.0519, -0.1965, 0.01, -0.0502, -0.1949, 0.01, -0.0529, -0.1972, 0.01, -0.0488, -0.1873, 0.0061, -0.067, -0.1873, 0.0083, -0.067, -0.1878, 0.0061, -0.0656, -0.1886, 0.0061, -0.0639, -0.1878, 0.0083, -0.0656, -0.1886, 0.0083, -0.0639, -0.1873, 0.01, -0.067, -0.1879, 0.01, -0.0654, -0.1886, 0.01, -0.064, -0.1686, 0.0061, -0.067, -0.1686, 0.0083, -0.067, -0.1686, 0.0061, -0.0656,
-0.1686, 0.0061, -0.0635, -0.1686, 0.0083, -0.0656, -0.1686, 0.0083, -0.0635, -0.1686, 0.01, -0.067, -0.1686, 0.01, -0.0656, -0.1686, 0.01, -0.0635, -0.1662, 0.0061, -0.067, -0.1662, 0.0083, -0.067, -0.1641, 0.0061, -0.067, -0.1641, 0.0083, -0.067, -0.1647, 0.0061, -0.0631, -0.1647, 0.0083, -0.0631, -0.1662, 0.01, -0.067, -0.1641, 0.01, -0.067, -0.1662, 0.01, -0.0656, -0.1662, 0.01, -0.0635, -0.1641, 0.01, -0.0656, -0.1647, 0.01, -0.0631, -0.1686, 0.0061, -0.0613, -0.1686, 0.0061, -0.0591, -0.1686, 0.0083, -0.0613, -0.1686, 0.0083, -0.0591, -0.1686, 0.0061, -0.057, -0.1686, 0.0061, -0.0548, -0.1686, 0.0083, -0.057, -0.1686, 0.0083, -0.0548, -0.1686, 0.01, -0.0613, -0.1686, 0.01, -0.0591, -0.1686, 0.01, -0.057, -0.1686, 0.01, -0.0548,
-0.1647, 0.0061, -0.0613, -0.1647, 0.0061, -0.0591, -0.1647, 0.0083, -0.0613, -0.1647, 0.0083, -0.0591, -0.1647, 0.0061, -0.057, -0.1647, 0.0061, -0.0548, -0.1647, 0.0083, -0.057, -0.1647, 0.0083, -0.0548, -0.1662, 0.01, -0.0613, -0.1662, 0.01, -0.0591, -0.1647, 0.01, -0.0613, -0.1647, 0.01, -0.0591, -0.1662, 0.01, -0.057, -0.1662, 0.01, -0.0548, -0.1647, 0.01, -0.057, -0.1647, 0.01, -0.0548, -0.1619, 0.0061, -0.067, -0.1619, 0.0083, -0.067, -0.1597, 0.0061, -0.067, -0.1597, 0.0083, -0.067, -0.1619, 0.0061, -0.0631, -0.1619, 0.0083, -0.0631, -0.1597, 0.0061, -0.0631, -0.1597, 0.0083, -0.0631, -0.1619, 0.01, -0.067, -0.1597, 0.01, -0.067, -0.1619, 0.01, -0.0656, -0.1619, 0.01, -0.0631, -0.1597, 0.01, -0.0656, -0.1597, 0.01, -0.0631,
-0.1575, 0.0061, -0.067, -0.1575, 0.0083, -0.067, -0.1554, 0.0061, -0.067, -0.1554, 0.0083, -0.067, -0.1575, 0.0061, -0.0631, -0.1575, 0.0083, -0.0631, -0.1554, 0.0061, -0.063, -0.1554, 0.0083, -0.063, -0.1575, 0.01, -0.067, -0.1554, 0.01, -0.067, -0.1575, 0.01, -0.0656, -0.1576, 0.01, -0.0631, -0.1554, 0.01, -0.0656, -0.1554, 0.01, -0.063, -0.1686, 0.0061, -0.0526, -0.1686, 0.0061, -0.0505, -0.1686, 0.0083, -0.0526, -0.1686, 0.0083, -0.0505, -0.1686, 0.0061, -0.0483, -0.1686, 0.0061, -0.0461, -0.1686, 0.0083, -0.0483, -0.1686, 0.0083, -0.0461, -0.1686, 0.01, -0.0526, -0.1686, 0.01, -0.0505, -0.1686, 0.01, -0.0483, -0.1686, 0.01, -0.0461, -0.1647, 0.0061, -0.0526, -0.1647, 0.0061, -0.0505, -0.1647, 0.0083, -0.0526, -0.1647, 0.0083, -0.0505,
-0.1647, 0.0061, -0.0483, -0.1647, 0.0061, -0.0454, -0.1647, 0.0083, -0.0483, -0.1647, 0.0083, -0.0454, -0.1662, 0.01, -0.0526, -0.1662, 0.01, -0.0505, -0.1647, 0.01, -0.0526, -0.1647, 0.01, -0.0505, -0.1662, 0.01, -0.0483, -0.1662, 0.01, -0.0461, -0.1647, 0.01, -0.0483, -0.1647, 0.01, -0.0454, -0.1686, 0.0061, -0.0439, -0.1686, 0.0061, -0.0425, -0.1686, 0.0083, -0.0439, -0.1686, 0.0083, -0.0425, -0.1686, 0.01, -0.0439, -0.1686, 0.01, -0.0425, -0.1662, 0.0061, -0.0425, -0.1662, 0.0083, -0.0425, -0.1641, 0.0061, -0.0425, -0.1641, 0.0083, -0.0425, -0.1662, 0.01, -0.0439, -0.1662, 0.01, -0.0425, -0.1641, 0.01, -0.0439, -0.1641, 0.01, -0.0425, -0.1619, 0.0061, -0.0454, -0.1619, 0.0083, -0.0454, -0.1597, 0.0061, -0.0454, -0.1597, 0.0083, -0.0454,
-0.1619, 0.01, -0.0454, -0.1597, 0.01, -0.0454, -0.1575, 0.0061, -0.0455, -0.1575, 0.0083, -0.0455, -0.1554, 0.0061, -0.0457, -0.1554, 0.0083, -0.0457, -0.1576, 0.01, -0.0455, -0.1554, 0.01, -0.0456, -0.1619, 0.0061, -0.0425, -0.1619, 0.0083, -0.0425, -0.1597, 0.0061, -0.0425, -0.1597, 0.0083, -0.0425, -0.1619, 0.01, -0.0439, -0.1619, 0.01, -0.0425, -0.1597, 0.01, -0.0439, -0.1597, 0.01, -0.0425, -0.1575, 0.0061, -0.0425, -0.1575, 0.0083, -0.0425, -0.1554, 0.0061, -0.0425, -0.1554, 0.0083, -0.0425, -0.1575, 0.01, -0.0439, -0.1575, 0.01, -0.0425, -0.1554, 0.01, -0.0439, -0.1554, 0.01, -0.0425, -0.1535, 0.0061, -0.0669, -0.1535, 0.0083, -0.0669, -0.1524, 0.0061, -0.0667, -0.1532, 0.0061, -0.0629, -0.1524, 0.0083, -0.0667, -0.1532, 0.0083, -0.0629,
-0.151, 0.0061, -0.0665, -0.1516, 0.0061, -0.0625, -0.151, 0.0083, -0.0665, -0.1516, 0.0083, -0.0625, -0.1537, 0.01, -0.0669, -0.1529, 0.01, -0.0668, -0.1533, 0.01, -0.0629, -0.1507, 0.01, -0.0664, -0.1513, 0.01, -0.0624, -0.1488, 0.0061, -0.0656, -0.1488, 0.0083, -0.0656, -0.1475, 0.0061, -0.0648, -0.1463, 0.0061, -0.064, -0.1475, 0.0083, -0.0648, -0.1463, 0.0083, -0.064, -0.1486, 0.01, -0.0655, -0.1489, 0.01, -0.0635, -0.1475, 0.01, -0.0648, -0.1463, 0.01, -0.064, -0.1504, 0.0061, -0.0621, -0.1504, 0.0083, -0.0621, -0.1504, 0.01, -0.0621, -0.1489, 0.0061, -0.0611, -0.148, 0.0061, -0.06, -0.1489, 0.0083, -0.0611, -0.148, 0.0083, -0.06, -0.1472, 0.0061, -0.059, -0.1472, 0.0083, -0.059, -0.1465, 0.0061, -0.057, -0.1462, 0.0061, -0.0548,
-0.1465, 0.0083, -0.057, -0.1462, 0.0083, -0.0548, -0.149, 0.01, -0.0612, -0.148, 0.01, -0.06, -0.1467, 0.01, -0.0613, -0.1472, 0.01, -0.059, -0.1464, 0.01, -0.0569, -0.1462, 0.01, -0.0548, -0.1452, 0.0061, -0.0628, -0.1452, 0.0083, -0.0628, -0.1452, 0.01, -0.0628, -0.144, 0.0061, -0.0614, -0.144, 0.0083, -0.0614, -0.1434, 0.0061, -0.0603, -0.1428, 0.0061, -0.0592, -0.1434, 0.0083, -0.0603, -0.1428, 0.0083, -0.0592, -0.1423, 0.0061, -0.057, -0.1422, 0.0061, -0.0548, -0.1423, 0.0083, -0.057, -0.1422, 0.0083, -0.0548, -0.1441, 0.01, -0.0615, -0.1445, 0.01, -0.0591, -0.1434, 0.01, -0.0603, -0.1429, 0.01, -0.0593, -0.1445, 0.01, -0.057, -0.1445, 0.01, -0.0548, -0.1423, 0.01, -0.0571, -0.1422, 0.01, -0.0548, -0.1532, 0.0061, -0.0461,
-0.1532, 0.0083, -0.0461, -0.1503, 0.0061, -0.0474, -0.1514, 0.0061, -0.0468, -0.1503, 0.0083, -0.0474, -0.1514, 0.0083, -0.0468, -0.1533, 0.01, -0.0461, -0.1502, 0.01, -0.0474, -0.1512, 0.01, -0.0469, -0.1479, 0.0061, -0.0495, -0.1479, 0.0083, -0.0495, -0.1464, 0.0061, -0.0526, -0.1469, 0.0061, -0.0509, -0.1464, 0.0083, -0.0526, -0.1469, 0.0083, -0.0509, -0.1489, 0.0061, -0.0484, -0.1489, 0.0083, -0.0484, -0.1458, 0.0061, -0.0452, -0.1458, 0.0083, -0.0452, -0.1479, 0.01, -0.0495, -0.1464, 0.01, -0.0527, -0.1471, 0.01, -0.0505, -0.1489, 0.01, -0.0484, -0.1489, 0.01, -0.0461, -0.1467, 0.01, -0.0483, -0.1458, 0.01, -0.0451, -0.1532, 0.0061, -0.0425, -0.1532, 0.0083, -0.0425, -0.1501, 0.0061, -0.0429, -0.1511, 0.0061, -0.0426, -0.1501, 0.0083, -0.0429,
-0.1511, 0.0083, -0.0426, -0.1532, 0.01, -0.0439, -0.1532, 0.01, -0.0425, -0.1506, 0.01, -0.0428, -0.1515, 0.01, -0.0426, -0.1488, 0.0061, -0.0432, -0.1488, 0.0083, -0.0432, -0.1468, 0.0061, -0.0443, -0.1468, 0.0083, -0.0443, -0.1485, 0.01, -0.0433, -0.1469, 0.01, -0.0443, -0.1422, 0.0061, -0.0526, -0.1425, 0.0061, -0.0504, -0.1422, 0.0083, -0.0526, -0.1425, 0.0083, -0.0504, -0.1437, 0.0061, -0.0477, -0.1447, 0.0061, -0.0462, -0.1437, 0.0083, -0.0477, -0.1447, 0.0083, -0.0462, -0.1431, 0.0061, -0.0487, -0.1431, 0.0083, -0.0487, -0.1445, 0.01, -0.0526, -0.1445, 0.01, -0.0505, -0.1422, 0.01, -0.0526, -0.1425, 0.01, -0.0503, -0.1436, 0.01, -0.0478, -0.1447, 0.01, -0.0462, -0.143, 0.01, -0.0489, -0.151, 0.0061, -0.0219, -0.151, 0.0061, -0.0201,
-0.151, 0.0083, -0.0219, -0.151, 0.0083, -0.0201, -0.151, 0.01, -0.0219, -0.151, 0.01, -0.0201, -0.1489, 0.0061, -0.0219, -0.1489, 0.0083, -0.0219, -0.1467, 0.0061, -0.0219, -0.1467, 0.0083, -0.0219, -0.1489, 0.01, -0.0219, -0.1489, 0.01, -0.0201, -0.1467, 0.01, -0.0219, -0.1467, 0.01, -0.0201, -0.1445, 0.0061, -0.0219, -0.1445, 0.0083, -0.0219, -0.1424, 0.0061, -0.0219, -0.1424, 0.0083, -0.0219, -0.1445, 0.01, -0.0219, -0.1445, 0.01, -0.0201, -0.1424, 0.01, -0.0219, -0.1424, 0.01, -0.0201, -0.1402, 0.0061, -0.0219, -0.1402, 0.0083, -0.0219, -0.138, 0.0061, -0.0219, -0.138, 0.0083, -0.0219, -0.1402, 0.01, -0.0219, -0.1402, 0.01, -0.0201, -0.138, 0.01, -0.0219, -0.138, 0.01, -0.0201, -0.151, 0.0061, -0.0179, -0.151, 0.0061, -0.0157,
-0.151, 0.0083, -0.0179, -0.151, 0.0083, -0.0157, -0.151, 0.0061, -0.0136, -0.151, 0.0061, -0.0114, -0.151, 0.0083, -0.0136, -0.151, 0.0083, -0.0114, -0.151, 0.01, -0.0179, -0.151, 0.01, -0.0157, -0.151, 0.01, -0.0136, -0.151, 0.01, -0.0114, -0.1489, 0.01, -0.0179, -0.1489, 0.01, -0.0157, -0.1467, 0.01, -0.0179, -0.1467, 0.01, -0.0157, -0.1489, 0.01, -0.0136, -0.1489, 0.01, -0.0114, -0.1467, 0.01, -0.0136, -0.1467, 0.01, -0.0114, -0.151, 0.0061, -0.0092, -0.151, 0.0061, -0.0071, -0.151, 0.0083, -0.0092, -0.151, 0.0083, -0.0071, -0.151, 0.0061, -0.0049, -0.151, 0.0061, -0.0027, -0.151, 0.0083, -0.0049, -0.151, 0.0083, -0.0027, -0.151, 0.01, -0.0092, -0.151, 0.01, -0.0071, -0.151, 0.01, -0.0049, -0.151, 0.01, -0.0027,
-0.1489, 0.01, -0.0092, -0.1489, 0.01, -0.0071, -0.1467, 0.01, -0.0092, -0.1467, 0.01, -0.0071, -0.1489, 0.01, -0.0049, -0.1489, 0.01, -0.0027, -0.1467, 0.01, -0.0049, -0.1467, 0.01, -0.0027, -0.1445, 0.01, -0.0179, -0.1445, 0.01, -0.0157, -0.1424, 0.01, -0.0179, -0.1424, 0.01, -0.0157, -0.1445, 0.01, -0.0136, -0.1445, 0.01, -0.0114, -0.1424, 0.01, -0.0136, -0.1424, 0.01, -0.0114, -0.1402, 0.01, -0.0179, -0.1402, 0.01, -0.0157, -0.138, 0.01, -0.0179, -0.138, 0.01, -0.0157, -0.1402, 0.01, -0.0136, -0.1402, 0.01, -0.0114, -0.138, 0.01, -0.0136, -0.138, 0.01, -0.0114, -0.1445, 0.01, -0.0092, -0.1445, 0.01, -0.0071, -0.1424, 0.01, -0.0092, -0.1424, 0.01, -0.0071, -0.1445, 0.01, -0.0049, -0.1445, 0.01, -0.0027,
-0.1424, 0.01, -0.0049, -0.1424, 0.01, -0.0027, -0.1376, 0.0061, -0.0086, -0.1387, 0.0061, -0.0072, -0.1376, 0.0083, -0.0086, -0.1387, 0.0083, -0.0072, -0.1391, 0.0061, -0.0043, -0.1392, 0.0061, -0.0027, -0.1391, 0.0083, -0.0043, -0.1392, 0.0083, -0.0027, -0.139, 0.0061, -0.0054, -0.139, 0.0083, -0.0054, -0.1402, 0.01, -0.0092, -0.1402, 0.01, -0.0071, -0.1375, 0.01, -0.0087, -0.1387, 0.01, -0.0072, -0.1391, 0.01, -0.0048, -0.1392, 0.01, -0.0027, -0.139, 0.01, -0.0056, -0.1225, 0.0061, -0.067, -0.1225, 0.0083, -0.067, -0.1207, 0.0061, -0.067, -0.1207, 0.0083, -0.067, -0.1225, 0.0061, -0.0656, -0.1225, 0.0061, -0.0635, -0.1225, 0.0083, -0.0656, -0.1225, 0.0083, -0.0635, -0.1225, 0.01, -0.067, -0.1207, 0.01, -0.067, -0.1225, 0.01, -0.0656,
-0.1225, 0.01, -0.0635, -0.1207, 0.01, -0.0656, -0.1207, 0.01, -0.0635, -0.1225, 0.0061, -0.0613, -0.1225, 0.0061, -0.0591, -0.1225, 0.0083, -0.0613, -0.1225, 0.0083, -0.0591, -0.1225, 0.0061, -0.057, -0.1225, 0.0061, -0.0548, -0.1225, 0.0083, -0.057, -0.1225, 0.0083, -0.0548, -0.1225, 0.01, -0.0613, -0.1225, 0.01, -0.0591, -0.1207, 0.01, -0.0613, -0.1207, 0.01, -0.0591, -0.1225, 0.01, -0.057, -0.1225, 0.01, -0.0548, -0.1207, 0.01, -0.057, -0.1207, 0.01, -0.0548, -0.1225, 0.0061, -0.0526, -0.1225, 0.0061, -0.0505, -0.1225, 0.0083, -0.0526, -0.1225, 0.0083, -0.0505, -0.1225, 0.0061, -0.0483, -0.1225, 0.0061, -0.0461, -0.1225, 0.0083, -0.0483, -0.1225, 0.0083, -0.0461, -0.1225, 0.01, -0.0526, -0.1225, 0.01, -0.0505, -0.1207, 0.01, -0.0526,
-0.1207, 0.01, -0.0505, -0.1225, 0.01, -0.0483, -0.1225, 0.01, -0.0461, -0.1207, 0.01, -0.0483, -0.1207, 0.01, -0.0461, -0.1225, 0.0061, -0.0439, -0.1225, 0.0061, -0.0425, -0.1225, 0.0083, -0.0439, -0.1225, 0.0083, -0.0425, -0.1207, 0.0061, -0.0425, -0.1207, 0.0083, -0.0425, -0.1225, 0.01, -0.0439, -0.1225, 0.01, -0.0425, -0.1207, 0.01, -0.0439, -0.1207, 0.01, -0.0425, -0.1185, 0.0061, -0.067, -0.1185, 0.0083, -0.067, -0.1163, 0.0061, -0.067, -0.1163, 0.0083, -0.067, -0.1186, 0.0061, -0.0631, -0.1186, 0.0083, -0.0631, -0.1163, 0.0061, -0.0631, -0.1163, 0.0083, -0.0631, -0.1185, 0.01, -0.067, -0.1163, 0.01, -0.067, -0.1185, 0.01, -0.0656, -0.1186, 0.01, -0.0631, -0.1163, 0.01, -0.0656, -0.1163, 0.01, -0.0631, -0.1141, 0.0061, -0.067,
-0.1141, 0.0083, -0.067, -0.112, 0.0061, -0.067, -0.112, 0.0083, -0.067, -0.1141, 0.0061, -0.0631, -0.1141, 0.0083, -0.0631, -0.112, 0.0061, -0.063, -0.112, 0.0083, -0.063, -0.1141, 0.01, -0.067, -0.112, 0.01, -0.067, -0.1141, 0.01, -0.0656, -0.1141, 0.01, -0.0631, -0.112, 0.01, -0.0656, -0.112, 0.01, -0.063, -0.1186, 0.0061, -0.0613, -0.1186, 0.0061, -0.0591, -0.1186, 0.0083, -0.0613, -0.1186, 0.0083, -0.0591, -0.1186, 0.0061, -0.057, -0.1186, 0.0061, -0.0548, -0.1186, 0.0083, -0.057, -0.1186, 0.0083, -0.0548, -0.1186, 0.01, -0.0613, -0.1186, 0.01, -0.0591, -0.1186, 0.01, -0.057, -0.1186, 0.01, -0.0548, -0.1098, 0.0061, -0.067, -0.1098, 0.0083, -0.067, -0.1076, 0.0061, -0.0669, -0.1076, 0.0083, -0.0669, -0.1098, 0.0061, -0.063,
-0.1098, 0.0083, -0.063, -0.1066, 0.0061, -0.0667, -0.1076, 0.0061, -0.0629, -0.1066, 0.0083, -0.0667, -0.1076, 0.0083, -0.0629, -0.1098, 0.01, -0.067, -0.108, 0.01, -0.0669, -0.1098, 0.01, -0.0656, -0.1098, 0.01, -0.063, -0.1072, 0.01, -0.0669, -0.1078, 0.01, -0.0629, -0.1054, 0.0061, -0.0666, -0.1061, 0.0061, -0.0625, -0.1054, 0.0083, -0.0666, -0.1061, 0.0083, -0.0625, -0.1033, 0.0061, -0.0658, -0.1033, 0.0083, -0.0658, -0.1052, 0.01, -0.0665, -0.1058, 0.01, -0.0624, -0.1031, 0.01, -0.0658, -0.1033, 0.01, -0.0635, -0.105, 0.0061, -0.0622, -0.105, 0.0083, -0.0622, -0.103, 0.0061, -0.0613, -0.103, 0.0083, -0.0613, -0.1048, 0.01, -0.0622, -0.1034, 0.01, -0.0616, -0.1186, 0.0061, -0.0526, -0.1186, 0.0061, -0.0505, -0.1186, 0.0083, -0.0526,
-0.1186, 0.0083, -0.0505, -0.1186, 0.0061, -0.0483, -0.1186, 0.0061, -0.0454, -0.1186, 0.0083, -0.0483, -0.1186, 0.0083, -0.0454, -0.1163, 0.0061, -0.0454, -0.1163, 0.0083, -0.0454, -0.1186, 0.01, -0.0526, -0.1186, 0.01, -0.0505, -0.1186, 0.01, -0.0483, -0.1186, 0.01, -0.0454, -0.1163, 0.01, -0.0454, -0.1141, 0.0061, -0.0454, -0.1141, 0.0083, -0.0454, -0.112, 0.0061, -0.0454, -0.112, 0.0083, -0.0454, -0.1141, 0.01, -0.0454, -0.112, 0.01, -0.0454, -0.1185, 0.0061, -0.0425, -0.1185, 0.0083, -0.0425, -0.1163, 0.0061, -0.0425, -0.1163, 0.0083, -0.0425, -0.1185, 0.01, -0.0439, -0.1185, 0.01, -0.0425, -0.1163, 0.01, -0.0439, -0.1163, 0.01, -0.0425, -0.1141, 0.0061, -0.0425, -0.1141, 0.0083, -0.0425, -0.112, 0.0061, -0.0425, -0.112, 0.0083, -0.0425,
-0.1141, 0.01, -0.0439, -0.1141, 0.01, -0.0425, -0.112, 0.01, -0.0439, -0.112, 0.01, -0.0425, -0.1098, 0.0061, -0.0455, -0.1098, 0.0083, -0.0455, -0.1076, 0.0061, -0.0459, -0.1076, 0.0083, -0.0459, -0.1098, 0.01, -0.0455, -0.1077, 0.01, -0.0459, -0.1054, 0.0061, -0.0465, -0.1054, 0.0083, -0.0465, -0.1031, 0.0061, -0.0478, -0.1043, 0.0061, -0.0471, -0.1031, 0.0083, -0.0478, -0.1043, 0.0083, -0.0471, -0.1057, 0.01, -0.0464, -0.103, 0.01, -0.0479, -0.104, 0.01, -0.0473, -0.1098, 0.0061, -0.0425, -0.1098, 0.0083, -0.0425, -0.1076, 0.0061, -0.0425, -0.1076, 0.0083, -0.0425, -0.1098, 0.01, -0.0439, -0.1098, 0.01, -0.0425, -0.1076, 0.01, -0.0439, -0.1076, 0.01, -0.0425, -0.1054, 0.0061, -0.0426, -0.1054, 0.0083, -0.0426, -0.1031, 0.0061, -0.0431,
-0.1042, 0.0061, -0.0428, -0.1031, 0.0083, -0.0431, -0.1042, 0.0083, -0.0428, -0.1055, 0.01, -0.0439, -0.1054, 0.01, -0.0426, -0.1032, 0.01, -0.0431, -0.1043, 0.01, -0.0428, -0.1358, 0.0061, -0.0219, -0.1358, 0.0083, -0.0219, -0.1337, 0.0061, -0.0219, -0.1337, 0.0083, -0.0219, -0.1358, 0.01, -0.0219, -0.1358, 0.01, -0.0201, -0.1337, 0.01, -0.0219, -0.1337, 0.01, -0.0201, -0.1315, 0.0061, -0.0219, -0.1315, 0.0083, -0.0219, -0.1293, 0.0061, -0.0219, -0.1293, 0.0083, -0.0219, -0.1315, 0.01, -0.0219, -0.1315, 0.01, -0.0201, -0.1293, 0.01, -0.0219, -0.1293, 0.01, -0.0201, -0.1272, 0.0061, -0.0219, -0.1272, 0.0083, -0.0219, -0.125, 0.0061, -0.0219, -0.125, 0.0083, -0.0219, -0.1272, 0.01, -0.0219, -0.1272, 0.01, -0.0201, -0.125, 0.01, -0.0219,
-0.125, 0.01, -0.0201, -0.1228, 0.0061, -0.0219, -0.1228, 0.0083, -0.0219, -0.1207, 0.0061, -0.0219, -0.1207, 0.0083, -0.0219, -0.1228, 0.01, -0.0219, -0.1228, 0.01, -0.0201, -0.1207, 0.01, -0.0219, -0.1207, 0.01, -0.0201, -0.1358, 0.01, -0.0179, -0.1358, 0.01, -0.0157, -0.1337, 0.01, -0.0179, -0.1337, 0.01, -0.0157, -0.1358, 0.01, -0.0136, -0.1358, 0.01, -0.0114, -0.1337, 0.01, -0.0136, -0.1337, 0.01, -0.0114, -0.1315, 0.01, -0.0179, -0.1315, 0.01, -0.0157, -0.1293, 0.01, -0.0179, -0.1293, 0.01, -0.0157, -0.1315, 0.01, -0.0136, -0.1315, 0.01, -0.0114, -0.1293, 0.01, -0.0136, -0.1293, 0.01, -0.0114, -0.1358, 0.0061, -0.0097, -0.1358, 0.0083, -0.0097, -0.1337, 0.0061, -0.01, -0.1337, 0.0083, -0.01, -0.1358, 0.01, -0.0097,
-0.1337, 0.01, -0.01, -0.1315, 0.0061, -0.0101, -0.1315, 0.0083, -0.0101, -0.1293, 0.0061, -0.0101, -0.1293, 0.0083, -0.0101, -0.1315, 0.01, -0.0101, -0.1293, 0.01, -0.0101, -0.1272, 0.01, -0.0179, -0.1272, 0.01, -0.0157, -0.125, 0.01, -0.0179, -0.125, 0.01, -0.0157, -0.1272, 0.01, -0.0136, -0.1272, 0.01, -0.0114, -0.125, 0.01, -0.0136, -0.125, 0.01, -0.0114, -0.1228, 0.01, -0.0179, -0.1228, 0.01, -0.0157, -0.1207, 0.01, -0.0179, -0.1207, 0.01, -0.0157, -0.1228, 0.01, -0.0136, -0.1228, 0.01, -0.0114, -0.1207, 0.01, -0.0136, -0.1207, 0.01, -0.0114, -0.1272, 0.0061, -0.0101, -0.1272, 0.0083, -0.0101, -0.125, 0.0061, -0.0101, -0.125, 0.0083, -0.0101, -0.1272, 0.01, -0.0101, -0.125, 0.01, -0.0101, -0.1228, 0.0061, -0.0101,
-0.1228, 0.0083, -0.0101, -0.1207, 0.0061, -0.0101, -0.1207, 0.0083, -0.0101, -0.1228, 0.01, -0.0101, -0.1207, 0.01, -0.0101, -0.1185, 0.0061, -0.0219, -0.1185, 0.0083, -0.0219, -0.1163, 0.0061, -0.0219, -0.1163, 0.0083, -0.0219, -0.1185, 0.01, -0.0219, -0.1185, 0.01, -0.0201, -0.1163, 0.01, -0.0219, -0.1163, 0.01, -0.0201, -0.1141, 0.0061, -0.0219, -0.1141, 0.0083, -0.0219, -0.112, 0.0061, -0.0219, -0.112, 0.0083, -0.0219, -0.1141, 0.01, -0.0219, -0.1141, 0.01, -0.0201, -0.112, 0.01, -0.0219, -0.112, 0.01, -0.0201, -0.1098, 0.0061, -0.0219, -0.1098, 0.0083, -0.0219, -0.1076, 0.0061, -0.0219, -0.1076, 0.0083, -0.0219, -0.1098, 0.01, -0.0219, -0.1098, 0.01, -0.0201, -0.1076, 0.01, -0.0219, -0.1076, 0.01, -0.0201, -0.1055, 0.0061, -0.0219,
-0.1055, 0.0083, -0.0219, -0.1033, 0.0061, -0.0219, -0.1033, 0.0083, -0.0219, -0.1055, 0.01, -0.0219, -0.1055, 0.01, -0.0201, -0.1033, 0.01, -0.0219, -0.1033, 0.01, -0.0201, -0.1185, 0.01, -0.0179, -0.1185, 0.01, -0.0157, -0.1163, 0.01, -0.0179, -0.1163, 0.01, -0.0157, -0.1185, 0.01, -0.0136, -0.1185, 0.01, -0.0114, -0.1163, 0.01, -0.0136, -0.1163, 0.01, -0.0114, -0.1141, 0.01, -0.0179, -0.1141, 0.01, -0.0157, -0.112, 0.01, -0.0179, -0.112, 0.01, -0.0157, -0.1141, 0.01, -0.0136, -0.1141, 0.01, -0.0114, -0.112, 0.01, -0.0136, -0.112, 0.01, -0.0114, -0.1185, 0.0061, -0.0101, -0.1185, 0.0083, -0.0101, -0.1163, 0.0061, -0.0101, -0.1163, 0.0083, -0.0101, -0.1185, 0.01, -0.0101, -0.1163, 0.01, -0.0101, -0.1141, 0.0061, -0.0101,
-0.1141, 0.0083, -0.0101, -0.112, 0.0061, -0.0101, -0.112, 0.0083, -0.0101, -0.1141, 0.01, -0.0101, -0.112, 0.01, -0.0101, -0.1098, 0.01, -0.0179, -0.1098, 0.01, -0.0157, -0.1076, 0.01, -0.0179, -0.1076, 0.01, -0.0157, -0.1098, 0.01, -0.0136, -0.1098, 0.01, -0.0114, -0.1076, 0.01, -0.0136, -0.1076, 0.01, -0.0114, -0.1055, 0.01, -0.0179, -0.1055, 0.01, -0.0157, -0.1033, 0.01, -0.0179, -0.1033, 0.01, -0.0157, -0.1055, 0.01, -0.0136, -0.1055, 0.01, -0.0114, -0.1033, 0.01, -0.0136, -0.1033, 0.01, -0.0114, -0.1098, 0.0061, -0.0101, -0.1098, 0.0083, -0.0101, -0.1076, 0.0061, -0.0101, -0.1076, 0.0083, -0.0101, -0.1098, 0.01, -0.0101, -0.1076, 0.01, -0.0101, -0.1055, 0.0061, -0.0101, -0.1055, 0.0083, -0.0101, -0.1033, 0.0061, -0.0101,
-0.1033, 0.0083, -0.0101, -0.1055, 0.01, -0.0101, -0.1033, 0.01, -0.0101, -0.1016, 0.0061, -0.065, -0.1005, 0.0061, -0.0642, -0.1016, 0.0083, -0.065, -0.1005, 0.0083, -0.0642, -0.0994, 0.0061, -0.0631, -0.0994, 0.0083, -0.0631, -0.1017, 0.01, -0.065, -0.1006, 0.01, -0.0642, -0.0994, 0.01, -0.0631, -0.102, 0.0061, -0.0604, -0.1007, 0.0061, -0.0591, -0.102, 0.0083, -0.0604, -0.1007, 0.0083, -0.0591, -0.0983, 0.0061, -0.0618, -0.0983, 0.0083, -0.0618, -0.1002, 0.0061, -0.0577, -0.1002, 0.0083, -0.0577, -0.0997, 0.0061, -0.0566, -0.0995, 0.0061, -0.0548, -0.0997, 0.0083, -0.0566, -0.0995, 0.0083, -0.0548, -0.102, 0.01, -0.0604, -0.1007, 0.01, -0.059, -0.0983, 0.01, -0.0618, -0.099, 0.01, -0.0591, -0.1002, 0.01, -0.0578, -0.0998, 0.01, -0.0569,
-0.0995, 0.01, -0.0548, -0.0976, 0.0061, -0.0607, -0.0968, 0.0061, -0.0592, -0.0976, 0.0083, -0.0607, -0.0968, 0.0083, -0.0592, -0.0962, 0.0061, -0.057, -0.0961, 0.0061, -0.0548, -0.0962, 0.0083, -0.057, -0.0961, 0.0083, -0.0548, -0.0975, 0.01, -0.0606, -0.0968, 0.01, -0.0594, -0.0962, 0.01, -0.0571, -0.0961, 0.01, -0.0548, -0.1002, 0.0061, -0.0518, -0.1008, 0.0061, -0.0504, -0.1002, 0.0083, -0.0518, -0.1008, 0.0083, -0.0504, -0.0997, 0.0061, -0.0529, -0.0997, 0.0083, -0.0529, -0.1019, 0.0061, -0.049, -0.1019, 0.0083, -0.049, -0.0979, 0.0061, -0.0473, -0.0988, 0.0061, -0.046, -0.0979, 0.0083, -0.0473, -0.0988, 0.0083, -0.046, -0.1002, 0.01, -0.0518, -0.1006, 0.01, -0.0506, -0.0998, 0.01, -0.0527, -0.099, 0.01, -0.0505, -0.1019, 0.01, -0.049,
-0.1011, 0.01, -0.0461, -0.0978, 0.01, -0.0474, -0.099, 0.01, -0.0457, -0.0961, 0.0061, -0.0526, -0.0964, 0.0061, -0.0504, -0.0961, 0.0083, -0.0526, -0.0964, 0.0083, -0.0504, -0.0972, 0.0061, -0.0483, -0.0972, 0.0083, -0.0483, -0.0961, 0.01, -0.0526, -0.0965, 0.01, -0.0503, -0.0971, 0.01, -0.0485, -0.101, 0.0061, -0.0441, -0.101, 0.0083, -0.0441, -0.1, 0.0061, -0.045, -0.1, 0.0083, -0.045, -0.1009, 0.01, -0.0442, -0.1, 0.01, -0.045, -0.0766, 0.0061, -0.0589, -0.0766, 0.0083, -0.0589, -0.0771, 0.0061, -0.057, -0.0774, 0.0061, -0.0548, -0.0771, 0.0083, -0.057, -0.0774, 0.0083, -0.0548, -0.0766, 0.01, -0.0587, -0.0771, 0.01, -0.057, -0.0774, 0.01, -0.0548, -0.0742, 0.0061, -0.0627, -0.0742, 0.0083, -0.0627, -0.072, 0.0061, -0.0647,
-0.0731, 0.0061, -0.0638, -0.072, 0.0083, -0.0647, -0.0731, 0.0083, -0.0638, -0.0742, 0.01, -0.0626, -0.072, 0.01, -0.0647, -0.0732, 0.01, -0.0638, -0.0708, 0.0061, -0.0655, -0.0708, 0.0083, -0.0655, -0.0686, 0.0061, -0.0664, -0.0676, 0.0061, -0.0624, -0.0686, 0.0083, -0.0664, -0.0676, 0.0083, -0.0624, -0.071, 0.01, -0.0654, -0.0707, 0.01, -0.0635, -0.0688, 0.01, -0.0663, -0.0681, 0.01, -0.0623, -0.0753, 0.0061, -0.0614, -0.0761, 0.0061, -0.06, -0.0753, 0.0083, -0.0614, -0.0761, 0.0083, -0.06, -0.0723, 0.0061, -0.0587, -0.0723, 0.0083, -0.0587, -0.0729, 0.0061, -0.057, -0.0733, 0.0061, -0.0548, -0.0729, 0.0083, -0.057, -0.0733, 0.0083, -0.0548, -0.0751, 0.01, -0.0617, -0.0762, 0.01, -0.0597, -0.0729, 0.01, -0.0613, -0.0722, 0.01, -0.0588,
-0.0751, 0.01, -0.057, -0.0751, 0.01, -0.0548, -0.073, 0.01, -0.0568, -0.0733, 0.01, -0.0548, -0.0707, 0.0061, -0.0608, -0.0715, 0.0061, -0.0598, -0.0707, 0.0083, -0.0608, -0.0715, 0.0083, -0.0598, -0.0687, 0.0061, -0.062, -0.0687, 0.0083, -0.062, -0.0705, 0.01, -0.061, -0.0714, 0.01, -0.0598, -0.0689, 0.01, -0.0619, -0.0773, 0.0061, -0.0526, -0.0769, 0.0061, -0.0504, -0.0773, 0.0083, -0.0526, -0.0769, 0.0083, -0.0504, -0.0763, 0.0061, -0.049, -0.0763, 0.0083, -0.049, -0.0773, 0.01, -0.0526, -0.0769, 0.01, -0.0503, -0.0764, 0.01, -0.0491, -0.0732, 0.0061, -0.0526, -0.0727, 0.0061, -0.0508, -0.0732, 0.0083, -0.0526, -0.0727, 0.0083, -0.0508, -0.0758, 0.0061, -0.0479, -0.0747, 0.0061, -0.0464, -0.0758, 0.0083, -0.0479, -0.0747, 0.0083, -0.0464,
-0.0736, 0.0061, -0.0453, -0.0736, 0.0083, -0.0453, -0.0751, 0.01, -0.0526, -0.0751, 0.01, -0.0505, -0.0732, 0.01, -0.0527, -0.0725, 0.01, -0.0505, -0.0758, 0.01, -0.048, -0.0748, 0.01, -0.0464, -0.0729, 0.01, -0.0483, -0.0736, 0.01, -0.0453, -0.0717, 0.0061, -0.0495, -0.0717, 0.0083, -0.0495, -0.0707, 0.0061, -0.0482, -0.0707, 0.0083, -0.0482, -0.0694, 0.0061, -0.0474, -0.0683, 0.0061, -0.0468, -0.0694, 0.0083, -0.0474, -0.0683, 0.0083, -0.0468, -0.0717, 0.01, -0.0495, -0.0708, 0.01, -0.0483, -0.0707, 0.01, -0.0461, -0.0694, 0.01, -0.0474, -0.0684, 0.01, -0.0468, -0.0726, 0.0061, -0.0445, -0.0726, 0.0083, -0.0445, -0.0725, 0.01, -0.0445, -0.0708, 0.0061, -0.0435, -0.0708, 0.0083, -0.0435, -0.0692, 0.0061, -0.043, -0.0681, 0.0061, -0.0427,
-0.0692, 0.0083, -0.043, -0.0681, 0.0083, -0.0427, -0.0711, 0.01, -0.0437, -0.0689, 0.01, -0.0429, -0.0679, 0.01, -0.0427, -0.1011, 0.0061, -0.0219, -0.1011, 0.0083, -0.0219, -0.099, 0.0061, -0.0219, -0.099, 0.0083, -0.0219, -0.1011, 0.01, -0.0219, -0.1011, 0.01, -0.0201, -0.099, 0.01, -0.0219, -0.099, 0.01, -0.0201, -0.0968, 0.0061, -0.0219, -0.0968, 0.0083, -0.0219, -0.0946, 0.0061, -0.0219, -0.0946, 0.0083, -0.0219, -0.0968, 0.01, -0.0219, -0.0968, 0.01, -0.0201, -0.0946, 0.01, -0.0219, -0.0946, 0.01, -0.0201, -0.0924, 0.0061, -0.0219, -0.0924, 0.0083, -0.0219, -0.0902, 0.0061, -0.0219, -0.0902, 0.0061, -0.0201, -0.0902, 0.0083, -0.0219, -0.0902, 0.0083, -0.0201, -0.0924, 0.01, -0.0219, -0.0924, 0.01, -0.0201, -0.0902, 0.01, -0.0219,
-0.0902, 0.01, -0.0201, -0.1011, 0.01, -0.0179, -0.1011, 0.01, -0.0157, -0.099, 0.01, -0.0179, -0.099, 0.01, -0.0157, -0.1011, 0.01, -0.0136, -0.1011, 0.01, -0.0114, -0.099, 0.01, -0.0136, -0.099, 0.01, -0.0114, -0.0968, 0.01, -0.0179, -0.0968, 0.01, -0.0157, -0.0946, 0.01, -0.0179, -0.0946, 0.01, -0.0157, -0.0968, 0.01, -0.0136, -0.0968, 0.01, -0.0114, -0.0946, 0.01, -0.0136, -0.0946, 0.01, -0.0114, -0.1011, 0.0061, -0.0101, -0.1011, 0.0083, -0.0101, -0.099, 0.0061, -0.0101, -0.099, 0.0083, -0.0101, -0.1011, 0.01, -0.0101, -0.099, 0.01, -0.0101, -0.0968, 0.0061, -0.0101, -0.0968, 0.0083, -0.0101, -0.0946, 0.0061, -0.0101, -0.0946, 0.0083, -0.0101, -0.0968, 0.01, -0.0101, -0.0946, 0.01, -0.0101, -0.0902, 0.0061, -0.0179,
-0.0902, 0.0061, -0.0157, -0.0902, 0.0083, -0.0179, -0.0902, 0.0083, -0.0157, -0.0902, 0.0061, -0.0136, -0.0902, 0.0061, -0.0114, -0.0902, 0.0083, -0.0136, -0.0902, 0.0083, -0.0114, -0.0924, 0.01, -0.0179, -0.0924, 0.01, -0.0157, -0.0902, 0.01, -0.0179, -0.0902, 0.01, -0.0157, -0.0924, 0.01, -0.0136, -0.0924, 0.01, -0.0114, -0.0902, 0.01, -0.0136, -0.0902, 0.01, -0.0114, -0.0924, 0.0061, -0.0101, -0.0924, 0.0083, -0.0101, -0.0902, 0.0061, -0.0101, -0.0902, 0.0083, -0.0101, -0.0924, 0.01, -0.0101, -0.0902, 0.01, -0.0101, -0.0775, 0.0061, -0.0219, -0.0775, 0.0061, -0.0201, -0.0775, 0.0083, -0.0219, -0.0775, 0.0083, -0.0201, -0.0775, 0.01, -0.0219, -0.0775, 0.01, -0.0201, -0.0751, 0.0061, -0.0219, -0.0751, 0.0083, -0.0219, -0.0729, 0.0061, -0.0219,
-0.0729, 0.0083, -0.0219, -0.0751, 0.01, -0.0219, -0.0751, 0.01, -0.0201, -0.0729, 0.01, -0.0219, -0.0729, 0.01, -0.0201, -0.0707, 0.0061, -0.0219, -0.0707, 0.0083, -0.0219, -0.0686, 0.0061, -0.0219, -0.0686, 0.0083, -0.0219, -0.0707, 0.01, -0.0219, -0.0707, 0.01, -0.0201, -0.0686, 0.01, -0.0219, -0.0686, 0.01, -0.0201, -0.0775, 0.0061, -0.0179, -0.0775, 0.0061, -0.0157, -0.0775, 0.0083, -0.0179, -0.0775, 0.0083, -0.0157, -0.0775, 0.0061, -0.0136, -0.0775, 0.0061, -0.0114, -0.0775, 0.0083, -0.0136, -0.0775, 0.0083, -0.0114, -0.0775, 0.01, -0.0179, -0.0775, 0.01, -0.0157, -0.0775, 0.01, -0.0136, -0.0775, 0.01, -0.0114, -0.0775, 0.0061, -0.0092, -0.0775, 0.0061, -0.0071, -0.0775, 0.0083, -0.0092, -0.0775, 0.0083, -0.0071, -0.0775, 0.0061, -0.0049,
-0.0775, 0.0061, -0.0027, -0.0775, 0.0083, -0.0049, -0.0775, 0.0083, -0.0027, -0.0775, 0.01, -0.0092, -0.0775, 0.01, -0.0071, -0.0775, 0.01, -0.0049, -0.0775, 0.01, -0.0027, -0.0751, 0.01, -0.0179, -0.0751, 0.01, -0.0157, -0.0729, 0.01, -0.0179, -0.0729, 0.01, -0.0157, -0.0751, 0.01, -0.0136, -0.0751, 0.01, -0.0114, -0.0729, 0.01, -0.0136, -0.0729, 0.01, -0.0114, -0.0707, 0.01, -0.0179, -0.0707, 0.01, -0.0157, -0.0686, 0.01, -0.0179, -0.0686, 0.01, -0.0157, -0.0707, 0.01, -0.0136, -0.0707, 0.01, -0.0114, -0.0686, 0.01, -0.0136, -0.0686, 0.01, -0.0114, -0.0751, 0.01, -0.0092, -0.0751, 0.01, -0.0071, -0.0729, 0.01, -0.0092, -0.0729, 0.01, -0.0071, -0.0751, 0.01, -0.0049, -0.0751, 0.01, -0.0027, -0.0729, 0.01, -0.0049,
-0.0729, 0.01, -0.0027, -0.0707, 0.01, -0.0092, -0.0707, 0.01, -0.0071, -0.0686, 0.01, -0.0092, -0.0686, 0.01, -0.0071, -0.0707, 0.01, -0.0049, -0.0707, 0.01, -0.0027, -0.0686, 0.01, -0.0049, -0.0686, 0.01, -0.0027, -0.0661, 0.0061, -0.0669, -0.0661, 0.0083, -0.0669, -0.0642, 0.0061, -0.067, -0.0642, 0.0083, -0.067, -0.0672, 0.0061, -0.0667, -0.0664, 0.0061, -0.0627, -0.0672, 0.0083, -0.0667, -0.0664, 0.0083, -0.0627, -0.0642, 0.0061, -0.063, -0.0642, 0.0083, -0.063, -0.0659, 0.01, -0.0669, -0.0642, 0.01, -0.067, -0.0667, 0.01, -0.0668, -0.0663, 0.01, -0.0628, -0.0642, 0.01, -0.0656, -0.0642, 0.01, -0.063, -0.0621, 0.0061, -0.067, -0.0621, 0.0083, -0.067, -0.0599, 0.0061, -0.067, -0.0599, 0.0083, -0.067, -0.0621, 0.0061, -0.063,
-0.0621, 0.0083, -0.063, -0.0599, 0.0061, -0.0631, -0.0599, 0.0083, -0.0631, -0.0621, 0.01, -0.067, -0.0599, 0.01, -0.067, -0.0621, 0.01, -0.0656, -0.062, 0.01, -0.063, -0.0599, 0.01, -0.0656, -0.0599, 0.01, -0.0631, -0.0577, 0.0061, -0.0669, -0.0577, 0.0083, -0.0669, -0.0565, 0.0061, -0.0667, -0.0565, 0.0083, -0.0667, -0.0577, 0.0061, -0.063, -0.0577, 0.0083, -0.063, -0.0554, 0.0061, -0.0666, -0.0559, 0.0061, -0.0626, -0.0554, 0.0083, -0.0666, -0.0559, 0.0083, -0.0626, -0.0577, 0.01, -0.0669, -0.0566, 0.01, -0.0668, -0.0577, 0.01, -0.0656, -0.0578, 0.01, -0.063, -0.0555, 0.01, -0.0666, -0.0557, 0.01, -0.0625, -0.0534, 0.0061, -0.066, -0.0534, 0.0083, -0.066, -0.0514, 0.0061, -0.0651, -0.0503, 0.0061, -0.0645, -0.0514, 0.0083, -0.0651,
-0.0503, 0.0083, -0.0645, -0.0531, 0.01, -0.0659, -0.0534, 0.01, -0.0635, -0.0515, 0.01, -0.0652, -0.0505, 0.01, -0.0646, -0.0548, 0.0061, -0.0623, -0.0548, 0.0083, -0.0623, -0.0547, 0.01, -0.0623, -0.0533, 0.0061, -0.0615, -0.0533, 0.0083, -0.0615, -0.052, 0.0061, -0.0605, -0.0509, 0.0061, -0.0592, -0.052, 0.0083, -0.0605, -0.0509, 0.0083, -0.0592, -0.0502, 0.0061, -0.058, -0.0502, 0.0083, -0.058, -0.0535, 0.01, -0.0617, -0.052, 0.01, -0.0605, -0.0507, 0.01, -0.059, -0.0502, 0.01, -0.058, -0.0664, 0.0061, -0.0461, -0.0664, 0.0083, -0.0461, -0.0642, 0.0061, -0.0456, -0.0642, 0.0083, -0.0456, -0.0663, 0.01, -0.0461, -0.0642, 0.01, -0.0456, -0.0621, 0.0061, -0.0454, -0.0621, 0.0083, -0.0454, -0.0599, 0.0061, -0.0455, -0.0599, 0.0083, -0.0455,
-0.0621, 0.01, -0.0454, -0.0599, 0.01, -0.0455, -0.0664, 0.0061, -0.0425, -0.0664, 0.0083, -0.0425, -0.0642, 0.0061, -0.0425, -0.0642, 0.0083, -0.0425, -0.0664, 0.01, -0.0439, -0.0664, 0.01, -0.0425, -0.0642, 0.01, -0.0439, -0.0642, 0.01, -0.0425, -0.0621, 0.0061, -0.0425, -0.0621, 0.0083, -0.0425, -0.0599, 0.0061, -0.0425, -0.0599, 0.0083, -0.0425, -0.0621, 0.01, -0.0439, -0.0621, 0.01, -0.0425, -0.0599, 0.01, -0.0439, -0.0599, 0.01, -0.0425, -0.0577, 0.0061, -0.0458, -0.0577, 0.0083, -0.0458, -0.0555, 0.0061, -0.0463, -0.0555, 0.0083, -0.0463, -0.0578, 0.01, -0.0457, -0.0557, 0.01, -0.0463, -0.0506, 0.0061, -0.0502, -0.0506, 0.0083, -0.0502, -0.0529, 0.0061, -0.0476, -0.054, 0.0061, -0.047, -0.0529, 0.0083, -0.0476, -0.054, 0.0083, -0.047,
-0.0517, 0.0061, -0.0487, -0.0517, 0.0083, -0.0487, -0.0507, 0.01, -0.0501, -0.0528, 0.01, -0.0477, -0.0538, 0.01, -0.0471, -0.0517, 0.01, -0.0487, -0.0512, 0.01, -0.0461, -0.0577, 0.0061, -0.0425, -0.0577, 0.0083, -0.0425, -0.0555, 0.0061, -0.0426, -0.0555, 0.0083, -0.0426, -0.0577, 0.01, -0.0439, -0.0577, 0.01, -0.0425, -0.0555, 0.01, -0.0439, -0.0555, 0.01, -0.0426, -0.0532, 0.0061, -0.0431, -0.0544, 0.0061, -0.0428, -0.0532, 0.0083, -0.0431, -0.0544, 0.0083, -0.0428, -0.0511, 0.0061, -0.044, -0.0511, 0.0083, -0.044, -0.0533, 0.01, -0.0431, -0.0544, 0.01, -0.0428, -0.0509, 0.01, -0.0442, -0.049, 0.0061, -0.0635, -0.049, 0.0083, -0.0635, -0.049, 0.01, -0.0636, -0.0481, 0.0061, -0.0623, -0.0481, 0.0083, -0.0623, -0.0473, 0.0061, -0.0613,
-0.0465, 0.0061, -0.0596, -0.0473, 0.0083, -0.0613, -0.0465, 0.0083, -0.0596, -0.0497, 0.0061, -0.0569, -0.0492, 0.0061, -0.0547, -0.0497, 0.0083, -0.0569, -0.0492, 0.0083, -0.0547, -0.0461, 0.0061, -0.057, -0.0461, 0.0061, -0.0548, -0.0461, 0.0083, -0.057, -0.0461, 0.0083, -0.0548, -0.048, 0.01, -0.0622, -0.049, 0.01, -0.0591, -0.0472, 0.01, -0.0611, -0.0465, 0.01, -0.0594, -0.0497, 0.01, -0.0569, -0.0492, 0.01, -0.0548, -0.0461, 0.01, -0.0571, -0.0461, 0.01, -0.0548, -0.0496, 0.0061, -0.0526, -0.05, 0.0061, -0.0513, -0.0496, 0.0083, -0.0526, -0.05, 0.0083, -0.0513, -0.0461, 0.0061, -0.0526, -0.0463, 0.0061, -0.0504, -0.0461, 0.0083, -0.0526, -0.0463, 0.0083, -0.0504, -0.0488, 0.0061, -0.0459, -0.0488, 0.0083, -0.0459, -0.047, 0.0061, -0.0482,
-0.0479, 0.0061, -0.0471, -0.047, 0.0083, -0.0482, -0.0479, 0.0083, -0.0471, -0.0495, 0.01, -0.0528, -0.0502, 0.01, -0.051, -0.0461, 0.01, -0.0526, -0.0464, 0.01, -0.0503, -0.049, 0.01, -0.0483, -0.0489, 0.01, -0.0459, -0.0471, 0.01, -0.048, -0.0479, 0.01, -0.0471, -0.05, 0.0061, -0.045, -0.05, 0.0083, -0.045, -0.05, 0.01, -0.0449, -0.0667, 0.0061, -0.0219, -0.0667, 0.0061, -0.0201, -0.0667, 0.0083, -0.0219, -0.0667, 0.0083, -0.0201, -0.0667, 0.01, -0.0219, -0.0667, 0.01, -0.0201, -0.0667, 0.0061, -0.0179, -0.0667, 0.0061, -0.0157, -0.0667, 0.0083, -0.0179, -0.0667, 0.0083, -0.0157, -0.0667, 0.0061, -0.0136, -0.0667, 0.0061, -0.0114, -0.0667, 0.0083, -0.0136, -0.0667, 0.0083, -0.0114, -0.0667, 0.01, -0.0179, -0.0667, 0.01, -0.0157,
-0.0667, 0.01, -0.0136, -0.0667, 0.01, -0.0114, -0.0667, 0.0061, -0.0092, -0.0667, 0.0061, -0.0071, -0.0667, 0.0083, -0.0092, -0.0667, 0.0083, -0.0071, -0.0667, 0.0061, -0.0049, -0.0667, 0.0061, -0.0027, -0.0667, 0.0083, -0.0049, -0.0667, 0.0083, -0.0027, -0.0667, 0.01, -0.0092, -0.0667, 0.01, -0.0071, -0.0667, 0.01, -0.0049, -0.0667, 0.01, -0.0027, -0.05, 0.0061, -0.0219, -0.05, 0.0061, -0.0201, -0.05, 0.0083, -0.0219, -0.05, 0.0083, -0.0201, -0.0469, 0.0061, -0.0219, -0.0469, 0.0083, -0.0219, -0.05, 0.01, -0.0219, -0.05, 0.01, -0.0201, -0.0469, 0.01, -0.0219, -0.0469, 0.01, -0.0201, -0.0447, 0.0061, -0.0219, -0.0447, 0.0083, -0.0219, -0.0425, 0.0061, -0.0219, -0.0425, 0.0083, -0.0219, -0.0447, 0.01, -0.0219, -0.0447, 0.01, -0.0201,
-0.0425, 0.01, -0.0219, -0.0425, 0.01, -0.0201, -0.0404, 0.0061, -0.0219, -0.0404, 0.0083, -0.0219, -0.0382, 0.0061, -0.0219, -0.0382, 0.0061, -0.0201, -0.0382, 0.0083, -0.0219, -0.0382, 0.0083, -0.0201, -0.0404, 0.01, -0.0219, -0.0404, 0.01, -0.0201, -0.0382, 0.01, -0.0219, -0.0382, 0.01, -0.0201, -0.05, 0.0061, -0.0179, -0.05, 0.0061, -0.0157, -0.05, 0.0083, -0.0179, -0.05, 0.0083, -0.0157, -0.05, 0.0061, -0.0136, -0.05, 0.0061, -0.0114, -0.05, 0.0083, -0.0136, -0.05, 0.0083, -0.0114, -0.05, 0.01, -0.0179, -0.05, 0.01, -0.0157, -0.0469, 0.01, -0.0179, -0.0469, 0.01, -0.0157, -0.05, 0.01, -0.0136, -0.05, 0.01, -0.0114, -0.0469, 0.01, -0.0136, -0.0469, 0.01, -0.0114, -0.0447, 0.01, -0.0179, -0.0447, 0.01, -0.0157,
-0.0425, 0.01, -0.0179, -0.0425, 0.01, -0.0157, -0.0447, 0.01, -0.0136, -0.0447, 0.01, -0.0114, -0.0425, 0.01, -0.0136, -0.0425, 0.01, -0.0114, -0.05, 0.0061, -0.0092, -0.05, 0.0061, -0.0071, -0.05, 0.0083, -0.0092, -0.05, 0.0083, -0.0071, -0.05, 0.0061, -0.0049, -0.05, 0.0061, -0.0027, -0.05, 0.0083, -0.0049, -0.05, 0.0083, -0.0027, -0.05, 0.01, -0.0092, -0.05, 0.01, -0.0071, -0.0469, 0.01, -0.0092, -0.0469, 0.01, -0.0071, -0.05, 0.01, -0.0049, -0.05, 0.01, -0.0027, -0.0469, 0.01, -0.0049, -0.0469, 0.01, -0.0027, -0.0447, 0.01, -0.0092, -0.0447, 0.01, -0.0071, -0.0425, 0.01, -0.0092, -0.0425, 0.01, -0.0071, -0.0447, 0.01, -0.0049, -0.0447, 0.01, -0.0027, -0.0425, 0.01, -0.0049, -0.0425, 0.01, -0.0027,
-0.0382, 0.0061, -0.0179, -0.0382, 0.0061, -0.0157, -0.0382, 0.0083, -0.0179, -0.0382, 0.0083, -0.0157, -0.0382, 0.0061, -0.0136, -0.0382, 0.0061, -0.0114, -0.0382, 0.0083, -0.0136, -0.0382, 0.0083, -0.0114, -0.0404, 0.01, -0.0179, -0.0404, 0.01, -0.0157, -0.0382, 0.01, -0.0179, -0.0382, 0.01, -0.0157, -0.0404, 0.01, -0.0136, -0.0404, 0.01, -0.0114, -0.0382, 0.01, -0.0136, -0.0382, 0.01, -0.0114, -0.0382, 0.0061, -0.0092, -0.0382, 0.0061, -0.0071, -0.0382, 0.0083, -0.0092, -0.0382, 0.0083, -0.0071, -0.0382, 0.0061, -0.0049, -0.0382, 0.0061, -0.0027, -0.0382, 0.0083, -0.0049, -0.0382, 0.0083, -0.0027, -0.0404, 0.01, -0.0092, -0.0404, 0.01, -0.0071, -0.0382, 0.01, -0.0092, -0.0382, 0.01, -0.0071, -0.0404, 0.01, -0.0049, -0.0404, 0.01, -0.0027,
-0.0382, 0.01, -0.0049, -0.0382, 0.01, -0.0027, -0.0265, 0.0061, -0.067, -0.0265, 0.0083, -0.067, -0.0252, 0.0061, -0.067, -0.0252, 0.0083, -0.067, -0.0265, 0.0061, -0.0656, -0.0265, 0.0061, -0.0635, -0.0265, 0.0083, -0.0656, -0.0265, 0.0083, -0.0635, -0.0265, 0.01, -0.067, -0.0252, 0.01, -0.067, -0.0265, 0.01, -0.0656, -0.0265, 0.01, -0.0635, -0.0252, 0.01, -0.0656, -0.0252, 0.01, -0.0635, -0.0265, 0.0061, -0.0613, -0.0265, 0.0061, -0.0591, -0.0265, 0.0083, -0.0613, -0.0265, 0.0083, -0.0591, -0.0265, 0.0061, -0.057, -0.0265, 0.0061, -0.0548, -0.0265, 0.0083, -0.057, -0.0265, 0.0083, -0.0548, -0.0265, 0.01, -0.0613, -0.0265, 0.01, -0.0591, -0.0252, 0.01, -0.0613, -0.0252, 0.01, -0.0591, -0.0265, 0.01, -0.057, -0.0265, 0.01, -0.0548,
-0.0252, 0.01, -0.057, -0.0252, 0.01, -0.0548, -0.0225, 0.0061, -0.067, -0.0225, 0.0083, -0.067, -0.0225, 0.0061, -0.0656, -0.0225, 0.0061, -0.0635, -0.0225, 0.0083, -0.0656, -0.0225, 0.0083, -0.0635, -0.0225, 0.01, -0.067, -0.0225, 0.01, -0.0656, -0.0225, 0.01, -0.0635, -0.0225, 0.0061, -0.0613, -0.0225, 0.0061, -0.0591, -0.0225, 0.0083, -0.0613, -0.0225, 0.0083, -0.0591, -0.0225, 0.0061, -0.057, -0.0225, 0.0061, -0.0548, -0.0225, 0.0083, -0.057, -0.0225, 0.0083, -0.0548, -0.0225, 0.01, -0.0613, -0.0225, 0.01, -0.0591, -0.0225, 0.01, -0.057, -0.0225, 0.01, -0.0548, -0.0179, 0.0061, -0.054, -0.0179, 0.0083, -0.054, -0.0157, 0.0061, -0.0562, -0.0168, 0.0061, -0.0551, -0.0157, 0.0083, -0.0562, -0.0168, 0.0083, -0.0551, -0.0179, 0.01, -0.054,
-0.0157, 0.01, -0.0562, -0.0168, 0.01, -0.0551, -0.0265, 0.0061, -0.0526, -0.0265, 0.0061, -0.0505, -0.0265, 0.0083, -0.0526, -0.0265, 0.0083, -0.0505, -0.0265, 0.0061, -0.0483, -0.0264, 0.0061, -0.0466, -0.0265, 0.0083, -0.0483, -0.0264, 0.0083, -0.0466, -0.0262, 0.0061, -0.0455, -0.0262, 0.0083, -0.0455, -0.0265, 0.01, -0.0526, -0.0265, 0.01, -0.0505, -0.0252, 0.01, -0.0526, -0.0252, 0.01, -0.0505, -0.0265, 0.01, -0.0483, -0.0264, 0.01, -0.0468, -0.0252, 0.01, -0.0483, -0.0262, 0.01, -0.0459, -0.0255, 0.0061, -0.0437, -0.0255, 0.0083, -0.0437, -0.0255, 0.01, -0.0437, -0.0225, 0.0061, -0.0526, -0.0225, 0.0061, -0.0503, -0.0225, 0.0083, -0.0526, -0.0225, 0.0083, -0.0503, -0.0202, 0.0061, -0.052, -0.0213, 0.0061, -0.0511, -0.0202, 0.0083, -0.052,
-0.0213, 0.0083, -0.0511, -0.02, 0.0061, -0.0452, -0.02, 0.0083, -0.0452, -0.0225, 0.01, -0.0526, -0.0225, 0.01, -0.0503, -0.0202, 0.01, -0.052, -0.0213, 0.01, -0.0511, -0.023, 0.01, -0.0483, -0.023, 0.01, -0.0461, -0.0208, 0.01, -0.0483, -0.0201, 0.01, -0.0452, -0.019, 0.0061, -0.053, -0.019, 0.0083, -0.053, -0.0155, 0.0061, -0.0495, -0.0155, 0.0083, -0.0495, -0.0177, 0.0061, -0.0473, -0.0188, 0.0061, -0.0463, -0.0177, 0.0083, -0.0473, -0.0188, 0.0083, -0.0463, -0.0166, 0.0061, -0.0484, -0.0166, 0.0083, -0.0484, -0.019, 0.01, -0.053, -0.0187, 0.01, -0.0505, -0.0165, 0.01, -0.0526, -0.0155, 0.01, -0.0495, -0.0177, 0.01, -0.0473, -0.0188, 0.01, -0.0463, -0.0166, 0.01, -0.0484, -0.0231, 0.0061, -0.0433, -0.0231, 0.0083, -0.0433,
-0.021, 0.0061, -0.0444, -0.021, 0.0083, -0.0444, -0.0231, 0.01, -0.0433, -0.0211, 0.01, -0.0443, -0.0092, 0.0061, -0.0627, -0.0092, 0.0083, -0.0627, -0.0069, 0.0061, -0.0647, -0.008, 0.0061, -0.0638, -0.0069, 0.0083, -0.0647, -0.008, 0.0083, -0.0638, -0.0092, 0.01, -0.0627, -0.0069, 0.01, -0.0647, -0.008, 0.01, -0.0638, -0.0135, 0.0061, -0.0584, -0.0135, 0.0083, -0.0584, -0.0114, 0.0061, -0.0605, -0.0124, 0.0061, -0.0594, -0.0114, 0.0083, -0.0605, -0.0124, 0.0083, -0.0594, -0.0146, 0.0061, -0.0573, -0.0146, 0.0083, -0.0573, -0.0112, 0.0061, -0.0538, -0.0112, 0.0083, -0.0538, -0.0135, 0.01, -0.0584, -0.0114, 0.01, -0.0605, -0.0124, 0.01, -0.0594, -0.0146, 0.01, -0.0573, -0.0143, 0.01, -0.0548, -0.0121, 0.01, -0.057, -0.0112, 0.01, -0.0538,
-0.0103, 0.0061, -0.0616, -0.0103, 0.0083, -0.0616, -0.0068, 0.0061, -0.0582, -0.0068, 0.0083, -0.0582, -0.009, 0.0061, -0.056, -0.0101, 0.0061, -0.0549, -0.009, 0.0083, -0.056, -0.0101, 0.0083, -0.0549, -0.0079, 0.0061, -0.0571, -0.0079, 0.0083, -0.0571, -0.0103, 0.01, -0.0616, -0.01, 0.01, -0.0591, -0.0078, 0.01, -0.0613, -0.0068, 0.01, -0.0582, -0.009, 0.01, -0.056, -0.0101, 0.01, -0.0549, -0.0079, 0.01, -0.0571, -0.0028, 0.0061, -0.0668, -0.0028, 0.0083, -0.0668, -0.0057, 0.0061, -0.0656, -0.0057, 0.0083, -0.0656, -0.0039, 0.0061, -0.0666, -0.0039, 0.0083, -0.0666, -0.0026, 0.01, -0.0668, -0.0059, 0.01, -0.0655, -0.0056, 0.01, -0.0635, -0.0037, 0.01, -0.0666, -0.0035, 0.01, -0.0635, -0.002, 0.0061, -0.067, -0.002, 0.0083, -0.067,
-0.002, 0.0061, -0.0656, -0.002, 0.0061, -0.0635, -0.002, 0.0083, -0.0656, -0.002, 0.0083, -0.0635, -0.002, 0.01, -0.067, -0.002, 0.01, -0.0656, -0.002, 0.01, -0.0635, -0.0059, 0.0061, -0.0591, -0.0059, 0.0083, -0.0591, -0.0059, 0.0061, -0.057, -0.0059, 0.0061, -0.0548, -0.0059, 0.0083, -0.057, -0.0059, 0.0083, -0.0548, -0.0056, 0.01, -0.0613, -0.0059, 0.01, -0.0591, -0.0035, 0.01, -0.0613, -0.0035, 0.01, -0.0591, -0.0059, 0.01, -0.057, -0.0059, 0.01, -0.0548, -0.0035, 0.01, -0.057, -0.0035, 0.01, -0.0548, -0.002, 0.0061, -0.0613, -0.002, 0.0061, -0.0591, -0.002, 0.0083, -0.0613, -0.002, 0.0083, -0.0591, -0.002, 0.0061, -0.057, -0.002, 0.0061, -0.0548, -0.002, 0.0083, -0.057, -0.002, 0.0083, -0.0548, -0.002, 0.01, -0.0613,
-0.002, 0.01, -0.0591, -0.002, 0.01, -0.057, -0.002, 0.01, -0.0548, -0.0134, 0.0061, -0.0517, -0.0144, 0.0061, -0.0506, -0.0134, 0.0083, -0.0517, -0.0144, 0.0083, -0.0506, -0.0123, 0.0061, -0.0528, -0.0123, 0.0083, -0.0528, -0.0134, 0.01, -0.0517, -0.0144, 0.01, -0.0506, -0.0123, 0.01, -0.0528, -0.0059, 0.0061, -0.0526, -0.0059, 0.0061, -0.0505, -0.0059, 0.0083, -0.0526, -0.0059, 0.0083, -0.0505, -0.0059, 0.0061, -0.0483, -0.0059, 0.0061, -0.0461, -0.0059, 0.0083, -0.0483, -0.0059, 0.0083, -0.0461, -0.0059, 0.01, -0.0526, -0.0059, 0.01, -0.0505, -0.0035, 0.01, -0.0526, -0.0035, 0.01, -0.0505, -0.0059, 0.01, -0.0483, -0.0059, 0.01, -0.0461, -0.0035, 0.01, -0.0483, -0.0035, 0.01, -0.0461, -0.002, 0.0061, -0.0526, -0.002, 0.0061, -0.0505,
-0.002, 0.0083, -0.0526, -0.002, 0.0083, -0.0505, -0.002, 0.0061, -0.0483, -0.002, 0.0061, -0.0461, -0.002, 0.0083, -0.0483, -0.002, 0.0083, -0.0461, -0.002, 0.01, -0.0526, -0.002, 0.01, -0.0505, -0.002, 0.01, -0.0483, -0.002, 0.01, -0.0461, -0.0059, 0.0061, -0.0439, -0.0059, 0.0061, -0.0425, -0.0059, 0.0083, -0.0439, -0.0059, 0.0083, -0.0425, -0.0035, 0.0061, -0.0425, -0.0035, 0.0083, -0.0425, -0.0059, 0.01, -0.0439, -0.0059, 0.01, -0.0425, -0.0035, 0.01, -0.0439, -0.0035, 0.01, -0.0425, -0.002, 0.0061, -0.0439, -0.002, 0.0061, -0.0425, -0.002, 0.0083, -0.0439, -0.002, 0.0083, -0.0425, -0.002, 0.01, -0.0439, -0.002, 0.01, -0.0425, -0.4807, 0.0061, -0.0005, -0.4801, 0.0061, 0.0016, -0.4807, 0.0083, -0.0005, -0.4801, 0.0083, 0.0016,
-0.4807, 0.01, -0.0004, -0.4802, 0.01, 0.0013, -0.4798, 0.0061, 0.0027, -0.4798, 0.0083, 0.0027, -0.4795, 0.0061, 0.0038, -0.4789, 0.0061, 0.006, -0.4795, 0.0083, 0.0038, -0.4789, 0.0083, 0.006, -0.4787, 0.01, -0.0005, -0.48, 0.01, 0.0021, -0.4766, 0.01, -0.0005, -0.4766, 0.01, 0.0016, -0.4795, 0.01, 0.004, -0.4789, 0.01, 0.0061, -0.4766, 0.01, 0.0038, -0.4766, 0.01, 0.006, -0.4784, 0.0061, 0.0081, -0.4779, 0.0061, 0.0102, -0.4784, 0.0083, 0.0081, -0.4779, 0.0083, 0.0102, -0.4776, 0.0061, 0.0112, -0.4776, 0.0083, 0.0112, -0.4772, 0.0061, 0.0125, -0.4765, 0.0061, 0.0147, -0.4772, 0.0083, 0.0125, -0.4765, 0.0083, 0.0147, -0.4784, 0.01, 0.0082, -0.478, 0.01, 0.0099, -0.4766, 0.01, 0.0081, -0.4777, 0.01, 0.0107,
-0.4772, 0.01, 0.0127, -0.4765, 0.01, 0.0148, -0.4744, 0.01, -0.0005, -0.4744, 0.01, 0.0016, -0.4722, 0.01, -0.0005, -0.4722, 0.01, 0.0016, -0.4744, 0.01, 0.0038, -0.4744, 0.01, 0.006, -0.4722, 0.01, 0.0038, -0.4722, 0.01, 0.006, -0.47, 0.01, -0.0005, -0.47, 0.01, 0.0016, -0.4679, 0.01, -0.0005, -0.4679, 0.01, 0.0016, -0.47, 0.01, 0.0038, -0.47, 0.01, 0.006, -0.4679, 0.01, 0.0038, -0.4679, 0.01, 0.006, -0.4744, 0.01, 0.0081, -0.4744, 0.01, 0.0103, -0.4722, 0.01, 0.0081, -0.4722, 0.01, 0.0103, -0.4744, 0.01, 0.0125, -0.4744, 0.01, 0.0146, -0.4722, 0.01, 0.0125, -0.4722, 0.01, 0.0146, -0.47, 0.01, 0.0081, -0.47, 0.01, 0.0103, -0.4679, 0.01, 0.0081, -0.4679, 0.01, 0.0103,
-0.47, 0.01, 0.0125, -0.47, 0.01, 0.0146, -0.4679, 0.01, 0.0125, -0.4679, 0.01, 0.0146, -0.4758, 0.0061, 0.0166, -0.4758, 0.0083, 0.0166, -0.4759, 0.01, 0.0164, -0.4754, 0.0061, 0.0177, -0.4749, 0.0061, 0.019, -0.4754, 0.0083, 0.0177, -0.4749, 0.0083, 0.019, -0.4742, 0.0061, 0.0212, -0.4736, 0.0061, 0.0229, -0.4742, 0.0083, 0.0212, -0.4736, 0.0083, 0.0229, -0.4732, 0.0061, 0.024, -0.4732, 0.0083, 0.024, -0.4755, 0.01, 0.0173, -0.4749, 0.01, 0.0192, -0.4722, 0.01, 0.0168, -0.4722, 0.01, 0.019, -0.4742, 0.01, 0.0213, -0.4736, 0.01, 0.0228, -0.4722, 0.01, 0.0212, -0.4733, 0.01, 0.0237, -0.47, 0.01, 0.0168, -0.47, 0.01, 0.019, -0.4679, 0.01, 0.0168, -0.4679, 0.01, 0.019, -0.47, 0.01, 0.0212,
-0.47, 0.01, 0.0233, -0.4679, 0.01, 0.0212, -0.4679, 0.01, 0.0233, -0.4726, 0.0061, 0.0255, -0.4716, 0.0061, 0.0277, -0.4726, 0.0083, 0.0255, -0.4716, 0.0083, 0.0277, -0.4711, 0.0061, 0.0288, -0.4711, 0.0083, 0.0288, -0.4724, 0.01, 0.0258, -0.4716, 0.01, 0.0278, -0.4711, 0.01, 0.0288, -0.4707, 0.0061, 0.0299, -0.4698, 0.0061, 0.032, -0.4707, 0.0083, 0.0299, -0.4698, 0.0083, 0.032, -0.47, 0.01, 0.0255, -0.47, 0.01, 0.0277, -0.4679, 0.01, 0.0255, -0.4679, 0.01, 0.0277, -0.4707, 0.01, 0.0299, -0.4697, 0.01, 0.0322, -0.4679, 0.01, 0.0298, -0.4679, 0.01, 0.032, -0.4628, 0.0061, -0.0005, -0.4625, 0.0061, 0.0006, -0.4628, 0.0083, -0.0005, -0.4625, 0.0083, 0.0006, -0.4657, 0.01, -0.0005, -0.4657, 0.01, 0.0016,
-0.4628, 0.01, -0.0008, -0.4623, 0.01, 0.0012, -0.4657, 0.01, 0.0038, -0.4657, 0.01, 0.006, -0.4635, 0.01, 0.0038, -0.4635, 0.01, 0.006, -0.4622, 0.0061, 0.0017, -0.4622, 0.0083, 0.0017, -0.4615, 0.0061, 0.0038, -0.4608, 0.0061, 0.006, -0.4615, 0.0083, 0.0038, -0.4608, 0.0083, 0.006, -0.4621, 0.01, 0.002, -0.4616, 0.01, 0.0037, -0.4609, 0.01, 0.0057, -0.4657, 0.01, 0.0081, -0.4657, 0.01, 0.0103, -0.4635, 0.01, 0.0081, -0.4635, 0.01, 0.0103, -0.4657, 0.01, 0.0125, -0.4657, 0.01, 0.0146, -0.4635, 0.01, 0.0125, -0.4635, 0.01, 0.0146, -0.4604, 0.0061, 0.0074, -0.4604, 0.0083, 0.0074, -0.4601, 0.0061, 0.0085, -0.4595, 0.0061, 0.0103, -0.4601, 0.0083, 0.0085, -0.4595, 0.0083, 0.0103, -0.4587, 0.0061, 0.0125,
-0.4582, 0.0061, 0.0138, -0.4587, 0.0083, 0.0125, -0.4582, 0.0083, 0.0138, -0.4603, 0.01, 0.0078, -0.4614, 0.01, 0.0103, -0.46, 0.01, 0.0087, -0.4595, 0.01, 0.0102, -0.4614, 0.01, 0.0125, -0.4614, 0.01, 0.0146, -0.4588, 0.01, 0.0122, -0.4581, 0.01, 0.0142, -0.4578, 0.0061, 0.0149, -0.4578, 0.0083, 0.0149, -0.4577, 0.01, 0.0151, -0.4657, 0.01, 0.0168, -0.4657, 0.01, 0.019, -0.4635, 0.01, 0.0168, -0.4635, 0.01, 0.019, -0.4657, 0.01, 0.0212, -0.4657, 0.01, 0.0233, -0.4635, 0.01, 0.0212, -0.4635, 0.01, 0.0233, -0.4614, 0.01, 0.0168, -0.4614, 0.01, 0.019, -0.4592, 0.01, 0.0168, -0.4592, 0.01, 0.019, -0.4614, 0.01, 0.0212, -0.4614, 0.01, 0.0233, -0.4592, 0.01, 0.0212, -0.4592, 0.01, 0.0233,
-0.4657, 0.01, 0.0255, -0.4657, 0.01, 0.0277, -0.4635, 0.01, 0.0255, -0.4635, 0.01, 0.0277, -0.4657, 0.01, 0.0298, -0.4657, 0.01, 0.032, -0.4635, 0.01, 0.0298, -0.4635, 0.01, 0.032, -0.4614, 0.01, 0.0255, -0.4614, 0.01, 0.0277, -0.4592, 0.01, 0.0255, -0.4592, 0.01, 0.0277, -0.4614, 0.01, 0.0298, -0.4614, 0.01, 0.032, -0.4592, 0.01, 0.0298, -0.4592, 0.01, 0.032, -0.4571, 0.0061, 0.0168, -0.4563, 0.0061, 0.019, -0.4571, 0.0083, 0.0168, -0.4563, 0.0083, 0.019, -0.4559, 0.0061, 0.0201, -0.4559, 0.0083, 0.0201, -0.4556, 0.0061, 0.0212, -0.4548, 0.0061, 0.0233, -0.4556, 0.0083, 0.0212, -0.4548, 0.0083, 0.0233, -0.4572, 0.01, 0.0167, -0.4564, 0.01, 0.0187, -0.4558, 0.01, 0.0206, -0.457, 0.01, 0.0233,
-0.4555, 0.01, 0.0214, -0.4549, 0.01, 0.0231, -0.4541, 0.0061, 0.0251, -0.4541, 0.0083, 0.0251, -0.457, 0.01, 0.0255, -0.457, 0.01, 0.0277, -0.454, 0.01, 0.0252, -0.4549, 0.01, 0.0277, -0.457, 0.01, 0.0298, -0.457, 0.01, 0.032, -0.4549, 0.01, 0.0298, -0.4549, 0.01, 0.032, -0.4536, 0.0061, 0.0262, -0.4529, 0.0061, 0.0277, -0.4536, 0.0083, 0.0262, -0.4529, 0.0083, 0.0277, -0.452, 0.0061, 0.0295, -0.452, 0.0083, 0.0295, -0.4514, 0.0061, 0.0306, -0.4507, 0.0061, 0.032, -0.4514, 0.0083, 0.0306, -0.4507, 0.0083, 0.032, -0.4536, 0.01, 0.0263, -0.453, 0.01, 0.0275, -0.4519, 0.01, 0.0296, -0.4527, 0.01, 0.032, -0.4514, 0.01, 0.0307, -0.4508, 0.01, 0.0318, -0.4691, 0.0061, 0.0334, -0.4691, 0.0083, 0.0334,
-0.4686, 0.0061, 0.0345, -0.4677, 0.0061, 0.0364, -0.4686, 0.0083, 0.0345, -0.4677, 0.0083, 0.0364, -0.467, 0.0061, 0.0378, -0.467, 0.0083, 0.0378, -0.4691, 0.01, 0.0333, -0.4686, 0.01, 0.0344, -0.4676, 0.01, 0.0365, -0.467, 0.01, 0.0377, -0.4664, 0.0061, 0.0389, -0.4655, 0.0061, 0.0407, -0.4664, 0.0083, 0.0389, -0.4655, 0.0083, 0.0407, -0.4657, 0.01, 0.0342, -0.4657, 0.01, 0.0363, -0.4635, 0.01, 0.0342, -0.4635, 0.01, 0.0363, -0.4664, 0.01, 0.0388, -0.4654, 0.01, 0.0409, -0.4635, 0.01, 0.0385, -0.4635, 0.01, 0.0407, -0.4614, 0.01, 0.0342, -0.4614, 0.01, 0.0363, -0.4592, 0.01, 0.0342, -0.4592, 0.01, 0.0363, -0.4614, 0.01, 0.0385, -0.4614, 0.01, 0.0407, -0.4592, 0.01, 0.0385, -0.4592, 0.01, 0.0407,
-0.4647, 0.0061, 0.042, -0.4647, 0.0083, 0.042, -0.4641, 0.0061, 0.0431, -0.463, 0.0061, 0.0449, -0.4641, 0.0083, 0.0431, -0.463, 0.0083, 0.0449, -0.4648, 0.01, 0.042, -0.4641, 0.01, 0.0431, -0.4631, 0.01, 0.0448, -0.4624, 0.0061, 0.046, -0.4624, 0.0083, 0.046, -0.4617, 0.0061, 0.0472, -0.4606, 0.0061, 0.0488, -0.4617, 0.0083, 0.0472, -0.4606, 0.0083, 0.0488, -0.4599, 0.0061, 0.0499, -0.4599, 0.0083, 0.0499, -0.4614, 0.01, 0.0429, -0.4625, 0.01, 0.0458, -0.4592, 0.01, 0.0429, -0.4592, 0.01, 0.045, -0.4615, 0.01, 0.0475, -0.4607, 0.01, 0.0488, -0.4592, 0.01, 0.0472, -0.46, 0.01, 0.0498, -0.457, 0.01, 0.0342, -0.457, 0.01, 0.0363, -0.4549, 0.01, 0.0342, -0.4549, 0.01, 0.0363, -0.457, 0.01, 0.0385,
-0.457, 0.01, 0.0407, -0.4549, 0.01, 0.0385, -0.4549, 0.01, 0.0407, -0.4498, 0.0061, 0.0339, -0.4498, 0.0083, 0.0339, -0.4517, 0.0061, 0.0417, -0.4517, 0.0083, 0.0417, -0.4498, 0.0061, 0.0392, -0.4505, 0.0061, 0.0408, -0.4498, 0.0083, 0.0392, -0.4505, 0.0083, 0.0408, -0.4527, 0.01, 0.0342, -0.4527, 0.01, 0.0363, -0.4498, 0.01, 0.0339, -0.4505, 0.01, 0.0363, -0.4527, 0.01, 0.0385, -0.4515, 0.01, 0.0415, -0.4499, 0.01, 0.0393, -0.4505, 0.01, 0.0408, -0.457, 0.01, 0.0429, -0.457, 0.01, 0.045, -0.4549, 0.01, 0.0429, -0.4549, 0.01, 0.045, -0.457, 0.01, 0.0472, -0.457, 0.01, 0.0494, -0.4549, 0.01, 0.0472, -0.4549, 0.01, 0.0494, -0.4523, 0.0061, 0.0426, -0.452, 0.0061, 0.0452, -0.4523, 0.0083, 0.0426,
-0.452, 0.0083, 0.0452, -0.4512, 0.0061, 0.0429, -0.4515, 0.0061, 0.0447, -0.4512, 0.0083, 0.0429, -0.4515, 0.0083, 0.0447, -0.4519, 0.0061, 0.0469, -0.4522, 0.0061, 0.0494, -0.4519, 0.0083, 0.0469, -0.4522, 0.0083, 0.0494, -0.4516, 0.0061, 0.0473, -0.4509, 0.0061, 0.0494, -0.4516, 0.0083, 0.0473, -0.4509, 0.0083, 0.0494, -0.4525, 0.01, 0.0431, -0.452, 0.01, 0.0453, -0.451, 0.01, 0.0425, -0.4515, 0.01, 0.045, -0.452, 0.01, 0.0468, -0.4523, 0.01, 0.0492, -0.4516, 0.01, 0.0472, -0.4508, 0.01, 0.0496, -0.4589, 0.0061, 0.0515, -0.4589, 0.0083, 0.0515, -0.459, 0.01, 0.0514, -0.4581, 0.0061, 0.0526, -0.4573, 0.0061, 0.0537, -0.4581, 0.0083, 0.0526, -0.4573, 0.0083, 0.0537, -0.4562, 0.0061, 0.0551, -0.4562, 0.0083, 0.0551,
-0.4553, 0.0061, 0.0562, -0.4543, 0.0061, 0.0576, -0.4553, 0.0083, 0.0562, -0.4543, 0.0083, 0.0576, -0.4582, 0.01, 0.0524, -0.457, 0.01, 0.054, -0.4549, 0.01, 0.0515, -0.4549, 0.01, 0.0537, -0.4562, 0.01, 0.0551, -0.4553, 0.01, 0.0562, -0.4543, 0.01, 0.0575, -0.4517, 0.0061, 0.0505, -0.4517, 0.0083, 0.0505, -0.4499, 0.0061, 0.0521, -0.4497, 0.0061, 0.0529, -0.4499, 0.0083, 0.0521, -0.4497, 0.0083, 0.0529, -0.4534, 0.0061, 0.0586, -0.4534, 0.0083, 0.0586, -0.4515, 0.01, 0.0508, -0.4527, 0.01, 0.0537, -0.4504, 0.01, 0.0514, -0.4498, 0.01, 0.0529, -0.4527, 0.01, 0.0559, -0.4534, 0.01, 0.0586, -0.4505, 0.01, 0.0559, -0.4505, 0.01, 0.058, -0.4523, 0.0061, 0.0599, -0.4523, 0.0083, 0.0599, -0.4513, 0.0061, 0.061,
-0.4502, 0.0061, 0.0621, -0.4513, 0.0083, 0.061, -0.4502, 0.0083, 0.0621, -0.4523, 0.01, 0.0599, -0.4513, 0.01, 0.061, -0.4502, 0.01, 0.0621, -0.4322, 0.0061, 0.0045, -0.4328, 0.0061, 0.0059, -0.4322, 0.0083, 0.0045, -0.4328, 0.0083, 0.0059, -0.4323, 0.01, 0.0046, -0.4328, 0.01, 0.0058, -0.4332, 0.0061, 0.0081, -0.4333, 0.0061, 0.0103, -0.4332, 0.0083, 0.0081, -0.4333, 0.0083, 0.0103, -0.4333, 0.0061, 0.0125, -0.4331, 0.0061, 0.0147, -0.4333, 0.0083, 0.0125, -0.4331, 0.0083, 0.0147, -0.4332, 0.01, 0.0081, -0.4333, 0.01, 0.0103, -0.4333, 0.01, 0.0125, -0.4331, 0.01, 0.0147, -0.4325, 0.0061, 0.0168, -0.4325, 0.0083, 0.0168, -0.4326, 0.01, 0.0165, -0.4306, 0.0061, 0.0019, -0.4306, 0.0083, 0.0019, -0.4285, 0.0061, -0.0001,
-0.4296, 0.0061, 0.0009, -0.4285, 0.0083, -0.0001, -0.4296, 0.0083, 0.0009, -0.4317, 0.0061, 0.0034, -0.4317, 0.0083, 0.0034, -0.4306, 0.01, 0.002, -0.4284, 0.01, -0.0001, -0.4295, 0.01, 0.0008, -0.4317, 0.01, 0.0035, -0.431, 0.01, 0.006, -0.4288, 0.01, 0.0038, -0.4288, 0.01, 0.006, -0.4269, 0.0061, -0.0012, -0.4269, 0.0083, -0.0012, -0.4269, 0.01, -0.0012, -0.4266, 0.01, 0.0016, -0.4245, 0.01, -0.0005, -0.4245, 0.01, 0.0016, -0.4266, 0.01, 0.0038, -0.4266, 0.01, 0.006, -0.4245, 0.01, 0.0038, -0.4245, 0.01, 0.006, -0.431, 0.01, 0.0081, -0.431, 0.01, 0.0103, -0.4288, 0.01, 0.0081, -0.4288, 0.01, 0.0103, -0.431, 0.01, 0.0125, -0.431, 0.01, 0.0146, -0.4288, 0.01, 0.0125, -0.4288, 0.01, 0.0146,
-0.4266, 0.01, 0.0081, -0.4266, 0.01, 0.0103, -0.4245, 0.01, 0.0081, -0.4245, 0.01, 0.0103, -0.4266, 0.01, 0.0125, -0.4266, 0.01, 0.0146, -0.4245, 0.01, 0.0125, -0.4245, 0.01, 0.0146, -0.4223, 0.01, -0.0005, -0.4223, 0.01, 0.0016, -0.4201, 0.01, -0.0005, -0.4201, 0.01, 0.0016, -0.4223, 0.01, 0.0038, -0.4223, 0.01, 0.006, -0.4201, 0.01, 0.0038, -0.4201, 0.01, 0.006, -0.418, 0.01, -0.0005, -0.418, 0.01, 0.0016, -0.4158, 0.01, -0.0005, -0.4158, 0.01, 0.0016, -0.418, 0.01, 0.0038, -0.418, 0.01, 0.006, -0.4158, 0.01, 0.0038, -0.4158, 0.01, 0.006, -0.4223, 0.01, 0.0081, -0.4223, 0.01, 0.0103, -0.4201, 0.01, 0.0081, -0.4201, 0.01, 0.0103, -0.4223, 0.01, 0.0125, -0.4223, 0.01, 0.0146,
-0.4201, 0.01, 0.0125, -0.4201, 0.01, 0.0146, -0.418, 0.01, 0.0081, -0.418, 0.01, 0.0103, -0.4158, 0.01, 0.0081, -0.4158, 0.01, 0.0103, -0.418, 0.01, 0.0125, -0.418, 0.01, 0.0146, -0.4158, 0.01, 0.0125, -0.4158, 0.01, 0.0146, -0.4321, 0.0061, 0.0179, -0.4315, 0.0061, 0.0191, -0.4321, 0.0083, 0.0179, -0.4315, 0.0083, 0.0191, -0.4304, 0.0061, 0.0206, -0.4304, 0.0083, 0.0206, -0.4293, 0.0061, 0.0217, -0.4282, 0.0061, 0.0226, -0.4293, 0.0083, 0.0217, -0.4282, 0.0083, 0.0226, -0.4322, 0.01, 0.0175, -0.4314, 0.01, 0.0194, -0.4288, 0.01, 0.0168, -0.4288, 0.01, 0.019, -0.4304, 0.01, 0.0206, -0.4293, 0.01, 0.0217, -0.4281, 0.01, 0.0226, -0.4267, 0.0061, 0.0236, -0.4267, 0.0083, 0.0236, -0.4249, 0.0061, 0.0242,
-0.4249, 0.0083, 0.0242, -0.4266, 0.01, 0.0168, -0.4266, 0.01, 0.019, -0.4245, 0.01, 0.0168, -0.4245, 0.01, 0.019, -0.4266, 0.01, 0.0212, -0.427, 0.01, 0.0234, -0.4245, 0.01, 0.0212, -0.4247, 0.01, 0.0243, -0.4238, 0.0061, 0.0246, -0.4238, 0.0083, 0.0246, -0.4237, 0.01, 0.0246, -0.4223, 0.01, 0.0168, -0.4223, 0.01, 0.019, -0.4201, 0.01, 0.0168, -0.4201, 0.01, 0.019, -0.4223, 0.01, 0.0212, -0.4223, 0.01, 0.0233, -0.4201, 0.01, 0.0212, -0.4201, 0.01, 0.0233, -0.4148, 0.0061, 0.0244, -0.4148, 0.0083, 0.0244, -0.418, 0.01, 0.0168, -0.418, 0.01, 0.019, -0.4158, 0.01, 0.0168, -0.4158, 0.01, 0.019, -0.418, 0.01, 0.0212, -0.418, 0.01, 0.0233, -0.4158, 0.01, 0.0212, -0.4153, 0.01, 0.0246,
-0.4223, 0.0061, 0.0249, -0.4223, 0.0083, 0.0249, -0.4201, 0.0061, 0.0251, -0.4201, 0.0083, 0.0251, -0.4223, 0.01, 0.0249, -0.4201, 0.01, 0.0251, -0.418, 0.0061, 0.025, -0.418, 0.0083, 0.025, -0.4159, 0.0061, 0.0247, -0.4159, 0.0083, 0.0247, -0.4179, 0.01, 0.025, -0.4162, 0.01, 0.0247, -0.4493, 0.0061, 0.035, -0.4487, 0.0061, 0.0363, -0.4493, 0.0083, 0.035, -0.4487, 0.0083, 0.0363, -0.4478, 0.0061, 0.0387, -0.4483, 0.0061, 0.0407, -0.4478, 0.0083, 0.0387, -0.4483, 0.0083, 0.0407, -0.4462, 0.0061, 0.0385, -0.4471, 0.0061, 0.0414, -0.4462, 0.0083, 0.0385, -0.4471, 0.0083, 0.0414, -0.4493, 0.01, 0.035, -0.4487, 0.01, 0.0362, -0.4478, 0.01, 0.0387, -0.4483, 0.01, 0.0407, -0.4463, 0.01, 0.0385, -0.4464, 0.01, 0.0407,
-0.444, 0.0061, 0.0384, -0.444, 0.0083, 0.0384, -0.4418, 0.0061, 0.0388, -0.4418, 0.0083, 0.0388, -0.444, 0.01, 0.0384, -0.444, 0.01, 0.0407, -0.4417, 0.01, 0.0388, -0.4418, 0.01, 0.0407, -0.4476, 0.0061, 0.0419, -0.4476, 0.0083, 0.0419, -0.4471, 0.0061, 0.0419, -0.4471, 0.0083, 0.0419, -0.4483, 0.01, 0.0426, -0.4483, 0.01, 0.045, -0.4471, 0.01, 0.0419, -0.4462, 0.01, 0.045, -0.4483, 0.01, 0.0472, -0.4483, 0.01, 0.0494, -0.4462, 0.01, 0.0472, -0.4462, 0.01, 0.0494, -0.444, 0.01, 0.0429, -0.444, 0.01, 0.045, -0.4418, 0.01, 0.0429, -0.4418, 0.01, 0.045, -0.444, 0.01, 0.0472, -0.444, 0.01, 0.0494, -0.4418, 0.01, 0.0472, -0.4418, 0.01, 0.0494, -0.44, 0.0061, 0.0393, -0.439, 0.0061, 0.0398,
-0.44, 0.0083, 0.0393, -0.439, 0.0083, 0.0398, -0.438, 0.0061, 0.0402, -0.438, 0.0083, 0.0402, -0.4402, 0.01, 0.0392, -0.4392, 0.01, 0.0397, -0.4375, 0.01, 0.0406, -0.4365, 0.0061, 0.0419, -0.4365, 0.0083, 0.0419, -0.4364, 0.0061, 0.0504, -0.4364, 0.0083, 0.0504, -0.4397, 0.01, 0.0429, -0.4397, 0.01, 0.045, -0.4364, 0.01, 0.0421, -0.4375, 0.01, 0.045, -0.4397, 0.01, 0.0472, -0.4397, 0.01, 0.0494, -0.4375, 0.01, 0.0472, -0.4363, 0.01, 0.0501, -0.4359, 0.0061, 0.043, -0.4355, 0.0061, 0.045, -0.4359, 0.0083, 0.043, -0.4355, 0.0083, 0.045, -0.4354, 0.0061, 0.0472, -0.4359, 0.0061, 0.0493, -0.4354, 0.0083, 0.0472, -0.4359, 0.0083, 0.0493, -0.4359, 0.01, 0.0433, -0.4355, 0.01, 0.045, -0.4354, 0.01, 0.0473,
-0.4358, 0.01, 0.0491, -0.4489, 0.0061, 0.0521, -0.4481, 0.0061, 0.0535, -0.4489, 0.0083, 0.0521, -0.4481, 0.0083, 0.0535, -0.4462, 0.0061, 0.0538, -0.4462, 0.0083, 0.0538, -0.4473, 0.0061, 0.0548, -0.4473, 0.0083, 0.0548, -0.4461, 0.0061, 0.0556, -0.4461, 0.0083, 0.0556, -0.4488, 0.01, 0.0521, -0.4481, 0.01, 0.0535, -0.4462, 0.01, 0.0515, -0.4464, 0.01, 0.0538, -0.4475, 0.01, 0.0547, -0.4483, 0.01, 0.058, -0.446, 0.01, 0.0556, -0.4462, 0.01, 0.058, -0.444, 0.0061, 0.0543, -0.444, 0.0083, 0.0543, -0.4418, 0.0061, 0.0541, -0.4418, 0.0083, 0.0541, -0.444, 0.0061, 0.0559, -0.444, 0.0083, 0.0559, -0.4418, 0.0061, 0.0557, -0.4418, 0.0083, 0.0557, -0.444, 0.01, 0.0515, -0.444, 0.01, 0.0543, -0.4418, 0.01, 0.0515,
-0.4417, 0.01, 0.0541, -0.444, 0.01, 0.0559, -0.444, 0.01, 0.058, -0.4419, 0.01, 0.0558, -0.4418, 0.01, 0.058, -0.4491, 0.0061, 0.0632, -0.4491, 0.0083, 0.0632, -0.4481, 0.0061, 0.0642, -0.4481, 0.0083, 0.0642, -0.4468, 0.0061, 0.0653, -0.4457, 0.0061, 0.0662, -0.4468, 0.0083, 0.0653, -0.4457, 0.0083, 0.0662, -0.4483, 0.01, 0.0602, -0.4491, 0.01, 0.0632, -0.4462, 0.01, 0.0602, -0.4462, 0.01, 0.0624, -0.448, 0.01, 0.0642, -0.4468, 0.01, 0.0653, -0.4457, 0.01, 0.0662, -0.4444, 0.0061, 0.0673, -0.4444, 0.0083, 0.0673, -0.444, 0.01, 0.0602, -0.444, 0.01, 0.0624, -0.4418, 0.01, 0.0602, -0.4418, 0.01, 0.0624, -0.444, 0.01, 0.0646, -0.4444, 0.01, 0.0673, -0.4418, 0.01, 0.0646, -0.4418, 0.01, 0.0667,
-0.4395, 0.0061, 0.0544, -0.4395, 0.0083, 0.0544, -0.4373, 0.0061, 0.0518, -0.4378, 0.0061, 0.0539, -0.4373, 0.0083, 0.0518, -0.4378, 0.0083, 0.0539, -0.44, 0.0061, 0.0551, -0.44, 0.0083, 0.0551, -0.4397, 0.01, 0.0515, -0.4393, 0.01, 0.0543, -0.4375, 0.01, 0.052, -0.4378, 0.0097, 0.054, -0.4398, 0.01, 0.055, -0.4397, 0.01, 0.058, -0.4375, 0.01, 0.0559, -0.4375, 0.01, 0.058, -0.4358, 0.0061, 0.0545, -0.4358, 0.0083, 0.0545, -0.4347, 0.0061, 0.0549, -0.4347, 0.0083, 0.0549, -0.4331, 0.0061, 0.0558, -0.4331, 0.0083, 0.0558, -0.436, 0.01, 0.0545, -0.4349, 0.01, 0.0548, -0.4353, 0.01, 0.058, -0.4329, 0.01, 0.056, -0.4332, 0.01, 0.058, -0.4397, 0.01, 0.0602, -0.4397, 0.01, 0.0624, -0.4375, 0.01, 0.0602,
-0.4375, 0.01, 0.0624, -0.4397, 0.01, 0.0646, -0.4397, 0.01, 0.0667, -0.4375, 0.01, 0.0646, -0.4375, 0.01, 0.0667, -0.4353, 0.01, 0.0602, -0.4353, 0.01, 0.0624, -0.4332, 0.01, 0.0602, -0.4332, 0.01, 0.0624, -0.4353, 0.01, 0.0646, -0.4353, 0.01, 0.0667, -0.4332, 0.01, 0.0646, -0.4332, 0.01, 0.0667, -0.4318, 0.0061, 0.0568, -0.4307, 0.0061, 0.0576, -0.4318, 0.0083, 0.0568, -0.4307, 0.0083, 0.0576, -0.4294, 0.0061, 0.0587, -0.4294, 0.0083, 0.0587, -0.4319, 0.01, 0.0568, -0.4307, 0.01, 0.0576, -0.4294, 0.01, 0.0587, -0.4283, 0.0061, 0.0596, -0.4283, 0.0083, 0.0596, -0.431, 0.01, 0.0602, -0.431, 0.01, 0.0624, -0.4283, 0.01, 0.0596, -0.4288, 0.01, 0.0624, -0.431, 0.01, 0.0646, -0.431, 0.01, 0.0667,
-0.4288, 0.01, 0.0646, -0.4288, 0.01, 0.0667, -0.427, 0.0061, 0.0607, -0.4259, 0.0061, 0.0616, -0.427, 0.0083, 0.0607, -0.4259, 0.0083, 0.0616, -0.4245, 0.0061, 0.0627, -0.4245, 0.0083, 0.0627, -0.4271, 0.01, 0.0607, -0.426, 0.01, 0.0615, -0.4243, 0.01, 0.0628, -0.4266, 0.01, 0.0646, -0.4266, 0.01, 0.0667, -0.4245, 0.01, 0.0646, -0.4245, 0.01, 0.0667, -0.4234, 0.0061, 0.0635, -0.4234, 0.0083, 0.0635, -0.4223, 0.0061, 0.0642, -0.4223, 0.0083, 0.0642, -0.4206, 0.0061, 0.0653, -0.4195, 0.0061, 0.066, -0.4206, 0.0083, 0.0653, -0.4195, 0.0083, 0.066, -0.4234, 0.01, 0.0635, -0.4222, 0.01, 0.0642, -0.4223, 0.01, 0.0667, -0.4207, 0.01, 0.0652, -0.4197, 0.01, 0.0659, -0.418, 0.0061, 0.0669, -0.418, 0.0083, 0.0669,
-0.4167, 0.0061, 0.0677, -0.4167, 0.0083, 0.0677, -0.4178, 0.01, 0.067, -0.4167, 0.01, 0.0677, -0.4434, 0.0061, 0.0682, -0.4434, 0.0083, 0.0682, -0.442, 0.0061, 0.0693, -0.441, 0.0061, 0.0701, -0.442, 0.0083, 0.0693, -0.441, 0.0083, 0.0701, -0.4433, 0.01, 0.0682, -0.4421, 0.01, 0.0693, -0.4409, 0.01, 0.0702, -0.4397, 0.0061, 0.0711, -0.4397, 0.0083, 0.0711, -0.4384, 0.0061, 0.0721, -0.4384, 0.0083, 0.0721, -0.4374, 0.0061, 0.0728, -0.4374, 0.0083, 0.0728, -0.4397, 0.01, 0.0689, -0.4399, 0.01, 0.0709, -0.4375, 0.01, 0.0689, -0.4383, 0.01, 0.0721, -0.4373, 0.01, 0.0729, -0.4358, 0.0061, 0.0739, -0.4347, 0.0061, 0.0746, -0.4358, 0.0083, 0.0739, -0.4347, 0.0083, 0.0746, -0.4332, 0.0061, 0.0755, -0.4332, 0.0083, 0.0755,
-0.4353, 0.01, 0.0689, -0.4353, 0.01, 0.0711, -0.4332, 0.01, 0.0689, -0.4332, 0.01, 0.0711, -0.4357, 0.01, 0.0739, -0.4346, 0.01, 0.0747, -0.4332, 0.01, 0.0732, -0.4334, 0.01, 0.0754, -0.4321, 0.0061, 0.0868, -0.4322, 0.0061, 0.0882, -0.4321, 0.0083, 0.0868, -0.4322, 0.0083, 0.0882, -0.4321, 0.01, 0.087, -0.4322, 0.01, 0.0879, -0.4309, 0.0061, 0.0763, -0.4309, 0.0083, 0.0763, -0.4288, 0.0061, 0.0758, -0.4288, 0.0083, 0.0758, -0.431, 0.01, 0.0689, -0.431, 0.01, 0.0711, -0.4288, 0.01, 0.0689, -0.4288, 0.01, 0.0711, -0.431, 0.01, 0.0732, -0.4309, 0.01, 0.0763, -0.4288, 0.01, 0.0732, -0.4286, 0.01, 0.0757, -0.4257, 0.0061, 0.0742, -0.4266, 0.0061, 0.0753, -0.4257, 0.0083, 0.0742, -0.4266, 0.0083, 0.0753,
-0.4242, 0.0061, 0.0742, -0.4246, 0.0061, 0.0748, -0.4242, 0.0083, 0.0742, -0.4246, 0.0083, 0.0748, -0.4266, 0.01, 0.0689, -0.4266, 0.01, 0.0711, -0.4245, 0.01, 0.0689, -0.4245, 0.01, 0.0711, -0.426, 0.01, 0.0744, -0.4266, 0.01, 0.0754, -0.4239, 0.01, 0.0741, -0.4246, 0.01, 0.0748, -0.4303, 0.0061, 0.0801, -0.4303, 0.0083, 0.0801, -0.4286, 0.0061, 0.0777, -0.4296, 0.0061, 0.079, -0.4286, 0.0083, 0.0777, -0.4296, 0.0083, 0.079, -0.4311, 0.0061, 0.0819, -0.4318, 0.0061, 0.0841, -0.4311, 0.0083, 0.0819, -0.4318, 0.0083, 0.0841, -0.4278, 0.0061, 0.0829, -0.4277, 0.0061, 0.083, -0.4278, 0.0083, 0.0829, -0.4277, 0.0083, 0.083, -0.4304, 0.01, 0.0802, -0.4286, 0.01, 0.0777, -0.4297, 0.01, 0.0791, -0.4311, 0.01, 0.0817,
-0.4317, 0.01, 0.0839, -0.4278, 0.01, 0.0829, -0.4276, 0.01, 0.0836, -0.4275, 0.0061, 0.0767, -0.4264, 0.0061, 0.0798, -0.4275, 0.0083, 0.0767, -0.4264, 0.0083, 0.0798, -0.4246, 0.0061, 0.0778, -0.4248, 0.0061, 0.0794, -0.4246, 0.0083, 0.0778, -0.4248, 0.0083, 0.0794, -0.4285, 0.0061, 0.0847, -0.4275, 0.0061, 0.084, -0.4285, 0.0083, 0.0847, -0.4275, 0.0083, 0.084, -0.4253, 0.0061, 0.0811, -0.4253, 0.0083, 0.0811, -0.4275, 0.01, 0.0766, -0.4264, 0.01, 0.0798, -0.4246, 0.01, 0.0778, -0.4248, 0.01, 0.0794, -0.4285, 0.01, 0.0847, -0.4275, 0.01, 0.084, -0.4253, 0.01, 0.0811, -0.4245, 0.01, 0.0841, -0.4223, 0.01, 0.0689, -0.4223, 0.01, 0.0711, -0.4201, 0.01, 0.0689, -0.4201, 0.01, 0.0711, -0.4222, 0.01, 0.0738,
-0.4223, 0.01, 0.0754, -0.4201, 0.01, 0.0734, -0.4201, 0.01, 0.0754, -0.4156, 0.0061, 0.0683, -0.4156, 0.0083, 0.0683, -0.418, 0.01, 0.0689, -0.418, 0.01, 0.0711, -0.4156, 0.01, 0.0683, -0.4158, 0.01, 0.0711, -0.418, 0.01, 0.0733, -0.418, 0.01, 0.0754, -0.4158, 0.01, 0.0736, -0.4158, 0.01, 0.0754, -0.4224, 0.0061, 0.078, -0.4229, 0.0061, 0.079, -0.4224, 0.0083, 0.078, -0.4229, 0.0083, 0.079, -0.422, 0.01, 0.0778, -0.4229, 0.01, 0.0789, -0.4204, 0.01, 0.0775, -0.4201, 0.01, 0.0798, -0.4223, 0.01, 0.0819, -0.4223, 0.01, 0.0841, -0.4201, 0.01, 0.0819, -0.4201, 0.01, 0.0841, -0.418, 0.01, 0.0776, -0.418, 0.01, 0.0798, -0.4158, 0.01, 0.0776, -0.4158, 0.01, 0.0798, -0.418, 0.01, 0.0819,
-0.418, 0.01, 0.0841, -0.4158, 0.01, 0.0819, -0.4158, 0.01, 0.0841, -0.432, 0.0061, 0.0858, -0.432, 0.0061, 0.0893, -0.432, 0.0083, 0.0858, -0.432, 0.0083, 0.0893, -0.4318, 0.0061, 0.0906, -0.431, 0.0061, 0.0928, -0.4318, 0.0083, 0.0906, -0.431, 0.0083, 0.0928, -0.432, 0.01, 0.0862, -0.4321, 0.01, 0.0887, -0.4288, 0.01, 0.0863, -0.4288, 0.01, 0.0884, -0.4317, 0.01, 0.0909, -0.4309, 0.01, 0.093, -0.4288, 0.01, 0.0906, -0.4288, 0.01, 0.0928, -0.4266, 0.01, 0.0863, -0.4266, 0.01, 0.0884, -0.4245, 0.01, 0.0863, -0.4245, 0.01, 0.0884, -0.4266, 0.01, 0.0906, -0.4266, 0.01, 0.0928, -0.4245, 0.01, 0.0906, -0.4245, 0.01, 0.0928, -0.4301, 0.0061, 0.0942, -0.4301, 0.0083, 0.0942, -0.4293, 0.0061, 0.0953,
-0.4282, 0.0061, 0.0965, -0.4293, 0.0083, 0.0953, -0.4282, 0.0083, 0.0965, -0.4301, 0.01, 0.0941, -0.4293, 0.01, 0.0953, -0.4282, 0.01, 0.0965, -0.427, 0.0061, 0.0976, -0.427, 0.0083, 0.0976, -0.4259, 0.0061, 0.0984, -0.4259, 0.0083, 0.0984, -0.4245, 0.0061, 0.0993, -0.4245, 0.0083, 0.0993, -0.4266, 0.01, 0.0949, -0.427, 0.01, 0.0976, -0.4245, 0.01, 0.0949, -0.4245, 0.01, 0.0971, -0.4258, 0.01, 0.0984, -0.4247, 0.01, 0.0992, -0.4223, 0.01, 0.0863, -0.4223, 0.01, 0.0884, -0.4201, 0.01, 0.0863, -0.4201, 0.01, 0.0884, -0.4223, 0.01, 0.0906, -0.4223, 0.01, 0.0928, -0.4201, 0.01, 0.0906, -0.4201, 0.01, 0.0928, -0.418, 0.01, 0.0863, -0.418, 0.01, 0.0884, -0.4158, 0.01, 0.0863, -0.4158, 0.01, 0.0884,
-0.418, 0.01, 0.0906, -0.418, 0.01, 0.0928, -0.4158, 0.01, 0.0906, -0.4158, 0.01, 0.0928, -0.4223, 0.0061, 0.1001, -0.4223, 0.0083, 0.1001, -0.4211, 0.0061, 0.1003, -0.42, 0.0061, 0.1005, -0.4211, 0.0083, 0.1003, -0.42, 0.0083, 0.1005, -0.4223, 0.01, 0.0949, -0.4223, 0.01, 0.0971, -0.4201, 0.01, 0.0949, -0.4201, 0.01, 0.0971, -0.4226, 0.01, 0.1, -0.4205, 0.01, 0.1005, -0.4197, 0.01, 0.1006, -0.418, 0.0061, 0.1006, -0.418, 0.0083, 0.1006, -0.4154, 0.0061, 0.1003, -0.4165, 0.0061, 0.1004, -0.4154, 0.0083, 0.1003, -0.4165, 0.0083, 0.1004, -0.418, 0.01, 0.0949, -0.418, 0.01, 0.0971, -0.4158, 0.01, 0.0949, -0.4158, 0.01, 0.0971, -0.418, 0.01, 0.0993, -0.418, 0.01, 0.1006, -0.4157, 0.01, 0.1003,
-0.4166, 0.01, 0.1004, -0.413, 0.0061, -0.0014, -0.413, 0.0083, -0.0014, -0.4114, 0.0061, -0.0004, -0.4114, 0.0083, -0.0004, -0.4131, 0.01, -0.0014, -0.4136, 0.01, 0.0016, -0.4112, 0.01, -0.0002, -0.4115, 0.01, 0.0016, -0.4136, 0.01, 0.0038, -0.4136, 0.01, 0.006, -0.4115, 0.01, 0.0038, -0.4115, 0.01, 0.006, -0.4104, 0.0061, 0.0005, -0.4091, 0.0061, 0.0015, -0.4104, 0.0083, 0.0005, -0.4091, 0.0083, 0.0015, -0.4082, 0.0061, 0.0027, -0.4082, 0.0083, 0.0027, -0.4075, 0.0061, 0.0038, -0.4064, 0.0061, 0.0059, -0.4075, 0.0083, 0.0038, -0.4064, 0.0083, 0.0059, -0.4104, 0.01, 0.0005, -0.4094, 0.01, 0.0012, -0.4081, 0.01, 0.003, -0.4093, 0.01, 0.006, -0.4074, 0.01, 0.004, -0.4065, 0.01, 0.0057, -0.4136, 0.01, 0.0081,
-0.4136, 0.01, 0.0103, -0.4115, 0.01, 0.0081, -0.4115, 0.01, 0.0103, -0.4136, 0.01, 0.0125, -0.4136, 0.01, 0.0146, -0.4115, 0.01, 0.0125, -0.4115, 0.01, 0.0146, -0.4061, 0.0061, 0.0076, -0.4061, 0.0083, 0.0076, -0.4061, 0.0061, 0.0147, -0.4061, 0.0083, 0.0147, -0.4093, 0.01, 0.0081, -0.4093, 0.01, 0.0103, -0.406, 0.01, 0.008, -0.4071, 0.01, 0.0103, -0.4093, 0.01, 0.0125, -0.4093, 0.01, 0.0146, -0.4071, 0.01, 0.0125, -0.4061, 0.01, 0.0146, -0.3981, 0.0061, -0.0012, -0.3981, 0.0083, -0.0012, -0.398, 0.01, -0.0011, -0.406, 0.0061, 0.0087, -0.4059, 0.0061, 0.0103, -0.406, 0.0083, 0.0087, -0.4059, 0.0083, 0.0103, -0.4059, 0.0061, 0.0125, -0.406, 0.0061, 0.0136, -0.4059, 0.0083, 0.0125, -0.406, 0.0083, 0.0136,
-0.4059, 0.01, 0.0088, -0.4059, 0.01, 0.0103, -0.4059, 0.01, 0.0125, -0.406, 0.01, 0.0136, -0.4136, 0.0061, 0.024, -0.4136, 0.0083, 0.024, -0.4105, 0.0061, 0.0221, -0.4115, 0.0061, 0.023, -0.4105, 0.0083, 0.0221, -0.4115, 0.0083, 0.023, -0.4136, 0.01, 0.0168, -0.4136, 0.01, 0.019, -0.4115, 0.01, 0.0168, -0.4115, 0.01, 0.019, -0.4136, 0.01, 0.0212, -0.4133, 0.01, 0.0239, -0.4105, 0.01, 0.0222, -0.4116, 0.01, 0.0231, -0.4083, 0.0061, 0.0199, -0.4083, 0.0083, 0.0199, -0.4067, 0.0061, 0.0168, -0.4076, 0.0061, 0.0189, -0.4067, 0.0083, 0.0168, -0.4076, 0.0083, 0.0189, -0.4093, 0.0061, 0.0211, -0.4093, 0.0083, 0.0211, -0.4093, 0.01, 0.0168, -0.4083, 0.01, 0.0198, -0.4068, 0.01, 0.0171, -0.4075, 0.01, 0.0187,
-0.4093, 0.01, 0.0211, -0.3963, 0.0061, -0.0001, -0.3963, 0.0083, -0.0001, -0.3951, 0.0061, 0.0005, -0.394, 0.0061, 0.0011, -0.3951, 0.0083, 0.0005, -0.394, 0.0083, 0.0011, -0.3966, 0.01, -0.0003, -0.3948, 0.01, 0.0007, -0.3939, 0.01, 0.0012, -0.3921, 0.0061, 0.0022, -0.3921, 0.0083, 0.0022, -0.3911, 0.0061, 0.0028, -0.3911, 0.0083, 0.0028, -0.3897, 0.0061, 0.0036, -0.3897, 0.0083, 0.0036, -0.3919, 0.01, -0.0005, -0.3921, 0.01, 0.0022, -0.3897, 0.01, -0.0005, -0.3897, 0.01, 0.0016, -0.391, 0.01, 0.0029, -0.3899, 0.01, 0.0035, -0.3881, 0.0061, 0.0045, -0.387, 0.0061, 0.0052, -0.3881, 0.0083, 0.0045, -0.387, 0.0083, 0.0052, -0.3854, 0.0061, 0.0061, -0.3854, 0.0083, 0.0061, -0.3876, 0.01, -0.0005, -0.3876, 0.01, 0.0016,
-0.3854, 0.01, -0.0005, -0.3854, 0.01, 0.0016, -0.388, 0.01, 0.0046, -0.3869, 0.01, 0.0052, -0.3854, 0.01, 0.0038, -0.3857, 0.01, 0.0059, -0.3841, 0.0061, 0.0069, -0.3841, 0.0083, 0.0069, -0.3832, 0.01, -0.0005, -0.3832, 0.01, 0.0016, -0.3811, 0.01, -0.0005, -0.3811, 0.01, 0.0016, -0.3832, 0.01, 0.0038, -0.3839, 0.01, 0.007, -0.3811, 0.01, 0.0038, -0.3811, 0.01, 0.006, -0.383, 0.0061, 0.0075, -0.383, 0.0083, 0.0075, -0.3811, 0.0061, 0.0086, -0.38, 0.0061, 0.0093, -0.3811, 0.0083, 0.0086, -0.38, 0.0083, 0.0093, -0.3828, 0.01, 0.0076, -0.3812, 0.01, 0.0086, -0.38, 0.01, 0.0093, -0.4023, 0.0061, 0.0451, -0.4023, 0.0083, 0.0451, -0.4026, 0.0061, 0.0472, -0.402, 0.0061, 0.0489, -0.4026, 0.0083, 0.0472,
-0.402, 0.0083, 0.0489, -0.4024, 0.01, 0.0452, -0.4026, 0.01, 0.0472, -0.4021, 0.01, 0.0488, -0.4011, 0.0061, 0.0429, -0.4017, 0.0061, 0.044, -0.4011, 0.0083, 0.0429, -0.4017, 0.0083, 0.044, -0.3987, 0.0061, 0.042, -0.3987, 0.0083, 0.042, -0.4014, 0.0061, 0.05, -0.4014, 0.0083, 0.05, -0.4011, 0.01, 0.0429, -0.4018, 0.01, 0.0442, -0.3986, 0.01, 0.042, -0.3984, 0.01, 0.045, -0.4006, 0.01, 0.0472, -0.4013, 0.01, 0.05, -0.3984, 0.01, 0.0472, -0.3984, 0.01, 0.0494, -0.4002, 0.0061, 0.0508, -0.4002, 0.0083, 0.0508, -0.3986, 0.0061, 0.0514, -0.3986, 0.0083, 0.0514, -0.4001, 0.01, 0.0509, -0.3986, 0.01, 0.0514, -0.396, 0.0061, 0.0423, -0.396, 0.0083, 0.0423, -0.3946, 0.0061, 0.0435, -0.3935, 0.0061, 0.0444,
-0.3946, 0.0083, 0.0435, -0.3935, 0.0083, 0.0444, -0.3928, 0.0061, 0.0472, -0.3938, 0.0061, 0.0494, -0.3928, 0.0083, 0.0472, -0.3938, 0.0083, 0.0494, -0.396, 0.01, 0.0423, -0.3963, 0.01, 0.045, -0.3946, 0.01, 0.0434, -0.3935, 0.01, 0.0444, -0.3963, 0.01, 0.0472, -0.3963, 0.01, 0.0494, -0.3928, 0.01, 0.0472, -0.3941, 0.01, 0.0498, -0.3959, 0.0061, 0.0512, -0.3959, 0.0083, 0.0512, -0.3951, 0.0061, 0.0505, -0.3951, 0.0083, 0.0505, -0.3959, 0.01, 0.0512, -0.3951, 0.01, 0.0505, -0.3789, 0.01, -0.0005, -0.3789, 0.01, 0.0016, -0.3767, 0.01, -0.0005, -0.3767, 0.01, 0.0016, -0.3789, 0.01, 0.0038, -0.3789, 0.01, 0.006, -0.3767, 0.01, 0.0038, -0.3767, 0.01, 0.006, -0.3746, 0.01, -0.0005, -0.3746, 0.01, 0.0016,
-0.3724, 0.01, -0.0005, -0.3724, 0.01, 0.0016, -0.3746, 0.01, 0.0038, -0.3746, 0.01, 0.006, -0.3724, 0.01, 0.0038, -0.3724, 0.01, 0.006, -0.3789, 0.0061, 0.0099, -0.3789, 0.0083, 0.0099, -0.377, 0.0061, 0.0109, -0.377, 0.0083, 0.0109, -0.3759, 0.0061, 0.0115, -0.3759, 0.0083, 0.0115, -0.3789, 0.01, 0.0081, -0.379, 0.01, 0.0098, -0.3767, 0.01, 0.0081, -0.3769, 0.01, 0.0109, -0.3758, 0.01, 0.0115, -0.3746, 0.0061, 0.0123, -0.3746, 0.0083, 0.0123, -0.3729, 0.0061, 0.0132, -0.3718, 0.0061, 0.0138, -0.3729, 0.0083, 0.0132, -0.3718, 0.0083, 0.0138, -0.3746, 0.01, 0.0081, -0.3746, 0.01, 0.0103, -0.3724, 0.01, 0.0081, -0.3724, 0.01, 0.0103, -0.3747, 0.01, 0.0122, -0.3728, 0.01, 0.0133, -0.3717, 0.01, 0.0139,
-0.3694, 0.0061, -0.0015, -0.3694, 0.0083, -0.0015, -0.3681, 0.0061, -0.0006, -0.3681, 0.0083, -0.0006, -0.3695, 0.01, -0.0016, -0.3702, 0.01, 0.0016, -0.3678, 0.01, -0.0004, -0.368, 0.01, 0.0016, -0.3702, 0.01, 0.0038, -0.3702, 0.01, 0.006, -0.368, 0.01, 0.0038, -0.368, 0.01, 0.006, -0.3667, 0.0061, 0.0003, -0.3656, 0.0061, 0.0011, -0.3667, 0.0083, 0.0003, -0.3656, 0.0083, 0.0011, -0.3639, 0.0061, 0.0021, -0.3639, 0.0083, 0.0021, -0.3628, 0.0061, 0.0029, -0.3628, 0.0083, 0.0029, -0.3667, 0.01, 0.0003, -0.3656, 0.01, 0.001, -0.364, 0.01, 0.0021, -0.3659, 0.01, 0.0038, -0.3659, 0.01, 0.006, -0.363, 0.01, 0.0027, -0.3637, 0.01, 0.006, -0.3702, 0.0061, 0.0147, -0.3702, 0.0083, 0.0147, -0.3687, 0.0061, 0.0155,
-0.3687, 0.0083, 0.0155, -0.3702, 0.01, 0.0081, -0.3702, 0.01, 0.0103, -0.368, 0.01, 0.0081, -0.368, 0.01, 0.0103, -0.3702, 0.01, 0.0125, -0.3705, 0.01, 0.0146, -0.368, 0.01, 0.0125, -0.3686, 0.01, 0.0156, -0.3659, 0.01, 0.0081, -0.3659, 0.01, 0.0103, -0.3637, 0.01, 0.0081, -0.3637, 0.01, 0.0103, -0.3659, 0.01, 0.0125, -0.3659, 0.01, 0.0146, -0.3637, 0.01, 0.0125, -0.3637, 0.01, 0.0146, -0.3677, 0.0061, 0.0161, -0.3677, 0.0083, 0.0161, -0.3675, 0.01, 0.0162, -0.3659, 0.0061, 0.0171, -0.3659, 0.0083, 0.0171, -0.3646, 0.0061, 0.0178, -0.3635, 0.0061, 0.0184, -0.3646, 0.0083, 0.0178, -0.3635, 0.0083, 0.0184, -0.3662, 0.01, 0.0169, -0.3643, 0.01, 0.0179, -0.3633, 0.01, 0.0184, -0.3615, 0.0061, 0.0037,
-0.3615, 0.0083, 0.0037, -0.36, 0.0061, 0.0046, -0.3589, 0.0061, 0.0053, -0.36, 0.0083, 0.0046, -0.3589, 0.0083, 0.0053, -0.3613, 0.01, 0.0039, -0.3615, 0.01, 0.006, -0.3601, 0.01, 0.0046, -0.359, 0.01, 0.0053, -0.3548, 0.0061, -0.0011, -0.3548, 0.0083, -0.0011, -0.3572, 0.0061, 0.0064, -0.3572, 0.0083, 0.0064, -0.3547, 0.01, -0.001, -0.3574, 0.01, 0.0063, -0.3615, 0.01, 0.0081, -0.3615, 0.01, 0.0103, -0.3594, 0.01, 0.0081, -0.3594, 0.01, 0.0103, -0.3615, 0.01, 0.0125, -0.3615, 0.01, 0.0146, -0.3594, 0.01, 0.0125, -0.3594, 0.01, 0.0146, -0.3561, 0.0061, 0.0071, -0.3561, 0.0083, 0.0071, -0.355, 0.0061, 0.0077, -0.355, 0.0083, 0.0077, -0.3563, 0.01, 0.0069, -0.3572, 0.01, 0.0103, -0.3547, 0.01, 0.0079,
-0.355, 0.01, 0.0103, -0.3572, 0.01, 0.0125, -0.3572, 0.01, 0.0146, -0.355, 0.01, 0.0125, -0.355, 0.01, 0.0146, -0.3531, 0.0061, -0.0, -0.352, 0.0061, 0.0007, -0.3531, 0.0083, -0.0, -0.352, 0.0083, 0.0007, -0.3507, 0.0061, 0.0015, -0.3507, 0.0083, 0.0015, -0.3531, 0.01, 0.0, -0.352, 0.01, 0.0007, -0.3507, 0.01, -0.0005, -0.3509, 0.01, 0.0014, -0.3492, 0.0061, 0.0025, -0.3492, 0.0083, 0.0025, -0.3481, 0.0061, 0.0031, -0.3481, 0.0083, 0.0031, -0.3463, 0.0061, 0.0042, -0.3463, 0.0083, 0.0042, -0.3485, 0.01, -0.0005, -0.3491, 0.01, 0.0025, -0.3463, 0.01, -0.0005, -0.3463, 0.01, 0.0016, -0.348, 0.01, 0.0032, -0.3467, 0.01, 0.004, -0.3533, 0.0061, 0.0088, -0.3522, 0.0061, 0.0095, -0.3533, 0.0083, 0.0088,
-0.3522, 0.0083, 0.0095, -0.3507, 0.0061, 0.0104, -0.3507, 0.0083, 0.0104, -0.3534, 0.01, 0.0088, -0.3523, 0.01, 0.0094, -0.3505, 0.01, 0.0105, -0.3529, 0.01, 0.0125, -0.3529, 0.01, 0.0146, -0.3507, 0.01, 0.0125, -0.3507, 0.01, 0.0146, -0.3493, 0.0061, 0.0112, -0.3493, 0.0083, 0.0112, -0.3483, 0.0061, 0.0119, -0.3483, 0.0083, 0.0119, -0.3464, 0.0061, 0.013, -0.3453, 0.0061, 0.0136, -0.3464, 0.0083, 0.013, -0.3453, 0.0083, 0.0136, -0.3494, 0.01, 0.0112, -0.3483, 0.01, 0.0119, -0.3485, 0.01, 0.0146, -0.3465, 0.01, 0.0129, -0.3456, 0.01, 0.0134, -0.3615, 0.0061, 0.0194, -0.3615, 0.0083, 0.0194, -0.3604, 0.0061, 0.02, -0.3604, 0.0083, 0.02, -0.3593, 0.0061, 0.0206, -0.3593, 0.0083, 0.0206, -0.3615, 0.01, 0.0168,
-0.3619, 0.01, 0.0192, -0.3594, 0.01, 0.0168, -0.3601, 0.01, 0.0202, -0.3591, 0.01, 0.0207, -0.3573, 0.0061, 0.0217, -0.3562, 0.0061, 0.0223, -0.3573, 0.0083, 0.0217, -0.3562, 0.0083, 0.0223, -0.355, 0.0061, 0.0228, -0.355, 0.0083, 0.0228, -0.3572, 0.01, 0.0168, -0.3572, 0.01, 0.019, -0.355, 0.01, 0.0168, -0.355, 0.01, 0.019, -0.3573, 0.01, 0.0217, -0.3562, 0.01, 0.0223, -0.355, 0.01, 0.0212, -0.3551, 0.01, 0.0228, -0.3529, 0.0061, 0.0238, -0.3529, 0.0083, 0.0238, -0.3516, 0.0061, 0.0243, -0.3516, 0.0083, 0.0243, -0.3529, 0.01, 0.0168, -0.3529, 0.01, 0.019, -0.3507, 0.01, 0.0168, -0.3507, 0.01, 0.019, -0.3529, 0.01, 0.0212, -0.3531, 0.01, 0.0237, -0.3507, 0.01, 0.0212, -0.3513, 0.01, 0.0245,
-0.3485, 0.01, 0.0168, -0.3485, 0.01, 0.019, -0.3463, 0.01, 0.0168, -0.3463, 0.01, 0.019, -0.3485, 0.01, 0.0212, -0.3485, 0.01, 0.0233, -0.3463, 0.01, 0.0212, -0.3463, 0.01, 0.0233, -0.3505, 0.0061, 0.0248, -0.3505, 0.0083, 0.0248, -0.3503, 0.01, 0.0249, -0.3485, 0.0061, 0.0257, -0.3485, 0.0083, 0.0257, -0.347, 0.0061, 0.0264, -0.3459, 0.0061, 0.0269, -0.347, 0.0083, 0.0264, -0.3459, 0.0083, 0.0269, -0.3488, 0.01, 0.0256, -0.3468, 0.01, 0.0265, -0.3458, 0.01, 0.0269, -0.3715, 0.0061, 0.0633, -0.3715, 0.0083, 0.0633, -0.3738, 0.0061, 0.0653, -0.3749, 0.0061, 0.0667, -0.3738, 0.0083, 0.0653, -0.3749, 0.0083, 0.0667, -0.3727, 0.0061, 0.0641, -0.3727, 0.0083, 0.0641, -0.3715, 0.01, 0.0633, -0.3738, 0.01, 0.0653,
-0.3747, 0.01, 0.0664, -0.3727, 0.01, 0.0641, -0.3724, 0.01, 0.0667, -0.3703, 0.0061, 0.0625, -0.3703, 0.0083, 0.0625, -0.3681, 0.0061, 0.0619, -0.3681, 0.0083, 0.0619, -0.3704, 0.01, 0.0626, -0.3682, 0.01, 0.0619, -0.3702, 0.01, 0.0646, -0.3702, 0.01, 0.0667, -0.368, 0.01, 0.0646, -0.368, 0.01, 0.0667, -0.3659, 0.0061, 0.0617, -0.3659, 0.0083, 0.0617, -0.3636, 0.0061, 0.062, -0.3636, 0.0083, 0.062, -0.3659, 0.01, 0.0617, -0.3635, 0.01, 0.062, -0.3659, 0.01, 0.0646, -0.3659, 0.01, 0.0667, -0.3637, 0.01, 0.0646, -0.3637, 0.01, 0.0667, -0.3548, 0.0061, 0.0406, -0.3548, 0.0083, 0.0406, -0.3549, 0.01, 0.0408, -0.3561, 0.0061, 0.0439, -0.3563, 0.0061, 0.045, -0.3561, 0.0083, 0.0439, -0.3563, 0.0083, 0.045,
-0.3558, 0.0061, 0.0427, -0.3558, 0.0083, 0.0427, -0.3563, 0.0061, 0.0472, -0.3563, 0.0083, 0.0472, -0.3561, 0.0061, 0.0483, -0.3559, 0.0061, 0.0494, -0.3561, 0.0083, 0.0483, -0.3559, 0.0083, 0.0494, -0.3561, 0.01, 0.0439, -0.3563, 0.01, 0.045, -0.3558, 0.01, 0.0428, -0.355, 0.01, 0.045, -0.3564, 0.01, 0.0468, -0.3562, 0.01, 0.0476, -0.3558, 0.01, 0.0497, -0.3502, 0.0061, 0.0372, -0.3502, 0.0083, 0.0372, -0.3533, 0.0061, 0.0389, -0.3539, 0.0061, 0.0396, -0.3533, 0.0083, 0.0389, -0.3539, 0.0083, 0.0396, -0.3513, 0.0061, 0.0377, -0.3513, 0.0083, 0.0377, -0.3501, 0.01, 0.0372, -0.3529, 0.01, 0.0386, -0.354, 0.01, 0.0397, -0.3511, 0.01, 0.0376, -0.3507, 0.01, 0.0407, -0.3485, 0.0061, 0.0366, -0.3485, 0.0083, 0.0366,
-0.3464, 0.0061, 0.0364, -0.3464, 0.0083, 0.0364, -0.3486, 0.01, 0.0367, -0.3464, 0.01, 0.0364, -0.3485, 0.01, 0.0385, -0.3485, 0.01, 0.0407, -0.3463, 0.01, 0.0385, -0.3463, 0.01, 0.0407, -0.3529, 0.01, 0.0429, -0.3529, 0.01, 0.045, -0.3507, 0.01, 0.0429, -0.3507, 0.01, 0.045, -0.3529, 0.01, 0.0472, -0.3529, 0.01, 0.0494, -0.3507, 0.01, 0.0472, -0.3507, 0.01, 0.0494, -0.3485, 0.01, 0.0429, -0.3485, 0.01, 0.045, -0.3463, 0.01, 0.0429, -0.3463, 0.01, 0.045, -0.3485, 0.01, 0.0472, -0.3485, 0.01, 0.0494, -0.3463, 0.01, 0.0472, -0.3463, 0.01, 0.0494, -0.3548, 0.0061, 0.0516, -0.3548, 0.0083, 0.0516, -0.3549, 0.01, 0.0515, -0.3617, 0.0061, 0.0629, -0.3617, 0.0083, 0.0629, -0.3625, 0.0061, 0.0622,
-0.3605, 0.0061, 0.0665, -0.3625, 0.0083, 0.0622, -0.3605, 0.0083, 0.0665, -0.3591, 0.0061, 0.0647, -0.3592, 0.0061, 0.0663, -0.3591, 0.0083, 0.0647, -0.3592, 0.0083, 0.0663, -0.3618, 0.01, 0.0628, -0.3625, 0.01, 0.0622, -0.3607, 0.01, 0.0661, -0.3593, 0.01, 0.0645, -0.3591, 0.01, 0.0664, -0.3578, 0.0061, 0.0669, -0.3578, 0.0083, 0.0669, -0.3577, 0.01, 0.0671, -0.3539, 0.0061, 0.0526, -0.3532, 0.0061, 0.0536, -0.3539, 0.0083, 0.0526, -0.3532, 0.0083, 0.0536, -0.3514, 0.0061, 0.0546, -0.3514, 0.0083, 0.0546, -0.3503, 0.0061, 0.0551, -0.3503, 0.0083, 0.0551, -0.354, 0.01, 0.0525, -0.353, 0.01, 0.0538, -0.3507, 0.01, 0.0515, -0.3512, 0.01, 0.0547, -0.3501, 0.01, 0.0551, -0.3485, 0.0061, 0.0554, -0.3485, 0.0083, 0.0554,
-0.3463, 0.0061, 0.0555, -0.3463, 0.0083, 0.0555, -0.3485, 0.01, 0.0515, -0.3485, 0.01, 0.0537, -0.3463, 0.01, 0.0515, -0.3463, 0.01, 0.0537, -0.3486, 0.01, 0.0554, -0.3463, 0.01, 0.0555, -0.4137, 0.0061, 0.0694, -0.4126, 0.0061, 0.07, -0.4137, 0.0083, 0.0694, -0.4126, 0.0083, 0.07, -0.4115, 0.0061, 0.0706, -0.4115, 0.0083, 0.0706, -0.4136, 0.0061, 0.0739, -0.4131, 0.0061, 0.0746, -0.4136, 0.0083, 0.0739, -0.4131, 0.0083, 0.0746, -0.4119, 0.0061, 0.074, -0.4114, 0.0061, 0.0747, -0.4119, 0.0083, 0.074, -0.4114, 0.0083, 0.0747, -0.4139, 0.01, 0.0693, -0.4129, 0.01, 0.0699, -0.4111, 0.01, 0.0708, -0.414, 0.01, 0.0738, -0.4133, 0.01, 0.0745, -0.4118, 0.01, 0.074, -0.411, 0.01, 0.0749, -0.4093, 0.0061, 0.0717,
-0.4093, 0.0083, 0.0717, -0.4082, 0.0061, 0.0722, -0.4082, 0.0083, 0.0722, -0.4071, 0.0061, 0.0727, -0.4071, 0.0083, 0.0727, -0.4095, 0.01, 0.0716, -0.4086, 0.01, 0.072, -0.4095, 0.01, 0.0757, -0.4068, 0.01, 0.0728, -0.4071, 0.01, 0.0754, -0.4136, 0.01, 0.0776, -0.4136, 0.01, 0.0798, -0.4115, 0.01, 0.0776, -0.4115, 0.01, 0.0798, -0.4136, 0.01, 0.0819, -0.4136, 0.01, 0.0841, -0.4115, 0.01, 0.0819, -0.4115, 0.01, 0.0841, -0.4068, 0.0061, 0.0797, -0.4068, 0.0083, 0.0797, -0.4062, 0.0061, 0.0814, -0.4062, 0.0083, 0.0814, -0.4086, 0.01, 0.0772, -0.4093, 0.01, 0.0798, -0.4081, 0.01, 0.0784, -0.4069, 0.01, 0.0794, -0.4093, 0.01, 0.0819, -0.4093, 0.01, 0.0841, -0.4061, 0.01, 0.0816, -0.4071, 0.01, 0.0841,
-0.4049, 0.0061, 0.0735, -0.4049, 0.0083, 0.0735, -0.4032, 0.0061, 0.0741, -0.4021, 0.0061, 0.0745, -0.4032, 0.0083, 0.0741, -0.4021, 0.0083, 0.0745, -0.4048, 0.01, 0.0735, -0.4049, 0.01, 0.0754, -0.4033, 0.01, 0.074, -0.4024, 0.01, 0.0743, -0.4006, 0.0061, 0.0749, -0.4006, 0.0083, 0.0749, -0.3984, 0.0061, 0.0757, -0.3984, 0.0083, 0.0757, -0.4004, 0.01, 0.075, -0.3983, 0.01, 0.0757, -0.4057, 0.0061, 0.0803, -0.4057, 0.0083, 0.0803, -0.4056, 0.0061, 0.0819, -0.4056, 0.0083, 0.0819, -0.4049, 0.01, 0.0776, -0.4057, 0.01, 0.0802, -0.4028, 0.01, 0.0776, -0.4028, 0.01, 0.0798, -0.4055, 0.01, 0.0823, -0.4054, 0.01, 0.0842, -0.4028, 0.01, 0.0819, -0.4028, 0.01, 0.0841, -0.4006, 0.01, 0.0776, -0.4006, 0.01, 0.0798,
-0.3984, 0.01, 0.0776, -0.3984, 0.01, 0.0798, -0.4006, 0.01, 0.0819, -0.4006, 0.01, 0.0841, -0.3984, 0.01, 0.0819, -0.3984, 0.01, 0.0841, -0.4136, 0.01, 0.0863, -0.4136, 0.01, 0.0884, -0.4115, 0.01, 0.0863, -0.4115, 0.01, 0.0884, -0.4136, 0.01, 0.0906, -0.4136, 0.01, 0.0928, -0.4115, 0.01, 0.0906, -0.4115, 0.01, 0.0928, -0.4064, 0.0061, 0.0928, -0.4064, 0.0083, 0.0928, -0.4093, 0.01, 0.0863, -0.4093, 0.01, 0.0884, -0.4071, 0.01, 0.0863, -0.4071, 0.01, 0.0884, -0.4093, 0.01, 0.0906, -0.4093, 0.01, 0.0928, -0.4071, 0.01, 0.0906, -0.4064, 0.01, 0.0929, -0.4106, 0.0061, 0.098, -0.4106, 0.0083, 0.098, -0.4136, 0.0061, 0.0998, -0.4136, 0.0083, 0.0998, -0.4117, 0.0061, 0.0987, -0.4117, 0.0083, 0.0987,
-0.4136, 0.01, 0.0949, -0.4136, 0.01, 0.0971, -0.4115, 0.01, 0.0949, -0.4107, 0.01, 0.0981, -0.4133, 0.01, 0.0997, -0.4118, 0.01, 0.0988, -0.4084, 0.0061, 0.0958, -0.4094, 0.0061, 0.097, -0.4084, 0.0083, 0.0958, -0.4094, 0.0083, 0.097, -0.4075, 0.0061, 0.0947, -0.4075, 0.0083, 0.0947, -0.4083, 0.01, 0.0957, -0.4094, 0.01, 0.097, -0.4075, 0.01, 0.0946, -0.4045, 0.0061, 0.0887, -0.4045, 0.0083, 0.0887, -0.4028, 0.0061, 0.089, -0.4028, 0.0083, 0.089, -0.4056, 0.0061, 0.0906, -0.406, 0.0061, 0.0917, -0.4056, 0.0083, 0.0906, -0.406, 0.0083, 0.0917, -0.4045, 0.01, 0.0862, -0.4045, 0.01, 0.0887, -0.4028, 0.01, 0.0863, -0.4029, 0.01, 0.089, -0.4056, 0.01, 0.0907, -0.406, 0.01, 0.0917, -0.4006, 0.0061, 0.0893,
-0.4006, 0.0083, 0.0893, -0.3995, 0.0061, 0.0895, -0.3995, 0.0083, 0.0895, -0.3984, 0.0061, 0.0897, -0.3984, 0.0083, 0.0897, -0.4006, 0.01, 0.0863, -0.4005, 0.01, 0.0893, -0.3984, 0.01, 0.0863, -0.3984, 0.01, 0.0884, -0.3995, 0.01, 0.0895, -0.3984, 0.01, 0.0897, -0.3966, 0.0061, 0.0763, -0.3966, 0.0083, 0.0763, -0.3968, 0.01, 0.0762, -0.3955, 0.0061, 0.0766, -0.3955, 0.0083, 0.0766, -0.3941, 0.0061, 0.077, -0.3941, 0.0083, 0.077, -0.3959, 0.01, 0.0765, -0.3963, 0.01, 0.0798, -0.3939, 0.01, 0.077, -0.3941, 0.01, 0.0798, -0.3963, 0.01, 0.0819, -0.3963, 0.01, 0.0841, -0.3941, 0.01, 0.0819, -0.3941, 0.01, 0.0841, -0.3919, 0.0061, 0.0775, -0.3919, 0.0083, 0.0775, -0.3898, 0.0061, 0.0779, -0.3898, 0.0083, 0.0779,
-0.3918, 0.01, 0.0775, -0.3919, 0.01, 0.0798, -0.3897, 0.01, 0.0779, -0.3897, 0.01, 0.0798, -0.3919, 0.01, 0.0819, -0.3919, 0.01, 0.0841, -0.3897, 0.01, 0.0819, -0.3897, 0.01, 0.0841, -0.3876, 0.0061, 0.0783, -0.3876, 0.0083, 0.0783, -0.3858, 0.0061, 0.0786, -0.3848, 0.0061, 0.0787, -0.3858, 0.0083, 0.0786, -0.3848, 0.0083, 0.0787, -0.3875, 0.01, 0.0783, -0.3876, 0.01, 0.0798, -0.3861, 0.01, 0.0785, -0.3852, 0.01, 0.0787, -0.3876, 0.01, 0.0819, -0.3876, 0.01, 0.0841, -0.3854, 0.01, 0.0819, -0.3854, 0.01, 0.0841, -0.3832, 0.0061, 0.0789, -0.3832, 0.0083, 0.0789, -0.3811, 0.0061, 0.079, -0.3811, 0.0083, 0.079, -0.3832, 0.01, 0.0789, -0.381, 0.01, 0.079, -0.3832, 0.01, 0.0819, -0.3832, 0.01, 0.0841,
-0.3811, 0.01, 0.0819, -0.3811, 0.01, 0.0841, -0.3963, 0.0061, 0.0901, -0.3963, 0.0083, 0.0901, -0.3941, 0.0061, 0.0904, -0.3941, 0.0083, 0.0904, -0.3963, 0.01, 0.0863, -0.3963, 0.01, 0.0884, -0.3941, 0.01, 0.0863, -0.3941, 0.01, 0.0884, -0.3963, 0.01, 0.0901, -0.3942, 0.01, 0.0904, -0.3919, 0.0061, 0.0908, -0.3919, 0.0083, 0.0908, -0.3898, 0.0061, 0.0912, -0.3898, 0.0083, 0.0912, -0.3919, 0.01, 0.0863, -0.3919, 0.01, 0.0884, -0.3897, 0.01, 0.0863, -0.3897, 0.01, 0.0884, -0.392, 0.01, 0.0908, -0.3899, 0.01, 0.0911, -0.3876, 0.0061, 0.0915, -0.3876, 0.0083, 0.0915, -0.3861, 0.0061, 0.0917, -0.3851, 0.0061, 0.0917, -0.3861, 0.0083, 0.0917, -0.3851, 0.0083, 0.0917, -0.3876, 0.01, 0.0863, -0.3876, 0.01, 0.0884,
-0.3854, 0.01, 0.0863, -0.3854, 0.01, 0.0884, -0.3877, 0.01, 0.0915, -0.3856, 0.01, 0.0917, -0.3848, 0.01, 0.0918, -0.3832, 0.0061, 0.0918, -0.3832, 0.0083, 0.0918, -0.3811, 0.0061, 0.0918, -0.3811, 0.0083, 0.0918, -0.3832, 0.01, 0.0863, -0.3832, 0.01, 0.0884, -0.3811, 0.01, 0.0863, -0.3811, 0.01, 0.0884, -0.3832, 0.01, 0.0906, -0.3832, 0.01, 0.0918, -0.3811, 0.01, 0.0906, -0.3811, 0.01, 0.0918, -0.3849, 0.0061, 0.0954, -0.3861, 0.0061, 0.0969, -0.3849, 0.0083, 0.0954, -0.3861, 0.0083, 0.0969, -0.3867, 0.0061, 0.0991, -0.3862, 0.0061, 0.1015, -0.3867, 0.0083, 0.0991, -0.3862, 0.0083, 0.1015, -0.3849, 0.01, 0.0954, -0.386, 0.01, 0.0967, -0.3867, 0.01, 0.0991, -0.3862, 0.01, 0.1015, -0.3832, 0.0061, 0.0941,
-0.3832, 0.0083, 0.0941, -0.3804, 0.0061, 0.0944, -0.3804, 0.0083, 0.0944, -0.3832, 0.01, 0.0941, -0.3832, 0.01, 0.0971, -0.3808, 0.01, 0.0943, -0.3811, 0.01, 0.0971, -0.3832, 0.01, 0.0993, -0.3832, 0.01, 0.1015, -0.3811, 0.01, 0.0993, -0.3811, 0.01, 0.1015, -0.3848, 0.0061, 0.1029, -0.3848, 0.0083, 0.1029, -0.3847, 0.01, 0.1029, -0.3832, 0.0061, 0.1037, -0.3832, 0.0083, 0.1037, -0.381, 0.0061, 0.1039, -0.381, 0.0083, 0.1039, -0.3834, 0.01, 0.1037, -0.381, 0.01, 0.1039, -0.3758, 0.0061, 0.0693, -0.3761, 0.0061, 0.0711, -0.3758, 0.0083, 0.0693, -0.3761, 0.0083, 0.0711, -0.376, 0.0061, 0.0731, -0.376, 0.0083, 0.0731, -0.3759, 0.01, 0.0695, -0.3761, 0.01, 0.071, -0.376, 0.01, 0.0728, -0.3755, 0.0061, 0.0682,
-0.3755, 0.0083, 0.0682, -0.3756, 0.0061, 0.0742, -0.3751, 0.0061, 0.0755, -0.3756, 0.0083, 0.0742, -0.3751, 0.0083, 0.0755, -0.3757, 0.01, 0.0685, -0.3746, 0.01, 0.0711, -0.3724, 0.01, 0.0689, -0.3724, 0.01, 0.0711, -0.3758, 0.01, 0.0738, -0.3749, 0.01, 0.0758, -0.3724, 0.01, 0.0732, -0.3724, 0.01, 0.0754, -0.3789, 0.0061, 0.0791, -0.3789, 0.0083, 0.0791, -0.3767, 0.0061, 0.0792, -0.3767, 0.0083, 0.0792, -0.3788, 0.01, 0.0791, -0.3766, 0.01, 0.0793, -0.3789, 0.01, 0.0819, -0.3789, 0.01, 0.0841, -0.3767, 0.01, 0.0819, -0.3767, 0.01, 0.0841, -0.374, 0.0061, 0.0773, -0.3745, 0.0061, 0.0798, -0.374, 0.0083, 0.0773, -0.3745, 0.0083, 0.0798, -0.3732, 0.0061, 0.0783, -0.3724, 0.0061, 0.0799, -0.3732, 0.0083, 0.0783,
-0.3724, 0.0083, 0.0799, -0.3717, 0.0061, 0.081, -0.3717, 0.0083, 0.081, -0.3741, 0.01, 0.0772, -0.3744, 0.01, 0.0799, -0.3731, 0.01, 0.0784, -0.3725, 0.01, 0.0799, -0.3746, 0.01, 0.0819, -0.3746, 0.01, 0.0841, -0.372, 0.01, 0.0809, -0.3724, 0.01, 0.0841, -0.3702, 0.01, 0.0689, -0.3702, 0.01, 0.0711, -0.368, 0.01, 0.0689, -0.368, 0.01, 0.0711, -0.3702, 0.01, 0.0732, -0.3702, 0.01, 0.0754, -0.368, 0.01, 0.0732, -0.368, 0.01, 0.0754, -0.3659, 0.01, 0.0689, -0.3659, 0.01, 0.0711, -0.3637, 0.01, 0.0689, -0.3637, 0.01, 0.0711, -0.3659, 0.01, 0.0732, -0.3659, 0.01, 0.0754, -0.3637, 0.01, 0.0732, -0.3637, 0.01, 0.0754, -0.3703, 0.0061, 0.0803, -0.3703, 0.0083, 0.0803, -0.3686, 0.0061, 0.0808,
-0.3686, 0.0083, 0.0808, -0.3702, 0.0061, 0.0816, -0.3702, 0.0083, 0.0816, -0.368, 0.0061, 0.0818, -0.368, 0.0083, 0.0818, -0.3702, 0.01, 0.0776, -0.3705, 0.01, 0.0802, -0.368, 0.01, 0.0776, -0.3682, 0.01, 0.0808, -0.37, 0.01, 0.0817, -0.3702, 0.01, 0.0841, -0.3678, 0.01, 0.0819, -0.368, 0.01, 0.0841, -0.365, 0.0061, 0.0808, -0.365, 0.0083, 0.0808, -0.3637, 0.0061, 0.0807, -0.3637, 0.0083, 0.0807, -0.367, 0.0061, 0.082, -0.367, 0.0083, 0.082, -0.3637, 0.0061, 0.0821, -0.3637, 0.0083, 0.0821, -0.3659, 0.01, 0.0776, -0.3657, 0.01, 0.0809, -0.3637, 0.01, 0.0776, -0.3634, 0.01, 0.0807, -0.366, 0.01, 0.082, -0.3659, 0.01, 0.0841, -0.3638, 0.01, 0.0821, -0.3637, 0.01, 0.0841, -0.3789, 0.0061, 0.0918,
-0.3789, 0.0083, 0.0918, -0.3767, 0.0061, 0.0918, -0.3767, 0.0083, 0.0918, -0.3789, 0.01, 0.0863, -0.3789, 0.01, 0.0884, -0.3767, 0.01, 0.0863, -0.3767, 0.01, 0.0884, -0.3789, 0.01, 0.0906, -0.3789, 0.01, 0.0918, -0.3767, 0.01, 0.0906, -0.3767, 0.01, 0.0918, -0.3746, 0.0061, 0.0918, -0.3746, 0.0083, 0.0918, -0.3724, 0.0061, 0.0918, -0.3724, 0.0083, 0.0918, -0.3746, 0.01, 0.0863, -0.3746, 0.01, 0.0884, -0.3724, 0.01, 0.0863, -0.3724, 0.01, 0.0884, -0.3746, 0.01, 0.0906, -0.3746, 0.01, 0.0918, -0.3724, 0.01, 0.0906, -0.3724, 0.01, 0.0918, -0.3789, 0.0061, 0.0952, -0.3779, 0.0061, 0.0962, -0.3789, 0.0083, 0.0952, -0.3779, 0.0083, 0.0962, -0.3773, 0.0061, 0.0972, -0.3773, 0.0083, 0.0972, -0.3782, 0.0061, 0.1021,
-0.3782, 0.0083, 0.1021, -0.3769, 0.0061, 0.0991, -0.3776, 0.0061, 0.1014, -0.3769, 0.0083, 0.0991, -0.3776, 0.0083, 0.1014, -0.379, 0.01, 0.0951, -0.3779, 0.01, 0.0963, -0.3773, 0.01, 0.0974, -0.3789, 0.01, 0.0993, -0.3782, 0.01, 0.102, -0.3769, 0.01, 0.0991, -0.3774, 0.01, 0.1009, -0.3701, 0.0061, 0.0916, -0.3712, 0.0061, 0.0917, -0.3701, 0.0083, 0.0916, -0.3712, 0.0083, 0.0917, -0.368, 0.0061, 0.0913, -0.368, 0.0083, 0.0913, -0.3702, 0.01, 0.0863, -0.3702, 0.01, 0.0884, -0.368, 0.01, 0.0863, -0.368, 0.01, 0.0884, -0.3703, 0.01, 0.0916, -0.3712, 0.01, 0.0917, -0.3679, 0.01, 0.0913, -0.3659, 0.0061, 0.0909, -0.3659, 0.0083, 0.0909, -0.3637, 0.0061, 0.0905, -0.3637, 0.0083, 0.0905, -0.3659, 0.01, 0.0863,
-0.3659, 0.01, 0.0884, -0.3637, 0.01, 0.0863, -0.3637, 0.01, 0.0884, -0.3658, 0.01, 0.0908, -0.3636, 0.01, 0.0904, -0.3605, 0.0061, 0.0679, -0.3605, 0.0083, 0.0679, -0.3588, 0.0061, 0.0704, -0.3588, 0.0083, 0.0704, -0.3583, 0.0061, 0.0765, -0.3583, 0.0083, 0.0765, -0.3603, 0.01, 0.0682, -0.3615, 0.01, 0.0711, -0.3596, 0.01, 0.0691, -0.3587, 0.01, 0.0703, -0.3615, 0.01, 0.0732, -0.3615, 0.01, 0.0754, -0.3594, 0.01, 0.0732, -0.3582, 0.01, 0.0763, -0.3571, 0.0061, 0.0689, -0.3569, 0.0061, 0.0711, -0.3571, 0.0083, 0.0689, -0.3569, 0.0083, 0.0711, -0.357, 0.0061, 0.0733, -0.3577, 0.0061, 0.0757, -0.357, 0.0083, 0.0733, -0.3577, 0.0083, 0.0757, -0.3571, 0.01, 0.0687, -0.3569, 0.01, 0.071, -0.3571, 0.01, 0.0733,
-0.3575, 0.01, 0.0752, -0.3614, 0.0061, 0.0805, -0.3614, 0.0083, 0.0805, -0.3591, 0.0061, 0.0777, -0.3596, 0.0061, 0.0793, -0.3591, 0.0083, 0.0777, -0.3596, 0.0083, 0.0793, -0.3619, 0.0061, 0.0812, -0.3619, 0.0083, 0.0812, -0.3615, 0.01, 0.0776, -0.3611, 0.01, 0.0803, -0.3594, 0.01, 0.078, -0.3599, 0.01, 0.0795, -0.3618, 0.01, 0.0812, -0.3615, 0.01, 0.0841, -0.3594, 0.01, 0.0819, -0.3594, 0.01, 0.0841, -0.357, 0.0061, 0.0781, -0.3581, 0.0061, 0.0788, -0.357, 0.0083, 0.0781, -0.3581, 0.0083, 0.0788, -0.3551, 0.0061, 0.0772, -0.3551, 0.0083, 0.0772, -0.3569, 0.01, 0.0781, -0.3579, 0.01, 0.0787, -0.3553, 0.01, 0.0773, -0.355, 0.01, 0.0798, -0.3572, 0.01, 0.0819, -0.3572, 0.01, 0.0841, -0.355, 0.01, 0.0819,
-0.355, 0.01, 0.0841, -0.3521, 0.0061, 0.0764, -0.3521, 0.0083, 0.0764, -0.3507, 0.0061, 0.076, -0.3507, 0.0083, 0.076, -0.352, 0.01, 0.0764, -0.3508, 0.01, 0.076, -0.3485, 0.0061, 0.0753, -0.3485, 0.0083, 0.0753, -0.3455, 0.0061, 0.0742, -0.3466, 0.0061, 0.0746, -0.3455, 0.0083, 0.0742, -0.3466, 0.0083, 0.0746, -0.3487, 0.01, 0.0754, -0.3454, 0.01, 0.0742, -0.3465, 0.01, 0.0746, -0.3532, 0.0061, 0.0767, -0.3532, 0.0083, 0.0767, -0.353, 0.01, 0.0767, -0.3529, 0.01, 0.0798, -0.3507, 0.01, 0.0776, -0.3507, 0.01, 0.0798, -0.3529, 0.01, 0.0819, -0.3529, 0.01, 0.0841, -0.3507, 0.01, 0.0819, -0.3507, 0.01, 0.0841, -0.3485, 0.01, 0.0776, -0.3485, 0.01, 0.0798, -0.3463, 0.01, 0.0776, -0.3463, 0.01, 0.0798,
-0.3485, 0.01, 0.0819, -0.3485, 0.01, 0.0841, -0.3463, 0.01, 0.0819, -0.3463, 0.01, 0.0841, -0.3583, 0.0061, 0.0895, -0.3583, 0.0083, 0.0895, -0.3615, 0.0061, 0.0901, -0.3615, 0.0083, 0.0901, -0.3594, 0.0061, 0.0897, -0.3594, 0.0083, 0.0897, -0.3615, 0.01, 0.0863, -0.3615, 0.01, 0.0884, -0.3594, 0.01, 0.0863, -0.359, 0.01, 0.0896, -0.3615, 0.01, 0.09, -0.3597, 0.01, 0.0898, -0.3572, 0.0061, 0.0893, -0.3572, 0.0083, 0.0893, -0.355, 0.0061, 0.0889, -0.355, 0.0083, 0.0889, -0.3572, 0.01, 0.0863, -0.357, 0.01, 0.0893, -0.355, 0.01, 0.0863, -0.3549, 0.01, 0.0889, -0.3529, 0.0061, 0.0885, -0.3529, 0.0083, 0.0885, -0.3507, 0.0061, 0.0879, -0.3507, 0.0083, 0.0879, -0.3529, 0.01, 0.0863, -0.3527, 0.01, 0.0884,
-0.3507, 0.01, 0.0863, -0.3506, 0.01, 0.0879, -0.3482, 0.0061, 0.0871, -0.3493, 0.0061, 0.0875, -0.3482, 0.0083, 0.0871, -0.3493, 0.0083, 0.0875, -0.3463, 0.0061, 0.0864, -0.3463, 0.0083, 0.0864, -0.3483, 0.01, 0.0871, -0.3494, 0.01, 0.0875, -0.3461, 0.01, 0.0864, -0.3791, 0.0061, 0.1031, -0.3791, 0.0083, 0.1031, -0.3791, 0.01, 0.1031, -0.3452, 0.0061, 0.0049, -0.3442, 0.0061, 0.0055, -0.3452, 0.0083, 0.0049, -0.3442, 0.0083, 0.0055, -0.3423, 0.0061, 0.0066, -0.3423, 0.0083, 0.0066, -0.3442, 0.01, -0.0005, -0.3442, 0.01, 0.0016, -0.342, 0.01, -0.0005, -0.342, 0.01, 0.0016, -0.345, 0.01, 0.005, -0.344, 0.01, 0.0056, -0.342, 0.01, 0.0038, -0.3423, 0.01, 0.0066, -0.3398, 0.01, -0.0005, -0.3398, 0.01, 0.0016,
-0.3377, 0.01, -0.0005, -0.3377, 0.01, 0.0016, -0.3398, 0.01, 0.0038, -0.3398, 0.01, 0.006, -0.3377, 0.01, 0.0038, -0.3377, 0.01, 0.006, -0.3412, 0.0061, 0.0072, -0.3412, 0.0083, 0.0072, -0.3442, 0.0061, 0.0142, -0.3442, 0.0083, 0.0142, -0.3423, 0.0061, 0.0153, -0.3423, 0.0083, 0.0153, -0.3412, 0.01, 0.0072, -0.3438, 0.01, 0.0144, -0.3424, 0.01, 0.0152, -0.3398, 0.0061, 0.008, -0.3398, 0.0083, 0.008, -0.3381, 0.0061, 0.0089, -0.3371, 0.0061, 0.0095, -0.3381, 0.0083, 0.0089, -0.3371, 0.0083, 0.0095, -0.34, 0.01, 0.0079, -0.338, 0.01, 0.009, -0.337, 0.01, 0.0095, -0.3355, 0.01, -0.0005, -0.3355, 0.01, 0.0016, -0.3333, 0.01, -0.0005, -0.3333, 0.01, 0.0016, -0.3355, 0.01, 0.0038, -0.3355, 0.01, 0.006,
-0.3333, 0.01, 0.0038, -0.3333, 0.01, 0.006, -0.3312, 0.01, -0.0005, -0.3312, 0.01, 0.0016, -0.329, 0.01, -0.0005, -0.329, 0.01, 0.0016, -0.3312, 0.01, 0.0038, -0.3312, 0.01, 0.006, -0.329, 0.01, 0.0038, -0.329, 0.01, 0.006, -0.3355, 0.0061, 0.0102, -0.3355, 0.0083, 0.0102, -0.3337, 0.0061, 0.0111, -0.3337, 0.0083, 0.0111, -0.3327, 0.0061, 0.0116, -0.3327, 0.0083, 0.0116, -0.3355, 0.01, 0.0081, -0.3357, 0.01, 0.0101, -0.3333, 0.01, 0.0081, -0.3336, 0.01, 0.0111, -0.3326, 0.01, 0.0116, -0.3312, 0.0061, 0.0123, -0.3312, 0.0083, 0.0123, -0.3293, 0.0061, 0.0132, -0.3282, 0.0061, 0.0137, -0.3293, 0.0083, 0.0132, -0.3282, 0.0083, 0.0137, -0.3312, 0.01, 0.0081, -0.3312, 0.01, 0.0103, -0.329, 0.01, 0.0081,
-0.329, 0.01, 0.0103, -0.3313, 0.01, 0.0122, -0.3292, 0.01, 0.0132, -0.3282, 0.01, 0.0137, -0.3412, 0.0061, 0.0159, -0.3412, 0.0083, 0.0159, -0.3442, 0.01, 0.0168, -0.3442, 0.01, 0.019, -0.3414, 0.01, 0.0158, -0.342, 0.01, 0.019, -0.3442, 0.01, 0.0212, -0.3442, 0.01, 0.0233, -0.342, 0.01, 0.0212, -0.342, 0.01, 0.0233, -0.3398, 0.0061, 0.0167, -0.3398, 0.0083, 0.0167, -0.3382, 0.0061, 0.0176, -0.3371, 0.0061, 0.0181, -0.3382, 0.0083, 0.0176, -0.3371, 0.0083, 0.0181, -0.3396, 0.01, 0.0168, -0.3398, 0.01, 0.019, -0.3383, 0.01, 0.0175, -0.3372, 0.01, 0.0181, -0.3398, 0.01, 0.0212, -0.3398, 0.01, 0.0233, -0.3377, 0.01, 0.0212, -0.3377, 0.01, 0.0233, -0.3442, 0.0061, 0.0277, -0.3442, 0.0083, 0.0277,
-0.3423, 0.0061, 0.0284, -0.3423, 0.0083, 0.0284, -0.3412, 0.0061, 0.0289, -0.3412, 0.0083, 0.0289, -0.3442, 0.01, 0.0255, -0.3444, 0.01, 0.0276, -0.342, 0.01, 0.0255, -0.3422, 0.01, 0.0285, -0.3411, 0.01, 0.0289, -0.3398, 0.0061, 0.0294, -0.3398, 0.0083, 0.0294, -0.3377, 0.0061, 0.0303, -0.3377, 0.0083, 0.0303, -0.3398, 0.01, 0.0255, -0.3398, 0.01, 0.0277, -0.3377, 0.01, 0.0255, -0.3377, 0.01, 0.0277, -0.34, 0.01, 0.0294, -0.3379, 0.01, 0.0302, -0.3355, 0.0061, 0.0189, -0.3355, 0.0083, 0.0189, -0.3337, 0.0061, 0.0198, -0.3337, 0.0083, 0.0198, -0.3327, 0.0061, 0.0203, -0.3327, 0.0083, 0.0203, -0.3353, 0.01, 0.019, -0.3339, 0.01, 0.0197, -0.3355, 0.01, 0.0212, -0.3355, 0.01, 0.0233, -0.3329, 0.01, 0.0202,
-0.3333, 0.01, 0.0233, -0.3312, 0.0061, 0.0209, -0.3312, 0.0083, 0.0209, -0.329, 0.0061, 0.0218, -0.329, 0.0083, 0.0218, -0.3309, 0.01, 0.021, -0.3312, 0.01, 0.0233, -0.3289, 0.01, 0.0218, -0.329, 0.01, 0.0233, -0.3363, 0.0061, 0.0308, -0.3352, 0.0061, 0.0312, -0.3363, 0.0083, 0.0308, -0.3352, 0.0083, 0.0312, -0.3333, 0.0061, 0.0318, -0.3333, 0.0083, 0.0318, -0.3355, 0.01, 0.0255, -0.3355, 0.01, 0.0277, -0.3333, 0.01, 0.0255, -0.3333, 0.01, 0.0277, -0.3359, 0.01, 0.0309, -0.335, 0.01, 0.0313, -0.3333, 0.01, 0.0298, -0.3335, 0.01, 0.0318, -0.3312, 0.0061, 0.0325, -0.3312, 0.0083, 0.0325, -0.3296, 0.0061, 0.033, -0.3296, 0.0083, 0.033, -0.3312, 0.01, 0.0255, -0.3312, 0.01, 0.0277, -0.329, 0.01, 0.0255,
-0.329, 0.01, 0.0277, -0.3312, 0.01, 0.0298, -0.3314, 0.01, 0.0324, -0.329, 0.01, 0.0298, -0.3293, 0.01, 0.0331, -0.3261, 0.0061, -0.0014, -0.3261, 0.0083, -0.0014, -0.3246, 0.0061, -0.0004, -0.3246, 0.0083, -0.0004, -0.3262, 0.01, -0.0015, -0.3268, 0.01, 0.0016, -0.3244, 0.01, -0.0003, -0.3246, 0.01, 0.0016, -0.3268, 0.01, 0.0038, -0.3268, 0.01, 0.006, -0.3246, 0.01, 0.0038, -0.3246, 0.01, 0.006, -0.3234, 0.0061, 0.0004, -0.3223, 0.0061, 0.0011, -0.3234, 0.0083, 0.0004, -0.3223, 0.0083, 0.0011, -0.3205, 0.0061, 0.0022, -0.3205, 0.0083, 0.0022, -0.3194, 0.0061, 0.0028, -0.3194, 0.0083, 0.0028, -0.3234, 0.01, 0.0004, -0.3223, 0.01, 0.0011, -0.3207, 0.01, 0.0021, -0.3225, 0.01, 0.0038, -0.3225, 0.01, 0.006,
-0.3197, 0.01, 0.0027, -0.3203, 0.01, 0.006, -0.3268, 0.0061, 0.0144, -0.3268, 0.0083, 0.0144, -0.3247, 0.0061, 0.0153, -0.3247, 0.0083, 0.0153, -0.3268, 0.01, 0.0081, -0.3268, 0.01, 0.0103, -0.3246, 0.01, 0.0081, -0.3246, 0.01, 0.0103, -0.3268, 0.01, 0.0125, -0.327, 0.01, 0.0143, -0.3246, 0.01, 0.0125, -0.3249, 0.01, 0.0151, -0.3235, 0.0061, 0.0157, -0.3235, 0.0083, 0.0157, -0.3225, 0.01, 0.0081, -0.3225, 0.01, 0.0103, -0.3203, 0.01, 0.0081, -0.3203, 0.01, 0.0103, -0.3225, 0.01, 0.0125, -0.323, 0.01, 0.0159, -0.3203, 0.01, 0.0125, -0.3203, 0.01, 0.0146, -0.3181, 0.0061, 0.0036, -0.3181, 0.0083, 0.0036, -0.3164, 0.0061, 0.0045, -0.3153, 0.0061, 0.0051, -0.3164, 0.0083, 0.0045, -0.3153, 0.0083, 0.0051,
-0.3179, 0.01, 0.0037, -0.3181, 0.01, 0.006, -0.3165, 0.01, 0.0044, -0.3154, 0.01, 0.005, -0.3138, 0.0061, 0.0059, -0.3138, 0.0083, 0.0059, -0.3122, 0.0061, 0.0068, -0.3122, 0.0083, 0.0068, -0.3136, 0.01, 0.0061, -0.3123, 0.01, 0.0067, -0.3181, 0.01, 0.0081, -0.3181, 0.01, 0.0103, -0.316, 0.01, 0.0081, -0.316, 0.01, 0.0103, -0.3181, 0.01, 0.0125, -0.3181, 0.01, 0.0146, -0.316, 0.01, 0.0125, -0.316, 0.01, 0.0146, -0.3111, 0.0061, 0.0074, -0.3111, 0.0083, 0.0074, -0.3138, 0.01, 0.0081, -0.3138, 0.01, 0.0103, -0.3113, 0.01, 0.0073, -0.3116, 0.01, 0.0103, -0.3138, 0.01, 0.0125, -0.3138, 0.01, 0.0146, -0.3116, 0.01, 0.0125, -0.3116, 0.01, 0.0146, -0.3279, 0.0061, 0.0222, -0.3268, 0.0061, 0.0227,
-0.3279, 0.0083, 0.0222, -0.3268, 0.0083, 0.0227, -0.3246, 0.0061, 0.0235, -0.3246, 0.0083, 0.0235, -0.3279, 0.01, 0.0222, -0.3268, 0.01, 0.0227, -0.3245, 0.01, 0.0236, -0.3224, 0.0061, 0.0161, -0.3224, 0.0083, 0.0161, -0.3203, 0.0061, 0.0169, -0.3203, 0.0083, 0.0169, -0.3229, 0.0061, 0.0242, -0.3229, 0.0083, 0.0242, -0.3221, 0.01, 0.0162, -0.3205, 0.01, 0.0168, -0.3231, 0.01, 0.0241, -0.3268, 0.01, 0.0255, -0.3268, 0.01, 0.0277, -0.3246, 0.01, 0.0255, -0.3246, 0.01, 0.0277, -0.3268, 0.01, 0.0298, -0.3268, 0.01, 0.032, -0.3246, 0.01, 0.0298, -0.3246, 0.01, 0.032, -0.3218, 0.0061, 0.0246, -0.3218, 0.0083, 0.0246, -0.3203, 0.0061, 0.0251, -0.3203, 0.0083, 0.0251, -0.3221, 0.01, 0.0245, -0.3225, 0.01, 0.0277,
-0.3201, 0.01, 0.0251, -0.3203, 0.01, 0.0277, -0.3225, 0.01, 0.0298, -0.3225, 0.01, 0.032, -0.3203, 0.01, 0.0298, -0.3203, 0.01, 0.032, -0.3184, 0.0061, 0.0176, -0.3173, 0.0061, 0.018, -0.3184, 0.0083, 0.0176, -0.3173, 0.0083, 0.018, -0.316, 0.0061, 0.0185, -0.316, 0.0083, 0.0185, -0.3183, 0.01, 0.0176, -0.3173, 0.01, 0.018, -0.316, 0.01, 0.0168, -0.3161, 0.01, 0.0185, -0.3138, 0.0061, 0.0192, -0.3138, 0.0083, 0.0192, -0.3118, 0.0061, 0.0198, -0.3118, 0.0083, 0.0198, -0.3107, 0.0061, 0.0201, -0.3107, 0.0083, 0.0201, -0.3138, 0.01, 0.0168, -0.314, 0.01, 0.0192, -0.3116, 0.01, 0.0168, -0.3117, 0.01, 0.0199, -0.3107, 0.01, 0.0201, -0.3181, 0.0061, 0.0257, -0.3181, 0.0083, 0.0257, -0.3162, 0.0061, 0.0263,
-0.3151, 0.0061, 0.0267, -0.3162, 0.0083, 0.0263, -0.3151, 0.0083, 0.0267, -0.318, 0.01, 0.0257, -0.3181, 0.01, 0.0277, -0.3164, 0.01, 0.0262, -0.3155, 0.01, 0.0265, -0.3181, 0.01, 0.0298, -0.3181, 0.01, 0.032, -0.316, 0.01, 0.0298, -0.316, 0.01, 0.032, -0.3138, 0.0061, 0.0271, -0.3138, 0.0083, 0.0271, -0.3116, 0.0061, 0.0277, -0.3116, 0.0083, 0.0277, -0.3136, 0.01, 0.0271, -0.3115, 0.01, 0.0277, -0.3138, 0.01, 0.0298, -0.3138, 0.01, 0.032, -0.3116, 0.01, 0.0298, -0.3116, 0.01, 0.032, -0.3441, 0.0061, 0.0367, -0.3441, 0.0083, 0.0367, -0.3427, 0.0061, 0.0373, -0.3427, 0.0083, 0.0373, -0.3416, 0.0061, 0.0378, -0.3416, 0.0083, 0.0378, -0.3441, 0.01, 0.0367, -0.3428, 0.01, 0.0372, -0.3442, 0.01, 0.0385,
-0.3442, 0.01, 0.0407, -0.3417, 0.01, 0.0377, -0.342, 0.01, 0.0407, -0.3401, 0.0061, 0.0388, -0.339, 0.0061, 0.04, -0.3401, 0.0083, 0.0388, -0.339, 0.0083, 0.04, -0.3384, 0.0061, 0.0411, -0.3384, 0.0083, 0.0411, -0.3401, 0.01, 0.0388, -0.3389, 0.01, 0.0401, -0.3383, 0.01, 0.0412, -0.3442, 0.01, 0.0429, -0.3442, 0.01, 0.045, -0.342, 0.01, 0.0429, -0.342, 0.01, 0.045, -0.3442, 0.01, 0.0472, -0.3442, 0.01, 0.0494, -0.342, 0.01, 0.0472, -0.342, 0.01, 0.0494, -0.3378, 0.0061, 0.0422, -0.3373, 0.0061, 0.045, -0.3378, 0.0083, 0.0422, -0.3373, 0.0083, 0.045, -0.3373, 0.0061, 0.0472, -0.3377, 0.0061, 0.0494, -0.3373, 0.0083, 0.0472, -0.3377, 0.0083, 0.0494, -0.3398, 0.01, 0.0429, -0.3398, 0.01, 0.045,
-0.3376, 0.01, 0.0426, -0.3373, 0.01, 0.045, -0.3398, 0.01, 0.0472, -0.3398, 0.01, 0.0494, -0.3373, 0.01, 0.0473, -0.3378, 0.01, 0.0496, -0.3286, 0.0061, 0.0333, -0.3286, 0.0083, 0.0333, -0.3284, 0.01, 0.0334, -0.3418, 0.0061, 0.0544, -0.3418, 0.0083, 0.0544, -0.3441, 0.0061, 0.0553, -0.3441, 0.0083, 0.0553, -0.343, 0.0061, 0.0548, -0.343, 0.0083, 0.0548, -0.3442, 0.01, 0.0515, -0.3442, 0.01, 0.0537, -0.342, 0.01, 0.0515, -0.3418, 0.01, 0.0544, -0.3441, 0.01, 0.0553, -0.343, 0.01, 0.0548, -0.3392, 0.0061, 0.0521, -0.3403, 0.0061, 0.0532, -0.3392, 0.0083, 0.0521, -0.3403, 0.0083, 0.0532, -0.3385, 0.0061, 0.0509, -0.3385, 0.0083, 0.0509, -0.3391, 0.01, 0.052, -0.3403, 0.01, 0.0532, -0.3385, 0.01, 0.0508,
-0.3311, 0.0061, 0.0673, -0.3311, 0.0083, 0.0673, -0.3282, 0.0061, 0.0655, -0.3292, 0.0061, 0.0662, -0.3282, 0.0083, 0.0655, -0.3292, 0.0083, 0.0662, -0.3309, 0.01, 0.0672, -0.3281, 0.01, 0.0655, -0.3292, 0.01, 0.0662, -0.3268, 0.0061, 0.0339, -0.3268, 0.0083, 0.0339, -0.3247, 0.0061, 0.0345, -0.3247, 0.0083, 0.0345, -0.3269, 0.01, 0.0338, -0.3248, 0.01, 0.0344, -0.3225, 0.0061, 0.035, -0.3225, 0.0083, 0.035, -0.3212, 0.0061, 0.0352, -0.3201, 0.0061, 0.0354, -0.3212, 0.0083, 0.0352, -0.3201, 0.0083, 0.0354, -0.3227, 0.01, 0.0349, -0.3206, 0.01, 0.0353, -0.3198, 0.01, 0.0355, -0.3181, 0.0061, 0.0358, -0.3181, 0.0083, 0.0358, -0.316, 0.0061, 0.0363, -0.316, 0.0083, 0.0363, -0.3181, 0.01, 0.0342, -0.3182, 0.01, 0.0358,
-0.316, 0.01, 0.0342, -0.3161, 0.01, 0.0363, -0.3138, 0.0061, 0.0366, -0.3138, 0.0083, 0.0366, -0.3116, 0.0061, 0.0369, -0.3116, 0.0083, 0.0369, -0.3138, 0.01, 0.0342, -0.3139, 0.01, 0.0366, -0.3116, 0.01, 0.0342, -0.3117, 0.01, 0.0368, -0.3106, 0.0061, 0.0504, -0.3106, 0.0083, 0.0504, -0.3106, 0.01, 0.0504, -0.3193, 0.0061, 0.059, -0.3193, 0.0083, 0.059, -0.3193, 0.01, 0.059, -0.3243, 0.0061, 0.0629, -0.3243, 0.0083, 0.0629, -0.3268, 0.0061, 0.0646, -0.3268, 0.0083, 0.0646, -0.3254, 0.0061, 0.0637, -0.3254, 0.0083, 0.0637, -0.3243, 0.01, 0.0629, -0.3271, 0.01, 0.0648, -0.3268, 0.01, 0.0667, -0.3253, 0.01, 0.0636, -0.3246, 0.01, 0.0667, -0.3217, 0.0061, 0.061, -0.3228, 0.0061, 0.0619, -0.3217, 0.0083, 0.061,
-0.3228, 0.0083, 0.0619, -0.3204, 0.0061, 0.06, -0.3204, 0.0083, 0.06, -0.3217, 0.01, 0.061, -0.3228, 0.01, 0.0619, -0.3204, 0.01, 0.06, -0.3203, 0.01, 0.0624, -0.3225, 0.01, 0.0646, -0.3225, 0.01, 0.0667, -0.3203, 0.01, 0.0646, -0.3203, 0.01, 0.0667, -0.3149, 0.0061, 0.0548, -0.3149, 0.0083, 0.0548, -0.3171, 0.0061, 0.0569, -0.3182, 0.0061, 0.058, -0.3171, 0.0083, 0.0569, -0.3182, 0.0083, 0.058, -0.316, 0.0061, 0.0558, -0.316, 0.0083, 0.0558, -0.3149, 0.01, 0.0548, -0.3171, 0.01, 0.0569, -0.3182, 0.01, 0.058, -0.316, 0.01, 0.0558, -0.316, 0.01, 0.058, -0.3127, 0.0061, 0.0526, -0.3138, 0.0061, 0.0537, -0.3127, 0.0083, 0.0526, -0.3138, 0.0083, 0.0537, -0.3117, 0.0061, 0.0515, -0.3117, 0.0083, 0.0515,
-0.3127, 0.01, 0.0526, -0.3138, 0.01, 0.0537, -0.3117, 0.01, 0.0515, -0.3116, 0.01, 0.0537, -0.3138, 0.01, 0.0559, -0.3138, 0.01, 0.058, -0.3116, 0.01, 0.0559, -0.3116, 0.01, 0.058, -0.3181, 0.01, 0.0602, -0.3181, 0.01, 0.0624, -0.316, 0.01, 0.0602, -0.316, 0.01, 0.0624, -0.3181, 0.01, 0.0646, -0.3181, 0.01, 0.0667, -0.316, 0.01, 0.0646, -0.316, 0.01, 0.0667, -0.313, 0.0061, 0.0676, -0.313, 0.0083, 0.0676, -0.3107, 0.0061, 0.0655, -0.3118, 0.0061, 0.0665, -0.3107, 0.0083, 0.0655, -0.3118, 0.0083, 0.0665, -0.3138, 0.01, 0.0602, -0.3138, 0.01, 0.0624, -0.3116, 0.01, 0.0602, -0.3116, 0.01, 0.0624, -0.3138, 0.01, 0.0646, -0.313, 0.01, 0.0676, -0.3107, 0.01, 0.0655, -0.3118, 0.01, 0.0665,
-0.3095, 0.0061, 0.0082, -0.3095, 0.0083, 0.0082, -0.3077, 0.0061, 0.0089, -0.3066, 0.0061, 0.0094, -0.3077, 0.0083, 0.0089, -0.3066, 0.0083, 0.0094, -0.3093, 0.01, 0.0083, -0.3095, 0.01, 0.0103, -0.3078, 0.01, 0.0089, -0.3069, 0.01, 0.0093, -0.3095, 0.01, 0.0125, -0.3095, 0.01, 0.0146, -0.3073, 0.01, 0.0125, -0.3073, 0.01, 0.0146, -0.3051, 0.0061, 0.01, -0.3051, 0.0083, 0.01, -0.3029, 0.0061, 0.0108, -0.3029, 0.0083, 0.0108, -0.3049, 0.01, 0.0101, -0.3029, 0.01, 0.0109, -0.3051, 0.01, 0.0125, -0.3051, 0.01, 0.0146, -0.3029, 0.01, 0.0125, -0.3029, 0.01, 0.0146, -0.3017, 0.0061, 0.0113, -0.3017, 0.0083, 0.0113, -0.3006, 0.0061, 0.0118, -0.3006, 0.0083, 0.0118, -0.2986, 0.0061, 0.0125, -0.2986, 0.0083, 0.0125,
-0.3017, 0.01, 0.0113, -0.3007, 0.01, 0.0117, -0.3008, 0.01, 0.0146, -0.2985, 0.01, 0.0125, -0.2986, 0.01, 0.0146, -0.2964, 0.0061, 0.0131, -0.2964, 0.0083, 0.0131, -0.2948, 0.0061, 0.0134, -0.2937, 0.0061, 0.0137, -0.2948, 0.0083, 0.0134, -0.2937, 0.0083, 0.0137, -0.2964, 0.01, 0.0131, -0.2964, 0.01, 0.0146, -0.295, 0.01, 0.0134, -0.2941, 0.01, 0.0136, -0.3095, 0.0061, 0.0205, -0.3095, 0.0083, 0.0205, -0.3073, 0.0061, 0.021, -0.3073, 0.0083, 0.021, -0.3095, 0.01, 0.0168, -0.3095, 0.01, 0.019, -0.3073, 0.01, 0.0168, -0.3073, 0.01, 0.019, -0.3095, 0.01, 0.0205, -0.3074, 0.01, 0.021, -0.3051, 0.0061, 0.0215, -0.3051, 0.0083, 0.0215, -0.3029, 0.0061, 0.022, -0.3029, 0.0083, 0.022, -0.3051, 0.01, 0.0168,
-0.3051, 0.01, 0.019, -0.3029, 0.01, 0.0168, -0.3029, 0.01, 0.019, -0.3053, 0.01, 0.0215, -0.3031, 0.01, 0.0219, -0.3095, 0.0061, 0.0282, -0.3095, 0.0083, 0.0282, -0.3074, 0.0061, 0.0286, -0.3074, 0.0083, 0.0286, -0.3063, 0.0061, 0.0288, -0.3063, 0.0083, 0.0288, -0.3094, 0.01, 0.0282, -0.3077, 0.01, 0.0285, -0.3095, 0.01, 0.0298, -0.3095, 0.01, 0.032, -0.3069, 0.01, 0.0287, -0.3073, 0.01, 0.032, -0.3051, 0.0061, 0.029, -0.3051, 0.0083, 0.029, -0.3029, 0.0061, 0.0293, -0.3029, 0.0083, 0.0293, -0.305, 0.01, 0.029, -0.3051, 0.01, 0.032, -0.3028, 0.01, 0.0293, -0.3029, 0.01, 0.032, -0.3016, 0.0061, 0.0222, -0.3005, 0.0061, 0.0224, -0.3016, 0.0083, 0.0222, -0.3005, 0.0083, 0.0224, -0.2986, 0.0061, 0.0227,
-0.2986, 0.0083, 0.0227, -0.3008, 0.01, 0.0168, -0.3008, 0.01, 0.019, -0.2986, 0.01, 0.0168, -0.2986, 0.01, 0.019, -0.301, 0.01, 0.0223, -0.3003, 0.01, 0.0224, -0.2986, 0.01, 0.0212, -0.2986, 0.01, 0.0227, -0.2964, 0.0061, 0.023, -0.2964, 0.0083, 0.023, -0.2943, 0.0061, 0.0232, -0.2943, 0.0083, 0.0232, -0.2964, 0.01, 0.0168, -0.2964, 0.01, 0.019, -0.2943, 0.01, 0.0168, -0.2943, 0.01, 0.019, -0.2964, 0.01, 0.0212, -0.2965, 0.01, 0.023, -0.2943, 0.01, 0.0212, -0.2943, 0.01, 0.0232, -0.3008, 0.0061, 0.0297, -0.3008, 0.0083, 0.0297, -0.2986, 0.0061, 0.0299, -0.2986, 0.0083, 0.0299, -0.3007, 0.01, 0.0297, -0.3008, 0.01, 0.032, -0.2986, 0.01, 0.03, -0.2986, 0.01, 0.032, -0.2964, 0.0061, 0.03,
-0.2964, 0.0083, 0.03, -0.2943, 0.0061, 0.0301, -0.2943, 0.0083, 0.0301, -0.2964, 0.01, 0.03, -0.2964, 0.01, 0.032, -0.2943, 0.01, 0.0301, -0.2943, 0.01, 0.032, -0.2889, 0.0061, 0.0064, -0.2889, 0.0083, 0.0064, -0.289, 0.01, 0.0066, -0.2882, 0.0061, -0.0005, -0.2883, 0.0061, 0.0016, -0.2882, 0.0083, -0.0005, -0.2883, 0.0083, 0.0016, -0.2885, 0.0061, 0.0038, -0.2888, 0.0061, 0.0053, -0.2885, 0.0083, 0.0038, -0.2888, 0.0083, 0.0053, -0.2882, 0.01, -0.0006, -0.2883, 0.01, 0.0016, -0.2856, 0.01, -0.0005, -0.2856, 0.01, 0.0016, -0.2885, 0.01, 0.0037, -0.2888, 0.01, 0.0058, -0.2856, 0.01, 0.0038, -0.2856, 0.01, 0.006, -0.2892, 0.0061, 0.0081, -0.2896, 0.0061, 0.0103, -0.2892, 0.0083, 0.0081, -0.2896, 0.0083, 0.0103,
-0.2921, 0.0061, 0.014, -0.2921, 0.0083, 0.014, -0.29, 0.0061, 0.0125, -0.2902, 0.0061, 0.0144, -0.29, 0.0083, 0.0125, -0.2902, 0.0083, 0.0144, -0.2892, 0.01, 0.0081, -0.2896, 0.01, 0.0102, -0.2919, 0.01, 0.014, -0.29, 0.01, 0.0124, -0.2902, 0.01, 0.0144, -0.2878, 0.01, 0.0081, -0.2878, 0.01, 0.0103, -0.2856, 0.01, 0.0081, -0.2856, 0.01, 0.0103, -0.2878, 0.01, 0.0125, -0.2878, 0.01, 0.0146, -0.2856, 0.01, 0.0125, -0.2856, 0.01, 0.0146, -0.2802, 0.0061, 0.0069, -0.2802, 0.0083, 0.0069, -0.2834, 0.01, -0.0005, -0.2834, 0.01, 0.0016, -0.2812, 0.01, -0.0005, -0.2812, 0.01, 0.0016, -0.2834, 0.01, 0.0038, -0.2834, 0.01, 0.006, -0.2812, 0.01, 0.0038, -0.2801, 0.01, 0.0063, -0.2794, 0.0061, -0.0005,
-0.2795, 0.0061, 0.0016, -0.2794, 0.0083, -0.0005, -0.2795, 0.0083, 0.0016, -0.2797, 0.0061, 0.0038, -0.28, 0.0061, 0.0058, -0.2797, 0.0083, 0.0038, -0.28, 0.0083, 0.0058, -0.2794, 0.01, -0.0005, -0.2795, 0.01, 0.0017, -0.2797, 0.01, 0.0038, -0.28, 0.01, 0.0055, -0.2803, 0.0061, 0.0081, -0.2806, 0.0061, 0.0103, -0.2803, 0.0083, 0.0081, -0.2806, 0.0083, 0.0103, -0.281, 0.0061, 0.0125, -0.2813, 0.0061, 0.0146, -0.281, 0.0083, 0.0125, -0.2813, 0.0083, 0.0146, -0.2834, 0.01, 0.0081, -0.2834, 0.01, 0.0103, -0.2804, 0.01, 0.0083, -0.2807, 0.01, 0.0104, -0.2834, 0.01, 0.0125, -0.2834, 0.01, 0.0146, -0.281, 0.01, 0.0126, -0.2813, 0.01, 0.0147, -0.2921, 0.0061, 0.0232, -0.2921, 0.0083, 0.0232, -0.2899, 0.0061, 0.0232,
-0.2899, 0.0083, 0.0232, -0.2921, 0.01, 0.0168, -0.2921, 0.01, 0.019, -0.2899, 0.01, 0.0168, -0.2899, 0.01, 0.019, -0.2921, 0.01, 0.0212, -0.2921, 0.01, 0.0232, -0.2899, 0.01, 0.0212, -0.2899, 0.01, 0.0232, -0.2878, 0.0061, 0.0232, -0.2878, 0.0083, 0.0232, -0.2856, 0.0061, 0.0232, -0.2856, 0.0083, 0.0232, -0.2878, 0.01, 0.0168, -0.2878, 0.01, 0.019, -0.2856, 0.01, 0.0168, -0.2856, 0.01, 0.019, -0.2878, 0.01, 0.0212, -0.2878, 0.01, 0.0232, -0.2856, 0.01, 0.0212, -0.2856, 0.01, 0.0232, -0.2921, 0.0061, 0.0301, -0.2921, 0.0083, 0.0301, -0.2899, 0.0061, 0.0301, -0.2899, 0.0083, 0.0301, -0.2921, 0.01, 0.0301, -0.2921, 0.01, 0.032, -0.2899, 0.01, 0.0301, -0.2899, 0.01, 0.032, -0.2878, 0.0061, 0.0301,
-0.2878, 0.0083, 0.0301, -0.2856, 0.0061, 0.03, -0.2856, 0.0083, 0.03, -0.2878, 0.01, 0.0301, -0.2878, 0.01, 0.032, -0.2856, 0.01, 0.03, -0.2856, 0.01, 0.032, -0.2814, 0.0061, 0.0163, -0.2814, 0.0083, 0.0163, -0.2834, 0.0061, 0.0232, -0.2834, 0.0083, 0.0232, -0.2812, 0.0061, 0.0232, -0.2812, 0.0083, 0.0232, -0.2834, 0.01, 0.0168, -0.2834, 0.01, 0.019, -0.2814, 0.01, 0.0163, -0.2812, 0.01, 0.019, -0.2834, 0.01, 0.0212, -0.2834, 0.01, 0.0232, -0.2812, 0.01, 0.0212, -0.2812, 0.01, 0.0232, -0.2791, 0.0061, 0.0163, -0.2791, 0.0083, 0.0163, -0.2769, 0.0061, 0.0163, -0.2769, 0.0083, 0.0163, -0.2791, 0.0061, 0.0232, -0.2791, 0.0083, 0.0232, -0.2769, 0.0061, 0.023, -0.2769, 0.0083, 0.023, -0.2791, 0.01, 0.0163,
-0.2791, 0.01, 0.019, -0.2769, 0.01, 0.0163, -0.2769, 0.01, 0.019, -0.2791, 0.01, 0.0212, -0.279, 0.01, 0.0232, -0.2769, 0.01, 0.0212, -0.2769, 0.01, 0.023, -0.2834, 0.0061, 0.0299, -0.2834, 0.0083, 0.0299, -0.2812, 0.0061, 0.0296, -0.2812, 0.0083, 0.0296, -0.2835, 0.01, 0.0299, -0.2834, 0.01, 0.032, -0.2813, 0.01, 0.0296, -0.2812, 0.01, 0.032, -0.2764, 0.0061, 0.0286, -0.2764, 0.0083, 0.0286, -0.2791, 0.0061, 0.0292, -0.2791, 0.0083, 0.0292, -0.2774, 0.0061, 0.0289, -0.2774, 0.0083, 0.0289, -0.2762, 0.01, 0.0286, -0.2792, 0.01, 0.0292, -0.2791, 0.01, 0.032, -0.2771, 0.01, 0.0288, -0.2769, 0.01, 0.032, -0.3095, 0.0061, 0.037, -0.3095, 0.0083, 0.037, -0.3073, 0.0061, 0.0371, -0.3073, 0.0083, 0.0371,
-0.3095, 0.01, 0.0342, -0.3095, 0.01, 0.037, -0.3073, 0.01, 0.0342, -0.3074, 0.01, 0.0371, -0.3058, 0.0061, 0.0374, -0.3058, 0.0083, 0.0374, -0.3047, 0.0061, 0.0375, -0.3047, 0.0083, 0.0375, -0.301, 0.0061, 0.038, -0.3029, 0.0061, 0.0407, -0.301, 0.0083, 0.038, -0.3029, 0.0083, 0.0407, -0.3051, 0.01, 0.0342, -0.3053, 0.01, 0.0374, -0.3029, 0.01, 0.0342, -0.3029, 0.01, 0.0363, -0.3045, 0.01, 0.0376, -0.301, 0.01, 0.038, -0.3027, 0.01, 0.0405, -0.3065, 0.0061, 0.0457, -0.3065, 0.0083, 0.0457, -0.3085, 0.0061, 0.0481, -0.3096, 0.0061, 0.0493, -0.3085, 0.0083, 0.0481, -0.3096, 0.0083, 0.0493, -0.3076, 0.0061, 0.0471, -0.3076, 0.0083, 0.0471, -0.3065, 0.01, 0.0457, -0.3085, 0.01, 0.0481, -0.3096, 0.01, 0.0493,
-0.3076, 0.01, 0.047, -0.3073, 0.01, 0.0494, -0.3046, 0.0061, 0.0432, -0.3057, 0.0061, 0.0446, -0.3046, 0.0083, 0.0432, -0.3057, 0.0083, 0.0446, -0.3038, 0.0061, 0.0421, -0.3038, 0.0083, 0.0421, -0.3047, 0.01, 0.0432, -0.3057, 0.01, 0.0447, -0.3039, 0.01, 0.0422, -0.3029, 0.01, 0.045, -0.3051, 0.01, 0.0472, -0.3051, 0.01, 0.0494, -0.3029, 0.01, 0.0472, -0.3029, 0.01, 0.0494, -0.3008, 0.01, 0.0342, -0.3008, 0.01, 0.0363, -0.2986, 0.01, 0.0342, -0.2986, 0.01, 0.0363, -0.3013, 0.01, 0.0384, -0.3008, 0.01, 0.0407, -0.2986, 0.01, 0.0385, -0.2986, 0.01, 0.0407, -0.2964, 0.01, 0.0342, -0.2964, 0.01, 0.0363, -0.2943, 0.01, 0.0342, -0.2943, 0.01, 0.0363, -0.2964, 0.01, 0.0385, -0.2964, 0.01, 0.0407,
-0.2943, 0.01, 0.0385, -0.2943, 0.01, 0.0407, -0.3008, 0.01, 0.0429, -0.3008, 0.01, 0.045, -0.2986, 0.01, 0.0429, -0.2986, 0.01, 0.045, -0.3008, 0.01, 0.0472, -0.3008, 0.01, 0.0494, -0.2986, 0.01, 0.0472, -0.2986, 0.01, 0.0494, -0.2932, 0.0061, 0.0438, -0.294, 0.0061, 0.045, -0.2932, 0.0083, 0.0438, -0.294, 0.0083, 0.045, -0.2957, 0.0061, 0.0477, -0.2968, 0.0061, 0.0494, -0.2957, 0.0083, 0.0477, -0.2968, 0.0083, 0.0494, -0.295, 0.0061, 0.0467, -0.295, 0.0083, 0.0467, -0.2964, 0.01, 0.0429, -0.2964, 0.01, 0.045, -0.2931, 0.01, 0.0436, -0.2942, 0.01, 0.0453, -0.2956, 0.01, 0.0477, -0.2969, 0.01, 0.0496, -0.295, 0.01, 0.0466, -0.3095, 0.01, 0.0515, -0.3095, 0.01, 0.0537, -0.3073, 0.01, 0.0515,
-0.3073, 0.01, 0.0537, -0.3095, 0.01, 0.0559, -0.3095, 0.01, 0.058, -0.3073, 0.01, 0.0559, -0.3073, 0.01, 0.058, -0.3043, 0.0061, 0.0589, -0.3043, 0.0083, 0.0589, -0.3022, 0.0061, 0.0565, -0.3033, 0.0061, 0.0578, -0.3022, 0.0083, 0.0565, -0.3033, 0.0083, 0.0578, -0.3051, 0.01, 0.0515, -0.3051, 0.01, 0.0537, -0.3029, 0.01, 0.0515, -0.3029, 0.01, 0.0537, -0.3051, 0.01, 0.0559, -0.3042, 0.01, 0.0588, -0.3022, 0.01, 0.0565, -0.3033, 0.01, 0.0577, -0.3084, 0.0061, 0.0634, -0.3084, 0.0083, 0.0634, -0.3063, 0.0061, 0.0612, -0.3074, 0.0061, 0.0624, -0.3063, 0.0083, 0.0612, -0.3074, 0.0083, 0.0624, -0.3095, 0.0061, 0.0644, -0.3095, 0.0083, 0.0644, -0.3095, 0.01, 0.0602, -0.3084, 0.01, 0.0634, -0.3063, 0.01, 0.0611,
-0.3073, 0.01, 0.0623, -0.3095, 0.01, 0.0645, -0.3053, 0.0061, 0.0601, -0.3053, 0.0083, 0.0601, -0.3053, 0.01, 0.0601, -0.3002, 0.0061, 0.0541, -0.3002, 0.0083, 0.0541, -0.2983, 0.0061, 0.0516, -0.2994, 0.0061, 0.053, -0.2983, 0.0083, 0.0516, -0.2994, 0.0083, 0.053, -0.3013, 0.0061, 0.0554, -0.3013, 0.0083, 0.0554, -0.3008, 0.01, 0.0515, -0.3002, 0.01, 0.0541, -0.2983, 0.01, 0.0516, -0.2994, 0.01, 0.053, -0.3013, 0.01, 0.0554, -0.2975, 0.0061, 0.0505, -0.2975, 0.0083, 0.0505, -0.2975, 0.01, 0.0505, -0.2902, 0.0061, 0.0369, -0.2902, 0.0083, 0.0369, -0.291, 0.0061, 0.0396, -0.2915, 0.0061, 0.0407, -0.291, 0.0083, 0.0396, -0.2915, 0.0083, 0.0407, -0.2906, 0.0061, 0.0385, -0.2906, 0.0083, 0.0385, -0.2921, 0.01, 0.0342,
-0.2921, 0.01, 0.0363, -0.2899, 0.01, 0.0342, -0.2902, 0.01, 0.0369, -0.2908, 0.01, 0.0391, -0.2916, 0.01, 0.041, -0.2905, 0.01, 0.0382, -0.2878, 0.0061, 0.0369, -0.2878, 0.0083, 0.0369, -0.2856, 0.0061, 0.0367, -0.2856, 0.0083, 0.0367, -0.2878, 0.01, 0.0342, -0.2877, 0.01, 0.0369, -0.2856, 0.01, 0.0342, -0.2855, 0.01, 0.0367, -0.2926, 0.0061, 0.0428, -0.2926, 0.0083, 0.0428, -0.2925, 0.01, 0.0426, -0.2834, 0.0061, 0.0363, -0.2834, 0.0083, 0.0363, -0.2812, 0.0061, 0.0359, -0.2812, 0.0083, 0.0359, -0.2834, 0.01, 0.0342, -0.2833, 0.01, 0.0363, -0.2812, 0.01, 0.0342, -0.2812, 0.01, 0.0359, -0.2783, 0.0061, 0.0352, -0.2794, 0.0061, 0.0355, -0.2783, 0.0083, 0.0352, -0.2794, 0.0083, 0.0355, -0.2769, 0.0061, 0.0348,
-0.2769, 0.0083, 0.0348, -0.2787, 0.01, 0.0353, -0.2796, 0.01, 0.0355, -0.2767, 0.01, 0.0347, -0.3442, 0.0061, 0.0737, -0.3442, 0.0083, 0.0737, -0.342, 0.0061, 0.0728, -0.342, 0.0083, 0.0728, -0.3443, 0.01, 0.0738, -0.3442, 0.01, 0.0754, -0.3423, 0.01, 0.0729, -0.342, 0.01, 0.0754, -0.3397, 0.0061, 0.0717, -0.3397, 0.0083, 0.0717, -0.3377, 0.0061, 0.0707, -0.3377, 0.0083, 0.0707, -0.3408, 0.0061, 0.0722, -0.3408, 0.0083, 0.0722, -0.3395, 0.01, 0.0716, -0.338, 0.01, 0.0708, -0.3405, 0.01, 0.0721, -0.3398, 0.01, 0.0754, -0.3377, 0.01, 0.0732, -0.3377, 0.01, 0.0754, -0.3419, 0.0061, 0.0848, -0.3419, 0.0083, 0.0848, -0.3442, 0.01, 0.0776, -0.3442, 0.01, 0.0798, -0.342, 0.01, 0.0776, -0.342, 0.01, 0.0798,
-0.3442, 0.01, 0.0819, -0.3442, 0.01, 0.0841, -0.342, 0.01, 0.0819, -0.3419, 0.01, 0.0849, -0.3398, 0.0061, 0.0841, -0.3398, 0.0083, 0.0841, -0.3368, 0.0061, 0.0829, -0.3379, 0.0061, 0.0834, -0.3368, 0.0083, 0.0829, -0.3379, 0.0083, 0.0834, -0.3398, 0.01, 0.0776, -0.3398, 0.01, 0.0798, -0.3377, 0.01, 0.0776, -0.3377, 0.01, 0.0798, -0.3398, 0.01, 0.0819, -0.3396, 0.01, 0.0841, -0.3371, 0.01, 0.0831, -0.3381, 0.01, 0.0835, -0.3353, 0.0061, 0.0695, -0.3364, 0.0061, 0.0701, -0.3353, 0.0083, 0.0695, -0.3364, 0.0083, 0.0701, -0.3333, 0.0061, 0.0685, -0.3333, 0.0083, 0.0685, -0.3351, 0.01, 0.0694, -0.3361, 0.01, 0.0699, -0.3336, 0.01, 0.0687, -0.3333, 0.01, 0.0711, -0.3355, 0.01, 0.0732, -0.3355, 0.01, 0.0754,
-0.3333, 0.01, 0.0732, -0.3333, 0.01, 0.0754, -0.3321, 0.0061, 0.0679, -0.3321, 0.0083, 0.0679, -0.3319, 0.01, 0.0677, -0.3312, 0.01, 0.0711, -0.329, 0.01, 0.0689, -0.329, 0.01, 0.0711, -0.3312, 0.01, 0.0732, -0.3312, 0.01, 0.0754, -0.329, 0.01, 0.0732, -0.329, 0.01, 0.0754, -0.3324, 0.0061, 0.0808, -0.3324, 0.0083, 0.0808, -0.3355, 0.0061, 0.0823, -0.3355, 0.0083, 0.0823, -0.3335, 0.0061, 0.0813, -0.3335, 0.0083, 0.0813, -0.3355, 0.01, 0.0776, -0.3355, 0.01, 0.0798, -0.3333, 0.01, 0.0776, -0.3327, 0.01, 0.0809, -0.3352, 0.01, 0.0822, -0.3336, 0.01, 0.0814, -0.3311, 0.0061, 0.0802, -0.3311, 0.0083, 0.0802, -0.3282, 0.0061, 0.0785, -0.3293, 0.0061, 0.0791, -0.3282, 0.0083, 0.0785, -0.3293, 0.0083, 0.0791,
-0.3312, 0.01, 0.0776, -0.3308, 0.01, 0.08, -0.3284, 0.01, 0.0786, -0.3294, 0.01, 0.0792, -0.3442, 0.0061, 0.0856, -0.3442, 0.0083, 0.0856, -0.343, 0.0061, 0.0852, -0.343, 0.0083, 0.0852, -0.3441, 0.01, 0.0856, -0.343, 0.01, 0.0852, -0.3242, 0.0061, 0.0761, -0.3242, 0.0083, 0.0761, -0.3268, 0.01, 0.0689, -0.3268, 0.01, 0.0711, -0.3246, 0.01, 0.0689, -0.3246, 0.01, 0.0711, -0.3268, 0.01, 0.0732, -0.3268, 0.01, 0.0754, -0.3246, 0.01, 0.0732, -0.3243, 0.01, 0.0761, -0.3215, 0.0061, 0.0742, -0.3226, 0.0061, 0.075, -0.3215, 0.0083, 0.0742, -0.3226, 0.0083, 0.075, -0.3203, 0.0061, 0.0734, -0.3203, 0.0083, 0.0734, -0.3225, 0.01, 0.0689, -0.3225, 0.01, 0.0711, -0.3203, 0.01, 0.0689, -0.3203, 0.01, 0.0711,
-0.3217, 0.01, 0.0743, -0.3227, 0.01, 0.0751, -0.32, 0.01, 0.0732, -0.3268, 0.0061, 0.0777, -0.3268, 0.0083, 0.0777, -0.3253, 0.0061, 0.0768, -0.3253, 0.0083, 0.0768, -0.3265, 0.01, 0.0775, -0.3254, 0.01, 0.0768, -0.3178, 0.0061, 0.0716, -0.3178, 0.0083, 0.0716, -0.3153, 0.0061, 0.0696, -0.3164, 0.0061, 0.0705, -0.3153, 0.0083, 0.0696, -0.3164, 0.0083, 0.0705, -0.3189, 0.0061, 0.0724, -0.3189, 0.0083, 0.0724, -0.3181, 0.01, 0.0689, -0.3178, 0.01, 0.0716, -0.3153, 0.01, 0.0696, -0.3164, 0.01, 0.0706, -0.3189, 0.01, 0.0724, -0.3141, 0.0061, 0.0685, -0.3141, 0.0083, 0.0685, -0.3141, 0.01, 0.0686, -0.2653, 0.0061, 0.0157, -0.2653, 0.0083, 0.0157, -0.2639, 0.0061, 0.0154, -0.2639, 0.0083, 0.0154, -0.2652, 0.01, 0.0157,
-0.2639, 0.01, 0.0154, -0.2617, 0.0061, 0.015, -0.2617, 0.0083, 0.015, -0.2595, 0.0061, 0.0146, -0.2595, 0.0083, 0.0146, -0.2618, 0.01, 0.015, -0.2596, 0.01, 0.0146, -0.2747, 0.0061, 0.0163, -0.2747, 0.0083, 0.0163, -0.2726, 0.0061, 0.0163, -0.2726, 0.0083, 0.0163, -0.2747, 0.0061, 0.0226, -0.2747, 0.0083, 0.0226, -0.2723, 0.0061, 0.022, -0.2734, 0.0061, 0.0223, -0.2723, 0.0083, 0.022, -0.2734, 0.0083, 0.0223, -0.2747, 0.01, 0.0163, -0.2747, 0.01, 0.019, -0.2726, 0.01, 0.0163, -0.2726, 0.01, 0.019, -0.2747, 0.01, 0.0212, -0.2747, 0.01, 0.0226, -0.2725, 0.01, 0.0221, -0.2735, 0.01, 0.0223, -0.2704, 0.0061, 0.0163, -0.2704, 0.0083, 0.0163, -0.2682, 0.0061, 0.0161, -0.2682, 0.0083, 0.0161, -0.2704, 0.0061, 0.0216,
-0.2704, 0.0083, 0.0216, -0.2682, 0.0061, 0.0211, -0.2682, 0.0083, 0.0211, -0.2704, 0.01, 0.0163, -0.2704, 0.01, 0.019, -0.2683, 0.01, 0.0161, -0.2682, 0.01, 0.019, -0.2702, 0.01, 0.0216, -0.2681, 0.01, 0.0211, -0.2747, 0.0061, 0.0283, -0.2747, 0.0083, 0.0283, -0.2726, 0.0061, 0.0277, -0.2726, 0.0083, 0.0277, -0.2717, 0.0061, 0.033, -0.2717, 0.0083, 0.033, -0.2748, 0.01, 0.0283, -0.2727, 0.01, 0.0277, -0.2747, 0.01, 0.0298, -0.2747, 0.01, 0.032, -0.2726, 0.01, 0.0298, -0.2721, 0.01, 0.0331, -0.2704, 0.0061, 0.027, -0.2704, 0.0083, 0.027, -0.2681, 0.0061, 0.0263, -0.2692, 0.0061, 0.0266, -0.2681, 0.0083, 0.0263, -0.2692, 0.0083, 0.0266, -0.2704, 0.0061, 0.0325, -0.2704, 0.0083, 0.0325, -0.2682, 0.0061, 0.0317,
-0.2682, 0.0083, 0.0317, -0.2706, 0.01, 0.0271, -0.2678, 0.01, 0.0262, -0.2687, 0.01, 0.0265, -0.2704, 0.01, 0.0298, -0.2701, 0.01, 0.0324, -0.2682, 0.01, 0.0298, -0.2681, 0.01, 0.0316, -0.2664, 0.0061, 0.0159, -0.2664, 0.0083, 0.0159, -0.2635, 0.0061, 0.0198, -0.2635, 0.0083, 0.0198, -0.266, 0.0061, 0.0206, -0.266, 0.0083, 0.0206, -0.2646, 0.0061, 0.0202, -0.2646, 0.0083, 0.0202, -0.2661, 0.01, 0.0158, -0.2661, 0.01, 0.019, -0.2639, 0.01, 0.0168, -0.2637, 0.01, 0.0199, -0.266, 0.01, 0.0206, -0.2647, 0.01, 0.0202, -0.2617, 0.0061, 0.0193, -0.2617, 0.0083, 0.0193, -0.2595, 0.0061, 0.0185, -0.2595, 0.0083, 0.0185, -0.2616, 0.0061, 0.0241, -0.2616, 0.0083, 0.0241, -0.2595, 0.0061, 0.0232, -0.2595, 0.0083, 0.0232,
-0.2617, 0.01, 0.0168, -0.2615, 0.01, 0.0192, -0.2595, 0.01, 0.0168, -0.2594, 0.01, 0.0185, -0.2614, 0.01, 0.024, -0.2598, 0.01, 0.0233, -0.2661, 0.0061, 0.0256, -0.2661, 0.0083, 0.0256, -0.2639, 0.0061, 0.0249, -0.2639, 0.0083, 0.0249, -0.2657, 0.0061, 0.0306, -0.2668, 0.0061, 0.0311, -0.2657, 0.0083, 0.0306, -0.2668, 0.0083, 0.0311, -0.2639, 0.0061, 0.0298, -0.2639, 0.0083, 0.0298, -0.2662, 0.01, 0.0256, -0.2661, 0.01, 0.0277, -0.2641, 0.01, 0.025, -0.2639, 0.01, 0.0277, -0.2658, 0.01, 0.0307, -0.2669, 0.01, 0.0311, -0.2637, 0.01, 0.0297, -0.2627, 0.0061, 0.0245, -0.2611, 0.0061, 0.0285, -0.2627, 0.0083, 0.0245, -0.2611, 0.0083, 0.0285, -0.2595, 0.0061, 0.0276, -0.2595, 0.0083, 0.0276, -0.2622, 0.0061, 0.029,
-0.2622, 0.0083, 0.029, -0.2622, 0.01, 0.0243, -0.2613, 0.01, 0.0286, -0.2595, 0.01, 0.0255, -0.2593, 0.01, 0.0275, -0.2623, 0.01, 0.0291, -0.2574, 0.0061, 0.0142, -0.2574, 0.0083, 0.0142, -0.2542, 0.0061, 0.0136, -0.2552, 0.0061, 0.0138, -0.2542, 0.0083, 0.0136, -0.2552, 0.0083, 0.0138, -0.2575, 0.01, 0.0142, -0.2541, 0.01, 0.0135, -0.2551, 0.01, 0.0138, -0.253, 0.0061, 0.0133, -0.2526, 0.0061, 0.0154, -0.253, 0.0083, 0.0133, -0.2526, 0.0083, 0.0154, -0.2509, 0.0061, 0.0128, -0.2508, 0.0061, 0.0145, -0.2509, 0.0083, 0.0128, -0.2508, 0.0083, 0.0145, -0.2531, 0.01, 0.0133, -0.2527, 0.01, 0.0155, -0.2509, 0.01, 0.0128, -0.2506, 0.01, 0.0144, -0.2467, 0.0061, 0.0121, -0.2496, 0.0061, 0.0137, -0.2467, 0.0083, 0.0121,
-0.2496, 0.0083, 0.0137, -0.2497, 0.01, 0.0131, -0.2496, 0.01, 0.0137, -0.2571, 0.0061, 0.0176, -0.2582, 0.0061, 0.018, -0.2571, 0.0083, 0.0176, -0.2582, 0.0083, 0.018, -0.2552, 0.0061, 0.0167, -0.2552, 0.0083, 0.0167, -0.2569, 0.0061, 0.022, -0.258, 0.0061, 0.0225, -0.2569, 0.0083, 0.022, -0.258, 0.0083, 0.0225, -0.2552, 0.0061, 0.0211, -0.2542, 0.0061, 0.0244, -0.2552, 0.0083, 0.0211, -0.2542, 0.0083, 0.0244, -0.2572, 0.01, 0.0176, -0.2583, 0.01, 0.018, -0.255, 0.01, 0.0166, -0.2568, 0.01, 0.0219, -0.2578, 0.01, 0.0224, -0.2554, 0.01, 0.0212, -0.2544, 0.01, 0.0245, -0.2537, 0.0061, 0.016, -0.2525, 0.0061, 0.0198, -0.2537, 0.0083, 0.016, -0.2525, 0.0083, 0.0198, -0.2509, 0.0061, 0.019, -0.2509, 0.0083, 0.019,
-0.2536, 0.0061, 0.0203, -0.253, 0.0061, 0.0235, -0.2536, 0.0083, 0.0203, -0.253, 0.0083, 0.0235, -0.2505, 0.0061, 0.0217, -0.2516, 0.0061, 0.0225, -0.2505, 0.0083, 0.0217, -0.2516, 0.0083, 0.0225, -0.2538, 0.01, 0.016, -0.2524, 0.01, 0.0197, -0.2511, 0.01, 0.0191, -0.2534, 0.01, 0.0203, -0.2527, 0.01, 0.0233, -0.2505, 0.01, 0.0217, -0.2516, 0.01, 0.0225, -0.257, 0.0061, 0.0262, -0.2581, 0.0061, 0.0268, -0.257, 0.0083, 0.0262, -0.2581, 0.0083, 0.0268, -0.2553, 0.0061, 0.0251, -0.2553, 0.0083, 0.0251, -0.257, 0.01, 0.0262, -0.2581, 0.01, 0.0268, -0.2554, 0.01, 0.0252, -0.2481, 0.0061, 0.0176, -0.2486, 0.0061, 0.0189, -0.2481, 0.0083, 0.0176, -0.2486, 0.0083, 0.0189, -0.2449, 0.0061, 0.0163, -0.2469, 0.0061, 0.0186,
-0.2449, 0.0083, 0.0163, -0.2469, 0.0083, 0.0186, -0.2491, 0.0061, 0.0206, -0.2491, 0.0083, 0.0206, -0.248, 0.01, 0.0176, -0.2486, 0.01, 0.019, -0.2449, 0.01, 0.0163, -0.2469, 0.01, 0.0186, -0.2491, 0.01, 0.0207, -0.2747, 0.0061, 0.0341, -0.2747, 0.0083, 0.0341, -0.2728, 0.0061, 0.0334, -0.2728, 0.0083, 0.0334, -0.2746, 0.01, 0.034, -0.273, 0.01, 0.0335, -0.2196, 0.0061, -0.0005, -0.2196, 0.0061, 0.0016, -0.2196, 0.0083, -0.0005, -0.2196, 0.0083, 0.0016, -0.2196, 0.0061, 0.0038, -0.2196, 0.0061, 0.006, -0.2196, 0.0083, 0.0038, -0.2196, 0.0083, 0.006, -0.2196, 0.01, -0.0005, -0.2196, 0.01, 0.0016, -0.2196, 0.01, 0.0038, -0.2196, 0.01, 0.006, -0.2183, 0.01, -0.0005, -0.2183, 0.01, 0.0016, -0.2161, 0.01, -0.0005,
-0.2161, 0.01, 0.0016, -0.2183, 0.01, 0.0038, -0.2183, 0.01, 0.006, -0.2161, 0.01, 0.0038, -0.2161, 0.01, 0.006, -0.2196, 0.0061, 0.0081, -0.2196, 0.0061, 0.0103, -0.2196, 0.0083, 0.0081, -0.2196, 0.0083, 0.0103, -0.2196, 0.0061, 0.0125, -0.2196, 0.0061, 0.0146, -0.2196, 0.0083, 0.0125, -0.2196, 0.0083, 0.0146, -0.2196, 0.01, 0.0081, -0.2196, 0.01, 0.0103, -0.2196, 0.01, 0.0125, -0.2196, 0.01, 0.0146, -0.2183, 0.01, 0.0081, -0.2183, 0.01, 0.0103, -0.2161, 0.01, 0.0081, -0.2161, 0.01, 0.0103, -0.2183, 0.01, 0.0125, -0.2183, 0.01, 0.0146, -0.2161, 0.01, 0.0125, -0.2161, 0.01, 0.0146, -0.214, 0.01, -0.0005, -0.214, 0.01, 0.0016, -0.2118, 0.01, -0.0005, -0.2118, 0.01, 0.0016, -0.214, 0.01, 0.0038,
-0.214, 0.01, 0.006, -0.2118, 0.01, 0.0038, -0.2118, 0.01, 0.006, -0.2088, 0.0061, -0.0005, -0.2088, 0.0061, 0.0016, -0.2088, 0.0083, -0.0005, -0.2088, 0.0083, 0.0016, -0.2088, 0.0061, 0.0038, -0.2088, 0.0061, 0.006, -0.2088, 0.0083, 0.0038, -0.2088, 0.0083, 0.006, -0.2088, 0.01, -0.0005, -0.2088, 0.01, 0.0016, -0.2088, 0.01, 0.0038, -0.2088, 0.01, 0.006, -0.214, 0.01, 0.0081, -0.214, 0.01, 0.0103, -0.2118, 0.01, 0.0081, -0.2118, 0.01, 0.0103, -0.214, 0.01, 0.0125, -0.214, 0.01, 0.0146, -0.2118, 0.01, 0.0125, -0.2118, 0.01, 0.0146, -0.2088, 0.0061, 0.0085, -0.2088, 0.0083, 0.0085, -0.2075, 0.0061, 0.0085, -0.2075, 0.0083, 0.0085, -0.2088, 0.01, 0.0085, -0.2096, 0.01, 0.0103, -0.2075, 0.01, 0.0085,
-0.2075, 0.01, 0.0103, -0.2096, 0.01, 0.0125, -0.2096, 0.01, 0.0146, -0.2075, 0.01, 0.0125, -0.2075, 0.01, 0.0146, -0.2196, 0.0061, 0.0168, -0.2196, 0.0061, 0.019, -0.2196, 0.0083, 0.0168, -0.2196, 0.0083, 0.019, -0.2196, 0.0061, 0.0212, -0.2196, 0.0061, 0.0233, -0.2196, 0.0083, 0.0212, -0.2196, 0.0083, 0.0233, -0.2196, 0.01, 0.0168, -0.2196, 0.01, 0.019, -0.2196, 0.01, 0.0212, -0.2196, 0.01, 0.0233, -0.2183, 0.01, 0.0168, -0.2183, 0.01, 0.019, -0.2161, 0.01, 0.0168, -0.2161, 0.01, 0.019, -0.2183, 0.01, 0.0212, -0.2183, 0.01, 0.0233, -0.2161, 0.01, 0.0212, -0.2161, 0.01, 0.0233, -0.2196, 0.0061, 0.0255, -0.2196, 0.0061, 0.0277, -0.2196, 0.0083, 0.0255, -0.2196, 0.0083, 0.0277, -0.2196, 0.0061, 0.0298,
-0.2196, 0.0061, 0.032, -0.2196, 0.0083, 0.0298, -0.2196, 0.0083, 0.032, -0.2196, 0.01, 0.0255, -0.2196, 0.01, 0.0277, -0.2196, 0.01, 0.0298, -0.2196, 0.01, 0.032, -0.2183, 0.01, 0.0255, -0.2183, 0.01, 0.0277, -0.2161, 0.01, 0.0255, -0.2161, 0.01, 0.0277, -0.2183, 0.01, 0.0298, -0.2183, 0.01, 0.032, -0.2161, 0.01, 0.0298, -0.2161, 0.01, 0.032, -0.214, 0.01, 0.0168, -0.214, 0.01, 0.019, -0.2118, 0.01, 0.0168, -0.2118, 0.01, 0.019, -0.214, 0.01, 0.0212, -0.214, 0.01, 0.0233, -0.2118, 0.01, 0.0212, -0.2118, 0.01, 0.0233, -0.2088, 0.0061, 0.0203, -0.2088, 0.0061, 0.0233, -0.2088, 0.0083, 0.0203, -0.2088, 0.0083, 0.0233, -0.2075, 0.0061, 0.0203, -0.2075, 0.0083, 0.0203, -0.2096, 0.01, 0.0168,
-0.2096, 0.01, 0.019, -0.2075, 0.01, 0.0168, -0.2075, 0.01, 0.019, -0.2088, 0.01, 0.0203, -0.2088, 0.01, 0.0233, -0.2075, 0.01, 0.0203, -0.214, 0.01, 0.0255, -0.214, 0.01, 0.0277, -0.2118, 0.01, 0.0255, -0.2118, 0.01, 0.0277, -0.214, 0.01, 0.0298, -0.214, 0.01, 0.032, -0.2118, 0.01, 0.0298, -0.2118, 0.01, 0.032, -0.2088, 0.0061, 0.0255, -0.2088, 0.0061, 0.0277, -0.2088, 0.0083, 0.0255, -0.2088, 0.0083, 0.0277, -0.2088, 0.0061, 0.0298, -0.2088, 0.0061, 0.032, -0.2088, 0.0083, 0.0298, -0.2088, 0.0083, 0.032, -0.2088, 0.01, 0.0255, -0.2088, 0.01, 0.0277, -0.2088, 0.01, 0.0298, -0.2088, 0.01, 0.0319, -0.2196, 0.0061, 0.0342, -0.2196, 0.0061, 0.0363, -0.2196, 0.0083, 0.0342, -0.2196, 0.0083, 0.0363,
-0.2196, 0.0061, 0.0385, -0.2196, 0.0061, 0.0407, -0.2196, 0.0083, 0.0385, -0.2196, 0.0083, 0.0407, -0.2196, 0.01, 0.0342, -0.2196, 0.01, 0.0363, -0.2196, 0.01, 0.0385, -0.2196, 0.01, 0.0407, -0.2183, 0.01, 0.0342, -0.2183, 0.01, 0.0363, -0.2161, 0.01, 0.0342, -0.2161, 0.01, 0.0363, -0.2183, 0.01, 0.0385, -0.2183, 0.01, 0.0407, -0.2161, 0.01, 0.0385, -0.2161, 0.01, 0.0407, -0.2196, 0.0061, 0.0429, -0.2196, 0.0061, 0.045, -0.2196, 0.0083, 0.0429, -0.2196, 0.0083, 0.045, -0.2196, 0.0061, 0.0472, -0.2196, 0.0061, 0.0487, -0.2196, 0.0083, 0.0472, -0.2196, 0.0083, 0.0487, -0.2196, 0.01, 0.0429, -0.2196, 0.01, 0.045, -0.2196, 0.01, 0.0472, -0.2196, 0.01, 0.0487, -0.2183, 0.0061, 0.0487, -0.2183, 0.0083, 0.0487,
-0.2161, 0.0061, 0.0487, -0.2161, 0.0083, 0.0487, -0.2183, 0.01, 0.0429, -0.2183, 0.01, 0.045, -0.2161, 0.01, 0.0429, -0.2161, 0.01, 0.045, -0.2183, 0.01, 0.0472, -0.2183, 0.01, 0.0487, -0.2161, 0.01, 0.0472, -0.2161, 0.01, 0.0487, -0.214, 0.01, 0.0342, -0.214, 0.01, 0.0363, -0.2118, 0.01, 0.0342, -0.2118, 0.01, 0.0363, -0.214, 0.01, 0.0385, -0.214, 0.01, 0.0407, -0.2118, 0.01, 0.0385, -0.2118, 0.01, 0.0407, -0.2086, 0.0061, 0.0334, -0.2086, 0.0083, 0.0334, -0.2083, 0.0061, 0.0345, -0.2076, 0.0061, 0.0361, -0.2083, 0.0083, 0.0345, -0.2076, 0.0083, 0.0361, -0.2085, 0.01, 0.0339, -0.2096, 0.01, 0.0363, -0.2083, 0.01, 0.0347, -0.2076, 0.01, 0.0361, -0.2096, 0.01, 0.0385, -0.2096, 0.01, 0.0407,
-0.2075, 0.01, 0.0385, -0.2075, 0.01, 0.0407, -0.214, 0.0061, 0.0487, -0.214, 0.0083, 0.0487, -0.2118, 0.0061, 0.0487, -0.2118, 0.0083, 0.0487, -0.214, 0.01, 0.0429, -0.214, 0.01, 0.045, -0.2118, 0.01, 0.0429, -0.2118, 0.01, 0.045, -0.214, 0.01, 0.0472, -0.214, 0.01, 0.0487, -0.2118, 0.01, 0.0472, -0.2118, 0.01, 0.0487, -0.2096, 0.0061, 0.0487, -0.2096, 0.0083, 0.0487, -0.2075, 0.0061, 0.0487, -0.2075, 0.0083, 0.0487, -0.2096, 0.01, 0.0429, -0.2096, 0.01, 0.045, -0.2075, 0.01, 0.0429, -0.2075, 0.01, 0.045, -0.2096, 0.01, 0.0472, -0.2096, 0.01, 0.0487, -0.2075, 0.01, 0.0472, -0.2075, 0.01, 0.0487, -0.2053, 0.0061, 0.0085, -0.2053, 0.0083, 0.0085, -0.2031, 0.0061, 0.0085, -0.2031, 0.0083, 0.0085,
-0.2053, 0.01, 0.0085, -0.2053, 0.01, 0.0103, -0.2031, 0.01, 0.0085, -0.2031, 0.01, 0.0103, -0.2053, 0.01, 0.0125, -0.2053, 0.01, 0.0146, -0.2031, 0.01, 0.0125, -0.2031, 0.01, 0.0146, -0.2009, 0.0061, 0.0085, -0.2009, 0.0083, 0.0085, -0.1988, 0.0061, 0.0085, -0.1988, 0.0083, 0.0085, -0.2009, 0.01, 0.0085, -0.2009, 0.01, 0.0103, -0.1988, 0.01, 0.0085, -0.1988, 0.01, 0.0103, -0.2009, 0.01, 0.0125, -0.2009, 0.01, 0.0146, -0.1988, 0.01, 0.0125, -0.1988, 0.01, 0.0146, -0.1966, 0.0061, 0.0085, -0.1966, 0.0083, 0.0085, -0.1944, 0.0061, 0.0085, -0.1944, 0.0083, 0.0085, -0.1966, 0.01, 0.0085, -0.1966, 0.01, 0.0103, -0.1944, 0.01, 0.0085, -0.1944, 0.01, 0.0103, -0.1966, 0.01, 0.0125, -0.1966, 0.01, 0.0146,
-0.1944, 0.01, 0.0125, -0.1944, 0.01, 0.0146, -0.1923, 0.0061, 0.0085, -0.1923, 0.0083, 0.0085, -0.1901, 0.0061, 0.0085, -0.1901, 0.0083, 0.0085, -0.1923, 0.01, 0.0085, -0.1923, 0.01, 0.0103, -0.1901, 0.01, 0.0085, -0.1901, 0.01, 0.0103, -0.1923, 0.01, 0.0125, -0.1923, 0.01, 0.0146, -0.1901, 0.01, 0.0125, -0.1901, 0.01, 0.0146, -0.2053, 0.0061, 0.0203, -0.2053, 0.0083, 0.0203, -0.2031, 0.0061, 0.0203, -0.2031, 0.0083, 0.0203, -0.2053, 0.01, 0.0168, -0.2053, 0.01, 0.019, -0.2031, 0.01, 0.0168, -0.2031, 0.01, 0.019, -0.2053, 0.01, 0.0203, -0.2031, 0.01, 0.0203, -0.2009, 0.0061, 0.0203, -0.2009, 0.0083, 0.0203, -0.1988, 0.0061, 0.0203, -0.1988, 0.0083, 0.0203, -0.2009, 0.01, 0.0168, -0.2009, 0.01, 0.019,
-0.1988, 0.01, 0.0168, -0.1988, 0.01, 0.019, -0.2009, 0.01, 0.0203, -0.1988, 0.01, 0.0203, -0.1966, 0.0061, 0.0203, -0.1966, 0.0083, 0.0203, -0.1944, 0.0061, 0.0203, -0.1944, 0.0083, 0.0203, -0.1966, 0.01, 0.0168, -0.1966, 0.01, 0.019, -0.1944, 0.01, 0.0168, -0.1944, 0.01, 0.019, -0.1966, 0.01, 0.0203, -0.1944, 0.01, 0.0203, -0.1923, 0.0061, 0.0203, -0.1923, 0.0083, 0.0203, -0.1901, 0.0061, 0.0203, -0.1901, 0.0083, 0.0203, -0.1923, 0.01, 0.0168, -0.1923, 0.01, 0.019, -0.1901, 0.01, 0.0168, -0.1901, 0.01, 0.019, -0.1923, 0.01, 0.0203, -0.1901, 0.01, 0.0203, -0.1879, 0.0061, 0.0085, -0.1879, 0.0083, 0.0085, -0.1858, 0.0061, 0.0085, -0.1858, 0.0083, 0.0085, -0.1879, 0.01, 0.0085, -0.1879, 0.01, 0.0103,
-0.1858, 0.01, 0.0085, -0.1858, 0.01, 0.0103, -0.1879, 0.01, 0.0125, -0.1879, 0.01, 0.0146, -0.1858, 0.01, 0.0125, -0.1858, 0.01, 0.0146, -0.1836, 0.0061, 0.0085, -0.1836, 0.0083, 0.0085, -0.1814, 0.0061, 0.0085, -0.1814, 0.0083, 0.0085, -0.1836, 0.01, 0.0085, -0.1836, 0.01, 0.0103, -0.1814, 0.01, 0.0085, -0.1814, 0.01, 0.0103, -0.1836, 0.01, 0.0125, -0.1836, 0.01, 0.0146, -0.1814, 0.01, 0.0125, -0.1814, 0.01, 0.0146, -0.1792, 0.0061, 0.0085, -0.1792, 0.0083, 0.0085, -0.1771, 0.0061, 0.0085, -0.1771, 0.0083, 0.0085, -0.1792, 0.01, 0.0085, -0.1792, 0.01, 0.0103, -0.1771, 0.01, 0.0085, -0.1771, 0.01, 0.0103, -0.1792, 0.01, 0.0125, -0.1792, 0.01, 0.0146, -0.1771, 0.01, 0.0125, -0.1771, 0.01, 0.0146,
-0.1749, 0.0061, 0.0085, -0.1749, 0.0083, 0.0085, -0.1727, 0.0061, 0.0085, -0.1727, 0.0083, 0.0085, -0.1749, 0.01, 0.0085, -0.1749, 0.01, 0.0103, -0.1727, 0.01, 0.0085, -0.1727, 0.01, 0.0103, -0.1749, 0.01, 0.0125, -0.1749, 0.01, 0.0146, -0.1727, 0.01, 0.0125, -0.1727, 0.01, 0.0146, -0.1879, 0.0061, 0.0203, -0.1879, 0.0083, 0.0203, -0.1858, 0.0061, 0.0203, -0.1858, 0.0083, 0.0203, -0.1879, 0.01, 0.0168, -0.1879, 0.01, 0.019, -0.1858, 0.01, 0.0168, -0.1858, 0.01, 0.019, -0.1879, 0.01, 0.0203, -0.1858, 0.01, 0.0203, -0.1836, 0.0061, 0.0203, -0.1836, 0.0083, 0.0203, -0.1814, 0.0061, 0.0203, -0.1814, 0.0083, 0.0203, -0.1836, 0.01, 0.0168, -0.1836, 0.01, 0.019, -0.1814, 0.01, 0.0168, -0.1814, 0.01, 0.019,
-0.1836, 0.01, 0.0203, -0.1814, 0.01, 0.0203, -0.1792, 0.0061, 0.0203, -0.1792, 0.0083, 0.0203, -0.1771, 0.0061, 0.0203, -0.1771, 0.0083, 0.0203, -0.1792, 0.01, 0.0168, -0.1792, 0.01, 0.019, -0.1771, 0.01, 0.0168, -0.1771, 0.01, 0.019, -0.1792, 0.01, 0.0203, -0.1771, 0.01, 0.0203, -0.1749, 0.0061, 0.0203, -0.1749, 0.0083, 0.0203, -0.1727, 0.0061, 0.0203, -0.1727, 0.0083, 0.0203, -0.1749, 0.01, 0.0168, -0.1749, 0.01, 0.019, -0.1727, 0.01, 0.0168, -0.1727, 0.01, 0.019, -0.1749, 0.01, 0.0203, -0.1727, 0.01, 0.0203, -0.2059, 0.0061, 0.0372, -0.2059, 0.0083, 0.0372, -0.2048, 0.0061, 0.0376, -0.2048, 0.0083, 0.0376, -0.2031, 0.0061, 0.0378, -0.2031, 0.0083, 0.0378, -0.206, 0.01, 0.0372, -0.2051, 0.01, 0.0375,
-0.2053, 0.01, 0.0407, -0.203, 0.01, 0.0378, -0.2031, 0.01, 0.0407, -0.2009, 0.0061, 0.0379, -0.2009, 0.0083, 0.0379, -0.1988, 0.0061, 0.0379, -0.1988, 0.0083, 0.0379, -0.2009, 0.01, 0.0379, -0.2009, 0.01, 0.0407, -0.1988, 0.01, 0.0379, -0.1988, 0.01, 0.0407, -0.2053, 0.0061, 0.0487, -0.2053, 0.0083, 0.0487, -0.2031, 0.0061, 0.0487, -0.2031, 0.0083, 0.0487, -0.2053, 0.01, 0.0429, -0.2053, 0.01, 0.045, -0.2031, 0.01, 0.0429, -0.2031, 0.01, 0.045, -0.2053, 0.01, 0.0472, -0.2053, 0.01, 0.0487, -0.2031, 0.01, 0.0472, -0.2031, 0.01, 0.0487, -0.2009, 0.0061, 0.0487, -0.2009, 0.0083, 0.0487, -0.1988, 0.0061, 0.0487, -0.1988, 0.0083, 0.0487, -0.2009, 0.01, 0.0429, -0.2009, 0.01, 0.045, -0.1988, 0.01, 0.0429,
-0.1988, 0.01, 0.045, -0.2009, 0.01, 0.0472, -0.2009, 0.01, 0.0487, -0.1988, 0.01, 0.0472, -0.1988, 0.01, 0.0487, -0.1966, 0.0061, 0.0379, -0.1966, 0.0083, 0.0379, -0.1944, 0.0061, 0.0379, -0.1944, 0.0083, 0.0379, -0.1966, 0.01, 0.0379, -0.1966, 0.01, 0.0407, -0.1944, 0.01, 0.0379, -0.1944, 0.01, 0.0407, -0.1923, 0.0061, 0.0379, -0.1923, 0.0083, 0.0379, -0.1901, 0.0061, 0.0379, -0.1901, 0.0083, 0.0379, -0.1923, 0.01, 0.0379, -0.1923, 0.01, 0.0407, -0.1901, 0.01, 0.0379, -0.1901, 0.01, 0.0407, -0.1966, 0.0061, 0.0487, -0.1966, 0.0083, 0.0487, -0.1944, 0.0061, 0.0487, -0.1944, 0.0083, 0.0487, -0.1966, 0.01, 0.0429, -0.1966, 0.01, 0.045, -0.1944, 0.01, 0.0429, -0.1944, 0.01, 0.045, -0.1966, 0.01, 0.0472,
-0.1966, 0.01, 0.0487, -0.1944, 0.01, 0.0472, -0.1944, 0.01, 0.0487, -0.1923, 0.0061, 0.0487, -0.1923, 0.0083, 0.0487, -0.1901, 0.0061, 0.0487, -0.1901, 0.0083, 0.0487, -0.1923, 0.01, 0.0429, -0.1923, 0.01, 0.045, -0.1901, 0.01, 0.0429, -0.1901, 0.01, 0.045, -0.1923, 0.01, 0.0472, -0.1923, 0.01, 0.0487, -0.1901, 0.01, 0.0472, -0.1901, 0.01, 0.0487, -0.1879, 0.0061, 0.0379, -0.1879, 0.0083, 0.0379, -0.1858, 0.0061, 0.0379, -0.1858, 0.0083, 0.0379, -0.1879, 0.01, 0.0379, -0.1879, 0.01, 0.0407, -0.1858, 0.01, 0.0379, -0.1858, 0.01, 0.0407, -0.1836, 0.0061, 0.0379, -0.1836, 0.0083, 0.0379, -0.1814, 0.0061, 0.0379, -0.1814, 0.0083, 0.0379, -0.1836, 0.01, 0.0379, -0.1836, 0.01, 0.0407, -0.1814, 0.01, 0.0379,
-0.1814, 0.01, 0.0407, -0.1879, 0.0061, 0.0487, -0.1879, 0.0083, 0.0487, -0.1858, 0.0061, 0.0487, -0.1858, 0.0083, 0.0487, -0.1879, 0.01, 0.0429, -0.1879, 0.01, 0.045, -0.1858, 0.01, 0.0429, -0.1858, 0.01, 0.045, -0.1879, 0.01, 0.0472, -0.1879, 0.01, 0.0487, -0.1858, 0.01, 0.0472, -0.1858, 0.01, 0.0487, -0.1836, 0.0061, 0.0487, -0.1836, 0.0083, 0.0487, -0.1814, 0.0061, 0.0487, -0.1814, 0.0083, 0.0487, -0.1836, 0.01, 0.0429, -0.1836, 0.01, 0.045, -0.1814, 0.01, 0.0429, -0.1814, 0.01, 0.045, -0.1836, 0.01, 0.0472, -0.1836, 0.01, 0.0487, -0.1814, 0.01, 0.0472, -0.1814, 0.01, 0.0487, -0.1792, 0.0061, 0.0379, -0.1792, 0.0083, 0.0379, -0.1771, 0.0061, 0.0379, -0.1771, 0.0083, 0.0379, -0.1792, 0.01, 0.0379,
-0.1792, 0.01, 0.0407, -0.1771, 0.01, 0.0379, -0.1771, 0.01, 0.0407, -0.1749, 0.0061, 0.0379, -0.1749, 0.0083, 0.0379, -0.1727, 0.0061, 0.0379, -0.1727, 0.0083, 0.0379, -0.1749, 0.01, 0.0379, -0.1749, 0.01, 0.0407, -0.1727, 0.01, 0.0379, -0.1727, 0.01, 0.0407, -0.1792, 0.0061, 0.0487, -0.1792, 0.0083, 0.0487, -0.1771, 0.0061, 0.0487, -0.1771, 0.0083, 0.0487, -0.1792, 0.01, 0.0429, -0.1792, 0.01, 0.045, -0.1771, 0.01, 0.0429, -0.1771, 0.01, 0.045, -0.1792, 0.01, 0.0472, -0.1792, 0.01, 0.0487, -0.1771, 0.01, 0.0472, -0.1771, 0.01, 0.0487, -0.1749, 0.0061, 0.0487, -0.1749, 0.0083, 0.0487, -0.1727, 0.0061, 0.0487, -0.1727, 0.0083, 0.0487, -0.1749, 0.01, 0.0429, -0.1749, 0.01, 0.045, -0.1727, 0.01, 0.0429,
-0.1727, 0.01, 0.045, -0.1749, 0.01, 0.0472, -0.1749, 0.01, 0.0487, -0.1727, 0.01, 0.0472, -0.1727, 0.01, 0.0487, -0.1706, 0.0061, 0.0085, -0.1706, 0.0083, 0.0085, -0.1684, 0.0061, 0.0085, -0.1684, 0.0083, 0.0085, -0.1706, 0.01, 0.0085, -0.1706, 0.01, 0.0103, -0.1684, 0.01, 0.0085, -0.1684, 0.01, 0.0103, -0.1706, 0.01, 0.0125, -0.1706, 0.01, 0.0146, -0.1684, 0.01, 0.0125, -0.1684, 0.01, 0.0146, -0.1662, 0.0061, 0.0085, -0.1662, 0.0083, 0.0085, -0.1641, 0.0061, 0.0085, -0.1641, 0.0083, 0.0085, -0.1662, 0.01, 0.0085, -0.1662, 0.01, 0.0103, -0.1641, 0.01, 0.0085, -0.1641, 0.01, 0.0103, -0.1662, 0.01, 0.0125, -0.1662, 0.01, 0.0146, -0.1641, 0.01, 0.0125, -0.1641, 0.01, 0.0146, -0.1627, 0.0061, 0.0085,
-0.1627, 0.0061, 0.0103, -0.1627, 0.0083, 0.0085, -0.1627, 0.0083, 0.0103, -0.1627, 0.0061, 0.0125, -0.1627, 0.0061, 0.0146, -0.1627, 0.0083, 0.0125, -0.1627, 0.0083, 0.0146, -0.1627, 0.01, 0.0085, -0.1627, 0.01, 0.0103, -0.1627, 0.01, 0.0125, -0.1627, 0.01, 0.0146, -0.1706, 0.0061, 0.0203, -0.1706, 0.0083, 0.0203, -0.1684, 0.0061, 0.0203, -0.1684, 0.0083, 0.0203, -0.1706, 0.01, 0.0168, -0.1706, 0.01, 0.019, -0.1684, 0.01, 0.0168, -0.1684, 0.01, 0.019, -0.1706, 0.01, 0.0203, -0.1684, 0.01, 0.0203, -0.1662, 0.0061, 0.0203, -0.1662, 0.0083, 0.0203, -0.1641, 0.0061, 0.0203, -0.1641, 0.0083, 0.0203, -0.1662, 0.01, 0.0168, -0.1662, 0.01, 0.019, -0.1641, 0.01, 0.0168, -0.1641, 0.01, 0.019, -0.1662, 0.01, 0.0203,
-0.1641, 0.01, 0.0203, -0.1627, 0.0061, 0.0168, -0.1627, 0.0061, 0.019, -0.1627, 0.0083, 0.0168, -0.1627, 0.0083, 0.019, -0.1627, 0.0061, 0.0203, -0.1627, 0.0083, 0.0203, -0.1627, 0.01, 0.0168, -0.1627, 0.01, 0.019, -0.1627, 0.01, 0.0203, -0.151, 0.0061, -0.0005, -0.151, 0.0061, 0.0016, -0.151, 0.0083, -0.0005, -0.151, 0.0083, 0.0016, -0.151, 0.0061, 0.0038, -0.151, 0.0061, 0.006, -0.151, 0.0083, 0.0038, -0.151, 0.0083, 0.006, -0.151, 0.01, -0.0005, -0.151, 0.01, 0.0016, -0.151, 0.01, 0.0038, -0.151, 0.01, 0.006, -0.1489, 0.01, -0.0005, -0.1489, 0.01, 0.0016, -0.1467, 0.01, -0.0005, -0.1467, 0.01, 0.0016, -0.1489, 0.01, 0.0038, -0.1489, 0.01, 0.006, -0.1467, 0.01, 0.0038, -0.1467, 0.01, 0.006,
-0.151, 0.0061, 0.0081, -0.151, 0.0061, 0.0103, -0.151, 0.0083, 0.0081, -0.151, 0.0083, 0.0103, -0.151, 0.0061, 0.0125, -0.151, 0.0061, 0.0146, -0.151, 0.0083, 0.0125, -0.151, 0.0083, 0.0146, -0.151, 0.01, 0.0081, -0.151, 0.01, 0.0103, -0.151, 0.01, 0.0125, -0.151, 0.01, 0.0146, -0.1489, 0.01, 0.0081, -0.1489, 0.01, 0.0103, -0.1467, 0.01, 0.0081, -0.1467, 0.01, 0.0103, -0.1489, 0.01, 0.0125, -0.1489, 0.01, 0.0146, -0.1467, 0.01, 0.0125, -0.1467, 0.01, 0.0146, -0.1445, 0.01, -0.0005, -0.1445, 0.01, 0.0016, -0.1424, 0.01, -0.0005, -0.1424, 0.01, 0.0016, -0.1445, 0.01, 0.0038, -0.1445, 0.01, 0.006, -0.1424, 0.01, 0.0038, -0.1424, 0.01, 0.006, -0.1392, 0.0061, -0.0005, -0.1392, 0.0061, 0.0016,
-0.1392, 0.0083, -0.0005, -0.1392, 0.0083, 0.0016, -0.1392, 0.0061, 0.0038, -0.1392, 0.0061, 0.006, -0.1392, 0.0083, 0.0038, -0.1392, 0.0083, 0.006, -0.1392, 0.01, -0.0005, -0.1392, 0.01, 0.0016, -0.1392, 0.01, 0.0038, -0.1392, 0.01, 0.006, -0.1445, 0.01, 0.0081, -0.1445, 0.01, 0.0103, -0.1424, 0.01, 0.0081, -0.1424, 0.01, 0.0103, -0.1445, 0.01, 0.0125, -0.1445, 0.01, 0.0146, -0.1424, 0.01, 0.0125, -0.1424, 0.01, 0.0146, -0.1392, 0.0061, 0.0081, -0.1392, 0.0061, 0.0103, -0.1392, 0.0083, 0.0081, -0.1392, 0.0083, 0.0103, -0.1392, 0.0061, 0.0125, -0.1392, 0.0061, 0.0146, -0.1392, 0.0083, 0.0125, -0.1392, 0.0083, 0.0146, -0.1392, 0.01, 0.0081, -0.1392, 0.01, 0.0103, -0.1392, 0.01, 0.0125, -0.1392, 0.01, 0.0146,
-0.151, 0.0061, 0.0168, -0.151, 0.0061, 0.019, -0.151, 0.0083, 0.0168, -0.151, 0.0083, 0.019, -0.151, 0.0061, 0.0212, -0.151, 0.0061, 0.0233, -0.151, 0.0083, 0.0212, -0.151, 0.0083, 0.0233, -0.151, 0.01, 0.0168, -0.151, 0.01, 0.019, -0.151, 0.01, 0.0212, -0.151, 0.01, 0.0233, -0.1489, 0.01, 0.0168, -0.1489, 0.01, 0.019, -0.1467, 0.01, 0.0168, -0.1467, 0.01, 0.019, -0.1489, 0.01, 0.0212, -0.1489, 0.01, 0.0233, -0.1467, 0.01, 0.0212, -0.1467, 0.01, 0.0233, -0.151, 0.0061, 0.0255, -0.151, 0.0061, 0.0277, -0.151, 0.0083, 0.0255, -0.151, 0.0083, 0.0277, -0.151, 0.0061, 0.0298, -0.151, 0.0061, 0.032, -0.151, 0.0083, 0.0298, -0.151, 0.0083, 0.032, -0.151, 0.01, 0.0255, -0.151, 0.01, 0.0277,
-0.151, 0.01, 0.0298, -0.151, 0.01, 0.032, -0.1489, 0.01, 0.0255, -0.1489, 0.01, 0.0277, -0.1467, 0.01, 0.0255, -0.1467, 0.01, 0.0277, -0.1489, 0.01, 0.0298, -0.1489, 0.01, 0.032, -0.1467, 0.01, 0.0298, -0.1467, 0.01, 0.032, -0.1445, 0.01, 0.0168, -0.1445, 0.01, 0.019, -0.1424, 0.01, 0.0168, -0.1424, 0.01, 0.019, -0.1445, 0.01, 0.0212, -0.1445, 0.01, 0.0233, -0.1424, 0.01, 0.0212, -0.1424, 0.01, 0.0233, -0.1392, 0.0061, 0.0168, -0.1392, 0.0061, 0.019, -0.1392, 0.0083, 0.0168, -0.1392, 0.0083, 0.019, -0.1392, 0.0061, 0.0212, -0.1392, 0.0061, 0.0233, -0.1392, 0.0083, 0.0212, -0.1392, 0.0083, 0.0233, -0.1392, 0.01, 0.0168, -0.1392, 0.01, 0.019, -0.1392, 0.01, 0.0212, -0.1392, 0.01, 0.0233,
-0.1445, 0.01, 0.0255, -0.1445, 0.01, 0.0277, -0.1424, 0.01, 0.0255, -0.1424, 0.01, 0.0277, -0.1445, 0.01, 0.0298, -0.1445, 0.01, 0.032, -0.1424, 0.01, 0.0298, -0.1424, 0.01, 0.032, -0.1392, 0.0061, 0.0255, -0.1392, 0.0061, 0.0277, -0.1392, 0.0083, 0.0255, -0.1392, 0.0083, 0.0277, -0.1392, 0.0061, 0.0298, -0.1392, 0.0061, 0.032, -0.1392, 0.0083, 0.0298, -0.1392, 0.0083, 0.032, -0.1392, 0.01, 0.0255, -0.1392, 0.01, 0.0277, -0.1392, 0.01, 0.0298, -0.1392, 0.01, 0.032, -0.1706, 0.0061, 0.0379, -0.1706, 0.0083, 0.0379, -0.1684, 0.0061, 0.0379, -0.1684, 0.0083, 0.0379, -0.1706, 0.01, 0.0379, -0.1706, 0.01, 0.0407, -0.1684, 0.01, 0.0379, -0.1684, 0.01, 0.0407, -0.1662, 0.0061, 0.0379, -0.1662, 0.0083, 0.0379,
-0.1641, 0.0061, 0.0379, -0.1641, 0.0083, 0.0379, -0.1662, 0.01, 0.0379, -0.1662, 0.01, 0.0407, -0.1641, 0.01, 0.0379, -0.1641, 0.01, 0.0407, -0.1706, 0.0061, 0.0487, -0.1706, 0.0083, 0.0487, -0.1684, 0.0061, 0.0487, -0.1684, 0.0083, 0.0487, -0.1706, 0.01, 0.0429, -0.1706, 0.01, 0.045, -0.1684, 0.01, 0.0429, -0.1684, 0.01, 0.045, -0.1706, 0.01, 0.0472, -0.1706, 0.01, 0.0487, -0.1684, 0.01, 0.0472, -0.1684, 0.01, 0.0487, -0.1662, 0.0061, 0.0487, -0.1662, 0.0083, 0.0487, -0.1641, 0.0061, 0.0487, -0.1641, 0.0083, 0.0487, -0.1662, 0.01, 0.0429, -0.1662, 0.01, 0.045, -0.1641, 0.01, 0.0429, -0.1641, 0.01, 0.045, -0.1662, 0.01, 0.0472, -0.1662, 0.01, 0.0487, -0.1641, 0.01, 0.0472, -0.1641, 0.01, 0.0487,
-0.1619, 0.0061, 0.0379, -0.1619, 0.0083, 0.0379, -0.1588, 0.0061, 0.0379, -0.1588, 0.0061, 0.0407, -0.1588, 0.0083, 0.0379, -0.1588, 0.0083, 0.0407, -0.1619, 0.01, 0.0379, -0.1619, 0.01, 0.0407, -0.1588, 0.01, 0.0379, -0.1588, 0.01, 0.0407, -0.1588, 0.0061, 0.0429, -0.1588, 0.0061, 0.045, -0.1588, 0.0083, 0.0429, -0.1588, 0.0083, 0.045, -0.1619, 0.0061, 0.0487, -0.1619, 0.0083, 0.0487, -0.1588, 0.0061, 0.0472, -0.1588, 0.0061, 0.0487, -0.1588, 0.0083, 0.0472, -0.1588, 0.0083, 0.0487, -0.1619, 0.01, 0.0429, -0.1619, 0.01, 0.045, -0.1588, 0.01, 0.0429, -0.1588, 0.01, 0.045, -0.1619, 0.01, 0.0472, -0.1619, 0.01, 0.0487, -0.1588, 0.01, 0.0472, -0.1588, 0.01, 0.0487, -0.151, 0.0061, 0.0342, -0.151, 0.0061, 0.0363,
-0.151, 0.0083, 0.0342, -0.151, 0.0083, 0.0363, -0.151, 0.0061, 0.0385, -0.151, 0.0061, 0.0407, -0.151, 0.0083, 0.0385, -0.151, 0.0083, 0.0407, -0.151, 0.01, 0.0342, -0.151, 0.01, 0.0363, -0.151, 0.01, 0.0385, -0.151, 0.01, 0.0407, -0.1489, 0.01, 0.0342, -0.1489, 0.01, 0.0363, -0.1467, 0.01, 0.0342, -0.1467, 0.01, 0.0363, -0.1489, 0.01, 0.0385, -0.1489, 0.01, 0.0407, -0.1467, 0.01, 0.0385, -0.1467, 0.01, 0.0407, -0.151, 0.0061, 0.0429, -0.151, 0.0061, 0.045, -0.151, 0.0083, 0.0429, -0.151, 0.0083, 0.045, -0.151, 0.0061, 0.0472, -0.151, 0.0061, 0.0487, -0.151, 0.0083, 0.0472, -0.151, 0.0083, 0.0487, -0.151, 0.01, 0.0429, -0.151, 0.01, 0.045, -0.151, 0.01, 0.0472, -0.151, 0.01, 0.0487,
-0.1489, 0.0061, 0.0487, -0.1489, 0.0083, 0.0487, -0.1467, 0.0061, 0.0487, -0.1467, 0.0083, 0.0487, -0.1489, 0.01, 0.0429, -0.1489, 0.01, 0.045, -0.1467, 0.01, 0.0429, -0.1467, 0.01, 0.045, -0.1489, 0.01, 0.0472, -0.1489, 0.01, 0.0487, -0.1467, 0.01, 0.0472, -0.1467, 0.01, 0.0487, -0.1445, 0.01, 0.0342, -0.1445, 0.01, 0.0363, -0.1424, 0.01, 0.0342, -0.1424, 0.01, 0.0363, -0.1445, 0.01, 0.0385, -0.1445, 0.01, 0.0407, -0.1424, 0.01, 0.0385, -0.1424, 0.01, 0.0407, -0.1392, 0.0061, 0.0342, -0.1392, 0.0061, 0.0363, -0.1392, 0.0083, 0.0342, -0.1392, 0.0083, 0.0363, -0.1392, 0.0061, 0.0385, -0.1392, 0.0061, 0.0407, -0.1392, 0.0083, 0.0385, -0.1392, 0.0083, 0.0407, -0.1392, 0.01, 0.0342, -0.1392, 0.01, 0.0363,
-0.1392, 0.01, 0.0385, -0.1392, 0.01, 0.0407, -0.1445, 0.0061, 0.0487, -0.1445, 0.0083, 0.0487, -0.1424, 0.0061, 0.0487, -0.1424, 0.0083, 0.0487, -0.1445, 0.01, 0.0429, -0.1445, 0.01, 0.045, -0.1424, 0.01, 0.0429, -0.1424, 0.01, 0.045, -0.1445, 0.01, 0.0472, -0.1445, 0.01, 0.0487, -0.1424, 0.01, 0.0472, -0.1424, 0.01, 0.0487, -0.1392, 0.0061, 0.0429, -0.1392, 0.0061, 0.045, -0.1392, 0.0083, 0.0429, -0.1392, 0.0083, 0.045, -0.1392, 0.0061, 0.0472, -0.1392, 0.0061, 0.0487, -0.1392, 0.0083, 0.0472, -0.1392, 0.0083, 0.0487, -0.1392, 0.01, 0.0429, -0.1392, 0.01, 0.045, -0.1392, 0.01, 0.0472, -0.1392, 0.01, 0.0487, -0.0775, 0.0061, -0.0005, -0.0775, 0.0061, 0.0016, -0.0775, 0.0083, -0.0005, -0.0775, 0.0083, 0.0016,
-0.0775, 0.0061, 0.0038, -0.0775, 0.0061, 0.006, -0.0775, 0.0083, 0.0038, -0.0775, 0.0083, 0.006, -0.0775, 0.01, -0.0005, -0.0775, 0.01, 0.0016, -0.0775, 0.01, 0.0038, -0.0775, 0.01, 0.006, -0.0775, 0.0061, 0.0081, -0.0775, 0.0061, 0.0103, -0.0775, 0.0083, 0.0081, -0.0775, 0.0083, 0.0103, -0.0775, 0.0061, 0.0125, -0.0775, 0.0061, 0.0146, -0.0775, 0.0083, 0.0125, -0.0775, 0.0083, 0.0146, -0.0775, 0.01, 0.0081, -0.0775, 0.01, 0.0103, -0.0775, 0.01, 0.0125, -0.0775, 0.01, 0.0146, -0.0751, 0.01, -0.0005, -0.0751, 0.01, 0.0016, -0.0729, 0.01, -0.0005, -0.0729, 0.01, 0.0016, -0.0751, 0.01, 0.0038, -0.0751, 0.01, 0.006, -0.0729, 0.01, 0.0038, -0.0729, 0.01, 0.006, -0.0707, 0.01, -0.0005, -0.0707, 0.01, 0.0016,
-0.0686, 0.01, -0.0005, -0.0686, 0.01, 0.0016, -0.0707, 0.01, 0.0038, -0.0707, 0.01, 0.006, -0.0686, 0.01, 0.0038, -0.0686, 0.01, 0.006, -0.0751, 0.01, 0.0081, -0.0751, 0.01, 0.0103, -0.0729, 0.01, 0.0081, -0.0729, 0.01, 0.0103, -0.0751, 0.01, 0.0125, -0.0751, 0.01, 0.0146, -0.0729, 0.01, 0.0125, -0.0729, 0.01, 0.0146, -0.0707, 0.01, 0.0081, -0.0707, 0.01, 0.0103, -0.0686, 0.01, 0.0081, -0.0686, 0.01, 0.0103, -0.0707, 0.01, 0.0125, -0.0707, 0.01, 0.0146, -0.0686, 0.01, 0.0125, -0.0686, 0.01, 0.0146, -0.0775, 0.0061, 0.0168, -0.0775, 0.0061, 0.019, -0.0775, 0.0083, 0.0168, -0.0775, 0.0083, 0.019, -0.0775, 0.0061, 0.0212, -0.0775, 0.0061, 0.0233, -0.0775, 0.0083, 0.0212, -0.0775, 0.0083, 0.0233,
-0.0775, 0.01, 0.0168, -0.0775, 0.01, 0.019, -0.0775, 0.01, 0.0212, -0.0775, 0.01, 0.0233, -0.0775, 0.0061, 0.0255, -0.0775, 0.0061, 0.0277, -0.0775, 0.0083, 0.0255, -0.0775, 0.0083, 0.0277, -0.0775, 0.0061, 0.0298, -0.0775, 0.0061, 0.032, -0.0775, 0.0083, 0.0298, -0.0775, 0.0083, 0.032, -0.0775, 0.01, 0.0255, -0.0775, 0.01, 0.0277, -0.0775, 0.01, 0.0298, -0.0775, 0.01, 0.032, -0.0751, 0.01, 0.0168, -0.0751, 0.01, 0.019, -0.0729, 0.01, 0.0168, -0.0729, 0.01, 0.019, -0.0751, 0.01, 0.0212, -0.0751, 0.01, 0.0233, -0.0729, 0.01, 0.0212, -0.0729, 0.01, 0.0233, -0.0707, 0.01, 0.0168, -0.0707, 0.01, 0.019, -0.0686, 0.01, 0.0168, -0.0686, 0.01, 0.019, -0.0707, 0.01, 0.0212, -0.0707, 0.01, 0.0233,
-0.0686, 0.01, 0.0212, -0.0686, 0.01, 0.0233, -0.0751, 0.01, 0.0255, -0.0751, 0.01, 0.0277, -0.0729, 0.01, 0.0255, -0.0729, 0.01, 0.0277, -0.0751, 0.01, 0.0298, -0.0751, 0.01, 0.032, -0.0729, 0.01, 0.0298, -0.0729, 0.01, 0.032, -0.0707, 0.01, 0.0255, -0.0707, 0.01, 0.0277, -0.0686, 0.01, 0.0255, -0.0686, 0.01, 0.0277, -0.0707, 0.01, 0.0298, -0.0707, 0.01, 0.032, -0.0686, 0.01, 0.0298, -0.0686, 0.01, 0.032, -0.0775, 0.0061, 0.0342, -0.0775, 0.0061, 0.0363, -0.0775, 0.0083, 0.0342, -0.0775, 0.0083, 0.0363, -0.0775, 0.0061, 0.0385, -0.0775, 0.0061, 0.0407, -0.0775, 0.0083, 0.0385, -0.0775, 0.0083, 0.0407, -0.0775, 0.01, 0.0342, -0.0775, 0.01, 0.0363, -0.0775, 0.01, 0.0385, -0.0775, 0.01, 0.0407,
-0.0775, 0.0061, 0.0429, -0.0775, 0.0061, 0.045, -0.0775, 0.0083, 0.0429, -0.0775, 0.0083, 0.045, -0.0775, 0.0061, 0.0472, -0.0775, 0.0061, 0.0487, -0.0775, 0.0083, 0.0472, -0.0775, 0.0083, 0.0487, -0.0775, 0.01, 0.0429, -0.0775, 0.01, 0.045, -0.0775, 0.01, 0.0472, -0.0775, 0.01, 0.0487, -0.0751, 0.01, 0.0342, -0.0751, 0.01, 0.0363, -0.0729, 0.01, 0.0342, -0.0729, 0.01, 0.0363, -0.0751, 0.01, 0.0385, -0.0751, 0.01, 0.0407, -0.0729, 0.01, 0.0385, -0.0729, 0.01, 0.0407, -0.0707, 0.01, 0.0342, -0.0707, 0.01, 0.0363, -0.0686, 0.01, 0.0342, -0.0686, 0.01, 0.0363, -0.0707, 0.01, 0.0385, -0.0707, 0.01, 0.0407, -0.0686, 0.01, 0.0385, -0.0686, 0.01, 0.0407, -0.0751, 0.0061, 0.0487, -0.0751, 0.0083, 0.0487,
-0.0729, 0.0061, 0.0487, -0.0729, 0.0083, 0.0487, -0.0751, 0.01, 0.0429, -0.0751, 0.01, 0.045, -0.0729, 0.01, 0.0429, -0.0729, 0.01, 0.045, -0.0751, 0.01, 0.0472, -0.0751, 0.01, 0.0487, -0.0729, 0.01, 0.0472, -0.0729, 0.01, 0.0487, -0.0707, 0.0061, 0.0487, -0.0707, 0.0083, 0.0487, -0.0686, 0.0061, 0.0487, -0.0686, 0.0083, 0.0487, -0.0707, 0.01, 0.0429, -0.0707, 0.01, 0.045, -0.0686, 0.01, 0.0429, -0.0686, 0.01, 0.045, -0.0707, 0.01, 0.0472, -0.0707, 0.01, 0.0487, -0.0686, 0.01, 0.0472, -0.0686, 0.01, 0.0487, -0.0667, 0.0061, -0.0005, -0.0667, 0.0061, 0.0016, -0.0667, 0.0083, -0.0005, -0.0667, 0.0083, 0.0016, -0.0667, 0.0061, 0.0038, -0.0667, 0.0061, 0.006, -0.0667, 0.0083, 0.0038, -0.0667, 0.0083, 0.006,
-0.0667, 0.01, -0.0005, -0.0667, 0.01, 0.0016, -0.0667, 0.01, 0.0038, -0.0667, 0.01, 0.006, -0.0667, 0.0061, 0.0081, -0.0667, 0.0061, 0.0103, -0.0667, 0.0083, 0.0081, -0.0667, 0.0083, 0.0103, -0.0667, 0.0061, 0.0125, -0.0667, 0.0061, 0.0146, -0.0667, 0.0083, 0.0125, -0.0667, 0.0083, 0.0146, -0.0667, 0.01, 0.0081, -0.0667, 0.01, 0.0103, -0.0667, 0.01, 0.0125, -0.0667, 0.01, 0.0146, -0.0667, 0.0061, 0.0168, -0.0667, 0.0061, 0.019, -0.0667, 0.0083, 0.0168, -0.0667, 0.0083, 0.019, -0.0667, 0.0061, 0.0212, -0.0667, 0.0061, 0.0233, -0.0667, 0.0083, 0.0212, -0.0667, 0.0083, 0.0233, -0.0667, 0.01, 0.0168, -0.0667, 0.01, 0.019, -0.0667, 0.01, 0.0212, -0.0667, 0.01, 0.0233, -0.0667, 0.0061, 0.0255, -0.0667, 0.0061, 0.0277,
-0.0667, 0.0083, 0.0255, -0.0667, 0.0083, 0.0277, -0.0667, 0.0061, 0.0298, -0.0667, 0.0061, 0.032, -0.0667, 0.0083, 0.0298, -0.0667, 0.0083, 0.032, -0.0667, 0.01, 0.0255, -0.0667, 0.01, 0.0277, -0.0667, 0.01, 0.0298, -0.0667, 0.01, 0.032, -0.05, 0.0061, -0.0005, -0.05, 0.0061, 0.0016, -0.05, 0.0083, -0.0005, -0.05, 0.0083, 0.0016, -0.05, 0.0061, 0.0038, -0.05, 0.0061, 0.006, -0.05, 0.0083, 0.0038, -0.05, 0.0083, 0.006, -0.05, 0.01, -0.0005, -0.05, 0.01, 0.0016, -0.0469, 0.01, -0.0005, -0.0469, 0.01, 0.0016, -0.05, 0.01, 0.0038, -0.05, 0.01, 0.006, -0.0469, 0.01, 0.0038, -0.0469, 0.01, 0.006, -0.0447, 0.01, -0.0005, -0.0447, 0.01, 0.0016, -0.0425, 0.01, -0.0005, -0.0425, 0.01, 0.0016,
-0.0447, 0.01, 0.0038, -0.0447, 0.01, 0.006, -0.0425, 0.01, 0.0038, -0.0425, 0.01, 0.006, -0.05, 0.0061, 0.0081, -0.05, 0.0061, 0.0103, -0.05, 0.0083, 0.0081, -0.05, 0.0083, 0.0103, -0.05, 0.0061, 0.0125, -0.05, 0.0061, 0.0146, -0.05, 0.0083, 0.0125, -0.05, 0.0083, 0.0146, -0.05, 0.01, 0.0081, -0.05, 0.01, 0.0103, -0.0469, 0.01, 0.0081, -0.0469, 0.01, 0.0103, -0.05, 0.01, 0.0125, -0.05, 0.01, 0.0146, -0.0469, 0.01, 0.0125, -0.0469, 0.01, 0.0146, -0.0447, 0.01, 0.0081, -0.0447, 0.01, 0.0103, -0.0425, 0.01, 0.0081, -0.0425, 0.01, 0.0103, -0.0447, 0.01, 0.0125, -0.0447, 0.01, 0.0146, -0.0425, 0.01, 0.0125, -0.0425, 0.01, 0.0146, -0.0382, 0.0061, -0.0013, -0.0382, 0.0083, -0.0013,
-0.0404, 0.01, -0.0005, -0.0404, 0.01, 0.0016, -0.0382, 0.01, -0.0013, -0.0382, 0.01, 0.0016, -0.0404, 0.01, 0.0038, -0.0404, 0.01, 0.006, -0.0382, 0.01, 0.0038, -0.0382, 0.01, 0.006, -0.036, 0.0061, -0.0013, -0.036, 0.0083, -0.0013, -0.0338, 0.0061, -0.0013, -0.0338, 0.0083, -0.0013, -0.036, 0.01, -0.0013, -0.036, 0.01, 0.0016, -0.0338, 0.01, -0.0013, -0.0338, 0.01, 0.0016, -0.036, 0.01, 0.0038, -0.036, 0.01, 0.006, -0.0338, 0.01, 0.0038, -0.0338, 0.01, 0.006, -0.0382, 0.0061, 0.0105, -0.0382, 0.0083, 0.0105, -0.0382, 0.0061, 0.0125, -0.0382, 0.0061, 0.0146, -0.0382, 0.0083, 0.0125, -0.0382, 0.0083, 0.0146, -0.0404, 0.01, 0.0081, -0.0404, 0.01, 0.0103, -0.0382, 0.01, 0.0081, -0.0382, 0.01, 0.0105,
-0.0404, 0.01, 0.0125, -0.0404, 0.01, 0.0146, -0.0382, 0.01, 0.0125, -0.0382, 0.01, 0.0146, -0.036, 0.0061, 0.0105, -0.036, 0.0083, 0.0105, -0.0338, 0.0061, 0.0105, -0.0338, 0.0083, 0.0105, -0.036, 0.01, 0.0081, -0.036, 0.01, 0.0105, -0.0338, 0.01, 0.0081, -0.0338, 0.01, 0.0105, -0.05, 0.0061, 0.0168, -0.05, 0.0061, 0.019, -0.05, 0.0083, 0.0168, -0.05, 0.0083, 0.019, -0.05, 0.0061, 0.0212, -0.05, 0.0061, 0.0233, -0.05, 0.0083, 0.0212, -0.05, 0.0083, 0.0233, -0.05, 0.01, 0.0168, -0.05, 0.01, 0.019, -0.0469, 0.01, 0.0168, -0.0469, 0.01, 0.019, -0.05, 0.01, 0.0212, -0.05, 0.01, 0.0233, -0.0469, 0.01, 0.0212, -0.0469, 0.01, 0.0233, -0.0447, 0.01, 0.0168, -0.0447, 0.01, 0.019,
-0.0425, 0.01, 0.0168, -0.0425, 0.01, 0.019, -0.0447, 0.01, 0.0212, -0.0447, 0.01, 0.0233, -0.0425, 0.01, 0.0212, -0.0425, 0.01, 0.0233, -0.05, 0.0061, 0.0255, -0.05, 0.0061, 0.0277, -0.05, 0.0083, 0.0255, -0.05, 0.0083, 0.0277, -0.05, 0.0061, 0.0298, -0.05, 0.0061, 0.032, -0.05, 0.0083, 0.0298, -0.05, 0.0083, 0.032, -0.05, 0.01, 0.0255, -0.05, 0.01, 0.0277, -0.0469, 0.01, 0.0255, -0.0469, 0.01, 0.0277, -0.05, 0.01, 0.0298, -0.05, 0.01, 0.032, -0.0469, 0.01, 0.0298, -0.0469, 0.01, 0.032, -0.0447, 0.01, 0.0255, -0.0447, 0.01, 0.0277, -0.0425, 0.01, 0.0255, -0.0425, 0.01, 0.0277, -0.0447, 0.01, 0.0298, -0.0447, 0.01, 0.032, -0.0425, 0.01, 0.0298, -0.0425, 0.01, 0.032,
-0.0382, 0.0061, 0.0168, -0.0382, 0.0061, 0.019, -0.0382, 0.0083, 0.0168, -0.0382, 0.0083, 0.019, -0.0382, 0.0061, 0.0212, -0.0382, 0.0061, 0.0233, -0.0382, 0.0083, 0.0212, -0.0382, 0.0083, 0.0233, -0.0404, 0.01, 0.0168, -0.0404, 0.01, 0.019, -0.0382, 0.01, 0.0168, -0.0382, 0.01, 0.019, -0.0404, 0.01, 0.0212, -0.0404, 0.01, 0.0233, -0.0382, 0.01, 0.0212, -0.0382, 0.01, 0.0233, -0.0382, 0.0061, 0.0255, -0.0382, 0.0061, 0.0277, -0.0382, 0.0083, 0.0255, -0.0382, 0.0083, 0.0277, -0.0382, 0.0061, 0.0298, -0.0382, 0.0061, 0.032, -0.0382, 0.0083, 0.0298, -0.0382, 0.0083, 0.032, -0.0404, 0.01, 0.0255, -0.0404, 0.01, 0.0277, -0.0382, 0.01, 0.0255, -0.0382, 0.01, 0.0277, -0.0404, 0.01, 0.0298, -0.0404, 0.01, 0.032,
-0.0382, 0.01, 0.0298, -0.0382, 0.01, 0.032, -0.0667, 0.0061, 0.0342, -0.0667, 0.0061, 0.0363, -0.0667, 0.0083, 0.0342, -0.0667, 0.0083, 0.0363, -0.0667, 0.0061, 0.0385, -0.0667, 0.0061, 0.0407, -0.0667, 0.0083, 0.0385, -0.0667, 0.0083, 0.0407, -0.0667, 0.01, 0.0342, -0.0667, 0.01, 0.0363, -0.0667, 0.01, 0.0385, -0.0667, 0.01, 0.0407, -0.0667, 0.0061, 0.0429, -0.0667, 0.0061, 0.045, -0.0667, 0.0083, 0.0429, -0.0667, 0.0083, 0.045, -0.0667, 0.0061, 0.0472, -0.0667, 0.0061, 0.0487, -0.0667, 0.0083, 0.0472, -0.0667, 0.0083, 0.0487, -0.0667, 0.01, 0.0429, -0.0667, 0.01, 0.045, -0.0667, 0.01, 0.0472, -0.0667, 0.01, 0.0487, -0.05, 0.0061, 0.0342, -0.05, 0.0061, 0.0363, -0.05, 0.0083, 0.0342, -0.05, 0.0083, 0.0363,
-0.05, 0.0061, 0.0385, -0.05, 0.0061, 0.0407, -0.05, 0.0083, 0.0385, -0.05, 0.0083, 0.0407, -0.05, 0.01, 0.0342, -0.05, 0.01, 0.0363, -0.0469, 0.01, 0.0342, -0.0469, 0.01, 0.0363, -0.05, 0.01, 0.0385, -0.05, 0.01, 0.0407, -0.0469, 0.01, 0.0385, -0.0469, 0.01, 0.0407, -0.0447, 0.01, 0.0342, -0.0447, 0.01, 0.0363, -0.0425, 0.01, 0.0342, -0.0425, 0.01, 0.0363, -0.0447, 0.01, 0.0385, -0.0447, 0.01, 0.0407, -0.0425, 0.01, 0.0385, -0.0425, 0.01, 0.0407, -0.05, 0.0061, 0.0429, -0.05, 0.0061, 0.045, -0.05, 0.0083, 0.0429, -0.05, 0.0083, 0.045, -0.05, 0.0061, 0.0472, -0.05, 0.0061, 0.0487, -0.05, 0.0083, 0.0472, -0.05, 0.0083, 0.0487, -0.0469, 0.0061, 0.0487, -0.0469, 0.0083, 0.0487,
-0.05, 0.01, 0.0429, -0.05, 0.01, 0.045, -0.0469, 0.01, 0.0429, -0.0469, 0.01, 0.045, -0.05, 0.01, 0.0472, -0.05, 0.01, 0.0487, -0.0469, 0.01, 0.0472, -0.0469, 0.01, 0.0487, -0.0447, 0.0061, 0.0487, -0.0447, 0.0083, 0.0487, -0.0425, 0.0061, 0.0487, -0.0425, 0.0083, 0.0487, -0.0447, 0.01, 0.0429, -0.0447, 0.01, 0.045, -0.0425, 0.01, 0.0429, -0.0425, 0.01, 0.045, -0.0447, 0.01, 0.0472, -0.0447, 0.01, 0.0487, -0.0425, 0.01, 0.0472, -0.0425, 0.01, 0.0487, -0.0382, 0.0061, 0.0342, -0.0382, 0.0061, 0.0363, -0.0382, 0.0083, 0.0342, -0.0382, 0.0083, 0.0363, -0.0382, 0.0061, 0.0379, -0.0382, 0.0083, 0.0379, -0.0404, 0.01, 0.0342, -0.0404, 0.01, 0.0363, -0.0382, 0.01, 0.0342, -0.0382, 0.01, 0.0363,
-0.0404, 0.01, 0.0385, -0.0404, 0.01, 0.0407, -0.0382, 0.01, 0.0379, -0.0382, 0.01, 0.0407, -0.036, 0.0061, 0.0379, -0.036, 0.0083, 0.0379, -0.0338, 0.0061, 0.0379, -0.0338, 0.0083, 0.0379, -0.036, 0.01, 0.0379, -0.036, 0.01, 0.0407, -0.0338, 0.01, 0.0379, -0.0338, 0.01, 0.0407, -0.0404, 0.0061, 0.0487, -0.0404, 0.0083, 0.0487, -0.0382, 0.0061, 0.0487, -0.0382, 0.0083, 0.0487, -0.0404, 0.01, 0.0429, -0.0404, 0.01, 0.045, -0.0382, 0.01, 0.0429, -0.0382, 0.01, 0.045, -0.0404, 0.01, 0.0472, -0.0404, 0.01, 0.0487, -0.0382, 0.01, 0.0472, -0.0382, 0.01, 0.0487, -0.036, 0.0061, 0.0487, -0.036, 0.0083, 0.0487, -0.0338, 0.0061, 0.0487, -0.0338, 0.0083, 0.0487, -0.036, 0.01, 0.0429, -0.036, 0.01, 0.045,
-0.0338, 0.01, 0.0429, -0.0338, 0.01, 0.045, -0.036, 0.01, 0.0472, -0.036, 0.01, 0.0487, -0.0338, 0.01, 0.0472, -0.0338, 0.01, 0.0487, -0.0317, 0.0061, -0.0013, -0.0317, 0.0083, -0.0013, -0.0295, 0.0061, -0.0013, -0.0295, 0.0083, -0.0013, -0.0317, 0.01, -0.0013, -0.0317, 0.01, 0.0016, -0.0295, 0.01, -0.0013, -0.0295, 0.01, 0.0016, -0.0317, 0.01, 0.0038, -0.0317, 0.01, 0.006, -0.0295, 0.01, 0.0038, -0.0295, 0.01, 0.006, -0.0273, 0.0061, -0.0013, -0.0273, 0.0083, -0.0013, -0.0252, 0.0061, -0.0013, -0.0252, 0.0083, -0.0013, -0.0273, 0.01, -0.0013, -0.0273, 0.01, 0.0016, -0.0252, 0.01, -0.0013, -0.0252, 0.01, 0.0016, -0.0273, 0.01, 0.0038, -0.0273, 0.01, 0.006, -0.0252, 0.01, 0.0038, -0.0252, 0.01, 0.006,
-0.0317, 0.0061, 0.0105, -0.0317, 0.0083, 0.0105, -0.0295, 0.0061, 0.0105, -0.0295, 0.0083, 0.0105, -0.0317, 0.01, 0.0081, -0.0317, 0.01, 0.0105, -0.0295, 0.01, 0.0081, -0.0295, 0.01, 0.0105, -0.0273, 0.0061, 0.0105, -0.0273, 0.0083, 0.0105, -0.0252, 0.0061, 0.0105, -0.0252, 0.0083, 0.0105, -0.0273, 0.01, 0.0081, -0.0273, 0.01, 0.0105, -0.0252, 0.01, 0.0081, -0.0252, 0.01, 0.0105, -0.023, 0.0061, -0.0013, -0.023, 0.0083, -0.0013, -0.0208, 0.0061, -0.0013, -0.0208, 0.0083, -0.0013, -0.023, 0.01, -0.0013, -0.023, 0.01, 0.0016, -0.0208, 0.01, -0.0013, -0.0208, 0.01, 0.0016, -0.023, 0.01, 0.0038, -0.023, 0.01, 0.006, -0.0208, 0.01, 0.0038, -0.0208, 0.01, 0.006, -0.0187, 0.0061, -0.0013, -0.0187, 0.0083, -0.0013,
-0.0165, 0.0061, -0.0013, -0.0165, 0.0083, -0.0013, -0.0187, 0.01, -0.0013, -0.0187, 0.01, 0.0016, -0.0165, 0.01, -0.0013, -0.0165, 0.01, 0.0016, -0.0187, 0.01, 0.0038, -0.0187, 0.01, 0.006, -0.0165, 0.01, 0.0038, -0.0165, 0.01, 0.006, -0.023, 0.0061, 0.0105, -0.023, 0.0083, 0.0105, -0.0208, 0.0061, 0.0105, -0.0208, 0.0083, 0.0105, -0.023, 0.01, 0.0081, -0.023, 0.01, 0.0105, -0.0208, 0.01, 0.0081, -0.0208, 0.01, 0.0105, -0.0187, 0.0061, 0.0105, -0.0187, 0.0083, 0.0105, -0.0165, 0.0061, 0.0105, -0.0165, 0.0083, 0.0105, -0.0187, 0.01, 0.0081, -0.0187, 0.01, 0.0105, -0.0165, 0.01, 0.0081, -0.0165, 0.01, 0.0105, -0.0143, 0.0061, -0.0013, -0.0143, 0.0083, -0.0013, -0.0121, 0.0061, -0.0013, -0.0121, 0.0083, -0.0013,
-0.0143, 0.01, -0.0013, -0.0143, 0.01, 0.0016, -0.0121, 0.01, -0.0013, -0.0121, 0.01, 0.0016, -0.0143, 0.01, 0.0038, -0.0143, 0.01, 0.006, -0.0121, 0.01, 0.0038, -0.0121, 0.01, 0.006, -0.01, 0.0061, -0.0013, -0.01, 0.0083, -0.0013, -0.0078, 0.0061, -0.0013, -0.0078, 0.0083, -0.0013, -0.01, 0.01, -0.0013, -0.01, 0.01, 0.0016, -0.0078, 0.01, -0.0013, -0.0078, 0.01, 0.0016, -0.01, 0.01, 0.0038, -0.01, 0.01, 0.006, -0.0078, 0.01, 0.0038, -0.0078, 0.01, 0.006, -0.0143, 0.0061, 0.0105, -0.0143, 0.0083, 0.0105, -0.0121, 0.0061, 0.0105, -0.0121, 0.0083, 0.0105, -0.0143, 0.01, 0.0081, -0.0143, 0.01, 0.0105, -0.0121, 0.01, 0.0081, -0.0121, 0.01, 0.0105, -0.01, 0.0061, 0.0105, -0.01, 0.0083, 0.0105,
-0.0078, 0.0061, 0.0105, -0.0078, 0.0083, 0.0105, -0.01, 0.01, 0.0081, -0.01, 0.01, 0.0105, -0.0078, 0.01, 0.0081, -0.0078, 0.01, 0.0105, -0.0056, 0.0061, -0.0013, -0.0056, 0.0083, -0.0013, -0.0035, 0.0061, -0.0013, -0.0035, 0.0083, -0.0013, -0.0056, 0.01, -0.0013, -0.0056, 0.01, 0.0016, -0.0035, 0.01, -0.0013, -0.0035, 0.01, 0.0016, -0.0056, 0.01, 0.0038, -0.0056, 0.01, 0.006, -0.0035, 0.01, 0.0038, -0.0035, 0.01, 0.006, -0.0013, 0.0061, -0.0013, -0.0013, 0.0083, -0.0013, 0.0009, 0.0061, -0.0013, 0.0009, 0.0083, -0.0013, -0.0013, 0.01, -0.0013, -0.0013, 0.01, 0.0016, 0.0009, 0.01, -0.0013, 0.0009, 0.01, 0.0016, -0.0013, 0.01, 0.0038, -0.0013, 0.01, 0.006, 0.0009, 0.01, 0.0038, 0.0009, 0.01, 0.006,
-0.0056, 0.0061, 0.0105, -0.0056, 0.0083, 0.0105, -0.0035, 0.0061, 0.0105, -0.0035, 0.0083, 0.0105, -0.0056, 0.01, 0.0081, -0.0056, 0.01, 0.0105, -0.0035, 0.01, 0.0081, -0.0035, 0.01, 0.0105, -0.0013, 0.0061, 0.0105, -0.0013, 0.0083, 0.0105, 0.0009, 0.0061, 0.0105, 0.0009, 0.0083, 0.0105, -0.0013, 0.01, 0.0081, -0.0013, 0.01, 0.0105, 0.0009, 0.01, 0.0081, 0.0009, 0.01, 0.0105, -0.0317, 0.0061, 0.0379, -0.0317, 0.0083, 0.0379, -0.0295, 0.0061, 0.0379, -0.0295, 0.0083, 0.0379, -0.0317, 0.01, 0.0379, -0.0317, 0.01, 0.0407, -0.0295, 0.01, 0.0379, -0.0295, 0.01, 0.0407, -0.0273, 0.0061, 0.0379, -0.0273, 0.0083, 0.0379, -0.0252, 0.0061, 0.0379, -0.0252, 0.0083, 0.0379, -0.0273, 0.01, 0.0379, -0.0273, 0.01, 0.0407,
-0.0252, 0.01, 0.0379, -0.0252, 0.01, 0.0407, -0.0317, 0.0061, 0.0487, -0.0317, 0.0083, 0.0487, -0.0295, 0.0061, 0.0487, -0.0295, 0.0083, 0.0487, -0.0317, 0.01, 0.0429, -0.0317, 0.01, 0.045, -0.0295, 0.01, 0.0429, -0.0295, 0.01, 0.045, -0.0317, 0.01, 0.0472, -0.0317, 0.01, 0.0487, -0.0295, 0.01, 0.0472, -0.0295, 0.01, 0.0487, -0.0273, 0.0061, 0.0487, -0.0273, 0.0083, 0.0487, -0.0252, 0.0061, 0.0487, -0.0252, 0.0083, 0.0487, -0.0273, 0.01, 0.0429, -0.0273, 0.01, 0.045, -0.0252, 0.01, 0.0429, -0.0252, 0.01, 0.045, -0.0273, 0.01, 0.0472, -0.0273, 0.01, 0.0487, -0.0252, 0.01, 0.0472, -0.0252, 0.01, 0.0487, -0.023, 0.0061, 0.0379, -0.023, 0.0083, 0.0379, -0.0208, 0.0061, 0.0379, -0.0208, 0.0083, 0.0379,
-0.023, 0.01, 0.0379, -0.023, 0.01, 0.0407, -0.0208, 0.01, 0.0379, -0.0208, 0.01, 0.0407, -0.0187, 0.0061, 0.0379, -0.0187, 0.0083, 0.0379, -0.0165, 0.0061, 0.0379, -0.0165, 0.0083, 0.0379, -0.0187, 0.01, 0.0379, -0.0187, 0.01, 0.0407, -0.0165, 0.01, 0.0379, -0.0165, 0.01, 0.0407, -0.023, 0.0061, 0.0487, -0.023, 0.0083, 0.0487, -0.0208, 0.0061, 0.0487, -0.0208, 0.0083, 0.0487, -0.023, 0.01, 0.0429, -0.023, 0.01, 0.045, -0.0208, 0.01, 0.0429, -0.0208, 0.01, 0.045, -0.023, 0.01, 0.0472, -0.023, 0.01, 0.0487, -0.0208, 0.01, 0.0472, -0.0208, 0.01, 0.0487, -0.0187, 0.0061, 0.0487, -0.0187, 0.0083, 0.0487, -0.0165, 0.0061, 0.0487, -0.0165, 0.0083, 0.0487, -0.0187, 0.01, 0.0429, -0.0187, 0.01, 0.045,
-0.0165, 0.01, 0.0429, -0.0165, 0.01, 0.045, -0.0187, 0.01, 0.0472, -0.0187, 0.01, 0.0487, -0.0165, 0.01, 0.0472, -0.0165, 0.01, 0.0487, -0.0143, 0.0061, 0.0379, -0.0143, 0.0083, 0.0379, -0.0121, 0.0061, 0.0379, -0.0121, 0.0083, 0.0379, -0.0143, 0.01, 0.0379, -0.0143, 0.01, 0.0407, -0.0121, 0.01, 0.0379, -0.0121, 0.01, 0.0407, -0.01, 0.0061, 0.0379, -0.01, 0.0083, 0.0379, -0.0078, 0.0061, 0.0379, -0.0078, 0.0083, 0.0379, -0.01, 0.01, 0.0379, -0.01, 0.01, 0.0407, -0.0078, 0.01, 0.0379, -0.0078, 0.01, 0.0407, -0.0143, 0.0061, 0.0487, -0.0143, 0.0083, 0.0487, -0.0121, 0.0061, 0.0487, -0.0121, 0.0083, 0.0487, -0.0143, 0.01, 0.0429, -0.0143, 0.01, 0.045, -0.0121, 0.01, 0.0429, -0.0121, 0.01, 0.045,
-0.0143, 0.01, 0.0472, -0.0143, 0.01, 0.0487, -0.0121, 0.01, 0.0472, -0.0121, 0.01, 0.0487, -0.01, 0.0061, 0.0487, -0.01, 0.0083, 0.0487, -0.0078, 0.0061, 0.0487, -0.0078, 0.0083, 0.0487, -0.01, 0.01, 0.0429, -0.01, 0.01, 0.045, -0.0078, 0.01, 0.0429, -0.0078, 0.01, 0.045, -0.01, 0.01, 0.0472, -0.01, 0.01, 0.0487, -0.0078, 0.01, 0.0472, -0.0078, 0.01, 0.0487, -0.0056, 0.0061, 0.0379, -0.0056, 0.0083, 0.0379, -0.0035, 0.0061, 0.0379, -0.0035, 0.0083, 0.0379, -0.0056, 0.01, 0.0379, -0.0056, 0.01, 0.0407, -0.0035, 0.01, 0.0379, -0.0035, 0.01, 0.0407, -0.0013, 0.0061, 0.0379, -0.0013, 0.0083, 0.0379, 0.0009, 0.0061, 0.0379, 0.0009, 0.0083, 0.0379, -0.0013, 0.01, 0.0379, -0.0013, 0.01, 0.0407,
0.0008, 0.01, 0.0379, 0.0009, 0.01, 0.0407, -0.0056, 0.0061, 0.0487, -0.0056, 0.0083, 0.0487, -0.0035, 0.0061, 0.0487, -0.0035, 0.0083, 0.0487, -0.0056, 0.01, 0.0429, -0.0056, 0.01, 0.045, -0.0035, 0.01, 0.0429, -0.0035, 0.01, 0.045, -0.0056, 0.01, 0.0472, -0.0056, 0.01, 0.0487, -0.0035, 0.01, 0.0472, -0.0035, 0.01, 0.0487, -0.0013, 0.0061, 0.0487, -0.0013, 0.0083, 0.0487, 0.0009, 0.0061, 0.0487, 0.0009, 0.0083, 0.0487, -0.0013, 0.01, 0.0429, -0.0013, 0.01, 0.045, 0.0009, 0.01, 0.0429, 0.0009, 0.01, 0.045, -0.0013, 0.01, 0.0472, -0.0013, 0.01, 0.0487, 0.0009, 0.01, 0.0472, 0.0009, 0.01, 0.0487, 0.0461, -0.0, -0.067, 0.0486, -0.0, -0.067, 0.0461, -0.0, -0.0656, 0.0461, -0.0, -0.0635,
0.0486, -0.0, -0.0656, 0.0486, -0.0, -0.0635, 0.0461, 0.0017, -0.067, 0.0461, 0.0039, -0.067, 0.0486, 0.0017, -0.067, 0.0486, 0.0039, -0.067, 0.0461, 0.0017, -0.0656, 0.0461, 0.0017, -0.0635, 0.0461, 0.0039, -0.0656, 0.0461, 0.0039, -0.0635, 0.05, -0.0, -0.067, 0.05, -0.0, -0.0656, 0.05, -0.0, -0.0635, 0.05, 0.0017, -0.067, 0.05, 0.0039, -0.067, 0.05, 0.0017, -0.0656, 0.05, 0.0017, -0.0635, 0.05, 0.0039, -0.0656, 0.05, 0.0039, -0.0635, 0.0461, -0.0, -0.0613, 0.0461, -0.0, -0.0591, 0.0486, -0.0, -0.0613, 0.0486, -0.0, -0.0591, 0.0461, -0.0, -0.057, 0.0461, -0.0, -0.0548, 0.0486, -0.0, -0.057, 0.0486, -0.0, -0.0548, 0.0461, 0.0017, -0.0613, 0.0461, 0.0017, -0.0591, 0.0461, 0.0039, -0.0613,
0.0461, 0.0039, -0.0591, 0.0461, 0.0017, -0.057, 0.0461, 0.0017, -0.0548, 0.0461, 0.0039, -0.057, 0.0461, 0.0039, -0.0548, 0.05, -0.0, -0.0613, 0.05, -0.0, -0.0591, 0.05, -0.0, -0.057, 0.05, -0.0, -0.0548, 0.0529, -0.0, -0.0562, 0.0529, -0.0, -0.0548, 0.05, 0.0017, -0.0613, 0.05, 0.0017, -0.0591, 0.05, 0.0039, -0.0613, 0.05, 0.0039, -0.0591, 0.05, 0.0017, -0.057, 0.05, 0.0017, -0.0548, 0.05, 0.0039, -0.057, 0.05, 0.0039, -0.0548, 0.0529, 0.0017, -0.0562, 0.0529, 0.0017, -0.0548, 0.0529, 0.0039, -0.0562, 0.0529, 0.0039, -0.0548, 0.0461, -0.0, -0.0526, 0.0461, -0.0, -0.0505, 0.0486, -0.0, -0.0526, 0.0486, -0.0, -0.0505, 0.0461, -0.0, -0.0483, 0.0461, -0.0, -0.0461, 0.0486, -0.0, -0.0483,
0.0486, -0.0, -0.0461, 0.0461, 0.0017, -0.0526, 0.0461, 0.0017, -0.0505, 0.0461, 0.0039, -0.0526, 0.0461, 0.0039, -0.0505, 0.0461, 0.0017, -0.0483, 0.0461, 0.0017, -0.0461, 0.0461, 0.0039, -0.0483, 0.0461, 0.0039, -0.0461, 0.05, -0.0, -0.0526, 0.05, -0.0, -0.0505, 0.0529, -0.0, -0.0523, 0.05, -0.0, -0.0483, 0.05, -0.0, -0.0454, 0.053, -0.0, -0.0454, 0.05, 0.0017, -0.0526, 0.05, 0.0017, -0.0505, 0.05, 0.0039, -0.0526, 0.05, 0.0039, -0.0505, 0.0529, 0.0017, -0.0523, 0.0529, 0.0039, -0.0523, 0.05, 0.0017, -0.0483, 0.05, 0.0017, -0.0454, 0.05, 0.0039, -0.0483, 0.05, 0.0039, -0.0454, 0.053, 0.0017, -0.0454, 0.053, 0.0039, -0.0454, 0.0461, -0.0, -0.0439, 0.0461, -0.0, -0.0425, 0.0486, -0.0, -0.0439,
0.0486, -0.0, -0.0425, 0.0461, 0.0017, -0.0439, 0.0461, 0.0017, -0.0425, 0.0461, 0.0039, -0.0439, 0.0461, 0.0039, -0.0425, 0.0486, 0.0017, -0.0425, 0.0486, 0.0039, -0.0425, 0.0508, -0.0, -0.0439, 0.0508, -0.0, -0.0425, 0.053, -0.0, -0.0439, 0.053, -0.0, -0.0425, 0.0508, 0.0017, -0.0425, 0.0508, 0.0039, -0.0425, 0.053, 0.0017, -0.0425, 0.053, 0.0039, -0.0425, 0.0551, -0.0, -0.0562, 0.0551, -0.0, -0.0548, 0.0573, -0.0, -0.0562, 0.0573, -0.0, -0.0548, 0.0551, 0.0017, -0.0562, 0.0551, 0.0039, -0.0562, 0.0573, 0.0017, -0.0562, 0.0573, 0.0039, -0.0562, 0.0595, -0.0, -0.0562, 0.0595, -0.0, -0.0548, 0.0616, -0.0, -0.0562, 0.0616, -0.0, -0.0548, 0.0595, 0.0017, -0.0562, 0.0595, 0.0039, -0.0562, 0.0616, 0.0017, -0.0562,
0.0616, 0.0039, -0.0562, 0.0638, -0.0, -0.0562, 0.0638, -0.0, -0.0548, 0.0657, -0.0, -0.0562, 0.0657, -0.0, -0.0548, 0.0638, 0.0017, -0.0562, 0.0638, 0.0039, -0.0562, 0.0657, 0.0017, -0.0562, 0.0657, 0.0017, -0.0548, 0.0657, 0.0039, -0.0562, 0.0657, 0.0039, -0.0548, 0.0551, -0.0, -0.0523, 0.0573, -0.0, -0.0523, 0.0551, -0.0, -0.0454, 0.0573, -0.0, -0.0454, 0.0551, 0.0017, -0.0523, 0.0551, 0.0039, -0.0523, 0.0573, 0.0017, -0.0523, 0.0573, 0.0039, -0.0523, 0.0551, 0.0017, -0.0454, 0.0551, 0.0039, -0.0454, 0.0573, 0.0017, -0.0454, 0.0573, 0.0039, -0.0454, 0.0595, -0.0, -0.0523, 0.0616, -0.0, -0.0523, 0.0595, -0.0, -0.0454, 0.0616, -0.0, -0.0454, 0.0595, 0.0017, -0.0523, 0.0595, 0.0039, -0.0523, 0.0616, 0.0017, -0.0523,
0.0616, 0.0039, -0.0523, 0.0595, 0.0017, -0.0454, 0.0595, 0.0039, -0.0454, 0.0616, 0.0017, -0.0454, 0.0616, 0.0039, -0.0454, 0.0551, -0.0, -0.0439, 0.0551, -0.0, -0.0425, 0.0573, -0.0, -0.0439, 0.0573, -0.0, -0.0425, 0.0551, 0.0017, -0.0425, 0.0551, 0.0039, -0.0425, 0.0573, 0.0017, -0.0425, 0.0573, 0.0039, -0.0425, 0.0595, -0.0, -0.0439, 0.0595, -0.0, -0.0425, 0.0616, -0.0, -0.0439, 0.0616, -0.0, -0.0425, 0.0595, 0.0017, -0.0425, 0.0595, 0.0039, -0.0425, 0.0616, 0.0017, -0.0425, 0.0616, 0.0039, -0.0425, 0.0638, -0.0, -0.0523, 0.0657, -0.0, -0.0523, 0.0638, -0.0, -0.0454, 0.0667, -0.0, -0.0454, 0.0638, 0.0017, -0.0523, 0.0638, 0.0039, -0.0523, 0.0657, 0.0017, -0.0523, 0.0657, 0.0039, -0.0523, 0.0638, 0.0017, -0.0454,
0.0638, 0.0039, -0.0454, 0.0667, 0.0017, -0.0454, 0.0667, 0.0039, -0.0454, 0.0638, -0.0, -0.0439, 0.0638, -0.0, -0.0425, 0.0667, -0.0, -0.0439, 0.0667, -0.0, -0.0425, 0.0638, 0.0017, -0.0425, 0.0638, 0.0039, -0.0425, 0.0667, 0.0017, -0.0439, 0.0667, 0.0017, -0.0425, 0.0667, 0.0039, -0.0439, 0.0667, 0.0039, -0.0425, 0.0887, -0.0, -0.0624, 0.0887, 0.0017, -0.0625, 0.0887, 0.0039, -0.0625, 0.088, -0.0, -0.0616, 0.087, -0.0, -0.0594, 0.0864, -0.0, -0.0565, 0.0863, -0.0, -0.0548, 0.0865, -0.0, -0.0573, 0.0877, -0.0, -0.0548, 0.0879, 0.0017, -0.0614, 0.0869, 0.0017, -0.0592, 0.0879, 0.0039, -0.0614, 0.0869, 0.0039, -0.0592, 0.0864, 0.0017, -0.0568, 0.0863, 0.0017, -0.0548, 0.0864, 0.0039, -0.0568, 0.0863, 0.0039, -0.0548,
0.0866, 0.0017, -0.0579, 0.0866, 0.0039, -0.0579, 0.0863, -0.0, -0.0526, 0.0865, -0.0, -0.0513, 0.0877, -0.0, -0.0526, 0.0867, -0.0, -0.0503, 0.0875, -0.0, -0.048, 0.0884, -0.0, -0.0468, 0.0863, 0.0017, -0.0526, 0.0865, 0.0017, -0.0511, 0.0863, 0.0039, -0.0526, 0.0865, 0.0039, -0.0511, 0.0867, 0.0017, -0.0501, 0.0867, 0.0039, -0.0501, 0.0874, 0.0017, -0.0482, 0.0884, 0.0017, -0.0468, 0.0874, 0.0039, -0.0482, 0.0884, 0.0039, -0.0468, 0.0909, -0.0, -0.0646, 0.0897, -0.0, -0.0637, 0.0918, -0.0, -0.0652, 0.092, -0.0, -0.0635, 0.0909, 0.0017, -0.0646, 0.0897, 0.0017, -0.0637, 0.0909, 0.0039, -0.0646, 0.0897, 0.0039, -0.0637, 0.092, 0.0017, -0.0653, 0.092, 0.0039, -0.0653, 0.0972, -0.0, -0.0668, 0.0939, -0.0, -0.0661,
0.0942, -0.0, -0.0635, 0.0962, -0.0, -0.0667, 0.0963, -0.0, -0.0627, 0.097, 0.0017, -0.0668, 0.097, 0.0039, -0.0668, 0.0942, 0.0017, -0.0662, 0.0942, 0.0039, -0.0662, 0.0959, 0.0017, -0.0666, 0.0964, 0.0017, -0.0627, 0.0959, 0.0039, -0.0666, 0.0964, 0.0039, -0.0627, 0.0899, -0.0, -0.0613, 0.091, -0.0, -0.0586, 0.0926, -0.0, -0.0608, 0.0915, -0.0, -0.0595, 0.0904, -0.0, -0.0568, 0.0902, -0.0, -0.0548, 0.0908, 0.0017, -0.0583, 0.0908, 0.0039, -0.0583, 0.0926, 0.0017, -0.0608, 0.0914, 0.0017, -0.0594, 0.0926, 0.0039, -0.0608, 0.0914, 0.0039, -0.0594, 0.0904, 0.0017, -0.057, 0.0903, 0.0017, -0.0548, 0.0904, 0.0039, -0.057, 0.0903, 0.0039, -0.0548, 0.0942, -0.0, -0.0621, 0.0953, -0.0, -0.0624, 0.0942, 0.0017, -0.0621,
0.0942, 0.0039, -0.0621, 0.0953, 0.0017, -0.0624, 0.0953, 0.0039, -0.0624, 0.0985, -0.0, -0.0669, 0.1007, -0.0, -0.067, 0.0985, -0.0, -0.0656, 0.0986, -0.0, -0.063, 0.1007, -0.0, -0.0656, 0.1007, -0.0, -0.0631, 0.0985, 0.0017, -0.0669, 0.0985, 0.0039, -0.0669, 0.1007, 0.0017, -0.067, 0.1007, 0.0039, -0.067, 0.0985, 0.0017, -0.063, 0.0985, 0.0039, -0.063, 0.1007, 0.0017, -0.0631, 0.1007, 0.0039, -0.0631, 0.1029, -0.0, -0.067, 0.105, -0.0, -0.0669, 0.1029, -0.0, -0.0656, 0.1029, -0.0, -0.0631, 0.105, -0.0, -0.0656, 0.105, -0.0, -0.063, 0.1029, 0.0017, -0.067, 0.1029, 0.0039, -0.067, 0.105, 0.0017, -0.067, 0.105, 0.0039, -0.067, 0.1029, 0.0017, -0.0631, 0.1029, 0.0039, -0.0631, 0.105, 0.0017, -0.063,
0.105, 0.0039, -0.063, 0.0904, -0.0, -0.0528, 0.0909, -0.0, -0.0508, 0.0913, -0.0, -0.0499, 0.0899, -0.0, -0.0483, 0.0894, -0.0, -0.0456, 0.0922, -0.0, -0.0485, 0.092, -0.0, -0.0461, 0.0904, 0.0017, -0.0526, 0.0908, 0.0017, -0.0512, 0.0904, 0.0039, -0.0526, 0.0908, 0.0039, -0.0512, 0.0912, 0.0017, -0.05, 0.0912, 0.0039, -0.05, 0.0894, 0.0017, -0.0456, 0.0894, 0.0039, -0.0456, 0.0923, 0.0017, -0.0485, 0.0923, 0.0039, -0.0485, 0.0934, -0.0, -0.0475, 0.0945, -0.0, -0.0469, 0.0965, -0.0, -0.046, 0.0935, 0.0017, -0.0474, 0.0946, 0.0017, -0.0468, 0.0935, 0.0039, -0.0474, 0.0946, 0.0039, -0.0468, 0.0963, 0.0017, -0.0461, 0.0963, 0.0039, -0.0461, 0.0905, -0.0, -0.0447, 0.0917, -0.0, -0.0438, 0.0905, 0.0017, -0.0447,
0.0905, 0.0039, -0.0447, 0.092, 0.0017, -0.0437, 0.092, 0.0039, -0.0437, 0.0939, -0.0, -0.0429, 0.0949, -0.0, -0.0427, 0.0964, -0.0, -0.0439, 0.0963, -0.0, -0.0425, 0.0937, 0.0017, -0.043, 0.0948, 0.0017, -0.0427, 0.0937, 0.0039, -0.043, 0.0948, 0.0039, -0.0427, 0.0964, 0.0017, -0.0425, 0.0964, 0.0039, -0.0425, 0.0986, -0.0, -0.0456, 0.1007, -0.0, -0.0454, 0.0985, 0.0017, -0.0456, 0.0985, 0.0039, -0.0456, 0.1007, 0.0017, -0.0454, 0.1007, 0.0039, -0.0454, 0.1029, -0.0, -0.0455, 0.105, -0.0, -0.0457, 0.1029, 0.0017, -0.0455, 0.1029, 0.0039, -0.0455, 0.1051, 0.0017, -0.0457, 0.1051, 0.0039, -0.0457, 0.0985, -0.0, -0.0439, 0.0985, -0.0, -0.0425, 0.1007, -0.0, -0.0439, 0.1007, -0.0, -0.0425, 0.0985, 0.0017, -0.0425,
0.0985, 0.0039, -0.0425, 0.1007, 0.0017, -0.0425, 0.1007, 0.0039, -0.0425, 0.1029, -0.0, -0.0439, 0.1029, -0.0, -0.0425, 0.105, -0.0, -0.0439, 0.105, -0.0, -0.0425, 0.1029, 0.0017, -0.0425, 0.1029, 0.0039, -0.0425, 0.105, 0.0017, -0.0425, 0.105, 0.0039, -0.0425, 0.0716, -0.0, -0.0219, 0.0716, -0.0, -0.0201, 0.0747, -0.0, -0.0219, 0.0747, -0.0, -0.0201, 0.0716, 0.0017, -0.0219, 0.0716, 0.0017, -0.0201, 0.0716, 0.0039, -0.0219, 0.0716, 0.0039, -0.0201, 0.0747, 0.0017, -0.0219, 0.0747, 0.0039, -0.0219, 0.0768, -0.0, -0.0219, 0.0768, -0.0, -0.0201, 0.079, -0.0, -0.0219, 0.079, -0.0, -0.0201, 0.0768, 0.0017, -0.0219, 0.0768, 0.0039, -0.0219, 0.079, 0.0017, -0.0219, 0.079, 0.0039, -0.0219, 0.0812, -0.0, -0.0219,
0.0812, -0.0, -0.0201, 0.0833, -0.0, -0.0219, 0.0833, -0.0, -0.0201, 0.0812, 0.0017, -0.0219, 0.0812, 0.0039, -0.0219, 0.0833, 0.0017, -0.0219, 0.0833, 0.0017, -0.0201, 0.0833, 0.0039, -0.0219, 0.0833, 0.0039, -0.0201, 0.0716, -0.0, -0.0179, 0.0716, -0.0, -0.0157, 0.0747, -0.0, -0.0179, 0.0747, -0.0, -0.0157, 0.0716, -0.0, -0.0136, 0.0716, -0.0, -0.0114, 0.0747, -0.0, -0.0136, 0.0747, -0.0, -0.0114, 0.0716, 0.0017, -0.0179, 0.0716, 0.0017, -0.0157, 0.0716, 0.0039, -0.0179, 0.0716, 0.0039, -0.0157, 0.0716, 0.0017, -0.0136, 0.0716, 0.0017, -0.0114, 0.0716, 0.0039, -0.0136, 0.0716, 0.0039, -0.0114, 0.0768, -0.0, -0.0179, 0.0768, -0.0, -0.0157, 0.079, -0.0, -0.0179, 0.079, -0.0, -0.0157, 0.0768, -0.0, -0.0136,
0.0768, -0.0, -0.0114, 0.079, -0.0, -0.0136, 0.079, -0.0, -0.0114, 0.0716, -0.0, -0.0092, 0.0716, -0.0, -0.0071, 0.0747, -0.0, -0.0092, 0.0747, -0.0, -0.0071, 0.0716, -0.0, -0.0049, 0.0716, -0.0, -0.0027, 0.0747, -0.0, -0.0049, 0.0747, -0.0, -0.0027, 0.0716, 0.0017, -0.0092, 0.0716, 0.0017, -0.0071, 0.0716, 0.0039, -0.0092, 0.0716, 0.0039, -0.0071, 0.0716, 0.0017, -0.0049, 0.0716, 0.0017, -0.0027, 0.0716, 0.0039, -0.0049, 0.0716, 0.0039, -0.0027, 0.0768, -0.0, -0.0092, 0.0768, -0.0, -0.0071, 0.079, -0.0, -0.0092, 0.079, -0.0, -0.0071, 0.0768, -0.0, -0.0049, 0.0768, -0.0, -0.0027, 0.079, -0.0, -0.0049, 0.079, -0.0, -0.0027, 0.0812, -0.0, -0.0179, 0.0812, -0.0, -0.0157, 0.0833, -0.0, -0.0179,
0.0833, -0.0, -0.0157, 0.0812, -0.0, -0.0136, 0.0812, -0.0, -0.0114, 0.0833, -0.0, -0.0136, 0.0833, -0.0, -0.0114, 0.0833, 0.0017, -0.0179, 0.0833, 0.0017, -0.0157, 0.0833, 0.0039, -0.0179, 0.0833, 0.0039, -0.0157, 0.0833, 0.0017, -0.0136, 0.0833, 0.0017, -0.0114, 0.0833, 0.0039, -0.0136, 0.0833, 0.0039, -0.0114, 0.0812, -0.0, -0.0092, 0.0812, -0.0, -0.0071, 0.0833, -0.0, -0.0092, 0.0833, -0.0, -0.0071, 0.0812, -0.0, -0.0049, 0.0812, -0.0, -0.0027, 0.0833, -0.0, -0.0049, 0.0833, -0.0, -0.0027, 0.0833, 0.0017, -0.0092, 0.0833, 0.0017, -0.0071, 0.0833, 0.0039, -0.0092, 0.0833, 0.0039, -0.0071, 0.0833, 0.0017, -0.0049, 0.0833, 0.0017, -0.0027, 0.0833, 0.0039, -0.0049, 0.0833, 0.0039, -0.0027, 0.1065, -0.0, -0.0668,
0.1074, -0.0, -0.0667, 0.1072, -0.0, -0.0626, 0.1097, -0.0, -0.0661, 0.1094, -0.0, -0.0635, 0.1066, 0.0017, -0.0668, 0.1066, 0.0039, -0.0668, 0.1077, 0.0017, -0.0666, 0.1071, 0.0017, -0.0627, 0.1077, 0.0039, -0.0666, 0.1071, 0.0039, -0.0627, 0.1094, 0.0017, -0.0662, 0.1094, 0.0039, -0.0662, 0.1113, -0.0, -0.0653, 0.1123, -0.0, -0.0646, 0.1138, -0.0, -0.0636, 0.1115, 0.0017, -0.0652, 0.1125, 0.0017, -0.0645, 0.1115, 0.0039, -0.0652, 0.1125, 0.0039, -0.0645, 0.1138, 0.0017, -0.0635, 0.1138, 0.0039, -0.0635, 0.1082, -0.0, -0.0623, 0.1093, -0.0, -0.062, 0.1082, 0.0017, -0.0624, 0.1082, 0.0039, -0.0624, 0.1094, 0.0017, -0.062, 0.1094, 0.0039, -0.062, 0.1112, -0.0, -0.0608, 0.1122, -0.0, -0.0596, 0.1148, -0.0, -0.0622,
0.1127, -0.0, -0.0587, 0.1134, -0.0, -0.0569, 0.1136, -0.0, -0.0548, 0.1111, 0.0017, -0.0608, 0.1122, 0.0017, -0.0596, 0.1111, 0.0039, -0.0608, 0.1122, 0.0039, -0.0596, 0.1148, 0.0017, -0.0623, 0.1129, 0.0017, -0.0584, 0.1148, 0.0039, -0.0623, 0.1129, 0.0039, -0.0584, 0.1134, 0.0017, -0.057, 0.1136, 0.0017, -0.0548, 0.1134, 0.0039, -0.057, 0.1136, 0.0039, -0.0548, 0.1156, -0.0, -0.0612, 0.1167, -0.0, -0.0593, 0.117, -0.0, -0.0581, 0.1159, -0.0, -0.057, 0.1159, -0.0, -0.0548, 0.1173, -0.0, -0.057, 0.1176, -0.0, -0.0548, 0.1155, 0.0017, -0.0613, 0.1167, 0.0017, -0.0593, 0.1155, 0.0039, -0.0613, 0.1167, 0.0039, -0.0593, 0.117, 0.0017, -0.0581, 0.117, 0.0039, -0.0581, 0.1173, 0.0017, -0.057, 0.1176, 0.0017, -0.0548,
0.1173, 0.0039, -0.057, 0.1176, 0.0039, -0.0548, 0.107, -0.0, -0.0462, 0.11, -0.0, -0.0476, 0.109, -0.0, -0.047, 0.1072, 0.0017, -0.0463, 0.1072, 0.0039, -0.0463, 0.1099, 0.0017, -0.0476, 0.1088, 0.0017, -0.047, 0.1099, 0.0039, -0.0476, 0.1088, 0.0039, -0.047, 0.1122, -0.0, -0.0498, 0.1134, -0.0, -0.0528, 0.1128, -0.0, -0.0508, 0.1112, -0.0, -0.0486, 0.1116, -0.0, -0.0461, 0.1137, -0.0, -0.0483, 0.1142, -0.0, -0.0456, 0.1123, 0.0017, -0.0499, 0.1123, 0.0039, -0.0499, 0.1134, 0.0017, -0.0526, 0.1129, 0.0017, -0.051, 0.1134, 0.0039, -0.0526, 0.1129, 0.0039, -0.051, 0.1112, 0.0017, -0.0486, 0.1112, 0.0039, -0.0486, 0.1142, 0.0017, -0.0456, 0.1142, 0.0039, -0.0456, 0.1072, -0.0, -0.0439, 0.1072, -0.0, -0.0426,
0.1096, -0.0, -0.043, 0.1086, -0.0, -0.0427, 0.1072, 0.0017, -0.0426, 0.1072, 0.0039, -0.0426, 0.1098, 0.0017, -0.043, 0.1088, 0.0017, -0.0427, 0.1098, 0.0039, -0.043, 0.1088, 0.0039, -0.0427, 0.1119, -0.0, -0.0437, 0.1131, -0.0, -0.0447, 0.1116, 0.0017, -0.0436, 0.1116, 0.0039, -0.0436, 0.1131, 0.0017, -0.0447, 0.1131, 0.0039, -0.0447, 0.1159, -0.0, -0.0526, 0.1169, -0.0, -0.0501, 0.1175, -0.0, -0.0526, 0.1172, -0.0, -0.0511, 0.1161, -0.0, -0.048, 0.1152, -0.0, -0.0468, 0.1168, 0.0017, -0.0498, 0.1168, 0.0039, -0.0498, 0.1175, 0.0017, -0.0526, 0.1172, 0.0017, -0.0509, 0.1175, 0.0039, -0.0526, 0.1172, 0.0039, -0.0509, 0.1162, 0.0017, -0.0482, 0.1152, 0.0017, -0.0467, 0.1162, 0.0039, -0.0482, 0.1152, 0.0039, -0.0467,
0.1363, -0.0, -0.067, 0.1363, -0.0, -0.0656, 0.1363, -0.0, -0.0635, 0.1363, 0.0017, -0.067, 0.1363, 0.0039, -0.067, 0.1363, 0.0017, -0.0656, 0.1363, 0.0017, -0.0635, 0.1363, 0.0039, -0.0656, 0.1363, 0.0039, -0.0635, 0.1376, -0.0, -0.067, 0.1402, -0.0, -0.067, 0.1376, -0.0, -0.0656, 0.1376, -0.0, -0.0635, 0.1402, -0.0, -0.0656, 0.1402, -0.0, -0.0635, 0.1376, 0.0017, -0.067, 0.1376, 0.0039, -0.067, 0.1402, 0.0017, -0.067, 0.1402, 0.0039, -0.067, 0.1402, 0.0017, -0.0656, 0.1402, 0.0017, -0.0635, 0.1402, 0.0039, -0.0656, 0.1402, 0.0039, -0.0635, 0.1363, -0.0, -0.0613, 0.1363, -0.0, -0.0591, 0.1363, -0.0, -0.057, 0.1363, -0.0, -0.0548, 0.1363, 0.0017, -0.0613, 0.1363, 0.0017, -0.0591, 0.1363, 0.0039, -0.0613,
0.1363, 0.0039, -0.0591, 0.1363, 0.0017, -0.057, 0.1363, 0.0017, -0.0548, 0.1363, 0.0039, -0.057, 0.1363, 0.0039, -0.0548, 0.1376, -0.0, -0.0613, 0.1376, -0.0, -0.0591, 0.1402, -0.0, -0.0613, 0.1402, -0.0, -0.0591, 0.1376, -0.0, -0.057, 0.1376, -0.0, -0.0548, 0.1402, -0.0, -0.057, 0.1402, -0.0, -0.0548, 0.1402, 0.0017, -0.0613, 0.1402, 0.0017, -0.0591, 0.1402, 0.0039, -0.0613, 0.1402, 0.0039, -0.0591, 0.1402, 0.0017, -0.057, 0.1402, 0.0017, -0.0548, 0.1402, 0.0039, -0.057, 0.1402, 0.0039, -0.0548, 0.1363, -0.0, -0.0526, 0.1363, -0.0, -0.0505, 0.1363, -0.0, -0.0483, 0.1363, -0.0, -0.0461, 0.1363, 0.0017, -0.0526, 0.1363, 0.0017, -0.0505, 0.1363, 0.0039, -0.0526, 0.1363, 0.0039, -0.0505, 0.1363, 0.0017, -0.0483,
0.1363, 0.0017, -0.0461, 0.1363, 0.0039, -0.0483, 0.1363, 0.0039, -0.0461, 0.1376, -0.0, -0.0526, 0.1376, -0.0, -0.0505, 0.1402, -0.0, -0.0526, 0.1402, -0.0, -0.0505, 0.1376, -0.0, -0.0483, 0.1376, -0.0, -0.0461, 0.1402, -0.0, -0.0483, 0.1402, -0.0, -0.0454, 0.1402, 0.0017, -0.0526, 0.1402, 0.0017, -0.0505, 0.1402, 0.0039, -0.0526, 0.1402, 0.0039, -0.0505, 0.1402, 0.0017, -0.0483, 0.1402, 0.0017, -0.0454, 0.1402, 0.0039, -0.0483, 0.1402, 0.0039, -0.0454, 0.1363, -0.0, -0.0439, 0.1363, -0.0, -0.0425, 0.1363, 0.0017, -0.0439, 0.1363, 0.0017, -0.0425, 0.1363, 0.0039, -0.0439, 0.1363, 0.0039, -0.0425, 0.1376, -0.0, -0.0439, 0.1376, -0.0, -0.0425, 0.1398, -0.0, -0.0439, 0.1398, -0.0, -0.0425, 0.1376, 0.0017, -0.0425,
0.1376, 0.0039, -0.0425, 0.1398, 0.0017, -0.0425, 0.1398, 0.0039, -0.0425, 0.1431, -0.0, -0.0591, 0.1431, -0.0, -0.0562, 0.1431, 0.0017, -0.0591, 0.1431, 0.0039, -0.0591, 0.1431, 0.0017, -0.0562, 0.1431, 0.0039, -0.0562, 0.1463, -0.0, -0.0591, 0.1484, -0.0, -0.0592, 0.1463, -0.0, -0.0562, 0.1484, -0.0, -0.0562, 0.1463, 0.0017, -0.0591, 0.1463, 0.0039, -0.0591, 0.1485, 0.0017, -0.0592, 0.1485, 0.0039, -0.0592, 0.1463, 0.0017, -0.0562, 0.1463, 0.0039, -0.0562, 0.1484, 0.0017, -0.0562, 0.1484, 0.0039, -0.0562, 0.1553, -0.0, -0.0632, 0.1576, -0.0, -0.0652, 0.1564, -0.0, -0.0643, 0.1553, 0.0017, -0.0632, 0.1553, 0.0039, -0.0632, 0.1575, 0.0017, -0.0651, 0.1564, 0.0017, -0.0643, 0.1575, 0.0039, -0.0651, 0.1564, 0.0039, -0.0643,
0.1504, -0.0, -0.0596, 0.153, -0.0, -0.061, 0.1521, -0.0, -0.0603, 0.1506, -0.0, -0.0562, 0.1527, -0.0, -0.0562, 0.1506, 0.0017, -0.0596, 0.1506, 0.0039, -0.0596, 0.1529, 0.0017, -0.0609, 0.1518, 0.0017, -0.0602, 0.1529, 0.0039, -0.0609, 0.1518, 0.0039, -0.0602, 0.1506, 0.0017, -0.0562, 0.1506, 0.0039, -0.0562, 0.1528, 0.0017, -0.0562, 0.1528, 0.0039, -0.0562, 0.1542, -0.0, -0.0621, 0.155, -0.0, -0.0591, 0.1571, -0.0, -0.0613, 0.1571, -0.0, -0.0591, 0.1549, -0.0, -0.056, 0.1557, -0.0, -0.0557, 0.1571, -0.0, -0.057, 0.1571, -0.0, -0.0553, 0.1542, 0.0017, -0.0621, 0.1542, 0.0039, -0.0621, 0.1546, 0.0017, -0.0561, 0.1556, 0.0017, -0.0558, 0.1546, 0.0039, -0.0561, 0.1556, 0.0039, -0.0558, 0.1572, 0.0017, -0.0553,
0.1572, 0.0039, -0.0553, 0.1419, -0.0, -0.0454, 0.1441, -0.0, -0.0454, 0.1419, 0.0017, -0.0454, 0.1419, 0.0039, -0.0454, 0.1441, 0.0017, -0.0454, 0.1441, 0.0039, -0.0454, 0.1463, -0.0, -0.0454, 0.1484, -0.0, -0.0454, 0.1463, 0.0017, -0.0454, 0.1463, 0.0039, -0.0454, 0.1484, 0.0017, -0.0454, 0.1484, 0.0039, -0.0454, 0.1419, -0.0, -0.0439, 0.1419, -0.0, -0.0425, 0.1441, -0.0, -0.0439, 0.1441, -0.0, -0.0425, 0.1419, 0.0017, -0.0425, 0.1419, 0.0039, -0.0425, 0.1441, 0.0017, -0.0425, 0.1441, 0.0039, -0.0425, 0.1463, -0.0, -0.0439, 0.1463, -0.0, -0.0425, 0.1484, -0.0, -0.0439, 0.1484, -0.0, -0.0425, 0.1463, 0.0017, -0.0425, 0.1463, 0.0039, -0.0425, 0.1484, 0.0017, -0.0425, 0.1484, 0.0039, -0.0425, 0.1506, -0.0, -0.0454,
0.1528, -0.0, -0.0456, 0.1506, 0.0017, -0.0454, 0.1506, 0.0039, -0.0454, 0.1528, 0.0017, -0.0456, 0.1528, 0.0039, -0.0456, 0.1549, -0.0, -0.0459, 0.1579, -0.0, -0.0475, 0.1569, -0.0, -0.0468, 0.155, 0.0017, -0.0459, 0.155, 0.0039, -0.0459, 0.1579, 0.0017, -0.0474, 0.1568, 0.0017, -0.0467, 0.1579, 0.0039, -0.0474, 0.1568, 0.0039, -0.0467, 0.1506, -0.0, -0.0439, 0.1506, -0.0, -0.0425, 0.1528, -0.0, -0.0439, 0.1528, -0.0, -0.0425, 0.1506, 0.0017, -0.0425, 0.1506, 0.0039, -0.0425, 0.1528, 0.0017, -0.0425, 0.1528, 0.0039, -0.0425, 0.155, -0.0, -0.0439, 0.155, -0.0, -0.0425, 0.1571, -0.0, -0.0439, 0.1571, -0.0, -0.0426, 0.155, 0.0017, -0.0425, 0.155, 0.0039, -0.0425, 0.1571, 0.0017, -0.0426, 0.1571, 0.0039, -0.0426,
0.1621, -0.0, -0.0669, 0.159, -0.0, -0.0661, 0.1593, -0.0, -0.0635, 0.1612, -0.0, -0.0667, 0.1619, -0.0, -0.0632, 0.1619, 0.0017, -0.0669, 0.1619, 0.0039, -0.0669, 0.1592, 0.0017, -0.0662, 0.1592, 0.0039, -0.0662, 0.1608, 0.0017, -0.0666, 0.1619, 0.0017, -0.0632, 0.1608, 0.0039, -0.0666, 0.1619, 0.0039, -0.0632, 0.1648, -0.0, -0.0671, 0.1636, -0.0, -0.0653, 0.1628, -0.0, -0.0643, 0.1648, 0.0017, -0.0671, 0.1648, 0.0039, -0.0671, 0.1638, 0.0017, -0.0656, 0.1628, 0.0017, -0.0643, 0.1638, 0.0039, -0.0656, 0.1628, 0.0039, -0.0643, 0.16, -0.0, -0.0609, 0.1588, -0.0, -0.0591, 0.1608, -0.0, -0.0619, 0.1606, -0.0, -0.0583, 0.1593, -0.0, -0.057, 0.1588, -0.0, -0.0543, 0.1618, -0.0, -0.0574, 0.1615, -0.0, -0.0548,
0.1599, 0.0017, -0.0608, 0.1588, 0.0017, -0.0591, 0.1599, 0.0039, -0.0608, 0.1588, 0.0039, -0.0591, 0.1608, 0.0017, -0.0619, 0.1607, 0.0017, -0.0583, 0.1608, 0.0039, -0.0619, 0.1607, 0.0039, -0.0583, 0.1587, 0.0017, -0.0543, 0.1587, 0.0039, -0.0543, 0.1618, 0.0017, -0.0574, 0.1618, 0.0039, -0.0574, 0.1629, -0.0, -0.0563, 0.1637, -0.0, -0.0551, 0.1629, 0.0017, -0.0563, 0.1638, 0.0017, -0.0549, 0.1629, 0.0039, -0.0563, 0.1638, 0.0039, -0.0549, 0.1598, -0.0, -0.0525, 0.1603, -0.0, -0.0503, 0.1615, -0.0, -0.0526, 0.1615, -0.0, -0.0505, 0.1591, -0.0, -0.0484, 0.1593, -0.0, -0.0461, 0.1615, -0.0, -0.0483, 0.1615, -0.0, -0.0461, 0.1596, 0.0017, -0.053, 0.1603, 0.0017, -0.0503, 0.1596, 0.0039, -0.053, 0.1603, 0.0039, -0.0503,
0.1585, 0.0017, -0.0479, 0.1585, 0.0039, -0.0479, 0.1645, -0.0, -0.0529, 0.1647, -0.0, -0.0505, 0.1644, -0.0, -0.048, 0.1633, -0.0, -0.0459, 0.1646, 0.0017, -0.0527, 0.1647, 0.0017, -0.0505, 0.1646, 0.0039, -0.0527, 0.1647, 0.0039, -0.0505, 0.1645, 0.0017, -0.048, 0.1634, 0.0017, -0.046, 0.1645, 0.0039, -0.048, 0.1634, 0.0039, -0.046, 0.1595, -0.0, -0.043, 0.1584, -0.0, -0.0428, 0.1618, -0.0, -0.0442, 0.1596, 0.0017, -0.0431, 0.1585, 0.0017, -0.0428, 0.1596, 0.0039, -0.0431, 0.1585, 0.0039, -0.0428, 0.162, 0.0017, -0.0444, 0.162, 0.0039, -0.0444, 0.1626, -0.0, -0.045, 0.1626, 0.0017, -0.045, 0.1626, 0.0039, -0.045, 0.1412, -0.0, -0.0219, 0.1412, -0.0, -0.0201, 0.1441, -0.0, -0.0219, 0.1441, -0.0, -0.0201,
0.1412, 0.0017, -0.0219, 0.1412, 0.0017, -0.0201, 0.1412, 0.0039, -0.0219, 0.1412, 0.0039, -0.0201, 0.1441, 0.0017, -0.0219, 0.1441, 0.0039, -0.0219, 0.1463, -0.0, -0.0219, 0.1463, -0.0, -0.0201, 0.1484, -0.0, -0.0219, 0.1484, -0.0, -0.0201, 0.1463, 0.0017, -0.0219, 0.1463, 0.0039, -0.0219, 0.1484, 0.0017, -0.0219, 0.1484, 0.0039, -0.0219, 0.1506, -0.0, -0.0219, 0.1506, -0.0, -0.0201, 0.1528, -0.0, -0.0219, 0.1528, -0.0, -0.0201, 0.1506, 0.0017, -0.0219, 0.1506, 0.0039, -0.0219, 0.1528, 0.0017, -0.0219, 0.1528, 0.0039, -0.0219, 0.155, -0.0, -0.0219, 0.155, -0.0, -0.0201, 0.1571, -0.0, -0.0219, 0.1571, -0.0, -0.0201, 0.155, 0.0017, -0.0219, 0.155, 0.0039, -0.0219, 0.1571, 0.0017, -0.0219, 0.1571, 0.0039, -0.0219,
0.1412, -0.0, -0.0179, 0.1412, -0.0, -0.0157, 0.1441, -0.0, -0.0179, 0.1441, -0.0, -0.0157, 0.1412, -0.0, -0.0136, 0.1412, -0.0, -0.0114, 0.1441, -0.0, -0.0136, 0.1441, -0.0, -0.0114, 0.1412, 0.0017, -0.0179, 0.1412, 0.0017, -0.0157, 0.1412, 0.0039, -0.0179, 0.1412, 0.0039, -0.0157, 0.1412, 0.0017, -0.0136, 0.1412, 0.0017, -0.0114, 0.1412, 0.0039, -0.0136, 0.1412, 0.0039, -0.0114, 0.1463, -0.0, -0.0179, 0.1463, -0.0, -0.0157, 0.1484, -0.0, -0.0179, 0.1484, -0.0, -0.0157, 0.1463, -0.0, -0.0136, 0.1463, -0.0, -0.0114, 0.1484, -0.0, -0.0136, 0.1484, -0.0, -0.0114, 0.1412, -0.0, -0.0092, 0.1412, -0.0, -0.0071, 0.1441, -0.0, -0.0092, 0.1441, -0.0, -0.0071, 0.1412, -0.0, -0.0049, 0.1412, -0.0, -0.0027,
0.1441, -0.0, -0.0049, 0.1441, -0.0, -0.0027, 0.1412, 0.0017, -0.0092, 0.1412, 0.0017, -0.0071, 0.1412, 0.0039, -0.0092, 0.1412, 0.0039, -0.0071, 0.1412, 0.0017, -0.0049, 0.1412, 0.0017, -0.0027, 0.1412, 0.0039, -0.0049, 0.1412, 0.0039, -0.0027, 0.1463, -0.0, -0.0092, 0.1463, -0.0, -0.0071, 0.1484, -0.0, -0.0092, 0.1484, -0.0, -0.0071, 0.1463, -0.0, -0.0049, 0.1463, -0.0, -0.0027, 0.1484, -0.0, -0.0049, 0.1484, -0.0, -0.0027, 0.1506, -0.0, -0.0179, 0.1506, -0.0, -0.0157, 0.1528, -0.0, -0.0179, 0.1528, -0.0, -0.0157, 0.1506, -0.0, -0.0136, 0.1506, -0.0, -0.0114, 0.1528, -0.0, -0.0136, 0.1528, -0.0, -0.0114, 0.155, -0.0, -0.0179, 0.155, -0.0, -0.0157, 0.1571, -0.0, -0.0179, 0.1571, -0.0, -0.0157,
0.155, -0.0, -0.0136, 0.155, -0.0, -0.0114, 0.1571, -0.0, -0.0136, 0.1571, -0.0, -0.0114, 0.1506, -0.0, -0.0092, 0.1506, -0.0, -0.0071, 0.152, -0.0, -0.0101, 0.152, -0.0, -0.0071, 0.1506, -0.0, -0.0049, 0.1506, -0.0, -0.0027, 0.152, -0.0, -0.0049, 0.152, -0.0, -0.0027, 0.152, 0.0017, -0.0101, 0.152, 0.0017, -0.0071, 0.152, 0.0039, -0.0101, 0.152, 0.0039, -0.0071, 0.152, 0.0017, -0.0049, 0.152, 0.0017, -0.0027, 0.152, 0.0039, -0.0049, 0.152, 0.0039, -0.0027, 0.155, -0.0, -0.0101, 0.1571, -0.0, -0.0101, 0.155, 0.0017, -0.0101, 0.155, 0.0039, -0.0101, 0.1571, 0.0017, -0.0101, 0.1571, 0.0039, -0.0101, 0.1593, -0.0, -0.0219, 0.1593, -0.0, -0.0201, 0.1615, -0.0, -0.0219, 0.1615, -0.0, -0.0201,
0.1593, 0.0017, -0.0219, 0.1593, 0.0039, -0.0219, 0.1615, 0.0017, -0.0219, 0.1615, 0.0039, -0.0219, 0.1636, -0.0, -0.0219, 0.1636, -0.0, -0.0201, 0.1658, -0.0, -0.0219, 0.1658, -0.0, -0.0201, 0.1636, 0.0017, -0.0219, 0.1636, 0.0039, -0.0219, 0.1658, 0.0017, -0.0219, 0.1658, 0.0039, -0.0219, 0.168, -0.0, -0.0219, 0.168, -0.0, -0.0201, 0.1701, -0.0, -0.0219, 0.1701, -0.0, -0.0201, 0.168, 0.0017, -0.0219, 0.168, 0.0039, -0.0219, 0.1701, 0.0017, -0.0219, 0.1701, 0.0039, -0.0219, 0.1723, -0.0, -0.0219, 0.1723, -0.0, -0.0201, 0.1745, -0.0, -0.0219, 0.1745, -0.0, -0.0201, 0.1723, 0.0017, -0.0219, 0.1723, 0.0039, -0.0219, 0.1745, 0.0017, -0.0219, 0.1745, 0.0039, -0.0219, 0.1593, -0.0, -0.0179, 0.1593, -0.0, -0.0157,
0.1615, -0.0, -0.0179, 0.1615, -0.0, -0.0157, 0.1593, -0.0, -0.0136, 0.1593, -0.0, -0.0114, 0.1615, -0.0, -0.0136, 0.1615, -0.0, -0.0114, 0.1636, -0.0, -0.0179, 0.1636, -0.0, -0.0157, 0.1658, -0.0, -0.0179, 0.1658, -0.0, -0.0157, 0.1636, -0.0, -0.0136, 0.1636, -0.0, -0.0114, 0.1658, -0.0, -0.0136, 0.1658, -0.0, -0.0114, 0.1593, -0.0, -0.0101, 0.1615, -0.0, -0.0101, 0.1593, 0.0017, -0.0101, 0.1593, 0.0039, -0.0101, 0.1615, 0.0017, -0.0101, 0.1615, 0.0039, -0.0101, 0.1636, -0.0, -0.0101, 0.1658, -0.0, -0.0101, 0.1636, 0.0017, -0.0101, 0.1636, 0.0039, -0.0101, 0.1658, 0.0017, -0.0101, 0.1658, 0.0039, -0.0101, 0.168, -0.0, -0.0179, 0.168, -0.0, -0.0157, 0.1701, -0.0, -0.0179, 0.1701, -0.0, -0.0157,
0.168, -0.0, -0.0136, 0.168, -0.0, -0.0114, 0.1701, -0.0, -0.0136, 0.1701, -0.0, -0.0114, 0.1723, -0.0, -0.0179, 0.1723, -0.0, -0.0157, 0.1745, -0.0, -0.0179, 0.1745, -0.0, -0.0157, 0.1723, -0.0, -0.0136, 0.1723, -0.0, -0.0114, 0.1745, -0.0, -0.0136, 0.1745, -0.0, -0.0114, 0.168, -0.0, -0.0101, 0.1701, -0.0, -0.0101, 0.168, 0.0017, -0.0101, 0.168, 0.0039, -0.0101, 0.1701, 0.0017, -0.0101, 0.1701, 0.0039, -0.0101, 0.1723, -0.0, -0.0101, 0.1745, -0.0, -0.0101, 0.1723, 0.0017, -0.0101, 0.1723, 0.0039, -0.0101, 0.1745, 0.0017, -0.0101, 0.1745, 0.0039, -0.0101, 0.1767, -0.0, -0.0219, 0.1767, -0.0, -0.0201, 0.1788, -0.0, -0.0219, 0.1788, -0.0, -0.0201, 0.1767, 0.0017, -0.0219, 0.1767, 0.0039, -0.0219,
0.1788, 0.0017, -0.0219, 0.1788, 0.0039, -0.0219, 0.181, -0.0, -0.0219, 0.181, -0.0, -0.0201, 0.1832, -0.0, -0.0219, 0.1832, -0.0, -0.0201, 0.181, 0.0017, -0.0219, 0.181, 0.0039, -0.0219, 0.1832, 0.0017, -0.0219, 0.1832, 0.0039, -0.0219, 0.1853, -0.0, -0.0219, 0.1853, -0.0, -0.0201, 0.1875, -0.0, -0.0219, 0.1875, -0.0, -0.0201, 0.1853, 0.0017, -0.0219, 0.1853, 0.0039, -0.0219, 0.1875, 0.0017, -0.0219, 0.1875, 0.0039, -0.0219, 0.1897, -0.0, -0.0219, 0.1897, -0.0, -0.0201, 0.1918, -0.0, -0.0219, 0.1918, -0.0, -0.0201, 0.1897, 0.0017, -0.0219, 0.1897, 0.0039, -0.0219, 0.1918, 0.0017, -0.0219, 0.1918, 0.0039, -0.0219, 0.1767, -0.0, -0.0179, 0.1767, -0.0, -0.0157, 0.1788, -0.0, -0.0179, 0.1788, -0.0, -0.0157,
0.1767, -0.0, -0.0136, 0.1767, -0.0, -0.0114, 0.1788, -0.0, -0.0136, 0.1788, -0.0, -0.0114, 0.181, -0.0, -0.0179, 0.181, -0.0, -0.0157, 0.1832, -0.0, -0.0179, 0.1832, -0.0, -0.0157, 0.181, -0.0, -0.0136, 0.181, -0.0, -0.0114, 0.1832, -0.0, -0.0136, 0.1832, -0.0, -0.0114, 0.1767, -0.0, -0.0101, 0.1788, -0.0, -0.0101, 0.1767, 0.0017, -0.0101, 0.1767, 0.0039, -0.0101, 0.1788, 0.0017, -0.0101, 0.1788, 0.0039, -0.0101, 0.181, -0.0, -0.0101, 0.1832, -0.0, -0.0101, 0.181, 0.0017, -0.0101, 0.181, 0.0039, -0.0101, 0.1832, 0.0017, -0.0101, 0.1832, 0.0039, -0.0101, 0.1853, -0.0, -0.0179, 0.1853, -0.0, -0.0157, 0.1875, -0.0, -0.0179, 0.1875, -0.0, -0.0157, 0.1853, -0.0, -0.0136, 0.1853, -0.0, -0.0114,
0.1875, -0.0, -0.0136, 0.1875, -0.0, -0.0114, 0.1897, -0.0, -0.0179, 0.1897, -0.0, -0.0157, 0.1918, -0.0, -0.0179, 0.1918, -0.0, -0.0157, 0.1897, -0.0, -0.0136, 0.1897, -0.0, -0.0114, 0.1918, -0.0, -0.0136, 0.1918, -0.0, -0.0114, 0.1853, -0.0, -0.0101, 0.1875, -0.0, -0.0101, 0.1853, 0.0017, -0.0101, 0.1853, 0.0039, -0.0101, 0.1875, 0.0017, -0.0101, 0.1875, 0.0039, -0.0101, 0.1897, -0.0, -0.0101, 0.1918, -0.0, -0.0101, 0.1897, 0.0017, -0.0101, 0.1897, 0.0039, -0.0101, 0.1918, 0.0017, -0.0101, 0.1918, 0.0039, -0.0101, 0.194, -0.0, -0.0219, 0.194, -0.0, -0.0201, 0.1962, -0.0, -0.0219, 0.1962, -0.0, -0.0201, 0.194, 0.0017, -0.0219, 0.194, 0.0039, -0.0219, 0.1962, 0.0017, -0.0219, 0.1962, 0.0039, -0.0219,
0.1984, -0.0, -0.0219, 0.1984, -0.0, -0.0201, 0.2005, -0.0, -0.0219, 0.2005, -0.0, -0.0201, 0.1984, 0.0017, -0.0219, 0.1984, 0.0039, -0.0219, 0.2005, 0.0017, -0.0219, 0.2005, 0.0039, -0.0219, 0.202, -0.0, -0.0219, 0.202, -0.0, -0.0201, 0.202, 0.0017, -0.0219, 0.202, 0.0017, -0.0201, 0.202, 0.0039, -0.0219, 0.202, 0.0039, -0.0201, 0.194, -0.0, -0.0179, 0.194, -0.0, -0.0157, 0.1962, -0.0, -0.0179, 0.1962, -0.0, -0.0157, 0.194, -0.0, -0.0136, 0.194, -0.0, -0.0114, 0.1962, -0.0, -0.0136, 0.1962, -0.0, -0.0114, 0.1984, -0.0, -0.0179, 0.1984, -0.0, -0.0157, 0.2005, -0.0, -0.0179, 0.2005, -0.0, -0.0157, 0.1984, -0.0, -0.0136, 0.1984, -0.0, -0.0114, 0.2005, -0.0, -0.0136, 0.2005, -0.0, -0.0114,
0.194, -0.0, -0.0101, 0.1962, -0.0, -0.0101, 0.194, 0.0017, -0.0101, 0.194, 0.0039, -0.0101, 0.1962, 0.0017, -0.0101, 0.1962, 0.0039, -0.0101, 0.1984, -0.0, -0.0101, 0.2005, -0.0, -0.0101, 0.1984, 0.0017, -0.0101, 0.1984, 0.0039, -0.0101, 0.2005, 0.0017, -0.0101, 0.2005, 0.0039, -0.0101, 0.202, -0.0, -0.0179, 0.202, -0.0, -0.0157, 0.202, -0.0, -0.0136, 0.202, -0.0, -0.0114, 0.202, 0.0017, -0.0179, 0.202, 0.0017, -0.0157, 0.202, 0.0039, -0.0179, 0.202, 0.0039, -0.0157, 0.202, 0.0017, -0.0136, 0.202, 0.0017, -0.0114, 0.202, 0.0039, -0.0136, 0.202, 0.0039, -0.0114, 0.202, -0.0, -0.0101, 0.202, 0.0017, -0.0101, 0.202, 0.0039, -0.0101, 0.2118, -0.0, -0.067, 0.2135, -0.0, -0.067, 0.2118, -0.0, -0.0656,
0.2118, -0.0, -0.0635, 0.2135, -0.0, -0.0656, 0.2135, -0.0, -0.0635, 0.2118, 0.0017, -0.067, 0.2118, 0.0039, -0.067, 0.2135, 0.0017, -0.067, 0.2135, 0.0039, -0.067, 0.2118, 0.0017, -0.0656, 0.2118, 0.0017, -0.0635, 0.2118, 0.0039, -0.0656, 0.2118, 0.0039, -0.0635, 0.2157, -0.0, -0.067, 0.2179, -0.0, -0.067, 0.2157, -0.0, -0.0656, 0.2157, -0.0, -0.0631, 0.2179, -0.0, -0.0656, 0.2179, -0.0, -0.0631, 0.2157, 0.0017, -0.067, 0.2157, 0.0039, -0.067, 0.2179, 0.0017, -0.067, 0.2179, 0.0039, -0.067, 0.2157, 0.0017, -0.0631, 0.2157, 0.0039, -0.0631, 0.2179, 0.0017, -0.0631, 0.2179, 0.0039, -0.0631, 0.2118, -0.0, -0.0613, 0.2118, -0.0, -0.0591, 0.2135, -0.0, -0.0613, 0.2135, -0.0, -0.0591, 0.2118, -0.0, -0.057,
0.2118, -0.0, -0.0548, 0.2135, -0.0, -0.057, 0.2135, -0.0, -0.0548, 0.2118, 0.0017, -0.0613, 0.2118, 0.0017, -0.0591, 0.2118, 0.0039, -0.0613, 0.2118, 0.0039, -0.0591, 0.2118, 0.0017, -0.057, 0.2118, 0.0017, -0.0548, 0.2118, 0.0039, -0.057, 0.2118, 0.0039, -0.0548, 0.2157, -0.0, -0.0613, 0.2157, -0.0, -0.0591, 0.2157, -0.0, -0.057, 0.2157, -0.0, -0.0548, 0.2157, 0.0017, -0.0613, 0.2157, 0.0017, -0.0591, 0.2157, 0.0039, -0.0613, 0.2157, 0.0039, -0.0591, 0.2157, 0.0017, -0.057, 0.2157, 0.0017, -0.0548, 0.2157, 0.0039, -0.057, 0.2157, 0.0039, -0.0548, 0.2201, -0.0, -0.067, 0.2222, -0.0, -0.067, 0.2201, -0.0, -0.0656, 0.2201, -0.0, -0.0631, 0.2222, -0.0, -0.0656, 0.2222, -0.0, -0.0631, 0.2201, 0.0017, -0.067,
0.2201, 0.0039, -0.067, 0.2222, 0.0017, -0.067, 0.2222, 0.0039, -0.067, 0.2201, 0.0017, -0.0631, 0.2201, 0.0039, -0.0631, 0.2222, 0.0017, -0.0631, 0.2222, 0.0039, -0.0631, 0.2244, -0.0, -0.067, 0.2266, -0.0, -0.067, 0.2244, -0.0, -0.0656, 0.2244, -0.0, -0.0631, 0.2266, -0.0, -0.0656, 0.2266, -0.0, -0.0631, 0.2244, 0.0017, -0.067, 0.2244, 0.0039, -0.067, 0.2266, 0.0017, -0.067, 0.2266, 0.0039, -0.067, 0.2244, 0.0017, -0.0631, 0.2244, 0.0039, -0.0631, 0.2266, 0.0017, -0.0631, 0.2266, 0.0039, -0.0631, 0.2196, -0.0, -0.0562, 0.2196, -0.0, -0.0548, 0.2222, -0.0, -0.0562, 0.2222, -0.0, -0.0548, 0.2196, 0.0017, -0.0562, 0.2196, 0.0017, -0.0548, 0.2196, 0.0039, -0.0562, 0.2196, 0.0039, -0.0548, 0.2222, 0.0017, -0.0562,
0.2222, 0.0039, -0.0562, 0.2244, -0.0, -0.0562, 0.2244, -0.0, -0.0548, 0.2266, -0.0, -0.0562, 0.2266, -0.0, -0.0548, 0.2244, 0.0017, -0.0562, 0.2244, 0.0039, -0.0562, 0.2266, 0.0017, -0.0562, 0.2266, 0.0039, -0.0562, 0.2118, -0.0, -0.0526, 0.2118, -0.0, -0.0505, 0.2135, -0.0, -0.0526, 0.2135, -0.0, -0.0505, 0.2118, -0.0, -0.0483, 0.2118, -0.0, -0.0461, 0.2135, -0.0, -0.0483, 0.2135, -0.0, -0.0461, 0.2118, 0.0017, -0.0526, 0.2118, 0.0017, -0.0505, 0.2118, 0.0039, -0.0526, 0.2118, 0.0039, -0.0505, 0.2118, 0.0017, -0.0483, 0.2118, 0.0017, -0.0461, 0.2118, 0.0039, -0.0483, 0.2118, 0.0039, -0.0461, 0.2157, -0.0, -0.0526, 0.2157, -0.0, -0.0505, 0.2157, -0.0, -0.0483, 0.2157, -0.0, -0.0454, 0.2179, -0.0, -0.0454,
0.2157, 0.0017, -0.0526, 0.2157, 0.0017, -0.0505, 0.2157, 0.0039, -0.0526, 0.2157, 0.0039, -0.0505, 0.2157, 0.0017, -0.0483, 0.2157, 0.0017, -0.0454, 0.2157, 0.0039, -0.0483, 0.2157, 0.0039, -0.0454, 0.2179, 0.0017, -0.0454, 0.2179, 0.0039, -0.0454, 0.2118, -0.0, -0.0439, 0.2118, -0.0, -0.0425, 0.2135, -0.0, -0.0439, 0.2135, -0.0, -0.0425, 0.2118, 0.0017, -0.0439, 0.2118, 0.0017, -0.0425, 0.2118, 0.0039, -0.0439, 0.2118, 0.0039, -0.0425, 0.2135, 0.0017, -0.0425, 0.2135, 0.0039, -0.0425, 0.2157, -0.0, -0.0439, 0.2157, -0.0, -0.0425, 0.2179, -0.0, -0.0439, 0.2179, -0.0, -0.0425, 0.2157, 0.0017, -0.0425, 0.2157, 0.0039, -0.0425, 0.2179, 0.0017, -0.0425, 0.2179, 0.0039, -0.0425, 0.2196, -0.0, -0.0523, 0.2222, -0.0, -0.0523,
0.2201, -0.0, -0.0454, 0.2222, -0.0, -0.0454, 0.2196, 0.0017, -0.0523, 0.2196, 0.0039, -0.0523, 0.2222, 0.0017, -0.0523, 0.2222, 0.0039, -0.0523, 0.2201, 0.0017, -0.0454, 0.2201, 0.0039, -0.0454, 0.2222, 0.0017, -0.0454, 0.2222, 0.0039, -0.0454, 0.2244, -0.0, -0.0523, 0.2266, -0.0, -0.0523, 0.2244, -0.0, -0.0454, 0.2266, -0.0, -0.0454, 0.2244, 0.0017, -0.0523, 0.2244, 0.0039, -0.0523, 0.2266, 0.0017, -0.0523, 0.2266, 0.0039, -0.0523, 0.2244, 0.0017, -0.0454, 0.2244, 0.0039, -0.0454, 0.2266, 0.0017, -0.0454, 0.2266, 0.0039, -0.0454, 0.2201, -0.0, -0.0439, 0.2201, -0.0, -0.0425, 0.2222, -0.0, -0.0439, 0.2222, -0.0, -0.0425, 0.2201, 0.0017, -0.0425, 0.2201, 0.0039, -0.0425, 0.2222, 0.0017, -0.0425, 0.2222, 0.0039, -0.0425,
0.2244, -0.0, -0.0439, 0.2244, -0.0, -0.0425, 0.2266, -0.0, -0.0439, 0.2266, -0.0, -0.0425, 0.2244, 0.0017, -0.0425, 0.2244, 0.0039, -0.0425, 0.2266, 0.0017, -0.0425, 0.2266, 0.0039, -0.0425, 0.2287, -0.0, -0.067, 0.2304, -0.0, -0.0669, 0.2287, -0.0, -0.0656, 0.2287, -0.0, -0.0631, 0.2312, -0.0, -0.0668, 0.2308, -0.0, -0.0629, 0.2287, 0.0017, -0.067, 0.2287, 0.0039, -0.067, 0.2307, 0.0017, -0.0669, 0.2307, 0.0039, -0.0669, 0.2287, 0.0017, -0.063, 0.2287, 0.0039, -0.063, 0.2318, 0.0017, -0.0667, 0.2309, 0.0017, -0.0629, 0.2318, 0.0039, -0.0667, 0.2309, 0.0039, -0.0629, 0.2333, -0.0, -0.0665, 0.2328, -0.0, -0.0624, 0.2356, -0.0, -0.0657, 0.2353, -0.0, -0.0635, 0.2331, 0.0017, -0.0665, 0.2325, 0.0017, -0.0625,
0.2331, 0.0039, -0.0665, 0.2325, 0.0039, -0.0625, 0.2355, 0.0017, -0.0657, 0.2355, 0.0039, -0.0657, 0.2287, -0.0, -0.0563, 0.2287, -0.0, -0.0548, 0.2309, -0.0, -0.0565, 0.2309, -0.0, -0.0548, 0.2287, 0.0017, -0.0563, 0.2287, 0.0039, -0.0563, 0.2309, 0.0017, -0.0565, 0.2309, 0.0039, -0.0565, 0.2337, -0.0, -0.0621, 0.2349, -0.0, -0.0613, 0.2363, -0.0, -0.0591, 0.2329, -0.0, -0.0571, 0.2331, -0.0, -0.0548, 0.2346, -0.0, -0.058, 0.2353, -0.0, -0.0548, 0.2336, 0.0017, -0.0621, 0.2336, 0.0039, -0.0621, 0.2349, 0.0017, -0.0613, 0.2363, 0.0017, -0.0591, 0.2349, 0.0039, -0.0613, 0.2363, 0.0039, -0.0591, 0.2331, 0.0017, -0.0571, 0.2331, 0.0039, -0.0571, 0.2345, 0.0017, -0.0579, 0.2345, 0.0039, -0.0579, 0.2365, -0.0, -0.0648,
0.238, -0.0, -0.0634, 0.2366, 0.0017, -0.0648, 0.238, 0.0017, -0.0634, 0.2366, 0.0039, -0.0648, 0.238, 0.0039, -0.0634, 0.2386, -0.0, -0.0618, 0.2374, -0.0, -0.0591, 0.2389, -0.0, -0.0609, 0.239, -0.0, -0.0591, 0.2374, -0.0, -0.057, 0.2384, -0.0, -0.0547, 0.2387, -0.0, -0.057, 0.2384, -0.0, -0.0546, 0.2384, 0.0017, -0.0622, 0.2384, 0.0039, -0.0622, 0.2388, 0.0017, -0.0612, 0.239, 0.0017, -0.0591, 0.2388, 0.0039, -0.0612, 0.239, 0.0039, -0.0591, 0.2387, 0.0017, -0.057, 0.2384, 0.0017, -0.0544, 0.2387, 0.0039, -0.057, 0.2384, 0.0039, -0.0544, 0.2287, -0.0, -0.0522, 0.2309, -0.0, -0.0522, 0.2287, -0.0, -0.0455, 0.2308, -0.0, -0.0458, 0.2287, 0.0017, -0.0522, 0.2287, 0.0039, -0.0522, 0.2309, 0.0017, -0.0522,
0.2309, 0.0039, -0.0522, 0.2287, 0.0017, -0.0455, 0.2287, 0.0039, -0.0455, 0.2309, 0.0017, -0.0459, 0.2309, 0.0039, -0.0459, 0.2329, -0.0, -0.0518, 0.2338, -0.0, -0.0512, 0.2353, -0.0, -0.0526, 0.2345, -0.0, -0.0504, 0.2339, -0.0, -0.0475, 0.2329, -0.0, -0.0466, 0.2348, -0.0, -0.0484, 0.2353, -0.0, -0.0461, 0.2326, 0.0017, -0.0519, 0.2337, 0.0017, -0.0513, 0.2326, 0.0039, -0.0519, 0.2337, 0.0039, -0.0513, 0.2345, 0.0017, -0.0504, 0.2345, 0.0039, -0.0504, 0.2339, 0.0017, -0.0475, 0.2328, 0.0017, -0.0465, 0.2339, 0.0039, -0.0475, 0.2328, 0.0039, -0.0465, 0.2348, 0.0017, -0.0484, 0.2348, 0.0039, -0.0484, 0.2287, -0.0, -0.0439, 0.2287, -0.0, -0.0425, 0.2309, -0.0, -0.0439, 0.2309, -0.0, -0.0425, 0.2287, 0.0017, -0.0425,
0.2287, 0.0039, -0.0425, 0.2309, 0.0017, -0.0425, 0.2309, 0.0039, -0.0425, 0.2334, -0.0, -0.0429, 0.2324, -0.0, -0.0427, 0.2356, -0.0, -0.0438, 0.2337, 0.0017, -0.043, 0.2326, 0.0017, -0.0427, 0.2337, 0.0039, -0.043, 0.2326, 0.0039, -0.0427, 0.2353, 0.0017, -0.0437, 0.2353, 0.0039, -0.0437, 0.2374, -0.0, -0.0526, 0.2374, -0.0, -0.0505, 0.2386, -0.0, -0.0526, 0.239, -0.0, -0.0505, 0.2387, -0.0, -0.0475, 0.2377, -0.0, -0.0457, 0.239, -0.0, -0.0485, 0.2386, 0.0017, -0.0526, 0.239, 0.0017, -0.0505, 0.2386, 0.0039, -0.0526, 0.239, 0.0039, -0.0505, 0.2385, 0.0017, -0.0472, 0.2379, 0.0017, -0.046, 0.2385, 0.0039, -0.0472, 0.2379, 0.0039, -0.046, 0.2388, 0.0017, -0.0478, 0.2388, 0.0039, -0.0478, 0.2367, -0.0, -0.0447,
0.2368, 0.0017, -0.0447, 0.2368, 0.0039, -0.0447, 0.2187, -0.0, -0.0106, 0.2187, 0.0017, -0.0107, 0.2187, 0.0039, -0.0107, 0.2143, -0.0, -0.0022, 0.2144, 0.0017, -0.0024, 0.2144, 0.0039, -0.0024, 0.2162, -0.0, -0.0067, 0.2178, -0.0, -0.0096, 0.2168, -0.0, -0.0077, 0.2154, -0.0, -0.0052, 0.2146, -0.0, -0.0032, 0.2179, -0.0, -0.0049, 0.2179, -0.0, -0.0027, 0.2163, 0.0017, -0.0069, 0.2163, 0.0039, -0.0069, 0.2176, 0.0017, -0.0093, 0.2169, 0.0017, -0.008, 0.2176, 0.0039, -0.0093, 0.2169, 0.0039, -0.008, 0.2153, 0.0017, -0.0049, 0.2147, 0.0017, -0.0035, 0.2153, 0.0039, -0.0049, 0.2147, 0.0039, -0.0035, 0.2231, -0.0, -0.0149, 0.2207, -0.0, -0.0129, 0.2196, -0.0, -0.0118, 0.2219, -0.0, -0.014, 0.2222, -0.0, -0.0114,
0.223, 0.0017, -0.0149, 0.223, 0.0039, -0.0149, 0.2207, 0.0017, -0.0129, 0.2196, 0.0017, -0.0118, 0.2207, 0.0039, -0.0129, 0.2196, 0.0039, -0.0118, 0.2219, 0.0017, -0.014, 0.2219, 0.0039, -0.014, 0.2241, -0.0, -0.0156, 0.227, -0.0, -0.0173, 0.226, -0.0, -0.0168, 0.2244, -0.0, -0.0136, 0.2244, -0.0, -0.0114, 0.2266, -0.0, -0.0136, 0.2266, -0.0, -0.0114, 0.2244, 0.0017, -0.0158, 0.2244, 0.0039, -0.0158, 0.2269, 0.0017, -0.0172, 0.2258, 0.0017, -0.0167, 0.2269, 0.0039, -0.0172, 0.2258, 0.0039, -0.0167, 0.2201, -0.0, -0.0092, 0.2201, -0.0, -0.0071, 0.2222, -0.0, -0.0092, 0.2222, -0.0, -0.0071, 0.2201, -0.0, -0.0049, 0.2201, -0.0, -0.0027, 0.2222, -0.0, -0.0049, 0.2222, -0.0, -0.0027, 0.2244, -0.0, -0.0092,
0.2244, -0.0, -0.0071, 0.2266, -0.0, -0.0092, 0.2266, -0.0, -0.0071, 0.2244, -0.0, -0.0049, 0.2244, -0.0, -0.0027, 0.2266, -0.0, -0.0049, 0.2266, -0.0, -0.0027, 0.2318, -0.0, -0.0191, 0.2318, 0.0017, -0.0191, 0.2318, 0.0039, -0.0191, 0.233, -0.0, -0.0194, 0.2351, -0.0, -0.02, 0.2331, 0.0017, -0.0194, 0.2331, 0.0039, -0.0194, 0.2352, 0.0017, -0.02, 0.2352, 0.0039, -0.02, 0.2373, -0.0, -0.0205, 0.2395, -0.0, -0.0208, 0.2374, 0.0017, -0.0205, 0.2374, 0.0039, -0.0205, 0.2396, 0.0017, -0.0208, 0.2396, 0.0039, -0.0208, 0.2416, -0.0, -0.021, 0.2443, -0.0, -0.0214, 0.2436, -0.0, -0.0213, 0.2418, 0.0017, -0.021, 0.2418, 0.0039, -0.021, 0.244, 0.0017, -0.0213, 0.2429, 0.0017, -0.0212, 0.244, 0.0039, -0.0213,
0.2429, 0.0039, -0.0212, 0.2285, -0.0, -0.0179, 0.2287, -0.0, -0.0157, 0.2308, -0.0, -0.0187, 0.2309, -0.0, -0.0157, 0.2287, -0.0, -0.0136, 0.2287, -0.0, -0.0114, 0.2309, -0.0, -0.0136, 0.2309, -0.0, -0.0114, 0.2287, 0.0017, -0.018, 0.2287, 0.0039, -0.018, 0.2307, 0.0017, -0.0187, 0.2307, 0.0039, -0.0187, 0.2331, -0.0, -0.0179, 0.2331, -0.0, -0.0157, 0.2353, -0.0, -0.0179, 0.2353, -0.0, -0.0157, 0.2331, -0.0, -0.0136, 0.2331, -0.0, -0.0114, 0.2353, -0.0, -0.0136, 0.2353, -0.0, -0.0114, 0.2287, -0.0, -0.0092, 0.2287, -0.0, -0.0071, 0.2309, -0.0, -0.0092, 0.2309, -0.0, -0.0071, 0.2287, -0.0, -0.0049, 0.2289, -0.0, -0.0026, 0.231, -0.0, -0.0047, 0.23, -0.0, -0.0036, 0.2289, 0.0017, -0.0026,
0.2289, 0.0039, -0.0026, 0.231, 0.0017, -0.0047, 0.23, 0.0017, -0.0036, 0.231, 0.0039, -0.0047, 0.23, 0.0039, -0.0036, 0.2331, -0.0, -0.0092, 0.2334, -0.0, -0.0063, 0.2353, -0.0, -0.0092, 0.2354, -0.0, -0.0075, 0.2323, -0.0, -0.0057, 0.2335, 0.0017, -0.0064, 0.2335, 0.0039, -0.0064, 0.2352, 0.0017, -0.0074, 0.2352, 0.0039, -0.0074, 0.2324, 0.0017, -0.0057, 0.2324, 0.0039, -0.0057, 0.2374, -0.0, -0.0179, 0.2374, -0.0, -0.0157, 0.2396, -0.0, -0.0179, 0.2396, -0.0, -0.0157, 0.2374, -0.0, -0.0136, 0.2374, -0.0, -0.0114, 0.2396, -0.0, -0.0136, 0.2396, -0.0, -0.0114, 0.2418, -0.0, -0.0179, 0.2418, -0.0, -0.0157, 0.2439, -0.0, -0.0179, 0.2439, -0.0, -0.0157, 0.2418, -0.0, -0.0136, 0.2418, -0.0, -0.0114,
0.2439, -0.0, -0.0136, 0.2439, -0.0, -0.0114, 0.2377, -0.0, -0.0081, 0.2369, -0.0, -0.0079, 0.2398, -0.0, -0.0086, 0.2381, 0.0017, -0.0082, 0.2371, 0.0017, -0.0079, 0.2381, 0.0039, -0.0082, 0.2371, 0.0039, -0.0079, 0.2396, 0.0017, -0.0086, 0.2396, 0.0039, -0.0086, 0.2419, -0.0, -0.009, 0.244, -0.0, -0.0093, 0.2418, 0.0017, -0.009, 0.2418, 0.0039, -0.009, 0.2439, 0.0017, -0.0093, 0.2439, 0.0039, -0.0093, 0.2588, -0.0, -0.067, 0.2613, -0.0, -0.067, 0.2588, -0.0, -0.0656, 0.2588, -0.0, -0.0635, 0.2613, -0.0, -0.0656, 0.2618, -0.0, -0.0631, 0.2588, 0.0017, -0.067, 0.2588, 0.0039, -0.067, 0.2613, 0.0017, -0.067, 0.2613, 0.0039, -0.067, 0.2588, 0.0017, -0.0656, 0.2588, 0.0017, -0.0635, 0.2588, 0.0039, -0.0656,
0.2588, 0.0039, -0.0635, 0.2618, 0.0017, -0.0631, 0.2618, 0.0039, -0.0631, 0.2588, -0.0, -0.0613, 0.2588, -0.0, -0.0591, 0.2618, -0.0, -0.0613, 0.2618, -0.0, -0.0591, 0.2588, -0.0, -0.057, 0.2588, -0.0, -0.0548, 0.2618, -0.0, -0.057, 0.2618, -0.0, -0.0548, 0.2588, 0.0017, -0.0613, 0.2588, 0.0017, -0.0591, 0.2588, 0.0039, -0.0613, 0.2588, 0.0039, -0.0591, 0.2618, 0.0017, -0.0613, 0.2618, 0.0017, -0.0591, 0.2618, 0.0039, -0.0613, 0.2618, 0.0039, -0.0591, 0.2588, 0.0017, -0.057, 0.2588, 0.0017, -0.0548, 0.2588, 0.0039, -0.057, 0.2588, 0.0039, -0.0548, 0.2618, 0.0017, -0.057, 0.2618, 0.0017, -0.0548, 0.2618, 0.0039, -0.057, 0.2618, 0.0039, -0.0548, 0.2588, -0.0, -0.0526, 0.2588, -0.0, -0.0505, 0.2618, -0.0, -0.0526,
0.2618, -0.0, -0.0505, 0.2588, -0.0, -0.0483, 0.2588, -0.0, -0.0461, 0.2618, -0.0, -0.0483, 0.2618, -0.0, -0.0461, 0.2588, 0.0017, -0.0526, 0.2588, 0.0017, -0.0505, 0.2588, 0.0039, -0.0526, 0.2588, 0.0039, -0.0505, 0.2618, 0.0017, -0.0526, 0.2618, 0.0017, -0.0505, 0.2618, 0.0039, -0.0526, 0.2618, 0.0039, -0.0505, 0.2588, 0.0017, -0.0483, 0.2588, 0.0017, -0.0461, 0.2588, 0.0039, -0.0483, 0.2588, 0.0039, -0.0461, 0.2618, 0.0017, -0.0483, 0.2618, 0.0017, -0.0461, 0.2618, 0.0039, -0.0483, 0.2618, 0.0039, -0.0461, 0.2588, -0.0, -0.0439, 0.2588, -0.0, -0.0425, 0.2618, -0.0, -0.0439, 0.2618, -0.0, -0.0425, 0.2588, 0.0017, -0.0439, 0.2588, 0.0017, -0.0425, 0.2588, 0.0039, -0.0439, 0.2588, 0.0039, -0.0425, 0.2618, 0.0017, -0.0439,
0.2618, 0.0017, -0.0425, 0.2618, 0.0039, -0.0439, 0.2618, 0.0039, -0.0425, 0.2635, -0.0, -0.067, 0.2656, -0.0, -0.067, 0.2635, -0.0, -0.0656, 0.2635, -0.0, -0.0631, 0.2656, -0.0, -0.0656, 0.2656, -0.0, -0.0631, 0.2635, 0.0017, -0.067, 0.2635, 0.0039, -0.067, 0.2656, 0.0017, -0.067, 0.2656, 0.0039, -0.067, 0.2635, 0.0017, -0.0631, 0.2635, 0.0039, -0.0631, 0.2656, 0.0017, -0.0631, 0.2656, 0.0039, -0.0631, 0.2678, -0.0, -0.067, 0.27, -0.0, -0.067, 0.2678, -0.0, -0.0656, 0.2678, -0.0, -0.0631, 0.27, -0.0, -0.0656, 0.27, -0.0, -0.0631, 0.2678, 0.0017, -0.067, 0.2678, 0.0039, -0.067, 0.27, 0.0017, -0.067, 0.27, 0.0039, -0.067, 0.2678, 0.0017, -0.0631, 0.2678, 0.0039, -0.0631, 0.27, 0.0017, -0.0631,
0.27, 0.0039, -0.0631, 0.2721, -0.0, -0.067, 0.2743, -0.0, -0.067, 0.2721, -0.0, -0.0656, 0.2721, -0.0, -0.0631, 0.2743, -0.0, -0.0656, 0.2743, -0.0, -0.0631, 0.2721, 0.0017, -0.067, 0.2721, 0.0039, -0.067, 0.2743, 0.0017, -0.067, 0.2743, 0.0039, -0.067, 0.2721, 0.0017, -0.0631, 0.2721, 0.0039, -0.0631, 0.2743, 0.0017, -0.0631, 0.2743, 0.0039, -0.0631, 0.2765, -0.0, -0.067, 0.2794, -0.0, -0.067, 0.2765, -0.0, -0.0656, 0.2765, -0.0, -0.0631, 0.2794, -0.0, -0.0656, 0.2794, -0.0, -0.0631, 0.2765, 0.0017, -0.067, 0.2765, 0.0039, -0.067, 0.2794, 0.0017, -0.067, 0.2794, 0.0039, -0.067, 0.2765, 0.0017, -0.0631, 0.2765, 0.0039, -0.0631, 0.2794, 0.0017, -0.0656, 0.2794, 0.0017, -0.0631, 0.2794, 0.0039, -0.0656,
0.2794, 0.0039, -0.0631, 0.2461, -0.0, -0.0216, 0.2461, -0.0, -0.0201, 0.2482, -0.0, -0.0218, 0.2483, -0.0, -0.0201, 0.2461, 0.0017, -0.0216, 0.2461, 0.0039, -0.0216, 0.2483, 0.0017, -0.0218, 0.2483, 0.0039, -0.0218, 0.2504, -0.0, -0.0219, 0.2504, -0.0, -0.0201, 0.2526, -0.0, -0.0219, 0.2526, -0.0, -0.0201, 0.2504, 0.0017, -0.0219, 0.2504, 0.0039, -0.0219, 0.2526, 0.0017, -0.0219, 0.2526, 0.0039, -0.0219, 0.2548, -0.0, -0.0219, 0.2548, -0.0, -0.0201, 0.257, -0.0, -0.0217, 0.257, -0.0, -0.0201, 0.2548, 0.0017, -0.0219, 0.2548, 0.0039, -0.0219, 0.257, 0.0017, -0.0217, 0.257, 0.0039, -0.0217, 0.2591, -0.0, -0.0214, 0.2591, -0.0, -0.0201, 0.2605, -0.0, -0.0212, 0.2613, -0.0, -0.0211, 0.2591, 0.0017, -0.0214,
0.2591, 0.0039, -0.0214, 0.2606, 0.0017, -0.0212, 0.2617, 0.0017, -0.0211, 0.2606, 0.0039, -0.0212, 0.2617, 0.0039, -0.0211, 0.2461, -0.0, -0.0179, 0.2461, -0.0, -0.0157, 0.2483, -0.0, -0.0179, 0.2483, -0.0, -0.0157, 0.2461, -0.0, -0.0136, 0.2461, -0.0, -0.0114, 0.2483, -0.0, -0.0136, 0.2483, -0.0, -0.0114, 0.2504, -0.0, -0.0179, 0.2504, -0.0, -0.0157, 0.2526, -0.0, -0.0179, 0.2526, -0.0, -0.0157, 0.2504, -0.0, -0.0136, 0.2504, -0.0, -0.0114, 0.2526, -0.0, -0.0136, 0.2526, -0.0, -0.0114, 0.2462, -0.0, -0.0096, 0.2483, -0.0, -0.0099, 0.2461, 0.0017, -0.0096, 0.2461, 0.0039, -0.0096, 0.2483, 0.0017, -0.0099, 0.2483, 0.0039, -0.0099, 0.2504, -0.0, -0.0101, 0.2526, -0.0, -0.0101, 0.2504, 0.0017, -0.0101,
0.2504, 0.0039, -0.0101, 0.2526, 0.0017, -0.0101, 0.2526, 0.0039, -0.0101, 0.2548, -0.0, -0.0179, 0.2548, -0.0, -0.0157, 0.257, -0.0, -0.0179, 0.257, -0.0, -0.0157, 0.2548, -0.0, -0.0136, 0.2548, -0.0, -0.0114, 0.257, -0.0, -0.0136, 0.257, -0.0, -0.0114, 0.2591, -0.0, -0.0179, 0.2591, -0.0, -0.0157, 0.2613, -0.0, -0.0179, 0.2613, -0.0, -0.0157, 0.2591, -0.0, -0.0136, 0.2591, -0.0, -0.0114, 0.2613, -0.0, -0.0136, 0.2613, -0.0, -0.0114, 0.2548, -0.0, -0.0101, 0.2569, -0.0, -0.01, 0.2548, 0.0017, -0.0101, 0.2548, 0.0039, -0.0101, 0.257, 0.0017, -0.01, 0.257, 0.0039, -0.01, 0.2591, -0.0, -0.0097, 0.2612, -0.0, -0.0093, 0.2591, 0.0017, -0.0097, 0.2591, 0.0039, -0.0097, 0.2613, 0.0017, -0.0093,
0.2613, 0.0039, -0.0093, 0.2636, -0.0, -0.0209, 0.2658, -0.0, -0.0206, 0.2635, 0.0017, -0.0209, 0.2635, 0.0039, -0.0209, 0.2656, 0.0017, -0.0206, 0.2656, 0.0039, -0.0206, 0.2679, -0.0, -0.0202, 0.2701, -0.0, -0.0197, 0.2678, 0.0017, -0.0202, 0.2678, 0.0039, -0.0202, 0.27, 0.0017, -0.0197, 0.27, 0.0039, -0.0197, 0.2715, -0.0, -0.0192, 0.2717, 0.0017, -0.0192, 0.2717, 0.0039, -0.0192, 0.2635, -0.0, -0.0179, 0.2635, -0.0, -0.0157, 0.2656, -0.0, -0.0179, 0.2656, -0.0, -0.0157, 0.2635, -0.0, -0.0136, 0.2635, -0.0, -0.0114, 0.2656, -0.0, -0.0136, 0.2656, -0.0, -0.0114, 0.2678, -0.0, -0.0179, 0.2678, -0.0, -0.0157, 0.27, -0.0, -0.0179, 0.27, -0.0, -0.0157, 0.2678, -0.0, -0.0136, 0.2678, -0.0, -0.0114,
0.27, -0.0, -0.0136, 0.27, -0.0, -0.0114, 0.2633, -0.0, -0.009, 0.2656, -0.0, -0.0084, 0.2666, -0.0, -0.0081, 0.2635, 0.0017, -0.0089, 0.2635, 0.0039, -0.0089, 0.2655, 0.0017, -0.0085, 0.2666, 0.0017, -0.0081, 0.2655, 0.0039, -0.0085, 0.2666, 0.0039, -0.0081, 0.2678, -0.0, -0.0092, 0.2677, -0.0, -0.0077, 0.27, -0.0, -0.0092, 0.2699, -0.0, -0.0066, 0.271, -0.0, -0.0059, 0.2678, 0.0017, -0.0076, 0.2678, 0.0039, -0.0076, 0.27, 0.0017, -0.0066, 0.27, 0.0039, -0.0066, 0.271, 0.0017, -0.0059, 0.271, 0.0039, -0.0059, 0.2725, -0.0, -0.019, 0.2721, -0.0, -0.0157, 0.2746, -0.0, -0.0182, 0.2743, -0.0, -0.0157, 0.2721, -0.0, -0.0136, 0.2721, -0.0, -0.0114, 0.2743, -0.0, -0.0136, 0.2743, -0.0, -0.0114,
0.2728, 0.0017, -0.0189, 0.2728, 0.0039, -0.0189, 0.2743, 0.0017, -0.0183, 0.2743, 0.0039, -0.0183, 0.2766, -0.0, -0.0174, 0.2765, -0.0, -0.0157, 0.2776, -0.0, -0.0169, 0.2788, -0.0, -0.0163, 0.2765, -0.0, -0.0136, 0.2765, -0.0, -0.0114, 0.2787, -0.0, -0.0136, 0.2787, -0.0, -0.0114, 0.2765, 0.0017, -0.0175, 0.2765, 0.0039, -0.0175, 0.2776, 0.0017, -0.0169, 0.2788, 0.0017, -0.0163, 0.2776, 0.0039, -0.0169, 0.2788, 0.0039, -0.0163, 0.2721, -0.0, -0.0092, 0.2721, -0.0, -0.0071, 0.2743, -0.0, -0.0092, 0.2743, -0.0, -0.0071, 0.272, -0.0, -0.0053, 0.2735, -0.0, -0.004, 0.2746, -0.0, -0.0029, 0.2722, 0.0017, -0.0051, 0.2722, 0.0039, -0.0051, 0.2735, 0.0017, -0.004, 0.2746, 0.0017, -0.0029, 0.2735, 0.0039, -0.004,
0.2746, 0.0039, -0.0029, 0.2765, -0.0, -0.0092, 0.2765, -0.0, -0.0071, 0.2787, -0.0, -0.0092, 0.2787, -0.0, -0.0071, 0.2765, -0.0, -0.0049, 0.2754, -0.0, -0.0019, 0.2787, -0.0, -0.0049, 0.2787, -0.0, -0.0027, 0.2755, 0.0017, -0.0017, 0.2755, 0.0039, -0.0017, 0.298, -0.0, -0.067, 0.3004, -0.0, -0.067, 0.298, -0.0, -0.0656, 0.298, -0.0, -0.0635, 0.3004, -0.0, -0.0656, 0.3004, -0.0, -0.0635, 0.298, 0.0017, -0.067, 0.298, 0.0039, -0.067, 0.3004, 0.0017, -0.067, 0.3004, 0.0039, -0.067, 0.298, 0.0017, -0.0656, 0.298, 0.0017, -0.0635, 0.298, 0.0039, -0.0656, 0.298, 0.0039, -0.0635, 0.3025, -0.0, -0.067, 0.3047, -0.0, -0.067, 0.3025, -0.0, -0.0656, 0.302, -0.0, -0.0631, 0.3047, -0.0, -0.0656,
0.3047, -0.0, -0.0631, 0.3025, 0.0017, -0.067, 0.3025, 0.0039, -0.067, 0.3047, 0.0017, -0.067, 0.3047, 0.0039, -0.067, 0.302, 0.0017, -0.0631, 0.302, 0.0039, -0.0631, 0.3047, 0.0017, -0.0631, 0.3047, 0.0039, -0.0631, 0.298, -0.0, -0.0613, 0.298, -0.0, -0.0591, 0.3004, -0.0, -0.0613, 0.3004, -0.0, -0.0591, 0.298, -0.0, -0.057, 0.298, -0.0, -0.0548, 0.3004, -0.0, -0.057, 0.3004, -0.0, -0.0548, 0.298, 0.0017, -0.0613, 0.298, 0.0017, -0.0591, 0.298, 0.0039, -0.0613, 0.298, 0.0039, -0.0591, 0.298, 0.0017, -0.057, 0.298, 0.0017, -0.0548, 0.298, 0.0039, -0.057, 0.298, 0.0039, -0.0548, 0.302, -0.0, -0.0613, 0.302, -0.0, -0.0591, 0.302, -0.0, -0.057, 0.302, -0.0, -0.0548, 0.3049, -0.0, -0.0562,
0.3049, -0.0, -0.0548, 0.302, 0.0017, -0.0613, 0.302, 0.0017, -0.0591, 0.302, 0.0039, -0.0613, 0.302, 0.0039, -0.0591, 0.302, 0.0017, -0.057, 0.302, 0.0017, -0.0548, 0.302, 0.0039, -0.057, 0.302, 0.0039, -0.0548, 0.3049, 0.0017, -0.0562, 0.3049, 0.0017, -0.0548, 0.3049, 0.0039, -0.0562, 0.3049, 0.0039, -0.0548, 0.3069, -0.0, -0.067, 0.309, -0.0, -0.067, 0.3069, -0.0, -0.0656, 0.3069, -0.0, -0.0631, 0.309, -0.0, -0.0656, 0.309, -0.0, -0.0631, 0.3069, 0.0017, -0.067, 0.3069, 0.0039, -0.067, 0.309, 0.0017, -0.067, 0.309, 0.0039, -0.067, 0.3069, 0.0017, -0.0631, 0.3069, 0.0039, -0.0631, 0.309, 0.0017, -0.0631, 0.309, 0.0039, -0.0631, 0.3112, -0.0, -0.067, 0.3134, -0.0, -0.067, 0.3112, -0.0, -0.0656,
0.3112, -0.0, -0.0631, 0.3134, -0.0, -0.0656, 0.3134, -0.0, -0.0631, 0.3112, 0.0017, -0.067, 0.3112, 0.0039, -0.067, 0.3134, 0.0017, -0.067, 0.3134, 0.0039, -0.067, 0.3112, 0.0017, -0.0631, 0.3112, 0.0039, -0.0631, 0.3134, 0.0017, -0.0631, 0.3134, 0.0039, -0.0631, 0.3069, -0.0, -0.0562, 0.3069, -0.0, -0.0548, 0.309, -0.0, -0.0562, 0.309, -0.0, -0.0548, 0.3069, 0.0017, -0.0562, 0.3069, 0.0039, -0.0562, 0.309, 0.0017, -0.0562, 0.309, 0.0039, -0.0562, 0.3112, -0.0, -0.0562, 0.3112, -0.0, -0.0548, 0.3134, -0.0, -0.0562, 0.3134, -0.0, -0.0548, 0.3112, 0.0017, -0.0562, 0.3112, 0.0039, -0.0562, 0.3134, 0.0017, -0.0562, 0.3134, 0.0039, -0.0562, 0.298, -0.0, -0.0526, 0.298, -0.0, -0.0505, 0.3004, -0.0, -0.0526,
0.3004, -0.0, -0.0505, 0.298, -0.0, -0.0483, 0.298, -0.0, -0.0461, 0.3004, -0.0, -0.0483, 0.3004, -0.0, -0.0461, 0.298, 0.0017, -0.0526, 0.298, 0.0017, -0.0505, 0.298, 0.0039, -0.0526, 0.298, 0.0039, -0.0505, 0.298, 0.0017, -0.0483, 0.298, 0.0017, -0.0461, 0.298, 0.0039, -0.0483, 0.298, 0.0039, -0.0461, 0.302, -0.0, -0.0526, 0.302, -0.0, -0.0505, 0.3049, -0.0, -0.0523, 0.302, -0.0, -0.0483, 0.302, -0.0, -0.0454, 0.3047, -0.0, -0.0454, 0.302, 0.0017, -0.0526, 0.302, 0.0017, -0.0505, 0.302, 0.0039, -0.0526, 0.302, 0.0039, -0.0505, 0.3049, 0.0017, -0.0523, 0.3049, 0.0039, -0.0523, 0.302, 0.0017, -0.0483, 0.302, 0.0017, -0.0454, 0.302, 0.0039, -0.0483, 0.302, 0.0039, -0.0454, 0.3047, 0.0017, -0.0454,
0.3047, 0.0039, -0.0454, 0.298, -0.0, -0.0439, 0.298, -0.0, -0.0425, 0.3004, -0.0, -0.0439, 0.3004, -0.0, -0.0425, 0.298, 0.0017, -0.0439, 0.298, 0.0017, -0.0425, 0.298, 0.0039, -0.0439, 0.298, 0.0039, -0.0425, 0.3004, 0.0017, -0.0425, 0.3004, 0.0039, -0.0425, 0.3025, -0.0, -0.0439, 0.3025, -0.0, -0.0425, 0.3047, -0.0, -0.0439, 0.3047, -0.0, -0.0425, 0.3025, 0.0017, -0.0425, 0.3025, 0.0039, -0.0425, 0.3047, 0.0017, -0.0425, 0.3047, 0.0039, -0.0425, 0.3069, -0.0, -0.0523, 0.309, -0.0, -0.0523, 0.3069, -0.0, -0.0454, 0.309, -0.0, -0.0454, 0.3069, 0.0017, -0.0523, 0.3069, 0.0039, -0.0523, 0.309, 0.0017, -0.0523, 0.309, 0.0039, -0.0523, 0.3069, 0.0017, -0.0454, 0.3069, 0.0039, -0.0454, 0.309, 0.0017, -0.0454,
0.309, 0.0039, -0.0454, 0.3112, -0.0, -0.0523, 0.3134, -0.0, -0.0523, 0.3112, -0.0, -0.0454, 0.3134, -0.0, -0.0454, 0.3112, 0.0017, -0.0523, 0.3112, 0.0039, -0.0523, 0.3134, 0.0017, -0.0523, 0.3134, 0.0039, -0.0523, 0.3112, 0.0017, -0.0454, 0.3112, 0.0039, -0.0454, 0.3134, 0.0017, -0.0454, 0.3134, 0.0039, -0.0454, 0.3069, -0.0, -0.0439, 0.3069, -0.0, -0.0425, 0.309, -0.0, -0.0439, 0.309, -0.0, -0.0425, 0.3069, 0.0017, -0.0425, 0.3069, 0.0039, -0.0425, 0.309, 0.0017, -0.0425, 0.309, 0.0039, -0.0425, 0.3112, -0.0, -0.0439, 0.3112, -0.0, -0.0425, 0.3134, -0.0, -0.0439, 0.3134, -0.0, -0.0425, 0.3112, 0.0017, -0.0425, 0.3112, 0.0039, -0.0425, 0.3134, 0.0017, -0.0425, 0.3134, 0.0039, -0.0425, 0.2804, -0.0, -0.0152,
0.2815, -0.0, -0.0144, 0.2808, -0.0, -0.0114, 0.2827, -0.0, -0.0133, 0.2838, -0.0, -0.0122, 0.2805, 0.0017, -0.0152, 0.2805, 0.0039, -0.0152, 0.2815, 0.0017, -0.0143, 0.2815, 0.0039, -0.0143, 0.2828, 0.0017, -0.0133, 0.2838, 0.0017, -0.0122, 0.2828, 0.0039, -0.0133, 0.2838, 0.0039, -0.0122, 0.2848, -0.0, -0.0111, 0.2847, 0.0017, -0.0111, 0.2847, 0.0039, -0.0111, 0.2808, -0.0, -0.0092, 0.2808, -0.0, -0.0071, 0.283, -0.0, -0.0092, 0.283, -0.0, -0.0071, 0.2808, -0.0, -0.0049, 0.2808, -0.0, -0.0027, 0.283, -0.0, -0.0049, 0.283, -0.0, -0.0027, 0.2858, -0.0, -0.0096, 0.2852, -0.0, -0.0071, 0.2866, -0.0, -0.0085, 0.2873, -0.0, -0.0073, 0.2852, -0.0, -0.0049, 0.2852, -0.0, -0.0027, 0.2883, -0.0, -0.0052,
0.2873, -0.0, -0.0027, 0.2858, 0.0017, -0.0097, 0.2858, 0.0039, -0.0097, 0.2865, 0.0017, -0.0086, 0.2874, 0.0017, -0.0071, 0.2865, 0.0039, -0.0086, 0.2874, 0.0039, -0.0071, 0.2882, 0.0017, -0.0054, 0.2882, 0.0039, -0.0054, 0.2887, -0.0, -0.0042, 0.2891, -0.0, -0.0028, 0.2886, 0.0017, -0.0043, 0.2892, 0.0017, -0.0027, 0.2886, 0.0039, -0.0043, 0.2892, 0.0039, -0.0027, 0.3088, -0.0, -0.0219, 0.3088, -0.0, -0.0201, 0.3088, 0.0017, -0.0219, 0.3088, 0.0017, -0.0201, 0.3088, 0.0039, -0.0219, 0.3088, 0.0039, -0.0201, 0.3112, -0.0, -0.0219, 0.3112, -0.0, -0.0201, 0.3134, -0.0, -0.0219, 0.3134, -0.0, -0.0201, 0.3112, 0.0017, -0.0219, 0.3112, 0.0039, -0.0219, 0.3134, 0.0017, -0.0219, 0.3134, 0.0039, -0.0219, 0.3088, -0.0, -0.0179,
0.3088, -0.0, -0.0157, 0.3088, -0.0, -0.0136, 0.3088, -0.0, -0.0114, 0.3088, 0.0017, -0.0179, 0.3088, 0.0017, -0.0157, 0.3088, 0.0039, -0.0179, 0.3088, 0.0039, -0.0157, 0.3088, 0.0017, -0.0136, 0.3088, 0.0017, -0.0114, 0.3088, 0.0039, -0.0136, 0.3088, 0.0039, -0.0114, 0.3112, -0.0, -0.0179, 0.3112, -0.0, -0.0157, 0.3134, -0.0, -0.0179, 0.3134, -0.0, -0.0157, 0.3112, -0.0, -0.0136, 0.3112, -0.0, -0.0114, 0.3134, -0.0, -0.0136, 0.3134, -0.0, -0.0114, 0.3088, -0.0, -0.0092, 0.3088, -0.0, -0.0071, 0.3088, -0.0, -0.0049, 0.3088, -0.0, -0.0027, 0.3088, 0.0017, -0.0092, 0.3088, 0.0017, -0.0071, 0.3088, 0.0039, -0.0092, 0.3088, 0.0039, -0.0071, 0.3088, 0.0017, -0.0049, 0.3088, 0.0017, -0.0027, 0.3088, 0.0039, -0.0049,
0.3088, 0.0039, -0.0027, 0.3112, -0.0, -0.0092, 0.3112, -0.0, -0.0071, 0.3134, -0.0, -0.0092, 0.3134, -0.0, -0.0071, 0.3112, -0.0, -0.0049, 0.3112, -0.0, -0.0027, 0.3134, -0.0, -0.0049, 0.3134, -0.0, -0.0027, 0.3155, -0.0, -0.067, 0.3177, -0.0, -0.067, 0.3155, -0.0, -0.0656, 0.3155, -0.0, -0.0631, 0.3177, -0.0, -0.0656, 0.3177, -0.0, -0.0631, 0.3155, 0.0017, -0.067, 0.3155, 0.0039, -0.067, 0.3177, 0.0017, -0.067, 0.3177, 0.0039, -0.067, 0.3155, 0.0017, -0.0631, 0.3155, 0.0039, -0.0631, 0.3177, 0.0017, -0.0631, 0.3177, 0.0039, -0.0631, 0.3196, -0.0, -0.067, 0.3196, -0.0, -0.0656, 0.3196, -0.0, -0.0631, 0.3196, 0.0017, -0.067, 0.3196, 0.0039, -0.067, 0.3196, 0.0017, -0.0656, 0.3196, 0.0017, -0.0631,
0.3196, 0.0039, -0.0656, 0.3196, 0.0039, -0.0631, 0.3155, -0.0, -0.0562, 0.3155, -0.0, -0.0548, 0.3176, -0.0, -0.0562, 0.3176, -0.0, -0.0548, 0.3155, 0.0017, -0.0562, 0.3155, 0.0039, -0.0562, 0.3176, 0.0017, -0.0562, 0.3176, 0.0017, -0.0548, 0.3176, 0.0039, -0.0562, 0.3176, 0.0039, -0.0548, 0.3155, -0.0, -0.0523, 0.3176, -0.0, -0.0523, 0.3155, -0.0, -0.0454, 0.3177, -0.0, -0.0454, 0.3155, 0.0017, -0.0523, 0.3155, 0.0039, -0.0523, 0.3176, 0.0017, -0.0523, 0.3176, 0.0039, -0.0523, 0.3155, 0.0017, -0.0454, 0.3155, 0.0039, -0.0454, 0.3177, 0.0017, -0.0454, 0.3177, 0.0039, -0.0454, 0.3196, -0.0, -0.0454, 0.3196, 0.0017, -0.0454, 0.3196, 0.0039, -0.0454, 0.3155, -0.0, -0.0439, 0.3155, -0.0, -0.0425, 0.3177, -0.0, -0.0439,
0.3177, -0.0, -0.0425, 0.3155, 0.0017, -0.0425, 0.3155, 0.0039, -0.0425, 0.3177, 0.0017, -0.0425, 0.3177, 0.0039, -0.0425, 0.3196, -0.0, -0.0439, 0.3196, -0.0, -0.0425, 0.3196, 0.0017, -0.0439, 0.3196, 0.0017, -0.0425, 0.3196, 0.0039, -0.0439, 0.3196, 0.0039, -0.0425, 0.3402, -0.0, -0.067, 0.3402, -0.0, -0.0656, 0.3402, -0.0, -0.0635, 0.3402, 0.0017, -0.067, 0.3402, 0.0039, -0.067, 0.3402, 0.0017, -0.0656, 0.3402, 0.0017, -0.0635, 0.3402, 0.0039, -0.0656, 0.3402, 0.0039, -0.0635, 0.3402, -0.0, -0.0613, 0.3402, -0.0, -0.0591, 0.3402, -0.0, -0.057, 0.3402, -0.0, -0.0548, 0.3402, 0.0017, -0.0613, 0.3402, 0.0017, -0.0591, 0.3402, 0.0039, -0.0613, 0.3402, 0.0039, -0.0591, 0.3402, 0.0017, -0.057, 0.3402, 0.0017, -0.0548,
0.3402, 0.0039, -0.057, 0.3402, 0.0039, -0.0548, 0.3416, -0.0, -0.067, 0.3431, -0.0, -0.067, 0.3416, -0.0, -0.0656, 0.3416, -0.0, -0.0635, 0.3431, -0.0, -0.0656, 0.3431, -0.0, -0.0635, 0.3416, 0.0017, -0.067, 0.3416, 0.0039, -0.067, 0.3431, 0.0017, -0.067, 0.3431, 0.0039, -0.067, 0.3431, 0.0017, -0.0656, 0.3431, 0.0017, -0.0635, 0.3431, 0.0039, -0.0656, 0.3431, 0.0039, -0.0635, 0.3416, -0.0, -0.0613, 0.3416, -0.0, -0.0591, 0.3431, -0.0, -0.0613, 0.3431, -0.0, -0.0591, 0.3416, -0.0, -0.057, 0.3416, -0.0, -0.0548, 0.3431, -0.0, -0.057, 0.3431, -0.0, -0.0548, 0.3431, 0.0017, -0.0613, 0.3431, 0.0017, -0.0591, 0.3431, 0.0039, -0.0613, 0.3431, 0.0039, -0.0591, 0.3431, 0.0017, -0.057, 0.3431, 0.0017, -0.0548,
0.3431, 0.0039, -0.057, 0.3431, 0.0039, -0.0548, 0.3483, -0.0, -0.0546, 0.3483, 0.0017, -0.0545, 0.3483, 0.0039, -0.0545, 0.3402, -0.0, -0.0526, 0.3402, -0.0, -0.0505, 0.3402, -0.0, -0.0483, 0.3402, -0.0, -0.0461, 0.3402, 0.0017, -0.0526, 0.3402, 0.0017, -0.0505, 0.3402, 0.0039, -0.0526, 0.3402, 0.0039, -0.0505, 0.3402, 0.0017, -0.0483, 0.3402, 0.0017, -0.0461, 0.3402, 0.0039, -0.0483, 0.3402, 0.0039, -0.0461, 0.3402, -0.0, -0.0439, 0.3402, -0.0, -0.0425, 0.3402, 0.0017, -0.0439, 0.3402, 0.0017, -0.0425, 0.3402, 0.0039, -0.0439, 0.3402, 0.0039, -0.0425, 0.3416, -0.0, -0.0526, 0.3416, -0.0, -0.0505, 0.3431, -0.0, -0.0526, 0.3431, -0.0, -0.0499, 0.3416, -0.0, -0.0483, 0.3416, -0.0, -0.0461, 0.3438, -0.0, -0.0483,
0.3438, -0.0, -0.0461, 0.3431, 0.0017, -0.0526, 0.3431, 0.0017, -0.0499, 0.3431, 0.0039, -0.0526, 0.3431, 0.0039, -0.0499, 0.346, -0.0, -0.0525, 0.345, -0.0, -0.0515, 0.3472, -0.0, -0.0535, 0.3481, -0.0, -0.0505, 0.3459, -0.0, -0.0483, 0.3462, -0.0, -0.0457, 0.3485, -0.0, -0.0479, 0.3474, -0.0, -0.0468, 0.346, 0.0017, -0.0524, 0.3449, 0.0017, -0.0515, 0.346, 0.0039, -0.0524, 0.3449, 0.0039, -0.0515, 0.3472, 0.0017, -0.0535, 0.3472, 0.0039, -0.0535, 0.3462, 0.0017, -0.0457, 0.3462, 0.0039, -0.0457, 0.3485, 0.0017, -0.0479, 0.3474, 0.0017, -0.0468, 0.3485, 0.0039, -0.0479, 0.3474, 0.0039, -0.0468, 0.3421, -0.0, -0.0429, 0.3411, -0.0, -0.0426, 0.3441, -0.0, -0.0439, 0.3423, 0.0017, -0.043, 0.3413, 0.0017, -0.0426,
0.3423, 0.0039, -0.043, 0.3413, 0.0039, -0.0426, 0.3438, 0.0017, -0.0438, 0.3438, 0.0039, -0.0438, 0.3451, -0.0, -0.0448, 0.3452, 0.0017, -0.0448, 0.3452, 0.0039, -0.0448, 0.3155, -0.0, -0.0219, 0.3155, -0.0, -0.0201, 0.3177, -0.0, -0.0219, 0.3177, -0.0, -0.0201, 0.3155, 0.0017, -0.0219, 0.3155, 0.0039, -0.0219, 0.3177, 0.0017, -0.0219, 0.3177, 0.0039, -0.0219, 0.3206, -0.0, -0.0219, 0.3206, -0.0, -0.0201, 0.3206, 0.0017, -0.0219, 0.3206, 0.0017, -0.0201, 0.3206, 0.0039, -0.0219, 0.3206, 0.0039, -0.0201, 0.3155, -0.0, -0.0179, 0.3155, -0.0, -0.0157, 0.3177, -0.0, -0.0179, 0.3177, -0.0, -0.0157, 0.3155, -0.0, -0.0136, 0.3155, -0.0, -0.0114, 0.3177, -0.0, -0.0136, 0.3177, -0.0, -0.0114, 0.3206, -0.0, -0.0179,
0.3206, -0.0, -0.0157, 0.3206, -0.0, -0.0136, 0.3206, -0.0, -0.0114, 0.3206, 0.0017, -0.0179, 0.3206, 0.0017, -0.0157, 0.3206, 0.0039, -0.0179, 0.3206, 0.0039, -0.0157, 0.3206, 0.0017, -0.0136, 0.3206, 0.0017, -0.0114, 0.3206, 0.0039, -0.0136, 0.3206, 0.0039, -0.0114, 0.3155, -0.0, -0.0092, 0.3155, -0.0, -0.0071, 0.3177, -0.0, -0.0092, 0.3177, -0.0, -0.0071, 0.3155, -0.0, -0.0049, 0.3155, -0.0, -0.0027, 0.3177, -0.0, -0.0049, 0.3177, -0.0, -0.0027, 0.3206, -0.0, -0.0092, 0.3206, -0.0, -0.0071, 0.3206, -0.0, -0.0049, 0.3206, -0.0, -0.0027, 0.3206, 0.0017, -0.0092, 0.3206, 0.0017, -0.0071, 0.3206, 0.0039, -0.0092, 0.3206, 0.0039, -0.0071, 0.3206, 0.0017, -0.0049, 0.3206, 0.0017, -0.0027, 0.3206, 0.0039, -0.0049,
0.3206, 0.0039, -0.0027, 0.3402, -0.0, -0.0219, 0.3402, -0.0, -0.0201, 0.3402, 0.0017, -0.0219, 0.3402, 0.0017, -0.0201, 0.3402, 0.0039, -0.0219, 0.3402, 0.0039, -0.0201, 0.3416, -0.0, -0.0219, 0.3416, -0.0, -0.0201, 0.3438, -0.0, -0.0219, 0.3438, -0.0, -0.0201, 0.3416, 0.0017, -0.0219, 0.3416, 0.0039, -0.0219, 0.3438, 0.0017, -0.0219, 0.3438, 0.0039, -0.0219, 0.3459, -0.0, -0.0219, 0.3459, -0.0, -0.0201, 0.3481, -0.0, -0.0219, 0.3481, -0.0, -0.0201, 0.3459, 0.0017, -0.0219, 0.3459, 0.0039, -0.0219, 0.3481, 0.0017, -0.0219, 0.3481, 0.0039, -0.0219, 0.3402, -0.0, -0.0179, 0.3402, -0.0, -0.0157, 0.3402, -0.0, -0.0136, 0.3402, -0.0, -0.0114, 0.3402, 0.0017, -0.0179, 0.3402, 0.0017, -0.0157, 0.3402, 0.0039, -0.0179,
0.3402, 0.0039, -0.0157, 0.3402, 0.0017, -0.0136, 0.3402, 0.0017, -0.0114, 0.3402, 0.0039, -0.0136, 0.3402, 0.0039, -0.0114, 0.3402, -0.0, -0.0092, 0.3402, -0.0, -0.0071, 0.3402, -0.0, -0.0049, 0.3402, -0.0, -0.0027, 0.3402, 0.0017, -0.0092, 0.3402, 0.0017, -0.0071, 0.3402, 0.0039, -0.0092, 0.3402, 0.0039, -0.0071, 0.3402, 0.0017, -0.0049, 0.3402, 0.0017, -0.0027, 0.3402, 0.0039, -0.0049, 0.3402, 0.0039, -0.0027, 0.3416, -0.0, -0.0179, 0.3416, -0.0, -0.0157, 0.3438, -0.0, -0.0179, 0.3438, -0.0, -0.0157, 0.3416, -0.0, -0.0136, 0.3416, -0.0, -0.0114, 0.3438, -0.0, -0.0136, 0.3438, -0.0, -0.0114, 0.3459, -0.0, -0.0179, 0.3459, -0.0, -0.0157, 0.3481, -0.0, -0.0179, 0.3481, -0.0, -0.0157, 0.3459, -0.0, -0.0136,
0.3459, -0.0, -0.0114, 0.3481, -0.0, -0.0136, 0.3481, -0.0, -0.0114, 0.3416, -0.0, -0.0092, 0.3416, -0.0, -0.0071, 0.3438, -0.0, -0.0092, 0.3438, -0.0, -0.0071, 0.3416, -0.0, -0.0049, 0.3416, -0.0, -0.0027, 0.3438, -0.0, -0.0049, 0.3438, -0.0, -0.0027, 0.3459, -0.0, -0.0092, 0.3459, -0.0, -0.0071, 0.3481, -0.0, -0.0092, 0.3481, -0.0, -0.0071, 0.3459, -0.0, -0.0049, 0.3459, -0.0, -0.0027, 0.3481, -0.0, -0.0049, 0.3481, -0.0, -0.0027, 0.3571, -0.0, -0.0631, 0.357, 0.0017, -0.0631, 0.357, 0.0039, -0.0631, 0.3527, -0.0, -0.0589, 0.3505, -0.0, -0.0567, 0.3494, -0.0, -0.0556, 0.3516, -0.0, -0.0578, 0.3524, -0.0, -0.0548, 0.3527, 0.0017, -0.0589, 0.3527, 0.0039, -0.0589, 0.3505, 0.0017, -0.0567,
0.3494, 0.0017, -0.0556, 0.3505, 0.0039, -0.0567, 0.3494, 0.0039, -0.0556, 0.3516, 0.0017, -0.0578, 0.3516, 0.0039, -0.0578, 0.3549, -0.0, -0.061, 0.3538, -0.0, -0.06, 0.356, -0.0, -0.0621, 0.3568, -0.0, -0.0591, 0.3546, -0.0, -0.057, 0.355, -0.0, -0.0544, 0.3572, -0.0, -0.0565, 0.3561, -0.0, -0.0554, 0.3549, 0.0017, -0.061, 0.3538, 0.0017, -0.06, 0.3549, 0.0039, -0.061, 0.3538, 0.0039, -0.06, 0.356, 0.0017, -0.0621, 0.356, 0.0039, -0.0621, 0.355, 0.0017, -0.0544, 0.355, 0.0039, -0.0544, 0.3572, 0.0017, -0.0565, 0.3562, 0.0017, -0.0554, 0.3572, 0.0039, -0.0565, 0.3562, 0.0039, -0.0554, 0.3594, -0.0, -0.0651, 0.3583, -0.0, -0.0642, 0.3608, -0.0, -0.066, 0.3611, -0.0, -0.0635, 0.3594, 0.0017, -0.065,
0.3583, 0.0017, -0.0642, 0.3594, 0.0039, -0.065, 0.3583, 0.0039, -0.0642, 0.361, 0.0017, -0.0662, 0.361, 0.0039, -0.0662, 0.3637, -0.0, -0.067, 0.3637, -0.0, -0.0635, 0.3637, 0.0017, -0.067, 0.3637, 0.0017, -0.0635, 0.3637, 0.0039, -0.067, 0.3637, 0.0039, -0.0635, 0.3589, -0.0, -0.0613, 0.3595, -0.0, -0.0583, 0.3611, -0.0, -0.0613, 0.3608, -0.0, -0.0591, 0.3585, -0.0, -0.0575, 0.3608, -0.0, -0.057, 0.3608, -0.0, -0.0548, 0.3596, 0.0017, -0.0584, 0.3596, 0.0039, -0.0584, 0.3608, 0.0017, -0.0591, 0.3608, 0.0039, -0.0591, 0.3585, 0.0017, -0.0575, 0.3585, 0.0039, -0.0575, 0.3608, 0.0017, -0.057, 0.3608, 0.0017, -0.0548, 0.3608, 0.0039, -0.057, 0.3608, 0.0039, -0.0548, 0.3637, -0.0, -0.0613, 0.3637, -0.0, -0.0591,
0.3637, -0.0, -0.057, 0.3637, -0.0, -0.0548, 0.3637, 0.0017, -0.0613, 0.3637, 0.0017, -0.0591, 0.3637, 0.0039, -0.0613, 0.3637, 0.0039, -0.0591, 0.3637, 0.0017, -0.057, 0.3637, 0.0017, -0.0548, 0.3637, 0.0039, -0.057, 0.3637, 0.0039, -0.0548, 0.3503, -0.0, -0.0526, 0.3507, -0.0, -0.05, 0.3529, -0.0, -0.0522, 0.3518, -0.0, -0.0511, 0.3496, -0.0, -0.0489, 0.3507, 0.0017, -0.05, 0.3507, 0.0039, -0.05, 0.3529, 0.0017, -0.0522, 0.3518, 0.0017, -0.0511, 0.3529, 0.0039, -0.0522, 0.3518, 0.0039, -0.0511, 0.3496, 0.0017, -0.049, 0.3496, 0.0039, -0.049, 0.354, -0.0, -0.0533, 0.354, 0.0017, -0.0533, 0.354, 0.0039, -0.0533, 0.3608, -0.0, -0.0526, 0.3608, -0.0, -0.0505, 0.3608, -0.0, -0.0483, 0.3608, -0.0, -0.0461,
0.3608, 0.0017, -0.0526, 0.3608, 0.0017, -0.0505, 0.3608, 0.0039, -0.0526, 0.3608, 0.0039, -0.0505, 0.3608, 0.0017, -0.0483, 0.3608, 0.0017, -0.0461, 0.3608, 0.0039, -0.0483, 0.3608, 0.0039, -0.0461, 0.3637, -0.0, -0.0526, 0.3637, -0.0, -0.0505, 0.3637, -0.0, -0.0483, 0.3637, -0.0, -0.0461, 0.3637, 0.0017, -0.0526, 0.3637, 0.0017, -0.0505, 0.3637, 0.0039, -0.0526, 0.3637, 0.0039, -0.0505, 0.3637, 0.0017, -0.0483, 0.3637, 0.0017, -0.0461, 0.3637, 0.0039, -0.0483, 0.3637, 0.0039, -0.0461, 0.3608, -0.0, -0.0439, 0.3608, -0.0, -0.0425, 0.3608, 0.0017, -0.0439, 0.3608, 0.0017, -0.0425, 0.3608, 0.0039, -0.0439, 0.3608, 0.0039, -0.0425, 0.3637, -0.0, -0.0439, 0.3637, -0.0, -0.0425, 0.3637, 0.0017, -0.0439, 0.3637, 0.0017, -0.0425,
0.3637, 0.0039, -0.0439, 0.3637, 0.0039, -0.0425, 0.3503, -0.0, -0.0219, 0.3503, -0.0, -0.0201, 0.3524, -0.0, -0.0219, 0.3524, -0.0, -0.0201, 0.3503, 0.0017, -0.0219, 0.3503, 0.0039, -0.0219, 0.3524, 0.0017, -0.0219, 0.3524, 0.0039, -0.0219, 0.3546, -0.0, -0.0219, 0.3546, -0.0, -0.0201, 0.3568, -0.0, -0.0219, 0.3568, -0.0, -0.0201, 0.3546, 0.0017, -0.0219, 0.3546, 0.0039, -0.0219, 0.3568, 0.0017, -0.0219, 0.3568, 0.0039, -0.0219, 0.3589, -0.0, -0.0219, 0.3589, -0.0, -0.0201, 0.3611, -0.0, -0.0219, 0.3611, -0.0, -0.0201, 0.3589, 0.0017, -0.0219, 0.3589, 0.0039, -0.0219, 0.3611, 0.0017, -0.0219, 0.3611, 0.0039, -0.0219, 0.3633, -0.0, -0.0219, 0.3633, -0.0, -0.0201, 0.3655, -0.0, -0.0219, 0.3655, -0.0, -0.0201,
0.3633, 0.0017, -0.0219, 0.3633, 0.0039, -0.0219, 0.3655, 0.0017, -0.0219, 0.3655, 0.0039, -0.0219, 0.3503, -0.0, -0.0179, 0.3503, -0.0, -0.0157, 0.3524, -0.0, -0.0179, 0.3524, -0.0, -0.0157, 0.3503, -0.0, -0.0136, 0.3503, -0.0, -0.0114, 0.3524, -0.0, -0.0136, 0.3524, -0.0, -0.0114, 0.3546, -0.0, -0.0179, 0.3546, -0.0, -0.0157, 0.3568, -0.0, -0.0179, 0.3568, -0.0, -0.0157, 0.3546, -0.0, -0.0136, 0.3546, -0.0, -0.0114, 0.3568, -0.0, -0.0136, 0.3568, -0.0, -0.0114, 0.3503, -0.0, -0.0092, 0.3503, -0.0, -0.0071, 0.352, -0.0, -0.0101, 0.352, -0.0, -0.0071, 0.3503, -0.0, -0.0049, 0.3503, -0.0, -0.0027, 0.352, -0.0, -0.0049, 0.352, -0.0, -0.0027, 0.352, 0.0017, -0.0101, 0.352, 0.0017, -0.0071,
0.352, 0.0039, -0.0101, 0.352, 0.0039, -0.0071, 0.352, 0.0017, -0.0049, 0.352, 0.0017, -0.0027, 0.352, 0.0039, -0.0049, 0.352, 0.0039, -0.0027, 0.3546, -0.0, -0.0101, 0.3568, -0.0, -0.0101, 0.3546, 0.0017, -0.0101, 0.3546, 0.0039, -0.0101, 0.3568, 0.0017, -0.0101, 0.3568, 0.0039, -0.0101, 0.3589, -0.0, -0.0179, 0.3589, -0.0, -0.0157, 0.3611, -0.0, -0.0179, 0.3611, -0.0, -0.0157, 0.3589, -0.0, -0.0136, 0.3589, -0.0, -0.0114, 0.3611, -0.0, -0.0136, 0.3611, -0.0, -0.0114, 0.3633, -0.0, -0.0179, 0.3633, -0.0, -0.0157, 0.3655, -0.0, -0.0179, 0.3655, -0.0, -0.0157, 0.3633, -0.0, -0.0136, 0.3633, -0.0, -0.0114, 0.3655, -0.0, -0.0136, 0.3655, -0.0, -0.0114, 0.3589, -0.0, -0.0101, 0.3611, -0.0, -0.0101,
0.3589, 0.0017, -0.0101, 0.3589, 0.0039, -0.0101, 0.3611, 0.0017, -0.0101, 0.3611, 0.0039, -0.0101, 0.3633, -0.0, -0.0101, 0.3655, -0.0, -0.0101, 0.3633, 0.0017, -0.0101, 0.3633, 0.0039, -0.0101, 0.3655, 0.0017, -0.0101, 0.3655, 0.0039, -0.0101, 0.3676, -0.0, -0.0219, 0.3676, -0.0, -0.0201, 0.3698, -0.0, -0.0219, 0.3698, -0.0, -0.0201, 0.3676, 0.0017, -0.0219, 0.3676, 0.0039, -0.0219, 0.3698, 0.0017, -0.0219, 0.3698, 0.0039, -0.0219, 0.372, -0.0, -0.0219, 0.372, -0.0, -0.0201, 0.3741, -0.0, -0.0219, 0.3741, -0.0, -0.0201, 0.372, 0.0017, -0.0219, 0.372, 0.0039, -0.0219, 0.3741, 0.0017, -0.0219, 0.3741, 0.0039, -0.0219, 0.3763, -0.0, -0.0219, 0.3763, -0.0, -0.0201, 0.3785, -0.0, -0.0219, 0.3785, -0.0, -0.0201,
0.3763, 0.0017, -0.0219, 0.3763, 0.0039, -0.0219, 0.3785, 0.0017, -0.0219, 0.3785, 0.0039, -0.0219, 0.3807, -0.0, -0.0218, 0.3806, -0.0, -0.0201, 0.3828, -0.0, -0.0215, 0.3828, -0.0, -0.0201, 0.3807, 0.0017, -0.0218, 0.3807, 0.0039, -0.0218, 0.3828, 0.0017, -0.0215, 0.3828, 0.0039, -0.0215, 0.3676, -0.0, -0.0179, 0.3676, -0.0, -0.0157, 0.3698, -0.0, -0.0179, 0.3698, -0.0, -0.0157, 0.3676, -0.0, -0.0136, 0.3676, -0.0, -0.0114, 0.3698, -0.0, -0.0136, 0.3698, -0.0, -0.0114, 0.372, -0.0, -0.0179, 0.372, -0.0, -0.0157, 0.3741, -0.0, -0.0179, 0.3741, -0.0, -0.0157, 0.372, -0.0, -0.0136, 0.372, -0.0, -0.0114, 0.3741, -0.0, -0.0136, 0.3741, -0.0, -0.0114, 0.3676, -0.0, -0.0101, 0.3698, -0.0, -0.0101,
0.3676, 0.0017, -0.0101, 0.3676, 0.0039, -0.0101, 0.3698, 0.0017, -0.0101, 0.3698, 0.0039, -0.0101, 0.372, -0.0, -0.0101, 0.3741, -0.0, -0.0101, 0.372, 0.0017, -0.0101, 0.372, 0.0039, -0.0101, 0.3741, 0.0017, -0.0101, 0.3741, 0.0039, -0.0101, 0.3763, -0.0, -0.0179, 0.3763, -0.0, -0.0157, 0.3785, -0.0, -0.0179, 0.3785, -0.0, -0.0157, 0.3763, -0.0, -0.0136, 0.3763, -0.0, -0.0114, 0.3785, -0.0, -0.0136, 0.3785, -0.0, -0.0114, 0.3806, -0.0, -0.0179, 0.3806, -0.0, -0.0157, 0.3828, -0.0, -0.0179, 0.3828, -0.0, -0.0157, 0.3806, -0.0, -0.0136, 0.3806, -0.0, -0.0114, 0.3828, -0.0, -0.0136, 0.3828, -0.0, -0.0114, 0.3763, -0.0, -0.0101, 0.3785, -0.0, -0.0101, 0.3763, 0.0017, -0.0101, 0.3763, 0.0039, -0.0101,
0.3785, 0.0017, -0.0101, 0.3785, 0.0039, -0.0101, 0.3806, -0.0, -0.0101, 0.3828, -0.0, -0.0099, 0.3807, 0.0017, -0.0101, 0.3807, 0.0039, -0.0101, 0.3828, 0.0017, -0.0099, 0.3828, 0.0039, -0.0099, 0.3863, -0.0, -0.067, 0.3863, -0.0, -0.0656, 0.3863, -0.0, -0.0635, 0.3863, 0.0017, -0.067, 0.3863, 0.0039, -0.067, 0.3863, 0.0017, -0.0656, 0.3863, 0.0017, -0.0635, 0.3863, 0.0039, -0.0656, 0.3863, 0.0039, -0.0635, 0.3893, -0.0, -0.067, 0.3915, -0.0, -0.067, 0.3893, -0.0, -0.0656, 0.3902, -0.0, -0.0631, 0.3915, -0.0, -0.0656, 0.3915, -0.0, -0.0631, 0.3893, 0.0017, -0.067, 0.3893, 0.0039, -0.067, 0.3915, 0.0017, -0.067, 0.3915, 0.0039, -0.067, 0.3902, 0.0017, -0.0631, 0.3902, 0.0039, -0.0631, 0.3915, 0.0017, -0.0631,
0.3915, 0.0039, -0.0631, 0.3863, -0.0, -0.0613, 0.3863, -0.0, -0.0591, 0.3863, -0.0, -0.057, 0.3863, -0.0, -0.0548, 0.3863, 0.0017, -0.0613, 0.3863, 0.0017, -0.0591, 0.3863, 0.0039, -0.0613, 0.3863, 0.0039, -0.0591, 0.3863, 0.0017, -0.057, 0.3863, 0.0017, -0.0548, 0.3863, 0.0039, -0.057, 0.3863, 0.0039, -0.0548, 0.3902, -0.0, -0.0613, 0.3902, -0.0, -0.0591, 0.3902, -0.0, -0.057, 0.3902, -0.0, -0.0548, 0.3902, 0.0017, -0.0613, 0.3902, 0.0017, -0.0591, 0.3902, 0.0039, -0.0613, 0.3902, 0.0039, -0.0591, 0.3902, 0.0017, -0.057, 0.3902, 0.0017, -0.0548, 0.3902, 0.0039, -0.057, 0.3902, 0.0039, -0.0548, 0.3937, -0.0, -0.067, 0.3958, -0.0, -0.067, 0.3937, -0.0, -0.0656, 0.3937, -0.0, -0.0631, 0.3958, -0.0, -0.0656,
0.3958, -0.0, -0.0631, 0.3937, 0.0017, -0.067, 0.3937, 0.0039, -0.067, 0.3958, 0.0017, -0.067, 0.3958, 0.0039, -0.067, 0.3937, 0.0017, -0.0631, 0.3937, 0.0039, -0.0631, 0.3958, 0.0017, -0.0631, 0.3958, 0.0039, -0.0631, 0.398, -0.0, -0.067, 0.4002, -0.0, -0.0669, 0.398, -0.0, -0.0656, 0.398, -0.0, -0.063, 0.4002, -0.0, -0.0656, 0.4001, -0.0, -0.063, 0.398, 0.0017, -0.067, 0.398, 0.0039, -0.067, 0.4002, 0.0017, -0.0669, 0.4002, 0.0039, -0.0669, 0.398, 0.0017, -0.063, 0.398, 0.0039, -0.063, 0.4002, 0.0017, -0.063, 0.4002, 0.0039, -0.063, 0.3863, -0.0, -0.0526, 0.3863, -0.0, -0.0505, 0.3863, -0.0, -0.0483, 0.3863, -0.0, -0.0461, 0.3863, 0.0017, -0.0526, 0.3863, 0.0017, -0.0505, 0.3863, 0.0039, -0.0526,
0.3863, 0.0039, -0.0505, 0.3863, 0.0017, -0.0483, 0.3863, 0.0017, -0.0461, 0.3863, 0.0039, -0.0483, 0.3863, 0.0039, -0.0461, 0.3902, -0.0, -0.0526, 0.3902, -0.0, -0.0505, 0.3902, -0.0, -0.0483, 0.3902, -0.0, -0.0454, 0.3915, -0.0, -0.0454, 0.3902, 0.0017, -0.0526, 0.3902, 0.0017, -0.0505, 0.3902, 0.0039, -0.0526, 0.3902, 0.0039, -0.0505, 0.3902, 0.0017, -0.0483, 0.3902, 0.0017, -0.0454, 0.3902, 0.0039, -0.0483, 0.3902, 0.0039, -0.0454, 0.3915, 0.0017, -0.0454, 0.3915, 0.0039, -0.0454, 0.3863, -0.0, -0.0439, 0.3863, -0.0, -0.0425, 0.3863, 0.0017, -0.0439, 0.3863, 0.0017, -0.0425, 0.3863, 0.0039, -0.0439, 0.3863, 0.0039, -0.0425, 0.3893, -0.0, -0.0439, 0.3893, -0.0, -0.0425, 0.3915, -0.0, -0.0439, 0.3915, -0.0, -0.0425,
0.3893, 0.0017, -0.0425, 0.3893, 0.0039, -0.0425, 0.3915, 0.0017, -0.0425, 0.3915, 0.0039, -0.0425, 0.3937, -0.0, -0.0454, 0.3958, -0.0, -0.0454, 0.3937, 0.0017, -0.0454, 0.3937, 0.0039, -0.0454, 0.3958, 0.0017, -0.0454, 0.3958, 0.0039, -0.0454, 0.398, -0.0, -0.0455, 0.4001, -0.0, -0.0457, 0.398, 0.0017, -0.0455, 0.398, 0.0039, -0.0455, 0.4002, 0.0017, -0.0458, 0.4002, 0.0039, -0.0458, 0.3937, -0.0, -0.0439, 0.3937, -0.0, -0.0425, 0.3958, -0.0, -0.0439, 0.3958, -0.0, -0.0425, 0.3937, 0.0017, -0.0425, 0.3937, 0.0039, -0.0425, 0.3958, 0.0017, -0.0425, 0.3958, 0.0039, -0.0425, 0.398, -0.0, -0.0439, 0.398, -0.0, -0.0425, 0.4002, -0.0, -0.0439, 0.4002, -0.0, -0.0425, 0.398, 0.0017, -0.0425, 0.398, 0.0039, -0.0425,
0.4002, 0.0017, -0.0425, 0.4002, 0.0039, -0.0425, 0.4016, -0.0, -0.0668, 0.4025, -0.0, -0.0667, 0.4023, -0.0, -0.0627, 0.4048, -0.0, -0.0661, 0.4045, -0.0, -0.0635, 0.4017, 0.0017, -0.0668, 0.4017, 0.0039, -0.0668, 0.4028, 0.0017, -0.0666, 0.4023, 0.0017, -0.0627, 0.4028, 0.0039, -0.0666, 0.4023, 0.0039, -0.0627, 0.4046, 0.0017, -0.0662, 0.4046, 0.0039, -0.0662, 0.4064, -0.0, -0.0653, 0.4075, -0.0, -0.0645, 0.4089, -0.0, -0.0635, 0.4066, 0.0017, -0.0652, 0.4076, 0.0017, -0.0644, 0.4066, 0.0039, -0.0652, 0.4076, 0.0039, -0.0644, 0.409, 0.0017, -0.0635, 0.409, 0.0039, -0.0635, 0.4033, -0.0, -0.0623, 0.4045, -0.0, -0.0619, 0.4033, 0.0017, -0.0623, 0.4033, 0.0039, -0.0623, 0.4046, 0.0017, -0.0619, 0.4046, 0.0039, -0.0619,
0.4062, -0.0, -0.0608, 0.4073, -0.0, -0.0596, 0.41, -0.0, -0.0623, 0.4079, -0.0, -0.0587, 0.4085, -0.0, -0.0569, 0.4087, -0.0, -0.0548, 0.4062, 0.0017, -0.0608, 0.4074, 0.0017, -0.0595, 0.4062, 0.0039, -0.0608, 0.4074, 0.0039, -0.0595, 0.4099, 0.0017, -0.0624, 0.408, 0.0017, -0.0584, 0.4099, 0.0039, -0.0624, 0.408, 0.0039, -0.0584, 0.4085, 0.0017, -0.057, 0.4087, 0.0017, -0.0548, 0.4085, 0.0039, -0.057, 0.4087, 0.0039, -0.0548, 0.4108, -0.0, -0.0612, 0.4119, -0.0, -0.0593, 0.4122, -0.0, -0.0582, 0.411, -0.0, -0.057, 0.411, -0.0, -0.0548, 0.4125, -0.0, -0.057, 0.4127, -0.0, -0.0548, 0.4107, 0.0017, -0.0613, 0.4119, 0.0017, -0.0594, 0.4107, 0.0039, -0.0613, 0.4119, 0.0039, -0.0594, 0.4122, 0.0017, -0.0582,
0.4122, 0.0039, -0.0582, 0.4125, 0.0017, -0.057, 0.4127, 0.0017, -0.0548, 0.4125, 0.0039, -0.057, 0.4127, 0.0039, -0.0548, 0.4022, -0.0, -0.0462, 0.4051, -0.0, -0.0476, 0.4041, -0.0, -0.0471, 0.4024, 0.0017, -0.0463, 0.4024, 0.0039, -0.0463, 0.405, 0.0017, -0.0476, 0.4039, 0.0017, -0.047, 0.405, 0.0039, -0.0476, 0.4039, 0.0039, -0.047, 0.4074, -0.0, -0.0498, 0.4086, -0.0, -0.0528, 0.408, -0.0, -0.0508, 0.4064, -0.0, -0.0486, 0.4067, -0.0, -0.0461, 0.4089, -0.0, -0.0483, 0.4093, -0.0, -0.0456, 0.4075, 0.0017, -0.0498, 0.4075, 0.0039, -0.0498, 0.4086, 0.0017, -0.0526, 0.4081, 0.0017, -0.0509, 0.4086, 0.0039, -0.0526, 0.4081, 0.0039, -0.0509, 0.4064, 0.0017, -0.0486, 0.4064, 0.0039, -0.0486, 0.4093, 0.0017, -0.0456,
0.4093, 0.0039, -0.0456, 0.4024, -0.0, -0.0439, 0.4024, -0.0, -0.0426, 0.4047, -0.0, -0.043, 0.4036, -0.0, -0.0428, 0.4024, 0.0017, -0.0426, 0.4024, 0.0039, -0.0426, 0.4048, 0.0017, -0.0431, 0.4037, 0.0017, -0.0428, 0.4048, 0.0039, -0.0431, 0.4037, 0.0039, -0.0428, 0.407, -0.0, -0.044, 0.4081, -0.0, -0.0447, 0.4067, 0.0017, -0.0439, 0.4067, 0.0039, -0.0439, 0.4082, 0.0017, -0.0448, 0.4082, 0.0039, -0.0448, 0.411, -0.0, -0.0526, 0.4121, -0.0, -0.0502, 0.4126, -0.0, -0.0526, 0.4123, -0.0, -0.0511, 0.4111, -0.0, -0.048, 0.4103, -0.0, -0.0468, 0.412, 0.0017, -0.0499, 0.412, 0.0039, -0.0499, 0.4126, 0.0017, -0.0526, 0.4122, 0.0017, -0.051, 0.4126, 0.0039, -0.0526, 0.4122, 0.0039, -0.051, 0.4112, 0.0017, -0.0482,
0.4103, 0.0017, -0.0468, 0.4112, 0.0039, -0.0482, 0.4103, 0.0039, -0.0468, 0.3844, -0.0, -0.0213, 0.3852, -0.0, -0.0212, 0.3873, -0.0, -0.0209, 0.3846, 0.0017, -0.0213, 0.3857, 0.0017, -0.0211, 0.3846, 0.0039, -0.0213, 0.3857, 0.0039, -0.0211, 0.3872, 0.0017, -0.0209, 0.3872, 0.0039, -0.0209, 0.3895, -0.0, -0.0206, 0.3917, -0.0, -0.02, 0.3893, 0.0017, -0.0206, 0.3893, 0.0039, -0.0206, 0.3915, 0.0017, -0.0201, 0.3915, 0.0039, -0.0201, 0.3933, -0.0, -0.0194, 0.3935, 0.0017, -0.0194, 0.3935, 0.0039, -0.0194, 0.385, -0.0, -0.0179, 0.385, -0.0, -0.0157, 0.3872, -0.0, -0.0179, 0.3872, -0.0, -0.0157, 0.385, -0.0, -0.0136, 0.385, -0.0, -0.0114, 0.3872, -0.0, -0.0136, 0.3872, -0.0, -0.0114, 0.3893, -0.0, -0.0179,
0.3893, -0.0, -0.0157, 0.3915, -0.0, -0.0179, 0.3915, -0.0, -0.0157, 0.3893, -0.0, -0.0136, 0.3893, -0.0, -0.0114, 0.3915, -0.0, -0.0136, 0.3915, -0.0, -0.0114, 0.3849, -0.0, -0.0096, 0.387, -0.0, -0.0091, 0.385, 0.0017, -0.0096, 0.385, 0.0039, -0.0096, 0.3872, 0.0017, -0.009, 0.3872, 0.0039, -0.009, 0.3892, -0.0, -0.0084, 0.3902, -0.0, -0.008, 0.3915, -0.0, -0.0092, 0.3914, -0.0, -0.0076, 0.3891, 0.0017, -0.0084, 0.3902, 0.0017, -0.0081, 0.3891, 0.0039, -0.0084, 0.3902, 0.0039, -0.0081, 0.3915, 0.0017, -0.0075, 0.3915, 0.0039, -0.0075, 0.3942, -0.0, -0.0191, 0.3937, -0.0, -0.0157, 0.3961, -0.0, -0.0183, 0.3958, -0.0, -0.0157, 0.3937, -0.0, -0.0136, 0.3937, -0.0, -0.0114, 0.3958, -0.0, -0.0136,
0.3958, -0.0, -0.0114, 0.3946, 0.0017, -0.0189, 0.3946, 0.0039, -0.0189, 0.3959, 0.0017, -0.0185, 0.3959, 0.0039, -0.0185, 0.3982, -0.0, -0.0174, 0.398, -0.0, -0.0157, 0.3991, -0.0, -0.0169, 0.4003, -0.0, -0.0162, 0.398, -0.0, -0.0136, 0.398, -0.0, -0.0114, 0.4002, -0.0, -0.0136, 0.4002, -0.0, -0.0114, 0.398, 0.0017, -0.0175, 0.398, 0.0039, -0.0175, 0.3992, 0.0017, -0.0169, 0.4003, 0.0017, -0.0162, 0.3992, 0.0039, -0.0169, 0.4003, 0.0039, -0.0162, 0.3937, -0.0, -0.0092, 0.3935, -0.0, -0.0065, 0.3958, -0.0, -0.0092, 0.3958, -0.0, -0.0071, 0.3945, -0.0, -0.0058, 0.3956, -0.0, -0.0051, 0.3935, 0.0017, -0.0065, 0.3935, 0.0039, -0.0065, 0.3945, 0.0017, -0.0058, 0.3945, 0.0039, -0.0058, 0.3959, 0.0017, -0.0049,
0.3959, 0.0039, -0.0049, 0.398, -0.0, -0.0092, 0.398, -0.0, -0.0071, 0.4002, -0.0, -0.0092, 0.4002, -0.0, -0.0071, 0.3971, -0.0, -0.0038, 0.3981, -0.0, -0.0028, 0.4002, -0.0, -0.0049, 0.399, -0.0, -0.0017, 0.397, 0.0017, -0.0039, 0.3982, 0.0017, -0.0028, 0.397, 0.0039, -0.0039, 0.3982, 0.0039, -0.0028, 0.3991, 0.0017, -0.0016, 0.3991, 0.0039, -0.0016, 0.4018, -0.0, -0.0151, 0.4029, -0.0, -0.0143, 0.4024, -0.0, -0.0114, 0.4042, -0.0, -0.0132, 0.4053, -0.0, -0.0122, 0.4018, 0.0017, -0.0151, 0.4018, 0.0039, -0.0151, 0.4029, 0.0017, -0.0142, 0.4029, 0.0039, -0.0142, 0.4042, 0.0017, -0.0132, 0.4053, 0.0017, -0.0122, 0.4042, 0.0039, -0.0132, 0.4053, 0.0039, -0.0122, 0.4064, -0.0, -0.0111, 0.4064, 0.0017, -0.0111,
0.4064, 0.0039, -0.0111, 0.4024, -0.0, -0.0092, 0.4024, -0.0, -0.0071, 0.4045, -0.0, -0.0092, 0.4045, -0.0, -0.0071, 0.4024, -0.0, -0.0049, 0.4024, -0.0, -0.0027, 0.4045, -0.0, -0.0049, 0.4045, -0.0, -0.0027, 0.4075, -0.0, -0.0099, 0.4067, -0.0, -0.0071, 0.4084, -0.0, -0.0088, 0.4095, -0.0, -0.0074, 0.4067, -0.0, -0.0049, 0.4067, -0.0, -0.0027, 0.4089, -0.0, -0.0049, 0.4089, -0.0, -0.0027, 0.4075, 0.0017, -0.0099, 0.4075, 0.0039, -0.0099, 0.4084, 0.0017, -0.0089, 0.4094, 0.0017, -0.0074, 0.4084, 0.0039, -0.0089, 0.4094, 0.0039, -0.0074, 0.4102, -0.0, -0.0063, 0.4109, -0.0, -0.0051, 0.412, -0.0, -0.0032, 0.4125, -0.0, -0.0021, 0.4102, 0.0017, -0.0063, 0.4102, 0.0039, -0.0063, 0.4111, 0.0017, -0.0049,
0.4119, 0.0017, -0.0034, 0.4111, 0.0039, -0.0049, 0.4119, 0.0039, -0.0034, 0.4124, 0.0017, -0.0023, 0.4124, 0.0039, -0.0023, 0.4324, -0.0, -0.067, 0.4349, -0.0, -0.067, 0.4324, -0.0, -0.0656, 0.4324, -0.0, -0.0635, 0.4349, -0.0, -0.0656, 0.4353, -0.0, -0.0631, 0.4324, 0.0017, -0.067, 0.4324, 0.0039, -0.067, 0.4349, 0.0017, -0.067, 0.4349, 0.0039, -0.067, 0.4324, 0.0017, -0.0656, 0.4324, 0.0017, -0.0635, 0.4324, 0.0039, -0.0656, 0.4324, 0.0039, -0.0635, 0.4353, 0.0017, -0.0631, 0.4353, 0.0039, -0.0631, 0.4324, -0.0, -0.0613, 0.4324, -0.0, -0.0591, 0.4353, -0.0, -0.0613, 0.4353, -0.0, -0.0591, 0.4324, -0.0, -0.057, 0.4324, -0.0, -0.0548, 0.4353, -0.0, -0.057, 0.4353, -0.0, -0.0548, 0.4324, 0.0017, -0.0613,
0.4324, 0.0017, -0.0591, 0.4324, 0.0039, -0.0613, 0.4324, 0.0039, -0.0591, 0.4353, 0.0017, -0.0613, 0.4353, 0.0017, -0.0591, 0.4353, 0.0039, -0.0613, 0.4353, 0.0039, -0.0591, 0.4324, 0.0017, -0.057, 0.4324, 0.0017, -0.0548, 0.4324, 0.0039, -0.057, 0.4324, 0.0039, -0.0548, 0.4353, 0.0017, -0.057, 0.4353, 0.0017, -0.0548, 0.4353, 0.0039, -0.057, 0.4353, 0.0039, -0.0548, 0.4324, -0.0, -0.0526, 0.4324, -0.0, -0.0505, 0.4353, -0.0, -0.0527, 0.4353, -0.0, -0.0506, 0.4324, -0.0, -0.0483, 0.4324, -0.0, -0.0461, 0.4359, -0.0, -0.0487, 0.4349, -0.0, -0.0461, 0.4324, 0.0017, -0.0526, 0.4324, 0.0017, -0.0505, 0.4324, 0.0039, -0.0526, 0.4324, 0.0039, -0.0505, 0.4353, 0.0017, -0.0526, 0.4355, 0.0017, -0.0497, 0.4353, 0.0039, -0.0526,
0.4355, 0.0039, -0.0497, 0.4324, 0.0017, -0.0483, 0.4324, 0.0017, -0.0461, 0.4324, 0.0039, -0.0483, 0.4324, 0.0039, -0.0461, 0.4358, 0.0017, -0.0489, 0.4358, 0.0039, -0.0489, 0.4324, -0.0, -0.0439, 0.4324, -0.0, -0.0425, 0.4349, -0.0, -0.0439, 0.4349, -0.0, -0.0425, 0.4324, 0.0017, -0.0439, 0.4324, 0.0017, -0.0425, 0.4324, 0.0039, -0.0439, 0.4324, 0.0039, -0.0425, 0.4349, 0.0017, -0.0425, 0.4349, 0.0039, -0.0425, 0.4371, -0.0, -0.067, 0.4392, -0.0, -0.067, 0.4371, -0.0, -0.0656, 0.4371, -0.0, -0.0631, 0.4392, -0.0, -0.0656, 0.4392, -0.0, -0.0631, 0.4371, 0.0017, -0.067, 0.4371, 0.0039, -0.067, 0.4392, 0.0017, -0.067, 0.4392, 0.0039, -0.067, 0.4371, 0.0017, -0.0631, 0.4371, 0.0039, -0.0631, 0.4392, 0.0017, -0.0631,
0.4392, 0.0039, -0.0631, 0.4414, -0.0, -0.067, 0.4436, -0.0, -0.067, 0.4414, -0.0, -0.0656, 0.4414, -0.0, -0.0631, 0.4436, -0.0, -0.0656, 0.4436, -0.0, -0.0631, 0.4414, 0.0017, -0.067, 0.4414, 0.0039, -0.067, 0.4436, 0.0017, -0.067, 0.4436, 0.0039, -0.067, 0.4414, 0.0017, -0.0631, 0.4414, 0.0039, -0.0631, 0.4436, 0.0017, -0.0631, 0.4436, 0.0039, -0.0631, 0.4382, -0.0, -0.0562, 0.4382, -0.0, -0.0548, 0.4382, 0.0017, -0.0562, 0.4382, 0.0017, -0.0548, 0.4382, 0.0039, -0.0562, 0.4382, 0.0039, -0.0548, 0.4414, -0.0, -0.0562, 0.4414, -0.0, -0.0548, 0.4436, -0.0, -0.0562, 0.4436, -0.0, -0.0548, 0.4414, 0.0017, -0.0562, 0.4414, 0.0039, -0.0562, 0.4436, 0.0017, -0.0562, 0.4436, 0.0039, -0.0562, 0.4458, -0.0, -0.067,
0.4479, -0.0, -0.067, 0.4458, -0.0, -0.0656, 0.4458, -0.0, -0.0631, 0.4479, -0.0, -0.0656, 0.4479, -0.0, -0.0631, 0.4458, 0.0017, -0.067, 0.4458, 0.0039, -0.067, 0.4479, 0.0017, -0.067, 0.4479, 0.0039, -0.067, 0.4458, 0.0017, -0.0631, 0.4458, 0.0039, -0.0631, 0.4479, 0.0017, -0.0631, 0.4479, 0.0039, -0.0631, 0.4501, -0.0, -0.067, 0.4529, -0.0, -0.067, 0.4501, -0.0, -0.0656, 0.4501, -0.0, -0.0631, 0.4529, -0.0, -0.0656, 0.4529, -0.0, -0.0631, 0.4501, 0.0017, -0.067, 0.4501, 0.0039, -0.067, 0.4529, 0.0017, -0.067, 0.4529, 0.0039, -0.067, 0.4501, 0.0017, -0.0631, 0.4501, 0.0039, -0.0631, 0.4529, 0.0017, -0.0656, 0.4529, 0.0017, -0.0631, 0.4529, 0.0039, -0.0656, 0.4529, 0.0039, -0.0631, 0.4458, -0.0, -0.0562,
0.4458, -0.0, -0.0548, 0.4479, -0.0, -0.0562, 0.4479, -0.0, -0.0548, 0.4458, 0.0017, -0.0562, 0.4458, 0.0039, -0.0562, 0.4479, 0.0017, -0.0562, 0.4479, 0.0039, -0.0562, 0.4501, -0.0, -0.0562, 0.4501, -0.0, -0.0548, 0.452, -0.0, -0.0562, 0.452, -0.0, -0.0548, 0.4501, 0.0017, -0.0562, 0.4501, 0.0039, -0.0562, 0.452, 0.0017, -0.0562, 0.452, 0.0017, -0.0548, 0.452, 0.0039, -0.0562, 0.452, 0.0039, -0.0548, 0.4382, -0.0, -0.0523, 0.4363, -0.0, -0.0477, 0.4372, -0.0, -0.0466, 0.4393, -0.0, -0.0457, 0.4382, 0.0017, -0.0523, 0.4382, 0.0039, -0.0523, 0.4362, 0.0017, -0.0478, 0.4372, 0.0017, -0.0466, 0.4362, 0.0039, -0.0478, 0.4372, 0.0039, -0.0466, 0.4396, 0.0017, -0.0456, 0.4396, 0.0039, -0.0456, 0.4414, -0.0, -0.0523,
0.4436, -0.0, -0.0523, 0.4414, -0.0, -0.0455, 0.4436, -0.0, -0.0454, 0.4414, 0.0017, -0.0523, 0.4414, 0.0039, -0.0523, 0.4436, 0.0017, -0.0523, 0.4436, 0.0039, -0.0523, 0.4414, 0.0017, -0.0455, 0.4414, 0.0039, -0.0455, 0.4436, 0.0017, -0.0454, 0.4436, 0.0039, -0.0454, 0.4371, -0.0, -0.0439, 0.4371, -0.0, -0.0425, 0.4392, -0.0, -0.0439, 0.4392, -0.0, -0.0425, 0.4371, 0.0017, -0.0425, 0.4371, 0.0039, -0.0425, 0.4392, 0.0017, -0.0425, 0.4392, 0.0039, -0.0425, 0.4414, -0.0, -0.0439, 0.4414, -0.0, -0.0425, 0.4436, -0.0, -0.0439, 0.4436, -0.0, -0.0425, 0.4414, 0.0017, -0.0425, 0.4414, 0.0039, -0.0425, 0.4436, 0.0017, -0.0425, 0.4436, 0.0039, -0.0425, 0.4458, -0.0, -0.0523, 0.4479, -0.0, -0.0523, 0.4458, -0.0, -0.0454,
0.4479, -0.0, -0.0454, 0.4458, 0.0017, -0.0523, 0.4458, 0.0039, -0.0523, 0.4479, 0.0017, -0.0523, 0.4479, 0.0039, -0.0523, 0.4458, 0.0017, -0.0454, 0.4458, 0.0039, -0.0454, 0.4479, 0.0017, -0.0454, 0.4479, 0.0039, -0.0454, 0.4501, -0.0, -0.0523, 0.452, -0.0, -0.0523, 0.4501, -0.0, -0.0454, 0.4529, -0.0, -0.0454, 0.4501, 0.0017, -0.0523, 0.4501, 0.0039, -0.0523, 0.452, 0.0017, -0.0523, 0.452, 0.0039, -0.0523, 0.4501, 0.0017, -0.0454, 0.4501, 0.0039, -0.0454, 0.4529, 0.0017, -0.0454, 0.4529, 0.0039, -0.0454, 0.4458, -0.0, -0.0439, 0.4458, -0.0, -0.0425, 0.4479, -0.0, -0.0439, 0.4479, -0.0, -0.0425, 0.4458, 0.0017, -0.0425, 0.4458, 0.0039, -0.0425, 0.4479, 0.0017, -0.0425, 0.4479, 0.0039, -0.0425, 0.4501, -0.0, -0.0439,
0.4501, -0.0, -0.0425, 0.4529, -0.0, -0.0439, 0.4529, -0.0, -0.0425, 0.4501, 0.0017, -0.0425, 0.4501, 0.0039, -0.0425, 0.4529, 0.0017, -0.0439, 0.4529, 0.0017, -0.0425, 0.4529, 0.0039, -0.0439, 0.4529, 0.0039, -0.0425, 0.4245, -0.0, -0.0219, 0.4245, -0.0, -0.0201, 0.4262, -0.0, -0.0219, 0.4262, -0.0, -0.0201, 0.4245, 0.0017, -0.0219, 0.4245, 0.0017, -0.0201, 0.4245, 0.0039, -0.0219, 0.4245, 0.0039, -0.0201, 0.4262, 0.0017, -0.0219, 0.4262, 0.0039, -0.0219, 0.4284, -0.0, -0.0219, 0.4284, -0.0, -0.0201, 0.4306, -0.0, -0.0219, 0.4306, -0.0, -0.0201, 0.4284, 0.0017, -0.0219, 0.4284, 0.0039, -0.0219, 0.4306, 0.0017, -0.0219, 0.4306, 0.0039, -0.0219, 0.4327, -0.0, -0.0219, 0.4327, -0.0, -0.0201, 0.4349, -0.0, -0.0219,
0.4349, -0.0, -0.0201, 0.4327, 0.0017, -0.0219, 0.4327, 0.0039, -0.0219, 0.4349, 0.0017, -0.0219, 0.4349, 0.0039, -0.0219, 0.4245, -0.0, -0.0179, 0.4245, -0.0, -0.0157, 0.4262, -0.0, -0.0179, 0.4262, -0.0, -0.0157, 0.4245, -0.0, -0.0136, 0.4245, -0.0, -0.0114, 0.4262, -0.0, -0.0136, 0.4262, -0.0, -0.0114, 0.4245, 0.0017, -0.0179, 0.4245, 0.0017, -0.0157, 0.4245, 0.0039, -0.0179, 0.4245, 0.0039, -0.0157, 0.4245, 0.0017, -0.0136, 0.4245, 0.0017, -0.0114, 0.4245, 0.0039, -0.0136, 0.4245, 0.0039, -0.0114, 0.4245, -0.0, -0.0101, 0.4262, -0.0, -0.0101, 0.4245, 0.0017, -0.0101, 0.4245, 0.0039, -0.0101, 0.4262, 0.0017, -0.0101, 0.4262, 0.0039, -0.0101, 0.4284, -0.0, -0.0179, 0.4284, -0.0, -0.0157, 0.4306, -0.0, -0.0179,
0.4306, -0.0, -0.0157, 0.4284, -0.0, -0.0136, 0.4284, -0.0, -0.0114, 0.4306, -0.0, -0.0136, 0.4306, -0.0, -0.0114, 0.4327, -0.0, -0.0179, 0.4327, -0.0, -0.0157, 0.4349, -0.0, -0.0179, 0.4349, -0.0, -0.0157, 0.4327, -0.0, -0.0136, 0.4327, -0.0, -0.0114, 0.4349, -0.0, -0.0136, 0.4349, -0.0, -0.0114, 0.4284, -0.0, -0.0101, 0.4306, -0.0, -0.0101, 0.4284, 0.0017, -0.0101, 0.4284, 0.0039, -0.0101, 0.4306, 0.0017, -0.0101, 0.4306, 0.0039, -0.0101, 0.4327, -0.0, -0.0101, 0.4349, -0.0, -0.0101, 0.4327, 0.0017, -0.0101, 0.4327, 0.0039, -0.0101, 0.4349, 0.0017, -0.0101, 0.4349, 0.0039, -0.0101, 0.4371, -0.0, -0.0219, 0.4371, -0.0, -0.0201, 0.4392, -0.0, -0.0219, 0.4392, -0.0, -0.0201, 0.4371, 0.0017, -0.0219,
0.4371, 0.0039, -0.0219, 0.4392, 0.0017, -0.0219, 0.4392, 0.0039, -0.0219, 0.4414, -0.0, -0.0219, 0.4414, -0.0, -0.0201, 0.4436, -0.0, -0.0219, 0.4436, -0.0, -0.0201, 0.4414, 0.0017, -0.0219, 0.4414, 0.0039, -0.0219, 0.4436, 0.0017, -0.0219, 0.4436, 0.0039, -0.0219, 0.4458, -0.0, -0.0219, 0.4458, -0.0, -0.0201, 0.4479, -0.0, -0.0219, 0.4479, -0.0, -0.0201, 0.4458, 0.0017, -0.0219, 0.4458, 0.0039, -0.0219, 0.4479, 0.0017, -0.0219, 0.4479, 0.0039, -0.0219, 0.4501, -0.0, -0.0219, 0.4501, -0.0, -0.0201, 0.4523, -0.0, -0.0219, 0.4523, -0.0, -0.0201, 0.4501, 0.0017, -0.0219, 0.4501, 0.0039, -0.0219, 0.4523, 0.0017, -0.0219, 0.4523, 0.0039, -0.0219, 0.4371, -0.0, -0.0179, 0.4371, -0.0, -0.0157, 0.4392, -0.0, -0.0179,
0.4392, -0.0, -0.0157, 0.4371, -0.0, -0.0136, 0.4371, -0.0, -0.0114, 0.4392, -0.0, -0.0136, 0.4392, -0.0, -0.0114, 0.4414, -0.0, -0.0179, 0.4414, -0.0, -0.0157, 0.4436, -0.0, -0.0179, 0.4436, -0.0, -0.0157, 0.4414, -0.0, -0.0136, 0.4414, -0.0, -0.0114, 0.4436, -0.0, -0.0136, 0.4436, -0.0, -0.0114, 0.4371, -0.0, -0.0101, 0.4392, -0.0, -0.0101, 0.4371, 0.0017, -0.0101, 0.4371, 0.0039, -0.0101, 0.4392, 0.0017, -0.0101, 0.4392, 0.0039, -0.0101, 0.4414, -0.0, -0.0101, 0.4436, -0.0, -0.0101, 0.4414, 0.0017, -0.0101, 0.4414, 0.0039, -0.0101, 0.4436, 0.0017, -0.0101, 0.4436, 0.0039, -0.0101, 0.4458, -0.0, -0.0179, 0.4458, -0.0, -0.0157, 0.4479, -0.0, -0.0179, 0.4479, -0.0, -0.0157, 0.4458, -0.0, -0.0136,
0.4458, -0.0, -0.0114, 0.4479, -0.0, -0.0136, 0.4479, -0.0, -0.0114, 0.4501, -0.0, -0.0179, 0.4501, -0.0, -0.0157, 0.4523, -0.0, -0.0179, 0.4523, -0.0, -0.0157, 0.4501, -0.0, -0.0136, 0.4501, -0.0, -0.0114, 0.4523, -0.0, -0.0136, 0.4523, -0.0, -0.0114, 0.4458, -0.0, -0.0101, 0.4479, -0.0, -0.0101, 0.4458, 0.0017, -0.0101, 0.4458, 0.0039, -0.0101, 0.4479, 0.0017, -0.0101, 0.4479, 0.0039, -0.0101, 0.4501, -0.0, -0.0101, 0.4523, -0.0, -0.0101, 0.4501, 0.0017, -0.0101, 0.4501, 0.0039, -0.0101, 0.4523, 0.0017, -0.0101, 0.4523, 0.0039, -0.0101, 0.4725, -0.0, -0.067, 0.474, -0.0, -0.067, 0.4725, -0.0, -0.0656, 0.4725, -0.0, -0.0635, 0.474, -0.0, -0.0656, 0.474, -0.0, -0.0635, 0.4725, 0.0017, -0.067,
0.4725, 0.0039, -0.067, 0.474, 0.0017, -0.067, 0.474, 0.0039, -0.067, 0.4725, 0.0017, -0.0656, 0.4725, 0.0017, -0.0635, 0.4725, 0.0039, -0.0656, 0.4725, 0.0039, -0.0635, 0.4765, -0.0, -0.067, 0.4765, -0.0, -0.0656, 0.4765, -0.0, -0.0635, 0.4765, 0.0017, -0.067, 0.4765, 0.0039, -0.067, 0.4765, 0.0017, -0.0656, 0.4765, 0.0017, -0.0635, 0.4765, 0.0039, -0.0656, 0.4765, 0.0039, -0.0635, 0.4725, -0.0, -0.0613, 0.4725, -0.0, -0.0591, 0.474, -0.0, -0.0613, 0.474, -0.0, -0.0591, 0.4725, -0.0, -0.057, 0.4725, -0.0, -0.0548, 0.474, -0.0, -0.057, 0.474, -0.0, -0.0548, 0.4725, 0.0017, -0.0613, 0.4725, 0.0017, -0.0591, 0.4725, 0.0039, -0.0613, 0.4725, 0.0039, -0.0591, 0.4725, 0.0017, -0.057, 0.4725, 0.0017, -0.0548,
0.4725, 0.0039, -0.057, 0.4725, 0.0039, -0.0548, 0.4765, -0.0, -0.0613, 0.4765, -0.0, -0.0591, 0.4765, -0.0, -0.057, 0.4765, -0.0, -0.0548, 0.4765, 0.0017, -0.0613, 0.4765, 0.0017, -0.0591, 0.4765, 0.0039, -0.0613, 0.4765, 0.0039, -0.0591, 0.4765, 0.0017, -0.057, 0.4765, 0.0017, -0.0548, 0.4765, 0.0039, -0.057, 0.4765, 0.0039, -0.0548, 0.4804, -0.0, -0.0591, 0.4826, -0.0, -0.0591, 0.4804, -0.0, -0.0562, 0.4826, -0.0, -0.0562, 0.4804, 0.0017, -0.0591, 0.4804, 0.0039, -0.0591, 0.4826, 0.0017, -0.0591, 0.4826, 0.0039, -0.0591, 0.4804, 0.0017, -0.0562, 0.4804, 0.0039, -0.0562, 0.4826, 0.0017, -0.0562, 0.4826, 0.0039, -0.0562, 0.4848, -0.0, -0.0592, 0.4868, -0.0, -0.0596, 0.4848, -0.0, -0.0562, 0.487, -0.0, -0.0562,
0.4848, 0.0017, -0.0592, 0.4848, 0.0039, -0.0592, 0.487, 0.0017, -0.0596, 0.487, 0.0039, -0.0596, 0.4848, 0.0017, -0.0562, 0.4848, 0.0039, -0.0562, 0.487, 0.0017, -0.0562, 0.487, 0.0039, -0.0562, 0.4725, -0.0, -0.0526, 0.4725, -0.0, -0.0505, 0.474, -0.0, -0.0526, 0.474, -0.0, -0.0505, 0.4725, -0.0, -0.0483, 0.4725, -0.0, -0.0461, 0.474, -0.0, -0.0483, 0.474, -0.0, -0.0461, 0.4725, 0.0017, -0.0526, 0.4725, 0.0017, -0.0505, 0.4725, 0.0039, -0.0526, 0.4725, 0.0039, -0.0505, 0.4725, 0.0017, -0.0483, 0.4725, 0.0017, -0.0461, 0.4725, 0.0039, -0.0483, 0.4725, 0.0039, -0.0461, 0.4765, -0.0, -0.0527, 0.4766, -0.0, -0.0506, 0.477, -0.0, -0.0485, 0.4761, -0.0, -0.0461, 0.4775, -0.0, -0.0476, 0.4786, -0.0, -0.0465,
0.4765, 0.0017, -0.0526, 0.4766, 0.0017, -0.0504, 0.4765, 0.0039, -0.0526, 0.4766, 0.0039, -0.0504, 0.4769, 0.0017, -0.0487, 0.4769, 0.0039, -0.0487, 0.4775, 0.0017, -0.0476, 0.478, 0.0017, -0.047, 0.4775, 0.0039, -0.0476, 0.478, 0.0039, -0.047, 0.4725, -0.0, -0.0439, 0.4725, -0.0, -0.0425, 0.474, -0.0, -0.0439, 0.474, -0.0, -0.0425, 0.4725, 0.0017, -0.0439, 0.4725, 0.0017, -0.0425, 0.4725, 0.0039, -0.0439, 0.4725, 0.0039, -0.0425, 0.474, 0.0017, -0.0425, 0.474, 0.0039, -0.0425, 0.4761, -0.0, -0.0439, 0.4761, -0.0, -0.0425, 0.4783, -0.0, -0.0439, 0.4783, -0.0, -0.0425, 0.4761, 0.0017, -0.0425, 0.4761, 0.0039, -0.0425, 0.4783, 0.0017, -0.0425, 0.4783, 0.0039, -0.0425, 0.4801, -0.0, -0.0456, 0.4827, -0.0, -0.0455,
0.4801, 0.0017, -0.0456, 0.4801, 0.0039, -0.0456, 0.4826, 0.0017, -0.0455, 0.4826, 0.0039, -0.0455, 0.4848, -0.0, -0.0454, 0.487, -0.0, -0.0454, 0.4848, 0.0017, -0.0454, 0.4848, 0.0039, -0.0454, 0.487, 0.0017, -0.0454, 0.487, 0.0039, -0.0454, 0.4805, -0.0, -0.0439, 0.4805, -0.0, -0.0425, 0.4826, -0.0, -0.0439, 0.4826, -0.0, -0.0425, 0.4805, 0.0017, -0.0425, 0.4805, 0.0039, -0.0425, 0.4826, 0.0017, -0.0425, 0.4826, 0.0039, -0.0425, 0.4848, -0.0, -0.0439, 0.4848, -0.0, -0.0425, 0.487, -0.0, -0.0439, 0.487, -0.0, -0.0425, 0.4848, 0.0017, -0.0425, 0.4848, 0.0039, -0.0425, 0.487, 0.0017, -0.0425, 0.487, 0.0039, -0.0425, 0.4544, -0.0, -0.0219, 0.4544, -0.0, -0.0201, 0.4566, -0.0, -0.0219, 0.4566, -0.0, -0.0201,
0.4544, 0.0017, -0.0219, 0.4544, 0.0039, -0.0219, 0.4566, 0.0017, -0.0219, 0.4566, 0.0039, -0.0219, 0.4588, -0.0, -0.0219, 0.4588, -0.0, -0.0201, 0.4609, -0.0, -0.0219, 0.4609, -0.0, -0.0201, 0.4588, 0.0017, -0.0219, 0.4588, 0.0039, -0.0219, 0.4609, 0.0017, -0.0219, 0.4609, 0.0039, -0.0219, 0.4631, -0.0, -0.0219, 0.4631, -0.0, -0.0201, 0.4653, -0.0, -0.0219, 0.4653, -0.0, -0.0201, 0.4631, 0.0017, -0.0219, 0.4631, 0.0039, -0.0219, 0.4653, 0.0017, -0.0219, 0.4653, 0.0039, -0.0219, 0.4675, -0.0, -0.0219, 0.4675, -0.0, -0.0201, 0.4696, -0.0, -0.0219, 0.4696, -0.0, -0.0201, 0.4675, 0.0017, -0.0219, 0.4675, 0.0039, -0.0219, 0.4696, 0.0017, -0.0219, 0.4696, 0.0039, -0.0219, 0.4544, -0.0, -0.0179, 0.4544, -0.0, -0.0157,
0.4566, -0.0, -0.0179, 0.4566, -0.0, -0.0157, 0.4544, -0.0, -0.0136, 0.4544, -0.0, -0.0114, 0.4566, -0.0, -0.0136, 0.4566, -0.0, -0.0114, 0.4588, -0.0, -0.0179, 0.4588, -0.0, -0.0157, 0.4609, -0.0, -0.0179, 0.4609, -0.0, -0.0157, 0.4588, -0.0, -0.0136, 0.4588, -0.0, -0.0114, 0.4609, -0.0, -0.0136, 0.4609, -0.0, -0.0114, 0.4544, -0.0, -0.0101, 0.4566, -0.0, -0.0101, 0.4544, 0.0017, -0.0101, 0.4544, 0.0039, -0.0101, 0.4566, 0.0017, -0.0101, 0.4566, 0.0039, -0.0101, 0.4588, -0.0, -0.0101, 0.4609, -0.0, -0.0101, 0.4588, 0.0017, -0.0101, 0.4588, 0.0039, -0.0101, 0.4609, 0.0017, -0.0101, 0.4609, 0.0039, -0.0101, 0.4631, -0.0, -0.0179, 0.4631, -0.0, -0.0157, 0.4653, -0.0, -0.0179, 0.4653, -0.0, -0.0157,
0.4631, -0.0, -0.0136, 0.4631, -0.0, -0.0114, 0.4653, -0.0, -0.0136, 0.4653, -0.0, -0.0114, 0.4675, -0.0, -0.0179, 0.4675, -0.0, -0.0157, 0.4696, -0.0, -0.0179, 0.4696, -0.0, -0.0157, 0.4675, -0.0, -0.0136, 0.4675, -0.0, -0.0114, 0.4696, -0.0, -0.0136, 0.4696, -0.0, -0.0114, 0.4631, -0.0, -0.0101, 0.4653, -0.0, -0.0101, 0.4631, 0.0017, -0.0101, 0.4631, 0.0039, -0.0101, 0.4653, 0.0017, -0.0101, 0.4653, 0.0039, -0.0101, 0.4675, -0.0, -0.0101, 0.4696, -0.0, -0.0101, 0.4675, 0.0017, -0.0101, 0.4675, 0.0039, -0.0101, 0.4696, 0.0017, -0.0101, 0.4696, 0.0039, -0.0101, 0.4718, -0.0, -0.0219, 0.4718, -0.0, -0.0201, 0.474, -0.0, -0.0219, 0.474, -0.0, -0.0201, 0.4718, 0.0017, -0.0219, 0.4718, 0.0039, -0.0219,
0.474, 0.0017, -0.0219, 0.474, 0.0039, -0.0219, 0.4761, -0.0, -0.0219, 0.4761, -0.0, -0.0201, 0.4783, -0.0, -0.0218, 0.4783, -0.0, -0.0201, 0.4761, 0.0017, -0.0219, 0.4761, 0.0039, -0.0219, 0.4783, 0.0017, -0.0218, 0.4783, 0.0039, -0.0218, 0.4805, -0.0, -0.0216, 0.4805, -0.0, -0.0201, 0.4821, -0.0, -0.0213, 0.4829, -0.0, -0.0212, 0.4805, 0.0017, -0.0216, 0.4805, 0.0039, -0.0216, 0.4823, 0.0017, -0.0213, 0.4834, 0.0017, -0.0211, 0.4823, 0.0039, -0.0213, 0.4834, 0.0039, -0.0211, 0.485, -0.0, -0.0207, 0.4872, -0.0, -0.0202, 0.4848, 0.0017, -0.0208, 0.4848, 0.0039, -0.0208, 0.487, 0.0017, -0.0203, 0.487, 0.0039, -0.0203, 0.4718, -0.0, -0.0179, 0.4718, -0.0, -0.0157, 0.474, -0.0, -0.0179, 0.474, -0.0, -0.0157,
0.4718, -0.0, -0.0136, 0.4718, -0.0, -0.0114, 0.474, -0.0, -0.0136, 0.474, -0.0, -0.0114, 0.4761, -0.0, -0.0179, 0.4761, -0.0, -0.0157, 0.4783, -0.0, -0.0179, 0.4783, -0.0, -0.0157, 0.4761, -0.0, -0.0136, 0.4761, -0.0, -0.0114, 0.4783, -0.0, -0.0136, 0.4783, -0.0, -0.0114, 0.4718, -0.0, -0.0101, 0.474, -0.0, -0.0101, 0.4718, 0.0017, -0.0101, 0.4718, 0.0039, -0.0101, 0.474, 0.0017, -0.0101, 0.474, 0.0039, -0.0101, 0.4761, -0.0, -0.0101, 0.4783, -0.0, -0.01, 0.4761, 0.0017, -0.0101, 0.4761, 0.0039, -0.0101, 0.4783, 0.0017, -0.01, 0.4783, 0.0039, -0.01, 0.4805, -0.0, -0.0179, 0.4805, -0.0, -0.0157, 0.4826, -0.0, -0.0179, 0.4826, -0.0, -0.0157, 0.4805, -0.0, -0.0136, 0.4805, -0.0, -0.0114,
0.4826, -0.0, -0.0136, 0.4826, -0.0, -0.0114, 0.4848, -0.0, -0.0179, 0.4848, -0.0, -0.0157, 0.487, -0.0, -0.0179, 0.487, -0.0, -0.0157, 0.4848, -0.0, -0.0136, 0.4848, -0.0, -0.0114, 0.487, -0.0, -0.0136, 0.487, -0.0, -0.0114, 0.4804, -0.0, -0.0096, 0.4824, -0.0, -0.0091, 0.4805, 0.0017, -0.0096, 0.4805, 0.0039, -0.0096, 0.4827, 0.0017, -0.009, 0.4827, 0.0039, -0.009, 0.4843, -0.0, -0.0082, 0.4853, -0.0, -0.0077, 0.487, -0.0, -0.0092, 0.4869, -0.0, -0.0068, 0.4879, -0.0, -0.0058, 0.4841, 0.0017, -0.0083, 0.4852, 0.0017, -0.0078, 0.4841, 0.0039, -0.0083, 0.4852, 0.0039, -0.0078, 0.4869, 0.0017, -0.0067, 0.4869, 0.0039, -0.0067, 0.4879, 0.0017, -0.0058, 0.4879, 0.0039, -0.0058, 0.4919, -0.0, -0.0629,
0.4919, 0.0017, -0.0628, 0.4919, 0.0039, -0.0628, 0.4943, -0.0, -0.0648, 0.4932, -0.0, -0.064, 0.4954, -0.0, -0.0657, 0.4957, -0.0, -0.0635, 0.4943, 0.0017, -0.0648, 0.4932, 0.0017, -0.0639, 0.4943, 0.0039, -0.0648, 0.4932, 0.0039, -0.0639, 0.4956, 0.0017, -0.0658, 0.4956, 0.0039, -0.0658, 0.4895, -0.0, -0.0608, 0.4886, -0.0, -0.0603, 0.4909, -0.0, -0.0619, 0.4913, -0.0, -0.0591, 0.4891, -0.0, -0.0562, 0.4912, -0.0, -0.056, 0.4921, -0.0, -0.0558, 0.4894, 0.0017, -0.0608, 0.4882, 0.0017, -0.0601, 0.4894, 0.0039, -0.0608, 0.4882, 0.0039, -0.0601, 0.4908, 0.0017, -0.0619, 0.4908, 0.0039, -0.0619, 0.4892, 0.0017, -0.0562, 0.4892, 0.0039, -0.0562, 0.4909, 0.0017, -0.056, 0.492, 0.0017, -0.0558, 0.4909, 0.0039, -0.056,
0.492, 0.0039, -0.0558, 0.4935, -0.0, -0.0613, 0.4935, -0.0, -0.0591, 0.4963, -0.0, -0.0609, 0.4953, -0.0, -0.0594, 0.4935, -0.0, -0.057, 0.4934, -0.0, -0.0553, 0.4957, -0.0, -0.057, 0.4952, -0.0, -0.0543, 0.4963, 0.0017, -0.0608, 0.4953, 0.0017, -0.0594, 0.4963, 0.0039, -0.0608, 0.4953, 0.0039, -0.0594, 0.4935, 0.0017, -0.0553, 0.4935, 0.0039, -0.0553, 0.4953, 0.0017, -0.0543, 0.4953, 0.0039, -0.0543, 0.4987, -0.0, -0.0668, 0.5, -0.0, -0.0669, 0.4977, -0.0, -0.0666, 0.4983, -0.0, -0.063, 0.5005, -0.0, -0.0652, 0.4994, -0.0, -0.0641, 0.4986, 0.0017, -0.0668, 0.4986, 0.0039, -0.0668, 0.5, 0.0017, -0.0669, 0.5, 0.0039, -0.0669, 0.4975, 0.0017, -0.0665, 0.4983, 0.0017, -0.063, 0.4975, 0.0039, -0.0665,
0.4983, 0.0039, -0.063, 0.5005, 0.0017, -0.0652, 0.4994, 0.0017, -0.0641, 0.5005, 0.0039, -0.0652, 0.4994, 0.0039, -0.0641, 0.502, -0.0, -0.067, 0.5014, -0.0, -0.0664, 0.502, 0.0017, -0.067, 0.502, 0.0039, -0.067, 0.5014, 0.0017, -0.0663, 0.5014, 0.0039, -0.0663, 0.4972, -0.0, -0.0619, 0.497, -0.0, -0.0582, 0.4981, -0.0, -0.0574, 0.4978, -0.0, -0.0548, 0.4994, -0.0, -0.0563, 0.5006, -0.0, -0.0549, 0.4972, 0.0017, -0.0619, 0.497, 0.0017, -0.0582, 0.4972, 0.0039, -0.0619, 0.497, 0.0039, -0.0582, 0.4981, 0.0017, -0.0574, 0.4981, 0.0039, -0.0574, 0.4994, 0.0017, -0.0563, 0.5006, 0.0017, -0.0549, 0.4994, 0.0039, -0.0563, 0.5006, 0.0039, -0.0549, 0.5011, -0.0, -0.0537, 0.5011, 0.0017, -0.0538, 0.5011, 0.0039, -0.0538,
0.4891, -0.0, -0.0455, 0.4912, -0.0, -0.0459, 0.4892, 0.0017, -0.0455, 0.4892, 0.0039, -0.0455, 0.4914, 0.0017, -0.046, 0.4914, 0.0039, -0.046, 0.4964, -0.0, -0.0526, 0.497, -0.0, -0.0503, 0.4943, -0.0, -0.0474, 0.4933, -0.0, -0.0468, 0.4955, -0.0, -0.0484, 0.4957, -0.0, -0.0461, 0.4962, 0.0017, -0.053, 0.497, 0.0017, -0.0503, 0.4962, 0.0039, -0.053, 0.497, 0.0039, -0.0503, 0.4943, 0.0017, -0.0474, 0.4932, 0.0017, -0.0468, 0.4943, 0.0039, -0.0474, 0.4932, 0.0039, -0.0468, 0.4955, 0.0017, -0.0484, 0.4955, 0.0039, -0.0484, 0.4892, -0.0, -0.0439, 0.4892, -0.0, -0.0425, 0.4913, -0.0, -0.0439, 0.4913, -0.0, -0.0425, 0.4892, 0.0017, -0.0425, 0.4892, 0.0039, -0.0425, 0.4913, 0.0017, -0.0425, 0.4913, 0.0039, -0.0425,
0.4935, -0.0, -0.0439, 0.4935, -0.0, -0.0426, 0.4958, -0.0, -0.0431, 0.4946, -0.0, -0.0428, 0.4935, 0.0017, -0.0426, 0.4935, 0.0039, -0.0426, 0.4958, 0.0017, -0.0431, 0.4947, 0.0017, -0.0428, 0.4958, 0.0039, -0.0431, 0.4947, 0.0039, -0.0428, 0.4978, -0.0, -0.0526, 0.4978, -0.0, -0.0505, 0.5, -0.0, -0.0526, 0.5, -0.0, -0.0505, 0.4978, -0.0, -0.0483, 0.4978, -0.0, -0.0461, 0.5011, -0.0, -0.0477, 0.4998, -0.0, -0.0458, 0.501, 0.0017, -0.0475, 0.5, 0.0017, -0.046, 0.501, 0.0039, -0.0475, 0.5, 0.0039, -0.046, 0.5015, -0.0, -0.0527, 0.5018, -0.0, -0.0505, 0.5015, -0.0, -0.0488, 0.5015, 0.0017, -0.0527, 0.5018, 0.0017, -0.0505, 0.5015, 0.0039, -0.0527, 0.5018, 0.0039, -0.0505, 0.5015, 0.0017, -0.0486,
0.5015, 0.0039, -0.0486, 0.4981, -0.0, -0.0442, 0.499, -0.0, -0.045, 0.4979, 0.0017, -0.0441, 0.4979, 0.0039, -0.0441, 0.499, 0.0017, -0.045, 0.499, 0.0039, -0.045, 0.4889, -0.0, -0.0196, 0.4892, 0.0017, -0.0195, 0.4892, 0.0039, -0.0195, 0.4898, -0.0, -0.0192, 0.4892, -0.0, -0.0157, 0.4917, -0.0, -0.0183, 0.4913, -0.0, -0.0157, 0.4892, -0.0, -0.0136, 0.4892, -0.0, -0.0114, 0.4913, -0.0, -0.0136, 0.4913, -0.0, -0.0114, 0.4902, 0.0017, -0.019, 0.4902, 0.0039, -0.019, 0.4914, 0.0017, -0.0185, 0.4914, 0.0039, -0.0185, 0.4931, -0.0, -0.0174, 0.4942, -0.0, -0.0166, 0.4955, -0.0, -0.0156, 0.4935, -0.0, -0.0136, 0.4935, -0.0, -0.0114, 0.4966, -0.0, -0.0146, 0.4957, -0.0, -0.0114, 0.4932, 0.0017, -0.0174,
0.4942, 0.0017, -0.0166, 0.4932, 0.0039, -0.0174, 0.4942, 0.0039, -0.0166, 0.4956, 0.0017, -0.0155, 0.4956, 0.0039, -0.0155, 0.4966, 0.0017, -0.0145, 0.4966, 0.0039, -0.0145, 0.4892, -0.0, -0.0092, 0.4892, -0.0, -0.0071, 0.4913, -0.0, -0.0092, 0.4913, -0.0, -0.0071, 0.489, -0.0, -0.0048, 0.4899, -0.0, -0.0034, 0.4913, -0.0, -0.0049, 0.4904, -0.0, -0.0025, 0.489, 0.0017, -0.0048, 0.49, 0.0017, -0.0034, 0.489, 0.0039, -0.0048, 0.49, 0.0039, -0.0034, 0.4905, 0.0017, -0.0023, 0.4905, 0.0039, -0.0023, 0.4935, -0.0, -0.0092, 0.4935, -0.0, -0.0071, 0.4957, -0.0, -0.0092, 0.4957, -0.0, -0.0071, 0.4935, -0.0, -0.0049, 0.4935, -0.0, -0.0027, 0.4957, -0.0, -0.0049, 0.4957, -0.0, -0.0027, 0.4977, -0.0, -0.0135,
0.4988, -0.0, -0.0121, 0.4995, -0.0, -0.011, 0.4977, 0.0017, -0.0135, 0.4987, 0.0017, -0.0122, 0.4977, 0.0039, -0.0135, 0.4987, 0.0039, -0.0122, 0.4994, 0.0017, -0.0111, 0.4994, 0.0039, -0.0111, 0.4978, -0.0, -0.0092, 0.4978, -0.0, -0.0071, 0.5004, -0.0, -0.0096, 0.5012, -0.0, -0.0076, 0.4978, -0.0, -0.0049, 0.4978, -0.0, -0.0027, 0.5, -0.0, -0.0049, 0.5, -0.0, -0.0027, 0.5005, 0.0017, -0.0093, 0.501, 0.0017, -0.008, 0.5005, 0.0039, -0.0093, 0.501, 0.0039, -0.008, 0.5015, -0.0, -0.0067, 0.5018, -0.0, -0.0049, 0.5019, -0.0, -0.0027, 0.5014, 0.0017, -0.0069, 0.5014, 0.0039, -0.0069, 0.5018, 0.0017, -0.0049, 0.5019, 0.0017, -0.0027, 0.5018, 0.0039, -0.0049, 0.5019, 0.0039, -0.0027, 0.0031, -0.0, -0.0013,
0.003, -0.0, 0.0016, 0.0053, -0.0, -0.0012, 0.0052, -0.0, 0.0016, 0.003, -0.0, 0.0038, 0.003, -0.0, 0.006, 0.0052, -0.0, 0.0038, 0.0052, -0.0, 0.006, 0.003, 0.0017, -0.0013, 0.003, 0.0039, -0.0013, 0.0052, 0.0017, -0.0012, 0.0052, 0.0039, -0.0012, 0.0075, -0.0, -0.0009, 0.0074, -0.0, 0.0016, 0.0097, -0.0, -0.0004, 0.0096, -0.0, 0.0016, 0.0074, -0.0, 0.0038, 0.0074, -0.0, 0.006, 0.0096, -0.0, 0.0038, 0.0096, -0.0, 0.006, 0.0074, 0.0017, -0.0009, 0.0074, 0.0039, -0.0009, 0.0096, 0.0017, -0.0005, 0.0096, 0.0039, -0.0005, 0.003, -0.0, 0.0081, 0.003, -0.0, 0.0105, 0.0052, -0.0, 0.0081, 0.0051, -0.0, 0.0105, 0.003, 0.0017, 0.0105, 0.003, 0.0039, 0.0105, 0.0052, 0.0017, 0.0106,
0.0052, 0.0039, 0.0106, 0.0074, -0.0, 0.0081, 0.0072, -0.0, 0.0109, 0.0096, -0.0, 0.0081, 0.0089, -0.0, 0.0115, 0.0099, -0.0, 0.0119, 0.0074, 0.0017, 0.0109, 0.0074, 0.0039, 0.0109, 0.0085, 0.0017, 0.0114, 0.0085, 0.0039, 0.0114, 0.0096, 0.0017, 0.0118, 0.0096, 0.0039, 0.0118, 0.0114, -0.0, 0.0001, 0.0123, -0.0, 0.0004, 0.0142, -0.0, 0.0012, 0.0117, -0.0, 0.0038, 0.0117, -0.0, 0.006, 0.0139, -0.0, 0.0038, 0.0139, -0.0, 0.006, 0.0116, 0.0017, 0.0002, 0.0127, 0.0017, 0.0006, 0.0116, 0.0039, 0.0002, 0.0127, 0.0039, 0.0006, 0.0139, 0.0017, 0.0011, 0.0139, 0.0039, 0.0011, 0.0162, -0.0, 0.0022, 0.0172, -0.0, 0.0027, 0.0161, -0.0, 0.0038, 0.0161, -0.0, 0.006, 0.0183, -0.0, 0.0034,
0.0182, -0.0, 0.006, 0.0161, 0.0017, 0.0021, 0.0161, 0.0039, 0.0021, 0.0172, 0.0017, 0.0027, 0.0172, 0.0039, 0.0027, 0.0183, 0.0017, 0.0033, 0.0183, 0.0039, 0.0033, 0.0117, -0.0, 0.0081, 0.0117, -0.0, 0.0103, 0.0139, -0.0, 0.0081, 0.0139, -0.0, 0.0103, 0.0115, -0.0, 0.0129, 0.0126, -0.0, 0.0137, 0.0139, -0.0, 0.0125, 0.0138, -0.0, 0.0147, 0.0116, 0.0017, 0.0129, 0.0126, 0.0017, 0.0137, 0.0116, 0.0039, 0.0129, 0.0126, 0.0039, 0.0137, 0.0138, 0.0017, 0.0147, 0.0138, 0.0039, 0.0147, 0.0161, -0.0, 0.0081, 0.0161, -0.0, 0.0103, 0.0182, -0.0, 0.0081, 0.0182, -0.0, 0.0103, 0.0161, -0.0, 0.0125, 0.0161, -0.0, 0.0146, 0.0182, -0.0, 0.0125, 0.0182, -0.0, 0.0146, 0.0148, -0.0, 0.0159,
0.0148, 0.0017, 0.016, 0.0148, 0.0039, 0.016, 0.0156, -0.0, 0.017, 0.0167, -0.0, 0.019, 0.0182, -0.0, 0.0168, 0.0182, -0.0, 0.019, 0.0171, -0.0, 0.0203, 0.0173, -0.0, 0.0211, 0.0176, -0.0, 0.0233, 0.0156, 0.0017, 0.017, 0.0167, 0.0017, 0.0189, 0.0156, 0.0039, 0.017, 0.0167, 0.0039, 0.0189, 0.0171, 0.0017, 0.0203, 0.0171, 0.0039, 0.0203, 0.0174, 0.0017, 0.0214, 0.0176, 0.0017, 0.0233, 0.0174, 0.0039, 0.0214, 0.0176, 0.0039, 0.0233, 0.0144, -0.0, 0.0325, 0.0144, 0.0017, 0.0324, 0.0144, 0.0039, 0.0324, 0.0169, -0.0, 0.0282, 0.0175, -0.0, 0.0254, 0.0171, -0.0, 0.0274, 0.0163, -0.0, 0.0297, 0.0152, -0.0, 0.0315, 0.0182, -0.0, 0.0298, 0.0182, -0.0, 0.032, 0.017, 0.0017, 0.028,
0.017, 0.0039, 0.028, 0.0175, 0.0017, 0.0255, 0.0172, 0.0017, 0.0269, 0.0175, 0.0039, 0.0255, 0.0172, 0.0039, 0.0269, 0.0162, 0.0017, 0.0299, 0.0153, 0.0017, 0.0313, 0.0162, 0.0039, 0.0299, 0.0153, 0.0039, 0.0313, 0.0198, -0.0, 0.0044, 0.021, -0.0, 0.0053, 0.0221, -0.0, 0.0064, 0.0199, 0.0017, 0.0044, 0.021, 0.0017, 0.0053, 0.0199, 0.0039, 0.0044, 0.021, 0.0039, 0.0053, 0.0221, 0.0017, 0.0064, 0.0221, 0.0039, 0.0064, 0.0204, -0.0, 0.0081, 0.0204, -0.0, 0.0103, 0.0232, -0.0, 0.0076, 0.0226, -0.0, 0.0103, 0.0204, -0.0, 0.0125, 0.0204, -0.0, 0.0146, 0.0226, -0.0, 0.0125, 0.0226, -0.0, 0.0146, 0.0232, 0.0017, 0.0076, 0.0232, 0.0039, 0.0076, 0.0241, -0.0, 0.0087, 0.0252, -0.0, 0.0102,
0.0259, -0.0, 0.0113, 0.0247, -0.0, 0.0125, 0.0247, -0.0, 0.0146, 0.0265, -0.0, 0.0123, 0.0273, -0.0, 0.0144, 0.0241, 0.0017, 0.0087, 0.0252, 0.0017, 0.0102, 0.0241, 0.0039, 0.0087, 0.0252, 0.0039, 0.0102, 0.0259, 0.0017, 0.0113, 0.0259, 0.0039, 0.0113, 0.0265, 0.0017, 0.0124, 0.0274, 0.0017, 0.0146, 0.0265, 0.0039, 0.0124, 0.0274, 0.0039, 0.0146, 0.0204, -0.0, 0.0168, 0.0204, -0.0, 0.019, 0.0226, -0.0, 0.0168, 0.0226, -0.0, 0.019, 0.0204, -0.0, 0.0212, 0.0204, -0.0, 0.0233, 0.0226, -0.0, 0.0212, 0.0226, -0.0, 0.0233, 0.0247, -0.0, 0.0168, 0.0247, -0.0, 0.019, 0.028, -0.0, 0.0166, 0.0269, -0.0, 0.019, 0.0247, -0.0, 0.0212, 0.0247, -0.0, 0.0233, 0.0269, -0.0, 0.0212,
0.0269, -0.0, 0.0233, 0.0279, 0.0017, 0.0163, 0.0279, 0.0039, 0.0163, 0.0204, -0.0, 0.0255, 0.0204, -0.0, 0.0277, 0.0226, -0.0, 0.0255, 0.0226, -0.0, 0.0277, 0.0204, -0.0, 0.0298, 0.0204, -0.0, 0.032, 0.0226, -0.0, 0.0298, 0.0226, -0.0, 0.032, 0.0247, -0.0, 0.0255, 0.0247, -0.0, 0.0277, 0.0269, -0.0, 0.0255, 0.0269, -0.0, 0.0277, 0.0247, -0.0, 0.0298, 0.0247, -0.0, 0.032, 0.0281, -0.0, 0.0302, 0.0276, -0.0, 0.0323, 0.028, 0.0017, 0.0308, 0.0277, 0.0017, 0.032, 0.028, 0.0039, 0.0308, 0.0277, 0.0039, 0.032, 0.0281, -0.0, 0.0175, 0.0283, -0.0, 0.019, 0.0284, -0.0, 0.0212, 0.0284, -0.0, 0.0233, 0.0281, 0.0017, 0.0174, 0.0283, 0.0017, 0.019, 0.0281, 0.0039, 0.0174,
0.0283, 0.0039, 0.019, 0.0284, 0.0017, 0.0212, 0.0284, 0.0017, 0.0233, 0.0284, 0.0039, 0.0212, 0.0284, 0.0039, 0.0233, 0.0284, -0.0, 0.0255, 0.0284, -0.0, 0.0277, 0.0282, -0.0, 0.0294, 0.0284, 0.0017, 0.0255, 0.0284, 0.0017, 0.0277, 0.0284, 0.0039, 0.0255, 0.0284, 0.0039, 0.0277, 0.0282, 0.0017, 0.0297, 0.0282, 0.0039, 0.0297, 0.0061, -0.0, 0.0373, 0.003, -0.0, 0.0378, 0.003, -0.0, 0.0407, 0.0052, -0.0, 0.0376, 0.0052, -0.0, 0.0407, 0.006, 0.0017, 0.0374, 0.006, 0.0039, 0.0374, 0.003, 0.0017, 0.0378, 0.003, 0.0039, 0.0378, 0.0049, 0.0017, 0.0376, 0.0049, 0.0039, 0.0376, 0.0073, -0.0, 0.037, 0.0093, -0.0, 0.0362, 0.0074, -0.0, 0.0385, 0.0074, -0.0, 0.0407, 0.0096, -0.0, 0.0385,
0.0096, -0.0, 0.0407, 0.0074, 0.0017, 0.037, 0.0074, 0.0039, 0.037, 0.0096, 0.0017, 0.0361, 0.0096, 0.0039, 0.0361, 0.003, -0.0, 0.0429, 0.003, -0.0, 0.045, 0.0052, -0.0, 0.0429, 0.0052, -0.0, 0.045, 0.003, -0.0, 0.0472, 0.003, -0.0, 0.0487, 0.0052, -0.0, 0.0472, 0.0052, -0.0, 0.0487, 0.003, 0.0017, 0.0487, 0.003, 0.0039, 0.0487, 0.0052, 0.0017, 0.0487, 0.0052, 0.0039, 0.0487, 0.0074, -0.0, 0.0429, 0.0074, -0.0, 0.045, 0.0096, -0.0, 0.0429, 0.0096, -0.0, 0.045, 0.0078, -0.0, 0.0484, 0.007, -0.0, 0.0485, 0.0098, -0.0, 0.048, 0.0084, 0.0017, 0.0483, 0.0073, 0.0017, 0.0485, 0.0084, 0.0039, 0.0483, 0.0073, 0.0039, 0.0485, 0.0096, 0.0017, 0.048, 0.0096, 0.0039, 0.048,
0.012, -0.0, 0.0346, 0.011, -0.0, 0.0352, 0.0133, -0.0, 0.0335, 0.0139, -0.0, 0.0363, 0.0117, -0.0, 0.0385, 0.0117, -0.0, 0.0407, 0.0139, -0.0, 0.0385, 0.0139, -0.0, 0.0407, 0.012, 0.0017, 0.0347, 0.0108, 0.0017, 0.0354, 0.012, 0.0039, 0.0347, 0.0108, 0.0039, 0.0354, 0.0133, 0.0017, 0.0335, 0.0133, 0.0039, 0.0335, 0.0161, -0.0, 0.0342, 0.0161, -0.0, 0.0363, 0.0182, -0.0, 0.0342, 0.0182, -0.0, 0.0363, 0.0161, -0.0, 0.0385, 0.0161, -0.0, 0.0407, 0.0182, -0.0, 0.0385, 0.0182, -0.0, 0.0407, 0.0117, -0.0, 0.0429, 0.0117, -0.0, 0.045, 0.0139, -0.0, 0.0429, 0.0146, -0.0, 0.0463, 0.0119, -0.0, 0.0473, 0.0136, -0.0, 0.0467, 0.0149, 0.0017, 0.0461, 0.0149, 0.0039, 0.0461,
0.0117, 0.0017, 0.0474, 0.0117, 0.0039, 0.0474, 0.0139, 0.0017, 0.0466, 0.0139, 0.0039, 0.0466, 0.0161, -0.0, 0.0429, 0.0164, -0.0, 0.0454, 0.0189, -0.0, 0.0438, 0.0178, -0.0, 0.0445, 0.0161, 0.0017, 0.0455, 0.0161, 0.0039, 0.0455, 0.019, 0.0017, 0.0437, 0.0179, 0.0017, 0.0445, 0.019, 0.0039, 0.0437, 0.0179, 0.0039, 0.0445, 0.0204, -0.0, 0.0342, 0.0204, -0.0, 0.0363, 0.0226, -0.0, 0.0342, 0.0226, -0.0, 0.0363, 0.0204, -0.0, 0.0385, 0.0214, -0.0, 0.0417, 0.0236, -0.0, 0.0394, 0.0225, -0.0, 0.0406, 0.0214, 0.0017, 0.0417, 0.0214, 0.0039, 0.0417, 0.0235, 0.0017, 0.0394, 0.0225, 0.0017, 0.0407, 0.0235, 0.0039, 0.0394, 0.0225, 0.0039, 0.0407, 0.0247, -0.0, 0.0342, 0.0255, -0.0, 0.0367,
0.0268, -0.0, 0.0344, 0.0261, -0.0, 0.0356, 0.0244, -0.0, 0.0383, 0.0254, 0.0017, 0.0368, 0.0254, 0.0039, 0.0368, 0.0269, 0.0017, 0.0342, 0.0261, 0.0017, 0.0357, 0.0269, 0.0039, 0.0342, 0.0261, 0.0039, 0.0357, 0.0244, 0.0017, 0.0384, 0.0244, 0.0039, 0.0384, 0.0203, -0.0, 0.0427, 0.0204, 0.0017, 0.0427, 0.0204, 0.0039, 0.0427, 0.0716, -0.0, -0.0005, 0.0716, -0.0, 0.0016, 0.0747, -0.0, -0.0005, 0.0747, -0.0, 0.0016, 0.0716, -0.0, 0.0038, 0.0716, -0.0, 0.006, 0.0747, -0.0, 0.0038, 0.0747, -0.0, 0.006, 0.0716, 0.0017, -0.0005, 0.0716, 0.0017, 0.0016, 0.0716, 0.0039, -0.0005, 0.0716, 0.0039, 0.0016, 0.0716, 0.0017, 0.0038, 0.0716, 0.0017, 0.006, 0.0716, 0.0039, 0.0038, 0.0716, 0.0039, 0.006,
0.0768, -0.0, -0.0005, 0.0768, -0.0, 0.0016, 0.079, -0.0, -0.0005, 0.079, -0.0, 0.0016, 0.0768, -0.0, 0.0038, 0.0768, -0.0, 0.006, 0.079, -0.0, 0.0038, 0.079, -0.0, 0.006, 0.0716, -0.0, 0.0081, 0.0716, -0.0, 0.0103, 0.0747, -0.0, 0.0081, 0.0747, -0.0, 0.0103, 0.0716, -0.0, 0.0125, 0.0716, -0.0, 0.0146, 0.0747, -0.0, 0.0125, 0.0747, -0.0, 0.0146, 0.0716, 0.0017, 0.0081, 0.0716, 0.0017, 0.0103, 0.0716, 0.0039, 0.0081, 0.0716, 0.0039, 0.0103, 0.0716, 0.0017, 0.0125, 0.0716, 0.0017, 0.0146, 0.0716, 0.0039, 0.0125, 0.0716, 0.0039, 0.0146, 0.0768, -0.0, 0.0081, 0.0768, -0.0, 0.0103, 0.079, -0.0, 0.0081, 0.079, -0.0, 0.0103, 0.0768, -0.0, 0.0125, 0.0768, -0.0, 0.0146,
0.079, -0.0, 0.0125, 0.079, -0.0, 0.0146, 0.0812, -0.0, -0.0005, 0.0812, -0.0, 0.0016, 0.0833, -0.0, -0.0005, 0.0833, -0.0, 0.0016, 0.0812, -0.0, 0.0038, 0.0812, -0.0, 0.006, 0.0833, -0.0, 0.0038, 0.0833, -0.0, 0.006, 0.0833, 0.0017, -0.0005, 0.0833, 0.0017, 0.0016, 0.0833, 0.0039, -0.0005, 0.0833, 0.0039, 0.0016, 0.0833, 0.0017, 0.0038, 0.0833, 0.0017, 0.006, 0.0833, 0.0039, 0.0038, 0.0833, 0.0039, 0.006, 0.0812, -0.0, 0.0081, 0.0812, -0.0, 0.0103, 0.0833, -0.0, 0.0085, 0.0833, -0.0, 0.0103, 0.0812, -0.0, 0.0125, 0.0812, -0.0, 0.0146, 0.0833, -0.0, 0.0125, 0.0833, -0.0, 0.0146, 0.0833, 0.0017, 0.0085, 0.0833, 0.0039, 0.0085, 0.0855, -0.0, 0.0085, 0.0855, -0.0, 0.0103,
0.0877, -0.0, 0.0085, 0.0877, -0.0, 0.0103, 0.0855, -0.0, 0.0125, 0.0855, -0.0, 0.0146, 0.0877, -0.0, 0.0125, 0.0877, -0.0, 0.0146, 0.0855, 0.0017, 0.0085, 0.0855, 0.0039, 0.0085, 0.0877, 0.0017, 0.0085, 0.0877, 0.0039, 0.0085, 0.0716, -0.0, 0.0168, 0.0716, -0.0, 0.019, 0.0747, -0.0, 0.0168, 0.0747, -0.0, 0.019, 0.0716, -0.0, 0.0212, 0.0716, -0.0, 0.0233, 0.0747, -0.0, 0.0212, 0.0747, -0.0, 0.0233, 0.0716, 0.0017, 0.0168, 0.0716, 0.0017, 0.019, 0.0716, 0.0039, 0.0168, 0.0716, 0.0039, 0.019, 0.0716, 0.0017, 0.0212, 0.0716, 0.0017, 0.0233, 0.0716, 0.0039, 0.0212, 0.0716, 0.0039, 0.0233, 0.0768, -0.0, 0.0168, 0.0768, -0.0, 0.019, 0.079, -0.0, 0.0168, 0.079, -0.0, 0.019,
0.0768, -0.0, 0.0212, 0.0768, -0.0, 0.0233, 0.079, -0.0, 0.0212, 0.079, -0.0, 0.0233, 0.0716, -0.0, 0.0255, 0.0716, -0.0, 0.0277, 0.0747, -0.0, 0.0255, 0.0747, -0.0, 0.0277, 0.0716, -0.0, 0.0298, 0.0716, -0.0, 0.032, 0.0747, -0.0, 0.0298, 0.0747, -0.0, 0.032, 0.0716, 0.0017, 0.0255, 0.0716, 0.0017, 0.0277, 0.0716, 0.0039, 0.0255, 0.0716, 0.0039, 0.0277, 0.0716, 0.0017, 0.0298, 0.0716, 0.0017, 0.032, 0.0716, 0.0039, 0.0298, 0.0716, 0.0039, 0.032, 0.0768, -0.0, 0.0255, 0.0768, -0.0, 0.0277, 0.079, -0.0, 0.0255, 0.079, -0.0, 0.0277, 0.0768, -0.0, 0.0298, 0.0768, -0.0, 0.032, 0.079, -0.0, 0.0298, 0.079, -0.0, 0.032, 0.0812, -0.0, 0.0168, 0.0812, -0.0, 0.019,
0.0833, -0.0, 0.0168, 0.0833, -0.0, 0.019, 0.0812, -0.0, 0.0212, 0.0812, -0.0, 0.0233, 0.0833, -0.0, 0.0203, 0.0833, -0.0, 0.0233, 0.0833, 0.0017, 0.0203, 0.0833, 0.0017, 0.0233, 0.0833, 0.0039, 0.0203, 0.0833, 0.0039, 0.0233, 0.0855, -0.0, 0.0168, 0.0855, -0.0, 0.019, 0.0877, -0.0, 0.0168, 0.0877, -0.0, 0.019, 0.0855, -0.0, 0.0203, 0.0877, -0.0, 0.0203, 0.0855, 0.0017, 0.0203, 0.0855, 0.0039, 0.0203, 0.0877, 0.0017, 0.0203, 0.0877, 0.0039, 0.0203, 0.0812, -0.0, 0.0255, 0.0812, -0.0, 0.0277, 0.0833, -0.0, 0.0255, 0.0833, -0.0, 0.0277, 0.0812, -0.0, 0.0298, 0.0812, -0.0, 0.032, 0.0833, -0.0, 0.0298, 0.0833, -0.0, 0.032, 0.0833, 0.0017, 0.0255, 0.0833, 0.0017, 0.0277,
0.0833, 0.0039, 0.0255, 0.0833, 0.0039, 0.0277, 0.0833, 0.0017, 0.0298, 0.0833, 0.0017, 0.032, 0.0833, 0.0039, 0.0298, 0.0833, 0.0039, 0.032, 0.0899, -0.0, 0.0085, 0.0899, -0.0, 0.0103, 0.092, -0.0, 0.0085, 0.092, -0.0, 0.0103, 0.0899, -0.0, 0.0125, 0.0899, -0.0, 0.0146, 0.092, -0.0, 0.0125, 0.092, -0.0, 0.0146, 0.0899, 0.0017, 0.0085, 0.0899, 0.0039, 0.0085, 0.092, 0.0017, 0.0085, 0.092, 0.0039, 0.0085, 0.0942, -0.0, 0.0085, 0.0942, -0.0, 0.0103, 0.0964, -0.0, 0.0085, 0.0964, -0.0, 0.0103, 0.0942, -0.0, 0.0125, 0.0942, -0.0, 0.0146, 0.0964, -0.0, 0.0125, 0.0964, -0.0, 0.0146, 0.0942, 0.0017, 0.0085, 0.0942, 0.0039, 0.0085, 0.0964, 0.0017, 0.0085, 0.0964, 0.0039, 0.0085,
0.0985, -0.0, 0.0085, 0.0985, -0.0, 0.0103, 0.1007, -0.0, 0.0085, 0.1007, -0.0, 0.0103, 0.0985, -0.0, 0.0125, 0.0985, -0.0, 0.0146, 0.1007, -0.0, 0.0125, 0.1007, -0.0, 0.0146, 0.0985, 0.0017, 0.0085, 0.0985, 0.0039, 0.0085, 0.1007, 0.0017, 0.0085, 0.1007, 0.0039, 0.0085, 0.1029, -0.0, 0.0085, 0.1029, -0.0, 0.0103, 0.105, -0.0, 0.0085, 0.105, -0.0, 0.0103, 0.1029, -0.0, 0.0125, 0.1029, -0.0, 0.0146, 0.105, -0.0, 0.0125, 0.105, -0.0, 0.0146, 0.1029, 0.0017, 0.0085, 0.1029, 0.0039, 0.0085, 0.105, 0.0017, 0.0085, 0.105, 0.0039, 0.0085, 0.0899, -0.0, 0.0168, 0.0899, -0.0, 0.019, 0.092, -0.0, 0.0168, 0.092, -0.0, 0.019, 0.0899, -0.0, 0.0203, 0.092, -0.0, 0.0203,
0.0899, 0.0017, 0.0203, 0.0899, 0.0039, 0.0203, 0.092, 0.0017, 0.0203, 0.092, 0.0039, 0.0203, 0.0942, -0.0, 0.0168, 0.0942, -0.0, 0.019, 0.0964, -0.0, 0.0168, 0.0964, -0.0, 0.019, 0.0942, -0.0, 0.0203, 0.0964, -0.0, 0.0203, 0.0942, 0.0017, 0.0203, 0.0942, 0.0039, 0.0203, 0.0964, 0.0017, 0.0203, 0.0964, 0.0039, 0.0203, 0.0985, -0.0, 0.0168, 0.0985, -0.0, 0.019, 0.1007, -0.0, 0.0168, 0.1007, -0.0, 0.019, 0.0985, -0.0, 0.0203, 0.1007, -0.0, 0.0203, 0.0985, 0.0017, 0.0203, 0.0985, 0.0039, 0.0203, 0.1007, 0.0017, 0.0203, 0.1007, 0.0039, 0.0203, 0.1029, -0.0, 0.0168, 0.1029, -0.0, 0.019, 0.105, -0.0, 0.0168, 0.105, -0.0, 0.019, 0.1029, -0.0, 0.0203, 0.105, -0.0, 0.0203,
0.1029, 0.0017, 0.0203, 0.1029, 0.0039, 0.0203, 0.105, 0.0017, 0.0203, 0.105, 0.0039, 0.0203, 0.0716, -0.0, 0.0342, 0.0716, -0.0, 0.0363, 0.0747, -0.0, 0.0342, 0.0747, -0.0, 0.0363, 0.0716, -0.0, 0.0385, 0.0716, -0.0, 0.0407, 0.0747, -0.0, 0.0385, 0.0747, -0.0, 0.0407, 0.0716, 0.0017, 0.0342, 0.0716, 0.0017, 0.0363, 0.0716, 0.0039, 0.0342, 0.0716, 0.0039, 0.0363, 0.0716, 0.0017, 0.0385, 0.0716, 0.0017, 0.0407, 0.0716, 0.0039, 0.0385, 0.0716, 0.0039, 0.0407, 0.0768, -0.0, 0.0342, 0.0768, -0.0, 0.0363, 0.079, -0.0, 0.0342, 0.079, -0.0, 0.0363, 0.0768, -0.0, 0.0385, 0.0768, -0.0, 0.0407, 0.079, -0.0, 0.0385, 0.079, -0.0, 0.0407, 0.0716, -0.0, 0.0429, 0.0716, -0.0, 0.045,
0.0747, -0.0, 0.0429, 0.0747, -0.0, 0.045, 0.0716, -0.0, 0.0472, 0.0716, -0.0, 0.0487, 0.0747, -0.0, 0.0472, 0.0747, -0.0, 0.0487, 0.0716, 0.0017, 0.0429, 0.0716, 0.0017, 0.045, 0.0716, 0.0039, 0.0429, 0.0716, 0.0039, 0.045, 0.0716, 0.0017, 0.0472, 0.0716, 0.0017, 0.0487, 0.0716, 0.0039, 0.0472, 0.0716, 0.0039, 0.0487, 0.0747, 0.0017, 0.0487, 0.0747, 0.0039, 0.0487, 0.0768, -0.0, 0.0429, 0.0768, -0.0, 0.045, 0.079, -0.0, 0.0429, 0.079, -0.0, 0.045, 0.0768, -0.0, 0.0472, 0.0768, -0.0, 0.0487, 0.079, -0.0, 0.0472, 0.079, -0.0, 0.0487, 0.0768, 0.0017, 0.0487, 0.0768, 0.0039, 0.0487, 0.079, 0.0017, 0.0487, 0.079, 0.0039, 0.0487, 0.0812, -0.0, 0.0342, 0.0812, -0.0, 0.0363,
0.0833, -0.0, 0.0342, 0.0833, -0.0, 0.0363, 0.0812, -0.0, 0.0385, 0.0812, -0.0, 0.0407, 0.0833, -0.0, 0.0379, 0.0833, -0.0, 0.0407, 0.0833, 0.0017, 0.0342, 0.0833, 0.0017, 0.0363, 0.0833, 0.0039, 0.0342, 0.0833, 0.0039, 0.0363, 0.0833, 0.0017, 0.0379, 0.0833, 0.0039, 0.0379, 0.0855, -0.0, 0.0379, 0.0855, -0.0, 0.0407, 0.0877, -0.0, 0.0379, 0.0877, -0.0, 0.0407, 0.0855, 0.0017, 0.0379, 0.0855, 0.0039, 0.0379, 0.0877, 0.0017, 0.0379, 0.0877, 0.0039, 0.0379, 0.0812, -0.0, 0.0429, 0.0812, -0.0, 0.045, 0.0833, -0.0, 0.0429, 0.0833, -0.0, 0.045, 0.0812, -0.0, 0.0472, 0.0812, -0.0, 0.0487, 0.0833, -0.0, 0.0472, 0.0833, -0.0, 0.0487, 0.0812, 0.0017, 0.0487, 0.0812, 0.0039, 0.0487,
0.0833, 0.0017, 0.0487, 0.0833, 0.0039, 0.0487, 0.0855, -0.0, 0.0429, 0.0855, -0.0, 0.045, 0.0877, -0.0, 0.0429, 0.0877, -0.0, 0.045, 0.0855, -0.0, 0.0472, 0.0855, -0.0, 0.0487, 0.0877, -0.0, 0.0472, 0.0877, -0.0, 0.0487, 0.0855, 0.0017, 0.0487, 0.0855, 0.0039, 0.0487, 0.0877, 0.0017, 0.0487, 0.0877, 0.0039, 0.0487, 0.0899, -0.0, 0.0379, 0.0899, -0.0, 0.0407, 0.092, -0.0, 0.0379, 0.092, -0.0, 0.0407, 0.0899, 0.0017, 0.0379, 0.0899, 0.0039, 0.0379, 0.092, 0.0017, 0.0379, 0.092, 0.0039, 0.0379, 0.0942, -0.0, 0.0379, 0.0942, -0.0, 0.0407, 0.0964, -0.0, 0.0379, 0.0964, -0.0, 0.0407, 0.0942, 0.0017, 0.0379, 0.0942, 0.0039, 0.0379, 0.0964, 0.0017, 0.0379, 0.0964, 0.0039, 0.0379,
0.0899, -0.0, 0.0429, 0.0899, -0.0, 0.045, 0.092, -0.0, 0.0429, 0.092, -0.0, 0.045, 0.0899, -0.0, 0.0472, 0.0899, -0.0, 0.0487, 0.092, -0.0, 0.0472, 0.092, -0.0, 0.0487, 0.0899, 0.0017, 0.0487, 0.0899, 0.0039, 0.0487, 0.092, 0.0017, 0.0487, 0.092, 0.0039, 0.0487, 0.0942, -0.0, 0.0429, 0.0942, -0.0, 0.045, 0.0964, -0.0, 0.0429, 0.0964, -0.0, 0.045, 0.0942, -0.0, 0.0472, 0.0942, -0.0, 0.0487, 0.0964, -0.0, 0.0472, 0.0964, -0.0, 0.0487, 0.0942, 0.0017, 0.0487, 0.0942, 0.0039, 0.0487, 0.0964, 0.0017, 0.0487, 0.0964, 0.0039, 0.0487, 0.0985, -0.0, 0.0379, 0.0985, -0.0, 0.0407, 0.1007, -0.0, 0.0379, 0.1007, -0.0, 0.0407, 0.0985, 0.0017, 0.0379, 0.0985, 0.0039, 0.0379,
0.1007, 0.0017, 0.0379, 0.1007, 0.0039, 0.0379, 0.1029, -0.0, 0.0379, 0.1029, -0.0, 0.0407, 0.105, -0.0, 0.0379, 0.105, -0.0, 0.0407, 0.1029, 0.0017, 0.0379, 0.1029, 0.0039, 0.0379, 0.105, 0.0017, 0.0379, 0.105, 0.0039, 0.0379, 0.0985, -0.0, 0.0429, 0.0985, -0.0, 0.045, 0.1007, -0.0, 0.0429, 0.1007, -0.0, 0.045, 0.0985, -0.0, 0.0472, 0.0985, -0.0, 0.0487, 0.1007, -0.0, 0.0472, 0.1007, -0.0, 0.0487, 0.0985, 0.0017, 0.0487, 0.0985, 0.0039, 0.0487, 0.1007, 0.0017, 0.0487, 0.1007, 0.0039, 0.0487, 0.1029, -0.0, 0.0429, 0.1029, -0.0, 0.045, 0.105, -0.0, 0.0429, 0.105, -0.0, 0.045, 0.1029, -0.0, 0.0472, 0.1029, -0.0, 0.0487, 0.105, -0.0, 0.0472, 0.105, -0.0, 0.0487,
0.1029, 0.0017, 0.0487, 0.1029, 0.0039, 0.0487, 0.105, 0.0017, 0.0487, 0.105, 0.0039, 0.0487, 0.1072, -0.0, 0.0085, 0.1072, -0.0, 0.0103, 0.1094, -0.0, 0.0085, 0.1094, -0.0, 0.0103, 0.1072, -0.0, 0.0125, 0.1072, -0.0, 0.0146, 0.1094, -0.0, 0.0125, 0.1094, -0.0, 0.0146, 0.1072, 0.0017, 0.0085, 0.1072, 0.0039, 0.0085, 0.1094, 0.0017, 0.0085, 0.1094, 0.0039, 0.0085, 0.1116, -0.0, 0.0085, 0.1116, -0.0, 0.0103, 0.1137, -0.0, 0.0085, 0.1137, -0.0, 0.0103, 0.1116, -0.0, 0.0125, 0.1116, -0.0, 0.0146, 0.1137, -0.0, 0.0125, 0.1137, -0.0, 0.0146, 0.1116, 0.0017, 0.0085, 0.1116, 0.0039, 0.0085, 0.1137, 0.0017, 0.0085, 0.1137, 0.0039, 0.0085, 0.1159, -0.0, 0.0085, 0.1159, -0.0, 0.0103,
0.1181, -0.0, 0.0085, 0.1181, -0.0, 0.0103, 0.1159, -0.0, 0.0125, 0.1159, -0.0, 0.0146, 0.1181, -0.0, 0.0125, 0.1181, -0.0, 0.0146, 0.1159, 0.0017, 0.0085, 0.1159, 0.0039, 0.0085, 0.1181, 0.0017, 0.0085, 0.1181, 0.0039, 0.0085, 0.1202, -0.0, 0.0085, 0.1202, -0.0, 0.0103, 0.1224, -0.0, 0.0085, 0.1224, -0.0, 0.0103, 0.1202, -0.0, 0.0125, 0.1202, -0.0, 0.0146, 0.1224, -0.0, 0.0125, 0.1224, -0.0, 0.0146, 0.1202, 0.0017, 0.0085, 0.1202, 0.0039, 0.0085, 0.1224, 0.0017, 0.0085, 0.1224, 0.0039, 0.0085, 0.1072, -0.0, 0.0168, 0.1072, -0.0, 0.019, 0.1094, -0.0, 0.0168, 0.1094, -0.0, 0.019, 0.1072, -0.0, 0.0203, 0.1094, -0.0, 0.0203, 0.1072, 0.0017, 0.0203, 0.1072, 0.0039, 0.0203,
0.1094, 0.0017, 0.0203, 0.1094, 0.0039, 0.0203, 0.1116, -0.0, 0.0168, 0.1116, -0.0, 0.019, 0.1137, -0.0, 0.0168, 0.1137, -0.0, 0.019, 0.1116, -0.0, 0.0203, 0.1137, -0.0, 0.0203, 0.1116, 0.0017, 0.0203, 0.1116, 0.0039, 0.0203, 0.1137, 0.0017, 0.0203, 0.1137, 0.0039, 0.0203, 0.1159, -0.0, 0.0168, 0.1159, -0.0, 0.019, 0.1181, -0.0, 0.0168, 0.1181, -0.0, 0.019, 0.1159, -0.0, 0.0203, 0.1181, -0.0, 0.0203, 0.1159, 0.0017, 0.0203, 0.1159, 0.0039, 0.0203, 0.1181, 0.0017, 0.0203, 0.1181, 0.0039, 0.0203, 0.1202, -0.0, 0.0168, 0.1202, -0.0, 0.019, 0.1224, -0.0, 0.0168, 0.1224, -0.0, 0.019, 0.1202, -0.0, 0.0203, 0.1224, -0.0, 0.0203, 0.1202, 0.0017, 0.0203, 0.1202, 0.0039, 0.0203,
0.1224, 0.0017, 0.0203, 0.1224, 0.0039, 0.0203, 0.1246, -0.0, 0.0085, 0.1246, -0.0, 0.0103, 0.1267, -0.0, 0.0085, 0.1267, -0.0, 0.0103, 0.1246, -0.0, 0.0125, 0.1246, -0.0, 0.0146, 0.1267, -0.0, 0.0125, 0.1267, -0.0, 0.0146, 0.1246, 0.0017, 0.0085, 0.1246, 0.0039, 0.0085, 0.1267, 0.0017, 0.0085, 0.1267, 0.0039, 0.0085, 0.1294, -0.0, 0.0085, 0.1294, -0.0, 0.0103, 0.1294, -0.0, 0.0125, 0.1294, -0.0, 0.0146, 0.1294, 0.0017, 0.0085, 0.1294, 0.0017, 0.0103, 0.1294, 0.0039, 0.0085, 0.1294, 0.0039, 0.0103, 0.1294, 0.0017, 0.0125, 0.1294, 0.0017, 0.0146, 0.1294, 0.0039, 0.0125, 0.1294, 0.0039, 0.0146, 0.1246, -0.0, 0.0168, 0.1246, -0.0, 0.019, 0.1267, -0.0, 0.0168, 0.1267, -0.0, 0.019,
0.1246, -0.0, 0.0203, 0.1267, -0.0, 0.0203, 0.1246, 0.0017, 0.0203, 0.1246, 0.0039, 0.0203, 0.1267, 0.0017, 0.0203, 0.1267, 0.0039, 0.0203, 0.1294, -0.0, 0.0168, 0.1294, -0.0, 0.019, 0.1294, -0.0, 0.0203, 0.1294, 0.0017, 0.0168, 0.1294, 0.0017, 0.019, 0.1294, 0.0039, 0.0168, 0.1294, 0.0039, 0.019, 0.1294, 0.0017, 0.0203, 0.1294, 0.0039, 0.0203, 0.1072, -0.0, 0.0379, 0.1072, -0.0, 0.0407, 0.1094, -0.0, 0.0379, 0.1094, -0.0, 0.0407, 0.1072, 0.0017, 0.0379, 0.1072, 0.0039, 0.0379, 0.1094, 0.0017, 0.0379, 0.1094, 0.0039, 0.0379, 0.1116, -0.0, 0.0379, 0.1116, -0.0, 0.0407, 0.1137, -0.0, 0.0379, 0.1137, -0.0, 0.0407, 0.1116, 0.0017, 0.0379, 0.1116, 0.0039, 0.0379, 0.1137, 0.0017, 0.0379,
0.1137, 0.0039, 0.0379, 0.1072, -0.0, 0.0429, 0.1072, -0.0, 0.045, 0.1094, -0.0, 0.0429, 0.1094, -0.0, 0.045, 0.1072, -0.0, 0.0472, 0.1072, -0.0, 0.0487, 0.1094, -0.0, 0.0472, 0.1094, -0.0, 0.0487, 0.1072, 0.0017, 0.0487, 0.1072, 0.0039, 0.0487, 0.1094, 0.0017, 0.0487, 0.1094, 0.0039, 0.0487, 0.1116, -0.0, 0.0429, 0.1116, -0.0, 0.045, 0.1137, -0.0, 0.0429, 0.1137, -0.0, 0.045, 0.1116, -0.0, 0.0472, 0.1116, -0.0, 0.0487, 0.1137, -0.0, 0.0472, 0.1137, -0.0, 0.0487, 0.1116, 0.0017, 0.0487, 0.1116, 0.0039, 0.0487, 0.1137, 0.0017, 0.0487, 0.1137, 0.0039, 0.0487, 0.1159, -0.0, 0.0379, 0.1159, -0.0, 0.0407, 0.1181, -0.0, 0.0379, 0.1181, -0.0, 0.0407, 0.1159, 0.0017, 0.0379,
0.1159, 0.0039, 0.0379, 0.1181, 0.0017, 0.0379, 0.1181, 0.0039, 0.0379, 0.1202, -0.0, 0.0379, 0.1202, -0.0, 0.0407, 0.1224, -0.0, 0.0379, 0.1224, -0.0, 0.0407, 0.1202, 0.0017, 0.0379, 0.1202, 0.0039, 0.0379, 0.1224, 0.0017, 0.0379, 0.1224, 0.0039, 0.0379, 0.1159, -0.0, 0.0429, 0.1159, -0.0, 0.045, 0.1181, -0.0, 0.0429, 0.1181, -0.0, 0.045, 0.1159, -0.0, 0.0472, 0.1159, -0.0, 0.0487, 0.1181, -0.0, 0.0472, 0.1181, -0.0, 0.0487, 0.1159, 0.0017, 0.0487, 0.1159, 0.0039, 0.0487, 0.1181, 0.0017, 0.0487, 0.1181, 0.0039, 0.0487, 0.1202, -0.0, 0.0429, 0.1202, -0.0, 0.045, 0.1224, -0.0, 0.0429, 0.1224, -0.0, 0.045, 0.1202, -0.0, 0.0472, 0.1202, -0.0, 0.0487, 0.1224, -0.0, 0.0472,
0.1224, -0.0, 0.0487, 0.1202, 0.0017, 0.0487, 0.1202, 0.0039, 0.0487, 0.1224, 0.0017, 0.0487, 0.1224, 0.0039, 0.0487, 0.1246, -0.0, 0.0379, 0.1246, -0.0, 0.0407, 0.1267, -0.0, 0.0379, 0.1267, -0.0, 0.0407, 0.1246, 0.0017, 0.0379, 0.1246, 0.0039, 0.0379, 0.1267, 0.0017, 0.0379, 0.1267, 0.0039, 0.0379, 0.1289, -0.0, 0.0379, 0.1289, -0.0, 0.0407, 0.1311, -0.0, 0.0379, 0.1311, -0.0, 0.0407, 0.1289, 0.0017, 0.0379, 0.1289, 0.0039, 0.0379, 0.1311, 0.0017, 0.0379, 0.1311, 0.0039, 0.0379, 0.1246, -0.0, 0.0429, 0.1246, -0.0, 0.045, 0.1267, -0.0, 0.0429, 0.1267, -0.0, 0.045, 0.1246, -0.0, 0.0472, 0.1246, -0.0, 0.0487, 0.1267, -0.0, 0.0472, 0.1267, -0.0, 0.0487, 0.1246, 0.0017, 0.0487,
0.1246, 0.0039, 0.0487, 0.1267, 0.0017, 0.0487, 0.1267, 0.0039, 0.0487, 0.1289, -0.0, 0.0429, 0.1289, -0.0, 0.045, 0.1311, -0.0, 0.0429, 0.1311, -0.0, 0.045, 0.1289, -0.0, 0.0472, 0.1289, -0.0, 0.0487, 0.1311, -0.0, 0.0472, 0.1311, -0.0, 0.0487, 0.1289, 0.0017, 0.0487, 0.1289, 0.0039, 0.0487, 0.1311, 0.0017, 0.0487, 0.1311, 0.0039, 0.0487, 0.1324, -0.0, 0.0379, 0.1324, -0.0, 0.0407, 0.1324, 0.0017, 0.0379, 0.1324, 0.0017, 0.0407, 0.1324, 0.0039, 0.0379, 0.1324, 0.0039, 0.0407, 0.1324, -0.0, 0.0429, 0.1324, -0.0, 0.045, 0.1324, -0.0, 0.0472, 0.1324, -0.0, 0.0487, 0.1324, 0.0017, 0.0429, 0.1324, 0.0017, 0.045, 0.1324, 0.0039, 0.0429, 0.1324, 0.0039, 0.045, 0.1324, 0.0017, 0.0472,
0.1324, 0.0017, 0.0487, 0.1324, 0.0039, 0.0472, 0.1324, 0.0039, 0.0487, 0.1412, -0.0, -0.0005, 0.1412, -0.0, 0.0016, 0.1441, -0.0, -0.0005, 0.1441, -0.0, 0.0016, 0.1412, -0.0, 0.0038, 0.1412, -0.0, 0.006, 0.1441, -0.0, 0.0038, 0.1441, -0.0, 0.006, 0.1412, 0.0017, -0.0005, 0.1412, 0.0017, 0.0016, 0.1412, 0.0039, -0.0005, 0.1412, 0.0039, 0.0016, 0.1412, 0.0017, 0.0038, 0.1412, 0.0017, 0.006, 0.1412, 0.0039, 0.0038, 0.1412, 0.0039, 0.006, 0.1463, -0.0, -0.0005, 0.1463, -0.0, 0.0016, 0.1484, -0.0, -0.0005, 0.1484, -0.0, 0.0016, 0.1463, -0.0, 0.0038, 0.1463, -0.0, 0.006, 0.1484, -0.0, 0.0038, 0.1484, -0.0, 0.006, 0.1412, -0.0, 0.0081, 0.1412, -0.0, 0.0103, 0.1441, -0.0, 0.0081,
0.1441, -0.0, 0.0103, 0.1412, -0.0, 0.0125, 0.1412, -0.0, 0.0146, 0.1441, -0.0, 0.0125, 0.1441, -0.0, 0.0146, 0.1412, 0.0017, 0.0081, 0.1412, 0.0017, 0.0103, 0.1412, 0.0039, 0.0081, 0.1412, 0.0039, 0.0103, 0.1412, 0.0017, 0.0125, 0.1412, 0.0017, 0.0146, 0.1412, 0.0039, 0.0125, 0.1412, 0.0039, 0.0146, 0.1463, -0.0, 0.0081, 0.1463, -0.0, 0.0103, 0.1484, -0.0, 0.0081, 0.1484, -0.0, 0.0103, 0.1463, -0.0, 0.0125, 0.1463, -0.0, 0.0146, 0.1484, -0.0, 0.0125, 0.1484, -0.0, 0.0146, 0.1506, -0.0, -0.0005, 0.1506, -0.0, 0.0016, 0.152, -0.0, -0.0005, 0.152, -0.0, 0.0016, 0.1506, -0.0, 0.0038, 0.1506, -0.0, 0.006, 0.152, -0.0, 0.0038, 0.152, -0.0, 0.006, 0.152, 0.0017, -0.0005,
0.152, 0.0017, 0.0016, 0.152, 0.0039, -0.0005, 0.152, 0.0039, 0.0016, 0.152, 0.0017, 0.0038, 0.152, 0.0017, 0.006, 0.152, 0.0039, 0.0038, 0.152, 0.0039, 0.006, 0.1506, -0.0, 0.0081, 0.1506, -0.0, 0.0103, 0.152, -0.0, 0.0081, 0.152, -0.0, 0.0103, 0.1506, -0.0, 0.0125, 0.1506, -0.0, 0.0146, 0.152, -0.0, 0.0125, 0.152, -0.0, 0.0146, 0.152, 0.0017, 0.0081, 0.152, 0.0017, 0.0103, 0.152, 0.0039, 0.0081, 0.152, 0.0039, 0.0103, 0.152, 0.0017, 0.0125, 0.152, 0.0017, 0.0146, 0.152, 0.0039, 0.0125, 0.152, 0.0039, 0.0146, 0.1412, -0.0, 0.0168, 0.1412, -0.0, 0.019, 0.1441, -0.0, 0.0168, 0.1441, -0.0, 0.019, 0.1412, -0.0, 0.0212, 0.1412, -0.0, 0.0233, 0.1441, -0.0, 0.0212,
0.1441, -0.0, 0.0233, 0.1412, 0.0017, 0.0168, 0.1412, 0.0017, 0.019, 0.1412, 0.0039, 0.0168, 0.1412, 0.0039, 0.019, 0.1412, 0.0017, 0.0212, 0.1412, 0.0017, 0.0233, 0.1412, 0.0039, 0.0212, 0.1412, 0.0039, 0.0233, 0.1463, -0.0, 0.0168, 0.1463, -0.0, 0.019, 0.1484, -0.0, 0.0168, 0.1484, -0.0, 0.019, 0.1463, -0.0, 0.0212, 0.1463, -0.0, 0.0233, 0.1484, -0.0, 0.0212, 0.1484, -0.0, 0.0233, 0.1412, -0.0, 0.0255, 0.1412, -0.0, 0.0277, 0.1441, -0.0, 0.0255, 0.1441, -0.0, 0.0277, 0.1412, -0.0, 0.0298, 0.1412, -0.0, 0.032, 0.1441, -0.0, 0.0298, 0.1441, -0.0, 0.032, 0.1412, 0.0017, 0.0255, 0.1412, 0.0017, 0.0277, 0.1412, 0.0039, 0.0255, 0.1412, 0.0039, 0.0277, 0.1412, 0.0017, 0.0298,
0.1412, 0.0017, 0.032, 0.1412, 0.0039, 0.0298, 0.1412, 0.0039, 0.032, 0.1463, -0.0, 0.0255, 0.1463, -0.0, 0.0277, 0.1484, -0.0, 0.0255, 0.1484, -0.0, 0.0277, 0.1463, -0.0, 0.0298, 0.1463, -0.0, 0.032, 0.1484, -0.0, 0.0298, 0.1484, -0.0, 0.032, 0.1506, -0.0, 0.0168, 0.1506, -0.0, 0.019, 0.152, -0.0, 0.0168, 0.152, -0.0, 0.019, 0.1506, -0.0, 0.0212, 0.1506, -0.0, 0.0233, 0.152, -0.0, 0.0212, 0.152, -0.0, 0.0233, 0.152, 0.0017, 0.0168, 0.152, 0.0017, 0.019, 0.152, 0.0039, 0.0168, 0.152, 0.0039, 0.019, 0.152, 0.0017, 0.0212, 0.152, 0.0017, 0.0233, 0.152, 0.0039, 0.0212, 0.152, 0.0039, 0.0233, 0.1506, -0.0, 0.0255, 0.1506, -0.0, 0.0277, 0.152, -0.0, 0.0255,
0.152, -0.0, 0.0277, 0.1506, -0.0, 0.0298, 0.1506, -0.0, 0.032, 0.152, -0.0, 0.0298, 0.152, -0.0, 0.032, 0.152, 0.0017, 0.0255, 0.152, 0.0017, 0.0277, 0.152, 0.0039, 0.0255, 0.152, 0.0039, 0.0277, 0.152, 0.0017, 0.0298, 0.152, 0.0017, 0.032, 0.152, 0.0039, 0.0298, 0.152, 0.0039, 0.032, 0.1412, -0.0, 0.0342, 0.1412, -0.0, 0.0363, 0.1441, -0.0, 0.0342, 0.1441, -0.0, 0.0363, 0.1412, -0.0, 0.0385, 0.1412, -0.0, 0.0407, 0.1441, -0.0, 0.0385, 0.1441, -0.0, 0.0407, 0.1412, 0.0017, 0.0342, 0.1412, 0.0017, 0.0363, 0.1412, 0.0039, 0.0342, 0.1412, 0.0039, 0.0363, 0.1412, 0.0017, 0.0385, 0.1412, 0.0017, 0.0407, 0.1412, 0.0039, 0.0385, 0.1412, 0.0039, 0.0407, 0.1463, -0.0, 0.0342,
0.1463, -0.0, 0.0363, 0.1484, -0.0, 0.0342, 0.1484, -0.0, 0.0363, 0.1463, -0.0, 0.0385, 0.1463, -0.0, 0.0407, 0.1484, -0.0, 0.0385, 0.1484, -0.0, 0.0407, 0.1412, -0.0, 0.0429, 0.1412, -0.0, 0.045, 0.1441, -0.0, 0.0429, 0.1441, -0.0, 0.045, 0.1412, -0.0, 0.0472, 0.1412, -0.0, 0.0487, 0.1441, -0.0, 0.0472, 0.1441, -0.0, 0.0487, 0.1412, 0.0017, 0.0429, 0.1412, 0.0017, 0.045, 0.1412, 0.0039, 0.0429, 0.1412, 0.0039, 0.045, 0.1412, 0.0017, 0.0472, 0.1412, 0.0017, 0.0487, 0.1412, 0.0039, 0.0472, 0.1412, 0.0039, 0.0487, 0.1441, 0.0017, 0.0487, 0.1441, 0.0039, 0.0487, 0.1463, -0.0, 0.0429, 0.1463, -0.0, 0.045, 0.1484, -0.0, 0.0429, 0.1484, -0.0, 0.045, 0.1463, -0.0, 0.0472,
0.1463, -0.0, 0.0487, 0.1484, -0.0, 0.0472, 0.1484, -0.0, 0.0487, 0.1463, 0.0017, 0.0487, 0.1463, 0.0039, 0.0487, 0.1484, 0.0017, 0.0487, 0.1484, 0.0039, 0.0487, 0.1506, -0.0, 0.0342, 0.1506, -0.0, 0.0363, 0.152, -0.0, 0.0342, 0.152, -0.0, 0.0363, 0.1506, -0.0, 0.0385, 0.1506, -0.0, 0.0407, 0.152, -0.0, 0.0385, 0.152, -0.0, 0.0407, 0.152, 0.0017, 0.0342, 0.152, 0.0017, 0.0363, 0.152, 0.0039, 0.0342, 0.152, 0.0039, 0.0363, 0.152, 0.0017, 0.0385, 0.152, 0.0017, 0.0407, 0.152, 0.0039, 0.0385, 0.152, 0.0039, 0.0407, 0.1506, -0.0, 0.0429, 0.1506, -0.0, 0.045, 0.152, -0.0, 0.0429, 0.152, -0.0, 0.045, 0.1506, -0.0, 0.0472, 0.1506, -0.0, 0.0487, 0.152, -0.0, 0.0472,
0.152, -0.0, 0.0487, 0.152, 0.0017, 0.0429, 0.152, 0.0017, 0.045, 0.152, 0.0039, 0.0429, 0.152, 0.0039, 0.045, 0.1506, 0.0017, 0.0487, 0.1506, 0.0039, 0.0487, 0.152, 0.0017, 0.0472, 0.152, 0.0017, 0.0487, 0.152, 0.0039, 0.0472, 0.152, 0.0039, 0.0487, 0.2139, -0.0, -0.0006, 0.2136, -0.0, 0.0015, 0.2133, -0.0, 0.0037, 0.213, -0.0, 0.0059, 0.2139, 0.0017, -0.0006, 0.2136, 0.0017, 0.0016, 0.2139, 0.0039, -0.0006, 0.2136, 0.0039, 0.0016, 0.2133, 0.0017, 0.0038, 0.213, 0.0017, 0.006, 0.2133, 0.0039, 0.0038, 0.213, 0.0039, 0.006, 0.2157, -0.0, -0.0005, 0.2157, -0.0, 0.0016, 0.2179, -0.0, -0.0005, 0.2179, -0.0, 0.0016, 0.2157, -0.0, 0.0038, 0.2157, -0.0, 0.006, 0.2179, -0.0, 0.0038,
0.2179, -0.0, 0.006, 0.2128, -0.0, 0.0081, 0.2127, -0.0, 0.0103, 0.2127, -0.0, 0.0125, 0.2127, -0.0, 0.0146, 0.2128, 0.0017, 0.0081, 0.2127, 0.0017, 0.0103, 0.2128, 0.0039, 0.0081, 0.2127, 0.0039, 0.0103, 0.2127, 0.0017, 0.0125, 0.2127, 0.0017, 0.0146, 0.2127, 0.0039, 0.0125, 0.2127, 0.0039, 0.0146, 0.2157, -0.0, 0.0081, 0.2157, -0.0, 0.0103, 0.2179, -0.0, 0.0081, 0.2179, -0.0, 0.0103, 0.2157, -0.0, 0.0125, 0.2157, -0.0, 0.0146, 0.2179, -0.0, 0.0125, 0.2179, -0.0, 0.0146, 0.2201, -0.0, -0.0005, 0.2201, -0.0, 0.0016, 0.2222, -0.0, -0.0005, 0.2222, -0.0, 0.0016, 0.2201, -0.0, 0.0038, 0.2201, -0.0, 0.006, 0.2222, -0.0, 0.0038, 0.2222, -0.0, 0.006, 0.2244, -0.0, -0.0005,
0.2244, -0.0, 0.0016, 0.2272, -0.0, -0.0002, 0.2259, -0.0, 0.0017, 0.2252, -0.0, 0.0038, 0.2247, -0.0, 0.0061, 0.2256, -0.0, 0.0028, 0.2271, 0.0017, -0.0002, 0.226, 0.0017, 0.0016, 0.2271, 0.0039, -0.0002, 0.226, 0.0039, 0.0016, 0.2252, 0.0017, 0.0039, 0.2248, 0.0017, 0.006, 0.2252, 0.0039, 0.0039, 0.2248, 0.0039, 0.006, 0.2255, 0.0017, 0.0029, 0.2255, 0.0039, 0.0029, 0.2201, -0.0, 0.0081, 0.2201, -0.0, 0.0103, 0.2222, -0.0, 0.0081, 0.2222, -0.0, 0.0103, 0.2201, -0.0, 0.0125, 0.2201, -0.0, 0.0146, 0.2222, -0.0, 0.0125, 0.2222, -0.0, 0.0146, 0.2245, -0.0, 0.0082, 0.2245, -0.0, 0.0103, 0.2245, -0.0, 0.0125, 0.2245, -0.0, 0.0146, 0.2245, 0.0017, 0.0081, 0.2245, 0.0017, 0.0103,
0.2245, 0.0039, 0.0081, 0.2245, 0.0039, 0.0103, 0.2245, 0.0017, 0.0125, 0.2245, 0.0017, 0.0146, 0.2245, 0.0039, 0.0125, 0.2245, 0.0039, 0.0146, 0.2127, -0.0, 0.0168, 0.2127, -0.0, 0.019, 0.2127, -0.0, 0.0212, 0.2127, -0.0, 0.0233, 0.2127, 0.0017, 0.0168, 0.2127, 0.0017, 0.019, 0.2127, 0.0039, 0.0168, 0.2127, 0.0039, 0.019, 0.2127, 0.0017, 0.0212, 0.2127, 0.0017, 0.0233, 0.2127, 0.0039, 0.0212, 0.2127, 0.0039, 0.0233, 0.2157, -0.0, 0.0168, 0.2157, -0.0, 0.019, 0.2179, -0.0, 0.0168, 0.2179, -0.0, 0.019, 0.2157, -0.0, 0.0212, 0.2157, -0.0, 0.0233, 0.2179, -0.0, 0.0212, 0.2179, -0.0, 0.0233, 0.2127, -0.0, 0.0255, 0.2127, -0.0, 0.0277, 0.2127, -0.0, 0.0298, 0.2127, -0.0, 0.032,
0.2127, 0.0017, 0.0255, 0.2127, 0.0017, 0.0277, 0.2127, 0.0039, 0.0255, 0.2127, 0.0039, 0.0277, 0.2127, 0.0017, 0.0298, 0.2127, 0.0017, 0.032, 0.2127, 0.0039, 0.0298, 0.2127, 0.0039, 0.032, 0.2157, -0.0, 0.0255, 0.2157, -0.0, 0.0277, 0.2179, -0.0, 0.0255, 0.2179, -0.0, 0.0277, 0.2157, -0.0, 0.0298, 0.2157, -0.0, 0.032, 0.2179, -0.0, 0.0298, 0.2179, -0.0, 0.032, 0.2201, -0.0, 0.0168, 0.2201, -0.0, 0.019, 0.2222, -0.0, 0.0168, 0.2222, -0.0, 0.019, 0.2201, -0.0, 0.0212, 0.2201, -0.0, 0.0233, 0.2222, -0.0, 0.0212, 0.2222, -0.0, 0.0233, 0.2245, -0.0, 0.0168, 0.2245, -0.0, 0.019, 0.2245, -0.0, 0.0212, 0.2245, -0.0, 0.0233, 0.2245, 0.0017, 0.0168, 0.2245, 0.0017, 0.019,
0.2245, 0.0039, 0.0168, 0.2245, 0.0039, 0.019, 0.2245, 0.0017, 0.0212, 0.2245, 0.0017, 0.0233, 0.2245, 0.0039, 0.0212, 0.2245, 0.0039, 0.0233, 0.2201, -0.0, 0.0255, 0.2201, -0.0, 0.0277, 0.2222, -0.0, 0.0255, 0.2222, -0.0, 0.0277, 0.2201, -0.0, 0.0298, 0.2201, -0.0, 0.032, 0.2222, -0.0, 0.0298, 0.2222, -0.0, 0.032, 0.2245, -0.0, 0.0255, 0.2245, -0.0, 0.0277, 0.2245, -0.0, 0.0298, 0.2245, -0.0, 0.032, 0.2245, 0.0017, 0.0255, 0.2245, 0.0017, 0.0277, 0.2245, 0.0039, 0.0255, 0.2245, 0.0039, 0.0277, 0.2245, 0.0017, 0.0298, 0.2245, 0.0017, 0.032, 0.2245, 0.0039, 0.0298, 0.2245, 0.0039, 0.032, 0.2279, -0.0, -0.0014, 0.2279, 0.0017, -0.0013, 0.2279, 0.0039, -0.0013, 0.2127, -0.0, 0.0342,
0.2127, -0.0, 0.0363, 0.2127, -0.0, 0.0385, 0.2127, -0.0, 0.0407, 0.2127, 0.0017, 0.0342, 0.2127, 0.0017, 0.0363, 0.2127, 0.0039, 0.0342, 0.2127, 0.0039, 0.0363, 0.2127, 0.0017, 0.0385, 0.2127, 0.0017, 0.0407, 0.2127, 0.0039, 0.0385, 0.2127, 0.0039, 0.0407, 0.2157, -0.0, 0.0342, 0.2157, -0.0, 0.0363, 0.2179, -0.0, 0.0342, 0.2179, -0.0, 0.0363, 0.2157, -0.0, 0.0385, 0.2157, -0.0, 0.0407, 0.2179, -0.0, 0.0385, 0.2179, -0.0, 0.0407, 0.2127, -0.0, 0.0429, 0.2127, -0.0, 0.045, 0.2127, -0.0, 0.0472, 0.2127, -0.0, 0.0487, 0.2127, 0.0017, 0.0429, 0.2127, 0.0017, 0.045, 0.2127, 0.0039, 0.0429, 0.2127, 0.0039, 0.045, 0.2127, 0.0017, 0.0472, 0.2127, 0.0017, 0.0487, 0.2127, 0.0039, 0.0472,
0.2127, 0.0039, 0.0487, 0.2157, -0.0, 0.0429, 0.2157, -0.0, 0.045, 0.2179, -0.0, 0.0429, 0.2179, -0.0, 0.045, 0.2157, -0.0, 0.0472, 0.2157, -0.0, 0.0487, 0.2179, -0.0, 0.0472, 0.2179, -0.0, 0.0487, 0.2157, 0.0017, 0.0487, 0.2157, 0.0039, 0.0487, 0.2179, 0.0017, 0.0487, 0.2179, 0.0039, 0.0487, 0.2201, -0.0, 0.0342, 0.2201, -0.0, 0.0363, 0.2222, -0.0, 0.0342, 0.2222, -0.0, 0.0363, 0.2201, -0.0, 0.0385, 0.2201, -0.0, 0.0407, 0.2222, -0.0, 0.0385, 0.2222, -0.0, 0.0407, 0.2245, -0.0, 0.0342, 0.2245, -0.0, 0.0363, 0.2245, -0.0, 0.0385, 0.2245, -0.0, 0.0407, 0.2245, 0.0017, 0.0342, 0.2245, 0.0017, 0.0363, 0.2245, 0.0039, 0.0342, 0.2245, 0.0039, 0.0363, 0.2245, 0.0017, 0.0385,
0.2245, 0.0017, 0.0407, 0.2245, 0.0039, 0.0385, 0.2245, 0.0039, 0.0407, 0.2201, -0.0, 0.0429, 0.2201, -0.0, 0.045, 0.2222, -0.0, 0.0429, 0.2222, -0.0, 0.045, 0.2201, -0.0, 0.0472, 0.2201, -0.0, 0.0487, 0.2222, -0.0, 0.0472, 0.2222, -0.0, 0.0487, 0.2201, 0.0017, 0.0487, 0.2201, 0.0039, 0.0487, 0.2222, 0.0017, 0.0487, 0.2222, 0.0039, 0.0487, 0.2245, -0.0, 0.0429, 0.2245, -0.0, 0.045, 0.2245, -0.0, 0.0472, 0.2245, -0.0, 0.0487, 0.2245, 0.0017, 0.0429, 0.2245, 0.0017, 0.045, 0.2245, 0.0039, 0.0429, 0.2245, 0.0039, 0.045, 0.2245, 0.0017, 0.0472, 0.2245, 0.0017, 0.0487, 0.2245, 0.0039, 0.0472, 0.2245, 0.0039, 0.0487, 0.2765, -0.0, -0.0003, 0.2772, -0.0, 0.0011, 0.2787, -0.0, -0.0005,
0.2776, -0.0, 0.002, 0.2783, -0.0, 0.004, 0.2787, -0.0, 0.0061, 0.2763, 0.0017, -0.0006, 0.2773, 0.0017, 0.0012, 0.2763, 0.0039, -0.0006, 0.2773, 0.0039, 0.0012, 0.2777, 0.0017, 0.0023, 0.2777, 0.0039, 0.0023, 0.2782, 0.0017, 0.0038, 0.2787, 0.0017, 0.006, 0.2782, 0.0039, 0.0038, 0.2787, 0.0039, 0.006, 0.2791, -0.0, 0.0082, 0.2793, -0.0, 0.0103, 0.2794, -0.0, 0.0125, 0.2794, -0.0, 0.0146, 0.2791, 0.0017, 0.0081, 0.2793, 0.0017, 0.0103, 0.2791, 0.0039, 0.0081, 0.2793, 0.0039, 0.0103, 0.2794, 0.0017, 0.0125, 0.2794, 0.0017, 0.0146, 0.2794, 0.0039, 0.0125, 0.2794, 0.0039, 0.0146, 0.2794, -0.0, 0.0168, 0.2794, -0.0, 0.019, 0.2794, -0.0, 0.0212, 0.2794, -0.0, 0.0233, 0.2794, 0.0017, 0.0168,
0.2794, 0.0017, 0.019, 0.2794, 0.0039, 0.0168, 0.2794, 0.0039, 0.019, 0.2794, 0.0017, 0.0212, 0.2794, 0.0017, 0.0233, 0.2794, 0.0039, 0.0212, 0.2794, 0.0039, 0.0233, 0.2794, -0.0, 0.0255, 0.2794, -0.0, 0.0277, 0.2794, -0.0, 0.0298, 0.2794, -0.0, 0.032, 0.2794, 0.0017, 0.0255, 0.2794, 0.0017, 0.0277, 0.2794, 0.0039, 0.0255, 0.2794, 0.0039, 0.0277, 0.2794, 0.0017, 0.0298, 0.2794, 0.0017, 0.032, 0.2794, 0.0039, 0.0298, 0.2794, 0.0039, 0.032, 0.2794, -0.0, 0.0342, 0.2794, -0.0, 0.0363, 0.2794, -0.0, 0.0385, 0.2794, -0.0, 0.0407, 0.2794, 0.0017, 0.0342, 0.2794, 0.0017, 0.0363, 0.2794, 0.0039, 0.0342, 0.2794, 0.0039, 0.0363, 0.2794, 0.0017, 0.0385, 0.2794, 0.0017, 0.0407, 0.2794, 0.0039, 0.0385,
0.2794, 0.0039, 0.0407, 0.2794, -0.0, 0.0429, 0.2794, -0.0, 0.045, 0.2794, -0.0, 0.0472, 0.2794, -0.0, 0.0487, 0.2794, 0.0017, 0.0429, 0.2794, 0.0017, 0.045, 0.2794, 0.0039, 0.0429, 0.2794, 0.0039, 0.045, 0.2794, 0.0017, 0.0472, 0.2794, 0.0017, 0.0487, 0.2794, 0.0039, 0.0472, 0.2794, 0.0039, 0.0487, 0.2808, -0.0, -0.0005, 0.2808, -0.0, 0.0016, 0.283, -0.0, -0.0005, 0.283, -0.0, 0.0016, 0.2808, -0.0, 0.0038, 0.2808, -0.0, 0.006, 0.283, -0.0, 0.0038, 0.283, -0.0, 0.006, 0.2852, -0.0, -0.0005, 0.2852, -0.0, 0.0016, 0.2873, -0.0, -0.0005, 0.2873, -0.0, 0.0016, 0.2852, -0.0, 0.0038, 0.2852, -0.0, 0.006, 0.2873, -0.0, 0.0038, 0.2873, -0.0, 0.006, 0.2808, -0.0, 0.0081,
0.2808, -0.0, 0.0103, 0.283, -0.0, 0.0081, 0.283, -0.0, 0.0103, 0.2808, -0.0, 0.0125, 0.2808, -0.0, 0.0146, 0.283, -0.0, 0.0125, 0.283, -0.0, 0.0146, 0.2852, -0.0, 0.0081, 0.2852, -0.0, 0.0103, 0.2873, -0.0, 0.0081, 0.2873, -0.0, 0.0103, 0.2852, -0.0, 0.0125, 0.2852, -0.0, 0.0146, 0.2873, -0.0, 0.0125, 0.2873, -0.0, 0.0146, 0.2897, -0.0, -0.0007, 0.2901, -0.0, 0.0015, 0.2903, -0.0, 0.0037, 0.2906, -0.0, 0.0057, 0.2908, -0.0, 0.0065, 0.2897, 0.0017, -0.0006, 0.2901, 0.0017, 0.0016, 0.2897, 0.0039, -0.0006, 0.2901, 0.0039, 0.0016, 0.2903, 0.0017, 0.0038, 0.2906, 0.0017, 0.0051, 0.2903, 0.0039, 0.0038, 0.2906, 0.0039, 0.0051, 0.2907, 0.0017, 0.0062, 0.2907, 0.0039, 0.0062,
0.2895, -0.0, 0.0081, 0.2895, -0.0, 0.0103, 0.291, -0.0, 0.0081, 0.2911, -0.0, 0.0103, 0.2895, -0.0, 0.0125, 0.2895, -0.0, 0.0146, 0.2912, -0.0, 0.0125, 0.2912, -0.0, 0.0146, 0.291, 0.0017, 0.0081, 0.2911, 0.0017, 0.0103, 0.291, 0.0039, 0.0081, 0.2911, 0.0039, 0.0103, 0.2912, 0.0017, 0.0125, 0.2912, 0.0017, 0.0146, 0.2912, 0.0039, 0.0125, 0.2912, 0.0039, 0.0146, 0.2808, -0.0, 0.0168, 0.2808, -0.0, 0.019, 0.283, -0.0, 0.0168, 0.283, -0.0, 0.019, 0.2808, -0.0, 0.0212, 0.2808, -0.0, 0.0233, 0.283, -0.0, 0.0212, 0.283, -0.0, 0.0233, 0.2852, -0.0, 0.0168, 0.2852, -0.0, 0.019, 0.2873, -0.0, 0.0168, 0.2873, -0.0, 0.019, 0.2852, -0.0, 0.0212, 0.2852, -0.0, 0.0233,
0.2873, -0.0, 0.0212, 0.2873, -0.0, 0.0233, 0.2808, -0.0, 0.0255, 0.2808, -0.0, 0.0277, 0.283, -0.0, 0.0255, 0.283, -0.0, 0.0277, 0.2808, -0.0, 0.0298, 0.2808, -0.0, 0.032, 0.283, -0.0, 0.0298, 0.283, -0.0, 0.032, 0.2852, -0.0, 0.0255, 0.2852, -0.0, 0.0277, 0.2873, -0.0, 0.0255, 0.2873, -0.0, 0.0277, 0.2852, -0.0, 0.0298, 0.2852, -0.0, 0.032, 0.2873, -0.0, 0.0298, 0.2873, -0.0, 0.032, 0.2895, -0.0, 0.0168, 0.2895, -0.0, 0.019, 0.2912, -0.0, 0.0168, 0.2912, -0.0, 0.019, 0.2895, -0.0, 0.0212, 0.2895, -0.0, 0.0233, 0.2912, -0.0, 0.0212, 0.2912, -0.0, 0.0233, 0.2912, 0.0017, 0.0168, 0.2912, 0.0017, 0.019, 0.2912, 0.0039, 0.0168, 0.2912, 0.0039, 0.019,
0.2912, 0.0017, 0.0212, 0.2912, 0.0017, 0.0233, 0.2912, 0.0039, 0.0212, 0.2912, 0.0039, 0.0233, 0.2895, -0.0, 0.0255, 0.2895, -0.0, 0.0277, 0.2912, -0.0, 0.0255, 0.2912, -0.0, 0.0277, 0.2895, -0.0, 0.0298, 0.2895, -0.0, 0.032, 0.2912, -0.0, 0.0298, 0.2912, -0.0, 0.032, 0.2912, 0.0017, 0.0255, 0.2912, 0.0017, 0.0277, 0.2912, 0.0039, 0.0255, 0.2912, 0.0039, 0.0277, 0.2912, 0.0017, 0.0298, 0.2912, 0.0017, 0.032, 0.2912, 0.0039, 0.0298, 0.2912, 0.0039, 0.032, 0.3088, -0.0, -0.0005, 0.3088, -0.0, 0.0016, 0.3088, -0.0, 0.0038, 0.3088, -0.0, 0.006, 0.3088, 0.0017, -0.0005, 0.3088, 0.0017, 0.0016, 0.3088, 0.0039, -0.0005, 0.3088, 0.0039, 0.0016, 0.3088, 0.0017, 0.0038, 0.3088, 0.0017, 0.006,
0.3088, 0.0039, 0.0038, 0.3088, 0.0039, 0.006, 0.3112, -0.0, -0.0005, 0.3112, -0.0, 0.0016, 0.3134, -0.0, -0.0005, 0.3134, -0.0, 0.0016, 0.3112, -0.0, 0.0038, 0.3112, -0.0, 0.006, 0.3134, -0.0, 0.0038, 0.3134, -0.0, 0.006, 0.3088, -0.0, 0.0081, 0.3088, -0.0, 0.0103, 0.3088, -0.0, 0.0125, 0.3088, -0.0, 0.0146, 0.3088, 0.0017, 0.0081, 0.3088, 0.0017, 0.0103, 0.3088, 0.0039, 0.0081, 0.3088, 0.0039, 0.0103, 0.3088, 0.0017, 0.0125, 0.3088, 0.0017, 0.0146, 0.3088, 0.0039, 0.0125, 0.3088, 0.0039, 0.0146, 0.3112, -0.0, 0.0081, 0.3112, -0.0, 0.0103, 0.3134, -0.0, 0.0081, 0.3134, -0.0, 0.0103, 0.3112, -0.0, 0.0125, 0.3112, -0.0, 0.0146, 0.3134, -0.0, 0.0125, 0.3134, -0.0, 0.0146,
0.3088, -0.0, 0.0168, 0.3088, -0.0, 0.019, 0.3088, -0.0, 0.0212, 0.3088, -0.0, 0.0233, 0.3088, 0.0017, 0.0168, 0.3088, 0.0017, 0.019, 0.3088, 0.0039, 0.0168, 0.3088, 0.0039, 0.019, 0.3088, 0.0017, 0.0212, 0.3088, 0.0017, 0.0233, 0.3088, 0.0039, 0.0212, 0.3088, 0.0039, 0.0233, 0.3112, -0.0, 0.0168, 0.3112, -0.0, 0.019, 0.3134, -0.0, 0.0168, 0.3134, -0.0, 0.019, 0.3112, -0.0, 0.0212, 0.3112, -0.0, 0.0233, 0.3134, -0.0, 0.0212, 0.3134, -0.0, 0.0233, 0.3088, -0.0, 0.0255, 0.3088, -0.0, 0.0277, 0.3088, -0.0, 0.0298, 0.3088, -0.0, 0.032, 0.3088, 0.0017, 0.0255, 0.3088, 0.0017, 0.0277, 0.3088, 0.0039, 0.0255, 0.3088, 0.0039, 0.0277, 0.3088, 0.0017, 0.0298, 0.3088, 0.0017, 0.032,
0.3088, 0.0039, 0.0298, 0.3088, 0.0039, 0.032, 0.3112, -0.0, 0.0255, 0.3112, -0.0, 0.0277, 0.3134, -0.0, 0.0255, 0.3134, -0.0, 0.0277, 0.3112, -0.0, 0.0298, 0.3112, -0.0, 0.032, 0.3134, -0.0, 0.0298, 0.3134, -0.0, 0.032, 0.2808, -0.0, 0.0342, 0.2808, -0.0, 0.0363, 0.283, -0.0, 0.0342, 0.283, -0.0, 0.0363, 0.2808, -0.0, 0.0385, 0.2808, -0.0, 0.0407, 0.283, -0.0, 0.0385, 0.283, -0.0, 0.0407, 0.2852, -0.0, 0.0342, 0.2852, -0.0, 0.0363, 0.2873, -0.0, 0.0342, 0.2873, -0.0, 0.0363, 0.2852, -0.0, 0.0385, 0.2852, -0.0, 0.0407, 0.2873, -0.0, 0.0385, 0.2873, -0.0, 0.0407, 0.2808, -0.0, 0.0429, 0.2808, -0.0, 0.045, 0.283, -0.0, 0.0429, 0.283, -0.0, 0.045,
0.2808, -0.0, 0.0472, 0.2808, -0.0, 0.0487, 0.283, -0.0, 0.0472, 0.283, -0.0, 0.0487, 0.2808, 0.0017, 0.0487, 0.2808, 0.0039, 0.0487, 0.283, 0.0017, 0.0487, 0.283, 0.0039, 0.0487, 0.2852, -0.0, 0.0429, 0.2852, -0.0, 0.045, 0.2873, -0.0, 0.0429, 0.2873, -0.0, 0.045, 0.2852, -0.0, 0.0472, 0.2852, -0.0, 0.0487, 0.2873, -0.0, 0.0472, 0.2873, -0.0, 0.0487, 0.2852, 0.0017, 0.0487, 0.2852, 0.0039, 0.0487, 0.2873, 0.0017, 0.0487, 0.2873, 0.0039, 0.0487, 0.2895, -0.0, 0.0342, 0.2895, -0.0, 0.0363, 0.2912, -0.0, 0.0342, 0.2912, -0.0, 0.0363, 0.2895, -0.0, 0.0385, 0.2895, -0.0, 0.0407, 0.2912, -0.0, 0.0385, 0.2912, -0.0, 0.0407, 0.2912, 0.0017, 0.0342, 0.2912, 0.0017, 0.0363,
0.2912, 0.0039, 0.0342, 0.2912, 0.0039, 0.0363, 0.2912, 0.0017, 0.0385, 0.2912, 0.0017, 0.0407, 0.2912, 0.0039, 0.0385, 0.2912, 0.0039, 0.0407, 0.2895, -0.0, 0.0429, 0.2895, -0.0, 0.045, 0.2912, -0.0, 0.0429, 0.2912, -0.0, 0.045, 0.2895, -0.0, 0.0472, 0.2895, -0.0, 0.0487, 0.2912, -0.0, 0.0472, 0.2912, -0.0, 0.0487, 0.2912, 0.0017, 0.0429, 0.2912, 0.0017, 0.045, 0.2912, 0.0039, 0.0429, 0.2912, 0.0039, 0.045, 0.2895, 0.0017, 0.0487, 0.2895, 0.0039, 0.0487, 0.2912, 0.0017, 0.0472, 0.2912, 0.0017, 0.0487, 0.2912, 0.0039, 0.0472, 0.2912, 0.0039, 0.0487, 0.3088, -0.0, 0.0342, 0.3088, -0.0, 0.0363, 0.3088, -0.0, 0.0385, 0.3088, -0.0, 0.0407, 0.3088, 0.0017, 0.0342, 0.3088, 0.0017, 0.0363,
0.3088, 0.0039, 0.0342, 0.3088, 0.0039, 0.0363, 0.3088, 0.0017, 0.0385, 0.3088, 0.0017, 0.0407, 0.3088, 0.0039, 0.0385, 0.3088, 0.0039, 0.0407, 0.3112, -0.0, 0.0342, 0.3112, -0.0, 0.0363, 0.3134, -0.0, 0.0342, 0.3134, -0.0, 0.0363, 0.3112, -0.0, 0.0385, 0.3112, -0.0, 0.0407, 0.3134, -0.0, 0.0385, 0.3134, -0.0, 0.0407, 0.3088, -0.0, 0.0429, 0.3088, -0.0, 0.045, 0.3088, -0.0, 0.0472, 0.3088, -0.0, 0.0487, 0.3088, 0.0017, 0.0429, 0.3088, 0.0017, 0.045, 0.3088, 0.0039, 0.0429, 0.3088, 0.0039, 0.045, 0.3088, 0.0017, 0.0472, 0.3088, 0.0017, 0.0487, 0.3088, 0.0039, 0.0472, 0.3088, 0.0039, 0.0487, 0.3112, -0.0, 0.0429, 0.3112, -0.0, 0.045, 0.3134, -0.0, 0.0429, 0.3134, -0.0, 0.045,
0.3112, -0.0, 0.0472, 0.3112, -0.0, 0.0487, 0.3134, -0.0, 0.0472, 0.3134, -0.0, 0.0487, 0.3112, 0.0017, 0.0487, 0.3112, 0.0039, 0.0487, 0.3134, 0.0017, 0.0487, 0.3134, 0.0039, 0.0487, 0.3155, -0.0, -0.0005, 0.3155, -0.0, 0.0016, 0.3177, -0.0, -0.0005, 0.3177, -0.0, 0.0016, 0.3155, -0.0, 0.0038, 0.3155, -0.0, 0.006, 0.3177, -0.0, 0.0038, 0.3177, -0.0, 0.006, 0.3206, -0.0, -0.0005, 0.3206, -0.0, 0.0016, 0.3206, -0.0, 0.0038, 0.3206, -0.0, 0.006, 0.3206, 0.0017, -0.0005, 0.3206, 0.0017, 0.0016, 0.3206, 0.0039, -0.0005, 0.3206, 0.0039, 0.0016, 0.3206, 0.0017, 0.0038, 0.3206, 0.0017, 0.006, 0.3206, 0.0039, 0.0038, 0.3206, 0.0039, 0.006, 0.3155, -0.0, 0.0081, 0.3155, -0.0, 0.0103,
0.3177, -0.0, 0.0081, 0.3177, -0.0, 0.0103, 0.3155, -0.0, 0.0125, 0.3155, -0.0, 0.0146, 0.3177, -0.0, 0.0125, 0.3177, -0.0, 0.0146, 0.3206, -0.0, 0.0081, 0.3206, -0.0, 0.0103, 0.3206, -0.0, 0.0125, 0.3206, -0.0, 0.0146, 0.3206, 0.0017, 0.0081, 0.3206, 0.0017, 0.0103, 0.3206, 0.0039, 0.0081, 0.3206, 0.0039, 0.0103, 0.3206, 0.0017, 0.0125, 0.3206, 0.0017, 0.0146, 0.3206, 0.0039, 0.0125, 0.3206, 0.0039, 0.0146, 0.3155, -0.0, 0.0168, 0.3155, -0.0, 0.019, 0.3177, -0.0, 0.0168, 0.3177, -0.0, 0.019, 0.3155, -0.0, 0.0212, 0.3155, -0.0, 0.0233, 0.3177, -0.0, 0.0212, 0.3177, -0.0, 0.0233, 0.3206, -0.0, 0.0168, 0.3206, -0.0, 0.019, 0.3206, -0.0, 0.0212, 0.3206, -0.0, 0.0233,
0.3206, 0.0017, 0.0168, 0.3206, 0.0017, 0.019, 0.3206, 0.0039, 0.0168, 0.3206, 0.0039, 0.019, 0.3206, 0.0017, 0.0212, 0.3206, 0.0017, 0.0233, 0.3206, 0.0039, 0.0212, 0.3206, 0.0039, 0.0233, 0.3155, -0.0, 0.0255, 0.3155, -0.0, 0.0277, 0.3177, -0.0, 0.0255, 0.3177, -0.0, 0.0277, 0.3155, -0.0, 0.0298, 0.3155, -0.0, 0.032, 0.3177, -0.0, 0.0298, 0.3177, -0.0, 0.032, 0.3206, -0.0, 0.0255, 0.3206, -0.0, 0.0277, 0.3206, -0.0, 0.0298, 0.3206, -0.0, 0.032, 0.3206, 0.0017, 0.0255, 0.3206, 0.0017, 0.0277, 0.3206, 0.0039, 0.0255, 0.3206, 0.0039, 0.0277, 0.3206, 0.0017, 0.0298, 0.3206, 0.0017, 0.032, 0.3206, 0.0039, 0.0298, 0.3206, 0.0039, 0.032, 0.3402, -0.0, -0.0005, 0.3402, -0.0, 0.0016,
0.3402, -0.0, 0.0038, 0.3402, -0.0, 0.006, 0.3402, 0.0017, -0.0005, 0.3402, 0.0017, 0.0016, 0.3402, 0.0039, -0.0005, 0.3402, 0.0039, 0.0016, 0.3402, 0.0017, 0.0038, 0.3402, 0.0017, 0.006, 0.3402, 0.0039, 0.0038, 0.3402, 0.0039, 0.006, 0.3402, -0.0, 0.0081, 0.3402, -0.0, 0.0103, 0.3402, -0.0, 0.0125, 0.3402, -0.0, 0.0146, 0.3402, 0.0017, 0.0081, 0.3402, 0.0017, 0.0103, 0.3402, 0.0039, 0.0081, 0.3402, 0.0039, 0.0103, 0.3402, 0.0017, 0.0125, 0.3402, 0.0017, 0.0146, 0.3402, 0.0039, 0.0125, 0.3402, 0.0039, 0.0146, 0.3416, -0.0, -0.0005, 0.3416, -0.0, 0.0016, 0.3438, -0.0, -0.0005, 0.3438, -0.0, 0.0016, 0.3416, -0.0, 0.0038, 0.3416, -0.0, 0.006, 0.3438, -0.0, 0.0038, 0.3438, -0.0, 0.006,
0.3459, -0.0, -0.0005, 0.3459, -0.0, 0.0016, 0.3481, -0.0, -0.0005, 0.3481, -0.0, 0.0016, 0.3459, -0.0, 0.0038, 0.3459, -0.0, 0.006, 0.3481, -0.0, 0.0038, 0.3481, -0.0, 0.006, 0.3416, -0.0, 0.0081, 0.3416, -0.0, 0.0103, 0.3438, -0.0, 0.0081, 0.3438, -0.0, 0.0103, 0.3416, -0.0, 0.0125, 0.3416, -0.0, 0.0146, 0.3438, -0.0, 0.0125, 0.3438, -0.0, 0.0146, 0.3459, -0.0, 0.0081, 0.3459, -0.0, 0.0103, 0.3481, -0.0, 0.0081, 0.3481, -0.0, 0.0103, 0.3459, -0.0, 0.0125, 0.3459, -0.0, 0.0146, 0.3481, -0.0, 0.0125, 0.3481, -0.0, 0.0146, 0.3402, -0.0, 0.0168, 0.3402, -0.0, 0.019, 0.3402, -0.0, 0.0212, 0.3402, -0.0, 0.0233, 0.3402, 0.0017, 0.0168, 0.3402, 0.0017, 0.019,
0.3402, 0.0039, 0.0168, 0.3402, 0.0039, 0.019, 0.3402, 0.0017, 0.0212, 0.3402, 0.0017, 0.0233, 0.3402, 0.0039, 0.0212, 0.3402, 0.0039, 0.0233, 0.3402, -0.0, 0.0255, 0.3402, -0.0, 0.0277, 0.3402, -0.0, 0.0298, 0.3402, -0.0, 0.032, 0.3402, 0.0017, 0.0255, 0.3402, 0.0017, 0.0277, 0.3402, 0.0039, 0.0255, 0.3402, 0.0039, 0.0277, 0.3402, 0.0017, 0.0298, 0.3402, 0.0017, 0.032, 0.3402, 0.0039, 0.0298, 0.3402, 0.0039, 0.032, 0.3416, -0.0, 0.0168, 0.3416, -0.0, 0.019, 0.3438, -0.0, 0.0168, 0.3438, -0.0, 0.019, 0.3416, -0.0, 0.0212, 0.3416, -0.0, 0.0233, 0.3438, -0.0, 0.0212, 0.3438, -0.0, 0.0233, 0.3459, -0.0, 0.0168, 0.3459, -0.0, 0.019, 0.3481, -0.0, 0.0168, 0.3481, -0.0, 0.019,
0.3459, -0.0, 0.0212, 0.3459, -0.0, 0.0233, 0.3481, -0.0, 0.0212, 0.3481, -0.0, 0.0233, 0.3416, -0.0, 0.0255, 0.3416, -0.0, 0.0277, 0.3438, -0.0, 0.0255, 0.3438, -0.0, 0.0277, 0.3416, -0.0, 0.0298, 0.3416, -0.0, 0.032, 0.3438, -0.0, 0.0298, 0.3438, -0.0, 0.032, 0.3459, -0.0, 0.0255, 0.3459, -0.0, 0.0277, 0.3481, -0.0, 0.0255, 0.3481, -0.0, 0.0277, 0.3459, -0.0, 0.0298, 0.3459, -0.0, 0.032, 0.3481, -0.0, 0.0298, 0.3481, -0.0, 0.032, 0.3155, -0.0, 0.0342, 0.3155, -0.0, 0.0363, 0.3177, -0.0, 0.0342, 0.3177, -0.0, 0.0363, 0.3155, -0.0, 0.0385, 0.3155, -0.0, 0.0407, 0.3177, -0.0, 0.0385, 0.3177, -0.0, 0.0407, 0.3206, -0.0, 0.0342, 0.3206, -0.0, 0.0363,
0.3206, -0.0, 0.0385, 0.3206, -0.0, 0.0407, 0.3206, 0.0017, 0.0342, 0.3206, 0.0017, 0.0363, 0.3206, 0.0039, 0.0342, 0.3206, 0.0039, 0.0363, 0.3206, 0.0017, 0.0385, 0.3206, 0.0017, 0.0407, 0.3206, 0.0039, 0.0385, 0.3206, 0.0039, 0.0407, 0.3155, -0.0, 0.0429, 0.3155, -0.0, 0.045, 0.3177, -0.0, 0.0429, 0.3177, -0.0, 0.045, 0.3155, -0.0, 0.0472, 0.3155, -0.0, 0.0487, 0.3177, -0.0, 0.0472, 0.3177, -0.0, 0.0487, 0.3155, 0.0017, 0.0487, 0.3155, 0.0039, 0.0487, 0.3177, 0.0017, 0.0487, 0.3177, 0.0039, 0.0487, 0.3206, -0.0, 0.0429, 0.3206, -0.0, 0.045, 0.3206, -0.0, 0.0472, 0.3206, -0.0, 0.0487, 0.3206, 0.0017, 0.0429, 0.3206, 0.0017, 0.045, 0.3206, 0.0039, 0.0429, 0.3206, 0.0039, 0.045,
0.3206, 0.0017, 0.0472, 0.3206, 0.0017, 0.0487, 0.3206, 0.0039, 0.0472, 0.3206, 0.0039, 0.0487, 0.3402, -0.0, 0.0342, 0.3402, -0.0, 0.0363, 0.3402, -0.0, 0.0385, 0.3402, -0.0, 0.0407, 0.3402, 0.0017, 0.0342, 0.3402, 0.0017, 0.0363, 0.3402, 0.0039, 0.0342, 0.3402, 0.0039, 0.0363, 0.3402, 0.0017, 0.0385, 0.3402, 0.0017, 0.0407, 0.3402, 0.0039, 0.0385, 0.3402, 0.0039, 0.0407, 0.3402, -0.0, 0.0429, 0.3402, -0.0, 0.045, 0.3402, -0.0, 0.0472, 0.3402, -0.0, 0.0487, 0.3402, 0.0017, 0.0429, 0.3402, 0.0017, 0.045, 0.3402, 0.0039, 0.0429, 0.3402, 0.0039, 0.045, 0.3402, 0.0017, 0.0472, 0.3402, 0.0017, 0.0487, 0.3402, 0.0039, 0.0472, 0.3402, 0.0039, 0.0487, 0.3416, -0.0, 0.0342, 0.3416, -0.0, 0.0363,
0.3438, -0.0, 0.0342, 0.3438, -0.0, 0.0363, 0.3416, -0.0, 0.0385, 0.3416, -0.0, 0.0407, 0.3438, -0.0, 0.0385, 0.3438, -0.0, 0.0407, 0.3459, -0.0, 0.0342, 0.3459, -0.0, 0.0363, 0.3481, -0.0, 0.0342, 0.3481, -0.0, 0.0363, 0.3459, -0.0, 0.0385, 0.3459, -0.0, 0.0407, 0.3481, -0.0, 0.0385, 0.3481, -0.0, 0.0407, 0.3416, -0.0, 0.0429, 0.3416, -0.0, 0.045, 0.3438, -0.0, 0.0429, 0.3438, -0.0, 0.045, 0.3416, -0.0, 0.0472, 0.3416, -0.0, 0.0487, 0.3438, -0.0, 0.0472, 0.3438, -0.0, 0.0487, 0.3416, 0.0017, 0.0487, 0.3416, 0.0039, 0.0487, 0.3438, 0.0017, 0.0487, 0.3438, 0.0039, 0.0487, 0.3459, -0.0, 0.0429, 0.3459, -0.0, 0.045, 0.3481, -0.0, 0.0429, 0.3481, -0.0, 0.045,
0.3459, -0.0, 0.0472, 0.3459, -0.0, 0.0487, 0.3481, -0.0, 0.0472, 0.3481, -0.0, 0.0487, 0.3459, 0.0017, 0.0487, 0.3459, 0.0039, 0.0487, 0.3481, 0.0017, 0.0487, 0.3481, 0.0039, 0.0487, 0.3503, -0.0, -0.0005, 0.3503, -0.0, 0.0016, 0.352, -0.0, -0.0005, 0.352, -0.0, 0.0016, 0.3503, -0.0, 0.0038, 0.3503, -0.0, 0.006, 0.352, -0.0, 0.0038, 0.352, -0.0, 0.006, 0.352, 0.0017, -0.0005, 0.352, 0.0017, 0.0016, 0.352, 0.0039, -0.0005, 0.352, 0.0039, 0.0016, 0.352, 0.0017, 0.0038, 0.352, 0.0017, 0.006, 0.352, 0.0039, 0.0038, 0.352, 0.0039, 0.006, 0.3503, -0.0, 0.0081, 0.3503, -0.0, 0.0103, 0.352, -0.0, 0.0081, 0.352, -0.0, 0.0103, 0.3503, -0.0, 0.0125, 0.3503, -0.0, 0.0146,
0.352, -0.0, 0.0125, 0.352, -0.0, 0.0146, 0.352, 0.0017, 0.0081, 0.352, 0.0017, 0.0103, 0.352, 0.0039, 0.0081, 0.352, 0.0039, 0.0103, 0.352, 0.0017, 0.0125, 0.352, 0.0017, 0.0146, 0.352, 0.0039, 0.0125, 0.352, 0.0039, 0.0146, 0.3503, -0.0, 0.0168, 0.3503, -0.0, 0.019, 0.352, -0.0, 0.0168, 0.352, -0.0, 0.019, 0.3503, -0.0, 0.0212, 0.3503, -0.0, 0.0233, 0.352, -0.0, 0.0212, 0.352, -0.0, 0.0233, 0.352, 0.0017, 0.0168, 0.352, 0.0017, 0.019, 0.352, 0.0039, 0.0168, 0.352, 0.0039, 0.019, 0.352, 0.0017, 0.0212, 0.352, 0.0017, 0.0233, 0.352, 0.0039, 0.0212, 0.352, 0.0039, 0.0233, 0.3503, -0.0, 0.0255, 0.3503, -0.0, 0.0277, 0.352, -0.0, 0.0255, 0.352, -0.0, 0.0277,
0.3503, -0.0, 0.0298, 0.3503, -0.0, 0.032, 0.352, -0.0, 0.0298, 0.352, -0.0, 0.032, 0.352, 0.0017, 0.0255, 0.352, 0.0017, 0.0277, 0.352, 0.0039, 0.0255, 0.352, 0.0039, 0.0277, 0.352, 0.0017, 0.0298, 0.352, 0.0017, 0.032, 0.352, 0.0039, 0.0298, 0.352, 0.0039, 0.032, 0.3503, -0.0, 0.0342, 0.3503, -0.0, 0.0363, 0.352, -0.0, 0.0342, 0.352, -0.0, 0.0363, 0.3503, -0.0, 0.0385, 0.3503, -0.0, 0.0407, 0.352, -0.0, 0.0379, 0.3524, -0.0, 0.0407, 0.352, 0.0017, 0.0342, 0.352, 0.0017, 0.0363, 0.352, 0.0039, 0.0342, 0.352, 0.0039, 0.0363, 0.352, 0.0017, 0.0379, 0.352, 0.0039, 0.0379, 0.3546, -0.0, 0.0379, 0.3546, -0.0, 0.0407, 0.3568, -0.0, 0.0379, 0.3568, -0.0, 0.0407,
0.3546, 0.0017, 0.0379, 0.3546, 0.0039, 0.0379, 0.3568, 0.0017, 0.0379, 0.3568, 0.0039, 0.0379, 0.3503, -0.0, 0.0429, 0.3503, -0.0, 0.045, 0.3524, -0.0, 0.0429, 0.3524, -0.0, 0.045, 0.3503, -0.0, 0.0472, 0.3503, -0.0, 0.0487, 0.3524, -0.0, 0.0472, 0.3524, -0.0, 0.0487, 0.3503, 0.0017, 0.0487, 0.3503, 0.0039, 0.0487, 0.3524, 0.0017, 0.0487, 0.3524, 0.0039, 0.0487, 0.3546, -0.0, 0.0429, 0.3546, -0.0, 0.045, 0.3568, -0.0, 0.0429, 0.3568, -0.0, 0.045, 0.3546, -0.0, 0.0472, 0.3546, -0.0, 0.0487, 0.3568, -0.0, 0.0472, 0.3568, -0.0, 0.0487, 0.3546, 0.0017, 0.0487, 0.3546, 0.0039, 0.0487, 0.3568, 0.0017, 0.0487, 0.3568, 0.0039, 0.0487, 0.3589, -0.0, 0.0379, 0.3589, -0.0, 0.0407,
0.3611, -0.0, 0.0379, 0.3611, -0.0, 0.0407, 0.3589, 0.0017, 0.0379, 0.3589, 0.0039, 0.0379, 0.3611, 0.0017, 0.0379, 0.3611, 0.0039, 0.0379, 0.3633, -0.0, 0.0379, 0.3633, -0.0, 0.0407, 0.3655, -0.0, 0.0379, 0.3655, -0.0, 0.0407, 0.3633, 0.0017, 0.0379, 0.3633, 0.0039, 0.0379, 0.3655, 0.0017, 0.0379, 0.3655, 0.0039, 0.0379, 0.3589, -0.0, 0.0429, 0.3589, -0.0, 0.045, 0.3611, -0.0, 0.0429, 0.3611, -0.0, 0.045, 0.3589, -0.0, 0.0472, 0.3589, -0.0, 0.0487, 0.3611, -0.0, 0.0472, 0.3611, -0.0, 0.0487, 0.3589, 0.0017, 0.0487, 0.3589, 0.0039, 0.0487, 0.3611, 0.0017, 0.0487, 0.3611, 0.0039, 0.0487, 0.3633, -0.0, 0.0429, 0.3633, -0.0, 0.045, 0.3655, -0.0, 0.0429, 0.3655, -0.0, 0.045,
0.3633, -0.0, 0.0472, 0.3633, -0.0, 0.0487, 0.3655, -0.0, 0.0472, 0.3655, -0.0, 0.0487, 0.3633, 0.0017, 0.0487, 0.3633, 0.0039, 0.0487, 0.3655, 0.0017, 0.0487, 0.3655, 0.0039, 0.0487, 0.3676, -0.0, 0.0379, 0.3676, -0.0, 0.0407, 0.3698, -0.0, 0.0379, 0.3698, -0.0, 0.0407, 0.3676, 0.0017, 0.0379, 0.3676, 0.0039, 0.0379, 0.3698, 0.0017, 0.0379, 0.3698, 0.0039, 0.0379, 0.372, -0.0, 0.0379, 0.372, -0.0, 0.0407, 0.3741, -0.0, 0.0379, 0.3741, -0.0, 0.0407, 0.372, 0.0017, 0.0379, 0.372, 0.0039, 0.0379, 0.3741, 0.0017, 0.0379, 0.3741, 0.0039, 0.0379, 0.3676, -0.0, 0.0429, 0.3676, -0.0, 0.045, 0.3698, -0.0, 0.0429, 0.3698, -0.0, 0.045, 0.3676, -0.0, 0.0472, 0.3676, -0.0, 0.0487,
0.3698, -0.0, 0.0472, 0.3698, -0.0, 0.0487, 0.3676, 0.0017, 0.0487, 0.3676, 0.0039, 0.0487, 0.3698, 0.0017, 0.0487, 0.3698, 0.0039, 0.0487, 0.372, -0.0, 0.0429, 0.372, -0.0, 0.045, 0.3741, -0.0, 0.0429, 0.3741, -0.0, 0.045, 0.372, -0.0, 0.0472, 0.372, -0.0, 0.0487, 0.3741, -0.0, 0.0472, 0.3741, -0.0, 0.0487, 0.372, 0.0017, 0.0487, 0.372, 0.0039, 0.0487, 0.3741, 0.0017, 0.0487, 0.3741, 0.0039, 0.0487, 0.3763, -0.0, 0.0379, 0.3763, -0.0, 0.0407, 0.3784, -0.0, 0.0379, 0.3785, -0.0, 0.0407, 0.3763, 0.0017, 0.0379, 0.3763, 0.0039, 0.0379, 0.3785, 0.0017, 0.0379, 0.3785, 0.0039, 0.0379, 0.3835, -0.0, 0.0373, 0.3806, -0.0, 0.0377, 0.3806, -0.0, 0.0407, 0.3827, -0.0, 0.0374,
0.3828, -0.0, 0.0407, 0.3833, 0.0017, 0.0373, 0.3833, 0.0039, 0.0373, 0.3807, 0.0017, 0.0377, 0.3807, 0.0039, 0.0377, 0.3822, 0.0017, 0.0375, 0.3822, 0.0039, 0.0375, 0.3763, -0.0, 0.0429, 0.3763, -0.0, 0.045, 0.3785, -0.0, 0.0429, 0.3785, -0.0, 0.045, 0.3763, -0.0, 0.0472, 0.3763, -0.0, 0.0487, 0.3785, -0.0, 0.0472, 0.3785, -0.0, 0.0487, 0.3763, 0.0017, 0.0487, 0.3763, 0.0039, 0.0487, 0.3785, 0.0017, 0.0487, 0.3785, 0.0039, 0.0487, 0.3806, -0.0, 0.0429, 0.3806, -0.0, 0.045, 0.3828, -0.0, 0.0429, 0.3828, -0.0, 0.045, 0.3806, -0.0, 0.0472, 0.3806, -0.0, 0.0487, 0.3828, -0.0, 0.0472, 0.3828, -0.0, 0.0487, 0.3806, 0.0017, 0.0487, 0.3806, 0.0039, 0.0487, 0.3828, 0.0017, 0.0487,
0.3828, 0.0039, 0.0487, 0.4002, -0.0, -0.0003, 0.4009, -0.0, 0.001, 0.3999, 0.0017, -0.0006, 0.401, 0.0017, 0.0011, 0.3999, 0.0039, -0.0006, 0.401, 0.0039, 0.0011, 0.3947, -0.0, 0.033, 0.3956, -0.0, 0.0324, 0.3947, 0.0017, 0.033, 0.3947, 0.0039, 0.033, 0.3959, 0.0017, 0.0321, 0.3959, 0.0039, 0.0321, 0.4009, -0.0, 0.0261, 0.4001, -0.0, 0.0274, 0.3982, -0.0, 0.03, 0.3971, -0.0, 0.031, 0.399, -0.0, 0.0289, 0.4002, -0.0, 0.032, 0.4009, 0.0017, 0.026, 0.3999, 0.0017, 0.0277, 0.4009, 0.0039, 0.026, 0.3999, 0.0039, 0.0277, 0.3982, 0.0017, 0.0299, 0.3971, 0.0017, 0.031, 0.3982, 0.0039, 0.0299, 0.3971, 0.0039, 0.031, 0.3991, 0.0017, 0.0288, 0.3991, 0.0039, 0.0288, 0.4024, -0.0, -0.0005,
0.4015, -0.0, 0.002, 0.4045, -0.0, -0.0005, 0.4045, -0.0, 0.0016, 0.4025, -0.0, 0.004, 0.4031, -0.0, 0.0056, 0.4045, -0.0, 0.0038, 0.4033, -0.0, 0.0064, 0.4016, 0.0017, 0.0021, 0.4016, 0.0039, 0.0021, 0.4024, 0.0017, 0.0038, 0.4032, 0.0017, 0.0058, 0.4024, 0.0039, 0.0038, 0.4032, 0.0039, 0.0058, 0.4035, 0.0017, 0.0069, 0.4035, 0.0039, 0.0069, 0.4067, -0.0, -0.0005, 0.4067, -0.0, 0.0016, 0.4089, -0.0, -0.0005, 0.4089, -0.0, 0.0016, 0.4067, -0.0, 0.0038, 0.4067, -0.0, 0.006, 0.4089, -0.0, 0.0038, 0.4089, -0.0, 0.006, 0.4038, -0.0, 0.0083, 0.4039, -0.0, 0.0104, 0.4039, -0.0, 0.0125, 0.4039, -0.0, 0.0146, 0.4037, 0.0017, 0.0081, 0.4039, 0.0017, 0.0103, 0.4037, 0.0039, 0.0081,
0.4039, 0.0039, 0.0103, 0.4039, 0.0017, 0.0125, 0.4039, 0.0017, 0.0146, 0.4039, 0.0039, 0.0125, 0.4039, 0.0039, 0.0146, 0.4067, -0.0, 0.0081, 0.4067, -0.0, 0.0103, 0.4089, -0.0, 0.0081, 0.4089, -0.0, 0.0103, 0.4067, -0.0, 0.0125, 0.4067, -0.0, 0.0146, 0.4089, -0.0, 0.0125, 0.4089, -0.0, 0.0146, 0.411, -0.0, -0.0005, 0.411, -0.0, 0.0016, 0.4131, -0.0, -0.0007, 0.414, -0.0, 0.0015, 0.411, -0.0, 0.0038, 0.411, -0.0, 0.006, 0.4132, -0.0, 0.0038, 0.4132, -0.0, 0.006, 0.4132, 0.0017, -0.0006, 0.414, 0.0017, 0.0015, 0.4132, 0.0039, -0.0006, 0.414, 0.0039, 0.0015, 0.4144, -0.0, 0.0026, 0.4147, -0.0, 0.0037, 0.4152, -0.0, 0.0059, 0.4143, 0.0017, 0.0026, 0.4143, 0.0039, 0.0026,
0.4147, 0.0017, 0.0038, 0.4153, 0.0017, 0.006, 0.4147, 0.0039, 0.0038, 0.4153, 0.0039, 0.006, 0.411, -0.0, 0.0081, 0.411, -0.0, 0.0103, 0.4132, -0.0, 0.0081, 0.4132, -0.0, 0.0103, 0.411, -0.0, 0.0125, 0.411, -0.0, 0.0146, 0.4132, -0.0, 0.0125, 0.4132, -0.0, 0.0146, 0.4156, -0.0, 0.0081, 0.4157, -0.0, 0.0103, 0.4157, -0.0, 0.0125, 0.4157, -0.0, 0.0146, 0.4156, 0.0017, 0.0081, 0.4157, 0.0017, 0.0103, 0.4156, 0.0039, 0.0081, 0.4157, 0.0039, 0.0103, 0.4157, 0.0017, 0.0125, 0.4157, 0.0017, 0.0146, 0.4157, 0.0039, 0.0125, 0.4157, 0.0039, 0.0146, 0.4039, -0.0, 0.0168, 0.4037, -0.0, 0.0188, 0.4031, -0.0, 0.0215, 0.4025, -0.0, 0.0231, 0.4033, -0.0, 0.0207, 0.4045, -0.0, 0.0233,
0.4039, 0.0017, 0.0168, 0.4037, 0.0017, 0.019, 0.4039, 0.0039, 0.0168, 0.4037, 0.0039, 0.019, 0.4032, 0.0017, 0.0212, 0.4024, 0.0017, 0.0234, 0.4032, 0.0039, 0.0212, 0.4024, 0.0039, 0.0234, 0.4034, 0.0017, 0.0201, 0.4034, 0.0039, 0.0201, 0.4067, -0.0, 0.0168, 0.4067, -0.0, 0.019, 0.4089, -0.0, 0.0168, 0.4089, -0.0, 0.019, 0.4067, -0.0, 0.0212, 0.4067, -0.0, 0.0233, 0.4089, -0.0, 0.0212, 0.4089, -0.0, 0.0233, 0.4015, -0.0, 0.0251, 0.4024, -0.0, 0.0277, 0.4045, -0.0, 0.0255, 0.4045, -0.0, 0.0277, 0.4024, -0.0, 0.0298, 0.4024, -0.0, 0.032, 0.4045, -0.0, 0.0298, 0.4045, -0.0, 0.032, 0.4016, 0.0017, 0.025, 0.4016, 0.0039, 0.025, 0.4067, -0.0, 0.0255, 0.4067, -0.0, 0.0277,
0.4089, -0.0, 0.0255, 0.4089, -0.0, 0.0277, 0.4067, -0.0, 0.0298, 0.4067, -0.0, 0.032, 0.4089, -0.0, 0.0298, 0.4089, -0.0, 0.032, 0.411, -0.0, 0.0168, 0.411, -0.0, 0.019, 0.4132, -0.0, 0.0168, 0.4132, -0.0, 0.019, 0.411, -0.0, 0.0212, 0.411, -0.0, 0.0233, 0.4132, -0.0, 0.0212, 0.4132, -0.0, 0.0233, 0.4157, -0.0, 0.0168, 0.4157, -0.0, 0.019, 0.4155, -0.0, 0.0212, 0.4152, -0.0, 0.0234, 0.4157, 0.0017, 0.0168, 0.4157, 0.0017, 0.019, 0.4157, 0.0039, 0.0168, 0.4157, 0.0039, 0.019, 0.4155, 0.0017, 0.0212, 0.4152, 0.0017, 0.0233, 0.4155, 0.0039, 0.0212, 0.4152, 0.0039, 0.0233, 0.411, -0.0, 0.0255, 0.411, -0.0, 0.0277, 0.4144, -0.0, 0.026, 0.4137, -0.0, 0.0279,
0.411, -0.0, 0.0298, 0.4117, -0.0, 0.0323, 0.4129, -0.0, 0.03, 0.4123, -0.0, 0.0312, 0.4142, 0.0017, 0.0265, 0.4138, 0.0017, 0.0277, 0.4142, 0.0039, 0.0265, 0.4138, 0.0039, 0.0277, 0.4117, 0.0017, 0.0323, 0.4117, 0.0039, 0.0323, 0.4129, 0.0017, 0.0299, 0.4123, 0.0017, 0.0312, 0.4129, 0.0039, 0.0299, 0.4123, 0.0039, 0.0312, 0.4147, -0.0, 0.0251, 0.4146, 0.0017, 0.0254, 0.4146, 0.0039, 0.0254, 0.385, -0.0, 0.0371, 0.3871, -0.0, 0.0367, 0.385, -0.0, 0.0385, 0.385, -0.0, 0.0407, 0.3872, -0.0, 0.0385, 0.3872, -0.0, 0.0407, 0.385, 0.0017, 0.0371, 0.385, 0.0039, 0.0371, 0.3872, 0.0017, 0.0366, 0.3872, 0.0039, 0.0366, 0.3891, -0.0, 0.0361, 0.3919, -0.0, 0.0348, 0.3909, -0.0, 0.0352,
0.3893, -0.0, 0.0385, 0.3893, -0.0, 0.0407, 0.3915, -0.0, 0.0385, 0.3915, -0.0, 0.0407, 0.3894, 0.0017, 0.036, 0.3894, 0.0039, 0.036, 0.3918, 0.0017, 0.0348, 0.3907, 0.0017, 0.0354, 0.3918, 0.0039, 0.0348, 0.3907, 0.0039, 0.0354, 0.385, -0.0, 0.0429, 0.385, -0.0, 0.045, 0.3872, -0.0, 0.0429, 0.3872, -0.0, 0.045, 0.385, -0.0, 0.0472, 0.385, -0.0, 0.0487, 0.3872, -0.0, 0.0472, 0.3872, -0.0, 0.0486, 0.385, 0.0017, 0.0487, 0.385, 0.0039, 0.0487, 0.3872, 0.0017, 0.0486, 0.3872, 0.0039, 0.0486, 0.3893, -0.0, 0.0429, 0.3893, -0.0, 0.045, 0.3915, -0.0, 0.0429, 0.3915, -0.0, 0.045, 0.3896, -0.0, 0.0483, 0.3887, -0.0, 0.0485, 0.3917, -0.0, 0.0478, 0.39, 0.0017, 0.0482,
0.389, 0.0017, 0.0484, 0.39, 0.0039, 0.0482, 0.389, 0.0039, 0.0484, 0.3915, 0.0017, 0.0479, 0.3915, 0.0039, 0.0479, 0.3936, -0.0, 0.0338, 0.3937, -0.0, 0.0363, 0.3958, -0.0, 0.0342, 0.3958, -0.0, 0.0363, 0.3937, -0.0, 0.0385, 0.3937, -0.0, 0.0407, 0.3958, -0.0, 0.0385, 0.3958, -0.0, 0.0407, 0.3936, 0.0017, 0.0338, 0.3936, 0.0039, 0.0338, 0.398, -0.0, 0.0342, 0.398, -0.0, 0.0363, 0.4002, -0.0, 0.0342, 0.4002, -0.0, 0.0363, 0.398, -0.0, 0.0385, 0.398, -0.0, 0.0407, 0.4002, -0.0, 0.0385, 0.4002, -0.0, 0.0407, 0.3937, -0.0, 0.0429, 0.3937, -0.0, 0.045, 0.3958, -0.0, 0.0429, 0.3964, -0.0, 0.0461, 0.3938, -0.0, 0.0471, 0.3954, -0.0, 0.0465, 0.3967, 0.0017, 0.046,
0.3967, 0.0039, 0.046, 0.3937, 0.0017, 0.0472, 0.3937, 0.0039, 0.0472, 0.3956, 0.0017, 0.0465, 0.3956, 0.0039, 0.0465, 0.398, -0.0, 0.0429, 0.3983, -0.0, 0.0453, 0.4008, -0.0, 0.0438, 0.3997, -0.0, 0.0445, 0.398, 0.0017, 0.0454, 0.398, 0.0039, 0.0454, 0.4009, 0.0017, 0.0437, 0.3998, 0.0017, 0.0444, 0.4009, 0.0039, 0.0437, 0.3998, 0.0039, 0.0444, 0.4024, -0.0, 0.0342, 0.4024, -0.0, 0.0363, 0.4045, -0.0, 0.0342, 0.4045, -0.0, 0.0363, 0.4024, -0.0, 0.0385, 0.4024, -0.0, 0.0407, 0.4045, -0.0, 0.0385, 0.4046, -0.0, 0.0409, 0.4046, 0.0017, 0.0409, 0.4046, 0.0039, 0.0409, 0.4067, -0.0, 0.0342, 0.4067, -0.0, 0.0363, 0.41, -0.0, 0.035, 0.4088, -0.0, 0.0367, 0.4069, -0.0, 0.0387,
0.4058, -0.0, 0.0398, 0.4079, -0.0, 0.0376, 0.4099, 0.0017, 0.0352, 0.409, 0.0017, 0.0364, 0.4099, 0.0039, 0.0352, 0.409, 0.0039, 0.0364, 0.4069, 0.0017, 0.0387, 0.4058, 0.0017, 0.0398, 0.4069, 0.0039, 0.0387, 0.4058, 0.0039, 0.0398, 0.4079, 0.0017, 0.0376, 0.4079, 0.0039, 0.0376, 0.4026, -0.0, 0.0426, 0.4035, -0.0, 0.0418, 0.4024, 0.0017, 0.0427, 0.4024, 0.0039, 0.0427, 0.4035, 0.0017, 0.0418, 0.4035, 0.0039, 0.0418, 0.4107, -0.0, 0.034, 0.4106, 0.0017, 0.0341, 0.4106, 0.0039, 0.0341, 0.4291, -0.0, 0.0153, 0.4316, -0.0, 0.0135, 0.4304, -0.0, 0.0142, 0.429, 0.0017, 0.0153, 0.429, 0.0039, 0.0153, 0.4316, 0.0017, 0.0135, 0.4305, 0.0017, 0.0142, 0.4316, 0.0039, 0.0135, 0.4305, 0.0039, 0.0142,
0.4326, -0.0, 0.013, 0.4327, -0.0, 0.0146, 0.4346, -0.0, 0.012, 0.4349, -0.0, 0.0146, 0.4327, 0.0017, 0.0129, 0.4327, 0.0039, 0.0129, 0.4349, 0.0017, 0.0119, 0.4349, 0.0039, 0.0119, 0.427, -0.0, 0.0175, 0.4261, -0.0, 0.0186, 0.4244, -0.0, 0.0214, 0.4238, -0.0, 0.0231, 0.425, -0.0, 0.0204, 0.4262, -0.0, 0.0233, 0.427, 0.0017, 0.0175, 0.4258, 0.0017, 0.0189, 0.427, 0.0039, 0.0175, 0.4258, 0.0039, 0.0189, 0.4245, 0.0017, 0.0211, 0.4237, 0.0017, 0.0233, 0.4245, 0.0039, 0.0211, 0.4237, 0.0039, 0.0233, 0.4252, 0.0017, 0.0201, 0.4252, 0.0039, 0.0201, 0.4229, -0.0, 0.0265, 0.4228, -0.0, 0.0277, 0.4226, -0.0, 0.0298, 0.4228, -0.0, 0.0316, 0.423, 0.0017, 0.0265, 0.4228, 0.0017, 0.0277,
0.423, 0.0039, 0.0265, 0.4228, 0.0039, 0.0277, 0.4226, 0.0017, 0.0298, 0.4228, 0.0017, 0.0319, 0.4226, 0.0039, 0.0298, 0.4228, 0.0039, 0.0319, 0.4231, -0.0, 0.0255, 0.4241, -0.0, 0.0277, 0.4262, -0.0, 0.0255, 0.4262, -0.0, 0.0277, 0.4241, -0.0, 0.0298, 0.4229, -0.0, 0.0324, 0.4262, -0.0, 0.0298, 0.4262, -0.0, 0.032, 0.4231, 0.0017, 0.0254, 0.4231, 0.0039, 0.0254, 0.423, 0.0017, 0.033, 0.423, 0.0039, 0.033, 0.4279, -0.0, 0.0164, 0.4284, -0.0, 0.019, 0.4306, -0.0, 0.0168, 0.4306, -0.0, 0.019, 0.4284, -0.0, 0.0212, 0.4284, -0.0, 0.0233, 0.4306, -0.0, 0.0212, 0.4306, -0.0, 0.0233, 0.4279, 0.0017, 0.0164, 0.4279, 0.0039, 0.0164, 0.4327, -0.0, 0.0168, 0.4327, -0.0, 0.019,
0.4349, -0.0, 0.0168, 0.4349, -0.0, 0.019, 0.4327, -0.0, 0.0212, 0.4327, -0.0, 0.0233, 0.4349, -0.0, 0.0212, 0.4349, -0.0, 0.0233, 0.4284, -0.0, 0.0255, 0.4284, -0.0, 0.0277, 0.4306, -0.0, 0.0255, 0.4306, -0.0, 0.0277, 0.4284, -0.0, 0.0298, 0.4284, -0.0, 0.032, 0.4306, -0.0, 0.0298, 0.4306, -0.0, 0.032, 0.4327, -0.0, 0.0255, 0.4327, -0.0, 0.0277, 0.4349, -0.0, 0.0255, 0.4349, -0.0, 0.0277, 0.4327, -0.0, 0.0298, 0.4327, -0.0, 0.032, 0.4349, -0.0, 0.0298, 0.4349, -0.0, 0.032, 0.4377, -0.0, 0.0112, 0.4392, -0.0, 0.0109, 0.4368, -0.0, 0.0114, 0.4371, -0.0, 0.0146, 0.4392, -0.0, 0.0125, 0.4392, -0.0, 0.0146, 0.4375, 0.0017, 0.0112, 0.4375, 0.0039, 0.0112,
0.4392, 0.0017, 0.0109, 0.4392, 0.0039, 0.0109, 0.4364, 0.0017, 0.0115, 0.4364, 0.0039, 0.0115, 0.4414, -0.0, 0.0106, 0.4436, -0.0, 0.0105, 0.4414, -0.0, 0.0125, 0.4414, -0.0, 0.0146, 0.4436, -0.0, 0.0125, 0.4436, -0.0, 0.0146, 0.4414, 0.0017, 0.0106, 0.4414, 0.0039, 0.0106, 0.4436, 0.0017, 0.0105, 0.4436, 0.0039, 0.0105, 0.4458, -0.0, 0.0105, 0.4479, -0.0, 0.0105, 0.4458, -0.0, 0.0125, 0.4458, -0.0, 0.0146, 0.4479, -0.0, 0.0125, 0.4479, -0.0, 0.0146, 0.4458, 0.0017, 0.0105, 0.4458, 0.0039, 0.0105, 0.4479, 0.0017, 0.0105, 0.4479, 0.0039, 0.0105, 0.4501, -0.0, 0.0105, 0.4523, -0.0, 0.0105, 0.4501, -0.0, 0.0125, 0.4501, -0.0, 0.0146, 0.4523, -0.0, 0.0125, 0.4523, -0.0, 0.0146,
0.4501, 0.0017, 0.0105, 0.4501, 0.0039, 0.0105, 0.4523, 0.0017, 0.0105, 0.4523, 0.0039, 0.0105, 0.4371, -0.0, 0.0168, 0.4371, -0.0, 0.019, 0.4392, -0.0, 0.0168, 0.4392, -0.0, 0.019, 0.4371, -0.0, 0.0212, 0.4371, -0.0, 0.0233, 0.4392, -0.0, 0.0212, 0.4398, -0.0, 0.0242, 0.4399, 0.0017, 0.0241, 0.4399, 0.0039, 0.0241, 0.4414, -0.0, 0.0168, 0.4414, -0.0, 0.019, 0.4436, -0.0, 0.0168, 0.4436, -0.0, 0.019, 0.4414, -0.0, 0.0212, 0.4416, -0.0, 0.0232, 0.4436, -0.0, 0.0212, 0.4436, -0.0, 0.0225, 0.4414, 0.0017, 0.0232, 0.4414, 0.0039, 0.0232, 0.4436, 0.0017, 0.0225, 0.4436, 0.0039, 0.0225, 0.4378, -0.0, 0.026, 0.4366, -0.0, 0.0278, 0.4388, -0.0, 0.025, 0.436, -0.0, 0.0297,
0.4366, -0.0, 0.0319, 0.4377, 0.0017, 0.0261, 0.4367, 0.0017, 0.0276, 0.4377, 0.0039, 0.0261, 0.4367, 0.0039, 0.0276, 0.4388, 0.0017, 0.025, 0.4388, 0.0039, 0.025, 0.436, 0.0017, 0.0297, 0.4367, 0.0017, 0.0321, 0.436, 0.0039, 0.0297, 0.4367, 0.0039, 0.0321, 0.4458, -0.0, 0.0168, 0.4458, -0.0, 0.019, 0.4479, -0.0, 0.0168, 0.4479, -0.0, 0.019, 0.4459, -0.0, 0.0223, 0.4453, -0.0, 0.0223, 0.4479, -0.0, 0.0222, 0.4467, 0.0017, 0.0222, 0.4456, 0.0017, 0.0223, 0.4467, 0.0039, 0.0222, 0.4456, 0.0039, 0.0223, 0.4479, 0.0017, 0.0222, 0.4479, 0.0039, 0.0222, 0.4501, -0.0, 0.0168, 0.4501, -0.0, 0.019, 0.4523, -0.0, 0.0168, 0.4523, -0.0, 0.019, 0.4501, -0.0, 0.0222, 0.4523, -0.0, 0.0222,
0.4501, 0.0017, 0.0222, 0.4501, 0.0039, 0.0222, 0.4523, 0.0017, 0.0222, 0.4523, 0.0039, 0.0222, 0.4232, -0.0, 0.0344, 0.4238, -0.0, 0.0366, 0.4262, -0.0, 0.0342, 0.4262, -0.0, 0.0363, 0.4244, -0.0, 0.0383, 0.4251, -0.0, 0.0393, 0.4262, -0.0, 0.041, 0.4232, 0.0017, 0.0342, 0.4237, 0.0017, 0.0364, 0.4232, 0.0039, 0.0342, 0.4237, 0.0039, 0.0364, 0.4245, 0.0017, 0.0385, 0.4245, 0.0039, 0.0385, 0.4252, 0.0017, 0.0395, 0.426, 0.0017, 0.0407, 0.4252, 0.0039, 0.0395, 0.426, 0.0039, 0.0407, 0.427, -0.0, 0.042, 0.427, 0.0017, 0.0421, 0.427, 0.0039, 0.0421, 0.4284, -0.0, 0.0342, 0.4284, -0.0, 0.0363, 0.4306, -0.0, 0.0342, 0.4306, -0.0, 0.0363, 0.4284, -0.0, 0.0385, 0.4284, -0.0, 0.0407,
0.4306, -0.0, 0.0385, 0.4306, -0.0, 0.0407, 0.4327, -0.0, 0.0342, 0.4327, -0.0, 0.0363, 0.4349, -0.0, 0.0342, 0.4349, -0.0, 0.0363, 0.4327, -0.0, 0.0385, 0.4327, -0.0, 0.0407, 0.4349, -0.0, 0.0385, 0.4349, -0.0, 0.0407, 0.4281, -0.0, 0.0432, 0.4292, -0.0, 0.0442, 0.4306, -0.0, 0.0429, 0.4302, -0.0, 0.045, 0.4281, 0.0017, 0.0432, 0.4292, 0.0017, 0.0442, 0.4281, 0.0039, 0.0432, 0.4292, 0.0039, 0.0442, 0.4305, 0.0017, 0.0452, 0.4305, 0.0039, 0.0452, 0.4327, -0.0, 0.0429, 0.432, -0.0, 0.0461, 0.4349, -0.0, 0.0429, 0.4349, -0.0, 0.045, 0.433, -0.0, 0.0468, 0.4346, -0.0, 0.0475, 0.4318, 0.0017, 0.046, 0.4318, 0.0039, 0.046, 0.4329, 0.0017, 0.0467, 0.4329, 0.0039, 0.0467,
0.4349, 0.0017, 0.0476, 0.4349, 0.0039, 0.0476, 0.4377, -0.0, 0.0337, 0.4371, -0.0, 0.0363, 0.4388, -0.0, 0.0347, 0.4398, -0.0, 0.0356, 0.4371, -0.0, 0.0385, 0.4371, -0.0, 0.0407, 0.4392, -0.0, 0.0385, 0.4392, -0.0, 0.0407, 0.4377, 0.0017, 0.0336, 0.4377, 0.0039, 0.0336, 0.4388, 0.0017, 0.0347, 0.4398, 0.0017, 0.0356, 0.4388, 0.0039, 0.0347, 0.4398, 0.0039, 0.0356, 0.4416, -0.0, 0.0367, 0.443, -0.0, 0.0372, 0.4414, -0.0, 0.0385, 0.4414, -0.0, 0.0407, 0.4439, -0.0, 0.0374, 0.4436, -0.0, 0.0407, 0.4414, 0.0017, 0.0366, 0.4414, 0.0039, 0.0366, 0.4432, 0.0017, 0.0373, 0.4432, 0.0039, 0.0373, 0.4443, 0.0017, 0.0375, 0.4443, 0.0039, 0.0375, 0.4371, -0.0, 0.0429, 0.4371, -0.0, 0.045,
0.4392, -0.0, 0.0429, 0.4392, -0.0, 0.045, 0.4368, -0.0, 0.0482, 0.4378, -0.0, 0.0484, 0.4392, -0.0, 0.0472, 0.4392, -0.0, 0.0486, 0.4366, 0.0017, 0.0481, 0.4376, 0.0017, 0.0484, 0.4366, 0.0039, 0.0481, 0.4376, 0.0039, 0.0484, 0.4392, 0.0017, 0.0486, 0.4392, 0.0039, 0.0486, 0.4414, -0.0, 0.0429, 0.4414, -0.0, 0.045, 0.4436, -0.0, 0.0429, 0.4436, -0.0, 0.045, 0.4414, -0.0, 0.0472, 0.4414, -0.0, 0.0487, 0.4436, -0.0, 0.0472, 0.4436, -0.0, 0.0487, 0.4414, 0.0017, 0.0487, 0.4414, 0.0039, 0.0487, 0.4436, 0.0017, 0.0487, 0.4436, 0.0039, 0.0487, 0.4459, -0.0, 0.0378, 0.4458, -0.0, 0.0407, 0.4479, -0.0, 0.0379, 0.4479, -0.0, 0.0407, 0.4457, 0.0017, 0.0378, 0.4457, 0.0039, 0.0378,
0.4479, 0.0017, 0.0379, 0.4479, 0.0039, 0.0379, 0.4501, -0.0, 0.0379, 0.4501, -0.0, 0.0407, 0.4523, -0.0, 0.0379, 0.4523, -0.0, 0.0407, 0.4501, 0.0017, 0.0379, 0.4501, 0.0039, 0.0379, 0.4523, 0.0017, 0.0379, 0.4523, 0.0039, 0.0379, 0.4458, -0.0, 0.0429, 0.4458, -0.0, 0.045, 0.4479, -0.0, 0.0429, 0.4479, -0.0, 0.045, 0.4458, -0.0, 0.0472, 0.4458, -0.0, 0.0487, 0.4479, -0.0, 0.0472, 0.4479, -0.0, 0.0487, 0.4458, 0.0017, 0.0487, 0.4458, 0.0039, 0.0487, 0.4479, 0.0017, 0.0487, 0.4479, 0.0039, 0.0487, 0.4501, -0.0, 0.0429, 0.4501, -0.0, 0.045, 0.4523, -0.0, 0.0429, 0.4523, -0.0, 0.045, 0.4501, -0.0, 0.0472, 0.4501, -0.0, 0.0487, 0.4523, -0.0, 0.0472, 0.4523, -0.0, 0.0487,
0.4501, 0.0017, 0.0487, 0.4501, 0.0039, 0.0487, 0.4523, 0.0017, 0.0487, 0.4523, 0.0039, 0.0487, 0.4544, -0.0, 0.0105, 0.4566, -0.0, 0.0105, 0.4544, -0.0, 0.0125, 0.4544, -0.0, 0.0146, 0.4566, -0.0, 0.0125, 0.4566, -0.0, 0.0146, 0.4544, 0.0017, 0.0105, 0.4544, 0.0039, 0.0105, 0.4566, 0.0017, 0.0105, 0.4566, 0.0039, 0.0105, 0.4588, -0.0, 0.0105, 0.4609, -0.0, 0.0105, 0.4588, -0.0, 0.0125, 0.4588, -0.0, 0.0146, 0.4609, -0.0, 0.0125, 0.4609, -0.0, 0.0146, 0.4588, 0.0017, 0.0105, 0.4588, 0.0039, 0.0105, 0.4609, 0.0017, 0.0105, 0.4609, 0.0039, 0.0105, 0.4631, -0.0, 0.0105, 0.4653, -0.0, 0.0105, 0.4631, -0.0, 0.0125, 0.4631, -0.0, 0.0146, 0.4653, -0.0, 0.0125, 0.4653, -0.0, 0.0146,
0.4631, 0.0017, 0.0105, 0.4631, 0.0039, 0.0105, 0.4653, 0.0017, 0.0105, 0.4653, 0.0039, 0.0105, 0.4675, -0.0, 0.0105, 0.4696, -0.0, 0.0105, 0.4675, -0.0, 0.0125, 0.4675, -0.0, 0.0146, 0.4696, -0.0, 0.0125, 0.4696, -0.0, 0.0146, 0.4675, 0.0017, 0.0105, 0.4675, 0.0039, 0.0105, 0.4696, 0.0017, 0.0105, 0.4696, 0.0039, 0.0105, 0.4544, -0.0, 0.0168, 0.4544, -0.0, 0.019, 0.4566, -0.0, 0.0168, 0.4566, -0.0, 0.019, 0.4544, -0.0, 0.0222, 0.4566, -0.0, 0.0222, 0.4544, 0.0017, 0.0222, 0.4544, 0.0039, 0.0222, 0.4566, 0.0017, 0.0222, 0.4566, 0.0039, 0.0222, 0.4588, -0.0, 0.0168, 0.4588, -0.0, 0.019, 0.4609, -0.0, 0.0168, 0.4609, -0.0, 0.019, 0.4588, -0.0, 0.0222, 0.4609, -0.0, 0.0222,
0.4588, 0.0017, 0.0222, 0.4588, 0.0039, 0.0222, 0.4609, 0.0017, 0.0222, 0.4609, 0.0039, 0.0222, 0.4631, -0.0, 0.0168, 0.4631, -0.0, 0.019, 0.4653, -0.0, 0.0168, 0.4653, -0.0, 0.019, 0.4631, -0.0, 0.0222, 0.4653, -0.0, 0.0222, 0.4631, 0.0017, 0.0222, 0.4631, 0.0039, 0.0222, 0.4653, 0.0017, 0.0222, 0.4653, 0.0039, 0.0222, 0.4675, -0.0, 0.0168, 0.4675, -0.0, 0.019, 0.4696, -0.0, 0.0168, 0.4696, -0.0, 0.019, 0.4675, -0.0, 0.0222, 0.4696, -0.0, 0.0222, 0.4675, 0.0017, 0.0222, 0.4675, 0.0039, 0.0222, 0.4696, 0.0017, 0.0222, 0.4696, 0.0039, 0.0222, 0.4718, -0.0, 0.0105, 0.474, -0.0, 0.0105, 0.4718, -0.0, 0.0125, 0.4718, -0.0, 0.0146, 0.474, -0.0, 0.0125, 0.474, -0.0, 0.0146,
0.4718, 0.0017, 0.0105, 0.4718, 0.0039, 0.0105, 0.474, 0.0017, 0.0105, 0.474, 0.0039, 0.0105, 0.4761, -0.0, 0.0104, 0.4783, -0.0, 0.0104, 0.4761, -0.0, 0.0125, 0.4761, -0.0, 0.0146, 0.4783, -0.0, 0.0125, 0.4783, -0.0, 0.0146, 0.4761, 0.0017, 0.0104, 0.4761, 0.0039, 0.0104, 0.4783, 0.0017, 0.0104, 0.4783, 0.0039, 0.0104, 0.4877, -0.0, 0.0067, 0.4877, 0.0017, 0.0067, 0.4877, 0.0039, 0.0067, 0.4804, -0.0, 0.0102, 0.4837, -0.0, 0.0092, 0.4826, -0.0, 0.0096, 0.4805, -0.0, 0.0125, 0.4805, -0.0, 0.0146, 0.4826, -0.0, 0.0125, 0.4826, -0.0, 0.0146, 0.4805, 0.0017, 0.0102, 0.4805, 0.0039, 0.0102, 0.4836, 0.0017, 0.0092, 0.4826, 0.0017, 0.0096, 0.4836, 0.0039, 0.0092, 0.4826, 0.0039, 0.0096,
0.4847, -0.0, 0.0087, 0.4848, -0.0, 0.0103, 0.4866, -0.0, 0.0076, 0.487, -0.0, 0.0103, 0.4848, -0.0, 0.0125, 0.4848, -0.0, 0.0146, 0.487, -0.0, 0.0125, 0.487, -0.0, 0.0146, 0.4849, 0.0017, 0.0087, 0.4849, 0.0039, 0.0087, 0.4866, 0.0017, 0.0076, 0.4866, 0.0039, 0.0076, 0.4718, -0.0, 0.0168, 0.4718, -0.0, 0.019, 0.474, -0.0, 0.0168, 0.474, -0.0, 0.019, 0.4718, -0.0, 0.0222, 0.474, -0.0, 0.0222, 0.4718, 0.0017, 0.0222, 0.4718, 0.0039, 0.0222, 0.474, 0.0017, 0.0222, 0.474, 0.0039, 0.0222, 0.4761, -0.0, 0.0168, 0.4761, -0.0, 0.019, 0.4783, -0.0, 0.0168, 0.4783, -0.0, 0.019, 0.4761, -0.0, 0.0222, 0.4783, -0.0, 0.0222, 0.4761, 0.0017, 0.0222, 0.4761, 0.0039, 0.0222,
0.4783, 0.0017, 0.0222, 0.4783, 0.0039, 0.0222, 0.4805, -0.0, 0.0168, 0.4805, -0.0, 0.019, 0.4826, -0.0, 0.0168, 0.4826, -0.0, 0.019, 0.4806, -0.0, 0.0221, 0.4828, -0.0, 0.0219, 0.4805, 0.0017, 0.0222, 0.4805, 0.0039, 0.0222, 0.4827, 0.0017, 0.0219, 0.4827, 0.0039, 0.0219, 0.4848, -0.0, 0.0168, 0.4848, -0.0, 0.019, 0.487, -0.0, 0.0168, 0.487, -0.0, 0.019, 0.485, -0.0, 0.0214, 0.4871, -0.0, 0.0208, 0.4848, 0.0017, 0.0214, 0.4848, 0.0039, 0.0214, 0.487, 0.0017, 0.0208, 0.487, 0.0039, 0.0208, 0.4544, -0.0, 0.0379, 0.4544, -0.0, 0.0407, 0.4566, -0.0, 0.0379, 0.4566, -0.0, 0.0407, 0.4544, 0.0017, 0.0379, 0.4544, 0.0039, 0.0379, 0.4566, 0.0017, 0.0379, 0.4566, 0.0039, 0.0379,
0.4588, -0.0, 0.0379, 0.4588, -0.0, 0.0407, 0.4609, -0.0, 0.0379, 0.4609, -0.0, 0.0407, 0.4588, 0.0017, 0.0379, 0.4588, 0.0039, 0.0379, 0.4609, 0.0017, 0.0379, 0.4609, 0.0039, 0.0379, 0.4544, -0.0, 0.0429, 0.4544, -0.0, 0.045, 0.4566, -0.0, 0.0429, 0.4566, -0.0, 0.045, 0.4544, -0.0, 0.0472, 0.4544, -0.0, 0.0487, 0.4566, -0.0, 0.0472, 0.4566, -0.0, 0.0487, 0.4544, 0.0017, 0.0487, 0.4544, 0.0039, 0.0487, 0.4566, 0.0017, 0.0487, 0.4566, 0.0039, 0.0487, 0.4588, -0.0, 0.0429, 0.4588, -0.0, 0.045, 0.4609, -0.0, 0.0429, 0.4609, -0.0, 0.045, 0.4588, -0.0, 0.0472, 0.4588, -0.0, 0.0487, 0.4609, -0.0, 0.0472, 0.4609, -0.0, 0.0487, 0.4588, 0.0017, 0.0487, 0.4588, 0.0039, 0.0487,
0.4609, 0.0017, 0.0487, 0.4609, 0.0039, 0.0487, 0.4631, -0.0, 0.0379, 0.4631, -0.0, 0.0407, 0.4653, -0.0, 0.0379, 0.4653, -0.0, 0.0407, 0.4631, 0.0017, 0.0379, 0.4631, 0.0039, 0.0379, 0.4653, 0.0017, 0.0379, 0.4653, 0.0039, 0.0379, 0.4675, -0.0, 0.0379, 0.4675, -0.0, 0.0407, 0.4696, -0.0, 0.0379, 0.4696, -0.0, 0.0407, 0.4675, 0.0017, 0.0379, 0.4675, 0.0039, 0.0379, 0.4696, 0.0017, 0.0379, 0.4696, 0.0039, 0.0379, 0.4631, -0.0, 0.0429, 0.4631, -0.0, 0.045, 0.4653, -0.0, 0.0429, 0.4653, -0.0, 0.045, 0.4631, -0.0, 0.0472, 0.4631, -0.0, 0.0487, 0.4653, -0.0, 0.0472, 0.4653, -0.0, 0.0487, 0.4631, 0.0017, 0.0487, 0.4631, 0.0039, 0.0487, 0.4653, 0.0017, 0.0487, 0.4653, 0.0039, 0.0487,
0.4675, -0.0, 0.0429, 0.4675, -0.0, 0.045, 0.4696, -0.0, 0.0429, 0.4696, -0.0, 0.045, 0.4675, -0.0, 0.0472, 0.4675, -0.0, 0.0487, 0.4696, -0.0, 0.0472, 0.4696, -0.0, 0.0487, 0.4675, 0.0017, 0.0487, 0.4675, 0.0039, 0.0487, 0.4696, 0.0017, 0.0487, 0.4696, 0.0039, 0.0487, 0.4718, -0.0, 0.0379, 0.4718, -0.0, 0.0407, 0.474, -0.0, 0.0379, 0.474, -0.0, 0.0407, 0.4718, 0.0017, 0.0379, 0.4718, 0.0039, 0.0379, 0.474, 0.0017, 0.0379, 0.474, 0.0039, 0.0379, 0.4761, -0.0, 0.0379, 0.4761, -0.0, 0.0407, 0.4783, -0.0, 0.0379, 0.4783, -0.0, 0.0407, 0.4761, 0.0017, 0.0379, 0.4761, 0.0039, 0.0379, 0.4783, 0.0017, 0.0379, 0.4783, 0.0039, 0.0379, 0.4718, -0.0, 0.0429, 0.4718, -0.0, 0.045,
0.474, -0.0, 0.0429, 0.474, -0.0, 0.045, 0.4718, -0.0, 0.0472, 0.4718, -0.0, 0.0487, 0.474, -0.0, 0.0472, 0.474, -0.0, 0.0487, 0.4718, 0.0017, 0.0487, 0.4718, 0.0039, 0.0487, 0.474, 0.0017, 0.0487, 0.474, 0.0039, 0.0487, 0.4761, -0.0, 0.0429, 0.4761, -0.0, 0.045, 0.4783, -0.0, 0.0429, 0.4783, -0.0, 0.045, 0.4761, -0.0, 0.0472, 0.4761, -0.0, 0.0487, 0.4783, -0.0, 0.0472, 0.4783, -0.0, 0.0487, 0.4761, 0.0017, 0.0487, 0.4761, 0.0039, 0.0487, 0.4783, 0.0017, 0.0487, 0.4783, 0.0039, 0.0487, 0.4805, -0.0, 0.0379, 0.4805, -0.0, 0.0407, 0.4826, -0.0, 0.0379, 0.4826, -0.0, 0.0407, 0.4805, 0.0017, 0.0379, 0.4805, 0.0039, 0.0379, 0.4826, 0.0017, 0.0379, 0.4826, 0.0039, 0.0379,
0.4848, -0.0, 0.0379, 0.4848, -0.0, 0.0407, 0.487, -0.0, 0.0379, 0.487, -0.0, 0.0407, 0.4848, 0.0017, 0.0379, 0.4848, 0.0039, 0.0379, 0.487, 0.0017, 0.0379, 0.487, 0.0039, 0.0379, 0.4805, -0.0, 0.0429, 0.4805, -0.0, 0.045, 0.4826, -0.0, 0.0429, 0.4826, -0.0, 0.045, 0.4805, -0.0, 0.0472, 0.4805, -0.0, 0.0487, 0.4826, -0.0, 0.0472, 0.4826, -0.0, 0.0487, 0.4805, 0.0017, 0.0487, 0.4805, 0.0039, 0.0487, 0.4826, 0.0017, 0.0487, 0.4826, 0.0039, 0.0487, 0.4848, -0.0, 0.0429, 0.4848, -0.0, 0.045, 0.487, -0.0, 0.0429, 0.487, -0.0, 0.045, 0.4848, -0.0, 0.0472, 0.4848, -0.0, 0.0487, 0.487, -0.0, 0.0472, 0.487, -0.0, 0.0487, 0.4848, 0.0017, 0.0487, 0.4848, 0.0039, 0.0487,
0.487, 0.0017, 0.0487, 0.487, 0.0039, 0.0487, 0.491, -0.0, -0.0004, 0.4911, -0.0, 0.0017, 0.4897, -0.0, 0.0043, 0.4888, -0.0, 0.0057, 0.4904, -0.0, 0.0033, 0.4913, -0.0, 0.006, 0.491, 0.0017, -0.0006, 0.4911, 0.0017, 0.0017, 0.491, 0.0039, -0.0006, 0.4911, 0.0039, 0.0017, 0.4898, 0.0017, 0.0042, 0.4888, 0.0017, 0.0057, 0.4898, 0.0039, 0.0042, 0.4888, 0.0039, 0.0057, 0.4905, 0.0017, 0.0032, 0.4905, 0.0039, 0.0032, 0.4935, -0.0, -0.0005, 0.4935, -0.0, 0.0016, 0.4957, -0.0, -0.0005, 0.4957, -0.0, 0.0016, 0.4935, -0.0, 0.0038, 0.4935, -0.0, 0.006, 0.4957, -0.0, 0.0038, 0.4957, -0.0, 0.006, 0.4892, -0.0, 0.0081, 0.4892, -0.0, 0.0103, 0.4913, -0.0, 0.0081, 0.4913, -0.0, 0.0103,
0.4892, -0.0, 0.0125, 0.4892, -0.0, 0.0146, 0.4913, -0.0, 0.0125, 0.4913, -0.0, 0.0146, 0.4935, -0.0, 0.0081, 0.4935, -0.0, 0.0103, 0.4957, -0.0, 0.0081, 0.4957, -0.0, 0.0103, 0.4935, -0.0, 0.0125, 0.4935, -0.0, 0.0146, 0.4957, -0.0, 0.0125, 0.4965, -0.0, 0.0154, 0.4965, 0.0017, 0.0155, 0.4965, 0.0039, 0.0155, 0.4978, -0.0, -0.0005, 0.4978, -0.0, 0.0016, 0.5, -0.0, -0.0005, 0.5, -0.0, 0.0016, 0.4978, -0.0, 0.0038, 0.4978, -0.0, 0.006, 0.5, -0.0, 0.0038, 0.5, -0.0, 0.006, 0.502, -0.0, -0.0005, 0.502, -0.0, 0.0016, 0.5019, -0.0, 0.0038, 0.5017, -0.0, 0.006, 0.502, 0.0017, -0.0005, 0.502, 0.0017, 0.0016, 0.502, 0.0039, -0.0005, 0.502, 0.0039, 0.0016,
0.5019, 0.0017, 0.0038, 0.5017, 0.0017, 0.006, 0.5019, 0.0039, 0.0038, 0.5017, 0.0039, 0.006, 0.4978, -0.0, 0.0081, 0.4978, -0.0, 0.0103, 0.501, -0.0, 0.0085, 0.5001, -0.0, 0.0106, 0.4986, -0.0, 0.013, 0.4975, -0.0, 0.0143, 0.4993, -0.0, 0.0118, 0.5009, 0.0017, 0.0087, 0.5002, 0.0017, 0.0103, 0.5009, 0.0039, 0.0087, 0.5002, 0.0039, 0.0103, 0.4985, 0.0017, 0.013, 0.4974, 0.0017, 0.0144, 0.4985, 0.0039, 0.013, 0.4974, 0.0039, 0.0144, 0.4992, 0.0017, 0.0119, 0.4992, 0.0039, 0.0119, 0.5013, -0.0, 0.0075, 0.5013, 0.0017, 0.0076, 0.5013, 0.0039, 0.0076, 0.4892, -0.0, 0.0168, 0.4895, -0.0, 0.0199, 0.4913, -0.0, 0.0168, 0.4916, -0.0, 0.019, 0.4884, -0.0, 0.0203, 0.4896, 0.0017, 0.0198,
0.4896, 0.0039, 0.0198, 0.4914, 0.0017, 0.0191, 0.4914, 0.0039, 0.0191, 0.4886, 0.0017, 0.0203, 0.4886, 0.0039, 0.0203, 0.4939, -0.0, 0.0176, 0.4928, -0.0, 0.0182, 0.4954, -0.0, 0.0165, 0.494, 0.0017, 0.0175, 0.4929, 0.0017, 0.0182, 0.494, 0.0039, 0.0175, 0.4929, 0.0039, 0.0182, 0.4954, 0.0017, 0.0165, 0.4954, 0.0039, 0.0165, 0.4892, -0.0, 0.0379, 0.4892, -0.0, 0.0407, 0.4913, -0.0, 0.0379, 0.4913, -0.0, 0.0407, 0.4892, 0.0017, 0.0379, 0.4892, 0.0039, 0.0379, 0.4913, 0.0017, 0.0379, 0.4913, 0.0039, 0.0379, 0.4935, -0.0, 0.0379, 0.4935, -0.0, 0.0407, 0.4957, -0.0, 0.0379, 0.4957, -0.0, 0.0407, 0.4935, 0.0017, 0.0379, 0.4935, 0.0039, 0.0379, 0.4957, 0.0017, 0.0379, 0.4957, 0.0039, 0.0379,
0.4892, -0.0, 0.0429, 0.4892, -0.0, 0.045, 0.4913, -0.0, 0.0429, 0.4913, -0.0, 0.045, 0.4892, -0.0, 0.0472, 0.4892, -0.0, 0.0487, 0.4913, -0.0, 0.0472, 0.4913, -0.0, 0.0487, 0.4892, 0.0017, 0.0487, 0.4892, 0.0039, 0.0487, 0.4913, 0.0017, 0.0487, 0.4913, 0.0039, 0.0487, 0.4935, -0.0, 0.0429, 0.4935, -0.0, 0.045, 0.4957, -0.0, 0.0429, 0.4957, -0.0, 0.045, 0.4935, -0.0, 0.0472, 0.4935, -0.0, 0.0487, 0.4957, -0.0, 0.0472, 0.4957, -0.0, 0.0487, 0.4935, 0.0017, 0.0487, 0.4935, 0.0039, 0.0487, 0.4957, 0.0017, 0.0487, 0.4957, 0.0039, 0.0487, 0.498, -0.0, 0.0379, 0.498, -0.0, 0.0407, 0.498, 0.0017, 0.0379, 0.498, 0.0017, 0.0407, 0.498, 0.0039, 0.0379, 0.498, 0.0039, 0.0407,
0.498, -0.0, 0.0429, 0.498, -0.0, 0.045, 0.498, -0.0, 0.0472, 0.498, -0.0, 0.0487, 0.498, 0.0017, 0.0429, 0.498, 0.0017, 0.045, 0.498, 0.0039, 0.0429, 0.498, 0.0039, 0.045, 0.498, 0.0017, 0.0472, 0.498, 0.0017, 0.0487, 0.498, 0.0039, 0.0472, 0.498, 0.0039, 0.0487, 0.0461, 0.0061, -0.067, 0.0461, 0.0083, -0.067, 0.0486, 0.0061, -0.067, 0.0486, 0.0083, -0.067, 0.0461, 0.0061, -0.0656, 0.0461, 0.0061, -0.0635, 0.0461, 0.0083, -0.0656, 0.0461, 0.0083, -0.0635, 0.0461, 0.01, -0.067, 0.0486, 0.01, -0.067, 0.0461, 0.01, -0.0656, 0.0461, 0.01, -0.0635, 0.0486, 0.01, -0.0656, 0.0486, 0.01, -0.0635, 0.05, 0.0061, -0.067, 0.05, 0.0083, -0.067, 0.05, 0.0061, -0.0656, 0.05, 0.0061, -0.0635,
0.05, 0.0083, -0.0656, 0.05, 0.0083, -0.0635, 0.05, 0.01, -0.067, 0.05, 0.01, -0.0656, 0.05, 0.01, -0.0635, 0.0461, 0.0061, -0.0613, 0.0461, 0.0061, -0.0591, 0.0461, 0.0083, -0.0613, 0.0461, 0.0083, -0.0591, 0.0461, 0.0061, -0.057, 0.0461, 0.0061, -0.0548, 0.0461, 0.0083, -0.057, 0.0461, 0.0083, -0.0548, 0.0461, 0.01, -0.0613, 0.0461, 0.01, -0.0591, 0.0486, 0.01, -0.0613, 0.0486, 0.01, -0.0591, 0.0461, 0.01, -0.057, 0.0461, 0.01, -0.0548, 0.0486, 0.01, -0.057, 0.0486, 0.01, -0.0548, 0.05, 0.0061, -0.0613, 0.05, 0.0061, -0.0591, 0.05, 0.0083, -0.0613, 0.05, 0.0083, -0.0591, 0.05, 0.0061, -0.057, 0.05, 0.0061, -0.0548, 0.05, 0.0083, -0.057, 0.05, 0.0083, -0.0548, 0.0529, 0.0061, -0.0562,
0.0529, 0.0061, -0.0548, 0.0529, 0.0083, -0.0562, 0.0529, 0.0083, -0.0548, 0.05, 0.01, -0.0613, 0.05, 0.01, -0.0591, 0.05, 0.01, -0.057, 0.05, 0.01, -0.0548, 0.0529, 0.01, -0.0562, 0.0529, 0.01, -0.0548, 0.0461, 0.0061, -0.0526, 0.0461, 0.0061, -0.0505, 0.0461, 0.0083, -0.0526, 0.0461, 0.0083, -0.0505, 0.0461, 0.0061, -0.0483, 0.0461, 0.0061, -0.0461, 0.0461, 0.0083, -0.0483, 0.0461, 0.0083, -0.0461, 0.0461, 0.01, -0.0526, 0.0461, 0.01, -0.0505, 0.0486, 0.01, -0.0526, 0.0486, 0.01, -0.0505, 0.0461, 0.01, -0.0483, 0.0461, 0.01, -0.0461, 0.0486, 0.01, -0.0483, 0.0486, 0.01, -0.0461, 0.05, 0.0061, -0.0526, 0.05, 0.0061, -0.0505, 0.05, 0.0083, -0.0526, 0.05, 0.0083, -0.0505, 0.0529, 0.0061, -0.0523,
0.0529, 0.0083, -0.0523, 0.05, 0.0061, -0.0483, 0.05, 0.0061, -0.0454, 0.05, 0.0083, -0.0483, 0.05, 0.0083, -0.0454, 0.053, 0.0061, -0.0454, 0.053, 0.0083, -0.0454, 0.05, 0.01, -0.0526, 0.05, 0.01, -0.0505, 0.0529, 0.01, -0.0523, 0.05, 0.01, -0.0483, 0.05, 0.01, -0.0454, 0.053, 0.01, -0.0454, 0.0461, 0.0061, -0.0439, 0.0461, 0.0061, -0.0425, 0.0461, 0.0083, -0.0439, 0.0461, 0.0083, -0.0425, 0.0486, 0.0061, -0.0425, 0.0486, 0.0083, -0.0425, 0.0461, 0.01, -0.0439, 0.0461, 0.01, -0.0425, 0.0486, 0.01, -0.0439, 0.0486, 0.01, -0.0425, 0.0508, 0.0061, -0.0425, 0.0508, 0.0083, -0.0425, 0.053, 0.0061, -0.0425, 0.053, 0.0083, -0.0425, 0.0508, 0.01, -0.0439, 0.0508, 0.01, -0.0425, 0.053, 0.01, -0.0439,
0.053, 0.01, -0.0425, 0.0551, 0.0061, -0.0562, 0.0551, 0.0083, -0.0562, 0.0573, 0.0061, -0.0562, 0.0573, 0.0083, -0.0562, 0.0551, 0.01, -0.0562, 0.0551, 0.01, -0.0548, 0.0573, 0.01, -0.0562, 0.0573, 0.01, -0.0548, 0.0595, 0.0061, -0.0562, 0.0595, 0.0083, -0.0562, 0.0616, 0.0061, -0.0562, 0.0616, 0.0083, -0.0562, 0.0595, 0.01, -0.0562, 0.0595, 0.01, -0.0548, 0.0616, 0.01, -0.0562, 0.0616, 0.01, -0.0548, 0.0638, 0.0061, -0.0562, 0.0638, 0.0083, -0.0562, 0.0657, 0.0061, -0.0562, 0.0657, 0.0061, -0.0548, 0.0657, 0.0083, -0.0562, 0.0657, 0.0083, -0.0548, 0.0638, 0.01, -0.0562, 0.0638, 0.01, -0.0548, 0.0657, 0.01, -0.0562, 0.0657, 0.01, -0.0548, 0.0551, 0.0061, -0.0523, 0.0551, 0.0083, -0.0523, 0.0573, 0.0061, -0.0523,
0.0573, 0.0083, -0.0523, 0.0551, 0.0061, -0.0454, 0.0551, 0.0083, -0.0454, 0.0573, 0.0061, -0.0454, 0.0573, 0.0083, -0.0454, 0.0551, 0.01, -0.0523, 0.0573, 0.01, -0.0523, 0.0551, 0.01, -0.0454, 0.0573, 0.01, -0.0454, 0.0595, 0.0061, -0.0523, 0.0595, 0.0083, -0.0523, 0.0616, 0.0061, -0.0523, 0.0616, 0.0083, -0.0523, 0.0595, 0.0061, -0.0454, 0.0595, 0.0083, -0.0454, 0.0616, 0.0061, -0.0454, 0.0616, 0.0083, -0.0454, 0.0595, 0.01, -0.0523, 0.0616, 0.01, -0.0523, 0.0595, 0.01, -0.0454, 0.0616, 0.01, -0.0454, 0.0551, 0.0061, -0.0425, 0.0551, 0.0083, -0.0425, 0.0573, 0.0061, -0.0425, 0.0573, 0.0083, -0.0425, 0.0551, 0.01, -0.0439, 0.0551, 0.01, -0.0425, 0.0573, 0.01, -0.0439, 0.0573, 0.01, -0.0425, 0.0595, 0.0061, -0.0425,
0.0595, 0.0083, -0.0425, 0.0616, 0.0061, -0.0425, 0.0616, 0.0083, -0.0425, 0.0595, 0.01, -0.0439, 0.0595, 0.01, -0.0425, 0.0616, 0.01, -0.0439, 0.0616, 0.01, -0.0425, 0.0638, 0.0061, -0.0523, 0.0638, 0.0083, -0.0523, 0.0657, 0.0061, -0.0523, 0.0657, 0.0083, -0.0523, 0.0638, 0.0061, -0.0454, 0.0638, 0.0083, -0.0454, 0.0667, 0.0061, -0.0454, 0.0667, 0.0083, -0.0454, 0.0638, 0.01, -0.0523, 0.0657, 0.01, -0.0523, 0.0638, 0.01, -0.0454, 0.0667, 0.01, -0.0454, 0.0638, 0.0061, -0.0425, 0.0638, 0.0083, -0.0425, 0.0667, 0.0061, -0.0439, 0.0667, 0.0061, -0.0425, 0.0667, 0.0083, -0.0439, 0.0667, 0.0083, -0.0425, 0.0638, 0.01, -0.0439, 0.0638, 0.01, -0.0425, 0.0667, 0.01, -0.0439, 0.0667, 0.01, -0.0425, 0.0887, 0.0061, -0.0625,
0.0887, 0.0083, -0.0625, 0.0887, 0.01, -0.0624, 0.0879, 0.0061, -0.0614, 0.0869, 0.0061, -0.0592, 0.0879, 0.0083, -0.0614, 0.0869, 0.0083, -0.0592, 0.0864, 0.0061, -0.0568, 0.0863, 0.0061, -0.0548, 0.0864, 0.0083, -0.0568, 0.0863, 0.0083, -0.0548, 0.0866, 0.0061, -0.0579, 0.0866, 0.0083, -0.0579, 0.088, 0.01, -0.0616, 0.087, 0.01, -0.0594, 0.0864, 0.01, -0.0565, 0.0863, 0.01, -0.0548, 0.0865, 0.01, -0.0573, 0.0877, 0.01, -0.0548, 0.0863, 0.0061, -0.0526, 0.0865, 0.0061, -0.0511, 0.0863, 0.0083, -0.0526, 0.0865, 0.0083, -0.0511, 0.0867, 0.0061, -0.0501, 0.0867, 0.0083, -0.0501, 0.0874, 0.0061, -0.0482, 0.0884, 0.0061, -0.0468, 0.0874, 0.0083, -0.0482, 0.0884, 0.0083, -0.0468, 0.0863, 0.01, -0.0526, 0.0865, 0.01, -0.0513,
0.0877, 0.01, -0.0526, 0.0867, 0.01, -0.0503, 0.0875, 0.01, -0.048, 0.0884, 0.01, -0.0468, 0.0909, 0.0061, -0.0646, 0.0897, 0.0061, -0.0637, 0.0909, 0.0083, -0.0646, 0.0897, 0.0083, -0.0637, 0.092, 0.0061, -0.0653, 0.092, 0.0083, -0.0653, 0.0909, 0.01, -0.0646, 0.0897, 0.01, -0.0637, 0.0918, 0.01, -0.0652, 0.092, 0.01, -0.0635, 0.097, 0.0061, -0.0668, 0.097, 0.0083, -0.0668, 0.0942, 0.0061, -0.0662, 0.0942, 0.0083, -0.0662, 0.0959, 0.0061, -0.0666, 0.0964, 0.0061, -0.0627, 0.0959, 0.0083, -0.0666, 0.0964, 0.0083, -0.0627, 0.0972, 0.01, -0.0668, 0.0939, 0.01, -0.0661, 0.0942, 0.01, -0.0635, 0.0962, 0.01, -0.0667, 0.0963, 0.01, -0.0627, 0.0908, 0.0061, -0.0583, 0.0908, 0.0083, -0.0583, 0.0926, 0.0061, -0.0608,
0.0914, 0.0061, -0.0594, 0.0926, 0.0083, -0.0608, 0.0914, 0.0083, -0.0594, 0.0904, 0.0061, -0.057, 0.0903, 0.0061, -0.0548, 0.0904, 0.0083, -0.057, 0.0903, 0.0083, -0.0548, 0.0899, 0.01, -0.0613, 0.091, 0.01, -0.0586, 0.0926, 0.01, -0.0608, 0.0915, 0.01, -0.0595, 0.0904, 0.01, -0.0568, 0.0902, 0.01, -0.0548, 0.0942, 0.0061, -0.0621, 0.0942, 0.0083, -0.0621, 0.0953, 0.0061, -0.0624, 0.0953, 0.0083, -0.0624, 0.0942, 0.01, -0.0621, 0.0953, 0.01, -0.0624, 0.0985, 0.0061, -0.0669, 0.0985, 0.0083, -0.0669, 0.1007, 0.0061, -0.067, 0.1007, 0.0083, -0.067, 0.0985, 0.0061, -0.063, 0.0985, 0.0083, -0.063, 0.1007, 0.0061, -0.0631, 0.1007, 0.0083, -0.0631, 0.0985, 0.01, -0.0669, 0.1007, 0.01, -0.067, 0.0985, 0.01, -0.0656,
0.0986, 0.01, -0.063, 0.1007, 0.01, -0.0656, 0.1007, 0.01, -0.0631, 0.1029, 0.0061, -0.067, 0.1029, 0.0083, -0.067, 0.105, 0.0061, -0.067, 0.105, 0.0083, -0.067, 0.1029, 0.0061, -0.0631, 0.1029, 0.0083, -0.0631, 0.105, 0.0061, -0.063, 0.105, 0.0083, -0.063, 0.1029, 0.01, -0.067, 0.105, 0.01, -0.0669, 0.1029, 0.01, -0.0656, 0.1029, 0.01, -0.0631, 0.105, 0.01, -0.0656, 0.105, 0.01, -0.063, 0.0904, 0.0061, -0.0526, 0.0908, 0.0061, -0.0512, 0.0904, 0.0083, -0.0526, 0.0908, 0.0083, -0.0512, 0.0912, 0.0061, -0.05, 0.0912, 0.0083, -0.05, 0.0894, 0.0061, -0.0456, 0.0894, 0.0083, -0.0456, 0.0923, 0.0061, -0.0485, 0.0923, 0.0083, -0.0485, 0.0904, 0.01, -0.0528, 0.0909, 0.01, -0.0508, 0.0913, 0.01, -0.0499,
0.0899, 0.01, -0.0483, 0.0894, 0.01, -0.0456, 0.0922, 0.01, -0.0485, 0.092, 0.01, -0.0461, 0.0935, 0.0061, -0.0474, 0.0946, 0.0061, -0.0468, 0.0935, 0.0083, -0.0474, 0.0946, 0.0083, -0.0468, 0.0963, 0.0061, -0.0461, 0.0963, 0.0083, -0.0461, 0.0934, 0.01, -0.0475, 0.0945, 0.01, -0.0469, 0.0965, 0.01, -0.046, 0.0905, 0.0061, -0.0447, 0.0905, 0.0083, -0.0447, 0.092, 0.0061, -0.0437, 0.092, 0.0083, -0.0437, 0.0905, 0.01, -0.0447, 0.0917, 0.01, -0.0438, 0.0937, 0.0061, -0.043, 0.0948, 0.0061, -0.0427, 0.0937, 0.0083, -0.043, 0.0948, 0.0083, -0.0427, 0.0964, 0.0061, -0.0425, 0.0964, 0.0083, -0.0425, 0.0939, 0.01, -0.0429, 0.0949, 0.01, -0.0427, 0.0964, 0.01, -0.0439, 0.0963, 0.01, -0.0425, 0.0985, 0.0061, -0.0456,
0.0985, 0.0083, -0.0456, 0.1007, 0.0061, -0.0454, 0.1007, 0.0083, -0.0454, 0.0986, 0.01, -0.0456, 0.1007, 0.01, -0.0454, 0.1029, 0.0061, -0.0455, 0.1029, 0.0083, -0.0455, 0.1051, 0.0061, -0.0457, 0.1051, 0.0083, -0.0457, 0.1029, 0.01, -0.0455, 0.105, 0.01, -0.0457, 0.0985, 0.0061, -0.0425, 0.0985, 0.0083, -0.0425, 0.1007, 0.0061, -0.0425, 0.1007, 0.0083, -0.0425, 0.0985, 0.01, -0.0439, 0.0985, 0.01, -0.0425, 0.1007, 0.01, -0.0439, 0.1007, 0.01, -0.0425, 0.1029, 0.0061, -0.0425, 0.1029, 0.0083, -0.0425, 0.105, 0.0061, -0.0425, 0.105, 0.0083, -0.0425, 0.1029, 0.01, -0.0439, 0.1029, 0.01, -0.0425, 0.105, 0.01, -0.0439, 0.105, 0.01, -0.0425, 0.0716, 0.0061, -0.0219, 0.0716, 0.0061, -0.0201, 0.0716, 0.0083, -0.0219,
0.0716, 0.0083, -0.0201, 0.0747, 0.0061, -0.0219, 0.0747, 0.0083, -0.0219, 0.0716, 0.01, -0.0219, 0.0716, 0.01, -0.0201, 0.0747, 0.01, -0.0219, 0.0747, 0.01, -0.0201, 0.0768, 0.0061, -0.0219, 0.0768, 0.0083, -0.0219, 0.079, 0.0061, -0.0219, 0.079, 0.0083, -0.0219, 0.0768, 0.01, -0.0219, 0.0768, 0.01, -0.0201, 0.079, 0.01, -0.0219, 0.079, 0.01, -0.0201, 0.0812, 0.0061, -0.0219, 0.0812, 0.0083, -0.0219, 0.0833, 0.0061, -0.0219, 0.0833, 0.0061, -0.0201, 0.0833, 0.0083, -0.0219, 0.0833, 0.0083, -0.0201, 0.0812, 0.01, -0.0219, 0.0812, 0.01, -0.0201, 0.0833, 0.01, -0.0219, 0.0833, 0.01, -0.0201, 0.0716, 0.0061, -0.0179, 0.0716, 0.0061, -0.0157, 0.0716, 0.0083, -0.0179, 0.0716, 0.0083, -0.0157, 0.0716, 0.0061, -0.0136,
0.0716, 0.0061, -0.0114, 0.0716, 0.0083, -0.0136, 0.0716, 0.0083, -0.0114, 0.0716, 0.01, -0.0179, 0.0716, 0.01, -0.0157, 0.0747, 0.01, -0.0179, 0.0747, 0.01, -0.0157, 0.0716, 0.01, -0.0136, 0.0716, 0.01, -0.0114, 0.0747, 0.01, -0.0136, 0.0747, 0.01, -0.0114, 0.0768, 0.01, -0.0179, 0.0768, 0.01, -0.0157, 0.079, 0.01, -0.0179, 0.079, 0.01, -0.0157, 0.0768, 0.01, -0.0136, 0.0768, 0.01, -0.0114, 0.079, 0.01, -0.0136, 0.079, 0.01, -0.0114, 0.0716, 0.0061, -0.0092, 0.0716, 0.0061, -0.0071, 0.0716, 0.0083, -0.0092, 0.0716, 0.0083, -0.0071, 0.0716, 0.0061, -0.0049, 0.0716, 0.0061, -0.0027, 0.0716, 0.0083, -0.0049, 0.0716, 0.0083, -0.0027, 0.0716, 0.01, -0.0092, 0.0716, 0.01, -0.0071, 0.0747, 0.01, -0.0092,
0.0747, 0.01, -0.0071, 0.0716, 0.01, -0.0049, 0.0716, 0.01, -0.0027, 0.0747, 0.01, -0.0049, 0.0747, 0.01, -0.0027, 0.0768, 0.01, -0.0092, 0.0768, 0.01, -0.0071, 0.079, 0.01, -0.0092, 0.079, 0.01, -0.0071, 0.0768, 0.01, -0.0049, 0.0768, 0.01, -0.0027, 0.079, 0.01, -0.0049, 0.079, 0.01, -0.0027, 0.0833, 0.0061, -0.0179, 0.0833, 0.0061, -0.0157, 0.0833, 0.0083, -0.0179, 0.0833, 0.0083, -0.0157, 0.0833, 0.0061, -0.0136, 0.0833, 0.0061, -0.0114, 0.0833, 0.0083, -0.0136, 0.0833, 0.0083, -0.0114, 0.0812, 0.01, -0.0179, 0.0812, 0.01, -0.0157, 0.0833, 0.01, -0.0179, 0.0833, 0.01, -0.0157, 0.0812, 0.01, -0.0136, 0.0812, 0.01, -0.0114, 0.0833, 0.01, -0.0136, 0.0833, 0.01, -0.0114, 0.0833, 0.0061, -0.0092,
0.0833, 0.0061, -0.0071, 0.0833, 0.0083, -0.0092, 0.0833, 0.0083, -0.0071, 0.0833, 0.0061, -0.0049, 0.0833, 0.0061, -0.0027, 0.0833, 0.0083, -0.0049, 0.0833, 0.0083, -0.0027, 0.0812, 0.01, -0.0092, 0.0812, 0.01, -0.0071, 0.0833, 0.01, -0.0092, 0.0833, 0.01, -0.0071, 0.0812, 0.01, -0.0049, 0.0812, 0.01, -0.0027, 0.0833, 0.01, -0.0049, 0.0833, 0.01, -0.0027, 0.1066, 0.0061, -0.0668, 0.1066, 0.0083, -0.0668, 0.1077, 0.0061, -0.0666, 0.1071, 0.0061, -0.0627, 0.1077, 0.0083, -0.0666, 0.1071, 0.0083, -0.0627, 0.1094, 0.0061, -0.0662, 0.1094, 0.0083, -0.0662, 0.1065, 0.01, -0.0668, 0.1074, 0.01, -0.0667, 0.1072, 0.01, -0.0626, 0.1097, 0.01, -0.0661, 0.1094, 0.01, -0.0635, 0.1115, 0.0061, -0.0652, 0.1125, 0.0061, -0.0645,
0.1115, 0.0083, -0.0652, 0.1125, 0.0083, -0.0645, 0.1138, 0.0061, -0.0635, 0.1138, 0.0083, -0.0635, 0.1113, 0.01, -0.0653, 0.1123, 0.01, -0.0646, 0.1138, 0.01, -0.0636, 0.1082, 0.0061, -0.0624, 0.1082, 0.0083, -0.0624, 0.1094, 0.0061, -0.062, 0.1094, 0.0083, -0.062, 0.1082, 0.01, -0.0623, 0.1093, 0.01, -0.062, 0.1111, 0.0061, -0.0608, 0.1122, 0.0061, -0.0596, 0.1111, 0.0083, -0.0608, 0.1122, 0.0083, -0.0596, 0.1148, 0.0061, -0.0623, 0.1129, 0.0061, -0.0584, 0.1148, 0.0083, -0.0623, 0.1129, 0.0083, -0.0584, 0.1134, 0.0061, -0.057, 0.1136, 0.0061, -0.0548, 0.1134, 0.0083, -0.057, 0.1136, 0.0083, -0.0548, 0.1112, 0.01, -0.0608, 0.1122, 0.01, -0.0596, 0.1148, 0.01, -0.0622, 0.1127, 0.01, -0.0587, 0.1134, 0.01, -0.0569,
0.1136, 0.01, -0.0548, 0.1155, 0.0061, -0.0613, 0.1167, 0.0061, -0.0593, 0.1155, 0.0083, -0.0613, 0.1167, 0.0083, -0.0593, 0.117, 0.0061, -0.0581, 0.117, 0.0083, -0.0581, 0.1173, 0.0061, -0.057, 0.1176, 0.0061, -0.0548, 0.1173, 0.0083, -0.057, 0.1176, 0.0083, -0.0548, 0.1156, 0.01, -0.0612, 0.1167, 0.01, -0.0593, 0.117, 0.01, -0.0581, 0.1159, 0.01, -0.057, 0.1159, 0.01, -0.0548, 0.1173, 0.01, -0.057, 0.1176, 0.01, -0.0548, 0.1072, 0.0061, -0.0463, 0.1072, 0.0083, -0.0463, 0.1099, 0.0061, -0.0476, 0.1088, 0.0061, -0.047, 0.1099, 0.0083, -0.0476, 0.1088, 0.0083, -0.047, 0.107, 0.01, -0.0462, 0.11, 0.01, -0.0476, 0.109, 0.01, -0.047, 0.1123, 0.0061, -0.0499, 0.1123, 0.0083, -0.0499, 0.1134, 0.0061, -0.0526,
0.1129, 0.0061, -0.051, 0.1134, 0.0083, -0.0526, 0.1129, 0.0083, -0.051, 0.1112, 0.0061, -0.0486, 0.1112, 0.0083, -0.0486, 0.1142, 0.0061, -0.0456, 0.1142, 0.0083, -0.0456, 0.1122, 0.01, -0.0498, 0.1134, 0.01, -0.0528, 0.1128, 0.01, -0.0508, 0.1112, 0.01, -0.0486, 0.1116, 0.01, -0.0461, 0.1137, 0.01, -0.0483, 0.1142, 0.01, -0.0456, 0.1072, 0.0061, -0.0426, 0.1072, 0.0083, -0.0426, 0.1098, 0.0061, -0.043, 0.1088, 0.0061, -0.0427, 0.1098, 0.0083, -0.043, 0.1088, 0.0083, -0.0427, 0.1072, 0.01, -0.0439, 0.1072, 0.01, -0.0426, 0.1096, 0.01, -0.043, 0.1086, 0.01, -0.0427, 0.1116, 0.0061, -0.0436, 0.1116, 0.0083, -0.0436, 0.1131, 0.0061, -0.0447, 0.1131, 0.0083, -0.0447, 0.1119, 0.01, -0.0437, 0.1131, 0.01, -0.0447,
0.1168, 0.0061, -0.0498, 0.1168, 0.0083, -0.0498, 0.1175, 0.0061, -0.0526, 0.1172, 0.0061, -0.0509, 0.1175, 0.0083, -0.0526, 0.1172, 0.0083, -0.0509, 0.1162, 0.0061, -0.0482, 0.1152, 0.0061, -0.0467, 0.1162, 0.0083, -0.0482, 0.1152, 0.0083, -0.0467, 0.1159, 0.01, -0.0526, 0.1169, 0.01, -0.0501, 0.1175, 0.01, -0.0526, 0.1172, 0.01, -0.0511, 0.1161, 0.01, -0.048, 0.1152, 0.01, -0.0468, 0.1363, 0.0061, -0.067, 0.1363, 0.0083, -0.067, 0.1363, 0.0061, -0.0656, 0.1363, 0.0061, -0.0635, 0.1363, 0.0083, -0.0656, 0.1363, 0.0083, -0.0635, 0.1363, 0.01, -0.067, 0.1363, 0.01, -0.0656, 0.1363, 0.01, -0.0635, 0.1376, 0.0061, -0.067, 0.1376, 0.0083, -0.067, 0.1402, 0.0061, -0.067, 0.1402, 0.0083, -0.067, 0.1402, 0.0061, -0.0656,
0.1402, 0.0061, -0.0635, 0.1402, 0.0083, -0.0656, 0.1402, 0.0083, -0.0635, 0.1376, 0.01, -0.067, 0.1402, 0.01, -0.067, 0.1376, 0.01, -0.0656, 0.1376, 0.01, -0.0635, 0.1402, 0.01, -0.0656, 0.1402, 0.01, -0.0635, 0.1363, 0.0061, -0.0613, 0.1363, 0.0061, -0.0591, 0.1363, 0.0083, -0.0613, 0.1363, 0.0083, -0.0591, 0.1363, 0.0061, -0.057, 0.1363, 0.0061, -0.0548, 0.1363, 0.0083, -0.057, 0.1363, 0.0083, -0.0548, 0.1363, 0.01, -0.0613, 0.1363, 0.01, -0.0591, 0.1363, 0.01, -0.057, 0.1363, 0.01, -0.0548, 0.1402, 0.0061, -0.0613, 0.1402, 0.0061, -0.0591, 0.1402, 0.0083, -0.0613, 0.1402, 0.0083, -0.0591, 0.1402, 0.0061, -0.057, 0.1402, 0.0061, -0.0548, 0.1402, 0.0083, -0.057, 0.1402, 0.0083, -0.0548, 0.1376, 0.01, -0.0613,
0.1376, 0.01, -0.0591, 0.1402, 0.01, -0.0613, 0.1402, 0.01, -0.0591, 0.1376, 0.01, -0.057, 0.1376, 0.01, -0.0548, 0.1402, 0.01, -0.057, 0.1402, 0.01, -0.0548, 0.1363, 0.0061, -0.0526, 0.1363, 0.0061, -0.0505, 0.1363, 0.0083, -0.0526, 0.1363, 0.0083, -0.0505, 0.1363, 0.0061, -0.0483, 0.1363, 0.0061, -0.0461, 0.1363, 0.0083, -0.0483, 0.1363, 0.0083, -0.0461, 0.1363, 0.01, -0.0526, 0.1363, 0.01, -0.0505, 0.1363, 0.01, -0.0483, 0.1363, 0.01, -0.0461, 0.1402, 0.0061, -0.0526, 0.1402, 0.0061, -0.0505, 0.1402, 0.0083, -0.0526, 0.1402, 0.0083, -0.0505, 0.1402, 0.0061, -0.0483, 0.1402, 0.0061, -0.0454, 0.1402, 0.0083, -0.0483, 0.1402, 0.0083, -0.0454, 0.1376, 0.01, -0.0526, 0.1376, 0.01, -0.0505, 0.1402, 0.01, -0.0526,
0.1402, 0.01, -0.0505, 0.1376, 0.01, -0.0483, 0.1376, 0.01, -0.0461, 0.1402, 0.01, -0.0483, 0.1402, 0.01, -0.0454, 0.1363, 0.0061, -0.0439, 0.1363, 0.0061, -0.0425, 0.1363, 0.0083, -0.0439, 0.1363, 0.0083, -0.0425, 0.1363, 0.01, -0.0439, 0.1363, 0.01, -0.0425, 0.1376, 0.0061, -0.0425, 0.1376, 0.0083, -0.0425, 0.1398, 0.0061, -0.0425, 0.1398, 0.0083, -0.0425, 0.1376, 0.01, -0.0439, 0.1376, 0.01, -0.0425, 0.1398, 0.01, -0.0439, 0.1398, 0.01, -0.0425, 0.1431, 0.0061, -0.0591, 0.1431, 0.0083, -0.0591, 0.1431, 0.0061, -0.0562, 0.1431, 0.0083, -0.0562, 0.1431, 0.01, -0.0591, 0.1431, 0.01, -0.0562, 0.1463, 0.0061, -0.0591, 0.1463, 0.0083, -0.0591, 0.1485, 0.0061, -0.0592, 0.1485, 0.0083, -0.0592, 0.1463, 0.0061, -0.0562,
0.1463, 0.0083, -0.0562, 0.1484, 0.0061, -0.0562, 0.1484, 0.0083, -0.0562, 0.1463, 0.01, -0.0591, 0.1484, 0.01, -0.0592, 0.1463, 0.01, -0.0562, 0.1484, 0.01, -0.0562, 0.1553, 0.0061, -0.0632, 0.1553, 0.0083, -0.0632, 0.1575, 0.0061, -0.0651, 0.1564, 0.0061, -0.0643, 0.1575, 0.0083, -0.0651, 0.1564, 0.0083, -0.0643, 0.1553, 0.01, -0.0632, 0.1576, 0.01, -0.0652, 0.1564, 0.01, -0.0643, 0.1506, 0.0061, -0.0596, 0.1506, 0.0083, -0.0596, 0.1529, 0.0061, -0.0609, 0.1518, 0.0061, -0.0602, 0.1529, 0.0083, -0.0609, 0.1518, 0.0083, -0.0602, 0.1506, 0.0061, -0.0562, 0.1506, 0.0083, -0.0562, 0.1528, 0.0061, -0.0562, 0.1528, 0.0083, -0.0562, 0.1504, 0.01, -0.0596, 0.153, 0.01, -0.061, 0.1521, 0.01, -0.0603, 0.1506, 0.01, -0.0562,
0.1527, 0.01, -0.0562, 0.1542, 0.0061, -0.0621, 0.1542, 0.0083, -0.0621, 0.1546, 0.0061, -0.0561, 0.1556, 0.0061, -0.0558, 0.1546, 0.0083, -0.0561, 0.1556, 0.0083, -0.0558, 0.1572, 0.0061, -0.0553, 0.1572, 0.0083, -0.0553, 0.1542, 0.01, -0.0621, 0.155, 0.01, -0.0591, 0.1571, 0.01, -0.0613, 0.1571, 0.01, -0.0591, 0.1549, 0.01, -0.056, 0.1557, 0.01, -0.0557, 0.1571, 0.01, -0.057, 0.1571, 0.01, -0.0553, 0.1419, 0.0061, -0.0454, 0.1419, 0.0083, -0.0454, 0.1441, 0.0061, -0.0454, 0.1441, 0.0083, -0.0454, 0.1419, 0.01, -0.0454, 0.1441, 0.01, -0.0454, 0.1463, 0.0061, -0.0454, 0.1463, 0.0083, -0.0454, 0.1484, 0.0061, -0.0454, 0.1484, 0.0083, -0.0454, 0.1463, 0.01, -0.0454, 0.1484, 0.01, -0.0454, 0.1419, 0.0061, -0.0425,
0.1419, 0.0083, -0.0425, 0.1441, 0.0061, -0.0425, 0.1441, 0.0083, -0.0425, 0.1419, 0.01, -0.0439, 0.1419, 0.01, -0.0425, 0.1441, 0.01, -0.0439, 0.1441, 0.01, -0.0425, 0.1463, 0.0061, -0.0425, 0.1463, 0.0083, -0.0425, 0.1484, 0.0061, -0.0425, 0.1484, 0.0083, -0.0425, 0.1463, 0.01, -0.0439, 0.1463, 0.01, -0.0425, 0.1484, 0.01, -0.0439, 0.1484, 0.01, -0.0425, 0.1506, 0.0061, -0.0454, 0.1506, 0.0083, -0.0454, 0.1528, 0.0061, -0.0456, 0.1528, 0.0083, -0.0456, 0.1506, 0.01, -0.0454, 0.1528, 0.01, -0.0456, 0.155, 0.0061, -0.0459, 0.155, 0.0083, -0.0459, 0.1579, 0.0061, -0.0474, 0.1568, 0.0061, -0.0467, 0.1579, 0.0083, -0.0474, 0.1568, 0.0083, -0.0467, 0.1549, 0.01, -0.0459, 0.1579, 0.01, -0.0475, 0.1569, 0.01, -0.0468,
0.1506, 0.0061, -0.0425, 0.1506, 0.0083, -0.0425, 0.1528, 0.0061, -0.0425, 0.1528, 0.0083, -0.0425, 0.1506, 0.01, -0.0439, 0.1506, 0.01, -0.0425, 0.1528, 0.01, -0.0439, 0.1528, 0.01, -0.0425, 0.155, 0.0061, -0.0425, 0.155, 0.0083, -0.0425, 0.1571, 0.0061, -0.0426, 0.1571, 0.0083, -0.0426, 0.155, 0.01, -0.0439, 0.155, 0.01, -0.0425, 0.1571, 0.01, -0.0439, 0.1571, 0.01, -0.0426, 0.1619, 0.0061, -0.0669, 0.1619, 0.0083, -0.0669, 0.1592, 0.0061, -0.0662, 0.1592, 0.0083, -0.0662, 0.1608, 0.0061, -0.0666, 0.1619, 0.0061, -0.0632, 0.1608, 0.0083, -0.0666, 0.1619, 0.0083, -0.0632, 0.1621, 0.01, -0.0669, 0.159, 0.01, -0.0661, 0.1593, 0.01, -0.0635, 0.1612, 0.01, -0.0667, 0.1619, 0.01, -0.0632, 0.1648, 0.0061, -0.0671,
0.1648, 0.0083, -0.0671, 0.1638, 0.0061, -0.0656, 0.1628, 0.0061, -0.0643, 0.1638, 0.0083, -0.0656, 0.1628, 0.0083, -0.0643, 0.1648, 0.01, -0.0671, 0.1636, 0.01, -0.0653, 0.1628, 0.01, -0.0643, 0.1599, 0.0061, -0.0608, 0.1588, 0.0061, -0.0591, 0.1599, 0.0083, -0.0608, 0.1588, 0.0083, -0.0591, 0.1608, 0.0061, -0.0619, 0.1607, 0.0061, -0.0583, 0.1608, 0.0083, -0.0619, 0.1607, 0.0083, -0.0583, 0.1587, 0.0061, -0.0543, 0.1587, 0.0083, -0.0543, 0.1618, 0.0061, -0.0574, 0.1618, 0.0083, -0.0574, 0.16, 0.01, -0.0609, 0.1588, 0.01, -0.0591, 0.1608, 0.01, -0.0619, 0.1606, 0.01, -0.0583, 0.1593, 0.01, -0.057, 0.1588, 0.01, -0.0543, 0.1618, 0.01, -0.0574, 0.1615, 0.01, -0.0548, 0.1629, 0.0061, -0.0563, 0.1638, 0.0061, -0.0549,
0.1629, 0.0083, -0.0563, 0.1638, 0.0083, -0.0549, 0.1629, 0.01, -0.0563, 0.1637, 0.01, -0.0551, 0.1596, 0.0061, -0.053, 0.1603, 0.0061, -0.0503, 0.1596, 0.0083, -0.053, 0.1603, 0.0083, -0.0503, 0.1585, 0.0061, -0.0479, 0.1585, 0.0083, -0.0479, 0.1598, 0.01, -0.0525, 0.1603, 0.01, -0.0503, 0.1615, 0.01, -0.0526, 0.1615, 0.01, -0.0505, 0.1591, 0.01, -0.0484, 0.1593, 0.01, -0.0461, 0.1615, 0.01, -0.0483, 0.1615, 0.01, -0.0461, 0.1646, 0.0061, -0.0527, 0.1647, 0.0061, -0.0505, 0.1646, 0.0083, -0.0527, 0.1647, 0.0083, -0.0505, 0.1645, 0.0061, -0.048, 0.1634, 0.0061, -0.046, 0.1645, 0.0083, -0.048, 0.1634, 0.0083, -0.046, 0.1645, 0.01, -0.0529, 0.1647, 0.01, -0.0505, 0.1644, 0.01, -0.048, 0.1633, 0.01, -0.0459,
0.1596, 0.0061, -0.0431, 0.1585, 0.0061, -0.0428, 0.1596, 0.0083, -0.0431, 0.1585, 0.0083, -0.0428, 0.162, 0.0061, -0.0444, 0.162, 0.0083, -0.0444, 0.1595, 0.01, -0.043, 0.1584, 0.01, -0.0428, 0.1618, 0.01, -0.0442, 0.1626, 0.0061, -0.045, 0.1626, 0.0083, -0.045, 0.1626, 0.01, -0.045, 0.1412, 0.0061, -0.0219, 0.1412, 0.0061, -0.0201, 0.1412, 0.0083, -0.0219, 0.1412, 0.0083, -0.0201, 0.1441, 0.0061, -0.0219, 0.1441, 0.0083, -0.0219, 0.1412, 0.01, -0.0219, 0.1412, 0.01, -0.0201, 0.1441, 0.01, -0.0219, 0.1441, 0.01, -0.0201, 0.1463, 0.0061, -0.0219, 0.1463, 0.0083, -0.0219, 0.1484, 0.0061, -0.0219, 0.1484, 0.0083, -0.0219, 0.1463, 0.01, -0.0219, 0.1463, 0.01, -0.0201, 0.1484, 0.01, -0.0219, 0.1484, 0.01, -0.0201,
0.1506, 0.0061, -0.0219, 0.1506, 0.0083, -0.0219, 0.1528, 0.0061, -0.0219, 0.1528, 0.0083, -0.0219, 0.1506, 0.01, -0.0219, 0.1506, 0.01, -0.0201, 0.1528, 0.01, -0.0219, 0.1528, 0.01, -0.0201, 0.155, 0.0061, -0.0219, 0.155, 0.0083, -0.0219, 0.1571, 0.0061, -0.0219, 0.1571, 0.0083, -0.0219, 0.155, 0.01, -0.0219, 0.155, 0.01, -0.0201, 0.1571, 0.01, -0.0219, 0.1571, 0.01, -0.0201, 0.1412, 0.0061, -0.0179, 0.1412, 0.0061, -0.0157, 0.1412, 0.0083, -0.0179, 0.1412, 0.0083, -0.0157, 0.1412, 0.0061, -0.0136, 0.1412, 0.0061, -0.0114, 0.1412, 0.0083, -0.0136, 0.1412, 0.0083, -0.0114, 0.1412, 0.01, -0.0179, 0.1412, 0.01, -0.0157, 0.1441, 0.01, -0.0179, 0.1441, 0.01, -0.0157, 0.1412, 0.01, -0.0136, 0.1412, 0.01, -0.0114,
0.1441, 0.01, -0.0136, 0.1441, 0.01, -0.0114, 0.1463, 0.01, -0.0179, 0.1463, 0.01, -0.0157, 0.1484, 0.01, -0.0179, 0.1484, 0.01, -0.0157, 0.1463, 0.01, -0.0136, 0.1463, 0.01, -0.0114, 0.1484, 0.01, -0.0136, 0.1484, 0.01, -0.0114, 0.1412, 0.0061, -0.0092, 0.1412, 0.0061, -0.0071, 0.1412, 0.0083, -0.0092, 0.1412, 0.0083, -0.0071, 0.1412, 0.0061, -0.0049, 0.1412, 0.0061, -0.0027, 0.1412, 0.0083, -0.0049, 0.1412, 0.0083, -0.0027, 0.1412, 0.01, -0.0092, 0.1412, 0.01, -0.0071, 0.1441, 0.01, -0.0092, 0.1441, 0.01, -0.0071, 0.1412, 0.01, -0.0049, 0.1412, 0.01, -0.0027, 0.1441, 0.01, -0.0049, 0.1441, 0.01, -0.0027, 0.1463, 0.01, -0.0092, 0.1463, 0.01, -0.0071, 0.1484, 0.01, -0.0092, 0.1484, 0.01, -0.0071,
0.1463, 0.01, -0.0049, 0.1463, 0.01, -0.0027, 0.1484, 0.01, -0.0049, 0.1484, 0.01, -0.0027, 0.1506, 0.01, -0.0179, 0.1506, 0.01, -0.0157, 0.1528, 0.01, -0.0179, 0.1528, 0.01, -0.0157, 0.1506, 0.01, -0.0136, 0.1506, 0.01, -0.0114, 0.1528, 0.01, -0.0136, 0.1528, 0.01, -0.0114, 0.155, 0.01, -0.0179, 0.155, 0.01, -0.0157, 0.1571, 0.01, -0.0179, 0.1571, 0.01, -0.0157, 0.155, 0.01, -0.0136, 0.155, 0.01, -0.0114, 0.1571, 0.01, -0.0136, 0.1571, 0.01, -0.0114, 0.152, 0.0061, -0.0101, 0.152, 0.0061, -0.0071, 0.152, 0.0083, -0.0101, 0.152, 0.0083, -0.0071, 0.152, 0.0061, -0.0049, 0.152, 0.0061, -0.0027, 0.152, 0.0083, -0.0049, 0.152, 0.0083, -0.0027, 0.1506, 0.01, -0.0092, 0.1506, 0.01, -0.0071,
0.152, 0.01, -0.0101, 0.152, 0.01, -0.0071, 0.1506, 0.01, -0.0049, 0.1506, 0.01, -0.0027, 0.152, 0.01, -0.0049, 0.152, 0.01, -0.0027, 0.155, 0.0061, -0.0101, 0.155, 0.0083, -0.0101, 0.1571, 0.0061, -0.0101, 0.1571, 0.0083, -0.0101, 0.155, 0.01, -0.0101, 0.1571, 0.01, -0.0101, 0.1593, 0.0061, -0.0219, 0.1593, 0.0083, -0.0219, 0.1615, 0.0061, -0.0219, 0.1615, 0.0083, -0.0219, 0.1593, 0.01, -0.0219, 0.1593, 0.01, -0.0201, 0.1615, 0.01, -0.0219, 0.1615, 0.01, -0.0201, 0.1636, 0.0061, -0.0219, 0.1636, 0.0083, -0.0219, 0.1658, 0.0061, -0.0219, 0.1658, 0.0083, -0.0219, 0.1636, 0.01, -0.0219, 0.1636, 0.01, -0.0201, 0.1658, 0.01, -0.0219, 0.1658, 0.01, -0.0201, 0.168, 0.0061, -0.0219, 0.168, 0.0083, -0.0219,
0.1701, 0.0061, -0.0219, 0.1701, 0.0083, -0.0219, 0.168, 0.01, -0.0219, 0.168, 0.01, -0.0201, 0.1701, 0.01, -0.0219, 0.1701, 0.01, -0.0201, 0.1723, 0.0061, -0.0219, 0.1723, 0.0083, -0.0219, 0.1745, 0.0061, -0.0219, 0.1745, 0.0083, -0.0219, 0.1723, 0.01, -0.0219, 0.1723, 0.01, -0.0201, 0.1745, 0.01, -0.0219, 0.1745, 0.01, -0.0201, 0.1593, 0.01, -0.0179, 0.1593, 0.01, -0.0157, 0.1615, 0.01, -0.0179, 0.1615, 0.01, -0.0157, 0.1593, 0.01, -0.0136, 0.1593, 0.01, -0.0114, 0.1615, 0.01, -0.0136, 0.1615, 0.01, -0.0114, 0.1636, 0.01, -0.0179, 0.1636, 0.01, -0.0157, 0.1658, 0.01, -0.0179, 0.1658, 0.01, -0.0157, 0.1636, 0.01, -0.0136, 0.1636, 0.01, -0.0114, 0.1658, 0.01, -0.0136, 0.1658, 0.01, -0.0114,
0.1593, 0.0061, -0.0101, 0.1593, 0.0083, -0.0101, 0.1615, 0.0061, -0.0101, 0.1615, 0.0083, -0.0101, 0.1593, 0.01, -0.0101, 0.1615, 0.01, -0.0101, 0.1636, 0.0061, -0.0101, 0.1636, 0.0083, -0.0101, 0.1658, 0.0061, -0.0101, 0.1658, 0.0083, -0.0101, 0.1636, 0.01, -0.0101, 0.1658, 0.01, -0.0101, 0.168, 0.01, -0.0179, 0.168, 0.01, -0.0157, 0.1701, 0.01, -0.0179, 0.1701, 0.01, -0.0157, 0.168, 0.01, -0.0136, 0.168, 0.01, -0.0114, 0.1701, 0.01, -0.0136, 0.1701, 0.01, -0.0114, 0.1723, 0.01, -0.0179, 0.1723, 0.01, -0.0157, 0.1745, 0.01, -0.0179, 0.1745, 0.01, -0.0157, 0.1723, 0.01, -0.0136, 0.1723, 0.01, -0.0114, 0.1745, 0.01, -0.0136, 0.1745, 0.01, -0.0114, 0.168, 0.0061, -0.0101, 0.168, 0.0083, -0.0101,
0.1701, 0.0061, -0.0101, 0.1701, 0.0083, -0.0101, 0.168, 0.01, -0.0101, 0.1701, 0.01, -0.0101, 0.1723, 0.0061, -0.0101, 0.1723, 0.0083, -0.0101, 0.1745, 0.0061, -0.0101, 0.1745, 0.0083, -0.0101, 0.1723, 0.01, -0.0101, 0.1745, 0.01, -0.0101, 0.1767, 0.0061, -0.0219, 0.1767, 0.0083, -0.0219, 0.1788, 0.0061, -0.0219, 0.1788, 0.0083, -0.0219, 0.1767, 0.01, -0.0219, 0.1767, 0.01, -0.0201, 0.1788, 0.01, -0.0219, 0.1788, 0.01, -0.0201, 0.181, 0.0061, -0.0219, 0.181, 0.0083, -0.0219, 0.1832, 0.0061, -0.0219, 0.1832, 0.0083, -0.0219, 0.181, 0.01, -0.0219, 0.181, 0.01, -0.0201, 0.1832, 0.01, -0.0219, 0.1832, 0.01, -0.0201, 0.1853, 0.0061, -0.0219, 0.1853, 0.0083, -0.0219, 0.1875, 0.0061, -0.0219, 0.1875, 0.0083, -0.0219,
0.1853, 0.01, -0.0219, 0.1853, 0.01, -0.0201, 0.1875, 0.01, -0.0219, 0.1875, 0.01, -0.0201, 0.1897, 0.0061, -0.0219, 0.1897, 0.0083, -0.0219, 0.1918, 0.0061, -0.0219, 0.1918, 0.0083, -0.0219, 0.1897, 0.01, -0.0219, 0.1897, 0.01, -0.0201, 0.1918, 0.01, -0.0219, 0.1918, 0.01, -0.0201, 0.1767, 0.01, -0.0179, 0.1767, 0.01, -0.0157, 0.1788, 0.01, -0.0179, 0.1788, 0.01, -0.0157, 0.1767, 0.01, -0.0136, 0.1767, 0.01, -0.0114, 0.1788, 0.01, -0.0136, 0.1788, 0.01, -0.0114, 0.181, 0.01, -0.0179, 0.181, 0.01, -0.0157, 0.1832, 0.01, -0.0179, 0.1832, 0.01, -0.0157, 0.181, 0.01, -0.0136, 0.181, 0.01, -0.0114, 0.1832, 0.01, -0.0136, 0.1832, 0.01, -0.0114, 0.1767, 0.0061, -0.0101, 0.1767, 0.0083, -0.0101,
0.1788, 0.0061, -0.0101, 0.1788, 0.0083, -0.0101, 0.1767, 0.01, -0.0101, 0.1788, 0.01, -0.0101, 0.181, 0.0061, -0.0101, 0.181, 0.0083, -0.0101, 0.1832, 0.0061, -0.0101, 0.1832, 0.0083, -0.0101, 0.181, 0.01, -0.0101, 0.1832, 0.01, -0.0101, 0.1853, 0.01, -0.0179, 0.1853, 0.01, -0.0157, 0.1875, 0.01, -0.0179, 0.1875, 0.01, -0.0157, 0.1853, 0.01, -0.0136, 0.1853, 0.01, -0.0114, 0.1875, 0.01, -0.0136, 0.1875, 0.01, -0.0114, 0.1897, 0.01, -0.0179, 0.1897, 0.01, -0.0157, 0.1918, 0.01, -0.0179, 0.1918, 0.01, -0.0157, 0.1897, 0.01, -0.0136, 0.1897, 0.01, -0.0114, 0.1918, 0.01, -0.0136, 0.1918, 0.01, -0.0114, 0.1853, 0.0061, -0.0101, 0.1853, 0.0083, -0.0101, 0.1875, 0.0061, -0.0101, 0.1875, 0.0083, -0.0101,
0.1853, 0.01, -0.0101, 0.1875, 0.01, -0.0101, 0.1897, 0.0061, -0.0101, 0.1897, 0.0083, -0.0101, 0.1918, 0.0061, -0.0101, 0.1918, 0.0083, -0.0101, 0.1897, 0.01, -0.0101, 0.1918, 0.01, -0.0101, 0.194, 0.0061, -0.0219, 0.194, 0.0083, -0.0219, 0.1962, 0.0061, -0.0219, 0.1962, 0.0083, -0.0219, 0.194, 0.01, -0.0219, 0.194, 0.01, -0.0201, 0.1962, 0.01, -0.0219, 0.1962, 0.01, -0.0201, 0.1984, 0.0061, -0.0219, 0.1984, 0.0083, -0.0219, 0.2005, 0.0061, -0.0219, 0.2005, 0.0083, -0.0219, 0.1984, 0.01, -0.0219, 0.1984, 0.01, -0.0201, 0.2005, 0.01, -0.0219, 0.2005, 0.01, -0.0201, 0.202, 0.0061, -0.0219, 0.202, 0.0061, -0.0201, 0.202, 0.0083, -0.0219, 0.202, 0.0083, -0.0201, 0.202, 0.01, -0.0219, 0.202, 0.01, -0.0201,
0.194, 0.01, -0.0179, 0.194, 0.01, -0.0157, 0.1962, 0.01, -0.0179, 0.1962, 0.01, -0.0157, 0.194, 0.01, -0.0136, 0.194, 0.01, -0.0114, 0.1962, 0.01, -0.0136, 0.1962, 0.01, -0.0114, 0.1984, 0.01, -0.0179, 0.1984, 0.01, -0.0157, 0.2005, 0.01, -0.0179, 0.2005, 0.01, -0.0157, 0.1984, 0.01, -0.0136, 0.1984, 0.01, -0.0114, 0.2005, 0.01, -0.0136, 0.2005, 0.01, -0.0114, 0.194, 0.0061, -0.0101, 0.194, 0.0083, -0.0101, 0.1962, 0.0061, -0.0101, 0.1962, 0.0083, -0.0101, 0.194, 0.01, -0.0101, 0.1962, 0.01, -0.0101, 0.1984, 0.0061, -0.0101, 0.1984, 0.0083, -0.0101, 0.2005, 0.0061, -0.0101, 0.2005, 0.0083, -0.0101, 0.1984, 0.01, -0.0101, 0.2005, 0.01, -0.0101, 0.202, 0.0061, -0.0179, 0.202, 0.0061, -0.0157,
0.202, 0.0083, -0.0179, 0.202, 0.0083, -0.0157, 0.202, 0.0061, -0.0136, 0.202, 0.0061, -0.0114, 0.202, 0.0083, -0.0136, 0.202, 0.0083, -0.0114, 0.202, 0.01, -0.0179, 0.202, 0.01, -0.0157, 0.202, 0.01, -0.0136, 0.202, 0.01, -0.0114, 0.202, 0.0061, -0.0101, 0.202, 0.0083, -0.0101, 0.202, 0.01, -0.0101, 0.2118, 0.0061, -0.067, 0.2118, 0.0083, -0.067, 0.2135, 0.0061, -0.067, 0.2135, 0.0083, -0.067, 0.2118, 0.0061, -0.0656, 0.2118, 0.0061, -0.0635, 0.2118, 0.0083, -0.0656, 0.2118, 0.0083, -0.0635, 0.2118, 0.01, -0.067, 0.2135, 0.01, -0.067, 0.2118, 0.01, -0.0656, 0.2118, 0.01, -0.0635, 0.2135, 0.01, -0.0656, 0.2135, 0.01, -0.0635, 0.2157, 0.0061, -0.067, 0.2157, 0.0083, -0.067, 0.2179, 0.0061, -0.067,
0.2179, 0.0083, -0.067, 0.2157, 0.0061, -0.0631, 0.2157, 0.0083, -0.0631, 0.2179, 0.0061, -0.0631, 0.2179, 0.0083, -0.0631, 0.2157, 0.01, -0.067, 0.2179, 0.01, -0.067, 0.2157, 0.01, -0.0656, 0.2157, 0.01, -0.0631, 0.2179, 0.01, -0.0656, 0.2179, 0.01, -0.0631, 0.2118, 0.0061, -0.0613, 0.2118, 0.0061, -0.0591, 0.2118, 0.0083, -0.0613, 0.2118, 0.0083, -0.0591, 0.2118, 0.0061, -0.057, 0.2118, 0.0061, -0.0548, 0.2118, 0.0083, -0.057, 0.2118, 0.0083, -0.0548, 0.2118, 0.01, -0.0613, 0.2118, 0.01, -0.0591, 0.2135, 0.01, -0.0613, 0.2135, 0.01, -0.0591, 0.2118, 0.01, -0.057, 0.2118, 0.01, -0.0548, 0.2135, 0.01, -0.057, 0.2135, 0.01, -0.0548, 0.2157, 0.0061, -0.0613, 0.2157, 0.0061, -0.0591, 0.2157, 0.0083, -0.0613,
0.2157, 0.0083, -0.0591, 0.2157, 0.0061, -0.057, 0.2157, 0.0061, -0.0548, 0.2157, 0.0083, -0.057, 0.2157, 0.0083, -0.0548, 0.2157, 0.01, -0.0613, 0.2157, 0.01, -0.0591, 0.2157, 0.01, -0.057, 0.2157, 0.01, -0.0548, 0.2201, 0.0061, -0.067, 0.2201, 0.0083, -0.067, 0.2222, 0.0061, -0.067, 0.2222, 0.0083, -0.067, 0.2201, 0.0061, -0.0631, 0.2201, 0.0083, -0.0631, 0.2222, 0.0061, -0.0631, 0.2222, 0.0083, -0.0631, 0.2201, 0.01, -0.067, 0.2222, 0.01, -0.067, 0.2201, 0.01, -0.0656, 0.2201, 0.01, -0.0631, 0.2222, 0.01, -0.0656, 0.2222, 0.01, -0.0631, 0.2244, 0.0061, -0.067, 0.2244, 0.0083, -0.067, 0.2266, 0.0061, -0.067, 0.2266, 0.0083, -0.067, 0.2244, 0.0061, -0.0631, 0.2244, 0.0083, -0.0631, 0.2266, 0.0061, -0.0631,
0.2266, 0.0083, -0.0631, 0.2244, 0.01, -0.067, 0.2266, 0.01, -0.067, 0.2244, 0.01, -0.0656, 0.2244, 0.01, -0.0631, 0.2266, 0.01, -0.0656, 0.2266, 0.01, -0.0631, 0.2196, 0.0061, -0.0562, 0.2196, 0.0061, -0.0548, 0.2196, 0.0083, -0.0562, 0.2196, 0.0083, -0.0548, 0.2222, 0.0061, -0.0562, 0.2222, 0.0083, -0.0562, 0.2196, 0.01, -0.0562, 0.2196, 0.01, -0.0548, 0.2222, 0.01, -0.0562, 0.2222, 0.01, -0.0548, 0.2244, 0.0061, -0.0562, 0.2244, 0.0083, -0.0562, 0.2266, 0.0061, -0.0562, 0.2266, 0.0083, -0.0562, 0.2244, 0.01, -0.0562, 0.2244, 0.01, -0.0548, 0.2266, 0.01, -0.0562, 0.2266, 0.01, -0.0548, 0.2118, 0.0061, -0.0526, 0.2118, 0.0061, -0.0505, 0.2118, 0.0083, -0.0526, 0.2118, 0.0083, -0.0505, 0.2118, 0.0061, -0.0483,
0.2118, 0.0061, -0.0461, 0.2118, 0.0083, -0.0483, 0.2118, 0.0083, -0.0461, 0.2118, 0.01, -0.0526, 0.2118, 0.01, -0.0505, 0.2135, 0.01, -0.0526, 0.2135, 0.01, -0.0505, 0.2118, 0.01, -0.0483, 0.2118, 0.01, -0.0461, 0.2135, 0.01, -0.0483, 0.2135, 0.01, -0.0461, 0.2157, 0.0061, -0.0526, 0.2157, 0.0061, -0.0505, 0.2157, 0.0083, -0.0526, 0.2157, 0.0083, -0.0505, 0.2157, 0.0061, -0.0483, 0.2157, 0.0061, -0.0454, 0.2157, 0.0083, -0.0483, 0.2157, 0.0083, -0.0454, 0.2179, 0.0061, -0.0454, 0.2179, 0.0083, -0.0454, 0.2157, 0.01, -0.0526, 0.2157, 0.01, -0.0505, 0.2157, 0.01, -0.0483, 0.2157, 0.01, -0.0454, 0.2179, 0.01, -0.0454, 0.2118, 0.0061, -0.0439, 0.2118, 0.0061, -0.0425, 0.2118, 0.0083, -0.0439, 0.2118, 0.0083, -0.0425,
0.2135, 0.0061, -0.0425, 0.2135, 0.0083, -0.0425, 0.2118, 0.01, -0.0439, 0.2118, 0.01, -0.0425, 0.2135, 0.01, -0.0439, 0.2135, 0.01, -0.0425, 0.2157, 0.0061, -0.0425, 0.2157, 0.0083, -0.0425, 0.2179, 0.0061, -0.0425, 0.2179, 0.0083, -0.0425, 0.2157, 0.01, -0.0439, 0.2157, 0.01, -0.0425, 0.2179, 0.01, -0.0439, 0.2179, 0.01, -0.0425, 0.2196, 0.0061, -0.0523, 0.2196, 0.0083, -0.0523, 0.2222, 0.0061, -0.0523, 0.2222, 0.0083, -0.0523, 0.2201, 0.0061, -0.0454, 0.2201, 0.0083, -0.0454, 0.2222, 0.0061, -0.0454, 0.2222, 0.0083, -0.0454, 0.2196, 0.01, -0.0523, 0.2222, 0.01, -0.0523, 0.2201, 0.01, -0.0454, 0.2222, 0.01, -0.0454, 0.2244, 0.0061, -0.0523, 0.2244, 0.0083, -0.0523, 0.2266, 0.0061, -0.0523, 0.2266, 0.0083, -0.0523,
0.2244, 0.0061, -0.0454, 0.2244, 0.0083, -0.0454, 0.2266, 0.0061, -0.0454, 0.2266, 0.0083, -0.0454, 0.2244, 0.01, -0.0523, 0.2266, 0.01, -0.0523, 0.2244, 0.01, -0.0454, 0.2266, 0.01, -0.0454, 0.2201, 0.0061, -0.0425, 0.2201, 0.0083, -0.0425, 0.2222, 0.0061, -0.0425, 0.2222, 0.0083, -0.0425, 0.2201, 0.01, -0.0439, 0.2201, 0.01, -0.0425, 0.2222, 0.01, -0.0439, 0.2222, 0.01, -0.0425, 0.2244, 0.0061, -0.0425, 0.2244, 0.0083, -0.0425, 0.2266, 0.0061, -0.0425, 0.2266, 0.0083, -0.0425, 0.2244, 0.01, -0.0439, 0.2244, 0.01, -0.0425, 0.2266, 0.01, -0.0439, 0.2266, 0.01, -0.0425, 0.2287, 0.0061, -0.067, 0.2287, 0.0083, -0.067, 0.2307, 0.0061, -0.0669, 0.2307, 0.0083, -0.0669, 0.2287, 0.0061, -0.063, 0.2287, 0.0083, -0.063,
0.2318, 0.0061, -0.0667, 0.2309, 0.0061, -0.0629, 0.2318, 0.0083, -0.0667, 0.2309, 0.0083, -0.0629, 0.2287, 0.01, -0.067, 0.2304, 0.01, -0.0669, 0.2287, 0.01, -0.0656, 0.2287, 0.01, -0.0631, 0.2312, 0.01, -0.0668, 0.2308, 0.01, -0.0629, 0.2331, 0.0061, -0.0665, 0.2325, 0.0061, -0.0625, 0.2331, 0.0083, -0.0665, 0.2325, 0.0083, -0.0625, 0.2355, 0.0061, -0.0657, 0.2355, 0.0083, -0.0657, 0.2333, 0.01, -0.0665, 0.2328, 0.01, -0.0624, 0.2356, 0.01, -0.0657, 0.2353, 0.01, -0.0635, 0.2287, 0.0061, -0.0563, 0.2287, 0.0083, -0.0563, 0.2309, 0.0061, -0.0565, 0.2309, 0.0083, -0.0565, 0.2287, 0.01, -0.0563, 0.2287, 0.01, -0.0548, 0.2309, 0.01, -0.0565, 0.2309, 0.01, -0.0548, 0.2336, 0.0061, -0.0621, 0.2336, 0.0083, -0.0621,
0.2349, 0.0061, -0.0613, 0.2363, 0.0061, -0.0591, 0.2349, 0.0083, -0.0613, 0.2363, 0.0083, -0.0591, 0.2331, 0.0061, -0.0571, 0.2331, 0.0083, -0.0571, 0.2345, 0.0061, -0.0579, 0.2345, 0.0083, -0.0579, 0.2337, 0.01, -0.0621, 0.2349, 0.01, -0.0613, 0.2363, 0.01, -0.0591, 0.2329, 0.01, -0.0571, 0.2331, 0.01, -0.0548, 0.2346, 0.01, -0.058, 0.2353, 0.01, -0.0548, 0.2366, 0.0061, -0.0648, 0.238, 0.0061, -0.0634, 0.2366, 0.0083, -0.0648, 0.238, 0.0083, -0.0634, 0.2365, 0.01, -0.0648, 0.238, 0.01, -0.0634, 0.2384, 0.0061, -0.0622, 0.2384, 0.0083, -0.0622, 0.2388, 0.0061, -0.0612, 0.239, 0.0061, -0.0591, 0.2388, 0.0083, -0.0612, 0.239, 0.0083, -0.0591, 0.2387, 0.0061, -0.057, 0.2384, 0.0061, -0.0544, 0.2387, 0.0083, -0.057,
0.2384, 0.0083, -0.0544, 0.2386, 0.01, -0.0618, 0.2374, 0.01, -0.0591, 0.2389, 0.01, -0.0609, 0.239, 0.01, -0.0591, 0.2374, 0.01, -0.057, 0.2384, 0.01, -0.0547, 0.2387, 0.01, -0.057, 0.2384, 0.01, -0.0546, 0.2287, 0.0061, -0.0522, 0.2287, 0.0083, -0.0522, 0.2309, 0.0061, -0.0522, 0.2309, 0.0083, -0.0522, 0.2287, 0.0061, -0.0455, 0.2287, 0.0083, -0.0455, 0.2309, 0.0061, -0.0459, 0.2309, 0.0083, -0.0459, 0.2287, 0.01, -0.0522, 0.2309, 0.01, -0.0522, 0.2287, 0.01, -0.0455, 0.2308, 0.01, -0.0458, 0.2326, 0.0061, -0.0519, 0.2337, 0.0061, -0.0513, 0.2326, 0.0083, -0.0519, 0.2337, 0.0083, -0.0513, 0.2345, 0.0061, -0.0504, 0.2345, 0.0083, -0.0504, 0.2339, 0.0061, -0.0475, 0.2328, 0.0061, -0.0465, 0.2339, 0.0083, -0.0475,
0.2328, 0.0083, -0.0465, 0.2348, 0.0061, -0.0484, 0.2348, 0.0083, -0.0484, 0.2329, 0.01, -0.0518, 0.2338, 0.01, -0.0512, 0.2353, 0.01, -0.0526, 0.2345, 0.01, -0.0504, 0.2339, 0.01, -0.0475, 0.2329, 0.01, -0.0466, 0.2348, 0.01, -0.0484, 0.2353, 0.01, -0.0461, 0.2287, 0.0061, -0.0425, 0.2287, 0.0083, -0.0425, 0.2309, 0.0061, -0.0425, 0.2309, 0.0083, -0.0425, 0.2287, 0.01, -0.0439, 0.2287, 0.01, -0.0425, 0.2309, 0.01, -0.0439, 0.2309, 0.01, -0.0425, 0.2337, 0.0061, -0.043, 0.2326, 0.0061, -0.0427, 0.2337, 0.0083, -0.043, 0.2326, 0.0083, -0.0427, 0.2353, 0.0061, -0.0437, 0.2353, 0.0083, -0.0437, 0.2334, 0.01, -0.0429, 0.2324, 0.01, -0.0427, 0.2356, 0.01, -0.0438, 0.2386, 0.0061, -0.0526, 0.239, 0.0061, -0.0505,
0.2386, 0.0083, -0.0526, 0.239, 0.0083, -0.0505, 0.2385, 0.0061, -0.0472, 0.2379, 0.0061, -0.046, 0.2385, 0.0083, -0.0472, 0.2379, 0.0083, -0.046, 0.2388, 0.0061, -0.0478, 0.2388, 0.0083, -0.0478, 0.2374, 0.01, -0.0526, 0.2374, 0.01, -0.0505, 0.2386, 0.01, -0.0526, 0.239, 0.01, -0.0505, 0.2387, 0.01, -0.0475, 0.2377, 0.01, -0.0457, 0.239, 0.01, -0.0485, 0.2368, 0.0061, -0.0447, 0.2368, 0.0083, -0.0447, 0.2367, 0.01, -0.0447, 0.2187, 0.0061, -0.0107, 0.2187, 0.0083, -0.0107, 0.2187, 0.01, -0.0106, 0.2144, 0.0061, -0.0024, 0.2144, 0.0083, -0.0024, 0.2143, 0.01, -0.0022, 0.2163, 0.0061, -0.0069, 0.2163, 0.0083, -0.0069, 0.2176, 0.0061, -0.0093, 0.2169, 0.0061, -0.008, 0.2176, 0.0083, -0.0093, 0.2169, 0.0083, -0.008,
0.2153, 0.0061, -0.0049, 0.2147, 0.0061, -0.0035, 0.2153, 0.0083, -0.0049, 0.2147, 0.0083, -0.0035, 0.2162, 0.01, -0.0067, 0.2178, 0.01, -0.0096, 0.2168, 0.01, -0.0077, 0.2154, 0.01, -0.0052, 0.2146, 0.01, -0.0032, 0.2179, 0.01, -0.0049, 0.2179, 0.01, -0.0027, 0.223, 0.0061, -0.0149, 0.223, 0.0083, -0.0149, 0.2207, 0.0061, -0.0129, 0.2196, 0.0061, -0.0118, 0.2207, 0.0083, -0.0129, 0.2196, 0.0083, -0.0118, 0.2219, 0.0061, -0.014, 0.2219, 0.0083, -0.014, 0.2231, 0.01, -0.0149, 0.2207, 0.01, -0.0129, 0.2196, 0.01, -0.0118, 0.2219, 0.01, -0.014, 0.2222, 0.01, -0.0114, 0.2244, 0.0061, -0.0158, 0.2244, 0.0083, -0.0158, 0.2269, 0.0061, -0.0172, 0.2258, 0.0061, -0.0167, 0.2269, 0.0083, -0.0172, 0.2258, 0.0083, -0.0167,
0.2241, 0.01, -0.0156, 0.227, 0.01, -0.0173, 0.226, 0.01, -0.0168, 0.2244, 0.01, -0.0136, 0.2244, 0.01, -0.0114, 0.2266, 0.01, -0.0136, 0.2266, 0.01, -0.0114, 0.2201, 0.01, -0.0092, 0.2201, 0.01, -0.0071, 0.2222, 0.01, -0.0092, 0.2222, 0.01, -0.0071, 0.2201, 0.01, -0.0049, 0.2201, 0.01, -0.0027, 0.2222, 0.01, -0.0049, 0.2222, 0.01, -0.0027, 0.2244, 0.01, -0.0092, 0.2244, 0.01, -0.0071, 0.2266, 0.01, -0.0092, 0.2266, 0.01, -0.0071, 0.2244, 0.01, -0.0049, 0.2244, 0.01, -0.0027, 0.2266, 0.01, -0.0049, 0.2266, 0.01, -0.0027, 0.2318, 0.0061, -0.0191, 0.2318, 0.0083, -0.0191, 0.2318, 0.01, -0.0191, 0.2331, 0.0061, -0.0194, 0.2331, 0.0083, -0.0194, 0.2352, 0.0061, -0.02, 0.2352, 0.0083, -0.02,
0.233, 0.01, -0.0194, 0.2351, 0.01, -0.02, 0.2374, 0.0061, -0.0205, 0.2374, 0.0083, -0.0205, 0.2396, 0.0061, -0.0208, 0.2396, 0.0083, -0.0208, 0.2373, 0.01, -0.0205, 0.2395, 0.01, -0.0208, 0.2418, 0.0061, -0.021, 0.2418, 0.0083, -0.021, 0.244, 0.0061, -0.0213, 0.2429, 0.0061, -0.0212, 0.244, 0.0083, -0.0213, 0.2429, 0.0083, -0.0212, 0.2416, 0.01, -0.021, 0.2443, 0.01, -0.0214, 0.2436, 0.01, -0.0213, 0.2287, 0.0061, -0.018, 0.2287, 0.0083, -0.018, 0.2307, 0.0061, -0.0187, 0.2307, 0.0083, -0.0187, 0.2285, 0.01, -0.0179, 0.2287, 0.01, -0.0157, 0.2308, 0.01, -0.0187, 0.2309, 0.01, -0.0157, 0.2287, 0.01, -0.0136, 0.2287, 0.01, -0.0114, 0.2309, 0.01, -0.0136, 0.2309, 0.01, -0.0114, 0.2331, 0.01, -0.0179,
0.2331, 0.01, -0.0157, 0.2353, 0.01, -0.0179, 0.2353, 0.01, -0.0157, 0.2331, 0.01, -0.0136, 0.2331, 0.01, -0.0114, 0.2353, 0.01, -0.0136, 0.2353, 0.01, -0.0114, 0.2289, 0.0061, -0.0026, 0.2289, 0.0083, -0.0026, 0.231, 0.0061, -0.0047, 0.23, 0.0061, -0.0036, 0.231, 0.0083, -0.0047, 0.23, 0.0083, -0.0036, 0.2287, 0.01, -0.0092, 0.2287, 0.01, -0.0071, 0.2309, 0.01, -0.0092, 0.2309, 0.01, -0.0071, 0.2287, 0.01, -0.0049, 0.2289, 0.01, -0.0026, 0.231, 0.01, -0.0047, 0.23, 0.01, -0.0036, 0.2335, 0.0061, -0.0064, 0.2335, 0.0083, -0.0064, 0.2352, 0.0061, -0.0074, 0.2352, 0.0083, -0.0074, 0.2324, 0.0061, -0.0057, 0.2324, 0.0083, -0.0057, 0.2331, 0.01, -0.0092, 0.2334, 0.01, -0.0063, 0.2353, 0.01, -0.0092,
0.2354, 0.01, -0.0075, 0.2323, 0.01, -0.0057, 0.2374, 0.01, -0.0179, 0.2374, 0.01, -0.0157, 0.2396, 0.01, -0.0179, 0.2396, 0.01, -0.0157, 0.2374, 0.01, -0.0136, 0.2374, 0.01, -0.0114, 0.2396, 0.01, -0.0136, 0.2396, 0.01, -0.0114, 0.2418, 0.01, -0.0179, 0.2418, 0.01, -0.0157, 0.2439, 0.01, -0.0179, 0.2439, 0.01, -0.0157, 0.2418, 0.01, -0.0136, 0.2418, 0.01, -0.0114, 0.2439, 0.01, -0.0136, 0.2439, 0.01, -0.0114, 0.2381, 0.0061, -0.0082, 0.2371, 0.0061, -0.0079, 0.2381, 0.0083, -0.0082, 0.2371, 0.0083, -0.0079, 0.2396, 0.0061, -0.0086, 0.2396, 0.0083, -0.0086, 0.2377, 0.01, -0.0081, 0.2369, 0.01, -0.0079, 0.2398, 0.01, -0.0086, 0.2418, 0.0061, -0.009, 0.2418, 0.0083, -0.009, 0.2439, 0.0061, -0.0093,
0.2439, 0.0083, -0.0093, 0.2419, 0.01, -0.009, 0.244, 0.01, -0.0093, 0.2588, 0.0061, -0.067, 0.2588, 0.0083, -0.067, 0.2613, 0.0061, -0.067, 0.2613, 0.0083, -0.067, 0.2588, 0.0061, -0.0656, 0.2588, 0.0061, -0.0635, 0.2588, 0.0083, -0.0656, 0.2588, 0.0083, -0.0635, 0.2618, 0.0061, -0.0631, 0.2618, 0.0083, -0.0631, 0.2588, 0.01, -0.067, 0.2613, 0.01, -0.067, 0.2588, 0.01, -0.0656, 0.2588, 0.01, -0.0635, 0.2613, 0.01, -0.0656, 0.2618, 0.01, -0.0631, 0.2588, 0.0061, -0.0613, 0.2588, 0.0061, -0.0591, 0.2588, 0.0083, -0.0613, 0.2588, 0.0083, -0.0591, 0.2618, 0.0061, -0.0613, 0.2618, 0.0061, -0.0591, 0.2618, 0.0083, -0.0613, 0.2618, 0.0083, -0.0591, 0.2588, 0.0061, -0.057, 0.2588, 0.0061, -0.0548, 0.2588, 0.0083, -0.057,
0.2588, 0.0083, -0.0548, 0.2618, 0.0061, -0.057, 0.2618, 0.0061, -0.0548, 0.2618, 0.0083, -0.057, 0.2618, 0.0083, -0.0548, 0.2588, 0.01, -0.0613, 0.2588, 0.01, -0.0591, 0.2618, 0.01, -0.0613, 0.2618, 0.01, -0.0591, 0.2588, 0.01, -0.057, 0.2588, 0.01, -0.0548, 0.2618, 0.01, -0.057, 0.2618, 0.01, -0.0548, 0.2588, 0.0061, -0.0526, 0.2588, 0.0061, -0.0505, 0.2588, 0.0083, -0.0526, 0.2588, 0.0083, -0.0505, 0.2618, 0.0061, -0.0526, 0.2618, 0.0061, -0.0505, 0.2618, 0.0083, -0.0526, 0.2618, 0.0083, -0.0505, 0.2588, 0.0061, -0.0483, 0.2588, 0.0061, -0.0461, 0.2588, 0.0083, -0.0483, 0.2588, 0.0083, -0.0461, 0.2618, 0.0061, -0.0483, 0.2618, 0.0061, -0.0461, 0.2618, 0.0083, -0.0483, 0.2618, 0.0083, -0.0461, 0.2588, 0.01, -0.0526,
0.2588, 0.01, -0.0505, 0.2618, 0.01, -0.0526, 0.2618, 0.01, -0.0505, 0.2588, 0.01, -0.0483, 0.2588, 0.01, -0.0461, 0.2618, 0.01, -0.0483, 0.2618, 0.01, -0.0461, 0.2588, 0.0061, -0.0439, 0.2588, 0.0061, -0.0425, 0.2588, 0.0083, -0.0439, 0.2588, 0.0083, -0.0425, 0.2618, 0.0061, -0.0439, 0.2618, 0.0061, -0.0425, 0.2618, 0.0083, -0.0439, 0.2618, 0.0083, -0.0425, 0.2588, 0.01, -0.0439, 0.2588, 0.01, -0.0425, 0.2618, 0.01, -0.0439, 0.2618, 0.01, -0.0425, 0.2635, 0.0061, -0.067, 0.2635, 0.0083, -0.067, 0.2656, 0.0061, -0.067, 0.2656, 0.0083, -0.067, 0.2635, 0.0061, -0.0631, 0.2635, 0.0083, -0.0631, 0.2656, 0.0061, -0.0631, 0.2656, 0.0083, -0.0631, 0.2635, 0.01, -0.067, 0.2656, 0.01, -0.067, 0.2635, 0.01, -0.0656,
0.2635, 0.01, -0.0631, 0.2656, 0.01, -0.0656, 0.2656, 0.01, -0.0631, 0.2678, 0.0061, -0.067, 0.2678, 0.0083, -0.067, 0.27, 0.0061, -0.067, 0.27, 0.0083, -0.067, 0.2678, 0.0061, -0.0631, 0.2678, 0.0083, -0.0631, 0.27, 0.0061, -0.0631, 0.27, 0.0083, -0.0631, 0.2678, 0.01, -0.067, 0.27, 0.01, -0.067, 0.2678, 0.01, -0.0656, 0.2678, 0.01, -0.0631, 0.27, 0.01, -0.0656, 0.27, 0.01, -0.0631, 0.2721, 0.0061, -0.067, 0.2721, 0.0083, -0.067, 0.2743, 0.0061, -0.067, 0.2743, 0.0083, -0.067, 0.2721, 0.0061, -0.0631, 0.2721, 0.0083, -0.0631, 0.2743, 0.0061, -0.0631, 0.2743, 0.0083, -0.0631, 0.2721, 0.01, -0.067, 0.2743, 0.01, -0.067, 0.2721, 0.01, -0.0656, 0.2721, 0.01, -0.0631, 0.2743, 0.01, -0.0656,
0.2743, 0.01, -0.0631, 0.2765, 0.0061, -0.067, 0.2765, 0.0083, -0.067, 0.2794, 0.0061, -0.067, 0.2794, 0.0083, -0.067, 0.2765, 0.0061, -0.0631, 0.2765, 0.0083, -0.0631, 0.2794, 0.0061, -0.0656, 0.2794, 0.0061, -0.0631, 0.2794, 0.0083, -0.0656, 0.2794, 0.0083, -0.0631, 0.2765, 0.01, -0.067, 0.2794, 0.01, -0.067, 0.2765, 0.01, -0.0656, 0.2765, 0.01, -0.0631, 0.2794, 0.01, -0.0656, 0.2794, 0.01, -0.0631, 0.2461, 0.0061, -0.0216, 0.2461, 0.0083, -0.0216, 0.2483, 0.0061, -0.0218, 0.2483, 0.0083, -0.0218, 0.2461, 0.01, -0.0216, 0.2461, 0.01, -0.0201, 0.2482, 0.01, -0.0218, 0.2483, 0.01, -0.0201, 0.2504, 0.0061, -0.0219, 0.2504, 0.0083, -0.0219, 0.2526, 0.0061, -0.0219, 0.2526, 0.0083, -0.0219, 0.2504, 0.01, -0.0219,
0.2504, 0.01, -0.0201, 0.2526, 0.01, -0.0219, 0.2526, 0.01, -0.0201, 0.2548, 0.0061, -0.0219, 0.2548, 0.0083, -0.0219, 0.257, 0.0061, -0.0217, 0.257, 0.0083, -0.0217, 0.2548, 0.01, -0.0219, 0.2548, 0.01, -0.0201, 0.257, 0.01, -0.0217, 0.257, 0.01, -0.0201, 0.2591, 0.0061, -0.0214, 0.2591, 0.0083, -0.0214, 0.2606, 0.0061, -0.0212, 0.2617, 0.0061, -0.0211, 0.2606, 0.0083, -0.0212, 0.2617, 0.0083, -0.0211, 0.2591, 0.01, -0.0214, 0.2591, 0.01, -0.0201, 0.2605, 0.01, -0.0212, 0.2613, 0.01, -0.0211, 0.2461, 0.01, -0.0179, 0.2461, 0.01, -0.0157, 0.2483, 0.01, -0.0179, 0.2483, 0.01, -0.0157, 0.2461, 0.01, -0.0136, 0.2461, 0.01, -0.0114, 0.2483, 0.01, -0.0136, 0.2483, 0.01, -0.0114, 0.2504, 0.01, -0.0179,
0.2504, 0.01, -0.0157, 0.2526, 0.01, -0.0179, 0.2526, 0.01, -0.0157, 0.2504, 0.01, -0.0136, 0.2504, 0.01, -0.0114, 0.2526, 0.01, -0.0136, 0.2526, 0.01, -0.0114, 0.2461, 0.0061, -0.0096, 0.2461, 0.0083, -0.0096, 0.2483, 0.0061, -0.0099, 0.2483, 0.0083, -0.0099, 0.2462, 0.01, -0.0096, 0.2483, 0.01, -0.0099, 0.2504, 0.0061, -0.0101, 0.2504, 0.0083, -0.0101, 0.2526, 0.0061, -0.0101, 0.2526, 0.0083, -0.0101, 0.2504, 0.01, -0.0101, 0.2526, 0.01, -0.0101, 0.2548, 0.01, -0.0179, 0.2548, 0.01, -0.0157, 0.257, 0.01, -0.0179, 0.257, 0.01, -0.0157, 0.2548, 0.01, -0.0136, 0.2548, 0.01, -0.0114, 0.257, 0.01, -0.0136, 0.257, 0.01, -0.0114, 0.2591, 0.01, -0.0179, 0.2591, 0.01, -0.0157, 0.2613, 0.01, -0.0179,
0.2613, 0.01, -0.0157, 0.2591, 0.01, -0.0136, 0.2591, 0.01, -0.0114, 0.2613, 0.01, -0.0136, 0.2613, 0.01, -0.0114, 0.2548, 0.0061, -0.0101, 0.2548, 0.0083, -0.0101, 0.257, 0.0061, -0.01, 0.257, 0.0083, -0.01, 0.2548, 0.01, -0.0101, 0.2569, 0.01, -0.01, 0.2591, 0.0061, -0.0097, 0.2591, 0.0083, -0.0097, 0.2613, 0.0061, -0.0093, 0.2613, 0.0083, -0.0093, 0.2591, 0.01, -0.0097, 0.2612, 0.01, -0.0093, 0.2635, 0.0061, -0.0209, 0.2635, 0.0083, -0.0209, 0.2656, 0.0061, -0.0206, 0.2656, 0.0083, -0.0206, 0.2636, 0.01, -0.0209, 0.2658, 0.01, -0.0206, 0.2678, 0.0061, -0.0202, 0.2678, 0.0083, -0.0202, 0.27, 0.0061, -0.0197, 0.27, 0.0083, -0.0197, 0.2679, 0.01, -0.0202, 0.2701, 0.01, -0.0197, 0.2717, 0.0061, -0.0192,
0.2717, 0.0083, -0.0192, 0.2715, 0.01, -0.0192, 0.2635, 0.01, -0.0179, 0.2635, 0.01, -0.0157, 0.2656, 0.01, -0.0179, 0.2656, 0.01, -0.0157, 0.2635, 0.01, -0.0136, 0.2635, 0.01, -0.0114, 0.2656, 0.01, -0.0136, 0.2656, 0.01, -0.0114, 0.2678, 0.01, -0.0179, 0.2678, 0.01, -0.0157, 0.27, 0.01, -0.0179, 0.27, 0.01, -0.0157, 0.2678, 0.01, -0.0136, 0.2678, 0.01, -0.0114, 0.27, 0.01, -0.0136, 0.27, 0.01, -0.0114, 0.2635, 0.0061, -0.0089, 0.2635, 0.0083, -0.0089, 0.2655, 0.0061, -0.0085, 0.2666, 0.0061, -0.0081, 0.2655, 0.0083, -0.0085, 0.2666, 0.0083, -0.0081, 0.2633, 0.01, -0.009, 0.2656, 0.01, -0.0084, 0.2666, 0.01, -0.0081, 0.2678, 0.0061, -0.0076, 0.2678, 0.0083, -0.0076, 0.27, 0.0061, -0.0066,
0.27, 0.0083, -0.0066, 0.271, 0.0061, -0.0059, 0.271, 0.0083, -0.0059, 0.2678, 0.01, -0.0092, 0.2677, 0.01, -0.0077, 0.27, 0.01, -0.0092, 0.2699, 0.01, -0.0066, 0.271, 0.01, -0.0059, 0.2728, 0.0061, -0.0189, 0.2728, 0.0083, -0.0189, 0.2743, 0.0061, -0.0183, 0.2743, 0.0083, -0.0183, 0.2725, 0.01, -0.019, 0.2721, 0.01, -0.0157, 0.2746, 0.01, -0.0182, 0.2743, 0.01, -0.0157, 0.2721, 0.01, -0.0136, 0.2721, 0.01, -0.0114, 0.2743, 0.01, -0.0136, 0.2743, 0.01, -0.0114, 0.2765, 0.0061, -0.0175, 0.2765, 0.0083, -0.0175, 0.2776, 0.0061, -0.0169, 0.2788, 0.0061, -0.0163, 0.2776, 0.0083, -0.0169, 0.2788, 0.0083, -0.0163, 0.2766, 0.01, -0.0174, 0.2765, 0.01, -0.0157, 0.2776, 0.01, -0.0169, 0.2788, 0.01, -0.0163,
0.2765, 0.01, -0.0136, 0.2765, 0.01, -0.0114, 0.2787, 0.01, -0.0136, 0.2787, 0.01, -0.0114, 0.2722, 0.0061, -0.0051, 0.2722, 0.0083, -0.0051, 0.2735, 0.0061, -0.004, 0.2746, 0.0061, -0.0029, 0.2735, 0.0083, -0.004, 0.2746, 0.0083, -0.0029, 0.2721, 0.01, -0.0092, 0.2721, 0.01, -0.0071, 0.2743, 0.01, -0.0092, 0.2743, 0.01, -0.0071, 0.272, 0.01, -0.0053, 0.2735, 0.01, -0.004, 0.2746, 0.01, -0.0029, 0.2755, 0.0061, -0.0017, 0.2755, 0.0083, -0.0017, 0.2765, 0.01, -0.0092, 0.2765, 0.01, -0.0071, 0.2787, 0.01, -0.0092, 0.2787, 0.01, -0.0071, 0.2765, 0.01, -0.0049, 0.2754, 0.01, -0.0019, 0.2787, 0.01, -0.0049, 0.2787, 0.01, -0.0027, 0.298, 0.0061, -0.067, 0.298, 0.0083, -0.067, 0.3004, 0.0061, -0.067,
0.3004, 0.0083, -0.067, 0.298, 0.0061, -0.0656, 0.298, 0.0061, -0.0635, 0.298, 0.0083, -0.0656, 0.298, 0.0083, -0.0635, 0.298, 0.01, -0.067, 0.3004, 0.01, -0.067, 0.298, 0.01, -0.0656, 0.298, 0.01, -0.0635, 0.3004, 0.01, -0.0656, 0.3004, 0.01, -0.0635, 0.3025, 0.0061, -0.067, 0.3025, 0.0083, -0.067, 0.3047, 0.0061, -0.067, 0.3047, 0.0083, -0.067, 0.302, 0.0061, -0.0631, 0.302, 0.0083, -0.0631, 0.3047, 0.0061, -0.0631, 0.3047, 0.0083, -0.0631, 0.3025, 0.01, -0.067, 0.3047, 0.01, -0.067, 0.3025, 0.01, -0.0656, 0.302, 0.01, -0.0631, 0.3047, 0.01, -0.0656, 0.3047, 0.01, -0.0631, 0.298, 0.0061, -0.0613, 0.298, 0.0061, -0.0591, 0.298, 0.0083, -0.0613, 0.298, 0.0083, -0.0591, 0.298, 0.0061, -0.057,
0.298, 0.0061, -0.0548, 0.298, 0.0083, -0.057, 0.298, 0.0083, -0.0548, 0.298, 0.01, -0.0613, 0.298, 0.01, -0.0591, 0.3004, 0.01, -0.0613, 0.3004, 0.01, -0.0591, 0.298, 0.01, -0.057, 0.298, 0.01, -0.0548, 0.3004, 0.01, -0.057, 0.3004, 0.01, -0.0548, 0.302, 0.0061, -0.0613, 0.302, 0.0061, -0.0591, 0.302, 0.0083, -0.0613, 0.302, 0.0083, -0.0591, 0.302, 0.0061, -0.057, 0.302, 0.0061, -0.0548, 0.302, 0.0083, -0.057, 0.302, 0.0083, -0.0548, 0.3049, 0.0061, -0.0562, 0.3049, 0.0061, -0.0548, 0.3049, 0.0083, -0.0562, 0.3049, 0.0083, -0.0548, 0.302, 0.01, -0.0613, 0.302, 0.01, -0.0591, 0.302, 0.01, -0.057, 0.302, 0.01, -0.0548, 0.3049, 0.01, -0.0562, 0.3049, 0.01, -0.0548, 0.3069, 0.0061, -0.067,
0.3069, 0.0083, -0.067, 0.309, 0.0061, -0.067, 0.309, 0.0083, -0.067, 0.3069, 0.0061, -0.0631, 0.3069, 0.0083, -0.0631, 0.309, 0.0061, -0.0631, 0.309, 0.0083, -0.0631, 0.3069, 0.01, -0.067, 0.309, 0.01, -0.067, 0.3069, 0.01, -0.0656, 0.3069, 0.01, -0.0631, 0.309, 0.01, -0.0656, 0.309, 0.01, -0.0631, 0.3112, 0.0061, -0.067, 0.3112, 0.0083, -0.067, 0.3134, 0.0061, -0.067, 0.3134, 0.0083, -0.067, 0.3112, 0.0061, -0.0631, 0.3112, 0.0083, -0.0631, 0.3134, 0.0061, -0.0631, 0.3134, 0.0083, -0.0631, 0.3112, 0.01, -0.067, 0.3134, 0.01, -0.067, 0.3112, 0.01, -0.0656, 0.3112, 0.01, -0.0631, 0.3134, 0.01, -0.0656, 0.3134, 0.01, -0.0631, 0.3069, 0.0061, -0.0562, 0.3069, 0.0083, -0.0562, 0.309, 0.0061, -0.0562,
0.309, 0.0083, -0.0562, 0.3069, 0.01, -0.0562, 0.3069, 0.01, -0.0548, 0.309, 0.01, -0.0562, 0.309, 0.01, -0.0548, 0.3112, 0.0061, -0.0562, 0.3112, 0.0083, -0.0562, 0.3134, 0.0061, -0.0562, 0.3134, 0.0083, -0.0562, 0.3112, 0.01, -0.0562, 0.3112, 0.01, -0.0548, 0.3134, 0.01, -0.0562, 0.3134, 0.01, -0.0548, 0.298, 0.0061, -0.0526, 0.298, 0.0061, -0.0505, 0.298, 0.0083, -0.0526, 0.298, 0.0083, -0.0505, 0.298, 0.0061, -0.0483, 0.298, 0.0061, -0.0461, 0.298, 0.0083, -0.0483, 0.298, 0.0083, -0.0461, 0.298, 0.01, -0.0526, 0.298, 0.01, -0.0505, 0.3004, 0.01, -0.0526, 0.3004, 0.01, -0.0505, 0.298, 0.01, -0.0483, 0.298, 0.01, -0.0461, 0.3004, 0.01, -0.0483, 0.3004, 0.01, -0.0461, 0.302, 0.0061, -0.0526,
0.302, 0.0061, -0.0505, 0.302, 0.0083, -0.0526, 0.302, 0.0083, -0.0505, 0.3049, 0.0061, -0.0523, 0.3049, 0.0083, -0.0523, 0.302, 0.0061, -0.0483, 0.302, 0.0061, -0.0454, 0.302, 0.0083, -0.0483, 0.302, 0.0083, -0.0454, 0.3047, 0.0061, -0.0454, 0.3047, 0.0083, -0.0454, 0.302, 0.01, -0.0526, 0.302, 0.01, -0.0505, 0.3049, 0.01, -0.0523, 0.302, 0.01, -0.0483, 0.302, 0.01, -0.0454, 0.3047, 0.01, -0.0454, 0.298, 0.0061, -0.0439, 0.298, 0.0061, -0.0425, 0.298, 0.0083, -0.0439, 0.298, 0.0083, -0.0425, 0.3004, 0.0061, -0.0425, 0.3004, 0.0083, -0.0425, 0.298, 0.01, -0.0439, 0.298, 0.01, -0.0425, 0.3004, 0.01, -0.0439, 0.3004, 0.01, -0.0425, 0.3025, 0.0061, -0.0425, 0.3025, 0.0083, -0.0425, 0.3047, 0.0061, -0.0425,
0.3047, 0.0083, -0.0425, 0.3025, 0.01, -0.0439, 0.3025, 0.01, -0.0425, 0.3047, 0.01, -0.0439, 0.3047, 0.01, -0.0425, 0.3069, 0.0061, -0.0523, 0.3069, 0.0083, -0.0523, 0.309, 0.0061, -0.0523, 0.309, 0.0083, -0.0523, 0.3069, 0.0061, -0.0454, 0.3069, 0.0083, -0.0454, 0.309, 0.0061, -0.0454, 0.309, 0.0083, -0.0454, 0.3069, 0.01, -0.0523, 0.309, 0.01, -0.0523, 0.3069, 0.01, -0.0454, 0.309, 0.01, -0.0454, 0.3112, 0.0061, -0.0523, 0.3112, 0.0083, -0.0523, 0.3134, 0.0061, -0.0523, 0.3134, 0.0083, -0.0523, 0.3112, 0.0061, -0.0454, 0.3112, 0.0083, -0.0454, 0.3134, 0.0061, -0.0454, 0.3134, 0.0083, -0.0454, 0.3112, 0.01, -0.0523, 0.3134, 0.01, -0.0523, 0.3112, 0.01, -0.0454, 0.3134, 0.01, -0.0454, 0.3069, 0.0061, -0.0425,
0.3069, 0.0083, -0.0425, 0.309, 0.0061, -0.0425, 0.309, 0.0083, -0.0425, 0.3069, 0.01, -0.0439, 0.3069, 0.01, -0.0425, 0.309, 0.01, -0.0439, 0.309, 0.01, -0.0425, 0.3112, 0.0061, -0.0425, 0.3112, 0.0083, -0.0425, 0.3134, 0.0061, -0.0425, 0.3134, 0.0083, -0.0425, 0.3112, 0.01, -0.0439, 0.3112, 0.01, -0.0425, 0.3134, 0.01, -0.0439, 0.3134, 0.01, -0.0425, 0.2805, 0.0061, -0.0152, 0.2805, 0.0083, -0.0152, 0.2815, 0.0061, -0.0143, 0.2815, 0.0083, -0.0143, 0.2828, 0.0061, -0.0133, 0.2838, 0.0061, -0.0122, 0.2828, 0.0083, -0.0133, 0.2838, 0.0083, -0.0122, 0.2804, 0.01, -0.0152, 0.2815, 0.01, -0.0144, 0.2808, 0.01, -0.0114, 0.2827, 0.01, -0.0133, 0.2838, 0.01, -0.0122, 0.2847, 0.0061, -0.0111, 0.2847, 0.0083, -0.0111,
0.2848, 0.01, -0.0111, 0.2808, 0.01, -0.0092, 0.2808, 0.01, -0.0071, 0.283, 0.01, -0.0092, 0.283, 0.01, -0.0071, 0.2808, 0.01, -0.0049, 0.2808, 0.01, -0.0027, 0.283, 0.01, -0.0049, 0.283, 0.01, -0.0027, 0.2858, 0.0061, -0.0097, 0.2858, 0.0083, -0.0097, 0.2865, 0.0061, -0.0086, 0.2874, 0.0061, -0.0071, 0.2865, 0.0083, -0.0086, 0.2874, 0.0083, -0.0071, 0.2882, 0.0061, -0.0054, 0.2882, 0.0083, -0.0054, 0.2858, 0.01, -0.0096, 0.2852, 0.01, -0.0071, 0.2866, 0.01, -0.0085, 0.2873, 0.01, -0.0073, 0.2852, 0.01, -0.0049, 0.2852, 0.01, -0.0027, 0.2883, 0.01, -0.0052, 0.2873, 0.01, -0.0027, 0.2886, 0.0061, -0.0043, 0.2892, 0.0061, -0.0027, 0.2886, 0.0083, -0.0043, 0.2892, 0.0083, -0.0027, 0.2887, 0.01, -0.0042,
0.2891, 0.01, -0.0028, 0.3088, 0.0061, -0.0219, 0.3088, 0.0061, -0.0201, 0.3088, 0.0083, -0.0219, 0.3088, 0.0083, -0.0201, 0.3088, 0.01, -0.0219, 0.3088, 0.01, -0.0201, 0.3112, 0.0061, -0.0219, 0.3112, 0.0083, -0.0219, 0.3134, 0.0061, -0.0219, 0.3134, 0.0083, -0.0219, 0.3112, 0.01, -0.0219, 0.3112, 0.01, -0.0201, 0.3134, 0.01, -0.0219, 0.3134, 0.01, -0.0201, 0.3088, 0.0061, -0.0179, 0.3088, 0.0061, -0.0157, 0.3088, 0.0083, -0.0179, 0.3088, 0.0083, -0.0157, 0.3088, 0.0061, -0.0136, 0.3088, 0.0061, -0.0114, 0.3088, 0.0083, -0.0136, 0.3088, 0.0083, -0.0114, 0.3088, 0.01, -0.0179, 0.3088, 0.01, -0.0157, 0.3088, 0.01, -0.0136, 0.3088, 0.01, -0.0114, 0.3112, 0.01, -0.0179, 0.3112, 0.01, -0.0157, 0.3134, 0.01, -0.0179,
0.3134, 0.01, -0.0157, 0.3112, 0.01, -0.0136, 0.3112, 0.01, -0.0114, 0.3134, 0.01, -0.0136, 0.3134, 0.01, -0.0114, 0.3088, 0.0061, -0.0092, 0.3088, 0.0061, -0.0071, 0.3088, 0.0083, -0.0092, 0.3088, 0.0083, -0.0071, 0.3088, 0.0061, -0.0049, 0.3088, 0.0061, -0.0027, 0.3088, 0.0083, -0.0049, 0.3088, 0.0083, -0.0027, 0.3088, 0.01, -0.0092, 0.3088, 0.01, -0.0071, 0.3088, 0.01, -0.0049, 0.3088, 0.01, -0.0027, 0.3112, 0.01, -0.0092, 0.3112, 0.01, -0.0071, 0.3134, 0.01, -0.0092, 0.3134, 0.01, -0.0071, 0.3112, 0.01, -0.0049, 0.3112, 0.01, -0.0027, 0.3134, 0.01, -0.0049, 0.3134, 0.01, -0.0027, 0.3155, 0.0061, -0.067, 0.3155, 0.0083, -0.067, 0.3177, 0.0061, -0.067, 0.3177, 0.0083, -0.067, 0.3155, 0.0061, -0.0631,
0.3155, 0.0083, -0.0631, 0.3177, 0.0061, -0.0631, 0.3177, 0.0083, -0.0631, 0.3155, 0.01, -0.067, 0.3177, 0.01, -0.067, 0.3155, 0.01, -0.0656, 0.3155, 0.01, -0.0631, 0.3177, 0.01, -0.0656, 0.3177, 0.01, -0.0631, 0.3196, 0.0061, -0.067, 0.3196, 0.0083, -0.067, 0.3196, 0.0061, -0.0656, 0.3196, 0.0061, -0.0631, 0.3196, 0.0083, -0.0656, 0.3196, 0.0083, -0.0631, 0.3196, 0.01, -0.067, 0.3196, 0.01, -0.0656, 0.3196, 0.01, -0.0631, 0.3155, 0.0061, -0.0562, 0.3155, 0.0083, -0.0562, 0.3176, 0.0061, -0.0562, 0.3176, 0.0061, -0.0548, 0.3176, 0.0083, -0.0562, 0.3176, 0.0083, -0.0548, 0.3155, 0.01, -0.0562, 0.3155, 0.01, -0.0548, 0.3176, 0.01, -0.0562, 0.3176, 0.01, -0.0548, 0.3155, 0.0061, -0.0523, 0.3155, 0.0083, -0.0523,
0.3176, 0.0061, -0.0523, 0.3176, 0.0083, -0.0523, 0.3155, 0.0061, -0.0454, 0.3155, 0.0083, -0.0454, 0.3177, 0.0061, -0.0454, 0.3177, 0.0083, -0.0454, 0.3155, 0.01, -0.0523, 0.3176, 0.01, -0.0523, 0.3155, 0.01, -0.0454, 0.3177, 0.01, -0.0454, 0.3196, 0.0061, -0.0454, 0.3196, 0.0083, -0.0454, 0.3196, 0.01, -0.0454, 0.3155, 0.0061, -0.0425, 0.3155, 0.0083, -0.0425, 0.3177, 0.0061, -0.0425, 0.3177, 0.0083, -0.0425, 0.3155, 0.01, -0.0439, 0.3155, 0.01, -0.0425, 0.3177, 0.01, -0.0439, 0.3177, 0.01, -0.0425, 0.3196, 0.0061, -0.0439, 0.3196, 0.0061, -0.0425, 0.3196, 0.0083, -0.0439, 0.3196, 0.0083, -0.0425, 0.3196, 0.01, -0.0439, 0.3196, 0.01, -0.0425, 0.3402, 0.0061, -0.067, 0.3402, 0.0083, -0.067, 0.3402, 0.0061, -0.0656,
0.3402, 0.0061, -0.0635, 0.3402, 0.0083, -0.0656, 0.3402, 0.0083, -0.0635, 0.3402, 0.01, -0.067, 0.3402, 0.01, -0.0656, 0.3402, 0.01, -0.0635, 0.3402, 0.0061, -0.0613, 0.3402, 0.0061, -0.0591, 0.3402, 0.0083, -0.0613, 0.3402, 0.0083, -0.0591, 0.3402, 0.0061, -0.057, 0.3402, 0.0061, -0.0548, 0.3402, 0.0083, -0.057, 0.3402, 0.0083, -0.0548, 0.3402, 0.01, -0.0613, 0.3402, 0.01, -0.0591, 0.3402, 0.01, -0.057, 0.3402, 0.01, -0.0548, 0.3416, 0.0061, -0.067, 0.3416, 0.0083, -0.067, 0.3431, 0.0061, -0.067, 0.3431, 0.0083, -0.067, 0.3431, 0.0061, -0.0656, 0.3431, 0.0061, -0.0635, 0.3431, 0.0083, -0.0656, 0.3431, 0.0083, -0.0635, 0.3416, 0.01, -0.067, 0.3431, 0.01, -0.067, 0.3416, 0.01, -0.0656, 0.3416, 0.01, -0.0635,
0.3431, 0.01, -0.0656, 0.3431, 0.01, -0.0635, 0.3431, 0.0061, -0.0613, 0.3431, 0.0061, -0.0591, 0.3431, 0.0083, -0.0613, 0.3431, 0.0083, -0.0591, 0.3431, 0.0061, -0.057, 0.3431, 0.0061, -0.0548, 0.3431, 0.0083, -0.057, 0.3431, 0.0083, -0.0548, 0.3416, 0.01, -0.0613, 0.3416, 0.01, -0.0591, 0.3431, 0.01, -0.0613, 0.3431, 0.01, -0.0591, 0.3416, 0.01, -0.057, 0.3416, 0.01, -0.0548, 0.3431, 0.01, -0.057, 0.3431, 0.01, -0.0548, 0.3483, 0.0061, -0.0545, 0.3483, 0.0083, -0.0545, 0.3483, 0.01, -0.0546, 0.3402, 0.0061, -0.0526, 0.3402, 0.0061, -0.0505, 0.3402, 0.0083, -0.0526, 0.3402, 0.0083, -0.0505, 0.3402, 0.0061, -0.0483, 0.3402, 0.0061, -0.0461, 0.3402, 0.0083, -0.0483, 0.3402, 0.0083, -0.0461, 0.3402, 0.01, -0.0526,
0.3402, 0.01, -0.0505, 0.3402, 0.01, -0.0483, 0.3402, 0.01, -0.0461, 0.3402, 0.0061, -0.0439, 0.3402, 0.0061, -0.0425, 0.3402, 0.0083, -0.0439, 0.3402, 0.0083, -0.0425, 0.3402, 0.01, -0.0439, 0.3402, 0.01, -0.0425, 0.3431, 0.0061, -0.0526, 0.3431, 0.0061, -0.0499, 0.3431, 0.0083, -0.0526, 0.3431, 0.0083, -0.0499, 0.3416, 0.01, -0.0526, 0.3416, 0.01, -0.0505, 0.3431, 0.01, -0.0526, 0.3431, 0.01, -0.0499, 0.3416, 0.01, -0.0483, 0.3416, 0.01, -0.0461, 0.3438, 0.01, -0.0483, 0.3438, 0.01, -0.0461, 0.346, 0.0061, -0.0524, 0.3449, 0.0061, -0.0515, 0.346, 0.0083, -0.0524, 0.3449, 0.0083, -0.0515, 0.3472, 0.0061, -0.0535, 0.3472, 0.0083, -0.0535, 0.3462, 0.0061, -0.0457, 0.3462, 0.0083, -0.0457, 0.3485, 0.0061, -0.0479,
0.3474, 0.0061, -0.0468, 0.3485, 0.0083, -0.0479, 0.3474, 0.0083, -0.0468, 0.346, 0.01, -0.0525, 0.345, 0.01, -0.0515, 0.3472, 0.01, -0.0535, 0.3481, 0.01, -0.0505, 0.3459, 0.01, -0.0483, 0.3462, 0.01, -0.0457, 0.3485, 0.01, -0.0479, 0.3474, 0.01, -0.0468, 0.3423, 0.0061, -0.043, 0.3413, 0.0061, -0.0426, 0.3423, 0.0083, -0.043, 0.3413, 0.0083, -0.0426, 0.3438, 0.0061, -0.0438, 0.3438, 0.0083, -0.0438, 0.3421, 0.01, -0.0429, 0.3411, 0.01, -0.0426, 0.3441, 0.01, -0.0439, 0.3452, 0.0061, -0.0448, 0.3452, 0.0083, -0.0448, 0.3451, 0.01, -0.0448, 0.3155, 0.0061, -0.0219, 0.3155, 0.0083, -0.0219, 0.3177, 0.0061, -0.0219, 0.3177, 0.0083, -0.0219, 0.3155, 0.01, -0.0219, 0.3155, 0.01, -0.0201, 0.3177, 0.01, -0.0219,
0.3177, 0.01, -0.0201, 0.3206, 0.0061, -0.0219, 0.3206, 0.0061, -0.0201, 0.3206, 0.0083, -0.0219, 0.3206, 0.0083, -0.0201, 0.3206, 0.01, -0.0219, 0.3206, 0.01, -0.0201, 0.3155, 0.01, -0.0179, 0.3155, 0.01, -0.0157, 0.3177, 0.01, -0.0179, 0.3177, 0.01, -0.0157, 0.3155, 0.01, -0.0136, 0.3155, 0.01, -0.0114, 0.3177, 0.01, -0.0136, 0.3177, 0.01, -0.0114, 0.3206, 0.0061, -0.0179, 0.3206, 0.0061, -0.0157, 0.3206, 0.0083, -0.0179, 0.3206, 0.0083, -0.0157, 0.3206, 0.0061, -0.0136, 0.3206, 0.0061, -0.0114, 0.3206, 0.0083, -0.0136, 0.3206, 0.0083, -0.0114, 0.3206, 0.01, -0.0179, 0.3206, 0.01, -0.0157, 0.3206, 0.01, -0.0136, 0.3206, 0.01, -0.0114, 0.3155, 0.01, -0.0092, 0.3155, 0.01, -0.0071, 0.3177, 0.01, -0.0092,
0.3177, 0.01, -0.0071, 0.3155, 0.01, -0.0049, 0.3155, 0.01, -0.0027, 0.3177, 0.01, -0.0049, 0.3177, 0.01, -0.0027, 0.3206, 0.0061, -0.0092, 0.3206, 0.0061, -0.0071, 0.3206, 0.0083, -0.0092, 0.3206, 0.0083, -0.0071, 0.3206, 0.0061, -0.0049, 0.3206, 0.0061, -0.0027, 0.3206, 0.0083, -0.0049, 0.3206, 0.0083, -0.0027, 0.3206, 0.01, -0.0092, 0.3206, 0.01, -0.0071, 0.3206, 0.01, -0.0049, 0.3206, 0.01, -0.0027, 0.3402, 0.0061, -0.0219, 0.3402, 0.0061, -0.0201, 0.3402, 0.0083, -0.0219, 0.3402, 0.0083, -0.0201, 0.3402, 0.01, -0.0219, 0.3402, 0.01, -0.0201, 0.3416, 0.0061, -0.0219, 0.3416, 0.0083, -0.0219, 0.3438, 0.0061, -0.0219, 0.3438, 0.0083, -0.0219, 0.3416, 0.01, -0.0219, 0.3416, 0.01, -0.0201, 0.3438, 0.01, -0.0219,
0.3438, 0.01, -0.0201, 0.3459, 0.0061, -0.0219, 0.3459, 0.0083, -0.0219, 0.3481, 0.0061, -0.0219, 0.3481, 0.0083, -0.0219, 0.3459, 0.01, -0.0219, 0.3459, 0.01, -0.0201, 0.3481, 0.01, -0.0219, 0.3481, 0.01, -0.0201, 0.3402, 0.0061, -0.0179, 0.3402, 0.0061, -0.0157, 0.3402, 0.0083, -0.0179, 0.3402, 0.0083, -0.0157, 0.3402, 0.0061, -0.0136, 0.3402, 0.0061, -0.0114, 0.3402, 0.0083, -0.0136, 0.3402, 0.0083, -0.0114, 0.3402, 0.01, -0.0179, 0.3402, 0.01, -0.0157, 0.3402, 0.01, -0.0136, 0.3402, 0.01, -0.0114, 0.3402, 0.0061, -0.0092, 0.3402, 0.0061, -0.0071, 0.3402, 0.0083, -0.0092, 0.3402, 0.0083, -0.0071, 0.3402, 0.0061, -0.0049, 0.3402, 0.0061, -0.0027, 0.3402, 0.0083, -0.0049, 0.3402, 0.0083, -0.0027, 0.3402, 0.01, -0.0092,
0.3402, 0.01, -0.0071, 0.3402, 0.01, -0.0049, 0.3402, 0.01, -0.0027, 0.3416, 0.01, -0.0179, 0.3416, 0.01, -0.0157, 0.3438, 0.01, -0.0179, 0.3438, 0.01, -0.0157, 0.3416, 0.01, -0.0136, 0.3416, 0.01, -0.0114, 0.3438, 0.01, -0.0136, 0.3438, 0.01, -0.0114, 0.3459, 0.01, -0.0179, 0.3459, 0.01, -0.0157, 0.3481, 0.01, -0.0179, 0.3481, 0.01, -0.0157, 0.3459, 0.01, -0.0136, 0.3459, 0.01, -0.0114, 0.3481, 0.01, -0.0136, 0.3481, 0.01, -0.0114, 0.3416, 0.01, -0.0092, 0.3416, 0.01, -0.0071, 0.3438, 0.01, -0.0092, 0.3438, 0.01, -0.0071, 0.3416, 0.01, -0.0049, 0.3416, 0.01, -0.0027, 0.3438, 0.01, -0.0049, 0.3438, 0.01, -0.0027, 0.3459, 0.01, -0.0092, 0.3459, 0.01, -0.0071, 0.3481, 0.01, -0.0092,
0.3481, 0.01, -0.0071, 0.3459, 0.01, -0.0049, 0.3459, 0.01, -0.0027, 0.3481, 0.01, -0.0049, 0.3481, 0.01, -0.0027, 0.357, 0.0061, -0.0631, 0.357, 0.0083, -0.0631, 0.3571, 0.01, -0.0631, 0.3527, 0.0061, -0.0589, 0.3527, 0.0083, -0.0589, 0.3505, 0.0061, -0.0567, 0.3494, 0.0061, -0.0556, 0.3505, 0.0083, -0.0567, 0.3494, 0.0083, -0.0556, 0.3516, 0.0061, -0.0578, 0.3516, 0.0083, -0.0578, 0.3527, 0.01, -0.0589, 0.3505, 0.01, -0.0567, 0.3494, 0.01, -0.0556, 0.3516, 0.01, -0.0578, 0.3524, 0.01, -0.0548, 0.3549, 0.0061, -0.061, 0.3538, 0.0061, -0.06, 0.3549, 0.0083, -0.061, 0.3538, 0.0083, -0.06, 0.356, 0.0061, -0.0621, 0.356, 0.0083, -0.0621, 0.355, 0.0061, -0.0544, 0.355, 0.0083, -0.0544, 0.3572, 0.0061, -0.0565,
0.3562, 0.0061, -0.0554, 0.3572, 0.0083, -0.0565, 0.3562, 0.0083, -0.0554, 0.3549, 0.01, -0.061, 0.3538, 0.01, -0.06, 0.356, 0.01, -0.0621, 0.3568, 0.01, -0.0591, 0.3546, 0.01, -0.057, 0.355, 0.01, -0.0544, 0.3572, 0.01, -0.0565, 0.3561, 0.01, -0.0554, 0.3594, 0.0061, -0.065, 0.3583, 0.0061, -0.0642, 0.3594, 0.0083, -0.065, 0.3583, 0.0083, -0.0642, 0.361, 0.0061, -0.0662, 0.361, 0.0083, -0.0662, 0.3594, 0.01, -0.0651, 0.3583, 0.01, -0.0642, 0.3608, 0.01, -0.066, 0.3611, 0.01, -0.0635, 0.3637, 0.0061, -0.067, 0.3637, 0.0061, -0.0635, 0.3637, 0.0083, -0.067, 0.3637, 0.0083, -0.0635, 0.3637, 0.01, -0.067, 0.3637, 0.01, -0.0635, 0.3596, 0.0061, -0.0584, 0.3596, 0.0083, -0.0584, 0.3608, 0.0061, -0.0591,
0.3608, 0.0083, -0.0591, 0.3585, 0.0061, -0.0575, 0.3585, 0.0083, -0.0575, 0.3608, 0.0061, -0.057, 0.3608, 0.0061, -0.0548, 0.3608, 0.0083, -0.057, 0.3608, 0.0083, -0.0548, 0.3589, 0.01, -0.0613, 0.3595, 0.01, -0.0583, 0.3611, 0.01, -0.0613, 0.3608, 0.01, -0.0591, 0.3585, 0.01, -0.0575, 0.3608, 0.01, -0.057, 0.3608, 0.01, -0.0548, 0.3637, 0.0061, -0.0613, 0.3637, 0.0061, -0.0591, 0.3637, 0.0083, -0.0613, 0.3637, 0.0083, -0.0591, 0.3637, 0.0061, -0.057, 0.3637, 0.0061, -0.0548, 0.3637, 0.0083, -0.057, 0.3637, 0.0083, -0.0548, 0.3637, 0.01, -0.0613, 0.3637, 0.01, -0.0591, 0.3637, 0.01, -0.057, 0.3637, 0.01, -0.0548, 0.3507, 0.0061, -0.05, 0.3507, 0.0083, -0.05, 0.3529, 0.0061, -0.0522, 0.3518, 0.0061, -0.0511,
0.3529, 0.0083, -0.0522, 0.3518, 0.0083, -0.0511, 0.3496, 0.0061, -0.049, 0.3496, 0.0083, -0.049, 0.3503, 0.01, -0.0526, 0.3507, 0.01, -0.05, 0.3529, 0.01, -0.0522, 0.3518, 0.01, -0.0511, 0.3496, 0.01, -0.0489, 0.354, 0.0061, -0.0533, 0.354, 0.0083, -0.0533, 0.354, 0.01, -0.0533, 0.3608, 0.0061, -0.0526, 0.3608, 0.0061, -0.0505, 0.3608, 0.0083, -0.0526, 0.3608, 0.0083, -0.0505, 0.3608, 0.0061, -0.0483, 0.3608, 0.0061, -0.0461, 0.3608, 0.0083, -0.0483, 0.3608, 0.0083, -0.0461, 0.3608, 0.01, -0.0526, 0.3608, 0.01, -0.0505, 0.3608, 0.01, -0.0483, 0.3608, 0.01, -0.0461, 0.3637, 0.0061, -0.0526, 0.3637, 0.0061, -0.0505, 0.3637, 0.0083, -0.0526, 0.3637, 0.0083, -0.0505, 0.3637, 0.0061, -0.0483, 0.3637, 0.0061, -0.0461,
0.3637, 0.0083, -0.0483, 0.3637, 0.0083, -0.0461, 0.3637, 0.01, -0.0526, 0.3637, 0.01, -0.0505, 0.3637, 0.01, -0.0483, 0.3637, 0.01, -0.0461, 0.3608, 0.0061, -0.0439, 0.3608, 0.0061, -0.0425, 0.3608, 0.0083, -0.0439, 0.3608, 0.0083, -0.0425, 0.3608, 0.01, -0.0439, 0.3608, 0.01, -0.0425, 0.3637, 0.0061, -0.0439, 0.3637, 0.0061, -0.0425, 0.3637, 0.0083, -0.0439, 0.3637, 0.0083, -0.0425, 0.3637, 0.01, -0.0439, 0.3637, 0.01, -0.0425, 0.3503, 0.0061, -0.0219, 0.3503, 0.0083, -0.0219, 0.3524, 0.0061, -0.0219, 0.3524, 0.0083, -0.0219, 0.3503, 0.01, -0.0219, 0.3503, 0.01, -0.0201, 0.3524, 0.01, -0.0219, 0.3524, 0.01, -0.0201, 0.3546, 0.0061, -0.0219, 0.3546, 0.0083, -0.0219, 0.3568, 0.0061, -0.0219, 0.3568, 0.0083, -0.0219,
0.3546, 0.01, -0.0219, 0.3546, 0.01, -0.0201, 0.3568, 0.01, -0.0219, 0.3568, 0.01, -0.0201, 0.3589, 0.0061, -0.0219, 0.3589, 0.0083, -0.0219, 0.3611, 0.0061, -0.0219, 0.3611, 0.0083, -0.0219, 0.3589, 0.01, -0.0219, 0.3589, 0.01, -0.0201, 0.3611, 0.01, -0.0219, 0.3611, 0.01, -0.0201, 0.3633, 0.0061, -0.0219, 0.3633, 0.0083, -0.0219, 0.3655, 0.0061, -0.0219, 0.3655, 0.0083, -0.0219, 0.3633, 0.01, -0.0219, 0.3633, 0.01, -0.0201, 0.3655, 0.01, -0.0219, 0.3655, 0.01, -0.0201, 0.3503, 0.01, -0.0179, 0.3503, 0.01, -0.0157, 0.3524, 0.01, -0.0179, 0.3524, 0.01, -0.0157, 0.3503, 0.01, -0.0136, 0.3503, 0.01, -0.0114, 0.3524, 0.01, -0.0136, 0.3524, 0.01, -0.0114, 0.3546, 0.01, -0.0179, 0.3546, 0.01, -0.0157,
0.3568, 0.01, -0.0179, 0.3568, 0.01, -0.0157, 0.3546, 0.01, -0.0136, 0.3546, 0.01, -0.0114, 0.3568, 0.01, -0.0136, 0.3568, 0.01, -0.0114, 0.352, 0.0061, -0.0101, 0.352, 0.0061, -0.0071, 0.352, 0.0083, -0.0101, 0.352, 0.0083, -0.0071, 0.352, 0.0061, -0.0049, 0.352, 0.0061, -0.0027, 0.352, 0.0083, -0.0049, 0.352, 0.0083, -0.0027, 0.3503, 0.01, -0.0092, 0.3503, 0.01, -0.0071, 0.352, 0.01, -0.0101, 0.352, 0.01, -0.0071, 0.3503, 0.01, -0.0049, 0.3503, 0.01, -0.0027, 0.352, 0.01, -0.0049, 0.352, 0.01, -0.0027, 0.3546, 0.0061, -0.0101, 0.3546, 0.0083, -0.0101, 0.3568, 0.0061, -0.0101, 0.3568, 0.0083, -0.0101, 0.3546, 0.01, -0.0101, 0.3568, 0.01, -0.0101, 0.3589, 0.01, -0.0179, 0.3589, 0.01, -0.0157,
0.3611, 0.01, -0.0179, 0.3611, 0.01, -0.0157, 0.3589, 0.01, -0.0136, 0.3589, 0.01, -0.0114, 0.3611, 0.01, -0.0136, 0.3611, 0.01, -0.0114, 0.3633, 0.01, -0.0179, 0.3633, 0.01, -0.0157, 0.3655, 0.01, -0.0179, 0.3655, 0.01, -0.0157, 0.3633, 0.01, -0.0136, 0.3633, 0.01, -0.0114, 0.3655, 0.01, -0.0136, 0.3655, 0.01, -0.0114, 0.3589, 0.0061, -0.0101, 0.3589, 0.0083, -0.0101, 0.3611, 0.0061, -0.0101, 0.3611, 0.0083, -0.0101, 0.3589, 0.01, -0.0101, 0.3611, 0.01, -0.0101, 0.3633, 0.0061, -0.0101, 0.3633, 0.0083, -0.0101, 0.3655, 0.0061, -0.0101, 0.3655, 0.0083, -0.0101, 0.3633, 0.01, -0.0101, 0.3655, 0.01, -0.0101, 0.3676, 0.0061, -0.0219, 0.3676, 0.0083, -0.0219, 0.3698, 0.0061, -0.0219, 0.3698, 0.0083, -0.0219,
0.3676, 0.01, -0.0219, 0.3676, 0.01, -0.0201, 0.3698, 0.01, -0.0219, 0.3698, 0.01, -0.0201, 0.372, 0.0061, -0.0219, 0.372, 0.0083, -0.0219, 0.3741, 0.0061, -0.0219, 0.3741, 0.0083, -0.0219, 0.372, 0.01, -0.0219, 0.372, 0.01, -0.0201, 0.3741, 0.01, -0.0219, 0.3741, 0.01, -0.0201, 0.3763, 0.0061, -0.0219, 0.3763, 0.0083, -0.0219, 0.3785, 0.0061, -0.0219, 0.3785, 0.0083, -0.0219, 0.3763, 0.01, -0.0219, 0.3763, 0.01, -0.0201, 0.3785, 0.01, -0.0219, 0.3785, 0.01, -0.0201, 0.3807, 0.0061, -0.0218, 0.3807, 0.0083, -0.0218, 0.3828, 0.0061, -0.0215, 0.3828, 0.0083, -0.0215, 0.3807, 0.01, -0.0218, 0.3806, 0.01, -0.0201, 0.3828, 0.01, -0.0215, 0.3828, 0.01, -0.0201, 0.3676, 0.01, -0.0179, 0.3676, 0.01, -0.0157,
0.3698, 0.01, -0.0179, 0.3698, 0.01, -0.0157, 0.3676, 0.01, -0.0136, 0.3676, 0.01, -0.0114, 0.3698, 0.01, -0.0136, 0.3698, 0.01, -0.0114, 0.372, 0.01, -0.0179, 0.372, 0.01, -0.0157, 0.3741, 0.01, -0.0179, 0.3741, 0.01, -0.0157, 0.372, 0.01, -0.0136, 0.372, 0.01, -0.0114, 0.3741, 0.01, -0.0136, 0.3741, 0.01, -0.0114, 0.3676, 0.0061, -0.0101, 0.3676, 0.0083, -0.0101, 0.3698, 0.0061, -0.0101, 0.3698, 0.0083, -0.0101, 0.3676, 0.01, -0.0101, 0.3698, 0.01, -0.0101, 0.372, 0.0061, -0.0101, 0.372, 0.0083, -0.0101, 0.3741, 0.0061, -0.0101, 0.3741, 0.0083, -0.0101, 0.372, 0.01, -0.0101, 0.3741, 0.01, -0.0101, 0.3763, 0.01, -0.0179, 0.3763, 0.01, -0.0157, 0.3785, 0.01, -0.0179, 0.3785, 0.01, -0.0157,
0.3763, 0.01, -0.0136, 0.3763, 0.01, -0.0114, 0.3785, 0.01, -0.0136, 0.3785, 0.01, -0.0114, 0.3806, 0.01, -0.0179, 0.3806, 0.01, -0.0157, 0.3828, 0.01, -0.0179, 0.3828, 0.01, -0.0157, 0.3806, 0.01, -0.0136, 0.3806, 0.01, -0.0114, 0.3828, 0.01, -0.0136, 0.3828, 0.01, -0.0114, 0.3763, 0.0061, -0.0101, 0.3763, 0.0083, -0.0101, 0.3785, 0.0061, -0.0101, 0.3785, 0.0083, -0.0101, 0.3763, 0.01, -0.0101, 0.3785, 0.01, -0.0101, 0.3807, 0.0061, -0.0101, 0.3807, 0.0083, -0.0101, 0.3828, 0.0061, -0.0099, 0.3828, 0.0083, -0.0099, 0.3806, 0.01, -0.0101, 0.3828, 0.01, -0.0099, 0.3863, 0.0061, -0.067, 0.3863, 0.0083, -0.067, 0.3863, 0.0061, -0.0656, 0.3863, 0.0061, -0.0635, 0.3863, 0.0083, -0.0656, 0.3863, 0.0083, -0.0635,
0.3863, 0.01, -0.067, 0.3863, 0.01, -0.0656, 0.3863, 0.01, -0.0635, 0.3893, 0.0061, -0.067, 0.3893, 0.0083, -0.067, 0.3915, 0.0061, -0.067, 0.3915, 0.0083, -0.067, 0.3902, 0.0061, -0.0631, 0.3902, 0.0083, -0.0631, 0.3915, 0.0061, -0.0631, 0.3915, 0.0083, -0.0631, 0.3893, 0.01, -0.067, 0.3915, 0.01, -0.067, 0.3893, 0.01, -0.0656, 0.3902, 0.01, -0.0631, 0.3915, 0.01, -0.0656, 0.3915, 0.01, -0.0631, 0.3863, 0.0061, -0.0613, 0.3863, 0.0061, -0.0591, 0.3863, 0.0083, -0.0613, 0.3863, 0.0083, -0.0591, 0.3863, 0.0061, -0.057, 0.3863, 0.0061, -0.0548, 0.3863, 0.0083, -0.057, 0.3863, 0.0083, -0.0548, 0.3863, 0.01, -0.0613, 0.3863, 0.01, -0.0591, 0.3863, 0.01, -0.057, 0.3863, 0.01, -0.0548, 0.3902, 0.0061, -0.0613,
0.3902, 0.0061, -0.0591, 0.3902, 0.0083, -0.0613, 0.3902, 0.0083, -0.0591, 0.3902, 0.0061, -0.057, 0.3902, 0.0061, -0.0548, 0.3902, 0.0083, -0.057, 0.3902, 0.0083, -0.0548, 0.3902, 0.01, -0.0613, 0.3902, 0.01, -0.0591, 0.3902, 0.01, -0.057, 0.3902, 0.01, -0.0548, 0.3937, 0.0061, -0.067, 0.3937, 0.0083, -0.067, 0.3958, 0.0061, -0.067, 0.3958, 0.0083, -0.067, 0.3937, 0.0061, -0.0631, 0.3937, 0.0083, -0.0631, 0.3958, 0.0061, -0.0631, 0.3958, 0.0083, -0.0631, 0.3937, 0.01, -0.067, 0.3958, 0.01, -0.067, 0.3937, 0.01, -0.0656, 0.3937, 0.01, -0.0631, 0.3958, 0.01, -0.0656, 0.3958, 0.01, -0.0631, 0.398, 0.0061, -0.067, 0.398, 0.0083, -0.067, 0.4002, 0.0061, -0.0669, 0.4002, 0.0083, -0.0669, 0.398, 0.0061, -0.063,
0.398, 0.0083, -0.063, 0.4002, 0.0061, -0.063, 0.4002, 0.0083, -0.063, 0.398, 0.01, -0.067, 0.4002, 0.01, -0.0669, 0.398, 0.01, -0.0656, 0.398, 0.01, -0.063, 0.4002, 0.01, -0.0656, 0.4001, 0.01, -0.063, 0.3863, 0.0061, -0.0526, 0.3863, 0.0061, -0.0505, 0.3863, 0.0083, -0.0526, 0.3863, 0.0083, -0.0505, 0.3863, 0.0061, -0.0483, 0.3863, 0.0061, -0.0461, 0.3863, 0.0083, -0.0483, 0.3863, 0.0083, -0.0461, 0.3863, 0.01, -0.0526, 0.3863, 0.01, -0.0505, 0.3863, 0.01, -0.0483, 0.3863, 0.01, -0.0461, 0.3902, 0.0061, -0.0526, 0.3902, 0.0061, -0.0505, 0.3902, 0.0083, -0.0526, 0.3902, 0.0083, -0.0505, 0.3902, 0.0061, -0.0483, 0.3902, 0.0061, -0.0454, 0.3902, 0.0083, -0.0483, 0.3902, 0.0083, -0.0454, 0.3915, 0.0061, -0.0454,
0.3915, 0.0083, -0.0454, 0.3902, 0.01, -0.0526, 0.3902, 0.01, -0.0505, 0.3902, 0.01, -0.0483, 0.3902, 0.01, -0.0454, 0.3915, 0.01, -0.0454, 0.3863, 0.0061, -0.0439, 0.3863, 0.0061, -0.0425, 0.3863, 0.0083, -0.0439, 0.3863, 0.0083, -0.0425, 0.3863, 0.01, -0.0439, 0.3863, 0.01, -0.0425, 0.3893, 0.0061, -0.0425, 0.3893, 0.0083, -0.0425, 0.3915, 0.0061, -0.0425, 0.3915, 0.0083, -0.0425, 0.3893, 0.01, -0.0439, 0.3893, 0.01, -0.0425, 0.3915, 0.01, -0.0439, 0.3915, 0.01, -0.0425, 0.3937, 0.0061, -0.0454, 0.3937, 0.0083, -0.0454, 0.3958, 0.0061, -0.0454, 0.3958, 0.0083, -0.0454, 0.3937, 0.01, -0.0454, 0.3958, 0.01, -0.0454, 0.398, 0.0061, -0.0455, 0.398, 0.0083, -0.0455, 0.4002, 0.0061, -0.0458, 0.4002, 0.0083, -0.0458,
0.398, 0.01, -0.0455, 0.4001, 0.01, -0.0457, 0.3937, 0.0061, -0.0425, 0.3937, 0.0083, -0.0425, 0.3958, 0.0061, -0.0425, 0.3958, 0.0083, -0.0425, 0.3937, 0.01, -0.0439, 0.3937, 0.01, -0.0425, 0.3958, 0.01, -0.0439, 0.3958, 0.01, -0.0425, 0.398, 0.0061, -0.0425, 0.398, 0.0083, -0.0425, 0.4002, 0.0061, -0.0425, 0.4002, 0.0083, -0.0425, 0.398, 0.01, -0.0439, 0.398, 0.01, -0.0425, 0.4002, 0.01, -0.0439, 0.4002, 0.01, -0.0425, 0.4017, 0.0061, -0.0668, 0.4017, 0.0083, -0.0668, 0.4028, 0.0061, -0.0666, 0.4023, 0.0061, -0.0627, 0.4028, 0.0083, -0.0666, 0.4023, 0.0083, -0.0627, 0.4046, 0.0061, -0.0662, 0.4046, 0.0083, -0.0662, 0.4016, 0.01, -0.0668, 0.4025, 0.01, -0.0667, 0.4023, 0.01, -0.0627, 0.4048, 0.01, -0.0661,
0.4045, 0.01, -0.0635, 0.4066, 0.0061, -0.0652, 0.4076, 0.0061, -0.0644, 0.4066, 0.0083, -0.0652, 0.4076, 0.0083, -0.0644, 0.409, 0.0061, -0.0635, 0.409, 0.0083, -0.0635, 0.4064, 0.01, -0.0653, 0.4075, 0.01, -0.0645, 0.4089, 0.01, -0.0635, 0.4033, 0.0061, -0.0623, 0.4033, 0.0083, -0.0623, 0.4046, 0.0061, -0.0619, 0.4046, 0.0083, -0.0619, 0.4033, 0.01, -0.0623, 0.4045, 0.01, -0.0619, 0.4062, 0.0061, -0.0608, 0.4074, 0.0061, -0.0595, 0.4062, 0.0083, -0.0608, 0.4074, 0.0083, -0.0595, 0.4099, 0.0061, -0.0624, 0.408, 0.0061, -0.0584, 0.4099, 0.0083, -0.0624, 0.408, 0.0083, -0.0584, 0.4085, 0.0061, -0.057, 0.4087, 0.0061, -0.0548, 0.4085, 0.0083, -0.057, 0.4087, 0.0083, -0.0548, 0.4062, 0.01, -0.0608, 0.4073, 0.01, -0.0596,
0.41, 0.01, -0.0623, 0.4079, 0.01, -0.0587, 0.4085, 0.01, -0.0569, 0.4087, 0.01, -0.0548, 0.4107, 0.0061, -0.0613, 0.4119, 0.0061, -0.0594, 0.4107, 0.0083, -0.0613, 0.4119, 0.0083, -0.0594, 0.4122, 0.0061, -0.0582, 0.4122, 0.0083, -0.0582, 0.4125, 0.0061, -0.057, 0.4127, 0.0061, -0.0548, 0.4125, 0.0083, -0.057, 0.4127, 0.0083, -0.0548, 0.4108, 0.01, -0.0612, 0.4119, 0.01, -0.0593, 0.4122, 0.01, -0.0582, 0.411, 0.01, -0.057, 0.411, 0.01, -0.0548, 0.4125, 0.01, -0.057, 0.4127, 0.01, -0.0548, 0.4024, 0.0061, -0.0463, 0.4024, 0.0083, -0.0463, 0.405, 0.0061, -0.0476, 0.4039, 0.0061, -0.047, 0.405, 0.0083, -0.0476, 0.4039, 0.0083, -0.047, 0.4022, 0.01, -0.0462, 0.4051, 0.01, -0.0476, 0.4041, 0.01, -0.0471,
0.4075, 0.0061, -0.0498, 0.4075, 0.0083, -0.0498, 0.4086, 0.0061, -0.0526, 0.4081, 0.0061, -0.0509, 0.4086, 0.0083, -0.0526, 0.4081, 0.0083, -0.0509, 0.4064, 0.0061, -0.0486, 0.4064, 0.0083, -0.0486, 0.4093, 0.0061, -0.0456, 0.4093, 0.0083, -0.0456, 0.4074, 0.01, -0.0498, 0.4086, 0.01, -0.0528, 0.408, 0.01, -0.0508, 0.4064, 0.01, -0.0486, 0.4067, 0.01, -0.0461, 0.4089, 0.01, -0.0483, 0.4093, 0.01, -0.0456, 0.4024, 0.0061, -0.0426, 0.4024, 0.0083, -0.0426, 0.4048, 0.0061, -0.0431, 0.4037, 0.0061, -0.0428, 0.4048, 0.0083, -0.0431, 0.4037, 0.0083, -0.0428, 0.4024, 0.01, -0.0439, 0.4024, 0.01, -0.0426, 0.4047, 0.01, -0.043, 0.4036, 0.01, -0.0428, 0.4067, 0.0061, -0.0439, 0.4067, 0.0083, -0.0439, 0.4082, 0.0061, -0.0448,
0.4082, 0.0083, -0.0448, 0.407, 0.01, -0.044, 0.4081, 0.01, -0.0447, 0.412, 0.0061, -0.0499, 0.412, 0.0083, -0.0499, 0.4126, 0.0061, -0.0526, 0.4122, 0.0061, -0.051, 0.4126, 0.0083, -0.0526, 0.4122, 0.0083, -0.051, 0.4112, 0.0061, -0.0482, 0.4103, 0.0061, -0.0468, 0.4112, 0.0083, -0.0482, 0.4103, 0.0083, -0.0468, 0.411, 0.01, -0.0526, 0.4121, 0.01, -0.0502, 0.4126, 0.01, -0.0526, 0.4123, 0.01, -0.0511, 0.4111, 0.01, -0.048, 0.4103, 0.01, -0.0468, 0.3846, 0.0061, -0.0213, 0.3857, 0.0061, -0.0211, 0.3846, 0.0083, -0.0213, 0.3857, 0.0083, -0.0211, 0.3872, 0.0061, -0.0209, 0.3872, 0.0083, -0.0209, 0.3844, 0.01, -0.0213, 0.3852, 0.01, -0.0212, 0.3873, 0.01, -0.0209, 0.3893, 0.0061, -0.0206, 0.3893, 0.0083, -0.0206,
0.3915, 0.0061, -0.0201, 0.3915, 0.0083, -0.0201, 0.3895, 0.01, -0.0206, 0.3917, 0.01, -0.02, 0.3935, 0.0061, -0.0194, 0.3935, 0.0083, -0.0194, 0.3933, 0.01, -0.0194, 0.385, 0.01, -0.0179, 0.385, 0.01, -0.0157, 0.3872, 0.01, -0.0179, 0.3872, 0.01, -0.0157, 0.385, 0.01, -0.0136, 0.385, 0.01, -0.0114, 0.3872, 0.01, -0.0136, 0.3872, 0.01, -0.0114, 0.3893, 0.01, -0.0179, 0.3893, 0.01, -0.0157, 0.3915, 0.01, -0.0179, 0.3915, 0.01, -0.0157, 0.3893, 0.01, -0.0136, 0.3893, 0.01, -0.0114, 0.3915, 0.01, -0.0136, 0.3915, 0.01, -0.0114, 0.385, 0.0061, -0.0096, 0.385, 0.0083, -0.0096, 0.3872, 0.0061, -0.009, 0.3872, 0.0083, -0.009, 0.3849, 0.01, -0.0096, 0.387, 0.01, -0.0091, 0.3891, 0.0061, -0.0084,
0.3902, 0.0061, -0.0081, 0.3891, 0.0083, -0.0084, 0.3902, 0.0083, -0.0081, 0.3915, 0.0061, -0.0075, 0.3915, 0.0083, -0.0075, 0.3892, 0.01, -0.0084, 0.3902, 0.01, -0.008, 0.3915, 0.01, -0.0092, 0.3914, 0.01, -0.0076, 0.3946, 0.0061, -0.0189, 0.3946, 0.0083, -0.0189, 0.3959, 0.0061, -0.0185, 0.3959, 0.0083, -0.0185, 0.3942, 0.01, -0.0191, 0.3937, 0.01, -0.0157, 0.3961, 0.01, -0.0183, 0.3958, 0.01, -0.0157, 0.3937, 0.01, -0.0136, 0.3937, 0.01, -0.0114, 0.3958, 0.01, -0.0136, 0.3958, 0.01, -0.0114, 0.398, 0.0061, -0.0175, 0.398, 0.0083, -0.0175, 0.3992, 0.0061, -0.0169, 0.4003, 0.0061, -0.0162, 0.3992, 0.0083, -0.0169, 0.4003, 0.0083, -0.0162, 0.3982, 0.01, -0.0174, 0.398, 0.01, -0.0157, 0.3991, 0.01, -0.0169,
0.4003, 0.01, -0.0162, 0.398, 0.01, -0.0136, 0.398, 0.01, -0.0114, 0.4002, 0.01, -0.0136, 0.4002, 0.01, -0.0114, 0.3935, 0.0061, -0.0065, 0.3935, 0.0083, -0.0065, 0.3945, 0.0061, -0.0058, 0.3945, 0.0083, -0.0058, 0.3959, 0.0061, -0.0049, 0.3959, 0.0083, -0.0049, 0.3937, 0.01, -0.0092, 0.3935, 0.01, -0.0065, 0.3958, 0.01, -0.0092, 0.3958, 0.01, -0.0071, 0.3945, 0.01, -0.0058, 0.3956, 0.01, -0.0051, 0.397, 0.0061, -0.0039, 0.3982, 0.0061, -0.0028, 0.397, 0.0083, -0.0039, 0.3982, 0.0083, -0.0028, 0.3991, 0.0061, -0.0016, 0.3991, 0.0083, -0.0016, 0.398, 0.01, -0.0092, 0.398, 0.01, -0.0071, 0.4002, 0.01, -0.0092, 0.4002, 0.01, -0.0071, 0.3971, 0.01, -0.0038, 0.3981, 0.01, -0.0028, 0.4002, 0.01, -0.0049,
0.399, 0.01, -0.0017, 0.4018, 0.0061, -0.0151, 0.4018, 0.0083, -0.0151, 0.4029, 0.0061, -0.0142, 0.4029, 0.0083, -0.0142, 0.4042, 0.0061, -0.0132, 0.4053, 0.0061, -0.0122, 0.4042, 0.0083, -0.0132, 0.4053, 0.0083, -0.0122, 0.4018, 0.01, -0.0151, 0.4029, 0.01, -0.0143, 0.4024, 0.01, -0.0114, 0.4042, 0.01, -0.0132, 0.4053, 0.01, -0.0122, 0.4064, 0.0061, -0.0111, 0.4064, 0.0083, -0.0111, 0.4064, 0.01, -0.0111, 0.4024, 0.01, -0.0092, 0.4024, 0.01, -0.0071, 0.4045, 0.01, -0.0092, 0.4045, 0.01, -0.0071, 0.4024, 0.01, -0.0049, 0.4024, 0.01, -0.0027, 0.4045, 0.01, -0.0049, 0.4045, 0.01, -0.0027, 0.4075, 0.0061, -0.0099, 0.4075, 0.0083, -0.0099, 0.4084, 0.0061, -0.0089, 0.4094, 0.0061, -0.0074, 0.4084, 0.0083, -0.0089,
0.4094, 0.0083, -0.0074, 0.4075, 0.01, -0.0099, 0.4067, 0.01, -0.0071, 0.4084, 0.01, -0.0088, 0.4095, 0.01, -0.0074, 0.4067, 0.01, -0.0049, 0.4067, 0.01, -0.0027, 0.4089, 0.01, -0.0049, 0.4089, 0.01, -0.0027, 0.4102, 0.0061, -0.0063, 0.4102, 0.0083, -0.0063, 0.4111, 0.0061, -0.0049, 0.4119, 0.0061, -0.0034, 0.4111, 0.0083, -0.0049, 0.4119, 0.0083, -0.0034, 0.4124, 0.0061, -0.0023, 0.4124, 0.0083, -0.0023, 0.4102, 0.01, -0.0063, 0.4109, 0.01, -0.0051, 0.412, 0.01, -0.0032, 0.4125, 0.01, -0.0021, 0.4324, 0.0061, -0.067, 0.4324, 0.0083, -0.067, 0.4349, 0.0061, -0.067, 0.4349, 0.0083, -0.067, 0.4324, 0.0061, -0.0656, 0.4324, 0.0061, -0.0635, 0.4324, 0.0083, -0.0656, 0.4324, 0.0083, -0.0635, 0.4353, 0.0061, -0.0631,
0.4353, 0.0083, -0.0631, 0.4324, 0.01, -0.067, 0.4349, 0.01, -0.067, 0.4324, 0.01, -0.0656, 0.4324, 0.01, -0.0635, 0.4349, 0.01, -0.0656, 0.4353, 0.01, -0.0631, 0.4324, 0.0061, -0.0613, 0.4324, 0.0061, -0.0591, 0.4324, 0.0083, -0.0613, 0.4324, 0.0083, -0.0591, 0.4353, 0.0061, -0.0613, 0.4353, 0.0061, -0.0591, 0.4353, 0.0083, -0.0613, 0.4353, 0.0083, -0.0591, 0.4324, 0.0061, -0.057, 0.4324, 0.0061, -0.0548, 0.4324, 0.0083, -0.057, 0.4324, 0.0083, -0.0548, 0.4353, 0.0061, -0.057, 0.4353, 0.0061, -0.0548, 0.4353, 0.0083, -0.057, 0.4353, 0.0083, -0.0548, 0.4324, 0.01, -0.0613, 0.4324, 0.01, -0.0591, 0.4353, 0.01, -0.0613, 0.4353, 0.01, -0.0591, 0.4324, 0.01, -0.057, 0.4324, 0.01, -0.0548, 0.4353, 0.01, -0.057,
0.4353, 0.01, -0.0548, 0.4324, 0.0061, -0.0526, 0.4324, 0.0061, -0.0505, 0.4324, 0.0083, -0.0526, 0.4324, 0.0083, -0.0505, 0.4353, 0.0061, -0.0526, 0.4355, 0.0061, -0.0497, 0.4353, 0.0083, -0.0526, 0.4355, 0.0083, -0.0497, 0.4324, 0.0061, -0.0483, 0.4324, 0.0061, -0.0461, 0.4324, 0.0083, -0.0483, 0.4324, 0.0083, -0.0461, 0.4358, 0.0061, -0.0489, 0.4358, 0.0083, -0.0489, 0.4324, 0.01, -0.0526, 0.4324, 0.01, -0.0505, 0.4353, 0.01, -0.0527, 0.4353, 0.01, -0.0506, 0.4324, 0.01, -0.0483, 0.4324, 0.01, -0.0461, 0.4359, 0.01, -0.0487, 0.4349, 0.01, -0.0461, 0.4324, 0.0061, -0.0439, 0.4324, 0.0061, -0.0425, 0.4324, 0.0083, -0.0439, 0.4324, 0.0083, -0.0425, 0.4349, 0.0061, -0.0425, 0.4349, 0.0083, -0.0425, 0.4324, 0.01, -0.0439,
0.4324, 0.01, -0.0425, 0.4349, 0.01, -0.0439, 0.4349, 0.01, -0.0425, 0.4371, 0.0061, -0.067, 0.4371, 0.0083, -0.067, 0.4392, 0.0061, -0.067, 0.4392, 0.0083, -0.067, 0.4371, 0.0061, -0.0631, 0.4371, 0.0083, -0.0631, 0.4392, 0.0061, -0.0631, 0.4392, 0.0083, -0.0631, 0.4371, 0.01, -0.067, 0.4392, 0.01, -0.067, 0.4371, 0.01, -0.0656, 0.4371, 0.01, -0.0631, 0.4392, 0.01, -0.0656, 0.4392, 0.01, -0.0631, 0.4414, 0.0061, -0.067, 0.4414, 0.0083, -0.067, 0.4436, 0.0061, -0.067, 0.4436, 0.0083, -0.067, 0.4414, 0.0061, -0.0631, 0.4414, 0.0083, -0.0631, 0.4436, 0.0061, -0.0631, 0.4436, 0.0083, -0.0631, 0.4414, 0.01, -0.067, 0.4436, 0.01, -0.067, 0.4414, 0.01, -0.0656, 0.4414, 0.01, -0.0631, 0.4436, 0.01, -0.0656,
0.4436, 0.01, -0.0631, 0.4382, 0.0061, -0.0562, 0.4382, 0.0061, -0.0548, 0.4382, 0.0083, -0.0562, 0.4382, 0.0083, -0.0548, 0.4382, 0.01, -0.0562, 0.4382, 0.01, -0.0548, 0.4414, 0.0061, -0.0562, 0.4414, 0.0083, -0.0562, 0.4436, 0.0061, -0.0562, 0.4436, 0.0083, -0.0562, 0.4414, 0.01, -0.0562, 0.4414, 0.01, -0.0548, 0.4436, 0.01, -0.0562, 0.4436, 0.01, -0.0548, 0.4458, 0.0061, -0.067, 0.4458, 0.0083, -0.067, 0.4479, 0.0061, -0.067, 0.4479, 0.0083, -0.067, 0.4458, 0.0061, -0.0631, 0.4458, 0.0083, -0.0631, 0.4479, 0.0061, -0.0631, 0.4479, 0.0083, -0.0631, 0.4458, 0.01, -0.067, 0.4479, 0.01, -0.067, 0.4458, 0.01, -0.0656, 0.4458, 0.01, -0.0631, 0.4479, 0.01, -0.0656, 0.4479, 0.01, -0.0631, 0.4501, 0.0061, -0.067,
0.4501, 0.0083, -0.067, 0.4529, 0.0061, -0.067, 0.4529, 0.0083, -0.067, 0.4501, 0.0061, -0.0631, 0.4501, 0.0083, -0.0631, 0.4529, 0.0061, -0.0656, 0.4529, 0.0061, -0.0631, 0.4529, 0.0083, -0.0656, 0.4529, 0.0083, -0.0631, 0.4501, 0.01, -0.067, 0.4529, 0.01, -0.067, 0.4501, 0.01, -0.0656, 0.4501, 0.01, -0.0631, 0.4529, 0.01, -0.0656, 0.4529, 0.01, -0.0631, 0.4458, 0.0061, -0.0562, 0.4458, 0.0083, -0.0562, 0.4479, 0.0061, -0.0562, 0.4479, 0.0083, -0.0562, 0.4458, 0.01, -0.0562, 0.4458, 0.01, -0.0548, 0.4479, 0.01, -0.0562, 0.4479, 0.01, -0.0548, 0.4501, 0.0061, -0.0562, 0.4501, 0.0083, -0.0562, 0.452, 0.0061, -0.0562, 0.452, 0.0061, -0.0548, 0.452, 0.0083, -0.0562, 0.452, 0.0083, -0.0548, 0.4501, 0.01, -0.0562,
0.4501, 0.01, -0.0548, 0.452, 0.01, -0.0562, 0.452, 0.01, -0.0548, 0.4382, 0.0061, -0.0523, 0.4382, 0.0083, -0.0523, 0.4362, 0.0061, -0.0478, 0.4372, 0.0061, -0.0466, 0.4362, 0.0083, -0.0478, 0.4372, 0.0083, -0.0466, 0.4396, 0.0061, -0.0456, 0.4396, 0.0083, -0.0456, 0.4382, 0.01, -0.0523, 0.4363, 0.01, -0.0477, 0.4372, 0.01, -0.0466, 0.4393, 0.01, -0.0457, 0.4414, 0.0061, -0.0523, 0.4414, 0.0083, -0.0523, 0.4436, 0.0061, -0.0523, 0.4436, 0.0083, -0.0523, 0.4414, 0.0061, -0.0455, 0.4414, 0.0083, -0.0455, 0.4436, 0.0061, -0.0454, 0.4436, 0.0083, -0.0454, 0.4414, 0.01, -0.0523, 0.4436, 0.01, -0.0523, 0.4414, 0.01, -0.0455, 0.4436, 0.01, -0.0454, 0.4371, 0.0061, -0.0425, 0.4371, 0.0083, -0.0425, 0.4392, 0.0061, -0.0425,
0.4392, 0.0083, -0.0425, 0.4371, 0.01, -0.0439, 0.4371, 0.01, -0.0425, 0.4392, 0.01, -0.0439, 0.4392, 0.01, -0.0425, 0.4414, 0.0061, -0.0425, 0.4414, 0.0083, -0.0425, 0.4436, 0.0061, -0.0425, 0.4436, 0.0083, -0.0425, 0.4414, 0.01, -0.0439, 0.4414, 0.01, -0.0425, 0.4436, 0.01, -0.0439, 0.4436, 0.01, -0.0425, 0.4458, 0.0061, -0.0523, 0.4458, 0.0083, -0.0523, 0.4479, 0.0061, -0.0523, 0.4479, 0.0083, -0.0523, 0.4458, 0.0061, -0.0454, 0.4458, 0.0083, -0.0454, 0.4479, 0.0061, -0.0454, 0.4479, 0.0083, -0.0454, 0.4458, 0.01, -0.0523, 0.4479, 0.01, -0.0523, 0.4458, 0.01, -0.0454, 0.4479, 0.01, -0.0454, 0.4501, 0.0061, -0.0523, 0.4501, 0.0083, -0.0523, 0.452, 0.0061, -0.0523, 0.452, 0.0083, -0.0523, 0.4501, 0.0061, -0.0454,
0.4501, 0.0083, -0.0454, 0.4529, 0.0061, -0.0454, 0.4529, 0.0083, -0.0454, 0.4501, 0.01, -0.0523, 0.452, 0.01, -0.0523, 0.4501, 0.01, -0.0454, 0.4529, 0.01, -0.0454, 0.4458, 0.0061, -0.0425, 0.4458, 0.0083, -0.0425, 0.4479, 0.0061, -0.0425, 0.4479, 0.0083, -0.0425, 0.4458, 0.01, -0.0439, 0.4458, 0.01, -0.0425, 0.4479, 0.01, -0.0439, 0.4479, 0.01, -0.0425, 0.4501, 0.0061, -0.0425, 0.4501, 0.0083, -0.0425, 0.4529, 0.0061, -0.0439, 0.4529, 0.0061, -0.0425, 0.4529, 0.0083, -0.0439, 0.4529, 0.0083, -0.0425, 0.4501, 0.01, -0.0439, 0.4501, 0.01, -0.0425, 0.4529, 0.01, -0.0439, 0.4529, 0.01, -0.0425, 0.4245, 0.0061, -0.0219, 0.4245, 0.0061, -0.0201, 0.4245, 0.0083, -0.0219, 0.4245, 0.0083, -0.0201, 0.4262, 0.0061, -0.0219,
0.4262, 0.0083, -0.0219, 0.4245, 0.01, -0.0219, 0.4245, 0.01, -0.0201, 0.4262, 0.01, -0.0219, 0.4262, 0.01, -0.0201, 0.4284, 0.0061, -0.0219, 0.4284, 0.0083, -0.0219, 0.4306, 0.0061, -0.0219, 0.4306, 0.0083, -0.0219, 0.4284, 0.01, -0.0219, 0.4284, 0.01, -0.0201, 0.4306, 0.01, -0.0219, 0.4306, 0.01, -0.0201, 0.4327, 0.0061, -0.0219, 0.4327, 0.0083, -0.0219, 0.4349, 0.0061, -0.0219, 0.4349, 0.0083, -0.0219, 0.4327, 0.01, -0.0219, 0.4327, 0.01, -0.0201, 0.4349, 0.01, -0.0219, 0.4349, 0.01, -0.0201, 0.4245, 0.0061, -0.0179, 0.4245, 0.0061, -0.0157, 0.4245, 0.0083, -0.0179, 0.4245, 0.0083, -0.0157, 0.4245, 0.0061, -0.0136, 0.4245, 0.0061, -0.0114, 0.4245, 0.0083, -0.0136, 0.4245, 0.0083, -0.0114, 0.4245, 0.01, -0.0179,
0.4245, 0.01, -0.0157, 0.4262, 0.01, -0.0179, 0.4262, 0.01, -0.0157, 0.4245, 0.01, -0.0136, 0.4245, 0.01, -0.0114, 0.4262, 0.01, -0.0136, 0.4262, 0.01, -0.0114, 0.4245, 0.0061, -0.0101, 0.4245, 0.0083, -0.0101, 0.4262, 0.0061, -0.0101, 0.4262, 0.0083, -0.0101, 0.4245, 0.01, -0.0101, 0.4262, 0.01, -0.0101, 0.4284, 0.01, -0.0179, 0.4284, 0.01, -0.0157, 0.4306, 0.01, -0.0179, 0.4306, 0.01, -0.0157, 0.4284, 0.01, -0.0136, 0.4284, 0.01, -0.0114, 0.4306, 0.01, -0.0136, 0.4306, 0.01, -0.0114, 0.4327, 0.01, -0.0179, 0.4327, 0.01, -0.0157, 0.4349, 0.01, -0.0179, 0.4349, 0.01, -0.0157, 0.4327, 0.01, -0.0136, 0.4327, 0.01, -0.0114, 0.4349, 0.01, -0.0136, 0.4349, 0.01, -0.0114, 0.4284, 0.0061, -0.0101,
0.4284, 0.0083, -0.0101, 0.4306, 0.0061, -0.0101, 0.4306, 0.0083, -0.0101, 0.4284, 0.01, -0.0101, 0.4306, 0.01, -0.0101, 0.4327, 0.0061, -0.0101, 0.4327, 0.0083, -0.0101, 0.4349, 0.0061, -0.0101, 0.4349, 0.0083, -0.0101, 0.4327, 0.01, -0.0101, 0.4349, 0.01, -0.0101, 0.4371, 0.0061, -0.0219, 0.4371, 0.0083, -0.0219, 0.4392, 0.0061, -0.0219, 0.4392, 0.0083, -0.0219, 0.4371, 0.01, -0.0219, 0.4371, 0.01, -0.0201, 0.4392, 0.01, -0.0219, 0.4392, 0.01, -0.0201, 0.4414, 0.0061, -0.0219, 0.4414, 0.0083, -0.0219, 0.4436, 0.0061, -0.0219, 0.4436, 0.0083, -0.0219, 0.4414, 0.01, -0.0219, 0.4414, 0.01, -0.0201, 0.4436, 0.01, -0.0219, 0.4436, 0.01, -0.0201, 0.4458, 0.0061, -0.0219, 0.4458, 0.0083, -0.0219, 0.4479, 0.0061, -0.0219,
0.4479, 0.0083, -0.0219, 0.4458, 0.01, -0.0219, 0.4458, 0.01, -0.0201, 0.4479, 0.01, -0.0219, 0.4479, 0.01, -0.0201, 0.4501, 0.0061, -0.0219, 0.4501, 0.0083, -0.0219, 0.4523, 0.0061, -0.0219, 0.4523, 0.0083, -0.0219, 0.4501, 0.01, -0.0219, 0.4501, 0.01, -0.0201, 0.4523, 0.01, -0.0219, 0.4523, 0.01, -0.0201, 0.4371, 0.01, -0.0179, 0.4371, 0.01, -0.0157, 0.4392, 0.01, -0.0179, 0.4392, 0.01, -0.0157, 0.4371, 0.01, -0.0136, 0.4371, 0.01, -0.0114, 0.4392, 0.01, -0.0136, 0.4392, 0.01, -0.0114, 0.4414, 0.01, -0.0179, 0.4414, 0.01, -0.0157, 0.4436, 0.01, -0.0179, 0.4436, 0.01, -0.0157, 0.4414, 0.01, -0.0136, 0.4414, 0.01, -0.0114, 0.4436, 0.01, -0.0136, 0.4436, 0.01, -0.0114, 0.4371, 0.0061, -0.0101,
0.4371, 0.0083, -0.0101, 0.4392, 0.0061, -0.0101, 0.4392, 0.0083, -0.0101, 0.4371, 0.01, -0.0101, 0.4392, 0.01, -0.0101, 0.4414, 0.0061, -0.0101, 0.4414, 0.0083, -0.0101, 0.4436, 0.0061, -0.0101, 0.4436, 0.0083, -0.0101, 0.4414, 0.01, -0.0101, 0.4436, 0.01, -0.0101, 0.4458, 0.01, -0.0179, 0.4458, 0.01, -0.0157, 0.4479, 0.01, -0.0179, 0.4479, 0.01, -0.0157, 0.4458, 0.01, -0.0136, 0.4458, 0.01, -0.0114, 0.4479, 0.01, -0.0136, 0.4479, 0.01, -0.0114, 0.4501, 0.01, -0.0179, 0.4501, 0.01, -0.0157, 0.4523, 0.01, -0.0179, 0.4523, 0.01, -0.0157, 0.4501, 0.01, -0.0136, 0.4501, 0.01, -0.0114, 0.4523, 0.01, -0.0136, 0.4523, 0.01, -0.0114, 0.4458, 0.0061, -0.0101, 0.4458, 0.0083, -0.0101, 0.4479, 0.0061, -0.0101,
0.4479, 0.0083, -0.0101, 0.4458, 0.01, -0.0101, 0.4479, 0.01, -0.0101, 0.4501, 0.0061, -0.0101, 0.4501, 0.0083, -0.0101, 0.4523, 0.0061, -0.0101, 0.4523, 0.0083, -0.0101, 0.4501, 0.01, -0.0101, 0.4523, 0.01, -0.0101, 0.4725, 0.0061, -0.067, 0.4725, 0.0083, -0.067, 0.474, 0.0061, -0.067, 0.474, 0.0083, -0.067, 0.4725, 0.0061, -0.0656, 0.4725, 0.0061, -0.0635, 0.4725, 0.0083, -0.0656, 0.4725, 0.0083, -0.0635, 0.4725, 0.01, -0.067, 0.474, 0.01, -0.067, 0.4725, 0.01, -0.0656, 0.4725, 0.01, -0.0635, 0.474, 0.01, -0.0656, 0.474, 0.01, -0.0635, 0.4765, 0.0061, -0.067, 0.4765, 0.0083, -0.067, 0.4765, 0.0061, -0.0656, 0.4765, 0.0061, -0.0635, 0.4765, 0.0083, -0.0656, 0.4765, 0.0083, -0.0635, 0.4765, 0.01, -0.067,
0.4765, 0.01, -0.0656, 0.4765, 0.01, -0.0635, 0.4725, 0.0061, -0.0613, 0.4725, 0.0061, -0.0591, 0.4725, 0.0083, -0.0613, 0.4725, 0.0083, -0.0591, 0.4725, 0.0061, -0.057, 0.4725, 0.0061, -0.0548, 0.4725, 0.0083, -0.057, 0.4725, 0.0083, -0.0548, 0.4725, 0.01, -0.0613, 0.4725, 0.01, -0.0591, 0.474, 0.01, -0.0613, 0.474, 0.01, -0.0591, 0.4725, 0.01, -0.057, 0.4725, 0.01, -0.0548, 0.474, 0.01, -0.057, 0.474, 0.01, -0.0548, 0.4765, 0.0061, -0.0613, 0.4765, 0.0061, -0.0591, 0.4765, 0.0083, -0.0613, 0.4765, 0.0083, -0.0591, 0.4765, 0.0061, -0.057, 0.4765, 0.0061, -0.0548, 0.4765, 0.0083, -0.057, 0.4765, 0.0083, -0.0548, 0.4765, 0.01, -0.0613, 0.4765, 0.01, -0.0591, 0.4765, 0.01, -0.057, 0.4765, 0.01, -0.0548,
0.4804, 0.0061, -0.0591, 0.4804, 0.0083, -0.0591, 0.4826, 0.0061, -0.0591, 0.4826, 0.0083, -0.0591, 0.4804, 0.0061, -0.0562, 0.4804, 0.0083, -0.0562, 0.4826, 0.0061, -0.0562, 0.4826, 0.0083, -0.0562, 0.4804, 0.01, -0.0591, 0.4826, 0.01, -0.0591, 0.4804, 0.01, -0.0562, 0.4826, 0.01, -0.0562, 0.4848, 0.0061, -0.0592, 0.4848, 0.0083, -0.0592, 0.487, 0.0061, -0.0596, 0.487, 0.0083, -0.0596, 0.4848, 0.0061, -0.0562, 0.4848, 0.0083, -0.0562, 0.487, 0.0061, -0.0562, 0.487, 0.0083, -0.0562, 0.4848, 0.01, -0.0592, 0.4868, 0.01, -0.0596, 0.4848, 0.01, -0.0562, 0.487, 0.01, -0.0562, 0.4725, 0.0061, -0.0526, 0.4725, 0.0061, -0.0505, 0.4725, 0.0083, -0.0526, 0.4725, 0.0083, -0.0505, 0.4725, 0.0061, -0.0483, 0.4725, 0.0061, -0.0461,
0.4725, 0.0083, -0.0483, 0.4725, 0.0083, -0.0461, 0.4725, 0.01, -0.0526, 0.4725, 0.01, -0.0505, 0.474, 0.01, -0.0526, 0.474, 0.01, -0.0505, 0.4725, 0.01, -0.0483, 0.4725, 0.01, -0.0461, 0.474, 0.01, -0.0483, 0.474, 0.01, -0.0461, 0.4765, 0.0061, -0.0526, 0.4766, 0.0061, -0.0504, 0.4765, 0.0083, -0.0526, 0.4766, 0.0083, -0.0504, 0.4769, 0.0061, -0.0487, 0.4769, 0.0083, -0.0487, 0.4775, 0.0061, -0.0476, 0.478, 0.0061, -0.047, 0.4775, 0.0083, -0.0476, 0.478, 0.0083, -0.047, 0.4765, 0.01, -0.0527, 0.4766, 0.01, -0.0506, 0.477, 0.01, -0.0485, 0.4761, 0.01, -0.0461, 0.4775, 0.01, -0.0476, 0.4786, 0.01, -0.0465, 0.4725, 0.0061, -0.0439, 0.4725, 0.0061, -0.0425, 0.4725, 0.0083, -0.0439, 0.4725, 0.0083, -0.0425,
0.474, 0.0061, -0.0425, 0.474, 0.0083, -0.0425, 0.4725, 0.01, -0.0439, 0.4725, 0.01, -0.0425, 0.474, 0.01, -0.0439, 0.474, 0.01, -0.0425, 0.4761, 0.0061, -0.0425, 0.4761, 0.0083, -0.0425, 0.4783, 0.0061, -0.0425, 0.4783, 0.0083, -0.0425, 0.4761, 0.01, -0.0439, 0.4761, 0.01, -0.0425, 0.4783, 0.01, -0.0439, 0.4783, 0.01, -0.0425, 0.4801, 0.0061, -0.0456, 0.4801, 0.0083, -0.0456, 0.4826, 0.0061, -0.0455, 0.4826, 0.0083, -0.0455, 0.4801, 0.01, -0.0456, 0.4827, 0.01, -0.0455, 0.4848, 0.0061, -0.0454, 0.4848, 0.0083, -0.0454, 0.487, 0.0061, -0.0454, 0.487, 0.0083, -0.0454, 0.4848, 0.01, -0.0454, 0.487, 0.01, -0.0454, 0.4805, 0.0061, -0.0425, 0.4805, 0.0083, -0.0425, 0.4826, 0.0061, -0.0425, 0.4826, 0.0083, -0.0425,
0.4805, 0.01, -0.0439, 0.4805, 0.01, -0.0425, 0.4826, 0.01, -0.0439, 0.4826, 0.01, -0.0425, 0.4848, 0.0061, -0.0425, 0.4848, 0.0083, -0.0425, 0.487, 0.0061, -0.0425, 0.487, 0.0083, -0.0425, 0.4848, 0.01, -0.0439, 0.4848, 0.01, -0.0425, 0.487, 0.01, -0.0439, 0.487, 0.01, -0.0425, 0.4544, 0.0061, -0.0219, 0.4544, 0.0083, -0.0219, 0.4566, 0.0061, -0.0219, 0.4566, 0.0083, -0.0219, 0.4544, 0.01, -0.0219, 0.4544, 0.01, -0.0201, 0.4566, 0.01, -0.0219, 0.4566, 0.01, -0.0201, 0.4588, 0.0061, -0.0219, 0.4588, 0.0083, -0.0219, 0.4609, 0.0061, -0.0219, 0.4609, 0.0083, -0.0219, 0.4588, 0.01, -0.0219, 0.4588, 0.01, -0.0201, 0.4609, 0.01, -0.0219, 0.4609, 0.01, -0.0201, 0.4631, 0.0061, -0.0219, 0.4631, 0.0083, -0.0219,
0.4653, 0.0061, -0.0219, 0.4653, 0.0083, -0.0219, 0.4631, 0.01, -0.0219, 0.4631, 0.01, -0.0201, 0.4653, 0.01, -0.0219, 0.4653, 0.01, -0.0201, 0.4675, 0.0061, -0.0219, 0.4675, 0.0083, -0.0219, 0.4696, 0.0061, -0.0219, 0.4696, 0.0083, -0.0219, 0.4675, 0.01, -0.0219, 0.4675, 0.01, -0.0201, 0.4696, 0.01, -0.0219, 0.4696, 0.01, -0.0201, 0.4544, 0.01, -0.0179, 0.4544, 0.01, -0.0157, 0.4566, 0.01, -0.0179, 0.4566, 0.01, -0.0157, 0.4544, 0.01, -0.0136, 0.4544, 0.01, -0.0114, 0.4566, 0.01, -0.0136, 0.4566, 0.01, -0.0114, 0.4588, 0.01, -0.0179, 0.4588, 0.01, -0.0157, 0.4609, 0.01, -0.0179, 0.4609, 0.01, -0.0157, 0.4588, 0.01, -0.0136, 0.4588, 0.01, -0.0114, 0.4609, 0.01, -0.0136, 0.4609, 0.01, -0.0114,
0.4544, 0.0061, -0.0101, 0.4544, 0.0083, -0.0101, 0.4566, 0.0061, -0.0101, 0.4566, 0.0083, -0.0101, 0.4544, 0.01, -0.0101, 0.4566, 0.01, -0.0101, 0.4588, 0.0061, -0.0101, 0.4588, 0.0083, -0.0101, 0.4609, 0.0061, -0.0101, 0.4609, 0.0083, -0.0101, 0.4588, 0.01, -0.0101, 0.4609, 0.01, -0.0101, 0.4631, 0.01, -0.0179, 0.4631, 0.01, -0.0157, 0.4653, 0.01, -0.0179, 0.4653, 0.01, -0.0157, 0.4631, 0.01, -0.0136, 0.4631, 0.01, -0.0114, 0.4653, 0.01, -0.0136, 0.4653, 0.01, -0.0114, 0.4675, 0.01, -0.0179, 0.4675, 0.01, -0.0157, 0.4696, 0.01, -0.0179, 0.4696, 0.01, -0.0157, 0.4675, 0.01, -0.0136, 0.4675, 0.01, -0.0114, 0.4696, 0.01, -0.0136, 0.4696, 0.01, -0.0114, 0.4631, 0.0061, -0.0101, 0.4631, 0.0083, -0.0101,
0.4653, 0.0061, -0.0101, 0.4653, 0.0083, -0.0101, 0.4631, 0.01, -0.0101, 0.4653, 0.01, -0.0101, 0.4675, 0.0061, -0.0101, 0.4675, 0.0083, -0.0101, 0.4696, 0.0061, -0.0101, 0.4696, 0.0083, -0.0101, 0.4675, 0.01, -0.0101, 0.4696, 0.01, -0.0101, 0.4718, 0.0061, -0.0219, 0.4718, 0.0083, -0.0219, 0.474, 0.0061, -0.0219, 0.474, 0.0083, -0.0219, 0.4718, 0.01, -0.0219, 0.4718, 0.01, -0.0201, 0.474, 0.01, -0.0219, 0.474, 0.01, -0.0201, 0.4761, 0.0061, -0.0219, 0.4761, 0.0083, -0.0219, 0.4783, 0.0061, -0.0218, 0.4783, 0.0083, -0.0218, 0.4761, 0.01, -0.0219, 0.4761, 0.01, -0.0201, 0.4783, 0.01, -0.0218, 0.4783, 0.01, -0.0201, 0.4805, 0.0061, -0.0216, 0.4805, 0.0083, -0.0216, 0.4823, 0.0061, -0.0213, 0.4834, 0.0061, -0.0211,
0.4823, 0.0083, -0.0213, 0.4834, 0.0083, -0.0211, 0.4805, 0.01, -0.0216, 0.4805, 0.01, -0.0201, 0.4821, 0.01, -0.0213, 0.4829, 0.01, -0.0212, 0.4848, 0.0061, -0.0208, 0.4848, 0.0083, -0.0208, 0.487, 0.0061, -0.0203, 0.487, 0.0083, -0.0203, 0.485, 0.01, -0.0207, 0.4872, 0.01, -0.0202, 0.4718, 0.01, -0.0179, 0.4718, 0.01, -0.0157, 0.474, 0.01, -0.0179, 0.474, 0.01, -0.0157, 0.4718, 0.01, -0.0136, 0.4718, 0.01, -0.0114, 0.474, 0.01, -0.0136, 0.474, 0.01, -0.0114, 0.4761, 0.01, -0.0179, 0.4761, 0.01, -0.0157, 0.4783, 0.01, -0.0179, 0.4783, 0.01, -0.0157, 0.4761, 0.01, -0.0136, 0.4761, 0.01, -0.0114, 0.4783, 0.01, -0.0136, 0.4783, 0.01, -0.0114, 0.4718, 0.0061, -0.0101, 0.4718, 0.0083, -0.0101,
0.474, 0.0061, -0.0101, 0.474, 0.0083, -0.0101, 0.4718, 0.01, -0.0101, 0.474, 0.01, -0.0101, 0.4761, 0.0061, -0.0101, 0.4761, 0.0083, -0.0101, 0.4783, 0.0061, -0.01, 0.4783, 0.0083, -0.01, 0.4761, 0.01, -0.0101, 0.4783, 0.01, -0.01, 0.4805, 0.01, -0.0179, 0.4805, 0.01, -0.0157, 0.4826, 0.01, -0.0179, 0.4826, 0.01, -0.0157, 0.4805, 0.01, -0.0136, 0.4805, 0.01, -0.0114, 0.4826, 0.01, -0.0136, 0.4826, 0.01, -0.0114, 0.4848, 0.01, -0.0179, 0.4848, 0.01, -0.0157, 0.487, 0.01, -0.0179, 0.487, 0.01, -0.0157, 0.4848, 0.01, -0.0136, 0.4848, 0.01, -0.0114, 0.487, 0.01, -0.0136, 0.487, 0.01, -0.0114, 0.4805, 0.0061, -0.0096, 0.4805, 0.0083, -0.0096, 0.4827, 0.0061, -0.009, 0.4827, 0.0083, -0.009,
0.4804, 0.01, -0.0096, 0.4824, 0.01, -0.0091, 0.4841, 0.0061, -0.0083, 0.4852, 0.0061, -0.0078, 0.4841, 0.0083, -0.0083, 0.4852, 0.0083, -0.0078, 0.4869, 0.0061, -0.0067, 0.4869, 0.0083, -0.0067, 0.4879, 0.0061, -0.0058, 0.4879, 0.0083, -0.0058, 0.4843, 0.01, -0.0082, 0.4853, 0.01, -0.0077, 0.487, 0.01, -0.0092, 0.4869, 0.01, -0.0068, 0.4879, 0.01, -0.0058, 0.4919, 0.0061, -0.0628, 0.4919, 0.0083, -0.0628, 0.4919, 0.01, -0.0629, 0.4943, 0.0061, -0.0648, 0.4932, 0.0061, -0.0639, 0.4943, 0.0083, -0.0648, 0.4932, 0.0083, -0.0639, 0.4956, 0.0061, -0.0658, 0.4956, 0.0083, -0.0658, 0.4943, 0.01, -0.0648, 0.4932, 0.01, -0.064, 0.4954, 0.01, -0.0657, 0.4957, 0.01, -0.0635, 0.4894, 0.0061, -0.0608, 0.4882, 0.0061, -0.0601,
0.4894, 0.0083, -0.0608, 0.4882, 0.0083, -0.0601, 0.4908, 0.0061, -0.0619, 0.4908, 0.0083, -0.0619, 0.4892, 0.0061, -0.0562, 0.4892, 0.0083, -0.0562, 0.4909, 0.0061, -0.056, 0.492, 0.0061, -0.0558, 0.4909, 0.0083, -0.056, 0.492, 0.0083, -0.0558, 0.4895, 0.01, -0.0608, 0.4886, 0.01, -0.0603, 0.4909, 0.01, -0.0619, 0.4913, 0.01, -0.0591, 0.4891, 0.01, -0.0562, 0.4912, 0.01, -0.056, 0.4921, 0.01, -0.0558, 0.4963, 0.0061, -0.0608, 0.4953, 0.0061, -0.0594, 0.4963, 0.0083, -0.0608, 0.4953, 0.0083, -0.0594, 0.4935, 0.0061, -0.0553, 0.4935, 0.0083, -0.0553, 0.4953, 0.0061, -0.0543, 0.4953, 0.0083, -0.0543, 0.4935, 0.01, -0.0613, 0.4935, 0.01, -0.0591, 0.4963, 0.01, -0.0609, 0.4953, 0.01, -0.0594, 0.4935, 0.01, -0.057,
0.4934, 0.01, -0.0553, 0.4957, 0.01, -0.057, 0.4952, 0.01, -0.0543, 0.4986, 0.0061, -0.0668, 0.4986, 0.0083, -0.0668, 0.5, 0.0061, -0.0669, 0.5, 0.0083, -0.0669, 0.4975, 0.0061, -0.0665, 0.4983, 0.0061, -0.063, 0.4975, 0.0083, -0.0665, 0.4983, 0.0083, -0.063, 0.5005, 0.0061, -0.0652, 0.4994, 0.0061, -0.0641, 0.5005, 0.0083, -0.0652, 0.4994, 0.0083, -0.0641, 0.4987, 0.01, -0.0668, 0.5, 0.01, -0.0669, 0.4977, 0.01, -0.0666, 0.4983, 0.01, -0.063, 0.5005, 0.01, -0.0652, 0.4994, 0.01, -0.0641, 0.502, 0.0061, -0.067, 0.502, 0.0083, -0.067, 0.5014, 0.0061, -0.0663, 0.5014, 0.0083, -0.0663, 0.502, 0.01, -0.067, 0.5014, 0.01, -0.0664, 0.4972, 0.0061, -0.0619, 0.497, 0.0061, -0.0582, 0.4972, 0.0083, -0.0619,
0.497, 0.0083, -0.0582, 0.4981, 0.0061, -0.0574, 0.4981, 0.0083, -0.0574, 0.4994, 0.0061, -0.0563, 0.5006, 0.0061, -0.0549, 0.4994, 0.0083, -0.0563, 0.5006, 0.0083, -0.0549, 0.4972, 0.01, -0.0619, 0.497, 0.01, -0.0582, 0.4981, 0.01, -0.0574, 0.4978, 0.01, -0.0548, 0.4994, 0.01, -0.0563, 0.5006, 0.01, -0.0549, 0.5011, 0.0061, -0.0538, 0.5011, 0.0083, -0.0538, 0.5011, 0.01, -0.0537, 0.4892, 0.0061, -0.0455, 0.4892, 0.0083, -0.0455, 0.4914, 0.0061, -0.046, 0.4914, 0.0083, -0.046, 0.4891, 0.01, -0.0455, 0.4912, 0.01, -0.0459, 0.4962, 0.0061, -0.053, 0.497, 0.0061, -0.0503, 0.4962, 0.0083, -0.053, 0.497, 0.0083, -0.0503, 0.4943, 0.0061, -0.0474, 0.4932, 0.0061, -0.0468, 0.4943, 0.0083, -0.0474, 0.4932, 0.0083, -0.0468,
0.4955, 0.0061, -0.0484, 0.4955, 0.0083, -0.0484, 0.4964, 0.01, -0.0526, 0.497, 0.01, -0.0503, 0.4943, 0.01, -0.0474, 0.4933, 0.01, -0.0468, 0.4955, 0.01, -0.0484, 0.4957, 0.01, -0.0461, 0.4892, 0.0061, -0.0425, 0.4892, 0.0083, -0.0425, 0.4913, 0.0061, -0.0425, 0.4913, 0.0083, -0.0425, 0.4892, 0.01, -0.0439, 0.4892, 0.01, -0.0425, 0.4913, 0.01, -0.0439, 0.4913, 0.01, -0.0425, 0.4935, 0.0061, -0.0426, 0.4935, 0.0083, -0.0426, 0.4958, 0.0061, -0.0431, 0.4947, 0.0061, -0.0428, 0.4958, 0.0083, -0.0431, 0.4947, 0.0083, -0.0428, 0.4935, 0.01, -0.0439, 0.4935, 0.01, -0.0426, 0.4958, 0.01, -0.0431, 0.4946, 0.01, -0.0428, 0.501, 0.0061, -0.0475, 0.5, 0.0061, -0.046, 0.501, 0.0083, -0.0475, 0.5, 0.0083, -0.046,
0.4978, 0.01, -0.0526, 0.4978, 0.01, -0.0505, 0.5, 0.01, -0.0526, 0.5, 0.01, -0.0505, 0.4978, 0.01, -0.0483, 0.4978, 0.01, -0.0461, 0.5011, 0.01, -0.0477, 0.4998, 0.01, -0.0458, 0.5015, 0.0061, -0.0527, 0.5018, 0.0061, -0.0505, 0.5015, 0.0083, -0.0527, 0.5018, 0.0083, -0.0505, 0.5015, 0.0061, -0.0486, 0.5015, 0.0083, -0.0486, 0.5015, 0.01, -0.0527, 0.5018, 0.01, -0.0505, 0.5015, 0.01, -0.0488, 0.4979, 0.0061, -0.0441, 0.4979, 0.0083, -0.0441, 0.499, 0.0061, -0.045, 0.499, 0.0083, -0.045, 0.4981, 0.01, -0.0442, 0.499, 0.01, -0.045, 0.4892, 0.0061, -0.0195, 0.4892, 0.0083, -0.0195, 0.4889, 0.01, -0.0196, 0.4902, 0.0061, -0.019, 0.4902, 0.0083, -0.019, 0.4914, 0.0061, -0.0185, 0.4914, 0.0083, -0.0185,
0.4898, 0.01, -0.0192, 0.4892, 0.01, -0.0157, 0.4917, 0.01, -0.0183, 0.4913, 0.01, -0.0157, 0.4892, 0.01, -0.0136, 0.4892, 0.01, -0.0114, 0.4913, 0.01, -0.0136, 0.4913, 0.01, -0.0114, 0.4932, 0.0061, -0.0174, 0.4942, 0.0061, -0.0166, 0.4932, 0.0083, -0.0174, 0.4942, 0.0083, -0.0166, 0.4956, 0.0061, -0.0155, 0.4956, 0.0083, -0.0155, 0.4966, 0.0061, -0.0145, 0.4966, 0.0083, -0.0145, 0.4931, 0.01, -0.0174, 0.4942, 0.01, -0.0166, 0.4955, 0.01, -0.0156, 0.4935, 0.01, -0.0136, 0.4935, 0.01, -0.0114, 0.4966, 0.01, -0.0146, 0.4957, 0.01, -0.0114, 0.489, 0.0061, -0.0048, 0.49, 0.0061, -0.0034, 0.489, 0.0083, -0.0048, 0.49, 0.0083, -0.0034, 0.4905, 0.0061, -0.0023, 0.4905, 0.0083, -0.0023, 0.4892, 0.01, -0.0092,
0.4892, 0.01, -0.0071, 0.4913, 0.01, -0.0092, 0.4913, 0.01, -0.0071, 0.489, 0.01, -0.0048, 0.4899, 0.01, -0.0034, 0.4913, 0.01, -0.0049, 0.4904, 0.01, -0.0025, 0.4935, 0.01, -0.0092, 0.4935, 0.01, -0.0071, 0.4957, 0.01, -0.0092, 0.4957, 0.01, -0.0071, 0.4935, 0.01, -0.0049, 0.4935, 0.01, -0.0027, 0.4957, 0.01, -0.0049, 0.4957, 0.01, -0.0027, 0.4977, 0.0061, -0.0135, 0.4987, 0.0061, -0.0122, 0.4977, 0.0083, -0.0135, 0.4987, 0.0083, -0.0122, 0.4994, 0.0061, -0.0111, 0.4994, 0.0083, -0.0111, 0.4977, 0.01, -0.0135, 0.4988, 0.01, -0.0121, 0.4995, 0.01, -0.011, 0.5005, 0.0061, -0.0093, 0.501, 0.0061, -0.008, 0.5005, 0.0083, -0.0093, 0.501, 0.0083, -0.008, 0.4978, 0.01, -0.0092, 0.4978, 0.01, -0.0071,
0.5004, 0.01, -0.0096, 0.5012, 0.01, -0.0076, 0.4978, 0.01, -0.0049, 0.4978, 0.01, -0.0027, 0.5, 0.01, -0.0049, 0.5, 0.01, -0.0027, 0.5014, 0.0061, -0.0069, 0.5014, 0.0083, -0.0069, 0.5018, 0.0061, -0.0049, 0.5019, 0.0061, -0.0027, 0.5018, 0.0083, -0.0049, 0.5019, 0.0083, -0.0027, 0.5015, 0.01, -0.0067, 0.5018, 0.01, -0.0049, 0.5019, 0.01, -0.0027, 0.003, 0.0061, -0.0013, 0.003, 0.0083, -0.0013, 0.0052, 0.0061, -0.0012, 0.0052, 0.0083, -0.0012, 0.0031, 0.01, -0.0013, 0.003, 0.01, 0.0016, 0.0053, 0.01, -0.0012, 0.0052, 0.01, 0.0016, 0.003, 0.01, 0.0038, 0.003, 0.01, 0.006, 0.0052, 0.01, 0.0038, 0.0052, 0.01, 0.006, 0.0074, 0.0061, -0.0009, 0.0074, 0.0083, -0.0009, 0.0096, 0.0061, -0.0005,
0.0096, 0.0083, -0.0005, 0.0075, 0.01, -0.0009, 0.0074, 0.01, 0.0016, 0.0097, 0.01, -0.0004, 0.0096, 0.01, 0.0016, 0.0074, 0.01, 0.0038, 0.0074, 0.01, 0.006, 0.0096, 0.01, 0.0038, 0.0096, 0.01, 0.006, 0.003, 0.0061, 0.0105, 0.003, 0.0083, 0.0105, 0.0052, 0.0061, 0.0106, 0.0052, 0.0083, 0.0106, 0.003, 0.01, 0.0081, 0.003, 0.01, 0.0105, 0.0052, 0.01, 0.0081, 0.0051, 0.01, 0.0105, 0.0074, 0.0061, 0.0109, 0.0074, 0.0083, 0.0109, 0.0085, 0.0061, 0.0114, 0.0085, 0.0083, 0.0114, 0.0096, 0.0061, 0.0118, 0.0096, 0.0083, 0.0118, 0.0074, 0.01, 0.0081, 0.0072, 0.01, 0.0109, 0.0096, 0.01, 0.0081, 0.0089, 0.01, 0.0115, 0.0099, 0.01, 0.0119, 0.0116, 0.0061, 0.0002, 0.0127, 0.0061, 0.0006,
0.0116, 0.0083, 0.0002, 0.0127, 0.0083, 0.0006, 0.0139, 0.0061, 0.0011, 0.0139, 0.0083, 0.0011, 0.0114, 0.01, 0.0001, 0.0123, 0.01, 0.0004, 0.0142, 0.01, 0.0012, 0.0117, 0.01, 0.0038, 0.0117, 0.01, 0.006, 0.0139, 0.01, 0.0038, 0.0139, 0.01, 0.006, 0.0161, 0.0061, 0.0021, 0.0161, 0.0083, 0.0021, 0.0172, 0.0061, 0.0027, 0.0172, 0.0083, 0.0027, 0.0183, 0.0061, 0.0033, 0.0183, 0.0083, 0.0033, 0.0162, 0.01, 0.0022, 0.0172, 0.01, 0.0027, 0.0161, 0.01, 0.0038, 0.0161, 0.01, 0.006, 0.0183, 0.01, 0.0034, 0.0182, 0.01, 0.006, 0.0116, 0.0061, 0.0129, 0.0126, 0.0061, 0.0137, 0.0116, 0.0083, 0.0129, 0.0126, 0.0083, 0.0137, 0.0138, 0.0061, 0.0147, 0.0138, 0.0083, 0.0147, 0.0117, 0.01, 0.0081,
0.0117, 0.01, 0.0103, 0.0139, 0.01, 0.0081, 0.0139, 0.01, 0.0103, 0.0115, 0.01, 0.0129, 0.0126, 0.01, 0.0137, 0.0139, 0.01, 0.0125, 0.0138, 0.01, 0.0147, 0.0161, 0.01, 0.0081, 0.0161, 0.01, 0.0103, 0.0182, 0.01, 0.0081, 0.0182, 0.01, 0.0103, 0.0161, 0.01, 0.0125, 0.0161, 0.01, 0.0146, 0.0182, 0.01, 0.0125, 0.0182, 0.01, 0.0146, 0.0148, 0.0061, 0.016, 0.0148, 0.0083, 0.016, 0.0148, 0.01, 0.0159, 0.0156, 0.0061, 0.017, 0.0167, 0.0061, 0.0189, 0.0156, 0.0083, 0.017, 0.0167, 0.0083, 0.0189, 0.0171, 0.0061, 0.0203, 0.0171, 0.0083, 0.0203, 0.0174, 0.0061, 0.0214, 0.0176, 0.0061, 0.0233, 0.0174, 0.0083, 0.0214, 0.0176, 0.0083, 0.0233, 0.0156, 0.01, 0.017, 0.0167, 0.01, 0.019,
0.0182, 0.01, 0.0168, 0.0182, 0.01, 0.019, 0.0171, 0.01, 0.0203, 0.0173, 0.01, 0.0211, 0.0176, 0.01, 0.0233, 0.0144, 0.0061, 0.0324, 0.0144, 0.0083, 0.0324, 0.0144, 0.01, 0.0325, 0.017, 0.0061, 0.028, 0.017, 0.0083, 0.028, 0.0175, 0.0061, 0.0255, 0.0172, 0.0061, 0.0269, 0.0175, 0.0083, 0.0255, 0.0172, 0.0083, 0.0269, 0.0162, 0.0061, 0.0299, 0.0153, 0.0061, 0.0313, 0.0162, 0.0083, 0.0299, 0.0153, 0.0083, 0.0313, 0.0169, 0.01, 0.0282, 0.0175, 0.01, 0.0254, 0.0171, 0.01, 0.0274, 0.0163, 0.01, 0.0297, 0.0152, 0.01, 0.0315, 0.0182, 0.01, 0.0298, 0.0182, 0.01, 0.032, 0.0199, 0.0061, 0.0044, 0.021, 0.0061, 0.0053, 0.0199, 0.0083, 0.0044, 0.021, 0.0083, 0.0053, 0.0221, 0.0061, 0.0064,
0.0221, 0.0083, 0.0064, 0.0198, 0.01, 0.0044, 0.021, 0.01, 0.0053, 0.0221, 0.01, 0.0064, 0.0232, 0.0061, 0.0076, 0.0232, 0.0083, 0.0076, 0.0204, 0.01, 0.0081, 0.0204, 0.01, 0.0103, 0.0232, 0.01, 0.0076, 0.0226, 0.01, 0.0103, 0.0204, 0.01, 0.0125, 0.0204, 0.01, 0.0146, 0.0226, 0.01, 0.0125, 0.0226, 0.01, 0.0146, 0.0241, 0.0061, 0.0087, 0.0252, 0.0061, 0.0102, 0.0241, 0.0083, 0.0087, 0.0252, 0.0083, 0.0102, 0.0259, 0.0061, 0.0113, 0.0259, 0.0083, 0.0113, 0.0265, 0.0061, 0.0124, 0.0274, 0.0061, 0.0146, 0.0265, 0.0083, 0.0124, 0.0274, 0.0083, 0.0146, 0.0241, 0.01, 0.0087, 0.0252, 0.01, 0.0102, 0.0259, 0.01, 0.0113, 0.0247, 0.01, 0.0125, 0.0247, 0.01, 0.0146, 0.0265, 0.01, 0.0123,
0.0273, 0.01, 0.0144, 0.0204, 0.01, 0.0168, 0.0204, 0.01, 0.019, 0.0226, 0.01, 0.0168, 0.0226, 0.01, 0.019, 0.0204, 0.01, 0.0212, 0.0204, 0.01, 0.0233, 0.0226, 0.01, 0.0212, 0.0226, 0.01, 0.0233, 0.0279, 0.0061, 0.0163, 0.0279, 0.0083, 0.0163, 0.0247, 0.01, 0.0168, 0.0247, 0.01, 0.019, 0.028, 0.01, 0.0166, 0.0269, 0.01, 0.019, 0.0247, 0.01, 0.0212, 0.0247, 0.01, 0.0233, 0.0269, 0.01, 0.0212, 0.0269, 0.01, 0.0233, 0.0204, 0.01, 0.0255, 0.0204, 0.01, 0.0277, 0.0226, 0.01, 0.0255, 0.0226, 0.01, 0.0277, 0.0204, 0.01, 0.0298, 0.0204, 0.01, 0.032, 0.0226, 0.01, 0.0298, 0.0226, 0.01, 0.032, 0.028, 0.0061, 0.0308, 0.0277, 0.0061, 0.032, 0.028, 0.0083, 0.0308,
0.0277, 0.0083, 0.032, 0.0247, 0.01, 0.0255, 0.0247, 0.01, 0.0277, 0.0269, 0.01, 0.0255, 0.0269, 0.01, 0.0277, 0.0247, 0.01, 0.0298, 0.0247, 0.01, 0.032, 0.0281, 0.01, 0.0302, 0.0276, 0.01, 0.0323, 0.0281, 0.0061, 0.0174, 0.0283, 0.0061, 0.019, 0.0281, 0.0083, 0.0174, 0.0283, 0.0083, 0.019, 0.0284, 0.0061, 0.0212, 0.0284, 0.0061, 0.0233, 0.0284, 0.0083, 0.0212, 0.0284, 0.0083, 0.0233, 0.0281, 0.01, 0.0175, 0.0283, 0.01, 0.019, 0.0284, 0.01, 0.0212, 0.0284, 0.01, 0.0233, 0.0284, 0.0061, 0.0255, 0.0284, 0.0061, 0.0277, 0.0284, 0.0083, 0.0255, 0.0284, 0.0083, 0.0277, 0.0282, 0.0061, 0.0297, 0.0282, 0.0083, 0.0297, 0.0284, 0.01, 0.0255, 0.0284, 0.01, 0.0277, 0.0282, 0.01, 0.0294,
0.006, 0.0061, 0.0374, 0.006, 0.0083, 0.0374, 0.003, 0.0061, 0.0378, 0.003, 0.0083, 0.0378, 0.0049, 0.0061, 0.0376, 0.0049, 0.0083, 0.0376, 0.0061, 0.01, 0.0373, 0.003, 0.01, 0.0378, 0.003, 0.01, 0.0407, 0.0052, 0.01, 0.0376, 0.0052, 0.01, 0.0407, 0.0074, 0.0061, 0.037, 0.0074, 0.0083, 0.037, 0.0096, 0.0061, 0.0361, 0.0096, 0.0083, 0.0361, 0.0073, 0.01, 0.037, 0.0093, 0.01, 0.0362, 0.0074, 0.01, 0.0385, 0.0074, 0.01, 0.0407, 0.0096, 0.01, 0.0385, 0.0096, 0.01, 0.0407, 0.003, 0.0061, 0.0487, 0.003, 0.0083, 0.0487, 0.0052, 0.0061, 0.0487, 0.0052, 0.0083, 0.0487, 0.003, 0.01, 0.0429, 0.003, 0.01, 0.045, 0.0052, 0.01, 0.0429, 0.0052, 0.01, 0.045, 0.003, 0.01, 0.0472,
0.003, 0.01, 0.0487, 0.0052, 0.01, 0.0472, 0.0052, 0.01, 0.0487, 0.0084, 0.0061, 0.0483, 0.0073, 0.0061, 0.0485, 0.0084, 0.0083, 0.0483, 0.0073, 0.0083, 0.0485, 0.0096, 0.0061, 0.048, 0.0096, 0.0083, 0.048, 0.0074, 0.01, 0.0429, 0.0074, 0.01, 0.045, 0.0096, 0.01, 0.0429, 0.0096, 0.01, 0.045, 0.0078, 0.01, 0.0484, 0.007, 0.01, 0.0485, 0.0098, 0.01, 0.048, 0.012, 0.0061, 0.0347, 0.0108, 0.0061, 0.0354, 0.012, 0.0083, 0.0347, 0.0108, 0.0083, 0.0354, 0.0133, 0.0061, 0.0335, 0.0133, 0.0083, 0.0335, 0.012, 0.01, 0.0346, 0.011, 0.01, 0.0352, 0.0133, 0.01, 0.0335, 0.0139, 0.01, 0.0363, 0.0117, 0.01, 0.0385, 0.0117, 0.01, 0.0407, 0.0139, 0.01, 0.0385, 0.0139, 0.01, 0.0407,
0.0161, 0.01, 0.0342, 0.0161, 0.01, 0.0363, 0.0182, 0.01, 0.0342, 0.0182, 0.01, 0.0363, 0.0161, 0.01, 0.0385, 0.0161, 0.01, 0.0407, 0.0182, 0.01, 0.0385, 0.0182, 0.01, 0.0407, 0.0149, 0.0061, 0.0461, 0.0149, 0.0083, 0.0461, 0.0117, 0.0061, 0.0474, 0.0117, 0.0083, 0.0474, 0.0139, 0.0061, 0.0466, 0.0139, 0.0083, 0.0466, 0.0117, 0.01, 0.0429, 0.0117, 0.01, 0.045, 0.0139, 0.01, 0.0429, 0.0146, 0.01, 0.0463, 0.0119, 0.01, 0.0473, 0.0136, 0.01, 0.0467, 0.0161, 0.0061, 0.0455, 0.0161, 0.0083, 0.0455, 0.019, 0.0061, 0.0437, 0.0179, 0.0061, 0.0445, 0.019, 0.0083, 0.0437, 0.0179, 0.0083, 0.0445, 0.0161, 0.01, 0.0429, 0.0164, 0.01, 0.0454, 0.0189, 0.01, 0.0438, 0.0178, 0.01, 0.0445,
0.0214, 0.0061, 0.0417, 0.0214, 0.0083, 0.0417, 0.0235, 0.0061, 0.0394, 0.0225, 0.0061, 0.0407, 0.0235, 0.0083, 0.0394, 0.0225, 0.0083, 0.0407, 0.0204, 0.01, 0.0342, 0.0204, 0.01, 0.0363, 0.0226, 0.01, 0.0342, 0.0226, 0.01, 0.0363, 0.0204, 0.01, 0.0385, 0.0214, 0.01, 0.0417, 0.0236, 0.01, 0.0394, 0.0225, 0.01, 0.0406, 0.0254, 0.0061, 0.0368, 0.0254, 0.0083, 0.0368, 0.0269, 0.0061, 0.0342, 0.0261, 0.0061, 0.0357, 0.0269, 0.0083, 0.0342, 0.0261, 0.0083, 0.0357, 0.0244, 0.0061, 0.0384, 0.0244, 0.0083, 0.0384, 0.0247, 0.01, 0.0342, 0.0255, 0.01, 0.0367, 0.0268, 0.01, 0.0344, 0.0261, 0.01, 0.0356, 0.0244, 0.01, 0.0383, 0.0204, 0.0061, 0.0427, 0.0204, 0.0083, 0.0427, 0.0203, 0.01, 0.0427,
0.0716, 0.0061, -0.0005, 0.0716, 0.0061, 0.0016, 0.0716, 0.0083, -0.0005, 0.0716, 0.0083, 0.0016, 0.0716, 0.0061, 0.0038, 0.0716, 0.0061, 0.006, 0.0716, 0.0083, 0.0038, 0.0716, 0.0083, 0.006, 0.0716, 0.01, -0.0005, 0.0716, 0.01, 0.0016, 0.0747, 0.01, -0.0005, 0.0747, 0.01, 0.0016, 0.0716, 0.01, 0.0038, 0.0716, 0.01, 0.006, 0.0747, 0.01, 0.0038, 0.0747, 0.01, 0.006, 0.0768, 0.01, -0.0005, 0.0768, 0.01, 0.0016, 0.079, 0.01, -0.0005, 0.079, 0.01, 0.0016, 0.0768, 0.01, 0.0038, 0.0768, 0.01, 0.006, 0.079, 0.01, 0.0038, 0.079, 0.01, 0.006, 0.0716, 0.0061, 0.0081, 0.0716, 0.0061, 0.0103, 0.0716, 0.0083, 0.0081, 0.0716, 0.0083, 0.0103, 0.0716, 0.0061, 0.0125, 0.0716, 0.0061, 0.0146,
0.0716, 0.0083, 0.0125, 0.0716, 0.0083, 0.0146, 0.0716, 0.01, 0.0081, 0.0716, 0.01, 0.0103, 0.0747, 0.01, 0.0081, 0.0747, 0.01, 0.0103, 0.0716, 0.01, 0.0125, 0.0716, 0.01, 0.0146, 0.0747, 0.01, 0.0125, 0.0747, 0.01, 0.0146, 0.0768, 0.01, 0.0081, 0.0768, 0.01, 0.0103, 0.079, 0.01, 0.0081, 0.079, 0.01, 0.0103, 0.0768, 0.01, 0.0125, 0.0768, 0.01, 0.0146, 0.079, 0.01, 0.0125, 0.079, 0.01, 0.0146, 0.0833, 0.0061, -0.0005, 0.0833, 0.0061, 0.0016, 0.0833, 0.0083, -0.0005, 0.0833, 0.0083, 0.0016, 0.0833, 0.0061, 0.0038, 0.0833, 0.0061, 0.006, 0.0833, 0.0083, 0.0038, 0.0833, 0.0083, 0.006, 0.0812, 0.01, -0.0005, 0.0812, 0.01, 0.0016, 0.0833, 0.01, -0.0005, 0.0833, 0.01, 0.0016,
0.0812, 0.01, 0.0038, 0.0812, 0.01, 0.006, 0.0833, 0.01, 0.0038, 0.0833, 0.01, 0.006, 0.0833, 0.0061, 0.0085, 0.0833, 0.0083, 0.0085, 0.0812, 0.01, 0.0081, 0.0812, 0.01, 0.0103, 0.0833, 0.01, 0.0085, 0.0833, 0.01, 0.0103, 0.0812, 0.01, 0.0125, 0.0812, 0.01, 0.0146, 0.0833, 0.01, 0.0125, 0.0833, 0.01, 0.0146, 0.0855, 0.0061, 0.0085, 0.0855, 0.0083, 0.0085, 0.0877, 0.0061, 0.0085, 0.0877, 0.0083, 0.0085, 0.0855, 0.01, 0.0085, 0.0855, 0.01, 0.0103, 0.0877, 0.01, 0.0085, 0.0877, 0.01, 0.0103, 0.0855, 0.01, 0.0125, 0.0855, 0.01, 0.0146, 0.0877, 0.01, 0.0125, 0.0877, 0.01, 0.0146, 0.0716, 0.0061, 0.0168, 0.0716, 0.0061, 0.019, 0.0716, 0.0083, 0.0168, 0.0716, 0.0083, 0.019,
0.0716, 0.0061, 0.0212, 0.0716, 0.0061, 0.0233, 0.0716, 0.0083, 0.0212, 0.0716, 0.0083, 0.0233, 0.0716, 0.01, 0.0168, 0.0716, 0.01, 0.019, 0.0747, 0.01, 0.0168, 0.0747, 0.01, 0.019, 0.0716, 0.01, 0.0212, 0.0716, 0.01, 0.0233, 0.0747, 0.01, 0.0212, 0.0747, 0.01, 0.0233, 0.0768, 0.01, 0.0168, 0.0768, 0.01, 0.019, 0.079, 0.01, 0.0168, 0.079, 0.01, 0.019, 0.0768, 0.01, 0.0212, 0.0768, 0.01, 0.0233, 0.079, 0.01, 0.0212, 0.079, 0.01, 0.0233, 0.0716, 0.0061, 0.0255, 0.0716, 0.0061, 0.0277, 0.0716, 0.0083, 0.0255, 0.0716, 0.0083, 0.0277, 0.0716, 0.0061, 0.0298, 0.0716, 0.0061, 0.032, 0.0716, 0.0083, 0.0298, 0.0716, 0.0083, 0.032, 0.0716, 0.01, 0.0255, 0.0716, 0.01, 0.0277,
0.0747, 0.01, 0.0255, 0.0747, 0.01, 0.0277, 0.0716, 0.01, 0.0298, 0.0716, 0.01, 0.032, 0.0747, 0.01, 0.0298, 0.0747, 0.01, 0.032, 0.0768, 0.01, 0.0255, 0.0768, 0.01, 0.0277, 0.079, 0.01, 0.0255, 0.079, 0.01, 0.0277, 0.0768, 0.01, 0.0298, 0.0768, 0.01, 0.032, 0.079, 0.01, 0.0298, 0.079, 0.01, 0.032, 0.0833, 0.0061, 0.0203, 0.0833, 0.0061, 0.0233, 0.0833, 0.0083, 0.0203, 0.0833, 0.0083, 0.0233, 0.0812, 0.01, 0.0168, 0.0812, 0.01, 0.019, 0.0833, 0.01, 0.0168, 0.0833, 0.01, 0.019, 0.0812, 0.01, 0.0212, 0.0812, 0.01, 0.0233, 0.0833, 0.01, 0.0203, 0.0833, 0.01, 0.0233, 0.0855, 0.0061, 0.0203, 0.0855, 0.0083, 0.0203, 0.0877, 0.0061, 0.0203, 0.0877, 0.0083, 0.0203,
0.0855, 0.01, 0.0168, 0.0855, 0.01, 0.019, 0.0877, 0.01, 0.0168, 0.0877, 0.01, 0.019, 0.0855, 0.01, 0.0203, 0.0877, 0.01, 0.0203, 0.0833, 0.0061, 0.0255, 0.0833, 0.0061, 0.0277, 0.0833, 0.0083, 0.0255, 0.0833, 0.0083, 0.0277, 0.0833, 0.0061, 0.0298, 0.0833, 0.0061, 0.032, 0.0833, 0.0083, 0.0298, 0.0833, 0.0083, 0.032, 0.0812, 0.01, 0.0255, 0.0812, 0.01, 0.0277, 0.0833, 0.01, 0.0255, 0.0833, 0.01, 0.0277, 0.0812, 0.01, 0.0298, 0.0812, 0.01, 0.032, 0.0833, 0.01, 0.0298, 0.0833, 0.01, 0.032, 0.0899, 0.0061, 0.0085, 0.0899, 0.0083, 0.0085, 0.092, 0.0061, 0.0085, 0.092, 0.0083, 0.0085, 0.0899, 0.01, 0.0085, 0.0899, 0.01, 0.0103, 0.092, 0.01, 0.0085, 0.092, 0.01, 0.0103,
0.0899, 0.01, 0.0125, 0.0899, 0.01, 0.0146, 0.092, 0.01, 0.0125, 0.092, 0.01, 0.0146, 0.0942, 0.0061, 0.0085, 0.0942, 0.0083, 0.0085, 0.0964, 0.0061, 0.0085, 0.0964, 0.0083, 0.0085, 0.0942, 0.01, 0.0085, 0.0942, 0.01, 0.0103, 0.0964, 0.01, 0.0085, 0.0964, 0.01, 0.0103, 0.0942, 0.01, 0.0125, 0.0942, 0.01, 0.0146, 0.0964, 0.01, 0.0125, 0.0964, 0.01, 0.0146, 0.0985, 0.0061, 0.0085, 0.0985, 0.0083, 0.0085, 0.1007, 0.0061, 0.0085, 0.1007, 0.0083, 0.0085, 0.0985, 0.01, 0.0085, 0.0985, 0.01, 0.0103, 0.1007, 0.01, 0.0085, 0.1007, 0.01, 0.0103, 0.0985, 0.01, 0.0125, 0.0985, 0.01, 0.0146, 0.1007, 0.01, 0.0125, 0.1007, 0.01, 0.0146, 0.1029, 0.0061, 0.0085, 0.1029, 0.0083, 0.0085,
0.105, 0.0061, 0.0085, 0.105, 0.0083, 0.0085, 0.1029, 0.01, 0.0085, 0.1029, 0.01, 0.0103, 0.105, 0.01, 0.0085, 0.105, 0.01, 0.0103, 0.1029, 0.01, 0.0125, 0.1029, 0.01, 0.0146, 0.105, 0.01, 0.0125, 0.105, 0.01, 0.0146, 0.0899, 0.0061, 0.0203, 0.0899, 0.0083, 0.0203, 0.092, 0.0061, 0.0203, 0.092, 0.0083, 0.0203, 0.0899, 0.01, 0.0168, 0.0899, 0.01, 0.019, 0.092, 0.01, 0.0168, 0.092, 0.01, 0.019, 0.0899, 0.01, 0.0203, 0.092, 0.01, 0.0203, 0.0942, 0.0061, 0.0203, 0.0942, 0.0083, 0.0203, 0.0964, 0.0061, 0.0203, 0.0964, 0.0083, 0.0203, 0.0942, 0.01, 0.0168, 0.0942, 0.01, 0.019, 0.0964, 0.01, 0.0168, 0.0964, 0.01, 0.019, 0.0942, 0.01, 0.0203, 0.0964, 0.01, 0.0203,
0.0985, 0.0061, 0.0203, 0.0985, 0.0083, 0.0203, 0.1007, 0.0061, 0.0203, 0.1007, 0.0083, 0.0203, 0.0985, 0.01, 0.0168, 0.0985, 0.01, 0.019, 0.1007, 0.01, 0.0168, 0.1007, 0.01, 0.019, 0.0985, 0.01, 0.0203, 0.1007, 0.01, 0.0203, 0.1029, 0.0061, 0.0203, 0.1029, 0.0083, 0.0203, 0.105, 0.0061, 0.0203, 0.105, 0.0083, 0.0203, 0.1029, 0.01, 0.0168, 0.1029, 0.01, 0.019, 0.105, 0.01, 0.0168, 0.105, 0.01, 0.019, 0.1029, 0.01, 0.0203, 0.105, 0.01, 0.0203, 0.0716, 0.0061, 0.0342, 0.0716, 0.0061, 0.0363, 0.0716, 0.0083, 0.0342, 0.0716, 0.0083, 0.0363, 0.0716, 0.0061, 0.0385, 0.0716, 0.0061, 0.0407, 0.0716, 0.0083, 0.0385, 0.0716, 0.0083, 0.0407, 0.0716, 0.01, 0.0342, 0.0716, 0.01, 0.0363,
0.0747, 0.01, 0.0342, 0.0747, 0.01, 0.0363, 0.0716, 0.01, 0.0385, 0.0716, 0.01, 0.0407, 0.0747, 0.01, 0.0385, 0.0747, 0.01, 0.0407, 0.0768, 0.01, 0.0342, 0.0768, 0.01, 0.0363, 0.079, 0.01, 0.0342, 0.079, 0.01, 0.0363, 0.0768, 0.01, 0.0385, 0.0768, 0.01, 0.0407, 0.079, 0.01, 0.0385, 0.079, 0.01, 0.0407, 0.0716, 0.0061, 0.0429, 0.0716, 0.0061, 0.045, 0.0716, 0.0083, 0.0429, 0.0716, 0.0083, 0.045, 0.0716, 0.0061, 0.0472, 0.0716, 0.0061, 0.0487, 0.0716, 0.0083, 0.0472, 0.0716, 0.0083, 0.0487, 0.0747, 0.0061, 0.0487, 0.0747, 0.0083, 0.0487, 0.0716, 0.01, 0.0429, 0.0716, 0.01, 0.045, 0.0747, 0.01, 0.0429, 0.0747, 0.01, 0.045, 0.0716, 0.01, 0.0472, 0.0716, 0.01, 0.0487,
0.0747, 0.01, 0.0472, 0.0747, 0.01, 0.0487, 0.0768, 0.0061, 0.0487, 0.0768, 0.0083, 0.0487, 0.079, 0.0061, 0.0487, 0.079, 0.0083, 0.0487, 0.0768, 0.01, 0.0429, 0.0768, 0.01, 0.045, 0.079, 0.01, 0.0429, 0.079, 0.01, 0.045, 0.0768, 0.01, 0.0472, 0.0768, 0.01, 0.0487, 0.079, 0.01, 0.0472, 0.079, 0.01, 0.0487, 0.0833, 0.0061, 0.0342, 0.0833, 0.0061, 0.0363, 0.0833, 0.0083, 0.0342, 0.0833, 0.0083, 0.0363, 0.0833, 0.0061, 0.0379, 0.0833, 0.0083, 0.0379, 0.0812, 0.01, 0.0342, 0.0812, 0.01, 0.0363, 0.0833, 0.01, 0.0342, 0.0833, 0.01, 0.0363, 0.0812, 0.01, 0.0385, 0.0812, 0.01, 0.0407, 0.0833, 0.01, 0.0379, 0.0833, 0.01, 0.0407, 0.0855, 0.0061, 0.0379, 0.0855, 0.0083, 0.0379,
0.0877, 0.0061, 0.0379, 0.0877, 0.0083, 0.0379, 0.0855, 0.01, 0.0379, 0.0855, 0.01, 0.0407, 0.0877, 0.01, 0.0379, 0.0877, 0.01, 0.0407, 0.0812, 0.0061, 0.0487, 0.0812, 0.0083, 0.0487, 0.0833, 0.0061, 0.0487, 0.0833, 0.0083, 0.0487, 0.0812, 0.01, 0.0429, 0.0812, 0.01, 0.045, 0.0833, 0.01, 0.0429, 0.0833, 0.01, 0.045, 0.0812, 0.01, 0.0472, 0.0812, 0.01, 0.0487, 0.0833, 0.01, 0.0472, 0.0833, 0.01, 0.0487, 0.0855, 0.0061, 0.0487, 0.0855, 0.0083, 0.0487, 0.0877, 0.0061, 0.0487, 0.0877, 0.0083, 0.0487, 0.0855, 0.01, 0.0429, 0.0855, 0.01, 0.045, 0.0877, 0.01, 0.0429, 0.0877, 0.01, 0.045, 0.0855, 0.01, 0.0472, 0.0855, 0.01, 0.0487, 0.0877, 0.01, 0.0472, 0.0877, 0.01, 0.0487,
0.0899, 0.0061, 0.0379, 0.0899, 0.0083, 0.0379, 0.092, 0.0061, 0.0379, 0.092, 0.0083, 0.0379, 0.0899, 0.01, 0.0379, 0.0899, 0.01, 0.0407, 0.092, 0.01, 0.0379, 0.092, 0.01, 0.0407, 0.0942, 0.0061, 0.0379, 0.0942, 0.0083, 0.0379, 0.0964, 0.0061, 0.0379, 0.0964, 0.0083, 0.0379, 0.0942, 0.01, 0.0379, 0.0942, 0.01, 0.0407, 0.0964, 0.01, 0.0379, 0.0964, 0.01, 0.0407, 0.0899, 0.0061, 0.0487, 0.0899, 0.0083, 0.0487, 0.092, 0.0061, 0.0487, 0.092, 0.0083, 0.0487, 0.0899, 0.01, 0.0429, 0.0899, 0.01, 0.045, 0.092, 0.01, 0.0429, 0.092, 0.01, 0.045, 0.0899, 0.01, 0.0472, 0.0899, 0.01, 0.0487, 0.092, 0.01, 0.0472, 0.092, 0.01, 0.0487, 0.0942, 0.0061, 0.0487, 0.0942, 0.0083, 0.0487,
0.0964, 0.0061, 0.0487, 0.0964, 0.0083, 0.0487, 0.0942, 0.01, 0.0429, 0.0942, 0.01, 0.045, 0.0964, 0.01, 0.0429, 0.0964, 0.01, 0.045, 0.0942, 0.01, 0.0472, 0.0942, 0.01, 0.0487, 0.0964, 0.01, 0.0472, 0.0964, 0.01, 0.0487, 0.0985, 0.0061, 0.0379, 0.0985, 0.0083, 0.0379, 0.1007, 0.0061, 0.0379, 0.1007, 0.0083, 0.0379, 0.0985, 0.01, 0.0379, 0.0985, 0.01, 0.0407, 0.1007, 0.01, 0.0379, 0.1007, 0.01, 0.0407, 0.1029, 0.0061, 0.0379, 0.1029, 0.0083, 0.0379, 0.105, 0.0061, 0.0379, 0.105, 0.0083, 0.0379, 0.1029, 0.01, 0.0379, 0.1029, 0.01, 0.0407, 0.105, 0.01, 0.0379, 0.105, 0.01, 0.0407, 0.0985, 0.0061, 0.0487, 0.0985, 0.0083, 0.0487, 0.1007, 0.0061, 0.0487, 0.1007, 0.0083, 0.0487,
0.0985, 0.01, 0.0429, 0.0985, 0.01, 0.045, 0.1007, 0.01, 0.0429, 0.1007, 0.01, 0.045, 0.0985, 0.01, 0.0472, 0.0985, 0.01, 0.0487, 0.1007, 0.01, 0.0472, 0.1007, 0.01, 0.0487, 0.1029, 0.0061, 0.0487, 0.1029, 0.0083, 0.0487, 0.105, 0.0061, 0.0487, 0.105, 0.0083, 0.0487, 0.1029, 0.01, 0.0429, 0.1029, 0.01, 0.045, 0.105, 0.01, 0.0429, 0.105, 0.01, 0.045, 0.1029, 0.01, 0.0472, 0.1029, 0.01, 0.0487, 0.105, 0.01, 0.0472, 0.105, 0.01, 0.0487, 0.1072, 0.0061, 0.0085, 0.1072, 0.0083, 0.0085, 0.1094, 0.0061, 0.0085, 0.1094, 0.0083, 0.0085, 0.1072, 0.01, 0.0085, 0.1072, 0.01, 0.0103, 0.1094, 0.01, 0.0085, 0.1094, 0.01, 0.0103, 0.1072, 0.01, 0.0125, 0.1072, 0.01, 0.0146,
0.1094, 0.01, 0.0125, 0.1094, 0.01, 0.0146, 0.1116, 0.0061, 0.0085, 0.1116, 0.0083, 0.0085, 0.1137, 0.0061, 0.0085, 0.1137, 0.0083, 0.0085, 0.1116, 0.01, 0.0085, 0.1116, 0.01, 0.0103, 0.1137, 0.01, 0.0085, 0.1137, 0.01, 0.0103, 0.1116, 0.01, 0.0125, 0.1116, 0.01, 0.0146, 0.1137, 0.01, 0.0125, 0.1137, 0.01, 0.0146, 0.1159, 0.0061, 0.0085, 0.1159, 0.0083, 0.0085, 0.1181, 0.0061, 0.0085, 0.1181, 0.0083, 0.0085, 0.1159, 0.01, 0.0085, 0.1159, 0.01, 0.0103, 0.1181, 0.01, 0.0085, 0.1181, 0.01, 0.0103, 0.1159, 0.01, 0.0125, 0.1159, 0.01, 0.0146, 0.1181, 0.01, 0.0125, 0.1181, 0.01, 0.0146, 0.1202, 0.0061, 0.0085, 0.1202, 0.0083, 0.0085, 0.1224, 0.0061, 0.0085, 0.1224, 0.0083, 0.0085,
0.1202, 0.01, 0.0085, 0.1202, 0.01, 0.0103, 0.1224, 0.01, 0.0085, 0.1224, 0.01, 0.0103, 0.1202, 0.01, 0.0125, 0.1202, 0.01, 0.0146, 0.1224, 0.01, 0.0125, 0.1224, 0.01, 0.0146, 0.1072, 0.0061, 0.0203, 0.1072, 0.0083, 0.0203, 0.1094, 0.0061, 0.0203, 0.1094, 0.0083, 0.0203, 0.1072, 0.01, 0.0168, 0.1072, 0.01, 0.019, 0.1094, 0.01, 0.0168, 0.1094, 0.01, 0.019, 0.1072, 0.01, 0.0203, 0.1094, 0.01, 0.0203, 0.1116, 0.0061, 0.0203, 0.1116, 0.0083, 0.0203, 0.1137, 0.0061, 0.0203, 0.1137, 0.0083, 0.0203, 0.1116, 0.01, 0.0168, 0.1116, 0.01, 0.019, 0.1137, 0.01, 0.0168, 0.1137, 0.01, 0.019, 0.1116, 0.01, 0.0203, 0.1137, 0.01, 0.0203, 0.1159, 0.0061, 0.0203, 0.1159, 0.0083, 0.0203,
0.1181, 0.0061, 0.0203, 0.1181, 0.0083, 0.0203, 0.1159, 0.01, 0.0168, 0.1159, 0.01, 0.019, 0.1181, 0.01, 0.0168, 0.1181, 0.01, 0.019, 0.1159, 0.01, 0.0203, 0.1181, 0.01, 0.0203, 0.1202, 0.0061, 0.0203, 0.1202, 0.0083, 0.0203, 0.1224, 0.0061, 0.0203, 0.1224, 0.0083, 0.0203, 0.1202, 0.01, 0.0168, 0.1202, 0.01, 0.019, 0.1224, 0.01, 0.0168, 0.1224, 0.01, 0.019, 0.1202, 0.01, 0.0203, 0.1224, 0.01, 0.0203, 0.1246, 0.0061, 0.0085, 0.1246, 0.0083, 0.0085, 0.1267, 0.0061, 0.0085, 0.1267, 0.0083, 0.0085, 0.1246, 0.01, 0.0085, 0.1246, 0.01, 0.0103, 0.1267, 0.01, 0.0085, 0.1267, 0.01, 0.0103, 0.1246, 0.01, 0.0125, 0.1246, 0.01, 0.0146, 0.1267, 0.01, 0.0125, 0.1267, 0.01, 0.0146,
0.1294, 0.0061, 0.0085, 0.1294, 0.0061, 0.0103, 0.1294, 0.0083, 0.0085, 0.1294, 0.0083, 0.0103, 0.1294, 0.0061, 0.0125, 0.1294, 0.0061, 0.0146, 0.1294, 0.0083, 0.0125, 0.1294, 0.0083, 0.0146, 0.1294, 0.01, 0.0085, 0.1294, 0.01, 0.0103, 0.1294, 0.01, 0.0125, 0.1294, 0.01, 0.0146, 0.1246, 0.0061, 0.0203, 0.1246, 0.0083, 0.0203, 0.1267, 0.0061, 0.0203, 0.1267, 0.0083, 0.0203, 0.1246, 0.01, 0.0168, 0.1246, 0.01, 0.019, 0.1267, 0.01, 0.0168, 0.1267, 0.01, 0.019, 0.1246, 0.01, 0.0203, 0.1267, 0.01, 0.0203, 0.1294, 0.0061, 0.0168, 0.1294, 0.0061, 0.019, 0.1294, 0.0083, 0.0168, 0.1294, 0.0083, 0.019, 0.1294, 0.0061, 0.0203, 0.1294, 0.0083, 0.0203, 0.1294, 0.01, 0.0168, 0.1294, 0.01, 0.019,
0.1294, 0.01, 0.0203, 0.1072, 0.0061, 0.0379, 0.1072, 0.0083, 0.0379, 0.1094, 0.0061, 0.0379, 0.1094, 0.0083, 0.0379, 0.1072, 0.01, 0.0379, 0.1072, 0.01, 0.0407, 0.1094, 0.01, 0.0379, 0.1094, 0.01, 0.0407, 0.1116, 0.0061, 0.0379, 0.1116, 0.0083, 0.0379, 0.1137, 0.0061, 0.0379, 0.1137, 0.0083, 0.0379, 0.1116, 0.01, 0.0379, 0.1116, 0.01, 0.0407, 0.1137, 0.01, 0.0379, 0.1137, 0.01, 0.0407, 0.1072, 0.0061, 0.0487, 0.1072, 0.0083, 0.0487, 0.1094, 0.0061, 0.0487, 0.1094, 0.0083, 0.0487, 0.1072, 0.01, 0.0429, 0.1072, 0.01, 0.045, 0.1094, 0.01, 0.0429, 0.1094, 0.01, 0.045, 0.1072, 0.01, 0.0472, 0.1072, 0.01, 0.0487, 0.1094, 0.01, 0.0472, 0.1094, 0.01, 0.0487, 0.1116, 0.0061, 0.0487,
0.1116, 0.0083, 0.0487, 0.1137, 0.0061, 0.0487, 0.1137, 0.0083, 0.0487, 0.1116, 0.01, 0.0429, 0.1116, 0.01, 0.045, 0.1137, 0.01, 0.0429, 0.1137, 0.01, 0.045, 0.1116, 0.01, 0.0472, 0.1116, 0.01, 0.0487, 0.1137, 0.01, 0.0472, 0.1137, 0.01, 0.0487, 0.1159, 0.0061, 0.0379, 0.1159, 0.0083, 0.0379, 0.1181, 0.0061, 0.0379, 0.1181, 0.0083, 0.0379, 0.1159, 0.01, 0.0379, 0.1159, 0.01, 0.0407, 0.1181, 0.01, 0.0379, 0.1181, 0.01, 0.0407, 0.1202, 0.0061, 0.0379, 0.1202, 0.0083, 0.0379, 0.1224, 0.0061, 0.0379, 0.1224, 0.0083, 0.0379, 0.1202, 0.01, 0.0379, 0.1202, 0.01, 0.0407, 0.1224, 0.01, 0.0379, 0.1224, 0.01, 0.0407, 0.1159, 0.0061, 0.0487, 0.1159, 0.0083, 0.0487, 0.1181, 0.0061, 0.0487,
0.1181, 0.0083, 0.0487, 0.1159, 0.01, 0.0429, 0.1159, 0.01, 0.045, 0.1181, 0.01, 0.0429, 0.1181, 0.01, 0.045, 0.1159, 0.01, 0.0472, 0.1159, 0.01, 0.0487, 0.1181, 0.01, 0.0472, 0.1181, 0.01, 0.0487, 0.1202, 0.0061, 0.0487, 0.1202, 0.0083, 0.0487, 0.1224, 0.0061, 0.0487, 0.1224, 0.0083, 0.0487, 0.1202, 0.01, 0.0429, 0.1202, 0.01, 0.045, 0.1224, 0.01, 0.0429, 0.1224, 0.01, 0.045, 0.1202, 0.01, 0.0472, 0.1202, 0.01, 0.0487, 0.1224, 0.01, 0.0472, 0.1224, 0.01, 0.0487, 0.1246, 0.0061, 0.0379, 0.1246, 0.0083, 0.0379, 0.1267, 0.0061, 0.0379, 0.1267, 0.0083, 0.0379, 0.1246, 0.01, 0.0379, 0.1246, 0.01, 0.0407, 0.1267, 0.01, 0.0379, 0.1267, 0.01, 0.0407, 0.1289, 0.0061, 0.0379,
0.1289, 0.0083, 0.0379, 0.1311, 0.0061, 0.0379, 0.1311, 0.0083, 0.0379, 0.1289, 0.01, 0.0379, 0.1289, 0.01, 0.0407, 0.1311, 0.01, 0.0379, 0.1311, 0.01, 0.0407, 0.1246, 0.0061, 0.0487, 0.1246, 0.0083, 0.0487, 0.1267, 0.0061, 0.0487, 0.1267, 0.0083, 0.0487, 0.1246, 0.01, 0.0429, 0.1246, 0.01, 0.045, 0.1267, 0.01, 0.0429, 0.1267, 0.01, 0.045, 0.1246, 0.01, 0.0472, 0.1246, 0.01, 0.0487, 0.1267, 0.01, 0.0472, 0.1267, 0.01, 0.0487, 0.1289, 0.0061, 0.0487, 0.1289, 0.0083, 0.0487, 0.1311, 0.0061, 0.0487, 0.1311, 0.0083, 0.0487, 0.1289, 0.01, 0.0429, 0.1289, 0.01, 0.045, 0.1311, 0.01, 0.0429, 0.1311, 0.01, 0.045, 0.1289, 0.01, 0.0472, 0.1289, 0.01, 0.0487, 0.1311, 0.01, 0.0472,
0.1311, 0.01, 0.0487, 0.1324, 0.0061, 0.0379, 0.1324, 0.0061, 0.0407, 0.1324, 0.0083, 0.0379, 0.1324, 0.0083, 0.0407, 0.1324, 0.01, 0.0379, 0.1324, 0.01, 0.0407, 0.1324, 0.0061, 0.0429, 0.1324, 0.0061, 0.045, 0.1324, 0.0083, 0.0429, 0.1324, 0.0083, 0.045, 0.1324, 0.0061, 0.0472, 0.1324, 0.0061, 0.0487, 0.1324, 0.0083, 0.0472, 0.1324, 0.0083, 0.0487, 0.1324, 0.01, 0.0429, 0.1324, 0.01, 0.045, 0.1324, 0.01, 0.0472, 0.1324, 0.01, 0.0487, 0.1412, 0.0061, -0.0005, 0.1412, 0.0061, 0.0016, 0.1412, 0.0083, -0.0005, 0.1412, 0.0083, 0.0016, 0.1412, 0.0061, 0.0038, 0.1412, 0.0061, 0.006, 0.1412, 0.0083, 0.0038, 0.1412, 0.0083, 0.006, 0.1412, 0.01, -0.0005, 0.1412, 0.01, 0.0016, 0.1441, 0.01, -0.0005,
0.1441, 0.01, 0.0016, 0.1412, 0.01, 0.0038, 0.1412, 0.01, 0.006, 0.1441, 0.01, 0.0038, 0.1441, 0.01, 0.006, 0.1463, 0.01, -0.0005, 0.1463, 0.01, 0.0016, 0.1484, 0.01, -0.0005, 0.1484, 0.01, 0.0016, 0.1463, 0.01, 0.0038, 0.1463, 0.01, 0.006, 0.1484, 0.01, 0.0038, 0.1484, 0.01, 0.006, 0.1412, 0.0061, 0.0081, 0.1412, 0.0061, 0.0103, 0.1412, 0.0083, 0.0081, 0.1412, 0.0083, 0.0103, 0.1412, 0.0061, 0.0125, 0.1412, 0.0061, 0.0146, 0.1412, 0.0083, 0.0125, 0.1412, 0.0083, 0.0146, 0.1412, 0.01, 0.0081, 0.1412, 0.01, 0.0103, 0.1441, 0.01, 0.0081, 0.1441, 0.01, 0.0103, 0.1412, 0.01, 0.0125, 0.1412, 0.01, 0.0146, 0.1441, 0.01, 0.0125, 0.1441, 0.01, 0.0146, 0.1463, 0.01, 0.0081,
0.1463, 0.01, 0.0103, 0.1484, 0.01, 0.0081, 0.1484, 0.01, 0.0103, 0.1463, 0.01, 0.0125, 0.1463, 0.01, 0.0146, 0.1484, 0.01, 0.0125, 0.1484, 0.01, 0.0146, 0.152, 0.0061, -0.0005, 0.152, 0.0061, 0.0016, 0.152, 0.0083, -0.0005, 0.152, 0.0083, 0.0016, 0.152, 0.0061, 0.0038, 0.152, 0.0061, 0.006, 0.152, 0.0083, 0.0038, 0.152, 0.0083, 0.006, 0.1506, 0.01, -0.0005, 0.1506, 0.01, 0.0016, 0.152, 0.01, -0.0005, 0.152, 0.01, 0.0016, 0.1506, 0.01, 0.0038, 0.1506, 0.01, 0.006, 0.152, 0.01, 0.0038, 0.152, 0.01, 0.006, 0.152, 0.0061, 0.0081, 0.152, 0.0061, 0.0103, 0.152, 0.0083, 0.0081, 0.152, 0.0083, 0.0103, 0.152, 0.0061, 0.0125, 0.152, 0.0061, 0.0146, 0.152, 0.0083, 0.0125,
0.152, 0.0083, 0.0146, 0.1506, 0.01, 0.0081, 0.1506, 0.01, 0.0103, 0.152, 0.01, 0.0081, 0.152, 0.01, 0.0103, 0.1506, 0.01, 0.0125, 0.1506, 0.01, 0.0146, 0.152, 0.01, 0.0125, 0.152, 0.01, 0.0146, 0.1412, 0.0061, 0.0168, 0.1412, 0.0061, 0.019, 0.1412, 0.0083, 0.0168, 0.1412, 0.0083, 0.019, 0.1412, 0.0061, 0.0212, 0.1412, 0.0061, 0.0233, 0.1412, 0.0083, 0.0212, 0.1412, 0.0083, 0.0233, 0.1412, 0.01, 0.0168, 0.1412, 0.01, 0.019, 0.1441, 0.01, 0.0168, 0.1441, 0.01, 0.019, 0.1412, 0.01, 0.0212, 0.1412, 0.01, 0.0233, 0.1441, 0.01, 0.0212, 0.1441, 0.01, 0.0233, 0.1463, 0.01, 0.0168, 0.1463, 0.01, 0.019, 0.1484, 0.01, 0.0168, 0.1484, 0.01, 0.019, 0.1463, 0.01, 0.0212,
0.1463, 0.01, 0.0233, 0.1484, 0.01, 0.0212, 0.1484, 0.01, 0.0233, 0.1412, 0.0061, 0.0255, 0.1412, 0.0061, 0.0277, 0.1412, 0.0083, 0.0255, 0.1412, 0.0083, 0.0277, 0.1412, 0.0061, 0.0298, 0.1412, 0.0061, 0.032, 0.1412, 0.0083, 0.0298, 0.1412, 0.0083, 0.032, 0.1412, 0.01, 0.0255, 0.1412, 0.01, 0.0277, 0.1441, 0.01, 0.0255, 0.1441, 0.01, 0.0277, 0.1412, 0.01, 0.0298, 0.1412, 0.01, 0.032, 0.1441, 0.01, 0.0298, 0.1441, 0.01, 0.032, 0.1463, 0.01, 0.0255, 0.1463, 0.01, 0.0277, 0.1484, 0.01, 0.0255, 0.1484, 0.01, 0.0277, 0.1463, 0.01, 0.0298, 0.1463, 0.01, 0.032, 0.1484, 0.01, 0.0298, 0.1484, 0.01, 0.032, 0.152, 0.0061, 0.0168, 0.152, 0.0061, 0.019, 0.152, 0.0083, 0.0168,
0.152, 0.0083, 0.019, 0.152, 0.0061, 0.0212, 0.152, 0.0061, 0.0233, 0.152, 0.0083, 0.0212, 0.152, 0.0083, 0.0233, 0.1506, 0.01, 0.0168, 0.1506, 0.01, 0.019, 0.152, 0.01, 0.0168, 0.152, 0.01, 0.019, 0.1506, 0.01, 0.0212, 0.1506, 0.01, 0.0233, 0.152, 0.01, 0.0212, 0.152, 0.01, 0.0233, 0.152, 0.0061, 0.0255, 0.152, 0.0061, 0.0277, 0.152, 0.0083, 0.0255, 0.152, 0.0083, 0.0277, 0.152, 0.0061, 0.0298, 0.152, 0.0061, 0.032, 0.152, 0.0083, 0.0298, 0.152, 0.0083, 0.032, 0.1506, 0.01, 0.0255, 0.1506, 0.01, 0.0277, 0.152, 0.01, 0.0255, 0.152, 0.01, 0.0277, 0.1506, 0.01, 0.0298, 0.1506, 0.01, 0.032, 0.152, 0.01, 0.0298, 0.152, 0.01, 0.032, 0.1412, 0.0061, 0.0342,
0.1412, 0.0061, 0.0363, 0.1412, 0.0083, 0.0342, 0.1412, 0.0083, 0.0363, 0.1412, 0.0061, 0.0385, 0.1412, 0.0061, 0.0407, 0.1412, 0.0083, 0.0385, 0.1412, 0.0083, 0.0407, 0.1412, 0.01, 0.0342, 0.1412, 0.01, 0.0363, 0.1441, 0.01, 0.0342, 0.1441, 0.01, 0.0363, 0.1412, 0.01, 0.0385, 0.1412, 0.01, 0.0407, 0.1441, 0.01, 0.0385, 0.1441, 0.01, 0.0407, 0.1463, 0.01, 0.0342, 0.1463, 0.01, 0.0363, 0.1484, 0.01, 0.0342, 0.1484, 0.01, 0.0363, 0.1463, 0.01, 0.0385, 0.1463, 0.01, 0.0407, 0.1484, 0.01, 0.0385, 0.1484, 0.01, 0.0407, 0.1412, 0.0061, 0.0429, 0.1412, 0.0061, 0.045, 0.1412, 0.0083, 0.0429, 0.1412, 0.0083, 0.045, 0.1412, 0.0061, 0.0472, 0.1412, 0.0061, 0.0487, 0.1412, 0.0083, 0.0472,
0.1412, 0.0083, 0.0487, 0.1441, 0.0061, 0.0487, 0.1441, 0.0083, 0.0487, 0.1412, 0.01, 0.0429, 0.1412, 0.01, 0.045, 0.1441, 0.01, 0.0429, 0.1441, 0.01, 0.045, 0.1412, 0.01, 0.0472, 0.1412, 0.01, 0.0487, 0.1441, 0.01, 0.0472, 0.1441, 0.01, 0.0487, 0.1463, 0.0061, 0.0487, 0.1463, 0.0083, 0.0487, 0.1484, 0.0061, 0.0487, 0.1484, 0.0083, 0.0487, 0.1463, 0.01, 0.0429, 0.1463, 0.01, 0.045, 0.1484, 0.01, 0.0429, 0.1484, 0.01, 0.045, 0.1463, 0.01, 0.0472, 0.1463, 0.01, 0.0487, 0.1484, 0.01, 0.0472, 0.1484, 0.01, 0.0487, 0.152, 0.0061, 0.0342, 0.152, 0.0061, 0.0363, 0.152, 0.0083, 0.0342, 0.152, 0.0083, 0.0363, 0.152, 0.0061, 0.0385, 0.152, 0.0061, 0.0407, 0.152, 0.0083, 0.0385,
0.152, 0.0083, 0.0407, 0.1506, 0.01, 0.0342, 0.1506, 0.01, 0.0363, 0.152, 0.01, 0.0342, 0.152, 0.01, 0.0363, 0.1506, 0.01, 0.0385, 0.1506, 0.01, 0.0407, 0.152, 0.01, 0.0385, 0.152, 0.01, 0.0407, 0.152, 0.0061, 0.0429, 0.152, 0.0061, 0.045, 0.152, 0.0083, 0.0429, 0.152, 0.0083, 0.045, 0.1506, 0.0061, 0.0487, 0.1506, 0.0083, 0.0487, 0.152, 0.0061, 0.0472, 0.152, 0.0061, 0.0487, 0.152, 0.0083, 0.0472, 0.152, 0.0083, 0.0487, 0.1506, 0.01, 0.0429, 0.1506, 0.01, 0.045, 0.152, 0.01, 0.0429, 0.152, 0.01, 0.045, 0.1506, 0.01, 0.0472, 0.1506, 0.01, 0.0487, 0.152, 0.01, 0.0472, 0.152, 0.01, 0.0487, 0.2139, 0.0061, -0.0006, 0.2136, 0.0061, 0.0016, 0.2139, 0.0083, -0.0006,
0.2136, 0.0083, 0.0016, 0.2133, 0.0061, 0.0038, 0.213, 0.0061, 0.006, 0.2133, 0.0083, 0.0038, 0.213, 0.0083, 0.006, 0.2139, 0.01, -0.0006, 0.2136, 0.01, 0.0015, 0.2133, 0.01, 0.0037, 0.213, 0.01, 0.0059, 0.2157, 0.01, -0.0005, 0.2157, 0.01, 0.0016, 0.2179, 0.01, -0.0005, 0.2179, 0.01, 0.0016, 0.2157, 0.01, 0.0038, 0.2157, 0.01, 0.006, 0.2179, 0.01, 0.0038, 0.2179, 0.01, 0.006, 0.2128, 0.0061, 0.0081, 0.2127, 0.0061, 0.0103, 0.2128, 0.0083, 0.0081, 0.2127, 0.0083, 0.0103, 0.2127, 0.0061, 0.0125, 0.2127, 0.0061, 0.0146, 0.2127, 0.0083, 0.0125, 0.2127, 0.0083, 0.0146, 0.2128, 0.01, 0.0081, 0.2127, 0.01, 0.0103, 0.2127, 0.01, 0.0125, 0.2127, 0.01, 0.0146, 0.2157, 0.01, 0.0081,
0.2157, 0.01, 0.0103, 0.2179, 0.01, 0.0081, 0.2179, 0.01, 0.0103, 0.2157, 0.01, 0.0125, 0.2157, 0.01, 0.0146, 0.2179, 0.01, 0.0125, 0.2179, 0.01, 0.0146, 0.2201, 0.01, -0.0005, 0.2201, 0.01, 0.0016, 0.2222, 0.01, -0.0005, 0.2222, 0.01, 0.0016, 0.2201, 0.01, 0.0038, 0.2201, 0.01, 0.006, 0.2222, 0.01, 0.0038, 0.2222, 0.01, 0.006, 0.2271, 0.0061, -0.0002, 0.226, 0.0061, 0.0016, 0.2271, 0.0083, -0.0002, 0.226, 0.0083, 0.0016, 0.2252, 0.0061, 0.0039, 0.2248, 0.0061, 0.006, 0.2252, 0.0083, 0.0039, 0.2248, 0.0083, 0.006, 0.2255, 0.0061, 0.0029, 0.2255, 0.0083, 0.0029, 0.2244, 0.01, -0.0005, 0.2244, 0.01, 0.0016, 0.2272, 0.01, -0.0002, 0.2259, 0.01, 0.0017, 0.2252, 0.01, 0.0038,
0.2247, 0.01, 0.0061, 0.2256, 0.01, 0.0028, 0.2201, 0.01, 0.0081, 0.2201, 0.01, 0.0103, 0.2222, 0.01, 0.0081, 0.2222, 0.01, 0.0103, 0.2201, 0.01, 0.0125, 0.2201, 0.01, 0.0146, 0.2222, 0.01, 0.0125, 0.2222, 0.01, 0.0146, 0.2245, 0.0061, 0.0081, 0.2245, 0.0061, 0.0103, 0.2245, 0.0083, 0.0081, 0.2245, 0.0083, 0.0103, 0.2245, 0.0061, 0.0125, 0.2245, 0.0061, 0.0146, 0.2245, 0.0083, 0.0125, 0.2245, 0.0083, 0.0146, 0.2245, 0.01, 0.0082, 0.2245, 0.01, 0.0103, 0.2245, 0.01, 0.0125, 0.2245, 0.01, 0.0146, 0.2127, 0.0061, 0.0168, 0.2127, 0.0061, 0.019, 0.2127, 0.0083, 0.0168, 0.2127, 0.0083, 0.019, 0.2127, 0.0061, 0.0212, 0.2127, 0.0061, 0.0233, 0.2127, 0.0083, 0.0212, 0.2127, 0.0083, 0.0233,
0.2127, 0.01, 0.0168, 0.2127, 0.01, 0.019, 0.2127, 0.01, 0.0212, 0.2127, 0.01, 0.0233, 0.2157, 0.01, 0.0168, 0.2157, 0.01, 0.019, 0.2179, 0.01, 0.0168, 0.2179, 0.01, 0.019, 0.2157, 0.01, 0.0212, 0.2157, 0.01, 0.0233, 0.2179, 0.01, 0.0212, 0.2179, 0.01, 0.0233, 0.2127, 0.0061, 0.0255, 0.2127, 0.0061, 0.0277, 0.2127, 0.0083, 0.0255, 0.2127, 0.0083, 0.0277, 0.2127, 0.0061, 0.0298, 0.2127, 0.0061, 0.032, 0.2127, 0.0083, 0.0298, 0.2127, 0.0083, 0.032, 0.2127, 0.01, 0.0255, 0.2127, 0.01, 0.0277, 0.2127, 0.01, 0.0298, 0.2127, 0.01, 0.032, 0.2157, 0.01, 0.0255, 0.2157, 0.01, 0.0277, 0.2179, 0.01, 0.0255, 0.2179, 0.01, 0.0277, 0.2157, 0.01, 0.0298, 0.2157, 0.01, 0.032,
0.2179, 0.01, 0.0298, 0.2179, 0.01, 0.032, 0.2201, 0.01, 0.0168, 0.2201, 0.01, 0.019, 0.2222, 0.01, 0.0168, 0.2222, 0.01, 0.019, 0.2201, 0.01, 0.0212, 0.2201, 0.01, 0.0233, 0.2222, 0.01, 0.0212, 0.2222, 0.01, 0.0233, 0.2245, 0.0061, 0.0168, 0.2245, 0.0061, 0.019, 0.2245, 0.0083, 0.0168, 0.2245, 0.0083, 0.019, 0.2245, 0.0061, 0.0212, 0.2245, 0.0061, 0.0233, 0.2245, 0.0083, 0.0212, 0.2245, 0.0083, 0.0233, 0.2245, 0.01, 0.0168, 0.2245, 0.01, 0.019, 0.2245, 0.01, 0.0212, 0.2245, 0.01, 0.0233, 0.2201, 0.01, 0.0255, 0.2201, 0.01, 0.0277, 0.2222, 0.01, 0.0255, 0.2222, 0.01, 0.0277, 0.2201, 0.01, 0.0298, 0.2201, 0.01, 0.032, 0.2222, 0.01, 0.0298, 0.2222, 0.01, 0.032,
0.2245, 0.0061, 0.0255, 0.2245, 0.0061, 0.0277, 0.2245, 0.0083, 0.0255, 0.2245, 0.0083, 0.0277, 0.2245, 0.0061, 0.0298, 0.2245, 0.0061, 0.032, 0.2245, 0.0083, 0.0298, 0.2245, 0.0083, 0.032, 0.2245, 0.01, 0.0255, 0.2245, 0.01, 0.0277, 0.2245, 0.01, 0.0298, 0.2245, 0.01, 0.032, 0.2279, 0.0061, -0.0013, 0.2279, 0.0083, -0.0013, 0.2279, 0.01, -0.0014, 0.2127, 0.0061, 0.0342, 0.2127, 0.0061, 0.0363, 0.2127, 0.0083, 0.0342, 0.2127, 0.0083, 0.0363, 0.2127, 0.0061, 0.0385, 0.2127, 0.0061, 0.0407, 0.2127, 0.0083, 0.0385, 0.2127, 0.0083, 0.0407, 0.2127, 0.01, 0.0342, 0.2127, 0.01, 0.0363, 0.2127, 0.01, 0.0385, 0.2127, 0.01, 0.0407, 0.2157, 0.01, 0.0342, 0.2157, 0.01, 0.0363, 0.2179, 0.01, 0.0342,
0.2179, 0.01, 0.0363, 0.2157, 0.01, 0.0385, 0.2157, 0.01, 0.0407, 0.2179, 0.01, 0.0385, 0.2179, 0.01, 0.0407, 0.2127, 0.0061, 0.0429, 0.2127, 0.0061, 0.045, 0.2127, 0.0083, 0.0429, 0.2127, 0.0083, 0.045, 0.2127, 0.0061, 0.0472, 0.2127, 0.0061, 0.0487, 0.2127, 0.0083, 0.0472, 0.2127, 0.0083, 0.0487, 0.2127, 0.01, 0.0429, 0.2127, 0.01, 0.045, 0.2127, 0.01, 0.0472, 0.2127, 0.01, 0.0487, 0.2157, 0.0061, 0.0487, 0.2157, 0.0083, 0.0487, 0.2179, 0.0061, 0.0487, 0.2179, 0.0083, 0.0487, 0.2157, 0.01, 0.0429, 0.2157, 0.01, 0.045, 0.2179, 0.01, 0.0429, 0.2179, 0.01, 0.045, 0.2157, 0.01, 0.0472, 0.2157, 0.01, 0.0487, 0.2179, 0.01, 0.0472, 0.2179, 0.01, 0.0487, 0.2201, 0.01, 0.0342,
0.2201, 0.01, 0.0363, 0.2222, 0.01, 0.0342, 0.2222, 0.01, 0.0363, 0.2201, 0.01, 0.0385, 0.2201, 0.01, 0.0407, 0.2222, 0.01, 0.0385, 0.2222, 0.01, 0.0407, 0.2245, 0.0061, 0.0342, 0.2245, 0.0061, 0.0363, 0.2245, 0.0083, 0.0342, 0.2245, 0.0083, 0.0363, 0.2245, 0.0061, 0.0385, 0.2245, 0.0061, 0.0407, 0.2245, 0.0083, 0.0385, 0.2245, 0.0083, 0.0407, 0.2245, 0.01, 0.0342, 0.2245, 0.01, 0.0363, 0.2245, 0.01, 0.0385, 0.2245, 0.01, 0.0407, 0.2201, 0.0061, 0.0487, 0.2201, 0.0083, 0.0487, 0.2222, 0.0061, 0.0487, 0.2222, 0.0083, 0.0487, 0.2201, 0.01, 0.0429, 0.2201, 0.01, 0.045, 0.2222, 0.01, 0.0429, 0.2222, 0.01, 0.045, 0.2201, 0.01, 0.0472, 0.2201, 0.01, 0.0487, 0.2222, 0.01, 0.0472,
0.2222, 0.01, 0.0487, 0.2245, 0.0061, 0.0429, 0.2245, 0.0061, 0.045, 0.2245, 0.0083, 0.0429, 0.2245, 0.0083, 0.045, 0.2245, 0.0061, 0.0472, 0.2245, 0.0061, 0.0487, 0.2245, 0.0083, 0.0472, 0.2245, 0.0083, 0.0487, 0.2245, 0.01, 0.0429, 0.2245, 0.01, 0.045, 0.2245, 0.01, 0.0472, 0.2245, 0.01, 0.0487, 0.2763, 0.0061, -0.0006, 0.2773, 0.0061, 0.0012, 0.2763, 0.0083, -0.0006, 0.2773, 0.0083, 0.0012, 0.2777, 0.0061, 0.0023, 0.2777, 0.0083, 0.0023, 0.2782, 0.0061, 0.0038, 0.2787, 0.0061, 0.006, 0.2782, 0.0083, 0.0038, 0.2787, 0.0083, 0.006, 0.2765, 0.01, -0.0003, 0.2772, 0.01, 0.0011, 0.2787, 0.01, -0.0005, 0.2776, 0.01, 0.002, 0.2783, 0.01, 0.004, 0.2787, 0.01, 0.0061, 0.2791, 0.0061, 0.0081,
0.2793, 0.0061, 0.0103, 0.2791, 0.0083, 0.0081, 0.2793, 0.0083, 0.0103, 0.2794, 0.0061, 0.0125, 0.2794, 0.0061, 0.0146, 0.2794, 0.0083, 0.0125, 0.2794, 0.0083, 0.0146, 0.2791, 0.01, 0.0082, 0.2793, 0.01, 0.0103, 0.2794, 0.01, 0.0125, 0.2794, 0.01, 0.0146, 0.2794, 0.0061, 0.0168, 0.2794, 0.0061, 0.019, 0.2794, 0.0083, 0.0168, 0.2794, 0.0083, 0.019, 0.2794, 0.0061, 0.0212, 0.2794, 0.0061, 0.0233, 0.2794, 0.0083, 0.0212, 0.2794, 0.0083, 0.0233, 0.2794, 0.01, 0.0168, 0.2794, 0.01, 0.019, 0.2794, 0.01, 0.0212, 0.2794, 0.01, 0.0233, 0.2794, 0.0061, 0.0255, 0.2794, 0.0061, 0.0277, 0.2794, 0.0083, 0.0255, 0.2794, 0.0083, 0.0277, 0.2794, 0.0061, 0.0298, 0.2794, 0.0061, 0.032, 0.2794, 0.0083, 0.0298,
0.2794, 0.0083, 0.032, 0.2794, 0.01, 0.0255, 0.2794, 0.01, 0.0277, 0.2794, 0.01, 0.0298, 0.2794, 0.01, 0.032, 0.2794, 0.0061, 0.0342, 0.2794, 0.0061, 0.0363, 0.2794, 0.0083, 0.0342, 0.2794, 0.0083, 0.0363, 0.2794, 0.0061, 0.0385, 0.2794, 0.0061, 0.0407, 0.2794, 0.0083, 0.0385, 0.2794, 0.0083, 0.0407, 0.2794, 0.01, 0.0342, 0.2794, 0.01, 0.0363, 0.2794, 0.01, 0.0385, 0.2794, 0.01, 0.0407, 0.2794, 0.0061, 0.0429, 0.2794, 0.0061, 0.045, 0.2794, 0.0083, 0.0429, 0.2794, 0.0083, 0.045, 0.2794, 0.0061, 0.0472, 0.2794, 0.0061, 0.0487, 0.2794, 0.0083, 0.0472, 0.2794, 0.0083, 0.0487, 0.2794, 0.01, 0.0429, 0.2794, 0.01, 0.045, 0.2794, 0.01, 0.0472, 0.2794, 0.01, 0.0487, 0.2808, 0.01, -0.0005,
0.2808, 0.01, 0.0016, 0.283, 0.01, -0.0005, 0.283, 0.01, 0.0016, 0.2808, 0.01, 0.0038, 0.2808, 0.01, 0.006, 0.283, 0.01, 0.0038, 0.283, 0.01, 0.006, 0.2852, 0.01, -0.0005, 0.2852, 0.01, 0.0016, 0.2873, 0.01, -0.0005, 0.2873, 0.01, 0.0016, 0.2852, 0.01, 0.0038, 0.2852, 0.01, 0.006, 0.2873, 0.01, 0.0038, 0.2873, 0.01, 0.006, 0.2808, 0.01, 0.0081, 0.2808, 0.01, 0.0103, 0.283, 0.01, 0.0081, 0.283, 0.01, 0.0103, 0.2808, 0.01, 0.0125, 0.2808, 0.01, 0.0146, 0.283, 0.01, 0.0125, 0.283, 0.01, 0.0146, 0.2852, 0.01, 0.0081, 0.2852, 0.01, 0.0103, 0.2873, 0.01, 0.0081, 0.2873, 0.01, 0.0103, 0.2852, 0.01, 0.0125, 0.2852, 0.01, 0.0146, 0.2873, 0.01, 0.0125,
0.2873, 0.01, 0.0146, 0.2897, 0.0061, -0.0006, 0.2901, 0.0061, 0.0016, 0.2897, 0.0083, -0.0006, 0.2901, 0.0083, 0.0016, 0.2903, 0.0061, 0.0038, 0.2906, 0.0061, 0.0051, 0.2903, 0.0083, 0.0038, 0.2906, 0.0083, 0.0051, 0.2907, 0.0061, 0.0062, 0.2907, 0.0083, 0.0062, 0.2897, 0.01, -0.0007, 0.2901, 0.01, 0.0015, 0.2903, 0.01, 0.0037, 0.2906, 0.01, 0.0057, 0.2908, 0.01, 0.0065, 0.291, 0.0061, 0.0081, 0.2911, 0.0061, 0.0103, 0.291, 0.0083, 0.0081, 0.2911, 0.0083, 0.0103, 0.2912, 0.0061, 0.0125, 0.2912, 0.0061, 0.0146, 0.2912, 0.0083, 0.0125, 0.2912, 0.0083, 0.0146, 0.2895, 0.01, 0.0081, 0.2895, 0.01, 0.0103, 0.291, 0.01, 0.0081, 0.2911, 0.01, 0.0103, 0.2895, 0.01, 0.0125, 0.2895, 0.01, 0.0146,
0.2912, 0.01, 0.0125, 0.2912, 0.01, 0.0146, 0.2808, 0.01, 0.0168, 0.2808, 0.01, 0.019, 0.283, 0.01, 0.0168, 0.283, 0.01, 0.019, 0.2808, 0.01, 0.0212, 0.2808, 0.01, 0.0233, 0.283, 0.01, 0.0212, 0.283, 0.01, 0.0233, 0.2852, 0.01, 0.0168, 0.2852, 0.01, 0.019, 0.2873, 0.01, 0.0168, 0.2873, 0.01, 0.019, 0.2852, 0.01, 0.0212, 0.2852, 0.01, 0.0233, 0.2873, 0.01, 0.0212, 0.2873, 0.01, 0.0233, 0.2808, 0.01, 0.0255, 0.2808, 0.01, 0.0277, 0.283, 0.01, 0.0255, 0.283, 0.01, 0.0277, 0.2808, 0.01, 0.0298, 0.2808, 0.01, 0.032, 0.283, 0.01, 0.0298, 0.283, 0.01, 0.032, 0.2852, 0.01, 0.0255, 0.2852, 0.01, 0.0277, 0.2873, 0.01, 0.0255, 0.2873, 0.01, 0.0277,
0.2852, 0.01, 0.0298, 0.2852, 0.01, 0.032, 0.2873, 0.01, 0.0298, 0.2873, 0.01, 0.032, 0.2912, 0.0061, 0.0168, 0.2912, 0.0061, 0.019, 0.2912, 0.0083, 0.0168, 0.2912, 0.0083, 0.019, 0.2912, 0.0061, 0.0212, 0.2912, 0.0061, 0.0233, 0.2912, 0.0083, 0.0212, 0.2912, 0.0083, 0.0233, 0.2895, 0.01, 0.0168, 0.2895, 0.01, 0.019, 0.2912, 0.01, 0.0168, 0.2912, 0.01, 0.019, 0.2895, 0.01, 0.0212, 0.2895, 0.01, 0.0233, 0.2912, 0.01, 0.0212, 0.2912, 0.01, 0.0233, 0.2912, 0.0061, 0.0255, 0.2912, 0.0061, 0.0277, 0.2912, 0.0083, 0.0255, 0.2912, 0.0083, 0.0277, 0.2912, 0.0061, 0.0298, 0.2912, 0.0061, 0.032, 0.2912, 0.0083, 0.0298, 0.2912, 0.0083, 0.032, 0.2895, 0.01, 0.0255, 0.2895, 0.01, 0.0277,
0.2912, 0.01, 0.0255, 0.2912, 0.01, 0.0277, 0.2895, 0.01, 0.0298, 0.2895, 0.01, 0.032, 0.2912, 0.01, 0.0298, 0.2912, 0.01, 0.032, 0.3088, 0.0061, -0.0005, 0.3088, 0.0061, 0.0016, 0.3088, 0.0083, -0.0005, 0.3088, 0.0083, 0.0016, 0.3088, 0.0061, 0.0038, 0.3088, 0.0061, 0.006, 0.3088, 0.0083, 0.0038, 0.3088, 0.0083, 0.006, 0.3088, 0.01, -0.0005, 0.3088, 0.01, 0.0016, 0.3088, 0.01, 0.0038, 0.3088, 0.01, 0.006, 0.3112, 0.01, -0.0005, 0.3112, 0.01, 0.0016, 0.3134, 0.01, -0.0005, 0.3134, 0.01, 0.0016, 0.3112, 0.01, 0.0038, 0.3112, 0.01, 0.006, 0.3134, 0.01, 0.0038, 0.3134, 0.01, 0.006, 0.3088, 0.0061, 0.0081, 0.3088, 0.0061, 0.0103, 0.3088, 0.0083, 0.0081, 0.3088, 0.0083, 0.0103,
0.3088, 0.0061, 0.0125, 0.3088, 0.0061, 0.0146, 0.3088, 0.0083, 0.0125, 0.3088, 0.0083, 0.0146, 0.3088, 0.01, 0.0081, 0.3088, 0.01, 0.0103, 0.3088, 0.01, 0.0125, 0.3088, 0.01, 0.0146, 0.3112, 0.01, 0.0081, 0.3112, 0.01, 0.0103, 0.3134, 0.01, 0.0081, 0.3134, 0.01, 0.0103, 0.3112, 0.01, 0.0125, 0.3112, 0.01, 0.0146, 0.3134, 0.01, 0.0125, 0.3134, 0.01, 0.0146, 0.3088, 0.0061, 0.0168, 0.3088, 0.0061, 0.019, 0.3088, 0.0083, 0.0168, 0.3088, 0.0083, 0.019, 0.3088, 0.0061, 0.0212, 0.3088, 0.0061, 0.0233, 0.3088, 0.0083, 0.0212, 0.3088, 0.0083, 0.0233, 0.3088, 0.01, 0.0168, 0.3088, 0.01, 0.019, 0.3088, 0.01, 0.0212, 0.3088, 0.01, 0.0233, 0.3112, 0.01, 0.0168, 0.3112, 0.01, 0.019,
0.3134, 0.01, 0.0168, 0.3134, 0.01, 0.019, 0.3112, 0.01, 0.0212, 0.3112, 0.01, 0.0233, 0.3134, 0.01, 0.0212, 0.3134, 0.01, 0.0233, 0.3088, 0.0061, 0.0255, 0.3088, 0.0061, 0.0277, 0.3088, 0.0083, 0.0255, 0.3088, 0.0083, 0.0277, 0.3088, 0.0061, 0.0298, 0.3088, 0.0061, 0.032, 0.3088, 0.0083, 0.0298, 0.3088, 0.0083, 0.032, 0.3088, 0.01, 0.0255, 0.3088, 0.01, 0.0277, 0.3088, 0.01, 0.0298, 0.3088, 0.01, 0.032, 0.3112, 0.01, 0.0255, 0.3112, 0.01, 0.0277, 0.3134, 0.01, 0.0255, 0.3134, 0.01, 0.0277, 0.3112, 0.01, 0.0298, 0.3112, 0.01, 0.032, 0.3134, 0.01, 0.0298, 0.3134, 0.01, 0.032, 0.2808, 0.01, 0.0342, 0.2808, 0.01, 0.0363, 0.283, 0.01, 0.0342, 0.283, 0.01, 0.0363,
0.2808, 0.01, 0.0385, 0.2808, 0.01, 0.0407, 0.283, 0.01, 0.0385, 0.283, 0.01, 0.0407, 0.2852, 0.01, 0.0342, 0.2852, 0.01, 0.0363, 0.2873, 0.01, 0.0342, 0.2873, 0.01, 0.0363, 0.2852, 0.01, 0.0385, 0.2852, 0.01, 0.0407, 0.2873, 0.01, 0.0385, 0.2873, 0.01, 0.0407, 0.2808, 0.0061, 0.0487, 0.2808, 0.0083, 0.0487, 0.283, 0.0061, 0.0487, 0.283, 0.0083, 0.0487, 0.2808, 0.01, 0.0429, 0.2808, 0.01, 0.045, 0.283, 0.01, 0.0429, 0.283, 0.01, 0.045, 0.2808, 0.01, 0.0472, 0.2808, 0.01, 0.0487, 0.283, 0.01, 0.0472, 0.283, 0.01, 0.0487, 0.2852, 0.0061, 0.0487, 0.2852, 0.0083, 0.0487, 0.2873, 0.0061, 0.0487, 0.2873, 0.0083, 0.0487, 0.2852, 0.01, 0.0429, 0.2852, 0.01, 0.045,
0.2873, 0.01, 0.0429, 0.2873, 0.01, 0.045, 0.2852, 0.01, 0.0472, 0.2852, 0.01, 0.0487, 0.2873, 0.01, 0.0472, 0.2873, 0.01, 0.0487, 0.2912, 0.0061, 0.0342, 0.2912, 0.0061, 0.0363, 0.2912, 0.0083, 0.0342, 0.2912, 0.0083, 0.0363, 0.2912, 0.0061, 0.0385, 0.2912, 0.0061, 0.0407, 0.2912, 0.0083, 0.0385, 0.2912, 0.0083, 0.0407, 0.2895, 0.01, 0.0342, 0.2895, 0.01, 0.0363, 0.2912, 0.01, 0.0342, 0.2912, 0.01, 0.0363, 0.2895, 0.01, 0.0385, 0.2895, 0.01, 0.0407, 0.2912, 0.01, 0.0385, 0.2912, 0.01, 0.0407, 0.2912, 0.0061, 0.0429, 0.2912, 0.0061, 0.045, 0.2912, 0.0083, 0.0429, 0.2912, 0.0083, 0.045, 0.2895, 0.0061, 0.0487, 0.2895, 0.0083, 0.0487, 0.2912, 0.0061, 0.0472, 0.2912, 0.0061, 0.0487,
0.2912, 0.0083, 0.0472, 0.2912, 0.0083, 0.0487, 0.2895, 0.01, 0.0429, 0.2895, 0.01, 0.045, 0.2912, 0.01, 0.0429, 0.2912, 0.01, 0.045, 0.2895, 0.01, 0.0472, 0.2895, 0.01, 0.0487, 0.2912, 0.01, 0.0472, 0.2912, 0.01, 0.0487, 0.3088, 0.0061, 0.0342, 0.3088, 0.0061, 0.0363, 0.3088, 0.0083, 0.0342, 0.3088, 0.0083, 0.0363, 0.3088, 0.0061, 0.0385, 0.3088, 0.0061, 0.0407, 0.3088, 0.0083, 0.0385, 0.3088, 0.0083, 0.0407, 0.3088, 0.01, 0.0342, 0.3088, 0.01, 0.0363, 0.3088, 0.01, 0.0385, 0.3088, 0.01, 0.0407, 0.3112, 0.01, 0.0342, 0.3112, 0.01, 0.0363, 0.3134, 0.01, 0.0342, 0.3134, 0.01, 0.0363, 0.3112, 0.01, 0.0385, 0.3112, 0.01, 0.0407, 0.3134, 0.01, 0.0385, 0.3134, 0.01, 0.0407,
0.3088, 0.0061, 0.0429, 0.3088, 0.0061, 0.045, 0.3088, 0.0083, 0.0429, 0.3088, 0.0083, 0.045, 0.3088, 0.0061, 0.0472, 0.3088, 0.0061, 0.0487, 0.3088, 0.0083, 0.0472, 0.3088, 0.0083, 0.0487, 0.3088, 0.01, 0.0429, 0.3088, 0.01, 0.045, 0.3088, 0.01, 0.0472, 0.3088, 0.01, 0.0487, 0.3112, 0.0061, 0.0487, 0.3112, 0.0083, 0.0487, 0.3134, 0.0061, 0.0487, 0.3134, 0.0083, 0.0487, 0.3112, 0.01, 0.0429, 0.3112, 0.01, 0.045, 0.3134, 0.01, 0.0429, 0.3134, 0.01, 0.045, 0.3112, 0.01, 0.0472, 0.3112, 0.01, 0.0487, 0.3134, 0.01, 0.0472, 0.3134, 0.01, 0.0487, 0.3155, 0.01, -0.0005, 0.3155, 0.01, 0.0016, 0.3177, 0.01, -0.0005, 0.3177, 0.01, 0.0016, 0.3155, 0.01, 0.0038, 0.3155, 0.01, 0.006,
0.3177, 0.01, 0.0038, 0.3177, 0.01, 0.006, 0.3206, 0.0061, -0.0005, 0.3206, 0.0061, 0.0016, 0.3206, 0.0083, -0.0005, 0.3206, 0.0083, 0.0016, 0.3206, 0.0061, 0.0038, 0.3206, 0.0061, 0.006, 0.3206, 0.0083, 0.0038, 0.3206, 0.0083, 0.006, 0.3206, 0.01, -0.0005, 0.3206, 0.01, 0.0016, 0.3206, 0.01, 0.0038, 0.3206, 0.01, 0.006, 0.3155, 0.01, 0.0081, 0.3155, 0.01, 0.0103, 0.3177, 0.01, 0.0081, 0.3177, 0.01, 0.0103, 0.3155, 0.01, 0.0125, 0.3155, 0.01, 0.0146, 0.3177, 0.01, 0.0125, 0.3177, 0.01, 0.0146, 0.3206, 0.0061, 0.0081, 0.3206, 0.0061, 0.0103, 0.3206, 0.0083, 0.0081, 0.3206, 0.0083, 0.0103, 0.3206, 0.0061, 0.0125, 0.3206, 0.0061, 0.0146, 0.3206, 0.0083, 0.0125, 0.3206, 0.0083, 0.0146,
0.3206, 0.01, 0.0081, 0.3206, 0.01, 0.0103, 0.3206, 0.01, 0.0125, 0.3206, 0.01, 0.0146, 0.3155, 0.01, 0.0168, 0.3155, 0.01, 0.019, 0.3177, 0.01, 0.0168, 0.3177, 0.01, 0.019, 0.3155, 0.01, 0.0212, 0.3155, 0.01, 0.0233, 0.3177, 0.01, 0.0212, 0.3177, 0.01, 0.0233, 0.3206, 0.0061, 0.0168, 0.3206, 0.0061, 0.019, 0.3206, 0.0083, 0.0168, 0.3206, 0.0083, 0.019, 0.3206, 0.0061, 0.0212, 0.3206, 0.0061, 0.0233, 0.3206, 0.0083, 0.0212, 0.3206, 0.0083, 0.0233, 0.3206, 0.01, 0.0168, 0.3206, 0.01, 0.019, 0.3206, 0.01, 0.0212, 0.3206, 0.01, 0.0233, 0.3155, 0.01, 0.0255, 0.3155, 0.01, 0.0277, 0.3177, 0.01, 0.0255, 0.3177, 0.01, 0.0277, 0.3155, 0.01, 0.0298, 0.3155, 0.01, 0.032,
0.3177, 0.01, 0.0298, 0.3177, 0.01, 0.032, 0.3206, 0.0061, 0.0255, 0.3206, 0.0061, 0.0277, 0.3206, 0.0083, 0.0255, 0.3206, 0.0083, 0.0277, 0.3206, 0.0061, 0.0298, 0.3206, 0.0061, 0.032, 0.3206, 0.0083, 0.0298, 0.3206, 0.0083, 0.032, 0.3206, 0.01, 0.0255, 0.3206, 0.01, 0.0277, 0.3206, 0.01, 0.0298, 0.3206, 0.01, 0.032, 0.3402, 0.0061, -0.0005, 0.3402, 0.0061, 0.0016, 0.3402, 0.0083, -0.0005, 0.3402, 0.0083, 0.0016, 0.3402, 0.0061, 0.0038, 0.3402, 0.0061, 0.006, 0.3402, 0.0083, 0.0038, 0.3402, 0.0083, 0.006, 0.3402, 0.01, -0.0005, 0.3402, 0.01, 0.0016, 0.3402, 0.01, 0.0038, 0.3402, 0.01, 0.006, 0.3402, 0.0061, 0.0081, 0.3402, 0.0061, 0.0103, 0.3402, 0.0083, 0.0081, 0.3402, 0.0083, 0.0103,
0.3402, 0.0061, 0.0125, 0.3402, 0.0061, 0.0146, 0.3402, 0.0083, 0.0125, 0.3402, 0.0083, 0.0146, 0.3402, 0.01, 0.0081, 0.3402, 0.01, 0.0103, 0.3402, 0.01, 0.0125, 0.3402, 0.01, 0.0146, 0.3416, 0.01, -0.0005, 0.3416, 0.01, 0.0016, 0.3438, 0.01, -0.0005, 0.3438, 0.01, 0.0016, 0.3416, 0.01, 0.0038, 0.3416, 0.01, 0.006, 0.3438, 0.01, 0.0038, 0.3438, 0.01, 0.006, 0.3459, 0.01, -0.0005, 0.3459, 0.01, 0.0016, 0.3481, 0.01, -0.0005, 0.3481, 0.01, 0.0016, 0.3459, 0.01, 0.0038, 0.3459, 0.01, 0.006, 0.3481, 0.01, 0.0038, 0.3481, 0.01, 0.006, 0.3416, 0.01, 0.0081, 0.3416, 0.01, 0.0103, 0.3438, 0.01, 0.0081, 0.3438, 0.01, 0.0103, 0.3416, 0.01, 0.0125, 0.3416, 0.01, 0.0146,
0.3438, 0.01, 0.0125, 0.3438, 0.01, 0.0146, 0.3459, 0.01, 0.0081, 0.3459, 0.01, 0.0103, 0.3481, 0.01, 0.0081, 0.3481, 0.01, 0.0103, 0.3459, 0.01, 0.0125, 0.3459, 0.01, 0.0146, 0.3481, 0.01, 0.0125, 0.3481, 0.01, 0.0146, 0.3402, 0.0061, 0.0168, 0.3402, 0.0061, 0.019, 0.3402, 0.0083, 0.0168, 0.3402, 0.0083, 0.019, 0.3402, 0.0061, 0.0212, 0.3402, 0.0061, 0.0233, 0.3402, 0.0083, 0.0212, 0.3402, 0.0083, 0.0233, 0.3402, 0.01, 0.0168, 0.3402, 0.01, 0.019, 0.3402, 0.01, 0.0212, 0.3402, 0.01, 0.0233, 0.3402, 0.0061, 0.0255, 0.3402, 0.0061, 0.0277, 0.3402, 0.0083, 0.0255, 0.3402, 0.0083, 0.0277, 0.3402, 0.0061, 0.0298, 0.3402, 0.0061, 0.032, 0.3402, 0.0083, 0.0298, 0.3402, 0.0083, 0.032,
0.3402, 0.01, 0.0255, 0.3402, 0.01, 0.0277, 0.3402, 0.01, 0.0298, 0.3402, 0.01, 0.032, 0.3416, 0.01, 0.0168, 0.3416, 0.01, 0.019, 0.3438, 0.01, 0.0168, 0.3438, 0.01, 0.019, 0.3416, 0.01, 0.0212, 0.3416, 0.01, 0.0233, 0.3438, 0.01, 0.0212, 0.3438, 0.01, 0.0233, 0.3459, 0.01, 0.0168, 0.3459, 0.01, 0.019, 0.3481, 0.01, 0.0168, 0.3481, 0.01, 0.019, 0.3459, 0.01, 0.0212, 0.3459, 0.01, 0.0233, 0.3481, 0.01, 0.0212, 0.3481, 0.01, 0.0233, 0.3416, 0.01, 0.0255, 0.3416, 0.01, 0.0277, 0.3438, 0.01, 0.0255, 0.3438, 0.01, 0.0277, 0.3416, 0.01, 0.0298, 0.3416, 0.01, 0.032, 0.3438, 0.01, 0.0298, 0.3438, 0.01, 0.032, 0.3459, 0.01, 0.0255, 0.3459, 0.01, 0.0277,
0.3481, 0.01, 0.0255, 0.3481, 0.01, 0.0277, 0.3459, 0.01, 0.0298, 0.3459, 0.01, 0.032, 0.3481, 0.01, 0.0298, 0.3481, 0.01, 0.032, 0.3155, 0.01, 0.0342, 0.3155, 0.01, 0.0363, 0.3177, 0.01, 0.0342, 0.3177, 0.01, 0.0363, 0.3155, 0.01, 0.0385, 0.3155, 0.01, 0.0407, 0.3177, 0.01, 0.0385, 0.3177, 0.01, 0.0407, 0.3206, 0.0061, 0.0342, 0.3206, 0.0061, 0.0363, 0.3206, 0.0083, 0.0342, 0.3206, 0.0083, 0.0363, 0.3206, 0.0061, 0.0385, 0.3206, 0.0061, 0.0407, 0.3206, 0.0083, 0.0385, 0.3206, 0.0083, 0.0407, 0.3206, 0.01, 0.0342, 0.3206, 0.01, 0.0363, 0.3206, 0.01, 0.0385, 0.3206, 0.01, 0.0407, 0.3155, 0.0061, 0.0487, 0.3155, 0.0083, 0.0487, 0.3177, 0.0061, 0.0487, 0.3177, 0.0083, 0.0487,
0.3155, 0.01, 0.0429, 0.3155, 0.01, 0.045, 0.3177, 0.01, 0.0429, 0.3177, 0.01, 0.045, 0.3155, 0.01, 0.0472, 0.3155, 0.01, 0.0487, 0.3177, 0.01, 0.0472, 0.3177, 0.01, 0.0487, 0.3206, 0.0061, 0.0429, 0.3206, 0.0061, 0.045, 0.3206, 0.0083, 0.0429, 0.3206, 0.0083, 0.045, 0.3206, 0.0061, 0.0472, 0.3206, 0.0061, 0.0487, 0.3206, 0.0083, 0.0472, 0.3206, 0.0083, 0.0487, 0.3206, 0.01, 0.0429, 0.3206, 0.01, 0.045, 0.3206, 0.01, 0.0472, 0.3206, 0.01, 0.0487, 0.3402, 0.0061, 0.0342, 0.3402, 0.0061, 0.0363, 0.3402, 0.0083, 0.0342, 0.3402, 0.0083, 0.0363, 0.3402, 0.0061, 0.0385, 0.3402, 0.0061, 0.0407, 0.3402, 0.0083, 0.0385, 0.3402, 0.0083, 0.0407, 0.3402, 0.01, 0.0342, 0.3402, 0.01, 0.0363,
0.3402, 0.01, 0.0385, 0.3402, 0.01, 0.0407, 0.3402, 0.0061, 0.0429, 0.3402, 0.0061, 0.045, 0.3402, 0.0083, 0.0429, 0.3402, 0.0083, 0.045, 0.3402, 0.0061, 0.0472, 0.3402, 0.0061, 0.0487, 0.3402, 0.0083, 0.0472, 0.3402, 0.0083, 0.0487, 0.3402, 0.01, 0.0429, 0.3402, 0.01, 0.045, 0.3402, 0.01, 0.0472, 0.3402, 0.01, 0.0487, 0.3416, 0.01, 0.0342, 0.3416, 0.01, 0.0363, 0.3438, 0.01, 0.0342, 0.3438, 0.01, 0.0363, 0.3416, 0.01, 0.0385, 0.3416, 0.01, 0.0407, 0.3438, 0.01, 0.0385, 0.3438, 0.01, 0.0407, 0.3459, 0.01, 0.0342, 0.3459, 0.01, 0.0363, 0.3481, 0.01, 0.0342, 0.3481, 0.01, 0.0363, 0.3459, 0.01, 0.0385, 0.3459, 0.01, 0.0407, 0.3481, 0.01, 0.0385, 0.3481, 0.01, 0.0407,
0.3416, 0.0061, 0.0487, 0.3416, 0.0083, 0.0487, 0.3438, 0.0061, 0.0487, 0.3438, 0.0083, 0.0487, 0.3416, 0.01, 0.0429, 0.3416, 0.01, 0.045, 0.3438, 0.01, 0.0429, 0.3438, 0.01, 0.045, 0.3416, 0.01, 0.0472, 0.3416, 0.01, 0.0487, 0.3438, 0.01, 0.0472, 0.3438, 0.01, 0.0487, 0.3459, 0.0061, 0.0487, 0.3459, 0.0083, 0.0487, 0.3481, 0.0061, 0.0487, 0.3481, 0.0083, 0.0487, 0.3459, 0.01, 0.0429, 0.3459, 0.01, 0.045, 0.3481, 0.01, 0.0429, 0.3481, 0.01, 0.045, 0.3459, 0.01, 0.0472, 0.3459, 0.01, 0.0487, 0.3481, 0.01, 0.0472, 0.3481, 0.01, 0.0487, 0.352, 0.0061, -0.0005, 0.352, 0.0061, 0.0016, 0.352, 0.0083, -0.0005, 0.352, 0.0083, 0.0016, 0.352, 0.0061, 0.0038, 0.352, 0.0061, 0.006,
0.352, 0.0083, 0.0038, 0.352, 0.0083, 0.006, 0.3503, 0.01, -0.0005, 0.3503, 0.01, 0.0016, 0.352, 0.01, -0.0005, 0.352, 0.01, 0.0016, 0.3503, 0.01, 0.0038, 0.3503, 0.01, 0.006, 0.352, 0.01, 0.0038, 0.352, 0.01, 0.006, 0.352, 0.0061, 0.0081, 0.352, 0.0061, 0.0103, 0.352, 0.0083, 0.0081, 0.352, 0.0083, 0.0103, 0.352, 0.0061, 0.0125, 0.352, 0.0061, 0.0146, 0.352, 0.0083, 0.0125, 0.352, 0.0083, 0.0146, 0.3503, 0.01, 0.0081, 0.3503, 0.01, 0.0103, 0.352, 0.01, 0.0081, 0.352, 0.01, 0.0103, 0.3503, 0.01, 0.0125, 0.3503, 0.01, 0.0146, 0.352, 0.01, 0.0125, 0.352, 0.01, 0.0146, 0.352, 0.0061, 0.0168, 0.352, 0.0061, 0.019, 0.352, 0.0083, 0.0168, 0.352, 0.0083, 0.019,
0.352, 0.0061, 0.0212, 0.352, 0.0061, 0.0233, 0.352, 0.0083, 0.0212, 0.352, 0.0083, 0.0233, 0.3503, 0.01, 0.0168, 0.3503, 0.01, 0.019, 0.352, 0.01, 0.0168, 0.352, 0.01, 0.019, 0.3503, 0.01, 0.0212, 0.3503, 0.01, 0.0233, 0.352, 0.01, 0.0212, 0.352, 0.01, 0.0233, 0.352, 0.0061, 0.0255, 0.352, 0.0061, 0.0277, 0.352, 0.0083, 0.0255, 0.352, 0.0083, 0.0277, 0.352, 0.0061, 0.0298, 0.352, 0.0061, 0.032, 0.352, 0.0083, 0.0298, 0.352, 0.0083, 0.032, 0.3503, 0.01, 0.0255, 0.3503, 0.01, 0.0277, 0.352, 0.01, 0.0255, 0.352, 0.01, 0.0277, 0.3503, 0.01, 0.0298, 0.3503, 0.01, 0.032, 0.352, 0.01, 0.0298, 0.352, 0.01, 0.032, 0.352, 0.0061, 0.0342, 0.352, 0.0061, 0.0363,
0.352, 0.0083, 0.0342, 0.352, 0.0083, 0.0363, 0.352, 0.0061, 0.0379, 0.352, 0.0083, 0.0379, 0.3503, 0.01, 0.0342, 0.3503, 0.01, 0.0363, 0.352, 0.01, 0.0342, 0.352, 0.01, 0.0363, 0.3503, 0.01, 0.0385, 0.3503, 0.01, 0.0407, 0.352, 0.01, 0.0379, 0.3524, 0.01, 0.0407, 0.3546, 0.0061, 0.0379, 0.3546, 0.0083, 0.0379, 0.3568, 0.0061, 0.0379, 0.3568, 0.0083, 0.0379, 0.3546, 0.01, 0.0379, 0.3546, 0.01, 0.0407, 0.3568, 0.01, 0.0379, 0.3568, 0.01, 0.0407, 0.3503, 0.0061, 0.0487, 0.3503, 0.0083, 0.0487, 0.3524, 0.0061, 0.0487, 0.3524, 0.0083, 0.0487, 0.3503, 0.01, 0.0429, 0.3503, 0.01, 0.045, 0.3524, 0.01, 0.0429, 0.3524, 0.01, 0.045, 0.3503, 0.01, 0.0472, 0.3503, 0.01, 0.0487,
0.3524, 0.01, 0.0472, 0.3524, 0.01, 0.0487, 0.3546, 0.0061, 0.0487, 0.3546, 0.0083, 0.0487, 0.3568, 0.0061, 0.0487, 0.3568, 0.0083, 0.0487, 0.3546, 0.01, 0.0429, 0.3546, 0.01, 0.045, 0.3568, 0.01, 0.0429, 0.3568, 0.01, 0.045, 0.3546, 0.01, 0.0472, 0.3546, 0.01, 0.0487, 0.3568, 0.01, 0.0472, 0.3568, 0.01, 0.0487, 0.3589, 0.0061, 0.0379, 0.3589, 0.0083, 0.0379, 0.3611, 0.0061, 0.0379, 0.3611, 0.0083, 0.0379, 0.3589, 0.01, 0.0379, 0.3589, 0.01, 0.0407, 0.3611, 0.01, 0.0379, 0.3611, 0.01, 0.0407, 0.3633, 0.0061, 0.0379, 0.3633, 0.0083, 0.0379, 0.3655, 0.0061, 0.0379, 0.3655, 0.0083, 0.0379, 0.3633, 0.01, 0.0379, 0.3633, 0.01, 0.0407, 0.3655, 0.01, 0.0379, 0.3655, 0.01, 0.0407,
0.3589, 0.0061, 0.0487, 0.3589, 0.0083, 0.0487, 0.3611, 0.0061, 0.0487, 0.3611, 0.0083, 0.0487, 0.3589, 0.01, 0.0429, 0.3589, 0.01, 0.045, 0.3611, 0.01, 0.0429, 0.3611, 0.01, 0.045, 0.3589, 0.01, 0.0472, 0.3589, 0.01, 0.0487, 0.3611, 0.01, 0.0472, 0.3611, 0.01, 0.0487, 0.3633, 0.0061, 0.0487, 0.3633, 0.0083, 0.0487, 0.3655, 0.0061, 0.0487, 0.3655, 0.0083, 0.0487, 0.3633, 0.01, 0.0429, 0.3633, 0.01, 0.045, 0.3655, 0.01, 0.0429, 0.3655, 0.01, 0.045, 0.3633, 0.01, 0.0472, 0.3633, 0.01, 0.0487, 0.3655, 0.01, 0.0472, 0.3655, 0.01, 0.0487, 0.3676, 0.0061, 0.0379, 0.3676, 0.0083, 0.0379, 0.3698, 0.0061, 0.0379, 0.3698, 0.0083, 0.0379, 0.3676, 0.01, 0.0379, 0.3676, 0.01, 0.0407,
0.3698, 0.01, 0.0379, 0.3698, 0.01, 0.0407, 0.372, 0.0061, 0.0379, 0.372, 0.0083, 0.0379, 0.3741, 0.0061, 0.0379, 0.3741, 0.0083, 0.0379, 0.372, 0.01, 0.0379, 0.372, 0.01, 0.0407, 0.3741, 0.01, 0.0379, 0.3741, 0.01, 0.0407, 0.3676, 0.0061, 0.0487, 0.3676, 0.0083, 0.0487, 0.3698, 0.0061, 0.0487, 0.3698, 0.0083, 0.0487, 0.3676, 0.01, 0.0429, 0.3676, 0.01, 0.045, 0.3698, 0.01, 0.0429, 0.3698, 0.01, 0.045, 0.3676, 0.01, 0.0472, 0.3676, 0.01, 0.0487, 0.3698, 0.01, 0.0472, 0.3698, 0.01, 0.0487, 0.372, 0.0061, 0.0487, 0.372, 0.0083, 0.0487, 0.3741, 0.0061, 0.0487, 0.3741, 0.0083, 0.0487, 0.372, 0.01, 0.0429, 0.372, 0.01, 0.045, 0.3741, 0.01, 0.0429, 0.3741, 0.01, 0.045,
0.372, 0.01, 0.0472, 0.372, 0.01, 0.0487, 0.3741, 0.01, 0.0472, 0.3741, 0.01, 0.0487, 0.3763, 0.0061, 0.0379, 0.3763, 0.0083, 0.0379, 0.3785, 0.0061, 0.0379, 0.3785, 0.0083, 0.0379, 0.3763, 0.01, 0.0379, 0.3763, 0.01, 0.0407, 0.3784, 0.01, 0.0379, 0.3785, 0.01, 0.0407, 0.3833, 0.0061, 0.0373, 0.3833, 0.0083, 0.0373, 0.3807, 0.0061, 0.0377, 0.3807, 0.0083, 0.0377, 0.3822, 0.0061, 0.0375, 0.3822, 0.0083, 0.0375, 0.3835, 0.01, 0.0373, 0.3806, 0.01, 0.0377, 0.3806, 0.01, 0.0407, 0.3827, 0.01, 0.0374, 0.3828, 0.01, 0.0407, 0.3763, 0.0061, 0.0487, 0.3763, 0.0083, 0.0487, 0.3785, 0.0061, 0.0487, 0.3785, 0.0083, 0.0487, 0.3763, 0.01, 0.0429, 0.3763, 0.01, 0.045, 0.3785, 0.01, 0.0429,
0.3785, 0.01, 0.045, 0.3763, 0.01, 0.0472, 0.3763, 0.01, 0.0487, 0.3785, 0.01, 0.0472, 0.3785, 0.01, 0.0487, 0.3806, 0.0061, 0.0487, 0.3806, 0.0083, 0.0487, 0.3828, 0.0061, 0.0487, 0.3828, 0.0083, 0.0487, 0.3806, 0.01, 0.0429, 0.3806, 0.01, 0.045, 0.3828, 0.01, 0.0429, 0.3828, 0.01, 0.045, 0.3806, 0.01, 0.0472, 0.3806, 0.01, 0.0487, 0.3828, 0.01, 0.0472, 0.3828, 0.01, 0.0487, 0.3999, 0.0061, -0.0006, 0.401, 0.0061, 0.0011, 0.3999, 0.0083, -0.0006, 0.401, 0.0083, 0.0011, 0.4002, 0.01, -0.0003, 0.4009, 0.01, 0.001, 0.3947, 0.0061, 0.033, 0.3947, 0.0083, 0.033, 0.3959, 0.0061, 0.0321, 0.3959, 0.0083, 0.0321, 0.3947, 0.01, 0.033, 0.3956, 0.01, 0.0324, 0.4009, 0.0061, 0.026,
0.3999, 0.0061, 0.0277, 0.4009, 0.0083, 0.026, 0.3999, 0.0083, 0.0277, 0.3982, 0.0061, 0.0299, 0.3971, 0.0061, 0.031, 0.3982, 0.0083, 0.0299, 0.3971, 0.0083, 0.031, 0.3991, 0.0061, 0.0288, 0.3991, 0.0083, 0.0288, 0.4009, 0.01, 0.0261, 0.4001, 0.01, 0.0274, 0.3982, 0.01, 0.03, 0.3971, 0.01, 0.031, 0.399, 0.01, 0.0289, 0.4002, 0.01, 0.032, 0.4016, 0.0061, 0.0021, 0.4016, 0.0083, 0.0021, 0.4024, 0.0061, 0.0038, 0.4032, 0.0061, 0.0058, 0.4024, 0.0083, 0.0038, 0.4032, 0.0083, 0.0058, 0.4035, 0.0061, 0.0069, 0.4035, 0.0083, 0.0069, 0.4024, 0.01, -0.0005, 0.4015, 0.01, 0.002, 0.4045, 0.01, -0.0005, 0.4045, 0.01, 0.0016, 0.4025, 0.01, 0.004, 0.4031, 0.01, 0.0056, 0.4045, 0.01, 0.0038,
0.4033, 0.01, 0.0064, 0.4067, 0.01, -0.0005, 0.4067, 0.01, 0.0016, 0.4089, 0.01, -0.0005, 0.4089, 0.01, 0.0016, 0.4067, 0.01, 0.0038, 0.4067, 0.01, 0.006, 0.4089, 0.01, 0.0038, 0.4089, 0.01, 0.006, 0.4037, 0.0061, 0.0081, 0.4039, 0.0061, 0.0103, 0.4037, 0.0083, 0.0081, 0.4039, 0.0083, 0.0103, 0.4039, 0.0061, 0.0125, 0.4039, 0.0061, 0.0146, 0.4039, 0.0083, 0.0125, 0.4039, 0.0083, 0.0146, 0.4038, 0.01, 0.0083, 0.4039, 0.01, 0.0104, 0.4039, 0.01, 0.0125, 0.4039, 0.01, 0.0146, 0.4067, 0.01, 0.0081, 0.4067, 0.01, 0.0103, 0.4089, 0.01, 0.0081, 0.4089, 0.01, 0.0103, 0.4067, 0.01, 0.0125, 0.4067, 0.01, 0.0146, 0.4089, 0.01, 0.0125, 0.4089, 0.01, 0.0146, 0.4132, 0.0061, -0.0006,
0.414, 0.0061, 0.0015, 0.4132, 0.0083, -0.0006, 0.414, 0.0083, 0.0015, 0.411, 0.01, -0.0005, 0.411, 0.01, 0.0016, 0.4131, 0.01, -0.0007, 0.414, 0.01, 0.0015, 0.411, 0.01, 0.0038, 0.411, 0.01, 0.006, 0.4132, 0.01, 0.0038, 0.4132, 0.01, 0.006, 0.4143, 0.0061, 0.0026, 0.4143, 0.0083, 0.0026, 0.4147, 0.0061, 0.0038, 0.4153, 0.0061, 0.006, 0.4147, 0.0083, 0.0038, 0.4153, 0.0083, 0.006, 0.4144, 0.01, 0.0026, 0.4147, 0.01, 0.0037, 0.4152, 0.01, 0.0059, 0.411, 0.01, 0.0081, 0.411, 0.01, 0.0103, 0.4132, 0.01, 0.0081, 0.4132, 0.01, 0.0103, 0.411, 0.01, 0.0125, 0.411, 0.01, 0.0146, 0.4132, 0.01, 0.0125, 0.4132, 0.01, 0.0146, 0.4156, 0.0061, 0.0081, 0.4157, 0.0061, 0.0103,
0.4156, 0.0083, 0.0081, 0.4157, 0.0083, 0.0103, 0.4157, 0.0061, 0.0125, 0.4157, 0.0061, 0.0146, 0.4157, 0.0083, 0.0125, 0.4157, 0.0083, 0.0146, 0.4156, 0.01, 0.0081, 0.4157, 0.01, 0.0103, 0.4157, 0.01, 0.0125, 0.4157, 0.01, 0.0146, 0.4039, 0.0061, 0.0168, 0.4037, 0.0061, 0.019, 0.4039, 0.0083, 0.0168, 0.4037, 0.0083, 0.019, 0.4032, 0.0061, 0.0212, 0.4024, 0.0061, 0.0234, 0.4032, 0.0083, 0.0212, 0.4024, 0.0083, 0.0234, 0.4034, 0.0061, 0.0201, 0.4034, 0.0083, 0.0201, 0.4039, 0.01, 0.0168, 0.4037, 0.01, 0.0188, 0.4031, 0.01, 0.0215, 0.4025, 0.01, 0.0231, 0.4033, 0.01, 0.0207, 0.4045, 0.01, 0.0233, 0.4067, 0.01, 0.0168, 0.4067, 0.01, 0.019, 0.4089, 0.01, 0.0168, 0.4089, 0.01, 0.019,
0.4067, 0.01, 0.0212, 0.4067, 0.01, 0.0233, 0.4089, 0.01, 0.0212, 0.4089, 0.01, 0.0233, 0.4016, 0.0061, 0.025, 0.4016, 0.0083, 0.025, 0.4015, 0.01, 0.0251, 0.4024, 0.01, 0.0277, 0.4045, 0.01, 0.0255, 0.4045, 0.01, 0.0277, 0.4024, 0.01, 0.0298, 0.4024, 0.01, 0.032, 0.4045, 0.01, 0.0298, 0.4045, 0.01, 0.032, 0.4067, 0.01, 0.0255, 0.4067, 0.01, 0.0277, 0.4089, 0.01, 0.0255, 0.4089, 0.01, 0.0277, 0.4067, 0.01, 0.0298, 0.4067, 0.01, 0.032, 0.4089, 0.01, 0.0298, 0.4089, 0.01, 0.032, 0.411, 0.01, 0.0168, 0.411, 0.01, 0.019, 0.4132, 0.01, 0.0168, 0.4132, 0.01, 0.019, 0.411, 0.01, 0.0212, 0.411, 0.01, 0.0233, 0.4132, 0.01, 0.0212, 0.4132, 0.01, 0.0233,
0.4157, 0.0061, 0.0168, 0.4157, 0.0061, 0.019, 0.4157, 0.0083, 0.0168, 0.4157, 0.0083, 0.019, 0.4155, 0.0061, 0.0212, 0.4152, 0.0061, 0.0233, 0.4155, 0.0083, 0.0212, 0.4152, 0.0083, 0.0233, 0.4157, 0.01, 0.0168, 0.4157, 0.01, 0.019, 0.4155, 0.01, 0.0212, 0.4152, 0.01, 0.0234, 0.4142, 0.0061, 0.0265, 0.4138, 0.0061, 0.0277, 0.4142, 0.0083, 0.0265, 0.4138, 0.0083, 0.0277, 0.4117, 0.0061, 0.0323, 0.4117, 0.0083, 0.0323, 0.4129, 0.0061, 0.0299, 0.4123, 0.0061, 0.0312, 0.4129, 0.0083, 0.0299, 0.4123, 0.0083, 0.0312, 0.411, 0.01, 0.0255, 0.411, 0.01, 0.0277, 0.4144, 0.01, 0.026, 0.4137, 0.01, 0.0279, 0.411, 0.01, 0.0298, 0.4117, 0.01, 0.0323, 0.4129, 0.01, 0.03, 0.4123, 0.01, 0.0312,
0.4146, 0.0061, 0.0254, 0.4146, 0.0083, 0.0254, 0.4147, 0.01, 0.0251, 0.385, 0.0061, 0.0371, 0.385, 0.0083, 0.0371, 0.3872, 0.0061, 0.0366, 0.3872, 0.0083, 0.0366, 0.385, 0.01, 0.0371, 0.3871, 0.01, 0.0367, 0.385, 0.01, 0.0385, 0.385, 0.01, 0.0407, 0.3872, 0.01, 0.0385, 0.3872, 0.01, 0.0407, 0.3894, 0.0061, 0.036, 0.3894, 0.0083, 0.036, 0.3918, 0.0061, 0.0348, 0.3907, 0.0061, 0.0354, 0.3918, 0.0083, 0.0348, 0.3907, 0.0083, 0.0354, 0.3891, 0.01, 0.0361, 0.3919, 0.01, 0.0348, 0.3909, 0.01, 0.0352, 0.3893, 0.01, 0.0385, 0.3893, 0.01, 0.0407, 0.3915, 0.01, 0.0385, 0.3915, 0.01, 0.0407, 0.385, 0.0061, 0.0487, 0.385, 0.0083, 0.0487, 0.3872, 0.0061, 0.0486, 0.3872, 0.0083, 0.0486,
0.385, 0.01, 0.0429, 0.385, 0.01, 0.045, 0.3872, 0.01, 0.0429, 0.3872, 0.01, 0.045, 0.385, 0.01, 0.0472, 0.385, 0.01, 0.0487, 0.3872, 0.01, 0.0472, 0.3872, 0.01, 0.0486, 0.39, 0.0061, 0.0482, 0.389, 0.0061, 0.0484, 0.39, 0.0083, 0.0482, 0.389, 0.0083, 0.0484, 0.3915, 0.0061, 0.0479, 0.3915, 0.0083, 0.0479, 0.3893, 0.01, 0.0429, 0.3893, 0.01, 0.045, 0.3915, 0.01, 0.0429, 0.3915, 0.01, 0.045, 0.3896, 0.01, 0.0483, 0.3887, 0.01, 0.0485, 0.3917, 0.01, 0.0478, 0.3936, 0.0061, 0.0338, 0.3936, 0.0083, 0.0338, 0.3936, 0.01, 0.0338, 0.3937, 0.01, 0.0363, 0.3958, 0.01, 0.0342, 0.3958, 0.01, 0.0363, 0.3937, 0.01, 0.0385, 0.3937, 0.01, 0.0407, 0.3958, 0.01, 0.0385,
0.3958, 0.01, 0.0407, 0.398, 0.01, 0.0342, 0.398, 0.01, 0.0363, 0.4002, 0.01, 0.0342, 0.4002, 0.01, 0.0363, 0.398, 0.01, 0.0385, 0.398, 0.01, 0.0407, 0.4002, 0.01, 0.0385, 0.4002, 0.01, 0.0407, 0.3967, 0.0061, 0.046, 0.3967, 0.0083, 0.046, 0.3937, 0.0061, 0.0472, 0.3937, 0.0083, 0.0472, 0.3956, 0.0061, 0.0465, 0.3956, 0.0083, 0.0465, 0.3937, 0.01, 0.0429, 0.3937, 0.01, 0.045, 0.3958, 0.01, 0.0429, 0.3964, 0.01, 0.0461, 0.3938, 0.01, 0.0471, 0.3954, 0.01, 0.0465, 0.398, 0.0061, 0.0454, 0.398, 0.0083, 0.0454, 0.4009, 0.0061, 0.0437, 0.3998, 0.0061, 0.0444, 0.4009, 0.0083, 0.0437, 0.3998, 0.0083, 0.0444, 0.398, 0.01, 0.0429, 0.3983, 0.01, 0.0453, 0.4008, 0.01, 0.0438,
0.3997, 0.01, 0.0445, 0.4046, 0.0061, 0.0409, 0.4046, 0.0083, 0.0409, 0.4024, 0.01, 0.0342, 0.4024, 0.01, 0.0363, 0.4045, 0.01, 0.0342, 0.4045, 0.01, 0.0363, 0.4024, 0.01, 0.0385, 0.4024, 0.01, 0.0407, 0.4045, 0.01, 0.0385, 0.4046, 0.01, 0.0409, 0.4099, 0.0061, 0.0352, 0.409, 0.0061, 0.0364, 0.4099, 0.0083, 0.0352, 0.409, 0.0083, 0.0364, 0.4069, 0.0061, 0.0387, 0.4058, 0.0061, 0.0398, 0.4069, 0.0083, 0.0387, 0.4058, 0.0083, 0.0398, 0.4079, 0.0061, 0.0376, 0.4079, 0.0083, 0.0376, 0.4067, 0.01, 0.0342, 0.4067, 0.01, 0.0363, 0.41, 0.01, 0.035, 0.4088, 0.01, 0.0367, 0.4069, 0.01, 0.0387, 0.4058, 0.01, 0.0398, 0.4079, 0.01, 0.0376, 0.4024, 0.0061, 0.0427, 0.4024, 0.0083, 0.0427,
0.4035, 0.0061, 0.0418, 0.4035, 0.0083, 0.0418, 0.4026, 0.01, 0.0426, 0.4035, 0.01, 0.0418, 0.4106, 0.0061, 0.0341, 0.4106, 0.0083, 0.0341, 0.4107, 0.01, 0.034, 0.429, 0.0061, 0.0153, 0.429, 0.0083, 0.0153, 0.4316, 0.0061, 0.0135, 0.4305, 0.0061, 0.0142, 0.4316, 0.0083, 0.0135, 0.4305, 0.0083, 0.0142, 0.4291, 0.01, 0.0153, 0.4316, 0.01, 0.0135, 0.4304, 0.01, 0.0142, 0.4327, 0.0061, 0.0129, 0.4327, 0.0083, 0.0129, 0.4349, 0.0061, 0.0119, 0.4349, 0.0083, 0.0119, 0.4326, 0.01, 0.013, 0.4327, 0.01, 0.0146, 0.4346, 0.01, 0.012, 0.4349, 0.01, 0.0146, 0.427, 0.0061, 0.0175, 0.4258, 0.0061, 0.0189, 0.427, 0.0083, 0.0175, 0.4258, 0.0083, 0.0189, 0.4245, 0.0061, 0.0211, 0.4237, 0.0061, 0.0233,
0.4245, 0.0083, 0.0211, 0.4237, 0.0083, 0.0233, 0.4252, 0.0061, 0.0201, 0.4252, 0.0083, 0.0201, 0.427, 0.01, 0.0175, 0.4261, 0.01, 0.0186, 0.4244, 0.01, 0.0214, 0.4238, 0.01, 0.0231, 0.425, 0.01, 0.0204, 0.4262, 0.01, 0.0233, 0.423, 0.0061, 0.0265, 0.4228, 0.0061, 0.0277, 0.423, 0.0083, 0.0265, 0.4228, 0.0083, 0.0277, 0.4226, 0.0061, 0.0298, 0.4228, 0.0061, 0.0319, 0.4226, 0.0083, 0.0298, 0.4228, 0.0083, 0.0319, 0.4229, 0.01, 0.0265, 0.4228, 0.01, 0.0277, 0.4226, 0.01, 0.0298, 0.4228, 0.01, 0.0316, 0.4231, 0.0061, 0.0254, 0.4231, 0.0083, 0.0254, 0.423, 0.0061, 0.033, 0.423, 0.0083, 0.033, 0.4231, 0.01, 0.0255, 0.4241, 0.01, 0.0277, 0.4262, 0.01, 0.0255, 0.4262, 0.01, 0.0277,
0.4241, 0.01, 0.0298, 0.4229, 0.01, 0.0324, 0.4262, 0.01, 0.0298, 0.4262, 0.01, 0.032, 0.4279, 0.0061, 0.0164, 0.4279, 0.0083, 0.0164, 0.4279, 0.01, 0.0164, 0.4284, 0.01, 0.019, 0.4306, 0.01, 0.0168, 0.4306, 0.01, 0.019, 0.4284, 0.01, 0.0212, 0.4284, 0.01, 0.0233, 0.4306, 0.01, 0.0212, 0.4306, 0.01, 0.0233, 0.4327, 0.01, 0.0168, 0.4327, 0.01, 0.019, 0.4349, 0.01, 0.0168, 0.4349, 0.01, 0.019, 0.4327, 0.01, 0.0212, 0.4327, 0.01, 0.0233, 0.4349, 0.01, 0.0212, 0.4349, 0.01, 0.0233, 0.4284, 0.01, 0.0255, 0.4284, 0.01, 0.0277, 0.4306, 0.01, 0.0255, 0.4306, 0.01, 0.0277, 0.4284, 0.01, 0.0298, 0.4284, 0.01, 0.032, 0.4306, 0.01, 0.0298, 0.4306, 0.01, 0.032,
0.4327, 0.01, 0.0255, 0.4327, 0.01, 0.0277, 0.4349, 0.01, 0.0255, 0.4349, 0.01, 0.0277, 0.4327, 0.01, 0.0298, 0.4327, 0.01, 0.032, 0.4349, 0.01, 0.0298, 0.4349, 0.01, 0.032, 0.4375, 0.0061, 0.0112, 0.4375, 0.0083, 0.0112, 0.4392, 0.0061, 0.0109, 0.4392, 0.0083, 0.0109, 0.4364, 0.0061, 0.0115, 0.4364, 0.0083, 0.0115, 0.4377, 0.01, 0.0112, 0.4392, 0.01, 0.0109, 0.4368, 0.01, 0.0114, 0.4371, 0.01, 0.0146, 0.4392, 0.01, 0.0125, 0.4392, 0.01, 0.0146, 0.4414, 0.0061, 0.0106, 0.4414, 0.0083, 0.0106, 0.4436, 0.0061, 0.0105, 0.4436, 0.0083, 0.0105, 0.4414, 0.01, 0.0106, 0.4436, 0.01, 0.0105, 0.4414, 0.01, 0.0125, 0.4414, 0.01, 0.0146, 0.4436, 0.01, 0.0125, 0.4436, 0.01, 0.0146,
0.4458, 0.0061, 0.0105, 0.4458, 0.0083, 0.0105, 0.4479, 0.0061, 0.0105, 0.4479, 0.0083, 0.0105, 0.4458, 0.01, 0.0105, 0.4479, 0.01, 0.0105, 0.4458, 0.01, 0.0125, 0.4458, 0.01, 0.0146, 0.4479, 0.01, 0.0125, 0.4479, 0.01, 0.0146, 0.4501, 0.0061, 0.0105, 0.4501, 0.0083, 0.0105, 0.4523, 0.0061, 0.0105, 0.4523, 0.0083, 0.0105, 0.4501, 0.01, 0.0105, 0.4523, 0.01, 0.0105, 0.4501, 0.01, 0.0125, 0.4501, 0.01, 0.0146, 0.4523, 0.01, 0.0125, 0.4523, 0.01, 0.0146, 0.4399, 0.0061, 0.0241, 0.4399, 0.0083, 0.0241, 0.4371, 0.01, 0.0168, 0.4371, 0.01, 0.019, 0.4392, 0.01, 0.0168, 0.4392, 0.01, 0.019, 0.4371, 0.01, 0.0212, 0.4371, 0.01, 0.0233, 0.4392, 0.01, 0.0212, 0.4398, 0.01, 0.0242,
0.4414, 0.0061, 0.0232, 0.4414, 0.0083, 0.0232, 0.4436, 0.0061, 0.0225, 0.4436, 0.0083, 0.0225, 0.4414, 0.01, 0.0168, 0.4414, 0.01, 0.019, 0.4436, 0.01, 0.0168, 0.4436, 0.01, 0.019, 0.4414, 0.01, 0.0212, 0.4416, 0.01, 0.0232, 0.4436, 0.01, 0.0212, 0.4436, 0.01, 0.0225, 0.4377, 0.0061, 0.0261, 0.4367, 0.0061, 0.0276, 0.4377, 0.0083, 0.0261, 0.4367, 0.0083, 0.0276, 0.4388, 0.0061, 0.025, 0.4388, 0.0083, 0.025, 0.436, 0.0061, 0.0297, 0.4367, 0.0061, 0.0321, 0.436, 0.0083, 0.0297, 0.4367, 0.0083, 0.0321, 0.4378, 0.01, 0.026, 0.4366, 0.01, 0.0278, 0.4388, 0.01, 0.025, 0.436, 0.01, 0.0297, 0.4366, 0.01, 0.0319, 0.4467, 0.0061, 0.0222, 0.4456, 0.0061, 0.0223, 0.4467, 0.0083, 0.0222,
0.4456, 0.0083, 0.0223, 0.4479, 0.0061, 0.0222, 0.4479, 0.0083, 0.0222, 0.4458, 0.01, 0.0168, 0.4458, 0.01, 0.019, 0.4479, 0.01, 0.0168, 0.4479, 0.01, 0.019, 0.4459, 0.01, 0.0223, 0.4453, 0.01, 0.0223, 0.4479, 0.01, 0.0222, 0.4501, 0.0061, 0.0222, 0.4501, 0.0083, 0.0222, 0.4523, 0.0061, 0.0222, 0.4523, 0.0083, 0.0222, 0.4501, 0.01, 0.0168, 0.4501, 0.01, 0.019, 0.4523, 0.01, 0.0168, 0.4523, 0.01, 0.019, 0.4501, 0.01, 0.0222, 0.4523, 0.01, 0.0222, 0.4232, 0.0061, 0.0342, 0.4237, 0.0061, 0.0364, 0.4232, 0.0083, 0.0342, 0.4237, 0.0083, 0.0364, 0.4245, 0.0061, 0.0385, 0.4245, 0.0083, 0.0385, 0.4252, 0.0061, 0.0395, 0.426, 0.0061, 0.0407, 0.4252, 0.0083, 0.0395, 0.426, 0.0083, 0.0407,
0.4232, 0.01, 0.0344, 0.4238, 0.01, 0.0366, 0.4262, 0.01, 0.0342, 0.4262, 0.01, 0.0363, 0.4244, 0.01, 0.0383, 0.4251, 0.01, 0.0393, 0.4262, 0.01, 0.041, 0.427, 0.0061, 0.0421, 0.427, 0.0083, 0.0421, 0.427, 0.01, 0.042, 0.4284, 0.01, 0.0342, 0.4284, 0.01, 0.0363, 0.4306, 0.01, 0.0342, 0.4306, 0.01, 0.0363, 0.4284, 0.01, 0.0385, 0.4284, 0.01, 0.0407, 0.4306, 0.01, 0.0385, 0.4306, 0.01, 0.0407, 0.4327, 0.01, 0.0342, 0.4327, 0.01, 0.0363, 0.4349, 0.01, 0.0342, 0.4349, 0.01, 0.0363, 0.4327, 0.01, 0.0385, 0.4327, 0.01, 0.0407, 0.4349, 0.01, 0.0385, 0.4349, 0.01, 0.0407, 0.4281, 0.0061, 0.0432, 0.4292, 0.0061, 0.0442, 0.4281, 0.0083, 0.0432, 0.4292, 0.0083, 0.0442,
0.4305, 0.0061, 0.0452, 0.4305, 0.0083, 0.0452, 0.4281, 0.01, 0.0432, 0.4292, 0.01, 0.0442, 0.4306, 0.01, 0.0429, 0.4302, 0.01, 0.045, 0.4318, 0.0061, 0.046, 0.4318, 0.0083, 0.046, 0.4329, 0.0061, 0.0467, 0.4329, 0.0083, 0.0467, 0.4349, 0.0061, 0.0476, 0.4349, 0.0083, 0.0476, 0.4327, 0.01, 0.0429, 0.432, 0.01, 0.0461, 0.4349, 0.01, 0.0429, 0.4349, 0.01, 0.045, 0.433, 0.01, 0.0468, 0.4346, 0.01, 0.0475, 0.4377, 0.0061, 0.0336, 0.4377, 0.0083, 0.0336, 0.4388, 0.0061, 0.0347, 0.4398, 0.0061, 0.0356, 0.4388, 0.0083, 0.0347, 0.4398, 0.0083, 0.0356, 0.4377, 0.01, 0.0337, 0.4371, 0.01, 0.0363, 0.4388, 0.01, 0.0347, 0.4398, 0.01, 0.0356, 0.4371, 0.01, 0.0385, 0.4371, 0.01, 0.0407,
0.4392, 0.01, 0.0385, 0.4392, 0.01, 0.0407, 0.4414, 0.0061, 0.0366, 0.4414, 0.0083, 0.0366, 0.4432, 0.0061, 0.0373, 0.4432, 0.0083, 0.0373, 0.4443, 0.0061, 0.0375, 0.4443, 0.0083, 0.0375, 0.4416, 0.01, 0.0367, 0.443, 0.01, 0.0372, 0.4414, 0.01, 0.0385, 0.4414, 0.01, 0.0407, 0.4439, 0.01, 0.0374, 0.4436, 0.01, 0.0407, 0.4366, 0.0061, 0.0481, 0.4376, 0.0061, 0.0484, 0.4366, 0.0083, 0.0481, 0.4376, 0.0083, 0.0484, 0.4392, 0.0061, 0.0486, 0.4392, 0.0083, 0.0486, 0.4371, 0.01, 0.0429, 0.4371, 0.01, 0.045, 0.4392, 0.01, 0.0429, 0.4392, 0.01, 0.045, 0.4368, 0.01, 0.0482, 0.4378, 0.01, 0.0484, 0.4392, 0.01, 0.0472, 0.4392, 0.01, 0.0486, 0.4414, 0.0061, 0.0487, 0.4414, 0.0083, 0.0487,
0.4436, 0.0061, 0.0487, 0.4436, 0.0083, 0.0487, 0.4414, 0.01, 0.0429, 0.4414, 0.01, 0.045, 0.4436, 0.01, 0.0429, 0.4436, 0.01, 0.045, 0.4414, 0.01, 0.0472, 0.4414, 0.01, 0.0487, 0.4436, 0.01, 0.0472, 0.4436, 0.01, 0.0487, 0.4457, 0.0061, 0.0378, 0.4457, 0.0083, 0.0378, 0.4479, 0.0061, 0.0379, 0.4479, 0.0083, 0.0379, 0.4459, 0.01, 0.0378, 0.4458, 0.01, 0.0407, 0.4479, 0.01, 0.0379, 0.4479, 0.01, 0.0407, 0.4501, 0.0061, 0.0379, 0.4501, 0.0083, 0.0379, 0.4523, 0.0061, 0.0379, 0.4523, 0.0083, 0.0379, 0.4501, 0.01, 0.0379, 0.4501, 0.01, 0.0407, 0.4523, 0.01, 0.0379, 0.4523, 0.01, 0.0407, 0.4458, 0.0061, 0.0487, 0.4458, 0.0083, 0.0487, 0.4479, 0.0061, 0.0487, 0.4479, 0.0083, 0.0487,
0.4458, 0.01, 0.0429, 0.4458, 0.01, 0.045, 0.4479, 0.01, 0.0429, 0.4479, 0.01, 0.045, 0.4458, 0.01, 0.0472, 0.4458, 0.01, 0.0487, 0.4479, 0.01, 0.0472, 0.4479, 0.01, 0.0487, 0.4501, 0.0061, 0.0487, 0.4501, 0.0083, 0.0487, 0.4523, 0.0061, 0.0487, 0.4523, 0.0083, 0.0487, 0.4501, 0.01, 0.0429, 0.4501, 0.01, 0.045, 0.4523, 0.01, 0.0429, 0.4523, 0.01, 0.045, 0.4501, 0.01, 0.0472, 0.4501, 0.01, 0.0487, 0.4523, 0.01, 0.0472, 0.4523, 0.01, 0.0487, 0.4544, 0.0061, 0.0105, 0.4544, 0.0083, 0.0105, 0.4566, 0.0061, 0.0105, 0.4566, 0.0083, 0.0105, 0.4544, 0.01, 0.0105, 0.4566, 0.01, 0.0105, 0.4544, 0.01, 0.0125, 0.4544, 0.01, 0.0146, 0.4566, 0.01, 0.0125, 0.4566, 0.01, 0.0146,
0.4588, 0.0061, 0.0105, 0.4588, 0.0083, 0.0105, 0.4609, 0.0061, 0.0105, 0.4609, 0.0083, 0.0105, 0.4588, 0.01, 0.0105, 0.4609, 0.01, 0.0105, 0.4588, 0.01, 0.0125, 0.4588, 0.01, 0.0146, 0.4609, 0.01, 0.0125, 0.4609, 0.01, 0.0146, 0.4631, 0.0061, 0.0105, 0.4631, 0.0083, 0.0105, 0.4653, 0.0061, 0.0105, 0.4653, 0.0083, 0.0105, 0.4631, 0.01, 0.0105, 0.4653, 0.01, 0.0105, 0.4631, 0.01, 0.0125, 0.4631, 0.01, 0.0146, 0.4653, 0.01, 0.0125, 0.4653, 0.01, 0.0146, 0.4675, 0.0061, 0.0105, 0.4675, 0.0083, 0.0105, 0.4696, 0.0061, 0.0105, 0.4696, 0.0083, 0.0105, 0.4675, 0.01, 0.0105, 0.4696, 0.01, 0.0105, 0.4675, 0.01, 0.0125, 0.4675, 0.01, 0.0146, 0.4696, 0.01, 0.0125, 0.4696, 0.01, 0.0146,
0.4544, 0.0061, 0.0222, 0.4544, 0.0083, 0.0222, 0.4566, 0.0061, 0.0222, 0.4566, 0.0083, 0.0222, 0.4544, 0.01, 0.0168, 0.4544, 0.01, 0.019, 0.4566, 0.01, 0.0168, 0.4566, 0.01, 0.019, 0.4544, 0.01, 0.0222, 0.4566, 0.01, 0.0222, 0.4588, 0.0061, 0.0222, 0.4588, 0.0083, 0.0222, 0.4609, 0.0061, 0.0222, 0.4609, 0.0083, 0.0222, 0.4588, 0.01, 0.0168, 0.4588, 0.01, 0.019, 0.4609, 0.01, 0.0168, 0.4609, 0.01, 0.019, 0.4588, 0.01, 0.0222, 0.4609, 0.01, 0.0222, 0.4631, 0.0061, 0.0222, 0.4631, 0.0083, 0.0222, 0.4653, 0.0061, 0.0222, 0.4653, 0.0083, 0.0222, 0.4631, 0.01, 0.0168, 0.4631, 0.01, 0.019, 0.4653, 0.01, 0.0168, 0.4653, 0.01, 0.019, 0.4631, 0.01, 0.0222, 0.4653, 0.01, 0.0222,
0.4675, 0.0061, 0.0222, 0.4675, 0.0083, 0.0222, 0.4696, 0.0061, 0.0222, 0.4696, 0.0083, 0.0222, 0.4675, 0.01, 0.0168, 0.4675, 0.01, 0.019, 0.4696, 0.01, 0.0168, 0.4696, 0.01, 0.019, 0.4675, 0.01, 0.0222, 0.4696, 0.01, 0.0222, 0.4718, 0.0061, 0.0105, 0.4718, 0.0083, 0.0105, 0.474, 0.0061, 0.0105, 0.474, 0.0083, 0.0105, 0.4718, 0.01, 0.0105, 0.474, 0.01, 0.0105, 0.4718, 0.01, 0.0125, 0.4718, 0.01, 0.0146, 0.474, 0.01, 0.0125, 0.474, 0.01, 0.0146, 0.4761, 0.0061, 0.0104, 0.4761, 0.0083, 0.0104, 0.4783, 0.0061, 0.0104, 0.4783, 0.0083, 0.0104, 0.4761, 0.01, 0.0104, 0.4783, 0.01, 0.0104, 0.4761, 0.01, 0.0125, 0.4761, 0.01, 0.0146, 0.4783, 0.01, 0.0125, 0.4783, 0.01, 0.0146,
0.4877, 0.0061, 0.0067, 0.4877, 0.0083, 0.0067, 0.4877, 0.01, 0.0067, 0.4805, 0.0061, 0.0102, 0.4805, 0.0083, 0.0102, 0.4836, 0.0061, 0.0092, 0.4826, 0.0061, 0.0096, 0.4836, 0.0083, 0.0092, 0.4826, 0.0083, 0.0096, 0.4804, 0.01, 0.0102, 0.4837, 0.01, 0.0092, 0.4826, 0.01, 0.0096, 0.4805, 0.01, 0.0125, 0.4805, 0.01, 0.0146, 0.4826, 0.01, 0.0125, 0.4826, 0.01, 0.0146, 0.4849, 0.0061, 0.0087, 0.4849, 0.0083, 0.0087, 0.4866, 0.0061, 0.0076, 0.4866, 0.0083, 0.0076, 0.4847, 0.01, 0.0087, 0.4848, 0.01, 0.0103, 0.4866, 0.01, 0.0076, 0.487, 0.01, 0.0103, 0.4848, 0.01, 0.0125, 0.4848, 0.01, 0.0146, 0.487, 0.01, 0.0125, 0.487, 0.01, 0.0146, 0.4718, 0.0061, 0.0222, 0.4718, 0.0083, 0.0222,
0.474, 0.0061, 0.0222, 0.474, 0.0083, 0.0222, 0.4718, 0.01, 0.0168, 0.4718, 0.01, 0.019, 0.474, 0.01, 0.0168, 0.474, 0.01, 0.019, 0.4718, 0.01, 0.0222, 0.474, 0.01, 0.0222, 0.4761, 0.0061, 0.0222, 0.4761, 0.0083, 0.0222, 0.4783, 0.0061, 0.0222, 0.4783, 0.0083, 0.0222, 0.4761, 0.01, 0.0168, 0.4761, 0.01, 0.019, 0.4783, 0.01, 0.0168, 0.4783, 0.01, 0.019, 0.4761, 0.01, 0.0222, 0.4783, 0.01, 0.0222, 0.4805, 0.0061, 0.0222, 0.4805, 0.0083, 0.0222, 0.4827, 0.0061, 0.0219, 0.4827, 0.0083, 0.0219, 0.4805, 0.01, 0.0168, 0.4805, 0.01, 0.019, 0.4826, 0.01, 0.0168, 0.4826, 0.01, 0.019, 0.4806, 0.01, 0.0221, 0.4828, 0.01, 0.0219, 0.4848, 0.0061, 0.0214, 0.4848, 0.0083, 0.0214,
0.487, 0.0061, 0.0208, 0.487, 0.0083, 0.0208, 0.4848, 0.01, 0.0168, 0.4848, 0.01, 0.019, 0.487, 0.01, 0.0168, 0.487, 0.01, 0.019, 0.485, 0.01, 0.0214, 0.4871, 0.01, 0.0208, 0.4544, 0.0061, 0.0379, 0.4544, 0.0083, 0.0379, 0.4566, 0.0061, 0.0379, 0.4566, 0.0083, 0.0379, 0.4544, 0.01, 0.0379, 0.4544, 0.01, 0.0407, 0.4566, 0.01, 0.0379, 0.4566, 0.01, 0.0407, 0.4588, 0.0061, 0.0379, 0.4588, 0.0083, 0.0379, 0.4609, 0.0061, 0.0379, 0.4609, 0.0083, 0.0379, 0.4588, 0.01, 0.0379, 0.4588, 0.01, 0.0407, 0.4609, 0.01, 0.0379, 0.4609, 0.01, 0.0407, 0.4544, 0.0061, 0.0487, 0.4544, 0.0083, 0.0487, 0.4566, 0.0061, 0.0487, 0.4566, 0.0083, 0.0487, 0.4544, 0.01, 0.0429, 0.4544, 0.01, 0.045,
0.4566, 0.01, 0.0429, 0.4566, 0.01, 0.045, 0.4544, 0.01, 0.0472, 0.4544, 0.01, 0.0487, 0.4566, 0.01, 0.0472, 0.4566, 0.01, 0.0487, 0.4588, 0.0061, 0.0487, 0.4588, 0.0083, 0.0487, 0.4609, 0.0061, 0.0487, 0.4609, 0.0083, 0.0487, 0.4588, 0.01, 0.0429, 0.4588, 0.01, 0.045, 0.4609, 0.01, 0.0429, 0.4609, 0.01, 0.045, 0.4588, 0.01, 0.0472, 0.4588, 0.01, 0.0487, 0.4609, 0.01, 0.0472, 0.4609, 0.01, 0.0487, 0.4631, 0.0061, 0.0379, 0.4631, 0.0083, 0.0379, 0.4653, 0.0061, 0.0379, 0.4653, 0.0083, 0.0379, 0.4631, 0.01, 0.0379, 0.4631, 0.01, 0.0407, 0.4653, 0.01, 0.0379, 0.4653, 0.01, 0.0407, 0.4675, 0.0061, 0.0379, 0.4675, 0.0083, 0.0379, 0.4696, 0.0061, 0.0379, 0.4696, 0.0083, 0.0379,
0.4675, 0.01, 0.0379, 0.4675, 0.01, 0.0407, 0.4696, 0.01, 0.0379, 0.4696, 0.01, 0.0407, 0.4631, 0.0061, 0.0487, 0.4631, 0.0083, 0.0487, 0.4653, 0.0061, 0.0487, 0.4653, 0.0083, 0.0487, 0.4631, 0.01, 0.0429, 0.4631, 0.01, 0.045, 0.4653, 0.01, 0.0429, 0.4653, 0.01, 0.045, 0.4631, 0.01, 0.0472, 0.4631, 0.01, 0.0487, 0.4653, 0.01, 0.0472, 0.4653, 0.01, 0.0487, 0.4675, 0.0061, 0.0487, 0.4675, 0.0083, 0.0487, 0.4696, 0.0061, 0.0487, 0.4696, 0.0083, 0.0487, 0.4675, 0.01, 0.0429, 0.4675, 0.01, 0.045, 0.4696, 0.01, 0.0429, 0.4696, 0.01, 0.045, 0.4675, 0.01, 0.0472, 0.4675, 0.01, 0.0487, 0.4696, 0.01, 0.0472, 0.4696, 0.01, 0.0487, 0.4718, 0.0061, 0.0379, 0.4718, 0.0083, 0.0379,
0.474, 0.0061, 0.0379, 0.474, 0.0083, 0.0379, 0.4718, 0.01, 0.0379, 0.4718, 0.01, 0.0407, 0.474, 0.01, 0.0379, 0.474, 0.01, 0.0407, 0.4761, 0.0061, 0.0379, 0.4761, 0.0083, 0.0379, 0.4783, 0.0061, 0.0379, 0.4783, 0.0083, 0.0379, 0.4761, 0.01, 0.0379, 0.4761, 0.01, 0.0407, 0.4783, 0.01, 0.0379, 0.4783, 0.01, 0.0407, 0.4718, 0.0061, 0.0487, 0.4718, 0.0083, 0.0487, 0.474, 0.0061, 0.0487, 0.474, 0.0083, 0.0487, 0.4718, 0.01, 0.0429, 0.4718, 0.01, 0.045, 0.474, 0.01, 0.0429, 0.474, 0.01, 0.045, 0.4718, 0.01, 0.0472, 0.4718, 0.01, 0.0487, 0.474, 0.01, 0.0472, 0.474, 0.01, 0.0487, 0.4761, 0.0061, 0.0487, 0.4761, 0.0083, 0.0487, 0.4783, 0.0061, 0.0487, 0.4783, 0.0083, 0.0487,
0.4761, 0.01, 0.0429, 0.4761, 0.01, 0.045, 0.4783, 0.01, 0.0429, 0.4783, 0.01, 0.045, 0.4761, 0.01, 0.0472, 0.4761, 0.01, 0.0487, 0.4783, 0.01, 0.0472, 0.4783, 0.01, 0.0487, 0.4805, 0.0061, 0.0379, 0.4805, 0.0083, 0.0379, 0.4826, 0.0061, 0.0379, 0.4826, 0.0083, 0.0379, 0.4805, 0.01, 0.0379, 0.4805, 0.01, 0.0407, 0.4826, 0.01, 0.0379, 0.4826, 0.01, 0.0407, 0.4848, 0.0061, 0.0379, 0.4848, 0.0083, 0.0379, 0.487, 0.0061, 0.0379, 0.487, 0.0083, 0.0379, 0.4848, 0.01, 0.0379, 0.4848, 0.01, 0.0407, 0.487, 0.01, 0.0379, 0.487, 0.01, 0.0407, 0.4805, 0.0061, 0.0487, 0.4805, 0.0083, 0.0487, 0.4826, 0.0061, 0.0487, 0.4826, 0.0083, 0.0487, 0.4805, 0.01, 0.0429, 0.4805, 0.01, 0.045,
0.4826, 0.01, 0.0429, 0.4826, 0.01, 0.045, 0.4805, 0.01, 0.0472, 0.4805, 0.01, 0.0487, 0.4826, 0.01, 0.0472, 0.4826, 0.01, 0.0487, 0.4848, 0.0061, 0.0487, 0.4848, 0.0083, 0.0487, 0.487, 0.0061, 0.0487, 0.487, 0.0083, 0.0487, 0.4848, 0.01, 0.0429, 0.4848, 0.01, 0.045, 0.487, 0.01, 0.0429, 0.487, 0.01, 0.045, 0.4848, 0.01, 0.0472, 0.4848, 0.01, 0.0487, 0.487, 0.01, 0.0472, 0.487, 0.01, 0.0487, 0.491, 0.0061, -0.0006, 0.4911, 0.0061, 0.0017, 0.491, 0.0083, -0.0006, 0.4911, 0.0083, 0.0017, 0.4898, 0.0061, 0.0042, 0.4888, 0.0061, 0.0057, 0.4898, 0.0083, 0.0042, 0.4888, 0.0083, 0.0057, 0.4905, 0.0061, 0.0032, 0.4905, 0.0083, 0.0032, 0.491, 0.01, -0.0004, 0.4911, 0.01, 0.0017,
0.4897, 0.01, 0.0043, 0.4888, 0.01, 0.0057, 0.4904, 0.01, 0.0033, 0.4913, 0.01, 0.006, 0.4935, 0.01, -0.0005, 0.4935, 0.01, 0.0016, 0.4957, 0.01, -0.0005, 0.4957, 0.01, 0.0016, 0.4935, 0.01, 0.0038, 0.4935, 0.01, 0.006, 0.4957, 0.01, 0.0038, 0.4957, 0.01, 0.006, 0.4892, 0.01, 0.0081, 0.4892, 0.01, 0.0103, 0.4913, 0.01, 0.0081, 0.4913, 0.01, 0.0103, 0.4892, 0.01, 0.0125, 0.4892, 0.01, 0.0146, 0.4913, 0.01, 0.0125, 0.4913, 0.01, 0.0146, 0.4965, 0.0061, 0.0155, 0.4965, 0.0083, 0.0155, 0.4935, 0.01, 0.0081, 0.4935, 0.01, 0.0103, 0.4957, 0.01, 0.0081, 0.4957, 0.01, 0.0103, 0.4935, 0.01, 0.0125, 0.4935, 0.01, 0.0146, 0.4957, 0.01, 0.0125, 0.4965, 0.01, 0.0154,
0.4978, 0.01, -0.0005, 0.4978, 0.01, 0.0016, 0.5, 0.01, -0.0005, 0.5, 0.01, 0.0016, 0.4978, 0.01, 0.0038, 0.4978, 0.01, 0.006, 0.5, 0.01, 0.0038, 0.5, 0.01, 0.006, 0.502, 0.0061, -0.0005, 0.502, 0.0061, 0.0016, 0.502, 0.0083, -0.0005, 0.502, 0.0083, 0.0016, 0.5019, 0.0061, 0.0038, 0.5017, 0.0061, 0.006, 0.5019, 0.0083, 0.0038, 0.5017, 0.0083, 0.006, 0.502, 0.01, -0.0005, 0.502, 0.01, 0.0016, 0.5019, 0.01, 0.0038, 0.5017, 0.01, 0.006, 0.5009, 0.0061, 0.0087, 0.5002, 0.0061, 0.0103, 0.5009, 0.0083, 0.0087, 0.5002, 0.0083, 0.0103, 0.4985, 0.0061, 0.013, 0.4974, 0.0061, 0.0144, 0.4985, 0.0083, 0.013, 0.4974, 0.0083, 0.0144, 0.4992, 0.0061, 0.0119, 0.4992, 0.0083, 0.0119,
0.4978, 0.01, 0.0081, 0.4978, 0.01, 0.0103, 0.501, 0.01, 0.0085, 0.5001, 0.01, 0.0106, 0.4986, 0.01, 0.013, 0.4975, 0.01, 0.0143, 0.4993, 0.01, 0.0118, 0.5013, 0.0061, 0.0076, 0.5013, 0.0083, 0.0076, 0.5013, 0.01, 0.0075, 0.4896, 0.0061, 0.0198, 0.4896, 0.0083, 0.0198, 0.4914, 0.0061, 0.0191, 0.4914, 0.0083, 0.0191, 0.4886, 0.0061, 0.0203, 0.4886, 0.0083, 0.0203, 0.4892, 0.01, 0.0168, 0.4895, 0.01, 0.0199, 0.4913, 0.01, 0.0168, 0.4916, 0.01, 0.019, 0.4884, 0.01, 0.0203, 0.494, 0.0061, 0.0175, 0.4929, 0.0061, 0.0182, 0.494, 0.0083, 0.0175, 0.4929, 0.0083, 0.0182, 0.4954, 0.0061, 0.0165, 0.4954, 0.0083, 0.0165, 0.4939, 0.01, 0.0176, 0.4928, 0.01, 0.0182, 0.4954, 0.01, 0.0165,
0.4892, 0.0061, 0.0379, 0.4892, 0.0083, 0.0379, 0.4913, 0.0061, 0.0379, 0.4913, 0.0083, 0.0379, 0.4892, 0.01, 0.0379, 0.4892, 0.01, 0.0407, 0.4913, 0.01, 0.0379, 0.4913, 0.01, 0.0407, 0.4935, 0.0061, 0.0379, 0.4935, 0.0083, 0.0379, 0.4957, 0.0061, 0.0379, 0.4957, 0.0083, 0.0379, 0.4935, 0.01, 0.0379, 0.4935, 0.01, 0.0407, 0.4957, 0.01, 0.0379, 0.4957, 0.01, 0.0407, 0.4892, 0.0061, 0.0487, 0.4892, 0.0083, 0.0487, 0.4913, 0.0061, 0.0487, 0.4913, 0.0083, 0.0487, 0.4892, 0.01, 0.0429, 0.4892, 0.01, 0.045, 0.4913, 0.01, 0.0429, 0.4913, 0.01, 0.045, 0.4892, 0.01, 0.0472, 0.4892, 0.01, 0.0487, 0.4913, 0.01, 0.0472, 0.4913, 0.01, 0.0487, 0.4935, 0.0061, 0.0487, 0.4935, 0.0083, 0.0487,
0.4957, 0.0061, 0.0487, 0.4957, 0.0083, 0.0487, 0.4935, 0.01, 0.0429, 0.4935, 0.01, 0.045, 0.4957, 0.01, 0.0429, 0.4957, 0.01, 0.045, 0.4935, 0.01, 0.0472, 0.4935, 0.01, 0.0487, 0.4957, 0.01, 0.0472, 0.4957, 0.01, 0.0487, 0.498, 0.0061, 0.0379, 0.498, 0.0061, 0.0407, 0.498, 0.0083, 0.0379, 0.498, 0.0083, 0.0407, 0.498, 0.01, 0.0379, 0.498, 0.01, 0.0407, 0.498, 0.0061, 0.0429, 0.498, 0.0061, 0.045, 0.498, 0.0083, 0.0429, 0.498, 0.0083, 0.045, 0.498, 0.0061, 0.0472, 0.498, 0.0061, 0.0487, 0.498, 0.0083, 0.0472, 0.498, 0.0083, 0.0487, 0.498, 0.01, 0.0429, 0.498, 0.01, 0.045, 0.498, 0.01, 0.0472, 0.498, 0.01, 0.0487});
    
    _trix = std::vector<int>({9, 10, 8, 7, 9, 6, 4, 9, 8, 6, 5, 3, 14, 17, 15, 19, 20, 18, 23, 24, 22, 24, 26, 22, 30, 29, 28, 32, 35, 33, 43, 40, 42, 43, 38, 41, 12, 29, 13, 32, 17, 16, 20, 36, 18, 13, 18, 12, 14, 23, 22, 22, 16, 14, 41, 34, 40, 35, 38, 33, 26, 21, 19, 31, 32, 29, 16, 19, 13, 13, 32, 16, 18, 28, 12, 2, 23, 1, 0, 23, 15, 3, 30, 28, 36, 7, 6, 3, 36, 6, 52, 53, 51, 50, 52, 49, 58, 55, 57, 46, 51, 45, 49, 46, 44, 57, 47, 48, 44, 48, 47, 55, 50, 49, 44, 55, 49, 60, 63, 61, 67, 68, 66, 68, 70, 66, 61, 66, 60, 66, 62, 60, 70, 65, 64, 59, 70, 64, 74, 77, 75, 85, 82, 84, 85, 80, 83, 92, 95, 93, 101, 102, 100, 102, 104, 100, 97, 82, 96, 89, 101, 88, 83, 76, 82, 77, 80, 75, 93, 100, 92, 92, 104, 94, 87, 80, 81, 104, 99, 98,
86, 75, 80, 94, 98, 91, 90, 82, 76, 79, 101, 93, 106, 109, 107, 114, 115, 113, 115, 117, 113, 122, 119, 121, 124, 119, 123, 125, 128, 126, 133, 130, 132, 125, 109, 108, 135, 123, 134, 106, 114, 113, 113, 108, 106, 132, 126, 128, 120, 110, 119, 123, 110, 112, 109, 110, 107, 135, 130, 131, 130, 129, 126, 109, 129, 112, 134, 112, 129, 45, 75, 46, 87, 57, 86, 46, 78, 48, 86, 48, 78, 110, 63, 62, 65, 114, 64, 122, 71, 120, 107, 62, 59, 64, 107, 59, 111, 71, 63, 117, 89, 88, 95, 125, 93, 79, 117, 88, 79, 125, 108, 139, 140, 138, 148, 145, 147, 150, 145, 149, 142, 137, 136, 140, 145, 138, 148, 139, 146, 136, 139, 138, 145, 143, 142, 144, 145, 142, 136, 145, 142, 143, 139, 137, 61, 144, 142, 145, 69, 67, 150, 71, 73, 142, 67, 61, 149, 63, 71, 51, 2, 1, 45, 1, 0, 17, 74, 15, 97, 38, 39, 96, 33, 38, 17, 90, 76, 0, 74, 45,
157, 158, 156, 155, 157, 154, 165, 164, 163, 170, 171, 169, 168, 170, 167, 152, 164, 153, 158, 173, 156, 161, 170, 169, 167, 162, 160, 152, 157, 156, 154, 153, 151, 160, 165, 163, 173, 168, 167, 160, 173, 167, 156, 163, 152, 178, 181, 179, 183, 184, 182, 184, 186, 182, 177, 183, 176, 179, 182, 178, 182, 180, 178, 175, 183, 179, 194, 195, 193, 192, 194, 191, 202, 201, 200, 207, 208, 206, 205, 207, 204, 189, 201, 190, 195, 210, 193, 198, 207, 206, 204, 199, 197, 190, 193, 189, 188, 194, 190, 197, 202, 200, 210, 205, 204, 200, 204, 197, 189, 210, 200, 214, 213, 212, 218, 217, 216, 224, 223, 222, 228, 227, 226, 222, 215, 214, 226, 219, 218, 215, 216, 213, 225, 226, 223, 223, 218, 215, 230, 233, 231, 236, 235, 234, 239, 240, 238, 244, 243, 242, 248, 247, 246, 242, 233, 232, 246, 237, 236, 238, 232, 230, 233, 234, 231, 245, 246, 243, 243, 236, 233, 252, 251, 250, 256, 255, 254, 260, 259, 258,
264, 263, 262, 258, 253, 252, 262, 257, 256, 253, 254, 251, 261, 262, 259, 259, 256, 253, 198, 231, 199, 234, 203, 202, 208, 238, 206, 231, 202, 199, 206, 230, 198, 250, 225, 224, 254, 229, 228, 251, 228, 225, 188, 214, 212, 220, 192, 191, 203, 222, 201, 212, 191, 188, 201, 214, 190, 237, 250, 235, 249, 258, 247, 247, 252, 237, 235, 224, 203, 274, 273, 272, 279, 280, 278, 277, 279, 276, 284, 287, 285, 290, 289, 288, 293, 294, 292, 270, 285, 271, 288, 275, 274, 280, 292, 278, 270, 279, 278, 269, 279, 271, 292, 286, 284, 269, 274, 272, 282, 277, 276, 287, 288, 285, 272, 276, 269, 285, 274, 271, 278, 284, 270, 299, 300, 298, 302, 305, 303, 307, 308, 306, 308, 310, 306, 301, 307, 300, 298, 297, 296, 303, 306, 302, 302, 310, 304, 297, 307, 303, 314, 313, 312, 318, 317, 316, 322, 321, 320, 326, 325, 324, 320, 315, 314, 324, 319, 318, 315, 316, 313, 323, 324, 321, 321, 318, 315, 268, 298, 267,
267, 296, 266, 312, 287, 286, 316, 291, 290, 313, 290, 287, 294, 268, 267, 266, 294, 267, 296, 314, 312, 305, 320, 303, 297, 320, 314, 266, 312, 286, 330, 329, 328, 334, 333, 332, 338, 337, 336, 342, 341, 340, 336, 331, 330, 340, 335, 334, 331, 332, 329, 339, 340, 337, 337, 334, 331, 346, 345, 344, 348, 351, 349, 357, 354, 356, 357, 352, 355, 358, 361, 359, 368, 365, 367, 368, 363, 366, 358, 347, 346, 370, 354, 369, 366, 360, 365, 361, 363, 359, 354, 351, 350, 351, 352, 349, 347, 348, 345, 370, 363, 364, 369, 359, 363, 347, 362, 350, 369, 350, 362, 373, 372, 371, 377, 376, 375, 383, 382, 381, 391, 388, 390, 393, 388, 392, 381, 374, 373, 377, 386, 378, 391, 379, 389, 389, 385, 388, 392, 385, 387, 374, 375, 372, 382, 387, 385, 382, 377, 374, 386, 379, 378, 399, 396, 398, 398, 394, 395, 371, 339, 338, 375, 343, 342, 372, 342, 339, 397, 365, 396, 396, 360, 394, 335, 344, 333, 343, 358, 341,
341, 346, 335, 376, 395, 394, 399, 379, 380, 379, 395, 378, 343, 394, 360, 180, 273, 181, 187, 282, 186, 186, 272, 180, 328, 245, 244, 332, 249, 248, 329, 248, 245, 344, 261, 260, 348, 265, 264, 345, 264, 261, 333, 260, 249, 240, 177, 176, 181, 242, 179, 232, 176, 175, 175, 242, 232, 275, 328, 273, 291, 336, 289, 289, 330, 275, 319, 371, 317, 327, 381, 325, 325, 373, 319, 317, 338, 291, 273, 244, 181, 161, 213, 162, 216, 166, 165, 171, 220, 169, 213, 165, 162, 169, 212, 161, 403, 406, 404, 408, 409, 407, 409, 411, 407, 402, 408, 401, 403, 408, 407, 407, 405, 403, 401, 404, 400, 416, 417, 415, 415, 414, 413, 424, 423, 422, 429, 430, 428, 427, 429, 426, 436, 435, 434, 440, 439, 438, 420, 435, 421, 438, 425, 424, 430, 442, 428, 421, 428, 420, 419, 429, 421, 419, 424, 422, 432, 427, 426, 437, 438, 435, 422, 426, 419, 435, 424, 421, 420, 442, 434, 405, 423, 406, 412, 432, 411, 411, 422, 405,
413, 436, 434, 442, 416, 415, 413, 442, 415, 444, 447, 445, 450, 449, 448, 453, 454, 452, 454, 456, 452, 462, 461, 460, 466, 465, 464, 460, 447, 446, 464, 451, 450, 445, 452, 444, 444, 456, 446, 445, 450, 448, 458, 455, 453, 463, 464, 461, 448, 453, 445, 461, 450, 447, 470, 469, 468, 479, 476, 478, 488, 485, 487, 490, 485, 489, 492, 493, 491, 470, 481, 471, 488, 474, 486, 486, 480, 485, 489, 480, 482, 478, 472, 473, 491, 484, 483, 469, 473, 472, 479, 474, 475, 478, 471, 474, 481, 474, 471, 497, 496, 495, 499, 502, 500, 506, 503, 505, 509, 508, 507, 511, 514, 512, 520, 517, 519, 520, 515, 518, 507, 498, 497, 501, 512, 502, 516, 505, 515, 505, 500, 502, 518, 513, 517, 518, 512, 514, 498, 499, 496, 510, 511, 508, 508, 501, 498, 515, 502, 512, 522, 525, 523, 529, 530, 528, 530, 532, 528, 535, 538, 536, 542, 543, 541, 524, 536, 525, 533, 541, 532, 523, 528, 522, 522, 532, 524, 541, 537, 535,
543, 540, 539, 537, 539, 534, 532, 535, 524, 495, 463, 462, 499, 467, 466, 496, 466, 463, 527, 489, 526, 494, 529, 493, 526, 482, 521, 493, 523, 484, 451, 468, 449, 465, 482, 480, 465, 470, 451, 527, 503, 504, 526, 500, 503, 467, 521, 482, 545, 548, 546, 550, 551, 549, 551, 553, 549, 560, 559, 558, 565, 566, 564, 563, 565, 562, 547, 559, 548, 554, 568, 553, 556, 565, 564, 555, 565, 557, 546, 549, 545, 549, 547, 545, 555, 560, 558, 568, 563, 562, 558, 562, 555, 553, 558, 547, 572, 571, 570, 576, 575, 574, 580, 579, 578, 584, 583, 582, 578, 573, 572, 582, 577, 576, 573, 574, 571, 581, 582, 579, 579, 576, 573, 588, 587, 586, 592, 591, 590, 598, 597, 596, 602, 601, 600, 596, 589, 588, 600, 593, 592, 589, 590, 587, 599, 600, 597, 597, 592, 589, 606, 605, 604, 610, 609, 608, 614, 613, 612, 618, 617, 616, 612, 607, 606, 616, 611, 610, 607, 608, 605, 615, 616, 613, 613, 610, 607, 556, 587, 557,
590, 561, 560, 566, 594, 564, 587, 560, 557, 556, 594, 586, 604, 581, 580, 608, 585, 584, 605, 584, 581, 548, 570, 546, 561, 578, 559, 559, 572, 548, 593, 604, 591, 603, 612, 601, 601, 606, 593, 591, 580, 561, 622, 621, 620, 629, 626, 628, 630, 633, 631, 638, 635, 637, 630, 623, 622, 640, 628, 639, 633, 635, 631, 628, 624, 625, 621, 625, 624, 640, 635, 636, 639, 631, 635, 623, 634, 625, 634, 628, 625, 645, 644, 643, 648, 649, 647, 655, 654, 653, 658, 659, 657, 653, 646, 645, 650, 657, 649, 647, 642, 641, 657, 652, 651, 641, 645, 643, 651, 655, 653, 642, 653, 645, 649, 651, 642, 661, 664, 662, 668, 665, 667, 671, 670, 669, 676, 673, 675, 663, 670, 664, 674, 667, 673, 667, 662, 664, 675, 670, 672, 673, 664, 670, 681, 680, 679, 684, 685, 683, 691, 690, 689, 694, 695, 693, 689, 682, 681, 686, 693, 685, 683, 678, 677, 687, 695, 688, 678, 679, 677, 687, 691, 689, 678, 689, 681, 685, 687, 678,
632, 662, 633, 666, 637, 665, 665, 633, 662, 679, 656, 655, 660, 683, 659, 652, 679, 655, 659, 677, 652, 418, 550, 417, 417, 546, 414, 570, 437, 436, 574, 441, 440, 571, 440, 437, 414, 570, 436, 620, 510, 509, 627, 517, 626, 510, 624, 513, 626, 513, 624, 643, 538, 537, 540, 647, 539, 534, 643, 537, 534, 647, 641, 456, 402, 401, 406, 460, 404, 446, 401, 400, 400, 460, 446, 425, 495, 423, 441, 507, 439, 439, 497, 425, 423, 462, 406, 577, 620, 575, 585, 630, 583, 583, 622, 577, 611, 661, 609, 619, 669, 617, 617, 663, 611, 609, 632, 585, 575, 509, 441, 697, 700, 698, 703, 702, 701, 706, 707, 705, 707, 709, 705, 713, 712, 711, 717, 716, 715, 699, 712, 700, 715, 704, 703, 710, 719, 709, 698, 705, 697, 705, 699, 697, 700, 701, 698, 714, 715, 712, 712, 703, 700, 709, 711, 699, 729, 726, 728, 731, 726, 730, 737, 734, 736, 735, 730, 734, 727, 723, 726, 730, 723, 725, 736, 732, 733, 734, 725, 732,
740, 739, 738, 742, 745, 743, 749, 746, 748, 752, 751, 750, 756, 755, 754, 761, 758, 760, 750, 741, 740, 744, 755, 745, 759, 748, 758, 748, 743, 745, 760, 755, 757, 741, 742, 739, 753, 754, 751, 751, 744, 741, 758, 745, 755, 738, 714, 713, 717, 743, 718, 747, 721, 746, 739, 717, 714, 743, 721, 718, 702, 725, 723, 716, 733, 732, 737, 721, 722, 736, 718, 721, 704, 732, 725, 763, 766, 764, 770, 771, 769, 771, 773, 769, 779, 778, 777, 782, 783, 781, 777, 766, 765, 768, 781, 767, 781, 776, 775, 763, 770, 769, 763, 773, 765, 773, 768, 767, 775, 779, 777, 765, 767, 762, 762, 777, 765, 762, 781, 775, 793, 794, 792, 794, 790, 792, 797, 798, 796, 802, 801, 800, 806, 805, 804, 800, 787, 786, 804, 789, 788, 787, 792, 786, 786, 790, 785, 788, 797, 796, 796, 795, 793, 803, 804, 801, 787, 796, 793, 801, 788, 787, 813, 812, 811, 818, 819, 817, 816, 818, 815, 821, 824, 822, 827, 826, 825, 830, 831, 829,
809, 822, 810, 825, 814, 813, 819, 829, 817, 810, 817, 809, 808, 818, 810, 829, 823, 821, 808, 813, 811, 824, 825, 822, 822, 813, 810, 809, 829, 821, 835, 834, 833, 839, 838, 837, 843, 842, 841, 847, 846, 845, 841, 836, 835, 845, 840, 839, 836, 837, 834, 844, 845, 842, 842, 839, 836, 811, 780, 779, 784, 815, 783, 776, 811, 779, 783, 808, 776, 833, 803, 802, 837, 807, 806, 834, 806, 803, 766, 785, 764, 790, 772, 770, 780, 800, 778, 785, 770, 764, 766, 800, 786, 814, 833, 812, 828, 841, 826, 826, 835, 814, 812, 802, 780, 851, 850, 849, 855, 854, 853, 860, 857, 859, 863, 862, 861, 867, 866, 865, 872, 869, 871, 861, 852, 851, 865, 856, 855, 870, 859, 869, 856, 857, 854, 868, 869, 866, 852, 853, 850, 864, 865, 862, 862, 855, 852, 866, 859, 856, 875, 874, 873, 877, 880, 878, 885, 884, 883, 889, 888, 887, 883, 876, 875, 887, 880, 879, 876, 877, 874, 886, 887, 884, 884, 879, 876, 896, 893, 895,
902, 899, 901, 900, 895, 899, 895, 891, 892, 898, 899, 897, 897, 895, 892, 873, 864, 863, 877, 868, 867, 882, 871, 881, 874, 867, 864, 881, 868, 878, 880, 891, 878, 893, 882, 881, 890, 897, 888, 878, 893, 881, 888, 892, 880, 910, 907, 909, 914, 911, 913, 907, 904, 903, 911, 906, 905, 913, 908, 910, 908, 905, 904, 917, 916, 915, 921, 920, 919, 928, 925, 927, 924, 925, 923, 932, 929, 931, 926, 917, 925, 915, 925, 917, 929, 922, 921, 918, 919, 916, 931, 926, 928, 926, 921, 918, 924, 912, 914, 923, 906, 912, 849, 753, 752, 756, 854, 757, 858, 760, 857, 850, 756, 753, 854, 760, 757, 903, 824, 823, 905, 828, 827, 832, 907, 831, 904, 827, 824, 831, 903, 823, 843, 916, 844, 919, 848, 847, 916, 847, 844, 906, 843, 828, 935, 934, 933, 939, 938, 937, 943, 942, 941, 947, 946, 945, 941, 936, 935, 945, 940, 939, 936, 937, 934, 944, 945, 942, 942, 939, 936, 951, 950, 949, 955, 954, 953, 959, 958, 957,
963, 962, 961, 957, 952, 951, 961, 956, 955, 952, 953, 950, 960, 961, 958, 958, 955, 952, 967, 966, 965, 971, 970, 969, 975, 974, 973, 979, 978, 977, 973, 968, 967, 977, 972, 971, 968, 969, 966, 976, 977, 974, 974, 971, 968, 983, 982, 981, 987, 986, 985, 991, 990, 989, 995, 994, 993, 989, 984, 983, 993, 988, 987, 984, 985, 982, 992, 993, 990, 990, 987, 984, 965, 944, 943, 969, 948, 947, 966, 947, 944, 981, 960, 959, 985, 964, 963, 982, 963, 960, 940, 949, 938, 948, 957, 946, 946, 951, 940, 972, 981, 970, 980, 989, 978, 978, 983, 972, 970, 959, 948, 999, 998, 997, 1004, 1001, 1003, 1007, 1006, 1005, 1013, 1010, 1012, 1005, 1000, 999, 1011, 1003, 1010, 1000, 1001, 998, 1012, 1006, 1008, 1014, 1013, 1012, 1008, 1014, 1012, 1006, 1003, 1000, 1020, 1019, 1018, 1023, 1024, 1022, 1030, 1029, 1028, 1033, 1034, 1032, 1028, 1021, 1020, 1025, 1032, 1024, 1016, 1024, 1017, 1026, 1034, 1027, 1017, 1018, 1016, 1027, 1028, 1026, 1026, 1020, 1017, 1017, 1032, 1026,
1038, 1037, 1036, 1045, 1042, 1044, 1048, 1047, 1046, 1055, 1052, 1054, 1046, 1039, 1038, 1053, 1044, 1052, 1041, 1042, 1040, 1051, 1052, 1050, 1039, 1040, 1037, 1049, 1050, 1047, 1047, 1041, 1039, 1050, 1044, 1041, 1058, 1061, 1059, 1063, 1064, 1062, 1070, 1069, 1068, 1073, 1074, 1072, 1068, 1061, 1060, 1067, 1072, 1066, 1062, 1057, 1056, 1068, 1074, 1070, 1057, 1058, 1056, 1067, 1064, 1065, 1064, 1060, 1057, 1066, 1068, 1060, 1007, 1037, 1008, 1043, 1014, 1042, 1037, 1009, 1008, 1042, 1009, 1040, 1058, 1031, 1030, 1035, 1062, 1034, 1056, 1030, 1027, 1027, 1062, 1056, 1081, 1080, 1079, 1086, 1083, 1085, 1086, 1087, 1084, 1093, 1092, 1091, 1096, 1097, 1095, 1091, 1082, 1081, 1088, 1095, 1087, 1083, 1077, 1076, 1084, 1078, 1077, 1089, 1097, 1090, 1078, 1079, 1077, 1090, 1091, 1089, 1089, 1081, 1078, 1078, 1095, 1089, 1101, 1100, 1099, 1105, 1104, 1103, 1109, 1108, 1107, 1113, 1112, 1111, 1107, 1102, 1101, 1111, 1106, 1105, 1102, 1103, 1100, 1110, 1111, 1108, 1108, 1105, 1102, 1118, 1117, 1116, 1123, 1124, 1122, 1128, 1127, 1126, 1135, 1132, 1134, 1131, 1132, 1130,
1126, 1119, 1118, 1125, 1130, 1124, 1122, 1118, 1116, 1132, 1129, 1128, 1126, 1132, 1128, 1123, 1120, 1121, 1120, 1116, 1115, 1118, 1130, 1126, 1138, 1137, 1136, 1142, 1141, 1140, 1146, 1145, 1144, 1150, 1149, 1148, 1144, 1139, 1138, 1148, 1143, 1142, 1139, 1140, 1137, 1147, 1148, 1145, 1145, 1142, 1139, 1093, 1117, 1094, 1098, 1120, 1097, 1115, 1093, 1090, 1097, 1115, 1090, 1136, 1110, 1109, 1140, 1114, 1113, 1137, 1113, 1110, 1082, 1099, 1080, 1094, 1107, 1092, 1092, 1101, 1082, 1119, 1136, 1117, 1127, 1146, 1144, 1127, 1138, 1119, 1117, 1109, 1094, 1154, 1153, 1152, 1158, 1157, 1156, 1161, 1162, 1160, 1165, 1161, 1164, 1168, 1167, 1166, 1170, 1173, 1171, 1166, 1155, 1154, 1170, 1159, 1158, 1175, 1164, 1174, 1156, 1161, 1160, 1164, 1157, 1159, 1153, 1158, 1156, 1169, 1170, 1167, 1167, 1158, 1155, 1171, 1164, 1159, 1178, 1177, 1176, 1185, 1182, 1184, 1181, 1182, 1180, 1193, 1190, 1192, 1195, 1196, 1194, 1185, 1194, 1183, 1187, 1190, 1186, 1182, 1179, 1178, 1176, 1182, 1178, 1188, 1196, 1189, 1179, 1194, 1188, 1200, 1199, 1198, 1204, 1203, 1202, 1208, 1207, 1206,
1212, 1211, 1210, 1206, 1201, 1200, 1210, 1205, 1204, 1201, 1202, 1199, 1209, 1210, 1207, 1207, 1204, 1201, 1220, 1221, 1219, 1221, 1217, 1219, 1225, 1224, 1223, 1231, 1228, 1230, 1223, 1216, 1215, 1229, 1220, 1228, 1215, 1220, 1219, 1219, 1214, 1215, 1226, 1228, 1224, 1232, 1231, 1230, 1226, 1232, 1230, 1224, 1220, 1216, 1198, 1169, 1168, 1202, 1173, 1172, 1199, 1172, 1169, 1218, 1192, 1217, 1214, 1192, 1187, 1173, 1186, 1171, 1190, 1175, 1174, 1171, 1190, 1174, 1205, 1214, 1203, 1213, 1223, 1211, 1205, 1223, 1215, 1203, 1187, 1173, 1076, 976, 975, 1079, 980, 979, 976, 1079, 979, 1099, 992, 991, 1103, 996, 995, 1100, 995, 992, 1080, 991, 980, 1152, 1049, 1048, 1162, 1054, 1160, 1153, 1051, 1049, 1160, 1051, 1156, 1176, 1071, 1070, 1075, 1180, 1074, 1070, 1180, 1176, 956, 997, 954, 964, 1005, 962, 962, 999, 956, 988, 1036, 986, 996, 1046, 994, 994, 1038, 988, 986, 1007, 964, 1106, 1152, 1104, 1114, 1166, 1112, 1112, 1154, 1106, 1143, 1198, 1141, 1151, 1206, 1149, 1149, 1200, 1143, 1141, 1168, 1114, 1104, 1048, 996, 1236, 1235, 1234,
1240, 1239, 1238, 1244, 1243, 1242, 1248, 1247, 1246, 1242, 1237, 1236, 1246, 1241, 1240, 1237, 1238, 1235, 1245, 1246, 1243, 1243, 1240, 1237, 1252, 1251, 1250, 1255, 1256, 1254, 1259, 1255, 1258, 1262, 1261, 1260, 1268, 1265, 1267, 1260, 1253, 1252, 1266, 1258, 1265, 1250, 1255, 1254, 1258, 1251, 1253, 1267, 1261, 1263, 1269, 1268, 1267, 1263, 1269, 1267, 1261, 1258, 1253, 1273, 1272, 1271, 1277, 1276, 1275, 1281, 1280, 1279, 1285, 1284, 1283, 1279, 1274, 1273, 1283, 1278, 1277, 1274, 1275, 1272, 1282, 1283, 1280, 1280, 1277, 1274, 1289, 1288, 1287, 1297, 1298, 1296, 1298, 1294, 1296, 1302, 1301, 1300, 1306, 1305, 1304, 1311, 1308, 1310, 1300, 1290, 1289, 1304, 1293, 1292, 1309, 1297, 1308, 1293, 1296, 1292, 1292, 1294, 1291, 1307, 1308, 1305, 1290, 1291, 1288, 1303, 1304, 1301, 1290, 1304, 1292, 1305, 1297, 1293, 1271, 1245, 1244, 1275, 1249, 1248, 1272, 1248, 1245, 1262, 1288, 1263, 1295, 1269, 1294, 1288, 1264, 1263, 1291, 1269, 1264, 1239, 1252, 1250, 1249, 1260, 1247, 1247, 1252, 1241, 1278, 1287, 1276, 1286, 1300, 1284, 1284, 1289, 1278, 1276, 1262, 1249,
1314, 1313, 1312, 1318, 1317, 1316, 1320, 1323, 1321, 1326, 1325, 1324, 1320, 1315, 1314, 1324, 1319, 1318, 1315, 1316, 1313, 1323, 1324, 1321, 1321, 1318, 1315, 1332, 1331, 1330, 1334, 1337, 1335, 1342, 1341, 1340, 1346, 1345, 1344, 1340, 1333, 1332, 1344, 1337, 1336, 1333, 1334, 1331, 1343, 1344, 1341, 1341, 1336, 1333, 1350, 1349, 1348, 1354, 1353, 1352, 1361, 1358, 1360, 1357, 1358, 1356, 1364, 1367, 1365, 1369, 1370, 1368, 1364, 1355, 1354, 1361, 1368, 1359, 1358, 1351, 1350, 1348, 1358, 1350, 1362, 1370, 1363, 1349, 1354, 1352, 1363, 1364, 1362, 1373, 1370, 1371, 1370, 1366, 1363, 1362, 1354, 1351, 1351, 1368, 1362, 1376, 1375, 1374, 1380, 1379, 1378, 1384, 1383, 1382, 1388, 1387, 1386, 1382, 1377, 1376, 1386, 1381, 1380, 1377, 1378, 1375, 1385, 1386, 1383, 1383, 1380, 1377, 1348, 1323, 1322, 1352, 1327, 1326, 1329, 1356, 1328, 1349, 1326, 1323, 1322, 1356, 1348, 1374, 1343, 1342, 1378, 1347, 1346, 1375, 1346, 1343, 1319, 1330, 1317, 1327, 1340, 1325, 1325, 1332, 1319, 1355, 1374, 1353, 1367, 1382, 1365, 1365, 1376, 1355, 1353, 1342, 1327, 1395, 1392, 1394,
1398, 1397, 1396, 1401, 1402, 1400, 1405, 1401, 1404, 1402, 1394, 1400, 1391, 1392, 1390, 1396, 1401, 1400, 1399, 1401, 1397, 1396, 1394, 1391, 1409, 1410, 1408, 1408, 1407, 1406, 1412, 1415, 1413, 1423, 1420, 1422, 1426, 1425, 1424, 1430, 1429, 1428, 1433, 1434, 1432, 1437, 1433, 1436, 1424, 1415, 1414, 1434, 1422, 1432, 1417, 1420, 1416, 1428, 1433, 1432, 1431, 1433, 1429, 1415, 1416, 1413, 1420, 1419, 1418, 1425, 1430, 1428, 1413, 1420, 1418, 1425, 1417, 1415, 1428, 1422, 1417, 1441, 1442, 1440, 1447, 1448, 1446, 1443, 1446, 1442, 1440, 1439, 1438, 1444, 1448, 1445, 1439, 1446, 1444, 1412, 1399, 1398, 1419, 1404, 1418, 1413, 1404, 1399, 1411, 1440, 1410, 1410, 1438, 1407, 1312, 1282, 1281, 1316, 1286, 1285, 1313, 1285, 1282, 1330, 1303, 1302, 1334, 1307, 1306, 1339, 1310, 1338, 1331, 1306, 1303, 1338, 1307, 1335, 1317, 1302, 1286, 1337, 1390, 1335, 1392, 1339, 1338, 1345, 1398, 1396, 1335, 1392, 1338, 1345, 1391, 1337, 1381, 1412, 1379, 1389, 1424, 1387, 1387, 1414, 1381, 1379, 1398, 1347, 933, 599, 598, 937, 603, 602, 934, 602, 599, 949, 615, 614,
953, 619, 618, 950, 618, 615, 938, 614, 603, 997, 672, 671, 1002, 675, 1001, 998, 675, 672, 1018, 692, 691, 696, 1022, 695, 1016, 691, 688, 688, 1022, 1016, 954, 671, 619, 1234, 886, 885, 1238, 890, 889, 1235, 889, 886, 1256, 901, 1254, 1254, 898, 1250, 1239, 898, 890, 483, 713, 711, 719, 492, 491, 711, 491, 483, 525, 738, 523, 538, 750, 536, 536, 740, 525, 484, 738, 713, 646, 849, 644, 656, 861, 654, 654, 851, 646, 682, 873, 680, 692, 883, 690, 690, 875, 682, 680, 863, 656, 644, 752, 538, 1021, 1234, 1019, 1031, 1242, 1029, 1029, 1236, 1021, 1061, 1271, 1059, 1071, 1279, 1069, 1069, 1273, 1061, 1059, 1244, 1031, 1177, 1314, 1312, 1189, 1320, 1188, 1329, 1196, 1197, 1196, 1322, 1189, 1188, 1314, 1179, 1177, 1281, 1071, 1019, 885, 692, 4, 449, 5, 10, 458, 8, 8, 448, 4, 468, 31, 30, 477, 40, 476, 31, 472, 34, 476, 34, 472, 449, 30, 5, 701, 95, 94, 99, 706, 98, 91, 701, 94, 91, 706, 698, 127, 724, 128, 729, 132, 727,
727, 128, 724, 702, 127, 95, 151, 588, 586, 594, 155, 154, 166, 596, 164, 586, 154, 151, 164, 588, 153, 219, 933, 217, 229, 941, 227, 227, 935, 219, 257, 965, 255, 265, 973, 263, 263, 967, 257, 255, 943, 229, 1085, 352, 353, 1083, 349, 352, 265, 1076, 975, 217, 598, 166, 1456, 1457, 1455, 1462, 1463, 1461, 1458, 1461, 1457, 1455, 1454, 1453, 1461, 1460, 1459, 1457, 1459, 1454, 1468, 1469, 1467, 1474, 1475, 1473, 1470, 1473, 1469, 1467, 1466, 1465, 1473, 1472, 1471, 1469, 1471, 1466, 1464, 1467, 1463, 1463, 1465, 1460, 1452, 1455, 1451, 1450, 1455, 1453, 1481, 1480, 1479, 1484, 1485, 1483, 1494, 1493, 1492, 1499, 1500, 1498, 1497, 1499, 1496, 1492, 1482, 1481, 1486, 1496, 1485, 1483, 1478, 1477, 1491, 1498, 1490, 1489, 1499, 1491, 1477, 1481, 1479, 1487, 1484, 1483, 1489, 1494, 1492, 1479, 1483, 1477, 1478, 1492, 1481, 1485, 1489, 1478, 1504, 1503, 1502, 1508, 1507, 1506, 1512, 1511, 1510, 1516, 1515, 1514, 1510, 1505, 1504, 1514, 1509, 1508, 1505, 1506, 1503, 1513, 1514, 1511, 1511, 1508, 1505, 1518, 1521, 1519,
1524, 1523, 1522, 1527, 1528, 1526, 1530, 1533, 1531, 1536, 1535, 1534, 1539, 1540, 1538, 1520, 1531, 1521, 1534, 1525, 1524, 1529, 1538, 1528, 1526, 1520, 1518, 1530, 1540, 1532, 1521, 1522, 1519, 1533, 1534, 1531, 1531, 1524, 1521, 1528, 1530, 1520, 1544, 1543, 1542, 1548, 1547, 1546, 1554, 1553, 1552, 1556, 1559, 1557, 1563, 1560, 1562, 1552, 1545, 1544, 1548, 1557, 1549, 1561, 1550, 1560, 1562, 1557, 1559, 1545, 1546, 1543, 1555, 1556, 1553, 1553, 1548, 1545, 1560, 1549, 1557, 1490, 1519, 1491, 1522, 1495, 1494, 1500, 1526, 1498, 1519, 1494, 1491, 1490, 1526, 1518, 1542, 1513, 1512, 1546, 1517, 1516, 1543, 1516, 1513, 1482, 1502, 1480, 1495, 1510, 1493, 1493, 1504, 1482, 1525, 1542, 1523, 1537, 1552, 1535, 1535, 1544, 1525, 1523, 1512, 1495, 1564, 1567, 1565, 1572, 1569, 1571, 1581, 1578, 1580, 1583, 1578, 1582, 1566, 1576, 1567, 1581, 1571, 1579, 1571, 1565, 1567, 1579, 1575, 1578, 1577, 1578, 1575, 1574, 1569, 1570, 1573, 1565, 1569, 1579, 1567, 1576, 1589, 1586, 1588, 1588, 1584, 1585, 1587, 1582, 1586, 1586, 1577, 1584, 1592, 1591, 1590, 1596, 1595, 1594,
1600, 1599, 1598, 1604, 1603, 1602, 1598, 1593, 1592, 1602, 1597, 1596, 1593, 1594, 1591, 1601, 1602, 1599, 1599, 1596, 1593, 1608, 1607, 1606, 1616, 1613, 1615, 1618, 1613, 1617, 1621, 1620, 1619, 1628, 1625, 1627, 1619, 1609, 1608, 1626, 1617, 1625, 1614, 1610, 1613, 1612, 1613, 1610, 1627, 1623, 1624, 1607, 1612, 1610, 1620, 1624, 1623, 1609, 1623, 1612, 1625, 1612, 1623, 1631, 1630, 1629, 1635, 1634, 1633, 1639, 1638, 1637, 1643, 1642, 1641, 1637, 1632, 1631, 1641, 1636, 1635, 1632, 1633, 1630, 1640, 1641, 1638, 1638, 1635, 1632, 1647, 1646, 1645, 1654, 1651, 1653, 1655, 1658, 1656, 1663, 1660, 1662, 1655, 1648, 1647, 1665, 1653, 1664, 1658, 1660, 1656, 1653, 1649, 1650, 1646, 1650, 1649, 1665, 1660, 1661, 1664, 1656, 1660, 1648, 1659, 1650, 1659, 1653, 1650, 1629, 1601, 1600, 1633, 1605, 1604, 1630, 1604, 1601, 1645, 1622, 1621, 1652, 1627, 1651, 1622, 1649, 1624, 1651, 1624, 1649, 1597, 1606, 1595, 1605, 1619, 1603, 1603, 1608, 1597, 1636, 1645, 1634, 1644, 1655, 1642, 1642, 1647, 1636, 1634, 1621, 1605, 1590, 1533, 1532, 1594, 1537, 1536, 1591, 1536, 1533,
1606, 1555, 1554, 1558, 1611, 1559, 1616, 1562, 1614, 1607, 1558, 1555, 1611, 1562, 1559, 1595, 1554, 1537, 1509, 1564, 1507, 1515, 1577, 1575, 1515, 1566, 1509, 1547, 1585, 1584, 1589, 1550, 1551, 1588, 1549, 1550, 1517, 1584, 1577, 1669, 1670, 1668, 1675, 1676, 1674, 1671, 1674, 1670, 1668, 1667, 1666, 1674, 1673, 1672, 1670, 1672, 1667, 1681, 1682, 1680, 1687, 1688, 1686, 1683, 1686, 1682, 1678, 1682, 1679, 1684, 1688, 1685, 1679, 1686, 1684, 1677, 1680, 1676, 1673, 1680, 1678, 1693, 1694, 1692, 1692, 1691, 1690, 1689, 1692, 1688, 1685, 1692, 1690, 1698, 1697, 1696, 1702, 1701, 1700, 1706, 1705, 1704, 1710, 1709, 1708, 1704, 1699, 1698, 1708, 1703, 1702, 1699, 1700, 1697, 1707, 1708, 1705, 1705, 1702, 1699, 1712, 1715, 1713, 1719, 1716, 1718, 1720, 1723, 1721, 1727, 1724, 1726, 1720, 1715, 1714, 1725, 1718, 1724, 1718, 1713, 1715, 1723, 1724, 1721, 1724, 1715, 1721, 1730, 1729, 1728, 1734, 1733, 1732, 1738, 1737, 1736, 1742, 1741, 1740, 1736, 1731, 1730, 1740, 1735, 1734, 1731, 1732, 1729, 1739, 1740, 1737, 1737, 1734, 1731, 1746, 1745, 1744, 1751, 1748, 1750,
1752, 1755, 1753, 1763, 1760, 1762, 1752, 1747, 1746, 1759, 1750, 1758, 1747, 1748, 1745, 1762, 1756, 1757, 1755, 1756, 1753, 1760, 1759, 1758, 1753, 1760, 1758, 1758, 1747, 1753, 1728, 1707, 1706, 1732, 1711, 1710, 1729, 1710, 1707, 1744, 1723, 1722, 1749, 1726, 1748, 1745, 1726, 1723, 1703, 1712, 1701, 1711, 1720, 1709, 1709, 1714, 1703, 1735, 1744, 1733, 1743, 1752, 1741, 1741, 1746, 1735, 1733, 1722, 1711, 1764, 1767, 1765, 1770, 1769, 1768, 1776, 1775, 1774, 1780, 1779, 1778, 1783, 1784, 1782, 1774, 1767, 1766, 1778, 1771, 1770, 1773, 1782, 1772, 1774, 1784, 1776, 1767, 1768, 1765, 1777, 1778, 1775, 1775, 1770, 1767, 1772, 1774, 1766, 1788, 1787, 1786, 1795, 1792, 1794, 1798, 1797, 1796, 1805, 1802, 1804, 1796, 1789, 1788, 1803, 1794, 1802, 1791, 1792, 1790, 1801, 1802, 1800, 1789, 1790, 1787, 1799, 1800, 1797, 1797, 1791, 1789, 1800, 1794, 1791, 1808, 1807, 1806, 1812, 1811, 1810, 1819, 1816, 1818, 1815, 1816, 1814, 1824, 1823, 1822, 1827, 1828, 1826, 1822, 1813, 1812, 1819, 1826, 1817, 1816, 1809, 1808, 1814, 1808, 1806, 1820, 1828, 1821, 1807, 1812, 1810,
1821, 1822, 1820, 1820, 1812, 1809, 1817, 1820, 1809, 1832, 1831, 1830, 1836, 1835, 1834, 1839, 1840, 1838, 1843, 1839, 1842, 1846, 1845, 1844, 1850, 1849, 1848, 1855, 1852, 1854, 1844, 1833, 1832, 1848, 1837, 1836, 1853, 1842, 1852, 1834, 1839, 1838, 1842, 1835, 1837, 1851, 1852, 1849, 1831, 1836, 1834, 1847, 1848, 1845, 1845, 1836, 1833, 1849, 1842, 1837, 1806, 1777, 1776, 1810, 1781, 1780, 1785, 1814, 1784, 1807, 1780, 1777, 1776, 1814, 1806, 1830, 1799, 1798, 1840, 1804, 1838, 1831, 1801, 1799, 1838, 1801, 1834, 1771, 1786, 1769, 1781, 1796, 1779, 1779, 1788, 1771, 1813, 1830, 1811, 1825, 1844, 1823, 1823, 1832, 1813, 1811, 1798, 1781, 1764, 1739, 1738, 1768, 1743, 1742, 1765, 1742, 1739, 1786, 1755, 1754, 1793, 1762, 1792, 1787, 1757, 1755, 1790, 1762, 1757, 1769, 1754, 1743, 1857, 1855, 1854, 1851, 1857, 1854, 1476, 1668, 1475, 1475, 1666, 1472, 1696, 1640, 1639, 1700, 1644, 1643, 1697, 1643, 1640, 1657, 1713, 1658, 1717, 1662, 1716, 1716, 1658, 1713, 1701, 1657, 1644, 1540, 1452, 1451, 1532, 1451, 1450, 1453, 1592, 1590, 1459, 1600, 1598, 1454, 1598, 1592,
1466, 1629, 1465, 1472, 1637, 1471, 1471, 1631, 1466, 1460, 1629, 1600, 1450, 1590, 1532, 1667, 1696, 1666, 1673, 1704, 1672, 1672, 1698, 1667, 1678, 1730, 1728, 1685, 1736, 1684, 1684, 1730, 1679, 1673, 1728, 1706, 1691, 1764, 1690, 1773, 1694, 1695, 1694, 1766, 1691, 1690, 1738, 1685, 1666, 1639, 1472, 1862, 1863, 1861, 1859, 1863, 1860, 1865, 1868, 1866, 1871, 1870, 1869, 1877, 1876, 1875, 1881, 1880, 1879, 1888, 1885, 1887, 1884, 1885, 1883, 1875, 1868, 1867, 1879, 1872, 1871, 1874, 1883, 1873, 1885, 1878, 1877, 1875, 1885, 1877, 1868, 1869, 1866, 1876, 1881, 1879, 1876, 1871, 1868, 1867, 1883, 1875, 1891, 1894, 1892, 1896, 1897, 1895, 1903, 1902, 1901, 1910, 1907, 1909, 1906, 1907, 1905, 1901, 1894, 1893, 1900, 1905, 1899, 1889, 1897, 1890, 1907, 1904, 1903, 1901, 1907, 1903, 1890, 1891, 1889, 1900, 1897, 1898, 1897, 1893, 1890, 1893, 1905, 1901, 1891, 1882, 1881, 1888, 1895, 1886, 1889, 1881, 1878, 1878, 1895, 1889, 1860, 1865, 1859, 1874, 1863, 1864, 1863, 1867, 1860, 1913, 1912, 1911, 1917, 1916, 1915, 1921, 1920, 1919, 1925, 1924, 1923, 1919, 1914, 1913,
1923, 1918, 1917, 1914, 1915, 1912, 1922, 1923, 1920, 1920, 1917, 1914, 1929, 1928, 1927, 1937, 1938, 1936, 1938, 1934, 1936, 1942, 1941, 1940, 1946, 1945, 1944, 1951, 1948, 1950, 1940, 1930, 1929, 1944, 1933, 1932, 1949, 1937, 1948, 1932, 1937, 1936, 1932, 1934, 1931, 1947, 1948, 1945, 1930, 1931, 1928, 1943, 1944, 1941, 1930, 1944, 1932, 1945, 1937, 1933, 1954, 1953, 1952, 1958, 1957, 1956, 1962, 1961, 1960, 1966, 1965, 1964, 1960, 1955, 1954, 1964, 1959, 1958, 1955, 1956, 1953, 1963, 1964, 1961, 1961, 1958, 1955, 1970, 1969, 1968, 1974, 1973, 1972, 1978, 1977, 1976, 1982, 1981, 1980, 1976, 1971, 1970, 1980, 1975, 1974, 1971, 1972, 1969, 1979, 1980, 1977, 1977, 1974, 1971, 1952, 1922, 1921, 1956, 1926, 1925, 1953, 1925, 1922, 1968, 1943, 1942, 1946, 1973, 1947, 1969, 1946, 1943, 1918, 1927, 1916, 1926, 1940, 1924, 1924, 1929, 1918, 1959, 1968, 1957, 1967, 1976, 1965, 1965, 1970, 1959, 1957, 1942, 1926, 1987, 1988, 1986, 1984, 1988, 1985, 1990, 1993, 1991, 1996, 1995, 1994, 2005, 2004, 2003, 2010, 2007, 2009, 2010, 2011, 2008, 2000, 1993, 1992, 2003, 1997, 1996,
1999, 2007, 1998, 2007, 2001, 2000, 2001, 2011, 2002, 1993, 1994, 1991, 2002, 2003, 2001, 1993, 2003, 1996, 1998, 2000, 1992, 2015, 2014, 2013, 2019, 2018, 2017, 2023, 2022, 2021, 2027, 2026, 2025, 2021, 2016, 2015, 2025, 2020, 2019, 2016, 2017, 2014, 2024, 2025, 2022, 2022, 2019, 2016, 2032, 2031, 2030, 2041, 2038, 2040, 2037, 2038, 2036, 2045, 2046, 2044, 2041, 2044, 2039, 2038, 2033, 2032, 2036, 2032, 2030, 2042, 2046, 2043, 2037, 2034, 2035, 2034, 2030, 2029, 2033, 2044, 2042, 2050, 2049, 2048, 2054, 2053, 2052, 2056, 2059, 2057, 2062, 2061, 2060, 2056, 2051, 2050, 2060, 2055, 2054, 2051, 2052, 2049, 2059, 2060, 2057, 2057, 2054, 2051, 2005, 2031, 2006, 2012, 2034, 2011, 2029, 2005, 2002, 2002, 2034, 2029, 2048, 2024, 2023, 2052, 2028, 2027, 2049, 2027, 2024, 1997, 2013, 1995, 2006, 2021, 2004, 2004, 2015, 1997, 2031, 2050, 2048, 2043, 2056, 2042, 2065, 2046, 2047, 2064, 2043, 2046, 2042, 2050, 2033, 2031, 2023, 2006, 1910, 1986, 1908, 1904, 1986, 1984, 1990, 1963, 1962, 1994, 1967, 1966, 1991, 1966, 1963, 2013, 1979, 1978, 2017, 1983, 1982, 2014, 1982, 1979,
1995, 1978, 1967, 1872, 1911, 1870, 1882, 1919, 1880, 1880, 1913, 1872, 1894, 1952, 1892, 1902, 1962, 1960, 1902, 1954, 1894, 1892, 1921, 1882, 1985, 1990, 1984, 1999, 1988, 1989, 1988, 1992, 1985, 1984, 1962, 1904, 2068, 2067, 2066, 2075, 2072, 2074, 2071, 2072, 2070, 2082, 2083, 2081, 2075, 2081, 2073, 2072, 2069, 2068, 2066, 2072, 2068, 2077, 2083, 2078, 2069, 2081, 2077, 2087, 2086, 2085, 2091, 2090, 2089, 2093, 2096, 2094, 2099, 2098, 2097, 2093, 2088, 2087, 2097, 2092, 2091, 2088, 2089, 2086, 2096, 2097, 2094, 2094, 2091, 2088, 2109, 2110, 2108, 2110, 2106, 2108, 2112, 2115, 2113, 2123, 2120, 2122, 2112, 2105, 2104, 2119, 2109, 2118, 2104, 2109, 2108, 2104, 2106, 2103, 2117, 2120, 2116, 2115, 2116, 2113, 2120, 2119, 2118, 2113, 2120, 2118, 2118, 2105, 2113, 2129, 2128, 2127, 2134, 2131, 2133, 2134, 2135, 2132, 2140, 2139, 2138, 2149, 2146, 2148, 2145, 2146, 2144, 2129, 2139, 2130, 2136, 2142, 2135, 2131, 2125, 2124, 2125, 2135, 2126, 2146, 2141, 2140, 2144, 2140, 2138, 2126, 2127, 2125, 2145, 2142, 2143, 2142, 2138, 2137, 2137, 2129, 2126, 2126, 2142, 2137,
2107, 2079, 2106, 2103, 2079, 2076, 2124, 2096, 2095, 2127, 2100, 2099, 2102, 2131, 2101, 2096, 2127, 2099, 2101, 2124, 2095, 2067, 2087, 2085, 2078, 2093, 2077, 2102, 2083, 2084, 2101, 2078, 2083, 2077, 2087, 2069, 2152, 2151, 2150, 2154, 2157, 2155, 2162, 2161, 2160, 2166, 2165, 2164, 2160, 2153, 2152, 2164, 2157, 2156, 2153, 2154, 2151, 2163, 2164, 2161, 2161, 2156, 2153, 2174, 2171, 2173, 2179, 2178, 2177, 2183, 2184, 2182, 2187, 2183, 2186, 2184, 2173, 2182, 2169, 2171, 2168, 2177, 2183, 2182, 2186, 2178, 2180, 2188, 2187, 2186, 2180, 2188, 2186, 2177, 2173, 2169, 2192, 2191, 2190, 2196, 2195, 2194, 2200, 2199, 2198, 2204, 2203, 2202, 2198, 2193, 2192, 2202, 2197, 2196, 2193, 2194, 2191, 2201, 2202, 2199, 2199, 2196, 2193, 2208, 2207, 2206, 2216, 2217, 2215, 2217, 2213, 2215, 2221, 2220, 2219, 2225, 2224, 2223, 2230, 2227, 2229, 2219, 2209, 2208, 2223, 2212, 2211, 2228, 2216, 2227, 2211, 2216, 2215, 2211, 2213, 2210, 2226, 2227, 2224, 2209, 2210, 2207, 2222, 2223, 2220, 2209, 2223, 2211, 2224, 2216, 2212, 2190, 2163, 2162, 2194, 2167, 2166, 2191, 2166, 2163,
2179, 2207, 2180, 2214, 2188, 2213, 2207, 2181, 2180, 2210, 2188, 2181, 2157, 2168, 2155, 2171, 2159, 2158, 2165, 2179, 2177, 2155, 2171, 2158, 2165, 2169, 2157, 2197, 2206, 2195, 2205, 2219, 2203, 2203, 2208, 2197, 2195, 2179, 2167, 2233, 2232, 2231, 2237, 2236, 2235, 2240, 2241, 2239, 2244, 2240, 2243, 2247, 2246, 2245, 2251, 2250, 2249, 2245, 2234, 2233, 2237, 2250, 2238, 2235, 2240, 2239, 2243, 2236, 2238, 2232, 2237, 2235, 2248, 2249, 2246, 2246, 2237, 2234, 2259, 2260, 2258, 2268, 2269, 2267, 2269, 2265, 2267, 2266, 2256, 2265, 2263, 2268, 2267, 2263, 2265, 2262, 2254, 2260, 2255, 2262, 2256, 2253, 2273, 2272, 2271, 2277, 2276, 2275, 2281, 2280, 2279, 2285, 2284, 2283, 2279, 2274, 2273, 2283, 2278, 2277, 2274, 2275, 2272, 2282, 2283, 2280, 2280, 2277, 2274, 2289, 2288, 2287, 2295, 2292, 2294, 2300, 2299, 2298, 2304, 2303, 2302, 2307, 2308, 2306, 2311, 2307, 2310, 2289, 2299, 2290, 2308, 2296, 2306, 2290, 2292, 2288, 2302, 2307, 2306, 2310, 2303, 2305, 2296, 2295, 2294, 2299, 2304, 2302, 2290, 2296, 2294, 2299, 2291, 2290, 2306, 2291, 2302, 2271, 2248, 2247,
2275, 2252, 2251, 2272, 2251, 2248, 2287, 2264, 2263, 2293, 2268, 2292, 2288, 2268, 2264, 2256, 2244, 2243, 2252, 2262, 2250, 2253, 2243, 2238, 2250, 2253, 2238, 2278, 2287, 2276, 2286, 2298, 2284, 2284, 2289, 2278, 2252, 2287, 2263, 2312, 2315, 2313, 2318, 2317, 2316, 2327, 2326, 2325, 2332, 2329, 2331, 2332, 2333, 2330, 2322, 2315, 2314, 2325, 2319, 2318, 2321, 2329, 2320, 2329, 2323, 2322, 2323, 2333, 2324, 2315, 2316, 2313, 2324, 2325, 2323, 2315, 2325, 2318, 2320, 2322, 2314, 2337, 2336, 2335, 2341, 2340, 2339, 2345, 2344, 2343, 2349, 2348, 2347, 2343, 2338, 2337, 2347, 2342, 2341, 2338, 2339, 2336, 2346, 2347, 2344, 2344, 2341, 2338, 2354, 2353, 2352, 2363, 2360, 2362, 2359, 2360, 2358, 2370, 2371, 2369, 2363, 2369, 2361, 2360, 2355, 2354, 2358, 2354, 2352, 2365, 2371, 2366, 2359, 2356, 2357, 2356, 2352, 2351, 2355, 2369, 2365, 2375, 2374, 2373, 2379, 2378, 2377, 2381, 2384, 2382, 2387, 2386, 2385, 2381, 2376, 2375, 2385, 2380, 2379, 2376, 2377, 2374, 2384, 2385, 2382, 2382, 2379, 2376, 2327, 2353, 2328, 2334, 2356, 2333, 2351, 2327, 2324, 2324, 2356, 2351,
2373, 2346, 2345, 2377, 2350, 2349, 2374, 2349, 2346, 2319, 2335, 2317, 2328, 2343, 2326, 2326, 2337, 2319, 2353, 2375, 2373, 2366, 2381, 2365, 2390, 2371, 2372, 2389, 2366, 2371, 2365, 2375, 2355, 2353, 2345, 2328, 2231, 2115, 2114, 2241, 2122, 2239, 2232, 2117, 2115, 2235, 2122, 2117, 2149, 2258, 2147, 2141, 2258, 2254, 2312, 2201, 2200, 2316, 2205, 2204, 2313, 2204, 2201, 2335, 2222, 2221, 2225, 2340, 2226, 2336, 2225, 2222, 2317, 2221, 2205, 2092, 2150, 2090, 2100, 2160, 2098, 2098, 2152, 2092, 2130, 2190, 2128, 2139, 2200, 2198, 2139, 2192, 2130, 2128, 2162, 2100, 2255, 2312, 2254, 2321, 2260, 2261, 2260, 2314, 2255, 2367, 2311, 2310, 2364, 2310, 2305, 2254, 2200, 2141, 2394, 2397, 2395, 2401, 2398, 2400, 2401, 2402, 2399, 2412, 2409, 2411, 2412, 2413, 2410, 2406, 2397, 2396, 2405, 2409, 2404, 2398, 2392, 2391, 2399, 2393, 2392, 2409, 2407, 2406, 2407, 2413, 2408, 2393, 2394, 2392, 2405, 2402, 2403, 2402, 2396, 2393, 2404, 2406, 2396, 2414, 2416, 2413, 2408, 2416, 2415, 2420, 2419, 2418, 2424, 2423, 2422, 2428, 2427, 2426, 2432, 2431, 2430, 2426, 2421, 2420,
2430, 2425, 2424, 2421, 2422, 2419, 2429, 2430, 2427, 2427, 2424, 2421, 2436, 2435, 2434, 2440, 2439, 2438, 2444, 2443, 2442, 2448, 2447, 2446, 2442, 2437, 2436, 2446, 2441, 2440, 2437, 2438, 2435, 2445, 2446, 2443, 2443, 2440, 2437, 2452, 2451, 2450, 2456, 2455, 2454, 2463, 2460, 2462, 2459, 2460, 2458, 2467, 2466, 2465, 2472, 2473, 2471, 2456, 2466, 2457, 2463, 2469, 2461, 2460, 2453, 2452, 2458, 2452, 2450, 2465, 2473, 2467, 2451, 2456, 2454, 2472, 2469, 2470, 2469, 2465, 2464, 2464, 2456, 2453, 2461, 2464, 2453, 2477, 2476, 2475, 2481, 2480, 2479, 2485, 2484, 2483, 2489, 2488, 2487, 2483, 2478, 2477, 2487, 2482, 2481, 2478, 2479, 2476, 2486, 2487, 2484, 2484, 2481, 2478, 2428, 2451, 2429, 2454, 2433, 2432, 2451, 2432, 2429, 2475, 2445, 2444, 2479, 2449, 2448, 2476, 2448, 2445, 2425, 2434, 2423, 2433, 2442, 2431, 2431, 2436, 2425, 2457, 2475, 2455, 2468, 2483, 2466, 2466, 2477, 2457, 2455, 2444, 2433, 2497, 2498, 2496, 2498, 2494, 2496, 2500, 2503, 2501, 2511, 2508, 2510, 2500, 2493, 2492, 2507, 2497, 2506, 2492, 2497, 2496, 2492, 2494, 2491, 2505, 2508, 2504,
2503, 2504, 2501, 2508, 2507, 2506, 2501, 2508, 2506, 2506, 2493, 2501, 2515, 2518, 2516, 2522, 2519, 2521, 2522, 2523, 2520, 2529, 2528, 2527, 2536, 2533, 2535, 2532, 2533, 2531, 2527, 2518, 2517, 2526, 2531, 2525, 2519, 2513, 2512, 2520, 2514, 2513, 2533, 2530, 2529, 2527, 2533, 2529, 2514, 2515, 2513, 2526, 2523, 2524, 2523, 2517, 2514, 2517, 2531, 2527, 2539, 2538, 2537, 2543, 2542, 2541, 2546, 2547, 2545, 2550, 2546, 2549, 2553, 2552, 2551, 2557, 2556, 2555, 2551, 2540, 2539, 2543, 2556, 2544, 2541, 2546, 2545, 2549, 2542, 2544, 2538, 2543, 2541, 2554, 2555, 2552, 2552, 2543, 2540, 2567, 2566, 2565, 2570, 2571, 2569, 2574, 2570, 2573, 2571, 2561, 2569, 2566, 2569, 2565, 2568, 2570, 2566, 2569, 2559, 2565, 2537, 2503, 2502, 2547, 2510, 2545, 2538, 2505, 2503, 2541, 2510, 2505, 2536, 2563, 2534, 2534, 2560, 2530, 2561, 2550, 2549, 2556, 2567, 2565, 2544, 2561, 2549, 2556, 2559, 2544, 2581, 2578, 2580, 2581, 2582, 2579, 2579, 2575, 2578, 2576, 2582, 2577, 2584, 2587, 2585, 2590, 2589, 2588, 2597, 2600, 2598, 2604, 2601, 2603, 2604, 2605, 2602, 2594, 2587, 2586,
2597, 2591, 2590, 2593, 2601, 2592, 2601, 2595, 2594, 2602, 2596, 2595, 2587, 2588, 2585, 2596, 2597, 2595, 2608, 2605, 2606, 2605, 2599, 2596, 2587, 2597, 2590, 2592, 2594, 2586, 2615, 2612, 2614, 2615, 2616, 2613, 2617, 2619, 2616, 2612, 2610, 2609, 2610, 2616, 2611, 2611, 2619, 2618, 2609, 2600, 2599, 2608, 2612, 2607, 2607, 2609, 2599, 2577, 2584, 2576, 2593, 2582, 2583, 2582, 2586, 2577, 2623, 2622, 2621, 2627, 2626, 2625, 2631, 2630, 2629, 2635, 2634, 2633, 2629, 2624, 2623, 2633, 2628, 2627, 2624, 2625, 2622, 2632, 2633, 2630, 2630, 2627, 2624, 2637, 2640, 2638, 2649, 2650, 2648, 2650, 2646, 2648, 2654, 2653, 2652, 2658, 2657, 2656, 2663, 2660, 2662, 2652, 2640, 2639, 2656, 2643, 2642, 2661, 2649, 2660, 2642, 2649, 2648, 2648, 2641, 2642, 2659, 2660, 2657, 2640, 2641, 2638, 2646, 2645, 2644, 2655, 2656, 2653, 2638, 2646, 2644, 2640, 2656, 2642, 2657, 2649, 2643, 2666, 2665, 2664, 2670, 2669, 2668, 2674, 2673, 2672, 2678, 2677, 2676, 2685, 2682, 2684, 2681, 2682, 2680, 2666, 2673, 2667, 2676, 2671, 2670, 2682, 2675, 2674, 2680, 2674, 2672, 2667, 2668, 2665,
2673, 2678, 2676, 2673, 2670, 2667, 2688, 2687, 2686, 2692, 2691, 2690, 2696, 2695, 2694, 2700, 2699, 2698, 2694, 2689, 2688, 2698, 2693, 2692, 2689, 2690, 2687, 2697, 2698, 2695, 2695, 2692, 2689, 2664, 2632, 2631, 2668, 2636, 2635, 2665, 2635, 2632, 2686, 2655, 2654, 2658, 2691, 2659, 2687, 2658, 2655, 2628, 2637, 2626, 2636, 2652, 2634, 2634, 2639, 2628, 2671, 2686, 2669, 2679, 2694, 2677, 2677, 2688, 2671, 2669, 2654, 2636, 2575, 2468, 2467, 2474, 2578, 2473, 2467, 2578, 2575, 2584, 2486, 2485, 2588, 2490, 2489, 2585, 2489, 2486, 2468, 2584, 2485, 2621, 2554, 2553, 2625, 2558, 2557, 2622, 2557, 2554, 2637, 2568, 2567, 2645, 2573, 2644, 2638, 2573, 2568, 2626, 2567, 2558, 2441, 2491, 2439, 2449, 2500, 2447, 2441, 2500, 2492, 2482, 2537, 2480, 2490, 2551, 2488, 2488, 2539, 2482, 2480, 2502, 2449, 2591, 2621, 2589, 2600, 2629, 2598, 2598, 2623, 2591, 2611, 2664, 2610, 2681, 2619, 2620, 2680, 2618, 2619, 2618, 2666, 2611, 2600, 2664, 2631, 2589, 2553, 2490, 2391, 2059, 2058, 2394, 2063, 2062, 2065, 2398, 2064, 2059, 2394, 2062, 2064, 2391, 2058, 2418, 2282, 2281,
2422, 2286, 2285, 2419, 2285, 2282, 2434, 2301, 2300, 2304, 2439, 2305, 2435, 2304, 2301, 2423, 2300, 2286, 2495, 2367, 2494, 2491, 2367, 2364, 2512, 2384, 2383, 2515, 2388, 2387, 2390, 2519, 2389, 2384, 2515, 2387, 2389, 2512, 2383, 2439, 2364, 2305, 2079, 1951, 1950, 1947, 2079, 1950, 1975, 2103, 1973, 1983, 2112, 1981, 1975, 2112, 2104, 1973, 2076, 1947, 2020, 2231, 2018, 2028, 2245, 2026, 2026, 2233, 2020, 2055, 2271, 2053, 2063, 2279, 2061, 2061, 2273, 2055, 2053, 2247, 2028, 2018, 2114, 1983, 2397, 2418, 2395, 2408, 2426, 2407, 2397, 2426, 2420, 2459, 2416, 2417, 2416, 2450, 2415, 2415, 2428, 2408, 2395, 2281, 2063, 2705, 2704, 2703, 2710, 2711, 2709, 2715, 2714, 2713, 2722, 2719, 2721, 2718, 2719, 2717, 2713, 2706, 2705, 2712, 2717, 2711, 2709, 2705, 2703, 2719, 2716, 2715, 2713, 2719, 2715, 2710, 2707, 2708, 2707, 2703, 2702, 2705, 2717, 2713, 2725, 2724, 2723, 2729, 2728, 2727, 2733, 2732, 2731, 2737, 2736, 2735, 2731, 2726, 2725, 2735, 2730, 2729, 2726, 2727, 2724, 2734, 2735, 2732, 2732, 2729, 2726, 2742, 2743, 2741, 2739, 2743, 2740, 2745, 2748, 2746,
2751, 2750, 2749, 2757, 2756, 2755, 2761, 2760, 2759, 2768, 2765, 2767, 2764, 2765, 2763, 2755, 2748, 2747, 2759, 2752, 2751, 2754, 2763, 2753, 2765, 2758, 2757, 2755, 2765, 2757, 2748, 2749, 2746, 2756, 2761, 2759, 2756, 2751, 2748, 2747, 2763, 2755, 2722, 2741, 2720, 2716, 2741, 2739, 2745, 2734, 2733, 2749, 2738, 2737, 2746, 2737, 2734, 2706, 2723, 2704, 2714, 2733, 2731, 2714, 2725, 2706, 2740, 2745, 2739, 2754, 2743, 2744, 2743, 2747, 2740, 2739, 2733, 2716, 2774, 2771, 2773, 2777, 2776, 2775, 2780, 2781, 2779, 2784, 2780, 2783, 2781, 2773, 2779, 2770, 2771, 2769, 2775, 2780, 2779, 2778, 2780, 2776, 2779, 2770, 2775, 2785, 2788, 2786, 2796, 2793, 2795, 2799, 2798, 2797, 2803, 2802, 2801, 2806, 2807, 2805, 2810, 2806, 2809, 2797, 2788, 2787, 2807, 2795, 2805, 2790, 2793, 2789, 2801, 2806, 2805, 2809, 2802, 2804, 2788, 2789, 2786, 2793, 2792, 2791, 2798, 2803, 2801, 2786, 2793, 2791, 2798, 2790, 2788, 2801, 2795, 2790, 2785, 2778, 2777, 2792, 2783, 2791, 2786, 2783, 2778, 2812, 2810, 2809, 2811, 2809, 2804, 2816, 2819, 2817, 2821, 2822, 2820, 2832, 2829, 2831,
2832, 2833, 2830, 2826, 2819, 2818, 2825, 2829, 2824, 2814, 2822, 2815, 2829, 2827, 2826, 2830, 2828, 2827, 2815, 2816, 2814, 2825, 2822, 2823, 2824, 2815, 2822, 2824, 2826, 2818, 2837, 2836, 2835, 2841, 2840, 2839, 2843, 2846, 2844, 2849, 2848, 2847, 2843, 2838, 2837, 2847, 2842, 2841, 2838, 2839, 2836, 2846, 2847, 2844, 2844, 2841, 2838, 2859, 2860, 2858, 2860, 2856, 2858, 2862, 2865, 2863, 2873, 2870, 2872, 2862, 2855, 2854, 2869, 2859, 2868, 2854, 2859, 2858, 2854, 2856, 2853, 2867, 2870, 2866, 2865, 2866, 2863, 2870, 2869, 2868, 2863, 2870, 2868, 2868, 2855, 2863, 2876, 2875, 2874, 2880, 2879, 2878, 2887, 2884, 2886, 2883, 2884, 2882, 2891, 2890, 2889, 2900, 2897, 2899, 2896, 2897, 2895, 2880, 2890, 2881, 2887, 2893, 2885, 2884, 2877, 2876, 2874, 2884, 2876, 2897, 2892, 2891, 2895, 2891, 2889, 2875, 2880, 2878, 2896, 2893, 2894, 2893, 2889, 2888, 2888, 2880, 2877, 2885, 2888, 2877, 2903, 2902, 2901, 2907, 2906, 2905, 2910, 2911, 2909, 2914, 2910, 2913, 2917, 2916, 2915, 2921, 2920, 2919, 2915, 2904, 2903, 2907, 2920, 2908, 2905, 2910, 2909, 2913, 2906, 2908,
2902, 2907, 2905, 2918, 2919, 2916, 2916, 2907, 2904, 2874, 2846, 2845, 2878, 2850, 2849, 2852, 2882, 2851, 2875, 2849, 2846, 2845, 2882, 2874, 2901, 2865, 2864, 2911, 2872, 2909, 2902, 2867, 2865, 2905, 2872, 2867, 2842, 2853, 2840, 2850, 2862, 2848, 2842, 2862, 2854, 2881, 2901, 2879, 2890, 2917, 2915, 2890, 2903, 2881, 2879, 2864, 2850, 2816, 2762, 2761, 2768, 2820, 2766, 2814, 2761, 2758, 2758, 2820, 2814, 2835, 2800, 2799, 2803, 2840, 2804, 2836, 2803, 2800, 2857, 2812, 2856, 2853, 2812, 2811, 2840, 2811, 2804, 2730, 2769, 2728, 2736, 2777, 2775, 2736, 2770, 2730, 2752, 2785, 2750, 2762, 2797, 2760, 2760, 2787, 2752, 2750, 2777, 2738, 2819, 2835, 2817, 2828, 2843, 2827, 2852, 2833, 2834, 2833, 2845, 2828, 2819, 2843, 2837, 2817, 2799, 2762, 2928, 2927, 2926, 2932, 2931, 2930, 2939, 2936, 2938, 2935, 2936, 2934, 2943, 2942, 2941, 2952, 2949, 2951, 2948, 2949, 2947, 2932, 2942, 2933, 2939, 2945, 2937, 2937, 2928, 2936, 2934, 2928, 2926, 2950, 2943, 2949, 2941, 2949, 2943, 2927, 2932, 2930, 2948, 2945, 2946, 2947, 2940, 2945, 2940, 2932, 2929, 2929, 2945, 2940,
2952, 2954, 2950, 2944, 2954, 2953, 2935, 2924, 2925, 2924, 2926, 2923, 2900, 2924, 2898, 2898, 2923, 2892, 2917, 2927, 2918, 2930, 2922, 2921, 2927, 2921, 2918, 2923, 2917, 2892, 2961, 2960, 2959, 2965, 2966, 2964, 2969, 2965, 2968, 2966, 2957, 2964, 2959, 2965, 2964, 2968, 2960, 2962, 2970, 2969, 2968, 2963, 2968, 2962, 2964, 2956, 2959, 2975, 2974, 2973, 2984, 2981, 2983, 2980, 2981, 2979, 2988, 2989, 2987, 2984, 2987, 2982, 2981, 2976, 2975, 2979, 2975, 2973, 2985, 2989, 2986, 2980, 2977, 2978, 2977, 2973, 2972, 2976, 2987, 2985, 2993, 2992, 2991, 2997, 2996, 2995, 2999, 3002, 3000, 3005, 3004, 3003, 2999, 2994, 2993, 3003, 2998, 2997, 2994, 2995, 2992, 3002, 3003, 3000, 3000, 2997, 2994, 3012, 3015, 3013, 3019, 3016, 3018, 3019, 3020, 3017, 3026, 3025, 3024, 3033, 3030, 3032, 3029, 3030, 3028, 3024, 3015, 3014, 3023, 3028, 3022, 3016, 3010, 3009, 3017, 3011, 3010, 3030, 3027, 3026, 3024, 3030, 3026, 3011, 3012, 3010, 3023, 3020, 3021, 3020, 3014, 3011, 3014, 3028, 3024, 3009, 3002, 3001, 3012, 3006, 3005, 3008, 3016, 3007, 3002, 3012, 3005, 3007, 3009, 3001,
2974, 2993, 2991, 2986, 2999, 2985, 3008, 2989, 2990, 2989, 3001, 2986, 2985, 2993, 2976, 3033, 3035, 3031, 3031, 3034, 3027, 3039, 3038, 3037, 3043, 3042, 3041, 3046, 3047, 3045, 3050, 3046, 3049, 3053, 3052, 3051, 3057, 3056, 3055, 3051, 3040, 3039, 3043, 3056, 3044, 3042, 3045, 3041, 3044, 3046, 3042, 3038, 3043, 3041, 3054, 3055, 3052, 3052, 3043, 3040, 3062, 3063, 3061, 3065, 3068, 3066, 3077, 3078, 3076, 3078, 3074, 3076, 3065, 3060, 3059, 3073, 3062, 3072, 3059, 3062, 3061, 3071, 3076, 3070, 3070, 3074, 3069, 3068, 3069, 3066, 3074, 3073, 3072, 3069, 3072, 3066, 3072, 3060, 3066, 3082, 3081, 3080, 3084, 3087, 3085, 3093, 3090, 3092, 3089, 3090, 3088, 3102, 3099, 3101, 3102, 3103, 3100, 3096, 3087, 3086, 3095, 3099, 3094, 3090, 3083, 3082, 3088, 3082, 3080, 3100, 3096, 3099, 3097, 3103, 3098, 3083, 3084, 3081, 3095, 3091, 3093, 3091, 3086, 3083, 3086, 3099, 3096, 3107, 3106, 3105, 3109, 3112, 3110, 3117, 3116, 3115, 3121, 3120, 3119, 3126, 3123, 3125, 3115, 3108, 3107, 3119, 3112, 3111, 3123, 3118, 3117, 3108, 3109, 3106, 3116, 3121, 3119, 3116, 3111, 3108,
3053, 3081, 3054, 3084, 3058, 3057, 3081, 3057, 3054, 3105, 3068, 3067, 3109, 3071, 3070, 3114, 3077, 3113, 3068, 3109, 3070, 3110, 3077, 3071, 3061, 3050, 3049, 3058, 3065, 3056, 3044, 3061, 3049, 3056, 3059, 3044, 3087, 3105, 3085, 3098, 3115, 3097, 3125, 3103, 3104, 3103, 3117, 3098, 3087, 3115, 3107, 3085, 3067, 3058, 3129, 3128, 3127, 3134, 3135, 3133, 3138, 3134, 3137, 3140, 3141, 3139, 3145, 3144, 3143, 3147, 3150, 3148, 3129, 3144, 3130, 3147, 3132, 3131, 3152, 3140, 3151, 3127, 3134, 3133, 3130, 3134, 3128, 3131, 3140, 3139, 3139, 3138, 3137, 3146, 3147, 3144, 3130, 3139, 3137, 3144, 3131, 3130, 3148, 3140, 3132, 3155, 3154, 3153, 3160, 3161, 3159, 3164, 3160, 3163, 3166, 3167, 3165, 3153, 3160, 3159, 3156, 3160, 3154, 3158, 3165, 3157, 3165, 3164, 3163, 3156, 3165, 3163, 3150, 3153, 3148, 3159, 3152, 3151, 3148, 3159, 3151, 3172, 3171, 3170, 3181, 3178, 3180, 3177, 3178, 3176, 3178, 3173, 3172, 3176, 3172, 3170, 3177, 3174, 3175, 3176, 3169, 3174, 3182, 3185, 3183, 3188, 3187, 3186, 3200, 3197, 3199, 3204, 3201, 3203, 3204, 3205, 3202, 3192, 3185, 3184,
3194, 3189, 3188, 3191, 3197, 3190, 3197, 3193, 3192, 3202, 3194, 3201, 3195, 3205, 3196, 3185, 3186, 3183, 3203, 3198, 3200, 3198, 3194, 3193, 3193, 3188, 3185, 3184, 3197, 3192, 3209, 3208, 3207, 3211, 3214, 3212, 3219, 3218, 3217, 3223, 3222, 3221, 3228, 3225, 3227, 3217, 3210, 3209, 3221, 3214, 3213, 3225, 3220, 3219, 3210, 3211, 3208, 3218, 3223, 3221, 3218, 3213, 3210, 3237, 3234, 3236, 3234, 3231, 3230, 3236, 3232, 3233, 3232, 3230, 3229, 3230, 3224, 3223, 3228, 3232, 3226, 3229, 3223, 3220, 3220, 3232, 3229, 3189, 3207, 3187, 3196, 3217, 3195, 3227, 3205, 3206, 3225, 3196, 3205, 3189, 3217, 3209, 3121, 3171, 3122, 3126, 3174, 3124, 3169, 3121, 3118, 3118, 3174, 3169, 3182, 3146, 3145, 3186, 3150, 3149, 3183, 3149, 3146, 3155, 3208, 3156, 3211, 3158, 3157, 3216, 3166, 3215, 3208, 3157, 3156, 3212, 3166, 3158, 3187, 3155, 3150, 3112, 3127, 3110, 3133, 3114, 3113, 3122, 3143, 3120, 3110, 3133, 3113, 3120, 3129, 3112, 3173, 3182, 3171, 3191, 3179, 3181, 3179, 3184, 3173, 3171, 3145, 3122, 3240, 3239, 3238, 3244, 3243, 3242, 3247, 3248, 3246, 3251, 3252, 3250,
3255, 3251, 3254, 3258, 3257, 3256, 3262, 3261, 3260, 3256, 3241, 3240, 3244, 3261, 3245, 3238, 3247, 3246, 3243, 3250, 3242, 3245, 3251, 3243, 3241, 3242, 3239, 3250, 3249, 3247, 3259, 3260, 3257, 3239, 3250, 3247, 3257, 3244, 3241, 3269, 3270, 3268, 3273, 3274, 3272, 3278, 3277, 3276, 3280, 3283, 3281, 3276, 3265, 3264, 3280, 3267, 3266, 3285, 3273, 3284, 3264, 3269, 3268, 3267, 3272, 3266, 3272, 3271, 3269, 3279, 3280, 3277, 3265, 3272, 3269, 3277, 3266, 3265, 3281, 3273, 3267, 3291, 3290, 3289, 3296, 3293, 3295, 3296, 3297, 3294, 3302, 3299, 3301, 3294, 3286, 3293, 3287, 3297, 3288, 3299, 3292, 3291, 3288, 3289, 3287, 3301, 3297, 3298, 3299, 3288, 3297, 3305, 3304, 3303, 3307, 3310, 3308, 3314, 3311, 3313, 3322, 3319, 3321, 3317, 3310, 3309, 3316, 3319, 3315, 3311, 3306, 3305, 3320, 3317, 3319, 3306, 3307, 3304, 3316, 3312, 3314, 3315, 3306, 3312, 3309, 3319, 3317, 3286, 3259, 3258, 3289, 3263, 3262, 3259, 3289, 3262, 3303, 3279, 3278, 3307, 3283, 3282, 3304, 3282, 3279, 3268, 3255, 3254, 3263, 3276, 3261, 3245, 3268, 3254, 3261, 3264, 3245, 3292, 3303, 3290,
3313, 3300, 3302, 3300, 3305, 3292, 3290, 3278, 3263, 3325, 3324, 3323, 3327, 3330, 3328, 3332, 3333, 3331, 3336, 3337, 3335, 3323, 3332, 3331, 3327, 3336, 3335, 3324, 3329, 3327, 3335, 3334, 3332, 3324, 3335, 3332, 3341, 3340, 3339, 3345, 3344, 3343, 3348, 3349, 3347, 3352, 3353, 3351, 3339, 3348, 3347, 3344, 3351, 3343, 3342, 3343, 3340, 3351, 3350, 3348, 3343, 3348, 3340, 3357, 3356, 3355, 3361, 3360, 3359, 3370, 3367, 3369, 3374, 3371, 3373, 3363, 3358, 3357, 3361, 3366, 3362, 3367, 3364, 3363, 3371, 3366, 3365, 3358, 3359, 3356, 3373, 3368, 3370, 3368, 3365, 3364, 3358, 3365, 3361, 3377, 3376, 3375, 3381, 3380, 3379, 3390, 3387, 3389, 3394, 3391, 3393, 3377, 3384, 3378, 3385, 3382, 3381, 3388, 3383, 3387, 3392, 3385, 3391, 3378, 3379, 3376, 3393, 3388, 3390, 3391, 3384, 3388, 3378, 3385, 3381, 3355, 3326, 3325, 3359, 3330, 3329, 3356, 3329, 3326, 3375, 3342, 3341, 3379, 3346, 3345, 3376, 3345, 3342, 3328, 3341, 3339, 3347, 3338, 3336, 3328, 3347, 3336, 3362, 3375, 3360, 3389, 3372, 3374, 3387, 3366, 3372, 3366, 3377, 3362, 3360, 3341, 3330, 3283, 3323, 3281,
3331, 3285, 3284, 3323, 3284, 3281, 3310, 3355, 3308, 3369, 3320, 3322, 3320, 3363, 3318, 3318, 3357, 3310, 3308, 3325, 3283, 2961, 3038, 2962, 3047, 2970, 3045, 3038, 2963, 2962, 3041, 2970, 2963, 3214, 3238, 3212, 3246, 3216, 3215, 3224, 3256, 3222, 3238, 3215, 3212, 3222, 3240, 3214, 3295, 3235, 3237, 3235, 3286, 3231, 3231, 3258, 3224, 1824, 2704, 1825, 1829, 2707, 1828, 2702, 1824, 1821, 1828, 2702, 1821, 2723, 1847, 1846, 1850, 2728, 1851, 2724, 1850, 1847, 2704, 1846, 1825, 2772, 1857, 2771, 2769, 1857, 1856, 2728, 1856, 1851, 2678, 2974, 2679, 2685, 2977, 2683, 2972, 2678, 2675, 2683, 2972, 2675, 2991, 2697, 2696, 2995, 2701, 2700, 2992, 2700, 2697, 2974, 2696, 2679, 2957, 2914, 2913, 2920, 2961, 2959, 2908, 2957, 2913, 2920, 2956, 2908, 2933, 3037, 2931, 2942, 3053, 3051, 2942, 3039, 2933, 3089, 2954, 2955, 3088, 2953, 2954, 2953, 3053, 2944, 2931, 2961, 2922, 304, 1480, 305, 311, 1487, 310, 310, 1479, 304, 1502, 323, 322, 1506, 327, 326, 1503, 326, 323, 1480, 322, 305, 1564, 384, 383, 1574, 392, 1573, 384, 1568, 387, 1573, 387, 1568,
1507, 383, 327, 1135, 1861, 1133, 1129, 1861, 1859, 1865, 1147, 1146, 1869, 1151, 1150, 1866, 1150, 1147, 1859, 1146, 1129, 1911, 1209, 1208, 1915, 1213, 1212, 1912, 1212, 1209, 1225, 1928, 1226, 1935, 1232, 1934, 1928, 1227, 1226, 1931, 1232, 1227, 1916, 1225, 1213, 1870, 1208, 1151, 2066, 1367, 1366, 1373, 2070, 1372, 1366, 2070, 2066, 2085, 1385, 1384, 2089, 1389, 1388, 2086, 1388, 1385, 2067, 1384, 1367, 2150, 1427, 1426, 2154, 1431, 1430, 2159, 1436, 2158, 2151, 1430, 1427, 2155, 1436, 1431, 1449, 2175, 1448, 1448, 2170, 1445, 2090, 1426, 1389, 3403, 3406, 3404, 3408, 3409, 3407, 3412, 3413, 3411, 3398, 3407, 3397, 3413, 3399, 3411, 3401, 3409, 3402, 3403, 3412, 3411, 3399, 3398, 3397, 3402, 3403, 3401, 3416, 3409, 3410, 3396, 3397, 3395, 3415, 3402, 3409, 3401, 3396, 3395, 3395, 3407, 3401, 3403, 3399, 3396, 3423, 3424, 3422, 3424, 3420, 3422, 3419, 3422, 3418, 3418, 3420, 3417, 3432, 3429, 3431, 3432, 3433, 3430, 3429, 3427, 3426, 3427, 3433, 3428, 3435, 3438, 3436, 3441, 3440, 3439, 3448, 3449, 3447, 3452, 3448, 3451, 3457, 3456, 3455, 3462, 3459, 3461,
3468, 3465, 3467, 3464, 3465, 3463, 3453, 3438, 3437, 3441, 3456, 3442, 3446, 3459, 3445, 3459, 3454, 3453, 3439, 3448, 3447, 3442, 3448, 3440, 3465, 3458, 3457, 3455, 3465, 3457, 3438, 3439, 3436, 3447, 3444, 3443, 3464, 3460, 3462, 3436, 3447, 3443, 3460, 3455, 3454, 3454, 3441, 3438, 3445, 3453, 3437, 3426, 3406, 3405, 3416, 3429, 3415, 3415, 3426, 3405, 3435, 3419, 3418, 3444, 3423, 3443, 3436, 3423, 3419, 3406, 3417, 3404, 3420, 3414, 3412, 3404, 3420, 3412, 3428, 3435, 3427, 3446, 3433, 3434, 3445, 3428, 3433, 3406, 3435, 3418, 3472, 3473, 3471, 3475, 3478, 3476, 3489, 3490, 3488, 3490, 3486, 3488, 3475, 3470, 3469, 3483, 3472, 3482, 3469, 3472, 3471, 3480, 3489, 3488, 3480, 3486, 3479, 3478, 3479, 3476, 3486, 3483, 3482, 3479, 3482, 3476, 3476, 3472, 3470, 3493, 3491, 3489, 3481, 3493, 3489, 3471, 3452, 3451, 3458, 3475, 3456, 3485, 3466, 3468, 3442, 3471, 3451, 3466, 3477, 3458, 3456, 3469, 3442, 3498, 3499, 3497, 3503, 3502, 3501, 3506, 3507, 3505, 3501, 3496, 3495, 3499, 3505, 3497, 3496, 3497, 3495, 3505, 3503, 3501, 3497, 3501, 3495, 3511, 3510, 3509,
3514, 3515, 3513, 3522, 3519, 3521, 3517, 3512, 3511, 3516, 3519, 3515, 3513, 3511, 3509, 3520, 3517, 3519, 3515, 3517, 3511, 3509, 3504, 3503, 3508, 3513, 3507, 3507, 3509, 3503, 3528, 3529, 3527, 3532, 3533, 3531, 3537, 3536, 3535, 3541, 3540, 3539, 3535, 3524, 3523, 3539, 3526, 3525, 3524, 3527, 3523, 3526, 3531, 3525, 3531, 3530, 3528, 3538, 3539, 3536, 3525, 3528, 3524, 3536, 3525, 3524, 3548, 3549, 3547, 3552, 3553, 3551, 3557, 3556, 3555, 3561, 3560, 3559, 3555, 3544, 3543, 3559, 3546, 3545, 3544, 3547, 3543, 3546, 3551, 3545, 3551, 3550, 3548, 3558, 3559, 3556, 3545, 3548, 3544, 3556, 3545, 3544, 3565, 3564, 3563, 3569, 3568, 3567, 3578, 3575, 3577, 3582, 3579, 3581, 3571, 3566, 3565, 3573, 3570, 3569, 3576, 3571, 3575, 3580, 3573, 3579, 3566, 3567, 3564, 3581, 3576, 3578, 3579, 3572, 3576, 3572, 3569, 3566, 3585, 3584, 3583, 3589, 3588, 3587, 3598, 3595, 3597, 3602, 3599, 3601, 3591, 3586, 3585, 3593, 3590, 3589, 3596, 3591, 3595, 3600, 3593, 3599, 3586, 3587, 3584, 3601, 3596, 3598, 3599, 3592, 3596, 3592, 3589, 3586, 3563, 3538, 3537, 3567, 3542, 3541,
3564, 3541, 3538, 3583, 3558, 3557, 3587, 3562, 3561, 3584, 3561, 3558, 3547, 3534, 3532, 3542, 3555, 3540, 3543, 3532, 3526, 3540, 3543, 3526, 3570, 3583, 3568, 3597, 3580, 3582, 3595, 3574, 3580, 3574, 3585, 3570, 3568, 3557, 3542, 3527, 3500, 3498, 3504, 3535, 3502, 3523, 3498, 3496, 3502, 3523, 3496, 3512, 3563, 3510, 3577, 3520, 3522, 3575, 3518, 3520, 3518, 3565, 3512, 3510, 3537, 3504, 3608, 3609, 3607, 3614, 3611, 3613, 3622, 3623, 3621, 3630, 3631, 3629, 3631, 3625, 3629, 3628, 3630, 3626, 3610, 3621, 3609, 3627, 3613, 3625, 3607, 3604, 3603, 3621, 3616, 3615, 3606, 3611, 3605, 3619, 3630, 3629, 3619, 3625, 3617, 3626, 3620, 3618, 3604, 3605, 3603, 3611, 3608, 3607, 3615, 3619, 3617, 3603, 3611, 3607, 3604, 3617, 3606, 3609, 3615, 3604, 3617, 3613, 3606, 3639, 3638, 3637, 3644, 3645, 3643, 3645, 3647, 3643, 3635, 3644, 3634, 3637, 3644, 3643, 3637, 3647, 3639, 3647, 3642, 3641, 3639, 3641, 3636, 3634, 3638, 3633, 3651, 3654, 3652, 3657, 3656, 3655, 3660, 3661, 3659, 3661, 3663, 3659, 3667, 3670, 3668, 3678, 3675, 3677, 3678, 3673, 3676, 3682, 3679, 3681,
3667, 3654, 3653, 3657, 3672, 3658, 3682, 3665, 3680, 3651, 3660, 3659, 3659, 3653, 3651, 3676, 3669, 3675, 3676, 3668, 3670, 3680, 3671, 3679, 3652, 3657, 3655, 3681, 3673, 3674, 3679, 3668, 3673, 3654, 3671, 3657, 3680, 3658, 3672, 3619, 3656, 3620, 3624, 3660, 3623, 3616, 3655, 3619, 3616, 3660, 3652, 3663, 3635, 3634, 3640, 3667, 3638, 3677, 3648, 3650, 3633, 3663, 3634, 3675, 3640, 3648, 3633, 3667, 3653, 3688, 3687, 3686, 3693, 3690, 3692, 3697, 3698, 3696, 3700, 3703, 3701, 3704, 3707, 3705, 3709, 3710, 3708, 3715, 3712, 3714, 3704, 3689, 3688, 3695, 3709, 3694, 3690, 3684, 3683, 3701, 3708, 3700, 3686, 3698, 3688, 3713, 3706, 3712, 3697, 3691, 3693, 3698, 3695, 3694, 3703, 3704, 3701, 3696, 3684, 3691, 3685, 3698, 3694, 3685, 3704, 3688, 3694, 3701, 3685, 3721, 3718, 3720, 3727, 3724, 3726, 3725, 3720, 3724, 3717, 3718, 3716, 3726, 3722, 3723, 3724, 3717, 3722, 3735, 3732, 3734, 3737, 3732, 3736, 3733, 3728, 3732, 3736, 3728, 3730, 3739, 3733, 3735, 3738, 3729, 3733, 3702, 3729, 3703, 3739, 3712, 3738, 3703, 3731, 3706, 3738, 3706, 3731, 3689, 3716, 3687,
3705, 3723, 3722, 3727, 3713, 3715, 3726, 3707, 3713, 3689, 3722, 3717, 3748, 3745, 3747, 3746, 3741, 3745, 3747, 3743, 3744, 3745, 3740, 3743, 3741, 3640, 3639, 3642, 3743, 3641, 3748, 3648, 3746, 3636, 3741, 3639, 3636, 3743, 3740, 3746, 3640, 3742, 3618, 3702, 3700, 3708, 3628, 3626, 3700, 3626, 3618, 3656, 3730, 3728, 3737, 3665, 3666, 3665, 3730, 3658, 3656, 3702, 3620, 3755, 3756, 3754, 3760, 3763, 3761, 3765, 3766, 3764, 3760, 3751, 3750, 3753, 3764, 3752, 3764, 3759, 3758, 3751, 3754, 3750, 3754, 3753, 3752, 3759, 3760, 3758, 3750, 3752, 3749, 3758, 3750, 3749, 3752, 3758, 3749, 3775, 3776, 3774, 3779, 3780, 3778, 3784, 3783, 3782, 3788, 3787, 3786, 3793, 3790, 3792, 3797, 3794, 3796, 3782, 3771, 3770, 3786, 3773, 3772, 3771, 3774, 3770, 3791, 3784, 3790, 3773, 3778, 3772, 3795, 3788, 3794, 3778, 3777, 3775, 3785, 3786, 3783, 3796, 3791, 3793, 3772, 3775, 3771, 3794, 3785, 3791, 3783, 3772, 3771, 3800, 3803, 3801, 3805, 3806, 3804, 3811, 3808, 3810, 3814, 3817, 3815, 3819, 3820, 3818, 3825, 3822, 3824, 3802, 3815, 3803, 3807, 3818, 3806, 3823, 3810, 3822,
3804, 3799, 3798, 3812, 3820, 3813, 3803, 3808, 3801, 3817, 3822, 3815, 3799, 3800, 3798, 3813, 3814, 3812, 3812, 3802, 3799, 3799, 3818, 3812, 3815, 3810, 3803, 3762, 3801, 3763, 3767, 3804, 3766, 3809, 3768, 3808, 3798, 3762, 3759, 3766, 3798, 3759, 3808, 3763, 3801, 3774, 3757, 3755, 3763, 3782, 3761, 3792, 3768, 3769, 3770, 3755, 3751, 3790, 3763, 3768, 3761, 3770, 3751, 3831, 3832, 3830, 3835, 3836, 3834, 3840, 3839, 3838, 3844, 3843, 3842, 3849, 3846, 3848, 3853, 3850, 3852, 3838, 3827, 3826, 3842, 3829, 3828, 3827, 3830, 3826, 3847, 3840, 3846, 3829, 3834, 3828, 3851, 3844, 3850, 3834, 3833, 3831, 3841, 3842, 3839, 3852, 3847, 3849, 3828, 3831, 3827, 3850, 3841, 3847, 3839, 3828, 3827, 3857, 3858, 3856, 3862, 3861, 3860, 3867, 3864, 3866, 3860, 3855, 3854, 3865, 3857, 3864, 3855, 3856, 3854, 3866, 3861, 3863, 3864, 3855, 3861, 3871, 3872, 3870, 3877, 3878, 3876, 3873, 3876, 3872, 3870, 3869, 3868, 3876, 3875, 3874, 3872, 3874, 3869, 3880, 3883, 3881, 3887, 3884, 3886, 3890, 3889, 3888, 3895, 3892, 3894, 3888, 3883, 3882, 3893, 3886, 3892, 3886, 3881, 3883,
3894, 3889, 3891, 3892, 3883, 3889, 3853, 3870, 3851, 3851, 3868, 3845, 3862, 3881, 3863, 3885, 3866, 3884, 3884, 3863, 3881, 3856, 3837, 3835, 3845, 3860, 3843, 3854, 3835, 3829, 3843, 3854, 3829, 3869, 3880, 3868, 3874, 3890, 3888, 3869, 3888, 3882, 3868, 3862, 3845, 3899, 3898, 3897, 3905, 3907, 3903, 3910, 3904, 3908, 3913, 3912, 3911, 3920, 3917, 3919, 3916, 3917, 3915, 3899, 3912, 3900, 3909, 3915, 3907, 3903, 3899, 3897, 3908, 3898, 3900, 3917, 3914, 3913, 3911, 3917, 3913, 3905, 3901, 3902, 3901, 3897, 3896, 3899, 3915, 3911, 3926, 3925, 3924, 3931, 3932, 3930, 3935, 3931, 3934, 3937, 3938, 3936, 3932, 3922, 3930, 3925, 3930, 3924, 3927, 3931, 3925, 3928, 3937, 3936, 3936, 3935, 3934, 3927, 3936, 3934, 3924, 3922, 3921, 3945, 3944, 3943, 3949, 3948, 3947, 3956, 3953, 3955, 3952, 3953, 3951, 3960, 3957, 3959, 3953, 3946, 3945, 3951, 3945, 3943, 3957, 3950, 3949, 3946, 3947, 3944, 3959, 3954, 3956, 3954, 3949, 3946, 3920, 3941, 3918, 3914, 3941, 3940, 3926, 3944, 3927, 3947, 3929, 3928, 3944, 3928, 3927, 3922, 3910, 3908, 3912, 3926, 3924, 3900, 3922, 3908,
3912, 3921, 3900, 3952, 3941, 3942, 3941, 3943, 3940, 3940, 3926, 3914, 3967, 3968, 3966, 3965, 3967, 3964, 3975, 3974, 3973, 3980, 3981, 3979, 3978, 3980, 3977, 3962, 3974, 3963, 3968, 3983, 3966, 3971, 3980, 3979, 3977, 3972, 3970, 3963, 3966, 3962, 3964, 3963, 3961, 3970, 3975, 3973, 3983, 3978, 3977, 3970, 3983, 3977, 3962, 3983, 3973, 3987, 3990, 3988, 3996, 3993, 3995, 3996, 3991, 3994, 4002, 3999, 4001, 4000, 3993, 3999, 3994, 3989, 3993, 3990, 3991, 3988, 4001, 3997, 3998, 3997, 3993, 3989, 4003, 4006, 4004, 4013, 4010, 4012, 4017, 4014, 4016, 4019, 4014, 4018, 4011, 4005, 4010, 4014, 4008, 4007, 4018, 4007, 4009, 4004, 4009, 4007, 4019, 4011, 4013, 4011, 4009, 4006, 3971, 4004, 3972, 3975, 4008, 3976, 4017, 3985, 4015, 4004, 3975, 3972, 4008, 3985, 3976, 3963, 3987, 3961, 3974, 3998, 3997, 4002, 3985, 3986, 4001, 3976, 3985, 3974, 3989, 3963, 3816, 3898, 3817, 3821, 3901, 3820, 3906, 3824, 3904, 3896, 3816, 3813, 3820, 3896, 3813, 3898, 3824, 3817, 3879, 3964, 3878, 3878, 3961, 3875, 3890, 3988, 3891, 3992, 3894, 3991, 3991, 3891, 3988, 3875, 3987, 3890,
3830, 3781, 3779, 3789, 3838, 3787, 3848, 3795, 3797, 3826, 3779, 3773, 3846, 3789, 3795, 3787, 3826, 3773, 3979, 3939, 3937, 3971, 3937, 3929, 3950, 4003, 3948, 4012, 3958, 3960, 4010, 3950, 3958, 3929, 4003, 3971, 4023, 4024, 4022, 4028, 4027, 4026, 4031, 4032, 4030, 4026, 4021, 4020, 4024, 4030, 4022, 4021, 4022, 4020, 4030, 4028, 4026, 4022, 4026, 4020, 4036, 4035, 4034, 4039, 4040, 4038, 4044, 4043, 4042, 4047, 4048, 4046, 4042, 4037, 4036, 4041, 4046, 4040, 4038, 4036, 4034, 4046, 4044, 4042, 4040, 4042, 4036, 4034, 4029, 4028, 4033, 4038, 4032, 4032, 4034, 4028, 4055, 4056, 4054, 4059, 4060, 4058, 4064, 4063, 4062, 4068, 4067, 4066, 4062, 4051, 4050, 4066, 4053, 4052, 4051, 4054, 4050, 4053, 4058, 4052, 4058, 4057, 4055, 4065, 4066, 4063, 4052, 4055, 4051, 4063, 4052, 4051, 4075, 4076, 4074, 4079, 4080, 4078, 4084, 4083, 4082, 4088, 4087, 4086, 4082, 4071, 4070, 4086, 4073, 4072, 4071, 4074, 4070, 4073, 4078, 4072, 4078, 4077, 4075, 4085, 4086, 4083, 4072, 4075, 4071, 4083, 4072, 4071, 4092, 4091, 4090, 4096, 4095, 4094, 4100, 4099, 4098, 4104, 4103, 4102,
4098, 4093, 4092, 4102, 4097, 4096, 4093, 4094, 4091, 4101, 4102, 4099, 4099, 4096, 4093, 4108, 4107, 4106, 4112, 4111, 4110, 4114, 4117, 4115, 4124, 4121, 4123, 4128, 4125, 4127, 4130, 4125, 4129, 4114, 4109, 4108, 4112, 4119, 4113, 4121, 4117, 4116, 4125, 4119, 4118, 4129, 4118, 4120, 4109, 4110, 4107, 4115, 4120, 4118, 4130, 4122, 4124, 4122, 4120, 4117, 4115, 4112, 4109, 4090, 4065, 4064, 4094, 4069, 4068, 4091, 4068, 4065, 4106, 4085, 4084, 4110, 4089, 4088, 4107, 4088, 4085, 4074, 4061, 4059, 4069, 4082, 4067, 4070, 4059, 4053, 4067, 4070, 4053, 4097, 4106, 4095, 4105, 4114, 4103, 4103, 4108, 4097, 4095, 4084, 4069, 4054, 4025, 4023, 4029, 4062, 4027, 4050, 4023, 4021, 4027, 4050, 4021, 4037, 4090, 4035, 4045, 4098, 4043, 4043, 4092, 4037, 4035, 4064, 4029, 3683, 3478, 3477, 3480, 3687, 3481, 3485, 3690, 3484, 3478, 3686, 3480, 3477, 3690, 3683, 3719, 3493, 3718, 3718, 3492, 3716, 3687, 3492, 3481, 4135, 4134, 4133, 4138, 4139, 4137, 4142, 4143, 4141, 4137, 4132, 4131, 4134, 4141, 4133, 4132, 4133, 4131, 4141, 4138, 4137, 4133, 4137, 4131, 4147, 4146, 4145,
4151, 4150, 4149, 4154, 4155, 4153, 4158, 4159, 4157, 4146, 4153, 4145, 4150, 4157, 4149, 4148, 4149, 4146, 4157, 4156, 4154, 4149, 4154, 4146, 4136, 4145, 4134, 4153, 4144, 4142, 4145, 4142, 4134, 4163, 4162, 4161, 4167, 4166, 4165, 4170, 4171, 4169, 4174, 4175, 4173, 4162, 4169, 4161, 4166, 4173, 4165, 4164, 4165, 4162, 4173, 4172, 4170, 4165, 4170, 4162, 4179, 4178, 4177, 4182, 4183, 4181, 4186, 4182, 4185, 4178, 4181, 4177, 4185, 4178, 4180, 4168, 4177, 4166, 4181, 4176, 4174, 4177, 4174, 4166, 4189, 4192, 4190, 4194, 4195, 4193, 4200, 4197, 4199, 4205, 4204, 4203, 4208, 4209, 4207, 4214, 4211, 4213, 4191, 4204, 4192, 4196, 4207, 4195, 4212, 4199, 4211, 4193, 4188, 4187, 4201, 4209, 4202, 4199, 4190, 4192, 4206, 4211, 4204, 4188, 4189, 4187, 4201, 4205, 4203, 4201, 4191, 4188, 4195, 4201, 4188, 4204, 4199, 4192, 4222, 4219, 4221, 4226, 4223, 4225, 4220, 4215, 4219, 4224, 4217, 4223, 4225, 4220, 4222, 4223, 4216, 4220, 4229, 4232, 4230, 4234, 4235, 4233, 4242, 4237, 4240, 4243, 4246, 4244, 4249, 4251, 4247, 4243, 4232, 4231, 4241, 4247, 4239, 4250, 4240, 4248,
4233, 4228, 4227, 4232, 4237, 4230, 4243, 4251, 4245, 4228, 4229, 4227, 4241, 4235, 4236, 4235, 4231, 4228, 4239, 4243, 4231, 4248, 4232, 4244, 4258, 4255, 4257, 4254, 4255, 4253, 4205, 4230, 4206, 4210, 4233, 4209, 4238, 4213, 4237, 4202, 4229, 4205, 4202, 4233, 4227, 4230, 4213, 4206, 4221, 4197, 4198, 4219, 4190, 4197, 4246, 4253, 4244, 4255, 4250, 4248, 4244, 4255, 4248, 4267, 4264, 4266, 4271, 4268, 4270, 4271, 4272, 4269, 4277, 4278, 4276, 4273, 4276, 4272, 4265, 4259, 4264, 4269, 4261, 4268, 4269, 4263, 4262, 4276, 4275, 4274, 4270, 4265, 4267, 4268, 4260, 4265, 4272, 4274, 4263, 4282, 4281, 4280, 4287, 4284, 4286, 4290, 4289, 4288, 4295, 4292, 4294, 4282, 4289, 4283, 4293, 4286, 4292, 4286, 4281, 4283, 4294, 4289, 4291, 4292, 4283, 4289, 4299, 4300, 4298, 4308, 4309, 4307, 4306, 4308, 4305, 4301, 4305, 4300, 4298, 4297, 4296, 4304, 4307, 4303, 4302, 4308, 4304, 4300, 4302, 4297, 4313, 4312, 4311, 4318, 4315, 4317, 4319, 4322, 4320, 4328, 4325, 4327, 4328, 4323, 4326, 4313, 4320, 4314, 4324, 4317, 4323, 4317, 4312, 4314, 4326, 4321, 4325, 4322, 4323, 4320,
4323, 4314, 4320, 4279, 4298, 4278, 4278, 4296, 4275, 4290, 4312, 4291, 4316, 4294, 4315, 4315, 4291, 4312, 4263, 4280, 4262, 4274, 4290, 4288, 4274, 4282, 4263, 4296, 4313, 4311, 4304, 4319, 4302, 4297, 4319, 4313, 4275, 4311, 4290, 4189, 4136, 4135, 4140, 4193, 4139, 4187, 4135, 4132, 4139, 4187, 4132, 4215, 4148, 4147, 4217, 4152, 4151, 4216, 4151, 4148, 4190, 4147, 4136, 4259, 4164, 4163, 4261, 4168, 4167, 4260, 4167, 4164, 4280, 4180, 4179, 4285, 4185, 4284, 4284, 4180, 4281, 4262, 4179, 4168, 4152, 4161, 4150, 4169, 4160, 4158, 4161, 4158, 4150, 4266, 4224, 4226, 4264, 4218, 4224, 4218, 4163, 4152, 4331, 4330, 4329, 4334, 4335, 4333, 4339, 4338, 4337, 4342, 4343, 4341, 4337, 4332, 4331, 4336, 4341, 4335, 4333, 4331, 4329, 4341, 4339, 4337, 4335, 4337, 4331, 4347, 4346, 4345, 4350, 4351, 4349, 4355, 4354, 4353, 4358, 4359, 4357, 4353, 4348, 4347, 4352, 4357, 4351, 4349, 4347, 4345, 4357, 4355, 4353, 4351, 4353, 4347, 4345, 4340, 4339, 4344, 4349, 4343, 4343, 4345, 4339, 4363, 4362, 4361, 4367, 4366, 4365, 4371, 4370, 4369, 4375, 4374, 4373, 4369, 4364, 4363,
4373, 4368, 4367, 4364, 4365, 4362, 4372, 4373, 4370, 4370, 4367, 4364, 4382, 4379, 4381, 4388, 4385, 4387, 4386, 4381, 4385, 4381, 4377, 4378, 4387, 4383, 4384, 4385, 4378, 4383, 4391, 4390, 4389, 4395, 4394, 4393, 4399, 4398, 4397, 4403, 4402, 4401, 4397, 4392, 4391, 4401, 4396, 4395, 4392, 4393, 4390, 4400, 4401, 4398, 4398, 4395, 4392, 4410, 4407, 4409, 4416, 4413, 4415, 4414, 4409, 4413, 4409, 4405, 4406, 4415, 4411, 4412, 4413, 4406, 4411, 4389, 4372, 4371, 4393, 4376, 4375, 4390, 4375, 4372, 4408, 4387, 4407, 4407, 4384, 4405, 4366, 4378, 4377, 4376, 4383, 4374, 4368, 4383, 4378, 4394, 4406, 4405, 4404, 4411, 4402, 4396, 4411, 4406, 4394, 4384, 4376, 4419, 4418, 4417, 4422, 4423, 4421, 4427, 4426, 4425, 4430, 4431, 4429, 4425, 4420, 4419, 4424, 4429, 4423, 4421, 4419, 4417, 4429, 4427, 4425, 4423, 4425, 4419, 4435, 4434, 4433, 4438, 4439, 4437, 4443, 4442, 4441, 4446, 4447, 4445, 4441, 4436, 4435, 4440, 4445, 4439, 4437, 4435, 4433, 4445, 4443, 4441, 4439, 4441, 4435, 4433, 4428, 4427, 4432, 4437, 4431, 4431, 4433, 4427, 4451, 4450, 4449, 4455, 4454, 4453,
4459, 4458, 4457, 4463, 4462, 4461, 4457, 4452, 4451, 4461, 4456, 4455, 4452, 4453, 4450, 4460, 4461, 4458, 4458, 4455, 4452, 4470, 4467, 4469, 4476, 4473, 4475, 4474, 4469, 4473, 4469, 4465, 4466, 4475, 4471, 4472, 4473, 4466, 4471, 4479, 4478, 4477, 4483, 4482, 4481, 4487, 4486, 4485, 4491, 4490, 4489, 4485, 4480, 4479, 4489, 4484, 4483, 4480, 4481, 4478, 4488, 4489, 4486, 4486, 4483, 4480, 4498, 4495, 4497, 4504, 4501, 4503, 4502, 4497, 4501, 4497, 4493, 4494, 4503, 4499, 4500, 4501, 4494, 4499, 4477, 4460, 4459, 4481, 4464, 4463, 4478, 4463, 4460, 4496, 4475, 4495, 4495, 4472, 4493, 4456, 4465, 4454, 4464, 4471, 4462, 4462, 4466, 4456, 4484, 4493, 4482, 4492, 4499, 4490, 4490, 4494, 4484, 4482, 4472, 4464, 4417, 4356, 4355, 4360, 4421, 4359, 4359, 4417, 4355, 4449, 4400, 4399, 4453, 4404, 4403, 4450, 4403, 4400, 4468, 4415, 4467, 4467, 4412, 4465, 4454, 4412, 4404, 4332, 4361, 4330, 4340, 4369, 4338, 4338, 4363, 4332, 4348, 4389, 4346, 4356, 4397, 4354, 4354, 4391, 4348, 4346, 4371, 4340, 4420, 4449, 4418, 4428, 4457, 4426, 4426, 4451, 4420, 4436, 4477, 4434,
4444, 4485, 4442, 4442, 4479, 4436, 4434, 4459, 4428, 4418, 4399, 4356, 4511, 4508, 4510, 4511, 4512, 4509, 4508, 4506, 4505, 4509, 4507, 4506, 4514, 4517, 4515, 4522, 4523, 4521, 4530, 4531, 4529, 4531, 4527, 4529, 4540, 4537, 4539, 4544, 4541, 4543, 4533, 4517, 4516, 4535, 4520, 4519, 4526, 4537, 4525, 4514, 4522, 4521, 4537, 4534, 4533, 4520, 4529, 4519, 4519, 4527, 4518, 4541, 4536, 4535, 4517, 4518, 4515, 4527, 4524, 4522, 4543, 4538, 4540, 4518, 4522, 4515, 4538, 4535, 4534, 4517, 4535, 4519, 4516, 4537, 4533, 4506, 4516, 4514, 4526, 4512, 4513, 4525, 4507, 4512, 4545, 4548, 4546, 4551, 4550, 4549, 4554, 4555, 4553, 4555, 4557, 4553, 4564, 4561, 4563, 4571, 4568, 4570, 4565, 4548, 4547, 4573, 4561, 4572, 4546, 4553, 4545, 4553, 4547, 4545, 4569, 4565, 4568, 4562, 4551, 4561, 4548, 4549, 4546, 4559, 4556, 4554, 4573, 4569, 4571, 4549, 4554, 4546, 4569, 4567, 4566, 4548, 4567, 4551, 4572, 4551, 4567, 4579, 4576, 4578, 4578, 4574, 4575, 4550, 4575, 4574, 4579, 4562, 4564, 4578, 4552, 4562, 4586, 4587, 4585, 4582, 4585, 4581, 4585, 4584, 4583, 4580, 4585, 4583,
4595, 4596, 4594, 4593, 4595, 4592, 4600, 4599, 4598, 4603, 4604, 4602, 4590, 4599, 4591, 4596, 4602, 4594, 4590, 4595, 4594, 4592, 4591, 4589, 4602, 4600, 4598, 4594, 4598, 4590, 4610, 4609, 4608, 4612, 4615, 4613, 4617, 4618, 4616, 4625, 4622, 4624, 4632, 4629, 4631, 4634, 4629, 4633, 4610, 4627, 4611, 4632, 4620, 4630, 4609, 4616, 4608, 4629, 4627, 4626, 4633, 4626, 4628, 4623, 4614, 4622, 4611, 4612, 4609, 4624, 4620, 4621, 4622, 4611, 4620, 4630, 4611, 4627, 4581, 4613, 4582, 4584, 4617, 4583, 4580, 4612, 4581, 4583, 4609, 4580, 4589, 4610, 4608, 4616, 4593, 4592, 4599, 4628, 4626, 4634, 4605, 4607, 4608, 4592, 4589, 4633, 4601, 4605, 4599, 4610, 4591, 4639, 4640, 4638, 4635, 4639, 4638, 4642, 4641, 4639, 4636, 4642, 4639, 4646, 4645, 4644, 4650, 4649, 4648, 4655, 4652, 4654, 4657, 4658, 4656, 4661, 4657, 4660, 4665, 4664, 4663, 4674, 4671, 4673, 4670, 4671, 4669, 4650, 4664, 4651, 4655, 4667, 4653, 4653, 4646, 4652, 4649, 4656, 4648, 4660, 4649, 4651, 4671, 4666, 4665, 4669, 4665, 4663, 4645, 4650, 4648, 4670, 4667, 4668, 4667, 4663, 4662, 4662, 4650, 4647,
4647, 4667, 4662, 4684, 4685, 4683, 4685, 4681, 4683, 4682, 4676, 4681, 4679, 4684, 4683, 4679, 4681, 4678, 4678, 4676, 4675, 4644, 4636, 4635, 4658, 4642, 4656, 4645, 4637, 4636, 4648, 4642, 4637, 4676, 4661, 4660, 4664, 4679, 4678, 4651, 4676, 4660, 4664, 4675, 4651, 4557, 4532, 4530, 4570, 4542, 4544, 4520, 4557, 4530, 4568, 4536, 4542, 4520, 4565, 4547, 4638, 4588, 4586, 4582, 4638, 4586, 4615, 4644, 4613, 4654, 4623, 4625, 4623, 4646, 4615, 4613, 4635, 4582, 4689, 4688, 4687, 4692, 4693, 4691, 4697, 4696, 4695, 4700, 4701, 4699, 4695, 4690, 4689, 4694, 4699, 4693, 4691, 4689, 4687, 4699, 4697, 4695, 4693, 4695, 4689, 4705, 4704, 4703, 4712, 4709, 4711, 4708, 4709, 4707, 4710, 4705, 4709, 4707, 4705, 4703, 4703, 4698, 4697, 4702, 4707, 4701, 4701, 4703, 4697, 4715, 4714, 4713, 4719, 4718, 4717, 4723, 4722, 4721, 4727, 4726, 4725, 4721, 4716, 4715, 4725, 4720, 4719, 4716, 4717, 4714, 4724, 4725, 4722, 4722, 4719, 4716, 4734, 4731, 4733, 4740, 4737, 4739, 4738, 4733, 4737, 4733, 4729, 4730, 4739, 4735, 4736, 4737, 4730, 4735, 4743, 4742, 4741, 4747, 4746, 4745,
4752, 4749, 4751, 4756, 4753, 4755, 4750, 4743, 4749, 4754, 4747, 4753, 4744, 4745, 4742, 4755, 4750, 4752, 4753, 4744, 4750, 4762, 4759, 4761, 4761, 4757, 4758, 4741, 4724, 4723, 4745, 4728, 4727, 4742, 4727, 4724, 4760, 4739, 4759, 4759, 4736, 4757, 4720, 4729, 4718, 4728, 4735, 4726, 4726, 4730, 4720, 4748, 4757, 4746, 4762, 4754, 4756, 4761, 4748, 4754, 4746, 4736, 4728, 4766, 4767, 4765, 4764, 4765, 4763, 4771, 4770, 4769, 4774, 4775, 4773, 4779, 4778, 4777, 4782, 4783, 4781, 4777, 4772, 4771, 4776, 4781, 4775, 4773, 4771, 4769, 4781, 4779, 4777, 4775, 4777, 4771, 4769, 4764, 4763, 4767, 4773, 4765, 4765, 4769, 4763, 4790, 4791, 4789, 4794, 4795, 4793, 4786, 4789, 4785, 4788, 4793, 4787, 4793, 4792, 4790, 4787, 4790, 4786, 4802, 4803, 4801, 4806, 4807, 4805, 4798, 4801, 4797, 4800, 4805, 4799, 4805, 4804, 4802, 4799, 4802, 4798, 4811, 4810, 4809, 4815, 4814, 4813, 4819, 4818, 4817, 4823, 4822, 4821, 4817, 4812, 4811, 4821, 4816, 4815, 4812, 4813, 4810, 4820, 4821, 4818, 4818, 4815, 4812, 4827, 4826, 4825, 4831, 4830, 4829, 4835, 4834, 4833, 4839, 4838, 4837,
4833, 4828, 4827, 4837, 4832, 4831, 4828, 4829, 4826, 4836, 4837, 4834, 4834, 4831, 4828, 4809, 4786, 4785, 4813, 4788, 4787, 4810, 4787, 4786, 4825, 4798, 4797, 4829, 4800, 4799, 4826, 4799, 4798, 4801, 4796, 4794, 4797, 4794, 4788, 4816, 4825, 4814, 4824, 4833, 4822, 4822, 4827, 4816, 4814, 4797, 4788, 4690, 4713, 4688, 4698, 4721, 4696, 4696, 4715, 4690, 4706, 4741, 4704, 4751, 4710, 4712, 4749, 4706, 4710, 4704, 4723, 4698, 4789, 4768, 4766, 4785, 4766, 4764, 4772, 4809, 4770, 4780, 4817, 4778, 4778, 4811, 4772, 4770, 4785, 4764, 4505, 4246, 4245, 4252, 4508, 4251, 4245, 4508, 4505, 4523, 4257, 4521, 4514, 4257, 4254, 4246, 4514, 4254, 4303, 4550, 4304, 4309, 4559, 4307, 4303, 4559, 4549, 4577, 4325, 4576, 4576, 4321, 4574, 4550, 4321, 4304, 4687, 4444, 4443, 4448, 4691, 4447, 4447, 4687, 4443, 4713, 4488, 4487, 4717, 4492, 4491, 4714, 4491, 4488, 4732, 4503, 4731, 4731, 4500, 4729, 4718, 4500, 4492, 4688, 4487, 4444, 4845, 4844, 4843, 4848, 4849, 4847, 4856, 4852, 4855, 4859, 4862, 4860, 4864, 4865, 4863, 4870, 4867, 4869, 4845, 4860, 4846, 4850, 4863, 4849,
4868, 4855, 4867, 4841, 4849, 4842, 4857, 4865, 4858, 4855, 4844, 4846, 4862, 4867, 4860, 4841, 4845, 4843, 4851, 4848, 4847, 4857, 4861, 4859, 4843, 4847, 4841, 4842, 4859, 4845, 4849, 4857, 4842, 4867, 4846, 4860, 4873, 4876, 4874, 4878, 4879, 4877, 4886, 4881, 4884, 4889, 4888, 4887, 4896, 4893, 4895, 4892, 4893, 4891, 4875, 4888, 4876, 4885, 4891, 4883, 4877, 4872, 4871, 4884, 4874, 4876, 4893, 4890, 4889, 4887, 4893, 4889, 4872, 4873, 4871, 4885, 4879, 4880, 4879, 4875, 4872, 4875, 4891, 4887, 4907, 4908, 4906, 4908, 4904, 4906, 4905, 4898, 4904, 4901, 4907, 4906, 4901, 4904, 4900, 4910, 4909, 4907, 4902, 4910, 4907, 4900, 4898, 4897, 4861, 4874, 4862, 4866, 4877, 4865, 4882, 4869, 4881, 4871, 4861, 4858, 4858, 4877, 4871, 4874, 4869, 4862, 4898, 4886, 4884, 4888, 4901, 4900, 4876, 4898, 4884, 4888, 4897, 4876, 4915, 4916, 4914, 4921, 4922, 4920, 4917, 4920, 4916, 4912, 4916, 4913, 4920, 4919, 4918, 4913, 4920, 4918, 4926, 4925, 4924, 4931, 4928, 4930, 4934, 4933, 4932, 4939, 4936, 4938, 4926, 4933, 4927, 4937, 4930, 4936, 4927, 4928, 4925, 4938, 4933, 4935,
4933, 4930, 4927, 4943, 4944, 4942, 4947, 4950, 4948, 4954, 4955, 4953, 4955, 4957, 4953, 4945, 4954, 4944, 4942, 4941, 4940, 4948, 4953, 4947, 4947, 4957, 4949, 4957, 4952, 4951, 4946, 4957, 4951, 4941, 4954, 4948, 4959, 4962, 4960, 4966, 4963, 4965, 4973, 4970, 4972, 4975, 4970, 4974, 4961, 4968, 4962, 4973, 4965, 4971, 4965, 4960, 4962, 4971, 4967, 4970, 4974, 4967, 4969, 4968, 4965, 4962, 4923, 4942, 4922, 4922, 4940, 4919, 4959, 4935, 4934, 4964, 4938, 4963, 4963, 4935, 4960, 4913, 4924, 4912, 4918, 4934, 4932, 4918, 4926, 4913, 4940, 4961, 4959, 4950, 4967, 4948, 4948, 4961, 4941, 4919, 4959, 4934, 4981, 4980, 4979, 4985, 4984, 4983, 4992, 4989, 4991, 4988, 4989, 4987, 4994, 4995, 4993, 5000, 4997, 4999, 4990, 4981, 4989, 4979, 4989, 4981, 4984, 4993, 4983, 4997, 4986, 4985, 4980, 4985, 4983, 4999, 4990, 4992, 4997, 4982, 4990, 4988, 4977, 4978, 4987, 4976, 4977, 5001, 5004, 5002, 5005, 5008, 5006, 5010, 5011, 5009, 5016, 5013, 5015, 5022, 5019, 5021, 5022, 5017, 5020, 5001, 5010, 5009, 5014, 5003, 5013, 5020, 5007, 5019, 5008, 5017, 5006, 5004, 5005, 5002,
5021, 5014, 5016, 5019, 5004, 5014, 5025, 5017, 5018, 5024, 5006, 5017, 4896, 4977, 4894, 4894, 4976, 4890, 4901, 4980, 4902, 4995, 4910, 4993, 4980, 4903, 4902, 4983, 4910, 4903, 4890, 4979, 4901, 5005, 4950, 4949, 4952, 5010, 4951, 4946, 5005, 4949, 4951, 5002, 4946, 5025, 4974, 5024, 5024, 4969, 5023, 4950, 5023, 4969, 4984, 5003, 5001, 5009, 4996, 4994, 5015, 4998, 5000, 5001, 4994, 4984, 5013, 4986, 4998, 5031, 5028, 5030, 5029, 5026, 5028, 5039, 5036, 5038, 5043, 5040, 5042, 5037, 5032, 5036, 5041, 5034, 5040, 5042, 5037, 5039, 5040, 5033, 5037, 5051, 5048, 5050, 5055, 5052, 5054, 5049, 5044, 5048, 5053, 5046, 5052, 5054, 5049, 5051, 5052, 5045, 5049, 5050, 5041, 5043, 5048, 5035, 5041, 5058, 5057, 5056, 5061, 5062, 5060, 5062, 5064, 5060, 5068, 5067, 5066, 5071, 5072, 5070, 5066, 5059, 5058, 5065, 5070, 5064, 5057, 5060, 5056, 5060, 5058, 5056, 5070, 5068, 5066, 5064, 5066, 5058, 5076, 5075, 5074, 5083, 5080, 5082, 5079, 5080, 5078, 5081, 5076, 5080, 5078, 5076, 5074, 5074, 5069, 5068, 5073, 5078, 5072, 5072, 5074, 5068, 5086, 5085, 5084, 5090, 5089, 5088,
5093, 5094, 5092, 5097, 5098, 5096, 5102, 5101, 5100, 5106, 5105, 5104, 5100, 5087, 5086, 5104, 5091, 5090, 5085, 5092, 5084, 5089, 5096, 5088, 5087, 5088, 5085, 5096, 5095, 5093, 5103, 5104, 5101, 5088, 5093, 5085, 5101, 5090, 5087, 5110, 5109, 5108, 5114, 5113, 5112, 5117, 5118, 5116, 5121, 5122, 5120, 5126, 5125, 5124, 5130, 5129, 5128, 5124, 5111, 5110, 5128, 5115, 5114, 5109, 5116, 5108, 5113, 5120, 5112, 5111, 5112, 5109, 5120, 5119, 5117, 5127, 5128, 5125, 5112, 5117, 5109, 5125, 5114, 5111, 5134, 5133, 5132, 5138, 5137, 5136, 5143, 5140, 5142, 5147, 5144, 5146, 5141, 5134, 5140, 5145, 5138, 5144, 5135, 5136, 5133, 5146, 5141, 5143, 5144, 5135, 5141, 5150, 5149, 5148, 5154, 5153, 5152, 5159, 5156, 5158, 5163, 5160, 5162, 5157, 5150, 5156, 5161, 5154, 5160, 5151, 5152, 5149, 5162, 5157, 5159, 5160, 5151, 5157, 5132, 5103, 5102, 5136, 5107, 5106, 5133, 5106, 5103, 5148, 5127, 5126, 5152, 5131, 5130, 5149, 5130, 5127, 5091, 5108, 5089, 5116, 5099, 5097, 5107, 5124, 5105, 5108, 5097, 5089, 5105, 5110, 5091, 5139, 5148, 5137, 5158, 5145, 5147, 5156, 5139, 5145,
5137, 5126, 5107, 5038, 5029, 5031, 5036, 5027, 5029, 5059, 5084, 5057, 5092, 5063, 5061, 5069, 5100, 5067, 5084, 5061, 5057, 5067, 5086, 5059, 5077, 5132, 5075, 5142, 5081, 5083, 5140, 5077, 5081, 5075, 5102, 5069, 5168, 5167, 5166, 5171, 5172, 5170, 5175, 5176, 5174, 5170, 5165, 5164, 5167, 5174, 5166, 5165, 5166, 5164, 5174, 5171, 5170, 5166, 5170, 5164, 5180, 5179, 5178, 5184, 5183, 5182, 5187, 5188, 5186, 5191, 5192, 5190, 5179, 5186, 5178, 5183, 5190, 5182, 5181, 5182, 5179, 5190, 5189, 5187, 5182, 5187, 5179, 5202, 5199, 5201, 5200, 5195, 5199, 5201, 5197, 5198, 5199, 5194, 5197, 5210, 5207, 5209, 5214, 5211, 5213, 5221, 5222, 5220, 5219, 5221, 5218, 5208, 5203, 5207, 5212, 5205, 5211, 5216, 5221, 5220, 5215, 5221, 5217, 5213, 5208, 5210, 5211, 5204, 5208, 5195, 5169, 5168, 5173, 5197, 5172, 5194, 5168, 5165, 5172, 5194, 5165, 5203, 5181, 5180, 5205, 5185, 5184, 5204, 5184, 5181, 5169, 5178, 5167, 5186, 5177, 5175, 5178, 5175, 5167, 5209, 5200, 5202, 5207, 5196, 5200, 5196, 5180, 5169, 5226, 5225, 5224, 5230, 5229, 5228, 5233, 5234, 5232, 5237, 5238, 5236,
5241, 5237, 5240, 5224, 5233, 5232, 5229, 5236, 5228, 5231, 5237, 5229, 5225, 5230, 5228, 5236, 5235, 5233, 5225, 5236, 5233, 5251, 5250, 5249, 5256, 5253, 5255, 5258, 5259, 5257, 5256, 5258, 5254, 5265, 5264, 5263, 5277, 5274, 5276, 5279, 5274, 5278, 5247, 5264, 5248, 5251, 5268, 5252, 5259, 5270, 5257, 5277, 5261, 5275, 5254, 5245, 5253, 5248, 5257, 5247, 5246, 5258, 5248, 5275, 5267, 5274, 5269, 5274, 5267, 5246, 5251, 5249, 5264, 5269, 5267, 5279, 5272, 5273, 5278, 5266, 5272, 5264, 5251, 5248, 5247, 5270, 5263, 5268, 5261, 5252, 5285, 5282, 5284, 5281, 5282, 5280, 5245, 5227, 5226, 5230, 5250, 5231, 5227, 5249, 5230, 5283, 5243, 5282, 5282, 5242, 5280, 5243, 5241, 5240, 5242, 5240, 5231, 5250, 5281, 5280, 5285, 5261, 5262, 5284, 5252, 5261, 5250, 5242, 5231, 5292, 5293, 5291, 5290, 5292, 5289, 5287, 5292, 5291, 5289, 5288, 5286, 5300, 5299, 5298, 5305, 5306, 5304, 5303, 5305, 5302, 5314, 5313, 5312, 5326, 5323, 5325, 5328, 5323, 5327, 5296, 5313, 5297, 5300, 5317, 5301, 5306, 5319, 5304, 5326, 5310, 5324, 5297, 5304, 5296, 5295, 5305, 5297, 5324, 5316, 5323,
5318, 5323, 5316, 5295, 5300, 5298, 5308, 5303, 5302, 5313, 5318, 5316, 5328, 5321, 5322, 5298, 5302, 5295, 5327, 5315, 5321, 5313, 5300, 5297, 5296, 5319, 5312, 5317, 5310, 5301, 5330, 5333, 5331, 5337, 5338, 5336, 5348, 5345, 5347, 5341, 5333, 5332, 5335, 5343, 5334, 5336, 5332, 5330, 5346, 5341, 5345, 5338, 5335, 5334, 5347, 5343, 5344, 5332, 5334, 5329, 5345, 5340, 5343, 5329, 5341, 5332, 5329, 5343, 5340, 5349, 5352, 5350, 5355, 5354, 5353, 5360, 5361, 5359, 5370, 5367, 5369, 5374, 5371, 5373, 5363, 5352, 5351, 5365, 5356, 5355, 5368, 5363, 5367, 5354, 5359, 5353, 5372, 5365, 5371, 5350, 5355, 5353, 5359, 5358, 5357, 5373, 5368, 5370, 5353, 5357, 5350, 5371, 5364, 5368, 5364, 5355, 5352, 5287, 5331, 5288, 5293, 5336, 5291, 5291, 5330, 5287, 5314, 5350, 5315, 5358, 5321, 5357, 5350, 5321, 5315, 5286, 5314, 5312, 5319, 5290, 5289, 5312, 5289, 5286, 5333, 5349, 5331, 5369, 5346, 5348, 5367, 5342, 5346, 5342, 5351, 5333, 5331, 5314, 5288, 5381, 5378, 5380, 5383, 5378, 5382, 5379, 5375, 5378, 5377, 5378, 5375, 5386, 5385, 5384, 5390, 5389, 5388, 5393, 5394, 5392,
5397, 5398, 5396, 5407, 5404, 5406, 5411, 5408, 5410, 5400, 5387, 5386, 5402, 5391, 5390, 5385, 5392, 5384, 5405, 5400, 5404, 5389, 5396, 5388, 5409, 5402, 5408, 5387, 5388, 5385, 5396, 5395, 5393, 5410, 5405, 5407, 5388, 5393, 5385, 5408, 5401, 5405, 5401, 5390, 5387, 5414, 5413, 5412, 5417, 5418, 5416, 5421, 5417, 5420, 5427, 5424, 5426, 5422, 5415, 5414, 5427, 5420, 5425, 5413, 5416, 5412, 5420, 5413, 5415, 5425, 5422, 5424, 5425, 5415, 5423, 5391, 5412, 5389, 5416, 5399, 5397, 5426, 5409, 5411, 5412, 5397, 5389, 5424, 5403, 5409, 5403, 5414, 5391, 5216, 5299, 5217, 5222, 5308, 5220, 5216, 5308, 5298, 5265, 5376, 5266, 5381, 5272, 5379, 5376, 5272, 5266, 5185, 5224, 5183, 5232, 5193, 5191, 5224, 5191, 5183, 5255, 5212, 5214, 5215, 5265, 5263, 5270, 5219, 5218, 5253, 5206, 5212, 5263, 5218, 5215, 5206, 5226, 5185, 5299, 5377, 5375, 5383, 5310, 5311, 5382, 5301, 5310, 5356, 5384, 5354, 5392, 5362, 5360, 5406, 5372, 5374, 5384, 5360, 5354, 5404, 5366, 5372, 5366, 5386, 5356, 5299, 5265, 5217, 4600, 4844, 4601, 4606, 4851, 4604, 4854, 4605, 4852, 4600, 4851, 4843,
4844, 4605, 4601, 4674, 4914, 4672, 4666, 4914, 4912, 4924, 4680, 4679, 4929, 4684, 4928, 4928, 4680, 4925, 4666, 4924, 4679, 5026, 4780, 4779, 4784, 5028, 4783, 4783, 5026, 4779, 5032, 4820, 4819, 5034, 4824, 4823, 5033, 4823, 4820, 5044, 4836, 4835, 5046, 4840, 4839, 5045, 4839, 4836, 5035, 4835, 4824, 5027, 4819, 4780, 4329, 4045, 4044, 4049, 4333, 4048, 4048, 4329, 4044, 4361, 4101, 4100, 4365, 4105, 4104, 4362, 4104, 4101, 4380, 4121, 4379, 4379, 4116, 4377, 4105, 4377, 4116, 4330, 4100, 4045, 5431, 5432, 5430, 5434, 5437, 5435, 5440, 5439, 5438, 5445, 5446, 5444, 5434, 5429, 5428, 5443, 5431, 5442, 5429, 5430, 5428, 5438, 5445, 5444, 5437, 5438, 5435, 5444, 5443, 5442, 5438, 5442, 5435, 5435, 5431, 5429, 5450, 5449, 5448, 5455, 5456, 5454, 5459, 5455, 5458, 5461, 5462, 5460, 5449, 5454, 5448, 5451, 5455, 5449, 5452, 5461, 5460, 5460, 5459, 5458, 5452, 5458, 5451, 5466, 5465, 5464, 5470, 5469, 5468, 5474, 5473, 5472, 5478, 5477, 5476, 5472, 5467, 5466, 5476, 5471, 5470, 5467, 5468, 5465, 5475, 5476, 5473, 5473, 5470, 5467, 5482, 5481, 5480, 5486, 5485, 5484,
5490, 5489, 5488, 5494, 5493, 5492, 5488, 5483, 5482, 5492, 5487, 5486, 5483, 5484, 5481, 5491, 5492, 5489, 5489, 5486, 5483, 5464, 5437, 5436, 5468, 5441, 5440, 5465, 5440, 5437, 5450, 5481, 5451, 5484, 5453, 5452, 5481, 5452, 5451, 5441, 5448, 5439, 5454, 5447, 5445, 5439, 5454, 5445, 5471, 5480, 5469, 5479, 5488, 5477, 5477, 5482, 5471, 5469, 5450, 5441, 5501, 5500, 5499, 5505, 5504, 5503, 5508, 5509, 5507, 5512, 5513, 5511, 5516, 5512, 5515, 5519, 5518, 5517, 5523, 5522, 5521, 5517, 5502, 5501, 5505, 5522, 5506, 5500, 5507, 5499, 5504, 5511, 5503, 5506, 5512, 5504, 5502, 5503, 5500, 5511, 5510, 5508, 5520, 5521, 5518, 5500, 5511, 5508, 5518, 5505, 5502, 5529, 5530, 5528, 5536, 5535, 5534, 5540, 5539, 5538, 5543, 5544, 5542, 5534, 5526, 5525, 5544, 5532, 5542, 5525, 5529, 5528, 5538, 5543, 5542, 5532, 5531, 5529, 5535, 5540, 5538, 5526, 5532, 5529, 5535, 5527, 5526, 5542, 5527, 5538, 5509, 5497, 5507, 5499, 5497, 5496, 5528, 5516, 5515, 5524, 5534, 5522, 5525, 5515, 5506, 5522, 5525, 5506, 5548, 5547, 5546, 5552, 5551, 5550, 5557, 5554, 5556, 5566, 5563, 5565,
5559, 5553, 5552, 5557, 5561, 5555, 5555, 5548, 5554, 5563, 5560, 5559, 5547, 5552, 5550, 5565, 5561, 5562, 5563, 5558, 5561, 5558, 5552, 5549, 5549, 5561, 5558, 5569, 5568, 5567, 5573, 5572, 5571, 5577, 5576, 5575, 5581, 5580, 5579, 5588, 5585, 5587, 5584, 5585, 5583, 5592, 5589, 5591, 5569, 5576, 5570, 5579, 5574, 5573, 5586, 5577, 5585, 5575, 5585, 5577, 5590, 5581, 5589, 5570, 5571, 5568, 5578, 5579, 5576, 5591, 5586, 5588, 5586, 5581, 5578, 5576, 5573, 5570, 5592, 5594, 5590, 5582, 5594, 5593, 5553, 5567, 5551, 5584, 5564, 5566, 5583, 5560, 5564, 5560, 5569, 5553, 5598, 5597, 5596, 5602, 5601, 5600, 5606, 5605, 5604, 5610, 5609, 5608, 5604, 5599, 5598, 5608, 5603, 5602, 5599, 5600, 5597, 5607, 5608, 5605, 5605, 5602, 5599, 5614, 5613, 5612, 5618, 5617, 5616, 5622, 5621, 5620, 5626, 5625, 5624, 5620, 5615, 5614, 5624, 5619, 5618, 5615, 5616, 5613, 5623, 5624, 5621, 5621, 5618, 5615, 5632, 5631, 5630, 5637, 5634, 5636, 5643, 5640, 5642, 5639, 5640, 5638, 5634, 5629, 5628, 5641, 5632, 5640, 5630, 5640, 5632, 5639, 5635, 5637, 5638, 5629, 5635, 5644, 5647, 5645,
5650, 5649, 5648, 5658, 5657, 5656, 5663, 5660, 5662, 5669, 5666, 5668, 5665, 5666, 5664, 5654, 5647, 5646, 5650, 5657, 5651, 5653, 5660, 5652, 5660, 5655, 5654, 5666, 5659, 5658, 5656, 5666, 5658, 5647, 5648, 5645, 5665, 5661, 5663, 5661, 5656, 5655, 5655, 5650, 5647, 5646, 5660, 5654, 5628, 5607, 5606, 5610, 5631, 5611, 5629, 5610, 5607, 5644, 5623, 5622, 5648, 5627, 5626, 5645, 5626, 5623, 5603, 5612, 5601, 5611, 5620, 5609, 5609, 5614, 5603, 5633, 5644, 5631, 5653, 5641, 5643, 5641, 5646, 5633, 5631, 5622, 5611, 5546, 5475, 5474, 5550, 5479, 5478, 5547, 5478, 5475, 5567, 5491, 5490, 5571, 5495, 5494, 5568, 5494, 5491, 5551, 5490, 5479, 5596, 5520, 5519, 5600, 5524, 5523, 5597, 5523, 5520, 5612, 5537, 5536, 5616, 5541, 5540, 5613, 5540, 5537, 5601, 5536, 5524, 5497, 5463, 5461, 5496, 5461, 5453, 5485, 5501, 5499, 5495, 5517, 5493, 5493, 5501, 5487, 5485, 5496, 5453, 5574, 5596, 5572, 5580, 5606, 5604, 5580, 5598, 5574, 5636, 5594, 5595, 5634, 5593, 5594, 5593, 5606, 5582, 5572, 5519, 5495, 5677, 5678, 5676, 5678, 5674, 5676, 5671, 5677, 5676, 5671, 5674, 5670,
5680, 5679, 5677, 5672, 5680, 5677, 5684, 5683, 5682, 5688, 5687, 5686, 5691, 5692, 5690, 5696, 5695, 5694, 5700, 5699, 5698, 5694, 5685, 5684, 5698, 5689, 5688, 5686, 5691, 5690, 5683, 5688, 5686, 5697, 5698, 5695, 5695, 5688, 5685, 5708, 5709, 5707, 5709, 5705, 5707, 5711, 5714, 5712, 5717, 5716, 5715, 5722, 5723, 5721, 5726, 5722, 5725, 5711, 5704, 5703, 5720, 5708, 5719, 5704, 5707, 5703, 5703, 5705, 5702, 5715, 5722, 5721, 5718, 5722, 5716, 5714, 5715, 5712, 5721, 5720, 5719, 5712, 5721, 5719, 5712, 5708, 5704, 5729, 5728, 5727, 5733, 5732, 5731, 5735, 5738, 5736, 5741, 5740, 5739, 5750, 5747, 5749, 5750, 5745, 5748, 5735, 5730, 5729, 5739, 5734, 5733, 5748, 5741, 5747, 5748, 5740, 5742, 5730, 5731, 5728, 5738, 5739, 5736, 5736, 5733, 5730, 5753, 5752, 5751, 5757, 5756, 5755, 5770, 5767, 5769, 5772, 5773, 5771, 5770, 5772, 5768, 5778, 5775, 5777, 5780, 5781, 5779, 5781, 5775, 5779, 5759, 5754, 5753, 5757, 5764, 5758, 5768, 5759, 5767, 5762, 5771, 5761, 5768, 5762, 5760, 5775, 5764, 5763, 5765, 5780, 5779, 5765, 5775, 5763, 5754, 5755, 5752, 5762, 5763, 5760,
5760, 5757, 5754, 5727, 5697, 5696, 5731, 5701, 5700, 5728, 5700, 5697, 5751, 5714, 5713, 5717, 5756, 5718, 5752, 5717, 5714, 5689, 5702, 5687, 5705, 5693, 5691, 5701, 5711, 5699, 5687, 5705, 5691, 5689, 5711, 5703, 5734, 5751, 5732, 5769, 5745, 5746, 5771, 5750, 5748, 5767, 5740, 5745, 5761, 5748, 5742, 5734, 5759, 5753, 5732, 5713, 5701, 5788, 5787, 5786, 5794, 5795, 5793, 5800, 5801, 5799, 5801, 5797, 5799, 5803, 5806, 5804, 5807, 5810, 5808, 5816, 5813, 5815, 5812, 5813, 5811, 5788, 5804, 5789, 5807, 5792, 5791, 5818, 5800, 5817, 5786, 5794, 5793, 5813, 5806, 5805, 5811, 5805, 5803, 5792, 5799, 5791, 5791, 5797, 5790, 5789, 5790, 5787, 5797, 5796, 5794, 5804, 5809, 5807, 5787, 5797, 5794, 5789, 5807, 5791, 5808, 5800, 5792, 5825, 5826, 5824, 5826, 5822, 5824, 5821, 5824, 5820, 5820, 5822, 5819, 5795, 5784, 5793, 5786, 5784, 5783, 5810, 5819, 5808, 5822, 5818, 5817, 5819, 5817, 5808, 5682, 5672, 5671, 5692, 5680, 5690, 5683, 5673, 5672, 5686, 5680, 5673, 5784, 5726, 5725, 5783, 5725, 5718, 5756, 5788, 5786, 5812, 5776, 5778, 5813, 5782, 5780, 5811, 5764, 5776,
5805, 5780, 5766, 5764, 5788, 5758, 5756, 5783, 5718, 5831, 5832, 5830, 5829, 5830, 5828, 5839, 5840, 5838, 5843, 5844, 5842, 5835, 5838, 5834, 5836, 5843, 5842, 5842, 5841, 5839, 5835, 5842, 5839, 5838, 5833, 5831, 5834, 5831, 5829, 5854, 5853, 5852, 5859, 5860, 5858, 5857, 5859, 5856, 5866, 5865, 5864, 5870, 5869, 5868, 5850, 5865, 5851, 5868, 5855, 5854, 5860, 5872, 5858, 5851, 5858, 5850, 5856, 5851, 5849, 5849, 5854, 5852, 5862, 5857, 5856, 5867, 5868, 5865, 5849, 5862, 5856, 5865, 5854, 5851, 5850, 5872, 5864, 5876, 5875, 5874, 5880, 5879, 5878, 5884, 5883, 5882, 5888, 5887, 5886, 5882, 5877, 5876, 5886, 5881, 5880, 5877, 5878, 5875, 5885, 5886, 5883, 5883, 5880, 5877, 5892, 5891, 5890, 5896, 5895, 5894, 5900, 5899, 5898, 5904, 5903, 5902, 5898, 5893, 5892, 5902, 5897, 5896, 5893, 5894, 5891, 5901, 5902, 5899, 5899, 5896, 5893, 5908, 5907, 5906, 5912, 5911, 5910, 5916, 5915, 5914, 5920, 5919, 5918, 5914, 5909, 5908, 5918, 5913, 5912, 5909, 5910, 5907, 5917, 5918, 5915, 5915, 5912, 5909, 5890, 5867, 5866, 5894, 5871, 5870, 5891, 5870, 5867, 5906, 5885, 5884,
5910, 5889, 5888, 5907, 5888, 5885, 5855, 5874, 5853, 5871, 5882, 5869, 5869, 5876, 5855, 5897, 5906, 5895, 5905, 5914, 5903, 5903, 5908, 5897, 5895, 5884, 5871, 5924, 5923, 5922, 5930, 5931, 5929, 5936, 5937, 5935, 5937, 5933, 5935, 5941, 5940, 5939, 5945, 5944, 5943, 5950, 5947, 5949, 5939, 5925, 5924, 5943, 5928, 5927, 5948, 5936, 5947, 5923, 5929, 5922, 5927, 5936, 5935, 5927, 5933, 5926, 5946, 5947, 5944, 5925, 5926, 5923, 5933, 5932, 5930, 5942, 5943, 5940, 5923, 5933, 5930, 5925, 5943, 5927, 5944, 5936, 5928, 5956, 5955, 5954, 5960, 5959, 5958, 5964, 5963, 5962, 5968, 5967, 5966, 5962, 5957, 5956, 5966, 5961, 5960, 5957, 5958, 5955, 5965, 5966, 5963, 5963, 5960, 5957, 5977, 5974, 5976, 5983, 5980, 5982, 5981, 5976, 5980, 5973, 5974, 5972, 5979, 5980, 5978, 5980, 5973, 5978, 5954, 5942, 5941, 5945, 5959, 5946, 5955, 5945, 5942, 5975, 5952, 5974, 5974, 5951, 5972, 5952, 5950, 5949, 5946, 5952, 5949, 5961, 5972, 5959, 5967, 5979, 5978, 5983, 5970, 5971, 5982, 5969, 5970, 5961, 5978, 5973, 5959, 5951, 5946, 5828, 5853, 5829, 5832, 5862, 5830, 5830, 5852, 5828,
5834, 5875, 5835, 5878, 5837, 5836, 5835, 5878, 5836, 5829, 5874, 5834, 5931, 5847, 5929, 5922, 5847, 5846, 5847, 5845, 5843, 5846, 5843, 5837, 5879, 5924, 5922, 5889, 5939, 5887, 5887, 5924, 5881, 5913, 5954, 5911, 5921, 5962, 5919, 5919, 5956, 5913, 5911, 5941, 5889, 5879, 5846, 5837, 5986, 5989, 5987, 5993, 5990, 5992, 5995, 5996, 5994, 5996, 5998, 5994, 6001, 5995, 5999, 6004, 6003, 6002, 6007, 6008, 6006, 5988, 6003, 5989, 6000, 6006, 5998, 5990, 5985, 5984, 5986, 5995, 5994, 5994, 5988, 5986, 5999, 5987, 5989, 6002, 6008, 6004, 5996, 5991, 5993, 5991, 5986, 5985, 5998, 6002, 5988, 6010, 6013, 6011, 6016, 6015, 6014, 6019, 6020, 6018, 6024, 6023, 6022, 6028, 6027, 6026, 6022, 6013, 6012, 6026, 6017, 6016, 6018, 6012, 6010, 6013, 6014, 6011, 6025, 6026, 6023, 6023, 6016, 6013, 6036, 6033, 6035, 6036, 6037, 6034, 6033, 6031, 6030, 6031, 6037, 6032, 6039, 6042, 6040, 6045, 6044, 6043, 6058, 6055, 6057, 6062, 6059, 6061, 6051, 6042, 6041, 6045, 6054, 6046, 6048, 6055, 6047, 6062, 6049, 6060, 6056, 6051, 6055, 6059, 6054, 6053, 6042, 6043, 6040, 6061, 6056, 6058,
6059, 6052, 6056, 6052, 6045, 6042, 6041, 6055, 6051, 6060, 6046, 6054, 6030, 6005, 6004, 6009, 6033, 6008, 6008, 6030, 6004, 6039, 6025, 6024, 6043, 6029, 6028, 6040, 6028, 6025, 6019, 5995, 5997, 6020, 6001, 5999, 6005, 6022, 6003, 6018, 5987, 5995, 6012, 5999, 5989, 5989, 6022, 6012, 6032, 6039, 6031, 6048, 6037, 6038, 6047, 6032, 6037, 6005, 6039, 6024, 6065, 6064, 6063, 6069, 6068, 6067, 6072, 6073, 6071, 6076, 6072, 6075, 6078, 6079, 6077, 6081, 6084, 6082, 6085, 6088, 6086, 6092, 6089, 6091, 6094, 6095, 6093, 6081, 6066, 6065, 6085, 6070, 6069, 6090, 6075, 6089, 6080, 6093, 6079, 6064, 6071, 6063, 6066, 6072, 6064, 6091, 6082, 6084, 6077, 6069, 6067, 6093, 6087, 6085, 6078, 6072, 6074, 6072, 6067, 6064, 6082, 6075, 6066, 6069, 6093, 6085, 6097, 6100, 6098, 6109, 6110, 6108, 6110, 6106, 6108, 6114, 6113, 6112, 6116, 6119, 6117, 6112, 6100, 6099, 6116, 6103, 6102, 6121, 6109, 6120, 6102, 6109, 6108, 6108, 6101, 6102, 6100, 6101, 6098, 6106, 6105, 6104, 6115, 6116, 6113, 6098, 6106, 6104, 6100, 6116, 6102, 6120, 6103, 6117, 6124, 6123, 6122, 6128, 6127, 6126,
6135, 6132, 6134, 6141, 6140, 6139, 6146, 6147, 6145, 6128, 6140, 6129, 6135, 6143, 6133, 6133, 6124, 6132, 6139, 6147, 6141, 6125, 6126, 6123, 6136, 6131, 6130, 6146, 6143, 6144, 6126, 6130, 6123, 6143, 6139, 6138, 6138, 6128, 6125, 6125, 6143, 6138, 6151, 6150, 6149, 6155, 6154, 6153, 6159, 6158, 6157, 6163, 6162, 6161, 6157, 6152, 6151, 6161, 6156, 6155, 6152, 6153, 6150, 6160, 6161, 6158, 6158, 6155, 6152, 6122, 6084, 6083, 6087, 6127, 6088, 6131, 6091, 6130, 6096, 6136, 6095, 6130, 6084, 6123, 6087, 6136, 6126, 6149, 6115, 6114, 6153, 6119, 6118, 6150, 6118, 6115, 6070, 6097, 6068, 6088, 6112, 6086, 6086, 6099, 6070, 6129, 6149, 6127, 6142, 6157, 6140, 6140, 6151, 6129, 6127, 6114, 6088, 6171, 6168, 6170, 6171, 6172, 6169, 6168, 6166, 6165, 6169, 6167, 6166, 6174, 6177, 6175, 6180, 6179, 6178, 6189, 6188, 6187, 6194, 6191, 6193, 6194, 6195, 6192, 6184, 6177, 6176, 6187, 6181, 6180, 6183, 6191, 6182, 6191, 6185, 6184, 6185, 6195, 6186, 6177, 6178, 6175, 6186, 6187, 6185, 6177, 6187, 6180, 6182, 6184, 6176, 6200, 6199, 6198, 6209, 6206, 6208, 6205, 6206, 6204,
6213, 6214, 6212, 6209, 6212, 6207, 6206, 6201, 6200, 6204, 6200, 6198, 6210, 6214, 6211, 6205, 6202, 6203, 6204, 6197, 6202, 6201, 6212, 6210, 6189, 6199, 6190, 6196, 6202, 6195, 6197, 6189, 6186, 6186, 6202, 6197, 6167, 6174, 6166, 6183, 6172, 6173, 6172, 6176, 6167, 6165, 6142, 6141, 6148, 6168, 6147, 6141, 6168, 6165, 6174, 6160, 6159, 6178, 6164, 6163, 6175, 6163, 6160, 6142, 6174, 6159, 6015, 6065, 6063, 6029, 6081, 6027, 6027, 6065, 6017, 6044, 6124, 6122, 6134, 6049, 6050, 6049, 6124, 6046, 6044, 6083, 6029, 5984, 5738, 5737, 5744, 5990, 5743, 5996, 5747, 5994, 5750, 5995, 5748, 5985, 5741, 5738, 5737, 5990, 5984, 5986, 5747, 5741, 5742, 5995, 5987, 5761, 6011, 5762, 6014, 5766, 5765, 5762, 6014, 5765, 5742, 6010, 5761, 5809, 6068, 5810, 6073, 5813, 6071, 5816, 6072, 5814, 5806, 6067, 5809, 6063, 5813, 5805, 5814, 6064, 5806, 6097, 5821, 5820, 6105, 5825, 6104, 6104, 5821, 6098, 5810, 6097, 5820, 6015, 5805, 5766, 5541, 5670, 5539, 5674, 5545, 5543, 5539, 5674, 5543, 5619, 5682, 5617, 5627, 5694, 5625, 5625, 5684, 5619, 5651, 5727, 5649, 5659, 5735, 5657,
5744, 5667, 5669, 5667, 5737, 5659, 5657, 5729, 5651, 5649, 5696, 5627, 5541, 5682, 5671, 6222, 6223, 6221, 6223, 6219, 6221, 6217, 6222, 6221, 6221, 6216, 6217, 6227, 6226, 6225, 6233, 6230, 6232, 6238, 6237, 6236, 6242, 6241, 6240, 6245, 6246, 6244, 6249, 6245, 6248, 6227, 6237, 6228, 6246, 6234, 6244, 6228, 6230, 6226, 6240, 6245, 6244, 6243, 6245, 6241, 6234, 6233, 6232, 6237, 6242, 6240, 6228, 6234, 6232, 6237, 6229, 6228, 6244, 6229, 6240, 6225, 6218, 6217, 6231, 6222, 6230, 6226, 6222, 6218, 6253, 6254, 6252, 6251, 6252, 6250, 6258, 6257, 6256, 6260, 6263, 6261, 6267, 6264, 6266, 6270, 6269, 6268, 6279, 6276, 6278, 6281, 6282, 6280, 6282, 6276, 6280, 6279, 6281, 6277, 6268, 6259, 6258, 6272, 6263, 6262, 6279, 6266, 6277, 6266, 6261, 6263, 6277, 6272, 6276, 6275, 6280, 6274, 6280, 6272, 6274, 6277, 6275, 6273, 6259, 6260, 6257, 6271, 6272, 6269, 6269, 6262, 6259, 6277, 6263, 6273, 6287, 6290, 6288, 6294, 6295, 6293, 6292, 6294, 6291, 6298, 6299, 6297, 6304, 6301, 6303, 6307, 6306, 6305, 6311, 6310, 6309, 6320, 6317, 6319, 6320, 6314, 6318, 6322, 6323, 6321,
6323, 6325, 6321, 6285, 6306, 6286, 6295, 6313, 6293, 6323, 6301, 6321, 6304, 6322, 6302, 6285, 6294, 6293, 6291, 6286, 6284, 6317, 6308, 6307, 6318, 6306, 6308, 6288, 6297, 6287, 6302, 6289, 6301, 6310, 6321, 6309, 6321, 6311, 6309, 6286, 6287, 6284, 6297, 6292, 6291, 6323, 6314, 6316, 6325, 6320, 6318, 6287, 6291, 6284, 6321, 6306, 6314, 6311, 6318, 6308, 6306, 6289, 6286, 6285, 6313, 6305, 6321, 6289, 6309, 6290, 6322, 6310, 6329, 6328, 6327, 6331, 6334, 6332, 6335, 6338, 6336, 6341, 6340, 6339, 6335, 6330, 6329, 6333, 6340, 6334, 6328, 6333, 6331, 6338, 6339, 6336, 6330, 6339, 6333, 6345, 6348, 6346, 6351, 6350, 6349, 6354, 6355, 6353, 6359, 6358, 6357, 6363, 6362, 6361, 6357, 6348, 6347, 6361, 6352, 6351, 6346, 6353, 6345, 6348, 6349, 6346, 6360, 6361, 6358, 6358, 6351, 6348, 6327, 6271, 6270, 6274, 6332, 6275, 6328, 6274, 6271, 6349, 6312, 6311, 6355, 6317, 6353, 6320, 6354, 6318, 6346, 6311, 6308, 6353, 6307, 6345, 6308, 6354, 6346, 6273, 6307, 6305, 6313, 6279, 6277, 6305, 6277, 6273, 6332, 6347, 6345, 6342, 6357, 6340, 6334, 6357, 6347, 6332, 6307, 6275,
6365, 6368, 6366, 6371, 6370, 6369, 6378, 6379, 6377, 6383, 6382, 6381, 6387, 6386, 6385, 6381, 6368, 6367, 6385, 6372, 6371, 6369, 6378, 6377, 6368, 6369, 6366, 6377, 6376, 6374, 6384, 6385, 6382, 6369, 6374, 6366, 6382, 6371, 6368, 6395, 6396, 6394, 6396, 6392, 6394, 6398, 6401, 6399, 6409, 6406, 6408, 6398, 6391, 6390, 6405, 6395, 6404, 6390, 6395, 6394, 6390, 6392, 6389, 6403, 6406, 6402, 6401, 6402, 6399, 6406, 6405, 6404, 6399, 6406, 6404, 6404, 6391, 6399, 6412, 6411, 6410, 6416, 6415, 6414, 6420, 6419, 6418, 6424, 6423, 6422, 6418, 6413, 6412, 6422, 6417, 6416, 6413, 6414, 6411, 6421, 6422, 6419, 6419, 6416, 6413, 6428, 6427, 6426, 6436, 6437, 6435, 6437, 6433, 6435, 6441, 6440, 6439, 6443, 6446, 6444, 6452, 6449, 6451, 6452, 6447, 6450, 6439, 6429, 6428, 6443, 6432, 6431, 6448, 6436, 6447, 6431, 6436, 6435, 6435, 6430, 6431, 6450, 6445, 6449, 6446, 6447, 6444, 6429, 6430, 6427, 6442, 6443, 6440, 6429, 6443, 6431, 6447, 6432, 6444, 6410, 6384, 6383, 6414, 6388, 6387, 6411, 6387, 6384, 6426, 6401, 6400, 6434, 6408, 6433, 6427, 6403, 6401, 6430, 6408, 6403,
6372, 6389, 6370, 6392, 6380, 6378, 6388, 6398, 6386, 6370, 6392, 6378, 6372, 6398, 6390, 6417, 6426, 6415, 6425, 6439, 6423, 6423, 6428, 6417, 6415, 6400, 6388, 6256, 6239, 6238, 6260, 6243, 6242, 6265, 6248, 6264, 6257, 6242, 6239, 6261, 6248, 6243, 6250, 6366, 6251, 6254, 6373, 6252, 6376, 6253, 6374, 6252, 6365, 6250, 6366, 6253, 6251, 6288, 6367, 6365, 6373, 6300, 6298, 6312, 6381, 6310, 6288, 6373, 6298, 6310, 6367, 6290, 6352, 6410, 6350, 6364, 6418, 6362, 6362, 6412, 6352, 6350, 6383, 6312, 6119, 6216, 6117, 6219, 6121, 6120, 6117, 6219, 6120, 6156, 6225, 6154, 6164, 6236, 6162, 6162, 6227, 6156, 6119, 6225, 6217, 6181, 6256, 6179, 6190, 6268, 6188, 6188, 6258, 6181, 6199, 6329, 6327, 6211, 6335, 6210, 6344, 6214, 6215, 6214, 6337, 6211, 6210, 6329, 6201, 6199, 6270, 6190, 6179, 6238, 6164, 6455, 6454, 6453, 6459, 6458, 6457, 6462, 6463, 6461, 6469, 6468, 6467, 6474, 6471, 6473, 6480, 6477, 6479, 6476, 6477, 6475, 6465, 6456, 6455, 6459, 6468, 6460, 6464, 6471, 6463, 6453, 6463, 6455, 6471, 6466, 6465, 6477, 6470, 6469, 6467, 6477, 6469, 6456, 6457, 6454,
6476, 6472, 6474, 6472, 6467, 6466, 6466, 6459, 6456, 6455, 6471, 6465, 6483, 6482, 6481, 6487, 6486, 6485, 6489, 6492, 6490, 6495, 6494, 6493, 6489, 6484, 6483, 6493, 6488, 6487, 6484, 6485, 6482, 6492, 6493, 6490, 6490, 6487, 6484, 6511, 6508, 6510, 6515, 6512, 6514, 6509, 6504, 6508, 6512, 6507, 6506, 6514, 6509, 6511, 6512, 6505, 6509, 6504, 6492, 6491, 6495, 6507, 6496, 6498, 6508, 6497, 6515, 6499, 6513, 6505, 6495, 6492, 6491, 6508, 6504, 6513, 6496, 6507, 6460, 6481, 6458, 6470, 6489, 6468, 6498, 6478, 6480, 6497, 6470, 6478, 6468, 6483, 6460, 6518, 6517, 6516, 6520, 6523, 6521, 6529, 6526, 6528, 6529, 6524, 6527, 6530, 6533, 6531, 6540, 6537, 6539, 6540, 6535, 6538, 6530, 6519, 6518, 6542, 6526, 6541, 6538, 6532, 6537, 6538, 6531, 6533, 6527, 6522, 6526, 6527, 6521, 6523, 6519, 6520, 6517, 6542, 6535, 6536, 6535, 6534, 6531, 6519, 6534, 6522, 6534, 6526, 6522, 6549, 6550, 6548, 6550, 6546, 6548, 6552, 6555, 6553, 6563, 6560, 6562, 6552, 6545, 6544, 6559, 6549, 6558, 6544, 6549, 6548, 6544, 6546, 6543, 6557, 6560, 6556, 6555, 6556, 6553, 6560, 6559, 6558,
6553, 6560, 6558, 6558, 6545, 6553, 6566, 6565, 6564, 6570, 6569, 6568, 6573, 6574, 6572, 6577, 6573, 6576, 6580, 6579, 6578, 6582, 6585, 6583, 6578, 6567, 6566, 6582, 6571, 6570, 6587, 6576, 6586, 6568, 6573, 6572, 6571, 6573, 6569, 6565, 6570, 6568, 6581, 6582, 6579, 6579, 6570, 6567, 6583, 6576, 6571, 6594, 6595, 6593, 6595, 6591, 6593, 6590, 6593, 6589, 6593, 6588, 6589, 6564, 6555, 6554, 6574, 6562, 6572, 6565, 6557, 6555, 6568, 6562, 6557, 6585, 6588, 6583, 6591, 6587, 6586, 6583, 6591, 6586, 6547, 6502, 6546, 6543, 6502, 6501, 6488, 6516, 6486, 6494, 6532, 6530, 6539, 6499, 6500, 6537, 6496, 6499, 6494, 6518, 6488, 6603, 6604, 6602, 6606, 6609, 6607, 6610, 6613, 6611, 6615, 6616, 6614, 6616, 6618, 6614, 6610, 6599, 6598, 6601, 6615, 6600, 6621, 6603, 6620, 6606, 6615, 6614, 6606, 6618, 6608, 6599, 6602, 6598, 6602, 6601, 6600, 6609, 6610, 6607, 6597, 6602, 6600, 6597, 6610, 6598, 6600, 6607, 6597, 6611, 6603, 6599, 6627, 6624, 6626, 6626, 6622, 6623, 6630, 6629, 6628, 6632, 6635, 6633, 6641, 6638, 6640, 6637, 6638, 6636, 6645, 6642, 6644, 6638, 6631, 6630,
6636, 6630, 6628, 6643, 6634, 6642, 6631, 6632, 6629, 6644, 6639, 6641, 6639, 6634, 6631, 6651, 6648, 6650, 6647, 6648, 6646, 6628, 6609, 6608, 6632, 6613, 6612, 6619, 6636, 6618, 6629, 6612, 6609, 6608, 6636, 6628, 6649, 6626, 6648, 6648, 6623, 6646, 6613, 6622, 6611, 6624, 6621, 6620, 6611, 6624, 6620, 6633, 6647, 6646, 6651, 6643, 6645, 6650, 6635, 6643, 6633, 6623, 6613, 6654, 6653, 6652, 6658, 6657, 6656, 6662, 6661, 6660, 6666, 6665, 6664, 6660, 6655, 6654, 6664, 6659, 6658, 6655, 6656, 6653, 6663, 6664, 6661, 6661, 6658, 6655, 6670, 6669, 6668, 6676, 6673, 6675, 6681, 6680, 6679, 6689, 6690, 6688, 6690, 6686, 6688, 6670, 6680, 6671, 6687, 6677, 6686, 6671, 6673, 6669, 6684, 6689, 6688, 6684, 6686, 6683, 6677, 6676, 6675, 6682, 6683, 6680, 6671, 6677, 6675, 6680, 6672, 6671, 6683, 6677, 6672, 6694, 6693, 6692, 6698, 6697, 6696, 6701, 6702, 6700, 6707, 6710, 6708, 6714, 6711, 6713, 6714, 6715, 6712, 6704, 6695, 6694, 6707, 6699, 6698, 6703, 6711, 6702, 6692, 6702, 6694, 6711, 6705, 6704, 6712, 6706, 6705, 6695, 6696, 6693, 6706, 6707, 6705, 6718, 6715, 6716,
6715, 6709, 6706, 6695, 6707, 6698, 6694, 6711, 6704, 6721, 6720, 6719, 6725, 6724, 6723, 6730, 6727, 6729, 6733, 6732, 6731, 6737, 6736, 6735, 6731, 6722, 6721, 6725, 6736, 6726, 6726, 6727, 6724, 6722, 6723, 6720, 6734, 6735, 6732, 6732, 6725, 6722, 6662, 6693, 6663, 6696, 6667, 6666, 6693, 6666, 6663, 6719, 6682, 6681, 6723, 6685, 6684, 6728, 6689, 6727, 6682, 6723, 6684, 6724, 6689, 6685, 6659, 6668, 6657, 6667, 6679, 6665, 6665, 6670, 6659, 6699, 6719, 6697, 6710, 6731, 6708, 6708, 6721, 6699, 6697, 6681, 6667, 6748, 6749, 6747, 6749, 6745, 6747, 6746, 6740, 6745, 6743, 6748, 6747, 6743, 6745, 6742, 6742, 6740, 6739, 6753, 6752, 6751, 6760, 6757, 6759, 6756, 6757, 6755, 6760, 6765, 6758, 6757, 6754, 6753, 6751, 6757, 6753, 6758, 6762, 6754, 6769, 6768, 6767, 6773, 6772, 6771, 6777, 6776, 6775, 6781, 6780, 6779, 6784, 6785, 6783, 6769, 6776, 6770, 6779, 6774, 6773, 6775, 6785, 6777, 6770, 6771, 6768, 6778, 6779, 6776, 6776, 6773, 6770, 6789, 6788, 6787, 6792, 6793, 6791, 6796, 6792, 6795, 6797, 6800, 6798, 6808, 6805, 6807, 6797, 6790, 6789, 6804, 6795, 6803,
6787, 6792, 6791, 6790, 6792, 6788, 6802, 6805, 6801, 6800, 6801, 6798, 6805, 6804, 6803, 6798, 6805, 6803, 6798, 6795, 6790, 6814, 6813, 6812, 6819, 6816, 6818, 6819, 6820, 6817, 6825, 6824, 6823, 6834, 6831, 6833, 6830, 6831, 6829, 6814, 6824, 6815, 6821, 6827, 6820, 6816, 6810, 6809, 6810, 6820, 6811, 6831, 6826, 6825, 6829, 6825, 6823, 6811, 6812, 6810, 6830, 6827, 6828, 6827, 6823, 6822, 6822, 6814, 6811, 6811, 6827, 6822, 6793, 6763, 6791, 6791, 6761, 6787, 6809, 6778, 6777, 6812, 6782, 6781, 6786, 6816, 6785, 6778, 6812, 6781, 6777, 6816, 6809, 6752, 6769, 6767, 6784, 6765, 6766, 6765, 6775, 6762, 6762, 6769, 6754, 6837, 6836, 6835, 6843, 6840, 6842, 6848, 6847, 6846, 6852, 6851, 6850, 6855, 6856, 6854, 6859, 6855, 6858, 6837, 6847, 6838, 6856, 6844, 6854, 6838, 6840, 6836, 6851, 6854, 6850, 6853, 6855, 6851, 6844, 6843, 6842, 6847, 6852, 6850, 6838, 6844, 6842, 6847, 6839, 6838, 6854, 6839, 6850, 6862, 6861, 6860, 6866, 6865, 6864, 6871, 6868, 6870, 6874, 6873, 6872, 6878, 6877, 6876, 6872, 6863, 6862, 6866, 6877, 6867, 6870, 6865, 6867, 6863, 6864, 6861,
6875, 6876, 6873, 6873, 6866, 6863, 6889, 6890, 6888, 6890, 6886, 6888, 6887, 6881, 6886, 6884, 6889, 6888, 6884, 6886, 6883, 6883, 6881, 6880, 6860, 6849, 6848, 6864, 6853, 6852, 6869, 6858, 6868, 6861, 6852, 6849, 6865, 6858, 6853, 6881, 6871, 6870, 6879, 6883, 6877, 6867, 6881, 6870, 6877, 6880, 6867, 6751, 6710, 6709, 6718, 6755, 6717, 6709, 6755, 6751, 6767, 6734, 6733, 6771, 6738, 6737, 6768, 6737, 6734, 6752, 6733, 6710, 6835, 6744, 6743, 6841, 6748, 6840, 6836, 6748, 6744, 6740, 6730, 6729, 6738, 6742, 6736, 6726, 6740, 6729, 6736, 6739, 6726, 6774, 6835, 6772, 6782, 6846, 6780, 6780, 6837, 6774, 6815, 6860, 6813, 6824, 6874, 6872, 6824, 6862, 6815, 6813, 6848, 6782, 6738, 6835, 6743, 6893, 6896, 6894, 6900, 6901, 6899, 6901, 6903, 6899, 6894, 6899, 6893, 6899, 6895, 6893, 6903, 6898, 6897, 6895, 6897, 6892, 6909, 6908, 6907, 6912, 6913, 6911, 6919, 6918, 6917, 6922, 6923, 6921, 6917, 6910, 6909, 6914, 6921, 6913, 6911, 6906, 6905, 6915, 6923, 6916, 6905, 6909, 6907, 6916, 6917, 6915, 6906, 6917, 6909, 6913, 6915, 6906, 6907, 6896, 6895, 6898, 6911, 6897,
6892, 6907, 6895, 6897, 6905, 6892, 6928, 6931, 6929, 6936, 6937, 6935, 6937, 6939, 6935, 6944, 6945, 6943, 6945, 6941, 6943, 6928, 6936, 6935, 6935, 6930, 6928, 6934, 6943, 6933, 6933, 6941, 6932, 6931, 6932, 6929, 6941, 6938, 6936, 6932, 6936, 6929, 6948, 6951, 6949, 6955, 6956, 6954, 6956, 6958, 6954, 6964, 6963, 6962, 6967, 6968, 6966, 6962, 6951, 6950, 6953, 6966, 6952, 6966, 6961, 6960, 6949, 6954, 6948, 6948, 6958, 6950, 6958, 6953, 6952, 6960, 6964, 6962, 6950, 6952, 6947, 6947, 6962, 6950, 6947, 6966, 6960, 6972, 6971, 6970, 6976, 6975, 6974, 6980, 6979, 6978, 6984, 6983, 6982, 6978, 6973, 6972, 6982, 6977, 6976, 6973, 6974, 6971, 6981, 6982, 6979, 6979, 6976, 6973, 6927, 6955, 6926, 6926, 6949, 6925, 6970, 6931, 6930, 6933, 6975, 6934, 6931, 6974, 6933, 6939, 6927, 6926, 6925, 6939, 6926, 6951, 6970, 6949, 6965, 6978, 6963, 6963, 6972, 6951, 6925, 6970, 6930, 6992, 6991, 6990, 7003, 7000, 7002, 6998, 7000, 6996, 7003, 6997, 7001, 7003, 7005, 7001, 7000, 6993, 6992, 6996, 6992, 6990, 6993, 6997, 6991, 6998, 6994, 6995, 6996, 6989, 6994, 6993, 7005, 7004,
7009, 7008, 7007, 7013, 7014, 7012, 7017, 7013, 7016, 7022, 7021, 7020, 7029, 7026, 7028, 7009, 7021, 7010, 7027, 7018, 7026, 7007, 7013, 7012, 7016, 7008, 7010, 7025, 7026, 7024, 7018, 7017, 7016, 7021, 7025, 7024, 7010, 7018, 7016, 7021, 7011, 7010, 7026, 7011, 7024, 7014, 6987, 7012, 7007, 6987, 6986, 6919, 6991, 6920, 6924, 6994, 6923, 6916, 6990, 6919, 6923, 6989, 6916, 6987, 6946, 6944, 6934, 6987, 6944, 6975, 7009, 7007, 6985, 7020, 6983, 6983, 7009, 6977, 6975, 6986, 6934, 6652, 6581, 6580, 6656, 6585, 6584, 6653, 6584, 6581, 6668, 6590, 6589, 6674, 6594, 6673, 6669, 6594, 6590, 6585, 6668, 6589, 7040, 7037, 7039, 7040, 7041, 7038, 7046, 7043, 7045, 7048, 7049, 7047, 7049, 7043, 7047, 7046, 7048, 7044, 7052, 7055, 7053, 7059, 7060, 7058, 7035, 7053, 7036, 7042, 7056, 7041, 7037, 7031, 7030, 7031, 7041, 7032, 7043, 7034, 7033, 7035, 7048, 7047, 7035, 7043, 7033, 7034, 7048, 7036, 7058, 7054, 7052, 7032, 7033, 7031, 7059, 7056, 7057, 7056, 7052, 7051, 7032, 7052, 7035, 7032, 7056, 7051, 7064, 7063, 7062, 7068, 7067, 7066, 7072, 7071, 7070, 7074, 7077, 7075,
7064, 7071, 7065, 7074, 7069, 7068, 7065, 7066, 7063, 7073, 7074, 7071, 7081, 7078, 7079, 7078, 7076, 7073, 7071, 7068, 7065, 7084, 7083, 7082, 7091, 7088, 7090, 7087, 7088, 7086, 7095, 7096, 7094, 7091, 7094, 7089, 7088, 7085, 7084, 7086, 7084, 7082, 7092, 7096, 7093, 7089, 7092, 7085, 7098, 7101, 7099, 7102, 7105, 7103, 7112, 7111, 7110, 7116, 7115, 7114, 7110, 7101, 7100, 7114, 7105, 7104, 7099, 7104, 7102, 7108, 7107, 7106, 7113, 7114, 7111, 7099, 7108, 7106, 7111, 7104, 7101, 7082, 7055, 7054, 7061, 7086, 7060, 7054, 7086, 7082, 7098, 7073, 7072, 7102, 7077, 7076, 7107, 7078, 7106, 7081, 7108, 7080, 7099, 7078, 7073, 7076, 7108, 7102, 7034, 7064, 7062, 7055, 7070, 7053, 7053, 7064, 7036, 7083, 7100, 7098, 7093, 7110, 7092, 7092, 7100, 7085, 7055, 7098, 7072, 7120, 7119, 7118, 7124, 7123, 7122, 7128, 7127, 7126, 7132, 7131, 7130, 7126, 7121, 7120, 7130, 7125, 7124, 7121, 7122, 7119, 7129, 7130, 7127, 7127, 7124, 7121, 7138, 7137, 7136, 7147, 7144, 7146, 7154, 7151, 7153, 7156, 7151, 7155, 7138, 7149, 7139, 7154, 7142, 7152, 7151, 7149, 7148, 7155, 7148, 7150,
7146, 7140, 7141, 7137, 7141, 7140, 7147, 7142, 7143, 7146, 7139, 7142, 7149, 7142, 7139, 7159, 7158, 7157, 7166, 7163, 7165, 7170, 7167, 7169, 7171, 7174, 7172, 7182, 7179, 7181, 7171, 7160, 7159, 7178, 7165, 7177, 7160, 7163, 7158, 7168, 7161, 7167, 7176, 7179, 7175, 7169, 7163, 7164, 7174, 7175, 7172, 7179, 7178, 7177, 7167, 7158, 7163, 7172, 7179, 7177, 7172, 7165, 7160, 7157, 7129, 7128, 7132, 7162, 7133, 7170, 7134, 7168, 7129, 7161, 7132, 7162, 7134, 7133, 7125, 7136, 7123, 7131, 7150, 7148, 7156, 7134, 7135, 7134, 7150, 7133, 7131, 7138, 7125, 7188, 7187, 7186, 7192, 7191, 7190, 7195, 7196, 7194, 7200, 7199, 7198, 7204, 7203, 7202, 7207, 7208, 7206, 7198, 7189, 7188, 7202, 7193, 7192, 7197, 7206, 7196, 7194, 7188, 7186, 7198, 7208, 7200, 7189, 7190, 7187, 7201, 7202, 7199, 7199, 7192, 7189, 7188, 7206, 7198, 7212, 7211, 7210, 7216, 7215, 7214, 7223, 7220, 7222, 7219, 7220, 7218, 7228, 7227, 7226, 7231, 7232, 7230, 7226, 7217, 7216, 7223, 7230, 7221, 7220, 7213, 7212, 7218, 7212, 7210, 7224, 7232, 7225, 7211, 7216, 7214, 7225, 7226, 7224, 7224, 7216, 7213,
7221, 7224, 7213, 7210, 7201, 7200, 7214, 7205, 7204, 7209, 7218, 7208, 7211, 7204, 7201, 7200, 7218, 7210, 7195, 7184, 7185, 7184, 7186, 7183, 7236, 7235, 7234, 7243, 7240, 7242, 7246, 7245, 7244, 7253, 7250, 7252, 7244, 7237, 7236, 7251, 7242, 7250, 7239, 7240, 7238, 7249, 7250, 7248, 7237, 7238, 7235, 7247, 7248, 7245, 7245, 7239, 7237, 7248, 7242, 7239, 7256, 7255, 7254, 7264, 7265, 7263, 7265, 7261, 7263, 7269, 7268, 7267, 7271, 7274, 7272, 7278, 7275, 7277, 7267, 7257, 7256, 7271, 7260, 7259, 7276, 7264, 7275, 7259, 7264, 7263, 7263, 7258, 7259, 7274, 7275, 7272, 7257, 7258, 7255, 7270, 7271, 7268, 7257, 7271, 7259, 7275, 7260, 7272, 7284, 7285, 7283, 7288, 7289, 7287, 7291, 7294, 7292, 7297, 7296, 7295, 7300, 7301, 7299, 7279, 7292, 7280, 7295, 7282, 7281, 7285, 7299, 7283, 7280, 7283, 7279, 7299, 7293, 7291, 7282, 7287, 7281, 7287, 7286, 7284, 7294, 7295, 7292, 7281, 7284, 7280, 7280, 7295, 7281, 7279, 7299, 7291, 7254, 7247, 7246, 7262, 7252, 7261, 7255, 7249, 7247, 7258, 7252, 7249, 7097, 7184, 7096, 7096, 7183, 7093, 7112, 7187, 7113, 7190, 7117, 7116,
7187, 7116, 7113, 7183, 7112, 7093, 7234, 7174, 7173, 7241, 7181, 7240, 7235, 7176, 7174, 7238, 7181, 7176, 7069, 7118, 7067, 7077, 7126, 7075, 7075, 7120, 7069, 7105, 7157, 7103, 7117, 7171, 7115, 7115, 7159, 7105, 7103, 7128, 7077, 7193, 7234, 7191, 7205, 7244, 7203, 7203, 7236, 7193, 7217, 7254, 7215, 7229, 7267, 7227, 7227, 7256, 7217, 7215, 7246, 7205, 7191, 7173, 7117, 7311, 7308, 7310, 7309, 7304, 7308, 7307, 7308, 7306, 7306, 7304, 7303, 7316, 7317, 7315, 7323, 7322, 7321, 7327, 7326, 7325, 7332, 7333, 7331, 7312, 7322, 7313, 7317, 7329, 7315, 7333, 7319, 7331, 7312, 7316, 7315, 7326, 7331, 7325, 7319, 7318, 7316, 7322, 7327, 7325, 7314, 7316, 7313, 7322, 7314, 7313, 7315, 7321, 7312, 7325, 7319, 7314, 7339, 7338, 7337, 7342, 7343, 7341, 7349, 7352, 7350, 7358, 7359, 7357, 7359, 7353, 7357, 7356, 7358, 7354, 7349, 7340, 7339, 7344, 7354, 7343, 7335, 7343, 7336, 7347, 7358, 7357, 7347, 7353, 7345, 7346, 7358, 7348, 7336, 7337, 7335, 7346, 7351, 7349, 7362, 7358, 7360, 7358, 7351, 7348, 7346, 7339, 7336, 7336, 7354, 7346, 7365, 7364, 7363, 7369, 7368, 7367,
7373, 7372, 7371, 7377, 7376, 7375, 7371, 7366, 7365, 7375, 7370, 7369, 7366, 7367, 7364, 7374, 7375, 7372, 7372, 7369, 7366, 7381, 7380, 7379, 7383, 7386, 7384, 7390, 7387, 7389, 7392, 7393, 7391, 7397, 7396, 7395, 7401, 7400, 7399, 7406, 7403, 7405, 7408, 7409, 7407, 7395, 7382, 7381, 7385, 7400, 7386, 7389, 7380, 7382, 7398, 7403, 7396, 7383, 7393, 7385, 7399, 7409, 7401, 7393, 7390, 7389, 7408, 7403, 7404, 7382, 7393, 7389, 7403, 7399, 7396, 7396, 7385, 7382, 7363, 7324, 7323, 7367, 7328, 7327, 7364, 7327, 7324, 7347, 7380, 7348, 7383, 7352, 7351, 7388, 7358, 7387, 7362, 7391, 7361, 7380, 7358, 7348, 7361, 7383, 7351, 7328, 7345, 7326, 7353, 7334, 7332, 7326, 7353, 7332, 7370, 7379, 7368, 7378, 7395, 7376, 7376, 7381, 7370, 7328, 7379, 7347, 7413, 7412, 7411, 7415, 7418, 7416, 7422, 7419, 7421, 7425, 7424, 7423, 7427, 7430, 7428, 7434, 7431, 7433, 7423, 7414, 7413, 7427, 7418, 7417, 7432, 7421, 7431, 7418, 7419, 7416, 7433, 7428, 7430, 7414, 7415, 7412, 7426, 7427, 7424, 7424, 7417, 7414, 7431, 7418, 7428, 7438, 7441, 7439, 7445, 7442, 7444, 7445, 7446, 7443,
7453, 7450, 7452, 7453, 7448, 7451, 7442, 7436, 7435, 7436, 7446, 7437, 7450, 7441, 7440, 7451, 7439, 7441, 7437, 7438, 7436, 7452, 7446, 7447, 7450, 7437, 7446, 7456, 7455, 7454, 7464, 7461, 7463, 7466, 7461, 7465, 7469, 7468, 7467, 7476, 7473, 7475, 7467, 7457, 7456, 7474, 7465, 7473, 7462, 7458, 7461, 7460, 7461, 7458, 7475, 7471, 7472, 7455, 7460, 7458, 7468, 7472, 7471, 7457, 7471, 7460, 7473, 7460, 7471, 7454, 7426, 7425, 7429, 7459, 7430, 7464, 7433, 7462, 7455, 7429, 7426, 7459, 7433, 7430, 7477, 7480, 7478, 7483, 7482, 7481, 7486, 7487, 7485, 7498, 7495, 7497, 7502, 7499, 7501, 7491, 7480, 7479, 7483, 7494, 7484, 7502, 7489, 7500, 7485, 7479, 7477, 7495, 7492, 7491, 7499, 7494, 7493, 7478, 7483, 7481, 7501, 7496, 7498, 7499, 7492, 7496, 7492, 7483, 7480, 7500, 7484, 7494, 7503, 7506, 7504, 7509, 7508, 7507, 7516, 7513, 7515, 7519, 7522, 7520, 7525, 7524, 7523, 7528, 7529, 7527, 7529, 7531, 7527, 7523, 7510, 7509, 7516, 7528, 7514, 7514, 7505, 7513, 7520, 7527, 7519, 7527, 7521, 7519, 7504, 7509, 7507, 7517, 7512, 7511, 7522, 7523, 7520, 7507, 7511, 7504,
7520, 7509, 7506, 7514, 7520, 7506, 7536, 7537, 7535, 7545, 7546, 7544, 7543, 7545, 7542, 7538, 7542, 7537, 7533, 7537, 7534, 7541, 7544, 7540, 7542, 7541, 7539, 7537, 7539, 7534, 7550, 7549, 7548, 7554, 7553, 7552, 7560, 7559, 7558, 7564, 7563, 7562, 7558, 7551, 7550, 7562, 7555, 7554, 7551, 7552, 7549, 7561, 7562, 7559, 7559, 7554, 7551, 7521, 7549, 7522, 7552, 7526, 7525, 7532, 7556, 7531, 7549, 7525, 7522, 7531, 7548, 7521, 7482, 7505, 7503, 7515, 7489, 7490, 7513, 7484, 7489, 7533, 7550, 7548, 7556, 7536, 7535, 7539, 7560, 7558, 7548, 7535, 7533, 7534, 7558, 7550, 7568, 7567, 7566, 7577, 7574, 7576, 7578, 7581, 7579, 7585, 7582, 7584, 7578, 7569, 7568, 7583, 7572, 7582, 7584, 7579, 7581, 7571, 7574, 7570, 7567, 7571, 7570, 7577, 7572, 7573, 7576, 7569, 7572, 7579, 7572, 7569, 7586, 7589, 7587, 7593, 7590, 7592, 7600, 7597, 7599, 7602, 7597, 7601, 7588, 7595, 7589, 7600, 7592, 7598, 7592, 7587, 7589, 7598, 7594, 7597, 7601, 7594, 7596, 7595, 7592, 7589, 7580, 7587, 7581, 7591, 7584, 7590, 7590, 7581, 7587, 7373, 7478, 7374, 7481, 7378, 7377, 7478, 7377, 7374,
7503, 7398, 7397, 7401, 7508, 7402, 7512, 7405, 7511, 7410, 7517, 7409, 7511, 7398, 7504, 7409, 7507, 7401, 7482, 7397, 7378, 7566, 7470, 7469, 7575, 7475, 7574, 7470, 7570, 7472, 7574, 7472, 7570, 7340, 7411, 7338, 7352, 7423, 7350, 7350, 7413, 7340, 7386, 7454, 7384, 7402, 7467, 7400, 7400, 7456, 7386, 7384, 7425, 7352, 7510, 7566, 7508, 7526, 7578, 7524, 7524, 7568, 7510, 7555, 7586, 7553, 7563, 7596, 7594, 7563, 7588, 7555, 7553, 7580, 7526, 7508, 7469, 7402, 7337, 7229, 7228, 7233, 7341, 7232, 7225, 7337, 7228, 7225, 7341, 7335, 7411, 7270, 7269, 7415, 7274, 7273, 7420, 7277, 7419, 7412, 7273, 7270, 7416, 7277, 7274, 7435, 7294, 7293, 7438, 7298, 7297, 7302, 7442, 7301, 7294, 7438, 7297, 7301, 7435, 7293, 7338, 7269, 7229, 7605, 7310, 7604, 7603, 7310, 7307, 7304, 7290, 7288, 7298, 7306, 7296, 7282, 7304, 7288, 7296, 7303, 7282, 7605, 7448, 7449, 7448, 7603, 7439, 7298, 7603, 7307, 7608, 7607, 7606, 7612, 7611, 7610, 7615, 7616, 7614, 7619, 7615, 7618, 7622, 7621, 7620, 7626, 7625, 7624, 7620, 7609, 7608, 7612, 7625, 7613, 7610, 7615, 7614, 7618, 7611, 7613,
7607, 7612, 7610, 7623, 7624, 7621, 7621, 7612, 7609, 7634, 7635, 7633, 7643, 7644, 7642, 7644, 7640, 7642, 7641, 7631, 7640, 7638, 7643, 7642, 7638, 7640, 7637, 7629, 7635, 7630, 7637, 7631, 7628, 7648, 7647, 7646, 7652, 7651, 7650, 7656, 7655, 7654, 7660, 7659, 7658, 7654, 7649, 7648, 7658, 7653, 7652, 7649, 7650, 7647, 7657, 7658, 7655, 7655, 7652, 7649, 7664, 7663, 7662, 7670, 7667, 7669, 7675, 7674, 7673, 7683, 7684, 7682, 7684, 7680, 7682, 7664, 7674, 7665, 7681, 7671, 7680, 7665, 7667, 7663, 7678, 7683, 7682, 7678, 7680, 7677, 7671, 7670, 7669, 7676, 7677, 7674, 7665, 7671, 7669, 7674, 7666, 7665, 7677, 7671, 7666, 7646, 7623, 7622, 7650, 7627, 7626, 7647, 7626, 7623, 7662, 7639, 7638, 7668, 7643, 7667, 7663, 7643, 7639, 7631, 7619, 7618, 7627, 7637, 7625, 7613, 7631, 7618, 7625, 7628, 7613, 7653, 7662, 7651, 7661, 7673, 7659, 7659, 7664, 7653, 7627, 7662, 7638, 7686, 7689, 7687, 7692, 7691, 7690, 7698, 7697, 7696, 7702, 7701, 7700, 7709, 7706, 7708, 7705, 7706, 7704, 7696, 7689, 7688, 7700, 7693, 7692, 7695, 7704, 7694, 7706, 7699, 7698, 7696, 7706, 7698,
7689, 7690, 7687, 7697, 7702, 7700, 7697, 7692, 7689, 7688, 7704, 7696, 7712, 7711, 7710, 7718, 7715, 7717, 7723, 7722, 7721, 7730, 7727, 7729, 7712, 7722, 7713, 7728, 7719, 7727, 7713, 7715, 7711, 7726, 7727, 7725, 7719, 7718, 7717, 7724, 7725, 7722, 7713, 7719, 7717, 7722, 7714, 7713, 7725, 7719, 7714, 7733, 7736, 7734, 7738, 7739, 7737, 7745, 7744, 7743, 7748, 7749, 7747, 7743, 7736, 7735, 7742, 7747, 7741, 7731, 7739, 7732, 7743, 7749, 7745, 7732, 7733, 7731, 7742, 7739, 7740, 7739, 7735, 7732, 7735, 7747, 7743, 7753, 7752, 7751, 7757, 7756, 7755, 7760, 7761, 7759, 7764, 7760, 7763, 7767, 7766, 7765, 7771, 7770, 7769, 7776, 7773, 7775, 7765, 7754, 7753, 7769, 7758, 7757, 7774, 7763, 7773, 7755, 7760, 7759, 7763, 7756, 7758, 7775, 7770, 7772, 7752, 7757, 7755, 7768, 7769, 7766, 7766, 7757, 7754, 7770, 7763, 7758, 7733, 7703, 7702, 7709, 7737, 7707, 7731, 7702, 7699, 7699, 7737, 7731, 7751, 7724, 7723, 7761, 7729, 7759, 7752, 7726, 7724, 7759, 7726, 7755, 7693, 7710, 7691, 7703, 7721, 7701, 7701, 7712, 7693, 7736, 7751, 7734, 7746, 7765, 7744, 7744, 7753, 7736,
7734, 7723, 7703, 7779, 7778, 7777, 7783, 7782, 7781, 7786, 7787, 7785, 7792, 7795, 7793, 7799, 7796, 7798, 7799, 7800, 7797, 7789, 7780, 7779, 7792, 7784, 7783, 7788, 7796, 7787, 7777, 7787, 7779, 7796, 7790, 7789, 7797, 7791, 7790, 7780, 7781, 7778, 7791, 7792, 7790, 7803, 7800, 7801, 7800, 7794, 7791, 7780, 7792, 7783, 7779, 7796, 7789, 7806, 7805, 7804, 7810, 7809, 7808, 7815, 7812, 7814, 7818, 7817, 7816, 7820, 7823, 7821, 7816, 7807, 7806, 7820, 7811, 7810, 7825, 7814, 7824, 7811, 7812, 7809, 7807, 7808, 7805, 7819, 7820, 7817, 7817, 7810, 7807, 7824, 7811, 7821, 7828, 7827, 7826, 7835, 7832, 7834, 7831, 7832, 7830, 7839, 7840, 7838, 7835, 7838, 7833, 7832, 7829, 7828, 7826, 7832, 7828, 7836, 7840, 7837, 7829, 7838, 7836, 7844, 7843, 7842, 7848, 7847, 7846, 7850, 7853, 7851, 7856, 7855, 7854, 7850, 7845, 7844, 7854, 7849, 7848, 7845, 7846, 7843, 7853, 7854, 7851, 7851, 7848, 7845, 7826, 7795, 7794, 7803, 7830, 7802, 7794, 7830, 7826, 7842, 7819, 7818, 7846, 7823, 7822, 7843, 7822, 7819, 7784, 7804, 7782, 7795, 7816, 7793, 7793, 7806, 7784, 7827, 7844, 7842,
7837, 7850, 7836, 7859, 7840, 7841, 7840, 7852, 7837, 7836, 7844, 7829, 7827, 7818, 7795, 7866, 7863, 7865, 7866, 7867, 7864, 7875, 7872, 7874, 7868, 7876, 7867, 7874, 7869, 7870, 7864, 7860, 7863, 7861, 7867, 7862, 7867, 7871, 7862, 7880, 7879, 7878, 7884, 7883, 7882, 7888, 7887, 7886, 7892, 7891, 7890, 7895, 7896, 7894, 7880, 7887, 7881, 7890, 7885, 7884, 7894, 7888, 7886, 7881, 7882, 7879, 7889, 7890, 7887, 7887, 7884, 7881, 7900, 7899, 7898, 7903, 7904, 7902, 7907, 7903, 7906, 7910, 7909, 7908, 7916, 7913, 7915, 7908, 7901, 7900, 7914, 7906, 7913, 7898, 7903, 7902, 7901, 7903, 7899, 7911, 7913, 7909, 7917, 7916, 7915, 7911, 7917, 7915, 7909, 7906, 7901, 7921, 7920, 7919, 7925, 7924, 7923, 7932, 7929, 7931, 7928, 7929, 7927, 7937, 7936, 7935, 7940, 7941, 7939, 7935, 7926, 7925, 7932, 7939, 7930, 7929, 7922, 7921, 7927, 7921, 7919, 7933, 7941, 7934, 7920, 7925, 7923, 7934, 7935, 7933, 7933, 7925, 7922, 7930, 7933, 7922, 7904, 7874, 7902, 7898, 7874, 7870, 7919, 7889, 7888, 7923, 7893, 7892, 7897, 7927, 7896, 7920, 7892, 7889, 7888, 7927, 7919, 7862, 7878, 7861,
7895, 7876, 7877, 7876, 7886, 7871, 7871, 7880, 7862, 7656, 7778, 7657, 7781, 7661, 7660, 7778, 7660, 7657, 7804, 7676, 7675, 7808, 7679, 7678, 7813, 7683, 7812, 7676, 7808, 7678, 7809, 7683, 7679, 7782, 7675, 7661, 7860, 7746, 7745, 7750, 7863, 7749, 7745, 7863, 7860, 7878, 7768, 7767, 7771, 7883, 7772, 7879, 7771, 7768, 7746, 7878, 7767, 7630, 7686, 7629, 7695, 7635, 7636, 7635, 7688, 7630, 7823, 7869, 7821, 7872, 7825, 7824, 7821, 7872, 7824, 7847, 7900, 7898, 7857, 7908, 7855, 7855, 7900, 7849, 7847, 7870, 7823, 7945, 7948, 7946, 7950, 7951, 7949, 7961, 7958, 7960, 7961, 7962, 7959, 7955, 7948, 7947, 7954, 7958, 7953, 7949, 7944, 7943, 7959, 7955, 7958, 7956, 7962, 7957, 7944, 7945, 7943, 7954, 7951, 7952, 7951, 7947, 7944, 7953, 7955, 7947, 7966, 7965, 7964, 7970, 7969, 7968, 7974, 7973, 7972, 7976, 7979, 7977, 7983, 7980, 7982, 7987, 7984, 7986, 7972, 7967, 7966, 7976, 7971, 7970, 7981, 7974, 7980, 7985, 7978, 7984, 7967, 7968, 7965, 7975, 7976, 7973, 7986, 7981, 7983, 7981, 7978, 7975, 7973, 7970, 7967, 7948, 7964, 7946, 7957, 7972, 7956, 7982, 7962, 7963,
7980, 7957, 7962, 7948, 7972, 7966, 7993, 7992, 7991, 8002, 7999, 8001, 8009, 8006, 8008, 8011, 8006, 8010, 7993, 8004, 7994, 8009, 7997, 8007, 8006, 8004, 8003, 8010, 8003, 8005, 8001, 7995, 7996, 7992, 7996, 7995, 8002, 7997, 7998, 8001, 7994, 7997, 8004, 7997, 7994, 8018, 8019, 8017, 8016, 8018, 8015, 8014, 8017, 8013, 8012, 8018, 8014, 8026, 8023, 8025, 8033, 8034, 8032, 8034, 8030, 8032, 8031, 8025, 8030, 8022, 8023, 8021, 8028, 8033, 8032, 8032, 8027, 8028, 8027, 8025, 8022, 8038, 8037, 8036, 8043, 8040, 8042, 8046, 8045, 8044, 8051, 8048, 8050, 8044, 8039, 8038, 8049, 8042, 8048, 8039, 8040, 8037, 8047, 8048, 8045, 8045, 8042, 8039, 8036, 8029, 8028, 8041, 8033, 8040, 8040, 8029, 8037, 8059, 8062, 8060, 8066, 8067, 8065, 8069, 8072, 8070, 8075, 8074, 8073, 8078, 8079, 8077, 8079, 8081, 8077, 8073, 8062, 8061, 8064, 8078, 8063, 8070, 8077, 8069, 8069, 8081, 8071, 8065, 8061, 8059, 8067, 8064, 8063, 8072, 8073, 8070, 8061, 8063, 8058, 8058, 8073, 8061, 8058, 8078, 8070, 8084, 8087, 8085, 8091, 8092, 8090, 8092, 8094, 8090, 8096, 8099, 8097, 8102, 8101, 8100,
8105, 8106, 8104, 8106, 8108, 8104, 8100, 8087, 8086, 8089, 8105, 8088, 8096, 8105, 8104, 8096, 8108, 8098, 8084, 8091, 8090, 8090, 8086, 8084, 8094, 8089, 8088, 8099, 8100, 8097, 8083, 8094, 8088, 8083, 8100, 8086, 8088, 8097, 8083, 8112, 8111, 8110, 8116, 8115, 8114, 8120, 8119, 8118, 8122, 8125, 8123, 8131, 8128, 8130, 8131, 8126, 8129, 8118, 8113, 8112, 8122, 8117, 8116, 8129, 8124, 8128, 8125, 8126, 8123, 8113, 8114, 8111, 8121, 8122, 8119, 8119, 8116, 8113, 8057, 8091, 8056, 8056, 8085, 8055, 8110, 8072, 8071, 8114, 8076, 8075, 8111, 8075, 8072, 8081, 8057, 8056, 8055, 8081, 8056, 8087, 8110, 8085, 8103, 8118, 8101, 8101, 8112, 8087, 8055, 8110, 8071, 8024, 7989, 8023, 8023, 7988, 8021, 8013, 8060, 8014, 8019, 8065, 8017, 8017, 8059, 8013, 7971, 7991, 7969, 7977, 8005, 8003, 8011, 7985, 7987, 8010, 7979, 7985, 7977, 7993, 7971, 8108, 8054, 8053, 8098, 8053, 8052, 8134, 8133, 8132, 8138, 8137, 8136, 8145, 8142, 8144, 8141, 8142, 8140, 8150, 8149, 8148, 8153, 8154, 8152, 8148, 8139, 8138, 8145, 8152, 8143, 8142, 8135, 8134, 8132, 8142, 8134, 8146, 8154, 8147,
8133, 8138, 8136, 8147, 8148, 8146, 8146, 8138, 8135, 8143, 8146, 8135, 8159, 8158, 8157, 8164, 8165, 8163, 8169, 8168, 8167, 8176, 8173, 8175, 8172, 8173, 8171, 8167, 8160, 8159, 8166, 8171, 8165, 8157, 8165, 8159, 8173, 8170, 8169, 8171, 8169, 8167, 8164, 8161, 8162, 8161, 8157, 8156, 8159, 8171, 8167, 8150, 8158, 8151, 8155, 8161, 8154, 8156, 8150, 8147, 8147, 8161, 8156, 8179, 8178, 8177, 8186, 8183, 8185, 8189, 8188, 8187, 8196, 8193, 8195, 8187, 8180, 8179, 8194, 8185, 8193, 8182, 8183, 8181, 8192, 8193, 8191, 8180, 8181, 8178, 8190, 8191, 8188, 8188, 8182, 8180, 8191, 8185, 8182, 8199, 8202, 8200, 8204, 8205, 8203, 8211, 8210, 8209, 8214, 8215, 8213, 8209, 8202, 8201, 8208, 8213, 8207, 8203, 8198, 8197, 8209, 8215, 8211, 8198, 8199, 8197, 8208, 8205, 8206, 8205, 8201, 8198, 8207, 8209, 8201, 8219, 8218, 8217, 8223, 8222, 8221, 8226, 8227, 8225, 8230, 8226, 8229, 8233, 8232, 8231, 8237, 8236, 8235, 8242, 8239, 8241, 8231, 8220, 8219, 8235, 8224, 8223, 8240, 8229, 8239, 8221, 8226, 8225, 8229, 8222, 8224, 8238, 8239, 8236, 8218, 8223, 8221, 8234, 8235, 8232,
8232, 8223, 8220, 8236, 8229, 8224, 8245, 8244, 8243, 8248, 8249, 8247, 8251, 8254, 8252, 8256, 8257, 8255, 8251, 8246, 8245, 8250, 8255, 8249, 8243, 8249, 8245, 8251, 8257, 8253, 8245, 8255, 8251, 8217, 8190, 8189, 8227, 8195, 8225, 8218, 8192, 8190, 8225, 8192, 8221, 8243, 8212, 8211, 8216, 8247, 8215, 8211, 8247, 8243, 8262, 8265, 8263, 8266, 8269, 8267, 8271, 8272, 8270, 8275, 8276, 8274, 8261, 8275, 8260, 8263, 8270, 8262, 8266, 8275, 8274, 8263, 8268, 8266, 8274, 8273, 8271, 8266, 8271, 8263, 8260, 8267, 8259, 8284, 8285, 8283, 8290, 8291, 8289, 8288, 8290, 8287, 8295, 8294, 8293, 8299, 8298, 8297, 8278, 8294, 8279, 8297, 8282, 8281, 8279, 8283, 8278, 8282, 8289, 8281, 8280, 8290, 8282, 8289, 8286, 8284, 8296, 8297, 8294, 8281, 8284, 8279, 8279, 8297, 8281, 8303, 8302, 8301, 8307, 8306, 8305, 8318, 8315, 8317, 8322, 8319, 8321, 8303, 8312, 8304, 8307, 8314, 8308, 8322, 8309, 8320, 8316, 8311, 8315, 8319, 8314, 8313, 8304, 8305, 8302, 8321, 8316, 8318, 8319, 8312, 8316, 8304, 8313, 8307, 8320, 8308, 8314, 8323, 8326, 8324, 8327, 8330, 8328, 8334, 8331, 8333,
8338, 8335, 8337, 8332, 8325, 8331, 8336, 8329, 8335, 8324, 8329, 8327, 8337, 8332, 8334, 8335, 8326, 8332, 8301, 8265, 8264, 8305, 8269, 8268, 8302, 8268, 8265, 8323, 8296, 8295, 8327, 8300, 8299, 8324, 8299, 8296, 8283, 8261, 8260, 8269, 8293, 8267, 8259, 8283, 8260, 8259, 8293, 8278, 8308, 8323, 8306, 8333, 8309, 8310, 8309, 8325, 8308, 8306, 8295, 8269, 8341, 8340, 8339, 8343, 8346, 8344, 8350, 8347, 8349, 8353, 8352, 8351, 8355, 8358, 8356, 8362, 8359, 8361, 8351, 8342, 8341, 8355, 8346, 8345, 8360, 8349, 8359, 8346, 8347, 8344, 8361, 8356, 8358, 8342, 8343, 8340, 8354, 8355, 8352, 8352, 8345, 8342, 8359, 8346, 8356, 8363, 8366, 8364, 8368, 8369, 8367, 8373, 8372, 8371, 8376, 8377, 8375, 8371, 8366, 8365, 8370, 8375, 8369, 8367, 8365, 8363, 8375, 8373, 8371, 8369, 8371, 8365, 8379, 8382, 8380, 8385, 8384, 8383, 8392, 8389, 8391, 8395, 8394, 8393, 8397, 8400, 8398, 8402, 8403, 8401, 8408, 8405, 8407, 8393, 8382, 8381, 8397, 8386, 8385, 8388, 8401, 8387, 8406, 8391, 8405, 8401, 8395, 8393, 8391, 8384, 8386, 8407, 8398, 8400, 8382, 8383, 8380, 8396, 8397, 8394,
8394, 8385, 8382, 8381, 8401, 8393, 8405, 8386, 8398, 8409, 8412, 8410, 8414, 8415, 8413, 8418, 8421, 8419, 8425, 8426, 8424, 8411, 8419, 8412, 8416, 8424, 8415, 8413, 8411, 8409, 8418, 8426, 8420, 8426, 8423, 8422, 8420, 8422, 8417, 8415, 8418, 8411, 8379, 8354, 8353, 8383, 8358, 8357, 8390, 8361, 8389, 8380, 8357, 8354, 8389, 8358, 8384, 8373, 8410, 8374, 8378, 8413, 8377, 8377, 8409, 8373, 8176, 8287, 8174, 8174, 8280, 8170, 8339, 8234, 8233, 8343, 8238, 8237, 8348, 8241, 8347, 8340, 8237, 8234, 8344, 8241, 8238, 8253, 8364, 8254, 8258, 8367, 8257, 8253, 8367, 8363, 8139, 8177, 8137, 8151, 8187, 8149, 8149, 8179, 8139, 8160, 8217, 8158, 8168, 8233, 8231, 8168, 8219, 8160, 8158, 8189, 8151, 8282, 8339, 8280, 8300, 8351, 8298, 8298, 8341, 8282, 8328, 8381, 8379, 8388, 8336, 8338, 8387, 8330, 8336, 8328, 8353, 8300, 8280, 8233, 8170, 8430, 8429, 8428, 8435, 8432, 8434, 8436, 8439, 8437, 8448, 8449, 8447, 8449, 8445, 8447, 8436, 8431, 8430, 8444, 8434, 8443, 8431, 8432, 8429, 8442, 8447, 8441, 8447, 8440, 8441, 8439, 8440, 8437, 8445, 8444, 8443, 8437, 8445, 8443,
8443, 8431, 8437, 8452, 8455, 8453, 8459, 8460, 8458, 8460, 8462, 8458, 8464, 8467, 8465, 8470, 8469, 8468, 8473, 8474, 8472, 8474, 8476, 8472, 8468, 8455, 8454, 8457, 8473, 8456, 8464, 8473, 8472, 8472, 8466, 8464, 8453, 8458, 8452, 8452, 8462, 8454, 8462, 8457, 8456, 8467, 8468, 8465, 8451, 8462, 8456, 8451, 8468, 8454, 8456, 8465, 8451, 8480, 8479, 8478, 8484, 8483, 8482, 8488, 8487, 8486, 8492, 8491, 8490, 8486, 8481, 8480, 8490, 8485, 8484, 8479, 8484, 8482, 8489, 8490, 8487, 8487, 8484, 8481, 8496, 8495, 8494, 8500, 8499, 8498, 8502, 8505, 8503, 8512, 8509, 8511, 8516, 8513, 8515, 8518, 8513, 8517, 8502, 8497, 8496, 8500, 8507, 8501, 8510, 8504, 8509, 8514, 8506, 8513, 8517, 8506, 8508, 8497, 8498, 8495, 8503, 8508, 8506, 8518, 8510, 8512, 8517, 8505, 8510, 8503, 8500, 8497, 8478, 8439, 8438, 8482, 8442, 8441, 8479, 8441, 8439, 8494, 8467, 8466, 8498, 8471, 8470, 8495, 8470, 8467, 8476, 8450, 8448, 8466, 8448, 8442, 8485, 8494, 8483, 8493, 8502, 8491, 8491, 8496, 8485, 8442, 8494, 8466, 8521, 8520, 8519, 8525, 8524, 8523, 8529, 8528, 8527, 8531, 8534, 8532,
8540, 8537, 8539, 8540, 8535, 8538, 8527, 8522, 8521, 8531, 8526, 8525, 8537, 8534, 8533, 8538, 8532, 8534, 8522, 8523, 8520, 8530, 8531, 8528, 8528, 8525, 8522, 8541, 8544, 8542, 8551, 8548, 8550, 8551, 8546, 8549, 8556, 8548, 8555, 8548, 8544, 8543, 8549, 8542, 8544, 8553, 8546, 8547, 8552, 8542, 8546, 8555, 8543, 8554, 8557, 8560, 8558, 8567, 8564, 8566, 8567, 8562, 8565, 8572, 8564, 8571, 8565, 8559, 8564, 8565, 8558, 8560, 8569, 8562, 8563, 8562, 8561, 8558, 8570, 8564, 8559, 8557, 8530, 8529, 8569, 8537, 8568, 8530, 8561, 8533, 8568, 8533, 8561, 8526, 8541, 8524, 8556, 8535, 8536, 8535, 8554, 8532, 8526, 8554, 8543, 8575, 8574, 8573, 8585, 8582, 8584, 8587, 8582, 8586, 8588, 8591, 8589, 8595, 8592, 8594, 8575, 8589, 8576, 8593, 8580, 8592, 8594, 8589, 8591, 8582, 8578, 8577, 8586, 8577, 8579, 8576, 8577, 8574, 8587, 8580, 8581, 8580, 8579, 8576, 8592, 8576, 8589, 8599, 8602, 8600, 8606, 8603, 8605, 8613, 8610, 8612, 8615, 8610, 8614, 8601, 8608, 8602, 8613, 8605, 8611, 8605, 8600, 8602, 8611, 8607, 8610, 8609, 8610, 8607, 8608, 8605, 8602, 8590, 8600, 8591,
8604, 8594, 8603, 8603, 8591, 8600, 8598, 8583, 8585, 8597, 8578, 8583, 8573, 8489, 8488, 8492, 8578, 8493, 8574, 8492, 8489, 8598, 8509, 8597, 8597, 8504, 8596, 8493, 8596, 8504, 8455, 8519, 8453, 8471, 8527, 8469, 8469, 8521, 8455, 8501, 8557, 8499, 8572, 8514, 8516, 8571, 8507, 8514, 8501, 8570, 8559, 8499, 8529, 8471, 8132, 7853, 7852, 8136, 7857, 7856, 7859, 8140, 7858, 8133, 7856, 7853, 7852, 8140, 8132, 7910, 8178, 7911, 8184, 7917, 8183, 8178, 7912, 7911, 8183, 7912, 8181, 8199, 7938, 7937, 7942, 8203, 7941, 8197, 7937, 7934, 7934, 8203, 8197, 8137, 7910, 7857, 8428, 8047, 8046, 8433, 8050, 8432, 8429, 8050, 8047, 8054, 8459, 8053, 8052, 8459, 8453, 8519, 8099, 8098, 8523, 8103, 8102, 8520, 8102, 8099, 8541, 8121, 8120, 8553, 8128, 8552, 8121, 8545, 8124, 8545, 8128, 8124, 8524, 8120, 8103, 8052, 8519, 8098, 7989, 7776, 7775, 7772, 7989, 7775, 7885, 8021, 7883, 7893, 8027, 7891, 7891, 8022, 7885, 7926, 8036, 7924, 7938, 8044, 7936, 7936, 8038, 7926, 7893, 8036, 8028, 7883, 7988, 7772, 8202, 8428, 8200, 8212, 8436, 8210, 8210, 8430, 8202, 8246, 8478, 8244,
8254, 8486, 8252, 8252, 8480, 8246, 8244, 8438, 8212, 8366, 8573, 8364, 8374, 8588, 8372, 8372, 8575, 8366, 8412, 8599, 8410, 8419, 8609, 8607, 8419, 8601, 8412, 8410, 8590, 8374, 8364, 8488, 8254, 8200, 8046, 7938, 8616, 8619, 8617, 8621, 8622, 8620, 8628, 8627, 8626, 8631, 8632, 8630, 8618, 8627, 8619, 8623, 8634, 8622, 8630, 8625, 8624, 8620, 8618, 8616, 8624, 8628, 8626, 8634, 8631, 8630, 8626, 8630, 8624, 8618, 8634, 8626, 8638, 8637, 8636, 8642, 8641, 8640, 8648, 8647, 8646, 8650, 8653, 8651, 8659, 8656, 8658, 8659, 8654, 8657, 8646, 8639, 8638, 8642, 8651, 8643, 8655, 8644, 8654, 8657, 8652, 8656, 8653, 8654, 8651, 8639, 8640, 8637, 8649, 8650, 8647, 8647, 8642, 8639, 8654, 8643, 8651, 8660, 8663, 8661, 8666, 8665, 8664, 8669, 8670, 8668, 8670, 8672, 8668, 8676, 8675, 8674, 8680, 8679, 8678, 8662, 8675, 8663, 8678, 8667, 8666, 8673, 8682, 8672, 8661, 8668, 8660, 8668, 8662, 8660, 8663, 8664, 8661, 8677, 8678, 8675, 8675, 8666, 8663, 8672, 8674, 8662, 8686, 8685, 8684, 8695, 8692, 8694, 8696, 8699, 8697, 8705, 8702, 8704, 8705, 8700, 8703, 8686, 8697, 8687,
8701, 8690, 8700, 8703, 8698, 8702, 8703, 8697, 8699, 8694, 8688, 8689, 8685, 8689, 8688, 8695, 8690, 8691, 8694, 8687, 8690, 8700, 8687, 8697, 8664, 8629, 8628, 8633, 8669, 8632, 8625, 8664, 8628, 8632, 8661, 8625, 8684, 8649, 8648, 8693, 8656, 8692, 8649, 8688, 8652, 8688, 8656, 8652, 8619, 8636, 8617, 8629, 8646, 8627, 8627, 8638, 8619, 8667, 8684, 8665, 8681, 8696, 8679, 8679, 8686, 8667, 8665, 8648, 8629, 8711, 8708, 8710, 8710, 8706, 8707, 8714, 8713, 8712, 8718, 8717, 8716, 8724, 8723, 8722, 8732, 8729, 8731, 8734, 8729, 8733, 8722, 8715, 8714, 8718, 8727, 8719, 8732, 8720, 8730, 8730, 8726, 8729, 8733, 8726, 8728, 8715, 8716, 8713, 8723, 8728, 8726, 8723, 8718, 8715, 8727, 8720, 8719, 8740, 8737, 8739, 8739, 8735, 8736, 8741, 8744, 8742, 8751, 8748, 8750, 8751, 8746, 8749, 8756, 8748, 8755, 8749, 8743, 8748, 8744, 8746, 8742, 8753, 8746, 8747, 8752, 8742, 8746, 8754, 8748, 8743, 8741, 8725, 8724, 8753, 8733, 8752, 8725, 8745, 8728, 8752, 8728, 8745, 8717, 8736, 8735, 8740, 8720, 8721, 8739, 8719, 8720, 8712, 8677, 8676, 8716, 8681, 8680, 8713, 8680, 8677,
8738, 8702, 8737, 8737, 8698, 8735, 8681, 8735, 8698, 8641, 8707, 8706, 8711, 8644, 8645, 8710, 8643, 8644, 7606, 6800, 6799, 7616, 6807, 7614, 7607, 6802, 6800, 7610, 6807, 6802, 6834, 7633, 6832, 6826, 7633, 7629, 7686, 6875, 6874, 7690, 6879, 6878, 7687, 6878, 6875, 7710, 6885, 6884, 7716, 6889, 7715, 7711, 6889, 6885, 6879, 7710, 6884, 7629, 6874, 6826, 7945, 6965, 6964, 6969, 7949, 6968, 7943, 6964, 6961, 6961, 7949, 7943, 7964, 6981, 6980, 7968, 6985, 6984, 7965, 6984, 6981, 7946, 6980, 6965, 7991, 7023, 7022, 8000, 7028, 7999, 7023, 7995, 7025, 7999, 7025, 7995, 7969, 7022, 6985, 7540, 8617, 7541, 7546, 8620, 7544, 7544, 8616, 7540, 8636, 7561, 7560, 8640, 7565, 7564, 8637, 7564, 7561, 8617, 7560, 7541, 8709, 7601, 8708, 8708, 7596, 8706, 7565, 8706, 7596, 6894, 7323, 7321, 7329, 6902, 6900, 6894, 7329, 6900, 6910, 7363, 6908, 6920, 7371, 6918, 6918, 7365, 6910, 6908, 7323, 6896, 7486, 6997, 6999, 7487, 7003, 7001, 7497, 7005, 7006, 6997, 7477, 6991, 7479, 7001, 6993, 7495, 7004, 7005, 7004, 7479, 6993, 6920, 7477, 7373, 8012, 8676, 8674, 8682, 8016, 8015,
8674, 8015, 8012, 8062, 8712, 8060, 8076, 8722, 8074, 8074, 8714, 8062, 8117, 8741, 8115, 8756, 8126, 8127, 8755, 8123, 8126, 8117, 8754, 8743, 8115, 8724, 8076, 8060, 8676, 8014, 5900, 6454, 5901, 6457, 5905, 5904, 6454, 5904, 5901, 6481, 5917, 5916, 6485, 5921, 5920, 6482, 5920, 5917, 6458, 5916, 5905, 6516, 5965, 5964, 5968, 6521, 5969, 6525, 5970, 6524, 6517, 5968, 5965, 6521, 5970, 5969, 6486, 5964, 5921, 7030, 6338, 6337, 7033, 6342, 6341, 6344, 7037, 6343, 6338, 7033, 6341, 6343, 7030, 6337, 7062, 6360, 6359, 7066, 6364, 6363, 7063, 6363, 6360, 6342, 7062, 6359, 7118, 6421, 6420, 7122, 6425, 6424, 7119, 6424, 6421, 7136, 6442, 6441, 7145, 6449, 7144, 6442, 7140, 6445, 7144, 6445, 7140, 7123, 6441, 6425, 7067, 6420, 6364, 8760, 8761, 8759, 8766, 8767, 8765, 8762, 8765, 8761, 8757, 8761, 8758, 8765, 8764, 8763, 8761, 8763, 8758, 8769, 8772, 8770, 8773, 8776, 8774, 8778, 8779, 8777, 8786, 8783, 8785, 8786, 8781, 8784, 8787, 8790, 8788, 8794, 8795, 8793, 8800, 8797, 8799, 8787, 8772, 8771, 8780, 8793, 8779, 8798, 8783, 8797, 8777, 8771, 8769, 8793, 8789, 8787,
8784, 8775, 8783, 8776, 8781, 8774, 8799, 8791, 8792, 8772, 8773, 8770, 8788, 8792, 8791, 8772, 8791, 8775, 8779, 8787, 8771, 8791, 8783, 8775, 8805, 8804, 8803, 8808, 8809, 8807, 8816, 8819, 8817, 8823, 8824, 8822, 8821, 8823, 8820, 8829, 8826, 8828, 8805, 8817, 8806, 8810, 8820, 8809, 8827, 8811, 8826, 8807, 8802, 8801, 8815, 8822, 8814, 8813, 8823, 8815, 8828, 8817, 8819, 8801, 8805, 8803, 8813, 8818, 8816, 8802, 8816, 8805, 8809, 8813, 8802, 8817, 8811, 8806, 8832, 8831, 8830, 8843, 8840, 8842, 8844, 8847, 8845, 8853, 8850, 8852, 8855, 8856, 8854, 8832, 8845, 8833, 8851, 8838, 8850, 8852, 8845, 8847, 8835, 8840, 8834, 8848, 8856, 8849, 8831, 8835, 8834, 8843, 8838, 8839, 8842, 8833, 8838, 8850, 8833, 8845, 8858, 8861, 8859, 8866, 8867, 8865, 8872, 8869, 8871, 8874, 8869, 8873, 8877, 8876, 8875, 8885, 8881, 8884, 8875, 8861, 8860, 8868, 8880, 8867, 8887, 8873, 8886, 8865, 8860, 8858, 8884, 8876, 8878, 8870, 8862, 8869, 8873, 8862, 8864, 8859, 8864, 8862, 8887, 8881, 8883, 8886, 8876, 8881, 8861, 8879, 8864, 8860, 8880, 8875, 8886, 8864, 8879, 8789, 8831, 8790,
8796, 8836, 8795, 8841, 8799, 8840, 8790, 8834, 8792, 8789, 8836, 8830, 8840, 8792, 8834, 8814, 8859, 8815, 8818, 8863, 8819, 8824, 8865, 8822, 8872, 8828, 8870, 8859, 8818, 8815, 8814, 8865, 8858, 8870, 8819, 8863, 8848, 8877, 8875, 8880, 8855, 8854, 8875, 8854, 8848, 8894, 8895, 8893, 8892, 8894, 8891, 8897, 8900, 8898, 8903, 8905, 8901, 8908, 8902, 8906, 8897, 8890, 8889, 8895, 8901, 8893, 8890, 8893, 8889, 8891, 8890, 8888, 8901, 8899, 8897, 8906, 8898, 8900, 8893, 8897, 8889, 8914, 8911, 8913, 8912, 8909, 8911, 8909, 8900, 8899, 8907, 8911, 8905, 8914, 8906, 8912, 8905, 8909, 8899, 8912, 8900, 8910, 8920, 8919, 8918, 8925, 8922, 8924, 8927, 8922, 8926, 8929, 8930, 8928, 8930, 8932, 8928, 8941, 8940, 8939, 8946, 8947, 8944, 8956, 8953, 8955, 8956, 8950, 8954, 8920, 8940, 8921, 8945, 8926, 8943, 8934, 8949, 8932, 8952, 8933, 8950, 8923, 8915, 8922, 8926, 8915, 8917, 8944, 8938, 8937, 8919, 8928, 8918, 8928, 8920, 8918, 8954, 8941, 8953, 8954, 8940, 8942, 8938, 8939, 8937, 8949, 8946, 8944, 8939, 8944, 8937, 8943, 8917, 8936, 8932, 8939, 8920, 8950, 8921, 8940,
8963, 8960, 8962, 8965, 8960, 8964, 8961, 8957, 8960, 8959, 8960, 8957, 8966, 8969, 8967, 8972, 8973, 8971, 8978, 8975, 8977, 8973, 8975, 8971, 8978, 8972, 8976, 8967, 8971, 8966, 8976, 8968, 8975, 8966, 8975, 8968, 8969, 8972, 8967, 8980, 8972, 8974, 8979, 8967, 8972, 8948, 8972, 8947, 8980, 8953, 8979, 8938, 8970, 8941, 8947, 8967, 8938, 8970, 8953, 8941, 8921, 8957, 8919, 8965, 8933, 8935, 8964, 8921, 8933, 8768, 8891, 8767, 8767, 8888, 8764, 8846, 8916, 8847, 8925, 8852, 8923, 8857, 8929, 8856, 8923, 8847, 8916, 8856, 8919, 8849, 8877, 8958, 8878, 8963, 8884, 8961, 8961, 8878, 8958, 8919, 8877, 8849, 8757, 8832, 8830, 8836, 8760, 8759, 8763, 8846, 8844, 8830, 8759, 8757, 8758, 8844, 8832, 8888, 8917, 8915, 8945, 8902, 8904, 8943, 8898, 8902, 8898, 8917, 8890, 8888, 8846, 8764, 8986, 8983, 8985, 8982, 8983, 8981, 8992, 8993, 8991, 8996, 8997, 8995, 9001, 9000, 8999, 9005, 9004, 9003, 8999, 8988, 8987, 9003, 8990, 8989, 8988, 8991, 8987, 8990, 8995, 8989, 8995, 8994, 8992, 9002, 9003, 9000, 8989, 8992, 8988, 9000, 8989, 8988, 9012, 9013, 9011, 9016, 9017, 9015,
9021, 9020, 9019, 9025, 9024, 9023, 9019, 9008, 9007, 9023, 9010, 9009, 9008, 9011, 9007, 9010, 9015, 9009, 9015, 9014, 9012, 9022, 9023, 9020, 9009, 9012, 9008, 9020, 9009, 9008, 9029, 9028, 9027, 9033, 9032, 9031, 9042, 9039, 9041, 9046, 9043, 9045, 9035, 9030, 9029, 9037, 9034, 9033, 9040, 9035, 9039, 9044, 9037, 9043, 9030, 9031, 9028, 9045, 9040, 9042, 9043, 9036, 9040, 9036, 9033, 9030, 9049, 9048, 9047, 9053, 9052, 9051, 9055, 9058, 9056, 9061, 9060, 9059, 9064, 9065, 9063, 9049, 9056, 9050, 9059, 9054, 9053, 9063, 9057, 9055, 9050, 9051, 9048, 9058, 9059, 9056, 9056, 9053, 9050, 9027, 9002, 9001, 9031, 9006, 9005, 9028, 9005, 9002, 9047, 9022, 9021, 9051, 9026, 9025, 9048, 9025, 9022, 9011, 8998, 8996, 9006, 9019, 9004, 9007, 8996, 8990, 9004, 9007, 8990, 9034, 9047, 9032, 9064, 9044, 9046, 9063, 9038, 9044, 9038, 9049, 9034, 9032, 9021, 9006, 9072, 9073, 9071, 9076, 9077, 9075, 9081, 9080, 9079, 9085, 9084, 9083, 9079, 9068, 9067, 9083, 9070, 9069, 9068, 9071, 9067, 9070, 9075, 9069, 9075, 9074, 9072, 9082, 9083, 9080, 9069, 9072, 9068, 9080, 9069, 9068,
9092, 9093, 9091, 9096, 9097, 9095, 9101, 9100, 9099, 9105, 9104, 9103, 9099, 9088, 9087, 9103, 9090, 9089, 9088, 9091, 9087, 9090, 9095, 9089, 9095, 9094, 9092, 9102, 9103, 9100, 9089, 9092, 9088, 9100, 9089, 9088, 9109, 9108, 9107, 9113, 9112, 9111, 9117, 9116, 9115, 9125, 9122, 9124, 9127, 9122, 9126, 9115, 9110, 9109, 9113, 9120, 9114, 9123, 9119, 9122, 9126, 9119, 9121, 9110, 9111, 9108, 9118, 9119, 9116, 9116, 9113, 9110, 9130, 9129, 9128, 9134, 9133, 9132, 9143, 9140, 9142, 9147, 9144, 9146, 9130, 9137, 9131, 9134, 9139, 9135, 9140, 9137, 9136, 9144, 9139, 9138, 9131, 9132, 9129, 9146, 9141, 9143, 9141, 9138, 9137, 9131, 9138, 9134, 9107, 9082, 9081, 9111, 9086, 9085, 9108, 9085, 9082, 9128, 9102, 9101, 9132, 9106, 9105, 9129, 9105, 9102, 9091, 9078, 9076, 9086, 9099, 9084, 9087, 9076, 9070, 9084, 9087, 9070, 9114, 9128, 9112, 9142, 9123, 9125, 9140, 9120, 9123, 9120, 9130, 9114, 9112, 9101, 9086, 9071, 9018, 9016, 9026, 9079, 9024, 9067, 9016, 9010, 9024, 9067, 9010, 9054, 9107, 9052, 9062, 9115, 9060, 9060, 9109, 9054, 9052, 9081, 9026, 9153, 9154, 9152,
9157, 9158, 9156, 9162, 9161, 9160, 9166, 9165, 9164, 9160, 9149, 9148, 9164, 9151, 9150, 9149, 9152, 9148, 9151, 9156, 9150, 9156, 9155, 9153, 9163, 9164, 9161, 9150, 9153, 9149, 9161, 9150, 9149, 9173, 9174, 9172, 9177, 9178, 9176, 9182, 9181, 9180, 9186, 9185, 9184, 9191, 9188, 9190, 9180, 9169, 9168, 9184, 9171, 9170, 9189, 9177, 9188, 9169, 9172, 9168, 9171, 9176, 9170, 9190, 9185, 9187, 9176, 9175, 9173, 9183, 9184, 9181, 9170, 9173, 9169, 9181, 9170, 9169, 9188, 9171, 9185, 9194, 9193, 9192, 9198, 9197, 9196, 9202, 9201, 9200, 9206, 9205, 9204, 9213, 9210, 9212, 9209, 9210, 9208, 9194, 9201, 9195, 9204, 9199, 9198, 9211, 9202, 9210, 9200, 9210, 9202, 9195, 9196, 9193, 9201, 9206, 9204, 9201, 9198, 9195, 9216, 9215, 9214, 9220, 9219, 9218, 9225, 9222, 9224, 9228, 9227, 9226, 9232, 9231, 9230, 9237, 9234, 9236, 9226, 9217, 9216, 9230, 9221, 9220, 9235, 9224, 9234, 9224, 9219, 9221, 9236, 9231, 9233, 9217, 9218, 9215, 9229, 9230, 9227, 9227, 9220, 9217, 9234, 9221, 9231, 9192, 9163, 9162, 9196, 9167, 9166, 9193, 9166, 9163, 9214, 9183, 9182, 9218, 9187, 9186,
9223, 9190, 9222, 9215, 9186, 9183, 9222, 9187, 9219, 9172, 9159, 9157, 9167, 9180, 9165, 9168, 9157, 9151, 9165, 9168, 9151, 9199, 9214, 9197, 9207, 9226, 9205, 9205, 9216, 9199, 9197, 9182, 9167, 8804, 8982, 8981, 8986, 8811, 8812, 8985, 8806, 8811, 9152, 9098, 9096, 9106, 9160, 9104, 9148, 9096, 9090, 9104, 9148, 9090, 9135, 9192, 9133, 9209, 9145, 9147, 9208, 9139, 9145, 9139, 9194, 9135, 9133, 9162, 9106, 9240, 9239, 9238, 9244, 9243, 9242, 9247, 9248, 9246, 9252, 9251, 9250, 9256, 9255, 9254, 9259, 9260, 9258, 9250, 9241, 9240, 9254, 9245, 9244, 9249, 9258, 9248, 9246, 9240, 9238, 9258, 9252, 9250, 9241, 9242, 9239, 9253, 9254, 9251, 9251, 9244, 9241, 9248, 9250, 9240, 9264, 9263, 9262, 9268, 9267, 9266, 9271, 9272, 9270, 9276, 9275, 9274, 9280, 9279, 9278, 9283, 9284, 9282, 9274, 9265, 9264, 9278, 9269, 9268, 9273, 9282, 9272, 9270, 9264, 9262, 9282, 9276, 9274, 9265, 9266, 9263, 9277, 9278, 9275, 9275, 9268, 9265, 9272, 9274, 9264, 9262, 9253, 9252, 9266, 9257, 9256, 9261, 9270, 9260, 9263, 9256, 9253, 9260, 9262, 9252, 9288, 9287, 9286, 9295, 9292, 9294,
9298, 9297, 9296, 9305, 9302, 9304, 9296, 9289, 9288, 9303, 9294, 9302, 9294, 9290, 9291, 9304, 9300, 9301, 9289, 9290, 9287, 9299, 9300, 9297, 9297, 9291, 9289, 9302, 9291, 9300, 9308, 9307, 9306, 9315, 9312, 9314, 9318, 9317, 9316, 9325, 9322, 9324, 9316, 9309, 9308, 9323, 9314, 9322, 9314, 9310, 9311, 9324, 9320, 9321, 9309, 9310, 9307, 9319, 9320, 9317, 9317, 9311, 9309, 9322, 9311, 9320, 9306, 9299, 9298, 9313, 9304, 9312, 9307, 9301, 9299, 9312, 9301, 9310, 9328, 9327, 9326, 9332, 9331, 9330, 9335, 9336, 9334, 9340, 9339, 9338, 9344, 9343, 9342, 9347, 9348, 9346, 9338, 9329, 9328, 9342, 9333, 9332, 9337, 9346, 9336, 9334, 9328, 9326, 9346, 9340, 9338, 9329, 9330, 9327, 9341, 9342, 9339, 9339, 9332, 9329, 9336, 9338, 9328, 9352, 9351, 9350, 9356, 9355, 9354, 9359, 9360, 9358, 9364, 9363, 9362, 9368, 9367, 9366, 9371, 9372, 9370, 9362, 9353, 9352, 9366, 9357, 9356, 9361, 9370, 9360, 9358, 9352, 9350, 9370, 9364, 9362, 9353, 9354, 9351, 9365, 9366, 9363, 9363, 9356, 9353, 9360, 9362, 9352, 9350, 9341, 9340, 9354, 9345, 9344, 9349, 9358, 9348, 9351, 9344, 9341,
9348, 9350, 9340, 9376, 9375, 9374, 9383, 9380, 9382, 9386, 9385, 9384, 9393, 9390, 9392, 9384, 9377, 9376, 9391, 9382, 9390, 9382, 9378, 9379, 9392, 9388, 9389, 9377, 9378, 9375, 9387, 9388, 9385, 9385, 9379, 9377, 9390, 9379, 9388, 9396, 9395, 9394, 9403, 9400, 9402, 9406, 9405, 9404, 9413, 9410, 9412, 9404, 9397, 9396, 9411, 9402, 9410, 9402, 9398, 9399, 9412, 9408, 9409, 9397, 9398, 9395, 9407, 9408, 9405, 9405, 9399, 9397, 9410, 9399, 9408, 9394, 9387, 9386, 9401, 9392, 9400, 9395, 9389, 9387, 9400, 9389, 9398, 9326, 9277, 9276, 9330, 9281, 9280, 9285, 9334, 9284, 9327, 9280, 9277, 9284, 9326, 9276, 9374, 9319, 9318, 9381, 9324, 9380, 9375, 9321, 9319, 9380, 9321, 9378, 9245, 9286, 9243, 9257, 9296, 9255, 9255, 9288, 9245, 9269, 9306, 9267, 9281, 9316, 9279, 9279, 9308, 9269, 9267, 9298, 9257, 9333, 9374, 9331, 9345, 9384, 9343, 9343, 9376, 9333, 9357, 9394, 9355, 9369, 9404, 9367, 9367, 9396, 9357, 9355, 9386, 9345, 9331, 9318, 9281, 9417, 9416, 9415, 9422, 9423, 9421, 9425, 9428, 9426, 9430, 9431, 9429, 9425, 9418, 9417, 9424, 9429, 9423, 9421, 9417, 9415,
9425, 9431, 9427, 9422, 9419, 9420, 9419, 9415, 9414, 9417, 9429, 9425, 9435, 9434, 9433, 9439, 9438, 9437, 9444, 9441, 9443, 9447, 9446, 9445, 9451, 9450, 9449, 9456, 9453, 9455, 9445, 9436, 9435, 9449, 9440, 9439, 9454, 9443, 9453, 9443, 9438, 9440, 9455, 9450, 9452, 9436, 9437, 9434, 9448, 9449, 9446, 9446, 9439, 9436, 9453, 9440, 9450, 9459, 9458, 9457, 9462, 9463, 9461, 9467, 9466, 9465, 9470, 9471, 9469, 9465, 9460, 9459, 9464, 9469, 9463, 9461, 9459, 9457, 9469, 9467, 9465, 9463, 9465, 9459, 9475, 9474, 9473, 9479, 9478, 9477, 9484, 9481, 9483, 9487, 9486, 9485, 9491, 9490, 9489, 9496, 9493, 9495, 9485, 9476, 9475, 9489, 9480, 9479, 9494, 9483, 9493, 9483, 9478, 9480, 9495, 9490, 9492, 9476, 9477, 9474, 9488, 9489, 9486, 9486, 9479, 9476, 9493, 9480, 9490, 9457, 9428, 9427, 9432, 9461, 9431, 9431, 9457, 9427, 9473, 9448, 9447, 9477, 9452, 9451, 9482, 9455, 9481, 9474, 9451, 9448, 9481, 9452, 9478, 9418, 9433, 9416, 9428, 9445, 9426, 9426, 9435, 9418, 9460, 9473, 9458, 9468, 9485, 9466, 9466, 9475, 9460, 9458, 9447, 9428, 9499, 9498, 9497, 9502, 9503, 9501,
9507, 9506, 9505, 9510, 9511, 9509, 9505, 9500, 9499, 9504, 9509, 9503, 9501, 9499, 9497, 9509, 9507, 9505, 9503, 9505, 9499, 9515, 9514, 9513, 9519, 9518, 9517, 9524, 9521, 9523, 9527, 9526, 9525, 9531, 9530, 9529, 9536, 9533, 9535, 9525, 9516, 9515, 9529, 9520, 9519, 9534, 9523, 9533, 9523, 9518, 9520, 9535, 9530, 9532, 9516, 9517, 9514, 9528, 9529, 9526, 9526, 9519, 9516, 9533, 9520, 9530, 9539, 9538, 9537, 9542, 9543, 9541, 9547, 9546, 9545, 9550, 9551, 9549, 9545, 9540, 9539, 9544, 9549, 9543, 9541, 9539, 9537, 9549, 9547, 9545, 9543, 9545, 9539, 9555, 9554, 9553, 9559, 9558, 9557, 9564, 9561, 9563, 9567, 9566, 9565, 9571, 9570, 9569, 9576, 9573, 9575, 9565, 9556, 9555, 9569, 9560, 9559, 9574, 9563, 9573, 9563, 9558, 9560, 9575, 9570, 9572, 9556, 9557, 9554, 9568, 9569, 9566, 9566, 9559, 9556, 9573, 9560, 9570, 9537, 9508, 9507, 9512, 9541, 9511, 9511, 9537, 9507, 9553, 9528, 9527, 9557, 9532, 9531, 9562, 9535, 9561, 9554, 9531, 9528, 9561, 9532, 9558, 9500, 9513, 9498, 9508, 9525, 9506, 9506, 9515, 9500, 9540, 9553, 9538, 9548, 9565, 9546, 9546, 9555, 9540,
9538, 9527, 9508, 9497, 9468, 9467, 9472, 9501, 9471, 9471, 9497, 9467, 9513, 9488, 9487, 9517, 9492, 9491, 9522, 9495, 9521, 9514, 9491, 9488, 9521, 9492, 9518, 9498, 9487, 9468, 9579, 9578, 9577, 9583, 9582, 9581, 9586, 9587, 9585, 9591, 9590, 9589, 9595, 9594, 9593, 9598, 9599, 9597, 9589, 9580, 9579, 9593, 9584, 9583, 9588, 9597, 9587, 9585, 9579, 9577, 9597, 9591, 9589, 9580, 9581, 9578, 9592, 9593, 9590, 9590, 9583, 9580, 9587, 9589, 9579, 9608, 9605, 9607, 9612, 9609, 9611, 9606, 9601, 9605, 9610, 9603, 9609, 9611, 9606, 9608, 9609, 9602, 9606, 9601, 9592, 9591, 9603, 9596, 9595, 9600, 9605, 9599, 9602, 9595, 9592, 9599, 9601, 9591, 9615, 9614, 9613, 9622, 9619, 9621, 9625, 9624, 9623, 9632, 9629, 9631, 9623, 9616, 9615, 9630, 9621, 9629, 9621, 9617, 9618, 9631, 9627, 9628, 9616, 9617, 9614, 9626, 9627, 9624, 9624, 9618, 9616, 9629, 9618, 9627, 9639, 9636, 9638, 9637, 9633, 9636, 9641, 9637, 9639, 9640, 9634, 9637, 9633, 9626, 9625, 9641, 9631, 9640, 9634, 9628, 9626, 9640, 9628, 9635, 9647, 9648, 9646, 9651, 9652, 9650, 9656, 9655, 9654, 9660, 9659, 9658,
9654, 9643, 9642, 9658, 9645, 9644, 9643, 9646, 9642, 9645, 9650, 9644, 9650, 9649, 9647, 9657, 9658, 9655, 9644, 9647, 9643, 9655, 9644, 9643, 9667, 9668, 9666, 9671, 9672, 9670, 9676, 9675, 9674, 9680, 9679, 9678, 9674, 9663, 9662, 9678, 9665, 9664, 9663, 9666, 9662, 9665, 9670, 9664, 9670, 9669, 9667, 9677, 9678, 9675, 9664, 9667, 9663, 9675, 9664, 9663, 9684, 9683, 9682, 9688, 9687, 9686, 9697, 9694, 9696, 9701, 9698, 9700, 9684, 9691, 9685, 9692, 9689, 9688, 9695, 9690, 9694, 9699, 9692, 9698, 9685, 9686, 9683, 9700, 9695, 9697, 9698, 9691, 9695, 9691, 9688, 9685, 9704, 9703, 9702, 9708, 9707, 9706, 9717, 9714, 9716, 9721, 9718, 9720, 9710, 9705, 9704, 9712, 9709, 9708, 9715, 9710, 9714, 9719, 9712, 9718, 9705, 9706, 9703, 9720, 9715, 9717, 9718, 9711, 9715, 9711, 9708, 9705, 9682, 9657, 9656, 9686, 9661, 9660, 9683, 9660, 9657, 9702, 9677, 9676, 9706, 9681, 9680, 9703, 9680, 9677, 9666, 9653, 9651, 9661, 9674, 9659, 9662, 9651, 9645, 9659, 9662, 9645, 9689, 9702, 9687, 9716, 9699, 9701, 9714, 9693, 9699, 9693, 9704, 9689, 9687, 9676, 9661, 9727, 9728, 9726,
9731, 9732, 9730, 9736, 9735, 9734, 9740, 9739, 9738, 9734, 9723, 9722, 9738, 9725, 9724, 9723, 9726, 9722, 9725, 9730, 9724, 9730, 9729, 9727, 9737, 9738, 9735, 9724, 9727, 9723, 9735, 9724, 9723, 9747, 9748, 9746, 9751, 9752, 9750, 9756, 9755, 9754, 9760, 9759, 9758, 9754, 9743, 9742, 9758, 9745, 9744, 9743, 9746, 9742, 9745, 9750, 9744, 9750, 9749, 9747, 9757, 9758, 9755, 9744, 9747, 9743, 9755, 9744, 9743, 9764, 9763, 9762, 9768, 9767, 9766, 9777, 9774, 9776, 9781, 9778, 9780, 9770, 9765, 9764, 9772, 9769, 9768, 9775, 9770, 9774, 9779, 9772, 9778, 9765, 9766, 9763, 9780, 9775, 9777, 9778, 9771, 9775, 9771, 9768, 9765, 9784, 9783, 9782, 9788, 9787, 9786, 9797, 9794, 9796, 9801, 9798, 9800, 9790, 9785, 9784, 9792, 9789, 9788, 9795, 9790, 9794, 9799, 9792, 9798, 9785, 9786, 9783, 9800, 9795, 9797, 9798, 9791, 9795, 9791, 9788, 9785, 9762, 9737, 9736, 9766, 9741, 9740, 9763, 9740, 9737, 9782, 9757, 9756, 9786, 9761, 9760, 9783, 9760, 9757, 9746, 9733, 9731, 9741, 9754, 9739, 9742, 9731, 9725, 9739, 9742, 9725, 9769, 9782, 9767, 9796, 9779, 9781, 9794, 9773, 9779,
9773, 9784, 9769, 9767, 9756, 9741, 9584, 9613, 9582, 9596, 9623, 9594, 9594, 9615, 9584, 9638, 9610, 9612, 9636, 9604, 9610, 9604, 9625, 9596, 9726, 9673, 9671, 9681, 9734, 9679, 9722, 9671, 9665, 9679, 9722, 9665, 9709, 9762, 9707, 9776, 9719, 9721, 9774, 9713, 9719, 9713, 9764, 9709, 9707, 9736, 9681, 9804, 9803, 9802, 9807, 9808, 9806, 9812, 9811, 9810, 9815, 9816, 9814, 9810, 9805, 9804, 9809, 9814, 9808, 9806, 9804, 9802, 9814, 9812, 9810, 9808, 9810, 9804, 9820, 9819, 9818, 9824, 9823, 9822, 9829, 9826, 9828, 9832, 9831, 9830, 9836, 9835, 9834, 9841, 9838, 9840, 9830, 9821, 9820, 9834, 9825, 9824, 9839, 9828, 9838, 9828, 9823, 9825, 9840, 9835, 9837, 9821, 9822, 9819, 9833, 9834, 9831, 9831, 9824, 9821, 9838, 9825, 9835, 9844, 9843, 9842, 9851, 9848, 9850, 9847, 9848, 9846, 9849, 9844, 9848, 9846, 9844, 9842, 9854, 9853, 9852, 9858, 9857, 9856, 9863, 9860, 9862, 9869, 9866, 9868, 9869, 9864, 9867, 9861, 9854, 9860, 9867, 9858, 9866, 9867, 9857, 9859, 9855, 9856, 9853, 9868, 9861, 9863, 9866, 9855, 9861, 9842, 9813, 9812, 9817, 9846, 9816, 9816, 9842, 9812,
9852, 9833, 9832, 9856, 9837, 9836, 9865, 9840, 9864, 9853, 9836, 9833, 9864, 9837, 9857, 9805, 9818, 9803, 9813, 9830, 9811, 9811, 9820, 9805, 9845, 9852, 9843, 9862, 9849, 9851, 9860, 9845, 9849, 9843, 9832, 9813, 9875, 9876, 9874, 9879, 9880, 9878, 9884, 9883, 9882, 9888, 9887, 9886, 9882, 9871, 9870, 9886, 9873, 9872, 9871, 9874, 9870, 9873, 9878, 9872, 9878, 9877, 9875, 9885, 9886, 9883, 9872, 9875, 9871, 9883, 9872, 9871, 9895, 9896, 9894, 9899, 9900, 9898, 9904, 9903, 9902, 9908, 9907, 9906, 9913, 9910, 9912, 9902, 9891, 9890, 9906, 9893, 9892, 9911, 9899, 9910, 9891, 9894, 9890, 9893, 9898, 9892, 9912, 9907, 9909, 9898, 9897, 9895, 9905, 9906, 9903, 9892, 9895, 9891, 9903, 9892, 9891, 9910, 9893, 9907, 9916, 9915, 9914, 9920, 9919, 9918, 9929, 9926, 9928, 9933, 9930, 9932, 9922, 9917, 9916, 9924, 9921, 9920, 9927, 9922, 9926, 9931, 9924, 9930, 9917, 9918, 9915, 9932, 9927, 9929, 9930, 9923, 9927, 9923, 9920, 9917, 9936, 9935, 9934, 9940, 9939, 9938, 9945, 9942, 9944, 9953, 9950, 9952, 9957, 9954, 9956, 9946, 9937, 9936, 9948, 9941, 9940, 9957, 9944, 9955,
9951, 9946, 9950, 9944, 9939, 9941, 9955, 9948, 9954, 9937, 9938, 9935, 9956, 9951, 9953, 9954, 9947, 9951, 9947, 9940, 9937, 9955, 9941, 9949, 9914, 9885, 9884, 9918, 9889, 9888, 9915, 9888, 9885, 9934, 9905, 9904, 9938, 9909, 9908, 9943, 9912, 9942, 9935, 9908, 9905, 9942, 9909, 9939, 9894, 9881, 9879, 9889, 9902, 9887, 9890, 9879, 9873, 9887, 9890, 9873, 9921, 9934, 9919, 9952, 9931, 9933, 9950, 9925, 9931, 9925, 9936, 9921, 9919, 9904, 9889, 9577, 9365, 9364, 9581, 9369, 9368, 9373, 9585, 9372, 9578, 9368, 9365, 9372, 9577, 9364, 9613, 9407, 9406, 9620, 9412, 9619, 9614, 9409, 9407, 9619, 9409, 9617, 9582, 9406, 9369, 9802, 9548, 9547, 9552, 9806, 9551, 9551, 9802, 9547, 9818, 9568, 9567, 9822, 9572, 9571, 9827, 9575, 9826, 9819, 9571, 9568, 9826, 9572, 9823, 9803, 9567, 9548, 9874, 9753, 9751, 9761, 9882, 9759, 9870, 9751, 9745, 9759, 9870, 9745, 9789, 9914, 9787, 9928, 9799, 9801, 9926, 9793, 9799, 9793, 9916, 9789, 9787, 9884, 9761, 9238, 9058, 9057, 9242, 9062, 9061, 9066, 9246, 9065, 9239, 9061, 9058, 9065, 9238, 9057, 9286, 9118, 9117, 9293, 9126, 9292,
9287, 9121, 9118, 9292, 9121, 9290, 9243, 9117, 9062, 9206, 9416, 9207, 9213, 9419, 9211, 9414, 9206, 9203, 9211, 9414, 9203, 9433, 9229, 9228, 9437, 9233, 9232, 9442, 9236, 9441, 9434, 9232, 9229, 9441, 9233, 9438, 9416, 9228, 9207, 9963, 9964, 9962, 9967, 9968, 9966, 9959, 9962, 9958, 9961, 9966, 9960, 9966, 9965, 9963, 9960, 9963, 9959, 9975, 9976, 9974, 9979, 9980, 9978, 9971, 9974, 9970, 9973, 9978, 9972, 9978, 9977, 9975, 9972, 9975, 9971, 9984, 9983, 9982, 9988, 9987, 9986, 9992, 9991, 9990, 9996, 9995, 9994, 9990, 9985, 9984, 9994, 9989, 9988, 9985, 9986, 9983, 9993, 9994, 9991, 9991, 9988, 9985, 10000, 9999, 9998, 10004, 10003, 10002, 10008, 10007, 10006, 10012, 10011, 10010, 10006, 10001, 10000, 10010, 10005, 10004, 10001, 10002, 9999, 10009, 10010, 10007, 10007, 10004, 10001, 9982, 9959, 9958, 9986, 9961, 9960, 9983, 9960, 9959, 9998, 9971, 9970, 10002, 9973, 9972, 9999, 9972, 9971, 9974, 9969, 9967, 9970, 9967, 9961, 9989, 9998, 9987, 9997, 10006, 9995, 9995, 10000, 9989, 9987, 9970, 9961, 10019, 10020, 10018, 10023, 10024, 10022, 10015, 10018, 10014, 10017, 10022, 10016, 10022, 10021, 10019,
10016, 10019, 10015, 10031, 10032, 10030, 10035, 10036, 10034, 10027, 10030, 10026, 10029, 10034, 10028, 10034, 10033, 10031, 10028, 10031, 10027, 10040, 10039, 10038, 10044, 10043, 10042, 10048, 10047, 10046, 10052, 10051, 10050, 10046, 10041, 10040, 10050, 10045, 10044, 10041, 10042, 10039, 10049, 10050, 10047, 10047, 10044, 10041, 10056, 10055, 10054, 10060, 10059, 10058, 10064, 10063, 10062, 10068, 10067, 10066, 10062, 10057, 10056, 10066, 10061, 10060, 10057, 10058, 10055, 10065, 10066, 10063, 10063, 10060, 10057, 10038, 10015, 10014, 10042, 10017, 10016, 10039, 10016, 10015, 10054, 10027, 10026, 10058, 10029, 10028, 10055, 10028, 10027, 10030, 10025, 10023, 10026, 10023, 10017, 10045, 10054, 10043, 10053, 10062, 10051, 10051, 10056, 10045, 10043, 10026, 10017, 10018, 9981, 9979, 10014, 9979, 9973, 10005, 10038, 10003, 10013, 10046, 10011, 10011, 10040, 10005, 10003, 10014, 9973, 10075, 10076, 10074, 10079, 10080, 10078, 10071, 10074, 10070, 10073, 10078, 10072, 10078, 10077, 10075, 10072, 10075, 10071, 10087, 10088, 10086, 10091, 10092, 10090, 10083, 10086, 10082, 10085, 10090, 10084, 10090, 10089, 10087, 10084, 10087, 10083, 10096, 10095, 10094, 10100, 10099, 10098, 10104, 10103, 10102, 10108, 10107, 10106, 10102, 10097, 10096,
10106, 10101, 10100, 10097, 10098, 10095, 10105, 10106, 10103, 10103, 10100, 10097, 10112, 10111, 10110, 10116, 10115, 10114, 10121, 10118, 10120, 10124, 10123, 10122, 10128, 10127, 10126, 10133, 10130, 10132, 10122, 10113, 10112, 10126, 10117, 10116, 10131, 10120, 10130, 10120, 10115, 10117, 10132, 10127, 10129, 10113, 10114, 10111, 10125, 10126, 10123, 10123, 10116, 10113, 10130, 10117, 10127, 10094, 10071, 10070, 10098, 10073, 10072, 10095, 10072, 10071, 10110, 10083, 10082, 10114, 10085, 10084, 10119, 10091, 10118, 10111, 10084, 10083, 10118, 10085, 10115, 10086, 10081, 10079, 10082, 10079, 10073, 10101, 10110, 10099, 10109, 10122, 10107, 10107, 10112, 10101, 10099, 10082, 10073, 10074, 10037, 10035, 10070, 10035, 10029, 10061, 10094, 10059, 10069, 10102, 10067, 10067, 10096, 10061, 10059, 10070, 10029, 10141, 10138, 10140, 10145, 10142, 10144, 10139, 10134, 10138, 10143, 10136, 10142, 10144, 10139, 10141, 10142, 10135, 10139, 10153, 10150, 10152, 10157, 10154, 10156, 10151, 10146, 10150, 10155, 10148, 10154, 10156, 10151, 10153, 10154, 10147, 10151, 10152, 10143, 10145, 10150, 10137, 10143, 10165, 10162, 10164, 10169, 10166, 10168, 10163, 10158, 10162, 10167, 10160, 10166, 10168, 10163, 10165, 10166, 10159, 10163, 10177, 10174, 10176,
10181, 10178, 10180, 10175, 10170, 10174, 10179, 10172, 10178, 10180, 10175, 10177, 10178, 10171, 10175, 10176, 10167, 10169, 10174, 10161, 10167, 10184, 10183, 10182, 10188, 10187, 10186, 10191, 10192, 10190, 10195, 10196, 10194, 10200, 10199, 10198, 10204, 10203, 10202, 10198, 10185, 10184, 10202, 10189, 10188, 10183, 10190, 10182, 10187, 10194, 10186, 10185, 10186, 10183, 10194, 10193, 10191, 10201, 10202, 10199, 10186, 10191, 10183, 10199, 10188, 10185, 10208, 10207, 10206, 10212, 10211, 10210, 10215, 10216, 10214, 10219, 10220, 10218, 10224, 10223, 10222, 10228, 10227, 10226, 10222, 10209, 10208, 10226, 10213, 10212, 10207, 10214, 10206, 10211, 10218, 10210, 10209, 10210, 10207, 10218, 10217, 10215, 10225, 10226, 10223, 10210, 10215, 10207, 10223, 10212, 10209, 10230, 10233, 10231, 10236, 10235, 10234, 10242, 10241, 10240, 10246, 10245, 10244, 10249, 10250, 10248, 10240, 10233, 10232, 10244, 10237, 10236, 10239, 10248, 10238, 10248, 10242, 10240, 10233, 10234, 10231, 10243, 10244, 10241, 10241, 10236, 10233, 10238, 10240, 10232, 10254, 10253, 10252, 10258, 10257, 10256, 10265, 10262, 10264, 10268, 10267, 10266, 10273, 10270, 10272, 10254, 10267, 10255, 10271, 10260, 10270, 10272, 10267, 10269, 10263, 10258, 10262, 10255, 10256, 10253,
10264, 10260, 10261, 10262, 10255, 10260, 10270, 10255, 10267, 10230, 10201, 10200, 10234, 10205, 10204, 10231, 10204, 10201, 10252, 10225, 10224, 10256, 10229, 10228, 10253, 10228, 10225, 10189, 10206, 10187, 10214, 10197, 10195, 10205, 10222, 10203, 10206, 10195, 10187, 10203, 10208, 10189, 10237, 10252, 10235, 10247, 10266, 10245, 10245, 10254, 10237, 10235, 10224, 10205, 10276, 10275, 10274, 10280, 10279, 10278, 10283, 10284, 10282, 10287, 10288, 10286, 10292, 10291, 10290, 10296, 10295, 10294, 10290, 10277, 10276, 10294, 10281, 10280, 10275, 10282, 10274, 10279, 10286, 10278, 10277, 10278, 10275, 10286, 10285, 10283, 10293, 10294, 10291, 10278, 10283, 10275, 10291, 10280, 10277, 10300, 10299, 10298, 10304, 10303, 10302, 10307, 10308, 10306, 10311, 10312, 10310, 10316, 10315, 10314, 10320, 10319, 10318, 10314, 10301, 10300, 10318, 10305, 10304, 10299, 10306, 10298, 10303, 10310, 10302, 10301, 10302, 10299, 10310, 10309, 10307, 10317, 10318, 10315, 10302, 10307, 10299, 10315, 10304, 10301, 10324, 10323, 10322, 10328, 10327, 10326, 10333, 10330, 10332, 10337, 10334, 10336, 10331, 10324, 10330, 10335, 10328, 10334, 10325, 10326, 10323, 10336, 10331, 10333, 10334, 10325, 10331, 10340, 10339, 10338, 10344, 10343, 10342, 10349, 10346, 10348,
10353, 10350, 10352, 10347, 10340, 10346, 10351, 10344, 10350, 10341, 10342, 10339, 10352, 10347, 10349, 10350, 10341, 10347, 10322, 10293, 10292, 10326, 10297, 10296, 10323, 10296, 10293, 10338, 10317, 10316, 10342, 10321, 10320, 10339, 10320, 10317, 10281, 10298, 10279, 10306, 10289, 10287, 10297, 10314, 10295, 10298, 10287, 10279, 10295, 10300, 10281, 10329, 10338, 10327, 10348, 10335, 10337, 10346, 10329, 10335, 10327, 10316, 10297, 10164, 10155, 10157, 10162, 10149, 10155, 10213, 10274, 10211, 10282, 10221, 10219, 10229, 10290, 10227, 10274, 10219, 10211, 10227, 10276, 10213, 10259, 10322, 10257, 10332, 10263, 10265, 10330, 10259, 10263, 10257, 10292, 10229, 10361, 10358, 10360, 10365, 10362, 10364, 10359, 10354, 10358, 10363, 10356, 10362, 10364, 10359, 10361, 10362, 10355, 10359, 10373, 10370, 10372, 10377, 10374, 10376, 10371, 10366, 10370, 10375, 10368, 10374, 10376, 10371, 10373, 10374, 10367, 10371, 10372, 10363, 10365, 10370, 10357, 10363, 10380, 10379, 10378, 10384, 10383, 10382, 10387, 10388, 10386, 10391, 10392, 10390, 10396, 10395, 10394, 10400, 10399, 10398, 10394, 10381, 10380, 10398, 10385, 10384, 10379, 10386, 10378, 10383, 10390, 10382, 10381, 10382, 10379, 10390, 10389, 10387, 10397, 10398, 10395, 10382, 10387, 10379,
10395, 10384, 10381, 10404, 10403, 10402, 10408, 10407, 10406, 10411, 10412, 10410, 10415, 10416, 10414, 10419, 10415, 10418, 10422, 10421, 10420, 10426, 10425, 10424, 10431, 10428, 10430, 10420, 10405, 10404, 10424, 10409, 10408, 10429, 10418, 10428, 10403, 10410, 10402, 10407, 10414, 10406, 10418, 10407, 10409, 10430, 10425, 10427, 10405, 10406, 10403, 10414, 10413, 10411, 10423, 10424, 10421, 10406, 10411, 10403, 10421, 10408, 10405, 10428, 10409, 10425, 10434, 10433, 10432, 10436, 10439, 10437, 10443, 10440, 10442, 10448, 10447, 10446, 10451, 10452, 10450, 10446, 10439, 10438, 10445, 10450, 10444, 10441, 10434, 10440, 10450, 10448, 10446, 10433, 10438, 10436, 10445, 10441, 10443, 10444, 10435, 10441, 10444, 10446, 10438, 10456, 10455, 10454, 10460, 10459, 10458, 10465, 10462, 10464, 10468, 10467, 10466, 10472, 10471, 10470, 10477, 10474, 10476, 10466, 10457, 10456, 10470, 10461, 10460, 10475, 10464, 10474, 10464, 10459, 10461, 10476, 10471, 10473, 10457, 10458, 10455, 10469, 10470, 10467, 10467, 10460, 10457, 10474, 10461, 10471, 10432, 10397, 10396, 10436, 10401, 10400, 10433, 10400, 10397, 10454, 10423, 10422, 10458, 10427, 10426, 10463, 10430, 10462, 10455, 10426, 10423, 10462, 10427, 10459, 10385, 10402, 10383, 10410, 10393, 10391,
10401, 10420, 10399, 10402, 10391, 10383, 10399, 10404, 10385, 10439, 10454, 10437, 10449, 10466, 10447, 10447, 10456, 10439, 10437, 10422, 10401, 10480, 10479, 10478, 10484, 10483, 10482, 10487, 10488, 10486, 10492, 10491, 10490, 10496, 10495, 10494, 10499, 10500, 10498, 10490, 10481, 10480, 10494, 10485, 10484, 10489, 10498, 10488, 10486, 10480, 10478, 10498, 10492, 10490, 10481, 10482, 10479, 10493, 10494, 10491, 10491, 10484, 10481, 10488, 10490, 10480, 10504, 10503, 10502, 10509, 10506, 10508, 10512, 10511, 10510, 10517, 10514, 10516, 10510, 10505, 10504, 10515, 10508, 10514, 10508, 10503, 10505, 10516, 10511, 10513, 10514, 10505, 10511, 10520, 10519, 10518, 10524, 10523, 10522, 10527, 10528, 10526, 10532, 10531, 10530, 10536, 10535, 10534, 10539, 10540, 10538, 10530, 10521, 10520, 10534, 10525, 10524, 10529, 10538, 10528, 10526, 10520, 10518, 10538, 10532, 10530, 10521, 10522, 10519, 10533, 10534, 10531, 10531, 10524, 10521, 10528, 10530, 10520, 10544, 10543, 10542, 10549, 10546, 10548, 10552, 10551, 10550, 10557, 10554, 10556, 10550, 10545, 10544, 10555, 10548, 10554, 10548, 10543, 10545, 10556, 10551, 10553, 10554, 10545, 10551, 10518, 10493, 10492, 10522, 10497, 10496, 10501, 10526, 10500, 10519, 10496, 10493, 10500, 10518, 10492,
10542, 10513, 10512, 10547, 10516, 10546, 10546, 10513, 10543, 10485, 10502, 10483, 10497, 10510, 10495, 10495, 10504, 10485, 10525, 10542, 10523, 10537, 10550, 10535, 10535, 10544, 10525, 10523, 10512, 10497, 10560, 10559, 10558, 10564, 10563, 10562, 10567, 10568, 10566, 10572, 10571, 10570, 10576, 10575, 10574, 10579, 10580, 10578, 10570, 10561, 10560, 10574, 10565, 10564, 10569, 10578, 10568, 10566, 10560, 10558, 10578, 10572, 10570, 10561, 10562, 10559, 10573, 10574, 10571, 10571, 10564, 10561, 10568, 10570, 10560, 10584, 10583, 10582, 10589, 10586, 10588, 10592, 10591, 10590, 10597, 10594, 10596, 10590, 10585, 10584, 10595, 10588, 10594, 10588, 10583, 10585, 10596, 10591, 10593, 10594, 10585, 10591, 10600, 10599, 10598, 10604, 10603, 10602, 10607, 10608, 10606, 10610, 10613, 10611, 10616, 10615, 10614, 10619, 10620, 10618, 10610, 10601, 10600, 10614, 10605, 10604, 10609, 10618, 10608, 10606, 10600, 10598, 10610, 10620, 10612, 10601, 10602, 10599, 10613, 10614, 10611, 10611, 10604, 10601, 10608, 10610, 10600, 10622, 10625, 10623, 10629, 10626, 10628, 10630, 10633, 10631, 10637, 10634, 10636, 10630, 10625, 10624, 10635, 10628, 10634, 10625, 10626, 10623, 10633, 10634, 10631, 10631, 10628, 10625, 10598, 10573, 10572, 10602, 10577, 10576,
10581, 10606, 10580, 10599, 10576, 10573, 10580, 10598, 10572, 10622, 10593, 10592, 10627, 10596, 10626, 10626, 10593, 10623, 10565, 10582, 10563, 10577, 10590, 10575, 10575, 10584, 10565, 10605, 10622, 10603, 10617, 10630, 10615, 10615, 10624, 10605, 10603, 10592, 10577, 10558, 10533, 10532, 10562, 10537, 10536, 10541, 10566, 10540, 10559, 10536, 10533, 10540, 10558, 10532, 10582, 10553, 10552, 10587, 10556, 10586, 10586, 10553, 10583, 10563, 10552, 10537, 10640, 10639, 10638, 10643, 10644, 10642, 10648, 10647, 10646, 10651, 10652, 10650, 10646, 10641, 10640, 10645, 10650, 10644, 10642, 10640, 10638, 10650, 10648, 10646, 10644, 10646, 10640, 10656, 10655, 10654, 10660, 10659, 10658, 10665, 10662, 10664, 10668, 10667, 10666, 10672, 10671, 10670, 10677, 10674, 10676, 10666, 10657, 10656, 10670, 10661, 10660, 10675, 10664, 10674, 10664, 10659, 10661, 10676, 10671, 10673, 10657, 10658, 10655, 10669, 10670, 10667, 10667, 10660, 10657, 10674, 10661, 10671, 10680, 10679, 10678, 10683, 10684, 10682, 10688, 10687, 10686, 10691, 10692, 10690, 10686, 10681, 10680, 10685, 10690, 10684, 10682, 10680, 10678, 10690, 10688, 10686, 10684, 10686, 10680, 10696, 10695, 10694, 10700, 10699, 10698, 10705, 10702, 10704, 10708, 10707, 10706, 10712, 10711, 10710,
10717, 10714, 10716, 10706, 10697, 10696, 10710, 10701, 10700, 10715, 10704, 10714, 10704, 10699, 10701, 10716, 10711, 10713, 10697, 10698, 10695, 10709, 10710, 10707, 10707, 10700, 10697, 10714, 10701, 10711, 10678, 10649, 10648, 10653, 10682, 10652, 10652, 10678, 10648, 10694, 10669, 10668, 10698, 10673, 10672, 10703, 10676, 10702, 10695, 10672, 10669, 10702, 10673, 10699, 10641, 10654, 10639, 10649, 10666, 10647, 10647, 10656, 10641, 10681, 10694, 10679, 10689, 10706, 10687, 10687, 10696, 10681, 10679, 10668, 10649, 10720, 10719, 10718, 10723, 10724, 10722, 10728, 10727, 10726, 10731, 10732, 10730, 10726, 10721, 10720, 10725, 10730, 10724, 10722, 10720, 10718, 10730, 10728, 10726, 10724, 10726, 10720, 10736, 10735, 10734, 10740, 10739, 10738, 10745, 10742, 10744, 10748, 10747, 10746, 10752, 10751, 10750, 10757, 10754, 10756, 10746, 10737, 10736, 10750, 10741, 10740, 10755, 10744, 10754, 10744, 10739, 10741, 10756, 10751, 10753, 10737, 10738, 10735, 10749, 10750, 10747, 10747, 10740, 10737, 10754, 10741, 10751, 10760, 10759, 10758, 10763, 10764, 10762, 10768, 10767, 10766, 10771, 10772, 10770, 10766, 10761, 10760, 10765, 10770, 10764, 10762, 10760, 10758, 10770, 10768, 10766, 10764, 10766, 10760, 10776, 10775, 10774, 10780, 10779, 10778,
10785, 10782, 10784, 10788, 10787, 10786, 10790, 10793, 10791, 10797, 10794, 10796, 10786, 10777, 10776, 10790, 10781, 10780, 10795, 10784, 10794, 10784, 10779, 10781, 10793, 10794, 10791, 10777, 10778, 10775, 10789, 10790, 10787, 10787, 10780, 10777, 10794, 10781, 10791, 10758, 10729, 10728, 10733, 10762, 10732, 10732, 10758, 10728, 10774, 10749, 10748, 10778, 10753, 10752, 10783, 10756, 10782, 10775, 10752, 10749, 10782, 10753, 10779, 10721, 10734, 10719, 10729, 10746, 10727, 10727, 10736, 10721, 10761, 10774, 10759, 10769, 10786, 10767, 10767, 10776, 10761, 10759, 10748, 10729, 10718, 10689, 10688, 10693, 10722, 10692, 10692, 10718, 10688, 10734, 10709, 10708, 10738, 10713, 10712, 10743, 10716, 10742, 10735, 10712, 10709, 10742, 10713, 10739, 10719, 10708, 10689, 10478, 10243, 10242, 10482, 10247, 10246, 10251, 10486, 10250, 10479, 10246, 10243, 10250, 10478, 10242, 10502, 10269, 10268, 10507, 10272, 10506, 10506, 10269, 10503, 10483, 10268, 10247, 10638, 10449, 10448, 10453, 10642, 10452, 10452, 10638, 10448, 10654, 10469, 10468, 10658, 10473, 10472, 10663, 10476, 10662, 10655, 10472, 10469, 10662, 10473, 10659, 10639, 10468, 10449, 10360, 10179, 10181, 10358, 10173, 10179, 10305, 10378, 10303, 10386, 10313, 10311, 10321, 10394, 10319,
10378, 10311, 10303, 10319, 10380, 10305, 10345, 10432, 10343, 10442, 10351, 10353, 10440, 10345, 10351, 10343, 10396, 10321, 10134, 9993, 9992, 10136, 9997, 9996, 10135, 9996, 9993, 10146, 10009, 10008, 10148, 10013, 10012, 10147, 10012, 10009, 10137, 10008, 9997, 10158, 10049, 10048, 10160, 10053, 10052, 10159, 10052, 10049, 10170, 10065, 10064, 10172, 10069, 10068, 10171, 10068, 10065, 10161, 10064, 10053, 10149, 10048, 10013, 10354, 10105, 10104, 10356, 10109, 10108, 10355, 10108, 10105, 10366, 10125, 10124, 10368, 10129, 10128, 10377, 10132, 10375, 10367, 10128, 10125, 10375, 10129, 10369, 10357, 10124, 10109, 10173, 10104, 10069, 8769, 8396, 8395, 8399, 8774, 8400, 8404, 8777, 8403, 8782, 8407, 8781, 8770, 8399, 8396, 8403, 8769, 8395, 8781, 8400, 8774, 8803, 8421, 8420, 8423, 8807, 8422, 8417, 8803, 8420, 8417, 8807, 8801, 8984, 8614, 8983, 8983, 8609, 8981, 8421, 8981, 8609, 10804, 10805, 10803, 10802, 10804, 10801, 10806, 10808, 10805, 10802, 10809, 10806, 10811, 10812, 10810, 10815, 10816, 10814, 10816, 10818, 10814, 10825, 10822, 10823, 10833, 10830, 10832, 10833, 10828, 10831, 10835, 10836, 10834, 10841, 10838, 10839, 10812, 10826, 10810, 10835, 10820, 10821, 10825, 10838, 10824,
10812, 10821, 10820, 10817, 10822, 10816, 10816, 10824, 10819, 10840, 10833, 10832, 10841, 10829, 10833, 10818, 10813, 10811, 10838, 10837, 10835, 10813, 10824, 10821, 10838, 10821, 10824, 10812, 10834, 10827, 10799, 10815, 10798, 10800, 10817, 10799, 10826, 10802, 10801, 10836, 10807, 10834, 10827, 10807, 10802, 10845, 10846, 10844, 10843, 10845, 10842, 10851, 10848, 10850, 10846, 10854, 10853, 10843, 10854, 10847, 10851, 10855, 10849, 10848, 10843, 10842, 10856, 10852, 10855, 10849, 10852, 10843, 10860, 10861, 10859, 10861, 10863, 10859, 10871, 10868, 10869, 10861, 10869, 10868, 10861, 10870, 10865, 10863, 10858, 10857, 10865, 10867, 10858, 10877, 10874, 10876, 10877, 10872, 10875, 10885, 10882, 10883, 10893, 10894, 10892, 10894, 10896, 10892, 10903, 10900, 10901, 10889, 10874, 10888, 10881, 10893, 10880, 10884, 10877, 10876, 10885, 10873, 10877, 10894, 10901, 10900, 10900, 10897, 10894, 10879, 10872, 10873, 10896, 10891, 10890, 10883, 10879, 10873, 10891, 10902, 10899, 10898, 10876, 10889, 10887, 10895, 10881, 10905, 10906, 10904, 10906, 10908, 10904, 10913, 10910, 10912, 10915, 10910, 10914, 10919, 10916, 10917, 10926, 10923, 10925, 10932, 10929, 10930, 10928, 10914, 10927, 10919, 10929, 10918, 10907, 10916, 10906, 10906, 10918, 10909,
10926, 10930, 10924, 10920, 10913, 10912, 10915, 10920, 10912, 10920, 10919, 10917, 10928, 10923, 10924, 10933, 10924, 10930, 10933, 10919, 10922, 10928, 10922, 10915, 10879, 10850, 10878, 10883, 10853, 10854, 10886, 10854, 10856, 10879, 10856, 10851, 10858, 10905, 10857, 10913, 10864, 10911, 10871, 10920, 10870, 10870, 10917, 10867, 10858, 10917, 10907, 10921, 10866, 10913, 10908, 10881, 10880, 10929, 10903, 10901, 10909, 10887, 10881, 10929, 10887, 10918, 10935, 10936, 10934, 10943, 10940, 10942, 10945, 10940, 10944, 10936, 10940, 10934, 10943, 10935, 10941, 10939, 10946, 10938, 10937, 10938, 10936, 10947, 10942, 10946, 10948, 10942, 10945, 10938, 10942, 10936, 10947, 10937, 10943, 10940, 10862, 10860, 10945, 10864, 10866, 10948, 10869, 10946, 10862, 10946, 10869, 10871, 10945, 10866, 10844, 10799, 10798, 10799, 10853, 10800, 10882, 10825, 10823, 10889, 10828, 10829, 10839, 10889, 10829, 10898, 10825, 10884, 10882, 10800, 10853, 10952, 10953, 10951, 10950, 10952, 10949, 10961, 10962, 10960, 10959, 10961, 10958, 10970, 10971, 10969, 10953, 10964, 10951, 10970, 10956, 10957, 10963, 10967, 10962, 10959, 10968, 10963, 10954, 10956, 10953, 10950, 10957, 10954, 10964, 10959, 10958, 10971, 10966, 10969, 10965, 10966, 10959, 10953, 10969, 10965,
10977, 10978, 10976, 10978, 10980, 10976, 10985, 10982, 10983, 10974, 10977, 10973, 10978, 10983, 10982, 10978, 10984, 10981, 10975, 10979, 10974, 10989, 10990, 10988, 10987, 10989, 10986, 10998, 10999, 10997, 10996, 10998, 10995, 11007, 11008, 11006, 10990, 11001, 10988, 11007, 10993, 10994, 11000, 11004, 10999, 10996, 11005, 11000, 10990, 10994, 10993, 10992, 10991, 10987, 11001, 10996, 10995, 11008, 11003, 11006, 10996, 11006, 11003, 10993, 11002, 10990, 11013, 11014, 11012, 11019, 11016, 11017, 11023, 11020, 11021, 11027, 11024, 11025, 11021, 11014, 11015, 11025, 11018, 11019, 11018, 11013, 11016, 11026, 11021, 11024, 11024, 11015, 11018, 11029, 11030, 11028, 11035, 11032, 11033, 11039, 11036, 11037, 11043, 11040, 11041, 11047, 11044, 11045, 11035, 11040, 11034, 11045, 11038, 11039, 11029, 11034, 11031, 11038, 11033, 11036, 11046, 11041, 11044, 11044, 11035, 11038, 11051, 11048, 11049, 11055, 11052, 11053, 11059, 11056, 11057, 11063, 11060, 11061, 11057, 11050, 11051, 11061, 11054, 11055, 11054, 11049, 11052, 11062, 11057, 11060, 11060, 11051, 11054, 10999, 11028, 10997, 11033, 11004, 11005, 11037, 11008, 11009, 11008, 11033, 11005, 10999, 11032, 11029, 11049, 11022, 11023, 11053, 11026, 11027, 11052, 11023, 11026, 11010, 10987, 10986,
11014, 10992, 11012, 11022, 11007, 11020, 10987, 11012, 10992, 11014, 11007, 10994, 11050, 11037, 11048, 11058, 11045, 11056, 11056, 11039, 11050, 11048, 11009, 11022, 11070, 11071, 11069, 11068, 11070, 11067, 11079, 11080, 11078, 11083, 11084, 11082, 11089, 11086, 11087, 11093, 11090, 11091, 11071, 11082, 11069, 11087, 11076, 11077, 11091, 11080, 11081, 11072, 11076, 11071, 11075, 11072, 11068, 11083, 11088, 11085, 11073, 11068, 11067, 11080, 11075, 11078, 11092, 11087, 11090, 11068, 11078, 11075, 11080, 11087, 11077, 11071, 11086, 11083, 11095, 11096, 11094, 11101, 11102, 11100, 11102, 11104, 11100, 11109, 11106, 11107, 11097, 11101, 11096, 11095, 11099, 11097, 11102, 11107, 11106, 11106, 11105, 11102, 11099, 11103, 11097, 11113, 11110, 11111, 11117, 11114, 11115, 11121, 11118, 11119, 11125, 11122, 11123, 11119, 11112, 11113, 11123, 11116, 11117, 11116, 11111, 11114, 11124, 11119, 11122, 11122, 11113, 11116, 11065, 11094, 11064, 11065, 11098, 11095, 11089, 11110, 11088, 11115, 11092, 11093, 11114, 11089, 11092, 11084, 11065, 11064, 11085, 11066, 11065, 11112, 11098, 11110, 11118, 11109, 11107, 11118, 11099, 11112, 11110, 11066, 11088, 11129, 11126, 11127, 11133, 11130, 11131, 11137, 11134, 11135, 11141, 11138, 11139, 11135, 11128, 11129,
11139, 11132, 11133, 11132, 11127, 11130, 11140, 11135, 11138, 11138, 11129, 11132, 11147, 11144, 11146, 11147, 11142, 11145, 11151, 11148, 11149, 11155, 11152, 11153, 11161, 11158, 11160, 11161, 11156, 11159, 11167, 11164, 11165, 11163, 11144, 11162, 11151, 11164, 11150, 11166, 11161, 11160, 11167, 11157, 11161, 11155, 11146, 11154, 11155, 11143, 11147, 11152, 11151, 11149, 11163, 11156, 11157, 11165, 11163, 11157, 11168, 11151, 11154, 11163, 11154, 11146, 11174, 11171, 11172, 11176, 11177, 11175, 11182, 11179, 11181, 11184, 11179, 11183, 11188, 11185, 11186, 11182, 11169, 11180, 11186, 11173, 11174, 11190, 11177, 11178, 11189, 11182, 11181, 11184, 11189, 11181, 11177, 11172, 11175, 11191, 11186, 11189, 11177, 11186, 11174, 11190, 11170, 11182, 11195, 11192, 11194, 11195, 11196, 11193, 11172, 11136, 11137, 11176, 11140, 11141, 11175, 11137, 11140, 11193, 11158, 11192, 11193, 11166, 11160, 11150, 11131, 11148, 11164, 11141, 11139, 11164, 11133, 11150, 11195, 11169, 11170, 11197, 11176, 11196, 11197, 11170, 11178, 11196, 11141, 11166, 10981, 11073, 10980, 11079, 10984, 10985, 10981, 11078, 11074, 11127, 11042, 11043, 11131, 11046, 11047, 11130, 11043, 11046, 11149, 11058, 11059, 11063, 11152, 11062, 11152, 11059, 11062, 11148, 11047, 11058,
11030, 10974, 10973, 11040, 10985, 10983, 10974, 11034, 10975, 11040, 10975, 11034, 11128, 11079, 11126, 11136, 11091, 11134, 11134, 11081, 11128, 11173, 11115, 11171, 11187, 11123, 11185, 11185, 11117, 11173, 11171, 11093, 11136, 11126, 10985, 11042, 10962, 11010, 10960, 11013, 10967, 10968, 11017, 10971, 10972, 10971, 11013, 10968, 10962, 11012, 11011, 11202, 11203, 11201, 11203, 11205, 11201, 11210, 11207, 11208, 11199, 11202, 11198, 11204, 11207, 11203, 11203, 11209, 11206, 11199, 11208, 11204, 11212, 11213, 11211, 11212, 11216, 11214, 11220, 11221, 11219, 11218, 11220, 11217, 11229, 11230, 11228, 11235, 11236, 11234, 11241, 11238, 11239, 11221, 11232, 11219, 11235, 11226, 11227, 11239, 11230, 11231, 11221, 11227, 11226, 11225, 11222, 11218, 11223, 11218, 11217, 11230, 11225, 11228, 11240, 11235, 11238, 11218, 11228, 11225, 11230, 11235, 11227, 11226, 11233, 11221, 11206, 11223, 11205, 11229, 11209, 11210, 11206, 11228, 11224, 11232, 11212, 11211, 11236, 11215, 11234, 11233, 11215, 11212, 11243, 11244, 11242, 11244, 11246, 11242, 11253, 11250, 11251, 11255, 11256, 11254, 11261, 11258, 11259, 11265, 11262, 11263, 11253, 11258, 11252, 11263, 11256, 11257, 11244, 11251, 11250, 11250, 11247, 11244, 11248, 11245, 11243, 11256, 11251, 11254,
11264, 11259, 11262, 11245, 11254, 11251, 11262, 11253, 11256, 11271, 11268, 11270, 11273, 11274, 11272, 11281, 11278, 11280, 11283, 11278, 11282, 11285, 11286, 11284, 11281, 11266, 11279, 11289, 11274, 11275, 11288, 11281, 11280, 11283, 11288, 11280, 11271, 11276, 11269, 11285, 11292, 11287, 11271, 11266, 11267, 11277, 11273, 11276, 11275, 11271, 11267, 11289, 11267, 11281, 11296, 11293, 11295, 11300, 11297, 11298, 11304, 11301, 11302, 11310, 11307, 11309, 11310, 11305, 11308, 11314, 11311, 11312, 11318, 11315, 11316, 11306, 11295, 11305, 11312, 11299, 11300, 11316, 11303, 11304, 11296, 11302, 11294, 11317, 11310, 11309, 11310, 11316, 11306, 11303, 11298, 11301, 11315, 11314, 11312, 11315, 11300, 11303, 11306, 11304, 11296, 11322, 11323, 11321, 11323, 11325, 11321, 11331, 11328, 11329, 11335, 11336, 11334, 11342, 11339, 11340, 11326, 11334, 11325, 11340, 11330, 11331, 11323, 11329, 11328, 11328, 11326, 11323, 11335, 11341, 11337, 11336, 11333, 11332, 11333, 11341, 11338, 11326, 11339, 11335, 11298, 11260, 11261, 11265, 11301, 11264, 11301, 11261, 11264, 11320, 11282, 11319, 11287, 11322, 11286, 11320, 11290, 11283, 11287, 11329, 11324, 11274, 11255, 11272, 11290, 11263, 11288, 11274, 11263, 11257, 11320, 11293, 11294, 11302, 11320, 11294,
11327, 11265, 11290, 11344, 11345, 11343, 11345, 11347, 11343, 11352, 11349, 11350, 11356, 11357, 11355, 11354, 11356, 11353, 11365, 11366, 11364, 11348, 11359, 11347, 11365, 11351, 11352, 11358, 11362, 11357, 11361, 11358, 11354, 11345, 11350, 11349, 11345, 11351, 11348, 11359, 11354, 11353, 11366, 11361, 11364, 11354, 11364, 11361, 11348, 11364, 11360, 11371, 11368, 11369, 11375, 11372, 11373, 11379, 11376, 11377, 11383, 11380, 11381, 11377, 11370, 11371, 11381, 11374, 11375, 11374, 11369, 11372, 11382, 11377, 11380, 11380, 11371, 11374, 11387, 11388, 11386, 11393, 11390, 11391, 11397, 11394, 11395, 11401, 11398, 11399, 11395, 11388, 11389, 11399, 11392, 11393, 11392, 11387, 11390, 11400, 11395, 11398, 11398, 11389, 11392, 11405, 11402, 11403, 11409, 11406, 11407, 11413, 11410, 11411, 11417, 11414, 11415, 11411, 11404, 11405, 11415, 11408, 11409, 11408, 11403, 11406, 11416, 11411, 11414, 11414, 11405, 11408, 11357, 11384, 11355, 11387, 11362, 11363, 11391, 11366, 11367, 11366, 11387, 11363, 11362, 11385, 11357, 11403, 11378, 11379, 11407, 11382, 11383, 11406, 11379, 11382, 11368, 11352, 11350, 11378, 11365, 11376, 11376, 11352, 11370, 11404, 11391, 11402, 11412, 11399, 11410, 11410, 11393, 11404, 11402, 11367, 11378, 11421, 11418, 11420,
11425, 11422, 11423, 11431, 11428, 11430, 11437, 11434, 11435, 11433, 11420, 11432, 11425, 11434, 11424, 11437, 11429, 11431, 11421, 11426, 11419, 11427, 11423, 11426, 11433, 11428, 11429, 11435, 11433, 11429, 11438, 11425, 11427, 11438, 11421, 11433, 11440, 11441, 11439, 11448, 11445, 11446, 11450, 11451, 11449, 11458, 11455, 11456, 11442, 11449, 11441, 11456, 11447, 11448, 11440, 11444, 11442, 11450, 11454, 11452, 11447, 11443, 11445, 11457, 11453, 11455, 11455, 11444, 11447, 11442, 11453, 11450, 11462, 11459, 11461, 11466, 11463, 11464, 11470, 11467, 11469, 11472, 11473, 11471, 11468, 11461, 11467, 11472, 11465, 11466, 11462, 11464, 11460, 11470, 11472, 11468, 11468, 11466, 11462, 11476, 11477, 11475, 11484, 11481, 11482, 11486, 11487, 11485, 11494, 11491, 11492, 11478, 11485, 11477, 11492, 11483, 11484, 11476, 11480, 11478, 11489, 11488, 11486, 11481, 11480, 11479, 11493, 11489, 11491, 11491, 11480, 11483, 11478, 11489, 11486, 11460, 11430, 11459, 11464, 11436, 11437, 11460, 11437, 11431, 11452, 11475, 11451, 11482, 11457, 11458, 11481, 11454, 11457, 11452, 11479, 11476, 11214, 11344, 11213, 11214, 11350, 11346, 11369, 11236, 11237, 11373, 11240, 11241, 11372, 11237, 11240, 11368, 11216, 11236, 11419, 11307, 11418, 11423, 11313, 11314,
11426, 11314, 11317, 11419, 11317, 11309, 11333, 11439, 11332, 11342, 11445, 11341, 11445, 11338, 11341, 11338, 11440, 11333, 11246, 11199, 11198, 11258, 11210, 11208, 11199, 11252, 11200, 11258, 11200, 11252, 11299, 11229, 11297, 11313, 11239, 11311, 11311, 11231, 11299, 11297, 11210, 11260, 11424, 11373, 11422, 11436, 11381, 11434, 11434, 11375, 11424, 11465, 11407, 11463, 11473, 11415, 11471, 11471, 11409, 11465, 11463, 11383, 11436, 11422, 11241, 11313, 11496, 11497, 11495, 11497, 11499, 11495, 11504, 11501, 11502, 11508, 11505, 11506, 11514, 11515, 11513, 11518, 11519, 11517, 11500, 11509, 11499, 11514, 11503, 11504, 11518, 11507, 11508, 11497, 11502, 11501, 11497, 11503, 11500, 11505, 11504, 11502, 11519, 11514, 11517, 11517, 11504, 11507, 11500, 11513, 11510, 11524, 11521, 11523, 11526, 11521, 11525, 11533, 11530, 11532, 11531, 11525, 11530, 11527, 11524, 11523, 11526, 11527, 11523, 11533, 11534, 11531, 11531, 11529, 11526, 11539, 11536, 11538, 11543, 11540, 11541, 11547, 11544, 11545, 11551, 11548, 11550, 11555, 11552, 11553, 11557, 11558, 11556, 11549, 11538, 11548, 11553, 11542, 11543, 11557, 11546, 11547, 11539, 11545, 11537, 11551, 11557, 11549, 11546, 11541, 11544, 11558, 11553, 11556, 11556, 11543, 11546, 11549, 11547, 11539,
11537, 11511, 11536, 11541, 11515, 11516, 11545, 11519, 11520, 11544, 11516, 11519, 11545, 11512, 11537, 11529, 11506, 11527, 11533, 11511, 11512, 11535, 11518, 11534, 11520, 11533, 11512, 11534, 11508, 11529, 11563, 11564, 11562, 11564, 11566, 11562, 11572, 11569, 11570, 11574, 11575, 11573, 11582, 11579, 11580, 11561, 11573, 11560, 11572, 11579, 11571, 11574, 11578, 11576, 11565, 11569, 11564, 11569, 11567, 11564, 11566, 11561, 11560, 11581, 11577, 11579, 11561, 11571, 11568, 11579, 11568, 11571, 11568, 11574, 11561, 11586, 11587, 11585, 11587, 11583, 11585, 11590, 11591, 11589, 11601, 11598, 11599, 11605, 11602, 11603, 11595, 11598, 11594, 11597, 11602, 11596, 11587, 11595, 11594, 11594, 11584, 11587, 11592, 11596, 11591, 11589, 11588, 11586, 11604, 11599, 11602, 11591, 11595, 11588, 11596, 11599, 11595, 11609, 11610, 11608, 11607, 11609, 11606, 11618, 11615, 11616, 11620, 11621, 11619, 11626, 11623, 11624, 11630, 11627, 11628, 11610, 11619, 11608, 11624, 11613, 11614, 11628, 11617, 11618, 11610, 11614, 11613, 11612, 11611, 11607, 11620, 11625, 11622, 11617, 11612, 11615, 11629, 11624, 11627, 11617, 11624, 11614, 11613, 11620, 11610, 11634, 11631, 11632, 11638, 11635, 11636, 11642, 11639, 11640, 11646, 11643, 11644, 11640, 11633, 11634,
11644, 11637, 11638, 11637, 11632, 11635, 11645, 11640, 11643, 11643, 11634, 11637, 11576, 11606, 11575, 11616, 11581, 11582, 11615, 11578, 11581, 11576, 11612, 11607, 11632, 11600, 11601, 11636, 11604, 11605, 11635, 11601, 11604, 11583, 11565, 11563, 11593, 11572, 11570, 11600, 11580, 11598, 11565, 11593, 11570, 11598, 11572, 11594, 11633, 11616, 11631, 11641, 11628, 11639, 11639, 11618, 11633, 11631, 11582, 11600, 11650, 11647, 11649, 11654, 11651, 11652, 11656, 11657, 11655, 11662, 11659, 11661, 11666, 11663, 11664, 11668, 11669, 11667, 11660, 11649, 11659, 11664, 11653, 11654, 11658, 11667, 11657, 11658, 11648, 11650, 11670, 11660, 11662, 11657, 11652, 11655, 11669, 11664, 11667, 11667, 11654, 11657, 11668, 11650, 11660, 11676, 11673, 11674, 11680, 11677, 11678, 11684, 11681, 11682, 11688, 11685, 11686, 11682, 11675, 11676, 11686, 11679, 11680, 11679, 11674, 11677, 11687, 11682, 11685, 11685, 11676, 11679, 11692, 11689, 11691, 11698, 11695, 11697, 11696, 11691, 11695, 11692, 11693, 11690, 11700, 11696, 11698, 11699, 11692, 11696, 11672, 11661, 11671, 11674, 11665, 11666, 11670, 11677, 11669, 11677, 11666, 11669, 11672, 11670, 11662, 11689, 11672, 11671, 11693, 11680, 11678, 11699, 11688, 11686, 11690, 11678, 11672, 11694, 11686, 11680,
11704, 11701, 11703, 11708, 11705, 11707, 11710, 11703, 11709, 11712, 11707, 11711, 11707, 11702, 11704, 11711, 11704, 11710, 11718, 11715, 11717, 11714, 11715, 11713, 11722, 11719, 11721, 11724, 11725, 11723, 11728, 11729, 11727, 11725, 11718, 11717, 11723, 11717, 11714, 11730, 11721, 11729, 11721, 11716, 11718, 11727, 11726, 11724, 11729, 11718, 11726, 11714, 11706, 11708, 11712, 11714, 11708, 11648, 11550, 11647, 11652, 11554, 11555, 11656, 11558, 11559, 11655, 11555, 11558, 11656, 11551, 11648, 11622, 11701, 11621, 11626, 11709, 11625, 11630, 11711, 11629, 11629, 11710, 11626, 11622, 11709, 11703, 11724, 11641, 11642, 11728, 11645, 11646, 11727, 11642, 11645, 11641, 11712, 11630, 11734, 11731, 11732, 11738, 11735, 11736, 11742, 11739, 11740, 11746, 11743, 11744, 11740, 11733, 11734, 11744, 11737, 11738, 11737, 11732, 11735, 11745, 11740, 11743, 11743, 11734, 11737, 11750, 11747, 11748, 11754, 11751, 11752, 11758, 11755, 11756, 11762, 11759, 11760, 11756, 11749, 11750, 11760, 11753, 11754, 11753, 11748, 11751, 11761, 11756, 11759, 11759, 11750, 11753, 11766, 11763, 11764, 11770, 11767, 11768, 11772, 11773, 11771, 11778, 11775, 11776, 11772, 11765, 11766, 11776, 11769, 11770, 11769, 11764, 11767, 11777, 11772, 11775, 11775, 11766, 11769,
11782, 11779, 11780, 11786, 11783, 11784, 11790, 11787, 11788, 11794, 11791, 11792, 11788, 11781, 11782, 11792, 11785, 11786, 11785, 11780, 11783, 11793, 11788, 11791, 11791, 11782, 11785, 11764, 11741, 11742, 11768, 11745, 11746, 11767, 11742, 11745, 11780, 11757, 11758, 11784, 11761, 11762, 11783, 11758, 11761, 11749, 11736, 11747, 11757, 11744, 11755, 11755, 11738, 11749, 11781, 11768, 11779, 11789, 11776, 11787, 11787, 11770, 11781, 11779, 11746, 11757, 11798, 11795, 11797, 11800, 11801, 11799, 11806, 11803, 11805, 11810, 11811, 11809, 11804, 11797, 11803, 11802, 11809, 11801, 11802, 11796, 11798, 11806, 11810, 11804, 11807, 11806, 11805, 11808, 11812, 11806, 11810, 11798, 11804, 11815, 11816, 11814, 11823, 11820, 11821, 11825, 11826, 11824, 11833, 11830, 11831, 11817, 11824, 11816, 11831, 11822, 11823, 11818, 11817, 11815, 11828, 11827, 11825, 11820, 11819, 11818, 11830, 11829, 11828, 11822, 11828, 11819, 11819, 11825, 11817, 11837, 11834, 11836, 11841, 11838, 11839, 11847, 11844, 11846, 11851, 11848, 11849, 11845, 11836, 11844, 11849, 11840, 11841, 11843, 11835, 11837, 11853, 11845, 11847, 11842, 11841, 11839, 11852, 11851, 11849, 11843, 11849, 11841, 11852, 11837, 11845, 11855, 11856, 11854, 11865, 11862, 11863, 11867, 11868, 11866,
11871, 11872, 11870, 11859, 11866, 11858, 11865, 11870, 11864, 11855, 11861, 11857, 11870, 11869, 11867, 11859, 11856, 11857, 11862, 11861, 11860, 11864, 11857, 11861, 11859, 11870, 11867, 11835, 11807, 11834, 11839, 11811, 11812, 11813, 11839, 11812, 11835, 11813, 11808, 11827, 11854, 11826, 11863, 11832, 11833, 11832, 11860, 11829, 11829, 11855, 11827, 11877, 11874, 11876, 11877, 11878, 11875, 11886, 11883, 11884, 11888, 11889, 11887, 11896, 11893, 11894, 11879, 11887, 11878, 11894, 11885, 11886, 11881, 11876, 11880, 11877, 11882, 11879, 11891, 11890, 11888, 11883, 11882, 11881, 11893, 11892, 11891, 11885, 11891, 11882, 11882, 11888, 11879, 11900, 11897, 11898, 11904, 11901, 11902, 11908, 11905, 11906, 11912, 11909, 11910, 11906, 11899, 11900, 11910, 11903, 11904, 11903, 11898, 11901, 11911, 11906, 11909, 11909, 11900, 11903, 11916, 11917, 11915, 11921, 11922, 11920, 11929, 11926, 11928, 11925, 11926, 11924, 11931, 11932, 11930, 11918, 11924, 11917, 11923, 11930, 11922, 11916, 11922, 11918, 11933, 11928, 11932, 11930, 11928, 11925, 11916, 11913, 11914, 11920, 11914, 11919, 11922, 11925, 11918, 11937, 11934, 11935, 11941, 11938, 11939, 11945, 11942, 11943, 11949, 11946, 11947, 11943, 11936, 11937, 11947, 11940, 11941, 11940, 11935, 11938,
11948, 11943, 11946, 11946, 11937, 11940, 11890, 11913, 11889, 11921, 11895, 11896, 11895, 11919, 11892, 11890, 11919, 11914, 11935, 11907, 11908, 11939, 11911, 11912, 11938, 11908, 11911, 11899, 11884, 11897, 11907, 11894, 11905, 11905, 11886, 11899, 11936, 11921, 11934, 11944, 11931, 11942, 11942, 11923, 11936, 11934, 11896, 11907, 11951, 11952, 11950, 11955, 11951, 11954, 11959, 11956, 11957, 11961, 11962, 11960, 11969, 11966, 11967, 11973, 11970, 11971, 11965, 11954, 11964, 11967, 11958, 11959, 11963, 11970, 11962, 11953, 11960, 11952, 11955, 11961, 11953, 11962, 11957, 11960, 11972, 11967, 11970, 11970, 11959, 11962, 11971, 11955, 11965, 11979, 11976, 11978, 11975, 11976, 11974, 11981, 11982, 11980, 11987, 11984, 11986, 11989, 11990, 11988, 11979, 11988, 11977, 11993, 11985, 11987, 11983, 11978, 11982, 11980, 11978, 11975, 11994, 11991, 11989, 11983, 11989, 11979, 11999, 11996, 11997, 12003, 12000, 12001, 12007, 12004, 12005, 12011, 12008, 12009, 12005, 11998, 11999, 12009, 12002, 12003, 12002, 11997, 12000, 12010, 12005, 12008, 12008, 11999, 12002, 12015, 12016, 12014, 12016, 12012, 12014, 12024, 12021, 12023, 12028, 12029, 12027, 12022, 12015, 12021, 12020, 12027, 12019, 12017, 12019, 12016, 12016, 12018, 12013, 12030, 12022, 12024,
12025, 12024, 12023, 12026, 12030, 12024, 12028, 12017, 12022, 11997, 11968, 11969, 12001, 11972, 11973, 12000, 11969, 11972, 12013, 11986, 12012, 12018, 11987, 12013, 11984, 11965, 11964, 11992, 11973, 11971, 11985, 11971, 11965, 12018, 12003, 12001, 12029, 12009, 12027, 12027, 12003, 12019, 11993, 12001, 11973, 11774, 11880, 11773, 11884, 11777, 11778, 11883, 11774, 11777, 11898, 11789, 11790, 11902, 11793, 11794, 11901, 11790, 11793, 11897, 11778, 11789, 11952, 11846, 11950, 11957, 11850, 11851, 11853, 11957, 11851, 11952, 11853, 11847, 11869, 11974, 11868, 11873, 11980, 11872, 11872, 11975, 11869, 11801, 11752, 11799, 11811, 11760, 11809, 11809, 11754, 11801, 11840, 11784, 11838, 11850, 11792, 11848, 11848, 11786, 11840, 11838, 11762, 11811, 11958, 11902, 11956, 11968, 11910, 11966, 11966, 11904, 11958, 11998, 11939, 11996, 12006, 11947, 12004, 12004, 11941, 11998, 11996, 11912, 11968, 11956, 11794, 11850, 12035, 12032, 12033, 12039, 12036, 12037, 12043, 12040, 12041, 12047, 12044, 12045, 12041, 12034, 12035, 12045, 12038, 12039, 12038, 12033, 12036, 12046, 12041, 12044, 12044, 12035, 12038, 12049, 12050, 12048, 12053, 12049, 12052, 12055, 12056, 12054, 12061, 12058, 12060, 12065, 12066, 12064, 12059, 12052, 12058, 12057, 12064, 12056,
12051, 12054, 12050, 12053, 12055, 12051, 12061, 12065, 12059, 12062, 12061, 12060, 12063, 12067, 12061, 12065, 12053, 12059, 12072, 12069, 12070, 12076, 12073, 12074, 12080, 12077, 12078, 12084, 12081, 12082, 12078, 12071, 12072, 12082, 12075, 12076, 12075, 12070, 12073, 12083, 12078, 12081, 12081, 12072, 12075, 12088, 12089, 12087, 12089, 12085, 12087, 12094, 12091, 12092, 12101, 12098, 12100, 12105, 12102, 12103, 12107, 12108, 12106, 12099, 12088, 12098, 12103, 12093, 12094, 12097, 12106, 12096, 12089, 12097, 12096, 12096, 12086, 12089, 12109, 12099, 12101, 12095, 12094, 12092, 12108, 12103, 12106, 12106, 12094, 12096, 12107, 12090, 12099, 12070, 12042, 12043, 12074, 12046, 12047, 12073, 12043, 12046, 12086, 12062, 12085, 12092, 12066, 12067, 12068, 12092, 12067, 12095, 12063, 12086, 12056, 12037, 12054, 12066, 12045, 12064, 12064, 12039, 12056, 12093, 12074, 12091, 12104, 12082, 12102, 12102, 12076, 12093, 12091, 12047, 12066, 12113, 12110, 12111, 12117, 12114, 12115, 12123, 12120, 12121, 12127, 12124, 12125, 12121, 12112, 12113, 12125, 12116, 12117, 12116, 12111, 12114, 12126, 12121, 12124, 12124, 12113, 12116, 12133, 12130, 12131, 12137, 12134, 12135, 12141, 12138, 12139, 12145, 12142, 12143, 12139, 12132, 12133, 12143, 12136, 12137,
12136, 12131, 12134, 12144, 12139, 12142, 12142, 12133, 12136, 12151, 12148, 12150, 12147, 12148, 12146, 12153, 12154, 12152, 12159, 12156, 12157, 12161, 12162, 12160, 12171, 12168, 12169, 12151, 12160, 12149, 12169, 12158, 12159, 12155, 12150, 12154, 12152, 12150, 12147, 12166, 12163, 12161, 12158, 12153, 12156, 12165, 12162, 12163, 12168, 12167, 12166, 12170, 12163, 12167, 12158, 12166, 12155, 12155, 12161, 12151, 12175, 12172, 12173, 12179, 12176, 12177, 12183, 12180, 12181, 12187, 12184, 12185, 12181, 12174, 12175, 12185, 12178, 12179, 12178, 12173, 12176, 12186, 12181, 12184, 12184, 12175, 12178, 12119, 12146, 12118, 12123, 12152, 12122, 12157, 12126, 12127, 12156, 12123, 12126, 12122, 12147, 12119, 12173, 12140, 12141, 12177, 12144, 12145, 12176, 12141, 12144, 12132, 12115, 12130, 12140, 12125, 12138, 12138, 12117, 12132, 12174, 12157, 12172, 12182, 12169, 12180, 12180, 12159, 12174, 12172, 12127, 12140, 12191, 12188, 12190, 12195, 12196, 12194, 12199, 12195, 12198, 12201, 12202, 12200, 12196, 12190, 12194, 12193, 12189, 12191, 12197, 12200, 12196, 12203, 12197, 12199, 12200, 12191, 12196, 12205, 12206, 12204, 12205, 12209, 12207, 12215, 12212, 12214, 12219, 12216, 12217, 12223, 12224, 12222, 12227, 12223, 12226, 12231, 12228, 12229,
12233, 12234, 12232, 12224, 12214, 12222, 12229, 12218, 12219, 12221, 12213, 12215, 12225, 12232, 12224, 12235, 12225, 12227, 12212, 12211, 12210, 12220, 12219, 12217, 12234, 12229, 12232, 12213, 12217, 12211, 12221, 12229, 12219, 12232, 12215, 12224, 12237, 12238, 12236, 12243, 12244, 12242, 12239, 12242, 12238, 12237, 12241, 12239, 12246, 12245, 12243, 12241, 12243, 12239, 12211, 12198, 12210, 12203, 12216, 12202, 12217, 12199, 12211, 12207, 12236, 12206, 12207, 12240, 12237, 12111, 12079, 12080, 12115, 12083, 12084, 12114, 12080, 12083, 12129, 12100, 12128, 12131, 12104, 12105, 12109, 12134, 12108, 12134, 12105, 12108, 12129, 12109, 12101, 12130, 12084, 12104, 12188, 12129, 12128, 12192, 12137, 12135, 12202, 12143, 12200, 12189, 12135, 12129, 12193, 12143, 12137, 12218, 12177, 12216, 12230, 12185, 12228, 12228, 12179, 12218, 12216, 12145, 12202, 11732, 11396, 11397, 11736, 11400, 11401, 11735, 11397, 11400, 11748, 11412, 11413, 11752, 11416, 11417, 11751, 11413, 11416, 11747, 11401, 11412, 11796, 11469, 11795, 11474, 11799, 11473, 11800, 11470, 11796, 11488, 11814, 11487, 11821, 11493, 11494, 11493, 11818, 11490, 11490, 11815, 11488, 11799, 11417, 11473, 12033, 11683, 11684, 12037, 11687, 11688, 12036, 11684, 11687, 12050, 11697, 12048,
12050, 11700, 11698, 11700, 12037, 11688, 11509, 11285, 11284, 11515, 11291, 11513, 11285, 11513, 11291, 11540, 11331, 11329, 11554, 11340, 11552, 11552, 11331, 11542, 11540, 11292, 11515, 11653, 11446, 11651, 11665, 11456, 11663, 11663, 11448, 11653, 11675, 11482, 11673, 11683, 11492, 11681, 11681, 11484, 11675, 11673, 11458, 11665, 11651, 11342, 11554, 12034, 11821, 12032, 12042, 11831, 12040, 12040, 11823, 12034, 12071, 11863, 12069, 12079, 11871, 12077, 12077, 11865, 12071, 12069, 11833, 12042, 12112, 11981, 12110, 12119, 11990, 11991, 12120, 11995, 11994, 12122, 11991, 11995, 12112, 11994, 11983, 12110, 11873, 12079, 12032, 11494, 11683, 10805, 11248, 10803, 11255, 10808, 10809, 10805, 11254, 11249, 11269, 10830, 11268, 11273, 10836, 10837, 11276, 10837, 10840, 11269, 10840, 10832, 10836, 11255, 10809, 10891, 11496, 10890, 10903, 11505, 10902, 11505, 10899, 10902, 10899, 11498, 10891, 11524, 10925, 11522, 11528, 10931, 10932, 11524, 10932, 10926, 10931, 11506, 10903, 11384, 10950, 10949, 11388, 10955, 11386, 11396, 10970, 11394, 10950, 11386, 10955, 11388, 10970, 10957, 11733, 11017, 11731, 11741, 11025, 11739, 11739, 11019, 11733, 11765, 11053, 11763, 11771, 11063, 11061, 11771, 11055, 11765, 11763, 11027, 11741, 11876, 11142, 11143,
11153, 11876, 11143, 11880, 11063, 11773, 11731, 10972, 11396, 12252, 12253, 12251, 12258, 12259, 12257, 12254, 12257, 12253, 12252, 12256, 12254, 12258, 12262, 12260, 12254, 12261, 12258, 12264, 12265, 12263, 12270, 12271, 12269, 12266, 12269, 12265, 12264, 12268, 12266, 12273, 12272, 12270, 12268, 12270, 12266, 12260, 12263, 12259, 12260, 12267, 12264, 12249, 12251, 12248, 12250, 12252, 12249, 12276, 12277, 12275, 12284, 12285, 12283, 12290, 12291, 12289, 12288, 12290, 12287, 12299, 12296, 12297, 12278, 12287, 12277, 12297, 12285, 12286, 12276, 12282, 12278, 12291, 12295, 12294, 12293, 12292, 12288, 12279, 12276, 12275, 12285, 12281, 12283, 12298, 12293, 12296, 12276, 12283, 12281, 12296, 12282, 12285, 12278, 12293, 12288, 12303, 12300, 12301, 12307, 12304, 12305, 12311, 12308, 12309, 12315, 12312, 12313, 12309, 12302, 12303, 12313, 12306, 12307, 12306, 12301, 12304, 12314, 12309, 12312, 12312, 12303, 12306, 12317, 12318, 12316, 12323, 12320, 12321, 12327, 12324, 12325, 12329, 12330, 12328, 12335, 12332, 12333, 12339, 12336, 12337, 12319, 12328, 12318, 12333, 12322, 12323, 12337, 12326, 12327, 12317, 12322, 12319, 12332, 12331, 12329, 12326, 12321, 12324, 12338, 12333, 12336, 12336, 12323, 12326, 12319, 12332, 12329, 12345, 12342, 12343,
12347, 12348, 12346, 12353, 12350, 12352, 12357, 12354, 12355, 12361, 12358, 12359, 12351, 12340, 12350, 12355, 12344, 12345, 12359, 12348, 12349, 12353, 12359, 12351, 12348, 12343, 12346, 12360, 12355, 12358, 12358, 12345, 12348, 12351, 12349, 12341, 12291, 12316, 12289, 12321, 12294, 12295, 12325, 12298, 12299, 12298, 12321, 12295, 12294, 12317, 12291, 12343, 12310, 12311, 12347, 12314, 12315, 12346, 12311, 12314, 12302, 12284, 12300, 12310, 12297, 12308, 12308, 12286, 12302, 12344, 12325, 12342, 12356, 12337, 12354, 12354, 12327, 12344, 12342, 12299, 12310, 12365, 12362, 12364, 12371, 12368, 12369, 12376, 12373, 12375, 12378, 12373, 12377, 12376, 12364, 12374, 12380, 12370, 12371, 12365, 12369, 12363, 12379, 12376, 12375, 12381, 12375, 12378, 12367, 12362, 12363, 12369, 12367, 12363, 12376, 12371, 12365, 12385, 12382, 12384, 12385, 12386, 12383, 12383, 12377, 12382, 12383, 12381, 12378, 12391, 12388, 12389, 12395, 12392, 12393, 12399, 12396, 12397, 12403, 12400, 12401, 12397, 12390, 12391, 12401, 12394, 12395, 12394, 12389, 12392, 12402, 12397, 12400, 12400, 12391, 12394, 12407, 12404, 12406, 12409, 12404, 12408, 12413, 12410, 12411, 12420, 12417, 12419, 12424, 12421, 12422, 12418, 12408, 12417, 12422, 12412, 12413, 12414, 12407, 12406,
12416, 12406, 12409, 12420, 12425, 12418, 12416, 12411, 12414, 12426, 12422, 12425, 12425, 12413, 12416, 12418, 12416, 12409, 12430, 12427, 12428, 12434, 12431, 12432, 12438, 12435, 12436, 12442, 12439, 12440, 12436, 12429, 12430, 12440, 12433, 12434, 12433, 12428, 12431, 12441, 12436, 12439, 12439, 12430, 12433, 12446, 12443, 12445, 12450, 12447, 12448, 12456, 12453, 12455, 12462, 12459, 12460, 12458, 12445, 12457, 12450, 12459, 12449, 12462, 12454, 12456, 12446, 12451, 12444, 12452, 12448, 12451, 12458, 12453, 12454, 12460, 12458, 12454, 12463, 12450, 12452, 12463, 12446, 12458, 12428, 12398, 12399, 12432, 12402, 12403, 12431, 12399, 12402, 12444, 12419, 12443, 12448, 12423, 12424, 12451, 12424, 12426, 12444, 12426, 12420, 12412, 12393, 12410, 12423, 12401, 12421, 12421, 12395, 12412, 12449, 12432, 12447, 12461, 12440, 12459, 12459, 12434, 12449, 12447, 12403, 12423, 12335, 12388, 12334, 12393, 12338, 12339, 12392, 12335, 12338, 12407, 12352, 12405, 12411, 12356, 12357, 12415, 12360, 12361, 12360, 12411, 12357, 12415, 12353, 12407, 12410, 12339, 12356, 12370, 12305, 12368, 12381, 12313, 12379, 12370, 12313, 12307, 12385, 12340, 12341, 12387, 12347, 12386, 12349, 12385, 12341, 12386, 12315, 12381, 12465, 12466, 12464, 12471, 12472, 12470,
12467, 12470, 12466, 12468, 12467, 12465, 12474, 12473, 12471, 12469, 12471, 12467, 12477, 12478, 12476, 12483, 12484, 12482, 12479, 12482, 12478, 12480, 12479, 12477, 12486, 12485, 12483, 12481, 12483, 12479, 12473, 12476, 12472, 12475, 12477, 12473, 12489, 12490, 12488, 12489, 12493, 12491, 12485, 12488, 12484, 12487, 12489, 12485, 12497, 12494, 12495, 12501, 12498, 12499, 12505, 12502, 12503, 12509, 12506, 12507, 12503, 12496, 12497, 12507, 12500, 12501, 12500, 12495, 12498, 12508, 12503, 12506, 12506, 12497, 12500, 12513, 12510, 12512, 12517, 12514, 12515, 12521, 12518, 12520, 12525, 12522, 12523, 12519, 12512, 12518, 12517, 12522, 12516, 12513, 12515, 12511, 12525, 12519, 12521, 12519, 12517, 12513, 12529, 12526, 12527, 12533, 12530, 12531, 12537, 12534, 12535, 12541, 12538, 12539, 12535, 12528, 12529, 12539, 12532, 12533, 12532, 12527, 12530, 12540, 12535, 12538, 12538, 12529, 12532, 12545, 12542, 12544, 12547, 12548, 12546, 12555, 12552, 12554, 12559, 12556, 12557, 12551, 12544, 12550, 12549, 12556, 12548, 12549, 12543, 12545, 12555, 12560, 12553, 12552, 12551, 12550, 12560, 12559, 12557, 12553, 12557, 12551, 12551, 12549, 12545, 12527, 12504, 12505, 12531, 12508, 12509, 12530, 12505, 12508, 12543, 12520, 12542, 12525, 12546, 12524,
12547, 12521, 12543, 12516, 12499, 12514, 12524, 12507, 12522, 12522, 12501, 12516, 12548, 12531, 12546, 12558, 12539, 12556, 12556, 12533, 12548, 12546, 12509, 12524, 12567, 12564, 12565, 12571, 12568, 12569, 12573, 12574, 12572, 12577, 12578, 12576, 12583, 12580, 12581, 12563, 12572, 12562, 12567, 12576, 12566, 12581, 12570, 12571, 12576, 12575, 12573, 12570, 12565, 12568, 12582, 12577, 12580, 12580, 12567, 12570, 12563, 12576, 12573, 12587, 12584, 12586, 12591, 12588, 12589, 12597, 12594, 12596, 12601, 12598, 12599, 12595, 12586, 12594, 12599, 12590, 12591, 12593, 12585, 12587, 12603, 12595, 12597, 12592, 12591, 12589, 12602, 12601, 12599, 12593, 12599, 12591, 12602, 12587, 12595, 12609, 12606, 12608, 12605, 12606, 12604, 12611, 12612, 12610, 12617, 12614, 12615, 12619, 12620, 12618, 12627, 12624, 12625, 12609, 12618, 12607, 12625, 12616, 12617, 12613, 12608, 12612, 12605, 12612, 12608, 12622, 12621, 12619, 12616, 12611, 12614, 12624, 12623, 12622, 12616, 12622, 12613, 12609, 12622, 12619, 12629, 12630, 12628, 12633, 12629, 12632, 12637, 12634, 12635, 12639, 12640, 12638, 12645, 12642, 12644, 12649, 12646, 12647, 12651, 12652, 12650, 12643, 12632, 12642, 12647, 12636, 12637, 12641, 12650, 12640, 12631, 12638, 12630, 12633, 12639, 12631,
12653, 12643, 12645, 12640, 12635, 12638, 12652, 12647, 12650, 12650, 12637, 12640, 12651, 12633, 12643, 12575, 12604, 12574, 12579, 12610, 12578, 12615, 12582, 12583, 12614, 12579, 12582, 12578, 12605, 12575, 12630, 12596, 12628, 12635, 12600, 12601, 12603, 12635, 12601, 12630, 12603, 12597, 12590, 12569, 12588, 12600, 12581, 12598, 12598, 12571, 12590, 12636, 12615, 12634, 12648, 12625, 12646, 12646, 12617, 12636, 12634, 12583, 12600, 12565, 12536, 12537, 12569, 12540, 12541, 12568, 12537, 12540, 12585, 12554, 12584, 12589, 12558, 12559, 12561, 12589, 12559, 12592, 12555, 12585, 12588, 12541, 12558, 12654, 12645, 12644, 12655, 12653, 12645, 12272, 12464, 12271, 12274, 12465, 12272, 12495, 12437, 12438, 12499, 12441, 12442, 12498, 12438, 12441, 12511, 12455, 12510, 12515, 12461, 12462, 12511, 12462, 12456, 12514, 12442, 12461, 12330, 12249, 12248, 12249, 12334, 12250, 12390, 12255, 12388, 12398, 12261, 12396, 12396, 12256, 12390, 12427, 12268, 12267, 12437, 12273, 12435, 12435, 12268, 12429, 12427, 12262, 12398, 12388, 12250, 12334, 12496, 12468, 12494, 12504, 12474, 12502, 12502, 12469, 12496, 12528, 12480, 12526, 12534, 12487, 12486, 12528, 12486, 12481, 12526, 12475, 12504, 12563, 12490, 12491, 12564, 12493, 12492, 12566, 12491, 12493,
12536, 12492, 12487, 12494, 12274, 12437, 12658, 12659, 12657, 12661, 12660, 12658, 12668, 12665, 12666, 12672, 12669, 12670, 12678, 12675, 12677, 12674, 12675, 12673, 12680, 12681, 12679, 12686, 12683, 12684, 12664, 12673, 12663, 12668, 12679, 12667, 12684, 12671, 12672, 12682, 12677, 12681, 12679, 12677, 12674, 12671, 12666, 12669, 12685, 12680, 12683, 12683, 12668, 12671, 12667, 12674, 12664, 12688, 12689, 12687, 12698, 12695, 12696, 12704, 12701, 12703, 12700, 12701, 12699, 12706, 12707, 12705, 12692, 12699, 12691, 12698, 12705, 12697, 12693, 12690, 12688, 12708, 12703, 12707, 12705, 12703, 12700, 12692, 12689, 12690, 12695, 12694, 12693, 12697, 12690, 12694, 12697, 12700, 12692, 12678, 12687, 12676, 12696, 12685, 12686, 12685, 12693, 12682, 12682, 12688, 12678, 12664, 12659, 12660, 12665, 12662, 12661, 12667, 12660, 12662, 12712, 12709, 12710, 12716, 12713, 12714, 12720, 12717, 12718, 12724, 12721, 12722, 12718, 12711, 12712, 12722, 12715, 12716, 12715, 12710, 12713, 12723, 12718, 12721, 12721, 12712, 12715, 12728, 12729, 12727, 12729, 12725, 12727, 12734, 12731, 12732, 12741, 12738, 12740, 12745, 12742, 12743, 12747, 12748, 12746, 12739, 12728, 12738, 12743, 12733, 12734, 12737, 12746, 12736, 12730, 12736, 12729, 12736, 12726, 12729,
12749, 12739, 12741, 12735, 12734, 12732, 12748, 12743, 12746, 12746, 12734, 12736, 12747, 12730, 12739, 12753, 12750, 12751, 12757, 12754, 12755, 12761, 12758, 12759, 12765, 12762, 12763, 12759, 12752, 12753, 12763, 12756, 12757, 12756, 12751, 12754, 12764, 12759, 12762, 12762, 12753, 12756, 12769, 12766, 12767, 12773, 12770, 12771, 12777, 12774, 12775, 12781, 12778, 12779, 12775, 12768, 12769, 12779, 12772, 12773, 12772, 12767, 12770, 12780, 12775, 12778, 12778, 12769, 12772, 12751, 12719, 12720, 12755, 12723, 12724, 12754, 12720, 12723, 12767, 12744, 12745, 12771, 12748, 12749, 12770, 12745, 12748, 12733, 12714, 12731, 12744, 12722, 12742, 12742, 12716, 12733, 12768, 12755, 12766, 12776, 12763, 12774, 12774, 12757, 12768, 12766, 12724, 12744, 12783, 12784, 12782, 12786, 12785, 12783, 12793, 12790, 12791, 12797, 12794, 12795, 12801, 12798, 12800, 12801, 12802, 12799, 12810, 12807, 12808, 12789, 12798, 12788, 12793, 12804, 12792, 12808, 12796, 12797, 12805, 12800, 12804, 12805, 12803, 12801, 12796, 12791, 12794, 12807, 12806, 12805, 12807, 12793, 12796, 12789, 12804, 12800, 12814, 12811, 12812, 12818, 12815, 12816, 12822, 12819, 12820, 12826, 12823, 12824, 12820, 12813, 12814, 12824, 12817, 12818, 12817, 12812, 12815, 12825, 12820, 12823,
12823, 12814, 12817, 12834, 12831, 12833, 12830, 12831, 12829, 12837, 12838, 12836, 12841, 12842, 12840, 12834, 12840, 12832, 12839, 12833, 12838, 12830, 12838, 12833, 12844, 12843, 12841, 12830, 12827, 12828, 12836, 12828, 12835, 12839, 12841, 12834, 12849, 12846, 12847, 12853, 12850, 12851, 12859, 12856, 12857, 12863, 12860, 12861, 12857, 12848, 12849, 12861, 12852, 12853, 12852, 12847, 12850, 12862, 12857, 12860, 12860, 12849, 12852, 12803, 12827, 12802, 12837, 12809, 12810, 12809, 12835, 12806, 12806, 12828, 12803, 12847, 12821, 12822, 12851, 12825, 12826, 12850, 12822, 12825, 12813, 12795, 12811, 12821, 12808, 12819, 12819, 12797, 12813, 12848, 12837, 12846, 12855, 12842, 12843, 12856, 12845, 12844, 12845, 12855, 12843, 12848, 12844, 12839, 12846, 12810, 12821, 12704, 12782, 12702, 12708, 12783, 12704, 12791, 12760, 12761, 12795, 12764, 12765, 12794, 12761, 12764, 12812, 12776, 12777, 12816, 12780, 12781, 12815, 12777, 12780, 12811, 12765, 12776, 12711, 12670, 12709, 12719, 12684, 12717, 12717, 12672, 12711, 12752, 12696, 12750, 12760, 12706, 12758, 12758, 12698, 12752, 12750, 12686, 12719, 12789, 12784, 12785, 12790, 12787, 12786, 12792, 12785, 12787, 12760, 12786, 12708, 12869, 12866, 12868, 12865, 12866, 12864, 12871, 12872, 12870,
12877, 12878, 12876, 12869, 12876, 12867, 12873, 12868, 12872, 12870, 12868, 12865, 12881, 12879, 12877, 12873, 12877, 12869, 12886, 12883, 12884, 12890, 12887, 12888, 12896, 12893, 12894, 12900, 12897, 12898, 12894, 12885, 12886, 12898, 12889, 12890, 12889, 12884, 12887, 12899, 12894, 12897, 12897, 12886, 12889, 12904, 12905, 12903, 12905, 12901, 12903, 12915, 12912, 12914, 12919, 12916, 12917, 12911, 12904, 12910, 12909, 12916, 12908, 12906, 12908, 12905, 12908, 12902, 12905, 12921, 12913, 12915, 12912, 12911, 12910, 12920, 12919, 12917, 12913, 12917, 12911, 12911, 12909, 12906, 12925, 12922, 12924, 12925, 12926, 12923, 12934, 12931, 12932, 12942, 12939, 12941, 12938, 12939, 12937, 12945, 12946, 12944, 12927, 12935, 12926, 12945, 12933, 12934, 12929, 12924, 12928, 12929, 12927, 12925, 12947, 12941, 12946, 12938, 12946, 12941, 12931, 12930, 12929, 12938, 12935, 12936, 12944, 12936, 12943, 12933, 12943, 12930, 12930, 12936, 12927, 12902, 12874, 12901, 12907, 12875, 12902, 12892, 12922, 12891, 12896, 12928, 12895, 12932, 12899, 12900, 12931, 12896, 12899, 12892, 12928, 12924, 12885, 12871, 12883, 12892, 12878, 12879, 12893, 12882, 12881, 12882, 12892, 12879, 12885, 12881, 12873, 12953, 12950, 12951, 12957, 12954, 12955, 12961, 12958, 12959,
12965, 12962, 12963, 12959, 12952, 12953, 12963, 12956, 12957, 12956, 12951, 12954, 12964, 12959, 12962, 12962, 12953, 12956, 12969, 12966, 12968, 12976, 12977, 12975, 12980, 12976, 12979, 12984, 12985, 12983, 12977, 12968, 12975, 12973, 12967, 12969, 12978, 12983, 12977, 12980, 12984, 12978, 12981, 12980, 12979, 12982, 12986, 12980, 12983, 12969, 12977, 12991, 12988, 12989, 12995, 12992, 12993, 12999, 12996, 12997, 13003, 13000, 13001, 12997, 12990, 12991, 13001, 12994, 12995, 12994, 12989, 12992, 13002, 12997, 13000, 13000, 12991, 12994, 13007, 13008, 13006, 13008, 13004, 13006, 13013, 13010, 13011, 13020, 13017, 13019, 13024, 13021, 13022, 13026, 13027, 13025, 13018, 13007, 13017, 13022, 13012, 13013, 13016, 13025, 13015, 13009, 13015, 13008, 13015, 13005, 13008, 13028, 13018, 13020, 13014, 13013, 13011, 13027, 13022, 13025, 13025, 13013, 13015, 13026, 13009, 13018, 12989, 12960, 12961, 12993, 12964, 12965, 12992, 12961, 12964, 13005, 12981, 13004, 13011, 12985, 12986, 12987, 13011, 12986, 13014, 12982, 13005, 12966, 12949, 12948, 12972, 12957, 12955, 12985, 12963, 12983, 12967, 12955, 12949, 12973, 12963, 12957, 13012, 12993, 13010, 13023, 13001, 13021, 13021, 12995, 13012, 13010, 12965, 12985, 13030, 13031, 13029, 13034, 13030, 13033,
13038, 13035, 13036, 13040, 13041, 13039, 13046, 13043, 13044, 13050, 13047, 13048, 13044, 13037, 13038, 13048, 13041, 13042, 13032, 13039, 13031, 13034, 13040, 13032, 13041, 13036, 13039, 13049, 13044, 13047, 13047, 13038, 13041, 13054, 13055, 13053, 13063, 13064, 13062, 13064, 13060, 13062, 13061, 13051, 13060, 13065, 13067, 13064, 13067, 13061, 13064, 13058, 13056, 13054, 13066, 13052, 13061, 13072, 13069, 13070, 13076, 13073, 13074, 13080, 13077, 13078, 13084, 13081, 13082, 13078, 13071, 13072, 13082, 13075, 13076, 13075, 13070, 13073, 13083, 13078, 13081, 13081, 13072, 13075, 13088, 13085, 13087, 13092, 13093, 13091, 13097, 13098, 13096, 13101, 13097, 13100, 13105, 13102, 13103, 13107, 13108, 13106, 13098, 13089, 13096, 13103, 13093, 13094, 13094, 13086, 13088, 13099, 13106, 13098, 13101, 13107, 13099, 13089, 13088, 13087, 13108, 13103, 13106, 13090, 13094, 13088, 13095, 13103, 13094, 13098, 13095, 13090, 13070, 13045, 13046, 13074, 13049, 13050, 13073, 13046, 13049, 13086, 13063, 13085, 13068, 13091, 13067, 13092, 13065, 13086, 13051, 13034, 13033, 13066, 13050, 13048, 13034, 13057, 13042, 13057, 13048, 13042, 13093, 13074, 13091, 13104, 13082, 13102, 13102, 13076, 13093, 13091, 13050, 13067, 13115, 13112, 13113, 13119, 13116, 13117,
13123, 13120, 13122, 13123, 13124, 13121, 13132, 13129, 13130, 13111, 13120, 13110, 13115, 13126, 13114, 13130, 13118, 13119, 13127, 13122, 13126, 13127, 13125, 13123, 13118, 13113, 13116, 13129, 13128, 13127, 13129, 13115, 13118, 13111, 13126, 13122, 13136, 13133, 13134, 13140, 13137, 13138, 13144, 13141, 13142, 13148, 13145, 13146, 13142, 13135, 13136, 13146, 13139, 13140, 13139, 13134, 13137, 13147, 13142, 13145, 13145, 13136, 13139, 13156, 13153, 13155, 13152, 13153, 13151, 13159, 13160, 13158, 13165, 13166, 13164, 13156, 13164, 13154, 13161, 13155, 13160, 13152, 13160, 13155, 13169, 13167, 13165, 13152, 13149, 13150, 13158, 13150, 13157, 13161, 13165, 13156, 13174, 13171, 13172, 13178, 13175, 13176, 13184, 13181, 13182, 13188, 13185, 13186, 13182, 13173, 13174, 13186, 13177, 13178, 13177, 13172, 13175, 13187, 13182, 13185, 13185, 13174, 13177, 13125, 13149, 13124, 13159, 13131, 13132, 13131, 13157, 13128, 13128, 13150, 13125, 13172, 13143, 13144, 13176, 13147, 13148, 13175, 13144, 13147, 13135, 13117, 13133, 13143, 13130, 13141, 13141, 13119, 13135, 13173, 13159, 13171, 13180, 13166, 13167, 13181, 13170, 13169, 13170, 13180, 13167, 13173, 13169, 13161, 13171, 13132, 13143, 13031, 12914, 13029, 13036, 12918, 12919, 12921, 13036, 12919,
13039, 12915, 13031, 12942, 13053, 12940, 12947, 13054, 12942, 13113, 12998, 12999, 13117, 13002, 13003, 13116, 12999, 13002, 13134, 13023, 13024, 13138, 13027, 13028, 13137, 13024, 13027, 13133, 13003, 13023, 12952, 12888, 12950, 12960, 12898, 12958, 12958, 12890, 12952, 12990, 12932, 12988, 12998, 12945, 12996, 12996, 12934, 12990, 12988, 12900, 12960, 13111, 13055, 13056, 13112, 13059, 13058, 13114, 13056, 13059, 13162, 13101, 13100, 13101, 13168, 13109, 12998, 13058, 12947, 13192, 13189, 13191, 13192, 13193, 13190, 13203, 13200, 13201, 13207, 13204, 13206, 13207, 13208, 13205, 13196, 13204, 13195, 13203, 13210, 13202, 13198, 13191, 13197, 13192, 13199, 13194, 13211, 13206, 13210, 13211, 13209, 13207, 13196, 13193, 13194, 13200, 13199, 13198, 13202, 13194, 13199, 13196, 13210, 13206, 13209, 13213, 13208, 13212, 13214, 13209, 13219, 13216, 13217, 13223, 13220, 13221, 13227, 13224, 13225, 13231, 13228, 13229, 13225, 13218, 13219, 13229, 13222, 13223, 13222, 13217, 13220, 13230, 13225, 13228, 13228, 13219, 13222, 13235, 13232, 13233, 13239, 13236, 13237, 13243, 13240, 13241, 13247, 13244, 13245, 13241, 13234, 13235, 13245, 13238, 13239, 13238, 13233, 13236, 13246, 13241, 13244, 13244, 13235, 13238, 13253, 13250, 13252, 13249, 13250, 13248,
13255, 13256, 13254, 13261, 13258, 13259, 13265, 13266, 13264, 13270, 13271, 13269, 13253, 13262, 13251, 13270, 13260, 13261, 13257, 13252, 13256, 13249, 13256, 13252, 13269, 13267, 13265, 13260, 13255, 13258, 13265, 13262, 13263, 13269, 13263, 13268, 13260, 13268, 13257, 13253, 13268, 13263, 13276, 13273, 13274, 13280, 13277, 13278, 13284, 13281, 13282, 13288, 13285, 13286, 13282, 13275, 13276, 13286, 13279, 13280, 13279, 13274, 13277, 13287, 13282, 13285, 13285, 13276, 13279, 13255, 13226, 13227, 13259, 13230, 13231, 13258, 13227, 13230, 13274, 13242, 13243, 13278, 13246, 13247, 13277, 13243, 13246, 13234, 13221, 13232, 13242, 13229, 13240, 13240, 13223, 13234, 13275, 13259, 13273, 13283, 13270, 13281, 13281, 13261, 13275, 13273, 13231, 13242, 13292, 13293, 13291, 13293, 13289, 13291, 13303, 13300, 13302, 13307, 13304, 13305, 13299, 13292, 13298, 13297, 13304, 13296, 13294, 13296, 13293, 13296, 13290, 13293, 13309, 13301, 13303, 13300, 13299, 13298, 13308, 13307, 13305, 13301, 13305, 13299, 13299, 13297, 13294, 13313, 13310, 13312, 13313, 13314, 13311, 13324, 13321, 13322, 13330, 13327, 13329, 13326, 13327, 13325, 13332, 13333, 13331, 13317, 13325, 13316, 13324, 13331, 13323, 13319, 13312, 13318, 13313, 13320, 13315, 13334, 13329, 13333,
13331, 13329, 13326, 13317, 13314, 13315, 13321, 13320, 13319, 13323, 13315, 13320, 13323, 13326, 13317, 13336, 13337, 13335, 13340, 13336, 13339, 13344, 13341, 13342, 13346, 13347, 13345, 13352, 13349, 13350, 13356, 13353, 13354, 13350, 13343, 13344, 13354, 13347, 13348, 13338, 13345, 13337, 13340, 13346, 13338, 13347, 13342, 13345, 13355, 13350, 13353, 13353, 13344, 13347, 13364, 13365, 13363, 13368, 13364, 13367, 13370, 13371, 13369, 13365, 13357, 13363, 13365, 13370, 13369, 13372, 13366, 13368, 13365, 13361, 13358, 13337, 13302, 13335, 13342, 13306, 13307, 13309, 13342, 13307, 13345, 13303, 13337, 13330, 13359, 13328, 13330, 13362, 13360, 13357, 13340, 13339, 13371, 13354, 13369, 13358, 13348, 13340, 13361, 13354, 13348, 13376, 13373, 13375, 13376, 13377, 13374, 13379, 13376, 13375, 13380, 13378, 13376, 13387, 13384, 13385, 13391, 13388, 13389, 13395, 13392, 13394, 13395, 13396, 13393, 13406, 13403, 13404, 13383, 13392, 13382, 13387, 13400, 13386, 13404, 13390, 13391, 13401, 13394, 13400, 13395, 13402, 13397, 13390, 13385, 13388, 13399, 13396, 13397, 13403, 13402, 13401, 13405, 13397, 13402, 13403, 13387, 13390, 13383, 13400, 13394, 13410, 13407, 13409, 13410, 13411, 13408, 13412, 13416, 13411, 13414, 13409, 13413, 13414, 13412, 13410,
13415, 13417, 13412, 13399, 13407, 13398, 13406, 13413, 13405, 13399, 13413, 13409, 13383, 13377, 13378, 13384, 13381, 13380, 13386, 13378, 13381, 13422, 13419, 13420, 13426, 13423, 13424, 13430, 13427, 13428, 13434, 13431, 13432, 13428, 13421, 13422, 13432, 13425, 13426, 13425, 13420, 13423, 13433, 13428, 13431, 13431, 13422, 13425, 13440, 13441, 13439, 13441, 13437, 13439, 13446, 13443, 13444, 13453, 13450, 13452, 13457, 13454, 13455, 13459, 13460, 13458, 13451, 13440, 13450, 13455, 13445, 13446, 13449, 13458, 13448, 13442, 13448, 13441, 13441, 13447, 13438, 13461, 13451, 13453, 13437, 13436, 13435, 13447, 13446, 13444, 13460, 13455, 13458, 13438, 13444, 13436, 13458, 13446, 13448, 13459, 13442, 13451, 13465, 13462, 13463, 13469, 13466, 13467, 13475, 13472, 13474, 13471, 13472, 13470, 13477, 13478, 13476, 13483, 13480, 13481, 13477, 13464, 13465, 13481, 13468, 13469, 13479, 13474, 13478, 13471, 13478, 13474, 13468, 13463, 13466, 13482, 13477, 13480, 13480, 13465, 13468, 13487, 13484, 13485, 13491, 13488, 13489, 13495, 13492, 13493, 13499, 13496, 13497, 13493, 13486, 13487, 13497, 13490, 13491, 13490, 13485, 13488, 13498, 13493, 13496, 13496, 13487, 13490, 13463, 13429, 13430, 13467, 13433, 13434, 13466, 13430, 13433, 13485, 13456, 13457,
13489, 13460, 13461, 13488, 13457, 13460, 13445, 13424, 13443, 13456, 13432, 13454, 13454, 13426, 13445, 13486, 13467, 13484, 13494, 13481, 13492, 13492, 13469, 13486, 13484, 13434, 13456, 13267, 13373, 13266, 13272, 13379, 13271, 13271, 13375, 13267, 13385, 13283, 13284, 13389, 13287, 13288, 13388, 13284, 13287, 13384, 13272, 13283, 13420, 13351, 13352, 13424, 13355, 13356, 13423, 13352, 13355, 13436, 13367, 13435, 13372, 13443, 13371, 13444, 13368, 13436, 13443, 13356, 13371, 13295, 13239, 13237, 13306, 13245, 13304, 13304, 13239, 13296, 13343, 13278, 13341, 13351, 13286, 13349, 13349, 13280, 13343, 13341, 13247, 13306, 13421, 13389, 13419, 13429, 13404, 13427, 13427, 13391, 13421, 13462, 13415, 13414, 13471, 13416, 13417, 13418, 13471, 13417, 13464, 13418, 13415, 13462, 13406, 13429, 13419, 13288, 13351, 12855, 13189, 12854, 12859, 13197, 12858, 13201, 12862, 12863, 13200, 12859, 12862, 12855, 13197, 13191, 13217, 13079, 13080, 13221, 13083, 13084, 13220, 13080, 13083, 13233, 13104, 13105, 13237, 13108, 13109, 13236, 13105, 13108, 13232, 13084, 13104, 13290, 13162, 13289, 13295, 13163, 13290, 13180, 13310, 13179, 13184, 13318, 13183, 13322, 13187, 13188, 13321, 13184, 13187, 13180, 13318, 13312, 13168, 13237, 13109, 12874, 12741, 12740,
12875, 12749, 12741, 12907, 12773, 12771, 12918, 12779, 12916, 12916, 12773, 12908, 12880, 12771, 12749, 13037, 12816, 13035, 13045, 12824, 13043, 13043, 12818, 13037, 13071, 12851, 13069, 13079, 12861, 13077, 13077, 12853, 13071, 13069, 12826, 13045, 13035, 12781, 12918, 13218, 13201, 13216, 13224, 13212, 13211, 13224, 13203, 13218, 13249, 13213, 13214, 13254, 13214, 13215, 13226, 13215, 13212, 13216, 12863, 13079, 13503, 13504, 13502, 13508, 13509, 13507, 13516, 13513, 13515, 13512, 13513, 13511, 13518, 13519, 13517, 13505, 13511, 13504, 13510, 13517, 13509, 13503, 13509, 13505, 13520, 13515, 13519, 13517, 13515, 13512, 13503, 13500, 13501, 13507, 13501, 13506, 13509, 13512, 13505, 13524, 13521, 13522, 13528, 13525, 13526, 13532, 13529, 13530, 13536, 13533, 13534, 13530, 13523, 13524, 13534, 13527, 13528, 13527, 13522, 13525, 13535, 13530, 13533, 13533, 13524, 13527, 13538, 13539, 13537, 13541, 13540, 13538, 13548, 13545, 13546, 13552, 13549, 13550, 13558, 13555, 13557, 13554, 13555, 13553, 13560, 13561, 13559, 13566, 13563, 13564, 13544, 13553, 13543, 13548, 13559, 13547, 13564, 13551, 13552, 13562, 13557, 13561, 13559, 13557, 13554, 13551, 13546, 13549, 13565, 13560, 13563, 13563, 13548, 13551, 13547, 13554, 13544, 13516, 13537, 13514,
13520, 13538, 13516, 13546, 13531, 13532, 13550, 13535, 13536, 13549, 13532, 13535, 13523, 13508, 13521, 13531, 13518, 13529, 13529, 13510, 13523, 13544, 13539, 13540, 13545, 13542, 13541, 13547, 13540, 13542, 13531, 13541, 13520, 13570, 13567, 13569, 13574, 13575, 13573, 13578, 13574, 13577, 13580, 13581, 13579, 13575, 13569, 13573, 13572, 13568, 13570, 13576, 13579, 13575, 13582, 13576, 13578, 13575, 13572, 13570, 13588, 13585, 13587, 13592, 13589, 13590, 13596, 13597, 13595, 13600, 13596, 13599, 13604, 13601, 13602, 13606, 13607, 13605, 13597, 13587, 13595, 13602, 13591, 13592, 13594, 13586, 13588, 13598, 13605, 13597, 13600, 13606, 13598, 13585, 13584, 13583, 13593, 13592, 13590, 13607, 13602, 13605, 13586, 13590, 13584, 13594, 13602, 13592, 13605, 13588, 13597, 13584, 13577, 13583, 13582, 13589, 13581, 13590, 13578, 13584, 13609, 13600, 13599, 13600, 13611, 13608, 13613, 13614, 13612, 13623, 13620, 13621, 13627, 13624, 13626, 13627, 13628, 13625, 13617, 13624, 13616, 13623, 13630, 13622, 13618, 13615, 13613, 13631, 13626, 13630, 13627, 13632, 13629, 13617, 13614, 13615, 13620, 13619, 13618, 13619, 13617, 13615, 13617, 13630, 13626, 13636, 13633, 13634, 13640, 13637, 13638, 13646, 13643, 13644, 13650, 13647, 13648, 13644, 13635, 13636,
13648, 13639, 13640, 13639, 13634, 13637, 13649, 13644, 13647, 13647, 13636, 13639, 13654, 13655, 13653, 13655, 13651, 13653, 13665, 13662, 13664, 13669, 13666, 13667, 13661, 13654, 13660, 13659, 13666, 13658, 13656, 13658, 13655, 13658, 13652, 13655, 13671, 13663, 13665, 13662, 13661, 13660, 13670, 13669, 13667, 13663, 13667, 13661, 13661, 13659, 13656, 13677, 13674, 13676, 13673, 13674, 13672, 13679, 13680, 13678, 13685, 13682, 13683, 13693, 13690, 13692, 13689, 13690, 13688, 13696, 13697, 13695, 13677, 13686, 13675, 13696, 13684, 13685, 13681, 13676, 13680, 13678, 13676, 13673, 13698, 13692, 13697, 13689, 13697, 13692, 13684, 13679, 13682, 13689, 13686, 13687, 13695, 13687, 13694, 13684, 13694, 13681, 13677, 13694, 13687, 13700, 13701, 13699, 13704, 13700, 13703, 13708, 13705, 13706, 13710, 13711, 13709, 13716, 13713, 13714, 13720, 13717, 13718, 13714, 13707, 13708, 13718, 13711, 13712, 13702, 13709, 13701, 13704, 13710, 13702, 13711, 13706, 13709, 13719, 13714, 13717, 13717, 13708, 13711, 13642, 13672, 13641, 13646, 13678, 13645, 13683, 13649, 13650, 13682, 13646, 13649, 13645, 13673, 13642, 13701, 13664, 13699, 13706, 13668, 13669, 13671, 13706, 13669, 13709, 13665, 13701, 13657, 13640, 13638, 13668, 13648, 13666, 13666, 13640, 13658,
13707, 13683, 13705, 13715, 13696, 13713, 13713, 13685, 13707, 13705, 13650, 13668, 13558, 13612, 13556, 13621, 13565, 13566, 13565, 13618, 13562, 13562, 13613, 13558, 13634, 13603, 13604, 13638, 13607, 13608, 13637, 13604, 13607, 13652, 13609, 13651, 13657, 13610, 13652, 13611, 13638, 13608, 13571, 13528, 13526, 13581, 13534, 13579, 13572, 13534, 13528, 13591, 13550, 13589, 13603, 13564, 13601, 13601, 13552, 13591, 13589, 13536, 13581, 13635, 13621, 13633, 13642, 13628, 13629, 13643, 13632, 13631, 13645, 13629, 13632, 13643, 13623, 13635, 13633, 13566, 13603, 13729, 13726, 13728, 13725, 13726, 13724, 13731, 13732, 13730, 13737, 13734, 13735, 13745, 13742, 13744, 13741, 13742, 13740, 13748, 13749, 13747, 13729, 13738, 13727, 13748, 13736, 13737, 13732, 13729, 13728, 13725, 13732, 13728, 13749, 13745, 13744, 13747, 13744, 13741, 13736, 13731, 13734, 13741, 13738, 13739, 13746, 13741, 13739, 13736, 13746, 13733, 13733, 13739, 13729, 13745, 13751, 13743, 13750, 13752, 13745, 13725, 13721, 13722, 13730, 13722, 13723, 13693, 13721, 13691, 13693, 13723, 13722, 13731, 13715, 13716, 13735, 13719, 13720, 13734, 13716, 13719, 13715, 13723, 13698, 13758, 13759, 13757, 13762, 13758, 13761, 13766, 13767, 13765, 13759, 13754, 13757, 13760, 13765, 13759,
13762, 13766, 13760, 13763, 13762, 13761, 13762, 13769, 13768, 13759, 13756, 13755, 13777, 13774, 13776, 13773, 13774, 13772, 13780, 13781, 13779, 13784, 13785, 13783, 13777, 13783, 13775, 13782, 13776, 13781, 13773, 13781, 13776, 13787, 13786, 13784, 13773, 13770, 13771, 13779, 13771, 13778, 13782, 13784, 13777, 13792, 13789, 13790, 13796, 13793, 13794, 13802, 13799, 13800, 13806, 13803, 13804, 13800, 13791, 13792, 13804, 13795, 13796, 13795, 13790, 13793, 13805, 13800, 13803, 13803, 13792, 13795, 13810, 13807, 13809, 13810, 13811, 13808, 13821, 13818, 13819, 13827, 13824, 13826, 13823, 13824, 13822, 13829, 13830, 13828, 13814, 13822, 13813, 13821, 13828, 13820, 13816, 13809, 13815, 13810, 13817, 13812, 13831, 13826, 13830, 13828, 13826, 13823, 13814, 13811, 13812, 13818, 13817, 13816, 13820, 13812, 13817, 13820, 13823, 13814, 13798, 13807, 13797, 13802, 13815, 13801, 13819, 13805, 13806, 13818, 13802, 13805, 13798, 13815, 13809, 13791, 13780, 13789, 13798, 13785, 13786, 13799, 13788, 13787, 13801, 13786, 13788, 13791, 13787, 13782, 13827, 13832, 13825, 13827, 13834, 13833, 13836, 13837, 13835, 13840, 13836, 13839, 13844, 13841, 13842, 13846, 13847, 13845, 13852, 13849, 13850, 13856, 13853, 13854, 13850, 13843, 13844, 13854, 13847, 13848,
13837, 13846, 13845, 13848, 13838, 13840, 13847, 13842, 13845, 13855, 13850, 13853, 13853, 13844, 13847, 13858, 13859, 13857, 13868, 13869, 13867, 13869, 13865, 13867, 13874, 13871, 13872, 13864, 13858, 13863, 13862, 13871, 13861, 13860, 13861, 13859, 13869, 13877, 13876, 13876, 13866, 13869, 13865, 13864, 13863, 13875, 13874, 13872, 13864, 13875, 13872, 13864, 13862, 13860, 13883, 13880, 13882, 13879, 13880, 13878, 13887, 13888, 13886, 13893, 13890, 13891, 13897, 13894, 13896, 13897, 13898, 13895, 13885, 13894, 13884, 13893, 13900, 13892, 13889, 13882, 13888, 13879, 13888, 13882, 13900, 13897, 13896, 13901, 13899, 13897, 13885, 13881, 13883, 13890, 13889, 13887, 13892, 13883, 13889, 13892, 13896, 13885, 13908, 13905, 13906, 13912, 13909, 13910, 13916, 13913, 13915, 13918, 13919, 13917, 13924, 13921, 13922, 13918, 13907, 13908, 13922, 13911, 13912, 13920, 13915, 13919, 13911, 13906, 13909, 13923, 13918, 13921, 13921, 13908, 13911, 13887, 13851, 13852, 13891, 13855, 13856, 13890, 13852, 13855, 13904, 13868, 13903, 13906, 13873, 13874, 13877, 13909, 13876, 13909, 13874, 13876, 13910, 13870, 13904, 13857, 13840, 13839, 13873, 13854, 13871, 13859, 13848, 13840, 13861, 13854, 13848, 13907, 13891, 13905, 13915, 13898, 13899, 13917, 13902, 13901,
13919, 13899, 13902, 13917, 13893, 13907, 13905, 13856, 13873, 13926, 13927, 13925, 13930, 13926, 13929, 13932, 13933, 13931, 13936, 13937, 13935, 13946, 13943, 13944, 13950, 13947, 13948, 13942, 13932, 13941, 13944, 13937, 13938, 13940, 13947, 13939, 13928, 13935, 13927, 13938, 13928, 13930, 13934, 13939, 13933, 13931, 13930, 13929, 13949, 13944, 13947, 13933, 13938, 13930, 13939, 13944, 13938, 13948, 13934, 13942, 13952, 13953, 13951, 13956, 13952, 13955, 13958, 13959, 13957, 13962, 13963, 13961, 13954, 13961, 13953, 13964, 13954, 13956, 13959, 13966, 13965, 13957, 13956, 13955, 13959, 13964, 13956, 13951, 13942, 13941, 13961, 13950, 13948, 13953, 13948, 13942, 13974, 13971, 13973, 13970, 13971, 13969, 13977, 13978, 13976, 13979, 13973, 13978, 13970, 13978, 13973, 13970, 13967, 13968, 13975, 13970, 13968, 13985, 13982, 13983, 13989, 13986, 13987, 13993, 13990, 13992, 13997, 13994, 13996, 13997, 13998, 13995, 13981, 13990, 13980, 13985, 14000, 13984, 13989, 14002, 13988, 14001, 13992, 14000, 14002, 13997, 13996, 14003, 13999, 13997, 13988, 13983, 13986, 13996, 13991, 13993, 14002, 13993, 14001, 13988, 14001, 13985, 13984, 13992, 13981, 14010, 14007, 14008, 14014, 14011, 14012, 14018, 14015, 14017, 14020, 14021, 14019, 14026, 14023, 14024,
14020, 14009, 14010, 14024, 14013, 14014, 14022, 14017, 14021, 14013, 14008, 14011, 14025, 14020, 14023, 14023, 14010, 14013, 14032, 14029, 14031, 14035, 14031, 14034, 14031, 14027, 14028, 14034, 14028, 14033, 14018, 14027, 14016, 14026, 14034, 14025, 14025, 14033, 14022, 14022, 14028, 14018, 14009, 13987, 14007, 14017, 13998, 13999, 14019, 14004, 14003, 14004, 14017, 13999, 14019, 13989, 14009, 13916, 13967, 13914, 13977, 13923, 13924, 13923, 13975, 13920, 13920, 13968, 13916, 13983, 13945, 13946, 13987, 13949, 13950, 13986, 13946, 13949, 14006, 13958, 14005, 14008, 13963, 13964, 13966, 14011, 13965, 13965, 14008, 13964, 14012, 13960, 14006, 14007, 13950, 13963, 13925, 13904, 13903, 13935, 13912, 13910, 13945, 13922, 13943, 13927, 13910, 13904, 13943, 13912, 13937, 13981, 13972, 13974, 13982, 13979, 13977, 13984, 13974, 13979, 13982, 13924, 13945, 14037, 14038, 14036, 14041, 14042, 14040, 14045, 14041, 14044, 14047, 14048, 14046, 14051, 14052, 14050, 14057, 14054, 14055, 14061, 14058, 14059, 14055, 14048, 14049, 14059, 14052, 14053, 14039, 14046, 14038, 14042, 14051, 14050, 14053, 14043, 14045, 14040, 14039, 14037, 14050, 14049, 14047, 14060, 14055, 14058, 14042, 14047, 14039, 14058, 14049, 14052, 14063, 14064, 14062, 14067, 14068, 14066,
14079, 14076, 14077, 14083, 14080, 14081, 14075, 14067, 14074, 14071, 14076, 14070, 14073, 14080, 14072, 14065, 14070, 14064, 14068, 14073, 14072, 14066, 14065, 14063, 14082, 14077, 14080, 14068, 14071, 14065, 14072, 14077, 14071, 14081, 14069, 14075, 14087, 14084, 14086, 14087, 14088, 14085, 14093, 14090, 14092, 14098, 14099, 14097, 14094, 14087, 14086, 14095, 14089, 14087, 14100, 14092, 14099, 14092, 14088, 14089, 14097, 14096, 14095, 14096, 14092, 14089, 14104, 14101, 14103, 14108, 14109, 14107, 14114, 14111, 14112, 14118, 14115, 14117, 14106, 14115, 14105, 14114, 14119, 14113, 14110, 14103, 14109, 14119, 14118, 14117, 14106, 14102, 14104, 14111, 14110, 14108, 14110, 14106, 14104, 14113, 14117, 14106, 14057, 14094, 14056, 14098, 14060, 14061, 14097, 14057, 14060, 14108, 14078, 14079, 14112, 14082, 14083, 14111, 14079, 14082, 14062, 14045, 14044, 14078, 14059, 14076, 14064, 14053, 14045, 14070, 14059, 14053, 14103, 14091, 14093, 14107, 14100, 14098, 14109, 14093, 14100, 14107, 14061, 14078, 14122, 14123, 14121, 14126, 14127, 14125, 14130, 14131, 14129, 14136, 14133, 14134, 14124, 14129, 14123, 14128, 14133, 14127, 14125, 14124, 14122, 14135, 14130, 14133, 14127, 14130, 14124, 14138, 14139, 14137, 14142, 14143, 14141, 14146, 14147, 14145,
14150, 14151, 14149, 14140, 14145, 14139, 14143, 14150, 14149, 14141, 14140, 14138, 14149, 14148, 14146, 14140, 14149, 14146, 14156, 14153, 14154, 14160, 14157, 14158, 14164, 14161, 14163, 14168, 14165, 14167, 14156, 14169, 14155, 14172, 14159, 14160, 14170, 14163, 14169, 14172, 14167, 14171, 14159, 14154, 14157, 14167, 14162, 14164, 14171, 14164, 14170, 14171, 14156, 14159, 14176, 14173, 14174, 14180, 14177, 14178, 14184, 14181, 14183, 14188, 14185, 14187, 14190, 14175, 14176, 14180, 14191, 14179, 14190, 14183, 14189, 14191, 14188, 14187, 14179, 14174, 14177, 14187, 14182, 14184, 14190, 14187, 14184, 14191, 14176, 14179, 14154, 14131, 14132, 14158, 14135, 14136, 14157, 14132, 14135, 14174, 14147, 14148, 14178, 14151, 14152, 14177, 14148, 14151, 14137, 14128, 14126, 14147, 14134, 14145, 14139, 14134, 14128, 14175, 14158, 14173, 14183, 14166, 14168, 14189, 14168, 14172, 14189, 14160, 14175, 14173, 14136, 14147, 14121, 14075, 14074, 14129, 14083, 14081, 14075, 14129, 14081, 14155, 14112, 14153, 14163, 14116, 14118, 14169, 14118, 14120, 14155, 14120, 14114, 14153, 14083, 14131, 13837, 13763, 13835, 13842, 13767, 13768, 13769, 13842, 13768, 13845, 13764, 13837, 14036, 14006, 14005, 14046, 14014, 14012, 14056, 14024, 14054, 14006, 14046, 14012,
14054, 14014, 14048, 14086, 14030, 14032, 14094, 14032, 14035, 14056, 14035, 14026, 12621, 13500, 12620, 13508, 12626, 12627, 12626, 13506, 12623, 12621, 13506, 13501, 13522, 12648, 12649, 13526, 12652, 12653, 13525, 12649, 12652, 13521, 12627, 12648, 13568, 12654, 13567, 13571, 12655, 13568, 12656, 13526, 12653, 13475, 13770, 13473, 13780, 13482, 13483, 13482, 13778, 13479, 13475, 13778, 13771, 13790, 13494, 13495, 13794, 13498, 13499, 13793, 13495, 13498, 13789, 13483, 13494, 13754, 13704, 13703, 13767, 13718, 13765, 13755, 13712, 13704, 13756, 13718, 13712, 13843, 13735, 13841, 13851, 13748, 13849, 13849, 13737, 13843, 13879, 13751, 13752, 13753, 13879, 13752, 13851, 13753, 13750, 13841, 13720, 13767, 11105, 12279, 11104, 12284, 11108, 11109, 11105, 12283, 12280, 12301, 11120, 11121, 12305, 11124, 11125, 12304, 11121, 11124, 12300, 11109, 11120, 12367, 11183, 12366, 11188, 12368, 11187, 12372, 11188, 11191, 12367, 11191, 11184, 12368, 11125, 11187, 11929, 12657, 11927, 11933, 12658, 11929, 12666, 11944, 11945, 12670, 11948, 11949, 12669, 11945, 11948, 11944, 12661, 11933, 12710, 12006, 12007, 12714, 12010, 12011, 12713, 12007, 12010, 12726, 12025, 12725, 12732, 12029, 12030, 12031, 12732, 12030, 12735, 12026, 12726, 12731, 12011, 12029,
12709, 11949, 12006, 12165, 12864, 12164, 12171, 12870, 12170, 12170, 12865, 12165, 12884, 12182, 12183, 12888, 12186, 12187, 12887, 12183, 12186, 12883, 12171, 12182, 12949, 12226, 12948, 12951, 12230, 12231, 12235, 12954, 12234, 12954, 12231, 12234, 12955, 12227, 12949, 12245, 12970, 12244, 12245, 12974, 12971, 12950, 12187, 12230, 14200, 14201, 14199, 14204, 14205, 14203, 14214, 14211, 14212, 14194, 14199, 14193, 14205, 14195, 14203, 14209, 14202, 14200, 14206, 14211, 14205, 14195, 14194, 14193, 14208, 14201, 14202, 14211, 14210, 14209, 14194, 14198, 14197, 14210, 14208, 14202, 14198, 14209, 14197, 14197, 14200, 14194, 14211, 14196, 14205, 14218, 14219, 14217, 14219, 14215, 14217, 14219, 14223, 14222, 14222, 14216, 14219, 14227, 14224, 14226, 14227, 14228, 14225, 14231, 14226, 14230, 14231, 14229, 14227, 14238, 14239, 14237, 14242, 14238, 14241, 14246, 14243, 14244, 14248, 14249, 14247, 14254, 14251, 14253, 14260, 14257, 14259, 14256, 14257, 14255, 14264, 14265, 14263, 14236, 14251, 14235, 14246, 14261, 14245, 14264, 14249, 14250, 14262, 14253, 14261, 14240, 14247, 14239, 14250, 14240, 14242, 14266, 14259, 14265, 14263, 14259, 14256, 14237, 14234, 14233, 14247, 14246, 14244, 14256, 14252, 14254, 14239, 14244, 14234, 14263, 14254, 14262,
14249, 14262, 14246, 14236, 14261, 14253, 14208, 14224, 14207, 14214, 14230, 14213, 14208, 14230, 14226, 14234, 14218, 14233, 14223, 14243, 14222, 14244, 14220, 14234, 14215, 14206, 14204, 14221, 14214, 14212, 14216, 14212, 14206, 14236, 14228, 14229, 14243, 14232, 14231, 14232, 14236, 14229, 14243, 14214, 14222, 14268, 14269, 14267, 14280, 14281, 14279, 14281, 14277, 14279, 14286, 14283, 14284, 14274, 14268, 14273, 14272, 14283, 14271, 14270, 14271, 14269, 14282, 14288, 14281, 14288, 14278, 14281, 14277, 14274, 14273, 14287, 14286, 14284, 14274, 14287, 14284, 14284, 14270, 14274, 14290, 14282, 14280, 14291, 14289, 14282, 14267, 14242, 14241, 14276, 14258, 14260, 14283, 14266, 14264, 14269, 14250, 14242, 14285, 14260, 14266, 14271, 14264, 14250, 14294, 14295, 14293, 14300, 14301, 14299, 14306, 14303, 14304, 14295, 14299, 14293, 14304, 14297, 14298, 14296, 14297, 14295, 14303, 14302, 14300, 14297, 14300, 14295, 14308, 14309, 14307, 14314, 14311, 14312, 14318, 14315, 14317, 14310, 14315, 14309, 14320, 14313, 14314, 14311, 14310, 14308, 14320, 14317, 14319, 14313, 14317, 14310, 14302, 14307, 14301, 14312, 14305, 14306, 14305, 14308, 14302, 14322, 14323, 14321, 14326, 14327, 14325, 14336, 14333, 14334, 14340, 14337, 14338, 14334, 14329, 14330,
14338, 14331, 14332, 14324, 14329, 14323, 14328, 14331, 14327, 14325, 14324, 14322, 14339, 14334, 14337, 14327, 14330, 14324, 14337, 14330, 14331, 14342, 14343, 14341, 14346, 14347, 14345, 14356, 14353, 14354, 14360, 14357, 14358, 14354, 14349, 14350, 14358, 14351, 14352, 14344, 14349, 14343, 14348, 14351, 14347, 14345, 14344, 14342, 14359, 14354, 14357, 14347, 14350, 14344, 14357, 14350, 14351, 14364, 14361, 14362, 14368, 14365, 14366, 14372, 14369, 14371, 14376, 14373, 14375, 14378, 14363, 14364, 14380, 14367, 14368, 14378, 14371, 14377, 14380, 14375, 14379, 14367, 14362, 14365, 14375, 14370, 14372, 14379, 14372, 14378, 14379, 14364, 14367, 14384, 14381, 14382, 14388, 14385, 14386, 14392, 14389, 14391, 14396, 14393, 14395, 14398, 14383, 14384, 14400, 14387, 14388, 14398, 14391, 14397, 14400, 14395, 14399, 14387, 14382, 14385, 14395, 14390, 14392, 14399, 14392, 14398, 14399, 14384, 14387, 14362, 14335, 14336, 14366, 14339, 14340, 14365, 14336, 14339, 14382, 14355, 14356, 14386, 14359, 14360, 14385, 14356, 14359, 14341, 14328, 14326, 14355, 14338, 14353, 14343, 14332, 14328, 14353, 14332, 14349, 14383, 14366, 14381, 14391, 14374, 14376, 14397, 14376, 14380, 14397, 14368, 14383, 14381, 14340, 14355, 14321, 14296, 14294, 14335, 14304, 14333,
14323, 14298, 14296, 14333, 14298, 14329, 14363, 14312, 14361, 14371, 14316, 14318, 14377, 14318, 14320, 14377, 14314, 14363, 14361, 14306, 14335, 14402, 14403, 14401, 14408, 14405, 14407, 14414, 14415, 14413, 14422, 14423, 14421, 14423, 14417, 14421, 14420, 14422, 14418, 14404, 14413, 14403, 14419, 14407, 14417, 14409, 14404, 14402, 14425, 14416, 14414, 14412, 14406, 14408, 14424, 14429, 14423, 14429, 14419, 14423, 14420, 14430, 14424, 14405, 14402, 14401, 14411, 14410, 14409, 14429, 14425, 14427, 14406, 14409, 14402, 14427, 14410, 14412, 14410, 14414, 14404, 14427, 14408, 14419, 14437, 14438, 14436, 14438, 14440, 14436, 14446, 14447, 14445, 14432, 14437, 14431, 14439, 14445, 14438, 14445, 14442, 14438, 14440, 14435, 14434, 14435, 14447, 14444, 14432, 14446, 14439, 14450, 14451, 14449, 14451, 14453, 14449, 14460, 14457, 14458, 14462, 14463, 14461, 14470, 14467, 14469, 14470, 14465, 14468, 14474, 14471, 14473, 14478, 14475, 14476, 14474, 14455, 14472, 14460, 14475, 14459, 14480, 14463, 14464, 14452, 14457, 14451, 14451, 14459, 14454, 14477, 14470, 14469, 14470, 14476, 14466, 14479, 14474, 14473, 14463, 14458, 14461, 14473, 14465, 14466, 14476, 14473, 14466, 14479, 14460, 14463, 14474, 14464, 14456, 14416, 14450, 14415, 14462, 14429, 14430,
14461, 14426, 14429, 14426, 14452, 14416, 14453, 14432, 14431, 14469, 14441, 14443, 14475, 14448, 14446, 14454, 14433, 14432, 14448, 14469, 14443, 14475, 14433, 14459, 14484, 14481, 14483, 14488, 14489, 14487, 14495, 14496, 14494, 14499, 14500, 14498, 14505, 14502, 14504, 14509, 14506, 14507, 14513, 14510, 14511, 14486, 14499, 14485, 14497, 14510, 14496, 14492, 14483, 14491, 14500, 14507, 14506, 14494, 14490, 14488, 14512, 14505, 14504, 14488, 14482, 14484, 14489, 14486, 14485, 14510, 14509, 14507, 14492, 14488, 14484, 14490, 14493, 14486, 14510, 14493, 14496, 14486, 14507, 14501, 14517, 14514, 14516, 14523, 14520, 14522, 14521, 14516, 14520, 14519, 14515, 14517, 14523, 14524, 14521, 14521, 14519, 14517, 14529, 14526, 14528, 14531, 14526, 14530, 14534, 14529, 14528, 14531, 14534, 14528, 14533, 14527, 14529, 14535, 14533, 14529, 14533, 14502, 14532, 14535, 14508, 14509, 14537, 14509, 14512, 14533, 14512, 14504, 14518, 14497, 14495, 14523, 14503, 14505, 14525, 14511, 14524, 14513, 14523, 14505, 14524, 14497, 14519, 14543, 14540, 14542, 14545, 14543, 14542, 14542, 14538, 14539, 14544, 14542, 14539, 14435, 14538, 14434, 14543, 14441, 14541, 14448, 14545, 14447, 14545, 14444, 14447, 14444, 14539, 14435, 14543, 14448, 14443, 14498, 14420, 14418,
14508, 14428, 14506, 14420, 14506, 14428, 14531, 14455, 14456, 14536, 14462, 14534, 14536, 14456, 14464, 14508, 14462, 14430, 14550, 14551, 14549, 14557, 14558, 14556, 14567, 14564, 14565, 14548, 14556, 14547, 14565, 14554, 14555, 14562, 14559, 14557, 14552, 14554, 14551, 14549, 14548, 14547, 14564, 14563, 14562, 14551, 14553, 14548, 14554, 14562, 14553, 14553, 14557, 14548, 14569, 14570, 14568, 14573, 14574, 14572, 14583, 14580, 14582, 14587, 14584, 14586, 14591, 14588, 14589, 14595, 14592, 14593, 14589, 14576, 14577, 14593, 14578, 14579, 14571, 14576, 14570, 14591, 14582, 14590, 14575, 14578, 14574, 14595, 14586, 14594, 14572, 14571, 14569, 14586, 14581, 14583, 14594, 14589, 14592, 14574, 14577, 14571, 14594, 14583, 14591, 14592, 14577, 14578, 14597, 14598, 14596, 14603, 14600, 14602, 14609, 14606, 14607, 14611, 14612, 14610, 14617, 14614, 14616, 14623, 14620, 14621, 14599, 14610, 14598, 14615, 14602, 14614, 14621, 14608, 14609, 14604, 14599, 14597, 14618, 14613, 14611, 14609, 14601, 14603, 14623, 14615, 14617, 14606, 14605, 14604, 14620, 14619, 14618, 14608, 14618, 14605, 14605, 14611, 14599, 14621, 14603, 14615, 14559, 14596, 14558, 14601, 14560, 14600, 14607, 14566, 14567, 14566, 14604, 14563, 14563, 14597, 14559, 14607, 14561, 14601,
14568, 14552, 14550, 14582, 14560, 14561, 14588, 14567, 14565, 14570, 14555, 14552, 14590, 14561, 14567, 14588, 14555, 14576, 14625, 14626, 14624, 14629, 14630, 14628, 14639, 14636, 14638, 14643, 14640, 14642, 14647, 14644, 14645, 14649, 14650, 14648, 14645, 14632, 14633, 14649, 14634, 14635, 14627, 14632, 14626, 14647, 14638, 14646, 14631, 14634, 14630, 14651, 14642, 14650, 14628, 14627, 14625, 14642, 14637, 14639, 14650, 14645, 14648, 14630, 14633, 14627, 14650, 14639, 14647, 14648, 14633, 14634, 14653, 14654, 14652, 14661, 14658, 14660, 14665, 14662, 14663, 14659, 14653, 14658, 14663, 14656, 14657, 14655, 14656, 14654, 14665, 14659, 14661, 14663, 14655, 14659, 14667, 14668, 14666, 14673, 14674, 14672, 14669, 14672, 14668, 14667, 14671, 14669, 14673, 14677, 14675, 14669, 14676, 14673, 14681, 14678, 14680, 14685, 14682, 14683, 14689, 14686, 14688, 14691, 14692, 14690, 14687, 14680, 14686, 14685, 14690, 14684, 14681, 14683, 14679, 14689, 14691, 14687, 14687, 14685, 14681, 14643, 14666, 14641, 14643, 14670, 14667, 14679, 14660, 14678, 14683, 14664, 14665, 14679, 14665, 14661, 14652, 14631, 14629, 14662, 14651, 14649, 14654, 14635, 14631, 14662, 14635, 14656, 14682, 14671, 14670, 14692, 14676, 14690, 14690, 14671, 14684, 14664, 14670, 14651,
14698, 14700, 14696, 14703, 14697, 14701, 14706, 14707, 14705, 14714, 14711, 14713, 14710, 14711, 14709, 14716, 14717, 14715, 14702, 14709, 14700, 14716, 14707, 14708, 14698, 14707, 14702, 14703, 14706, 14699, 14718, 14713, 14717, 14715, 14713, 14710, 14698, 14694, 14695, 14705, 14695, 14704, 14707, 14710, 14702, 14723, 14724, 14722, 14727, 14723, 14726, 14729, 14730, 14728, 14733, 14734, 14732, 14724, 14719, 14722, 14724, 14733, 14732, 14735, 14725, 14727, 14731, 14736, 14730, 14728, 14727, 14726, 14730, 14735, 14727, 14732, 14720, 14724, 14746, 14743, 14745, 14742, 14743, 14741, 14750, 14747, 14749, 14752, 14753, 14751, 14756, 14757, 14755, 14754, 14745, 14753, 14742, 14753, 14745, 14758, 14749, 14757, 14749, 14744, 14746, 14755, 14754, 14752, 14757, 14746, 14754, 14714, 14738, 14712, 14718, 14739, 14714, 14752, 14734, 14735, 14737, 14755, 14736, 14736, 14752, 14735, 14719, 14703, 14701, 14734, 14716, 14732, 14720, 14708, 14703, 14721, 14716, 14708, 14742, 14738, 14739, 14751, 14739, 14740, 14734, 14740, 14718, 14762, 14763, 14761, 14760, 14762, 14759, 14771, 14772, 14770, 14769, 14771, 14768, 14782, 14783, 14781, 14763, 14774, 14761, 14782, 14766, 14767, 14773, 14779, 14772, 14769, 14780, 14773, 14763, 14767, 14766, 14760, 14767, 14764,
14774, 14769, 14768, 14783, 14778, 14781, 14775, 14778, 14769, 14766, 14775, 14763, 14790, 14787, 14789, 14790, 14785, 14788, 14794, 14791, 14792, 14798, 14795, 14797, 14796, 14787, 14795, 14793, 14790, 14789, 14794, 14786, 14790, 14798, 14799, 14796, 14799, 14789, 14796, 14804, 14801, 14803, 14808, 14805, 14807, 14810, 14805, 14809, 14814, 14811, 14812, 14813, 14804, 14803, 14816, 14807, 14815, 14810, 14815, 14807, 14810, 14802, 14804, 14817, 14812, 14815, 14817, 14804, 14814, 14808, 14776, 14806, 14812, 14779, 14780, 14816, 14783, 14784, 14783, 14812, 14780, 14816, 14777, 14808, 14791, 14767, 14765, 14798, 14776, 14777, 14800, 14782, 14799, 14784, 14798, 14777, 14793, 14782, 14767, 14613, 14694, 14612, 14699, 14616, 14697, 14706, 14622, 14623, 14622, 14704, 14619, 14613, 14704, 14695, 14706, 14617, 14699, 14675, 14759, 14674, 14675, 14765, 14760, 14786, 14688, 14785, 14792, 14692, 14693, 14786, 14693, 14689, 14791, 14677, 14692, 14624, 14575, 14573, 14638, 14585, 14587, 14646, 14593, 14644, 14626, 14579, 14575, 14646, 14587, 14595, 14644, 14579, 14632, 14770, 14731, 14729, 14731, 14779, 14737, 14803, 14748, 14750, 14811, 14758, 14756, 14758, 14803, 14750, 14811, 14737, 14779, 14819, 14820, 14818, 14825, 14826, 14824, 14831, 14828, 14829,
14820, 14824, 14818, 14829, 14822, 14823, 14821, 14822, 14820, 14828, 14827, 14825, 14822, 14825, 14820, 14833, 14834, 14832, 14839, 14836, 14837, 14841, 14842, 14840, 14847, 14844, 14845, 14835, 14840, 14834, 14845, 14838, 14839, 14836, 14835, 14833, 14844, 14843, 14841, 14838, 14841, 14835, 14827, 14832, 14826, 14837, 14830, 14831, 14830, 14833, 14827, 14849, 14850, 14848, 14853, 14854, 14852, 14863, 14860, 14861, 14867, 14864, 14865, 14861, 14856, 14857, 14865, 14858, 14859, 14851, 14856, 14850, 14855, 14858, 14854, 14852, 14851, 14849, 14866, 14861, 14864, 14854, 14857, 14851, 14864, 14857, 14858, 14869, 14870, 14868, 14873, 14874, 14872, 14883, 14880, 14881, 14887, 14884, 14885, 14881, 14876, 14877, 14885, 14878, 14879, 14871, 14876, 14870, 14875, 14878, 14874, 14872, 14871, 14869, 14886, 14881, 14884, 14874, 14877, 14871, 14884, 14877, 14878, 14891, 14888, 14889, 14895, 14892, 14893, 14899, 14896, 14897, 14903, 14900, 14901, 14897, 14890, 14891, 14901, 14894, 14895, 14894, 14889, 14892, 14902, 14897, 14900, 14900, 14891, 14894, 14907, 14904, 14905, 14911, 14908, 14909, 14915, 14912, 14914, 14919, 14916, 14918, 14921, 14916, 14920, 14925, 14922, 14923, 14923, 14906, 14907, 14927, 14910, 14911, 14925, 14914, 14924, 14927, 14918, 14926,
14921, 14926, 14918, 14910, 14905, 14908, 14921, 14913, 14915, 14928, 14923, 14926, 14928, 14915, 14925, 14910, 14923, 14907, 14889, 14862, 14863, 14893, 14866, 14867, 14892, 14863, 14866, 14905, 14882, 14883, 14909, 14886, 14887, 14908, 14883, 14886, 14868, 14855, 14853, 14882, 14865, 14880, 14870, 14859, 14855, 14880, 14859, 14876, 14906, 14893, 14904, 14922, 14903, 14901, 14922, 14895, 14906, 14904, 14867, 14882, 14848, 14821, 14819, 14862, 14829, 14860, 14850, 14823, 14821, 14860, 14823, 14856, 14890, 14837, 14888, 14898, 14845, 14896, 14896, 14839, 14890, 14888, 14831, 14862, 14276, 14481, 14275, 14286, 14491, 14285, 14495, 14288, 14289, 14494, 14286, 14288, 14285, 14483, 14276, 14515, 14290, 14514, 14515, 14292, 14291, 14292, 14495, 14289, 14930, 14931, 14929, 14934, 14935, 14933, 14942, 14939, 14940, 14937, 14932, 14930, 14936, 14939, 14935, 14933, 14930, 14929, 14939, 14938, 14937, 14935, 14937, 14930, 14944, 14945, 14943, 14948, 14949, 14947, 14954, 14951, 14952, 14958, 14955, 14956, 14946, 14951, 14945, 14950, 14955, 14949, 14947, 14946, 14944, 14957, 14952, 14955, 14949, 14952, 14946, 14943, 14936, 14934, 14953, 14940, 14951, 14945, 14940, 14936, 14960, 14961, 14959, 14964, 14965, 14963, 14970, 14967, 14968, 14974, 14971, 14972,
14962, 14967, 14961, 14966, 14971, 14965, 14963, 14962, 14960, 14973, 14968, 14971, 14965, 14968, 14962, 14976, 14977, 14975, 14980, 14976, 14979, 14984, 14981, 14982, 14978, 14981, 14977, 14984, 14978, 14980, 14975, 14966, 14964, 14983, 14972, 14981, 14977, 14972, 14966, 14986, 14987, 14985, 14992, 14989, 14991, 14998, 14995, 14996, 15000, 15001, 14999, 15006, 15003, 15005, 15010, 15011, 15009, 14988, 14999, 14987, 15004, 14991, 15003, 15010, 14997, 14998, 14993, 14988, 14986, 15007, 15002, 15000, 14998, 14990, 14992, 15012, 15004, 15006, 14995, 14994, 14993, 15011, 15007, 15009, 14997, 15007, 14994, 14994, 15000, 14988, 15010, 14992, 15004, 15016, 15013, 15015, 15020, 15017, 15019, 15022, 15015, 15021, 15024, 15019, 15023, 15019, 15014, 15016, 15023, 15016, 15022, 15026, 15027, 15025, 15034, 15029, 15032, 15040, 15037, 15038, 15043, 15045, 15041, 15050, 15047, 15048, 15033, 15041, 15031, 15044, 15032, 15042, 15040, 15047, 15039, 15026, 15036, 15028, 15040, 15030, 15034, 15047, 15046, 15043, 15033, 15027, 15028, 15037, 15036, 15035, 15039, 15028, 15036, 15033, 15047, 15043, 15044, 15040, 15034, 15054, 15051, 15053, 15056, 15052, 15054, 15002, 15025, 15001, 15030, 15005, 15029, 15038, 15011, 15012, 15037, 15008, 15011, 15008, 15026, 15002,
15038, 15006, 15030, 15015, 14989, 14990, 15021, 14990, 14996, 15051, 15044, 15042, 15055, 15050, 15048, 15052, 15048, 15044, 15060, 15057, 15059, 15064, 15061, 15063, 15064, 15065, 15062, 15073, 15074, 15072, 15066, 15072, 15065, 15068, 15059, 15067, 15070, 15063, 15069, 15064, 15071, 15066, 15073, 15077, 15075, 15063, 15058, 15060, 15069, 15060, 15068, 15066, 15076, 15073, 15081, 15078, 15080, 15083, 15084, 15082, 15089, 15086, 15088, 15091, 15092, 15090, 15087, 15080, 15086, 15091, 15084, 15085, 15081, 15083, 15079, 15089, 15091, 15087, 15087, 15085, 15081, 15095, 15096, 15094, 15103, 15104, 15102, 15101, 15103, 15100, 15097, 15100, 15096, 15095, 15099, 15097, 15104, 15108, 15107, 15106, 15105, 15101, 15097, 15106, 15101, 15112, 15109, 15111, 15114, 15115, 15113, 15122, 15119, 15121, 15122, 15117, 15120, 15126, 15123, 15124, 15118, 15111, 15117, 15124, 15115, 15116, 15112, 15114, 15110, 15125, 15122, 15121, 15126, 15118, 15122, 15118, 15116, 15112, 15075, 15094, 15074, 15075, 15098, 15095, 15110, 15088, 15109, 15114, 15092, 15093, 15110, 15093, 15089, 15082, 15071, 15070, 15092, 15076, 15090, 15084, 15076, 15071, 15115, 15098, 15113, 15123, 15108, 15106, 15123, 15099, 15115, 15113, 15077, 15092, 14932, 14985, 14931, 14942, 14995, 14941,
14941, 14993, 14938, 14938, 14986, 14932, 15022, 14953, 14954, 15024, 14957, 14958, 15023, 14954, 14957, 14953, 14996, 14942, 15068, 14969, 14970, 14974, 15069, 14973, 15069, 14970, 14973, 15079, 14979, 15078, 15083, 14983, 14984, 15083, 14980, 15079, 14983, 15070, 14974, 14959, 14950, 14948, 14969, 14956, 14967, 14961, 14956, 14950, 15059, 15018, 15020, 15067, 15020, 15024, 15067, 14958, 14969, 15128, 15129, 15127, 15134, 15131, 15132, 15136, 15137, 15135, 15142, 15139, 15140, 15130, 15135, 15129, 15140, 15133, 15134, 15131, 15130, 15128, 15139, 15138, 15136, 15133, 15136, 15130, 15144, 15145, 15143, 15150, 15147, 15148, 15152, 15153, 15151, 15158, 15155, 15156, 15146, 15151, 15145, 15156, 15149, 15150, 15147, 15146, 15144, 15155, 15154, 15152, 15149, 15152, 15146, 15138, 15143, 15137, 15148, 15141, 15142, 15141, 15144, 15138, 15162, 15159, 15160, 15166, 15163, 15164, 15170, 15167, 15168, 15174, 15171, 15172, 15168, 15161, 15162, 15172, 15165, 15166, 15165, 15160, 15163, 15173, 15168, 15171, 15171, 15162, 15165, 15178, 15175, 15177, 15184, 15181, 15183, 15182, 15177, 15181, 15178, 15179, 15176, 15184, 15185, 15182, 15182, 15180, 15178, 15190, 15187, 15188, 15194, 15191, 15192, 15198, 15195, 15196, 15202, 15199, 15200, 15196, 15189, 15190,
15200, 15193, 15194, 15193, 15188, 15191, 15201, 15196, 15199, 15199, 15190, 15193, 15206, 15203, 15205, 15212, 15209, 15211, 15210, 15205, 15209, 15208, 15204, 15206, 15214, 15210, 15212, 15213, 15206, 15210, 15188, 15169, 15170, 15192, 15173, 15174, 15191, 15170, 15173, 15204, 15183, 15203, 15204, 15186, 15184, 15180, 15164, 15179, 15185, 15174, 15172, 15185, 15166, 15180, 15208, 15192, 15207, 15214, 15200, 15213, 15213, 15194, 15208, 15186, 15192, 15174, 15216, 15217, 15215, 15222, 15219, 15220, 15224, 15225, 15223, 15230, 15227, 15228, 15218, 15223, 15217, 15228, 15221, 15222, 15219, 15218, 15216, 15227, 15226, 15224, 15221, 15224, 15218, 15232, 15233, 15231, 15238, 15235, 15236, 15240, 15241, 15239, 15246, 15243, 15244, 15234, 15239, 15233, 15244, 15237, 15238, 15235, 15234, 15232, 15243, 15242, 15240, 15237, 15240, 15234, 15226, 15231, 15225, 15236, 15229, 15230, 15229, 15232, 15226, 15250, 15247, 15248, 15254, 15251, 15252, 15258, 15255, 15256, 15262, 15259, 15260, 15256, 15249, 15250, 15260, 15253, 15254, 15253, 15248, 15251, 15261, 15256, 15259, 15259, 15250, 15253, 15266, 15263, 15265, 15272, 15269, 15271, 15270, 15265, 15269, 15268, 15264, 15266, 15274, 15270, 15272, 15273, 15266, 15270, 15278, 15275, 15276, 15282, 15279, 15280,
15286, 15283, 15284, 15290, 15287, 15288, 15284, 15277, 15278, 15288, 15281, 15282, 15281, 15276, 15279, 15289, 15284, 15287, 15287, 15278, 15281, 15294, 15291, 15293, 15300, 15297, 15299, 15298, 15293, 15297, 15296, 15292, 15294, 15302, 15298, 15300, 15301, 15294, 15298, 15276, 15257, 15258, 15280, 15261, 15262, 15279, 15258, 15261, 15292, 15271, 15291, 15295, 15272, 15292, 15268, 15252, 15267, 15274, 15260, 15273, 15273, 15254, 15268, 15296, 15280, 15295, 15302, 15288, 15301, 15301, 15282, 15296, 15295, 15262, 15274, 15154, 15215, 15153, 15220, 15157, 15158, 15157, 15216, 15154, 15248, 15197, 15198, 15252, 15201, 15202, 15251, 15198, 15201, 15264, 15211, 15263, 15267, 15212, 15264, 15267, 15202, 15214, 15161, 15132, 15159, 15169, 15140, 15167, 15167, 15134, 15161, 15189, 15148, 15187, 15197, 15156, 15195, 15195, 15150, 15189, 15187, 15142, 15169, 15249, 15220, 15247, 15257, 15228, 15255, 15255, 15222, 15249, 15277, 15236, 15275, 15285, 15244, 15283, 15283, 15238, 15277, 15275, 15230, 15257, 15247, 15158, 15197, 15306, 15303, 15305, 15306, 15307, 15304, 15310, 15305, 15309, 15306, 15311, 15308, 15313, 15314, 15312, 15321, 15322, 15320, 15322, 15318, 15320, 15327, 15324, 15325, 15334, 15331, 15333, 15338, 15335, 15337, 15317, 15331, 15316,
15327, 15339, 15326, 15330, 15341, 15329, 15315, 15324, 15314, 15340, 15333, 15339, 15322, 15330, 15329, 15329, 15319, 15322, 15342, 15337, 15341, 15318, 15315, 15313, 15328, 15327, 15325, 15337, 15332, 15334, 15315, 15328, 15325, 15341, 15334, 15340, 15341, 15327, 15329, 15326, 15333, 15317, 15317, 15307, 15308, 15326, 15310, 15324, 15311, 15317, 15308, 15344, 15345, 15343, 15345, 15347, 15343, 15354, 15351, 15353, 15358, 15355, 15356, 15360, 15361, 15359, 15366, 15363, 15365, 15368, 15351, 15367, 15358, 15369, 15357, 15345, 15356, 15355, 15345, 15357, 15348, 15369, 15366, 15365, 15361, 15354, 15353, 15349, 15346, 15344, 15359, 15358, 15356, 15368, 15364, 15366, 15346, 15359, 15356, 15371, 15366, 15370, 15371, 15358, 15361, 15368, 15361, 15353, 15375, 15372, 15374, 15375, 15376, 15373, 15375, 15352, 15354, 15377, 15360, 15376, 15362, 15375, 15354, 15381, 15382, 15380, 15382, 15386, 15385, 15380, 15379, 15378, 15382, 15384, 15379, 15390, 15391, 15389, 15388, 15390, 15387, 15397, 15398, 15396, 15403, 15404, 15402, 15391, 15396, 15389, 15403, 15394, 15395, 15392, 15394, 15391, 15388, 15395, 15392, 15397, 15404, 15400, 15391, 15402, 15397, 15407, 15408, 15406, 15415, 15412, 15414, 15417, 15418, 15416, 15423, 15420, 15421, 15427, 15424, 15426,
15429, 15424, 15428, 15427, 15410, 15425, 15431, 15418, 15419, 15408, 15417, 15416, 15431, 15426, 15430, 15429, 15430, 15426, 15422, 15415, 15414, 15414, 15410, 15411, 15420, 15419, 15417, 15419, 15414, 15411, 15427, 15419, 15411, 15379, 15407, 15378, 15421, 15385, 15386, 15420, 15384, 15385, 15379, 15417, 15409, 15406, 15388, 15387, 15418, 15393, 15416, 15429, 15399, 15401, 15432, 15403, 15430, 15388, 15416, 15393, 15405, 15429, 15401, 15418, 15403, 15395, 15434, 15435, 15433, 15436, 15439, 15435, 15437, 15436, 15434, 15438, 15440, 15436, 15445, 15442, 15444, 15447, 15448, 15446, 15451, 15447, 15450, 15453, 15454, 15452, 15457, 15458, 15456, 15467, 15464, 15466, 15463, 15464, 15462, 15470, 15471, 15469, 15445, 15460, 15443, 15470, 15458, 15459, 15454, 15445, 15444, 15448, 15457, 15456, 15451, 15457, 15449, 15472, 15466, 15471, 15463, 15471, 15466, 15458, 15453, 15456, 15463, 15460, 15461, 15469, 15461, 15468, 15458, 15468, 15455, 15455, 15461, 15445, 15479, 15480, 15478, 15480, 15476, 15478, 15477, 15473, 15476, 15481, 15483, 15480, 15483, 15477, 15480, 15482, 15474, 15477, 15448, 15437, 15446, 15440, 15452, 15439, 15441, 15453, 15440, 15456, 15438, 15448, 15473, 15451, 15450, 15483, 15470, 15482, 15474, 15459, 15451, 15475, 15470, 15459,
15347, 15323, 15321, 15365, 15336, 15338, 15348, 15330, 15323, 15342, 15365, 15338, 15369, 15330, 15357, 15433, 15383, 15381, 15435, 15386, 15383, 15444, 15413, 15415, 15452, 15423, 15421, 15454, 15415, 15423, 15439, 15421, 15386, 15486, 15487, 15485, 15492, 15489, 15490, 15494, 15495, 15493, 15500, 15497, 15498, 15488, 15493, 15487, 15498, 15491, 15492, 15489, 15488, 15486, 15497, 15496, 15494, 15491, 15494, 15488, 15506, 15503, 15505, 15502, 15503, 15501, 15510, 15507, 15508, 15510, 15505, 15509, 15507, 15505, 15502, 15496, 15501, 15495, 15508, 15499, 15500, 15499, 15502, 15496, 15514, 15511, 15512, 15518, 15515, 15516, 15522, 15519, 15520, 15526, 15523, 15524, 15520, 15513, 15514, 15524, 15517, 15518, 15517, 15512, 15515, 15525, 15520, 15523, 15523, 15514, 15517, 15530, 15527, 15529, 15536, 15533, 15535, 15534, 15529, 15533, 15532, 15528, 15530, 15538, 15534, 15536, 15537, 15530, 15534, 15542, 15539, 15541, 15546, 15543, 15545, 15550, 15547, 15548, 15554, 15551, 15552, 15550, 15541, 15549, 15554, 15545, 15553, 15545, 15540, 15542, 15553, 15548, 15551, 15553, 15542, 15550, 15558, 15555, 15557, 15560, 15556, 15558, 15548, 15521, 15522, 15552, 15525, 15526, 15551, 15522, 15525, 15556, 15535, 15555, 15559, 15536, 15556, 15532, 15516, 15531,
15538, 15524, 15537, 15537, 15518, 15532, 15558, 15544, 15546, 15560, 15552, 15559, 15560, 15546, 15554, 15559, 15526, 15538, 15562, 15563, 15561, 15564, 15565, 15563, 15568, 15569, 15567, 15574, 15571, 15572, 15576, 15577, 15575, 15582, 15579, 15580, 15570, 15575, 15569, 15580, 15573, 15574, 15571, 15570, 15568, 15579, 15578, 15576, 15573, 15576, 15570, 15563, 15567, 15561, 15572, 15565, 15566, 15565, 15568, 15563, 15584, 15585, 15583, 15588, 15589, 15587, 15586, 15591, 15585, 15590, 15593, 15589, 15587, 15586, 15584, 15589, 15592, 15586, 15596, 15597, 15595, 15600, 15601, 15599, 15598, 15603, 15597, 15602, 15605, 15601, 15599, 15598, 15596, 15601, 15604, 15598, 15610, 15607, 15608, 15614, 15611, 15612, 15618, 15615, 15616, 15622, 15619, 15620, 15616, 15609, 15610, 15620, 15613, 15614, 15613, 15608, 15611, 15621, 15616, 15619, 15619, 15610, 15613, 15626, 15623, 15624, 15630, 15627, 15628, 15634, 15631, 15632, 15638, 15635, 15636, 15632, 15625, 15626, 15636, 15629, 15630, 15629, 15624, 15627, 15637, 15632, 15635, 15635, 15626, 15629, 15608, 15591, 15592, 15612, 15593, 15594, 15611, 15592, 15593, 15624, 15603, 15604, 15628, 15605, 15606, 15627, 15604, 15605, 15595, 15590, 15588, 15597, 15594, 15590, 15625, 15612, 15623, 15633, 15620, 15631,
15631, 15614, 15625, 15623, 15594, 15603, 15513, 15490, 15511, 15521, 15498, 15519, 15519, 15492, 15513, 15541, 15504, 15506, 15549, 15508, 15547, 15549, 15506, 15510, 15547, 15500, 15521, 15583, 15564, 15562, 15585, 15566, 15564, 15609, 15572, 15607, 15617, 15580, 15615, 15615, 15574, 15609, 15607, 15566, 15591, 15046, 15303, 15045, 15050, 15309, 15049, 15049, 15305, 15046, 15314, 15053, 15312, 15324, 15054, 15314, 15324, 15050, 15056, 15104, 15349, 15102, 15360, 15107, 15108, 15107, 15350, 15104, 15373, 15119, 15372, 15373, 15125, 15121, 15125, 15360, 15108, 15242, 15485, 15241, 15490, 15245, 15246, 15245, 15486, 15242, 15512, 15285, 15286, 15516, 15289, 15290, 15515, 15286, 15289, 15528, 15299, 15527, 15531, 15300, 15528, 15531, 15290, 15302, 15511, 15246, 15285, 15640, 15641, 15639, 15648, 15644, 15647, 15652, 15653, 15651, 15656, 15657, 15655, 15662, 15659, 15661, 15668, 15665, 15666, 15642, 15655, 15641, 15660, 15647, 15659, 15666, 15653, 15654, 15649, 15642, 15640, 15663, 15658, 15656, 15648, 15652, 15646, 15668, 15660, 15662, 15643, 15640, 15639, 15653, 15649, 15651, 15667, 15663, 15665, 15640, 15651, 15649, 15665, 15650, 15653, 15642, 15663, 15656, 15660, 15654, 15648, 15670, 15671, 15669, 15678, 15673, 15676, 15684, 15681, 15682,
15690, 15687, 15689, 15686, 15687, 15685, 15692, 15693, 15691, 15677, 15685, 15675, 15692, 15683, 15684, 15670, 15680, 15672, 15678, 15682, 15674, 15694, 15689, 15693, 15691, 15689, 15686, 15677, 15671, 15672, 15681, 15680, 15679, 15683, 15672, 15680, 15683, 15686, 15677, 15701, 15702, 15700, 15702, 15698, 15700, 15699, 15695, 15698, 15703, 15707, 15702, 15707, 15699, 15702, 15704, 15703, 15701, 15705, 15708, 15703, 15706, 15696, 15699, 15658, 15669, 15657, 15674, 15661, 15673, 15682, 15667, 15668, 15667, 15679, 15664, 15664, 15670, 15658, 15682, 15662, 15674, 15695, 15678, 15676, 15707, 15692, 15706, 15696, 15684, 15678, 15697, 15692, 15684, 15711, 15712, 15710, 15717, 15718, 15716, 15713, 15716, 15712, 15714, 15713, 15711, 15717, 15721, 15719, 15715, 15717, 15713, 15725, 15722, 15724, 15727, 15728, 15726, 15733, 15730, 15732, 15735, 15736, 15734, 15731, 15724, 15730, 15735, 15728, 15729, 15729, 15723, 15725, 15733, 15735, 15731, 15735, 15725, 15731, 15739, 15740, 15738, 15747, 15748, 15746, 15748, 15750, 15746, 15756, 15753, 15754, 15741, 15747, 15740, 15739, 15743, 15741, 15748, 15754, 15753, 15753, 15751, 15748, 15750, 15745, 15744, 15751, 15752, 15745, 15743, 15749, 15741, 15760, 15757, 15759, 15764, 15761, 15762, 15768, 15765, 15767,
15770, 15765, 15769, 15768, 15759, 15766, 15772, 15763, 15764, 15760, 15762, 15758, 15771, 15768, 15767, 15770, 15771, 15767, 15772, 15760, 15768, 15719, 15738, 15718, 15719, 15742, 15739, 15758, 15732, 15757, 15737, 15761, 15736, 15758, 15737, 15733, 15726, 15715, 15714, 15736, 15720, 15734, 15728, 15720, 15715, 15763, 15742, 15761, 15771, 15756, 15754, 15763, 15754, 15743, 15761, 15721, 15736, 15782, 15779, 15781, 15778, 15779, 15777, 15784, 15785, 15783, 15790, 15787, 15789, 15792, 15793, 15791, 15796, 15797, 15795, 15793, 15782, 15781, 15791, 15781, 15778, 15785, 15796, 15795, 15798, 15789, 15797, 15789, 15780, 15782, 15797, 15792, 15795, 15794, 15789, 15782, 15778, 15774, 15775, 15776, 15778, 15775, 15800, 15801, 15799, 15806, 15803, 15805, 15812, 15809, 15811, 15812, 15807, 15810, 15816, 15813, 15814, 15820, 15817, 15818, 15802, 15813, 15801, 15815, 15806, 15805, 15819, 15812, 15811, 15820, 15808, 15812, 15811, 15804, 15806, 15817, 15816, 15814, 15816, 15811, 15806, 15822, 15807, 15808, 15818, 15822, 15808, 15690, 15774, 15688, 15690, 15776, 15775, 15785, 15704, 15783, 15792, 15707, 15708, 15709, 15792, 15708, 15795, 15705, 15785, 15791, 15694, 15707, 15745, 15800, 15744, 15756, 15817, 15755, 15817, 15752, 15755, 15745, 15814, 15802,
15822, 15769, 15821, 15822, 15773, 15770, 15823, 15756, 15773, 15799, 15786, 15784, 15805, 15788, 15790, 15815, 15796, 15813, 15786, 15813, 15796, 15798, 15805, 15790, 15827, 15824, 15826, 15829, 15826, 15828, 15833, 15830, 15832, 15837, 15834, 15836, 15839, 15832, 15838, 15841, 15836, 15840, 15836, 15831, 15833, 15840, 15833, 15839, 15845, 15842, 15844, 15849, 15846, 15848, 15851, 15844, 15850, 15853, 15848, 15852, 15848, 15843, 15845, 15852, 15845, 15851, 15844, 15835, 15837, 15850, 15837, 15841, 15855, 15856, 15854, 15856, 15858, 15854, 15863, 15860, 15861, 15865, 15866, 15864, 15871, 15868, 15869, 15859, 15864, 15858, 15869, 15862, 15863, 15857, 15860, 15856, 15860, 15859, 15856, 15868, 15867, 15865, 15862, 15865, 15859, 15877, 15874, 15876, 15873, 15874, 15872, 15881, 15878, 15879, 15881, 15876, 15880, 15878, 15876, 15873, 15867, 15872, 15866, 15879, 15870, 15871, 15870, 15873, 15867, 15883, 15884, 15882, 15887, 15888, 15886, 15893, 15890, 15891, 15897, 15894, 15895, 15901, 15898, 15899, 15905, 15902, 15903, 15899, 15892, 15893, 15903, 15896, 15897, 15885, 15890, 15884, 15889, 15894, 15888, 15886, 15885, 15883, 15896, 15891, 15894, 15904, 15899, 15902, 15888, 15891, 15885, 15902, 15893, 15896, 15907, 15908, 15906, 15911, 15912, 15910,
15917, 15914, 15915, 15921, 15918, 15919, 15925, 15922, 15923, 15929, 15926, 15927, 15923, 15916, 15917, 15927, 15920, 15921, 15909, 15914, 15908, 15913, 15918, 15912, 15910, 15909, 15907, 15920, 15915, 15918, 15928, 15923, 15926, 15912, 15915, 15909, 15926, 15917, 15920, 15933, 15930, 15932, 15937, 15934, 15936, 15941, 15938, 15939, 15945, 15942, 15943, 15941, 15932, 15940, 15945, 15936, 15944, 15936, 15931, 15933, 15944, 15939, 15942, 15944, 15933, 15941, 15949, 15946, 15948, 15953, 15950, 15952, 15957, 15954, 15955, 15961, 15958, 15959, 15957, 15948, 15956, 15961, 15952, 15960, 15952, 15947, 15949, 15960, 15955, 15958, 15960, 15949, 15957, 15939, 15900, 15901, 15943, 15904, 15905, 15942, 15901, 15904, 15955, 15924, 15925, 15959, 15928, 15929, 15958, 15925, 15928, 15906, 15889, 15887, 15916, 15895, 15914, 15924, 15903, 15922, 15908, 15895, 15889, 15922, 15897, 15916, 15948, 15935, 15937, 15956, 15943, 15954, 15956, 15937, 15945, 15954, 15905, 15924, 15832, 15825, 15827, 15838, 15827, 15829, 15882, 15857, 15855, 15892, 15861, 15890, 15900, 15869, 15898, 15884, 15861, 15857, 15898, 15863, 15892, 15932, 15875, 15877, 15940, 15879, 15938, 15940, 15877, 15881, 15938, 15871, 15900, 15963, 15964, 15962, 15967, 15968, 15966, 15975, 15972, 15973,
15970, 15965, 15963, 15969, 15972, 15968, 15966, 15963, 15962, 15974, 15970, 15972, 15968, 15970, 15963, 15977, 15978, 15976, 15981, 15982, 15980, 15987, 15984, 15985, 15991, 15988, 15989, 15979, 15984, 15978, 15983, 15988, 15982, 15980, 15979, 15977, 15990, 15985, 15988, 15982, 15985, 15979, 15997, 15994, 15996, 16000, 15996, 15999, 15996, 15992, 15993, 15999, 15993, 15998, 16004, 16001, 16003, 16008, 16005, 16007, 16016, 16017, 16015, 16014, 16016, 16013, 16010, 16003, 16009, 16012, 16007, 16011, 16018, 16020, 16017, 16019, 16018, 16014, 16007, 16002, 16004, 16011, 16004, 16010, 15965, 15992, 15964, 16000, 15974, 15975, 15999, 15971, 15974, 15971, 15993, 15965, 16010, 15986, 15987, 16012, 15990, 15991, 16011, 15987, 15990, 15976, 15969, 15967, 15986, 15973, 15984, 15978, 15973, 15969, 16003, 15995, 15997, 16009, 15997, 16000, 16009, 15975, 15986, 16023, 16024, 16022, 16027, 16028, 16026, 16031, 16027, 16030, 16033, 16034, 16032, 16037, 16038, 16036, 16025, 16032, 16024, 16028, 16037, 16036, 16039, 16029, 16031, 16026, 16025, 16023, 16038, 16033, 16036, 16028, 16033, 16025, 16046, 16043, 16045, 16048, 16049, 16047, 16046, 16048, 16044, 16058, 16059, 16057, 16068, 16065, 16067, 16070, 16065, 16069, 16072, 16073, 16071, 16049, 16061, 16047,
16068, 16051, 16066, 16072, 16055, 16056, 16076, 16059, 16060, 16054, 16045, 16053, 16049, 16056, 16055, 16054, 16050, 16046, 16075, 16068, 16067, 16077, 16067, 16070, 16059, 16054, 16057, 16070, 16063, 16064, 16077, 16072, 16075, 16074, 16070, 16064, 16059, 16072, 16056, 16055, 16062, 16049, 16076, 16052, 16068, 16081, 16078, 16080, 16083, 16079, 16081, 16035, 16053, 16034, 16058, 16038, 16039, 16057, 16035, 16038, 16079, 16040, 16078, 16079, 16042, 16041, 16040, 16031, 16030, 16031, 16042, 16039, 16081, 16051, 16052, 16083, 16058, 16082, 16060, 16081, 16052, 16042, 16058, 16039, 16087, 16088, 16086, 16085, 16087, 16084, 16089, 16091, 16088, 16085, 16092, 16089, 16096, 16097, 16095, 16094, 16096, 16093, 16107, 16108, 16106, 16117, 16114, 16116, 16119, 16114, 16118, 16121, 16122, 16120, 16097, 16110, 16095, 16117, 16101, 16115, 16121, 16104, 16105, 16125, 16108, 16109, 16097, 16105, 16104, 16103, 16098, 16094, 16124, 16117, 16116, 16126, 16116, 16119, 16099, 16094, 16093, 16108, 16103, 16106, 16119, 16112, 16113, 16126, 16121, 16124, 16094, 16106, 16103, 16123, 16119, 16113, 16108, 16121, 16105, 16104, 16111, 16097, 16125, 16102, 16117, 16130, 16131, 16129, 16137, 16134, 16135, 16143, 16140, 16142, 16128, 16138, 16127, 16137, 16145, 16136,
16130, 16136, 16132, 16146, 16142, 16145, 16131, 16128, 16127, 16142, 16138, 16139, 16128, 16136, 16133, 16145, 16139, 16144, 16145, 16133, 16136, 16133, 16139, 16128, 16150, 16151, 16149, 16156, 16153, 16154, 16160, 16157, 16158, 16164, 16161, 16163, 16168, 16165, 16167, 16170, 16155, 16156, 16172, 16159, 16160, 16170, 16163, 16169, 16152, 16157, 16151, 16172, 16167, 16171, 16149, 16148, 16147, 16159, 16154, 16157, 16167, 16162, 16164, 16148, 16157, 16154, 16171, 16164, 16170, 16171, 16156, 16159, 16088, 16129, 16086, 16135, 16091, 16092, 16088, 16134, 16130, 16148, 16112, 16147, 16154, 16122, 16123, 16154, 16113, 16148, 16110, 16085, 16084, 16122, 16090, 16120, 16085, 16120, 16090, 16155, 16135, 16153, 16163, 16141, 16143, 16169, 16143, 16146, 16169, 16137, 16155, 16122, 16135, 16092, 16176, 16173, 16175, 16178, 16173, 16177, 16179, 16176, 16175, 16181, 16175, 16178, 16183, 16184, 16182, 16187, 16188, 16186, 16193, 16190, 16191, 16197, 16194, 16195, 16201, 16198, 16200, 16205, 16202, 16204, 16207, 16192, 16193, 16209, 16196, 16197, 16185, 16190, 16184, 16207, 16200, 16206, 16189, 16194, 16188, 16209, 16204, 16208, 16186, 16185, 16183, 16196, 16191, 16194, 16204, 16199, 16201, 16188, 16191, 16185, 16208, 16201, 16207, 16208, 16193, 16196,
16211, 16212, 16210, 16215, 16211, 16214, 16219, 16216, 16217, 16223, 16220, 16222, 16223, 16214, 16221, 16225, 16218, 16219, 16213, 16216, 16212, 16219, 16213, 16215, 16225, 16222, 16224, 16225, 16215, 16223, 16210, 16189, 16187, 16218, 16195, 16216, 16222, 16203, 16205, 16212, 16195, 16189, 16224, 16205, 16209, 16224, 16197, 16218, 16017, 16099, 16015, 16107, 16020, 16021, 16020, 16100, 16017, 16176, 16063, 16174, 16180, 16073, 16074, 16180, 16064, 16176, 16022, 15983, 15981, 16034, 15989, 16032, 16024, 15989, 15983, 16045, 16006, 16008, 16061, 16014, 16013, 16073, 16019, 16071, 16053, 16008, 16012, 16014, 16071, 16019, 16053, 15991, 16034, 16178, 16101, 16102, 16181, 16107, 16179, 16109, 16178, 16102, 16182, 16152, 16150, 16192, 16158, 16190, 16200, 16166, 16168, 16184, 16158, 16152, 16206, 16168, 16172, 16206, 16160, 16192, 16073, 16107, 16021, 15400, 15643, 15398, 15646, 15399, 15644, 15652, 15404, 15405, 15404, 15645, 15400, 15652, 15401, 15646, 15467, 15710, 15465, 15472, 15711, 15467, 15723, 15479, 15722, 15484, 15726, 15483, 15723, 15484, 15481, 15726, 15472, 15483, 15578, 15824, 15577, 15829, 15581, 15582, 15581, 15826, 15578, 15839, 15617, 15618, 15841, 15621, 15622, 15840, 15618, 15621, 15851, 15633, 15634, 15853, 15637, 15638,
15852, 15634, 15637, 15850, 15622, 15633, 15838, 15582, 15617, 14843, 15127, 14842, 15132, 14846, 14847, 14846, 15128, 14843, 15160, 14898, 14899, 15164, 14902, 14903, 15163, 14899, 14902, 15176, 14912, 15175, 15176, 14924, 14914, 15179, 14903, 14924, 15159, 14847, 14898, 16227, 16228, 16226, 16235, 16236, 16234, 16241, 16238, 16239, 16243, 16244, 16242, 16233, 16227, 16232, 16231, 16238, 16230, 16228, 16231, 16230, 16237, 16242, 16236, 16234, 16233, 16232, 16242, 16241, 16239, 16233, 16242, 16239, 16239, 16229, 16233, 16247, 16248, 16246, 16251, 16247, 16250, 16253, 16254, 16252, 16257, 16258, 16256, 16248, 16257, 16256, 16259, 16249, 16251, 16255, 16260, 16254, 16252, 16251, 16250, 16251, 16260, 16259, 16265, 16262, 16263, 16269, 16266, 16267, 16273, 16270, 16271, 16277, 16274, 16275, 16271, 16264, 16265, 16275, 16268, 16269, 16268, 16263, 16266, 16276, 16271, 16274, 16274, 16265, 16268, 16281, 16278, 16279, 16285, 16282, 16283, 16289, 16286, 16287, 16293, 16290, 16291, 16287, 16280, 16281, 16291, 16284, 16285, 16284, 16279, 16282, 16292, 16287, 16290, 16290, 16281, 16284, 16263, 16240, 16241, 16267, 16244, 16245, 16266, 16241, 16244, 16279, 16258, 16259, 16261, 16282, 16260, 16260, 16279, 16259, 16246, 16237, 16235, 16256, 16245, 16243,
16248, 16243, 16237, 16280, 16267, 16278, 16288, 16275, 16286, 16286, 16269, 16280, 16278, 16245, 16258, 16298, 16299, 16297, 16302, 16303, 16301, 16306, 16302, 16305, 16308, 16309, 16307, 16312, 16313, 16311, 16318, 16315, 16316, 16322, 16319, 16320, 16316, 16309, 16310, 16320, 16313, 16314, 16299, 16308, 16307, 16303, 16312, 16311, 16314, 16304, 16306, 16301, 16300, 16298, 16311, 16310, 16308, 16321, 16316, 16319, 16303, 16308, 16300, 16319, 16310, 16313, 16324, 16325, 16323, 16333, 16334, 16332, 16339, 16336, 16337, 16341, 16342, 16340, 16334, 16327, 16332, 16330, 16336, 16329, 16326, 16329, 16325, 16335, 16340, 16334, 16327, 16326, 16324, 16342, 16337, 16340, 16328, 16330, 16326, 16331, 16337, 16330, 16334, 16331, 16328, 16299, 16294, 16297, 16307, 16295, 16299, 16323, 16306, 16305, 16338, 16320, 16336, 16306, 16329, 16314, 16329, 16320, 16314, 16347, 16344, 16346, 16349, 16350, 16348, 16355, 16352, 16353, 16361, 16358, 16360, 16347, 16356, 16345, 16355, 16363, 16354, 16350, 16347, 16346, 16364, 16360, 16363, 16354, 16349, 16352, 16360, 16356, 16357, 16362, 16360, 16357, 16354, 16362, 16351, 16351, 16357, 16347, 16368, 16365, 16366, 16372, 16369, 16370, 16378, 16375, 16377, 16374, 16375, 16373, 16382, 16379, 16381, 16384, 16385, 16383,
16388, 16389, 16387, 16384, 16367, 16368, 16388, 16371, 16372, 16385, 16378, 16377, 16383, 16377, 16374, 16389, 16382, 16381, 16371, 16366, 16369, 16381, 16376, 16378, 16387, 16386, 16384, 16389, 16378, 16386, 16387, 16368, 16371, 16382, 16391, 16380, 16390, 16392, 16382, 16367, 16353, 16365, 16374, 16359, 16361, 16364, 16374, 16361, 16367, 16364, 16355, 16397, 16394, 16395, 16401, 16398, 16399, 16405, 16402, 16403, 16409, 16406, 16407, 16403, 16396, 16397, 16407, 16400, 16401, 16400, 16395, 16398, 16408, 16403, 16406, 16406, 16397, 16400, 16413, 16410, 16411, 16417, 16414, 16415, 16421, 16418, 16419, 16425, 16422, 16423, 16419, 16412, 16413, 16423, 16416, 16417, 16416, 16411, 16414, 16424, 16419, 16422, 16422, 16413, 16416, 16429, 16426, 16428, 16435, 16432, 16434, 16431, 16432, 16430, 16439, 16440, 16438, 16437, 16428, 16436, 16440, 16435, 16434, 16438, 16434, 16431, 16431, 16427, 16429, 16437, 16431, 16429, 16447, 16444, 16445, 16451, 16448, 16449, 16455, 16452, 16454, 16461, 16458, 16460, 16457, 16458, 16456, 16465, 16466, 16464, 16443, 16452, 16442, 16447, 16462, 16446, 16465, 16450, 16451, 16463, 16454, 16462, 16467, 16460, 16466, 16464, 16460, 16457, 16450, 16445, 16448, 16457, 16453, 16455, 16464, 16455, 16463, 16450, 16463, 16447,
16446, 16454, 16443, 16405, 16436, 16404, 16439, 16408, 16409, 16408, 16437, 16405, 16445, 16420, 16421, 16449, 16424, 16425, 16448, 16421, 16424, 16412, 16399, 16410, 16420, 16407, 16418, 16418, 16401, 16412, 16443, 16433, 16435, 16444, 16441, 16439, 16446, 16435, 16441, 16444, 16409, 16420, 16349, 16272, 16273, 16353, 16276, 16277, 16352, 16273, 16276, 16366, 16288, 16289, 16370, 16292, 16293, 16369, 16289, 16292, 16365, 16277, 16288, 16395, 16317, 16318, 16399, 16321, 16322, 16398, 16318, 16321, 16411, 16338, 16339, 16415, 16342, 16343, 16414, 16339, 16342, 16410, 16322, 16338, 16294, 16255, 16253, 16255, 16296, 16261, 16309, 16283, 16307, 16317, 16291, 16315, 16315, 16285, 16309, 16296, 16283, 16261, 16396, 16370, 16394, 16404, 16388, 16402, 16402, 16372, 16396, 16428, 16391, 16392, 16393, 16428, 16392, 16404, 16393, 16390, 16394, 16293, 16317, 16471, 16472, 16470, 16472, 16468, 16470, 16473, 16477, 16472, 16477, 16469, 16472, 16474, 16473, 16471, 16475, 16478, 16473, 16481, 16482, 16480, 16487, 16484, 16485, 16489, 16490, 16488, 16495, 16492, 16493, 16499, 16496, 16497, 16493, 16486, 16487, 16497, 16490, 16491, 16483, 16488, 16482, 16490, 16485, 16488, 16498, 16493, 16496, 16496, 16487, 16490, 16503, 16504, 16502, 16504, 16500, 16502,
16512, 16513, 16511, 16516, 16512, 16515, 16520, 16517, 16518, 16522, 16523, 16521, 16510, 16503, 16509, 16508, 16517, 16507, 16504, 16508, 16507, 16507, 16501, 16504, 16514, 16521, 16513, 16524, 16514, 16516, 16511, 16510, 16509, 16521, 16520, 16518, 16513, 16518, 16510, 16518, 16505, 16510, 16528, 16525, 16526, 16532, 16529, 16530, 16540, 16537, 16539, 16540, 16535, 16538, 16544, 16541, 16542, 16546, 16547, 16545, 16542, 16527, 16528, 16532, 16545, 16531, 16547, 16540, 16539, 16540, 16546, 16536, 16531, 16526, 16529, 16545, 16544, 16542, 16545, 16528, 16531, 16552, 16549, 16550, 16556, 16553, 16554, 16560, 16557, 16559, 16562, 16563, 16561, 16560, 16562, 16558, 16568, 16565, 16567, 16570, 16571, 16569, 16571, 16565, 16569, 16552, 16573, 16551, 16578, 16555, 16556, 16573, 16560, 16559, 16563, 16576, 16575, 16560, 16576, 16564, 16578, 16567, 16577, 16572, 16579, 16571, 16579, 16567, 16571, 16555, 16550, 16553, 16577, 16576, 16574, 16555, 16574, 16552, 16526, 16494, 16495, 16530, 16498, 16499, 16529, 16495, 16498, 16550, 16519, 16520, 16554, 16523, 16524, 16553, 16520, 16523, 16500, 16483, 16481, 16506, 16491, 16489, 16519, 16497, 16517, 16501, 16489, 16483, 16517, 16491, 16507, 16551, 16530, 16549, 16559, 16535, 16536, 16561, 16540, 16538,
16546, 16559, 16536, 16540, 16575, 16548, 16573, 16532, 16551, 16549, 16499, 16519, 16585, 16586, 16584, 16591, 16592, 16590, 16592, 16588, 16590, 16595, 16596, 16594, 16606, 16603, 16605, 16602, 16603, 16601, 16612, 16609, 16610, 16616, 16613, 16614, 16608, 16591, 16607, 16610, 16596, 16597, 16600, 16613, 16599, 16587, 16594, 16586, 16612, 16605, 16611, 16602, 16611, 16605, 16592, 16600, 16599, 16599, 16589, 16592, 16588, 16587, 16585, 16598, 16597, 16595, 16615, 16610, 16613, 16589, 16595, 16587, 16613, 16597, 16599, 16614, 16593, 16608, 16620, 16621, 16619, 16621, 16617, 16619, 16621, 16625, 16624, 16624, 16618, 16621, 16586, 16581, 16584, 16594, 16582, 16586, 16617, 16608, 16607, 16623, 16616, 16614, 16608, 16623, 16614, 16482, 16474, 16480, 16478, 16484, 16477, 16479, 16485, 16478, 16488, 16475, 16482, 16581, 16516, 16515, 16516, 16583, 16524, 16596, 16554, 16594, 16602, 16566, 16568, 16603, 16572, 16570, 16578, 16602, 16568, 16572, 16611, 16580, 16596, 16578, 16556, 16583, 16554, 16524, 16627, 16628, 16626, 16628, 16631, 16630, 16633, 16634, 16632, 16637, 16638, 16636, 16634, 16641, 16640, 16639, 16642, 16638, 16636, 16635, 16633, 16638, 16641, 16635, 16632, 16629, 16627, 16629, 16640, 16631, 16650, 16651, 16649, 16648, 16650, 16647,
16659, 16660, 16658, 16665, 16666, 16664, 16671, 16668, 16669, 16651, 16662, 16649, 16665, 16656, 16657, 16669, 16660, 16661, 16651, 16657, 16656, 16648, 16657, 16652, 16653, 16648, 16647, 16660, 16655, 16658, 16670, 16665, 16668, 16654, 16655, 16648, 16660, 16665, 16657, 16656, 16663, 16651, 16675, 16672, 16673, 16679, 16676, 16677, 16683, 16680, 16681, 16687, 16684, 16685, 16681, 16674, 16675, 16685, 16678, 16679, 16678, 16673, 16676, 16686, 16681, 16684, 16684, 16675, 16678, 16691, 16688, 16689, 16695, 16692, 16693, 16699, 16696, 16697, 16703, 16700, 16701, 16697, 16690, 16691, 16701, 16694, 16695, 16694, 16689, 16692, 16702, 16697, 16700, 16700, 16691, 16694, 16707, 16704, 16705, 16711, 16708, 16709, 16715, 16712, 16713, 16719, 16716, 16717, 16713, 16706, 16707, 16717, 16710, 16711, 16710, 16705, 16708, 16718, 16713, 16716, 16716, 16707, 16710, 16689, 16666, 16667, 16693, 16670, 16671, 16692, 16667, 16670, 16705, 16682, 16683, 16709, 16686, 16687, 16708, 16683, 16686, 16674, 16659, 16672, 16682, 16669, 16680, 16680, 16661, 16674, 16706, 16693, 16704, 16714, 16701, 16712, 16712, 16695, 16706, 16704, 16671, 16682, 16721, 16722, 16720, 16727, 16728, 16726, 16728, 16724, 16726, 16731, 16732, 16730, 16740, 16737, 16739, 16744, 16741, 16742,
16746, 16747, 16745, 16738, 16727, 16737, 16742, 16732, 16733, 16736, 16745, 16735, 16722, 16731, 16730, 16729, 16735, 16728, 16735, 16725, 16728, 16748, 16738, 16740, 16724, 16723, 16721, 16734, 16733, 16731, 16747, 16742, 16745, 16725, 16731, 16723, 16745, 16733, 16735, 16746, 16729, 16738, 16755, 16752, 16753, 16759, 16756, 16757, 16765, 16762, 16763, 16767, 16768, 16766, 16763, 16754, 16755, 16767, 16758, 16759, 16758, 16753, 16756, 16768, 16763, 16766, 16766, 16755, 16758, 16773, 16770, 16772, 16779, 16776, 16778, 16777, 16772, 16776, 16775, 16771, 16773, 16781, 16777, 16779, 16777, 16775, 16773, 16753, 16743, 16744, 16757, 16747, 16748, 16756, 16744, 16747, 16771, 16749, 16770, 16771, 16751, 16750, 16749, 16740, 16739, 16750, 16748, 16740, 16774, 16759, 16757, 16779, 16760, 16761, 16781, 16767, 16780, 16769, 16779, 16761, 16780, 16759, 16775, 16751, 16757, 16748, 16628, 16653, 16626, 16659, 16630, 16631, 16628, 16658, 16654, 16673, 16640, 16641, 16643, 16676, 16642, 16676, 16641, 16642, 16672, 16631, 16640, 16722, 16644, 16720, 16730, 16645, 16722, 16644, 16639, 16637, 16639, 16646, 16643, 16732, 16677, 16730, 16743, 16685, 16741, 16741, 16679, 16732, 16754, 16709, 16752, 16764, 16717, 16762, 16762, 16711, 16754, 16752, 16687, 16743,
16646, 16677, 16643, 16785, 16782, 16784, 16787, 16788, 16786, 16788, 16790, 16786, 16793, 16787, 16791, 16799, 16796, 16797, 16801, 16802, 16800, 16805, 16806, 16804, 16792, 16800, 16790, 16805, 16798, 16799, 16795, 16784, 16794, 16789, 16796, 16788, 16788, 16798, 16792, 16793, 16797, 16789, 16804, 16803, 16801, 16788, 16783, 16785, 16796, 16785, 16795, 16792, 16804, 16801, 16809, 16810, 16808, 16815, 16812, 16813, 16819, 16816, 16817, 16823, 16820, 16821, 16827, 16824, 16825, 16815, 16820, 16814, 16825, 16818, 16819, 16809, 16814, 16811, 16818, 16813, 16816, 16826, 16821, 16824, 16824, 16815, 16818, 16831, 16828, 16830, 16831, 16832, 16829, 16835, 16830, 16834, 16835, 16833, 16831, 16844, 16841, 16842, 16846, 16847, 16845, 16852, 16849, 16851, 16856, 16853, 16855, 16838, 16849, 16837, 16856, 16839, 16854, 16844, 16857, 16843, 16860, 16847, 16848, 16857, 16852, 16851, 16860, 16855, 16859, 16847, 16842, 16845, 16855, 16850, 16852, 16858, 16855, 16852, 16847, 16858, 16844, 16843, 16851, 16838, 16856, 16848, 16840, 16803, 16828, 16802, 16807, 16834, 16806, 16803, 16834, 16830, 16842, 16822, 16823, 16846, 16826, 16827, 16845, 16823, 16826, 16809, 16787, 16789, 16810, 16793, 16791, 16822, 16805, 16820, 16797, 16809, 16789, 16793, 16814, 16799,
16820, 16799, 16814, 16838, 16832, 16833, 16841, 16836, 16835, 16836, 16838, 16833, 16841, 16807, 16822, 16862, 16863, 16861, 16866, 16862, 16865, 16868, 16869, 16867, 16872, 16873, 16871, 16876, 16877, 16875, 16882, 16879, 16881, 16884, 16885, 16883, 16890, 16887, 16888, 16894, 16891, 16892, 16880, 16865, 16879, 16870, 16883, 16869, 16874, 16887, 16873, 16878, 16891, 16877, 16863, 16872, 16871, 16874, 16864, 16866, 16882, 16888, 16880, 16868, 16877, 16870, 16884, 16893, 16886, 16868, 16862, 16864, 16875, 16864, 16872, 16888, 16866, 16880, 16877, 16884, 16870, 16900, 16901, 16899, 16901, 16897, 16899, 16906, 16903, 16904, 16915, 16912, 16913, 16919, 16916, 16917, 16911, 16900, 16910, 16913, 16905, 16906, 16909, 16916, 16908, 16902, 16908, 16901, 16901, 16907, 16898, 16897, 16896, 16895, 16907, 16906, 16904, 16918, 16913, 16916, 16898, 16904, 16896, 16916, 16906, 16908, 16911, 16909, 16902, 16925, 16922, 16924, 16929, 16930, 16928, 16933, 16934, 16932, 16939, 16940, 16938, 16944, 16945, 16943, 16925, 16936, 16923, 16944, 16934, 16935, 16930, 16925, 16924, 16943, 16941, 16939, 16926, 16921, 16920, 16932, 16931, 16929, 16939, 16936, 16937, 16921, 16932, 16929, 16943, 16937, 16942, 16934, 16942, 16931, 16931, 16937, 16925, 16950, 16947, 16948,
16954, 16951, 16952, 16958, 16955, 16956, 16962, 16959, 16960, 16956, 16949, 16950, 16960, 16953, 16954, 16953, 16948, 16951, 16961, 16956, 16959, 16959, 16950, 16953, 16921, 16881, 16920, 16886, 16926, 16885, 16890, 16928, 16889, 16933, 16893, 16894, 16921, 16890, 16882, 16893, 16927, 16886, 16948, 16914, 16915, 16952, 16918, 16919, 16951, 16915, 16918, 16905, 16876, 16903, 16914, 16892, 16912, 16912, 16878, 16905, 16949, 16933, 16947, 16957, 16944, 16955, 16955, 16935, 16949, 16947, 16894, 16914, 16966, 16963, 16965, 16966, 16967, 16964, 16970, 16965, 16969, 16966, 16971, 16968, 16977, 16974, 16975, 16981, 16978, 16979, 16985, 16982, 16984, 16985, 16986, 16983, 16994, 16991, 16992, 16973, 16982, 16972, 16977, 16988, 16976, 16992, 16980, 16981, 16989, 16984, 16988, 16989, 16987, 16985, 16980, 16975, 16978, 16991, 16990, 16989, 16991, 16977, 16980, 16973, 16988, 16984, 17002, 16999, 17001, 16998, 16999, 16997, 17005, 17006, 17004, 17009, 17010, 17008, 17002, 17008, 17000, 17007, 17001, 17006, 16998, 17006, 17001, 17012, 17011, 17009, 16998, 16995, 16996, 17003, 16998, 16996, 17007, 17009, 17002, 16987, 16995, 16986, 17005, 16993, 16994, 16993, 17003, 16990, 16990, 16996, 16987, 16973, 16967, 16968, 16974, 16971, 16970, 16976, 16968, 16971,
16941, 16963, 16940, 16946, 16969, 16945, 16945, 16965, 16941, 16975, 16957, 16958, 16979, 16961, 16962, 16978, 16958, 16961, 16974, 16946, 16957, 16873, 16817, 16871, 16889, 16825, 16887, 16887, 16819, 16873, 16924, 16839, 16840, 16930, 16846, 16928, 16930, 16840, 16848, 16928, 16827, 16889, 16534, 16782, 16533, 16788, 16537, 16786, 16540, 16787, 16538, 16544, 16794, 16543, 16547, 16795, 16544, 16543, 16784, 16534, 16796, 16539, 16788, 16548, 16789, 16540, 16813, 16575, 16576, 16580, 16816, 16579, 16816, 16576, 16579, 16812, 16548, 16575, 16863, 16603, 16861, 16606, 16862, 16604, 16876, 16615, 16616, 16875, 16612, 16615, 16871, 16605, 16863, 16606, 16872, 16864, 16896, 16620, 16895, 16625, 16903, 16624, 16896, 16625, 16622, 16903, 16616, 16624, 16611, 16817, 16580, 16468, 16335, 16333, 16476, 16343, 16341, 16469, 16341, 16335, 16486, 16415, 16484, 16494, 16423, 16492, 16492, 16417, 16486, 16527, 16449, 16525, 16534, 16459, 16461, 16541, 16467, 16465, 16543, 16461, 16467, 16541, 16451, 16527, 16525, 16425, 16494, 16484, 16343, 16477, 17017, 17018, 17016, 17018, 17014, 17016, 17019, 17021, 17018, 17018, 17020, 17015, 17026, 17023, 17025, 17030, 17031, 17029, 17035, 17036, 17034, 17039, 17035, 17038, 17043, 17040, 17041, 17045, 17046, 17044,
17036, 17027, 17034, 17041, 17031, 17032, 17032, 17024, 17026, 17037, 17044, 17036, 17047, 17037, 17039, 17027, 17026, 17025, 17046, 17041, 17044, 17028, 17032, 17026, 17033, 17041, 17032, 17036, 17033, 17028, 17024, 17017, 17023, 17022, 17029, 17021, 17030, 17019, 17024, 17049, 17050, 17048, 17050, 17053, 17052, 17057, 17054, 17056, 17061, 17058, 17059, 17065, 17062, 17063, 17069, 17066, 17068, 17071, 17072, 17070, 17072, 17066, 17070, 17069, 17071, 17067, 17077, 17074, 17075, 17069, 17056, 17067, 17075, 17060, 17061, 17065, 17078, 17064, 17057, 17063, 17055, 17078, 17069, 17068, 17072, 17081, 17080, 17072, 17078, 17068, 17069, 17081, 17073, 17064, 17059, 17062, 17078, 17077, 17075, 17064, 17075, 17061, 17069, 17065, 17057, 17085, 17086, 17084, 17083, 17085, 17082, 17089, 17090, 17088, 17095, 17092, 17094, 17102, 17099, 17100, 17110, 17107, 17109, 17110, 17104, 17108, 17112, 17113, 17111, 17113, 17115, 17111, 17118, 17119, 17117, 17122, 17123, 17121, 17086, 17103, 17084, 17113, 17092, 17111, 17095, 17112, 17093, 17118, 17097, 17098, 17087, 17097, 17086, 17083, 17098, 17087, 17120, 17109, 17119, 17110, 17118, 17106, 17090, 17100, 17099, 17101, 17095, 17094, 17113, 17122, 17121, 17113, 17123, 17116, 17088, 17083, 17082, 17099, 17098, 17096,
17113, 17104, 17106, 17115, 17110, 17108, 17083, 17099, 17096, 17118, 17113, 17106, 17110, 17123, 17120, 17101, 17118, 17098, 17097, 17105, 17086, 17121, 17094, 17113, 17102, 17114, 17095, 17128, 17125, 17126, 17132, 17129, 17130, 17138, 17135, 17136, 17140, 17141, 17139, 17136, 17127, 17128, 17140, 17131, 17132, 17131, 17126, 17129, 17139, 17138, 17136, 17139, 17128, 17131, 17144, 17145, 17143, 17150, 17147, 17148, 17154, 17151, 17152, 17158, 17155, 17156, 17162, 17159, 17160, 17150, 17155, 17149, 17160, 17153, 17154, 17145, 17148, 17147, 17151, 17150, 17148, 17161, 17156, 17159, 17159, 17150, 17153, 17126, 17076, 17077, 17130, 17080, 17081, 17080, 17126, 17077, 17145, 17107, 17143, 17110, 17144, 17108, 17124, 17151, 17123, 17123, 17148, 17120, 17145, 17119, 17109, 17120, 17146, 17110, 17103, 17069, 17067, 17119, 17079, 17117, 17069, 17117, 17079, 17149, 17130, 17147, 17157, 17140, 17155, 17155, 17132, 17149, 17119, 17130, 17081, 17168, 17169, 17167, 17174, 17171, 17172, 17176, 17177, 17175, 17182, 17179, 17180, 17186, 17183, 17184, 17180, 17173, 17174, 17184, 17177, 17178, 17170, 17175, 17169, 17167, 17166, 17164, 17175, 17174, 17172, 17185, 17180, 17183, 17166, 17175, 17172, 17183, 17174, 17177, 17190, 17191, 17189, 17191, 17187, 17189,
17201, 17198, 17200, 17205, 17202, 17203, 17197, 17190, 17196, 17195, 17202, 17194, 17192, 17194, 17191, 17194, 17188, 17191, 17207, 17199, 17201, 17198, 17197, 17196, 17206, 17205, 17203, 17199, 17203, 17197, 17197, 17195, 17192, 17211, 17208, 17209, 17215, 17212, 17213, 17219, 17216, 17217, 17223, 17220, 17221, 17217, 17210, 17211, 17221, 17214, 17215, 17214, 17209, 17212, 17222, 17217, 17220, 17220, 17211, 17214, 17227, 17228, 17226, 17228, 17224, 17226, 17233, 17230, 17231, 17242, 17239, 17241, 17242, 17237, 17240, 17246, 17243, 17244, 17250, 17247, 17248, 17238, 17227, 17237, 17244, 17232, 17233, 17236, 17247, 17235, 17229, 17235, 17228, 17228, 17234, 17225, 17249, 17242, 17241, 17250, 17238, 17242, 17234, 17233, 17231, 17247, 17246, 17244, 17247, 17233, 17235, 17238, 17236, 17229, 17209, 17181, 17182, 17213, 17185, 17186, 17212, 17182, 17185, 17225, 17200, 17224, 17231, 17204, 17205, 17207, 17231, 17205, 17234, 17201, 17225, 17187, 17170, 17168, 17193, 17178, 17176, 17204, 17184, 17202, 17188, 17176, 17170, 17202, 17178, 17194, 17232, 17213, 17230, 17245, 17221, 17243, 17243, 17215, 17232, 17230, 17186, 17204, 17055, 17038, 17054, 17059, 17042, 17043, 17047, 17062, 17046, 17062, 17043, 17046, 17063, 17039, 17055, 17050, 17163, 17048,
17166, 17049, 17164, 17172, 17052, 17053, 17050, 17171, 17165, 17172, 17051, 17166, 17163, 17091, 17089, 17173, 17100, 17171, 17179, 17124, 17122, 17165, 17100, 17091, 17173, 17122, 17102, 17210, 17152, 17208, 17218, 17160, 17216, 17216, 17154, 17210, 17208, 17124, 17181, 17014, 16911, 16910, 17020, 16919, 16917, 17015, 16917, 16911, 17031, 16952, 17029, 17042, 16960, 17040, 17040, 16954, 17031, 17029, 16919, 17021, 17060, 16979, 17058, 17076, 16992, 17074, 17074, 16981, 17060, 17127, 17005, 17125, 17134, 17010, 17011, 17135, 17013, 17012, 17137, 17011, 17013, 17127, 17012, 17007, 17125, 16994, 17076, 17058, 16962, 17042, 17252, 17253, 17251, 17256, 17257, 17255, 17262, 17259, 17260, 17266, 17263, 17265, 17272, 17269, 17271, 17268, 17269, 17267, 17276, 17277, 17275, 17254, 17263, 17253, 17258, 17273, 17257, 17276, 17261, 17262, 17255, 17254, 17252, 17274, 17265, 17273, 17278, 17271, 17277, 17275, 17271, 17268, 17261, 17256, 17259, 17268, 17264, 17266, 17275, 17266, 17274, 17261, 17274, 17258, 17257, 17265, 17254, 17282, 17279, 17280, 17286, 17283, 17284, 17294, 17291, 17292, 17296, 17297, 17295, 17292, 17281, 17282, 17296, 17285, 17286, 17285, 17280, 17283, 17297, 17292, 17295, 17295, 17282, 17285, 17305, 17302, 17304, 17309, 17306, 17308,
17310, 17305, 17304, 17313, 17308, 17312, 17308, 17303, 17305, 17311, 17308, 17305, 17288, 17302, 17287, 17309, 17289, 17307, 17294, 17310, 17293, 17313, 17297, 17298, 17297, 17311, 17294, 17293, 17304, 17288, 17309, 17298, 17290, 17281, 17260, 17279, 17288, 17270, 17272, 17291, 17278, 17276, 17278, 17288, 17272, 17291, 17262, 17281, 17319, 17316, 17318, 17319, 17314, 17317, 17323, 17320, 17321, 17327, 17324, 17325, 17333, 17330, 17332, 17333, 17328, 17331, 17339, 17336, 17337, 17335, 17316, 17334, 17323, 17336, 17322, 17338, 17333, 17332, 17333, 17337, 17329, 17326, 17319, 17318, 17319, 17325, 17315, 17324, 17323, 17321, 17335, 17328, 17329, 17340, 17329, 17337, 17340, 17323, 17326, 17340, 17318, 17335, 17344, 17345, 17343, 17345, 17341, 17343, 17355, 17352, 17354, 17359, 17356, 17357, 17351, 17344, 17350, 17349, 17356, 17348, 17346, 17348, 17345, 17348, 17342, 17345, 17361, 17353, 17355, 17352, 17351, 17350, 17360, 17359, 17357, 17353, 17357, 17351, 17351, 17349, 17346, 17363, 17364, 17362, 17367, 17363, 17366, 17371, 17368, 17369, 17373, 17374, 17372, 17381, 17378, 17379, 17385, 17382, 17383, 17377, 17366, 17376, 17379, 17370, 17371, 17375, 17382, 17374, 17365, 17372, 17364, 17375, 17365, 17367, 17374, 17369, 17372, 17384, 17379, 17382,
17382, 17371, 17374, 17383, 17367, 17377, 17389, 17390, 17388, 17390, 17386, 17388, 17390, 17394, 17393, 17390, 17392, 17387, 17364, 17354, 17362, 17369, 17358, 17359, 17361, 17369, 17359, 17372, 17355, 17364, 17386, 17377, 17376, 17392, 17385, 17383, 17387, 17383, 17377, 17342, 17299, 17341, 17347, 17300, 17342, 17322, 17284, 17320, 17332, 17289, 17290, 17338, 17296, 17336, 17298, 17332, 17290, 17336, 17286, 17322, 17398, 17399, 17397, 17405, 17406, 17404, 17406, 17408, 17404, 17415, 17412, 17413, 17419, 17416, 17417, 17396, 17405, 17395, 17411, 17398, 17410, 17403, 17416, 17402, 17407, 17412, 17406, 17412, 17409, 17406, 17399, 17403, 17402, 17397, 17396, 17395, 17416, 17415, 17413, 17399, 17401, 17396, 17416, 17401, 17402, 17396, 17413, 17407, 17417, 17400, 17411, 17423, 17420, 17422, 17423, 17424, 17421, 17431, 17428, 17430, 17427, 17428, 17426, 17435, 17432, 17434, 17437, 17438, 17436, 17443, 17440, 17441, 17439, 17430, 17438, 17427, 17438, 17430, 17442, 17435, 17434, 17434, 17429, 17431, 17440, 17439, 17437, 17442, 17431, 17439, 17447, 17444, 17446, 17449, 17445, 17447, 17409, 17426, 17408, 17415, 17436, 17414, 17441, 17418, 17419, 17440, 17415, 17418, 17414, 17427, 17409, 17445, 17422, 17444, 17445, 17425, 17423, 17420, 17411, 17410,
17424, 17419, 17417, 17421, 17417, 17411, 17447, 17433, 17435, 17449, 17441, 17448, 17443, 17447, 17435, 17425, 17441, 17419, 17453, 17450, 17451, 17457, 17454, 17455, 17461, 17458, 17459, 17465, 17462, 17463, 17459, 17452, 17453, 17463, 17456, 17457, 17456, 17451, 17454, 17464, 17459, 17462, 17462, 17453, 17456, 17469, 17466, 17468, 17473, 17474, 17472, 17480, 17481, 17479, 17481, 17477, 17479, 17486, 17483, 17484, 17478, 17470, 17477, 17484, 17474, 17475, 17475, 17467, 17469, 17482, 17488, 17481, 17488, 17478, 17481, 17470, 17469, 17468, 17487, 17486, 17484, 17471, 17475, 17469, 17476, 17484, 17475, 17487, 17471, 17478, 17491, 17492, 17490, 17495, 17496, 17494, 17501, 17498, 17499, 17505, 17502, 17504, 17505, 17506, 17503, 17516, 17513, 17514, 17493, 17502, 17492, 17497, 17510, 17496, 17514, 17500, 17501, 17494, 17493, 17491, 17511, 17504, 17510, 17505, 17512, 17507, 17500, 17495, 17498, 17509, 17506, 17507, 17513, 17512, 17511, 17515, 17507, 17512, 17513, 17497, 17500, 17496, 17504, 17493, 17520, 17517, 17519, 17524, 17521, 17522, 17526, 17527, 17525, 17532, 17529, 17530, 17536, 17533, 17534, 17530, 17523, 17524, 17534, 17527, 17528, 17528, 17518, 17520, 17527, 17522, 17525, 17535, 17530, 17533, 17533, 17524, 17527, 17495, 17460, 17461,
17499, 17464, 17465, 17498, 17461, 17464, 17518, 17480, 17517, 17522, 17485, 17486, 17489, 17525, 17488, 17525, 17486, 17488, 17526, 17482, 17518, 17474, 17455, 17472, 17485, 17463, 17483, 17483, 17457, 17474, 17523, 17499, 17521, 17531, 17514, 17529, 17529, 17501, 17523, 17521, 17465, 17485, 17543, 17544, 17542, 17544, 17540, 17542, 17541, 17537, 17540, 17545, 17547, 17544, 17547, 17541, 17544, 17546, 17538, 17541, 17554, 17551, 17553, 17550, 17551, 17549, 17556, 17557, 17555, 17554, 17561, 17552, 17558, 17553, 17557, 17555, 17553, 17550, 17554, 17564, 17562, 17568, 17565, 17566, 17572, 17569, 17570, 17574, 17575, 17573, 17578, 17579, 17577, 17584, 17581, 17582, 17578, 17567, 17568, 17582, 17571, 17572, 17577, 17576, 17574, 17571, 17566, 17569, 17583, 17578, 17581, 17581, 17568, 17571, 17586, 17587, 17585, 17590, 17586, 17589, 17592, 17593, 17591, 17600, 17597, 17599, 17604, 17601, 17602, 17596, 17589, 17595, 17594, 17601, 17593, 17588, 17591, 17587, 17594, 17588, 17590, 17606, 17598, 17600, 17597, 17596, 17595, 17605, 17604, 17602, 17598, 17602, 17596, 17602, 17590, 17596, 17610, 17607, 17609, 17610, 17611, 17608, 17619, 17616, 17617, 17627, 17624, 17626, 17623, 17624, 17622, 17630, 17631, 17629, 17612, 17620, 17611, 17630, 17618, 17619,
17614, 17609, 17613, 17614, 17612, 17610, 17632, 17626, 17631, 17623, 17631, 17626, 17616, 17615, 17614, 17623, 17620, 17621, 17629, 17621, 17628, 17618, 17628, 17615, 17615, 17621, 17612, 17587, 17559, 17585, 17587, 17563, 17560, 17576, 17607, 17575, 17580, 17613, 17579, 17617, 17583, 17584, 17616, 17580, 17583, 17579, 17609, 17576, 17567, 17556, 17565, 17574, 17561, 17562, 17577, 17562, 17564, 17567, 17564, 17558, 17636, 17633, 17635, 17640, 17641, 17639, 17645, 17646, 17644, 17649, 17645, 17648, 17653, 17650, 17651, 17655, 17656, 17654, 17646, 17637, 17644, 17651, 17641, 17642, 17642, 17634, 17636, 17646, 17655, 17654, 17657, 17647, 17649, 17637, 17636, 17635, 17656, 17651, 17654, 17638, 17642, 17636, 17643, 17651, 17642, 17646, 17643, 17638, 17661, 17658, 17660, 17665, 17662, 17663, 17667, 17668, 17666, 17673, 17670, 17671, 17677, 17674, 17675, 17671, 17664, 17665, 17675, 17668, 17669, 17661, 17667, 17659, 17668, 17663, 17666, 17676, 17671, 17674, 17674, 17665, 17668, 17684, 17685, 17683, 17685, 17681, 17683, 17682, 17678, 17681, 17686, 17688, 17685, 17688, 17682, 17685, 17687, 17679, 17682, 17659, 17648, 17658, 17663, 17652, 17653, 17657, 17666, 17656, 17666, 17653, 17656, 17667, 17649, 17659, 17678, 17661, 17660, 17687, 17677, 17675,
17679, 17669, 17661, 17680, 17675, 17669, 17509, 17549, 17508, 17516, 17555, 17515, 17515, 17550, 17509, 17566, 17531, 17532, 17570, 17535, 17536, 17569, 17532, 17535, 17565, 17516, 17531, 17634, 17543, 17633, 17548, 17639, 17547, 17640, 17545, 17634, 17537, 17520, 17519, 17546, 17536, 17534, 17538, 17528, 17520, 17539, 17534, 17528, 17641, 17570, 17639, 17652, 17582, 17650, 17650, 17572, 17641, 17664, 17617, 17662, 17672, 17630, 17670, 17670, 17619, 17664, 17662, 17584, 17652, 17639, 17536, 17547, 17693, 17694, 17692, 17694, 17696, 17692, 17702, 17699, 17700, 17694, 17700, 17699, 17694, 17701, 17697, 17696, 17691, 17690, 17691, 17701, 17698, 17704, 17705, 17703, 17712, 17709, 17710, 17714, 17715, 17713, 17722, 17719, 17720, 17706, 17713, 17705, 17720, 17711, 17712, 17704, 17708, 17706, 17717, 17716, 17714, 17711, 17707, 17709, 17719, 17718, 17717, 17719, 17708, 17711, 17706, 17717, 17714, 17691, 17703, 17690, 17702, 17709, 17701, 17709, 17698, 17701, 17691, 17707, 17704, 17727, 17728, 17726, 17728, 17730, 17726, 17735, 17736, 17734, 17736, 17732, 17734, 17741, 17738, 17739, 17729, 17738, 17728, 17728, 17740, 17731, 17736, 17744, 17743, 17743, 17733, 17736, 17732, 17729, 17727, 17742, 17741, 17739, 17729, 17742, 17739, 17748, 17749, 17747,
17749, 17751, 17747, 17757, 17754, 17755, 17759, 17760, 17758, 17767, 17764, 17765, 17746, 17758, 17745, 17757, 17764, 17756, 17759, 17763, 17761, 17749, 17755, 17754, 17754, 17752, 17749, 17751, 17746, 17745, 17766, 17762, 17764, 17746, 17756, 17753, 17764, 17753, 17756, 17753, 17759, 17746, 17771, 17768, 17769, 17775, 17772, 17773, 17779, 17776, 17777, 17783, 17780, 17781, 17777, 17770, 17771, 17781, 17774, 17775, 17774, 17769, 17772, 17782, 17777, 17780, 17780, 17771, 17774, 17724, 17748, 17723, 17724, 17755, 17750, 17741, 17768, 17740, 17773, 17743, 17744, 17772, 17741, 17743, 17730, 17724, 17723, 17731, 17725, 17724, 17768, 17757, 17755, 17778, 17765, 17776, 17776, 17757, 17770, 17768, 17725, 17740, 17796, 17793, 17795, 17791, 17793, 17789, 17796, 17790, 17794, 17799, 17800, 17798, 17796, 17802, 17794, 17801, 17795, 17800, 17791, 17800, 17795, 17801, 17792, 17796, 17791, 17787, 17788, 17797, 17791, 17788, 17801, 17803, 17796, 17806, 17807, 17805, 17810, 17806, 17809, 17814, 17815, 17813, 17821, 17818, 17820, 17825, 17822, 17823, 17819, 17811, 17818, 17823, 17815, 17816, 17808, 17813, 17807, 17810, 17814, 17808, 17827, 17819, 17821, 17811, 17810, 17809, 17827, 17823, 17826, 17812, 17816, 17810, 17817, 17823, 17816, 17819, 17817, 17812,
17807, 17784, 17805, 17813, 17785, 17807, 17716, 17787, 17715, 17799, 17721, 17722, 17798, 17718, 17721, 17716, 17797, 17788, 17784, 17737, 17735, 17785, 17744, 17737, 17815, 17773, 17813, 17824, 17781, 17822, 17822, 17775, 17815, 17786, 17773, 17744, 17451, 17380, 17381, 17455, 17384, 17385, 17454, 17381, 17384, 17467, 17389, 17466, 17394, 17472, 17393, 17473, 17391, 17467, 17472, 17385, 17393, 17831, 17828, 17830, 17831, 17832, 17829, 17837, 17834, 17836, 17839, 17840, 17838, 17840, 17834, 17838, 17837, 17839, 17835, 17852, 17853, 17851, 17859, 17856, 17857, 17833, 17849, 17832, 17857, 17847, 17848, 17843, 17830, 17842, 17843, 17833, 17831, 17846, 17836, 17845, 17841, 17847, 17840, 17847, 17836, 17840, 17846, 17841, 17837, 17852, 17858, 17854, 17845, 17844, 17843, 17852, 17849, 17850, 17856, 17850, 17855, 17856, 17844, 17847, 17844, 17850, 17833, 17863, 17860, 17861, 17867, 17864, 17865, 17873, 17874, 17872, 17879, 17876, 17877, 17873, 17862, 17863, 17877, 17866, 17867, 17866, 17861, 17864, 17871, 17868, 17869, 17876, 17875, 17873, 17878, 17869, 17875, 17876, 17863, 17866, 17885, 17882, 17884, 17881, 17882, 17880, 17887, 17888, 17886, 17891, 17892, 17890, 17885, 17890, 17883, 17889, 17884, 17888, 17881, 17888, 17884, 17894, 17893, 17891,
17885, 17894, 17891, 17903, 17900, 17901, 17907, 17904, 17905, 17911, 17908, 17909, 17915, 17912, 17913, 17909, 17902, 17903, 17913, 17906, 17907, 17898, 17897, 17896, 17906, 17901, 17904, 17914, 17909, 17912, 17899, 17901, 17897, 17912, 17903, 17906, 17854, 17880, 17853, 17859, 17886, 17858, 17858, 17881, 17854, 17897, 17868, 17896, 17871, 17898, 17870, 17875, 17900, 17874, 17879, 17904, 17878, 17901, 17869, 17897, 17878, 17899, 17871, 17862, 17846, 17860, 17872, 17859, 17857, 17862, 17857, 17848, 17902, 17887, 17900, 17908, 17895, 17894, 17902, 17894, 17889, 17900, 17859, 17874, 17919, 17916, 17917, 17923, 17920, 17921, 17929, 17926, 17927, 17931, 17932, 17930, 17927, 17918, 17919, 17931, 17922, 17923, 17922, 17917, 17920, 17932, 17927, 17930, 17930, 17919, 17922, 17939, 17936, 17938, 17941, 17942, 17940, 17949, 17946, 17948, 17951, 17946, 17950, 17949, 17934, 17947, 17953, 17942, 17943, 17953, 17948, 17952, 17951, 17952, 17948, 17939, 17944, 17937, 17939, 17934, 17935, 17945, 17941, 17944, 17943, 17939, 17935, 17953, 17935, 17949, 17958, 17955, 17957, 17962, 17959, 17961, 17964, 17965, 17963, 17974, 17971, 17973, 17978, 17975, 17976, 17970, 17957, 17969, 17966, 17975, 17965, 17966, 17956, 17958, 17967, 17962, 17961, 17980, 17972, 17974,
17961, 17955, 17956, 17971, 17970, 17969, 17979, 17978, 17976, 17964, 17961, 17956, 17972, 17976, 17970, 17976, 17958, 17970, 17962, 17924, 17960, 17929, 17963, 17928, 17968, 17932, 17933, 17967, 17929, 17932, 17968, 17925, 17962, 17942, 17921, 17940, 17951, 17924, 17925, 17954, 17931, 17952, 17954, 17925, 17933, 17942, 17931, 17923, 17985, 17986, 17984, 17989, 17990, 17988, 17995, 17992, 17993, 17997, 17998, 17996, 18001, 18002, 18000, 18007, 18004, 18005, 17987, 17996, 17986, 17991, 18000, 17990, 18005, 17994, 17995, 17985, 17990, 17987, 18000, 17999, 17997, 17994, 17989, 17992, 18006, 18001, 18004, 18004, 17991, 17994, 17990, 17997, 17987, 18013, 18010, 18012, 18009, 18010, 18008, 18015, 18016, 18014, 18021, 18018, 18019, 18023, 18024, 18022, 18031, 18028, 18029, 18013, 18022, 18011, 18029, 18020, 18021, 18017, 18012, 18016, 18009, 18016, 18012, 18026, 18025, 18023, 18020, 18015, 18018, 18028, 18027, 18026, 18020, 18026, 18017, 18013, 18026, 18023, 17999, 18008, 17998, 18003, 18014, 18002, 18019, 18006, 18007, 18018, 18003, 18006, 18002, 18009, 17999, 17985, 17981, 17982, 17988, 17982, 17983, 18035, 18032, 18034, 18039, 18036, 18037, 18045, 18042, 18044, 18049, 18046, 18047, 18043, 18034, 18042, 18047, 18038, 18039, 18041, 18033, 18035,
18051, 18043, 18045, 18040, 18039, 18037, 18050, 18049, 18047, 18041, 18047, 18039, 18050, 18035, 18043, 18055, 18056, 18054, 18056, 18052, 18054, 18061, 18058, 18059, 18068, 18065, 18067, 18072, 18069, 18070, 18076, 18073, 18074, 18066, 18055, 18065, 18070, 18060, 18061, 18064, 18073, 18063, 18057, 18063, 18056, 18056, 18062, 18053, 18076, 18066, 18068, 18062, 18061, 18059, 18075, 18070, 18073, 18073, 18061, 18063, 18066, 18064, 18057, 18078, 18079, 18077, 18082, 18083, 18081, 18090, 18091, 18089, 18096, 18093, 18094, 18100, 18097, 18098, 18079, 18089, 18077, 18094, 18085, 18086, 18088, 18097, 18087, 18079, 18086, 18085, 18090, 18095, 18092, 18083, 18088, 18087, 18081, 18080, 18078, 18099, 18094, 18097, 18080, 18087, 18086, 18097, 18086, 18087, 18085, 18090, 18079, 18053, 18044, 18052, 18059, 18048, 18049, 18051, 18059, 18049, 18062, 18045, 18053, 17893, 17981, 17892, 17893, 17983, 17982, 17989, 17910, 17911, 17993, 17914, 17915, 17992, 17911, 17914, 17910, 17983, 17895, 18033, 17973, 18032, 18037, 17977, 17978, 17980, 18037, 17978, 18040, 17974, 18033, 17918, 17865, 17916, 17928, 17877, 17926, 17926, 17867, 17918, 17963, 17907, 17905, 17977, 17913, 17975, 17975, 17907, 17965, 17928, 17905, 17879, 18038, 17993, 18036, 18048, 18005, 18046,
18046, 17995, 18038, 18060, 18019, 18058, 18071, 18029, 18069, 18069, 18021, 18060, 18058, 18007, 18048, 18036, 17915, 17977, 18107, 18104, 18106, 18105, 18101, 18104, 18109, 18105, 18107, 18108, 18102, 18105, 18111, 18112, 18110, 18122, 18123, 18121, 18126, 18127, 18125, 18130, 18131, 18129, 18112, 18119, 18110, 18123, 18114, 18121, 18126, 18116, 18117, 18113, 18116, 18112, 18123, 18130, 18129, 18114, 18113, 18111, 18131, 18126, 18129, 18113, 18118, 18117, 18118, 18126, 18117, 18112, 18125, 18120, 18129, 18115, 18123, 18134, 18135, 18133, 18142, 18139, 18140, 18148, 18149, 18147, 18149, 18143, 18147, 18146, 18148, 18144, 18160, 18157, 18158, 18136, 18144, 18135, 18158, 18141, 18142, 18137, 18136, 18134, 18150, 18155, 18149, 18155, 18145, 18149, 18154, 18150, 18146, 18139, 18138, 18137, 18152, 18148, 18150, 18159, 18154, 18157, 18159, 18150, 18156, 18141, 18154, 18138, 18138, 18146, 18136, 18164, 18161, 18162, 18168, 18165, 18166, 18172, 18169, 18170, 18176, 18173, 18174, 18170, 18163, 18164, 18174, 18167, 18168, 18167, 18162, 18165, 18175, 18170, 18173, 18173, 18164, 18167, 18180, 18177, 18179, 18182, 18183, 18181, 18186, 18187, 18185, 18192, 18189, 18190, 18196, 18193, 18195, 18198, 18199, 18197, 18202, 18203, 18201, 18206, 18207, 18205,
18188, 18201, 18187, 18206, 18191, 18192, 18180, 18186, 18178, 18204, 18194, 18196, 18189, 18184, 18182, 18205, 18200, 18198, 18183, 18180, 18179, 18198, 18193, 18194, 18184, 18188, 18180, 18205, 18194, 18202, 18191, 18202, 18188, 18162, 18127, 18128, 18166, 18131, 18132, 18165, 18128, 18131, 18178, 18148, 18177, 18152, 18181, 18151, 18186, 18155, 18156, 18160, 18189, 18159, 18186, 18150, 18178, 18152, 18189, 18182, 18143, 18124, 18122, 18153, 18132, 18130, 18145, 18130, 18124, 18187, 18166, 18185, 18203, 18174, 18201, 18201, 18168, 18187, 18185, 18132, 18155, 18212, 18209, 18211, 18216, 18213, 18214, 18220, 18217, 18218, 18224, 18221, 18223, 18228, 18225, 18226, 18232, 18229, 18230, 18222, 18211, 18221, 18226, 18215, 18216, 18220, 18229, 18219, 18220, 18210, 18212, 18224, 18230, 18222, 18219, 18214, 18217, 18231, 18226, 18229, 18229, 18216, 18219, 18222, 18220, 18212, 18236, 18233, 18235, 18236, 18237, 18234, 18244, 18241, 18243, 18244, 18239, 18242, 18251, 18248, 18249, 18246, 18235, 18245, 18246, 18238, 18236, 18251, 18243, 18250, 18244, 18249, 18240, 18243, 18237, 18238, 18248, 18247, 18246, 18247, 18243, 18238, 18255, 18252, 18254, 18257, 18252, 18256, 18261, 18258, 18259, 18268, 18265, 18267, 18272, 18269, 18270, 18266, 18256, 18265,
18270, 18260, 18261, 18262, 18255, 18254, 18264, 18254, 18257, 18268, 18273, 18266, 18264, 18259, 18262, 18274, 18270, 18273, 18273, 18261, 18264, 18266, 18264, 18257, 18255, 18223, 18253, 18259, 18227, 18228, 18263, 18231, 18232, 18231, 18259, 18228, 18263, 18224, 18255, 18276, 18277, 18275, 18284, 18281, 18282, 18286, 18287, 18285, 18292, 18289, 18291, 18296, 18293, 18295, 18296, 18279, 18294, 18284, 18297, 18283, 18300, 18287, 18288, 18276, 18283, 18278, 18298, 18291, 18297, 18300, 18295, 18299, 18287, 18282, 18285, 18295, 18290, 18292, 18298, 18295, 18292, 18287, 18298, 18284, 18296, 18288, 18280, 18306, 18303, 18305, 18312, 18309, 18310, 18314, 18315, 18313, 18318, 18319, 18317, 18319, 18321, 18317, 18326, 18323, 18324, 18330, 18327, 18328, 18306, 18318, 18304, 18328, 18315, 18316, 18311, 18306, 18305, 18319, 18324, 18323, 18319, 18325, 18322, 18307, 18302, 18301, 18315, 18310, 18313, 18327, 18326, 18324, 18302, 18313, 18310, 18315, 18324, 18312, 18306, 18324, 18320, 18332, 18333, 18331, 18340, 18341, 18339, 18338, 18340, 18337, 18334, 18337, 18333, 18335, 18334, 18332, 18341, 18345, 18344, 18338, 18345, 18342, 18334, 18343, 18338, 18349, 18350, 18348, 18355, 18352, 18353, 18359, 18356, 18357, 18363, 18360, 18361, 18357, 18350, 18351,
18361, 18354, 18355, 18354, 18349, 18352, 18362, 18357, 18360, 18360, 18351, 18354, 18322, 18346, 18321, 18349, 18325, 18326, 18353, 18329, 18330, 18352, 18326, 18329, 18322, 18348, 18347, 18305, 18279, 18280, 18311, 18286, 18309, 18288, 18305, 18280, 18346, 18332, 18331, 18350, 18335, 18348, 18358, 18343, 18356, 18332, 18348, 18335, 18356, 18336, 18350, 18369, 18366, 18368, 18371, 18372, 18370, 18379, 18376, 18378, 18383, 18380, 18381, 18377, 18364, 18376, 18373, 18380, 18372, 18379, 18381, 18377, 18375, 18367, 18369, 18369, 18364, 18365, 18375, 18371, 18374, 18373, 18369, 18365, 18381, 18365, 18377, 18387, 18384, 18386, 18391, 18388, 18389, 18395, 18392, 18394, 18397, 18392, 18396, 18395, 18386, 18393, 18399, 18390, 18391, 18387, 18389, 18385, 18398, 18395, 18394, 18397, 18398, 18394, 18399, 18387, 18395, 18385, 18378, 18384, 18389, 18382, 18383, 18385, 18383, 18379, 18282, 18171, 18172, 18286, 18175, 18176, 18285, 18172, 18175, 18302, 18195, 18301, 18200, 18307, 18199, 18204, 18309, 18203, 18314, 18207, 18208, 18302, 18204, 18196, 18200, 18313, 18308, 18309, 18176, 18203, 18367, 18267, 18366, 18371, 18271, 18272, 18374, 18272, 18274, 18367, 18274, 18268, 18215, 18140, 18213, 18227, 18158, 18225, 18225, 18142, 18215, 18260, 18190, 18258,
18271, 18206, 18269, 18269, 18192, 18260, 18258, 18160, 18227, 18372, 18314, 18370, 18382, 18328, 18380, 18380, 18316, 18372, 18390, 18353, 18388, 18400, 18361, 18398, 18390, 18361, 18355, 18388, 18330, 18382, 18370, 18208, 18271, 18025, 18133, 18024, 18140, 18030, 18031, 18139, 18027, 18030, 18027, 18134, 18025, 18210, 18067, 18209, 18214, 18071, 18072, 18076, 18217, 18075, 18217, 18072, 18075, 18218, 18068, 18210, 18092, 18233, 18091, 18096, 18245, 18095, 18100, 18248, 18099, 18248, 18096, 18099, 18092, 18245, 18235, 18213, 18031, 18071, 18402, 18106, 18401, 18403, 18107, 18402, 18101, 18084, 18082, 18108, 18100, 18098, 18102, 18088, 18084, 18103, 18098, 18088, 18402, 18239, 18240, 18403, 18240, 18249, 18403, 18100, 18109, 18405, 18406, 18404, 18409, 18405, 18408, 18413, 18410, 18411, 18415, 18416, 18414, 18421, 18418, 18419, 18425, 18422, 18423, 18419, 18412, 18413, 18423, 18416, 18417, 18407, 18414, 18406, 18409, 18415, 18407, 18416, 18411, 18414, 18424, 18419, 18422, 18422, 18413, 18416, 18429, 18430, 18428, 18438, 18439, 18437, 18439, 18435, 18437, 18436, 18426, 18435, 18440, 18442, 18439, 18442, 18436, 18439, 18433, 18431, 18429, 18441, 18427, 18436, 18447, 18444, 18445, 18451, 18448, 18449, 18455, 18452, 18453, 18459, 18456, 18457,
18453, 18446, 18447, 18457, 18450, 18451, 18450, 18445, 18448, 18458, 18453, 18456, 18456, 18447, 18450, 18463, 18460, 18462, 18467, 18468, 18466, 18474, 18475, 18473, 18475, 18471, 18473, 18480, 18477, 18478, 18472, 18464, 18471, 18478, 18468, 18469, 18469, 18461, 18463, 18476, 18482, 18475, 18482, 18472, 18475, 18464, 18463, 18462, 18481, 18480, 18478, 18465, 18469, 18463, 18470, 18478, 18469, 18481, 18465, 18472, 18445, 18420, 18421, 18449, 18424, 18425, 18448, 18421, 18424, 18461, 18438, 18460, 18443, 18466, 18442, 18467, 18440, 18461, 18426, 18409, 18408, 18441, 18425, 18423, 18427, 18417, 18409, 18432, 18423, 18417, 18468, 18449, 18466, 18479, 18457, 18477, 18477, 18451, 18468, 18466, 18425, 18442, 18489, 18486, 18487, 18493, 18490, 18491, 18499, 18496, 18498, 18495, 18496, 18494, 18501, 18502, 18500, 18507, 18504, 18505, 18485, 18494, 18484, 18489, 18500, 18488, 18505, 18492, 18493, 18503, 18498, 18502, 18500, 18498, 18495, 18492, 18487, 18490, 18506, 18501, 18504, 18504, 18489, 18492, 18488, 18495, 18485, 18511, 18508, 18510, 18515, 18516, 18514, 18522, 18519, 18521, 18526, 18523, 18524, 18520, 18512, 18519, 18524, 18516, 18517, 18517, 18509, 18511, 18528, 18520, 18522, 18512, 18511, 18510, 18527, 18526, 18524, 18513, 18517, 18511,
18518, 18524, 18517, 18527, 18513, 18520, 18530, 18531, 18529, 18540, 18537, 18538, 18542, 18543, 18541, 18546, 18547, 18545, 18534, 18541, 18533, 18540, 18545, 18539, 18535, 18532, 18530, 18545, 18544, 18542, 18534, 18531, 18532, 18537, 18536, 18535, 18539, 18532, 18536, 18539, 18542, 18534, 18550, 18551, 18549, 18554, 18550, 18553, 18558, 18555, 18556, 18560, 18561, 18559, 18566, 18563, 18565, 18570, 18567, 18568, 18572, 18573, 18571, 18564, 18553, 18563, 18568, 18557, 18558, 18562, 18571, 18561, 18552, 18559, 18551, 18554, 18560, 18552, 18566, 18572, 18564, 18561, 18556, 18559, 18573, 18568, 18571, 18571, 18558, 18561, 18572, 18554, 18564, 18499, 18529, 18497, 18538, 18506, 18507, 18506, 18535, 18503, 18503, 18530, 18499, 18551, 18521, 18549, 18556, 18525, 18526, 18528, 18556, 18526, 18551, 18528, 18522, 18516, 18491, 18514, 18525, 18505, 18523, 18523, 18493, 18516, 18557, 18538, 18555, 18569, 18546, 18567, 18567, 18540, 18557, 18555, 18507, 18525, 18576, 18577, 18575, 18580, 18581, 18579, 18586, 18583, 18584, 18590, 18587, 18589, 18590, 18591, 18588, 18601, 18598, 18599, 18578, 18587, 18577, 18582, 18595, 18581, 18599, 18585, 18586, 18579, 18578, 18576, 18596, 18589, 18595, 18590, 18597, 18592, 18585, 18580, 18583, 18594, 18591, 18592,
18598, 18597, 18596, 18600, 18592, 18597, 18598, 18582, 18585, 18581, 18589, 18578, 18605, 18602, 18604, 18609, 18606, 18607, 18611, 18612, 18610, 18619, 18616, 18617, 18623, 18620, 18621, 18615, 18604, 18614, 18617, 18608, 18609, 18613, 18620, 18612, 18613, 18603, 18605, 18612, 18607, 18610, 18622, 18617, 18620, 18620, 18609, 18612, 18615, 18613, 18605, 18629, 18626, 18628, 18625, 18626, 18624, 18631, 18632, 18630, 18635, 18636, 18634, 18629, 18634, 18627, 18633, 18628, 18632, 18630, 18628, 18625, 18638, 18637, 18635, 18633, 18635, 18629, 18643, 18640, 18641, 18647, 18644, 18645, 18653, 18650, 18651, 18657, 18654, 18655, 18651, 18642, 18643, 18655, 18646, 18647, 18646, 18641, 18644, 18656, 18651, 18654, 18654, 18643, 18646, 18594, 18624, 18593, 18601, 18630, 18600, 18600, 18625, 18594, 18641, 18618, 18619, 18645, 18622, 18623, 18644, 18619, 18622, 18608, 18584, 18606, 18618, 18599, 18616, 18616, 18586, 18608, 18642, 18631, 18640, 18649, 18636, 18637, 18650, 18639, 18638, 18652, 18637, 18639, 18642, 18638, 18633, 18640, 18601, 18618, 18661, 18658, 18660, 18661, 18662, 18659, 18670, 18667, 18669, 18663, 18671, 18662, 18670, 18673, 18668, 18664, 18661, 18660, 18665, 18663, 18661, 18663, 18675, 18672, 18679, 18676, 18677, 18683, 18680, 18681,
18685, 18686, 18684, 18689, 18690, 18688, 18695, 18692, 18693, 18689, 18678, 18679, 18693, 18682, 18683, 18685, 18690, 18687, 18682, 18677, 18680, 18694, 18689, 18692, 18692, 18679, 18682, 18697, 18698, 18696, 18701, 18697, 18700, 18703, 18704, 18702, 18709, 18706, 18708, 18713, 18714, 18712, 18707, 18700, 18706, 18705, 18712, 18704, 18699, 18702, 18698, 18705, 18699, 18701, 18715, 18707, 18709, 18710, 18709, 18708, 18711, 18715, 18709, 18713, 18701, 18707, 18722, 18719, 18721, 18718, 18719, 18717, 18724, 18725, 18723, 18730, 18727, 18728, 18732, 18733, 18731, 18740, 18737, 18738, 18722, 18731, 18720, 18738, 18729, 18730, 18726, 18721, 18725, 18718, 18725, 18721, 18735, 18734, 18732, 18729, 18724, 18727, 18737, 18736, 18735, 18729, 18735, 18726, 18722, 18735, 18732, 18698, 18669, 18696, 18702, 18670, 18698, 18687, 18717, 18686, 18691, 18723, 18690, 18728, 18694, 18695, 18727, 18691, 18694, 18690, 18718, 18687, 18676, 18666, 18665, 18685, 18671, 18672, 18688, 18672, 18675, 18678, 18675, 18666, 18580, 18454, 18455, 18584, 18458, 18459, 18583, 18455, 18458, 18603, 18474, 18602, 18607, 18479, 18480, 18483, 18610, 18482, 18610, 18480, 18482, 18611, 18476, 18603, 18606, 18459, 18479, 18544, 18658, 18543, 18548, 18664, 18547, 18547, 18660, 18544,
18677, 18569, 18570, 18681, 18573, 18574, 18680, 18570, 18573, 18676, 18548, 18569, 18485, 18430, 18431, 18486, 18434, 18433, 18488, 18431, 18434, 18667, 18615, 18614, 18673, 18623, 18621, 18668, 18621, 18615, 18704, 18645, 18702, 18714, 18655, 18712, 18712, 18647, 18704, 18674, 18645, 18623, 18742, 18743, 18741, 18752, 18749, 18750, 18756, 18753, 18755, 18756, 18757, 18754, 18746, 18753, 18745, 18752, 18759, 18751, 18742, 18748, 18744, 18759, 18756, 18755, 18760, 18758, 18756, 18746, 18743, 18744, 18749, 18748, 18747, 18751, 18744, 18748, 18746, 18759, 18755, 18765, 18762, 18763, 18769, 18766, 18767, 18773, 18770, 18772, 18777, 18774, 18776, 18779, 18780, 18778, 18785, 18782, 18783, 18779, 18764, 18765, 18783, 18768, 18769, 18780, 18773, 18772, 18784, 18777, 18776, 18768, 18763, 18766, 18776, 18771, 18773, 18782, 18781, 18779, 18784, 18773, 18781, 18782, 18765, 18768, 18764, 18750, 18762, 18772, 18757, 18758, 18778, 18761, 18760, 18761, 18772, 18758, 18778, 18752, 18764, 18794, 18791, 18793, 18796, 18797, 18795, 18804, 18801, 18803, 18806, 18801, 18805, 18804, 18789, 18802, 18808, 18797, 18798, 18808, 18803, 18807, 18806, 18807, 18803, 18794, 18799, 18792, 18794, 18789, 18790, 18800, 18796, 18799, 18798, 18794, 18790, 18808, 18790, 18804,
18813, 18814, 18812, 18811, 18813, 18810, 18814, 18818, 18817, 18816, 18815, 18811, 18822, 18819, 18821, 18828, 18829, 18827, 18829, 18825, 18827, 18826, 18821, 18825, 18824, 18820, 18822, 18830, 18832, 18829, 18829, 18831, 18826, 18831, 18822, 18826, 18837, 18834, 18836, 18839, 18840, 18838, 18845, 18842, 18844, 18847, 18848, 18846, 18843, 18836, 18842, 18841, 18846, 18840, 18841, 18835, 18837, 18849, 18843, 18845, 18847, 18837, 18843, 18835, 18828, 18834, 18833, 18838, 18832, 18835, 18833, 18830, 18859, 18860, 18858, 18866, 18863, 18864, 18868, 18869, 18867, 18869, 18871, 18867, 18876, 18873, 18874, 18880, 18877, 18878, 18857, 18868, 18856, 18866, 18877, 18865, 18869, 18874, 18873, 18873, 18872, 18869, 18859, 18865, 18861, 18860, 18857, 18856, 18877, 18876, 18874, 18857, 18865, 18862, 18877, 18862, 18865, 18862, 18870, 18857, 18884, 18885, 18883, 18885, 18887, 18883, 18893, 18890, 18891, 18895, 18896, 18894, 18896, 18898, 18894, 18903, 18900, 18901, 18907, 18904, 18905, 18882, 18895, 18881, 18893, 18904, 18892, 18897, 18900, 18896, 18900, 18899, 18896, 18886, 18890, 18885, 18885, 18892, 18888, 18887, 18882, 18881, 18904, 18903, 18901, 18888, 18889, 18882, 18904, 18889, 18892, 18882, 18901, 18897, 18911, 18908, 18909, 18915, 18912, 18913,
18921, 18918, 18920, 18921, 18916, 18919, 18925, 18922, 18923, 18929, 18926, 18927, 18923, 18910, 18911, 18915, 18926, 18914, 18928, 18921, 18920, 18929, 18917, 18921, 18914, 18909, 18912, 18926, 18925, 18923, 18926, 18911, 18914, 18854, 18884, 18853, 18854, 18891, 18886, 18876, 18908, 18875, 18913, 18879, 18880, 18912, 18876, 18879, 18871, 18854, 18853, 18872, 18855, 18854, 18908, 18893, 18891, 18924, 18905, 18922, 18922, 18893, 18910, 18908, 18855, 18875, 18820, 18786, 18819, 18820, 18788, 18787, 18814, 18858, 18812, 18864, 18817, 18818, 18814, 18863, 18859, 18797, 18767, 18795, 18806, 18775, 18777, 18809, 18783, 18807, 18785, 18806, 18777, 18797, 18783, 18769, 18898, 18851, 18850, 18851, 18902, 18852, 18935, 18932, 18934, 18931, 18932, 18930, 18937, 18938, 18936, 18943, 18940, 18941, 18945, 18946, 18944, 18953, 18950, 18951, 18935, 18944, 18933, 18951, 18942, 18943, 18939, 18934, 18938, 18936, 18934, 18931, 18948, 18947, 18945, 18942, 18937, 18940, 18950, 18949, 18948, 18942, 18948, 18939, 18935, 18948, 18945, 18957, 18958, 18956, 18962, 18963, 18961, 18970, 18967, 18969, 18966, 18967, 18965, 18972, 18973, 18971, 18959, 18965, 18958, 18964, 18971, 18963, 18961, 18959, 18957, 18974, 18969, 18973, 18966, 18973, 18969, 18957, 18954, 18955,
18961, 18955, 18960, 18963, 18966, 18959, 18947, 18954, 18946, 18962, 18952, 18953, 18952, 18960, 18949, 18949, 18955, 18947, 18978, 18975, 18977, 18982, 18979, 18980, 18988, 18985, 18987, 18992, 18989, 18990, 18986, 18977, 18985, 18990, 18981, 18982, 18984, 18976, 18978, 18994, 18986, 18988, 18983, 18982, 18980, 18993, 18992, 18990, 18984, 18990, 18982, 18993, 18978, 18986, 18996, 18997, 18995, 19006, 19003, 19004, 19008, 19009, 19007, 19012, 19013, 19011, 19000, 19007, 18999, 19006, 19011, 19005, 18996, 19002, 18998, 19011, 19010, 19008, 19000, 18997, 18998, 19003, 19002, 19001, 19005, 18998, 19002, 19000, 19011, 19008, 19016, 19017, 19015, 19020, 19016, 19019, 19024, 19021, 19022, 19026, 19027, 19025, 19032, 19029, 19031, 19036, 19033, 19034, 19038, 19039, 19037, 19030, 19019, 19029, 19034, 19023, 19024, 19028, 19037, 19027, 19018, 19025, 19017, 19020, 19026, 19018, 19040, 19030, 19032, 19027, 19022, 19025, 19039, 19034, 19037, 19037, 19024, 19027, 19038, 19020, 19030, 19042, 19043, 19041, 19046, 19047, 19045, 19050, 19051, 19049, 19056, 19053, 19054, 19044, 19049, 19043, 19048, 19053, 19047, 19045, 19044, 19042, 19053, 19052, 19050, 19047, 19050, 19044, 19017, 18987, 19015, 19022, 18991, 18992, 18994, 19022, 18992, 19017, 18994, 18988,
19010, 19041, 19009, 19014, 19045, 19013, 19013, 19042, 19010, 19061, 19062, 19060, 19065, 19066, 19064, 19071, 19068, 19069, 19075, 19072, 19073, 19058, 19065, 19057, 19062, 19069, 19068, 19067, 19072, 19066, 19064, 19063, 19061, 19074, 19069, 19072, 19063, 19072, 19069, 19058, 19073, 19067, 19077, 19078, 19076, 19083, 19084, 19082, 19081, 19083, 19080, 19094, 19091, 19092, 19098, 19095, 19096, 19092, 19086, 19087, 19090, 19095, 19089, 19078, 19087, 19086, 19084, 19090, 19089, 19088, 19085, 19081, 19082, 19079, 19077, 19097, 19092, 19095, 19079, 19089, 19087, 19095, 19087, 19089, 19104, 19101, 19102, 19106, 19107, 19105, 19112, 19109, 19111, 19116, 19113, 19115, 19116, 19099, 19114, 19118, 19103, 19104, 19120, 19107, 19108, 19117, 19112, 19111, 19120, 19115, 19119, 19107, 19102, 19105, 19115, 19110, 19112, 19118, 19115, 19112, 19119, 19104, 19107, 19116, 19108, 19100, 19124, 19121, 19123, 19128, 19125, 19127, 19132, 19129, 19130, 19136, 19133, 19134, 19131, 19124, 19123, 19135, 19128, 19127, 19127, 19122, 19124, 19135, 19130, 19133, 19132, 19127, 19124, 19102, 19070, 19071, 19106, 19074, 19075, 19105, 19071, 19074, 19130, 19093, 19094, 19134, 19097, 19098, 19133, 19094, 19097, 19076, 19058, 19057, 19091, 19075, 19073, 19078, 19059, 19058,
19091, 19059, 19086, 19123, 19099, 19100, 19129, 19108, 19106, 19131, 19100, 19108, 19129, 19075, 19093, 19140, 19137, 19139, 19144, 19141, 19142, 19148, 19145, 19146, 19152, 19149, 19151, 19156, 19153, 19154, 19160, 19157, 19158, 19150, 19139, 19149, 19154, 19143, 19144, 19158, 19147, 19148, 19148, 19138, 19140, 19160, 19150, 19152, 19147, 19142, 19145, 19159, 19154, 19157, 19157, 19144, 19147, 19158, 19140, 19150, 19162, 19163, 19161, 19168, 19165, 19166, 19170, 19171, 19169, 19174, 19175, 19173, 19164, 19169, 19163, 19168, 19173, 19167, 19162, 19167, 19164, 19170, 19175, 19172, 19164, 19173, 19170, 19182, 19179, 19181, 19186, 19183, 19184, 19188, 19189, 19187, 19192, 19193, 19191, 19198, 19195, 19197, 19202, 19199, 19200, 19206, 19203, 19204, 19178, 19191, 19177, 19196, 19181, 19195, 19186, 19199, 19185, 19190, 19203, 19189, 19199, 19194, 19192, 19182, 19188, 19180, 19198, 19204, 19196, 19189, 19184, 19187, 19205, 19200, 19203, 19203, 19186, 19189, 19185, 19192, 19178, 19196, 19190, 19182, 19208, 19209, 19207, 19214, 19211, 19212, 19218, 19219, 19217, 19225, 19222, 19223, 19210, 19217, 19209, 19223, 19213, 19214, 19208, 19213, 19210, 19222, 19220, 19218, 19219, 19216, 19215, 19216, 19224, 19221, 19210, 19222, 19218, 19180, 19151, 19179,
19184, 19155, 19156, 19160, 19187, 19159, 19187, 19156, 19159, 19180, 19160, 19152, 19172, 19207, 19171, 19212, 19175, 19176, 19172, 19211, 19208, 18970, 19080, 18968, 18970, 19088, 19081, 19138, 19031, 19137, 19142, 19035, 19036, 19040, 19145, 19039, 19145, 19036, 19039, 19146, 19032, 19138, 19052, 19161, 19051, 19166, 19055, 19056, 19055, 19162, 19052, 18981, 18941, 18979, 18991, 18951, 18989, 18989, 18943, 18981, 19023, 18962, 19021, 19035, 18972, 19033, 19033, 18964, 19023, 19021, 18953, 18991, 19141, 19090, 19088, 19155, 19096, 19153, 19143, 19096, 19090, 19178, 19126, 19128, 19185, 19134, 19183, 19136, 19178, 19128, 19183, 19098, 19155, 19035, 19088, 18974, 19229, 19226, 19228, 19231, 19232, 19230, 19239, 19240, 19238, 19240, 19236, 19238, 19245, 19242, 19243, 19235, 19228, 19234, 19233, 19242, 19232, 19233, 19227, 19229, 19240, 19248, 19247, 19240, 19246, 19237, 19236, 19235, 19234, 19246, 19245, 19243, 19237, 19243, 19235, 19235, 19233, 19229, 19252, 19253, 19251, 19253, 19255, 19251, 19261, 19258, 19259, 19263, 19264, 19262, 19264, 19266, 19262, 19271, 19268, 19269, 19275, 19272, 19273, 19250, 19263, 19249, 19261, 19272, 19260, 19265, 19268, 19264, 19264, 19270, 19267, 19253, 19259, 19258, 19258, 19256, 19253, 19255, 19250, 19249,
19272, 19271, 19269, 19256, 19257, 19250, 19272, 19257, 19260, 19250, 19269, 19265, 19279, 19276, 19277, 19281, 19282, 19280, 19287, 19284, 19285, 19291, 19288, 19289, 19285, 19278, 19279, 19289, 19282, 19283, 19282, 19277, 19280, 19290, 19285, 19288, 19288, 19279, 19282, 19295, 19292, 19293, 19299, 19296, 19297, 19303, 19300, 19302, 19307, 19304, 19306, 19309, 19304, 19308, 19313, 19310, 19311, 19311, 19294, 19295, 19315, 19298, 19299, 19312, 19303, 19302, 19314, 19307, 19306, 19309, 19314, 19306, 19298, 19293, 19296, 19309, 19301, 19303, 19316, 19311, 19314, 19313, 19309, 19303, 19298, 19311, 19295, 19277, 19244, 19245, 19248, 19280, 19247, 19247, 19277, 19245, 19271, 19292, 19270, 19297, 19274, 19275, 19296, 19271, 19274, 19266, 19241, 19239, 19241, 19270, 19248, 19294, 19281, 19292, 19310, 19291, 19289, 19310, 19283, 19294, 19292, 19248, 19270, 19320, 19317, 19318, 19324, 19321, 19322, 19330, 19327, 19329, 19330, 19325, 19328, 19334, 19331, 19332, 19338, 19335, 19336, 19332, 19319, 19320, 19324, 19335, 19323, 19338, 19329, 19337, 19330, 19336, 19326, 19323, 19318, 19321, 19335, 19334, 19332, 19335, 19320, 19323, 19344, 19341, 19343, 19344, 19339, 19342, 19350, 19347, 19348, 19353, 19341, 19352, 19350, 19343, 19349, 19344, 19348, 19340,
19346, 19339, 19340, 19348, 19346, 19340, 19353, 19349, 19343, 19360, 19357, 19359, 19360, 19355, 19358, 19366, 19363, 19364, 19369, 19357, 19368, 19365, 19360, 19359, 19360, 19364, 19356, 19362, 19355, 19356, 19367, 19356, 19364, 19370, 19359, 19369, 19362, 19327, 19361, 19334, 19363, 19333, 19367, 19334, 19337, 19362, 19337, 19329, 19347, 19324, 19322, 19353, 19325, 19326, 19354, 19326, 19336, 19354, 19324, 19349, 19376, 19373, 19375, 19378, 19373, 19377, 19380, 19381, 19379, 19389, 19386, 19388, 19393, 19390, 19391, 19387, 19371, 19386, 19391, 19381, 19382, 19389, 19391, 19387, 19384, 19375, 19383, 19378, 19383, 19375, 19378, 19371, 19372, 19383, 19382, 19380, 19385, 19372, 19382, 19387, 19382, 19372, 19400, 19397, 19399, 19404, 19401, 19402, 19408, 19405, 19407, 19410, 19405, 19409, 19408, 19399, 19406, 19412, 19403, 19404, 19400, 19402, 19398, 19411, 19408, 19407, 19413, 19407, 19410, 19412, 19400, 19408, 19398, 19388, 19397, 19402, 19392, 19393, 19398, 19393, 19389, 19395, 19374, 19376, 19384, 19395, 19376, 19380, 19286, 19287, 19384, 19290, 19291, 19290, 19380, 19287, 19395, 19300, 19394, 19395, 19312, 19302, 19396, 19291, 19312, 19317, 19261, 19259, 19333, 19273, 19331, 19331, 19261, 19319, 19363, 19299, 19297, 19369, 19305, 19307,
19315, 19369, 19307, 19370, 19299, 19365, 19363, 19275, 19333, 18649, 18930, 18648, 18653, 18936, 18652, 18941, 18656, 18657, 18940, 18653, 18656, 18652, 18931, 18649, 18976, 18710, 18975, 18980, 18714, 18715, 18716, 18980, 18715, 18976, 18716, 18711, 18734, 18995, 18733, 19004, 18739, 18740, 18739, 19001, 18736, 18736, 18996, 18734, 18979, 18657, 18714, 19227, 18844, 19226, 18849, 19230, 18848, 19231, 18845, 19227, 18851, 19252, 18850, 18852, 19254, 18851, 18903, 19317, 18902, 19322, 18906, 18907, 19321, 18903, 18906, 19346, 18918, 19345, 18925, 19347, 18924, 19351, 18925, 18928, 19351, 18920, 19346, 19347, 18907, 18924, 19317, 18852, 18902, 18786, 18566, 18565, 18787, 18574, 18566, 18823, 18683, 18681, 18831, 18695, 18693, 18824, 18693, 18683, 18840, 18728, 18838, 18848, 18738, 18846, 18846, 18730, 18840, 18838, 18695, 18832, 18788, 18681, 18574, 19232, 19004, 19230, 19244, 19012, 19242, 19242, 19006, 19232, 19278, 19046, 19276, 19286, 19054, 19284, 19284, 19048, 19278, 19276, 19014, 19244, 19381, 19166, 19379, 19392, 19174, 19390, 19390, 19168, 19381, 19403, 19212, 19401, 19413, 19223, 19411, 19403, 19223, 19214, 19401, 19176, 19392, 19379, 19056, 19286, 19230, 18740, 18848, 19415, 19416, 19414, 19421, 19418, 19419, 19423, 19424, 19422,
19431, 19432, 19430, 19417, 19426, 19416, 19431, 19420, 19421, 19423, 19429, 19425, 19415, 19420, 19417, 19426, 19423, 19422, 19432, 19428, 19430, 19423, 19430, 19428, 19420, 19427, 19417, 19439, 19436, 19437, 19441, 19442, 19440, 19449, 19446, 19448, 19449, 19444, 19447, 19453, 19450, 19451, 19457, 19454, 19455, 19445, 19434, 19444, 19451, 19438, 19439, 19455, 19442, 19443, 19456, 19449, 19448, 19457, 19445, 19449, 19442, 19437, 19440, 19454, 19453, 19451, 19454, 19439, 19442, 19445, 19443, 19435, 19459, 19460, 19458, 19460, 19462, 19458, 19467, 19464, 19465, 19471, 19468, 19469, 19475, 19476, 19474, 19481, 19478, 19479, 19463, 19472, 19462, 19475, 19466, 19467, 19479, 19470, 19471, 19460, 19465, 19464, 19460, 19466, 19463, 19468, 19467, 19465, 19480, 19475, 19478, 19478, 19467, 19470, 19463, 19474, 19473, 19487, 19484, 19486, 19489, 19490, 19488, 19499, 19496, 19498, 19499, 19494, 19497, 19503, 19500, 19501, 19495, 19482, 19494, 19501, 19490, 19491, 19502, 19499, 19498, 19499, 19501, 19495, 19487, 19492, 19485, 19487, 19482, 19483, 19493, 19489, 19492, 19491, 19487, 19483, 19495, 19491, 19483, 19425, 19459, 19424, 19469, 19432, 19433, 19468, 19429, 19432, 19425, 19465, 19461, 19485, 19446, 19484, 19489, 19452, 19453, 19492, 19453, 19456,
19492, 19448, 19485, 19438, 19419, 19436, 19452, 19431, 19450, 19450, 19421, 19438, 19490, 19469, 19488, 19500, 19481, 19479, 19500, 19471, 19490, 19488, 19433, 19452, 19507, 19504, 19506, 19507, 19508, 19505, 19515, 19512, 19513, 19517, 19518, 19516, 19523, 19520, 19522, 19525, 19520, 19524, 19529, 19526, 19527, 19523, 19510, 19521, 19527, 19514, 19515, 19531, 19518, 19519, 19530, 19523, 19522, 19525, 19530, 19522, 19518, 19513, 19516, 19532, 19527, 19530, 19518, 19527, 19515, 19531, 19511, 19523, 19536, 19533, 19535, 19536, 19537, 19534, 19544, 19541, 19543, 19544, 19539, 19542, 19550, 19547, 19548, 19553, 19541, 19552, 19549, 19544, 19543, 19550, 19540, 19544, 19546, 19539, 19540, 19548, 19546, 19540, 19554, 19543, 19553, 19546, 19524, 19545, 19529, 19547, 19528, 19551, 19529, 19532, 19546, 19532, 19525, 19536, 19510, 19511, 19538, 19517, 19537, 19519, 19536, 19511, 19513, 19476, 19477, 19517, 19480, 19481, 19516, 19477, 19480, 19534, 19496, 19533, 19534, 19502, 19498, 19537, 19481, 19502, 19507, 19434, 19435, 19509, 19441, 19508, 19443, 19507, 19435, 18406, 17599, 18404, 18411, 17603, 17604, 17606, 18411, 17604, 18414, 17600, 18406, 17627, 18428, 17625, 17632, 18429, 17627, 18487, 17672, 17673, 18491, 17676, 17677, 18490, 17673, 17676,
18509, 17684, 18508, 17689, 18514, 17688, 18515, 17686, 18509, 18514, 17677, 17688, 17672, 18433, 17632, 17761, 18741, 17760, 18750, 17766, 17767, 17766, 18747, 17763, 17763, 18742, 17761, 18763, 17778, 17779, 18767, 17782, 17783, 18766, 17779, 17782, 18762, 17767, 17778, 18792, 17820, 18791, 18796, 17824, 17825, 18799, 17825, 17827, 18792, 17827, 17821, 18795, 17783, 17824, 18341, 19414, 18339, 19419, 18344, 18345, 18341, 19418, 19415, 19437, 18358, 18359, 19441, 18362, 18363, 19440, 18359, 18362, 18358, 19419, 18345, 19505, 18396, 19504, 19505, 18400, 18397, 19508, 18363, 18400, 18119, 17695, 17693, 18127, 17700, 18125, 18120, 17700, 17695, 18163, 17710, 18161, 18171, 17720, 18169, 18169, 17712, 18163, 18161, 17702, 18127, 18276, 17790, 17792, 18277, 17796, 17794, 18291, 17802, 17803, 18281, 17792, 17799, 17796, 18283, 17801, 17804, 18291, 17803, 18283, 17804, 17801, 18281, 17722, 18171, 19472, 18811, 18810, 19476, 18816, 19474, 18811, 19474, 18816, 19514, 18864, 19512, 19528, 18878, 19526, 19526, 18866, 19514, 19547, 18915, 18913, 19553, 18916, 18917, 18927, 19553, 18917, 19554, 18915, 19549, 19547, 18880, 19528, 19476, 18864, 18818, 17256, 16698, 16699, 17260, 16702, 16703, 17259, 16699, 16702, 17280, 16714, 16715, 17284, 16718, 16719,
17283, 16715, 16718, 17279, 16703, 16714, 17315, 16760, 17314, 17321, 16764, 16765, 17325, 16768, 16769, 17324, 16765, 16768, 17325, 16761, 17315, 17320, 16719, 16764, 17134, 17828, 17133, 17138, 17842, 17137, 17142, 17845, 17141, 17845, 17138, 17141, 17134, 17842, 17830, 17861, 17157, 17158, 17865, 17161, 17162, 17864, 17158, 17161, 17860, 17142, 17157, 17917, 17218, 17219, 17921, 17222, 17223, 17920, 17219, 17222, 17937, 17239, 17936, 17941, 17245, 17246, 17944, 17246, 17249, 17937, 17249, 17241, 17940, 17223, 17245, 17916, 17162, 17218, 19556, 19557, 19555, 19562, 19563, 19561, 19558, 19561, 19557, 19559, 19558, 19556, 19562, 19566, 19564, 19558, 19565, 19562, 19568, 19569, 19567, 19576, 19573, 19575, 19576, 19571, 19574, 19580, 19577, 19578, 19584, 19581, 19582, 19586, 19587, 19585, 19592, 19589, 19591, 19596, 19593, 19594, 19570, 19585, 19569, 19590, 19573, 19589, 19580, 19593, 19579, 19568, 19579, 19570, 19586, 19595, 19588, 19583, 19576, 19575, 19584, 19572, 19576, 19592, 19597, 19590, 19581, 19580, 19578, 19598, 19594, 19597, 19597, 19580, 19583, 19570, 19593, 19586, 19597, 19575, 19590, 19600, 19601, 19599, 19608, 19609, 19607, 19614, 19615, 19613, 19612, 19614, 19611, 19620, 19617, 19619, 19627, 19624, 19625, 19602, 19611, 19601,
19618, 19603, 19617, 19625, 19609, 19610, 19600, 19606, 19602, 19615, 19623, 19622, 19621, 19616, 19612, 19620, 19625, 19618, 19609, 19605, 19607, 19626, 19621, 19624, 19624, 19606, 19609, 19602, 19621, 19612, 19625, 19604, 19618, 19635, 19632, 19634, 19637, 19638, 19636, 19645, 19642, 19644, 19647, 19648, 19646, 19653, 19650, 19651, 19643, 19630, 19642, 19651, 19638, 19639, 19645, 19651, 19643, 19641, 19633, 19635, 19654, 19649, 19647, 19635, 19630, 19631, 19641, 19637, 19640, 19639, 19635, 19631, 19643, 19639, 19631, 19657, 19658, 19656, 19663, 19660, 19662, 19665, 19660, 19664, 19669, 19666, 19667, 19678, 19674, 19677, 19682, 19683, 19681, 19659, 19673, 19658, 19680, 19664, 19679, 19669, 19681, 19668, 19657, 19668, 19659, 19678, 19682, 19676, 19670, 19663, 19662, 19665, 19670, 19662, 19672, 19667, 19670, 19680, 19674, 19676, 19682, 19680, 19676, 19685, 19669, 19672, 19668, 19675, 19659, 19680, 19672, 19665, 19588, 19628, 19587, 19633, 19591, 19632, 19637, 19595, 19596, 19640, 19596, 19598, 19595, 19629, 19588, 19633, 19598, 19592, 19615, 19656, 19613, 19663, 19619, 19661, 19667, 19622, 19623, 19671, 19626, 19627, 19626, 19667, 19623, 19622, 19657, 19615, 19663, 19627, 19620, 19673, 19647, 19646, 19683, 19654, 19681, 19647, 19681, 19654,
19689, 19690, 19688, 19687, 19689, 19686, 19697, 19699, 19695, 19702, 19696, 19700, 19706, 19703, 19704, 19690, 19695, 19688, 19694, 19703, 19693, 19690, 19694, 19693, 19687, 19694, 19691, 19697, 19705, 19701, 19702, 19704, 19698, 19690, 19703, 19697, 19710, 19707, 19709, 19711, 19710, 19709, 19701, 19707, 19699, 19710, 19700, 19708, 19706, 19711, 19705, 19701, 19711, 19709, 19710, 19706, 19702, 19716, 19713, 19715, 19718, 19713, 19717, 19720, 19721, 19719, 19721, 19723, 19719, 19731, 19732, 19730, 19737, 19738, 19735, 19747, 19744, 19746, 19747, 19741, 19745, 19752, 19753, 19751, 19736, 19717, 19734, 19725, 19740, 19723, 19743, 19724, 19741, 19752, 19732, 19733, 19727, 19716, 19715, 19718, 19727, 19715, 19737, 19750, 19739, 19721, 19731, 19730, 19721, 19732, 19725, 19753, 19747, 19746, 19747, 19752, 19743, 19740, 19737, 19735, 19751, 19750, 19749, 19737, 19751, 19749, 19736, 19729, 19718, 19725, 19751, 19742, 19743, 19733, 19726, 19758, 19755, 19757, 19760, 19755, 19759, 19761, 19758, 19757, 19763, 19757, 19760, 19765, 19766, 19764, 19771, 19768, 19770, 19766, 19768, 19764, 19771, 19765, 19769, 19777, 19774, 19775, 19766, 19775, 19774, 19776, 19771, 19770, 19774, 19770, 19766, 19777, 19767, 19771, 19773, 19765, 19767, 19775, 19773, 19767,
19739, 19765, 19738, 19773, 19744, 19772, 19778, 19750, 19753, 19739, 19775, 19767, 19778, 19746, 19773, 19760, 19724, 19726, 19761, 19733, 19731, 19733, 19760, 19726, 19564, 19686, 19563, 19564, 19692, 19687, 19716, 19644, 19714, 19649, 19720, 19648, 19728, 19652, 19653, 19716, 19653, 19645, 19649, 19731, 19722, 19758, 19677, 19756, 19762, 19683, 19684, 19758, 19684, 19678, 19683, 19731, 19655, 19628, 19556, 19555, 19638, 19559, 19636, 19652, 19565, 19650, 19556, 19636, 19559, 19650, 19560, 19638, 19729, 19692, 19727, 19736, 19696, 19698, 19704, 19736, 19698, 19729, 19704, 19694, 19652, 19692, 19566, 19782, 19779, 19781, 19784, 19780, 19782, 19786, 19787, 19785, 19790, 19791, 19789, 19800, 19797, 19798, 19804, 19801, 19802, 19798, 19793, 19794, 19802, 19795, 19796, 19788, 19793, 19787, 19792, 19795, 19791, 19789, 19788, 19786, 19803, 19798, 19801, 19791, 19794, 19788, 19801, 19794, 19795, 19806, 19807, 19805, 19810, 19811, 19809, 19820, 19817, 19818, 19824, 19821, 19822, 19818, 19813, 19814, 19822, 19815, 19816, 19808, 19813, 19807, 19812, 19815, 19811, 19809, 19808, 19806, 19823, 19818, 19821, 19811, 19814, 19808, 19821, 19814, 19815, 19828, 19825, 19826, 19832, 19829, 19830, 19836, 19833, 19835, 19840, 19837, 19839, 19842, 19827, 19828,
19844, 19831, 19832, 19842, 19835, 19841, 19844, 19839, 19843, 19831, 19826, 19829, 19839, 19834, 19836, 19843, 19836, 19842, 19843, 19828, 19831, 19848, 19845, 19846, 19852, 19849, 19850, 19854, 19855, 19853, 19860, 19857, 19858, 19864, 19861, 19862, 19858, 19847, 19848, 19862, 19851, 19852, 19857, 19856, 19854, 19851, 19846, 19849, 19863, 19858, 19861, 19861, 19848, 19851, 19826, 19799, 19800, 19830, 19803, 19804, 19829, 19800, 19803, 19846, 19819, 19820, 19850, 19823, 19824, 19849, 19820, 19823, 19805, 19792, 19790, 19819, 19802, 19817, 19807, 19796, 19792, 19817, 19796, 19813, 19847, 19830, 19845, 19854, 19838, 19840, 19857, 19840, 19844, 19847, 19844, 19832, 19845, 19804, 19819, 19866, 19867, 19865, 19870, 19871, 19869, 19880, 19877, 19878, 19884, 19881, 19882, 19878, 19873, 19874, 19882, 19875, 19876, 19868, 19873, 19867, 19872, 19875, 19871, 19869, 19868, 19866, 19883, 19878, 19881, 19871, 19874, 19868, 19881, 19874, 19875, 19886, 19887, 19885, 19890, 19891, 19889, 19900, 19897, 19898, 19904, 19901, 19902, 19898, 19893, 19894, 19902, 19895, 19896, 19888, 19893, 19887, 19892, 19895, 19891, 19889, 19888, 19886, 19903, 19898, 19901, 19891, 19894, 19888, 19901, 19894, 19895, 19908, 19905, 19906, 19912, 19909, 19910, 19916, 19913, 19915,
19918, 19913, 19917, 19922, 19919, 19920, 19920, 19907, 19908, 19924, 19911, 19912, 19924, 19915, 19923, 19925, 19915, 19918, 19911, 19906, 19909, 19923, 19922, 19920, 19911, 19920, 19908, 19929, 19926, 19927, 19933, 19930, 19931, 19937, 19934, 19936, 19941, 19938, 19940, 19943, 19928, 19929, 19945, 19932, 19933, 19943, 19936, 19942, 19945, 19940, 19944, 19932, 19927, 19930, 19940, 19935, 19937, 19944, 19937, 19943, 19944, 19929, 19932, 19906, 19879, 19880, 19910, 19883, 19884, 19909, 19880, 19883, 19927, 19899, 19900, 19931, 19903, 19904, 19930, 19900, 19903, 19885, 19872, 19870, 19899, 19882, 19897, 19887, 19876, 19872, 19897, 19876, 19893, 19928, 19910, 19926, 19936, 19914, 19916, 19942, 19916, 19924, 19942, 19912, 19928, 19926, 19884, 19899, 19865, 19812, 19810, 19879, 19822, 19877, 19867, 19816, 19812, 19877, 19816, 19873, 19907, 19850, 19905, 19921, 19862, 19919, 19919, 19852, 19907, 19905, 19824, 19879, 19947, 19948, 19946, 19951, 19952, 19950, 19961, 19958, 19959, 19965, 19962, 19963, 19959, 19954, 19955, 19963, 19956, 19957, 19949, 19954, 19948, 19953, 19956, 19952, 19950, 19949, 19947, 19964, 19959, 19962, 19952, 19955, 19949, 19962, 19955, 19956, 19967, 19968, 19966, 19971, 19972, 19970, 19981, 19978, 19980, 19985, 19982, 19983,
19989, 19986, 19987, 19979, 19971, 19978, 19983, 19974, 19975, 19987, 19976, 19977, 19969, 19974, 19968, 19973, 19976, 19972, 19989, 19979, 19981, 19970, 19969, 19967, 19988, 19983, 19986, 19972, 19975, 19969, 19986, 19975, 19976, 19987, 19973, 19979, 19993, 19990, 19991, 19997, 19994, 19995, 20003, 20000, 20002, 19999, 20000, 19998, 20005, 20006, 20004, 20011, 20008, 20009, 20005, 19992, 19993, 20009, 19996, 19997, 20006, 20003, 20002, 20004, 20002, 19999, 19996, 19991, 19994, 20010, 20005, 20008, 20008, 19993, 19996, 20015, 20012, 20014, 20019, 20016, 20017, 20023, 20020, 20021, 20027, 20024, 20026, 20031, 20028, 20029, 20035, 20032, 20033, 20025, 20014, 20024, 20029, 20018, 20019, 20033, 20022, 20023, 20023, 20013, 20015, 20035, 20025, 20027, 20022, 20017, 20020, 20034, 20029, 20032, 20032, 20019, 20022, 20033, 20015, 20025, 19991, 19960, 19961, 19995, 19964, 19965, 19994, 19961, 19964, 20013, 19980, 20012, 20017, 19984, 19985, 20021, 19988, 19989, 20020, 19985, 19988, 20021, 19981, 20013, 19966, 19953, 19951, 19984, 19963, 19982, 19968, 19957, 19953, 19982, 19957, 19974, 20018, 19995, 20016, 20030, 20009, 20028, 20028, 19997, 20018, 20016, 19965, 19984, 19782, 19603, 19604, 19784, 19608, 19783, 19610, 19782, 19604, 19946, 19892, 19890,
19960, 19902, 19958, 19948, 19896, 19892, 19958, 19896, 19954, 19992, 19931, 19990, 19999, 19939, 19941, 19945, 19999, 19941, 19992, 19945, 19933, 19990, 19904, 19960, 20037, 20038, 20036, 20043, 20040, 20041, 20047, 20044, 20045, 20049, 20050, 20048, 20055, 20052, 20053, 20059, 20056, 20057, 20039, 20048, 20038, 20053, 20042, 20043, 20057, 20046, 20047, 20040, 20039, 20037, 20052, 20051, 20049, 20046, 20041, 20044, 20058, 20053, 20056, 20056, 20043, 20046, 20042, 20049, 20039, 20061, 20062, 20060, 20067, 20064, 20065, 20071, 20068, 20069, 20073, 20074, 20072, 20079, 20076, 20077, 20083, 20080, 20081, 20063, 20072, 20062, 20077, 20066, 20067, 20081, 20070, 20071, 20064, 20063, 20061, 20076, 20075, 20073, 20070, 20065, 20068, 20082, 20077, 20080, 20080, 20067, 20070, 20066, 20073, 20063, 20051, 20060, 20050, 20065, 20054, 20055, 20069, 20058, 20059, 20068, 20055, 20058, 20054, 20061, 20051, 20087, 20084, 20086, 20091, 20088, 20089, 20097, 20094, 20096, 20101, 20098, 20099, 20095, 20086, 20094, 20099, 20090, 20091, 20093, 20085, 20087, 20103, 20095, 20097, 20093, 20089, 20092, 20103, 20099, 20102, 20102, 20091, 20093, 20102, 20087, 20095, 20107, 20104, 20106, 20111, 20108, 20109, 20117, 20114, 20116, 20121, 20118, 20119, 20115, 20106, 20114,
20119, 20110, 20111, 20113, 20105, 20107, 20123, 20115, 20117, 20113, 20109, 20112, 20123, 20119, 20122, 20122, 20111, 20113, 20122, 20107, 20115, 20105, 20096, 20104, 20109, 20100, 20101, 20112, 20101, 20103, 20112, 20097, 20105, 20125, 20126, 20124, 20131, 20128, 20129, 20135, 20132, 20133, 20137, 20138, 20136, 20143, 20140, 20141, 20147, 20144, 20145, 20127, 20136, 20126, 20141, 20130, 20131, 20145, 20134, 20135, 20128, 20127, 20125, 20140, 20139, 20137, 20134, 20129, 20132, 20146, 20141, 20144, 20144, 20131, 20134, 20130, 20137, 20127, 20149, 20150, 20148, 20155, 20152, 20153, 20159, 20156, 20157, 20161, 20162, 20160, 20167, 20164, 20165, 20171, 20168, 20169, 20151, 20160, 20150, 20165, 20154, 20155, 20169, 20158, 20159, 20152, 20151, 20149, 20164, 20163, 20161, 20158, 20153, 20156, 20170, 20165, 20168, 20168, 20155, 20158, 20154, 20161, 20151, 20139, 20148, 20138, 20153, 20142, 20143, 20157, 20146, 20147, 20156, 20143, 20146, 20142, 20149, 20139, 20175, 20172, 20174, 20179, 20176, 20177, 20185, 20182, 20184, 20189, 20186, 20187, 20183, 20174, 20182, 20187, 20178, 20179, 20181, 20173, 20175, 20191, 20183, 20185, 20181, 20177, 20180, 20191, 20187, 20190, 20190, 20179, 20181, 20190, 20175, 20183, 20195, 20192, 20194, 20199, 20196, 20197,
20205, 20202, 20204, 20209, 20206, 20207, 20203, 20194, 20202, 20207, 20198, 20199, 20201, 20193, 20195, 20211, 20203, 20205, 20201, 20197, 20200, 20211, 20207, 20210, 20210, 20199, 20201, 20210, 20195, 20203, 20193, 20184, 20192, 20197, 20188, 20189, 20200, 20189, 20191, 20200, 20185, 20193, 20075, 20124, 20074, 20129, 20078, 20079, 20133, 20082, 20083, 20132, 20079, 20082, 20078, 20125, 20075, 20173, 20116, 20172, 20177, 20120, 20121, 20180, 20121, 20123, 20180, 20117, 20173, 20090, 20045, 20088, 20100, 20057, 20098, 20098, 20047, 20090, 20110, 20069, 20108, 20120, 20081, 20118, 20118, 20071, 20110, 20108, 20059, 20100, 20178, 20133, 20176, 20188, 20145, 20186, 20186, 20135, 20178, 20198, 20157, 20196, 20208, 20169, 20206, 20206, 20159, 20198, 20196, 20147, 20188, 20176, 20083, 20120, 20215, 20216, 20214, 20220, 20221, 20219, 20224, 20225, 20223, 20230, 20227, 20228, 20217, 20223, 20216, 20222, 20227, 20221, 20215, 20221, 20217, 20227, 20226, 20224, 20215, 20212, 20213, 20219, 20213, 20218, 20221, 20224, 20217, 20234, 20231, 20233, 20238, 20235, 20236, 20242, 20239, 20240, 20246, 20243, 20245, 20250, 20247, 20248, 20254, 20251, 20252, 20244, 20233, 20243, 20248, 20237, 20238, 20252, 20241, 20242, 20242, 20232, 20234, 20254, 20244, 20246,
20241, 20236, 20239, 20253, 20248, 20251, 20251, 20238, 20241, 20252, 20234, 20244, 20256, 20257, 20255, 20262, 20259, 20260, 20264, 20265, 20263, 20270, 20267, 20268, 20258, 20263, 20257, 20268, 20261, 20262, 20259, 20258, 20256, 20267, 20266, 20264, 20261, 20264, 20258, 20274, 20271, 20273, 20278, 20275, 20276, 20282, 20279, 20280, 20286, 20283, 20285, 20290, 20287, 20288, 20294, 20291, 20292, 20284, 20273, 20283, 20288, 20277, 20278, 20292, 20281, 20282, 20282, 20272, 20274, 20294, 20284, 20286, 20281, 20276, 20279, 20293, 20288, 20291, 20291, 20278, 20281, 20292, 20274, 20284, 20226, 20255, 20225, 20260, 20229, 20230, 20229, 20256, 20226, 20272, 20245, 20271, 20276, 20249, 20250, 20280, 20253, 20254, 20279, 20250, 20253, 20280, 20246, 20272, 20237, 20220, 20235, 20249, 20228, 20247, 20247, 20222, 20237, 20277, 20260, 20275, 20289, 20268, 20287, 20287, 20262, 20277, 20275, 20230, 20249, 20296, 20297, 20295, 20302, 20299, 20300, 20304, 20305, 20303, 20310, 20307, 20308, 20298, 20303, 20297, 20308, 20301, 20302, 20299, 20298, 20296, 20307, 20306, 20304, 20301, 20304, 20298, 20314, 20311, 20313, 20318, 20315, 20316, 20322, 20319, 20320, 20326, 20323, 20325, 20330, 20327, 20328, 20334, 20331, 20332, 20324, 20313, 20323, 20328, 20317, 20318,
20332, 20321, 20322, 20322, 20312, 20314, 20334, 20324, 20326, 20321, 20316, 20319, 20333, 20328, 20331, 20331, 20318, 20321, 20332, 20314, 20324, 20336, 20337, 20335, 20342, 20339, 20340, 20344, 20345, 20343, 20350, 20347, 20348, 20338, 20343, 20337, 20348, 20341, 20342, 20339, 20338, 20336, 20347, 20346, 20344, 20341, 20344, 20338, 20354, 20351, 20353, 20358, 20355, 20356, 20362, 20359, 20360, 20366, 20363, 20365, 20370, 20367, 20368, 20374, 20371, 20372, 20364, 20353, 20363, 20368, 20357, 20358, 20372, 20361, 20362, 20362, 20352, 20354, 20374, 20364, 20366, 20361, 20356, 20359, 20373, 20368, 20371, 20371, 20358, 20361, 20372, 20354, 20364, 20306, 20335, 20305, 20340, 20309, 20310, 20309, 20336, 20306, 20352, 20325, 20351, 20356, 20329, 20330, 20360, 20333, 20334, 20359, 20330, 20333, 20360, 20326, 20352, 20317, 20300, 20315, 20329, 20308, 20327, 20327, 20302, 20317, 20357, 20340, 20355, 20369, 20348, 20367, 20367, 20342, 20357, 20355, 20310, 20329, 20266, 20295, 20265, 20300, 20269, 20270, 20269, 20296, 20266, 20312, 20285, 20311, 20316, 20289, 20290, 20320, 20293, 20294, 20319, 20290, 20293, 20320, 20286, 20312, 20315, 20270, 20289, 20376, 20377, 20375, 20382, 20379, 20380, 20386, 20383, 20384, 20388, 20389, 20387, 20394, 20391, 20392,
20398, 20395, 20396, 20378, 20387, 20377, 20392, 20381, 20382, 20396, 20385, 20386, 20379, 20378, 20376, 20391, 20390, 20388, 20385, 20380, 20383, 20397, 20392, 20395, 20395, 20382, 20385, 20381, 20388, 20378, 20402, 20399, 20401, 20406, 20403, 20405, 20408, 20401, 20407, 20410, 20405, 20409, 20405, 20400, 20402, 20409, 20402, 20408, 20390, 20399, 20389, 20408, 20393, 20394, 20410, 20397, 20398, 20409, 20394, 20397, 20393, 20401, 20390, 20414, 20411, 20413, 20418, 20415, 20416, 20424, 20421, 20423, 20428, 20425, 20426, 20422, 20413, 20421, 20426, 20417, 20418, 20420, 20412, 20414, 20430, 20422, 20424, 20420, 20416, 20419, 20430, 20426, 20429, 20429, 20418, 20420, 20429, 20414, 20422, 20434, 20431, 20433, 20438, 20433, 20437, 20436, 20432, 20434, 20439, 20434, 20438, 20436, 20423, 20435, 20438, 20427, 20428, 20439, 20428, 20430, 20439, 20424, 20436, 20441, 20442, 20440, 20445, 20446, 20444, 20455, 20452, 20453, 20459, 20456, 20457, 20453, 20448, 20449, 20457, 20450, 20451, 20443, 20448, 20442, 20447, 20450, 20446, 20444, 20443, 20441, 20458, 20453, 20456, 20446, 20449, 20443, 20456, 20449, 20450, 20461, 20462, 20460, 20465, 20466, 20464, 20475, 20472, 20473, 20479, 20476, 20477, 20473, 20468, 20469, 20477, 20470, 20471, 20463, 20468, 20462,
20467, 20470, 20466, 20464, 20463, 20461, 20478, 20473, 20476, 20466, 20469, 20463, 20476, 20469, 20470, 20483, 20480, 20481, 20487, 20484, 20485, 20491, 20488, 20490, 20495, 20492, 20494, 20497, 20482, 20483, 20499, 20486, 20487, 20496, 20491, 20490, 20499, 20494, 20498, 20486, 20481, 20484, 20494, 20489, 20491, 20497, 20494, 20491, 20486, 20497, 20483, 20503, 20500, 20501, 20507, 20504, 20505, 20511, 20508, 20510, 20515, 20512, 20514, 20517, 20502, 20503, 20519, 20506, 20507, 20517, 20510, 20516, 20519, 20514, 20518, 20506, 20501, 20504, 20514, 20509, 20511, 20518, 20511, 20517, 20518, 20503, 20506, 20481, 20454, 20455, 20485, 20458, 20459, 20484, 20455, 20458, 20501, 20474, 20475, 20505, 20478, 20479, 20504, 20475, 20478, 20460, 20447, 20445, 20474, 20457, 20472, 20462, 20451, 20447, 20472, 20451, 20468, 20502, 20485, 20500, 20510, 20493, 20495, 20516, 20495, 20499, 20516, 20487, 20502, 20500, 20459, 20474, 20521, 20522, 20520, 20525, 20526, 20524, 20535, 20532, 20533, 20539, 20536, 20537, 20533, 20528, 20529, 20537, 20530, 20531, 20523, 20528, 20522, 20527, 20530, 20526, 20524, 20523, 20521, 20538, 20533, 20536, 20526, 20529, 20523, 20536, 20529, 20530, 20541, 20542, 20540, 20545, 20546, 20544, 20555, 20552, 20553, 20559, 20556, 20557,
20553, 20548, 20549, 20557, 20550, 20551, 20543, 20548, 20542, 20547, 20550, 20546, 20544, 20543, 20541, 20558, 20553, 20556, 20546, 20549, 20543, 20556, 20549, 20550, 20563, 20560, 20561, 20567, 20564, 20565, 20571, 20568, 20570, 20575, 20572, 20574, 20577, 20562, 20563, 20579, 20566, 20567, 20577, 20570, 20576, 20579, 20574, 20578, 20566, 20561, 20564, 20574, 20569, 20571, 20578, 20571, 20577, 20578, 20563, 20566, 20583, 20580, 20581, 20587, 20584, 20585, 20591, 20588, 20590, 20595, 20592, 20594, 20597, 20582, 20583, 20599, 20586, 20587, 20597, 20590, 20596, 20599, 20594, 20598, 20586, 20581, 20584, 20594, 20589, 20591, 20598, 20591, 20597, 20598, 20583, 20586, 20561, 20534, 20535, 20565, 20538, 20539, 20564, 20535, 20538, 20581, 20554, 20555, 20585, 20558, 20559, 20584, 20555, 20558, 20540, 20527, 20525, 20554, 20537, 20552, 20542, 20531, 20527, 20552, 20531, 20548, 20582, 20565, 20580, 20590, 20573, 20575, 20596, 20575, 20579, 20596, 20567, 20582, 20580, 20539, 20554, 20417, 20384, 20415, 20427, 20396, 20425, 20425, 20386, 20417, 20433, 20404, 20406, 20437, 20406, 20410, 20437, 20398, 20427, 20520, 20467, 20465, 20534, 20477, 20532, 20522, 20471, 20467, 20532, 20471, 20528, 20562, 20505, 20560, 20570, 20513, 20515, 20576, 20515, 20519,
20576, 20507, 20562, 20560, 20479, 20534, 20601, 20602, 20600, 20607, 20604, 20605, 20609, 20610, 20608, 20615, 20612, 20613, 20603, 20608, 20602, 20613, 20606, 20607, 20604, 20603, 20601, 20612, 20611, 20609, 20606, 20609, 20603, 20619, 20616, 20618, 20623, 20620, 20621, 20627, 20624, 20625, 20631, 20628, 20630, 20635, 20632, 20633, 20639, 20636, 20637, 20629, 20618, 20628, 20633, 20622, 20623, 20637, 20626, 20627, 20627, 20617, 20619, 20639, 20629, 20631, 20626, 20621, 20624, 20638, 20633, 20636, 20636, 20623, 20626, 20637, 20619, 20629, 20645, 20642, 20644, 20641, 20642, 20640, 20649, 20646, 20647, 20649, 20644, 20648, 20646, 20644, 20641, 20653, 20650, 20652, 20659, 20656, 20658, 20659, 20654, 20657, 20663, 20660, 20661, 20667, 20664, 20665, 20663, 20652, 20662, 20667, 20658, 20666, 20667, 20655, 20659, 20658, 20651, 20653, 20666, 20661, 20664, 20666, 20653, 20663, 20611, 20640, 20610, 20647, 20614, 20615, 20614, 20641, 20611, 20655, 20630, 20654, 20661, 20634, 20635, 20665, 20638, 20639, 20664, 20635, 20638, 20665, 20631, 20655, 20622, 20605, 20620, 20634, 20613, 20632, 20632, 20607, 20622, 20652, 20643, 20645, 20662, 20647, 20660, 20662, 20645, 20649, 20660, 20615, 20634, 20669, 20670, 20668, 20673, 20674, 20672, 20683, 20680, 20681,
20687, 20684, 20685, 20681, 20676, 20677, 20685, 20678, 20679, 20671, 20676, 20670, 20675, 20678, 20674, 20672, 20671, 20669, 20686, 20681, 20684, 20674, 20677, 20671, 20684, 20677, 20678, 20689, 20690, 20688, 20693, 20694, 20692, 20703, 20700, 20702, 20707, 20704, 20705, 20711, 20708, 20709, 20701, 20693, 20700, 20705, 20696, 20697, 20709, 20698, 20699, 20691, 20696, 20690, 20695, 20698, 20694, 20711, 20701, 20703, 20692, 20691, 20689, 20710, 20705, 20708, 20694, 20697, 20691, 20708, 20697, 20698, 20709, 20695, 20701, 20715, 20712, 20713, 20719, 20716, 20717, 20723, 20720, 20722, 20727, 20724, 20726, 20729, 20714, 20715, 20731, 20718, 20719, 20729, 20722, 20728, 20731, 20726, 20730, 20718, 20713, 20716, 20726, 20721, 20723, 20730, 20723, 20729, 20730, 20715, 20718, 20735, 20732, 20734, 20739, 20736, 20737, 20743, 20740, 20741, 20747, 20744, 20746, 20751, 20748, 20750, 20751, 20734, 20749, 20753, 20738, 20739, 20755, 20742, 20743, 20753, 20746, 20752, 20743, 20733, 20735, 20755, 20750, 20754, 20742, 20737, 20740, 20750, 20745, 20747, 20754, 20747, 20753, 20754, 20739, 20742, 20755, 20735, 20751, 20713, 20682, 20683, 20717, 20686, 20687, 20716, 20683, 20686, 20733, 20702, 20732, 20737, 20706, 20707, 20741, 20710, 20711, 20740, 20707, 20710,
20741, 20703, 20733, 20688, 20675, 20673, 20706, 20685, 20704, 20690, 20679, 20675, 20704, 20679, 20696, 20738, 20717, 20736, 20746, 20725, 20727, 20752, 20727, 20731, 20752, 20719, 20738, 20736, 20687, 20706, 20163, 20375, 20162, 20380, 20166, 20167, 20384, 20170, 20171, 20383, 20167, 20170, 20166, 20376, 20163, 20412, 20204, 20411, 20416, 20208, 20209, 20419, 20209, 20211, 20419, 20205, 20412, 20415, 20171, 20208, 20346, 20600, 20345, 20605, 20349, 20350, 20349, 20601, 20346, 20617, 20365, 20616, 20621, 20369, 20370, 20625, 20373, 20374, 20624, 20370, 20373, 20625, 20366, 20617, 20620, 20350, 20369, 20668, 20547, 20545, 20682, 20557, 20680, 20670, 20551, 20547, 20680, 20551, 20676, 20714, 20585, 20712, 20722, 20593, 20595, 20728, 20595, 20599, 20728, 20587, 20714, 20712, 20559, 20682, 19856, 20036, 19855, 20041, 19859, 19860, 20045, 19863, 19864, 20044, 19860, 19863, 19859, 20037, 19856, 20085, 19917, 20084, 20089, 19921, 19922, 20092, 19922, 19925, 20092, 19918, 20085, 20088, 19864, 19921, 20003, 20212, 20001, 20220, 20010, 20011, 20010, 20218, 20007, 20003, 20218, 20213, 20232, 20026, 20231, 20236, 20030, 20031, 20240, 20034, 20035, 20239, 20031, 20034, 20240, 20027, 20232, 20235, 20011, 20030, 20757, 20758, 20756, 20761, 20762, 20760,
20759, 20764, 20758, 20763, 20766, 20762, 20760, 20759, 20757, 20762, 20765, 20759, 20769, 20770, 20768, 20773, 20774, 20772, 20771, 20776, 20770, 20775, 20778, 20774, 20772, 20771, 20769, 20774, 20777, 20771, 20783, 20780, 20781, 20787, 20784, 20785, 20791, 20788, 20789, 20795, 20792, 20793, 20789, 20782, 20783, 20793, 20786, 20787, 20786, 20781, 20784, 20794, 20789, 20792, 20792, 20783, 20786, 20799, 20796, 20797, 20803, 20800, 20801, 20807, 20804, 20805, 20811, 20808, 20809, 20805, 20798, 20799, 20809, 20802, 20803, 20802, 20797, 20800, 20810, 20805, 20808, 20808, 20799, 20802, 20781, 20764, 20765, 20785, 20766, 20767, 20784, 20765, 20766, 20797, 20776, 20777, 20801, 20778, 20779, 20800, 20777, 20778, 20768, 20763, 20761, 20770, 20767, 20763, 20798, 20785, 20796, 20806, 20793, 20804, 20804, 20787, 20798, 20796, 20767, 20776, 20813, 20814, 20812, 20817, 20818, 20816, 20815, 20820, 20814, 20819, 20822, 20818, 20816, 20815, 20813, 20818, 20821, 20815, 20825, 20826, 20824, 20829, 20830, 20828, 20827, 20832, 20826, 20831, 20834, 20830, 20828, 20827, 20825, 20830, 20833, 20827, 20839, 20836, 20837, 20843, 20840, 20841, 20847, 20844, 20845, 20851, 20848, 20849, 20845, 20838, 20839, 20849, 20842, 20843, 20842, 20837, 20840, 20850, 20845, 20848,
20848, 20839, 20842, 20855, 20852, 20853, 20859, 20856, 20857, 20863, 20860, 20861, 20867, 20864, 20865, 20861, 20854, 20855, 20865, 20858, 20859, 20858, 20853, 20856, 20866, 20861, 20864, 20864, 20855, 20858, 20837, 20820, 20821, 20841, 20822, 20823, 20840, 20821, 20822, 20853, 20832, 20833, 20857, 20834, 20835, 20856, 20833, 20834, 20824, 20819, 20817, 20826, 20823, 20819, 20854, 20841, 20852, 20862, 20849, 20860, 20860, 20843, 20854, 20852, 20823, 20832, 20812, 20775, 20773, 20814, 20779, 20775, 20838, 20801, 20836, 20846, 20809, 20844, 20844, 20803, 20838, 20836, 20779, 20820, 20869, 20870, 20868, 20873, 20874, 20872, 20871, 20876, 20870, 20875, 20878, 20874, 20872, 20871, 20869, 20874, 20877, 20871, 20881, 20882, 20880, 20885, 20886, 20884, 20883, 20888, 20882, 20887, 20890, 20886, 20884, 20883, 20881, 20886, 20889, 20883, 20895, 20892, 20893, 20899, 20896, 20897, 20903, 20900, 20901, 20907, 20904, 20905, 20901, 20894, 20895, 20905, 20898, 20899, 20898, 20893, 20896, 20906, 20901, 20904, 20904, 20895, 20898, 20911, 20908, 20910, 20915, 20912, 20913, 20919, 20916, 20917, 20923, 20920, 20922, 20927, 20924, 20925, 20931, 20928, 20929, 20921, 20910, 20920, 20925, 20914, 20915, 20929, 20918, 20919, 20919, 20909, 20911, 20931, 20921, 20923,
20918, 20913, 20916, 20930, 20925, 20928, 20928, 20915, 20918, 20929, 20911, 20921, 20893, 20876, 20877, 20897, 20878, 20879, 20896, 20877, 20878, 20909, 20885, 20908, 20913, 20888, 20889, 20917, 20890, 20891, 20916, 20889, 20890, 20917, 20887, 20909, 20880, 20875, 20873, 20882, 20879, 20875, 20914, 20897, 20912, 20926, 20905, 20924, 20924, 20899, 20914, 20912, 20879, 20888, 20868, 20831, 20829, 20870, 20835, 20831, 20894, 20857, 20892, 20902, 20865, 20900, 20900, 20859, 20894, 20892, 20835, 20876, 20935, 20932, 20934, 20939, 20936, 20938, 20941, 20934, 20940, 20943, 20938, 20942, 20938, 20933, 20935, 20942, 20935, 20941, 20947, 20944, 20946, 20951, 20948, 20950, 20953, 20946, 20952, 20955, 20950, 20954, 20950, 20945, 20947, 20954, 20947, 20953, 20946, 20937, 20939, 20952, 20939, 20943, 20959, 20956, 20958, 20963, 20960, 20962, 20965, 20958, 20964, 20967, 20962, 20966, 20962, 20957, 20959, 20966, 20959, 20965, 20971, 20968, 20970, 20975, 20972, 20974, 20977, 20970, 20976, 20979, 20974, 20978, 20974, 20969, 20971, 20978, 20971, 20977, 20970, 20961, 20963, 20976, 20963, 20967, 20981, 20982, 20980, 20985, 20986, 20984, 20991, 20988, 20989, 20995, 20992, 20993, 20999, 20996, 20997, 21003, 21000, 21001, 20997, 20990, 20991, 21001, 20994, 20995,
20983, 20988, 20982, 20987, 20992, 20986, 20984, 20983, 20981, 20994, 20989, 20992, 21002, 20997, 21000, 20986, 20989, 20983, 21000, 20991, 20994, 21005, 21006, 21004, 21009, 21010, 21008, 21015, 21012, 21013, 21019, 21016, 21017, 21023, 21020, 21021, 21027, 21024, 21025, 21021, 21014, 21015, 21025, 21018, 21019, 21007, 21012, 21006, 21011, 21016, 21010, 21008, 21007, 21005, 21018, 21013, 21016, 21026, 21021, 21024, 21010, 21013, 21007, 21024, 21015, 21018, 21033, 21030, 21031, 21037, 21034, 21035, 21039, 21040, 21038, 21045, 21042, 21043, 21049, 21046, 21047, 21029, 21038, 21028, 21033, 21042, 21032, 21047, 21036, 21037, 21042, 21041, 21039, 21036, 21031, 21034, 21048, 21043, 21046, 21046, 21033, 21036, 21032, 21039, 21029, 21055, 21052, 21054, 21057, 21058, 21056, 21063, 21060, 21061, 21067, 21064, 21066, 21071, 21068, 21069, 21065, 21050, 21064, 21069, 21058, 21059, 21071, 21065, 21067, 21063, 21054, 21062, 21054, 21050, 21051, 21060, 21059, 21057, 21062, 21051, 21059, 21069, 21051, 21065, 21031, 20998, 20999, 21035, 21002, 21003, 21034, 20999, 21002, 21057, 21022, 21023, 21061, 21026, 21027, 21060, 21023, 21026, 21004, 20987, 20985, 21014, 20993, 21012, 21022, 21001, 21020, 21006, 20993, 20987, 21020, 20995, 21014, 21058, 21035, 21056,
21070, 21047, 21068, 21068, 21037, 21058, 21056, 21003, 21022, 21073, 21074, 21072, 21077, 21078, 21076, 21083, 21080, 21081, 21087, 21084, 21085, 21091, 21088, 21089, 21095, 21092, 21093, 21089, 21082, 21083, 21093, 21086, 21087, 21075, 21080, 21074, 21079, 21084, 21078, 21076, 21075, 21073, 21086, 21081, 21084, 21094, 21089, 21092, 21078, 21081, 21075, 21092, 21083, 21086, 21097, 21098, 21096, 21101, 21102, 21100, 21107, 21104, 21105, 21111, 21108, 21109, 21115, 21112, 21113, 21119, 21116, 21117, 21113, 21106, 21107, 21117, 21110, 21111, 21099, 21104, 21098, 21103, 21108, 21102, 21100, 21099, 21097, 21110, 21105, 21108, 21118, 21113, 21116, 21102, 21105, 21099, 21116, 21107, 21110, 21123, 21120, 21122, 21127, 21124, 21126, 21131, 21128, 21129, 21135, 21132, 21133, 21131, 21122, 21130, 21135, 21126, 21134, 21126, 21121, 21123, 21134, 21129, 21132, 21134, 21123, 21131, 21139, 21136, 21138, 21143, 21140, 21142, 21147, 21144, 21145, 21151, 21148, 21149, 21147, 21138, 21146, 21151, 21142, 21150, 21142, 21137, 21139, 21150, 21145, 21148, 21150, 21139, 21147, 21129, 21090, 21091, 21133, 21094, 21095, 21132, 21091, 21094, 21145, 21114, 21115, 21149, 21118, 21119, 21148, 21115, 21118, 21096, 21079, 21077, 21106, 21085, 21104, 21114, 21093, 21112,
21098, 21085, 21079, 21112, 21087, 21106, 21138, 21125, 21127, 21146, 21133, 21144, 21146, 21127, 21135, 21144, 21095, 21114, 20958, 20949, 20951, 20964, 20951, 20955, 21072, 21011, 21009, 21082, 21017, 21080, 21090, 21025, 21088, 21074, 21017, 21011, 21088, 21019, 21082, 21122, 21053, 21055, 21130, 21061, 21128, 21130, 21055, 21063, 21128, 21027, 21090, 21155, 21152, 21154, 21159, 21156, 21158, 21161, 21154, 21160, 21163, 21158, 21162, 21158, 21153, 21155, 21162, 21155, 21161, 21167, 21164, 21166, 21171, 21168, 21170, 21173, 21166, 21172, 21175, 21170, 21174, 21170, 21165, 21167, 21174, 21167, 21173, 21166, 21157, 21159, 21172, 21159, 21163, 21177, 21178, 21176, 21181, 21182, 21180, 21187, 21184, 21185, 21191, 21188, 21189, 21195, 21192, 21193, 21199, 21196, 21197, 21193, 21186, 21187, 21197, 21190, 21191, 21179, 21184, 21178, 21183, 21188, 21182, 21180, 21179, 21177, 21190, 21185, 21188, 21198, 21193, 21196, 21182, 21185, 21179, 21196, 21187, 21190, 21201, 21202, 21200, 21205, 21206, 21204, 21209, 21205, 21208, 21213, 21210, 21211, 21217, 21214, 21215, 21221, 21218, 21220, 21225, 21222, 21223, 21229, 21226, 21227, 21219, 21208, 21218, 21223, 21212, 21213, 21227, 21216, 21217, 21203, 21210, 21202, 21207, 21214, 21206, 21217, 21207, 21209,
21229, 21219, 21221, 21204, 21203, 21201, 21216, 21211, 21214, 21228, 21223, 21226, 21206, 21211, 21203, 21226, 21213, 21216, 21227, 21209, 21219, 21233, 21230, 21232, 21239, 21236, 21237, 21243, 21240, 21241, 21245, 21246, 21244, 21251, 21248, 21249, 21235, 21244, 21234, 21243, 21248, 21242, 21239, 21232, 21238, 21248, 21247, 21245, 21235, 21231, 21233, 21242, 21237, 21240, 21242, 21233, 21239, 21242, 21245, 21235, 21255, 21252, 21254, 21259, 21256, 21257, 21263, 21260, 21261, 21267, 21264, 21266, 21271, 21268, 21269, 21275, 21272, 21273, 21265, 21254, 21264, 21269, 21258, 21259, 21273, 21262, 21263, 21263, 21253, 21255, 21275, 21265, 21267, 21262, 21257, 21260, 21274, 21269, 21272, 21272, 21259, 21262, 21273, 21255, 21265, 21237, 21194, 21195, 21241, 21198, 21199, 21240, 21195, 21198, 21253, 21220, 21252, 21257, 21224, 21225, 21261, 21228, 21229, 21260, 21225, 21228, 21261, 21221, 21253, 21200, 21183, 21181, 21212, 21189, 21210, 21224, 21197, 21222, 21202, 21189, 21183, 21222, 21191, 21212, 21258, 21241, 21256, 21270, 21249, 21268, 21268, 21243, 21258, 21256, 21199, 21224, 21277, 21278, 21276, 21283, 21280, 21281, 21287, 21284, 21285, 21289, 21290, 21288, 21295, 21292, 21293, 21299, 21296, 21297, 21279, 21288, 21278, 21293, 21282, 21283,
21297, 21286, 21287, 21280, 21279, 21277, 21292, 21291, 21289, 21286, 21281, 21284, 21298, 21293, 21296, 21296, 21283, 21286, 21282, 21289, 21279, 21303, 21300, 21302, 21307, 21304, 21305, 21311, 21308, 21310, 21315, 21312, 21313, 21309, 21302, 21308, 21313, 21306, 21307, 21307, 21301, 21303, 21315, 21309, 21311, 21313, 21303, 21309, 21317, 21318, 21316, 21323, 21320, 21321, 21327, 21324, 21325, 21329, 21330, 21328, 21335, 21332, 21333, 21339, 21336, 21337, 21319, 21328, 21318, 21333, 21322, 21323, 21337, 21326, 21327, 21320, 21319, 21317, 21332, 21331, 21329, 21326, 21321, 21324, 21338, 21333, 21336, 21336, 21323, 21326, 21322, 21329, 21319, 21343, 21340, 21342, 21347, 21344, 21345, 21351, 21348, 21350, 21355, 21352, 21353, 21349, 21342, 21348, 21353, 21346, 21347, 21347, 21341, 21343, 21355, 21349, 21351, 21353, 21343, 21349, 21291, 21316, 21290, 21321, 21294, 21295, 21325, 21298, 21299, 21324, 21295, 21298, 21294, 21317, 21291, 21341, 21310, 21340, 21345, 21314, 21315, 21345, 21311, 21341, 21306, 21285, 21304, 21314, 21297, 21312, 21312, 21287, 21306, 21346, 21325, 21344, 21354, 21337, 21352, 21352, 21327, 21346, 21344, 21299, 21314, 21357, 21358, 21356, 21363, 21360, 21361, 21367, 21364, 21365, 21369, 21370, 21368, 21375, 21372, 21373,
21379, 21376, 21377, 21359, 21368, 21358, 21373, 21362, 21363, 21377, 21366, 21367, 21360, 21359, 21357, 21372, 21371, 21369, 21366, 21361, 21364, 21378, 21373, 21376, 21376, 21363, 21366, 21362, 21369, 21359, 21383, 21380, 21382, 21387, 21384, 21385, 21391, 21388, 21390, 21395, 21392, 21393, 21389, 21382, 21388, 21393, 21386, 21387, 21387, 21381, 21383, 21395, 21389, 21391, 21393, 21383, 21389, 21397, 21398, 21396, 21403, 21400, 21401, 21407, 21404, 21405, 21409, 21410, 21408, 21415, 21412, 21413, 21419, 21416, 21417, 21399, 21408, 21398, 21413, 21402, 21403, 21417, 21406, 21407, 21400, 21399, 21397, 21412, 21411, 21409, 21406, 21401, 21404, 21418, 21413, 21416, 21416, 21403, 21406, 21402, 21409, 21399, 21423, 21420, 21422, 21427, 21424, 21425, 21431, 21428, 21430, 21435, 21432, 21433, 21429, 21422, 21428, 21427, 21432, 21426, 21427, 21421, 21423, 21435, 21429, 21431, 21433, 21423, 21429, 21371, 21396, 21370, 21401, 21374, 21375, 21405, 21378, 21379, 21404, 21375, 21378, 21374, 21397, 21371, 21421, 21390, 21420, 21425, 21394, 21395, 21425, 21391, 21421, 21386, 21365, 21384, 21394, 21377, 21392, 21392, 21367, 21386, 21426, 21405, 21424, 21434, 21417, 21432, 21432, 21407, 21426, 21424, 21379, 21394, 21331, 21356, 21330, 21361, 21334, 21335,
21365, 21338, 21339, 21364, 21335, 21338, 21334, 21357, 21331, 21381, 21350, 21380, 21385, 21354, 21355, 21385, 21351, 21381, 21384, 21339, 21354, 21437, 21438, 21436, 21443, 21440, 21441, 21445, 21446, 21444, 21451, 21448, 21449, 21439, 21444, 21438, 21449, 21442, 21443, 21440, 21439, 21437, 21448, 21447, 21445, 21442, 21445, 21439, 21455, 21452, 21454, 21459, 21456, 21457, 21463, 21460, 21461, 21467, 21464, 21466, 21471, 21468, 21469, 21475, 21472, 21473, 21465, 21454, 21464, 21469, 21458, 21459, 21473, 21462, 21463, 21463, 21453, 21455, 21475, 21465, 21467, 21462, 21457, 21460, 21474, 21469, 21472, 21472, 21459, 21462, 21473, 21455, 21465, 21477, 21478, 21476, 21483, 21480, 21481, 21485, 21486, 21484, 21491, 21488, 21489, 21479, 21484, 21478, 21489, 21482, 21483, 21480, 21479, 21477, 21488, 21487, 21485, 21482, 21485, 21479, 21495, 21492, 21494, 21499, 21496, 21497, 21503, 21500, 21501, 21507, 21504, 21506, 21511, 21508, 21509, 21515, 21512, 21513, 21505, 21494, 21504, 21509, 21498, 21499, 21513, 21502, 21503, 21503, 21493, 21495, 21515, 21505, 21507, 21502, 21497, 21500, 21514, 21509, 21512, 21512, 21499, 21502, 21513, 21495, 21505, 21447, 21476, 21446, 21481, 21450, 21451, 21450, 21477, 21447, 21493, 21466, 21492, 21497, 21470, 21471,
21501, 21474, 21475, 21500, 21471, 21474, 21501, 21467, 21493, 21458, 21441, 21456, 21470, 21449, 21468, 21468, 21443, 21458, 21498, 21481, 21496, 21510, 21489, 21508, 21508, 21483, 21498, 21496, 21451, 21470, 21517, 21518, 21516, 21523, 21520, 21521, 21525, 21526, 21524, 21531, 21528, 21529, 21519, 21524, 21518, 21529, 21522, 21523, 21520, 21519, 21517, 21528, 21527, 21525, 21522, 21525, 21519, 21535, 21532, 21534, 21539, 21536, 21537, 21543, 21540, 21541, 21547, 21544, 21546, 21551, 21548, 21549, 21555, 21552, 21553, 21545, 21534, 21544, 21549, 21538, 21539, 21553, 21542, 21543, 21543, 21533, 21535, 21555, 21545, 21547, 21542, 21537, 21540, 21554, 21549, 21552, 21552, 21539, 21542, 21553, 21535, 21545, 21557, 21558, 21556, 21563, 21560, 21561, 21565, 21566, 21564, 21569, 21570, 21568, 21559, 21564, 21558, 21563, 21568, 21562, 21560, 21559, 21557, 21565, 21570, 21567, 21559, 21568, 21565, 21575, 21572, 21574, 21579, 21576, 21577, 21583, 21580, 21581, 21587, 21584, 21586, 21591, 21588, 21589, 21595, 21592, 21593, 21585, 21574, 21584, 21589, 21578, 21579, 21593, 21582, 21583, 21583, 21573, 21575, 21595, 21585, 21587, 21582, 21577, 21580, 21594, 21589, 21592, 21592, 21579, 21582, 21593, 21575, 21585, 21527, 21556, 21526, 21561, 21530, 21531,
21530, 21557, 21527, 21573, 21546, 21572, 21577, 21550, 21551, 21581, 21554, 21555, 21580, 21551, 21554, 21581, 21547, 21573, 21538, 21521, 21536, 21550, 21529, 21548, 21548, 21523, 21538, 21578, 21561, 21576, 21590, 21569, 21588, 21588, 21563, 21578, 21576, 21531, 21550, 21487, 21516, 21486, 21521, 21490, 21491, 21490, 21517, 21487, 21533, 21506, 21532, 21537, 21510, 21511, 21541, 21514, 21515, 21540, 21511, 21514, 21541, 21507, 21533, 21536, 21491, 21510, 21041, 21276, 21040, 21281, 21044, 21045, 21285, 21048, 21049, 21284, 21045, 21048, 21044, 21277, 21041, 21301, 21066, 21300, 21305, 21070, 21071, 21305, 21067, 21301, 21304, 21049, 21070, 21247, 21436, 21246, 21441, 21250, 21251, 21250, 21437, 21247, 21453, 21266, 21452, 21457, 21270, 21271, 21461, 21274, 21275, 21460, 21271, 21274, 21461, 21267, 21453, 21456, 21251, 21270, 21154, 20973, 20975, 21160, 20975, 20979, 21176, 21103, 21101, 21186, 21109, 21184, 21194, 21117, 21192, 21178, 21109, 21103, 21192, 21111, 21186, 21232, 21141, 21143, 21238, 21149, 21236, 21238, 21143, 21151, 21236, 21119, 21194, 20941, 20790, 20791, 20943, 20794, 20795, 20942, 20791, 20794, 20953, 20806, 20807, 20955, 20810, 20811, 20954, 20807, 20810, 20952, 20795, 20806, 20965, 20846, 20847, 20967, 20850, 20851,
20966, 20847, 20850, 20977, 20862, 20863, 20979, 20866, 20867, 20978, 20863, 20866, 20976, 20851, 20862, 20964, 20811, 20846, 21161, 20902, 20903, 21163, 20906, 20907, 21162, 20903, 20906, 21171, 20922, 21169, 21173, 20926, 20927, 21175, 20930, 20931, 21174, 20927, 20930, 21175, 20923, 21171, 21172, 20907, 20926, 21160, 20867, 20902, 19194, 19567, 19193, 19572, 19197, 19571, 19202, 19577, 19201, 19582, 19205, 19206, 19581, 19202, 19205, 19194, 19577, 19568, 19572, 19206, 19198, 19216, 19599, 19215, 19225, 19607, 19224, 19607, 19221, 19224, 19221, 19600, 19216, 19780, 19409, 19779, 19780, 19413, 19410, 19783, 19225, 19413, 21600, 21599, 21598, 21603, 21604, 21602, 21607, 21608, 21606, 21618, 21615, 21617, 21611, 21601, 21600, 21605, 21613, 21604, 21602, 21597, 21596, 21599, 21606, 21598, 21616, 21611, 21615, 21597, 21598, 21596, 21606, 21603, 21602, 21617, 21613, 21614, 21598, 21602, 21596, 21615, 21610, 21613, 21610, 21600, 21597, 21604, 21610, 21597, 21621, 21620, 21619, 21625, 21624, 21623, 21628, 21629, 21627, 21632, 21633, 21631, 21644, 21641, 21643, 21648, 21645, 21647, 21650, 21651, 21649, 21635, 21622, 21621, 21637, 21626, 21625, 21620, 21627, 21619, 21642, 21635, 21641, 21624, 21631, 21623, 21646, 21637, 21645, 21640, 21649, 21639,
21622, 21623, 21620, 21631, 21630, 21628, 21647, 21642, 21644, 21623, 21628, 21620, 21645, 21636, 21642, 21636, 21625, 21622, 21601, 21619, 21599, 21627, 21609, 21607, 21643, 21616, 21618, 21619, 21607, 21599, 21641, 21612, 21616, 21612, 21621, 21601, 21655, 21654, 21653, 21659, 21658, 21657, 21662, 21663, 21661, 21666, 21667, 21665, 21678, 21675, 21677, 21684, 21681, 21683, 21684, 21685, 21682, 21669, 21656, 21655, 21672, 21660, 21659, 21654, 21661, 21653, 21676, 21669, 21675, 21658, 21665, 21657, 21682, 21672, 21681, 21682, 21674, 21673, 21656, 21657, 21654, 21665, 21664, 21662, 21683, 21676, 21678, 21657, 21662, 21654, 21681, 21670, 21676, 21670, 21659, 21656, 21689, 21688, 21687, 21692, 21693, 21691, 21696, 21692, 21695, 21699, 21698, 21697, 21704, 21701, 21703, 21697, 21690, 21689, 21702, 21695, 21701, 21688, 21691, 21687, 21695, 21688, 21690, 21703, 21698, 21700, 21701, 21690, 21698, 21660, 21687, 21658, 21691, 21668, 21666, 21674, 21697, 21673, 21687, 21666, 21658, 21673, 21689, 21660, 21707, 21706, 21705, 21710, 21711, 21709, 21715, 21714, 21713, 21718, 21719, 21717, 21713, 21708, 21707, 21712, 21717, 21711, 21709, 21707, 21705, 21717, 21715, 21713, 21711, 21713, 21707, 21723, 21722, 21721, 21730, 21727, 21729, 21726, 21727, 21725,
21728, 21723, 21727, 21725, 21723, 21721, 21721, 21716, 21715, 21720, 21725, 21719, 21719, 21721, 21715, 21738, 21735, 21737, 21740, 21741, 21739, 21750, 21747, 21749, 21752, 21753, 21751, 21748, 21737, 21747, 21742, 21751, 21741, 21737, 21731, 21732, 21749, 21743, 21744, 21739, 21734, 21733, 21751, 21746, 21745, 21747, 21732, 21743, 21741, 21745, 21734, 21757, 21756, 21755, 21762, 21759, 21761, 21765, 21764, 21763, 21770, 21767, 21769, 21763, 21758, 21757, 21768, 21761, 21767, 21761, 21756, 21758, 21769, 21764, 21766, 21767, 21758, 21764, 21778, 21775, 21777, 21780, 21781, 21779, 21777, 21771, 21772, 21779, 21774, 21773, 21785, 21784, 21783, 21792, 21789, 21791, 21792, 21787, 21790, 21790, 21785, 21789, 21790, 21784, 21786, 21776, 21749, 21775, 21754, 21779, 21753, 21775, 21744, 21771, 21753, 21773, 21746, 21783, 21766, 21765, 21788, 21769, 21787, 21787, 21766, 21784, 21734, 21755, 21733, 21746, 21763, 21745, 21745, 21757, 21734, 21774, 21783, 21773, 21791, 21781, 21782, 21789, 21774, 21781, 21773, 21765, 21746, 21705, 21640, 21639, 21651, 21709, 21649, 21649, 21705, 21639, 21736, 21679, 21735, 21686, 21739, 21685, 21735, 21671, 21731, 21685, 21733, 21674, 21755, 21700, 21699, 21760, 21703, 21759, 21759, 21700, 21756, 21733, 21699, 21674,
21626, 21653, 21624, 21661, 21634, 21632, 21677, 21646, 21648, 21679, 21652, 21650, 21653, 21632, 21624, 21675, 21638, 21646, 21671, 21650, 21640, 21638, 21655, 21626, 21708, 21731, 21706, 21716, 21743, 21714, 21714, 21732, 21708, 21724, 21771, 21722, 21778, 21728, 21730, 21777, 21724, 21728, 21722, 21744, 21716, 21706, 21671, 21640, 21798, 21801, 21799, 21803, 21804, 21802, 21807, 21808, 21806, 21808, 21810, 21806, 21797, 21802, 21796, 21798, 21807, 21806, 21806, 21800, 21798, 21810, 21805, 21803, 21797, 21810, 21803, 21802, 21795, 21794, 21796, 21794, 21793, 21814, 21813, 21812, 21819, 21820, 21818, 21823, 21819, 21822, 21825, 21826, 21824, 21813, 21818, 21812, 21815, 21819, 21813, 21816, 21825, 21824, 21824, 21823, 21822, 21815, 21824, 21822, 21828, 21831, 21829, 21833, 21834, 21832, 21834, 21836, 21832, 21841, 21840, 21839, 21846, 21847, 21845, 21830, 21840, 21831, 21837, 21845, 21836, 21828, 21833, 21832, 21832, 21830, 21828, 21839, 21847, 21841, 21847, 21844, 21843, 21841, 21843, 21838, 21836, 21839, 21830, 21851, 21854, 21852, 21862, 21859, 21861, 21862, 21857, 21860, 21866, 21863, 21865, 21872, 21869, 21871, 21870, 21859, 21869, 21859, 21854, 21853, 21860, 21852, 21854, 21871, 21867, 21868, 21864, 21855, 21863, 21865, 21857, 21858,
21857, 21855, 21852, 21869, 21853, 21867, 21875, 21878, 21876, 21880, 21881, 21879, 21886, 21883, 21885, 21891, 21890, 21889, 21894, 21895, 21893, 21900, 21897, 21899, 21889, 21878, 21877, 21882, 21893, 21881, 21898, 21885, 21897, 21879, 21874, 21873, 21887, 21895, 21888, 21885, 21876, 21878, 21892, 21897, 21890, 21873, 21877, 21875, 21888, 21889, 21887, 21874, 21889, 21877, 21881, 21887, 21874, 21890, 21885, 21878, 21875, 21842, 21841, 21844, 21879, 21843, 21884, 21848, 21883, 21838, 21875, 21841, 21838, 21879, 21873, 21876, 21848, 21842, 21831, 21851, 21829, 21840, 21868, 21867, 21872, 21848, 21850, 21871, 21842, 21848, 21831, 21867, 21853, 21904, 21907, 21905, 21911, 21908, 21910, 21911, 21912, 21909, 21924, 21921, 21923, 21924, 21925, 21922, 21918, 21907, 21906, 21917, 21921, 21916, 21909, 21901, 21908, 21902, 21912, 21903, 21921, 21919, 21918, 21919, 21925, 21920, 21903, 21904, 21902, 21917, 21912, 21913, 21916, 21903, 21912, 21916, 21918, 21906, 21932, 21929, 21931, 21935, 21934, 21933, 21938, 21939, 21937, 21942, 21938, 21941, 21939, 21931, 21937, 21928, 21929, 21927, 21933, 21938, 21937, 21941, 21934, 21936, 21933, 21931, 21928, 21946, 21947, 21945, 21952, 21953, 21951, 21948, 21951, 21947, 21943, 21947, 21944, 21951, 21950, 21949,
21947, 21949, 21944, 21955, 21958, 21956, 21962, 21959, 21961, 21963, 21966, 21964, 21970, 21967, 21969, 21963, 21958, 21957, 21968, 21961, 21967, 21958, 21959, 21956, 21969, 21964, 21966, 21967, 21958, 21964, 21926, 21945, 21925, 21920, 21945, 21943, 21955, 21936, 21935, 21960, 21941, 21959, 21956, 21941, 21936, 21907, 21927, 21905, 21929, 21915, 21914, 21919, 21935, 21933, 21905, 21929, 21914, 21907, 21933, 21928, 21944, 21955, 21943, 21949, 21965, 21963, 21944, 21963, 21957, 21943, 21935, 21920, 21795, 21833, 21794, 21794, 21829, 21793, 21851, 21797, 21796, 21855, 21801, 21800, 21797, 21855, 21800, 21793, 21851, 21796, 21814, 21902, 21815, 21904, 21817, 21816, 21915, 21825, 21914, 21815, 21904, 21816, 21905, 21825, 21817, 21801, 21812, 21799, 21818, 21809, 21807, 21799, 21818, 21807, 21910, 21864, 21866, 21864, 21901, 21856, 21856, 21814, 21801, 21973, 21972, 21971, 21976, 21977, 21975, 21977, 21979, 21975, 21983, 21982, 21981, 21986, 21987, 21985, 21981, 21974, 21973, 21980, 21985, 21979, 21972, 21975, 21971, 21975, 21973, 21971, 21985, 21983, 21981, 21979, 21981, 21973, 21991, 21990, 21989, 21998, 21995, 21997, 21994, 21995, 21993, 21996, 21991, 21995, 21993, 21991, 21989, 21989, 21984, 21983, 21988, 21993, 21987, 21987, 21989, 21983,
22001, 22000, 21999, 22005, 22004, 22003, 22008, 22009, 22007, 22012, 22013, 22011, 22017, 22016, 22015, 22021, 22020, 22019, 22015, 22002, 22001, 22019, 22006, 22005, 22000, 22007, 21999, 22004, 22011, 22003, 22002, 22003, 22000, 22011, 22010, 22008, 22018, 22019, 22016, 22003, 22008, 22000, 22016, 22005, 22002, 22025, 22024, 22023, 22029, 22028, 22027, 22032, 22033, 22031, 22036, 22037, 22035, 22041, 22040, 22039, 22045, 22044, 22043, 22039, 22026, 22025, 22043, 22030, 22029, 22024, 22031, 22023, 22028, 22035, 22027, 22026, 22027, 22024, 22035, 22034, 22032, 22042, 22043, 22040, 22027, 22032, 22024, 22040, 22029, 22026, 22049, 22048, 22047, 22053, 22052, 22051, 22058, 22055, 22057, 22062, 22059, 22061, 22056, 22049, 22055, 22060, 22053, 22059, 22050, 22051, 22048, 22061, 22056, 22058, 22059, 22050, 22056, 22065, 22064, 22063, 22069, 22068, 22067, 22074, 22071, 22073, 22078, 22075, 22077, 22072, 22065, 22071, 22076, 22069, 22075, 22066, 22067, 22064, 22077, 22072, 22074, 22075, 22066, 22072, 22047, 22018, 22017, 22051, 22022, 22021, 22048, 22021, 22018, 22063, 22042, 22041, 22067, 22046, 22045, 22064, 22045, 22042, 22006, 22023, 22004, 22031, 22014, 22012, 22022, 22039, 22020, 22023, 22012, 22004, 22020, 22025, 22006, 22054, 22063, 22052,
22073, 22060, 22062, 22071, 22054, 22060, 22052, 22041, 22022, 21974, 21999, 21972, 22007, 21978, 21976, 21984, 22015, 21982, 21999, 21976, 21972, 21982, 22001, 21974, 21992, 22047, 21990, 22057, 21996, 21998, 22055, 21992, 21996, 21990, 22017, 21984, 22080, 22083, 22081, 22088, 22090, 22086, 22098, 22095, 22097, 22098, 22099, 22096, 22092, 22083, 22082, 22091, 22095, 22090, 22086, 22082, 22080, 22095, 22093, 22092, 22096, 22094, 22093, 22088, 22084, 22085, 22084, 22080, 22079, 22082, 22095, 22092, 22106, 22103, 22105, 22107, 22110, 22108, 22114, 22115, 22113, 22120, 22114, 22118, 22122, 22123, 22121, 22115, 22105, 22113, 22102, 22103, 22101, 22108, 22113, 22107, 22110, 22114, 22108, 22111, 22122, 22121, 22121, 22120, 22118, 22111, 22118, 22110, 22107, 22105, 22102, 22130, 22129, 22128, 22135, 22132, 22134, 22135, 22136, 22133, 22141, 22138, 22140, 22133, 22125, 22132, 22126, 22136, 22127, 22138, 22131, 22130, 22127, 22128, 22126, 22140, 22136, 22137, 22136, 22130, 22127, 22109, 22126, 22110, 22128, 22112, 22111, 22119, 22132, 22117, 22110, 22128, 22111, 22109, 22132, 22125, 22083, 22101, 22081, 22103, 22089, 22087, 22093, 22109, 22107, 22119, 22099, 22100, 22081, 22103, 22087, 22117, 22094, 22099, 22083, 22107, 22102, 22148, 22149, 22147,
22146, 22148, 22145, 22156, 22155, 22154, 22161, 22162, 22160, 22159, 22161, 22158, 22143, 22155, 22144, 22149, 22164, 22147, 22152, 22161, 22160, 22158, 22153, 22151, 22144, 22147, 22143, 22145, 22144, 22142, 22151, 22156, 22154, 22164, 22159, 22158, 22151, 22164, 22158, 22143, 22164, 22154, 22168, 22171, 22169, 22177, 22174, 22176, 22177, 22172, 22175, 22183, 22180, 22182, 22181, 22174, 22180, 22175, 22170, 22174, 22171, 22172, 22169, 22182, 22178, 22179, 22178, 22174, 22170, 22184, 22187, 22185, 22195, 22192, 22194, 22195, 22190, 22193, 22199, 22196, 22198, 22192, 22187, 22186, 22193, 22185, 22187, 22197, 22188, 22196, 22198, 22190, 22191, 22190, 22188, 22185, 22184, 22153, 22152, 22156, 22189, 22157, 22199, 22166, 22197, 22185, 22156, 22153, 22197, 22157, 22189, 22144, 22168, 22142, 22155, 22179, 22178, 22183, 22166, 22167, 22182, 22157, 22166, 22155, 22170, 22144, 22206, 22207, 22205, 22213, 22212, 22211, 22216, 22217, 22215, 22222, 22219, 22221, 22211, 22202, 22201, 22204, 22215, 22203, 22215, 22210, 22209, 22202, 22205, 22201, 22220, 22213, 22219, 22205, 22204, 22203, 22210, 22211, 22209, 22221, 22217, 22218, 22201, 22203, 22200, 22219, 22210, 22217, 22209, 22201, 22200, 22203, 22209, 22200, 22228, 22229, 22227, 22232, 22233, 22231,
22237, 22236, 22235, 22241, 22240, 22239, 22246, 22243, 22245, 22250, 22247, 22249, 22235, 22224, 22223, 22239, 22226, 22225, 22224, 22227, 22223, 22244, 22237, 22243, 22226, 22231, 22225, 22248, 22241, 22247, 22231, 22230, 22228, 22238, 22239, 22236, 22249, 22244, 22246, 22225, 22228, 22224, 22247, 22238, 22244, 22236, 22225, 22224, 22227, 22208, 22206, 22214, 22235, 22212, 22245, 22220, 22222, 22223, 22206, 22202, 22243, 22214, 22220, 22212, 22223, 22202, 22256, 22257, 22255, 22260, 22261, 22259, 22265, 22264, 22263, 22269, 22268, 22267, 22274, 22271, 22273, 22278, 22275, 22277, 22263, 22252, 22251, 22267, 22254, 22253, 22252, 22255, 22251, 22272, 22265, 22271, 22254, 22259, 22253, 22276, 22269, 22275, 22259, 22258, 22256, 22266, 22267, 22264, 22277, 22272, 22274, 22253, 22256, 22252, 22275, 22266, 22272, 22264, 22253, 22252, 22282, 22283, 22281, 22287, 22286, 22285, 22292, 22289, 22291, 22285, 22280, 22279, 22290, 22282, 22289, 22280, 22281, 22279, 22291, 22286, 22288, 22289, 22280, 22286, 22281, 22262, 22260, 22270, 22285, 22268, 22279, 22260, 22254, 22268, 22279, 22254, 22160, 22124, 22122, 22152, 22122, 22112, 22129, 22186, 22184, 22194, 22139, 22141, 22192, 22131, 22139, 22112, 22184, 22152, 22255, 22234, 22232, 22242, 22263, 22240,
22273, 22248, 22250, 22251, 22232, 22226, 22271, 22242, 22248, 22240, 22251, 22226, 21891, 22081, 21892, 21896, 22084, 21895, 22089, 21899, 22087, 22079, 21891, 21888, 21895, 22079, 21888, 22087, 21892, 22081, 21954, 22145, 21953, 21953, 22142, 21950, 21965, 22169, 21966, 22173, 21969, 22172, 22172, 21966, 22169, 21950, 22168, 21965, 22304, 22305, 22303, 22310, 22307, 22309, 22296, 22303, 22295, 22308, 22297, 22307, 22303, 22300, 22299, 22309, 22301, 22302, 22297, 22296, 22295, 22299, 22302, 22301, 22294, 22295, 22293, 22299, 22294, 22293, 22295, 22299, 22293, 22307, 22294, 22301, 22317, 22318, 22316, 22315, 22317, 22314, 22313, 22316, 22312, 22311, 22317, 22313, 22328, 22329, 22327, 22326, 22328, 22325, 22334, 22331, 22333, 22337, 22336, 22335, 22341, 22340, 22339, 22346, 22347, 22345, 22350, 22346, 22349, 22321, 22336, 22322, 22329, 22343, 22327, 22347, 22333, 22345, 22322, 22327, 22321, 22320, 22328, 22322, 22324, 22331, 22323, 22339, 22346, 22345, 22342, 22346, 22340, 22320, 22324, 22323, 22336, 22341, 22339, 22336, 22324, 22322, 22321, 22343, 22335, 22345, 22324, 22339, 22306, 22325, 22305, 22332, 22309, 22331, 22300, 22323, 22302, 22305, 22320, 22300, 22323, 22309, 22302, 22311, 22337, 22335, 22343, 22315, 22314, 22335, 22314, 22311,
22354, 22355, 22353, 22360, 22361, 22359, 22356, 22359, 22355, 22353, 22352, 22351, 22359, 22358, 22357, 22355, 22357, 22352, 22365, 22364, 22363, 22370, 22367, 22369, 22373, 22372, 22371, 22378, 22375, 22377, 22371, 22366, 22365, 22376, 22369, 22375, 22369, 22364, 22366, 22377, 22372, 22374, 22375, 22366, 22372, 22382, 22383, 22381, 22391, 22392, 22390, 22389, 22391, 22388, 22384, 22388, 22383, 22381, 22380, 22379, 22387, 22390, 22386, 22388, 22387, 22385, 22383, 22385, 22380, 22396, 22395, 22394, 22401, 22398, 22400, 22402, 22405, 22403, 22409, 22406, 22408, 22396, 22403, 22397, 22407, 22400, 22406, 22400, 22395, 22397, 22408, 22403, 22405, 22406, 22397, 22403, 22362, 22381, 22361, 22361, 22379, 22358, 22373, 22395, 22374, 22399, 22377, 22398, 22395, 22377, 22374, 22352, 22363, 22351, 22357, 22373, 22371, 22357, 22365, 22352, 22379, 22396, 22394, 22385, 22404, 22402, 22380, 22402, 22396, 22358, 22394, 22373, 22413, 22412, 22411, 22418, 22419, 22417, 22431, 22428, 22430, 22424, 22414, 22413, 22416, 22426, 22415, 22431, 22420, 22429, 22411, 22419, 22413, 22429, 22424, 22428, 22419, 22416, 22415, 22430, 22426, 22427, 22413, 22415, 22410, 22428, 22423, 22426, 22410, 22424, 22413, 22410, 22426, 22423, 22429, 22414, 22425, 22436, 22439, 22437,
22443, 22440, 22442, 22446, 22440, 22444, 22443, 22445, 22441, 22457, 22454, 22456, 22452, 22439, 22438, 22451, 22454, 22450, 22440, 22433, 22432, 22444, 22432, 22434, 22441, 22435, 22433, 22454, 22453, 22452, 22435, 22436, 22433, 22451, 22445, 22447, 22445, 22438, 22435, 22438, 22454, 22452, 22412, 22434, 22432, 22446, 22420, 22422, 22420, 22434, 22414, 22460, 22459, 22458, 22464, 22463, 22462, 22467, 22468, 22466, 22479, 22476, 22478, 22483, 22480, 22482, 22460, 22473, 22461, 22464, 22475, 22465, 22458, 22467, 22466, 22476, 22473, 22472, 22481, 22474, 22480, 22459, 22464, 22462, 22470, 22469, 22467, 22482, 22477, 22479, 22462, 22467, 22459, 22480, 22473, 22477, 22473, 22464, 22461, 22490, 22487, 22489, 22492, 22487, 22491, 22495, 22491, 22494, 22488, 22484, 22487, 22486, 22487, 22484, 22493, 22491, 22486, 22463, 22486, 22484, 22495, 22481, 22483, 22494, 22475, 22481, 22465, 22493, 22486, 22312, 22412, 22313, 22318, 22417, 22316, 22316, 22411, 22312, 22337, 22433, 22338, 22436, 22342, 22341, 22449, 22349, 22448, 22433, 22341, 22338, 22437, 22349, 22342, 22412, 22337, 22313, 22386, 22463, 22387, 22392, 22470, 22390, 22386, 22470, 22462, 22404, 22485, 22405, 22490, 22408, 22488, 22485, 22408, 22405, 22463, 22404, 22387, 22439, 22458, 22437,
22466, 22449, 22448, 22478, 22455, 22457, 22437, 22466, 22448, 22455, 22472, 22453, 22453, 22460, 22439, 22498, 22497, 22496, 22501, 22502, 22500, 22502, 22504, 22500, 22508, 22507, 22506, 22511, 22512, 22510, 22506, 22499, 22498, 22505, 22510, 22504, 22497, 22500, 22496, 22500, 22498, 22496, 22510, 22508, 22506, 22504, 22506, 22498, 22516, 22515, 22514, 22519, 22520, 22518, 22524, 22523, 22522, 22527, 22528, 22526, 22522, 22517, 22516, 22521, 22526, 22520, 22518, 22516, 22514, 22526, 22524, 22522, 22520, 22522, 22516, 22514, 22509, 22508, 22513, 22518, 22512, 22512, 22514, 22508, 22532, 22531, 22530, 22536, 22535, 22534, 22539, 22540, 22538, 22543, 22544, 22542, 22548, 22547, 22546, 22552, 22551, 22550, 22546, 22533, 22532, 22550, 22537, 22536, 22531, 22538, 22530, 22535, 22542, 22534, 22533, 22534, 22531, 22542, 22541, 22539, 22549, 22550, 22547, 22534, 22539, 22531, 22547, 22536, 22533, 22556, 22555, 22554, 22560, 22559, 22558, 22563, 22564, 22562, 22567, 22568, 22566, 22572, 22571, 22570, 22576, 22575, 22574, 22570, 22557, 22556, 22574, 22561, 22560, 22555, 22562, 22554, 22559, 22566, 22558, 22557, 22558, 22555, 22566, 22565, 22563, 22573, 22574, 22571, 22558, 22563, 22555, 22571, 22560, 22557, 22580, 22579, 22578, 22584, 22583, 22582,
22588, 22587, 22586, 22592, 22591, 22590, 22586, 22581, 22580, 22590, 22585, 22584, 22581, 22582, 22579, 22589, 22590, 22587, 22587, 22584, 22581, 22594, 22597, 22595, 22600, 22599, 22598, 22605, 22602, 22604, 22609, 22606, 22608, 22615, 22612, 22614, 22613, 22602, 22612, 22603, 22596, 22602, 22614, 22610, 22611, 22607, 22600, 22606, 22597, 22598, 22595, 22608, 22603, 22605, 22606, 22597, 22603, 22612, 22596, 22610, 22578, 22549, 22548, 22582, 22553, 22552, 22579, 22552, 22549, 22594, 22573, 22572, 22598, 22577, 22576, 22595, 22576, 22573, 22537, 22554, 22535, 22562, 22545, 22543, 22553, 22570, 22551, 22554, 22543, 22535, 22551, 22556, 22537, 22583, 22596, 22594, 22593, 22610, 22591, 22585, 22610, 22596, 22583, 22572, 22553, 22618, 22617, 22616, 22621, 22622, 22620, 22626, 22625, 22624, 22629, 22630, 22628, 22624, 22619, 22618, 22623, 22628, 22622, 22620, 22618, 22616, 22628, 22626, 22624, 22622, 22624, 22618, 22634, 22633, 22632, 22637, 22638, 22636, 22642, 22641, 22640, 22645, 22646, 22644, 22640, 22635, 22634, 22639, 22644, 22638, 22636, 22634, 22632, 22644, 22642, 22640, 22638, 22640, 22634, 22632, 22627, 22626, 22631, 22636, 22630, 22630, 22632, 22626, 22650, 22649, 22648, 22654, 22653, 22652, 22658, 22657, 22656, 22662, 22661, 22660,
22656, 22651, 22650, 22660, 22655, 22654, 22651, 22652, 22649, 22659, 22660, 22657, 22657, 22654, 22651, 22669, 22666, 22668, 22675, 22672, 22674, 22673, 22668, 22672, 22668, 22664, 22665, 22674, 22670, 22671, 22672, 22665, 22670, 22678, 22677, 22676, 22682, 22681, 22680, 22686, 22685, 22684, 22690, 22689, 22688, 22684, 22679, 22678, 22688, 22683, 22682, 22679, 22680, 22677, 22687, 22688, 22685, 22685, 22682, 22679, 22697, 22694, 22696, 22703, 22700, 22702, 22701, 22696, 22700, 22696, 22692, 22693, 22702, 22698, 22699, 22700, 22693, 22698, 22676, 22659, 22658, 22680, 22663, 22662, 22677, 22662, 22659, 22695, 22674, 22694, 22694, 22671, 22692, 22655, 22664, 22653, 22663, 22670, 22661, 22661, 22665, 22655, 22683, 22692, 22681, 22691, 22698, 22689, 22689, 22693, 22683, 22681, 22671, 22663, 22616, 22525, 22524, 22529, 22620, 22528, 22528, 22616, 22524, 22648, 22589, 22588, 22652, 22593, 22592, 22649, 22592, 22589, 22667, 22614, 22666, 22666, 22611, 22664, 22653, 22611, 22593, 22499, 22530, 22497, 22538, 22503, 22501, 22509, 22546, 22507, 22530, 22501, 22497, 22507, 22532, 22499, 22517, 22578, 22515, 22525, 22586, 22523, 22523, 22580, 22517, 22515, 22548, 22509, 22619, 22648, 22617, 22627, 22656, 22625, 22625, 22650, 22619, 22635, 22676, 22633,
22643, 22684, 22641, 22641, 22678, 22635, 22633, 22658, 22627, 22617, 22588, 22525, 22706, 22705, 22704, 22709, 22710, 22708, 22714, 22713, 22712, 22717, 22718, 22716, 22712, 22707, 22706, 22711, 22716, 22710, 22708, 22706, 22704, 22716, 22714, 22712, 22710, 22712, 22706, 22722, 22721, 22720, 22725, 22726, 22724, 22730, 22729, 22728, 22733, 22734, 22732, 22728, 22723, 22722, 22727, 22732, 22726, 22724, 22722, 22720, 22732, 22730, 22728, 22726, 22728, 22722, 22720, 22715, 22714, 22719, 22724, 22718, 22718, 22720, 22714, 22738, 22737, 22736, 22742, 22741, 22740, 22746, 22745, 22744, 22750, 22749, 22748, 22744, 22739, 22738, 22748, 22743, 22742, 22739, 22740, 22737, 22747, 22748, 22745, 22745, 22742, 22739, 22757, 22754, 22756, 22763, 22760, 22762, 22761, 22756, 22760, 22756, 22752, 22753, 22762, 22758, 22759, 22760, 22753, 22758, 22766, 22765, 22764, 22770, 22769, 22768, 22774, 22773, 22772, 22778, 22777, 22776, 22772, 22767, 22766, 22776, 22771, 22770, 22767, 22768, 22765, 22775, 22776, 22773, 22773, 22770, 22767, 22785, 22782, 22784, 22791, 22788, 22790, 22789, 22784, 22788, 22784, 22780, 22781, 22790, 22786, 22787, 22788, 22781, 22786, 22764, 22747, 22746, 22768, 22751, 22750, 22765, 22750, 22747, 22783, 22762, 22782, 22782, 22759, 22780,
22743, 22752, 22741, 22751, 22758, 22749, 22749, 22753, 22743, 22771, 22780, 22769, 22779, 22786, 22777, 22777, 22781, 22771, 22769, 22759, 22751, 22794, 22793, 22792, 22797, 22798, 22796, 22802, 22801, 22800, 22805, 22806, 22804, 22800, 22795, 22794, 22799, 22804, 22798, 22796, 22794, 22792, 22804, 22802, 22800, 22798, 22800, 22794, 22813, 22810, 22812, 22811, 22808, 22810, 22808, 22803, 22802, 22807, 22810, 22806, 22806, 22808, 22802, 22816, 22815, 22814, 22820, 22819, 22818, 22824, 22823, 22822, 22828, 22827, 22826, 22822, 22817, 22816, 22826, 22821, 22820, 22817, 22818, 22815, 22825, 22826, 22823, 22823, 22820, 22817, 22835, 22832, 22834, 22841, 22838, 22840, 22839, 22834, 22838, 22834, 22830, 22831, 22840, 22836, 22837, 22838, 22831, 22836, 22849, 22846, 22848, 22853, 22850, 22852, 22847, 22842, 22846, 22851, 22844, 22850, 22852, 22847, 22849, 22850, 22843, 22847, 22842, 22825, 22824, 22844, 22829, 22828, 22843, 22828, 22825, 22856, 22840, 22855, 22855, 22837, 22854, 22821, 22830, 22819, 22829, 22836, 22827, 22827, 22831, 22821, 22856, 22851, 22853, 22855, 22845, 22851, 22845, 22837, 22829, 22792, 22731, 22730, 22735, 22796, 22734, 22734, 22792, 22730, 22814, 22775, 22774, 22818, 22779, 22778, 22815, 22778, 22775, 22833, 22790, 22832,
22832, 22787, 22830, 22819, 22787, 22779, 22707, 22736, 22705, 22715, 22744, 22713, 22713, 22738, 22707, 22723, 22764, 22721, 22731, 22772, 22729, 22729, 22766, 22723, 22721, 22746, 22715, 22795, 22814, 22793, 22803, 22822, 22801, 22801, 22816, 22795, 22848, 22811, 22813, 22846, 22809, 22811, 22809, 22824, 22803, 22793, 22774, 22731, 22704, 22643, 22642, 22647, 22708, 22646, 22646, 22704, 22642, 22736, 22687, 22686, 22740, 22691, 22690, 22737, 22690, 22687, 22755, 22702, 22754, 22754, 22699, 22752, 22741, 22699, 22691, 22705, 22686, 22643, 22861, 22860, 22859, 22864, 22865, 22863, 22868, 22869, 22867, 22873, 22876, 22874, 22878, 22879, 22877, 22884, 22881, 22883, 22873, 22862, 22861, 22866, 22877, 22865, 22863, 22858, 22857, 22877, 22872, 22871, 22860, 22867, 22859, 22883, 22874, 22876, 22858, 22859, 22857, 22867, 22864, 22863, 22872, 22873, 22871, 22859, 22863, 22857, 22871, 22861, 22858, 22865, 22871, 22858, 22887, 22886, 22885, 22891, 22890, 22889, 22894, 22895, 22893, 22898, 22899, 22897, 22908, 22905, 22907, 22912, 22909, 22911, 22901, 22888, 22887, 22903, 22892, 22891, 22886, 22893, 22885, 22906, 22901, 22905, 22890, 22897, 22889, 22910, 22903, 22909, 22888, 22889, 22886, 22897, 22896, 22894, 22911, 22906, 22908, 22889, 22894, 22886,
22909, 22902, 22906, 22902, 22891, 22888, 22917, 22916, 22915, 22920, 22921, 22919, 22926, 22923, 22925, 22931, 22930, 22929, 22934, 22935, 22933, 22940, 22937, 22939, 22929, 22918, 22917, 22922, 22933, 22921, 22938, 22925, 22937, 22919, 22914, 22913, 22927, 22935, 22928, 22925, 22916, 22918, 22932, 22937, 22930, 22914, 22915, 22913, 22927, 22931, 22929, 22927, 22917, 22914, 22921, 22927, 22914, 22937, 22918, 22930, 22941, 22944, 22942, 22946, 22947, 22945, 22947, 22949, 22945, 22951, 22954, 22952, 22956, 22957, 22955, 22943, 22952, 22944, 22950, 22955, 22949, 22941, 22946, 22945, 22941, 22949, 22943, 22955, 22953, 22951, 22949, 22951, 22943, 22915, 22876, 22875, 22880, 22919, 22879, 22924, 22883, 22923, 22913, 22875, 22872, 22879, 22913, 22872, 22923, 22876, 22916, 22862, 22885, 22860, 22893, 22870, 22868, 22907, 22881, 22882, 22885, 22868, 22860, 22905, 22874, 22881, 22874, 22887, 22862, 22961, 22960, 22959, 22965, 22964, 22963, 22968, 22969, 22967, 22972, 22973, 22971, 22983, 22980, 22982, 22987, 22984, 22986, 22987, 22988, 22985, 22975, 22962, 22961, 22977, 22966, 22965, 22960, 22967, 22959, 22981, 22975, 22980, 22964, 22971, 22963, 22985, 22977, 22984, 22985, 22979, 22978, 22962, 22963, 22960, 22971, 22970, 22968, 22986, 22981, 22983,
22963, 22968, 22960, 22984, 22976, 22981, 22976, 22965, 22962, 22992, 22991, 22990, 22995, 22996, 22994, 22999, 22995, 22998, 23002, 23001, 23000, 23007, 23004, 23006, 23000, 22993, 22992, 23005, 22998, 23004, 22991, 22994, 22990, 22998, 22991, 22993, 23006, 23001, 23003, 23004, 22993, 23001, 23015, 23012, 23014, 23017, 23018, 23016, 23027, 23024, 23026, 23029, 23030, 23028, 23025, 23014, 23024, 23019, 23028, 23018, 23009, 23012, 23008, 23021, 23024, 23020, 23016, 23011, 23010, 23028, 23023, 23022, 23020, 23014, 23009, 23018, 23022, 23011, 23034, 23033, 23032, 23039, 23036, 23038, 23042, 23041, 23040, 23047, 23044, 23046, 23040, 23035, 23034, 23045, 23038, 23044, 23038, 23033, 23035, 23046, 23041, 23043, 23044, 23035, 23041, 22989, 23016, 22988, 22988, 23010, 22979, 23032, 23003, 23002, 23037, 23006, 23036, 23036, 23003, 23033, 22966, 22990, 22964, 22994, 22974, 22972, 22979, 23000, 22978, 22990, 22972, 22964, 22978, 22992, 22966, 23011, 23032, 23010, 23023, 23040, 23022, 23011, 23040, 23034, 23010, 23002, 22979, 23050, 23053, 23051, 23055, 23056, 23054, 23063, 23058, 23061, 23064, 23067, 23065, 23070, 23072, 23068, 23064, 23053, 23052, 23062, 23068, 23060, 23071, 23061, 23069, 23054, 23049, 23048, 23053, 23058, 23051, 23064, 23072, 23066,
23049, 23050, 23048, 23062, 23056, 23057, 23056, 23052, 23049, 23060, 23064, 23052, 23069, 23053, 23065, 23074, 23077, 23075, 23079, 23080, 23078, 23085, 23088, 23086, 23092, 23093, 23091, 23093, 23089, 23091, 23096, 23097, 23095, 23076, 23086, 23077, 23081, 23095, 23080, 23083, 23092, 23091, 23091, 23082, 23083, 23078, 23076, 23074, 23095, 23087, 23085, 23097, 23094, 23092, 23087, 23092, 23084, 23080, 23085, 23076, 23104, 23101, 23103, 23101, 23100, 23099, 23107, 23106, 23105, 23111, 23110, 23109, 23118, 23115, 23117, 23114, 23115, 23113, 23122, 23119, 23121, 23116, 23107, 23115, 23105, 23115, 23107, 23119, 23112, 23111, 23106, 23111, 23109, 23121, 23116, 23118, 23119, 23108, 23116, 23099, 23067, 23066, 23073, 23101, 23072, 23066, 23101, 23099, 23083, 23106, 23084, 23109, 23088, 23087, 23084, 23109, 23087, 23067, 23082, 23065, 23089, 23071, 23069, 23065, 23089, 23069, 23114, 23102, 23104, 23113, 23100, 23102, 23067, 23105, 23083, 23130, 23127, 23129, 23132, 23133, 23131, 23137, 23136, 23135, 23139, 23142, 23140, 23144, 23145, 23143, 23148, 23144, 23147, 23150, 23151, 23149, 23151, 23153, 23149, 23145, 23129, 23143, 23134, 23150, 23133, 23124, 23127, 23123, 23135, 23144, 23143, 23147, 23136, 23138, 23131, 23126, 23125, 23140, 23149, 23139,
23139, 23153, 23141, 23153, 23148, 23147, 23141, 23147, 23138, 23143, 23124, 23135, 23126, 23150, 23140, 23155, 23158, 23156, 23162, 23159, 23161, 23169, 23166, 23168, 23171, 23166, 23170, 23157, 23164, 23158, 23169, 23161, 23167, 23161, 23156, 23158, 23167, 23163, 23166, 23165, 23166, 23163, 23164, 23161, 23158, 23174, 23173, 23172, 23182, 23179, 23181, 23186, 23183, 23185, 23188, 23183, 23187, 23179, 23175, 23174, 23184, 23176, 23183, 23187, 23176, 23178, 23173, 23178, 23176, 23188, 23180, 23182, 23180, 23178, 23175, 23137, 23173, 23138, 23176, 23142, 23141, 23173, 23141, 23138, 23191, 23170, 23190, 23190, 23165, 23189, 23125, 23157, 23155, 23142, 23163, 23140, 23140, 23157, 23126, 23191, 23184, 23186, 23190, 23177, 23184, 23142, 23189, 23165, 22931, 23051, 22932, 22936, 23054, 22935, 23059, 22939, 23058, 23048, 22931, 22928, 22928, 23054, 23048, 23051, 22939, 22932, 22953, 23075, 22954, 22958, 23078, 22957, 22957, 23074, 22953, 23128, 23026, 23127, 23031, 23131, 23030, 23123, 23026, 23021, 23030, 23125, 23023, 23155, 23043, 23042, 23160, 23046, 23159, 23159, 23043, 23156, 23023, 23155, 23042, 22892, 22959, 22890, 22967, 22900, 22898, 22982, 22910, 22912, 22959, 22898, 22890, 22980, 22904, 22910, 22904, 22961, 22892, 22944, 23008, 22942,
23012, 22948, 22946, 22952, 23021, 23020, 23008, 22946, 22942, 22952, 23009, 22944, 23077, 23123, 23075, 23086, 23137, 23135, 23086, 23124, 23077, 23110, 23174, 23172, 23181, 23120, 23122, 23120, 23174, 23112, 23088, 23172, 23137, 23075, 23021, 22954, 23201, 23204, 23202, 23208, 23209, 23207, 23206, 23208, 23205, 23212, 23213, 23211, 23197, 23212, 23196, 23200, 23207, 23199, 23205, 23200, 23198, 23201, 23212, 23211, 23198, 23203, 23201, 23211, 23206, 23205, 23201, 23205, 23198, 23196, 23202, 23195, 23216, 23219, 23217, 23223, 23224, 23222, 23224, 23226, 23222, 23233, 23232, 23231, 23238, 23239, 23237, 23236, 23238, 23235, 23231, 23219, 23218, 23221, 23235, 23220, 23230, 23237, 23229, 23235, 23230, 23228, 23216, 23223, 23222, 23216, 23226, 23218, 23226, 23221, 23220, 23228, 23233, 23231, 23218, 23220, 23215, 23215, 23231, 23218, 23220, 23228, 23215, 23243, 23242, 23241, 23247, 23246, 23245, 23251, 23250, 23249, 23255, 23254, 23253, 23249, 23244, 23243, 23253, 23248, 23247, 23244, 23245, 23242, 23252, 23253, 23250, 23250, 23247, 23244, 23194, 23223, 23193, 23193, 23217, 23192, 23199, 23242, 23200, 23245, 23204, 23203, 23242, 23203, 23200, 23207, 23194, 23193, 23199, 23193, 23192, 23219, 23241, 23217, 23234, 23249, 23232, 23232, 23243, 23219,
23192, 23241, 23199, 23263, 23264, 23262, 23259, 23262, 23258, 23262, 23261, 23260, 23258, 23260, 23257, 23269, 23270, 23268, 23278, 23279, 23277, 23276, 23278, 23275, 23271, 23275, 23270, 23268, 23267, 23266, 23274, 23277, 23273, 23272, 23278, 23274, 23270, 23272, 23267, 23265, 23268, 23264, 23264, 23266, 23261, 23281, 23284, 23282, 23287, 23286, 23285, 23290, 23291, 23289, 23295, 23294, 23293, 23299, 23298, 23297, 23293, 23284, 23283, 23297, 23288, 23287, 23289, 23283, 23281, 23284, 23285, 23282, 23296, 23297, 23294, 23294, 23287, 23284, 23303, 23302, 23301, 23305, 23308, 23306, 23314, 23311, 23313, 23314, 23309, 23312, 23315, 23318, 23316, 23323, 23320, 23322, 23315, 23304, 23303, 23325, 23311, 23324, 23322, 23316, 23318, 23311, 23308, 23307, 23308, 23309, 23306, 23304, 23305, 23302, 23325, 23320, 23321, 23324, 23316, 23320, 23304, 23319, 23307, 23319, 23311, 23307, 23328, 23327, 23326, 23332, 23331, 23330, 23336, 23335, 23334, 23340, 23339, 23338, 23334, 23329, 23328, 23338, 23333, 23332, 23329, 23330, 23327, 23337, 23338, 23335, 23335, 23332, 23329, 23348, 23345, 23347, 23350, 23345, 23349, 23356, 23353, 23355, 23354, 23349, 23353, 23346, 23342, 23345, 23344, 23345, 23342, 23355, 23351, 23352, 23353, 23344, 23351, 23326, 23296, 23295,
23330, 23300, 23299, 23327, 23299, 23296, 23317, 23343, 23318, 23348, 23322, 23346, 23343, 23322, 23318, 23288, 23301, 23286, 23300, 23315, 23298, 23298, 23303, 23288, 23331, 23344, 23342, 23339, 23352, 23351, 23333, 23351, 23344, 23331, 23317, 23300, 23229, 23282, 23230, 23285, 23234, 23233, 23239, 23289, 23237, 23282, 23233, 23230, 23237, 23281, 23229, 23301, 23252, 23251, 23305, 23256, 23255, 23302, 23255, 23252, 23286, 23251, 23234, 23291, 23259, 23258, 23260, 23295, 23293, 23283, 23258, 23257, 23257, 23293, 23283, 23266, 23328, 23326, 23274, 23334, 23272, 23267, 23334, 23328, 23261, 23326, 23295, 23361, 23360, 23359, 23364, 23365, 23363, 23368, 23369, 23367, 23363, 23358, 23357, 23360, 23367, 23359, 23358, 23359, 23357, 23367, 23364, 23363, 23359, 23363, 23357, 23375, 23374, 23373, 23379, 23378, 23377, 23382, 23383, 23381, 23388, 23385, 23387, 23390, 23391, 23389, 23396, 23393, 23395, 23374, 23381, 23373, 23386, 23375, 23385, 23378, 23389, 23377, 23394, 23379, 23393, 23376, 23377, 23374, 23389, 23384, 23382, 23395, 23386, 23388, 23377, 23382, 23374, 23393, 23376, 23386, 23362, 23373, 23360, 23381, 23370, 23368, 23387, 23371, 23372, 23373, 23368, 23360, 23385, 23362, 23371, 23399, 23398, 23397, 23403, 23402, 23401, 23406, 23407, 23405,
23412, 23409, 23411, 23414, 23415, 23413, 23420, 23417, 23419, 23398, 23405, 23397, 23410, 23399, 23409, 23402, 23413, 23401, 23418, 23403, 23417, 23400, 23401, 23398, 23413, 23408, 23406, 23419, 23410, 23412, 23401, 23406, 23398, 23417, 23400, 23410, 23423, 23422, 23421, 23426, 23427, 23425, 23432, 23429, 23431, 23432, 23426, 23430, 23422, 23425, 23421, 23430, 23423, 23429, 23430, 23422, 23424, 23404, 23421, 23402, 23425, 23416, 23414, 23431, 23418, 23420, 23421, 23414, 23402, 23429, 23404, 23418, 23437, 23436, 23435, 23440, 23441, 23439, 23446, 23443, 23445, 23451, 23450, 23449, 23454, 23455, 23453, 23460, 23457, 23459, 23449, 23438, 23437, 23442, 23453, 23441, 23458, 23445, 23457, 23439, 23434, 23433, 23453, 23448, 23447, 23445, 23436, 23438, 23459, 23450, 23452, 23434, 23435, 23433, 23448, 23449, 23447, 23447, 23437, 23434, 23441, 23447, 23434, 23457, 23438, 23450, 23465, 23464, 23463, 23468, 23469, 23467, 23474, 23471, 23473, 23479, 23478, 23477, 23482, 23483, 23481, 23490, 23487, 23489, 23490, 23485, 23488, 23477, 23466, 23465, 23470, 23481, 23469, 23486, 23473, 23485, 23467, 23462, 23461, 23481, 23476, 23475, 23473, 23464, 23466, 23488, 23479, 23487, 23488, 23478, 23480, 23462, 23463, 23461, 23476, 23477, 23475, 23489, 23483, 23484,
23487, 23476, 23483, 23475, 23465, 23462, 23469, 23475, 23462, 23485, 23466, 23478, 23463, 23452, 23451, 23456, 23467, 23455, 23472, 23459, 23471, 23461, 23451, 23448, 23455, 23461, 23448, 23471, 23452, 23464, 23435, 23362, 23361, 23366, 23439, 23365, 23444, 23371, 23443, 23433, 23361, 23358, 23365, 23433, 23358, 23443, 23362, 23436, 23380, 23397, 23378, 23405, 23392, 23390, 23411, 23394, 23396, 23397, 23390, 23378, 23409, 23380, 23394, 23491, 23494, 23492, 23496, 23497, 23495, 23501, 23500, 23499, 23504, 23505, 23503, 23493, 23500, 23494, 23498, 23503, 23497, 23495, 23493, 23491, 23503, 23501, 23499, 23497, 23499, 23493, 23509, 23508, 23507, 23512, 23513, 23511, 23517, 23516, 23515, 23524, 23521, 23523, 23520, 23521, 23519, 23515, 23510, 23509, 23514, 23519, 23513, 23507, 23513, 23509, 23521, 23518, 23517, 23519, 23517, 23515, 23509, 23519, 23515, 23507, 23502, 23501, 23506, 23511, 23505, 23501, 23511, 23507, 23527, 23526, 23525, 23531, 23530, 23529, 23535, 23534, 23533, 23539, 23538, 23537, 23533, 23528, 23527, 23537, 23532, 23531, 23528, 23529, 23526, 23536, 23537, 23534, 23534, 23531, 23528, 23546, 23543, 23545, 23552, 23549, 23551, 23550, 23545, 23549, 23545, 23541, 23542, 23551, 23547, 23548, 23549, 23542, 23547, 23555, 23554, 23553,
23559, 23558, 23557, 23563, 23562, 23561, 23567, 23566, 23565, 23561, 23556, 23555, 23565, 23560, 23559, 23556, 23557, 23554, 23564, 23565, 23562, 23562, 23559, 23556, 23574, 23571, 23573, 23580, 23577, 23579, 23578, 23573, 23577, 23570, 23571, 23569, 23576, 23577, 23575, 23575, 23573, 23570, 23553, 23536, 23535, 23557, 23540, 23539, 23554, 23539, 23536, 23572, 23551, 23571, 23569, 23551, 23548, 23530, 23542, 23541, 23538, 23548, 23547, 23532, 23547, 23542, 23560, 23569, 23558, 23568, 23575, 23566, 23566, 23570, 23560, 23558, 23548, 23540, 23584, 23585, 23583, 23590, 23591, 23589, 23586, 23589, 23585, 23581, 23585, 23582, 23587, 23591, 23588, 23582, 23589, 23587, 23592, 23594, 23591, 23591, 23593, 23588, 23598, 23597, 23596, 23602, 23601, 23600, 23606, 23605, 23604, 23610, 23609, 23608, 23604, 23599, 23598, 23608, 23603, 23602, 23599, 23600, 23597, 23607, 23608, 23605, 23605, 23602, 23599, 23618, 23619, 23617, 23619, 23615, 23617, 23623, 23622, 23621, 23629, 23626, 23628, 23621, 23614, 23613, 23627, 23618, 23626, 23613, 23618, 23617, 23613, 23615, 23612, 23624, 23626, 23622, 23630, 23629, 23628, 23625, 23628, 23624, 23622, 23618, 23614, 23634, 23633, 23632, 23638, 23637, 23636, 23641, 23642, 23640, 23646, 23645, 23644, 23650, 23649, 23648,
23657, 23654, 23656, 23653, 23654, 23652, 23644, 23635, 23634, 23648, 23639, 23638, 23643, 23652, 23642, 23640, 23634, 23632, 23655, 23646, 23654, 23644, 23654, 23646, 23635, 23636, 23633, 23645, 23650, 23648, 23645, 23638, 23635, 23634, 23652, 23644, 23660, 23659, 23658, 23668, 23669, 23667, 23669, 23665, 23667, 23673, 23672, 23671, 23675, 23678, 23676, 23671, 23661, 23660, 23675, 23664, 23663, 23680, 23668, 23679, 23664, 23667, 23663, 23663, 23665, 23662, 23661, 23662, 23659, 23674, 23675, 23672, 23661, 23675, 23663, 23676, 23668, 23664, 23606, 23633, 23607, 23636, 23611, 23610, 23633, 23610, 23607, 23623, 23659, 23624, 23666, 23630, 23665, 23659, 23625, 23624, 23662, 23630, 23625, 23603, 23612, 23601, 23611, 23621, 23609, 23603, 23621, 23613, 23639, 23658, 23637, 23651, 23671, 23649, 23649, 23660, 23639, 23637, 23623, 23611, 23524, 23583, 23522, 23522, 23581, 23518, 23596, 23564, 23563, 23600, 23568, 23567, 23597, 23567, 23564, 23616, 23579, 23615, 23612, 23579, 23576, 23601, 23576, 23568, 23494, 23525, 23492, 23502, 23533, 23500, 23500, 23527, 23494, 23510, 23553, 23508, 23516, 23563, 23561, 23516, 23555, 23510, 23508, 23535, 23502, 23582, 23596, 23581, 23588, 23604, 23587, 23587, 23598, 23582, 23641, 23594, 23595, 23594, 23632, 23593,
23593, 23606, 23588, 23581, 23563, 23518, 23273, 23492, 23274, 23279, 23495, 23277, 23273, 23495, 23491, 23525, 23337, 23336, 23529, 23341, 23340, 23526, 23340, 23337, 23544, 23355, 23543, 23543, 23352, 23541, 23341, 23541, 23352, 23492, 23336, 23274, 22278, 22353, 22276, 22276, 22351, 22270, 22363, 22288, 22287, 22368, 22291, 22367, 22367, 22288, 22364, 22270, 22363, 22287, 23685, 23684, 23683, 23688, 23689, 23687, 23692, 23693, 23691, 23697, 23700, 23698, 23702, 23703, 23701, 23708, 23705, 23707, 23685, 23698, 23686, 23690, 23701, 23689, 23687, 23682, 23681, 23701, 23696, 23695, 23684, 23691, 23683, 23707, 23698, 23700, 23682, 23683, 23681, 23691, 23688, 23687, 23696, 23697, 23695, 23683, 23687, 23681, 23695, 23685, 23682, 23689, 23695, 23682, 23711, 23710, 23709, 23715, 23714, 23713, 23718, 23719, 23717, 23722, 23723, 23721, 23734, 23731, 23733, 23738, 23735, 23737, 23740, 23741, 23739, 23725, 23712, 23711, 23727, 23716, 23715, 23710, 23717, 23709, 23732, 23725, 23731, 23714, 23721, 23713, 23736, 23727, 23735, 23730, 23739, 23729, 23712, 23713, 23710, 23721, 23720, 23718, 23737, 23732, 23734, 23713, 23718, 23710, 23735, 23726, 23732, 23726, 23715, 23712, 23747, 23746, 23745, 23750, 23751, 23749, 23756, 23753, 23755, 23761, 23760, 23759,
23764, 23765, 23763, 23770, 23767, 23769, 23759, 23748, 23747, 23752, 23763, 23751, 23768, 23755, 23767, 23749, 23744, 23743, 23763, 23758, 23757, 23755, 23746, 23748, 23769, 23760, 23762, 23744, 23745, 23743, 23758, 23759, 23757, 23757, 23747, 23744, 23751, 23757, 23744, 23767, 23748, 23760, 23773, 23772, 23771, 23776, 23777, 23775, 23781, 23780, 23779, 23784, 23785, 23783, 23779, 23774, 23773, 23778, 23783, 23777, 23775, 23773, 23771, 23783, 23781, 23779, 23777, 23779, 23773, 23745, 23700, 23699, 23704, 23749, 23703, 23754, 23707, 23753, 23743, 23699, 23696, 23703, 23743, 23696, 23753, 23700, 23746, 23771, 23730, 23729, 23741, 23775, 23739, 23739, 23771, 23729, 23686, 23709, 23684, 23717, 23694, 23692, 23733, 23705, 23706, 23709, 23692, 23684, 23731, 23698, 23705, 23698, 23711, 23686, 23789, 23788, 23787, 23793, 23792, 23791, 23796, 23797, 23795, 23800, 23801, 23799, 23812, 23809, 23811, 23818, 23815, 23817, 23818, 23819, 23816, 23803, 23790, 23789, 23806, 23794, 23793, 23788, 23795, 23787, 23810, 23803, 23809, 23792, 23799, 23791, 23816, 23806, 23815, 23816, 23808, 23807, 23790, 23791, 23788, 23799, 23798, 23796, 23817, 23810, 23812, 23791, 23796, 23788, 23815, 23804, 23810, 23804, 23793, 23790, 23823, 23822, 23821, 23826, 23827, 23825,
23830, 23826, 23829, 23833, 23832, 23831, 23838, 23835, 23837, 23831, 23824, 23823, 23836, 23829, 23835, 23822, 23825, 23821, 23829, 23822, 23824, 23837, 23832, 23834, 23835, 23824, 23832, 23846, 23843, 23845, 23848, 23849, 23847, 23858, 23855, 23857, 23860, 23861, 23859, 23856, 23845, 23855, 23850, 23859, 23849, 23845, 23839, 23840, 23857, 23851, 23852, 23847, 23842, 23841, 23859, 23854, 23853, 23855, 23840, 23851, 23849, 23853, 23842, 23865, 23864, 23863, 23870, 23867, 23869, 23873, 23872, 23871, 23878, 23875, 23877, 23871, 23866, 23865, 23876, 23869, 23875, 23869, 23864, 23866, 23877, 23872, 23874, 23875, 23866, 23872, 23844, 23813, 23843, 23820, 23847, 23819, 23843, 23805, 23839, 23819, 23841, 23808, 23863, 23834, 23833, 23868, 23837, 23867, 23867, 23834, 23864, 23794, 23821, 23792, 23825, 23802, 23800, 23808, 23831, 23807, 23821, 23800, 23792, 23807, 23823, 23794, 23842, 23863, 23841, 23854, 23871, 23853, 23853, 23865, 23842, 23841, 23833, 23808, 23716, 23787, 23714, 23795, 23724, 23722, 23811, 23736, 23738, 23813, 23742, 23740, 23787, 23722, 23714, 23809, 23728, 23736, 23805, 23740, 23730, 23728, 23789, 23716, 23774, 23839, 23772, 23782, 23851, 23780, 23780, 23840, 23774, 23772, 23805, 23730, 23882, 23881, 23880, 23891, 23888, 23890,
23887, 23888, 23886, 23891, 23893, 23889, 23889, 23882, 23888, 23886, 23882, 23880, 23887, 23884, 23885, 23884, 23880, 23879, 23883, 23893, 23892, 23897, 23896, 23895, 23901, 23900, 23899, 23905, 23904, 23903, 23907, 23910, 23908, 23916, 23913, 23915, 23912, 23913, 23911, 23897, 23904, 23898, 23907, 23902, 23901, 23913, 23906, 23905, 23903, 23913, 23905, 23898, 23899, 23896, 23906, 23907, 23904, 23918, 23914, 23916, 23914, 23909, 23906, 23904, 23901, 23898, 23924, 23921, 23923, 23922, 23919, 23921, 23919, 23910, 23909, 23918, 23921, 23917, 23909, 23921, 23919, 23881, 23897, 23895, 23912, 23893, 23894, 23911, 23892, 23893, 23892, 23897, 23883, 23928, 23929, 23927, 23933, 23932, 23931, 23936, 23937, 23935, 23931, 23926, 23925, 23929, 23935, 23927, 23926, 23927, 23925, 23935, 23933, 23931, 23927, 23931, 23925, 23944, 23945, 23943, 23948, 23949, 23947, 23953, 23952, 23951, 23957, 23956, 23955, 23951, 23940, 23939, 23955, 23942, 23941, 23940, 23943, 23939, 23942, 23947, 23941, 23947, 23946, 23944, 23954, 23955, 23952, 23941, 23944, 23940, 23952, 23941, 23940, 23964, 23965, 23963, 23968, 23969, 23967, 23973, 23972, 23971, 23977, 23976, 23975, 23971, 23960, 23959, 23975, 23962, 23961, 23960, 23963, 23959, 23962, 23967, 23961, 23967, 23966, 23964,
23974, 23975, 23972, 23961, 23964, 23960, 23972, 23961, 23960, 23963, 23950, 23948, 23958, 23971, 23956, 23959, 23948, 23942, 23956, 23959, 23942, 23943, 23930, 23928, 23934, 23951, 23932, 23939, 23928, 23926, 23932, 23939, 23926, 23983, 23982, 23981, 23986, 23987, 23985, 23992, 23989, 23991, 24001, 23998, 24000, 23994, 23984, 23983, 23988, 23996, 23987, 24001, 23991, 23999, 23985, 23980, 23979, 23991, 23982, 23984, 23999, 23994, 23998, 23980, 23981, 23979, 24000, 23996, 23997, 23998, 23993, 23996, 23993, 23983, 23980, 23987, 23993, 23980, 23999, 23984, 23995, 24004, 24003, 24002, 24011, 24008, 24010, 24007, 24008, 24006, 24009, 24004, 24008, 24006, 24004, 24002, 24019, 24016, 24018, 24021, 24022, 24020, 24023, 24025, 24022, 24018, 24012, 24013, 24020, 24015, 24014, 24022, 24024, 24015, 24029, 24028, 24027, 24034, 24031, 24033, 24040, 24037, 24039, 24035, 24030, 24029, 24040, 24033, 24038, 24033, 24028, 24030, 24038, 24035, 24037, 24038, 24030, 24036, 24015, 24027, 24014, 24039, 24025, 24026, 24037, 24024, 24025, 24024, 24029, 24015, 24047, 24048, 24046, 24043, 24046, 24042, 24046, 24045, 24044, 24042, 24044, 24041, 24055, 24056, 24054, 24059, 24060, 24058, 24051, 24054, 24050, 24053, 24058, 24052, 24058, 24057, 24055, 24052, 24055, 24051,
24066, 24065, 24064, 24069, 24070, 24068, 24075, 24072, 24074, 24068, 24063, 24062, 24073, 24066, 24072, 24063, 24064, 24062, 24074, 24070, 24071, 24072, 24063, 24070, 24078, 24077, 24076, 24082, 24081, 24080, 24087, 24084, 24086, 24091, 24088, 24090, 24085, 24078, 24084, 24089, 24082, 24088, 24079, 24080, 24077, 24090, 24085, 24087, 24088, 24079, 24085, 24064, 24043, 24042, 24045, 24068, 24044, 24062, 24042, 24041, 24044, 24062, 24041, 24076, 24051, 24050, 24080, 24053, 24052, 24077, 24052, 24051, 24054, 24049, 24047, 24050, 24047, 24043, 24067, 24076, 24065, 24086, 24073, 24075, 24084, 24067, 24073, 24065, 24050, 24043, 24100, 24101, 24099, 24104, 24105, 24103, 24096, 24099, 24095, 24098, 24103, 24097, 24103, 24102, 24100, 24097, 24100, 24096, 24110, 24111, 24109, 24108, 24109, 24107, 24115, 24114, 24113, 24119, 24118, 24117, 24124, 24121, 24123, 24125, 24128, 24126, 24129, 24132, 24130, 24134, 24135, 24133, 24135, 24137, 24133, 24144, 24141, 24143, 24144, 24139, 24142, 24129, 24120, 24119, 24124, 24134, 24122, 24122, 24115, 24121, 24126, 24133, 24125, 24125, 24137, 24127, 24142, 24131, 24141, 24132, 24139, 24130, 24116, 24117, 24114, 24128, 24129, 24126, 24116, 24129, 24119, 24116, 24134, 24126, 24151, 24148, 24150, 24153, 24148, 24152,
24156, 24152, 24155, 24149, 24145, 24148, 24152, 24145, 24147, 24155, 24147, 24154, 24113, 24096, 24095, 24117, 24098, 24097, 24114, 24097, 24096, 24107, 24146, 24108, 24151, 24110, 24149, 24146, 24110, 24108, 24109, 24106, 24104, 24107, 24104, 24098, 24118, 24147, 24145, 24156, 24139, 24140, 24155, 24130, 24139, 24120, 24154, 24147, 24118, 24107, 24098, 24005, 24012, 24003, 24019, 24009, 24011, 24018, 24005, 24009, 24099, 24061, 24059, 24095, 24059, 24053, 24083, 24113, 24081, 24123, 24089, 24091, 24137, 24094, 24093, 24121, 24083, 24089, 24127, 24093, 24092, 24081, 24095, 24053, 24159, 24158, 24157, 24162, 24163, 24161, 24170, 24167, 24169, 24165, 24160, 24159, 24164, 24167, 24163, 24161, 24159, 24157, 24168, 24165, 24167, 24163, 24165, 24159, 24173, 24172, 24171, 24177, 24176, 24175, 24186, 24183, 24185, 24190, 24187, 24189, 24179, 24174, 24173, 24181, 24178, 24177, 24184, 24179, 24183, 24188, 24181, 24187, 24174, 24175, 24172, 24189, 24184, 24186, 24187, 24180, 24184, 24180, 24177, 24174, 24193, 24192, 24191, 24197, 24196, 24195, 24206, 24203, 24205, 24210, 24207, 24209, 24199, 24194, 24193, 24201, 24198, 24197, 24204, 24199, 24203, 24208, 24201, 24207, 24194, 24195, 24192, 24209, 24204, 24206, 24207, 24200, 24204, 24200, 24197, 24194,
24178, 24191, 24176, 24205, 24188, 24190, 24203, 24182, 24188, 24182, 24193, 24178, 24214, 24215, 24213, 24212, 24213, 24211, 24219, 24218, 24217, 24222, 24223, 24221, 24227, 24226, 24225, 24230, 24231, 24229, 24225, 24220, 24219, 24224, 24229, 24223, 24221, 24219, 24217, 24229, 24227, 24225, 24223, 24225, 24219, 24217, 24212, 24211, 24215, 24221, 24213, 24213, 24217, 24211, 24238, 24239, 24237, 24242, 24243, 24241, 24234, 24237, 24233, 24236, 24241, 24235, 24241, 24240, 24238, 24235, 24238, 24234, 24250, 24251, 24249, 24254, 24255, 24253, 24246, 24249, 24245, 24248, 24253, 24247, 24253, 24252, 24250, 24247, 24250, 24246, 24259, 24258, 24257, 24263, 24262, 24261, 24267, 24266, 24265, 24271, 24270, 24269, 24265, 24260, 24259, 24269, 24264, 24263, 24260, 24261, 24258, 24268, 24269, 24266, 24266, 24263, 24260, 24275, 24274, 24273, 24279, 24278, 24277, 24283, 24282, 24281, 24287, 24286, 24285, 24281, 24276, 24275, 24285, 24280, 24279, 24276, 24277, 24274, 24284, 24285, 24282, 24282, 24279, 24276, 24257, 24234, 24233, 24261, 24236, 24235, 24258, 24235, 24234, 24273, 24246, 24245, 24277, 24248, 24247, 24274, 24247, 24246, 24249, 24244, 24242, 24245, 24242, 24236, 24264, 24273, 24262, 24272, 24281, 24270, 24270, 24275, 24264, 24262, 24245, 24236,
24160, 24171, 24158, 24185, 24168, 24170, 24183, 24166, 24168, 24166, 24173, 24160, 24237, 24216, 24214, 24233, 24214, 24212, 24220, 24257, 24218, 24228, 24265, 24226, 24226, 24259, 24220, 24218, 24233, 24212, 23981, 23762, 23761, 23766, 23985, 23765, 23990, 23769, 23989, 23979, 23761, 23758, 23765, 23979, 23758, 23989, 23762, 23982, 24002, 23782, 23781, 23786, 24006, 23785, 23785, 24002, 23781, 24017, 23857, 24016, 23862, 24020, 23861, 24016, 23852, 24012, 23861, 24014, 23854, 24027, 23874, 23873, 24032, 23877, 24031, 24031, 23874, 24028, 24014, 23873, 23854, 24003, 23852, 23782, 24157, 23934, 23933, 23938, 24161, 23937, 23937, 24157, 23933, 24171, 23954, 23953, 24175, 23958, 23957, 24172, 23957, 23954, 24191, 23974, 23973, 24195, 23978, 23977, 24192, 23977, 23974, 24176, 23973, 23958, 24158, 23953, 23934, 24293, 24296, 24294, 24300, 24301, 24299, 24301, 24303, 24299, 24305, 24308, 24306, 24309, 24312, 24310, 24314, 24315, 24313, 24315, 24317, 24313, 24324, 24321, 24323, 24324, 24319, 24322, 24309, 24296, 24295, 24298, 24314, 24297, 24306, 24313, 24305, 24305, 24317, 24307, 24294, 24299, 24293, 24293, 24303, 24295, 24322, 24311, 24321, 24312, 24319, 24310, 24303, 24298, 24297, 24308, 24309, 24306, 24295, 24297, 24292, 24292, 24309, 24295,
24292, 24314, 24306, 24325, 24328, 24326, 24330, 24331, 24329, 24331, 24333, 24329, 24340, 24337, 24339, 24327, 24336, 24328, 24334, 24337, 24333, 24326, 24329, 24325, 24329, 24327, 24325, 24338, 24335, 24337, 24333, 24335, 24327, 24341, 24344, 24342, 24351, 24348, 24350, 24355, 24356, 24354, 24365, 24362, 24364, 24369, 24366, 24368, 24343, 24359, 24344, 24360, 24347, 24346, 24344, 24348, 24342, 24363, 24358, 24362, 24347, 24354, 24346, 24367, 24360, 24366, 24353, 24348, 24349, 24354, 24351, 24350, 24368, 24363, 24365, 24352, 24342, 24348, 24344, 24354, 24350, 24366, 24359, 24363, 24344, 24360, 24346, 24291, 24330, 24290, 24289, 24330, 24326, 24341, 24308, 24307, 24353, 24321, 24352, 24308, 24345, 24311, 24345, 24321, 24311, 24317, 24291, 24290, 24307, 24290, 24289, 24328, 24341, 24326, 24364, 24338, 24340, 24362, 24336, 24338, 24336, 24343, 24328, 24289, 24341, 24307, 24370, 24373, 24371, 24380, 24377, 24379, 24380, 24375, 24378, 24385, 24377, 24384, 24378, 24372, 24377, 24378, 24371, 24373, 24382, 24375, 24376, 24381, 24371, 24375, 24383, 24377, 24372, 24391, 24392, 24390, 24395, 24396, 24394, 24405, 24402, 24404, 24409, 24406, 24408, 24398, 24387, 24386, 24400, 24389, 24388, 24387, 24390, 24386, 24403, 24398, 24402, 24389, 24394, 24388,
24407, 24400, 24406, 24394, 24393, 24391, 24408, 24403, 24405, 24388, 24391, 24387, 24406, 24399, 24403, 24399, 24388, 24387, 24413, 24414, 24412, 24421, 24418, 24420, 24416, 24411, 24410, 24421, 24413, 24419, 24411, 24412, 24410, 24419, 24416, 24418, 24419, 24411, 24417, 24412, 24397, 24395, 24420, 24407, 24409, 24410, 24395, 24389, 24418, 24401, 24407, 24401, 24410, 24389, 24296, 24370, 24294, 24385, 24319, 24320, 24384, 24310, 24319, 24296, 24383, 24372, 24390, 24357, 24355, 24404, 24367, 24369, 24386, 24355, 24347, 24402, 24361, 24367, 24361, 24386, 24347, 24424, 24423, 24422, 24427, 24428, 24426, 24432, 24431, 24430, 24435, 24436, 24434, 24430, 24425, 24424, 24429, 24434, 24428, 24426, 24424, 24422, 24434, 24432, 24430, 24428, 24430, 24424, 24440, 24439, 24438, 24443, 24444, 24442, 24448, 24447, 24446, 24451, 24452, 24450, 24446, 24441, 24440, 24445, 24450, 24444, 24442, 24440, 24438, 24450, 24448, 24446, 24444, 24446, 24440, 24438, 24433, 24432, 24437, 24442, 24436, 24436, 24438, 24432, 24456, 24455, 24454, 24460, 24459, 24458, 24464, 24463, 24462, 24468, 24467, 24466, 24462, 24457, 24456, 24466, 24461, 24460, 24457, 24458, 24455, 24465, 24466, 24463, 24463, 24460, 24457, 24470, 24473, 24471, 24476, 24475, 24474, 24481, 24478, 24480,
24485, 24482, 24484, 24491, 24488, 24490, 24489, 24478, 24488, 24479, 24472, 24478, 24490, 24486, 24487, 24483, 24476, 24482, 24473, 24474, 24471, 24484, 24479, 24481, 24482, 24473, 24479, 24488, 24472, 24486, 24494, 24493, 24492, 24498, 24497, 24496, 24502, 24501, 24500, 24506, 24505, 24504, 24500, 24495, 24494, 24504, 24499, 24498, 24495, 24496, 24493, 24503, 24504, 24501, 24501, 24498, 24495, 24513, 24510, 24512, 24519, 24516, 24518, 24517, 24512, 24516, 24512, 24508, 24509, 24518, 24514, 24515, 24516, 24509, 24514, 24492, 24465, 24464, 24496, 24469, 24468, 24493, 24468, 24465, 24511, 24490, 24510, 24510, 24487, 24508, 24459, 24472, 24470, 24469, 24486, 24467, 24461, 24486, 24472, 24499, 24508, 24497, 24507, 24514, 24505, 24505, 24509, 24499, 24497, 24487, 24469, 24520, 24523, 24521, 24525, 24526, 24524, 24530, 24529, 24528, 24533, 24534, 24532, 24528, 24523, 24522, 24527, 24532, 24526, 24524, 24522, 24520, 24528, 24534, 24530, 24522, 24532, 24528, 24536, 24539, 24537, 24541, 24542, 24540, 24546, 24545, 24544, 24549, 24550, 24548, 24544, 24539, 24538, 24543, 24548, 24542, 24536, 24542, 24538, 24544, 24550, 24546, 24538, 24548, 24544, 24536, 24531, 24530, 24535, 24540, 24534, 24530, 24540, 24536, 24554, 24553, 24552, 24558, 24557, 24556,
24562, 24561, 24560, 24566, 24565, 24564, 24560, 24555, 24554, 24564, 24559, 24558, 24555, 24556, 24553, 24563, 24564, 24561, 24561, 24558, 24555, 24573, 24570, 24572, 24579, 24576, 24578, 24577, 24572, 24576, 24572, 24568, 24569, 24578, 24574, 24575, 24576, 24569, 24574, 24582, 24581, 24580, 24586, 24585, 24584, 24590, 24589, 24588, 24594, 24593, 24592, 24588, 24583, 24582, 24592, 24587, 24586, 24583, 24584, 24581, 24591, 24592, 24589, 24589, 24586, 24583, 24601, 24598, 24600, 24607, 24604, 24606, 24605, 24600, 24604, 24597, 24598, 24596, 24603, 24604, 24602, 24602, 24600, 24597, 24580, 24563, 24562, 24584, 24567, 24566, 24581, 24566, 24563, 24599, 24578, 24598, 24596, 24578, 24575, 24559, 24568, 24557, 24567, 24574, 24565, 24565, 24569, 24559, 24585, 24597, 24596, 24595, 24602, 24593, 24593, 24597, 24587, 24585, 24575, 24567, 24520, 24449, 24448, 24453, 24524, 24452, 24452, 24520, 24448, 24552, 24503, 24502, 24556, 24507, 24506, 24553, 24506, 24503, 24571, 24518, 24570, 24570, 24515, 24568, 24557, 24515, 24507, 24425, 24454, 24423, 24433, 24462, 24431, 24431, 24456, 24425, 24441, 24492, 24439, 24449, 24500, 24447, 24447, 24494, 24441, 24439, 24464, 24433, 24523, 24552, 24521, 24531, 24560, 24529, 24529, 24554, 24523, 24539, 24580, 24537,
24547, 24588, 24545, 24545, 24582, 24539, 24537, 24562, 24531, 24521, 24502, 24449, 24614, 24615, 24613, 24621, 24620, 24619, 24624, 24625, 24623, 24630, 24627, 24629, 24619, 24610, 24609, 24612, 24623, 24611, 24623, 24618, 24617, 24610, 24613, 24609, 24629, 24620, 24622, 24613, 24612, 24611, 24618, 24619, 24617, 24609, 24611, 24608, 24617, 24609, 24608, 24608, 24623, 24617, 24636, 24637, 24635, 24640, 24641, 24639, 24650, 24647, 24649, 24654, 24651, 24653, 24643, 24632, 24631, 24645, 24634, 24633, 24632, 24635, 24631, 24648, 24643, 24647, 24634, 24639, 24633, 24652, 24645, 24651, 24639, 24638, 24636, 24653, 24648, 24650, 24633, 24636, 24632, 24651, 24644, 24648, 24644, 24633, 24632, 24659, 24658, 24657, 24662, 24663, 24661, 24668, 24665, 24667, 24671, 24674, 24672, 24676, 24677, 24675, 24682, 24679, 24681, 24659, 24672, 24660, 24664, 24675, 24663, 24680, 24667, 24679, 24661, 24656, 24655, 24669, 24677, 24670, 24660, 24665, 24658, 24674, 24679, 24672, 24656, 24657, 24655, 24670, 24671, 24669, 24669, 24659, 24656, 24656, 24675, 24669, 24672, 24667, 24660, 24657, 24622, 24621, 24626, 24661, 24625, 24666, 24629, 24665, 24655, 24621, 24618, 24625, 24655, 24618, 24665, 24622, 24658, 24635, 24616, 24614, 24649, 24627, 24628, 24631, 24614, 24610,
24647, 24620, 24627, 24620, 24631, 24610, 24688, 24689, 24687, 24692, 24693, 24691, 24703, 24700, 24702, 24707, 24704, 24706, 24707, 24708, 24705, 24695, 24684, 24683, 24697, 24686, 24685, 24684, 24687, 24683, 24701, 24695, 24700, 24686, 24691, 24685, 24705, 24697, 24704, 24705, 24699, 24698, 24691, 24690, 24688, 24706, 24701, 24703, 24685, 24688, 24684, 24704, 24696, 24701, 24696, 24685, 24684, 24713, 24714, 24712, 24716, 24719, 24717, 24723, 24720, 24722, 24710, 24717, 24711, 24721, 24713, 24720, 24711, 24712, 24710, 24722, 24717, 24719, 24720, 24711, 24717, 24727, 24728, 24726, 24733, 24734, 24732, 24729, 24732, 24728, 24726, 24725, 24724, 24732, 24731, 24730, 24728, 24730, 24725, 24736, 24739, 24737, 24743, 24740, 24742, 24746, 24745, 24744, 24751, 24748, 24750, 24744, 24739, 24738, 24749, 24742, 24748, 24742, 24737, 24739, 24750, 24745, 24747, 24748, 24739, 24745, 24709, 24726, 24708, 24708, 24724, 24699, 24718, 24737, 24719, 24741, 24722, 24740, 24740, 24719, 24737, 24712, 24694, 24692, 24698, 24718, 24716, 24710, 24692, 24686, 24686, 24716, 24710, 24724, 24738, 24736, 24730, 24746, 24744, 24725, 24744, 24738, 24699, 24736, 24718, 24753, 24756, 24754, 24761, 24763, 24759, 24771, 24768, 24770, 24771, 24772, 24769, 24765, 24756, 24755,
24764, 24768, 24763, 24759, 24755, 24753, 24768, 24766, 24765, 24769, 24767, 24766, 24761, 24757, 24758, 24757, 24753, 24752, 24755, 24768, 24765, 24779, 24776, 24778, 24780, 24783, 24781, 24787, 24788, 24786, 24793, 24787, 24791, 24795, 24796, 24794, 24788, 24778, 24786, 24775, 24776, 24774, 24781, 24786, 24780, 24783, 24787, 24781, 24784, 24795, 24794, 24794, 24793, 24791, 24784, 24791, 24783, 24780, 24778, 24775, 24803, 24802, 24801, 24808, 24805, 24807, 24808, 24809, 24806, 24814, 24811, 24813, 24806, 24798, 24805, 24799, 24809, 24800, 24811, 24804, 24803, 24800, 24801, 24799, 24813, 24809, 24810, 24811, 24800, 24809, 24782, 24799, 24783, 24801, 24785, 24784, 24792, 24805, 24790, 24783, 24801, 24784, 24782, 24805, 24798, 24756, 24774, 24754, 24776, 24762, 24760, 24766, 24782, 24780, 24792, 24772, 24773, 24754, 24776, 24760, 24790, 24767, 24772, 24756, 24780, 24775, 24821, 24822, 24820, 24819, 24821, 24818, 24829, 24828, 24827, 24834, 24835, 24833, 24832, 24834, 24831, 24816, 24828, 24817, 24822, 24837, 24820, 24826, 24833, 24825, 24831, 24826, 24824, 24817, 24820, 24816, 24818, 24817, 24815, 24824, 24829, 24827, 24837, 24832, 24831, 24824, 24837, 24831, 24816, 24837, 24827, 24841, 24844, 24842, 24850, 24847, 24849, 24850, 24845, 24848,
24856, 24853, 24855, 24854, 24847, 24853, 24848, 24843, 24847, 24844, 24845, 24842, 24855, 24851, 24852, 24853, 24843, 24851, 24857, 24860, 24858, 24868, 24865, 24867, 24868, 24863, 24866, 24872, 24869, 24871, 24865, 24860, 24859, 24866, 24858, 24860, 24870, 24861, 24869, 24871, 24863, 24864, 24863, 24861, 24858, 24857, 24826, 24825, 24829, 24862, 24830, 24872, 24839, 24870, 24858, 24829, 24826, 24870, 24830, 24862, 24817, 24841, 24815, 24828, 24852, 24851, 24856, 24839, 24840, 24855, 24830, 24839, 24828, 24843, 24817, 24673, 24754, 24674, 24678, 24757, 24677, 24762, 24681, 24760, 24752, 24673, 24670, 24677, 24752, 24670, 24754, 24681, 24674, 24735, 24818, 24734, 24734, 24815, 24731, 24746, 24842, 24747, 24846, 24750, 24845, 24845, 24747, 24842, 24731, 24841, 24746, 24687, 24642, 24640, 24702, 24652, 24654, 24683, 24640, 24634, 24700, 24646, 24652, 24646, 24683, 24634, 24833, 24797, 24795, 24825, 24795, 24785, 24802, 24859, 24857, 24867, 24812, 24814, 24865, 24804, 24812, 24802, 24825, 24785, 24879, 24876, 24878, 24879, 24880, 24877, 24885, 24886, 24884, 24881, 24884, 24880, 24876, 24874, 24873, 24877, 24875, 24874, 24882, 24886, 24883, 24875, 24884, 24882, 24887, 24889, 24886, 24886, 24888, 24883, 24893, 24892, 24891, 24897, 24896, 24895,
24901, 24900, 24899, 24905, 24904, 24903, 24899, 24894, 24893, 24903, 24898, 24897, 24894, 24895, 24892, 24902, 24903, 24900, 24900, 24897, 24894, 24912, 24909, 24911, 24915, 24914, 24913, 24918, 24919, 24917, 24922, 24918, 24921, 24919, 24911, 24917, 24908, 24909, 24907, 24913, 24918, 24917, 24916, 24918, 24914, 24913, 24911, 24908, 24925, 24924, 24923, 24929, 24928, 24927, 24932, 24933, 24931, 24937, 24936, 24935, 24941, 24940, 24939, 24948, 24945, 24947, 24944, 24945, 24943, 24935, 24926, 24925, 24939, 24930, 24929, 24934, 24943, 24933, 24931, 24925, 24923, 24946, 24937, 24945, 24935, 24945, 24937, 24926, 24927, 24924, 24936, 24941, 24939, 24936, 24929, 24926, 24925, 24943, 24935, 24949, 24952, 24950, 24960, 24957, 24959, 24963, 24962, 24961, 24965, 24968, 24966, 24970, 24971, 24969, 24974, 24970, 24973, 24961, 24952, 24951, 24971, 24959, 24969, 24954, 24957, 24953, 24965, 24970, 24969, 24968, 24970, 24966, 24952, 24953, 24950, 24957, 24956, 24955, 24962, 24967, 24965, 24950, 24957, 24955, 24962, 24954, 24952, 24965, 24959, 24954, 24901, 24924, 24902, 24927, 24906, 24905, 24924, 24905, 24902, 24949, 24916, 24915, 24956, 24921, 24955, 24950, 24921, 24916, 24898, 24907, 24896, 24904, 24915, 24913, 24904, 24908, 24898, 24930, 24949, 24928,
24942, 24961, 24940, 24940, 24951, 24930, 24928, 24915, 24906, 24978, 24977, 24976, 24987, 24984, 24986, 24983, 24984, 24982, 24987, 24989, 24985, 24984, 24979, 24978, 24982, 24978, 24976, 24983, 24980, 24981, 24980, 24976, 24975, 24985, 24988, 24979, 24993, 24992, 24991, 24997, 24996, 24995, 25001, 25000, 24999, 25005, 25004, 25003, 25012, 25009, 25011, 25008, 25009, 25007, 24993, 25000, 24994, 25003, 24998, 24997, 25010, 25001, 25009, 24999, 25009, 25001, 24994, 24995, 24992, 25000, 25005, 25003, 25000, 24997, 24994, 25022, 25019, 25021, 25022, 25023, 25020, 25019, 25015, 25014, 25015, 25023, 25016, 25021, 25017, 25018, 25017, 25014, 25013, 25014, 25006, 25005, 25012, 25017, 25010, 25013, 25005, 25002, 25002, 25017, 25013, 24977, 24993, 24991, 25008, 24989, 24990, 25007, 24988, 24989, 24988, 24993, 24979, 24941, 24977, 24942, 24948, 24980, 24946, 24975, 24941, 24938, 24946, 24975, 24938, 24991, 24964, 24963, 24967, 24996, 24968, 24992, 24967, 24964, 24977, 24963, 24942, 24875, 24891, 24874, 24883, 24899, 24882, 24882, 24893, 24875, 24932, 24889, 24890, 24889, 24923, 24888, 24888, 24901, 24883, 24873, 24547, 24546, 24551, 24876, 24550, 24550, 24873, 24546, 24891, 24591, 24590, 24895, 24595, 24594, 24892, 24594, 24591, 24910, 24606, 24909,
24907, 24606, 24603, 24896, 24603, 24595, 24547, 24891, 24590, 24094, 24300, 24093, 24092, 24300, 24294, 24370, 24128, 24127, 24382, 24141, 24381, 24128, 24374, 24131, 24374, 24141, 24131, 24092, 24370, 24127, 24422, 24228, 24227, 24232, 24426, 24231, 24231, 24422, 24227, 24454, 24268, 24267, 24458, 24272, 24271, 24455, 24271, 24268, 24470, 24284, 24283, 24474, 24288, 24287, 24471, 24287, 24284, 24459, 24283, 24272, 24423, 24267, 24228, 25029, 25028, 25027, 25032, 25033, 25031, 25036, 25037, 25035, 25031, 25026, 25025, 25028, 25035, 25027, 25026, 25027, 25025, 25035, 25032, 25031, 25027, 25031, 25025, 25043, 25042, 25041, 25045, 25048, 25046, 25050, 25051, 25049, 25056, 25053, 25055, 25058, 25059, 25057, 25064, 25061, 25063, 25042, 25049, 25041, 25054, 25043, 25053, 25046, 25057, 25045, 25061, 25048, 25047, 25042, 25047, 25045, 25057, 25052, 25050, 25063, 25054, 25056, 25045, 25050, 25042, 25054, 25047, 25044, 25030, 25041, 25028, 25049, 25038, 25036, 25055, 25039, 25040, 25041, 25036, 25028, 25053, 25030, 25039, 25065, 25068, 25066, 25069, 25072, 25070, 25074, 25075, 25073, 25080, 25077, 25079, 25082, 25083, 25081, 25066, 25073, 25065, 25077, 25068, 25067, 25070, 25081, 25069, 25068, 25069, 25066, 25081, 25076, 25074, 25086, 25078, 25080,
25069, 25074, 25066, 25078, 25071, 25068, 25089, 25088, 25087, 25092, 25093, 25091, 25096, 25092, 25095, 25088, 25091, 25087, 25095, 25088, 25090, 25072, 25087, 25070, 25091, 25084, 25082, 25087, 25082, 25070, 25101, 25100, 25099, 25104, 25105, 25103, 25110, 25107, 25109, 25115, 25114, 25113, 25118, 25119, 25117, 25124, 25121, 25123, 25113, 25102, 25101, 25106, 25117, 25105, 25122, 25109, 25121, 25103, 25098, 25097, 25117, 25112, 25111, 25109, 25100, 25102, 25123, 25114, 25116, 25098, 25099, 25097, 25112, 25113, 25111, 25111, 25101, 25098, 25105, 25111, 25098, 25121, 25102, 25114, 25128, 25129, 25127, 25133, 25132, 25131, 25136, 25137, 25135, 25131, 25126, 25125, 25129, 25135, 25127, 25126, 25127, 25125, 25135, 25133, 25131, 25127, 25131, 25125, 25143, 25142, 25141, 25146, 25147, 25145, 25152, 25149, 25151, 25157, 25156, 25155, 25160, 25161, 25159, 25168, 25165, 25167, 25168, 25163, 25166, 25155, 25144, 25143, 25148, 25159, 25147, 25164, 25151, 25163, 25145, 25140, 25139, 25159, 25154, 25153, 25151, 25142, 25144, 25166, 25157, 25165, 25166, 25156, 25158, 25140, 25141, 25139, 25154, 25155, 25153, 25167, 25161, 25162, 25165, 25154, 25161, 25153, 25143, 25140, 25147, 25153, 25140, 25163, 25144, 25156, 25171, 25170, 25169, 25174, 25175, 25173,
25179, 25178, 25177, 25186, 25183, 25185, 25182, 25183, 25181, 25177, 25172, 25171, 25176, 25181, 25175, 25173, 25171, 25169, 25184, 25179, 25183, 25181, 25179, 25177, 25175, 25177, 25171, 25141, 25116, 25115, 25120, 25145, 25119, 25150, 25123, 25149, 25139, 25115, 25112, 25119, 25139, 25112, 25149, 25116, 25142, 25169, 25134, 25133, 25138, 25173, 25137, 25137, 25169, 25133, 25196, 25193, 25195, 25196, 25197, 25194, 25206, 25203, 25205, 25208, 25209, 25207, 25204, 25191, 25203, 25198, 25207, 25197, 25205, 25199, 25200, 25194, 25188, 25193, 25194, 25190, 25189, 25201, 25209, 25202, 25203, 25187, 25199, 25197, 25201, 25190, 25213, 25212, 25211, 25218, 25215, 25217, 25221, 25220, 25219, 25226, 25223, 25225, 25219, 25214, 25213, 25224, 25217, 25223, 25217, 25212, 25214, 25225, 25220, 25222, 25223, 25214, 25220, 25234, 25231, 25233, 25236, 25237, 25235, 25246, 25243, 25245, 25248, 25249, 25247, 25244, 25233, 25243, 25238, 25247, 25237, 25233, 25227, 25228, 25245, 25239, 25240, 25229, 25237, 25230, 25247, 25242, 25241, 25243, 25228, 25239, 25237, 25241, 25230, 25253, 25252, 25251, 25258, 25255, 25257, 25261, 25260, 25259, 25268, 25265, 25267, 25268, 25263, 25266, 25259, 25254, 25253, 25264, 25257, 25263, 25257, 25252, 25254, 25266, 25261, 25265,
25266, 25260, 25262, 25263, 25254, 25260, 25232, 25205, 25231, 25210, 25235, 25209, 25231, 25200, 25227, 25202, 25235, 25229, 25251, 25222, 25221, 25256, 25225, 25255, 25255, 25222, 25252, 25190, 25211, 25189, 25202, 25219, 25201, 25201, 25213, 25190, 25230, 25251, 25229, 25242, 25259, 25241, 25267, 25249, 25250, 25265, 25242, 25249, 25241, 25253, 25230, 25202, 25251, 25221, 25099, 25030, 25029, 25034, 25103, 25033, 25108, 25039, 25107, 25097, 25029, 25026, 25033, 25097, 25026, 25107, 25030, 25100, 25188, 25072, 25071, 25086, 25193, 25085, 25071, 25193, 25188, 25211, 25090, 25089, 25216, 25095, 25215, 25215, 25090, 25212, 25072, 25211, 25089, 25046, 25067, 25065, 25073, 25060, 25058, 25079, 25062, 25064, 25065, 25058, 25046, 25062, 25067, 25048, 25191, 25130, 25128, 25134, 25199, 25132, 25187, 25128, 25126, 25132, 25187, 25126, 25172, 25227, 25170, 25180, 25239, 25178, 25246, 25184, 25186, 25245, 25180, 25184, 25178, 25228, 25172, 25170, 25200, 25134, 25271, 25270, 25269, 25274, 25275, 25273, 25275, 25277, 25273, 25270, 25273, 25269, 25273, 25271, 25269, 25281, 25280, 25279, 25284, 25285, 25283, 25289, 25288, 25287, 25292, 25293, 25291, 25287, 25282, 25281, 25286, 25291, 25285, 25283, 25281, 25279, 25291, 25289, 25287, 25285, 25287, 25281,
25279, 25272, 25271, 25278, 25283, 25277, 25277, 25279, 25271, 25297, 25296, 25295, 25301, 25300, 25299, 25304, 25305, 25303, 25308, 25309, 25307, 25296, 25303, 25295, 25300, 25307, 25299, 25298, 25299, 25296, 25307, 25306, 25304, 25299, 25304, 25296, 25316, 25313, 25315, 25315, 25311, 25312, 25319, 25318, 25317, 25323, 25322, 25321, 25327, 25326, 25325, 25331, 25330, 25329, 25325, 25320, 25319, 25329, 25324, 25323, 25320, 25321, 25318, 25328, 25329, 25326, 25326, 25323, 25320, 25338, 25335, 25337, 25344, 25341, 25343, 25342, 25337, 25341, 25337, 25333, 25334, 25343, 25339, 25340, 25341, 25334, 25339, 25317, 25298, 25297, 25321, 25302, 25301, 25318, 25301, 25298, 25336, 25315, 25335, 25335, 25312, 25333, 25302, 25311, 25300, 25313, 25310, 25308, 25311, 25308, 25300, 25324, 25333, 25322, 25332, 25339, 25330, 25330, 25334, 25324, 25322, 25312, 25302, 25347, 25346, 25345, 25350, 25351, 25349, 25355, 25354, 25353, 25358, 25359, 25357, 25353, 25348, 25347, 25352, 25357, 25351, 25349, 25347, 25345, 25357, 25355, 25353, 25351, 25353, 25347, 25363, 25362, 25361, 25366, 25367, 25365, 25371, 25370, 25369, 25374, 25375, 25373, 25369, 25364, 25363, 25368, 25373, 25367, 25365, 25363, 25361, 25373, 25371, 25369, 25367, 25369, 25363, 25361, 25356, 25355,
25360, 25365, 25359, 25359, 25361, 25355, 25379, 25378, 25377, 25383, 25382, 25381, 25387, 25386, 25385, 25391, 25390, 25389, 25385, 25380, 25379, 25389, 25384, 25383, 25380, 25381, 25378, 25388, 25389, 25386, 25386, 25383, 25380, 25398, 25395, 25397, 25404, 25401, 25403, 25402, 25397, 25401, 25397, 25393, 25394, 25403, 25399, 25400, 25401, 25394, 25399, 25407, 25406, 25405, 25411, 25410, 25409, 25415, 25414, 25413, 25419, 25418, 25417, 25413, 25408, 25407, 25417, 25412, 25411, 25408, 25409, 25406, 25416, 25417, 25414, 25414, 25411, 25408, 25426, 25423, 25425, 25432, 25429, 25431, 25430, 25425, 25429, 25425, 25421, 25422, 25431, 25427, 25428, 25429, 25422, 25427, 25405, 25388, 25387, 25409, 25392, 25391, 25406, 25391, 25388, 25424, 25403, 25423, 25423, 25400, 25421, 25384, 25393, 25382, 25392, 25399, 25390, 25390, 25394, 25384, 25412, 25421, 25410, 25420, 25427, 25418, 25418, 25422, 25412, 25410, 25400, 25392, 25345, 25290, 25289, 25294, 25349, 25293, 25293, 25345, 25289, 25377, 25328, 25327, 25381, 25332, 25331, 25378, 25331, 25328, 25396, 25343, 25395, 25395, 25340, 25393, 25382, 25340, 25332, 25272, 25295, 25270, 25303, 25276, 25274, 25295, 25274, 25270, 25282, 25317, 25280, 25290, 25325, 25288, 25288, 25319, 25282, 25280, 25297, 25272,
25348, 25377, 25346, 25356, 25385, 25354, 25354, 25379, 25348, 25364, 25405, 25362, 25372, 25413, 25370, 25370, 25407, 25364, 25362, 25387, 25356, 25346, 25327, 25290, 25437, 25436, 25435, 25440, 25441, 25439, 25444, 25445, 25443, 25455, 25452, 25454, 25448, 25438, 25437, 25442, 25450, 25441, 25439, 25434, 25433, 25436, 25443, 25435, 25453, 25448, 25452, 25434, 25435, 25433, 25443, 25440, 25439, 25454, 25450, 25451, 25435, 25439, 25433, 25452, 25447, 25450, 25447, 25437, 25434, 25441, 25447, 25434, 25458, 25457, 25456, 25462, 25461, 25460, 25465, 25466, 25464, 25469, 25470, 25468, 25479, 25476, 25478, 25483, 25480, 25482, 25458, 25473, 25459, 25462, 25475, 25463, 25457, 25464, 25456, 25476, 25473, 25472, 25461, 25468, 25460, 25480, 25475, 25474, 25459, 25460, 25457, 25468, 25467, 25465, 25482, 25477, 25479, 25460, 25465, 25457, 25477, 25474, 25473, 25473, 25462, 25459, 25489, 25490, 25488, 25495, 25492, 25494, 25501, 25502, 25500, 25507, 25504, 25506, 25491, 25500, 25490, 25505, 25494, 25504, 25488, 25485, 25484, 25500, 25497, 25496, 25494, 25486, 25487, 25499, 25504, 25498, 25485, 25486, 25484, 25492, 25489, 25488, 25496, 25499, 25498, 25486, 25488, 25484, 25485, 25498, 25487, 25490, 25496, 25485, 25498, 25494, 25487, 25438, 25456, 25436,
25464, 25446, 25444, 25478, 25453, 25455, 25456, 25444, 25436, 25476, 25449, 25453, 25449, 25458, 25438, 25510, 25509, 25508, 25514, 25513, 25512, 25517, 25518, 25516, 25521, 25522, 25520, 25528, 25527, 25526, 25533, 25530, 25532, 25539, 25536, 25538, 25535, 25536, 25534, 25510, 25525, 25511, 25514, 25527, 25515, 25509, 25516, 25508, 25530, 25525, 25524, 25513, 25520, 25512, 25537, 25528, 25536, 25526, 25536, 25528, 25511, 25512, 25509, 25520, 25519, 25517, 25535, 25531, 25533, 25512, 25517, 25509, 25534, 25525, 25531, 25525, 25514, 25511, 25542, 25541, 25540, 25545, 25546, 25544, 25549, 25545, 25548, 25552, 25551, 25550, 25557, 25554, 25556, 25550, 25543, 25542, 25555, 25548, 25554, 25541, 25544, 25540, 25548, 25541, 25543, 25556, 25551, 25553, 25554, 25543, 25551, 25561, 25562, 25560, 25567, 25568, 25566, 25563, 25566, 25562, 25558, 25562, 25559, 25566, 25565, 25564, 25559, 25566, 25564, 25572, 25571, 25570, 25577, 25574, 25576, 25580, 25579, 25578, 25585, 25582, 25584, 25578, 25573, 25572, 25583, 25576, 25582, 25576, 25571, 25573, 25584, 25579, 25581, 25582, 25573, 25579, 25539, 25560, 25537, 25529, 25560, 25558, 25570, 25553, 25552, 25575, 25556, 25574, 25574, 25553, 25571, 25515, 25540, 25513, 25544, 25523, 25521, 25527, 25552, 25550,
25540, 25521, 25513, 25527, 25542, 25515, 25559, 25570, 25558, 25564, 25580, 25578, 25564, 25572, 25559, 25558, 25552, 25529, 25463, 25508, 25461, 25516, 25471, 25469, 25532, 25481, 25483, 25508, 25469, 25461, 25481, 25524, 25475, 25463, 25524, 25510, 25588, 25587, 25586, 25591, 25592, 25590, 25596, 25595, 25594, 25599, 25600, 25598, 25594, 25589, 25588, 25593, 25598, 25592, 25590, 25588, 25586, 25598, 25596, 25594, 25592, 25594, 25588, 25604, 25603, 25602, 25607, 25608, 25606, 25612, 25611, 25610, 25615, 25616, 25614, 25610, 25605, 25604, 25609, 25614, 25608, 25606, 25604, 25602, 25614, 25612, 25610, 25608, 25610, 25604, 25602, 25597, 25596, 25601, 25606, 25600, 25600, 25602, 25596, 25620, 25619, 25618, 25624, 25623, 25622, 25628, 25627, 25626, 25632, 25631, 25630, 25626, 25621, 25620, 25630, 25625, 25624, 25621, 25622, 25619, 25629, 25630, 25627, 25627, 25624, 25621, 25639, 25636, 25638, 25645, 25642, 25644, 25643, 25638, 25642, 25638, 25634, 25635, 25644, 25640, 25641, 25642, 25635, 25640, 25648, 25647, 25646, 25652, 25651, 25650, 25656, 25655, 25654, 25660, 25659, 25658, 25654, 25649, 25648, 25658, 25653, 25652, 25649, 25650, 25647, 25657, 25658, 25655, 25655, 25652, 25649, 25667, 25664, 25666, 25673, 25670, 25672, 25671, 25666, 25670,
25666, 25662, 25663, 25672, 25668, 25669, 25670, 25663, 25668, 25646, 25629, 25628, 25650, 25633, 25632, 25647, 25632, 25629, 25665, 25644, 25664, 25664, 25641, 25662, 25625, 25634, 25623, 25633, 25640, 25631, 25631, 25635, 25625, 25653, 25662, 25651, 25661, 25668, 25659, 25659, 25663, 25653, 25651, 25641, 25633, 25676, 25675, 25674, 25679, 25680, 25678, 25684, 25683, 25682, 25687, 25688, 25686, 25676, 25683, 25677, 25681, 25686, 25680, 25674, 25680, 25676, 25682, 25688, 25684, 25676, 25686, 25682, 25692, 25691, 25690, 25699, 25696, 25698, 25695, 25696, 25694, 25703, 25704, 25702, 25699, 25702, 25697, 25696, 25693, 25692, 25694, 25692, 25690, 25700, 25704, 25701, 25697, 25700, 25693, 25690, 25685, 25684, 25689, 25694, 25688, 25684, 25694, 25690, 25708, 25707, 25706, 25712, 25711, 25710, 25716, 25715, 25714, 25720, 25719, 25718, 25714, 25709, 25708, 25718, 25713, 25712, 25709, 25710, 25707, 25717, 25718, 25715, 25715, 25712, 25709, 25727, 25724, 25726, 25733, 25730, 25732, 25731, 25726, 25730, 25723, 25724, 25722, 25729, 25730, 25728, 25728, 25726, 25723, 25736, 25735, 25734, 25740, 25739, 25738, 25744, 25743, 25742, 25748, 25747, 25746, 25742, 25737, 25736, 25746, 25741, 25740, 25737, 25738, 25735, 25745, 25746, 25743, 25743, 25740, 25737,
25755, 25752, 25754, 25758, 25757, 25756, 25762, 25763, 25761, 25766, 25762, 25765, 25763, 25754, 25761, 25751, 25752, 25750, 25756, 25762, 25761, 25765, 25757, 25759, 25767, 25766, 25765, 25760, 25765, 25759, 25761, 25751, 25756, 25734, 25717, 25716, 25738, 25721, 25720, 25735, 25720, 25717, 25753, 25732, 25752, 25750, 25732, 25729, 25711, 25723, 25722, 25721, 25728, 25719, 25719, 25723, 25713, 25741, 25750, 25739, 25747, 25758, 25756, 25747, 25751, 25741, 25739, 25729, 25721, 25612, 25675, 25613, 25617, 25678, 25616, 25612, 25678, 25674, 25706, 25657, 25656, 25710, 25661, 25660, 25707, 25660, 25657, 25725, 25672, 25724, 25724, 25669, 25722, 25711, 25669, 25661, 25589, 25618, 25587, 25597, 25626, 25595, 25595, 25620, 25589, 25605, 25646, 25603, 25613, 25654, 25611, 25611, 25648, 25605, 25603, 25628, 25597, 25677, 25706, 25675, 25685, 25714, 25683, 25683, 25708, 25677, 25691, 25736, 25734, 25701, 25742, 25700, 25700, 25736, 25693, 25691, 25716, 25685, 25675, 25656, 25613, 25586, 25372, 25371, 25376, 25590, 25375, 25375, 25586, 25371, 25618, 25416, 25415, 25622, 25420, 25419, 25619, 25419, 25416, 25637, 25431, 25636, 25636, 25428, 25634, 25623, 25428, 25420, 25587, 25415, 25372, 25772, 25775, 25773, 25777, 25778, 25776, 25778, 25780, 25776,
25771, 25777, 25770, 25773, 25776, 25772, 25776, 25774, 25772, 25769, 25777, 25773, 25782, 25785, 25783, 25790, 25791, 25789, 25791, 25793, 25789, 25798, 25799, 25797, 25799, 25795, 25797, 25801, 25804, 25802, 25807, 25806, 25805, 25812, 25809, 25811, 25816, 25813, 25815, 25801, 25785, 25784, 25805, 25788, 25787, 25814, 25798, 25813, 25783, 25789, 25782, 25782, 25793, 25784, 25809, 25804, 25803, 25787, 25798, 25797, 25797, 25786, 25787, 25808, 25813, 25806, 25785, 25786, 25783, 25804, 25805, 25802, 25785, 25805, 25787, 25806, 25798, 25788, 25819, 25822, 25820, 25824, 25825, 25823, 25834, 25831, 25833, 25834, 25828, 25832, 25826, 25837, 25825, 25840, 25831, 25839, 25817, 25825, 25818, 25831, 25822, 25821, 25832, 25820, 25822, 25818, 25819, 25817, 25827, 25824, 25823, 25840, 25837, 25838, 25819, 25823, 25817, 25837, 25836, 25835, 25818, 25836, 25821, 25825, 25835, 25818, 25839, 25821, 25836, 25845, 25844, 25843, 25856, 25853, 25855, 25852, 25853, 25851, 25856, 25858, 25854, 25854, 25845, 25853, 25843, 25853, 25845, 25852, 25848, 25850, 25848, 25843, 25842, 25846, 25858, 25857, 25819, 25775, 25774, 25781, 25827, 25780, 25780, 25819, 25774, 25807, 25844, 25808, 25849, 25809, 25847, 25812, 25848, 25810, 25842, 25807, 25804, 25847, 25803, 25841,
25810, 25842, 25804, 25793, 25771, 25770, 25775, 25801, 25773, 25784, 25770, 25769, 25769, 25801, 25784, 25849, 25828, 25830, 25828, 25841, 25820, 25775, 25841, 25803, 25863, 25864, 25862, 25868, 25871, 25869, 25873, 25874, 25872, 25877, 25878, 25876, 25878, 25880, 25876, 25865, 25877, 25864, 25866, 25873, 25872, 25862, 25861, 25860, 25869, 25876, 25868, 25868, 25880, 25870, 25880, 25875, 25873, 25870, 25873, 25867, 25864, 25869, 25861, 25884, 25883, 25882, 25889, 25886, 25888, 25890, 25893, 25891, 25899, 25896, 25898, 25899, 25894, 25897, 25884, 25891, 25885, 25895, 25888, 25894, 25888, 25883, 25885, 25897, 25892, 25896, 25893, 25894, 25891, 25894, 25885, 25891, 25902, 25901, 25900, 25904, 25907, 25905, 25911, 25908, 25910, 25918, 25915, 25917, 25912, 25903, 25902, 25920, 25908, 25919, 25915, 25913, 25912, 25909, 25906, 25908, 25903, 25904, 25901, 25920, 25916, 25918, 25916, 25914, 25913, 25903, 25914, 25906, 25919, 25906, 25914, 25926, 25923, 25925, 25925, 25921, 25922, 25900, 25867, 25866, 25904, 25871, 25870, 25867, 25904, 25870, 25924, 25896, 25923, 25923, 25892, 25921, 25860, 25884, 25882, 25871, 25890, 25869, 25861, 25890, 25884, 25905, 25922, 25921, 25926, 25909, 25911, 25925, 25907, 25909, 25871, 25921, 25892, 25872, 25816, 25815,
25808, 25872, 25815, 25844, 25902, 25900, 25917, 25858, 25859, 25858, 25912, 25857, 25857, 25902, 25846, 25844, 25866, 25808, 25932, 25931, 25930, 25936, 25935, 25934, 25939, 25940, 25938, 25945, 25948, 25946, 25952, 25949, 25951, 25952, 25953, 25950, 25942, 25933, 25932, 25945, 25937, 25936, 25941, 25949, 25940, 25938, 25932, 25930, 25949, 25943, 25942, 25950, 25944, 25943, 25933, 25934, 25931, 25944, 25945, 25943, 25956, 25953, 25954, 25953, 25947, 25944, 25933, 25945, 25936, 25932, 25949, 25942, 25959, 25958, 25957, 25963, 25962, 25961, 25966, 25967, 25965, 25970, 25966, 25969, 25973, 25972, 25971, 25977, 25976, 25975, 25971, 25960, 25959, 25963, 25976, 25964, 25962, 25965, 25961, 25964, 25966, 25962, 25958, 25963, 25961, 25974, 25975, 25972, 25972, 25963, 25960, 25985, 25982, 25984, 25985, 25986, 25983, 25983, 25979, 25982, 25980, 25986, 25981, 25990, 25989, 25988, 25994, 25993, 25992, 25999, 25996, 25998, 26008, 26005, 26007, 26001, 25995, 25994, 25999, 26003, 25997, 25997, 25990, 25996, 26005, 26002, 26001, 25989, 25994, 25992, 26007, 26003, 26004, 26005, 26000, 26003, 26000, 25994, 25991, 25991, 26003, 26000, 25979, 25948, 25947, 25956, 25982, 25955, 25947, 25982, 25979, 25988, 25974, 25973, 25992, 25978, 25977, 25989, 25977, 25974,
25937, 25957, 25935, 25948, 25971, 25946, 25946, 25959, 25937, 25981, 25988, 25980, 25998, 25986, 25987, 25986, 25990, 25981, 25948, 25988, 25973, 25939, 25928, 25929, 25928, 25930, 25927, 25503, 25790, 25502, 25796, 25506, 25795, 25497, 25786, 25499, 25497, 25790, 25783, 25786, 25506, 25499, 25569, 25862, 25568, 25568, 25860, 25565, 25882, 25581, 25580, 25887, 25584, 25886, 25886, 25581, 25883, 25565, 25882, 25580, 25705, 25928, 25704, 25704, 25927, 25701, 25744, 25931, 25745, 25934, 25749, 25748, 25931, 25748, 25745, 25758, 25958, 25759, 25967, 25767, 25965, 25958, 25760, 25759, 25965, 25760, 25961, 25935, 25758, 25749, 25927, 25744, 25701, 23650, 23881, 23651, 23657, 23884, 23655, 23879, 23650, 23647, 23655, 23879, 23647, 23895, 23674, 23673, 23899, 23678, 23677, 23896, 23677, 23674, 23881, 23673, 23651, 26011, 26010, 26009, 26015, 26014, 26013, 26018, 26019, 26017, 26023, 26022, 26021, 26027, 26026, 26025, 26030, 26031, 26029, 26021, 26012, 26011, 26025, 26016, 26015, 26020, 26029, 26019, 26009, 26019, 26011, 26021, 26031, 26023, 26012, 26013, 26010, 26024, 26025, 26022, 26022, 26015, 26012, 26011, 26029, 26021, 26035, 26034, 26033, 26040, 26037, 26039, 26043, 26042, 26041, 26049, 26046, 26048, 26041, 26036, 26035, 26047, 26039, 26046,
26036, 26037, 26034, 26048, 26042, 26044, 26050, 26049, 26048, 26044, 26050, 26048, 26042, 26039, 26036, 26057, 26056, 26055, 26062, 26059, 26061, 26062, 26063, 26060, 26067, 26070, 26068, 26072, 26073, 26071, 26067, 26058, 26057, 26064, 26071, 26063, 26059, 26053, 26052, 26060, 26054, 26053, 26065, 26073, 26066, 26054, 26055, 26053, 26066, 26067, 26065, 26076, 26073, 26074, 26075, 26066, 26073, 26065, 26057, 26054, 26054, 26071, 26065, 26079, 26078, 26077, 26083, 26082, 26081, 26086, 26087, 26085, 26090, 26086, 26089, 26093, 26092, 26091, 26097, 26096, 26095, 26091, 26080, 26079, 26083, 26096, 26084, 26081, 26086, 26085, 26084, 26086, 26082, 26078, 26083, 26081, 26094, 26095, 26092, 26092, 26083, 26080, 26052, 26024, 26023, 26055, 26028, 26027, 26032, 26059, 26031, 26024, 26055, 26027, 26031, 26052, 26023, 26043, 26078, 26044, 26087, 26050, 26085, 26078, 26045, 26044, 26085, 26045, 26081, 26016, 26033, 26014, 26028, 26041, 26026, 26026, 26035, 26016, 26058, 26077, 26056, 26070, 26091, 26068, 26068, 26079, 26058, 26056, 26043, 26028, 26104, 26103, 26102, 26110, 26111, 26109, 26116, 26117, 26115, 26117, 26113, 26115, 26111, 26100, 26109, 26102, 26110, 26109, 26108, 26115, 26107, 26107, 26113, 26106, 26105, 26106, 26103, 26113, 26112, 26110,
26103, 26113, 26110, 26102, 26100, 26099, 26125, 26128, 26126, 26132, 26133, 26131, 26130, 26132, 26129, 26136, 26137, 26135, 26121, 26136, 26120, 26123, 26132, 26131, 26129, 26124, 26122, 26126, 26135, 26125, 26122, 26127, 26125, 26135, 26130, 26129, 26125, 26129, 26122, 26120, 26126, 26119, 26131, 26118, 26116, 26123, 26116, 26108, 26145, 26142, 26144, 26145, 26146, 26143, 26142, 26140, 26139, 26143, 26141, 26140, 26148, 26151, 26149, 26154, 26153, 26152, 26163, 26162, 26161, 26168, 26165, 26167, 26168, 26169, 26166, 26174, 26171, 26173, 26158, 26151, 26150, 26161, 26155, 26154, 26157, 26165, 26156, 26166, 26158, 26165, 26166, 26160, 26159, 26171, 26164, 26163, 26151, 26152, 26149, 26160, 26161, 26159, 26173, 26169, 26170, 26169, 26163, 26160, 26151, 26161, 26154, 26150, 26165, 26158, 26141, 26148, 26140, 26157, 26146, 26147, 26156, 26141, 26146, 26177, 26176, 26175, 26181, 26180, 26179, 26183, 26186, 26184, 26189, 26188, 26187, 26183, 26178, 26177, 26187, 26182, 26181, 26178, 26179, 26176, 26186, 26187, 26184, 26184, 26181, 26178, 26195, 26194, 26193, 26199, 26198, 26197, 26203, 26202, 26201, 26205, 26208, 26206, 26212, 26209, 26211, 26201, 26196, 26195, 26205, 26200, 26199, 26209, 26208, 26207, 26196, 26197, 26194, 26204, 26205, 26202,
26202, 26199, 26196, 26220, 26217, 26219, 26224, 26221, 26223, 26218, 26213, 26217, 26221, 26216, 26215, 26223, 26218, 26220, 26218, 26215, 26214, 26231, 26228, 26230, 26229, 26225, 26228, 26233, 26229, 26231, 26229, 26227, 26226, 26213, 26186, 26185, 26215, 26190, 26189, 26192, 26217, 26191, 26214, 26189, 26186, 26185, 26217, 26213, 26203, 26226, 26204, 26233, 26209, 26232, 26204, 26227, 26207, 26232, 26207, 26227, 26182, 26193, 26180, 26190, 26201, 26188, 26188, 26195, 26182, 26230, 26222, 26224, 26228, 26216, 26222, 26216, 26203, 26190, 26139, 26070, 26069, 26076, 26142, 26075, 26075, 26139, 26069, 26148, 26094, 26093, 26152, 26098, 26097, 26149, 26097, 26094, 26070, 26148, 26093, 26175, 26105, 26104, 26179, 26108, 26107, 26105, 26179, 26107, 26123, 26194, 26124, 26197, 26128, 26127, 26194, 26127, 26124, 26180, 26123, 26108, 26100, 26090, 26089, 26096, 26104, 26102, 26099, 26089, 26084, 26096, 26099, 26084, 26155, 26175, 26153, 26164, 26183, 26162, 26192, 26172, 26174, 26191, 26164, 26172, 26162, 26177, 26155, 26153, 26104, 26098, 26235, 26238, 26236, 26242, 26243, 26241, 26249, 26248, 26247, 26252, 26253, 26251, 26247, 26238, 26237, 26240, 26251, 26239, 26251, 26246, 26245, 26235, 26243, 26237, 26243, 26240, 26239, 26245, 26249, 26247,
26237, 26239, 26234, 26234, 26247, 26237, 26234, 26251, 26245, 26257, 26256, 26255, 26259, 26262, 26260, 26266, 26263, 26265, 26269, 26268, 26267, 26277, 26274, 26276, 26279, 26274, 26278, 26267, 26258, 26257, 26261, 26272, 26262, 26277, 26265, 26275, 26265, 26260, 26262, 26275, 26271, 26274, 26273, 26274, 26271, 26258, 26259, 26256, 26268, 26273, 26271, 26268, 26261, 26258, 26272, 26265, 26262, 26280, 26283, 26281, 26286, 26285, 26284, 26289, 26290, 26288, 26290, 26292, 26288, 26296, 26295, 26294, 26300, 26299, 26298, 26294, 26283, 26282, 26298, 26287, 26286, 26281, 26288, 26280, 26280, 26292, 26282, 26283, 26284, 26281, 26297, 26298, 26295, 26295, 26286, 26283, 26304, 26303, 26302, 26313, 26310, 26312, 26314, 26317, 26315, 26323, 26320, 26322, 26323, 26318, 26321, 26304, 26315, 26305, 26319, 26308, 26318, 26321, 26316, 26320, 26321, 26315, 26317, 26307, 26310, 26306, 26303, 26307, 26306, 26313, 26308, 26309, 26312, 26305, 26308, 26318, 26305, 26315, 26284, 26250, 26249, 26254, 26289, 26253, 26246, 26284, 26249, 26253, 26281, 26246, 26302, 26270, 26269, 26311, 26278, 26310, 26270, 26306, 26273, 26310, 26273, 26306, 26238, 26255, 26236, 26250, 26267, 26248, 26248, 26257, 26238, 26287, 26302, 26285, 26301, 26314, 26299, 26299, 26304, 26287,
26285, 26269, 26250, 26326, 26325, 26324, 26328, 26331, 26329, 26337, 26334, 26336, 26337, 26332, 26335, 26338, 26341, 26339, 26348, 26345, 26347, 26348, 26343, 26346, 26338, 26327, 26326, 26350, 26334, 26349, 26346, 26340, 26345, 26346, 26339, 26341, 26334, 26331, 26330, 26331, 26332, 26329, 26327, 26328, 26325, 26350, 26343, 26344, 26343, 26342, 26339, 26327, 26342, 26330, 26349, 26330, 26342, 26353, 26332, 26333, 26352, 26329, 26332, 26324, 26297, 26296, 26328, 26301, 26300, 26325, 26300, 26297, 26353, 26320, 26352, 26351, 26320, 26316, 26301, 26351, 26316, 26292, 26121, 26120, 26128, 26294, 26126, 26119, 26292, 26120, 26119, 26294, 26282, 26200, 26324, 26198, 26206, 26340, 26338, 26347, 26210, 26212, 26345, 26208, 26210, 26206, 26326, 26200, 26198, 26296, 26128, 26356, 26355, 26354, 26360, 26359, 26358, 26363, 26364, 26362, 26367, 26368, 26366, 26372, 26371, 26370, 26376, 26375, 26374, 26370, 26357, 26356, 26374, 26361, 26360, 26355, 26362, 26354, 26359, 26366, 26358, 26357, 26358, 26355, 26366, 26365, 26363, 26373, 26374, 26371, 26358, 26363, 26355, 26371, 26360, 26357, 26380, 26379, 26378, 26384, 26383, 26382, 26387, 26388, 26386, 26391, 26392, 26390, 26396, 26395, 26394, 26400, 26399, 26398, 26394, 26381, 26380, 26398, 26385, 26384,
26379, 26386, 26378, 26383, 26390, 26382, 26381, 26382, 26379, 26390, 26389, 26387, 26397, 26398, 26395, 26382, 26387, 26379, 26395, 26384, 26381, 26404, 26403, 26402, 26408, 26407, 26406, 26413, 26410, 26412, 26417, 26414, 26416, 26411, 26404, 26410, 26415, 26408, 26414, 26405, 26406, 26403, 26416, 26411, 26413, 26414, 26405, 26411, 26420, 26419, 26418, 26424, 26423, 26422, 26430, 26429, 26428, 26434, 26433, 26432, 26437, 26438, 26436, 26428, 26421, 26420, 26432, 26425, 26424, 26427, 26436, 26426, 26436, 26430, 26428, 26421, 26422, 26419, 26431, 26432, 26429, 26429, 26424, 26421, 26426, 26428, 26420, 26402, 26373, 26372, 26406, 26377, 26376, 26403, 26376, 26373, 26418, 26397, 26396, 26422, 26401, 26400, 26419, 26400, 26397, 26361, 26378, 26359, 26386, 26369, 26367, 26377, 26394, 26375, 26378, 26367, 26359, 26375, 26380, 26361, 26409, 26418, 26407, 26427, 26415, 26417, 26426, 26409, 26415, 26407, 26396, 26377, 26442, 26441, 26440, 26446, 26445, 26444, 26449, 26450, 26448, 26453, 26454, 26452, 26458, 26457, 26456, 26462, 26461, 26460, 26456, 26443, 26442, 26460, 26447, 26446, 26441, 26448, 26440, 26445, 26452, 26444, 26443, 26444, 26441, 26452, 26451, 26449, 26459, 26460, 26457, 26444, 26449, 26441, 26457, 26446, 26443, 26466, 26465, 26464,
26470, 26469, 26468, 26473, 26474, 26472, 26477, 26478, 26476, 26482, 26481, 26480, 26486, 26485, 26484, 26480, 26467, 26466, 26484, 26471, 26470, 26465, 26472, 26464, 26469, 26476, 26468, 26467, 26468, 26465, 26476, 26475, 26473, 26483, 26484, 26481, 26468, 26473, 26465, 26481, 26470, 26467, 26490, 26489, 26488, 26492, 26495, 26493, 26499, 26496, 26498, 26502, 26501, 26500, 26509, 26506, 26508, 26500, 26491, 26490, 26507, 26496, 26506, 26497, 26494, 26496, 26508, 26504, 26505, 26489, 26494, 26492, 26503, 26504, 26501, 26491, 26504, 26494, 26506, 26494, 26504, 26512, 26511, 26510, 26516, 26515, 26514, 26521, 26518, 26520, 26525, 26522, 26524, 26519, 26512, 26518, 26523, 26516, 26522, 26513, 26514, 26511, 26524, 26519, 26521, 26522, 26513, 26519, 26488, 26459, 26458, 26492, 26463, 26462, 26489, 26462, 26459, 26510, 26483, 26482, 26514, 26487, 26486, 26511, 26486, 26483, 26447, 26464, 26445, 26472, 26455, 26453, 26463, 26480, 26461, 26464, 26453, 26445, 26461, 26466, 26447, 26495, 26510, 26493, 26520, 26497, 26499, 26518, 26495, 26497, 26493, 26482, 26463, 26528, 26527, 26526, 26532, 26531, 26530, 26535, 26536, 26534, 26540, 26539, 26538, 26544, 26543, 26542, 26547, 26548, 26546, 26538, 26529, 26528, 26542, 26533, 26532, 26537, 26546, 26536,
26534, 26528, 26526, 26546, 26540, 26538, 26529, 26530, 26527, 26541, 26542, 26539, 26539, 26532, 26529, 26536, 26538, 26528, 26552, 26551, 26550, 26556, 26555, 26554, 26559, 26560, 26558, 26564, 26563, 26562, 26568, 26567, 26566, 26571, 26572, 26570, 26562, 26553, 26552, 26566, 26557, 26556, 26561, 26570, 26560, 26558, 26552, 26550, 26570, 26564, 26562, 26553, 26554, 26551, 26565, 26566, 26563, 26563, 26556, 26553, 26560, 26562, 26552, 26550, 26541, 26540, 26554, 26545, 26544, 26549, 26558, 26548, 26551, 26544, 26541, 26548, 26550, 26540, 26576, 26575, 26574, 26583, 26580, 26582, 26586, 26585, 26584, 26593, 26590, 26592, 26584, 26577, 26576, 26591, 26582, 26590, 26582, 26578, 26579, 26592, 26588, 26589, 26577, 26578, 26575, 26587, 26588, 26585, 26585, 26579, 26577, 26590, 26579, 26588, 26596, 26595, 26594, 26603, 26600, 26602, 26606, 26605, 26604, 26613, 26610, 26612, 26604, 26597, 26596, 26611, 26602, 26610, 26602, 26598, 26599, 26612, 26608, 26609, 26597, 26598, 26595, 26607, 26608, 26605, 26605, 26599, 26597, 26610, 26599, 26608, 26594, 26587, 26586, 26601, 26592, 26600, 26595, 26589, 26587, 26600, 26589, 26598, 26526, 26431, 26430, 26530, 26435, 26434, 26439, 26534, 26438, 26527, 26434, 26431, 26438, 26526, 26430, 26574, 26503, 26502,
26581, 26508, 26580, 26575, 26505, 26503, 26580, 26505, 26578, 26385, 26440, 26383, 26448, 26393, 26391, 26401, 26456, 26399, 26440, 26391, 26383, 26399, 26442, 26385, 26425, 26488, 26423, 26435, 26500, 26433, 26433, 26490, 26425, 26423, 26458, 26401, 26533, 26574, 26531, 26545, 26584, 26543, 26543, 26576, 26533, 26557, 26594, 26555, 26569, 26604, 26567, 26567, 26596, 26557, 26555, 26586, 26545, 26531, 26502, 26435, 26616, 26615, 26614, 26620, 26619, 26618, 26623, 26624, 26622, 26627, 26628, 26626, 26632, 26631, 26630, 26636, 26635, 26634, 26630, 26617, 26616, 26634, 26621, 26620, 26615, 26622, 26614, 26619, 26626, 26618, 26617, 26618, 26615, 26626, 26625, 26623, 26633, 26634, 26631, 26618, 26623, 26615, 26631, 26620, 26617, 26640, 26639, 26638, 26644, 26643, 26642, 26647, 26648, 26646, 26651, 26652, 26650, 26655, 26651, 26654, 26658, 26657, 26656, 26662, 26661, 26660, 26667, 26664, 26666, 26656, 26641, 26640, 26660, 26645, 26644, 26665, 26654, 26664, 26639, 26646, 26638, 26643, 26650, 26642, 26654, 26643, 26645, 26666, 26661, 26663, 26641, 26642, 26639, 26650, 26649, 26647, 26659, 26660, 26657, 26642, 26647, 26639, 26657, 26644, 26641, 26664, 26645, 26661, 26670, 26669, 26668, 26672, 26675, 26673, 26679, 26676, 26678, 26684, 26683, 26682,
26687, 26688, 26686, 26682, 26675, 26674, 26681, 26686, 26680, 26677, 26670, 26676, 26686, 26684, 26682, 26669, 26674, 26672, 26681, 26677, 26679, 26680, 26671, 26677, 26680, 26682, 26674, 26692, 26691, 26690, 26696, 26695, 26694, 26701, 26698, 26700, 26704, 26703, 26702, 26708, 26707, 26706, 26713, 26710, 26712, 26702, 26693, 26692, 26706, 26697, 26696, 26711, 26700, 26710, 26700, 26695, 26697, 26712, 26707, 26709, 26693, 26694, 26691, 26705, 26706, 26703, 26703, 26696, 26693, 26710, 26697, 26707, 26668, 26633, 26632, 26672, 26637, 26636, 26669, 26636, 26633, 26690, 26659, 26658, 26694, 26663, 26662, 26699, 26666, 26698, 26691, 26662, 26659, 26698, 26663, 26695, 26621, 26638, 26619, 26646, 26629, 26627, 26637, 26656, 26635, 26638, 26627, 26619, 26635, 26640, 26621, 26675, 26690, 26673, 26685, 26702, 26683, 26683, 26692, 26675, 26673, 26658, 26637, 26716, 26715, 26714, 26719, 26720, 26718, 26724, 26723, 26722, 26727, 26728, 26726, 26722, 26717, 26716, 26721, 26726, 26720, 26718, 26716, 26714, 26726, 26724, 26722, 26720, 26722, 26716, 26732, 26731, 26730, 26736, 26735, 26734, 26741, 26738, 26740, 26744, 26743, 26742, 26748, 26747, 26746, 26753, 26750, 26752, 26742, 26733, 26732, 26746, 26737, 26736, 26751, 26740, 26750, 26740, 26735, 26737,
26752, 26747, 26749, 26733, 26734, 26731, 26745, 26746, 26743, 26743, 26736, 26733, 26750, 26737, 26747, 26756, 26755, 26754, 26759, 26760, 26758, 26764, 26763, 26762, 26767, 26768, 26766, 26762, 26757, 26756, 26761, 26766, 26760, 26758, 26756, 26754, 26766, 26764, 26762, 26760, 26762, 26756, 26772, 26771, 26770, 26776, 26775, 26774, 26781, 26778, 26780, 26784, 26783, 26782, 26788, 26787, 26786, 26793, 26790, 26792, 26782, 26773, 26772, 26786, 26777, 26776, 26791, 26780, 26790, 26780, 26775, 26777, 26792, 26787, 26789, 26773, 26774, 26771, 26785, 26786, 26783, 26783, 26776, 26773, 26790, 26777, 26787, 26754, 26725, 26724, 26729, 26758, 26728, 26728, 26754, 26724, 26770, 26745, 26744, 26774, 26749, 26748, 26779, 26752, 26778, 26771, 26748, 26745, 26778, 26749, 26775, 26717, 26730, 26715, 26725, 26742, 26723, 26723, 26732, 26717, 26757, 26770, 26755, 26765, 26782, 26763, 26763, 26772, 26757, 26755, 26744, 26725, 26714, 26685, 26684, 26689, 26718, 26688, 26688, 26714, 26684, 26730, 26705, 26704, 26734, 26709, 26708, 26739, 26712, 26738, 26731, 26708, 26705, 26738, 26709, 26735, 26715, 26704, 26685, 26796, 26795, 26794, 26800, 26799, 26798, 26803, 26804, 26802, 26808, 26807, 26806, 26812, 26811, 26810, 26815, 26816, 26814, 26806, 26797, 26796,
26810, 26801, 26800, 26805, 26814, 26804, 26802, 26796, 26794, 26814, 26808, 26806, 26797, 26798, 26795, 26809, 26810, 26807, 26807, 26800, 26797, 26804, 26806, 26796, 26820, 26819, 26818, 26824, 26823, 26822, 26827, 26828, 26826, 26832, 26831, 26830, 26836, 26835, 26834, 26839, 26840, 26838, 26830, 26821, 26820, 26834, 26825, 26824, 26829, 26838, 26828, 26826, 26820, 26818, 26838, 26832, 26830, 26821, 26822, 26819, 26833, 26834, 26831, 26831, 26824, 26821, 26828, 26830, 26820, 26818, 26809, 26808, 26822, 26813, 26812, 26817, 26826, 26816, 26819, 26812, 26809, 26816, 26818, 26808, 26844, 26843, 26842, 26851, 26848, 26850, 26854, 26853, 26852, 26861, 26858, 26860, 26852, 26845, 26844, 26859, 26850, 26858, 26850, 26846, 26847, 26860, 26856, 26857, 26845, 26846, 26843, 26855, 26856, 26853, 26853, 26847, 26845, 26858, 26847, 26856, 26864, 26863, 26862, 26871, 26868, 26870, 26874, 26873, 26872, 26881, 26878, 26880, 26872, 26865, 26864, 26879, 26870, 26878, 26870, 26866, 26867, 26880, 26876, 26877, 26865, 26866, 26863, 26875, 26876, 26873, 26873, 26867, 26865, 26878, 26867, 26876, 26862, 26855, 26854, 26869, 26860, 26868, 26863, 26857, 26855, 26868, 26857, 26866, 26884, 26883, 26882, 26888, 26887, 26886, 26891, 26892, 26890, 26901, 26898, 26900,
26905, 26902, 26904, 26894, 26885, 26884, 26896, 26889, 26888, 26893, 26898, 26892, 26890, 26884, 26882, 26899, 26894, 26898, 26903, 26896, 26902, 26885, 26886, 26883, 26904, 26899, 26901, 26902, 26895, 26899, 26895, 26888, 26885, 26892, 26894, 26884, 26908, 26907, 26906, 26915, 26912, 26914, 26922, 26919, 26921, 26916, 26909, 26908, 26924, 26914, 26923, 26920, 26916, 26919, 26914, 26910, 26911, 26909, 26910, 26907, 26924, 26920, 26922, 26923, 26917, 26920, 26917, 26911, 26909, 26923, 26911, 26918, 26882, 26833, 26832, 26886, 26837, 26836, 26841, 26890, 26840, 26883, 26836, 26833, 26840, 26882, 26832, 26906, 26875, 26874, 26913, 26880, 26912, 26907, 26877, 26875, 26912, 26877, 26910, 26801, 26842, 26799, 26813, 26852, 26811, 26811, 26844, 26801, 26825, 26862, 26823, 26837, 26872, 26835, 26835, 26864, 26825, 26823, 26854, 26813, 26889, 26906, 26887, 26921, 26903, 26905, 26919, 26897, 26903, 26897, 26908, 26889, 26887, 26874, 26837, 26927, 26926, 26925, 26930, 26931, 26929, 26935, 26934, 26933, 26938, 26939, 26937, 26933, 26928, 26927, 26932, 26937, 26931, 26929, 26927, 26925, 26937, 26935, 26933, 26931, 26933, 26927, 26943, 26942, 26941, 26947, 26946, 26945, 26952, 26949, 26951, 26955, 26954, 26953, 26959, 26958, 26957, 26964, 26961, 26963,
26953, 26944, 26943, 26957, 26948, 26947, 26962, 26951, 26961, 26951, 26946, 26948, 26963, 26958, 26960, 26944, 26945, 26942, 26956, 26957, 26954, 26954, 26947, 26944, 26961, 26948, 26958, 26967, 26966, 26965, 26970, 26971, 26969, 26975, 26974, 26973, 26978, 26979, 26977, 26973, 26968, 26967, 26972, 26977, 26971, 26969, 26967, 26965, 26977, 26975, 26973, 26971, 26973, 26967, 26983, 26982, 26981, 26987, 26986, 26985, 26992, 26989, 26991, 26995, 26994, 26993, 26999, 26998, 26997, 27004, 27001, 27003, 26993, 26984, 26983, 26997, 26988, 26987, 27002, 26991, 27001, 26991, 26986, 26988, 27003, 26998, 27000, 26984, 26985, 26982, 26996, 26997, 26994, 26994, 26987, 26984, 27001, 26988, 26998, 26965, 26936, 26935, 26940, 26969, 26939, 26939, 26965, 26935, 26981, 26956, 26955, 26985, 26960, 26959, 26990, 26963, 26989, 26982, 26959, 26956, 26989, 26960, 26986, 26928, 26941, 26926, 26936, 26953, 26934, 26934, 26943, 26928, 26968, 26981, 26966, 26976, 26993, 26974, 26974, 26983, 26968, 26966, 26955, 26936, 27007, 27006, 27005, 27010, 27011, 27009, 27015, 27014, 27013, 27018, 27019, 27017, 27013, 27008, 27007, 27012, 27017, 27011, 27009, 27007, 27005, 27017, 27015, 27013, 27011, 27013, 27007, 27023, 27022, 27021, 27027, 27026, 27025, 27032, 27029, 27031,
27035, 27034, 27033, 27039, 27038, 27037, 27044, 27041, 27043, 27033, 27024, 27023, 27037, 27028, 27027, 27042, 27031, 27041, 27031, 27026, 27028, 27043, 27038, 27040, 27024, 27025, 27022, 27036, 27037, 27034, 27034, 27027, 27024, 27041, 27028, 27038, 27050, 27047, 27049, 27048, 27045, 27047, 27058, 27055, 27057, 27062, 27059, 27061, 27056, 27051, 27055, 27060, 27053, 27059, 27061, 27056, 27058, 27059, 27052, 27056, 27045, 27016, 27015, 27020, 27047, 27019, 27019, 27045, 27015, 27051, 27036, 27035, 27053, 27040, 27039, 27062, 27043, 27060, 27052, 27039, 27036, 27060, 27040, 27054, 27008, 27021, 27006, 27016, 27033, 27014, 27014, 27023, 27008, 27057, 27048, 27050, 27055, 27046, 27048, 27046, 27035, 27016, 27005, 26976, 26975, 26980, 27009, 26979, 26979, 27005, 26975, 27021, 26996, 26995, 27025, 27000, 26999, 27030, 27003, 27029, 27022, 26999, 26996, 27029, 27000, 27026, 27006, 26995, 26976, 26794, 26565, 26564, 26798, 26569, 26568, 26573, 26802, 26572, 26795, 26568, 26565, 26572, 26794, 26564, 26842, 26607, 26606, 26849, 26612, 26848, 26843, 26609, 26607, 26848, 26609, 26846, 26799, 26606, 26569, 26925, 26765, 26764, 26769, 26929, 26768, 26768, 26925, 26764, 26941, 26785, 26784, 26945, 26789, 26788, 26950, 26792, 26949, 26942, 26788, 26785,
26949, 26789, 26946, 26926, 26784, 26765, 26471, 26614, 26469, 26622, 26479, 26477, 26487, 26630, 26485, 26614, 26477, 26469, 26485, 26616, 26471, 26517, 26668, 26515, 26678, 26523, 26525, 26676, 26517, 26523, 26515, 26632, 26487, 27065, 27064, 27063, 27069, 27068, 27067, 27072, 27073, 27071, 27076, 27077, 27075, 27081, 27080, 27079, 27085, 27084, 27083, 27079, 27066, 27065, 27083, 27070, 27069, 27064, 27071, 27063, 27068, 27075, 27067, 27066, 27067, 27064, 27075, 27074, 27072, 27082, 27083, 27080, 27067, 27072, 27064, 27080, 27069, 27066, 27089, 27088, 27087, 27093, 27092, 27091, 27096, 27097, 27095, 27100, 27101, 27099, 27105, 27104, 27103, 27109, 27108, 27107, 27103, 27090, 27089, 27107, 27094, 27093, 27088, 27095, 27087, 27092, 27099, 27091, 27090, 27091, 27088, 27099, 27098, 27096, 27106, 27107, 27104, 27091, 27096, 27088, 27104, 27093, 27090, 27113, 27112, 27111, 27117, 27116, 27115, 27122, 27119, 27121, 27126, 27123, 27125, 27120, 27113, 27119, 27124, 27117, 27123, 27114, 27115, 27112, 27125, 27120, 27122, 27123, 27114, 27120, 27129, 27128, 27127, 27133, 27132, 27131, 27138, 27135, 27137, 27142, 27139, 27141, 27136, 27129, 27135, 27140, 27133, 27139, 27130, 27131, 27128, 27141, 27136, 27138, 27139, 27130, 27136, 27111, 27082, 27081,
27115, 27086, 27085, 27112, 27085, 27082, 27127, 27106, 27105, 27131, 27110, 27109, 27128, 27109, 27106, 27070, 27087, 27068, 27095, 27078, 27076, 27086, 27103, 27084, 27087, 27076, 27068, 27084, 27089, 27070, 27118, 27127, 27116, 27137, 27124, 27126, 27135, 27118, 27124, 27116, 27105, 27086, 27145, 27144, 27143, 27149, 27148, 27147, 27152, 27153, 27151, 27156, 27157, 27155, 27161, 27160, 27159, 27165, 27164, 27163, 27159, 27146, 27145, 27163, 27150, 27149, 27144, 27151, 27143, 27148, 27155, 27147, 27146, 27147, 27144, 27155, 27154, 27152, 27162, 27163, 27160, 27147, 27152, 27144, 27160, 27149, 27146, 27169, 27168, 27167, 27173, 27172, 27171, 27176, 27177, 27175, 27180, 27181, 27179, 27185, 27184, 27183, 27189, 27188, 27187, 27183, 27170, 27169, 27187, 27174, 27173, 27168, 27175, 27167, 27172, 27179, 27171, 27170, 27171, 27168, 27179, 27178, 27176, 27186, 27187, 27184, 27171, 27176, 27168, 27184, 27173, 27170, 27193, 27192, 27191, 27197, 27196, 27195, 27202, 27199, 27201, 27206, 27203, 27205, 27200, 27193, 27199, 27204, 27197, 27203, 27194, 27195, 27192, 27205, 27200, 27202, 27203, 27194, 27200, 27209, 27208, 27207, 27213, 27212, 27211, 27218, 27215, 27217, 27222, 27219, 27221, 27216, 27209, 27215, 27220, 27213, 27219, 27210, 27211, 27208,
27221, 27216, 27218, 27219, 27210, 27216, 27191, 27162, 27161, 27195, 27166, 27165, 27192, 27165, 27162, 27207, 27186, 27185, 27211, 27190, 27189, 27208, 27189, 27186, 27150, 27167, 27148, 27175, 27158, 27156, 27166, 27183, 27164, 27167, 27156, 27148, 27164, 27169, 27150, 27198, 27207, 27196, 27217, 27204, 27206, 27215, 27198, 27204, 27196, 27185, 27166, 27094, 27143, 27092, 27151, 27102, 27100, 27110, 27159, 27108, 27143, 27100, 27092, 27108, 27145, 27094, 27134, 27191, 27132, 27201, 27140, 27142, 27199, 27134, 27140, 27132, 27161, 27110, 27225, 27224, 27223, 27229, 27228, 27227, 27232, 27233, 27231, 27236, 27237, 27235, 27241, 27240, 27239, 27245, 27244, 27243, 27239, 27226, 27225, 27243, 27230, 27229, 27224, 27231, 27223, 27228, 27235, 27227, 27226, 27227, 27224, 27235, 27234, 27232, 27242, 27243, 27240, 27227, 27232, 27224, 27240, 27229, 27226, 27249, 27248, 27247, 27253, 27252, 27251, 27256, 27257, 27255, 27260, 27261, 27259, 27264, 27260, 27263, 27267, 27266, 27265, 27271, 27270, 27269, 27276, 27273, 27275, 27265, 27250, 27249, 27269, 27254, 27253, 27274, 27263, 27273, 27248, 27255, 27247, 27252, 27259, 27251, 27263, 27252, 27254, 27275, 27270, 27272, 27250, 27251, 27248, 27259, 27258, 27256, 27268, 27269, 27266, 27251, 27256, 27248,
27266, 27253, 27250, 27273, 27254, 27270, 27279, 27278, 27277, 27283, 27282, 27281, 27288, 27285, 27287, 27292, 27289, 27291, 27286, 27279, 27285, 27290, 27283, 27289, 27280, 27281, 27278, 27291, 27286, 27288, 27289, 27280, 27286, 27295, 27294, 27293, 27299, 27298, 27297, 27304, 27301, 27303, 27310, 27307, 27309, 27310, 27305, 27308, 27302, 27295, 27301, 27308, 27299, 27307, 27308, 27298, 27300, 27296, 27297, 27294, 27309, 27302, 27304, 27307, 27296, 27302, 27277, 27242, 27241, 27281, 27246, 27245, 27278, 27245, 27242, 27293, 27268, 27267, 27297, 27272, 27271, 27306, 27275, 27305, 27294, 27271, 27268, 27305, 27272, 27298, 27230, 27247, 27228, 27255, 27238, 27236, 27246, 27265, 27244, 27247, 27236, 27228, 27244, 27249, 27230, 27284, 27293, 27282, 27303, 27290, 27292, 27301, 27284, 27290, 27282, 27267, 27246, 27174, 27223, 27172, 27231, 27182, 27180, 27190, 27239, 27188, 27223, 27180, 27172, 27188, 27225, 27174, 27214, 27277, 27212, 27287, 27220, 27222, 27285, 27214, 27220, 27212, 27241, 27190, 27316, 27317, 27315, 27320, 27321, 27319, 27325, 27324, 27323, 27329, 27328, 27327, 27311, 27324, 27312, 27313, 27328, 27314, 27312, 27315, 27311, 27314, 27319, 27313, 27319, 27318, 27316, 27326, 27327, 27324, 27313, 27316, 27312, 27312, 27327, 27313,
27336, 27337, 27335, 27340, 27341, 27339, 27345, 27344, 27343, 27349, 27348, 27347, 27343, 27332, 27331, 27347, 27334, 27333, 27332, 27335, 27331, 27334, 27339, 27333, 27339, 27338, 27336, 27346, 27347, 27344, 27333, 27336, 27332, 27344, 27333, 27332, 27353, 27352, 27351, 27357, 27356, 27355, 27359, 27362, 27360, 27369, 27366, 27368, 27373, 27370, 27372, 27375, 27370, 27374, 27359, 27354, 27353, 27357, 27364, 27358, 27367, 27361, 27366, 27370, 27364, 27363, 27374, 27363, 27365, 27354, 27355, 27352, 27360, 27365, 27363, 27375, 27367, 27369, 27374, 27362, 27367, 27360, 27357, 27354, 27378, 27377, 27376, 27382, 27381, 27380, 27391, 27388, 27390, 27395, 27392, 27394, 27384, 27379, 27378, 27386, 27383, 27382, 27389, 27384, 27388, 27393, 27386, 27392, 27379, 27380, 27377, 27394, 27389, 27391, 27392, 27385, 27389, 27385, 27382, 27379, 27351, 27326, 27325, 27355, 27330, 27329, 27352, 27329, 27326, 27376, 27346, 27345, 27380, 27350, 27349, 27377, 27349, 27346, 27335, 27322, 27320, 27330, 27343, 27328, 27331, 27320, 27314, 27314, 27343, 27331, 27358, 27376, 27356, 27390, 27371, 27373, 27388, 27364, 27371, 27358, 27384, 27378, 27356, 27345, 27330, 27401, 27402, 27400, 27405, 27406, 27404, 27410, 27409, 27408, 27414, 27413, 27412, 27408, 27397, 27396,
27412, 27399, 27398, 27397, 27400, 27396, 27399, 27404, 27398, 27404, 27403, 27401, 27411, 27412, 27409, 27398, 27401, 27397, 27409, 27398, 27397, 27421, 27422, 27420, 27425, 27426, 27424, 27430, 27429, 27428, 27434, 27433, 27432, 27428, 27417, 27416, 27432, 27419, 27418, 27417, 27420, 27416, 27419, 27424, 27418, 27424, 27423, 27421, 27431, 27432, 27429, 27418, 27421, 27417, 27429, 27418, 27417, 27438, 27437, 27436, 27442, 27441, 27440, 27451, 27448, 27450, 27455, 27452, 27454, 27444, 27439, 27438, 27446, 27443, 27442, 27449, 27444, 27448, 27453, 27446, 27452, 27439, 27440, 27437, 27454, 27449, 27451, 27452, 27445, 27449, 27445, 27442, 27439, 27458, 27457, 27456, 27462, 27461, 27460, 27471, 27468, 27470, 27475, 27472, 27474, 27464, 27459, 27458, 27466, 27463, 27462, 27469, 27464, 27468, 27473, 27466, 27472, 27459, 27460, 27457, 27474, 27469, 27471, 27472, 27465, 27469, 27465, 27462, 27459, 27436, 27411, 27410, 27440, 27415, 27414, 27437, 27414, 27411, 27456, 27431, 27430, 27460, 27435, 27434, 27457, 27434, 27431, 27420, 27407, 27405, 27415, 27428, 27413, 27416, 27405, 27399, 27413, 27416, 27399, 27443, 27456, 27441, 27470, 27453, 27455, 27468, 27447, 27453, 27447, 27458, 27443, 27441, 27430, 27415, 27478, 27366, 27477, 27477, 27361, 27476,
27400, 27342, 27340, 27350, 27408, 27348, 27396, 27340, 27334, 27348, 27396, 27334, 27383, 27436, 27381, 27450, 27393, 27395, 27448, 27387, 27393, 27383, 27444, 27438, 27381, 27410, 27350, 27484, 27485, 27483, 27488, 27489, 27487, 27493, 27492, 27491, 27497, 27496, 27495, 27491, 27480, 27479, 27495, 27482, 27481, 27480, 27483, 27479, 27482, 27487, 27481, 27487, 27486, 27484, 27494, 27495, 27492, 27481, 27484, 27480, 27492, 27481, 27480, 27504, 27505, 27503, 27508, 27509, 27507, 27513, 27512, 27511, 27517, 27516, 27515, 27522, 27519, 27521, 27511, 27500, 27499, 27515, 27502, 27501, 27520, 27508, 27519, 27500, 27503, 27499, 27502, 27507, 27501, 27521, 27516, 27518, 27507, 27506, 27504, 27514, 27515, 27512, 27501, 27504, 27500, 27512, 27501, 27500, 27519, 27502, 27516, 27525, 27524, 27523, 27529, 27528, 27527, 27538, 27535, 27537, 27542, 27539, 27541, 27531, 27526, 27525, 27533, 27530, 27529, 27536, 27531, 27535, 27540, 27533, 27539, 27526, 27527, 27524, 27541, 27536, 27538, 27539, 27532, 27536, 27532, 27529, 27526, 27545, 27544, 27543, 27549, 27548, 27547, 27554, 27551, 27553, 27562, 27559, 27561, 27566, 27563, 27565, 27555, 27546, 27545, 27557, 27550, 27549, 27566, 27553, 27564, 27560, 27555, 27559, 27553, 27548, 27550, 27564, 27557, 27563,
27546, 27547, 27544, 27565, 27560, 27562, 27563, 27556, 27560, 27556, 27549, 27546, 27564, 27550, 27558, 27523, 27494, 27493, 27527, 27498, 27497, 27524, 27497, 27494, 27543, 27514, 27513, 27547, 27518, 27517, 27552, 27521, 27551, 27544, 27517, 27514, 27551, 27518, 27548, 27503, 27490, 27488, 27498, 27511, 27496, 27499, 27488, 27482, 27496, 27499, 27482, 27530, 27543, 27528, 27561, 27540, 27542, 27559, 27534, 27540, 27534, 27545, 27530, 27528, 27513, 27498, 27569, 27568, 27567, 27574, 27575, 27573, 27578, 27574, 27577, 27580, 27581, 27579, 27567, 27574, 27573, 27570, 27574, 27568, 27571, 27580, 27579, 27579, 27578, 27577, 27571, 27577, 27570, 27588, 27589, 27587, 27592, 27593, 27591, 27583, 27588, 27587, 27585, 27592, 27591, 27591, 27590, 27588, 27584, 27591, 27588, 27587, 27582, 27580, 27572, 27587, 27580, 27600, 27601, 27599, 27604, 27605, 27603, 27595, 27600, 27599, 27598, 27603, 27597, 27603, 27602, 27600, 27597, 27600, 27596, 27612, 27613, 27611, 27616, 27617, 27615, 27608, 27611, 27607, 27610, 27615, 27609, 27615, 27614, 27612, 27609, 27612, 27608, 27611, 27606, 27604, 27607, 27604, 27598, 27599, 27594, 27592, 27586, 27599, 27592, 27624, 27625, 27623, 27628, 27629, 27627, 27620, 27623, 27619, 27622, 27627, 27621, 27627, 27626, 27624,
27621, 27624, 27620, 27636, 27637, 27635, 27640, 27641, 27639, 27632, 27635, 27631, 27634, 27639, 27633, 27639, 27638, 27636, 27633, 27636, 27632, 27635, 27630, 27628, 27631, 27628, 27622, 27483, 27427, 27425, 27435, 27491, 27433, 27479, 27425, 27419, 27433, 27479, 27419, 27463, 27523, 27461, 27537, 27473, 27475, 27535, 27467, 27473, 27467, 27525, 27463, 27461, 27493, 27435, 27623, 27618, 27616, 27619, 27616, 27610, 27645, 27644, 27643, 27649, 27648, 27647, 27653, 27652, 27651, 27657, 27656, 27655, 27651, 27646, 27645, 27655, 27650, 27649, 27646, 27647, 27644, 27654, 27655, 27652, 27652, 27649, 27646, 27661, 27660, 27659, 27665, 27664, 27663, 27669, 27668, 27667, 27673, 27672, 27671, 27667, 27662, 27661, 27671, 27666, 27665, 27662, 27663, 27660, 27670, 27671, 27668, 27668, 27665, 27662, 27683, 27680, 27682, 27687, 27684, 27686, 27687, 27688, 27685, 27680, 27676, 27675, 27685, 27677, 27684, 27678, 27688, 27679, 27686, 27681, 27683, 27681, 27677, 27676, 27692, 27691, 27690, 27694, 27697, 27695, 27701, 27698, 27700, 27705, 27702, 27704, 27698, 27693, 27692, 27702, 27697, 27696, 27693, 27694, 27691, 27704, 27699, 27701, 27699, 27696, 27693, 27675, 27654, 27653, 27677, 27658, 27657, 27676, 27657, 27654, 27690, 27670, 27669, 27694, 27674, 27673,
27691, 27673, 27670, 27650, 27659, 27648, 27658, 27667, 27656, 27656, 27661, 27650, 27679, 27690, 27678, 27700, 27688, 27689, 27698, 27679, 27688, 27658, 27690, 27669, 27708, 27707, 27706, 27712, 27711, 27710, 27716, 27715, 27714, 27720, 27719, 27718, 27714, 27709, 27708, 27718, 27713, 27712, 27709, 27710, 27707, 27717, 27718, 27715, 27715, 27712, 27709, 27724, 27723, 27722, 27728, 27727, 27726, 27732, 27731, 27730, 27736, 27735, 27734, 27730, 27725, 27724, 27734, 27729, 27728, 27725, 27726, 27723, 27733, 27734, 27731, 27731, 27728, 27725, 27740, 27739, 27738, 27744, 27743, 27742, 27749, 27746, 27748, 27753, 27750, 27752, 27747, 27740, 27746, 27751, 27744, 27750, 27741, 27742, 27739, 27752, 27747, 27749, 27750, 27741, 27747, 27756, 27755, 27754, 27760, 27759, 27758, 27765, 27762, 27764, 27769, 27766, 27768, 27763, 27756, 27762, 27767, 27760, 27766, 27757, 27758, 27755, 27768, 27763, 27765, 27766, 27757, 27763, 27738, 27717, 27716, 27742, 27721, 27720, 27739, 27720, 27717, 27754, 27733, 27732, 27758, 27737, 27736, 27755, 27736, 27733, 27713, 27722, 27711, 27721, 27730, 27719, 27719, 27724, 27713, 27745, 27754, 27743, 27764, 27751, 27753, 27762, 27745, 27751, 27743, 27732, 27721, 27775, 27776, 27774, 27779, 27780, 27778, 27784, 27783, 27782,
27788, 27787, 27786, 27782, 27771, 27770, 27786, 27773, 27772, 27771, 27774, 27770, 27773, 27778, 27772, 27778, 27777, 27775, 27785, 27786, 27783, 27772, 27775, 27771, 27783, 27772, 27771, 27795, 27796, 27794, 27799, 27800, 27798, 27804, 27803, 27802, 27808, 27807, 27806, 27802, 27791, 27790, 27806, 27793, 27792, 27791, 27794, 27790, 27793, 27798, 27792, 27798, 27797, 27795, 27805, 27806, 27803, 27792, 27795, 27791, 27803, 27792, 27791, 27794, 27781, 27779, 27789, 27802, 27787, 27790, 27779, 27773, 27787, 27790, 27773, 27815, 27816, 27814, 27819, 27820, 27818, 27824, 27823, 27822, 27828, 27827, 27826, 27822, 27811, 27810, 27826, 27813, 27812, 27811, 27814, 27810, 27813, 27818, 27812, 27818, 27817, 27815, 27825, 27826, 27823, 27812, 27815, 27811, 27823, 27812, 27811, 27835, 27836, 27834, 27839, 27840, 27838, 27844, 27843, 27842, 27848, 27847, 27846, 27842, 27831, 27830, 27846, 27833, 27832, 27831, 27834, 27830, 27833, 27838, 27832, 27838, 27837, 27835, 27845, 27846, 27843, 27832, 27835, 27831, 27843, 27832, 27831, 27834, 27821, 27819, 27829, 27842, 27827, 27830, 27819, 27813, 27827, 27830, 27813, 27666, 27706, 27664, 27674, 27714, 27672, 27672, 27708, 27666, 27697, 27738, 27695, 27748, 27703, 27705, 27746, 27697, 27703, 27695, 27716, 27674,
27814, 27801, 27799, 27809, 27822, 27807, 27810, 27799, 27793, 27807, 27810, 27793, 27852, 27851, 27850, 27856, 27855, 27854, 27860, 27859, 27858, 27864, 27863, 27862, 27858, 27853, 27852, 27862, 27857, 27856, 27853, 27854, 27851, 27861, 27862, 27859, 27859, 27856, 27853, 27868, 27867, 27866, 27872, 27871, 27870, 27877, 27874, 27876, 27880, 27879, 27878, 27884, 27883, 27882, 27889, 27886, 27888, 27878, 27869, 27868, 27882, 27873, 27872, 27887, 27876, 27886, 27876, 27871, 27873, 27888, 27883, 27885, 27869, 27870, 27867, 27881, 27882, 27879, 27879, 27872, 27869, 27886, 27873, 27883, 27892, 27891, 27890, 27896, 27895, 27894, 27901, 27898, 27900, 27905, 27902, 27904, 27899, 27892, 27898, 27903, 27896, 27902, 27893, 27894, 27891, 27904, 27899, 27901, 27902, 27893, 27899, 27908, 27907, 27906, 27912, 27911, 27910, 27917, 27914, 27916, 27923, 27920, 27922, 27923, 27918, 27921, 27915, 27908, 27914, 27921, 27912, 27920, 27921, 27911, 27913, 27909, 27910, 27907, 27922, 27915, 27917, 27920, 27909, 27915, 27890, 27861, 27860, 27894, 27865, 27864, 27891, 27864, 27861, 27906, 27881, 27880, 27910, 27885, 27884, 27919, 27888, 27918, 27907, 27884, 27881, 27918, 27885, 27911, 27857, 27866, 27855, 27865, 27878, 27863, 27863, 27868, 27857, 27897, 27906, 27895,
27916, 27903, 27905, 27914, 27897, 27903, 27895, 27880, 27865, 27929, 27930, 27928, 27933, 27934, 27932, 27938, 27937, 27936, 27942, 27941, 27940, 27936, 27925, 27924, 27940, 27927, 27926, 27925, 27928, 27924, 27927, 27932, 27926, 27932, 27931, 27929, 27939, 27940, 27937, 27926, 27929, 27925, 27937, 27926, 27925, 27949, 27950, 27948, 27953, 27954, 27952, 27958, 27957, 27956, 27962, 27961, 27960, 27967, 27964, 27966, 27956, 27945, 27944, 27960, 27947, 27946, 27965, 27953, 27964, 27945, 27948, 27944, 27947, 27952, 27946, 27966, 27961, 27963, 27952, 27951, 27949, 27959, 27960, 27957, 27946, 27949, 27945, 27957, 27946, 27945, 27964, 27947, 27961, 27948, 27935, 27933, 27943, 27956, 27941, 27944, 27933, 27927, 27941, 27944, 27927, 27970, 27969, 27968, 27974, 27973, 27972, 27983, 27980, 27982, 27987, 27984, 27986, 27976, 27971, 27970, 27978, 27975, 27974, 27981, 27976, 27980, 27985, 27978, 27984, 27971, 27972, 27969, 27986, 27981, 27983, 27984, 27977, 27981, 27977, 27974, 27971, 27990, 27989, 27988, 27994, 27993, 27992, 28003, 28000, 28002, 28007, 28004, 28006, 27996, 27991, 27990, 27998, 27995, 27994, 28001, 27996, 28000, 28005, 27998, 28004, 27991, 27992, 27989, 28006, 28001, 28003, 28004, 27997, 28001, 27997, 27994, 27991, 27975, 27988, 27973,
28002, 27985, 27987, 28000, 27979, 27985, 27979, 27990, 27975, 28010, 28009, 28008, 28014, 28013, 28012, 28023, 28020, 28022, 28027, 28024, 28026, 28016, 28011, 28010, 28018, 28015, 28014, 28021, 28016, 28020, 28025, 28018, 28024, 28011, 28012, 28009, 28026, 28021, 28023, 28024, 28017, 28021, 28017, 28014, 28011, 28030, 28029, 28028, 28034, 28033, 28032, 28043, 28040, 28042, 28047, 28044, 28046, 28036, 28031, 28030, 28038, 28035, 28034, 28041, 28036, 28040, 28045, 28038, 28044, 28031, 28032, 28029, 28046, 28041, 28043, 28044, 28037, 28041, 28037, 28034, 28031, 28015, 28028, 28013, 28042, 28025, 28027, 28040, 28019, 28025, 28019, 28030, 28015, 28053, 28054, 28052, 28057, 28058, 28056, 28049, 28052, 28048, 28051, 28056, 28050, 28056, 28055, 28053, 28050, 28053, 28049, 28065, 28066, 28064, 28069, 28070, 28068, 28061, 28064, 28060, 28063, 28068, 28062, 28068, 28067, 28065, 28062, 28065, 28061, 28074, 28073, 28072, 28078, 28077, 28076, 28082, 28081, 28080, 28086, 28085, 28084, 28080, 28075, 28074, 28084, 28079, 28078, 28075, 28076, 28073, 28083, 28084, 28081, 28081, 28078, 28075, 28090, 28089, 28088, 28094, 28093, 28092, 28098, 28097, 28096, 28102, 28101, 28100, 28096, 28091, 28090, 28100, 28095, 28094, 28091, 28092, 28089, 28099, 28100, 28097,
28097, 28094, 28091, 28072, 28049, 28048, 28076, 28051, 28050, 28073, 28050, 28049, 28088, 28061, 28060, 28092, 28063, 28062, 28089, 28062, 28061, 28064, 28059, 28057, 28060, 28057, 28051, 28079, 28088, 28077, 28087, 28096, 28085, 28085, 28090, 28079, 28077, 28060, 28051, 28109, 28110, 28108, 28113, 28114, 28112, 28105, 28108, 28104, 28107, 28112, 28106, 28112, 28111, 28109, 28106, 28109, 28105, 28121, 28122, 28120, 28125, 28126, 28124, 28117, 28120, 28116, 28119, 28124, 28118, 28124, 28123, 28121, 28118, 28121, 28117, 28130, 28129, 28128, 28134, 28133, 28132, 28138, 28137, 28136, 28142, 28141, 28140, 28136, 28131, 28130, 28140, 28135, 28134, 28131, 28132, 28129, 28139, 28140, 28137, 28137, 28134, 28131, 28146, 28145, 28144, 28150, 28149, 28148, 28154, 28153, 28152, 28158, 28157, 28156, 28152, 28147, 28146, 28156, 28151, 28150, 28147, 28148, 28145, 28155, 28156, 28153, 28153, 28150, 28147, 28128, 28105, 28104, 28132, 28107, 28106, 28129, 28106, 28105, 28144, 28117, 28116, 28148, 28119, 28118, 28145, 28118, 28117, 28120, 28115, 28113, 28116, 28113, 28107, 28135, 28144, 28133, 28143, 28152, 28141, 28141, 28146, 28135, 28133, 28116, 28107, 27995, 28008, 27993, 28022, 28005, 28007, 28020, 27999, 28005, 27999, 28010, 27995, 28108, 28071, 28069,
28104, 28069, 28063, 28095, 28128, 28093, 28103, 28136, 28101, 28101, 28130, 28095, 28093, 28104, 28063, 28162, 28161, 28160, 28166, 28165, 28164, 28175, 28172, 28174, 28179, 28176, 28178, 28168, 28163, 28162, 28170, 28167, 28166, 28173, 28168, 28172, 28177, 28170, 28176, 28163, 28164, 28161, 28178, 28173, 28175, 28176, 28169, 28173, 28169, 28166, 28163, 28182, 28181, 28180, 28186, 28185, 28184, 28191, 28188, 28190, 28199, 28196, 28198, 28203, 28200, 28202, 28192, 28183, 28182, 28194, 28187, 28186, 28203, 28190, 28201, 28197, 28192, 28196, 28190, 28185, 28187, 28201, 28194, 28200, 28183, 28184, 28181, 28202, 28197, 28199, 28200, 28193, 28197, 28193, 28186, 28183, 28201, 28187, 28195, 28167, 28180, 28165, 28198, 28177, 28179, 28196, 28171, 28177, 28171, 28182, 28167, 28209, 28210, 28208, 28213, 28214, 28212, 28205, 28208, 28204, 28207, 28212, 28206, 28212, 28211, 28209, 28206, 28209, 28205, 28221, 28222, 28220, 28225, 28226, 28224, 28217, 28220, 28216, 28219, 28224, 28218, 28224, 28223, 28221, 28218, 28221, 28217, 28230, 28229, 28228, 28234, 28233, 28232, 28238, 28237, 28236, 28242, 28241, 28240, 28236, 28231, 28230, 28240, 28235, 28234, 28231, 28232, 28229, 28239, 28240, 28237, 28237, 28234, 28231, 28246, 28245, 28244, 28250, 28249, 28248,
28255, 28252, 28254, 28258, 28257, 28256, 28262, 28261, 28260, 28267, 28264, 28266, 28256, 28247, 28246, 28260, 28251, 28250, 28265, 28254, 28264, 28254, 28249, 28251, 28266, 28261, 28263, 28247, 28248, 28245, 28259, 28260, 28257, 28257, 28250, 28247, 28264, 28251, 28261, 28228, 28205, 28204, 28232, 28207, 28206, 28229, 28206, 28205, 28244, 28217, 28216, 28248, 28219, 28218, 28253, 28225, 28252, 28245, 28218, 28217, 28252, 28219, 28249, 28220, 28215, 28213, 28216, 28213, 28207, 28235, 28244, 28233, 28243, 28256, 28241, 28241, 28246, 28235, 28233, 28216, 28207, 27968, 27785, 27784, 27972, 27789, 27788, 27969, 27788, 27785, 27988, 27805, 27804, 27992, 27809, 27808, 27989, 27808, 27805, 27973, 27804, 27789, 28008, 27825, 27824, 28012, 27829, 27828, 28009, 27828, 27825, 28028, 27845, 27844, 28032, 27849, 27848, 28029, 27848, 27845, 28013, 27844, 27829, 27993, 27824, 27809, 28160, 27939, 27938, 28164, 27943, 27942, 28161, 27942, 27939, 28180, 27959, 27958, 28184, 27963, 27962, 28189, 27966, 28188, 28181, 27962, 27959, 28188, 27963, 28185, 28165, 27958, 27943, 27729, 27850, 27727, 27737, 27858, 27735, 27735, 27852, 27729, 27761, 27890, 27759, 27900, 27767, 27769, 27898, 27761, 27767, 27759, 27860, 27737, 27928, 27841, 27839, 27849, 27936, 27847,
27924, 27839, 27833, 27847, 27924, 27833, 28035, 28160, 28033, 28174, 28045, 28047, 28172, 28039, 28045, 28039, 28162, 28035, 28208, 28127, 28125, 28204, 28125, 28119, 28151, 28228, 28149, 28159, 28236, 28157, 28157, 28230, 28151, 28149, 28204, 28119, 28033, 27938, 27849, 28270, 28269, 28268, 28274, 28273, 28272, 28279, 28276, 28278, 28283, 28280, 28282, 28277, 28270, 28276, 28281, 28274, 28280, 28271, 28272, 28269, 28282, 28277, 28279, 28280, 28271, 28277, 28286, 28285, 28284, 28290, 28289, 28288, 28295, 28292, 28294, 28299, 28296, 28298, 28293, 28286, 28292, 28297, 28290, 28296, 28287, 28288, 28285, 28298, 28293, 28295, 28296, 28287, 28293, 28275, 28284, 28273, 28294, 28281, 28283, 28292, 28275, 28281, 28302, 28301, 28300, 28306, 28305, 28304, 28311, 28308, 28310, 28315, 28312, 28314, 28309, 28302, 28308, 28313, 28306, 28312, 28303, 28304, 28301, 28314, 28309, 28311, 28312, 28303, 28309, 28318, 28317, 28316, 28322, 28321, 28320, 28327, 28324, 28326, 28331, 28328, 28330, 28325, 28318, 28324, 28329, 28322, 28328, 28319, 28320, 28317, 28330, 28325, 28327, 28328, 28319, 28325, 28307, 28316, 28305, 28326, 28313, 28315, 28324, 28307, 28313, 28291, 28300, 28289, 28310, 28297, 28299, 28308, 28291, 28297, 28334, 28333, 28332, 28336, 28339, 28337,
28343, 28340, 28342, 28348, 28347, 28346, 28351, 28352, 28350, 28346, 28339, 28338, 28345, 28350, 28344, 28341, 28334, 28340, 28350, 28348, 28346, 28333, 28338, 28336, 28345, 28341, 28343, 28344, 28335, 28341, 28344, 28346, 28338, 28356, 28355, 28354, 28360, 28359, 28358, 28365, 28362, 28364, 28368, 28367, 28366, 28372, 28371, 28370, 28377, 28374, 28376, 28366, 28357, 28356, 28370, 28361, 28360, 28375, 28364, 28374, 28364, 28359, 28361, 28376, 28371, 28373, 28357, 28358, 28355, 28369, 28370, 28367, 28367, 28360, 28357, 28374, 28361, 28371, 28380, 28379, 28378, 28383, 28384, 28382, 28388, 28387, 28386, 28391, 28392, 28390, 28386, 28381, 28380, 28385, 28390, 28384, 28382, 28380, 28378, 28390, 28388, 28386, 28384, 28386, 28380, 28396, 28395, 28394, 28400, 28399, 28398, 28405, 28402, 28404, 28408, 28407, 28406, 28412, 28411, 28410, 28417, 28414, 28416, 28406, 28397, 28396, 28410, 28401, 28400, 28415, 28404, 28414, 28404, 28399, 28401, 28416, 28411, 28413, 28397, 28398, 28395, 28409, 28410, 28407, 28407, 28400, 28397, 28414, 28401, 28411, 28378, 28349, 28348, 28353, 28382, 28352, 28352, 28378, 28348, 28394, 28369, 28368, 28398, 28373, 28372, 28403, 28376, 28402, 28395, 28372, 28369, 28402, 28373, 28399, 28339, 28354, 28337, 28349, 28366, 28347,
28347, 28356, 28339, 28381, 28394, 28379, 28389, 28406, 28387, 28387, 28396, 28381, 28379, 28368, 28349, 28420, 28419, 28418, 28423, 28424, 28422, 28426, 28429, 28427, 28431, 28432, 28430, 28426, 28421, 28420, 28425, 28430, 28424, 28422, 28420, 28418, 28430, 28428, 28426, 28424, 28426, 28420, 28436, 28435, 28434, 28440, 28439, 28438, 28445, 28442, 28444, 28448, 28447, 28446, 28452, 28451, 28450, 28457, 28454, 28456, 28446, 28437, 28436, 28450, 28441, 28440, 28455, 28444, 28454, 28444, 28439, 28441, 28456, 28451, 28453, 28437, 28438, 28435, 28449, 28450, 28447, 28447, 28440, 28437, 28454, 28441, 28451, 28460, 28459, 28458, 28463, 28464, 28462, 28467, 28470, 28468, 28474, 28475, 28473, 28460, 28468, 28461, 28465, 28473, 28464, 28462, 28460, 28458, 28467, 28475, 28469, 28475, 28472, 28471, 28469, 28471, 28466, 28464, 28467, 28460, 28479, 28478, 28477, 28483, 28482, 28481, 28488, 28485, 28487, 28491, 28490, 28489, 28495, 28494, 28493, 28500, 28497, 28499, 28489, 28480, 28479, 28493, 28484, 28483, 28498, 28487, 28497, 28487, 28482, 28484, 28499, 28494, 28496, 28480, 28481, 28478, 28492, 28493, 28490, 28490, 28483, 28480, 28497, 28484, 28494, 28458, 28429, 28428, 28433, 28462, 28432, 28432, 28458, 28428, 28477, 28449, 28448, 28481, 28453, 28452,
28486, 28456, 28485, 28478, 28452, 28449, 28485, 28453, 28482, 28421, 28434, 28419, 28429, 28446, 28427, 28427, 28436, 28421, 28461, 28477, 28459, 28470, 28489, 28468, 28468, 28479, 28461, 28459, 28448, 28429, 28418, 28389, 28388, 28393, 28422, 28392, 28392, 28418, 28388, 28434, 28409, 28408, 28438, 28413, 28412, 28443, 28416, 28442, 28435, 28412, 28409, 28442, 28413, 28439, 28419, 28408, 28389, 28504, 28505, 28503, 28501, 28504, 28503, 28510, 28511, 28509, 28515, 28518, 28516, 28520, 28521, 28519, 28524, 28525, 28523, 28525, 28527, 28523, 28512, 28524, 28511, 28514, 28519, 28513, 28509, 28508, 28507, 28515, 28524, 28523, 28523, 28517, 28515, 28527, 28522, 28520, 28517, 28520, 28514, 28511, 28516, 28508, 28529, 28532, 28530, 28535, 28534, 28533, 28540, 28541, 28539, 28544, 28540, 28543, 28547, 28546, 28545, 28551, 28550, 28549, 28545, 28532, 28531, 28535, 28550, 28536, 28534, 28539, 28533, 28536, 28540, 28534, 28532, 28533, 28530, 28539, 28538, 28537, 28548, 28549, 28546, 28530, 28539, 28537, 28546, 28535, 28532, 28558, 28559, 28557, 28562, 28563, 28561, 28567, 28566, 28565, 28571, 28570, 28569, 28565, 28554, 28553, 28555, 28570, 28556, 28553, 28558, 28557, 28556, 28561, 28555, 28561, 28560, 28558, 28568, 28569, 28566, 28554, 28561, 28558,
28554, 28569, 28555, 28575, 28574, 28573, 28579, 28578, 28577, 28584, 28581, 28583, 28593, 28590, 28592, 28586, 28580, 28579, 28584, 28588, 28582, 28581, 28576, 28575, 28590, 28587, 28586, 28574, 28579, 28577, 28592, 28588, 28589, 28588, 28586, 28585, 28585, 28579, 28576, 28576, 28588, 28585, 28596, 28595, 28594, 28600, 28599, 28598, 28609, 28606, 28608, 28613, 28610, 28612, 28602, 28597, 28596, 28604, 28601, 28600, 28606, 28603, 28602, 28610, 28605, 28604, 28597, 28598, 28595, 28612, 28607, 28609, 28607, 28604, 28603, 28597, 28604, 28600, 28573, 28548, 28547, 28577, 28552, 28551, 28574, 28551, 28548, 28594, 28568, 28567, 28598, 28572, 28571, 28595, 28571, 28568, 28557, 28544, 28543, 28552, 28565, 28550, 28553, 28543, 28536, 28550, 28553, 28536, 28580, 28594, 28578, 28608, 28591, 28593, 28591, 28602, 28587, 28587, 28596, 28580, 28578, 28567, 28552, 28616, 28619, 28617, 28621, 28622, 28620, 28625, 28626, 28624, 28626, 28628, 28624, 28632, 28631, 28630, 28636, 28635, 28634, 28614, 28631, 28615, 28634, 28619, 28618, 28615, 28620, 28614, 28616, 28625, 28624, 28624, 28618, 28616, 28628, 28623, 28621, 28633, 28634, 28631, 28615, 28628, 28621, 28615, 28634, 28618, 28640, 28639, 28638, 28644, 28643, 28642, 28650, 28649, 28648, 28654, 28653, 28652,
28648, 28641, 28640, 28652, 28645, 28644, 28641, 28642, 28639, 28651, 28652, 28649, 28649, 28644, 28641, 28658, 28657, 28656, 28662, 28661, 28660, 28671, 28668, 28670, 28675, 28672, 28674, 28658, 28665, 28659, 28662, 28667, 28663, 28669, 28664, 28668, 28673, 28666, 28672, 28659, 28660, 28657, 28674, 28669, 28671, 28672, 28665, 28669, 28659, 28666, 28662, 28676, 28679, 28677, 28680, 28683, 28681, 28687, 28684, 28686, 28693, 28690, 28692, 28693, 28688, 28691, 28696, 28684, 28695, 28684, 28679, 28678, 28691, 28682, 28690, 28691, 28681, 28683, 28677, 28682, 28680, 28692, 28685, 28687, 28690, 28679, 28685, 28695, 28678, 28694, 28656, 28633, 28632, 28660, 28637, 28636, 28657, 28636, 28633, 28676, 28651, 28650, 28680, 28655, 28654, 28677, 28654, 28651, 28617, 28640, 28638, 28646, 28627, 28625, 28637, 28648, 28635, 28638, 28625, 28617, 28635, 28640, 28619, 28663, 28676, 28661, 28696, 28673, 28675, 28673, 28694, 28667, 28663, 28694, 28678, 28661, 28650, 28637, 28529, 28502, 28501, 28538, 28504, 28537, 28530, 28504, 28502, 28513, 28639, 28514, 28642, 28518, 28517, 28521, 28646, 28519, 28514, 28642, 28517, 28519, 28638, 28513, 28620, 28564, 28562, 28572, 28630, 28570, 28614, 28562, 28556, 28570, 28614, 28556, 28601, 28656, 28599, 28670, 28611, 28613,
28668, 28605, 28611, 28605, 28658, 28601, 28599, 28632, 28572, 28701, 28700, 28699, 28704, 28705, 28703, 28712, 28711, 28710, 28717, 28718, 28716, 28715, 28717, 28714, 28710, 28702, 28701, 28706, 28714, 28705, 28703, 28698, 28697, 28709, 28716, 28708, 28707, 28717, 28709, 28697, 28701, 28699, 28707, 28712, 28710, 28698, 28710, 28701, 28705, 28707, 28698, 28722, 28721, 28720, 28724, 28727, 28725, 28731, 28728, 28730, 28734, 28733, 28732, 28742, 28739, 28741, 28744, 28739, 28743, 28732, 28723, 28722, 28726, 28737, 28727, 28742, 28730, 28740, 28730, 28725, 28727, 28740, 28736, 28739, 28738, 28739, 28736, 28723, 28724, 28721, 28733, 28738, 28736, 28733, 28726, 28723, 28737, 28730, 28727, 28747, 28746, 28745, 28751, 28750, 28749, 28757, 28756, 28755, 28761, 28760, 28759, 28755, 28748, 28747, 28759, 28752, 28751, 28748, 28749, 28746, 28758, 28759, 28756, 28756, 28751, 28748, 28765, 28764, 28763, 28774, 28771, 28773, 28775, 28778, 28776, 28784, 28781, 28783, 28784, 28779, 28782, 28765, 28776, 28766, 28780, 28769, 28779, 28782, 28777, 28781, 28782, 28776, 28778, 28768, 28771, 28767, 28764, 28768, 28767, 28774, 28769, 28770, 28773, 28766, 28769, 28776, 28769, 28766, 28708, 28746, 28709, 28749, 28713, 28712, 28718, 28753, 28716, 28746, 28712, 28709,
28708, 28753, 28745, 28763, 28735, 28734, 28772, 28743, 28771, 28735, 28767, 28738, 28771, 28738, 28767, 28702, 28720, 28700, 28713, 28732, 28711, 28711, 28722, 28702, 28752, 28763, 28750, 28762, 28775, 28760, 28760, 28765, 28752, 28750, 28734, 28713, 28787, 28786, 28785, 28791, 28790, 28789, 28795, 28798, 28796, 28805, 28802, 28804, 28809, 28806, 28808, 28811, 28806, 28810, 28795, 28788, 28787, 28791, 28800, 28792, 28809, 28793, 28807, 28803, 28797, 28802, 28807, 28799, 28806, 28810, 28799, 28801, 28788, 28789, 28786, 28796, 28801, 28799, 28811, 28803, 28805, 28810, 28798, 28803, 28796, 28791, 28788, 28807, 28792, 28800, 28817, 28814, 28816, 28816, 28812, 28813, 28820, 28802, 28819, 28819, 28797, 28818, 28790, 28813, 28812, 28817, 28793, 28794, 28793, 28813, 28792, 28785, 28758, 28757, 28789, 28762, 28761, 28786, 28761, 28758, 28815, 28781, 28814, 28814, 28777, 28812, 28762, 28812, 28777, 28699, 28470, 28469, 28472, 28703, 28471, 28466, 28699, 28469, 28466, 28703, 28697, 28720, 28492, 28491, 28724, 28496, 28495, 28729, 28499, 28728, 28721, 28495, 28492, 28728, 28496, 28725, 28700, 28491, 28470, 28323, 28332, 28321, 28342, 28329, 28331, 28340, 28323, 28329, 28507, 28747, 28745, 28753, 28510, 28509, 28518, 28755, 28516, 28507, 28753, 28509,
28508, 28755, 28747, 28645, 28785, 28643, 28655, 28795, 28653, 28653, 28787, 28645, 28820, 28688, 28689, 28688, 28818, 28681, 28655, 28818, 28797, 28643, 28757, 28518, 28268, 28083, 28082, 28272, 28087, 28086, 28269, 28086, 28083, 28284, 28099, 28098, 28288, 28103, 28102, 28285, 28102, 28099, 28273, 28098, 28087, 28300, 28139, 28138, 28304, 28143, 28142, 28301, 28142, 28139, 28316, 28155, 28154, 28320, 28159, 28158, 28317, 28158, 28155, 28305, 28154, 28143, 28289, 28138, 28103, 28332, 28239, 28238, 28336, 28243, 28242, 28333, 28242, 28239, 28354, 28259, 28258, 28358, 28263, 28262, 28363, 28266, 28362, 28355, 28262, 28259, 28362, 28263, 28359, 28337, 28258, 28243, 28321, 28238, 28159, 28827, 28828, 28826, 28825, 28827, 28824, 28830, 28833, 28831, 28835, 28836, 28834, 28822, 28831, 28823, 28828, 28834, 28826, 28822, 28827, 28826, 28824, 28823, 28821, 28834, 28832, 28830, 28826, 28830, 28822, 28840, 28843, 28841, 28845, 28846, 28844, 28849, 28850, 28848, 28850, 28852, 28848, 28839, 28844, 28838, 28840, 28849, 28848, 28848, 28842, 28840, 28852, 28847, 28845, 28842, 28845, 28839, 28859, 28860, 28858, 28863, 28864, 28862, 28868, 28867, 28866, 28870, 28873, 28871, 28854, 28867, 28855, 28870, 28857, 28856, 28860, 28874, 28858, 28877, 28863, 28876,
28854, 28859, 28858, 28857, 28862, 28856, 28862, 28861, 28859, 28869, 28870, 28867, 28856, 28859, 28855, 28855, 28870, 28856, 28858, 28866, 28854, 28871, 28863, 28857, 28880, 28879, 28878, 28884, 28883, 28882, 28890, 28889, 28888, 28894, 28893, 28892, 28888, 28881, 28880, 28892, 28885, 28884, 28881, 28882, 28879, 28891, 28892, 28889, 28889, 28884, 28881, 28898, 28897, 28896, 28902, 28901, 28900, 28906, 28905, 28904, 28910, 28909, 28908, 28904, 28899, 28898, 28908, 28903, 28902, 28899, 28900, 28897, 28907, 28908, 28905, 28905, 28902, 28899, 28838, 28879, 28839, 28882, 28843, 28842, 28846, 28886, 28844, 28839, 28882, 28842, 28844, 28878, 28838, 28896, 28869, 28868, 28900, 28873, 28872, 28897, 28872, 28869, 28841, 28868, 28866, 28874, 28851, 28849, 28866, 28849, 28841, 28885, 28896, 28883, 28895, 28904, 28893, 28893, 28898, 28885, 28883, 28868, 28843, 28916, 28915, 28914, 28919, 28920, 28918, 28928, 28927, 28926, 28931, 28932, 28930, 28926, 28917, 28916, 28921, 28930, 28920, 28912, 28920, 28913, 28930, 28925, 28924, 28912, 28916, 28914, 28922, 28919, 28918, 28924, 28928, 28926, 28914, 28918, 28912, 28913, 28926, 28916, 28920, 28924, 28913, 28938, 28937, 28936, 28941, 28942, 28940, 28948, 28947, 28946, 28951, 28952, 28950, 28946, 28939, 28938,
28943, 28950, 28942, 28940, 28935, 28934, 28950, 28945, 28944, 28935, 28936, 28934, 28945, 28946, 28944, 28935, 28946, 28938, 28942, 28944, 28935, 28936, 28929, 28928, 28933, 28940, 28932, 28925, 28936, 28928, 28932, 28934, 28925, 28956, 28955, 28954, 28960, 28959, 28958, 28966, 28965, 28964, 28968, 28971, 28969, 28975, 28972, 28974, 28964, 28957, 28956, 28960, 28969, 28961, 28973, 28962, 28972, 28974, 28969, 28971, 28957, 28958, 28955, 28967, 28968, 28965, 28965, 28960, 28957, 28972, 28961, 28969, 28984, 28981, 28983, 28986, 28981, 28985, 28990, 28987, 28989, 28982, 28976, 28981, 28985, 28976, 28978, 28987, 28980, 28979, 28989, 28982, 28984, 28987, 28977, 28982, 28993, 28992, 28991, 29001, 28998, 29000, 29003, 28998, 29002, 29006, 29005, 29004, 29013, 29010, 29012, 29004, 28994, 28993, 29011, 29002, 29010, 28999, 28995, 28998, 28997, 28998, 28995, 29012, 29008, 29009, 28994, 28995, 28992, 29005, 29009, 29008, 29005, 28997, 28994, 29010, 28997, 29008, 28991, 28967, 28966, 28970, 28996, 28971, 29001, 28974, 28999, 28992, 28970, 28967, 28996, 28974, 28971, 28959, 28978, 28976, 28986, 28962, 28963, 28985, 28961, 28962, 28832, 28915, 28833, 28837, 28922, 28836, 28832, 28922, 28914, 28954, 28891, 28890, 28958, 28895, 28894, 28955, 28894, 28891,
28906, 28977, 28907, 28979, 28911, 28910, 28907, 28979, 28910, 28959, 28906, 28895, 28821, 28880, 28878, 28886, 28825, 28824, 28833, 28888, 28831, 28878, 28824, 28821, 28831, 28880, 28823, 28917, 28954, 28915, 28929, 28964, 28927, 28927, 28956, 28917, 28939, 28991, 28937, 28949, 29004, 28947, 28947, 28993, 28939, 28937, 28966, 28929, 28915, 28890, 28833, 29016, 29015, 29014, 29022, 29023, 29021, 29028, 29029, 29027, 29029, 29025, 29027, 29014, 29022, 29021, 29019, 29028, 29027, 29019, 29025, 29018, 29017, 29018, 29015, 29025, 29024, 29022, 29018, 29022, 29015, 29036, 29035, 29034, 29040, 29039, 29038, 29044, 29043, 29042, 29048, 29047, 29046, 29042, 29037, 29036, 29046, 29041, 29040, 29037, 29038, 29035, 29045, 29046, 29043, 29043, 29040, 29037, 29052, 29051, 29050, 29055, 29056, 29054, 29059, 29055, 29058, 29060, 29063, 29061, 29071, 29068, 29070, 29060, 29053, 29052, 29067, 29058, 29066, 29051, 29054, 29050, 29053, 29055, 29051, 29065, 29068, 29064, 29063, 29064, 29061, 29068, 29067, 29066, 29061, 29068, 29066, 29061, 29058, 29053, 29034, 29017, 29016, 29038, 29020, 29019, 29017, 29038, 29019, 29056, 29032, 29054, 29050, 29032, 29031, 29032, 29030, 29028, 29020, 29032, 29028, 29039, 29052, 29050, 29049, 29060, 29047, 29047, 29052, 29041,
29039, 29031, 29020, 29074, 29073, 29072, 29078, 29077, 29076, 29085, 29082, 29084, 29081, 29082, 29080, 29088, 29091, 29089, 29093, 29094, 29092, 29088, 29079, 29078, 29085, 29092, 29083, 29082, 29075, 29074, 29072, 29082, 29074, 29086, 29094, 29087, 29073, 29078, 29076, 29087, 29088, 29086, 29097, 29094, 29095, 29094, 29090, 29087, 29086, 29078, 29075, 29075, 29092, 29086, 29100, 29099, 29098, 29104, 29103, 29102, 29107, 29108, 29106, 29111, 29107, 29110, 29114, 29113, 29112, 29116, 29119, 29117, 29123, 29120, 29122, 29112, 29101, 29100, 29116, 29105, 29104, 29121, 29110, 29120, 29102, 29107, 29106, 29110, 29103, 29105, 29119, 29120, 29117, 29099, 29104, 29102, 29115, 29116, 29113, 29113, 29104, 29101, 29117, 29110, 29105, 29124, 29127, 29125, 29129, 29130, 29128, 29134, 29133, 29132, 29137, 29138, 29136, 29126, 29133, 29127, 29131, 29136, 29130, 29124, 29130, 29126, 29132, 29138, 29134, 29126, 29136, 29132, 29142, 29141, 29140, 29146, 29145, 29144, 29151, 29148, 29150, 29154, 29153, 29152, 29158, 29157, 29156, 29163, 29160, 29162, 29152, 29143, 29142, 29156, 29147, 29146, 29161, 29150, 29160, 29150, 29145, 29147, 29162, 29157, 29159, 29143, 29144, 29141, 29155, 29156, 29153, 29153, 29146, 29143, 29160, 29147, 29157, 29124, 29091, 29090,
29097, 29128, 29096, 29096, 29124, 29090, 29140, 29115, 29114, 29118, 29145, 29119, 29149, 29122, 29148, 29141, 29118, 29115, 29145, 29122, 29119, 29079, 29098, 29077, 29091, 29112, 29089, 29089, 29100, 29079, 29127, 29140, 29125, 29135, 29152, 29133, 29133, 29142, 29127, 29125, 29114, 29091, 29044, 29073, 29045, 29076, 29049, 29048, 29073, 29048, 29045, 29098, 29063, 29062, 29108, 29070, 29106, 29099, 29065, 29063, 29106, 29065, 29102, 29077, 29062, 29049, 29168, 29167, 29166, 29171, 29172, 29170, 29178, 29177, 29176, 29181, 29182, 29180, 29176, 29169, 29168, 29173, 29180, 29172, 29170, 29165, 29164, 29180, 29175, 29174, 29165, 29166, 29164, 29175, 29176, 29174, 29174, 29168, 29165, 29172, 29174, 29165, 29188, 29187, 29186, 29191, 29192, 29190, 29198, 29197, 29196, 29201, 29202, 29200, 29196, 29189, 29188, 29193, 29200, 29192, 29190, 29185, 29184, 29200, 29195, 29194, 29185, 29186, 29184, 29195, 29196, 29194, 29194, 29188, 29185, 29192, 29194, 29185, 29186, 29179, 29178, 29183, 29190, 29182, 29184, 29178, 29175, 29182, 29184, 29175, 29206, 29205, 29204, 29213, 29210, 29212, 29216, 29215, 29214, 29223, 29220, 29222, 29214, 29207, 29206, 29221, 29212, 29220, 29212, 29208, 29209, 29222, 29218, 29219, 29207, 29208, 29205, 29217, 29218, 29215,
29215, 29209, 29207, 29220, 29209, 29218, 29226, 29225, 29224, 29233, 29230, 29232, 29236, 29235, 29234, 29243, 29240, 29242, 29234, 29227, 29226, 29241, 29232, 29240, 29232, 29228, 29229, 29242, 29238, 29239, 29227, 29228, 29225, 29237, 29238, 29235, 29235, 29229, 29227, 29240, 29229, 29238, 29224, 29217, 29216, 29231, 29222, 29230, 29225, 29219, 29217, 29230, 29219, 29228, 29248, 29247, 29246, 29251, 29252, 29250, 29258, 29257, 29256, 29261, 29262, 29260, 29256, 29249, 29248, 29253, 29260, 29252, 29250, 29245, 29244, 29260, 29255, 29254, 29244, 29248, 29246, 29254, 29258, 29256, 29245, 29256, 29248, 29252, 29254, 29245, 29272, 29271, 29270, 29277, 29278, 29276, 29275, 29277, 29274, 29280, 29283, 29281, 29286, 29285, 29284, 29289, 29290, 29288, 29268, 29281, 29269, 29284, 29273, 29272, 29278, 29288, 29276, 29269, 29276, 29268, 29274, 29269, 29267, 29288, 29282, 29280, 29267, 29272, 29270, 29283, 29284, 29281, 29281, 29272, 29269, 29276, 29280, 29268, 29270, 29259, 29258, 29263, 29274, 29262, 29255, 29270, 29258, 29262, 29267, 29255, 29290, 29266, 29265, 29282, 29265, 29264, 29294, 29293, 29292, 29301, 29298, 29300, 29304, 29303, 29302, 29311, 29308, 29310, 29302, 29295, 29294, 29309, 29300, 29308, 29300, 29296, 29297, 29310, 29306, 29307,
29295, 29296, 29293, 29305, 29306, 29303, 29303, 29297, 29295, 29308, 29297, 29306, 29314, 29313, 29312, 29321, 29318, 29320, 29324, 29323, 29322, 29331, 29328, 29330, 29322, 29315, 29314, 29329, 29320, 29328, 29320, 29316, 29317, 29330, 29326, 29327, 29313, 29317, 29316, 29323, 29327, 29326, 29315, 29326, 29317, 29328, 29317, 29326, 29312, 29305, 29304, 29319, 29310, 29318, 29305, 29316, 29307, 29318, 29307, 29316, 29246, 29199, 29198, 29203, 29250, 29202, 29195, 29246, 29198, 29202, 29244, 29195, 29292, 29237, 29236, 29299, 29242, 29298, 29293, 29239, 29237, 29298, 29239, 29296, 29169, 29204, 29167, 29179, 29214, 29177, 29177, 29206, 29169, 29189, 29224, 29187, 29199, 29234, 29197, 29197, 29226, 29189, 29187, 29216, 29179, 29249, 29292, 29247, 29259, 29302, 29257, 29257, 29294, 29249, 29273, 29312, 29271, 29287, 29322, 29285, 29285, 29314, 29273, 29271, 29304, 29259, 29247, 29236, 29199, 29334, 29333, 29332, 29337, 29338, 29336, 29342, 29341, 29340, 29345, 29346, 29344, 29340, 29335, 29334, 29339, 29344, 29338, 29336, 29334, 29332, 29344, 29342, 29340, 29338, 29340, 29334, 29350, 29349, 29348, 29354, 29353, 29352, 29359, 29356, 29358, 29362, 29361, 29360, 29366, 29365, 29364, 29371, 29368, 29370, 29360, 29351, 29350, 29364, 29355, 29354,
29369, 29358, 29368, 29358, 29353, 29355, 29370, 29365, 29367, 29351, 29352, 29349, 29363, 29364, 29361, 29361, 29354, 29351, 29368, 29355, 29365, 29374, 29373, 29372, 29377, 29378, 29376, 29382, 29381, 29380, 29385, 29386, 29384, 29380, 29375, 29374, 29379, 29384, 29378, 29376, 29374, 29372, 29384, 29382, 29380, 29378, 29380, 29374, 29390, 29389, 29388, 29394, 29393, 29392, 29399, 29396, 29398, 29402, 29401, 29400, 29406, 29405, 29404, 29411, 29408, 29410, 29400, 29391, 29390, 29404, 29395, 29394, 29409, 29398, 29408, 29398, 29393, 29395, 29410, 29405, 29407, 29391, 29392, 29389, 29403, 29404, 29401, 29401, 29394, 29391, 29408, 29395, 29405, 29372, 29343, 29342, 29347, 29376, 29346, 29346, 29372, 29342, 29388, 29363, 29362, 29392, 29367, 29366, 29397, 29370, 29396, 29389, 29366, 29363, 29396, 29367, 29393, 29335, 29348, 29333, 29343, 29360, 29341, 29341, 29350, 29335, 29375, 29388, 29373, 29383, 29400, 29381, 29381, 29390, 29375, 29373, 29362, 29343, 29414, 29413, 29412, 29417, 29418, 29416, 29422, 29421, 29420, 29425, 29426, 29424, 29420, 29415, 29414, 29419, 29424, 29418, 29416, 29414, 29412, 29424, 29422, 29420, 29418, 29420, 29414, 29430, 29429, 29428, 29434, 29433, 29432, 29439, 29436, 29438, 29442, 29441, 29440, 29446, 29445, 29444,
29451, 29448, 29450, 29440, 29431, 29430, 29444, 29435, 29434, 29449, 29438, 29448, 29438, 29433, 29435, 29450, 29445, 29447, 29431, 29432, 29429, 29443, 29444, 29441, 29441, 29434, 29431, 29448, 29435, 29445, 29454, 29453, 29452, 29457, 29458, 29456, 29462, 29461, 29460, 29465, 29466, 29464, 29460, 29455, 29454, 29459, 29464, 29458, 29456, 29454, 29452, 29464, 29462, 29460, 29458, 29460, 29454, 29470, 29469, 29468, 29474, 29473, 29472, 29479, 29476, 29478, 29482, 29481, 29480, 29486, 29485, 29484, 29491, 29488, 29490, 29480, 29471, 29470, 29484, 29475, 29474, 29489, 29478, 29488, 29478, 29473, 29475, 29490, 29485, 29487, 29471, 29472, 29469, 29483, 29484, 29481, 29481, 29474, 29471, 29488, 29475, 29485, 29452, 29423, 29422, 29427, 29456, 29426, 29426, 29452, 29422, 29468, 29443, 29442, 29472, 29447, 29446, 29477, 29450, 29476, 29469, 29446, 29443, 29476, 29447, 29473, 29415, 29428, 29413, 29423, 29440, 29421, 29421, 29430, 29415, 29455, 29468, 29453, 29463, 29480, 29461, 29461, 29470, 29455, 29453, 29442, 29423, 29412, 29383, 29382, 29387, 29416, 29386, 29386, 29412, 29382, 29428, 29403, 29402, 29432, 29407, 29406, 29437, 29410, 29436, 29429, 29406, 29403, 29436, 29407, 29433, 29413, 29402, 29383, 29166, 28949, 28948, 28953, 29170, 28952,
29164, 28948, 28945, 28952, 29164, 28945, 29204, 29007, 29006, 29211, 29012, 29210, 29205, 29009, 29007, 29210, 29009, 29208, 29167, 29006, 28949, 29134, 29333, 29135, 29139, 29336, 29138, 29134, 29336, 29332, 29348, 29155, 29154, 29352, 29159, 29158, 29357, 29162, 29356, 29349, 29158, 29155, 29356, 29159, 29353, 29333, 29154, 29135, 28871, 29016, 29014, 29021, 28877, 28876, 29014, 28876, 28871, 28903, 29034, 28901, 28911, 29042, 28909, 28909, 29036, 28903, 28901, 29016, 28873, 29081, 28988, 28990, 28988, 29072, 28980, 28980, 29044, 28911, 29494, 29497, 29495, 29499, 29500, 29498, 29503, 29504, 29502, 29504, 29506, 29502, 29510, 29509, 29508, 29514, 29513, 29512, 29492, 29509, 29493, 29512, 29497, 29496, 29492, 29499, 29498, 29494, 29503, 29502, 29502, 29496, 29494, 29506, 29501, 29499, 29511, 29512, 29509, 29493, 29506, 29499, 29493, 29512, 29496, 29518, 29517, 29516, 29522, 29521, 29520, 29526, 29525, 29524, 29530, 29529, 29528, 29524, 29519, 29518, 29528, 29523, 29522, 29519, 29520, 29517, 29527, 29528, 29525, 29525, 29522, 29519, 29536, 29535, 29534, 29540, 29539, 29538, 29549, 29546, 29548, 29553, 29550, 29552, 29536, 29543, 29537, 29540, 29545, 29541, 29547, 29542, 29546, 29551, 29544, 29550, 29537, 29538, 29535, 29552, 29547, 29549,
29550, 29543, 29547, 29537, 29544, 29540, 29554, 29557, 29555, 29564, 29561, 29563, 29568, 29565, 29567, 29570, 29565, 29569, 29573, 29561, 29572, 29562, 29556, 29561, 29565, 29559, 29558, 29569, 29558, 29560, 29555, 29560, 29558, 29570, 29562, 29564, 29569, 29557, 29562, 29572, 29556, 29571, 29534, 29511, 29510, 29538, 29515, 29514, 29535, 29514, 29511, 29554, 29527, 29526, 29530, 29559, 29531, 29568, 29532, 29566, 29555, 29530, 29527, 29566, 29531, 29559, 29497, 29516, 29495, 29515, 29524, 29513, 29513, 29518, 29497, 29541, 29554, 29539, 29573, 29551, 29553, 29551, 29571, 29545, 29541, 29571, 29556, 29539, 29526, 29515, 29574, 29577, 29575, 29582, 29579, 29581, 29591, 29588, 29590, 29593, 29588, 29592, 29576, 29586, 29577, 29591, 29581, 29589, 29581, 29575, 29577, 29589, 29585, 29588, 29587, 29588, 29585, 29584, 29579, 29580, 29583, 29575, 29579, 29589, 29577, 29586, 29523, 29574, 29521, 29529, 29587, 29585, 29593, 29532, 29533, 29592, 29531, 29532, 29529, 29576, 29523, 29596, 29595, 29594, 29599, 29600, 29598, 29604, 29603, 29602, 29607, 29608, 29606, 29602, 29597, 29596, 29601, 29606, 29600, 29598, 29596, 29594, 29606, 29604, 29602, 29600, 29602, 29596, 29612, 29611, 29610, 29616, 29615, 29614, 29621, 29618, 29620, 29624, 29623, 29622,
29628, 29627, 29626, 29633, 29630, 29632, 29622, 29613, 29612, 29626, 29617, 29616, 29631, 29620, 29630, 29620, 29615, 29617, 29632, 29627, 29629, 29613, 29614, 29611, 29625, 29626, 29623, 29623, 29616, 29613, 29630, 29617, 29627, 29639, 29636, 29638, 29637, 29634, 29636, 29647, 29644, 29646, 29651, 29648, 29650, 29645, 29640, 29644, 29649, 29642, 29648, 29650, 29645, 29647, 29648, 29641, 29645, 29634, 29605, 29604, 29609, 29636, 29608, 29608, 29634, 29604, 29640, 29625, 29624, 29642, 29629, 29628, 29651, 29632, 29649, 29641, 29628, 29625, 29649, 29629, 29643, 29597, 29610, 29595, 29605, 29622, 29603, 29603, 29612, 29597, 29646, 29637, 29639, 29644, 29635, 29637, 29635, 29624, 29605, 29266, 29503, 29265, 29264, 29503, 29495, 29516, 29283, 29282, 29520, 29287, 29286, 29517, 29286, 29283, 29264, 29516, 29282, 29574, 29325, 29324, 29584, 29330, 29583, 29325, 29578, 29327, 29578, 29330, 29327, 29521, 29324, 29287, 29594, 29463, 29462, 29467, 29598, 29466, 29466, 29594, 29462, 29610, 29483, 29482, 29614, 29487, 29486, 29619, 29490, 29618, 29611, 29486, 29483, 29618, 29487, 29615, 29595, 29482, 29463, 27569, 27644, 27570, 27647, 27572, 27571, 27644, 27571, 27570, 27659, 27584, 27583, 27585, 27664, 27586, 27660, 27585, 27584, 27648, 27583, 27572,
27706, 27596, 27595, 27710, 27598, 27597, 27596, 27710, 27597, 27722, 27608, 27607, 27726, 27610, 27609, 27723, 27609, 27608, 27711, 27607, 27598, 27586, 27706, 27595, 27850, 27620, 27619, 27854, 27622, 27621, 27851, 27621, 27620, 27866, 27632, 27631, 27870, 27634, 27633, 27875, 27640, 27874, 27867, 27633, 27632, 27874, 27634, 27871, 27855, 27631, 27622, 27727, 27619, 27610, 29653, 29654, 29652, 29657, 29658, 29656, 29665, 29662, 29663, 29671, 29668, 29670, 29655, 29666, 29654, 29674, 29664, 29665, 29660, 29655, 29653, 29659, 29662, 29658, 29674, 29670, 29673, 29656, 29653, 29652, 29664, 29660, 29662, 29670, 29666, 29667, 29658, 29660, 29653, 29673, 29667, 29672, 29673, 29661, 29664, 29661, 29667, 29655, 29676, 29677, 29675, 29680, 29681, 29679, 29686, 29683, 29684, 29690, 29687, 29688, 29694, 29691, 29693, 29698, 29695, 29697, 29700, 29701, 29699, 29704, 29685, 29686, 29706, 29689, 29690, 29678, 29683, 29677, 29704, 29693, 29703, 29682, 29687, 29681, 29706, 29697, 29705, 29702, 29707, 29701, 29679, 29678, 29676, 29689, 29684, 29687, 29697, 29692, 29694, 29681, 29684, 29678, 29705, 29694, 29704, 29705, 29686, 29689, 29675, 29659, 29657, 29685, 29663, 29683, 29693, 29669, 29671, 29677, 29663, 29659, 29703, 29671, 29674, 29703, 29665, 29685,
29710, 29711, 29709, 29714, 29715, 29713, 29720, 29717, 29718, 29724, 29721, 29722, 29728, 29725, 29727, 29734, 29731, 29733, 29734, 29735, 29732, 29738, 29719, 29720, 29724, 29740, 29723, 29712, 29717, 29711, 29738, 29727, 29737, 29716, 29721, 29715, 29741, 29733, 29740, 29741, 29736, 29734, 29713, 29712, 29710, 29723, 29718, 29721, 29733, 29726, 29728, 29715, 29718, 29712, 29740, 29728, 29738, 29740, 29720, 29723, 29744, 29745, 29743, 29748, 29744, 29747, 29752, 29749, 29750, 29756, 29753, 29755, 29760, 29757, 29758, 29754, 29747, 29753, 29758, 29751, 29752, 29746, 29749, 29745, 29752, 29746, 29748, 29760, 29754, 29756, 29758, 29748, 29754, 29743, 29716, 29714, 29751, 29722, 29749, 29757, 29742, 29741, 29745, 29722, 29716, 29751, 29741, 29724, 29762, 29763, 29761, 29768, 29765, 29766, 29770, 29771, 29769, 29776, 29773, 29774, 29764, 29769, 29763, 29774, 29767, 29768, 29765, 29764, 29762, 29773, 29772, 29770, 29767, 29770, 29764, 29782, 29779, 29781, 29778, 29779, 29777, 29786, 29783, 29784, 29786, 29781, 29785, 29783, 29781, 29778, 29772, 29777, 29771, 29784, 29775, 29776, 29775, 29778, 29772, 29790, 29787, 29789, 29792, 29793, 29791, 29802, 29799, 29801, 29804, 29805, 29803, 29800, 29789, 29799, 29794, 29803, 29793, 29796, 29788, 29790,
29808, 29800, 29802, 29797, 29794, 29792, 29809, 29806, 29804, 29807, 29790, 29800, 29798, 29804, 29794, 29814, 29811, 29813, 29818, 29815, 29816, 29822, 29819, 29821, 29826, 29823, 29824, 29820, 29813, 29819, 29824, 29817, 29818, 29818, 29812, 29814, 29826, 29820, 29822, 29824, 29814, 29820, 29830, 29827, 29829, 29832, 29833, 29831, 29836, 29828, 29830, 29837, 29834, 29832, 29844, 29841, 29843, 29844, 29839, 29842, 29848, 29845, 29846, 29848, 29843, 29847, 29848, 29840, 29844, 29828, 29801, 29827, 29806, 29831, 29805, 29835, 29802, 29828, 29810, 29832, 29806, 29840, 29821, 29839, 29846, 29825, 29826, 29846, 29822, 29840, 29817, 29797, 29815, 29825, 29809, 29823, 29823, 29798, 29817, 29843, 29833, 29834, 29847, 29837, 29845, 29847, 29834, 29838, 29845, 29810, 29825, 29701, 29761, 29699, 29766, 29707, 29708, 29707, 29762, 29701, 29788, 29729, 29787, 29736, 29791, 29735, 29795, 29730, 29788, 29742, 29792, 29736, 29812, 29755, 29811, 29816, 29759, 29760, 29816, 29756, 29812, 29815, 29742, 29759, 29709, 29682, 29680, 29719, 29688, 29717, 29727, 29696, 29698, 29729, 29702, 29700, 29711, 29688, 29682, 29737, 29698, 29706, 29730, 29708, 29702, 29737, 29690, 29719, 29796, 29766, 29795, 29808, 29774, 29807, 29807, 29768, 29796, 29830, 29780, 29782,
29836, 29784, 29835, 29836, 29782, 29786, 29835, 29776, 29808, 29795, 29708, 29739, 29853, 29854, 29852, 29857, 29858, 29856, 29858, 29860, 29856, 29867, 29864, 29865, 29854, 29863, 29862, 29859, 29864, 29858, 29858, 29866, 29861, 29860, 29855, 29853, 29861, 29863, 29855, 29852, 29850, 29849, 29850, 29862, 29851, 29869, 29870, 29868, 29873, 29869, 29872, 29875, 29876, 29874, 29879, 29880, 29878, 29870, 29879, 29878, 29881, 29871, 29873, 29877, 29882, 29876, 29874, 29873, 29872, 29876, 29881, 29873, 29885, 29886, 29884, 29886, 29888, 29884, 29893, 29890, 29891, 29897, 29898, 29896, 29904, 29905, 29903, 29889, 29896, 29888, 29904, 29892, 29893, 29887, 29890, 29886, 29886, 29892, 29889, 29903, 29900, 29897, 29898, 29895, 29894, 29895, 29905, 29902, 29889, 29903, 29897, 29912, 29909, 29911, 29912, 29907, 29910, 29916, 29913, 29915, 29920, 29917, 29918, 29926, 29923, 29925, 29924, 29909, 29923, 29920, 29911, 29919, 29912, 29918, 29908, 29926, 29927, 29924, 29921, 29916, 29915, 29915, 29907, 29908, 29921, 29908, 29918, 29924, 29919, 29911, 29930, 29931, 29929, 29936, 29933, 29935, 29942, 29939, 29940, 29944, 29945, 29943, 29950, 29947, 29949, 29954, 29955, 29953, 29932, 29943, 29931, 29948, 29935, 29947, 29942, 29953, 29941, 29930, 29938, 29932,
29951, 29946, 29944, 29936, 29940, 29934, 29956, 29948, 29950, 29941, 29937, 29939, 29953, 29952, 29951, 29953, 29938, 29941, 29932, 29951, 29944, 29954, 29936, 29948, 29895, 29929, 29894, 29934, 29899, 29933, 29906, 29939, 29905, 29939, 29902, 29905, 29902, 29930, 29895, 29940, 29901, 29934, 29917, 29893, 29891, 29926, 29899, 29901, 29928, 29904, 29927, 29906, 29926, 29901, 29927, 29893, 29919, 29960, 29957, 29959, 29960, 29961, 29958, 29973, 29970, 29971, 29977, 29974, 29976, 29977, 29978, 29975, 29966, 29974, 29965, 29973, 29980, 29972, 29967, 29960, 29959, 29968, 29962, 29960, 29981, 29976, 29980, 29981, 29979, 29977, 29966, 29961, 29962, 29970, 29969, 29968, 29969, 29966, 29962, 29966, 29980, 29976, 29986, 29983, 29985, 29990, 29991, 29989, 29994, 29990, 29993, 29996, 29997, 29995, 29991, 29985, 29989, 29988, 29984, 29986, 29992, 29995, 29991, 29994, 29996, 29992, 29995, 29986, 29991, 30000, 30001, 29999, 30006, 30007, 30005, 30002, 30005, 30001, 30003, 30002, 30000, 30006, 30010, 30008, 30002, 30009, 30006, 30014, 30011, 30013, 30018, 30015, 30016, 30022, 30019, 30021, 30026, 30023, 30024, 30020, 30013, 30019, 30018, 30023, 30017, 30018, 30012, 30014, 30022, 30024, 30020, 30020, 30018, 30014, 29979, 29999, 29978, 29982, 30000, 29979,
30012, 29993, 30011, 29998, 30015, 29997, 30016, 29994, 30012, 29983, 29964, 29963, 29987, 29973, 29971, 29997, 29981, 29995, 29984, 29971, 29964, 29995, 29973, 29988, 30015, 30004, 30003, 30025, 30009, 30023, 30023, 30004, 30017, 29997, 30003, 29982, 29850, 29885, 29849, 29850, 29891, 29887, 29863, 29917, 29862, 29867, 29921, 29866, 29921, 29863, 29866, 29917, 29851, 29862, 29964, 29875, 29963, 29968, 29880, 29881, 29883, 29970, 29882, 29970, 29881, 29882, 29971, 29877, 29964, 29868, 29859, 29857, 29878, 29867, 29865, 29870, 29865, 29859, 29959, 29914, 29916, 29967, 29916, 29922, 29880, 29922, 29867, 30028, 30029, 30027, 30029, 30031, 30027, 30036, 30033, 30034, 30038, 30039, 30037, 30044, 30041, 30042, 30032, 30037, 30031, 30042, 30035, 30036, 30030, 30033, 30029, 30033, 30032, 30029, 30041, 30040, 30038, 30035, 30038, 30032, 30050, 30047, 30049, 30046, 30047, 30045, 30054, 30051, 30052, 30054, 30049, 30053, 30051, 30049, 30046, 30040, 30045, 30039, 30052, 30043, 30044, 30043, 30046, 30040, 30056, 30057, 30055, 30060, 30061, 30059, 30066, 30063, 30064, 30070, 30067, 30068, 30074, 30071, 30072, 30078, 30075, 30076, 30072, 30065, 30066, 30076, 30069, 30070, 30058, 30063, 30057, 30062, 30067, 30061, 30059, 30058, 30056, 30069, 30064, 30067,
30077, 30072, 30075, 30061, 30064, 30058, 30075, 30066, 30069, 30080, 30081, 30079, 30084, 30085, 30083, 30090, 30087, 30088, 30094, 30091, 30092, 30098, 30095, 30096, 30102, 30099, 30100, 30096, 30089, 30090, 30100, 30093, 30094, 30082, 30087, 30081, 30086, 30091, 30085, 30083, 30082, 30080, 30093, 30088, 30091, 30101, 30096, 30099, 30085, 30088, 30082, 30099, 30090, 30093, 30106, 30103, 30105, 30110, 30107, 30109, 30114, 30111, 30112, 30118, 30115, 30116, 30114, 30105, 30113, 30118, 30109, 30117, 30109, 30104, 30106, 30117, 30112, 30115, 30117, 30106, 30114, 30122, 30119, 30121, 30126, 30123, 30125, 30130, 30127, 30128, 30134, 30131, 30132, 30130, 30121, 30129, 30134, 30125, 30133, 30125, 30120, 30122, 30133, 30128, 30131, 30133, 30122, 30130, 30112, 30073, 30074, 30116, 30077, 30078, 30115, 30074, 30077, 30128, 30097, 30098, 30132, 30101, 30102, 30131, 30098, 30101, 30079, 30062, 30060, 30089, 30068, 30087, 30097, 30076, 30095, 30081, 30068, 30062, 30095, 30070, 30089, 30121, 30108, 30110, 30129, 30116, 30127, 30129, 30110, 30118, 30127, 30078, 30097, 30055, 30030, 30028, 30065, 30034, 30063, 30073, 30042, 30071, 30057, 30034, 30030, 30071, 30036, 30065, 30105, 30048, 30050, 30113, 30052, 30111, 30113, 30050, 30054, 30111, 30044, 30073,
30139, 30141, 30137, 30147, 30144, 30145, 30151, 30148, 30150, 30151, 30152, 30149, 30142, 30148, 30141, 30147, 30154, 30146, 30139, 30146, 30142, 30155, 30150, 30154, 30151, 30156, 30153, 30139, 30135, 30136, 30144, 30136, 30143, 30146, 30150, 30142, 30160, 30157, 30159, 30164, 30165, 30163, 30170, 30164, 30168, 30172, 30173, 30171, 30178, 30175, 30176, 30165, 30159, 30163, 30162, 30158, 30160, 30165, 30176, 30175, 30178, 30166, 30170, 30174, 30179, 30173, 30171, 30170, 30168, 30170, 30179, 30178, 30175, 30160, 30165, 30184, 30181, 30183, 30184, 30185, 30182, 30190, 30187, 30189, 30195, 30196, 30194, 30191, 30184, 30183, 30192, 30186, 30184, 30197, 30189, 30196, 30189, 30185, 30186, 30194, 30193, 30192, 30196, 30186, 30193, 30169, 30181, 30167, 30192, 30177, 30178, 30180, 30194, 30179, 30194, 30178, 30179, 30177, 30183, 30169, 30157, 30140, 30138, 30161, 30147, 30145, 30169, 30152, 30153, 30177, 30155, 30175, 30158, 30145, 30140, 30156, 30169, 30153, 30175, 30147, 30162, 30201, 30202, 30200, 30199, 30201, 30198, 30210, 30211, 30209, 30208, 30210, 30207, 30221, 30222, 30220, 30202, 30213, 30200, 30221, 30205, 30206, 30212, 30218, 30211, 30208, 30219, 30212, 30202, 30206, 30205, 30199, 30206, 30203, 30213, 30208, 30207, 30222, 30217, 30220,
30214, 30217, 30208, 30205, 30214, 30202, 30229, 30226, 30228, 30229, 30224, 30227, 30233, 30230, 30231, 30237, 30234, 30236, 30235, 30226, 30234, 30232, 30229, 30228, 30233, 30225, 30229, 30237, 30238, 30235, 30238, 30228, 30235, 30245, 30242, 30244, 30245, 30240, 30243, 30249, 30246, 30248, 30253, 30250, 30251, 30253, 30244, 30252, 30245, 30251, 30241, 30254, 30249, 30248, 30248, 30240, 30241, 30254, 30241, 30251, 30249, 30215, 30247, 30219, 30250, 30218, 30255, 30222, 30223, 30222, 30251, 30219, 30249, 30223, 30216, 30230, 30206, 30204, 30237, 30215, 30216, 30239, 30221, 30238, 30223, 30237, 30216, 30232, 30221, 30206, 30259, 30260, 30258, 30266, 30267, 30265, 30272, 30269, 30271, 30278, 30275, 30276, 30257, 30265, 30256, 30276, 30263, 30264, 30273, 30268, 30266, 30261, 30263, 30260, 30278, 30271, 30277, 30258, 30257, 30256, 30271, 30267, 30268, 30277, 30273, 30275, 30260, 30262, 30257, 30277, 30268, 30274, 30275, 30262, 30263, 30262, 30266, 30257, 30280, 30281, 30279, 30284, 30285, 30283, 30294, 30291, 30293, 30298, 30295, 30297, 30302, 30299, 30300, 30306, 30303, 30304, 30300, 30287, 30288, 30304, 30289, 30290, 30282, 30287, 30281, 30302, 30293, 30301, 30286, 30289, 30285, 30306, 30297, 30305, 30283, 30282, 30280, 30297, 30292, 30294,
30305, 30300, 30303, 30285, 30288, 30282, 30305, 30294, 30302, 30303, 30288, 30289, 30279, 30261, 30259, 30293, 30270, 30272, 30301, 30276, 30299, 30281, 30264, 30261, 30301, 30272, 30278, 30299, 30264, 30287, 30308, 30309, 30307, 30312, 30313, 30311, 30322, 30319, 30321, 30326, 30323, 30325, 30330, 30327, 30328, 30332, 30333, 30331, 30328, 30315, 30316, 30332, 30317, 30318, 30310, 30315, 30309, 30330, 30321, 30329, 30314, 30317, 30313, 30334, 30325, 30333, 30311, 30310, 30308, 30325, 30320, 30322, 30333, 30328, 30331, 30313, 30316, 30310, 30333, 30322, 30330, 30331, 30316, 30317, 30336, 30337, 30335, 30344, 30341, 30343, 30348, 30345, 30346, 30342, 30336, 30341, 30346, 30339, 30340, 30338, 30339, 30337, 30348, 30342, 30344, 30346, 30338, 30342, 30335, 30314, 30312, 30345, 30334, 30332, 30337, 30318, 30314, 30345, 30318, 30339, 30209, 30174, 30172, 30174, 30218, 30180, 30244, 30188, 30190, 30252, 30195, 30250, 30197, 30244, 30190, 30250, 30180, 30218, 30307, 30286, 30284, 30321, 30296, 30298, 30329, 30304, 30327, 30309, 30290, 30286, 30329, 30298, 30306, 30327, 30290, 30315, 29946, 30135, 29945, 30140, 29949, 30138, 30145, 29955, 29956, 29955, 30143, 29952, 29946, 30143, 30136, 30140, 29956, 29950, 30008, 30198, 30007, 30008, 30204, 30199,
30225, 30021, 30224, 30231, 30025, 30026, 30225, 30026, 30022, 30230, 30010, 30025, 30356, 30357, 30355, 30362, 30359, 30361, 30350, 30355, 30349, 30360, 30351, 30359, 30356, 30364, 30358, 30366, 30360, 30362, 30351, 30350, 30349, 30366, 30363, 30365, 30350, 30354, 30353, 30354, 30363, 30353, 30350, 30363, 30356, 30365, 30352, 30360, 30370, 30371, 30369, 30368, 30370, 30367, 30371, 30375, 30374, 30373, 30372, 30368, 30379, 30380, 30378, 30377, 30379, 30376, 30385, 30382, 30384, 30394, 30395, 30393, 30398, 30394, 30397, 30400, 30401, 30399, 30404, 30405, 30403, 30380, 30391, 30378, 30395, 30384, 30393, 30400, 30387, 30388, 30380, 30388, 30387, 30386, 30381, 30377, 30390, 30383, 30385, 30396, 30403, 30395, 30406, 30396, 30398, 30390, 30386, 30389, 30405, 30400, 30403, 30390, 30400, 30388, 30387, 30392, 30380, 30395, 30390, 30385, 30358, 30376, 30357, 30383, 30361, 30382, 30389, 30364, 30366, 30358, 30386, 30377, 30389, 30362, 30383, 30391, 30368, 30367, 30401, 30373, 30399, 30368, 30399, 30373, 30408, 30409, 30407, 30414, 30415, 30413, 30410, 30413, 30409, 30411, 30410, 30408, 30414, 30418, 30416, 30410, 30417, 30414, 30422, 30419, 30421, 30426, 30423, 30424, 30430, 30427, 30429, 30434, 30431, 30432, 30428, 30421, 30427, 30432, 30425, 30426,
30426, 30420, 30422, 30434, 30428, 30430, 30432, 30422, 30428, 30436, 30437, 30435, 30444, 30445, 30443, 30442, 30444, 30441, 30438, 30441, 30437, 30436, 30440, 30438, 30445, 30449, 30448, 30442, 30449, 30446, 30438, 30447, 30442, 30453, 30450, 30452, 30455, 30456, 30454, 30461, 30458, 30460, 30465, 30462, 30463, 30459, 30452, 30458, 30463, 30456, 30457, 30453, 30455, 30451, 30461, 30463, 30459, 30459, 30457, 30453, 30416, 30435, 30415, 30416, 30439, 30436, 30451, 30429, 30450, 30455, 30433, 30434, 30455, 30430, 30451, 30425, 30411, 30423, 30433, 30417, 30431, 30425, 30417, 30412, 30456, 30439, 30454, 30464, 30447, 30462, 30462, 30440, 30456, 30454, 30418, 30433, 30469, 30470, 30468, 30476, 30477, 30475, 30484, 30481, 30483, 30467, 30479, 30466, 30484, 30471, 30482, 30478, 30486, 30477, 30475, 30472, 30469, 30486, 30484, 30483, 30470, 30467, 30466, 30483, 30479, 30480, 30467, 30477, 30474, 30485, 30483, 30480, 30486, 30474, 30477, 30474, 30480, 30467, 30484, 30478, 30473, 30491, 30488, 30490, 30494, 30488, 30492, 30491, 30493, 30489, 30507, 30504, 30505, 30511, 30508, 30510, 30499, 30508, 30498, 30507, 30512, 30506, 30501, 30490, 30500, 30494, 30500, 30490, 30491, 30503, 30495, 30513, 30510, 30512, 30499, 30493, 30495, 30504, 30503, 30501,
30506, 30495, 30503, 30506, 30510, 30499, 30494, 30471, 30473, 30502, 30476, 30500, 30502, 30473, 30478, 30515, 30516, 30514, 30521, 30522, 30520, 30525, 30526, 30524, 30531, 30528, 30530, 30535, 30532, 30534, 30537, 30522, 30523, 30539, 30526, 30527, 30517, 30520, 30516, 30537, 30530, 30536, 30538, 30535, 30534, 30518, 30517, 30515, 30526, 30521, 30524, 30534, 30529, 30531, 30517, 30524, 30521, 30537, 30534, 30531, 30526, 30537, 30523, 30543, 30540, 30542, 30545, 30540, 30544, 30550, 30544, 30549, 30546, 30543, 30542, 30548, 30542, 30545, 30551, 30545, 30550, 30548, 30525, 30546, 30550, 30533, 30535, 30539, 30550, 30535, 30551, 30527, 30548, 30371, 30468, 30369, 30476, 30374, 30375, 30371, 30475, 30469, 30497, 30397, 30496, 30501, 30401, 30402, 30406, 30504, 30405, 30405, 30501, 30402, 30505, 30398, 30497, 30401, 30476, 30375, 30445, 30518, 30443, 30525, 30448, 30449, 30448, 30519, 30445, 30543, 30460, 30541, 30547, 30464, 30465, 30547, 30461, 30543, 30464, 30525, 30449, 30514, 30497, 30496, 30520, 30507, 30505, 30530, 30509, 30511, 30516, 30505, 30497, 30536, 30511, 30513, 30522, 30513, 30507, 30553, 30554, 30552, 30554, 30556, 30552, 30561, 30558, 30559, 30563, 30564, 30562, 30569, 30566, 30567, 30557, 30562, 30556, 30567, 30560, 30561,
30555, 30558, 30554, 30558, 30557, 30554, 30566, 30565, 30563, 30560, 30563, 30557, 30571, 30572, 30570, 30577, 30574, 30575, 30579, 30580, 30578, 30585, 30582, 30583, 30573, 30578, 30572, 30583, 30576, 30577, 30574, 30573, 30571, 30582, 30581, 30579, 30576, 30579, 30573, 30565, 30570, 30564, 30575, 30568, 30569, 30568, 30571, 30565, 30587, 30588, 30586, 30591, 30592, 30590, 30597, 30594, 30595, 30601, 30598, 30599, 30605, 30602, 30603, 30609, 30606, 30607, 30603, 30596, 30597, 30607, 30600, 30601, 30589, 30594, 30588, 30593, 30598, 30592, 30590, 30589, 30587, 30600, 30595, 30598, 30608, 30603, 30606, 30592, 30595, 30589, 30606, 30597, 30600, 30611, 30612, 30610, 30615, 30616, 30614, 30621, 30618, 30619, 30625, 30622, 30623, 30629, 30626, 30627, 30633, 30630, 30631, 30627, 30620, 30621, 30631, 30624, 30625, 30613, 30618, 30612, 30617, 30622, 30616, 30614, 30613, 30611, 30624, 30619, 30622, 30632, 30627, 30630, 30616, 30619, 30613, 30630, 30621, 30624, 30637, 30634, 30635, 30641, 30638, 30639, 30645, 30642, 30643, 30649, 30646, 30647, 30643, 30636, 30637, 30647, 30640, 30641, 30640, 30635, 30638, 30648, 30643, 30646, 30646, 30637, 30640, 30653, 30650, 30652, 30657, 30654, 30656, 30661, 30658, 30659, 30665, 30662, 30663, 30669, 30666, 30668,
30667, 30650, 30666, 30661, 30652, 30660, 30671, 30667, 30669, 30665, 30656, 30664, 30656, 30651, 30653, 30664, 30659, 30662, 30664, 30653, 30661, 30670, 30652, 30667, 30635, 30604, 30605, 30639, 30608, 30609, 30638, 30605, 30608, 30659, 30628, 30629, 30663, 30632, 30633, 30662, 30629, 30632, 30610, 30593, 30591, 30620, 30599, 30618, 30628, 30607, 30626, 30612, 30599, 30593, 30626, 30601, 30620, 30660, 30639, 30658, 30671, 30647, 30670, 30670, 30641, 30660, 30658, 30609, 30628, 30673, 30674, 30672, 30679, 30676, 30677, 30681, 30682, 30680, 30687, 30684, 30685, 30675, 30680, 30674, 30685, 30678, 30679, 30676, 30675, 30673, 30684, 30683, 30681, 30678, 30681, 30675, 30689, 30690, 30688, 30695, 30692, 30693, 30697, 30698, 30696, 30703, 30700, 30701, 30691, 30696, 30690, 30701, 30694, 30695, 30692, 30691, 30689, 30700, 30699, 30697, 30694, 30697, 30691, 30683, 30688, 30682, 30693, 30686, 30687, 30686, 30689, 30683, 30707, 30704, 30705, 30711, 30708, 30709, 30715, 30712, 30713, 30719, 30716, 30717, 30713, 30706, 30707, 30717, 30710, 30711, 30710, 30705, 30708, 30718, 30713, 30716, 30716, 30707, 30710, 30723, 30720, 30722, 30729, 30726, 30728, 30727, 30722, 30726, 30725, 30721, 30723, 30731, 30727, 30729, 30730, 30723, 30727, 30735, 30732, 30733,
30739, 30736, 30737, 30743, 30740, 30741, 30747, 30744, 30745, 30741, 30734, 30735, 30745, 30738, 30739, 30738, 30733, 30736, 30746, 30741, 30744, 30744, 30735, 30738, 30751, 30748, 30750, 30757, 30754, 30756, 30755, 30750, 30754, 30753, 30749, 30751, 30759, 30755, 30757, 30758, 30751, 30755, 30733, 30714, 30715, 30737, 30718, 30719, 30736, 30715, 30718, 30749, 30728, 30748, 30752, 30729, 30749, 30725, 30709, 30724, 30731, 30717, 30730, 30730, 30711, 30725, 30753, 30737, 30752, 30759, 30745, 30758, 30758, 30739, 30753, 30752, 30719, 30731, 30581, 30672, 30580, 30677, 30584, 30585, 30584, 30673, 30581, 30705, 30644, 30645, 30709, 30648, 30649, 30708, 30645, 30648, 30721, 30668, 30720, 30724, 30669, 30721, 30724, 30649, 30671, 30586, 30555, 30553, 30596, 30559, 30594, 30604, 30567, 30602, 30588, 30559, 30555, 30602, 30561, 30596, 30636, 30575, 30634, 30644, 30583, 30642, 30642, 30577, 30636, 30634, 30569, 30604, 30706, 30677, 30704, 30714, 30685, 30712, 30712, 30679, 30706, 30734, 30693, 30732, 30742, 30701, 30740, 30740, 30695, 30734, 30732, 30687, 30714, 30704, 30585, 30644, 30761, 30762, 30760, 30767, 30764, 30765, 30769, 30770, 30768, 30775, 30772, 30773, 30763, 30768, 30762, 30773, 30766, 30767, 30764, 30763, 30761, 30772, 30771, 30769,
30766, 30769, 30763, 30777, 30778, 30776, 30783, 30780, 30781, 30785, 30786, 30784, 30791, 30788, 30789, 30779, 30784, 30778, 30789, 30782, 30783, 30780, 30779, 30777, 30788, 30787, 30785, 30782, 30785, 30779, 30771, 30776, 30770, 30781, 30774, 30775, 30774, 30777, 30771, 30795, 30792, 30793, 30799, 30796, 30797, 30803, 30800, 30801, 30807, 30804, 30805, 30801, 30794, 30795, 30805, 30798, 30799, 30798, 30793, 30796, 30806, 30801, 30804, 30804, 30795, 30798, 30811, 30808, 30810, 30817, 30814, 30816, 30815, 30810, 30814, 30813, 30809, 30811, 30819, 30815, 30817, 30818, 30811, 30815, 30823, 30820, 30821, 30827, 30824, 30825, 30831, 30828, 30829, 30835, 30832, 30833, 30829, 30822, 30823, 30833, 30826, 30827, 30826, 30821, 30824, 30834, 30829, 30832, 30832, 30823, 30826, 30839, 30836, 30838, 30845, 30842, 30844, 30843, 30838, 30842, 30841, 30837, 30839, 30847, 30843, 30845, 30846, 30839, 30843, 30821, 30802, 30803, 30825, 30806, 30807, 30824, 30803, 30806, 30837, 30816, 30836, 30840, 30817, 30837, 30813, 30797, 30812, 30819, 30805, 30818, 30818, 30799, 30813, 30841, 30825, 30840, 30847, 30833, 30846, 30846, 30827, 30841, 30840, 30807, 30819, 30849, 30850, 30848, 30855, 30852, 30853, 30857, 30858, 30856, 30863, 30860, 30861, 30851, 30856, 30850,
30861, 30854, 30855, 30852, 30851, 30849, 30860, 30859, 30857, 30854, 30857, 30851, 30867, 30864, 30866, 30869, 30866, 30868, 30859, 30864, 30858, 30869, 30862, 30863, 30862, 30866, 30859, 30873, 30870, 30871, 30877, 30874, 30875, 30881, 30878, 30879, 30885, 30882, 30883, 30879, 30872, 30873, 30883, 30876, 30877, 30876, 30871, 30874, 30884, 30879, 30882, 30882, 30873, 30876, 30889, 30886, 30888, 30895, 30892, 30894, 30893, 30888, 30892, 30891, 30887, 30889, 30897, 30893, 30895, 30896, 30889, 30893, 30901, 30898, 30900, 30905, 30902, 30904, 30907, 30900, 30906, 30909, 30904, 30908, 30904, 30899, 30901, 30908, 30901, 30907, 30907, 30880, 30881, 30909, 30884, 30885, 30908, 30881, 30884, 30911, 30894, 30910, 30912, 30895, 30911, 30891, 30875, 30890, 30897, 30883, 30896, 30896, 30877, 30891, 30911, 30903, 30905, 30912, 30905, 30909, 30912, 30885, 30897, 30787, 30848, 30786, 30853, 30790, 30791, 30790, 30849, 30787, 30871, 30830, 30831, 30875, 30834, 30835, 30874, 30831, 30834, 30887, 30844, 30886, 30890, 30845, 30887, 30890, 30835, 30847, 30794, 30765, 30792, 30802, 30773, 30800, 30800, 30767, 30794, 30822, 30781, 30820, 30830, 30789, 30828, 30828, 30783, 30822, 30820, 30775, 30802, 30872, 30853, 30870, 30880, 30861, 30878, 30878, 30855, 30872,
30900, 30865, 30867, 30906, 30867, 30869, 30906, 30863, 30880, 30870, 30791, 30830, 30699, 30760, 30698, 30765, 30702, 30703, 30702, 30761, 30699, 30793, 30742, 30743, 30797, 30746, 30747, 30796, 30743, 30746, 30809, 30756, 30808, 30812, 30757, 30809, 30812, 30747, 30759, 30792, 30703, 30742, 30914, 30915, 30913, 30918, 30919, 30917, 30926, 30923, 30924, 30928, 30929, 30927, 30934, 30931, 30933, 30940, 30937, 30938, 30916, 30927, 30915, 30926, 30937, 30925, 30921, 30916, 30914, 30935, 30930, 30928, 30920, 30923, 30919, 30940, 30932, 30934, 30917, 30914, 30913, 30925, 30921, 30923, 30939, 30935, 30937, 30919, 30921, 30914, 30937, 30922, 30925, 30922, 30928, 30916, 30942, 30943, 30941, 30946, 30947, 30945, 30952, 30949, 30950, 30956, 30953, 30954, 30960, 30957, 30959, 30964, 30961, 30963, 30966, 30951, 30952, 30968, 30955, 30956, 30944, 30949, 30943, 30966, 30959, 30965, 30948, 30953, 30947, 30968, 30963, 30967, 30945, 30944, 30942, 30955, 30950, 30953, 30963, 30958, 30960, 30947, 30950, 30944, 30967, 30960, 30966, 30967, 30952, 30955, 30970, 30971, 30969, 30976, 30973, 30975, 30982, 30979, 30980, 30984, 30985, 30983, 30990, 30987, 30989, 30994, 30995, 30993, 30972, 30983, 30971, 30988, 30975, 30987, 30994, 30981, 30982, 30977, 30972, 30970,
30991, 30986, 30984, 30982, 30974, 30976, 30996, 30988, 30990, 30979, 30978, 30977, 30995, 30991, 30993, 30981, 30991, 30978, 30978, 30984, 30972, 30994, 30976, 30988, 30998, 30999, 30997, 30999, 31001, 30997, 31006, 31003, 31004, 31008, 31009, 31007, 31014, 31011, 31012, 31002, 31007, 31001, 31012, 31005, 31006, 31000, 31003, 30999, 31003, 31002, 30999, 31008, 31013, 31010, 31005, 31008, 31002, 30930, 30969, 30929, 30974, 30933, 30973, 30980, 30939, 30940, 30979, 30936, 30939, 30936, 30970, 30930, 30980, 30934, 30974, 30941, 30920, 30918, 30951, 30924, 30949, 30959, 30931, 30932, 30943, 30924, 30920, 30965, 30932, 30938, 30951, 30938, 30926, 31016, 31017, 31015, 31020, 31021, 31019, 31026, 31023, 31024, 31030, 31027, 31028, 31034, 31031, 31033, 31038, 31035, 31037, 31038, 31039, 31036, 31042, 31025, 31026, 31030, 31043, 31029, 31018, 31023, 31017, 31042, 31033, 31041, 31022, 31027, 31021, 31044, 31037, 31043, 31044, 31040, 31038, 31019, 31018, 31016, 31029, 31024, 31027, 31037, 31032, 31034, 31021, 31024, 31018, 31043, 31034, 31042, 31043, 31026, 31029, 31047, 31048, 31046, 31051, 31047, 31050, 31055, 31052, 31053, 31059, 31056, 31058, 31063, 31060, 31061, 31057, 31050, 31056, 31061, 31054, 31055, 31049, 31052, 31048, 31055, 31049, 31051,
31063, 31057, 31059, 31061, 31051, 31057, 31067, 31064, 31066, 31069, 31070, 31068, 31079, 31076, 31078, 31081, 31082, 31080, 31077, 31066, 31076, 31071, 31080, 31070, 31073, 31065, 31067, 31085, 31077, 31079, 31074, 31071, 31069, 31081, 31087, 31083, 31084, 31067, 31077, 31071, 31086, 31081, 31091, 31088, 31090, 31095, 31092, 31093, 31099, 31096, 31098, 31101, 31102, 31100, 31097, 31090, 31096, 31101, 31094, 31095, 31095, 31089, 31091, 31099, 31101, 31097, 31101, 31091, 31097, 31040, 31068, 31039, 31045, 31069, 31040, 31089, 31058, 31088, 31093, 31062, 31063, 31093, 31059, 31089, 31046, 31022, 31020, 31054, 31028, 31052, 31060, 31045, 31044, 31048, 31028, 31022, 31054, 31044, 31030, 31094, 31074, 31092, 31100, 31087, 31086, 31100, 31075, 31094, 31092, 31045, 31062, 31105, 31106, 31104, 31113, 31108, 31111, 31119, 31116, 31117, 31122, 31124, 31120, 31129, 31126, 31127, 31112, 31120, 31110, 31123, 31111, 31121, 31119, 31126, 31118, 31105, 31115, 31107, 31119, 31109, 31113, 31126, 31125, 31122, 31112, 31106, 31107, 31116, 31115, 31114, 31118, 31107, 31115, 31112, 31126, 31122, 31123, 31119, 31113, 31131, 31132, 31130, 31137, 31134, 31135, 31141, 31142, 31140, 31142, 31138, 31140, 31145, 31146, 31144, 31154, 31151, 31152, 31133, 31144, 31132,
31152, 31136, 31137, 31143, 31149, 31142, 31142, 31148, 31139, 31131, 31136, 31133, 31145, 31153, 31147, 31146, 31143, 31141, 31143, 31153, 31150, 31133, 31151, 31145, 31158, 31155, 31157, 31160, 31157, 31159, 31166, 31163, 31165, 31162, 31163, 31161, 31170, 31167, 31169, 31172, 31173, 31171, 31176, 31177, 31175, 31173, 31166, 31165, 31171, 31165, 31162, 31178, 31169, 31177, 31169, 31164, 31166, 31177, 31172, 31175, 31174, 31169, 31166, 31125, 31155, 31124, 31129, 31159, 31128, 31128, 31157, 31125, 31172, 31149, 31150, 31154, 31175, 31153, 31175, 31150, 31153, 31138, 31123, 31121, 31148, 31129, 31127, 31139, 31127, 31123, 31162, 31156, 31158, 31160, 31162, 31158, 31171, 31129, 31149, 31182, 31179, 31181, 31184, 31185, 31183, 31192, 31193, 31191, 31196, 31192, 31195, 31198, 31199, 31197, 31199, 31201, 31197, 31204, 31205, 31203, 31210, 31207, 31208, 31193, 31181, 31191, 31186, 31198, 31185, 31188, 31180, 31182, 31194, 31203, 31193, 31196, 31204, 31194, 31184, 31190, 31186, 31199, 31208, 31207, 31207, 31202, 31199, 31201, 31196, 31195, 31196, 31209, 31206, 31193, 31188, 31182, 31190, 31200, 31186, 31214, 31211, 31213, 31218, 31215, 31216, 31222, 31219, 31221, 31224, 31219, 31223, 31222, 31213, 31220, 31226, 31217, 31218, 31214, 31216, 31212,
31225, 31222, 31221, 31227, 31221, 31224, 31226, 31214, 31222, 31231, 31228, 31230, 31235, 31232, 31234, 31237, 31232, 31236, 31239, 31240, 31238, 31241, 31230, 31240, 31242, 31235, 31234, 31237, 31242, 31234, 31237, 31229, 31231, 31244, 31239, 31242, 31244, 31231, 31241, 31239, 31205, 31206, 31210, 31242, 31209, 31209, 31239, 31206, 31246, 31223, 31245, 31246, 31227, 31224, 31217, 31189, 31215, 31225, 31210, 31208, 31217, 31208, 31190, 31246, 31233, 31235, 31243, 31246, 31235, 31247, 31210, 31227, 30986, 31104, 30985, 31109, 30989, 31108, 31117, 30995, 30996, 30995, 31114, 30992, 30992, 31105, 30986, 31117, 30990, 31109, 31010, 31130, 31009, 31135, 31013, 31014, 31010, 31134, 31131, 31180, 31078, 31179, 31083, 31183, 31082, 31187, 31079, 31180, 31083, 31189, 31184, 31212, 31098, 31211, 31103, 31215, 31102, 31212, 31103, 31099, 31215, 31087, 31102, 31015, 30948, 30946, 31025, 30954, 31023, 31033, 30962, 30964, 31017, 30954, 30948, 31041, 30964, 30968, 31041, 30956, 31025, 31064, 31000, 30998, 31072, 31006, 31004, 31085, 31012, 31084, 31065, 31004, 31000, 31073, 31012, 31006, 31187, 31137, 31135, 31205, 31152, 31203, 31188, 31152, 31137, 31230, 31168, 31170, 31240, 31176, 31238, 31240, 31170, 31178, 31238, 31154, 31205, 31085, 31135, 31014,
31257, 31258, 31256, 31255, 31257, 31254, 31261, 31262, 31260, 31270, 31267, 31268, 31252, 31261, 31251, 31258, 31266, 31265, 31255, 31266, 31259, 31263, 31267, 31262, 31260, 31255, 31254, 31269, 31264, 31267, 31255, 31267, 31264, 31252, 31268, 31263, 31274, 31275, 31273, 31275, 31277, 31273, 31283, 31280, 31281, 31287, 31288, 31286, 31285, 31287, 31284, 31296, 31293, 31294, 31272, 31284, 31271, 31283, 31293, 31282, 31288, 31292, 31291, 31285, 31292, 31289, 31276, 31280, 31275, 31280, 31278, 31275, 31277, 31272, 31271, 31295, 31290, 31293, 31272, 31282, 31279, 31293, 31279, 31282, 31272, 31290, 31285, 31300, 31297, 31298, 31304, 31301, 31302, 31308, 31305, 31306, 31312, 31309, 31310, 31306, 31299, 31300, 31310, 31303, 31304, 31303, 31298, 31301, 31311, 31306, 31309, 31309, 31300, 31303, 31249, 31274, 31248, 31249, 31281, 31276, 31298, 31265, 31266, 31302, 31269, 31270, 31269, 31298, 31266, 31256, 31249, 31248, 31249, 31265, 31250, 31297, 31283, 31281, 31307, 31294, 31305, 31305, 31283, 31299, 31297, 31250, 31265, 31317, 31318, 31316, 31314, 31316, 31313, 31317, 31321, 31319, 31314, 31320, 31317, 31323, 31324, 31322, 31331, 31332, 31330, 31329, 31331, 31328, 31325, 31328, 31324, 31323, 31327, 31325, 31332, 31336, 31335, 31334, 31333, 31329,
31325, 31334, 31329, 31319, 31322, 31318, 31319, 31326, 31323, 31338, 31339, 31337, 31344, 31341, 31342, 31348, 31345, 31346, 31352, 31349, 31350, 31356, 31353, 31354, 31344, 31349, 31343, 31354, 31347, 31348, 31338, 31343, 31340, 31347, 31342, 31345, 31355, 31350, 31353, 31353, 31344, 31347, 31362, 31359, 31361, 31362, 31357, 31360, 31366, 31363, 31364, 31370, 31367, 31368, 31374, 31371, 31373, 31380, 31377, 31378, 31376, 31359, 31375, 31366, 31377, 31365, 31374, 31378, 31372, 31370, 31361, 31369, 31370, 31358, 31362, 31367, 31366, 31364, 31376, 31371, 31372, 31378, 31376, 31372, 31381, 31366, 31369, 31381, 31361, 31376, 31385, 31382, 31383, 31389, 31386, 31387, 31393, 31390, 31391, 31397, 31394, 31395, 31391, 31384, 31385, 31395, 31388, 31389, 31388, 31383, 31386, 31396, 31391, 31394, 31394, 31385, 31388, 31401, 31398, 31400, 31403, 31398, 31402, 31410, 31407, 31409, 31408, 31402, 31407, 31404, 31401, 31400, 31406, 31400, 31403, 31410, 31411, 31408, 31408, 31406, 31403, 31383, 31351, 31352, 31387, 31355, 31356, 31386, 31352, 31355, 31401, 31373, 31399, 31405, 31379, 31380, 31405, 31374, 31401, 31365, 31346, 31363, 31379, 31354, 31377, 31377, 31348, 31365, 31406, 31387, 31404, 31412, 31395, 31411, 31411, 31389, 31406, 31379, 31387, 31356,
31288, 31337, 31286, 31342, 31291, 31292, 31346, 31295, 31296, 31295, 31342, 31292, 31288, 31341, 31338, 31364, 31307, 31308, 31312, 31367, 31311, 31367, 31308, 31311, 31363, 31296, 31307, 31339, 31314, 31313, 31351, 31320, 31349, 31314, 31343, 31315, 31349, 31315, 31343, 31384, 31326, 31382, 31390, 31336, 31334, 31390, 31327, 31384, 31382, 31321, 31351, 31414, 31415, 31413, 31418, 31419, 31417, 31426, 31427, 31425, 31423, 31416, 31414, 31420, 31425, 31419, 31417, 31414, 31413, 31427, 31423, 31425, 31419, 31423, 31414, 31430, 31431, 31429, 31436, 31433, 31435, 31438, 31439, 31437, 31444, 31441, 31443, 31448, 31445, 31446, 31452, 31449, 31450, 31432, 31445, 31431, 31448, 31435, 31447, 31440, 31449, 31439, 31452, 31443, 31451, 31437, 31432, 31430, 31443, 31434, 31436, 31451, 31446, 31449, 31439, 31446, 31432, 31451, 31436, 31448, 31429, 31420, 31418, 31435, 31421, 31422, 31445, 31428, 31426, 31431, 31426, 31420, 31447, 31422, 31428, 31454, 31455, 31453, 31460, 31457, 31459, 31462, 31463, 31461, 31468, 31465, 31467, 31472, 31469, 31470, 31476, 31473, 31474, 31456, 31469, 31455, 31472, 31459, 31471, 31464, 31473, 31463, 31476, 31467, 31475, 31461, 31456, 31454, 31467, 31458, 31460, 31475, 31470, 31473, 31463, 31470, 31456, 31475, 31460, 31472,
31478, 31479, 31477, 31484, 31481, 31483, 31484, 31478, 31482, 31488, 31485, 31486, 31480, 31485, 31479, 31488, 31483, 31487, 31488, 31480, 31484, 31477, 31464, 31462, 31483, 31466, 31468, 31487, 31474, 31485, 31479, 31474, 31464, 31487, 31468, 31476, 31490, 31491, 31489, 31496, 31493, 31495, 31502, 31499, 31500, 31504, 31505, 31503, 31510, 31507, 31509, 31516, 31513, 31514, 31492, 31503, 31491, 31508, 31495, 31507, 31514, 31501, 31502, 31497, 31492, 31490, 31511, 31506, 31504, 31502, 31494, 31496, 31516, 31508, 31510, 31501, 31497, 31499, 31515, 31511, 31513, 31513, 31498, 31501, 31498, 31504, 31492, 31514, 31496, 31508, 31518, 31519, 31517, 31524, 31521, 31523, 31530, 31527, 31528, 31532, 31533, 31531, 31540, 31537, 31539, 31540, 31535, 31538, 31546, 31543, 31544, 31520, 31531, 31519, 31536, 31523, 31535, 31544, 31529, 31530, 31525, 31520, 31518, 31541, 31534, 31532, 31530, 31522, 31524, 31546, 31539, 31545, 31546, 31536, 31540, 31529, 31525, 31527, 31539, 31533, 31534, 31545, 31541, 31543, 31545, 31534, 31542, 31543, 31526, 31529, 31526, 31532, 31520, 31544, 31524, 31536, 31506, 31517, 31505, 31522, 31509, 31521, 31528, 31515, 31516, 31527, 31512, 31515, 31512, 31518, 31506, 31528, 31510, 31522, 31416, 31489, 31415, 31494, 31421, 31493,
31428, 31499, 31427, 31499, 31424, 31427, 31424, 31490, 31416, 31500, 31422, 31494, 31453, 31440, 31438, 31459, 31442, 31444, 31471, 31450, 31469, 31455, 31450, 31440, 31471, 31444, 31452, 31548, 31549, 31547, 31554, 31551, 31552, 31556, 31557, 31555, 31560, 31561, 31559, 31550, 31555, 31549, 31560, 31553, 31554, 31548, 31553, 31550, 31556, 31561, 31558, 31550, 31559, 31556, 31564, 31565, 31563, 31568, 31569, 31567, 31576, 31573, 31575, 31572, 31573, 31571, 31578, 31579, 31577, 31566, 31571, 31565, 31570, 31577, 31569, 31567, 31566, 31564, 31580, 31575, 31579, 31572, 31579, 31575, 31569, 31572, 31566, 31558, 31563, 31557, 31562, 31567, 31561, 31561, 31564, 31558, 31584, 31581, 31582, 31588, 31585, 31586, 31592, 31589, 31590, 31596, 31593, 31594, 31590, 31583, 31584, 31594, 31587, 31588, 31587, 31582, 31585, 31595, 31590, 31593, 31593, 31584, 31587, 31600, 31597, 31599, 31606, 31603, 31605, 31604, 31599, 31603, 31600, 31601, 31598, 31606, 31607, 31604, 31604, 31602, 31600, 31612, 31609, 31610, 31616, 31613, 31614, 31620, 31617, 31618, 31624, 31621, 31622, 31618, 31611, 31612, 31622, 31615, 31616, 31615, 31610, 31613, 31623, 31618, 31621, 31621, 31612, 31615, 31628, 31625, 31627, 31634, 31631, 31633, 31632, 31627, 31631, 31630, 31626, 31628,
31636, 31632, 31634, 31635, 31628, 31632, 31610, 31591, 31592, 31614, 31595, 31596, 31613, 31592, 31595, 31626, 31605, 31625, 31629, 31606, 31626, 31602, 31586, 31601, 31608, 31594, 31607, 31607, 31588, 31602, 31629, 31616, 31614, 31635, 31624, 31622, 31630, 31622, 31616, 31608, 31614, 31596, 31638, 31639, 31637, 31644, 31645, 31643, 31640, 31643, 31639, 31641, 31640, 31638, 31647, 31646, 31644, 31642, 31644, 31640, 31646, 31649, 31645, 31646, 31651, 31650, 31655, 31652, 31653, 31659, 31656, 31657, 31663, 31660, 31661, 31667, 31664, 31665, 31661, 31654, 31655, 31665, 31658, 31659, 31658, 31653, 31656, 31666, 31661, 31664, 31664, 31655, 31658, 31671, 31672, 31670, 31672, 31668, 31670, 31680, 31677, 31679, 31684, 31685, 31683, 31678, 31671, 31677, 31676, 31683, 31675, 31673, 31675, 31672, 31675, 31669, 31672, 31686, 31678, 31680, 31681, 31680, 31679, 31680, 31687, 31686, 31684, 31673, 31678, 31689, 31690, 31688, 31693, 31694, 31692, 31699, 31696, 31697, 31705, 31702, 31704, 31701, 31702, 31700, 31707, 31708, 31706, 31713, 31710, 31711, 31691, 31700, 31690, 31695, 31706, 31694, 31711, 31698, 31699, 31689, 31694, 31691, 31708, 31705, 31704, 31706, 31704, 31701, 31698, 31693, 31696, 31712, 31707, 31710, 31710, 31695, 31698, 31694, 31701, 31691,
31717, 31718, 31716, 31718, 31714, 31716, 31723, 31720, 31721, 31732, 31729, 31730, 31736, 31733, 31734, 31728, 31717, 31727, 31730, 31722, 31723, 31726, 31733, 31725, 31718, 31726, 31725, 31725, 31715, 31718, 31724, 31723, 31721, 31735, 31730, 31733, 31733, 31723, 31725, 31734, 31719, 31728, 31693, 31662, 31663, 31697, 31666, 31667, 31696, 31663, 31666, 31715, 31681, 31714, 31721, 31685, 31686, 31687, 31721, 31686, 31724, 31682, 31715, 31674, 31659, 31657, 31685, 31665, 31683, 31683, 31659, 31675, 31722, 31697, 31720, 31731, 31711, 31729, 31729, 31699, 31722, 31720, 31667, 31685, 31576, 31637, 31574, 31576, 31641, 31638, 31653, 31619, 31620, 31657, 31623, 31624, 31656, 31620, 31623, 31669, 31633, 31668, 31674, 31634, 31669, 31636, 31657, 31624, 31583, 31552, 31581, 31591, 31560, 31589, 31589, 31554, 31583, 31611, 31568, 31609, 31619, 31578, 31617, 31617, 31570, 31611, 31609, 31562, 31591, 31652, 31642, 31641, 31660, 31648, 31647, 31654, 31647, 31642, 31689, 31649, 31650, 31692, 31650, 31651, 31662, 31651, 31648, 31619, 31641, 31580, 31332, 31547, 31330, 31552, 31335, 31336, 31335, 31548, 31332, 31582, 31392, 31393, 31586, 31396, 31397, 31585, 31393, 31396, 31598, 31409, 31597, 31598, 31412, 31410, 31601, 31397, 31412, 31392, 31552, 31336,
30326, 30407, 30324, 30334, 30408, 30326, 30420, 30343, 30419, 30424, 30347, 30348, 30424, 30344, 30420, 30423, 30334, 30347, 31738, 31739, 31737, 31742, 31743, 31741, 31750, 31747, 31748, 31752, 31753, 31751, 31758, 31755, 31757, 31764, 31761, 31762, 31740, 31751, 31739, 31762, 31749, 31750, 31745, 31740, 31738, 31759, 31754, 31752, 31744, 31747, 31743, 31764, 31756, 31758, 31741, 31738, 31737, 31749, 31745, 31747, 31763, 31759, 31761, 31743, 31745, 31738, 31761, 31746, 31749, 31746, 31752, 31740, 31766, 31767, 31765, 31770, 31771, 31769, 31776, 31773, 31774, 31780, 31777, 31778, 31784, 31781, 31783, 31788, 31785, 31787, 31790, 31791, 31789, 31794, 31775, 31776, 31796, 31779, 31780, 31768, 31773, 31767, 31794, 31783, 31793, 31772, 31777, 31771, 31796, 31787, 31795, 31792, 31797, 31791, 31769, 31768, 31766, 31779, 31774, 31777, 31787, 31782, 31784, 31771, 31774, 31768, 31795, 31784, 31794, 31795, 31776, 31779, 31800, 31801, 31799, 31806, 31803, 31805, 31812, 31809, 31810, 31814, 31815, 31813, 31820, 31817, 31819, 31826, 31823, 31824, 31802, 31813, 31801, 31818, 31805, 31817, 31824, 31811, 31812, 31807, 31802, 31800, 31821, 31816, 31814, 31812, 31804, 31806, 31826, 31818, 31820, 31811, 31807, 31809, 31825, 31821, 31823, 31823, 31808, 31811,
31808, 31814, 31802, 31824, 31806, 31818, 31828, 31829, 31827, 31834, 31831, 31832, 31836, 31837, 31835, 31842, 31839, 31840, 31830, 31835, 31829, 31840, 31833, 31834, 31831, 31830, 31828, 31839, 31838, 31836, 31833, 31836, 31830, 31754, 31799, 31753, 31804, 31757, 31803, 31810, 31763, 31764, 31809, 31760, 31763, 31760, 31800, 31754, 31810, 31758, 31804, 31791, 31827, 31789, 31832, 31797, 31798, 31797, 31828, 31791, 31765, 31744, 31742, 31775, 31748, 31773, 31783, 31755, 31756, 31767, 31748, 31744, 31793, 31756, 31762, 31775, 31762, 31750, 31844, 31845, 31843, 31848, 31849, 31847, 31854, 31851, 31852, 31858, 31855, 31856, 31862, 31859, 31861, 31868, 31865, 31867, 31868, 31869, 31866, 31872, 31853, 31854, 31858, 31874, 31857, 31846, 31851, 31845, 31872, 31861, 31871, 31850, 31855, 31849, 31875, 31867, 31874, 31875, 31870, 31868, 31847, 31846, 31844, 31857, 31852, 31855, 31867, 31860, 31862, 31849, 31852, 31846, 31874, 31862, 31872, 31874, 31854, 31857, 31878, 31879, 31877, 31882, 31878, 31881, 31886, 31883, 31884, 31890, 31887, 31889, 31894, 31891, 31892, 31888, 31881, 31887, 31892, 31885, 31886, 31880, 31883, 31879, 31886, 31880, 31882, 31894, 31888, 31890, 31892, 31882, 31888, 31898, 31895, 31897, 31900, 31901, 31899, 31910, 31907, 31909,
31912, 31913, 31911, 31908, 31897, 31907, 31902, 31911, 31901, 31904, 31896, 31898, 31916, 31908, 31910, 31905, 31902, 31900, 31917, 31914, 31912, 31915, 31898, 31908, 31906, 31912, 31902, 31922, 31919, 31921, 31926, 31923, 31924, 31930, 31927, 31929, 31934, 31931, 31932, 31928, 31921, 31927, 31932, 31925, 31926, 31926, 31920, 31922, 31934, 31928, 31930, 31932, 31922, 31928, 31896, 31863, 31895, 31870, 31899, 31869, 31903, 31864, 31896, 31876, 31900, 31870, 31920, 31889, 31919, 31924, 31893, 31894, 31924, 31890, 31920, 31877, 31850, 31848, 31885, 31856, 31883, 31891, 31876, 31875, 31879, 31856, 31850, 31885, 31875, 31858, 31925, 31905, 31923, 31933, 31917, 31931, 31931, 31906, 31925, 31923, 31876, 31893, 31843, 31772, 31770, 31853, 31778, 31851, 31861, 31786, 31788, 31863, 31792, 31790, 31845, 31778, 31772, 31871, 31788, 31796, 31864, 31798, 31792, 31871, 31780, 31853, 31904, 31832, 31903, 31916, 31840, 31915, 31915, 31834, 31904, 31903, 31798, 31873, 31942, 31939, 31941, 31938, 31939, 31937, 31945, 31946, 31944, 31942, 31948, 31940, 31946, 31942, 31941, 31938, 31946, 31941, 31938, 31935, 31936, 31944, 31936, 31943, 31947, 31949, 31942, 31954, 31951, 31952, 31958, 31955, 31956, 31964, 31961, 31963, 31960, 31961, 31959, 31968, 31969, 31967,
31974, 31971, 31972, 31968, 31953, 31954, 31972, 31957, 31958, 31970, 31963, 31969, 31967, 31963, 31960, 31957, 31952, 31955, 31966, 31962, 31964, 31971, 31970, 31968, 31973, 31964, 31970, 31971, 31954, 31957, 31978, 31975, 31977, 31979, 31978, 31977, 31966, 31975, 31965, 31974, 31979, 31973, 31973, 31977, 31966, 31953, 31945, 31951, 31960, 31948, 31949, 31950, 31960, 31949, 31953, 31950, 31947, 31982, 31983, 31981, 31988, 31989, 31987, 31994, 31991, 31992, 31983, 31987, 31981, 31992, 31985, 31986, 31984, 31985, 31983, 31991, 31990, 31988, 31985, 31988, 31983, 31996, 31997, 31995, 32000, 32001, 31999, 32010, 32007, 32008, 32014, 32011, 32012, 32008, 32003, 32004, 32012, 32005, 32006, 31998, 32003, 31997, 32002, 32005, 32001, 31999, 31998, 31996, 32013, 32008, 32011, 32001, 32004, 31998, 32011, 32004, 32005, 32016, 32017, 32015, 32020, 32021, 32019, 32030, 32027, 32028, 32034, 32031, 32032, 32028, 32023, 32024, 32032, 32025, 32026, 32018, 32023, 32017, 32022, 32025, 32021, 32019, 32018, 32016, 32033, 32028, 32031, 32021, 32024, 32018, 32031, 32024, 32025, 32015, 32002, 32000, 32029, 32012, 32027, 32017, 32006, 32002, 32027, 32006, 32023, 31995, 31984, 31982, 32009, 31992, 32007, 31997, 31986, 31984, 32007, 31986, 32003, 32036, 32037, 32035,
32042, 32039, 32041, 32048, 32045, 32046, 32054, 32051, 32053, 32038, 32049, 32037, 32054, 32041, 32052, 32057, 32047, 32048, 32043, 32038, 32036, 32048, 32040, 32042, 32057, 32053, 32056, 32047, 32043, 32045, 32053, 32049, 32050, 32056, 32050, 32055, 32056, 32044, 32047, 32044, 32050, 32038, 32057, 32042, 32054, 32063, 32060, 32062, 32059, 32060, 32058, 32067, 32064, 32065, 32067, 32062, 32066, 32064, 32062, 32059, 32071, 32068, 32070, 32073, 32074, 32072, 32075, 32080, 32074, 32077, 32069, 32071, 32078, 32075, 32073, 32079, 32081, 32075, 32086, 32083, 32085, 32090, 32087, 32088, 32094, 32091, 32093, 32094, 32085, 32092, 32096, 32089, 32090, 32090, 32084, 32086, 32096, 32093, 32095, 32096, 32086, 32094, 32089, 32078, 32087, 32093, 32080, 32081, 32095, 32081, 32082, 32095, 32079, 32089, 32100, 32101, 32099, 32102, 32104, 32101, 32099, 32098, 32097, 32101, 32103, 32098, 32107, 32108, 32106, 32111, 32112, 32110, 32109, 32114, 32108, 32113, 32116, 32112, 32110, 32109, 32107, 32112, 32115, 32109, 32119, 32120, 32118, 32125, 32122, 32124, 32131, 32128, 32129, 32126, 32121, 32119, 32131, 32124, 32130, 32124, 32120, 32121, 32130, 32126, 32128, 32130, 32121, 32127, 32135, 32132, 32134, 32139, 32136, 32138, 32143, 32140, 32141, 32147, 32144, 32145,
32143, 32134, 32142, 32147, 32138, 32146, 32138, 32133, 32135, 32146, 32141, 32144, 32146, 32135, 32143, 32098, 32118, 32097, 32129, 32104, 32105, 32128, 32103, 32104, 32103, 32119, 32098, 32141, 32114, 32115, 32145, 32116, 32117, 32144, 32115, 32116, 32106, 32102, 32100, 32108, 32105, 32102, 32134, 32123, 32125, 32142, 32129, 32140, 32142, 32125, 32131, 32140, 32105, 32114, 32152, 32153, 32151, 32156, 32157, 32155, 32154, 32159, 32153, 32158, 32161, 32157, 32155, 32154, 32152, 32157, 32160, 32154, 32164, 32165, 32163, 32166, 32167, 32165, 32172, 32169, 32171, 32174, 32175, 32173, 32180, 32177, 32178, 32182, 32183, 32181, 32183, 32185, 32181, 32192, 32189, 32191, 32192, 32187, 32190, 32196, 32193, 32194, 32200, 32197, 32198, 32172, 32182, 32170, 32180, 32197, 32179, 32175, 32172, 32171, 32183, 32194, 32193, 32193, 32186, 32183, 32199, 32192, 32191, 32200, 32188, 32192, 32177, 32176, 32174, 32197, 32196, 32194, 32197, 32176, 32179, 32176, 32184, 32172, 32204, 32201, 32203, 32206, 32201, 32205, 32211, 32205, 32210, 32207, 32204, 32203, 32206, 32207, 32203, 32211, 32209, 32206, 32174, 32159, 32160, 32178, 32161, 32162, 32177, 32160, 32161, 32204, 32164, 32202, 32208, 32167, 32168, 32208, 32166, 32204, 32163, 32158, 32156, 32165, 32162, 32158,
32209, 32178, 32207, 32211, 32187, 32188, 32198, 32211, 32188, 32212, 32180, 32209, 32167, 32178, 32162, 32071, 32061, 32063, 32077, 32065, 32076, 32077, 32063, 32067, 32151, 32113, 32111, 32153, 32117, 32113, 32171, 32137, 32139, 32175, 32145, 32173, 32185, 32149, 32148, 32175, 32139, 32147, 32149, 32195, 32150, 32173, 32117, 32159, 32214, 32215, 32213, 32220, 32217, 32218, 32224, 32221, 32223, 32216, 32221, 32215, 32226, 32219, 32220, 32217, 32216, 32214, 32226, 32223, 32225, 32219, 32223, 32216, 32230, 32227, 32228, 32234, 32231, 32232, 32238, 32235, 32237, 32242, 32239, 32241, 32244, 32229, 32230, 32246, 32233, 32234, 32244, 32237, 32243, 32246, 32241, 32245, 32233, 32228, 32231, 32241, 32236, 32238, 32245, 32238, 32244, 32245, 32230, 32233, 32250, 32247, 32248, 32254, 32251, 32252, 32258, 32255, 32257, 32262, 32259, 32261, 32264, 32249, 32250, 32266, 32253, 32254, 32264, 32257, 32263, 32266, 32261, 32265, 32253, 32248, 32251, 32261, 32256, 32258, 32265, 32258, 32264, 32265, 32250, 32253, 32249, 32232, 32247, 32257, 32240, 32242, 32263, 32242, 32246, 32263, 32234, 32249, 32268, 32269, 32267, 32270, 32271, 32269, 32274, 32275, 32273, 32280, 32277, 32278, 32282, 32283, 32281, 32288, 32285, 32286, 32276, 32281, 32275, 32286, 32279, 32280,
32277, 32276, 32274, 32285, 32284, 32282, 32279, 32282, 32276, 32269, 32273, 32267, 32278, 32271, 32272, 32271, 32274, 32269, 32290, 32291, 32289, 32294, 32295, 32293, 32292, 32297, 32291, 32296, 32299, 32295, 32293, 32292, 32290, 32295, 32298, 32292, 32302, 32303, 32301, 32306, 32307, 32305, 32304, 32309, 32303, 32308, 32311, 32307, 32305, 32304, 32302, 32307, 32310, 32304, 32316, 32313, 32314, 32320, 32317, 32318, 32324, 32321, 32322, 32328, 32325, 32326, 32322, 32315, 32316, 32326, 32319, 32320, 32319, 32314, 32317, 32327, 32322, 32325, 32325, 32316, 32319, 32332, 32329, 32330, 32336, 32333, 32334, 32340, 32337, 32338, 32344, 32341, 32342, 32338, 32331, 32332, 32342, 32335, 32336, 32335, 32330, 32333, 32343, 32338, 32341, 32341, 32332, 32335, 32314, 32297, 32298, 32318, 32299, 32300, 32317, 32298, 32299, 32330, 32309, 32310, 32334, 32311, 32312, 32333, 32310, 32311, 32301, 32296, 32294, 32303, 32300, 32296, 32331, 32318, 32329, 32339, 32326, 32337, 32337, 32320, 32331, 32329, 32300, 32309, 32229, 32218, 32227, 32237, 32222, 32224, 32243, 32224, 32226, 32243, 32220, 32229, 32289, 32270, 32268, 32291, 32272, 32270, 32315, 32278, 32313, 32323, 32286, 32321, 32321, 32280, 32315, 32313, 32272, 32297, 31816, 32035, 31815, 32040, 31819, 32039,
32046, 31825, 31826, 32045, 31822, 31825, 31822, 32036, 31816, 32046, 31820, 32040, 31838, 32058, 31837, 32065, 31841, 31842, 31841, 32059, 31838, 32069, 31909, 32068, 31914, 32072, 31913, 32076, 31910, 32069, 31918, 32073, 31914, 32084, 31929, 32083, 32088, 31933, 31934, 32088, 31930, 32084, 32087, 31918, 31933, 32076, 31842, 31916, 31990, 32213, 31989, 32218, 31993, 31994, 31993, 32214, 31990, 32228, 32009, 32010, 32232, 32013, 32014, 32231, 32010, 32013, 32248, 32029, 32030, 32252, 32033, 32034, 32251, 32030, 32033, 32247, 32014, 32029, 32227, 31994, 32009, 32351, 32352, 32350, 32352, 32354, 32350, 32360, 32357, 32358, 32362, 32363, 32361, 32363, 32365, 32361, 32372, 32369, 32371, 32372, 32367, 32370, 32376, 32373, 32374, 32380, 32377, 32378, 32349, 32362, 32348, 32360, 32377, 32359, 32363, 32374, 32373, 32373, 32366, 32363, 32352, 32358, 32357, 32357, 32355, 32352, 32379, 32372, 32371, 32380, 32368, 32372, 32354, 32349, 32348, 32377, 32376, 32374, 32349, 32359, 32356, 32377, 32356, 32359, 32356, 32364, 32349, 32382, 32383, 32381, 32383, 32385, 32381, 32390, 32387, 32388, 32394, 32391, 32393, 32386, 32391, 32385, 32396, 32389, 32390, 32383, 32388, 32387, 32383, 32389, 32386, 32395, 32394, 32393, 32386, 32395, 32393, 32400, 32397, 32399,
32404, 32405, 32403, 32410, 32407, 32408, 32417, 32414, 32416, 32421, 32418, 32420, 32423, 32409, 32410, 32425, 32412, 32413, 32410, 32398, 32400, 32423, 32416, 32422, 32406, 32412, 32405, 32425, 32420, 32424, 32402, 32397, 32398, 32403, 32400, 32399, 32420, 32415, 32417, 32408, 32402, 32398, 32405, 32410, 32400, 32424, 32417, 32423, 32424, 32410, 32412, 32346, 32382, 32345, 32347, 32384, 32346, 32402, 32369, 32401, 32376, 32407, 32375, 32411, 32376, 32379, 32411, 32371, 32402, 32365, 32346, 32345, 32346, 32375, 32347, 32407, 32390, 32388, 32416, 32392, 32394, 32422, 32394, 32396, 32422, 32390, 32409, 32407, 32347, 32375, 32431, 32428, 32430, 32431, 32426, 32429, 32437, 32434, 32435, 32440, 32428, 32439, 32436, 32431, 32430, 32431, 32435, 32427, 32433, 32426, 32427, 32435, 32433, 32427, 32441, 32430, 32440, 32443, 32444, 32442, 32447, 32448, 32446, 32457, 32454, 32456, 32461, 32458, 32460, 32463, 32450, 32451, 32465, 32452, 32453, 32445, 32450, 32444, 32463, 32456, 32462, 32449, 32452, 32448, 32465, 32460, 32464, 32446, 32445, 32443, 32460, 32455, 32457, 32448, 32451, 32445, 32464, 32457, 32463, 32464, 32451, 32452, 32467, 32468, 32466, 32475, 32472, 32474, 32475, 32467, 32473, 32477, 32470, 32471, 32469, 32470, 32468, 32477, 32474, 32476,
32477, 32469, 32475, 32466, 32449, 32447, 32474, 32459, 32461, 32468, 32453, 32449, 32476, 32461, 32465, 32476, 32453, 32470, 32434, 32360, 32358, 32440, 32367, 32368, 32378, 32440, 32368, 32441, 32360, 32436, 32442, 32406, 32404, 32456, 32419, 32421, 32444, 32413, 32406, 32462, 32421, 32425, 32462, 32413, 32450, 32479, 32480, 32478, 32485, 32482, 32483, 32487, 32488, 32486, 32493, 32490, 32491, 32481, 32486, 32480, 32491, 32484, 32485, 32482, 32481, 32479, 32490, 32489, 32487, 32484, 32487, 32481, 32495, 32496, 32494, 32501, 32498, 32499, 32503, 32504, 32502, 32509, 32506, 32507, 32497, 32502, 32496, 32507, 32500, 32501, 32498, 32497, 32495, 32506, 32505, 32503, 32500, 32503, 32497, 32489, 32494, 32488, 32499, 32492, 32493, 32492, 32495, 32489, 32513, 32510, 32511, 32517, 32514, 32515, 32521, 32518, 32519, 32525, 32522, 32523, 32519, 32512, 32513, 32523, 32516, 32517, 32516, 32511, 32514, 32524, 32519, 32522, 32522, 32513, 32516, 32529, 32526, 32528, 32533, 32530, 32532, 32537, 32534, 32535, 32541, 32538, 32539, 32545, 32542, 32544, 32543, 32526, 32542, 32537, 32528, 32536, 32547, 32543, 32545, 32541, 32532, 32540, 32532, 32527, 32529, 32540, 32535, 32538, 32540, 32529, 32537, 32546, 32528, 32543, 32551, 32548, 32549, 32555, 32552, 32553,
32559, 32556, 32557, 32563, 32560, 32561, 32557, 32550, 32551, 32561, 32554, 32555, 32554, 32549, 32552, 32562, 32557, 32560, 32560, 32551, 32554, 32567, 32564, 32566, 32573, 32570, 32572, 32571, 32566, 32570, 32569, 32565, 32567, 32575, 32571, 32573, 32574, 32567, 32571, 32549, 32520, 32521, 32553, 32524, 32525, 32552, 32521, 32524, 32565, 32544, 32564, 32568, 32545, 32565, 32536, 32515, 32534, 32547, 32523, 32546, 32546, 32517, 32536, 32569, 32553, 32568, 32575, 32561, 32574, 32574, 32555, 32569, 32568, 32525, 32547, 32577, 32578, 32576, 32583, 32580, 32581, 32585, 32586, 32584, 32589, 32590, 32588, 32579, 32584, 32578, 32583, 32588, 32582, 32580, 32579, 32577, 32588, 32587, 32585, 32582, 32585, 32579, 32593, 32594, 32592, 32599, 32596, 32597, 32601, 32602, 32600, 32605, 32606, 32604, 32595, 32600, 32594, 32599, 32604, 32598, 32596, 32595, 32593, 32604, 32603, 32601, 32598, 32601, 32595, 32587, 32592, 32586, 32591, 32596, 32590, 32590, 32593, 32587, 32611, 32608, 32609, 32615, 32612, 32613, 32619, 32616, 32617, 32623, 32620, 32621, 32617, 32610, 32611, 32621, 32614, 32615, 32614, 32609, 32612, 32622, 32617, 32620, 32620, 32611, 32614, 32627, 32624, 32626, 32633, 32630, 32632, 32631, 32626, 32630, 32629, 32625, 32627, 32635, 32631, 32633,
32634, 32627, 32631, 32639, 32636, 32637, 32643, 32640, 32641, 32647, 32644, 32645, 32651, 32648, 32649, 32645, 32638, 32639, 32649, 32642, 32643, 32642, 32637, 32640, 32650, 32645, 32648, 32648, 32639, 32642, 32655, 32652, 32654, 32661, 32658, 32660, 32659, 32654, 32658, 32657, 32653, 32655, 32663, 32659, 32661, 32662, 32655, 32659, 32637, 32618, 32619, 32641, 32622, 32623, 32640, 32619, 32622, 32653, 32632, 32652, 32656, 32633, 32653, 32629, 32613, 32628, 32634, 32623, 32621, 32634, 32615, 32629, 32657, 32641, 32656, 32662, 32651, 32649, 32657, 32649, 32643, 32635, 32641, 32623, 32505, 32576, 32504, 32581, 32508, 32509, 32508, 32577, 32505, 32609, 32558, 32559, 32613, 32562, 32563, 32612, 32559, 32562, 32625, 32572, 32624, 32628, 32573, 32625, 32628, 32563, 32575, 32512, 32483, 32510, 32520, 32491, 32518, 32518, 32485, 32512, 32550, 32499, 32548, 32558, 32507, 32556, 32556, 32501, 32550, 32548, 32493, 32520, 32610, 32581, 32608, 32618, 32589, 32616, 32616, 32583, 32610, 32638, 32597, 32636, 32646, 32605, 32644, 32644, 32599, 32638, 32636, 32591, 32618, 32608, 32509, 32558, 32667, 32668, 32666, 32674, 32675, 32673, 32680, 32677, 32679, 32684, 32685, 32683, 32665, 32673, 32664, 32672, 32683, 32671, 32681, 32676, 32674, 32669, 32671, 32668,
32686, 32678, 32680, 32666, 32665, 32664, 32683, 32682, 32681, 32668, 32670, 32665, 32671, 32681, 32670, 32670, 32674, 32665, 32688, 32689, 32687, 32692, 32693, 32691, 32702, 32699, 32701, 32706, 32703, 32705, 32708, 32695, 32696, 32710, 32697, 32698, 32690, 32695, 32689, 32708, 32701, 32707, 32694, 32697, 32693, 32710, 32705, 32709, 32691, 32690, 32688, 32705, 32700, 32702, 32693, 32696, 32690, 32709, 32702, 32708, 32709, 32696, 32697, 32712, 32713, 32711, 32718, 32715, 32717, 32722, 32723, 32721, 32726, 32727, 32725, 32732, 32729, 32731, 32738, 32735, 32736, 32714, 32725, 32713, 32730, 32717, 32729, 32736, 32723, 32724, 32719, 32714, 32712, 32733, 32728, 32726, 32724, 32716, 32718, 32738, 32730, 32732, 32721, 32720, 32719, 32735, 32734, 32733, 32723, 32733, 32720, 32720, 32726, 32714, 32736, 32718, 32730, 32676, 32711, 32675, 32716, 32679, 32715, 32722, 32685, 32686, 32685, 32719, 32682, 32682, 32712, 32676, 32722, 32680, 32716, 32687, 32669, 32667, 32701, 32677, 32678, 32689, 32672, 32669, 32707, 32678, 32684, 32695, 32684, 32672, 32740, 32741, 32739, 32744, 32745, 32743, 32754, 32751, 32753, 32758, 32755, 32757, 32758, 32759, 32756, 32762, 32747, 32748, 32750, 32763, 32749, 32742, 32747, 32741, 32762, 32753, 32761, 32746, 32749, 32745,
32764, 32757, 32763, 32758, 32765, 32760, 32743, 32742, 32740, 32757, 32752, 32754, 32745, 32748, 32742, 32763, 32754, 32762, 32763, 32748, 32749, 32767, 32768, 32766, 32775, 32772, 32774, 32779, 32776, 32777, 32773, 32767, 32772, 32777, 32770, 32771, 32769, 32770, 32768, 32779, 32773, 32775, 32777, 32769, 32773, 32781, 32782, 32780, 32787, 32788, 32786, 32783, 32786, 32782, 32781, 32785, 32783, 32787, 32791, 32789, 32783, 32790, 32787, 32795, 32792, 32794, 32799, 32796, 32797, 32803, 32800, 32802, 32805, 32806, 32804, 32801, 32794, 32800, 32799, 32804, 32798, 32799, 32793, 32795, 32803, 32805, 32801, 32801, 32799, 32795, 32760, 32780, 32759, 32760, 32784, 32781, 32793, 32774, 32792, 32797, 32778, 32779, 32797, 32775, 32793, 32766, 32746, 32744, 32778, 32764, 32776, 32768, 32750, 32746, 32776, 32750, 32770, 32798, 32784, 32796, 32806, 32790, 32804, 32804, 32785, 32798, 32796, 32765, 32778, 32812, 32814, 32810, 32820, 32817, 32818, 32824, 32821, 32823, 32824, 32825, 32822, 32815, 32821, 32814, 32820, 32827, 32819, 32812, 32819, 32815, 32828, 32823, 32827, 32824, 32829, 32826, 32812, 32808, 32809, 32817, 32809, 32816, 32819, 32823, 32815, 32833, 32830, 32832, 32837, 32838, 32836, 32843, 32837, 32841, 32845, 32846, 32844, 32851, 32848, 32849,
32838, 32832, 32836, 32835, 32831, 32833, 32838, 32849, 32848, 32851, 32839, 32843, 32847, 32852, 32846, 32844, 32843, 32841, 32843, 32852, 32851, 32848, 32833, 32838, 32857, 32854, 32856, 32857, 32858, 32855, 32863, 32860, 32862, 32868, 32869, 32867, 32864, 32857, 32856, 32865, 32859, 32857, 32870, 32862, 32869, 32862, 32858, 32859, 32867, 32866, 32865, 32866, 32862, 32859, 32842, 32854, 32840, 32865, 32850, 32851, 32853, 32867, 32852, 32867, 32851, 32852, 32850, 32856, 32842, 32830, 32813, 32811, 32834, 32820, 32818, 32842, 32825, 32826, 32850, 32828, 32848, 32831, 32818, 32813, 32829, 32842, 32826, 32848, 32820, 32835, 32874, 32875, 32873, 32872, 32874, 32871, 32883, 32884, 32882, 32881, 32883, 32880, 32894, 32895, 32893, 32875, 32886, 32873, 32894, 32878, 32879, 32884, 32892, 32891, 32881, 32892, 32885, 32875, 32879, 32878, 32872, 32879, 32876, 32886, 32881, 32880, 32895, 32890, 32893, 32887, 32890, 32881, 32878, 32887, 32875, 32902, 32899, 32901, 32902, 32897, 32900, 32906, 32903, 32904, 32910, 32907, 32909, 32908, 32899, 32907, 32905, 32902, 32901, 32906, 32898, 32902, 32910, 32911, 32908, 32908, 32905, 32901, 32918, 32915, 32917, 32918, 32913, 32916, 32922, 32919, 32921, 32926, 32923, 32924, 32926, 32917, 32925, 32918, 32924, 32914,
32927, 32922, 32921, 32921, 32913, 32914, 32927, 32914, 32924, 32922, 32888, 32920, 32892, 32923, 32891, 32928, 32895, 32896, 32895, 32924, 32892, 32922, 32896, 32889, 32903, 32879, 32877, 32910, 32888, 32889, 32912, 32894, 32911, 32896, 32910, 32889, 32905, 32894, 32879, 32728, 32808, 32727, 32813, 32731, 32811, 32818, 32737, 32738, 32737, 32816, 32734, 32728, 32816, 32809, 32818, 32732, 32813, 32789, 32871, 32788, 32789, 32877, 32872, 32898, 32802, 32897, 32904, 32806, 32807, 32898, 32807, 32803, 32903, 32791, 32806, 32739, 32694, 32692, 32753, 32704, 32706, 32741, 32698, 32694, 32761, 32706, 32710, 32761, 32698, 32747, 32882, 32847, 32845, 32847, 32891, 32853, 32917, 32861, 32863, 32925, 32868, 32923, 32870, 32917, 32863, 32891, 32868, 32853, 32932, 32929, 32931, 32932, 32933, 32930, 32939, 32940, 32938, 32934, 32938, 32933, 32936, 32931, 32935, 32932, 32937, 32934, 32942, 32941, 32939, 32937, 32939, 32934, 32941, 32944, 32940, 32941, 32946, 32945, 32950, 32947, 32948, 32954, 32951, 32952, 32958, 32955, 32956, 32962, 32959, 32960, 32956, 32949, 32950, 32960, 32953, 32954, 32953, 32948, 32951, 32961, 32956, 32959, 32959, 32950, 32953, 32966, 32963, 32965, 32970, 32971, 32969, 32974, 32970, 32973, 32976, 32977, 32975, 32971, 32965, 32969,
32968, 32964, 32966, 32972, 32975, 32971, 32978, 32972, 32974, 32975, 32966, 32971, 32980, 32981, 32979, 32984, 32985, 32983, 32990, 32987, 32988, 32996, 32993, 32995, 32992, 32993, 32991, 32998, 32999, 32997, 33004, 33001, 33002, 32982, 32991, 32981, 32986, 32997, 32985, 33002, 32989, 32990, 32980, 32985, 32982, 32999, 32996, 32995, 32997, 32995, 32992, 32989, 32984, 32987, 33003, 32998, 33001, 33001, 32986, 32989, 32985, 32992, 32982, 33010, 33007, 33009, 33014, 33011, 33012, 33018, 33019, 33017, 33022, 33018, 33021, 33026, 33023, 33024, 33030, 33027, 33028, 33019, 33009, 33017, 33024, 33013, 33014, 33016, 33008, 33010, 33020, 33027, 33019, 33030, 33020, 33022, 33007, 33006, 33005, 33015, 33014, 33012, 33029, 33024, 33027, 33008, 33012, 33006, 33016, 33024, 33014, 33027, 33010, 33019, 32984, 32957, 32958, 32988, 32961, 32962, 32987, 32958, 32961, 33006, 32973, 33005, 32978, 33011, 32977, 33012, 32974, 33006, 32967, 32954, 32952, 32977, 32960, 32975, 32968, 32960, 32954, 33013, 32988, 33011, 33025, 33002, 33023, 33023, 32990, 33013, 33011, 32962, 32977, 33038, 33035, 33037, 33034, 33035, 33033, 33041, 33042, 33040, 33038, 33044, 33036, 33043, 33037, 33042, 33034, 33042, 33037, 33034, 33031, 33032, 33040, 33032, 33039, 33038, 33046, 33045,
33050, 33047, 33048, 33054, 33051, 33052, 33060, 33057, 33059, 33056, 33057, 33055, 33062, 33063, 33061, 33068, 33065, 33066, 33062, 33049, 33050, 33066, 33053, 33054, 33063, 33060, 33059, 33061, 33059, 33056, 33053, 33048, 33051, 33067, 33062, 33065, 33065, 33050, 33053, 33074, 33071, 33073, 33074, 33075, 33072, 33079, 33073, 33078, 33079, 33076, 33074, 33073, 33069, 33070, 33078, 33070, 33077, 33060, 33069, 33058, 33068, 33078, 33067, 33067, 33077, 33064, 33064, 33070, 33060, 33049, 33041, 33047, 33056, 33044, 33045, 33046, 33056, 33045, 33049, 33046, 33043, 32996, 33031, 32994, 33041, 33003, 33004, 33003, 33039, 33000, 32996, 33039, 33032, 33048, 33025, 33026, 33052, 33029, 33030, 33051, 33026, 33029, 33047, 33004, 33025, 32947, 32937, 32936, 32955, 32943, 32942, 32949, 32942, 32937, 32980, 32944, 32945, 32983, 32945, 32946, 32957, 32946, 32943, 32603, 32929, 32602, 32607, 32935, 32606, 32603, 32935, 32931, 32948, 32646, 32647, 32952, 32650, 32651, 32951, 32647, 32650, 32964, 32660, 32963, 32967, 32661, 32964, 32663, 32952, 32651, 32947, 32607, 32646, 32149, 32351, 32148, 32150, 32353, 32149, 32433, 32189, 32432, 32196, 32434, 32195, 32438, 32196, 32199, 32438, 32191, 32433, 32434, 32150, 32195, 32284, 32478, 32283, 32483, 32287, 32288,
32287, 32479, 32284, 32511, 32323, 32324, 32515, 32327, 32328, 32514, 32324, 32327, 32535, 32339, 32340, 32539, 32343, 32344, 32538, 32340, 32343, 32534, 32328, 32339, 32510, 32288, 32323, 33082, 33083, 33081, 33086, 33087, 33085, 33094, 33095, 33093, 33091, 33084, 33082, 33088, 33093, 33087, 33085, 33082, 33081, 33095, 33091, 33093, 33087, 33091, 33082, 33098, 33099, 33097, 33104, 33101, 33103, 33106, 33107, 33105, 33112, 33109, 33111, 33116, 33113, 33114, 33120, 33117, 33118, 33100, 33113, 33099, 33116, 33103, 33115, 33108, 33117, 33107, 33120, 33111, 33119, 33105, 33100, 33098, 33111, 33102, 33104, 33119, 33114, 33117, 33107, 33114, 33100, 33119, 33104, 33116, 33097, 33088, 33086, 33103, 33089, 33090, 33113, 33096, 33094, 33099, 33094, 33088, 33115, 33090, 33096, 33122, 33123, 33121, 33128, 33125, 33127, 33130, 33131, 33129, 33138, 33135, 33136, 33142, 33139, 33140, 33124, 33135, 33123, 33138, 33127, 33137, 33132, 33139, 33131, 33129, 33124, 33122, 33134, 33126, 33128, 33139, 33138, 33136, 33131, 33136, 33124, 33141, 33128, 33138, 33144, 33145, 33143, 33148, 33144, 33147, 33152, 33149, 33150, 33146, 33149, 33145, 33152, 33146, 33148, 33143, 33132, 33130, 33151, 33140, 33149, 33145, 33140, 33132, 33154, 33155, 33153, 33160, 33157, 33159,
33166, 33163, 33164, 33168, 33169, 33167, 33174, 33171, 33173, 33180, 33177, 33178, 33156, 33167, 33155, 33172, 33159, 33171, 33178, 33165, 33166, 33161, 33156, 33154, 33175, 33170, 33168, 33166, 33158, 33160, 33180, 33172, 33174, 33165, 33161, 33163, 33179, 33175, 33177, 33177, 33162, 33165, 33162, 33168, 33156, 33178, 33160, 33172, 33182, 33183, 33181, 33188, 33189, 33187, 33194, 33191, 33192, 33183, 33187, 33181, 33192, 33185, 33186, 33184, 33185, 33183, 33191, 33190, 33188, 33185, 33188, 33183, 33196, 33197, 33195, 33202, 33199, 33201, 33208, 33205, 33206, 33210, 33211, 33209, 33218, 33215, 33217, 33218, 33213, 33216, 33224, 33221, 33222, 33198, 33209, 33197, 33214, 33201, 33213, 33222, 33207, 33208, 33203, 33198, 33196, 33219, 33212, 33210, 33208, 33200, 33202, 33224, 33217, 33223, 33224, 33214, 33218, 33207, 33203, 33205, 33217, 33211, 33212, 33223, 33219, 33221, 33223, 33212, 33220, 33221, 33204, 33207, 33204, 33210, 33198, 33222, 33202, 33214, 33226, 33227, 33225, 33232, 33229, 33230, 33238, 33235, 33237, 33234, 33235, 33233, 33242, 33239, 33240, 33228, 33233, 33227, 33240, 33231, 33232, 33229, 33228, 33226, 33242, 33237, 33241, 33239, 33237, 33234, 33231, 33234, 33228, 33170, 33195, 33169, 33200, 33173, 33199, 33206, 33179, 33180,
33205, 33176, 33179, 33176, 33196, 33170, 33206, 33174, 33200, 33190, 33225, 33189, 33230, 33193, 33194, 33193, 33226, 33190, 33248, 33245, 33247, 33248, 33249, 33246, 33258, 33255, 33257, 33260, 33261, 33259, 33256, 33243, 33255, 33250, 33259, 33249, 33264, 33256, 33258, 33252, 33248, 33247, 33248, 33254, 33250, 33265, 33262, 33260, 33263, 33244, 33256, 33250, 33265, 33260, 33270, 33267, 33269, 33274, 33271, 33272, 33278, 33275, 33277, 33282, 33279, 33280, 33276, 33269, 33275, 33280, 33273, 33274, 33274, 33268, 33270, 33282, 33276, 33278, 33280, 33270, 33276, 33286, 33283, 33285, 33288, 33289, 33287, 33298, 33295, 33297, 33300, 33301, 33299, 33296, 33285, 33295, 33290, 33299, 33289, 33292, 33284, 33286, 33304, 33296, 33298, 33293, 33290, 33288, 33305, 33302, 33300, 33303, 33286, 33296, 33294, 33300, 33290, 33310, 33307, 33309, 33314, 33311, 33312, 33320, 33317, 33319, 33320, 33315, 33318, 33324, 33321, 33322, 33316, 33309, 33315, 33322, 33313, 33314, 33314, 33308, 33310, 33324, 33319, 33323, 33324, 33316, 33320, 33322, 33310, 33316, 33284, 33257, 33283, 33262, 33287, 33261, 33291, 33258, 33284, 33266, 33288, 33262, 33308, 33277, 33307, 33312, 33281, 33282, 33312, 33278, 33308, 33271, 33254, 33253, 33279, 33266, 33265, 33273, 33265, 33254,
33311, 33294, 33293, 33319, 33301, 33302, 33323, 33305, 33321, 33323, 33302, 33306, 33321, 33294, 33313, 33311, 33266, 33281, 33084, 33153, 33083, 33158, 33089, 33157, 33096, 33163, 33095, 33163, 33092, 33095, 33092, 33154, 33084, 33164, 33090, 33158, 33134, 33245, 33133, 33142, 33252, 33141, 33141, 33247, 33134, 33268, 33147, 33267, 33272, 33151, 33152, 33272, 33148, 33268, 33271, 33142, 33151, 33121, 33108, 33106, 33127, 33110, 33112, 33137, 33118, 33135, 33123, 33118, 33108, 33137, 33112, 33120, 33243, 33184, 33182, 33264, 33192, 33263, 33244, 33186, 33184, 33263, 33186, 33251, 33292, 33230, 33291, 33298, 33236, 33238, 33304, 33240, 33303, 33304, 33238, 33242, 33303, 33232, 33292, 33291, 33194, 33264, 33326, 33327, 33325, 33327, 33329, 33325, 33334, 33331, 33332, 33328, 33331, 33327, 33331, 33330, 33327, 33336, 33337, 33335, 33342, 33339, 33340, 33344, 33345, 33343, 33350, 33347, 33348, 33338, 33343, 33337, 33348, 33341, 33342, 33339, 33338, 33336, 33347, 33346, 33344, 33341, 33344, 33338, 33330, 33335, 33329, 33340, 33333, 33334, 33333, 33336, 33330, 33352, 33353, 33351, 33356, 33357, 33355, 33362, 33359, 33360, 33366, 33363, 33364, 33354, 33359, 33353, 33358, 33363, 33357, 33355, 33354, 33352, 33365, 33360, 33363, 33357, 33360, 33354,
33370, 33367, 33369, 33372, 33368, 33370, 33376, 33373, 33374, 33380, 33377, 33378, 33384, 33381, 33382, 33388, 33385, 33386, 33382, 33375, 33376, 33386, 33379, 33380, 33379, 33374, 33377, 33387, 33382, 33385, 33385, 33376, 33379, 33392, 33389, 33391, 33398, 33395, 33397, 33396, 33391, 33395, 33394, 33390, 33392, 33400, 33396, 33398, 33399, 33392, 33396, 33374, 33361, 33362, 33378, 33365, 33366, 33377, 33362, 33365, 33390, 33369, 33389, 33393, 33370, 33390, 33367, 33358, 33356, 33372, 33364, 33371, 33368, 33364, 33358, 33394, 33378, 33393, 33400, 33386, 33399, 33399, 33380, 33394, 33393, 33366, 33372, 33402, 33403, 33401, 33408, 33405, 33406, 33410, 33411, 33409, 33416, 33413, 33414, 33404, 33409, 33403, 33414, 33407, 33408, 33405, 33404, 33402, 33413, 33412, 33410, 33407, 33410, 33404, 33418, 33419, 33417, 33424, 33421, 33422, 33426, 33427, 33425, 33432, 33429, 33430, 33420, 33425, 33419, 33430, 33423, 33424, 33421, 33420, 33418, 33429, 33428, 33426, 33423, 33426, 33420, 33412, 33417, 33411, 33422, 33415, 33416, 33415, 33418, 33412, 33436, 33433, 33434, 33440, 33437, 33438, 33444, 33441, 33442, 33448, 33445, 33446, 33442, 33435, 33436, 33446, 33439, 33440, 33439, 33434, 33437, 33447, 33442, 33445, 33445, 33436, 33439, 33452, 33449, 33451,
33458, 33455, 33457, 33456, 33451, 33455, 33454, 33450, 33452, 33460, 33456, 33458, 33459, 33452, 33456, 33464, 33461, 33462, 33468, 33465, 33466, 33472, 33469, 33470, 33476, 33473, 33474, 33470, 33463, 33464, 33474, 33467, 33468, 33467, 33462, 33465, 33475, 33470, 33473, 33473, 33464, 33467, 33480, 33477, 33479, 33486, 33483, 33485, 33484, 33479, 33483, 33482, 33478, 33480, 33488, 33484, 33486, 33487, 33480, 33484, 33462, 33443, 33444, 33466, 33447, 33448, 33465, 33444, 33447, 33478, 33457, 33477, 33481, 33458, 33478, 33454, 33438, 33453, 33460, 33446, 33459, 33459, 33440, 33454, 33482, 33466, 33481, 33488, 33474, 33487, 33487, 33468, 33482, 33481, 33448, 33460, 33346, 33401, 33345, 33406, 33349, 33350, 33349, 33402, 33346, 33434, 33383, 33384, 33438, 33387, 33388, 33437, 33384, 33387, 33450, 33397, 33449, 33453, 33398, 33450, 33453, 33388, 33400, 33351, 33328, 33326, 33361, 33332, 33359, 33353, 33332, 33328, 33375, 33340, 33373, 33383, 33348, 33381, 33381, 33342, 33375, 33373, 33334, 33361, 33435, 33406, 33433, 33443, 33414, 33441, 33441, 33408, 33435, 33463, 33422, 33461, 33471, 33430, 33469, 33469, 33424, 33463, 33461, 33416, 33443, 33433, 33350, 33383, 33490, 33491, 33489, 33494, 33495, 33493, 33502, 33499, 33500, 33508, 33505, 33507,
33492, 33503, 33491, 33511, 33501, 33502, 33497, 33492, 33490, 33496, 33499, 33495, 33511, 33507, 33510, 33493, 33490, 33489, 33501, 33497, 33499, 33507, 33503, 33504, 33495, 33497, 33490, 33510, 33504, 33509, 33510, 33498, 33501, 33498, 33504, 33492, 33513, 33514, 33512, 33517, 33518, 33516, 33523, 33520, 33521, 33527, 33524, 33525, 33531, 33528, 33530, 33535, 33532, 33534, 33537, 33522, 33523, 33539, 33526, 33527, 33515, 33520, 33514, 33537, 33530, 33536, 33519, 33524, 33518, 33539, 33534, 33538, 33516, 33515, 33513, 33526, 33521, 33524, 33534, 33529, 33531, 33518, 33521, 33515, 33538, 33531, 33537, 33526, 33537, 33523, 33541, 33542, 33540, 33547, 33544, 33546, 33553, 33554, 33552, 33559, 33556, 33558, 33543, 33552, 33542, 33557, 33546, 33556, 33541, 33549, 33543, 33553, 33561, 33555, 33551, 33545, 33547, 33563, 33557, 33559, 33544, 33541, 33540, 33550, 33549, 33548, 33563, 33560, 33562, 33545, 33548, 33541, 33562, 33549, 33551, 33543, 33560, 33553, 33562, 33547, 33557, 33512, 33496, 33494, 33522, 33500, 33520, 33530, 33506, 33508, 33514, 33500, 33496, 33536, 33508, 33511, 33536, 33502, 33522, 33565, 33566, 33564, 33569, 33570, 33568, 33575, 33572, 33573, 33579, 33576, 33577, 33583, 33580, 33582, 33589, 33586, 33588, 33585, 33586, 33584,
33593, 33594, 33592, 33591, 33574, 33575, 33593, 33578, 33579, 33567, 33572, 33566, 33591, 33582, 33590, 33571, 33576, 33570, 33594, 33589, 33588, 33592, 33588, 33585, 33568, 33567, 33565, 33578, 33573, 33576, 33585, 33581, 33583, 33570, 33573, 33567, 33591, 33585, 33583, 33578, 33591, 33575, 33597, 33598, 33596, 33601, 33597, 33600, 33605, 33602, 33603, 33609, 33606, 33608, 33613, 33610, 33611, 33607, 33600, 33606, 33611, 33604, 33605, 33599, 33602, 33598, 33605, 33599, 33601, 33613, 33607, 33609, 33611, 33601, 33607, 33615, 33616, 33614, 33621, 33622, 33620, 33617, 33620, 33616, 33618, 33617, 33615, 33621, 33625, 33623, 33619, 33621, 33617, 33629, 33626, 33628, 33633, 33630, 33631, 33637, 33634, 33636, 33639, 33640, 33638, 33635, 33628, 33634, 33639, 33632, 33633, 33633, 33627, 33629, 33637, 33639, 33635, 33639, 33629, 33635, 33589, 33614, 33587, 33595, 33615, 33589, 33627, 33608, 33626, 33631, 33612, 33613, 33631, 33609, 33627, 33596, 33571, 33569, 33604, 33577, 33602, 33612, 33593, 33610, 33598, 33577, 33571, 33610, 33579, 33604, 33630, 33619, 33618, 33640, 33624, 33638, 33632, 33624, 33619, 33612, 33618, 33595, 33564, 33519, 33517, 33574, 33525, 33572, 33582, 33533, 33535, 33566, 33525, 33519, 33590, 33535, 33539, 33590, 33527, 33574,
33643, 33644, 33642, 33649, 33646, 33647, 33651, 33652, 33650, 33657, 33654, 33655, 33645, 33650, 33644, 33655, 33648, 33649, 33646, 33645, 33643, 33654, 33653, 33651, 33648, 33651, 33645, 33659, 33660, 33658, 33665, 33662, 33663, 33667, 33668, 33666, 33673, 33670, 33671, 33661, 33666, 33660, 33671, 33664, 33665, 33662, 33661, 33659, 33670, 33669, 33667, 33664, 33667, 33661, 33653, 33658, 33652, 33663, 33656, 33657, 33656, 33659, 33653, 33677, 33674, 33675, 33681, 33678, 33679, 33685, 33682, 33683, 33689, 33686, 33687, 33683, 33676, 33677, 33687, 33680, 33681, 33680, 33675, 33678, 33688, 33683, 33686, 33686, 33677, 33680, 33693, 33690, 33692, 33699, 33696, 33698, 33697, 33692, 33696, 33695, 33691, 33693, 33701, 33697, 33699, 33700, 33693, 33697, 33705, 33702, 33703, 33709, 33706, 33707, 33713, 33710, 33711, 33717, 33714, 33715, 33711, 33704, 33705, 33715, 33708, 33709, 33708, 33703, 33706, 33716, 33711, 33714, 33714, 33705, 33708, 33721, 33718, 33720, 33727, 33724, 33726, 33725, 33720, 33724, 33723, 33719, 33721, 33729, 33725, 33727, 33728, 33721, 33725, 33703, 33684, 33685, 33707, 33688, 33689, 33706, 33685, 33688, 33719, 33698, 33718, 33722, 33699, 33719, 33695, 33679, 33694, 33701, 33687, 33700, 33700, 33681, 33695, 33723, 33707, 33722,
33729, 33715, 33728, 33728, 33709, 33723, 33722, 33689, 33701, 33731, 33732, 33730, 33735, 33736, 33734, 33739, 33740, 33738, 33743, 33744, 33742, 33733, 33738, 33732, 33743, 33736, 33737, 33734, 33733, 33731, 33742, 33741, 33739, 33736, 33739, 33733, 33751, 33748, 33750, 33747, 33748, 33746, 33753, 33754, 33752, 33757, 33758, 33756, 33751, 33756, 33749, 33755, 33750, 33754, 33747, 33754, 33750, 33760, 33759, 33757, 33751, 33760, 33757, 33741, 33746, 33740, 33745, 33752, 33744, 33744, 33747, 33741, 33765, 33762, 33763, 33769, 33766, 33767, 33773, 33770, 33771, 33777, 33774, 33775, 33771, 33764, 33765, 33775, 33768, 33769, 33768, 33763, 33766, 33776, 33771, 33774, 33774, 33765, 33768, 33781, 33778, 33780, 33787, 33784, 33786, 33785, 33780, 33784, 33783, 33779, 33781, 33789, 33785, 33787, 33788, 33781, 33785, 33793, 33790, 33791, 33797, 33794, 33795, 33801, 33798, 33799, 33805, 33802, 33803, 33799, 33792, 33793, 33803, 33796, 33797, 33796, 33791, 33794, 33804, 33799, 33802, 33802, 33793, 33796, 33809, 33806, 33808, 33813, 33814, 33812, 33817, 33813, 33816, 33821, 33822, 33820, 33814, 33808, 33812, 33811, 33807, 33809, 33815, 33820, 33814, 33817, 33821, 33815, 33818, 33817, 33816, 33817, 33824, 33823, 33814, 33811, 33809, 33791, 33772, 33773,
33795, 33776, 33777, 33794, 33773, 33776, 33807, 33786, 33806, 33810, 33787, 33807, 33783, 33767, 33782, 33788, 33777, 33775, 33783, 33775, 33769, 33810, 33797, 33795, 33822, 33803, 33820, 33811, 33803, 33797, 33789, 33795, 33777, 33669, 33730, 33668, 33735, 33672, 33673, 33672, 33731, 33669, 33763, 33712, 33713, 33767, 33716, 33717, 33766, 33713, 33716, 33779, 33726, 33778, 33782, 33727, 33779, 33729, 33767, 33717, 33676, 33647, 33674, 33684, 33655, 33682, 33682, 33649, 33676, 33704, 33663, 33702, 33712, 33671, 33710, 33710, 33665, 33704, 33702, 33657, 33684, 33764, 33735, 33762, 33772, 33743, 33770, 33770, 33737, 33764, 33792, 33753, 33790, 33798, 33761, 33760, 33792, 33760, 33755, 33790, 33745, 33772, 33762, 33673, 33712, 33428, 33642, 33427, 33647, 33431, 33432, 33431, 33643, 33428, 33675, 33471, 33472, 33679, 33475, 33476, 33678, 33472, 33475, 33691, 33485, 33690, 33694, 33486, 33691, 33694, 33476, 33488, 33674, 33432, 33471, 33829, 33830, 33828, 33830, 33832, 33828, 33837, 33834, 33835, 33826, 33829, 33825, 33830, 33835, 33834, 33830, 33836, 33833, 33827, 33831, 33826, 33839, 33840, 33838, 33840, 33842, 33838, 33847, 33848, 33846, 33848, 33844, 33846, 33853, 33850, 33851, 33860, 33857, 33859, 33864, 33861, 33863, 33868, 33865, 33866,
33870, 33871, 33869, 33862, 33847, 33861, 33853, 33865, 33852, 33856, 33869, 33855, 33840, 33851, 33850, 33850, 33843, 33840, 33868, 33859, 33867, 33849, 33855, 33848, 33848, 33854, 33845, 33872, 33862, 33864, 33854, 33853, 33851, 33869, 33868, 33866, 33869, 33853, 33855, 33870, 33849, 33862, 33874, 33875, 33873, 33884, 33881, 33883, 33884, 33878, 33882, 33890, 33887, 33888, 33876, 33891, 33875, 33894, 33881, 33893, 33885, 33876, 33874, 33890, 33883, 33889, 33884, 33888, 33880, 33877, 33874, 33873, 33887, 33886, 33885, 33894, 33891, 33892, 33874, 33887, 33885, 33896, 33892, 33895, 33896, 33886, 33889, 33876, 33895, 33892, 33894, 33889, 33883, 33906, 33903, 33905, 33902, 33903, 33901, 33910, 33911, 33909, 33906, 33913, 33904, 33911, 33906, 33905, 33909, 33905, 33902, 33902, 33898, 33900, 33909, 33900, 33908, 33912, 33914, 33906, 33833, 33877, 33832, 33837, 33887, 33836, 33833, 33887, 33879, 33899, 33857, 33897, 33860, 33898, 33858, 33910, 33871, 33872, 33871, 33908, 33868, 33899, 33867, 33859, 33860, 33908, 33900, 33842, 33826, 33825, 33865, 33837, 33835, 33826, 33852, 33827, 33865, 33827, 33852, 33899, 33878, 33880, 33907, 33880, 33888, 33907, 33837, 33867, 33917, 33918, 33916, 33923, 33924, 33922, 33927, 33928, 33926, 33928, 33930, 33926,
33937, 33934, 33935, 33919, 33927, 33918, 33925, 33932, 33924, 33917, 33921, 33919, 33928, 33935, 33934, 33934, 33931, 33928, 33930, 33925, 33923, 33925, 33936, 33933, 33919, 33935, 33929, 33941, 33938, 33940, 33943, 33944, 33942, 33951, 33948, 33950, 33951, 33946, 33949, 33955, 33952, 33953, 33947, 33940, 33946, 33953, 33944, 33945, 33941, 33943, 33939, 33954, 33951, 33950, 33955, 33947, 33951, 33947, 33945, 33941, 33959, 33956, 33958, 33963, 33960, 33961, 33967, 33964, 33965, 33971, 33968, 33970, 33973, 33956, 33972, 33963, 33974, 33962, 33975, 33970, 33974, 33966, 33959, 33958, 33964, 33963, 33961, 33973, 33969, 33971, 33976, 33971, 33975, 33976, 33963, 33966, 33973, 33966, 33958, 33980, 33977, 33979, 33980, 33981, 33978, 33933, 33960, 33932, 33937, 33964, 33936, 33964, 33933, 33936, 33978, 33948, 33977, 33978, 33954, 33950, 33944, 33920, 33942, 33952, 33937, 33935, 33952, 33921, 33944, 33980, 33957, 33959, 33982, 33965, 33981, 33967, 33980, 33959, 33981, 33937, 33954, 33922, 33864, 33863, 33924, 33872, 33864, 33962, 33910, 33960, 33970, 33913, 33914, 33974, 33914, 33915, 33962, 33915, 33912, 33932, 33910, 33872, 33987, 33988, 33986, 33991, 33992, 33990, 33997, 33994, 33995, 34001, 33998, 34000, 34001, 34002, 33999, 34012, 34009, 34010,
33989, 33998, 33988, 33993, 34006, 33992, 34010, 33996, 33997, 33987, 33992, 33989, 34007, 34000, 34006, 34001, 34008, 34003, 33996, 33991, 33994, 34005, 34002, 34003, 34009, 34008, 34007, 34011, 34003, 34008, 34009, 33993, 33996, 33992, 34000, 33989, 34014, 34015, 34013, 34018, 34014, 34017, 34022, 34019, 34020, 34024, 34025, 34023, 34030, 34027, 34028, 34034, 34031, 34032, 34028, 34021, 34022, 34032, 34025, 34026, 34015, 34024, 34023, 34026, 34016, 34018, 34025, 34020, 34023, 34033, 34028, 34031, 34031, 34022, 34025, 34038, 34035, 34037, 34038, 34039, 34036, 34041, 34038, 34037, 34042, 34040, 34038, 34047, 34044, 34046, 34049, 34050, 34048, 34055, 34052, 34053, 34061, 34058, 34060, 34047, 34056, 34045, 34055, 34063, 34054, 34050, 34047, 34046, 34064, 34060, 34063, 34054, 34049, 34052, 34060, 34056, 34057, 34062, 34060, 34057, 34054, 34062, 34051, 34051, 34057, 34047, 34005, 34035, 34004, 34012, 34041, 34011, 34011, 34037, 34005, 34049, 34029, 34030, 34053, 34033, 34034, 34052, 34030, 34033, 34021, 33995, 34019, 34029, 34010, 34027, 34027, 33997, 34021, 34046, 34039, 34040, 34048, 34043, 34042, 34050, 34040, 34043, 34048, 34012, 34029, 33987, 33983, 33984, 33990, 33984, 33985, 33555, 33839, 33554, 33845, 33558, 33844, 33854, 33561, 33563,
33561, 33841, 33555, 33854, 33559, 33845, 33623, 33916, 33622, 33623, 33920, 33917, 33939, 33636, 33938, 33641, 33942, 33640, 33939, 33641, 33637, 33942, 33625, 33640, 33759, 33983, 33758, 33759, 33985, 33984, 33991, 33800, 33801, 33995, 33804, 33805, 33994, 33801, 33804, 34015, 33818, 34013, 34020, 33822, 33823, 33824, 34020, 33823, 34015, 33824, 33819, 34019, 33805, 33822, 33800, 33985, 33761, 31705, 31935, 31703, 31945, 31712, 31713, 31712, 31943, 31709, 31705, 31943, 31936, 31952, 31731, 31732, 31956, 31735, 31736, 31955, 31732, 31735, 31951, 31713, 31731, 34066, 34067, 34065, 34070, 34071, 34069, 34076, 34073, 34074, 34078, 34079, 34077, 34082, 34083, 34081, 34088, 34085, 34086, 34068, 34077, 34067, 34072, 34081, 34071, 34086, 34075, 34076, 34069, 34068, 34066, 34081, 34080, 34078, 34075, 34070, 34073, 34087, 34082, 34085, 34085, 34072, 34075, 34071, 34078, 34068, 34092, 34089, 34091, 34094, 34095, 34093, 34100, 34097, 34099, 34104, 34105, 34103, 34098, 34091, 34097, 34096, 34103, 34095, 34096, 34090, 34092, 34100, 34104, 34098, 34101, 34100, 34099, 34102, 34106, 34100, 34104, 34092, 34098, 34111, 34108, 34110, 34111, 34112, 34109, 34120, 34117, 34118, 34122, 34123, 34121, 34132, 34129, 34130, 34113, 34121, 34112, 34130, 34119, 34120,
34115, 34110, 34114, 34111, 34116, 34113, 34127, 34124, 34122, 34117, 34116, 34115, 34126, 34123, 34124, 34129, 34128, 34127, 34128, 34126, 34124, 34119, 34127, 34116, 34116, 34122, 34113, 34134, 34135, 34133, 34138, 34134, 34137, 34142, 34139, 34140, 34144, 34145, 34143, 34150, 34147, 34148, 34154, 34151, 34152, 34148, 34141, 34142, 34152, 34145, 34146, 34136, 34143, 34135, 34146, 34136, 34138, 34145, 34140, 34143, 34153, 34148, 34151, 34151, 34142, 34145, 34080, 34108, 34079, 34084, 34114, 34083, 34118, 34087, 34088, 34117, 34084, 34087, 34080, 34114, 34110, 34135, 34101, 34133, 34140, 34105, 34106, 34107, 34140, 34106, 34135, 34107, 34102, 34095, 34074, 34093, 34105, 34086, 34103, 34103, 34076, 34095, 34141, 34118, 34139, 34149, 34130, 34147, 34147, 34120, 34141, 34139, 34088, 34105, 34159, 34160, 34158, 34165, 34166, 34164, 34166, 34162, 34164, 34169, 34170, 34168, 34160, 34155, 34158, 34161, 34168, 34160, 34166, 34174, 34173, 34173, 34163, 34166, 34162, 34161, 34159, 34172, 34171, 34169, 34163, 34169, 34161, 34168, 34156, 34160, 34181, 34182, 34180, 34179, 34181, 34178, 34185, 34186, 34184, 34194, 34191, 34192, 34176, 34185, 34175, 34183, 34189, 34182, 34179, 34190, 34183, 34186, 34192, 34191, 34184, 34179, 34178, 34193, 34188, 34191,
34179, 34191, 34188, 34176, 34192, 34187, 34180, 34167, 34165, 34167, 34189, 34174, 34198, 34195, 34197, 34198, 34199, 34196, 34202, 34197, 34201, 34198, 34203, 34200, 34209, 34206, 34207, 34213, 34210, 34211, 34217, 34214, 34216, 34217, 34218, 34215, 34223, 34220, 34222, 34228, 34229, 34227, 34205, 34214, 34204, 34209, 34224, 34208, 34228, 34212, 34213, 34224, 34217, 34216, 34217, 34226, 34219, 34230, 34222, 34229, 34212, 34207, 34210, 34222, 34218, 34219, 34227, 34226, 34225, 34229, 34219, 34226, 34227, 34209, 34212, 34208, 34216, 34205, 34205, 34199, 34200, 34206, 34203, 34202, 34203, 34205, 34200, 34234, 34231, 34232, 34238, 34235, 34236, 34244, 34241, 34242, 34248, 34245, 34246, 34242, 34233, 34234, 34246, 34237, 34238, 34237, 34232, 34235, 34247, 34242, 34245, 34245, 34234, 34237, 34252, 34249, 34250, 34256, 34253, 34254, 34260, 34257, 34259, 34264, 34261, 34262, 34268, 34265, 34266, 34262, 34251, 34252, 34266, 34255, 34256, 34268, 34259, 34267, 34255, 34250, 34253, 34265, 34264, 34262, 34265, 34252, 34255, 34272, 34269, 34271, 34276, 34273, 34275, 34277, 34272, 34271, 34280, 34275, 34279, 34275, 34270, 34272, 34279, 34272, 34278, 34284, 34281, 34283, 34287, 34284, 34283, 34286, 34282, 34284, 34289, 34284, 34288, 34240, 34269, 34239,
34244, 34277, 34243, 34248, 34279, 34247, 34247, 34278, 34244, 34243, 34271, 34240, 34286, 34257, 34285, 34288, 34263, 34264, 34289, 34264, 34267, 34286, 34267, 34259, 34251, 34236, 34249, 34263, 34246, 34261, 34261, 34238, 34251, 34283, 34274, 34276, 34280, 34283, 34276, 34263, 34280, 34248, 34126, 34195, 34125, 34132, 34201, 34131, 34126, 34201, 34197, 34207, 34149, 34150, 34211, 34153, 34154, 34210, 34150, 34153, 34206, 34132, 34149, 34232, 34170, 34171, 34174, 34235, 34173, 34235, 34171, 34173, 34250, 34189, 34190, 34254, 34193, 34194, 34193, 34250, 34190, 34189, 34236, 34174, 34155, 34138, 34137, 34170, 34152, 34168, 34138, 34157, 34146, 34157, 34152, 34146, 34233, 34211, 34231, 34240, 34221, 34223, 34241, 34230, 34228, 34230, 34240, 34223, 34241, 34213, 34233, 34231, 34154, 34170, 34293, 34294, 34292, 34300, 34297, 34298, 34302, 34303, 34301, 34310, 34307, 34308, 34291, 34301, 34290, 34300, 34307, 34299, 34302, 34306, 34304, 34297, 34295, 34293, 34294, 34291, 34290, 34309, 34305, 34307, 34291, 34299, 34296, 34307, 34296, 34299, 34296, 34302, 34291, 34314, 34311, 34313, 34318, 34315, 34316, 34322, 34319, 34320, 34326, 34323, 34325, 34328, 34323, 34327, 34332, 34329, 34330, 34326, 34313, 34324, 34330, 34317, 34318, 34334, 34321, 34322,
34314, 34320, 34312, 34333, 34326, 34325, 34335, 34325, 34328, 34321, 34316, 34319, 34335, 34330, 34333, 34321, 34330, 34318, 34334, 34314, 34326, 34337, 34338, 34336, 34338, 34340, 34336, 34345, 34342, 34343, 34349, 34346, 34347, 34353, 34350, 34351, 34357, 34354, 34355, 34345, 34350, 34344, 34355, 34348, 34349, 34338, 34343, 34342, 34342, 34341, 34338, 34346, 34345, 34343, 34356, 34351, 34354, 34354, 34345, 34348, 34363, 34360, 34362, 34365, 34366, 34364, 34375, 34372, 34374, 34375, 34370, 34373, 34379, 34376, 34377, 34371, 34358, 34370, 34377, 34366, 34367, 34378, 34375, 34374, 34375, 34377, 34371, 34369, 34361, 34363, 34363, 34358, 34359, 34369, 34365, 34368, 34367, 34363, 34359, 34371, 34367, 34359, 34304, 34337, 34303, 34347, 34309, 34310, 34346, 34306, 34309, 34304, 34343, 34339, 34361, 34327, 34360, 34365, 34331, 34332, 34368, 34332, 34335, 34361, 34335, 34328, 34317, 34298, 34315, 34331, 34308, 34329, 34329, 34300, 34317, 34366, 34347, 34364, 34376, 34357, 34355, 34376, 34349, 34366, 34364, 34310, 34331, 34385, 34382, 34384, 34385, 34380, 34383, 34389, 34386, 34387, 34393, 34390, 34391, 34399, 34396, 34398, 34399, 34394, 34397, 34405, 34402, 34403, 34401, 34382, 34400, 34389, 34402, 34388, 34404, 34399, 34398, 34399, 34403, 34395,
34393, 34384, 34392, 34393, 34381, 34385, 34390, 34389, 34387, 34401, 34394, 34395, 34406, 34395, 34403, 34406, 34389, 34392, 34401, 34392, 34384, 34408, 34380, 34381, 34391, 34408, 34381, 34387, 34352, 34353, 34357, 34390, 34356, 34390, 34353, 34356, 34408, 34372, 34407, 34409, 34374, 34408, 34409, 34357, 34378, 34340, 34176, 34175, 34350, 34194, 34192, 34341, 34177, 34176, 34350, 34177, 34344, 34388, 34254, 34386, 34398, 34258, 34260, 34404, 34266, 34402, 34268, 34398, 34260, 34402, 34256, 34388, 34386, 34194, 34352, 34411, 34412, 34410, 34415, 34416, 34414, 34421, 34418, 34419, 34425, 34422, 34423, 34429, 34426, 34427, 34433, 34430, 34431, 34427, 34420, 34421, 34431, 34424, 34425, 34413, 34418, 34412, 34417, 34422, 34416, 34414, 34413, 34411, 34424, 34419, 34422, 34432, 34427, 34430, 34416, 34419, 34413, 34430, 34421, 34424, 34435, 34436, 34434, 34439, 34440, 34438, 34445, 34442, 34443, 34449, 34446, 34447, 34453, 34450, 34451, 34457, 34454, 34455, 34451, 34444, 34445, 34455, 34448, 34449, 34437, 34442, 34436, 34441, 34446, 34440, 34438, 34437, 34435, 34448, 34443, 34446, 34456, 34451, 34454, 34440, 34443, 34437, 34454, 34445, 34448, 34461, 34458, 34460, 34465, 34462, 34464, 34469, 34466, 34467, 34473, 34470, 34471, 34469, 34460, 34468,
34473, 34464, 34472, 34464, 34459, 34461, 34472, 34467, 34470, 34472, 34461, 34469, 34477, 34478, 34476, 34483, 34480, 34481, 34485, 34486, 34484, 34491, 34488, 34489, 34495, 34492, 34493, 34475, 34484, 34474, 34479, 34488, 34478, 34493, 34482, 34483, 34488, 34487, 34485, 34482, 34477, 34480, 34494, 34489, 34492, 34492, 34479, 34482, 34478, 34485, 34475, 34467, 34428, 34429, 34471, 34432, 34433, 34470, 34429, 34432, 34477, 34452, 34453, 34481, 34456, 34457, 34480, 34453, 34456, 34434, 34417, 34415, 34444, 34423, 34442, 34452, 34431, 34450, 34436, 34423, 34417, 34450, 34425, 34444, 34475, 34463, 34465, 34476, 34473, 34471, 34478, 34465, 34473, 34476, 34433, 34452, 34497, 34498, 34496, 34501, 34502, 34500, 34507, 34504, 34505, 34511, 34508, 34509, 34515, 34512, 34513, 34519, 34516, 34517, 34513, 34506, 34507, 34517, 34510, 34511, 34499, 34504, 34498, 34503, 34508, 34502, 34500, 34499, 34497, 34510, 34505, 34508, 34518, 34513, 34516, 34502, 34505, 34499, 34516, 34507, 34510, 34521, 34522, 34520, 34525, 34526, 34524, 34531, 34528, 34529, 34535, 34532, 34533, 34539, 34536, 34537, 34543, 34540, 34541, 34537, 34530, 34531, 34541, 34534, 34535, 34523, 34528, 34522, 34527, 34532, 34526, 34524, 34523, 34521, 34534, 34529, 34532, 34542, 34537, 34540,
34526, 34529, 34523, 34540, 34531, 34534, 34547, 34544, 34546, 34551, 34548, 34549, 34555, 34552, 34553, 34559, 34556, 34558, 34563, 34560, 34561, 34557, 34544, 34556, 34561, 34550, 34551, 34555, 34546, 34554, 34565, 34557, 34559, 34554, 34549, 34552, 34565, 34561, 34564, 34564, 34551, 34554, 34564, 34546, 34557, 34569, 34566, 34568, 34573, 34570, 34572, 34577, 34574, 34575, 34581, 34578, 34579, 34577, 34568, 34576, 34581, 34572, 34580, 34572, 34567, 34569, 34580, 34575, 34578, 34580, 34569, 34577, 34549, 34514, 34515, 34553, 34518, 34519, 34552, 34515, 34518, 34575, 34538, 34539, 34579, 34542, 34543, 34578, 34539, 34542, 34520, 34503, 34501, 34530, 34509, 34528, 34538, 34517, 34536, 34522, 34509, 34503, 34536, 34511, 34530, 34568, 34545, 34547, 34576, 34553, 34574, 34576, 34547, 34555, 34574, 34519, 34538, 34583, 34584, 34582, 34589, 34586, 34587, 34593, 34590, 34591, 34595, 34596, 34594, 34601, 34598, 34599, 34605, 34602, 34603, 34585, 34594, 34584, 34599, 34588, 34589, 34603, 34592, 34593, 34586, 34585, 34583, 34598, 34597, 34595, 34592, 34587, 34590, 34604, 34599, 34602, 34602, 34589, 34592, 34588, 34595, 34585, 34607, 34608, 34606, 34613, 34610, 34611, 34617, 34614, 34615, 34619, 34620, 34618, 34625, 34622, 34623, 34629, 34626, 34627,
34609, 34618, 34608, 34623, 34612, 34613, 34627, 34616, 34617, 34610, 34609, 34607, 34622, 34621, 34619, 34616, 34611, 34614, 34628, 34623, 34626, 34626, 34613, 34616, 34612, 34619, 34609, 34597, 34606, 34596, 34611, 34600, 34601, 34615, 34604, 34605, 34614, 34601, 34604, 34600, 34607, 34597, 34633, 34630, 34632, 34637, 34634, 34635, 34643, 34640, 34642, 34647, 34644, 34645, 34641, 34632, 34640, 34645, 34636, 34637, 34639, 34631, 34633, 34649, 34641, 34643, 34639, 34635, 34638, 34649, 34645, 34648, 34648, 34637, 34639, 34648, 34633, 34641, 34653, 34650, 34652, 34657, 34654, 34655, 34663, 34660, 34662, 34667, 34664, 34665, 34661, 34652, 34660, 34665, 34656, 34657, 34659, 34651, 34653, 34669, 34661, 34663, 34659, 34655, 34658, 34669, 34665, 34668, 34668, 34657, 34659, 34668, 34653, 34661, 34651, 34642, 34650, 34655, 34646, 34647, 34658, 34647, 34649, 34658, 34643, 34651, 34487, 34582, 34486, 34587, 34490, 34491, 34591, 34494, 34495, 34590, 34491, 34494, 34490, 34583, 34487, 34631, 34558, 34630, 34635, 34562, 34563, 34638, 34563, 34565, 34638, 34559, 34631, 34496, 34441, 34439, 34506, 34447, 34504, 34514, 34455, 34512, 34498, 34447, 34441, 34512, 34449, 34506, 34550, 34481, 34548, 34562, 34493, 34560, 34560, 34483, 34550, 34548, 34457, 34514,
34636, 34591, 34634, 34646, 34603, 34644, 34644, 34593, 34636, 34656, 34615, 34654, 34666, 34627, 34664, 34664, 34617, 34656, 34654, 34605, 34646, 34634, 34495, 34562, 34671, 34672, 34670, 34675, 34676, 34674, 34681, 34678, 34679, 34685, 34682, 34683, 34689, 34686, 34687, 34693, 34690, 34691, 34687, 34680, 34681, 34691, 34684, 34685, 34673, 34678, 34672, 34677, 34682, 34676, 34674, 34673, 34671, 34684, 34679, 34682, 34692, 34687, 34690, 34676, 34679, 34673, 34690, 34681, 34684, 34695, 34696, 34694, 34699, 34700, 34698, 34703, 34699, 34702, 34707, 34704, 34705, 34711, 34708, 34709, 34715, 34712, 34714, 34719, 34716, 34717, 34723, 34720, 34721, 34713, 34702, 34712, 34717, 34706, 34707, 34721, 34710, 34711, 34697, 34704, 34696, 34701, 34708, 34700, 34711, 34701, 34703, 34723, 34713, 34715, 34698, 34697, 34695, 34710, 34705, 34708, 34722, 34717, 34720, 34700, 34705, 34697, 34720, 34707, 34710, 34721, 34703, 34713, 34727, 34724, 34726, 34733, 34730, 34731, 34737, 34734, 34735, 34739, 34740, 34738, 34745, 34742, 34743, 34729, 34738, 34728, 34737, 34742, 34736, 34733, 34726, 34732, 34742, 34741, 34739, 34729, 34725, 34727, 34736, 34731, 34734, 34736, 34727, 34733, 34736, 34739, 34729, 34749, 34746, 34748, 34753, 34750, 34751, 34757, 34754, 34755,
34761, 34758, 34760, 34765, 34762, 34763, 34769, 34766, 34767, 34759, 34748, 34758, 34763, 34752, 34753, 34767, 34756, 34757, 34757, 34747, 34749, 34769, 34759, 34761, 34756, 34751, 34754, 34768, 34763, 34766, 34766, 34753, 34756, 34767, 34749, 34759, 34731, 34688, 34689, 34735, 34692, 34693, 34734, 34689, 34692, 34747, 34714, 34746, 34751, 34718, 34719, 34755, 34722, 34723, 34754, 34719, 34722, 34755, 34715, 34747, 34694, 34677, 34675, 34706, 34683, 34704, 34718, 34691, 34716, 34696, 34683, 34677, 34716, 34685, 34706, 34752, 34735, 34750, 34764, 34743, 34762, 34762, 34737, 34752, 34750, 34693, 34718, 34771, 34772, 34770, 34777, 34774, 34775, 34779, 34780, 34778, 34785, 34782, 34783, 34773, 34778, 34772, 34783, 34776, 34777, 34774, 34773, 34771, 34782, 34781, 34779, 34776, 34779, 34773, 34789, 34786, 34788, 34793, 34790, 34791, 34797, 34794, 34795, 34801, 34798, 34800, 34805, 34802, 34803, 34809, 34806, 34807, 34799, 34788, 34798, 34803, 34792, 34793, 34807, 34796, 34797, 34797, 34787, 34789, 34809, 34799, 34801, 34796, 34791, 34794, 34808, 34803, 34806, 34806, 34793, 34796, 34807, 34789, 34799, 34811, 34812, 34810, 34817, 34814, 34815, 34819, 34820, 34818, 34825, 34822, 34823, 34813, 34818, 34812, 34823, 34816, 34817, 34814, 34813, 34811,
34822, 34821, 34819, 34816, 34819, 34813, 34829, 34826, 34828, 34833, 34830, 34831, 34837, 34834, 34835, 34841, 34838, 34840, 34845, 34842, 34843, 34849, 34846, 34847, 34839, 34828, 34838, 34843, 34832, 34833, 34847, 34836, 34837, 34837, 34827, 34829, 34849, 34839, 34841, 34836, 34831, 34834, 34848, 34843, 34846, 34846, 34833, 34836, 34847, 34829, 34839, 34781, 34810, 34780, 34815, 34784, 34785, 34784, 34811, 34781, 34827, 34800, 34826, 34831, 34804, 34805, 34835, 34808, 34809, 34834, 34805, 34808, 34835, 34801, 34827, 34792, 34775, 34790, 34804, 34783, 34802, 34802, 34777, 34792, 34832, 34815, 34830, 34844, 34823, 34842, 34842, 34817, 34832, 34830, 34785, 34804, 34741, 34770, 34740, 34775, 34744, 34745, 34744, 34771, 34741, 34787, 34760, 34786, 34791, 34764, 34765, 34795, 34768, 34769, 34794, 34765, 34768, 34795, 34761, 34787, 34790, 34745, 34764, 34851, 34852, 34850, 34857, 34854, 34855, 34861, 34858, 34859, 34863, 34864, 34862, 34869, 34866, 34867, 34873, 34870, 34871, 34853, 34862, 34852, 34867, 34856, 34857, 34871, 34860, 34861, 34854, 34853, 34851, 34866, 34865, 34863, 34860, 34855, 34858, 34872, 34867, 34870, 34870, 34857, 34860, 34856, 34863, 34853, 34875, 34876, 34874, 34881, 34878, 34879, 34885, 34882, 34883, 34887, 34888, 34886,
34893, 34890, 34891, 34897, 34894, 34895, 34877, 34886, 34876, 34891, 34880, 34881, 34895, 34884, 34885, 34878, 34877, 34875, 34890, 34889, 34887, 34884, 34879, 34882, 34896, 34891, 34894, 34894, 34881, 34884, 34880, 34887, 34877, 34865, 34874, 34864, 34879, 34868, 34869, 34883, 34872, 34873, 34882, 34869, 34872, 34868, 34875, 34865, 34901, 34898, 34900, 34905, 34902, 34903, 34911, 34908, 34910, 34915, 34912, 34913, 34909, 34900, 34908, 34913, 34904, 34905, 34907, 34899, 34901, 34917, 34909, 34911, 34907, 34903, 34906, 34917, 34913, 34916, 34916, 34905, 34907, 34916, 34901, 34909, 34921, 34918, 34920, 34925, 34922, 34923, 34931, 34928, 34930, 34935, 34932, 34933, 34929, 34920, 34928, 34933, 34924, 34925, 34927, 34919, 34921, 34937, 34929, 34931, 34927, 34923, 34926, 34937, 34933, 34936, 34936, 34925, 34927, 34936, 34921, 34929, 34919, 34910, 34918, 34923, 34914, 34915, 34926, 34915, 34917, 34926, 34911, 34919, 34939, 34940, 34938, 34945, 34942, 34943, 34949, 34946, 34947, 34953, 34950, 34952, 34957, 34954, 34956, 34941, 34950, 34940, 34959, 34944, 34945, 34961, 34948, 34949, 34942, 34941, 34939, 34959, 34952, 34958, 34961, 34956, 34960, 34948, 34943, 34946, 34956, 34951, 34953, 34960, 34953, 34959, 34960, 34945, 34948, 34944, 34952, 34941,
34965, 34962, 34964, 34969, 34966, 34967, 34975, 34972, 34974, 34977, 34964, 34976, 34979, 34968, 34969, 34979, 34974, 34978, 34971, 34963, 34965, 34971, 34967, 34970, 34977, 34973, 34975, 34980, 34975, 34979, 34980, 34969, 34971, 34980, 34965, 34977, 34889, 34938, 34888, 34943, 34892, 34893, 34947, 34896, 34897, 34946, 34893, 34896, 34892, 34939, 34889, 34963, 34930, 34962, 34967, 34934, 34935, 34970, 34935, 34937, 34970, 34931, 34963, 34904, 34859, 34902, 34914, 34871, 34912, 34912, 34861, 34904, 34924, 34883, 34922, 34934, 34895, 34932, 34932, 34885, 34924, 34922, 34873, 34914, 34968, 34947, 34966, 34974, 34955, 34957, 34978, 34957, 34961, 34978, 34949, 34968, 34966, 34897, 34934, 34982, 34983, 34981, 34988, 34985, 34986, 34990, 34991, 34989, 34996, 34993, 34994, 34984, 34989, 34983, 34994, 34987, 34988, 34985, 34984, 34982, 34993, 34992, 34990, 34987, 34990, 34984, 35000, 34997, 34999, 35004, 35001, 35002, 35008, 35005, 35006, 35012, 35009, 35011, 35016, 35013, 35014, 35020, 35017, 35018, 35010, 34999, 35009, 35014, 35003, 35004, 35018, 35007, 35008, 35008, 34998, 35000, 35020, 35010, 35012, 35007, 35002, 35005, 35019, 35014, 35017, 35017, 35004, 35007, 35018, 35000, 35010, 35022, 35023, 35021, 35028, 35025, 35026, 35030, 35031, 35029,
35036, 35033, 35034, 35024, 35029, 35023, 35034, 35027, 35028, 35025, 35024, 35022, 35033, 35032, 35030, 35027, 35030, 35024, 35040, 35037, 35039, 35044, 35041, 35042, 35048, 35045, 35046, 35052, 35049, 35051, 35056, 35053, 35054, 35060, 35057, 35058, 35050, 35039, 35049, 35054, 35043, 35044, 35058, 35047, 35048, 35048, 35038, 35040, 35060, 35050, 35052, 35047, 35042, 35045, 35059, 35054, 35057, 35057, 35044, 35047, 35058, 35040, 35050, 34992, 35021, 34991, 35026, 34995, 34996, 34995, 35022, 34992, 35038, 35011, 35037, 35042, 35015, 35016, 35046, 35019, 35020, 35045, 35016, 35019, 35046, 35012, 35038, 35003, 34986, 35001, 35015, 34994, 35013, 35013, 34988, 35003, 35043, 35026, 35041, 35055, 35034, 35053, 35053, 35028, 35043, 35041, 34996, 35015, 35062, 35063, 35061, 35068, 35065, 35066, 35070, 35071, 35069, 35076, 35073, 35074, 35064, 35069, 35063, 35074, 35067, 35068, 35065, 35064, 35062, 35073, 35072, 35070, 35067, 35070, 35064, 35080, 35077, 35079, 35084, 35081, 35082, 35088, 35085, 35086, 35092, 35089, 35091, 35096, 35093, 35094, 35100, 35097, 35098, 35090, 35079, 35089, 35094, 35083, 35084, 35098, 35087, 35088, 35088, 35078, 35080, 35100, 35090, 35092, 35087, 35082, 35085, 35099, 35094, 35097, 35097, 35084, 35087, 35098, 35080, 35090,
35104, 35101, 35103, 35106, 35103, 35105, 35110, 35107, 35109, 35114, 35111, 35113, 35116, 35109, 35115, 35118, 35113, 35117, 35113, 35108, 35110, 35117, 35110, 35116, 35072, 35101, 35071, 35106, 35075, 35076, 35075, 35103, 35072, 35114, 35091, 35112, 35116, 35095, 35096, 35118, 35099, 35100, 35117, 35096, 35099, 35118, 35092, 35114, 35083, 35066, 35081, 35095, 35074, 35093, 35093, 35068, 35083, 35109, 35102, 35104, 35115, 35104, 35106, 35115, 35076, 35095, 35032, 35061, 35031, 35066, 35035, 35036, 35035, 35062, 35032, 35078, 35051, 35077, 35082, 35055, 35056, 35086, 35059, 35060, 35085, 35056, 35059, 35086, 35052, 35078, 35081, 35036, 35055, 34621, 34850, 34620, 34855, 34624, 34625, 34859, 34628, 34629, 34858, 34625, 34628, 34624, 34851, 34621, 34899, 34662, 34898, 34903, 34666, 34667, 34906, 34667, 34669, 34906, 34663, 34899, 34902, 34629, 34666, 34821, 34981, 34820, 34986, 34824, 34825, 34824, 34982, 34821, 34998, 34840, 34997, 35002, 34844, 34845, 35006, 34848, 34849, 35005, 34845, 34848, 35006, 34841, 34998, 35001, 34825, 34844, 34670, 34527, 34525, 34680, 34533, 34678, 34688, 34541, 34686, 34672, 34533, 34527, 34686, 34535, 34680, 34726, 34571, 34573, 34732, 34579, 34730, 34732, 34573, 34581, 34730, 34543, 34688, 35120, 35121, 35119,
35124, 35125, 35123, 35130, 35127, 35128, 35134, 35131, 35132, 35138, 35135, 35136, 35142, 35139, 35140, 35136, 35129, 35130, 35140, 35133, 35134, 35122, 35127, 35121, 35126, 35131, 35125, 35123, 35122, 35120, 35133, 35128, 35131, 35141, 35136, 35139, 35125, 35128, 35122, 35139, 35130, 35133, 35144, 35145, 35143, 35148, 35149, 35147, 35154, 35151, 35152, 35158, 35155, 35156, 35162, 35159, 35160, 35166, 35163, 35164, 35160, 35153, 35154, 35164, 35157, 35158, 35146, 35151, 35145, 35150, 35155, 35149, 35147, 35146, 35144, 35157, 35152, 35155, 35165, 35160, 35163, 35149, 35152, 35146, 35163, 35154, 35157, 35170, 35167, 35169, 35174, 35171, 35173, 35178, 35175, 35176, 35182, 35179, 35180, 35178, 35169, 35177, 35182, 35173, 35181, 35173, 35168, 35170, 35181, 35176, 35179, 35181, 35170, 35178, 35186, 35183, 35185, 35190, 35187, 35189, 35194, 35191, 35192, 35198, 35195, 35196, 35194, 35185, 35193, 35198, 35189, 35197, 35189, 35184, 35186, 35197, 35192, 35195, 35197, 35186, 35194, 35176, 35137, 35138, 35180, 35141, 35142, 35179, 35138, 35141, 35192, 35161, 35162, 35196, 35165, 35166, 35195, 35162, 35165, 35143, 35126, 35124, 35153, 35132, 35151, 35161, 35140, 35159, 35145, 35132, 35126, 35159, 35134, 35153, 35185, 35172, 35174, 35193, 35180, 35191,
35193, 35174, 35182, 35191, 35142, 35161, 35200, 35201, 35199, 35204, 35205, 35203, 35210, 35207, 35208, 35214, 35211, 35212, 35218, 35215, 35216, 35222, 35219, 35220, 35216, 35209, 35210, 35220, 35213, 35214, 35202, 35207, 35201, 35206, 35211, 35205, 35203, 35202, 35200, 35213, 35208, 35211, 35221, 35216, 35219, 35205, 35208, 35202, 35219, 35210, 35213, 35224, 35225, 35223, 35228, 35229, 35227, 35234, 35231, 35232, 35238, 35235, 35236, 35242, 35239, 35240, 35246, 35243, 35244, 35240, 35233, 35234, 35244, 35237, 35238, 35226, 35231, 35225, 35230, 35235, 35229, 35227, 35226, 35224, 35237, 35232, 35235, 35245, 35240, 35243, 35229, 35232, 35226, 35243, 35234, 35237, 35250, 35247, 35249, 35254, 35251, 35253, 35258, 35255, 35256, 35262, 35259, 35260, 35258, 35249, 35257, 35262, 35253, 35261, 35253, 35248, 35250, 35261, 35256, 35259, 35261, 35250, 35258, 35266, 35263, 35265, 35270, 35267, 35269, 35274, 35271, 35272, 35278, 35275, 35276, 35274, 35265, 35273, 35278, 35269, 35277, 35269, 35264, 35266, 35277, 35272, 35275, 35277, 35266, 35274, 35256, 35217, 35218, 35260, 35221, 35222, 35259, 35218, 35221, 35272, 35241, 35242, 35276, 35245, 35246, 35275, 35242, 35245, 35223, 35206, 35204, 35233, 35212, 35231, 35241, 35220, 35239, 35225, 35212, 35206,
35239, 35214, 35233, 35265, 35252, 35254, 35273, 35260, 35271, 35273, 35254, 35262, 35271, 35222, 35241, 35199, 35150, 35148, 35209, 35156, 35207, 35217, 35164, 35215, 35201, 35156, 35150, 35215, 35158, 35209, 35249, 35188, 35190, 35257, 35196, 35255, 35257, 35190, 35198, 35255, 35166, 35217, 35280, 35281, 35279, 35284, 35285, 35283, 35290, 35287, 35288, 35294, 35291, 35292, 35298, 35295, 35296, 35302, 35299, 35300, 35296, 35289, 35290, 35300, 35293, 35294, 35282, 35287, 35281, 35286, 35291, 35285, 35283, 35282, 35280, 35293, 35288, 35291, 35301, 35296, 35299, 35285, 35288, 35282, 35299, 35290, 35293, 35304, 35305, 35303, 35308, 35309, 35307, 35312, 35308, 35311, 35316, 35313, 35314, 35320, 35317, 35318, 35324, 35321, 35323, 35328, 35325, 35326, 35332, 35329, 35330, 35322, 35311, 35321, 35326, 35315, 35316, 35330, 35319, 35320, 35306, 35313, 35305, 35310, 35317, 35309, 35320, 35310, 35312, 35332, 35322, 35324, 35307, 35306, 35304, 35319, 35314, 35317, 35331, 35326, 35329, 35309, 35314, 35306, 35329, 35316, 35319, 35330, 35312, 35322, 35336, 35333, 35335, 35340, 35337, 35339, 35344, 35341, 35342, 35348, 35345, 35346, 35344, 35335, 35343, 35348, 35339, 35347, 35339, 35334, 35336, 35347, 35342, 35345, 35347, 35336, 35344, 35352, 35349, 35351,
35358, 35355, 35357, 35358, 35353, 35356, 35362, 35359, 35360, 35366, 35363, 35364, 35362, 35351, 35361, 35366, 35357, 35365, 35366, 35354, 35358, 35357, 35350, 35352, 35365, 35360, 35363, 35365, 35352, 35362, 35342, 35297, 35298, 35346, 35301, 35302, 35345, 35298, 35301, 35354, 35323, 35353, 35360, 35327, 35328, 35364, 35331, 35332, 35363, 35328, 35331, 35364, 35324, 35354, 35303, 35286, 35284, 35315, 35292, 35313, 35327, 35300, 35325, 35305, 35292, 35286, 35325, 35294, 35315, 35351, 35338, 35340, 35361, 35346, 35359, 35361, 35340, 35348, 35359, 35302, 35327, 35279, 35230, 35228, 35289, 35236, 35287, 35297, 35244, 35295, 35281, 35236, 35230, 35295, 35238, 35289, 35335, 35268, 35270, 35343, 35276, 35341, 35343, 35270, 35278, 35341, 35246, 35297, 35368, 35369, 35367, 35372, 35373, 35371, 35382, 35379, 35380, 35386, 35383, 35384, 35380, 35375, 35376, 35384, 35377, 35378, 35369, 35376, 35375, 35373, 35378, 35377, 35371, 35370, 35368, 35385, 35380, 35383, 35370, 35377, 35376, 35383, 35376, 35377, 35388, 35389, 35387, 35392, 35393, 35391, 35402, 35399, 35400, 35406, 35403, 35404, 35396, 35399, 35395, 35398, 35403, 35397, 35389, 35396, 35395, 35393, 35398, 35397, 35391, 35390, 35388, 35405, 35400, 35403, 35390, 35397, 35396, 35397, 35400, 35396,
35410, 35407, 35408, 35414, 35411, 35412, 35418, 35415, 35417, 35422, 35419, 35421, 35424, 35419, 35423, 35428, 35425, 35426, 35426, 35409, 35410, 35430, 35413, 35414, 35427, 35418, 35417, 35430, 35421, 35429, 35424, 35429, 35421, 35413, 35408, 35411, 35424, 35416, 35418, 35431, 35426, 35429, 35428, 35424, 35418, 35413, 35426, 35410, 35435, 35432, 35433, 35439, 35436, 35437, 35443, 35440, 35442, 35447, 35444, 35446, 35435, 35448, 35434, 35439, 35450, 35438, 35448, 35443, 35442, 35450, 35447, 35446, 35438, 35433, 35436, 35446, 35441, 35443, 35449, 35446, 35443, 35438, 35449, 35435, 35408, 35381, 35382, 35412, 35385, 35386, 35411, 35382, 35385, 35433, 35401, 35402, 35437, 35405, 35406, 35436, 35402, 35405, 35387, 35374, 35372, 35401, 35384, 35399, 35374, 35395, 35378, 35399, 35378, 35395, 35434, 35412, 35432, 35442, 35420, 35422, 35430, 35442, 35422, 35448, 35414, 35434, 35432, 35386, 35401, 35453, 35454, 35452, 35457, 35458, 35456, 35467, 35464, 35465, 35471, 35468, 35469, 35465, 35460, 35461, 35469, 35462, 35463, 35455, 35460, 35454, 35459, 35462, 35458, 35456, 35455, 35453, 35470, 35465, 35468, 35458, 35461, 35455, 35468, 35461, 35462, 35473, 35474, 35472, 35477, 35478, 35476, 35487, 35484, 35485, 35491, 35488, 35489, 35485, 35480, 35481,
35489, 35482, 35483, 35475, 35480, 35474, 35479, 35482, 35478, 35476, 35475, 35473, 35490, 35485, 35488, 35478, 35481, 35475, 35488, 35481, 35482, 35495, 35492, 35493, 35499, 35496, 35497, 35503, 35500, 35502, 35507, 35504, 35506, 35495, 35508, 35494, 35511, 35498, 35499, 35508, 35503, 35502, 35511, 35506, 35510, 35498, 35493, 35496, 35506, 35501, 35503, 35510, 35503, 35509, 35510, 35495, 35498, 35515, 35512, 35513, 35519, 35516, 35517, 35523, 35520, 35522, 35527, 35524, 35526, 35529, 35514, 35515, 35531, 35518, 35519, 35529, 35522, 35528, 35531, 35526, 35530, 35518, 35513, 35516, 35526, 35521, 35523, 35530, 35523, 35529, 35530, 35515, 35518, 35493, 35466, 35467, 35497, 35470, 35471, 35496, 35467, 35470, 35513, 35486, 35487, 35517, 35490, 35491, 35516, 35487, 35490, 35472, 35459, 35457, 35486, 35469, 35484, 35474, 35463, 35459, 35484, 35463, 35480, 35514, 35497, 35512, 35522, 35505, 35507, 35528, 35507, 35511, 35528, 35499, 35514, 35512, 35471, 35486, 35533, 35415, 35532, 35533, 35427, 35417, 35452, 35394, 35392, 35466, 35404, 35464, 35454, 35398, 35394, 35464, 35398, 35460, 35494, 35437, 35492, 35502, 35445, 35447, 35451, 35502, 35447, 35508, 35439, 35494, 35492, 35406, 35466, 35536, 35537, 35535, 35540, 35541, 35539, 35550, 35547, 35548,
35554, 35551, 35552, 35548, 35543, 35544, 35552, 35545, 35546, 35538, 35543, 35537, 35542, 35545, 35541, 35539, 35538, 35536, 35553, 35548, 35551, 35541, 35544, 35538, 35551, 35544, 35545, 35556, 35557, 35555, 35560, 35561, 35559, 35570, 35567, 35569, 35574, 35571, 35572, 35578, 35575, 35576, 35568, 35560, 35567, 35572, 35563, 35564, 35576, 35565, 35566, 35558, 35563, 35557, 35562, 35565, 35561, 35578, 35568, 35570, 35559, 35558, 35556, 35577, 35572, 35575, 35561, 35564, 35558, 35575, 35564, 35565, 35576, 35562, 35568, 35582, 35579, 35580, 35586, 35583, 35584, 35590, 35587, 35589, 35594, 35591, 35593, 35596, 35581, 35582, 35598, 35585, 35586, 35596, 35589, 35595, 35598, 35593, 35597, 35585, 35580, 35583, 35593, 35588, 35590, 35597, 35590, 35596, 35597, 35582, 35585, 35602, 35599, 35601, 35606, 35603, 35604, 35610, 35607, 35608, 35614, 35611, 35613, 35618, 35615, 35617, 35618, 35601, 35616, 35620, 35605, 35606, 35622, 35609, 35610, 35620, 35613, 35619, 35610, 35600, 35602, 35622, 35617, 35621, 35609, 35604, 35607, 35617, 35612, 35614, 35621, 35614, 35620, 35621, 35606, 35609, 35622, 35602, 35618, 35580, 35549, 35550, 35584, 35553, 35554, 35583, 35550, 35553, 35600, 35569, 35599, 35604, 35573, 35574, 35608, 35577, 35578, 35607, 35574, 35577,
35608, 35570, 35600, 35555, 35542, 35540, 35573, 35552, 35571, 35557, 35546, 35542, 35571, 35546, 35563, 35605, 35584, 35603, 35613, 35592, 35594, 35619, 35594, 35598, 35619, 35586, 35605, 35603, 35554, 35573, 35624, 35625, 35623, 35628, 35624, 35627, 35630, 35631, 35629, 35634, 35635, 35633, 35626, 35633, 35625, 35636, 35626, 35628, 35632, 35637, 35631, 35629, 35628, 35627, 35628, 35637, 35636, 35640, 35641, 35639, 35644, 35645, 35643, 35642, 35647, 35641, 35646, 35649, 35645, 35643, 35642, 35640, 35645, 35648, 35642, 35639, 35632, 35630, 35641, 35638, 35632, 35652, 35653, 35651, 35656, 35657, 35655, 35654, 35659, 35653, 35658, 35661, 35657, 35655, 35654, 35652, 35657, 35660, 35654, 35664, 35665, 35663, 35668, 35669, 35667, 35666, 35671, 35665, 35670, 35673, 35669, 35667, 35666, 35664, 35669, 35672, 35666, 35663, 35658, 35656, 35665, 35662, 35658, 35651, 35646, 35644, 35653, 35650, 35646, 35676, 35677, 35675, 35680, 35681, 35679, 35678, 35683, 35677, 35682, 35685, 35681, 35679, 35678, 35676, 35681, 35684, 35678, 35688, 35689, 35687, 35692, 35693, 35691, 35690, 35695, 35689, 35694, 35697, 35693, 35691, 35690, 35688, 35693, 35696, 35690, 35687, 35682, 35680, 35689, 35686, 35682, 35535, 35479, 35477, 35549, 35489, 35547, 35537, 35483, 35479,
35547, 35483, 35543, 35581, 35517, 35579, 35589, 35525, 35527, 35595, 35527, 35531, 35595, 35519, 35581, 35579, 35491, 35549, 35675, 35670, 35668, 35677, 35674, 35670, 35702, 35699, 35700, 35706, 35703, 35704, 35710, 35707, 35708, 35714, 35711, 35712, 35708, 35701, 35702, 35712, 35705, 35706, 35705, 35700, 35703, 35713, 35708, 35711, 35711, 35702, 35705, 35718, 35715, 35716, 35722, 35719, 35720, 35726, 35723, 35724, 35730, 35727, 35728, 35724, 35717, 35718, 35728, 35721, 35722, 35721, 35716, 35719, 35729, 35724, 35727, 35727, 35718, 35721, 35734, 35731, 35733, 35738, 35735, 35737, 35738, 35739, 35736, 35742, 35733, 35741, 35743, 35738, 35737, 35744, 35740, 35738, 35737, 35732, 35734, 35743, 35734, 35742, 35749, 35746, 35748, 35753, 35750, 35752, 35755, 35756, 35754, 35761, 35758, 35759, 35757, 35748, 35756, 35761, 35752, 35760, 35752, 35747, 35749, 35758, 35757, 35755, 35760, 35749, 35757, 35710, 35741, 35709, 35714, 35743, 35713, 35713, 35742, 35710, 35755, 35725, 35726, 35759, 35729, 35730, 35758, 35726, 35729, 35717, 35704, 35715, 35725, 35712, 35723, 35723, 35706, 35717, 35748, 35739, 35740, 35754, 35745, 35744, 35745, 35748, 35740, 35754, 35714, 35725, 35765, 35762, 35763, 35769, 35766, 35767, 35773, 35770, 35771, 35777, 35774, 35775,
35771, 35764, 35765, 35775, 35768, 35769, 35768, 35763, 35766, 35776, 35771, 35774, 35774, 35765, 35768, 35781, 35778, 35779, 35785, 35782, 35783, 35789, 35786, 35787, 35793, 35790, 35791, 35787, 35780, 35781, 35791, 35784, 35785, 35784, 35779, 35782, 35792, 35787, 35790, 35790, 35781, 35784, 35797, 35794, 35796, 35801, 35798, 35800, 35803, 35804, 35802, 35809, 35806, 35807, 35805, 35796, 35804, 35809, 35800, 35808, 35800, 35795, 35797, 35808, 35803, 35806, 35808, 35797, 35805, 35813, 35810, 35812, 35817, 35814, 35816, 35821, 35818, 35819, 35825, 35822, 35823, 35821, 35812, 35820, 35825, 35816, 35824, 35816, 35811, 35813, 35824, 35819, 35822, 35824, 35813, 35821, 35803, 35772, 35773, 35807, 35776, 35777, 35806, 35773, 35776, 35819, 35788, 35789, 35823, 35792, 35793, 35822, 35789, 35792, 35780, 35767, 35778, 35788, 35775, 35786, 35786, 35769, 35780, 35812, 35799, 35801, 35820, 35807, 35818, 35820, 35801, 35809, 35818, 35777, 35788, 35827, 35828, 35826, 35831, 35832, 35830, 35841, 35838, 35839, 35845, 35842, 35843, 35839, 35834, 35835, 35843, 35836, 35837, 35829, 35834, 35828, 35833, 35836, 35832, 35830, 35829, 35827, 35844, 35839, 35842, 35832, 35835, 35829, 35842, 35835, 35836, 35847, 35848, 35846, 35851, 35852, 35850, 35861, 35858, 35859,
35865, 35862, 35863, 35859, 35854, 35855, 35863, 35856, 35857, 35849, 35854, 35848, 35853, 35856, 35852, 35850, 35849, 35847, 35864, 35859, 35862, 35852, 35855, 35849, 35862, 35855, 35856, 35846, 35833, 35831, 35860, 35843, 35858, 35848, 35837, 35833, 35858, 35837, 35854, 35867, 35868, 35866, 35871, 35872, 35870, 35881, 35878, 35879, 35885, 35882, 35883, 35879, 35874, 35875, 35883, 35876, 35877, 35869, 35874, 35868, 35873, 35876, 35872, 35870, 35869, 35867, 35884, 35879, 35882, 35872, 35875, 35869, 35882, 35875, 35876, 35887, 35888, 35886, 35891, 35892, 35890, 35901, 35898, 35899, 35905, 35902, 35903, 35899, 35894, 35895, 35903, 35896, 35897, 35889, 35894, 35888, 35893, 35896, 35892, 35890, 35889, 35887, 35904, 35899, 35902, 35892, 35895, 35889, 35902, 35895, 35896, 35886, 35873, 35871, 35900, 35883, 35898, 35888, 35877, 35873, 35898, 35877, 35894, 35764, 35720, 35762, 35772, 35728, 35770, 35770, 35722, 35764, 35796, 35751, 35753, 35802, 35761, 35759, 35804, 35753, 35761, 35802, 35730, 35772, 35866, 35853, 35851, 35880, 35863, 35878, 35868, 35857, 35853, 35878, 35857, 35874, 35909, 35906, 35907, 35913, 35910, 35911, 35917, 35914, 35915, 35921, 35918, 35919, 35915, 35908, 35909, 35919, 35912, 35913, 35912, 35907, 35910, 35920, 35915, 35918,
35918, 35909, 35912, 35925, 35922, 35924, 35929, 35926, 35927, 35933, 35930, 35931, 35937, 35934, 35936, 35941, 35938, 35939, 35945, 35942, 35943, 35935, 35924, 35934, 35939, 35928, 35929, 35943, 35932, 35933, 35933, 35923, 35925, 35945, 35935, 35937, 35932, 35927, 35930, 35944, 35939, 35942, 35942, 35929, 35932, 35943, 35925, 35935, 35949, 35946, 35948, 35953, 35950, 35952, 35957, 35954, 35955, 35961, 35958, 35959, 35957, 35948, 35956, 35961, 35952, 35960, 35952, 35947, 35949, 35960, 35955, 35958, 35960, 35949, 35957, 35965, 35962, 35964, 35971, 35968, 35970, 35971, 35966, 35969, 35975, 35972, 35973, 35979, 35976, 35977, 35975, 35964, 35974, 35979, 35970, 35978, 35979, 35967, 35971, 35970, 35963, 35965, 35978, 35973, 35976, 35978, 35965, 35975, 35955, 35916, 35917, 35959, 35920, 35921, 35958, 35917, 35920, 35967, 35936, 35966, 35973, 35940, 35941, 35977, 35944, 35945, 35976, 35941, 35944, 35977, 35937, 35967, 35928, 35911, 35926, 35940, 35919, 35938, 35938, 35913, 35928, 35964, 35951, 35953, 35974, 35959, 35972, 35974, 35953, 35961, 35972, 35921, 35940, 35981, 35982, 35980, 35985, 35986, 35984, 35995, 35992, 35993, 35999, 35996, 35997, 35993, 35988, 35989, 35997, 35990, 35991, 35983, 35988, 35982, 35987, 35990, 35986, 35984, 35983, 35981,
35998, 35993, 35996, 35986, 35989, 35983, 35996, 35989, 35990, 36001, 36002, 36000, 36005, 36006, 36004, 36015, 36012, 36014, 36019, 36016, 36017, 36023, 36020, 36021, 36013, 36005, 36012, 36017, 36008, 36009, 36021, 36010, 36011, 36003, 36008, 36002, 36007, 36010, 36006, 36023, 36013, 36015, 36004, 36003, 36001, 36022, 36017, 36020, 36006, 36009, 36003, 36020, 36009, 36010, 36021, 36007, 36013, 36000, 35987, 35985, 36018, 35997, 36016, 36002, 35991, 35987, 36016, 35991, 36008, 36027, 36024, 36025, 36031, 36028, 36029, 36035, 36032, 36034, 36039, 36036, 36038, 36041, 36026, 36027, 36043, 36030, 36031, 36041, 36034, 36040, 36043, 36038, 36042, 36030, 36025, 36028, 36038, 36033, 36035, 36042, 36035, 36041, 36042, 36027, 36030, 36047, 36044, 36045, 36051, 36048, 36049, 36055, 36052, 36054, 36059, 36056, 36058, 36061, 36046, 36047, 36063, 36050, 36051, 36061, 36054, 36060, 36063, 36058, 36062, 36050, 36045, 36048, 36058, 36053, 36055, 36062, 36055, 36061, 36062, 36047, 36050, 36046, 36029, 36044, 36054, 36037, 36039, 36060, 36039, 36043, 36060, 36031, 36046, 36067, 36064, 36065, 36071, 36068, 36069, 36075, 36072, 36074, 36079, 36076, 36078, 36081, 36066, 36067, 36083, 36070, 36071, 36081, 36074, 36080, 36083, 36078, 36082, 36070, 36065, 36068,
36078, 36073, 36075, 36082, 36075, 36081, 36082, 36067, 36070, 36087, 36084, 36085, 36091, 36088, 36089, 36095, 36092, 36094, 36099, 36096, 36098, 36101, 36086, 36087, 36103, 36090, 36091, 36101, 36094, 36100, 36103, 36098, 36102, 36090, 36085, 36088, 36098, 36093, 36095, 36102, 36095, 36101, 36102, 36087, 36090, 36086, 36069, 36084, 36094, 36077, 36079, 36100, 36079, 36083, 36100, 36071, 36086, 36105, 36106, 36104, 36109, 36110, 36108, 36107, 36112, 36106, 36111, 36114, 36110, 36108, 36107, 36105, 36110, 36113, 36107, 36117, 36118, 36116, 36121, 36122, 36120, 36119, 36124, 36118, 36123, 36126, 36122, 36120, 36119, 36117, 36122, 36125, 36119, 36131, 36128, 36129, 36135, 36132, 36133, 36139, 36136, 36137, 36143, 36140, 36141, 36137, 36130, 36131, 36141, 36134, 36135, 36134, 36129, 36132, 36142, 36137, 36140, 36140, 36131, 36134, 36147, 36144, 36145, 36151, 36148, 36149, 36155, 36152, 36153, 36159, 36156, 36157, 36153, 36146, 36147, 36157, 36150, 36151, 36150, 36145, 36148, 36158, 36153, 36156, 36156, 36147, 36150, 36129, 36112, 36113, 36133, 36114, 36115, 36132, 36113, 36114, 36145, 36124, 36125, 36149, 36126, 36127, 36148, 36125, 36126, 36116, 36111, 36109, 36118, 36115, 36111, 36146, 36133, 36144, 36154, 36141, 36152, 36152, 36135, 36146,
36144, 36115, 36124, 36161, 36162, 36160, 36165, 36166, 36164, 36163, 36168, 36162, 36167, 36170, 36166, 36164, 36163, 36161, 36166, 36169, 36163, 36173, 36174, 36172, 36177, 36178, 36176, 36175, 36180, 36174, 36179, 36182, 36178, 36176, 36175, 36173, 36178, 36181, 36175, 36187, 36184, 36185, 36191, 36188, 36189, 36195, 36192, 36193, 36199, 36196, 36197, 36193, 36186, 36187, 36197, 36190, 36191, 36190, 36185, 36188, 36198, 36193, 36196, 36196, 36187, 36190, 36203, 36200, 36201, 36207, 36204, 36205, 36211, 36208, 36209, 36215, 36212, 36213, 36209, 36202, 36203, 36213, 36206, 36207, 36206, 36201, 36204, 36214, 36209, 36212, 36212, 36203, 36206, 36185, 36168, 36169, 36189, 36170, 36171, 36188, 36169, 36170, 36201, 36180, 36181, 36205, 36182, 36183, 36204, 36181, 36182, 36172, 36167, 36165, 36174, 36171, 36167, 36202, 36189, 36200, 36210, 36197, 36208, 36208, 36191, 36202, 36200, 36171, 36180, 36066, 36049, 36064, 36074, 36057, 36059, 36080, 36059, 36063, 36080, 36051, 36066, 36160, 36123, 36121, 36162, 36127, 36123, 36186, 36149, 36184, 36194, 36157, 36192, 36192, 36151, 36186, 36184, 36127, 36168, 36219, 36216, 36217, 36223, 36220, 36221, 36227, 36224, 36226, 36231, 36228, 36230, 36233, 36218, 36219, 36235, 36222, 36223, 36233, 36226, 36232,
36235, 36230, 36234, 36222, 36217, 36220, 36230, 36225, 36227, 36234, 36227, 36233, 36234, 36219, 36222, 36239, 36236, 36238, 36243, 36240, 36241, 36247, 36244, 36245, 36251, 36248, 36250, 36255, 36252, 36254, 36255, 36238, 36253, 36257, 36242, 36243, 36259, 36246, 36247, 36257, 36250, 36256, 36247, 36237, 36239, 36259, 36254, 36258, 36246, 36241, 36244, 36254, 36249, 36251, 36258, 36251, 36257, 36258, 36243, 36246, 36259, 36239, 36255, 36242, 36221, 36240, 36250, 36229, 36231, 36256, 36231, 36235, 36256, 36223, 36242, 36261, 36262, 36260, 36265, 36266, 36264, 36263, 36268, 36262, 36267, 36270, 36266, 36264, 36263, 36261, 36266, 36269, 36263, 36273, 36274, 36272, 36277, 36278, 36276, 36275, 36280, 36274, 36279, 36282, 36278, 36276, 36275, 36273, 36278, 36281, 36275, 36287, 36284, 36285, 36291, 36288, 36289, 36295, 36292, 36293, 36299, 36296, 36297, 36293, 36286, 36287, 36297, 36290, 36291, 36290, 36285, 36288, 36298, 36293, 36296, 36296, 36287, 36290, 36303, 36300, 36302, 36307, 36304, 36305, 36311, 36308, 36309, 36315, 36312, 36314, 36319, 36316, 36317, 36323, 36320, 36321, 36313, 36302, 36312, 36317, 36306, 36307, 36321, 36310, 36311, 36311, 36301, 36303, 36323, 36313, 36315, 36310, 36305, 36308, 36322, 36317, 36320, 36320, 36307, 36310,
36321, 36303, 36313, 36285, 36268, 36269, 36289, 36270, 36271, 36288, 36269, 36270, 36301, 36277, 36300, 36305, 36280, 36281, 36309, 36282, 36283, 36308, 36281, 36282, 36309, 36279, 36301, 36272, 36267, 36265, 36274, 36271, 36267, 36306, 36289, 36304, 36318, 36297, 36316, 36316, 36291, 36306, 36304, 36271, 36280, 36025, 35840, 35841, 36029, 35844, 35845, 36028, 35841, 35844, 36045, 35860, 35861, 36049, 35864, 35865, 36048, 35861, 35864, 36044, 35845, 35860, 36065, 35880, 35881, 36069, 35884, 35885, 36068, 35881, 35884, 36085, 35900, 35901, 36089, 35904, 35905, 36088, 35901, 35904, 36084, 35885, 35900, 36064, 35865, 35880, 36217, 35994, 35995, 36221, 35998, 35999, 36220, 35995, 35998, 36237, 36014, 36236, 36241, 36018, 36019, 36245, 36022, 36023, 36244, 36019, 36022, 36245, 36015, 36237, 36240, 35999, 36018, 35908, 35783, 35906, 35916, 35791, 35914, 35914, 35785, 35908, 35948, 35815, 35817, 35956, 35823, 35954, 35956, 35817, 35825, 35954, 35793, 35916, 35980, 35893, 35891, 35994, 35903, 35992, 35982, 35897, 35893, 35992, 35897, 35988, 36218, 36089, 36216, 36226, 36097, 36099, 36232, 36099, 36103, 36232, 36091, 36218, 36260, 36179, 36177, 36262, 36183, 36179, 36286, 36205, 36284, 36294, 36213, 36292, 36292, 36207, 36286, 36284, 36183, 36268,
36216, 35905, 35994, 36327, 36324, 36326, 36331, 36328, 36330, 36335, 36332, 36333, 36339, 36336, 36337, 36335, 36326, 36334, 36339, 36330, 36338, 36330, 36325, 36327, 36338, 36333, 36336, 36338, 36327, 36335, 36343, 36340, 36342, 36347, 36344, 36346, 36351, 36348, 36349, 36355, 36352, 36353, 36351, 36342, 36350, 36355, 36346, 36354, 36346, 36341, 36343, 36354, 36349, 36352, 36354, 36343, 36351, 36342, 36329, 36331, 36350, 36337, 36348, 36350, 36331, 36339, 36359, 36356, 36358, 36363, 36360, 36362, 36367, 36364, 36365, 36371, 36368, 36369, 36367, 36358, 36366, 36371, 36362, 36370, 36362, 36357, 36359, 36370, 36365, 36368, 36370, 36359, 36367, 36375, 36372, 36374, 36379, 36376, 36378, 36383, 36380, 36381, 36387, 36384, 36385, 36383, 36374, 36382, 36387, 36378, 36386, 36378, 36373, 36375, 36386, 36381, 36384, 36386, 36375, 36383, 36374, 36361, 36363, 36382, 36369, 36380, 36382, 36363, 36371, 36358, 36345, 36347, 36366, 36353, 36364, 36366, 36347, 36355, 36391, 36388, 36390, 36397, 36394, 36395, 36401, 36398, 36399, 36403, 36404, 36402, 36409, 36406, 36407, 36393, 36402, 36392, 36401, 36406, 36400, 36397, 36390, 36396, 36406, 36405, 36403, 36393, 36389, 36391, 36400, 36395, 36398, 36400, 36391, 36397, 36400, 36403, 36393, 36413, 36410, 36412,
36417, 36414, 36415, 36421, 36418, 36419, 36425, 36422, 36424, 36429, 36426, 36427, 36433, 36430, 36431, 36423, 36412, 36422, 36427, 36416, 36417, 36431, 36420, 36421, 36421, 36411, 36413, 36433, 36423, 36425, 36420, 36415, 36418, 36432, 36427, 36430, 36430, 36417, 36420, 36431, 36413, 36423, 36435, 36436, 36434, 36441, 36438, 36439, 36443, 36444, 36442, 36449, 36446, 36447, 36437, 36442, 36436, 36447, 36440, 36441, 36438, 36437, 36435, 36446, 36445, 36443, 36440, 36443, 36437, 36453, 36450, 36452, 36457, 36454, 36455, 36461, 36458, 36459, 36465, 36462, 36464, 36469, 36466, 36467, 36473, 36470, 36471, 36463, 36452, 36462, 36467, 36456, 36457, 36471, 36460, 36461, 36461, 36451, 36453, 36473, 36463, 36465, 36460, 36455, 36458, 36472, 36467, 36470, 36470, 36457, 36460, 36471, 36453, 36463, 36405, 36434, 36404, 36439, 36408, 36409, 36408, 36435, 36405, 36451, 36424, 36450, 36455, 36428, 36429, 36459, 36432, 36433, 36458, 36429, 36432, 36459, 36425, 36451, 36416, 36399, 36414, 36428, 36407, 36426, 36426, 36401, 36416, 36456, 36439, 36454, 36468, 36447, 36466, 36466, 36441, 36456, 36454, 36409, 36428, 36475, 36476, 36474, 36481, 36478, 36479, 36483, 36484, 36482, 36489, 36486, 36487, 36477, 36482, 36476, 36481, 36486, 36480, 36478, 36477, 36475,
36483, 36488, 36485, 36477, 36486, 36483, 36493, 36490, 36492, 36497, 36494, 36495, 36501, 36498, 36499, 36505, 36502, 36504, 36509, 36506, 36507, 36513, 36510, 36511, 36503, 36492, 36502, 36507, 36496, 36497, 36511, 36500, 36501, 36501, 36491, 36493, 36513, 36503, 36505, 36500, 36495, 36498, 36512, 36507, 36510, 36510, 36497, 36500, 36511, 36493, 36503, 36515, 36516, 36514, 36519, 36520, 36518, 36525, 36526, 36524, 36532, 36529, 36530, 36517, 36524, 36516, 36530, 36520, 36521, 36515, 36520, 36517, 36529, 36527, 36525, 36526, 36523, 36522, 36523, 36531, 36528, 36517, 36529, 36525, 36536, 36533, 36535, 36540, 36537, 36538, 36544, 36541, 36542, 36548, 36545, 36547, 36552, 36549, 36550, 36554, 36555, 36553, 36546, 36535, 36545, 36550, 36539, 36540, 36554, 36543, 36544, 36544, 36534, 36536, 36556, 36546, 36548, 36543, 36538, 36541, 36555, 36550, 36553, 36553, 36540, 36543, 36554, 36536, 36546, 36485, 36514, 36484, 36489, 36518, 36488, 36485, 36518, 36515, 36534, 36504, 36533, 36538, 36508, 36509, 36542, 36512, 36513, 36541, 36509, 36512, 36542, 36505, 36534, 36496, 36479, 36494, 36508, 36487, 36506, 36506, 36481, 36496, 36539, 36519, 36537, 36551, 36530, 36549, 36549, 36521, 36539, 36537, 36489, 36508, 36445, 36474, 36444, 36479, 36448, 36449,
36448, 36475, 36445, 36491, 36464, 36490, 36495, 36468, 36469, 36499, 36472, 36473, 36498, 36469, 36472, 36499, 36465, 36491, 36494, 36449, 36468, 36558, 36559, 36557, 36560, 36561, 36559, 36564, 36565, 36563, 36570, 36571, 36569, 36574, 36575, 36573, 36575, 36577, 36573, 36584, 36581, 36582, 36566, 36574, 36565, 36571, 36580, 36579, 36564, 36568, 36566, 36576, 36581, 36575, 36575, 36583, 36578, 36577, 36572, 36570, 36572, 36583, 36580, 36566, 36582, 36576, 36588, 36589, 36587, 36592, 36588, 36591, 36596, 36593, 36594, 36598, 36599, 36597, 36604, 36601, 36602, 36608, 36605, 36606, 36602, 36595, 36596, 36606, 36599, 36600, 36589, 36598, 36597, 36600, 36590, 36592, 36587, 36586, 36585, 36597, 36596, 36594, 36607, 36602, 36605, 36589, 36594, 36586, 36605, 36596, 36599, 36610, 36611, 36609, 36614, 36615, 36613, 36624, 36621, 36622, 36628, 36625, 36626, 36618, 36621, 36617, 36626, 36619, 36620, 36612, 36617, 36611, 36615, 36620, 36619, 36613, 36612, 36610, 36627, 36622, 36625, 36615, 36618, 36612, 36625, 36618, 36619, 36632, 36629, 36631, 36634, 36635, 36633, 36640, 36637, 36638, 36646, 36643, 36645, 36632, 36641, 36630, 36640, 36648, 36639, 36636, 36631, 36635, 36649, 36645, 36648, 36639, 36634, 36637, 36645, 36641, 36642, 36648, 36642, 36647,
36639, 36647, 36636, 36636, 36642, 36632, 36653, 36650, 36651, 36657, 36654, 36655, 36661, 36658, 36660, 36665, 36662, 36664, 36653, 36666, 36652, 36657, 36668, 36656, 36667, 36660, 36666, 36669, 36664, 36668, 36656, 36651, 36654, 36664, 36659, 36661, 36668, 36661, 36667, 36668, 36653, 36656, 36634, 36603, 36604, 36638, 36607, 36608, 36637, 36604, 36607, 36651, 36623, 36624, 36655, 36627, 36628, 36654, 36624, 36627, 36609, 36592, 36591, 36623, 36606, 36621, 36592, 36617, 36600, 36617, 36606, 36600, 36652, 36638, 36650, 36660, 36644, 36646, 36666, 36646, 36649, 36652, 36649, 36640, 36650, 36608, 36623, 36671, 36672, 36670, 36675, 36676, 36674, 36676, 36678, 36674, 36685, 36682, 36683, 36689, 36686, 36687, 36693, 36690, 36691, 36687, 36680, 36681, 36685, 36690, 36684, 36672, 36681, 36680, 36677, 36682, 36676, 36676, 36684, 36679, 36678, 36673, 36671, 36692, 36687, 36690, 36679, 36681, 36673, 36690, 36681, 36684, 36697, 36698, 36696, 36703, 36700, 36701, 36707, 36704, 36705, 36711, 36708, 36709, 36705, 36698, 36699, 36709, 36702, 36703, 36702, 36697, 36700, 36710, 36705, 36708, 36708, 36699, 36702, 36715, 36712, 36713, 36719, 36716, 36717, 36723, 36720, 36722, 36727, 36724, 36726, 36729, 36714, 36715, 36731, 36718, 36719, 36728, 36723, 36722,
36730, 36727, 36726, 36718, 36713, 36716, 36726, 36721, 36723, 36729, 36726, 36723, 36730, 36715, 36718, 36735, 36732, 36734, 36741, 36738, 36740, 36741, 36736, 36739, 36745, 36742, 36743, 36749, 36746, 36747, 36751, 36732, 36750, 36745, 36734, 36744, 36748, 36741, 36740, 36741, 36747, 36737, 36740, 36733, 36735, 36748, 36743, 36746, 36745, 36740, 36735, 36751, 36744, 36734, 36713, 36688, 36689, 36717, 36692, 36693, 36716, 36689, 36692, 36743, 36706, 36707, 36711, 36746, 36710, 36746, 36707, 36710, 36694, 36677, 36675, 36698, 36683, 36696, 36706, 36691, 36704, 36677, 36696, 36683, 36704, 36685, 36698, 36742, 36719, 36717, 36751, 36725, 36727, 36752, 36727, 36731, 36752, 36719, 36744, 36742, 36693, 36706, 36586, 36558, 36585, 36562, 36593, 36561, 36594, 36560, 36586, 36571, 36694, 36569, 36697, 36579, 36580, 36584, 36700, 36583, 36700, 36580, 36583, 36571, 36696, 36695, 36670, 36616, 36614, 36688, 36626, 36686, 36616, 36680, 36620, 36680, 36626, 36620, 36714, 36655, 36712, 36722, 36663, 36665, 36669, 36722, 36665, 36714, 36669, 36657, 36712, 36628, 36688, 36754, 36755, 36753, 36762, 36759, 36760, 36766, 36767, 36765, 36764, 36766, 36763, 36775, 36772, 36773, 36756, 36763, 36755, 36773, 36761, 36762, 36754, 36758, 36756, 36767, 36771, 36770,
36769, 36768, 36764, 36761, 36757, 36759, 36774, 36769, 36772, 36772, 36758, 36761, 36756, 36769, 36764, 36779, 36776, 36778, 36783, 36780, 36781, 36787, 36784, 36785, 36791, 36788, 36790, 36793, 36788, 36792, 36797, 36794, 36795, 36791, 36778, 36789, 36795, 36782, 36783, 36799, 36786, 36787, 36779, 36785, 36777, 36798, 36791, 36790, 36800, 36790, 36793, 36786, 36781, 36784, 36800, 36795, 36798, 36786, 36795, 36783, 36799, 36779, 36791, 36804, 36805, 36803, 36810, 36807, 36808, 36814, 36811, 36812, 36818, 36815, 36816, 36812, 36805, 36806, 36816, 36809, 36810, 36809, 36804, 36807, 36817, 36812, 36815, 36815, 36806, 36809, 36824, 36821, 36823, 36826, 36827, 36825, 36836, 36833, 36835, 36836, 36831, 36834, 36840, 36837, 36838, 36832, 36819, 36831, 36838, 36827, 36828, 36839, 36836, 36835, 36836, 36838, 36832, 36830, 36822, 36824, 36824, 36819, 36820, 36830, 36826, 36829, 36828, 36824, 36820, 36838, 36820, 36832, 36767, 36801, 36765, 36804, 36770, 36771, 36808, 36774, 36775, 36774, 36804, 36771, 36770, 36802, 36767, 36822, 36792, 36821, 36826, 36796, 36797, 36829, 36797, 36800, 36822, 36800, 36793, 36782, 36760, 36780, 36796, 36773, 36794, 36794, 36762, 36782, 36827, 36808, 36825, 36837, 36818, 36816, 36837, 36810, 36827, 36825, 36775, 36796,
36846, 36843, 36844, 36848, 36849, 36847, 36854, 36851, 36853, 36858, 36855, 36857, 36860, 36855, 36859, 36864, 36861, 36862, 36858, 36841, 36856, 36862, 36845, 36846, 36866, 36849, 36850, 36863, 36854, 36853, 36865, 36858, 36857, 36860, 36865, 36857, 36849, 36844, 36847, 36860, 36852, 36854, 36867, 36862, 36865, 36864, 36860, 36854, 36849, 36862, 36846, 36858, 36850, 36842, 36871, 36868, 36870, 36871, 36872, 36869, 36875, 36851, 36874, 36875, 36863, 36853, 36871, 36841, 36842, 36873, 36848, 36872, 36873, 36842, 36850, 36844, 36813, 36814, 36848, 36817, 36818, 36847, 36814, 36817, 36869, 36833, 36868, 36869, 36839, 36835, 36872, 36818, 36839, 36523, 36753, 36522, 36532, 36759, 36531, 36759, 36528, 36531, 36528, 36754, 36523, 36777, 36547, 36776, 36781, 36551, 36552, 36556, 36784, 36555, 36784, 36552, 36555, 36777, 36556, 36548, 36780, 36532, 36551, 36390, 36377, 36379, 36396, 36385, 36394, 36396, 36379, 36387, 36801, 36564, 36563, 36805, 36567, 36803, 36811, 36584, 36582, 36802, 36567, 36564, 36811, 36568, 36805, 36845, 36701, 36843, 36861, 36711, 36709, 36861, 36703, 36845, 36875, 36736, 36737, 36876, 36737, 36747, 36876, 36711, 36863, 36843, 36584, 36813, 36333, 36138, 36139, 36337, 36142, 36143, 36336, 36139, 36142, 36349, 36154, 36155,
36353, 36158, 36159, 36352, 36155, 36158, 36348, 36143, 36154, 36365, 36194, 36195, 36369, 36198, 36199, 36368, 36195, 36198, 36381, 36210, 36211, 36385, 36214, 36215, 36384, 36211, 36214, 36380, 36199, 36210, 36364, 36159, 36194, 36395, 36294, 36295, 36399, 36298, 36299, 36398, 36295, 36298, 36411, 36314, 36410, 36415, 36318, 36319, 36419, 36322, 36323, 36418, 36319, 36322, 36419, 36315, 36411, 36414, 36299, 36318, 36394, 36215, 36294, 36880, 36881, 36879, 36878, 36880, 36877, 36887, 36888, 36886, 36893, 36890, 36891, 36881, 36886, 36879, 36891, 36884, 36885, 36882, 36884, 36881, 36878, 36885, 36882, 36887, 36892, 36889, 36881, 36890, 36887, 36895, 36896, 36894, 36899, 36900, 36898, 36900, 36902, 36898, 36909, 36906, 36907, 36896, 36905, 36904, 36901, 36906, 36900, 36900, 36908, 36903, 36902, 36897, 36895, 36897, 36908, 36905, 36911, 36912, 36910, 36915, 36916, 36914, 36927, 36928, 36926, 36933, 36930, 36931, 36912, 36922, 36910, 36925, 36915, 36924, 36927, 36918, 36919, 36921, 36930, 36920, 36913, 36918, 36912, 36916, 36921, 36920, 36914, 36913, 36911, 36932, 36927, 36930, 36913, 36920, 36919, 36930, 36919, 36920, 36912, 36926, 36923, 36931, 36917, 36925, 36937, 36938, 36936, 36943, 36940, 36941, 36947, 36944, 36945, 36951, 36948, 36949,
36945, 36938, 36939, 36949, 36942, 36943, 36942, 36937, 36940, 36950, 36945, 36948, 36948, 36939, 36942, 36955, 36952, 36953, 36959, 36956, 36957, 36963, 36960, 36961, 36967, 36964, 36965, 36961, 36954, 36955, 36965, 36958, 36959, 36958, 36953, 36956, 36966, 36961, 36964, 36964, 36955, 36958, 36896, 36934, 36894, 36937, 36904, 36905, 36909, 36940, 36908, 36940, 36905, 36908, 36896, 36936, 36935, 36953, 36928, 36929, 36957, 36932, 36933, 36956, 36929, 36932, 36922, 36901, 36899, 36928, 36907, 36926, 36901, 36926, 36907, 36954, 36941, 36952, 36962, 36949, 36960, 36960, 36943, 36954, 36952, 36909, 36928, 36969, 36970, 36968, 36977, 36978, 36976, 36981, 36982, 36980, 36989, 36986, 36987, 36971, 36980, 36970, 36987, 36978, 36979, 36974, 36971, 36969, 36981, 36985, 36983, 36972, 36969, 36968, 36978, 36974, 36976, 36988, 36984, 36986, 36969, 36976, 36974, 36986, 36975, 36978, 36971, 36984, 36981, 36991, 36992, 36990, 36999, 36996, 36997, 37001, 37002, 37000, 37009, 37006, 37007, 36993, 37000, 36992, 37007, 36998, 36999, 36991, 36995, 36993, 37004, 37003, 37001, 36996, 36995, 36994, 37008, 37004, 37006, 37006, 36995, 36998, 36995, 37001, 36993, 36983, 36990, 36982, 36997, 36988, 36989, 36996, 36985, 36988, 36983, 36994, 36991, 37015, 37012, 37013,
37017, 37018, 37016, 37023, 37020, 37022, 37027, 37024, 37025, 37031, 37028, 37029, 37021, 37010, 37020, 37025, 37014, 37015, 37029, 37018, 37019, 37023, 37029, 37021, 37018, 37013, 37016, 37030, 37025, 37028, 37028, 37015, 37018, 37021, 37019, 37011, 37035, 37032, 37034, 37037, 37032, 37036, 37041, 37038, 37040, 37042, 37035, 37034, 37037, 37042, 37034, 37046, 37040, 37045, 37040, 37033, 37035, 37043, 37040, 37035, 37050, 37047, 37049, 37052, 37047, 37051, 37056, 37053, 37054, 37063, 37060, 37062, 37067, 37064, 37065, 37061, 37051, 37060, 37065, 37055, 37056, 37057, 37050, 37049, 37059, 37049, 37052, 37063, 37068, 37061, 37057, 37056, 37054, 37069, 37065, 37068, 37059, 37065, 37056, 37061, 37059, 37052, 37050, 37022, 37048, 37054, 37026, 37027, 37058, 37030, 37031, 37030, 37054, 37027, 37058, 37023, 37050, 37037, 37010, 37011, 37044, 37017, 37042, 37019, 37037, 37011, 36889, 36972, 36888, 36977, 36892, 36893, 36892, 36973, 36889, 37013, 36946, 36947, 37017, 36950, 36951, 37016, 36947, 36950, 37043, 36962, 36963, 36967, 37045, 36966, 37045, 36963, 36966, 36962, 37017, 36951, 36934, 36878, 36877, 36938, 36883, 36936, 36946, 36891, 36944, 36878, 36936, 36883, 36938, 36891, 36885, 37014, 36977, 37012, 37026, 36987, 37024, 37024, 36979, 37014,
37055, 36997, 37053, 37066, 37007, 37064, 37064, 36999, 37055, 37053, 36989, 37026, 37012, 36893, 36946, 37071, 37072, 37070, 37077, 37078, 37076, 37078, 37074, 37076, 37081, 37082, 37080, 37073, 37080, 37072, 37079, 37085, 37078, 37085, 37075, 37078, 37074, 37073, 37071, 37084, 37083, 37081, 37073, 37084, 37081, 37093, 37090, 37091, 37097, 37094, 37095, 37101, 37098, 37099, 37105, 37102, 37103, 37099, 37092, 37093, 37103, 37096, 37097, 37096, 37091, 37094, 37104, 37099, 37102, 37102, 37093, 37096, 37107, 37108, 37106, 37111, 37107, 37110, 37113, 37114, 37112, 37121, 37118, 37120, 37125, 37122, 37123, 37117, 37110, 37116, 37115, 37122, 37114, 37108, 37113, 37112, 37115, 37109, 37111, 37127, 37119, 37121, 37118, 37117, 37116, 37126, 37125, 37123, 37119, 37123, 37117, 37123, 37111, 37117, 37091, 37082, 37083, 37086, 37094, 37085, 37094, 37083, 37085, 37108, 37087, 37106, 37112, 37088, 37108, 37087, 37079, 37077, 37088, 37086, 37079, 37114, 37095, 37112, 37124, 37103, 37122, 37122, 37097, 37114, 37089, 37095, 37086, 37133, 37130, 37132, 37129, 37130, 37128, 37135, 37136, 37134, 37141, 37138, 37139, 37143, 37144, 37142, 37153, 37150, 37151, 37133, 37142, 37131, 37151, 37140, 37141, 37137, 37132, 37136, 37134, 37132, 37129, 37148, 37145, 37143,
37140, 37135, 37138, 37147, 37144, 37145, 37150, 37149, 37148, 37152, 37145, 37149, 37140, 37148, 37137, 37137, 37143, 37133, 37155, 37156, 37154, 37159, 37155, 37158, 37163, 37160, 37161, 37165, 37166, 37164, 37171, 37168, 37170, 37175, 37172, 37173, 37179, 37176, 37177, 37169, 37158, 37168, 37173, 37162, 37163, 37167, 37176, 37166, 37157, 37164, 37156, 37159, 37165, 37157, 37179, 37169, 37171, 37166, 37161, 37164, 37178, 37173, 37176, 37176, 37163, 37166, 37177, 37159, 37169, 37181, 37182, 37180, 37187, 37184, 37185, 37189, 37190, 37188, 37193, 37194, 37192, 37183, 37188, 37182, 37193, 37186, 37187, 37184, 37183, 37181, 37192, 37191, 37189, 37186, 37189, 37183, 37199, 37196, 37198, 37203, 37200, 37201, 37207, 37204, 37205, 37211, 37208, 37210, 37215, 37212, 37213, 37219, 37216, 37217, 37209, 37198, 37208, 37213, 37202, 37203, 37217, 37206, 37207, 37207, 37197, 37199, 37219, 37209, 37211, 37206, 37201, 37204, 37218, 37213, 37216, 37216, 37203, 37206, 37217, 37199, 37209, 37147, 37180, 37146, 37153, 37184, 37152, 37147, 37184, 37181, 37197, 37170, 37196, 37201, 37174, 37175, 37205, 37178, 37179, 37204, 37175, 37178, 37205, 37171, 37197, 37162, 37139, 37160, 37174, 37151, 37172, 37172, 37141, 37162, 37202, 37185, 37200, 37214, 37193, 37212,
37212, 37187, 37202, 37200, 37153, 37174, 37135, 37100, 37101, 37139, 37104, 37105, 37138, 37101, 37104, 37156, 37120, 37154, 37161, 37124, 37125, 37127, 37161, 37125, 37156, 37127, 37121, 37160, 37105, 37124, 37221, 37222, 37220, 37229, 37226, 37227, 37231, 37232, 37230, 37239, 37236, 37237, 37223, 37230, 37222, 37237, 37228, 37229, 37224, 37223, 37221, 37234, 37233, 37231, 37228, 37224, 37226, 37238, 37234, 37236, 37236, 37225, 37228, 37225, 37231, 37223, 37241, 37242, 37240, 37249, 37246, 37247, 37251, 37252, 37250, 37259, 37256, 37257, 37243, 37250, 37242, 37257, 37248, 37249, 37244, 37243, 37241, 37251, 37255, 37253, 37248, 37244, 37246, 37256, 37255, 37254, 37256, 37245, 37248, 37245, 37251, 37243, 37233, 37240, 37232, 37247, 37238, 37239, 37246, 37235, 37238, 37235, 37241, 37233, 37263, 37260, 37262, 37267, 37264, 37265, 37273, 37270, 37272, 37277, 37274, 37275, 37271, 37262, 37270, 37275, 37266, 37267, 37269, 37261, 37263, 37279, 37271, 37273, 37269, 37265, 37268, 37279, 37275, 37278, 37278, 37267, 37269, 37278, 37263, 37271, 37283, 37280, 37282, 37287, 37284, 37285, 37293, 37290, 37292, 37297, 37294, 37295, 37291, 37282, 37290, 37295, 37286, 37287, 37289, 37281, 37283, 37299, 37291, 37293, 37289, 37285, 37288, 37299, 37295, 37298,
37298, 37287, 37289, 37298, 37283, 37291, 37281, 37272, 37280, 37285, 37276, 37277, 37288, 37277, 37279, 37288, 37273, 37281, 37301, 37302, 37300, 37309, 37306, 37307, 37311, 37312, 37310, 37319, 37316, 37317, 37303, 37310, 37302, 37317, 37308, 37309, 37301, 37305, 37303, 37311, 37315, 37313, 37308, 37304, 37306, 37318, 37314, 37316, 37316, 37305, 37308, 37303, 37314, 37311, 37326, 37327, 37325, 37324, 37326, 37323, 37335, 37332, 37333, 37337, 37338, 37336, 37343, 37340, 37341, 37347, 37344, 37345, 37327, 37336, 37325, 37341, 37330, 37331, 37345, 37334, 37335, 37327, 37331, 37330, 37324, 37331, 37328, 37337, 37342, 37339, 37334, 37329, 37332, 37346, 37341, 37344, 37334, 37341, 37331, 37327, 37340, 37337, 37313, 37323, 37312, 37333, 37318, 37319, 37332, 37315, 37318, 37313, 37329, 37324, 37338, 37321, 37320, 37321, 37342, 37322, 37351, 37348, 37350, 37355, 37352, 37353, 37361, 37358, 37360, 37365, 37362, 37363, 37359, 37350, 37358, 37363, 37354, 37355, 37357, 37349, 37351, 37361, 37366, 37359, 37357, 37353, 37356, 37366, 37365, 37363, 37357, 37363, 37355, 37359, 37357, 37351, 37371, 37368, 37370, 37375, 37372, 37373, 37381, 37378, 37380, 37385, 37382, 37383, 37379, 37370, 37378, 37383, 37374, 37375, 37371, 37376, 37369, 37381, 37386, 37379,
37377, 37373, 37376, 37387, 37383, 37386, 37386, 37375, 37377, 37379, 37377, 37371, 37369, 37360, 37368, 37373, 37364, 37365, 37376, 37365, 37367, 37369, 37367, 37361, 37253, 37300, 37252, 37307, 37258, 37259, 37306, 37255, 37258, 37253, 37304, 37301, 37349, 37292, 37348, 37353, 37296, 37297, 37356, 37297, 37299, 37356, 37293, 37349, 37266, 37227, 37264, 37276, 37237, 37274, 37274, 37229, 37266, 37286, 37247, 37284, 37296, 37257, 37294, 37294, 37249, 37286, 37284, 37239, 37276, 37354, 37307, 37352, 37364, 37317, 37362, 37362, 37309, 37354, 37374, 37333, 37372, 37384, 37345, 37382, 37382, 37335, 37374, 37372, 37319, 37364, 37352, 37259, 37296, 37389, 37390, 37388, 37395, 37392, 37393, 37397, 37398, 37396, 37403, 37400, 37401, 37391, 37396, 37390, 37401, 37394, 37395, 37392, 37391, 37389, 37400, 37399, 37397, 37394, 37397, 37391, 37407, 37404, 37406, 37411, 37408, 37409, 37415, 37412, 37413, 37419, 37416, 37418, 37423, 37420, 37421, 37427, 37424, 37425, 37417, 37406, 37416, 37421, 37410, 37411, 37425, 37414, 37415, 37415, 37405, 37407, 37427, 37417, 37419, 37414, 37409, 37412, 37426, 37421, 37424, 37424, 37411, 37414, 37425, 37407, 37417, 37429, 37430, 37428, 37435, 37432, 37433, 37437, 37438, 37436, 37443, 37440, 37441, 37431, 37436, 37430,
37441, 37434, 37435, 37432, 37431, 37429, 37440, 37439, 37437, 37434, 37437, 37431, 37447, 37444, 37446, 37451, 37448, 37449, 37455, 37452, 37453, 37459, 37456, 37458, 37463, 37460, 37461, 37467, 37464, 37465, 37457, 37446, 37456, 37461, 37450, 37451, 37465, 37454, 37455, 37455, 37445, 37447, 37467, 37457, 37459, 37454, 37449, 37452, 37466, 37461, 37464, 37464, 37451, 37454, 37465, 37447, 37457, 37399, 37428, 37398, 37433, 37402, 37403, 37402, 37429, 37399, 37445, 37418, 37444, 37449, 37422, 37423, 37453, 37426, 37427, 37452, 37423, 37426, 37453, 37419, 37445, 37410, 37393, 37408, 37422, 37401, 37420, 37420, 37395, 37410, 37450, 37433, 37448, 37462, 37441, 37460, 37460, 37435, 37450, 37448, 37403, 37422, 37469, 37470, 37468, 37475, 37472, 37473, 37477, 37478, 37476, 37483, 37480, 37481, 37471, 37476, 37470, 37481, 37474, 37475, 37472, 37471, 37469, 37480, 37479, 37477, 37474, 37477, 37471, 37487, 37484, 37486, 37491, 37488, 37489, 37495, 37492, 37493, 37499, 37496, 37498, 37503, 37500, 37501, 37507, 37504, 37505, 37497, 37486, 37496, 37501, 37490, 37491, 37505, 37494, 37495, 37495, 37485, 37487, 37507, 37497, 37499, 37494, 37489, 37492, 37506, 37501, 37504, 37504, 37491, 37494, 37505, 37487, 37497, 37509, 37510, 37508, 37515, 37512, 37513,
37517, 37518, 37516, 37523, 37520, 37521, 37511, 37516, 37510, 37521, 37514, 37515, 37512, 37511, 37509, 37520, 37519, 37517, 37514, 37517, 37511, 37527, 37524, 37526, 37531, 37528, 37529, 37535, 37532, 37533, 37539, 37536, 37538, 37543, 37540, 37541, 37547, 37544, 37545, 37537, 37526, 37536, 37541, 37530, 37531, 37545, 37534, 37535, 37535, 37525, 37527, 37547, 37537, 37539, 37534, 37529, 37532, 37546, 37541, 37544, 37544, 37531, 37534, 37545, 37527, 37537, 37479, 37508, 37478, 37513, 37482, 37483, 37482, 37509, 37479, 37525, 37498, 37524, 37529, 37502, 37503, 37533, 37506, 37507, 37532, 37503, 37506, 37533, 37499, 37525, 37490, 37473, 37488, 37502, 37481, 37500, 37500, 37475, 37490, 37530, 37513, 37528, 37542, 37521, 37540, 37540, 37515, 37530, 37528, 37483, 37502, 37439, 37468, 37438, 37473, 37442, 37443, 37442, 37469, 37439, 37485, 37458, 37484, 37489, 37462, 37463, 37493, 37466, 37467, 37492, 37463, 37466, 37493, 37459, 37485, 37488, 37443, 37462, 37003, 37220, 37002, 37227, 37008, 37009, 37226, 37005, 37008, 37005, 37221, 37003, 37261, 37062, 37260, 37265, 37066, 37067, 37268, 37067, 37069, 37268, 37063, 37261, 37264, 37009, 37066, 37191, 37388, 37190, 37393, 37194, 37195, 37194, 37389, 37191, 37405, 37210, 37404, 37409, 37214, 37215,
37413, 37218, 37219, 37412, 37215, 37218, 37413, 37211, 37405, 37408, 37195, 37214, 37070, 36925, 36924, 37082, 36931, 37080, 36925, 37080, 36931, 37092, 36957, 37090, 37100, 36965, 37098, 37098, 36959, 37092, 37090, 36933, 37082, 37129, 37039, 37041, 37134, 37041, 37046, 37100, 37046, 36967, 37549, 37550, 37548, 37553, 37554, 37552, 37554, 37556, 37552, 37563, 37560, 37561, 37567, 37564, 37565, 37571, 37568, 37569, 37565, 37558, 37559, 37563, 37568, 37562, 37551, 37558, 37550, 37555, 37560, 37554, 37554, 37562, 37557, 37556, 37551, 37549, 37570, 37565, 37568, 37557, 37559, 37551, 37568, 37559, 37562, 37575, 37572, 37573, 37579, 37576, 37577, 37585, 37582, 37583, 37587, 37588, 37586, 37583, 37574, 37575, 37587, 37578, 37579, 37578, 37573, 37576, 37588, 37583, 37586, 37586, 37575, 37578, 37593, 37590, 37591, 37597, 37594, 37595, 37601, 37598, 37600, 37605, 37602, 37604, 37607, 37592, 37593, 37609, 37596, 37597, 37606, 37601, 37600, 37608, 37605, 37604, 37596, 37591, 37594, 37604, 37599, 37601, 37607, 37604, 37601, 37608, 37593, 37596, 37613, 37610, 37612, 37617, 37614, 37616, 37619, 37614, 37618, 37623, 37620, 37621, 37628, 37610, 37627, 37622, 37613, 37612, 37625, 37616, 37624, 37619, 37624, 37616, 37619, 37611, 37613, 37626, 37621, 37624,
37623, 37619, 37613, 37628, 37622, 37612, 37591, 37566, 37567, 37595, 37570, 37571, 37594, 37567, 37570, 37617, 37580, 37615, 37621, 37584, 37585, 37625, 37588, 37589, 37588, 37621, 37585, 37617, 37589, 37581, 37572, 37563, 37561, 37584, 37569, 37582, 37582, 37563, 37574, 37620, 37597, 37595, 37628, 37603, 37605, 37629, 37605, 37609, 37629, 37597, 37622, 37620, 37571, 37584, 37633, 37630, 37632, 37639, 37636, 37637, 37644, 37641, 37643, 37646, 37641, 37645, 37644, 37632, 37642, 37648, 37638, 37639, 37633, 37637, 37631, 37647, 37644, 37643, 37649, 37643, 37646, 37635, 37630, 37631, 37637, 37635, 37631, 37644, 37639, 37633, 37638, 37577, 37636, 37646, 37580, 37581, 37649, 37587, 37647, 37589, 37646, 37581, 37638, 37587, 37579, 37651, 37652, 37650, 37657, 37654, 37655, 37659, 37660, 37658, 37665, 37662, 37663, 37653, 37658, 37652, 37663, 37656, 37657, 37654, 37653, 37651, 37662, 37661, 37659, 37656, 37659, 37653, 37669, 37666, 37668, 37673, 37670, 37671, 37677, 37674, 37675, 37681, 37678, 37680, 37685, 37682, 37683, 37689, 37686, 37687, 37679, 37668, 37678, 37683, 37672, 37673, 37687, 37676, 37677, 37677, 37667, 37669, 37689, 37679, 37681, 37676, 37671, 37674, 37688, 37683, 37686, 37686, 37673, 37676, 37687, 37669, 37679, 37693, 37690, 37692,
37695, 37692, 37694, 37699, 37696, 37698, 37703, 37700, 37702, 37705, 37698, 37704, 37707, 37702, 37706, 37702, 37697, 37699, 37706, 37699, 37705, 37661, 37690, 37660, 37695, 37664, 37665, 37664, 37692, 37661, 37703, 37680, 37701, 37705, 37684, 37685, 37707, 37688, 37689, 37706, 37685, 37688, 37707, 37681, 37703, 37672, 37655, 37670, 37684, 37663, 37682, 37682, 37657, 37672, 37698, 37691, 37693, 37704, 37693, 37695, 37704, 37665, 37684, 37321, 37553, 37320, 37322, 37555, 37321, 37343, 37572, 37342, 37577, 37346, 37347, 37576, 37343, 37346, 37572, 37322, 37342, 37635, 37380, 37634, 37385, 37636, 37384, 37640, 37385, 37387, 37640, 37381, 37635, 37636, 37347, 37384, 37519, 37650, 37518, 37655, 37522, 37523, 37522, 37651, 37519, 37667, 37538, 37666, 37671, 37542, 37543, 37675, 37546, 37547, 37674, 37543, 37546, 37675, 37539, 37667, 37670, 37523, 37542, 35700, 35635, 35636, 35638, 35703, 35637, 35637, 35700, 35636, 35648, 35715, 35647, 35720, 35649, 35650, 35649, 35716, 35648, 35647, 35704, 35638, 35660, 35762, 35659, 35767, 35661, 35662, 35766, 35660, 35661, 35779, 35671, 35672, 35783, 35673, 35674, 35782, 35672, 35673, 35778, 35662, 35671, 35762, 35650, 35659, 35907, 35683, 35684, 35911, 35685, 35686, 35910, 35684, 35685, 35923, 35692, 35922,
35927, 35695, 35696, 35931, 35697, 35698, 35930, 35696, 35697, 35931, 35694, 35923, 35926, 35686, 35695, 35906, 35674, 35683, 10612, 26010, 10613, 26013, 10617, 10616, 10621, 26017, 10620, 26010, 10616, 10613, 10612, 26017, 26009, 10632, 26034, 10633, 26038, 10636, 26037, 26034, 10636, 10633, 26014, 10632, 10617, 10768, 26236, 10769, 10773, 26241, 10772, 10772, 26235, 10768, 26255, 10789, 10788, 26259, 10793, 10792, 26264, 10796, 26263, 26256, 10792, 10789, 26263, 10793, 26260, 26236, 10788, 10769, 21411, 34065, 21410, 34070, 21414, 21415, 34074, 21418, 21419, 34073, 21415, 21418, 21414, 34066, 21411, 34090, 21430, 34089, 34094, 21434, 21435, 34094, 21431, 34090, 34093, 21419, 21434, 21567, 34292, 21566, 34298, 21570, 21571, 21567, 34297, 34293, 34312, 21586, 34311, 34316, 21590, 21591, 21595, 34319, 21594, 34319, 21591, 21594, 34312, 21595, 21587, 34315, 21571, 21590, 172, 10960, 171, 10958, 172, 168, 159, 10951, 158, 10949, 159, 155, 174, 10958, 168, 10951, 174, 158, 185, 10976, 184, 10976, 187, 184, 10973, 185, 177, 11094, 301, 299, 309, 11100, 308, 11100, 311, 308, 11096, 309, 301, 281, 11069, 280, 11067, 281, 277, 11082, 295, 293, 283, 11067, 277, 11069, 293, 280,
295, 11064, 268, 11064, 299, 268, 209, 10997, 208, 10995, 209, 205, 196, 10988, 195, 10986, 196, 192, 211, 10995, 205, 10988, 211, 195, 11028, 241, 239, 221, 10986, 192, 10997, 239, 208, 11180, 390, 11179, 11183, 390, 393, 11180, 380, 391, 11159, 367, 11158, 11159, 364, 368, 11145, 356, 11144, 11145, 353, 357, 11162, 364, 11156, 11162, 356, 370, 11194, 397, 399, 11194, 380, 11169, 11192, 367, 397, 241, 10973, 177, 10980, 283, 187, 10960, 221, 171, 11, 10803, 10, 10801, 11, 7, 21, 10810, 20, 25, 10814, 24, 10814, 27, 24, 10831, 42, 10830, 10831, 39, 43, 27, 10811, 21, 10810, 37, 20, 37, 10801, 7, 10798, 25, 2, 54, 10844, 53, 10842, 54, 50, 10850, 56, 58, 56, 10842, 50, 10875, 84, 10874, 10875, 81, 85, 103, 10892, 102, 10892, 105, 102, 10878, 81, 10872, 105, 10890, 99, 10888, 84, 97, 10880, 103, 89, 69, 10859, 68, 10859, 72, 68, 72, 10857, 65, 116, 10904, 115, 10904, 118, 115, 10925, 131, 133, 10911, 121, 10910, 10914, 121, 124, 10927, 131, 10923, 10927, 124, 135, 118, 10880, 89,
10878, 58, 87, 10857, 116, 65, 10911, 73, 122, 141, 10934, 140, 10941, 147, 10940, 10944, 147, 150, 10934, 147, 140, 10941, 141, 148, 147, 10860, 69, 10944, 73, 10864, 53, 10798, 2, 10888, 39, 10828, 11211, 418, 416, 410, 11201, 409, 11201, 412, 409, 11198, 410, 402, 431, 11219, 430, 11217, 431, 427, 433, 11217, 427, 11219, 443, 430, 443, 11211, 416, 11205, 433, 412, 567, 11355, 566, 11353, 567, 563, 552, 11343, 551, 11343, 554, 551, 569, 11353, 563, 11347, 569, 554, 11355, 595, 566, 455, 11242, 454, 11242, 457, 454, 459, 11243, 455, 11295, 504, 506, 11308, 519, 11307, 11308, 516, 520, 11305, 506, 516, 11279, 487, 11278, 11282, 487, 490, 11270, 477, 479, 11284, 494, 492, 11270, 475, 11266, 11279, 475, 488, 531, 11321, 530, 11321, 533, 530, 11334, 544, 542, 544, 11332, 540, 11325, 542, 533, 11319, 504, 11293, 11319, 490, 527, 11286, 531, 494, 11430, 636, 638, 11420, 627, 629, 11432, 636, 11428, 11432, 629, 640, 11461, 666, 668, 11469, 674, 676, 11467, 668, 674, 11439, 650, 648, 11449, 660, 658, 11441, 658, 650,
11475, 686, 684, 11485, 696, 694, 11477, 694, 686, 11459, 638, 666, 11451, 684, 660, 457, 11198, 402, 11213, 552, 418, 11418, 519, 627, 11332, 648, 540, 11875, 1085, 11874, 11875, 1088, 1086, 11887, 1098, 1096, 11878, 1096, 1088, 11915, 1125, 1123, 11927, 1134, 11926, 11924, 1134, 1131, 11915, 1121, 11913, 11917, 1131, 1125, 11889, 1121, 1098, 11797, 1002, 1004, 11805, 1011, 1013, 1015, 11805, 1013, 11803, 1004, 1011, 11836, 1043, 1045, 11846, 1053, 1055, 11844, 1045, 1053, 11814, 1025, 1023, 11824, 1035, 1033, 11816, 1033, 1025, 11854, 1065, 1063, 11866, 1075, 1073, 11858, 1065, 11856, 11858, 1073, 1067, 11834, 1015, 1043, 11826, 1063, 1035, 1163, 11950, 1162, 11954, 1163, 1165, 11964, 1165, 1175, 11986, 1191, 1193, 11977, 1184, 11976, 11974, 1184, 1181, 11988, 1197, 1195, 11977, 1195, 1185, 1222, 12014, 1221, 12014, 1218, 1221, 12023, 1229, 1231, 1233, 12023, 1231, 12021, 1222, 1229, 1191, 11964, 1175, 12012, 1193, 1218, 11950, 1055, 1162, 11868, 1181, 1075, 708, 11495, 707, 11495, 710, 707, 11499, 720, 710, 11538, 747, 749, 11550, 759, 761, 11548, 749, 759, 11522, 728, 11521, 11525, 728, 731,
11532, 735, 737, 11530, 731, 735, 11532, 722, 11511, 11536, 722, 747, 11649, 858, 860, 11661, 870, 872, 11659, 860, 870, 11691, 894, 896, 11697, 900, 902, 11695, 896, 900, 894, 11671, 882, 11671, 872, 882, 11573, 784, 782, 772, 11562, 771, 11562, 774, 771, 774, 11560, 768, 11560, 782, 768, 820, 11608, 819, 11606, 820, 816, 11619, 832, 830, 11608, 830, 819, 795, 11585, 794, 11585, 791, 794, 799, 11589, 798, 798, 11586, 795, 791, 11563, 772, 11575, 816, 784, 11702, 909, 11701, 11706, 913, 11705, 11705, 910, 11702, 11716, 927, 11715, 11713, 927, 924, 11720, 931, 11719, 11719, 928, 11716, 11713, 914, 11706, 11647, 761, 858, 11621, 909, 832, 1257, 12048, 1256, 12052, 1257, 1259, 12060, 1266, 1268, 1270, 12060, 1268, 12058, 1259, 1266, 1299, 12087, 1298, 12087, 1295, 1298, 12100, 1309, 1311, 12098, 1299, 1309, 12085, 1270, 1295, 12149, 1360, 12148, 12146, 1360, 1357, 12160, 1371, 1369, 12164, 1371, 12162, 12149, 1369, 1361, 12118, 1357, 1329, 12190, 1393, 1395, 1403, 12194, 1402, 12198, 1403, 1405, 12194, 1395, 1402, 12214, 1421, 1423, 1435, 12222, 1434, 12226, 1435, 1437,
1421, 12210, 1419, 12222, 1423, 1434, 12204, 1411, 1409, 12236, 1443, 1441, 12242, 1449, 1447, 12238, 1447, 1443, 12210, 1405, 1419, 12206, 1441, 1411, 1393, 12128, 1339, 12128, 1311, 1339, 720, 11284, 492, 12118, 1197, 11990, 11795, 676, 1002, 11487, 1023, 696, 12048, 902, 1256, 595, 10949, 155, 11874, 353, 11142, 10803, 459, 10, 11268, 42, 477, 10890, 708, 99, 11522, 133, 729, 12251, 1458, 1456, 12257, 1464, 1462, 12253, 1462, 1458, 12263, 1470, 1468, 12269, 1476, 1474, 12265, 1474, 1470, 12259, 1468, 1464, 12248, 1456, 1452, 12464, 1671, 1669, 12470, 1677, 1675, 12466, 1675, 1671, 12476, 1683, 1681, 12482, 1689, 1687, 12478, 1687, 1683, 12472, 1681, 1677, 12488, 1695, 1693, 12484, 1693, 1689, 12275, 1486, 1484, 1501, 12289, 1500, 12287, 1501, 1497, 1488, 12275, 1484, 12277, 1497, 1486, 12316, 1529, 1527, 12328, 1541, 1539, 12318, 1539, 1529, 12352, 1561, 1563, 12350, 1551, 1561, 12289, 1527, 1500, 12405, 1615, 12404, 12408, 1615, 1618, 12419, 1626, 1628, 12417, 1618, 1626, 12455, 1661, 1663, 12445, 1652, 1654, 12457, 1661, 12453, 12457, 1654, 1665, 12443, 1628, 1652, 12364, 1570, 1572, 12374, 1580, 12373,
12377, 1580, 1583, 12366, 1570, 12362, 12374, 1572, 1581, 12384, 1587, 1589, 12382, 1583, 1587, 12384, 1551, 12340, 12405, 1563, 1616, 12512, 1717, 1719, 12520, 1725, 1727, 12518, 1719, 1725, 12544, 1749, 1751, 12554, 1761, 1763, 1761, 12550, 1759, 12550, 1751, 1759, 12542, 1727, 1749, 12572, 1785, 1783, 12562, 1783, 1773, 12607, 1818, 12606, 12604, 1818, 1815, 12618, 1829, 1827, 12607, 1827, 1819, 12586, 1793, 1795, 12596, 1803, 1805, 12594, 1795, 1803, 1841, 12628, 1840, 12632, 1841, 1843, 12644, 1853, 1855, 12642, 1843, 1853, 12574, 1815, 1785, 12628, 1805, 1840, 1858, 12644, 1855, 12584, 1763, 1793, 1541, 12248, 1452, 12562, 1695, 12490, 12271, 1669, 1476, 12510, 1663, 1717, 13502, 2712, 2710, 13514, 2721, 13513, 13511, 2721, 2718, 13502, 2708, 13500, 13504, 2718, 2712, 13537, 2744, 2742, 13556, 2767, 13555, 13553, 2767, 2764, 13543, 2764, 2754, 13543, 2744, 13539, 13514, 2742, 2722, 13612, 2823, 2821, 13625, 2831, 13624, 13625, 2834, 2832, 13616, 2823, 13614, 13616, 2831, 2825, 13569, 2772, 2774, 2782, 13573, 2781, 13577, 2782, 2784, 13573, 2774, 2781, 13587, 2794, 2796, 2808, 13595, 2807, 13599, 2808, 2810, 2794, 13583, 2792,
13595, 2796, 2807, 2813, 13599, 2810, 13583, 2784, 2792, 13675, 2886, 13674, 13672, 2886, 2883, 13691, 2899, 13690, 13688, 2899, 2896, 13688, 2894, 13686, 13675, 2894, 2887, 2861, 13653, 2860, 13653, 2857, 2860, 13664, 2871, 2873, 2871, 13660, 2869, 13660, 2861, 2869, 2912, 13699, 2911, 13703, 2912, 2914, 13641, 2883, 2852, 13699, 2873, 2911, 13641, 2834, 13628, 13556, 2821, 2768, 13651, 2813, 2857, 13727, 2938, 13726, 13724, 2938, 2935, 13743, 2951, 13742, 13740, 2951, 2948, 13740, 2946, 13738, 13727, 2946, 2939, 13724, 2925, 13721, 13743, 2955, 2952, 13691, 2925, 2900, 12657, 1864, 1862, 12676, 1887, 12675, 12673, 1887, 1884, 12663, 1884, 1874, 12687, 1898, 1896, 12702, 1909, 12701, 12699, 1909, 1906, 12691, 1898, 12689, 12691, 1906, 1900, 12663, 1864, 12659, 12676, 1896, 1888, 12782, 1989, 1987, 1939, 12727, 1938, 12727, 1935, 1938, 12740, 1949, 1951, 12738, 1939, 1949, 12799, 2009, 12798, 12799, 2012, 2010, 12788, 2009, 1999, 12832, 2040, 12831, 12829, 2040, 2037, 12840, 2047, 2045, 12829, 2035, 12827, 12832, 2045, 2041, 12854, 2047, 12842, 12802, 2035, 2012, 12788, 1989, 12784, 12702, 1987, 1910, 13190, 2400, 13189, 13190, 2403, 2401,
13205, 2411, 13204, 13205, 2414, 2412, 13195, 2403, 13193, 13195, 2411, 2405, 13208, 2417, 2414, 12867, 2074, 12866, 12864, 2074, 2071, 12876, 2084, 2082, 12867, 2082, 2075, 2111, 12903, 2110, 12903, 2107, 2110, 12914, 2121, 2123, 2121, 12910, 2119, 12910, 2111, 2119, 12923, 2133, 12922, 12923, 2136, 2134, 12940, 2148, 12939, 12937, 2148, 2145, 12937, 2143, 12935, 12926, 2143, 2136, 12891, 2084, 12878, 12901, 2080, 2107, 12891, 2133, 2102, 2242, 13029, 2241, 13033, 2242, 2244, 2270, 13062, 2269, 13062, 2266, 2269, 13053, 2261, 2259, 13060, 2257, 2266, 13087, 2293, 2295, 2309, 13096, 2308, 13100, 2309, 2311, 2297, 13087, 2295, 13096, 2297, 2308, 2257, 13033, 2244, 13085, 2270, 2293, 12968, 2172, 2174, 2185, 12975, 2184, 12979, 2185, 2187, 2189, 12979, 2187, 12975, 2174, 2184, 2218, 13006, 2217, 13006, 2214, 2217, 13019, 2228, 2230, 13017, 2218, 2228, 2172, 12948, 2159, 13004, 2189, 2214, 13121, 2331, 13120, 13121, 2334, 2332, 13110, 2331, 2321, 13154, 2362, 13153, 13151, 2362, 2359, 13164, 2372, 2370, 13151, 2357, 13149, 13154, 2370, 2363, 13179, 2372, 13166, 13124, 2357, 2334, 13110, 2261, 13055, 2368, 13100, 2311, 13029, 2123, 2241,
12940, 2259, 2149, 13251, 2462, 13250, 13248, 2462, 2459, 13264, 2474, 2472, 13264, 2470, 13262, 13251, 2470, 2463, 13374, 2580, 13373, 13374, 2583, 2581, 13393, 2603, 13392, 13393, 2606, 2604, 13398, 2606, 13396, 13382, 2603, 2593, 13408, 2614, 13407, 13408, 2617, 2615, 13411, 2620, 2617, 13382, 2583, 13377, 13398, 2614, 2608, 2499, 13291, 2498, 13291, 2495, 2498, 13302, 2509, 2511, 2509, 13298, 2507, 13298, 2499, 2507, 2548, 13335, 2547, 13339, 2548, 2550, 13311, 2521, 13310, 13311, 2524, 2522, 13328, 2535, 13327, 13325, 2535, 2532, 13316, 2524, 13314, 13316, 2532, 2526, 2572, 13363, 2571, 13367, 2572, 2574, 13363, 2562, 2571, 2562, 13339, 2550, 13335, 2511, 2547, 13328, 2564, 2536, 13473, 2684, 13472, 13470, 2684, 2681, 2651, 13439, 2650, 13439, 2647, 2650, 13452, 2661, 2663, 2647, 13435, 2645, 13450, 2651, 2661, 13470, 2620, 13416, 13266, 2580, 2474, 13435, 2574, 2645, 2080, 12740, 1951, 13248, 2417, 13213, 12854, 2400, 2065, 13289, 2368, 2495, 13179, 2521, 2390, 2967, 13757, 2966, 13761, 2967, 2969, 2971, 13761, 2969, 13757, 2958, 2966, 3048, 13835, 3047, 13839, 3048, 3050, 13881, 3092, 13880, 13878, 3092, 3089, 13895, 3101, 13894,
13895, 3104, 3102, 13884, 3093, 13881, 13884, 3101, 3095, 3064, 13857, 3063, 3079, 13867, 3078, 13867, 3075, 3078, 3075, 13863, 3073, 13863, 3064, 3073, 13914, 3125, 13913, 3063, 13839, 3050, 13913, 3104, 13898, 13903, 3079, 3114, 13972, 3180, 13971, 13969, 3180, 3177, 13969, 3175, 13967, 3136, 13925, 3135, 13929, 3136, 3138, 3142, 13931, 3141, 3141, 13929, 3138, 13941, 3142, 3152, 3162, 13951, 3161, 13955, 3162, 3164, 3168, 13957, 3167, 3167, 13955, 3164, 3161, 13941, 3152, 13991, 3199, 13990, 13995, 3203, 13994, 13995, 3206, 3204, 13994, 3200, 13991, 13980, 3199, 3191, 14016, 3227, 14015, 14030, 3236, 14029, 14029, 3233, 14027, 14015, 3206, 13998, 14016, 3233, 3228, 3135, 13903, 3114, 13980, 3181, 13972, 13914, 3175, 3126, 14005, 3168, 3216, 13775, 2983, 13774, 13772, 2983, 2980, 13783, 2990, 2988, 13772, 2978, 13770, 13775, 2988, 2984, 13808, 3018, 13807, 13808, 3021, 3019, 13825, 3032, 13824, 13822, 3032, 3029, 13813, 3021, 13811, 13813, 3029, 3023, 13797, 2990, 13785, 13797, 3018, 3008, 13825, 3036, 3033, 3249, 14036, 3248, 3253, 14040, 3252, 14044, 3253, 3255, 3252, 14037, 3249, 14085, 3295, 14084, 14085, 3298, 3296, 14091, 3301, 14090,
14090, 3298, 14088, 3271, 14062, 3270, 3275, 14066, 3274, 3274, 14063, 3271, 14074, 3275, 3285, 14102, 3313, 14101, 14116, 3321, 14115, 14105, 3314, 14102, 14105, 3321, 3316, 3270, 14044, 3255, 14101, 3302, 14091, 3334, 14121, 3333, 3338, 14125, 3337, 3337, 14122, 3334, 14162, 3369, 14161, 14166, 3373, 14165, 14165, 3370, 14162, 3350, 14137, 3349, 3354, 14141, 3353, 3353, 14138, 3350, 14182, 3389, 14181, 14186, 3393, 14185, 14185, 3390, 14182, 3349, 14126, 3338, 14181, 3374, 14166, 3333, 14074, 3285, 14161, 3322, 14116, 3248, 14005, 3216, 14084, 3237, 14030, 13835, 2971, 3047, 2958, 13703, 2914, 13878, 2955, 13751, 12620, 2708, 1829, 13567, 1858, 2772, 13473, 2978, 2685, 11104, 1488, 311, 12366, 393, 1574, 11927, 1862, 1135, 12725, 1233, 1935, 12164, 2071, 1373, 12948, 1437, 2159, 12244, 2176, 1449, 14199, 3410, 3408, 3414, 14203, 3413, 3400, 14193, 3398, 14207, 3410, 14201, 14193, 3408, 3398, 14203, 3400, 3413, 14225, 3431, 14224, 14225, 3434, 3432, 3425, 14217, 3424, 14217, 3421, 3424, 14252, 3461, 14251, 3450, 14237, 3449, 14241, 3450, 3452, 14258, 3467, 14257, 14255, 3467, 3464, 3449, 14233, 3444, 14255, 3462, 14252, 14235, 3461, 3446,
3421, 14204, 3414, 14235, 3434, 14228, 14207, 3431, 3416, 14233, 3425, 3444, 3474, 14267, 3473, 3491, 14279, 3490, 14279, 3487, 3490, 3487, 14273, 3483, 14273, 3474, 3483, 3494, 14280, 3491, 3473, 14241, 3452, 14275, 3468, 14258, 3500, 14293, 3499, 14299, 3508, 3506, 14293, 3506, 3499, 14307, 3516, 3514, 14316, 3521, 14315, 14309, 3521, 3516, 14301, 3514, 3508, 3530, 14321, 3529, 3534, 14325, 3533, 3533, 14322, 3530, 14370, 3577, 14369, 14374, 3581, 14373, 14373, 3578, 14370, 3550, 14341, 3549, 3554, 14345, 3553, 3553, 14342, 3550, 14390, 3597, 14389, 14394, 3601, 14393, 14393, 3598, 14390, 3549, 14326, 3534, 14389, 3582, 14374, 3529, 14294, 3500, 14369, 3522, 14316, 3646, 14436, 3645, 14436, 3649, 3645, 3649, 14434, 3642, 14431, 3646, 3635, 14401, 3610, 3608, 14413, 3624, 3622, 14407, 3612, 3614, 3632, 14421, 3631, 14421, 3627, 3631, 14418, 3632, 3628, 3612, 14401, 3608, 14403, 3622, 3610, 14417, 3614, 3627, 3662, 14449, 3661, 14449, 3664, 3661, 14468, 3677, 14467, 14468, 3674, 3678, 14472, 3681, 14471, 14471, 3674, 14465, 14472, 3666, 3682, 3664, 14431, 3635, 14467, 3650, 14441, 14415, 3662, 3624, 14541, 3747, 14540, 14540, 3744, 14538,
14482, 3692, 14481, 3711, 14498, 3710, 14487, 3699, 3697, 14503, 3714, 14502, 14487, 3693, 14482, 3699, 14485, 3695, 14485, 3711, 3695, 14527, 3734, 14526, 14530, 3734, 3737, 14532, 3735, 14527, 14516, 3719, 3721, 14522, 3725, 3727, 14520, 3721, 3725, 14522, 3715, 14503, 14532, 3714, 3739, 3710, 14418, 3628, 14530, 3666, 14455, 14434, 3744, 3642, 14541, 3650, 3748, 14556, 3767, 3765, 3757, 14549, 3756, 3756, 14547, 3753, 14547, 3765, 3753, 14596, 3807, 3805, 14610, 3821, 3819, 14602, 3809, 3811, 14616, 3823, 3825, 14598, 3819, 3807, 14614, 3811, 3823, 3777, 14568, 3776, 14581, 3792, 14580, 3781, 14572, 3780, 14585, 3796, 14584, 3780, 14569, 3777, 14584, 3793, 14581, 3776, 14550, 3757, 14580, 3769, 14560, 14558, 3805, 3767, 14600, 3769, 3809, 14696, 3909, 3905, 14701, 3906, 3910, 14712, 3919, 14711, 14709, 3919, 3916, 14696, 3902, 14694, 14700, 3916, 3909, 3933, 14722, 3932, 14726, 3933, 3935, 3939, 14728, 3938, 3938, 14726, 3935, 14722, 3923, 3932, 14744, 3955, 14743, 14741, 3955, 3952, 14748, 3959, 14747, 14747, 3956, 14744, 3923, 14701, 3910, 14741, 3942, 14738, 14712, 3942, 3920, 3833, 14624, 3832, 14637, 3848, 14636, 3837, 14628, 3836,
14641, 3852, 14640, 3836, 14625, 3833, 14640, 3849, 14637, 14666, 3873, 3871, 14672, 3879, 3877, 14668, 3877, 3873, 3859, 14652, 3858, 14660, 3865, 3867, 14658, 3859, 3865, 14680, 3885, 3887, 14688, 3893, 3895, 14686, 3887, 3893, 3858, 14629, 3837, 14641, 3871, 3853, 14678, 3867, 3885, 3982, 14770, 3981, 14768, 3982, 3978, 3969, 14761, 3968, 14759, 3969, 3965, 3984, 14768, 3978, 14761, 3984, 3968, 14802, 4012, 14801, 14806, 4016, 14805, 14809, 4016, 4019, 14809, 4013, 14802, 14788, 3995, 14787, 14788, 3992, 3996, 14797, 4000, 4002, 14795, 3995, 4000, 14797, 3986, 14776, 14806, 3986, 4017, 3832, 14573, 3781, 14636, 3797, 14585, 3981, 14729, 3939, 14801, 3960, 14748, 14612, 3902, 3821, 14697, 3825, 3906, 14674, 3965, 3879, 14785, 3895, 3992, 4025, 14818, 4024, 14824, 4033, 4031, 14818, 4031, 4024, 14832, 4041, 4039, 14840, 4049, 4047, 14834, 4047, 4041, 14826, 4039, 4033, 4057, 14848, 4056, 4061, 14852, 4060, 4060, 14849, 4057, 4077, 14868, 4076, 4081, 14872, 4080, 4080, 14869, 4077, 14913, 4123, 14912, 14917, 4127, 14916, 14920, 4127, 4130, 14920, 4124, 14913, 4076, 14853, 4061, 4056, 14819, 4025, 14275, 3692, 3485, 14514, 3494, 3719,
14929, 4140, 4138, 4144, 14933, 4143, 4143, 14929, 4138, 4156, 14943, 4155, 4160, 14947, 4159, 4159, 14944, 4156, 4155, 14934, 4144, 14985, 4196, 4194, 14999, 4210, 4208, 14991, 4198, 4200, 15005, 4212, 4214, 14987, 4208, 4196, 15003, 4200, 4212, 15025, 4236, 4234, 15032, 4238, 4242, 15041, 4252, 4249, 15031, 4236, 15027, 15031, 4249, 4241, 15042, 4242, 4250, 15014, 4221, 15013, 15018, 4225, 15017, 15017, 4222, 15014, 15053, 4256, 4258, 15013, 4198, 14989, 4256, 15042, 4250, 15001, 4234, 4210, 15029, 4214, 4238, 4172, 14959, 4171, 4176, 14963, 4175, 4175, 14960, 4172, 4184, 14975, 4183, 14979, 4184, 4186, 4183, 14964, 4176, 15058, 4266, 15057, 15062, 4270, 15061, 15062, 4273, 4271, 15072, 4279, 4277, 15061, 4267, 15058, 15065, 4277, 4273, 15094, 4301, 4299, 4310, 15102, 4309, 15100, 4310, 4306, 15096, 4306, 4301, 15080, 4285, 4287, 15088, 4293, 4295, 15086, 4287, 4293, 15111, 4316, 4318, 15120, 4327, 15119, 15120, 4324, 4328, 15117, 4318, 4324, 15074, 4299, 4279, 15109, 4295, 4316, 4171, 14948, 4160, 15057, 4226, 15018, 14931, 4194, 4140, 15078, 4186, 4285, 15304, 4510, 15303, 15304, 4513, 4511, 4524, 15312, 4523, 15332, 4539, 15331,
4532, 15320, 4531, 15320, 4528, 4531, 15336, 4543, 15335, 4528, 15313, 4524, 15335, 4540, 15332, 15316, 4539, 4526, 15316, 4513, 15307, 4597, 15389, 4596, 15387, 4597, 4593, 15396, 4606, 4603, 15389, 4603, 4596, 4588, 15380, 4587, 4587, 15378, 4584, 4619, 15406, 4618, 15425, 4631, 15424, 15428, 4631, 4634, 15413, 4624, 15412, 15412, 4621, 15410, 15425, 4621, 4632, 4618, 15387, 4593, 15428, 4607, 15399, 15378, 4619, 4584, 4556, 15343, 4555, 15343, 4558, 4555, 15364, 4570, 15363, 15352, 4563, 15351, 4560, 15344, 4556, 15367, 4571, 15364, 15367, 4563, 4573, 15374, 4577, 4579, 15374, 4564, 15352, 4641, 15433, 4640, 4643, 15434, 4641, 15443, 4654, 15442, 4659, 15446, 4658, 15450, 4659, 4661, 15465, 4673, 15464, 15462, 4673, 4670, 15462, 4668, 15460, 15443, 4668, 4655, 4686, 15478, 4685, 15478, 4682, 4685, 15476, 4677, 4682, 4677, 15450, 4661, 15446, 4643, 4658, 4558, 15321, 4532, 15363, 4544, 15336, 4640, 15381, 4588, 15442, 4625, 15413, 15127, 4336, 4334, 15135, 4344, 4342, 15129, 4342, 4336, 15143, 4352, 4350, 15151, 4360, 4358, 15145, 4358, 4352, 15137, 4350, 4344, 15215, 4424, 4422, 15223, 4432, 4430, 15217, 4430, 4424, 15231, 4440, 4438,
15239, 4448, 4446, 15233, 4446, 4440, 15225, 4438, 4432, 15177, 4380, 4382, 15183, 4386, 4388, 15181, 4382, 4386, 15205, 4408, 4410, 15211, 4414, 4416, 15209, 4410, 4414, 15203, 4388, 4408, 15265, 4468, 4470, 15271, 4474, 4476, 15269, 4470, 4474, 15293, 4496, 4498, 15299, 4502, 4504, 15297, 4498, 4502, 15291, 4476, 4496, 15153, 4422, 4360, 15263, 4416, 4468, 15485, 4694, 4692, 15493, 4702, 4700, 15487, 4700, 4694, 15504, 4711, 15503, 15501, 4711, 4708, 15495, 4708, 4702, 4768, 15561, 4767, 15567, 4776, 4774, 15575, 4784, 4782, 15569, 4782, 4776, 15561, 4774, 4767, 15540, 4751, 15539, 15544, 4755, 15543, 15543, 4752, 15540, 15529, 4732, 4734, 15535, 4738, 4740, 15533, 4734, 4738, 15557, 4760, 4762, 15557, 4756, 15544, 15555, 4740, 4760, 4792, 15583, 4791, 4796, 15587, 4795, 4795, 15584, 4792, 4804, 15595, 4803, 4808, 15599, 4807, 4807, 15596, 4804, 4803, 15588, 4796, 15539, 4712, 15504, 4791, 15562, 4768, 15045, 4510, 4252, 15312, 4258, 4523, 15102, 4560, 4309, 15372, 4327, 4577, 15241, 4692, 4448, 15527, 4504, 4732, 15639, 4850, 4848, 15655, 4866, 4864, 15647, 4854, 4856, 15661, 4868, 4870, 4853, 15639, 4848, 15641, 4864, 4850,
15659, 4856, 4868, 15669, 4880, 4878, 15676, 4882, 4886, 15688, 4895, 15687, 15685, 4895, 4892, 15675, 4880, 15671, 15675, 4892, 4885, 4909, 15700, 4908, 15700, 4905, 4908, 4911, 15701, 4909, 15698, 4899, 4905, 4899, 15676, 4886, 15657, 4878, 4866, 15673, 4870, 4882, 15780, 4991, 15779, 15777, 4991, 4988, 4996, 15783, 4995, 15788, 4999, 15787, 15787, 4992, 15780, 15777, 4978, 15774, 15710, 4917, 4915, 15716, 4923, 4921, 15712, 4921, 4917, 15738, 4945, 4943, 4956, 15746, 4955, 15746, 4958, 4955, 4958, 15744, 4952, 15740, 4956, 4945, 15724, 4929, 4931, 15732, 4937, 4939, 15730, 4931, 4937, 15759, 4964, 4966, 15766, 4972, 15765, 15769, 4972, 4975, 15766, 4966, 4973, 15718, 4943, 4923, 15757, 4939, 4964, 5012, 15799, 5011, 15804, 5015, 15803, 15810, 5021, 15809, 15810, 5018, 5022, 15809, 5016, 15804, 15821, 5018, 15807, 5011, 15784, 4996, 15803, 5000, 15788, 15688, 4978, 4896, 15783, 4911, 4995, 15744, 5012, 4952, 15821, 4975, 5025, 15962, 5173, 5171, 5177, 15966, 5176, 5176, 15962, 5171, 15995, 5201, 15994, 15994, 5198, 15992, 5189, 15976, 5188, 5193, 15980, 5192, 5192, 15977, 5189, 16002, 5209, 16001, 16006, 5213, 16005, 5223, 16015, 5222,
16013, 5223, 5219, 16005, 5210, 16002, 5188, 15967, 5177, 16001, 5202, 15995, 15964, 5198, 5173, 5294, 16086, 5293, 16084, 5294, 5290, 16129, 5339, 5337, 16141, 5347, 16140, 5339, 16127, 5335, 16140, 5344, 16138, 16127, 5344, 5335, 5307, 16095, 5306, 16093, 5307, 5303, 16115, 5325, 16114, 16118, 5325, 5328, 5309, 16093, 5303, 16118, 5322, 16112, 16095, 5320, 5306, 16115, 5311, 5326, 16162, 5369, 16161, 5362, 16149, 5361, 16166, 5373, 16165, 5361, 16147, 5358, 16165, 5370, 16162, 5320, 16084, 5290, 16161, 5348, 16141, 16086, 5337, 5293, 16147, 5322, 5358, 5235, 16022, 5234, 5239, 16026, 5238, 16030, 5239, 5241, 5238, 16023, 5235, 16044, 5255, 16043, 5260, 16047, 5259, 16044, 5260, 5256, 16066, 5276, 16065, 16069, 5276, 5279, 16069, 5273, 16063, 16047, 5271, 5259, 16066, 5262, 5277, 16080, 5283, 5285, 5244, 16030, 5241, 16080, 5262, 16051, 16078, 5244, 5283, 16174, 5380, 16173, 16177, 5380, 5383, 5395, 16182, 5394, 16199, 5406, 16198, 5399, 16186, 5398, 16203, 5410, 16202, 5398, 16183, 5395, 16202, 5407, 16199, 5419, 16210, 5418, 16214, 5419, 5421, 16221, 5426, 16220, 16221, 5421, 5427, 5418, 16187, 5399, 16220, 5411, 16203, 5234, 15981, 5193,
16043, 5214, 16006, 5271, 16013, 5219, 16177, 5311, 16101, 5394, 16150, 5362, 16198, 5374, 16166, 16015, 5309, 5222, 16174, 5273, 5381, 15825, 5030, 15824, 5063, 15854, 5062, 15854, 5065, 5062, 15864, 5073, 5071, 15858, 5071, 5065, 15875, 5082, 15874, 15872, 5082, 5079, 15866, 5079, 5073, 15831, 5038, 15830, 15835, 5042, 15834, 15834, 5039, 15831, 15843, 5050, 15842, 15847, 5054, 15846, 15846, 5051, 15843, 15842, 5043, 15835, 5095, 15882, 5094, 5099, 15886, 5098, 5098, 15883, 5095, 15931, 5142, 15930, 15935, 5146, 15934, 15934, 5143, 15931, 5119, 15906, 5118, 5123, 15910, 5122, 5122, 15907, 5119, 15947, 5158, 15946, 15951, 5162, 15950, 15950, 5159, 15947, 5118, 15887, 5099, 15946, 5147, 15935, 15830, 5031, 15825, 5094, 15855, 5063, 15930, 5083, 15875, 15398, 4853, 4606, 15644, 4607, 4854, 15465, 4915, 4674, 15722, 4686, 4929, 15577, 5030, 4784, 14842, 4334, 4049, 15175, 4123, 4380, 29652, 21605, 21603, 21609, 29656, 21608, 29669, 21617, 29668, 21608, 29652, 21603, 29668, 21614, 29666, 29654, 21614, 21605, 21630, 29675, 21629, 29692, 21643, 29691, 21634, 29679, 21633, 29696, 21647, 29695, 21652, 29699, 21651, 21633, 29676, 21630, 29695, 21644, 29692, 21629, 29657, 21609,
29691, 21618, 29669, 29761, 21712, 21710, 29769, 21720, 21718, 29763, 21718, 21712, 29780, 21729, 29779, 29777, 21729, 21726, 29771, 21726, 21720, 21664, 29709, 21663, 29726, 21677, 29725, 21668, 29713, 21667, 29732, 21683, 29731, 29732, 21686, 21684, 21667, 29710, 21664, 29731, 21678, 29726, 21694, 29743, 21693, 29747, 21694, 21696, 29755, 21702, 21704, 29753, 21696, 21702, 21693, 29714, 21668, 29789, 21736, 21738, 29801, 21748, 21750, 29791, 21742, 21740, 29803, 21754, 21752, 29799, 21738, 21748, 29793, 21752, 21742, 29829, 21776, 21778, 29831, 21782, 21780, 29813, 21760, 21762, 29821, 21768, 21770, 29819, 21762, 21768, 29842, 21791, 29841, 29842, 21788, 21792, 29841, 21782, 29833, 29827, 21750, 21776, 29805, 21780, 21754, 29839, 21770, 21788, 21663, 29680, 21634, 29725, 21648, 29696, 21680, 29700, 21652, 29829, 21730, 29780, 29699, 21710, 21651, 29787, 21680, 21736, 29735, 21740, 21686, 29811, 21704, 21760, 21805, 29852, 21804, 21809, 29856, 21808, 29856, 21811, 21808, 21811, 29853, 21805, 21804, 29849, 21795, 21835, 29884, 21834, 29884, 21837, 21834, 29896, 21849, 21846, 21849, 29894, 21844, 29888, 21846, 21837, 29929, 21882, 21880, 29943, 21896, 21894, 29935, 21884, 21886, 29949, 21898, 21900, 29931, 21894, 21882, 29947, 21886, 21898,
29910, 21861, 29909, 29910, 21858, 21862, 29925, 21870, 21872, 29914, 21865, 29913, 29913, 21858, 29907, 29923, 21861, 21870, 29925, 21850, 29899, 29894, 21880, 21844, 29933, 21850, 21884, 21821, 29868, 21820, 29872, 21821, 21823, 21827, 29874, 21826, 21826, 29872, 21823, 29958, 21910, 29957, 29958, 21913, 21911, 29975, 21923, 29974, 29975, 21926, 21924, 29965, 21913, 29961, 29965, 21923, 21917, 29999, 21948, 21946, 30005, 21954, 21952, 30001, 21952, 21948, 29985, 21930, 21932, 21940, 29989, 21939, 29993, 21940, 21942, 29989, 21932, 21939, 30013, 21960, 21962, 30021, 21968, 21970, 30019, 21962, 21968, 21930, 29963, 21915, 29978, 21946, 21926, 30011, 21942, 21960, 21820, 29857, 21809, 29957, 21866, 29914, 29849, 21835, 21795, 29963, 21827, 21915, 30137, 22091, 22088, 30149, 22097, 30148, 30149, 22100, 22098, 30137, 22085, 30135, 30141, 22097, 22091, 30159, 22104, 22106, 22116, 30163, 22115, 30168, 22116, 22120, 22124, 30171, 22123, 22123, 30168, 22120, 30163, 22106, 22115, 30182, 22134, 30181, 30182, 22137, 22135, 30188, 22140, 30187, 30187, 22137, 30185, 22104, 30138, 22089, 30167, 22100, 30152, 30167, 22134, 22119, 30265, 22218, 22216, 22208, 30258, 22207, 30270, 22221, 30269, 22207, 30256, 22204, 30269, 22218, 30267, 30256, 22216, 22204,
22230, 30279, 22229, 30292, 22245, 30291, 22234, 30283, 22233, 30296, 22249, 30295, 22233, 30280, 22230, 30295, 22246, 30292, 22229, 30259, 22208, 30291, 22222, 30270, 22163, 30209, 22162, 30207, 22163, 22159, 22150, 30200, 22149, 30198, 22150, 22146, 22165, 30207, 22159, 30200, 22165, 22149, 30243, 22194, 30242, 30243, 22191, 22195, 30247, 22198, 30246, 30246, 22191, 30240, 30227, 22176, 30226, 30227, 22173, 22177, 30236, 22181, 22183, 30234, 22176, 22181, 30236, 22167, 30215, 30247, 22167, 22199, 22258, 30307, 22257, 30320, 22273, 30319, 22262, 30311, 22261, 30324, 22277, 30323, 22261, 30308, 22258, 30323, 22274, 30320, 22284, 30335, 22283, 30343, 22290, 22292, 30341, 22284, 22290, 22283, 30312, 22262, 22162, 30172, 22124, 30242, 22141, 30188, 22257, 30284, 22234, 30319, 22250, 30296, 21978, 30027, 21977, 30027, 21980, 21977, 30037, 21988, 21986, 30031, 21986, 21980, 30048, 21997, 30047, 30045, 21997, 21994, 30039, 21994, 21988, 22010, 30055, 22009, 22014, 30059, 22013, 22013, 30056, 22010, 30104, 22057, 30103, 30108, 22061, 30107, 30107, 22058, 30104, 22034, 30079, 22033, 22038, 30083, 22037, 22037, 30080, 22034, 30120, 22073, 30119, 30124, 22077, 30123, 30123, 22074, 30120, 22033, 30060, 22014, 30119, 22062, 30108, 22009, 30028, 21978,
30103, 21998, 30048, 29945, 22085, 21896, 30138, 21900, 22089, 30007, 22146, 21954, 30224, 21970, 22173, 22319, 30369, 22318, 30367, 22319, 22315, 30355, 22306, 22304, 30361, 22308, 22310, 22298, 30349, 22296, 30349, 22304, 22296, 30359, 22298, 22308, 22330, 30378, 22329, 30376, 22330, 22326, 30384, 22332, 22334, 22348, 30393, 22347, 30397, 22348, 22350, 30378, 22344, 22329, 30393, 22334, 22347, 22344, 30367, 22315, 30357, 22326, 22306, 30382, 22310, 22332, 30468, 22421, 22418, 30482, 22430, 30481, 22421, 30466, 22416, 30481, 22427, 30479, 30466, 22427, 22416, 30482, 22422, 22431, 30489, 22442, 30488, 30492, 22442, 22446, 30489, 22447, 22443, 30509, 22456, 30508, 30498, 22447, 30493, 30498, 22456, 22451, 30492, 22422, 30471, 30407, 22356, 22354, 30413, 22362, 22360, 30409, 22360, 22356, 30435, 22384, 22382, 22393, 30443, 22392, 30441, 22393, 22389, 30437, 22389, 22384, 30421, 22368, 22370, 30429, 22376, 22378, 30427, 22370, 22376, 30452, 22399, 22401, 30460, 22407, 22409, 30458, 22401, 22407, 30415, 22382, 22362, 30450, 22378, 22399, 22469, 30514, 22468, 30529, 22478, 30528, 30533, 22482, 30532, 22471, 30515, 22469, 30532, 22479, 30529, 30541, 22489, 30540, 30544, 22489, 22492, 30549, 22492, 22495, 30549, 22483, 30533, 22468, 30496, 22449,
30528, 22457, 30509, 30369, 22418, 22318, 30496, 22350, 22449, 30443, 22471, 22392, 30541, 22409, 22490, 22503, 30552, 22502, 30552, 22505, 22502, 30562, 22513, 22511, 30556, 22511, 22505, 30570, 22521, 22519, 30578, 22529, 22527, 30572, 22527, 22521, 30564, 22519, 22513, 30672, 22623, 22621, 30680, 22631, 22629, 30674, 22629, 22623, 30688, 22639, 22637, 30696, 22647, 22645, 30690, 22645, 22639, 30682, 22637, 22631, 22541, 30586, 22540, 22545, 30590, 22544, 22544, 30587, 22541, 22565, 30610, 22564, 22569, 30614, 22568, 22568, 30611, 22565, 30651, 22604, 30650, 30668, 22613, 22615, 30655, 22608, 30654, 30654, 22605, 30651, 30666, 22604, 22613, 22564, 30591, 22545, 30722, 22667, 22669, 30728, 22673, 22675, 30726, 22669, 22673, 30750, 22695, 22697, 30756, 22701, 22703, 30754, 22697, 22701, 30748, 22675, 22695, 22540, 30553, 22503, 30580, 22621, 22529, 30720, 22615, 22667, 30760, 22711, 22709, 30768, 22719, 22717, 30762, 22717, 22711, 30776, 22727, 22725, 30784, 22735, 22733, 30778, 22733, 22727, 30770, 22725, 22719, 30848, 22799, 22797, 30856, 22807, 22805, 30850, 22805, 22799, 30865, 22812, 30864, 30858, 22812, 22807, 30810, 22755, 22757, 30816, 22761, 22763, 30814, 22757, 22761, 30838, 22783, 22785, 30844, 22789, 22791, 30842, 22785, 22789,
30836, 22763, 22783, 30899, 22848, 30898, 30903, 22852, 30902, 30902, 22849, 30899, 30888, 22833, 22835, 30894, 22839, 22841, 30892, 22835, 22839, 30910, 22853, 30903, 30910, 22841, 22856, 30898, 22813, 30865, 30786, 22797, 22735, 30886, 22791, 22833, 30698, 22709, 22647, 30808, 22703, 22755, 30913, 22866, 22864, 30927, 22880, 22878, 22870, 30917, 22869, 30933, 22882, 22884, 22869, 30913, 22864, 30915, 22878, 22866, 30969, 22922, 22920, 30983, 22936, 22934, 30975, 22924, 22926, 30989, 22938, 22940, 30971, 22934, 22922, 30987, 22926, 22938, 22896, 30941, 22895, 30958, 22907, 30957, 22900, 30945, 22899, 30962, 22911, 30961, 22899, 30942, 22896, 30961, 22908, 30958, 22948, 30997, 22947, 30997, 22950, 22947, 31007, 22958, 22956, 31001, 22956, 22950, 22895, 30918, 22870, 30957, 22882, 30931, 30929, 22920, 22880, 30973, 22884, 22924, 31104, 23057, 23055, 31111, 23059, 23063, 31120, 23073, 23070, 31110, 23057, 31106, 31110, 23070, 23062, 31121, 23063, 23071, 31156, 23103, 31155, 23094, 31140, 23093, 31140, 23090, 23093, 31130, 23081, 23079, 31144, 23098, 23096, 23098, 31141, 23094, 31132, 23096, 23081, 31164, 23117, 31163, 31161, 23117, 23114, 31168, 23121, 31167, 31167, 23118, 31164, 23090, 31121, 23071, 31161, 23104, 31156, 31124, 23103, 23073,
22970, 31015, 22969, 31032, 22982, 31031, 22974, 31019, 22973, 31036, 22986, 31035, 31036, 22989, 22987, 22973, 31016, 22970, 31035, 22983, 31032, 31066, 23013, 23015, 31078, 23025, 23027, 31068, 23019, 23017, 31080, 23031, 23029, 31076, 23015, 23025, 31070, 23029, 23019, 22997, 31046, 22996, 31050, 22997, 22999, 31058, 23005, 23007, 31056, 22999, 23005, 31090, 23037, 23039, 31098, 23045, 23047, 31096, 23039, 23045, 22996, 31020, 22974, 31039, 23017, 22989, 31088, 23007, 23037, 31181, 23128, 23130, 23146, 31191, 23145, 31195, 23146, 23148, 31183, 23134, 23132, 23152, 31197, 23151, 31197, 23154, 23151, 23154, 31195, 23148, 31191, 23130, 23145, 31185, 23152, 23134, 31229, 23181, 31228, 31233, 23185, 31232, 31236, 23185, 23188, 31236, 23182, 31229, 31213, 23160, 23162, 31220, 23168, 31219, 31223, 23168, 23171, 31220, 23162, 23169, 31245, 23186, 31233, 31245, 23171, 23191, 22969, 30946, 22900, 31031, 22912, 30962, 23013, 30998, 22948, 31228, 23122, 31168, 30985, 23055, 22936, 31108, 22940, 23059, 31009, 23079, 22958, 31179, 23027, 23128, 31082, 23132, 23031, 31211, 23047, 23160, 31413, 23366, 23364, 23370, 31417, 23369, 23369, 31413, 23364, 23384, 31429, 23383, 31434, 23387, 31433, 23392, 31437, 23391, 31442, 23395, 31441, 23391, 31430, 23384,
31441, 23388, 31434, 23383, 31418, 23370, 31433, 23372, 31421, 31489, 23442, 23440, 31503, 23456, 23454, 31495, 23444, 23446, 31509, 23458, 23460, 31491, 23454, 23442, 31507, 23446, 23458, 31517, 23470, 23468, 31531, 23484, 23482, 31523, 23472, 23474, 31538, 23489, 31537, 31538, 23486, 23490, 31537, 23484, 31533, 31519, 23482, 23470, 31535, 23474, 23486, 31505, 23468, 23456, 31521, 23460, 23472, 23408, 31453, 23407, 31458, 23411, 31457, 23416, 31461, 23415, 31466, 23419, 31465, 23415, 31454, 23408, 31465, 23412, 31458, 23428, 31477, 23427, 31482, 23431, 31481, 31482, 23428, 23432, 23427, 31462, 23416, 31481, 23420, 31466, 23407, 31438, 23392, 31457, 23396, 31442, 31415, 23440, 23366, 31493, 23372, 23444, 31316, 23265, 23263, 31313, 23263, 23259, 31322, 23271, 23269, 23280, 31330, 23279, 31328, 23280, 23276, 31324, 23276, 23271, 31318, 23269, 23265, 23240, 31286, 23239, 31284, 23240, 23236, 23225, 31273, 23224, 31273, 23227, 23224, 23227, 31271, 23221, 31271, 23236, 23221, 23210, 31256, 23209, 31254, 23210, 23206, 23214, 31260, 23213, 23213, 31254, 23206, 31251, 23214, 23197, 23209, 31248, 23194, 31248, 23225, 23194, 31337, 23292, 23290, 31373, 23321, 23323, 31360, 23313, 31359, 31360, 23310, 23314, 31375, 23321, 31371, 31375, 23313, 23325,
31399, 23347, 31398, 31402, 23347, 23350, 31409, 23354, 23356, 31407, 23350, 23354, 31399, 23323, 23348, 23292, 31313, 23259, 31286, 23290, 23239, 31547, 23498, 23496, 31555, 23506, 23504, 31549, 23504, 23498, 31563, 23514, 23512, 31574, 23523, 31573, 31571, 23523, 23520, 31565, 23520, 23514, 31557, 23512, 23506, 31637, 23586, 23584, 31643, 23592, 23590, 31639, 23590, 23586, 31645, 23595, 23592, 31599, 23544, 23546, 31605, 23550, 23552, 31603, 23546, 23550, 31627, 23572, 23574, 31633, 23578, 23580, 31631, 23574, 23578, 31625, 23552, 23572, 31688, 23643, 23641, 31703, 23656, 31702, 31700, 23656, 23653, 31690, 23653, 23643, 23620, 31670, 23619, 31670, 23616, 23619, 31679, 23627, 23629, 23631, 31679, 23629, 31677, 23620, 23627, 23670, 31716, 23669, 31716, 23666, 23669, 31727, 23670, 23680, 31714, 23631, 23666, 31688, 23595, 31649, 31574, 23584, 23524, 31668, 23580, 23616, 31330, 23496, 23279, 31597, 23356, 23544, 30324, 22354, 22278, 30419, 22292, 22368, 31737, 23690, 23688, 31751, 23704, 23702, 23694, 31741, 23693, 31757, 23706, 23708, 23693, 31737, 23688, 31739, 23702, 23690, 31799, 23752, 23750, 31813, 23766, 23764, 31805, 23754, 23756, 31819, 23768, 23770, 31801, 23764, 23752, 31817, 23756, 23768, 23720, 31765, 23719, 31782, 23733, 31781,
23724, 31769, 23723, 31786, 23737, 31785, 23742, 31789, 23741, 23723, 31766, 23720, 31785, 23734, 31782, 31827, 23778, 23776, 31835, 23786, 23784, 31829, 23784, 23778, 23719, 31742, 23694, 31781, 23706, 31755, 31753, 23750, 23704, 31803, 23708, 23754, 31789, 23776, 23741, 23798, 31843, 23797, 31860, 23811, 31859, 23802, 31847, 23801, 31866, 23817, 31865, 31866, 23820, 23818, 23801, 31844, 23798, 31865, 23812, 31860, 31897, 23844, 23846, 31909, 23856, 23858, 31899, 23850, 23848, 31911, 23862, 23860, 31907, 23846, 23856, 31901, 23860, 23850, 23828, 31877, 23827, 31881, 23828, 23830, 31889, 23836, 23838, 31887, 23830, 23836, 31921, 23868, 23870, 31929, 23876, 23878, 31927, 23870, 23876, 23827, 31848, 23802, 31895, 23814, 23844, 31869, 23848, 23820, 31919, 23838, 23868, 23797, 31770, 23724, 31859, 23738, 31786, 23814, 31790, 23742, 32035, 23988, 23986, 32041, 23990, 23992, 32052, 24000, 32051, 32051, 23997, 32049, 32037, 23997, 23988, 32052, 23992, 24001, 32061, 24010, 32060, 32058, 24010, 24007, 24049, 32099, 24048, 24048, 32097, 24045, 32118, 24071, 24069, 32123, 24074, 32122, 32122, 24071, 32120, 24057, 32106, 24056, 24061, 32110, 24060, 24060, 32107, 24057, 32133, 24086, 32132, 32137, 24090, 32136, 32136, 24087, 32133, 24056, 32100, 24049,
32132, 24075, 32123, 32097, 24069, 24045, 32070, 24017, 24019, 32072, 24023, 24021, 32074, 24026, 24023, 32085, 24032, 24034, 32092, 24039, 32091, 32092, 24034, 24040, 32091, 24026, 32080, 24102, 32151, 24101, 24106, 32155, 24105, 24105, 32152, 24102, 32170, 24123, 32169, 24136, 32181, 24135, 32181, 24138, 24135, 32190, 24143, 32189, 32190, 24140, 24144, 32170, 24136, 24124, 24112, 32163, 24111, 32202, 24150, 32201, 32205, 24150, 24153, 32210, 24153, 24156, 24111, 32156, 24106, 32210, 24140, 32187, 32202, 24112, 24151, 32070, 24011, 32061, 24101, 32111, 24061, 32169, 24091, 32137, 24138, 32148, 24094, 23930, 31981, 23929, 31987, 23938, 23936, 31981, 23936, 23929, 31940, 23890, 31939, 31937, 23890, 23887, 31937, 23885, 31935, 31940, 23894, 23891, 31962, 23915, 31961, 31959, 23915, 23912, 31965, 23916, 31962, 31976, 23923, 31975, 31959, 23894, 31948, 31965, 23923, 23918, 23946, 31995, 23945, 23950, 31999, 23949, 23949, 31996, 23946, 23966, 32015, 23965, 23970, 32019, 23969, 23969, 32016, 23966, 23965, 32000, 23950, 23945, 31982, 23930, 32213, 24164, 24162, 32222, 24169, 32221, 32215, 24169, 24164, 24216, 32267, 24215, 32273, 24224, 24222, 32281, 24232, 24230, 32275, 24230, 24224, 32267, 24222, 24215, 32236, 24185, 32235, 32240, 24189, 32239,
32239, 24186, 32236, 32256, 24205, 32255, 32260, 24209, 32259, 32259, 24206, 32256, 32255, 24190, 32240, 24240, 32289, 24239, 24244, 32293, 24243, 24243, 32290, 24240, 24252, 32301, 24251, 24256, 32305, 24255, 24255, 32302, 24252, 24251, 32294, 24244, 32235, 24170, 32222, 24239, 32268, 24216, 31815, 23986, 23766, 32039, 23770, 23990, 31837, 24007, 23786, 32068, 23858, 24017, 31913, 24021, 23862, 32083, 23878, 24032, 31989, 24162, 23938, 24332, 32381, 24331, 32381, 24334, 24331, 32392, 24339, 32391, 32385, 24339, 24334, 24316, 32361, 24315, 32361, 24318, 24315, 24302, 32350, 24301, 32350, 24304, 24301, 32370, 24323, 32369, 32370, 24320, 24324, 24304, 32348, 24298, 32348, 24316, 24298, 32399, 24349, 24351, 32415, 24364, 32414, 24357, 32403, 24356, 32419, 24368, 32418, 32401, 24349, 32397, 24356, 32399, 24351, 32418, 24365, 32415, 24318, 32345, 24291, 32414, 24340, 32392, 32345, 24332, 24291, 32401, 24323, 24353, 32429, 24379, 32428, 32429, 24376, 24380, 32432, 24376, 32426, 32439, 24379, 24385, 24393, 32442, 24392, 32455, 24404, 32454, 24397, 32446, 24396, 32459, 24408, 32458, 24396, 32443, 24393, 32458, 24405, 32455, 24415, 32466, 24414, 32473, 24420, 32472, 32473, 24415, 24421, 24414, 32447, 24397, 32472, 24409, 32459, 32439, 24320, 32367,
24392, 32404, 24357, 32454, 24369, 32419, 32673, 24626, 24624, 24616, 32666, 24615, 32679, 24628, 24630, 24615, 32664, 24612, 32664, 24624, 24612, 32711, 24664, 24662, 32725, 24678, 24676, 32717, 24666, 24668, 32731, 24680, 24682, 32713, 24676, 24664, 32729, 24668, 24680, 24638, 32687, 24637, 32700, 24649, 32699, 24642, 32691, 24641, 32704, 24653, 32703, 24641, 32688, 24638, 32703, 24650, 32700, 24637, 32667, 24616, 32699, 24628, 32677, 32675, 24662, 24626, 32715, 24630, 24666, 32810, 24764, 24761, 32822, 24770, 32821, 32822, 24773, 24771, 32810, 24758, 32808, 32814, 24770, 24764, 32832, 24777, 24779, 24789, 32836, 24788, 32841, 24789, 24793, 24797, 32844, 24796, 24796, 32841, 24793, 32836, 24779, 24788, 32855, 24807, 32854, 32855, 24810, 24808, 32861, 24813, 32860, 32860, 24810, 32858, 24777, 32811, 24762, 32840, 24773, 32825, 32840, 24807, 24792, 24690, 32739, 24689, 32752, 24702, 32751, 24694, 32743, 24693, 32756, 24706, 32755, 32756, 24709, 24707, 24693, 32740, 24690, 32755, 24703, 32752, 32780, 24729, 24727, 32786, 24735, 24733, 32782, 24733, 24729, 24715, 32766, 24714, 32774, 24721, 24723, 32772, 24715, 24721, 32794, 24741, 24743, 32802, 24749, 24751, 32800, 24743, 24749, 24714, 32744, 24694, 32759, 24727, 24709, 32792, 24723, 24741,
24836, 32882, 24835, 32880, 24836, 24832, 24823, 32873, 24822, 32871, 24823, 24819, 24838, 32880, 24832, 32873, 24838, 24822, 32916, 24867, 32915, 32916, 24864, 24868, 32920, 24871, 32919, 32919, 24864, 32913, 32900, 24849, 32899, 32900, 24846, 24850, 32909, 24854, 24856, 32907, 24849, 24854, 32909, 24840, 32888, 32920, 24840, 24872, 24689, 32692, 24642, 32751, 24654, 32704, 24835, 32845, 24797, 32915, 24814, 32861, 32727, 24758, 24678, 32811, 24682, 24762, 32788, 24819, 24735, 32897, 24751, 24846, 32478, 24429, 24427, 32486, 24437, 24435, 32480, 24435, 24429, 32494, 24445, 24443, 32502, 24453, 24451, 32496, 24451, 24445, 32488, 24443, 24437, 32576, 24527, 24525, 32584, 24535, 24533, 32578, 24533, 24527, 32592, 24543, 24541, 32600, 24551, 24549, 32594, 24549, 24543, 32586, 24541, 24535, 32527, 24480, 32526, 32544, 24489, 24491, 32531, 24484, 32530, 32530, 24481, 32527, 32542, 24480, 24489, 32566, 24511, 24513, 32572, 24517, 24519, 32570, 24513, 24517, 32564, 24491, 24511, 32626, 24571, 24573, 32632, 24577, 24579, 32630, 24573, 24577, 32654, 24599, 24601, 32660, 24605, 24607, 32658, 24601, 24605, 32652, 24579, 24599, 32504, 24525, 24453, 32624, 24519, 24571, 32930, 24878, 32929, 32930, 24881, 24879, 32938, 24887, 24885, 32933, 24885, 24881,
32940, 24890, 24887, 32979, 24934, 24932, 32994, 24947, 32993, 32991, 24947, 24944, 32981, 24944, 24934, 32965, 24910, 24912, 24920, 32969, 24919, 32973, 24920, 24922, 32969, 24912, 24919, 33009, 24958, 24960, 24972, 33017, 24971, 33021, 24972, 24974, 24958, 33005, 24956, 33017, 24960, 24971, 33005, 24922, 24956, 33036, 24986, 33035, 33033, 24986, 24983, 33033, 24981, 33031, 33036, 24990, 24987, 33058, 25011, 33057, 33055, 25011, 25008, 33072, 25021, 33071, 33072, 25024, 25022, 33071, 25018, 33069, 33055, 24990, 33044, 33058, 25018, 25012, 32979, 24890, 32944, 32994, 24981, 24948, 32602, 24878, 24551, 32963, 24607, 24910, 32148, 24302, 24094, 32432, 24143, 24382, 32283, 24427, 24232, 33081, 25034, 25032, 25038, 33085, 25037, 25037, 33081, 25032, 25052, 33097, 25051, 33102, 25055, 33101, 25060, 33105, 25059, 33110, 25063, 33109, 25059, 33098, 25052, 33109, 25056, 33102, 25051, 33086, 25038, 33101, 25040, 33089, 33153, 25106, 25104, 33167, 25120, 25118, 33159, 25108, 25110, 33173, 25122, 25124, 33155, 25118, 25106, 33171, 25110, 25122, 33195, 25148, 25146, 33209, 25162, 25160, 33201, 25150, 25152, 33216, 25167, 33215, 33216, 25164, 25168, 33215, 25162, 33211, 33197, 25160, 25148, 33213, 25152, 25164, 25130, 33181, 25129, 33187, 25138, 25136,
33181, 25136, 25129, 33225, 25176, 25174, 33236, 25185, 33235, 33233, 25185, 25182, 33227, 25182, 25176, 33169, 25146, 25120, 33199, 25124, 25150, 33189, 25174, 25138, 25076, 33121, 25075, 33126, 25079, 33125, 25084, 33129, 25083, 25083, 33122, 25076, 33133, 25080, 33126, 25094, 33143, 25093, 33147, 25094, 25096, 25093, 33130, 25084, 33257, 25204, 25206, 33246, 25195, 33245, 33246, 25198, 25196, 33259, 25210, 25208, 33255, 25192, 25204, 33249, 25208, 25198, 33285, 25232, 25234, 33297, 25244, 25246, 33287, 25238, 25236, 33299, 25250, 25248, 33295, 25234, 25244, 33289, 25248, 25238, 33269, 25216, 25218, 33277, 25224, 25226, 33275, 25218, 25224, 33309, 25256, 25258, 33318, 25267, 33317, 33318, 25264, 25268, 33315, 25258, 25264, 33317, 25250, 33301, 33283, 25206, 25232, 33261, 25236, 25210, 33307, 25226, 25256, 25075, 33106, 25060, 33125, 25064, 33110, 25192, 33182, 25130, 33297, 25186, 33236, 33083, 25104, 25034, 33157, 25040, 25108, 33133, 25195, 25086, 33267, 25096, 25216, 33489, 25442, 25440, 25446, 33493, 25445, 33506, 25454, 33505, 25445, 33489, 25440, 33505, 25451, 33503, 33491, 25451, 25442, 25467, 33512, 25466, 33529, 25478, 33528, 25471, 33516, 25470, 33533, 25482, 33532, 25470, 33513, 25467, 33532, 25479, 33529, 33540, 25491, 25489,
33552, 25503, 25501, 33546, 25493, 25495, 33558, 25505, 25507, 25493, 33540, 25489, 33542, 25501, 25491, 33556, 25495, 25505, 25466, 33494, 25446, 33528, 25455, 33506, 25519, 33564, 25518, 33581, 25532, 33580, 25523, 33568, 25522, 33587, 25538, 33586, 33584, 25538, 25535, 25522, 33565, 25519, 33584, 25533, 33581, 33614, 25563, 25561, 33620, 25569, 25567, 33616, 25567, 25563, 25547, 33596, 25546, 33600, 25547, 25549, 33608, 25555, 25557, 33606, 25549, 25555, 33628, 25575, 25577, 33636, 25583, 25585, 33634, 25577, 25583, 25546, 33569, 25523, 33587, 25561, 25539, 33626, 25557, 25575, 25518, 33517, 25471, 33580, 25483, 33533, 25276, 33325, 25275, 33325, 25278, 25275, 33335, 25286, 25284, 33343, 25294, 25292, 33337, 25292, 25286, 33329, 25284, 25278, 33401, 25352, 25350, 33409, 25360, 25358, 33403, 25358, 25352, 33417, 25368, 25366, 33425, 25376, 25374, 33419, 25374, 25368, 33411, 25366, 25360, 25306, 33351, 25305, 25310, 33355, 25309, 25309, 33352, 25306, 33369, 25314, 25316, 33391, 25336, 25338, 33397, 25342, 25344, 33395, 25338, 25342, 25314, 33356, 25310, 33389, 25316, 25336, 33451, 25396, 25398, 33457, 25402, 25404, 33455, 25398, 25402, 33479, 25424, 25426, 33485, 25430, 25432, 33483, 25426, 25430, 33477, 25404, 25424, 25305, 33326, 25276,
33345, 25350, 25294, 33449, 25344, 25396, 33642, 25593, 25591, 33650, 25601, 25599, 33644, 25599, 25593, 33658, 25609, 25607, 33666, 25617, 25615, 33660, 25615, 25609, 33652, 25607, 25601, 33730, 25681, 25679, 33738, 25689, 25687, 33732, 25687, 25681, 33749, 25698, 33748, 33746, 25698, 25695, 33756, 25705, 25703, 33749, 25703, 25699, 33740, 25695, 25689, 33692, 25637, 25639, 33698, 25643, 25645, 33696, 25639, 25643, 33720, 25665, 25667, 33726, 25671, 25673, 33724, 25667, 25671, 33718, 25645, 25665, 33780, 25725, 25727, 33786, 25731, 25733, 33784, 25727, 25731, 33808, 25753, 25755, 25764, 33812, 25763, 33816, 25764, 25766, 25768, 33816, 25766, 33812, 25755, 25763, 33806, 25733, 25753, 33668, 25679, 25617, 33778, 25673, 25725, 33427, 25591, 25376, 33690, 25432, 25637, 25779, 33828, 25778, 33828, 25781, 25778, 33825, 25779, 25771, 33873, 25826, 25824, 33882, 25833, 33881, 33882, 25830, 25834, 25829, 33873, 25824, 33893, 25838, 33891, 33875, 25838, 25826, 33893, 25833, 25840, 25792, 33838, 25791, 33838, 25794, 25791, 33858, 25811, 33857, 25800, 33846, 25799, 33846, 25796, 25799, 33863, 25814, 25816, 33861, 25800, 25814, 33904, 25855, 33903, 33901, 25855, 25852, 33901, 25850, 33898, 33904, 25859, 25856, 25794, 33825, 25771, 33897, 25830, 33878,
33832, 25829, 25781, 33897, 25811, 25849, 33858, 25850, 25812, 25875, 33922, 25874, 33916, 25865, 25863, 25879, 33926, 25878, 33926, 25881, 25878, 25881, 33923, 25875, 33918, 25879, 25865, 33969, 25917, 33968, 33957, 25910, 33956, 33972, 25918, 33969, 33972, 25910, 25920, 33940, 25887, 25889, 33949, 25898, 33948, 33949, 25895, 25899, 33946, 25889, 25895, 33979, 25924, 25926, 33979, 25911, 33957, 33977, 25898, 25924, 25874, 33863, 25816, 33968, 25859, 33913, 33986, 25941, 25939, 33999, 25951, 33998, 33999, 25954, 25952, 34004, 25954, 34002, 33988, 25951, 25941, 34036, 25984, 34035, 34036, 25987, 25985, 25968, 34013, 25967, 34017, 25968, 25970, 34045, 25998, 34044, 34059, 26007, 34058, 34058, 26004, 34056, 34045, 26004, 25999, 34044, 25987, 34039, 34004, 25984, 25956, 33986, 25929, 33983, 33554, 25792, 25503, 33844, 25507, 25796, 33622, 25863, 25569, 33938, 25585, 25887, 33758, 25929, 25705, 34013, 25768, 25967, 31703, 23885, 23657, 5433, 16226, 5432, 5447, 16234, 5446, 5446, 16232, 5443, 16232, 5433, 5443, 5457, 16246, 5456, 16250, 5457, 5459, 5463, 16252, 5462, 5462, 16250, 5459, 5456, 16235, 5447, 16345, 5556, 16344, 16359, 5565, 16358, 16358, 5562, 16356, 16345, 5562, 5557, 16376, 5587, 16375, 16373, 5587, 5584,
16380, 5591, 16379, 16379, 5588, 16376, 16373, 5566, 16359, 16380, 5595, 5592, 5510, 16297, 5509, 5514, 16301, 5513, 16305, 5514, 5516, 5513, 16298, 5510, 5531, 16323, 5530, 5545, 16332, 5544, 5533, 16324, 5531, 16332, 5533, 5544, 5530, 16305, 5516, 16297, 5498, 5509, 16427, 5636, 16426, 16433, 5642, 16432, 16430, 5642, 5639, 16430, 5637, 16427, 16453, 5662, 16452, 16459, 5668, 16458, 16456, 5668, 5665, 16456, 5663, 16453, 16442, 5662, 5653, 16442, 5643, 16433, 5498, 16253, 5463, 16426, 5595, 16391, 5833, 16626, 5832, 5841, 16632, 5840, 5845, 16636, 5844, 5844, 16633, 5841, 5840, 16627, 5833, 5861, 16649, 5860, 16647, 5861, 5857, 5863, 16647, 5857, 16649, 5873, 5860, 5932, 16720, 5931, 5938, 16726, 5937, 16726, 5934, 5937, 16739, 5948, 5950, 5934, 16721, 5932, 16737, 5938, 5948, 16772, 5975, 5977, 16778, 5981, 5983, 16776, 5977, 5981, 5953, 16739, 5950, 16778, 5971, 16760, 16770, 5953, 5975, 5848, 16637, 5845, 16626, 5863, 5832, 16720, 5848, 5931, 5679, 16470, 5678, 16470, 5675, 5678, 5681, 16471, 5679, 5693, 16480, 5692, 16538, 5749, 16537, 16538, 5746, 5750, 5710, 16502, 5709, 16502, 5706, 5709, 5724, 16511, 5723, 16515, 5724, 5726,
5723, 16509, 5720, 16509, 5710, 5720, 16558, 5769, 16557, 5774, 16561, 5773, 16558, 5774, 5770, 16566, 5777, 16565, 5782, 16569, 5781, 16569, 5777, 5781, 5706, 16481, 5693, 16557, 5746, 16535, 5773, 16538, 5750, 5796, 16584, 5795, 16604, 5815, 16603, 16601, 5815, 5812, 5802, 16590, 5801, 16590, 5798, 5801, 5798, 16585, 5796, 16607, 5802, 5818, 5827, 16619, 5826, 16619, 5823, 5826, 5823, 16607, 5818, 16584, 5785, 5795, 5785, 16515, 5726, 16601, 5778, 16566, 5815, 16570, 5782, 16480, 5681, 5692, 16783, 5992, 16782, 5997, 16786, 5996, 16786, 6000, 5996, 16791, 5997, 6001, 16800, 6009, 6007, 16786, 5993, 16783, 16790, 6007, 6000, 16829, 6035, 16828, 16829, 6038, 6036, 16808, 6021, 6019, 16850, 6057, 16849, 16854, 6061, 16853, 16853, 6058, 16850, 16837, 6057, 6048, 16854, 6050, 6062, 16808, 5997, 16787, 6021, 16791, 6001, 16837, 6038, 16832, 16802, 6035, 6009, 6074, 16861, 6073, 16865, 6074, 6076, 16881, 6090, 6092, 16867, 6080, 6078, 16883, 6096, 6094, 16867, 6074, 16862, 16879, 6076, 6090, 16869, 6094, 6080, 16923, 6134, 16922, 16938, 6148, 6146, 6137, 16920, 6131, 16938, 6144, 16936, 16923, 6144, 6135, 6111, 16899, 6110, 16899, 6107, 6110,
6107, 16895, 6105, 16910, 6111, 6121, 16920, 6092, 6131, 16885, 6137, 6096, 16964, 6170, 16963, 16964, 6173, 6171, 16983, 6193, 16982, 16983, 6196, 6194, 16972, 6193, 6183, 17000, 6208, 16999, 16997, 6208, 6205, 17008, 6215, 6213, 16997, 6203, 16995, 17000, 6213, 6209, 16972, 6173, 16967, 16986, 6203, 6196, 16922, 6050, 16839, 16940, 6170, 6148, 5675, 16333, 5545, 16533, 5669, 16459, 16533, 5992, 5744, 16786, 5749, 5996, 16538, 5997, 5750, 16861, 5815, 6073, 16604, 6074, 5816, 16895, 5827, 6105, 6224, 17016, 6223, 17016, 6220, 6223, 17025, 6231, 6233, 6247, 17034, 6246, 17038, 6247, 6249, 6235, 17025, 6233, 17034, 6235, 6246, 17023, 6224, 6231, 17056, 6265, 6267, 17067, 6278, 17066, 6283, 17070, 6282, 17070, 6278, 6282, 17067, 6283, 6279, 17067, 6267, 6279, 6296, 17084, 6295, 17082, 6296, 6292, 17108, 6319, 17107, 17108, 6316, 6320, 6300, 17088, 6299, 17093, 6303, 17092, 6324, 17111, 6323, 17111, 6326, 6323, 6299, 17082, 6292, 17111, 6316, 17104, 6326, 17108, 6320, 17084, 6315, 6295, 17111, 6303, 6323, 17093, 6324, 6304, 6356, 17143, 6355, 6315, 17067, 6279, 17143, 6319, 6355, 17108, 6356, 6320, 6255, 17048, 6254, 6380, 17167, 6379,
6379, 17164, 6376, 6397, 17189, 6396, 17189, 6393, 6396, 17200, 6407, 6409, 6407, 17196, 6405, 17196, 6397, 6405, 6438, 17226, 6437, 17226, 6434, 6437, 17240, 6451, 17239, 17240, 6448, 6452, 17237, 6438, 6448, 6393, 17168, 6380, 17224, 6409, 6434, 6375, 17089, 6300, 17054, 6249, 6265, 17048, 6375, 6254, 17164, 6255, 6376, 6220, 16910, 6121, 17133, 6215, 17010, 17251, 6464, 6462, 17264, 6473, 17263, 17270, 6479, 17269, 17267, 6479, 6476, 17267, 6474, 17264, 17253, 6473, 6464, 17303, 6510, 17302, 17307, 6514, 17306, 17306, 6511, 17303, 17287, 6480, 17270, 17287, 6510, 6498, 17307, 6500, 6515, 6551, 17343, 6550, 17343, 6547, 6550, 17354, 6561, 6563, 6561, 17350, 6559, 17350, 6551, 6559, 6575, 17362, 6574, 17366, 6575, 6577, 17376, 6577, 6587, 6596, 17388, 6595, 17388, 6592, 6595, 6592, 17376, 6587, 17362, 6563, 6574, 17331, 6539, 17330, 17331, 6536, 6540, 17317, 6528, 17316, 17317, 6525, 6529, 17334, 6536, 17328, 17334, 6528, 6542, 17330, 6500, 17289, 17341, 6503, 6547, 17490, 6703, 6701, 17503, 6713, 17502, 17503, 6716, 6714, 17508, 6716, 17506, 17492, 6713, 6703, 17468, 6674, 6676, 6691, 17479, 6690, 17479, 6687, 6690, 6678, 17468, 6676,
17477, 6678, 6687, 17519, 6728, 6730, 17517, 6691, 6728, 17552, 6759, 17551, 17549, 6759, 6756, 17552, 6766, 6760, 6794, 17585, 6793, 17589, 6794, 6796, 17599, 6806, 6808, 6806, 17595, 6804, 17595, 6796, 6804, 17573, 6786, 6784, 17608, 6818, 17607, 17608, 6821, 6819, 17625, 6833, 17624, 17622, 6833, 6830, 17622, 6828, 17620, 17611, 6828, 6821, 17573, 6766, 17561, 17585, 6764, 6793, 17575, 6818, 6786, 6750, 17542, 6749, 17542, 6746, 6749, 17540, 6741, 6746, 17635, 6841, 6843, 6857, 17644, 6856, 17648, 6857, 6859, 6845, 17635, 6843, 17644, 6845, 6856, 17660, 6869, 6871, 6891, 17683, 6890, 17683, 6887, 6890, 17681, 6882, 6887, 6882, 17660, 6871, 17658, 6859, 6869, 6741, 17519, 6730, 17508, 6756, 6718, 17633, 6750, 6841, 6617, 17404, 6616, 17404, 6619, 6616, 6605, 17397, 6604, 6604, 17395, 6601, 17395, 6617, 6601, 17410, 6605, 6621, 17429, 6640, 17428, 17426, 6640, 6637, 17433, 6644, 17432, 17432, 6641, 17429, 17422, 6625, 6627, 17446, 6649, 6651, 6625, 17410, 6621, 17446, 6645, 17433, 17408, 6637, 6619, 17444, 6627, 6649, 17758, 6969, 6967, 6957, 17747, 6956, 17747, 6959, 6956, 6959, 17745, 6953, 17745, 6967, 6953, 6938, 17726, 6937,
17726, 6940, 6937, 6946, 17734, 6945, 17734, 6942, 6945, 6942, 17727, 6938, 6940, 17723, 6927, 17723, 6957, 6927, 6902, 17692, 6901, 17692, 6904, 6901, 6904, 17690, 6898, 17703, 6914, 6912, 17713, 6924, 6922, 17705, 6922, 6914, 17690, 6912, 6898, 7015, 17805, 7014, 17809, 7015, 7017, 17820, 7027, 7029, 7019, 17809, 7017, 17818, 7019, 7027, 17794, 7002, 17793, 17789, 7002, 6998, 17794, 6999, 7003, 17789, 6995, 17787, 17794, 7006, 7003, 17805, 6988, 7014, 6988, 17735, 6946, 17715, 6995, 6924, 17466, 6596, 6674, 7617, 18404, 7616, 18408, 7617, 7619, 7645, 18437, 7644, 18437, 7641, 7644, 18428, 7636, 7634, 18435, 7632, 7641, 18462, 7668, 7670, 7685, 18473, 7684, 18473, 7681, 7684, 7672, 18462, 7670, 18471, 7672, 7681, 7632, 18408, 7619, 18460, 7645, 7668, 18575, 7788, 7786, 18588, 7798, 18587, 18588, 7801, 7799, 18593, 7801, 18591, 18577, 7798, 7788, 18627, 7834, 18626, 18624, 7834, 7831, 18634, 7841, 7839, 18627, 7839, 7835, 18604, 7813, 7815, 18614, 7815, 7825, 18648, 7841, 18636, 18593, 7831, 7803, 18497, 7708, 18496, 18494, 7708, 7705, 18484, 7705, 7695, 18529, 7740, 7738, 18541, 7750, 7748, 18533, 7740, 18531, 18533, 7748, 7742,
18510, 7716, 7718, 18521, 7728, 7730, 7720, 18510, 7718, 18519, 7720, 7728, 7762, 18549, 7761, 18553, 7762, 7764, 18565, 7774, 7776, 18563, 7764, 7774, 18497, 7738, 7709, 18549, 7730, 7761, 18669, 7873, 7875, 18659, 7865, 18658, 18659, 7868, 7866, 18662, 7877, 7868, 7905, 18696, 7904, 18700, 7905, 7907, 18708, 7914, 7916, 7918, 18708, 7916, 18706, 7907, 7914, 18684, 7897, 7895, 18720, 7931, 18719, 18717, 7931, 7928, 18731, 7942, 7940, 18720, 7940, 7932, 18684, 7877, 18671, 18696, 7875, 7904, 18686, 7928, 7897, 18484, 7636, 18430, 7873, 18614, 7825, 18602, 7685, 7813, 18543, 7865, 7750, 18933, 8144, 18932, 18930, 8144, 8141, 18944, 8155, 8153, 18933, 8153, 8145, 18956, 8166, 8164, 18968, 8175, 18967, 18965, 8175, 8172, 18956, 8162, 18954, 18958, 8172, 8166, 18946, 8162, 8155, 8273, 19060, 8272, 8277, 19064, 8276, 8276, 19061, 8273, 19057, 8277, 8261, 19110, 8317, 19109, 19114, 8321, 19113, 19113, 8318, 19110, 19114, 8310, 8322, 8286, 19076, 8285, 8292, 19082, 8291, 19080, 8292, 8288, 8291, 19077, 8286, 19122, 8333, 19121, 19126, 8337, 19125, 19125, 8334, 19122, 8285, 19057, 8261, 19121, 8310, 19099, 18977, 8184, 8186, 18987, 8194, 8196,
18985, 8186, 8194, 8228, 19015, 8227, 19019, 8228, 8230, 19031, 8240, 8242, 19029, 8230, 8240, 18995, 8206, 8204, 19007, 8216, 8214, 18999, 8206, 18997, 18999, 8214, 8208, 19041, 8250, 8248, 19049, 8258, 8256, 19043, 8256, 8250, 19015, 8196, 8227, 19009, 8248, 8216, 19139, 8348, 8350, 19151, 8360, 8362, 19149, 8350, 8360, 19191, 8404, 8402, 19181, 8390, 8392, 19197, 8406, 8408, 19177, 8402, 8388, 19195, 8392, 8406, 19161, 8370, 8368, 19169, 8378, 8376, 19163, 8376, 8370, 19207, 8416, 8414, 19217, 8427, 8425, 8427, 19215, 8423, 19209, 8425, 8416, 19179, 8362, 8390, 19171, 8414, 8378, 19177, 8338, 19126, 18968, 8288, 8176, 19137, 8242, 8348, 19051, 8368, 8258, 18741, 7952, 7950, 18754, 7960, 18753, 18754, 7963, 7961, 18745, 7952, 18743, 18745, 7960, 7954, 18771, 7982, 18770, 18775, 7986, 18774, 18774, 7983, 18771, 18770, 7963, 18757, 18821, 8024, 8026, 8035, 18827, 8034, 18827, 8031, 8034, 18825, 8026, 8031, 18836, 8041, 8043, 18844, 8049, 8051, 18842, 8043, 8049, 18834, 8035, 8041, 18802, 8008, 18801, 18805, 8008, 8011, 18793, 8000, 8002, 18793, 7998, 18789, 18802, 7998, 8009, 8020, 18812, 8019, 18810, 8020, 8016, 8107, 18894, 8106,
18894, 8109, 8106, 8093, 18883, 8092, 18883, 8095, 8092, 8095, 18881, 8089, 18881, 8107, 8089, 8080, 18867, 8079, 18867, 8082, 8079, 18858, 8068, 8066, 8068, 18856, 8064, 18856, 8080, 8064, 18919, 8130, 18918, 18919, 8127, 8131, 8082, 18853, 8057, 18853, 8093, 8057, 18805, 7987, 18775, 8109, 18850, 8054, 18819, 7990, 8024, 18812, 8066, 8019, 19228, 8433, 8435, 8450, 19238, 8449, 19238, 8446, 8449, 8446, 19234, 8444, 19234, 8435, 8444, 8475, 19262, 8474, 19262, 8477, 8474, 8461, 19251, 8460, 19251, 8463, 8460, 8463, 19249, 8457, 19249, 8475, 8457, 19301, 8511, 19300, 19305, 8515, 19304, 19308, 8515, 8518, 19308, 8512, 19301, 8477, 19239, 8450, 19388, 8593, 8595, 19374, 8584, 19373, 19377, 8584, 8587, 19377, 8581, 19371, 19386, 8581, 8593, 19399, 8604, 8606, 19406, 8612, 19405, 19409, 8612, 8615, 19406, 8606, 8613, 19394, 8585, 19374, 19397, 8595, 8604, 19328, 8539, 19327, 19328, 8536, 8540, 19358, 8566, 19357, 19358, 8563, 8567, 19361, 8563, 19355, 19368, 8566, 8572, 19342, 8550, 19341, 19342, 8547, 8551, 19345, 8547, 19339, 19352, 8550, 8556, 19352, 8536, 19325, 19361, 8539, 8569, 19368, 8516, 19305, 19394, 8511, 8598, 7990, 18565, 7776,
18648, 8141, 7859, 18975, 7918, 8184, 18733, 8204, 7942, 19226, 8051, 8433, 18850, 8461, 8054, 19345, 8130, 8553, 17829, 7039, 17828, 17829, 7042, 7040, 17835, 7045, 17834, 7050, 17838, 7049, 17838, 7045, 7049, 17835, 7050, 7046, 17851, 7061, 7059, 17851, 7057, 17849, 17832, 7057, 7042, 17883, 7090, 17882, 17880, 7090, 7087, 17890, 7097, 7095, 17883, 7095, 7091, 17870, 7079, 17868, 7109, 17896, 7107, 17853, 7087, 7061, 17896, 7079, 7107, 17870, 7109, 7081, 17984, 7197, 7195, 17996, 7209, 7207, 17986, 7207, 7197, 18011, 7222, 18010, 18008, 7222, 7219, 18022, 7233, 7231, 18011, 7231, 7223, 17984, 7185, 17981, 17998, 7219, 7209, 17957, 7164, 7166, 17960, 7169, 17959, 17973, 7180, 7182, 17959, 7164, 17955, 7180, 17969, 7178, 17969, 7166, 7178, 17947, 7153, 17946, 17950, 7153, 7156, 17938, 7145, 7147, 17938, 7143, 17934, 17947, 7143, 7154, 17950, 7135, 17924, 17960, 7135, 7170, 18034, 7241, 7243, 18044, 7251, 7253, 18042, 7243, 7251, 7266, 18054, 7265, 18054, 7262, 7265, 18067, 7276, 7278, 18065, 7266, 7276, 7286, 18077, 7285, 18089, 7302, 7300, 7290, 18081, 7289, 7289, 18078, 7286, 18077, 7300, 7285, 18052, 7253, 7262, 17892, 7185, 7097,
18032, 7182, 7241, 7318, 18110, 7317, 7334, 18121, 7333, 7320, 18111, 7318, 18110, 7330, 7317, 18121, 7320, 7333, 18133, 7344, 7342, 7360, 18147, 7359, 18147, 7355, 7359, 18144, 7360, 7356, 18151, 7360, 18148, 18135, 7356, 7344, 18179, 7388, 7390, 18195, 7404, 7406, 18181, 7394, 7392, 18197, 7410, 7408, 7394, 18179, 7390, 18197, 7404, 18193, 7355, 18122, 7334, 18177, 7360, 7388, 18151, 7392, 7362, 18275, 7488, 7486, 18290, 7497, 18289, 18294, 7501, 18293, 18293, 7498, 18290, 18294, 7490, 7502, 18331, 7538, 7536, 7547, 18339, 7546, 18337, 7547, 7543, 18333, 7543, 7538, 18304, 7515, 18303, 7530, 18317, 7529, 18317, 7532, 7529, 7518, 18301, 7512, 18304, 7530, 7516, 18303, 7490, 18279, 7557, 18331, 7536, 18321, 7557, 7532, 18211, 7420, 7422, 18223, 7432, 7434, 18221, 7422, 7432, 18253, 7463, 18252, 18256, 7463, 7466, 18267, 7474, 7476, 18265, 7466, 7474, 18234, 7444, 18233, 18234, 7447, 7445, 18242, 7452, 18241, 18242, 7449, 7453, 18241, 7447, 18237, 18253, 7434, 7464, 18378, 7583, 7585, 18368, 7575, 7577, 18368, 7573, 18364, 18376, 7573, 7583, 18386, 7591, 7593, 18393, 7599, 18392, 18396, 7599, 7602, 18393, 7593, 7600, 18384, 7585, 7591,
18301, 7406, 7512, 18199, 7518, 7410, 18366, 7476, 7575, 18106, 7309, 7311, 18104, 7305, 7309, 7305, 18082, 7290, 18401, 7449, 18239, 18024, 7342, 7233, 18209, 7278, 7420, 18091, 7444, 7302, 18401, 7311, 7605, 19422, 8633, 8631, 19414, 8623, 8621, 8635, 19422, 8631, 19416, 8635, 8623, 8671, 19458, 8670, 19458, 8673, 8670, 19462, 8683, 8673, 19447, 8658, 19446, 19447, 8655, 8659, 19444, 8645, 8655, 19497, 8704, 19496, 19497, 8701, 8705, 19486, 8693, 8695, 19486, 8691, 19482, 19494, 8691, 8701, 19424, 8671, 8633, 19484, 8658, 8693, 19521, 8731, 19520, 19524, 8731, 8734, 19521, 8721, 8732, 19542, 8750, 19541, 19542, 8747, 8751, 19545, 8747, 19539, 19552, 8750, 8756, 19535, 8738, 8740, 19535, 8721, 19510, 19545, 8734, 8753, 19506, 8709, 8711, 19506, 8645, 19434, 19533, 8704, 8738, 7330, 17693, 6902, 18275, 6999, 17790, 7488, 17794, 7003, 18289, 7006, 17802, 8683, 18810, 8016, 19552, 8127, 18916, 18404, 6808, 7616, 17625, 7634, 6834, 18508, 6891, 7716, 17760, 7950, 6969, 18791, 7029, 8000, 18339, 8621, 7546, 19504, 7602, 8709, 17314, 5971, 6525, 17133, 7039, 6344, 17936, 6451, 7145, 19555, 8762, 8760, 19561, 8768, 8766, 19557, 8766, 8762,
8896, 19688, 8895, 19686, 8896, 8892, 19695, 8907, 8903, 19700, 8904, 8908, 19688, 8903, 8895, 19708, 8913, 19707, 19699, 8913, 8907, 19708, 8908, 8914, 19567, 8780, 8778, 19585, 8796, 8794, 19574, 8785, 19573, 19574, 8782, 8786, 19591, 8798, 8800, 19569, 8794, 8780, 19589, 8785, 8798, 19644, 8851, 8853, 19634, 8841, 8843, 19646, 8857, 8855, 19634, 8839, 19630, 19642, 8839, 8851, 19599, 8810, 8808, 8825, 19613, 8824, 19611, 8825, 8821, 19619, 8827, 8829, 19601, 8821, 8810, 19617, 8812, 8827, 19656, 8868, 8866, 19677, 8883, 8885, 19661, 8871, 19660, 19664, 8871, 8874, 19679, 8883, 19674, 19658, 8882, 8868, 19679, 8874, 8887, 8882, 19646, 8855, 19587, 8837, 8796, 19632, 8800, 8841, 19613, 8866, 8824, 19661, 8829, 8872, 19714, 8924, 19713, 19717, 8924, 8927, 19735, 8948, 8946, 8931, 19719, 8930, 19719, 8934, 8930, 19745, 8955, 19744, 19745, 8952, 8956, 8951, 19735, 8946, 19734, 8927, 8945, 19723, 8951, 8934, 19741, 8935, 8952, 8974, 19764, 8973, 19769, 8977, 19768, 19764, 8977, 8973, 19769, 8974, 8978, 19772, 8974, 19765, 19756, 8962, 19755, 19759, 8962, 8965, 19759, 8935, 19724, 19738, 8974, 8948, 19772, 8955, 8980, 8837, 19555, 8760,
19734, 8904, 19696, 19563, 8892, 8768, 19714, 8853, 8925, 19648, 8931, 8857, 19756, 8885, 8963, 8994, 19785, 8993, 8998, 19789, 8997, 8997, 19786, 8994, 19834, 9041, 19833, 19838, 9045, 19837, 19837, 9042, 19834, 9014, 19805, 9013, 9018, 19809, 9017, 9017, 19806, 9014, 19853, 9066, 9064, 9013, 19790, 8998, 19853, 9046, 19838, 9074, 19865, 9073, 9078, 19869, 9077, 9077, 19866, 9074, 19914, 9124, 19913, 19917, 9124, 9127, 9094, 19885, 9093, 9098, 19889, 9097, 9097, 19886, 9094, 19935, 9142, 19934, 19939, 9146, 19938, 19938, 9143, 19935, 9093, 19870, 9078, 19934, 9125, 19914, 9073, 19810, 9018, 19781, 8984, 8986, 9155, 19946, 9154, 9159, 19950, 9158, 9158, 19947, 9155, 20001, 9212, 20000, 19998, 9212, 9209, 9175, 19966, 9174, 9179, 19970, 9178, 19980, 9189, 9191, 9178, 19967, 9175, 19978, 9179, 9189, 20014, 9223, 9225, 20026, 9235, 9237, 20024, 9225, 9235, 9174, 19951, 9159, 20012, 9191, 9223, 19781, 8812, 19603, 9154, 19890, 9098, 19998, 9147, 19939, 20036, 9249, 9247, 20048, 9261, 9259, 20038, 9259, 9249, 20060, 9273, 9271, 20072, 9285, 9283, 20062, 9283, 9273, 20050, 9271, 9261, 20124, 9337, 9335, 20136, 9349, 9347, 20126, 9347, 9337,
20148, 9361, 9359, 20160, 9373, 9371, 20150, 9371, 9361, 20138, 9359, 9349, 20086, 9293, 9295, 20096, 9303, 9305, 20094, 9295, 9303, 20106, 9313, 9315, 20116, 9323, 9325, 20114, 9315, 9323, 20104, 9305, 9313, 20174, 9381, 9383, 20184, 9391, 9393, 20182, 9383, 9391, 20194, 9401, 9403, 20204, 9411, 9413, 20202, 9403, 9411, 20192, 9393, 9401, 20074, 9335, 9285, 20172, 9325, 9381, 20375, 9588, 9586, 20387, 9600, 9598, 20377, 9598, 9588, 20400, 9607, 20399, 20404, 9611, 20403, 20403, 9608, 20400, 20389, 9607, 9600, 9649, 20440, 9648, 9653, 20444, 9652, 9652, 20441, 9649, 20489, 9696, 20488, 20493, 9700, 20492, 20492, 9697, 20489, 9669, 20460, 9668, 9673, 20464, 9672, 9672, 20461, 9669, 20509, 9716, 20508, 20513, 9720, 20512, 20512, 9717, 20509, 9668, 20445, 9653, 20508, 9701, 20493, 20413, 9620, 9622, 20423, 9630, 9632, 20421, 9622, 9630, 20432, 9638, 20431, 20435, 9639, 20432, 20435, 9632, 9641, 9729, 20520, 9728, 9733, 20524, 9732, 9732, 20521, 9729, 20569, 9776, 20568, 20573, 9780, 20572, 20572, 9777, 20569, 9749, 20540, 9748, 9753, 20544, 9752, 9752, 20541, 9749, 20589, 9796, 20588, 20593, 9800, 20592, 20592, 9797, 20589, 9748, 20525, 9733,
20588, 9781, 20573, 20431, 9612, 20404, 9728, 20465, 9673, 20568, 9721, 20513, 20214, 9424, 9422, 20223, 9432, 9430, 20214, 9420, 20212, 20216, 9430, 9424, 20255, 9464, 9462, 20263, 9472, 9470, 20257, 9470, 9464, 20233, 9442, 9444, 20245, 9454, 9456, 20243, 9444, 9454, 20273, 9482, 9484, 20285, 9494, 9496, 20283, 9484, 9494, 20225, 9462, 9432, 20271, 9456, 9482, 20295, 9504, 9502, 20303, 9512, 9510, 20297, 9510, 9504, 20335, 9544, 9542, 20343, 9552, 9550, 20337, 9550, 9544, 20313, 9522, 9524, 20325, 9534, 9536, 20323, 9524, 9534, 20353, 9562, 9564, 20365, 9574, 9576, 20363, 9564, 9574, 20305, 9542, 9512, 20351, 9536, 9562, 20265, 9502, 9472, 20311, 9496, 9522, 20600, 9809, 9807, 20608, 9817, 9815, 20602, 9815, 9809, 20643, 9850, 20642, 20640, 9850, 9847, 20618, 9827, 9829, 20630, 9839, 9841, 20628, 9829, 9839, 20651, 9862, 20650, 20657, 9868, 20656, 20657, 9865, 9869, 20656, 9863, 20651, 20650, 9851, 20643, 20610, 9847, 9817, 20654, 9841, 9865, 9877, 20668, 9876, 9881, 20672, 9880, 9880, 20669, 9877, 20721, 9928, 20720, 20725, 9932, 20724, 20724, 9929, 20721, 9897, 20688, 9896, 9901, 20692, 9900, 20702, 9911, 9913, 9900, 20689, 9897,
20700, 9901, 9911, 20745, 9952, 20744, 20734, 9943, 9945, 20749, 9956, 20748, 20748, 9953, 20745, 20749, 9945, 9957, 9896, 20673, 9881, 20744, 9933, 20725, 20732, 9913, 9943, 9876, 20545, 9753, 20720, 9801, 20593, 20162, 9586, 9373, 20411, 9413, 9620, 20345, 9807, 9552, 20616, 9576, 9827, 19855, 9247, 9066, 20084, 9127, 9293, 20001, 9420, 9213, 20231, 9237, 9442, 9965, 20756, 9964, 9969, 20760, 9968, 9968, 20757, 9965, 9977, 20768, 9976, 9981, 20772, 9980, 9980, 20769, 9977, 9976, 20761, 9969, 10021, 20812, 10020, 10025, 20816, 10024, 10024, 20813, 10021, 10033, 20824, 10032, 10037, 20828, 10036, 10036, 20825, 10033, 10032, 20817, 10025, 10020, 20773, 9981, 10077, 20868, 10076, 10081, 20872, 10080, 10080, 20869, 10077, 10089, 20880, 10088, 10093, 20884, 10092, 10092, 20881, 10089, 20910, 10119, 10121, 20922, 10131, 10133, 20920, 10121, 10131, 10088, 20873, 10081, 20908, 10093, 10119, 10076, 20829, 10037, 20933, 10140, 20932, 20937, 10144, 20936, 20936, 10141, 20933, 20945, 10152, 20944, 20949, 10156, 20948, 20948, 10153, 20945, 20944, 10145, 20937, 10193, 20980, 10192, 10197, 20984, 10196, 10196, 20981, 10193, 21038, 10251, 10249, 21028, 10249, 10239, 10217, 21004, 10216, 10221, 21008, 10220,
10220, 21005, 10217, 21066, 10271, 10273, 21053, 10264, 21052, 21052, 10261, 21050, 21064, 10261, 10271, 10216, 20985, 10197, 20957, 10164, 20956, 20961, 10168, 20960, 20960, 10165, 20957, 20969, 10176, 20968, 20973, 10180, 20972, 20972, 10177, 20969, 20968, 10169, 20961, 10285, 21072, 10284, 10289, 21076, 10288, 10288, 21073, 10285, 21121, 10332, 21120, 21125, 10336, 21124, 21124, 10333, 21121, 10309, 21096, 10308, 10313, 21100, 10312, 10312, 21097, 10309, 21137, 10348, 21136, 21141, 10352, 21140, 21140, 10349, 21137, 10308, 21077, 10289, 21136, 10337, 21125, 20956, 10157, 20949, 10284, 21009, 10221, 21120, 10265, 21053, 21276, 10489, 10487, 21288, 10501, 10499, 21278, 10499, 10489, 21316, 10529, 10527, 21328, 10541, 10539, 21318, 10539, 10529, 21302, 10507, 10509, 21310, 10515, 10517, 21308, 10509, 10515, 21342, 10547, 10549, 21350, 10555, 10557, 21348, 10549, 10555, 21290, 10527, 10501, 21340, 10517, 10547, 21356, 10569, 10567, 21368, 10581, 10579, 21358, 10579, 10569, 21396, 10609, 10607, 21408, 10621, 10619, 21398, 10619, 10609, 21382, 10587, 10589, 21390, 10595, 10597, 21388, 10589, 10595, 21422, 10627, 10629, 21430, 10635, 10637, 21428, 10629, 10635, 21370, 10607, 10581, 21420, 10597, 10627, 21330, 10567, 10541, 21380, 10557, 10587,
21153, 10360, 21152, 21157, 10364, 21156, 21156, 10361, 21153, 21165, 10372, 21164, 21169, 10376, 21168, 21168, 10373, 21165, 21164, 10365, 21157, 10389, 21176, 10388, 10393, 21180, 10392, 10392, 21177, 10389, 21231, 10442, 21230, 21244, 10453, 10451, 21234, 10443, 21231, 21234, 10451, 10445, 10413, 21200, 10412, 10417, 21204, 10416, 21208, 10417, 10419, 21220, 10429, 10431, 10416, 21201, 10413, 21218, 10419, 10429, 21254, 10463, 10465, 21266, 10475, 10477, 21264, 10465, 10475, 10412, 21181, 10393, 21252, 10431, 10463, 21436, 10645, 10643, 21444, 10653, 10651, 21438, 10651, 10645, 21476, 10685, 10683, 21484, 10693, 10691, 21478, 10691, 10685, 21454, 10663, 10665, 21466, 10675, 10677, 21464, 10665, 10675, 21494, 10703, 10705, 21506, 10715, 10717, 21504, 10705, 10715, 21446, 10683, 10653, 21492, 10677, 10703, 21516, 10725, 10723, 21524, 10733, 10731, 21518, 10731, 10725, 21556, 10765, 10763, 21564, 10773, 10771, 21558, 10771, 10765, 21534, 10743, 10745, 21546, 10755, 10757, 21544, 10745, 10755, 21574, 10783, 10785, 21586, 10795, 10797, 21584, 10785, 10795, 21526, 10763, 10733, 21572, 10757, 10783, 21486, 10723, 10693, 21532, 10717, 10743, 21152, 10181, 20973, 10388, 21101, 10313, 21230, 10353, 21141, 21040, 10487, 10251, 21300, 10273, 10507,
21246, 10643, 10453, 21452, 10477, 10663, 21169, 10133, 10377, 19193, 8778, 8404, 19571, 8408, 8782, 19215, 8808, 8423, 19779, 8615, 8984, 34065, 26020, 26018, 34077, 26032, 26030, 34067, 26030, 26020, 34109, 26061, 34108, 34109, 26064, 26062, 34121, 26074, 26072, 34125, 26074, 34123, 34112, 26072, 26064, 34091, 26038, 26040, 34099, 26047, 26049, 26051, 34099, 26049, 34097, 26040, 26047, 26088, 34133, 26087, 34137, 26088, 26090, 34079, 26061, 26032, 34133, 26051, 26087, 34196, 26144, 34195, 34196, 26147, 26145, 34215, 26167, 34214, 34215, 26170, 26168, 34221, 26173, 34220, 34220, 26170, 34218, 34204, 26167, 26157, 34204, 26147, 34199, 26112, 34158, 26111, 26118, 34164, 26117, 34164, 26114, 26117, 26114, 34159, 26112, 34158, 26101, 26111, 26134, 34180, 26133, 34178, 26134, 26130, 26138, 34184, 26137, 26137, 34178, 26130, 34175, 26138, 26121, 26133, 34165, 26118, 34270, 26219, 34269, 34274, 26223, 34273, 34273, 26220, 34270, 34258, 26211, 34257, 34282, 26230, 34281, 34285, 26231, 34282, 34281, 26224, 34274, 34239, 26219, 26192, 34285, 26211, 26233, 26101, 34137, 26090, 34239, 26174, 34221, 34125, 26144, 26076, 34301, 26254, 26252, 34292, 26244, 26242, 26244, 34290, 26240, 34290, 26252, 26240, 26291, 34336, 26290, 34336, 26293, 26290,
34313, 26264, 26266, 34324, 26276, 34323, 34327, 26276, 26279, 34324, 26266, 26277, 34373, 26322, 34372, 34373, 26319, 26323, 34362, 26311, 26313, 34362, 26309, 34358, 34370, 26309, 26319, 34303, 26291, 26254, 34360, 26279, 26311, 34397, 26347, 34396, 34397, 26344, 26348, 34383, 26336, 34382, 34383, 26333, 26337, 34400, 26344, 34394, 34400, 26336, 26350, 34407, 26333, 34380, 34407, 26322, 26353, 26293, 34175, 26121, 34396, 26212, 34258, 26365, 34410, 26364, 26369, 34414, 26368, 26368, 34411, 26365, 34459, 26412, 34458, 34463, 26416, 34462, 34462, 26413, 34459, 26389, 34434, 26388, 26393, 34438, 26392, 26392, 34435, 26389, 34484, 26439, 26437, 34474, 26437, 26427, 26388, 34415, 26369, 34474, 26417, 34463, 34582, 26537, 26535, 34594, 26549, 26547, 34584, 26547, 26537, 34606, 26561, 26559, 34618, 26573, 26571, 34608, 26571, 26561, 34596, 26559, 26549, 26451, 34496, 26450, 26455, 34500, 26454, 26454, 34497, 26451, 34545, 26498, 34544, 34558, 26507, 26509, 34556, 26498, 26507, 26475, 34520, 26474, 26479, 34524, 26478, 26478, 34521, 26475, 34567, 26520, 34566, 34571, 26524, 34570, 34570, 26521, 34567, 26474, 34501, 26455, 34566, 26499, 34545, 34632, 26581, 26583, 34642, 26591, 26593, 34640, 26583, 26591, 34652, 26601, 26603, 34662, 26611, 26613,
34660, 26603, 26611, 34650, 26593, 26601, 26450, 34439, 26393, 34486, 26535, 26439, 34630, 26509, 26581, 34850, 26805, 26803, 34862, 26817, 26815, 34852, 26815, 26805, 34874, 26829, 26827, 34886, 26841, 26839, 34876, 26839, 26829, 34864, 26827, 26817, 34938, 26893, 26891, 34951, 26900, 34950, 34955, 26904, 34954, 34954, 26901, 34951, 34940, 26900, 26893, 34900, 26849, 26851, 34910, 26859, 26861, 34908, 26851, 26859, 34920, 26869, 26871, 34930, 26879, 26881, 34928, 26871, 26879, 34918, 26861, 26869, 34973, 26921, 34972, 34964, 26913, 26915, 34976, 26922, 34973, 34976, 26915, 26924, 34972, 26905, 34955, 34888, 26891, 26841, 34962, 26881, 26913, 26625, 34670, 26624, 26629, 34674, 26628, 26628, 34671, 26625, 34725, 26678, 34724, 34738, 26689, 26687, 34728, 26679, 34725, 34728, 26687, 26681, 26649, 34694, 26648, 26653, 34698, 26652, 34702, 26653, 26655, 34714, 26665, 26667, 26652, 34695, 26649, 34712, 26655, 26665, 34748, 26699, 26701, 34760, 26711, 26713, 34758, 26701, 26711, 26648, 34675, 26629, 34746, 26667, 26699, 34770, 26721, 26719, 34778, 26729, 26727, 34772, 26727, 26721, 34810, 26761, 26759, 34818, 26769, 26767, 34812, 26767, 26761, 34788, 26739, 26741, 34800, 26751, 26753, 34798, 26741, 26751, 34828, 26779, 26781, 34840, 26791, 26793,
34838, 26781, 26791, 34780, 26759, 26729, 34826, 26753, 26779, 34740, 26719, 26689, 34786, 26713, 26739, 34981, 26932, 26930, 34989, 26940, 26938, 34983, 26938, 26932, 35021, 26972, 26970, 35029, 26980, 26978, 35023, 26978, 26972, 34999, 26950, 26952, 35011, 26962, 26964, 35009, 26952, 26962, 35039, 26990, 26992, 35051, 27002, 27004, 35049, 26992, 27002, 34991, 26970, 26940, 35037, 26964, 26990, 35061, 27012, 27010, 35069, 27020, 27018, 35063, 27018, 27012, 35102, 27049, 35101, 35079, 27030, 27032, 35091, 27042, 27044, 35089, 27032, 27042, 35108, 27057, 35107, 35112, 27061, 35111, 35111, 27058, 35108, 35107, 27050, 35102, 35071, 27049, 27020, 35112, 27044, 27062, 35031, 27010, 26980, 35077, 27004, 27030, 26624, 34525, 26479, 34724, 26525, 34571, 34620, 26803, 26573, 34898, 26613, 26849, 34820, 26930, 26769, 34997, 26793, 26950, 27074, 35119, 27073, 27078, 35123, 27077, 27077, 35120, 27074, 35168, 27121, 35167, 35172, 27125, 35171, 35171, 27122, 35168, 27098, 35143, 27097, 27102, 35147, 27101, 27101, 35144, 27098, 35184, 27137, 35183, 35188, 27141, 35187, 35187, 27138, 35184, 27097, 35124, 27078, 35183, 27126, 35172, 27154, 35199, 27153, 27158, 35203, 27157, 27157, 35200, 27154, 35248, 27201, 35247, 35252, 27205, 35251, 35251, 27202, 35248,
27178, 35223, 27177, 27182, 35227, 27181, 27181, 35224, 27178, 35264, 27217, 35263, 35268, 27221, 35267, 35267, 27218, 35264, 27177, 35204, 27158, 35263, 27206, 35252, 27153, 35148, 27102, 35247, 27142, 35188, 27234, 35279, 27233, 27238, 35283, 27237, 27237, 35280, 27234, 35334, 27287, 35333, 35338, 27291, 35337, 35337, 27288, 35334, 27258, 35303, 27257, 27262, 35307, 27261, 35311, 27262, 27264, 35323, 27274, 27276, 27261, 35304, 27258, 35321, 27264, 27274, 35350, 27303, 35349, 35356, 27309, 35355, 35356, 27306, 27310, 35355, 27304, 35350, 27257, 35284, 27238, 35349, 27292, 35338, 35353, 27276, 27306, 27233, 35228, 27182, 35333, 27222, 35268, 27318, 35367, 27317, 27322, 35371, 27321, 27321, 35368, 27318, 35416, 27368, 35415, 35420, 27372, 35419, 35423, 27372, 27375, 35423, 27369, 35416, 27338, 35387, 27337, 27342, 35391, 27341, 27341, 35388, 27338, 35441, 27390, 35440, 35445, 27394, 35444, 35444, 27391, 35441, 27337, 35372, 27322, 35440, 27373, 35420, 27403, 35452, 27402, 27407, 35456, 27406, 27406, 35453, 27403, 35501, 27450, 35500, 35505, 27454, 35504, 35504, 27451, 35501, 27423, 35472, 27422, 27427, 35476, 27426, 27426, 35473, 27423, 35521, 27470, 35520, 35525, 27474, 35524, 35524, 27471, 35521, 27422, 35457, 27407, 35520, 27455, 35505,
27402, 35392, 27342, 35500, 27395, 35445, 35532, 27368, 27478, 27576, 35623, 27575, 35627, 27576, 27578, 27582, 35629, 27581, 27581, 35627, 27578, 27590, 35639, 27589, 27594, 35643, 27593, 27593, 35640, 27590, 27589, 35630, 27582, 27602, 35651, 27601, 27606, 35655, 27605, 27605, 35652, 27602, 27614, 35663, 27613, 27618, 35667, 27617, 27617, 35664, 27614, 27613, 35656, 27606, 27601, 35644, 27594, 27486, 35535, 27485, 27490, 35539, 27489, 27489, 35536, 27486, 35588, 27537, 35587, 35592, 27541, 35591, 35591, 27538, 35588, 27506, 35555, 27505, 27510, 35559, 27509, 35569, 27520, 27522, 27509, 35556, 27506, 35567, 27510, 27520, 35612, 27561, 35611, 35601, 27552, 27554, 35616, 27565, 35615, 35615, 27562, 35612, 35616, 27554, 27566, 27505, 35540, 27490, 35611, 27542, 35592, 35599, 27522, 27552, 27626, 35675, 27625, 27630, 35679, 27629, 27629, 35676, 27626, 27638, 35687, 27637, 27642, 35691, 27641, 27641, 35688, 27638, 27637, 35680, 27630, 27485, 35477, 27427, 35587, 27475, 35525, 27625, 35668, 27618, 35732, 27682, 35731, 35736, 27686, 35735, 35736, 27689, 27687, 35735, 27683, 35732, 35747, 27700, 35746, 35751, 27704, 35750, 35750, 27701, 35747, 35746, 27689, 35739, 27777, 35826, 27776, 27781, 35830, 27780, 27780, 35827, 27777, 27797, 35846, 27796,
27801, 35850, 27800, 27800, 35847, 27797, 27796, 35831, 27781, 35795, 27748, 35794, 35799, 27752, 35798, 35798, 27749, 35795, 35811, 27764, 35810, 35815, 27768, 35814, 35814, 27765, 35811, 35810, 27753, 35799, 27817, 35866, 27816, 27821, 35870, 27820, 27820, 35867, 27817, 27837, 35886, 27836, 27841, 35890, 27840, 27840, 35887, 27837, 27836, 35871, 27821, 35794, 27705, 35751, 27816, 35851, 27801, 36033, 27982, 36032, 36037, 27986, 36036, 36036, 27983, 36033, 36053, 28002, 36052, 36057, 28006, 36056, 36056, 28003, 36053, 36052, 27987, 36037, 28055, 36104, 28054, 28059, 36108, 28058, 28058, 36105, 28055, 28067, 36116, 28066, 28071, 36120, 28070, 28070, 36117, 28067, 28066, 36109, 28059, 36073, 28022, 36072, 36077, 28026, 36076, 36076, 28023, 36073, 36093, 28042, 36092, 36097, 28046, 36096, 36096, 28043, 36093, 36092, 28027, 36077, 28111, 36160, 28110, 28115, 36164, 28114, 28114, 36161, 28111, 28123, 36172, 28122, 28127, 36176, 28126, 28126, 36173, 28123, 28122, 36165, 28115, 36072, 28007, 36057, 28110, 36121, 28071, 35947, 27900, 35946, 35951, 27904, 35950, 35950, 27901, 35947, 35924, 27875, 27877, 35936, 27887, 27889, 35934, 27877, 27887, 35963, 27916, 35962, 35969, 27922, 35968, 35969, 27919, 27923, 35968, 27917, 35963, 35962, 27905, 35951,
35966, 27889, 27919, 27931, 35980, 27930, 27935, 35984, 27934, 27934, 35981, 27931, 27951, 36000, 27950, 27955, 36004, 27954, 36014, 27965, 27967, 27954, 36001, 27951, 36012, 27955, 27965, 27950, 35985, 27935, 36225, 28174, 36224, 36229, 28178, 36228, 36228, 28175, 36225, 36249, 28198, 36248, 36238, 28189, 28191, 36253, 28202, 36252, 36252, 28199, 36249, 36253, 28191, 28203, 36248, 28179, 36229, 28211, 36260, 28210, 28215, 36264, 28214, 28214, 36261, 28211, 28223, 36272, 28222, 28227, 36276, 28226, 28226, 36273, 28223, 36302, 28253, 28255, 36314, 28265, 28267, 36312, 28255, 28265, 28222, 36265, 28215, 36300, 28227, 28253, 35946, 27769, 35815, 27930, 35891, 27841, 36224, 28047, 36097, 28210, 36177, 28127, 36236, 27967, 28189, 36325, 28278, 36324, 36329, 28282, 36328, 36328, 28279, 36325, 36341, 28294, 36340, 36345, 28298, 36344, 36344, 28295, 36341, 36340, 28283, 36329, 36357, 28310, 36356, 36361, 28314, 36360, 36360, 28311, 36357, 36373, 28326, 36372, 36377, 28330, 36376, 36376, 28327, 36373, 36372, 28315, 36361, 36356, 28299, 36345, 28506, 36557, 28505, 28542, 36587, 28541, 36591, 28542, 28544, 28541, 36585, 28538, 36630, 28583, 36629, 36644, 28592, 36643, 36643, 28589, 36641, 36630, 28589, 28584, 28560, 36609, 28559, 28564, 36613, 28563,
28563, 36610, 28560, 36659, 28608, 36658, 36663, 28612, 36662, 36662, 28609, 36659, 28559, 36591, 28544, 36658, 28593, 36644, 28522, 36569, 28521, 36563, 28512, 28510, 28526, 36573, 28525, 36573, 28528, 28525, 28528, 36570, 28522, 36565, 28526, 28512, 28623, 36670, 28622, 28627, 36674, 28626, 36674, 28629, 28626, 28629, 36671, 28623, 36721, 28670, 36720, 36725, 28674, 36724, 36724, 28671, 36721, 36733, 28686, 36732, 36739, 28692, 36738, 36739, 28689, 28693, 36738, 28687, 36733, 36750, 28686, 28696, 28647, 36675, 28627, 36750, 28675, 36725, 28622, 36614, 28564, 36720, 28613, 36663, 36585, 28506, 28538, 36569, 28647, 28521, 36389, 28342, 36388, 36402, 28353, 28351, 36392, 28343, 36389, 36392, 28351, 28345, 36434, 28385, 28383, 36442, 28393, 28391, 36436, 28391, 28385, 36412, 28363, 28365, 36424, 28375, 28377, 36422, 28365, 28375, 36452, 28403, 28405, 36464, 28415, 28417, 36462, 28405, 28415, 36404, 28383, 28353, 36450, 28377, 28403, 36474, 28425, 28423, 36482, 28433, 28431, 36476, 28431, 28425, 36514, 28465, 28463, 36524, 28476, 28474, 28476, 36522, 28472, 36516, 28474, 28465, 36492, 28443, 28445, 36504, 28455, 28457, 36502, 28445, 28455, 36535, 28486, 28488, 36547, 28498, 28500, 36545, 28488, 28498, 36484, 28463, 28433, 36533, 28457, 28486,
36444, 28423, 28393, 36490, 28417, 28443, 36753, 28706, 28704, 28719, 36765, 28718, 36763, 28719, 28715, 36755, 28715, 28706, 36778, 28729, 28731, 36789, 28741, 36788, 36792, 28741, 28744, 36789, 28731, 28742, 36834, 28783, 36833, 36834, 28780, 28784, 36823, 28772, 28774, 36823, 28770, 36819, 36831, 28770, 28780, 36765, 28754, 28718, 36821, 28744, 28772, 36852, 28804, 36851, 36856, 28808, 36855, 36859, 28808, 28811, 36859, 28805, 36852, 36856, 28794, 28809, 36870, 28815, 28817, 36870, 28794, 36841, 36874, 28804, 28820, 36868, 28783, 28815, 36388, 28331, 36377, 28754, 36563, 28510, 36874, 28689, 36736, 36522, 28704, 28472, 36776, 28500, 28729, 36410, 28267, 28363, 28829, 36879, 28828, 36877, 28829, 28825, 36886, 28837, 28835, 36879, 28835, 28828, 36968, 28921, 28919, 36980, 28933, 28931, 28923, 36968, 28919, 36970, 28931, 28921, 36990, 28943, 28941, 37000, 28953, 28951, 36992, 28951, 28943, 36982, 28941, 28933, 28847, 36894, 28846, 28851, 36898, 28850, 36898, 28853, 28850, 28853, 36895, 28847, 28861, 36910, 28860, 28865, 36914, 28864, 28864, 36911, 28861, 36910, 28875, 28860, 36924, 28865, 28877, 28875, 36899, 28851, 36894, 28887, 28846, 37022, 28973, 28975, 37020, 28963, 28973, 37048, 29000, 37047, 37051, 29000, 29003, 37062, 29011, 29013,
37060, 29003, 29011, 37033, 28983, 37032, 37036, 28983, 28986, 37039, 28989, 37038, 37038, 28984, 37033, 37036, 28963, 37010, 37048, 28975, 29001, 28887, 36877, 28825, 36888, 28923, 28837, 37220, 29173, 29171, 37230, 29183, 29181, 37222, 29181, 29173, 37240, 29193, 29191, 37250, 29203, 29201, 37242, 29201, 29193, 37232, 29191, 29183, 37300, 29253, 29251, 37310, 29263, 29261, 37302, 29261, 29253, 29279, 37325, 29278, 37323, 29279, 29275, 37336, 29291, 29289, 37325, 29289, 29278, 29291, 37320, 29266, 37312, 29275, 29263, 37262, 29211, 29213, 37272, 29221, 29223, 37270, 29213, 29221, 37282, 29231, 29233, 37292, 29241, 29243, 37290, 29233, 29241, 37280, 29223, 29231, 37350, 29299, 29301, 37360, 29309, 29311, 37358, 29301, 29309, 37370, 29319, 29321, 37380, 29329, 29331, 37378, 29321, 29329, 37368, 29311, 29319, 37252, 29251, 29203, 37348, 29243, 29299, 29024, 37070, 29023, 29030, 37076, 29029, 37076, 29026, 29029, 29026, 37071, 29024, 29057, 37106, 29056, 37110, 29057, 29059, 37120, 29069, 29071, 29069, 37116, 29067, 37116, 29059, 29067, 29033, 37077, 29030, 37106, 29033, 29056, 37131, 29084, 37130, 37128, 29084, 29081, 37142, 29095, 29093, 37146, 29095, 37144, 37131, 29093, 29085, 37180, 29131, 29129, 37188, 29139, 29137, 37182, 29137, 29131,
29109, 37154, 29108, 37158, 29109, 29111, 37170, 29121, 29123, 37168, 29111, 29121, 37198, 29149, 29151, 37210, 29161, 29163, 37208, 29151, 29161, 37146, 29129, 29097, 37196, 29123, 29149, 37154, 29071, 29108, 37388, 29339, 29337, 37396, 29347, 29345, 37390, 29345, 29339, 37428, 29379, 29377, 37436, 29387, 29385, 37430, 29385, 29379, 37406, 29357, 29359, 37418, 29369, 29371, 37416, 29359, 29369, 37446, 29397, 29399, 37458, 29409, 29411, 37456, 29399, 29409, 37398, 29377, 29347, 37444, 29371, 29397, 37468, 29419, 29417, 37476, 29427, 29425, 37470, 29425, 29419, 37508, 29459, 29457, 37516, 29467, 29465, 37510, 29465, 29459, 37486, 29437, 29439, 37498, 29449, 29451, 37496, 29439, 29449, 37526, 29477, 29479, 37538, 29489, 29491, 37536, 29479, 29489, 37478, 29457, 29427, 37524, 29451, 29477, 37438, 29417, 29387, 37484, 29411, 29437, 29023, 36924, 28877, 37128, 28990, 37039, 37002, 29171, 28953, 37260, 29013, 29211, 37190, 29337, 29139, 37404, 29163, 29357, 29501, 37548, 29500, 29505, 37552, 29504, 37552, 29507, 29504, 29507, 37549, 29501, 37599, 29548, 37598, 37603, 29552, 37602, 37602, 29549, 37599, 37611, 29563, 37610, 37615, 29567, 37614, 37618, 29567, 29570, 37618, 29564, 37611, 37627, 29563, 29573, 37627, 29553, 37603, 37615, 29533, 29568,
37632, 29580, 29582, 37642, 29590, 37641, 37645, 29590, 29593, 37634, 29580, 37630, 37642, 29582, 29591, 37645, 29533, 37580, 37650, 29601, 29599, 37658, 29609, 29607, 37652, 29607, 29601, 37691, 29638, 37690, 37668, 29619, 29621, 37680, 29631, 29633, 37678, 29621, 29631, 37697, 29646, 37696, 37701, 29650, 37700, 37700, 29647, 37697, 37696, 29639, 37691, 37660, 29638, 29609, 37701, 29633, 29651, 37320, 29505, 29266, 37634, 29331, 29584, 37518, 29599, 29467, 37666, 29491, 29619, 35922, 27642, 27875, 5430, 799, 797, 807, 5434, 805, 789, 5430, 797, 805, 5428, 789, 840, 5464, 838, 848, 5472, 846, 846, 5466, 840, 838, 5436, 807, 922, 5546, 920, 5556, 930, 932, 930, 5548, 922, 920, 5474, 848, 1406, 5866, 5864, 5872, 1409, 1408, 5864, 1408, 1406, 1438, 5892, 5890, 1445, 5898, 1444, 1444, 5892, 1439, 1407, 5890, 5866, 6462, 2175, 2176, 2175, 6453, 2170, 6502, 2230, 2229, 2226, 6502, 2229, 2342, 6543, 2340, 2350, 6552, 2348, 2342, 6552, 6544, 2380, 6564, 2378, 2388, 6578, 2386, 2386, 6566, 2380, 2378, 6554, 2350, 2340, 6501, 2226, 2518, 6652, 2516, 2528, 6662, 6660, 2528, 6654, 2518, 6701, 2563, 2564, 2563, 6692, 2560,
2560, 6662, 2530, 6763, 2663, 2662, 2659, 6763, 2662, 2691, 6789, 6787, 2701, 6797, 2699, 2699, 6789, 2693, 2691, 6761, 2659, 2516, 6580, 2388, 2998, 7606, 2996, 3006, 7620, 3004, 3004, 7608, 2998, 3015, 7646, 3013, 3025, 7656, 7654, 3025, 7648, 3015, 3013, 7622, 3006, 7786, 3035, 3036, 3035, 7777, 3034, 3034, 7656, 3027, 3346, 8262, 3344, 8270, 3354, 3352, 8262, 3352, 3344, 3382, 8301, 3380, 8317, 3392, 3394, 8315, 3386, 3392, 3386, 8303, 3382, 3380, 8264, 3346, 2996, 6799, 2701, 2170, 5900, 1445, 8991, 3554, 3552, 3562, 8999, 3560, 8987, 3552, 3546, 3560, 8987, 3546, 3590, 9027, 3588, 9041, 3600, 3602, 9039, 3594, 3600, 3594, 9029, 3590, 3588, 9001, 3562, 9646, 4081, 4079, 4089, 9654, 4087, 9642, 4079, 4073, 4087, 9642, 4073, 4113, 9682, 4111, 9696, 4126, 4128, 9694, 4119, 4126, 4119, 9684, 4113, 4111, 9656, 4089, 9962, 4808, 4806, 9958, 4806, 4800, 4832, 9982, 4830, 4840, 9990, 4838, 4838, 9984, 4832, 4830, 9958, 4800, 10140, 5053, 5055, 10138, 5047, 5053, 5115, 10182, 5113, 10190, 5123, 5121, 5131, 10198, 5129, 10182, 5121, 5113, 5129, 10184, 5115, 5153, 10232, 10230,
10239, 5161, 5163, 10238, 5155, 5161, 5153, 10200, 5131, 5047, 9992, 4840, 16226, 11592, 11590, 16240, 11603, 16238, 16228, 11597, 11592, 16230, 11603, 11597, 16264, 11636, 16262, 16272, 11644, 16270, 16270, 11638, 16264, 16262, 11605, 16240, 16346, 11720, 11722, 16348, 11730, 11728, 16350, 11722, 11730, 16348, 11646, 16272, 16662, 12205, 12204, 16666, 12208, 16664, 12205, 16664, 12208, 16690, 12240, 16688, 16696, 12247, 12246, 16690, 12246, 12241, 16688, 12209, 16666, 17252, 12970, 12971, 17255, 12971, 12974, 17299, 13020, 13019, 17300, 13028, 13020, 17347, 13140, 13138, 17358, 13146, 17356, 17356, 13140, 17348, 17370, 13176, 17368, 17380, 13186, 17378, 17378, 13178, 17370, 17368, 13148, 17358, 17301, 13138, 13028, 17452, 13322, 17450, 17460, 13332, 17458, 17458, 13324, 17452, 17491, 13359, 13360, 17494, 13360, 13362, 17460, 13362, 13334, 17559, 13453, 13452, 17560, 13461, 13453, 17593, 13489, 17591, 17603, 13497, 17601, 17601, 13491, 17593, 17563, 13489, 13461, 17450, 13188, 17380, 18412, 13794, 18410, 18420, 13804, 18418, 18418, 13796, 18412, 18446, 13819, 18444, 18454, 13829, 18452, 18452, 13821, 18446, 18444, 13806, 18420, 18576, 13832, 13833, 18579, 13833, 13834, 18454, 13834, 13831, 19060, 14144, 14142, 19068, 14152, 14150,
14144, 19068, 14150, 19103, 14178, 19101, 19111, 14186, 14188, 14192, 19111, 14188, 19103, 14192, 14180, 19101, 14152, 19070, 18410, 13499, 17603, 16698, 12974, 12247, 19785, 14348, 14346, 19799, 14358, 19797, 19787, 14352, 14348, 19797, 14352, 19793, 19827, 14386, 19825, 19835, 14394, 14396, 19841, 14396, 14400, 19841, 14388, 19827, 19825, 14360, 19799, 20440, 14875, 14873, 20454, 14885, 20452, 20442, 14879, 14875, 20452, 14879, 20448, 20482, 14909, 20480, 20490, 14917, 14919, 14927, 20490, 14919, 20482, 14927, 14911, 20480, 14887, 20454, 20756, 15602, 15600, 20758, 15606, 15602, 20782, 15628, 20780, 20790, 15636, 20788, 20788, 15630, 20782, 20780, 15606, 20764, 20934, 15847, 15849, 20940, 15849, 15853, 20980, 15913, 15911, 20990, 15919, 20988, 20998, 15927, 20996, 20982, 15919, 15913, 20996, 15921, 20990, 21029, 15951, 15953, 21032, 15959, 21030, 21032, 15953, 15961, 21030, 15929, 20998, 20940, 15638, 20790, 22030, 26354, 22028, 26362, 22038, 22036, 22046, 26370, 22044, 26354, 22036, 22028, 22044, 26356, 22030, 22070, 26402, 22068, 26412, 22076, 22078, 26410, 22070, 22076, 22068, 26372, 22046, 22561, 27063, 22559, 27071, 22569, 22567, 22577, 27079, 22575, 27063, 22567, 22559, 22575, 27065, 22561, 22601, 27111, 22599, 27121, 22607, 22609,
27119, 22601, 22607, 22599, 27081, 22577, 27315, 23197, 23196, 23204, 27323, 23202, 23195, 27315, 23196, 23195, 27323, 27311, 23248, 27351, 23246, 23256, 27359, 23254, 23254, 27353, 23248, 23246, 27325, 23204, 27478, 23309, 23310, 23309, 27476, 23306, 23256, 27476, 27361, 23678, 27567, 23676, 27573, 23680, 23679, 23676, 27573, 23679, 23902, 27643, 23900, 23910, 27651, 23908, 23908, 27645, 23902, 27682, 23922, 23924, 23922, 27675, 23920, 23920, 27653, 23910, 27774, 23970, 23968, 23978, 27782, 23976, 27770, 23968, 23962, 23976, 27770, 23962, 24198, 27968, 24196, 27982, 24208, 24210, 27980, 24202, 24208, 24202, 27970, 24198, 28052, 24256, 24254, 28048, 24254, 24248, 24280, 28072, 24278, 24288, 28080, 24286, 24286, 28074, 24280, 24278, 28048, 24248, 24196, 27784, 23978, 24477, 28268, 24475, 28278, 24483, 24485, 28276, 24477, 24483, 28503, 24974, 24973, 24968, 28503, 24973, 24998, 28529, 24996, 25006, 28545, 25004, 25004, 28531, 24998, 25016, 28573, 25015, 28583, 25023, 25024, 25023, 28575, 25016, 25006, 28573, 28547, 24996, 28501, 24968, 24475, 28082, 24288, 29498, 25970, 25969, 25978, 29508, 25976, 29492, 25969, 25964, 25976, 29492, 25964, 25995, 29534, 25993, 29548, 26006, 26008, 26006, 29542, 26002, 26002, 29536, 25995, 25993, 29510, 25978,
23900, 27569, 23678, 34410, 30086, 30084, 34420, 30092, 34418, 34428, 30100, 34426, 34412, 30092, 30086, 34426, 30094, 34420, 34460, 30124, 30126, 34468, 30132, 34466, 34468, 30126, 30134, 34466, 30102, 34428, 35119, 30617, 30615, 35129, 30623, 35127, 35137, 30631, 35135, 35121, 30623, 30617, 35135, 30625, 35129, 35169, 30655, 30657, 35177, 30663, 35175, 35177, 30657, 30665, 35175, 30633, 35137, 35367, 31252, 31251, 35379, 31270, 31268, 35369, 31253, 31252, 35379, 31253, 35375, 35409, 31302, 35407, 35425, 31312, 31310, 35425, 31304, 35409, 35407, 31270, 35381, 35533, 31357, 31358, 35534, 31358, 31368, 35534, 31312, 35427, 35623, 31728, 31727, 35633, 31736, 31734, 35625, 31734, 31728, 35701, 31956, 35699, 35709, 31972, 35707, 35707, 31958, 35701, 35733, 31976, 31978, 35741, 31978, 31980, 35709, 31980, 31974, 35826, 32022, 32020, 35840, 32032, 35838, 35828, 32026, 32022, 35838, 32026, 35834, 36026, 32252, 36024, 36034, 32260, 32262, 36040, 32262, 32266, 36040, 32254, 36026, 36104, 32308, 32306, 36106, 32312, 32308, 36130, 32334, 36128, 36138, 32342, 36136, 36136, 32336, 36130, 36128, 32312, 36112, 36024, 32034, 35840, 36326, 32531, 32533, 36334, 32539, 36332, 36334, 32533, 32541, 36557, 33022, 33021, 36559, 33030, 33022, 36595, 33052, 36593,
36603, 33066, 36601, 36601, 33054, 36595, 36631, 33075, 33076, 36633, 33080, 33079, 36635, 33076, 33080, 36633, 33068, 36603, 36561, 33052, 33030, 36332, 32344, 36138, 37548, 34018, 34017, 37566, 34032, 37564, 34018, 37558, 34026, 37558, 34032, 34026, 37592, 34053, 37590, 37600, 34059, 34061, 37606, 34061, 34064, 37592, 34064, 34055, 37590, 34034, 37566, 35699, 31736, 35635, 5432, 11590, 799, 16344, 932, 11720, 5873, 12204, 1409, 17251, 2176, 12970, 6503, 13019, 2230, 17490, 2564, 13359, 6764, 13452, 2663, 18575, 3036, 13832, 8272, 14142, 3354, 19109, 3394, 14186, 8993, 14346, 3554, 19833, 3602, 14394, 9648, 14873, 4081, 20488, 4128, 14917, 9964, 15600, 4808, 20932, 5055, 15847, 10192, 15911, 5123, 21028, 5163, 15951, 26364, 30084, 22038, 34458, 22078, 30124, 27073, 30615, 22569, 35167, 22609, 30655, 27317, 31251, 23197, 35532, 23310, 31357, 27575, 31727, 23680, 35731, 23924, 31976, 27776, 32020, 23970, 36032, 24210, 32260, 28054, 32306, 24256, 36324, 24485, 32531, 28505, 33021, 24974, 36629, 25024, 33075, 29500, 34017, 25970, 37598, 26008, 34059, 21410, 26018, 10621, 34089, 10637, 26038, 21566, 26242, 10773, 34311, 10797, 26264, 9, 11, 10, 7, 11, 9, 4, 5, 9, 6, 9, 5,
14, 16, 17, 19, 21, 20, 23, 25, 24, 24, 27, 26, 30, 31, 29, 32, 34, 35, 43, 41, 40, 43, 39, 38, 12, 28, 29, 32, 33, 17, 20, 37, 36, 13, 19, 18, 14, 15, 23, 22, 26, 16, 41, 35, 34, 35, 41, 38, 26, 27, 21, 31, 34, 32, 16, 26, 19, 13, 29, 32, 18, 36, 28, 2, 25, 23, 0, 1, 23, 3, 5, 30, 36, 37, 7, 3, 28, 36, 52, 54, 53, 50, 54, 52, 58, 56, 55, 46, 52, 51, 49, 52, 46, 57, 55, 47, 44, 46, 48, 55, 56, 50, 44, 47, 55, 60, 62, 63, 67, 69, 68, 68, 72, 70, 61, 67, 66, 66, 70, 62, 70, 72, 65, 59, 62, 70, 74, 76, 77, 85, 83, 82, 85, 81, 80, 92, 94, 95, 101, 103, 102, 102, 105, 104, 97, 84, 82, 89, 103, 101, 83, 77, 76, 77, 83, 80, 93, 101, 100, 92, 100, 104, 87, 86, 80, 104, 105, 99, 86, 78, 75, 94, 104, 98, 90, 96, 82, 79, 88, 101,
106, 108, 109, 114, 116, 115, 115, 118, 117, 122, 120, 119, 124, 121, 119, 125, 127, 128, 133, 131, 130, 125, 126, 109, 135, 124, 123, 106, 107, 114, 113, 117, 108, 132, 130, 126, 120, 111, 110, 123, 119, 110, 109, 112, 110, 135, 134, 130, 130, 134, 129, 109, 126, 129, 134, 123, 112, 45, 74, 75, 87, 58, 57, 46, 75, 78, 86, 57, 48, 110, 111, 63, 65, 116, 114, 122, 73, 71, 107, 110, 62, 64, 114, 107, 111, 120, 71, 117, 118, 89, 95, 127, 125, 79, 108, 117, 79, 93, 125, 139, 141, 140, 148, 146, 145, 150, 147, 145, 142, 143, 137, 140, 147, 145, 148, 141, 139, 136, 137, 139, 145, 146, 143, 144, 149, 145, 136, 138, 145, 143, 146, 139, 61, 63, 144, 145, 147, 69, 150, 149, 71, 142, 145, 67, 149, 144, 63, 51, 53, 2, 45, 51, 1, 17, 76, 74, 97, 96, 38, 96, 90, 33, 17, 33, 90, 0, 15, 74, 157, 159, 158, 155, 159, 157, 165, 166, 164, 170, 172, 171,
168, 172, 170, 152, 163, 164, 158, 174, 173, 161, 162, 170, 167, 170, 162, 152, 153, 157, 154, 157, 153, 160, 162, 165, 173, 174, 168, 160, 163, 173, 156, 173, 163, 178, 180, 181, 183, 185, 184, 184, 187, 186, 177, 185, 183, 179, 183, 182, 182, 186, 180, 175, 176, 183, 194, 196, 195, 192, 196, 194, 202, 203, 201, 207, 209, 208, 205, 209, 207, 189, 200, 201, 195, 211, 210, 198, 199, 207, 204, 207, 199, 190, 194, 193, 188, 191, 194, 197, 199, 202, 210, 211, 205, 200, 210, 204, 189, 193, 210, 214, 215, 213, 218, 219, 217, 224, 225, 223, 228, 229, 227, 222, 223, 215, 226, 227, 219, 215, 218, 216, 225, 228, 226, 223, 226, 218, 230, 232, 233, 236, 237, 235, 239, 241, 240, 244, 245, 243, 248, 249, 247, 242, 243, 233, 246, 247, 237, 238, 240, 232, 233, 236, 234, 245, 248, 246, 243, 246, 236, 252, 253, 251, 256, 257, 255, 260, 261, 259, 264, 265, 263, 258, 259, 253, 262, 263, 257, 253, 256, 254,
261, 264, 262, 259, 262, 256, 198, 230, 231, 234, 235, 203, 208, 239, 238, 231, 234, 202, 206, 238, 230, 250, 251, 225, 254, 255, 229, 251, 254, 228, 188, 190, 214, 220, 221, 192, 203, 224, 222, 212, 220, 191, 201, 222, 214, 237, 252, 250, 249, 260, 258, 247, 258, 252, 235, 250, 224, 274, 275, 273, 279, 281, 280, 277, 281, 279, 284, 286, 287, 290, 291, 289, 293, 295, 294, 270, 284, 285, 288, 289, 275, 280, 293, 292, 270, 271, 279, 269, 276, 279, 292, 294, 286, 269, 271, 274, 282, 283, 277, 287, 290, 288, 272, 282, 276, 285, 288, 274, 278, 292, 284, 299, 301, 300, 302, 304, 305, 307, 309, 308, 308, 311, 310, 301, 309, 307, 298, 300, 297, 303, 307, 306, 302, 306, 310, 297, 300, 307, 314, 315, 313, 318, 319, 317, 322, 323, 321, 326, 327, 325, 320, 321, 315, 324, 325, 319, 315, 318, 316, 323, 326, 324, 321, 324, 318, 268, 299, 298, 267, 298, 296, 312, 313, 287, 316, 317, 291, 313, 316, 290,
294, 295, 268, 266, 286, 294, 296, 297, 314, 305, 322, 320, 297, 303, 320, 266, 296, 312, 330, 331, 329, 334, 335, 333, 338, 339, 337, 342, 343, 341, 336, 337, 331, 340, 341, 335, 331, 334, 332, 339, 342, 340, 337, 340, 334, 346, 347, 345, 348, 350, 351, 357, 355, 354, 357, 353, 352, 358, 360, 361, 368, 366, 365, 368, 364, 363, 358, 359, 347, 370, 356, 354, 366, 361, 360, 361, 366, 363, 354, 355, 351, 351, 355, 352, 347, 350, 348, 370, 369, 363, 369, 362, 359, 347, 359, 362, 369, 354, 350, 373, 374, 372, 377, 378, 376, 383, 384, 382, 391, 389, 388, 393, 390, 388, 381, 382, 374, 377, 385, 386, 391, 380, 379, 389, 386, 385, 392, 388, 385, 374, 377, 375, 382, 384, 387, 382, 385, 377, 386, 389, 379, 399, 397, 396, 398, 396, 394, 371, 372, 339, 375, 376, 343, 372, 375, 342, 397, 367, 365, 396, 365, 360, 335, 346, 344, 343, 360, 358, 341, 358, 346, 376, 378, 395, 399, 398, 379, 379, 398, 395,
343, 376, 394, 180, 272, 273, 187, 283, 282, 186, 282, 272, 328, 329, 245, 332, 333, 249, 329, 332, 248, 344, 345, 261, 348, 349, 265, 345, 348, 264, 333, 344, 260, 240, 241, 177, 181, 244, 242, 232, 240, 176, 175, 179, 242, 275, 330, 328, 291, 338, 336, 289, 336, 330, 319, 373, 371, 327, 383, 381, 325, 381, 373, 317, 371, 338, 273, 328, 244, 161, 212, 213, 216, 217, 166, 171, 221, 220, 213, 216, 165, 169, 220, 212, 403, 405, 406, 408, 410, 409, 409, 412, 411, 402, 410, 408, 403, 404, 408, 407, 411, 405, 401, 408, 404, 416, 418, 417, 415, 417, 414, 424, 425, 423, 429, 431, 430, 427, 431, 429, 436, 437, 435, 440, 441, 439, 420, 434, 435, 438, 439, 425, 430, 443, 442, 421, 429, 428, 419, 426, 429, 419, 421, 424, 432, 433, 427, 437, 440, 438, 422, 432, 426, 435, 438, 424, 420, 428, 442, 405, 422, 423, 412, 433, 432, 411, 432, 422, 413, 414, 436, 442, 443, 416, 413, 434, 442, 444, 446, 447,
450, 451, 449, 453, 455, 454, 454, 457, 456, 462, 463, 461, 466, 467, 465, 460, 461, 447, 464, 465, 451, 445, 453, 452, 444, 452, 456, 445, 447, 450, 458, 459, 455, 463, 466, 464, 448, 458, 453, 461, 464, 450, 470, 471, 469, 479, 477, 476, 488, 486, 485, 490, 487, 485, 492, 494, 493, 470, 480, 481, 488, 475, 474, 486, 481, 480, 489, 485, 480, 478, 476, 472, 491, 493, 484, 469, 471, 473, 479, 478, 474, 478, 473, 471, 481, 486, 474, 497, 498, 496, 499, 501, 502, 506, 504, 503, 509, 510, 508, 511, 513, 514, 520, 518, 517, 520, 516, 515, 507, 508, 498, 501, 511, 512, 516, 506, 505, 505, 503, 500, 518, 514, 513, 518, 515, 512, 498, 501, 499, 510, 513, 511, 508, 511, 501, 515, 505, 502, 522, 524, 525, 529, 531, 530, 530, 533, 532, 535, 537, 538, 542, 544, 543, 524, 535, 536, 533, 542, 541, 523, 529, 528, 522, 528, 532, 541, 543, 537, 543, 544, 540, 537, 543, 539, 532, 541, 535, 495, 496, 463,
499, 500, 467, 496, 499, 466, 527, 490, 489, 494, 531, 529, 526, 489, 482, 493, 529, 523, 451, 470, 468, 465, 467, 482, 465, 480, 470, 527, 526, 503, 526, 521, 500, 467, 500, 521, 545, 547, 548, 550, 552, 551, 551, 554, 553, 560, 561, 559, 565, 567, 566, 563, 567, 565, 547, 558, 559, 554, 569, 568, 556, 557, 565, 555, 562, 565, 546, 550, 549, 549, 553, 547, 555, 557, 560, 568, 569, 563, 558, 568, 562, 553, 568, 558, 572, 573, 571, 576, 577, 575, 580, 581, 579, 584, 585, 583, 578, 579, 573, 582, 583, 577, 573, 576, 574, 581, 584, 582, 579, 582, 576, 588, 589, 587, 592, 593, 591, 598, 599, 597, 602, 603, 601, 596, 597, 589, 600, 601, 593, 589, 592, 590, 599, 602, 600, 597, 600, 592, 606, 607, 605, 610, 611, 609, 614, 615, 613, 618, 619, 617, 612, 613, 607, 616, 617, 611, 607, 610, 608, 615, 618, 616, 613, 616, 610, 556, 586, 587, 590, 591, 561, 566, 595, 594, 587, 590, 560, 556, 564, 594,
604, 605, 581, 608, 609, 585, 605, 608, 584, 548, 572, 570, 561, 580, 578, 559, 578, 572, 593, 606, 604, 603, 614, 612, 601, 612, 606, 591, 604, 580, 622, 623, 621, 629, 627, 626, 630, 632, 633, 638, 636, 635, 630, 631, 623, 640, 629, 628, 633, 637, 635, 628, 626, 624, 621, 623, 625, 640, 639, 635, 639, 634, 631, 623, 631, 634, 634, 639, 628, 645, 646, 644, 648, 650, 649, 655, 656, 654, 658, 660, 659, 653, 654, 646, 650, 658, 657, 647, 649, 642, 657, 659, 652, 641, 642, 645, 651, 652, 655, 642, 651, 653, 649, 657, 651, 661, 663, 664, 668, 666, 665, 671, 672, 670, 676, 674, 673, 663, 669, 670, 674, 668, 667, 667, 665, 662, 675, 673, 670, 673, 667, 664, 681, 682, 680, 684, 686, 685, 691, 692, 690, 694, 696, 695, 689, 690, 682, 686, 694, 693, 683, 685, 678, 687, 693, 695, 678, 681, 679, 687, 688, 691, 678, 687, 689, 685, 693, 687, 632, 661, 662, 666, 638, 637, 665, 637, 633, 679, 680, 656,
660, 684, 683, 652, 677, 679, 659, 683, 677, 418, 552, 550, 417, 550, 546, 570, 571, 437, 574, 575, 441, 571, 574, 440, 414, 546, 570, 620, 621, 510, 627, 519, 517, 510, 621, 624, 626, 517, 513, 643, 644, 538, 540, 648, 647, 534, 641, 643, 534, 539, 647, 456, 457, 402, 406, 462, 460, 446, 456, 401, 400, 404, 460, 425, 497, 495, 441, 509, 507, 439, 507, 497, 423, 495, 462, 577, 622, 620, 585, 632, 630, 583, 630, 622, 611, 663, 661, 619, 671, 669, 617, 669, 663, 609, 661, 632, 575, 620, 509, 697, 699, 700, 703, 704, 702, 706, 708, 707, 707, 710, 709, 713, 714, 712, 717, 718, 716, 699, 711, 712, 715, 716, 704, 710, 720, 719, 698, 706, 705, 705, 709, 699, 700, 703, 701, 714, 717, 715, 712, 715, 703, 709, 719, 711, 729, 727, 726, 731, 728, 726, 737, 735, 734, 735, 731, 730, 727, 724, 723, 730, 726, 723, 736, 734, 732, 734, 730, 725, 740, 741, 739, 742, 744, 745, 749, 747, 746, 752, 753, 751,
756, 757, 755, 761, 759, 758, 750, 751, 741, 744, 754, 755, 759, 749, 748, 748, 746, 743, 760, 758, 755, 741, 744, 742, 753, 756, 754, 751, 754, 744, 758, 748, 745, 738, 739, 714, 717, 742, 743, 747, 722, 721, 739, 742, 717, 743, 746, 721, 702, 704, 725, 716, 718, 733, 737, 736, 721, 736, 733, 718, 704, 716, 732, 763, 765, 766, 770, 772, 771, 771, 774, 773, 779, 780, 778, 782, 784, 783, 777, 778, 766, 768, 782, 781, 781, 783, 776, 763, 764, 770, 763, 769, 773, 773, 774, 768, 775, 776, 779, 765, 773, 767, 762, 775, 777, 762, 767, 781, 793, 795, 794, 794, 791, 790, 797, 799, 798, 802, 803, 801, 806, 807, 805, 800, 801, 787, 804, 805, 789, 787, 793, 792, 786, 792, 790, 788, 789, 797, 796, 798, 795, 803, 806, 804, 787, 788, 796, 801, 804, 788, 813, 814, 812, 818, 820, 819, 816, 820, 818, 821, 823, 824, 827, 828, 826, 830, 832, 831, 809, 821, 822, 825, 826, 814, 819, 830, 829, 810, 818, 817,
808, 815, 818, 829, 831, 823, 808, 810, 813, 824, 827, 825, 822, 825, 813, 809, 817, 829, 835, 836, 834, 839, 840, 838, 843, 844, 842, 847, 848, 846, 841, 842, 836, 845, 846, 840, 836, 839, 837, 844, 847, 845, 842, 845, 839, 811, 812, 780, 784, 816, 815, 776, 808, 811, 783, 815, 808, 833, 834, 803, 837, 838, 807, 834, 837, 806, 766, 786, 785, 790, 791, 772, 780, 802, 800, 785, 790, 770, 766, 778, 800, 814, 835, 833, 828, 843, 841, 826, 841, 835, 812, 833, 802, 851, 852, 850, 855, 856, 854, 860, 858, 857, 863, 864, 862, 867, 868, 866, 872, 870, 869, 861, 862, 852, 865, 866, 856, 870, 860, 859, 856, 859, 857, 868, 871, 869, 852, 855, 853, 864, 867, 865, 862, 865, 855, 866, 869, 859, 875, 876, 874, 877, 879, 880, 885, 886, 884, 889, 890, 888, 883, 884, 876, 887, 888, 880, 876, 879, 877, 886, 889, 887, 884, 887, 879, 896, 894, 893, 902, 900, 899, 900, 896, 895, 895, 893, 891, 898, 901, 899,
897, 899, 895, 873, 874, 864, 877, 878, 868, 882, 872, 871, 874, 877, 867, 881, 871, 868, 880, 892, 891, 893, 894, 882, 890, 898, 897, 878, 891, 893, 888, 897, 892, 910, 908, 907, 914, 912, 911, 907, 908, 904, 911, 912, 906, 913, 911, 908, 908, 911, 905, 917, 918, 916, 921, 922, 920, 928, 926, 925, 924, 927, 925, 932, 930, 929, 926, 918, 917, 915, 923, 925, 929, 930, 922, 918, 921, 919, 931, 929, 926, 926, 929, 921, 924, 923, 912, 923, 915, 906, 849, 850, 753, 756, 853, 854, 858, 761, 760, 850, 853, 756, 854, 857, 760, 903, 904, 824, 905, 906, 828, 832, 909, 907, 904, 905, 827, 831, 907, 903, 843, 915, 916, 919, 920, 848, 916, 919, 847, 906, 915, 843, 935, 936, 934, 939, 940, 938, 943, 944, 942, 947, 948, 946, 941, 942, 936, 945, 946, 940, 936, 939, 937, 944, 947, 945, 942, 945, 939, 951, 952, 950, 955, 956, 954, 959, 960, 958, 963, 964, 962, 957, 958, 952, 961, 962, 956, 952, 955, 953,
960, 963, 961, 958, 961, 955, 967, 968, 966, 971, 972, 970, 975, 976, 974, 979, 980, 978, 973, 974, 968, 977, 978, 972, 968, 971, 969, 976, 979, 977, 974, 977, 971, 983, 984, 982, 987, 988, 986, 991, 992, 990, 995, 996, 994, 989, 990, 984, 993, 994, 988, 984, 987, 985, 992, 995, 993, 990, 993, 987, 965, 966, 944, 969, 970, 948, 966, 969, 947, 981, 982, 960, 985, 986, 964, 982, 985, 963, 940, 951, 949, 948, 959, 957, 946, 957, 951, 972, 983, 981, 980, 991, 989, 978, 989, 983, 970, 981, 959, 999, 1000, 998, 1004, 1002, 1001, 1007, 1008, 1006, 1013, 1011, 1010, 1005, 1006, 1000, 1011, 1004, 1003, 1000, 1003, 1001, 1012, 1010, 1006, 1014, 1015, 1013, 1008, 1009, 1014, 1006, 1010, 1003, 1020, 1021, 1019, 1023, 1025, 1024, 1030, 1031, 1029, 1033, 1035, 1034, 1028, 1029, 1021, 1025, 1033, 1032, 1016, 1022, 1024, 1026, 1032, 1034, 1017, 1020, 1018, 1027, 1030, 1028, 1026, 1028, 1020, 1017, 1024, 1032, 1038, 1039, 1037, 1045, 1043, 1042, 1048, 1049, 1047, 1055, 1053, 1052,
1046, 1047, 1039, 1053, 1045, 1044, 1041, 1044, 1042, 1051, 1054, 1052, 1039, 1041, 1040, 1049, 1051, 1050, 1047, 1050, 1041, 1050, 1052, 1044, 1058, 1060, 1061, 1063, 1065, 1064, 1070, 1071, 1069, 1073, 1075, 1074, 1068, 1069, 1061, 1067, 1073, 1072, 1062, 1064, 1057, 1068, 1072, 1074, 1057, 1060, 1058, 1067, 1066, 1064, 1064, 1066, 1060, 1066, 1072, 1068, 1007, 1036, 1037, 1043, 1015, 1014, 1037, 1040, 1009, 1042, 1014, 1009, 1058, 1059, 1031, 1035, 1063, 1062, 1056, 1058, 1030, 1027, 1034, 1062, 1081, 1082, 1080, 1086, 1084, 1083, 1086, 1088, 1087, 1093, 1094, 1092, 1096, 1098, 1097, 1091, 1092, 1082, 1088, 1096, 1095, 1083, 1084, 1077, 1084, 1087, 1078, 1089, 1095, 1097, 1078, 1081, 1079, 1090, 1093, 1091, 1089, 1091, 1081, 1078, 1087, 1095, 1101, 1102, 1100, 1105, 1106, 1104, 1109, 1110, 1108, 1113, 1114, 1112, 1107, 1108, 1102, 1111, 1112, 1106, 1102, 1105, 1103, 1110, 1113, 1111, 1108, 1111, 1105, 1118, 1119, 1117, 1123, 1125, 1124, 1128, 1129, 1127, 1135, 1133, 1132, 1131, 1134, 1132, 1126, 1127, 1119, 1125, 1131, 1130, 1122, 1124, 1118, 1132, 1133, 1129,
1126, 1130, 1132, 1123, 1122, 1120, 1120, 1122, 1116, 1118, 1124, 1130, 1138, 1139, 1137, 1142, 1143, 1141, 1146, 1147, 1145, 1150, 1151, 1149, 1144, 1145, 1139, 1148, 1149, 1143, 1139, 1142, 1140, 1147, 1150, 1148, 1145, 1148, 1142, 1093, 1116, 1117, 1098, 1121, 1120, 1115, 1116, 1093, 1097, 1120, 1115, 1136, 1137, 1110, 1140, 1141, 1114, 1137, 1140, 1113, 1082, 1101, 1099, 1094, 1109, 1107, 1092, 1107, 1101, 1119, 1138, 1136, 1127, 1129, 1146, 1127, 1144, 1138, 1117, 1136, 1109, 1154, 1155, 1153, 1158, 1159, 1157, 1161, 1163, 1162, 1165, 1163, 1161, 1168, 1169, 1167, 1170, 1172, 1173, 1166, 1167, 1155, 1170, 1171, 1159, 1175, 1165, 1164, 1156, 1157, 1161, 1164, 1161, 1157, 1153, 1155, 1158, 1169, 1172, 1170, 1167, 1170, 1158, 1171, 1174, 1164, 1178, 1179, 1177, 1185, 1183, 1182, 1181, 1184, 1182, 1193, 1191, 1190, 1195, 1197, 1196, 1185, 1195, 1194, 1187, 1192, 1190, 1182, 1183, 1179, 1176, 1180, 1182, 1188, 1194, 1196, 1179, 1183, 1194, 1200, 1201, 1199, 1204, 1205, 1203, 1208, 1209, 1207, 1212, 1213, 1211, 1206, 1207, 1201, 1210, 1211, 1205, 1201, 1204, 1202,
1209, 1212, 1210, 1207, 1210, 1204, 1220, 1222, 1221, 1221, 1218, 1217, 1225, 1226, 1224, 1231, 1229, 1228, 1223, 1224, 1216, 1229, 1222, 1220, 1215, 1216, 1220, 1219, 1217, 1214, 1226, 1230, 1228, 1232, 1233, 1231, 1226, 1227, 1232, 1224, 1228, 1220, 1198, 1199, 1169, 1202, 1203, 1173, 1199, 1202, 1172, 1218, 1193, 1192, 1214, 1217, 1192, 1173, 1187, 1186, 1190, 1191, 1175, 1171, 1186, 1190, 1205, 1215, 1214, 1213, 1225, 1223, 1205, 1211, 1223, 1203, 1214, 1187, 1076, 1077, 976, 1079, 1080, 980, 976, 1077, 1079, 1099, 1100, 992, 1103, 1104, 996, 1100, 1103, 995, 1080, 1099, 991, 1152, 1153, 1049, 1162, 1055, 1054, 1153, 1156, 1051, 1160, 1054, 1051, 1176, 1177, 1071, 1075, 1181, 1180, 1070, 1074, 1180, 956, 999, 997, 964, 1007, 1005, 962, 1005, 999, 988, 1038, 1036, 996, 1048, 1046, 994, 1046, 1038, 986, 1036, 1007, 1106, 1154, 1152, 1114, 1168, 1166, 1112, 1166, 1154, 1143, 1200, 1198, 1151, 1208, 1206, 1149, 1206, 1200, 1141, 1198, 1168, 1104, 1152, 1048, 1236, 1237, 1235, 1240, 1241, 1239, 1244, 1245, 1243, 1248, 1249, 1247, 1242, 1243, 1237,
1246, 1247, 1241, 1237, 1240, 1238, 1245, 1248, 1246, 1243, 1246, 1240, 1252, 1253, 1251, 1255, 1257, 1256, 1259, 1257, 1255, 1262, 1263, 1261, 1268, 1266, 1265, 1260, 1261, 1253, 1266, 1259, 1258, 1250, 1251, 1255, 1258, 1255, 1251, 1267, 1265, 1261, 1269, 1270, 1268, 1263, 1264, 1269, 1261, 1265, 1258, 1273, 1274, 1272, 1277, 1278, 1276, 1281, 1282, 1280, 1285, 1286, 1284, 1279, 1280, 1274, 1283, 1284, 1278, 1274, 1277, 1275, 1282, 1285, 1283, 1280, 1283, 1277, 1289, 1290, 1288, 1297, 1299, 1298, 1298, 1295, 1294, 1302, 1303, 1301, 1306, 1307, 1305, 1311, 1309, 1308, 1300, 1301, 1290, 1304, 1305, 1293, 1309, 1299, 1297, 1293, 1297, 1296, 1292, 1296, 1294, 1307, 1310, 1308, 1290, 1292, 1291, 1303, 1306, 1304, 1290, 1301, 1304, 1305, 1308, 1297, 1271, 1272, 1245, 1275, 1276, 1249, 1272, 1275, 1248, 1262, 1287, 1288, 1295, 1270, 1269, 1288, 1291, 1264, 1291, 1294, 1269, 1239, 1241, 1252, 1249, 1262, 1260, 1247, 1260, 1252, 1278, 1289, 1287, 1286, 1302, 1300, 1284, 1300, 1289, 1276, 1287, 1262, 1314, 1315, 1313, 1318, 1319, 1317, 1320, 1322, 1323, 1326, 1327, 1325,
1320, 1321, 1315, 1324, 1325, 1319, 1315, 1318, 1316, 1323, 1326, 1324, 1321, 1324, 1318, 1332, 1333, 1331, 1334, 1336, 1337, 1342, 1343, 1341, 1346, 1347, 1345, 1340, 1341, 1333, 1344, 1345, 1337, 1333, 1336, 1334, 1343, 1346, 1344, 1341, 1344, 1336, 1350, 1351, 1349, 1354, 1355, 1353, 1361, 1359, 1358, 1357, 1360, 1358, 1364, 1366, 1367, 1369, 1371, 1370, 1364, 1365, 1355, 1361, 1369, 1368, 1358, 1359, 1351, 1348, 1356, 1358, 1362, 1368, 1370, 1349, 1351, 1354, 1363, 1366, 1364, 1373, 1372, 1370, 1370, 1372, 1366, 1362, 1364, 1354, 1351, 1359, 1368, 1376, 1377, 1375, 1380, 1381, 1379, 1384, 1385, 1383, 1388, 1389, 1387, 1382, 1383, 1377, 1386, 1387, 1381, 1377, 1380, 1378, 1385, 1388, 1386, 1383, 1386, 1380, 1348, 1349, 1323, 1352, 1353, 1327, 1329, 1357, 1356, 1349, 1352, 1326, 1322, 1328, 1356, 1374, 1375, 1343, 1378, 1379, 1347, 1375, 1378, 1346, 1319, 1332, 1330, 1327, 1342, 1340, 1325, 1340, 1332, 1355, 1376, 1374, 1367, 1384, 1382, 1365, 1382, 1376, 1353, 1374, 1342, 1395, 1393, 1392, 1398, 1399, 1397, 1401, 1403, 1402, 1405, 1403, 1401, 1402, 1395, 1394,
1391, 1394, 1392, 1396, 1397, 1401, 1399, 1404, 1401, 1396, 1400, 1394, 1409, 1411, 1410, 1408, 1410, 1407, 1412, 1414, 1415, 1423, 1421, 1420, 1426, 1427, 1425, 1430, 1431, 1429, 1433, 1435, 1434, 1437, 1435, 1433, 1424, 1425, 1415, 1434, 1423, 1422, 1417, 1422, 1420, 1428, 1429, 1433, 1431, 1436, 1433, 1415, 1417, 1416, 1420, 1421, 1419, 1425, 1427, 1430, 1413, 1416, 1420, 1425, 1428, 1417, 1428, 1432, 1422, 1441, 1443, 1442, 1447, 1449, 1448, 1443, 1447, 1446, 1440, 1442, 1439, 1444, 1446, 1448, 1439, 1442, 1446, 1412, 1413, 1399, 1419, 1405, 1404, 1413, 1418, 1404, 1411, 1441, 1440, 1410, 1440, 1438, 1312, 1313, 1282, 1316, 1317, 1286, 1313, 1316, 1285, 1330, 1331, 1303, 1334, 1335, 1307, 1339, 1311, 1310, 1331, 1334, 1306, 1338, 1310, 1307, 1317, 1330, 1302, 1337, 1391, 1390, 1392, 1393, 1339, 1345, 1347, 1398, 1335, 1390, 1392, 1345, 1396, 1391, 1381, 1414, 1412, 1389, 1426, 1424, 1387, 1424, 1414, 1379, 1412, 1398, 933, 934, 599, 937, 938, 603, 934, 937, 602, 949, 950, 615, 953, 954, 619, 950, 953, 618, 938, 949, 614, 997, 998, 672,
1002, 676, 675, 998, 1001, 675, 1018, 1019, 692, 696, 1023, 1022, 1016, 1018, 691, 688, 695, 1022, 954, 997, 671, 1234, 1235, 886, 1238, 1239, 890, 1235, 1238, 889, 1256, 902, 901, 1254, 901, 898, 1239, 1250, 898, 483, 484, 713, 719, 720, 492, 711, 719, 491, 525, 740, 738, 538, 752, 750, 536, 750, 740, 484, 523, 738, 646, 851, 849, 656, 863, 861, 654, 861, 851, 682, 875, 873, 692, 885, 883, 690, 883, 875, 680, 873, 863, 644, 849, 752, 1021, 1236, 1234, 1031, 1244, 1242, 1029, 1242, 1236, 1061, 1273, 1271, 1071, 1281, 1279, 1069, 1279, 1273, 1059, 1271, 1244, 1177, 1179, 1314, 1189, 1322, 1320, 1329, 1328, 1196, 1196, 1328, 1322, 1188, 1320, 1314, 1177, 1312, 1281, 1019, 1234, 885, 4, 448, 449, 10, 459, 458, 8, 458, 448, 468, 469, 31, 477, 42, 40, 31, 469, 472, 476, 40, 34, 449, 468, 30, 701, 702, 95, 99, 708, 706, 91, 698, 701, 91, 98, 706, 127, 723, 724, 729, 133, 132, 727, 132, 128, 702, 723, 127, 151, 153, 588, 594, 595, 155,
166, 598, 596, 586, 594, 154, 164, 596, 588, 219, 935, 933, 229, 943, 941, 227, 941, 935, 257, 967, 965, 265, 975, 973, 263, 973, 967, 255, 965, 943, 1085, 1083, 352, 1083, 1076, 349, 265, 349, 1076, 217, 933, 598, 1456, 1458, 1457, 1462, 1464, 1463, 1458, 1462, 1461, 1455, 1457, 1454, 1461, 1463, 1460, 1457, 1461, 1459, 1468, 1470, 1469, 1474, 1476, 1475, 1470, 1474, 1473, 1467, 1469, 1466, 1473, 1475, 1472, 1469, 1473, 1471, 1464, 1468, 1467, 1463, 1467, 1465, 1452, 1456, 1455, 1450, 1451, 1455, 1481, 1482, 1480, 1484, 1486, 1485, 1494, 1495, 1493, 1499, 1501, 1500, 1497, 1501, 1499, 1492, 1493, 1482, 1486, 1497, 1496, 1483, 1485, 1478, 1491, 1499, 1498, 1489, 1496, 1499, 1477, 1478, 1481, 1487, 1488, 1484, 1489, 1491, 1494, 1479, 1487, 1483, 1478, 1489, 1492, 1485, 1496, 1489, 1504, 1505, 1503, 1508, 1509, 1507, 1512, 1513, 1511, 1516, 1517, 1515, 1510, 1511, 1505, 1514, 1515, 1509, 1505, 1508, 1506, 1513, 1516, 1514, 1511, 1514, 1508, 1518, 1520, 1521, 1524, 1525, 1523, 1527, 1529, 1528, 1530, 1532, 1533, 1536, 1537, 1535,
1539, 1541, 1540, 1520, 1530, 1531, 1534, 1535, 1525, 1529, 1539, 1538, 1526, 1528, 1520, 1530, 1538, 1540, 1521, 1524, 1522, 1533, 1536, 1534, 1531, 1534, 1524, 1528, 1538, 1530, 1544, 1545, 1543, 1548, 1549, 1547, 1554, 1555, 1553, 1556, 1558, 1559, 1563, 1561, 1560, 1552, 1553, 1545, 1548, 1556, 1557, 1561, 1551, 1550, 1562, 1560, 1557, 1545, 1548, 1546, 1555, 1558, 1556, 1553, 1556, 1548, 1560, 1550, 1549, 1490, 1518, 1519, 1522, 1523, 1495, 1500, 1527, 1526, 1519, 1522, 1494, 1490, 1498, 1526, 1542, 1543, 1513, 1546, 1547, 1517, 1543, 1546, 1516, 1482, 1504, 1502, 1495, 1512, 1510, 1493, 1510, 1504, 1525, 1544, 1542, 1537, 1554, 1552, 1535, 1552, 1544, 1523, 1542, 1512, 1564, 1566, 1567, 1572, 1570, 1569, 1581, 1579, 1578, 1583, 1580, 1578, 1566, 1575, 1576, 1581, 1572, 1571, 1571, 1569, 1565, 1579, 1576, 1575, 1577, 1582, 1578, 1574, 1573, 1569, 1573, 1568, 1565, 1579, 1571, 1567, 1589, 1587, 1586, 1588, 1586, 1584, 1587, 1583, 1582, 1586, 1582, 1577, 1592, 1593, 1591, 1596, 1597, 1595, 1600, 1601, 1599, 1604, 1605, 1603, 1598, 1599, 1593, 1602, 1603, 1597,
1593, 1596, 1594, 1601, 1604, 1602, 1599, 1602, 1596, 1608, 1609, 1607, 1616, 1614, 1613, 1618, 1615, 1613, 1621, 1622, 1620, 1628, 1626, 1625, 1619, 1620, 1609, 1626, 1618, 1617, 1614, 1611, 1610, 1612, 1617, 1613, 1627, 1625, 1623, 1607, 1609, 1612, 1620, 1622, 1624, 1609, 1620, 1623, 1625, 1617, 1612, 1631, 1632, 1630, 1635, 1636, 1634, 1639, 1640, 1638, 1643, 1644, 1642, 1637, 1638, 1632, 1641, 1642, 1636, 1632, 1635, 1633, 1640, 1643, 1641, 1638, 1641, 1635, 1647, 1648, 1646, 1654, 1652, 1651, 1655, 1657, 1658, 1663, 1661, 1660, 1655, 1656, 1648, 1665, 1654, 1653, 1658, 1662, 1660, 1653, 1651, 1649, 1646, 1648, 1650, 1665, 1664, 1660, 1664, 1659, 1656, 1648, 1656, 1659, 1659, 1664, 1653, 1629, 1630, 1601, 1633, 1634, 1605, 1630, 1633, 1604, 1645, 1646, 1622, 1652, 1628, 1627, 1622, 1646, 1649, 1651, 1627, 1624, 1597, 1608, 1606, 1605, 1621, 1619, 1603, 1619, 1608, 1636, 1647, 1645, 1644, 1657, 1655, 1642, 1655, 1647, 1634, 1645, 1621, 1590, 1591, 1533, 1594, 1595, 1537, 1591, 1594, 1536, 1606, 1607, 1555, 1558, 1610, 1611, 1616, 1563, 1562, 1607, 1610, 1558,
1611, 1614, 1562, 1595, 1606, 1554, 1509, 1566, 1564, 1515, 1517, 1577, 1515, 1575, 1566, 1547, 1549, 1585, 1589, 1588, 1550, 1588, 1585, 1549, 1517, 1547, 1584, 1669, 1671, 1670, 1675, 1677, 1676, 1671, 1675, 1674, 1668, 1670, 1667, 1674, 1676, 1673, 1670, 1674, 1672, 1681, 1683, 1682, 1687, 1689, 1688, 1683, 1687, 1686, 1678, 1680, 1682, 1684, 1686, 1688, 1679, 1682, 1686, 1677, 1681, 1680, 1673, 1676, 1680, 1693, 1695, 1694, 1692, 1694, 1691, 1689, 1693, 1692, 1685, 1688, 1692, 1698, 1699, 1697, 1702, 1703, 1701, 1706, 1707, 1705, 1710, 1711, 1709, 1704, 1705, 1699, 1708, 1709, 1703, 1699, 1702, 1700, 1707, 1710, 1708, 1705, 1708, 1702, 1712, 1714, 1715, 1719, 1717, 1716, 1720, 1722, 1723, 1727, 1725, 1724, 1720, 1721, 1715, 1725, 1719, 1718, 1718, 1716, 1713, 1723, 1726, 1724, 1724, 1718, 1715, 1730, 1731, 1729, 1734, 1735, 1733, 1738, 1739, 1737, 1742, 1743, 1741, 1736, 1737, 1731, 1740, 1741, 1735, 1731, 1734, 1732, 1739, 1742, 1740, 1737, 1740, 1734, 1746, 1747, 1745, 1751, 1749, 1748, 1752, 1754, 1755, 1763, 1761, 1760, 1752, 1753, 1747, 1759, 1751, 1750,
1747, 1750, 1748, 1762, 1760, 1756, 1755, 1757, 1756, 1760, 1761, 1759, 1753, 1756, 1760, 1758, 1750, 1747, 1728, 1729, 1707, 1732, 1733, 1711, 1729, 1732, 1710, 1744, 1745, 1723, 1749, 1727, 1726, 1745, 1748, 1726, 1703, 1714, 1712, 1711, 1722, 1720, 1709, 1720, 1714, 1735, 1746, 1744, 1743, 1754, 1752, 1741, 1752, 1746, 1733, 1744, 1722, 1764, 1766, 1767, 1770, 1771, 1769, 1776, 1777, 1775, 1780, 1781, 1779, 1783, 1785, 1784, 1774, 1775, 1767, 1778, 1779, 1771, 1773, 1783, 1782, 1774, 1782, 1784, 1767, 1770, 1768, 1777, 1780, 1778, 1775, 1778, 1770, 1772, 1782, 1774, 1788, 1789, 1787, 1795, 1793, 1792, 1798, 1799, 1797, 1805, 1803, 1802, 1796, 1797, 1789, 1803, 1795, 1794, 1791, 1794, 1792, 1801, 1804, 1802, 1789, 1791, 1790, 1799, 1801, 1800, 1797, 1800, 1791, 1800, 1802, 1794, 1808, 1809, 1807, 1812, 1813, 1811, 1819, 1817, 1816, 1815, 1818, 1816, 1824, 1825, 1823, 1827, 1829, 1828, 1822, 1823, 1813, 1819, 1827, 1826, 1816, 1817, 1809, 1814, 1816, 1808, 1820, 1826, 1828, 1807, 1809, 1812, 1821, 1824, 1822, 1820, 1822, 1812, 1817, 1826, 1820, 1832, 1833, 1831,
1836, 1837, 1835, 1839, 1841, 1840, 1843, 1841, 1839, 1846, 1847, 1845, 1850, 1851, 1849, 1855, 1853, 1852, 1844, 1845, 1833, 1848, 1849, 1837, 1853, 1843, 1842, 1834, 1835, 1839, 1842, 1839, 1835, 1851, 1854, 1852, 1831, 1833, 1836, 1847, 1850, 1848, 1845, 1848, 1836, 1849, 1852, 1842, 1806, 1807, 1777, 1810, 1811, 1781, 1785, 1815, 1814, 1807, 1810, 1780, 1776, 1784, 1814, 1830, 1831, 1799, 1840, 1805, 1804, 1831, 1834, 1801, 1838, 1804, 1801, 1771, 1788, 1786, 1781, 1798, 1796, 1779, 1796, 1788, 1813, 1832, 1830, 1825, 1846, 1844, 1823, 1844, 1832, 1811, 1830, 1798, 1764, 1765, 1739, 1768, 1769, 1743, 1765, 1768, 1742, 1786, 1787, 1755, 1793, 1763, 1762, 1787, 1790, 1757, 1790, 1792, 1762, 1769, 1786, 1754, 1857, 1858, 1855, 1851, 1856, 1857, 1476, 1669, 1668, 1475, 1668, 1666, 1696, 1697, 1640, 1700, 1701, 1644, 1697, 1700, 1643, 1657, 1712, 1713, 1717, 1663, 1662, 1716, 1662, 1658, 1701, 1712, 1657, 1540, 1541, 1452, 1532, 1540, 1451, 1453, 1454, 1592, 1459, 1460, 1600, 1454, 1459, 1598, 1466, 1631, 1629, 1472, 1639, 1637, 1471, 1637, 1631, 1460, 1465, 1629,
1450, 1453, 1590, 1667, 1698, 1696, 1673, 1706, 1704, 1672, 1704, 1698, 1678, 1679, 1730, 1685, 1738, 1736, 1684, 1736, 1730, 1673, 1678, 1728, 1691, 1766, 1764, 1773, 1772, 1694, 1694, 1772, 1766, 1690, 1764, 1738, 1666, 1696, 1639, 1862, 1864, 1863, 1859, 1861, 1863, 1865, 1867, 1868, 1871, 1872, 1870, 1877, 1878, 1876, 1881, 1882, 1880, 1888, 1886, 1885, 1884, 1887, 1885, 1875, 1876, 1868, 1879, 1880, 1872, 1874, 1884, 1883, 1885, 1886, 1878, 1875, 1883, 1885, 1868, 1871, 1869, 1876, 1878, 1881, 1876, 1879, 1871, 1867, 1873, 1883, 1891, 1893, 1894, 1896, 1898, 1897, 1903, 1904, 1902, 1910, 1908, 1907, 1906, 1909, 1907, 1901, 1902, 1894, 1900, 1906, 1905, 1889, 1895, 1897, 1907, 1908, 1904, 1901, 1905, 1907, 1890, 1893, 1891, 1900, 1899, 1897, 1897, 1899, 1893, 1893, 1899, 1905, 1891, 1892, 1882, 1888, 1896, 1895, 1889, 1891, 1881, 1878, 1886, 1895, 1860, 1867, 1865, 1874, 1873, 1863, 1863, 1873, 1867, 1913, 1914, 1912, 1917, 1918, 1916, 1921, 1922, 1920, 1925, 1926, 1924, 1919, 1920, 1914, 1923, 1924, 1918, 1914, 1917, 1915, 1922, 1925, 1923, 1920, 1923, 1917,
1929, 1930, 1928, 1937, 1939, 1938, 1938, 1935, 1934, 1942, 1943, 1941, 1946, 1947, 1945, 1951, 1949, 1948, 1940, 1941, 1930, 1944, 1945, 1933, 1949, 1939, 1937, 1932, 1933, 1937, 1932, 1936, 1934, 1947, 1950, 1948, 1930, 1932, 1931, 1943, 1946, 1944, 1930, 1941, 1944, 1945, 1948, 1937, 1954, 1955, 1953, 1958, 1959, 1957, 1962, 1963, 1961, 1966, 1967, 1965, 1960, 1961, 1955, 1964, 1965, 1959, 1955, 1958, 1956, 1963, 1966, 1964, 1961, 1964, 1958, 1970, 1971, 1969, 1974, 1975, 1973, 1978, 1979, 1977, 1982, 1983, 1981, 1976, 1977, 1971, 1980, 1981, 1975, 1971, 1974, 1972, 1979, 1982, 1980, 1977, 1980, 1974, 1952, 1953, 1922, 1956, 1957, 1926, 1953, 1956, 1925, 1968, 1969, 1943, 1946, 1972, 1973, 1969, 1972, 1946, 1918, 1929, 1927, 1926, 1942, 1940, 1924, 1940, 1929, 1959, 1970, 1968, 1967, 1978, 1976, 1965, 1976, 1970, 1957, 1968, 1942, 1987, 1989, 1988, 1984, 1986, 1988, 1990, 1992, 1993, 1996, 1997, 1995, 2005, 2006, 2004, 2010, 2008, 2007, 2010, 2012, 2011, 2000, 2001, 1993, 2003, 2004, 1997, 1999, 2009, 2007, 2007, 2008, 2001, 2001, 2008, 2011, 1993, 1996, 1994,
2002, 2005, 2003, 1993, 2001, 2003, 1998, 2007, 2000, 2015, 2016, 2014, 2019, 2020, 2018, 2023, 2024, 2022, 2027, 2028, 2026, 2021, 2022, 2016, 2025, 2026, 2020, 2016, 2019, 2017, 2024, 2027, 2025, 2022, 2025, 2019, 2032, 2033, 2031, 2041, 2039, 2038, 2037, 2040, 2038, 2045, 2047, 2046, 2041, 2045, 2044, 2038, 2039, 2033, 2036, 2038, 2032, 2042, 2044, 2046, 2037, 2036, 2034, 2034, 2036, 2030, 2033, 2039, 2044, 2050, 2051, 2049, 2054, 2055, 2053, 2056, 2058, 2059, 2062, 2063, 2061, 2056, 2057, 2051, 2060, 2061, 2055, 2051, 2054, 2052, 2059, 2062, 2060, 2057, 2060, 2054, 2005, 2030, 2031, 2012, 2035, 2034, 2029, 2030, 2005, 2002, 2011, 2034, 2048, 2049, 2024, 2052, 2053, 2028, 2049, 2052, 2027, 1997, 2015, 2013, 2006, 2023, 2021, 2004, 2021, 2015, 2031, 2033, 2050, 2043, 2058, 2056, 2065, 2064, 2046, 2064, 2058, 2043, 2042, 2056, 2050, 2031, 2048, 2023, 1910, 1987, 1986, 1904, 1908, 1986, 1990, 1991, 1963, 1994, 1995, 1967, 1991, 1994, 1966, 2013, 2014, 1979, 2017, 2018, 1983, 2014, 2017, 1982, 1995, 2013, 1978, 1872, 1913, 1911, 1882, 1921, 1919, 1880, 1919, 1913,
1894, 1954, 1952, 1902, 1904, 1962, 1902, 1960, 1954, 1892, 1952, 1921, 1985, 1992, 1990, 1999, 1998, 1988, 1988, 1998, 1992, 1984, 1990, 1962, 2068, 2069, 2067, 2075, 2073, 2072, 2071, 2074, 2072, 2082, 2084, 2083, 2075, 2082, 2081, 2072, 2073, 2069, 2066, 2070, 2072, 2077, 2081, 2083, 2069, 2073, 2081, 2087, 2088, 2086, 2091, 2092, 2090, 2093, 2095, 2096, 2099, 2100, 2098, 2093, 2094, 2088, 2097, 2098, 2092, 2088, 2091, 2089, 2096, 2099, 2097, 2094, 2097, 2091, 2109, 2111, 2110, 2110, 2107, 2106, 2112, 2114, 2115, 2123, 2121, 2120, 2112, 2113, 2105, 2119, 2111, 2109, 2104, 2105, 2109, 2104, 2108, 2106, 2117, 2122, 2120, 2115, 2117, 2116, 2120, 2121, 2119, 2113, 2116, 2120, 2118, 2109, 2105, 2129, 2130, 2128, 2134, 2132, 2131, 2134, 2136, 2135, 2140, 2141, 2139, 2149, 2147, 2146, 2145, 2148, 2146, 2129, 2138, 2139, 2136, 2143, 2142, 2131, 2132, 2125, 2125, 2132, 2135, 2146, 2147, 2141, 2144, 2146, 2140, 2126, 2129, 2127, 2145, 2144, 2142, 2142, 2144, 2138, 2137, 2138, 2129, 2126, 2135, 2142, 2107, 2080, 2079, 2103, 2106, 2079, 2124, 2125, 2096, 2127, 2128, 2100,
2102, 2133, 2131, 2096, 2125, 2127, 2101, 2131, 2124, 2067, 2069, 2087, 2078, 2095, 2093, 2102, 2101, 2083, 2101, 2095, 2078, 2077, 2093, 2087, 2152, 2153, 2151, 2154, 2156, 2157, 2162, 2163, 2161, 2166, 2167, 2165, 2160, 2161, 2153, 2164, 2165, 2157, 2153, 2156, 2154, 2163, 2166, 2164, 2161, 2164, 2156, 2174, 2172, 2171, 2179, 2180, 2178, 2183, 2185, 2184, 2187, 2185, 2183, 2184, 2174, 2173, 2169, 2173, 2171, 2177, 2178, 2183, 2186, 2183, 2178, 2188, 2189, 2187, 2180, 2181, 2188, 2177, 2182, 2173, 2192, 2193, 2191, 2196, 2197, 2195, 2200, 2201, 2199, 2204, 2205, 2203, 2198, 2199, 2193, 2202, 2203, 2197, 2193, 2196, 2194, 2201, 2204, 2202, 2199, 2202, 2196, 2208, 2209, 2207, 2216, 2218, 2217, 2217, 2214, 2213, 2221, 2222, 2220, 2225, 2226, 2224, 2230, 2228, 2227, 2219, 2220, 2209, 2223, 2224, 2212, 2228, 2218, 2216, 2211, 2212, 2216, 2211, 2215, 2213, 2226, 2229, 2227, 2209, 2211, 2210, 2222, 2225, 2223, 2209, 2220, 2223, 2224, 2227, 2216, 2190, 2191, 2163, 2194, 2195, 2167, 2191, 2194, 2166, 2179, 2206, 2207, 2214, 2189, 2188, 2207, 2210, 2181, 2210, 2213, 2188,
2157, 2169, 2168, 2171, 2172, 2159, 2165, 2167, 2179, 2155, 2168, 2171, 2165, 2177, 2169, 2197, 2208, 2206, 2205, 2221, 2219, 2203, 2219, 2208, 2195, 2206, 2179, 2233, 2234, 2232, 2237, 2238, 2236, 2240, 2242, 2241, 2244, 2242, 2240, 2247, 2248, 2246, 2251, 2252, 2250, 2245, 2246, 2234, 2237, 2249, 2250, 2235, 2236, 2240, 2243, 2240, 2236, 2232, 2234, 2237, 2248, 2251, 2249, 2246, 2249, 2237, 2259, 2261, 2260, 2268, 2270, 2269, 2269, 2266, 2265, 2266, 2257, 2256, 2263, 2264, 2268, 2263, 2267, 2265, 2254, 2258, 2260, 2262, 2265, 2256, 2273, 2274, 2272, 2277, 2278, 2276, 2281, 2282, 2280, 2285, 2286, 2284, 2279, 2280, 2274, 2283, 2284, 2278, 2274, 2277, 2275, 2282, 2285, 2283, 2280, 2283, 2277, 2289, 2290, 2288, 2295, 2293, 2292, 2300, 2301, 2299, 2304, 2305, 2303, 2307, 2309, 2308, 2311, 2309, 2307, 2289, 2298, 2299, 2308, 2297, 2296, 2290, 2294, 2292, 2302, 2303, 2307, 2310, 2307, 2303, 2296, 2297, 2295, 2299, 2301, 2304, 2290, 2291, 2296, 2299, 2302, 2291, 2306, 2296, 2291, 2271, 2272, 2248, 2275, 2276, 2252, 2272, 2275, 2251, 2287, 2288, 2264, 2293, 2270, 2268,
2288, 2292, 2268, 2256, 2257, 2244, 2252, 2263, 2262, 2253, 2256, 2243, 2250, 2262, 2253, 2278, 2289, 2287, 2286, 2300, 2298, 2284, 2298, 2289, 2252, 2276, 2287, 2312, 2314, 2315, 2318, 2319, 2317, 2327, 2328, 2326, 2332, 2330, 2329, 2332, 2334, 2333, 2322, 2323, 2315, 2325, 2326, 2319, 2321, 2331, 2329, 2329, 2330, 2323, 2323, 2330, 2333, 2315, 2318, 2316, 2324, 2327, 2325, 2315, 2323, 2325, 2320, 2329, 2322, 2337, 2338, 2336, 2341, 2342, 2340, 2345, 2346, 2344, 2349, 2350, 2348, 2343, 2344, 2338, 2347, 2348, 2342, 2338, 2341, 2339, 2346, 2349, 2347, 2344, 2347, 2341, 2354, 2355, 2353, 2363, 2361, 2360, 2359, 2362, 2360, 2370, 2372, 2371, 2363, 2370, 2369, 2360, 2361, 2355, 2358, 2360, 2354, 2365, 2369, 2371, 2359, 2358, 2356, 2356, 2358, 2352, 2355, 2361, 2369, 2375, 2376, 2374, 2379, 2380, 2378, 2381, 2383, 2384, 2387, 2388, 2386, 2381, 2382, 2376, 2385, 2386, 2380, 2376, 2379, 2377, 2384, 2387, 2385, 2382, 2385, 2379, 2327, 2352, 2353, 2334, 2357, 2356, 2351, 2352, 2327, 2324, 2333, 2356, 2373, 2374, 2346, 2377, 2378, 2350, 2374, 2377, 2349, 2319, 2337, 2335,
2328, 2345, 2343, 2326, 2343, 2337, 2353, 2355, 2375, 2366, 2383, 2381, 2390, 2389, 2371, 2389, 2383, 2366, 2365, 2381, 2375, 2353, 2373, 2345, 2231, 2232, 2115, 2241, 2123, 2122, 2232, 2235, 2117, 2235, 2239, 2122, 2149, 2259, 2258, 2141, 2147, 2258, 2312, 2313, 2201, 2316, 2317, 2205, 2313, 2316, 2204, 2335, 2336, 2222, 2225, 2339, 2340, 2336, 2339, 2225, 2317, 2335, 2221, 2092, 2152, 2150, 2100, 2162, 2160, 2098, 2160, 2152, 2130, 2192, 2190, 2139, 2141, 2200, 2139, 2198, 2192, 2128, 2190, 2162, 2255, 2314, 2312, 2321, 2320, 2260, 2260, 2320, 2314, 2367, 2368, 2311, 2364, 2367, 2310, 2254, 2312, 2200, 2394, 2396, 2397, 2401, 2399, 2398, 2401, 2403, 2402, 2412, 2410, 2409, 2412, 2414, 2413, 2406, 2407, 2397, 2405, 2411, 2409, 2398, 2399, 2392, 2399, 2402, 2393, 2409, 2410, 2407, 2407, 2410, 2413, 2393, 2396, 2394, 2405, 2404, 2402, 2402, 2404, 2396, 2404, 2409, 2406, 2414, 2417, 2416, 2408, 2413, 2416, 2420, 2421, 2419, 2424, 2425, 2423, 2428, 2429, 2427, 2432, 2433, 2431, 2426, 2427, 2421, 2430, 2431, 2425, 2421, 2424, 2422, 2429, 2432, 2430, 2427, 2430, 2424,
2436, 2437, 2435, 2440, 2441, 2439, 2444, 2445, 2443, 2448, 2449, 2447, 2442, 2443, 2437, 2446, 2447, 2441, 2437, 2440, 2438, 2445, 2448, 2446, 2443, 2446, 2440, 2452, 2453, 2451, 2456, 2457, 2455, 2463, 2461, 2460, 2459, 2462, 2460, 2467, 2468, 2466, 2472, 2474, 2473, 2456, 2465, 2466, 2463, 2470, 2469, 2460, 2461, 2453, 2458, 2460, 2452, 2465, 2471, 2473, 2451, 2453, 2456, 2472, 2471, 2469, 2469, 2471, 2465, 2464, 2465, 2456, 2461, 2469, 2464, 2477, 2478, 2476, 2481, 2482, 2480, 2485, 2486, 2484, 2489, 2490, 2488, 2483, 2484, 2478, 2487, 2488, 2482, 2478, 2481, 2479, 2486, 2489, 2487, 2484, 2487, 2481, 2428, 2450, 2451, 2454, 2455, 2433, 2451, 2454, 2432, 2475, 2476, 2445, 2479, 2480, 2449, 2476, 2479, 2448, 2425, 2436, 2434, 2433, 2444, 2442, 2431, 2442, 2436, 2457, 2477, 2475, 2468, 2485, 2483, 2466, 2483, 2477, 2455, 2475, 2444, 2497, 2499, 2498, 2498, 2495, 2494, 2500, 2502, 2503, 2511, 2509, 2508, 2500, 2501, 2493, 2507, 2499, 2497, 2492, 2493, 2497, 2492, 2496, 2494, 2505, 2510, 2508, 2503, 2505, 2504, 2508, 2509, 2507, 2501, 2504, 2508, 2506, 2497, 2493,
2515, 2517, 2518, 2522, 2520, 2519, 2522, 2524, 2523, 2529, 2530, 2528, 2536, 2534, 2533, 2532, 2535, 2533, 2527, 2528, 2518, 2526, 2532, 2531, 2519, 2520, 2513, 2520, 2523, 2514, 2533, 2534, 2530, 2527, 2531, 2533, 2514, 2517, 2515, 2526, 2525, 2523, 2523, 2525, 2517, 2517, 2525, 2531, 2539, 2540, 2538, 2543, 2544, 2542, 2546, 2548, 2547, 2550, 2548, 2546, 2553, 2554, 2552, 2557, 2558, 2556, 2551, 2552, 2540, 2543, 2555, 2556, 2541, 2542, 2546, 2549, 2546, 2542, 2538, 2540, 2543, 2554, 2557, 2555, 2552, 2555, 2543, 2567, 2568, 2566, 2570, 2572, 2571, 2574, 2572, 2570, 2571, 2562, 2561, 2566, 2570, 2569, 2568, 2573, 2570, 2569, 2561, 2559, 2537, 2538, 2503, 2547, 2511, 2510, 2538, 2541, 2505, 2541, 2545, 2510, 2536, 2564, 2563, 2534, 2563, 2560, 2561, 2562, 2550, 2556, 2558, 2567, 2544, 2559, 2561, 2556, 2565, 2559, 2581, 2579, 2578, 2581, 2583, 2582, 2579, 2576, 2575, 2576, 2579, 2582, 2584, 2586, 2587, 2590, 2591, 2589, 2597, 2599, 2600, 2604, 2602, 2601, 2604, 2606, 2605, 2594, 2595, 2587, 2597, 2598, 2591, 2593, 2603, 2601, 2601, 2602, 2595, 2602, 2605, 2596,
2587, 2590, 2588, 2596, 2599, 2597, 2608, 2607, 2605, 2605, 2607, 2599, 2587, 2595, 2597, 2592, 2601, 2594, 2615, 2613, 2612, 2615, 2617, 2616, 2617, 2620, 2619, 2612, 2613, 2610, 2610, 2613, 2616, 2611, 2616, 2619, 2609, 2610, 2600, 2608, 2614, 2612, 2607, 2612, 2609, 2577, 2586, 2584, 2593, 2592, 2582, 2582, 2592, 2586, 2623, 2624, 2622, 2627, 2628, 2626, 2631, 2632, 2630, 2635, 2636, 2634, 2629, 2630, 2624, 2633, 2634, 2628, 2624, 2627, 2625, 2632, 2635, 2633, 2630, 2633, 2627, 2637, 2639, 2640, 2649, 2651, 2650, 2650, 2647, 2646, 2654, 2655, 2653, 2658, 2659, 2657, 2663, 2661, 2660, 2652, 2653, 2640, 2656, 2657, 2643, 2661, 2651, 2649, 2642, 2643, 2649, 2648, 2646, 2641, 2659, 2662, 2660, 2640, 2642, 2641, 2646, 2647, 2645, 2655, 2658, 2656, 2638, 2641, 2646, 2640, 2653, 2656, 2657, 2660, 2649, 2666, 2667, 2665, 2670, 2671, 2669, 2674, 2675, 2673, 2678, 2679, 2677, 2685, 2683, 2682, 2681, 2684, 2682, 2666, 2672, 2673, 2676, 2677, 2671, 2682, 2683, 2675, 2680, 2682, 2674, 2667, 2670, 2668, 2673, 2675, 2678, 2673, 2676, 2670, 2688, 2689, 2687, 2692, 2693, 2691,
2696, 2697, 2695, 2700, 2701, 2699, 2694, 2695, 2689, 2698, 2699, 2693, 2689, 2692, 2690, 2697, 2700, 2698, 2695, 2698, 2692, 2664, 2665, 2632, 2668, 2669, 2636, 2665, 2668, 2635, 2686, 2687, 2655, 2658, 2690, 2691, 2687, 2690, 2658, 2628, 2639, 2637, 2636, 2654, 2652, 2634, 2652, 2639, 2671, 2688, 2686, 2679, 2696, 2694, 2677, 2694, 2688, 2669, 2686, 2654, 2575, 2576, 2468, 2474, 2580, 2578, 2467, 2473, 2578, 2584, 2585, 2486, 2588, 2589, 2490, 2585, 2588, 2489, 2468, 2576, 2584, 2621, 2622, 2554, 2625, 2626, 2558, 2622, 2625, 2557, 2637, 2638, 2568, 2645, 2574, 2573, 2638, 2644, 2573, 2626, 2637, 2567, 2441, 2492, 2491, 2449, 2502, 2500, 2441, 2447, 2500, 2482, 2539, 2537, 2490, 2553, 2551, 2488, 2551, 2539, 2480, 2537, 2502, 2591, 2623, 2621, 2600, 2631, 2629, 2598, 2629, 2623, 2611, 2666, 2664, 2681, 2680, 2619, 2680, 2672, 2618, 2618, 2672, 2666, 2600, 2610, 2664, 2589, 2621, 2553, 2391, 2392, 2059, 2394, 2395, 2063, 2065, 2400, 2398, 2059, 2392, 2394, 2064, 2398, 2391, 2418, 2419, 2282, 2422, 2423, 2286, 2419, 2422, 2285, 2434, 2435, 2301, 2304, 2438, 2439,
2435, 2438, 2304, 2423, 2434, 2300, 2495, 2368, 2367, 2491, 2494, 2367, 2512, 2513, 2384, 2515, 2516, 2388, 2390, 2521, 2519, 2384, 2513, 2515, 2389, 2519, 2512, 2439, 2491, 2364, 2079, 2080, 1951, 1947, 2076, 2079, 1975, 2104, 2103, 1983, 2114, 2112, 1975, 1981, 2112, 1973, 2103, 2076, 2020, 2233, 2231, 2028, 2247, 2245, 2026, 2245, 2233, 2055, 2273, 2271, 2063, 2281, 2279, 2061, 2279, 2273, 2053, 2271, 2247, 2018, 2231, 2114, 2397, 2420, 2418, 2408, 2428, 2426, 2397, 2407, 2426, 2459, 2458, 2416, 2416, 2458, 2450, 2415, 2450, 2428, 2395, 2418, 2281, 2705, 2706, 2704, 2710, 2712, 2711, 2715, 2716, 2714, 2722, 2720, 2719, 2718, 2721, 2719, 2713, 2714, 2706, 2712, 2718, 2717, 2709, 2711, 2705, 2719, 2720, 2716, 2713, 2717, 2719, 2710, 2709, 2707, 2707, 2709, 2703, 2705, 2711, 2717, 2725, 2726, 2724, 2729, 2730, 2728, 2733, 2734, 2732, 2737, 2738, 2736, 2731, 2732, 2726, 2735, 2736, 2730, 2726, 2729, 2727, 2734, 2737, 2735, 2732, 2735, 2729, 2742, 2744, 2743, 2739, 2741, 2743, 2745, 2747, 2748, 2751, 2752, 2750, 2757, 2758, 2756, 2761, 2762, 2760, 2768, 2766, 2765,
2764, 2767, 2765, 2755, 2756, 2748, 2759, 2760, 2752, 2754, 2764, 2763, 2765, 2766, 2758, 2755, 2763, 2765, 2748, 2751, 2749, 2756, 2758, 2761, 2756, 2759, 2751, 2747, 2753, 2763, 2722, 2742, 2741, 2716, 2720, 2741, 2745, 2746, 2734, 2749, 2750, 2738, 2746, 2749, 2737, 2706, 2725, 2723, 2714, 2716, 2733, 2714, 2731, 2725, 2740, 2747, 2745, 2754, 2753, 2743, 2743, 2753, 2747, 2739, 2745, 2733, 2774, 2772, 2771, 2777, 2778, 2776, 2780, 2782, 2781, 2784, 2782, 2780, 2781, 2774, 2773, 2770, 2773, 2771, 2775, 2776, 2780, 2778, 2783, 2780, 2779, 2773, 2770, 2785, 2787, 2788, 2796, 2794, 2793, 2799, 2800, 2798, 2803, 2804, 2802, 2806, 2808, 2807, 2810, 2808, 2806, 2797, 2798, 2788, 2807, 2796, 2795, 2790, 2795, 2793, 2801, 2802, 2806, 2809, 2806, 2802, 2788, 2790, 2789, 2793, 2794, 2792, 2798, 2800, 2803, 2786, 2789, 2793, 2798, 2801, 2790, 2801, 2805, 2795, 2785, 2786, 2778, 2792, 2784, 2783, 2786, 2791, 2783, 2812, 2813, 2810, 2811, 2812, 2809, 2816, 2818, 2819, 2821, 2823, 2822, 2832, 2830, 2829, 2832, 2834, 2833, 2826, 2827, 2819, 2825, 2831, 2829, 2814, 2820, 2822,
2829, 2830, 2827, 2830, 2833, 2828, 2815, 2818, 2816, 2825, 2824, 2822, 2824, 2818, 2815, 2824, 2829, 2826, 2837, 2838, 2836, 2841, 2842, 2840, 2843, 2845, 2846, 2849, 2850, 2848, 2843, 2844, 2838, 2847, 2848, 2842, 2838, 2841, 2839, 2846, 2849, 2847, 2844, 2847, 2841, 2859, 2861, 2860, 2860, 2857, 2856, 2862, 2864, 2865, 2873, 2871, 2870, 2862, 2863, 2855, 2869, 2861, 2859, 2854, 2855, 2859, 2854, 2858, 2856, 2867, 2872, 2870, 2865, 2867, 2866, 2870, 2871, 2869, 2863, 2866, 2870, 2868, 2859, 2855, 2876, 2877, 2875, 2880, 2881, 2879, 2887, 2885, 2884, 2883, 2886, 2884, 2891, 2892, 2890, 2900, 2898, 2897, 2896, 2899, 2897, 2880, 2889, 2890, 2887, 2894, 2893, 2884, 2885, 2877, 2874, 2882, 2884, 2897, 2898, 2892, 2895, 2897, 2891, 2875, 2877, 2880, 2896, 2895, 2893, 2893, 2895, 2889, 2888, 2889, 2880, 2885, 2893, 2888, 2903, 2904, 2902, 2907, 2908, 2906, 2910, 2912, 2911, 2914, 2912, 2910, 2917, 2918, 2916, 2921, 2922, 2920, 2915, 2916, 2904, 2907, 2919, 2920, 2905, 2906, 2910, 2913, 2910, 2906, 2902, 2904, 2907, 2918, 2921, 2919, 2916, 2919, 2907, 2874, 2875, 2846,
2878, 2879, 2850, 2852, 2883, 2882, 2875, 2878, 2849, 2845, 2851, 2882, 2901, 2902, 2865, 2911, 2873, 2872, 2902, 2905, 2867, 2905, 2909, 2872, 2842, 2854, 2853, 2850, 2864, 2862, 2842, 2848, 2862, 2881, 2903, 2901, 2890, 2892, 2917, 2890, 2915, 2903, 2879, 2901, 2864, 2816, 2817, 2762, 2768, 2821, 2820, 2814, 2816, 2761, 2758, 2766, 2820, 2835, 2836, 2800, 2803, 2839, 2840, 2836, 2839, 2803, 2857, 2813, 2812, 2853, 2856, 2812, 2840, 2853, 2811, 2730, 2770, 2769, 2736, 2738, 2777, 2736, 2775, 2770, 2752, 2787, 2785, 2762, 2799, 2797, 2760, 2797, 2787, 2750, 2785, 2777, 2819, 2837, 2835, 2828, 2845, 2843, 2852, 2851, 2833, 2833, 2851, 2845, 2819, 2827, 2843, 2817, 2835, 2799, 2928, 2929, 2927, 2932, 2933, 2931, 2939, 2937, 2936, 2935, 2938, 2936, 2943, 2944, 2942, 2952, 2950, 2949, 2948, 2951, 2949, 2932, 2941, 2942, 2939, 2946, 2945, 2937, 2929, 2928, 2934, 2936, 2928, 2950, 2944, 2943, 2941, 2947, 2949, 2927, 2929, 2932, 2948, 2947, 2945, 2947, 2941, 2940, 2940, 2941, 2932, 2929, 2937, 2945, 2952, 2955, 2954, 2944, 2950, 2954, 2935, 2934, 2924, 2924, 2934, 2926,
2900, 2925, 2924, 2898, 2924, 2923, 2917, 2926, 2927, 2930, 2931, 2922, 2927, 2930, 2921, 2923, 2926, 2917, 2961, 2962, 2960, 2965, 2967, 2966, 2969, 2967, 2965, 2966, 2958, 2957, 2959, 2960, 2965, 2968, 2965, 2960, 2970, 2971, 2969, 2963, 2970, 2968, 2964, 2957, 2956, 2975, 2976, 2974, 2984, 2982, 2981, 2980, 2983, 2981, 2988, 2990, 2989, 2984, 2988, 2987, 2981, 2982, 2976, 2979, 2981, 2975, 2985, 2987, 2989, 2980, 2979, 2977, 2977, 2979, 2973, 2976, 2982, 2987, 2993, 2994, 2992, 2997, 2998, 2996, 2999, 3001, 3002, 3005, 3006, 3004, 2999, 3000, 2994, 3003, 3004, 2998, 2994, 2997, 2995, 3002, 3005, 3003, 3000, 3003, 2997, 3012, 3014, 3015, 3019, 3017, 3016, 3019, 3021, 3020, 3026, 3027, 3025, 3033, 3031, 3030, 3029, 3032, 3030, 3024, 3025, 3015, 3023, 3029, 3028, 3016, 3017, 3010, 3017, 3020, 3011, 3030, 3031, 3027, 3024, 3028, 3030, 3011, 3014, 3012, 3023, 3022, 3020, 3020, 3022, 3014, 3014, 3022, 3028, 3009, 3010, 3002, 3012, 3013, 3006, 3008, 3018, 3016, 3002, 3010, 3012, 3007, 3016, 3009, 2974, 2976, 2993, 2986, 3001, 2999, 3008, 3007, 2989, 2989, 3007, 3001,
2985, 2999, 2993, 3033, 3036, 3035, 3031, 3035, 3034, 3039, 3040, 3038, 3043, 3044, 3042, 3046, 3048, 3047, 3050, 3048, 3046, 3053, 3054, 3052, 3057, 3058, 3056, 3051, 3052, 3040, 3043, 3055, 3056, 3042, 3046, 3045, 3044, 3049, 3046, 3038, 3040, 3043, 3054, 3057, 3055, 3052, 3055, 3043, 3062, 3064, 3063, 3065, 3067, 3068, 3077, 3079, 3078, 3078, 3075, 3074, 3065, 3066, 3060, 3073, 3064, 3062, 3059, 3060, 3062, 3071, 3077, 3076, 3070, 3076, 3074, 3068, 3070, 3069, 3074, 3075, 3073, 3069, 3074, 3072, 3072, 3062, 3060, 3082, 3083, 3081, 3084, 3086, 3087, 3093, 3091, 3090, 3089, 3092, 3090, 3102, 3100, 3099, 3102, 3104, 3103, 3096, 3097, 3087, 3095, 3101, 3099, 3090, 3091, 3083, 3088, 3090, 3082, 3100, 3097, 3096, 3097, 3100, 3103, 3083, 3086, 3084, 3095, 3094, 3091, 3091, 3094, 3086, 3086, 3094, 3099, 3107, 3108, 3106, 3109, 3111, 3112, 3117, 3118, 3116, 3121, 3122, 3120, 3126, 3124, 3123, 3115, 3116, 3108, 3119, 3120, 3112, 3123, 3124, 3118, 3108, 3111, 3109, 3116, 3118, 3121, 3116, 3119, 3111, 3053, 3080, 3081, 3084, 3085, 3058, 3081, 3084, 3057, 3105, 3106, 3068,
3109, 3110, 3071, 3114, 3079, 3077, 3068, 3106, 3109, 3110, 3113, 3077, 3061, 3063, 3050, 3058, 3067, 3065, 3044, 3059, 3061, 3056, 3065, 3059, 3087, 3107, 3105, 3098, 3117, 3115, 3125, 3123, 3103, 3103, 3123, 3117, 3087, 3097, 3115, 3085, 3105, 3067, 3129, 3130, 3128, 3134, 3136, 3135, 3138, 3136, 3134, 3140, 3142, 3141, 3145, 3146, 3144, 3147, 3149, 3150, 3129, 3143, 3144, 3147, 3148, 3132, 3152, 3142, 3140, 3127, 3128, 3134, 3130, 3137, 3134, 3131, 3132, 3140, 3139, 3141, 3138, 3146, 3149, 3147, 3130, 3131, 3139, 3144, 3147, 3131, 3148, 3151, 3140, 3155, 3156, 3154, 3160, 3162, 3161, 3164, 3162, 3160, 3166, 3168, 3167, 3153, 3154, 3160, 3156, 3163, 3160, 3158, 3166, 3165, 3165, 3167, 3164, 3156, 3157, 3165, 3150, 3155, 3153, 3159, 3161, 3152, 3148, 3153, 3159, 3172, 3173, 3171, 3181, 3179, 3178, 3177, 3180, 3178, 3178, 3179, 3173, 3176, 3178, 3172, 3177, 3176, 3174, 3176, 3170, 3169, 3182, 3184, 3185, 3188, 3189, 3187, 3200, 3198, 3197, 3204, 3202, 3201, 3204, 3206, 3205, 3192, 3193, 3185, 3194, 3195, 3189, 3191, 3199, 3197, 3197, 3198, 3193, 3202, 3195, 3194,
3195, 3202, 3205, 3185, 3188, 3186, 3203, 3201, 3198, 3198, 3201, 3194, 3193, 3194, 3188, 3184, 3190, 3197, 3209, 3210, 3208, 3211, 3213, 3214, 3219, 3220, 3218, 3223, 3224, 3222, 3228, 3226, 3225, 3217, 3218, 3210, 3221, 3222, 3214, 3225, 3226, 3220, 3210, 3213, 3211, 3218, 3220, 3223, 3218, 3221, 3213, 3237, 3235, 3234, 3234, 3235, 3231, 3236, 3234, 3232, 3232, 3234, 3230, 3230, 3231, 3224, 3228, 3233, 3232, 3229, 3230, 3223, 3220, 3226, 3232, 3189, 3209, 3207, 3196, 3219, 3217, 3227, 3225, 3205, 3225, 3219, 3196, 3189, 3195, 3217, 3121, 3170, 3171, 3126, 3175, 3174, 3169, 3170, 3121, 3118, 3124, 3174, 3182, 3183, 3146, 3186, 3187, 3150, 3183, 3186, 3149, 3155, 3207, 3208, 3211, 3212, 3158, 3216, 3168, 3166, 3208, 3211, 3157, 3212, 3215, 3166, 3187, 3207, 3155, 3112, 3129, 3127, 3133, 3135, 3114, 3122, 3145, 3143, 3110, 3127, 3133, 3120, 3143, 3129, 3173, 3184, 3182, 3191, 3190, 3179, 3179, 3190, 3184, 3171, 3182, 3145, 3240, 3241, 3239, 3244, 3245, 3243, 3247, 3249, 3248, 3251, 3253, 3252, 3255, 3253, 3251, 3258, 3259, 3257, 3262, 3263, 3261, 3256, 3257, 3241,
3244, 3260, 3261, 3238, 3239, 3247, 3243, 3251, 3250, 3245, 3254, 3251, 3241, 3244, 3242, 3250, 3252, 3249, 3259, 3262, 3260, 3239, 3242, 3250, 3257, 3260, 3244, 3269, 3271, 3270, 3273, 3275, 3274, 3278, 3279, 3277, 3280, 3282, 3283, 3276, 3277, 3265, 3280, 3281, 3267, 3285, 3275, 3273, 3264, 3265, 3269, 3267, 3273, 3272, 3272, 3274, 3271, 3279, 3282, 3280, 3265, 3266, 3272, 3277, 3280, 3266, 3281, 3284, 3273, 3291, 3292, 3290, 3296, 3294, 3293, 3296, 3298, 3297, 3302, 3300, 3299, 3294, 3287, 3286, 3287, 3294, 3297, 3299, 3300, 3292, 3288, 3291, 3289, 3301, 3299, 3297, 3299, 3291, 3288, 3305, 3306, 3304, 3307, 3309, 3310, 3314, 3312, 3311, 3322, 3320, 3319, 3317, 3318, 3310, 3316, 3321, 3319, 3311, 3312, 3306, 3320, 3318, 3317, 3306, 3309, 3307, 3316, 3315, 3312, 3315, 3309, 3306, 3309, 3315, 3319, 3286, 3287, 3259, 3289, 3290, 3263, 3259, 3287, 3289, 3303, 3304, 3279, 3307, 3308, 3283, 3304, 3307, 3282, 3268, 3270, 3255, 3263, 3278, 3276, 3245, 3264, 3268, 3261, 3276, 3264, 3292, 3305, 3303, 3313, 3311, 3300, 3300, 3311, 3305, 3290, 3303, 3278, 3325, 3326, 3324,
3327, 3329, 3330, 3332, 3334, 3333, 3336, 3338, 3337, 3323, 3324, 3332, 3327, 3328, 3336, 3324, 3326, 3329, 3335, 3337, 3334, 3324, 3327, 3335, 3341, 3342, 3340, 3345, 3346, 3344, 3348, 3350, 3349, 3352, 3354, 3353, 3339, 3340, 3348, 3344, 3352, 3351, 3342, 3345, 3343, 3351, 3353, 3350, 3343, 3351, 3348, 3357, 3358, 3356, 3361, 3362, 3360, 3370, 3368, 3367, 3374, 3372, 3371, 3363, 3364, 3358, 3361, 3365, 3366, 3367, 3368, 3364, 3371, 3372, 3366, 3358, 3361, 3359, 3373, 3371, 3368, 3368, 3371, 3365, 3358, 3364, 3365, 3377, 3378, 3376, 3381, 3382, 3380, 3390, 3388, 3387, 3394, 3392, 3391, 3377, 3383, 3384, 3385, 3386, 3382, 3388, 3384, 3383, 3392, 3386, 3385, 3378, 3381, 3379, 3393, 3391, 3388, 3391, 3385, 3384, 3378, 3384, 3385, 3355, 3356, 3326, 3359, 3360, 3330, 3356, 3359, 3329, 3375, 3376, 3342, 3379, 3380, 3346, 3376, 3379, 3345, 3328, 3330, 3341, 3347, 3349, 3338, 3328, 3339, 3347, 3362, 3377, 3375, 3389, 3387, 3372, 3387, 3383, 3366, 3366, 3383, 3377, 3360, 3375, 3341, 3283, 3325, 3323, 3331, 3333, 3285, 3323, 3331, 3284, 3310, 3357, 3355, 3369, 3367, 3320,
3320, 3367, 3363, 3318, 3363, 3357, 3308, 3355, 3325, 2961, 3037, 3038, 3047, 2971, 2970, 3038, 3041, 2963, 3041, 3045, 2970, 3214, 3240, 3238, 3246, 3248, 3216, 3224, 3258, 3256, 3238, 3246, 3215, 3222, 3256, 3240, 3295, 3293, 3235, 3235, 3293, 3286, 3231, 3286, 3258, 1824, 2703, 2704, 1829, 2708, 2707, 2702, 2703, 1824, 1828, 2707, 2702, 2723, 2724, 1847, 1850, 2727, 2728, 2724, 2727, 1850, 2704, 2723, 1846, 2772, 1858, 1857, 2769, 2771, 1857, 2728, 2769, 1856, 2678, 2973, 2974, 2685, 2978, 2977, 2972, 2973, 2678, 2683, 2977, 2972, 2991, 2992, 2697, 2995, 2996, 2701, 2992, 2995, 2700, 2974, 2991, 2696, 2957, 2958, 2914, 2920, 2922, 2961, 2908, 2956, 2957, 2920, 2959, 2956, 2933, 3039, 3037, 2942, 2944, 3053, 2942, 3051, 3039, 3089, 3088, 2954, 3088, 3080, 2953, 2953, 3080, 3053, 2931, 3037, 2961, 304, 1479, 1480, 311, 1488, 1487, 310, 1487, 1479, 1502, 1503, 323, 1506, 1507, 327, 1503, 1506, 326, 1480, 1502, 322, 1564, 1565, 384, 1574, 393, 392, 384, 1565, 1568, 1573, 392, 387, 1507, 1564, 383, 1135, 1862, 1861, 1129, 1133, 1861, 1865, 1866, 1147,
1869, 1870, 1151, 1866, 1869, 1150, 1859, 1865, 1146, 1911, 1912, 1209, 1915, 1916, 1213, 1912, 1915, 1212, 1225, 1927, 1928, 1935, 1233, 1232, 1928, 1931, 1227, 1931, 1934, 1232, 1916, 1927, 1225, 1870, 1911, 1208, 2066, 2067, 1367, 1373, 2071, 2070, 1366, 1372, 2070, 2085, 2086, 1385, 2089, 2090, 1389, 2086, 2089, 1388, 2067, 2085, 1384, 2150, 2151, 1427, 2154, 2155, 1431, 2159, 1437, 1436, 2151, 2154, 1430, 2155, 2158, 1436, 1449, 2176, 2175, 1448, 2175, 2170, 2090, 2150, 1426, 3403, 3405, 3406, 3408, 3410, 3409, 3412, 3414, 3413, 3398, 3408, 3407, 3413, 3400, 3399, 3401, 3407, 3409, 3403, 3404, 3412, 3399, 3400, 3398, 3402, 3405, 3403, 3416, 3415, 3409, 3396, 3399, 3397, 3415, 3405, 3402, 3401, 3403, 3396, 3395, 3397, 3407, 3403, 3411, 3399, 3423, 3425, 3424, 3424, 3421, 3420, 3419, 3423, 3422, 3418, 3422, 3420, 3432, 3430, 3429, 3432, 3434, 3433, 3429, 3430, 3427, 3427, 3430, 3433, 3435, 3437, 3438, 3441, 3442, 3440, 3448, 3450, 3449, 3452, 3450, 3448, 3457, 3458, 3456, 3462, 3460, 3459, 3468, 3466, 3465, 3464, 3467, 3465, 3453, 3454, 3438, 3441, 3455, 3456,
3446, 3461, 3459, 3459, 3460, 3454, 3439, 3440, 3448, 3442, 3451, 3448, 3465, 3466, 3458, 3455, 3463, 3465, 3438, 3441, 3439, 3447, 3449, 3444, 3464, 3463, 3460, 3436, 3439, 3447, 3460, 3463, 3455, 3454, 3455, 3441, 3445, 3459, 3453, 3426, 3427, 3406, 3416, 3431, 3429, 3415, 3429, 3426, 3435, 3436, 3419, 3444, 3425, 3423, 3436, 3443, 3423, 3406, 3418, 3417, 3420, 3421, 3414, 3404, 3417, 3420, 3428, 3437, 3435, 3446, 3445, 3433, 3445, 3437, 3428, 3406, 3427, 3435, 3472, 3474, 3473, 3475, 3477, 3478, 3489, 3491, 3490, 3490, 3487, 3486, 3475, 3476, 3470, 3483, 3474, 3472, 3469, 3470, 3472, 3480, 3481, 3489, 3480, 3488, 3486, 3478, 3480, 3479, 3486, 3487, 3483, 3479, 3486, 3482, 3476, 3482, 3472, 3493, 3494, 3491, 3481, 3492, 3493, 3471, 3473, 3452, 3458, 3477, 3475, 3485, 3484, 3466, 3442, 3469, 3471, 3466, 3484, 3477, 3456, 3475, 3469, 3498, 3500, 3499, 3503, 3504, 3502, 3506, 3508, 3507, 3501, 3502, 3496, 3499, 3506, 3505, 3496, 3498, 3497, 3505, 3507, 3503, 3497, 3505, 3501, 3511, 3512, 3510, 3514, 3516, 3515, 3522, 3520, 3519, 3517, 3518, 3512, 3516, 3521, 3519,
3513, 3515, 3511, 3520, 3518, 3517, 3515, 3519, 3517, 3509, 3510, 3504, 3508, 3514, 3513, 3507, 3513, 3509, 3528, 3530, 3529, 3532, 3534, 3533, 3537, 3538, 3536, 3541, 3542, 3540, 3535, 3536, 3524, 3539, 3540, 3526, 3524, 3528, 3527, 3526, 3532, 3531, 3531, 3533, 3530, 3538, 3541, 3539, 3525, 3531, 3528, 3536, 3539, 3525, 3548, 3550, 3549, 3552, 3554, 3553, 3557, 3558, 3556, 3561, 3562, 3560, 3555, 3556, 3544, 3559, 3560, 3546, 3544, 3548, 3547, 3546, 3552, 3551, 3551, 3553, 3550, 3558, 3561, 3559, 3545, 3551, 3548, 3556, 3559, 3545, 3565, 3566, 3564, 3569, 3570, 3568, 3578, 3576, 3575, 3582, 3580, 3579, 3571, 3572, 3566, 3573, 3574, 3570, 3576, 3572, 3571, 3580, 3574, 3573, 3566, 3569, 3567, 3581, 3579, 3576, 3579, 3573, 3572, 3572, 3573, 3569, 3585, 3586, 3584, 3589, 3590, 3588, 3598, 3596, 3595, 3602, 3600, 3599, 3591, 3592, 3586, 3593, 3594, 3590, 3596, 3592, 3591, 3600, 3594, 3593, 3586, 3589, 3587, 3601, 3599, 3596, 3599, 3593, 3592, 3592, 3593, 3589, 3563, 3564, 3538, 3567, 3568, 3542, 3564, 3567, 3541, 3583, 3584, 3558, 3587, 3588, 3562, 3584, 3587, 3561,
3547, 3549, 3534, 3542, 3557, 3555, 3543, 3547, 3532, 3540, 3555, 3543, 3570, 3585, 3583, 3597, 3595, 3580, 3595, 3591, 3574, 3574, 3591, 3585, 3568, 3583, 3557, 3527, 3529, 3500, 3504, 3537, 3535, 3523, 3527, 3498, 3502, 3535, 3523, 3512, 3565, 3563, 3577, 3575, 3520, 3575, 3571, 3518, 3518, 3571, 3565, 3510, 3563, 3537, 3608, 3610, 3609, 3614, 3612, 3611, 3622, 3624, 3623, 3630, 3632, 3631, 3631, 3627, 3625, 3628, 3632, 3630, 3610, 3622, 3621, 3627, 3614, 3613, 3607, 3609, 3604, 3621, 3623, 3616, 3606, 3613, 3611, 3619, 3620, 3630, 3619, 3629, 3625, 3626, 3630, 3620, 3604, 3606, 3605, 3611, 3612, 3608, 3615, 3616, 3619, 3603, 3605, 3611, 3604, 3615, 3617, 3609, 3621, 3615, 3617, 3625, 3613, 3639, 3640, 3638, 3644, 3646, 3645, 3645, 3649, 3647, 3635, 3646, 3644, 3637, 3638, 3644, 3637, 3643, 3647, 3647, 3649, 3642, 3639, 3647, 3641, 3634, 3644, 3638, 3651, 3653, 3654, 3657, 3658, 3656, 3660, 3662, 3661, 3661, 3664, 3663, 3667, 3669, 3670, 3678, 3676, 3675, 3678, 3674, 3673, 3682, 3680, 3679, 3667, 3668, 3654, 3657, 3671, 3672, 3682, 3666, 3665, 3651, 3652, 3660,
3659, 3663, 3653, 3676, 3670, 3669, 3676, 3673, 3668, 3680, 3672, 3671, 3652, 3654, 3657, 3681, 3679, 3673, 3679, 3671, 3668, 3654, 3668, 3671, 3680, 3665, 3658, 3619, 3655, 3656, 3624, 3662, 3660, 3616, 3652, 3655, 3616, 3623, 3660, 3663, 3664, 3635, 3640, 3669, 3667, 3677, 3675, 3648, 3633, 3653, 3663, 3675, 3669, 3640, 3633, 3638, 3667, 3688, 3689, 3687, 3693, 3691, 3690, 3697, 3699, 3698, 3700, 3702, 3703, 3704, 3706, 3707, 3709, 3711, 3710, 3715, 3713, 3712, 3704, 3705, 3689, 3695, 3711, 3709, 3690, 3691, 3684, 3701, 3709, 3708, 3686, 3696, 3698, 3713, 3707, 3706, 3697, 3696, 3691, 3698, 3699, 3695, 3703, 3706, 3704, 3696, 3686, 3684, 3685, 3688, 3698, 3685, 3701, 3704, 3694, 3709, 3701, 3721, 3719, 3718, 3727, 3725, 3724, 3725, 3721, 3720, 3717, 3720, 3718, 3726, 3724, 3722, 3724, 3720, 3717, 3735, 3733, 3732, 3737, 3734, 3732, 3733, 3729, 3728, 3736, 3732, 3728, 3739, 3738, 3733, 3738, 3731, 3729, 3702, 3728, 3729, 3739, 3714, 3712, 3703, 3729, 3731, 3738, 3712, 3706, 3689, 3717, 3716, 3705, 3707, 3723, 3727, 3726, 3713, 3726, 3723, 3707, 3689, 3705, 3722,
3748, 3746, 3745, 3746, 3742, 3741, 3747, 3745, 3743, 3745, 3741, 3740, 3741, 3742, 3640, 3642, 3744, 3743, 3748, 3650, 3648, 3636, 3740, 3741, 3636, 3641, 3743, 3746, 3648, 3640, 3618, 3620, 3702, 3708, 3710, 3628, 3700, 3708, 3626, 3656, 3658, 3730, 3737, 3736, 3665, 3665, 3736, 3730, 3656, 3728, 3702, 3755, 3757, 3756, 3760, 3762, 3763, 3765, 3767, 3766, 3760, 3761, 3751, 3753, 3765, 3764, 3764, 3766, 3759, 3751, 3755, 3754, 3754, 3756, 3753, 3759, 3762, 3760, 3750, 3754, 3752, 3758, 3760, 3750, 3752, 3764, 3758, 3775, 3777, 3776, 3779, 3781, 3780, 3784, 3785, 3783, 3788, 3789, 3787, 3793, 3791, 3790, 3797, 3795, 3794, 3782, 3783, 3771, 3786, 3787, 3773, 3771, 3775, 3774, 3791, 3785, 3784, 3773, 3779, 3778, 3795, 3789, 3788, 3778, 3780, 3777, 3785, 3788, 3786, 3796, 3794, 3791, 3772, 3778, 3775, 3794, 3788, 3785, 3783, 3786, 3772, 3800, 3802, 3803, 3805, 3807, 3806, 3811, 3809, 3808, 3814, 3816, 3817, 3819, 3821, 3820, 3825, 3823, 3822, 3802, 3814, 3815, 3807, 3819, 3818, 3823, 3811, 3810, 3804, 3806, 3799, 3812, 3818, 3820, 3803, 3810, 3808, 3817, 3824, 3822,
3799, 3802, 3800, 3813, 3816, 3814, 3812, 3814, 3802, 3799, 3806, 3818, 3815, 3822, 3810, 3762, 3800, 3801, 3767, 3805, 3804, 3809, 3769, 3768, 3798, 3800, 3762, 3766, 3804, 3798, 3808, 3768, 3763, 3774, 3776, 3757, 3763, 3784, 3782, 3792, 3790, 3768, 3770, 3774, 3755, 3790, 3784, 3763, 3761, 3782, 3770, 3831, 3833, 3832, 3835, 3837, 3836, 3840, 3841, 3839, 3844, 3845, 3843, 3849, 3847, 3846, 3853, 3851, 3850, 3838, 3839, 3827, 3842, 3843, 3829, 3827, 3831, 3830, 3847, 3841, 3840, 3829, 3835, 3834, 3851, 3845, 3844, 3834, 3836, 3833, 3841, 3844, 3842, 3852, 3850, 3847, 3828, 3834, 3831, 3850, 3844, 3841, 3839, 3842, 3828, 3857, 3859, 3858, 3862, 3863, 3861, 3867, 3865, 3864, 3860, 3861, 3855, 3865, 3859, 3857, 3855, 3857, 3856, 3866, 3864, 3861, 3864, 3857, 3855, 3871, 3873, 3872, 3877, 3879, 3878, 3873, 3877, 3876, 3870, 3872, 3869, 3876, 3878, 3875, 3872, 3876, 3874, 3880, 3882, 3883, 3887, 3885, 3884, 3890, 3891, 3889, 3895, 3893, 3892, 3888, 3889, 3883, 3893, 3887, 3886, 3886, 3884, 3881, 3894, 3892, 3889, 3892, 3886, 3883, 3853, 3871, 3870, 3851, 3870, 3868,
3862, 3880, 3881, 3885, 3867, 3866, 3884, 3866, 3863, 3856, 3858, 3837, 3845, 3862, 3860, 3854, 3856, 3835, 3843, 3860, 3854, 3869, 3882, 3880, 3874, 3875, 3890, 3869, 3874, 3888, 3868, 3880, 3862, 3899, 3900, 3898, 3905, 3909, 3907, 3910, 3906, 3904, 3913, 3914, 3912, 3920, 3918, 3917, 3916, 3919, 3917, 3899, 3911, 3912, 3909, 3916, 3915, 3903, 3907, 3899, 3908, 3904, 3898, 3917, 3918, 3914, 3911, 3915, 3917, 3905, 3903, 3901, 3901, 3903, 3897, 3899, 3907, 3915, 3926, 3927, 3925, 3931, 3933, 3932, 3935, 3933, 3931, 3937, 3939, 3938, 3932, 3923, 3922, 3925, 3931, 3930, 3927, 3934, 3931, 3928, 3929, 3937, 3936, 3938, 3935, 3927, 3928, 3936, 3924, 3930, 3922, 3945, 3946, 3944, 3949, 3950, 3948, 3956, 3954, 3953, 3952, 3955, 3953, 3960, 3958, 3957, 3953, 3954, 3946, 3951, 3953, 3945, 3957, 3958, 3950, 3946, 3949, 3947, 3959, 3957, 3954, 3954, 3957, 3949, 3920, 3942, 3941, 3914, 3918, 3941, 3926, 3943, 3944, 3947, 3948, 3929, 3944, 3947, 3928, 3922, 3923, 3910, 3912, 3914, 3926, 3900, 3921, 3922, 3912, 3924, 3921, 3952, 3951, 3941, 3941, 3951, 3943, 3940, 3943, 3926,
3967, 3969, 3968, 3965, 3969, 3967, 3975, 3976, 3974, 3980, 3982, 3981, 3978, 3982, 3980, 3962, 3973, 3974, 3968, 3984, 3983, 3971, 3972, 3980, 3977, 3980, 3972, 3963, 3967, 3966, 3964, 3967, 3963, 3970, 3972, 3975, 3983, 3984, 3978, 3970, 3973, 3983, 3962, 3966, 3983, 3987, 3989, 3990, 3996, 3994, 3993, 3996, 3992, 3991, 4002, 4000, 3999, 4000, 3995, 3993, 3994, 3990, 3989, 3990, 3994, 3991, 4001, 3999, 3997, 3997, 3999, 3993, 4003, 4005, 4006, 4013, 4011, 4010, 4017, 4015, 4014, 4019, 4016, 4014, 4011, 4006, 4005, 4014, 4015, 4008, 4018, 4014, 4007, 4004, 4006, 4009, 4019, 4018, 4011, 4011, 4018, 4009, 3971, 4003, 4004, 3975, 4007, 4008, 4017, 3986, 3985, 4004, 4007, 3975, 4008, 4015, 3985, 3963, 3989, 3987, 3974, 3976, 3998, 4002, 4001, 3985, 4001, 3998, 3976, 3974, 3997, 3989, 3816, 3897, 3898, 3821, 3902, 3901, 3906, 3825, 3824, 3896, 3897, 3816, 3820, 3901, 3896, 3898, 3904, 3824, 3879, 3965, 3964, 3878, 3964, 3961, 3890, 3987, 3988, 3992, 3895, 3894, 3991, 3894, 3891, 3875, 3961, 3987, 3830, 3832, 3781, 3789, 3840, 3838, 3848, 3846, 3795, 3826, 3830, 3779,
3846, 3840, 3789, 3787, 3838, 3826, 3979, 3981, 3939, 3971, 3979, 3937, 3950, 4005, 4003, 4012, 4010, 3958, 4010, 4005, 3950, 3929, 3948, 4003, 4023, 4025, 4024, 4028, 4029, 4027, 4031, 4033, 4032, 4026, 4027, 4021, 4024, 4031, 4030, 4021, 4023, 4022, 4030, 4032, 4028, 4022, 4030, 4026, 4036, 4037, 4035, 4039, 4041, 4040, 4044, 4045, 4043, 4047, 4049, 4048, 4042, 4043, 4037, 4041, 4047, 4046, 4038, 4040, 4036, 4046, 4048, 4044, 4040, 4046, 4042, 4034, 4035, 4029, 4033, 4039, 4038, 4032, 4038, 4034, 4055, 4057, 4056, 4059, 4061, 4060, 4064, 4065, 4063, 4068, 4069, 4067, 4062, 4063, 4051, 4066, 4067, 4053, 4051, 4055, 4054, 4053, 4059, 4058, 4058, 4060, 4057, 4065, 4068, 4066, 4052, 4058, 4055, 4063, 4066, 4052, 4075, 4077, 4076, 4079, 4081, 4080, 4084, 4085, 4083, 4088, 4089, 4087, 4082, 4083, 4071, 4086, 4087, 4073, 4071, 4075, 4074, 4073, 4079, 4078, 4078, 4080, 4077, 4085, 4088, 4086, 4072, 4078, 4075, 4083, 4086, 4072, 4092, 4093, 4091, 4096, 4097, 4095, 4100, 4101, 4099, 4104, 4105, 4103, 4098, 4099, 4093, 4102, 4103, 4097, 4093, 4096, 4094, 4101, 4104, 4102,
4099, 4102, 4096, 4108, 4109, 4107, 4112, 4113, 4111, 4114, 4116, 4117, 4124, 4122, 4121, 4128, 4126, 4125, 4130, 4127, 4125, 4114, 4115, 4109, 4112, 4118, 4119, 4121, 4122, 4117, 4125, 4126, 4119, 4129, 4125, 4118, 4109, 4112, 4110, 4115, 4117, 4120, 4130, 4129, 4122, 4122, 4129, 4120, 4115, 4118, 4112, 4090, 4091, 4065, 4094, 4095, 4069, 4091, 4094, 4068, 4106, 4107, 4085, 4110, 4111, 4089, 4107, 4110, 4088, 4074, 4076, 4061, 4069, 4084, 4082, 4070, 4074, 4059, 4067, 4082, 4070, 4097, 4108, 4106, 4105, 4116, 4114, 4103, 4114, 4108, 4095, 4106, 4084, 4054, 4056, 4025, 4029, 4064, 4062, 4050, 4054, 4023, 4027, 4062, 4050, 4037, 4092, 4090, 4045, 4100, 4098, 4043, 4098, 4092, 4035, 4090, 4064, 3683, 3684, 3478, 3480, 3686, 3687, 3485, 3692, 3690, 3478, 3684, 3686, 3477, 3484, 3690, 3719, 3494, 3493, 3718, 3493, 3492, 3687, 3716, 3492, 4135, 4136, 4134, 4138, 4140, 4139, 4142, 4144, 4143, 4137, 4139, 4132, 4134, 4142, 4141, 4132, 4135, 4133, 4141, 4143, 4138, 4133, 4141, 4137, 4147, 4148, 4146, 4151, 4152, 4150, 4154, 4156, 4155, 4158, 4160, 4159, 4146, 4154, 4153,
4150, 4158, 4157, 4148, 4151, 4149, 4157, 4159, 4156, 4149, 4157, 4154, 4136, 4147, 4145, 4153, 4155, 4144, 4145, 4153, 4142, 4163, 4164, 4162, 4167, 4168, 4166, 4170, 4172, 4171, 4174, 4176, 4175, 4162, 4170, 4169, 4166, 4174, 4173, 4164, 4167, 4165, 4173, 4175, 4172, 4165, 4173, 4170, 4179, 4180, 4178, 4182, 4184, 4183, 4186, 4184, 4182, 4178, 4182, 4181, 4185, 4182, 4178, 4168, 4179, 4177, 4181, 4183, 4176, 4177, 4181, 4174, 4189, 4191, 4192, 4194, 4196, 4195, 4200, 4198, 4197, 4205, 4206, 4204, 4208, 4210, 4209, 4214, 4212, 4211, 4191, 4203, 4204, 4196, 4208, 4207, 4212, 4200, 4199, 4193, 4195, 4188, 4201, 4207, 4209, 4199, 4197, 4190, 4206, 4213, 4211, 4188, 4191, 4189, 4201, 4202, 4205, 4201, 4203, 4191, 4195, 4207, 4201, 4204, 4211, 4199, 4222, 4220, 4219, 4226, 4224, 4223, 4220, 4216, 4215, 4224, 4218, 4217, 4225, 4223, 4220, 4223, 4217, 4216, 4229, 4231, 4232, 4234, 4236, 4235, 4242, 4238, 4237, 4243, 4245, 4246, 4249, 4252, 4251, 4243, 4244, 4232, 4241, 4249, 4247, 4250, 4242, 4240, 4233, 4235, 4228, 4232, 4240, 4237, 4243, 4247, 4251, 4228, 4231, 4229,
4241, 4239, 4235, 4235, 4239, 4231, 4239, 4247, 4243, 4248, 4240, 4232, 4258, 4256, 4255, 4254, 4257, 4255, 4205, 4229, 4230, 4210, 4234, 4233, 4238, 4214, 4213, 4202, 4227, 4229, 4202, 4209, 4233, 4230, 4237, 4213, 4221, 4219, 4197, 4219, 4215, 4190, 4246, 4254, 4253, 4255, 4256, 4250, 4244, 4253, 4255, 4267, 4265, 4264, 4271, 4269, 4268, 4271, 4273, 4272, 4277, 4279, 4278, 4273, 4277, 4276, 4265, 4260, 4259, 4269, 4262, 4261, 4269, 4272, 4263, 4276, 4278, 4275, 4270, 4268, 4265, 4268, 4261, 4260, 4272, 4276, 4274, 4282, 4283, 4281, 4287, 4285, 4284, 4290, 4291, 4289, 4295, 4293, 4292, 4282, 4288, 4289, 4293, 4287, 4286, 4286, 4284, 4281, 4294, 4292, 4289, 4292, 4286, 4283, 4299, 4301, 4300, 4308, 4310, 4309, 4306, 4310, 4308, 4301, 4306, 4305, 4298, 4300, 4297, 4304, 4308, 4307, 4302, 4305, 4308, 4300, 4305, 4302, 4313, 4314, 4312, 4318, 4316, 4315, 4319, 4321, 4322, 4328, 4326, 4325, 4328, 4324, 4323, 4313, 4319, 4320, 4324, 4318, 4317, 4317, 4315, 4312, 4326, 4322, 4321, 4322, 4326, 4323, 4323, 4317, 4314, 4279, 4299, 4298, 4278, 4298, 4296, 4290, 4311, 4312,
4316, 4295, 4294, 4315, 4294, 4291, 4263, 4282, 4280, 4274, 4275, 4290, 4274, 4288, 4282, 4296, 4297, 4313, 4304, 4321, 4319, 4297, 4302, 4319, 4275, 4296, 4311, 4189, 4190, 4136, 4140, 4194, 4193, 4187, 4189, 4135, 4139, 4193, 4187, 4215, 4216, 4148, 4217, 4218, 4152, 4216, 4217, 4151, 4190, 4215, 4147, 4259, 4260, 4164, 4261, 4262, 4168, 4260, 4261, 4167, 4280, 4281, 4180, 4285, 4186, 4185, 4284, 4185, 4180, 4262, 4280, 4179, 4152, 4163, 4161, 4169, 4171, 4160, 4161, 4169, 4158, 4266, 4264, 4224, 4264, 4259, 4218, 4218, 4259, 4163, 4331, 4332, 4330, 4334, 4336, 4335, 4339, 4340, 4338, 4342, 4344, 4343, 4337, 4338, 4332, 4336, 4342, 4341, 4333, 4335, 4331, 4341, 4343, 4339, 4335, 4341, 4337, 4347, 4348, 4346, 4350, 4352, 4351, 4355, 4356, 4354, 4358, 4360, 4359, 4353, 4354, 4348, 4352, 4358, 4357, 4349, 4351, 4347, 4357, 4359, 4355, 4351, 4357, 4353, 4345, 4346, 4340, 4344, 4350, 4349, 4343, 4349, 4345, 4363, 4364, 4362, 4367, 4368, 4366, 4371, 4372, 4370, 4375, 4376, 4374, 4369, 4370, 4364, 4373, 4374, 4368, 4364, 4367, 4365, 4372, 4375, 4373, 4370, 4373, 4367,
4382, 4380, 4379, 4388, 4386, 4385, 4386, 4382, 4381, 4381, 4379, 4377, 4387, 4385, 4383, 4385, 4381, 4378, 4391, 4392, 4390, 4395, 4396, 4394, 4399, 4400, 4398, 4403, 4404, 4402, 4397, 4398, 4392, 4401, 4402, 4396, 4392, 4395, 4393, 4400, 4403, 4401, 4398, 4401, 4395, 4410, 4408, 4407, 4416, 4414, 4413, 4414, 4410, 4409, 4409, 4407, 4405, 4415, 4413, 4411, 4413, 4409, 4406, 4389, 4390, 4372, 4393, 4394, 4376, 4390, 4393, 4375, 4408, 4388, 4387, 4407, 4387, 4384, 4366, 4368, 4378, 4376, 4384, 4383, 4368, 4374, 4383, 4394, 4396, 4406, 4404, 4412, 4411, 4396, 4402, 4411, 4394, 4405, 4384, 4419, 4420, 4418, 4422, 4424, 4423, 4427, 4428, 4426, 4430, 4432, 4431, 4425, 4426, 4420, 4424, 4430, 4429, 4421, 4423, 4419, 4429, 4431, 4427, 4423, 4429, 4425, 4435, 4436, 4434, 4438, 4440, 4439, 4443, 4444, 4442, 4446, 4448, 4447, 4441, 4442, 4436, 4440, 4446, 4445, 4437, 4439, 4435, 4445, 4447, 4443, 4439, 4445, 4441, 4433, 4434, 4428, 4432, 4438, 4437, 4431, 4437, 4433, 4451, 4452, 4450, 4455, 4456, 4454, 4459, 4460, 4458, 4463, 4464, 4462, 4457, 4458, 4452, 4461, 4462, 4456,
4452, 4455, 4453, 4460, 4463, 4461, 4458, 4461, 4455, 4470, 4468, 4467, 4476, 4474, 4473, 4474, 4470, 4469, 4469, 4467, 4465, 4475, 4473, 4471, 4473, 4469, 4466, 4479, 4480, 4478, 4483, 4484, 4482, 4487, 4488, 4486, 4491, 4492, 4490, 4485, 4486, 4480, 4489, 4490, 4484, 4480, 4483, 4481, 4488, 4491, 4489, 4486, 4489, 4483, 4498, 4496, 4495, 4504, 4502, 4501, 4502, 4498, 4497, 4497, 4495, 4493, 4503, 4501, 4499, 4501, 4497, 4494, 4477, 4478, 4460, 4481, 4482, 4464, 4478, 4481, 4463, 4496, 4476, 4475, 4495, 4475, 4472, 4456, 4466, 4465, 4464, 4472, 4471, 4462, 4471, 4466, 4484, 4494, 4493, 4492, 4500, 4499, 4490, 4499, 4494, 4482, 4493, 4472, 4417, 4418, 4356, 4360, 4422, 4421, 4359, 4421, 4417, 4449, 4450, 4400, 4453, 4454, 4404, 4450, 4453, 4403, 4468, 4416, 4415, 4467, 4415, 4412, 4454, 4465, 4412, 4332, 4363, 4361, 4340, 4371, 4369, 4338, 4369, 4363, 4348, 4391, 4389, 4356, 4399, 4397, 4354, 4397, 4391, 4346, 4389, 4371, 4420, 4451, 4449, 4428, 4459, 4457, 4426, 4457, 4451, 4436, 4479, 4477, 4444, 4487, 4485, 4442, 4485, 4479, 4434, 4477, 4459, 4418, 4449, 4399,
4511, 4509, 4508, 4511, 4513, 4512, 4508, 4509, 4506, 4509, 4512, 4507, 4514, 4516, 4517, 4522, 4524, 4523, 4530, 4532, 4531, 4531, 4528, 4527, 4540, 4538, 4537, 4544, 4542, 4541, 4533, 4534, 4517, 4535, 4536, 4520, 4526, 4539, 4537, 4514, 4515, 4522, 4537, 4538, 4534, 4520, 4530, 4529, 4519, 4529, 4527, 4541, 4542, 4536, 4517, 4519, 4518, 4527, 4528, 4524, 4543, 4541, 4538, 4518, 4527, 4522, 4538, 4541, 4535, 4517, 4534, 4535, 4516, 4525, 4537, 4506, 4507, 4516, 4526, 4525, 4512, 4525, 4516, 4507, 4545, 4547, 4548, 4551, 4552, 4550, 4554, 4556, 4555, 4555, 4558, 4557, 4564, 4562, 4561, 4571, 4569, 4568, 4565, 4566, 4548, 4573, 4563, 4561, 4546, 4554, 4553, 4553, 4557, 4547, 4569, 4566, 4565, 4562, 4552, 4551, 4548, 4551, 4549, 4559, 4560, 4556, 4573, 4572, 4569, 4549, 4559, 4554, 4569, 4572, 4567, 4548, 4566, 4567, 4572, 4561, 4551, 4579, 4577, 4576, 4578, 4576, 4574, 4550, 4552, 4575, 4579, 4578, 4562, 4578, 4575, 4552, 4586, 4588, 4587, 4582, 4586, 4585, 4585, 4587, 4584, 4580, 4581, 4585, 4595, 4597, 4596, 4593, 4597, 4595, 4600, 4601, 4599, 4603, 4606, 4604,
4590, 4598, 4599, 4596, 4603, 4602, 4590, 4591, 4595, 4592, 4595, 4591, 4602, 4604, 4600, 4594, 4602, 4598, 4610, 4611, 4609, 4612, 4614, 4615, 4617, 4619, 4618, 4625, 4623, 4622, 4632, 4630, 4629, 4634, 4631, 4629, 4610, 4626, 4627, 4632, 4621, 4620, 4609, 4617, 4616, 4629, 4630, 4627, 4633, 4629, 4626, 4623, 4615, 4614, 4611, 4614, 4612, 4624, 4622, 4620, 4622, 4614, 4611, 4630, 4620, 4611, 4581, 4612, 4613, 4584, 4619, 4617, 4580, 4609, 4612, 4583, 4617, 4609, 4589, 4591, 4610, 4616, 4618, 4593, 4599, 4601, 4628, 4634, 4633, 4605, 4608, 4616, 4592, 4633, 4628, 4601, 4599, 4626, 4610, 4639, 4641, 4640, 4635, 4636, 4639, 4642, 4643, 4641, 4636, 4637, 4642, 4646, 4647, 4645, 4650, 4651, 4649, 4655, 4653, 4652, 4657, 4659, 4658, 4661, 4659, 4657, 4665, 4666, 4664, 4674, 4672, 4671, 4670, 4673, 4671, 4650, 4663, 4664, 4655, 4668, 4667, 4653, 4647, 4646, 4649, 4657, 4656, 4660, 4657, 4649, 4671, 4672, 4666, 4669, 4671, 4665, 4645, 4647, 4650, 4670, 4669, 4667, 4667, 4669, 4663, 4662, 4663, 4650, 4647, 4653, 4667, 4684, 4686, 4685, 4685, 4682, 4681, 4682, 4677, 4676,
4679, 4680, 4684, 4679, 4683, 4681, 4678, 4681, 4676, 4644, 4645, 4636, 4658, 4643, 4642, 4645, 4648, 4637, 4648, 4656, 4642, 4676, 4677, 4661, 4664, 4666, 4679, 4651, 4675, 4676, 4664, 4678, 4675, 4557, 4558, 4532, 4570, 4568, 4542, 4520, 4547, 4557, 4568, 4565, 4536, 4520, 4536, 4565, 4638, 4640, 4588, 4582, 4635, 4638, 4615, 4646, 4644, 4654, 4652, 4623, 4623, 4652, 4646, 4613, 4644, 4635, 4689, 4690, 4688, 4692, 4694, 4693, 4697, 4698, 4696, 4700, 4702, 4701, 4695, 4696, 4690, 4694, 4700, 4699, 4691, 4693, 4689, 4699, 4701, 4697, 4693, 4699, 4695, 4705, 4706, 4704, 4712, 4710, 4709, 4708, 4711, 4709, 4710, 4706, 4705, 4707, 4709, 4705, 4703, 4704, 4698, 4702, 4708, 4707, 4701, 4707, 4703, 4715, 4716, 4714, 4719, 4720, 4718, 4723, 4724, 4722, 4727, 4728, 4726, 4721, 4722, 4716, 4725, 4726, 4720, 4716, 4719, 4717, 4724, 4727, 4725, 4722, 4725, 4719, 4734, 4732, 4731, 4740, 4738, 4737, 4738, 4734, 4733, 4733, 4731, 4729, 4739, 4737, 4735, 4737, 4733, 4730, 4743, 4744, 4742, 4747, 4748, 4746, 4752, 4750, 4749, 4756, 4754, 4753, 4750, 4744, 4743, 4754, 4748, 4747,
4744, 4747, 4745, 4755, 4753, 4750, 4753, 4747, 4744, 4762, 4760, 4759, 4761, 4759, 4757, 4741, 4742, 4724, 4745, 4746, 4728, 4742, 4745, 4727, 4760, 4740, 4739, 4759, 4739, 4736, 4720, 4730, 4729, 4728, 4736, 4735, 4726, 4735, 4730, 4748, 4758, 4757, 4762, 4761, 4754, 4761, 4758, 4748, 4746, 4757, 4736, 4766, 4768, 4767, 4764, 4766, 4765, 4771, 4772, 4770, 4774, 4776, 4775, 4779, 4780, 4778, 4782, 4784, 4783, 4777, 4778, 4772, 4776, 4782, 4781, 4773, 4775, 4771, 4781, 4783, 4779, 4775, 4781, 4777, 4769, 4770, 4764, 4767, 4774, 4773, 4765, 4773, 4769, 4790, 4792, 4791, 4794, 4796, 4795, 4786, 4790, 4789, 4788, 4794, 4793, 4793, 4795, 4792, 4787, 4793, 4790, 4802, 4804, 4803, 4806, 4808, 4807, 4798, 4802, 4801, 4800, 4806, 4805, 4805, 4807, 4804, 4799, 4805, 4802, 4811, 4812, 4810, 4815, 4816, 4814, 4819, 4820, 4818, 4823, 4824, 4822, 4817, 4818, 4812, 4821, 4822, 4816, 4812, 4815, 4813, 4820, 4823, 4821, 4818, 4821, 4815, 4827, 4828, 4826, 4831, 4832, 4830, 4835, 4836, 4834, 4839, 4840, 4838, 4833, 4834, 4828, 4837, 4838, 4832, 4828, 4831, 4829, 4836, 4839, 4837,
4834, 4837, 4831, 4809, 4810, 4786, 4813, 4814, 4788, 4810, 4813, 4787, 4825, 4826, 4798, 4829, 4830, 4800, 4826, 4829, 4799, 4801, 4803, 4796, 4797, 4801, 4794, 4816, 4827, 4825, 4824, 4835, 4833, 4822, 4833, 4827, 4814, 4825, 4797, 4690, 4715, 4713, 4698, 4723, 4721, 4696, 4721, 4715, 4706, 4743, 4741, 4751, 4749, 4710, 4749, 4743, 4706, 4704, 4741, 4723, 4789, 4791, 4768, 4785, 4789, 4766, 4772, 4811, 4809, 4780, 4819, 4817, 4778, 4817, 4811, 4770, 4809, 4785, 4505, 4506, 4246, 4252, 4510, 4508, 4245, 4251, 4508, 4523, 4258, 4257, 4514, 4521, 4257, 4246, 4506, 4514, 4303, 4549, 4550, 4309, 4560, 4559, 4303, 4307, 4559, 4577, 4327, 4325, 4576, 4325, 4321, 4550, 4574, 4321, 4687, 4688, 4444, 4448, 4692, 4691, 4447, 4691, 4687, 4713, 4714, 4488, 4717, 4718, 4492, 4714, 4717, 4491, 4732, 4504, 4503, 4731, 4503, 4500, 4718, 4729, 4500, 4688, 4713, 4487, 4845, 4846, 4844, 4848, 4850, 4849, 4856, 4854, 4852, 4859, 4861, 4862, 4864, 4866, 4865, 4870, 4868, 4867, 4845, 4859, 4860, 4850, 4864, 4863, 4868, 4856, 4855, 4841, 4847, 4849, 4857, 4863, 4865, 4855, 4852, 4844,
4862, 4869, 4867, 4841, 4842, 4845, 4851, 4853, 4848, 4857, 4858, 4861, 4843, 4851, 4847, 4842, 4857, 4859, 4849, 4863, 4857, 4867, 4855, 4846, 4873, 4875, 4876, 4878, 4880, 4879, 4886, 4882, 4881, 4889, 4890, 4888, 4896, 4894, 4893, 4892, 4895, 4893, 4875, 4887, 4888, 4885, 4892, 4891, 4877, 4879, 4872, 4884, 4881, 4874, 4893, 4894, 4890, 4887, 4891, 4893, 4872, 4875, 4873, 4885, 4883, 4879, 4879, 4883, 4875, 4875, 4883, 4891, 4907, 4909, 4908, 4908, 4905, 4904, 4905, 4899, 4898, 4901, 4902, 4907, 4901, 4906, 4904, 4910, 4911, 4909, 4902, 4903, 4910, 4900, 4904, 4898, 4861, 4873, 4874, 4866, 4878, 4877, 4882, 4870, 4869, 4871, 4873, 4861, 4858, 4865, 4877, 4874, 4881, 4869, 4898, 4899, 4886, 4888, 4890, 4901, 4876, 4897, 4898, 4888, 4900, 4897, 4915, 4917, 4916, 4921, 4923, 4922, 4917, 4921, 4920, 4912, 4914, 4916, 4920, 4922, 4919, 4913, 4916, 4920, 4926, 4927, 4925, 4931, 4929, 4928, 4934, 4935, 4933, 4939, 4937, 4936, 4926, 4932, 4933, 4937, 4931, 4930, 4927, 4930, 4928, 4938, 4936, 4933, 4933, 4936, 4930, 4943, 4945, 4944, 4947, 4949, 4950, 4954, 4956, 4955,
4955, 4958, 4957, 4945, 4956, 4954, 4942, 4944, 4941, 4948, 4954, 4953, 4947, 4953, 4957, 4957, 4958, 4952, 4946, 4949, 4957, 4941, 4944, 4954, 4959, 4961, 4962, 4966, 4964, 4963, 4973, 4971, 4970, 4975, 4972, 4970, 4961, 4967, 4968, 4973, 4966, 4965, 4965, 4963, 4960, 4971, 4968, 4967, 4974, 4970, 4967, 4968, 4971, 4965, 4923, 4943, 4942, 4922, 4942, 4940, 4959, 4960, 4935, 4964, 4939, 4938, 4963, 4938, 4935, 4913, 4926, 4924, 4918, 4919, 4934, 4918, 4932, 4926, 4940, 4941, 4961, 4950, 4969, 4967, 4948, 4967, 4961, 4919, 4940, 4959, 4981, 4982, 4980, 4985, 4986, 4984, 4992, 4990, 4989, 4988, 4991, 4989, 4994, 4996, 4995, 5000, 4998, 4997, 4990, 4982, 4981, 4979, 4987, 4989, 4984, 4994, 4993, 4997, 4998, 4986, 4980, 4982, 4985, 4999, 4997, 4990, 4997, 4985, 4982, 4988, 4987, 4977, 4987, 4979, 4976, 5001, 5003, 5004, 5005, 5007, 5008, 5010, 5012, 5011, 5016, 5014, 5013, 5022, 5020, 5019, 5022, 5018, 5017, 5001, 5002, 5010, 5014, 5004, 5003, 5020, 5008, 5007, 5008, 5020, 5017, 5004, 5007, 5005, 5021, 5019, 5014, 5019, 5007, 5004, 5025, 5024, 5017, 5024, 5023, 5006,
4896, 4978, 4977, 4894, 4977, 4976, 4901, 4979, 4980, 4995, 4911, 4910, 4980, 4983, 4903, 4983, 4993, 4910, 4890, 4976, 4979, 5005, 5006, 4950, 4952, 5012, 5010, 4946, 5002, 5005, 4951, 5010, 5002, 5025, 4975, 4974, 5024, 4974, 4969, 4950, 5006, 5023, 4984, 4986, 5003, 5009, 5011, 4996, 5015, 5013, 4998, 5001, 5009, 4994, 5013, 5003, 4986, 5031, 5029, 5028, 5029, 5027, 5026, 5039, 5037, 5036, 5043, 5041, 5040, 5037, 5033, 5032, 5041, 5035, 5034, 5042, 5040, 5037, 5040, 5034, 5033, 5051, 5049, 5048, 5055, 5053, 5052, 5049, 5045, 5044, 5053, 5047, 5046, 5054, 5052, 5049, 5052, 5046, 5045, 5050, 5048, 5041, 5048, 5044, 5035, 5058, 5059, 5057, 5061, 5063, 5062, 5062, 5065, 5064, 5068, 5069, 5067, 5071, 5073, 5072, 5066, 5067, 5059, 5065, 5071, 5070, 5057, 5061, 5060, 5060, 5064, 5058, 5070, 5072, 5068, 5064, 5070, 5066, 5076, 5077, 5075, 5083, 5081, 5080, 5079, 5082, 5080, 5081, 5077, 5076, 5078, 5080, 5076, 5074, 5075, 5069, 5073, 5079, 5078, 5072, 5078, 5074, 5086, 5087, 5085, 5090, 5091, 5089, 5093, 5095, 5094, 5097, 5099, 5098, 5102, 5103, 5101, 5106, 5107, 5105,
5100, 5101, 5087, 5104, 5105, 5091, 5085, 5093, 5092, 5089, 5097, 5096, 5087, 5090, 5088, 5096, 5098, 5095, 5103, 5106, 5104, 5088, 5096, 5093, 5101, 5104, 5090, 5110, 5111, 5109, 5114, 5115, 5113, 5117, 5119, 5118, 5121, 5123, 5122, 5126, 5127, 5125, 5130, 5131, 5129, 5124, 5125, 5111, 5128, 5129, 5115, 5109, 5117, 5116, 5113, 5121, 5120, 5111, 5114, 5112, 5120, 5122, 5119, 5127, 5130, 5128, 5112, 5120, 5117, 5125, 5128, 5114, 5134, 5135, 5133, 5138, 5139, 5137, 5143, 5141, 5140, 5147, 5145, 5144, 5141, 5135, 5134, 5145, 5139, 5138, 5135, 5138, 5136, 5146, 5144, 5141, 5144, 5138, 5135, 5150, 5151, 5149, 5154, 5155, 5153, 5159, 5157, 5156, 5163, 5161, 5160, 5157, 5151, 5150, 5161, 5155, 5154, 5151, 5154, 5152, 5162, 5160, 5157, 5160, 5154, 5151, 5132, 5133, 5103, 5136, 5137, 5107, 5133, 5136, 5106, 5148, 5149, 5127, 5152, 5153, 5131, 5149, 5152, 5130, 5091, 5110, 5108, 5116, 5118, 5099, 5107, 5126, 5124, 5108, 5116, 5097, 5105, 5124, 5110, 5139, 5150, 5148, 5158, 5156, 5145, 5156, 5150, 5139, 5137, 5148, 5126, 5038, 5036, 5029, 5036, 5032, 5027, 5059, 5086, 5084,
5092, 5094, 5063, 5069, 5102, 5100, 5084, 5092, 5061, 5067, 5100, 5086, 5077, 5134, 5132, 5142, 5140, 5081, 5140, 5134, 5077, 5075, 5132, 5102, 5168, 5169, 5167, 5171, 5173, 5172, 5175, 5177, 5176, 5170, 5172, 5165, 5167, 5175, 5174, 5165, 5168, 5166, 5174, 5176, 5171, 5166, 5174, 5170, 5180, 5181, 5179, 5184, 5185, 5183, 5187, 5189, 5188, 5191, 5193, 5192, 5179, 5187, 5186, 5183, 5191, 5190, 5181, 5184, 5182, 5190, 5192, 5189, 5182, 5190, 5187, 5202, 5200, 5199, 5200, 5196, 5195, 5201, 5199, 5197, 5199, 5195, 5194, 5210, 5208, 5207, 5214, 5212, 5211, 5221, 5223, 5222, 5219, 5223, 5221, 5208, 5204, 5203, 5212, 5206, 5205, 5216, 5217, 5221, 5215, 5218, 5221, 5213, 5211, 5208, 5211, 5205, 5204, 5195, 5196, 5169, 5173, 5198, 5197, 5194, 5195, 5168, 5172, 5197, 5194, 5203, 5204, 5181, 5205, 5206, 5185, 5204, 5205, 5184, 5169, 5180, 5178, 5186, 5188, 5177, 5178, 5186, 5175, 5209, 5207, 5200, 5207, 5203, 5196, 5196, 5203, 5180, 5226, 5227, 5225, 5230, 5231, 5229, 5233, 5235, 5234, 5237, 5239, 5238, 5241, 5239, 5237, 5224, 5225, 5233, 5229, 5237, 5236, 5231, 5240, 5237,
5225, 5227, 5230, 5236, 5238, 5235, 5225, 5228, 5236, 5251, 5252, 5250, 5256, 5254, 5253, 5258, 5260, 5259, 5256, 5260, 5258, 5265, 5266, 5264, 5277, 5275, 5274, 5279, 5276, 5274, 5247, 5263, 5264, 5251, 5267, 5268, 5259, 5271, 5270, 5277, 5262, 5261, 5254, 5246, 5245, 5248, 5258, 5257, 5246, 5254, 5258, 5275, 5268, 5267, 5269, 5278, 5274, 5246, 5248, 5251, 5264, 5266, 5269, 5279, 5278, 5272, 5278, 5269, 5266, 5264, 5267, 5251, 5247, 5257, 5270, 5268, 5275, 5261, 5285, 5283, 5282, 5281, 5284, 5282, 5245, 5246, 5227, 5230, 5249, 5250, 5227, 5246, 5249, 5283, 5244, 5243, 5282, 5243, 5242, 5243, 5244, 5241, 5242, 5243, 5240, 5250, 5252, 5281, 5285, 5284, 5261, 5284, 5281, 5252, 5250, 5280, 5242, 5292, 5294, 5293, 5290, 5294, 5292, 5287, 5288, 5292, 5289, 5292, 5288, 5300, 5301, 5299, 5305, 5307, 5306, 5303, 5307, 5305, 5314, 5315, 5313, 5326, 5324, 5323, 5328, 5325, 5323, 5296, 5312, 5313, 5300, 5316, 5317, 5306, 5320, 5319, 5326, 5311, 5310, 5297, 5305, 5304, 5295, 5302, 5305, 5324, 5317, 5316, 5318, 5327, 5323, 5295, 5297, 5300, 5308, 5309, 5303, 5313, 5315, 5318,
5328, 5327, 5321, 5298, 5308, 5302, 5327, 5318, 5315, 5313, 5316, 5300, 5296, 5304, 5319, 5317, 5324, 5310, 5330, 5332, 5333, 5337, 5339, 5338, 5348, 5346, 5345, 5341, 5342, 5333, 5335, 5344, 5343, 5336, 5338, 5332, 5346, 5342, 5341, 5338, 5339, 5335, 5347, 5345, 5343, 5332, 5338, 5334, 5345, 5341, 5340, 5329, 5340, 5341, 5329, 5334, 5343, 5349, 5351, 5352, 5355, 5356, 5354, 5360, 5362, 5361, 5370, 5368, 5367, 5374, 5372, 5371, 5363, 5364, 5352, 5365, 5366, 5356, 5368, 5364, 5363, 5354, 5360, 5359, 5372, 5366, 5365, 5350, 5352, 5355, 5359, 5361, 5358, 5373, 5371, 5368, 5353, 5359, 5357, 5371, 5365, 5364, 5364, 5365, 5355, 5287, 5330, 5331, 5293, 5337, 5336, 5291, 5336, 5330, 5314, 5349, 5350, 5358, 5322, 5321, 5350, 5357, 5321, 5286, 5288, 5314, 5319, 5320, 5290, 5312, 5319, 5289, 5333, 5351, 5349, 5369, 5367, 5346, 5367, 5363, 5342, 5342, 5363, 5351, 5331, 5349, 5314, 5381, 5379, 5378, 5383, 5380, 5378, 5379, 5376, 5375, 5377, 5382, 5378, 5386, 5387, 5385, 5390, 5391, 5389, 5393, 5395, 5394, 5397, 5399, 5398, 5407, 5405, 5404, 5411, 5409, 5408, 5400, 5401, 5387,
5402, 5403, 5391, 5385, 5393, 5392, 5405, 5401, 5400, 5389, 5397, 5396, 5409, 5403, 5402, 5387, 5390, 5388, 5396, 5398, 5395, 5410, 5408, 5405, 5388, 5396, 5393, 5408, 5402, 5401, 5401, 5402, 5390, 5414, 5415, 5413, 5417, 5419, 5418, 5421, 5419, 5417, 5427, 5425, 5424, 5422, 5423, 5415, 5427, 5421, 5420, 5413, 5417, 5416, 5420, 5417, 5413, 5425, 5423, 5422, 5425, 5420, 5415, 5391, 5414, 5412, 5416, 5418, 5399, 5426, 5424, 5409, 5412, 5416, 5397, 5424, 5422, 5403, 5403, 5422, 5414, 5216, 5298, 5299, 5222, 5309, 5308, 5216, 5220, 5308, 5265, 5375, 5376, 5381, 5273, 5272, 5376, 5379, 5272, 5185, 5226, 5224, 5232, 5234, 5193, 5224, 5232, 5191, 5255, 5253, 5212, 5215, 5217, 5265, 5270, 5271, 5219, 5253, 5245, 5206, 5263, 5270, 5218, 5206, 5245, 5226, 5299, 5301, 5377, 5383, 5382, 5310, 5382, 5377, 5301, 5356, 5386, 5384, 5392, 5394, 5362, 5406, 5404, 5372, 5384, 5392, 5360, 5404, 5400, 5366, 5366, 5400, 5386, 5299, 5375, 5265, 4600, 4843, 4844, 4606, 4853, 4851, 4854, 4607, 4605, 4600, 4604, 4851, 4844, 4852, 4605, 4674, 4915, 4914, 4666, 4672, 4914, 4924, 4925, 4680,
4929, 4686, 4684, 4928, 4684, 4680, 4666, 4912, 4924, 5026, 5027, 4780, 4784, 5030, 5028, 4783, 5028, 5026, 5032, 5033, 4820, 5034, 5035, 4824, 5033, 5034, 4823, 5044, 5045, 4836, 5046, 5047, 4840, 5045, 5046, 4839, 5035, 5044, 4835, 5027, 5032, 4819, 4329, 4330, 4045, 4049, 4334, 4333, 4048, 4333, 4329, 4361, 4362, 4101, 4365, 4366, 4105, 4362, 4365, 4104, 4380, 4123, 4121, 4379, 4121, 4116, 4105, 4366, 4377, 4330, 4361, 4100, 5431, 5433, 5432, 5434, 5436, 5437, 5440, 5441, 5439, 5445, 5447, 5446, 5434, 5435, 5429, 5443, 5433, 5431, 5429, 5431, 5430, 5438, 5439, 5445, 5437, 5440, 5438, 5444, 5446, 5443, 5438, 5444, 5442, 5435, 5442, 5431, 5450, 5451, 5449, 5455, 5457, 5456, 5459, 5457, 5455, 5461, 5463, 5462, 5449, 5455, 5454, 5451, 5458, 5455, 5452, 5453, 5461, 5460, 5462, 5459, 5452, 5460, 5458, 5466, 5467, 5465, 5470, 5471, 5469, 5474, 5475, 5473, 5478, 5479, 5477, 5472, 5473, 5467, 5476, 5477, 5471, 5467, 5470, 5468, 5475, 5478, 5476, 5473, 5476, 5470, 5482, 5483, 5481, 5486, 5487, 5485, 5490, 5491, 5489, 5494, 5495, 5493, 5488, 5489, 5483, 5492, 5493, 5487,
5483, 5486, 5484, 5491, 5494, 5492, 5489, 5492, 5486, 5464, 5465, 5437, 5468, 5469, 5441, 5465, 5468, 5440, 5450, 5480, 5481, 5484, 5485, 5453, 5481, 5484, 5452, 5441, 5450, 5448, 5454, 5456, 5447, 5439, 5448, 5454, 5471, 5482, 5480, 5479, 5490, 5488, 5477, 5488, 5482, 5469, 5480, 5450, 5501, 5502, 5500, 5505, 5506, 5504, 5508, 5510, 5509, 5512, 5514, 5513, 5516, 5514, 5512, 5519, 5520, 5518, 5523, 5524, 5522, 5517, 5518, 5502, 5505, 5521, 5522, 5500, 5508, 5507, 5504, 5512, 5511, 5506, 5515, 5512, 5502, 5505, 5503, 5511, 5513, 5510, 5520, 5523, 5521, 5500, 5503, 5511, 5518, 5521, 5505, 5529, 5531, 5530, 5536, 5537, 5535, 5540, 5541, 5539, 5543, 5545, 5544, 5534, 5535, 5526, 5544, 5533, 5532, 5525, 5526, 5529, 5538, 5539, 5543, 5532, 5533, 5531, 5535, 5537, 5540, 5526, 5527, 5532, 5535, 5538, 5527, 5542, 5532, 5527, 5509, 5498, 5497, 5499, 5507, 5497, 5528, 5530, 5516, 5524, 5536, 5534, 5525, 5528, 5515, 5522, 5534, 5525, 5548, 5549, 5547, 5552, 5553, 5551, 5557, 5555, 5554, 5566, 5564, 5563, 5559, 5560, 5553, 5557, 5562, 5561, 5555, 5549, 5548, 5563, 5564, 5560,
5547, 5549, 5552, 5565, 5563, 5561, 5563, 5559, 5558, 5558, 5559, 5552, 5549, 5555, 5561, 5569, 5570, 5568, 5573, 5574, 5572, 5577, 5578, 5576, 5581, 5582, 5580, 5588, 5586, 5585, 5584, 5587, 5585, 5592, 5590, 5589, 5569, 5575, 5576, 5579, 5580, 5574, 5586, 5578, 5577, 5575, 5583, 5585, 5590, 5582, 5581, 5570, 5573, 5571, 5578, 5581, 5579, 5591, 5589, 5586, 5586, 5589, 5581, 5576, 5579, 5573, 5592, 5595, 5594, 5582, 5590, 5594, 5553, 5569, 5567, 5584, 5583, 5564, 5583, 5575, 5560, 5560, 5575, 5569, 5598, 5599, 5597, 5602, 5603, 5601, 5606, 5607, 5605, 5610, 5611, 5609, 5604, 5605, 5599, 5608, 5609, 5603, 5599, 5602, 5600, 5607, 5610, 5608, 5605, 5608, 5602, 5614, 5615, 5613, 5618, 5619, 5617, 5622, 5623, 5621, 5626, 5627, 5625, 5620, 5621, 5615, 5624, 5625, 5619, 5615, 5618, 5616, 5623, 5626, 5624, 5621, 5624, 5618, 5632, 5633, 5631, 5637, 5635, 5634, 5643, 5641, 5640, 5639, 5642, 5640, 5634, 5635, 5629, 5641, 5633, 5632, 5630, 5638, 5640, 5639, 5638, 5635, 5638, 5630, 5629, 5644, 5646, 5647, 5650, 5651, 5649, 5658, 5659, 5657, 5663, 5661, 5660, 5669, 5667, 5666,
5665, 5668, 5666, 5654, 5655, 5647, 5650, 5656, 5657, 5653, 5662, 5660, 5660, 5661, 5655, 5666, 5667, 5659, 5656, 5664, 5666, 5647, 5650, 5648, 5665, 5664, 5661, 5661, 5664, 5656, 5655, 5656, 5650, 5646, 5652, 5660, 5628, 5629, 5607, 5610, 5630, 5631, 5629, 5630, 5610, 5644, 5645, 5623, 5648, 5649, 5627, 5645, 5648, 5626, 5603, 5614, 5612, 5611, 5622, 5620, 5609, 5620, 5614, 5633, 5646, 5644, 5653, 5652, 5641, 5641, 5652, 5646, 5631, 5644, 5622, 5546, 5547, 5475, 5550, 5551, 5479, 5547, 5550, 5478, 5567, 5568, 5491, 5571, 5572, 5495, 5568, 5571, 5494, 5551, 5567, 5490, 5596, 5597, 5520, 5600, 5601, 5524, 5597, 5600, 5523, 5612, 5613, 5537, 5616, 5617, 5541, 5613, 5616, 5540, 5601, 5612, 5536, 5497, 5498, 5463, 5496, 5497, 5461, 5485, 5487, 5501, 5495, 5519, 5517, 5493, 5517, 5501, 5485, 5499, 5496, 5574, 5598, 5596, 5580, 5582, 5606, 5580, 5604, 5598, 5636, 5634, 5594, 5634, 5628, 5593, 5593, 5628, 5606, 5572, 5596, 5519, 5677, 5679, 5678, 5678, 5675, 5674, 5671, 5672, 5677, 5671, 5676, 5674, 5680, 5681, 5679, 5672, 5673, 5680, 5684, 5685, 5683, 5688, 5689, 5687,
5691, 5693, 5692, 5696, 5697, 5695, 5700, 5701, 5699, 5694, 5695, 5685, 5698, 5699, 5689, 5686, 5687, 5691, 5683, 5685, 5688, 5697, 5700, 5698, 5695, 5698, 5688, 5708, 5710, 5709, 5709, 5706, 5705, 5711, 5713, 5714, 5717, 5718, 5716, 5722, 5724, 5723, 5726, 5724, 5722, 5711, 5712, 5704, 5720, 5710, 5708, 5704, 5708, 5707, 5703, 5707, 5705, 5715, 5716, 5722, 5718, 5725, 5722, 5714, 5717, 5715, 5721, 5723, 5720, 5712, 5715, 5721, 5712, 5719, 5708, 5729, 5730, 5728, 5733, 5734, 5732, 5735, 5737, 5738, 5741, 5742, 5740, 5750, 5748, 5747, 5750, 5746, 5745, 5735, 5736, 5730, 5739, 5740, 5734, 5748, 5742, 5741, 5748, 5745, 5740, 5730, 5733, 5731, 5738, 5741, 5739, 5736, 5739, 5733, 5753, 5754, 5752, 5757, 5758, 5756, 5770, 5768, 5767, 5772, 5774, 5773, 5770, 5774, 5772, 5778, 5776, 5775, 5780, 5782, 5781, 5781, 5777, 5775, 5759, 5760, 5754, 5757, 5763, 5764, 5768, 5760, 5759, 5762, 5772, 5771, 5768, 5772, 5762, 5775, 5776, 5764, 5765, 5766, 5780, 5765, 5779, 5775, 5754, 5757, 5755, 5762, 5765, 5763, 5760, 5763, 5757, 5727, 5728, 5697, 5731, 5732, 5701, 5728, 5731, 5700,
5751, 5752, 5714, 5717, 5755, 5756, 5752, 5755, 5717, 5689, 5703, 5702, 5705, 5706, 5693, 5701, 5713, 5711, 5687, 5702, 5705, 5689, 5699, 5711, 5734, 5753, 5751, 5769, 5767, 5745, 5771, 5773, 5750, 5767, 5759, 5740, 5761, 5771, 5748, 5734, 5740, 5759, 5732, 5751, 5713, 5788, 5789, 5787, 5794, 5796, 5795, 5800, 5802, 5801, 5801, 5798, 5797, 5803, 5805, 5806, 5807, 5809, 5810, 5816, 5814, 5813, 5812, 5815, 5813, 5788, 5803, 5804, 5807, 5808, 5792, 5818, 5802, 5800, 5786, 5787, 5794, 5813, 5814, 5806, 5811, 5813, 5805, 5792, 5800, 5799, 5791, 5799, 5797, 5789, 5791, 5790, 5797, 5798, 5796, 5804, 5806, 5809, 5787, 5790, 5797, 5789, 5804, 5807, 5808, 5817, 5800, 5825, 5827, 5826, 5826, 5823, 5822, 5821, 5825, 5824, 5820, 5824, 5822, 5795, 5785, 5784, 5786, 5793, 5784, 5810, 5820, 5819, 5822, 5823, 5818, 5819, 5822, 5817, 5682, 5683, 5672, 5692, 5681, 5680, 5683, 5686, 5673, 5686, 5690, 5680, 5784, 5785, 5726, 5783, 5784, 5725, 5756, 5758, 5788, 5812, 5811, 5776, 5813, 5815, 5782, 5811, 5803, 5764, 5805, 5813, 5780, 5764, 5803, 5788, 5756, 5786, 5783, 5831, 5833, 5832,
5829, 5831, 5830, 5839, 5841, 5840, 5843, 5845, 5844, 5835, 5839, 5838, 5836, 5837, 5843, 5842, 5844, 5841, 5835, 5836, 5842, 5838, 5840, 5833, 5834, 5838, 5831, 5854, 5855, 5853, 5859, 5861, 5860, 5857, 5861, 5859, 5866, 5867, 5865, 5870, 5871, 5869, 5850, 5864, 5865, 5868, 5869, 5855, 5860, 5873, 5872, 5851, 5859, 5858, 5856, 5859, 5851, 5849, 5851, 5854, 5862, 5863, 5857, 5867, 5870, 5868, 5849, 5852, 5862, 5865, 5868, 5854, 5850, 5858, 5872, 5876, 5877, 5875, 5880, 5881, 5879, 5884, 5885, 5883, 5888, 5889, 5887, 5882, 5883, 5877, 5886, 5887, 5881, 5877, 5880, 5878, 5885, 5888, 5886, 5883, 5886, 5880, 5892, 5893, 5891, 5896, 5897, 5895, 5900, 5901, 5899, 5904, 5905, 5903, 5898, 5899, 5893, 5902, 5903, 5897, 5893, 5896, 5894, 5901, 5904, 5902, 5899, 5902, 5896, 5908, 5909, 5907, 5912, 5913, 5911, 5916, 5917, 5915, 5920, 5921, 5919, 5914, 5915, 5909, 5918, 5919, 5913, 5909, 5912, 5910, 5917, 5920, 5918, 5915, 5918, 5912, 5890, 5891, 5867, 5894, 5895, 5871, 5891, 5894, 5870, 5906, 5907, 5885, 5910, 5911, 5889, 5907, 5910, 5888, 5855, 5876, 5874, 5871, 5884, 5882,
5869, 5882, 5876, 5897, 5908, 5906, 5905, 5916, 5914, 5903, 5914, 5908, 5895, 5906, 5884, 5924, 5925, 5923, 5930, 5932, 5931, 5936, 5938, 5937, 5937, 5934, 5933, 5941, 5942, 5940, 5945, 5946, 5944, 5950, 5948, 5947, 5939, 5940, 5925, 5943, 5944, 5928, 5948, 5938, 5936, 5923, 5930, 5929, 5927, 5928, 5936, 5927, 5935, 5933, 5946, 5949, 5947, 5925, 5927, 5926, 5933, 5934, 5932, 5942, 5945, 5943, 5923, 5926, 5933, 5925, 5940, 5943, 5944, 5947, 5936, 5956, 5957, 5955, 5960, 5961, 5959, 5964, 5965, 5963, 5968, 5969, 5967, 5962, 5963, 5957, 5966, 5967, 5961, 5957, 5960, 5958, 5965, 5968, 5966, 5963, 5966, 5960, 5977, 5975, 5974, 5983, 5981, 5980, 5981, 5977, 5976, 5973, 5976, 5974, 5979, 5982, 5980, 5980, 5976, 5973, 5954, 5955, 5942, 5945, 5958, 5959, 5955, 5958, 5945, 5975, 5953, 5952, 5974, 5952, 5951, 5952, 5953, 5950, 5946, 5951, 5952, 5961, 5973, 5972, 5967, 5969, 5979, 5983, 5982, 5970, 5982, 5979, 5969, 5961, 5967, 5978, 5959, 5972, 5951, 5828, 5852, 5853, 5832, 5863, 5862, 5830, 5862, 5852, 5834, 5874, 5875, 5878, 5879, 5837, 5835, 5875, 5878, 5829, 5853, 5874,
5931, 5848, 5847, 5922, 5929, 5847, 5847, 5848, 5845, 5846, 5847, 5843, 5879, 5881, 5924, 5889, 5941, 5939, 5887, 5939, 5924, 5913, 5956, 5954, 5921, 5964, 5962, 5919, 5962, 5956, 5911, 5954, 5941, 5879, 5922, 5846, 5986, 5988, 5989, 5993, 5991, 5990, 5995, 5997, 5996, 5996, 6000, 5998, 6001, 5997, 5995, 6004, 6005, 6003, 6007, 6009, 6008, 5988, 6002, 6003, 6000, 6007, 6006, 5990, 5991, 5985, 5986, 5987, 5995, 5994, 5998, 5988, 5999, 5995, 5987, 6002, 6006, 6008, 5996, 5994, 5991, 5991, 5994, 5986, 5998, 6006, 6002, 6010, 6012, 6013, 6016, 6017, 6015, 6019, 6021, 6020, 6024, 6025, 6023, 6028, 6029, 6027, 6022, 6023, 6013, 6026, 6027, 6017, 6018, 6020, 6012, 6013, 6016, 6014, 6025, 6028, 6026, 6023, 6026, 6016, 6036, 6034, 6033, 6036, 6038, 6037, 6033, 6034, 6031, 6031, 6034, 6037, 6039, 6041, 6042, 6045, 6046, 6044, 6058, 6056, 6055, 6062, 6060, 6059, 6051, 6052, 6042, 6045, 6053, 6054, 6048, 6057, 6055, 6062, 6050, 6049, 6056, 6052, 6051, 6059, 6060, 6054, 6042, 6045, 6043, 6061, 6059, 6056, 6059, 6053, 6052, 6052, 6053, 6045, 6041, 6047, 6055, 6060, 6049, 6046,
6030, 6031, 6005, 6009, 6035, 6033, 6008, 6033, 6030, 6039, 6040, 6025, 6043, 6044, 6029, 6040, 6043, 6028, 6019, 6018, 5995, 6020, 6021, 6001, 6005, 6024, 6022, 6018, 6010, 5987, 6012, 6020, 5999, 5989, 6003, 6022, 6032, 6041, 6039, 6048, 6047, 6037, 6047, 6041, 6032, 6005, 6031, 6039, 6065, 6066, 6064, 6069, 6070, 6068, 6072, 6074, 6073, 6076, 6074, 6072, 6078, 6080, 6079, 6081, 6083, 6084, 6085, 6087, 6088, 6092, 6090, 6089, 6094, 6096, 6095, 6081, 6082, 6066, 6085, 6086, 6070, 6090, 6076, 6075, 6080, 6094, 6093, 6064, 6072, 6071, 6066, 6075, 6072, 6091, 6089, 6082, 6077, 6079, 6069, 6093, 6095, 6087, 6078, 6077, 6072, 6072, 6077, 6067, 6082, 6089, 6075, 6069, 6079, 6093, 6097, 6099, 6100, 6109, 6111, 6110, 6110, 6107, 6106, 6114, 6115, 6113, 6116, 6118, 6119, 6112, 6113, 6100, 6116, 6117, 6103, 6121, 6111, 6109, 6102, 6103, 6109, 6108, 6106, 6101, 6100, 6102, 6101, 6106, 6107, 6105, 6115, 6118, 6116, 6098, 6101, 6106, 6100, 6113, 6116, 6120, 6109, 6103, 6124, 6125, 6123, 6128, 6129, 6127, 6135, 6133, 6132, 6141, 6142, 6140, 6146, 6148, 6147, 6128, 6139, 6140,
6135, 6144, 6143, 6133, 6125, 6124, 6139, 6145, 6147, 6125, 6128, 6126, 6136, 6137, 6131, 6146, 6145, 6143, 6126, 6136, 6130, 6143, 6145, 6139, 6138, 6139, 6128, 6125, 6133, 6143, 6151, 6152, 6150, 6155, 6156, 6154, 6159, 6160, 6158, 6163, 6164, 6162, 6157, 6158, 6152, 6161, 6162, 6156, 6152, 6155, 6153, 6160, 6163, 6161, 6158, 6161, 6155, 6122, 6123, 6084, 6087, 6126, 6127, 6131, 6092, 6091, 6096, 6137, 6136, 6130, 6091, 6084, 6087, 6095, 6136, 6149, 6150, 6115, 6153, 6154, 6119, 6150, 6153, 6118, 6070, 6099, 6097, 6088, 6114, 6112, 6086, 6112, 6099, 6129, 6151, 6149, 6142, 6159, 6157, 6140, 6157, 6151, 6127, 6149, 6114, 6171, 6169, 6168, 6171, 6173, 6172, 6168, 6169, 6166, 6169, 6172, 6167, 6174, 6176, 6177, 6180, 6181, 6179, 6189, 6190, 6188, 6194, 6192, 6191, 6194, 6196, 6195, 6184, 6185, 6177, 6187, 6188, 6181, 6183, 6193, 6191, 6191, 6192, 6185, 6185, 6192, 6195, 6177, 6180, 6178, 6186, 6189, 6187, 6177, 6185, 6187, 6182, 6191, 6184, 6200, 6201, 6199, 6209, 6207, 6206, 6205, 6208, 6206, 6213, 6215, 6214, 6209, 6213, 6212, 6206, 6207, 6201, 6204, 6206, 6200,
6210, 6212, 6214, 6205, 6204, 6202, 6204, 6198, 6197, 6201, 6207, 6212, 6189, 6198, 6199, 6196, 6203, 6202, 6197, 6198, 6189, 6186, 6195, 6202, 6167, 6176, 6174, 6183, 6182, 6172, 6172, 6182, 6176, 6165, 6166, 6142, 6148, 6170, 6168, 6141, 6147, 6168, 6174, 6175, 6160, 6178, 6179, 6164, 6175, 6178, 6163, 6142, 6166, 6174, 6015, 6017, 6065, 6029, 6083, 6081, 6027, 6081, 6065, 6044, 6046, 6124, 6134, 6132, 6049, 6049, 6132, 6124, 6044, 6122, 6083, 5984, 5985, 5738, 5744, 5992, 5990, 5996, 5749, 5747, 5750, 5997, 5995, 5985, 5986, 5741, 5737, 5743, 5990, 5986, 5994, 5747, 5742, 5748, 5995, 5761, 6010, 6011, 6014, 6015, 5766, 5762, 6011, 6014, 5742, 5987, 6010, 5809, 6067, 6068, 6073, 5815, 5813, 5816, 6074, 6072, 5806, 6064, 6067, 6063, 6071, 5813, 5814, 6072, 6064, 6097, 6098, 5821, 6105, 5827, 5825, 6104, 5825, 5821, 5810, 6068, 6097, 6015, 6063, 5805, 5541, 5671, 5670, 5674, 5675, 5545, 5539, 5670, 5674, 5619, 5684, 5682, 5627, 5696, 5694, 5625, 5694, 5684, 5651, 5729, 5727, 5659, 5737, 5735, 5744, 5743, 5667, 5667, 5743, 5737, 5657, 5735, 5729, 5649, 5727, 5696,
5541, 5617, 5682, 6222, 6224, 6223, 6223, 6220, 6219, 6217, 6218, 6222, 6221, 6219, 6216, 6227, 6228, 6226, 6233, 6231, 6230, 6238, 6239, 6237, 6242, 6243, 6241, 6245, 6247, 6246, 6249, 6247, 6245, 6227, 6236, 6237, 6246, 6235, 6234, 6228, 6232, 6230, 6240, 6241, 6245, 6243, 6248, 6245, 6234, 6235, 6233, 6237, 6239, 6242, 6228, 6229, 6234, 6237, 6240, 6229, 6244, 6234, 6229, 6225, 6226, 6218, 6231, 6224, 6222, 6226, 6230, 6222, 6253, 6255, 6254, 6251, 6253, 6252, 6258, 6259, 6257, 6260, 6262, 6263, 6267, 6265, 6264, 6270, 6271, 6269, 6279, 6277, 6276, 6281, 6283, 6282, 6282, 6278, 6276, 6279, 6283, 6281, 6268, 6269, 6259, 6272, 6273, 6263, 6279, 6267, 6266, 6266, 6264, 6261, 6277, 6273, 6272, 6275, 6281, 6280, 6280, 6276, 6272, 6277, 6281, 6275, 6259, 6262, 6260, 6271, 6274, 6272, 6269, 6272, 6262, 6277, 6266, 6263, 6287, 6289, 6290, 6294, 6296, 6295, 6292, 6296, 6294, 6298, 6300, 6299, 6304, 6302, 6301, 6307, 6308, 6306, 6311, 6312, 6310, 6320, 6318, 6317, 6320, 6316, 6314, 6322, 6324, 6323, 6323, 6326, 6325, 6285, 6305, 6306, 6295, 6315, 6313, 6323, 6303, 6301,
6304, 6324, 6322, 6285, 6286, 6294, 6291, 6294, 6286, 6317, 6318, 6308, 6318, 6314, 6306, 6288, 6298, 6297, 6302, 6290, 6289, 6310, 6322, 6321, 6321, 6325, 6311, 6286, 6289, 6287, 6297, 6299, 6292, 6323, 6321, 6314, 6325, 6326, 6320, 6287, 6297, 6291, 6321, 6309, 6306, 6311, 6325, 6318, 6306, 6309, 6289, 6285, 6293, 6313, 6321, 6301, 6289, 6290, 6302, 6322, 6329, 6330, 6328, 6331, 6333, 6334, 6335, 6337, 6338, 6341, 6342, 6340, 6335, 6336, 6330, 6333, 6339, 6340, 6328, 6330, 6333, 6338, 6341, 6339, 6330, 6336, 6339, 6345, 6347, 6348, 6351, 6352, 6350, 6354, 6356, 6355, 6359, 6360, 6358, 6363, 6364, 6362, 6357, 6358, 6348, 6361, 6362, 6352, 6346, 6354, 6353, 6348, 6351, 6349, 6360, 6363, 6361, 6358, 6361, 6351, 6327, 6328, 6271, 6274, 6331, 6332, 6328, 6331, 6274, 6349, 6350, 6312, 6355, 6319, 6317, 6320, 6356, 6354, 6346, 6349, 6311, 6353, 6317, 6307, 6308, 6318, 6354, 6273, 6275, 6307, 6313, 6315, 6279, 6305, 6313, 6277, 6332, 6334, 6347, 6342, 6359, 6357, 6334, 6340, 6357, 6332, 6345, 6307, 6365, 6367, 6368, 6371, 6372, 6370, 6378, 6380, 6379, 6383, 6384, 6382,
6387, 6388, 6386, 6381, 6382, 6368, 6385, 6386, 6372, 6369, 6370, 6378, 6368, 6371, 6369, 6377, 6379, 6376, 6384, 6387, 6385, 6369, 6377, 6374, 6382, 6385, 6371, 6395, 6397, 6396, 6396, 6393, 6392, 6398, 6400, 6401, 6409, 6407, 6406, 6398, 6399, 6391, 6405, 6397, 6395, 6390, 6391, 6395, 6390, 6394, 6392, 6403, 6408, 6406, 6401, 6403, 6402, 6406, 6407, 6405, 6399, 6402, 6406, 6404, 6395, 6391, 6412, 6413, 6411, 6416, 6417, 6415, 6420, 6421, 6419, 6424, 6425, 6423, 6418, 6419, 6413, 6422, 6423, 6417, 6413, 6416, 6414, 6421, 6424, 6422, 6419, 6422, 6416, 6428, 6429, 6427, 6436, 6438, 6437, 6437, 6434, 6433, 6441, 6442, 6440, 6443, 6445, 6446, 6452, 6450, 6449, 6452, 6448, 6447, 6439, 6440, 6429, 6443, 6444, 6432, 6448, 6438, 6436, 6431, 6432, 6436, 6435, 6433, 6430, 6450, 6446, 6445, 6446, 6450, 6447, 6429, 6431, 6430, 6442, 6445, 6443, 6429, 6440, 6443, 6447, 6436, 6432, 6410, 6411, 6384, 6414, 6415, 6388, 6411, 6414, 6387, 6426, 6427, 6401, 6434, 6409, 6408, 6427, 6430, 6403, 6430, 6433, 6408, 6372, 6390, 6389, 6392, 6393, 6380, 6388, 6400, 6398, 6370, 6389, 6392,
6372, 6386, 6398, 6417, 6428, 6426, 6425, 6441, 6439, 6423, 6439, 6428, 6415, 6426, 6400, 6256, 6257, 6239, 6260, 6261, 6243, 6265, 6249, 6248, 6257, 6260, 6242, 6261, 6264, 6248, 6250, 6365, 6366, 6254, 6375, 6373, 6376, 6255, 6253, 6252, 6373, 6365, 6366, 6374, 6253, 6288, 6290, 6367, 6373, 6375, 6300, 6312, 6383, 6381, 6288, 6365, 6373, 6310, 6381, 6367, 6352, 6412, 6410, 6364, 6420, 6418, 6362, 6418, 6412, 6350, 6410, 6383, 6119, 6217, 6216, 6219, 6220, 6121, 6117, 6216, 6219, 6156, 6227, 6225, 6164, 6238, 6236, 6162, 6236, 6227, 6119, 6154, 6225, 6181, 6258, 6256, 6190, 6270, 6268, 6188, 6268, 6258, 6199, 6201, 6329, 6211, 6337, 6335, 6344, 6343, 6214, 6214, 6343, 6337, 6210, 6335, 6329, 6199, 6327, 6270, 6179, 6256, 6238, 6455, 6456, 6454, 6459, 6460, 6458, 6462, 6464, 6463, 6469, 6470, 6468, 6474, 6472, 6471, 6480, 6478, 6477, 6476, 6479, 6477, 6465, 6466, 6456, 6459, 6467, 6468, 6464, 6473, 6471, 6453, 6461, 6463, 6471, 6472, 6466, 6477, 6478, 6470, 6467, 6475, 6477, 6456, 6459, 6457, 6476, 6475, 6472, 6472, 6475, 6467, 6466, 6467, 6459, 6455, 6463, 6471,
6483, 6484, 6482, 6487, 6488, 6486, 6489, 6491, 6492, 6495, 6496, 6494, 6489, 6490, 6484, 6493, 6494, 6488, 6484, 6487, 6485, 6492, 6495, 6493, 6490, 6493, 6487, 6511, 6509, 6508, 6515, 6513, 6512, 6509, 6505, 6504, 6512, 6513, 6507, 6514, 6512, 6509, 6512, 6506, 6505, 6504, 6505, 6492, 6495, 6506, 6507, 6498, 6510, 6508, 6515, 6500, 6499, 6505, 6506, 6495, 6491, 6497, 6508, 6513, 6499, 6496, 6460, 6483, 6481, 6470, 6491, 6489, 6498, 6497, 6478, 6497, 6491, 6470, 6468, 6489, 6483, 6518, 6519, 6517, 6520, 6522, 6523, 6529, 6527, 6526, 6529, 6525, 6524, 6530, 6532, 6533, 6540, 6538, 6537, 6540, 6536, 6535, 6530, 6531, 6519, 6542, 6528, 6526, 6538, 6533, 6532, 6538, 6535, 6531, 6527, 6523, 6522, 6527, 6524, 6521, 6519, 6522, 6520, 6542, 6541, 6535, 6535, 6541, 6534, 6519, 6531, 6534, 6534, 6541, 6526, 6549, 6551, 6550, 6550, 6547, 6546, 6552, 6554, 6555, 6563, 6561, 6560, 6552, 6553, 6545, 6559, 6551, 6549, 6544, 6545, 6549, 6544, 6548, 6546, 6557, 6562, 6560, 6555, 6557, 6556, 6560, 6561, 6559, 6553, 6556, 6560, 6558, 6549, 6545, 6566, 6567, 6565, 6570, 6571, 6569,
6573, 6575, 6574, 6577, 6575, 6573, 6580, 6581, 6579, 6582, 6584, 6585, 6578, 6579, 6567, 6582, 6583, 6571, 6587, 6577, 6576, 6568, 6569, 6573, 6571, 6576, 6573, 6565, 6567, 6570, 6581, 6584, 6582, 6579, 6582, 6570, 6583, 6586, 6576, 6594, 6596, 6595, 6595, 6592, 6591, 6590, 6594, 6593, 6593, 6591, 6588, 6564, 6565, 6555, 6574, 6563, 6562, 6565, 6568, 6557, 6568, 6572, 6562, 6585, 6589, 6588, 6591, 6592, 6587, 6583, 6588, 6591, 6547, 6503, 6502, 6543, 6546, 6502, 6488, 6518, 6516, 6494, 6496, 6532, 6539, 6537, 6499, 6537, 6532, 6496, 6494, 6530, 6518, 6603, 6605, 6604, 6606, 6608, 6609, 6610, 6612, 6613, 6615, 6617, 6616, 6616, 6619, 6618, 6610, 6611, 6599, 6601, 6617, 6615, 6621, 6605, 6603, 6606, 6607, 6615, 6606, 6614, 6618, 6599, 6603, 6602, 6602, 6604, 6601, 6609, 6612, 6610, 6597, 6598, 6602, 6597, 6607, 6610, 6600, 6615, 6607, 6611, 6620, 6603, 6627, 6625, 6624, 6626, 6624, 6622, 6630, 6631, 6629, 6632, 6634, 6635, 6641, 6639, 6638, 6637, 6640, 6638, 6645, 6643, 6642, 6638, 6639, 6631, 6636, 6638, 6630, 6643, 6635, 6634, 6631, 6634, 6632, 6644, 6642, 6639,
6639, 6642, 6634, 6651, 6649, 6648, 6647, 6650, 6648, 6628, 6629, 6609, 6632, 6633, 6613, 6619, 6637, 6636, 6629, 6632, 6612, 6608, 6618, 6636, 6649, 6627, 6626, 6648, 6626, 6623, 6613, 6623, 6622, 6624, 6625, 6621, 6611, 6622, 6624, 6633, 6635, 6647, 6651, 6650, 6643, 6650, 6647, 6635, 6633, 6646, 6623, 6654, 6655, 6653, 6658, 6659, 6657, 6662, 6663, 6661, 6666, 6667, 6665, 6660, 6661, 6655, 6664, 6665, 6659, 6655, 6658, 6656, 6663, 6666, 6664, 6661, 6664, 6658, 6670, 6671, 6669, 6676, 6674, 6673, 6681, 6682, 6680, 6689, 6691, 6690, 6690, 6687, 6686, 6670, 6679, 6680, 6687, 6678, 6677, 6671, 6675, 6673, 6684, 6685, 6689, 6684, 6688, 6686, 6677, 6678, 6676, 6682, 6684, 6683, 6671, 6672, 6677, 6680, 6683, 6672, 6683, 6686, 6677, 6694, 6695, 6693, 6698, 6699, 6697, 6701, 6703, 6702, 6707, 6709, 6710, 6714, 6712, 6711, 6714, 6716, 6715, 6704, 6705, 6695, 6707, 6708, 6699, 6703, 6713, 6711, 6692, 6700, 6702, 6711, 6712, 6705, 6712, 6715, 6706, 6695, 6698, 6696, 6706, 6709, 6707, 6718, 6717, 6715, 6715, 6717, 6709, 6695, 6705, 6707, 6694, 6702, 6711, 6721, 6722, 6720,
6725, 6726, 6724, 6730, 6728, 6727, 6733, 6734, 6732, 6737, 6738, 6736, 6731, 6732, 6722, 6725, 6735, 6736, 6726, 6729, 6727, 6722, 6725, 6723, 6734, 6737, 6735, 6732, 6735, 6725, 6662, 6692, 6693, 6696, 6697, 6667, 6693, 6696, 6666, 6719, 6720, 6682, 6723, 6724, 6685, 6728, 6691, 6689, 6682, 6720, 6723, 6724, 6727, 6689, 6659, 6670, 6668, 6667, 6681, 6679, 6665, 6679, 6670, 6699, 6721, 6719, 6710, 6733, 6731, 6708, 6731, 6721, 6697, 6719, 6681, 6748, 6750, 6749, 6749, 6746, 6745, 6746, 6741, 6740, 6743, 6744, 6748, 6743, 6747, 6745, 6742, 6745, 6740, 6753, 6754, 6752, 6760, 6758, 6757, 6756, 6759, 6757, 6760, 6766, 6765, 6757, 6758, 6754, 6751, 6755, 6757, 6758, 6765, 6762, 6769, 6770, 6768, 6773, 6774, 6772, 6777, 6778, 6776, 6781, 6782, 6780, 6784, 6786, 6785, 6769, 6775, 6776, 6779, 6780, 6774, 6775, 6783, 6785, 6770, 6773, 6771, 6778, 6781, 6779, 6776, 6779, 6773, 6789, 6790, 6788, 6792, 6794, 6793, 6796, 6794, 6792, 6797, 6799, 6800, 6808, 6806, 6805, 6797, 6798, 6790, 6804, 6796, 6795, 6787, 6788, 6792, 6790, 6795, 6792, 6802, 6807, 6805, 6800, 6802, 6801,
6805, 6806, 6804, 6798, 6801, 6805, 6798, 6803, 6795, 6814, 6815, 6813, 6819, 6817, 6816, 6819, 6821, 6820, 6825, 6826, 6824, 6834, 6832, 6831, 6830, 6833, 6831, 6814, 6823, 6824, 6821, 6828, 6827, 6816, 6817, 6810, 6810, 6817, 6820, 6831, 6832, 6826, 6829, 6831, 6825, 6811, 6814, 6812, 6830, 6829, 6827, 6827, 6829, 6823, 6822, 6823, 6814, 6811, 6820, 6827, 6793, 6764, 6763, 6791, 6763, 6761, 6809, 6810, 6778, 6812, 6813, 6782, 6786, 6818, 6816, 6778, 6810, 6812, 6777, 6785, 6816, 6752, 6754, 6769, 6784, 6783, 6765, 6765, 6783, 6775, 6762, 6775, 6769, 6837, 6838, 6836, 6843, 6841, 6840, 6848, 6849, 6847, 6852, 6853, 6851, 6855, 6857, 6856, 6859, 6857, 6855, 6837, 6846, 6847, 6856, 6845, 6844, 6838, 6842, 6840, 6851, 6855, 6854, 6853, 6858, 6855, 6844, 6845, 6843, 6847, 6849, 6852, 6838, 6839, 6844, 6847, 6850, 6839, 6854, 6844, 6839, 6862, 6863, 6861, 6866, 6867, 6865, 6871, 6869, 6868, 6874, 6875, 6873, 6878, 6879, 6877, 6872, 6873, 6863, 6866, 6876, 6877, 6870, 6868, 6865, 6863, 6866, 6864, 6875, 6878, 6876, 6873, 6876, 6866, 6889, 6891, 6890, 6890, 6887, 6886,
6887, 6882, 6881, 6884, 6885, 6889, 6884, 6888, 6886, 6883, 6886, 6881, 6860, 6861, 6849, 6864, 6865, 6853, 6869, 6859, 6858, 6861, 6864, 6852, 6865, 6868, 6858, 6881, 6882, 6871, 6879, 6884, 6883, 6867, 6880, 6881, 6877, 6883, 6880, 6751, 6752, 6710, 6718, 6756, 6755, 6709, 6717, 6755, 6767, 6768, 6734, 6771, 6772, 6738, 6768, 6771, 6737, 6752, 6767, 6733, 6835, 6836, 6744, 6841, 6750, 6748, 6836, 6840, 6748, 6740, 6741, 6730, 6738, 6743, 6742, 6726, 6739, 6740, 6736, 6742, 6739, 6774, 6837, 6835, 6782, 6848, 6846, 6780, 6846, 6837, 6815, 6862, 6860, 6824, 6826, 6874, 6824, 6872, 6862, 6813, 6860, 6848, 6738, 6772, 6835, 6893, 6895, 6896, 6900, 6902, 6901, 6901, 6904, 6903, 6894, 6900, 6899, 6899, 6903, 6895, 6903, 6904, 6898, 6895, 6903, 6897, 6909, 6910, 6908, 6912, 6914, 6913, 6919, 6920, 6918, 6922, 6924, 6923, 6917, 6918, 6910, 6914, 6922, 6921, 6911, 6913, 6906, 6915, 6921, 6923, 6905, 6906, 6909, 6916, 6919, 6917, 6906, 6915, 6917, 6913, 6921, 6915, 6907, 6908, 6896, 6898, 6912, 6911, 6892, 6905, 6907, 6897, 6911, 6905, 6928, 6930, 6931, 6936, 6938, 6937,
6937, 6940, 6939, 6944, 6946, 6945, 6945, 6942, 6941, 6928, 6929, 6936, 6935, 6939, 6930, 6934, 6944, 6943, 6933, 6943, 6941, 6931, 6933, 6932, 6941, 6942, 6938, 6932, 6941, 6936, 6948, 6950, 6951, 6955, 6957, 6956, 6956, 6959, 6958, 6964, 6965, 6963, 6967, 6969, 6968, 6962, 6963, 6951, 6953, 6967, 6966, 6966, 6968, 6961, 6949, 6955, 6954, 6948, 6954, 6958, 6958, 6959, 6953, 6960, 6961, 6964, 6950, 6958, 6952, 6947, 6960, 6962, 6947, 6952, 6966, 6972, 6973, 6971, 6976, 6977, 6975, 6980, 6981, 6979, 6984, 6985, 6983, 6978, 6979, 6973, 6982, 6983, 6977, 6973, 6976, 6974, 6981, 6984, 6982, 6979, 6982, 6976, 6927, 6957, 6955, 6926, 6955, 6949, 6970, 6971, 6931, 6933, 6974, 6975, 6931, 6971, 6974, 6939, 6940, 6927, 6925, 6930, 6939, 6951, 6972, 6970, 6965, 6980, 6978, 6963, 6978, 6972, 6925, 6949, 6970, 6992, 6993, 6991, 7003, 7001, 7000, 6998, 7002, 7000, 7003, 6999, 6997, 7003, 7006, 7005, 7000, 7001, 6993, 6996, 7000, 6992, 6993, 7001, 6997, 6998, 6996, 6994, 6996, 6990, 6989, 6993, 7001, 7005, 7009, 7010, 7008, 7013, 7015, 7014, 7017, 7015, 7013, 7022, 7023, 7021,
7029, 7027, 7026, 7009, 7020, 7021, 7027, 7019, 7018, 7007, 7008, 7013, 7016, 7013, 7008, 7025, 7028, 7026, 7018, 7019, 7017, 7021, 7023, 7025, 7010, 7011, 7018, 7021, 7024, 7011, 7026, 7018, 7011, 7014, 6988, 6987, 7007, 7012, 6987, 6919, 6990, 6991, 6924, 6995, 6994, 6916, 6989, 6990, 6923, 6994, 6989, 6987, 6988, 6946, 6934, 6986, 6987, 6975, 6977, 7009, 6985, 7022, 7020, 6983, 7020, 7009, 6975, 7007, 6986, 6652, 6653, 6581, 6656, 6657, 6585, 6653, 6656, 6584, 6668, 6669, 6590, 6674, 6596, 6594, 6669, 6673, 6594, 6585, 6657, 6668, 7040, 7038, 7037, 7040, 7042, 7041, 7046, 7044, 7043, 7048, 7050, 7049, 7049, 7045, 7043, 7046, 7050, 7048, 7052, 7054, 7055, 7059, 7061, 7060, 7035, 7052, 7053, 7042, 7057, 7056, 7037, 7038, 7031, 7031, 7038, 7041, 7043, 7044, 7034, 7035, 7036, 7048, 7035, 7047, 7043, 7034, 7044, 7048, 7058, 7060, 7054, 7032, 7035, 7033, 7059, 7058, 7056, 7056, 7058, 7052, 7032, 7051, 7052, 7032, 7041, 7056, 7064, 7065, 7063, 7068, 7069, 7067, 7072, 7073, 7071, 7074, 7076, 7077, 7064, 7070, 7071, 7074, 7075, 7069, 7065, 7068, 7066, 7073, 7076, 7074,
7081, 7080, 7078, 7078, 7080, 7076, 7071, 7074, 7068, 7084, 7085, 7083, 7091, 7089, 7088, 7087, 7090, 7088, 7095, 7097, 7096, 7091, 7095, 7094, 7088, 7089, 7085, 7086, 7088, 7084, 7092, 7094, 7096, 7089, 7094, 7092, 7098, 7100, 7101, 7102, 7104, 7105, 7112, 7113, 7111, 7116, 7117, 7115, 7110, 7111, 7101, 7114, 7115, 7105, 7099, 7101, 7104, 7108, 7109, 7107, 7113, 7116, 7114, 7099, 7102, 7108, 7111, 7114, 7104, 7082, 7083, 7055, 7061, 7087, 7086, 7054, 7060, 7086, 7098, 7099, 7073, 7102, 7103, 7077, 7107, 7079, 7078, 7081, 7109, 7108, 7099, 7106, 7078, 7076, 7080, 7108, 7034, 7036, 7064, 7055, 7072, 7070, 7053, 7070, 7064, 7083, 7085, 7100, 7093, 7112, 7110, 7092, 7110, 7100, 7055, 7083, 7098, 7120, 7121, 7119, 7124, 7125, 7123, 7128, 7129, 7127, 7132, 7133, 7131, 7126, 7127, 7121, 7130, 7131, 7125, 7121, 7124, 7122, 7129, 7132, 7130, 7127, 7130, 7124, 7138, 7139, 7137, 7147, 7145, 7144, 7154, 7152, 7151, 7156, 7153, 7151, 7138, 7148, 7149, 7154, 7143, 7142, 7151, 7152, 7149, 7155, 7151, 7148, 7146, 7144, 7140, 7137, 7139, 7141, 7147, 7146, 7142, 7146, 7141, 7139,
7149, 7152, 7142, 7159, 7160, 7158, 7166, 7164, 7163, 7170, 7168, 7167, 7171, 7173, 7174, 7182, 7180, 7179, 7171, 7172, 7160, 7178, 7166, 7165, 7160, 7165, 7163, 7168, 7162, 7161, 7176, 7181, 7179, 7169, 7167, 7163, 7174, 7176, 7175, 7179, 7180, 7178, 7167, 7161, 7158, 7172, 7175, 7179, 7172, 7177, 7165, 7157, 7158, 7129, 7132, 7161, 7162, 7170, 7135, 7134, 7129, 7158, 7161, 7162, 7168, 7134, 7125, 7138, 7136, 7131, 7133, 7150, 7156, 7155, 7134, 7134, 7155, 7150, 7131, 7148, 7138, 7188, 7189, 7187, 7192, 7193, 7191, 7195, 7197, 7196, 7200, 7201, 7199, 7204, 7205, 7203, 7207, 7209, 7208, 7198, 7199, 7189, 7202, 7203, 7193, 7197, 7207, 7206, 7194, 7196, 7188, 7198, 7206, 7208, 7189, 7192, 7190, 7201, 7204, 7202, 7199, 7202, 7192, 7188, 7196, 7206, 7212, 7213, 7211, 7216, 7217, 7215, 7223, 7221, 7220, 7219, 7222, 7220, 7228, 7229, 7227, 7231, 7233, 7232, 7226, 7227, 7217, 7223, 7231, 7230, 7220, 7221, 7213, 7218, 7220, 7212, 7224, 7230, 7232, 7211, 7213, 7216, 7225, 7228, 7226, 7224, 7226, 7216, 7221, 7230, 7224, 7210, 7211, 7201, 7214, 7215, 7205, 7209, 7219, 7218,
7211, 7214, 7204, 7200, 7208, 7218, 7195, 7194, 7184, 7184, 7194, 7186, 7236, 7237, 7235, 7243, 7241, 7240, 7246, 7247, 7245, 7253, 7251, 7250, 7244, 7245, 7237, 7251, 7243, 7242, 7239, 7242, 7240, 7249, 7252, 7250, 7237, 7239, 7238, 7247, 7249, 7248, 7245, 7248, 7239, 7248, 7250, 7242, 7256, 7257, 7255, 7264, 7266, 7265, 7265, 7262, 7261, 7269, 7270, 7268, 7271, 7273, 7274, 7278, 7276, 7275, 7267, 7268, 7257, 7271, 7272, 7260, 7276, 7266, 7264, 7259, 7260, 7264, 7263, 7261, 7258, 7274, 7277, 7275, 7257, 7259, 7258, 7270, 7273, 7271, 7257, 7268, 7271, 7275, 7264, 7260, 7284, 7286, 7285, 7288, 7290, 7289, 7291, 7293, 7294, 7297, 7298, 7296, 7300, 7302, 7301, 7279, 7291, 7292, 7295, 7296, 7282, 7285, 7300, 7299, 7280, 7284, 7283, 7299, 7301, 7293, 7282, 7288, 7287, 7287, 7289, 7286, 7294, 7297, 7295, 7281, 7287, 7284, 7280, 7292, 7295, 7279, 7283, 7299, 7254, 7255, 7247, 7262, 7253, 7252, 7255, 7258, 7249, 7258, 7261, 7252, 7097, 7185, 7184, 7096, 7184, 7183, 7112, 7186, 7187, 7190, 7191, 7117, 7187, 7190, 7116, 7183, 7186, 7112, 7234, 7235, 7174, 7241, 7182, 7181,
7235, 7238, 7176, 7238, 7240, 7181, 7069, 7120, 7118, 7077, 7128, 7126, 7075, 7126, 7120, 7105, 7159, 7157, 7117, 7173, 7171, 7115, 7171, 7159, 7103, 7157, 7128, 7193, 7236, 7234, 7205, 7246, 7244, 7203, 7244, 7236, 7217, 7256, 7254, 7229, 7269, 7267, 7227, 7267, 7256, 7215, 7254, 7246, 7191, 7234, 7173, 7311, 7309, 7308, 7309, 7305, 7304, 7307, 7310, 7308, 7306, 7308, 7304, 7316, 7318, 7317, 7323, 7324, 7322, 7327, 7328, 7326, 7332, 7334, 7333, 7312, 7321, 7322, 7317, 7330, 7329, 7333, 7320, 7319, 7312, 7313, 7316, 7326, 7332, 7331, 7319, 7320, 7318, 7322, 7324, 7327, 7314, 7319, 7316, 7322, 7325, 7314, 7315, 7329, 7321, 7325, 7331, 7319, 7339, 7340, 7338, 7342, 7344, 7343, 7349, 7351, 7352, 7358, 7360, 7359, 7359, 7355, 7353, 7356, 7360, 7358, 7349, 7350, 7340, 7344, 7356, 7354, 7335, 7341, 7343, 7347, 7348, 7358, 7347, 7357, 7353, 7346, 7354, 7358, 7336, 7339, 7337, 7346, 7348, 7351, 7362, 7361, 7358, 7358, 7361, 7351, 7346, 7349, 7339, 7336, 7343, 7354, 7365, 7366, 7364, 7369, 7370, 7368, 7373, 7374, 7372, 7377, 7378, 7376, 7371, 7372, 7366, 7375, 7376, 7370,
7366, 7369, 7367, 7374, 7377, 7375, 7372, 7375, 7369, 7381, 7382, 7380, 7383, 7385, 7386, 7390, 7388, 7387, 7392, 7394, 7393, 7397, 7398, 7396, 7401, 7402, 7400, 7406, 7404, 7403, 7408, 7410, 7409, 7395, 7396, 7382, 7385, 7399, 7400, 7389, 7387, 7380, 7398, 7405, 7403, 7383, 7391, 7393, 7399, 7407, 7409, 7393, 7394, 7390, 7408, 7407, 7403, 7382, 7385, 7393, 7403, 7407, 7399, 7396, 7399, 7385, 7363, 7364, 7324, 7367, 7368, 7328, 7364, 7367, 7327, 7347, 7379, 7380, 7383, 7384, 7352, 7388, 7360, 7358, 7362, 7392, 7391, 7380, 7387, 7358, 7361, 7391, 7383, 7328, 7347, 7345, 7353, 7355, 7334, 7326, 7345, 7353, 7370, 7381, 7379, 7378, 7397, 7395, 7376, 7395, 7381, 7328, 7368, 7379, 7413, 7414, 7412, 7415, 7417, 7418, 7422, 7420, 7419, 7425, 7426, 7424, 7427, 7429, 7430, 7434, 7432, 7431, 7423, 7424, 7414, 7427, 7428, 7418, 7432, 7422, 7421, 7418, 7421, 7419, 7433, 7431, 7428, 7414, 7417, 7415, 7426, 7429, 7427, 7424, 7427, 7417, 7431, 7421, 7418, 7438, 7440, 7441, 7445, 7443, 7442, 7445, 7447, 7446, 7453, 7451, 7450, 7453, 7449, 7448, 7442, 7443, 7436, 7436, 7443, 7446,
7450, 7451, 7441, 7451, 7448, 7439, 7437, 7440, 7438, 7452, 7450, 7446, 7450, 7440, 7437, 7456, 7457, 7455, 7464, 7462, 7461, 7466, 7463, 7461, 7469, 7470, 7468, 7476, 7474, 7473, 7467, 7468, 7457, 7474, 7466, 7465, 7462, 7459, 7458, 7460, 7465, 7461, 7475, 7473, 7471, 7455, 7457, 7460, 7468, 7470, 7472, 7457, 7468, 7471, 7473, 7465, 7460, 7454, 7455, 7426, 7429, 7458, 7459, 7464, 7434, 7433, 7455, 7458, 7429, 7459, 7462, 7433, 7477, 7479, 7480, 7483, 7484, 7482, 7486, 7488, 7487, 7498, 7496, 7495, 7502, 7500, 7499, 7491, 7492, 7480, 7483, 7493, 7494, 7502, 7490, 7489, 7485, 7487, 7479, 7495, 7496, 7492, 7499, 7500, 7494, 7478, 7480, 7483, 7501, 7499, 7496, 7499, 7493, 7492, 7492, 7493, 7483, 7500, 7489, 7484, 7503, 7505, 7506, 7509, 7510, 7508, 7516, 7514, 7513, 7519, 7521, 7522, 7525, 7526, 7524, 7528, 7530, 7529, 7529, 7532, 7531, 7523, 7524, 7510, 7516, 7530, 7528, 7514, 7506, 7505, 7520, 7528, 7527, 7527, 7531, 7521, 7504, 7506, 7509, 7517, 7518, 7512, 7522, 7525, 7523, 7507, 7517, 7511, 7520, 7523, 7509, 7514, 7528, 7520, 7536, 7538, 7537, 7545, 7547, 7546,
7543, 7547, 7545, 7538, 7543, 7542, 7533, 7535, 7537, 7541, 7545, 7544, 7542, 7545, 7541, 7537, 7542, 7539, 7550, 7551, 7549, 7554, 7555, 7553, 7560, 7561, 7559, 7564, 7565, 7563, 7558, 7559, 7551, 7562, 7563, 7555, 7551, 7554, 7552, 7561, 7564, 7562, 7559, 7562, 7554, 7521, 7548, 7549, 7552, 7553, 7526, 7532, 7557, 7556, 7549, 7552, 7525, 7531, 7556, 7548, 7482, 7484, 7505, 7515, 7513, 7489, 7513, 7505, 7484, 7533, 7534, 7550, 7556, 7557, 7536, 7539, 7541, 7560, 7548, 7556, 7535, 7534, 7539, 7558, 7568, 7569, 7567, 7577, 7575, 7574, 7578, 7580, 7581, 7585, 7583, 7582, 7578, 7579, 7569, 7583, 7573, 7572, 7584, 7582, 7579, 7571, 7576, 7574, 7567, 7569, 7571, 7577, 7576, 7572, 7576, 7571, 7569, 7579, 7582, 7572, 7586, 7588, 7589, 7593, 7591, 7590, 7600, 7598, 7597, 7602, 7599, 7597, 7588, 7594, 7595, 7600, 7593, 7592, 7592, 7590, 7587, 7598, 7595, 7594, 7601, 7597, 7594, 7595, 7598, 7592, 7580, 7586, 7587, 7591, 7585, 7584, 7590, 7584, 7581, 7373, 7477, 7478, 7481, 7482, 7378, 7478, 7481, 7377, 7503, 7504, 7398, 7401, 7507, 7508, 7512, 7406, 7405, 7410, 7518, 7517,
7511, 7405, 7398, 7409, 7517, 7507, 7482, 7503, 7397, 7566, 7567, 7470, 7575, 7476, 7475, 7470, 7567, 7570, 7574, 7475, 7472, 7340, 7413, 7411, 7352, 7425, 7423, 7350, 7423, 7413, 7386, 7456, 7454, 7402, 7469, 7467, 7400, 7467, 7456, 7384, 7454, 7425, 7510, 7568, 7566, 7526, 7580, 7578, 7524, 7578, 7568, 7555, 7588, 7586, 7563, 7565, 7596, 7563, 7594, 7588, 7553, 7586, 7580, 7508, 7566, 7469, 7337, 7338, 7229, 7233, 7342, 7341, 7225, 7335, 7337, 7225, 7232, 7341, 7411, 7412, 7270, 7415, 7416, 7274, 7420, 7278, 7277, 7412, 7415, 7273, 7416, 7419, 7277, 7435, 7436, 7294, 7438, 7439, 7298, 7302, 7444, 7442, 7294, 7436, 7438, 7301, 7442, 7435, 7338, 7411, 7269, 7605, 7311, 7310, 7603, 7604, 7310, 7304, 7305, 7290, 7298, 7307, 7306, 7282, 7303, 7304, 7296, 7306, 7303, 7605, 7604, 7448, 7448, 7604, 7603, 7298, 7439, 7603, 7608, 7609, 7607, 7612, 7613, 7611, 7615, 7617, 7616, 7619, 7617, 7615, 7622, 7623, 7621, 7626, 7627, 7625, 7620, 7621, 7609, 7612, 7624, 7625, 7610, 7611, 7615, 7618, 7615, 7611, 7607, 7609, 7612, 7623, 7626, 7624, 7621, 7624, 7612, 7634, 7636, 7635,
7643, 7645, 7644, 7644, 7641, 7640, 7641, 7632, 7631, 7638, 7639, 7643, 7638, 7642, 7640, 7629, 7633, 7635, 7637, 7640, 7631, 7648, 7649, 7647, 7652, 7653, 7651, 7656, 7657, 7655, 7660, 7661, 7659, 7654, 7655, 7649, 7658, 7659, 7653, 7649, 7652, 7650, 7657, 7660, 7658, 7655, 7658, 7652, 7664, 7665, 7663, 7670, 7668, 7667, 7675, 7676, 7674, 7683, 7685, 7684, 7684, 7681, 7680, 7664, 7673, 7674, 7681, 7672, 7671, 7665, 7669, 7667, 7678, 7679, 7683, 7678, 7682, 7680, 7671, 7672, 7670, 7676, 7678, 7677, 7665, 7666, 7671, 7674, 7677, 7666, 7677, 7680, 7671, 7646, 7647, 7623, 7650, 7651, 7627, 7647, 7650, 7626, 7662, 7663, 7639, 7668, 7645, 7643, 7663, 7667, 7643, 7631, 7632, 7619, 7627, 7638, 7637, 7613, 7628, 7631, 7625, 7637, 7628, 7653, 7664, 7662, 7661, 7675, 7673, 7659, 7673, 7664, 7627, 7651, 7662, 7686, 7688, 7689, 7692, 7693, 7691, 7698, 7699, 7697, 7702, 7703, 7701, 7709, 7707, 7706, 7705, 7708, 7706, 7696, 7697, 7689, 7700, 7701, 7693, 7695, 7705, 7704, 7706, 7707, 7699, 7696, 7704, 7706, 7689, 7692, 7690, 7697, 7699, 7702, 7697, 7700, 7692, 7688, 7694, 7704,
7712, 7713, 7711, 7718, 7716, 7715, 7723, 7724, 7722, 7730, 7728, 7727, 7712, 7721, 7722, 7728, 7720, 7719, 7713, 7717, 7715, 7726, 7729, 7727, 7719, 7720, 7718, 7724, 7726, 7725, 7713, 7714, 7719, 7722, 7725, 7714, 7725, 7727, 7719, 7733, 7735, 7736, 7738, 7740, 7739, 7745, 7746, 7744, 7748, 7750, 7749, 7743, 7744, 7736, 7742, 7748, 7747, 7731, 7737, 7739, 7743, 7747, 7749, 7732, 7735, 7733, 7742, 7741, 7739, 7739, 7741, 7735, 7735, 7741, 7747, 7753, 7754, 7752, 7757, 7758, 7756, 7760, 7762, 7761, 7764, 7762, 7760, 7767, 7768, 7766, 7771, 7772, 7770, 7776, 7774, 7773, 7765, 7766, 7754, 7769, 7770, 7758, 7774, 7764, 7763, 7755, 7756, 7760, 7763, 7760, 7756, 7775, 7773, 7770, 7752, 7754, 7757, 7768, 7771, 7769, 7766, 7769, 7757, 7770, 7773, 7763, 7733, 7734, 7703, 7709, 7738, 7737, 7731, 7733, 7702, 7699, 7707, 7737, 7751, 7752, 7724, 7761, 7730, 7729, 7752, 7755, 7726, 7759, 7729, 7726, 7693, 7712, 7710, 7703, 7723, 7721, 7701, 7721, 7712, 7736, 7753, 7751, 7746, 7767, 7765, 7744, 7765, 7753, 7734, 7751, 7723, 7779, 7780, 7778, 7783, 7784, 7782, 7786, 7788, 7787,
7792, 7794, 7795, 7799, 7797, 7796, 7799, 7801, 7800, 7789, 7790, 7780, 7792, 7793, 7784, 7788, 7798, 7796, 7777, 7785, 7787, 7796, 7797, 7790, 7797, 7800, 7791, 7780, 7783, 7781, 7791, 7794, 7792, 7803, 7802, 7800, 7800, 7802, 7794, 7780, 7790, 7792, 7779, 7787, 7796, 7806, 7807, 7805, 7810, 7811, 7809, 7815, 7813, 7812, 7818, 7819, 7817, 7820, 7822, 7823, 7816, 7817, 7807, 7820, 7821, 7811, 7825, 7815, 7814, 7811, 7814, 7812, 7807, 7810, 7808, 7819, 7822, 7820, 7817, 7820, 7810, 7824, 7814, 7811, 7828, 7829, 7827, 7835, 7833, 7832, 7831, 7834, 7832, 7839, 7841, 7840, 7835, 7839, 7838, 7832, 7833, 7829, 7826, 7830, 7832, 7836, 7838, 7840, 7829, 7833, 7838, 7844, 7845, 7843, 7848, 7849, 7847, 7850, 7852, 7853, 7856, 7857, 7855, 7850, 7851, 7845, 7854, 7855, 7849, 7845, 7848, 7846, 7853, 7856, 7854, 7851, 7854, 7848, 7826, 7827, 7795, 7803, 7831, 7830, 7794, 7802, 7830, 7842, 7843, 7819, 7846, 7847, 7823, 7843, 7846, 7822, 7784, 7806, 7804, 7795, 7818, 7816, 7793, 7816, 7806, 7827, 7829, 7844, 7837, 7852, 7850, 7859, 7858, 7840, 7840, 7858, 7852, 7836, 7850, 7844,
7827, 7842, 7818, 7866, 7864, 7863, 7866, 7868, 7867, 7875, 7873, 7872, 7868, 7877, 7876, 7874, 7872, 7869, 7864, 7861, 7860, 7861, 7864, 7867, 7867, 7876, 7871, 7880, 7881, 7879, 7884, 7885, 7883, 7888, 7889, 7887, 7892, 7893, 7891, 7895, 7897, 7896, 7880, 7886, 7887, 7890, 7891, 7885, 7894, 7896, 7888, 7881, 7884, 7882, 7889, 7892, 7890, 7887, 7890, 7884, 7900, 7901, 7899, 7903, 7905, 7904, 7907, 7905, 7903, 7910, 7911, 7909, 7916, 7914, 7913, 7908, 7909, 7901, 7914, 7907, 7906, 7898, 7899, 7903, 7901, 7906, 7903, 7911, 7915, 7913, 7917, 7918, 7916, 7911, 7912, 7917, 7909, 7913, 7906, 7921, 7922, 7920, 7925, 7926, 7924, 7932, 7930, 7929, 7928, 7931, 7929, 7937, 7938, 7936, 7940, 7942, 7941, 7935, 7936, 7926, 7932, 7940, 7939, 7929, 7930, 7922, 7927, 7929, 7921, 7933, 7939, 7941, 7920, 7922, 7925, 7934, 7937, 7935, 7933, 7935, 7925, 7930, 7939, 7933, 7904, 7875, 7874, 7898, 7902, 7874, 7919, 7920, 7889, 7923, 7924, 7893, 7897, 7928, 7927, 7920, 7923, 7892, 7888, 7896, 7927, 7862, 7880, 7878, 7895, 7894, 7876, 7876, 7894, 7886, 7871, 7886, 7880, 7656, 7777, 7778,
7781, 7782, 7661, 7778, 7781, 7660, 7804, 7805, 7676, 7808, 7809, 7679, 7813, 7685, 7683, 7676, 7805, 7808, 7809, 7812, 7683, 7782, 7804, 7675, 7860, 7861, 7746, 7750, 7865, 7863, 7745, 7749, 7863, 7878, 7879, 7768, 7771, 7882, 7883, 7879, 7882, 7771, 7746, 7861, 7878, 7630, 7688, 7686, 7695, 7694, 7635, 7635, 7694, 7688, 7823, 7870, 7869, 7872, 7873, 7825, 7821, 7869, 7872, 7847, 7849, 7900, 7857, 7910, 7908, 7855, 7908, 7900, 7847, 7898, 7870, 7945, 7947, 7948, 7950, 7952, 7951, 7961, 7959, 7958, 7961, 7963, 7962, 7955, 7956, 7948, 7954, 7960, 7958, 7949, 7951, 7944, 7959, 7956, 7955, 7956, 7959, 7962, 7944, 7947, 7945, 7954, 7953, 7951, 7951, 7953, 7947, 7953, 7958, 7955, 7966, 7967, 7965, 7970, 7971, 7969, 7974, 7975, 7973, 7976, 7978, 7979, 7983, 7981, 7980, 7987, 7985, 7984, 7972, 7973, 7967, 7976, 7977, 7971, 7981, 7975, 7974, 7985, 7979, 7978, 7967, 7970, 7968, 7975, 7978, 7976, 7986, 7984, 7981, 7981, 7984, 7978, 7973, 7976, 7970, 7948, 7966, 7964, 7957, 7974, 7972, 7982, 7980, 7962, 7980, 7974, 7957, 7948, 7956, 7972, 7993, 7994, 7992, 8002, 8000, 7999,
8009, 8007, 8006, 8011, 8008, 8006, 7993, 8003, 8004, 8009, 7998, 7997, 8006, 8007, 8004, 8010, 8006, 8003, 8001, 7999, 7995, 7992, 7994, 7996, 8002, 8001, 7997, 8001, 7996, 7994, 8004, 8007, 7997, 8018, 8020, 8019, 8016, 8020, 8018, 8014, 8018, 8017, 8012, 8015, 8018, 8026, 8024, 8023, 8033, 8035, 8034, 8034, 8031, 8030, 8031, 8026, 8025, 8022, 8025, 8023, 8028, 8029, 8033, 8032, 8030, 8027, 8027, 8030, 8025, 8038, 8039, 8037, 8043, 8041, 8040, 8046, 8047, 8045, 8051, 8049, 8048, 8044, 8045, 8039, 8049, 8043, 8042, 8039, 8042, 8040, 8047, 8050, 8048, 8045, 8048, 8042, 8036, 8037, 8029, 8041, 8035, 8033, 8040, 8033, 8029, 8059, 8061, 8062, 8066, 8068, 8067, 8069, 8071, 8072, 8075, 8076, 8074, 8078, 8080, 8079, 8079, 8082, 8081, 8073, 8074, 8062, 8064, 8080, 8078, 8070, 8078, 8077, 8069, 8077, 8081, 8065, 8067, 8061, 8067, 8068, 8064, 8072, 8075, 8073, 8061, 8067, 8063, 8058, 8070, 8073, 8058, 8063, 8078, 8084, 8086, 8087, 8091, 8093, 8092, 8092, 8095, 8094, 8096, 8098, 8099, 8102, 8103, 8101, 8105, 8107, 8106, 8106, 8109, 8108, 8100, 8101, 8087, 8089, 8107, 8105,
8096, 8097, 8105, 8096, 8104, 8108, 8084, 8085, 8091, 8090, 8094, 8086, 8094, 8095, 8089, 8099, 8102, 8100, 8083, 8086, 8094, 8083, 8097, 8100, 8088, 8105, 8097, 8112, 8113, 8111, 8116, 8117, 8115, 8120, 8121, 8119, 8122, 8124, 8125, 8131, 8129, 8128, 8131, 8127, 8126, 8118, 8119, 8113, 8122, 8123, 8117, 8129, 8125, 8124, 8125, 8129, 8126, 8113, 8116, 8114, 8121, 8124, 8122, 8119, 8122, 8116, 8057, 8093, 8091, 8056, 8091, 8085, 8110, 8111, 8072, 8114, 8115, 8076, 8111, 8114, 8075, 8081, 8082, 8057, 8055, 8071, 8081, 8087, 8112, 8110, 8103, 8120, 8118, 8101, 8118, 8112, 8055, 8085, 8110, 8024, 7990, 7989, 8023, 7989, 7988, 8013, 8059, 8060, 8019, 8066, 8065, 8017, 8065, 8059, 7971, 7993, 7991, 7977, 7979, 8005, 8011, 8010, 7985, 8010, 8005, 7979, 7977, 8003, 7993, 8108, 8109, 8054, 8098, 8108, 8053, 8134, 8135, 8133, 8138, 8139, 8137, 8145, 8143, 8142, 8141, 8144, 8142, 8150, 8151, 8149, 8153, 8155, 8154, 8148, 8149, 8139, 8145, 8153, 8152, 8142, 8143, 8135, 8132, 8140, 8142, 8146, 8152, 8154, 8133, 8135, 8138, 8147, 8150, 8148, 8146, 8148, 8138, 8143, 8152, 8146,
8159, 8160, 8158, 8164, 8166, 8165, 8169, 8170, 8168, 8176, 8174, 8173, 8172, 8175, 8173, 8167, 8168, 8160, 8166, 8172, 8171, 8157, 8163, 8165, 8173, 8174, 8170, 8171, 8173, 8169, 8164, 8163, 8161, 8161, 8163, 8157, 8159, 8165, 8171, 8150, 8157, 8158, 8155, 8162, 8161, 8156, 8157, 8150, 8147, 8154, 8161, 8179, 8180, 8178, 8186, 8184, 8183, 8189, 8190, 8188, 8196, 8194, 8193, 8187, 8188, 8180, 8194, 8186, 8185, 8182, 8185, 8183, 8192, 8195, 8193, 8180, 8182, 8181, 8190, 8192, 8191, 8188, 8191, 8182, 8191, 8193, 8185, 8199, 8201, 8202, 8204, 8206, 8205, 8211, 8212, 8210, 8214, 8216, 8215, 8209, 8210, 8202, 8208, 8214, 8213, 8203, 8205, 8198, 8209, 8213, 8215, 8198, 8201, 8199, 8208, 8207, 8205, 8205, 8207, 8201, 8207, 8213, 8209, 8219, 8220, 8218, 8223, 8224, 8222, 8226, 8228, 8227, 8230, 8228, 8226, 8233, 8234, 8232, 8237, 8238, 8236, 8242, 8240, 8239, 8231, 8232, 8220, 8235, 8236, 8224, 8240, 8230, 8229, 8221, 8222, 8226, 8229, 8226, 8222, 8238, 8241, 8239, 8218, 8220, 8223, 8234, 8237, 8235, 8232, 8235, 8223, 8236, 8239, 8229, 8245, 8246, 8244, 8248, 8250, 8249,
8251, 8253, 8254, 8256, 8258, 8257, 8251, 8252, 8246, 8250, 8256, 8255, 8243, 8247, 8249, 8251, 8255, 8257, 8245, 8249, 8255, 8217, 8218, 8190, 8227, 8196, 8195, 8218, 8221, 8192, 8225, 8195, 8192, 8243, 8244, 8212, 8216, 8248, 8247, 8211, 8215, 8247, 8262, 8264, 8265, 8266, 8268, 8269, 8271, 8273, 8272, 8275, 8277, 8276, 8261, 8277, 8275, 8263, 8271, 8270, 8266, 8267, 8275, 8263, 8265, 8268, 8274, 8276, 8273, 8266, 8274, 8271, 8260, 8275, 8267, 8284, 8286, 8285, 8290, 8292, 8291, 8288, 8292, 8290, 8295, 8296, 8294, 8299, 8300, 8298, 8278, 8293, 8294, 8297, 8298, 8282, 8279, 8284, 8283, 8282, 8290, 8289, 8280, 8287, 8290, 8289, 8291, 8286, 8296, 8299, 8297, 8281, 8289, 8284, 8279, 8294, 8297, 8303, 8304, 8302, 8307, 8308, 8306, 8318, 8316, 8315, 8322, 8320, 8319, 8303, 8311, 8312, 8307, 8313, 8314, 8322, 8310, 8309, 8316, 8312, 8311, 8319, 8320, 8314, 8304, 8307, 8305, 8321, 8319, 8316, 8319, 8313, 8312, 8304, 8312, 8313, 8320, 8309, 8308, 8323, 8325, 8326, 8327, 8329, 8330, 8334, 8332, 8331, 8338, 8336, 8335, 8332, 8326, 8325, 8336, 8330, 8329, 8324, 8326, 8329,
8337, 8335, 8332, 8335, 8329, 8326, 8301, 8302, 8265, 8305, 8306, 8269, 8302, 8305, 8268, 8323, 8324, 8296, 8327, 8328, 8300, 8324, 8327, 8299, 8283, 8285, 8261, 8269, 8295, 8293, 8259, 8278, 8283, 8259, 8267, 8293, 8308, 8325, 8323, 8333, 8331, 8309, 8309, 8331, 8325, 8306, 8323, 8295, 8341, 8342, 8340, 8343, 8345, 8346, 8350, 8348, 8347, 8353, 8354, 8352, 8355, 8357, 8358, 8362, 8360, 8359, 8351, 8352, 8342, 8355, 8356, 8346, 8360, 8350, 8349, 8346, 8349, 8347, 8361, 8359, 8356, 8342, 8345, 8343, 8354, 8357, 8355, 8352, 8355, 8345, 8359, 8349, 8346, 8363, 8365, 8366, 8368, 8370, 8369, 8373, 8374, 8372, 8376, 8378, 8377, 8371, 8372, 8366, 8370, 8376, 8375, 8367, 8369, 8365, 8375, 8377, 8373, 8369, 8375, 8371, 8379, 8381, 8382, 8385, 8386, 8384, 8392, 8390, 8389, 8395, 8396, 8394, 8397, 8399, 8400, 8402, 8404, 8403, 8408, 8406, 8405, 8393, 8394, 8382, 8397, 8398, 8386, 8388, 8402, 8401, 8406, 8392, 8391, 8401, 8403, 8395, 8391, 8389, 8384, 8407, 8405, 8398, 8382, 8385, 8383, 8396, 8399, 8397, 8394, 8397, 8385, 8381, 8387, 8401, 8405, 8391, 8386, 8409, 8411, 8412,
8414, 8416, 8415, 8418, 8420, 8421, 8425, 8427, 8426, 8411, 8418, 8419, 8416, 8425, 8424, 8413, 8415, 8411, 8418, 8424, 8426, 8426, 8427, 8423, 8420, 8426, 8422, 8415, 8424, 8418, 8379, 8380, 8354, 8383, 8384, 8358, 8390, 8362, 8361, 8380, 8383, 8357, 8389, 8361, 8358, 8373, 8409, 8410, 8378, 8414, 8413, 8377, 8413, 8409, 8176, 8288, 8287, 8174, 8287, 8280, 8339, 8340, 8234, 8343, 8344, 8238, 8348, 8242, 8241, 8340, 8343, 8237, 8344, 8347, 8241, 8253, 8363, 8364, 8258, 8368, 8367, 8253, 8257, 8367, 8139, 8179, 8177, 8151, 8189, 8187, 8149, 8187, 8179, 8160, 8219, 8217, 8168, 8170, 8233, 8168, 8231, 8219, 8158, 8217, 8189, 8282, 8341, 8339, 8300, 8353, 8351, 8298, 8351, 8341, 8328, 8330, 8381, 8388, 8387, 8336, 8387, 8381, 8330, 8328, 8379, 8353, 8280, 8339, 8233, 8430, 8431, 8429, 8435, 8433, 8432, 8436, 8438, 8439, 8448, 8450, 8449, 8449, 8446, 8445, 8436, 8437, 8431, 8444, 8435, 8434, 8431, 8434, 8432, 8442, 8448, 8447, 8447, 8445, 8440, 8439, 8441, 8440, 8445, 8446, 8444, 8437, 8440, 8445, 8443, 8434, 8431, 8452, 8454, 8455, 8459, 8461, 8460, 8460, 8463, 8462,
8464, 8466, 8467, 8470, 8471, 8469, 8473, 8475, 8474, 8474, 8477, 8476, 8468, 8469, 8455, 8457, 8475, 8473, 8464, 8465, 8473, 8472, 8476, 8466, 8453, 8459, 8458, 8452, 8458, 8462, 8462, 8463, 8457, 8467, 8470, 8468, 8451, 8454, 8462, 8451, 8465, 8468, 8456, 8473, 8465, 8480, 8481, 8479, 8484, 8485, 8483, 8488, 8489, 8487, 8492, 8493, 8491, 8486, 8487, 8481, 8490, 8491, 8485, 8479, 8481, 8484, 8489, 8492, 8490, 8487, 8490, 8484, 8496, 8497, 8495, 8500, 8501, 8499, 8502, 8504, 8505, 8512, 8510, 8509, 8516, 8514, 8513, 8518, 8515, 8513, 8502, 8503, 8497, 8500, 8506, 8507, 8510, 8505, 8504, 8514, 8507, 8506, 8517, 8513, 8506, 8497, 8500, 8498, 8503, 8505, 8508, 8518, 8517, 8510, 8517, 8508, 8505, 8503, 8506, 8500, 8478, 8479, 8439, 8482, 8483, 8442, 8479, 8482, 8441, 8494, 8495, 8467, 8498, 8499, 8471, 8495, 8498, 8470, 8476, 8477, 8450, 8466, 8476, 8448, 8485, 8496, 8494, 8493, 8504, 8502, 8491, 8502, 8496, 8442, 8483, 8494, 8521, 8522, 8520, 8525, 8526, 8524, 8529, 8530, 8528, 8531, 8533, 8534, 8540, 8538, 8537, 8540, 8536, 8535, 8527, 8528, 8522, 8531, 8532, 8526,
8537, 8538, 8534, 8538, 8535, 8532, 8522, 8525, 8523, 8530, 8533, 8531, 8528, 8531, 8525, 8541, 8543, 8544, 8551, 8549, 8548, 8551, 8547, 8546, 8556, 8550, 8548, 8548, 8549, 8544, 8549, 8546, 8542, 8553, 8552, 8546, 8552, 8545, 8542, 8555, 8548, 8543, 8557, 8559, 8560, 8567, 8565, 8564, 8567, 8563, 8562, 8572, 8566, 8564, 8565, 8560, 8559, 8565, 8562, 8558, 8569, 8568, 8562, 8562, 8568, 8561, 8570, 8571, 8564, 8557, 8558, 8530, 8569, 8539, 8537, 8530, 8558, 8561, 8568, 8537, 8533, 8526, 8543, 8541, 8556, 8555, 8535, 8535, 8555, 8554, 8526, 8532, 8554, 8575, 8576, 8574, 8585, 8583, 8582, 8587, 8584, 8582, 8588, 8590, 8591, 8595, 8593, 8592, 8575, 8588, 8589, 8593, 8581, 8580, 8594, 8592, 8589, 8582, 8583, 8578, 8586, 8582, 8577, 8576, 8579, 8577, 8587, 8586, 8580, 8580, 8586, 8579, 8592, 8580, 8576, 8599, 8601, 8602, 8606, 8604, 8603, 8613, 8611, 8610, 8615, 8612, 8610, 8601, 8607, 8608, 8613, 8606, 8605, 8605, 8603, 8600, 8611, 8608, 8607, 8609, 8614, 8610, 8608, 8611, 8605, 8590, 8599, 8600, 8604, 8595, 8594, 8603, 8594, 8591, 8598, 8597, 8583, 8597, 8596, 8578,
8573, 8574, 8489, 8492, 8577, 8578, 8574, 8577, 8492, 8598, 8511, 8509, 8597, 8509, 8504, 8493, 8578, 8596, 8455, 8521, 8519, 8471, 8529, 8527, 8469, 8527, 8521, 8501, 8559, 8557, 8572, 8571, 8514, 8571, 8570, 8507, 8501, 8507, 8570, 8499, 8557, 8529, 8132, 8133, 7853, 8136, 8137, 7857, 7859, 8141, 8140, 8133, 8136, 7856, 7852, 7858, 8140, 7910, 8177, 8178, 8184, 7918, 7917, 8178, 8181, 7912, 8183, 7917, 7912, 8199, 8200, 7938, 7942, 8204, 8203, 8197, 8199, 7937, 7934, 7941, 8203, 8137, 8177, 7910, 8428, 8429, 8047, 8433, 8051, 8050, 8429, 8432, 8050, 8054, 8461, 8459, 8052, 8053, 8459, 8519, 8520, 8099, 8523, 8524, 8103, 8520, 8523, 8102, 8541, 8542, 8121, 8553, 8130, 8128, 8121, 8542, 8545, 8545, 8552, 8128, 8524, 8541, 8120, 8052, 8453, 8519, 7989, 7990, 7776, 7772, 7988, 7989, 7885, 8022, 8021, 7893, 8028, 8027, 7891, 8027, 8022, 7926, 8038, 8036, 7938, 8046, 8044, 7936, 8044, 8038, 7893, 7924, 8036, 7883, 8021, 7988, 8202, 8430, 8428, 8212, 8438, 8436, 8210, 8436, 8430, 8246, 8480, 8478, 8254, 8488, 8486, 8252, 8486, 8480, 8244, 8478, 8438, 8366, 8575, 8573,
8374, 8590, 8588, 8372, 8588, 8575, 8412, 8601, 8599, 8419, 8421, 8609, 8419, 8607, 8601, 8410, 8599, 8590, 8364, 8573, 8488, 8200, 8428, 8046, 8616, 8618, 8619, 8621, 8623, 8622, 8628, 8629, 8627, 8631, 8633, 8632, 8618, 8626, 8627, 8623, 8635, 8634, 8630, 8632, 8625, 8620, 8622, 8618, 8624, 8625, 8628, 8634, 8635, 8631, 8626, 8634, 8630, 8618, 8622, 8634, 8638, 8639, 8637, 8642, 8643, 8641, 8648, 8649, 8647, 8650, 8652, 8653, 8659, 8657, 8656, 8659, 8655, 8654, 8646, 8647, 8639, 8642, 8650, 8651, 8655, 8645, 8644, 8657, 8653, 8652, 8653, 8657, 8654, 8639, 8642, 8640, 8649, 8652, 8650, 8647, 8650, 8642, 8654, 8644, 8643, 8660, 8662, 8663, 8666, 8667, 8665, 8669, 8671, 8670, 8670, 8673, 8672, 8676, 8677, 8675, 8680, 8681, 8679, 8662, 8674, 8675, 8678, 8679, 8667, 8673, 8683, 8682, 8661, 8669, 8668, 8668, 8672, 8662, 8663, 8666, 8664, 8677, 8680, 8678, 8675, 8678, 8666, 8672, 8682, 8674, 8686, 8687, 8685, 8695, 8693, 8692, 8696, 8698, 8699, 8705, 8703, 8702, 8705, 8701, 8700, 8686, 8696, 8697, 8701, 8691, 8690, 8703, 8699, 8698, 8703, 8700, 8697, 8694, 8692, 8688,
8685, 8687, 8689, 8695, 8694, 8690, 8694, 8689, 8687, 8700, 8690, 8687, 8664, 8665, 8629, 8633, 8671, 8669, 8625, 8661, 8664, 8632, 8669, 8661, 8684, 8685, 8649, 8693, 8658, 8656, 8649, 8685, 8688, 8688, 8692, 8656, 8619, 8638, 8636, 8629, 8648, 8646, 8627, 8646, 8638, 8667, 8686, 8684, 8681, 8698, 8696, 8679, 8696, 8686, 8665, 8684, 8648, 8711, 8709, 8708, 8710, 8708, 8706, 8714, 8715, 8713, 8718, 8719, 8717, 8724, 8725, 8723, 8732, 8730, 8729, 8734, 8731, 8729, 8722, 8723, 8715, 8718, 8726, 8727, 8732, 8721, 8720, 8730, 8727, 8726, 8733, 8729, 8726, 8715, 8718, 8716, 8723, 8725, 8728, 8723, 8726, 8718, 8727, 8730, 8720, 8740, 8738, 8737, 8739, 8737, 8735, 8741, 8743, 8744, 8751, 8749, 8748, 8751, 8747, 8746, 8756, 8750, 8748, 8749, 8744, 8743, 8744, 8749, 8746, 8753, 8752, 8746, 8752, 8745, 8742, 8754, 8755, 8748, 8741, 8742, 8725, 8753, 8734, 8733, 8725, 8742, 8745, 8752, 8733, 8728, 8717, 8719, 8736, 8740, 8739, 8720, 8739, 8736, 8719, 8712, 8713, 8677, 8716, 8717, 8681, 8713, 8716, 8680, 8738, 8704, 8702, 8737, 8702, 8698, 8681, 8717, 8735, 8641, 8643, 8707,
8711, 8710, 8644, 8710, 8707, 8643, 7606, 7607, 6800, 7616, 6808, 6807, 7607, 7610, 6802, 7610, 7614, 6807, 6834, 7634, 7633, 6826, 6832, 7633, 7686, 7687, 6875, 7690, 7691, 6879, 7687, 7690, 6878, 7710, 7711, 6885, 7716, 6891, 6889, 7711, 7715, 6889, 6879, 7691, 7710, 7629, 7686, 6874, 7945, 7946, 6965, 6969, 7950, 7949, 7943, 7945, 6964, 6961, 6968, 7949, 7964, 7965, 6981, 7968, 7969, 6985, 7965, 7968, 6984, 7946, 7964, 6980, 7991, 7992, 7023, 8000, 7029, 7028, 7023, 7992, 7995, 7999, 7028, 7025, 7969, 7991, 7022, 7540, 8616, 8617, 7546, 8621, 8620, 7544, 8620, 8616, 8636, 8637, 7561, 8640, 8641, 7565, 8637, 8640, 7564, 8617, 8636, 7560, 8709, 7602, 7601, 8708, 7601, 7596, 7565, 8641, 8706, 6894, 6896, 7323, 7329, 7330, 6902, 6894, 7321, 7329, 6910, 7365, 7363, 6920, 7373, 7371, 6918, 7371, 7365, 6908, 7363, 7323, 7486, 7485, 6997, 7487, 7488, 7003, 7497, 7495, 7005, 6997, 7485, 7477, 7479, 7487, 7001, 7495, 7491, 7004, 7004, 7491, 7479, 6920, 6991, 7477, 8012, 8014, 8676, 8682, 8683, 8016, 8674, 8682, 8015, 8062, 8714, 8712, 8076, 8724, 8722, 8074, 8722, 8714,
8117, 8743, 8741, 8756, 8755, 8126, 8755, 8754, 8123, 8117, 8123, 8754, 8115, 8741, 8724, 8060, 8712, 8676, 5900, 6453, 6454, 6457, 6458, 5905, 6454, 6457, 5904, 6481, 6482, 5917, 6485, 6486, 5921, 6482, 6485, 5920, 6458, 6481, 5916, 6516, 6517, 5965, 5968, 6520, 6521, 6525, 5971, 5970, 6517, 6520, 5968, 6521, 6524, 5970, 6486, 6516, 5964, 7030, 7031, 6338, 7033, 7034, 6342, 6344, 7039, 7037, 6338, 7031, 7033, 6343, 7037, 7030, 7062, 7063, 6360, 7066, 7067, 6364, 7063, 7066, 6363, 6342, 7034, 7062, 7118, 7119, 6421, 7122, 7123, 6425, 7119, 7122, 6424, 7136, 7137, 6442, 7145, 6451, 6449, 6442, 7137, 7140, 7144, 6449, 6445, 7123, 7136, 6441, 7067, 7118, 6420, 8760, 8762, 8761, 8766, 8768, 8767, 8762, 8766, 8765, 8757, 8759, 8761, 8765, 8767, 8764, 8761, 8765, 8763, 8769, 8771, 8772, 8773, 8775, 8776, 8778, 8780, 8779, 8786, 8784, 8783, 8786, 8782, 8781, 8787, 8789, 8790, 8794, 8796, 8795, 8800, 8798, 8797, 8787, 8788, 8772, 8780, 8794, 8793, 8798, 8785, 8783, 8777, 8779, 8771, 8793, 8795, 8789, 8784, 8776, 8775, 8776, 8784, 8781, 8799, 8797, 8791, 8772, 8775, 8773,
8788, 8790, 8792, 8772, 8788, 8791, 8779, 8793, 8787, 8791, 8797, 8783, 8805, 8806, 8804, 8808, 8810, 8809, 8816, 8818, 8819, 8823, 8825, 8824, 8821, 8825, 8823, 8829, 8827, 8826, 8805, 8816, 8817, 8810, 8821, 8820, 8827, 8812, 8811, 8807, 8809, 8802, 8815, 8823, 8822, 8813, 8820, 8823, 8828, 8826, 8817, 8801, 8802, 8805, 8813, 8815, 8818, 8802, 8813, 8816, 8809, 8820, 8813, 8817, 8826, 8811, 8832, 8833, 8831, 8843, 8841, 8840, 8844, 8846, 8847, 8853, 8851, 8850, 8855, 8857, 8856, 8832, 8844, 8845, 8851, 8839, 8838, 8852, 8850, 8845, 8835, 8842, 8840, 8848, 8854, 8856, 8831, 8833, 8835, 8843, 8842, 8838, 8842, 8835, 8833, 8850, 8838, 8833, 8858, 8860, 8861, 8866, 8868, 8867, 8872, 8870, 8869, 8874, 8871, 8869, 8877, 8878, 8876, 8885, 8883, 8881, 8875, 8876, 8861, 8868, 8882, 8880, 8887, 8874, 8873, 8865, 8867, 8860, 8884, 8881, 8876, 8870, 8863, 8862, 8873, 8869, 8862, 8859, 8861, 8864, 8887, 8886, 8881, 8886, 8879, 8876, 8861, 8876, 8879, 8860, 8867, 8880, 8886, 8873, 8864, 8789, 8830, 8831, 8796, 8837, 8836, 8841, 8800, 8799, 8790, 8831, 8834, 8789, 8795, 8836,
8840, 8799, 8792, 8814, 8858, 8859, 8818, 8862, 8863, 8824, 8866, 8865, 8872, 8829, 8828, 8859, 8862, 8818, 8814, 8822, 8865, 8870, 8828, 8819, 8848, 8849, 8877, 8880, 8882, 8855, 8875, 8880, 8854, 8894, 8896, 8895, 8892, 8896, 8894, 8897, 8899, 8900, 8903, 8907, 8905, 8908, 8904, 8902, 8897, 8898, 8890, 8895, 8903, 8901, 8890, 8894, 8893, 8891, 8894, 8890, 8901, 8905, 8899, 8906, 8902, 8898, 8893, 8901, 8897, 8914, 8912, 8911, 8912, 8910, 8909, 8909, 8910, 8900, 8907, 8913, 8911, 8914, 8908, 8906, 8905, 8911, 8909, 8912, 8906, 8900, 8920, 8921, 8919, 8925, 8923, 8922, 8927, 8924, 8922, 8929, 8931, 8930, 8930, 8934, 8932, 8941, 8942, 8940, 8946, 8948, 8947, 8956, 8954, 8953, 8956, 8952, 8950, 8920, 8939, 8940, 8945, 8927, 8926, 8934, 8951, 8949, 8952, 8935, 8933, 8923, 8916, 8915, 8926, 8922, 8915, 8944, 8947, 8938, 8919, 8929, 8928, 8928, 8932, 8920, 8954, 8942, 8941, 8954, 8950, 8940, 8938, 8941, 8939, 8949, 8951, 8946, 8939, 8949, 8944, 8943, 8926, 8917, 8932, 8949, 8939, 8950, 8933, 8921, 8963, 8961, 8960, 8965, 8962, 8960, 8961, 8958, 8957, 8959, 8964, 8960,
8966, 8968, 8969, 8972, 8974, 8973, 8978, 8976, 8975, 8973, 8977, 8975, 8978, 8974, 8972, 8967, 8972, 8971, 8976, 8969, 8968, 8966, 8971, 8975, 8969, 8976, 8972, 8980, 8979, 8972, 8979, 8970, 8967, 8948, 8974, 8972, 8980, 8955, 8953, 8938, 8967, 8970, 8947, 8972, 8967, 8970, 8979, 8953, 8921, 8959, 8957, 8965, 8964, 8933, 8964, 8959, 8921, 8768, 8892, 8891, 8767, 8891, 8888, 8846, 8915, 8916, 8925, 8853, 8852, 8857, 8931, 8929, 8923, 8852, 8847, 8856, 8929, 8919, 8877, 8957, 8958, 8963, 8885, 8884, 8961, 8884, 8878, 8919, 8957, 8877, 8757, 8758, 8832, 8836, 8837, 8760, 8763, 8764, 8846, 8830, 8836, 8759, 8758, 8763, 8844, 8888, 8890, 8917, 8945, 8943, 8902, 8943, 8936, 8898, 8898, 8936, 8917, 8888, 8915, 8846, 8986, 8984, 8983, 8982, 8985, 8983, 8992, 8994, 8993, 8996, 8998, 8997, 9001, 9002, 9000, 9005, 9006, 9004, 8999, 9000, 8988, 9003, 9004, 8990, 8988, 8992, 8991, 8990, 8996, 8995, 8995, 8997, 8994, 9002, 9005, 9003, 8989, 8995, 8992, 9000, 9003, 8989, 9012, 9014, 9013, 9016, 9018, 9017, 9021, 9022, 9020, 9025, 9026, 9024, 9019, 9020, 9008, 9023, 9024, 9010,
9008, 9012, 9011, 9010, 9016, 9015, 9015, 9017, 9014, 9022, 9025, 9023, 9009, 9015, 9012, 9020, 9023, 9009, 9029, 9030, 9028, 9033, 9034, 9032, 9042, 9040, 9039, 9046, 9044, 9043, 9035, 9036, 9030, 9037, 9038, 9034, 9040, 9036, 9035, 9044, 9038, 9037, 9030, 9033, 9031, 9045, 9043, 9040, 9043, 9037, 9036, 9036, 9037, 9033, 9049, 9050, 9048, 9053, 9054, 9052, 9055, 9057, 9058, 9061, 9062, 9060, 9064, 9066, 9065, 9049, 9055, 9056, 9059, 9060, 9054, 9063, 9065, 9057, 9050, 9053, 9051, 9058, 9061, 9059, 9056, 9059, 9053, 9027, 9028, 9002, 9031, 9032, 9006, 9028, 9031, 9005, 9047, 9048, 9022, 9051, 9052, 9026, 9048, 9051, 9025, 9011, 9013, 8998, 9006, 9021, 9019, 9007, 9011, 8996, 9004, 9019, 9007, 9034, 9049, 9047, 9064, 9063, 9044, 9063, 9055, 9038, 9038, 9055, 9049, 9032, 9047, 9021, 9072, 9074, 9073, 9076, 9078, 9077, 9081, 9082, 9080, 9085, 9086, 9084, 9079, 9080, 9068, 9083, 9084, 9070, 9068, 9072, 9071, 9070, 9076, 9075, 9075, 9077, 9074, 9082, 9085, 9083, 9069, 9075, 9072, 9080, 9083, 9069, 9092, 9094, 9093, 9096, 9098, 9097, 9101, 9102, 9100, 9105, 9106, 9104,
9099, 9100, 9088, 9103, 9104, 9090, 9088, 9092, 9091, 9090, 9096, 9095, 9095, 9097, 9094, 9102, 9105, 9103, 9089, 9095, 9092, 9100, 9103, 9089, 9109, 9110, 9108, 9113, 9114, 9112, 9117, 9118, 9116, 9125, 9123, 9122, 9127, 9124, 9122, 9115, 9116, 9110, 9113, 9119, 9120, 9123, 9120, 9119, 9126, 9122, 9119, 9110, 9113, 9111, 9118, 9121, 9119, 9116, 9119, 9113, 9130, 9131, 9129, 9134, 9135, 9133, 9143, 9141, 9140, 9147, 9145, 9144, 9130, 9136, 9137, 9134, 9138, 9139, 9140, 9141, 9137, 9144, 9145, 9139, 9131, 9134, 9132, 9146, 9144, 9141, 9141, 9144, 9138, 9131, 9137, 9138, 9107, 9108, 9082, 9111, 9112, 9086, 9108, 9111, 9085, 9128, 9129, 9102, 9132, 9133, 9106, 9129, 9132, 9105, 9091, 9093, 9078, 9086, 9101, 9099, 9087, 9091, 9076, 9084, 9099, 9087, 9114, 9130, 9128, 9142, 9140, 9123, 9140, 9136, 9120, 9120, 9136, 9130, 9112, 9128, 9101, 9071, 9073, 9018, 9026, 9081, 9079, 9067, 9071, 9016, 9024, 9079, 9067, 9054, 9109, 9107, 9062, 9117, 9115, 9060, 9115, 9109, 9052, 9107, 9081, 9153, 9155, 9154, 9157, 9159, 9158, 9162, 9163, 9161, 9166, 9167, 9165, 9160, 9161, 9149,
9164, 9165, 9151, 9149, 9153, 9152, 9151, 9157, 9156, 9156, 9158, 9155, 9163, 9166, 9164, 9150, 9156, 9153, 9161, 9164, 9150, 9173, 9175, 9174, 9177, 9179, 9178, 9182, 9183, 9181, 9186, 9187, 9185, 9191, 9189, 9188, 9180, 9181, 9169, 9184, 9185, 9171, 9189, 9179, 9177, 9169, 9173, 9172, 9171, 9177, 9176, 9190, 9188, 9185, 9176, 9178, 9175, 9183, 9186, 9184, 9170, 9176, 9173, 9181, 9184, 9170, 9188, 9177, 9171, 9194, 9195, 9193, 9198, 9199, 9197, 9202, 9203, 9201, 9206, 9207, 9205, 9213, 9211, 9210, 9209, 9212, 9210, 9194, 9200, 9201, 9204, 9205, 9199, 9211, 9203, 9202, 9200, 9208, 9210, 9195, 9198, 9196, 9201, 9203, 9206, 9201, 9204, 9198, 9216, 9217, 9215, 9220, 9221, 9219, 9225, 9223, 9222, 9228, 9229, 9227, 9232, 9233, 9231, 9237, 9235, 9234, 9226, 9227, 9217, 9230, 9231, 9221, 9235, 9225, 9224, 9224, 9222, 9219, 9236, 9234, 9231, 9217, 9220, 9218, 9229, 9232, 9230, 9227, 9230, 9220, 9234, 9224, 9221, 9192, 9193, 9163, 9196, 9197, 9167, 9193, 9196, 9166, 9214, 9215, 9183, 9218, 9219, 9187, 9223, 9191, 9190, 9215, 9218, 9186, 9222, 9190, 9187, 9172, 9174, 9159,
9167, 9182, 9180, 9168, 9172, 9157, 9165, 9180, 9168, 9199, 9216, 9214, 9207, 9228, 9226, 9205, 9226, 9216, 9197, 9214, 9182, 8804, 8806, 8982, 8986, 8985, 8811, 8985, 8982, 8806, 9152, 9154, 9098, 9106, 9162, 9160, 9148, 9152, 9096, 9104, 9160, 9148, 9135, 9194, 9192, 9209, 9208, 9145, 9208, 9200, 9139, 9139, 9200, 9194, 9133, 9192, 9162, 9240, 9241, 9239, 9244, 9245, 9243, 9247, 9249, 9248, 9252, 9253, 9251, 9256, 9257, 9255, 9259, 9261, 9260, 9250, 9251, 9241, 9254, 9255, 9245, 9249, 9259, 9258, 9246, 9248, 9240, 9258, 9260, 9252, 9241, 9244, 9242, 9253, 9256, 9254, 9251, 9254, 9244, 9248, 9258, 9250, 9264, 9265, 9263, 9268, 9269, 9267, 9271, 9273, 9272, 9276, 9277, 9275, 9280, 9281, 9279, 9283, 9285, 9284, 9274, 9275, 9265, 9278, 9279, 9269, 9273, 9283, 9282, 9270, 9272, 9264, 9282, 9284, 9276, 9265, 9268, 9266, 9277, 9280, 9278, 9275, 9278, 9268, 9272, 9282, 9274, 9262, 9263, 9253, 9266, 9267, 9257, 9261, 9271, 9270, 9263, 9266, 9256, 9260, 9270, 9262, 9288, 9289, 9287, 9295, 9293, 9292, 9298, 9299, 9297, 9305, 9303, 9302, 9296, 9297, 9289, 9303, 9295, 9294,
9294, 9292, 9290, 9304, 9302, 9300, 9289, 9291, 9290, 9299, 9301, 9300, 9297, 9300, 9291, 9302, 9294, 9291, 9308, 9309, 9307, 9315, 9313, 9312, 9318, 9319, 9317, 9325, 9323, 9322, 9316, 9317, 9309, 9323, 9315, 9314, 9314, 9312, 9310, 9324, 9322, 9320, 9309, 9311, 9310, 9319, 9321, 9320, 9317, 9320, 9311, 9322, 9314, 9311, 9306, 9307, 9299, 9313, 9305, 9304, 9307, 9310, 9301, 9312, 9304, 9301, 9328, 9329, 9327, 9332, 9333, 9331, 9335, 9337, 9336, 9340, 9341, 9339, 9344, 9345, 9343, 9347, 9349, 9348, 9338, 9339, 9329, 9342, 9343, 9333, 9337, 9347, 9346, 9334, 9336, 9328, 9346, 9348, 9340, 9329, 9332, 9330, 9341, 9344, 9342, 9339, 9342, 9332, 9336, 9346, 9338, 9352, 9353, 9351, 9356, 9357, 9355, 9359, 9361, 9360, 9364, 9365, 9363, 9368, 9369, 9367, 9371, 9373, 9372, 9362, 9363, 9353, 9366, 9367, 9357, 9361, 9371, 9370, 9358, 9360, 9352, 9370, 9372, 9364, 9353, 9356, 9354, 9365, 9368, 9366, 9363, 9366, 9356, 9360, 9370, 9362, 9350, 9351, 9341, 9354, 9355, 9345, 9349, 9359, 9358, 9351, 9354, 9344, 9348, 9358, 9350, 9376, 9377, 9375, 9383, 9381, 9380, 9386, 9387, 9385,
9393, 9391, 9390, 9384, 9385, 9377, 9391, 9383, 9382, 9382, 9380, 9378, 9392, 9390, 9388, 9377, 9379, 9378, 9387, 9389, 9388, 9385, 9388, 9379, 9390, 9382, 9379, 9396, 9397, 9395, 9403, 9401, 9400, 9406, 9407, 9405, 9413, 9411, 9410, 9404, 9405, 9397, 9411, 9403, 9402, 9402, 9400, 9398, 9412, 9410, 9408, 9397, 9399, 9398, 9407, 9409, 9408, 9405, 9408, 9399, 9410, 9402, 9399, 9394, 9395, 9387, 9401, 9393, 9392, 9395, 9398, 9389, 9400, 9392, 9389, 9326, 9327, 9277, 9330, 9331, 9281, 9285, 9335, 9334, 9327, 9330, 9280, 9284, 9334, 9326, 9374, 9375, 9319, 9381, 9325, 9324, 9375, 9378, 9321, 9380, 9324, 9321, 9245, 9288, 9286, 9257, 9298, 9296, 9255, 9296, 9288, 9269, 9308, 9306, 9281, 9318, 9316, 9279, 9316, 9308, 9267, 9306, 9298, 9333, 9376, 9374, 9345, 9386, 9384, 9343, 9384, 9376, 9357, 9396, 9394, 9369, 9406, 9404, 9367, 9404, 9396, 9355, 9394, 9386, 9331, 9374, 9318, 9417, 9418, 9416, 9422, 9424, 9423, 9425, 9427, 9428, 9430, 9432, 9431, 9425, 9426, 9418, 9424, 9430, 9429, 9421, 9423, 9417, 9425, 9429, 9431, 9422, 9421, 9419, 9419, 9421, 9415, 9417, 9423, 9429,
9435, 9436, 9434, 9439, 9440, 9438, 9444, 9442, 9441, 9447, 9448, 9446, 9451, 9452, 9450, 9456, 9454, 9453, 9445, 9446, 9436, 9449, 9450, 9440, 9454, 9444, 9443, 9443, 9441, 9438, 9455, 9453, 9450, 9436, 9439, 9437, 9448, 9451, 9449, 9446, 9449, 9439, 9453, 9443, 9440, 9459, 9460, 9458, 9462, 9464, 9463, 9467, 9468, 9466, 9470, 9472, 9471, 9465, 9466, 9460, 9464, 9470, 9469, 9461, 9463, 9459, 9469, 9471, 9467, 9463, 9469, 9465, 9475, 9476, 9474, 9479, 9480, 9478, 9484, 9482, 9481, 9487, 9488, 9486, 9491, 9492, 9490, 9496, 9494, 9493, 9485, 9486, 9476, 9489, 9490, 9480, 9494, 9484, 9483, 9483, 9481, 9478, 9495, 9493, 9490, 9476, 9479, 9477, 9488, 9491, 9489, 9486, 9489, 9479, 9493, 9483, 9480, 9457, 9458, 9428, 9432, 9462, 9461, 9431, 9461, 9457, 9473, 9474, 9448, 9477, 9478, 9452, 9482, 9456, 9455, 9474, 9477, 9451, 9481, 9455, 9452, 9418, 9435, 9433, 9428, 9447, 9445, 9426, 9445, 9435, 9460, 9475, 9473, 9468, 9487, 9485, 9466, 9485, 9475, 9458, 9473, 9447, 9499, 9500, 9498, 9502, 9504, 9503, 9507, 9508, 9506, 9510, 9512, 9511, 9505, 9506, 9500, 9504, 9510, 9509,
9501, 9503, 9499, 9509, 9511, 9507, 9503, 9509, 9505, 9515, 9516, 9514, 9519, 9520, 9518, 9524, 9522, 9521, 9527, 9528, 9526, 9531, 9532, 9530, 9536, 9534, 9533, 9525, 9526, 9516, 9529, 9530, 9520, 9534, 9524, 9523, 9523, 9521, 9518, 9535, 9533, 9530, 9516, 9519, 9517, 9528, 9531, 9529, 9526, 9529, 9519, 9533, 9523, 9520, 9539, 9540, 9538, 9542, 9544, 9543, 9547, 9548, 9546, 9550, 9552, 9551, 9545, 9546, 9540, 9544, 9550, 9549, 9541, 9543, 9539, 9549, 9551, 9547, 9543, 9549, 9545, 9555, 9556, 9554, 9559, 9560, 9558, 9564, 9562, 9561, 9567, 9568, 9566, 9571, 9572, 9570, 9576, 9574, 9573, 9565, 9566, 9556, 9569, 9570, 9560, 9574, 9564, 9563, 9563, 9561, 9558, 9575, 9573, 9570, 9556, 9559, 9557, 9568, 9571, 9569, 9566, 9569, 9559, 9573, 9563, 9560, 9537, 9538, 9508, 9512, 9542, 9541, 9511, 9541, 9537, 9553, 9554, 9528, 9557, 9558, 9532, 9562, 9536, 9535, 9554, 9557, 9531, 9561, 9535, 9532, 9500, 9515, 9513, 9508, 9527, 9525, 9506, 9525, 9515, 9540, 9555, 9553, 9548, 9567, 9565, 9546, 9565, 9555, 9538, 9553, 9527, 9497, 9498, 9468, 9472, 9502, 9501, 9471, 9501, 9497,
9513, 9514, 9488, 9517, 9518, 9492, 9522, 9496, 9495, 9514, 9517, 9491, 9521, 9495, 9492, 9498, 9513, 9487, 9579, 9580, 9578, 9583, 9584, 9582, 9586, 9588, 9587, 9591, 9592, 9590, 9595, 9596, 9594, 9598, 9600, 9599, 9589, 9590, 9580, 9593, 9594, 9584, 9588, 9598, 9597, 9585, 9587, 9579, 9597, 9599, 9591, 9580, 9583, 9581, 9592, 9595, 9593, 9590, 9593, 9583, 9587, 9597, 9589, 9608, 9606, 9605, 9612, 9610, 9609, 9606, 9602, 9601, 9610, 9604, 9603, 9611, 9609, 9606, 9609, 9603, 9602, 9601, 9602, 9592, 9603, 9604, 9596, 9600, 9607, 9605, 9602, 9603, 9595, 9599, 9605, 9601, 9615, 9616, 9614, 9622, 9620, 9619, 9625, 9626, 9624, 9632, 9630, 9629, 9623, 9624, 9616, 9630, 9622, 9621, 9621, 9619, 9617, 9631, 9629, 9627, 9616, 9618, 9617, 9626, 9628, 9627, 9624, 9627, 9618, 9629, 9621, 9618, 9639, 9637, 9636, 9637, 9634, 9633, 9641, 9640, 9637, 9640, 9635, 9634, 9633, 9634, 9626, 9641, 9632, 9631, 9634, 9635, 9628, 9640, 9631, 9628, 9647, 9649, 9648, 9651, 9653, 9652, 9656, 9657, 9655, 9660, 9661, 9659, 9654, 9655, 9643, 9658, 9659, 9645, 9643, 9647, 9646, 9645, 9651, 9650,
9650, 9652, 9649, 9657, 9660, 9658, 9644, 9650, 9647, 9655, 9658, 9644, 9667, 9669, 9668, 9671, 9673, 9672, 9676, 9677, 9675, 9680, 9681, 9679, 9674, 9675, 9663, 9678, 9679, 9665, 9663, 9667, 9666, 9665, 9671, 9670, 9670, 9672, 9669, 9677, 9680, 9678, 9664, 9670, 9667, 9675, 9678, 9664, 9684, 9685, 9683, 9688, 9689, 9687, 9697, 9695, 9694, 9701, 9699, 9698, 9684, 9690, 9691, 9692, 9693, 9689, 9695, 9691, 9690, 9699, 9693, 9692, 9685, 9688, 9686, 9700, 9698, 9695, 9698, 9692, 9691, 9691, 9692, 9688, 9704, 9705, 9703, 9708, 9709, 9707, 9717, 9715, 9714, 9721, 9719, 9718, 9710, 9711, 9705, 9712, 9713, 9709, 9715, 9711, 9710, 9719, 9713, 9712, 9705, 9708, 9706, 9720, 9718, 9715, 9718, 9712, 9711, 9711, 9712, 9708, 9682, 9683, 9657, 9686, 9687, 9661, 9683, 9686, 9660, 9702, 9703, 9677, 9706, 9707, 9681, 9703, 9706, 9680, 9666, 9668, 9653, 9661, 9676, 9674, 9662, 9666, 9651, 9659, 9674, 9662, 9689, 9704, 9702, 9716, 9714, 9699, 9714, 9710, 9693, 9693, 9710, 9704, 9687, 9702, 9676, 9727, 9729, 9728, 9731, 9733, 9732, 9736, 9737, 9735, 9740, 9741, 9739, 9734, 9735, 9723,
9738, 9739, 9725, 9723, 9727, 9726, 9725, 9731, 9730, 9730, 9732, 9729, 9737, 9740, 9738, 9724, 9730, 9727, 9735, 9738, 9724, 9747, 9749, 9748, 9751, 9753, 9752, 9756, 9757, 9755, 9760, 9761, 9759, 9754, 9755, 9743, 9758, 9759, 9745, 9743, 9747, 9746, 9745, 9751, 9750, 9750, 9752, 9749, 9757, 9760, 9758, 9744, 9750, 9747, 9755, 9758, 9744, 9764, 9765, 9763, 9768, 9769, 9767, 9777, 9775, 9774, 9781, 9779, 9778, 9770, 9771, 9765, 9772, 9773, 9769, 9775, 9771, 9770, 9779, 9773, 9772, 9765, 9768, 9766, 9780, 9778, 9775, 9778, 9772, 9771, 9771, 9772, 9768, 9784, 9785, 9783, 9788, 9789, 9787, 9797, 9795, 9794, 9801, 9799, 9798, 9790, 9791, 9785, 9792, 9793, 9789, 9795, 9791, 9790, 9799, 9793, 9792, 9785, 9788, 9786, 9800, 9798, 9795, 9798, 9792, 9791, 9791, 9792, 9788, 9762, 9763, 9737, 9766, 9767, 9741, 9763, 9766, 9740, 9782, 9783, 9757, 9786, 9787, 9761, 9783, 9786, 9760, 9746, 9748, 9733, 9741, 9756, 9754, 9742, 9746, 9731, 9739, 9754, 9742, 9769, 9784, 9782, 9796, 9794, 9779, 9794, 9790, 9773, 9773, 9790, 9784, 9767, 9782, 9756, 9584, 9615, 9613, 9596, 9625, 9623,
9594, 9623, 9615, 9638, 9636, 9610, 9636, 9633, 9604, 9604, 9633, 9625, 9726, 9728, 9673, 9681, 9736, 9734, 9722, 9726, 9671, 9679, 9734, 9722, 9709, 9764, 9762, 9776, 9774, 9719, 9774, 9770, 9713, 9713, 9770, 9764, 9707, 9762, 9736, 9804, 9805, 9803, 9807, 9809, 9808, 9812, 9813, 9811, 9815, 9817, 9816, 9810, 9811, 9805, 9809, 9815, 9814, 9806, 9808, 9804, 9814, 9816, 9812, 9808, 9814, 9810, 9820, 9821, 9819, 9824, 9825, 9823, 9829, 9827, 9826, 9832, 9833, 9831, 9836, 9837, 9835, 9841, 9839, 9838, 9830, 9831, 9821, 9834, 9835, 9825, 9839, 9829, 9828, 9828, 9826, 9823, 9840, 9838, 9835, 9821, 9824, 9822, 9833, 9836, 9834, 9831, 9834, 9824, 9838, 9828, 9825, 9844, 9845, 9843, 9851, 9849, 9848, 9847, 9850, 9848, 9849, 9845, 9844, 9846, 9848, 9844, 9854, 9855, 9853, 9858, 9859, 9857, 9863, 9861, 9860, 9869, 9867, 9866, 9869, 9865, 9864, 9861, 9855, 9854, 9867, 9859, 9858, 9867, 9864, 9857, 9855, 9858, 9856, 9868, 9866, 9861, 9866, 9858, 9855, 9842, 9843, 9813, 9817, 9847, 9846, 9816, 9846, 9842, 9852, 9853, 9833, 9856, 9857, 9837, 9865, 9841, 9840, 9853, 9856, 9836,
9864, 9840, 9837, 9805, 9820, 9818, 9813, 9832, 9830, 9811, 9830, 9820, 9845, 9854, 9852, 9862, 9860, 9849, 9860, 9854, 9845, 9843, 9852, 9832, 9875, 9877, 9876, 9879, 9881, 9880, 9884, 9885, 9883, 9888, 9889, 9887, 9882, 9883, 9871, 9886, 9887, 9873, 9871, 9875, 9874, 9873, 9879, 9878, 9878, 9880, 9877, 9885, 9888, 9886, 9872, 9878, 9875, 9883, 9886, 9872, 9895, 9897, 9896, 9899, 9901, 9900, 9904, 9905, 9903, 9908, 9909, 9907, 9913, 9911, 9910, 9902, 9903, 9891, 9906, 9907, 9893, 9911, 9901, 9899, 9891, 9895, 9894, 9893, 9899, 9898, 9912, 9910, 9907, 9898, 9900, 9897, 9905, 9908, 9906, 9892, 9898, 9895, 9903, 9906, 9892, 9910, 9899, 9893, 9916, 9917, 9915, 9920, 9921, 9919, 9929, 9927, 9926, 9933, 9931, 9930, 9922, 9923, 9917, 9924, 9925, 9921, 9927, 9923, 9922, 9931, 9925, 9924, 9917, 9920, 9918, 9932, 9930, 9927, 9930, 9924, 9923, 9923, 9924, 9920, 9936, 9937, 9935, 9940, 9941, 9939, 9945, 9943, 9942, 9953, 9951, 9950, 9957, 9955, 9954, 9946, 9947, 9937, 9948, 9949, 9941, 9957, 9945, 9944, 9951, 9947, 9946, 9944, 9942, 9939, 9955, 9949, 9948, 9937, 9940, 9938,
9956, 9954, 9951, 9954, 9948, 9947, 9947, 9948, 9940, 9955, 9944, 9941, 9914, 9915, 9885, 9918, 9919, 9889, 9915, 9918, 9888, 9934, 9935, 9905, 9938, 9939, 9909, 9943, 9913, 9912, 9935, 9938, 9908, 9942, 9912, 9909, 9894, 9896, 9881, 9889, 9904, 9902, 9890, 9894, 9879, 9887, 9902, 9890, 9921, 9936, 9934, 9952, 9950, 9931, 9950, 9946, 9925, 9925, 9946, 9936, 9919, 9934, 9904, 9577, 9578, 9365, 9581, 9582, 9369, 9373, 9586, 9585, 9578, 9581, 9368, 9372, 9585, 9577, 9613, 9614, 9407, 9620, 9413, 9412, 9614, 9617, 9409, 9619, 9412, 9409, 9582, 9613, 9406, 9802, 9803, 9548, 9552, 9807, 9806, 9551, 9806, 9802, 9818, 9819, 9568, 9822, 9823, 9572, 9827, 9576, 9575, 9819, 9822, 9571, 9826, 9575, 9572, 9803, 9818, 9567, 9874, 9876, 9753, 9761, 9884, 9882, 9870, 9874, 9751, 9759, 9882, 9870, 9789, 9916, 9914, 9928, 9926, 9799, 9926, 9922, 9793, 9793, 9922, 9916, 9787, 9914, 9884, 9238, 9239, 9058, 9242, 9243, 9062, 9066, 9247, 9246, 9239, 9242, 9061, 9065, 9246, 9238, 9286, 9287, 9118, 9293, 9127, 9126, 9287, 9290, 9121, 9292, 9126, 9121, 9243, 9286, 9117, 9206, 9415, 9416,
9213, 9420, 9419, 9414, 9415, 9206, 9211, 9419, 9414, 9433, 9434, 9229, 9437, 9438, 9233, 9442, 9237, 9236, 9434, 9437, 9232, 9441, 9236, 9233, 9416, 9433, 9228, 9963, 9965, 9964, 9967, 9969, 9968, 9959, 9963, 9962, 9961, 9967, 9966, 9966, 9968, 9965, 9960, 9966, 9963, 9975, 9977, 9976, 9979, 9981, 9980, 9971, 9975, 9974, 9973, 9979, 9978, 9978, 9980, 9977, 9972, 9978, 9975, 9984, 9985, 9983, 9988, 9989, 9987, 9992, 9993, 9991, 9996, 9997, 9995, 9990, 9991, 9985, 9994, 9995, 9989, 9985, 9988, 9986, 9993, 9996, 9994, 9991, 9994, 9988, 10000, 10001, 9999, 10004, 10005, 10003, 10008, 10009, 10007, 10012, 10013, 10011, 10006, 10007, 10001, 10010, 10011, 10005, 10001, 10004, 10002, 10009, 10012, 10010, 10007, 10010, 10004, 9982, 9983, 9959, 9986, 9987, 9961, 9983, 9986, 9960, 9998, 9999, 9971, 10002, 10003, 9973, 9999, 10002, 9972, 9974, 9976, 9969, 9970, 9974, 9967, 9989, 10000, 9998, 9997, 10008, 10006, 9995, 10006, 10000, 9987, 9998, 9970, 10019, 10021, 10020, 10023, 10025, 10024, 10015, 10019, 10018, 10017, 10023, 10022, 10022, 10024, 10021, 10016, 10022, 10019, 10031, 10033, 10032, 10035, 10037, 10036, 10027, 10031, 10030,
10029, 10035, 10034, 10034, 10036, 10033, 10028, 10034, 10031, 10040, 10041, 10039, 10044, 10045, 10043, 10048, 10049, 10047, 10052, 10053, 10051, 10046, 10047, 10041, 10050, 10051, 10045, 10041, 10044, 10042, 10049, 10052, 10050, 10047, 10050, 10044, 10056, 10057, 10055, 10060, 10061, 10059, 10064, 10065, 10063, 10068, 10069, 10067, 10062, 10063, 10057, 10066, 10067, 10061, 10057, 10060, 10058, 10065, 10068, 10066, 10063, 10066, 10060, 10038, 10039, 10015, 10042, 10043, 10017, 10039, 10042, 10016, 10054, 10055, 10027, 10058, 10059, 10029, 10055, 10058, 10028, 10030, 10032, 10025, 10026, 10030, 10023, 10045, 10056, 10054, 10053, 10064, 10062, 10051, 10062, 10056, 10043, 10054, 10026, 10018, 10020, 9981, 10014, 10018, 9979, 10005, 10040, 10038, 10013, 10048, 10046, 10011, 10046, 10040, 10003, 10038, 10014, 10075, 10077, 10076, 10079, 10081, 10080, 10071, 10075, 10074, 10073, 10079, 10078, 10078, 10080, 10077, 10072, 10078, 10075, 10087, 10089, 10088, 10091, 10093, 10092, 10083, 10087, 10086, 10085, 10091, 10090, 10090, 10092, 10089, 10084, 10090, 10087, 10096, 10097, 10095, 10100, 10101, 10099, 10104, 10105, 10103, 10108, 10109, 10107, 10102, 10103, 10097, 10106, 10107, 10101, 10097, 10100, 10098, 10105, 10108, 10106, 10103, 10106, 10100,
10112, 10113, 10111, 10116, 10117, 10115, 10121, 10119, 10118, 10124, 10125, 10123, 10128, 10129, 10127, 10133, 10131, 10130, 10122, 10123, 10113, 10126, 10127, 10117, 10131, 10121, 10120, 10120, 10118, 10115, 10132, 10130, 10127, 10113, 10116, 10114, 10125, 10128, 10126, 10123, 10126, 10116, 10130, 10120, 10117, 10094, 10095, 10071, 10098, 10099, 10073, 10095, 10098, 10072, 10110, 10111, 10083, 10114, 10115, 10085, 10119, 10093, 10091, 10111, 10114, 10084, 10118, 10091, 10085, 10086, 10088, 10081, 10082, 10086, 10079, 10101, 10112, 10110, 10109, 10124, 10122, 10107, 10122, 10112, 10099, 10110, 10082, 10074, 10076, 10037, 10070, 10074, 10035, 10061, 10096, 10094, 10069, 10104, 10102, 10067, 10102, 10096, 10059, 10094, 10070, 10141, 10139, 10138, 10145, 10143, 10142, 10139, 10135, 10134, 10143, 10137, 10136, 10144, 10142, 10139, 10142, 10136, 10135, 10153, 10151, 10150, 10157, 10155, 10154, 10151, 10147, 10146, 10155, 10149, 10148, 10156, 10154, 10151, 10154, 10148, 10147, 10152, 10150, 10143, 10150, 10146, 10137, 10165, 10163, 10162, 10169, 10167, 10166, 10163, 10159, 10158, 10167, 10161, 10160, 10168, 10166, 10163, 10166, 10160, 10159, 10177, 10175, 10174, 10181, 10179, 10178, 10175, 10171, 10170, 10179, 10173, 10172, 10180, 10178, 10175,
10178, 10172, 10171, 10176, 10174, 10167, 10174, 10170, 10161, 10184, 10185, 10183, 10188, 10189, 10187, 10191, 10193, 10192, 10195, 10197, 10196, 10200, 10201, 10199, 10204, 10205, 10203, 10198, 10199, 10185, 10202, 10203, 10189, 10183, 10191, 10190, 10187, 10195, 10194, 10185, 10188, 10186, 10194, 10196, 10193, 10201, 10204, 10202, 10186, 10194, 10191, 10199, 10202, 10188, 10208, 10209, 10207, 10212, 10213, 10211, 10215, 10217, 10216, 10219, 10221, 10220, 10224, 10225, 10223, 10228, 10229, 10227, 10222, 10223, 10209, 10226, 10227, 10213, 10207, 10215, 10214, 10211, 10219, 10218, 10209, 10212, 10210, 10218, 10220, 10217, 10225, 10228, 10226, 10210, 10218, 10215, 10223, 10226, 10212, 10230, 10232, 10233, 10236, 10237, 10235, 10242, 10243, 10241, 10246, 10247, 10245, 10249, 10251, 10250, 10240, 10241, 10233, 10244, 10245, 10237, 10239, 10249, 10248, 10248, 10250, 10242, 10233, 10236, 10234, 10243, 10246, 10244, 10241, 10244, 10236, 10238, 10248, 10240, 10254, 10255, 10253, 10258, 10259, 10257, 10265, 10263, 10262, 10268, 10269, 10267, 10273, 10271, 10270, 10254, 10266, 10267, 10271, 10261, 10260, 10272, 10270, 10267, 10263, 10259, 10258, 10255, 10258, 10256, 10264, 10262, 10260, 10262, 10258, 10255, 10270, 10260, 10255, 10230, 10231, 10201,
10234, 10235, 10205, 10231, 10234, 10204, 10252, 10253, 10225, 10256, 10257, 10229, 10253, 10256, 10228, 10189, 10208, 10206, 10214, 10216, 10197, 10205, 10224, 10222, 10206, 10214, 10195, 10203, 10222, 10208, 10237, 10254, 10252, 10247, 10268, 10266, 10245, 10266, 10254, 10235, 10252, 10224, 10276, 10277, 10275, 10280, 10281, 10279, 10283, 10285, 10284, 10287, 10289, 10288, 10292, 10293, 10291, 10296, 10297, 10295, 10290, 10291, 10277, 10294, 10295, 10281, 10275, 10283, 10282, 10279, 10287, 10286, 10277, 10280, 10278, 10286, 10288, 10285, 10293, 10296, 10294, 10278, 10286, 10283, 10291, 10294, 10280, 10300, 10301, 10299, 10304, 10305, 10303, 10307, 10309, 10308, 10311, 10313, 10312, 10316, 10317, 10315, 10320, 10321, 10319, 10314, 10315, 10301, 10318, 10319, 10305, 10299, 10307, 10306, 10303, 10311, 10310, 10301, 10304, 10302, 10310, 10312, 10309, 10317, 10320, 10318, 10302, 10310, 10307, 10315, 10318, 10304, 10324, 10325, 10323, 10328, 10329, 10327, 10333, 10331, 10330, 10337, 10335, 10334, 10331, 10325, 10324, 10335, 10329, 10328, 10325, 10328, 10326, 10336, 10334, 10331, 10334, 10328, 10325, 10340, 10341, 10339, 10344, 10345, 10343, 10349, 10347, 10346, 10353, 10351, 10350, 10347, 10341, 10340, 10351, 10345, 10344, 10341, 10344, 10342,
10352, 10350, 10347, 10350, 10344, 10341, 10322, 10323, 10293, 10326, 10327, 10297, 10323, 10326, 10296, 10338, 10339, 10317, 10342, 10343, 10321, 10339, 10342, 10320, 10281, 10300, 10298, 10306, 10308, 10289, 10297, 10316, 10314, 10298, 10306, 10287, 10295, 10314, 10300, 10329, 10340, 10338, 10348, 10346, 10335, 10346, 10340, 10329, 10327, 10338, 10316, 10164, 10162, 10155, 10162, 10158, 10149, 10213, 10276, 10274, 10282, 10284, 10221, 10229, 10292, 10290, 10274, 10282, 10219, 10227, 10290, 10276, 10259, 10324, 10322, 10332, 10330, 10263, 10330, 10324, 10259, 10257, 10322, 10292, 10361, 10359, 10358, 10365, 10363, 10362, 10359, 10355, 10354, 10363, 10357, 10356, 10364, 10362, 10359, 10362, 10356, 10355, 10373, 10371, 10370, 10377, 10375, 10374, 10371, 10367, 10366, 10375, 10369, 10368, 10376, 10374, 10371, 10374, 10368, 10367, 10372, 10370, 10363, 10370, 10366, 10357, 10380, 10381, 10379, 10384, 10385, 10383, 10387, 10389, 10388, 10391, 10393, 10392, 10396, 10397, 10395, 10400, 10401, 10399, 10394, 10395, 10381, 10398, 10399, 10385, 10379, 10387, 10386, 10383, 10391, 10390, 10381, 10384, 10382, 10390, 10392, 10389, 10397, 10400, 10398, 10382, 10390, 10387, 10395, 10398, 10384, 10404, 10405, 10403, 10408, 10409, 10407, 10411, 10413, 10412,
10415, 10417, 10416, 10419, 10417, 10415, 10422, 10423, 10421, 10426, 10427, 10425, 10431, 10429, 10428, 10420, 10421, 10405, 10424, 10425, 10409, 10429, 10419, 10418, 10403, 10411, 10410, 10407, 10415, 10414, 10418, 10415, 10407, 10430, 10428, 10425, 10405, 10408, 10406, 10414, 10416, 10413, 10423, 10426, 10424, 10406, 10414, 10411, 10421, 10424, 10408, 10428, 10418, 10409, 10434, 10435, 10433, 10436, 10438, 10439, 10443, 10441, 10440, 10448, 10449, 10447, 10451, 10453, 10452, 10446, 10447, 10439, 10445, 10451, 10450, 10441, 10435, 10434, 10450, 10452, 10448, 10433, 10435, 10438, 10445, 10444, 10441, 10444, 10438, 10435, 10444, 10450, 10446, 10456, 10457, 10455, 10460, 10461, 10459, 10465, 10463, 10462, 10468, 10469, 10467, 10472, 10473, 10471, 10477, 10475, 10474, 10466, 10467, 10457, 10470, 10471, 10461, 10475, 10465, 10464, 10464, 10462, 10459, 10476, 10474, 10471, 10457, 10460, 10458, 10469, 10472, 10470, 10467, 10470, 10460, 10474, 10464, 10461, 10432, 10433, 10397, 10436, 10437, 10401, 10433, 10436, 10400, 10454, 10455, 10423, 10458, 10459, 10427, 10463, 10431, 10430, 10455, 10458, 10426, 10462, 10430, 10427, 10385, 10404, 10402, 10410, 10412, 10393, 10401, 10422, 10420, 10402, 10410, 10391, 10399, 10420, 10404, 10439, 10456, 10454,
10449, 10468, 10466, 10447, 10466, 10456, 10437, 10454, 10422, 10480, 10481, 10479, 10484, 10485, 10483, 10487, 10489, 10488, 10492, 10493, 10491, 10496, 10497, 10495, 10499, 10501, 10500, 10490, 10491, 10481, 10494, 10495, 10485, 10489, 10499, 10498, 10486, 10488, 10480, 10498, 10500, 10492, 10481, 10484, 10482, 10493, 10496, 10494, 10491, 10494, 10484, 10488, 10498, 10490, 10504, 10505, 10503, 10509, 10507, 10506, 10512, 10513, 10511, 10517, 10515, 10514, 10510, 10511, 10505, 10515, 10509, 10508, 10508, 10506, 10503, 10516, 10514, 10511, 10514, 10508, 10505, 10520, 10521, 10519, 10524, 10525, 10523, 10527, 10529, 10528, 10532, 10533, 10531, 10536, 10537, 10535, 10539, 10541, 10540, 10530, 10531, 10521, 10534, 10535, 10525, 10529, 10539, 10538, 10526, 10528, 10520, 10538, 10540, 10532, 10521, 10524, 10522, 10533, 10536, 10534, 10531, 10534, 10524, 10528, 10538, 10530, 10544, 10545, 10543, 10549, 10547, 10546, 10552, 10553, 10551, 10557, 10555, 10554, 10550, 10551, 10545, 10555, 10549, 10548, 10548, 10546, 10543, 10556, 10554, 10551, 10554, 10548, 10545, 10518, 10519, 10493, 10522, 10523, 10497, 10501, 10527, 10526, 10519, 10522, 10496, 10500, 10526, 10518, 10542, 10543, 10513, 10547, 10517, 10516, 10546, 10516, 10513, 10485, 10504, 10502,
10497, 10512, 10510, 10495, 10510, 10504, 10525, 10544, 10542, 10537, 10552, 10550, 10535, 10550, 10544, 10523, 10542, 10512, 10560, 10561, 10559, 10564, 10565, 10563, 10567, 10569, 10568, 10572, 10573, 10571, 10576, 10577, 10575, 10579, 10581, 10580, 10570, 10571, 10561, 10574, 10575, 10565, 10569, 10579, 10578, 10566, 10568, 10560, 10578, 10580, 10572, 10561, 10564, 10562, 10573, 10576, 10574, 10571, 10574, 10564, 10568, 10578, 10570, 10584, 10585, 10583, 10589, 10587, 10586, 10592, 10593, 10591, 10597, 10595, 10594, 10590, 10591, 10585, 10595, 10589, 10588, 10588, 10586, 10583, 10596, 10594, 10591, 10594, 10588, 10585, 10600, 10601, 10599, 10604, 10605, 10603, 10607, 10609, 10608, 10610, 10612, 10613, 10616, 10617, 10615, 10619, 10621, 10620, 10610, 10611, 10601, 10614, 10615, 10605, 10609, 10619, 10618, 10606, 10608, 10600, 10610, 10618, 10620, 10601, 10604, 10602, 10613, 10616, 10614, 10611, 10614, 10604, 10608, 10618, 10610, 10622, 10624, 10625, 10629, 10627, 10626, 10630, 10632, 10633, 10637, 10635, 10634, 10630, 10631, 10625, 10635, 10629, 10628, 10625, 10628, 10626, 10633, 10636, 10634, 10631, 10634, 10628, 10598, 10599, 10573, 10602, 10603, 10577, 10581, 10607, 10606, 10599, 10602, 10576, 10580, 10606, 10598, 10622, 10623, 10593,
10627, 10597, 10596, 10626, 10596, 10593, 10565, 10584, 10582, 10577, 10592, 10590, 10575, 10590, 10584, 10605, 10624, 10622, 10617, 10632, 10630, 10615, 10630, 10624, 10603, 10622, 10592, 10558, 10559, 10533, 10562, 10563, 10537, 10541, 10567, 10566, 10559, 10562, 10536, 10540, 10566, 10558, 10582, 10583, 10553, 10587, 10557, 10556, 10586, 10556, 10553, 10563, 10582, 10552, 10640, 10641, 10639, 10643, 10645, 10644, 10648, 10649, 10647, 10651, 10653, 10652, 10646, 10647, 10641, 10645, 10651, 10650, 10642, 10644, 10640, 10650, 10652, 10648, 10644, 10650, 10646, 10656, 10657, 10655, 10660, 10661, 10659, 10665, 10663, 10662, 10668, 10669, 10667, 10672, 10673, 10671, 10677, 10675, 10674, 10666, 10667, 10657, 10670, 10671, 10661, 10675, 10665, 10664, 10664, 10662, 10659, 10676, 10674, 10671, 10657, 10660, 10658, 10669, 10672, 10670, 10667, 10670, 10660, 10674, 10664, 10661, 10680, 10681, 10679, 10683, 10685, 10684, 10688, 10689, 10687, 10691, 10693, 10692, 10686, 10687, 10681, 10685, 10691, 10690, 10682, 10684, 10680, 10690, 10692, 10688, 10684, 10690, 10686, 10696, 10697, 10695, 10700, 10701, 10699, 10705, 10703, 10702, 10708, 10709, 10707, 10712, 10713, 10711, 10717, 10715, 10714, 10706, 10707, 10697, 10710, 10711, 10701, 10715, 10705, 10704,
10704, 10702, 10699, 10716, 10714, 10711, 10697, 10700, 10698, 10709, 10712, 10710, 10707, 10710, 10700, 10714, 10704, 10701, 10678, 10679, 10649, 10653, 10683, 10682, 10652, 10682, 10678, 10694, 10695, 10669, 10698, 10699, 10673, 10703, 10677, 10676, 10695, 10698, 10672, 10702, 10676, 10673, 10641, 10656, 10654, 10649, 10668, 10666, 10647, 10666, 10656, 10681, 10696, 10694, 10689, 10708, 10706, 10687, 10706, 10696, 10679, 10694, 10668, 10720, 10721, 10719, 10723, 10725, 10724, 10728, 10729, 10727, 10731, 10733, 10732, 10726, 10727, 10721, 10725, 10731, 10730, 10722, 10724, 10720, 10730, 10732, 10728, 10724, 10730, 10726, 10736, 10737, 10735, 10740, 10741, 10739, 10745, 10743, 10742, 10748, 10749, 10747, 10752, 10753, 10751, 10757, 10755, 10754, 10746, 10747, 10737, 10750, 10751, 10741, 10755, 10745, 10744, 10744, 10742, 10739, 10756, 10754, 10751, 10737, 10740, 10738, 10749, 10752, 10750, 10747, 10750, 10740, 10754, 10744, 10741, 10760, 10761, 10759, 10763, 10765, 10764, 10768, 10769, 10767, 10771, 10773, 10772, 10766, 10767, 10761, 10765, 10771, 10770, 10762, 10764, 10760, 10770, 10772, 10768, 10764, 10770, 10766, 10776, 10777, 10775, 10780, 10781, 10779, 10785, 10783, 10782, 10788, 10789, 10787, 10790, 10792, 10793, 10797, 10795, 10794,
10786, 10787, 10777, 10790, 10791, 10781, 10795, 10785, 10784, 10784, 10782, 10779, 10793, 10796, 10794, 10777, 10780, 10778, 10789, 10792, 10790, 10787, 10790, 10780, 10794, 10784, 10781, 10758, 10759, 10729, 10733, 10763, 10762, 10732, 10762, 10758, 10774, 10775, 10749, 10778, 10779, 10753, 10783, 10757, 10756, 10775, 10778, 10752, 10782, 10756, 10753, 10721, 10736, 10734, 10729, 10748, 10746, 10727, 10746, 10736, 10761, 10776, 10774, 10769, 10788, 10786, 10767, 10786, 10776, 10759, 10774, 10748, 10718, 10719, 10689, 10693, 10723, 10722, 10692, 10722, 10718, 10734, 10735, 10709, 10738, 10739, 10713, 10743, 10717, 10716, 10735, 10738, 10712, 10742, 10716, 10713, 10719, 10734, 10708, 10478, 10479, 10243, 10482, 10483, 10247, 10251, 10487, 10486, 10479, 10482, 10246, 10250, 10486, 10478, 10502, 10503, 10269, 10507, 10273, 10272, 10506, 10272, 10269, 10483, 10502, 10268, 10638, 10639, 10449, 10453, 10643, 10642, 10452, 10642, 10638, 10654, 10655, 10469, 10658, 10659, 10473, 10663, 10477, 10476, 10655, 10658, 10472, 10662, 10476, 10473, 10639, 10654, 10468, 10360, 10358, 10179, 10358, 10354, 10173, 10305, 10380, 10378, 10386, 10388, 10313, 10321, 10396, 10394, 10378, 10386, 10311, 10319, 10394, 10380, 10345, 10434, 10432, 10442, 10440, 10351,
10440, 10434, 10345, 10343, 10432, 10396, 10134, 10135, 9993, 10136, 10137, 9997, 10135, 10136, 9996, 10146, 10147, 10009, 10148, 10149, 10013, 10147, 10148, 10012, 10137, 10146, 10008, 10158, 10159, 10049, 10160, 10161, 10053, 10159, 10160, 10052, 10170, 10171, 10065, 10172, 10173, 10069, 10171, 10172, 10068, 10161, 10170, 10064, 10149, 10158, 10048, 10354, 10355, 10105, 10356, 10357, 10109, 10355, 10356, 10108, 10366, 10367, 10125, 10368, 10369, 10129, 10377, 10133, 10132, 10367, 10368, 10128, 10375, 10132, 10129, 10357, 10366, 10124, 10173, 10354, 10104, 8769, 8770, 8396, 8399, 8773, 8774, 8404, 8778, 8777, 8782, 8408, 8407, 8770, 8773, 8399, 8403, 8777, 8769, 8781, 8407, 8400, 8803, 8804, 8421, 8423, 8808, 8807, 8417, 8801, 8803, 8417, 8422, 8807, 8984, 8615, 8614, 8983, 8614, 8609, 8421, 8804, 8981, 10804, 10806, 10805, 10802, 10806, 10804, 10806, 10809, 10808, 10802, 10807, 10809, 10811, 10813, 10812, 10815, 10817, 10816, 10816, 10819, 10818, 10825, 10824, 10822, 10833, 10831, 10830, 10833, 10829, 10828, 10835, 10837, 10836, 10841, 10840, 10838, 10812, 10827, 10826, 10835, 10834, 10820, 10825, 10839, 10838, 10812, 10813, 10821, 10817, 10823, 10822, 10816, 10822, 10824, 10840, 10841, 10833,
10841, 10839, 10829, 10818, 10819, 10813, 10838, 10840, 10837, 10813, 10819, 10824, 10838, 10835, 10821, 10812, 10820, 10834, 10799, 10817, 10815, 10800, 10823, 10817, 10826, 10827, 10802, 10836, 10809, 10807, 10827, 10834, 10807, 10845, 10847, 10846, 10843, 10847, 10845, 10851, 10849, 10848, 10846, 10847, 10854, 10843, 10852, 10854, 10851, 10856, 10855, 10848, 10849, 10843, 10856, 10854, 10852, 10849, 10855, 10852, 10860, 10862, 10861, 10861, 10865, 10863, 10871, 10870, 10868, 10861, 10862, 10869, 10861, 10868, 10870, 10863, 10865, 10858, 10865, 10870, 10867, 10877, 10875, 10874, 10877, 10873, 10872, 10885, 10884, 10882, 10893, 10895, 10894, 10894, 10897, 10896, 10903, 10902, 10900, 10889, 10876, 10874, 10881, 10895, 10893, 10884, 10885, 10877, 10885, 10883, 10873, 10894, 10895, 10901, 10900, 10902, 10897, 10879, 10878, 10872, 10896, 10897, 10891, 10883, 10886, 10879, 10891, 10897, 10902, 10898, 10884, 10876, 10887, 10901, 10895, 10905, 10907, 10906, 10906, 10909, 10908, 10913, 10911, 10910, 10915, 10912, 10910, 10919, 10918, 10916, 10926, 10924, 10923, 10932, 10931, 10929, 10928, 10915, 10914, 10919, 10930, 10929, 10907, 10917, 10916, 10906, 10916, 10918, 10926, 10932, 10930, 10920, 10921, 10913, 10915, 10922, 10920, 10920, 10922, 10919,
10928, 10927, 10923, 10933, 10928, 10924, 10933, 10930, 10919, 10928, 10933, 10922, 10879, 10851, 10850, 10883, 10882, 10853, 10886, 10883, 10854, 10879, 10886, 10856, 10858, 10907, 10905, 10913, 10866, 10864, 10871, 10921, 10920, 10870, 10920, 10917, 10858, 10867, 10917, 10921, 10871, 10866, 10908, 10909, 10881, 10929, 10931, 10903, 10909, 10918, 10887, 10929, 10901, 10887, 10935, 10937, 10936, 10943, 10941, 10940, 10945, 10942, 10940, 10936, 10942, 10940, 10943, 10937, 10935, 10939, 10947, 10946, 10937, 10939, 10938, 10947, 10943, 10942, 10948, 10946, 10942, 10938, 10946, 10942, 10947, 10939, 10937, 10940, 10942, 10862, 10945, 10944, 10864, 10948, 10871, 10869, 10862, 10942, 10946, 10871, 10948, 10945, 10844, 10846, 10799, 10799, 10846, 10853, 10882, 10884, 10825, 10889, 10888, 10828, 10839, 10898, 10889, 10898, 10839, 10825, 10882, 10823, 10800, 10952, 10954, 10953, 10950, 10954, 10952, 10961, 10963, 10962, 10959, 10963, 10961, 10970, 10972, 10971, 10953, 10965, 10964, 10970, 10969, 10956, 10963, 10968, 10967, 10959, 10966, 10968, 10954, 10957, 10956, 10950, 10955, 10957, 10964, 10965, 10959, 10971, 10968, 10966, 10965, 10969, 10966, 10953, 10956, 10969, 10977, 10979, 10978, 10978, 10981, 10980, 10985, 10984, 10982, 10974, 10979, 10977,
10978, 10979, 10983, 10978, 10982, 10984, 10975, 10983, 10979, 10989, 10991, 10990, 10987, 10991, 10989, 10998, 11000, 10999, 10996, 11000, 10998, 11007, 11009, 11008, 10990, 11002, 11001, 11007, 11006, 10993, 11000, 11005, 11004, 10996, 11003, 11005, 10990, 10991, 10994, 10992, 10994, 10991, 11001, 11002, 10996, 11008, 11005, 11003, 10996, 11002, 11006, 10993, 11006, 11002, 11013, 11015, 11014, 11019, 11018, 11016, 11023, 11022, 11020, 11027, 11026, 11024, 11021, 11020, 11014, 11025, 11024, 11018, 11018, 11015, 11013, 11026, 11023, 11021, 11024, 11021, 11015, 11029, 11031, 11030, 11035, 11034, 11032, 11039, 11038, 11036, 11043, 11042, 11040, 11047, 11046, 11044, 11035, 11041, 11040, 11045, 11044, 11038, 11029, 11032, 11034, 11038, 11035, 11033, 11046, 11043, 11041, 11044, 11041, 11035, 11051, 11050, 11048, 11055, 11054, 11052, 11059, 11058, 11056, 11063, 11062, 11060, 11057, 11056, 11050, 11061, 11060, 11054, 11054, 11051, 11049, 11062, 11059, 11057, 11060, 11057, 11051, 10999, 11029, 11028, 11033, 11032, 11004, 11037, 11036, 11008, 11008, 11036, 11033, 10999, 11004, 11032, 11049, 11048, 11022, 11053, 11052, 11026, 11052, 11049, 11023, 11010, 11011, 10987, 11014, 10994, 10992, 11022, 11009, 11007, 10987, 11011, 11012, 11014, 11020, 11007,
11050, 11039, 11037, 11058, 11047, 11045, 11056, 11045, 11039, 11048, 11037, 11009, 11070, 11072, 11071, 11068, 11072, 11070, 11079, 11081, 11080, 11083, 11085, 11084, 11089, 11088, 11086, 11093, 11092, 11090, 11071, 11083, 11082, 11087, 11086, 11076, 11091, 11090, 11080, 11072, 11077, 11076, 11075, 11077, 11072, 11083, 11086, 11088, 11073, 11074, 11068, 11080, 11077, 11075, 11092, 11089, 11087, 11068, 11074, 11078, 11080, 11090, 11087, 11071, 11076, 11086, 11095, 11097, 11096, 11101, 11103, 11102, 11102, 11105, 11104, 11109, 11108, 11106, 11097, 11103, 11101, 11095, 11098, 11099, 11102, 11103, 11107, 11106, 11108, 11105, 11099, 11107, 11103, 11113, 11112, 11110, 11117, 11116, 11114, 11121, 11120, 11118, 11125, 11124, 11122, 11119, 11118, 11112, 11123, 11122, 11116, 11116, 11113, 11111, 11124, 11121, 11119, 11122, 11119, 11113, 11065, 11095, 11094, 11065, 11066, 11098, 11089, 11111, 11110, 11115, 11114, 11092, 11114, 11111, 11089, 11084, 11085, 11065, 11085, 11088, 11066, 11112, 11099, 11098, 11118, 11120, 11109, 11118, 11107, 11099, 11110, 11098, 11066, 11129, 11128, 11126, 11133, 11132, 11130, 11137, 11136, 11134, 11141, 11140, 11138, 11135, 11134, 11128, 11139, 11138, 11132, 11132, 11129, 11127, 11140, 11137, 11135, 11138, 11135, 11129,
11147, 11145, 11144, 11147, 11143, 11142, 11151, 11150, 11148, 11155, 11154, 11152, 11161, 11159, 11158, 11161, 11157, 11156, 11167, 11166, 11164, 11163, 11146, 11144, 11151, 11165, 11164, 11166, 11167, 11161, 11167, 11165, 11157, 11155, 11147, 11146, 11155, 11153, 11143, 11152, 11154, 11151, 11163, 11162, 11156, 11165, 11168, 11163, 11168, 11165, 11151, 11163, 11168, 11154, 11174, 11173, 11171, 11176, 11178, 11177, 11182, 11180, 11179, 11184, 11181, 11179, 11188, 11187, 11185, 11182, 11170, 11169, 11186, 11185, 11173, 11190, 11189, 11177, 11189, 11190, 11182, 11184, 11191, 11189, 11177, 11174, 11172, 11191, 11188, 11186, 11177, 11189, 11186, 11190, 11178, 11170, 11195, 11193, 11192, 11195, 11197, 11196, 11172, 11171, 11136, 11176, 11175, 11140, 11175, 11172, 11137, 11193, 11160, 11158, 11193, 11196, 11166, 11150, 11133, 11131, 11164, 11166, 11141, 11164, 11139, 11133, 11195, 11194, 11169, 11197, 11178, 11176, 11197, 11195, 11170, 11196, 11176, 11141, 10981, 11074, 11073, 11079, 11078, 10984, 10981, 10984, 11078, 11127, 11126, 11042, 11131, 11130, 11046, 11130, 11127, 11043, 11149, 11148, 11058, 11063, 11153, 11152, 11152, 11149, 11059, 11148, 11131, 11047, 11030, 11031, 10974, 11040, 11042, 10985, 10974, 11031, 11034, 11040, 10983, 10975,
11128, 11081, 11079, 11136, 11093, 11091, 11134, 11091, 11081, 11173, 11117, 11115, 11187, 11125, 11123, 11185, 11123, 11117, 11171, 11115, 11093, 11126, 11079, 10985, 10962, 11011, 11010, 11013, 11012, 10967, 11017, 11016, 10971, 10971, 11016, 11013, 10962, 10967, 11012, 11202, 11204, 11203, 11203, 11206, 11205, 11210, 11209, 11207, 11199, 11204, 11202, 11204, 11208, 11207, 11203, 11207, 11209, 11199, 11200, 11208, 11212, 11214, 11213, 11212, 11215, 11216, 11220, 11222, 11221, 11218, 11222, 11220, 11229, 11231, 11230, 11235, 11237, 11236, 11241, 11240, 11238, 11221, 11233, 11232, 11235, 11234, 11226, 11239, 11238, 11230, 11221, 11222, 11227, 11225, 11227, 11222, 11223, 11224, 11218, 11230, 11227, 11225, 11240, 11237, 11235, 11218, 11224, 11228, 11230, 11238, 11235, 11226, 11234, 11233, 11206, 11224, 11223, 11229, 11228, 11209, 11206, 11209, 11228, 11232, 11233, 11212, 11236, 11216, 11215, 11233, 11234, 11215, 11243, 11245, 11244, 11244, 11247, 11246, 11253, 11252, 11250, 11255, 11257, 11256, 11261, 11260, 11258, 11265, 11264, 11262, 11253, 11259, 11258, 11263, 11262, 11256, 11244, 11245, 11251, 11250, 11252, 11247, 11248, 11249, 11245, 11256, 11253, 11251, 11264, 11261, 11259, 11245, 11249, 11254, 11262, 11259, 11253, 11271, 11269, 11268,
11273, 11275, 11274, 11281, 11279, 11278, 11283, 11280, 11278, 11285, 11287, 11286, 11281, 11267, 11266, 11289, 11288, 11274, 11288, 11289, 11281, 11283, 11290, 11288, 11271, 11277, 11276, 11285, 11291, 11292, 11271, 11270, 11266, 11277, 11275, 11273, 11275, 11277, 11271, 11289, 11275, 11267, 11296, 11294, 11293, 11300, 11299, 11297, 11304, 11303, 11301, 11310, 11308, 11307, 11310, 11306, 11305, 11314, 11313, 11311, 11318, 11317, 11315, 11306, 11296, 11295, 11312, 11311, 11299, 11316, 11315, 11303, 11296, 11304, 11302, 11317, 11318, 11310, 11310, 11318, 11316, 11303, 11300, 11298, 11315, 11317, 11314, 11315, 11312, 11300, 11306, 11316, 11304, 11322, 11324, 11323, 11323, 11326, 11325, 11331, 11330, 11328, 11335, 11337, 11336, 11342, 11341, 11339, 11326, 11335, 11334, 11340, 11339, 11330, 11323, 11324, 11329, 11328, 11330, 11326, 11335, 11339, 11341, 11336, 11337, 11333, 11333, 11337, 11341, 11326, 11330, 11339, 11298, 11297, 11260, 11265, 11302, 11301, 11301, 11298, 11261, 11320, 11283, 11282, 11287, 11324, 11322, 11320, 11327, 11290, 11287, 11292, 11329, 11274, 11257, 11255, 11290, 11265, 11263, 11274, 11288, 11263, 11320, 11319, 11293, 11302, 11327, 11320, 11327, 11302, 11265, 11344, 11346, 11345, 11345, 11348, 11347, 11352, 11351, 11349,
11356, 11358, 11357, 11354, 11358, 11356, 11365, 11367, 11366, 11348, 11360, 11359, 11365, 11364, 11351, 11358, 11363, 11362, 11361, 11363, 11358, 11345, 11346, 11350, 11345, 11349, 11351, 11359, 11360, 11354, 11366, 11363, 11361, 11354, 11360, 11364, 11348, 11351, 11364, 11371, 11370, 11368, 11375, 11374, 11372, 11379, 11378, 11376, 11383, 11382, 11380, 11377, 11376, 11370, 11381, 11380, 11374, 11374, 11371, 11369, 11382, 11379, 11377, 11380, 11377, 11371, 11387, 11389, 11388, 11393, 11392, 11390, 11397, 11396, 11394, 11401, 11400, 11398, 11395, 11394, 11388, 11399, 11398, 11392, 11392, 11389, 11387, 11400, 11397, 11395, 11398, 11395, 11389, 11405, 11404, 11402, 11409, 11408, 11406, 11413, 11412, 11410, 11417, 11416, 11414, 11411, 11410, 11404, 11415, 11414, 11408, 11408, 11405, 11403, 11416, 11413, 11411, 11414, 11411, 11405, 11357, 11385, 11384, 11387, 11386, 11362, 11391, 11390, 11366, 11366, 11390, 11387, 11362, 11386, 11385, 11403, 11402, 11378, 11407, 11406, 11382, 11406, 11403, 11379, 11368, 11370, 11352, 11378, 11367, 11365, 11376, 11365, 11352, 11404, 11393, 11391, 11412, 11401, 11399, 11410, 11399, 11393, 11402, 11391, 11367, 11421, 11419, 11418, 11425, 11424, 11422, 11431, 11429, 11428, 11437, 11436, 11434, 11433, 11421, 11420,
11425, 11435, 11434, 11437, 11435, 11429, 11421, 11427, 11426, 11427, 11425, 11423, 11433, 11432, 11428, 11435, 11438, 11433, 11438, 11435, 11425, 11438, 11427, 11421, 11440, 11442, 11441, 11448, 11447, 11445, 11450, 11452, 11451, 11458, 11457, 11455, 11442, 11450, 11449, 11456, 11455, 11447, 11440, 11443, 11444, 11450, 11453, 11454, 11447, 11444, 11443, 11457, 11454, 11453, 11455, 11453, 11444, 11442, 11444, 11453, 11462, 11460, 11459, 11466, 11465, 11463, 11470, 11468, 11467, 11472, 11474, 11473, 11468, 11462, 11461, 11472, 11471, 11465, 11462, 11466, 11464, 11470, 11474, 11472, 11468, 11472, 11466, 11476, 11478, 11477, 11484, 11483, 11481, 11486, 11488, 11487, 11494, 11493, 11491, 11478, 11486, 11485, 11492, 11491, 11483, 11476, 11479, 11480, 11489, 11490, 11488, 11481, 11483, 11480, 11493, 11490, 11489, 11491, 11489, 11480, 11478, 11480, 11489, 11460, 11431, 11430, 11464, 11463, 11436, 11460, 11464, 11437, 11452, 11476, 11475, 11482, 11481, 11457, 11481, 11479, 11454, 11452, 11454, 11479, 11214, 11346, 11344, 11214, 11216, 11350, 11369, 11368, 11236, 11373, 11372, 11240, 11372, 11369, 11237, 11368, 11350, 11216, 11419, 11309, 11307, 11423, 11422, 11313, 11426, 11423, 11314, 11419, 11426, 11317, 11333, 11440, 11439, 11342, 11446, 11445,
11445, 11443, 11338, 11338, 11443, 11440, 11246, 11247, 11199, 11258, 11260, 11210, 11199, 11247, 11252, 11258, 11208, 11200, 11299, 11231, 11229, 11313, 11241, 11239, 11311, 11239, 11231, 11297, 11229, 11210, 11424, 11375, 11373, 11436, 11383, 11381, 11434, 11381, 11375, 11465, 11409, 11407, 11473, 11417, 11415, 11471, 11415, 11409, 11463, 11407, 11383, 11422, 11373, 11241, 11496, 11498, 11497, 11497, 11500, 11499, 11504, 11503, 11501, 11508, 11507, 11505, 11514, 11516, 11515, 11518, 11520, 11519, 11500, 11510, 11509, 11514, 11513, 11503, 11518, 11517, 11507, 11497, 11498, 11502, 11497, 11501, 11503, 11505, 11507, 11504, 11519, 11516, 11514, 11517, 11514, 11504, 11500, 11503, 11513, 11524, 11522, 11521, 11526, 11523, 11521, 11533, 11531, 11530, 11531, 11526, 11525, 11527, 11528, 11524, 11526, 11529, 11527, 11533, 11535, 11534, 11531, 11534, 11529, 11539, 11537, 11536, 11543, 11542, 11540, 11547, 11546, 11544, 11551, 11549, 11548, 11555, 11554, 11552, 11557, 11559, 11558, 11549, 11539, 11538, 11553, 11552, 11542, 11557, 11556, 11546, 11539, 11547, 11545, 11551, 11559, 11557, 11546, 11543, 11541, 11558, 11555, 11553, 11556, 11553, 11543, 11549, 11557, 11547, 11537, 11512, 11511, 11541, 11540, 11515, 11545, 11544, 11519, 11544, 11541, 11516,
11545, 11520, 11512, 11529, 11508, 11506, 11533, 11532, 11511, 11535, 11520, 11518, 11520, 11535, 11533, 11534, 11518, 11508, 11563, 11565, 11564, 11564, 11567, 11566, 11572, 11571, 11569, 11574, 11576, 11575, 11582, 11581, 11579, 11561, 11574, 11573, 11572, 11580, 11579, 11574, 11577, 11578, 11565, 11570, 11569, 11569, 11571, 11567, 11566, 11567, 11561, 11581, 11578, 11577, 11561, 11567, 11571, 11579, 11577, 11568, 11568, 11577, 11574, 11586, 11588, 11587, 11587, 11584, 11583, 11590, 11592, 11591, 11601, 11600, 11598, 11605, 11604, 11602, 11595, 11599, 11598, 11597, 11603, 11602, 11587, 11588, 11595, 11594, 11593, 11584, 11592, 11597, 11596, 11589, 11591, 11588, 11604, 11601, 11599, 11591, 11596, 11595, 11596, 11602, 11599, 11609, 11611, 11610, 11607, 11611, 11609, 11618, 11617, 11615, 11620, 11622, 11621, 11626, 11625, 11623, 11630, 11629, 11627, 11610, 11620, 11619, 11624, 11623, 11613, 11628, 11627, 11617, 11610, 11611, 11614, 11612, 11614, 11611, 11620, 11623, 11625, 11617, 11614, 11612, 11629, 11626, 11624, 11617, 11627, 11624, 11613, 11623, 11620, 11634, 11633, 11631, 11638, 11637, 11635, 11642, 11641, 11639, 11646, 11645, 11643, 11640, 11639, 11633, 11644, 11643, 11637, 11637, 11634, 11632, 11645, 11642, 11640, 11643, 11640, 11634,
11576, 11607, 11606, 11616, 11615, 11581, 11615, 11612, 11578, 11576, 11578, 11612, 11632, 11631, 11600, 11636, 11635, 11604, 11635, 11632, 11601, 11583, 11584, 11565, 11593, 11594, 11572, 11600, 11582, 11580, 11565, 11584, 11593, 11598, 11580, 11572, 11633, 11618, 11616, 11641, 11630, 11628, 11639, 11628, 11618, 11631, 11616, 11582, 11650, 11648, 11647, 11654, 11653, 11651, 11656, 11658, 11657, 11662, 11660, 11659, 11666, 11665, 11663, 11668, 11670, 11669, 11660, 11650, 11649, 11664, 11663, 11653, 11658, 11668, 11667, 11658, 11656, 11648, 11670, 11668, 11660, 11657, 11654, 11652, 11669, 11666, 11664, 11667, 11664, 11654, 11668, 11658, 11650, 11676, 11675, 11673, 11680, 11679, 11677, 11684, 11683, 11681, 11688, 11687, 11685, 11682, 11681, 11675, 11686, 11685, 11679, 11679, 11676, 11674, 11687, 11684, 11682, 11685, 11682, 11676, 11692, 11690, 11689, 11698, 11696, 11695, 11696, 11692, 11691, 11692, 11694, 11693, 11700, 11699, 11696, 11699, 11694, 11692, 11672, 11662, 11661, 11674, 11673, 11665, 11670, 11678, 11677, 11677, 11674, 11666, 11672, 11678, 11670, 11689, 11690, 11672, 11693, 11694, 11680, 11699, 11700, 11688, 11690, 11693, 11678, 11694, 11699, 11686, 11704, 11702, 11701, 11708, 11706, 11705, 11710, 11704, 11703, 11712, 11708, 11707,
11707, 11705, 11702, 11711, 11707, 11704, 11718, 11716, 11715, 11714, 11717, 11715, 11722, 11720, 11719, 11724, 11726, 11725, 11728, 11730, 11729, 11725, 11726, 11718, 11723, 11725, 11717, 11730, 11722, 11721, 11721, 11719, 11716, 11727, 11729, 11726, 11729, 11721, 11718, 11714, 11713, 11706, 11712, 11723, 11714, 11648, 11551, 11550, 11652, 11651, 11554, 11656, 11655, 11558, 11655, 11652, 11555, 11656, 11559, 11551, 11622, 11703, 11701, 11626, 11710, 11709, 11630, 11712, 11711, 11629, 11711, 11710, 11622, 11625, 11709, 11724, 11723, 11641, 11728, 11727, 11645, 11727, 11724, 11642, 11641, 11723, 11712, 11734, 11733, 11731, 11738, 11737, 11735, 11742, 11741, 11739, 11746, 11745, 11743, 11740, 11739, 11733, 11744, 11743, 11737, 11737, 11734, 11732, 11745, 11742, 11740, 11743, 11740, 11734, 11750, 11749, 11747, 11754, 11753, 11751, 11758, 11757, 11755, 11762, 11761, 11759, 11756, 11755, 11749, 11760, 11759, 11753, 11753, 11750, 11748, 11761, 11758, 11756, 11759, 11756, 11750, 11766, 11765, 11763, 11770, 11769, 11767, 11772, 11774, 11773, 11778, 11777, 11775, 11772, 11771, 11765, 11776, 11775, 11769, 11769, 11766, 11764, 11777, 11774, 11772, 11775, 11772, 11766, 11782, 11781, 11779, 11786, 11785, 11783, 11790, 11789, 11787, 11794, 11793, 11791,
11788, 11787, 11781, 11792, 11791, 11785, 11785, 11782, 11780, 11793, 11790, 11788, 11791, 11788, 11782, 11764, 11763, 11741, 11768, 11767, 11745, 11767, 11764, 11742, 11780, 11779, 11757, 11784, 11783, 11761, 11783, 11780, 11758, 11749, 11738, 11736, 11757, 11746, 11744, 11755, 11744, 11738, 11781, 11770, 11768, 11789, 11778, 11776, 11787, 11776, 11770, 11779, 11768, 11746, 11798, 11796, 11795, 11800, 11802, 11801, 11806, 11804, 11803, 11810, 11812, 11811, 11804, 11798, 11797, 11802, 11810, 11809, 11802, 11800, 11796, 11806, 11812, 11810, 11807, 11808, 11806, 11808, 11813, 11812, 11810, 11802, 11798, 11815, 11817, 11816, 11823, 11822, 11820, 11825, 11827, 11826, 11833, 11832, 11830, 11817, 11825, 11824, 11831, 11830, 11822, 11818, 11819, 11817, 11828, 11829, 11827, 11820, 11822, 11819, 11830, 11832, 11829, 11822, 11830, 11828, 11819, 11828, 11825, 11837, 11835, 11834, 11841, 11840, 11838, 11847, 11845, 11844, 11851, 11850, 11848, 11845, 11837, 11836, 11849, 11848, 11840, 11843, 11842, 11835, 11853, 11852, 11845, 11842, 11843, 11841, 11852, 11853, 11851, 11843, 11852, 11849, 11852, 11843, 11837, 11855, 11857, 11856, 11865, 11864, 11862, 11867, 11869, 11868, 11871, 11873, 11872, 11859, 11867, 11866, 11865, 11871, 11870, 11855, 11860, 11861,
11870, 11872, 11869, 11859, 11858, 11856, 11862, 11864, 11861, 11864, 11859, 11857, 11859, 11864, 11870, 11835, 11808, 11807, 11839, 11838, 11811, 11813, 11842, 11839, 11835, 11842, 11813, 11827, 11855, 11854, 11863, 11862, 11832, 11832, 11862, 11860, 11829, 11860, 11855, 11877, 11875, 11874, 11877, 11879, 11878, 11886, 11885, 11883, 11888, 11890, 11889, 11896, 11895, 11893, 11879, 11888, 11887, 11894, 11893, 11885, 11881, 11877, 11876, 11877, 11881, 11882, 11891, 11892, 11890, 11883, 11885, 11882, 11893, 11895, 11892, 11885, 11893, 11891, 11882, 11891, 11888, 11900, 11899, 11897, 11904, 11903, 11901, 11908, 11907, 11905, 11912, 11911, 11909, 11906, 11905, 11899, 11910, 11909, 11903, 11903, 11900, 11898, 11911, 11908, 11906, 11909, 11906, 11900, 11916, 11918, 11917, 11921, 11923, 11922, 11929, 11927, 11926, 11925, 11928, 11926, 11931, 11933, 11932, 11918, 11925, 11924, 11923, 11931, 11930, 11916, 11920, 11922, 11933, 11929, 11928, 11930, 11932, 11928, 11916, 11915, 11913, 11920, 11916, 11914, 11922, 11930, 11925, 11937, 11936, 11934, 11941, 11940, 11938, 11945, 11944, 11942, 11949, 11948, 11946, 11943, 11942, 11936, 11947, 11946, 11940, 11940, 11937, 11935, 11948, 11945, 11943, 11946, 11943, 11937, 11890, 11914, 11913, 11921, 11920, 11895,
11895, 11920, 11919, 11890, 11892, 11919, 11935, 11934, 11907, 11939, 11938, 11911, 11938, 11935, 11908, 11899, 11886, 11884, 11907, 11896, 11894, 11905, 11894, 11886, 11936, 11923, 11921, 11944, 11933, 11931, 11942, 11931, 11923, 11934, 11921, 11896, 11951, 11953, 11952, 11955, 11953, 11951, 11959, 11958, 11956, 11961, 11963, 11962, 11969, 11968, 11966, 11973, 11972, 11970, 11965, 11955, 11954, 11967, 11966, 11958, 11963, 11971, 11970, 11953, 11961, 11960, 11955, 11963, 11961, 11962, 11959, 11957, 11972, 11969, 11967, 11970, 11967, 11959, 11971, 11963, 11955, 11979, 11977, 11976, 11975, 11978, 11976, 11981, 11983, 11982, 11987, 11985, 11984, 11989, 11991, 11990, 11979, 11989, 11988, 11993, 11992, 11985, 11983, 11979, 11978, 11980, 11982, 11978, 11994, 11995, 11991, 11983, 11994, 11989, 11999, 11998, 11996, 12003, 12002, 12000, 12007, 12006, 12004, 12011, 12010, 12008, 12005, 12004, 11998, 12009, 12008, 12002, 12002, 11999, 11997, 12010, 12007, 12005, 12008, 12005, 11999, 12015, 12017, 12016, 12016, 12013, 12012, 12024, 12022, 12021, 12028, 12030, 12029, 12022, 12017, 12015, 12020, 12028, 12027, 12017, 12020, 12019, 12016, 12019, 12018, 12030, 12028, 12022, 12025, 12026, 12024, 12026, 12031, 12030, 12028, 12020, 12017, 11997, 11996, 11968,
12001, 12000, 11972, 12000, 11997, 11969, 12013, 11987, 11986, 12018, 11993, 11987, 11984, 11985, 11965, 11992, 11993, 11973, 11985, 11992, 11971, 12018, 12019, 12003, 12029, 12011, 12009, 12027, 12009, 12003, 11993, 12018, 12001, 11774, 11881, 11880, 11884, 11883, 11777, 11883, 11881, 11774, 11898, 11897, 11789, 11902, 11901, 11793, 11901, 11898, 11790, 11897, 11884, 11778, 11952, 11847, 11846, 11957, 11956, 11850, 11853, 11960, 11957, 11952, 11960, 11853, 11869, 11975, 11974, 11873, 11981, 11980, 11872, 11980, 11975, 11801, 11754, 11752, 11811, 11762, 11760, 11809, 11760, 11754, 11840, 11786, 11784, 11850, 11794, 11792, 11848, 11792, 11786, 11838, 11784, 11762, 11958, 11904, 11902, 11968, 11912, 11910, 11966, 11910, 11904, 11998, 11941, 11939, 12006, 11949, 11947, 12004, 11947, 11941, 11996, 11939, 11912, 11956, 11902, 11794, 12035, 12034, 12032, 12039, 12038, 12036, 12043, 12042, 12040, 12047, 12046, 12044, 12041, 12040, 12034, 12045, 12044, 12038, 12038, 12035, 12033, 12046, 12043, 12041, 12044, 12041, 12035, 12049, 12051, 12050, 12053, 12051, 12049, 12055, 12057, 12056, 12061, 12059, 12058, 12065, 12067, 12066, 12059, 12053, 12052, 12057, 12065, 12064, 12051, 12055, 12054, 12053, 12057, 12055, 12061, 12067, 12065, 12062, 12063, 12061,
12063, 12068, 12067, 12065, 12057, 12053, 12072, 12071, 12069, 12076, 12075, 12073, 12080, 12079, 12077, 12084, 12083, 12081, 12078, 12077, 12071, 12082, 12081, 12075, 12075, 12072, 12070, 12083, 12080, 12078, 12081, 12078, 12072, 12088, 12090, 12089, 12089, 12086, 12085, 12094, 12093, 12091, 12101, 12099, 12098, 12105, 12104, 12102, 12107, 12109, 12108, 12099, 12090, 12088, 12103, 12102, 12093, 12097, 12107, 12106, 12089, 12090, 12097, 12096, 12095, 12086, 12109, 12107, 12099, 12095, 12096, 12094, 12108, 12105, 12103, 12106, 12103, 12094, 12107, 12097, 12090, 12070, 12069, 12042, 12074, 12073, 12046, 12073, 12070, 12043, 12086, 12063, 12062, 12092, 12091, 12066, 12068, 12095, 12092, 12095, 12068, 12063, 12056, 12039, 12037, 12066, 12047, 12045, 12064, 12045, 12039, 12093, 12076, 12074, 12104, 12084, 12082, 12102, 12082, 12076, 12091, 12074, 12047, 12113, 12112, 12110, 12117, 12116, 12114, 12123, 12122, 12120, 12127, 12126, 12124, 12121, 12120, 12112, 12125, 12124, 12116, 12116, 12113, 12111, 12126, 12123, 12121, 12124, 12121, 12113, 12133, 12132, 12130, 12137, 12136, 12134, 12141, 12140, 12138, 12145, 12144, 12142, 12139, 12138, 12132, 12143, 12142, 12136, 12136, 12133, 12131, 12144, 12141, 12139, 12142, 12139, 12133, 12151, 12149, 12148,
12147, 12150, 12148, 12153, 12155, 12154, 12159, 12158, 12156, 12161, 12163, 12162, 12171, 12170, 12168, 12151, 12161, 12160, 12169, 12168, 12158, 12155, 12151, 12150, 12152, 12154, 12150, 12166, 12167, 12163, 12158, 12155, 12153, 12165, 12164, 12162, 12168, 12170, 12167, 12170, 12165, 12163, 12158, 12168, 12166, 12155, 12166, 12161, 12175, 12174, 12172, 12179, 12178, 12176, 12183, 12182, 12180, 12187, 12186, 12184, 12181, 12180, 12174, 12185, 12184, 12178, 12178, 12175, 12173, 12186, 12183, 12181, 12184, 12181, 12175, 12119, 12147, 12146, 12123, 12153, 12152, 12157, 12156, 12126, 12156, 12153, 12123, 12122, 12152, 12147, 12173, 12172, 12140, 12177, 12176, 12144, 12176, 12173, 12141, 12132, 12117, 12115, 12140, 12127, 12125, 12138, 12125, 12117, 12174, 12159, 12157, 12182, 12171, 12169, 12180, 12169, 12159, 12172, 12157, 12127, 12191, 12189, 12188, 12195, 12197, 12196, 12199, 12197, 12195, 12201, 12203, 12202, 12196, 12191, 12190, 12193, 12192, 12189, 12197, 12201, 12200, 12203, 12201, 12197, 12200, 12193, 12191, 12205, 12207, 12206, 12205, 12208, 12209, 12215, 12213, 12212, 12219, 12218, 12216, 12223, 12225, 12224, 12227, 12225, 12223, 12231, 12230, 12228, 12233, 12235, 12234, 12224, 12215, 12214, 12229, 12228, 12218, 12221, 12220, 12213,
12225, 12233, 12232, 12235, 12233, 12225, 12212, 12213, 12211, 12220, 12221, 12219, 12234, 12231, 12229, 12213, 12220, 12217, 12221, 12232, 12229, 12232, 12221, 12215, 12237, 12239, 12238, 12243, 12245, 12244, 12239, 12243, 12242, 12237, 12240, 12241, 12246, 12247, 12245, 12241, 12246, 12243, 12211, 12199, 12198, 12203, 12217, 12216, 12217, 12203, 12199, 12207, 12237, 12236, 12207, 12209, 12240, 12111, 12110, 12079, 12115, 12114, 12083, 12114, 12111, 12080, 12129, 12101, 12100, 12131, 12130, 12104, 12109, 12135, 12134, 12134, 12131, 12105, 12129, 12135, 12109, 12130, 12115, 12084, 12188, 12189, 12129, 12192, 12193, 12137, 12202, 12145, 12143, 12189, 12192, 12135, 12193, 12200, 12143, 12218, 12179, 12177, 12230, 12187, 12185, 12228, 12185, 12179, 12216, 12177, 12145, 11732, 11731, 11396, 11736, 11735, 11400, 11735, 11732, 11397, 11748, 11747, 11412, 11752, 11751, 11416, 11751, 11748, 11413, 11747, 11736, 11401, 11796, 11470, 11469, 11474, 11800, 11799, 11800, 11474, 11470, 11488, 11815, 11814, 11821, 11820, 11493, 11493, 11820, 11818, 11490, 11818, 11815, 11799, 11752, 11417, 12033, 12032, 11683, 12037, 12036, 11687, 12036, 12033, 11684, 12050, 11698, 11697, 12050, 12054, 11700, 11700, 12054, 12037, 11509, 11510, 11285, 11515, 11292, 11291,
11285, 11510, 11513, 11540, 11542, 11331, 11554, 11342, 11340, 11552, 11340, 11331, 11540, 11329, 11292, 11653, 11448, 11446, 11665, 11458, 11456, 11663, 11456, 11448, 11675, 11484, 11482, 11683, 11494, 11492, 11681, 11492, 11484, 11673, 11482, 11458, 11651, 11446, 11342, 12034, 11823, 11821, 12042, 11833, 11831, 12040, 11831, 11823, 12071, 11865, 11863, 12079, 11873, 11871, 12077, 11871, 11865, 12069, 11863, 11833, 12112, 11983, 11981, 12119, 12118, 11990, 12120, 12122, 11995, 12122, 12119, 11991, 12112, 12120, 11994, 12110, 11981, 11873, 12032, 11821, 11494, 10805, 11249, 11248, 11255, 11254, 10808, 10805, 10808, 11254, 11269, 10832, 10830, 11273, 11272, 10836, 11276, 11273, 10837, 11269, 11276, 10840, 10836, 11272, 11255, 10891, 11498, 11496, 10903, 11506, 11505, 11505, 11502, 10899, 10899, 11502, 11498, 11524, 10926, 10925, 11528, 11527, 10931, 11524, 11528, 10932, 10931, 11527, 11506, 11384, 11385, 10950, 11388, 10957, 10955, 11396, 10972, 10970, 10950, 11385, 11386, 11388, 11394, 10970, 11733, 11019, 11017, 11741, 11027, 11025, 11739, 11025, 11019, 11765, 11055, 11053, 11771, 11773, 11063, 11771, 11061, 11055, 11763, 11053, 11027, 11876, 11874, 11142, 11153, 11880, 11876, 11880, 11153, 11063, 11731, 11017, 10972, 12252, 12254, 12253,
12258, 12260, 12259, 12254, 12258, 12257, 12252, 12255, 12256, 12258, 12261, 12262, 12254, 12256, 12261, 12264, 12266, 12265, 12270, 12272, 12271, 12266, 12270, 12269, 12264, 12267, 12268, 12273, 12274, 12272, 12268, 12273, 12270, 12260, 12264, 12263, 12260, 12262, 12267, 12249, 12252, 12251, 12250, 12255, 12252, 12276, 12278, 12277, 12284, 12286, 12285, 12290, 12292, 12291, 12288, 12292, 12290, 12299, 12298, 12296, 12278, 12288, 12287, 12297, 12296, 12285, 12276, 12281, 12282, 12291, 12292, 12295, 12293, 12295, 12292, 12279, 12280, 12276, 12285, 12282, 12281, 12298, 12295, 12293, 12276, 12280, 12283, 12296, 12293, 12282, 12278, 12282, 12293, 12303, 12302, 12300, 12307, 12306, 12304, 12311, 12310, 12308, 12315, 12314, 12312, 12309, 12308, 12302, 12313, 12312, 12306, 12306, 12303, 12301, 12314, 12311, 12309, 12312, 12309, 12303, 12317, 12319, 12318, 12323, 12322, 12320, 12327, 12326, 12324, 12329, 12331, 12330, 12335, 12334, 12332, 12339, 12338, 12336, 12319, 12329, 12328, 12333, 12332, 12322, 12337, 12336, 12326, 12317, 12320, 12322, 12332, 12334, 12331, 12326, 12323, 12321, 12338, 12335, 12333, 12336, 12333, 12323, 12319, 12322, 12332, 12345, 12344, 12342, 12347, 12349, 12348, 12353, 12351, 12350, 12357, 12356, 12354, 12361, 12360, 12358,
12351, 12341, 12340, 12355, 12354, 12344, 12359, 12358, 12348, 12353, 12361, 12359, 12348, 12345, 12343, 12360, 12357, 12355, 12358, 12355, 12345, 12351, 12359, 12349, 12291, 12317, 12316, 12321, 12320, 12294, 12325, 12324, 12298, 12298, 12324, 12321, 12294, 12320, 12317, 12343, 12342, 12310, 12347, 12346, 12314, 12346, 12343, 12311, 12302, 12286, 12284, 12310, 12299, 12297, 12308, 12297, 12286, 12344, 12327, 12325, 12356, 12339, 12337, 12354, 12337, 12327, 12342, 12325, 12299, 12365, 12363, 12362, 12371, 12370, 12368, 12376, 12374, 12373, 12378, 12375, 12373, 12376, 12365, 12364, 12380, 12379, 12370, 12365, 12371, 12369, 12379, 12380, 12376, 12381, 12379, 12375, 12367, 12366, 12362, 12369, 12372, 12367, 12376, 12380, 12371, 12385, 12383, 12382, 12385, 12387, 12386, 12383, 12378, 12377, 12383, 12386, 12381, 12391, 12390, 12388, 12395, 12394, 12392, 12399, 12398, 12396, 12403, 12402, 12400, 12397, 12396, 12390, 12401, 12400, 12394, 12394, 12391, 12389, 12402, 12399, 12397, 12400, 12397, 12391, 12407, 12405, 12404, 12409, 12406, 12404, 12413, 12412, 12410, 12420, 12418, 12417, 12424, 12423, 12421, 12418, 12409, 12408, 12422, 12421, 12412, 12414, 12415, 12407, 12416, 12414, 12406, 12420, 12426, 12425, 12416, 12413, 12411, 12426, 12424, 12422,
12425, 12422, 12413, 12418, 12425, 12416, 12430, 12429, 12427, 12434, 12433, 12431, 12438, 12437, 12435, 12442, 12441, 12439, 12436, 12435, 12429, 12440, 12439, 12433, 12433, 12430, 12428, 12441, 12438, 12436, 12439, 12436, 12430, 12446, 12444, 12443, 12450, 12449, 12447, 12456, 12454, 12453, 12462, 12461, 12459, 12458, 12446, 12445, 12450, 12460, 12459, 12462, 12460, 12454, 12446, 12452, 12451, 12452, 12450, 12448, 12458, 12457, 12453, 12460, 12463, 12458, 12463, 12460, 12450, 12463, 12452, 12446, 12428, 12427, 12398, 12432, 12431, 12402, 12431, 12428, 12399, 12444, 12420, 12419, 12448, 12447, 12423, 12451, 12448, 12424, 12444, 12451, 12426, 12412, 12395, 12393, 12423, 12403, 12401, 12421, 12401, 12395, 12449, 12434, 12432, 12461, 12442, 12440, 12459, 12440, 12434, 12447, 12432, 12403, 12335, 12389, 12388, 12393, 12392, 12338, 12392, 12389, 12335, 12407, 12353, 12352, 12411, 12410, 12356, 12415, 12414, 12360, 12360, 12414, 12411, 12415, 12361, 12353, 12410, 12393, 12339, 12370, 12307, 12305, 12381, 12315, 12313, 12370, 12379, 12313, 12385, 12384, 12340, 12387, 12349, 12347, 12349, 12387, 12385, 12386, 12347, 12315, 12465, 12467, 12466, 12471, 12473, 12472, 12467, 12471, 12470, 12468, 12469, 12467, 12474, 12475, 12473, 12469, 12474, 12471,
12477, 12479, 12478, 12483, 12485, 12484, 12479, 12483, 12482, 12480, 12481, 12479, 12486, 12487, 12485, 12481, 12486, 12483, 12473, 12477, 12476, 12475, 12480, 12477, 12489, 12491, 12490, 12489, 12492, 12493, 12485, 12489, 12488, 12487, 12492, 12489, 12497, 12496, 12494, 12501, 12500, 12498, 12505, 12504, 12502, 12509, 12508, 12506, 12503, 12502, 12496, 12507, 12506, 12500, 12500, 12497, 12495, 12508, 12505, 12503, 12506, 12503, 12497, 12513, 12511, 12510, 12517, 12516, 12514, 12521, 12519, 12518, 12525, 12524, 12522, 12519, 12513, 12512, 12517, 12523, 12522, 12513, 12517, 12515, 12525, 12523, 12519, 12519, 12523, 12517, 12529, 12528, 12526, 12533, 12532, 12530, 12537, 12536, 12534, 12541, 12540, 12538, 12535, 12534, 12528, 12539, 12538, 12532, 12532, 12529, 12527, 12540, 12537, 12535, 12538, 12535, 12529, 12545, 12543, 12542, 12547, 12549, 12548, 12555, 12553, 12552, 12559, 12558, 12556, 12551, 12545, 12544, 12549, 12557, 12556, 12549, 12547, 12543, 12555, 12561, 12560, 12552, 12553, 12551, 12560, 12561, 12559, 12553, 12560, 12557, 12551, 12557, 12549, 12527, 12526, 12504, 12531, 12530, 12508, 12530, 12527, 12505, 12543, 12521, 12520, 12525, 12547, 12546, 12547, 12525, 12521, 12516, 12501, 12499, 12524, 12509, 12507, 12522, 12507, 12501,
12548, 12533, 12531, 12558, 12541, 12539, 12556, 12539, 12533, 12546, 12531, 12509, 12567, 12566, 12564, 12571, 12570, 12568, 12573, 12575, 12574, 12577, 12579, 12578, 12583, 12582, 12580, 12563, 12573, 12572, 12567, 12577, 12576, 12581, 12580, 12570, 12576, 12578, 12575, 12570, 12567, 12565, 12582, 12579, 12577, 12580, 12577, 12567, 12563, 12566, 12576, 12587, 12585, 12584, 12591, 12590, 12588, 12597, 12595, 12594, 12601, 12600, 12598, 12595, 12587, 12586, 12599, 12598, 12590, 12593, 12592, 12585, 12603, 12602, 12595, 12592, 12593, 12591, 12602, 12603, 12601, 12593, 12602, 12599, 12602, 12593, 12587, 12609, 12607, 12606, 12605, 12608, 12606, 12611, 12613, 12612, 12617, 12616, 12614, 12619, 12621, 12620, 12627, 12626, 12624, 12609, 12619, 12618, 12625, 12624, 12616, 12613, 12609, 12608, 12605, 12610, 12612, 12622, 12623, 12621, 12616, 12613, 12611, 12624, 12626, 12623, 12616, 12624, 12622, 12609, 12613, 12622, 12629, 12631, 12630, 12633, 12631, 12629, 12637, 12636, 12634, 12639, 12641, 12640, 12645, 12643, 12642, 12649, 12648, 12646, 12651, 12653, 12652, 12643, 12633, 12632, 12647, 12646, 12636, 12641, 12651, 12650, 12631, 12639, 12638, 12633, 12641, 12639, 12653, 12651, 12643, 12640, 12637, 12635, 12652, 12649, 12647, 12650, 12647, 12637,
12651, 12641, 12633, 12575, 12605, 12604, 12579, 12611, 12610, 12615, 12614, 12582, 12614, 12611, 12579, 12578, 12610, 12605, 12630, 12597, 12596, 12635, 12634, 12600, 12603, 12638, 12635, 12630, 12638, 12603, 12590, 12571, 12569, 12600, 12583, 12581, 12598, 12581, 12571, 12636, 12617, 12615, 12648, 12627, 12625, 12646, 12625, 12617, 12634, 12615, 12583, 12565, 12564, 12536, 12569, 12568, 12540, 12568, 12565, 12537, 12585, 12555, 12554, 12589, 12588, 12558, 12561, 12592, 12589, 12592, 12561, 12555, 12588, 12569, 12541, 12654, 12655, 12645, 12655, 12656, 12653, 12272, 12465, 12464, 12274, 12468, 12465, 12495, 12494, 12437, 12499, 12498, 12441, 12498, 12495, 12438, 12511, 12456, 12455, 12515, 12514, 12461, 12511, 12515, 12462, 12514, 12499, 12442, 12330, 12331, 12249, 12249, 12331, 12334, 12390, 12256, 12255, 12398, 12262, 12261, 12396, 12261, 12256, 12427, 12429, 12268, 12437, 12274, 12273, 12435, 12273, 12268, 12427, 12267, 12262, 12388, 12255, 12250, 12496, 12469, 12468, 12504, 12475, 12474, 12502, 12474, 12469, 12528, 12481, 12480, 12534, 12536, 12487, 12528, 12534, 12486, 12526, 12480, 12475, 12563, 12562, 12490, 12564, 12566, 12493, 12566, 12563, 12491, 12536, 12564, 12492, 12494, 12468, 12274, 12658, 12660, 12659, 12661, 12662, 12660,
12668, 12667, 12665, 12672, 12671, 12669, 12678, 12676, 12675, 12674, 12677, 12675, 12680, 12682, 12681, 12686, 12685, 12683, 12664, 12674, 12673, 12668, 12680, 12679, 12684, 12683, 12671, 12682, 12678, 12677, 12679, 12681, 12677, 12671, 12668, 12666, 12685, 12682, 12680, 12683, 12680, 12668, 12667, 12679, 12674, 12688, 12690, 12689, 12698, 12697, 12695, 12704, 12702, 12701, 12700, 12703, 12701, 12706, 12708, 12707, 12692, 12700, 12699, 12698, 12706, 12705, 12693, 12694, 12690, 12708, 12704, 12703, 12705, 12707, 12703, 12692, 12691, 12689, 12695, 12697, 12694, 12697, 12692, 12690, 12697, 12705, 12700, 12678, 12688, 12687, 12696, 12695, 12685, 12685, 12695, 12693, 12682, 12693, 12688, 12664, 12663, 12659, 12665, 12667, 12662, 12667, 12664, 12660, 12712, 12711, 12709, 12716, 12715, 12713, 12720, 12719, 12717, 12724, 12723, 12721, 12718, 12717, 12711, 12722, 12721, 12715, 12715, 12712, 12710, 12723, 12720, 12718, 12721, 12718, 12712, 12728, 12730, 12729, 12729, 12726, 12725, 12734, 12733, 12731, 12741, 12739, 12738, 12745, 12744, 12742, 12747, 12749, 12748, 12739, 12730, 12728, 12743, 12742, 12733, 12737, 12747, 12746, 12730, 12737, 12736, 12736, 12735, 12726, 12749, 12747, 12739, 12735, 12736, 12734, 12748, 12745, 12743, 12746, 12743, 12734,
12747, 12737, 12730, 12753, 12752, 12750, 12757, 12756, 12754, 12761, 12760, 12758, 12765, 12764, 12762, 12759, 12758, 12752, 12763, 12762, 12756, 12756, 12753, 12751, 12764, 12761, 12759, 12762, 12759, 12753, 12769, 12768, 12766, 12773, 12772, 12770, 12777, 12776, 12774, 12781, 12780, 12778, 12775, 12774, 12768, 12779, 12778, 12772, 12772, 12769, 12767, 12780, 12777, 12775, 12778, 12775, 12769, 12751, 12750, 12719, 12755, 12754, 12723, 12754, 12751, 12720, 12767, 12766, 12744, 12771, 12770, 12748, 12770, 12767, 12745, 12733, 12716, 12714, 12744, 12724, 12722, 12742, 12722, 12716, 12768, 12757, 12755, 12776, 12765, 12763, 12774, 12763, 12757, 12766, 12755, 12724, 12783, 12785, 12784, 12786, 12787, 12785, 12793, 12792, 12790, 12797, 12796, 12794, 12801, 12799, 12798, 12801, 12803, 12802, 12810, 12809, 12807, 12789, 12800, 12798, 12793, 12805, 12804, 12808, 12807, 12796, 12805, 12801, 12800, 12805, 12806, 12803, 12796, 12793, 12791, 12807, 12809, 12806, 12807, 12805, 12793, 12789, 12792, 12804, 12814, 12813, 12811, 12818, 12817, 12815, 12822, 12821, 12819, 12826, 12825, 12823, 12820, 12819, 12813, 12824, 12823, 12817, 12817, 12814, 12812, 12825, 12822, 12820, 12823, 12820, 12814, 12834, 12832, 12831, 12830, 12833, 12831, 12837, 12839, 12838,
12841, 12843, 12842, 12834, 12841, 12840, 12839, 12834, 12833, 12830, 12836, 12838, 12844, 12845, 12843, 12830, 12829, 12827, 12836, 12830, 12828, 12839, 12844, 12841, 12849, 12848, 12846, 12853, 12852, 12850, 12859, 12858, 12856, 12863, 12862, 12860, 12857, 12856, 12848, 12861, 12860, 12852, 12852, 12849, 12847, 12862, 12859, 12857, 12860, 12857, 12849, 12803, 12828, 12827, 12837, 12836, 12809, 12809, 12836, 12835, 12806, 12835, 12828, 12847, 12846, 12821, 12851, 12850, 12825, 12850, 12847, 12822, 12813, 12797, 12795, 12821, 12810, 12808, 12819, 12808, 12797, 12848, 12839, 12837, 12855, 12854, 12842, 12856, 12858, 12845, 12845, 12858, 12855, 12848, 12856, 12844, 12846, 12837, 12810, 12704, 12783, 12782, 12708, 12786, 12783, 12791, 12790, 12760, 12795, 12794, 12764, 12794, 12791, 12761, 12812, 12811, 12776, 12816, 12815, 12780, 12815, 12812, 12777, 12811, 12795, 12765, 12711, 12672, 12670, 12719, 12686, 12684, 12717, 12684, 12672, 12752, 12698, 12696, 12760, 12708, 12706, 12758, 12706, 12698, 12750, 12696, 12686, 12789, 12788, 12784, 12790, 12792, 12787, 12792, 12789, 12785, 12760, 12790, 12786, 12869, 12867, 12866, 12865, 12868, 12866, 12871, 12873, 12872, 12877, 12879, 12878, 12869, 12877, 12876, 12873, 12869, 12868, 12870, 12872, 12868,
12881, 12882, 12879, 12873, 12881, 12877, 12886, 12885, 12883, 12890, 12889, 12887, 12896, 12895, 12893, 12900, 12899, 12897, 12894, 12893, 12885, 12898, 12897, 12889, 12889, 12886, 12884, 12899, 12896, 12894, 12897, 12894, 12886, 12904, 12906, 12905, 12905, 12902, 12901, 12915, 12913, 12912, 12919, 12918, 12916, 12911, 12906, 12904, 12909, 12917, 12916, 12906, 12909, 12908, 12908, 12907, 12902, 12921, 12920, 12913, 12912, 12913, 12911, 12920, 12921, 12919, 12913, 12920, 12917, 12911, 12917, 12909, 12925, 12923, 12922, 12925, 12927, 12926, 12934, 12933, 12931, 12942, 12940, 12939, 12938, 12941, 12939, 12945, 12947, 12946, 12927, 12936, 12935, 12945, 12944, 12933, 12929, 12925, 12924, 12929, 12930, 12927, 12947, 12942, 12941, 12938, 12944, 12946, 12931, 12933, 12930, 12938, 12937, 12935, 12944, 12938, 12936, 12933, 12944, 12943, 12930, 12943, 12936, 12902, 12875, 12874, 12907, 12880, 12875, 12892, 12924, 12922, 12896, 12929, 12928, 12932, 12931, 12899, 12931, 12929, 12896, 12892, 12895, 12928, 12885, 12873, 12871, 12892, 12891, 12878, 12893, 12895, 12882, 12882, 12895, 12892, 12885, 12893, 12881, 12953, 12952, 12950, 12957, 12956, 12954, 12961, 12960, 12958, 12965, 12964, 12962, 12959, 12958, 12952, 12963, 12962, 12956, 12956, 12953, 12951,
12964, 12961, 12959, 12962, 12959, 12953, 12969, 12967, 12966, 12976, 12978, 12977, 12980, 12978, 12976, 12984, 12986, 12985, 12977, 12969, 12968, 12973, 12972, 12967, 12978, 12984, 12983, 12980, 12986, 12984, 12981, 12982, 12980, 12982, 12987, 12986, 12983, 12973, 12969, 12991, 12990, 12988, 12995, 12994, 12992, 12999, 12998, 12996, 13003, 13002, 13000, 12997, 12996, 12990, 13001, 13000, 12994, 12994, 12991, 12989, 13002, 12999, 12997, 13000, 12997, 12991, 13007, 13009, 13008, 13008, 13005, 13004, 13013, 13012, 13010, 13020, 13018, 13017, 13024, 13023, 13021, 13026, 13028, 13027, 13018, 13009, 13007, 13022, 13021, 13012, 13016, 13026, 13025, 13009, 13016, 13015, 13015, 13014, 13005, 13028, 13026, 13018, 13014, 13015, 13013, 13027, 13024, 13022, 13025, 13022, 13013, 13026, 13016, 13009, 12989, 12988, 12960, 12993, 12992, 12964, 12992, 12989, 12961, 13005, 12982, 12981, 13011, 13010, 12985, 12987, 13014, 13011, 13014, 12987, 12982, 12966, 12967, 12949, 12972, 12973, 12957, 12985, 12965, 12963, 12967, 12972, 12955, 12973, 12983, 12963, 13012, 12995, 12993, 13023, 13003, 13001, 13021, 13001, 12995, 13010, 12993, 12965, 13030, 13032, 13031, 13034, 13032, 13030, 13038, 13037, 13035, 13040, 13042, 13041, 13046, 13045, 13043, 13050, 13049, 13047,
13044, 13043, 13037, 13048, 13047, 13041, 13032, 13040, 13039, 13034, 13042, 13040, 13041, 13038, 13036, 13049, 13046, 13044, 13047, 13044, 13038, 13054, 13056, 13055, 13063, 13065, 13064, 13064, 13061, 13060, 13061, 13052, 13051, 13065, 13068, 13067, 13067, 13066, 13061, 13058, 13059, 13056, 13066, 13057, 13052, 13072, 13071, 13069, 13076, 13075, 13073, 13080, 13079, 13077, 13084, 13083, 13081, 13078, 13077, 13071, 13082, 13081, 13075, 13075, 13072, 13070, 13083, 13080, 13078, 13081, 13078, 13072, 13088, 13086, 13085, 13092, 13094, 13093, 13097, 13099, 13098, 13101, 13099, 13097, 13105, 13104, 13102, 13107, 13109, 13108, 13098, 13090, 13089, 13103, 13102, 13093, 13094, 13092, 13086, 13099, 13107, 13106, 13101, 13109, 13107, 13089, 13090, 13088, 13108, 13105, 13103, 13090, 13095, 13094, 13095, 13106, 13103, 13098, 13106, 13095, 13070, 13069, 13045, 13074, 13073, 13049, 13073, 13070, 13046, 13086, 13065, 13063, 13068, 13092, 13091, 13092, 13068, 13065, 13051, 13052, 13034, 13066, 13067, 13050, 13034, 13052, 13057, 13057, 13066, 13048, 13093, 13076, 13074, 13104, 13084, 13082, 13102, 13082, 13076, 13091, 13074, 13050, 13115, 13114, 13112, 13119, 13118, 13116, 13123, 13121, 13120, 13123, 13125, 13124, 13132, 13131, 13129, 13111, 13122, 13120,
13115, 13127, 13126, 13130, 13129, 13118, 13127, 13123, 13122, 13127, 13128, 13125, 13118, 13115, 13113, 13129, 13131, 13128, 13129, 13127, 13115, 13111, 13114, 13126, 13136, 13135, 13133, 13140, 13139, 13137, 13144, 13143, 13141, 13148, 13147, 13145, 13142, 13141, 13135, 13146, 13145, 13139, 13139, 13136, 13134, 13147, 13144, 13142, 13145, 13142, 13136, 13156, 13154, 13153, 13152, 13155, 13153, 13159, 13161, 13160, 13165, 13167, 13166, 13156, 13165, 13164, 13161, 13156, 13155, 13152, 13158, 13160, 13169, 13170, 13167, 13152, 13151, 13149, 13158, 13152, 13150, 13161, 13169, 13165, 13174, 13173, 13171, 13178, 13177, 13175, 13184, 13183, 13181, 13188, 13187, 13185, 13182, 13181, 13173, 13186, 13185, 13177, 13177, 13174, 13172, 13187, 13184, 13182, 13185, 13182, 13174, 13125, 13150, 13149, 13159, 13158, 13131, 13131, 13158, 13157, 13128, 13157, 13150, 13172, 13171, 13143, 13176, 13175, 13147, 13175, 13172, 13144, 13135, 13119, 13117, 13143, 13132, 13130, 13141, 13130, 13119, 13173, 13161, 13159, 13180, 13179, 13166, 13181, 13183, 13170, 13170, 13183, 13180, 13173, 13181, 13169, 13171, 13159, 13132, 13031, 12915, 12914, 13036, 13035, 12918, 12921, 13039, 13036, 13039, 12921, 12915, 12942, 13054, 13053, 12947, 13058, 13054, 13113, 13112, 12998,
13117, 13116, 13002, 13116, 13113, 12999, 13134, 13133, 13023, 13138, 13137, 13027, 13137, 13134, 13024, 13133, 13117, 13003, 12952, 12890, 12888, 12960, 12900, 12898, 12958, 12898, 12890, 12990, 12934, 12932, 12998, 12947, 12945, 12996, 12945, 12934, 12988, 12932, 12900, 13111, 13110, 13055, 13112, 13114, 13059, 13114, 13111, 13056, 13162, 13163, 13101, 13101, 13163, 13168, 12998, 13112, 13058, 13192, 13190, 13189, 13192, 13194, 13193, 13203, 13202, 13200, 13207, 13205, 13204, 13207, 13209, 13208, 13196, 13206, 13204, 13203, 13211, 13210, 13198, 13192, 13191, 13192, 13198, 13199, 13211, 13207, 13206, 13211, 13212, 13209, 13196, 13195, 13193, 13200, 13202, 13199, 13202, 13196, 13194, 13196, 13202, 13210, 13209, 13214, 13213, 13212, 13215, 13214, 13219, 13218, 13216, 13223, 13222, 13220, 13227, 13226, 13224, 13231, 13230, 13228, 13225, 13224, 13218, 13229, 13228, 13222, 13222, 13219, 13217, 13230, 13227, 13225, 13228, 13225, 13219, 13235, 13234, 13232, 13239, 13238, 13236, 13243, 13242, 13240, 13247, 13246, 13244, 13241, 13240, 13234, 13245, 13244, 13238, 13238, 13235, 13233, 13246, 13243, 13241, 13244, 13241, 13235, 13253, 13251, 13250, 13249, 13252, 13250, 13255, 13257, 13256, 13261, 13260, 13258, 13265, 13267, 13266, 13270, 13272, 13271,
13253, 13263, 13262, 13270, 13269, 13260, 13257, 13253, 13252, 13249, 13254, 13256, 13269, 13271, 13267, 13260, 13257, 13255, 13265, 13264, 13262, 13269, 13265, 13263, 13260, 13269, 13268, 13253, 13257, 13268, 13276, 13275, 13273, 13280, 13279, 13277, 13284, 13283, 13281, 13288, 13287, 13285, 13282, 13281, 13275, 13286, 13285, 13279, 13279, 13276, 13274, 13287, 13284, 13282, 13285, 13282, 13276, 13255, 13254, 13226, 13259, 13258, 13230, 13258, 13255, 13227, 13274, 13273, 13242, 13278, 13277, 13246, 13277, 13274, 13243, 13234, 13223, 13221, 13242, 13231, 13229, 13240, 13229, 13223, 13275, 13261, 13259, 13283, 13272, 13270, 13281, 13270, 13261, 13273, 13259, 13231, 13292, 13294, 13293, 13293, 13290, 13289, 13303, 13301, 13300, 13307, 13306, 13304, 13299, 13294, 13292, 13297, 13305, 13304, 13294, 13297, 13296, 13296, 13295, 13290, 13309, 13308, 13301, 13300, 13301, 13299, 13308, 13309, 13307, 13301, 13308, 13305, 13299, 13305, 13297, 13313, 13311, 13310, 13313, 13315, 13314, 13324, 13323, 13321, 13330, 13328, 13327, 13326, 13329, 13327, 13332, 13334, 13333, 13317, 13326, 13325, 13324, 13332, 13331, 13319, 13313, 13312, 13313, 13319, 13320, 13334, 13330, 13329, 13331, 13333, 13329, 13317, 13316, 13314, 13321, 13323, 13320, 13323, 13317, 13315,
13323, 13331, 13326, 13336, 13338, 13337, 13340, 13338, 13336, 13344, 13343, 13341, 13346, 13348, 13347, 13352, 13351, 13349, 13356, 13355, 13353, 13350, 13349, 13343, 13354, 13353, 13347, 13338, 13346, 13345, 13340, 13348, 13346, 13347, 13344, 13342, 13355, 13352, 13350, 13353, 13350, 13344, 13364, 13366, 13365, 13368, 13366, 13364, 13370, 13372, 13371, 13365, 13358, 13357, 13365, 13366, 13370, 13372, 13370, 13366, 13365, 13369, 13361, 13337, 13303, 13302, 13342, 13341, 13306, 13309, 13345, 13342, 13345, 13309, 13303, 13330, 13360, 13359, 13330, 13334, 13362, 13357, 13358, 13340, 13371, 13356, 13354, 13358, 13361, 13348, 13361, 13369, 13354, 13376, 13374, 13373, 13376, 13378, 13377, 13379, 13380, 13376, 13380, 13381, 13378, 13387, 13386, 13384, 13391, 13390, 13388, 13395, 13393, 13392, 13395, 13397, 13396, 13406, 13405, 13403, 13383, 13394, 13392, 13387, 13401, 13400, 13404, 13403, 13390, 13401, 13395, 13394, 13395, 13401, 13402, 13390, 13387, 13385, 13399, 13398, 13396, 13403, 13405, 13402, 13405, 13399, 13397, 13403, 13401, 13387, 13383, 13386, 13400, 13410, 13408, 13407, 13410, 13412, 13411, 13412, 13417, 13416, 13414, 13410, 13409, 13414, 13415, 13412, 13415, 13418, 13417, 13399, 13409, 13407, 13406, 13414, 13413, 13399, 13405, 13413,
13383, 13382, 13377, 13384, 13386, 13381, 13386, 13383, 13378, 13422, 13421, 13419, 13426, 13425, 13423, 13430, 13429, 13427, 13434, 13433, 13431, 13428, 13427, 13421, 13432, 13431, 13425, 13425, 13422, 13420, 13433, 13430, 13428, 13431, 13428, 13422, 13440, 13442, 13441, 13441, 13438, 13437, 13446, 13445, 13443, 13453, 13451, 13450, 13457, 13456, 13454, 13459, 13461, 13460, 13451, 13442, 13440, 13455, 13454, 13445, 13449, 13459, 13458, 13442, 13449, 13448, 13441, 13448, 13447, 13461, 13459, 13451, 13437, 13438, 13436, 13447, 13448, 13446, 13460, 13457, 13455, 13438, 13447, 13444, 13458, 13455, 13446, 13459, 13449, 13442, 13465, 13464, 13462, 13469, 13468, 13466, 13475, 13473, 13472, 13471, 13474, 13472, 13477, 13479, 13478, 13483, 13482, 13480, 13477, 13476, 13464, 13481, 13480, 13468, 13479, 13475, 13474, 13471, 13476, 13478, 13468, 13465, 13463, 13482, 13479, 13477, 13480, 13477, 13465, 13487, 13486, 13484, 13491, 13490, 13488, 13495, 13494, 13492, 13499, 13498, 13496, 13493, 13492, 13486, 13497, 13496, 13490, 13490, 13487, 13485, 13498, 13495, 13493, 13496, 13493, 13487, 13463, 13462, 13429, 13467, 13466, 13433, 13466, 13463, 13430, 13485, 13484, 13456, 13489, 13488, 13460, 13488, 13485, 13457, 13445, 13426, 13424, 13456, 13434, 13432,
13454, 13432, 13426, 13486, 13469, 13467, 13494, 13483, 13481, 13492, 13481, 13469, 13484, 13467, 13434, 13267, 13375, 13373, 13272, 13380, 13379, 13271, 13379, 13375, 13385, 13384, 13283, 13389, 13388, 13287, 13388, 13385, 13284, 13384, 13380, 13272, 13420, 13419, 13351, 13424, 13423, 13355, 13423, 13420, 13352, 13436, 13368, 13367, 13372, 13444, 13443, 13444, 13372, 13368, 13443, 13424, 13356, 13295, 13296, 13239, 13306, 13247, 13245, 13304, 13245, 13239, 13343, 13280, 13278, 13351, 13288, 13286, 13349, 13286, 13280, 13341, 13278, 13247, 13421, 13391, 13389, 13429, 13406, 13404, 13427, 13404, 13391, 13462, 13464, 13415, 13471, 13470, 13416, 13418, 13476, 13471, 13464, 13476, 13418, 13462, 13414, 13406, 13419, 13389, 13288, 12855, 13191, 13189, 12859, 13198, 13197, 13201, 13200, 12862, 13200, 13198, 12859, 12855, 12858, 13197, 13217, 13216, 13079, 13221, 13220, 13083, 13220, 13217, 13080, 13233, 13232, 13104, 13237, 13236, 13108, 13236, 13233, 13105, 13232, 13221, 13084, 13290, 13163, 13162, 13295, 13168, 13163, 13180, 13312, 13310, 13184, 13319, 13318, 13322, 13321, 13187, 13321, 13319, 13184, 13180, 13183, 13318, 13168, 13295, 13237, 12874, 12875, 12741, 12875, 12880, 12749, 12907, 12908, 12773, 12918, 12781, 12779, 12916, 12779, 12773,
12880, 12907, 12771, 13037, 12818, 12816, 13045, 12826, 12824, 13043, 12824, 12818, 13071, 12853, 12851, 13079, 12863, 12861, 13077, 12861, 12853, 13069, 12851, 12826, 13035, 12816, 12781, 13218, 13203, 13201, 13224, 13226, 13212, 13224, 13211, 13203, 13249, 13248, 13213, 13254, 13249, 13214, 13226, 13254, 13215, 13216, 13201, 12863, 13503, 13505, 13504, 13508, 13510, 13509, 13516, 13514, 13513, 13512, 13515, 13513, 13518, 13520, 13519, 13505, 13512, 13511, 13510, 13518, 13517, 13503, 13507, 13509, 13520, 13516, 13515, 13517, 13519, 13515, 13503, 13502, 13500, 13507, 13503, 13501, 13509, 13517, 13512, 13524, 13523, 13521, 13528, 13527, 13525, 13532, 13531, 13529, 13536, 13535, 13533, 13530, 13529, 13523, 13534, 13533, 13527, 13527, 13524, 13522, 13535, 13532, 13530, 13533, 13530, 13524, 13538, 13540, 13539, 13541, 13542, 13540, 13548, 13547, 13545, 13552, 13551, 13549, 13558, 13556, 13555, 13554, 13557, 13555, 13560, 13562, 13561, 13566, 13565, 13563, 13544, 13554, 13553, 13548, 13560, 13559, 13564, 13563, 13551, 13562, 13558, 13557, 13559, 13561, 13557, 13551, 13548, 13546, 13565, 13562, 13560, 13563, 13560, 13548, 13547, 13559, 13554, 13516, 13538, 13537, 13520, 13541, 13538, 13546, 13545, 13531, 13550, 13549, 13535, 13549, 13546, 13532,
13523, 13510, 13508, 13531, 13520, 13518, 13529, 13518, 13510, 13544, 13543, 13539, 13545, 13547, 13542, 13547, 13544, 13540, 13531, 13545, 13541, 13570, 13568, 13567, 13574, 13576, 13575, 13578, 13576, 13574, 13580, 13582, 13581, 13575, 13570, 13569, 13572, 13571, 13568, 13576, 13580, 13579, 13582, 13580, 13576, 13575, 13579, 13572, 13588, 13586, 13585, 13592, 13591, 13589, 13596, 13598, 13597, 13600, 13598, 13596, 13604, 13603, 13601, 13606, 13608, 13607, 13597, 13588, 13587, 13602, 13601, 13591, 13594, 13593, 13586, 13598, 13606, 13605, 13600, 13608, 13606, 13585, 13586, 13584, 13593, 13594, 13592, 13607, 13604, 13602, 13586, 13593, 13590, 13594, 13605, 13602, 13605, 13594, 13588, 13584, 13578, 13577, 13582, 13590, 13589, 13590, 13582, 13578, 13609, 13610, 13600, 13600, 13610, 13611, 13613, 13615, 13614, 13623, 13622, 13620, 13627, 13625, 13624, 13627, 13629, 13628, 13617, 13626, 13624, 13623, 13631, 13630, 13618, 13619, 13615, 13631, 13627, 13626, 13627, 13631, 13632, 13617, 13616, 13614, 13620, 13622, 13619, 13619, 13622, 13617, 13617, 13622, 13630, 13636, 13635, 13633, 13640, 13639, 13637, 13646, 13645, 13643, 13650, 13649, 13647, 13644, 13643, 13635, 13648, 13647, 13639, 13639, 13636, 13634, 13649, 13646, 13644, 13647, 13644, 13636,
13654, 13656, 13655, 13655, 13652, 13651, 13665, 13663, 13662, 13669, 13668, 13666, 13661, 13656, 13654, 13659, 13667, 13666, 13656, 13659, 13658, 13658, 13657, 13652, 13671, 13670, 13663, 13662, 13663, 13661, 13670, 13671, 13669, 13663, 13670, 13667, 13661, 13667, 13659, 13677, 13675, 13674, 13673, 13676, 13674, 13679, 13681, 13680, 13685, 13684, 13682, 13693, 13691, 13690, 13689, 13692, 13690, 13696, 13698, 13697, 13677, 13687, 13686, 13696, 13695, 13684, 13681, 13677, 13676, 13678, 13680, 13676, 13698, 13693, 13692, 13689, 13695, 13697, 13684, 13681, 13679, 13689, 13688, 13686, 13695, 13689, 13687, 13684, 13695, 13694, 13677, 13681, 13694, 13700, 13702, 13701, 13704, 13702, 13700, 13708, 13707, 13705, 13710, 13712, 13711, 13716, 13715, 13713, 13720, 13719, 13717, 13714, 13713, 13707, 13718, 13717, 13711, 13702, 13710, 13709, 13704, 13712, 13710, 13711, 13708, 13706, 13719, 13716, 13714, 13717, 13714, 13708, 13642, 13673, 13672, 13646, 13679, 13678, 13683, 13682, 13649, 13682, 13679, 13646, 13645, 13678, 13673, 13701, 13665, 13664, 13706, 13705, 13668, 13671, 13709, 13706, 13709, 13671, 13665, 13657, 13658, 13640, 13668, 13650, 13648, 13666, 13648, 13640, 13707, 13685, 13683, 13715, 13698, 13696, 13713, 13696, 13685, 13705, 13683, 13650,
13558, 13613, 13612, 13621, 13620, 13565, 13565, 13620, 13618, 13562, 13618, 13613, 13634, 13633, 13603, 13638, 13637, 13607, 13637, 13634, 13604, 13652, 13610, 13609, 13657, 13611, 13610, 13611, 13657, 13638, 13571, 13572, 13528, 13581, 13536, 13534, 13572, 13579, 13534, 13591, 13552, 13550, 13603, 13566, 13564, 13601, 13564, 13552, 13589, 13550, 13536, 13635, 13623, 13621, 13642, 13641, 13628, 13643, 13645, 13632, 13645, 13642, 13629, 13643, 13631, 13623, 13633, 13621, 13566, 13729, 13727, 13726, 13725, 13728, 13726, 13731, 13733, 13732, 13737, 13736, 13734, 13745, 13743, 13742, 13741, 13744, 13742, 13748, 13750, 13749, 13729, 13739, 13738, 13748, 13747, 13736, 13732, 13733, 13729, 13725, 13730, 13732, 13749, 13750, 13745, 13747, 13749, 13744, 13736, 13733, 13731, 13741, 13740, 13738, 13746, 13747, 13741, 13736, 13747, 13746, 13733, 13746, 13739, 13745, 13752, 13751, 13750, 13753, 13752, 13725, 13724, 13721, 13730, 13725, 13722, 13693, 13722, 13721, 13693, 13698, 13723, 13731, 13730, 13715, 13735, 13734, 13719, 13734, 13731, 13716, 13715, 13730, 13723, 13758, 13760, 13759, 13762, 13760, 13758, 13766, 13768, 13767, 13759, 13755, 13754, 13760, 13766, 13765, 13762, 13768, 13766, 13763, 13764, 13762, 13762, 13764, 13769, 13759, 13765, 13756,
13777, 13775, 13774, 13773, 13776, 13774, 13780, 13782, 13781, 13784, 13786, 13785, 13777, 13784, 13783, 13782, 13777, 13776, 13773, 13779, 13781, 13787, 13788, 13786, 13773, 13772, 13770, 13779, 13773, 13771, 13782, 13787, 13784, 13792, 13791, 13789, 13796, 13795, 13793, 13802, 13801, 13799, 13806, 13805, 13803, 13800, 13799, 13791, 13804, 13803, 13795, 13795, 13792, 13790, 13805, 13802, 13800, 13803, 13800, 13792, 13810, 13808, 13807, 13810, 13812, 13811, 13821, 13820, 13818, 13827, 13825, 13824, 13823, 13826, 13824, 13829, 13831, 13830, 13814, 13823, 13822, 13821, 13829, 13828, 13816, 13810, 13809, 13810, 13816, 13817, 13831, 13827, 13826, 13828, 13830, 13826, 13814, 13813, 13811, 13818, 13820, 13817, 13820, 13814, 13812, 13820, 13828, 13823, 13798, 13809, 13807, 13802, 13816, 13815, 13819, 13818, 13805, 13818, 13816, 13802, 13798, 13801, 13815, 13791, 13782, 13780, 13798, 13797, 13785, 13799, 13801, 13788, 13801, 13798, 13786, 13791, 13799, 13787, 13827, 13833, 13832, 13827, 13831, 13834, 13836, 13838, 13837, 13840, 13838, 13836, 13844, 13843, 13841, 13846, 13848, 13847, 13852, 13851, 13849, 13856, 13855, 13853, 13850, 13849, 13843, 13854, 13853, 13847, 13837, 13838, 13846, 13848, 13846, 13838, 13847, 13844, 13842, 13855, 13852, 13850,
13853, 13850, 13844, 13858, 13860, 13859, 13868, 13870, 13869, 13869, 13866, 13865, 13874, 13873, 13871, 13864, 13860, 13858, 13862, 13872, 13871, 13860, 13862, 13861, 13869, 13870, 13877, 13876, 13875, 13866, 13865, 13866, 13864, 13875, 13876, 13874, 13864, 13866, 13875, 13864, 13872, 13862, 13883, 13881, 13880, 13879, 13882, 13880, 13887, 13889, 13888, 13893, 13892, 13890, 13897, 13895, 13894, 13897, 13899, 13898, 13885, 13896, 13894, 13893, 13901, 13900, 13889, 13883, 13882, 13879, 13886, 13888, 13900, 13901, 13897, 13901, 13902, 13899, 13885, 13884, 13881, 13890, 13892, 13889, 13892, 13885, 13883, 13892, 13900, 13896, 13908, 13907, 13905, 13912, 13911, 13909, 13916, 13914, 13913, 13918, 13920, 13919, 13924, 13923, 13921, 13918, 13917, 13907, 13922, 13921, 13911, 13920, 13916, 13915, 13911, 13908, 13906, 13923, 13920, 13918, 13921, 13918, 13908, 13887, 13886, 13851, 13891, 13890, 13855, 13890, 13887, 13852, 13904, 13870, 13868, 13906, 13905, 13873, 13877, 13910, 13909, 13909, 13906, 13874, 13910, 13877, 13870, 13857, 13859, 13840, 13873, 13856, 13854, 13859, 13861, 13848, 13861, 13871, 13854, 13907, 13893, 13891, 13915, 13913, 13898, 13917, 13919, 13902, 13919, 13915, 13899, 13917, 13901, 13893, 13905, 13891, 13856, 13926, 13928, 13927,
13930, 13928, 13926, 13932, 13934, 13933, 13936, 13938, 13937, 13946, 13945, 13943, 13950, 13949, 13947, 13942, 13934, 13932, 13944, 13943, 13937, 13940, 13948, 13947, 13928, 13936, 13935, 13938, 13936, 13928, 13934, 13940, 13939, 13931, 13933, 13930, 13949, 13946, 13944, 13933, 13939, 13938, 13939, 13947, 13944, 13948, 13940, 13934, 13952, 13954, 13953, 13956, 13954, 13952, 13958, 13960, 13959, 13962, 13964, 13963, 13954, 13962, 13961, 13964, 13962, 13954, 13959, 13960, 13966, 13957, 13959, 13956, 13959, 13965, 13964, 13951, 13953, 13942, 13961, 13963, 13950, 13953, 13961, 13948, 13974, 13972, 13971, 13970, 13973, 13971, 13977, 13979, 13978, 13979, 13974, 13973, 13970, 13976, 13978, 13970, 13969, 13967, 13975, 13976, 13970, 13985, 13984, 13982, 13989, 13988, 13986, 13993, 13991, 13990, 13997, 13995, 13994, 13997, 13999, 13998, 13981, 13992, 13990, 13985, 14001, 14000, 13989, 14003, 14002, 14001, 13993, 13992, 14002, 14003, 13997, 14003, 14004, 13999, 13988, 13985, 13983, 13996, 13994, 13991, 14002, 13996, 13993, 13988, 14002, 14001, 13984, 14000, 13992, 14010, 14009, 14007, 14014, 14013, 14011, 14018, 14016, 14015, 14020, 14022, 14021, 14026, 14025, 14023, 14020, 14019, 14009, 14024, 14023, 14013, 14022, 14018, 14017, 14013, 14010, 14008,
14025, 14022, 14020, 14023, 14020, 14010, 14032, 14030, 14029, 14035, 14032, 14031, 14031, 14029, 14027, 14034, 14031, 14028, 14018, 14028, 14027, 14026, 14035, 14034, 14025, 14034, 14033, 14022, 14033, 14028, 14009, 13989, 13987, 14017, 14015, 13998, 14019, 14021, 14004, 14004, 14021, 14017, 14019, 14003, 13989, 13916, 13968, 13967, 13977, 13976, 13923, 13923, 13976, 13975, 13920, 13975, 13968, 13983, 13982, 13945, 13987, 13986, 13949, 13986, 13983, 13946, 14006, 13960, 13958, 14008, 14007, 13963, 13966, 14012, 14011, 13965, 14011, 14008, 14012, 13966, 13960, 14007, 13987, 13950, 13925, 13927, 13904, 13935, 13937, 13912, 13945, 13924, 13922, 13927, 13935, 13910, 13943, 13922, 13912, 13981, 13980, 13972, 13982, 13984, 13979, 13984, 13981, 13974, 13982, 13977, 13924, 14037, 14039, 14038, 14041, 14043, 14042, 14045, 14043, 14041, 14047, 14049, 14048, 14051, 14053, 14052, 14057, 14056, 14054, 14061, 14060, 14058, 14055, 14054, 14048, 14059, 14058, 14052, 14039, 14047, 14046, 14042, 14043, 14051, 14053, 14051, 14043, 14040, 14042, 14039, 14050, 14052, 14049, 14060, 14057, 14055, 14042, 14050, 14047, 14058, 14055, 14049, 14063, 14065, 14064, 14067, 14069, 14068, 14079, 14078, 14076, 14083, 14082, 14080, 14075, 14069, 14067, 14071, 14077, 14076,
14073, 14081, 14080, 14065, 14071, 14070, 14068, 14069, 14073, 14066, 14068, 14065, 14082, 14079, 14077, 14068, 14072, 14071, 14072, 14080, 14077, 14081, 14073, 14069, 14087, 14085, 14084, 14087, 14089, 14088, 14093, 14091, 14090, 14098, 14100, 14099, 14094, 14095, 14087, 14095, 14096, 14089, 14100, 14093, 14092, 14092, 14090, 14088, 14097, 14099, 14096, 14096, 14099, 14092, 14104, 14102, 14101, 14108, 14110, 14109, 14114, 14113, 14111, 14118, 14116, 14115, 14106, 14117, 14115, 14114, 14120, 14119, 14110, 14104, 14103, 14119, 14120, 14118, 14106, 14105, 14102, 14111, 14113, 14110, 14110, 14113, 14106, 14113, 14119, 14117, 14057, 14095, 14094, 14098, 14097, 14060, 14097, 14095, 14057, 14108, 14107, 14078, 14112, 14111, 14082, 14111, 14108, 14079, 14062, 14064, 14045, 14078, 14061, 14059, 14064, 14070, 14053, 14070, 14076, 14059, 14103, 14101, 14091, 14107, 14109, 14100, 14109, 14103, 14093, 14107, 14098, 14061, 14122, 14124, 14123, 14126, 14128, 14127, 14130, 14132, 14131, 14136, 14135, 14133, 14124, 14130, 14129, 14128, 14134, 14133, 14125, 14127, 14124, 14135, 14132, 14130, 14127, 14133, 14130, 14138, 14140, 14139, 14142, 14144, 14143, 14146, 14148, 14147, 14150, 14152, 14151, 14140, 14146, 14145, 14143, 14144, 14150, 14141, 14143, 14140,
14149, 14151, 14148, 14140, 14143, 14149, 14156, 14155, 14153, 14160, 14159, 14157, 14164, 14162, 14161, 14168, 14166, 14165, 14156, 14170, 14169, 14172, 14171, 14159, 14170, 14164, 14163, 14172, 14168, 14167, 14159, 14156, 14154, 14167, 14165, 14162, 14171, 14167, 14164, 14171, 14170, 14156, 14176, 14175, 14173, 14180, 14179, 14177, 14184, 14182, 14181, 14188, 14186, 14185, 14190, 14189, 14175, 14180, 14192, 14191, 14190, 14184, 14183, 14191, 14192, 14188, 14179, 14176, 14174, 14187, 14185, 14182, 14190, 14191, 14187, 14191, 14190, 14176, 14154, 14153, 14131, 14158, 14157, 14135, 14157, 14154, 14132, 14174, 14173, 14147, 14178, 14177, 14151, 14177, 14174, 14148, 14137, 14139, 14128, 14147, 14136, 14134, 14139, 14145, 14134, 14175, 14160, 14158, 14183, 14181, 14166, 14189, 14183, 14168, 14189, 14172, 14160, 14173, 14158, 14136, 14121, 14123, 14075, 14129, 14131, 14083, 14075, 14123, 14129, 14155, 14114, 14112, 14163, 14161, 14116, 14169, 14163, 14118, 14155, 14169, 14120, 14153, 14112, 14083, 13837, 13764, 13763, 13842, 13841, 13767, 13769, 13845, 13842, 13845, 13769, 13764, 14036, 14038, 14006, 14046, 14048, 14014, 14056, 14026, 14024, 14006, 14038, 14046, 14054, 14024, 14014, 14086, 14084, 14030, 14094, 14086, 14032, 14056, 14094, 14035,
12621, 13501, 13500, 13508, 13507, 12626, 12626, 13507, 13506, 12621, 12623, 13506, 13522, 13521, 12648, 13526, 13525, 12652, 13525, 13522, 12649, 13521, 13508, 12627, 13568, 12655, 12654, 13571, 12656, 12655, 12656, 13571, 13526, 13475, 13771, 13770, 13780, 13779, 13482, 13482, 13779, 13778, 13475, 13479, 13778, 13790, 13789, 13494, 13794, 13793, 13498, 13793, 13790, 13495, 13789, 13780, 13483, 13754, 13755, 13704, 13767, 13720, 13718, 13755, 13756, 13712, 13756, 13765, 13718, 13843, 13737, 13735, 13851, 13750, 13748, 13849, 13748, 13737, 13879, 13878, 13751, 13753, 13886, 13879, 13851, 13886, 13753, 13841, 13735, 13720, 11105, 12280, 12279, 12284, 12283, 11108, 11105, 11108, 12283, 12301, 12300, 11120, 12305, 12304, 11124, 12304, 12301, 11121, 12300, 12284, 11109, 12367, 11184, 11183, 11188, 12369, 12368, 12372, 12369, 11188, 12367, 12372, 11191, 12368, 12305, 11125, 11929, 12658, 12657, 11933, 12661, 12658, 12666, 12665, 11944, 12670, 12669, 11948, 12669, 12666, 11945, 11944, 12665, 12661, 12710, 12709, 12006, 12714, 12713, 12010, 12713, 12710, 12007, 12726, 12026, 12025, 12732, 12731, 12029, 12031, 12735, 12732, 12735, 12031, 12026, 12731, 12714, 12011, 12709, 12670, 11949, 12165, 12865, 12864, 12171, 12871, 12870, 12170, 12870, 12865,
12884, 12883, 12182, 12888, 12887, 12186, 12887, 12884, 12183, 12883, 12871, 12171, 12949, 12227, 12226, 12951, 12950, 12230, 12235, 12955, 12954, 12954, 12951, 12231, 12955, 12235, 12227, 12245, 12971, 12970, 12245, 12247, 12974, 12950, 12888, 12187, 14200, 14202, 14201, 14204, 14206, 14205, 14214, 14213, 14211, 14194, 14200, 14199, 14205, 14196, 14195, 14209, 14210, 14202, 14206, 14212, 14211, 14195, 14196, 14194, 14208, 14207, 14201, 14211, 14213, 14210, 14194, 14196, 14198, 14210, 14213, 14208, 14198, 14211, 14209, 14197, 14209, 14200, 14211, 14198, 14196, 14218, 14220, 14219, 14219, 14216, 14215, 14219, 14220, 14223, 14222, 14221, 14216, 14227, 14225, 14224, 14227, 14229, 14228, 14231, 14227, 14226, 14231, 14232, 14229, 14238, 14240, 14239, 14242, 14240, 14238, 14246, 14245, 14243, 14248, 14250, 14249, 14254, 14252, 14251, 14260, 14258, 14257, 14256, 14259, 14257, 14264, 14266, 14265, 14236, 14253, 14251, 14246, 14262, 14261, 14264, 14263, 14249, 14262, 14254, 14253, 14240, 14248, 14247, 14250, 14248, 14240, 14266, 14260, 14259, 14263, 14265, 14259, 14237, 14239, 14234, 14247, 14249, 14246, 14256, 14255, 14252, 14239, 14247, 14244, 14263, 14256, 14254, 14249, 14263, 14262, 14236, 14245, 14261, 14208, 14226, 14224, 14214, 14231, 14230,
14208, 14213, 14230, 14234, 14220, 14218, 14223, 14244, 14243, 14244, 14223, 14220, 14215, 14216, 14206, 14221, 14222, 14214, 14216, 14221, 14212, 14236, 14235, 14228, 14243, 14245, 14232, 14232, 14245, 14236, 14243, 14231, 14214, 14268, 14270, 14269, 14280, 14282, 14281, 14281, 14278, 14277, 14286, 14285, 14283, 14274, 14270, 14268, 14272, 14284, 14283, 14270, 14272, 14271, 14282, 14289, 14288, 14288, 14287, 14278, 14277, 14278, 14274, 14287, 14288, 14286, 14274, 14278, 14287, 14284, 14272, 14270, 14290, 14291, 14282, 14291, 14292, 14289, 14267, 14269, 14242, 14276, 14275, 14258, 14283, 14285, 14266, 14269, 14271, 14250, 14285, 14276, 14260, 14271, 14283, 14264, 14294, 14296, 14295, 14300, 14302, 14301, 14306, 14305, 14303, 14295, 14300, 14299, 14304, 14303, 14297, 14296, 14298, 14297, 14303, 14305, 14302, 14297, 14303, 14300, 14308, 14310, 14309, 14314, 14313, 14311, 14318, 14316, 14315, 14310, 14317, 14315, 14320, 14319, 14313, 14311, 14313, 14310, 14320, 14318, 14317, 14313, 14319, 14317, 14302, 14308, 14307, 14312, 14311, 14305, 14305, 14311, 14308, 14322, 14324, 14323, 14326, 14328, 14327, 14336, 14335, 14333, 14340, 14339, 14337, 14334, 14333, 14329, 14338, 14337, 14331, 14324, 14330, 14329, 14328, 14332, 14331, 14325, 14327, 14324,
14339, 14336, 14334, 14327, 14331, 14330, 14337, 14334, 14330, 14342, 14344, 14343, 14346, 14348, 14347, 14356, 14355, 14353, 14360, 14359, 14357, 14354, 14353, 14349, 14358, 14357, 14351, 14344, 14350, 14349, 14348, 14352, 14351, 14345, 14347, 14344, 14359, 14356, 14354, 14347, 14351, 14350, 14357, 14354, 14350, 14364, 14363, 14361, 14368, 14367, 14365, 14372, 14370, 14369, 14376, 14374, 14373, 14378, 14377, 14363, 14380, 14379, 14367, 14378, 14372, 14371, 14380, 14376, 14375, 14367, 14364, 14362, 14375, 14373, 14370, 14379, 14375, 14372, 14379, 14378, 14364, 14384, 14383, 14381, 14388, 14387, 14385, 14392, 14390, 14389, 14396, 14394, 14393, 14398, 14397, 14383, 14400, 14399, 14387, 14398, 14392, 14391, 14400, 14396, 14395, 14387, 14384, 14382, 14395, 14393, 14390, 14399, 14395, 14392, 14399, 14398, 14384, 14362, 14361, 14335, 14366, 14365, 14339, 14365, 14362, 14336, 14382, 14381, 14355, 14386, 14385, 14359, 14385, 14382, 14356, 14341, 14343, 14328, 14355, 14340, 14338, 14343, 14349, 14332, 14353, 14338, 14332, 14383, 14368, 14366, 14391, 14389, 14374, 14397, 14391, 14376, 14397, 14380, 14368, 14381, 14366, 14340, 14321, 14323, 14296, 14335, 14306, 14304, 14323, 14329, 14298, 14333, 14304, 14298, 14363, 14314, 14312, 14371, 14369, 14316,
14377, 14371, 14318, 14377, 14320, 14314, 14361, 14312, 14306, 14402, 14404, 14403, 14408, 14406, 14405, 14414, 14416, 14415, 14422, 14424, 14423, 14423, 14419, 14417, 14420, 14424, 14422, 14404, 14414, 14413, 14419, 14408, 14407, 14409, 14410, 14404, 14425, 14426, 14416, 14412, 14411, 14406, 14424, 14430, 14429, 14429, 14427, 14419, 14420, 14428, 14430, 14405, 14406, 14402, 14411, 14412, 14410, 14429, 14426, 14425, 14406, 14411, 14409, 14427, 14425, 14410, 14410, 14425, 14414, 14427, 14412, 14408, 14437, 14439, 14438, 14438, 14442, 14440, 14446, 14448, 14447, 14432, 14439, 14437, 14439, 14446, 14445, 14445, 14447, 14442, 14440, 14442, 14435, 14435, 14442, 14447, 14432, 14433, 14446, 14450, 14452, 14451, 14451, 14454, 14453, 14460, 14459, 14457, 14462, 14464, 14463, 14470, 14468, 14467, 14470, 14466, 14465, 14474, 14472, 14471, 14478, 14477, 14475, 14474, 14456, 14455, 14460, 14476, 14475, 14480, 14479, 14463, 14452, 14458, 14457, 14451, 14457, 14459, 14477, 14478, 14470, 14470, 14478, 14476, 14479, 14480, 14474, 14463, 14460, 14458, 14473, 14471, 14465, 14476, 14479, 14473, 14479, 14476, 14460, 14474, 14480, 14464, 14416, 14452, 14450, 14462, 14461, 14429, 14461, 14458, 14426, 14426, 14458, 14452, 14453, 14454, 14432, 14469, 14467, 14441,
14475, 14477, 14448, 14454, 14459, 14433, 14448, 14477, 14469, 14475, 14446, 14433, 14484, 14482, 14481, 14488, 14490, 14489, 14495, 14497, 14496, 14499, 14501, 14500, 14505, 14503, 14502, 14509, 14508, 14506, 14513, 14512, 14510, 14486, 14501, 14499, 14497, 14511, 14510, 14492, 14484, 14483, 14500, 14501, 14507, 14494, 14496, 14490, 14512, 14513, 14505, 14488, 14487, 14482, 14489, 14490, 14486, 14510, 14512, 14509, 14492, 14494, 14488, 14490, 14496, 14493, 14510, 14507, 14493, 14486, 14493, 14507, 14517, 14515, 14514, 14523, 14521, 14520, 14521, 14517, 14516, 14519, 14518, 14515, 14523, 14525, 14524, 14521, 14524, 14519, 14529, 14527, 14526, 14531, 14528, 14526, 14534, 14535, 14529, 14531, 14536, 14534, 14533, 14532, 14527, 14535, 14537, 14533, 14533, 14504, 14502, 14535, 14534, 14508, 14537, 14535, 14509, 14533, 14537, 14512, 14518, 14519, 14497, 14523, 14522, 14503, 14525, 14513, 14511, 14513, 14525, 14523, 14524, 14511, 14497, 14543, 14541, 14540, 14545, 14546, 14543, 14542, 14540, 14538, 14544, 14545, 14542, 14435, 14539, 14538, 14543, 14443, 14441, 14448, 14546, 14545, 14545, 14544, 14444, 14444, 14544, 14539, 14543, 14546, 14448, 14498, 14500, 14420, 14508, 14430, 14428, 14420, 14500, 14506, 14531, 14530, 14455, 14536, 14464, 14462,
14536, 14531, 14456, 14508, 14534, 14462, 14550, 14552, 14551, 14557, 14559, 14558, 14567, 14566, 14564, 14548, 14557, 14556, 14565, 14564, 14554, 14562, 14563, 14559, 14552, 14555, 14554, 14549, 14551, 14548, 14564, 14566, 14563, 14551, 14554, 14553, 14554, 14564, 14562, 14553, 14562, 14557, 14569, 14571, 14570, 14573, 14575, 14574, 14583, 14581, 14580, 14587, 14585, 14584, 14591, 14590, 14588, 14595, 14594, 14592, 14589, 14588, 14576, 14593, 14592, 14578, 14571, 14577, 14576, 14591, 14583, 14582, 14575, 14579, 14578, 14595, 14587, 14586, 14572, 14574, 14571, 14586, 14584, 14581, 14594, 14591, 14589, 14574, 14578, 14577, 14594, 14586, 14583, 14592, 14589, 14577, 14597, 14599, 14598, 14603, 14601, 14600, 14609, 14608, 14606, 14611, 14613, 14612, 14617, 14615, 14614, 14623, 14622, 14620, 14599, 14611, 14610, 14615, 14603, 14602, 14621, 14620, 14608, 14604, 14605, 14599, 14618, 14619, 14613, 14609, 14607, 14601, 14623, 14621, 14615, 14606, 14608, 14605, 14620, 14622, 14619, 14608, 14620, 14618, 14605, 14618, 14611, 14621, 14609, 14603, 14559, 14597, 14596, 14601, 14561, 14560, 14607, 14606, 14566, 14566, 14606, 14604, 14563, 14604, 14597, 14607, 14567, 14561, 14568, 14570, 14552, 14582, 14580, 14560, 14588, 14590, 14567, 14570, 14576, 14555,
14590, 14582, 14561, 14588, 14565, 14555, 14625, 14627, 14626, 14629, 14631, 14630, 14639, 14637, 14636, 14643, 14641, 14640, 14647, 14646, 14644, 14649, 14651, 14650, 14645, 14644, 14632, 14649, 14648, 14634, 14627, 14633, 14632, 14647, 14639, 14638, 14631, 14635, 14634, 14651, 14643, 14642, 14628, 14630, 14627, 14642, 14640, 14637, 14650, 14647, 14645, 14630, 14634, 14633, 14650, 14642, 14639, 14648, 14645, 14633, 14653, 14655, 14654, 14661, 14659, 14658, 14665, 14664, 14662, 14659, 14655, 14653, 14663, 14662, 14656, 14655, 14657, 14656, 14665, 14663, 14659, 14663, 14657, 14655, 14667, 14669, 14668, 14673, 14675, 14674, 14669, 14673, 14672, 14667, 14670, 14671, 14673, 14676, 14677, 14669, 14671, 14676, 14681, 14679, 14678, 14685, 14684, 14682, 14689, 14687, 14686, 14691, 14693, 14692, 14687, 14681, 14680, 14685, 14691, 14690, 14681, 14685, 14683, 14689, 14693, 14691, 14687, 14691, 14685, 14643, 14667, 14666, 14643, 14651, 14670, 14679, 14661, 14660, 14683, 14682, 14664, 14679, 14683, 14665, 14652, 14654, 14631, 14662, 14664, 14651, 14654, 14656, 14635, 14662, 14649, 14635, 14682, 14684, 14671, 14692, 14677, 14676, 14690, 14676, 14671, 14664, 14682, 14670, 14698, 14702, 14700, 14703, 14699, 14697, 14706, 14708, 14707, 14714, 14712, 14711,
14710, 14713, 14711, 14716, 14718, 14717, 14702, 14710, 14709, 14716, 14715, 14707, 14698, 14705, 14707, 14703, 14708, 14706, 14718, 14714, 14713, 14715, 14717, 14713, 14698, 14696, 14694, 14705, 14698, 14695, 14707, 14715, 14710, 14723, 14725, 14724, 14727, 14725, 14723, 14729, 14731, 14730, 14733, 14735, 14734, 14724, 14720, 14719, 14724, 14725, 14733, 14735, 14733, 14725, 14731, 14737, 14736, 14728, 14730, 14727, 14730, 14736, 14735, 14732, 14721, 14720, 14746, 14744, 14743, 14742, 14745, 14743, 14750, 14748, 14747, 14752, 14754, 14753, 14756, 14758, 14757, 14754, 14746, 14745, 14742, 14751, 14753, 14758, 14750, 14749, 14749, 14747, 14744, 14755, 14757, 14754, 14757, 14749, 14746, 14714, 14739, 14738, 14718, 14740, 14739, 14752, 14751, 14734, 14737, 14756, 14755, 14736, 14755, 14752, 14719, 14720, 14703, 14734, 14718, 14716, 14720, 14721, 14708, 14721, 14732, 14716, 14742, 14741, 14738, 14751, 14742, 14739, 14734, 14751, 14740, 14762, 14764, 14763, 14760, 14764, 14762, 14771, 14773, 14772, 14769, 14773, 14771, 14782, 14784, 14783, 14763, 14775, 14774, 14782, 14781, 14766, 14773, 14780, 14779, 14769, 14778, 14780, 14763, 14764, 14767, 14760, 14765, 14767, 14774, 14775, 14769, 14783, 14780, 14778, 14775, 14781, 14778, 14766, 14781, 14775,
14790, 14788, 14787, 14790, 14786, 14785, 14794, 14793, 14791, 14798, 14796, 14795, 14796, 14789, 14787, 14793, 14794, 14790, 14794, 14792, 14786, 14798, 14800, 14799, 14799, 14793, 14789, 14804, 14802, 14801, 14808, 14806, 14805, 14810, 14807, 14805, 14814, 14813, 14811, 14813, 14814, 14804, 14816, 14808, 14807, 14810, 14817, 14815, 14810, 14809, 14802, 14817, 14814, 14812, 14817, 14810, 14804, 14808, 14777, 14776, 14812, 14811, 14779, 14816, 14815, 14783, 14783, 14815, 14812, 14816, 14784, 14777, 14791, 14793, 14767, 14798, 14797, 14776, 14800, 14784, 14782, 14784, 14800, 14798, 14793, 14799, 14782, 14613, 14695, 14694, 14699, 14617, 14616, 14706, 14705, 14622, 14622, 14705, 14704, 14613, 14619, 14704, 14706, 14623, 14617, 14675, 14760, 14759, 14675, 14677, 14765, 14786, 14689, 14688, 14792, 14791, 14692, 14786, 14792, 14693, 14791, 14765, 14677, 14624, 14626, 14575, 14638, 14636, 14585, 14646, 14595, 14593, 14626, 14632, 14579, 14646, 14638, 14587, 14644, 14593, 14579, 14770, 14772, 14731, 14731, 14772, 14779, 14803, 14801, 14748, 14811, 14813, 14758, 14758, 14813, 14803, 14811, 14756, 14737, 14819, 14821, 14820, 14825, 14827, 14826, 14831, 14830, 14828, 14820, 14825, 14824, 14829, 14828, 14822, 14821, 14823, 14822, 14828, 14830, 14827,
14822, 14828, 14825, 14833, 14835, 14834, 14839, 14838, 14836, 14841, 14843, 14842, 14847, 14846, 14844, 14835, 14841, 14840, 14845, 14844, 14838, 14836, 14838, 14835, 14844, 14846, 14843, 14838, 14844, 14841, 14827, 14833, 14832, 14837, 14836, 14830, 14830, 14836, 14833, 14849, 14851, 14850, 14853, 14855, 14854, 14863, 14862, 14860, 14867, 14866, 14864, 14861, 14860, 14856, 14865, 14864, 14858, 14851, 14857, 14856, 14855, 14859, 14858, 14852, 14854, 14851, 14866, 14863, 14861, 14854, 14858, 14857, 14864, 14861, 14857, 14869, 14871, 14870, 14873, 14875, 14874, 14883, 14882, 14880, 14887, 14886, 14884, 14881, 14880, 14876, 14885, 14884, 14878, 14871, 14877, 14876, 14875, 14879, 14878, 14872, 14874, 14871, 14886, 14883, 14881, 14874, 14878, 14877, 14884, 14881, 14877, 14891, 14890, 14888, 14895, 14894, 14892, 14899, 14898, 14896, 14903, 14902, 14900, 14897, 14896, 14890, 14901, 14900, 14894, 14894, 14891, 14889, 14902, 14899, 14897, 14900, 14897, 14891, 14907, 14906, 14904, 14911, 14910, 14908, 14915, 14913, 14912, 14919, 14917, 14916, 14921, 14918, 14916, 14925, 14924, 14922, 14923, 14922, 14906, 14927, 14926, 14910, 14925, 14915, 14914, 14927, 14919, 14918, 14921, 14928, 14926, 14910, 14907, 14905, 14921, 14920, 14913, 14928, 14925, 14923,
14928, 14921, 14915, 14910, 14926, 14923, 14889, 14888, 14862, 14893, 14892, 14866, 14892, 14889, 14863, 14905, 14904, 14882, 14909, 14908, 14886, 14908, 14905, 14883, 14868, 14870, 14855, 14882, 14867, 14865, 14870, 14876, 14859, 14880, 14865, 14859, 14906, 14895, 14893, 14922, 14924, 14903, 14922, 14901, 14895, 14904, 14893, 14867, 14848, 14850, 14821, 14862, 14831, 14829, 14850, 14856, 14823, 14860, 14829, 14823, 14890, 14839, 14837, 14898, 14847, 14845, 14896, 14845, 14839, 14888, 14837, 14831, 14276, 14483, 14481, 14286, 14492, 14491, 14495, 14494, 14288, 14494, 14492, 14286, 14285, 14491, 14483, 14515, 14291, 14290, 14515, 14518, 14292, 14292, 14518, 14495, 14930, 14932, 14931, 14934, 14936, 14935, 14942, 14941, 14939, 14937, 14938, 14932, 14936, 14940, 14939, 14933, 14935, 14930, 14939, 14941, 14938, 14935, 14939, 14937, 14944, 14946, 14945, 14948, 14950, 14949, 14954, 14953, 14951, 14958, 14957, 14955, 14946, 14952, 14951, 14950, 14956, 14955, 14947, 14949, 14946, 14957, 14954, 14952, 14949, 14955, 14952, 14943, 14945, 14936, 14953, 14942, 14940, 14945, 14951, 14940, 14960, 14962, 14961, 14964, 14966, 14965, 14970, 14969, 14967, 14974, 14973, 14971, 14962, 14968, 14967, 14966, 14972, 14971, 14963, 14965, 14962, 14973, 14970, 14968,
14965, 14971, 14968, 14976, 14978, 14977, 14980, 14978, 14976, 14984, 14983, 14981, 14978, 14982, 14981, 14984, 14982, 14978, 14975, 14977, 14966, 14983, 14974, 14972, 14977, 14981, 14972, 14986, 14988, 14987, 14992, 14990, 14989, 14998, 14997, 14995, 15000, 15002, 15001, 15006, 15004, 15003, 15010, 15012, 15011, 14988, 15000, 14999, 15004, 14992, 14991, 15010, 15009, 14997, 14993, 14994, 14988, 15007, 15008, 15002, 14998, 14996, 14990, 15012, 15010, 15004, 14995, 14997, 14994, 15011, 15008, 15007, 14997, 15009, 15007, 14994, 15007, 15000, 15010, 14998, 14992, 15016, 15014, 15013, 15020, 15018, 15017, 15022, 15016, 15015, 15024, 15020, 15019, 15019, 15017, 15014, 15023, 15019, 15016, 15026, 15028, 15027, 15034, 15030, 15029, 15040, 15039, 15037, 15043, 15046, 15045, 15050, 15049, 15047, 15033, 15043, 15041, 15044, 15034, 15032, 15040, 15048, 15047, 15026, 15035, 15036, 15040, 15038, 15030, 15047, 15049, 15046, 15033, 15031, 15027, 15037, 15039, 15036, 15039, 15033, 15028, 15033, 15039, 15047, 15044, 15048, 15040, 15054, 15052, 15051, 15056, 15055, 15052, 15002, 15026, 15025, 15030, 15006, 15005, 15038, 15037, 15011, 15037, 15035, 15008, 15008, 15035, 15026, 15038, 15012, 15006, 15015, 15013, 14989, 15021, 15015, 14990, 15051, 15052, 15044,
15055, 15056, 15050, 15052, 15055, 15048, 15060, 15058, 15057, 15064, 15062, 15061, 15064, 15066, 15065, 15073, 15075, 15074, 15066, 15073, 15072, 15068, 15060, 15059, 15070, 15064, 15063, 15064, 15070, 15071, 15073, 15076, 15077, 15063, 15061, 15058, 15069, 15063, 15060, 15066, 15071, 15076, 15081, 15079, 15078, 15083, 15085, 15084, 15089, 15087, 15086, 15091, 15093, 15092, 15087, 15081, 15080, 15091, 15090, 15084, 15081, 15085, 15083, 15089, 15093, 15091, 15087, 15091, 15085, 15095, 15097, 15096, 15103, 15105, 15104, 15101, 15105, 15103, 15097, 15101, 15100, 15095, 15098, 15099, 15104, 15105, 15108, 15106, 15108, 15105, 15097, 15099, 15106, 15112, 15110, 15109, 15114, 15116, 15115, 15122, 15120, 15119, 15122, 15118, 15117, 15126, 15125, 15123, 15118, 15112, 15111, 15124, 15123, 15115, 15112, 15116, 15114, 15125, 15126, 15122, 15126, 15124, 15118, 15118, 15124, 15116, 15075, 15095, 15094, 15075, 15077, 15098, 15110, 15089, 15088, 15114, 15113, 15092, 15110, 15114, 15093, 15082, 15084, 15071, 15092, 15077, 15076, 15084, 15090, 15076, 15115, 15099, 15098, 15123, 15125, 15108, 15123, 15106, 15099, 15113, 15098, 15077, 14932, 14986, 14985, 14942, 14996, 14995, 14941, 14995, 14993, 14938, 14993, 14986, 15022, 15021, 14953, 15024, 15023, 14957,
15023, 15022, 14954, 14953, 15021, 14996, 15068, 15067, 14969, 14974, 15070, 15069, 15069, 15068, 14970, 15079, 14980, 14979, 15083, 15082, 14983, 15083, 14984, 14980, 14983, 15082, 15070, 14959, 14961, 14950, 14969, 14958, 14956, 14961, 14967, 14956, 15059, 15057, 15018, 15067, 15059, 15020, 15067, 15024, 14958, 15128, 15130, 15129, 15134, 15133, 15131, 15136, 15138, 15137, 15142, 15141, 15139, 15130, 15136, 15135, 15140, 15139, 15133, 15131, 15133, 15130, 15139, 15141, 15138, 15133, 15139, 15136, 15144, 15146, 15145, 15150, 15149, 15147, 15152, 15154, 15153, 15158, 15157, 15155, 15146, 15152, 15151, 15156, 15155, 15149, 15147, 15149, 15146, 15155, 15157, 15154, 15149, 15155, 15152, 15138, 15144, 15143, 15148, 15147, 15141, 15141, 15147, 15144, 15162, 15161, 15159, 15166, 15165, 15163, 15170, 15169, 15167, 15174, 15173, 15171, 15168, 15167, 15161, 15172, 15171, 15165, 15165, 15162, 15160, 15173, 15170, 15168, 15171, 15168, 15162, 15178, 15176, 15175, 15184, 15182, 15181, 15182, 15178, 15177, 15178, 15180, 15179, 15184, 15186, 15185, 15182, 15185, 15180, 15190, 15189, 15187, 15194, 15193, 15191, 15198, 15197, 15195, 15202, 15201, 15199, 15196, 15195, 15189, 15200, 15199, 15193, 15193, 15190, 15188, 15201, 15198, 15196, 15199, 15196, 15190,
15206, 15204, 15203, 15212, 15210, 15209, 15210, 15206, 15205, 15208, 15207, 15204, 15214, 15213, 15210, 15213, 15208, 15206, 15188, 15187, 15169, 15192, 15191, 15173, 15191, 15188, 15170, 15204, 15184, 15183, 15204, 15207, 15186, 15180, 15166, 15164, 15185, 15186, 15174, 15185, 15172, 15166, 15208, 15194, 15192, 15214, 15202, 15200, 15213, 15200, 15194, 15186, 15207, 15192, 15216, 15218, 15217, 15222, 15221, 15219, 15224, 15226, 15225, 15230, 15229, 15227, 15218, 15224, 15223, 15228, 15227, 15221, 15219, 15221, 15218, 15227, 15229, 15226, 15221, 15227, 15224, 15232, 15234, 15233, 15238, 15237, 15235, 15240, 15242, 15241, 15246, 15245, 15243, 15234, 15240, 15239, 15244, 15243, 15237, 15235, 15237, 15234, 15243, 15245, 15242, 15237, 15243, 15240, 15226, 15232, 15231, 15236, 15235, 15229, 15229, 15235, 15232, 15250, 15249, 15247, 15254, 15253, 15251, 15258, 15257, 15255, 15262, 15261, 15259, 15256, 15255, 15249, 15260, 15259, 15253, 15253, 15250, 15248, 15261, 15258, 15256, 15259, 15256, 15250, 15266, 15264, 15263, 15272, 15270, 15269, 15270, 15266, 15265, 15268, 15267, 15264, 15274, 15273, 15270, 15273, 15268, 15266, 15278, 15277, 15275, 15282, 15281, 15279, 15286, 15285, 15283, 15290, 15289, 15287, 15284, 15283, 15277, 15288, 15287, 15281,
15281, 15278, 15276, 15289, 15286, 15284, 15287, 15284, 15278, 15294, 15292, 15291, 15300, 15298, 15297, 15298, 15294, 15293, 15296, 15295, 15292, 15302, 15301, 15298, 15301, 15296, 15294, 15276, 15275, 15257, 15280, 15279, 15261, 15279, 15276, 15258, 15292, 15272, 15271, 15295, 15274, 15272, 15268, 15254, 15252, 15274, 15262, 15260, 15273, 15260, 15254, 15296, 15282, 15280, 15302, 15290, 15288, 15301, 15288, 15282, 15295, 15280, 15262, 15154, 15216, 15215, 15220, 15219, 15157, 15157, 15219, 15216, 15248, 15247, 15197, 15252, 15251, 15201, 15251, 15248, 15198, 15264, 15212, 15211, 15267, 15214, 15212, 15267, 15252, 15202, 15161, 15134, 15132, 15169, 15142, 15140, 15167, 15140, 15134, 15189, 15150, 15148, 15197, 15158, 15156, 15195, 15156, 15150, 15187, 15148, 15142, 15249, 15222, 15220, 15257, 15230, 15228, 15255, 15228, 15222, 15277, 15238, 15236, 15285, 15246, 15244, 15283, 15244, 15238, 15275, 15236, 15230, 15247, 15220, 15158, 15306, 15304, 15303, 15306, 15308, 15307, 15310, 15306, 15305, 15306, 15310, 15311, 15313, 15315, 15314, 15321, 15323, 15322, 15322, 15319, 15318, 15327, 15326, 15324, 15334, 15332, 15331, 15338, 15336, 15335, 15317, 15333, 15331, 15327, 15340, 15339, 15330, 15342, 15341, 15315, 15325, 15324, 15340, 15334, 15333,
15322, 15323, 15330, 15329, 15328, 15319, 15342, 15338, 15337, 15318, 15319, 15315, 15328, 15329, 15327, 15337, 15335, 15332, 15315, 15319, 15328, 15341, 15337, 15334, 15341, 15340, 15327, 15326, 15339, 15333, 15317, 15316, 15307, 15326, 15311, 15310, 15311, 15326, 15317, 15344, 15346, 15345, 15345, 15348, 15347, 15354, 15352, 15351, 15358, 15357, 15355, 15360, 15362, 15361, 15366, 15364, 15363, 15368, 15353, 15351, 15358, 15370, 15369, 15345, 15346, 15356, 15345, 15355, 15357, 15369, 15370, 15366, 15361, 15362, 15354, 15349, 15350, 15346, 15359, 15361, 15358, 15368, 15367, 15364, 15346, 15350, 15359, 15371, 15368, 15366, 15371, 15370, 15358, 15368, 15371, 15361, 15375, 15373, 15372, 15375, 15377, 15376, 15375, 15374, 15352, 15377, 15362, 15360, 15362, 15377, 15375, 15381, 15383, 15382, 15382, 15383, 15386, 15380, 15382, 15379, 15382, 15385, 15384, 15390, 15392, 15391, 15388, 15392, 15390, 15397, 15400, 15398, 15403, 15405, 15404, 15391, 15397, 15396, 15403, 15402, 15394, 15392, 15395, 15394, 15388, 15393, 15395, 15397, 15402, 15404, 15391, 15394, 15402, 15407, 15409, 15408, 15415, 15413, 15412, 15417, 15419, 15418, 15423, 15422, 15420, 15427, 15425, 15424, 15429, 15426, 15424, 15427, 15411, 15410, 15431, 15430, 15418, 15408, 15409, 15417,
15431, 15427, 15426, 15429, 15432, 15430, 15422, 15423, 15415, 15414, 15412, 15410, 15420, 15422, 15419, 15419, 15422, 15414, 15427, 15431, 15419, 15379, 15409, 15407, 15421, 15420, 15385, 15420, 15417, 15384, 15379, 15384, 15417, 15406, 15408, 15388, 15418, 15395, 15393, 15429, 15428, 15399, 15432, 15405, 15403, 15388, 15408, 15416, 15405, 15432, 15429, 15418, 15430, 15403, 15434, 15436, 15435, 15436, 15440, 15439, 15437, 15438, 15436, 15438, 15441, 15440, 15445, 15443, 15442, 15447, 15449, 15448, 15451, 15449, 15447, 15453, 15455, 15454, 15457, 15459, 15458, 15467, 15465, 15464, 15463, 15466, 15464, 15470, 15472, 15471, 15445, 15461, 15460, 15470, 15469, 15458, 15454, 15455, 15445, 15448, 15449, 15457, 15451, 15459, 15457, 15472, 15467, 15466, 15463, 15469, 15471, 15458, 15455, 15453, 15463, 15462, 15460, 15469, 15463, 15461, 15458, 15469, 15468, 15455, 15468, 15461, 15479, 15481, 15480, 15480, 15477, 15476, 15477, 15474, 15473, 15481, 15484, 15483, 15483, 15482, 15477, 15482, 15475, 15474, 15448, 15438, 15437, 15440, 15453, 15452, 15441, 15456, 15453, 15456, 15441, 15438, 15473, 15474, 15451, 15483, 15472, 15470, 15474, 15475, 15459, 15475, 15482, 15470, 15347, 15348, 15323, 15365, 15363, 15336, 15348, 15357, 15330, 15342, 15369, 15365,
15369, 15342, 15330, 15433, 15435, 15383, 15435, 15439, 15386, 15444, 15442, 15413, 15452, 15454, 15423, 15454, 15444, 15415, 15439, 15452, 15421, 15486, 15488, 15487, 15492, 15491, 15489, 15494, 15496, 15495, 15500, 15499, 15497, 15488, 15494, 15493, 15498, 15497, 15491, 15489, 15491, 15488, 15497, 15499, 15496, 15491, 15497, 15494, 15506, 15504, 15503, 15502, 15505, 15503, 15510, 15509, 15507, 15510, 15506, 15505, 15507, 15509, 15505, 15496, 15502, 15501, 15508, 15507, 15499, 15499, 15507, 15502, 15514, 15513, 15511, 15518, 15517, 15515, 15522, 15521, 15519, 15526, 15525, 15523, 15520, 15519, 15513, 15524, 15523, 15517, 15517, 15514, 15512, 15525, 15522, 15520, 15523, 15520, 15514, 15530, 15528, 15527, 15536, 15534, 15533, 15534, 15530, 15529, 15532, 15531, 15528, 15538, 15537, 15534, 15537, 15532, 15530, 15542, 15540, 15539, 15546, 15544, 15543, 15550, 15549, 15547, 15554, 15553, 15551, 15550, 15542, 15541, 15554, 15546, 15545, 15545, 15543, 15540, 15553, 15550, 15548, 15553, 15545, 15542, 15558, 15556, 15555, 15560, 15559, 15556, 15548, 15547, 15521, 15552, 15551, 15525, 15551, 15548, 15522, 15556, 15536, 15535, 15559, 15538, 15536, 15532, 15518, 15516, 15538, 15526, 15524, 15537, 15524, 15518, 15558, 15557, 15544, 15560, 15554, 15552,
15560, 15558, 15546, 15559, 15552, 15526, 15562, 15564, 15563, 15564, 15566, 15565, 15568, 15570, 15569, 15574, 15573, 15571, 15576, 15578, 15577, 15582, 15581, 15579, 15570, 15576, 15575, 15580, 15579, 15573, 15571, 15573, 15570, 15579, 15581, 15578, 15573, 15579, 15576, 15563, 15568, 15567, 15572, 15571, 15565, 15565, 15571, 15568, 15584, 15586, 15585, 15588, 15590, 15589, 15586, 15592, 15591, 15590, 15594, 15593, 15587, 15589, 15586, 15589, 15593, 15592, 15596, 15598, 15597, 15600, 15602, 15601, 15598, 15604, 15603, 15602, 15606, 15605, 15599, 15601, 15598, 15601, 15605, 15604, 15610, 15609, 15607, 15614, 15613, 15611, 15618, 15617, 15615, 15622, 15621, 15619, 15616, 15615, 15609, 15620, 15619, 15613, 15613, 15610, 15608, 15621, 15618, 15616, 15619, 15616, 15610, 15626, 15625, 15623, 15630, 15629, 15627, 15634, 15633, 15631, 15638, 15637, 15635, 15632, 15631, 15625, 15636, 15635, 15629, 15629, 15626, 15624, 15637, 15634, 15632, 15635, 15632, 15626, 15608, 15607, 15591, 15612, 15611, 15593, 15611, 15608, 15592, 15624, 15623, 15603, 15628, 15627, 15605, 15627, 15624, 15604, 15595, 15597, 15590, 15597, 15603, 15594, 15625, 15614, 15612, 15633, 15622, 15620, 15631, 15620, 15614, 15623, 15612, 15594, 15513, 15492, 15490, 15521, 15500, 15498,
15519, 15498, 15492, 15541, 15539, 15504, 15549, 15510, 15508, 15549, 15541, 15506, 15547, 15508, 15500, 15583, 15585, 15564, 15585, 15591, 15566, 15609, 15574, 15572, 15617, 15582, 15580, 15615, 15580, 15574, 15607, 15572, 15566, 15046, 15305, 15303, 15050, 15310, 15309, 15049, 15309, 15305, 15314, 15054, 15053, 15324, 15056, 15054, 15324, 15310, 15050, 15104, 15350, 15349, 15360, 15359, 15107, 15107, 15359, 15350, 15373, 15121, 15119, 15373, 15376, 15125, 15125, 15376, 15360, 15242, 15486, 15485, 15490, 15489, 15245, 15245, 15489, 15486, 15512, 15511, 15285, 15516, 15515, 15289, 15515, 15512, 15286, 15528, 15300, 15299, 15531, 15302, 15300, 15531, 15516, 15290, 15511, 15490, 15246, 15640, 15642, 15641, 15648, 15646, 15644, 15652, 15654, 15653, 15656, 15658, 15657, 15662, 15660, 15659, 15668, 15667, 15665, 15642, 15656, 15655, 15660, 15648, 15647, 15666, 15665, 15653, 15649, 15650, 15642, 15663, 15664, 15658, 15648, 15654, 15652, 15668, 15666, 15660, 15643, 15645, 15640, 15653, 15650, 15649, 15667, 15664, 15663, 15640, 15645, 15651, 15665, 15663, 15650, 15642, 15650, 15663, 15660, 15666, 15654, 15670, 15672, 15671, 15678, 15674, 15673, 15684, 15683, 15681, 15690, 15688, 15687, 15686, 15689, 15687, 15692, 15694, 15693, 15677, 15686, 15685,
15692, 15691, 15683, 15670, 15679, 15680, 15678, 15684, 15682, 15694, 15690, 15689, 15691, 15693, 15689, 15677, 15675, 15671, 15681, 15683, 15680, 15683, 15677, 15672, 15683, 15691, 15686, 15701, 15703, 15702, 15702, 15699, 15698, 15699, 15696, 15695, 15703, 15708, 15707, 15707, 15706, 15699, 15704, 15705, 15703, 15705, 15709, 15708, 15706, 15697, 15696, 15658, 15670, 15669, 15674, 15662, 15661, 15682, 15681, 15667, 15667, 15681, 15679, 15664, 15679, 15670, 15682, 15668, 15662, 15695, 15696, 15678, 15707, 15694, 15692, 15696, 15697, 15684, 15697, 15706, 15692, 15711, 15713, 15712, 15717, 15719, 15718, 15713, 15717, 15716, 15714, 15715, 15713, 15717, 15720, 15721, 15715, 15720, 15717, 15725, 15723, 15722, 15727, 15729, 15728, 15733, 15731, 15730, 15735, 15737, 15736, 15731, 15725, 15724, 15735, 15734, 15728, 15729, 15727, 15723, 15733, 15737, 15735, 15735, 15729, 15725, 15739, 15741, 15740, 15747, 15749, 15748, 15748, 15751, 15750, 15756, 15755, 15753, 15741, 15749, 15747, 15739, 15742, 15743, 15748, 15749, 15754, 15753, 15755, 15751, 15750, 15751, 15745, 15751, 15755, 15752, 15743, 15754, 15749, 15760, 15758, 15757, 15764, 15763, 15761, 15768, 15766, 15765, 15770, 15767, 15765, 15768, 15760, 15759, 15772, 15771, 15763, 15760, 15764, 15762,
15771, 15772, 15768, 15770, 15773, 15771, 15772, 15764, 15760, 15719, 15739, 15738, 15719, 15721, 15742, 15758, 15733, 15732, 15737, 15762, 15761, 15758, 15762, 15737, 15726, 15728, 15715, 15736, 15721, 15720, 15728, 15734, 15720, 15763, 15743, 15742, 15771, 15773, 15756, 15763, 15771, 15754, 15761, 15742, 15721, 15782, 15780, 15779, 15778, 15781, 15779, 15784, 15786, 15785, 15790, 15788, 15787, 15792, 15794, 15793, 15796, 15798, 15797, 15793, 15794, 15782, 15791, 15793, 15781, 15785, 15786, 15796, 15798, 15790, 15789, 15789, 15787, 15780, 15797, 15794, 15792, 15794, 15797, 15789, 15778, 15777, 15774, 15776, 15791, 15778, 15800, 15802, 15801, 15806, 15804, 15803, 15812, 15810, 15809, 15812, 15808, 15807, 15816, 15815, 15813, 15820, 15819, 15817, 15802, 15814, 15813, 15815, 15816, 15806, 15819, 15820, 15812, 15820, 15818, 15808, 15811, 15809, 15804, 15817, 15819, 15816, 15816, 15819, 15811, 15822, 15821, 15807, 15818, 15823, 15822, 15690, 15775, 15774, 15690, 15694, 15776, 15785, 15705, 15704, 15792, 15791, 15707, 15709, 15795, 15792, 15795, 15709, 15705, 15791, 15776, 15694, 15745, 15802, 15800, 15756, 15818, 15817, 15817, 15814, 15752, 15745, 15752, 15814, 15822, 15770, 15769, 15822, 15823, 15773, 15823, 15818, 15756, 15799, 15801, 15786,
15805, 15803, 15788, 15815, 15798, 15796, 15786, 15801, 15813, 15798, 15815, 15805, 15827, 15825, 15824, 15829, 15827, 15826, 15833, 15831, 15830, 15837, 15835, 15834, 15839, 15833, 15832, 15841, 15837, 15836, 15836, 15834, 15831, 15840, 15836, 15833, 15845, 15843, 15842, 15849, 15847, 15846, 15851, 15845, 15844, 15853, 15849, 15848, 15848, 15846, 15843, 15852, 15848, 15845, 15844, 15842, 15835, 15850, 15844, 15837, 15855, 15857, 15856, 15856, 15859, 15858, 15863, 15862, 15860, 15865, 15867, 15866, 15871, 15870, 15868, 15859, 15865, 15864, 15869, 15868, 15862, 15857, 15861, 15860, 15860, 15862, 15859, 15868, 15870, 15867, 15862, 15868, 15865, 15877, 15875, 15874, 15873, 15876, 15874, 15881, 15880, 15878, 15881, 15877, 15876, 15878, 15880, 15876, 15867, 15873, 15872, 15879, 15878, 15870, 15870, 15878, 15873, 15883, 15885, 15884, 15887, 15889, 15888, 15893, 15892, 15890, 15897, 15896, 15894, 15901, 15900, 15898, 15905, 15904, 15902, 15899, 15898, 15892, 15903, 15902, 15896, 15885, 15891, 15890, 15889, 15895, 15894, 15886, 15888, 15885, 15896, 15893, 15891, 15904, 15901, 15899, 15888, 15894, 15891, 15902, 15899, 15893, 15907, 15909, 15908, 15911, 15913, 15912, 15917, 15916, 15914, 15921, 15920, 15918, 15925, 15924, 15922, 15929, 15928, 15926,
15923, 15922, 15916, 15927, 15926, 15920, 15909, 15915, 15914, 15913, 15919, 15918, 15910, 15912, 15909, 15920, 15917, 15915, 15928, 15925, 15923, 15912, 15918, 15915, 15926, 15923, 15917, 15933, 15931, 15930, 15937, 15935, 15934, 15941, 15940, 15938, 15945, 15944, 15942, 15941, 15933, 15932, 15945, 15937, 15936, 15936, 15934, 15931, 15944, 15941, 15939, 15944, 15936, 15933, 15949, 15947, 15946, 15953, 15951, 15950, 15957, 15956, 15954, 15961, 15960, 15958, 15957, 15949, 15948, 15961, 15953, 15952, 15952, 15950, 15947, 15960, 15957, 15955, 15960, 15952, 15949, 15939, 15938, 15900, 15943, 15942, 15904, 15942, 15939, 15901, 15955, 15954, 15924, 15959, 15958, 15928, 15958, 15955, 15925, 15906, 15908, 15889, 15916, 15897, 15895, 15924, 15905, 15903, 15908, 15914, 15895, 15922, 15903, 15897, 15948, 15946, 15935, 15956, 15945, 15943, 15956, 15948, 15937, 15954, 15943, 15905, 15832, 15830, 15825, 15838, 15832, 15827, 15882, 15884, 15857, 15892, 15863, 15861, 15900, 15871, 15869, 15884, 15890, 15861, 15898, 15869, 15863, 15932, 15930, 15875, 15940, 15881, 15879, 15940, 15932, 15877, 15938, 15879, 15871, 15963, 15965, 15964, 15967, 15969, 15968, 15975, 15974, 15972, 15970, 15971, 15965, 15969, 15973, 15972, 15966, 15968, 15963, 15974, 15971, 15970,
15968, 15972, 15970, 15977, 15979, 15978, 15981, 15983, 15982, 15987, 15986, 15984, 15991, 15990, 15988, 15979, 15985, 15984, 15983, 15989, 15988, 15980, 15982, 15979, 15990, 15987, 15985, 15982, 15988, 15985, 15997, 15995, 15994, 16000, 15997, 15996, 15996, 15994, 15992, 15999, 15996, 15993, 16004, 16002, 16001, 16008, 16006, 16005, 16016, 16018, 16017, 16014, 16018, 16016, 16010, 16004, 16003, 16012, 16008, 16007, 16018, 16021, 16020, 16019, 16021, 16018, 16007, 16005, 16002, 16011, 16007, 16004, 15965, 15993, 15992, 16000, 15999, 15974, 15999, 15998, 15971, 15971, 15998, 15993, 16010, 16009, 15986, 16012, 16011, 15990, 16011, 16010, 15987, 15976, 15978, 15969, 15986, 15975, 15973, 15978, 15984, 15973, 16003, 16001, 15995, 16009, 16003, 15997, 16009, 16000, 15975, 16023, 16025, 16024, 16027, 16029, 16028, 16031, 16029, 16027, 16033, 16035, 16034, 16037, 16039, 16038, 16025, 16033, 16032, 16028, 16029, 16037, 16039, 16037, 16029, 16026, 16028, 16025, 16038, 16035, 16033, 16028, 16036, 16033, 16046, 16044, 16043, 16048, 16050, 16049, 16046, 16050, 16048, 16058, 16060, 16059, 16068, 16066, 16065, 16070, 16067, 16065, 16072, 16074, 16073, 16049, 16062, 16061, 16068, 16052, 16051, 16072, 16071, 16055, 16076, 16075, 16059, 16054, 16046, 16045,
16049, 16050, 16056, 16054, 16056, 16050, 16075, 16076, 16068, 16077, 16075, 16067, 16059, 16056, 16054, 16070, 16069, 16063, 16077, 16074, 16072, 16074, 16077, 16070, 16059, 16075, 16072, 16055, 16071, 16062, 16076, 16060, 16052, 16081, 16079, 16078, 16083, 16082, 16079, 16035, 16054, 16053, 16058, 16057, 16038, 16057, 16054, 16035, 16079, 16041, 16040, 16079, 16082, 16042, 16040, 16041, 16031, 16031, 16041, 16042, 16081, 16080, 16051, 16083, 16060, 16058, 16060, 16083, 16081, 16042, 16082, 16058, 16087, 16089, 16088, 16085, 16089, 16087, 16089, 16092, 16091, 16085, 16090, 16092, 16096, 16098, 16097, 16094, 16098, 16096, 16107, 16109, 16108, 16117, 16115, 16114, 16119, 16116, 16114, 16121, 16123, 16122, 16097, 16111, 16110, 16117, 16102, 16101, 16121, 16120, 16104, 16125, 16124, 16108, 16097, 16098, 16105, 16103, 16105, 16098, 16124, 16125, 16117, 16126, 16124, 16116, 16099, 16100, 16094, 16108, 16105, 16103, 16119, 16118, 16112, 16126, 16123, 16121, 16094, 16100, 16106, 16123, 16126, 16119, 16108, 16124, 16121, 16104, 16120, 16111, 16125, 16109, 16102, 16130, 16132, 16131, 16137, 16136, 16134, 16143, 16141, 16140, 16128, 16139, 16138, 16137, 16146, 16145, 16130, 16134, 16136, 16146, 16143, 16142, 16131, 16132, 16128, 16142, 16140, 16138,
16128, 16132, 16136, 16145, 16142, 16139, 16145, 16144, 16133, 16133, 16144, 16139, 16150, 16152, 16151, 16156, 16155, 16153, 16160, 16159, 16157, 16164, 16162, 16161, 16168, 16166, 16165, 16170, 16169, 16155, 16172, 16171, 16159, 16170, 16164, 16163, 16152, 16158, 16157, 16172, 16168, 16167, 16149, 16151, 16148, 16159, 16156, 16154, 16167, 16165, 16162, 16148, 16151, 16157, 16171, 16167, 16164, 16171, 16170, 16156, 16088, 16130, 16129, 16135, 16134, 16091, 16088, 16091, 16134, 16148, 16113, 16112, 16154, 16153, 16122, 16154, 16123, 16113, 16110, 16111, 16085, 16122, 16092, 16090, 16085, 16111, 16120, 16155, 16137, 16135, 16163, 16161, 16141, 16169, 16163, 16143, 16169, 16146, 16137, 16122, 16153, 16135, 16176, 16174, 16173, 16178, 16175, 16173, 16179, 16180, 16176, 16181, 16179, 16175, 16183, 16185, 16184, 16187, 16189, 16188, 16193, 16192, 16190, 16197, 16196, 16194, 16201, 16199, 16198, 16205, 16203, 16202, 16207, 16206, 16192, 16209, 16208, 16196, 16185, 16191, 16190, 16207, 16201, 16200, 16189, 16195, 16194, 16209, 16205, 16204, 16186, 16188, 16185, 16196, 16193, 16191, 16204, 16202, 16199, 16188, 16194, 16191, 16208, 16204, 16201, 16208, 16207, 16193, 16211, 16213, 16212, 16215, 16213, 16211, 16219, 16218, 16216, 16223, 16221, 16220,
16223, 16215, 16214, 16225, 16224, 16218, 16213, 16217, 16216, 16219, 16217, 16213, 16225, 16223, 16222, 16225, 16219, 16215, 16210, 16212, 16189, 16218, 16197, 16195, 16222, 16220, 16203, 16212, 16216, 16195, 16224, 16222, 16205, 16224, 16209, 16197, 16017, 16100, 16099, 16107, 16106, 16020, 16020, 16106, 16100, 16176, 16064, 16063, 16180, 16179, 16073, 16180, 16074, 16064, 16022, 16024, 15983, 16034, 15991, 15989, 16024, 16032, 15989, 16045, 16043, 16006, 16061, 16062, 16014, 16073, 16021, 16019, 16053, 16045, 16008, 16014, 16062, 16071, 16053, 16012, 15991, 16178, 16177, 16101, 16181, 16109, 16107, 16109, 16181, 16178, 16182, 16184, 16152, 16192, 16160, 16158, 16200, 16198, 16166, 16184, 16190, 16158, 16206, 16200, 16168, 16206, 16172, 16160, 16073, 16179, 16107, 15400, 15645, 15643, 15646, 15401, 15399, 15652, 15651, 15404, 15404, 15651, 15645, 15652, 15405, 15401, 15467, 15711, 15710, 15472, 15714, 15711, 15723, 15481, 15479, 15484, 15727, 15726, 15723, 15727, 15484, 15726, 15714, 15472, 15578, 15826, 15824, 15829, 15828, 15581, 15581, 15828, 15826, 15839, 15838, 15617, 15841, 15840, 15621, 15840, 15839, 15618, 15851, 15850, 15633, 15853, 15852, 15637, 15852, 15851, 15634, 15850, 15841, 15622, 15838, 15829, 15582, 14843, 15128, 15127,
15132, 15131, 14846, 14846, 15131, 15128, 15160, 15159, 14898, 15164, 15163, 14902, 15163, 15160, 14899, 15176, 14914, 14912, 15176, 15179, 14924, 15179, 15164, 14903, 15159, 15132, 14847, 16227, 16229, 16228, 16235, 16237, 16236, 16241, 16240, 16238, 16243, 16245, 16244, 16233, 16229, 16227, 16231, 16239, 16238, 16228, 16229, 16231, 16237, 16243, 16242, 16234, 16236, 16233, 16242, 16244, 16241, 16233, 16236, 16242, 16239, 16231, 16229, 16247, 16249, 16248, 16251, 16249, 16247, 16253, 16255, 16254, 16257, 16259, 16258, 16248, 16249, 16257, 16259, 16257, 16249, 16255, 16261, 16260, 16252, 16254, 16251, 16251, 16254, 16260, 16265, 16264, 16262, 16269, 16268, 16266, 16273, 16272, 16270, 16277, 16276, 16274, 16271, 16270, 16264, 16275, 16274, 16268, 16268, 16265, 16263, 16276, 16273, 16271, 16274, 16271, 16265, 16281, 16280, 16278, 16285, 16284, 16282, 16289, 16288, 16286, 16293, 16292, 16290, 16287, 16286, 16280, 16291, 16290, 16284, 16284, 16281, 16279, 16292, 16289, 16287, 16290, 16287, 16281, 16263, 16262, 16240, 16267, 16266, 16244, 16266, 16263, 16241, 16279, 16278, 16258, 16261, 16283, 16282, 16260, 16282, 16279, 16246, 16248, 16237, 16256, 16258, 16245, 16248, 16256, 16243, 16280, 16269, 16267, 16288, 16277, 16275, 16286, 16275, 16269,
16278, 16267, 16245, 16298, 16300, 16299, 16302, 16304, 16303, 16306, 16304, 16302, 16308, 16310, 16309, 16312, 16314, 16313, 16318, 16317, 16315, 16322, 16321, 16319, 16316, 16315, 16309, 16320, 16319, 16313, 16299, 16300, 16308, 16303, 16304, 16312, 16314, 16312, 16304, 16301, 16303, 16300, 16311, 16313, 16310, 16321, 16318, 16316, 16303, 16311, 16308, 16319, 16316, 16310, 16324, 16326, 16325, 16333, 16335, 16334, 16339, 16338, 16336, 16341, 16343, 16342, 16334, 16328, 16327, 16330, 16337, 16336, 16326, 16330, 16329, 16335, 16341, 16340, 16327, 16328, 16326, 16342, 16339, 16337, 16328, 16331, 16330, 16331, 16340, 16337, 16334, 16340, 16331, 16299, 16295, 16294, 16307, 16296, 16295, 16323, 16325, 16306, 16338, 16322, 16320, 16306, 16325, 16329, 16329, 16336, 16320, 16347, 16345, 16344, 16349, 16351, 16350, 16355, 16354, 16352, 16361, 16359, 16358, 16347, 16357, 16356, 16355, 16364, 16363, 16350, 16351, 16347, 16364, 16361, 16360, 16354, 16351, 16349, 16360, 16358, 16356, 16362, 16363, 16360, 16354, 16363, 16362, 16351, 16362, 16357, 16368, 16367, 16365, 16372, 16371, 16369, 16378, 16376, 16375, 16374, 16377, 16375, 16382, 16380, 16379, 16384, 16386, 16385, 16388, 16390, 16389, 16384, 16383, 16367, 16388, 16387, 16371, 16385, 16386, 16378,
16383, 16385, 16377, 16389, 16390, 16382, 16371, 16368, 16366, 16381, 16379, 16376, 16387, 16389, 16386, 16389, 16381, 16378, 16387, 16384, 16368, 16382, 16392, 16391, 16390, 16393, 16392, 16367, 16355, 16353, 16374, 16373, 16359, 16364, 16383, 16374, 16367, 16383, 16364, 16397, 16396, 16394, 16401, 16400, 16398, 16405, 16404, 16402, 16409, 16408, 16406, 16403, 16402, 16396, 16407, 16406, 16400, 16400, 16397, 16395, 16408, 16405, 16403, 16406, 16403, 16397, 16413, 16412, 16410, 16417, 16416, 16414, 16421, 16420, 16418, 16425, 16424, 16422, 16419, 16418, 16412, 16423, 16422, 16416, 16416, 16413, 16411, 16424, 16421, 16419, 16422, 16419, 16413, 16429, 16427, 16426, 16435, 16433, 16432, 16431, 16434, 16432, 16439, 16441, 16440, 16437, 16429, 16428, 16440, 16441, 16435, 16438, 16440, 16434, 16431, 16430, 16427, 16437, 16438, 16431, 16447, 16446, 16444, 16451, 16450, 16448, 16455, 16453, 16452, 16461, 16459, 16458, 16457, 16460, 16458, 16465, 16467, 16466, 16443, 16454, 16452, 16447, 16463, 16462, 16465, 16464, 16450, 16463, 16455, 16454, 16467, 16461, 16460, 16464, 16466, 16460, 16450, 16447, 16445, 16457, 16456, 16453, 16464, 16457, 16455, 16450, 16464, 16463, 16446, 16462, 16454, 16405, 16437, 16436, 16439, 16438, 16408, 16408, 16438, 16437,
16445, 16444, 16420, 16449, 16448, 16424, 16448, 16445, 16421, 16412, 16401, 16399, 16420, 16409, 16407, 16418, 16407, 16401, 16443, 16442, 16433, 16444, 16446, 16441, 16446, 16443, 16435, 16444, 16439, 16409, 16349, 16348, 16272, 16353, 16352, 16276, 16352, 16349, 16273, 16366, 16365, 16288, 16370, 16369, 16292, 16369, 16366, 16289, 16365, 16353, 16277, 16395, 16394, 16317, 16399, 16398, 16321, 16398, 16395, 16318, 16411, 16410, 16338, 16415, 16414, 16342, 16414, 16411, 16339, 16410, 16399, 16322, 16294, 16295, 16255, 16255, 16295, 16296, 16309, 16285, 16283, 16317, 16293, 16291, 16315, 16291, 16285, 16296, 16307, 16283, 16396, 16372, 16370, 16404, 16390, 16388, 16402, 16388, 16372, 16428, 16426, 16391, 16393, 16436, 16428, 16404, 16436, 16393, 16394, 16370, 16293, 16471, 16473, 16472, 16472, 16469, 16468, 16473, 16478, 16477, 16477, 16476, 16469, 16474, 16475, 16473, 16475, 16479, 16478, 16481, 16483, 16482, 16487, 16486, 16484, 16489, 16491, 16490, 16495, 16494, 16492, 16499, 16498, 16496, 16493, 16492, 16486, 16497, 16496, 16490, 16483, 16489, 16488, 16490, 16487, 16485, 16498, 16495, 16493, 16496, 16493, 16487, 16503, 16505, 16504, 16504, 16501, 16500, 16512, 16514, 16513, 16516, 16514, 16512, 16520, 16519, 16517, 16522, 16524, 16523,
16510, 16505, 16503, 16508, 16518, 16517, 16504, 16505, 16508, 16507, 16506, 16501, 16514, 16522, 16521, 16524, 16522, 16514, 16511, 16513, 16510, 16521, 16523, 16520, 16513, 16521, 16518, 16518, 16508, 16505, 16528, 16527, 16525, 16532, 16531, 16529, 16540, 16538, 16537, 16540, 16536, 16535, 16544, 16543, 16541, 16546, 16548, 16547, 16542, 16541, 16527, 16532, 16546, 16545, 16547, 16548, 16540, 16540, 16548, 16546, 16531, 16528, 16526, 16545, 16547, 16544, 16545, 16542, 16528, 16552, 16551, 16549, 16556, 16555, 16553, 16560, 16558, 16557, 16562, 16564, 16563, 16560, 16564, 16562, 16568, 16566, 16565, 16570, 16572, 16571, 16571, 16567, 16565, 16552, 16574, 16573, 16578, 16577, 16555, 16573, 16574, 16560, 16563, 16564, 16576, 16560, 16574, 16576, 16578, 16568, 16567, 16572, 16580, 16579, 16579, 16577, 16567, 16555, 16552, 16550, 16577, 16579, 16576, 16555, 16577, 16574, 16526, 16525, 16494, 16530, 16529, 16498, 16529, 16526, 16495, 16550, 16549, 16519, 16554, 16553, 16523, 16553, 16550, 16520, 16500, 16501, 16483, 16506, 16507, 16491, 16519, 16499, 16497, 16501, 16506, 16489, 16517, 16497, 16491, 16551, 16532, 16530, 16559, 16557, 16535, 16561, 16563, 16540, 16546, 16573, 16559, 16540, 16563, 16575, 16573, 16546, 16532, 16549, 16530, 16499,
16585, 16587, 16586, 16591, 16593, 16592, 16592, 16589, 16588, 16595, 16597, 16596, 16606, 16604, 16603, 16602, 16605, 16603, 16612, 16611, 16609, 16616, 16615, 16613, 16608, 16593, 16591, 16610, 16609, 16596, 16600, 16614, 16613, 16587, 16595, 16594, 16612, 16606, 16605, 16602, 16609, 16611, 16592, 16593, 16600, 16599, 16598, 16589, 16588, 16589, 16587, 16598, 16599, 16597, 16615, 16612, 16610, 16589, 16598, 16595, 16613, 16610, 16597, 16614, 16600, 16593, 16620, 16622, 16621, 16621, 16618, 16617, 16621, 16622, 16625, 16624, 16623, 16618, 16586, 16582, 16581, 16594, 16583, 16582, 16617, 16618, 16608, 16623, 16624, 16616, 16608, 16618, 16623, 16482, 16475, 16474, 16478, 16485, 16484, 16479, 16488, 16485, 16488, 16479, 16475, 16581, 16582, 16516, 16516, 16582, 16583, 16596, 16556, 16554, 16602, 16601, 16566, 16603, 16605, 16572, 16578, 16609, 16602, 16572, 16605, 16611, 16596, 16609, 16578, 16583, 16594, 16554, 16627, 16629, 16628, 16628, 16629, 16631, 16633, 16635, 16634, 16637, 16639, 16638, 16634, 16635, 16641, 16639, 16643, 16642, 16636, 16638, 16635, 16638, 16642, 16641, 16632, 16634, 16629, 16629, 16634, 16640, 16650, 16652, 16651, 16648, 16652, 16650, 16659, 16661, 16660, 16665, 16667, 16666, 16671, 16670, 16668, 16651, 16663, 16662,
16665, 16664, 16656, 16669, 16668, 16660, 16651, 16652, 16657, 16648, 16655, 16657, 16653, 16654, 16648, 16660, 16657, 16655, 16670, 16667, 16665, 16654, 16658, 16655, 16660, 16668, 16665, 16656, 16664, 16663, 16675, 16674, 16672, 16679, 16678, 16676, 16683, 16682, 16680, 16687, 16686, 16684, 16681, 16680, 16674, 16685, 16684, 16678, 16678, 16675, 16673, 16686, 16683, 16681, 16684, 16681, 16675, 16691, 16690, 16688, 16695, 16694, 16692, 16699, 16698, 16696, 16703, 16702, 16700, 16697, 16696, 16690, 16701, 16700, 16694, 16694, 16691, 16689, 16702, 16699, 16697, 16700, 16697, 16691, 16707, 16706, 16704, 16711, 16710, 16708, 16715, 16714, 16712, 16719, 16718, 16716, 16713, 16712, 16706, 16717, 16716, 16710, 16710, 16707, 16705, 16718, 16715, 16713, 16716, 16713, 16707, 16689, 16688, 16666, 16693, 16692, 16670, 16692, 16689, 16667, 16705, 16704, 16682, 16709, 16708, 16686, 16708, 16705, 16683, 16674, 16661, 16659, 16682, 16671, 16669, 16680, 16669, 16661, 16706, 16695, 16693, 16714, 16703, 16701, 16712, 16701, 16695, 16704, 16693, 16671, 16721, 16723, 16722, 16727, 16729, 16728, 16728, 16725, 16724, 16731, 16733, 16732, 16740, 16738, 16737, 16744, 16743, 16741, 16746, 16748, 16747, 16738, 16729, 16727, 16742, 16741, 16732, 16736, 16746, 16745,
16722, 16723, 16731, 16729, 16736, 16735, 16735, 16734, 16725, 16748, 16746, 16738, 16724, 16725, 16723, 16734, 16735, 16733, 16747, 16744, 16742, 16725, 16734, 16731, 16745, 16742, 16733, 16746, 16736, 16729, 16755, 16754, 16752, 16759, 16758, 16756, 16765, 16764, 16762, 16767, 16769, 16768, 16763, 16762, 16754, 16767, 16766, 16758, 16758, 16755, 16753, 16768, 16765, 16763, 16766, 16763, 16755, 16773, 16771, 16770, 16779, 16777, 16776, 16777, 16773, 16772, 16775, 16774, 16771, 16781, 16780, 16777, 16777, 16780, 16775, 16753, 16752, 16743, 16757, 16756, 16747, 16756, 16753, 16744, 16771, 16750, 16749, 16771, 16774, 16751, 16749, 16750, 16740, 16750, 16751, 16748, 16774, 16775, 16759, 16779, 16778, 16760, 16781, 16769, 16767, 16769, 16781, 16779, 16780, 16767, 16759, 16751, 16774, 16757, 16628, 16654, 16653, 16659, 16658, 16630, 16628, 16630, 16658, 16673, 16672, 16640, 16643, 16677, 16676, 16676, 16673, 16641, 16672, 16659, 16631, 16722, 16645, 16644, 16730, 16646, 16645, 16644, 16645, 16639, 16639, 16645, 16646, 16732, 16679, 16677, 16743, 16687, 16685, 16741, 16685, 16679, 16754, 16711, 16709, 16764, 16719, 16717, 16762, 16717, 16711, 16752, 16709, 16687, 16646, 16730, 16677, 16785, 16783, 16782, 16787, 16789, 16788, 16788, 16792, 16790,
16793, 16789, 16787, 16799, 16798, 16796, 16801, 16803, 16802, 16805, 16807, 16806, 16792, 16801, 16800, 16805, 16804, 16798, 16795, 16785, 16784, 16789, 16797, 16796, 16788, 16796, 16798, 16793, 16799, 16797, 16804, 16806, 16803, 16788, 16786, 16783, 16796, 16788, 16785, 16792, 16798, 16804, 16809, 16811, 16810, 16815, 16814, 16812, 16819, 16818, 16816, 16823, 16822, 16820, 16827, 16826, 16824, 16815, 16821, 16820, 16825, 16824, 16818, 16809, 16812, 16814, 16818, 16815, 16813, 16826, 16823, 16821, 16824, 16821, 16815, 16831, 16829, 16828, 16831, 16833, 16832, 16835, 16831, 16830, 16835, 16836, 16833, 16844, 16843, 16841, 16846, 16848, 16847, 16852, 16850, 16849, 16856, 16854, 16853, 16838, 16851, 16849, 16856, 16840, 16839, 16844, 16858, 16857, 16860, 16859, 16847, 16857, 16858, 16852, 16860, 16856, 16855, 16847, 16844, 16842, 16855, 16853, 16850, 16858, 16859, 16855, 16847, 16859, 16858, 16843, 16857, 16851, 16856, 16860, 16848, 16803, 16830, 16828, 16807, 16835, 16834, 16803, 16806, 16834, 16842, 16841, 16822, 16846, 16845, 16826, 16845, 16842, 16823, 16809, 16808, 16787, 16810, 16811, 16793, 16822, 16807, 16805, 16797, 16812, 16809, 16793, 16811, 16814, 16820, 16805, 16799, 16838, 16837, 16832, 16841, 16843, 16836, 16836, 16843, 16838,
16841, 16835, 16807, 16862, 16864, 16863, 16866, 16864, 16862, 16868, 16870, 16869, 16872, 16874, 16873, 16876, 16878, 16877, 16882, 16880, 16879, 16884, 16886, 16885, 16890, 16889, 16887, 16894, 16893, 16891, 16880, 16866, 16865, 16870, 16884, 16883, 16874, 16888, 16887, 16878, 16892, 16891, 16863, 16864, 16872, 16874, 16872, 16864, 16882, 16890, 16888, 16868, 16875, 16877, 16884, 16891, 16893, 16868, 16867, 16862, 16875, 16868, 16864, 16888, 16874, 16866, 16877, 16891, 16884, 16900, 16902, 16901, 16901, 16898, 16897, 16906, 16905, 16903, 16915, 16914, 16912, 16919, 16918, 16916, 16911, 16902, 16900, 16913, 16912, 16905, 16909, 16917, 16916, 16902, 16909, 16908, 16901, 16908, 16907, 16897, 16898, 16896, 16907, 16908, 16906, 16918, 16915, 16913, 16898, 16907, 16904, 16916, 16913, 16906, 16911, 16917, 16909, 16925, 16923, 16922, 16929, 16931, 16930, 16933, 16935, 16934, 16939, 16941, 16940, 16944, 16946, 16945, 16925, 16937, 16936, 16944, 16943, 16934, 16930, 16931, 16925, 16943, 16945, 16941, 16926, 16927, 16921, 16932, 16934, 16931, 16939, 16938, 16936, 16921, 16927, 16932, 16943, 16939, 16937, 16934, 16943, 16942, 16931, 16942, 16937, 16950, 16949, 16947, 16954, 16953, 16951, 16958, 16957, 16955, 16962, 16961, 16959, 16956, 16955, 16949,
16960, 16959, 16953, 16953, 16950, 16948, 16961, 16958, 16956, 16959, 16956, 16950, 16921, 16882, 16881, 16886, 16927, 16926, 16890, 16929, 16928, 16933, 16932, 16893, 16921, 16929, 16890, 16893, 16932, 16927, 16948, 16947, 16914, 16952, 16951, 16918, 16951, 16948, 16915, 16905, 16878, 16876, 16914, 16894, 16892, 16912, 16892, 16878, 16949, 16935, 16933, 16957, 16946, 16944, 16955, 16944, 16935, 16947, 16933, 16894, 16966, 16964, 16963, 16966, 16968, 16967, 16970, 16966, 16965, 16966, 16970, 16971, 16977, 16976, 16974, 16981, 16980, 16978, 16985, 16983, 16982, 16985, 16987, 16986, 16994, 16993, 16991, 16973, 16984, 16982, 16977, 16989, 16988, 16992, 16991, 16980, 16989, 16985, 16984, 16989, 16990, 16987, 16980, 16977, 16975, 16991, 16993, 16990, 16991, 16989, 16977, 16973, 16976, 16988, 17002, 17000, 16999, 16998, 17001, 16999, 17005, 17007, 17006, 17009, 17011, 17010, 17002, 17009, 17008, 17007, 17002, 17001, 16998, 17004, 17006, 17012, 17013, 17011, 16998, 16997, 16995, 17003, 17004, 16998, 17007, 17012, 17009, 16987, 16996, 16995, 17005, 17004, 16993, 16993, 17004, 17003, 16990, 17003, 16996, 16973, 16972, 16967, 16974, 16976, 16971, 16976, 16973, 16968, 16941, 16965, 16963, 16946, 16970, 16969, 16945, 16969, 16965, 16975, 16974, 16957,
16979, 16978, 16961, 16978, 16975, 16958, 16974, 16970, 16946, 16873, 16819, 16817, 16889, 16827, 16825, 16887, 16825, 16819, 16924, 16922, 16839, 16930, 16848, 16846, 16930, 16924, 16840, 16928, 16846, 16827, 16534, 16784, 16782, 16788, 16539, 16537, 16540, 16789, 16787, 16544, 16795, 16794, 16547, 16796, 16795, 16543, 16794, 16784, 16796, 16547, 16539, 16548, 16797, 16789, 16813, 16812, 16575, 16580, 16817, 16816, 16816, 16813, 16576, 16812, 16797, 16548, 16863, 16605, 16603, 16606, 16864, 16862, 16876, 16875, 16615, 16875, 16872, 16612, 16871, 16611, 16605, 16606, 16612, 16872, 16896, 16622, 16620, 16625, 16904, 16903, 16896, 16904, 16625, 16903, 16876, 16616, 16611, 16871, 16817, 16468, 16469, 16335, 16476, 16477, 16343, 16469, 16476, 16341, 16486, 16417, 16415, 16494, 16425, 16423, 16492, 16423, 16417, 16527, 16451, 16449, 16534, 16533, 16459, 16541, 16543, 16467, 16543, 16534, 16461, 16541, 16465, 16451, 16525, 16449, 16425, 16484, 16415, 16343, 17017, 17019, 17018, 17018, 17015, 17014, 17019, 17022, 17021, 17018, 17021, 17020, 17026, 17024, 17023, 17030, 17032, 17031, 17035, 17037, 17036, 17039, 17037, 17035, 17043, 17042, 17040, 17045, 17047, 17046, 17036, 17028, 17027, 17041, 17040, 17031, 17032, 17030, 17024, 17037, 17045, 17044,
17047, 17045, 17037, 17027, 17028, 17026, 17046, 17043, 17041, 17028, 17033, 17032, 17033, 17044, 17041, 17036, 17044, 17033, 17024, 17019, 17017, 17022, 17030, 17029, 17030, 17022, 17019, 17049, 17051, 17050, 17050, 17051, 17053, 17057, 17055, 17054, 17061, 17060, 17058, 17065, 17064, 17062, 17069, 17067, 17066, 17071, 17073, 17072, 17072, 17068, 17066, 17069, 17073, 17071, 17077, 17076, 17074, 17069, 17057, 17056, 17075, 17074, 17060, 17065, 17079, 17078, 17057, 17065, 17063, 17078, 17079, 17069, 17072, 17073, 17081, 17072, 17080, 17078, 17069, 17079, 17081, 17064, 17061, 17059, 17078, 17080, 17077, 17064, 17078, 17075, 17069, 17079, 17065, 17085, 17087, 17086, 17083, 17087, 17085, 17089, 17091, 17090, 17095, 17093, 17092, 17102, 17101, 17099, 17110, 17108, 17107, 17110, 17106, 17104, 17112, 17114, 17113, 17113, 17116, 17115, 17118, 17120, 17119, 17122, 17124, 17123, 17086, 17105, 17103, 17113, 17094, 17092, 17095, 17114, 17112, 17118, 17117, 17097, 17087, 17098, 17097, 17083, 17096, 17098, 17120, 17110, 17109, 17110, 17120, 17118, 17090, 17091, 17100, 17101, 17102, 17095, 17113, 17114, 17122, 17113, 17121, 17123, 17088, 17090, 17083, 17099, 17101, 17098, 17113, 17111, 17104, 17115, 17116, 17110, 17083, 17090, 17099, 17118, 17121, 17113,
17110, 17116, 17123, 17101, 17121, 17118, 17097, 17117, 17105, 17121, 17101, 17094, 17102, 17122, 17114, 17128, 17127, 17125, 17132, 17131, 17129, 17138, 17137, 17135, 17140, 17142, 17141, 17136, 17135, 17127, 17140, 17139, 17131, 17131, 17128, 17126, 17139, 17141, 17138, 17139, 17136, 17128, 17144, 17146, 17145, 17150, 17149, 17147, 17154, 17153, 17151, 17158, 17157, 17155, 17162, 17161, 17159, 17150, 17156, 17155, 17160, 17159, 17153, 17145, 17146, 17148, 17151, 17153, 17150, 17161, 17158, 17156, 17159, 17156, 17150, 17126, 17125, 17076, 17130, 17129, 17080, 17080, 17129, 17126, 17145, 17109, 17107, 17110, 17146, 17144, 17124, 17152, 17151, 17123, 17151, 17148, 17145, 17147, 17119, 17120, 17148, 17146, 17103, 17105, 17069, 17119, 17081, 17079, 17069, 17105, 17117, 17149, 17132, 17130, 17157, 17142, 17140, 17155, 17140, 17132, 17119, 17147, 17130, 17168, 17170, 17169, 17174, 17173, 17171, 17176, 17178, 17177, 17182, 17181, 17179, 17186, 17185, 17183, 17180, 17179, 17173, 17184, 17183, 17177, 17170, 17176, 17175, 17167, 17169, 17166, 17175, 17177, 17174, 17185, 17182, 17180, 17166, 17169, 17175, 17183, 17180, 17174, 17190, 17192, 17191, 17191, 17188, 17187, 17201, 17199, 17198, 17205, 17204, 17202, 17197, 17192, 17190, 17195, 17203, 17202,
17192, 17195, 17194, 17194, 17193, 17188, 17207, 17206, 17199, 17198, 17199, 17197, 17206, 17207, 17205, 17199, 17206, 17203, 17197, 17203, 17195, 17211, 17210, 17208, 17215, 17214, 17212, 17219, 17218, 17216, 17223, 17222, 17220, 17217, 17216, 17210, 17221, 17220, 17214, 17214, 17211, 17209, 17222, 17219, 17217, 17220, 17217, 17211, 17227, 17229, 17228, 17228, 17225, 17224, 17233, 17232, 17230, 17242, 17240, 17239, 17242, 17238, 17237, 17246, 17245, 17243, 17250, 17249, 17247, 17238, 17229, 17227, 17244, 17243, 17232, 17236, 17248, 17247, 17229, 17236, 17235, 17228, 17235, 17234, 17249, 17250, 17242, 17250, 17248, 17238, 17234, 17235, 17233, 17247, 17249, 17246, 17247, 17244, 17233, 17238, 17248, 17236, 17209, 17208, 17181, 17213, 17212, 17185, 17212, 17209, 17182, 17225, 17201, 17200, 17231, 17230, 17204, 17207, 17234, 17231, 17234, 17207, 17201, 17187, 17188, 17170, 17193, 17194, 17178, 17204, 17186, 17184, 17188, 17193, 17176, 17202, 17184, 17178, 17232, 17215, 17213, 17245, 17223, 17221, 17243, 17221, 17215, 17230, 17213, 17186, 17055, 17039, 17038, 17059, 17058, 17042, 17047, 17063, 17062, 17062, 17059, 17043, 17063, 17047, 17039, 17050, 17165, 17163, 17166, 17051, 17049, 17172, 17171, 17052, 17050, 17052, 17171, 17172, 17053, 17051,
17163, 17165, 17091, 17173, 17102, 17100, 17179, 17181, 17124, 17165, 17171, 17100, 17173, 17179, 17122, 17210, 17154, 17152, 17218, 17162, 17160, 17216, 17160, 17154, 17208, 17152, 17124, 17014, 17015, 16911, 17020, 17021, 16919, 17015, 17020, 16917, 17031, 16954, 16952, 17042, 16962, 16960, 17040, 16960, 16954, 17029, 16952, 16919, 17060, 16981, 16979, 17076, 16994, 16992, 17074, 16992, 16981, 17127, 17007, 17005, 17134, 17133, 17010, 17135, 17137, 17013, 17137, 17134, 17011, 17127, 17135, 17012, 17125, 17005, 16994, 17058, 16979, 16962, 17252, 17254, 17253, 17256, 17258, 17257, 17262, 17261, 17259, 17266, 17264, 17263, 17272, 17270, 17269, 17268, 17271, 17269, 17276, 17278, 17277, 17254, 17265, 17263, 17258, 17274, 17273, 17276, 17275, 17261, 17255, 17257, 17254, 17274, 17266, 17265, 17278, 17272, 17271, 17275, 17277, 17271, 17261, 17258, 17256, 17268, 17267, 17264, 17275, 17268, 17266, 17261, 17275, 17274, 17257, 17273, 17265, 17282, 17281, 17279, 17286, 17285, 17283, 17294, 17293, 17291, 17296, 17298, 17297, 17292, 17291, 17281, 17296, 17295, 17285, 17285, 17282, 17280, 17297, 17294, 17292, 17295, 17292, 17282, 17305, 17303, 17302, 17309, 17307, 17306, 17310, 17311, 17305, 17313, 17309, 17308, 17308, 17306, 17303, 17311, 17312, 17308,
17288, 17304, 17302, 17309, 17290, 17289, 17294, 17311, 17310, 17313, 17312, 17297, 17297, 17312, 17311, 17293, 17310, 17304, 17309, 17313, 17298, 17281, 17262, 17260, 17288, 17287, 17270, 17291, 17293, 17278, 17278, 17293, 17288, 17291, 17276, 17262, 17319, 17317, 17316, 17319, 17315, 17314, 17323, 17322, 17320, 17327, 17326, 17324, 17333, 17331, 17330, 17333, 17329, 17328, 17339, 17338, 17336, 17335, 17318, 17316, 17323, 17337, 17336, 17338, 17339, 17333, 17333, 17339, 17337, 17326, 17327, 17319, 17319, 17327, 17325, 17324, 17326, 17323, 17335, 17334, 17328, 17340, 17335, 17329, 17340, 17337, 17323, 17340, 17326, 17318, 17344, 17346, 17345, 17345, 17342, 17341, 17355, 17353, 17352, 17359, 17358, 17356, 17351, 17346, 17344, 17349, 17357, 17356, 17346, 17349, 17348, 17348, 17347, 17342, 17361, 17360, 17353, 17352, 17353, 17351, 17360, 17361, 17359, 17353, 17360, 17357, 17351, 17357, 17349, 17363, 17365, 17364, 17367, 17365, 17363, 17371, 17370, 17368, 17373, 17375, 17374, 17381, 17380, 17378, 17385, 17384, 17382, 17377, 17367, 17366, 17379, 17378, 17370, 17375, 17383, 17382, 17365, 17373, 17372, 17375, 17373, 17365, 17374, 17371, 17369, 17384, 17381, 17379, 17382, 17379, 17371, 17383, 17375, 17367, 17389, 17391, 17390, 17390, 17387, 17386,
17390, 17391, 17394, 17390, 17393, 17392, 17364, 17355, 17354, 17369, 17368, 17358, 17361, 17372, 17369, 17372, 17361, 17355, 17386, 17387, 17377, 17392, 17393, 17385, 17387, 17392, 17383, 17342, 17300, 17299, 17347, 17301, 17300, 17322, 17286, 17284, 17332, 17330, 17289, 17338, 17298, 17296, 17298, 17338, 17332, 17336, 17296, 17286, 17398, 17400, 17399, 17405, 17407, 17406, 17406, 17409, 17408, 17415, 17414, 17412, 17419, 17418, 17416, 17396, 17407, 17405, 17411, 17400, 17398, 17403, 17417, 17416, 17407, 17413, 17412, 17412, 17414, 17409, 17399, 17400, 17403, 17397, 17399, 17396, 17416, 17418, 17415, 17399, 17402, 17401, 17416, 17413, 17401, 17396, 17401, 17413, 17417, 17403, 17400, 17423, 17421, 17420, 17423, 17425, 17424, 17431, 17429, 17428, 17427, 17430, 17428, 17435, 17433, 17432, 17437, 17439, 17438, 17443, 17442, 17440, 17439, 17431, 17430, 17427, 17436, 17438, 17442, 17443, 17435, 17434, 17432, 17429, 17440, 17442, 17439, 17442, 17434, 17431, 17447, 17445, 17444, 17449, 17448, 17445, 17409, 17427, 17426, 17415, 17437, 17436, 17441, 17440, 17418, 17440, 17437, 17415, 17414, 17436, 17427, 17445, 17423, 17422, 17445, 17448, 17425, 17420, 17421, 17411, 17424, 17425, 17419, 17421, 17424, 17417, 17447, 17446, 17433, 17449, 17443, 17441,
17443, 17449, 17447, 17425, 17448, 17441, 17453, 17452, 17450, 17457, 17456, 17454, 17461, 17460, 17458, 17465, 17464, 17462, 17459, 17458, 17452, 17463, 17462, 17456, 17456, 17453, 17451, 17464, 17461, 17459, 17462, 17459, 17453, 17469, 17467, 17466, 17473, 17475, 17474, 17480, 17482, 17481, 17481, 17478, 17477, 17486, 17485, 17483, 17478, 17471, 17470, 17484, 17483, 17474, 17475, 17473, 17467, 17482, 17489, 17488, 17488, 17487, 17478, 17470, 17471, 17469, 17487, 17488, 17486, 17471, 17476, 17475, 17476, 17487, 17484, 17487, 17476, 17471, 17491, 17493, 17492, 17495, 17497, 17496, 17501, 17500, 17498, 17505, 17503, 17502, 17505, 17507, 17506, 17516, 17515, 17513, 17493, 17504, 17502, 17497, 17511, 17510, 17514, 17513, 17500, 17494, 17496, 17493, 17511, 17505, 17504, 17505, 17511, 17512, 17500, 17497, 17495, 17509, 17508, 17506, 17513, 17515, 17512, 17515, 17509, 17507, 17513, 17511, 17497, 17496, 17510, 17504, 17520, 17518, 17517, 17524, 17523, 17521, 17526, 17528, 17527, 17532, 17531, 17529, 17536, 17535, 17533, 17530, 17529, 17523, 17534, 17533, 17527, 17528, 17526, 17518, 17527, 17524, 17522, 17535, 17532, 17530, 17533, 17530, 17524, 17495, 17494, 17460, 17499, 17498, 17464, 17498, 17495, 17461, 17518, 17482, 17480, 17522, 17521, 17485,
17489, 17526, 17525, 17525, 17522, 17486, 17526, 17489, 17482, 17474, 17457, 17455, 17485, 17465, 17463, 17483, 17463, 17457, 17523, 17501, 17499, 17531, 17516, 17514, 17529, 17514, 17501, 17521, 17499, 17465, 17543, 17545, 17544, 17544, 17541, 17540, 17541, 17538, 17537, 17545, 17548, 17547, 17547, 17546, 17541, 17546, 17539, 17538, 17554, 17552, 17551, 17550, 17553, 17551, 17556, 17558, 17557, 17554, 17562, 17561, 17558, 17554, 17553, 17555, 17557, 17553, 17554, 17558, 17564, 17568, 17567, 17565, 17572, 17571, 17569, 17574, 17576, 17575, 17578, 17580, 17579, 17584, 17583, 17581, 17578, 17577, 17567, 17582, 17581, 17571, 17577, 17579, 17576, 17571, 17568, 17566, 17583, 17580, 17578, 17581, 17578, 17568, 17586, 17588, 17587, 17590, 17588, 17586, 17592, 17594, 17593, 17600, 17598, 17597, 17604, 17603, 17601, 17596, 17590, 17589, 17594, 17602, 17601, 17588, 17592, 17591, 17594, 17592, 17588, 17606, 17605, 17598, 17597, 17598, 17596, 17605, 17606, 17604, 17598, 17605, 17602, 17602, 17594, 17590, 17610, 17608, 17607, 17610, 17612, 17611, 17619, 17618, 17616, 17627, 17625, 17624, 17623, 17626, 17624, 17630, 17632, 17631, 17612, 17621, 17620, 17630, 17629, 17618, 17614, 17610, 17609, 17614, 17615, 17612, 17632, 17627, 17626, 17623, 17629, 17631,
17616, 17618, 17615, 17623, 17622, 17620, 17629, 17623, 17621, 17618, 17629, 17628, 17615, 17628, 17621, 17587, 17560, 17559, 17587, 17591, 17563, 17576, 17609, 17607, 17580, 17614, 17613, 17617, 17616, 17583, 17616, 17614, 17580, 17579, 17613, 17609, 17567, 17558, 17556, 17574, 17573, 17561, 17577, 17574, 17562, 17567, 17577, 17564, 17636, 17634, 17633, 17640, 17642, 17641, 17645, 17647, 17646, 17649, 17647, 17645, 17653, 17652, 17650, 17655, 17657, 17656, 17646, 17638, 17637, 17651, 17650, 17641, 17642, 17640, 17634, 17646, 17647, 17655, 17657, 17655, 17647, 17637, 17638, 17636, 17656, 17653, 17651, 17638, 17643, 17642, 17643, 17654, 17651, 17646, 17654, 17643, 17661, 17659, 17658, 17665, 17664, 17662, 17667, 17669, 17668, 17673, 17672, 17670, 17677, 17676, 17674, 17671, 17670, 17664, 17675, 17674, 17668, 17661, 17669, 17667, 17668, 17665, 17663, 17676, 17673, 17671, 17674, 17671, 17665, 17684, 17686, 17685, 17685, 17682, 17681, 17682, 17679, 17678, 17686, 17689, 17688, 17688, 17687, 17682, 17687, 17680, 17679, 17659, 17649, 17648, 17663, 17662, 17652, 17657, 17667, 17666, 17666, 17663, 17653, 17667, 17657, 17649, 17678, 17679, 17661, 17687, 17688, 17677, 17679, 17680, 17669, 17680, 17687, 17675, 17509, 17550, 17549, 17516, 17556, 17555,
17515, 17555, 17550, 17566, 17565, 17531, 17570, 17569, 17535, 17569, 17566, 17532, 17565, 17556, 17516, 17634, 17545, 17543, 17548, 17640, 17639, 17640, 17548, 17545, 17537, 17538, 17520, 17546, 17547, 17536, 17538, 17539, 17528, 17539, 17546, 17534, 17641, 17572, 17570, 17652, 17584, 17582, 17650, 17582, 17572, 17664, 17619, 17617, 17672, 17632, 17630, 17670, 17630, 17619, 17662, 17617, 17584, 17639, 17570, 17536, 17693, 17695, 17694, 17694, 17697, 17696, 17702, 17701, 17699, 17694, 17695, 17700, 17694, 17699, 17701, 17696, 17697, 17691, 17691, 17697, 17701, 17704, 17706, 17705, 17712, 17711, 17709, 17714, 17716, 17715, 17722, 17721, 17719, 17706, 17714, 17713, 17720, 17719, 17711, 17704, 17707, 17708, 17717, 17718, 17716, 17711, 17708, 17707, 17719, 17721, 17718, 17719, 17717, 17708, 17706, 17708, 17717, 17691, 17704, 17703, 17702, 17710, 17709, 17709, 17707, 17698, 17691, 17698, 17707, 17727, 17729, 17728, 17728, 17731, 17730, 17735, 17737, 17736, 17736, 17733, 17732, 17741, 17740, 17738, 17729, 17739, 17738, 17728, 17738, 17740, 17736, 17737, 17744, 17743, 17742, 17733, 17732, 17733, 17729, 17742, 17743, 17741, 17729, 17733, 17742, 17748, 17750, 17749, 17749, 17752, 17751, 17757, 17756, 17754, 17759, 17761, 17760, 17767, 17766, 17764,
17746, 17759, 17758, 17757, 17765, 17764, 17759, 17762, 17763, 17749, 17750, 17755, 17754, 17756, 17752, 17751, 17752, 17746, 17766, 17763, 17762, 17746, 17752, 17756, 17764, 17762, 17753, 17753, 17762, 17759, 17771, 17770, 17768, 17775, 17774, 17772, 17779, 17778, 17776, 17783, 17782, 17780, 17777, 17776, 17770, 17781, 17780, 17774, 17774, 17771, 17769, 17782, 17779, 17777, 17780, 17777, 17771, 17724, 17750, 17748, 17724, 17725, 17755, 17741, 17769, 17768, 17773, 17772, 17743, 17772, 17769, 17741, 17730, 17731, 17724, 17731, 17740, 17725, 17768, 17770, 17757, 17778, 17767, 17765, 17776, 17765, 17757, 17768, 17755, 17725, 17796, 17794, 17793, 17791, 17795, 17793, 17796, 17792, 17790, 17799, 17801, 17800, 17796, 17803, 17802, 17801, 17796, 17795, 17791, 17798, 17800, 17801, 17799, 17792, 17791, 17789, 17787, 17797, 17798, 17791, 17801, 17804, 17803, 17806, 17808, 17807, 17810, 17808, 17806, 17814, 17816, 17815, 17821, 17819, 17818, 17825, 17824, 17822, 17819, 17812, 17811, 17823, 17822, 17815, 17808, 17814, 17813, 17810, 17816, 17814, 17827, 17826, 17819, 17811, 17812, 17810, 17827, 17825, 17823, 17812, 17817, 17816, 17817, 17826, 17823, 17819, 17826, 17817, 17807, 17785, 17784, 17813, 17786, 17785, 17716, 17788, 17787, 17799, 17798, 17721,
17798, 17797, 17718, 17716, 17718, 17797, 17784, 17785, 17737, 17785, 17786, 17744, 17815, 17775, 17773, 17824, 17783, 17781, 17822, 17781, 17775, 17786, 17813, 17773, 17451, 17450, 17380, 17455, 17454, 17384, 17454, 17451, 17381, 17467, 17391, 17389, 17394, 17473, 17472, 17473, 17394, 17391, 17472, 17455, 17385, 17831, 17829, 17828, 17831, 17833, 17832, 17837, 17835, 17834, 17839, 17841, 17840, 17840, 17836, 17834, 17837, 17841, 17839, 17852, 17854, 17853, 17859, 17858, 17856, 17833, 17850, 17849, 17857, 17856, 17847, 17843, 17831, 17830, 17843, 17844, 17833, 17846, 17837, 17836, 17841, 17848, 17847, 17847, 17845, 17836, 17846, 17848, 17841, 17852, 17856, 17858, 17845, 17847, 17844, 17852, 17851, 17849, 17856, 17852, 17850, 17856, 17855, 17844, 17844, 17855, 17850, 17863, 17862, 17860, 17867, 17866, 17864, 17873, 17875, 17874, 17879, 17878, 17876, 17873, 17872, 17862, 17877, 17876, 17866, 17866, 17863, 17861, 17871, 17870, 17868, 17876, 17878, 17875, 17878, 17871, 17869, 17876, 17873, 17863, 17885, 17883, 17882, 17881, 17884, 17882, 17887, 17889, 17888, 17891, 17893, 17892, 17885, 17891, 17890, 17889, 17885, 17884, 17881, 17886, 17888, 17894, 17895, 17893, 17885, 17889, 17894, 17903, 17902, 17900, 17907, 17906, 17904, 17911, 17910, 17908,
17915, 17914, 17912, 17909, 17908, 17902, 17913, 17912, 17906, 17898, 17899, 17897, 17906, 17903, 17901, 17914, 17911, 17909, 17899, 17904, 17901, 17912, 17909, 17903, 17854, 17881, 17880, 17859, 17887, 17886, 17858, 17886, 17881, 17897, 17869, 17868, 17871, 17899, 17898, 17875, 17901, 17900, 17879, 17905, 17904, 17901, 17875, 17869, 17878, 17904, 17899, 17862, 17848, 17846, 17872, 17874, 17859, 17862, 17872, 17857, 17902, 17889, 17887, 17908, 17910, 17895, 17902, 17908, 17894, 17900, 17887, 17859, 17919, 17918, 17916, 17923, 17922, 17920, 17929, 17928, 17926, 17931, 17933, 17932, 17927, 17926, 17918, 17931, 17930, 17922, 17922, 17919, 17917, 17932, 17929, 17927, 17930, 17927, 17919, 17939, 17937, 17936, 17941, 17943, 17942, 17949, 17947, 17946, 17951, 17948, 17946, 17949, 17935, 17934, 17953, 17952, 17942, 17953, 17949, 17948, 17951, 17954, 17952, 17939, 17945, 17944, 17939, 17938, 17934, 17945, 17943, 17941, 17943, 17945, 17939, 17953, 17943, 17935, 17958, 17956, 17955, 17962, 17960, 17959, 17964, 17966, 17965, 17974, 17972, 17971, 17978, 17977, 17975, 17970, 17958, 17957, 17966, 17976, 17975, 17966, 17964, 17956, 17967, 17968, 17962, 17980, 17979, 17972, 17961, 17959, 17955, 17971, 17972, 17970, 17979, 17980, 17978, 17964, 17967, 17961,
17972, 17979, 17976, 17976, 17966, 17958, 17962, 17925, 17924, 17929, 17964, 17963, 17968, 17967, 17932, 17967, 17964, 17929, 17968, 17933, 17925, 17942, 17923, 17921, 17951, 17950, 17924, 17954, 17933, 17931, 17954, 17951, 17925, 17942, 17952, 17931, 17985, 17987, 17986, 17989, 17991, 17990, 17995, 17994, 17992, 17997, 17999, 17998, 18001, 18003, 18002, 18007, 18006, 18004, 17987, 17997, 17996, 17991, 18001, 18000, 18005, 18004, 17994, 17985, 17988, 17990, 18000, 18002, 17999, 17994, 17991, 17989, 18006, 18003, 18001, 18004, 18001, 17991, 17990, 18000, 17997, 18013, 18011, 18010, 18009, 18012, 18010, 18015, 18017, 18016, 18021, 18020, 18018, 18023, 18025, 18024, 18031, 18030, 18028, 18013, 18023, 18022, 18029, 18028, 18020, 18017, 18013, 18012, 18009, 18014, 18016, 18026, 18027, 18025, 18020, 18017, 18015, 18028, 18030, 18027, 18020, 18028, 18026, 18013, 18017, 18026, 17999, 18009, 18008, 18003, 18015, 18014, 18019, 18018, 18006, 18018, 18015, 18003, 18002, 18014, 18009, 17985, 17984, 17981, 17988, 17985, 17982, 18035, 18033, 18032, 18039, 18038, 18036, 18045, 18043, 18042, 18049, 18048, 18046, 18043, 18035, 18034, 18047, 18046, 18038, 18041, 18040, 18033, 18051, 18050, 18043, 18040, 18041, 18039, 18050, 18051, 18049, 18041, 18050, 18047,
18050, 18041, 18035, 18055, 18057, 18056, 18056, 18053, 18052, 18061, 18060, 18058, 18068, 18066, 18065, 18072, 18071, 18069, 18076, 18075, 18073, 18066, 18057, 18055, 18070, 18069, 18060, 18064, 18074, 18073, 18057, 18064, 18063, 18056, 18063, 18062, 18076, 18074, 18066, 18062, 18063, 18061, 18075, 18072, 18070, 18073, 18070, 18061, 18066, 18074, 18064, 18078, 18080, 18079, 18082, 18084, 18083, 18090, 18092, 18091, 18096, 18095, 18093, 18100, 18099, 18097, 18079, 18090, 18089, 18094, 18093, 18085, 18088, 18098, 18097, 18079, 18080, 18086, 18090, 18093, 18095, 18083, 18084, 18088, 18081, 18083, 18080, 18099, 18096, 18094, 18080, 18083, 18087, 18097, 18094, 18086, 18085, 18093, 18090, 18053, 18045, 18044, 18059, 18058, 18048, 18051, 18062, 18059, 18062, 18051, 18045, 17893, 17982, 17981, 17893, 17895, 17983, 17989, 17988, 17910, 17993, 17992, 17914, 17992, 17989, 17911, 17910, 17988, 17983, 18033, 17974, 17973, 18037, 18036, 17977, 17980, 18040, 18037, 18040, 17980, 17974, 17918, 17867, 17865, 17928, 17879, 17877, 17926, 17877, 17867, 17963, 17965, 17907, 17977, 17915, 17913, 17975, 17913, 17907, 17928, 17963, 17905, 18038, 17995, 17993, 18048, 18007, 18005, 18046, 18005, 17995, 18060, 18021, 18019, 18071, 18031, 18029, 18069, 18029, 18021,
18058, 18019, 18007, 18036, 17993, 17915, 18107, 18105, 18104, 18105, 18102, 18101, 18109, 18108, 18105, 18108, 18103, 18102, 18111, 18113, 18112, 18122, 18124, 18123, 18126, 18128, 18127, 18130, 18132, 18131, 18112, 18120, 18119, 18123, 18115, 18114, 18126, 18125, 18116, 18113, 18117, 18116, 18123, 18124, 18130, 18114, 18115, 18113, 18131, 18128, 18126, 18113, 18115, 18118, 18118, 18129, 18126, 18112, 18116, 18125, 18129, 18118, 18115, 18134, 18136, 18135, 18142, 18141, 18139, 18148, 18150, 18149, 18149, 18145, 18143, 18146, 18150, 18148, 18160, 18159, 18157, 18136, 18146, 18144, 18158, 18157, 18141, 18137, 18138, 18136, 18150, 18156, 18155, 18155, 18153, 18145, 18154, 18156, 18150, 18139, 18141, 18138, 18152, 18151, 18148, 18159, 18156, 18154, 18159, 18152, 18150, 18141, 18157, 18154, 18138, 18154, 18146, 18164, 18163, 18161, 18168, 18167, 18165, 18172, 18171, 18169, 18176, 18175, 18173, 18170, 18169, 18163, 18174, 18173, 18167, 18167, 18164, 18162, 18175, 18172, 18170, 18173, 18170, 18164, 18180, 18178, 18177, 18182, 18184, 18183, 18186, 18188, 18187, 18192, 18191, 18189, 18196, 18194, 18193, 18198, 18200, 18199, 18202, 18204, 18203, 18206, 18208, 18207, 18188, 18202, 18201, 18206, 18205, 18191, 18180, 18188, 18186, 18204, 18202, 18194,
18189, 18191, 18184, 18205, 18207, 18200, 18183, 18184, 18180, 18198, 18197, 18193, 18184, 18191, 18188, 18205, 18198, 18194, 18191, 18205, 18202, 18162, 18161, 18127, 18166, 18165, 18131, 18165, 18162, 18128, 18178, 18150, 18148, 18152, 18182, 18181, 18186, 18185, 18155, 18160, 18190, 18189, 18186, 18156, 18150, 18152, 18159, 18189, 18143, 18145, 18124, 18153, 18155, 18132, 18145, 18153, 18130, 18187, 18168, 18166, 18203, 18176, 18174, 18201, 18174, 18168, 18185, 18166, 18132, 18212, 18210, 18209, 18216, 18215, 18213, 18220, 18219, 18217, 18224, 18222, 18221, 18228, 18227, 18225, 18232, 18231, 18229, 18222, 18212, 18211, 18226, 18225, 18215, 18220, 18230, 18229, 18220, 18218, 18210, 18224, 18232, 18230, 18219, 18216, 18214, 18231, 18228, 18226, 18229, 18226, 18216, 18222, 18230, 18220, 18236, 18234, 18233, 18236, 18238, 18237, 18244, 18242, 18241, 18244, 18240, 18239, 18251, 18250, 18248, 18246, 18236, 18235, 18246, 18247, 18238, 18251, 18244, 18243, 18244, 18251, 18249, 18243, 18241, 18237, 18248, 18250, 18247, 18247, 18250, 18243, 18255, 18253, 18252, 18257, 18254, 18252, 18261, 18260, 18258, 18268, 18266, 18265, 18272, 18271, 18269, 18266, 18257, 18256, 18270, 18269, 18260, 18262, 18263, 18255, 18264, 18262, 18254, 18268, 18274, 18273,
18264, 18261, 18259, 18274, 18272, 18270, 18273, 18270, 18261, 18266, 18273, 18264, 18255, 18224, 18223, 18259, 18258, 18227, 18263, 18262, 18231, 18231, 18262, 18259, 18263, 18232, 18224, 18276, 18278, 18277, 18284, 18283, 18281, 18286, 18288, 18287, 18292, 18290, 18289, 18296, 18294, 18293, 18296, 18280, 18279, 18284, 18298, 18297, 18300, 18299, 18287, 18276, 18281, 18283, 18298, 18292, 18291, 18300, 18296, 18295, 18287, 18284, 18282, 18295, 18293, 18290, 18298, 18299, 18295, 18287, 18299, 18298, 18296, 18300, 18288, 18306, 18304, 18303, 18312, 18311, 18309, 18314, 18316, 18315, 18318, 18320, 18319, 18319, 18322, 18321, 18326, 18325, 18323, 18330, 18329, 18327, 18306, 18320, 18318, 18328, 18327, 18315, 18311, 18312, 18306, 18319, 18320, 18324, 18319, 18323, 18325, 18307, 18308, 18302, 18315, 18312, 18310, 18327, 18329, 18326, 18302, 18308, 18313, 18315, 18327, 18324, 18306, 18312, 18324, 18332, 18334, 18333, 18340, 18342, 18341, 18338, 18342, 18340, 18334, 18338, 18337, 18335, 18336, 18334, 18341, 18342, 18345, 18338, 18343, 18345, 18334, 18336, 18343, 18349, 18351, 18350, 18355, 18354, 18352, 18359, 18358, 18356, 18363, 18362, 18360, 18357, 18356, 18350, 18361, 18360, 18354, 18354, 18351, 18349, 18362, 18359, 18357, 18360, 18357, 18351,
18322, 18347, 18346, 18349, 18348, 18325, 18353, 18352, 18329, 18352, 18349, 18326, 18322, 18325, 18348, 18305, 18303, 18279, 18311, 18288, 18286, 18288, 18311, 18305, 18346, 18347, 18332, 18350, 18336, 18335, 18358, 18345, 18343, 18332, 18347, 18348, 18356, 18343, 18336, 18369, 18367, 18366, 18371, 18373, 18372, 18379, 18377, 18376, 18383, 18382, 18380, 18377, 18365, 18364, 18373, 18381, 18380, 18379, 18383, 18381, 18375, 18374, 18367, 18369, 18368, 18364, 18375, 18373, 18371, 18373, 18375, 18369, 18381, 18373, 18365, 18387, 18385, 18384, 18391, 18390, 18388, 18395, 18393, 18392, 18397, 18394, 18392, 18395, 18387, 18386, 18399, 18398, 18390, 18387, 18391, 18389, 18398, 18399, 18395, 18397, 18400, 18398, 18399, 18391, 18387, 18385, 18379, 18378, 18389, 18388, 18382, 18385, 18389, 18383, 18282, 18281, 18171, 18286, 18285, 18175, 18285, 18282, 18172, 18302, 18196, 18195, 18200, 18308, 18307, 18204, 18310, 18309, 18314, 18313, 18207, 18302, 18310, 18204, 18200, 18207, 18313, 18309, 18286, 18176, 18367, 18268, 18267, 18371, 18370, 18271, 18374, 18371, 18272, 18367, 18374, 18274, 18215, 18142, 18140, 18227, 18160, 18158, 18225, 18158, 18142, 18260, 18192, 18190, 18271, 18208, 18206, 18269, 18206, 18192, 18258, 18190, 18160, 18372, 18316, 18314,
18382, 18330, 18328, 18380, 18328, 18316, 18390, 18355, 18353, 18400, 18363, 18361, 18390, 18398, 18361, 18388, 18353, 18330, 18370, 18314, 18208, 18025, 18134, 18133, 18140, 18139, 18030, 18139, 18137, 18027, 18027, 18137, 18134, 18210, 18068, 18067, 18214, 18213, 18071, 18076, 18218, 18217, 18217, 18214, 18072, 18218, 18076, 18068, 18092, 18235, 18233, 18096, 18246, 18245, 18100, 18249, 18248, 18248, 18246, 18096, 18092, 18095, 18245, 18213, 18140, 18031, 18402, 18107, 18106, 18403, 18109, 18107, 18101, 18102, 18084, 18108, 18109, 18100, 18102, 18103, 18088, 18103, 18108, 18098, 18402, 18401, 18239, 18403, 18402, 18240, 18403, 18249, 18100, 18405, 18407, 18406, 18409, 18407, 18405, 18413, 18412, 18410, 18415, 18417, 18416, 18421, 18420, 18418, 18425, 18424, 18422, 18419, 18418, 18412, 18423, 18422, 18416, 18407, 18415, 18414, 18409, 18417, 18415, 18416, 18413, 18411, 18424, 18421, 18419, 18422, 18419, 18413, 18429, 18431, 18430, 18438, 18440, 18439, 18439, 18436, 18435, 18436, 18427, 18426, 18440, 18443, 18442, 18442, 18441, 18436, 18433, 18434, 18431, 18441, 18432, 18427, 18447, 18446, 18444, 18451, 18450, 18448, 18455, 18454, 18452, 18459, 18458, 18456, 18453, 18452, 18446, 18457, 18456, 18450, 18450, 18447, 18445, 18458, 18455, 18453,
18456, 18453, 18447, 18463, 18461, 18460, 18467, 18469, 18468, 18474, 18476, 18475, 18475, 18472, 18471, 18480, 18479, 18477, 18472, 18465, 18464, 18478, 18477, 18468, 18469, 18467, 18461, 18476, 18483, 18482, 18482, 18481, 18472, 18464, 18465, 18463, 18481, 18482, 18480, 18465, 18470, 18469, 18470, 18481, 18478, 18481, 18470, 18465, 18445, 18444, 18420, 18449, 18448, 18424, 18448, 18445, 18421, 18461, 18440, 18438, 18443, 18467, 18466, 18467, 18443, 18440, 18426, 18427, 18409, 18441, 18442, 18425, 18427, 18432, 18417, 18432, 18441, 18423, 18468, 18451, 18449, 18479, 18459, 18457, 18477, 18457, 18451, 18466, 18449, 18425, 18489, 18488, 18486, 18493, 18492, 18490, 18499, 18497, 18496, 18495, 18498, 18496, 18501, 18503, 18502, 18507, 18506, 18504, 18485, 18495, 18494, 18489, 18501, 18500, 18505, 18504, 18492, 18503, 18499, 18498, 18500, 18502, 18498, 18492, 18489, 18487, 18506, 18503, 18501, 18504, 18501, 18489, 18488, 18500, 18495, 18511, 18509, 18508, 18515, 18517, 18516, 18522, 18520, 18519, 18526, 18525, 18523, 18520, 18513, 18512, 18524, 18523, 18516, 18517, 18515, 18509, 18528, 18527, 18520, 18512, 18513, 18511, 18527, 18528, 18526, 18513, 18518, 18517, 18518, 18527, 18524, 18527, 18518, 18513, 18530, 18532, 18531, 18540, 18539, 18537,
18542, 18544, 18543, 18546, 18548, 18547, 18534, 18542, 18541, 18540, 18546, 18545, 18535, 18536, 18532, 18545, 18547, 18544, 18534, 18533, 18531, 18537, 18539, 18536, 18539, 18534, 18532, 18539, 18545, 18542, 18550, 18552, 18551, 18554, 18552, 18550, 18558, 18557, 18555, 18560, 18562, 18561, 18566, 18564, 18563, 18570, 18569, 18567, 18572, 18574, 18573, 18564, 18554, 18553, 18568, 18567, 18557, 18562, 18572, 18571, 18552, 18560, 18559, 18554, 18562, 18560, 18566, 18574, 18572, 18561, 18558, 18556, 18573, 18570, 18568, 18571, 18568, 18558, 18572, 18562, 18554, 18499, 18530, 18529, 18538, 18537, 18506, 18506, 18537, 18535, 18503, 18535, 18530, 18551, 18522, 18521, 18556, 18555, 18525, 18528, 18559, 18556, 18551, 18559, 18528, 18516, 18493, 18491, 18525, 18507, 18505, 18523, 18505, 18493, 18557, 18540, 18538, 18569, 18548, 18546, 18567, 18546, 18540, 18555, 18538, 18507, 18576, 18578, 18577, 18580, 18582, 18581, 18586, 18585, 18583, 18590, 18588, 18587, 18590, 18592, 18591, 18601, 18600, 18598, 18578, 18589, 18587, 18582, 18596, 18595, 18599, 18598, 18585, 18579, 18581, 18578, 18596, 18590, 18589, 18590, 18596, 18597, 18585, 18582, 18580, 18594, 18593, 18591, 18598, 18600, 18597, 18600, 18594, 18592, 18598, 18596, 18582, 18581, 18595, 18589,
18605, 18603, 18602, 18609, 18608, 18606, 18611, 18613, 18612, 18619, 18618, 18616, 18623, 18622, 18620, 18615, 18605, 18604, 18617, 18616, 18608, 18613, 18621, 18620, 18613, 18611, 18603, 18612, 18609, 18607, 18622, 18619, 18617, 18620, 18617, 18609, 18615, 18621, 18613, 18629, 18627, 18626, 18625, 18628, 18626, 18631, 18633, 18632, 18635, 18637, 18636, 18629, 18635, 18634, 18633, 18629, 18628, 18630, 18632, 18628, 18638, 18639, 18637, 18633, 18638, 18635, 18643, 18642, 18640, 18647, 18646, 18644, 18653, 18652, 18650, 18657, 18656, 18654, 18651, 18650, 18642, 18655, 18654, 18646, 18646, 18643, 18641, 18656, 18653, 18651, 18654, 18651, 18643, 18594, 18625, 18624, 18601, 18631, 18630, 18600, 18630, 18625, 18641, 18640, 18618, 18645, 18644, 18622, 18644, 18641, 18619, 18608, 18586, 18584, 18618, 18601, 18599, 18616, 18599, 18586, 18642, 18633, 18631, 18649, 18648, 18636, 18650, 18652, 18639, 18652, 18649, 18637, 18642, 18650, 18638, 18640, 18631, 18601, 18661, 18659, 18658, 18661, 18663, 18662, 18670, 18668, 18667, 18663, 18672, 18671, 18670, 18674, 18673, 18664, 18665, 18661, 18665, 18666, 18663, 18663, 18666, 18675, 18679, 18678, 18676, 18683, 18682, 18680, 18685, 18687, 18686, 18689, 18691, 18690, 18695, 18694, 18692, 18689, 18688, 18678,
18693, 18692, 18682, 18685, 18688, 18690, 18682, 18679, 18677, 18694, 18691, 18689, 18692, 18689, 18679, 18697, 18699, 18698, 18701, 18699, 18697, 18703, 18705, 18704, 18709, 18707, 18706, 18713, 18715, 18714, 18707, 18701, 18700, 18705, 18713, 18712, 18699, 18703, 18702, 18705, 18703, 18699, 18715, 18713, 18707, 18710, 18711, 18709, 18711, 18716, 18715, 18713, 18705, 18701, 18722, 18720, 18719, 18718, 18721, 18719, 18724, 18726, 18725, 18730, 18729, 18727, 18732, 18734, 18733, 18740, 18739, 18737, 18722, 18732, 18731, 18738, 18737, 18729, 18726, 18722, 18721, 18718, 18723, 18725, 18735, 18736, 18734, 18729, 18726, 18724, 18737, 18739, 18736, 18729, 18737, 18735, 18722, 18726, 18735, 18698, 18670, 18669, 18702, 18674, 18670, 18687, 18718, 18717, 18691, 18724, 18723, 18728, 18727, 18694, 18727, 18724, 18691, 18690, 18723, 18718, 18676, 18678, 18666, 18685, 18684, 18671, 18688, 18685, 18672, 18678, 18688, 18675, 18580, 18579, 18454, 18584, 18583, 18458, 18583, 18580, 18455, 18603, 18476, 18474, 18607, 18606, 18479, 18483, 18611, 18610, 18610, 18607, 18480, 18611, 18483, 18476, 18606, 18584, 18459, 18544, 18660, 18658, 18548, 18665, 18664, 18547, 18664, 18660, 18677, 18676, 18569, 18681, 18680, 18573, 18680, 18677, 18570, 18676, 18665, 18548,
18485, 18484, 18430, 18486, 18488, 18434, 18488, 18485, 18431, 18667, 18668, 18615, 18673, 18674, 18623, 18668, 18673, 18621, 18704, 18647, 18645, 18714, 18657, 18655, 18712, 18655, 18647, 18674, 18702, 18645, 18742, 18744, 18743, 18752, 18751, 18749, 18756, 18754, 18753, 18756, 18758, 18757, 18746, 18755, 18753, 18752, 18760, 18759, 18742, 18747, 18748, 18759, 18760, 18756, 18760, 18761, 18758, 18746, 18745, 18743, 18749, 18751, 18748, 18751, 18746, 18744, 18746, 18751, 18759, 18765, 18764, 18762, 18769, 18768, 18766, 18773, 18771, 18770, 18777, 18775, 18774, 18779, 18781, 18780, 18785, 18784, 18782, 18779, 18778, 18764, 18783, 18782, 18768, 18780, 18781, 18773, 18784, 18785, 18777, 18768, 18765, 18763, 18776, 18774, 18771, 18782, 18784, 18781, 18784, 18776, 18773, 18782, 18779, 18765, 18764, 18752, 18750, 18772, 18770, 18757, 18778, 18780, 18761, 18761, 18780, 18772, 18778, 18760, 18752, 18794, 18792, 18791, 18796, 18798, 18797, 18804, 18802, 18801, 18806, 18803, 18801, 18804, 18790, 18789, 18808, 18807, 18797, 18808, 18804, 18803, 18806, 18809, 18807, 18794, 18800, 18799, 18794, 18793, 18789, 18800, 18798, 18796, 18798, 18800, 18794, 18808, 18798, 18790, 18813, 18815, 18814, 18811, 18815, 18813, 18814, 18815, 18818, 18816, 18818, 18815,
18822, 18820, 18819, 18828, 18830, 18829, 18829, 18826, 18825, 18826, 18822, 18821, 18824, 18823, 18820, 18830, 18833, 18832, 18829, 18832, 18831, 18831, 18824, 18822, 18837, 18835, 18834, 18839, 18841, 18840, 18845, 18843, 18842, 18847, 18849, 18848, 18843, 18837, 18836, 18841, 18847, 18846, 18841, 18839, 18835, 18849, 18847, 18843, 18847, 18841, 18837, 18835, 18830, 18828, 18833, 18839, 18838, 18835, 18839, 18833, 18859, 18861, 18860, 18866, 18865, 18863, 18868, 18870, 18869, 18869, 18872, 18871, 18876, 18875, 18873, 18880, 18879, 18877, 18857, 18870, 18868, 18866, 18878, 18877, 18869, 18870, 18874, 18873, 18875, 18872, 18859, 18863, 18865, 18860, 18861, 18857, 18877, 18879, 18876, 18857, 18861, 18865, 18877, 18874, 18862, 18862, 18874, 18870, 18884, 18886, 18885, 18885, 18888, 18887, 18893, 18892, 18890, 18895, 18897, 18896, 18896, 18899, 18898, 18903, 18902, 18900, 18907, 18906, 18904, 18882, 18897, 18895, 18893, 18905, 18904, 18897, 18901, 18900, 18900, 18902, 18899, 18886, 18891, 18890, 18885, 18890, 18892, 18887, 18888, 18882, 18904, 18906, 18903, 18888, 18892, 18889, 18904, 18901, 18889, 18882, 18889, 18901, 18911, 18910, 18908, 18915, 18914, 18912, 18921, 18919, 18918, 18921, 18917, 18916, 18925, 18924, 18922, 18929, 18928, 18926,
18923, 18922, 18910, 18915, 18927, 18926, 18928, 18929, 18921, 18929, 18927, 18917, 18914, 18911, 18909, 18926, 18928, 18925, 18926, 18923, 18911, 18854, 18886, 18884, 18854, 18855, 18891, 18876, 18909, 18908, 18913, 18912, 18879, 18912, 18909, 18876, 18871, 18872, 18854, 18872, 18875, 18855, 18908, 18910, 18893, 18924, 18907, 18905, 18922, 18905, 18893, 18908, 18891, 18855, 18820, 18787, 18786, 18820, 18823, 18788, 18814, 18859, 18858, 18864, 18863, 18817, 18814, 18817, 18863, 18797, 18769, 18767, 18806, 18805, 18775, 18809, 18785, 18783, 18785, 18809, 18806, 18797, 18807, 18783, 18898, 18899, 18851, 18851, 18899, 18902, 18935, 18933, 18932, 18931, 18934, 18932, 18937, 18939, 18938, 18943, 18942, 18940, 18945, 18947, 18946, 18953, 18952, 18950, 18935, 18945, 18944, 18951, 18950, 18942, 18939, 18935, 18934, 18936, 18938, 18934, 18948, 18949, 18947, 18942, 18939, 18937, 18950, 18952, 18949, 18942, 18950, 18948, 18935, 18939, 18948, 18957, 18959, 18958, 18962, 18964, 18963, 18970, 18968, 18967, 18966, 18969, 18967, 18972, 18974, 18973, 18959, 18966, 18965, 18964, 18972, 18971, 18961, 18963, 18959, 18974, 18970, 18969, 18966, 18971, 18973, 18957, 18956, 18954, 18961, 18957, 18955, 18963, 18971, 18966, 18947, 18955, 18954, 18962, 18961, 18952,
18952, 18961, 18960, 18949, 18960, 18955, 18978, 18976, 18975, 18982, 18981, 18979, 18988, 18986, 18985, 18992, 18991, 18989, 18986, 18978, 18977, 18990, 18989, 18981, 18984, 18983, 18976, 18994, 18993, 18986, 18983, 18984, 18982, 18993, 18994, 18992, 18984, 18993, 18990, 18993, 18984, 18978, 18996, 18998, 18997, 19006, 19005, 19003, 19008, 19010, 19009, 19012, 19014, 19013, 19000, 19008, 19007, 19006, 19012, 19011, 18996, 19001, 19002, 19011, 19013, 19010, 19000, 18999, 18997, 19003, 19005, 19002, 19005, 19000, 18998, 19000, 19005, 19011, 19016, 19018, 19017, 19020, 19018, 19016, 19024, 19023, 19021, 19026, 19028, 19027, 19032, 19030, 19029, 19036, 19035, 19033, 19038, 19040, 19039, 19030, 19020, 19019, 19034, 19033, 19023, 19028, 19038, 19037, 19018, 19026, 19025, 19020, 19028, 19026, 19040, 19038, 19030, 19027, 19024, 19022, 19039, 19036, 19034, 19037, 19034, 19024, 19038, 19028, 19020, 19042, 19044, 19043, 19046, 19048, 19047, 19050, 19052, 19051, 19056, 19055, 19053, 19044, 19050, 19049, 19048, 19054, 19053, 19045, 19047, 19044, 19053, 19055, 19052, 19047, 19053, 19050, 19017, 18988, 18987, 19022, 19021, 18991, 18994, 19025, 19022, 19017, 19025, 18994, 19010, 19042, 19041, 19014, 19046, 19045, 19013, 19045, 19042, 19061, 19063, 19062,
19065, 19067, 19066, 19071, 19070, 19068, 19075, 19074, 19072, 19058, 19067, 19065, 19062, 19063, 19069, 19067, 19073, 19072, 19064, 19066, 19063, 19074, 19071, 19069, 19063, 19066, 19072, 19058, 19059, 19073, 19077, 19079, 19078, 19083, 19085, 19084, 19081, 19085, 19083, 19094, 19093, 19091, 19098, 19097, 19095, 19092, 19091, 19086, 19090, 19096, 19095, 19078, 19079, 19087, 19084, 19085, 19090, 19088, 19090, 19085, 19082, 19084, 19079, 19097, 19094, 19092, 19079, 19084, 19089, 19095, 19092, 19087, 19104, 19103, 19101, 19106, 19108, 19107, 19112, 19110, 19109, 19116, 19114, 19113, 19116, 19100, 19099, 19118, 19117, 19103, 19120, 19119, 19107, 19117, 19118, 19112, 19120, 19116, 19115, 19107, 19104, 19102, 19115, 19113, 19110, 19118, 19119, 19115, 19119, 19118, 19104, 19116, 19120, 19108, 19124, 19122, 19121, 19128, 19126, 19125, 19132, 19131, 19129, 19136, 19135, 19133, 19131, 19132, 19124, 19135, 19136, 19128, 19127, 19125, 19122, 19135, 19132, 19130, 19132, 19135, 19127, 19102, 19101, 19070, 19106, 19105, 19074, 19105, 19102, 19071, 19130, 19129, 19093, 19134, 19133, 19097, 19133, 19130, 19094, 19076, 19078, 19058, 19091, 19093, 19075, 19078, 19086, 19059, 19091, 19073, 19059, 19123, 19121, 19099, 19129, 19131, 19108, 19131, 19123, 19100,
19129, 19106, 19075, 19140, 19138, 19137, 19144, 19143, 19141, 19148, 19147, 19145, 19152, 19150, 19149, 19156, 19155, 19153, 19160, 19159, 19157, 19150, 19140, 19139, 19154, 19153, 19143, 19158, 19157, 19147, 19148, 19146, 19138, 19160, 19158, 19150, 19147, 19144, 19142, 19159, 19156, 19154, 19157, 19154, 19144, 19158, 19148, 19140, 19162, 19164, 19163, 19168, 19167, 19165, 19170, 19172, 19171, 19174, 19176, 19175, 19164, 19170, 19169, 19168, 19174, 19173, 19162, 19165, 19167, 19170, 19173, 19175, 19164, 19167, 19173, 19182, 19180, 19179, 19186, 19185, 19183, 19188, 19190, 19189, 19192, 19194, 19193, 19198, 19196, 19195, 19202, 19201, 19199, 19206, 19205, 19203, 19178, 19192, 19191, 19196, 19182, 19181, 19186, 19200, 19199, 19190, 19204, 19203, 19199, 19201, 19194, 19182, 19190, 19188, 19198, 19206, 19204, 19189, 19186, 19184, 19205, 19202, 19200, 19203, 19200, 19186, 19185, 19199, 19192, 19196, 19204, 19190, 19208, 19210, 19209, 19214, 19213, 19211, 19218, 19220, 19219, 19225, 19224, 19222, 19210, 19218, 19217, 19223, 19222, 19213, 19208, 19211, 19213, 19222, 19224, 19220, 19219, 19220, 19216, 19216, 19220, 19224, 19210, 19213, 19222, 19180, 19152, 19151, 19184, 19183, 19155, 19160, 19188, 19187, 19187, 19184, 19156, 19180, 19188, 19160,
19172, 19208, 19207, 19212, 19211, 19175, 19172, 19175, 19211, 18970, 19081, 19080, 18970, 18974, 19088, 19138, 19032, 19031, 19142, 19141, 19035, 19040, 19146, 19145, 19145, 19142, 19036, 19146, 19040, 19032, 19052, 19162, 19161, 19166, 19165, 19055, 19055, 19165, 19162, 18981, 18943, 18941, 18991, 18953, 18951, 18989, 18951, 18943, 19023, 18964, 18962, 19035, 18974, 18972, 19033, 18972, 18964, 19021, 18962, 18953, 19141, 19143, 19090, 19155, 19098, 19096, 19143, 19153, 19096, 19178, 19177, 19126, 19185, 19136, 19134, 19136, 19185, 19178, 19183, 19134, 19098, 19035, 19141, 19088, 19229, 19227, 19226, 19231, 19233, 19232, 19239, 19241, 19240, 19240, 19237, 19236, 19245, 19244, 19242, 19235, 19229, 19228, 19233, 19243, 19242, 19233, 19231, 19227, 19240, 19241, 19248, 19240, 19247, 19246, 19236, 19237, 19235, 19246, 19247, 19245, 19237, 19246, 19243, 19235, 19243, 19233, 19252, 19254, 19253, 19253, 19256, 19255, 19261, 19260, 19258, 19263, 19265, 19264, 19264, 19267, 19266, 19271, 19270, 19268, 19275, 19274, 19272, 19250, 19265, 19263, 19261, 19273, 19272, 19265, 19269, 19268, 19264, 19268, 19270, 19253, 19254, 19259, 19258, 19260, 19256, 19255, 19256, 19250, 19272, 19274, 19271, 19256, 19260, 19257, 19272, 19269, 19257, 19250, 19257, 19269,
19279, 19278, 19276, 19281, 19283, 19282, 19287, 19286, 19284, 19291, 19290, 19288, 19285, 19284, 19278, 19289, 19288, 19282, 19282, 19279, 19277, 19290, 19287, 19285, 19288, 19285, 19279, 19295, 19294, 19292, 19299, 19298, 19296, 19303, 19301, 19300, 19307, 19305, 19304, 19309, 19306, 19304, 19313, 19312, 19310, 19311, 19310, 19294, 19315, 19314, 19298, 19312, 19313, 19303, 19314, 19315, 19307, 19309, 19316, 19314, 19298, 19295, 19293, 19309, 19308, 19301, 19316, 19313, 19311, 19313, 19316, 19309, 19298, 19314, 19311, 19277, 19276, 19244, 19248, 19281, 19280, 19247, 19280, 19277, 19271, 19293, 19292, 19297, 19296, 19274, 19296, 19293, 19271, 19266, 19267, 19241, 19241, 19267, 19270, 19294, 19283, 19281, 19310, 19312, 19291, 19310, 19289, 19283, 19292, 19281, 19248, 19320, 19319, 19317, 19324, 19323, 19321, 19330, 19328, 19327, 19330, 19326, 19325, 19334, 19333, 19331, 19338, 19337, 19335, 19332, 19331, 19319, 19324, 19336, 19335, 19338, 19330, 19329, 19330, 19338, 19336, 19323, 19320, 19318, 19335, 19337, 19334, 19335, 19332, 19320, 19344, 19342, 19341, 19344, 19340, 19339, 19350, 19349, 19347, 19353, 19343, 19341, 19350, 19344, 19343, 19344, 19350, 19348, 19346, 19345, 19339, 19348, 19351, 19346, 19353, 19354, 19349, 19360, 19358, 19357,
19360, 19356, 19355, 19366, 19365, 19363, 19369, 19359, 19357, 19365, 19366, 19360, 19360, 19366, 19364, 19362, 19361, 19355, 19367, 19362, 19356, 19370, 19365, 19359, 19362, 19329, 19327, 19334, 19364, 19363, 19367, 19364, 19334, 19362, 19367, 19337, 19347, 19349, 19324, 19353, 19352, 19325, 19354, 19353, 19326, 19354, 19336, 19324, 19376, 19374, 19373, 19378, 19375, 19373, 19380, 19382, 19381, 19389, 19387, 19386, 19393, 19392, 19390, 19387, 19372, 19371, 19391, 19390, 19381, 19389, 19393, 19391, 19384, 19376, 19375, 19378, 19385, 19383, 19378, 19377, 19371, 19383, 19385, 19382, 19385, 19378, 19372, 19387, 19391, 19382, 19400, 19398, 19397, 19404, 19403, 19401, 19408, 19406, 19405, 19410, 19407, 19405, 19408, 19400, 19399, 19412, 19411, 19403, 19400, 19404, 19402, 19411, 19412, 19408, 19413, 19411, 19407, 19412, 19404, 19400, 19398, 19389, 19388, 19402, 19401, 19392, 19398, 19402, 19393, 19395, 19394, 19374, 19384, 19396, 19395, 19380, 19379, 19286, 19384, 19383, 19290, 19290, 19383, 19380, 19395, 19302, 19300, 19395, 19396, 19312, 19396, 19384, 19291, 19317, 19319, 19261, 19333, 19275, 19273, 19331, 19273, 19261, 19363, 19365, 19299, 19369, 19368, 19305, 19315, 19370, 19369, 19370, 19315, 19299, 19363, 19297, 19275, 18649, 18931, 18930,
18653, 18937, 18936, 18941, 18940, 18656, 18940, 18937, 18653, 18652, 18936, 18931, 18976, 18711, 18710, 18980, 18979, 18714, 18716, 18983, 18980, 18976, 18983, 18716, 18734, 18996, 18995, 19004, 19003, 18739, 18739, 19003, 19001, 18736, 19001, 18996, 18979, 18941, 18657, 19227, 18845, 18844, 18849, 19231, 19230, 19231, 18849, 18845, 18851, 19254, 19252, 18852, 19259, 19254, 18903, 19318, 19317, 19322, 19321, 18906, 19321, 19318, 18903, 19346, 18920, 18918, 18925, 19348, 19347, 19351, 19348, 18925, 19351, 18928, 18920, 19347, 19322, 18907, 19317, 19259, 18852, 18786, 18787, 18566, 18787, 18788, 18574, 18823, 18824, 18683, 18831, 18832, 18695, 18824, 18831, 18693, 18840, 18730, 18728, 18848, 18740, 18738, 18846, 18738, 18730, 18838, 18728, 18695, 18788, 18823, 18681, 19232, 19006, 19004, 19244, 19014, 19012, 19242, 19012, 19006, 19278, 19048, 19046, 19286, 19056, 19054, 19284, 19054, 19048, 19276, 19046, 19014, 19381, 19168, 19166, 19392, 19176, 19174, 19390, 19174, 19168, 19403, 19214, 19212, 19413, 19225, 19223, 19403, 19411, 19223, 19401, 19212, 19176, 19379, 19166, 19056, 19230, 19004, 18740, 19415, 19417, 19416, 19421, 19420, 19418, 19423, 19425, 19424, 19431, 19433, 19432, 19417, 19427, 19426, 19431, 19430, 19420, 19423, 19428, 19429,
19415, 19418, 19420, 19426, 19427, 19423, 19432, 19429, 19428, 19423, 19427, 19430, 19420, 19430, 19427, 19439, 19438, 19436, 19441, 19443, 19442, 19449, 19447, 19446, 19449, 19445, 19444, 19453, 19452, 19450, 19457, 19456, 19454, 19445, 19435, 19434, 19451, 19450, 19438, 19455, 19454, 19442, 19456, 19457, 19449, 19457, 19455, 19445, 19442, 19439, 19437, 19454, 19456, 19453, 19454, 19451, 19439, 19445, 19455, 19443, 19459, 19461, 19460, 19460, 19463, 19462, 19467, 19466, 19464, 19471, 19470, 19468, 19475, 19477, 19476, 19481, 19480, 19478, 19463, 19473, 19472, 19475, 19474, 19466, 19479, 19478, 19470, 19460, 19461, 19465, 19460, 19464, 19466, 19468, 19470, 19467, 19480, 19477, 19475, 19478, 19475, 19467, 19463, 19466, 19474, 19487, 19485, 19484, 19489, 19491, 19490, 19499, 19497, 19496, 19499, 19495, 19494, 19503, 19502, 19500, 19495, 19483, 19482, 19501, 19500, 19490, 19502, 19503, 19499, 19499, 19503, 19501, 19487, 19493, 19492, 19487, 19486, 19482, 19493, 19491, 19489, 19491, 19493, 19487, 19495, 19501, 19491, 19425, 19461, 19459, 19469, 19468, 19432, 19468, 19465, 19429, 19425, 19429, 19465, 19485, 19448, 19446, 19489, 19488, 19452, 19492, 19489, 19453, 19492, 19456, 19448, 19438, 19421, 19419, 19452, 19433, 19431, 19450, 19431, 19421,
19490, 19471, 19469, 19500, 19502, 19481, 19500, 19479, 19471, 19488, 19469, 19433, 19507, 19505, 19504, 19507, 19509, 19508, 19515, 19514, 19512, 19517, 19519, 19518, 19523, 19521, 19520, 19525, 19522, 19520, 19529, 19528, 19526, 19523, 19511, 19510, 19527, 19526, 19514, 19531, 19530, 19518, 19530, 19531, 19523, 19525, 19532, 19530, 19518, 19515, 19513, 19532, 19529, 19527, 19518, 19530, 19527, 19531, 19519, 19511, 19536, 19534, 19533, 19536, 19538, 19537, 19544, 19542, 19541, 19544, 19540, 19539, 19550, 19549, 19547, 19553, 19543, 19541, 19549, 19550, 19544, 19550, 19548, 19540, 19546, 19545, 19539, 19548, 19551, 19546, 19554, 19549, 19543, 19546, 19525, 19524, 19529, 19548, 19547, 19551, 19548, 19529, 19546, 19551, 19532, 19536, 19535, 19510, 19538, 19519, 19517, 19519, 19538, 19536, 19513, 19512, 19476, 19517, 19516, 19480, 19516, 19513, 19477, 19534, 19498, 19496, 19534, 19537, 19502, 19537, 19517, 19481, 19507, 19506, 19434, 19509, 19443, 19441, 19443, 19509, 19507, 18406, 17600, 17599, 18411, 18410, 17603, 17606, 18414, 18411, 18414, 17606, 17600, 17627, 18429, 18428, 17632, 18433, 18429, 18487, 18486, 17672, 18491, 18490, 17676, 18490, 18487, 17673, 18509, 17686, 17684, 17689, 18515, 18514, 18515, 17689, 17686, 18514, 18491, 17677,
17672, 18486, 18433, 17761, 18742, 18741, 18750, 18749, 17766, 17766, 18749, 18747, 17763, 18747, 18742, 18763, 18762, 17778, 18767, 18766, 17782, 18766, 18763, 17779, 18762, 18750, 17767, 18792, 17821, 17820, 18796, 18795, 17824, 18799, 18796, 17825, 18792, 18799, 17827, 18795, 18767, 17783, 18341, 19415, 19414, 19419, 19418, 18344, 18341, 18344, 19418, 19437, 19436, 18358, 19441, 19440, 18362, 19440, 19437, 18359, 18358, 19436, 19419, 19505, 18397, 18396, 19505, 19508, 18400, 19508, 19441, 18363, 18119, 18120, 17695, 18127, 17702, 17700, 18120, 18125, 17700, 18163, 17712, 17710, 18171, 17722, 17720, 18169, 17720, 17712, 18161, 17710, 17702, 18276, 18275, 17790, 18277, 18278, 17796, 18291, 18289, 17802, 18281, 18276, 17792, 17796, 18278, 18283, 17804, 18297, 18291, 18283, 18297, 17804, 18281, 17799, 17722, 19472, 19473, 18811, 19476, 18818, 18816, 18811, 19473, 19474, 19514, 18866, 18864, 19528, 18880, 18878, 19526, 18878, 18866, 19547, 19549, 18915, 19553, 19552, 18916, 18927, 19554, 19553, 19554, 18927, 18915, 19547, 18913, 18880, 19476, 19512, 18864, 17256, 17255, 16698, 17260, 17259, 16702, 17259, 17256, 16699, 17280, 17279, 16714, 17284, 17283, 16718, 17283, 17280, 16715, 17279, 17260, 16703, 17315, 16761, 16760, 17321, 17320, 16764,
17325, 17324, 16768, 17324, 17321, 16765, 17325, 16769, 16761, 17320, 17284, 16719, 17134, 17830, 17828, 17138, 17843, 17842, 17142, 17846, 17845, 17845, 17843, 17138, 17134, 17137, 17842, 17861, 17860, 17157, 17865, 17864, 17161, 17864, 17861, 17158, 17860, 17846, 17142, 17917, 17916, 17218, 17921, 17920, 17222, 17920, 17917, 17219, 17937, 17241, 17239, 17941, 17940, 17245, 17944, 17941, 17246, 17937, 17944, 17249, 17940, 17921, 17223, 17916, 17865, 17162, 19556, 19558, 19557, 19562, 19564, 19563, 19558, 19562, 19561, 19559, 19560, 19558, 19562, 19565, 19566, 19558, 19560, 19565, 19568, 19570, 19569, 19576, 19574, 19573, 19576, 19572, 19571, 19580, 19579, 19577, 19584, 19583, 19581, 19586, 19588, 19587, 19592, 19590, 19589, 19596, 19595, 19593, 19570, 19586, 19585, 19590, 19575, 19573, 19580, 19594, 19593, 19568, 19577, 19579, 19586, 19593, 19595, 19583, 19584, 19576, 19584, 19582, 19572, 19592, 19598, 19597, 19581, 19583, 19580, 19598, 19596, 19594, 19597, 19594, 19580, 19570, 19579, 19593, 19597, 19583, 19575, 19600, 19602, 19601, 19608, 19610, 19609, 19614, 19616, 19615, 19612, 19616, 19614, 19620, 19618, 19617, 19627, 19626, 19624, 19602, 19612, 19611, 19618, 19604, 19603, 19625, 19624, 19609, 19600, 19605, 19606, 19615, 19616, 19623,
19621, 19623, 19616, 19620, 19627, 19625, 19609, 19606, 19605, 19626, 19623, 19621, 19624, 19621, 19606, 19602, 19606, 19621, 19625, 19610, 19604, 19635, 19633, 19632, 19637, 19639, 19638, 19645, 19643, 19642, 19647, 19649, 19648, 19653, 19652, 19650, 19643, 19631, 19630, 19651, 19650, 19638, 19645, 19653, 19651, 19641, 19640, 19633, 19654, 19655, 19649, 19635, 19634, 19630, 19641, 19639, 19637, 19639, 19641, 19635, 19643, 19651, 19639, 19657, 19659, 19658, 19663, 19661, 19660, 19665, 19662, 19660, 19669, 19668, 19666, 19678, 19676, 19674, 19682, 19684, 19683, 19659, 19675, 19673, 19680, 19665, 19664, 19669, 19682, 19681, 19657, 19666, 19668, 19678, 19684, 19682, 19670, 19671, 19663, 19665, 19672, 19670, 19672, 19669, 19667, 19680, 19679, 19674, 19682, 19685, 19680, 19685, 19682, 19669, 19668, 19681, 19675, 19680, 19685, 19672, 19588, 19629, 19628, 19633, 19592, 19591, 19637, 19636, 19595, 19640, 19637, 19596, 19595, 19636, 19629, 19633, 19640, 19598, 19615, 19657, 19656, 19663, 19620, 19619, 19667, 19666, 19622, 19671, 19670, 19626, 19626, 19670, 19667, 19622, 19666, 19657, 19663, 19671, 19627, 19673, 19675, 19647, 19683, 19655, 19654, 19647, 19675, 19681, 19689, 19691, 19690, 19687, 19691, 19689, 19697, 19701, 19699, 19702, 19698, 19696,
19706, 19705, 19703, 19690, 19697, 19695, 19694, 19704, 19703, 19690, 19691, 19694, 19687, 19692, 19694, 19697, 19703, 19705, 19702, 19706, 19704, 19690, 19693, 19703, 19710, 19708, 19707, 19711, 19712, 19710, 19701, 19709, 19707, 19710, 19702, 19700, 19706, 19712, 19711, 19701, 19705, 19711, 19710, 19712, 19706, 19716, 19714, 19713, 19718, 19715, 19713, 19720, 19722, 19721, 19721, 19725, 19723, 19731, 19733, 19732, 19737, 19739, 19738, 19747, 19745, 19744, 19747, 19743, 19741, 19752, 19754, 19753, 19736, 19718, 19717, 19725, 19742, 19740, 19743, 19726, 19724, 19752, 19751, 19732, 19727, 19728, 19716, 19718, 19729, 19727, 19737, 19749, 19750, 19721, 19722, 19731, 19721, 19730, 19732, 19753, 19754, 19747, 19747, 19754, 19752, 19740, 19742, 19737, 19751, 19753, 19750, 19737, 19742, 19751, 19736, 19748, 19729, 19725, 19732, 19751, 19743, 19752, 19733, 19758, 19756, 19755, 19760, 19757, 19755, 19761, 19762, 19758, 19763, 19761, 19757, 19765, 19767, 19766, 19771, 19769, 19768, 19766, 19770, 19768, 19771, 19767, 19765, 19777, 19776, 19774, 19766, 19767, 19775, 19776, 19777, 19771, 19774, 19776, 19770, 19777, 19775, 19767, 19773, 19772, 19765, 19775, 19778, 19773, 19739, 19767, 19765, 19773, 19746, 19744, 19778, 19775, 19750, 19739, 19750, 19775,
19778, 19753, 19746, 19760, 19759, 19724, 19761, 19763, 19733, 19733, 19763, 19760, 19564, 19687, 19686, 19564, 19566, 19692, 19716, 19645, 19644, 19649, 19722, 19720, 19728, 19727, 19652, 19716, 19728, 19653, 19649, 19655, 19731, 19758, 19678, 19677, 19762, 19761, 19683, 19758, 19762, 19684, 19683, 19761, 19731, 19628, 19629, 19556, 19638, 19560, 19559, 19652, 19566, 19565, 19556, 19629, 19636, 19650, 19565, 19560, 19729, 19694, 19692, 19736, 19734, 19696, 19704, 19748, 19736, 19729, 19748, 19704, 19652, 19727, 19692, 19782, 19780, 19779, 19784, 19783, 19780, 19786, 19788, 19787, 19790, 19792, 19791, 19800, 19799, 19797, 19804, 19803, 19801, 19798, 19797, 19793, 19802, 19801, 19795, 19788, 19794, 19793, 19792, 19796, 19795, 19789, 19791, 19788, 19803, 19800, 19798, 19791, 19795, 19794, 19801, 19798, 19794, 19806, 19808, 19807, 19810, 19812, 19811, 19820, 19819, 19817, 19824, 19823, 19821, 19818, 19817, 19813, 19822, 19821, 19815, 19808, 19814, 19813, 19812, 19816, 19815, 19809, 19811, 19808, 19823, 19820, 19818, 19811, 19815, 19814, 19821, 19818, 19814, 19828, 19827, 19825, 19832, 19831, 19829, 19836, 19834, 19833, 19840, 19838, 19837, 19842, 19841, 19827, 19844, 19843, 19831, 19842, 19836, 19835, 19844, 19840, 19839, 19831, 19828, 19826,
19839, 19837, 19834, 19843, 19839, 19836, 19843, 19842, 19828, 19848, 19847, 19845, 19852, 19851, 19849, 19854, 19856, 19855, 19860, 19859, 19857, 19864, 19863, 19861, 19858, 19857, 19847, 19862, 19861, 19851, 19857, 19859, 19856, 19851, 19848, 19846, 19863, 19860, 19858, 19861, 19858, 19848, 19826, 19825, 19799, 19830, 19829, 19803, 19829, 19826, 19800, 19846, 19845, 19819, 19850, 19849, 19823, 19849, 19846, 19820, 19805, 19807, 19792, 19819, 19804, 19802, 19807, 19813, 19796, 19817, 19802, 19796, 19847, 19832, 19830, 19854, 19853, 19838, 19857, 19854, 19840, 19847, 19857, 19844, 19845, 19830, 19804, 19866, 19868, 19867, 19870, 19872, 19871, 19880, 19879, 19877, 19884, 19883, 19881, 19878, 19877, 19873, 19882, 19881, 19875, 19868, 19874, 19873, 19872, 19876, 19875, 19869, 19871, 19868, 19883, 19880, 19878, 19871, 19875, 19874, 19881, 19878, 19874, 19886, 19888, 19887, 19890, 19892, 19891, 19900, 19899, 19897, 19904, 19903, 19901, 19898, 19897, 19893, 19902, 19901, 19895, 19888, 19894, 19893, 19892, 19896, 19895, 19889, 19891, 19888, 19903, 19900, 19898, 19891, 19895, 19894, 19901, 19898, 19894, 19908, 19907, 19905, 19912, 19911, 19909, 19916, 19914, 19913, 19918, 19915, 19913, 19922, 19921, 19919, 19920, 19919, 19907, 19924, 19923, 19911,
19924, 19916, 19915, 19925, 19923, 19915, 19911, 19908, 19906, 19923, 19925, 19922, 19911, 19923, 19920, 19929, 19928, 19926, 19933, 19932, 19930, 19937, 19935, 19934, 19941, 19939, 19938, 19943, 19942, 19928, 19945, 19944, 19932, 19943, 19937, 19936, 19945, 19941, 19940, 19932, 19929, 19927, 19940, 19938, 19935, 19944, 19940, 19937, 19944, 19943, 19929, 19906, 19905, 19879, 19910, 19909, 19883, 19909, 19906, 19880, 19927, 19926, 19899, 19931, 19930, 19903, 19930, 19927, 19900, 19885, 19887, 19872, 19899, 19884, 19882, 19887, 19893, 19876, 19897, 19882, 19876, 19928, 19912, 19910, 19936, 19934, 19914, 19942, 19936, 19916, 19942, 19924, 19912, 19926, 19910, 19884, 19865, 19867, 19812, 19879, 19824, 19822, 19867, 19873, 19816, 19877, 19822, 19816, 19907, 19852, 19850, 19921, 19864, 19862, 19919, 19862, 19852, 19905, 19850, 19824, 19947, 19949, 19948, 19951, 19953, 19952, 19961, 19960, 19958, 19965, 19964, 19962, 19959, 19958, 19954, 19963, 19962, 19956, 19949, 19955, 19954, 19953, 19957, 19956, 19950, 19952, 19949, 19964, 19961, 19959, 19952, 19956, 19955, 19962, 19959, 19955, 19967, 19969, 19968, 19971, 19973, 19972, 19981, 19979, 19978, 19985, 19984, 19982, 19989, 19988, 19986, 19979, 19973, 19971, 19983, 19982, 19974, 19987, 19986, 19976,
19969, 19975, 19974, 19973, 19977, 19976, 19989, 19987, 19979, 19970, 19972, 19969, 19988, 19985, 19983, 19972, 19976, 19975, 19986, 19983, 19975, 19987, 19977, 19973, 19993, 19992, 19990, 19997, 19996, 19994, 20003, 20001, 20000, 19999, 20002, 20000, 20005, 20007, 20006, 20011, 20010, 20008, 20005, 20004, 19992, 20009, 20008, 19996, 20006, 20007, 20003, 20004, 20006, 20002, 19996, 19993, 19991, 20010, 20007, 20005, 20008, 20005, 19993, 20015, 20013, 20012, 20019, 20018, 20016, 20023, 20022, 20020, 20027, 20025, 20024, 20031, 20030, 20028, 20035, 20034, 20032, 20025, 20015, 20014, 20029, 20028, 20018, 20033, 20032, 20022, 20023, 20021, 20013, 20035, 20033, 20025, 20022, 20019, 20017, 20034, 20031, 20029, 20032, 20029, 20019, 20033, 20023, 20015, 19991, 19990, 19960, 19995, 19994, 19964, 19994, 19991, 19961, 20013, 19981, 19980, 20017, 20016, 19984, 20021, 20020, 19988, 20020, 20017, 19985, 20021, 19989, 19981, 19966, 19968, 19953, 19984, 19965, 19963, 19968, 19974, 19957, 19982, 19963, 19957, 20018, 19997, 19995, 20030, 20011, 20009, 20028, 20009, 19997, 20016, 19995, 19965, 19782, 19781, 19603, 19784, 19610, 19608, 19610, 19784, 19782, 19946, 19948, 19892, 19960, 19904, 19902, 19948, 19954, 19896, 19958, 19902, 19896, 19992, 19933, 19931,
19999, 19998, 19939, 19945, 20004, 19999, 19992, 20004, 19945, 19990, 19931, 19904, 20037, 20039, 20038, 20043, 20042, 20040, 20047, 20046, 20044, 20049, 20051, 20050, 20055, 20054, 20052, 20059, 20058, 20056, 20039, 20049, 20048, 20053, 20052, 20042, 20057, 20056, 20046, 20040, 20042, 20039, 20052, 20054, 20051, 20046, 20043, 20041, 20058, 20055, 20053, 20056, 20053, 20043, 20042, 20052, 20049, 20061, 20063, 20062, 20067, 20066, 20064, 20071, 20070, 20068, 20073, 20075, 20074, 20079, 20078, 20076, 20083, 20082, 20080, 20063, 20073, 20072, 20077, 20076, 20066, 20081, 20080, 20070, 20064, 20066, 20063, 20076, 20078, 20075, 20070, 20067, 20065, 20082, 20079, 20077, 20080, 20077, 20067, 20066, 20076, 20073, 20051, 20061, 20060, 20065, 20064, 20054, 20069, 20068, 20058, 20068, 20065, 20055, 20054, 20064, 20061, 20087, 20085, 20084, 20091, 20090, 20088, 20097, 20095, 20094, 20101, 20100, 20098, 20095, 20087, 20086, 20099, 20098, 20090, 20093, 20092, 20085, 20103, 20102, 20095, 20093, 20091, 20089, 20103, 20101, 20099, 20102, 20099, 20091, 20102, 20093, 20087, 20107, 20105, 20104, 20111, 20110, 20108, 20117, 20115, 20114, 20121, 20120, 20118, 20115, 20107, 20106, 20119, 20118, 20110, 20113, 20112, 20105, 20123, 20122, 20115, 20113, 20111, 20109,
20123, 20121, 20119, 20122, 20119, 20111, 20122, 20113, 20107, 20105, 20097, 20096, 20109, 20108, 20100, 20112, 20109, 20101, 20112, 20103, 20097, 20125, 20127, 20126, 20131, 20130, 20128, 20135, 20134, 20132, 20137, 20139, 20138, 20143, 20142, 20140, 20147, 20146, 20144, 20127, 20137, 20136, 20141, 20140, 20130, 20145, 20144, 20134, 20128, 20130, 20127, 20140, 20142, 20139, 20134, 20131, 20129, 20146, 20143, 20141, 20144, 20141, 20131, 20130, 20140, 20137, 20149, 20151, 20150, 20155, 20154, 20152, 20159, 20158, 20156, 20161, 20163, 20162, 20167, 20166, 20164, 20171, 20170, 20168, 20151, 20161, 20160, 20165, 20164, 20154, 20169, 20168, 20158, 20152, 20154, 20151, 20164, 20166, 20163, 20158, 20155, 20153, 20170, 20167, 20165, 20168, 20165, 20155, 20154, 20164, 20161, 20139, 20149, 20148, 20153, 20152, 20142, 20157, 20156, 20146, 20156, 20153, 20143, 20142, 20152, 20149, 20175, 20173, 20172, 20179, 20178, 20176, 20185, 20183, 20182, 20189, 20188, 20186, 20183, 20175, 20174, 20187, 20186, 20178, 20181, 20180, 20173, 20191, 20190, 20183, 20181, 20179, 20177, 20191, 20189, 20187, 20190, 20187, 20179, 20190, 20181, 20175, 20195, 20193, 20192, 20199, 20198, 20196, 20205, 20203, 20202, 20209, 20208, 20206, 20203, 20195, 20194, 20207, 20206, 20198,
20201, 20200, 20193, 20211, 20210, 20203, 20201, 20199, 20197, 20211, 20209, 20207, 20210, 20207, 20199, 20210, 20201, 20195, 20193, 20185, 20184, 20197, 20196, 20188, 20200, 20197, 20189, 20200, 20191, 20185, 20075, 20125, 20124, 20129, 20128, 20078, 20133, 20132, 20082, 20132, 20129, 20079, 20078, 20128, 20125, 20173, 20117, 20116, 20177, 20176, 20120, 20180, 20177, 20121, 20180, 20123, 20117, 20090, 20047, 20045, 20100, 20059, 20057, 20098, 20057, 20047, 20110, 20071, 20069, 20120, 20083, 20081, 20118, 20081, 20071, 20108, 20069, 20059, 20178, 20135, 20133, 20188, 20147, 20145, 20186, 20145, 20135, 20198, 20159, 20157, 20208, 20171, 20169, 20206, 20169, 20159, 20196, 20157, 20147, 20176, 20133, 20083, 20215, 20217, 20216, 20220, 20222, 20221, 20224, 20226, 20225, 20230, 20229, 20227, 20217, 20224, 20223, 20222, 20228, 20227, 20215, 20219, 20221, 20227, 20229, 20226, 20215, 20214, 20212, 20219, 20215, 20213, 20221, 20227, 20224, 20234, 20232, 20231, 20238, 20237, 20235, 20242, 20241, 20239, 20246, 20244, 20243, 20250, 20249, 20247, 20254, 20253, 20251, 20244, 20234, 20233, 20248, 20247, 20237, 20252, 20251, 20241, 20242, 20240, 20232, 20254, 20252, 20244, 20241, 20238, 20236, 20253, 20250, 20248, 20251, 20248, 20238, 20252, 20242, 20234,
20256, 20258, 20257, 20262, 20261, 20259, 20264, 20266, 20265, 20270, 20269, 20267, 20258, 20264, 20263, 20268, 20267, 20261, 20259, 20261, 20258, 20267, 20269, 20266, 20261, 20267, 20264, 20274, 20272, 20271, 20278, 20277, 20275, 20282, 20281, 20279, 20286, 20284, 20283, 20290, 20289, 20287, 20294, 20293, 20291, 20284, 20274, 20273, 20288, 20287, 20277, 20292, 20291, 20281, 20282, 20280, 20272, 20294, 20292, 20284, 20281, 20278, 20276, 20293, 20290, 20288, 20291, 20288, 20278, 20292, 20282, 20274, 20226, 20256, 20255, 20260, 20259, 20229, 20229, 20259, 20256, 20272, 20246, 20245, 20276, 20275, 20249, 20280, 20279, 20253, 20279, 20276, 20250, 20280, 20254, 20246, 20237, 20222, 20220, 20249, 20230, 20228, 20247, 20228, 20222, 20277, 20262, 20260, 20289, 20270, 20268, 20287, 20268, 20262, 20275, 20260, 20230, 20296, 20298, 20297, 20302, 20301, 20299, 20304, 20306, 20305, 20310, 20309, 20307, 20298, 20304, 20303, 20308, 20307, 20301, 20299, 20301, 20298, 20307, 20309, 20306, 20301, 20307, 20304, 20314, 20312, 20311, 20318, 20317, 20315, 20322, 20321, 20319, 20326, 20324, 20323, 20330, 20329, 20327, 20334, 20333, 20331, 20324, 20314, 20313, 20328, 20327, 20317, 20332, 20331, 20321, 20322, 20320, 20312, 20334, 20332, 20324, 20321, 20318, 20316,
20333, 20330, 20328, 20331, 20328, 20318, 20332, 20322, 20314, 20336, 20338, 20337, 20342, 20341, 20339, 20344, 20346, 20345, 20350, 20349, 20347, 20338, 20344, 20343, 20348, 20347, 20341, 20339, 20341, 20338, 20347, 20349, 20346, 20341, 20347, 20344, 20354, 20352, 20351, 20358, 20357, 20355, 20362, 20361, 20359, 20366, 20364, 20363, 20370, 20369, 20367, 20374, 20373, 20371, 20364, 20354, 20353, 20368, 20367, 20357, 20372, 20371, 20361, 20362, 20360, 20352, 20374, 20372, 20364, 20361, 20358, 20356, 20373, 20370, 20368, 20371, 20368, 20358, 20372, 20362, 20354, 20306, 20336, 20335, 20340, 20339, 20309, 20309, 20339, 20336, 20352, 20326, 20325, 20356, 20355, 20329, 20360, 20359, 20333, 20359, 20356, 20330, 20360, 20334, 20326, 20317, 20302, 20300, 20329, 20310, 20308, 20327, 20308, 20302, 20357, 20342, 20340, 20369, 20350, 20348, 20367, 20348, 20342, 20355, 20340, 20310, 20266, 20296, 20295, 20300, 20299, 20269, 20269, 20299, 20296, 20312, 20286, 20285, 20316, 20315, 20289, 20320, 20319, 20293, 20319, 20316, 20290, 20320, 20294, 20286, 20315, 20300, 20270, 20376, 20378, 20377, 20382, 20381, 20379, 20386, 20385, 20383, 20388, 20390, 20389, 20394, 20393, 20391, 20398, 20397, 20395, 20378, 20388, 20387, 20392, 20391, 20381, 20396, 20395, 20385,
20379, 20381, 20378, 20391, 20393, 20390, 20385, 20382, 20380, 20397, 20394, 20392, 20395, 20392, 20382, 20381, 20391, 20388, 20402, 20400, 20399, 20406, 20404, 20403, 20408, 20402, 20401, 20410, 20406, 20405, 20405, 20403, 20400, 20409, 20405, 20402, 20390, 20401, 20399, 20408, 20407, 20393, 20410, 20409, 20397, 20409, 20408, 20394, 20393, 20407, 20401, 20414, 20412, 20411, 20418, 20417, 20415, 20424, 20422, 20421, 20428, 20427, 20425, 20422, 20414, 20413, 20426, 20425, 20417, 20420, 20419, 20412, 20430, 20429, 20422, 20420, 20418, 20416, 20430, 20428, 20426, 20429, 20426, 20418, 20429, 20420, 20414, 20434, 20432, 20431, 20438, 20434, 20433, 20436, 20435, 20432, 20439, 20436, 20434, 20436, 20424, 20423, 20438, 20437, 20427, 20439, 20438, 20428, 20439, 20430, 20424, 20441, 20443, 20442, 20445, 20447, 20446, 20455, 20454, 20452, 20459, 20458, 20456, 20453, 20452, 20448, 20457, 20456, 20450, 20443, 20449, 20448, 20447, 20451, 20450, 20444, 20446, 20443, 20458, 20455, 20453, 20446, 20450, 20449, 20456, 20453, 20449, 20461, 20463, 20462, 20465, 20467, 20466, 20475, 20474, 20472, 20479, 20478, 20476, 20473, 20472, 20468, 20477, 20476, 20470, 20463, 20469, 20468, 20467, 20471, 20470, 20464, 20466, 20463, 20478, 20475, 20473, 20466, 20470, 20469,
20476, 20473, 20469, 20483, 20482, 20480, 20487, 20486, 20484, 20491, 20489, 20488, 20495, 20493, 20492, 20497, 20496, 20482, 20499, 20498, 20486, 20496, 20497, 20491, 20499, 20495, 20494, 20486, 20483, 20481, 20494, 20492, 20489, 20497, 20498, 20494, 20486, 20498, 20497, 20503, 20502, 20500, 20507, 20506, 20504, 20511, 20509, 20508, 20515, 20513, 20512, 20517, 20516, 20502, 20519, 20518, 20506, 20517, 20511, 20510, 20519, 20515, 20514, 20506, 20503, 20501, 20514, 20512, 20509, 20518, 20514, 20511, 20518, 20517, 20503, 20481, 20480, 20454, 20485, 20484, 20458, 20484, 20481, 20455, 20501, 20500, 20474, 20505, 20504, 20478, 20504, 20501, 20475, 20460, 20462, 20447, 20474, 20459, 20457, 20462, 20468, 20451, 20472, 20457, 20451, 20502, 20487, 20485, 20510, 20508, 20493, 20516, 20510, 20495, 20516, 20499, 20487, 20500, 20485, 20459, 20521, 20523, 20522, 20525, 20527, 20526, 20535, 20534, 20532, 20539, 20538, 20536, 20533, 20532, 20528, 20537, 20536, 20530, 20523, 20529, 20528, 20527, 20531, 20530, 20524, 20526, 20523, 20538, 20535, 20533, 20526, 20530, 20529, 20536, 20533, 20529, 20541, 20543, 20542, 20545, 20547, 20546, 20555, 20554, 20552, 20559, 20558, 20556, 20553, 20552, 20548, 20557, 20556, 20550, 20543, 20549, 20548, 20547, 20551, 20550,
20544, 20546, 20543, 20558, 20555, 20553, 20546, 20550, 20549, 20556, 20553, 20549, 20563, 20562, 20560, 20567, 20566, 20564, 20571, 20569, 20568, 20575, 20573, 20572, 20577, 20576, 20562, 20579, 20578, 20566, 20577, 20571, 20570, 20579, 20575, 20574, 20566, 20563, 20561, 20574, 20572, 20569, 20578, 20574, 20571, 20578, 20577, 20563, 20583, 20582, 20580, 20587, 20586, 20584, 20591, 20589, 20588, 20595, 20593, 20592, 20597, 20596, 20582, 20599, 20598, 20586, 20597, 20591, 20590, 20599, 20595, 20594, 20586, 20583, 20581, 20594, 20592, 20589, 20598, 20594, 20591, 20598, 20597, 20583, 20561, 20560, 20534, 20565, 20564, 20538, 20564, 20561, 20535, 20581, 20580, 20554, 20585, 20584, 20558, 20584, 20581, 20555, 20540, 20542, 20527, 20554, 20539, 20537, 20542, 20548, 20531, 20552, 20537, 20531, 20582, 20567, 20565, 20590, 20588, 20573, 20596, 20590, 20575, 20596, 20579, 20567, 20580, 20565, 20539, 20417, 20386, 20384, 20427, 20398, 20396, 20425, 20396, 20386, 20433, 20431, 20404, 20437, 20433, 20406, 20437, 20410, 20398, 20520, 20522, 20467, 20534, 20479, 20477, 20522, 20528, 20471, 20532, 20477, 20471, 20562, 20507, 20505, 20570, 20568, 20513, 20576, 20570, 20515, 20576, 20519, 20507, 20560, 20505, 20479, 20601, 20603, 20602, 20607, 20606, 20604,
20609, 20611, 20610, 20615, 20614, 20612, 20603, 20609, 20608, 20613, 20612, 20606, 20604, 20606, 20603, 20612, 20614, 20611, 20606, 20612, 20609, 20619, 20617, 20616, 20623, 20622, 20620, 20627, 20626, 20624, 20631, 20629, 20628, 20635, 20634, 20632, 20639, 20638, 20636, 20629, 20619, 20618, 20633, 20632, 20622, 20637, 20636, 20626, 20627, 20625, 20617, 20639, 20637, 20629, 20626, 20623, 20621, 20638, 20635, 20633, 20636, 20633, 20623, 20637, 20627, 20619, 20645, 20643, 20642, 20641, 20644, 20642, 20649, 20648, 20646, 20649, 20645, 20644, 20646, 20648, 20644, 20653, 20651, 20650, 20659, 20657, 20656, 20659, 20655, 20654, 20663, 20662, 20660, 20667, 20666, 20664, 20663, 20653, 20652, 20667, 20659, 20658, 20667, 20665, 20655, 20658, 20656, 20651, 20666, 20663, 20661, 20666, 20658, 20653, 20611, 20641, 20640, 20647, 20646, 20614, 20614, 20646, 20641, 20655, 20631, 20630, 20661, 20660, 20634, 20665, 20664, 20638, 20664, 20661, 20635, 20665, 20639, 20631, 20622, 20607, 20605, 20634, 20615, 20613, 20632, 20613, 20607, 20652, 20650, 20643, 20662, 20649, 20647, 20662, 20652, 20645, 20660, 20647, 20615, 20669, 20671, 20670, 20673, 20675, 20674, 20683, 20682, 20680, 20687, 20686, 20684, 20681, 20680, 20676, 20685, 20684, 20678, 20671, 20677, 20676,
20675, 20679, 20678, 20672, 20674, 20671, 20686, 20683, 20681, 20674, 20678, 20677, 20684, 20681, 20677, 20689, 20691, 20690, 20693, 20695, 20694, 20703, 20701, 20700, 20707, 20706, 20704, 20711, 20710, 20708, 20701, 20695, 20693, 20705, 20704, 20696, 20709, 20708, 20698, 20691, 20697, 20696, 20695, 20699, 20698, 20711, 20709, 20701, 20692, 20694, 20691, 20710, 20707, 20705, 20694, 20698, 20697, 20708, 20705, 20697, 20709, 20699, 20695, 20715, 20714, 20712, 20719, 20718, 20716, 20723, 20721, 20720, 20727, 20725, 20724, 20729, 20728, 20714, 20731, 20730, 20718, 20729, 20723, 20722, 20731, 20727, 20726, 20718, 20715, 20713, 20726, 20724, 20721, 20730, 20726, 20723, 20730, 20729, 20715, 20735, 20733, 20732, 20739, 20738, 20736, 20743, 20742, 20740, 20747, 20745, 20744, 20751, 20749, 20748, 20751, 20735, 20734, 20753, 20752, 20738, 20755, 20754, 20742, 20753, 20747, 20746, 20743, 20741, 20733, 20755, 20751, 20750, 20742, 20739, 20737, 20750, 20748, 20745, 20754, 20750, 20747, 20754, 20753, 20739, 20755, 20743, 20735, 20713, 20712, 20682, 20717, 20716, 20686, 20716, 20713, 20683, 20733, 20703, 20702, 20737, 20736, 20706, 20741, 20740, 20710, 20740, 20737, 20707, 20741, 20711, 20703, 20688, 20690, 20675, 20706, 20687, 20685, 20690, 20696, 20679,
20704, 20685, 20679, 20738, 20719, 20717, 20746, 20744, 20725, 20752, 20746, 20727, 20752, 20731, 20719, 20736, 20717, 20687, 20163, 20376, 20375, 20380, 20379, 20166, 20384, 20383, 20170, 20383, 20380, 20167, 20166, 20379, 20376, 20412, 20205, 20204, 20416, 20415, 20208, 20419, 20416, 20209, 20419, 20211, 20205, 20415, 20384, 20171, 20346, 20601, 20600, 20605, 20604, 20349, 20349, 20604, 20601, 20617, 20366, 20365, 20621, 20620, 20369, 20625, 20624, 20373, 20624, 20621, 20370, 20625, 20374, 20366, 20620, 20605, 20350, 20668, 20670, 20547, 20682, 20559, 20557, 20670, 20676, 20551, 20680, 20557, 20551, 20714, 20587, 20585, 20722, 20720, 20593, 20728, 20722, 20595, 20728, 20599, 20587, 20712, 20585, 20559, 19856, 20037, 20036, 20041, 20040, 19859, 20045, 20044, 19863, 20044, 20041, 19860, 19859, 20040, 20037, 20085, 19918, 19917, 20089, 20088, 19921, 20092, 20089, 19922, 20092, 19925, 19918, 20088, 20045, 19864, 20003, 20213, 20212, 20220, 20219, 20010, 20010, 20219, 20218, 20003, 20007, 20218, 20232, 20027, 20026, 20236, 20235, 20030, 20240, 20239, 20034, 20239, 20236, 20031, 20240, 20035, 20027, 20235, 20220, 20011, 20757, 20759, 20758, 20761, 20763, 20762, 20759, 20765, 20764, 20763, 20767, 20766, 20760, 20762, 20759, 20762, 20766, 20765,
20769, 20771, 20770, 20773, 20775, 20774, 20771, 20777, 20776, 20775, 20779, 20778, 20772, 20774, 20771, 20774, 20778, 20777, 20783, 20782, 20780, 20787, 20786, 20784, 20791, 20790, 20788, 20795, 20794, 20792, 20789, 20788, 20782, 20793, 20792, 20786, 20786, 20783, 20781, 20794, 20791, 20789, 20792, 20789, 20783, 20799, 20798, 20796, 20803, 20802, 20800, 20807, 20806, 20804, 20811, 20810, 20808, 20805, 20804, 20798, 20809, 20808, 20802, 20802, 20799, 20797, 20810, 20807, 20805, 20808, 20805, 20799, 20781, 20780, 20764, 20785, 20784, 20766, 20784, 20781, 20765, 20797, 20796, 20776, 20801, 20800, 20778, 20800, 20797, 20777, 20768, 20770, 20763, 20770, 20776, 20767, 20798, 20787, 20785, 20806, 20795, 20793, 20804, 20793, 20787, 20796, 20785, 20767, 20813, 20815, 20814, 20817, 20819, 20818, 20815, 20821, 20820, 20819, 20823, 20822, 20816, 20818, 20815, 20818, 20822, 20821, 20825, 20827, 20826, 20829, 20831, 20830, 20827, 20833, 20832, 20831, 20835, 20834, 20828, 20830, 20827, 20830, 20834, 20833, 20839, 20838, 20836, 20843, 20842, 20840, 20847, 20846, 20844, 20851, 20850, 20848, 20845, 20844, 20838, 20849, 20848, 20842, 20842, 20839, 20837, 20850, 20847, 20845, 20848, 20845, 20839, 20855, 20854, 20852, 20859, 20858, 20856, 20863, 20862, 20860,
20867, 20866, 20864, 20861, 20860, 20854, 20865, 20864, 20858, 20858, 20855, 20853, 20866, 20863, 20861, 20864, 20861, 20855, 20837, 20836, 20820, 20841, 20840, 20822, 20840, 20837, 20821, 20853, 20852, 20832, 20857, 20856, 20834, 20856, 20853, 20833, 20824, 20826, 20819, 20826, 20832, 20823, 20854, 20843, 20841, 20862, 20851, 20849, 20860, 20849, 20843, 20852, 20841, 20823, 20812, 20814, 20775, 20814, 20820, 20779, 20838, 20803, 20801, 20846, 20811, 20809, 20844, 20809, 20803, 20836, 20801, 20779, 20869, 20871, 20870, 20873, 20875, 20874, 20871, 20877, 20876, 20875, 20879, 20878, 20872, 20874, 20871, 20874, 20878, 20877, 20881, 20883, 20882, 20885, 20887, 20886, 20883, 20889, 20888, 20887, 20891, 20890, 20884, 20886, 20883, 20886, 20890, 20889, 20895, 20894, 20892, 20899, 20898, 20896, 20903, 20902, 20900, 20907, 20906, 20904, 20901, 20900, 20894, 20905, 20904, 20898, 20898, 20895, 20893, 20906, 20903, 20901, 20904, 20901, 20895, 20911, 20909, 20908, 20915, 20914, 20912, 20919, 20918, 20916, 20923, 20921, 20920, 20927, 20926, 20924, 20931, 20930, 20928, 20921, 20911, 20910, 20925, 20924, 20914, 20929, 20928, 20918, 20919, 20917, 20909, 20931, 20929, 20921, 20918, 20915, 20913, 20930, 20927, 20925, 20928, 20925, 20915, 20929, 20919, 20911,
20893, 20892, 20876, 20897, 20896, 20878, 20896, 20893, 20877, 20909, 20887, 20885, 20913, 20912, 20888, 20917, 20916, 20890, 20916, 20913, 20889, 20917, 20891, 20887, 20880, 20882, 20875, 20882, 20888, 20879, 20914, 20899, 20897, 20926, 20907, 20905, 20924, 20905, 20899, 20912, 20897, 20879, 20868, 20870, 20831, 20870, 20876, 20835, 20894, 20859, 20857, 20902, 20867, 20865, 20900, 20865, 20859, 20892, 20857, 20835, 20935, 20933, 20932, 20939, 20937, 20936, 20941, 20935, 20934, 20943, 20939, 20938, 20938, 20936, 20933, 20942, 20938, 20935, 20947, 20945, 20944, 20951, 20949, 20948, 20953, 20947, 20946, 20955, 20951, 20950, 20950, 20948, 20945, 20954, 20950, 20947, 20946, 20944, 20937, 20952, 20946, 20939, 20959, 20957, 20956, 20963, 20961, 20960, 20965, 20959, 20958, 20967, 20963, 20962, 20962, 20960, 20957, 20966, 20962, 20959, 20971, 20969, 20968, 20975, 20973, 20972, 20977, 20971, 20970, 20979, 20975, 20974, 20974, 20972, 20969, 20978, 20974, 20971, 20970, 20968, 20961, 20976, 20970, 20963, 20981, 20983, 20982, 20985, 20987, 20986, 20991, 20990, 20988, 20995, 20994, 20992, 20999, 20998, 20996, 21003, 21002, 21000, 20997, 20996, 20990, 21001, 21000, 20994, 20983, 20989, 20988, 20987, 20993, 20992, 20984, 20986, 20983, 20994, 20991, 20989,
21002, 20999, 20997, 20986, 20992, 20989, 21000, 20997, 20991, 21005, 21007, 21006, 21009, 21011, 21010, 21015, 21014, 21012, 21019, 21018, 21016, 21023, 21022, 21020, 21027, 21026, 21024, 21021, 21020, 21014, 21025, 21024, 21018, 21007, 21013, 21012, 21011, 21017, 21016, 21008, 21010, 21007, 21018, 21015, 21013, 21026, 21023, 21021, 21010, 21016, 21013, 21024, 21021, 21015, 21033, 21032, 21030, 21037, 21036, 21034, 21039, 21041, 21040, 21045, 21044, 21042, 21049, 21048, 21046, 21029, 21039, 21038, 21033, 21043, 21042, 21047, 21046, 21036, 21042, 21044, 21041, 21036, 21033, 21031, 21048, 21045, 21043, 21046, 21043, 21033, 21032, 21042, 21039, 21055, 21053, 21052, 21057, 21059, 21058, 21063, 21062, 21060, 21067, 21065, 21064, 21071, 21070, 21068, 21065, 21051, 21050, 21069, 21068, 21058, 21071, 21069, 21065, 21063, 21055, 21054, 21054, 21052, 21050, 21060, 21062, 21059, 21062, 21054, 21051, 21069, 21059, 21051, 21031, 21030, 20998, 21035, 21034, 21002, 21034, 21031, 20999, 21057, 21056, 21022, 21061, 21060, 21026, 21060, 21057, 21023, 21004, 21006, 20987, 21014, 20995, 20993, 21022, 21003, 21001, 21006, 21012, 20993, 21020, 21001, 20995, 21058, 21037, 21035, 21070, 21049, 21047, 21068, 21047, 21037, 21056, 21035, 21003, 21073, 21075, 21074,
21077, 21079, 21078, 21083, 21082, 21080, 21087, 21086, 21084, 21091, 21090, 21088, 21095, 21094, 21092, 21089, 21088, 21082, 21093, 21092, 21086, 21075, 21081, 21080, 21079, 21085, 21084, 21076, 21078, 21075, 21086, 21083, 21081, 21094, 21091, 21089, 21078, 21084, 21081, 21092, 21089, 21083, 21097, 21099, 21098, 21101, 21103, 21102, 21107, 21106, 21104, 21111, 21110, 21108, 21115, 21114, 21112, 21119, 21118, 21116, 21113, 21112, 21106, 21117, 21116, 21110, 21099, 21105, 21104, 21103, 21109, 21108, 21100, 21102, 21099, 21110, 21107, 21105, 21118, 21115, 21113, 21102, 21108, 21105, 21116, 21113, 21107, 21123, 21121, 21120, 21127, 21125, 21124, 21131, 21130, 21128, 21135, 21134, 21132, 21131, 21123, 21122, 21135, 21127, 21126, 21126, 21124, 21121, 21134, 21131, 21129, 21134, 21126, 21123, 21139, 21137, 21136, 21143, 21141, 21140, 21147, 21146, 21144, 21151, 21150, 21148, 21147, 21139, 21138, 21151, 21143, 21142, 21142, 21140, 21137, 21150, 21147, 21145, 21150, 21142, 21139, 21129, 21128, 21090, 21133, 21132, 21094, 21132, 21129, 21091, 21145, 21144, 21114, 21149, 21148, 21118, 21148, 21145, 21115, 21096, 21098, 21079, 21106, 21087, 21085, 21114, 21095, 21093, 21098, 21104, 21085, 21112, 21093, 21087, 21138, 21136, 21125, 21146, 21135, 21133,
21146, 21138, 21127, 21144, 21133, 21095, 20958, 20956, 20949, 20964, 20958, 20951, 21072, 21074, 21011, 21082, 21019, 21017, 21090, 21027, 21025, 21074, 21080, 21017, 21088, 21025, 21019, 21122, 21120, 21053, 21130, 21063, 21061, 21130, 21122, 21055, 21128, 21061, 21027, 21155, 21153, 21152, 21159, 21157, 21156, 21161, 21155, 21154, 21163, 21159, 21158, 21158, 21156, 21153, 21162, 21158, 21155, 21167, 21165, 21164, 21171, 21169, 21168, 21173, 21167, 21166, 21175, 21171, 21170, 21170, 21168, 21165, 21174, 21170, 21167, 21166, 21164, 21157, 21172, 21166, 21159, 21177, 21179, 21178, 21181, 21183, 21182, 21187, 21186, 21184, 21191, 21190, 21188, 21195, 21194, 21192, 21199, 21198, 21196, 21193, 21192, 21186, 21197, 21196, 21190, 21179, 21185, 21184, 21183, 21189, 21188, 21180, 21182, 21179, 21190, 21187, 21185, 21198, 21195, 21193, 21182, 21188, 21185, 21196, 21193, 21187, 21201, 21203, 21202, 21205, 21207, 21206, 21209, 21207, 21205, 21213, 21212, 21210, 21217, 21216, 21214, 21221, 21219, 21218, 21225, 21224, 21222, 21229, 21228, 21226, 21219, 21209, 21208, 21223, 21222, 21212, 21227, 21226, 21216, 21203, 21211, 21210, 21207, 21215, 21214, 21217, 21215, 21207, 21229, 21227, 21219, 21204, 21206, 21203, 21216, 21213, 21211, 21228, 21225, 21223,
21206, 21214, 21211, 21226, 21223, 21213, 21227, 21217, 21209, 21233, 21231, 21230, 21239, 21238, 21236, 21243, 21242, 21240, 21245, 21247, 21246, 21251, 21250, 21248, 21235, 21245, 21244, 21243, 21249, 21248, 21239, 21233, 21232, 21248, 21250, 21247, 21235, 21234, 21231, 21242, 21239, 21237, 21242, 21235, 21233, 21242, 21248, 21245, 21255, 21253, 21252, 21259, 21258, 21256, 21263, 21262, 21260, 21267, 21265, 21264, 21271, 21270, 21268, 21275, 21274, 21272, 21265, 21255, 21254, 21269, 21268, 21258, 21273, 21272, 21262, 21263, 21261, 21253, 21275, 21273, 21265, 21262, 21259, 21257, 21274, 21271, 21269, 21272, 21269, 21259, 21273, 21263, 21255, 21237, 21236, 21194, 21241, 21240, 21198, 21240, 21237, 21195, 21253, 21221, 21220, 21257, 21256, 21224, 21261, 21260, 21228, 21260, 21257, 21225, 21261, 21229, 21221, 21200, 21202, 21183, 21212, 21191, 21189, 21224, 21199, 21197, 21202, 21210, 21189, 21222, 21197, 21191, 21258, 21243, 21241, 21270, 21251, 21249, 21268, 21249, 21243, 21256, 21241, 21199, 21277, 21279, 21278, 21283, 21282, 21280, 21287, 21286, 21284, 21289, 21291, 21290, 21295, 21294, 21292, 21299, 21298, 21296, 21279, 21289, 21288, 21293, 21292, 21282, 21297, 21296, 21286, 21280, 21282, 21279, 21292, 21294, 21291, 21286, 21283, 21281,
21298, 21295, 21293, 21296, 21293, 21283, 21282, 21292, 21289, 21303, 21301, 21300, 21307, 21306, 21304, 21311, 21309, 21308, 21315, 21314, 21312, 21309, 21303, 21302, 21313, 21312, 21306, 21307, 21305, 21301, 21315, 21313, 21309, 21313, 21307, 21303, 21317, 21319, 21318, 21323, 21322, 21320, 21327, 21326, 21324, 21329, 21331, 21330, 21335, 21334, 21332, 21339, 21338, 21336, 21319, 21329, 21328, 21333, 21332, 21322, 21337, 21336, 21326, 21320, 21322, 21319, 21332, 21334, 21331, 21326, 21323, 21321, 21338, 21335, 21333, 21336, 21333, 21323, 21322, 21332, 21329, 21343, 21341, 21340, 21347, 21346, 21344, 21351, 21349, 21348, 21355, 21354, 21352, 21349, 21343, 21342, 21353, 21352, 21346, 21347, 21345, 21341, 21355, 21353, 21349, 21353, 21347, 21343, 21291, 21317, 21316, 21321, 21320, 21294, 21325, 21324, 21298, 21324, 21321, 21295, 21294, 21320, 21317, 21341, 21311, 21310, 21345, 21344, 21314, 21345, 21315, 21311, 21306, 21287, 21285, 21314, 21299, 21297, 21312, 21297, 21287, 21346, 21327, 21325, 21354, 21339, 21337, 21352, 21337, 21327, 21344, 21325, 21299, 21357, 21359, 21358, 21363, 21362, 21360, 21367, 21366, 21364, 21369, 21371, 21370, 21375, 21374, 21372, 21379, 21378, 21376, 21359, 21369, 21368, 21373, 21372, 21362, 21377, 21376, 21366,
21360, 21362, 21359, 21372, 21374, 21371, 21366, 21363, 21361, 21378, 21375, 21373, 21376, 21373, 21363, 21362, 21372, 21369, 21383, 21381, 21380, 21387, 21386, 21384, 21391, 21389, 21388, 21395, 21394, 21392, 21389, 21383, 21382, 21393, 21392, 21386, 21387, 21385, 21381, 21395, 21393, 21389, 21393, 21387, 21383, 21397, 21399, 21398, 21403, 21402, 21400, 21407, 21406, 21404, 21409, 21411, 21410, 21415, 21414, 21412, 21419, 21418, 21416, 21399, 21409, 21408, 21413, 21412, 21402, 21417, 21416, 21406, 21400, 21402, 21399, 21412, 21414, 21411, 21406, 21403, 21401, 21418, 21415, 21413, 21416, 21413, 21403, 21402, 21412, 21409, 21423, 21421, 21420, 21427, 21426, 21424, 21431, 21429, 21428, 21435, 21434, 21432, 21429, 21423, 21422, 21427, 21433, 21432, 21427, 21425, 21421, 21435, 21433, 21429, 21433, 21427, 21423, 21371, 21397, 21396, 21401, 21400, 21374, 21405, 21404, 21378, 21404, 21401, 21375, 21374, 21400, 21397, 21421, 21391, 21390, 21425, 21424, 21394, 21425, 21395, 21391, 21386, 21367, 21365, 21394, 21379, 21377, 21392, 21377, 21367, 21426, 21407, 21405, 21434, 21419, 21417, 21432, 21417, 21407, 21424, 21405, 21379, 21331, 21357, 21356, 21361, 21360, 21334, 21365, 21364, 21338, 21364, 21361, 21335, 21334, 21360, 21357, 21381, 21351, 21350,
21385, 21384, 21354, 21385, 21355, 21351, 21384, 21365, 21339, 21437, 21439, 21438, 21443, 21442, 21440, 21445, 21447, 21446, 21451, 21450, 21448, 21439, 21445, 21444, 21449, 21448, 21442, 21440, 21442, 21439, 21448, 21450, 21447, 21442, 21448, 21445, 21455, 21453, 21452, 21459, 21458, 21456, 21463, 21462, 21460, 21467, 21465, 21464, 21471, 21470, 21468, 21475, 21474, 21472, 21465, 21455, 21454, 21469, 21468, 21458, 21473, 21472, 21462, 21463, 21461, 21453, 21475, 21473, 21465, 21462, 21459, 21457, 21474, 21471, 21469, 21472, 21469, 21459, 21473, 21463, 21455, 21477, 21479, 21478, 21483, 21482, 21480, 21485, 21487, 21486, 21491, 21490, 21488, 21479, 21485, 21484, 21489, 21488, 21482, 21480, 21482, 21479, 21488, 21490, 21487, 21482, 21488, 21485, 21495, 21493, 21492, 21499, 21498, 21496, 21503, 21502, 21500, 21507, 21505, 21504, 21511, 21510, 21508, 21515, 21514, 21512, 21505, 21495, 21494, 21509, 21508, 21498, 21513, 21512, 21502, 21503, 21501, 21493, 21515, 21513, 21505, 21502, 21499, 21497, 21514, 21511, 21509, 21512, 21509, 21499, 21513, 21503, 21495, 21447, 21477, 21476, 21481, 21480, 21450, 21450, 21480, 21477, 21493, 21467, 21466, 21497, 21496, 21470, 21501, 21500, 21474, 21500, 21497, 21471, 21501, 21475, 21467, 21458, 21443, 21441,
21470, 21451, 21449, 21468, 21449, 21443, 21498, 21483, 21481, 21510, 21491, 21489, 21508, 21489, 21483, 21496, 21481, 21451, 21517, 21519, 21518, 21523, 21522, 21520, 21525, 21527, 21526, 21531, 21530, 21528, 21519, 21525, 21524, 21529, 21528, 21522, 21520, 21522, 21519, 21528, 21530, 21527, 21522, 21528, 21525, 21535, 21533, 21532, 21539, 21538, 21536, 21543, 21542, 21540, 21547, 21545, 21544, 21551, 21550, 21548, 21555, 21554, 21552, 21545, 21535, 21534, 21549, 21548, 21538, 21553, 21552, 21542, 21543, 21541, 21533, 21555, 21553, 21545, 21542, 21539, 21537, 21554, 21551, 21549, 21552, 21549, 21539, 21553, 21543, 21535, 21557, 21559, 21558, 21563, 21562, 21560, 21565, 21567, 21566, 21569, 21571, 21570, 21559, 21565, 21564, 21563, 21569, 21568, 21560, 21562, 21559, 21565, 21568, 21570, 21559, 21562, 21568, 21575, 21573, 21572, 21579, 21578, 21576, 21583, 21582, 21580, 21587, 21585, 21584, 21591, 21590, 21588, 21595, 21594, 21592, 21585, 21575, 21574, 21589, 21588, 21578, 21593, 21592, 21582, 21583, 21581, 21573, 21595, 21593, 21585, 21582, 21579, 21577, 21594, 21591, 21589, 21592, 21589, 21579, 21593, 21583, 21575, 21527, 21557, 21556, 21561, 21560, 21530, 21530, 21560, 21557, 21573, 21547, 21546, 21577, 21576, 21550, 21581, 21580, 21554,
21580, 21577, 21551, 21581, 21555, 21547, 21538, 21523, 21521, 21550, 21531, 21529, 21548, 21529, 21523, 21578, 21563, 21561, 21590, 21571, 21569, 21588, 21569, 21563, 21576, 21561, 21531, 21487, 21517, 21516, 21521, 21520, 21490, 21490, 21520, 21517, 21533, 21507, 21506, 21537, 21536, 21510, 21541, 21540, 21514, 21540, 21537, 21511, 21541, 21515, 21507, 21536, 21521, 21491, 21041, 21277, 21276, 21281, 21280, 21044, 21285, 21284, 21048, 21284, 21281, 21045, 21044, 21280, 21277, 21301, 21067, 21066, 21305, 21304, 21070, 21305, 21071, 21067, 21304, 21285, 21049, 21247, 21437, 21436, 21441, 21440, 21250, 21250, 21440, 21437, 21453, 21267, 21266, 21457, 21456, 21270, 21461, 21460, 21274, 21460, 21457, 21271, 21461, 21275, 21267, 21456, 21441, 21251, 21154, 21152, 20973, 21160, 21154, 20975, 21176, 21178, 21103, 21186, 21111, 21109, 21194, 21119, 21117, 21178, 21184, 21109, 21192, 21117, 21111, 21232, 21230, 21141, 21238, 21151, 21149, 21238, 21232, 21143, 21236, 21149, 21119, 20941, 20940, 20790, 20943, 20942, 20794, 20942, 20941, 20791, 20953, 20952, 20806, 20955, 20954, 20810, 20954, 20953, 20807, 20952, 20943, 20795, 20965, 20964, 20846, 20967, 20966, 20850, 20966, 20965, 20847, 20977, 20976, 20862, 20979, 20978, 20866, 20978, 20977, 20863,
20976, 20967, 20851, 20964, 20955, 20811, 21161, 21160, 20902, 21163, 21162, 20906, 21162, 21161, 20903, 21171, 20923, 20922, 21173, 21172, 20926, 21175, 21174, 20930, 21174, 21173, 20927, 21175, 20931, 20923, 21172, 21163, 20907, 21160, 20979, 20867, 19194, 19568, 19567, 19572, 19198, 19197, 19202, 19578, 19577, 19582, 19581, 19205, 19581, 19578, 19202, 19194, 19201, 19577, 19572, 19582, 19206, 19216, 19600, 19599, 19225, 19608, 19607, 19607, 19605, 19221, 19221, 19605, 19600, 19780, 19410, 19409, 19780, 19783, 19413, 19783, 19608, 19225, 21600, 21601, 21599, 21603, 21605, 21604, 21607, 21609, 21608, 21618, 21616, 21615, 21611, 21612, 21601, 21605, 21614, 21613, 21602, 21604, 21597, 21599, 21607, 21606, 21616, 21612, 21611, 21597, 21600, 21598, 21606, 21608, 21603, 21617, 21615, 21613, 21598, 21606, 21602, 21615, 21611, 21610, 21610, 21611, 21600, 21604, 21613, 21610, 21621, 21622, 21620, 21625, 21626, 21624, 21628, 21630, 21629, 21632, 21634, 21633, 21644, 21642, 21641, 21648, 21646, 21645, 21650, 21652, 21651, 21635, 21636, 21622, 21637, 21638, 21626, 21620, 21628, 21627, 21642, 21636, 21635, 21624, 21632, 21631, 21646, 21638, 21637, 21640, 21650, 21649, 21622, 21625, 21623, 21631, 21633, 21630, 21647, 21645, 21642, 21623, 21631, 21628,
21645, 21637, 21636, 21636, 21637, 21625, 21601, 21621, 21619, 21627, 21629, 21609, 21643, 21641, 21616, 21619, 21627, 21607, 21641, 21635, 21612, 21612, 21635, 21621, 21655, 21656, 21654, 21659, 21660, 21658, 21662, 21664, 21663, 21666, 21668, 21667, 21678, 21676, 21675, 21684, 21682, 21681, 21684, 21686, 21685, 21669, 21670, 21656, 21672, 21673, 21660, 21654, 21662, 21661, 21676, 21670, 21669, 21658, 21666, 21665, 21682, 21673, 21672, 21682, 21685, 21674, 21656, 21659, 21657, 21665, 21667, 21664, 21683, 21681, 21676, 21657, 21665, 21662, 21681, 21672, 21670, 21670, 21672, 21659, 21689, 21690, 21688, 21692, 21694, 21693, 21696, 21694, 21692, 21699, 21700, 21698, 21704, 21702, 21701, 21697, 21698, 21690, 21702, 21696, 21695, 21688, 21692, 21691, 21695, 21692, 21688, 21703, 21701, 21698, 21701, 21695, 21690, 21660, 21689, 21687, 21691, 21693, 21668, 21674, 21699, 21697, 21687, 21691, 21666, 21673, 21697, 21689, 21707, 21708, 21706, 21710, 21712, 21711, 21715, 21716, 21714, 21718, 21720, 21719, 21713, 21714, 21708, 21712, 21718, 21717, 21709, 21711, 21707, 21717, 21719, 21715, 21711, 21717, 21713, 21723, 21724, 21722, 21730, 21728, 21727, 21726, 21729, 21727, 21728, 21724, 21723, 21725, 21727, 21723, 21721, 21722, 21716, 21720, 21726, 21725,
21719, 21725, 21721, 21738, 21736, 21735, 21740, 21742, 21741, 21750, 21748, 21747, 21752, 21754, 21753, 21748, 21738, 21737, 21742, 21752, 21751, 21737, 21735, 21731, 21749, 21747, 21743, 21739, 21741, 21734, 21751, 21753, 21746, 21747, 21737, 21732, 21741, 21751, 21745, 21757, 21758, 21756, 21762, 21760, 21759, 21765, 21766, 21764, 21770, 21768, 21767, 21763, 21764, 21758, 21768, 21762, 21761, 21761, 21759, 21756, 21769, 21767, 21764, 21767, 21761, 21758, 21778, 21776, 21775, 21780, 21782, 21781, 21777, 21775, 21771, 21779, 21781, 21774, 21785, 21786, 21784, 21792, 21790, 21789, 21792, 21788, 21787, 21790, 21786, 21785, 21790, 21787, 21784, 21776, 21750, 21749, 21754, 21780, 21779, 21775, 21749, 21744, 21753, 21779, 21773, 21783, 21784, 21766, 21788, 21770, 21769, 21787, 21769, 21766, 21734, 21757, 21755, 21746, 21765, 21763, 21745, 21763, 21757, 21774, 21785, 21783, 21791, 21789, 21781, 21789, 21785, 21774, 21773, 21783, 21765, 21705, 21706, 21640, 21651, 21710, 21709, 21649, 21709, 21705, 21736, 21680, 21679, 21686, 21740, 21739, 21735, 21679, 21671, 21685, 21739, 21733, 21755, 21756, 21700, 21760, 21704, 21703, 21759, 21703, 21700, 21733, 21755, 21699, 21626, 21655, 21653, 21661, 21663, 21634, 21677, 21675, 21646, 21679, 21680, 21652,
21653, 21661, 21632, 21675, 21669, 21638, 21671, 21679, 21650, 21638, 21669, 21655, 21708, 21732, 21731, 21716, 21744, 21743, 21714, 21743, 21732, 21724, 21772, 21771, 21778, 21777, 21728, 21777, 21772, 21724, 21722, 21771, 21744, 21706, 21731, 21671, 21798, 21800, 21801, 21803, 21805, 21804, 21807, 21809, 21808, 21808, 21811, 21810, 21797, 21803, 21802, 21798, 21799, 21807, 21806, 21810, 21800, 21810, 21811, 21805, 21797, 21800, 21810, 21802, 21804, 21795, 21796, 21802, 21794, 21814, 21815, 21813, 21819, 21821, 21820, 21823, 21821, 21819, 21825, 21827, 21826, 21813, 21819, 21818, 21815, 21822, 21819, 21816, 21817, 21825, 21824, 21826, 21823, 21815, 21816, 21824, 21828, 21830, 21831, 21833, 21835, 21834, 21834, 21837, 21836, 21841, 21842, 21840, 21846, 21849, 21847, 21830, 21839, 21840, 21837, 21846, 21845, 21828, 21829, 21833, 21832, 21836, 21830, 21839, 21845, 21847, 21847, 21849, 21844, 21841, 21847, 21843, 21836, 21845, 21839, 21851, 21853, 21854, 21862, 21860, 21859, 21862, 21858, 21857, 21866, 21864, 21863, 21872, 21870, 21869, 21870, 21861, 21859, 21859, 21860, 21854, 21860, 21857, 21852, 21871, 21869, 21867, 21864, 21856, 21855, 21865, 21863, 21857, 21857, 21863, 21855, 21869, 21859, 21853, 21875, 21877, 21878, 21880, 21882, 21881,
21886, 21884, 21883, 21891, 21892, 21890, 21894, 21896, 21895, 21900, 21898, 21897, 21889, 21890, 21878, 21882, 21894, 21893, 21898, 21886, 21885, 21879, 21881, 21874, 21887, 21893, 21895, 21885, 21883, 21876, 21892, 21899, 21897, 21873, 21874, 21877, 21888, 21891, 21889, 21874, 21887, 21889, 21881, 21893, 21887, 21890, 21897, 21885, 21875, 21876, 21842, 21844, 21880, 21879, 21884, 21850, 21848, 21838, 21873, 21875, 21838, 21843, 21879, 21876, 21883, 21848, 21831, 21853, 21851, 21840, 21842, 21868, 21872, 21871, 21848, 21871, 21868, 21842, 21831, 21840, 21867, 21904, 21906, 21907, 21911, 21909, 21908, 21911, 21913, 21912, 21924, 21922, 21921, 21924, 21926, 21925, 21918, 21919, 21907, 21917, 21923, 21921, 21909, 21902, 21901, 21902, 21909, 21912, 21921, 21922, 21919, 21919, 21922, 21925, 21903, 21906, 21904, 21917, 21916, 21912, 21916, 21906, 21903, 21916, 21921, 21918, 21932, 21930, 21929, 21935, 21936, 21934, 21938, 21940, 21939, 21942, 21940, 21938, 21939, 21932, 21931, 21928, 21931, 21929, 21933, 21934, 21938, 21941, 21938, 21934, 21933, 21937, 21931, 21946, 21948, 21947, 21952, 21954, 21953, 21948, 21952, 21951, 21943, 21945, 21947, 21951, 21953, 21950, 21947, 21951, 21949, 21955, 21957, 21958, 21962, 21960, 21959, 21963, 21965, 21966,
21970, 21968, 21967, 21963, 21964, 21958, 21968, 21962, 21961, 21958, 21961, 21959, 21969, 21967, 21964, 21967, 21961, 21958, 21926, 21946, 21945, 21920, 21925, 21945, 21955, 21956, 21936, 21960, 21942, 21941, 21956, 21959, 21941, 21907, 21928, 21927, 21929, 21930, 21915, 21919, 21920, 21935, 21905, 21927, 21929, 21907, 21919, 21933, 21944, 21957, 21955, 21949, 21950, 21965, 21944, 21949, 21963, 21943, 21955, 21935, 21795, 21835, 21833, 21794, 21833, 21829, 21851, 21852, 21797, 21855, 21856, 21801, 21797, 21852, 21855, 21793, 21829, 21851, 21814, 21901, 21902, 21904, 21905, 21817, 21915, 21827, 21825, 21815, 21902, 21904, 21905, 21914, 21825, 21801, 21814, 21812, 21818, 21820, 21809, 21799, 21812, 21818, 21910, 21908, 21864, 21864, 21908, 21901, 21856, 21901, 21814, 21973, 21974, 21972, 21976, 21978, 21977, 21977, 21980, 21979, 21983, 21984, 21982, 21986, 21988, 21987, 21981, 21982, 21974, 21980, 21986, 21985, 21972, 21976, 21975, 21975, 21979, 21973, 21985, 21987, 21983, 21979, 21985, 21981, 21991, 21992, 21990, 21998, 21996, 21995, 21994, 21997, 21995, 21996, 21992, 21991, 21993, 21995, 21991, 21989, 21990, 21984, 21988, 21994, 21993, 21987, 21993, 21989, 22001, 22002, 22000, 22005, 22006, 22004, 22008, 22010, 22009, 22012, 22014, 22013,
22017, 22018, 22016, 22021, 22022, 22020, 22015, 22016, 22002, 22019, 22020, 22006, 22000, 22008, 22007, 22004, 22012, 22011, 22002, 22005, 22003, 22011, 22013, 22010, 22018, 22021, 22019, 22003, 22011, 22008, 22016, 22019, 22005, 22025, 22026, 22024, 22029, 22030, 22028, 22032, 22034, 22033, 22036, 22038, 22037, 22041, 22042, 22040, 22045, 22046, 22044, 22039, 22040, 22026, 22043, 22044, 22030, 22024, 22032, 22031, 22028, 22036, 22035, 22026, 22029, 22027, 22035, 22037, 22034, 22042, 22045, 22043, 22027, 22035, 22032, 22040, 22043, 22029, 22049, 22050, 22048, 22053, 22054, 22052, 22058, 22056, 22055, 22062, 22060, 22059, 22056, 22050, 22049, 22060, 22054, 22053, 22050, 22053, 22051, 22061, 22059, 22056, 22059, 22053, 22050, 22065, 22066, 22064, 22069, 22070, 22068, 22074, 22072, 22071, 22078, 22076, 22075, 22072, 22066, 22065, 22076, 22070, 22069, 22066, 22069, 22067, 22077, 22075, 22072, 22075, 22069, 22066, 22047, 22048, 22018, 22051, 22052, 22022, 22048, 22051, 22021, 22063, 22064, 22042, 22067, 22068, 22046, 22064, 22067, 22045, 22006, 22025, 22023, 22031, 22033, 22014, 22022, 22041, 22039, 22023, 22031, 22012, 22020, 22039, 22025, 22054, 22065, 22063, 22073, 22071, 22060, 22071, 22065, 22054, 22052, 22063, 22041, 21974, 22001, 21999,
22007, 22009, 21978, 21984, 22017, 22015, 21999, 22007, 21976, 21982, 22015, 22001, 21992, 22049, 22047, 22057, 22055, 21996, 22055, 22049, 21992, 21990, 22047, 22017, 22080, 22082, 22083, 22088, 22091, 22090, 22098, 22096, 22095, 22098, 22100, 22099, 22092, 22093, 22083, 22091, 22097, 22095, 22086, 22090, 22082, 22095, 22096, 22093, 22096, 22099, 22094, 22088, 22086, 22084, 22084, 22086, 22080, 22082, 22090, 22095, 22106, 22104, 22103, 22107, 22109, 22110, 22114, 22116, 22115, 22120, 22116, 22114, 22122, 22124, 22123, 22115, 22106, 22105, 22102, 22105, 22103, 22108, 22114, 22113, 22110, 22118, 22114, 22111, 22112, 22122, 22121, 22123, 22120, 22111, 22121, 22118, 22107, 22113, 22105, 22130, 22131, 22129, 22135, 22133, 22132, 22135, 22137, 22136, 22141, 22139, 22138, 22133, 22126, 22125, 22126, 22133, 22136, 22138, 22139, 22131, 22127, 22130, 22128, 22140, 22138, 22136, 22136, 22138, 22130, 22109, 22125, 22126, 22128, 22129, 22112, 22119, 22134, 22132, 22110, 22126, 22128, 22109, 22117, 22132, 22083, 22102, 22101, 22103, 22104, 22089, 22093, 22094, 22109, 22119, 22117, 22099, 22081, 22101, 22103, 22117, 22109, 22094, 22083, 22093, 22107, 22148, 22150, 22149, 22146, 22150, 22148, 22156, 22157, 22155, 22161, 22163, 22162, 22159, 22163, 22161,
22143, 22154, 22155, 22149, 22165, 22164, 22152, 22153, 22161, 22158, 22161, 22153, 22144, 22148, 22147, 22145, 22148, 22144, 22151, 22153, 22156, 22164, 22165, 22159, 22151, 22154, 22164, 22143, 22147, 22164, 22168, 22170, 22171, 22177, 22175, 22174, 22177, 22173, 22172, 22183, 22181, 22180, 22181, 22176, 22174, 22175, 22171, 22170, 22171, 22175, 22172, 22182, 22180, 22178, 22178, 22180, 22174, 22184, 22186, 22187, 22195, 22193, 22192, 22195, 22191, 22190, 22199, 22197, 22196, 22192, 22193, 22187, 22193, 22190, 22185, 22197, 22189, 22188, 22198, 22196, 22190, 22190, 22196, 22188, 22184, 22185, 22153, 22156, 22188, 22189, 22199, 22167, 22166, 22185, 22188, 22156, 22197, 22166, 22157, 22144, 22170, 22168, 22155, 22157, 22179, 22183, 22182, 22166, 22182, 22179, 22157, 22155, 22178, 22170, 22206, 22208, 22207, 22213, 22214, 22212, 22216, 22218, 22217, 22222, 22220, 22219, 22211, 22212, 22202, 22204, 22216, 22215, 22215, 22217, 22210, 22202, 22206, 22205, 22220, 22214, 22213, 22205, 22207, 22204, 22210, 22213, 22211, 22221, 22219, 22217, 22201, 22205, 22203, 22219, 22213, 22210, 22209, 22211, 22201, 22203, 22215, 22209, 22228, 22230, 22229, 22232, 22234, 22233, 22237, 22238, 22236, 22241, 22242, 22240, 22246, 22244, 22243, 22250, 22248, 22247,
22235, 22236, 22224, 22239, 22240, 22226, 22224, 22228, 22227, 22244, 22238, 22237, 22226, 22232, 22231, 22248, 22242, 22241, 22231, 22233, 22230, 22238, 22241, 22239, 22249, 22247, 22244, 22225, 22231, 22228, 22247, 22241, 22238, 22236, 22239, 22225, 22227, 22229, 22208, 22214, 22237, 22235, 22245, 22243, 22220, 22223, 22227, 22206, 22243, 22237, 22214, 22212, 22235, 22223, 22256, 22258, 22257, 22260, 22262, 22261, 22265, 22266, 22264, 22269, 22270, 22268, 22274, 22272, 22271, 22278, 22276, 22275, 22263, 22264, 22252, 22267, 22268, 22254, 22252, 22256, 22255, 22272, 22266, 22265, 22254, 22260, 22259, 22276, 22270, 22269, 22259, 22261, 22258, 22266, 22269, 22267, 22277, 22275, 22272, 22253, 22259, 22256, 22275, 22269, 22266, 22264, 22267, 22253, 22282, 22284, 22283, 22287, 22288, 22286, 22292, 22290, 22289, 22285, 22286, 22280, 22290, 22284, 22282, 22280, 22282, 22281, 22291, 22289, 22286, 22289, 22282, 22280, 22281, 22283, 22262, 22270, 22287, 22285, 22279, 22281, 22260, 22268, 22285, 22279, 22160, 22162, 22124, 22152, 22160, 22122, 22129, 22131, 22186, 22194, 22192, 22139, 22192, 22186, 22131, 22112, 22129, 22184, 22255, 22257, 22234, 22242, 22265, 22263, 22273, 22271, 22248, 22251, 22255, 22232, 22271, 22265, 22242, 22240, 22263, 22251,
21891, 22080, 22081, 21896, 22085, 22084, 22089, 21900, 21899, 22079, 22080, 21891, 21895, 22084, 22079, 22087, 21899, 21892, 21954, 22146, 22145, 21953, 22145, 22142, 21965, 22168, 22169, 22173, 21970, 21969, 22172, 21969, 21966, 21950, 22142, 22168, 22304, 22306, 22305, 22310, 22308, 22307, 22296, 22304, 22303, 22308, 22298, 22297, 22303, 22305, 22300, 22309, 22307, 22301, 22297, 22298, 22296, 22299, 22300, 22302, 22294, 22297, 22295, 22299, 22301, 22294, 22295, 22303, 22299, 22307, 22297, 22294, 22317, 22319, 22318, 22315, 22319, 22317, 22313, 22317, 22316, 22311, 22314, 22317, 22328, 22330, 22329, 22326, 22330, 22328, 22334, 22332, 22331, 22337, 22338, 22336, 22341, 22342, 22340, 22346, 22348, 22347, 22350, 22348, 22346, 22321, 22335, 22336, 22329, 22344, 22343, 22347, 22334, 22333, 22322, 22328, 22327, 22320, 22325, 22328, 22324, 22333, 22331, 22339, 22340, 22346, 22342, 22349, 22346, 22320, 22322, 22324, 22336, 22338, 22341, 22336, 22339, 22324, 22321, 22327, 22343, 22345, 22333, 22324, 22306, 22326, 22325, 22332, 22310, 22309, 22300, 22320, 22323, 22305, 22325, 22320, 22323, 22331, 22309, 22311, 22313, 22337, 22343, 22344, 22315, 22335, 22343, 22314, 22354, 22356, 22355, 22360, 22362, 22361, 22356, 22360, 22359, 22353, 22355, 22352,
22359, 22361, 22358, 22355, 22359, 22357, 22365, 22366, 22364, 22370, 22368, 22367, 22373, 22374, 22372, 22378, 22376, 22375, 22371, 22372, 22366, 22376, 22370, 22369, 22369, 22367, 22364, 22377, 22375, 22372, 22375, 22369, 22366, 22382, 22384, 22383, 22391, 22393, 22392, 22389, 22393, 22391, 22384, 22389, 22388, 22381, 22383, 22380, 22387, 22391, 22390, 22388, 22391, 22387, 22383, 22388, 22385, 22396, 22397, 22395, 22401, 22399, 22398, 22402, 22404, 22405, 22409, 22407, 22406, 22396, 22402, 22403, 22407, 22401, 22400, 22400, 22398, 22395, 22408, 22406, 22403, 22406, 22400, 22397, 22362, 22382, 22381, 22361, 22381, 22379, 22373, 22394, 22395, 22399, 22378, 22377, 22395, 22398, 22377, 22352, 22365, 22363, 22357, 22358, 22373, 22357, 22371, 22365, 22379, 22380, 22396, 22385, 22387, 22404, 22380, 22385, 22402, 22358, 22379, 22394, 22413, 22414, 22412, 22418, 22421, 22419, 22431, 22429, 22428, 22424, 22425, 22414, 22416, 22427, 22426, 22431, 22422, 22420, 22411, 22417, 22419, 22429, 22425, 22424, 22419, 22421, 22416, 22430, 22428, 22426, 22413, 22419, 22415, 22428, 22424, 22423, 22410, 22423, 22424, 22410, 22415, 22426, 22429, 22420, 22414, 22436, 22438, 22439, 22443, 22441, 22440, 22446, 22442, 22440, 22443, 22447, 22445, 22457, 22455, 22454,
22452, 22453, 22439, 22451, 22456, 22454, 22440, 22441, 22433, 22444, 22440, 22432, 22441, 22445, 22435, 22454, 22455, 22453, 22435, 22438, 22436, 22451, 22450, 22445, 22445, 22450, 22438, 22438, 22450, 22454, 22412, 22414, 22434, 22446, 22444, 22420, 22420, 22444, 22434, 22460, 22461, 22459, 22464, 22465, 22463, 22467, 22469, 22468, 22479, 22477, 22476, 22483, 22481, 22480, 22460, 22472, 22473, 22464, 22474, 22475, 22458, 22459, 22467, 22476, 22477, 22473, 22481, 22475, 22474, 22459, 22461, 22464, 22470, 22471, 22469, 22482, 22480, 22477, 22462, 22470, 22467, 22480, 22474, 22473, 22473, 22474, 22464, 22490, 22488, 22487, 22492, 22489, 22487, 22495, 22492, 22491, 22488, 22485, 22484, 22486, 22491, 22487, 22493, 22494, 22491, 22463, 22465, 22486, 22495, 22494, 22481, 22494, 22493, 22475, 22465, 22475, 22493, 22312, 22411, 22412, 22318, 22418, 22417, 22316, 22417, 22411, 22337, 22432, 22433, 22436, 22437, 22342, 22449, 22350, 22349, 22433, 22436, 22341, 22437, 22448, 22349, 22412, 22432, 22337, 22386, 22462, 22463, 22392, 22471, 22470, 22386, 22390, 22470, 22404, 22484, 22485, 22490, 22409, 22408, 22485, 22488, 22408, 22463, 22484, 22404, 22439, 22460, 22458, 22466, 22468, 22449, 22478, 22476, 22455, 22437, 22458, 22466, 22455, 22476, 22472,
22453, 22472, 22460, 22498, 22499, 22497, 22501, 22503, 22502, 22502, 22505, 22504, 22508, 22509, 22507, 22511, 22513, 22512, 22506, 22507, 22499, 22505, 22511, 22510, 22497, 22501, 22500, 22500, 22504, 22498, 22510, 22512, 22508, 22504, 22510, 22506, 22516, 22517, 22515, 22519, 22521, 22520, 22524, 22525, 22523, 22527, 22529, 22528, 22522, 22523, 22517, 22521, 22527, 22526, 22518, 22520, 22516, 22526, 22528, 22524, 22520, 22526, 22522, 22514, 22515, 22509, 22513, 22519, 22518, 22512, 22518, 22514, 22532, 22533, 22531, 22536, 22537, 22535, 22539, 22541, 22540, 22543, 22545, 22544, 22548, 22549, 22547, 22552, 22553, 22551, 22546, 22547, 22533, 22550, 22551, 22537, 22531, 22539, 22538, 22535, 22543, 22542, 22533, 22536, 22534, 22542, 22544, 22541, 22549, 22552, 22550, 22534, 22542, 22539, 22547, 22550, 22536, 22556, 22557, 22555, 22560, 22561, 22559, 22563, 22565, 22564, 22567, 22569, 22568, 22572, 22573, 22571, 22576, 22577, 22575, 22570, 22571, 22557, 22574, 22575, 22561, 22555, 22563, 22562, 22559, 22567, 22566, 22557, 22560, 22558, 22566, 22568, 22565, 22573, 22576, 22574, 22558, 22566, 22563, 22571, 22574, 22560, 22580, 22581, 22579, 22584, 22585, 22583, 22588, 22589, 22587, 22592, 22593, 22591, 22586, 22587, 22581, 22590, 22591, 22585,
22581, 22584, 22582, 22589, 22592, 22590, 22587, 22590, 22584, 22594, 22596, 22597, 22600, 22601, 22599, 22605, 22603, 22602, 22609, 22607, 22606, 22615, 22613, 22612, 22613, 22604, 22602, 22603, 22597, 22596, 22614, 22612, 22610, 22607, 22601, 22600, 22597, 22600, 22598, 22608, 22606, 22603, 22606, 22600, 22597, 22612, 22602, 22596, 22578, 22579, 22549, 22582, 22583, 22553, 22579, 22582, 22552, 22594, 22595, 22573, 22598, 22599, 22577, 22595, 22598, 22576, 22537, 22556, 22554, 22562, 22564, 22545, 22553, 22572, 22570, 22554, 22562, 22543, 22551, 22570, 22556, 22583, 22585, 22596, 22593, 22611, 22610, 22585, 22591, 22610, 22583, 22594, 22572, 22618, 22619, 22617, 22621, 22623, 22622, 22626, 22627, 22625, 22629, 22631, 22630, 22624, 22625, 22619, 22623, 22629, 22628, 22620, 22622, 22618, 22628, 22630, 22626, 22622, 22628, 22624, 22634, 22635, 22633, 22637, 22639, 22638, 22642, 22643, 22641, 22645, 22647, 22646, 22640, 22641, 22635, 22639, 22645, 22644, 22636, 22638, 22634, 22644, 22646, 22642, 22638, 22644, 22640, 22632, 22633, 22627, 22631, 22637, 22636, 22630, 22636, 22632, 22650, 22651, 22649, 22654, 22655, 22653, 22658, 22659, 22657, 22662, 22663, 22661, 22656, 22657, 22651, 22660, 22661, 22655, 22651, 22654, 22652, 22659, 22662, 22660,
22657, 22660, 22654, 22669, 22667, 22666, 22675, 22673, 22672, 22673, 22669, 22668, 22668, 22666, 22664, 22674, 22672, 22670, 22672, 22668, 22665, 22678, 22679, 22677, 22682, 22683, 22681, 22686, 22687, 22685, 22690, 22691, 22689, 22684, 22685, 22679, 22688, 22689, 22683, 22679, 22682, 22680, 22687, 22690, 22688, 22685, 22688, 22682, 22697, 22695, 22694, 22703, 22701, 22700, 22701, 22697, 22696, 22696, 22694, 22692, 22702, 22700, 22698, 22700, 22696, 22693, 22676, 22677, 22659, 22680, 22681, 22663, 22677, 22680, 22662, 22695, 22675, 22674, 22694, 22674, 22671, 22655, 22665, 22664, 22663, 22671, 22670, 22661, 22670, 22665, 22683, 22693, 22692, 22691, 22699, 22698, 22689, 22698, 22693, 22681, 22692, 22671, 22616, 22617, 22525, 22529, 22621, 22620, 22528, 22620, 22616, 22648, 22649, 22589, 22652, 22653, 22593, 22649, 22652, 22592, 22667, 22615, 22614, 22666, 22614, 22611, 22653, 22664, 22611, 22499, 22532, 22530, 22538, 22540, 22503, 22509, 22548, 22546, 22530, 22538, 22501, 22507, 22546, 22532, 22517, 22580, 22578, 22525, 22588, 22586, 22523, 22586, 22580, 22515, 22578, 22548, 22619, 22650, 22648, 22627, 22658, 22656, 22625, 22656, 22650, 22635, 22678, 22676, 22643, 22686, 22684, 22641, 22684, 22678, 22633, 22676, 22658, 22617, 22648, 22588,
22706, 22707, 22705, 22709, 22711, 22710, 22714, 22715, 22713, 22717, 22719, 22718, 22712, 22713, 22707, 22711, 22717, 22716, 22708, 22710, 22706, 22716, 22718, 22714, 22710, 22716, 22712, 22722, 22723, 22721, 22725, 22727, 22726, 22730, 22731, 22729, 22733, 22735, 22734, 22728, 22729, 22723, 22727, 22733, 22732, 22724, 22726, 22722, 22732, 22734, 22730, 22726, 22732, 22728, 22720, 22721, 22715, 22719, 22725, 22724, 22718, 22724, 22720, 22738, 22739, 22737, 22742, 22743, 22741, 22746, 22747, 22745, 22750, 22751, 22749, 22744, 22745, 22739, 22748, 22749, 22743, 22739, 22742, 22740, 22747, 22750, 22748, 22745, 22748, 22742, 22757, 22755, 22754, 22763, 22761, 22760, 22761, 22757, 22756, 22756, 22754, 22752, 22762, 22760, 22758, 22760, 22756, 22753, 22766, 22767, 22765, 22770, 22771, 22769, 22774, 22775, 22773, 22778, 22779, 22777, 22772, 22773, 22767, 22776, 22777, 22771, 22767, 22770, 22768, 22775, 22778, 22776, 22773, 22776, 22770, 22785, 22783, 22782, 22791, 22789, 22788, 22789, 22785, 22784, 22784, 22782, 22780, 22790, 22788, 22786, 22788, 22784, 22781, 22764, 22765, 22747, 22768, 22769, 22751, 22765, 22768, 22750, 22783, 22763, 22762, 22782, 22762, 22759, 22743, 22753, 22752, 22751, 22759, 22758, 22749, 22758, 22753, 22771, 22781, 22780,
22779, 22787, 22786, 22777, 22786, 22781, 22769, 22780, 22759, 22794, 22795, 22793, 22797, 22799, 22798, 22802, 22803, 22801, 22805, 22807, 22806, 22800, 22801, 22795, 22799, 22805, 22804, 22796, 22798, 22794, 22804, 22806, 22802, 22798, 22804, 22800, 22813, 22811, 22810, 22811, 22809, 22808, 22808, 22809, 22803, 22807, 22812, 22810, 22806, 22810, 22808, 22816, 22817, 22815, 22820, 22821, 22819, 22824, 22825, 22823, 22828, 22829, 22827, 22822, 22823, 22817, 22826, 22827, 22821, 22817, 22820, 22818, 22825, 22828, 22826, 22823, 22826, 22820, 22835, 22833, 22832, 22841, 22839, 22838, 22839, 22835, 22834, 22834, 22832, 22830, 22840, 22838, 22836, 22838, 22834, 22831, 22849, 22847, 22846, 22853, 22851, 22850, 22847, 22843, 22842, 22851, 22845, 22844, 22852, 22850, 22847, 22850, 22844, 22843, 22842, 22843, 22825, 22844, 22845, 22829, 22843, 22844, 22828, 22856, 22841, 22840, 22855, 22840, 22837, 22821, 22831, 22830, 22829, 22837, 22836, 22827, 22836, 22831, 22856, 22855, 22851, 22855, 22854, 22845, 22845, 22854, 22837, 22792, 22793, 22731, 22735, 22797, 22796, 22734, 22796, 22792, 22814, 22815, 22775, 22818, 22819, 22779, 22815, 22818, 22778, 22833, 22791, 22790, 22832, 22790, 22787, 22819, 22830, 22787, 22707, 22738, 22736, 22715, 22746, 22744,
22713, 22744, 22738, 22723, 22766, 22764, 22731, 22774, 22772, 22729, 22772, 22766, 22721, 22764, 22746, 22795, 22816, 22814, 22803, 22824, 22822, 22801, 22822, 22816, 22848, 22846, 22811, 22846, 22842, 22809, 22809, 22842, 22824, 22793, 22814, 22774, 22704, 22705, 22643, 22647, 22709, 22708, 22646, 22708, 22704, 22736, 22737, 22687, 22740, 22741, 22691, 22737, 22740, 22690, 22755, 22703, 22702, 22754, 22702, 22699, 22741, 22752, 22699, 22705, 22736, 22686, 22861, 22862, 22860, 22864, 22866, 22865, 22868, 22870, 22869, 22873, 22875, 22876, 22878, 22880, 22879, 22884, 22882, 22881, 22873, 22874, 22862, 22866, 22878, 22877, 22863, 22865, 22858, 22877, 22879, 22872, 22860, 22868, 22867, 22883, 22881, 22874, 22858, 22861, 22859, 22867, 22869, 22864, 22872, 22875, 22873, 22859, 22867, 22863, 22871, 22873, 22861, 22865, 22877, 22871, 22887, 22888, 22886, 22891, 22892, 22890, 22894, 22896, 22895, 22898, 22900, 22899, 22908, 22906, 22905, 22912, 22910, 22909, 22901, 22902, 22888, 22903, 22904, 22892, 22886, 22894, 22893, 22906, 22902, 22901, 22890, 22898, 22897, 22910, 22904, 22903, 22888, 22891, 22889, 22897, 22899, 22896, 22911, 22909, 22906, 22889, 22897, 22894, 22909, 22903, 22902, 22902, 22903, 22891, 22917, 22918, 22916, 22920, 22922, 22921,
22926, 22924, 22923, 22931, 22932, 22930, 22934, 22936, 22935, 22940, 22938, 22937, 22929, 22930, 22918, 22922, 22934, 22933, 22938, 22926, 22925, 22919, 22921, 22914, 22927, 22933, 22935, 22925, 22923, 22916, 22932, 22939, 22937, 22914, 22917, 22915, 22927, 22928, 22931, 22927, 22929, 22917, 22921, 22933, 22927, 22937, 22925, 22918, 22941, 22943, 22944, 22946, 22948, 22947, 22947, 22950, 22949, 22951, 22953, 22954, 22956, 22958, 22957, 22943, 22951, 22952, 22950, 22956, 22955, 22941, 22942, 22946, 22941, 22945, 22949, 22955, 22957, 22953, 22949, 22955, 22951, 22915, 22916, 22876, 22880, 22920, 22919, 22924, 22884, 22883, 22913, 22915, 22875, 22879, 22919, 22913, 22923, 22883, 22876, 22862, 22887, 22885, 22893, 22895, 22870, 22907, 22905, 22881, 22885, 22893, 22868, 22905, 22901, 22874, 22874, 22901, 22887, 22961, 22962, 22960, 22965, 22966, 22964, 22968, 22970, 22969, 22972, 22974, 22973, 22983, 22981, 22980, 22987, 22985, 22984, 22987, 22989, 22988, 22975, 22976, 22962, 22977, 22978, 22966, 22960, 22968, 22967, 22981, 22976, 22975, 22964, 22972, 22971, 22985, 22978, 22977, 22985, 22988, 22979, 22962, 22965, 22963, 22971, 22973, 22970, 22986, 22984, 22981, 22963, 22971, 22968, 22984, 22977, 22976, 22976, 22977, 22965, 22992, 22993, 22991,
22995, 22997, 22996, 22999, 22997, 22995, 23002, 23003, 23001, 23007, 23005, 23004, 23000, 23001, 22993, 23005, 22999, 22998, 22991, 22995, 22994, 22998, 22995, 22991, 23006, 23004, 23001, 23004, 22998, 22993, 23015, 23013, 23012, 23017, 23019, 23018, 23027, 23025, 23024, 23029, 23031, 23030, 23025, 23015, 23014, 23019, 23029, 23028, 23009, 23014, 23012, 23021, 23026, 23024, 23016, 23018, 23011, 23028, 23030, 23023, 23020, 23024, 23014, 23018, 23028, 23022, 23034, 23035, 23033, 23039, 23037, 23036, 23042, 23043, 23041, 23047, 23045, 23044, 23040, 23041, 23035, 23045, 23039, 23038, 23038, 23036, 23033, 23046, 23044, 23041, 23044, 23038, 23035, 22989, 23017, 23016, 22988, 23016, 23010, 23032, 23033, 23003, 23037, 23007, 23006, 23036, 23006, 23003, 22966, 22992, 22990, 22994, 22996, 22974, 22979, 23002, 23000, 22990, 22994, 22972, 22978, 23000, 22992, 23011, 23034, 23032, 23023, 23042, 23040, 23011, 23022, 23040, 23010, 23032, 23002, 23050, 23052, 23053, 23055, 23057, 23056, 23063, 23059, 23058, 23064, 23066, 23067, 23070, 23073, 23072, 23064, 23065, 23053, 23062, 23070, 23068, 23071, 23063, 23061, 23054, 23056, 23049, 23053, 23061, 23058, 23064, 23068, 23072, 23049, 23052, 23050, 23062, 23060, 23056, 23056, 23060, 23052, 23060, 23068, 23064,
23069, 23061, 23053, 23074, 23076, 23077, 23079, 23081, 23080, 23085, 23087, 23088, 23092, 23094, 23093, 23093, 23090, 23089, 23096, 23098, 23097, 23076, 23085, 23086, 23081, 23096, 23095, 23083, 23084, 23092, 23091, 23089, 23082, 23078, 23080, 23076, 23095, 23097, 23087, 23097, 23098, 23094, 23087, 23097, 23092, 23080, 23095, 23085, 23104, 23102, 23101, 23101, 23102, 23100, 23107, 23108, 23106, 23111, 23112, 23110, 23118, 23116, 23115, 23114, 23117, 23115, 23122, 23120, 23119, 23116, 23108, 23107, 23105, 23113, 23115, 23119, 23120, 23112, 23106, 23108, 23111, 23121, 23119, 23116, 23119, 23111, 23108, 23099, 23100, 23067, 23073, 23103, 23101, 23066, 23072, 23101, 23083, 23105, 23106, 23109, 23110, 23088, 23084, 23106, 23109, 23067, 23083, 23082, 23089, 23090, 23071, 23065, 23082, 23089, 23114, 23113, 23102, 23113, 23105, 23100, 23067, 23100, 23105, 23130, 23128, 23127, 23132, 23134, 23133, 23137, 23138, 23136, 23139, 23141, 23142, 23144, 23146, 23145, 23148, 23146, 23144, 23150, 23152, 23151, 23151, 23154, 23153, 23145, 23130, 23129, 23134, 23152, 23150, 23124, 23129, 23127, 23135, 23136, 23144, 23147, 23144, 23136, 23131, 23133, 23126, 23140, 23150, 23149, 23139, 23149, 23153, 23153, 23154, 23148, 23141, 23153, 23147, 23143, 23129, 23124,
23126, 23133, 23150, 23155, 23157, 23158, 23162, 23160, 23159, 23169, 23167, 23166, 23171, 23168, 23166, 23157, 23163, 23164, 23169, 23162, 23161, 23161, 23159, 23156, 23167, 23164, 23163, 23165, 23170, 23166, 23164, 23167, 23161, 23174, 23175, 23173, 23182, 23180, 23179, 23186, 23184, 23183, 23188, 23185, 23183, 23179, 23180, 23175, 23184, 23177, 23176, 23187, 23183, 23176, 23173, 23175, 23178, 23188, 23187, 23180, 23180, 23187, 23178, 23137, 23172, 23173, 23176, 23177, 23142, 23173, 23176, 23141, 23191, 23171, 23170, 23190, 23170, 23165, 23125, 23126, 23157, 23142, 23165, 23163, 23140, 23163, 23157, 23191, 23190, 23184, 23190, 23189, 23177, 23142, 23177, 23189, 22931, 23050, 23051, 22936, 23055, 23054, 23059, 22940, 22939, 23048, 23050, 22931, 22928, 22935, 23054, 23051, 23058, 22939, 22953, 23074, 23075, 22958, 23079, 23078, 22957, 23078, 23074, 23128, 23027, 23026, 23031, 23132, 23131, 23123, 23127, 23026, 23030, 23131, 23125, 23155, 23156, 23043, 23160, 23047, 23046, 23159, 23046, 23043, 23023, 23125, 23155, 22892, 22961, 22959, 22967, 22969, 22900, 22982, 22980, 22910, 22959, 22967, 22898, 22980, 22975, 22904, 22904, 22975, 22961, 22944, 23009, 23008, 23012, 23013, 22948, 22952, 22954, 23021, 23008, 23012, 22946, 22952, 23020, 23009,
23077, 23124, 23123, 23086, 23088, 23137, 23086, 23135, 23124, 23110, 23112, 23174, 23181, 23179, 23120, 23120, 23179, 23174, 23088, 23110, 23172, 23075, 23123, 23021, 23201, 23203, 23204, 23208, 23210, 23209, 23206, 23210, 23208, 23212, 23214, 23213, 23197, 23214, 23212, 23200, 23208, 23207, 23205, 23208, 23200, 23201, 23202, 23212, 23198, 23200, 23203, 23211, 23213, 23206, 23201, 23211, 23205, 23196, 23212, 23202, 23216, 23218, 23219, 23223, 23225, 23224, 23224, 23227, 23226, 23233, 23234, 23232, 23238, 23240, 23239, 23236, 23240, 23238, 23231, 23232, 23219, 23221, 23236, 23235, 23230, 23238, 23237, 23235, 23238, 23230, 23216, 23217, 23223, 23216, 23222, 23226, 23226, 23227, 23221, 23228, 23230, 23233, 23218, 23226, 23220, 23215, 23228, 23231, 23220, 23235, 23228, 23243, 23244, 23242, 23247, 23248, 23246, 23251, 23252, 23250, 23255, 23256, 23254, 23249, 23250, 23244, 23253, 23254, 23248, 23244, 23247, 23245, 23252, 23255, 23253, 23250, 23253, 23247, 23194, 23225, 23223, 23193, 23223, 23217, 23199, 23241, 23242, 23245, 23246, 23204, 23242, 23245, 23203, 23207, 23209, 23194, 23199, 23207, 23193, 23219, 23243, 23241, 23234, 23251, 23249, 23232, 23249, 23243, 23192, 23217, 23241, 23263, 23265, 23264, 23259, 23263, 23262, 23262, 23264, 23261,
23258, 23262, 23260, 23269, 23271, 23270, 23278, 23280, 23279, 23276, 23280, 23278, 23271, 23276, 23275, 23268, 23270, 23267, 23274, 23278, 23277, 23272, 23275, 23278, 23270, 23275, 23272, 23265, 23269, 23268, 23264, 23268, 23266, 23281, 23283, 23284, 23287, 23288, 23286, 23290, 23292, 23291, 23295, 23296, 23294, 23299, 23300, 23298, 23293, 23294, 23284, 23297, 23298, 23288, 23289, 23291, 23283, 23284, 23287, 23285, 23296, 23299, 23297, 23294, 23297, 23287, 23303, 23304, 23302, 23305, 23307, 23308, 23314, 23312, 23311, 23314, 23310, 23309, 23315, 23317, 23318, 23323, 23321, 23320, 23315, 23316, 23304, 23325, 23313, 23311, 23322, 23320, 23316, 23311, 23312, 23308, 23308, 23312, 23309, 23304, 23307, 23305, 23325, 23324, 23320, 23324, 23319, 23316, 23304, 23316, 23319, 23319, 23324, 23311, 23328, 23329, 23327, 23332, 23333, 23331, 23336, 23337, 23335, 23340, 23341, 23339, 23334, 23335, 23329, 23338, 23339, 23333, 23329, 23332, 23330, 23337, 23340, 23338, 23335, 23338, 23332, 23348, 23346, 23345, 23350, 23347, 23345, 23356, 23354, 23353, 23354, 23350, 23349, 23346, 23343, 23342, 23344, 23349, 23345, 23355, 23353, 23351, 23353, 23349, 23344, 23326, 23327, 23296, 23330, 23331, 23300, 23327, 23330, 23299, 23317, 23342, 23343, 23348, 23323, 23322,
23343, 23346, 23322, 23288, 23303, 23301, 23300, 23317, 23315, 23298, 23315, 23303, 23331, 23333, 23344, 23339, 23341, 23352, 23333, 23339, 23351, 23331, 23342, 23317, 23229, 23281, 23282, 23285, 23286, 23234, 23239, 23290, 23289, 23282, 23285, 23233, 23237, 23289, 23281, 23301, 23302, 23252, 23305, 23306, 23256, 23302, 23305, 23255, 23286, 23301, 23251, 23291, 23292, 23259, 23260, 23261, 23295, 23283, 23291, 23258, 23257, 23260, 23293, 23266, 23267, 23328, 23274, 23336, 23334, 23267, 23272, 23334, 23261, 23266, 23326, 23361, 23362, 23360, 23364, 23366, 23365, 23368, 23370, 23369, 23363, 23365, 23358, 23360, 23368, 23367, 23358, 23361, 23359, 23367, 23369, 23364, 23359, 23367, 23363, 23375, 23376, 23374, 23379, 23380, 23378, 23382, 23384, 23383, 23388, 23386, 23385, 23390, 23392, 23391, 23396, 23394, 23393, 23374, 23382, 23381, 23386, 23376, 23375, 23378, 23390, 23389, 23394, 23380, 23379, 23376, 23379, 23377, 23389, 23391, 23384, 23395, 23393, 23386, 23377, 23389, 23382, 23393, 23379, 23376, 23362, 23375, 23373, 23381, 23383, 23370, 23387, 23385, 23371, 23373, 23381, 23368, 23385, 23375, 23362, 23399, 23400, 23398, 23403, 23404, 23402, 23406, 23408, 23407, 23412, 23410, 23409, 23414, 23416, 23415, 23420, 23418, 23417, 23398, 23406, 23405,
23410, 23400, 23399, 23402, 23414, 23413, 23418, 23404, 23403, 23400, 23403, 23401, 23413, 23415, 23408, 23419, 23417, 23410, 23401, 23413, 23406, 23417, 23403, 23400, 23423, 23424, 23422, 23426, 23428, 23427, 23432, 23430, 23429, 23432, 23428, 23426, 23422, 23426, 23425, 23430, 23424, 23423, 23430, 23426, 23422, 23404, 23423, 23421, 23425, 23427, 23416, 23431, 23429, 23418, 23421, 23425, 23414, 23429, 23423, 23404, 23437, 23438, 23436, 23440, 23442, 23441, 23446, 23444, 23443, 23451, 23452, 23450, 23454, 23456, 23455, 23460, 23458, 23457, 23449, 23450, 23438, 23442, 23454, 23453, 23458, 23446, 23445, 23439, 23441, 23434, 23453, 23455, 23448, 23445, 23443, 23436, 23459, 23457, 23450, 23434, 23437, 23435, 23448, 23451, 23449, 23447, 23449, 23437, 23441, 23453, 23447, 23457, 23445, 23438, 23465, 23466, 23464, 23468, 23470, 23469, 23474, 23472, 23471, 23479, 23480, 23478, 23482, 23484, 23483, 23490, 23488, 23487, 23490, 23486, 23485, 23477, 23478, 23466, 23470, 23482, 23481, 23486, 23474, 23473, 23467, 23469, 23462, 23481, 23483, 23476, 23473, 23471, 23464, 23488, 23480, 23479, 23488, 23485, 23478, 23462, 23465, 23463, 23476, 23479, 23477, 23489, 23487, 23483, 23487, 23479, 23476, 23475, 23477, 23465, 23469, 23481, 23475, 23485, 23473, 23466,
23463, 23464, 23452, 23456, 23468, 23467, 23472, 23460, 23459, 23461, 23463, 23451, 23455, 23467, 23461, 23471, 23459, 23452, 23435, 23436, 23362, 23366, 23440, 23439, 23444, 23372, 23371, 23433, 23435, 23361, 23365, 23439, 23433, 23443, 23371, 23362, 23380, 23399, 23397, 23405, 23407, 23392, 23411, 23409, 23394, 23397, 23405, 23390, 23409, 23399, 23380, 23491, 23493, 23494, 23496, 23498, 23497, 23501, 23502, 23500, 23504, 23506, 23505, 23493, 23499, 23500, 23498, 23504, 23503, 23495, 23497, 23493, 23503, 23505, 23501, 23497, 23503, 23499, 23509, 23510, 23508, 23512, 23514, 23513, 23517, 23518, 23516, 23524, 23522, 23521, 23520, 23523, 23521, 23515, 23516, 23510, 23514, 23520, 23519, 23507, 23511, 23513, 23521, 23522, 23518, 23519, 23521, 23517, 23509, 23513, 23519, 23507, 23508, 23502, 23506, 23512, 23511, 23501, 23505, 23511, 23527, 23528, 23526, 23531, 23532, 23530, 23535, 23536, 23534, 23539, 23540, 23538, 23533, 23534, 23528, 23537, 23538, 23532, 23528, 23531, 23529, 23536, 23539, 23537, 23534, 23537, 23531, 23546, 23544, 23543, 23552, 23550, 23549, 23550, 23546, 23545, 23545, 23543, 23541, 23551, 23549, 23547, 23549, 23545, 23542, 23555, 23556, 23554, 23559, 23560, 23558, 23563, 23564, 23562, 23567, 23568, 23566, 23561, 23562, 23556,
23565, 23566, 23560, 23556, 23559, 23557, 23564, 23567, 23565, 23562, 23565, 23559, 23574, 23572, 23571, 23580, 23578, 23577, 23578, 23574, 23573, 23570, 23573, 23571, 23576, 23579, 23577, 23575, 23577, 23573, 23553, 23554, 23536, 23557, 23558, 23540, 23554, 23557, 23539, 23572, 23552, 23551, 23569, 23571, 23551, 23530, 23532, 23542, 23538, 23540, 23548, 23532, 23538, 23547, 23560, 23570, 23569, 23568, 23576, 23575, 23566, 23575, 23570, 23558, 23569, 23548, 23584, 23586, 23585, 23590, 23592, 23591, 23586, 23590, 23589, 23581, 23583, 23585, 23587, 23589, 23591, 23582, 23585, 23589, 23592, 23595, 23594, 23591, 23594, 23593, 23598, 23599, 23597, 23602, 23603, 23601, 23606, 23607, 23605, 23610, 23611, 23609, 23604, 23605, 23599, 23608, 23609, 23603, 23599, 23602, 23600, 23607, 23610, 23608, 23605, 23608, 23602, 23618, 23620, 23619, 23619, 23616, 23615, 23623, 23624, 23622, 23629, 23627, 23626, 23621, 23622, 23614, 23627, 23620, 23618, 23613, 23614, 23618, 23613, 23617, 23615, 23624, 23628, 23626, 23630, 23631, 23629, 23625, 23630, 23628, 23622, 23626, 23618, 23634, 23635, 23633, 23638, 23639, 23637, 23641, 23643, 23642, 23646, 23647, 23645, 23650, 23651, 23649, 23657, 23655, 23654, 23653, 23656, 23654, 23644, 23645, 23635, 23648, 23649, 23639,
23643, 23653, 23652, 23640, 23642, 23634, 23655, 23647, 23646, 23644, 23652, 23654, 23635, 23638, 23636, 23645, 23647, 23650, 23645, 23648, 23638, 23634, 23642, 23652, 23660, 23661, 23659, 23668, 23670, 23669, 23669, 23666, 23665, 23673, 23674, 23672, 23675, 23677, 23678, 23671, 23672, 23661, 23675, 23676, 23664, 23680, 23670, 23668, 23664, 23668, 23667, 23663, 23667, 23665, 23661, 23663, 23662, 23674, 23677, 23675, 23661, 23672, 23675, 23676, 23679, 23668, 23606, 23632, 23633, 23636, 23637, 23611, 23633, 23636, 23610, 23623, 23658, 23659, 23666, 23631, 23630, 23659, 23662, 23625, 23662, 23665, 23630, 23603, 23613, 23612, 23611, 23623, 23621, 23603, 23609, 23621, 23639, 23660, 23658, 23651, 23673, 23671, 23649, 23671, 23660, 23637, 23658, 23623, 23524, 23584, 23583, 23522, 23583, 23581, 23596, 23597, 23564, 23600, 23601, 23568, 23597, 23600, 23567, 23616, 23580, 23579, 23612, 23615, 23579, 23601, 23612, 23576, 23494, 23527, 23525, 23502, 23535, 23533, 23500, 23533, 23527, 23510, 23555, 23553, 23516, 23518, 23563, 23516, 23561, 23555, 23508, 23553, 23535, 23582, 23598, 23596, 23588, 23606, 23604, 23587, 23604, 23598, 23641, 23640, 23594, 23594, 23640, 23632, 23593, 23632, 23606, 23581, 23596, 23563, 23273, 23491, 23492, 23279, 23496, 23495,
23273, 23277, 23495, 23525, 23526, 23337, 23529, 23530, 23341, 23526, 23529, 23340, 23544, 23356, 23355, 23543, 23355, 23352, 23341, 23530, 23541, 23492, 23525, 23336, 22278, 22354, 22353, 22276, 22353, 22351, 22363, 22364, 22288, 22368, 22292, 22291, 22367, 22291, 22288, 22270, 22351, 22363, 23685, 23686, 23684, 23688, 23690, 23689, 23692, 23694, 23693, 23697, 23699, 23700, 23702, 23704, 23703, 23708, 23706, 23705, 23685, 23697, 23698, 23690, 23702, 23701, 23687, 23689, 23682, 23701, 23703, 23696, 23684, 23692, 23691, 23707, 23705, 23698, 23682, 23685, 23683, 23691, 23693, 23688, 23696, 23699, 23697, 23683, 23691, 23687, 23695, 23697, 23685, 23689, 23701, 23695, 23711, 23712, 23710, 23715, 23716, 23714, 23718, 23720, 23719, 23722, 23724, 23723, 23734, 23732, 23731, 23738, 23736, 23735, 23740, 23742, 23741, 23725, 23726, 23712, 23727, 23728, 23716, 23710, 23718, 23717, 23732, 23726, 23725, 23714, 23722, 23721, 23736, 23728, 23727, 23730, 23740, 23739, 23712, 23715, 23713, 23721, 23723, 23720, 23737, 23735, 23732, 23713, 23721, 23718, 23735, 23727, 23726, 23726, 23727, 23715, 23747, 23748, 23746, 23750, 23752, 23751, 23756, 23754, 23753, 23761, 23762, 23760, 23764, 23766, 23765, 23770, 23768, 23767, 23759, 23760, 23748, 23752, 23764, 23763,
23768, 23756, 23755, 23749, 23751, 23744, 23763, 23765, 23758, 23755, 23753, 23746, 23769, 23767, 23760, 23744, 23747, 23745, 23758, 23761, 23759, 23757, 23759, 23747, 23751, 23763, 23757, 23767, 23755, 23748, 23773, 23774, 23772, 23776, 23778, 23777, 23781, 23782, 23780, 23784, 23786, 23785, 23779, 23780, 23774, 23778, 23784, 23783, 23775, 23777, 23773, 23783, 23785, 23781, 23777, 23783, 23779, 23745, 23746, 23700, 23704, 23750, 23749, 23754, 23708, 23707, 23743, 23745, 23699, 23703, 23749, 23743, 23753, 23707, 23700, 23771, 23772, 23730, 23741, 23776, 23775, 23739, 23775, 23771, 23686, 23711, 23709, 23717, 23719, 23694, 23733, 23731, 23705, 23709, 23717, 23692, 23731, 23725, 23698, 23698, 23725, 23711, 23789, 23790, 23788, 23793, 23794, 23792, 23796, 23798, 23797, 23800, 23802, 23801, 23812, 23810, 23809, 23818, 23816, 23815, 23818, 23820, 23819, 23803, 23804, 23790, 23806, 23807, 23794, 23788, 23796, 23795, 23810, 23804, 23803, 23792, 23800, 23799, 23816, 23807, 23806, 23816, 23819, 23808, 23790, 23793, 23791, 23799, 23801, 23798, 23817, 23815, 23810, 23791, 23799, 23796, 23815, 23806, 23804, 23804, 23806, 23793, 23823, 23824, 23822, 23826, 23828, 23827, 23830, 23828, 23826, 23833, 23834, 23832, 23838, 23836, 23835, 23831, 23832, 23824,
23836, 23830, 23829, 23822, 23826, 23825, 23829, 23826, 23822, 23837, 23835, 23832, 23835, 23829, 23824, 23846, 23844, 23843, 23848, 23850, 23849, 23858, 23856, 23855, 23860, 23862, 23861, 23856, 23846, 23845, 23850, 23860, 23859, 23845, 23843, 23839, 23857, 23855, 23851, 23847, 23849, 23842, 23859, 23861, 23854, 23855, 23845, 23840, 23849, 23859, 23853, 23865, 23866, 23864, 23870, 23868, 23867, 23873, 23874, 23872, 23878, 23876, 23875, 23871, 23872, 23866, 23876, 23870, 23869, 23869, 23867, 23864, 23877, 23875, 23872, 23875, 23869, 23866, 23844, 23814, 23813, 23820, 23848, 23847, 23843, 23813, 23805, 23819, 23847, 23841, 23863, 23864, 23834, 23868, 23838, 23837, 23867, 23837, 23834, 23794, 23823, 23821, 23825, 23827, 23802, 23808, 23833, 23831, 23821, 23825, 23800, 23807, 23831, 23823, 23842, 23865, 23863, 23854, 23873, 23871, 23853, 23871, 23865, 23841, 23863, 23833, 23716, 23789, 23787, 23795, 23797, 23724, 23811, 23809, 23736, 23813, 23814, 23742, 23787, 23795, 23722, 23809, 23803, 23728, 23805, 23813, 23740, 23728, 23803, 23789, 23774, 23840, 23839, 23782, 23852, 23851, 23780, 23851, 23840, 23772, 23839, 23805, 23882, 23883, 23881, 23891, 23889, 23888, 23887, 23890, 23888, 23891, 23894, 23893, 23889, 23883, 23882, 23886, 23888, 23882,
23887, 23886, 23884, 23884, 23886, 23880, 23883, 23889, 23893, 23897, 23898, 23896, 23901, 23902, 23900, 23905, 23906, 23904, 23907, 23909, 23910, 23916, 23914, 23913, 23912, 23915, 23913, 23897, 23903, 23904, 23907, 23908, 23902, 23913, 23914, 23906, 23903, 23911, 23913, 23898, 23901, 23899, 23906, 23909, 23907, 23918, 23917, 23914, 23914, 23917, 23909, 23904, 23907, 23901, 23924, 23922, 23921, 23922, 23920, 23919, 23919, 23920, 23910, 23918, 23923, 23921, 23909, 23917, 23921, 23881, 23883, 23897, 23912, 23911, 23893, 23911, 23903, 23892, 23892, 23903, 23897, 23928, 23930, 23929, 23933, 23934, 23932, 23936, 23938, 23937, 23931, 23932, 23926, 23929, 23936, 23935, 23926, 23928, 23927, 23935, 23937, 23933, 23927, 23935, 23931, 23944, 23946, 23945, 23948, 23950, 23949, 23953, 23954, 23952, 23957, 23958, 23956, 23951, 23952, 23940, 23955, 23956, 23942, 23940, 23944, 23943, 23942, 23948, 23947, 23947, 23949, 23946, 23954, 23957, 23955, 23941, 23947, 23944, 23952, 23955, 23941, 23964, 23966, 23965, 23968, 23970, 23969, 23973, 23974, 23972, 23977, 23978, 23976, 23971, 23972, 23960, 23975, 23976, 23962, 23960, 23964, 23963, 23962, 23968, 23967, 23967, 23969, 23966, 23974, 23977, 23975, 23961, 23967, 23964, 23972, 23975, 23961, 23963, 23965, 23950,
23958, 23973, 23971, 23959, 23963, 23948, 23956, 23971, 23959, 23943, 23945, 23930, 23934, 23953, 23951, 23939, 23943, 23928, 23932, 23951, 23939, 23983, 23984, 23982, 23986, 23988, 23987, 23992, 23990, 23989, 24001, 23999, 23998, 23994, 23995, 23984, 23988, 23997, 23996, 24001, 23992, 23991, 23985, 23987, 23980, 23991, 23989, 23982, 23999, 23995, 23994, 23980, 23983, 23981, 24000, 23998, 23996, 23998, 23994, 23993, 23993, 23994, 23983, 23987, 23996, 23993, 23999, 23991, 23984, 24004, 24005, 24003, 24011, 24009, 24008, 24007, 24010, 24008, 24009, 24005, 24004, 24006, 24008, 24004, 24019, 24017, 24016, 24021, 24023, 24022, 24023, 24026, 24025, 24018, 24016, 24012, 24020, 24022, 24015, 24022, 24025, 24024, 24029, 24030, 24028, 24034, 24032, 24031, 24040, 24038, 24037, 24035, 24036, 24030, 24040, 24034, 24033, 24033, 24031, 24028, 24038, 24036, 24035, 24038, 24033, 24030, 24015, 24029, 24027, 24039, 24037, 24025, 24037, 24035, 24024, 24024, 24035, 24029, 24047, 24049, 24048, 24043, 24047, 24046, 24046, 24048, 24045, 24042, 24046, 24044, 24055, 24057, 24056, 24059, 24061, 24060, 24051, 24055, 24054, 24053, 24059, 24058, 24058, 24060, 24057, 24052, 24058, 24055, 24066, 24067, 24065, 24069, 24071, 24070, 24075, 24073, 24072, 24068, 24070, 24063,
24073, 24067, 24066, 24063, 24066, 24064, 24074, 24072, 24070, 24072, 24066, 24063, 24078, 24079, 24077, 24082, 24083, 24081, 24087, 24085, 24084, 24091, 24089, 24088, 24085, 24079, 24078, 24089, 24083, 24082, 24079, 24082, 24080, 24090, 24088, 24085, 24088, 24082, 24079, 24064, 24065, 24043, 24045, 24069, 24068, 24062, 24064, 24042, 24044, 24068, 24062, 24076, 24077, 24051, 24080, 24081, 24053, 24077, 24080, 24052, 24054, 24056, 24049, 24050, 24054, 24047, 24067, 24078, 24076, 24086, 24084, 24073, 24084, 24078, 24067, 24065, 24076, 24050, 24100, 24102, 24101, 24104, 24106, 24105, 24096, 24100, 24099, 24098, 24104, 24103, 24103, 24105, 24102, 24097, 24103, 24100, 24110, 24112, 24111, 24108, 24110, 24109, 24115, 24116, 24114, 24119, 24120, 24118, 24124, 24122, 24121, 24125, 24127, 24128, 24129, 24131, 24132, 24134, 24136, 24135, 24135, 24138, 24137, 24144, 24142, 24141, 24144, 24140, 24139, 24129, 24130, 24120, 24124, 24136, 24134, 24122, 24116, 24115, 24126, 24134, 24133, 24125, 24133, 24137, 24142, 24132, 24131, 24132, 24142, 24139, 24116, 24119, 24117, 24128, 24131, 24129, 24116, 24126, 24129, 24116, 24122, 24134, 24151, 24149, 24148, 24153, 24150, 24148, 24156, 24153, 24152, 24149, 24146, 24145, 24152, 24148, 24145, 24155, 24152, 24147,
24113, 24114, 24096, 24117, 24118, 24098, 24114, 24117, 24097, 24107, 24145, 24146, 24151, 24112, 24110, 24146, 24149, 24110, 24109, 24111, 24106, 24107, 24109, 24104, 24118, 24120, 24147, 24156, 24155, 24139, 24155, 24154, 24130, 24120, 24130, 24154, 24118, 24145, 24107, 24005, 24013, 24012, 24019, 24018, 24009, 24018, 24013, 24005, 24099, 24101, 24061, 24095, 24099, 24059, 24083, 24115, 24113, 24123, 24121, 24089, 24137, 24138, 24094, 24121, 24115, 24083, 24127, 24137, 24093, 24081, 24113, 24095, 24159, 24160, 24158, 24162, 24164, 24163, 24170, 24168, 24167, 24165, 24166, 24160, 24164, 24169, 24167, 24161, 24163, 24159, 24168, 24166, 24165, 24163, 24167, 24165, 24173, 24174, 24172, 24177, 24178, 24176, 24186, 24184, 24183, 24190, 24188, 24187, 24179, 24180, 24174, 24181, 24182, 24178, 24184, 24180, 24179, 24188, 24182, 24181, 24174, 24177, 24175, 24189, 24187, 24184, 24187, 24181, 24180, 24180, 24181, 24177, 24193, 24194, 24192, 24197, 24198, 24196, 24206, 24204, 24203, 24210, 24208, 24207, 24199, 24200, 24194, 24201, 24202, 24198, 24204, 24200, 24199, 24208, 24202, 24201, 24194, 24197, 24195, 24209, 24207, 24204, 24207, 24201, 24200, 24200, 24201, 24197, 24178, 24193, 24191, 24205, 24203, 24188, 24203, 24199, 24182, 24182, 24199, 24193,
24214, 24216, 24215, 24212, 24214, 24213, 24219, 24220, 24218, 24222, 24224, 24223, 24227, 24228, 24226, 24230, 24232, 24231, 24225, 24226, 24220, 24224, 24230, 24229, 24221, 24223, 24219, 24229, 24231, 24227, 24223, 24229, 24225, 24217, 24218, 24212, 24215, 24222, 24221, 24213, 24221, 24217, 24238, 24240, 24239, 24242, 24244, 24243, 24234, 24238, 24237, 24236, 24242, 24241, 24241, 24243, 24240, 24235, 24241, 24238, 24250, 24252, 24251, 24254, 24256, 24255, 24246, 24250, 24249, 24248, 24254, 24253, 24253, 24255, 24252, 24247, 24253, 24250, 24259, 24260, 24258, 24263, 24264, 24262, 24267, 24268, 24266, 24271, 24272, 24270, 24265, 24266, 24260, 24269, 24270, 24264, 24260, 24263, 24261, 24268, 24271, 24269, 24266, 24269, 24263, 24275, 24276, 24274, 24279, 24280, 24278, 24283, 24284, 24282, 24287, 24288, 24286, 24281, 24282, 24276, 24285, 24286, 24280, 24276, 24279, 24277, 24284, 24287, 24285, 24282, 24285, 24279, 24257, 24258, 24234, 24261, 24262, 24236, 24258, 24261, 24235, 24273, 24274, 24246, 24277, 24278, 24248, 24274, 24277, 24247, 24249, 24251, 24244, 24245, 24249, 24242, 24264, 24275, 24273, 24272, 24283, 24281, 24270, 24281, 24275, 24262, 24273, 24245, 24160, 24173, 24171, 24185, 24183, 24168, 24183, 24179, 24166, 24166, 24179, 24173,
24237, 24239, 24216, 24233, 24237, 24214, 24220, 24259, 24257, 24228, 24267, 24265, 24226, 24265, 24259, 24218, 24257, 24233, 23981, 23982, 23762, 23766, 23986, 23985, 23990, 23770, 23769, 23979, 23981, 23761, 23765, 23985, 23979, 23989, 23769, 23762, 24002, 24003, 23782, 23786, 24007, 24006, 23785, 24006, 24002, 24017, 23858, 23857, 23862, 24021, 24020, 24016, 23857, 23852, 23861, 24020, 24014, 24027, 24028, 23874, 24032, 23878, 23877, 24031, 23877, 23874, 24014, 24027, 23873, 24003, 24012, 23852, 24157, 24158, 23934, 23938, 24162, 24161, 23937, 24161, 24157, 24171, 24172, 23954, 24175, 24176, 23958, 24172, 24175, 23957, 24191, 24192, 23974, 24195, 24196, 23978, 24192, 24195, 23977, 24176, 24191, 23973, 24158, 24171, 23953, 24293, 24295, 24296, 24300, 24302, 24301, 24301, 24304, 24303, 24305, 24307, 24308, 24309, 24311, 24312, 24314, 24316, 24315, 24315, 24318, 24317, 24324, 24322, 24321, 24324, 24320, 24319, 24309, 24310, 24296, 24298, 24316, 24314, 24306, 24314, 24313, 24305, 24313, 24317, 24294, 24300, 24299, 24293, 24299, 24303, 24322, 24312, 24311, 24312, 24322, 24319, 24303, 24304, 24298, 24308, 24311, 24309, 24295, 24303, 24297, 24292, 24306, 24309, 24292, 24297, 24314, 24325, 24327, 24328, 24330, 24332, 24331, 24331, 24334, 24333,
24340, 24338, 24337, 24327, 24335, 24336, 24334, 24339, 24337, 24326, 24330, 24329, 24329, 24333, 24327, 24338, 24336, 24335, 24333, 24337, 24335, 24341, 24343, 24344, 24351, 24349, 24348, 24355, 24357, 24356, 24365, 24363, 24362, 24369, 24367, 24366, 24343, 24358, 24359, 24360, 24361, 24347, 24344, 24350, 24348, 24363, 24359, 24358, 24347, 24355, 24354, 24367, 24361, 24360, 24353, 24352, 24348, 24354, 24356, 24351, 24368, 24366, 24363, 24352, 24345, 24342, 24344, 24346, 24354, 24366, 24360, 24359, 24344, 24359, 24360, 24291, 24332, 24330, 24289, 24290, 24330, 24341, 24342, 24308, 24353, 24323, 24321, 24308, 24342, 24345, 24345, 24352, 24321, 24317, 24318, 24291, 24307, 24317, 24290, 24328, 24343, 24341, 24364, 24362, 24338, 24362, 24358, 24336, 24336, 24358, 24343, 24289, 24326, 24341, 24370, 24372, 24373, 24380, 24378, 24377, 24380, 24376, 24375, 24385, 24379, 24377, 24378, 24373, 24372, 24378, 24375, 24371, 24382, 24381, 24375, 24381, 24374, 24371, 24383, 24384, 24377, 24391, 24393, 24392, 24395, 24397, 24396, 24405, 24403, 24402, 24409, 24407, 24406, 24398, 24399, 24387, 24400, 24401, 24389, 24387, 24391, 24390, 24403, 24399, 24398, 24389, 24395, 24394, 24407, 24401, 24400, 24394, 24396, 24393, 24408, 24406, 24403, 24388, 24394, 24391,
24406, 24400, 24399, 24399, 24400, 24388, 24413, 24415, 24414, 24421, 24419, 24418, 24416, 24417, 24411, 24421, 24415, 24413, 24411, 24413, 24412, 24419, 24417, 24416, 24419, 24413, 24411, 24412, 24414, 24397, 24420, 24418, 24407, 24410, 24412, 24395, 24418, 24416, 24401, 24401, 24416, 24410, 24296, 24372, 24370, 24385, 24384, 24319, 24384, 24383, 24310, 24296, 24310, 24383, 24390, 24392, 24357, 24404, 24402, 24367, 24386, 24390, 24355, 24402, 24398, 24361, 24361, 24398, 24386, 24424, 24425, 24423, 24427, 24429, 24428, 24432, 24433, 24431, 24435, 24437, 24436, 24430, 24431, 24425, 24429, 24435, 24434, 24426, 24428, 24424, 24434, 24436, 24432, 24428, 24434, 24430, 24440, 24441, 24439, 24443, 24445, 24444, 24448, 24449, 24447, 24451, 24453, 24452, 24446, 24447, 24441, 24445, 24451, 24450, 24442, 24444, 24440, 24450, 24452, 24448, 24444, 24450, 24446, 24438, 24439, 24433, 24437, 24443, 24442, 24436, 24442, 24438, 24456, 24457, 24455, 24460, 24461, 24459, 24464, 24465, 24463, 24468, 24469, 24467, 24462, 24463, 24457, 24466, 24467, 24461, 24457, 24460, 24458, 24465, 24468, 24466, 24463, 24466, 24460, 24470, 24472, 24473, 24476, 24477, 24475, 24481, 24479, 24478, 24485, 24483, 24482, 24491, 24489, 24488, 24489, 24480, 24478, 24479, 24473, 24472,
24490, 24488, 24486, 24483, 24477, 24476, 24473, 24476, 24474, 24484, 24482, 24479, 24482, 24476, 24473, 24488, 24478, 24472, 24494, 24495, 24493, 24498, 24499, 24497, 24502, 24503, 24501, 24506, 24507, 24505, 24500, 24501, 24495, 24504, 24505, 24499, 24495, 24498, 24496, 24503, 24506, 24504, 24501, 24504, 24498, 24513, 24511, 24510, 24519, 24517, 24516, 24517, 24513, 24512, 24512, 24510, 24508, 24518, 24516, 24514, 24516, 24512, 24509, 24492, 24493, 24465, 24496, 24497, 24469, 24493, 24496, 24468, 24511, 24491, 24490, 24510, 24490, 24487, 24459, 24461, 24472, 24469, 24487, 24486, 24461, 24467, 24486, 24499, 24509, 24508, 24507, 24515, 24514, 24505, 24514, 24509, 24497, 24508, 24487, 24520, 24522, 24523, 24525, 24527, 24526, 24530, 24531, 24529, 24533, 24535, 24534, 24528, 24529, 24523, 24527, 24533, 24532, 24524, 24526, 24522, 24528, 24532, 24534, 24522, 24526, 24532, 24536, 24538, 24539, 24541, 24543, 24542, 24546, 24547, 24545, 24549, 24551, 24550, 24544, 24545, 24539, 24543, 24549, 24548, 24536, 24540, 24542, 24544, 24548, 24550, 24538, 24542, 24548, 24536, 24537, 24531, 24535, 24541, 24540, 24530, 24534, 24540, 24554, 24555, 24553, 24558, 24559, 24557, 24562, 24563, 24561, 24566, 24567, 24565, 24560, 24561, 24555, 24564, 24565, 24559,
24555, 24558, 24556, 24563, 24566, 24564, 24561, 24564, 24558, 24573, 24571, 24570, 24579, 24577, 24576, 24577, 24573, 24572, 24572, 24570, 24568, 24578, 24576, 24574, 24576, 24572, 24569, 24582, 24583, 24581, 24586, 24587, 24585, 24590, 24591, 24589, 24594, 24595, 24593, 24588, 24589, 24583, 24592, 24593, 24587, 24583, 24586, 24584, 24591, 24594, 24592, 24589, 24592, 24586, 24601, 24599, 24598, 24607, 24605, 24604, 24605, 24601, 24600, 24597, 24600, 24598, 24603, 24606, 24604, 24602, 24604, 24600, 24580, 24581, 24563, 24584, 24585, 24567, 24581, 24584, 24566, 24599, 24579, 24578, 24596, 24598, 24578, 24559, 24569, 24568, 24567, 24575, 24574, 24565, 24574, 24569, 24585, 24587, 24597, 24595, 24603, 24602, 24593, 24602, 24597, 24585, 24596, 24575, 24520, 24521, 24449, 24453, 24525, 24524, 24452, 24524, 24520, 24552, 24553, 24503, 24556, 24557, 24507, 24553, 24556, 24506, 24571, 24519, 24518, 24570, 24518, 24515, 24557, 24568, 24515, 24425, 24456, 24454, 24433, 24464, 24462, 24431, 24462, 24456, 24441, 24494, 24492, 24449, 24502, 24500, 24447, 24500, 24494, 24439, 24492, 24464, 24523, 24554, 24552, 24531, 24562, 24560, 24529, 24560, 24554, 24539, 24582, 24580, 24547, 24590, 24588, 24545, 24588, 24582, 24537, 24580, 24562, 24521, 24552, 24502,
24614, 24616, 24615, 24621, 24622, 24620, 24624, 24626, 24625, 24630, 24628, 24627, 24619, 24620, 24610, 24612, 24624, 24623, 24623, 24625, 24618, 24610, 24614, 24613, 24629, 24627, 24620, 24613, 24615, 24612, 24618, 24621, 24619, 24609, 24613, 24611, 24617, 24619, 24609, 24608, 24611, 24623, 24636, 24638, 24637, 24640, 24642, 24641, 24650, 24648, 24647, 24654, 24652, 24651, 24643, 24644, 24632, 24645, 24646, 24634, 24632, 24636, 24635, 24648, 24644, 24643, 24634, 24640, 24639, 24652, 24646, 24645, 24639, 24641, 24638, 24653, 24651, 24648, 24633, 24639, 24636, 24651, 24645, 24644, 24644, 24645, 24633, 24659, 24660, 24658, 24662, 24664, 24663, 24668, 24666, 24665, 24671, 24673, 24674, 24676, 24678, 24677, 24682, 24680, 24679, 24659, 24671, 24672, 24664, 24676, 24675, 24680, 24668, 24667, 24661, 24663, 24656, 24669, 24675, 24677, 24660, 24667, 24665, 24674, 24681, 24679, 24656, 24659, 24657, 24670, 24673, 24671, 24669, 24671, 24659, 24656, 24663, 24675, 24672, 24679, 24667, 24657, 24658, 24622, 24626, 24662, 24661, 24666, 24630, 24629, 24655, 24657, 24621, 24625, 24661, 24655, 24665, 24629, 24622, 24635, 24637, 24616, 24649, 24647, 24627, 24631, 24635, 24614, 24647, 24643, 24620, 24620, 24643, 24631, 24688, 24690, 24689, 24692, 24694, 24693,
24703, 24701, 24700, 24707, 24705, 24704, 24707, 24709, 24708, 24695, 24696, 24684, 24697, 24698, 24686, 24684, 24688, 24687, 24701, 24696, 24695, 24686, 24692, 24691, 24705, 24698, 24697, 24705, 24708, 24699, 24691, 24693, 24690, 24706, 24704, 24701, 24685, 24691, 24688, 24704, 24697, 24696, 24696, 24697, 24685, 24713, 24715, 24714, 24716, 24718, 24719, 24723, 24721, 24720, 24710, 24716, 24717, 24721, 24715, 24713, 24711, 24713, 24712, 24722, 24720, 24717, 24720, 24713, 24711, 24727, 24729, 24728, 24733, 24735, 24734, 24729, 24733, 24732, 24726, 24728, 24725, 24732, 24734, 24731, 24728, 24732, 24730, 24736, 24738, 24739, 24743, 24741, 24740, 24746, 24747, 24745, 24751, 24749, 24748, 24744, 24745, 24739, 24749, 24743, 24742, 24742, 24740, 24737, 24750, 24748, 24745, 24748, 24742, 24739, 24709, 24727, 24726, 24708, 24726, 24724, 24718, 24736, 24737, 24741, 24723, 24722, 24740, 24722, 24719, 24712, 24714, 24694, 24698, 24699, 24718, 24710, 24712, 24692, 24686, 24698, 24716, 24724, 24725, 24738, 24730, 24731, 24746, 24725, 24730, 24744, 24699, 24724, 24736, 24753, 24755, 24756, 24761, 24764, 24763, 24771, 24769, 24768, 24771, 24773, 24772, 24765, 24766, 24756, 24764, 24770, 24768, 24759, 24763, 24755, 24768, 24769, 24766, 24769, 24772, 24767,
24761, 24759, 24757, 24757, 24759, 24753, 24755, 24763, 24768, 24779, 24777, 24776, 24780, 24782, 24783, 24787, 24789, 24788, 24793, 24789, 24787, 24795, 24797, 24796, 24788, 24779, 24778, 24775, 24778, 24776, 24781, 24787, 24786, 24783, 24791, 24787, 24784, 24785, 24795, 24794, 24796, 24793, 24784, 24794, 24791, 24780, 24786, 24778, 24803, 24804, 24802, 24808, 24806, 24805, 24808, 24810, 24809, 24814, 24812, 24811, 24806, 24799, 24798, 24799, 24806, 24809, 24811, 24812, 24804, 24800, 24803, 24801, 24813, 24811, 24809, 24811, 24803, 24800, 24782, 24798, 24799, 24801, 24802, 24785, 24792, 24807, 24805, 24783, 24799, 24801, 24782, 24790, 24805, 24756, 24775, 24774, 24776, 24777, 24762, 24766, 24767, 24782, 24792, 24790, 24772, 24754, 24774, 24776, 24790, 24782, 24767, 24756, 24766, 24780, 24821, 24823, 24822, 24819, 24823, 24821, 24829, 24830, 24828, 24834, 24836, 24835, 24832, 24836, 24834, 24816, 24827, 24828, 24822, 24838, 24837, 24826, 24834, 24833, 24831, 24834, 24826, 24817, 24821, 24820, 24818, 24821, 24817, 24824, 24826, 24829, 24837, 24838, 24832, 24824, 24827, 24837, 24816, 24820, 24837, 24841, 24843, 24844, 24850, 24848, 24847, 24850, 24846, 24845, 24856, 24854, 24853, 24854, 24849, 24847, 24848, 24844, 24843, 24844, 24848, 24845,
24855, 24853, 24851, 24853, 24847, 24843, 24857, 24859, 24860, 24868, 24866, 24865, 24868, 24864, 24863, 24872, 24870, 24869, 24865, 24866, 24860, 24866, 24863, 24858, 24870, 24862, 24861, 24871, 24869, 24863, 24863, 24869, 24861, 24857, 24858, 24826, 24829, 24861, 24862, 24872, 24840, 24839, 24858, 24861, 24829, 24870, 24839, 24830, 24817, 24843, 24841, 24828, 24830, 24852, 24856, 24855, 24839, 24855, 24852, 24830, 24828, 24851, 24843, 24673, 24753, 24754, 24678, 24758, 24757, 24762, 24682, 24681, 24752, 24753, 24673, 24677, 24757, 24752, 24754, 24760, 24681, 24735, 24819, 24818, 24734, 24818, 24815, 24746, 24841, 24842, 24846, 24751, 24750, 24845, 24750, 24747, 24731, 24815, 24841, 24687, 24689, 24642, 24702, 24700, 24652, 24683, 24687, 24640, 24700, 24695, 24646, 24646, 24695, 24683, 24833, 24835, 24797, 24825, 24833, 24795, 24802, 24804, 24859, 24867, 24865, 24812, 24865, 24859, 24804, 24802, 24857, 24825, 24879, 24877, 24876, 24879, 24881, 24880, 24885, 24887, 24886, 24881, 24885, 24884, 24876, 24877, 24874, 24877, 24880, 24875, 24882, 24884, 24886, 24875, 24880, 24884, 24887, 24890, 24889, 24886, 24889, 24888, 24893, 24894, 24892, 24897, 24898, 24896, 24901, 24902, 24900, 24905, 24906, 24904, 24899, 24900, 24894, 24903, 24904, 24898,
24894, 24897, 24895, 24902, 24905, 24903, 24900, 24903, 24897, 24912, 24910, 24909, 24915, 24916, 24914, 24918, 24920, 24919, 24922, 24920, 24918, 24919, 24912, 24911, 24908, 24911, 24909, 24913, 24914, 24918, 24916, 24921, 24918, 24913, 24917, 24911, 24925, 24926, 24924, 24929, 24930, 24928, 24932, 24934, 24933, 24937, 24938, 24936, 24941, 24942, 24940, 24948, 24946, 24945, 24944, 24947, 24945, 24935, 24936, 24926, 24939, 24940, 24930, 24934, 24944, 24943, 24931, 24933, 24925, 24946, 24938, 24937, 24935, 24943, 24945, 24926, 24929, 24927, 24936, 24938, 24941, 24936, 24939, 24929, 24925, 24933, 24943, 24949, 24951, 24952, 24960, 24958, 24957, 24963, 24964, 24962, 24965, 24967, 24968, 24970, 24972, 24971, 24974, 24972, 24970, 24961, 24962, 24952, 24971, 24960, 24959, 24954, 24959, 24957, 24965, 24966, 24970, 24968, 24973, 24970, 24952, 24954, 24953, 24957, 24958, 24956, 24962, 24964, 24967, 24950, 24953, 24957, 24962, 24965, 24954, 24965, 24969, 24959, 24901, 24923, 24924, 24927, 24928, 24906, 24924, 24927, 24905, 24949, 24950, 24916, 24956, 24922, 24921, 24950, 24955, 24921, 24898, 24908, 24907, 24904, 24906, 24915, 24904, 24913, 24908, 24930, 24951, 24949, 24942, 24963, 24961, 24940, 24961, 24951, 24928, 24949, 24915, 24978, 24979, 24977,
24987, 24985, 24984, 24983, 24986, 24984, 24987, 24990, 24989, 24984, 24985, 24979, 24982, 24984, 24978, 24983, 24982, 24980, 24980, 24982, 24976, 24985, 24989, 24988, 24993, 24994, 24992, 24997, 24998, 24996, 25001, 25002, 25000, 25005, 25006, 25004, 25012, 25010, 25009, 25008, 25011, 25009, 24993, 24999, 25000, 25003, 25004, 24998, 25010, 25002, 25001, 24999, 25007, 25009, 24994, 24997, 24995, 25000, 25002, 25005, 25000, 25003, 24997, 25022, 25020, 25019, 25022, 25024, 25023, 25019, 25020, 25015, 25015, 25020, 25023, 25021, 25019, 25017, 25017, 25019, 25014, 25014, 25015, 25006, 25012, 25018, 25017, 25013, 25014, 25005, 25002, 25010, 25017, 24977, 24979, 24993, 25008, 25007, 24989, 25007, 24999, 24988, 24988, 24999, 24993, 24941, 24976, 24977, 24948, 24981, 24980, 24975, 24976, 24941, 24946, 24980, 24975, 24991, 24992, 24964, 24967, 24995, 24996, 24992, 24995, 24967, 24977, 24991, 24963, 24875, 24893, 24891, 24883, 24901, 24899, 24882, 24899, 24893, 24932, 24931, 24889, 24889, 24931, 24923, 24888, 24923, 24901, 24873, 24874, 24547, 24551, 24878, 24876, 24550, 24876, 24873, 24891, 24892, 24591, 24895, 24896, 24595, 24892, 24895, 24594, 24910, 24607, 24606, 24907, 24909, 24606, 24896, 24907, 24603, 24547, 24874, 24891, 24094, 24302, 24300,
24092, 24093, 24300, 24370, 24371, 24128, 24382, 24143, 24141, 24128, 24371, 24374, 24374, 24381, 24141, 24092, 24294, 24370, 24422, 24423, 24228, 24232, 24427, 24426, 24231, 24426, 24422, 24454, 24455, 24268, 24458, 24459, 24272, 24455, 24458, 24271, 24470, 24471, 24284, 24474, 24475, 24288, 24471, 24474, 24287, 24459, 24470, 24283, 24423, 24454, 24267, 25029, 25030, 25028, 25032, 25034, 25033, 25036, 25038, 25037, 25031, 25033, 25026, 25028, 25036, 25035, 25026, 25029, 25027, 25035, 25037, 25032, 25027, 25035, 25031, 25043, 25044, 25042, 25045, 25047, 25048, 25050, 25052, 25051, 25056, 25054, 25053, 25058, 25060, 25059, 25064, 25062, 25061, 25042, 25050, 25049, 25054, 25044, 25043, 25046, 25058, 25057, 25061, 25062, 25048, 25042, 25044, 25047, 25057, 25059, 25052, 25063, 25061, 25054, 25045, 25057, 25050, 25054, 25061, 25047, 25030, 25043, 25041, 25049, 25051, 25038, 25055, 25053, 25039, 25041, 25049, 25036, 25053, 25043, 25030, 25065, 25067, 25068, 25069, 25071, 25072, 25074, 25076, 25075, 25080, 25078, 25077, 25082, 25084, 25083, 25066, 25074, 25073, 25077, 25078, 25068, 25070, 25082, 25081, 25068, 25071, 25069, 25081, 25083, 25076, 25086, 25085, 25078, 25069, 25081, 25074, 25078, 25085, 25071, 25089, 25090, 25088, 25092, 25094, 25093,
25096, 25094, 25092, 25088, 25092, 25091, 25095, 25092, 25088, 25072, 25089, 25087, 25091, 25093, 25084, 25087, 25091, 25082, 25101, 25102, 25100, 25104, 25106, 25105, 25110, 25108, 25107, 25115, 25116, 25114, 25118, 25120, 25119, 25124, 25122, 25121, 25113, 25114, 25102, 25106, 25118, 25117, 25122, 25110, 25109, 25103, 25105, 25098, 25117, 25119, 25112, 25109, 25107, 25100, 25123, 25121, 25114, 25098, 25101, 25099, 25112, 25115, 25113, 25111, 25113, 25101, 25105, 25117, 25111, 25121, 25109, 25102, 25128, 25130, 25129, 25133, 25134, 25132, 25136, 25138, 25137, 25131, 25132, 25126, 25129, 25136, 25135, 25126, 25128, 25127, 25135, 25137, 25133, 25127, 25135, 25131, 25143, 25144, 25142, 25146, 25148, 25147, 25152, 25150, 25149, 25157, 25158, 25156, 25160, 25162, 25161, 25168, 25166, 25165, 25168, 25164, 25163, 25155, 25156, 25144, 25148, 25160, 25159, 25164, 25152, 25151, 25145, 25147, 25140, 25159, 25161, 25154, 25151, 25149, 25142, 25166, 25158, 25157, 25166, 25163, 25156, 25140, 25143, 25141, 25154, 25157, 25155, 25167, 25165, 25161, 25165, 25157, 25154, 25153, 25155, 25143, 25147, 25159, 25153, 25163, 25151, 25144, 25171, 25172, 25170, 25174, 25176, 25175, 25179, 25180, 25178, 25186, 25184, 25183, 25182, 25185, 25183, 25177, 25178, 25172,
25176, 25182, 25181, 25173, 25175, 25171, 25184, 25180, 25179, 25181, 25183, 25179, 25175, 25181, 25177, 25141, 25142, 25116, 25120, 25146, 25145, 25150, 25124, 25123, 25139, 25141, 25115, 25119, 25145, 25139, 25149, 25123, 25116, 25169, 25170, 25134, 25138, 25174, 25173, 25137, 25173, 25169, 25196, 25194, 25193, 25196, 25198, 25197, 25206, 25204, 25203, 25208, 25210, 25209, 25204, 25192, 25191, 25198, 25208, 25207, 25205, 25203, 25199, 25194, 25189, 25188, 25194, 25197, 25190, 25201, 25207, 25209, 25203, 25191, 25187, 25197, 25207, 25201, 25213, 25214, 25212, 25218, 25216, 25215, 25221, 25222, 25220, 25226, 25224, 25223, 25219, 25220, 25214, 25224, 25218, 25217, 25217, 25215, 25212, 25225, 25223, 25220, 25223, 25217, 25214, 25234, 25232, 25231, 25236, 25238, 25237, 25246, 25244, 25243, 25248, 25250, 25249, 25244, 25234, 25233, 25238, 25248, 25247, 25233, 25231, 25227, 25245, 25243, 25239, 25229, 25235, 25237, 25247, 25249, 25242, 25243, 25233, 25228, 25237, 25247, 25241, 25253, 25254, 25252, 25258, 25256, 25255, 25261, 25262, 25260, 25268, 25266, 25265, 25268, 25264, 25263, 25259, 25260, 25254, 25264, 25258, 25257, 25257, 25255, 25252, 25266, 25262, 25261, 25266, 25263, 25260, 25263, 25257, 25254, 25232, 25206, 25205, 25210, 25236, 25235,
25231, 25205, 25200, 25202, 25209, 25235, 25251, 25252, 25222, 25256, 25226, 25225, 25255, 25225, 25222, 25190, 25213, 25211, 25202, 25221, 25219, 25201, 25219, 25213, 25230, 25253, 25251, 25242, 25261, 25259, 25267, 25265, 25249, 25265, 25261, 25242, 25241, 25259, 25253, 25202, 25229, 25251, 25099, 25100, 25030, 25034, 25104, 25103, 25108, 25040, 25039, 25097, 25099, 25029, 25033, 25103, 25097, 25107, 25039, 25030, 25188, 25189, 25072, 25086, 25195, 25193, 25071, 25085, 25193, 25211, 25212, 25090, 25216, 25096, 25095, 25215, 25095, 25090, 25072, 25189, 25211, 25046, 25048, 25067, 25073, 25075, 25060, 25079, 25077, 25062, 25065, 25073, 25058, 25062, 25077, 25067, 25191, 25192, 25130, 25134, 25200, 25199, 25187, 25191, 25128, 25132, 25199, 25187, 25172, 25228, 25227, 25180, 25240, 25239, 25246, 25245, 25184, 25245, 25240, 25180, 25178, 25239, 25228, 25170, 25227, 25200, 25271, 25272, 25270, 25274, 25276, 25275, 25275, 25278, 25277, 25270, 25274, 25273, 25273, 25277, 25271, 25281, 25282, 25280, 25284, 25286, 25285, 25289, 25290, 25288, 25292, 25294, 25293, 25287, 25288, 25282, 25286, 25292, 25291, 25283, 25285, 25281, 25291, 25293, 25289, 25285, 25291, 25287, 25279, 25280, 25272, 25278, 25284, 25283, 25277, 25283, 25279, 25297, 25298, 25296,
25301, 25302, 25300, 25304, 25306, 25305, 25308, 25310, 25309, 25296, 25304, 25303, 25300, 25308, 25307, 25298, 25301, 25299, 25307, 25309, 25306, 25299, 25307, 25304, 25316, 25314, 25313, 25315, 25313, 25311, 25319, 25320, 25318, 25323, 25324, 25322, 25327, 25328, 25326, 25331, 25332, 25330, 25325, 25326, 25320, 25329, 25330, 25324, 25320, 25323, 25321, 25328, 25331, 25329, 25326, 25329, 25323, 25338, 25336, 25335, 25344, 25342, 25341, 25342, 25338, 25337, 25337, 25335, 25333, 25343, 25341, 25339, 25341, 25337, 25334, 25317, 25318, 25298, 25321, 25322, 25302, 25318, 25321, 25301, 25336, 25316, 25315, 25335, 25315, 25312, 25302, 25312, 25311, 25313, 25314, 25310, 25311, 25313, 25308, 25324, 25334, 25333, 25332, 25340, 25339, 25330, 25339, 25334, 25322, 25333, 25312, 25347, 25348, 25346, 25350, 25352, 25351, 25355, 25356, 25354, 25358, 25360, 25359, 25353, 25354, 25348, 25352, 25358, 25357, 25349, 25351, 25347, 25357, 25359, 25355, 25351, 25357, 25353, 25363, 25364, 25362, 25366, 25368, 25367, 25371, 25372, 25370, 25374, 25376, 25375, 25369, 25370, 25364, 25368, 25374, 25373, 25365, 25367, 25363, 25373, 25375, 25371, 25367, 25373, 25369, 25361, 25362, 25356, 25360, 25366, 25365, 25359, 25365, 25361, 25379, 25380, 25378, 25383, 25384, 25382,
25387, 25388, 25386, 25391, 25392, 25390, 25385, 25386, 25380, 25389, 25390, 25384, 25380, 25383, 25381, 25388, 25391, 25389, 25386, 25389, 25383, 25398, 25396, 25395, 25404, 25402, 25401, 25402, 25398, 25397, 25397, 25395, 25393, 25403, 25401, 25399, 25401, 25397, 25394, 25407, 25408, 25406, 25411, 25412, 25410, 25415, 25416, 25414, 25419, 25420, 25418, 25413, 25414, 25408, 25417, 25418, 25412, 25408, 25411, 25409, 25416, 25419, 25417, 25414, 25417, 25411, 25426, 25424, 25423, 25432, 25430, 25429, 25430, 25426, 25425, 25425, 25423, 25421, 25431, 25429, 25427, 25429, 25425, 25422, 25405, 25406, 25388, 25409, 25410, 25392, 25406, 25409, 25391, 25424, 25404, 25403, 25423, 25403, 25400, 25384, 25394, 25393, 25392, 25400, 25399, 25390, 25399, 25394, 25412, 25422, 25421, 25420, 25428, 25427, 25418, 25427, 25422, 25410, 25421, 25400, 25345, 25346, 25290, 25294, 25350, 25349, 25293, 25349, 25345, 25377, 25378, 25328, 25381, 25382, 25332, 25378, 25381, 25331, 25396, 25344, 25343, 25395, 25343, 25340, 25382, 25393, 25340, 25272, 25297, 25295, 25303, 25305, 25276, 25295, 25303, 25274, 25282, 25319, 25317, 25290, 25327, 25325, 25288, 25325, 25319, 25280, 25317, 25297, 25348, 25379, 25377, 25356, 25387, 25385, 25354, 25385, 25379, 25364, 25407, 25405,
25372, 25415, 25413, 25370, 25413, 25407, 25362, 25405, 25387, 25346, 25377, 25327, 25437, 25438, 25436, 25440, 25442, 25441, 25444, 25446, 25445, 25455, 25453, 25452, 25448, 25449, 25438, 25442, 25451, 25450, 25439, 25441, 25434, 25436, 25444, 25443, 25453, 25449, 25448, 25434, 25437, 25435, 25443, 25445, 25440, 25454, 25452, 25450, 25435, 25443, 25439, 25452, 25448, 25447, 25447, 25448, 25437, 25441, 25450, 25447, 25458, 25459, 25457, 25462, 25463, 25461, 25465, 25467, 25466, 25469, 25471, 25470, 25479, 25477, 25476, 25483, 25481, 25480, 25458, 25472, 25473, 25462, 25474, 25475, 25457, 25465, 25464, 25476, 25477, 25473, 25461, 25469, 25468, 25480, 25481, 25475, 25459, 25462, 25460, 25468, 25470, 25467, 25482, 25480, 25477, 25460, 25468, 25465, 25477, 25480, 25474, 25473, 25474, 25462, 25489, 25491, 25490, 25495, 25493, 25492, 25501, 25503, 25502, 25507, 25505, 25504, 25491, 25501, 25500, 25505, 25495, 25494, 25488, 25490, 25485, 25500, 25502, 25497, 25494, 25492, 25486, 25499, 25506, 25504, 25485, 25487, 25486, 25492, 25493, 25489, 25496, 25497, 25499, 25486, 25492, 25488, 25485, 25496, 25498, 25490, 25500, 25496, 25498, 25504, 25494, 25438, 25458, 25456, 25464, 25466, 25446, 25478, 25476, 25453, 25456, 25464, 25444, 25476, 25472, 25449,
25449, 25472, 25458, 25510, 25511, 25509, 25514, 25515, 25513, 25517, 25519, 25518, 25521, 25523, 25522, 25528, 25529, 25527, 25533, 25531, 25530, 25539, 25537, 25536, 25535, 25538, 25536, 25510, 25524, 25525, 25514, 25526, 25527, 25509, 25517, 25516, 25530, 25531, 25525, 25513, 25521, 25520, 25537, 25529, 25528, 25526, 25534, 25536, 25511, 25514, 25512, 25520, 25522, 25519, 25535, 25534, 25531, 25512, 25520, 25517, 25534, 25526, 25525, 25525, 25526, 25514, 25542, 25543, 25541, 25545, 25547, 25546, 25549, 25547, 25545, 25552, 25553, 25551, 25557, 25555, 25554, 25550, 25551, 25543, 25555, 25549, 25548, 25541, 25545, 25544, 25548, 25545, 25541, 25556, 25554, 25551, 25554, 25548, 25543, 25561, 25563, 25562, 25567, 25569, 25568, 25563, 25567, 25566, 25558, 25560, 25562, 25566, 25568, 25565, 25559, 25562, 25566, 25572, 25573, 25571, 25577, 25575, 25574, 25580, 25581, 25579, 25585, 25583, 25582, 25578, 25579, 25573, 25583, 25577, 25576, 25576, 25574, 25571, 25584, 25582, 25579, 25582, 25576, 25573, 25539, 25561, 25560, 25529, 25537, 25560, 25570, 25571, 25553, 25575, 25557, 25556, 25574, 25556, 25553, 25515, 25542, 25540, 25544, 25546, 25523, 25527, 25529, 25552, 25540, 25544, 25521, 25527, 25550, 25542, 25559, 25572, 25570, 25564, 25565, 25580,
25564, 25578, 25572, 25558, 25570, 25552, 25463, 25510, 25508, 25516, 25518, 25471, 25532, 25530, 25481, 25508, 25516, 25469, 25481, 25530, 25524, 25463, 25475, 25524, 25588, 25589, 25587, 25591, 25593, 25592, 25596, 25597, 25595, 25599, 25601, 25600, 25594, 25595, 25589, 25593, 25599, 25598, 25590, 25592, 25588, 25598, 25600, 25596, 25592, 25598, 25594, 25604, 25605, 25603, 25607, 25609, 25608, 25612, 25613, 25611, 25615, 25617, 25616, 25610, 25611, 25605, 25609, 25615, 25614, 25606, 25608, 25604, 25614, 25616, 25612, 25608, 25614, 25610, 25602, 25603, 25597, 25601, 25607, 25606, 25600, 25606, 25602, 25620, 25621, 25619, 25624, 25625, 25623, 25628, 25629, 25627, 25632, 25633, 25631, 25626, 25627, 25621, 25630, 25631, 25625, 25621, 25624, 25622, 25629, 25632, 25630, 25627, 25630, 25624, 25639, 25637, 25636, 25645, 25643, 25642, 25643, 25639, 25638, 25638, 25636, 25634, 25644, 25642, 25640, 25642, 25638, 25635, 25648, 25649, 25647, 25652, 25653, 25651, 25656, 25657, 25655, 25660, 25661, 25659, 25654, 25655, 25649, 25658, 25659, 25653, 25649, 25652, 25650, 25657, 25660, 25658, 25655, 25658, 25652, 25667, 25665, 25664, 25673, 25671, 25670, 25671, 25667, 25666, 25666, 25664, 25662, 25672, 25670, 25668, 25670, 25666, 25663, 25646, 25647, 25629,
25650, 25651, 25633, 25647, 25650, 25632, 25665, 25645, 25644, 25664, 25644, 25641, 25625, 25635, 25634, 25633, 25641, 25640, 25631, 25640, 25635, 25653, 25663, 25662, 25661, 25669, 25668, 25659, 25668, 25663, 25651, 25662, 25641, 25676, 25677, 25675, 25679, 25681, 25680, 25684, 25685, 25683, 25687, 25689, 25688, 25676, 25682, 25683, 25681, 25687, 25686, 25674, 25678, 25680, 25682, 25686, 25688, 25676, 25680, 25686, 25692, 25693, 25691, 25699, 25697, 25696, 25695, 25698, 25696, 25703, 25705, 25704, 25699, 25703, 25702, 25696, 25697, 25693, 25694, 25696, 25692, 25700, 25702, 25704, 25697, 25702, 25700, 25690, 25691, 25685, 25689, 25695, 25694, 25684, 25688, 25694, 25708, 25709, 25707, 25712, 25713, 25711, 25716, 25717, 25715, 25720, 25721, 25719, 25714, 25715, 25709, 25718, 25719, 25713, 25709, 25712, 25710, 25717, 25720, 25718, 25715, 25718, 25712, 25727, 25725, 25724, 25733, 25731, 25730, 25731, 25727, 25726, 25723, 25726, 25724, 25729, 25732, 25730, 25728, 25730, 25726, 25736, 25737, 25735, 25740, 25741, 25739, 25744, 25745, 25743, 25748, 25749, 25747, 25742, 25743, 25737, 25746, 25747, 25741, 25737, 25740, 25738, 25745, 25748, 25746, 25743, 25746, 25740, 25755, 25753, 25752, 25758, 25759, 25757, 25762, 25764, 25763, 25766, 25764, 25762,
25763, 25755, 25754, 25751, 25754, 25752, 25756, 25757, 25762, 25765, 25762, 25757, 25767, 25768, 25766, 25760, 25767, 25765, 25761, 25754, 25751, 25734, 25735, 25717, 25738, 25739, 25721, 25735, 25738, 25720, 25753, 25733, 25732, 25750, 25752, 25732, 25711, 25713, 25723, 25721, 25729, 25728, 25719, 25728, 25723, 25741, 25751, 25750, 25747, 25749, 25758, 25747, 25756, 25751, 25739, 25750, 25729, 25612, 25674, 25675, 25617, 25679, 25678, 25612, 25616, 25678, 25706, 25707, 25657, 25710, 25711, 25661, 25707, 25710, 25660, 25725, 25673, 25672, 25724, 25672, 25669, 25711, 25722, 25669, 25589, 25620, 25618, 25597, 25628, 25626, 25595, 25626, 25620, 25605, 25648, 25646, 25613, 25656, 25654, 25611, 25654, 25648, 25603, 25646, 25628, 25677, 25708, 25706, 25685, 25716, 25714, 25683, 25714, 25708, 25691, 25693, 25736, 25701, 25744, 25742, 25700, 25742, 25736, 25691, 25734, 25716, 25675, 25706, 25656, 25586, 25587, 25372, 25376, 25591, 25590, 25375, 25590, 25586, 25618, 25619, 25416, 25622, 25623, 25420, 25619, 25622, 25419, 25637, 25432, 25431, 25636, 25431, 25428, 25623, 25634, 25428, 25587, 25618, 25415, 25772, 25774, 25775, 25777, 25779, 25778, 25778, 25781, 25780, 25771, 25779, 25777, 25773, 25777, 25776, 25776, 25780, 25774, 25769, 25770, 25777,
25782, 25784, 25785, 25790, 25792, 25791, 25791, 25794, 25793, 25798, 25800, 25799, 25799, 25796, 25795, 25801, 25803, 25804, 25807, 25808, 25806, 25812, 25810, 25809, 25816, 25814, 25813, 25801, 25802, 25785, 25805, 25806, 25788, 25814, 25800, 25798, 25783, 25790, 25789, 25782, 25789, 25793, 25809, 25810, 25804, 25787, 25788, 25798, 25797, 25795, 25786, 25808, 25815, 25813, 25785, 25787, 25786, 25804, 25807, 25805, 25785, 25802, 25805, 25806, 25813, 25798, 25819, 25821, 25822, 25824, 25826, 25825, 25834, 25832, 25831, 25834, 25830, 25828, 25826, 25838, 25837, 25840, 25833, 25831, 25817, 25823, 25825, 25831, 25832, 25822, 25832, 25828, 25820, 25818, 25821, 25819, 25827, 25829, 25824, 25840, 25839, 25837, 25819, 25827, 25823, 25837, 25839, 25836, 25818, 25835, 25836, 25825, 25837, 25835, 25839, 25831, 25821, 25845, 25846, 25844, 25856, 25854, 25853, 25852, 25855, 25853, 25856, 25859, 25858, 25854, 25846, 25845, 25843, 25851, 25853, 25852, 25851, 25848, 25848, 25851, 25843, 25846, 25854, 25858, 25819, 25820, 25775, 25781, 25829, 25827, 25780, 25827, 25819, 25807, 25843, 25844, 25849, 25811, 25809, 25812, 25850, 25848, 25842, 25843, 25807, 25847, 25809, 25803, 25810, 25848, 25842, 25793, 25794, 25771, 25775, 25803, 25801, 25784, 25793, 25770,
25769, 25773, 25801, 25849, 25847, 25828, 25828, 25847, 25841, 25775, 25820, 25841, 25863, 25865, 25864, 25868, 25870, 25871, 25873, 25875, 25874, 25877, 25879, 25878, 25878, 25881, 25880, 25865, 25879, 25877, 25866, 25867, 25873, 25862, 25864, 25861, 25869, 25877, 25876, 25868, 25876, 25880, 25880, 25881, 25875, 25870, 25880, 25873, 25864, 25877, 25869, 25884, 25885, 25883, 25889, 25887, 25886, 25890, 25892, 25893, 25899, 25897, 25896, 25899, 25895, 25894, 25884, 25890, 25891, 25895, 25889, 25888, 25888, 25886, 25883, 25897, 25893, 25892, 25893, 25897, 25894, 25894, 25888, 25885, 25902, 25903, 25901, 25904, 25906, 25907, 25911, 25909, 25908, 25918, 25916, 25915, 25912, 25913, 25903, 25920, 25910, 25908, 25915, 25916, 25913, 25909, 25907, 25906, 25903, 25906, 25904, 25920, 25919, 25916, 25916, 25919, 25914, 25903, 25913, 25914, 25919, 25908, 25906, 25926, 25924, 25923, 25925, 25923, 25921, 25900, 25901, 25867, 25904, 25905, 25871, 25867, 25901, 25904, 25924, 25898, 25896, 25923, 25896, 25892, 25860, 25861, 25884, 25871, 25892, 25890, 25861, 25869, 25890, 25905, 25907, 25922, 25926, 25925, 25909, 25925, 25922, 25907, 25871, 25905, 25921, 25872, 25874, 25816, 25808, 25866, 25872, 25844, 25846, 25902, 25917, 25915, 25858, 25858, 25915, 25912,
25857, 25912, 25902, 25844, 25900, 25866, 25932, 25933, 25931, 25936, 25937, 25935, 25939, 25941, 25940, 25945, 25947, 25948, 25952, 25950, 25949, 25952, 25954, 25953, 25942, 25943, 25933, 25945, 25946, 25937, 25941, 25951, 25949, 25938, 25940, 25932, 25949, 25950, 25943, 25950, 25953, 25944, 25933, 25936, 25934, 25944, 25947, 25945, 25956, 25955, 25953, 25953, 25955, 25947, 25933, 25943, 25945, 25932, 25940, 25949, 25959, 25960, 25958, 25963, 25964, 25962, 25966, 25968, 25967, 25970, 25968, 25966, 25973, 25974, 25972, 25977, 25978, 25976, 25971, 25972, 25960, 25963, 25975, 25976, 25962, 25966, 25965, 25964, 25969, 25966, 25958, 25960, 25963, 25974, 25977, 25975, 25972, 25975, 25963, 25985, 25983, 25982, 25985, 25987, 25986, 25983, 25980, 25979, 25980, 25983, 25986, 25990, 25991, 25989, 25994, 25995, 25993, 25999, 25997, 25996, 26008, 26006, 26005, 26001, 26002, 25995, 25999, 26004, 26003, 25997, 25991, 25990, 26005, 26006, 26002, 25989, 25991, 25994, 26007, 26005, 26003, 26005, 26001, 26000, 26000, 26001, 25994, 25991, 25997, 26003, 25979, 25980, 25948, 25956, 25984, 25982, 25947, 25955, 25982, 25988, 25989, 25974, 25992, 25993, 25978, 25989, 25992, 25977, 25937, 25959, 25957, 25948, 25973, 25971, 25946, 25971, 25959, 25981, 25990, 25988,
25998, 25996, 25986, 25986, 25996, 25990, 25948, 25980, 25988, 25939, 25938, 25928, 25928, 25938, 25930, 25503, 25792, 25790, 25796, 25507, 25506, 25497, 25783, 25786, 25497, 25502, 25790, 25786, 25795, 25506, 25569, 25863, 25862, 25568, 25862, 25860, 25882, 25883, 25581, 25887, 25585, 25584, 25886, 25584, 25581, 25565, 25860, 25882, 25705, 25929, 25928, 25704, 25928, 25927, 25744, 25930, 25931, 25934, 25935, 25749, 25931, 25934, 25748, 25758, 25957, 25958, 25967, 25768, 25767, 25958, 25961, 25760, 25965, 25767, 25760, 25935, 25957, 25758, 25927, 25930, 25744, 23650, 23880, 23881, 23657, 23885, 23884, 23879, 23880, 23650, 23655, 23884, 23879, 23895, 23896, 23674, 23899, 23900, 23678, 23896, 23899, 23677, 23881, 23895, 23673, 26011, 26012, 26010, 26015, 26016, 26014, 26018, 26020, 26019, 26023, 26024, 26022, 26027, 26028, 26026, 26030, 26032, 26031, 26021, 26022, 26012, 26025, 26026, 26016, 26020, 26030, 26029, 26009, 26017, 26019, 26021, 26029, 26031, 26012, 26015, 26013, 26024, 26027, 26025, 26022, 26025, 26015, 26011, 26019, 26029, 26035, 26036, 26034, 26040, 26038, 26037, 26043, 26044, 26042, 26049, 26047, 26046, 26041, 26042, 26036, 26047, 26040, 26039, 26036, 26039, 26037, 26048, 26046, 26042, 26050, 26051, 26049, 26044, 26045, 26050,
26042, 26046, 26039, 26057, 26058, 26056, 26062, 26060, 26059, 26062, 26064, 26063, 26067, 26069, 26070, 26072, 26074, 26073, 26067, 26068, 26058, 26064, 26072, 26071, 26059, 26060, 26053, 26060, 26063, 26054, 26065, 26071, 26073, 26054, 26057, 26055, 26066, 26069, 26067, 26076, 26075, 26073, 26075, 26069, 26066, 26065, 26067, 26057, 26054, 26063, 26071, 26079, 26080, 26078, 26083, 26084, 26082, 26086, 26088, 26087, 26090, 26088, 26086, 26093, 26094, 26092, 26097, 26098, 26096, 26091, 26092, 26080, 26083, 26095, 26096, 26081, 26082, 26086, 26084, 26089, 26086, 26078, 26080, 26083, 26094, 26097, 26095, 26092, 26095, 26083, 26052, 26053, 26024, 26055, 26056, 26028, 26032, 26061, 26059, 26024, 26053, 26055, 26031, 26059, 26052, 26043, 26077, 26078, 26087, 26051, 26050, 26078, 26081, 26045, 26085, 26050, 26045, 26016, 26035, 26033, 26028, 26043, 26041, 26026, 26041, 26035, 26058, 26079, 26077, 26070, 26093, 26091, 26068, 26091, 26079, 26056, 26077, 26043, 26104, 26105, 26103, 26110, 26112, 26111, 26116, 26118, 26117, 26117, 26114, 26113, 26111, 26101, 26100, 26102, 26103, 26110, 26108, 26116, 26115, 26107, 26115, 26113, 26105, 26107, 26106, 26113, 26114, 26112, 26103, 26106, 26113, 26102, 26109, 26100, 26125, 26127, 26128, 26132, 26134, 26133,
26130, 26134, 26132, 26136, 26138, 26137, 26121, 26138, 26136, 26123, 26124, 26132, 26129, 26132, 26124, 26126, 26136, 26135, 26122, 26124, 26127, 26135, 26137, 26130, 26125, 26135, 26129, 26120, 26136, 26126, 26131, 26133, 26118, 26123, 26131, 26116, 26145, 26143, 26142, 26145, 26147, 26146, 26142, 26143, 26140, 26143, 26146, 26141, 26148, 26150, 26151, 26154, 26155, 26153, 26163, 26164, 26162, 26168, 26166, 26165, 26168, 26170, 26169, 26174, 26172, 26171, 26158, 26159, 26151, 26161, 26162, 26155, 26157, 26167, 26165, 26166, 26159, 26158, 26166, 26169, 26160, 26171, 26172, 26164, 26151, 26154, 26152, 26160, 26163, 26161, 26173, 26171, 26169, 26169, 26171, 26163, 26151, 26159, 26161, 26150, 26156, 26165, 26141, 26150, 26148, 26157, 26156, 26146, 26156, 26150, 26141, 26177, 26178, 26176, 26181, 26182, 26180, 26183, 26185, 26186, 26189, 26190, 26188, 26183, 26184, 26178, 26187, 26188, 26182, 26178, 26181, 26179, 26186, 26189, 26187, 26184, 26187, 26181, 26195, 26196, 26194, 26199, 26200, 26198, 26203, 26204, 26202, 26205, 26207, 26208, 26212, 26210, 26209, 26201, 26202, 26196, 26205, 26206, 26200, 26209, 26210, 26208, 26196, 26199, 26197, 26204, 26207, 26205, 26202, 26205, 26199, 26220, 26218, 26217, 26224, 26222, 26221, 26218, 26214, 26213,
26221, 26222, 26216, 26223, 26221, 26218, 26218, 26221, 26215, 26231, 26229, 26228, 26229, 26226, 26225, 26233, 26232, 26229, 26229, 26232, 26227, 26213, 26214, 26186, 26215, 26216, 26190, 26192, 26219, 26217, 26214, 26215, 26189, 26185, 26191, 26217, 26203, 26225, 26226, 26233, 26211, 26209, 26204, 26226, 26227, 26232, 26209, 26207, 26182, 26195, 26193, 26190, 26203, 26201, 26188, 26201, 26195, 26230, 26228, 26222, 26228, 26225, 26216, 26216, 26225, 26203, 26139, 26140, 26070, 26076, 26144, 26142, 26075, 26142, 26139, 26148, 26149, 26094, 26152, 26153, 26098, 26149, 26152, 26097, 26070, 26140, 26148, 26175, 26176, 26105, 26179, 26180, 26108, 26105, 26176, 26179, 26123, 26193, 26194, 26197, 26198, 26128, 26194, 26197, 26127, 26180, 26193, 26123, 26100, 26101, 26090, 26096, 26098, 26104, 26099, 26100, 26089, 26096, 26102, 26099, 26155, 26177, 26175, 26164, 26185, 26183, 26192, 26191, 26172, 26191, 26185, 26164, 26162, 26183, 26177, 26153, 26175, 26104, 26235, 26237, 26238, 26242, 26244, 26243, 26249, 26250, 26248, 26252, 26254, 26253, 26247, 26248, 26238, 26240, 26252, 26251, 26251, 26253, 26246, 26235, 26241, 26243, 26243, 26244, 26240, 26245, 26246, 26249, 26237, 26243, 26239, 26234, 26245, 26247, 26234, 26239, 26251, 26257, 26258, 26256,
26259, 26261, 26262, 26266, 26264, 26263, 26269, 26270, 26268, 26277, 26275, 26274, 26279, 26276, 26274, 26267, 26268, 26258, 26261, 26271, 26272, 26277, 26266, 26265, 26265, 26263, 26260, 26275, 26272, 26271, 26273, 26278, 26274, 26258, 26261, 26259, 26268, 26270, 26273, 26268, 26271, 26261, 26272, 26275, 26265, 26280, 26282, 26283, 26286, 26287, 26285, 26289, 26291, 26290, 26290, 26293, 26292, 26296, 26297, 26295, 26300, 26301, 26299, 26294, 26295, 26283, 26298, 26299, 26287, 26281, 26289, 26288, 26280, 26288, 26292, 26283, 26286, 26284, 26297, 26300, 26298, 26295, 26298, 26286, 26304, 26305, 26303, 26313, 26311, 26310, 26314, 26316, 26317, 26323, 26321, 26320, 26323, 26319, 26318, 26304, 26314, 26315, 26319, 26309, 26308, 26321, 26317, 26316, 26321, 26318, 26315, 26307, 26312, 26310, 26303, 26305, 26307, 26313, 26312, 26308, 26312, 26307, 26305, 26318, 26308, 26305, 26284, 26285, 26250, 26254, 26291, 26289, 26246, 26281, 26284, 26253, 26289, 26281, 26302, 26303, 26270, 26311, 26279, 26278, 26270, 26303, 26306, 26310, 26278, 26273, 26238, 26257, 26255, 26250, 26269, 26267, 26248, 26267, 26257, 26287, 26304, 26302, 26301, 26316, 26314, 26299, 26314, 26304, 26285, 26302, 26269, 26326, 26327, 26325, 26328, 26330, 26331, 26337, 26335, 26334,
26337, 26333, 26332, 26338, 26340, 26341, 26348, 26346, 26345, 26348, 26344, 26343, 26338, 26339, 26327, 26350, 26336, 26334, 26346, 26341, 26340, 26346, 26343, 26339, 26334, 26335, 26331, 26331, 26335, 26332, 26327, 26330, 26328, 26350, 26349, 26343, 26343, 26349, 26342, 26327, 26339, 26342, 26349, 26334, 26330, 26353, 26352, 26332, 26352, 26351, 26329, 26324, 26325, 26297, 26328, 26329, 26301, 26325, 26328, 26300, 26353, 26322, 26320, 26351, 26352, 26320, 26301, 26329, 26351, 26292, 26293, 26121, 26128, 26296, 26294, 26119, 26282, 26292, 26119, 26126, 26294, 26200, 26326, 26324, 26206, 26208, 26340, 26347, 26345, 26210, 26345, 26340, 26208, 26206, 26338, 26326, 26198, 26324, 26296, 26356, 26357, 26355, 26360, 26361, 26359, 26363, 26365, 26364, 26367, 26369, 26368, 26372, 26373, 26371, 26376, 26377, 26375, 26370, 26371, 26357, 26374, 26375, 26361, 26355, 26363, 26362, 26359, 26367, 26366, 26357, 26360, 26358, 26366, 26368, 26365, 26373, 26376, 26374, 26358, 26366, 26363, 26371, 26374, 26360, 26380, 26381, 26379, 26384, 26385, 26383, 26387, 26389, 26388, 26391, 26393, 26392, 26396, 26397, 26395, 26400, 26401, 26399, 26394, 26395, 26381, 26398, 26399, 26385, 26379, 26387, 26386, 26383, 26391, 26390, 26381, 26384, 26382, 26390, 26392, 26389,
26397, 26400, 26398, 26382, 26390, 26387, 26395, 26398, 26384, 26404, 26405, 26403, 26408, 26409, 26407, 26413, 26411, 26410, 26417, 26415, 26414, 26411, 26405, 26404, 26415, 26409, 26408, 26405, 26408, 26406, 26416, 26414, 26411, 26414, 26408, 26405, 26420, 26421, 26419, 26424, 26425, 26423, 26430, 26431, 26429, 26434, 26435, 26433, 26437, 26439, 26438, 26428, 26429, 26421, 26432, 26433, 26425, 26427, 26437, 26436, 26436, 26438, 26430, 26421, 26424, 26422, 26431, 26434, 26432, 26429, 26432, 26424, 26426, 26436, 26428, 26402, 26403, 26373, 26406, 26407, 26377, 26403, 26406, 26376, 26418, 26419, 26397, 26422, 26423, 26401, 26419, 26422, 26400, 26361, 26380, 26378, 26386, 26388, 26369, 26377, 26396, 26394, 26378, 26386, 26367, 26375, 26394, 26380, 26409, 26420, 26418, 26427, 26426, 26415, 26426, 26420, 26409, 26407, 26418, 26396, 26442, 26443, 26441, 26446, 26447, 26445, 26449, 26451, 26450, 26453, 26455, 26454, 26458, 26459, 26457, 26462, 26463, 26461, 26456, 26457, 26443, 26460, 26461, 26447, 26441, 26449, 26448, 26445, 26453, 26452, 26443, 26446, 26444, 26452, 26454, 26451, 26459, 26462, 26460, 26444, 26452, 26449, 26457, 26460, 26446, 26466, 26467, 26465, 26470, 26471, 26469, 26473, 26475, 26474, 26477, 26479, 26478, 26482, 26483, 26481,
26486, 26487, 26485, 26480, 26481, 26467, 26484, 26485, 26471, 26465, 26473, 26472, 26469, 26477, 26476, 26467, 26470, 26468, 26476, 26478, 26475, 26483, 26486, 26484, 26468, 26476, 26473, 26481, 26484, 26470, 26490, 26491, 26489, 26492, 26494, 26495, 26499, 26497, 26496, 26502, 26503, 26501, 26509, 26507, 26506, 26500, 26501, 26491, 26507, 26498, 26496, 26497, 26495, 26494, 26508, 26506, 26504, 26489, 26491, 26494, 26503, 26505, 26504, 26491, 26501, 26504, 26506, 26496, 26494, 26512, 26513, 26511, 26516, 26517, 26515, 26521, 26519, 26518, 26525, 26523, 26522, 26519, 26513, 26512, 26523, 26517, 26516, 26513, 26516, 26514, 26524, 26522, 26519, 26522, 26516, 26513, 26488, 26489, 26459, 26492, 26493, 26463, 26489, 26492, 26462, 26510, 26511, 26483, 26514, 26515, 26487, 26511, 26514, 26486, 26447, 26466, 26464, 26472, 26474, 26455, 26463, 26482, 26480, 26464, 26472, 26453, 26461, 26480, 26466, 26495, 26512, 26510, 26520, 26518, 26497, 26518, 26512, 26495, 26493, 26510, 26482, 26528, 26529, 26527, 26532, 26533, 26531, 26535, 26537, 26536, 26540, 26541, 26539, 26544, 26545, 26543, 26547, 26549, 26548, 26538, 26539, 26529, 26542, 26543, 26533, 26537, 26547, 26546, 26534, 26536, 26528, 26546, 26548, 26540, 26529, 26532, 26530, 26541, 26544, 26542,
26539, 26542, 26532, 26536, 26546, 26538, 26552, 26553, 26551, 26556, 26557, 26555, 26559, 26561, 26560, 26564, 26565, 26563, 26568, 26569, 26567, 26571, 26573, 26572, 26562, 26563, 26553, 26566, 26567, 26557, 26561, 26571, 26570, 26558, 26560, 26552, 26570, 26572, 26564, 26553, 26556, 26554, 26565, 26568, 26566, 26563, 26566, 26556, 26560, 26570, 26562, 26550, 26551, 26541, 26554, 26555, 26545, 26549, 26559, 26558, 26551, 26554, 26544, 26548, 26558, 26550, 26576, 26577, 26575, 26583, 26581, 26580, 26586, 26587, 26585, 26593, 26591, 26590, 26584, 26585, 26577, 26591, 26583, 26582, 26582, 26580, 26578, 26592, 26590, 26588, 26577, 26579, 26578, 26587, 26589, 26588, 26585, 26588, 26579, 26590, 26582, 26579, 26596, 26597, 26595, 26603, 26601, 26600, 26606, 26607, 26605, 26613, 26611, 26610, 26604, 26605, 26597, 26611, 26603, 26602, 26602, 26600, 26598, 26612, 26610, 26608, 26597, 26599, 26598, 26607, 26609, 26608, 26605, 26608, 26599, 26610, 26602, 26599, 26594, 26595, 26587, 26601, 26593, 26592, 26595, 26598, 26589, 26600, 26592, 26589, 26526, 26527, 26431, 26530, 26531, 26435, 26439, 26535, 26534, 26527, 26530, 26434, 26438, 26534, 26526, 26574, 26575, 26503, 26581, 26509, 26508, 26575, 26578, 26505, 26580, 26508, 26505, 26385, 26442, 26440,
26448, 26450, 26393, 26401, 26458, 26456, 26440, 26448, 26391, 26399, 26456, 26442, 26425, 26490, 26488, 26435, 26502, 26500, 26433, 26500, 26490, 26423, 26488, 26458, 26533, 26576, 26574, 26545, 26586, 26584, 26543, 26584, 26576, 26557, 26596, 26594, 26569, 26606, 26604, 26567, 26604, 26596, 26555, 26594, 26586, 26531, 26574, 26502, 26616, 26617, 26615, 26620, 26621, 26619, 26623, 26625, 26624, 26627, 26629, 26628, 26632, 26633, 26631, 26636, 26637, 26635, 26630, 26631, 26617, 26634, 26635, 26621, 26615, 26623, 26622, 26619, 26627, 26626, 26617, 26620, 26618, 26626, 26628, 26625, 26633, 26636, 26634, 26618, 26626, 26623, 26631, 26634, 26620, 26640, 26641, 26639, 26644, 26645, 26643, 26647, 26649, 26648, 26651, 26653, 26652, 26655, 26653, 26651, 26658, 26659, 26657, 26662, 26663, 26661, 26667, 26665, 26664, 26656, 26657, 26641, 26660, 26661, 26645, 26665, 26655, 26654, 26639, 26647, 26646, 26643, 26651, 26650, 26654, 26651, 26643, 26666, 26664, 26661, 26641, 26644, 26642, 26650, 26652, 26649, 26659, 26662, 26660, 26642, 26650, 26647, 26657, 26660, 26644, 26664, 26654, 26645, 26670, 26671, 26669, 26672, 26674, 26675, 26679, 26677, 26676, 26684, 26685, 26683, 26687, 26689, 26688, 26682, 26683, 26675, 26681, 26687, 26686, 26677, 26671, 26670,
26686, 26688, 26684, 26669, 26671, 26674, 26681, 26680, 26677, 26680, 26674, 26671, 26680, 26686, 26682, 26692, 26693, 26691, 26696, 26697, 26695, 26701, 26699, 26698, 26704, 26705, 26703, 26708, 26709, 26707, 26713, 26711, 26710, 26702, 26703, 26693, 26706, 26707, 26697, 26711, 26701, 26700, 26700, 26698, 26695, 26712, 26710, 26707, 26693, 26696, 26694, 26705, 26708, 26706, 26703, 26706, 26696, 26710, 26700, 26697, 26668, 26669, 26633, 26672, 26673, 26637, 26669, 26672, 26636, 26690, 26691, 26659, 26694, 26695, 26663, 26699, 26667, 26666, 26691, 26694, 26662, 26698, 26666, 26663, 26621, 26640, 26638, 26646, 26648, 26629, 26637, 26658, 26656, 26638, 26646, 26627, 26635, 26656, 26640, 26675, 26692, 26690, 26685, 26704, 26702, 26683, 26702, 26692, 26673, 26690, 26658, 26716, 26717, 26715, 26719, 26721, 26720, 26724, 26725, 26723, 26727, 26729, 26728, 26722, 26723, 26717, 26721, 26727, 26726, 26718, 26720, 26716, 26726, 26728, 26724, 26720, 26726, 26722, 26732, 26733, 26731, 26736, 26737, 26735, 26741, 26739, 26738, 26744, 26745, 26743, 26748, 26749, 26747, 26753, 26751, 26750, 26742, 26743, 26733, 26746, 26747, 26737, 26751, 26741, 26740, 26740, 26738, 26735, 26752, 26750, 26747, 26733, 26736, 26734, 26745, 26748, 26746, 26743, 26746, 26736,
26750, 26740, 26737, 26756, 26757, 26755, 26759, 26761, 26760, 26764, 26765, 26763, 26767, 26769, 26768, 26762, 26763, 26757, 26761, 26767, 26766, 26758, 26760, 26756, 26766, 26768, 26764, 26760, 26766, 26762, 26772, 26773, 26771, 26776, 26777, 26775, 26781, 26779, 26778, 26784, 26785, 26783, 26788, 26789, 26787, 26793, 26791, 26790, 26782, 26783, 26773, 26786, 26787, 26777, 26791, 26781, 26780, 26780, 26778, 26775, 26792, 26790, 26787, 26773, 26776, 26774, 26785, 26788, 26786, 26783, 26786, 26776, 26790, 26780, 26777, 26754, 26755, 26725, 26729, 26759, 26758, 26728, 26758, 26754, 26770, 26771, 26745, 26774, 26775, 26749, 26779, 26753, 26752, 26771, 26774, 26748, 26778, 26752, 26749, 26717, 26732, 26730, 26725, 26744, 26742, 26723, 26742, 26732, 26757, 26772, 26770, 26765, 26784, 26782, 26763, 26782, 26772, 26755, 26770, 26744, 26714, 26715, 26685, 26689, 26719, 26718, 26688, 26718, 26714, 26730, 26731, 26705, 26734, 26735, 26709, 26739, 26713, 26712, 26731, 26734, 26708, 26738, 26712, 26709, 26715, 26730, 26704, 26796, 26797, 26795, 26800, 26801, 26799, 26803, 26805, 26804, 26808, 26809, 26807, 26812, 26813, 26811, 26815, 26817, 26816, 26806, 26807, 26797, 26810, 26811, 26801, 26805, 26815, 26814, 26802, 26804, 26796, 26814, 26816, 26808,
26797, 26800, 26798, 26809, 26812, 26810, 26807, 26810, 26800, 26804, 26814, 26806, 26820, 26821, 26819, 26824, 26825, 26823, 26827, 26829, 26828, 26832, 26833, 26831, 26836, 26837, 26835, 26839, 26841, 26840, 26830, 26831, 26821, 26834, 26835, 26825, 26829, 26839, 26838, 26826, 26828, 26820, 26838, 26840, 26832, 26821, 26824, 26822, 26833, 26836, 26834, 26831, 26834, 26824, 26828, 26838, 26830, 26818, 26819, 26809, 26822, 26823, 26813, 26817, 26827, 26826, 26819, 26822, 26812, 26816, 26826, 26818, 26844, 26845, 26843, 26851, 26849, 26848, 26854, 26855, 26853, 26861, 26859, 26858, 26852, 26853, 26845, 26859, 26851, 26850, 26850, 26848, 26846, 26860, 26858, 26856, 26845, 26847, 26846, 26855, 26857, 26856, 26853, 26856, 26847, 26858, 26850, 26847, 26864, 26865, 26863, 26871, 26869, 26868, 26874, 26875, 26873, 26881, 26879, 26878, 26872, 26873, 26865, 26879, 26871, 26870, 26870, 26868, 26866, 26880, 26878, 26876, 26865, 26867, 26866, 26875, 26877, 26876, 26873, 26876, 26867, 26878, 26870, 26867, 26862, 26863, 26855, 26869, 26861, 26860, 26863, 26866, 26857, 26868, 26860, 26857, 26884, 26885, 26883, 26888, 26889, 26887, 26891, 26893, 26892, 26901, 26899, 26898, 26905, 26903, 26902, 26894, 26895, 26885, 26896, 26897, 26889, 26893, 26900, 26898,
26890, 26892, 26884, 26899, 26895, 26894, 26903, 26897, 26896, 26885, 26888, 26886, 26904, 26902, 26899, 26902, 26896, 26895, 26895, 26896, 26888, 26892, 26898, 26894, 26908, 26909, 26907, 26915, 26913, 26912, 26922, 26920, 26919, 26916, 26917, 26909, 26924, 26915, 26914, 26920, 26917, 26916, 26914, 26912, 26910, 26909, 26911, 26910, 26924, 26923, 26920, 26923, 26918, 26917, 26917, 26918, 26911, 26923, 26914, 26911, 26882, 26883, 26833, 26886, 26887, 26837, 26841, 26891, 26890, 26883, 26886, 26836, 26840, 26890, 26882, 26906, 26907, 26875, 26913, 26881, 26880, 26907, 26910, 26877, 26912, 26880, 26877, 26801, 26844, 26842, 26813, 26854, 26852, 26811, 26852, 26844, 26825, 26864, 26862, 26837, 26874, 26872, 26835, 26872, 26864, 26823, 26862, 26854, 26889, 26908, 26906, 26921, 26919, 26903, 26919, 26916, 26897, 26897, 26916, 26908, 26887, 26906, 26874, 26927, 26928, 26926, 26930, 26932, 26931, 26935, 26936, 26934, 26938, 26940, 26939, 26933, 26934, 26928, 26932, 26938, 26937, 26929, 26931, 26927, 26937, 26939, 26935, 26931, 26937, 26933, 26943, 26944, 26942, 26947, 26948, 26946, 26952, 26950, 26949, 26955, 26956, 26954, 26959, 26960, 26958, 26964, 26962, 26961, 26953, 26954, 26944, 26957, 26958, 26948, 26962, 26952, 26951, 26951, 26949, 26946,
26963, 26961, 26958, 26944, 26947, 26945, 26956, 26959, 26957, 26954, 26957, 26947, 26961, 26951, 26948, 26967, 26968, 26966, 26970, 26972, 26971, 26975, 26976, 26974, 26978, 26980, 26979, 26973, 26974, 26968, 26972, 26978, 26977, 26969, 26971, 26967, 26977, 26979, 26975, 26971, 26977, 26973, 26983, 26984, 26982, 26987, 26988, 26986, 26992, 26990, 26989, 26995, 26996, 26994, 26999, 27000, 26998, 27004, 27002, 27001, 26993, 26994, 26984, 26997, 26998, 26988, 27002, 26992, 26991, 26991, 26989, 26986, 27003, 27001, 26998, 26984, 26987, 26985, 26996, 26999, 26997, 26994, 26997, 26987, 27001, 26991, 26988, 26965, 26966, 26936, 26940, 26970, 26969, 26939, 26969, 26965, 26981, 26982, 26956, 26985, 26986, 26960, 26990, 26964, 26963, 26982, 26985, 26959, 26989, 26963, 26960, 26928, 26943, 26941, 26936, 26955, 26953, 26934, 26953, 26943, 26968, 26983, 26981, 26976, 26995, 26993, 26974, 26993, 26983, 26966, 26981, 26955, 27007, 27008, 27006, 27010, 27012, 27011, 27015, 27016, 27014, 27018, 27020, 27019, 27013, 27014, 27008, 27012, 27018, 27017, 27009, 27011, 27007, 27017, 27019, 27015, 27011, 27017, 27013, 27023, 27024, 27022, 27027, 27028, 27026, 27032, 27030, 27029, 27035, 27036, 27034, 27039, 27040, 27038, 27044, 27042, 27041, 27033, 27034, 27024,
27037, 27038, 27028, 27042, 27032, 27031, 27031, 27029, 27026, 27043, 27041, 27038, 27024, 27027, 27025, 27036, 27039, 27037, 27034, 27037, 27027, 27041, 27031, 27028, 27050, 27048, 27047, 27048, 27046, 27045, 27058, 27056, 27055, 27062, 27060, 27059, 27056, 27052, 27051, 27060, 27054, 27053, 27061, 27059, 27056, 27059, 27053, 27052, 27045, 27046, 27016, 27020, 27049, 27047, 27019, 27047, 27045, 27051, 27052, 27036, 27053, 27054, 27040, 27062, 27044, 27043, 27052, 27053, 27039, 27060, 27043, 27040, 27008, 27023, 27021, 27016, 27035, 27033, 27014, 27033, 27023, 27057, 27055, 27048, 27055, 27051, 27046, 27046, 27051, 27035, 27005, 27006, 26976, 26980, 27010, 27009, 26979, 27009, 27005, 27021, 27022, 26996, 27025, 27026, 27000, 27030, 27004, 27003, 27022, 27025, 26999, 27029, 27003, 27000, 27006, 27021, 26995, 26794, 26795, 26565, 26798, 26799, 26569, 26573, 26803, 26802, 26795, 26798, 26568, 26572, 26802, 26794, 26842, 26843, 26607, 26849, 26613, 26612, 26843, 26846, 26609, 26848, 26612, 26609, 26799, 26842, 26606, 26925, 26926, 26765, 26769, 26930, 26929, 26768, 26929, 26925, 26941, 26942, 26785, 26945, 26946, 26789, 26950, 26793, 26792, 26942, 26945, 26788, 26949, 26792, 26789, 26926, 26941, 26784, 26471, 26616, 26614, 26622, 26624, 26479,
26487, 26632, 26630, 26614, 26622, 26477, 26485, 26630, 26616, 26517, 26670, 26668, 26678, 26676, 26523, 26676, 26670, 26517, 26515, 26668, 26632, 27065, 27066, 27064, 27069, 27070, 27068, 27072, 27074, 27073, 27076, 27078, 27077, 27081, 27082, 27080, 27085, 27086, 27084, 27079, 27080, 27066, 27083, 27084, 27070, 27064, 27072, 27071, 27068, 27076, 27075, 27066, 27069, 27067, 27075, 27077, 27074, 27082, 27085, 27083, 27067, 27075, 27072, 27080, 27083, 27069, 27089, 27090, 27088, 27093, 27094, 27092, 27096, 27098, 27097, 27100, 27102, 27101, 27105, 27106, 27104, 27109, 27110, 27108, 27103, 27104, 27090, 27107, 27108, 27094, 27088, 27096, 27095, 27092, 27100, 27099, 27090, 27093, 27091, 27099, 27101, 27098, 27106, 27109, 27107, 27091, 27099, 27096, 27104, 27107, 27093, 27113, 27114, 27112, 27117, 27118, 27116, 27122, 27120, 27119, 27126, 27124, 27123, 27120, 27114, 27113, 27124, 27118, 27117, 27114, 27117, 27115, 27125, 27123, 27120, 27123, 27117, 27114, 27129, 27130, 27128, 27133, 27134, 27132, 27138, 27136, 27135, 27142, 27140, 27139, 27136, 27130, 27129, 27140, 27134, 27133, 27130, 27133, 27131, 27141, 27139, 27136, 27139, 27133, 27130, 27111, 27112, 27082, 27115, 27116, 27086, 27112, 27115, 27085, 27127, 27128, 27106, 27131, 27132, 27110,
27128, 27131, 27109, 27070, 27089, 27087, 27095, 27097, 27078, 27086, 27105, 27103, 27087, 27095, 27076, 27084, 27103, 27089, 27118, 27129, 27127, 27137, 27135, 27124, 27135, 27129, 27118, 27116, 27127, 27105, 27145, 27146, 27144, 27149, 27150, 27148, 27152, 27154, 27153, 27156, 27158, 27157, 27161, 27162, 27160, 27165, 27166, 27164, 27159, 27160, 27146, 27163, 27164, 27150, 27144, 27152, 27151, 27148, 27156, 27155, 27146, 27149, 27147, 27155, 27157, 27154, 27162, 27165, 27163, 27147, 27155, 27152, 27160, 27163, 27149, 27169, 27170, 27168, 27173, 27174, 27172, 27176, 27178, 27177, 27180, 27182, 27181, 27185, 27186, 27184, 27189, 27190, 27188, 27183, 27184, 27170, 27187, 27188, 27174, 27168, 27176, 27175, 27172, 27180, 27179, 27170, 27173, 27171, 27179, 27181, 27178, 27186, 27189, 27187, 27171, 27179, 27176, 27184, 27187, 27173, 27193, 27194, 27192, 27197, 27198, 27196, 27202, 27200, 27199, 27206, 27204, 27203, 27200, 27194, 27193, 27204, 27198, 27197, 27194, 27197, 27195, 27205, 27203, 27200, 27203, 27197, 27194, 27209, 27210, 27208, 27213, 27214, 27212, 27218, 27216, 27215, 27222, 27220, 27219, 27216, 27210, 27209, 27220, 27214, 27213, 27210, 27213, 27211, 27221, 27219, 27216, 27219, 27213, 27210, 27191, 27192, 27162, 27195, 27196, 27166,
27192, 27195, 27165, 27207, 27208, 27186, 27211, 27212, 27190, 27208, 27211, 27189, 27150, 27169, 27167, 27175, 27177, 27158, 27166, 27185, 27183, 27167, 27175, 27156, 27164, 27183, 27169, 27198, 27209, 27207, 27217, 27215, 27204, 27215, 27209, 27198, 27196, 27207, 27185, 27094, 27145, 27143, 27151, 27153, 27102, 27110, 27161, 27159, 27143, 27151, 27100, 27108, 27159, 27145, 27134, 27193, 27191, 27201, 27199, 27140, 27199, 27193, 27134, 27132, 27191, 27161, 27225, 27226, 27224, 27229, 27230, 27228, 27232, 27234, 27233, 27236, 27238, 27237, 27241, 27242, 27240, 27245, 27246, 27244, 27239, 27240, 27226, 27243, 27244, 27230, 27224, 27232, 27231, 27228, 27236, 27235, 27226, 27229, 27227, 27235, 27237, 27234, 27242, 27245, 27243, 27227, 27235, 27232, 27240, 27243, 27229, 27249, 27250, 27248, 27253, 27254, 27252, 27256, 27258, 27257, 27260, 27262, 27261, 27264, 27262, 27260, 27267, 27268, 27266, 27271, 27272, 27270, 27276, 27274, 27273, 27265, 27266, 27250, 27269, 27270, 27254, 27274, 27264, 27263, 27248, 27256, 27255, 27252, 27260, 27259, 27263, 27260, 27252, 27275, 27273, 27270, 27250, 27253, 27251, 27259, 27261, 27258, 27268, 27271, 27269, 27251, 27259, 27256, 27266, 27269, 27253, 27273, 27263, 27254, 27279, 27280, 27278, 27283, 27284, 27282,
27288, 27286, 27285, 27292, 27290, 27289, 27286, 27280, 27279, 27290, 27284, 27283, 27280, 27283, 27281, 27291, 27289, 27286, 27289, 27283, 27280, 27295, 27296, 27294, 27299, 27300, 27298, 27304, 27302, 27301, 27310, 27308, 27307, 27310, 27306, 27305, 27302, 27296, 27295, 27308, 27300, 27299, 27308, 27305, 27298, 27296, 27299, 27297, 27309, 27307, 27302, 27307, 27299, 27296, 27277, 27278, 27242, 27281, 27282, 27246, 27278, 27281, 27245, 27293, 27294, 27268, 27297, 27298, 27272, 27306, 27276, 27275, 27294, 27297, 27271, 27305, 27275, 27272, 27230, 27249, 27247, 27255, 27257, 27238, 27246, 27267, 27265, 27247, 27255, 27236, 27244, 27265, 27249, 27284, 27295, 27293, 27303, 27301, 27290, 27301, 27295, 27284, 27282, 27293, 27267, 27174, 27225, 27223, 27231, 27233, 27182, 27190, 27241, 27239, 27223, 27231, 27180, 27188, 27239, 27225, 27214, 27279, 27277, 27287, 27285, 27220, 27285, 27279, 27214, 27212, 27277, 27241, 27316, 27318, 27317, 27320, 27322, 27321, 27325, 27326, 27324, 27329, 27330, 27328, 27311, 27323, 27324, 27313, 27327, 27328, 27312, 27316, 27315, 27314, 27320, 27319, 27319, 27321, 27318, 27326, 27329, 27327, 27313, 27319, 27316, 27312, 27324, 27327, 27336, 27338, 27337, 27340, 27342, 27341, 27345, 27346, 27344, 27349, 27350, 27348,
27343, 27344, 27332, 27347, 27348, 27334, 27332, 27336, 27335, 27334, 27340, 27339, 27339, 27341, 27338, 27346, 27349, 27347, 27333, 27339, 27336, 27344, 27347, 27333, 27353, 27354, 27352, 27357, 27358, 27356, 27359, 27361, 27362, 27369, 27367, 27366, 27373, 27371, 27370, 27375, 27372, 27370, 27359, 27360, 27354, 27357, 27363, 27364, 27367, 27362, 27361, 27370, 27371, 27364, 27374, 27370, 27363, 27354, 27357, 27355, 27360, 27362, 27365, 27375, 27374, 27367, 27374, 27365, 27362, 27360, 27363, 27357, 27378, 27379, 27377, 27382, 27383, 27381, 27391, 27389, 27388, 27395, 27393, 27392, 27384, 27385, 27379, 27386, 27387, 27383, 27389, 27385, 27384, 27393, 27387, 27386, 27379, 27382, 27380, 27394, 27392, 27389, 27392, 27386, 27385, 27385, 27386, 27382, 27351, 27352, 27326, 27355, 27356, 27330, 27352, 27355, 27329, 27376, 27377, 27346, 27380, 27381, 27350, 27377, 27380, 27349, 27335, 27337, 27322, 27330, 27345, 27343, 27331, 27335, 27320, 27314, 27328, 27343, 27358, 27378, 27376, 27390, 27388, 27371, 27388, 27384, 27364, 27358, 27364, 27384, 27356, 27376, 27345, 27401, 27403, 27402, 27405, 27407, 27406, 27410, 27411, 27409, 27414, 27415, 27413, 27408, 27409, 27397, 27412, 27413, 27399, 27397, 27401, 27400, 27399, 27405, 27404, 27404, 27406, 27403,
27411, 27414, 27412, 27398, 27404, 27401, 27409, 27412, 27398, 27421, 27423, 27422, 27425, 27427, 27426, 27430, 27431, 27429, 27434, 27435, 27433, 27428, 27429, 27417, 27432, 27433, 27419, 27417, 27421, 27420, 27419, 27425, 27424, 27424, 27426, 27423, 27431, 27434, 27432, 27418, 27424, 27421, 27429, 27432, 27418, 27438, 27439, 27437, 27442, 27443, 27441, 27451, 27449, 27448, 27455, 27453, 27452, 27444, 27445, 27439, 27446, 27447, 27443, 27449, 27445, 27444, 27453, 27447, 27446, 27439, 27442, 27440, 27454, 27452, 27449, 27452, 27446, 27445, 27445, 27446, 27442, 27458, 27459, 27457, 27462, 27463, 27461, 27471, 27469, 27468, 27475, 27473, 27472, 27464, 27465, 27459, 27466, 27467, 27463, 27469, 27465, 27464, 27473, 27467, 27466, 27459, 27462, 27460, 27474, 27472, 27469, 27472, 27466, 27465, 27465, 27466, 27462, 27436, 27437, 27411, 27440, 27441, 27415, 27437, 27440, 27414, 27456, 27457, 27431, 27460, 27461, 27435, 27457, 27460, 27434, 27420, 27422, 27407, 27415, 27430, 27428, 27416, 27420, 27405, 27413, 27428, 27416, 27443, 27458, 27456, 27470, 27468, 27453, 27468, 27464, 27447, 27447, 27464, 27458, 27441, 27456, 27430, 27478, 27368, 27366, 27477, 27366, 27361, 27400, 27402, 27342, 27350, 27410, 27408, 27396, 27400, 27340, 27348, 27408, 27396,
27383, 27438, 27436, 27450, 27448, 27393, 27448, 27444, 27387, 27383, 27387, 27444, 27381, 27436, 27410, 27484, 27486, 27485, 27488, 27490, 27489, 27493, 27494, 27492, 27497, 27498, 27496, 27491, 27492, 27480, 27495, 27496, 27482, 27480, 27484, 27483, 27482, 27488, 27487, 27487, 27489, 27486, 27494, 27497, 27495, 27481, 27487, 27484, 27492, 27495, 27481, 27504, 27506, 27505, 27508, 27510, 27509, 27513, 27514, 27512, 27517, 27518, 27516, 27522, 27520, 27519, 27511, 27512, 27500, 27515, 27516, 27502, 27520, 27510, 27508, 27500, 27504, 27503, 27502, 27508, 27507, 27521, 27519, 27516, 27507, 27509, 27506, 27514, 27517, 27515, 27501, 27507, 27504, 27512, 27515, 27501, 27519, 27508, 27502, 27525, 27526, 27524, 27529, 27530, 27528, 27538, 27536, 27535, 27542, 27540, 27539, 27531, 27532, 27526, 27533, 27534, 27530, 27536, 27532, 27531, 27540, 27534, 27533, 27526, 27529, 27527, 27541, 27539, 27536, 27539, 27533, 27532, 27532, 27533, 27529, 27545, 27546, 27544, 27549, 27550, 27548, 27554, 27552, 27551, 27562, 27560, 27559, 27566, 27564, 27563, 27555, 27556, 27546, 27557, 27558, 27550, 27566, 27554, 27553, 27560, 27556, 27555, 27553, 27551, 27548, 27564, 27558, 27557, 27546, 27549, 27547, 27565, 27563, 27560, 27563, 27557, 27556, 27556, 27557, 27549,
27564, 27553, 27550, 27523, 27524, 27494, 27527, 27528, 27498, 27524, 27527, 27497, 27543, 27544, 27514, 27547, 27548, 27518, 27552, 27522, 27521, 27544, 27547, 27517, 27551, 27521, 27518, 27503, 27505, 27490, 27498, 27513, 27511, 27499, 27503, 27488, 27496, 27511, 27499, 27530, 27545, 27543, 27561, 27559, 27540, 27559, 27555, 27534, 27534, 27555, 27545, 27528, 27543, 27513, 27569, 27570, 27568, 27574, 27576, 27575, 27578, 27576, 27574, 27580, 27582, 27581, 27567, 27568, 27574, 27570, 27577, 27574, 27571, 27572, 27580, 27579, 27581, 27578, 27571, 27579, 27577, 27588, 27590, 27589, 27592, 27594, 27593, 27583, 27584, 27588, 27585, 27586, 27592, 27591, 27593, 27590, 27584, 27585, 27591, 27587, 27589, 27582, 27572, 27583, 27587, 27600, 27602, 27601, 27604, 27606, 27605, 27595, 27596, 27600, 27598, 27604, 27603, 27603, 27605, 27602, 27597, 27603, 27600, 27612, 27614, 27613, 27616, 27618, 27617, 27608, 27612, 27611, 27610, 27616, 27615, 27615, 27617, 27614, 27609, 27615, 27612, 27611, 27613, 27606, 27607, 27611, 27604, 27599, 27601, 27594, 27586, 27595, 27599, 27624, 27626, 27625, 27628, 27630, 27629, 27620, 27624, 27623, 27622, 27628, 27627, 27627, 27629, 27626, 27621, 27627, 27624, 27636, 27638, 27637, 27640, 27642, 27641, 27632, 27636, 27635,
27634, 27640, 27639, 27639, 27641, 27638, 27633, 27639, 27636, 27635, 27637, 27630, 27631, 27635, 27628, 27483, 27485, 27427, 27435, 27493, 27491, 27479, 27483, 27425, 27433, 27491, 27479, 27463, 27525, 27523, 27537, 27535, 27473, 27535, 27531, 27467, 27467, 27531, 27525, 27461, 27523, 27493, 27623, 27625, 27618, 27619, 27623, 27616, 27645, 27646, 27644, 27649, 27650, 27648, 27653, 27654, 27652, 27657, 27658, 27656, 27651, 27652, 27646, 27655, 27656, 27650, 27646, 27649, 27647, 27654, 27657, 27655, 27652, 27655, 27649, 27661, 27662, 27660, 27665, 27666, 27664, 27669, 27670, 27668, 27673, 27674, 27672, 27667, 27668, 27662, 27671, 27672, 27666, 27662, 27665, 27663, 27670, 27673, 27671, 27668, 27671, 27665, 27683, 27681, 27680, 27687, 27685, 27684, 27687, 27689, 27688, 27680, 27681, 27676, 27685, 27678, 27677, 27678, 27685, 27688, 27686, 27684, 27681, 27681, 27684, 27677, 27692, 27693, 27691, 27694, 27696, 27697, 27701, 27699, 27698, 27705, 27703, 27702, 27698, 27699, 27693, 27702, 27703, 27697, 27693, 27696, 27694, 27704, 27702, 27699, 27699, 27702, 27696, 27675, 27676, 27654, 27677, 27678, 27658, 27676, 27677, 27657, 27690, 27691, 27670, 27694, 27695, 27674, 27691, 27694, 27673, 27650, 27661, 27659, 27658, 27669, 27667, 27656, 27667, 27661,
27679, 27692, 27690, 27700, 27698, 27688, 27698, 27692, 27679, 27658, 27678, 27690, 27708, 27709, 27707, 27712, 27713, 27711, 27716, 27717, 27715, 27720, 27721, 27719, 27714, 27715, 27709, 27718, 27719, 27713, 27709, 27712, 27710, 27717, 27720, 27718, 27715, 27718, 27712, 27724, 27725, 27723, 27728, 27729, 27727, 27732, 27733, 27731, 27736, 27737, 27735, 27730, 27731, 27725, 27734, 27735, 27729, 27725, 27728, 27726, 27733, 27736, 27734, 27731, 27734, 27728, 27740, 27741, 27739, 27744, 27745, 27743, 27749, 27747, 27746, 27753, 27751, 27750, 27747, 27741, 27740, 27751, 27745, 27744, 27741, 27744, 27742, 27752, 27750, 27747, 27750, 27744, 27741, 27756, 27757, 27755, 27760, 27761, 27759, 27765, 27763, 27762, 27769, 27767, 27766, 27763, 27757, 27756, 27767, 27761, 27760, 27757, 27760, 27758, 27768, 27766, 27763, 27766, 27760, 27757, 27738, 27739, 27717, 27742, 27743, 27721, 27739, 27742, 27720, 27754, 27755, 27733, 27758, 27759, 27737, 27755, 27758, 27736, 27713, 27724, 27722, 27721, 27732, 27730, 27719, 27730, 27724, 27745, 27756, 27754, 27764, 27762, 27751, 27762, 27756, 27745, 27743, 27754, 27732, 27775, 27777, 27776, 27779, 27781, 27780, 27784, 27785, 27783, 27788, 27789, 27787, 27782, 27783, 27771, 27786, 27787, 27773, 27771, 27775, 27774,
27773, 27779, 27778, 27778, 27780, 27777, 27785, 27788, 27786, 27772, 27778, 27775, 27783, 27786, 27772, 27795, 27797, 27796, 27799, 27801, 27800, 27804, 27805, 27803, 27808, 27809, 27807, 27802, 27803, 27791, 27806, 27807, 27793, 27791, 27795, 27794, 27793, 27799, 27798, 27798, 27800, 27797, 27805, 27808, 27806, 27792, 27798, 27795, 27803, 27806, 27792, 27794, 27796, 27781, 27789, 27804, 27802, 27790, 27794, 27779, 27787, 27802, 27790, 27815, 27817, 27816, 27819, 27821, 27820, 27824, 27825, 27823, 27828, 27829, 27827, 27822, 27823, 27811, 27826, 27827, 27813, 27811, 27815, 27814, 27813, 27819, 27818, 27818, 27820, 27817, 27825, 27828, 27826, 27812, 27818, 27815, 27823, 27826, 27812, 27835, 27837, 27836, 27839, 27841, 27840, 27844, 27845, 27843, 27848, 27849, 27847, 27842, 27843, 27831, 27846, 27847, 27833, 27831, 27835, 27834, 27833, 27839, 27838, 27838, 27840, 27837, 27845, 27848, 27846, 27832, 27838, 27835, 27843, 27846, 27832, 27834, 27836, 27821, 27829, 27844, 27842, 27830, 27834, 27819, 27827, 27842, 27830, 27666, 27708, 27706, 27674, 27716, 27714, 27672, 27714, 27708, 27697, 27740, 27738, 27748, 27746, 27703, 27746, 27740, 27697, 27695, 27738, 27716, 27814, 27816, 27801, 27809, 27824, 27822, 27810, 27814, 27799, 27807, 27822, 27810,
27852, 27853, 27851, 27856, 27857, 27855, 27860, 27861, 27859, 27864, 27865, 27863, 27858, 27859, 27853, 27862, 27863, 27857, 27853, 27856, 27854, 27861, 27864, 27862, 27859, 27862, 27856, 27868, 27869, 27867, 27872, 27873, 27871, 27877, 27875, 27874, 27880, 27881, 27879, 27884, 27885, 27883, 27889, 27887, 27886, 27878, 27879, 27869, 27882, 27883, 27873, 27887, 27877, 27876, 27876, 27874, 27871, 27888, 27886, 27883, 27869, 27872, 27870, 27881, 27884, 27882, 27879, 27882, 27872, 27886, 27876, 27873, 27892, 27893, 27891, 27896, 27897, 27895, 27901, 27899, 27898, 27905, 27903, 27902, 27899, 27893, 27892, 27903, 27897, 27896, 27893, 27896, 27894, 27904, 27902, 27899, 27902, 27896, 27893, 27908, 27909, 27907, 27912, 27913, 27911, 27917, 27915, 27914, 27923, 27921, 27920, 27923, 27919, 27918, 27915, 27909, 27908, 27921, 27913, 27912, 27921, 27918, 27911, 27909, 27912, 27910, 27922, 27920, 27915, 27920, 27912, 27909, 27890, 27891, 27861, 27894, 27895, 27865, 27891, 27894, 27864, 27906, 27907, 27881, 27910, 27911, 27885, 27919, 27889, 27888, 27907, 27910, 27884, 27918, 27888, 27885, 27857, 27868, 27866, 27865, 27880, 27878, 27863, 27878, 27868, 27897, 27908, 27906, 27916, 27914, 27903, 27914, 27908, 27897, 27895, 27906, 27880, 27929, 27931, 27930,
27933, 27935, 27934, 27938, 27939, 27937, 27942, 27943, 27941, 27936, 27937, 27925, 27940, 27941, 27927, 27925, 27929, 27928, 27927, 27933, 27932, 27932, 27934, 27931, 27939, 27942, 27940, 27926, 27932, 27929, 27937, 27940, 27926, 27949, 27951, 27950, 27953, 27955, 27954, 27958, 27959, 27957, 27962, 27963, 27961, 27967, 27965, 27964, 27956, 27957, 27945, 27960, 27961, 27947, 27965, 27955, 27953, 27945, 27949, 27948, 27947, 27953, 27952, 27966, 27964, 27961, 27952, 27954, 27951, 27959, 27962, 27960, 27946, 27952, 27949, 27957, 27960, 27946, 27964, 27953, 27947, 27948, 27950, 27935, 27943, 27958, 27956, 27944, 27948, 27933, 27941, 27956, 27944, 27970, 27971, 27969, 27974, 27975, 27973, 27983, 27981, 27980, 27987, 27985, 27984, 27976, 27977, 27971, 27978, 27979, 27975, 27981, 27977, 27976, 27985, 27979, 27978, 27971, 27974, 27972, 27986, 27984, 27981, 27984, 27978, 27977, 27977, 27978, 27974, 27990, 27991, 27989, 27994, 27995, 27993, 28003, 28001, 28000, 28007, 28005, 28004, 27996, 27997, 27991, 27998, 27999, 27995, 28001, 27997, 27996, 28005, 27999, 27998, 27991, 27994, 27992, 28006, 28004, 28001, 28004, 27998, 27997, 27997, 27998, 27994, 27975, 27990, 27988, 28002, 28000, 27985, 28000, 27996, 27979, 27979, 27996, 27990, 28010, 28011, 28009,
28014, 28015, 28013, 28023, 28021, 28020, 28027, 28025, 28024, 28016, 28017, 28011, 28018, 28019, 28015, 28021, 28017, 28016, 28025, 28019, 28018, 28011, 28014, 28012, 28026, 28024, 28021, 28024, 28018, 28017, 28017, 28018, 28014, 28030, 28031, 28029, 28034, 28035, 28033, 28043, 28041, 28040, 28047, 28045, 28044, 28036, 28037, 28031, 28038, 28039, 28035, 28041, 28037, 28036, 28045, 28039, 28038, 28031, 28034, 28032, 28046, 28044, 28041, 28044, 28038, 28037, 28037, 28038, 28034, 28015, 28030, 28028, 28042, 28040, 28025, 28040, 28036, 28019, 28019, 28036, 28030, 28053, 28055, 28054, 28057, 28059, 28058, 28049, 28053, 28052, 28051, 28057, 28056, 28056, 28058, 28055, 28050, 28056, 28053, 28065, 28067, 28066, 28069, 28071, 28070, 28061, 28065, 28064, 28063, 28069, 28068, 28068, 28070, 28067, 28062, 28068, 28065, 28074, 28075, 28073, 28078, 28079, 28077, 28082, 28083, 28081, 28086, 28087, 28085, 28080, 28081, 28075, 28084, 28085, 28079, 28075, 28078, 28076, 28083, 28086, 28084, 28081, 28084, 28078, 28090, 28091, 28089, 28094, 28095, 28093, 28098, 28099, 28097, 28102, 28103, 28101, 28096, 28097, 28091, 28100, 28101, 28095, 28091, 28094, 28092, 28099, 28102, 28100, 28097, 28100, 28094, 28072, 28073, 28049, 28076, 28077, 28051, 28073, 28076, 28050,
28088, 28089, 28061, 28092, 28093, 28063, 28089, 28092, 28062, 28064, 28066, 28059, 28060, 28064, 28057, 28079, 28090, 28088, 28087, 28098, 28096, 28085, 28096, 28090, 28077, 28088, 28060, 28109, 28111, 28110, 28113, 28115, 28114, 28105, 28109, 28108, 28107, 28113, 28112, 28112, 28114, 28111, 28106, 28112, 28109, 28121, 28123, 28122, 28125, 28127, 28126, 28117, 28121, 28120, 28119, 28125, 28124, 28124, 28126, 28123, 28118, 28124, 28121, 28130, 28131, 28129, 28134, 28135, 28133, 28138, 28139, 28137, 28142, 28143, 28141, 28136, 28137, 28131, 28140, 28141, 28135, 28131, 28134, 28132, 28139, 28142, 28140, 28137, 28140, 28134, 28146, 28147, 28145, 28150, 28151, 28149, 28154, 28155, 28153, 28158, 28159, 28157, 28152, 28153, 28147, 28156, 28157, 28151, 28147, 28150, 28148, 28155, 28158, 28156, 28153, 28156, 28150, 28128, 28129, 28105, 28132, 28133, 28107, 28129, 28132, 28106, 28144, 28145, 28117, 28148, 28149, 28119, 28145, 28148, 28118, 28120, 28122, 28115, 28116, 28120, 28113, 28135, 28146, 28144, 28143, 28154, 28152, 28141, 28152, 28146, 28133, 28144, 28116, 27995, 28010, 28008, 28022, 28020, 28005, 28020, 28016, 27999, 27999, 28016, 28010, 28108, 28110, 28071, 28104, 28108, 28069, 28095, 28130, 28128, 28103, 28138, 28136, 28101, 28136, 28130,
28093, 28128, 28104, 28162, 28163, 28161, 28166, 28167, 28165, 28175, 28173, 28172, 28179, 28177, 28176, 28168, 28169, 28163, 28170, 28171, 28167, 28173, 28169, 28168, 28177, 28171, 28170, 28163, 28166, 28164, 28178, 28176, 28173, 28176, 28170, 28169, 28169, 28170, 28166, 28182, 28183, 28181, 28186, 28187, 28185, 28191, 28189, 28188, 28199, 28197, 28196, 28203, 28201, 28200, 28192, 28193, 28183, 28194, 28195, 28187, 28203, 28191, 28190, 28197, 28193, 28192, 28190, 28188, 28185, 28201, 28195, 28194, 28183, 28186, 28184, 28202, 28200, 28197, 28200, 28194, 28193, 28193, 28194, 28186, 28201, 28190, 28187, 28167, 28182, 28180, 28198, 28196, 28177, 28196, 28192, 28171, 28171, 28192, 28182, 28209, 28211, 28210, 28213, 28215, 28214, 28205, 28209, 28208, 28207, 28213, 28212, 28212, 28214, 28211, 28206, 28212, 28209, 28221, 28223, 28222, 28225, 28227, 28226, 28217, 28221, 28220, 28219, 28225, 28224, 28224, 28226, 28223, 28218, 28224, 28221, 28230, 28231, 28229, 28234, 28235, 28233, 28238, 28239, 28237, 28242, 28243, 28241, 28236, 28237, 28231, 28240, 28241, 28235, 28231, 28234, 28232, 28239, 28242, 28240, 28237, 28240, 28234, 28246, 28247, 28245, 28250, 28251, 28249, 28255, 28253, 28252, 28258, 28259, 28257, 28262, 28263, 28261, 28267, 28265, 28264,
28256, 28257, 28247, 28260, 28261, 28251, 28265, 28255, 28254, 28254, 28252, 28249, 28266, 28264, 28261, 28247, 28250, 28248, 28259, 28262, 28260, 28257, 28260, 28250, 28264, 28254, 28251, 28228, 28229, 28205, 28232, 28233, 28207, 28229, 28232, 28206, 28244, 28245, 28217, 28248, 28249, 28219, 28253, 28227, 28225, 28245, 28248, 28218, 28252, 28225, 28219, 28220, 28222, 28215, 28216, 28220, 28213, 28235, 28246, 28244, 28243, 28258, 28256, 28241, 28256, 28246, 28233, 28244, 28216, 27968, 27969, 27785, 27972, 27973, 27789, 27969, 27972, 27788, 27988, 27989, 27805, 27992, 27993, 27809, 27989, 27992, 27808, 27973, 27988, 27804, 28008, 28009, 27825, 28012, 28013, 27829, 28009, 28012, 27828, 28028, 28029, 27845, 28032, 28033, 27849, 28029, 28032, 27848, 28013, 28028, 27844, 27993, 28008, 27824, 28160, 28161, 27939, 28164, 28165, 27943, 28161, 28164, 27942, 28180, 28181, 27959, 28184, 28185, 27963, 28189, 27967, 27966, 28181, 28184, 27962, 28188, 27966, 27963, 28165, 28180, 27958, 27729, 27852, 27850, 27737, 27860, 27858, 27735, 27858, 27852, 27761, 27892, 27890, 27900, 27898, 27767, 27898, 27892, 27761, 27759, 27890, 27860, 27928, 27930, 27841, 27849, 27938, 27936, 27924, 27928, 27839, 27847, 27936, 27924, 28035, 28162, 28160, 28174, 28172, 28045,
28172, 28168, 28039, 28039, 28168, 28162, 28208, 28210, 28127, 28204, 28208, 28125, 28151, 28230, 28228, 28159, 28238, 28236, 28157, 28236, 28230, 28149, 28228, 28204, 28033, 28160, 27938, 28270, 28271, 28269, 28274, 28275, 28273, 28279, 28277, 28276, 28283, 28281, 28280, 28277, 28271, 28270, 28281, 28275, 28274, 28271, 28274, 28272, 28282, 28280, 28277, 28280, 28274, 28271, 28286, 28287, 28285, 28290, 28291, 28289, 28295, 28293, 28292, 28299, 28297, 28296, 28293, 28287, 28286, 28297, 28291, 28290, 28287, 28290, 28288, 28298, 28296, 28293, 28296, 28290, 28287, 28275, 28286, 28284, 28294, 28292, 28281, 28292, 28286, 28275, 28302, 28303, 28301, 28306, 28307, 28305, 28311, 28309, 28308, 28315, 28313, 28312, 28309, 28303, 28302, 28313, 28307, 28306, 28303, 28306, 28304, 28314, 28312, 28309, 28312, 28306, 28303, 28318, 28319, 28317, 28322, 28323, 28321, 28327, 28325, 28324, 28331, 28329, 28328, 28325, 28319, 28318, 28329, 28323, 28322, 28319, 28322, 28320, 28330, 28328, 28325, 28328, 28322, 28319, 28307, 28318, 28316, 28326, 28324, 28313, 28324, 28318, 28307, 28291, 28302, 28300, 28310, 28308, 28297, 28308, 28302, 28291, 28334, 28335, 28333, 28336, 28338, 28339, 28343, 28341, 28340, 28348, 28349, 28347, 28351, 28353, 28352, 28346, 28347, 28339,
28345, 28351, 28350, 28341, 28335, 28334, 28350, 28352, 28348, 28333, 28335, 28338, 28345, 28344, 28341, 28344, 28338, 28335, 28344, 28350, 28346, 28356, 28357, 28355, 28360, 28361, 28359, 28365, 28363, 28362, 28368, 28369, 28367, 28372, 28373, 28371, 28377, 28375, 28374, 28366, 28367, 28357, 28370, 28371, 28361, 28375, 28365, 28364, 28364, 28362, 28359, 28376, 28374, 28371, 28357, 28360, 28358, 28369, 28372, 28370, 28367, 28370, 28360, 28374, 28364, 28361, 28380, 28381, 28379, 28383, 28385, 28384, 28388, 28389, 28387, 28391, 28393, 28392, 28386, 28387, 28381, 28385, 28391, 28390, 28382, 28384, 28380, 28390, 28392, 28388, 28384, 28390, 28386, 28396, 28397, 28395, 28400, 28401, 28399, 28405, 28403, 28402, 28408, 28409, 28407, 28412, 28413, 28411, 28417, 28415, 28414, 28406, 28407, 28397, 28410, 28411, 28401, 28415, 28405, 28404, 28404, 28402, 28399, 28416, 28414, 28411, 28397, 28400, 28398, 28409, 28412, 28410, 28407, 28410, 28400, 28414, 28404, 28401, 28378, 28379, 28349, 28353, 28383, 28382, 28352, 28382, 28378, 28394, 28395, 28369, 28398, 28399, 28373, 28403, 28377, 28376, 28395, 28398, 28372, 28402, 28376, 28373, 28339, 28356, 28354, 28349, 28368, 28366, 28347, 28366, 28356, 28381, 28396, 28394, 28389, 28408, 28406, 28387, 28406, 28396,
28379, 28394, 28368, 28420, 28421, 28419, 28423, 28425, 28424, 28426, 28428, 28429, 28431, 28433, 28432, 28426, 28427, 28421, 28425, 28431, 28430, 28422, 28424, 28420, 28430, 28432, 28428, 28424, 28430, 28426, 28436, 28437, 28435, 28440, 28441, 28439, 28445, 28443, 28442, 28448, 28449, 28447, 28452, 28453, 28451, 28457, 28455, 28454, 28446, 28447, 28437, 28450, 28451, 28441, 28455, 28445, 28444, 28444, 28442, 28439, 28456, 28454, 28451, 28437, 28440, 28438, 28449, 28452, 28450, 28447, 28450, 28440, 28454, 28444, 28441, 28460, 28461, 28459, 28463, 28465, 28464, 28467, 28469, 28470, 28474, 28476, 28475, 28460, 28467, 28468, 28465, 28474, 28473, 28462, 28464, 28460, 28467, 28473, 28475, 28475, 28476, 28472, 28469, 28475, 28471, 28464, 28473, 28467, 28479, 28480, 28478, 28483, 28484, 28482, 28488, 28486, 28485, 28491, 28492, 28490, 28495, 28496, 28494, 28500, 28498, 28497, 28489, 28490, 28480, 28493, 28494, 28484, 28498, 28488, 28487, 28487, 28485, 28482, 28499, 28497, 28494, 28480, 28483, 28481, 28492, 28495, 28493, 28490, 28493, 28483, 28497, 28487, 28484, 28458, 28459, 28429, 28433, 28463, 28462, 28432, 28462, 28458, 28477, 28478, 28449, 28481, 28482, 28453, 28486, 28457, 28456, 28478, 28481, 28452, 28485, 28456, 28453, 28421, 28436, 28434,
28429, 28448, 28446, 28427, 28446, 28436, 28461, 28479, 28477, 28470, 28491, 28489, 28468, 28489, 28479, 28459, 28477, 28448, 28418, 28419, 28389, 28393, 28423, 28422, 28392, 28422, 28418, 28434, 28435, 28409, 28438, 28439, 28413, 28443, 28417, 28416, 28435, 28438, 28412, 28442, 28416, 28413, 28419, 28434, 28408, 28504, 28506, 28505, 28501, 28502, 28504, 28510, 28512, 28511, 28515, 28517, 28518, 28520, 28522, 28521, 28524, 28526, 28525, 28525, 28528, 28527, 28512, 28526, 28524, 28514, 28520, 28519, 28509, 28511, 28508, 28515, 28516, 28524, 28523, 28527, 28517, 28527, 28528, 28522, 28517, 28527, 28520, 28511, 28524, 28516, 28529, 28531, 28532, 28535, 28536, 28534, 28540, 28542, 28541, 28544, 28542, 28540, 28547, 28548, 28546, 28551, 28552, 28550, 28545, 28546, 28532, 28535, 28549, 28550, 28534, 28540, 28539, 28536, 28543, 28540, 28532, 28535, 28533, 28539, 28541, 28538, 28548, 28551, 28549, 28530, 28533, 28539, 28546, 28549, 28535, 28558, 28560, 28559, 28562, 28564, 28563, 28567, 28568, 28566, 28571, 28572, 28570, 28565, 28566, 28554, 28555, 28569, 28570, 28553, 28554, 28558, 28556, 28562, 28561, 28561, 28563, 28560, 28568, 28571, 28569, 28554, 28555, 28561, 28554, 28566, 28569, 28575, 28576, 28574, 28579, 28580, 28578, 28584, 28582, 28581,
28593, 28591, 28590, 28586, 28587, 28580, 28584, 28589, 28588, 28581, 28582, 28576, 28590, 28591, 28587, 28574, 28576, 28579, 28592, 28590, 28588, 28588, 28590, 28586, 28585, 28586, 28579, 28576, 28582, 28588, 28596, 28597, 28595, 28600, 28601, 28599, 28609, 28607, 28606, 28613, 28611, 28610, 28602, 28603, 28597, 28604, 28605, 28601, 28606, 28607, 28603, 28610, 28611, 28605, 28597, 28600, 28598, 28612, 28610, 28607, 28607, 28610, 28604, 28597, 28603, 28604, 28573, 28574, 28548, 28577, 28578, 28552, 28574, 28577, 28551, 28594, 28595, 28568, 28598, 28599, 28572, 28595, 28598, 28571, 28557, 28559, 28544, 28552, 28567, 28565, 28553, 28557, 28543, 28550, 28565, 28553, 28580, 28596, 28594, 28608, 28606, 28591, 28591, 28606, 28602, 28587, 28602, 28596, 28578, 28594, 28567, 28616, 28618, 28619, 28621, 28623, 28622, 28625, 28627, 28626, 28626, 28629, 28628, 28632, 28633, 28631, 28636, 28637, 28635, 28614, 28630, 28631, 28634, 28635, 28619, 28615, 28621, 28620, 28616, 28617, 28625, 28624, 28628, 28618, 28628, 28629, 28623, 28633, 28636, 28634, 28615, 28618, 28628, 28615, 28631, 28634, 28640, 28641, 28639, 28644, 28645, 28643, 28650, 28651, 28649, 28654, 28655, 28653, 28648, 28649, 28641, 28652, 28653, 28645, 28641, 28644, 28642, 28651, 28654, 28652,
28649, 28652, 28644, 28658, 28659, 28657, 28662, 28663, 28661, 28671, 28669, 28668, 28675, 28673, 28672, 28658, 28664, 28665, 28662, 28666, 28667, 28669, 28665, 28664, 28673, 28667, 28666, 28659, 28662, 28660, 28674, 28672, 28669, 28672, 28666, 28665, 28659, 28665, 28666, 28676, 28678, 28679, 28680, 28682, 28683, 28687, 28685, 28684, 28693, 28691, 28690, 28693, 28689, 28688, 28696, 28686, 28684, 28684, 28685, 28679, 28691, 28683, 28682, 28691, 28688, 28681, 28677, 28679, 28682, 28692, 28690, 28685, 28690, 28682, 28679, 28695, 28684, 28678, 28656, 28657, 28633, 28660, 28661, 28637, 28657, 28660, 28636, 28676, 28677, 28651, 28680, 28681, 28655, 28677, 28680, 28654, 28617, 28619, 28640, 28646, 28647, 28627, 28637, 28650, 28648, 28638, 28646, 28625, 28635, 28648, 28640, 28663, 28678, 28676, 28696, 28695, 28673, 28673, 28695, 28694, 28663, 28667, 28694, 28661, 28676, 28650, 28529, 28530, 28502, 28538, 28506, 28504, 28530, 28537, 28504, 28513, 28638, 28639, 28642, 28643, 28518, 28521, 28647, 28646, 28514, 28639, 28642, 28519, 28646, 28638, 28620, 28622, 28564, 28572, 28632, 28630, 28614, 28620, 28562, 28570, 28630, 28614, 28601, 28658, 28656, 28670, 28668, 28611, 28668, 28664, 28605, 28605, 28664, 28658, 28599, 28656, 28632, 28701, 28702, 28700,
28704, 28706, 28705, 28712, 28713, 28711, 28717, 28719, 28718, 28715, 28719, 28717, 28710, 28711, 28702, 28706, 28715, 28714, 28703, 28705, 28698, 28709, 28717, 28716, 28707, 28714, 28717, 28697, 28698, 28701, 28707, 28709, 28712, 28698, 28707, 28710, 28705, 28714, 28707, 28722, 28723, 28721, 28724, 28726, 28727, 28731, 28729, 28728, 28734, 28735, 28733, 28742, 28740, 28739, 28744, 28741, 28739, 28732, 28733, 28723, 28726, 28736, 28737, 28742, 28731, 28730, 28730, 28728, 28725, 28740, 28737, 28736, 28738, 28743, 28739, 28723, 28726, 28724, 28733, 28735, 28738, 28733, 28736, 28726, 28737, 28740, 28730, 28747, 28748, 28746, 28751, 28752, 28750, 28757, 28758, 28756, 28761, 28762, 28760, 28755, 28756, 28748, 28759, 28760, 28752, 28748, 28751, 28749, 28758, 28761, 28759, 28756, 28759, 28751, 28765, 28766, 28764, 28774, 28772, 28771, 28775, 28777, 28778, 28784, 28782, 28781, 28784, 28780, 28779, 28765, 28775, 28776, 28780, 28770, 28769, 28782, 28778, 28777, 28782, 28779, 28776, 28768, 28773, 28771, 28764, 28766, 28768, 28774, 28773, 28769, 28773, 28768, 28766, 28776, 28779, 28769, 28708, 28745, 28746, 28749, 28750, 28713, 28718, 28754, 28753, 28746, 28749, 28712, 28708, 28716, 28753, 28763, 28764, 28735, 28772, 28744, 28743, 28735, 28764, 28767,
28771, 28743, 28738, 28702, 28722, 28720, 28713, 28734, 28732, 28711, 28732, 28722, 28752, 28765, 28763, 28762, 28777, 28775, 28760, 28775, 28765, 28750, 28763, 28734, 28787, 28788, 28786, 28791, 28792, 28790, 28795, 28797, 28798, 28805, 28803, 28802, 28809, 28807, 28806, 28811, 28808, 28806, 28795, 28796, 28788, 28791, 28799, 28800, 28809, 28794, 28793, 28803, 28798, 28797, 28807, 28800, 28799, 28810, 28806, 28799, 28788, 28791, 28789, 28796, 28798, 28801, 28811, 28810, 28803, 28810, 28801, 28798, 28796, 28799, 28791, 28807, 28793, 28792, 28817, 28815, 28814, 28816, 28814, 28812, 28820, 28804, 28802, 28819, 28802, 28797, 28790, 28792, 28813, 28817, 28816, 28793, 28793, 28816, 28813, 28785, 28786, 28758, 28789, 28790, 28762, 28786, 28789, 28761, 28815, 28783, 28781, 28814, 28781, 28777, 28762, 28790, 28812, 28699, 28700, 28470, 28472, 28704, 28703, 28466, 28697, 28699, 28466, 28471, 28703, 28720, 28721, 28492, 28724, 28725, 28496, 28729, 28500, 28499, 28721, 28724, 28495, 28728, 28499, 28496, 28700, 28720, 28491, 28323, 28334, 28332, 28342, 28340, 28329, 28340, 28334, 28323, 28507, 28508, 28747, 28753, 28754, 28510, 28518, 28757, 28755, 28507, 28745, 28753, 28508, 28516, 28755, 28645, 28787, 28785, 28655, 28797, 28795, 28653, 28795, 28787,
28820, 28819, 28688, 28688, 28819, 28818, 28655, 28681, 28818, 28643, 28785, 28757, 28268, 28269, 28083, 28272, 28273, 28087, 28269, 28272, 28086, 28284, 28285, 28099, 28288, 28289, 28103, 28285, 28288, 28102, 28273, 28284, 28098, 28300, 28301, 28139, 28304, 28305, 28143, 28301, 28304, 28142, 28316, 28317, 28155, 28320, 28321, 28159, 28317, 28320, 28158, 28305, 28316, 28154, 28289, 28300, 28138, 28332, 28333, 28239, 28336, 28337, 28243, 28333, 28336, 28242, 28354, 28355, 28259, 28358, 28359, 28263, 28363, 28267, 28266, 28355, 28358, 28262, 28362, 28266, 28263, 28337, 28354, 28258, 28321, 28332, 28238, 28827, 28829, 28828, 28825, 28829, 28827, 28830, 28832, 28833, 28835, 28837, 28836, 28822, 28830, 28831, 28828, 28835, 28834, 28822, 28823, 28827, 28824, 28827, 28823, 28834, 28836, 28832, 28826, 28834, 28830, 28840, 28842, 28843, 28845, 28847, 28846, 28849, 28851, 28850, 28850, 28853, 28852, 28839, 28845, 28844, 28840, 28841, 28849, 28848, 28852, 28842, 28852, 28853, 28847, 28842, 28852, 28845, 28859, 28861, 28860, 28863, 28865, 28864, 28868, 28869, 28867, 28870, 28872, 28873, 28854, 28866, 28867, 28870, 28871, 28857, 28860, 28875, 28874, 28877, 28865, 28863, 28854, 28855, 28859, 28857, 28863, 28862, 28862, 28864, 28861, 28869, 28872, 28870,
28856, 28862, 28859, 28855, 28867, 28870, 28858, 28874, 28866, 28871, 28876, 28863, 28880, 28881, 28879, 28884, 28885, 28883, 28890, 28891, 28889, 28894, 28895, 28893, 28888, 28889, 28881, 28892, 28893, 28885, 28881, 28884, 28882, 28891, 28894, 28892, 28889, 28892, 28884, 28898, 28899, 28897, 28902, 28903, 28901, 28906, 28907, 28905, 28910, 28911, 28909, 28904, 28905, 28899, 28908, 28909, 28903, 28899, 28902, 28900, 28907, 28910, 28908, 28905, 28908, 28902, 28838, 28878, 28879, 28882, 28883, 28843, 28846, 28887, 28886, 28839, 28879, 28882, 28844, 28886, 28878, 28896, 28897, 28869, 28900, 28901, 28873, 28897, 28900, 28872, 28841, 28843, 28868, 28874, 28875, 28851, 28866, 28874, 28849, 28885, 28898, 28896, 28895, 28906, 28904, 28893, 28904, 28898, 28883, 28896, 28868, 28916, 28917, 28915, 28919, 28921, 28920, 28928, 28929, 28927, 28931, 28933, 28932, 28926, 28927, 28917, 28921, 28931, 28930, 28912, 28918, 28920, 28930, 28932, 28925, 28912, 28913, 28916, 28922, 28923, 28919, 28924, 28925, 28928, 28914, 28922, 28918, 28913, 28924, 28926, 28920, 28930, 28924, 28938, 28939, 28937, 28941, 28943, 28942, 28948, 28949, 28947, 28951, 28953, 28952, 28946, 28947, 28939, 28943, 28951, 28950, 28940, 28942, 28935, 28950, 28952, 28945, 28935, 28938, 28936,
28945, 28948, 28946, 28935, 28944, 28946, 28942, 28950, 28944, 28936, 28937, 28929, 28933, 28941, 28940, 28925, 28934, 28936, 28932, 28940, 28934, 28956, 28957, 28955, 28960, 28961, 28959, 28966, 28967, 28965, 28968, 28970, 28971, 28975, 28973, 28972, 28964, 28965, 28957, 28960, 28968, 28969, 28973, 28963, 28962, 28974, 28972, 28969, 28957, 28960, 28958, 28967, 28970, 28968, 28965, 28968, 28960, 28972, 28962, 28961, 28984, 28982, 28981, 28986, 28983, 28981, 28990, 28988, 28987, 28982, 28977, 28976, 28985, 28981, 28976, 28987, 28988, 28980, 28989, 28987, 28982, 28987, 28979, 28977, 28993, 28994, 28992, 29001, 28999, 28998, 29003, 29000, 28998, 29006, 29007, 29005, 29013, 29011, 29010, 29004, 29005, 28994, 29011, 29003, 29002, 28999, 28996, 28995, 28997, 29002, 28998, 29012, 29010, 29008, 28994, 28997, 28995, 29005, 29007, 29009, 29005, 29008, 28997, 29010, 29002, 28997, 28991, 28992, 28967, 28970, 28995, 28996, 29001, 28975, 28974, 28992, 28995, 28970, 28996, 28999, 28974, 28959, 28961, 28978, 28986, 28985, 28962, 28985, 28978, 28961, 28832, 28914, 28915, 28837, 28923, 28922, 28832, 28836, 28922, 28954, 28955, 28891, 28958, 28959, 28895, 28955, 28958, 28894, 28906, 28976, 28977, 28979, 28980, 28911, 28907, 28977, 28979, 28959, 28976, 28906,
28821, 28823, 28880, 28886, 28887, 28825, 28833, 28890, 28888, 28878, 28886, 28824, 28831, 28888, 28880, 28917, 28956, 28954, 28929, 28966, 28964, 28927, 28964, 28956, 28939, 28993, 28991, 28949, 29006, 29004, 28947, 29004, 28993, 28937, 28991, 28966, 28915, 28954, 28890, 29016, 29017, 29015, 29022, 29024, 29023, 29028, 29030, 29029, 29029, 29026, 29025, 29014, 29015, 29022, 29019, 29020, 29028, 29019, 29027, 29025, 29017, 29019, 29018, 29025, 29026, 29024, 29018, 29025, 29022, 29036, 29037, 29035, 29040, 29041, 29039, 29044, 29045, 29043, 29048, 29049, 29047, 29042, 29043, 29037, 29046, 29047, 29041, 29037, 29040, 29038, 29045, 29048, 29046, 29043, 29046, 29040, 29052, 29053, 29051, 29055, 29057, 29056, 29059, 29057, 29055, 29060, 29062, 29063, 29071, 29069, 29068, 29060, 29061, 29053, 29067, 29059, 29058, 29051, 29055, 29054, 29053, 29058, 29055, 29065, 29070, 29068, 29063, 29065, 29064, 29068, 29069, 29067, 29061, 29064, 29068, 29061, 29066, 29058, 29034, 29035, 29017, 29038, 29039, 29020, 29017, 29035, 29038, 29056, 29033, 29032, 29050, 29054, 29032, 29032, 29033, 29030, 29020, 29031, 29032, 29039, 29041, 29052, 29049, 29062, 29060, 29047, 29060, 29052, 29039, 29050, 29031, 29074, 29075, 29073, 29078, 29079, 29077, 29085, 29083, 29082,
29081, 29084, 29082, 29088, 29090, 29091, 29093, 29095, 29094, 29088, 29089, 29079, 29085, 29093, 29092, 29082, 29083, 29075, 29072, 29080, 29082, 29086, 29092, 29094, 29073, 29075, 29078, 29087, 29090, 29088, 29097, 29096, 29094, 29094, 29096, 29090, 29086, 29088, 29078, 29075, 29083, 29092, 29100, 29101, 29099, 29104, 29105, 29103, 29107, 29109, 29108, 29111, 29109, 29107, 29114, 29115, 29113, 29116, 29118, 29119, 29123, 29121, 29120, 29112, 29113, 29101, 29116, 29117, 29105, 29121, 29111, 29110, 29102, 29103, 29107, 29110, 29107, 29103, 29119, 29122, 29120, 29099, 29101, 29104, 29115, 29118, 29116, 29113, 29116, 29104, 29117, 29120, 29110, 29124, 29126, 29127, 29129, 29131, 29130, 29134, 29135, 29133, 29137, 29139, 29138, 29126, 29132, 29133, 29131, 29137, 29136, 29124, 29128, 29130, 29132, 29136, 29138, 29126, 29130, 29136, 29142, 29143, 29141, 29146, 29147, 29145, 29151, 29149, 29148, 29154, 29155, 29153, 29158, 29159, 29157, 29163, 29161, 29160, 29152, 29153, 29143, 29156, 29157, 29147, 29161, 29151, 29150, 29150, 29148, 29145, 29162, 29160, 29157, 29143, 29146, 29144, 29155, 29158, 29156, 29153, 29156, 29146, 29160, 29150, 29147, 29124, 29125, 29091, 29097, 29129, 29128, 29096, 29128, 29124, 29140, 29141, 29115, 29118, 29144, 29145,
29149, 29123, 29122, 29141, 29144, 29118, 29145, 29148, 29122, 29079, 29100, 29098, 29091, 29114, 29112, 29089, 29112, 29100, 29127, 29142, 29140, 29135, 29154, 29152, 29133, 29152, 29142, 29125, 29140, 29114, 29044, 29072, 29073, 29076, 29077, 29049, 29073, 29076, 29048, 29098, 29099, 29063, 29108, 29071, 29070, 29099, 29102, 29065, 29106, 29070, 29065, 29077, 29098, 29062, 29168, 29169, 29167, 29171, 29173, 29172, 29178, 29179, 29177, 29181, 29183, 29182, 29176, 29177, 29169, 29173, 29181, 29180, 29170, 29172, 29165, 29180, 29182, 29175, 29165, 29168, 29166, 29175, 29178, 29176, 29174, 29176, 29168, 29172, 29180, 29174, 29188, 29189, 29187, 29191, 29193, 29192, 29198, 29199, 29197, 29201, 29203, 29202, 29196, 29197, 29189, 29193, 29201, 29200, 29190, 29192, 29185, 29200, 29202, 29195, 29185, 29188, 29186, 29195, 29198, 29196, 29194, 29196, 29188, 29192, 29200, 29194, 29186, 29187, 29179, 29183, 29191, 29190, 29184, 29186, 29178, 29182, 29190, 29184, 29206, 29207, 29205, 29213, 29211, 29210, 29216, 29217, 29215, 29223, 29221, 29220, 29214, 29215, 29207, 29221, 29213, 29212, 29212, 29210, 29208, 29222, 29220, 29218, 29207, 29209, 29208, 29217, 29219, 29218, 29215, 29218, 29209, 29220, 29212, 29209, 29226, 29227, 29225, 29233, 29231, 29230,
29236, 29237, 29235, 29243, 29241, 29240, 29234, 29235, 29227, 29241, 29233, 29232, 29232, 29230, 29228, 29242, 29240, 29238, 29227, 29229, 29228, 29237, 29239, 29238, 29235, 29238, 29229, 29240, 29232, 29229, 29224, 29225, 29217, 29231, 29223, 29222, 29225, 29228, 29219, 29230, 29222, 29219, 29248, 29249, 29247, 29251, 29253, 29252, 29258, 29259, 29257, 29261, 29263, 29262, 29256, 29257, 29249, 29253, 29261, 29260, 29250, 29252, 29245, 29260, 29262, 29255, 29244, 29245, 29248, 29254, 29255, 29258, 29245, 29254, 29256, 29252, 29260, 29254, 29272, 29273, 29271, 29277, 29279, 29278, 29275, 29279, 29277, 29280, 29282, 29283, 29286, 29287, 29285, 29289, 29291, 29290, 29268, 29280, 29281, 29284, 29285, 29273, 29278, 29289, 29288, 29269, 29277, 29276, 29274, 29277, 29269, 29288, 29290, 29282, 29267, 29269, 29272, 29283, 29286, 29284, 29281, 29284, 29272, 29276, 29288, 29280, 29270, 29271, 29259, 29263, 29275, 29274, 29255, 29267, 29270, 29262, 29274, 29267, 29290, 29291, 29266, 29282, 29290, 29265, 29294, 29295, 29293, 29301, 29299, 29298, 29304, 29305, 29303, 29311, 29309, 29308, 29302, 29303, 29295, 29309, 29301, 29300, 29300, 29298, 29296, 29310, 29308, 29306, 29295, 29297, 29296, 29305, 29307, 29306, 29303, 29306, 29297, 29308, 29300, 29297,
29314, 29315, 29313, 29321, 29319, 29318, 29324, 29325, 29323, 29331, 29329, 29328, 29322, 29323, 29315, 29329, 29321, 29320, 29320, 29318, 29316, 29330, 29328, 29326, 29313, 29315, 29317, 29323, 29325, 29327, 29315, 29323, 29326, 29328, 29320, 29317, 29312, 29313, 29305, 29319, 29311, 29310, 29305, 29313, 29316, 29318, 29310, 29307, 29246, 29247, 29199, 29203, 29251, 29250, 29195, 29244, 29246, 29202, 29250, 29244, 29292, 29293, 29237, 29299, 29243, 29242, 29293, 29296, 29239, 29298, 29242, 29239, 29169, 29206, 29204, 29179, 29216, 29214, 29177, 29214, 29206, 29189, 29226, 29224, 29199, 29236, 29234, 29197, 29234, 29226, 29187, 29224, 29216, 29249, 29294, 29292, 29259, 29304, 29302, 29257, 29302, 29294, 29273, 29314, 29312, 29287, 29324, 29322, 29285, 29322, 29314, 29271, 29312, 29304, 29247, 29292, 29236, 29334, 29335, 29333, 29337, 29339, 29338, 29342, 29343, 29341, 29345, 29347, 29346, 29340, 29341, 29335, 29339, 29345, 29344, 29336, 29338, 29334, 29344, 29346, 29342, 29338, 29344, 29340, 29350, 29351, 29349, 29354, 29355, 29353, 29359, 29357, 29356, 29362, 29363, 29361, 29366, 29367, 29365, 29371, 29369, 29368, 29360, 29361, 29351, 29364, 29365, 29355, 29369, 29359, 29358, 29358, 29356, 29353, 29370, 29368, 29365, 29351, 29354, 29352,
29363, 29366, 29364, 29361, 29364, 29354, 29368, 29358, 29355, 29374, 29375, 29373, 29377, 29379, 29378, 29382, 29383, 29381, 29385, 29387, 29386, 29380, 29381, 29375, 29379, 29385, 29384, 29376, 29378, 29374, 29384, 29386, 29382, 29378, 29384, 29380, 29390, 29391, 29389, 29394, 29395, 29393, 29399, 29397, 29396, 29402, 29403, 29401, 29406, 29407, 29405, 29411, 29409, 29408, 29400, 29401, 29391, 29404, 29405, 29395, 29409, 29399, 29398, 29398, 29396, 29393, 29410, 29408, 29405, 29391, 29394, 29392, 29403, 29406, 29404, 29401, 29404, 29394, 29408, 29398, 29395, 29372, 29373, 29343, 29347, 29377, 29376, 29346, 29376, 29372, 29388, 29389, 29363, 29392, 29393, 29367, 29397, 29371, 29370, 29389, 29392, 29366, 29396, 29370, 29367, 29335, 29350, 29348, 29343, 29362, 29360, 29341, 29360, 29350, 29375, 29390, 29388, 29383, 29402, 29400, 29381, 29400, 29390, 29373, 29388, 29362, 29414, 29415, 29413, 29417, 29419, 29418, 29422, 29423, 29421, 29425, 29427, 29426, 29420, 29421, 29415, 29419, 29425, 29424, 29416, 29418, 29414, 29424, 29426, 29422, 29418, 29424, 29420, 29430, 29431, 29429, 29434, 29435, 29433, 29439, 29437, 29436, 29442, 29443, 29441, 29446, 29447, 29445, 29451, 29449, 29448, 29440, 29441, 29431, 29444, 29445, 29435, 29449, 29439, 29438,
29438, 29436, 29433, 29450, 29448, 29445, 29431, 29434, 29432, 29443, 29446, 29444, 29441, 29444, 29434, 29448, 29438, 29435, 29454, 29455, 29453, 29457, 29459, 29458, 29462, 29463, 29461, 29465, 29467, 29466, 29460, 29461, 29455, 29459, 29465, 29464, 29456, 29458, 29454, 29464, 29466, 29462, 29458, 29464, 29460, 29470, 29471, 29469, 29474, 29475, 29473, 29479, 29477, 29476, 29482, 29483, 29481, 29486, 29487, 29485, 29491, 29489, 29488, 29480, 29481, 29471, 29484, 29485, 29475, 29489, 29479, 29478, 29478, 29476, 29473, 29490, 29488, 29485, 29471, 29474, 29472, 29483, 29486, 29484, 29481, 29484, 29474, 29488, 29478, 29475, 29452, 29453, 29423, 29427, 29457, 29456, 29426, 29456, 29452, 29468, 29469, 29443, 29472, 29473, 29447, 29477, 29451, 29450, 29469, 29472, 29446, 29476, 29450, 29447, 29415, 29430, 29428, 29423, 29442, 29440, 29421, 29440, 29430, 29455, 29470, 29468, 29463, 29482, 29480, 29461, 29480, 29470, 29453, 29468, 29442, 29412, 29413, 29383, 29387, 29417, 29416, 29386, 29416, 29412, 29428, 29429, 29403, 29432, 29433, 29407, 29437, 29411, 29410, 29429, 29432, 29406, 29436, 29410, 29407, 29413, 29428, 29402, 29166, 29167, 28949, 28953, 29171, 29170, 29164, 29166, 28948, 28952, 29170, 29164, 29204, 29205, 29007, 29211, 29013, 29012,
29205, 29208, 29009, 29210, 29012, 29009, 29167, 29204, 29006, 29134, 29332, 29333, 29139, 29337, 29336, 29134, 29138, 29336, 29348, 29349, 29155, 29352, 29353, 29159, 29357, 29163, 29162, 29349, 29352, 29158, 29356, 29162, 29159, 29333, 29348, 29154, 28871, 28873, 29016, 29021, 29023, 28877, 29014, 29021, 28876, 28903, 29036, 29034, 28911, 29044, 29042, 28909, 29042, 29036, 28901, 29034, 29016, 29081, 29080, 28988, 28988, 29080, 29072, 28980, 29072, 29044, 29494, 29496, 29497, 29499, 29501, 29500, 29503, 29505, 29504, 29504, 29507, 29506, 29510, 29511, 29509, 29514, 29515, 29513, 29492, 29508, 29509, 29512, 29513, 29497, 29492, 29493, 29499, 29494, 29495, 29503, 29502, 29506, 29496, 29506, 29507, 29501, 29511, 29514, 29512, 29493, 29496, 29506, 29493, 29509, 29512, 29518, 29519, 29517, 29522, 29523, 29521, 29526, 29527, 29525, 29530, 29531, 29529, 29524, 29525, 29519, 29528, 29529, 29523, 29519, 29522, 29520, 29527, 29530, 29528, 29525, 29528, 29522, 29536, 29537, 29535, 29540, 29541, 29539, 29549, 29547, 29546, 29553, 29551, 29550, 29536, 29542, 29543, 29540, 29544, 29545, 29547, 29543, 29542, 29551, 29545, 29544, 29537, 29540, 29538, 29552, 29550, 29547, 29550, 29544, 29543, 29537, 29543, 29544, 29554, 29556, 29557, 29564, 29562, 29561,
29568, 29566, 29565, 29570, 29567, 29565, 29573, 29563, 29561, 29562, 29557, 29556, 29565, 29566, 29559, 29569, 29565, 29558, 29555, 29557, 29560, 29570, 29569, 29562, 29569, 29560, 29557, 29572, 29561, 29556, 29534, 29535, 29511, 29538, 29539, 29515, 29535, 29538, 29514, 29554, 29555, 29527, 29530, 29558, 29559, 29568, 29533, 29532, 29555, 29558, 29530, 29566, 29532, 29531, 29497, 29518, 29516, 29515, 29526, 29524, 29513, 29524, 29518, 29541, 29556, 29554, 29573, 29572, 29551, 29551, 29572, 29571, 29541, 29545, 29571, 29539, 29554, 29526, 29574, 29576, 29577, 29582, 29580, 29579, 29591, 29589, 29588, 29593, 29590, 29588, 29576, 29585, 29586, 29591, 29582, 29581, 29581, 29579, 29575, 29589, 29586, 29585, 29587, 29592, 29588, 29584, 29583, 29579, 29583, 29578, 29575, 29589, 29581, 29577, 29523, 29576, 29574, 29529, 29531, 29587, 29593, 29592, 29532, 29592, 29587, 29531, 29529, 29585, 29576, 29596, 29597, 29595, 29599, 29601, 29600, 29604, 29605, 29603, 29607, 29609, 29608, 29602, 29603, 29597, 29601, 29607, 29606, 29598, 29600, 29596, 29606, 29608, 29604, 29600, 29606, 29602, 29612, 29613, 29611, 29616, 29617, 29615, 29621, 29619, 29618, 29624, 29625, 29623, 29628, 29629, 29627, 29633, 29631, 29630, 29622, 29623, 29613, 29626, 29627, 29617,
29631, 29621, 29620, 29620, 29618, 29615, 29632, 29630, 29627, 29613, 29616, 29614, 29625, 29628, 29626, 29623, 29626, 29616, 29630, 29620, 29617, 29639, 29637, 29636, 29637, 29635, 29634, 29647, 29645, 29644, 29651, 29649, 29648, 29645, 29641, 29640, 29649, 29643, 29642, 29650, 29648, 29645, 29648, 29642, 29641, 29634, 29635, 29605, 29609, 29638, 29636, 29608, 29636, 29634, 29640, 29641, 29625, 29642, 29643, 29629, 29651, 29633, 29632, 29641, 29642, 29628, 29649, 29632, 29629, 29597, 29612, 29610, 29605, 29624, 29622, 29603, 29622, 29612, 29646, 29644, 29637, 29644, 29640, 29635, 29635, 29640, 29624, 29266, 29505, 29503, 29264, 29265, 29503, 29516, 29517, 29283, 29520, 29521, 29287, 29517, 29520, 29286, 29264, 29495, 29516, 29574, 29575, 29325, 29584, 29331, 29330, 29325, 29575, 29578, 29578, 29583, 29330, 29521, 29574, 29324, 29594, 29595, 29463, 29467, 29599, 29598, 29466, 29598, 29594, 29610, 29611, 29483, 29614, 29615, 29487, 29619, 29491, 29490, 29611, 29614, 29486, 29618, 29490, 29487, 29595, 29610, 29482, 27569, 27643, 27644, 27647, 27648, 27572, 27644, 27647, 27571, 27659, 27660, 27584, 27585, 27663, 27664, 27660, 27663, 27585, 27648, 27659, 27583, 27706, 27707, 27596, 27710, 27711, 27598, 27596, 27707, 27710, 27722, 27723, 27608,
27726, 27727, 27610, 27723, 27726, 27609, 27711, 27722, 27607, 27586, 27664, 27706, 27850, 27851, 27620, 27854, 27855, 27622, 27851, 27854, 27621, 27866, 27867, 27632, 27870, 27871, 27634, 27875, 27642, 27640, 27867, 27870, 27633, 27874, 27640, 27634, 27855, 27866, 27631, 27727, 27850, 27619, 29653, 29655, 29654, 29657, 29659, 29658, 29665, 29664, 29662, 29671, 29669, 29668, 29655, 29667, 29666, 29674, 29673, 29664, 29660, 29661, 29655, 29659, 29663, 29662, 29674, 29671, 29670, 29656, 29658, 29653, 29664, 29661, 29660, 29670, 29668, 29666, 29658, 29662, 29660, 29673, 29670, 29667, 29673, 29672, 29661, 29661, 29672, 29667, 29676, 29678, 29677, 29680, 29682, 29681, 29686, 29685, 29683, 29690, 29689, 29687, 29694, 29692, 29691, 29698, 29696, 29695, 29700, 29702, 29701, 29704, 29703, 29685, 29706, 29705, 29689, 29678, 29684, 29683, 29704, 29694, 29693, 29682, 29688, 29687, 29706, 29698, 29697, 29702, 29708, 29707, 29679, 29681, 29678, 29689, 29686, 29684, 29697, 29695, 29692, 29681, 29687, 29684, 29705, 29697, 29694, 29705, 29704, 29686, 29675, 29677, 29659, 29685, 29665, 29663, 29693, 29691, 29669, 29677, 29683, 29663, 29703, 29693, 29671, 29703, 29674, 29665, 29710, 29712, 29711, 29714, 29716, 29715, 29720, 29719, 29717, 29724, 29723, 29721,
29728, 29726, 29725, 29734, 29732, 29731, 29734, 29736, 29735, 29738, 29737, 29719, 29724, 29741, 29740, 29712, 29718, 29717, 29738, 29728, 29727, 29716, 29722, 29721, 29741, 29734, 29733, 29741, 29742, 29736, 29713, 29715, 29712, 29723, 29720, 29718, 29733, 29731, 29726, 29715, 29721, 29718, 29740, 29733, 29728, 29740, 29738, 29720, 29744, 29746, 29745, 29748, 29746, 29744, 29752, 29751, 29749, 29756, 29754, 29753, 29760, 29759, 29757, 29754, 29748, 29747, 29758, 29757, 29751, 29746, 29750, 29749, 29752, 29750, 29746, 29760, 29758, 29754, 29758, 29752, 29748, 29743, 29745, 29716, 29751, 29724, 29722, 29757, 29759, 29742, 29745, 29749, 29722, 29751, 29757, 29741, 29762, 29764, 29763, 29768, 29767, 29765, 29770, 29772, 29771, 29776, 29775, 29773, 29764, 29770, 29769, 29774, 29773, 29767, 29765, 29767, 29764, 29773, 29775, 29772, 29767, 29773, 29770, 29782, 29780, 29779, 29778, 29781, 29779, 29786, 29785, 29783, 29786, 29782, 29781, 29783, 29785, 29781, 29772, 29778, 29777, 29784, 29783, 29775, 29775, 29783, 29778, 29790, 29788, 29787, 29792, 29794, 29793, 29802, 29800, 29799, 29804, 29806, 29805, 29800, 29790, 29789, 29794, 29804, 29803, 29796, 29795, 29788, 29808, 29807, 29800, 29797, 29798, 29794, 29809, 29810, 29806, 29807, 29796, 29790,
29798, 29809, 29804, 29814, 29812, 29811, 29818, 29817, 29815, 29822, 29820, 29819, 29826, 29825, 29823, 29820, 29814, 29813, 29824, 29823, 29817, 29818, 29816, 29812, 29826, 29824, 29820, 29824, 29818, 29814, 29830, 29828, 29827, 29832, 29834, 29833, 29836, 29835, 29828, 29837, 29838, 29834, 29844, 29842, 29841, 29844, 29840, 29839, 29848, 29847, 29845, 29848, 29844, 29843, 29848, 29846, 29840, 29828, 29802, 29801, 29806, 29832, 29831, 29835, 29808, 29802, 29810, 29837, 29832, 29840, 29822, 29821, 29846, 29845, 29825, 29846, 29826, 29822, 29817, 29798, 29797, 29825, 29810, 29809, 29823, 29809, 29798, 29843, 29841, 29833, 29847, 29838, 29837, 29847, 29843, 29834, 29845, 29837, 29810, 29701, 29762, 29761, 29766, 29765, 29707, 29707, 29765, 29762, 29788, 29730, 29729, 29736, 29792, 29791, 29795, 29739, 29730, 29742, 29797, 29792, 29812, 29756, 29755, 29816, 29815, 29759, 29816, 29760, 29756, 29815, 29797, 29742, 29709, 29711, 29682, 29719, 29690, 29688, 29727, 29725, 29696, 29729, 29730, 29702, 29711, 29717, 29688, 29737, 29727, 29698, 29730, 29739, 29708, 29737, 29706, 29690, 29796, 29768, 29766, 29808, 29776, 29774, 29807, 29774, 29768, 29830, 29829, 29780, 29836, 29786, 29784, 29836, 29830, 29782, 29835, 29784, 29776, 29795, 29766, 29708,
29853, 29855, 29854, 29857, 29859, 29858, 29858, 29861, 29860, 29867, 29866, 29864, 29854, 29855, 29863, 29859, 29865, 29864, 29858, 29864, 29866, 29860, 29861, 29855, 29861, 29866, 29863, 29852, 29854, 29850, 29850, 29854, 29862, 29869, 29871, 29870, 29873, 29871, 29869, 29875, 29877, 29876, 29879, 29881, 29880, 29870, 29871, 29879, 29881, 29879, 29871, 29877, 29883, 29882, 29874, 29876, 29873, 29876, 29882, 29881, 29885, 29887, 29886, 29886, 29889, 29888, 29893, 29892, 29890, 29897, 29900, 29898, 29904, 29906, 29905, 29889, 29897, 29896, 29904, 29903, 29892, 29887, 29891, 29890, 29886, 29890, 29892, 29903, 29905, 29900, 29898, 29900, 29895, 29895, 29900, 29905, 29889, 29892, 29903, 29912, 29910, 29909, 29912, 29908, 29907, 29916, 29914, 29913, 29920, 29919, 29917, 29926, 29924, 29923, 29924, 29911, 29909, 29920, 29912, 29911, 29912, 29920, 29918, 29926, 29928, 29927, 29921, 29922, 29916, 29915, 29913, 29907, 29921, 29915, 29908, 29924, 29927, 29919, 29930, 29932, 29931, 29936, 29934, 29933, 29942, 29941, 29939, 29944, 29946, 29945, 29950, 29948, 29947, 29954, 29956, 29955, 29932, 29944, 29943, 29948, 29936, 29935, 29942, 29954, 29953, 29930, 29937, 29938, 29951, 29952, 29946, 29936, 29942, 29940, 29956, 29954, 29948, 29941, 29938, 29937,
29953, 29955, 29952, 29953, 29951, 29938, 29932, 29938, 29951, 29954, 29942, 29936, 29895, 29930, 29929, 29934, 29901, 29899, 29906, 29940, 29939, 29939, 29937, 29902, 29902, 29937, 29930, 29940, 29906, 29901, 29917, 29919, 29893, 29926, 29925, 29899, 29928, 29906, 29904, 29906, 29928, 29926, 29927, 29904, 29893, 29960, 29958, 29957, 29960, 29962, 29961, 29973, 29972, 29970, 29977, 29975, 29974, 29977, 29979, 29978, 29966, 29976, 29974, 29973, 29981, 29980, 29967, 29968, 29960, 29968, 29969, 29962, 29981, 29977, 29976, 29981, 29982, 29979, 29966, 29965, 29961, 29970, 29972, 29969, 29969, 29972, 29966, 29966, 29972, 29980, 29986, 29984, 29983, 29990, 29992, 29991, 29994, 29992, 29990, 29996, 29998, 29997, 29991, 29986, 29985, 29988, 29987, 29984, 29992, 29996, 29995, 29994, 29998, 29996, 29995, 29988, 29986, 30000, 30002, 30001, 30006, 30008, 30007, 30002, 30006, 30005, 30003, 30004, 30002, 30006, 30009, 30010, 30002, 30004, 30009, 30014, 30012, 30011, 30018, 30017, 30015, 30022, 30020, 30019, 30026, 30025, 30023, 30020, 30014, 30013, 30018, 30024, 30023, 30018, 30016, 30012, 30022, 30026, 30024, 30020, 30024, 30018, 29979, 30000, 29999, 29982, 30003, 30000, 30012, 29994, 29993, 29998, 30016, 30015, 30016, 29998, 29994, 29983, 29984, 29964,
29987, 29988, 29973, 29997, 29982, 29981, 29984, 29987, 29971, 29995, 29981, 29973, 30015, 30017, 30004, 30025, 30010, 30009, 30023, 30009, 30004, 29997, 30015, 30003, 29850, 29887, 29885, 29850, 29851, 29891, 29863, 29918, 29917, 29867, 29922, 29921, 29921, 29918, 29863, 29917, 29891, 29851, 29964, 29877, 29875, 29968, 29967, 29880, 29883, 29971, 29970, 29970, 29968, 29881, 29971, 29883, 29877, 29868, 29870, 29859, 29878, 29880, 29867, 29870, 29878, 29865, 29959, 29957, 29914, 29967, 29959, 29916, 29880, 29967, 29922, 30028, 30030, 30029, 30029, 30032, 30031, 30036, 30035, 30033, 30038, 30040, 30039, 30044, 30043, 30041, 30032, 30038, 30037, 30042, 30041, 30035, 30030, 30034, 30033, 30033, 30035, 30032, 30041, 30043, 30040, 30035, 30041, 30038, 30050, 30048, 30047, 30046, 30049, 30047, 30054, 30053, 30051, 30054, 30050, 30049, 30051, 30053, 30049, 30040, 30046, 30045, 30052, 30051, 30043, 30043, 30051, 30046, 30056, 30058, 30057, 30060, 30062, 30061, 30066, 30065, 30063, 30070, 30069, 30067, 30074, 30073, 30071, 30078, 30077, 30075, 30072, 30071, 30065, 30076, 30075, 30069, 30058, 30064, 30063, 30062, 30068, 30067, 30059, 30061, 30058, 30069, 30066, 30064, 30077, 30074, 30072, 30061, 30067, 30064, 30075, 30072, 30066, 30080, 30082, 30081,
30084, 30086, 30085, 30090, 30089, 30087, 30094, 30093, 30091, 30098, 30097, 30095, 30102, 30101, 30099, 30096, 30095, 30089, 30100, 30099, 30093, 30082, 30088, 30087, 30086, 30092, 30091, 30083, 30085, 30082, 30093, 30090, 30088, 30101, 30098, 30096, 30085, 30091, 30088, 30099, 30096, 30090, 30106, 30104, 30103, 30110, 30108, 30107, 30114, 30113, 30111, 30118, 30117, 30115, 30114, 30106, 30105, 30118, 30110, 30109, 30109, 30107, 30104, 30117, 30114, 30112, 30117, 30109, 30106, 30122, 30120, 30119, 30126, 30124, 30123, 30130, 30129, 30127, 30134, 30133, 30131, 30130, 30122, 30121, 30134, 30126, 30125, 30125, 30123, 30120, 30133, 30130, 30128, 30133, 30125, 30122, 30112, 30111, 30073, 30116, 30115, 30077, 30115, 30112, 30074, 30128, 30127, 30097, 30132, 30131, 30101, 30131, 30128, 30098, 30079, 30081, 30062, 30089, 30070, 30068, 30097, 30078, 30076, 30081, 30087, 30068, 30095, 30076, 30070, 30121, 30119, 30108, 30129, 30118, 30116, 30129, 30121, 30110, 30127, 30116, 30078, 30055, 30057, 30030, 30065, 30036, 30034, 30073, 30044, 30042, 30057, 30063, 30034, 30071, 30042, 30036, 30105, 30103, 30048, 30113, 30054, 30052, 30113, 30105, 30050, 30111, 30052, 30044, 30139, 30142, 30141, 30147, 30146, 30144, 30151, 30149, 30148, 30151, 30153, 30152,
30142, 30150, 30148, 30147, 30155, 30154, 30139, 30144, 30146, 30155, 30151, 30150, 30151, 30155, 30156, 30139, 30137, 30135, 30144, 30139, 30136, 30146, 30154, 30150, 30160, 30158, 30157, 30164, 30166, 30165, 30170, 30166, 30164, 30172, 30174, 30173, 30178, 30177, 30175, 30165, 30160, 30159, 30162, 30161, 30158, 30165, 30166, 30176, 30178, 30176, 30166, 30174, 30180, 30179, 30171, 30173, 30170, 30170, 30173, 30179, 30175, 30162, 30160, 30184, 30182, 30181, 30184, 30186, 30185, 30190, 30188, 30187, 30195, 30197, 30196, 30191, 30192, 30184, 30192, 30193, 30186, 30197, 30190, 30189, 30189, 30187, 30185, 30194, 30196, 30193, 30196, 30189, 30186, 30169, 30183, 30181, 30192, 30191, 30177, 30180, 30195, 30194, 30194, 30192, 30178, 30177, 30191, 30183, 30157, 30158, 30140, 30161, 30162, 30147, 30169, 30167, 30152, 30177, 30156, 30155, 30158, 30161, 30145, 30156, 30177, 30169, 30175, 30155, 30147, 30201, 30203, 30202, 30199, 30203, 30201, 30210, 30212, 30211, 30208, 30212, 30210, 30221, 30223, 30222, 30202, 30214, 30213, 30221, 30220, 30205, 30212, 30219, 30218, 30208, 30217, 30219, 30202, 30203, 30206, 30199, 30204, 30206, 30213, 30214, 30208, 30222, 30219, 30217, 30214, 30220, 30217, 30205, 30220, 30214, 30229, 30227, 30226, 30229, 30225, 30224,
30233, 30232, 30230, 30237, 30235, 30234, 30235, 30228, 30226, 30232, 30233, 30229, 30233, 30231, 30225, 30237, 30239, 30238, 30238, 30232, 30228, 30245, 30243, 30242, 30245, 30241, 30240, 30249, 30247, 30246, 30253, 30252, 30250, 30253, 30245, 30244, 30245, 30253, 30251, 30254, 30255, 30249, 30248, 30246, 30240, 30254, 30248, 30241, 30249, 30216, 30215, 30219, 30251, 30250, 30255, 30254, 30222, 30222, 30254, 30251, 30249, 30255, 30223, 30230, 30232, 30206, 30237, 30236, 30215, 30239, 30223, 30221, 30223, 30239, 30237, 30232, 30238, 30221, 30259, 30261, 30260, 30266, 30268, 30267, 30272, 30270, 30269, 30278, 30277, 30275, 30257, 30266, 30265, 30276, 30275, 30263, 30273, 30274, 30268, 30261, 30264, 30263, 30278, 30272, 30271, 30258, 30260, 30257, 30271, 30269, 30267, 30277, 30274, 30273, 30260, 30263, 30262, 30277, 30271, 30268, 30275, 30273, 30262, 30262, 30273, 30266, 30280, 30282, 30281, 30284, 30286, 30285, 30294, 30292, 30291, 30298, 30296, 30295, 30302, 30301, 30299, 30306, 30305, 30303, 30300, 30299, 30287, 30304, 30303, 30289, 30282, 30288, 30287, 30302, 30294, 30293, 30286, 30290, 30289, 30306, 30298, 30297, 30283, 30285, 30282, 30297, 30295, 30292, 30305, 30302, 30300, 30285, 30289, 30288, 30305, 30297, 30294, 30303, 30300, 30288,
30279, 30281, 30261, 30293, 30291, 30270, 30301, 30278, 30276, 30281, 30287, 30264, 30301, 30293, 30272, 30299, 30276, 30264, 30308, 30310, 30309, 30312, 30314, 30313, 30322, 30320, 30319, 30326, 30324, 30323, 30330, 30329, 30327, 30332, 30334, 30333, 30328, 30327, 30315, 30332, 30331, 30317, 30310, 30316, 30315, 30330, 30322, 30321, 30314, 30318, 30317, 30334, 30326, 30325, 30311, 30313, 30310, 30325, 30323, 30320, 30333, 30330, 30328, 30313, 30317, 30316, 30333, 30325, 30322, 30331, 30328, 30316, 30336, 30338, 30337, 30344, 30342, 30341, 30348, 30347, 30345, 30342, 30338, 30336, 30346, 30345, 30339, 30338, 30340, 30339, 30348, 30346, 30342, 30346, 30340, 30338, 30335, 30337, 30314, 30345, 30347, 30334, 30337, 30339, 30318, 30345, 30332, 30318, 30209, 30211, 30174, 30174, 30211, 30218, 30244, 30242, 30188, 30252, 30197, 30195, 30197, 30252, 30244, 30250, 30195, 30180, 30307, 30309, 30286, 30321, 30319, 30296, 30329, 30306, 30304, 30309, 30315, 30290, 30329, 30321, 30298, 30327, 30304, 30290, 29946, 30136, 30135, 30140, 29950, 29949, 30145, 30144, 29955, 29955, 30144, 30143, 29946, 29952, 30143, 30140, 30145, 29956, 30008, 30199, 30198, 30008, 30010, 30204, 30225, 30022, 30021, 30231, 30230, 30025, 30225, 30231, 30026, 30230, 30204, 30010,
30356, 30358, 30357, 30362, 30360, 30359, 30350, 30356, 30355, 30360, 30352, 30351, 30356, 30363, 30364, 30366, 30365, 30360, 30351, 30352, 30350, 30366, 30364, 30363, 30350, 30352, 30354, 30354, 30365, 30363, 30350, 30353, 30363, 30365, 30354, 30352, 30370, 30372, 30371, 30368, 30372, 30370, 30371, 30372, 30375, 30373, 30375, 30372, 30379, 30381, 30380, 30377, 30381, 30379, 30385, 30383, 30382, 30394, 30396, 30395, 30398, 30396, 30394, 30400, 30402, 30401, 30404, 30406, 30405, 30380, 30392, 30391, 30395, 30385, 30384, 30400, 30399, 30387, 30380, 30381, 30388, 30386, 30388, 30381, 30390, 30389, 30383, 30396, 30404, 30403, 30406, 30404, 30396, 30390, 30388, 30386, 30405, 30402, 30400, 30390, 30403, 30400, 30387, 30399, 30392, 30395, 30403, 30390, 30358, 30377, 30376, 30383, 30362, 30361, 30389, 30386, 30364, 30358, 30364, 30386, 30389, 30366, 30362, 30391, 30392, 30368, 30401, 30375, 30373, 30368, 30392, 30399, 30408, 30410, 30409, 30414, 30416, 30415, 30410, 30414, 30413, 30411, 30412, 30410, 30414, 30417, 30418, 30410, 30412, 30417, 30422, 30420, 30419, 30426, 30425, 30423, 30430, 30428, 30427, 30434, 30433, 30431, 30428, 30422, 30421, 30432, 30431, 30425, 30426, 30424, 30420, 30434, 30432, 30428, 30432, 30426, 30422, 30436, 30438, 30437,
30444, 30446, 30445, 30442, 30446, 30444, 30438, 30442, 30441, 30436, 30439, 30440, 30445, 30446, 30449, 30442, 30447, 30449, 30438, 30440, 30447, 30453, 30451, 30450, 30455, 30457, 30456, 30461, 30459, 30458, 30465, 30464, 30462, 30459, 30453, 30452, 30463, 30462, 30456, 30453, 30457, 30455, 30461, 30465, 30463, 30459, 30463, 30457, 30416, 30436, 30435, 30416, 30418, 30439, 30451, 30430, 30429, 30455, 30454, 30433, 30455, 30434, 30430, 30425, 30412, 30411, 30433, 30418, 30417, 30425, 30431, 30417, 30456, 30440, 30439, 30464, 30449, 30447, 30462, 30447, 30440, 30454, 30439, 30418, 30469, 30472, 30470, 30476, 30478, 30477, 30484, 30482, 30481, 30467, 30480, 30479, 30484, 30473, 30471, 30478, 30487, 30486, 30475, 30477, 30472, 30486, 30487, 30484, 30470, 30472, 30467, 30483, 30481, 30479, 30467, 30472, 30477, 30485, 30486, 30483, 30486, 30485, 30474, 30474, 30485, 30480, 30484, 30487, 30478, 30491, 30489, 30488, 30494, 30490, 30488, 30491, 30495, 30493, 30507, 30506, 30504, 30511, 30509, 30508, 30499, 30510, 30508, 30507, 30513, 30512, 30501, 30491, 30490, 30494, 30502, 30500, 30491, 30501, 30503, 30513, 30511, 30510, 30499, 30498, 30493, 30504, 30506, 30503, 30506, 30499, 30495, 30506, 30512, 30510, 30494, 30492, 30471, 30502, 30478, 30476,
30502, 30494, 30473, 30515, 30517, 30516, 30521, 30523, 30522, 30525, 30527, 30526, 30531, 30529, 30528, 30535, 30533, 30532, 30537, 30536, 30522, 30539, 30538, 30526, 30517, 30521, 30520, 30537, 30531, 30530, 30538, 30539, 30535, 30518, 30519, 30517, 30526, 30523, 30521, 30534, 30532, 30529, 30517, 30519, 30524, 30537, 30538, 30534, 30526, 30538, 30537, 30543, 30541, 30540, 30545, 30542, 30540, 30550, 30545, 30544, 30546, 30547, 30543, 30548, 30546, 30542, 30551, 30548, 30545, 30548, 30527, 30525, 30550, 30549, 30533, 30539, 30551, 30550, 30551, 30539, 30527, 30371, 30469, 30468, 30476, 30475, 30374, 30371, 30374, 30475, 30497, 30398, 30397, 30501, 30500, 30401, 30406, 30505, 30504, 30405, 30504, 30501, 30505, 30406, 30398, 30401, 30500, 30476, 30445, 30519, 30518, 30525, 30524, 30448, 30448, 30524, 30519, 30543, 30461, 30460, 30547, 30546, 30464, 30547, 30465, 30461, 30464, 30546, 30525, 30514, 30516, 30497, 30520, 30522, 30507, 30530, 30528, 30509, 30516, 30520, 30505, 30536, 30530, 30511, 30522, 30536, 30513, 30553, 30555, 30554, 30554, 30557, 30556, 30561, 30560, 30558, 30563, 30565, 30564, 30569, 30568, 30566, 30557, 30563, 30562, 30567, 30566, 30560, 30555, 30559, 30558, 30558, 30560, 30557, 30566, 30568, 30565, 30560, 30566, 30563,
30571, 30573, 30572, 30577, 30576, 30574, 30579, 30581, 30580, 30585, 30584, 30582, 30573, 30579, 30578, 30583, 30582, 30576, 30574, 30576, 30573, 30582, 30584, 30581, 30576, 30582, 30579, 30565, 30571, 30570, 30575, 30574, 30568, 30568, 30574, 30571, 30587, 30589, 30588, 30591, 30593, 30592, 30597, 30596, 30594, 30601, 30600, 30598, 30605, 30604, 30602, 30609, 30608, 30606, 30603, 30602, 30596, 30607, 30606, 30600, 30589, 30595, 30594, 30593, 30599, 30598, 30590, 30592, 30589, 30600, 30597, 30595, 30608, 30605, 30603, 30592, 30598, 30595, 30606, 30603, 30597, 30611, 30613, 30612, 30615, 30617, 30616, 30621, 30620, 30618, 30625, 30624, 30622, 30629, 30628, 30626, 30633, 30632, 30630, 30627, 30626, 30620, 30631, 30630, 30624, 30613, 30619, 30618, 30617, 30623, 30622, 30614, 30616, 30613, 30624, 30621, 30619, 30632, 30629, 30627, 30616, 30622, 30619, 30630, 30627, 30621, 30637, 30636, 30634, 30641, 30640, 30638, 30645, 30644, 30642, 30649, 30648, 30646, 30643, 30642, 30636, 30647, 30646, 30640, 30640, 30637, 30635, 30648, 30645, 30643, 30646, 30643, 30637, 30653, 30651, 30650, 30657, 30655, 30654, 30661, 30660, 30658, 30665, 30664, 30662, 30669, 30667, 30666, 30667, 30652, 30650, 30661, 30653, 30652, 30671, 30670, 30667, 30665, 30657, 30656,
30656, 30654, 30651, 30664, 30661, 30659, 30664, 30656, 30653, 30670, 30660, 30652, 30635, 30634, 30604, 30639, 30638, 30608, 30638, 30635, 30605, 30659, 30658, 30628, 30663, 30662, 30632, 30662, 30659, 30629, 30610, 30612, 30593, 30620, 30601, 30599, 30628, 30609, 30607, 30612, 30618, 30599, 30626, 30607, 30601, 30660, 30641, 30639, 30671, 30649, 30647, 30670, 30647, 30641, 30658, 30639, 30609, 30673, 30675, 30674, 30679, 30678, 30676, 30681, 30683, 30682, 30687, 30686, 30684, 30675, 30681, 30680, 30685, 30684, 30678, 30676, 30678, 30675, 30684, 30686, 30683, 30678, 30684, 30681, 30689, 30691, 30690, 30695, 30694, 30692, 30697, 30699, 30698, 30703, 30702, 30700, 30691, 30697, 30696, 30701, 30700, 30694, 30692, 30694, 30691, 30700, 30702, 30699, 30694, 30700, 30697, 30683, 30689, 30688, 30693, 30692, 30686, 30686, 30692, 30689, 30707, 30706, 30704, 30711, 30710, 30708, 30715, 30714, 30712, 30719, 30718, 30716, 30713, 30712, 30706, 30717, 30716, 30710, 30710, 30707, 30705, 30718, 30715, 30713, 30716, 30713, 30707, 30723, 30721, 30720, 30729, 30727, 30726, 30727, 30723, 30722, 30725, 30724, 30721, 30731, 30730, 30727, 30730, 30725, 30723, 30735, 30734, 30732, 30739, 30738, 30736, 30743, 30742, 30740, 30747, 30746, 30744, 30741, 30740, 30734,
30745, 30744, 30738, 30738, 30735, 30733, 30746, 30743, 30741, 30744, 30741, 30735, 30751, 30749, 30748, 30757, 30755, 30754, 30755, 30751, 30750, 30753, 30752, 30749, 30759, 30758, 30755, 30758, 30753, 30751, 30733, 30732, 30714, 30737, 30736, 30718, 30736, 30733, 30715, 30749, 30729, 30728, 30752, 30731, 30729, 30725, 30711, 30709, 30731, 30719, 30717, 30730, 30717, 30711, 30753, 30739, 30737, 30759, 30747, 30745, 30758, 30745, 30739, 30752, 30737, 30719, 30581, 30673, 30672, 30677, 30676, 30584, 30584, 30676, 30673, 30705, 30704, 30644, 30709, 30708, 30648, 30708, 30705, 30645, 30721, 30669, 30668, 30724, 30671, 30669, 30724, 30709, 30649, 30586, 30588, 30555, 30596, 30561, 30559, 30604, 30569, 30567, 30588, 30594, 30559, 30602, 30567, 30561, 30636, 30577, 30575, 30644, 30585, 30583, 30642, 30583, 30577, 30634, 30575, 30569, 30706, 30679, 30677, 30714, 30687, 30685, 30712, 30685, 30679, 30734, 30695, 30693, 30742, 30703, 30701, 30740, 30701, 30695, 30732, 30693, 30687, 30704, 30677, 30585, 30761, 30763, 30762, 30767, 30766, 30764, 30769, 30771, 30770, 30775, 30774, 30772, 30763, 30769, 30768, 30773, 30772, 30766, 30764, 30766, 30763, 30772, 30774, 30771, 30766, 30772, 30769, 30777, 30779, 30778, 30783, 30782, 30780, 30785, 30787, 30786,
30791, 30790, 30788, 30779, 30785, 30784, 30789, 30788, 30782, 30780, 30782, 30779, 30788, 30790, 30787, 30782, 30788, 30785, 30771, 30777, 30776, 30781, 30780, 30774, 30774, 30780, 30777, 30795, 30794, 30792, 30799, 30798, 30796, 30803, 30802, 30800, 30807, 30806, 30804, 30801, 30800, 30794, 30805, 30804, 30798, 30798, 30795, 30793, 30806, 30803, 30801, 30804, 30801, 30795, 30811, 30809, 30808, 30817, 30815, 30814, 30815, 30811, 30810, 30813, 30812, 30809, 30819, 30818, 30815, 30818, 30813, 30811, 30823, 30822, 30820, 30827, 30826, 30824, 30831, 30830, 30828, 30835, 30834, 30832, 30829, 30828, 30822, 30833, 30832, 30826, 30826, 30823, 30821, 30834, 30831, 30829, 30832, 30829, 30823, 30839, 30837, 30836, 30845, 30843, 30842, 30843, 30839, 30838, 30841, 30840, 30837, 30847, 30846, 30843, 30846, 30841, 30839, 30821, 30820, 30802, 30825, 30824, 30806, 30824, 30821, 30803, 30837, 30817, 30816, 30840, 30819, 30817, 30813, 30799, 30797, 30819, 30807, 30805, 30818, 30805, 30799, 30841, 30827, 30825, 30847, 30835, 30833, 30846, 30833, 30827, 30840, 30825, 30807, 30849, 30851, 30850, 30855, 30854, 30852, 30857, 30859, 30858, 30863, 30862, 30860, 30851, 30857, 30856, 30861, 30860, 30854, 30852, 30854, 30851, 30860, 30862, 30859, 30854, 30860, 30857,
30867, 30865, 30864, 30869, 30867, 30866, 30859, 30866, 30864, 30869, 30868, 30862, 30862, 30868, 30866, 30873, 30872, 30870, 30877, 30876, 30874, 30881, 30880, 30878, 30885, 30884, 30882, 30879, 30878, 30872, 30883, 30882, 30876, 30876, 30873, 30871, 30884, 30881, 30879, 30882, 30879, 30873, 30889, 30887, 30886, 30895, 30893, 30892, 30893, 30889, 30888, 30891, 30890, 30887, 30897, 30896, 30893, 30896, 30891, 30889, 30901, 30899, 30898, 30905, 30903, 30902, 30907, 30901, 30900, 30909, 30905, 30904, 30904, 30902, 30899, 30908, 30904, 30901, 30907, 30906, 30880, 30909, 30908, 30884, 30908, 30907, 30881, 30911, 30895, 30894, 30912, 30897, 30895, 30891, 30877, 30875, 30897, 30885, 30883, 30896, 30883, 30877, 30911, 30910, 30903, 30912, 30911, 30905, 30912, 30909, 30885, 30787, 30849, 30848, 30853, 30852, 30790, 30790, 30852, 30849, 30871, 30870, 30830, 30875, 30874, 30834, 30874, 30871, 30831, 30887, 30845, 30844, 30890, 30847, 30845, 30890, 30875, 30835, 30794, 30767, 30765, 30802, 30775, 30773, 30800, 30773, 30767, 30822, 30783, 30781, 30830, 30791, 30789, 30828, 30789, 30783, 30820, 30781, 30775, 30872, 30855, 30853, 30880, 30863, 30861, 30878, 30861, 30855, 30900, 30898, 30865, 30906, 30900, 30867, 30906, 30869, 30863, 30870, 30853, 30791,
30699, 30761, 30760, 30765, 30764, 30702, 30702, 30764, 30761, 30793, 30792, 30742, 30797, 30796, 30746, 30796, 30793, 30743, 30809, 30757, 30756, 30812, 30759, 30757, 30812, 30797, 30747, 30792, 30765, 30703, 30914, 30916, 30915, 30918, 30920, 30919, 30926, 30925, 30923, 30928, 30930, 30929, 30934, 30932, 30931, 30940, 30939, 30937, 30916, 30928, 30927, 30926, 30938, 30937, 30921, 30922, 30916, 30935, 30936, 30930, 30920, 30924, 30923, 30940, 30938, 30932, 30917, 30919, 30914, 30925, 30922, 30921, 30939, 30936, 30935, 30919, 30923, 30921, 30937, 30935, 30922, 30922, 30935, 30928, 30942, 30944, 30943, 30946, 30948, 30947, 30952, 30951, 30949, 30956, 30955, 30953, 30960, 30958, 30957, 30964, 30962, 30961, 30966, 30965, 30951, 30968, 30967, 30955, 30944, 30950, 30949, 30966, 30960, 30959, 30948, 30954, 30953, 30968, 30964, 30963, 30945, 30947, 30944, 30955, 30952, 30950, 30963, 30961, 30958, 30947, 30953, 30950, 30967, 30963, 30960, 30967, 30966, 30952, 30970, 30972, 30971, 30976, 30974, 30973, 30982, 30981, 30979, 30984, 30986, 30985, 30990, 30988, 30987, 30994, 30996, 30995, 30972, 30984, 30983, 30988, 30976, 30975, 30994, 30993, 30981, 30977, 30978, 30972, 30991, 30992, 30986, 30982, 30980, 30974, 30996, 30994, 30988, 30979, 30981, 30978,
30995, 30992, 30991, 30981, 30993, 30991, 30978, 30991, 30984, 30994, 30982, 30976, 30998, 31000, 30999, 30999, 31002, 31001, 31006, 31005, 31003, 31008, 31010, 31009, 31014, 31013, 31011, 31002, 31008, 31007, 31012, 31011, 31005, 31000, 31004, 31003, 31003, 31005, 31002, 31008, 31011, 31013, 31005, 31011, 31008, 30930, 30970, 30969, 30974, 30934, 30933, 30980, 30979, 30939, 30979, 30977, 30936, 30936, 30977, 30970, 30980, 30940, 30934, 30941, 30943, 30920, 30951, 30926, 30924, 30959, 30957, 30931, 30943, 30949, 30924, 30965, 30959, 30932, 30951, 30965, 30938, 31016, 31018, 31017, 31020, 31022, 31021, 31026, 31025, 31023, 31030, 31029, 31027, 31034, 31032, 31031, 31038, 31036, 31035, 31038, 31040, 31039, 31042, 31041, 31025, 31030, 31044, 31043, 31018, 31024, 31023, 31042, 31034, 31033, 31022, 31028, 31027, 31044, 31038, 31037, 31044, 31045, 31040, 31019, 31021, 31018, 31029, 31026, 31024, 31037, 31035, 31032, 31021, 31027, 31024, 31043, 31037, 31034, 31043, 31042, 31026, 31047, 31049, 31048, 31051, 31049, 31047, 31055, 31054, 31052, 31059, 31057, 31056, 31063, 31062, 31060, 31057, 31051, 31050, 31061, 31060, 31054, 31049, 31053, 31052, 31055, 31053, 31049, 31063, 31061, 31057, 31061, 31055, 31051, 31067, 31065, 31064, 31069, 31071, 31070,
31079, 31077, 31076, 31081, 31083, 31082, 31077, 31067, 31066, 31071, 31081, 31080, 31073, 31072, 31065, 31085, 31084, 31077, 31074, 31075, 31071, 31081, 31086, 31087, 31084, 31073, 31067, 31071, 31075, 31086, 31091, 31089, 31088, 31095, 31094, 31092, 31099, 31097, 31096, 31101, 31103, 31102, 31097, 31091, 31090, 31101, 31100, 31094, 31095, 31093, 31089, 31099, 31103, 31101, 31101, 31095, 31091, 31040, 31069, 31068, 31045, 31074, 31069, 31089, 31059, 31058, 31093, 31092, 31062, 31093, 31063, 31059, 31046, 31048, 31022, 31054, 31030, 31028, 31060, 31062, 31045, 31048, 31052, 31028, 31054, 31060, 31044, 31094, 31075, 31074, 31100, 31102, 31087, 31100, 31086, 31075, 31092, 31074, 31045, 31105, 31107, 31106, 31113, 31109, 31108, 31119, 31118, 31116, 31122, 31125, 31124, 31129, 31128, 31126, 31112, 31122, 31120, 31123, 31113, 31111, 31119, 31127, 31126, 31105, 31114, 31115, 31119, 31117, 31109, 31126, 31128, 31125, 31112, 31110, 31106, 31116, 31118, 31115, 31118, 31112, 31107, 31112, 31118, 31126, 31123, 31127, 31119, 31131, 31133, 31132, 31137, 31136, 31134, 31141, 31143, 31142, 31142, 31139, 31138, 31145, 31147, 31146, 31154, 31153, 31151, 31133, 31145, 31144, 31152, 31151, 31136, 31143, 31150, 31149, 31142, 31149, 31148, 31131, 31134, 31136,
31145, 31151, 31153, 31146, 31147, 31143, 31143, 31147, 31153, 31133, 31136, 31151, 31158, 31156, 31155, 31160, 31158, 31157, 31166, 31164, 31163, 31162, 31165, 31163, 31170, 31168, 31167, 31172, 31174, 31173, 31176, 31178, 31177, 31173, 31174, 31166, 31171, 31173, 31165, 31178, 31170, 31169, 31169, 31167, 31164, 31177, 31174, 31172, 31174, 31177, 31169, 31125, 31157, 31155, 31129, 31160, 31159, 31128, 31159, 31157, 31172, 31171, 31149, 31154, 31176, 31175, 31175, 31172, 31150, 31138, 31139, 31123, 31148, 31149, 31129, 31139, 31148, 31127, 31162, 31161, 31156, 31160, 31171, 31162, 31171, 31160, 31129, 31182, 31180, 31179, 31184, 31186, 31185, 31192, 31194, 31193, 31196, 31194, 31192, 31198, 31200, 31199, 31199, 31202, 31201, 31204, 31206, 31205, 31210, 31209, 31207, 31193, 31182, 31181, 31186, 31200, 31198, 31188, 31187, 31180, 31194, 31204, 31203, 31196, 31206, 31204, 31184, 31189, 31190, 31199, 31200, 31208, 31207, 31209, 31202, 31201, 31202, 31196, 31196, 31202, 31209, 31193, 31203, 31188, 31190, 31208, 31200, 31214, 31212, 31211, 31218, 31217, 31215, 31222, 31220, 31219, 31224, 31221, 31219, 31222, 31214, 31213, 31226, 31225, 31217, 31214, 31218, 31216, 31225, 31226, 31222, 31227, 31225, 31221, 31226, 31218, 31214, 31231, 31229, 31228,
31235, 31233, 31232, 31237, 31234, 31232, 31239, 31241, 31240, 31241, 31231, 31230, 31242, 31243, 31235, 31237, 31244, 31242, 31237, 31236, 31229, 31244, 31241, 31239, 31244, 31237, 31231, 31239, 31238, 31205, 31210, 31243, 31242, 31209, 31242, 31239, 31246, 31224, 31223, 31246, 31247, 31227, 31217, 31190, 31189, 31225, 31227, 31210, 31217, 31225, 31208, 31246, 31245, 31233, 31243, 31247, 31246, 31247, 31243, 31210, 30986, 31105, 31104, 31109, 30990, 30989, 31117, 31116, 30995, 30995, 31116, 31114, 30992, 31114, 31105, 31117, 30996, 30990, 31010, 31131, 31130, 31135, 31134, 31013, 31010, 31013, 31134, 31180, 31079, 31078, 31083, 31184, 31183, 31187, 31085, 31079, 31083, 31087, 31189, 31212, 31099, 31098, 31103, 31216, 31215, 31212, 31216, 31103, 31215, 31189, 31087, 31015, 31017, 30948, 31025, 30956, 30954, 31033, 31031, 30962, 31017, 31023, 30954, 31041, 31033, 30964, 31041, 30968, 30956, 31064, 31065, 31000, 31072, 31073, 31006, 31085, 31014, 31012, 31065, 31072, 31004, 31073, 31084, 31012, 31187, 31188, 31137, 31205, 31154, 31152, 31188, 31203, 31152, 31230, 31228, 31168, 31240, 31178, 31176, 31240, 31230, 31170, 31238, 31176, 31154, 31085, 31187, 31135, 31257, 31259, 31258, 31255, 31259, 31257, 31261, 31263, 31262, 31270, 31269, 31267,
31252, 31263, 31261, 31258, 31259, 31266, 31255, 31264, 31266, 31263, 31268, 31267, 31260, 31262, 31255, 31269, 31266, 31264, 31255, 31262, 31267, 31252, 31253, 31268, 31274, 31276, 31275, 31275, 31278, 31277, 31283, 31282, 31280, 31287, 31289, 31288, 31285, 31289, 31287, 31296, 31295, 31293, 31272, 31285, 31284, 31283, 31294, 31293, 31288, 31289, 31292, 31285, 31290, 31292, 31276, 31281, 31280, 31280, 31282, 31278, 31277, 31278, 31272, 31295, 31292, 31290, 31272, 31278, 31282, 31293, 31290, 31279, 31272, 31279, 31290, 31300, 31299, 31297, 31304, 31303, 31301, 31308, 31307, 31305, 31312, 31311, 31309, 31306, 31305, 31299, 31310, 31309, 31303, 31303, 31300, 31298, 31311, 31308, 31306, 31309, 31306, 31300, 31249, 31276, 31274, 31249, 31250, 31281, 31298, 31297, 31265, 31302, 31301, 31269, 31269, 31301, 31298, 31256, 31258, 31249, 31249, 31258, 31265, 31297, 31299, 31283, 31307, 31296, 31294, 31305, 31294, 31283, 31297, 31281, 31250, 31317, 31319, 31318, 31314, 31317, 31316, 31317, 31320, 31321, 31314, 31315, 31320, 31323, 31325, 31324, 31331, 31333, 31332, 31329, 31333, 31331, 31325, 31329, 31328, 31323, 31326, 31327, 31332, 31333, 31336, 31334, 31336, 31333, 31325, 31327, 31334, 31319, 31323, 31322, 31319, 31321, 31326, 31338, 31340, 31339,
31344, 31343, 31341, 31348, 31347, 31345, 31352, 31351, 31349, 31356, 31355, 31353, 31344, 31350, 31349, 31354, 31353, 31347, 31338, 31341, 31343, 31347, 31344, 31342, 31355, 31352, 31350, 31353, 31350, 31344, 31362, 31360, 31359, 31362, 31358, 31357, 31366, 31365, 31363, 31370, 31369, 31367, 31374, 31372, 31371, 31380, 31379, 31377, 31376, 31361, 31359, 31366, 31378, 31377, 31374, 31380, 31378, 31370, 31362, 31361, 31370, 31368, 31358, 31367, 31369, 31366, 31376, 31375, 31371, 31378, 31381, 31376, 31381, 31378, 31366, 31381, 31369, 31361, 31385, 31384, 31382, 31389, 31388, 31386, 31393, 31392, 31390, 31397, 31396, 31394, 31391, 31390, 31384, 31395, 31394, 31388, 31388, 31385, 31383, 31396, 31393, 31391, 31394, 31391, 31385, 31401, 31399, 31398, 31403, 31400, 31398, 31410, 31408, 31407, 31408, 31403, 31402, 31404, 31405, 31401, 31406, 31404, 31400, 31410, 31412, 31411, 31408, 31411, 31406, 31383, 31382, 31351, 31387, 31386, 31355, 31386, 31383, 31352, 31401, 31374, 31373, 31405, 31404, 31379, 31405, 31380, 31374, 31365, 31348, 31346, 31379, 31356, 31354, 31377, 31354, 31348, 31406, 31389, 31387, 31412, 31397, 31395, 31411, 31395, 31389, 31379, 31404, 31387, 31288, 31338, 31337, 31342, 31341, 31291, 31346, 31345, 31295, 31295, 31345, 31342,
31288, 31291, 31341, 31364, 31363, 31307, 31312, 31368, 31367, 31367, 31364, 31308, 31363, 31346, 31296, 31339, 31340, 31314, 31351, 31321, 31320, 31314, 31340, 31343, 31349, 31320, 31315, 31384, 31327, 31326, 31390, 31392, 31336, 31390, 31334, 31327, 31382, 31326, 31321, 31414, 31416, 31415, 31418, 31420, 31419, 31426, 31428, 31427, 31423, 31424, 31416, 31420, 31426, 31425, 31417, 31419, 31414, 31427, 31424, 31423, 31419, 31425, 31423, 31430, 31432, 31431, 31436, 31434, 31433, 31438, 31440, 31439, 31444, 31442, 31441, 31448, 31447, 31445, 31452, 31451, 31449, 31432, 31446, 31445, 31448, 31436, 31435, 31440, 31450, 31449, 31452, 31444, 31443, 31437, 31439, 31432, 31443, 31441, 31434, 31451, 31448, 31446, 31439, 31449, 31446, 31451, 31443, 31436, 31429, 31431, 31420, 31435, 31433, 31421, 31445, 31447, 31428, 31431, 31445, 31426, 31447, 31435, 31422, 31454, 31456, 31455, 31460, 31458, 31457, 31462, 31464, 31463, 31468, 31466, 31465, 31472, 31471, 31469, 31476, 31475, 31473, 31456, 31470, 31469, 31472, 31460, 31459, 31464, 31474, 31473, 31476, 31468, 31467, 31461, 31463, 31456, 31467, 31465, 31458, 31475, 31472, 31470, 31463, 31473, 31470, 31475, 31467, 31460, 31478, 31480, 31479, 31484, 31482, 31481, 31484, 31480, 31478, 31488, 31487, 31485,
31480, 31486, 31485, 31488, 31484, 31483, 31488, 31486, 31480, 31477, 31479, 31464, 31483, 31481, 31466, 31487, 31476, 31474, 31479, 31485, 31474, 31487, 31483, 31468, 31490, 31492, 31491, 31496, 31494, 31493, 31502, 31501, 31499, 31504, 31506, 31505, 31510, 31508, 31507, 31516, 31515, 31513, 31492, 31504, 31503, 31508, 31496, 31495, 31514, 31513, 31501, 31497, 31498, 31492, 31511, 31512, 31506, 31502, 31500, 31494, 31516, 31514, 31508, 31501, 31498, 31497, 31515, 31512, 31511, 31513, 31511, 31498, 31498, 31511, 31504, 31514, 31502, 31496, 31518, 31520, 31519, 31524, 31522, 31521, 31530, 31529, 31527, 31532, 31534, 31533, 31540, 31538, 31537, 31540, 31536, 31535, 31546, 31545, 31543, 31520, 31532, 31531, 31536, 31524, 31523, 31544, 31543, 31529, 31525, 31526, 31520, 31541, 31542, 31534, 31530, 31528, 31522, 31546, 31540, 31539, 31546, 31544, 31536, 31529, 31526, 31525, 31539, 31537, 31533, 31545, 31542, 31541, 31545, 31539, 31534, 31543, 31541, 31526, 31526, 31541, 31532, 31544, 31530, 31524, 31506, 31518, 31517, 31522, 31510, 31509, 31528, 31527, 31515, 31527, 31525, 31512, 31512, 31525, 31518, 31528, 31516, 31510, 31416, 31490, 31489, 31494, 31422, 31421, 31428, 31500, 31499, 31499, 31497, 31424, 31424, 31497, 31490, 31500, 31428, 31422,
31453, 31455, 31440, 31459, 31457, 31442, 31471, 31452, 31450, 31455, 31469, 31450, 31471, 31459, 31444, 31548, 31550, 31549, 31554, 31553, 31551, 31556, 31558, 31557, 31560, 31562, 31561, 31550, 31556, 31555, 31560, 31559, 31553, 31548, 31551, 31553, 31556, 31559, 31561, 31550, 31553, 31559, 31564, 31566, 31565, 31568, 31570, 31569, 31576, 31574, 31573, 31572, 31575, 31573, 31578, 31580, 31579, 31566, 31572, 31571, 31570, 31578, 31577, 31567, 31569, 31566, 31580, 31576, 31575, 31572, 31577, 31579, 31569, 31577, 31572, 31558, 31564, 31563, 31562, 31568, 31567, 31561, 31567, 31564, 31584, 31583, 31581, 31588, 31587, 31585, 31592, 31591, 31589, 31596, 31595, 31593, 31590, 31589, 31583, 31594, 31593, 31587, 31587, 31584, 31582, 31595, 31592, 31590, 31593, 31590, 31584, 31600, 31598, 31597, 31606, 31604, 31603, 31604, 31600, 31599, 31600, 31602, 31601, 31606, 31608, 31607, 31604, 31607, 31602, 31612, 31611, 31609, 31616, 31615, 31613, 31620, 31619, 31617, 31624, 31623, 31621, 31618, 31617, 31611, 31622, 31621, 31615, 31615, 31612, 31610, 31623, 31620, 31618, 31621, 31618, 31612, 31628, 31626, 31625, 31634, 31632, 31631, 31632, 31628, 31627, 31630, 31629, 31626, 31636, 31635, 31632, 31635, 31630, 31628, 31610, 31609, 31591, 31614, 31613, 31595,
31613, 31610, 31592, 31626, 31606, 31605, 31629, 31608, 31606, 31602, 31588, 31586, 31608, 31596, 31594, 31607, 31594, 31588, 31629, 31630, 31616, 31635, 31636, 31624, 31630, 31635, 31622, 31608, 31629, 31614, 31638, 31640, 31639, 31644, 31646, 31645, 31640, 31644, 31643, 31641, 31642, 31640, 31647, 31648, 31646, 31642, 31647, 31644, 31646, 31650, 31649, 31646, 31648, 31651, 31655, 31654, 31652, 31659, 31658, 31656, 31663, 31662, 31660, 31667, 31666, 31664, 31661, 31660, 31654, 31665, 31664, 31658, 31658, 31655, 31653, 31666, 31663, 31661, 31664, 31661, 31655, 31671, 31673, 31672, 31672, 31669, 31668, 31680, 31678, 31677, 31684, 31686, 31685, 31678, 31673, 31671, 31676, 31684, 31683, 31673, 31676, 31675, 31675, 31674, 31669, 31686, 31684, 31678, 31681, 31682, 31680, 31680, 31682, 31687, 31684, 31676, 31673, 31689, 31691, 31690, 31693, 31695, 31694, 31699, 31698, 31696, 31705, 31703, 31702, 31701, 31704, 31702, 31707, 31709, 31708, 31713, 31712, 31710, 31691, 31701, 31700, 31695, 31707, 31706, 31711, 31710, 31698, 31689, 31692, 31694, 31708, 31709, 31705, 31706, 31708, 31704, 31698, 31695, 31693, 31712, 31709, 31707, 31710, 31707, 31695, 31694, 31706, 31701, 31717, 31719, 31718, 31718, 31715, 31714, 31723, 31722, 31720, 31732, 31731, 31729,
31736, 31735, 31733, 31728, 31719, 31717, 31730, 31729, 31722, 31726, 31734, 31733, 31718, 31719, 31726, 31725, 31724, 31715, 31724, 31725, 31723, 31735, 31732, 31730, 31733, 31730, 31723, 31734, 31726, 31719, 31693, 31692, 31662, 31697, 31696, 31666, 31696, 31693, 31663, 31715, 31682, 31681, 31721, 31720, 31685, 31687, 31724, 31721, 31724, 31687, 31682, 31674, 31675, 31659, 31685, 31667, 31665, 31683, 31665, 31659, 31722, 31699, 31697, 31731, 31713, 31711, 31729, 31711, 31699, 31720, 31697, 31667, 31576, 31638, 31637, 31576, 31580, 31641, 31653, 31652, 31619, 31657, 31656, 31623, 31656, 31653, 31620, 31669, 31634, 31633, 31674, 31636, 31634, 31636, 31674, 31657, 31583, 31554, 31552, 31591, 31562, 31560, 31589, 31560, 31554, 31611, 31570, 31568, 31619, 31580, 31578, 31617, 31578, 31570, 31609, 31568, 31562, 31652, 31654, 31642, 31660, 31662, 31648, 31654, 31660, 31647, 31689, 31688, 31649, 31692, 31689, 31650, 31662, 31692, 31651, 31619, 31652, 31641, 31332, 31548, 31547, 31552, 31551, 31335, 31335, 31551, 31548, 31582, 31581, 31392, 31586, 31585, 31396, 31585, 31582, 31393, 31598, 31410, 31409, 31598, 31601, 31412, 31601, 31586, 31397, 31392, 31581, 31552, 30326, 30408, 30407, 30334, 30411, 30408, 30420, 30344, 30343, 30424, 30423, 30347,
30424, 30348, 30344, 30423, 30411, 30334, 31738, 31740, 31739, 31742, 31744, 31743, 31750, 31749, 31747, 31752, 31754, 31753, 31758, 31756, 31755, 31764, 31763, 31761, 31740, 31752, 31751, 31762, 31761, 31749, 31745, 31746, 31740, 31759, 31760, 31754, 31744, 31748, 31747, 31764, 31762, 31756, 31741, 31743, 31738, 31749, 31746, 31745, 31763, 31760, 31759, 31743, 31747, 31745, 31761, 31759, 31746, 31746, 31759, 31752, 31766, 31768, 31767, 31770, 31772, 31771, 31776, 31775, 31773, 31780, 31779, 31777, 31784, 31782, 31781, 31788, 31786, 31785, 31790, 31792, 31791, 31794, 31793, 31775, 31796, 31795, 31779, 31768, 31774, 31773, 31794, 31784, 31783, 31772, 31778, 31777, 31796, 31788, 31787, 31792, 31798, 31797, 31769, 31771, 31768, 31779, 31776, 31774, 31787, 31785, 31782, 31771, 31777, 31774, 31795, 31787, 31784, 31795, 31794, 31776, 31800, 31802, 31801, 31806, 31804, 31803, 31812, 31811, 31809, 31814, 31816, 31815, 31820, 31818, 31817, 31826, 31825, 31823, 31802, 31814, 31813, 31818, 31806, 31805, 31824, 31823, 31811, 31807, 31808, 31802, 31821, 31822, 31816, 31812, 31810, 31804, 31826, 31824, 31818, 31811, 31808, 31807, 31825, 31822, 31821, 31823, 31821, 31808, 31808, 31821, 31814, 31824, 31812, 31806, 31828, 31830, 31829, 31834, 31833, 31831,
31836, 31838, 31837, 31842, 31841, 31839, 31830, 31836, 31835, 31840, 31839, 31833, 31831, 31833, 31830, 31839, 31841, 31838, 31833, 31839, 31836, 31754, 31800, 31799, 31804, 31758, 31757, 31810, 31809, 31763, 31809, 31807, 31760, 31760, 31807, 31800, 31810, 31764, 31758, 31791, 31828, 31827, 31832, 31831, 31797, 31797, 31831, 31828, 31765, 31767, 31744, 31775, 31750, 31748, 31783, 31781, 31755, 31767, 31773, 31748, 31793, 31783, 31756, 31775, 31793, 31762, 31844, 31846, 31845, 31848, 31850, 31849, 31854, 31853, 31851, 31858, 31857, 31855, 31862, 31860, 31859, 31868, 31866, 31865, 31868, 31870, 31869, 31872, 31871, 31853, 31858, 31875, 31874, 31846, 31852, 31851, 31872, 31862, 31861, 31850, 31856, 31855, 31875, 31868, 31867, 31875, 31876, 31870, 31847, 31849, 31846, 31857, 31854, 31852, 31867, 31865, 31860, 31849, 31855, 31852, 31874, 31867, 31862, 31874, 31872, 31854, 31878, 31880, 31879, 31882, 31880, 31878, 31886, 31885, 31883, 31890, 31888, 31887, 31894, 31893, 31891, 31888, 31882, 31881, 31892, 31891, 31885, 31880, 31884, 31883, 31886, 31884, 31880, 31894, 31892, 31888, 31892, 31886, 31882, 31898, 31896, 31895, 31900, 31902, 31901, 31910, 31908, 31907, 31912, 31914, 31913, 31908, 31898, 31897, 31902, 31912, 31911, 31904, 31903, 31896,
31916, 31915, 31908, 31905, 31906, 31902, 31917, 31918, 31914, 31915, 31904, 31898, 31906, 31917, 31912, 31922, 31920, 31919, 31926, 31925, 31923, 31930, 31928, 31927, 31934, 31933, 31931, 31928, 31922, 31921, 31932, 31931, 31925, 31926, 31924, 31920, 31934, 31932, 31928, 31932, 31926, 31922, 31896, 31864, 31863, 31870, 31900, 31899, 31903, 31873, 31864, 31876, 31905, 31900, 31920, 31890, 31889, 31924, 31923, 31893, 31924, 31894, 31890, 31877, 31879, 31850, 31885, 31858, 31856, 31891, 31893, 31876, 31879, 31883, 31856, 31885, 31891, 31875, 31925, 31906, 31905, 31933, 31918, 31917, 31931, 31917, 31906, 31923, 31905, 31876, 31843, 31845, 31772, 31853, 31780, 31778, 31861, 31859, 31786, 31863, 31864, 31792, 31845, 31851, 31778, 31871, 31861, 31788, 31864, 31873, 31798, 31871, 31796, 31780, 31904, 31834, 31832, 31916, 31842, 31840, 31915, 31840, 31834, 31903, 31832, 31798, 31942, 31940, 31939, 31938, 31941, 31939, 31945, 31947, 31946, 31942, 31949, 31948, 31946, 31947, 31942, 31938, 31944, 31946, 31938, 31937, 31935, 31944, 31938, 31936, 31947, 31950, 31949, 31954, 31953, 31951, 31958, 31957, 31955, 31964, 31962, 31961, 31960, 31963, 31961, 31968, 31970, 31969, 31974, 31973, 31971, 31968, 31967, 31953, 31972, 31971, 31957, 31970, 31964, 31963,
31967, 31969, 31963, 31957, 31954, 31952, 31966, 31965, 31962, 31971, 31973, 31970, 31973, 31966, 31964, 31971, 31968, 31954, 31978, 31976, 31975, 31979, 31980, 31978, 31966, 31977, 31975, 31974, 31980, 31979, 31973, 31979, 31977, 31953, 31947, 31945, 31960, 31959, 31948, 31950, 31967, 31960, 31953, 31967, 31950, 31982, 31984, 31983, 31988, 31990, 31989, 31994, 31993, 31991, 31983, 31988, 31987, 31992, 31991, 31985, 31984, 31986, 31985, 31991, 31993, 31990, 31985, 31991, 31988, 31996, 31998, 31997, 32000, 32002, 32001, 32010, 32009, 32007, 32014, 32013, 32011, 32008, 32007, 32003, 32012, 32011, 32005, 31998, 32004, 32003, 32002, 32006, 32005, 31999, 32001, 31998, 32013, 32010, 32008, 32001, 32005, 32004, 32011, 32008, 32004, 32016, 32018, 32017, 32020, 32022, 32021, 32030, 32029, 32027, 32034, 32033, 32031, 32028, 32027, 32023, 32032, 32031, 32025, 32018, 32024, 32023, 32022, 32026, 32025, 32019, 32021, 32018, 32033, 32030, 32028, 32021, 32025, 32024, 32031, 32028, 32024, 32015, 32017, 32002, 32029, 32014, 32012, 32017, 32023, 32006, 32027, 32012, 32006, 31995, 31997, 31984, 32009, 31994, 31992, 31997, 32003, 31986, 32007, 31992, 31986, 32036, 32038, 32037, 32042, 32040, 32039, 32048, 32047, 32045, 32054, 32052, 32051, 32038, 32050, 32049,
32054, 32042, 32041, 32057, 32056, 32047, 32043, 32044, 32038, 32048, 32046, 32040, 32057, 32054, 32053, 32047, 32044, 32043, 32053, 32051, 32049, 32056, 32053, 32050, 32056, 32055, 32044, 32044, 32055, 32050, 32057, 32048, 32042, 32063, 32061, 32060, 32059, 32062, 32060, 32067, 32066, 32064, 32067, 32063, 32062, 32064, 32066, 32062, 32071, 32069, 32068, 32073, 32075, 32074, 32075, 32081, 32080, 32077, 32076, 32069, 32078, 32079, 32075, 32079, 32082, 32081, 32086, 32084, 32083, 32090, 32089, 32087, 32094, 32092, 32091, 32094, 32086, 32085, 32096, 32095, 32089, 32090, 32088, 32084, 32096, 32094, 32093, 32096, 32090, 32086, 32089, 32079, 32078, 32093, 32091, 32080, 32095, 32093, 32081, 32095, 32082, 32079, 32100, 32102, 32101, 32102, 32105, 32104, 32099, 32101, 32098, 32101, 32104, 32103, 32107, 32109, 32108, 32111, 32113, 32112, 32109, 32115, 32114, 32113, 32117, 32116, 32110, 32112, 32109, 32112, 32116, 32115, 32119, 32121, 32120, 32125, 32123, 32122, 32131, 32130, 32128, 32126, 32127, 32121, 32131, 32125, 32124, 32124, 32122, 32120, 32130, 32127, 32126, 32130, 32124, 32121, 32135, 32133, 32132, 32139, 32137, 32136, 32143, 32142, 32140, 32147, 32146, 32144, 32143, 32135, 32134, 32147, 32139, 32138, 32138, 32136, 32133, 32146, 32143, 32141,
32146, 32138, 32135, 32098, 32119, 32118, 32129, 32128, 32104, 32128, 32126, 32103, 32103, 32126, 32119, 32141, 32140, 32114, 32145, 32144, 32116, 32144, 32141, 32115, 32106, 32108, 32102, 32108, 32114, 32105, 32134, 32132, 32123, 32142, 32131, 32129, 32142, 32134, 32125, 32140, 32129, 32105, 32152, 32154, 32153, 32156, 32158, 32157, 32154, 32160, 32159, 32158, 32162, 32161, 32155, 32157, 32154, 32157, 32161, 32160, 32164, 32166, 32165, 32166, 32168, 32167, 32172, 32170, 32169, 32174, 32176, 32175, 32180, 32179, 32177, 32182, 32184, 32183, 32183, 32186, 32185, 32192, 32190, 32189, 32192, 32188, 32187, 32196, 32195, 32193, 32200, 32199, 32197, 32172, 32184, 32182, 32180, 32198, 32197, 32175, 32176, 32172, 32183, 32184, 32194, 32193, 32195, 32186, 32199, 32200, 32192, 32200, 32198, 32188, 32177, 32179, 32176, 32197, 32199, 32196, 32197, 32194, 32176, 32176, 32194, 32184, 32204, 32202, 32201, 32206, 32203, 32201, 32211, 32206, 32205, 32207, 32208, 32204, 32206, 32209, 32207, 32211, 32212, 32209, 32174, 32173, 32159, 32178, 32177, 32161, 32177, 32174, 32160, 32204, 32166, 32164, 32208, 32207, 32167, 32208, 32168, 32166, 32163, 32165, 32158, 32165, 32167, 32162, 32209, 32180, 32178, 32211, 32210, 32187, 32198, 32212, 32211, 32212, 32198, 32180,
32167, 32207, 32178, 32071, 32070, 32061, 32077, 32067, 32065, 32077, 32071, 32063, 32151, 32153, 32113, 32153, 32159, 32117, 32171, 32169, 32137, 32175, 32147, 32145, 32185, 32186, 32149, 32175, 32171, 32139, 32149, 32186, 32195, 32173, 32145, 32117, 32214, 32216, 32215, 32220, 32219, 32217, 32224, 32222, 32221, 32216, 32223, 32221, 32226, 32225, 32219, 32217, 32219, 32216, 32226, 32224, 32223, 32219, 32225, 32223, 32230, 32229, 32227, 32234, 32233, 32231, 32238, 32236, 32235, 32242, 32240, 32239, 32244, 32243, 32229, 32246, 32245, 32233, 32244, 32238, 32237, 32246, 32242, 32241, 32233, 32230, 32228, 32241, 32239, 32236, 32245, 32241, 32238, 32245, 32244, 32230, 32250, 32249, 32247, 32254, 32253, 32251, 32258, 32256, 32255, 32262, 32260, 32259, 32264, 32263, 32249, 32266, 32265, 32253, 32264, 32258, 32257, 32266, 32262, 32261, 32253, 32250, 32248, 32261, 32259, 32256, 32265, 32261, 32258, 32265, 32264, 32250, 32249, 32234, 32232, 32257, 32255, 32240, 32263, 32257, 32242, 32263, 32246, 32234, 32268, 32270, 32269, 32270, 32272, 32271, 32274, 32276, 32275, 32280, 32279, 32277, 32282, 32284, 32283, 32288, 32287, 32285, 32276, 32282, 32281, 32286, 32285, 32279, 32277, 32279, 32276, 32285, 32287, 32284, 32279, 32285, 32282, 32269, 32274, 32273,
32278, 32277, 32271, 32271, 32277, 32274, 32290, 32292, 32291, 32294, 32296, 32295, 32292, 32298, 32297, 32296, 32300, 32299, 32293, 32295, 32292, 32295, 32299, 32298, 32302, 32304, 32303, 32306, 32308, 32307, 32304, 32310, 32309, 32308, 32312, 32311, 32305, 32307, 32304, 32307, 32311, 32310, 32316, 32315, 32313, 32320, 32319, 32317, 32324, 32323, 32321, 32328, 32327, 32325, 32322, 32321, 32315, 32326, 32325, 32319, 32319, 32316, 32314, 32327, 32324, 32322, 32325, 32322, 32316, 32332, 32331, 32329, 32336, 32335, 32333, 32340, 32339, 32337, 32344, 32343, 32341, 32338, 32337, 32331, 32342, 32341, 32335, 32335, 32332, 32330, 32343, 32340, 32338, 32341, 32338, 32332, 32314, 32313, 32297, 32318, 32317, 32299, 32317, 32314, 32298, 32330, 32329, 32309, 32334, 32333, 32311, 32333, 32330, 32310, 32301, 32303, 32296, 32303, 32309, 32300, 32331, 32320, 32318, 32339, 32328, 32326, 32337, 32326, 32320, 32329, 32318, 32300, 32229, 32220, 32218, 32237, 32235, 32222, 32243, 32237, 32224, 32243, 32226, 32220, 32289, 32291, 32270, 32291, 32297, 32272, 32315, 32280, 32278, 32323, 32288, 32286, 32321, 32286, 32280, 32313, 32278, 32272, 31816, 32036, 32035, 32040, 31820, 31819, 32046, 32045, 31825, 32045, 32043, 31822, 31822, 32043, 32036, 32046, 31826, 31820,
31838, 32059, 32058, 32065, 32064, 31841, 31841, 32064, 32059, 32069, 31910, 31909, 31914, 32073, 32072, 32076, 31916, 31910, 31918, 32078, 32073, 32084, 31930, 31929, 32088, 32087, 31933, 32088, 31934, 31930, 32087, 32078, 31918, 32076, 32065, 31842, 31990, 32214, 32213, 32218, 32217, 31993, 31993, 32217, 32214, 32228, 32227, 32009, 32232, 32231, 32013, 32231, 32228, 32010, 32248, 32247, 32029, 32252, 32251, 32033, 32251, 32248, 32030, 32247, 32232, 32014, 32227, 32218, 31994, 32351, 32353, 32352, 32352, 32355, 32354, 32360, 32359, 32357, 32362, 32364, 32363, 32363, 32366, 32365, 32372, 32370, 32369, 32372, 32368, 32367, 32376, 32375, 32373, 32380, 32379, 32377, 32349, 32364, 32362, 32360, 32378, 32377, 32363, 32364, 32374, 32373, 32375, 32366, 32352, 32353, 32358, 32357, 32359, 32355, 32379, 32380, 32372, 32380, 32378, 32368, 32354, 32355, 32349, 32377, 32379, 32376, 32349, 32355, 32359, 32377, 32374, 32356, 32356, 32374, 32364, 32382, 32384, 32383, 32383, 32386, 32385, 32390, 32389, 32387, 32394, 32392, 32391, 32386, 32393, 32391, 32396, 32395, 32389, 32383, 32384, 32388, 32383, 32387, 32389, 32395, 32396, 32394, 32386, 32389, 32395, 32400, 32398, 32397, 32404, 32406, 32405, 32410, 32409, 32407, 32417, 32415, 32414, 32421, 32419, 32418,
32423, 32422, 32409, 32425, 32424, 32412, 32410, 32408, 32398, 32423, 32417, 32416, 32406, 32413, 32412, 32425, 32421, 32420, 32402, 32401, 32397, 32403, 32405, 32400, 32420, 32418, 32415, 32408, 32411, 32402, 32405, 32412, 32410, 32424, 32420, 32417, 32424, 32423, 32410, 32346, 32384, 32382, 32347, 32388, 32384, 32402, 32371, 32369, 32376, 32408, 32407, 32411, 32408, 32376, 32411, 32379, 32371, 32365, 32366, 32346, 32346, 32366, 32375, 32407, 32409, 32390, 32416, 32414, 32392, 32422, 32416, 32394, 32422, 32396, 32390, 32407, 32388, 32347, 32431, 32429, 32428, 32431, 32427, 32426, 32437, 32436, 32434, 32440, 32430, 32428, 32436, 32437, 32431, 32431, 32437, 32435, 32433, 32432, 32426, 32435, 32438, 32433, 32441, 32436, 32430, 32443, 32445, 32444, 32447, 32449, 32448, 32457, 32455, 32454, 32461, 32459, 32458, 32463, 32462, 32450, 32465, 32464, 32452, 32445, 32451, 32450, 32463, 32457, 32456, 32449, 32453, 32452, 32465, 32461, 32460, 32446, 32448, 32445, 32460, 32458, 32455, 32448, 32452, 32451, 32464, 32460, 32457, 32464, 32463, 32451, 32467, 32469, 32468, 32475, 32473, 32472, 32475, 32469, 32467, 32477, 32476, 32470, 32469, 32471, 32470, 32477, 32475, 32474, 32477, 32471, 32469, 32466, 32468, 32449, 32474, 32472, 32459, 32468, 32470, 32453,
32476, 32474, 32461, 32476, 32465, 32453, 32434, 32436, 32360, 32440, 32439, 32367, 32378, 32441, 32440, 32441, 32378, 32360, 32442, 32444, 32406, 32456, 32454, 32419, 32444, 32450, 32413, 32462, 32456, 32421, 32462, 32425, 32413, 32479, 32481, 32480, 32485, 32484, 32482, 32487, 32489, 32488, 32493, 32492, 32490, 32481, 32487, 32486, 32491, 32490, 32484, 32482, 32484, 32481, 32490, 32492, 32489, 32484, 32490, 32487, 32495, 32497, 32496, 32501, 32500, 32498, 32503, 32505, 32504, 32509, 32508, 32506, 32497, 32503, 32502, 32507, 32506, 32500, 32498, 32500, 32497, 32506, 32508, 32505, 32500, 32506, 32503, 32489, 32495, 32494, 32499, 32498, 32492, 32492, 32498, 32495, 32513, 32512, 32510, 32517, 32516, 32514, 32521, 32520, 32518, 32525, 32524, 32522, 32519, 32518, 32512, 32523, 32522, 32516, 32516, 32513, 32511, 32524, 32521, 32519, 32522, 32519, 32513, 32529, 32527, 32526, 32533, 32531, 32530, 32537, 32536, 32534, 32541, 32540, 32538, 32545, 32543, 32542, 32543, 32528, 32526, 32537, 32529, 32528, 32547, 32546, 32543, 32541, 32533, 32532, 32532, 32530, 32527, 32540, 32537, 32535, 32540, 32532, 32529, 32546, 32536, 32528, 32551, 32550, 32548, 32555, 32554, 32552, 32559, 32558, 32556, 32563, 32562, 32560, 32557, 32556, 32550, 32561, 32560, 32554,
32554, 32551, 32549, 32562, 32559, 32557, 32560, 32557, 32551, 32567, 32565, 32564, 32573, 32571, 32570, 32571, 32567, 32566, 32569, 32568, 32565, 32575, 32574, 32571, 32574, 32569, 32567, 32549, 32548, 32520, 32553, 32552, 32524, 32552, 32549, 32521, 32565, 32545, 32544, 32568, 32547, 32545, 32536, 32517, 32515, 32547, 32525, 32523, 32546, 32523, 32517, 32569, 32555, 32553, 32575, 32563, 32561, 32574, 32561, 32555, 32568, 32553, 32525, 32577, 32579, 32578, 32583, 32582, 32580, 32585, 32587, 32586, 32589, 32591, 32590, 32579, 32585, 32584, 32583, 32589, 32588, 32580, 32582, 32579, 32588, 32590, 32587, 32582, 32588, 32585, 32593, 32595, 32594, 32599, 32598, 32596, 32601, 32603, 32602, 32605, 32607, 32606, 32595, 32601, 32600, 32599, 32605, 32604, 32596, 32598, 32595, 32604, 32606, 32603, 32598, 32604, 32601, 32587, 32593, 32592, 32591, 32597, 32596, 32590, 32596, 32593, 32611, 32610, 32608, 32615, 32614, 32612, 32619, 32618, 32616, 32623, 32622, 32620, 32617, 32616, 32610, 32621, 32620, 32614, 32614, 32611, 32609, 32622, 32619, 32617, 32620, 32617, 32611, 32627, 32625, 32624, 32633, 32631, 32630, 32631, 32627, 32626, 32629, 32628, 32625, 32635, 32634, 32631, 32634, 32629, 32627, 32639, 32638, 32636, 32643, 32642, 32640, 32647, 32646, 32644,
32651, 32650, 32648, 32645, 32644, 32638, 32649, 32648, 32642, 32642, 32639, 32637, 32650, 32647, 32645, 32648, 32645, 32639, 32655, 32653, 32652, 32661, 32659, 32658, 32659, 32655, 32654, 32657, 32656, 32653, 32663, 32662, 32659, 32662, 32657, 32655, 32637, 32636, 32618, 32641, 32640, 32622, 32640, 32637, 32619, 32653, 32633, 32632, 32656, 32635, 32633, 32629, 32615, 32613, 32634, 32635, 32623, 32634, 32621, 32615, 32657, 32643, 32641, 32662, 32663, 32651, 32657, 32662, 32649, 32635, 32656, 32641, 32505, 32577, 32576, 32581, 32580, 32508, 32508, 32580, 32577, 32609, 32608, 32558, 32613, 32612, 32562, 32612, 32609, 32559, 32625, 32573, 32572, 32628, 32575, 32573, 32628, 32613, 32563, 32512, 32485, 32483, 32520, 32493, 32491, 32518, 32491, 32485, 32550, 32501, 32499, 32558, 32509, 32507, 32556, 32507, 32501, 32548, 32499, 32493, 32610, 32583, 32581, 32618, 32591, 32589, 32616, 32589, 32583, 32638, 32599, 32597, 32646, 32607, 32605, 32644, 32605, 32599, 32636, 32597, 32591, 32608, 32581, 32509, 32667, 32669, 32668, 32674, 32676, 32675, 32680, 32678, 32677, 32684, 32686, 32685, 32665, 32674, 32673, 32672, 32684, 32683, 32681, 32682, 32676, 32669, 32672, 32671, 32686, 32684, 32678, 32666, 32668, 32665, 32683, 32685, 32682, 32668, 32671, 32670,
32671, 32683, 32681, 32670, 32681, 32674, 32688, 32690, 32689, 32692, 32694, 32693, 32702, 32700, 32699, 32706, 32704, 32703, 32708, 32707, 32695, 32710, 32709, 32697, 32690, 32696, 32695, 32708, 32702, 32701, 32694, 32698, 32697, 32710, 32706, 32705, 32691, 32693, 32690, 32705, 32703, 32700, 32693, 32697, 32696, 32709, 32705, 32702, 32709, 32708, 32696, 32712, 32714, 32713, 32718, 32716, 32715, 32722, 32724, 32723, 32726, 32728, 32727, 32732, 32730, 32729, 32738, 32737, 32735, 32714, 32726, 32725, 32730, 32718, 32717, 32736, 32735, 32723, 32719, 32720, 32714, 32733, 32734, 32728, 32724, 32722, 32716, 32738, 32736, 32730, 32721, 32723, 32720, 32735, 32737, 32734, 32723, 32735, 32733, 32720, 32733, 32726, 32736, 32724, 32718, 32676, 32712, 32711, 32716, 32680, 32679, 32722, 32721, 32685, 32685, 32721, 32719, 32682, 32719, 32712, 32722, 32686, 32680, 32687, 32689, 32669, 32701, 32699, 32677, 32689, 32695, 32672, 32707, 32701, 32678, 32695, 32707, 32684, 32740, 32742, 32741, 32744, 32746, 32745, 32754, 32752, 32751, 32758, 32756, 32755, 32758, 32760, 32759, 32762, 32761, 32747, 32750, 32764, 32763, 32742, 32748, 32747, 32762, 32754, 32753, 32746, 32750, 32749, 32764, 32758, 32757, 32758, 32764, 32765, 32743, 32745, 32742, 32757, 32755, 32752,
32745, 32749, 32748, 32763, 32757, 32754, 32763, 32762, 32748, 32767, 32769, 32768, 32775, 32773, 32772, 32779, 32778, 32776, 32773, 32769, 32767, 32777, 32776, 32770, 32769, 32771, 32770, 32779, 32777, 32773, 32777, 32771, 32769, 32781, 32783, 32782, 32787, 32789, 32788, 32783, 32787, 32786, 32781, 32784, 32785, 32787, 32790, 32791, 32783, 32785, 32790, 32795, 32793, 32792, 32799, 32798, 32796, 32803, 32801, 32800, 32805, 32807, 32806, 32801, 32795, 32794, 32799, 32805, 32804, 32799, 32797, 32793, 32803, 32807, 32805, 32801, 32805, 32799, 32760, 32781, 32780, 32760, 32765, 32784, 32793, 32775, 32774, 32797, 32796, 32778, 32797, 32779, 32775, 32766, 32768, 32746, 32778, 32765, 32764, 32768, 32770, 32750, 32776, 32764, 32750, 32798, 32785, 32784, 32806, 32791, 32790, 32804, 32790, 32785, 32796, 32784, 32765, 32812, 32815, 32814, 32820, 32819, 32817, 32824, 32822, 32821, 32824, 32826, 32825, 32815, 32823, 32821, 32820, 32828, 32827, 32812, 32817, 32819, 32828, 32824, 32823, 32824, 32828, 32829, 32812, 32810, 32808, 32817, 32812, 32809, 32819, 32827, 32823, 32833, 32831, 32830, 32837, 32839, 32838, 32843, 32839, 32837, 32845, 32847, 32846, 32851, 32850, 32848, 32838, 32833, 32832, 32835, 32834, 32831, 32838, 32839, 32849, 32851, 32849, 32839,
32847, 32853, 32852, 32844, 32846, 32843, 32843, 32846, 32852, 32848, 32835, 32833, 32857, 32855, 32854, 32857, 32859, 32858, 32863, 32861, 32860, 32868, 32870, 32869, 32864, 32865, 32857, 32865, 32866, 32859, 32870, 32863, 32862, 32862, 32860, 32858, 32867, 32869, 32866, 32866, 32869, 32862, 32842, 32856, 32854, 32865, 32864, 32850, 32853, 32868, 32867, 32867, 32865, 32851, 32850, 32864, 32856, 32830, 32831, 32813, 32834, 32835, 32820, 32842, 32840, 32825, 32850, 32829, 32828, 32831, 32834, 32818, 32829, 32850, 32842, 32848, 32828, 32820, 32874, 32876, 32875, 32872, 32876, 32874, 32883, 32885, 32884, 32881, 32885, 32883, 32894, 32896, 32895, 32875, 32887, 32886, 32894, 32893, 32878, 32884, 32885, 32892, 32881, 32890, 32892, 32875, 32876, 32879, 32872, 32877, 32879, 32886, 32887, 32881, 32895, 32892, 32890, 32887, 32893, 32890, 32878, 32893, 32887, 32902, 32900, 32899, 32902, 32898, 32897, 32906, 32905, 32903, 32910, 32908, 32907, 32908, 32901, 32899, 32905, 32906, 32902, 32906, 32904, 32898, 32910, 32912, 32911, 32908, 32911, 32905, 32918, 32916, 32915, 32918, 32914, 32913, 32922, 32920, 32919, 32926, 32925, 32923, 32926, 32918, 32917, 32918, 32926, 32924, 32927, 32928, 32922, 32921, 32919, 32913, 32927, 32921, 32914, 32922, 32889, 32888,
32892, 32924, 32923, 32928, 32927, 32895, 32895, 32927, 32924, 32922, 32928, 32896, 32903, 32905, 32879, 32910, 32909, 32888, 32912, 32896, 32894, 32896, 32912, 32910, 32905, 32911, 32894, 32728, 32809, 32808, 32813, 32732, 32731, 32818, 32817, 32737, 32737, 32817, 32816, 32728, 32734, 32816, 32818, 32738, 32732, 32789, 32872, 32871, 32789, 32791, 32877, 32898, 32803, 32802, 32904, 32903, 32806, 32898, 32904, 32807, 32903, 32877, 32791, 32739, 32741, 32694, 32753, 32751, 32704, 32741, 32747, 32698, 32761, 32753, 32706, 32761, 32710, 32698, 32882, 32884, 32847, 32847, 32884, 32891, 32917, 32915, 32861, 32925, 32870, 32868, 32870, 32925, 32917, 32891, 32923, 32868, 32932, 32930, 32929, 32932, 32934, 32933, 32939, 32941, 32940, 32934, 32939, 32938, 32936, 32932, 32931, 32932, 32936, 32937, 32942, 32943, 32941, 32937, 32942, 32939, 32941, 32945, 32944, 32941, 32943, 32946, 32950, 32949, 32947, 32954, 32953, 32951, 32958, 32957, 32955, 32962, 32961, 32959, 32956, 32955, 32949, 32960, 32959, 32953, 32953, 32950, 32948, 32961, 32958, 32956, 32959, 32956, 32950, 32966, 32964, 32963, 32970, 32972, 32971, 32974, 32972, 32970, 32976, 32978, 32977, 32971, 32966, 32965, 32968, 32967, 32964, 32972, 32976, 32975, 32978, 32976, 32972, 32975, 32968, 32966,
32980, 32982, 32981, 32984, 32986, 32985, 32990, 32989, 32987, 32996, 32994, 32993, 32992, 32995, 32993, 32998, 33000, 32999, 33004, 33003, 33001, 32982, 32992, 32991, 32986, 32998, 32997, 33002, 33001, 32989, 32980, 32983, 32985, 32999, 33000, 32996, 32997, 32999, 32995, 32989, 32986, 32984, 33003, 33000, 32998, 33001, 32998, 32986, 32985, 32997, 32992, 33010, 33008, 33007, 33014, 33013, 33011, 33018, 33020, 33019, 33022, 33020, 33018, 33026, 33025, 33023, 33030, 33029, 33027, 33019, 33010, 33009, 33024, 33023, 33013, 33016, 33015, 33008, 33020, 33028, 33027, 33030, 33028, 33020, 33007, 33008, 33006, 33015, 33016, 33014, 33029, 33026, 33024, 33008, 33015, 33012, 33016, 33027, 33024, 33027, 33016, 33010, 32984, 32983, 32957, 32988, 32987, 32961, 32987, 32984, 32958, 33006, 32974, 32973, 32978, 33012, 33011, 33012, 32978, 32974, 32967, 32968, 32954, 32977, 32962, 32960, 32968, 32975, 32960, 33013, 32990, 32988, 33025, 33004, 33002, 33023, 33002, 32990, 33011, 32988, 32962, 33038, 33036, 33035, 33034, 33037, 33035, 33041, 33043, 33042, 33038, 33045, 33044, 33043, 33038, 33037, 33034, 33040, 33042, 33034, 33033, 33031, 33040, 33034, 33032, 33038, 33043, 33046, 33050, 33049, 33047, 33054, 33053, 33051, 33060, 33058, 33057, 33056, 33059, 33057,
33062, 33064, 33063, 33068, 33067, 33065, 33062, 33061, 33049, 33066, 33065, 33053, 33063, 33064, 33060, 33061, 33063, 33059, 33053, 33050, 33048, 33067, 33064, 33062, 33065, 33062, 33050, 33074, 33072, 33071, 33074, 33076, 33075, 33079, 33074, 33073, 33079, 33080, 33076, 33073, 33071, 33069, 33078, 33073, 33070, 33060, 33070, 33069, 33068, 33079, 33078, 33067, 33078, 33077, 33064, 33077, 33070, 33049, 33043, 33041, 33056, 33055, 33044, 33046, 33061, 33056, 33049, 33061, 33046, 32996, 33032, 33031, 33041, 33040, 33003, 33003, 33040, 33039, 32996, 33000, 33039, 33048, 33047, 33025, 33052, 33051, 33029, 33051, 33048, 33026, 33047, 33041, 33004, 32947, 32949, 32937, 32955, 32957, 32943, 32949, 32955, 32942, 32980, 32979, 32944, 32983, 32980, 32945, 32957, 32983, 32946, 32603, 32931, 32929, 32607, 32936, 32935, 32603, 32606, 32935, 32948, 32947, 32646, 32952, 32951, 32650, 32951, 32948, 32647, 32964, 32661, 32660, 32967, 32663, 32661, 32663, 32967, 32952, 32947, 32936, 32607, 32149, 32353, 32351, 32150, 32358, 32353, 32433, 32191, 32189, 32196, 32435, 32434, 32438, 32435, 32196, 32438, 32199, 32191, 32434, 32358, 32150, 32284, 32479, 32478, 32483, 32482, 32287, 32287, 32482, 32479, 32511, 32510, 32323, 32515, 32514, 32327, 32514, 32511, 32324,
32535, 32534, 32339, 32539, 32538, 32343, 32538, 32535, 32340, 32534, 32515, 32328, 32510, 32483, 32288, 33082, 33084, 33083, 33086, 33088, 33087, 33094, 33096, 33095, 33091, 33092, 33084, 33088, 33094, 33093, 33085, 33087, 33082, 33095, 33092, 33091, 33087, 33093, 33091, 33098, 33100, 33099, 33104, 33102, 33101, 33106, 33108, 33107, 33112, 33110, 33109, 33116, 33115, 33113, 33120, 33119, 33117, 33100, 33114, 33113, 33116, 33104, 33103, 33108, 33118, 33117, 33120, 33112, 33111, 33105, 33107, 33100, 33111, 33109, 33102, 33119, 33116, 33114, 33107, 33117, 33114, 33119, 33111, 33104, 33097, 33099, 33088, 33103, 33101, 33089, 33113, 33115, 33096, 33099, 33113, 33094, 33115, 33103, 33090, 33122, 33124, 33123, 33128, 33126, 33125, 33130, 33132, 33131, 33138, 33137, 33135, 33142, 33141, 33139, 33124, 33136, 33135, 33138, 33128, 33127, 33132, 33140, 33139, 33129, 33131, 33124, 33134, 33133, 33126, 33139, 33141, 33138, 33131, 33139, 33136, 33141, 33134, 33128, 33144, 33146, 33145, 33148, 33146, 33144, 33152, 33151, 33149, 33146, 33150, 33149, 33152, 33150, 33146, 33143, 33145, 33132, 33151, 33142, 33140, 33145, 33149, 33140, 33154, 33156, 33155, 33160, 33158, 33157, 33166, 33165, 33163, 33168, 33170, 33169, 33174, 33172, 33171, 33180, 33179, 33177,
33156, 33168, 33167, 33172, 33160, 33159, 33178, 33177, 33165, 33161, 33162, 33156, 33175, 33176, 33170, 33166, 33164, 33158, 33180, 33178, 33172, 33165, 33162, 33161, 33179, 33176, 33175, 33177, 33175, 33162, 33162, 33175, 33168, 33178, 33166, 33160, 33182, 33184, 33183, 33188, 33190, 33189, 33194, 33193, 33191, 33183, 33188, 33187, 33192, 33191, 33185, 33184, 33186, 33185, 33191, 33193, 33190, 33185, 33191, 33188, 33196, 33198, 33197, 33202, 33200, 33199, 33208, 33207, 33205, 33210, 33212, 33211, 33218, 33216, 33215, 33218, 33214, 33213, 33224, 33223, 33221, 33198, 33210, 33209, 33214, 33202, 33201, 33222, 33221, 33207, 33203, 33204, 33198, 33219, 33220, 33212, 33208, 33206, 33200, 33224, 33218, 33217, 33224, 33222, 33214, 33207, 33204, 33203, 33217, 33215, 33211, 33223, 33220, 33219, 33223, 33217, 33212, 33221, 33219, 33204, 33204, 33219, 33210, 33222, 33208, 33202, 33226, 33228, 33227, 33232, 33231, 33229, 33238, 33236, 33235, 33234, 33237, 33235, 33242, 33241, 33239, 33228, 33234, 33233, 33240, 33239, 33231, 33229, 33231, 33228, 33242, 33238, 33237, 33239, 33241, 33237, 33231, 33239, 33234, 33170, 33196, 33195, 33200, 33174, 33173, 33206, 33205, 33179, 33205, 33203, 33176, 33176, 33203, 33196, 33206, 33180, 33174, 33190, 33226, 33225,
33230, 33229, 33193, 33193, 33229, 33226, 33248, 33246, 33245, 33248, 33250, 33249, 33258, 33256, 33255, 33260, 33262, 33261, 33256, 33244, 33243, 33250, 33260, 33259, 33264, 33263, 33256, 33252, 33253, 33248, 33248, 33253, 33254, 33265, 33266, 33262, 33263, 33251, 33244, 33250, 33254, 33265, 33270, 33268, 33267, 33274, 33273, 33271, 33278, 33276, 33275, 33282, 33281, 33279, 33276, 33270, 33269, 33280, 33279, 33273, 33274, 33272, 33268, 33282, 33280, 33276, 33280, 33274, 33270, 33286, 33284, 33283, 33288, 33290, 33289, 33298, 33296, 33295, 33300, 33302, 33301, 33296, 33286, 33285, 33290, 33300, 33299, 33292, 33291, 33284, 33304, 33303, 33296, 33293, 33294, 33290, 33305, 33306, 33302, 33303, 33292, 33286, 33294, 33305, 33300, 33310, 33308, 33307, 33314, 33313, 33311, 33320, 33318, 33317, 33320, 33316, 33315, 33324, 33323, 33321, 33316, 33310, 33309, 33322, 33321, 33313, 33314, 33312, 33308, 33324, 33320, 33319, 33324, 33322, 33316, 33322, 33314, 33310, 33284, 33258, 33257, 33262, 33288, 33287, 33291, 33264, 33258, 33266, 33293, 33288, 33308, 33278, 33277, 33312, 33311, 33281, 33312, 33282, 33278, 33271, 33273, 33254, 33279, 33281, 33266, 33273, 33279, 33265, 33311, 33313, 33294, 33319, 33317, 33301, 33323, 33306, 33305, 33323, 33319, 33302,
33321, 33305, 33294, 33311, 33293, 33266, 33084, 33154, 33153, 33158, 33090, 33089, 33096, 33164, 33163, 33163, 33161, 33092, 33092, 33161, 33154, 33164, 33096, 33090, 33134, 33247, 33245, 33142, 33253, 33252, 33141, 33252, 33247, 33268, 33148, 33147, 33272, 33271, 33151, 33272, 33152, 33148, 33271, 33253, 33142, 33121, 33123, 33108, 33127, 33125, 33110, 33137, 33120, 33118, 33123, 33135, 33118, 33137, 33127, 33112, 33243, 33244, 33184, 33264, 33194, 33192, 33244, 33251, 33186, 33263, 33192, 33186, 33292, 33232, 33230, 33298, 33297, 33236, 33304, 33242, 33240, 33304, 33298, 33238, 33303, 33240, 33232, 33291, 33230, 33194, 33326, 33328, 33327, 33327, 33330, 33329, 33334, 33333, 33331, 33328, 33332, 33331, 33331, 33333, 33330, 33336, 33338, 33337, 33342, 33341, 33339, 33344, 33346, 33345, 33350, 33349, 33347, 33338, 33344, 33343, 33348, 33347, 33341, 33339, 33341, 33338, 33347, 33349, 33346, 33341, 33347, 33344, 33330, 33336, 33335, 33340, 33339, 33333, 33333, 33339, 33336, 33352, 33354, 33353, 33356, 33358, 33357, 33362, 33361, 33359, 33366, 33365, 33363, 33354, 33360, 33359, 33358, 33364, 33363, 33355, 33357, 33354, 33365, 33362, 33360, 33357, 33363, 33360, 33370, 33368, 33367, 33372, 33371, 33368, 33376, 33375, 33373, 33380, 33379, 33377,
33384, 33383, 33381, 33388, 33387, 33385, 33382, 33381, 33375, 33386, 33385, 33379, 33379, 33376, 33374, 33387, 33384, 33382, 33385, 33382, 33376, 33392, 33390, 33389, 33398, 33396, 33395, 33396, 33392, 33391, 33394, 33393, 33390, 33400, 33399, 33396, 33399, 33394, 33392, 33374, 33373, 33361, 33378, 33377, 33365, 33377, 33374, 33362, 33390, 33370, 33369, 33393, 33372, 33370, 33367, 33368, 33358, 33372, 33366, 33364, 33368, 33371, 33364, 33394, 33380, 33378, 33400, 33388, 33386, 33399, 33386, 33380, 33393, 33378, 33366, 33402, 33404, 33403, 33408, 33407, 33405, 33410, 33412, 33411, 33416, 33415, 33413, 33404, 33410, 33409, 33414, 33413, 33407, 33405, 33407, 33404, 33413, 33415, 33412, 33407, 33413, 33410, 33418, 33420, 33419, 33424, 33423, 33421, 33426, 33428, 33427, 33432, 33431, 33429, 33420, 33426, 33425, 33430, 33429, 33423, 33421, 33423, 33420, 33429, 33431, 33428, 33423, 33429, 33426, 33412, 33418, 33417, 33422, 33421, 33415, 33415, 33421, 33418, 33436, 33435, 33433, 33440, 33439, 33437, 33444, 33443, 33441, 33448, 33447, 33445, 33442, 33441, 33435, 33446, 33445, 33439, 33439, 33436, 33434, 33447, 33444, 33442, 33445, 33442, 33436, 33452, 33450, 33449, 33458, 33456, 33455, 33456, 33452, 33451, 33454, 33453, 33450, 33460, 33459, 33456,
33459, 33454, 33452, 33464, 33463, 33461, 33468, 33467, 33465, 33472, 33471, 33469, 33476, 33475, 33473, 33470, 33469, 33463, 33474, 33473, 33467, 33467, 33464, 33462, 33475, 33472, 33470, 33473, 33470, 33464, 33480, 33478, 33477, 33486, 33484, 33483, 33484, 33480, 33479, 33482, 33481, 33478, 33488, 33487, 33484, 33487, 33482, 33480, 33462, 33461, 33443, 33466, 33465, 33447, 33465, 33462, 33444, 33478, 33458, 33457, 33481, 33460, 33458, 33454, 33440, 33438, 33460, 33448, 33446, 33459, 33446, 33440, 33482, 33468, 33466, 33488, 33476, 33474, 33487, 33474, 33468, 33481, 33466, 33448, 33346, 33402, 33401, 33406, 33405, 33349, 33349, 33405, 33402, 33434, 33433, 33383, 33438, 33437, 33387, 33437, 33434, 33384, 33450, 33398, 33397, 33453, 33400, 33398, 33453, 33438, 33388, 33351, 33353, 33328, 33361, 33334, 33332, 33353, 33359, 33332, 33375, 33342, 33340, 33383, 33350, 33348, 33381, 33348, 33342, 33373, 33340, 33334, 33435, 33408, 33406, 33443, 33416, 33414, 33441, 33414, 33408, 33463, 33424, 33422, 33471, 33432, 33430, 33469, 33430, 33424, 33461, 33422, 33416, 33433, 33406, 33350, 33490, 33492, 33491, 33494, 33496, 33495, 33502, 33501, 33499, 33508, 33506, 33505, 33492, 33504, 33503, 33511, 33510, 33501, 33497, 33498, 33492, 33496, 33500, 33499,
33511, 33508, 33507, 33493, 33495, 33490, 33501, 33498, 33497, 33507, 33505, 33503, 33495, 33499, 33497, 33510, 33507, 33504, 33510, 33509, 33498, 33498, 33509, 33504, 33513, 33515, 33514, 33517, 33519, 33518, 33523, 33522, 33520, 33527, 33526, 33524, 33531, 33529, 33528, 33535, 33533, 33532, 33537, 33536, 33522, 33539, 33538, 33526, 33515, 33521, 33520, 33537, 33531, 33530, 33519, 33525, 33524, 33539, 33535, 33534, 33516, 33518, 33515, 33526, 33523, 33521, 33534, 33532, 33529, 33518, 33524, 33521, 33538, 33534, 33531, 33526, 33538, 33537, 33541, 33543, 33542, 33547, 33545, 33544, 33553, 33555, 33554, 33559, 33557, 33556, 33543, 33553, 33552, 33557, 33547, 33546, 33541, 33548, 33549, 33553, 33560, 33561, 33551, 33550, 33545, 33563, 33562, 33557, 33544, 33545, 33541, 33550, 33551, 33549, 33563, 33561, 33560, 33545, 33550, 33548, 33562, 33560, 33549, 33543, 33549, 33560, 33562, 33551, 33547, 33512, 33514, 33496, 33522, 33502, 33500, 33530, 33528, 33506, 33514, 33520, 33500, 33536, 33530, 33508, 33536, 33511, 33502, 33565, 33567, 33566, 33569, 33571, 33570, 33575, 33574, 33572, 33579, 33578, 33576, 33583, 33581, 33580, 33589, 33587, 33586, 33585, 33588, 33586, 33593, 33595, 33594, 33591, 33590, 33574, 33593, 33592, 33578, 33567, 33573, 33572,
33591, 33583, 33582, 33571, 33577, 33576, 33594, 33595, 33589, 33592, 33594, 33588, 33568, 33570, 33567, 33578, 33575, 33573, 33585, 33584, 33581, 33570, 33576, 33573, 33591, 33592, 33585, 33578, 33592, 33591, 33597, 33599, 33598, 33601, 33599, 33597, 33605, 33604, 33602, 33609, 33607, 33606, 33613, 33612, 33610, 33607, 33601, 33600, 33611, 33610, 33604, 33599, 33603, 33602, 33605, 33603, 33599, 33613, 33611, 33607, 33611, 33605, 33601, 33615, 33617, 33616, 33621, 33623, 33622, 33617, 33621, 33620, 33618, 33619, 33617, 33621, 33624, 33625, 33619, 33624, 33621, 33629, 33627, 33626, 33633, 33632, 33630, 33637, 33635, 33634, 33639, 33641, 33640, 33635, 33629, 33628, 33639, 33638, 33632, 33633, 33631, 33627, 33637, 33641, 33639, 33639, 33633, 33629, 33589, 33615, 33614, 33595, 33618, 33615, 33627, 33609, 33608, 33631, 33630, 33612, 33631, 33613, 33609, 33596, 33598, 33571, 33604, 33579, 33577, 33612, 33595, 33593, 33598, 33602, 33577, 33610, 33593, 33579, 33630, 33632, 33619, 33640, 33625, 33624, 33632, 33638, 33624, 33612, 33630, 33618, 33564, 33566, 33519, 33574, 33527, 33525, 33582, 33580, 33533, 33566, 33572, 33525, 33590, 33582, 33535, 33590, 33539, 33527, 33643, 33645, 33644, 33649, 33648, 33646, 33651, 33653, 33652, 33657, 33656, 33654,
33645, 33651, 33650, 33655, 33654, 33648, 33646, 33648, 33645, 33654, 33656, 33653, 33648, 33654, 33651, 33659, 33661, 33660, 33665, 33664, 33662, 33667, 33669, 33668, 33673, 33672, 33670, 33661, 33667, 33666, 33671, 33670, 33664, 33662, 33664, 33661, 33670, 33672, 33669, 33664, 33670, 33667, 33653, 33659, 33658, 33663, 33662, 33656, 33656, 33662, 33659, 33677, 33676, 33674, 33681, 33680, 33678, 33685, 33684, 33682, 33689, 33688, 33686, 33683, 33682, 33676, 33687, 33686, 33680, 33680, 33677, 33675, 33688, 33685, 33683, 33686, 33683, 33677, 33693, 33691, 33690, 33699, 33697, 33696, 33697, 33693, 33692, 33695, 33694, 33691, 33701, 33700, 33697, 33700, 33695, 33693, 33705, 33704, 33702, 33709, 33708, 33706, 33713, 33712, 33710, 33717, 33716, 33714, 33711, 33710, 33704, 33715, 33714, 33708, 33708, 33705, 33703, 33716, 33713, 33711, 33714, 33711, 33705, 33721, 33719, 33718, 33727, 33725, 33724, 33725, 33721, 33720, 33723, 33722, 33719, 33729, 33728, 33725, 33728, 33723, 33721, 33703, 33702, 33684, 33707, 33706, 33688, 33706, 33703, 33685, 33719, 33699, 33698, 33722, 33701, 33699, 33695, 33681, 33679, 33701, 33689, 33687, 33700, 33687, 33681, 33723, 33709, 33707, 33729, 33717, 33715, 33728, 33715, 33709, 33722, 33707, 33689, 33731, 33733, 33732,
33735, 33737, 33736, 33739, 33741, 33740, 33743, 33745, 33744, 33733, 33739, 33738, 33743, 33742, 33736, 33734, 33736, 33733, 33742, 33744, 33741, 33736, 33742, 33739, 33751, 33749, 33748, 33747, 33750, 33748, 33753, 33755, 33754, 33757, 33759, 33758, 33751, 33757, 33756, 33755, 33751, 33750, 33747, 33752, 33754, 33760, 33761, 33759, 33751, 33755, 33760, 33741, 33747, 33746, 33745, 33753, 33752, 33744, 33752, 33747, 33765, 33764, 33762, 33769, 33768, 33766, 33773, 33772, 33770, 33777, 33776, 33774, 33771, 33770, 33764, 33775, 33774, 33768, 33768, 33765, 33763, 33776, 33773, 33771, 33774, 33771, 33765, 33781, 33779, 33778, 33787, 33785, 33784, 33785, 33781, 33780, 33783, 33782, 33779, 33789, 33788, 33785, 33788, 33783, 33781, 33793, 33792, 33790, 33797, 33796, 33794, 33801, 33800, 33798, 33805, 33804, 33802, 33799, 33798, 33792, 33803, 33802, 33796, 33796, 33793, 33791, 33804, 33801, 33799, 33802, 33799, 33793, 33809, 33807, 33806, 33813, 33815, 33814, 33817, 33815, 33813, 33821, 33823, 33822, 33814, 33809, 33808, 33811, 33810, 33807, 33815, 33821, 33820, 33817, 33823, 33821, 33818, 33819, 33817, 33817, 33819, 33824, 33814, 33820, 33811, 33791, 33790, 33772, 33795, 33794, 33776, 33794, 33791, 33773, 33807, 33787, 33786, 33810, 33789, 33787,
33783, 33769, 33767, 33788, 33789, 33777, 33783, 33788, 33775, 33810, 33811, 33797, 33822, 33805, 33803, 33811, 33820, 33803, 33789, 33810, 33795, 33669, 33731, 33730, 33735, 33734, 33672, 33672, 33734, 33731, 33763, 33762, 33712, 33767, 33766, 33716, 33766, 33763, 33713, 33779, 33727, 33726, 33782, 33729, 33727, 33729, 33782, 33767, 33676, 33649, 33647, 33684, 33657, 33655, 33682, 33655, 33649, 33704, 33665, 33663, 33712, 33673, 33671, 33710, 33671, 33665, 33702, 33663, 33657, 33764, 33737, 33735, 33772, 33745, 33743, 33770, 33743, 33737, 33792, 33755, 33753, 33798, 33800, 33761, 33792, 33798, 33760, 33790, 33753, 33745, 33762, 33735, 33673, 33428, 33643, 33642, 33647, 33646, 33431, 33431, 33646, 33643, 33675, 33674, 33471, 33679, 33678, 33475, 33678, 33675, 33472, 33691, 33486, 33485, 33694, 33488, 33486, 33694, 33679, 33476, 33674, 33647, 33432, 33829, 33831, 33830, 33830, 33833, 33832, 33837, 33836, 33834, 33826, 33831, 33829, 33830, 33831, 33835, 33830, 33834, 33836, 33827, 33835, 33831, 33839, 33841, 33840, 33840, 33843, 33842, 33847, 33849, 33848, 33848, 33845, 33844, 33853, 33852, 33850, 33860, 33858, 33857, 33864, 33862, 33861, 33868, 33867, 33865, 33870, 33872, 33871, 33862, 33849, 33847, 33853, 33866, 33865, 33856, 33870, 33869,
33840, 33841, 33851, 33850, 33852, 33843, 33868, 33860, 33859, 33849, 33856, 33855, 33848, 33855, 33854, 33872, 33870, 33862, 33854, 33855, 33853, 33869, 33871, 33868, 33869, 33866, 33853, 33870, 33856, 33849, 33874, 33876, 33875, 33884, 33882, 33881, 33884, 33880, 33878, 33890, 33889, 33887, 33876, 33892, 33891, 33894, 33883, 33881, 33885, 33886, 33876, 33890, 33884, 33883, 33884, 33890, 33888, 33877, 33879, 33874, 33887, 33889, 33886, 33894, 33893, 33891, 33874, 33879, 33887, 33896, 33894, 33892, 33896, 33895, 33886, 33876, 33886, 33895, 33894, 33896, 33889, 33906, 33904, 33903, 33902, 33905, 33903, 33910, 33912, 33911, 33906, 33914, 33913, 33911, 33912, 33906, 33909, 33911, 33905, 33902, 33901, 33898, 33909, 33902, 33900, 33912, 33915, 33914, 33833, 33879, 33877, 33837, 33888, 33887, 33833, 33836, 33887, 33899, 33859, 33857, 33860, 33900, 33898, 33910, 33909, 33871, 33871, 33909, 33908, 33899, 33907, 33867, 33860, 33868, 33908, 33842, 33843, 33826, 33865, 33867, 33837, 33826, 33843, 33852, 33865, 33835, 33827, 33899, 33897, 33878, 33907, 33899, 33880, 33907, 33888, 33837, 33917, 33919, 33918, 33923, 33925, 33924, 33927, 33929, 33928, 33928, 33931, 33930, 33937, 33936, 33934, 33919, 33929, 33927, 33925, 33933, 33932, 33917, 33920, 33921,
33928, 33929, 33935, 33934, 33936, 33931, 33930, 33931, 33925, 33925, 33931, 33936, 33919, 33921, 33935, 33941, 33939, 33938, 33943, 33945, 33944, 33951, 33949, 33948, 33951, 33947, 33946, 33955, 33954, 33952, 33947, 33941, 33940, 33953, 33952, 33944, 33941, 33945, 33943, 33954, 33955, 33951, 33955, 33953, 33947, 33947, 33953, 33945, 33959, 33957, 33956, 33963, 33962, 33960, 33967, 33966, 33964, 33971, 33969, 33968, 33973, 33958, 33956, 33963, 33975, 33974, 33975, 33971, 33970, 33966, 33967, 33959, 33964, 33966, 33963, 33973, 33972, 33969, 33976, 33973, 33971, 33976, 33975, 33963, 33973, 33976, 33966, 33980, 33978, 33977, 33980, 33982, 33981, 33933, 33961, 33960, 33937, 33965, 33964, 33964, 33961, 33933, 33978, 33950, 33948, 33978, 33981, 33954, 33944, 33921, 33920, 33952, 33954, 33937, 33952, 33935, 33921, 33980, 33979, 33957, 33982, 33967, 33965, 33967, 33982, 33980, 33981, 33965, 33937, 33922, 33924, 33864, 33924, 33932, 33872, 33962, 33912, 33910, 33970, 33968, 33913, 33974, 33970, 33914, 33962, 33974, 33915, 33932, 33960, 33910, 33987, 33989, 33988, 33991, 33993, 33992, 33997, 33996, 33994, 34001, 33999, 33998, 34001, 34003, 34002, 34012, 34011, 34009, 33989, 34000, 33998, 33993, 34007, 34006, 34010, 34009, 33996, 33987, 33990, 33992,
34007, 34001, 34000, 34001, 34007, 34008, 33996, 33993, 33991, 34005, 34004, 34002, 34009, 34011, 34008, 34011, 34005, 34003, 34009, 34007, 33993, 33992, 34006, 34000, 34014, 34016, 34015, 34018, 34016, 34014, 34022, 34021, 34019, 34024, 34026, 34025, 34030, 34029, 34027, 34034, 34033, 34031, 34028, 34027, 34021, 34032, 34031, 34025, 34015, 34016, 34024, 34026, 34024, 34016, 34025, 34022, 34020, 34033, 34030, 34028, 34031, 34028, 34022, 34038, 34036, 34035, 34038, 34040, 34039, 34041, 34042, 34038, 34042, 34043, 34040, 34047, 34045, 34044, 34049, 34051, 34050, 34055, 34054, 34052, 34061, 34059, 34058, 34047, 34057, 34056, 34055, 34064, 34063, 34050, 34051, 34047, 34064, 34061, 34060, 34054, 34051, 34049, 34060, 34058, 34056, 34062, 34063, 34060, 34054, 34063, 34062, 34051, 34062, 34057, 34005, 34037, 34035, 34012, 34042, 34041, 34011, 34041, 34037, 34049, 34048, 34029, 34053, 34052, 34033, 34052, 34049, 34030, 34021, 33997, 33995, 34029, 34012, 34010, 34027, 34010, 33997, 34046, 34044, 34039, 34048, 34050, 34043, 34050, 34046, 34040, 34048, 34042, 34012, 33987, 33986, 33983, 33990, 33987, 33984, 33555, 33841, 33839, 33845, 33559, 33558, 33854, 33851, 33561, 33561, 33851, 33841, 33854, 33563, 33559, 33623, 33917, 33916, 33623, 33625, 33920,
33939, 33637, 33636, 33641, 33943, 33942, 33939, 33943, 33641, 33942, 33920, 33625, 33759, 33984, 33983, 33759, 33761, 33985, 33991, 33990, 33800, 33995, 33994, 33804, 33994, 33991, 33801, 34015, 33819, 33818, 34020, 34019, 33822, 33824, 34023, 34020, 34015, 34023, 33824, 34019, 33995, 33805, 33800, 33990, 33985, 31705, 31936, 31935, 31945, 31944, 31712, 31712, 31944, 31943, 31705, 31709, 31943, 31952, 31951, 31731, 31956, 31955, 31735, 31955, 31952, 31732, 31951, 31945, 31713, 34066, 34068, 34067, 34070, 34072, 34071, 34076, 34075, 34073, 34078, 34080, 34079, 34082, 34084, 34083, 34088, 34087, 34085, 34068, 34078, 34077, 34072, 34082, 34081, 34086, 34085, 34075, 34069, 34071, 34068, 34081, 34083, 34080, 34075, 34072, 34070, 34087, 34084, 34082, 34085, 34082, 34072, 34071, 34081, 34078, 34092, 34090, 34089, 34094, 34096, 34095, 34100, 34098, 34097, 34104, 34106, 34105, 34098, 34092, 34091, 34096, 34104, 34103, 34096, 34094, 34090, 34100, 34106, 34104, 34101, 34102, 34100, 34102, 34107, 34106, 34104, 34096, 34092, 34111, 34109, 34108, 34111, 34113, 34112, 34120, 34119, 34117, 34122, 34124, 34123, 34132, 34131, 34129, 34113, 34122, 34121, 34130, 34129, 34119, 34115, 34111, 34110, 34111, 34115, 34116, 34127, 34128, 34124, 34117, 34119, 34116,
34126, 34125, 34123, 34129, 34131, 34128, 34128, 34131, 34126, 34119, 34129, 34127, 34116, 34127, 34122, 34134, 34136, 34135, 34138, 34136, 34134, 34142, 34141, 34139, 34144, 34146, 34145, 34150, 34149, 34147, 34154, 34153, 34151, 34148, 34147, 34141, 34152, 34151, 34145, 34136, 34144, 34143, 34146, 34144, 34136, 34145, 34142, 34140, 34153, 34150, 34148, 34151, 34148, 34142, 34080, 34110, 34108, 34084, 34115, 34114, 34118, 34117, 34087, 34117, 34115, 34084, 34080, 34083, 34114, 34135, 34102, 34101, 34140, 34139, 34105, 34107, 34143, 34140, 34135, 34143, 34107, 34095, 34076, 34074, 34105, 34088, 34086, 34103, 34086, 34076, 34141, 34120, 34118, 34149, 34132, 34130, 34147, 34130, 34120, 34139, 34118, 34088, 34159, 34161, 34160, 34165, 34167, 34166, 34166, 34163, 34162, 34169, 34171, 34170, 34160, 34156, 34155, 34161, 34169, 34168, 34166, 34167, 34174, 34173, 34172, 34163, 34162, 34163, 34161, 34172, 34173, 34171, 34163, 34172, 34169, 34168, 34157, 34156, 34181, 34183, 34182, 34179, 34183, 34181, 34185, 34187, 34186, 34194, 34193, 34191, 34176, 34187, 34185, 34183, 34190, 34189, 34179, 34188, 34190, 34186, 34187, 34192, 34184, 34186, 34179, 34193, 34190, 34188, 34179, 34186, 34191, 34176, 34177, 34192, 34180, 34182, 34167, 34167, 34182, 34189,
34198, 34196, 34195, 34198, 34200, 34199, 34202, 34198, 34197, 34198, 34202, 34203, 34209, 34208, 34206, 34213, 34212, 34210, 34217, 34215, 34214, 34217, 34219, 34218, 34223, 34221, 34220, 34228, 34230, 34229, 34205, 34216, 34214, 34209, 34225, 34224, 34228, 34227, 34212, 34224, 34225, 34217, 34217, 34225, 34226, 34230, 34223, 34222, 34212, 34209, 34207, 34222, 34220, 34218, 34227, 34229, 34226, 34229, 34222, 34219, 34227, 34225, 34209, 34208, 34224, 34216, 34205, 34204, 34199, 34206, 34208, 34203, 34203, 34208, 34205, 34234, 34233, 34231, 34238, 34237, 34235, 34244, 34243, 34241, 34248, 34247, 34245, 34242, 34241, 34233, 34246, 34245, 34237, 34237, 34234, 34232, 34247, 34244, 34242, 34245, 34242, 34234, 34252, 34251, 34249, 34256, 34255, 34253, 34260, 34258, 34257, 34264, 34263, 34261, 34268, 34267, 34265, 34262, 34261, 34251, 34266, 34265, 34255, 34268, 34260, 34259, 34255, 34252, 34250, 34265, 34267, 34264, 34265, 34262, 34252, 34272, 34270, 34269, 34276, 34274, 34273, 34277, 34278, 34272, 34280, 34276, 34275, 34275, 34273, 34270, 34279, 34275, 34272, 34284, 34282, 34281, 34287, 34288, 34284, 34286, 34285, 34282, 34289, 34286, 34284, 34240, 34271, 34269, 34244, 34278, 34277, 34248, 34280, 34279, 34247, 34279, 34278, 34243, 34277, 34271,
34286, 34259, 34257, 34288, 34287, 34263, 34289, 34288, 34264, 34286, 34289, 34267, 34251, 34238, 34236, 34263, 34248, 34246, 34261, 34246, 34238, 34283, 34281, 34274, 34280, 34287, 34283, 34263, 34287, 34280, 34126, 34197, 34195, 34132, 34202, 34201, 34126, 34131, 34201, 34207, 34206, 34149, 34211, 34210, 34153, 34210, 34207, 34150, 34206, 34202, 34132, 34232, 34231, 34170, 34174, 34236, 34235, 34235, 34232, 34171, 34250, 34249, 34189, 34254, 34253, 34193, 34193, 34253, 34250, 34189, 34249, 34236, 34155, 34156, 34138, 34170, 34154, 34152, 34138, 34156, 34157, 34157, 34168, 34152, 34233, 34213, 34211, 34240, 34239, 34221, 34241, 34243, 34230, 34230, 34243, 34240, 34241, 34228, 34213, 34231, 34211, 34154, 34293, 34295, 34294, 34300, 34299, 34297, 34302, 34304, 34303, 34310, 34309, 34307, 34291, 34302, 34301, 34300, 34308, 34307, 34302, 34305, 34306, 34297, 34299, 34295, 34294, 34295, 34291, 34309, 34306, 34305, 34291, 34295, 34299, 34307, 34305, 34296, 34296, 34305, 34302, 34314, 34312, 34311, 34318, 34317, 34315, 34322, 34321, 34319, 34326, 34324, 34323, 34328, 34325, 34323, 34332, 34331, 34329, 34326, 34314, 34313, 34330, 34329, 34317, 34334, 34333, 34321, 34314, 34322, 34320, 34333, 34334, 34326, 34335, 34333, 34325, 34321, 34318, 34316,
34335, 34332, 34330, 34321, 34333, 34330, 34334, 34322, 34314, 34337, 34339, 34338, 34338, 34341, 34340, 34345, 34344, 34342, 34349, 34348, 34346, 34353, 34352, 34350, 34357, 34356, 34354, 34345, 34351, 34350, 34355, 34354, 34348, 34338, 34339, 34343, 34342, 34344, 34341, 34346, 34348, 34345, 34356, 34353, 34351, 34354, 34351, 34345, 34363, 34361, 34360, 34365, 34367, 34366, 34375, 34373, 34372, 34375, 34371, 34370, 34379, 34378, 34376, 34371, 34359, 34358, 34377, 34376, 34366, 34378, 34379, 34375, 34375, 34379, 34377, 34369, 34368, 34361, 34363, 34362, 34358, 34369, 34367, 34365, 34367, 34369, 34363, 34371, 34377, 34367, 34304, 34339, 34337, 34347, 34346, 34309, 34346, 34343, 34306, 34304, 34306, 34343, 34361, 34328, 34327, 34365, 34364, 34331, 34368, 34365, 34332, 34361, 34368, 34335, 34317, 34300, 34298, 34331, 34310, 34308, 34329, 34308, 34300, 34366, 34349, 34347, 34376, 34378, 34357, 34376, 34355, 34349, 34364, 34347, 34310, 34385, 34383, 34382, 34385, 34381, 34380, 34389, 34388, 34386, 34393, 34392, 34390, 34399, 34397, 34396, 34399, 34395, 34394, 34405, 34404, 34402, 34401, 34384, 34382, 34389, 34403, 34402, 34404, 34405, 34399, 34399, 34405, 34403, 34393, 34385, 34384, 34393, 34391, 34381, 34390, 34392, 34389, 34401, 34400, 34394,
34406, 34401, 34395, 34406, 34403, 34389, 34401, 34406, 34392, 34408, 34407, 34380, 34391, 34409, 34408, 34387, 34386, 34352, 34357, 34391, 34390, 34390, 34387, 34353, 34408, 34374, 34372, 34409, 34378, 34374, 34409, 34391, 34357, 34340, 34341, 34176, 34350, 34352, 34194, 34341, 34344, 34177, 34350, 34192, 34177, 34388, 34256, 34254, 34398, 34396, 34258, 34404, 34268, 34266, 34268, 34404, 34398, 34402, 34266, 34256, 34386, 34254, 34194, 34411, 34413, 34412, 34415, 34417, 34416, 34421, 34420, 34418, 34425, 34424, 34422, 34429, 34428, 34426, 34433, 34432, 34430, 34427, 34426, 34420, 34431, 34430, 34424, 34413, 34419, 34418, 34417, 34423, 34422, 34414, 34416, 34413, 34424, 34421, 34419, 34432, 34429, 34427, 34416, 34422, 34419, 34430, 34427, 34421, 34435, 34437, 34436, 34439, 34441, 34440, 34445, 34444, 34442, 34449, 34448, 34446, 34453, 34452, 34450, 34457, 34456, 34454, 34451, 34450, 34444, 34455, 34454, 34448, 34437, 34443, 34442, 34441, 34447, 34446, 34438, 34440, 34437, 34448, 34445, 34443, 34456, 34453, 34451, 34440, 34446, 34443, 34454, 34451, 34445, 34461, 34459, 34458, 34465, 34463, 34462, 34469, 34468, 34466, 34473, 34472, 34470, 34469, 34461, 34460, 34473, 34465, 34464, 34464, 34462, 34459, 34472, 34469, 34467, 34472, 34464, 34461,
34477, 34479, 34478, 34483, 34482, 34480, 34485, 34487, 34486, 34491, 34490, 34488, 34495, 34494, 34492, 34475, 34485, 34484, 34479, 34489, 34488, 34493, 34492, 34482, 34488, 34490, 34487, 34482, 34479, 34477, 34494, 34491, 34489, 34492, 34489, 34479, 34478, 34488, 34485, 34467, 34466, 34428, 34471, 34470, 34432, 34470, 34467, 34429, 34477, 34476, 34452, 34481, 34480, 34456, 34480, 34477, 34453, 34434, 34436, 34417, 34444, 34425, 34423, 34452, 34433, 34431, 34436, 34442, 34423, 34450, 34431, 34425, 34475, 34474, 34463, 34476, 34478, 34473, 34478, 34475, 34465, 34476, 34471, 34433, 34497, 34499, 34498, 34501, 34503, 34502, 34507, 34506, 34504, 34511, 34510, 34508, 34515, 34514, 34512, 34519, 34518, 34516, 34513, 34512, 34506, 34517, 34516, 34510, 34499, 34505, 34504, 34503, 34509, 34508, 34500, 34502, 34499, 34510, 34507, 34505, 34518, 34515, 34513, 34502, 34508, 34505, 34516, 34513, 34507, 34521, 34523, 34522, 34525, 34527, 34526, 34531, 34530, 34528, 34535, 34534, 34532, 34539, 34538, 34536, 34543, 34542, 34540, 34537, 34536, 34530, 34541, 34540, 34534, 34523, 34529, 34528, 34527, 34533, 34532, 34524, 34526, 34523, 34534, 34531, 34529, 34542, 34539, 34537, 34526, 34532, 34529, 34540, 34537, 34531, 34547, 34545, 34544, 34551, 34550, 34548,
34555, 34554, 34552, 34559, 34557, 34556, 34563, 34562, 34560, 34557, 34546, 34544, 34561, 34560, 34550, 34555, 34547, 34546, 34565, 34564, 34557, 34554, 34551, 34549, 34565, 34563, 34561, 34564, 34561, 34551, 34564, 34554, 34546, 34569, 34567, 34566, 34573, 34571, 34570, 34577, 34576, 34574, 34581, 34580, 34578, 34577, 34569, 34568, 34581, 34573, 34572, 34572, 34570, 34567, 34580, 34577, 34575, 34580, 34572, 34569, 34549, 34548, 34514, 34553, 34552, 34518, 34552, 34549, 34515, 34575, 34574, 34538, 34579, 34578, 34542, 34578, 34575, 34539, 34520, 34522, 34503, 34530, 34511, 34509, 34538, 34519, 34517, 34522, 34528, 34509, 34536, 34517, 34511, 34568, 34566, 34545, 34576, 34555, 34553, 34576, 34568, 34547, 34574, 34553, 34519, 34583, 34585, 34584, 34589, 34588, 34586, 34593, 34592, 34590, 34595, 34597, 34596, 34601, 34600, 34598, 34605, 34604, 34602, 34585, 34595, 34594, 34599, 34598, 34588, 34603, 34602, 34592, 34586, 34588, 34585, 34598, 34600, 34597, 34592, 34589, 34587, 34604, 34601, 34599, 34602, 34599, 34589, 34588, 34598, 34595, 34607, 34609, 34608, 34613, 34612, 34610, 34617, 34616, 34614, 34619, 34621, 34620, 34625, 34624, 34622, 34629, 34628, 34626, 34609, 34619, 34618, 34623, 34622, 34612, 34627, 34626, 34616, 34610, 34612, 34609,
34622, 34624, 34621, 34616, 34613, 34611, 34628, 34625, 34623, 34626, 34623, 34613, 34612, 34622, 34619, 34597, 34607, 34606, 34611, 34610, 34600, 34615, 34614, 34604, 34614, 34611, 34601, 34600, 34610, 34607, 34633, 34631, 34630, 34637, 34636, 34634, 34643, 34641, 34640, 34647, 34646, 34644, 34641, 34633, 34632, 34645, 34644, 34636, 34639, 34638, 34631, 34649, 34648, 34641, 34639, 34637, 34635, 34649, 34647, 34645, 34648, 34645, 34637, 34648, 34639, 34633, 34653, 34651, 34650, 34657, 34656, 34654, 34663, 34661, 34660, 34667, 34666, 34664, 34661, 34653, 34652, 34665, 34664, 34656, 34659, 34658, 34651, 34669, 34668, 34661, 34659, 34657, 34655, 34669, 34667, 34665, 34668, 34665, 34657, 34668, 34659, 34653, 34651, 34643, 34642, 34655, 34654, 34646, 34658, 34655, 34647, 34658, 34649, 34643, 34487, 34583, 34582, 34587, 34586, 34490, 34591, 34590, 34494, 34590, 34587, 34491, 34490, 34586, 34583, 34631, 34559, 34558, 34635, 34634, 34562, 34638, 34635, 34563, 34638, 34565, 34559, 34496, 34498, 34441, 34506, 34449, 34447, 34514, 34457, 34455, 34498, 34504, 34447, 34512, 34455, 34449, 34550, 34483, 34481, 34562, 34495, 34493, 34560, 34493, 34483, 34548, 34481, 34457, 34636, 34593, 34591, 34646, 34605, 34603, 34644, 34603, 34593, 34656, 34617, 34615,
34666, 34629, 34627, 34664, 34627, 34617, 34654, 34615, 34605, 34634, 34591, 34495, 34671, 34673, 34672, 34675, 34677, 34676, 34681, 34680, 34678, 34685, 34684, 34682, 34689, 34688, 34686, 34693, 34692, 34690, 34687, 34686, 34680, 34691, 34690, 34684, 34673, 34679, 34678, 34677, 34683, 34682, 34674, 34676, 34673, 34684, 34681, 34679, 34692, 34689, 34687, 34676, 34682, 34679, 34690, 34687, 34681, 34695, 34697, 34696, 34699, 34701, 34700, 34703, 34701, 34699, 34707, 34706, 34704, 34711, 34710, 34708, 34715, 34713, 34712, 34719, 34718, 34716, 34723, 34722, 34720, 34713, 34703, 34702, 34717, 34716, 34706, 34721, 34720, 34710, 34697, 34705, 34704, 34701, 34709, 34708, 34711, 34709, 34701, 34723, 34721, 34713, 34698, 34700, 34697, 34710, 34707, 34705, 34722, 34719, 34717, 34700, 34708, 34705, 34720, 34717, 34707, 34721, 34711, 34703, 34727, 34725, 34724, 34733, 34732, 34730, 34737, 34736, 34734, 34739, 34741, 34740, 34745, 34744, 34742, 34729, 34739, 34738, 34737, 34743, 34742, 34733, 34727, 34726, 34742, 34744, 34741, 34729, 34728, 34725, 34736, 34733, 34731, 34736, 34729, 34727, 34736, 34742, 34739, 34749, 34747, 34746, 34753, 34752, 34750, 34757, 34756, 34754, 34761, 34759, 34758, 34765, 34764, 34762, 34769, 34768, 34766, 34759, 34749, 34748,
34763, 34762, 34752, 34767, 34766, 34756, 34757, 34755, 34747, 34769, 34767, 34759, 34756, 34753, 34751, 34768, 34765, 34763, 34766, 34763, 34753, 34767, 34757, 34749, 34731, 34730, 34688, 34735, 34734, 34692, 34734, 34731, 34689, 34747, 34715, 34714, 34751, 34750, 34718, 34755, 34754, 34722, 34754, 34751, 34719, 34755, 34723, 34715, 34694, 34696, 34677, 34706, 34685, 34683, 34718, 34693, 34691, 34696, 34704, 34683, 34716, 34691, 34685, 34752, 34737, 34735, 34764, 34745, 34743, 34762, 34743, 34737, 34750, 34735, 34693, 34771, 34773, 34772, 34777, 34776, 34774, 34779, 34781, 34780, 34785, 34784, 34782, 34773, 34779, 34778, 34783, 34782, 34776, 34774, 34776, 34773, 34782, 34784, 34781, 34776, 34782, 34779, 34789, 34787, 34786, 34793, 34792, 34790, 34797, 34796, 34794, 34801, 34799, 34798, 34805, 34804, 34802, 34809, 34808, 34806, 34799, 34789, 34788, 34803, 34802, 34792, 34807, 34806, 34796, 34797, 34795, 34787, 34809, 34807, 34799, 34796, 34793, 34791, 34808, 34805, 34803, 34806, 34803, 34793, 34807, 34797, 34789, 34811, 34813, 34812, 34817, 34816, 34814, 34819, 34821, 34820, 34825, 34824, 34822, 34813, 34819, 34818, 34823, 34822, 34816, 34814, 34816, 34813, 34822, 34824, 34821, 34816, 34822, 34819, 34829, 34827, 34826, 34833, 34832, 34830,
34837, 34836, 34834, 34841, 34839, 34838, 34845, 34844, 34842, 34849, 34848, 34846, 34839, 34829, 34828, 34843, 34842, 34832, 34847, 34846, 34836, 34837, 34835, 34827, 34849, 34847, 34839, 34836, 34833, 34831, 34848, 34845, 34843, 34846, 34843, 34833, 34847, 34837, 34829, 34781, 34811, 34810, 34815, 34814, 34784, 34784, 34814, 34811, 34827, 34801, 34800, 34831, 34830, 34804, 34835, 34834, 34808, 34834, 34831, 34805, 34835, 34809, 34801, 34792, 34777, 34775, 34804, 34785, 34783, 34802, 34783, 34777, 34832, 34817, 34815, 34844, 34825, 34823, 34842, 34823, 34817, 34830, 34815, 34785, 34741, 34771, 34770, 34775, 34774, 34744, 34744, 34774, 34771, 34787, 34761, 34760, 34791, 34790, 34764, 34795, 34794, 34768, 34794, 34791, 34765, 34795, 34769, 34761, 34790, 34775, 34745, 34851, 34853, 34852, 34857, 34856, 34854, 34861, 34860, 34858, 34863, 34865, 34864, 34869, 34868, 34866, 34873, 34872, 34870, 34853, 34863, 34862, 34867, 34866, 34856, 34871, 34870, 34860, 34854, 34856, 34853, 34866, 34868, 34865, 34860, 34857, 34855, 34872, 34869, 34867, 34870, 34867, 34857, 34856, 34866, 34863, 34875, 34877, 34876, 34881, 34880, 34878, 34885, 34884, 34882, 34887, 34889, 34888, 34893, 34892, 34890, 34897, 34896, 34894, 34877, 34887, 34886, 34891, 34890, 34880,
34895, 34894, 34884, 34878, 34880, 34877, 34890, 34892, 34889, 34884, 34881, 34879, 34896, 34893, 34891, 34894, 34891, 34881, 34880, 34890, 34887, 34865, 34875, 34874, 34879, 34878, 34868, 34883, 34882, 34872, 34882, 34879, 34869, 34868, 34878, 34875, 34901, 34899, 34898, 34905, 34904, 34902, 34911, 34909, 34908, 34915, 34914, 34912, 34909, 34901, 34900, 34913, 34912, 34904, 34907, 34906, 34899, 34917, 34916, 34909, 34907, 34905, 34903, 34917, 34915, 34913, 34916, 34913, 34905, 34916, 34907, 34901, 34921, 34919, 34918, 34925, 34924, 34922, 34931, 34929, 34928, 34935, 34934, 34932, 34929, 34921, 34920, 34933, 34932, 34924, 34927, 34926, 34919, 34937, 34936, 34929, 34927, 34925, 34923, 34937, 34935, 34933, 34936, 34933, 34925, 34936, 34927, 34921, 34919, 34911, 34910, 34923, 34922, 34914, 34926, 34923, 34915, 34926, 34917, 34911, 34939, 34941, 34940, 34945, 34944, 34942, 34949, 34948, 34946, 34953, 34951, 34950, 34957, 34955, 34954, 34941, 34952, 34950, 34959, 34958, 34944, 34961, 34960, 34948, 34942, 34944, 34941, 34959, 34953, 34952, 34961, 34957, 34956, 34948, 34945, 34943, 34956, 34954, 34951, 34960, 34956, 34953, 34960, 34959, 34945, 34944, 34958, 34952, 34965, 34963, 34962, 34969, 34968, 34966, 34975, 34973, 34972, 34977, 34965, 34964,
34979, 34978, 34968, 34979, 34975, 34974, 34971, 34970, 34963, 34971, 34969, 34967, 34977, 34976, 34973, 34980, 34977, 34975, 34980, 34979, 34969, 34980, 34971, 34965, 34889, 34939, 34938, 34943, 34942, 34892, 34947, 34946, 34896, 34946, 34943, 34893, 34892, 34942, 34939, 34963, 34931, 34930, 34967, 34966, 34934, 34970, 34967, 34935, 34970, 34937, 34931, 34904, 34861, 34859, 34914, 34873, 34871, 34912, 34871, 34861, 34924, 34885, 34883, 34934, 34897, 34895, 34932, 34895, 34885, 34922, 34883, 34873, 34968, 34949, 34947, 34974, 34972, 34955, 34978, 34974, 34957, 34978, 34961, 34949, 34966, 34947, 34897, 34982, 34984, 34983, 34988, 34987, 34985, 34990, 34992, 34991, 34996, 34995, 34993, 34984, 34990, 34989, 34994, 34993, 34987, 34985, 34987, 34984, 34993, 34995, 34992, 34987, 34993, 34990, 35000, 34998, 34997, 35004, 35003, 35001, 35008, 35007, 35005, 35012, 35010, 35009, 35016, 35015, 35013, 35020, 35019, 35017, 35010, 35000, 34999, 35014, 35013, 35003, 35018, 35017, 35007, 35008, 35006, 34998, 35020, 35018, 35010, 35007, 35004, 35002, 35019, 35016, 35014, 35017, 35014, 35004, 35018, 35008, 35000, 35022, 35024, 35023, 35028, 35027, 35025, 35030, 35032, 35031, 35036, 35035, 35033, 35024, 35030, 35029, 35034, 35033, 35027, 35025, 35027, 35024,
35033, 35035, 35032, 35027, 35033, 35030, 35040, 35038, 35037, 35044, 35043, 35041, 35048, 35047, 35045, 35052, 35050, 35049, 35056, 35055, 35053, 35060, 35059, 35057, 35050, 35040, 35039, 35054, 35053, 35043, 35058, 35057, 35047, 35048, 35046, 35038, 35060, 35058, 35050, 35047, 35044, 35042, 35059, 35056, 35054, 35057, 35054, 35044, 35058, 35048, 35040, 34992, 35022, 35021, 35026, 35025, 34995, 34995, 35025, 35022, 35038, 35012, 35011, 35042, 35041, 35015, 35046, 35045, 35019, 35045, 35042, 35016, 35046, 35020, 35012, 35003, 34988, 34986, 35015, 34996, 34994, 35013, 34994, 34988, 35043, 35028, 35026, 35055, 35036, 35034, 35053, 35034, 35028, 35041, 35026, 34996, 35062, 35064, 35063, 35068, 35067, 35065, 35070, 35072, 35071, 35076, 35075, 35073, 35064, 35070, 35069, 35074, 35073, 35067, 35065, 35067, 35064, 35073, 35075, 35072, 35067, 35073, 35070, 35080, 35078, 35077, 35084, 35083, 35081, 35088, 35087, 35085, 35092, 35090, 35089, 35096, 35095, 35093, 35100, 35099, 35097, 35090, 35080, 35079, 35094, 35093, 35083, 35098, 35097, 35087, 35088, 35086, 35078, 35100, 35098, 35090, 35087, 35084, 35082, 35099, 35096, 35094, 35097, 35094, 35084, 35098, 35088, 35080, 35104, 35102, 35101, 35106, 35104, 35103, 35110, 35108, 35107, 35114, 35112, 35111,
35116, 35110, 35109, 35118, 35114, 35113, 35113, 35111, 35108, 35117, 35113, 35110, 35072, 35103, 35101, 35106, 35105, 35075, 35075, 35105, 35103, 35114, 35092, 35091, 35116, 35115, 35095, 35118, 35117, 35099, 35117, 35116, 35096, 35118, 35100, 35092, 35083, 35068, 35066, 35095, 35076, 35074, 35093, 35074, 35068, 35109, 35107, 35102, 35115, 35109, 35104, 35115, 35106, 35076, 35032, 35062, 35061, 35066, 35065, 35035, 35035, 35065, 35062, 35078, 35052, 35051, 35082, 35081, 35055, 35086, 35085, 35059, 35085, 35082, 35056, 35086, 35060, 35052, 35081, 35066, 35036, 34621, 34851, 34850, 34855, 34854, 34624, 34859, 34858, 34628, 34858, 34855, 34625, 34624, 34854, 34851, 34899, 34663, 34662, 34903, 34902, 34666, 34906, 34903, 34667, 34906, 34669, 34663, 34902, 34859, 34629, 34821, 34982, 34981, 34986, 34985, 34824, 34824, 34985, 34982, 34998, 34841, 34840, 35002, 35001, 34844, 35006, 35005, 34848, 35005, 35002, 34845, 35006, 34849, 34841, 35001, 34986, 34825, 34670, 34672, 34527, 34680, 34535, 34533, 34688, 34543, 34541, 34672, 34678, 34533, 34686, 34541, 34535, 34726, 34724, 34571, 34732, 34581, 34579, 34732, 34726, 34573, 34730, 34579, 34543, 35120, 35122, 35121, 35124, 35126, 35125, 35130, 35129, 35127, 35134, 35133, 35131, 35138, 35137, 35135,
35142, 35141, 35139, 35136, 35135, 35129, 35140, 35139, 35133, 35122, 35128, 35127, 35126, 35132, 35131, 35123, 35125, 35122, 35133, 35130, 35128, 35141, 35138, 35136, 35125, 35131, 35128, 35139, 35136, 35130, 35144, 35146, 35145, 35148, 35150, 35149, 35154, 35153, 35151, 35158, 35157, 35155, 35162, 35161, 35159, 35166, 35165, 35163, 35160, 35159, 35153, 35164, 35163, 35157, 35146, 35152, 35151, 35150, 35156, 35155, 35147, 35149, 35146, 35157, 35154, 35152, 35165, 35162, 35160, 35149, 35155, 35152, 35163, 35160, 35154, 35170, 35168, 35167, 35174, 35172, 35171, 35178, 35177, 35175, 35182, 35181, 35179, 35178, 35170, 35169, 35182, 35174, 35173, 35173, 35171, 35168, 35181, 35178, 35176, 35181, 35173, 35170, 35186, 35184, 35183, 35190, 35188, 35187, 35194, 35193, 35191, 35198, 35197, 35195, 35194, 35186, 35185, 35198, 35190, 35189, 35189, 35187, 35184, 35197, 35194, 35192, 35197, 35189, 35186, 35176, 35175, 35137, 35180, 35179, 35141, 35179, 35176, 35138, 35192, 35191, 35161, 35196, 35195, 35165, 35195, 35192, 35162, 35143, 35145, 35126, 35153, 35134, 35132, 35161, 35142, 35140, 35145, 35151, 35132, 35159, 35140, 35134, 35185, 35183, 35172, 35193, 35182, 35180, 35193, 35185, 35174, 35191, 35180, 35142, 35200, 35202, 35201, 35204, 35206, 35205,
35210, 35209, 35207, 35214, 35213, 35211, 35218, 35217, 35215, 35222, 35221, 35219, 35216, 35215, 35209, 35220, 35219, 35213, 35202, 35208, 35207, 35206, 35212, 35211, 35203, 35205, 35202, 35213, 35210, 35208, 35221, 35218, 35216, 35205, 35211, 35208, 35219, 35216, 35210, 35224, 35226, 35225, 35228, 35230, 35229, 35234, 35233, 35231, 35238, 35237, 35235, 35242, 35241, 35239, 35246, 35245, 35243, 35240, 35239, 35233, 35244, 35243, 35237, 35226, 35232, 35231, 35230, 35236, 35235, 35227, 35229, 35226, 35237, 35234, 35232, 35245, 35242, 35240, 35229, 35235, 35232, 35243, 35240, 35234, 35250, 35248, 35247, 35254, 35252, 35251, 35258, 35257, 35255, 35262, 35261, 35259, 35258, 35250, 35249, 35262, 35254, 35253, 35253, 35251, 35248, 35261, 35258, 35256, 35261, 35253, 35250, 35266, 35264, 35263, 35270, 35268, 35267, 35274, 35273, 35271, 35278, 35277, 35275, 35274, 35266, 35265, 35278, 35270, 35269, 35269, 35267, 35264, 35277, 35274, 35272, 35277, 35269, 35266, 35256, 35255, 35217, 35260, 35259, 35221, 35259, 35256, 35218, 35272, 35271, 35241, 35276, 35275, 35245, 35275, 35272, 35242, 35223, 35225, 35206, 35233, 35214, 35212, 35241, 35222, 35220, 35225, 35231, 35212, 35239, 35220, 35214, 35265, 35263, 35252, 35273, 35262, 35260, 35273, 35265, 35254,
35271, 35260, 35222, 35199, 35201, 35150, 35209, 35158, 35156, 35217, 35166, 35164, 35201, 35207, 35156, 35215, 35164, 35158, 35249, 35247, 35188, 35257, 35198, 35196, 35257, 35249, 35190, 35255, 35196, 35166, 35280, 35282, 35281, 35284, 35286, 35285, 35290, 35289, 35287, 35294, 35293, 35291, 35298, 35297, 35295, 35302, 35301, 35299, 35296, 35295, 35289, 35300, 35299, 35293, 35282, 35288, 35287, 35286, 35292, 35291, 35283, 35285, 35282, 35293, 35290, 35288, 35301, 35298, 35296, 35285, 35291, 35288, 35299, 35296, 35290, 35304, 35306, 35305, 35308, 35310, 35309, 35312, 35310, 35308, 35316, 35315, 35313, 35320, 35319, 35317, 35324, 35322, 35321, 35328, 35327, 35325, 35332, 35331, 35329, 35322, 35312, 35311, 35326, 35325, 35315, 35330, 35329, 35319, 35306, 35314, 35313, 35310, 35318, 35317, 35320, 35318, 35310, 35332, 35330, 35322, 35307, 35309, 35306, 35319, 35316, 35314, 35331, 35328, 35326, 35309, 35317, 35314, 35329, 35326, 35316, 35330, 35320, 35312, 35336, 35334, 35333, 35340, 35338, 35337, 35344, 35343, 35341, 35348, 35347, 35345, 35344, 35336, 35335, 35348, 35340, 35339, 35339, 35337, 35334, 35347, 35344, 35342, 35347, 35339, 35336, 35352, 35350, 35349, 35358, 35356, 35355, 35358, 35354, 35353, 35362, 35361, 35359, 35366, 35365, 35363,
35362, 35352, 35351, 35366, 35358, 35357, 35366, 35364, 35354, 35357, 35355, 35350, 35365, 35362, 35360, 35365, 35357, 35352, 35342, 35341, 35297, 35346, 35345, 35301, 35345, 35342, 35298, 35354, 35324, 35323, 35360, 35359, 35327, 35364, 35363, 35331, 35363, 35360, 35328, 35364, 35332, 35324, 35303, 35305, 35286, 35315, 35294, 35292, 35327, 35302, 35300, 35305, 35313, 35292, 35325, 35300, 35294, 35351, 35349, 35338, 35361, 35348, 35346, 35361, 35351, 35340, 35359, 35346, 35302, 35279, 35281, 35230, 35289, 35238, 35236, 35297, 35246, 35244, 35281, 35287, 35236, 35295, 35244, 35238, 35335, 35333, 35268, 35343, 35278, 35276, 35343, 35335, 35270, 35341, 35276, 35246, 35368, 35370, 35369, 35372, 35374, 35373, 35382, 35381, 35379, 35386, 35385, 35383, 35380, 35379, 35375, 35384, 35383, 35377, 35369, 35370, 35376, 35373, 35374, 35378, 35371, 35373, 35370, 35385, 35382, 35380, 35370, 35373, 35377, 35383, 35380, 35376, 35388, 35390, 35389, 35392, 35394, 35393, 35402, 35401, 35399, 35406, 35405, 35403, 35396, 35400, 35399, 35398, 35404, 35403, 35389, 35390, 35396, 35393, 35394, 35398, 35391, 35393, 35390, 35405, 35402, 35400, 35390, 35393, 35397, 35397, 35403, 35400, 35410, 35409, 35407, 35414, 35413, 35411, 35418, 35416, 35415, 35422, 35420, 35419,
35424, 35421, 35419, 35428, 35427, 35425, 35426, 35425, 35409, 35430, 35429, 35413, 35427, 35428, 35418, 35430, 35422, 35421, 35424, 35431, 35429, 35413, 35410, 35408, 35424, 35423, 35416, 35431, 35428, 35426, 35428, 35431, 35424, 35413, 35429, 35426, 35435, 35434, 35432, 35439, 35438, 35436, 35443, 35441, 35440, 35447, 35445, 35444, 35435, 35449, 35448, 35439, 35451, 35450, 35448, 35449, 35443, 35450, 35451, 35447, 35438, 35435, 35433, 35446, 35444, 35441, 35449, 35450, 35446, 35438, 35450, 35449, 35408, 35407, 35381, 35412, 35411, 35385, 35411, 35408, 35382, 35433, 35432, 35401, 35437, 35436, 35405, 35436, 35433, 35402, 35387, 35389, 35374, 35401, 35386, 35384, 35374, 35389, 35395, 35399, 35384, 35378, 35434, 35414, 35412, 35442, 35440, 35420, 35430, 35448, 35442, 35448, 35430, 35414, 35432, 35412, 35386, 35453, 35455, 35454, 35457, 35459, 35458, 35467, 35466, 35464, 35471, 35470, 35468, 35465, 35464, 35460, 35469, 35468, 35462, 35455, 35461, 35460, 35459, 35463, 35462, 35456, 35458, 35455, 35470, 35467, 35465, 35458, 35462, 35461, 35468, 35465, 35461, 35473, 35475, 35474, 35477, 35479, 35478, 35487, 35486, 35484, 35491, 35490, 35488, 35485, 35484, 35480, 35489, 35488, 35482, 35475, 35481, 35480, 35479, 35483, 35482, 35476, 35478, 35475,
35490, 35487, 35485, 35478, 35482, 35481, 35488, 35485, 35481, 35495, 35494, 35492, 35499, 35498, 35496, 35503, 35501, 35500, 35507, 35505, 35504, 35495, 35509, 35508, 35511, 35510, 35498, 35508, 35509, 35503, 35511, 35507, 35506, 35498, 35495, 35493, 35506, 35504, 35501, 35510, 35506, 35503, 35510, 35509, 35495, 35515, 35514, 35512, 35519, 35518, 35516, 35523, 35521, 35520, 35527, 35525, 35524, 35529, 35528, 35514, 35531, 35530, 35518, 35529, 35523, 35522, 35531, 35527, 35526, 35518, 35515, 35513, 35526, 35524, 35521, 35530, 35526, 35523, 35530, 35529, 35515, 35493, 35492, 35466, 35497, 35496, 35470, 35496, 35493, 35467, 35513, 35512, 35486, 35517, 35516, 35490, 35516, 35513, 35487, 35472, 35474, 35459, 35486, 35471, 35469, 35474, 35480, 35463, 35484, 35469, 35463, 35514, 35499, 35497, 35522, 35520, 35505, 35528, 35522, 35507, 35528, 35511, 35499, 35512, 35497, 35471, 35533, 35417, 35415, 35533, 35534, 35427, 35452, 35454, 35394, 35466, 35406, 35404, 35454, 35460, 35398, 35464, 35404, 35398, 35494, 35439, 35437, 35502, 35500, 35445, 35451, 35508, 35502, 35508, 35451, 35439, 35492, 35437, 35406, 35536, 35538, 35537, 35540, 35542, 35541, 35550, 35549, 35547, 35554, 35553, 35551, 35548, 35547, 35543, 35552, 35551, 35545, 35538, 35544, 35543,
35542, 35546, 35545, 35539, 35541, 35538, 35553, 35550, 35548, 35541, 35545, 35544, 35551, 35548, 35544, 35556, 35558, 35557, 35560, 35562, 35561, 35570, 35568, 35567, 35574, 35573, 35571, 35578, 35577, 35575, 35568, 35562, 35560, 35572, 35571, 35563, 35576, 35575, 35565, 35558, 35564, 35563, 35562, 35566, 35565, 35578, 35576, 35568, 35559, 35561, 35558, 35577, 35574, 35572, 35561, 35565, 35564, 35575, 35572, 35564, 35576, 35566, 35562, 35582, 35581, 35579, 35586, 35585, 35583, 35590, 35588, 35587, 35594, 35592, 35591, 35596, 35595, 35581, 35598, 35597, 35585, 35596, 35590, 35589, 35598, 35594, 35593, 35585, 35582, 35580, 35593, 35591, 35588, 35597, 35593, 35590, 35597, 35596, 35582, 35602, 35600, 35599, 35606, 35605, 35603, 35610, 35609, 35607, 35614, 35612, 35611, 35618, 35616, 35615, 35618, 35602, 35601, 35620, 35619, 35605, 35622, 35621, 35609, 35620, 35614, 35613, 35610, 35608, 35600, 35622, 35618, 35617, 35609, 35606, 35604, 35617, 35615, 35612, 35621, 35617, 35614, 35621, 35620, 35606, 35622, 35610, 35602, 35580, 35579, 35549, 35584, 35583, 35553, 35583, 35580, 35550, 35600, 35570, 35569, 35604, 35603, 35573, 35608, 35607, 35577, 35607, 35604, 35574, 35608, 35578, 35570, 35555, 35557, 35542, 35573, 35554, 35552, 35557, 35563, 35546,
35571, 35552, 35546, 35605, 35586, 35584, 35613, 35611, 35592, 35619, 35613, 35594, 35619, 35598, 35586, 35603, 35584, 35554, 35624, 35626, 35625, 35628, 35626, 35624, 35630, 35632, 35631, 35634, 35636, 35635, 35626, 35634, 35633, 35636, 35634, 35626, 35632, 35638, 35637, 35629, 35631, 35628, 35628, 35631, 35637, 35640, 35642, 35641, 35644, 35646, 35645, 35642, 35648, 35647, 35646, 35650, 35649, 35643, 35645, 35642, 35645, 35649, 35648, 35639, 35641, 35632, 35641, 35647, 35638, 35652, 35654, 35653, 35656, 35658, 35657, 35654, 35660, 35659, 35658, 35662, 35661, 35655, 35657, 35654, 35657, 35661, 35660, 35664, 35666, 35665, 35668, 35670, 35669, 35666, 35672, 35671, 35670, 35674, 35673, 35667, 35669, 35666, 35669, 35673, 35672, 35663, 35665, 35658, 35665, 35671, 35662, 35651, 35653, 35646, 35653, 35659, 35650, 35676, 35678, 35677, 35680, 35682, 35681, 35678, 35684, 35683, 35682, 35686, 35685, 35679, 35681, 35678, 35681, 35685, 35684, 35688, 35690, 35689, 35692, 35694, 35693, 35690, 35696, 35695, 35694, 35698, 35697, 35691, 35693, 35690, 35693, 35697, 35696, 35687, 35689, 35682, 35689, 35695, 35686, 35535, 35537, 35479, 35549, 35491, 35489, 35537, 35543, 35483, 35547, 35489, 35483, 35581, 35519, 35517, 35589, 35587, 35525, 35595, 35589, 35527,
35595, 35531, 35519, 35579, 35517, 35491, 35675, 35677, 35670, 35677, 35683, 35674, 35702, 35701, 35699, 35706, 35705, 35703, 35710, 35709, 35707, 35714, 35713, 35711, 35708, 35707, 35701, 35712, 35711, 35705, 35705, 35702, 35700, 35713, 35710, 35708, 35711, 35708, 35702, 35718, 35717, 35715, 35722, 35721, 35719, 35726, 35725, 35723, 35730, 35729, 35727, 35724, 35723, 35717, 35728, 35727, 35721, 35721, 35718, 35716, 35729, 35726, 35724, 35727, 35724, 35718, 35734, 35732, 35731, 35738, 35736, 35735, 35738, 35740, 35739, 35742, 35734, 35733, 35743, 35744, 35738, 35744, 35745, 35740, 35737, 35735, 35732, 35743, 35737, 35734, 35749, 35747, 35746, 35753, 35751, 35750, 35755, 35757, 35756, 35761, 35760, 35758, 35757, 35749, 35748, 35761, 35753, 35752, 35752, 35750, 35747, 35758, 35760, 35757, 35760, 35752, 35749, 35710, 35742, 35741, 35714, 35744, 35743, 35713, 35743, 35742, 35755, 35754, 35725, 35759, 35758, 35729, 35758, 35755, 35726, 35717, 35706, 35704, 35725, 35714, 35712, 35723, 35712, 35706, 35748, 35746, 35739, 35754, 35756, 35745, 35745, 35756, 35748, 35754, 35744, 35714, 35765, 35764, 35762, 35769, 35768, 35766, 35773, 35772, 35770, 35777, 35776, 35774, 35771, 35770, 35764, 35775, 35774, 35768, 35768, 35765, 35763, 35776, 35773, 35771,
35774, 35771, 35765, 35781, 35780, 35778, 35785, 35784, 35782, 35789, 35788, 35786, 35793, 35792, 35790, 35787, 35786, 35780, 35791, 35790, 35784, 35784, 35781, 35779, 35792, 35789, 35787, 35790, 35787, 35781, 35797, 35795, 35794, 35801, 35799, 35798, 35803, 35805, 35804, 35809, 35808, 35806, 35805, 35797, 35796, 35809, 35801, 35800, 35800, 35798, 35795, 35808, 35805, 35803, 35808, 35800, 35797, 35813, 35811, 35810, 35817, 35815, 35814, 35821, 35820, 35818, 35825, 35824, 35822, 35821, 35813, 35812, 35825, 35817, 35816, 35816, 35814, 35811, 35824, 35821, 35819, 35824, 35816, 35813, 35803, 35802, 35772, 35807, 35806, 35776, 35806, 35803, 35773, 35819, 35818, 35788, 35823, 35822, 35792, 35822, 35819, 35789, 35780, 35769, 35767, 35788, 35777, 35775, 35786, 35775, 35769, 35812, 35810, 35799, 35820, 35809, 35807, 35820, 35812, 35801, 35818, 35807, 35777, 35827, 35829, 35828, 35831, 35833, 35832, 35841, 35840, 35838, 35845, 35844, 35842, 35839, 35838, 35834, 35843, 35842, 35836, 35829, 35835, 35834, 35833, 35837, 35836, 35830, 35832, 35829, 35844, 35841, 35839, 35832, 35836, 35835, 35842, 35839, 35835, 35847, 35849, 35848, 35851, 35853, 35852, 35861, 35860, 35858, 35865, 35864, 35862, 35859, 35858, 35854, 35863, 35862, 35856, 35849, 35855, 35854,
35853, 35857, 35856, 35850, 35852, 35849, 35864, 35861, 35859, 35852, 35856, 35855, 35862, 35859, 35855, 35846, 35848, 35833, 35860, 35845, 35843, 35848, 35854, 35837, 35858, 35843, 35837, 35867, 35869, 35868, 35871, 35873, 35872, 35881, 35880, 35878, 35885, 35884, 35882, 35879, 35878, 35874, 35883, 35882, 35876, 35869, 35875, 35874, 35873, 35877, 35876, 35870, 35872, 35869, 35884, 35881, 35879, 35872, 35876, 35875, 35882, 35879, 35875, 35887, 35889, 35888, 35891, 35893, 35892, 35901, 35900, 35898, 35905, 35904, 35902, 35899, 35898, 35894, 35903, 35902, 35896, 35889, 35895, 35894, 35893, 35897, 35896, 35890, 35892, 35889, 35904, 35901, 35899, 35892, 35896, 35895, 35902, 35899, 35895, 35886, 35888, 35873, 35900, 35885, 35883, 35888, 35894, 35877, 35898, 35883, 35877, 35764, 35722, 35720, 35772, 35730, 35728, 35770, 35728, 35722, 35796, 35794, 35751, 35802, 35804, 35761, 35804, 35796, 35753, 35802, 35759, 35730, 35866, 35868, 35853, 35880, 35865, 35863, 35868, 35874, 35857, 35878, 35863, 35857, 35909, 35908, 35906, 35913, 35912, 35910, 35917, 35916, 35914, 35921, 35920, 35918, 35915, 35914, 35908, 35919, 35918, 35912, 35912, 35909, 35907, 35920, 35917, 35915, 35918, 35915, 35909, 35925, 35923, 35922, 35929, 35928, 35926, 35933, 35932, 35930,
35937, 35935, 35934, 35941, 35940, 35938, 35945, 35944, 35942, 35935, 35925, 35924, 35939, 35938, 35928, 35943, 35942, 35932, 35933, 35931, 35923, 35945, 35943, 35935, 35932, 35929, 35927, 35944, 35941, 35939, 35942, 35939, 35929, 35943, 35933, 35925, 35949, 35947, 35946, 35953, 35951, 35950, 35957, 35956, 35954, 35961, 35960, 35958, 35957, 35949, 35948, 35961, 35953, 35952, 35952, 35950, 35947, 35960, 35957, 35955, 35960, 35952, 35949, 35965, 35963, 35962, 35971, 35969, 35968, 35971, 35967, 35966, 35975, 35974, 35972, 35979, 35978, 35976, 35975, 35965, 35964, 35979, 35971, 35970, 35979, 35977, 35967, 35970, 35968, 35963, 35978, 35975, 35973, 35978, 35970, 35965, 35955, 35954, 35916, 35959, 35958, 35920, 35958, 35955, 35917, 35967, 35937, 35936, 35973, 35972, 35940, 35977, 35976, 35944, 35976, 35973, 35941, 35977, 35945, 35937, 35928, 35913, 35911, 35940, 35921, 35919, 35938, 35919, 35913, 35964, 35962, 35951, 35974, 35961, 35959, 35974, 35964, 35953, 35972, 35959, 35921, 35981, 35983, 35982, 35985, 35987, 35986, 35995, 35994, 35992, 35999, 35998, 35996, 35993, 35992, 35988, 35997, 35996, 35990, 35983, 35989, 35988, 35987, 35991, 35990, 35984, 35986, 35983, 35998, 35995, 35993, 35986, 35990, 35989, 35996, 35993, 35989, 36001, 36003, 36002,
36005, 36007, 36006, 36015, 36013, 36012, 36019, 36018, 36016, 36023, 36022, 36020, 36013, 36007, 36005, 36017, 36016, 36008, 36021, 36020, 36010, 36003, 36009, 36008, 36007, 36011, 36010, 36023, 36021, 36013, 36004, 36006, 36003, 36022, 36019, 36017, 36006, 36010, 36009, 36020, 36017, 36009, 36021, 36011, 36007, 36000, 36002, 35987, 36018, 35999, 35997, 36002, 36008, 35991, 36016, 35997, 35991, 36027, 36026, 36024, 36031, 36030, 36028, 36035, 36033, 36032, 36039, 36037, 36036, 36041, 36040, 36026, 36043, 36042, 36030, 36041, 36035, 36034, 36043, 36039, 36038, 36030, 36027, 36025, 36038, 36036, 36033, 36042, 36038, 36035, 36042, 36041, 36027, 36047, 36046, 36044, 36051, 36050, 36048, 36055, 36053, 36052, 36059, 36057, 36056, 36061, 36060, 36046, 36063, 36062, 36050, 36061, 36055, 36054, 36063, 36059, 36058, 36050, 36047, 36045, 36058, 36056, 36053, 36062, 36058, 36055, 36062, 36061, 36047, 36046, 36031, 36029, 36054, 36052, 36037, 36060, 36054, 36039, 36060, 36043, 36031, 36067, 36066, 36064, 36071, 36070, 36068, 36075, 36073, 36072, 36079, 36077, 36076, 36081, 36080, 36066, 36083, 36082, 36070, 36081, 36075, 36074, 36083, 36079, 36078, 36070, 36067, 36065, 36078, 36076, 36073, 36082, 36078, 36075, 36082, 36081, 36067, 36087, 36086, 36084,
36091, 36090, 36088, 36095, 36093, 36092, 36099, 36097, 36096, 36101, 36100, 36086, 36103, 36102, 36090, 36101, 36095, 36094, 36103, 36099, 36098, 36090, 36087, 36085, 36098, 36096, 36093, 36102, 36098, 36095, 36102, 36101, 36087, 36086, 36071, 36069, 36094, 36092, 36077, 36100, 36094, 36079, 36100, 36083, 36071, 36105, 36107, 36106, 36109, 36111, 36110, 36107, 36113, 36112, 36111, 36115, 36114, 36108, 36110, 36107, 36110, 36114, 36113, 36117, 36119, 36118, 36121, 36123, 36122, 36119, 36125, 36124, 36123, 36127, 36126, 36120, 36122, 36119, 36122, 36126, 36125, 36131, 36130, 36128, 36135, 36134, 36132, 36139, 36138, 36136, 36143, 36142, 36140, 36137, 36136, 36130, 36141, 36140, 36134, 36134, 36131, 36129, 36142, 36139, 36137, 36140, 36137, 36131, 36147, 36146, 36144, 36151, 36150, 36148, 36155, 36154, 36152, 36159, 36158, 36156, 36153, 36152, 36146, 36157, 36156, 36150, 36150, 36147, 36145, 36158, 36155, 36153, 36156, 36153, 36147, 36129, 36128, 36112, 36133, 36132, 36114, 36132, 36129, 36113, 36145, 36144, 36124, 36149, 36148, 36126, 36148, 36145, 36125, 36116, 36118, 36111, 36118, 36124, 36115, 36146, 36135, 36133, 36154, 36143, 36141, 36152, 36141, 36135, 36144, 36133, 36115, 36161, 36163, 36162, 36165, 36167, 36166, 36163, 36169, 36168,
36167, 36171, 36170, 36164, 36166, 36163, 36166, 36170, 36169, 36173, 36175, 36174, 36177, 36179, 36178, 36175, 36181, 36180, 36179, 36183, 36182, 36176, 36178, 36175, 36178, 36182, 36181, 36187, 36186, 36184, 36191, 36190, 36188, 36195, 36194, 36192, 36199, 36198, 36196, 36193, 36192, 36186, 36197, 36196, 36190, 36190, 36187, 36185, 36198, 36195, 36193, 36196, 36193, 36187, 36203, 36202, 36200, 36207, 36206, 36204, 36211, 36210, 36208, 36215, 36214, 36212, 36209, 36208, 36202, 36213, 36212, 36206, 36206, 36203, 36201, 36214, 36211, 36209, 36212, 36209, 36203, 36185, 36184, 36168, 36189, 36188, 36170, 36188, 36185, 36169, 36201, 36200, 36180, 36205, 36204, 36182, 36204, 36201, 36181, 36172, 36174, 36167, 36174, 36180, 36171, 36202, 36191, 36189, 36210, 36199, 36197, 36208, 36197, 36191, 36200, 36189, 36171, 36066, 36051, 36049, 36074, 36072, 36057, 36080, 36074, 36059, 36080, 36063, 36051, 36160, 36162, 36123, 36162, 36168, 36127, 36186, 36151, 36149, 36194, 36159, 36157, 36192, 36157, 36151, 36184, 36149, 36127, 36219, 36218, 36216, 36223, 36222, 36220, 36227, 36225, 36224, 36231, 36229, 36228, 36233, 36232, 36218, 36235, 36234, 36222, 36233, 36227, 36226, 36235, 36231, 36230, 36222, 36219, 36217, 36230, 36228, 36225, 36234, 36230, 36227,
36234, 36233, 36219, 36239, 36237, 36236, 36243, 36242, 36240, 36247, 36246, 36244, 36251, 36249, 36248, 36255, 36253, 36252, 36255, 36239, 36238, 36257, 36256, 36242, 36259, 36258, 36246, 36257, 36251, 36250, 36247, 36245, 36237, 36259, 36255, 36254, 36246, 36243, 36241, 36254, 36252, 36249, 36258, 36254, 36251, 36258, 36257, 36243, 36259, 36247, 36239, 36242, 36223, 36221, 36250, 36248, 36229, 36256, 36250, 36231, 36256, 36235, 36223, 36261, 36263, 36262, 36265, 36267, 36266, 36263, 36269, 36268, 36267, 36271, 36270, 36264, 36266, 36263, 36266, 36270, 36269, 36273, 36275, 36274, 36277, 36279, 36278, 36275, 36281, 36280, 36279, 36283, 36282, 36276, 36278, 36275, 36278, 36282, 36281, 36287, 36286, 36284, 36291, 36290, 36288, 36295, 36294, 36292, 36299, 36298, 36296, 36293, 36292, 36286, 36297, 36296, 36290, 36290, 36287, 36285, 36298, 36295, 36293, 36296, 36293, 36287, 36303, 36301, 36300, 36307, 36306, 36304, 36311, 36310, 36308, 36315, 36313, 36312, 36319, 36318, 36316, 36323, 36322, 36320, 36313, 36303, 36302, 36317, 36316, 36306, 36321, 36320, 36310, 36311, 36309, 36301, 36323, 36321, 36313, 36310, 36307, 36305, 36322, 36319, 36317, 36320, 36317, 36307, 36321, 36311, 36303, 36285, 36284, 36268, 36289, 36288, 36270, 36288, 36285, 36269,
36301, 36279, 36277, 36305, 36304, 36280, 36309, 36308, 36282, 36308, 36305, 36281, 36309, 36283, 36279, 36272, 36274, 36267, 36274, 36280, 36271, 36306, 36291, 36289, 36318, 36299, 36297, 36316, 36297, 36291, 36304, 36289, 36271, 36025, 36024, 35840, 36029, 36028, 35844, 36028, 36025, 35841, 36045, 36044, 35860, 36049, 36048, 35864, 36048, 36045, 35861, 36044, 36029, 35845, 36065, 36064, 35880, 36069, 36068, 35884, 36068, 36065, 35881, 36085, 36084, 35900, 36089, 36088, 35904, 36088, 36085, 35901, 36084, 36069, 35885, 36064, 36049, 35865, 36217, 36216, 35994, 36221, 36220, 35998, 36220, 36217, 35995, 36237, 36015, 36014, 36241, 36240, 36018, 36245, 36244, 36022, 36244, 36241, 36019, 36245, 36023, 36015, 36240, 36221, 35999, 35908, 35785, 35783, 35916, 35793, 35791, 35914, 35791, 35785, 35948, 35946, 35815, 35956, 35825, 35823, 35956, 35948, 35817, 35954, 35823, 35793, 35980, 35982, 35893, 35994, 35905, 35903, 35982, 35988, 35897, 35992, 35903, 35897, 36218, 36091, 36089, 36226, 36224, 36097, 36232, 36226, 36099, 36232, 36103, 36091, 36260, 36262, 36179, 36262, 36268, 36183, 36286, 36207, 36205, 36294, 36215, 36213, 36292, 36213, 36207, 36284, 36205, 36183, 36216, 36089, 35905, 36327, 36325, 36324, 36331, 36329, 36328, 36335, 36334, 36332,
36339, 36338, 36336, 36335, 36327, 36326, 36339, 36331, 36330, 36330, 36328, 36325, 36338, 36335, 36333, 36338, 36330, 36327, 36343, 36341, 36340, 36347, 36345, 36344, 36351, 36350, 36348, 36355, 36354, 36352, 36351, 36343, 36342, 36355, 36347, 36346, 36346, 36344, 36341, 36354, 36351, 36349, 36354, 36346, 36343, 36342, 36340, 36329, 36350, 36339, 36337, 36350, 36342, 36331, 36359, 36357, 36356, 36363, 36361, 36360, 36367, 36366, 36364, 36371, 36370, 36368, 36367, 36359, 36358, 36371, 36363, 36362, 36362, 36360, 36357, 36370, 36367, 36365, 36370, 36362, 36359, 36375, 36373, 36372, 36379, 36377, 36376, 36383, 36382, 36380, 36387, 36386, 36384, 36383, 36375, 36374, 36387, 36379, 36378, 36378, 36376, 36373, 36386, 36383, 36381, 36386, 36378, 36375, 36374, 36372, 36361, 36382, 36371, 36369, 36382, 36374, 36363, 36358, 36356, 36345, 36366, 36355, 36353, 36366, 36358, 36347, 36391, 36389, 36388, 36397, 36396, 36394, 36401, 36400, 36398, 36403, 36405, 36404, 36409, 36408, 36406, 36393, 36403, 36402, 36401, 36407, 36406, 36397, 36391, 36390, 36406, 36408, 36405, 36393, 36392, 36389, 36400, 36397, 36395, 36400, 36393, 36391, 36400, 36406, 36403, 36413, 36411, 36410, 36417, 36416, 36414, 36421, 36420, 36418, 36425, 36423, 36422, 36429, 36428, 36426,
36433, 36432, 36430, 36423, 36413, 36412, 36427, 36426, 36416, 36431, 36430, 36420, 36421, 36419, 36411, 36433, 36431, 36423, 36420, 36417, 36415, 36432, 36429, 36427, 36430, 36427, 36417, 36431, 36421, 36413, 36435, 36437, 36436, 36441, 36440, 36438, 36443, 36445, 36444, 36449, 36448, 36446, 36437, 36443, 36442, 36447, 36446, 36440, 36438, 36440, 36437, 36446, 36448, 36445, 36440, 36446, 36443, 36453, 36451, 36450, 36457, 36456, 36454, 36461, 36460, 36458, 36465, 36463, 36462, 36469, 36468, 36466, 36473, 36472, 36470, 36463, 36453, 36452, 36467, 36466, 36456, 36471, 36470, 36460, 36461, 36459, 36451, 36473, 36471, 36463, 36460, 36457, 36455, 36472, 36469, 36467, 36470, 36467, 36457, 36471, 36461, 36453, 36405, 36435, 36434, 36439, 36438, 36408, 36408, 36438, 36435, 36451, 36425, 36424, 36455, 36454, 36428, 36459, 36458, 36432, 36458, 36455, 36429, 36459, 36433, 36425, 36416, 36401, 36399, 36428, 36409, 36407, 36426, 36407, 36401, 36456, 36441, 36439, 36468, 36449, 36447, 36466, 36447, 36441, 36454, 36439, 36409, 36475, 36477, 36476, 36481, 36480, 36478, 36483, 36485, 36484, 36489, 36488, 36486, 36477, 36483, 36482, 36481, 36487, 36486, 36478, 36480, 36477, 36483, 36486, 36488, 36477, 36480, 36486, 36493, 36491, 36490, 36497, 36496, 36494,
36501, 36500, 36498, 36505, 36503, 36502, 36509, 36508, 36506, 36513, 36512, 36510, 36503, 36493, 36492, 36507, 36506, 36496, 36511, 36510, 36500, 36501, 36499, 36491, 36513, 36511, 36503, 36500, 36497, 36495, 36512, 36509, 36507, 36510, 36507, 36497, 36511, 36501, 36493, 36515, 36517, 36516, 36519, 36521, 36520, 36525, 36527, 36526, 36532, 36531, 36529, 36517, 36525, 36524, 36530, 36529, 36520, 36515, 36518, 36520, 36529, 36531, 36527, 36526, 36527, 36523, 36523, 36527, 36531, 36517, 36520, 36529, 36536, 36534, 36533, 36540, 36539, 36537, 36544, 36543, 36541, 36548, 36546, 36545, 36552, 36551, 36549, 36554, 36556, 36555, 36546, 36536, 36535, 36550, 36549, 36539, 36554, 36553, 36543, 36544, 36542, 36534, 36556, 36554, 36546, 36543, 36540, 36538, 36555, 36552, 36550, 36553, 36550, 36540, 36554, 36544, 36536, 36485, 36515, 36514, 36489, 36519, 36518, 36485, 36488, 36518, 36534, 36505, 36504, 36538, 36537, 36508, 36542, 36541, 36512, 36541, 36538, 36509, 36542, 36513, 36505, 36496, 36481, 36479, 36508, 36489, 36487, 36506, 36487, 36481, 36539, 36521, 36519, 36551, 36532, 36530, 36549, 36530, 36521, 36537, 36519, 36489, 36445, 36475, 36474, 36479, 36478, 36448, 36448, 36478, 36475, 36491, 36465, 36464, 36495, 36494, 36468, 36499, 36498, 36472,
36498, 36495, 36469, 36499, 36473, 36465, 36494, 36479, 36449, 36558, 36560, 36559, 36560, 36562, 36561, 36564, 36566, 36565, 36570, 36572, 36571, 36574, 36576, 36575, 36575, 36578, 36577, 36584, 36583, 36581, 36566, 36576, 36574, 36571, 36572, 36580, 36564, 36567, 36568, 36576, 36582, 36581, 36575, 36581, 36583, 36577, 36578, 36572, 36572, 36578, 36583, 36566, 36568, 36582, 36588, 36590, 36589, 36592, 36590, 36588, 36596, 36595, 36593, 36598, 36600, 36599, 36604, 36603, 36601, 36608, 36607, 36605, 36602, 36601, 36595, 36606, 36605, 36599, 36589, 36590, 36598, 36600, 36598, 36590, 36587, 36589, 36586, 36597, 36599, 36596, 36607, 36604, 36602, 36589, 36597, 36594, 36605, 36602, 36596, 36610, 36612, 36611, 36614, 36616, 36615, 36624, 36623, 36621, 36628, 36627, 36625, 36618, 36622, 36621, 36626, 36625, 36619, 36612, 36618, 36617, 36615, 36616, 36620, 36613, 36615, 36612, 36627, 36624, 36622, 36615, 36619, 36618, 36625, 36622, 36618, 36632, 36630, 36629, 36634, 36636, 36635, 36640, 36639, 36637, 36646, 36644, 36643, 36632, 36642, 36641, 36640, 36649, 36648, 36636, 36632, 36631, 36649, 36646, 36645, 36639, 36636, 36634, 36645, 36643, 36641, 36648, 36645, 36642, 36639, 36648, 36647, 36636, 36647, 36642, 36653, 36652, 36650, 36657, 36656, 36654,
36661, 36659, 36658, 36665, 36663, 36662, 36653, 36667, 36666, 36657, 36669, 36668, 36667, 36661, 36660, 36669, 36665, 36664, 36656, 36653, 36651, 36664, 36662, 36659, 36668, 36664, 36661, 36668, 36667, 36653, 36634, 36633, 36603, 36638, 36637, 36607, 36637, 36634, 36604, 36651, 36650, 36623, 36655, 36654, 36627, 36654, 36651, 36624, 36609, 36611, 36592, 36623, 36608, 36606, 36592, 36611, 36617, 36617, 36621, 36606, 36652, 36640, 36638, 36660, 36658, 36644, 36666, 36660, 36646, 36652, 36666, 36649, 36650, 36638, 36608, 36671, 36673, 36672, 36675, 36677, 36676, 36676, 36679, 36678, 36685, 36684, 36682, 36689, 36688, 36686, 36693, 36692, 36690, 36687, 36686, 36680, 36685, 36691, 36690, 36672, 36673, 36681, 36677, 36683, 36682, 36676, 36682, 36684, 36678, 36679, 36673, 36692, 36689, 36687, 36679, 36684, 36681, 36690, 36687, 36681, 36697, 36699, 36698, 36703, 36702, 36700, 36707, 36706, 36704, 36711, 36710, 36708, 36705, 36704, 36698, 36709, 36708, 36702, 36702, 36699, 36697, 36710, 36707, 36705, 36708, 36705, 36699, 36715, 36714, 36712, 36719, 36718, 36716, 36723, 36721, 36720, 36727, 36725, 36724, 36729, 36728, 36714, 36731, 36730, 36718, 36728, 36729, 36723, 36730, 36731, 36727, 36718, 36715, 36713, 36726, 36724, 36721, 36729, 36730, 36726,
36730, 36729, 36715, 36735, 36733, 36732, 36741, 36739, 36738, 36741, 36737, 36736, 36745, 36744, 36742, 36749, 36748, 36746, 36751, 36734, 36732, 36745, 36735, 36734, 36748, 36749, 36741, 36741, 36749, 36747, 36740, 36738, 36733, 36748, 36745, 36743, 36745, 36748, 36740, 36751, 36752, 36744, 36713, 36712, 36688, 36717, 36716, 36692, 36716, 36713, 36689, 36743, 36742, 36706, 36711, 36747, 36746, 36746, 36743, 36707, 36694, 36695, 36677, 36698, 36685, 36683, 36706, 36693, 36691, 36677, 36695, 36696, 36704, 36691, 36685, 36742, 36744, 36719, 36751, 36750, 36725, 36752, 36751, 36727, 36752, 36731, 36719, 36742, 36717, 36693, 36586, 36560, 36558, 36562, 36594, 36593, 36594, 36562, 36560, 36571, 36695, 36694, 36697, 36696, 36579, 36584, 36701, 36700, 36700, 36697, 36580, 36571, 36579, 36696, 36670, 36672, 36616, 36688, 36628, 36626, 36616, 36672, 36680, 36680, 36686, 36626, 36714, 36657, 36655, 36722, 36720, 36663, 36669, 36728, 36722, 36714, 36728, 36669, 36712, 36655, 36628, 36754, 36756, 36755, 36762, 36761, 36759, 36766, 36768, 36767, 36764, 36768, 36766, 36775, 36774, 36772, 36756, 36764, 36763, 36773, 36772, 36761, 36754, 36757, 36758, 36767, 36768, 36771, 36769, 36771, 36768, 36761, 36758, 36757, 36774, 36771, 36769, 36772, 36769, 36758,
36756, 36758, 36769, 36779, 36777, 36776, 36783, 36782, 36780, 36787, 36786, 36784, 36791, 36789, 36788, 36793, 36790, 36788, 36797, 36796, 36794, 36791, 36779, 36778, 36795, 36794, 36782, 36799, 36798, 36786, 36779, 36787, 36785, 36798, 36799, 36791, 36800, 36798, 36790, 36786, 36783, 36781, 36800, 36797, 36795, 36786, 36798, 36795, 36799, 36787, 36779, 36804, 36806, 36805, 36810, 36809, 36807, 36814, 36813, 36811, 36818, 36817, 36815, 36812, 36811, 36805, 36816, 36815, 36809, 36809, 36806, 36804, 36817, 36814, 36812, 36815, 36812, 36806, 36824, 36822, 36821, 36826, 36828, 36827, 36836, 36834, 36833, 36836, 36832, 36831, 36840, 36839, 36837, 36832, 36820, 36819, 36838, 36837, 36827, 36839, 36840, 36836, 36836, 36840, 36838, 36830, 36829, 36822, 36824, 36823, 36819, 36830, 36828, 36826, 36828, 36830, 36824, 36838, 36828, 36820, 36767, 36802, 36801, 36804, 36803, 36770, 36808, 36807, 36774, 36774, 36807, 36804, 36770, 36803, 36802, 36822, 36793, 36792, 36826, 36825, 36796, 36829, 36826, 36797, 36822, 36829, 36800, 36782, 36762, 36760, 36796, 36775, 36773, 36794, 36773, 36762, 36827, 36810, 36808, 36837, 36839, 36818, 36837, 36816, 36810, 36825, 36808, 36775, 36846, 36845, 36843, 36848, 36850, 36849, 36854, 36852, 36851, 36858, 36856, 36855,
36860, 36857, 36855, 36864, 36863, 36861, 36858, 36842, 36841, 36862, 36861, 36845, 36866, 36865, 36849, 36863, 36864, 36854, 36865, 36866, 36858, 36860, 36867, 36865, 36849, 36846, 36844, 36860, 36859, 36852, 36867, 36864, 36862, 36864, 36867, 36860, 36849, 36865, 36862, 36858, 36866, 36850, 36871, 36869, 36868, 36871, 36873, 36872, 36875, 36853, 36851, 36875, 36876, 36863, 36871, 36870, 36841, 36873, 36850, 36848, 36873, 36871, 36842, 36844, 36843, 36813, 36848, 36847, 36817, 36847, 36844, 36814, 36869, 36835, 36833, 36869, 36872, 36839, 36872, 36848, 36818, 36523, 36754, 36753, 36532, 36760, 36759, 36759, 36757, 36528, 36528, 36757, 36754, 36777, 36548, 36547, 36781, 36780, 36551, 36556, 36785, 36784, 36784, 36781, 36552, 36777, 36785, 36556, 36780, 36760, 36532, 36390, 36388, 36377, 36396, 36387, 36385, 36396, 36390, 36379, 36801, 36802, 36564, 36805, 36568, 36567, 36811, 36813, 36584, 36802, 36803, 36567, 36811, 36582, 36568, 36845, 36703, 36701, 36861, 36863, 36711, 36861, 36709, 36703, 36875, 36874, 36736, 36876, 36875, 36737, 36876, 36747, 36711, 36843, 36701, 36584, 36333, 36332, 36138, 36337, 36336, 36142, 36336, 36333, 36139, 36349, 36348, 36154, 36353, 36352, 36158, 36352, 36349, 36155, 36348, 36337, 36143, 36365, 36364, 36194,
36369, 36368, 36198, 36368, 36365, 36195, 36381, 36380, 36210, 36385, 36384, 36214, 36384, 36381, 36211, 36380, 36369, 36199, 36364, 36353, 36159, 36395, 36394, 36294, 36399, 36398, 36298, 36398, 36395, 36295, 36411, 36315, 36314, 36415, 36414, 36318, 36419, 36418, 36322, 36418, 36415, 36319, 36419, 36323, 36315, 36414, 36399, 36299, 36394, 36385, 36215, 36880, 36882, 36881, 36878, 36882, 36880, 36887, 36889, 36888, 36893, 36892, 36890, 36881, 36887, 36886, 36891, 36890, 36884, 36882, 36885, 36884, 36878, 36883, 36885, 36887, 36890, 36892, 36881, 36884, 36890, 36895, 36897, 36896, 36899, 36901, 36900, 36900, 36903, 36902, 36909, 36908, 36906, 36896, 36897, 36905, 36901, 36907, 36906, 36900, 36906, 36908, 36902, 36903, 36897, 36897, 36903, 36908, 36911, 36913, 36912, 36915, 36917, 36916, 36927, 36929, 36928, 36933, 36932, 36930, 36912, 36923, 36922, 36925, 36917, 36915, 36927, 36926, 36918, 36921, 36931, 36930, 36913, 36919, 36918, 36916, 36917, 36921, 36914, 36916, 36913, 36932, 36929, 36927, 36913, 36916, 36920, 36930, 36927, 36919, 36912, 36918, 36926, 36931, 36921, 36917, 36937, 36939, 36938, 36943, 36942, 36940, 36947, 36946, 36944, 36951, 36950, 36948, 36945, 36944, 36938, 36949, 36948, 36942, 36942, 36939, 36937, 36950, 36947, 36945,
36948, 36945, 36939, 36955, 36954, 36952, 36959, 36958, 36956, 36963, 36962, 36960, 36967, 36966, 36964, 36961, 36960, 36954, 36965, 36964, 36958, 36958, 36955, 36953, 36966, 36963, 36961, 36964, 36961, 36955, 36896, 36935, 36934, 36937, 36936, 36904, 36909, 36941, 36940, 36940, 36937, 36905, 36896, 36904, 36936, 36953, 36952, 36928, 36957, 36956, 36932, 36956, 36953, 36929, 36922, 36923, 36901, 36928, 36909, 36907, 36901, 36923, 36926, 36954, 36943, 36941, 36962, 36951, 36949, 36960, 36949, 36943, 36952, 36941, 36909, 36969, 36971, 36970, 36977, 36979, 36978, 36981, 36983, 36982, 36989, 36988, 36986, 36971, 36981, 36980, 36987, 36986, 36978, 36974, 36975, 36971, 36981, 36984, 36985, 36972, 36973, 36969, 36978, 36975, 36974, 36988, 36985, 36984, 36969, 36973, 36976, 36986, 36984, 36975, 36971, 36975, 36984, 36991, 36993, 36992, 36999, 36998, 36996, 37001, 37003, 37002, 37009, 37008, 37006, 36993, 37001, 37000, 37007, 37006, 36998, 36991, 36994, 36995, 37004, 37005, 37003, 36996, 36998, 36995, 37008, 37005, 37004, 37006, 37004, 36995, 36995, 37004, 37001, 36983, 36991, 36990, 36997, 36996, 36988, 36996, 36994, 36985, 36983, 36985, 36994, 37015, 37014, 37012, 37017, 37019, 37018, 37023, 37021, 37020, 37027, 37026, 37024, 37031, 37030, 37028,
37021, 37011, 37010, 37025, 37024, 37014, 37029, 37028, 37018, 37023, 37031, 37029, 37018, 37015, 37013, 37030, 37027, 37025, 37028, 37025, 37015, 37021, 37029, 37019, 37035, 37033, 37032, 37037, 37034, 37032, 37041, 37039, 37038, 37042, 37043, 37035, 37037, 37044, 37042, 37046, 37041, 37040, 37040, 37038, 37033, 37043, 37045, 37040, 37050, 37048, 37047, 37052, 37049, 37047, 37056, 37055, 37053, 37063, 37061, 37060, 37067, 37066, 37064, 37061, 37052, 37051, 37065, 37064, 37055, 37057, 37058, 37050, 37059, 37057, 37049, 37063, 37069, 37068, 37057, 37059, 37056, 37069, 37067, 37065, 37059, 37068, 37065, 37061, 37068, 37059, 37050, 37023, 37022, 37054, 37053, 37026, 37058, 37057, 37030, 37030, 37057, 37054, 37058, 37031, 37023, 37037, 37036, 37010, 37044, 37019, 37017, 37019, 37044, 37037, 36889, 36973, 36972, 36977, 36976, 36892, 36892, 36976, 36973, 37013, 37012, 36946, 37017, 37016, 36950, 37016, 37013, 36947, 37043, 37042, 36962, 36967, 37046, 37045, 37045, 37043, 36963, 36962, 37042, 37017, 36934, 36935, 36878, 36938, 36885, 36883, 36946, 36893, 36891, 36878, 36935, 36936, 36938, 36944, 36891, 37014, 36979, 36977, 37026, 36989, 36987, 37024, 36987, 36979, 37055, 36999, 36997, 37066, 37009, 37007, 37064, 37007, 36999, 37053, 36997, 36989,
37012, 36977, 36893, 37071, 37073, 37072, 37077, 37079, 37078, 37078, 37075, 37074, 37081, 37083, 37082, 37073, 37081, 37080, 37079, 37086, 37085, 37085, 37084, 37075, 37074, 37075, 37073, 37084, 37085, 37083, 37073, 37075, 37084, 37093, 37092, 37090, 37097, 37096, 37094, 37101, 37100, 37098, 37105, 37104, 37102, 37099, 37098, 37092, 37103, 37102, 37096, 37096, 37093, 37091, 37104, 37101, 37099, 37102, 37099, 37093, 37107, 37109, 37108, 37111, 37109, 37107, 37113, 37115, 37114, 37121, 37119, 37118, 37125, 37124, 37122, 37117, 37111, 37110, 37115, 37123, 37122, 37108, 37109, 37113, 37115, 37113, 37109, 37127, 37126, 37119, 37118, 37119, 37117, 37126, 37127, 37125, 37119, 37126, 37123, 37123, 37115, 37111, 37091, 37090, 37082, 37086, 37095, 37094, 37094, 37091, 37083, 37108, 37088, 37087, 37112, 37089, 37088, 37087, 37088, 37079, 37088, 37089, 37086, 37114, 37097, 37095, 37124, 37105, 37103, 37122, 37103, 37097, 37089, 37112, 37095, 37133, 37131, 37130, 37129, 37132, 37130, 37135, 37137, 37136, 37141, 37140, 37138, 37143, 37145, 37144, 37153, 37152, 37150, 37133, 37143, 37142, 37151, 37150, 37140, 37137, 37133, 37132, 37134, 37136, 37132, 37148, 37149, 37145, 37140, 37137, 37135, 37147, 37146, 37144, 37150, 37152, 37149, 37152, 37147, 37145,
37140, 37150, 37148, 37137, 37148, 37143, 37155, 37157, 37156, 37159, 37157, 37155, 37163, 37162, 37160, 37165, 37167, 37166, 37171, 37169, 37168, 37175, 37174, 37172, 37179, 37178, 37176, 37169, 37159, 37158, 37173, 37172, 37162, 37167, 37177, 37176, 37157, 37165, 37164, 37159, 37167, 37165, 37179, 37177, 37169, 37166, 37163, 37161, 37178, 37175, 37173, 37176, 37173, 37163, 37177, 37167, 37159, 37181, 37183, 37182, 37187, 37186, 37184, 37189, 37191, 37190, 37193, 37195, 37194, 37183, 37189, 37188, 37193, 37192, 37186, 37184, 37186, 37183, 37192, 37194, 37191, 37186, 37192, 37189, 37199, 37197, 37196, 37203, 37202, 37200, 37207, 37206, 37204, 37211, 37209, 37208, 37215, 37214, 37212, 37219, 37218, 37216, 37209, 37199, 37198, 37213, 37212, 37202, 37217, 37216, 37206, 37207, 37205, 37197, 37219, 37217, 37209, 37206, 37203, 37201, 37218, 37215, 37213, 37216, 37213, 37203, 37217, 37207, 37199, 37147, 37181, 37180, 37153, 37185, 37184, 37147, 37152, 37184, 37197, 37171, 37170, 37201, 37200, 37174, 37205, 37204, 37178, 37204, 37201, 37175, 37205, 37179, 37171, 37162, 37141, 37139, 37174, 37153, 37151, 37172, 37151, 37141, 37202, 37187, 37185, 37214, 37195, 37193, 37212, 37193, 37187, 37200, 37185, 37153, 37135, 37134, 37100, 37139, 37138, 37104,
37138, 37135, 37101, 37156, 37121, 37120, 37161, 37160, 37124, 37127, 37164, 37161, 37156, 37164, 37127, 37160, 37139, 37105, 37221, 37223, 37222, 37229, 37228, 37226, 37231, 37233, 37232, 37239, 37238, 37236, 37223, 37231, 37230, 37237, 37236, 37228, 37224, 37225, 37223, 37234, 37235, 37233, 37228, 37225, 37224, 37238, 37235, 37234, 37236, 37234, 37225, 37225, 37234, 37231, 37241, 37243, 37242, 37249, 37248, 37246, 37251, 37253, 37252, 37259, 37258, 37256, 37243, 37251, 37250, 37257, 37256, 37248, 37244, 37245, 37243, 37251, 37254, 37255, 37248, 37245, 37244, 37256, 37258, 37255, 37256, 37254, 37245, 37245, 37254, 37251, 37233, 37241, 37240, 37247, 37246, 37238, 37246, 37244, 37235, 37235, 37244, 37241, 37263, 37261, 37260, 37267, 37266, 37264, 37273, 37271, 37270, 37277, 37276, 37274, 37271, 37263, 37262, 37275, 37274, 37266, 37269, 37268, 37261, 37279, 37278, 37271, 37269, 37267, 37265, 37279, 37277, 37275, 37278, 37275, 37267, 37278, 37269, 37263, 37283, 37281, 37280, 37287, 37286, 37284, 37293, 37291, 37290, 37297, 37296, 37294, 37291, 37283, 37282, 37295, 37294, 37286, 37289, 37288, 37281, 37299, 37298, 37291, 37289, 37287, 37285, 37299, 37297, 37295, 37298, 37295, 37287, 37298, 37289, 37283, 37281, 37273, 37272, 37285, 37284, 37276,
37288, 37285, 37277, 37288, 37279, 37273, 37301, 37303, 37302, 37309, 37308, 37306, 37311, 37313, 37312, 37319, 37318, 37316, 37303, 37311, 37310, 37317, 37316, 37308, 37301, 37304, 37305, 37311, 37314, 37315, 37308, 37305, 37304, 37318, 37315, 37314, 37316, 37314, 37305, 37303, 37305, 37314, 37326, 37328, 37327, 37324, 37328, 37326, 37335, 37334, 37332, 37337, 37339, 37338, 37343, 37342, 37340, 37347, 37346, 37344, 37327, 37337, 37336, 37341, 37340, 37330, 37345, 37344, 37334, 37327, 37328, 37331, 37324, 37329, 37331, 37337, 37340, 37342, 37334, 37331, 37329, 37346, 37343, 37341, 37334, 37344, 37341, 37327, 37330, 37340, 37313, 37324, 37323, 37333, 37332, 37318, 37332, 37329, 37315, 37313, 37315, 37329, 37338, 37339, 37321, 37321, 37339, 37342, 37351, 37349, 37348, 37355, 37354, 37352, 37361, 37359, 37358, 37365, 37364, 37362, 37359, 37351, 37350, 37363, 37362, 37354, 37357, 37356, 37349, 37361, 37367, 37366, 37357, 37355, 37353, 37366, 37367, 37365, 37357, 37366, 37363, 37359, 37366, 37357, 37371, 37369, 37368, 37375, 37374, 37372, 37381, 37379, 37378, 37385, 37384, 37382, 37379, 37371, 37370, 37383, 37382, 37374, 37371, 37377, 37376, 37381, 37387, 37386, 37377, 37375, 37373, 37387, 37385, 37383, 37386, 37383, 37375, 37379, 37386, 37377,
37369, 37361, 37360, 37373, 37372, 37364, 37376, 37373, 37365, 37369, 37376, 37367, 37253, 37301, 37300, 37307, 37306, 37258, 37306, 37304, 37255, 37253, 37255, 37304, 37349, 37293, 37292, 37353, 37352, 37296, 37356, 37353, 37297, 37356, 37299, 37293, 37266, 37229, 37227, 37276, 37239, 37237, 37274, 37237, 37229, 37286, 37249, 37247, 37296, 37259, 37257, 37294, 37257, 37249, 37284, 37247, 37239, 37354, 37309, 37307, 37364, 37319, 37317, 37362, 37317, 37309, 37374, 37335, 37333, 37384, 37347, 37345, 37382, 37345, 37335, 37372, 37333, 37319, 37352, 37307, 37259, 37389, 37391, 37390, 37395, 37394, 37392, 37397, 37399, 37398, 37403, 37402, 37400, 37391, 37397, 37396, 37401, 37400, 37394, 37392, 37394, 37391, 37400, 37402, 37399, 37394, 37400, 37397, 37407, 37405, 37404, 37411, 37410, 37408, 37415, 37414, 37412, 37419, 37417, 37416, 37423, 37422, 37420, 37427, 37426, 37424, 37417, 37407, 37406, 37421, 37420, 37410, 37425, 37424, 37414, 37415, 37413, 37405, 37427, 37425, 37417, 37414, 37411, 37409, 37426, 37423, 37421, 37424, 37421, 37411, 37425, 37415, 37407, 37429, 37431, 37430, 37435, 37434, 37432, 37437, 37439, 37438, 37443, 37442, 37440, 37431, 37437, 37436, 37441, 37440, 37434, 37432, 37434, 37431, 37440, 37442, 37439, 37434, 37440, 37437,
37447, 37445, 37444, 37451, 37450, 37448, 37455, 37454, 37452, 37459, 37457, 37456, 37463, 37462, 37460, 37467, 37466, 37464, 37457, 37447, 37446, 37461, 37460, 37450, 37465, 37464, 37454, 37455, 37453, 37445, 37467, 37465, 37457, 37454, 37451, 37449, 37466, 37463, 37461, 37464, 37461, 37451, 37465, 37455, 37447, 37399, 37429, 37428, 37433, 37432, 37402, 37402, 37432, 37429, 37445, 37419, 37418, 37449, 37448, 37422, 37453, 37452, 37426, 37452, 37449, 37423, 37453, 37427, 37419, 37410, 37395, 37393, 37422, 37403, 37401, 37420, 37401, 37395, 37450, 37435, 37433, 37462, 37443, 37441, 37460, 37441, 37435, 37448, 37433, 37403, 37469, 37471, 37470, 37475, 37474, 37472, 37477, 37479, 37478, 37483, 37482, 37480, 37471, 37477, 37476, 37481, 37480, 37474, 37472, 37474, 37471, 37480, 37482, 37479, 37474, 37480, 37477, 37487, 37485, 37484, 37491, 37490, 37488, 37495, 37494, 37492, 37499, 37497, 37496, 37503, 37502, 37500, 37507, 37506, 37504, 37497, 37487, 37486, 37501, 37500, 37490, 37505, 37504, 37494, 37495, 37493, 37485, 37507, 37505, 37497, 37494, 37491, 37489, 37506, 37503, 37501, 37504, 37501, 37491, 37505, 37495, 37487, 37509, 37511, 37510, 37515, 37514, 37512, 37517, 37519, 37518, 37523, 37522, 37520, 37511, 37517, 37516, 37521, 37520, 37514,
37512, 37514, 37511, 37520, 37522, 37519, 37514, 37520, 37517, 37527, 37525, 37524, 37531, 37530, 37528, 37535, 37534, 37532, 37539, 37537, 37536, 37543, 37542, 37540, 37547, 37546, 37544, 37537, 37527, 37526, 37541, 37540, 37530, 37545, 37544, 37534, 37535, 37533, 37525, 37547, 37545, 37537, 37534, 37531, 37529, 37546, 37543, 37541, 37544, 37541, 37531, 37545, 37535, 37527, 37479, 37509, 37508, 37513, 37512, 37482, 37482, 37512, 37509, 37525, 37499, 37498, 37529, 37528, 37502, 37533, 37532, 37506, 37532, 37529, 37503, 37533, 37507, 37499, 37490, 37475, 37473, 37502, 37483, 37481, 37500, 37481, 37475, 37530, 37515, 37513, 37542, 37523, 37521, 37540, 37521, 37515, 37528, 37513, 37483, 37439, 37469, 37468, 37473, 37472, 37442, 37442, 37472, 37469, 37485, 37459, 37458, 37489, 37488, 37462, 37493, 37492, 37466, 37492, 37489, 37463, 37493, 37467, 37459, 37488, 37473, 37443, 37003, 37221, 37220, 37227, 37226, 37008, 37226, 37224, 37005, 37005, 37224, 37221, 37261, 37063, 37062, 37265, 37264, 37066, 37268, 37265, 37067, 37268, 37069, 37063, 37264, 37227, 37009, 37191, 37389, 37388, 37393, 37392, 37194, 37194, 37392, 37389, 37405, 37211, 37210, 37409, 37408, 37214, 37413, 37412, 37218, 37412, 37409, 37215, 37413, 37219, 37211, 37408, 37393, 37195,
37070, 37072, 36925, 37082, 36933, 36931, 36925, 37072, 37080, 37092, 36959, 36957, 37100, 36967, 36965, 37098, 36965, 36959, 37090, 36957, 36933, 37129, 37128, 37039, 37134, 37129, 37041, 37100, 37134, 37046, 37549, 37551, 37550, 37553, 37555, 37554, 37554, 37557, 37556, 37563, 37562, 37560, 37567, 37566, 37564, 37571, 37570, 37568, 37565, 37564, 37558, 37563, 37569, 37568, 37551, 37559, 37558, 37555, 37561, 37560, 37554, 37560, 37562, 37556, 37557, 37551, 37570, 37567, 37565, 37557, 37562, 37559, 37568, 37565, 37559, 37575, 37574, 37572, 37579, 37578, 37576, 37585, 37584, 37582, 37587, 37589, 37588, 37583, 37582, 37574, 37587, 37586, 37578, 37578, 37575, 37573, 37588, 37585, 37583, 37586, 37583, 37575, 37593, 37592, 37590, 37597, 37596, 37594, 37601, 37599, 37598, 37605, 37603, 37602, 37607, 37606, 37592, 37609, 37608, 37596, 37606, 37607, 37601, 37608, 37609, 37605, 37596, 37593, 37591, 37604, 37602, 37599, 37607, 37608, 37604, 37608, 37607, 37593, 37613, 37611, 37610, 37617, 37615, 37614, 37619, 37616, 37614, 37623, 37622, 37620, 37628, 37612, 37610, 37622, 37623, 37613, 37625, 37617, 37616, 37619, 37626, 37624, 37619, 37618, 37611, 37626, 37623, 37621, 37623, 37626, 37619, 37628, 37629, 37622, 37591, 37590, 37566, 37595, 37594, 37570,
37594, 37591, 37567, 37617, 37581, 37580, 37621, 37620, 37584, 37625, 37624, 37588, 37588, 37624, 37621, 37617, 37625, 37589, 37572, 37574, 37563, 37584, 37571, 37569, 37582, 37569, 37563, 37620, 37622, 37597, 37628, 37627, 37603, 37629, 37628, 37605, 37629, 37609, 37597, 37620, 37595, 37571, 37633, 37631, 37630, 37639, 37638, 37636, 37644, 37642, 37641, 37646, 37643, 37641, 37644, 37633, 37632, 37648, 37647, 37638, 37633, 37639, 37637, 37647, 37648, 37644, 37649, 37647, 37643, 37635, 37634, 37630, 37637, 37640, 37635, 37644, 37648, 37639, 37638, 37579, 37577, 37646, 37645, 37580, 37649, 37589, 37587, 37589, 37649, 37646, 37638, 37647, 37587, 37651, 37653, 37652, 37657, 37656, 37654, 37659, 37661, 37660, 37665, 37664, 37662, 37653, 37659, 37658, 37663, 37662, 37656, 37654, 37656, 37653, 37662, 37664, 37661, 37656, 37662, 37659, 37669, 37667, 37666, 37673, 37672, 37670, 37677, 37676, 37674, 37681, 37679, 37678, 37685, 37684, 37682, 37689, 37688, 37686, 37679, 37669, 37668, 37683, 37682, 37672, 37687, 37686, 37676, 37677, 37675, 37667, 37689, 37687, 37679, 37676, 37673, 37671, 37688, 37685, 37683, 37686, 37683, 37673, 37687, 37677, 37669, 37693, 37691, 37690, 37695, 37693, 37692, 37699, 37697, 37696, 37703, 37701, 37700, 37705, 37699, 37698,
37707, 37703, 37702, 37702, 37700, 37697, 37706, 37702, 37699, 37661, 37692, 37690, 37695, 37694, 37664, 37664, 37694, 37692, 37703, 37681, 37680, 37705, 37704, 37684, 37707, 37706, 37688, 37706, 37705, 37685, 37707, 37689, 37681, 37672, 37657, 37655, 37684, 37665, 37663, 37682, 37663, 37657, 37698, 37696, 37691, 37704, 37698, 37693, 37704, 37695, 37665, 37321, 37555, 37553, 37322, 37561, 37555, 37343, 37573, 37572, 37577, 37576, 37346, 37576, 37573, 37343, 37572, 37561, 37322, 37635, 37381, 37380, 37385, 37637, 37636, 37640, 37637, 37385, 37640, 37387, 37381, 37636, 37577, 37347, 37519, 37651, 37650, 37655, 37654, 37522, 37522, 37654, 37651, 37667, 37539, 37538, 37671, 37670, 37542, 37675, 37674, 37546, 37674, 37671, 37543, 37675, 37547, 37539, 37670, 37655, 37523, 35700, 35699, 35635, 35638, 35704, 35703, 35637, 35703, 35700, 35648, 35716, 35715, 35720, 35719, 35649, 35649, 35719, 35716, 35647, 35715, 35704, 35660, 35763, 35762, 35767, 35766, 35661, 35766, 35763, 35660, 35779, 35778, 35671, 35783, 35782, 35673, 35782, 35779, 35672, 35778, 35767, 35662, 35762, 35720, 35650, 35907, 35906, 35683, 35911, 35910, 35685, 35910, 35907, 35684, 35923, 35694, 35692, 35927, 35926, 35695, 35931, 35930, 35697, 35930, 35927, 35696, 35931, 35698, 35694,
35926, 35911, 35686, 35906, 35783, 35674, 10612, 26009, 26010, 26013, 26014, 10617, 10621, 26018, 26017, 26010, 26013, 10616, 10612, 10620, 26017, 10632, 26033, 26034, 26038, 10637, 10636, 26034, 26037, 10636, 26014, 26033, 10632, 10768, 26235, 26236, 10773, 26242, 26241, 10772, 26241, 26235, 26255, 26256, 10789, 26259, 26260, 10793, 26264, 10797, 10796, 26256, 26259, 10792, 26263, 10796, 10793, 26236, 26255, 10788, 21411, 34066, 34065, 34070, 34069, 21414, 34074, 34073, 21418, 34073, 34070, 21415, 21414, 34069, 34066, 34090, 21431, 21430, 34094, 34093, 21434, 34094, 21435, 21431, 34093, 34074, 21419, 21567, 34293, 34292, 34298, 34297, 21570, 21567, 21570, 34297, 34312, 21587, 21586, 34316, 34315, 21590, 21595, 34320, 34319, 34319, 34316, 21591, 34312, 34320, 21595, 34315, 34298, 21571, 172, 10961, 10960, 10958, 10961, 172, 159, 10952, 10951, 10949, 10952, 159, 174, 10964, 10958, 10951, 10964, 174, 185, 10977, 10976, 10976, 10980, 187, 10973, 10977, 185, 11094, 11096, 301, 309, 11101, 11100, 11100, 11104, 311, 11096, 11101, 309, 281, 11070, 11069, 11067, 11070, 281, 11082, 11084, 295, 283, 11073, 11067, 11069, 11082, 293, 295, 11084, 11064, 11064, 11094, 299, 209, 10998, 10997, 10995, 10998, 209,
196, 10989, 10988, 10986, 10989, 196, 211, 11001, 10995, 10988, 11001, 211, 11028, 11030, 241, 221, 11010, 10986, 10997, 11028, 239, 11180, 391, 390, 11183, 11179, 390, 11180, 11169, 380, 11159, 368, 367, 11159, 11156, 364, 11145, 357, 356, 11145, 11142, 353, 11162, 370, 364, 11162, 11144, 356, 11194, 11192, 397, 11194, 399, 380, 11192, 11158, 367, 241, 11030, 10973, 10980, 11073, 283, 10960, 11010, 221, 11, 10804, 10803, 10801, 10804, 11, 21, 10811, 10810, 25, 10815, 10814, 10814, 10818, 27, 10831, 43, 42, 10831, 10828, 39, 27, 10818, 10811, 10810, 10826, 37, 37, 10826, 10801, 10798, 10815, 25, 54, 10845, 10844, 10842, 10845, 54, 10850, 10848, 56, 56, 10848, 10842, 10875, 85, 84, 10875, 10872, 81, 103, 10893, 10892, 10892, 10896, 105, 10878, 87, 81, 105, 10896, 10890, 10888, 10874, 84, 10880, 10893, 103, 69, 10860, 10859, 10859, 10863, 72, 72, 10863, 10857, 116, 10905, 10904, 10904, 10908, 118, 10925, 10923, 131, 10911, 122, 121, 10914, 10910, 121, 10927, 135, 131, 10927, 10914, 124, 118, 10908, 10880, 10878, 10850, 58, 10857, 10905, 116, 10911, 10864, 73, 141, 10935, 10934,
10941, 148, 147, 10944, 10940, 147, 10934, 10940, 147, 10941, 10935, 141, 147, 10940, 10860, 10944, 150, 73, 53, 10844, 10798, 10888, 97, 39, 11211, 11213, 418, 410, 11202, 11201, 11201, 11205, 412, 11198, 11202, 410, 431, 11220, 11219, 11217, 11220, 431, 433, 11223, 11217, 11219, 11232, 443, 443, 11232, 11211, 11205, 11223, 433, 567, 11356, 11355, 11353, 11356, 567, 552, 11344, 11343, 11343, 11347, 554, 569, 11359, 11353, 11347, 11359, 569, 11355, 11384, 595, 455, 11243, 11242, 11242, 11246, 457, 459, 11248, 11243, 11295, 11293, 504, 11308, 520, 519, 11308, 11305, 516, 11305, 11295, 506, 11279, 488, 487, 11282, 11278, 487, 11270, 11268, 477, 11284, 11286, 494, 11270, 479, 475, 11279, 11266, 475, 531, 11322, 11321, 11321, 11325, 533, 11334, 11336, 544, 544, 11336, 11332, 11325, 11334, 542, 11319, 527, 504, 11319, 11282, 490, 11286, 11322, 531, 11430, 11428, 636, 11420, 11418, 627, 11432, 640, 636, 11432, 11420, 629, 11461, 11459, 666, 11469, 11467, 674, 11467, 11461, 668, 11439, 11441, 650, 11449, 11451, 660, 11441, 11449, 658, 11475, 11477, 686, 11485, 11487, 696, 11477, 11485, 694, 11459, 11430, 638,
11451, 11475, 684, 457, 11246, 11198, 11213, 11344, 552, 11418, 11307, 519, 11332, 11439, 648, 11875, 1086, 1085, 11875, 11878, 1088, 11887, 11889, 1098, 11878, 11887, 1096, 11915, 11917, 1125, 11927, 1135, 1134, 11924, 11926, 1134, 11915, 1123, 1121, 11917, 11924, 1131, 11889, 11913, 1121, 11797, 11795, 1002, 11805, 11803, 1011, 1015, 11807, 11805, 11803, 11797, 1004, 11836, 11834, 1043, 11846, 11844, 1053, 11844, 11836, 1045, 11814, 11816, 1025, 11824, 11826, 1035, 11816, 11824, 1033, 11854, 11856, 1065, 11866, 11868, 1075, 11858, 1067, 1065, 11858, 11866, 1073, 11834, 11807, 1015, 11826, 11854, 1063, 1163, 11951, 11950, 11954, 11951, 1163, 11964, 11954, 1165, 11986, 11984, 1191, 11977, 1185, 1184, 11974, 11976, 1184, 11988, 11990, 1197, 11977, 11988, 1195, 1222, 12015, 12014, 12014, 12012, 1218, 12023, 12021, 1229, 1233, 12025, 12023, 12021, 12015, 1222, 1191, 11984, 11964, 12012, 11986, 1193, 11950, 11846, 1055, 11868, 11974, 1181, 708, 11496, 11495, 11495, 11499, 710, 11499, 11509, 720, 11538, 11536, 747, 11550, 11548, 759, 11548, 11538, 749, 11522, 729, 728, 11525, 11521, 728, 11532, 11530, 735, 11530, 11525, 731, 11532, 737, 722, 11536, 11511, 722,
11649, 11647, 858, 11661, 11659, 870, 11659, 11649, 860, 11691, 11689, 894, 11697, 11695, 900, 11695, 11691, 896, 894, 11689, 11671, 11671, 11661, 872, 11573, 11575, 784, 772, 11563, 11562, 11562, 11566, 774, 774, 11566, 11560, 11560, 11573, 782, 820, 11609, 11608, 11606, 11609, 820, 11619, 11621, 832, 11608, 11619, 830, 795, 11586, 11585, 11585, 11583, 791, 799, 11590, 11589, 798, 11589, 11586, 791, 11583, 11563, 11575, 11606, 816, 11702, 910, 909, 11706, 914, 913, 11705, 913, 910, 11716, 928, 927, 11713, 11715, 927, 11720, 932, 931, 11719, 931, 928, 11713, 924, 914, 11647, 11550, 761, 11621, 11701, 909, 1257, 12049, 12048, 12052, 12049, 1257, 12060, 12058, 1266, 1270, 12062, 12060, 12058, 12052, 1259, 1299, 12088, 12087, 12087, 12085, 1295, 12100, 12098, 1309, 12098, 12088, 1299, 12085, 12062, 1270, 12149, 1361, 1360, 12146, 12148, 1360, 12160, 12162, 1371, 12164, 1373, 1371, 12149, 12160, 1369, 12118, 12146, 1357, 12190, 12188, 1393, 1403, 12195, 12194, 12198, 12195, 1403, 12194, 12190, 1395, 12214, 12212, 1421, 1435, 12223, 12222, 12226, 12223, 1435, 1421, 12212, 12210, 12222, 12214, 1423, 12204, 12206, 1411, 12236, 12238, 1443,
12242, 12244, 1449, 12238, 12242, 1447, 12210, 12198, 1405, 12206, 12236, 1441, 1393, 12188, 12128, 12128, 12100, 1311, 720, 11509, 11284, 12118, 1329, 1197, 11795, 11469, 676, 11487, 11814, 1023, 12048, 11697, 902, 595, 11384, 10949, 11874, 1085, 353, 10803, 11248, 459, 11268, 10830, 42, 10890, 11496, 708, 11522, 10925, 133, 12251, 12253, 1458, 12257, 12259, 1464, 12253, 12257, 1462, 12263, 12265, 1470, 12269, 12271, 1476, 12265, 12269, 1474, 12259, 12263, 1468, 12248, 12251, 1456, 12464, 12466, 1671, 12470, 12472, 1677, 12466, 12470, 1675, 12476, 12478, 1683, 12482, 12484, 1689, 12478, 12482, 1687, 12472, 12476, 1681, 12488, 12490, 1695, 12484, 12488, 1693, 12275, 12277, 1486, 1501, 12290, 12289, 12287, 12290, 1501, 1488, 12279, 12275, 12277, 12287, 1497, 12316, 12318, 1529, 12328, 12330, 1541, 12318, 12328, 1539, 12352, 12350, 1561, 12350, 12340, 1551, 12289, 12316, 1527, 12405, 1616, 1615, 12408, 12404, 1615, 12419, 12417, 1626, 12417, 12408, 1618, 12455, 12453, 1661, 12445, 12443, 1652, 12457, 1665, 1661, 12457, 12445, 1654, 12443, 12419, 1628, 12364, 12362, 1570, 12374, 1581, 1580, 12377, 12373, 1580, 12366, 1574, 1570, 12374, 12364, 1572, 12384, 12382, 1587,
12382, 12377, 1583, 12384, 1589, 1551, 12405, 12352, 1563, 12512, 12510, 1717, 12520, 12518, 1725, 12518, 12512, 1719, 12544, 12542, 1749, 12554, 12552, 1761, 1761, 12552, 12550, 12550, 12544, 1751, 12542, 12520, 1727, 12572, 12574, 1785, 12562, 12572, 1783, 12607, 1819, 1818, 12604, 12606, 1818, 12618, 12620, 1829, 12607, 12618, 1827, 12586, 12584, 1793, 12596, 12594, 1803, 12594, 12586, 1795, 1841, 12629, 12628, 12632, 12629, 1841, 12644, 12642, 1853, 12642, 12632, 1843, 12574, 12604, 1815, 12628, 12596, 1805, 1858, 12654, 12644, 12584, 12554, 1763, 1541, 12330, 12248, 12562, 1773, 1695, 12271, 12464, 1669, 12510, 12455, 1663, 13502, 13504, 2712, 13514, 2722, 2721, 13511, 13513, 2721, 13502, 2710, 2708, 13504, 13511, 2718, 13537, 13539, 2744, 13556, 2768, 2767, 13553, 13555, 2767, 13543, 13553, 2764, 13543, 2754, 2744, 13514, 13537, 2742, 13612, 13614, 2823, 13625, 2832, 2831, 13625, 13628, 2834, 13616, 2825, 2823, 13616, 13624, 2831, 13569, 13567, 2772, 2782, 13574, 13573, 13577, 13574, 2782, 13573, 13569, 2774, 13587, 13585, 2794, 2808, 13596, 13595, 13599, 13596, 2808, 2794, 13585, 13583, 13595, 13587, 2796, 2813, 13609, 13599, 13583, 13577, 2784, 13675, 2887, 2886,
13672, 13674, 2886, 13691, 2900, 2899, 13688, 13690, 2899, 13688, 2896, 2894, 13675, 13686, 2894, 2861, 13654, 13653, 13653, 13651, 2857, 13664, 13662, 2871, 2871, 13662, 13660, 13660, 13654, 2861, 2912, 13700, 13699, 13703, 13700, 2912, 13641, 13672, 2883, 13699, 13664, 2873, 13641, 2852, 2834, 13556, 13612, 2821, 13651, 13609, 2813, 13727, 2939, 2938, 13724, 13726, 2938, 13743, 2952, 2951, 13740, 13742, 2951, 13740, 2948, 2946, 13727, 13738, 2946, 13724, 2935, 2925, 13743, 13751, 2955, 13691, 13721, 2925, 12657, 12659, 1864, 12676, 1888, 1887, 12673, 12675, 1887, 12663, 12673, 1884, 12687, 12689, 1898, 12702, 1910, 1909, 12699, 12701, 1909, 12691, 1900, 1898, 12691, 12699, 1906, 12663, 1874, 1864, 12676, 12687, 1896, 12782, 12784, 1989, 1939, 12728, 12727, 12727, 12725, 1935, 12740, 12738, 1949, 12738, 12728, 1939, 12799, 2010, 2009, 12799, 12802, 2012, 12788, 12798, 2009, 12832, 2041, 2040, 12829, 12831, 2040, 12840, 12842, 2047, 12829, 2037, 2035, 12832, 12840, 2045, 12854, 2065, 2047, 12802, 12827, 2035, 12788, 1999, 1989, 12702, 12782, 1987, 13190, 2401, 2400, 13190, 13193, 2403, 13205, 2412, 2411, 13205, 13208, 2414, 13195, 2405, 2403, 13195, 13204, 2411,
13208, 13213, 2417, 12867, 2075, 2074, 12864, 12866, 2074, 12876, 12878, 2084, 12867, 12876, 2082, 2111, 12904, 12903, 12903, 12901, 2107, 12914, 12912, 2121, 2121, 12912, 12910, 12910, 12904, 2111, 12923, 2134, 2133, 12923, 12926, 2136, 12940, 2149, 2148, 12937, 12939, 2148, 12937, 2145, 2143, 12926, 12935, 2143, 12891, 2102, 2084, 12901, 12874, 2080, 12891, 12922, 2133, 2242, 13030, 13029, 13033, 13030, 2242, 2270, 13063, 13062, 13062, 13060, 2266, 13053, 13055, 2261, 13060, 13051, 2257, 13087, 13085, 2293, 2309, 13097, 13096, 13100, 13097, 2309, 2297, 13089, 13087, 13096, 13089, 2297, 2257, 13051, 13033, 13085, 13063, 2270, 12968, 12966, 2172, 2185, 12976, 12975, 12979, 12976, 2185, 2189, 12981, 12979, 12975, 12968, 2174, 2218, 13007, 13006, 13006, 13004, 2214, 13019, 13017, 2228, 13017, 13007, 2218, 2172, 12966, 12948, 13004, 12981, 2189, 13121, 2332, 2331, 13121, 13124, 2334, 13110, 13120, 2331, 13154, 2363, 2362, 13151, 13153, 2362, 13164, 13166, 2372, 13151, 2359, 2357, 13154, 13164, 2370, 13179, 2390, 2372, 13124, 13149, 2357, 13110, 2321, 2261, 2368, 13162, 13100, 13029, 12914, 2123, 12940, 13053, 2259, 13251, 2463, 2462, 13248, 13250, 2462, 13264, 13266, 2474,
13264, 2472, 2470, 13251, 13262, 2470, 13374, 2581, 2580, 13374, 13377, 2583, 13393, 2604, 2603, 13393, 13396, 2606, 13398, 2608, 2606, 13382, 13392, 2603, 13408, 2615, 2614, 13408, 13411, 2617, 13411, 13416, 2620, 13382, 2593, 2583, 13398, 13407, 2614, 2499, 13292, 13291, 13291, 13289, 2495, 13302, 13300, 2509, 2509, 13300, 13298, 13298, 13292, 2499, 2548, 13336, 13335, 13339, 13336, 2548, 13311, 2522, 2521, 13311, 13314, 2524, 13328, 2536, 2535, 13325, 13327, 2535, 13316, 2526, 2524, 13316, 13325, 2532, 2572, 13364, 13363, 13367, 13364, 2572, 13363, 13357, 2562, 2562, 13357, 13339, 13335, 13302, 2511, 13328, 13359, 2564, 13473, 2685, 2684, 13470, 13472, 2684, 2651, 13440, 13439, 13439, 13437, 2647, 13452, 13450, 2661, 2647, 13437, 13435, 13450, 13440, 2651, 13470, 2681, 2620, 13266, 13373, 2580, 13435, 13367, 2574, 2080, 12874, 12740, 13248, 2459, 2417, 12854, 13189, 2400, 13289, 13162, 2368, 13179, 13310, 2521, 2967, 13758, 13757, 13761, 13758, 2967, 2971, 13763, 13761, 13757, 13754, 2958, 3048, 13836, 13835, 13839, 13836, 3048, 13881, 3093, 3092, 13878, 13880, 3092, 13895, 3102, 3101, 13895, 13898, 3104, 13884, 3095, 3093, 13884, 13894, 3101, 3064, 13858, 13857,
3079, 13868, 13867, 13867, 13865, 3075, 3075, 13865, 13863, 13863, 13858, 3064, 13914, 3126, 3125, 3063, 13857, 13839, 13913, 3125, 3104, 13903, 13868, 3079, 13972, 3181, 3180, 13969, 13971, 3180, 13969, 3177, 3175, 3136, 13926, 13925, 13929, 13926, 3136, 3142, 13932, 13931, 3141, 13931, 13929, 13941, 13932, 3142, 3162, 13952, 13951, 13955, 13952, 3162, 3168, 13958, 13957, 3167, 13957, 13955, 3161, 13951, 13941, 13991, 3200, 3199, 13995, 3204, 3203, 13995, 13998, 3206, 13994, 3203, 3200, 13980, 13990, 3199, 14016, 3228, 3227, 14030, 3237, 3236, 14029, 3236, 3233, 14015, 3227, 3206, 14016, 14027, 3233, 3135, 13925, 13903, 13980, 3191, 3181, 13914, 13967, 3175, 14005, 13958, 3168, 13775, 2984, 2983, 13772, 13774, 2983, 13783, 13785, 2990, 13772, 2980, 2978, 13775, 13783, 2988, 13808, 3019, 3018, 13808, 13811, 3021, 13825, 3033, 3032, 13822, 13824, 3032, 13813, 3023, 3021, 13813, 13822, 3029, 13797, 3008, 2990, 13797, 13807, 3018, 13825, 13832, 3036, 3249, 14037, 14036, 3253, 14041, 14040, 14044, 14041, 3253, 3252, 14040, 14037, 14085, 3296, 3295, 14085, 14088, 3298, 14091, 3302, 3301, 14090, 3301, 3298, 3271, 14063, 14062, 3275, 14067, 14066, 3274, 14066, 14063,
14074, 14067, 3275, 14102, 3314, 3313, 14116, 3322, 3321, 14105, 3316, 3314, 14105, 14115, 3321, 3270, 14062, 14044, 14101, 3313, 3302, 3334, 14122, 14121, 3338, 14126, 14125, 3337, 14125, 14122, 14162, 3370, 3369, 14166, 3374, 3373, 14165, 3373, 3370, 3350, 14138, 14137, 3354, 14142, 14141, 3353, 14141, 14138, 14182, 3390, 3389, 14186, 3394, 3393, 14185, 3393, 3390, 3349, 14137, 14126, 14181, 3389, 3374, 3333, 14121, 14074, 14161, 3369, 3322, 3248, 14036, 14005, 14084, 3295, 3237, 13835, 13763, 2971, 2958, 13754, 13703, 13878, 3089, 2955, 12620, 13500, 2708, 13567, 12654, 1858, 13473, 13770, 2978, 11104, 12279, 1488, 12366, 11183, 393, 11927, 12657, 1862, 12725, 12025, 1233, 12164, 12864, 2071, 12948, 12226, 1437, 12244, 12970, 2176, 14199, 14201, 3410, 3414, 14204, 14203, 3400, 14195, 14193, 14207, 3416, 3410, 14193, 14199, 3408, 14203, 14195, 3400, 14225, 3432, 3431, 14225, 14228, 3434, 3425, 14218, 14217, 14217, 14215, 3421, 14252, 3462, 3461, 3450, 14238, 14237, 14241, 14238, 3450, 14258, 3468, 3467, 14255, 14257, 3467, 3449, 14237, 14233, 14255, 3464, 3462, 14235, 14251, 3461, 3421, 14215, 14204, 14235, 3446, 3434, 14207, 14224, 3431, 14233, 14218, 3425,
3474, 14268, 14267, 3491, 14280, 14279, 14279, 14277, 3487, 3487, 14277, 14273, 14273, 14268, 3474, 3494, 14290, 14280, 3473, 14267, 14241, 14275, 3485, 3468, 3500, 14294, 14293, 14299, 14301, 3508, 14293, 14299, 3506, 14307, 14309, 3516, 14316, 3522, 3521, 14309, 14315, 3521, 14301, 14307, 3514, 3530, 14322, 14321, 3534, 14326, 14325, 3533, 14325, 14322, 14370, 3578, 3577, 14374, 3582, 3581, 14373, 3581, 3578, 3550, 14342, 14341, 3554, 14346, 14345, 3553, 14345, 14342, 14390, 3598, 3597, 14394, 3602, 3601, 14393, 3601, 3598, 3549, 14341, 14326, 14389, 3597, 3582, 3529, 14321, 14294, 14369, 3577, 3522, 3646, 14437, 14436, 14436, 14440, 3649, 3649, 14440, 14434, 14431, 14437, 3646, 14401, 14403, 3610, 14413, 14415, 3624, 14407, 14405, 3612, 3632, 14422, 14421, 14421, 14417, 3627, 14418, 14422, 3632, 3612, 14405, 14401, 14403, 14413, 3622, 14417, 14407, 3614, 3662, 14450, 14449, 14449, 14453, 3664, 14468, 3678, 3677, 14468, 14465, 3674, 14472, 3682, 3681, 14471, 3681, 3674, 14472, 14455, 3666, 3664, 14453, 14431, 14467, 3677, 3650, 14415, 14450, 3662, 14541, 3748, 3747, 14540, 3747, 3744, 14482, 3693, 3692, 3711, 14499, 14498, 14487, 14489, 3699, 14503, 3715, 3714,
14487, 3697, 3693, 3699, 14489, 14485, 14485, 14499, 3711, 14527, 3735, 3734, 14530, 14526, 3734, 14532, 3739, 3735, 14516, 14514, 3719, 14522, 14520, 3725, 14520, 14516, 3721, 14522, 3727, 3715, 14532, 14502, 3714, 3710, 14498, 14418, 14530, 3737, 3666, 14434, 14538, 3744, 14541, 14441, 3650, 14556, 14558, 3767, 3757, 14550, 14549, 3756, 14549, 14547, 14547, 14556, 3765, 14596, 14598, 3807, 14610, 14612, 3821, 14602, 14600, 3809, 14616, 14614, 3823, 14598, 14610, 3819, 14614, 14602, 3811, 3777, 14569, 14568, 14581, 3793, 3792, 3781, 14573, 14572, 14585, 3797, 3796, 3780, 14572, 14569, 14584, 3796, 3793, 3776, 14568, 14550, 14580, 3792, 3769, 14558, 14596, 3805, 14600, 14560, 3769, 14696, 14700, 3909, 14701, 14697, 3906, 14712, 3920, 3919, 14709, 14711, 3919, 14696, 3905, 3902, 14700, 14709, 3916, 3933, 14723, 14722, 14726, 14723, 3933, 3939, 14729, 14728, 3938, 14728, 14726, 14722, 14719, 3923, 14744, 3956, 3955, 14741, 14743, 3955, 14748, 3960, 3959, 14747, 3959, 3956, 3923, 14719, 14701, 14741, 3952, 3942, 14712, 14738, 3942, 3833, 14625, 14624, 14637, 3849, 3848, 3837, 14629, 14628, 14641, 3853, 3852, 3836, 14628, 14625, 14640, 3852, 3849, 14666, 14668, 3873,
14672, 14674, 3879, 14668, 14672, 3877, 3859, 14653, 14652, 14660, 14658, 3865, 14658, 14653, 3859, 14680, 14678, 3885, 14688, 14686, 3893, 14686, 14680, 3887, 3858, 14652, 14629, 14641, 14666, 3871, 14678, 14660, 3867, 3982, 14771, 14770, 14768, 14771, 3982, 3969, 14762, 14761, 14759, 14762, 3969, 3984, 14774, 14768, 14761, 14774, 3984, 14802, 4013, 4012, 14806, 4017, 4016, 14809, 14805, 4016, 14809, 4019, 4013, 14788, 3996, 3995, 14788, 14785, 3992, 14797, 14795, 4000, 14795, 14787, 3995, 14797, 4002, 3986, 14806, 14776, 3986, 3832, 14624, 14573, 14636, 3848, 3797, 3981, 14770, 14729, 14801, 4012, 3960, 14612, 14694, 3902, 14697, 14616, 3825, 14674, 14759, 3965, 14785, 14688, 3895, 4025, 14819, 14818, 14824, 14826, 4033, 14818, 14824, 4031, 14832, 14834, 4041, 14840, 14842, 4049, 14834, 14840, 4047, 14826, 14832, 4039, 4057, 14849, 14848, 4061, 14853, 14852, 4060, 14852, 14849, 4077, 14869, 14868, 4081, 14873, 14872, 4080, 14872, 14869, 14913, 4124, 4123, 14917, 4128, 4127, 14920, 14916, 4127, 14920, 4130, 4124, 4076, 14868, 14853, 4056, 14848, 14819, 14275, 14481, 3692, 14514, 14290, 3494, 14929, 14931, 4140, 4144, 14934, 14933, 4143, 14933, 14929, 4156, 14944, 14943,
4160, 14948, 14947, 4159, 14947, 14944, 4155, 14943, 14934, 14985, 14987, 4196, 14999, 15001, 4210, 14991, 14989, 4198, 15005, 15003, 4212, 14987, 14999, 4208, 15003, 14991, 4200, 15025, 15027, 4236, 15032, 15029, 4238, 15041, 15045, 4252, 15031, 4241, 4236, 15031, 15041, 4249, 15042, 15032, 4242, 15014, 4222, 4221, 15018, 4226, 4225, 15017, 4225, 4222, 15053, 15051, 4256, 15013, 4221, 4198, 4256, 15051, 15042, 15001, 15025, 4234, 15029, 15005, 4214, 4172, 14960, 14959, 4176, 14964, 14963, 4175, 14963, 14960, 4184, 14976, 14975, 14979, 14976, 4184, 4183, 14975, 14964, 15058, 4267, 4266, 15062, 4271, 4270, 15062, 15065, 4273, 15072, 15074, 4279, 15061, 4270, 4267, 15065, 15072, 4277, 15094, 15096, 4301, 4310, 15103, 15102, 15100, 15103, 4310, 15096, 15100, 4306, 15080, 15078, 4285, 15088, 15086, 4293, 15086, 15080, 4287, 15111, 15109, 4316, 15120, 4328, 4327, 15120, 15117, 4324, 15117, 15111, 4318, 15074, 15094, 4299, 15109, 15088, 4295, 4171, 14959, 14948, 15057, 4266, 4226, 14931, 14985, 4194, 15078, 14979, 4186, 15304, 4511, 4510, 15304, 15307, 4513, 4524, 15313, 15312, 15332, 4540, 4539, 4532, 15321, 15320, 15320, 15318, 4528, 15336, 4544, 4543, 4528, 15318, 15313,
15335, 4543, 4540, 15316, 15331, 4539, 15316, 4526, 4513, 4597, 15390, 15389, 15387, 15390, 4597, 15396, 15398, 4606, 15389, 15396, 4603, 4588, 15381, 15380, 4587, 15380, 15378, 4619, 15407, 15406, 15425, 4632, 4631, 15428, 15424, 4631, 15413, 4625, 4624, 15412, 4624, 4621, 15425, 15410, 4621, 4618, 15406, 15387, 15428, 4634, 4607, 15378, 15407, 4619, 4556, 15344, 15343, 15343, 15347, 4558, 15364, 4571, 4570, 15352, 4564, 4563, 4560, 15349, 15344, 15367, 4573, 4571, 15367, 15351, 4563, 15374, 15372, 4577, 15374, 4579, 4564, 4641, 15434, 15433, 4643, 15437, 15434, 15443, 4655, 4654, 4659, 15447, 15446, 15450, 15447, 4659, 15465, 4674, 4673, 15462, 15464, 4673, 15462, 4670, 4668, 15443, 15460, 4668, 4686, 15479, 15478, 15478, 15476, 4682, 15476, 15473, 4677, 4677, 15473, 15450, 15446, 15437, 4643, 4558, 15347, 15321, 15363, 4570, 4544, 4640, 15433, 15381, 15442, 4654, 4625, 15127, 15129, 4336, 15135, 15137, 4344, 15129, 15135, 4342, 15143, 15145, 4352, 15151, 15153, 4360, 15145, 15151, 4358, 15137, 15143, 4350, 15215, 15217, 4424, 15223, 15225, 4432, 15217, 15223, 4430, 15231, 15233, 4440, 15239, 15241, 4448, 15233, 15239, 4446, 15225, 15231, 4438, 15177, 15175, 4380,
15183, 15181, 4386, 15181, 15177, 4382, 15205, 15203, 4408, 15211, 15209, 4414, 15209, 15205, 4410, 15203, 15183, 4388, 15265, 15263, 4468, 15271, 15269, 4474, 15269, 15265, 4470, 15293, 15291, 4496, 15299, 15297, 4502, 15297, 15293, 4498, 15291, 15271, 4476, 15153, 15215, 4422, 15263, 15211, 4416, 15485, 15487, 4694, 15493, 15495, 4702, 15487, 15493, 4700, 15504, 4712, 4711, 15501, 15503, 4711, 15495, 15501, 4708, 4768, 15562, 15561, 15567, 15569, 4776, 15575, 15577, 4784, 15569, 15575, 4782, 15561, 15567, 4774, 15540, 4752, 4751, 15544, 4756, 4755, 15543, 4755, 4752, 15529, 15527, 4732, 15535, 15533, 4738, 15533, 15529, 4734, 15557, 15555, 4760, 15557, 4762, 4756, 15555, 15535, 4740, 4792, 15584, 15583, 4796, 15588, 15587, 4795, 15587, 15584, 4804, 15596, 15595, 4808, 15600, 15599, 4807, 15599, 15596, 4803, 15595, 15588, 15539, 4751, 4712, 4791, 15583, 15562, 15045, 15303, 4510, 15312, 15053, 4258, 15102, 15349, 4560, 15372, 15119, 4327, 15241, 15485, 4692, 15527, 15299, 4504, 15639, 15641, 4850, 15655, 15657, 4866, 15647, 15644, 4854, 15661, 15659, 4868, 4853, 15643, 15639, 15641, 15655, 4864, 15659, 15647, 4856, 15669, 15671, 4880, 15676, 15673, 4882, 15688, 4896, 4895,
15685, 15687, 4895, 15675, 4885, 4880, 15675, 15685, 4892, 4909, 15701, 15700, 15700, 15698, 4905, 4911, 15704, 15701, 15698, 15695, 4899, 4899, 15695, 15676, 15657, 15669, 4878, 15673, 15661, 4870, 15780, 4992, 4991, 15777, 15779, 4991, 4996, 15784, 15783, 15788, 5000, 4999, 15787, 4999, 4992, 15777, 4988, 4978, 15710, 15712, 4917, 15716, 15718, 4923, 15712, 15716, 4921, 15738, 15740, 4945, 4956, 15747, 15746, 15746, 15750, 4958, 4958, 15750, 15744, 15740, 15747, 4956, 15724, 15722, 4929, 15732, 15730, 4937, 15730, 15724, 4931, 15759, 15757, 4964, 15766, 4973, 4972, 15769, 15765, 4972, 15766, 15759, 4966, 15718, 15738, 4943, 15757, 15732, 4939, 5012, 15800, 15799, 15804, 5016, 5015, 15810, 5022, 5021, 15810, 15807, 5018, 15809, 5021, 5016, 15821, 5025, 5018, 5011, 15799, 15784, 15803, 5015, 5000, 15688, 15774, 4978, 15783, 15704, 4911, 15744, 15800, 5012, 15821, 15769, 4975, 15962, 15964, 5173, 5177, 15967, 15966, 5176, 15966, 15962, 15995, 5202, 5201, 15994, 5201, 5198, 5189, 15977, 15976, 5193, 15981, 15980, 5192, 15980, 15977, 16002, 5210, 5209, 16006, 5214, 5213, 5223, 16016, 16015, 16013, 16016, 5223, 16005, 5213, 5210, 5188, 15976, 15967, 16001, 5209, 5202,
15964, 15992, 5198, 5294, 16087, 16086, 16084, 16087, 5294, 16129, 16131, 5339, 16141, 5348, 5347, 5339, 16131, 16127, 16140, 5347, 5344, 16127, 16138, 5344, 5307, 16096, 16095, 16093, 16096, 5307, 16115, 5326, 5325, 16118, 16114, 5325, 5309, 16099, 16093, 16118, 5328, 5322, 16095, 16110, 5320, 16115, 16101, 5311, 16162, 5370, 5369, 5362, 16150, 16149, 16166, 5374, 5373, 5361, 16149, 16147, 16165, 5373, 5370, 5320, 16110, 16084, 16161, 5369, 5348, 16086, 16129, 5337, 16147, 16112, 5322, 5235, 16023, 16022, 5239, 16027, 16026, 16030, 16027, 5239, 5238, 16026, 16023, 16044, 5256, 5255, 5260, 16048, 16047, 16044, 16048, 5260, 16066, 5277, 5276, 16069, 16065, 5276, 16069, 5279, 5273, 16047, 16061, 5271, 16066, 16051, 5262, 16080, 16078, 5283, 5244, 16040, 16030, 16080, 5285, 5262, 16078, 16040, 5244, 16174, 5381, 5380, 16177, 16173, 5380, 5395, 16183, 16182, 16199, 5407, 5406, 5399, 16187, 16186, 16203, 5411, 5410, 5398, 16186, 16183, 16202, 5410, 5407, 5419, 16211, 16210, 16214, 16211, 5419, 16221, 5427, 5426, 16221, 16214, 5421, 5418, 16210, 16187, 16220, 5426, 5411, 5234, 16022, 15981, 16043, 5255, 5214, 5271, 16061, 16013, 16177, 5383, 5311, 5394, 16182, 16150,
16198, 5406, 5374, 16015, 16099, 5309, 16174, 16063, 5273, 15825, 5031, 5030, 5063, 15855, 15854, 15854, 15858, 5065, 15864, 15866, 5073, 15858, 15864, 5071, 15875, 5083, 5082, 15872, 15874, 5082, 15866, 15872, 5079, 15831, 5039, 5038, 15835, 5043, 5042, 15834, 5042, 5039, 15843, 5051, 5050, 15847, 5055, 5054, 15846, 5054, 5051, 15842, 5050, 5043, 5095, 15883, 15882, 5099, 15887, 15886, 5098, 15886, 15883, 15931, 5143, 5142, 15935, 5147, 5146, 15934, 5146, 5143, 5119, 15907, 15906, 5123, 15911, 15910, 5122, 15910, 15907, 15947, 5159, 5158, 15951, 5163, 5162, 15950, 5162, 5159, 5118, 15906, 15887, 15946, 5158, 5147, 15830, 5038, 5031, 5094, 15882, 15855, 15930, 5142, 5083, 15398, 15643, 4853, 15644, 15399, 4607, 15465, 15710, 4915, 15722, 15479, 4686, 15577, 15824, 5030, 14842, 15127, 4334, 15175, 14912, 4123, 29652, 29654, 21605, 21609, 29657, 29656, 29669, 21618, 21617, 21608, 29656, 29652, 29668, 21617, 21614, 29654, 29666, 21614, 21630, 29676, 29675, 29692, 21644, 21643, 21634, 29680, 29679, 29696, 21648, 21647, 21652, 29700, 29699, 21633, 29679, 29676, 29695, 21647, 21644, 21629, 29675, 29657, 29691, 21643, 21618, 29761, 29763, 21712, 29769, 29771, 21720, 29763, 29769, 21718,
29780, 21730, 21729, 29777, 29779, 21729, 29771, 29777, 21726, 21664, 29710, 29709, 29726, 21678, 21677, 21668, 29714, 29713, 29732, 21684, 21683, 29732, 29735, 21686, 21667, 29713, 29710, 29731, 21683, 21678, 21694, 29744, 29743, 29747, 29744, 21694, 29755, 29753, 21702, 29753, 29747, 21696, 21693, 29743, 29714, 29789, 29787, 21736, 29801, 29799, 21748, 29791, 29793, 21742, 29803, 29805, 21754, 29799, 29789, 21738, 29793, 29803, 21752, 29829, 29827, 21776, 29831, 29833, 21782, 29813, 29811, 21760, 29821, 29819, 21768, 29819, 29813, 21762, 29842, 21792, 21791, 29842, 29839, 21788, 29841, 21791, 21782, 29827, 29801, 21750, 29805, 29831, 21780, 29839, 29821, 21770, 21663, 29709, 29680, 29725, 21677, 21648, 21680, 29729, 29700, 29829, 21778, 21730, 29699, 29761, 21710, 29787, 29729, 21680, 29735, 29791, 21740, 29811, 29755, 21704, 21805, 29853, 29852, 21809, 29857, 29856, 29856, 29860, 21811, 21811, 29860, 29853, 21804, 29852, 29849, 21835, 29885, 29884, 29884, 29888, 21837, 29896, 29898, 21849, 21849, 29898, 29894, 29888, 29896, 21846, 29929, 29931, 21882, 29943, 29945, 21896, 29935, 29933, 21884, 29949, 29947, 21898, 29931, 29943, 21894, 29947, 29935, 21886, 29910, 21862, 21861, 29910, 29907, 21858, 29925, 29923, 21870, 29914, 21866, 21865,
29913, 21865, 21858, 29923, 29909, 21861, 29925, 21872, 21850, 29894, 29929, 21880, 29933, 29899, 21850, 21821, 29869, 29868, 29872, 29869, 21821, 21827, 29875, 29874, 21826, 29874, 29872, 29958, 21911, 21910, 29958, 29961, 21913, 29975, 21924, 21923, 29975, 29978, 21926, 29965, 21917, 21913, 29965, 29974, 21923, 29999, 30001, 21948, 30005, 30007, 21954, 30001, 30005, 21952, 29985, 29983, 21930, 21940, 29990, 29989, 29993, 29990, 21940, 29989, 29985, 21932, 30013, 30011, 21960, 30021, 30019, 21968, 30019, 30013, 21962, 21930, 29983, 29963, 29978, 29999, 21946, 30011, 29993, 21942, 21820, 29868, 29857, 29957, 21910, 21866, 29849, 29885, 21835, 29963, 29875, 21827, 30137, 30141, 22091, 30149, 22098, 22097, 30149, 30152, 22100, 30137, 22088, 22085, 30141, 30148, 22097, 30159, 30157, 22104, 22116, 30164, 30163, 30168, 30164, 22116, 22124, 30172, 30171, 22123, 30171, 30168, 30163, 30159, 22106, 30182, 22135, 22134, 30182, 30185, 22137, 30188, 22141, 22140, 30187, 22140, 22137, 22104, 30157, 30138, 30167, 22119, 22100, 30167, 30181, 22134, 30265, 30267, 22218, 22208, 30259, 30258, 30270, 22222, 22221, 22207, 30258, 30256, 30269, 22221, 22218, 30256, 30265, 22216, 22230, 30280, 30279, 30292, 22246, 22245, 22234, 30284, 30283, 30296, 22250, 22249,
22233, 30283, 30280, 30295, 22249, 22246, 22229, 30279, 30259, 30291, 22245, 22222, 22163, 30210, 30209, 30207, 30210, 22163, 22150, 30201, 30200, 30198, 30201, 22150, 22165, 30213, 30207, 30200, 30213, 22165, 30243, 22195, 22194, 30243, 30240, 22191, 30247, 22199, 22198, 30246, 22198, 22191, 30227, 22177, 22176, 30227, 30224, 22173, 30236, 30234, 22181, 30234, 30226, 22176, 30236, 22183, 22167, 30247, 30215, 22167, 22258, 30308, 30307, 30320, 22274, 22273, 22262, 30312, 30311, 30324, 22278, 22277, 22261, 30311, 30308, 30323, 22277, 22274, 22284, 30336, 30335, 30343, 30341, 22290, 30341, 30336, 22284, 22283, 30335, 30312, 22162, 30209, 30172, 30242, 22194, 22141, 22257, 30307, 30284, 30319, 22273, 22250, 21978, 30028, 30027, 30027, 30031, 21980, 30037, 30039, 21988, 30031, 30037, 21986, 30048, 21998, 21997, 30045, 30047, 21997, 30039, 30045, 21994, 22010, 30056, 30055, 22014, 30060, 30059, 22013, 30059, 30056, 30104, 22058, 22057, 30108, 22062, 22061, 30107, 22061, 22058, 22034, 30080, 30079, 22038, 30084, 30083, 22037, 30083, 30080, 30120, 22074, 22073, 30124, 22078, 22077, 30123, 22077, 22074, 22033, 30079, 30060, 30119, 22073, 22062, 22009, 30055, 30028, 30103, 22057, 21998, 29945, 30135, 22085, 30138, 29949, 21900, 30007, 30198, 22146,
30224, 30021, 21970, 22319, 30370, 30369, 30367, 30370, 22319, 30355, 30357, 22306, 30361, 30359, 22308, 22298, 30351, 30349, 30349, 30355, 22304, 30359, 30351, 22298, 22330, 30379, 30378, 30376, 30379, 22330, 30384, 30382, 22332, 22348, 30394, 30393, 30397, 30394, 22348, 30378, 30391, 22344, 30393, 30384, 22334, 22344, 30391, 30367, 30357, 30376, 22326, 30382, 30361, 22310, 30468, 30470, 22421, 30482, 22431, 22430, 22421, 30470, 30466, 30481, 22430, 22427, 30466, 30479, 22427, 30482, 30471, 22422, 30489, 22443, 22442, 30492, 30488, 22442, 30489, 30493, 22447, 30509, 22457, 22456, 30498, 22451, 22447, 30498, 30508, 22456, 30492, 22446, 22422, 30407, 30409, 22356, 30413, 30415, 22362, 30409, 30413, 22360, 30435, 30437, 22384, 22393, 30444, 30443, 30441, 30444, 22393, 30437, 30441, 22389, 30421, 30419, 22368, 30429, 30427, 22376, 30427, 30421, 22370, 30452, 30450, 22399, 30460, 30458, 22407, 30458, 30452, 22401, 30415, 30435, 22382, 30450, 30429, 22378, 22469, 30515, 30514, 30529, 22479, 22478, 30533, 22483, 22482, 22471, 30518, 30515, 30532, 22482, 22479, 30541, 22490, 22489, 30544, 30540, 22489, 30549, 30544, 22492, 30549, 22495, 22483, 22468, 30514, 30496, 30528, 22478, 22457, 30369, 30468, 22418, 30496, 30397, 22350, 30443, 30518, 22471,
30541, 30460, 22409, 22503, 30553, 30552, 30552, 30556, 22505, 30562, 30564, 22513, 30556, 30562, 22511, 30570, 30572, 22521, 30578, 30580, 22529, 30572, 30578, 22527, 30564, 30570, 22519, 30672, 30674, 22623, 30680, 30682, 22631, 30674, 30680, 22629, 30688, 30690, 22639, 30696, 30698, 22647, 30690, 30696, 22645, 30682, 30688, 22637, 22541, 30587, 30586, 22545, 30591, 30590, 22544, 30590, 30587, 22565, 30611, 30610, 22569, 30615, 30614, 22568, 30614, 30611, 30651, 22605, 22604, 30668, 30666, 22613, 30655, 22609, 22608, 30654, 22608, 22605, 30666, 30650, 22604, 22564, 30610, 30591, 30722, 30720, 22667, 30728, 30726, 22673, 30726, 30722, 22669, 30750, 30748, 22695, 30756, 30754, 22701, 30754, 30750, 22697, 30748, 30728, 22675, 22540, 30586, 30553, 30580, 30672, 22621, 30720, 30668, 22615, 30760, 30762, 22711, 30768, 30770, 22719, 30762, 30768, 22717, 30776, 30778, 22727, 30784, 30786, 22735, 30778, 30784, 22733, 30770, 30776, 22725, 30848, 30850, 22799, 30856, 30858, 22807, 30850, 30856, 22805, 30865, 22813, 22812, 30858, 30864, 22812, 30810, 30808, 22755, 30816, 30814, 22761, 30814, 30810, 22757, 30838, 30836, 22783, 30844, 30842, 22789, 30842, 30838, 22785, 30836, 30816, 22763, 30899, 22849, 22848, 30903, 22853, 22852, 30902, 22852, 22849,
30888, 30886, 22833, 30894, 30892, 22839, 30892, 30888, 22835, 30910, 22856, 22853, 30910, 30894, 22841, 30898, 22848, 22813, 30786, 30848, 22797, 30886, 30844, 22791, 30698, 30760, 22709, 30808, 30756, 22703, 30913, 30915, 22866, 30927, 30929, 22880, 22870, 30918, 30917, 30933, 30931, 22882, 22869, 30917, 30913, 30915, 30927, 22878, 30969, 30971, 22922, 30983, 30985, 22936, 30975, 30973, 22924, 30989, 30987, 22938, 30971, 30983, 22934, 30987, 30975, 22926, 22896, 30942, 30941, 30958, 22908, 22907, 22900, 30946, 30945, 30962, 22912, 22911, 22899, 30945, 30942, 30961, 22911, 22908, 22948, 30998, 30997, 30997, 31001, 22950, 31007, 31009, 22958, 31001, 31007, 22956, 22895, 30941, 30918, 30957, 22907, 22882, 30929, 30969, 22920, 30973, 30933, 22884, 31104, 31106, 23057, 31111, 31108, 23059, 31120, 31124, 23073, 31110, 23062, 23057, 31110, 31120, 23070, 31121, 31111, 23063, 31156, 23104, 23103, 23094, 31141, 31140, 31140, 31138, 23090, 31130, 31132, 23081, 31144, 31146, 23098, 23098, 31146, 31141, 31132, 31144, 23096, 31164, 23118, 23117, 31161, 31163, 23117, 31168, 23122, 23121, 31167, 23121, 23118, 23090, 31138, 31121, 31161, 23114, 23104, 31124, 31155, 23103, 22970, 31016, 31015, 31032, 22983, 22982, 22974, 31020, 31019, 31036, 22987, 22986,
31036, 31039, 22989, 22973, 31019, 31016, 31035, 22986, 22983, 31066, 31064, 23013, 31078, 31076, 23025, 31068, 31070, 23019, 31080, 31082, 23031, 31076, 31066, 23015, 31070, 31080, 23029, 22997, 31047, 31046, 31050, 31047, 22997, 31058, 31056, 23005, 31056, 31050, 22999, 31090, 31088, 23037, 31098, 31096, 23045, 31096, 31090, 23039, 22996, 31046, 31020, 31039, 31068, 23017, 31088, 31058, 23007, 31181, 31179, 23128, 23146, 31192, 31191, 31195, 31192, 23146, 31183, 31185, 23134, 23152, 31198, 31197, 31197, 31201, 23154, 23154, 31201, 31195, 31191, 31181, 23130, 31185, 31198, 23152, 31229, 23182, 23181, 31233, 23186, 23185, 31236, 31232, 23185, 31236, 23188, 23182, 31213, 31211, 23160, 31220, 23169, 23168, 31223, 31219, 23168, 31220, 31213, 23162, 31245, 23191, 23186, 31245, 31223, 23171, 22969, 31015, 30946, 31031, 22982, 22912, 23013, 31064, 30998, 31228, 23181, 23122, 30985, 31104, 23055, 31108, 30989, 22940, 31009, 31130, 23079, 31179, 31078, 23027, 31082, 31183, 23132, 31211, 31098, 23047, 31413, 31415, 23366, 23370, 31418, 31417, 23369, 31417, 31413, 23384, 31430, 31429, 31434, 23388, 23387, 23392, 31438, 31437, 31442, 23396, 23395, 23391, 31437, 31430, 31441, 23395, 23388, 23383, 31429, 31418, 31433, 23387, 23372, 31489, 31491, 23442,
31503, 31505, 23456, 31495, 31493, 23444, 31509, 31507, 23458, 31491, 31503, 23454, 31507, 31495, 23446, 31517, 31519, 23470, 31531, 31533, 23484, 31523, 31521, 23472, 31538, 23490, 23489, 31538, 31535, 23486, 31537, 23489, 23484, 31519, 31531, 23482, 31535, 31523, 23474, 31505, 31517, 23468, 31521, 31509, 23460, 23408, 31454, 31453, 31458, 23412, 23411, 23416, 31462, 31461, 31466, 23420, 23419, 23415, 31461, 31454, 31465, 23419, 23412, 23428, 31478, 31477, 31482, 23432, 23431, 31482, 31478, 23428, 23427, 31477, 31462, 31481, 23431, 23420, 23407, 31453, 31438, 31457, 23411, 23396, 31415, 31489, 23440, 31493, 31421, 23372, 31316, 31318, 23265, 31313, 31316, 23263, 31322, 31324, 23271, 23280, 31331, 31330, 31328, 31331, 23280, 31324, 31328, 23276, 31318, 31322, 23269, 23240, 31287, 31286, 31284, 31287, 23240, 23225, 31274, 31273, 31273, 31277, 23227, 23227, 31277, 31271, 31271, 31284, 23236, 23210, 31257, 31256, 31254, 31257, 23210, 23214, 31261, 31260, 23213, 31260, 31254, 31251, 31261, 23214, 23209, 31256, 31248, 31248, 31274, 23225, 31337, 31339, 23292, 31373, 31371, 23321, 31360, 23314, 23313, 31360, 31357, 23310, 31375, 23325, 23321, 31375, 31359, 23313, 31399, 23348, 23347, 31402, 31398, 23347, 31409, 31407, 23354, 31407, 31402, 23350,
31399, 31373, 23323, 23292, 31339, 31313, 31286, 31337, 23290, 31547, 31549, 23498, 31555, 31557, 23506, 31549, 31555, 23504, 31563, 31565, 23514, 31574, 23524, 23523, 31571, 31573, 23523, 31565, 31571, 23520, 31557, 31563, 23512, 31637, 31639, 23586, 31643, 31645, 23592, 31639, 31643, 23590, 31645, 31649, 23595, 31599, 31597, 23544, 31605, 31603, 23550, 31603, 31599, 23546, 31627, 31625, 23572, 31633, 31631, 23578, 31631, 31627, 23574, 31625, 31605, 23552, 31688, 31690, 23643, 31703, 23657, 23656, 31700, 31702, 23656, 31690, 31700, 23653, 23620, 31671, 31670, 31670, 31668, 23616, 31679, 31677, 23627, 23631, 31681, 31679, 31677, 31671, 23620, 23670, 31717, 31716, 31716, 31714, 23666, 31727, 31717, 23670, 31714, 31681, 23631, 31688, 23641, 23595, 31574, 31637, 23584, 31668, 31633, 23580, 31330, 31547, 23496, 31597, 31409, 23356, 30324, 30407, 22354, 30419, 30343, 22292, 31737, 31739, 23690, 31751, 31753, 23704, 23694, 31742, 31741, 31757, 31755, 23706, 23693, 31741, 31737, 31739, 31751, 23702, 31799, 31801, 23752, 31813, 31815, 23766, 31805, 31803, 23754, 31819, 31817, 23768, 31801, 31813, 23764, 31817, 31805, 23756, 23720, 31766, 31765, 31782, 23734, 23733, 23724, 31770, 31769, 31786, 23738, 23737, 23742, 31790, 31789, 23723, 31769, 31766,
31785, 23737, 23734, 31827, 31829, 23778, 31835, 31837, 23786, 31829, 31835, 23784, 23719, 31765, 31742, 31781, 23733, 23706, 31753, 31799, 23750, 31803, 31757, 23708, 31789, 31827, 23776, 23798, 31844, 31843, 31860, 23812, 23811, 23802, 31848, 31847, 31866, 23818, 23817, 31866, 31869, 23820, 23801, 31847, 31844, 31865, 23817, 23812, 31897, 31895, 23844, 31909, 31907, 23856, 31899, 31901, 23850, 31911, 31913, 23862, 31907, 31897, 23846, 31901, 31911, 23860, 23828, 31878, 31877, 31881, 31878, 23828, 31889, 31887, 23836, 31887, 31881, 23830, 31921, 31919, 23868, 31929, 31927, 23876, 31927, 31921, 23870, 23827, 31877, 31848, 31895, 31863, 23814, 31869, 31899, 23848, 31919, 31889, 23838, 23797, 31843, 31770, 31859, 23811, 23738, 23814, 31863, 31790, 32035, 32037, 23988, 32041, 32039, 23990, 32052, 24001, 24000, 32051, 24000, 23997, 32037, 32049, 23997, 32052, 32041, 23992, 32061, 24011, 24010, 32058, 32060, 24010, 24049, 32100, 32099, 24048, 32099, 32097, 32118, 32120, 24071, 32123, 24075, 24074, 32122, 24074, 24071, 24057, 32107, 32106, 24061, 32111, 32110, 24060, 32110, 32107, 32133, 24087, 24086, 32137, 24091, 24090, 32136, 24090, 24087, 24056, 32106, 32100, 32132, 24086, 24075, 32097, 32118, 24069, 32070, 32068, 24017, 32072, 32074, 24023,
32074, 32080, 24026, 32085, 32083, 24032, 32092, 24040, 24039, 32092, 32085, 24034, 32091, 24039, 24026, 24102, 32152, 32151, 24106, 32156, 32155, 24105, 32155, 32152, 32170, 24124, 24123, 24136, 32182, 32181, 32181, 32185, 24138, 32190, 24144, 24143, 32190, 32187, 24140, 32170, 32182, 24136, 24112, 32164, 32163, 32202, 24151, 24150, 32205, 32201, 24150, 32210, 32205, 24153, 24111, 32163, 32156, 32210, 24156, 24140, 32202, 32164, 24112, 32070, 24019, 24011, 24101, 32151, 32111, 32169, 24123, 24091, 24138, 32185, 32148, 23930, 31982, 31981, 31987, 31989, 23938, 31981, 31987, 23936, 31940, 23891, 23890, 31937, 31939, 23890, 31937, 23887, 23885, 31940, 31948, 23894, 31962, 23916, 23915, 31959, 31961, 23915, 31965, 23918, 23916, 31976, 23924, 23923, 31959, 23912, 23894, 31965, 31975, 23923, 23946, 31996, 31995, 23950, 32000, 31999, 23949, 31999, 31996, 23966, 32016, 32015, 23970, 32020, 32019, 23969, 32019, 32016, 23965, 32015, 32000, 23945, 31995, 31982, 32213, 32215, 24164, 32222, 24170, 24169, 32215, 32221, 24169, 24216, 32268, 32267, 32273, 32275, 24224, 32281, 32283, 24232, 32275, 32281, 24230, 32267, 32273, 24222, 32236, 24186, 24185, 32240, 24190, 24189, 32239, 24189, 24186, 32256, 24206, 24205, 32260, 24210, 24209, 32259, 24209, 24206,
32255, 24205, 24190, 24240, 32290, 32289, 24244, 32294, 32293, 24243, 32293, 32290, 24252, 32302, 32301, 24256, 32306, 32305, 24255, 32305, 32302, 24251, 32301, 32294, 32235, 24185, 24170, 24239, 32289, 32268, 31815, 32035, 23986, 32039, 31819, 23770, 31837, 32058, 24007, 32068, 31909, 23858, 31913, 32072, 24021, 32083, 31929, 23878, 31989, 32213, 24162, 24332, 32382, 32381, 32381, 32385, 24334, 32392, 24340, 24339, 32385, 32391, 24339, 24316, 32362, 32361, 32361, 32365, 24318, 24302, 32351, 32350, 32350, 32354, 24304, 32370, 24324, 24323, 32370, 32367, 24320, 24304, 32354, 32348, 32348, 32362, 24316, 32399, 32397, 24349, 32415, 24365, 24364, 24357, 32404, 32403, 32419, 24369, 24368, 32401, 24353, 24349, 24356, 32403, 32399, 32418, 24368, 24365, 24318, 32365, 32345, 32414, 24364, 24340, 32345, 32382, 24332, 32401, 32369, 24323, 32429, 24380, 24379, 32429, 32426, 24376, 32432, 24382, 24376, 32439, 32428, 24379, 24393, 32443, 32442, 32455, 24405, 24404, 24397, 32447, 32446, 32459, 24409, 24408, 24396, 32446, 32443, 32458, 24408, 24405, 24415, 32467, 32466, 32473, 24421, 24420, 32473, 32467, 24415, 24414, 32466, 32447, 32472, 24420, 24409, 32439, 24385, 24320, 24392, 32442, 32404, 32454, 24404, 24369, 32673, 32675, 24626, 24616, 32667, 32666,
32679, 32677, 24628, 24615, 32666, 32664, 32664, 32673, 24624, 32711, 32713, 24664, 32725, 32727, 24678, 32717, 32715, 24666, 32731, 32729, 24680, 32713, 32725, 24676, 32729, 32717, 24668, 24638, 32688, 32687, 32700, 24650, 24649, 24642, 32692, 32691, 32704, 24654, 24653, 24641, 32691, 32688, 32703, 24653, 24650, 24637, 32687, 32667, 32699, 24649, 24628, 32675, 32711, 24662, 32715, 32679, 24630, 32810, 32814, 24764, 32822, 24771, 24770, 32822, 32825, 24773, 32810, 24761, 24758, 32814, 32821, 24770, 32832, 32830, 24777, 24789, 32837, 32836, 32841, 32837, 24789, 24797, 32845, 32844, 24796, 32844, 32841, 32836, 32832, 24779, 32855, 24808, 24807, 32855, 32858, 24810, 32861, 24814, 24813, 32860, 24813, 24810, 24777, 32830, 32811, 32840, 24792, 24773, 32840, 32854, 24807, 24690, 32740, 32739, 32752, 24703, 24702, 24694, 32744, 32743, 32756, 24707, 24706, 32756, 32759, 24709, 24693, 32743, 32740, 32755, 24706, 24703, 32780, 32782, 24729, 32786, 32788, 24735, 32782, 32786, 24733, 24715, 32767, 32766, 32774, 32772, 24721, 32772, 32767, 24715, 32794, 32792, 24741, 32802, 32800, 24749, 32800, 32794, 24743, 24714, 32766, 32744, 32759, 32780, 24727, 32792, 32774, 24723, 24836, 32883, 32882, 32880, 32883, 24836, 24823, 32874, 32873, 32871, 32874, 24823,
24838, 32886, 32880, 32873, 32886, 24838, 32916, 24868, 24867, 32916, 32913, 24864, 32920, 24872, 24871, 32919, 24871, 24864, 32900, 24850, 24849, 32900, 32897, 24846, 32909, 32907, 24854, 32907, 32899, 24849, 32909, 24856, 24840, 32920, 32888, 24840, 24689, 32739, 32692, 32751, 24702, 24654, 24835, 32882, 32845, 32915, 24867, 24814, 32727, 32808, 24758, 32811, 32731, 24682, 32788, 32871, 24819, 32897, 32802, 24751, 32478, 32480, 24429, 32486, 32488, 24437, 32480, 32486, 24435, 32494, 32496, 24445, 32502, 32504, 24453, 32496, 32502, 24451, 32488, 32494, 24443, 32576, 32578, 24527, 32584, 32586, 24535, 32578, 32584, 24533, 32592, 32594, 24543, 32600, 32602, 24551, 32594, 32600, 24549, 32586, 32592, 24541, 32527, 24481, 24480, 32544, 32542, 24489, 32531, 24485, 24484, 32530, 24484, 24481, 32542, 32526, 24480, 32566, 32564, 24511, 32572, 32570, 24517, 32570, 32566, 24513, 32564, 32544, 24491, 32626, 32624, 24571, 32632, 32630, 24577, 32630, 32626, 24573, 32654, 32652, 24599, 32660, 32658, 24605, 32658, 32654, 24601, 32652, 32632, 24579, 32504, 32576, 24525, 32624, 32572, 24519, 32930, 24879, 24878, 32930, 32933, 24881, 32938, 32940, 24887, 32933, 32938, 24885, 32940, 32944, 24890, 32979, 32981, 24934, 32994, 24948, 24947, 32991, 32993, 24947,
32981, 32991, 24944, 32965, 32963, 24910, 24920, 32970, 32969, 32973, 32970, 24920, 32969, 32965, 24912, 33009, 33007, 24958, 24972, 33018, 33017, 33021, 33018, 24972, 24958, 33007, 33005, 33017, 33009, 24960, 33005, 32973, 24922, 33036, 24987, 24986, 33033, 33035, 24986, 33033, 24983, 24981, 33036, 33044, 24990, 33058, 25012, 25011, 33055, 33057, 25011, 33072, 25022, 25021, 33072, 33075, 25024, 33071, 25021, 25018, 33055, 25008, 24990, 33058, 33069, 25018, 32979, 24932, 24890, 32994, 33031, 24981, 32602, 32929, 24878, 32963, 32660, 24607, 32148, 32351, 24302, 32432, 32189, 24143, 32283, 32478, 24427, 33081, 33083, 25034, 25038, 33086, 33085, 25037, 33085, 33081, 25052, 33098, 33097, 33102, 25056, 25055, 25060, 33106, 33105, 33110, 25064, 25063, 25059, 33105, 33098, 33109, 25063, 25056, 25051, 33097, 33086, 33101, 25055, 25040, 33153, 33155, 25106, 33167, 33169, 25120, 33159, 33157, 25108, 33173, 33171, 25122, 33155, 33167, 25118, 33171, 33159, 25110, 33195, 33197, 25148, 33209, 33211, 25162, 33201, 33199, 25150, 33216, 25168, 25167, 33216, 33213, 25164, 33215, 25167, 25162, 33197, 33209, 25160, 33213, 33201, 25152, 25130, 33182, 33181, 33187, 33189, 25138, 33181, 33187, 25136, 33225, 33227, 25176, 33236, 25186, 25185, 33233, 33235, 25185,
33227, 33233, 25182, 33169, 33195, 25146, 33199, 33173, 25124, 33189, 33225, 25174, 25076, 33122, 33121, 33126, 25080, 25079, 25084, 33130, 33129, 25083, 33129, 33122, 33133, 25086, 25080, 25094, 33144, 33143, 33147, 33144, 25094, 25093, 33143, 33130, 33257, 33255, 25204, 33246, 25196, 25195, 33246, 33249, 25198, 33259, 33261, 25210, 33255, 33243, 25192, 33249, 33259, 25208, 33285, 33283, 25232, 33297, 33295, 25244, 33287, 33289, 25238, 33299, 33301, 25250, 33295, 33285, 25234, 33289, 33299, 25248, 33269, 33267, 25216, 33277, 33275, 25224, 33275, 33269, 25218, 33309, 33307, 25256, 33318, 25268, 25267, 33318, 33315, 25264, 33315, 33309, 25258, 33317, 25267, 25250, 33283, 33257, 25206, 33261, 33287, 25236, 33307, 33277, 25226, 25075, 33121, 33106, 33125, 25079, 25064, 25192, 33243, 33182, 33297, 25246, 25186, 33083, 33153, 25104, 33157, 33089, 25040, 33133, 33245, 25195, 33267, 33147, 25096, 33489, 33491, 25442, 25446, 33494, 33493, 33506, 25455, 25454, 25445, 33493, 33489, 33505, 25454, 25451, 33491, 33503, 25451, 25467, 33513, 33512, 33529, 25479, 25478, 25471, 33517, 33516, 33533, 25483, 25482, 25470, 33516, 33513, 33532, 25482, 25479, 33540, 33542, 25491, 33552, 33554, 25503, 33546, 33544, 25493, 33558, 33556, 25505, 25493, 33544, 33540,
33542, 33552, 25501, 33556, 33546, 25495, 25466, 33512, 33494, 33528, 25478, 25455, 25519, 33565, 33564, 33581, 25533, 25532, 25523, 33569, 33568, 33587, 25539, 25538, 33584, 33586, 25538, 25522, 33568, 33565, 33584, 25535, 25533, 33614, 33616, 25563, 33620, 33622, 25569, 33616, 33620, 25567, 25547, 33597, 33596, 33600, 33597, 25547, 33608, 33606, 25555, 33606, 33600, 25549, 33628, 33626, 25575, 33636, 33634, 25583, 33634, 33628, 25577, 25546, 33596, 33569, 33587, 33614, 25561, 33626, 33608, 25557, 25518, 33564, 33517, 33580, 25532, 25483, 25276, 33326, 33325, 33325, 33329, 25278, 33335, 33337, 25286, 33343, 33345, 25294, 33337, 33343, 25292, 33329, 33335, 25284, 33401, 33403, 25352, 33409, 33411, 25360, 33403, 33409, 25358, 33417, 33419, 25368, 33425, 33427, 25376, 33419, 33425, 25374, 33411, 33417, 25366, 25306, 33352, 33351, 25310, 33356, 33355, 25309, 33355, 33352, 33369, 33367, 25314, 33391, 33389, 25336, 33397, 33395, 25342, 33395, 33391, 25338, 25314, 33367, 33356, 33389, 33369, 25316, 33451, 33449, 25396, 33457, 33455, 25402, 33455, 33451, 25398, 33479, 33477, 25424, 33485, 33483, 25430, 33483, 33479, 25426, 33477, 33457, 25404, 25305, 33351, 33326, 33345, 33401, 25350, 33449, 33397, 25344, 33642, 33644, 25593, 33650, 33652, 25601,
33644, 33650, 25599, 33658, 33660, 25609, 33666, 33668, 25617, 33660, 33666, 25615, 33652, 33658, 25607, 33730, 33732, 25681, 33738, 33740, 25689, 33732, 33738, 25687, 33749, 25699, 25698, 33746, 33748, 25698, 33756, 33758, 25705, 33749, 33756, 25703, 33740, 33746, 25695, 33692, 33690, 25637, 33698, 33696, 25643, 33696, 33692, 25639, 33720, 33718, 25665, 33726, 33724, 25671, 33724, 33720, 25667, 33718, 33698, 25645, 33780, 33778, 25725, 33786, 33784, 25731, 33784, 33780, 25727, 33808, 33806, 25753, 25764, 33813, 33812, 33816, 33813, 25764, 25768, 33818, 33816, 33812, 33808, 25755, 33806, 33786, 25733, 33668, 33730, 25679, 33778, 33726, 25673, 33427, 33642, 25591, 33690, 33485, 25432, 25779, 33829, 33828, 33828, 33832, 25781, 33825, 33829, 25779, 33873, 33875, 25826, 33882, 25834, 25833, 33882, 33878, 25830, 25829, 33877, 33873, 33893, 25840, 25838, 33875, 33891, 25838, 33893, 33881, 25833, 25792, 33839, 33838, 33838, 33842, 25794, 33858, 25812, 25811, 25800, 33847, 33846, 33846, 33844, 25796, 33863, 33861, 25814, 33861, 33847, 25800, 33904, 25856, 25855, 33901, 33903, 25855, 33901, 25852, 25850, 33904, 33913, 25859, 25794, 33842, 33825, 33897, 25849, 25830, 33832, 33877, 25829, 33897, 33857, 25811, 33858, 33898, 25850, 25875, 33923, 33922,
33916, 33918, 25865, 25879, 33927, 33926, 33926, 33930, 25881, 25881, 33930, 33923, 33918, 33927, 25879, 33969, 25918, 25917, 33957, 25911, 25910, 33972, 25920, 25918, 33972, 33956, 25910, 33940, 33938, 25887, 33949, 25899, 25898, 33949, 33946, 25895, 33946, 33940, 25889, 33979, 33977, 25924, 33979, 25926, 25911, 33977, 33948, 25898, 25874, 33922, 33863, 33968, 25917, 25859, 33986, 33988, 25941, 33999, 25952, 25951, 33999, 34002, 25954, 34004, 25956, 25954, 33988, 33998, 25951, 34036, 25985, 25984, 34036, 34039, 25987, 25968, 34014, 34013, 34017, 34014, 25968, 34045, 25999, 25998, 34059, 26008, 26007, 34058, 26007, 26004, 34045, 34056, 26004, 34044, 25998, 25987, 34004, 34035, 25984, 33986, 25939, 25929, 33554, 33839, 25792, 33844, 33558, 25507, 33622, 33916, 25863, 33938, 33636, 25585, 33758, 33983, 25929, 34013, 33818, 25768, 31703, 31935, 23885, 5433, 16227, 16226, 5447, 16235, 16234, 5446, 16234, 16232, 16232, 16227, 5433, 5457, 16247, 16246, 16250, 16247, 5457, 5463, 16253, 16252, 5462, 16252, 16250, 5456, 16246, 16235, 16345, 5557, 5556, 16359, 5566, 5565, 16358, 5565, 5562, 16345, 16356, 5562, 16376, 5588, 5587, 16373, 16375, 5587, 16380, 5592, 5591, 16379, 5591, 5588, 16373, 5584, 5566, 16380, 16391, 5595,
5510, 16298, 16297, 5514, 16302, 16301, 16305, 16302, 5514, 5513, 16301, 16298, 5531, 16324, 16323, 5545, 16333, 16332, 5533, 16327, 16324, 16332, 16327, 5533, 5530, 16323, 16305, 16297, 16294, 5498, 16427, 5637, 5636, 16433, 5643, 5642, 16430, 16432, 5642, 16430, 5639, 5637, 16453, 5663, 5662, 16459, 5669, 5668, 16456, 16458, 5668, 16456, 5665, 5663, 16442, 16452, 5662, 16442, 5653, 5643, 5498, 16294, 16253, 16426, 5636, 5595, 5833, 16627, 16626, 5841, 16633, 16632, 5845, 16637, 16636, 5844, 16636, 16633, 5840, 16632, 16627, 5861, 16650, 16649, 16647, 16650, 5861, 5863, 16653, 16647, 16649, 16662, 5873, 5932, 16721, 16720, 5938, 16727, 16726, 16726, 16724, 5934, 16739, 16737, 5948, 5934, 16724, 16721, 16737, 16727, 5938, 16772, 16770, 5975, 16778, 16776, 5981, 16776, 16772, 5977, 5953, 16749, 16739, 16778, 5983, 5971, 16770, 16749, 5953, 5848, 16644, 16637, 16626, 16653, 5863, 16720, 16644, 5848, 5679, 16471, 16470, 16470, 16468, 5675, 5681, 16474, 16471, 5693, 16481, 16480, 16538, 5750, 5749, 16538, 16535, 5746, 5710, 16503, 16502, 16502, 16500, 5706, 5724, 16512, 16511, 16515, 16512, 5724, 5723, 16511, 16509, 16509, 16503, 5710, 16558, 5770, 5769, 5774, 16562, 16561,
16558, 16562, 5774, 16566, 5778, 5777, 5782, 16570, 16569, 16569, 16565, 5777, 5706, 16500, 16481, 16557, 5769, 5746, 5773, 16561, 16538, 5796, 16585, 16584, 16604, 5816, 5815, 16601, 16603, 5815, 5802, 16591, 16590, 16590, 16588, 5798, 5798, 16588, 16585, 16607, 16591, 5802, 5827, 16620, 16619, 16619, 16617, 5823, 5823, 16617, 16607, 16584, 16581, 5785, 5785, 16581, 16515, 16601, 5812, 5778, 5815, 16603, 16570, 16480, 16474, 5681, 16783, 5993, 5992, 5997, 16787, 16786, 16786, 16790, 6000, 16791, 16787, 5997, 16800, 16802, 6009, 16786, 5996, 5993, 16790, 16800, 6007, 16829, 6036, 6035, 16829, 16832, 6038, 16808, 16810, 6021, 16850, 6058, 6057, 16854, 6062, 6061, 16853, 6061, 6058, 16837, 16849, 6057, 16854, 16839, 6050, 16808, 6019, 5997, 6021, 16810, 16791, 16837, 6048, 6038, 16802, 16828, 6035, 6074, 16862, 16861, 16865, 16862, 6074, 16881, 16879, 6090, 16867, 16869, 6080, 16883, 16885, 6096, 16867, 6078, 6074, 16879, 16865, 6076, 16869, 16883, 6094, 16923, 6135, 6134, 16938, 16940, 6148, 6137, 16926, 16920, 16938, 6146, 6144, 16923, 16936, 6144, 6111, 16900, 16899, 16899, 16897, 6107, 6107, 16897, 16895, 16910, 16900, 6111, 16920, 16881, 6092, 16885, 16926, 6137,
16964, 6171, 6170, 16964, 16967, 6173, 16983, 6194, 6193, 16983, 16986, 6196, 16972, 16982, 6193, 17000, 6209, 6208, 16997, 16999, 6208, 17008, 17010, 6215, 16997, 6205, 6203, 17000, 17008, 6213, 16972, 6183, 6173, 16986, 16995, 6203, 16922, 6134, 6050, 16940, 16963, 6170, 5675, 16468, 16333, 16533, 5744, 5669, 16533, 16782, 5992, 16786, 16537, 5749, 16538, 16787, 5997, 16861, 16603, 5815, 16604, 16862, 6074, 16895, 16620, 5827, 6224, 17017, 17016, 17016, 17014, 6220, 17025, 17023, 6231, 6247, 17035, 17034, 17038, 17035, 6247, 6235, 17027, 17025, 17034, 17027, 6235, 17023, 17017, 6224, 17056, 17054, 6265, 17067, 6279, 6278, 6283, 17071, 17070, 17070, 17066, 6278, 17067, 17071, 6283, 17067, 17056, 6267, 6296, 17085, 17084, 17082, 17085, 6296, 17108, 6320, 6319, 17108, 17104, 6316, 6300, 17089, 17088, 17093, 6304, 6303, 6324, 17112, 17111, 17111, 17115, 6326, 6299, 17088, 17082, 17111, 6323, 6316, 6326, 17115, 17108, 17084, 17103, 6315, 17111, 17092, 6303, 17093, 17112, 6324, 6356, 17144, 17143, 6315, 17103, 17067, 17143, 17107, 6319, 17108, 17144, 6356, 6255, 17049, 17048, 6380, 17168, 17167, 6379, 17167, 17164, 6397, 17190, 17189, 17189, 17187, 6393, 17200, 17198, 6407,
6407, 17198, 17196, 17196, 17190, 6397, 6438, 17227, 17226, 17226, 17224, 6434, 17240, 6452, 6451, 17240, 17237, 6448, 17237, 17227, 6438, 6393, 17187, 17168, 17224, 17200, 6409, 6375, 17163, 17089, 17054, 17038, 6249, 17048, 17163, 6375, 17164, 17049, 6255, 6220, 17014, 16910, 17133, 6344, 6215, 17251, 17253, 6464, 17264, 6474, 6473, 17270, 6480, 6479, 17267, 17269, 6479, 17267, 6476, 6474, 17253, 17263, 6473, 17303, 6511, 6510, 17307, 6515, 6514, 17306, 6514, 6511, 17287, 6498, 6480, 17287, 17302, 6510, 17307, 17289, 6500, 6551, 17344, 17343, 17343, 17341, 6547, 17354, 17352, 6561, 6561, 17352, 17350, 17350, 17344, 6551, 6575, 17363, 17362, 17366, 17363, 6575, 17376, 17366, 6577, 6596, 17389, 17388, 17388, 17386, 6592, 6592, 17386, 17376, 17362, 17354, 6563, 17331, 6540, 6539, 17331, 17328, 6536, 17317, 6529, 6528, 17317, 17314, 6525, 17334, 6542, 6536, 17334, 17316, 6528, 17330, 6539, 6500, 17341, 17299, 6503, 17490, 17492, 6703, 17503, 6714, 6713, 17503, 17506, 6716, 17508, 6718, 6716, 17492, 17502, 6713, 17468, 17466, 6674, 6691, 17480, 17479, 17479, 17477, 6687, 6678, 17470, 17468, 17477, 17470, 6678, 17519, 17517, 6728, 17517, 17480, 6691, 17552, 6760, 6759,
17549, 17551, 6759, 17552, 17561, 6766, 6794, 17586, 17585, 17589, 17586, 6794, 17599, 17597, 6806, 6806, 17597, 17595, 17595, 17589, 6796, 17573, 17575, 6786, 17608, 6819, 6818, 17608, 17611, 6821, 17625, 6834, 6833, 17622, 17624, 6833, 17622, 6830, 6828, 17611, 17620, 6828, 17573, 6784, 6766, 17585, 17559, 6764, 17575, 17607, 6818, 6750, 17543, 17542, 17542, 17540, 6746, 17540, 17537, 6741, 17635, 17633, 6841, 6857, 17645, 17644, 17648, 17645, 6857, 6845, 17637, 17635, 17644, 17637, 6845, 17660, 17658, 6869, 6891, 17684, 17683, 17683, 17681, 6887, 17681, 17678, 6882, 6882, 17678, 17660, 17658, 17648, 6859, 6741, 17537, 17519, 17508, 17549, 6756, 17633, 17543, 6750, 6617, 17405, 17404, 17404, 17408, 6619, 6605, 17398, 17397, 6604, 17397, 17395, 17395, 17405, 6617, 17410, 17398, 6605, 17429, 6641, 6640, 17426, 17428, 6640, 17433, 6645, 6644, 17432, 6644, 6641, 17422, 17420, 6625, 17446, 17444, 6649, 6625, 17420, 17410, 17446, 6651, 6645, 17408, 17426, 6637, 17444, 17422, 6627, 17758, 17760, 6969, 6957, 17748, 17747, 17747, 17751, 6959, 6959, 17751, 17745, 17745, 17758, 6967, 6938, 17727, 17726, 17726, 17730, 6940, 6946, 17735, 17734, 17734, 17732, 6942, 6942, 17732, 17727,
6940, 17730, 17723, 17723, 17748, 6957, 6902, 17693, 17692, 17692, 17696, 6904, 6904, 17696, 17690, 17703, 17705, 6914, 17713, 17715, 6924, 17705, 17713, 6922, 17690, 17703, 6912, 7015, 17806, 17805, 17809, 17806, 7015, 17820, 17818, 7027, 7019, 17811, 17809, 17818, 17811, 7019, 17794, 7003, 7002, 17789, 17793, 7002, 17794, 17790, 6999, 17789, 6998, 6995, 17794, 17802, 7006, 17805, 17784, 6988, 6988, 17784, 17735, 17715, 17787, 6995, 17466, 17389, 6596, 7617, 18405, 18404, 18408, 18405, 7617, 7645, 18438, 18437, 18437, 18435, 7641, 18428, 18430, 7636, 18435, 18426, 7632, 18462, 18460, 7668, 7685, 18474, 18473, 18473, 18471, 7681, 7672, 18464, 18462, 18471, 18464, 7672, 7632, 18426, 18408, 18460, 18438, 7645, 18575, 18577, 7788, 18588, 7799, 7798, 18588, 18591, 7801, 18593, 7803, 7801, 18577, 18587, 7798, 18627, 7835, 7834, 18624, 18626, 7834, 18634, 18636, 7841, 18627, 18634, 7839, 18604, 18602, 7813, 18614, 18604, 7815, 18648, 7859, 7841, 18593, 18624, 7831, 18497, 7709, 7708, 18494, 18496, 7708, 18484, 18494, 7705, 18529, 18531, 7740, 18541, 18543, 7750, 18533, 7742, 7740, 18533, 18541, 7748, 18510, 18508, 7716, 18521, 18519, 7728, 7720, 18512, 18510, 18519, 18512, 7720,
7762, 18550, 18549, 18553, 18550, 7762, 18565, 18563, 7774, 18563, 18553, 7764, 18497, 18529, 7738, 18549, 18521, 7730, 18669, 18667, 7873, 18659, 7866, 7865, 18659, 18662, 7868, 18662, 18671, 7877, 7905, 18697, 18696, 18700, 18697, 7905, 18708, 18706, 7914, 7918, 18710, 18708, 18706, 18700, 7907, 18684, 18686, 7897, 18720, 7932, 7931, 18717, 18719, 7931, 18731, 18733, 7942, 18720, 18731, 7940, 18684, 7895, 7877, 18696, 18669, 7875, 18686, 18717, 7928, 18484, 7695, 7636, 7873, 18667, 18614, 18602, 18474, 7685, 18543, 18658, 7865, 18933, 8145, 8144, 18930, 18932, 8144, 18944, 18946, 8155, 18933, 18944, 8153, 18956, 18958, 8166, 18968, 8176, 8175, 18965, 18967, 8175, 18956, 8164, 8162, 18958, 18965, 8172, 18946, 18954, 8162, 8273, 19061, 19060, 8277, 19065, 19064, 8276, 19064, 19061, 19057, 19065, 8277, 19110, 8318, 8317, 19114, 8322, 8321, 19113, 8321, 8318, 19114, 19099, 8310, 8286, 19077, 19076, 8292, 19083, 19082, 19080, 19083, 8292, 8291, 19082, 19077, 19122, 8334, 8333, 19126, 8338, 8337, 19125, 8337, 8334, 8285, 19076, 19057, 19121, 8333, 8310, 18977, 18975, 8184, 18987, 18985, 8194, 18985, 18977, 8186, 8228, 19016, 19015, 19019, 19016, 8228, 19031, 19029, 8240,
19029, 19019, 8230, 18995, 18997, 8206, 19007, 19009, 8216, 18999, 8208, 8206, 18999, 19007, 8214, 19041, 19043, 8250, 19049, 19051, 8258, 19043, 19049, 8256, 19015, 18987, 8196, 19009, 19041, 8248, 19139, 19137, 8348, 19151, 19149, 8360, 19149, 19139, 8350, 19191, 19193, 8404, 19181, 19179, 8390, 19197, 19195, 8406, 19177, 19191, 8402, 19195, 19181, 8392, 19161, 19163, 8370, 19169, 19171, 8378, 19163, 19169, 8376, 19207, 19209, 8416, 19217, 19219, 8427, 8427, 19219, 19215, 19209, 19217, 8425, 19179, 19151, 8362, 19171, 19207, 8414, 19177, 8388, 8338, 18968, 19080, 8288, 19137, 19031, 8242, 19051, 19161, 8368, 18741, 18743, 7952, 18754, 7961, 7960, 18754, 18757, 7963, 18745, 7954, 7952, 18745, 18753, 7960, 18771, 7983, 7982, 18775, 7987, 7986, 18774, 7986, 7983, 18770, 7982, 7963, 18821, 18819, 8024, 8035, 18828, 18827, 18827, 18825, 8031, 18825, 18821, 8026, 18836, 18834, 8041, 18844, 18842, 8049, 18842, 18836, 8043, 18834, 18828, 8035, 18802, 8009, 8008, 18805, 18801, 8008, 18793, 18791, 8000, 18793, 8002, 7998, 18802, 18789, 7998, 8020, 18813, 18812, 18810, 18813, 8020, 8107, 18895, 18894, 18894, 18898, 8109, 8093, 18884, 18883, 18883, 18887, 8095, 8095, 18887, 18881,
18881, 18895, 8107, 8080, 18868, 18867, 18867, 18871, 8082, 18858, 18860, 8068, 8068, 18860, 18856, 18856, 18868, 8080, 18919, 8131, 8130, 18919, 18916, 8127, 8082, 18871, 18853, 18853, 18884, 8093, 18805, 8011, 7987, 8109, 18898, 18850, 18819, 18786, 7990, 18812, 18858, 8066, 19228, 19226, 8433, 8450, 19239, 19238, 19238, 19236, 8446, 8446, 19236, 19234, 19234, 19228, 8435, 8475, 19263, 19262, 19262, 19266, 8477, 8461, 19252, 19251, 19251, 19255, 8463, 8463, 19255, 19249, 19249, 19263, 8475, 19301, 8512, 8511, 19305, 8516, 8515, 19308, 19304, 8515, 19308, 8518, 8512, 8477, 19266, 19239, 19388, 19386, 8593, 19374, 8585, 8584, 19377, 19373, 8584, 19377, 8587, 8581, 19386, 19371, 8581, 19399, 19397, 8604, 19406, 8613, 8612, 19409, 19405, 8612, 19406, 19399, 8606, 19394, 8598, 8585, 19397, 19388, 8595, 19328, 8540, 8539, 19328, 19325, 8536, 19358, 8567, 8566, 19358, 19355, 8563, 19361, 8569, 8563, 19368, 19357, 8566, 19342, 8551, 8550, 19342, 19339, 8547, 19345, 8553, 8547, 19352, 19341, 8550, 19352, 8556, 8536, 19361, 19327, 8539, 19368, 8572, 8516, 19394, 19300, 8511, 7990, 18786, 18565, 18648, 18930, 8141, 18975, 18710, 7918, 18733, 18995, 8204, 19226, 18844, 8051,
18850, 19252, 8461, 19345, 18918, 8130, 17829, 7040, 7039, 17829, 17832, 7042, 17835, 7046, 7045, 7050, 17839, 17838, 17838, 17834, 7045, 17835, 17839, 7050, 17851, 17853, 7061, 17851, 7059, 7057, 17832, 17849, 7057, 17883, 7091, 7090, 17880, 17882, 7090, 17890, 17892, 7097, 17883, 17890, 7095, 17870, 7081, 7079, 7109, 17898, 17896, 17853, 17880, 7087, 17896, 17868, 7079, 17870, 17898, 7109, 17984, 17986, 7197, 17996, 17998, 7209, 17986, 17996, 7207, 18011, 7223, 7222, 18008, 18010, 7222, 18022, 18024, 7233, 18011, 18022, 7231, 17984, 7195, 7185, 17998, 18008, 7219, 17957, 17955, 7164, 17960, 7170, 7169, 17973, 17971, 7180, 17959, 7169, 7164, 7180, 17971, 17969, 17969, 17957, 7166, 17947, 7154, 7153, 17950, 17946, 7153, 17938, 17936, 7145, 17938, 7147, 7143, 17947, 17934, 7143, 17950, 7156, 7135, 17960, 17924, 7135, 18034, 18032, 7241, 18044, 18042, 7251, 18042, 18034, 7243, 7266, 18055, 18054, 18054, 18052, 7262, 18067, 18065, 7276, 18065, 18055, 7266, 7286, 18078, 18077, 18089, 18091, 7302, 7290, 18082, 18081, 7289, 18081, 18078, 18077, 18089, 7300, 18052, 18044, 7253, 17892, 17981, 7185, 18032, 17973, 7182, 7318, 18111, 18110, 7334, 18122, 18121, 7320, 18114, 18111,
18110, 18119, 7330, 18121, 18114, 7320, 18133, 18135, 7344, 7360, 18148, 18147, 18147, 18143, 7355, 18144, 18148, 7360, 18151, 7362, 7360, 18135, 18144, 7356, 18179, 18177, 7388, 18195, 18193, 7404, 18181, 18183, 7394, 18197, 18199, 7410, 7394, 18183, 18179, 18197, 7408, 7404, 7355, 18143, 18122, 18177, 18148, 7360, 18151, 18181, 7392, 18275, 18277, 7488, 18290, 7498, 7497, 18294, 7502, 7501, 18293, 7501, 7498, 18294, 18279, 7490, 18331, 18333, 7538, 7547, 18340, 18339, 18337, 18340, 7547, 18333, 18337, 7543, 18304, 7516, 7515, 7530, 18318, 18317, 18317, 18321, 7532, 7518, 18307, 18301, 18304, 18318, 7530, 18303, 7515, 7490, 7557, 18346, 18331, 18321, 18346, 7557, 18211, 18209, 7420, 18223, 18221, 7432, 18221, 18211, 7422, 18253, 7464, 7463, 18256, 18252, 7463, 18267, 18265, 7474, 18265, 18256, 7466, 18234, 7445, 7444, 18234, 18237, 7447, 18242, 7453, 7452, 18242, 18239, 7449, 18241, 7452, 7447, 18253, 18223, 7434, 18378, 18376, 7583, 18368, 18366, 7575, 18368, 7577, 7573, 18376, 18364, 7573, 18386, 18384, 7591, 18393, 7600, 7599, 18396, 18392, 7599, 18393, 18386, 7593, 18384, 18378, 7585, 18301, 18195, 7406, 18199, 18307, 7518, 18366, 18267, 7476, 18106, 18104, 7309,
18104, 18101, 7305, 7305, 18101, 18082, 18401, 7605, 7449, 18024, 18133, 7342, 18209, 18067, 7278, 18091, 18233, 7444, 18401, 18106, 7311, 19422, 19424, 8633, 19414, 19416, 8623, 8635, 19426, 19422, 19416, 19426, 8635, 8671, 19459, 19458, 19458, 19462, 8673, 19462, 19472, 8683, 19447, 8659, 8658, 19447, 19444, 8655, 19444, 19434, 8645, 19497, 8705, 8704, 19497, 19494, 8701, 19486, 19484, 8693, 19486, 8695, 8691, 19494, 19482, 8691, 19424, 19459, 8671, 19484, 19446, 8658, 19521, 8732, 8731, 19524, 19520, 8731, 19521, 19510, 8721, 19542, 8751, 8750, 19542, 19539, 8747, 19545, 8753, 8747, 19552, 19541, 8750, 19535, 19533, 8738, 19535, 8740, 8721, 19545, 19524, 8734, 19506, 19504, 8709, 19506, 8711, 8645, 19533, 19496, 8704, 7330, 18119, 17693, 18275, 7486, 6999, 7488, 18277, 17794, 18289, 7497, 7006, 8683, 19472, 18810, 19552, 8756, 8127, 18404, 17599, 6808, 17625, 18428, 7634, 18508, 17684, 6891, 17760, 18741, 7950, 18791, 17820, 7029, 18339, 19414, 8621, 19504, 18396, 7602, 17314, 16760, 5971, 17133, 17828, 7039, 17936, 17239, 6451, 19555, 19557, 8762, 19561, 19563, 8768, 19557, 19561, 8766, 8896, 19689, 19688, 19686, 19689, 8896, 19695, 19699, 8907, 19700, 19696, 8904,
19688, 19695, 8903, 19708, 8914, 8913, 19699, 19707, 8913, 19708, 19700, 8908, 19567, 19569, 8780, 19585, 19587, 8796, 19574, 8786, 8785, 19574, 19571, 8782, 19591, 19589, 8798, 19569, 19585, 8794, 19589, 19573, 8785, 19644, 19642, 8851, 19634, 19632, 8841, 19646, 19648, 8857, 19634, 8843, 8839, 19642, 19630, 8839, 19599, 19601, 8810, 8825, 19614, 19613, 19611, 19614, 8825, 19619, 19617, 8827, 19601, 19611, 8821, 19617, 19603, 8812, 19656, 19658, 8868, 19677, 19674, 8883, 19661, 8872, 8871, 19664, 19660, 8871, 19679, 8887, 8883, 19658, 19673, 8882, 19679, 19664, 8874, 8882, 19673, 19646, 19587, 19628, 8837, 19632, 19591, 8800, 19613, 19656, 8866, 19661, 19619, 8829, 19714, 8925, 8924, 19717, 19713, 8924, 19735, 19738, 8948, 8931, 19720, 19719, 19719, 19723, 8934, 19745, 8956, 8955, 19745, 19741, 8952, 8951, 19740, 19735, 19734, 19717, 8927, 19723, 19740, 8951, 19741, 19724, 8935, 8974, 19765, 19764, 19769, 8978, 8977, 19764, 19768, 8977, 19769, 19765, 8974, 19772, 8980, 8974, 19756, 8963, 8962, 19759, 19755, 8962, 19759, 8965, 8935, 19738, 19765, 8974, 19772, 19744, 8955, 8837, 19628, 19555, 19734, 8945, 8904, 19563, 19686, 8892, 19714, 19644, 8853, 19648, 19720, 8931,
19756, 19677, 8885, 8994, 19786, 19785, 8998, 19790, 19789, 8997, 19789, 19786, 19834, 9042, 9041, 19838, 9046, 9045, 19837, 9045, 9042, 9014, 19806, 19805, 9018, 19810, 19809, 9017, 19809, 19806, 19853, 19855, 9066, 9013, 19805, 19790, 19853, 9064, 9046, 9074, 19866, 19865, 9078, 19870, 19869, 9077, 19869, 19866, 19914, 9125, 9124, 19917, 19913, 9124, 9094, 19886, 19885, 9098, 19890, 19889, 9097, 19889, 19886, 19935, 9143, 9142, 19939, 9147, 9146, 19938, 9146, 9143, 9093, 19885, 19870, 19934, 9142, 9125, 9073, 19865, 19810, 19781, 19779, 8984, 9155, 19947, 19946, 9159, 19951, 19950, 9158, 19950, 19947, 20001, 9213, 9212, 19998, 20000, 9212, 9175, 19967, 19966, 9179, 19971, 19970, 19980, 19978, 9189, 9178, 19970, 19967, 19978, 19971, 9179, 20014, 20012, 9223, 20026, 20024, 9235, 20024, 20014, 9225, 9174, 19966, 19951, 20012, 19980, 9191, 19781, 8986, 8812, 9154, 19946, 19890, 19998, 9209, 9147, 20036, 20038, 9249, 20048, 20050, 9261, 20038, 20048, 9259, 20060, 20062, 9273, 20072, 20074, 9285, 20062, 20072, 9283, 20050, 20060, 9271, 20124, 20126, 9337, 20136, 20138, 9349, 20126, 20136, 9347, 20148, 20150, 9361, 20160, 20162, 9373, 20150, 20160, 9371, 20138, 20148, 9359,
20086, 20084, 9293, 20096, 20094, 9303, 20094, 20086, 9295, 20106, 20104, 9313, 20116, 20114, 9323, 20114, 20106, 9315, 20104, 20096, 9305, 20174, 20172, 9381, 20184, 20182, 9391, 20182, 20174, 9383, 20194, 20192, 9401, 20204, 20202, 9411, 20202, 20194, 9403, 20192, 20184, 9393, 20074, 20124, 9335, 20172, 20116, 9325, 20375, 20377, 9588, 20387, 20389, 9600, 20377, 20387, 9598, 20400, 9608, 9607, 20404, 9612, 9611, 20403, 9611, 9608, 20389, 20399, 9607, 9649, 20441, 20440, 9653, 20445, 20444, 9652, 20444, 20441, 20489, 9697, 9696, 20493, 9701, 9700, 20492, 9700, 9697, 9669, 20461, 20460, 9673, 20465, 20464, 9672, 20464, 20461, 20509, 9717, 9716, 20513, 9721, 9720, 20512, 9720, 9717, 9668, 20460, 20445, 20508, 9716, 9701, 20413, 20411, 9620, 20423, 20421, 9630, 20421, 20413, 9622, 20432, 9639, 9638, 20435, 9641, 9639, 20435, 20423, 9632, 9729, 20521, 20520, 9733, 20525, 20524, 9732, 20524, 20521, 20569, 9777, 9776, 20573, 9781, 9780, 20572, 9780, 9777, 9749, 20541, 20540, 9753, 20545, 20544, 9752, 20544, 20541, 20589, 9797, 9796, 20593, 9801, 9800, 20592, 9800, 9797, 9748, 20540, 20525, 20588, 9796, 9781, 20431, 9638, 9612, 9728, 20520, 20465, 20568, 9776, 9721,
20214, 20216, 9424, 20223, 20225, 9432, 20214, 9422, 9420, 20216, 20223, 9430, 20255, 20257, 9464, 20263, 20265, 9472, 20257, 20263, 9470, 20233, 20231, 9442, 20245, 20243, 9454, 20243, 20233, 9444, 20273, 20271, 9482, 20285, 20283, 9494, 20283, 20273, 9484, 20225, 20255, 9462, 20271, 20245, 9456, 20295, 20297, 9504, 20303, 20305, 9512, 20297, 20303, 9510, 20335, 20337, 9544, 20343, 20345, 9552, 20337, 20343, 9550, 20313, 20311, 9522, 20325, 20323, 9534, 20323, 20313, 9524, 20353, 20351, 9562, 20365, 20363, 9574, 20363, 20353, 9564, 20305, 20335, 9542, 20351, 20325, 9536, 20265, 20295, 9502, 20311, 20285, 9496, 20600, 20602, 9809, 20608, 20610, 9817, 20602, 20608, 9815, 20643, 9851, 9850, 20640, 20642, 9850, 20618, 20616, 9827, 20630, 20628, 9839, 20628, 20618, 9829, 20651, 9863, 9862, 20657, 9869, 9868, 20657, 20654, 9865, 20656, 9868, 9863, 20650, 9862, 9851, 20610, 20640, 9847, 20654, 20630, 9841, 9877, 20669, 20668, 9881, 20673, 20672, 9880, 20672, 20669, 20721, 9929, 9928, 20725, 9933, 9932, 20724, 9932, 9929, 9897, 20689, 20688, 9901, 20693, 20692, 20702, 20700, 9911, 9900, 20692, 20689, 20700, 20693, 9901, 20745, 9953, 9952, 20734, 20732, 9943, 20749, 9957, 9956,
20748, 9956, 9953, 20749, 20734, 9945, 9896, 20688, 20673, 20744, 9952, 9933, 20732, 20702, 9913, 9876, 20668, 20545, 20720, 9928, 9801, 20162, 20375, 9586, 20411, 20204, 9413, 20345, 20600, 9807, 20616, 20365, 9576, 19855, 20036, 9247, 20084, 19917, 9127, 20001, 20212, 9420, 20231, 20026, 9237, 9965, 20757, 20756, 9969, 20761, 20760, 9968, 20760, 20757, 9977, 20769, 20768, 9981, 20773, 20772, 9980, 20772, 20769, 9976, 20768, 20761, 10021, 20813, 20812, 10025, 20817, 20816, 10024, 20816, 20813, 10033, 20825, 20824, 10037, 20829, 20828, 10036, 20828, 20825, 10032, 20824, 20817, 10020, 20812, 20773, 10077, 20869, 20868, 10081, 20873, 20872, 10080, 20872, 20869, 10089, 20881, 20880, 10093, 20885, 20884, 10092, 20884, 20881, 20910, 20908, 10119, 20922, 20920, 10131, 20920, 20910, 10121, 10088, 20880, 20873, 20908, 20885, 10093, 10076, 20868, 20829, 20933, 10141, 10140, 20937, 10145, 10144, 20936, 10144, 10141, 20945, 10153, 10152, 20949, 10157, 10156, 20948, 10156, 10153, 20944, 10152, 10145, 10193, 20981, 20980, 10197, 20985, 20984, 10196, 20984, 20981, 21038, 21040, 10251, 21028, 21038, 10249, 10217, 21005, 21004, 10221, 21009, 21008, 10220, 21008, 21005, 21066, 21064, 10271, 21053, 10265, 10264, 21052, 10264, 10261,
21064, 21050, 10261, 10216, 21004, 20985, 20957, 10165, 10164, 20961, 10169, 10168, 20960, 10168, 10165, 20969, 10177, 10176, 20973, 10181, 10180, 20972, 10180, 10177, 20968, 10176, 10169, 10285, 21073, 21072, 10289, 21077, 21076, 10288, 21076, 21073, 21121, 10333, 10332, 21125, 10337, 10336, 21124, 10336, 10333, 10309, 21097, 21096, 10313, 21101, 21100, 10312, 21100, 21097, 21137, 10349, 10348, 21141, 10353, 10352, 21140, 10352, 10349, 10308, 21096, 21077, 21136, 10348, 10337, 20956, 10164, 10157, 10284, 21072, 21009, 21120, 10332, 10265, 21276, 21278, 10489, 21288, 21290, 10501, 21278, 21288, 10499, 21316, 21318, 10529, 21328, 21330, 10541, 21318, 21328, 10539, 21302, 21300, 10507, 21310, 21308, 10515, 21308, 21302, 10509, 21342, 21340, 10547, 21350, 21348, 10555, 21348, 21342, 10549, 21290, 21316, 10527, 21340, 21310, 10517, 21356, 21358, 10569, 21368, 21370, 10581, 21358, 21368, 10579, 21396, 21398, 10609, 21408, 21410, 10621, 21398, 21408, 10619, 21382, 21380, 10587, 21390, 21388, 10595, 21388, 21382, 10589, 21422, 21420, 10627, 21430, 21428, 10635, 21428, 21422, 10629, 21370, 21396, 10607, 21420, 21390, 10597, 21330, 21356, 10567, 21380, 21350, 10557, 21153, 10361, 10360, 21157, 10365, 10364, 21156, 10364, 10361, 21165, 10373, 10372,
21169, 10377, 10376, 21168, 10376, 10373, 21164, 10372, 10365, 10389, 21177, 21176, 10393, 21181, 21180, 10392, 21180, 21177, 21231, 10443, 10442, 21244, 21246, 10453, 21234, 10445, 10443, 21234, 21244, 10451, 10413, 21201, 21200, 10417, 21205, 21204, 21208, 21205, 10417, 21220, 21218, 10429, 10416, 21204, 21201, 21218, 21208, 10419, 21254, 21252, 10463, 21266, 21264, 10475, 21264, 21254, 10465, 10412, 21200, 21181, 21252, 21220, 10431, 21436, 21438, 10645, 21444, 21446, 10653, 21438, 21444, 10651, 21476, 21478, 10685, 21484, 21486, 10693, 21478, 21484, 10691, 21454, 21452, 10663, 21466, 21464, 10675, 21464, 21454, 10665, 21494, 21492, 10703, 21506, 21504, 10715, 21504, 21494, 10705, 21446, 21476, 10683, 21492, 21466, 10677, 21516, 21518, 10725, 21524, 21526, 10733, 21518, 21524, 10731, 21556, 21558, 10765, 21564, 21566, 10773, 21558, 21564, 10771, 21534, 21532, 10743, 21546, 21544, 10755, 21544, 21534, 10745, 21574, 21572, 10783, 21586, 21584, 10795, 21584, 21574, 10785, 21526, 21556, 10763, 21572, 21546, 10757, 21486, 21516, 10723, 21532, 21506, 10717, 21152, 10360, 10181, 10388, 21176, 21101, 21230, 10442, 10353, 21040, 21276, 10487, 21300, 21066, 10273, 21246, 21436, 10643, 21452, 21266, 10477, 21169, 20922, 10133, 19193, 19567, 8778,
19571, 19197, 8408, 19215, 19599, 8808, 19779, 19409, 8615, 34065, 34067, 26020, 34077, 34079, 26032, 34067, 34077, 26030, 34109, 26062, 26061, 34109, 34112, 26064, 34121, 34123, 26074, 34125, 26076, 26074, 34112, 34121, 26072, 34091, 34089, 26038, 34099, 34097, 26047, 26051, 34101, 34099, 34097, 34091, 26040, 26088, 34134, 34133, 34137, 34134, 26088, 34079, 34108, 26061, 34133, 34101, 26051, 34196, 26145, 26144, 34196, 34199, 26147, 34215, 26168, 26167, 34215, 34218, 26170, 34221, 26174, 26173, 34220, 26173, 26170, 34204, 34214, 26167, 34204, 26157, 26147, 26112, 34159, 34158, 26118, 34165, 34164, 34164, 34162, 26114, 26114, 34162, 34159, 34158, 34155, 26101, 26134, 34181, 34180, 34178, 34181, 26134, 26138, 34185, 34184, 26137, 34184, 34178, 34175, 34185, 26138, 26133, 34180, 34165, 34270, 26220, 26219, 34274, 26224, 26223, 34273, 26223, 26220, 34258, 26212, 26211, 34282, 26231, 26230, 34285, 26233, 26231, 34281, 26230, 26224, 34239, 34269, 26219, 34285, 34257, 26211, 26101, 34155, 34137, 34239, 26192, 26174, 34125, 34195, 26144, 34301, 34303, 26254, 34292, 34294, 26244, 26244, 34294, 34290, 34290, 34301, 26252, 26291, 34337, 34336, 34336, 34340, 26293, 34313, 34311, 26264, 34324, 26277, 26276, 34327, 34323, 26276, 34324, 34313, 26266,
34373, 26323, 26322, 34373, 34370, 26319, 34362, 34360, 26311, 34362, 26313, 26309, 34370, 34358, 26309, 34303, 34337, 26291, 34360, 34327, 26279, 34397, 26348, 26347, 34397, 34394, 26344, 34383, 26337, 26336, 34383, 34380, 26333, 34400, 26350, 26344, 34400, 34382, 26336, 34407, 26353, 26333, 34407, 34372, 26322, 26293, 34340, 34175, 34396, 26347, 26212, 26365, 34411, 34410, 26369, 34415, 34414, 26368, 34414, 34411, 34459, 26413, 26412, 34463, 26417, 26416, 34462, 26416, 26413, 26389, 34435, 34434, 26393, 34439, 34438, 26392, 34438, 34435, 34484, 34486, 26439, 34474, 34484, 26437, 26388, 34434, 34415, 34474, 26427, 26417, 34582, 34584, 26537, 34594, 34596, 26549, 34584, 34594, 26547, 34606, 34608, 26561, 34618, 34620, 26573, 34608, 34618, 26571, 34596, 34606, 26559, 26451, 34497, 34496, 26455, 34501, 34500, 26454, 34500, 34497, 34545, 26499, 26498, 34558, 34556, 26507, 34556, 34544, 26498, 26475, 34521, 34520, 26479, 34525, 34524, 26478, 34524, 34521, 34567, 26521, 26520, 34571, 26525, 26524, 34570, 26524, 26521, 26474, 34520, 34501, 34566, 26520, 26499, 34632, 34630, 26581, 34642, 34640, 26591, 34640, 34632, 26583, 34652, 34650, 26601, 34662, 34660, 26611, 34660, 34652, 26603, 34650, 34642, 26593, 26450, 34496, 34439, 34486, 34582, 26535,
34630, 34558, 26509, 34850, 34852, 26805, 34862, 34864, 26817, 34852, 34862, 26815, 34874, 34876, 26829, 34886, 34888, 26841, 34876, 34886, 26839, 34864, 34874, 26827, 34938, 34940, 26893, 34951, 26901, 26900, 34955, 26905, 26904, 34954, 26904, 26901, 34940, 34950, 26900, 34900, 34898, 26849, 34910, 34908, 26859, 34908, 34900, 26851, 34920, 34918, 26869, 34930, 34928, 26879, 34928, 34920, 26871, 34918, 34910, 26861, 34973, 26922, 26921, 34964, 34962, 26913, 34976, 26924, 26922, 34976, 34964, 26915, 34972, 26921, 26905, 34888, 34938, 26891, 34962, 34930, 26881, 26625, 34671, 34670, 26629, 34675, 34674, 26628, 34674, 34671, 34725, 26679, 26678, 34738, 34740, 26689, 34728, 26681, 26679, 34728, 34738, 26687, 26649, 34695, 34694, 26653, 34699, 34698, 34702, 34699, 26653, 34714, 34712, 26665, 26652, 34698, 34695, 34712, 34702, 26655, 34748, 34746, 26699, 34760, 34758, 26711, 34758, 34748, 26701, 26648, 34694, 34675, 34746, 34714, 26667, 34770, 34772, 26721, 34778, 34780, 26729, 34772, 34778, 26727, 34810, 34812, 26761, 34818, 34820, 26769, 34812, 34818, 26767, 34788, 34786, 26739, 34800, 34798, 26751, 34798, 34788, 26741, 34828, 34826, 26779, 34840, 34838, 26791, 34838, 34828, 26781, 34780, 34810, 26759, 34826, 34800, 26753, 34740, 34770, 26719,
34786, 34760, 26713, 34981, 34983, 26932, 34989, 34991, 26940, 34983, 34989, 26938, 35021, 35023, 26972, 35029, 35031, 26980, 35023, 35029, 26978, 34999, 34997, 26950, 35011, 35009, 26962, 35009, 34999, 26952, 35039, 35037, 26990, 35051, 35049, 27002, 35049, 35039, 26992, 34991, 35021, 26970, 35037, 35011, 26964, 35061, 35063, 27012, 35069, 35071, 27020, 35063, 35069, 27018, 35102, 27050, 27049, 35079, 35077, 27030, 35091, 35089, 27042, 35089, 35079, 27032, 35108, 27058, 27057, 35112, 27062, 27061, 35111, 27061, 27058, 35107, 27057, 27050, 35071, 35101, 27049, 35112, 35091, 27044, 35031, 35061, 27010, 35077, 35051, 27004, 26624, 34670, 34525, 34724, 26678, 26525, 34620, 34850, 26803, 34898, 34662, 26613, 34820, 34981, 26930, 34997, 34840, 26793, 27074, 35120, 35119, 27078, 35124, 35123, 27077, 35123, 35120, 35168, 27122, 27121, 35172, 27126, 27125, 35171, 27125, 27122, 27098, 35144, 35143, 27102, 35148, 35147, 27101, 35147, 35144, 35184, 27138, 27137, 35188, 27142, 27141, 35187, 27141, 27138, 27097, 35143, 35124, 35183, 27137, 27126, 27154, 35200, 35199, 27158, 35204, 35203, 27157, 35203, 35200, 35248, 27202, 27201, 35252, 27206, 27205, 35251, 27205, 27202, 27178, 35224, 35223, 27182, 35228, 35227, 27181, 35227, 35224, 35264, 27218, 27217,
35268, 27222, 27221, 35267, 27221, 27218, 27177, 35223, 35204, 35263, 27217, 27206, 27153, 35199, 35148, 35247, 27201, 27142, 27234, 35280, 35279, 27238, 35284, 35283, 27237, 35283, 35280, 35334, 27288, 27287, 35338, 27292, 27291, 35337, 27291, 27288, 27258, 35304, 35303, 27262, 35308, 35307, 35311, 35308, 27262, 35323, 35321, 27274, 27261, 35307, 35304, 35321, 35311, 27264, 35350, 27304, 27303, 35356, 27310, 27309, 35356, 35353, 27306, 35355, 27309, 27304, 27257, 35303, 35284, 35349, 27303, 27292, 35353, 35323, 27276, 27233, 35279, 35228, 35333, 27287, 27222, 27318, 35368, 35367, 27322, 35372, 35371, 27321, 35371, 35368, 35416, 27369, 27368, 35420, 27373, 27372, 35423, 35419, 27372, 35423, 27375, 27369, 27338, 35388, 35387, 27342, 35392, 35391, 27341, 35391, 35388, 35441, 27391, 27390, 35445, 27395, 27394, 35444, 27394, 27391, 27337, 35387, 35372, 35440, 27390, 27373, 27403, 35453, 35452, 27407, 35457, 35456, 27406, 35456, 35453, 35501, 27451, 27450, 35505, 27455, 27454, 35504, 27454, 27451, 27423, 35473, 35472, 27427, 35477, 35476, 27426, 35476, 35473, 35521, 27471, 27470, 35525, 27475, 27474, 35524, 27474, 27471, 27422, 35472, 35457, 35520, 27470, 27455, 27402, 35452, 35392, 35500, 27450, 27395, 35532, 35415, 27368, 27576, 35624, 35623,
35627, 35624, 27576, 27582, 35630, 35629, 27581, 35629, 35627, 27590, 35640, 35639, 27594, 35644, 35643, 27593, 35643, 35640, 27589, 35639, 35630, 27602, 35652, 35651, 27606, 35656, 35655, 27605, 35655, 35652, 27614, 35664, 35663, 27618, 35668, 35667, 27617, 35667, 35664, 27613, 35663, 35656, 27601, 35651, 35644, 27486, 35536, 35535, 27490, 35540, 35539, 27489, 35539, 35536, 35588, 27538, 27537, 35592, 27542, 27541, 35591, 27541, 27538, 27506, 35556, 35555, 27510, 35560, 35559, 35569, 35567, 27520, 27509, 35559, 35556, 35567, 35560, 27510, 35612, 27562, 27561, 35601, 35599, 27552, 35616, 27566, 27565, 35615, 27565, 27562, 35616, 35601, 27554, 27505, 35555, 35540, 35611, 27561, 27542, 35599, 35569, 27522, 27626, 35676, 35675, 27630, 35680, 35679, 27629, 35679, 35676, 27638, 35688, 35687, 27642, 35692, 35691, 27641, 35691, 35688, 27637, 35687, 35680, 27485, 35535, 35477, 35587, 27537, 27475, 27625, 35675, 35668, 35732, 27683, 27682, 35736, 27687, 27686, 35736, 35739, 27689, 35735, 27686, 27683, 35747, 27701, 27700, 35751, 27705, 27704, 35750, 27704, 27701, 35746, 27700, 27689, 27777, 35827, 35826, 27781, 35831, 35830, 27780, 35830, 35827, 27797, 35847, 35846, 27801, 35851, 35850, 27800, 35850, 35847, 27796, 35846, 35831, 35795, 27749, 27748,
35799, 27753, 27752, 35798, 27752, 27749, 35811, 27765, 27764, 35815, 27769, 27768, 35814, 27768, 27765, 35810, 27764, 27753, 27817, 35867, 35866, 27821, 35871, 35870, 27820, 35870, 35867, 27837, 35887, 35886, 27841, 35891, 35890, 27840, 35890, 35887, 27836, 35886, 35871, 35794, 27748, 27705, 27816, 35866, 35851, 36033, 27983, 27982, 36037, 27987, 27986, 36036, 27986, 27983, 36053, 28003, 28002, 36057, 28007, 28006, 36056, 28006, 28003, 36052, 28002, 27987, 28055, 36105, 36104, 28059, 36109, 36108, 28058, 36108, 36105, 28067, 36117, 36116, 28071, 36121, 36120, 28070, 36120, 36117, 28066, 36116, 36109, 36073, 28023, 28022, 36077, 28027, 28026, 36076, 28026, 28023, 36093, 28043, 28042, 36097, 28047, 28046, 36096, 28046, 28043, 36092, 28042, 28027, 28111, 36161, 36160, 28115, 36165, 36164, 28114, 36164, 36161, 28123, 36173, 36172, 28127, 36177, 36176, 28126, 36176, 36173, 28122, 36172, 36165, 36072, 28022, 28007, 28110, 36160, 36121, 35947, 27901, 27900, 35951, 27905, 27904, 35950, 27904, 27901, 35924, 35922, 27875, 35936, 35934, 27887, 35934, 35924, 27877, 35963, 27917, 27916, 35969, 27923, 27922, 35969, 35966, 27919, 35968, 27922, 27917, 35962, 27916, 27905, 35966, 35936, 27889, 27931, 35981, 35980, 27935, 35985, 35984, 27934, 35984, 35981,
27951, 36001, 36000, 27955, 36005, 36004, 36014, 36012, 27965, 27954, 36004, 36001, 36012, 36005, 27955, 27950, 36000, 35985, 36225, 28175, 28174, 36229, 28179, 28178, 36228, 28178, 28175, 36249, 28199, 28198, 36238, 36236, 28189, 36253, 28203, 28202, 36252, 28202, 28199, 36253, 36238, 28191, 36248, 28198, 28179, 28211, 36261, 36260, 28215, 36265, 36264, 28214, 36264, 36261, 28223, 36273, 36272, 28227, 36277, 36276, 28226, 36276, 36273, 36302, 36300, 28253, 36314, 36312, 28265, 36312, 36302, 28255, 28222, 36272, 36265, 36300, 36277, 28227, 35946, 27900, 27769, 27930, 35980, 35891, 36224, 28174, 28047, 28210, 36260, 36177, 36236, 36014, 27967, 36325, 28279, 28278, 36329, 28283, 28282, 36328, 28282, 28279, 36341, 28295, 28294, 36345, 28299, 28298, 36344, 28298, 28295, 36340, 28294, 28283, 36357, 28311, 28310, 36361, 28315, 28314, 36360, 28314, 28311, 36373, 28327, 28326, 36377, 28331, 28330, 36376, 28330, 28327, 36372, 28326, 28315, 36356, 28310, 28299, 28506, 36558, 36557, 28542, 36588, 36587, 36591, 36588, 28542, 28541, 36587, 36585, 36630, 28584, 28583, 36644, 28593, 28592, 36643, 28592, 28589, 36630, 36641, 28589, 28560, 36610, 36609, 28564, 36614, 36613, 28563, 36613, 36610, 36659, 28609, 28608, 36663, 28613, 28612, 36662, 28612, 28609,
28559, 36609, 36591, 36658, 28608, 28593, 28522, 36570, 36569, 36563, 36565, 28512, 28526, 36574, 36573, 36573, 36577, 28528, 28528, 36577, 36570, 36565, 36574, 28526, 28623, 36671, 36670, 28627, 36675, 36674, 36674, 36678, 28629, 28629, 36678, 36671, 36721, 28671, 28670, 36725, 28675, 28674, 36724, 28674, 28671, 36733, 28687, 28686, 36739, 28693, 28692, 36739, 36736, 28689, 36738, 28692, 28687, 36750, 36732, 28686, 28647, 36694, 36675, 36750, 28696, 28675, 28622, 36670, 36614, 36720, 28670, 28613, 36585, 36558, 28506, 36569, 36694, 28647, 36389, 28343, 28342, 36402, 36404, 28353, 36392, 28345, 28343, 36392, 36402, 28351, 36434, 36436, 28385, 36442, 36444, 28393, 36436, 36442, 28391, 36412, 36410, 28363, 36424, 36422, 28375, 36422, 36412, 28365, 36452, 36450, 28403, 36464, 36462, 28415, 36462, 36452, 28405, 36404, 36434, 28383, 36450, 36424, 28377, 36474, 36476, 28425, 36482, 36484, 28433, 36476, 36482, 28431, 36514, 36516, 28465, 36524, 36526, 28476, 28476, 36526, 36522, 36516, 36524, 28474, 36492, 36490, 28443, 36504, 36502, 28455, 36502, 36492, 28445, 36535, 36533, 28486, 36547, 36545, 28498, 36545, 36535, 28488, 36484, 36514, 28463, 36533, 36504, 28457, 36444, 36474, 28423, 36490, 36464, 28417, 36753, 36755, 28706, 28719, 36766, 36765,
36763, 36766, 28719, 36755, 36763, 28715, 36778, 36776, 28729, 36789, 28742, 28741, 36792, 36788, 28741, 36789, 36778, 28731, 36834, 28784, 28783, 36834, 36831, 28780, 36823, 36821, 28772, 36823, 28774, 28770, 36831, 36819, 28770, 36765, 36801, 28754, 36821, 36792, 28744, 36852, 28805, 28804, 36856, 28809, 28808, 36859, 36855, 28808, 36859, 28811, 28805, 36856, 36841, 28794, 36870, 36868, 28815, 36870, 28817, 28794, 36874, 36851, 28804, 36868, 36833, 28783, 36388, 28342, 28331, 28754, 36801, 36563, 36874, 28820, 28689, 36522, 36753, 28704, 36776, 36547, 28500, 36410, 36314, 28267, 28829, 36880, 36879, 36877, 36880, 28829, 36886, 36888, 28837, 36879, 36886, 28835, 36968, 36970, 28921, 36980, 36982, 28933, 28923, 36972, 36968, 36970, 36980, 28931, 36990, 36992, 28943, 37000, 37002, 28953, 36992, 37000, 28951, 36982, 36990, 28941, 28847, 36895, 36894, 28851, 36899, 36898, 36898, 36902, 28853, 28853, 36902, 36895, 28861, 36911, 36910, 28865, 36915, 36914, 28864, 36914, 36911, 36910, 36922, 28875, 36924, 36915, 28865, 28875, 36922, 36899, 36894, 36934, 28887, 37022, 37020, 28973, 37020, 37010, 28963, 37048, 29001, 29000, 37051, 37047, 29000, 37062, 37060, 29011, 37060, 37051, 29003, 37033, 28984, 28983, 37036, 37032, 28983, 37039, 28990, 28989,
37038, 28989, 28984, 37036, 28986, 28963, 37048, 37022, 28975, 28887, 36934, 36877, 36888, 36972, 28923, 37220, 37222, 29173, 37230, 37232, 29183, 37222, 37230, 29181, 37240, 37242, 29193, 37250, 37252, 29203, 37242, 37250, 29201, 37232, 37240, 29191, 37300, 37302, 29253, 37310, 37312, 29263, 37302, 37310, 29261, 29279, 37326, 37325, 37323, 37326, 29279, 37336, 37338, 29291, 37325, 37336, 29289, 29291, 37338, 37320, 37312, 37323, 29275, 37262, 37260, 29211, 37272, 37270, 29221, 37270, 37262, 29213, 37282, 37280, 29231, 37292, 37290, 29241, 37290, 37282, 29233, 37280, 37272, 29223, 37350, 37348, 29299, 37360, 37358, 29309, 37358, 37350, 29301, 37370, 37368, 29319, 37380, 37378, 29329, 37378, 37370, 29321, 37368, 37360, 29311, 37252, 37300, 29251, 37348, 37292, 29243, 29024, 37071, 37070, 29030, 37077, 37076, 37076, 37074, 29026, 29026, 37074, 37071, 29057, 37107, 37106, 37110, 37107, 29057, 37120, 37118, 29069, 29069, 37118, 37116, 37116, 37110, 29059, 29033, 37087, 37077, 37106, 37087, 29033, 37131, 29085, 29084, 37128, 37130, 29084, 37142, 37144, 29095, 37146, 29097, 29095, 37131, 37142, 29093, 37180, 37182, 29131, 37188, 37190, 29139, 37182, 37188, 29137, 29109, 37155, 37154, 37158, 37155, 29109, 37170, 37168, 29121, 37168, 37158, 29111,
37198, 37196, 29149, 37210, 37208, 29161, 37208, 37198, 29151, 37146, 37180, 29129, 37196, 37170, 29123, 37154, 37120, 29071, 37388, 37390, 29339, 37396, 37398, 29347, 37390, 37396, 29345, 37428, 37430, 29379, 37436, 37438, 29387, 37430, 37436, 29385, 37406, 37404, 29357, 37418, 37416, 29369, 37416, 37406, 29359, 37446, 37444, 29397, 37458, 37456, 29409, 37456, 37446, 29399, 37398, 37428, 29377, 37444, 37418, 29371, 37468, 37470, 29419, 37476, 37478, 29427, 37470, 37476, 29425, 37508, 37510, 29459, 37516, 37518, 29467, 37510, 37516, 29465, 37486, 37484, 29437, 37498, 37496, 29449, 37496, 37486, 29439, 37526, 37524, 29477, 37538, 37536, 29489, 37536, 37526, 29479, 37478, 37508, 29457, 37524, 37498, 29451, 37438, 37468, 29417, 37484, 37458, 29411, 29023, 37070, 36924, 37128, 29081, 28990, 37002, 37220, 29171, 37260, 37062, 29013, 37190, 37388, 29337, 37404, 37210, 29163, 29501, 37549, 37548, 29505, 37553, 37552, 37552, 37556, 29507, 29507, 37556, 37549, 37599, 29549, 29548, 37603, 29553, 29552, 37602, 29552, 29549, 37611, 29564, 29563, 37615, 29568, 29567, 37618, 37614, 29567, 37618, 29570, 29564, 37627, 37610, 29563, 37627, 29573, 29553, 37615, 37580, 29533, 37632, 37630, 29580, 37642, 29591, 29590, 37645, 37641, 29590, 37634, 29584, 29580,
37642, 37632, 29582, 37645, 29593, 29533, 37650, 37652, 29601, 37658, 37660, 29609, 37652, 37658, 29607, 37691, 29639, 29638, 37668, 37666, 29619, 37680, 37678, 29631, 37678, 37668, 29621, 37697, 29647, 29646, 37701, 29651, 29650, 37700, 29650, 29647, 37696, 29646, 29639, 37660, 37690, 29638, 37701, 37680, 29633, 37320, 37553, 29505, 37634, 37380, 29331, 37518, 37650, 29599, 37666, 37538, 29491, 35922, 35692, 27642, 5430, 5432, 799, 807, 5436, 5434, 789, 5428, 5430, 805, 5434, 5428, 840, 5466, 5464, 848, 5474, 5472, 846, 5472, 5466, 838, 5464, 5436, 922, 5548, 5546, 5556, 5554, 930, 930, 5554, 5548, 920, 5546, 5474, 1406, 1407, 5866, 5872, 5873, 1409, 5864, 5872, 1408, 1438, 1439, 5892, 1445, 5900, 5898, 1444, 5898, 5892, 1407, 1438, 5890, 6462, 6461, 2175, 2175, 6461, 6453, 6502, 6503, 2230, 2226, 6501, 6502, 2342, 6544, 6543, 2350, 6554, 6552, 2342, 2348, 6552, 2380, 6566, 6564, 2388, 6580, 6578, 2386, 6578, 6566, 2378, 6564, 6554, 2340, 6543, 6501, 2518, 6654, 6652, 2528, 2530, 6662, 2528, 6660, 6654, 6701, 6700, 2563, 2563, 6700, 6692, 2560, 6692, 6662, 6763, 6764, 2663, 2659, 6761, 6763, 2691, 2693, 6789,
2701, 6799, 6797, 2699, 6797, 6789, 2691, 6787, 6761, 2516, 6652, 6580, 2998, 7608, 7606, 3006, 7622, 7620, 3004, 7620, 7608, 3015, 7648, 7646, 3025, 3027, 7656, 3025, 7654, 7648, 3013, 7646, 7622, 7786, 7785, 3035, 3035, 7785, 7777, 3034, 7777, 7656, 3346, 8264, 8262, 8270, 8272, 3354, 8262, 8270, 3352, 3382, 8303, 8301, 8317, 8315, 3392, 8315, 8311, 3386, 3386, 8311, 8303, 3380, 8301, 8264, 2996, 7606, 6799, 2170, 6453, 5900, 8991, 8993, 3554, 3562, 9001, 8999, 8987, 8991, 3552, 3560, 8999, 8987, 3590, 9029, 9027, 9041, 9039, 3600, 9039, 9035, 3594, 3594, 9035, 9029, 3588, 9027, 9001, 9646, 9648, 4081, 4089, 9656, 9654, 9642, 9646, 4079, 4087, 9654, 9642, 4113, 9684, 9682, 9696, 9694, 4126, 9694, 9690, 4119, 4119, 9690, 9684, 4111, 9682, 9656, 9962, 9964, 4808, 9958, 9962, 4806, 4832, 9984, 9982, 4840, 9992, 9990, 4838, 9990, 9984, 4830, 9982, 9958, 10140, 10138, 5053, 10138, 10134, 5047, 5115, 10184, 10182, 10190, 10192, 5123, 5131, 10200, 10198, 10182, 10190, 5121, 5129, 10198, 10184, 5153, 5155, 10232, 10239, 10238, 5161, 10238, 10232, 5155, 5153, 10230, 10200, 5047, 10134, 9992,
16226, 16228, 11592, 16240, 11605, 11603, 16228, 16230, 11597, 16230, 16238, 11603, 16264, 11638, 11636, 16272, 11646, 11644, 16270, 11644, 11638, 16262, 11636, 11605, 16346, 16344, 11720, 16348, 16350, 11730, 16350, 16346, 11722, 16348, 11728, 11646, 16662, 16663, 12205, 16666, 12209, 12208, 12205, 16663, 16664, 16690, 12241, 12240, 16696, 16698, 12247, 16690, 16696, 12246, 16688, 12240, 12209, 17252, 17251, 12970, 17255, 17252, 12971, 17299, 17300, 13020, 17300, 17301, 13028, 17347, 17348, 13140, 17358, 13148, 13146, 17356, 13146, 13140, 17370, 13178, 13176, 17380, 13188, 13186, 17378, 13186, 13178, 17368, 13176, 13148, 17301, 17347, 13138, 17452, 13324, 13322, 17460, 13334, 13332, 17458, 13332, 13324, 17491, 17490, 13359, 17494, 17491, 13360, 17460, 17494, 13362, 17559, 17560, 13453, 17560, 17563, 13461, 17593, 13491, 13489, 17603, 13499, 13497, 17601, 13497, 13491, 17563, 17591, 13489, 17450, 13322, 13188, 18412, 13796, 13794, 18420, 13806, 13804, 18418, 13804, 13796, 18446, 13821, 13819, 18454, 13831, 13829, 18452, 13829, 13821, 18444, 13819, 13806, 18576, 18575, 13832, 18579, 18576, 13833, 18454, 18579, 13834, 19060, 19062, 14144, 19068, 19070, 14152, 14144, 19062, 19068, 19103, 14180, 14178, 19111, 19109, 14186, 14192, 19117, 19111,
19103, 19117, 14192, 19101, 14178, 14152, 18410, 13794, 13499, 16698, 17255, 12974, 19785, 19787, 14348, 19799, 14360, 14358, 19787, 19793, 14352, 19797, 14358, 14352, 19827, 14388, 14386, 19835, 19833, 14394, 19841, 19835, 14396, 19841, 14400, 14388, 19825, 14386, 14360, 20440, 20442, 14875, 20454, 14887, 14885, 20442, 20448, 14879, 20452, 14885, 14879, 20482, 14911, 14909, 20490, 20488, 14917, 14927, 20496, 20490, 20482, 20496, 14927, 20480, 14909, 14887, 20756, 20758, 15602, 20758, 20764, 15606, 20782, 15630, 15628, 20790, 15638, 15636, 20788, 15636, 15630, 20780, 15628, 15606, 20934, 20932, 15847, 20940, 20934, 15849, 20980, 20982, 15913, 20990, 15921, 15919, 20998, 15929, 15927, 20982, 20988, 15919, 20996, 15927, 15921, 21029, 21028, 15951, 21032, 15961, 15959, 21032, 21029, 15953, 21030, 15959, 15929, 20940, 15853, 15638, 22030, 26356, 26354, 26362, 26364, 22038, 22046, 26372, 26370, 26354, 26362, 22036, 22044, 26370, 26356, 22070, 26404, 26402, 26412, 26410, 22076, 26410, 26404, 22070, 22068, 26402, 26372, 22561, 27065, 27063, 27071, 27073, 22569, 22577, 27081, 27079, 27063, 27071, 22567, 22575, 27079, 27065, 22601, 27113, 27111, 27121, 27119, 22607, 27119, 27113, 22601, 22599, 27111, 27081, 27315, 27317, 23197, 23204, 27325, 27323,
23195, 27311, 27315, 23195, 23202, 27323, 23248, 27353, 27351, 23256, 27361, 27359, 23254, 27359, 27353, 23246, 27351, 27325, 27478, 27477, 23309, 23309, 27477, 27476, 23256, 23306, 27476, 23678, 27569, 27567, 27573, 27575, 23680, 23676, 27567, 27573, 23902, 27645, 27643, 23910, 27653, 27651, 23908, 27651, 27645, 27682, 27680, 23922, 23922, 27680, 27675, 23920, 27675, 27653, 27774, 27776, 23970, 23978, 27784, 27782, 27770, 27774, 23968, 23976, 27782, 27770, 24198, 27970, 27968, 27982, 27980, 24208, 27980, 27976, 24202, 24202, 27976, 27970, 28052, 28054, 24256, 28048, 28052, 24254, 24280, 28074, 28072, 24288, 28082, 28080, 24286, 28080, 28074, 24278, 28072, 28048, 24196, 27968, 27784, 24477, 28270, 28268, 28278, 28276, 24483, 28276, 28270, 24477, 28503, 28505, 24974, 24968, 28501, 28503, 24998, 28531, 28529, 25006, 28547, 28545, 25004, 28545, 28531, 25016, 28575, 28573, 28583, 28581, 25023, 25023, 28581, 28575, 25006, 25015, 28573, 24996, 28529, 28501, 24475, 28268, 28082, 29498, 29500, 25970, 25978, 29510, 29508, 29492, 29498, 25969, 25976, 29508, 29492, 25995, 29536, 29534, 29548, 29546, 26006, 26006, 29546, 29542, 26002, 29542, 29536, 25993, 29534, 29510, 23900, 27643, 27569, 34410, 34412, 30086, 34420, 30094, 30092, 34428, 30102, 30100,
34412, 34418, 30092, 34426, 30100, 30094, 34460, 34458, 30124, 34468, 30134, 30132, 34468, 34460, 30126, 34466, 30132, 30102, 35119, 35121, 30617, 35129, 30625, 30623, 35137, 30633, 30631, 35121, 35127, 30623, 35135, 30631, 30625, 35169, 35167, 30655, 35177, 30665, 30663, 35177, 35169, 30657, 35175, 30663, 30633, 35367, 35369, 31252, 35379, 35381, 31270, 35369, 35375, 31253, 35379, 31268, 31253, 35409, 31304, 31302, 35425, 35427, 31312, 35425, 31310, 31304, 35407, 31302, 31270, 35533, 35532, 31357, 35534, 35533, 31358, 35534, 31368, 31312, 35623, 35625, 31728, 35633, 35635, 31736, 35625, 35633, 31734, 35701, 31958, 31956, 35709, 31974, 31972, 35707, 31972, 31958, 35733, 35731, 31976, 35741, 35733, 31978, 35709, 35741, 31980, 35826, 35828, 32022, 35840, 32034, 32032, 35828, 35834, 32026, 35838, 32032, 32026, 36026, 32254, 32252, 36034, 36032, 32260, 36040, 36034, 32262, 36040, 32266, 32254, 36104, 36106, 32308, 36106, 36112, 32312, 36130, 32336, 32334, 36138, 32344, 32342, 36136, 32342, 32336, 36128, 32334, 32312, 36024, 32252, 32034, 36326, 36324, 32531, 36334, 32541, 32539, 36334, 36326, 32533, 36557, 36559, 33022, 36559, 36561, 33030, 36595, 33054, 33052, 36603, 33068, 33066, 36601, 33066, 33054, 36631, 36629, 33075, 36633, 36635, 33080,
36635, 36631, 33076, 36633, 33079, 33068, 36561, 36593, 33052, 36332, 32539, 32344, 37548, 37550, 34018, 37566, 34034, 34032, 34018, 37550, 37558, 37558, 37564, 34032, 37592, 34055, 34053, 37600, 37598, 34059, 37606, 37600, 34061, 37592, 37606, 34064, 37590, 34053, 34034, 35699, 31956, 31736, 5432, 16226, 11590, 16344, 5556, 932, 5873, 16662, 12204, 17251, 6462, 2176, 6503, 17299, 13019, 17490, 6701, 2564, 6764, 17559, 13452, 18575, 7786, 3036, 8272, 19060, 14142, 19109, 8317, 3394, 8993, 19785, 14346, 19833, 9041, 3602, 9648, 20440, 14873, 20488, 9696, 4128, 9964, 20756, 15600, 20932, 10140, 5055, 10192, 20980, 15911, 21028, 10239, 5163, 26364, 34410, 30084, 34458, 26412, 22078, 27073, 35119, 30615, 35167, 27121, 22609, 27317, 35367, 31251, 35532, 27478, 23310, 27575, 35623, 31727, 35731, 27682, 23924, 27776, 35826, 32020, 36032, 27982, 24210, 28054, 36104, 32306, 36324, 28278, 24485, 28505, 36557, 33021, 36629, 28583, 25024, 29500, 37548, 34017, 37598, 29548, 26008, 21410, 34065, 26018, 34089, 21430, 10637, 21566, 34292, 26242, 34311, 21586, 10797});

    _verty = std::vector<float>({0.0, -0.4928, -0.0364, -0.0017, -0.4928, -0.0364, -0.0039, -0.4928, -0.0364, 0.0, -0.4868, -0.0456, 0.0, -0.485, -0.0481, 0.0, -0.486, -0.0467, -0.0017, -0.4868, -0.0457, -0.0039, -0.4868, -0.0457, -0.0017, -0.4849, -0.0482, -0.0017, -0.486, -0.0468, -0.0039, -0.4849, -0.0482, -0.0039, -0.486, -0.0468, 0.0, -0.4886, -0.043, 0.0, -0.4893, -0.042, 0.0, -0.4912, -0.0391, 0.0, -0.4922, -0.0375, 0.0, -0.4905, -0.0402, 0.0, -0.4896, -0.0374, -0.0017, -0.4886, -0.043, -0.0017, -0.4895, -0.0418, -0.0039, -0.4886, -0.043, -0.0039, -0.4895, -0.0418, -0.0017, -0.4911, -0.0393, -0.0017, -0.4922, -0.0375, -0.0039, -0.4911, -0.0393, -0.0039, -0.4922, -0.0375, -0.0017, -0.4904, -0.0403, -0.0039, -0.4904, -0.0403, 0.0, -0.4879, -0.0441, 0.0, -0.4874, -0.0418,
0.0, -0.4852, -0.0439, 0.0, -0.4852, -0.0418, 0.0, -0.4874, -0.0396, 0.0, -0.4869, -0.0368, 0.0, -0.4845, -0.0386, 0.0, -0.4856, -0.0378, -0.0017, -0.4879, -0.0441, -0.0039, -0.4879, -0.0441, -0.0017, -0.4868, -0.0368, -0.0039, -0.4868, -0.0368, -0.0017, -0.4844, -0.0387, -0.0017, -0.4855, -0.0379, -0.0039, -0.4844, -0.0387, -0.0039, -0.4855, -0.0379, 0.0, -0.4951, -0.0321, 0.0, -0.4934, -0.0354, 0.0, -0.4945, -0.0332, 0.0, -0.4961, -0.0297, 0.0, -0.4935, -0.0308, -0.0017, -0.4951, -0.0321, -0.0039, -0.4951, -0.0321, -0.0017, -0.4935, -0.0353, -0.0017, -0.4945, -0.0332, -0.0039, -0.4935, -0.0353, -0.0039, -0.4945, -0.0332, -0.0017, -0.4961, -0.0297, -0.0039, -0.4961, -0.0297, -0.0017, -0.4938, -0.0306, -0.0039, -0.4938, -0.0306, 0.0, -0.4934, -0.024,
0.0, -0.4953, -0.0215, 0.0, -0.496, -0.0203, 0.0, -0.4945, -0.0226, 0.0, -0.4935, -0.0195, -0.0017, -0.4934, -0.0241, -0.0039, -0.4934, -0.0241, -0.0017, -0.4952, -0.0215, -0.0017, -0.4961, -0.0201, -0.0039, -0.4952, -0.0215, -0.0039, -0.4961, -0.0201, -0.0017, -0.4945, -0.0226, -0.0017, -0.4935, -0.0195, -0.0039, -0.4945, -0.0226, -0.0039, -0.4935, -0.0195, 0.0, -0.4917, -0.0353, 0.0, -0.4914, -0.0326, 0.0, -0.4891, -0.0347, 0.0, -0.4902, -0.0337, 0.0, -0.4925, -0.0317, 0.0, -0.4894, -0.0286, -0.0017, -0.4914, -0.0326, -0.0039, -0.4914, -0.0326, -0.0017, -0.4891, -0.0347, -0.0017, -0.4902, -0.0337, -0.0039, -0.4891, -0.0347, -0.0039, -0.4902, -0.0337, -0.0017, -0.4925, -0.0317, -0.0039, -0.4925, -0.0317, -0.0017, -0.4894, -0.0286, -0.0039, -0.4894, -0.0286,
0.0, -0.4879, -0.0358, 0.0, -0.4849, -0.0327, 0.0, -0.4872, -0.0307, 0.0, -0.4883, -0.0297, 0.0, -0.486, -0.0318, 0.0, -0.4852, -0.0288, -0.0017, -0.4879, -0.0358, -0.0039, -0.4879, -0.0358, -0.0017, -0.4849, -0.0326, -0.0039, -0.4849, -0.0326, -0.0017, -0.4872, -0.0307, -0.0017, -0.4883, -0.0296, -0.0039, -0.4872, -0.0307, -0.0039, -0.4883, -0.0296, -0.0017, -0.486, -0.0317, -0.0039, -0.486, -0.0317, 0.0, -0.4915, -0.0263, 0.0, -0.4926, -0.0251, 0.0, -0.4905, -0.0274, 0.0, -0.4896, -0.0244, 0.0, -0.4911, -0.0214, 0.0, -0.4922, -0.0206, 0.0, -0.4894, -0.0227, -0.0017, -0.4915, -0.0264, -0.0017, -0.4925, -0.0252, -0.0039, -0.4915, -0.0264, -0.0039, -0.4925, -0.0252, -0.0017, -0.4905, -0.0274, -0.0039, -0.4905, -0.0274, -0.0017, -0.4911, -0.0215,
-0.0017, -0.4922, -0.0206, -0.0039, -0.4911, -0.0215, -0.0039, -0.4922, -0.0206, -0.0017, -0.4896, -0.0225, -0.0039, -0.4896, -0.0225, 0.0, -0.4874, -0.0266, 0.0, -0.4873, -0.0239, 0.0, -0.4852, -0.0266, 0.0, -0.4851, -0.025, 0.0, -0.4884, -0.0233, -0.0017, -0.4873, -0.0239, -0.0039, -0.4873, -0.0239, -0.0017, -0.4853, -0.025, -0.0039, -0.4853, -0.025, -0.0017, -0.4884, -0.0233, -0.0039, -0.4884, -0.0233, 0.0, -0.4975, -0.0173, 0.0, -0.498, -0.016, -0.0017, -0.4975, -0.0175, -0.0017, -0.498, -0.016, -0.0039, -0.4975, -0.0175, -0.0039, -0.498, -0.016, 0.0, -0.4983, -0.0157, 0.0, -0.497, -0.0167, 0.0, -0.4946, -0.0186, -0.0017, -0.4983, -0.0157, -0.0017, -0.497, -0.0167, -0.0039, -0.4983, -0.0157, -0.0039, -0.497, -0.0167, -0.0017, -0.4945, -0.0186,
-0.0039, -0.4945, -0.0186, 0.0, -0.4561, -0.0713, 0.0, -0.4541, -0.0736, 0.0, -0.4552, -0.0724, -0.0017, -0.4561, -0.0713, -0.0039, -0.4561, -0.0713, -0.0017, -0.4541, -0.0736, -0.0017, -0.4552, -0.0724, -0.0039, -0.4541, -0.0736, -0.0039, -0.4552, -0.0724, 0.0, -0.452, -0.0759, 0.0, -0.4499, -0.0781, 0.0, -0.451, -0.077, 0.0, -0.4531, -0.0747, 0.0, -0.4527, -0.0722, 0.0, -0.4505, -0.0743, 0.0, -0.4505, -0.0722, -0.0017, -0.452, -0.0759, -0.0039, -0.452, -0.0759, -0.0017, -0.4499, -0.0781, -0.0017, -0.451, -0.077, -0.0039, -0.4499, -0.0781, -0.0039, -0.451, -0.077, -0.0017, -0.4531, -0.0747, -0.0039, -0.4531, -0.0747, 0.0, -0.4371, -0.0891, -0.0017, -0.4372, -0.089, -0.0039, -0.4372, -0.089, 0.0, -0.4345, -0.0908, 0.0, -0.4356, -0.0901,
0.0, -0.4334, -0.0916, 0.0, -0.4332, -0.0895, -0.0017, -0.4345, -0.0908, -0.0017, -0.4356, -0.0901, -0.0039, -0.4345, -0.0908, -0.0039, -0.4356, -0.0901, -0.0017, -0.4332, -0.0917, -0.0039, -0.4332, -0.0917, 0.0, -0.4478, -0.0802, 0.0, -0.4455, -0.0823, 0.0, -0.4466, -0.0813, -0.0017, -0.4478, -0.0802, -0.0039, -0.4478, -0.0802, -0.0017, -0.4455, -0.0823, -0.0017, -0.4466, -0.0813, -0.0039, -0.4455, -0.0823, -0.0039, -0.4466, -0.0813, 0.0, -0.4432, -0.0844, 0.0, -0.4408, -0.0863, 0.0, -0.442, -0.0855, 0.0, -0.4443, -0.0834, 0.0, -0.444, -0.0808, 0.0, -0.4418, -0.083, 0.0, -0.4418, -0.0808, -0.0017, -0.4432, -0.0844, -0.0039, -0.4432, -0.0844, -0.0017, -0.4408, -0.0863, -0.0017, -0.4419, -0.0855, -0.0039, -0.4408, -0.0863, -0.0039, -0.4419, -0.0855,
-0.0017, -0.4443, -0.0834, -0.0039, -0.4443, -0.0834, 0.0, -0.4488, -0.0792, 0.0, -0.4483, -0.0765, 0.0, -0.4462, -0.0787, 0.0, -0.4462, -0.0765, 0.0, -0.4483, -0.0743, 0.0, -0.4483, -0.0722, 0.0, -0.4462, -0.0743, 0.0, -0.4462, -0.0722, -0.0017, -0.4488, -0.0792, -0.0039, -0.4488, -0.0792, 0.0, -0.444, -0.0787, 0.0, -0.444, -0.0765, 0.0, -0.4418, -0.0787, 0.0, -0.4418, -0.0765, 0.0, -0.444, -0.0743, 0.0, -0.444, -0.0722, 0.0, -0.4418, -0.0743, 0.0, -0.4418, -0.0722, 0.0, -0.4399, -0.087, 0.0, -0.4397, -0.0852, 0.0, -0.4382, -0.0883, 0.0, -0.4375, -0.0852, 0.0, -0.4397, -0.083, 0.0, -0.4397, -0.0808, 0.0, -0.4375, -0.083, 0.0, -0.4375, -0.0808, -0.0017, -0.4397, -0.0872, -0.0039, -0.4397, -0.0872,
-0.0017, -0.4383, -0.0882, -0.0039, -0.4383, -0.0882, 0.0, -0.4353, -0.0874, 0.0, -0.4353, -0.0852, 0.0, -0.4332, -0.0874, 0.0, -0.4332, -0.0852, 0.0, -0.4353, -0.083, 0.0, -0.4353, -0.0808, 0.0, -0.4332, -0.083, 0.0, -0.4332, -0.0808, 0.0, -0.4397, -0.0787, 0.0, -0.4397, -0.0765, 0.0, -0.4375, -0.0787, 0.0, -0.4375, -0.0765, 0.0, -0.4397, -0.0743, 0.0, -0.4397, -0.0722, 0.0, -0.4375, -0.0743, 0.0, -0.4375, -0.0722, 0.0, -0.4353, -0.0787, 0.0, -0.4353, -0.0765, 0.0, -0.4332, -0.0787, 0.0, -0.4332, -0.0765, 0.0, -0.4353, -0.0743, 0.0, -0.4353, -0.0722, 0.0, -0.4332, -0.0743, 0.0, -0.4332, -0.0722, 0.0, -0.4234, -0.0972, -0.0017, -0.4235, -0.0972, -0.0039, -0.4235, -0.0972, 0.0, -0.4306, -0.0933,
0.0, -0.4277, -0.0949, 0.0, -0.4289, -0.0943, 0.0, -0.4316, -0.0927, 0.0, -0.431, -0.0895, 0.0, -0.4288, -0.0917, 0.0, -0.4288, -0.0895, -0.0017, -0.4307, -0.0933, -0.0039, -0.4307, -0.0933, -0.0017, -0.4277, -0.0949, -0.0017, -0.4288, -0.0943, -0.0039, -0.4277, -0.0949, -0.0039, -0.4288, -0.0943, -0.0017, -0.4318, -0.0926, -0.0039, -0.4318, -0.0926, 0.0, -0.4268, -0.0955, 0.0, -0.4266, -0.0939, 0.0, -0.4246, -0.0966, 0.0, -0.4245, -0.0939, 0.0, -0.4266, -0.0917, 0.0, -0.4266, -0.0895, 0.0, -0.4245, -0.0917, 0.0, -0.4245, -0.0895, -0.0017, -0.4266, -0.0955, -0.0039, -0.4266, -0.0955, -0.0017, -0.4245, -0.0966, -0.0039, -0.4245, -0.0966, 0.0, -0.4224, -0.0977, 0.0, -0.4205, -0.0986, -0.0017, -0.4223, -0.0977, -0.0039, -0.4223, -0.0977,
-0.0017, -0.4201, -0.0988, -0.0039, -0.4201, -0.0988, 0.0, -0.4177, -0.0998, 0.0, -0.4186, -0.0995, 0.0, -0.416, -0.1005, 0.0, -0.4158, -0.0982, -0.0017, -0.4179, -0.0997, -0.0017, -0.419, -0.0993, -0.0039, -0.4179, -0.0997, -0.0039, -0.419, -0.0993, -0.0017, -0.4158, -0.1006, -0.0039, -0.4158, -0.1006, 0.0, -0.4223, -0.096, 0.0, -0.4223, -0.0939, 0.0, -0.4201, -0.096, 0.0, -0.4201, -0.0939, 0.0, -0.4223, -0.0917, 0.0, -0.4223, -0.0895, 0.0, -0.4201, -0.0917, 0.0, -0.4201, -0.0895, 0.0, -0.418, -0.096, 0.0, -0.418, -0.0939, 0.0, -0.4158, -0.096, 0.0, -0.4158, -0.0939, 0.0, -0.418, -0.0917, 0.0, -0.418, -0.0895, 0.0, -0.4158, -0.0917, 0.0, -0.4158, -0.0895, 0.0, -0.431, -0.0874, 0.0, -0.431, -0.0852,
0.0, -0.4288, -0.0874, 0.0, -0.4288, -0.0852, 0.0, -0.431, -0.083, 0.0, -0.431, -0.0808, 0.0, -0.4288, -0.083, 0.0, -0.4288, -0.0808, 0.0, -0.4266, -0.0874, 0.0, -0.4266, -0.0852, 0.0, -0.4245, -0.0874, 0.0, -0.4245, -0.0852, 0.0, -0.4266, -0.083, 0.0, -0.4266, -0.0808, 0.0, -0.4245, -0.083, 0.0, -0.4245, -0.0808, 0.0, -0.431, -0.0787, 0.0, -0.431, -0.0765, 0.0, -0.4288, -0.0787, 0.0, -0.4288, -0.0765, 0.0, -0.431, -0.0743, 0.0, -0.4307, -0.072, 0.0, -0.4287, -0.0741, 0.0, -0.4297, -0.0731, -0.0017, -0.4307, -0.072, -0.0039, -0.4307, -0.072, -0.0017, -0.4287, -0.0741, -0.0017, -0.4297, -0.0731, -0.0039, -0.4287, -0.0741, -0.0039, -0.4297, -0.0731, 0.0, -0.4266, -0.0787, 0.0, -0.4263, -0.076,
0.0, -0.4239, -0.0778, 0.0, -0.4249, -0.077, 0.0, -0.4274, -0.0751, -0.0017, -0.4263, -0.076, -0.0039, -0.4263, -0.076, -0.0017, -0.4238, -0.0779, -0.0017, -0.4249, -0.0771, -0.0039, -0.4238, -0.0779, -0.0039, -0.4249, -0.0771, -0.0017, -0.4274, -0.0752, -0.0039, -0.4274, -0.0752, 0.0, -0.4223, -0.0874, 0.0, -0.4223, -0.0852, 0.0, -0.4201, -0.0874, 0.0, -0.4201, -0.0852, 0.0, -0.4223, -0.083, 0.0, -0.4223, -0.0808, 0.0, -0.4201, -0.083, 0.0, -0.42, -0.0804, -0.0017, -0.42, -0.0804, -0.0039, -0.42, -0.0804, 0.0, -0.418, -0.0874, 0.0, -0.418, -0.0852, 0.0, -0.4158, -0.0874, 0.0, -0.4158, -0.0852, 0.0, -0.4174, -0.082, 0.0, -0.4184, -0.0814, 0.0, -0.4155, -0.083, -0.0017, -0.4172, -0.0821, -0.0017, -0.4183, -0.0815,
-0.0039, -0.4172, -0.0821, -0.0039, -0.4183, -0.0815, -0.0017, -0.4158, -0.0828, -0.0039, -0.4158, -0.0828, 0.0, -0.4221, -0.079, 0.0, -0.4211, -0.0797, -0.0017, -0.4223, -0.0789, -0.0039, -0.4223, -0.0789, -0.0017, -0.4211, -0.0797, -0.0039, -0.4211, -0.0797, 0.0, -0.4799, -0.0538, -0.0017, -0.4799, -0.0538, -0.0039, -0.4799, -0.0538, 0.0, -0.4777, -0.0559, 0.0, -0.4788, -0.0549, 0.0, -0.4768, -0.0567, 0.0, -0.4766, -0.0548, -0.0017, -0.4777, -0.0559, -0.0017, -0.4788, -0.0549, -0.0039, -0.4777, -0.0559, -0.0039, -0.4788, -0.0549, -0.0017, -0.4766, -0.0569, -0.0039, -0.4766, -0.0569, 0.0, -0.469, -0.0624, 0.0, -0.468, -0.063, -0.0017, -0.469, -0.0624, -0.0039, -0.469, -0.0624, -0.0017, -0.4679, -0.0631, -0.0039, -0.4679, -0.0631, 0.0, -0.4742, -0.0589,
0.0, -0.4717, -0.0607, 0.0, -0.4728, -0.0599, 0.0, -0.4753, -0.058, 0.0, -0.4744, -0.0548, 0.0, -0.4722, -0.057, 0.0, -0.4722, -0.0548, -0.0017, -0.4743, -0.0588, -0.0039, -0.4743, -0.0588, -0.0017, -0.4717, -0.0606, -0.0017, -0.4728, -0.0599, -0.0039, -0.4717, -0.0606, -0.0039, -0.4728, -0.0599, -0.0017, -0.4753, -0.058, -0.0039, -0.4753, -0.058, 0.0, -0.4701, -0.0617, 0.0, -0.47, -0.0591, 0.0, -0.4679, -0.0613, 0.0, -0.4679, -0.0591, 0.0, -0.47, -0.057, 0.0, -0.47, -0.0548, 0.0, -0.4679, -0.057, 0.0, -0.4679, -0.0548, -0.0017, -0.4701, -0.0617, -0.0039, -0.4701, -0.0617, 0.0, -0.482, -0.0516, 0.0, -0.4828, -0.0507, 0.0, -0.481, -0.0527, 0.0, -0.4809, -0.0505, 0.0, -0.4841, -0.0492, 0.0, -0.4831, -0.0461,
0.0, -0.4809, -0.0483, 0.0, -0.4809, -0.0461, -0.0017, -0.482, -0.0516, -0.0017, -0.483, -0.0505, -0.0039, -0.482, -0.0516, -0.0039, -0.483, -0.0505, -0.0017, -0.481, -0.0527, -0.0039, -0.481, -0.0527, -0.0017, -0.4841, -0.0493, -0.0039, -0.4841, -0.0493, 0.0, -0.4787, -0.0526, 0.0, -0.4787, -0.0505, 0.0, -0.4766, -0.0526, 0.0, -0.4766, -0.0505, 0.0, -0.4787, -0.0483, 0.0, -0.4787, -0.0461, 0.0, -0.4766, -0.0483, 0.0, -0.4766, -0.0461, 0.0, -0.4831, -0.0439, 0.0, -0.4831, -0.0418, 0.0, -0.4809, -0.0439, 0.0, -0.4807, -0.0413, 0.0, -0.4828, -0.0398, 0.0, -0.4818, -0.0405, -0.0017, -0.4807, -0.0413, -0.0039, -0.4807, -0.0413, -0.0017, -0.4831, -0.0397, -0.0039, -0.4831, -0.0397, -0.0017, -0.4818, -0.0406, -0.0039, -0.4818, -0.0406,
0.0, -0.4781, -0.043, 0.0, -0.4791, -0.0423, 0.0, -0.4763, -0.044, 0.0, -0.4783, -0.0369, 0.0, -0.4768, -0.0376, -0.0017, -0.4779, -0.0431, -0.0017, -0.479, -0.0424, -0.0039, -0.4779, -0.0431, -0.0039, -0.479, -0.0424, -0.0017, -0.4766, -0.0439, -0.0039, -0.4766, -0.0439, -0.0017, -0.4785, -0.0368, -0.0039, -0.4785, -0.0368, -0.0017, -0.4766, -0.0377, -0.0039, -0.4766, -0.0377, 0.0, -0.4744, -0.0526, 0.0, -0.4744, -0.0505, 0.0, -0.4722, -0.0526, 0.0, -0.4722, -0.0505, 0.0, -0.4744, -0.0483, 0.0, -0.474, -0.0452, 0.0, -0.4722, -0.0483, 0.0, -0.472, -0.046, -0.0017, -0.4738, -0.0453, -0.0039, -0.4738, -0.0453, -0.0017, -0.4722, -0.0459, -0.0039, -0.4722, -0.0459, 0.0, -0.47, -0.0526, 0.0, -0.47, -0.0505, 0.0, -0.4679, -0.0526,
0.0, -0.4679, -0.0505, 0.0, -0.47, -0.0483, 0.0, -0.47, -0.0468, 0.0, -0.4679, -0.0476, 0.0, -0.4689, -0.0472, -0.0017, -0.4701, -0.0468, -0.0039, -0.4701, -0.0468, -0.0017, -0.4678, -0.0476, -0.0017, -0.4689, -0.0472, -0.0039, -0.4678, -0.0476, -0.0039, -0.4689, -0.0472, 0.0, -0.475, -0.0447, 0.0, -0.4737, -0.0388, 0.0, -0.4747, -0.0384, 0.0, -0.4723, -0.0393, 0.0, -0.4722, -0.0374, -0.0017, -0.4749, -0.0448, -0.0039, -0.4749, -0.0448, -0.0017, -0.4739, -0.0387, -0.0017, -0.475, -0.0383, -0.0039, -0.4739, -0.0387, -0.0039, -0.475, -0.0383, -0.0017, -0.4722, -0.0393, -0.0039, -0.4722, -0.0393, 0.0, -0.467, -0.0408, 0.0, -0.4702, -0.0399, 0.0, -0.47, -0.0374, 0.0, -0.468, -0.0405, 0.0, -0.4679, -0.0374, -0.0017, -0.4671, -0.0408,
-0.0039, -0.4671, -0.0408, -0.0017, -0.47, -0.04, -0.0039, -0.47, -0.04, -0.0017, -0.4682, -0.0405, -0.0039, -0.4682, -0.0405, 0.0, -0.4648, -0.0647, 0.0, -0.4659, -0.0641, 0.0, -0.4637, -0.0653, 0.0, -0.4635, -0.0635, -0.0017, -0.4649, -0.0647, -0.0017, -0.466, -0.0641, -0.0039, -0.4649, -0.0647, -0.0039, -0.466, -0.0641, -0.0017, -0.4635, -0.0654, -0.0039, -0.4635, -0.0654, 0.0, -0.4606, -0.067, 0.0, -0.4582, -0.069, 0.0, -0.4593, -0.068, 0.0, -0.4617, -0.0664, 0.0, -0.4614, -0.0635, 0.0, -0.4592, -0.0656, 0.0, -0.4592, -0.0635, -0.0017, -0.4607, -0.067, -0.0039, -0.4607, -0.067, -0.0017, -0.4582, -0.069, -0.0017, -0.4593, -0.068, -0.0039, -0.4582, -0.069, -0.0039, -0.4593, -0.068, -0.0017, -0.4618, -0.0663, -0.0039, -0.4618, -0.0663,
0.0, -0.4657, -0.0613, 0.0, -0.4657, -0.0591, 0.0, -0.4635, -0.0613, 0.0, -0.4635, -0.0591, 0.0, -0.4657, -0.057, 0.0, -0.4657, -0.0548, 0.0, -0.4635, -0.057, 0.0, -0.4635, -0.0548, 0.0, -0.4614, -0.0613, 0.0, -0.4614, -0.0591, 0.0, -0.4592, -0.0613, 0.0, -0.4592, -0.0591, 0.0, -0.4614, -0.057, 0.0, -0.4614, -0.0548, 0.0, -0.4592, -0.057, 0.0, -0.4592, -0.0548, 0.0, -0.4572, -0.0701, 0.0, -0.457, -0.0678, 0.0, -0.4549, -0.07, 0.0, -0.4549, -0.0678, 0.0, -0.457, -0.0656, 0.0, -0.457, -0.0635, 0.0, -0.4549, -0.0656, 0.0, -0.4549, -0.0635, -0.0017, -0.4572, -0.0701, -0.0039, -0.4572, -0.0701, 0.0, -0.4527, -0.07, 0.0, -0.4527, -0.0678, 0.0, -0.4505, -0.07, 0.0, -0.4505, -0.0678,
0.0, -0.4527, -0.0656, 0.0, -0.4527, -0.0635, 0.0, -0.4505, -0.0656, 0.0, -0.4505, -0.0635, 0.0, -0.457, -0.0613, 0.0, -0.457, -0.0591, 0.0, -0.4549, -0.0613, 0.0, -0.4549, -0.0591, 0.0, -0.457, -0.057, 0.0, -0.457, -0.0548, 0.0, -0.4549, -0.057, 0.0, -0.4549, -0.0548, 0.0, -0.4527, -0.0613, 0.0, -0.4527, -0.0591, 0.0, -0.4505, -0.0613, 0.0, -0.4505, -0.0591, 0.0, -0.4527, -0.057, 0.0, -0.4527, -0.0548, 0.0, -0.4505, -0.057, 0.0, -0.4505, -0.0548, 0.0, -0.4657, -0.0526, 0.0, -0.4657, -0.0505, 0.0, -0.4635, -0.0526, 0.0, -0.4635, -0.0505, 0.0, -0.4656, -0.0483, 0.0, -0.4635, -0.0488, -0.0017, -0.4657, -0.0482, -0.0039, -0.4657, -0.0482, -0.0017, -0.4635, -0.0487, -0.0039, -0.4635, -0.0487,
0.0, -0.4614, -0.0526, 0.0, -0.461, -0.0493, 0.0, -0.4592, -0.0526, 0.0, -0.459, -0.0497, 0.0, -0.4618, -0.0491, -0.0017, -0.4604, -0.0494, -0.0039, -0.4604, -0.0494, -0.0017, -0.4592, -0.0497, -0.0039, -0.4592, -0.0497, -0.0017, -0.4615, -0.0492, -0.0039, -0.4615, -0.0492, 0.0, -0.4658, -0.0411, 0.0, -0.4636, -0.0416, 0.0, -0.4657, -0.0396, 0.0, -0.4657, -0.0374, 0.0, -0.4635, -0.0396, 0.0, -0.4635, -0.0374, -0.0017, -0.4657, -0.0411, -0.0039, -0.4657, -0.0411, -0.0017, -0.4635, -0.0416, -0.0039, -0.4635, -0.0416, 0.0, -0.4615, -0.0421, 0.0, -0.4593, -0.0424, 0.0, -0.4614, -0.0396, 0.0, -0.4614, -0.0374, 0.0, -0.4592, -0.0396, 0.0, -0.4592, -0.0374, -0.0017, -0.4614, -0.0421, -0.0039, -0.4614, -0.0421, -0.0017, -0.4592, -0.0424,
-0.0039, -0.4592, -0.0424, 0.0, -0.457, -0.0526, 0.0, -0.4569, -0.05, 0.0, -0.4549, -0.0526, 0.0, -0.4548, -0.0502, -0.0017, -0.457, -0.05, -0.0039, -0.457, -0.05, -0.0017, -0.4549, -0.0502, -0.0039, -0.4549, -0.0502, 0.0, -0.4527, -0.0526, 0.0, -0.4527, -0.0503, 0.0, -0.4505, -0.0526, 0.0, -0.4505, -0.0503, -0.0017, -0.4527, -0.0503, -0.0039, -0.4527, -0.0503, -0.0017, -0.4505, -0.0503, -0.0039, -0.4505, -0.0503, 0.0, -0.457, -0.0425, 0.0, -0.4549, -0.0425, 0.0, -0.457, -0.0396, 0.0, -0.457, -0.0374, 0.0, -0.4549, -0.0396, 0.0, -0.4549, -0.0374, -0.0017, -0.457, -0.0425, -0.0039, -0.457, -0.0425, -0.0017, -0.4549, -0.0425, -0.0039, -0.4549, -0.0425, 0.0, -0.4527, -0.0425, 0.0, -0.4505, -0.0425, 0.0, -0.4527, -0.0396,
0.0, -0.4527, -0.0374, 0.0, -0.4505, -0.0396, 0.0, -0.4505, -0.0374, -0.0017, -0.4527, -0.0425, -0.0039, -0.4527, -0.0425, -0.0017, -0.4505, -0.0425, -0.0039, -0.4505, -0.0425, 0.0, -0.4824, -0.0345, 0.0, -0.4835, -0.0338, 0.0, -0.4812, -0.0353, 0.0, -0.4809, -0.0331, 0.0, -0.4831, -0.0309, 0.0, -0.4831, -0.0288, 0.0, -0.4809, -0.0309, 0.0, -0.4809, -0.0288, -0.0017, -0.4824, -0.0345, -0.0017, -0.4835, -0.0337, -0.0039, -0.4824, -0.0345, -0.0039, -0.4835, -0.0337, -0.0017, -0.4809, -0.0355, -0.0039, -0.4809, -0.0355, 0.0, -0.4793, -0.0363, 0.0, -0.4787, -0.0331, 0.0, -0.4766, -0.0353, 0.0, -0.4766, -0.0331, 0.0, -0.4787, -0.0309, 0.0, -0.4787, -0.0288, 0.0, -0.4766, -0.0309, 0.0, -0.4765, -0.0279, -0.0017, -0.4795, -0.0362,
-0.0039, -0.4795, -0.0362, -0.0017, -0.4764, -0.0279, -0.0039, -0.4764, -0.0279, 0.0, -0.483, -0.0259, 0.0, -0.4841, -0.0255, 0.0, -0.4807, -0.0267, -0.0017, -0.483, -0.0259, -0.0017, -0.4841, -0.0255, -0.0039, -0.483, -0.0259, -0.0039, -0.4841, -0.0255, -0.0017, -0.4809, -0.0266, -0.0039, -0.4809, -0.0266, 0.0, -0.4787, -0.0273, 0.0, -0.4775, -0.0276, -0.0017, -0.4787, -0.0273, -0.0039, -0.4787, -0.0273, -0.0017, -0.4775, -0.0276, -0.0039, -0.4775, -0.0276, 0.0, -0.4744, -0.0353, 0.0, -0.4744, -0.0331, 0.0, -0.4722, -0.0353, 0.0, -0.4722, -0.0331, 0.0, -0.4744, -0.0309, 0.0, -0.4743, -0.0284, 0.0, -0.4722, -0.0309, 0.0, -0.4722, -0.0286, -0.0017, -0.4744, -0.0283, -0.0039, -0.4744, -0.0283, -0.0017, -0.4722, -0.0286, -0.0039, -0.4722, -0.0286,
0.0, -0.47, -0.0353, 0.0, -0.47, -0.0331, 0.0, -0.4679, -0.0353, 0.0, -0.4679, -0.0331, 0.0, -0.47, -0.0309, 0.0, -0.47, -0.0287, 0.0, -0.4679, -0.0309, 0.0, -0.4679, -0.0287, -0.0017, -0.47, -0.0287, -0.0039, -0.47, -0.0287, -0.0017, -0.4679, -0.0287, -0.0039, -0.4679, -0.0287, 0.0, -0.4801, -0.0148, 0.0, -0.4824, -0.014, 0.0, -0.4823, -0.0114, 0.0, -0.4811, -0.0145, 0.0, -0.4809, -0.0114, -0.0017, -0.4801, -0.0148, -0.0039, -0.4801, -0.0148, -0.0017, -0.4824, -0.014, -0.0017, -0.4823, -0.0114, -0.0039, -0.4824, -0.014, -0.0039, -0.4823, -0.0114, -0.0017, -0.4812, -0.0144, -0.0039, -0.4812, -0.0144, 0.0, -0.4788, -0.0152, 0.0, -0.4767, -0.0158, 0.0, -0.4787, -0.0136, 0.0, -0.4787, -0.0114, 0.0, -0.4766, -0.0136,
0.0, -0.4766, -0.0114, -0.0017, -0.4787, -0.0152, -0.0039, -0.4787, -0.0152, -0.0017, -0.4766, -0.0159, -0.0039, -0.4766, -0.0159, 0.0, -0.4821, -0.0097, 0.0, -0.482, -0.0089, 0.0, -0.4817, -0.0069, 0.0, -0.4814, -0.0048, 0.0, -0.4811, -0.0026, -0.0017, -0.4821, -0.0094, -0.0039, -0.4821, -0.0094, -0.0017, -0.482, -0.0083, -0.0017, -0.4818, -0.0071, -0.0039, -0.482, -0.0083, -0.0039, -0.4818, -0.0071, -0.0017, -0.4815, -0.0049, -0.0017, -0.4812, -0.0027, -0.0039, -0.4815, -0.0049, -0.0039, -0.4812, -0.0027, 0.0, -0.4787, -0.0092, 0.0, -0.4787, -0.0071, 0.0, -0.4766, -0.0092, 0.0, -0.4766, -0.0071, 0.0, -0.4787, -0.0049, 0.0, -0.4787, -0.0027, 0.0, -0.4766, -0.0049, 0.0, -0.4766, -0.0027, 0.0, -0.4746, -0.0164, 0.0, -0.4713, -0.0168,
0.0, -0.4722, -0.0168, 0.0, -0.4744, -0.0136, 0.0, -0.4744, -0.0114, 0.0, -0.4722, -0.0136, 0.0, -0.4722, -0.0114, -0.0017, -0.4744, -0.0164, -0.0039, -0.4744, -0.0164, -0.0017, -0.4713, -0.0168, -0.0017, -0.4724, -0.0167, -0.0039, -0.4713, -0.0168, -0.0039, -0.4724, -0.0167, 0.0, -0.47, -0.0169, 0.0, -0.47, -0.0157, 0.0, -0.4679, -0.017, 0.0, -0.4679, -0.0157, 0.0, -0.47, -0.0136, 0.0, -0.47, -0.0114, 0.0, -0.4679, -0.0136, 0.0, -0.4679, -0.0114, -0.0017, -0.47, -0.0169, -0.0039, -0.47, -0.0169, -0.0017, -0.4679, -0.017, -0.0039, -0.4679, -0.017, 0.0, -0.4744, -0.0092, 0.0, -0.4744, -0.0071, 0.0, -0.4722, -0.0092, 0.0, -0.4722, -0.0071, 0.0, -0.4744, -0.0049, 0.0, -0.4744, -0.0027, 0.0, -0.4722, -0.0049,
0.0, -0.4722, -0.0027, 0.0, -0.47, -0.0092, 0.0, -0.47, -0.0071, 0.0, -0.4679, -0.0092, 0.0, -0.4679, -0.0071, 0.0, -0.47, -0.0049, 0.0, -0.47, -0.0027, 0.0, -0.4679, -0.0049, 0.0, -0.4679, -0.0027, 0.0, -0.4657, -0.0353, 0.0, -0.4657, -0.0331, 0.0, -0.4635, -0.0353, 0.0, -0.4635, -0.0331, 0.0, -0.4657, -0.0309, 0.0, -0.4657, -0.0287, 0.0, -0.4635, -0.0309, 0.0, -0.4636, -0.0287, -0.0017, -0.4657, -0.0287, -0.0039, -0.4657, -0.0287, -0.0017, -0.4635, -0.0287, -0.0039, -0.4635, -0.0287, 0.0, -0.4614, -0.0353, 0.0, -0.4614, -0.0331, 0.0, -0.4592, -0.0353, 0.0, -0.4592, -0.0331, 0.0, -0.4614, -0.0309, 0.0, -0.4615, -0.0284, 0.0, -0.4592, -0.0309, 0.0, -0.4594, -0.028, -0.0017, -0.4614, -0.0284,
-0.0039, -0.4614, -0.0284, -0.0017, -0.4592, -0.028, -0.0039, -0.4592, -0.028, 0.0, -0.457, -0.0353, 0.0, -0.457, -0.0331, 0.0, -0.4549, -0.0353, 0.0, -0.4549, -0.0331, 0.0, -0.457, -0.0309, 0.0, -0.4574, -0.0276, 0.0, -0.4549, -0.0309, 0.0, -0.4549, -0.0288, -0.0017, -0.458, -0.0277, -0.0039, -0.458, -0.0277, 0.0, -0.4527, -0.0353, 0.0, -0.4527, -0.0331, 0.0, -0.4505, -0.0353, 0.0, -0.4505, -0.0331, 0.0, -0.4527, -0.0309, 0.0, -0.4527, -0.0288, 0.0, -0.4505, -0.0309, 0.0, -0.4505, -0.0288, 0.0, -0.4566, -0.0274, 0.0, -0.4549, -0.027, -0.0017, -0.4569, -0.0274, -0.0039, -0.4569, -0.0274, -0.0017, -0.4549, -0.027, -0.0039, -0.4549, -0.027, 0.0, -0.4528, -0.0265, 0.0, -0.4507, -0.0261, -0.0017, -0.4527, -0.0265,
-0.0039, -0.4527, -0.0265, -0.0017, -0.4505, -0.026, -0.0039, -0.4505, -0.026, 0.0, -0.4657, -0.017, 0.0, -0.4657, -0.0158, 0.0, -0.4655, -0.0136, 0.0, -0.4652, -0.0115, -0.0017, -0.4657, -0.017, -0.0017, -0.4657, -0.0157, -0.0039, -0.4657, -0.017, -0.0039, -0.4657, -0.0157, -0.0017, -0.4655, -0.0136, -0.0017, -0.4651, -0.0114, -0.0039, -0.4655, -0.0136, -0.0039, -0.4651, -0.0114, 0.0, -0.4647, -0.0093, 0.0, -0.4657, -0.0071, 0.0, -0.4645, -0.0083, 0.0, -0.4643, -0.0071, 0.0, -0.4657, -0.0049, 0.0, -0.4657, -0.0027, 0.0, -0.4638, -0.005, 0.0, -0.4633, -0.0029, -0.0017, -0.4648, -0.0095, -0.0039, -0.4648, -0.0095, -0.0017, -0.4646, -0.0084, -0.0017, -0.4643, -0.0071, -0.0039, -0.4646, -0.0084, -0.0039, -0.4643, -0.0071, -0.0017, -0.4638, -0.0049,
-0.0017, -0.4633, -0.0027, -0.0039, -0.4638, -0.0049, -0.0039, -0.4633, -0.0027, 0.0, -0.4483, -0.07, 0.0, -0.4483, -0.0678, 0.0, -0.4462, -0.07, 0.0, -0.4462, -0.0678, 0.0, -0.4483, -0.0656, 0.0, -0.4483, -0.0635, 0.0, -0.4462, -0.0656, 0.0, -0.4462, -0.0635, 0.0, -0.444, -0.07, 0.0, -0.444, -0.0678, 0.0, -0.4418, -0.07, 0.0, -0.4418, -0.0678, 0.0, -0.444, -0.0656, 0.0, -0.444, -0.0635, 0.0, -0.4418, -0.0656, 0.0, -0.4418, -0.0635, 0.0, -0.4483, -0.0613, 0.0, -0.4483, -0.0591, 0.0, -0.4462, -0.0613, 0.0, -0.4462, -0.0591, 0.0, -0.4483, -0.057, 0.0, -0.4483, -0.0548, 0.0, -0.4462, -0.057, 0.0, -0.4462, -0.0548, 0.0, -0.444, -0.0613, 0.0, -0.444, -0.0591, 0.0, -0.4418, -0.0613,
0.0, -0.4418, -0.0591, 0.0, -0.444, -0.057, 0.0, -0.444, -0.0548, 0.0, -0.4418, -0.057, 0.0, -0.4418, -0.0548, 0.0, -0.4397, -0.07, 0.0, -0.4397, -0.0678, 0.0, -0.4375, -0.07, 0.0, -0.4375, -0.0678, 0.0, -0.4397, -0.0656, 0.0, -0.4397, -0.0635, 0.0, -0.4375, -0.0656, 0.0, -0.4375, -0.0635, 0.0, -0.4353, -0.07, 0.0, -0.4353, -0.0678, 0.0, -0.433, -0.069, 0.0, -0.4332, -0.0678, 0.0, -0.4353, -0.0656, 0.0, -0.4353, -0.0635, 0.0, -0.4332, -0.0656, 0.0, -0.4332, -0.0635, 0.0, -0.4397, -0.0613, 0.0, -0.4397, -0.0591, 0.0, -0.4375, -0.0613, 0.0, -0.4375, -0.0591, 0.0, -0.4397, -0.057, 0.0, -0.4397, -0.0548, 0.0, -0.4375, -0.057, 0.0, -0.4375, -0.0548, 0.0, -0.4353, -0.0613,
0.0, -0.4353, -0.0591, 0.0, -0.4332, -0.0613, 0.0, -0.4332, -0.0591, 0.0, -0.4353, -0.057, 0.0, -0.4353, -0.0548, 0.0, -0.4332, -0.057, 0.0, -0.4332, -0.0548, 0.0, -0.4483, -0.0526, 0.0, -0.4484, -0.0503, 0.0, -0.4462, -0.0526, 0.0, -0.4463, -0.0501, -0.0017, -0.4483, -0.0503, -0.0039, -0.4483, -0.0503, -0.0017, -0.4462, -0.0501, -0.0039, -0.4462, -0.0501, 0.0, -0.444, -0.0526, 0.0, -0.4441, -0.0498, 0.0, -0.4418, -0.0526, 0.0, -0.442, -0.0494, 0.0, -0.4411, -0.0492, -0.0017, -0.444, -0.0497, -0.0039, -0.444, -0.0497, -0.0017, -0.4424, -0.0495, -0.0039, -0.4424, -0.0495, -0.0017, -0.4413, -0.0493, -0.0039, -0.4413, -0.0493, 0.0, -0.4483, -0.0425, 0.0, -0.4461, -0.0423, 0.0, -0.4483, -0.0396, 0.0, -0.4483, -0.0374,
0.0, -0.4462, -0.0396, 0.0, -0.4462, -0.0374, -0.0017, -0.4483, -0.0425, -0.0039, -0.4483, -0.0425, -0.0017, -0.4462, -0.0423, -0.0039, -0.4462, -0.0423, 0.0, -0.4439, -0.042, 0.0, -0.4417, -0.0416, 0.0, -0.444, -0.0396, 0.0, -0.444, -0.0374, 0.0, -0.4418, -0.0396, 0.0, -0.4418, -0.0374, -0.0017, -0.444, -0.042, -0.0039, -0.444, -0.042, -0.0017, -0.4418, -0.0416, -0.0039, -0.4418, -0.0416, 0.0, -0.4397, -0.0526, 0.0, -0.4397, -0.0505, 0.0, -0.4375, -0.0526, 0.0, -0.4375, -0.0505, 0.0, -0.4397, -0.049, 0.0, -0.4376, -0.0486, -0.0017, -0.4397, -0.049, -0.0039, -0.4397, -0.049, -0.0017, -0.4375, -0.0486, -0.0039, -0.4375, -0.0486, 0.0, -0.4353, -0.0526, 0.0, -0.4353, -0.0505, 0.0, -0.4332, -0.0526, 0.0, -0.4332, -0.0505,
0.0, -0.4355, -0.0482, 0.0, -0.4334, -0.0476, -0.0017, -0.4353, -0.0481, -0.0039, -0.4353, -0.0481, -0.0017, -0.4331, -0.0476, -0.0039, -0.4331, -0.0476, 0.0, -0.4396, -0.0412, 0.0, -0.4382, -0.0409, 0.0, -0.4397, -0.0396, 0.0, -0.4397, -0.0374, 0.0, -0.4373, -0.0406, 0.0, -0.4375, -0.0374, -0.0017, -0.4397, -0.0412, -0.0039, -0.4397, -0.0412, -0.0017, -0.4381, -0.0408, -0.0039, -0.4381, -0.0408, -0.0017, -0.437, -0.0405, -0.0039, -0.437, -0.0405, 0.0, -0.4351, -0.0401, 0.0, -0.4353, -0.0374, 0.0, -0.433, -0.0396, 0.0, -0.4332, -0.0374, -0.0017, -0.4353, -0.0401, -0.0039, -0.4353, -0.0401, -0.0017, -0.4331, -0.0396, -0.0039, -0.4331, -0.0396, 0.0, -0.4328, -0.069, 0.0, -0.4311, -0.0689, 0.0, -0.4287, -0.0686, 0.0, -0.431, -0.0656,
0.0, -0.431, -0.0635, 0.0, -0.4288, -0.0656, 0.0, -0.4288, -0.0635, -0.0017, -0.4328, -0.069, -0.0017, -0.4308, -0.0689, -0.0039, -0.4328, -0.069, -0.0039, -0.4308, -0.0689, -0.0017, -0.4288, -0.0686, -0.0039, -0.4288, -0.0686, 0.0, -0.4265, -0.0681, 0.0, -0.4244, -0.0675, 0.0, -0.4266, -0.0656, 0.0, -0.4266, -0.0635, 0.0, -0.4245, -0.0656, 0.0, -0.4245, -0.0635, -0.0017, -0.4266, -0.0681, -0.0039, -0.4266, -0.0681, -0.0017, -0.4245, -0.0675, -0.0039, -0.4245, -0.0675, 0.0, -0.431, -0.0613, 0.0, -0.431, -0.0591, 0.0, -0.4288, -0.0613, 0.0, -0.4288, -0.0591, 0.0, -0.431, -0.057, 0.0, -0.431, -0.0548, 0.0, -0.4288, -0.057, 0.0, -0.4288, -0.0548, 0.0, -0.4266, -0.0613, 0.0, -0.4266, -0.0591, 0.0, -0.4245, -0.0613,
0.0, -0.4245, -0.0591, 0.0, -0.4266, -0.057, 0.0, -0.4266, -0.0548, 0.0, -0.4245, -0.057, 0.0, -0.4245, -0.0548, 0.0, -0.4229, -0.067, 0.0, -0.4219, -0.0667, 0.0, -0.4223, -0.0635, 0.0, -0.4199, -0.0659, 0.0, -0.4201, -0.0635, -0.0017, -0.4227, -0.067, -0.0039, -0.4227, -0.067, -0.0017, -0.4216, -0.0666, -0.0039, -0.4216, -0.0666, -0.0017, -0.4201, -0.066, -0.0039, -0.4201, -0.066, 0.0, -0.4179, -0.0651, 0.0, -0.418, -0.0635, 0.0, -0.4167, -0.0647, 0.0, -0.4157, -0.0642, -0.0017, -0.418, -0.0651, -0.0039, -0.418, -0.0651, -0.0017, -0.4167, -0.0646, -0.0017, -0.4156, -0.0642, -0.0039, -0.4167, -0.0646, -0.0039, -0.4156, -0.0642, 0.0, -0.4223, -0.0613, 0.0, -0.4223, -0.0591, 0.0, -0.4201, -0.0613, 0.0, -0.4201, -0.0591,
0.0, -0.4223, -0.057, 0.0, -0.4223, -0.0548, 0.0, -0.4201, -0.057, 0.0, -0.4201, -0.0548, 0.0, -0.418, -0.0613, 0.0, -0.418, -0.0591, 0.0, -0.4158, -0.0613, 0.0, -0.4158, -0.0591, 0.0, -0.418, -0.057, 0.0, -0.418, -0.0548, 0.0, -0.4158, -0.057, 0.0, -0.4158, -0.0548, 0.0, -0.431, -0.0526, 0.0, -0.431, -0.0505, 0.0, -0.4288, -0.0526, 0.0, -0.4288, -0.0505, 0.0, -0.4315, -0.0471, 0.0, -0.4306, -0.0468, 0.0, -0.4288, -0.0483, 0.0, -0.429, -0.0463, -0.0017, -0.432, -0.0472, -0.0017, -0.4309, -0.0469, -0.0039, -0.432, -0.0472, -0.0039, -0.4309, -0.0469, -0.0017, -0.4288, -0.0462, -0.0039, -0.4288, -0.0462, 0.0, -0.4266, -0.0526, 0.0, -0.4266, -0.0505, 0.0, -0.4245, -0.0526, 0.0, -0.4245, -0.0505,
0.0, -0.4266, -0.0483, 0.0, -0.4267, -0.0454, 0.0, -0.4245, -0.0483, 0.0, -0.4245, -0.0461, -0.0017, -0.4267, -0.0454, -0.0039, -0.4267, -0.0454, 0.0, -0.4309, -0.039, 0.0, -0.431, -0.0374, 0.0, -0.4298, -0.0386, 0.0, -0.4287, -0.0382, -0.0017, -0.431, -0.039, -0.0039, -0.431, -0.039, -0.0017, -0.4297, -0.0386, -0.0017, -0.4286, -0.0382, -0.0039, -0.4297, -0.0386, -0.0039, -0.4286, -0.0382, 0.0, -0.4256, -0.045, 0.0, -0.4245, -0.0446, 0.0, -0.4265, -0.0373, 0.0, -0.4251, -0.0367, -0.0017, -0.4256, -0.045, -0.0039, -0.4256, -0.045, -0.0017, -0.4245, -0.0446, -0.0039, -0.4245, -0.0446, -0.0017, -0.4266, -0.0373, -0.0039, -0.4266, -0.0373, -0.0017, -0.425, -0.0366, -0.0039, -0.425, -0.0366, 0.0, -0.4223, -0.0526, 0.0, -0.4223, -0.0505,
0.0, -0.4201, -0.0526, 0.0, -0.4201, -0.0505, 0.0, -0.4223, -0.0483, 0.0, -0.4223, -0.0461, 0.0, -0.4201, -0.0483, 0.0, -0.4201, -0.0461, 0.0, -0.418, -0.0526, 0.0, -0.418, -0.0505, 0.0, -0.4158, -0.0526, 0.0, -0.4158, -0.0505, 0.0, -0.418, -0.0483, 0.0, -0.418, -0.0461, 0.0, -0.4158, -0.0483, 0.0, -0.4158, -0.0461, 0.0, -0.4226, -0.0437, 0.0, -0.4206, -0.0429, 0.0, -0.4196, -0.0425, -0.0017, -0.4223, -0.0436, -0.0039, -0.4223, -0.0436, -0.0017, -0.4209, -0.043, -0.0017, -0.4198, -0.0425, -0.0039, -0.4209, -0.043, -0.0039, -0.4198, -0.0425, 0.0, -0.418, -0.0439, 0.0, -0.4182, -0.0419, 0.0, -0.4158, -0.0439, 0.0, -0.4161, -0.0409, 0.0, -0.4151, -0.0404, -0.0017, -0.418, -0.0418, -0.0039, -0.418, -0.0418,
-0.0017, -0.4162, -0.041, -0.0039, -0.4162, -0.041, -0.0017, -0.4152, -0.0405, -0.0039, -0.4152, -0.0405, 0.0, -0.4483, -0.0353, 0.0, -0.4483, -0.0331, 0.0, -0.4462, -0.0353, 0.0, -0.4462, -0.0331, 0.0, -0.4483, -0.0309, 0.0, -0.4483, -0.0288, 0.0, -0.4462, -0.0309, 0.0, -0.4462, -0.0288, 0.0, -0.444, -0.0353, 0.0, -0.444, -0.0331, 0.0, -0.4418, -0.0353, 0.0, -0.4418, -0.0331, 0.0, -0.444, -0.0309, 0.0, -0.444, -0.0288, 0.0, -0.4418, -0.0309, 0.0, -0.4418, -0.0288, 0.0, -0.4486, -0.0255, 0.0, -0.4477, -0.0253, 0.0, -0.4462, -0.0266, 0.0, -0.4463, -0.0249, -0.0017, -0.4489, -0.0256, -0.0017, -0.4479, -0.0253, -0.0039, -0.4489, -0.0256, -0.0039, -0.4479, -0.0253, -0.0017, -0.4462, -0.0248, -0.0039, -0.4462, -0.0248,
0.0, -0.444, -0.0266, 0.0, -0.4442, -0.0242, 0.0, -0.4418, -0.0266, 0.0, -0.4422, -0.0234, 0.0, -0.4412, -0.0231, -0.0017, -0.444, -0.0241, -0.0039, -0.444, -0.0241, -0.0017, -0.4424, -0.0235, -0.0039, -0.4424, -0.0235, -0.0017, -0.4413, -0.0231, -0.0039, -0.4413, -0.0231, 0.0, -0.4397, -0.0353, 0.0, -0.4397, -0.0331, 0.0, -0.4375, -0.0353, 0.0, -0.4375, -0.0331, 0.0, -0.4397, -0.0309, 0.0, -0.4397, -0.0288, 0.0, -0.4375, -0.0309, 0.0, -0.4375, -0.0288, 0.0, -0.4353, -0.0353, 0.0, -0.4353, -0.0331, 0.0, -0.4332, -0.0353, 0.0, -0.4332, -0.0331, 0.0, -0.4353, -0.0309, 0.0, -0.4353, -0.0288, 0.0, -0.4332, -0.0309, 0.0, -0.4332, -0.0288, 0.0, -0.4397, -0.0266, 0.0, -0.4397, -0.0244, 0.0, -0.4375, -0.0266,
0.0, -0.4375, -0.0244, 0.0, -0.4398, -0.0225, 0.0, -0.4375, -0.0216, 0.0, -0.4364, -0.0212, -0.0017, -0.4397, -0.0225, -0.0039, -0.4397, -0.0225, -0.0017, -0.4375, -0.0216, -0.0017, -0.4364, -0.0212, -0.0039, -0.4375, -0.0216, -0.0039, -0.4364, -0.0212, 0.0, -0.4353, -0.0266, 0.0, -0.4353, -0.0244, 0.0, -0.4332, -0.0266, 0.0, -0.4332, -0.0244, 0.0, -0.4353, -0.0222, 0.0, -0.4354, -0.0207, 0.0, -0.4332, -0.0222, 0.0, -0.4334, -0.0199, -0.0017, -0.4353, -0.0207, -0.0039, -0.4353, -0.0207, -0.0017, -0.4332, -0.0197, -0.0039, -0.4332, -0.0197, 0.0, -0.431, -0.0353, 0.0, -0.431, -0.0331, 0.0, -0.4288, -0.0353, 0.0, -0.4288, -0.0331, 0.0, -0.431, -0.0309, 0.0, -0.431, -0.0288, 0.0, -0.4288, -0.0309, 0.0, -0.4288, -0.0288,
0.0, -0.4266, -0.0353, 0.0, -0.4266, -0.0331, 0.0, -0.4241, -0.0362, 0.0, -0.4245, -0.0331, 0.0, -0.4266, -0.0309, 0.0, -0.4266, -0.0288, 0.0, -0.4245, -0.0309, 0.0, -0.4245, -0.0288, -0.0017, -0.4239, -0.0361, -0.0039, -0.4239, -0.0361, 0.0, -0.431, -0.0266, 0.0, -0.431, -0.0244, 0.0, -0.4288, -0.0266, 0.0, -0.4288, -0.0244, 0.0, -0.431, -0.0222, 0.0, -0.4315, -0.019, 0.0, -0.4288, -0.0222, 0.0, -0.4288, -0.0201, -0.0017, -0.4318, -0.0191, -0.0039, -0.4318, -0.0191, 0.0, -0.4266, -0.0266, 0.0, -0.4266, -0.0244, 0.0, -0.4245, -0.0266, 0.0, -0.4245, -0.0244, 0.0, -0.4266, -0.0222, 0.0, -0.4266, -0.0201, 0.0, -0.4245, -0.0222, 0.0, -0.4245, -0.0201, 0.0, -0.4221, -0.0353, 0.0, -0.4223, -0.0331,
0.0, -0.4206, -0.0346, 0.0, -0.4196, -0.0342, 0.0, -0.4223, -0.0309, 0.0, -0.4223, -0.0288, 0.0, -0.4201, -0.0309, 0.0, -0.4201, -0.0288, -0.0017, -0.4223, -0.0354, -0.0039, -0.4223, -0.0354, -0.0017, -0.4204, -0.0345, -0.0017, -0.4194, -0.034, -0.0039, -0.4204, -0.0345, -0.0039, -0.4194, -0.034, 0.0, -0.4177, -0.0332, 0.0, -0.4163, -0.0325, 0.0, -0.418, -0.0309, 0.0, -0.418, -0.0288, 0.0, -0.4153, -0.0319, 0.0, -0.4158, -0.0288, -0.0017, -0.4179, -0.0333, -0.0039, -0.4179, -0.0333, -0.0017, -0.4162, -0.0324, -0.0039, -0.4162, -0.0324, -0.0017, -0.4151, -0.0318, -0.0039, -0.4151, -0.0318, 0.0, -0.4223, -0.0266, 0.0, -0.4223, -0.0244, 0.0, -0.4201, -0.0266, 0.0, -0.4201, -0.0244, 0.0, -0.4223, -0.0222, 0.0, -0.4223, -0.0201,
0.0, -0.4201, -0.0222, 0.0, -0.4201, -0.0201, 0.0, -0.418, -0.0266, 0.0, -0.418, -0.0244, 0.0, -0.4158, -0.0266, 0.0, -0.4158, -0.0244, 0.0, -0.418, -0.0222, 0.0, -0.418, -0.0201, 0.0, -0.4158, -0.0222, 0.0, -0.4158, -0.0201, 0.0, -0.4305, -0.0185, 0.0, -0.429, -0.0179, -0.0017, -0.4307, -0.0186, -0.0039, -0.4307, -0.0186, -0.0017, -0.4288, -0.0178, -0.0039, -0.4288, -0.0178, 0.0, -0.427, -0.017, 0.0, -0.426, -0.0165, 0.0, -0.4245, -0.0179, 0.0, -0.4247, -0.0158, -0.0017, -0.4272, -0.017, -0.0017, -0.4262, -0.0165, -0.0039, -0.4272, -0.017, -0.0039, -0.4262, -0.0165, -0.0017, -0.4245, -0.0157, -0.0039, -0.4245, -0.0157, 0.0, -0.4257, -0.0018, 0.0, -0.4246, -0.0023, -0.0017, -0.4258, -0.0017, -0.0039, -0.4258, -0.0017,
-0.0017, -0.4245, -0.0023, -0.0039, -0.4245, -0.0023, 0.0, -0.4223, -0.0179, 0.0, -0.4228, -0.0148, 0.0, -0.4201, -0.0179, 0.0, -0.4201, -0.0157, 0.0, -0.4217, -0.0142, 0.0, -0.4204, -0.0135, -0.0017, -0.423, -0.0149, -0.0039, -0.423, -0.0149, -0.0017, -0.4219, -0.0143, -0.0039, -0.4219, -0.0143, -0.0017, -0.4201, -0.0134, -0.0039, -0.4201, -0.0134, 0.0, -0.418, -0.0179, 0.0, -0.418, -0.0157, 0.0, -0.4158, -0.0179, 0.0, -0.4158, -0.0157, 0.0, -0.4185, -0.0126, 0.0, -0.4175, -0.012, 0.0, -0.4158, -0.0136, 0.0, -0.4161, -0.0113, -0.0017, -0.4187, -0.0127, -0.0017, -0.4176, -0.0121, -0.0039, -0.4187, -0.0127, -0.0039, -0.4176, -0.0121, -0.0017, -0.4158, -0.0111, -0.0039, -0.4158, -0.0111, 0.0, -0.4224, -0.0029, 0.0, -0.4201, -0.0032,
-0.0017, -0.4223, -0.003, -0.0039, -0.4223, -0.003, -0.0017, -0.4201, -0.0032, -0.0039, -0.4201, -0.0032, 0.0, -0.4179, -0.003, 0.0, -0.4156, -0.0025, -0.0017, -0.4179, -0.003, -0.0039, -0.4179, -0.003, -0.0017, -0.4158, -0.0026, -0.0039, -0.4158, -0.0026, 0.0, -0.3976, -0.1059, -0.0017, -0.3977, -0.1059, -0.0039, -0.3977, -0.1059, 0.0, -0.3963, -0.1061, 0.0, -0.3942, -0.1066, -0.0017, -0.3963, -0.1061, -0.0039, -0.3963, -0.1061, -0.0017, -0.3941, -0.1066, -0.0039, -0.3941, -0.1066, 0.0, -0.392, -0.1069, 0.0, -0.3898, -0.1071, -0.0017, -0.3919, -0.1069, -0.0039, -0.3919, -0.1069, -0.0017, -0.3898, -0.1071, -0.0039, -0.3898, -0.1071, 0.0, -0.3876, -0.1072, 0.0, -0.3854, -0.1072, -0.0017, -0.3876, -0.1072, -0.0039, -0.3876, -0.1072, -0.0017, -0.3854, -0.1072,
-0.0039, -0.3854, -0.1072, 0.0, -0.3832, -0.1072, 0.0, -0.3811, -0.1072, -0.0017, -0.3832, -0.1072, -0.0039, -0.3832, -0.1072, -0.0017, -0.3811, -0.1072, -0.0039, -0.3811, -0.1072, 0.0, -0.4129, -0.1017, 0.0, -0.4116, -0.1022, 0.0, -0.4139, -0.1013, 0.0, -0.4136, -0.0982, 0.0, -0.4115, -0.1004, 0.0, -0.4115, -0.0982, -0.0017, -0.413, -0.1016, -0.0039, -0.413, -0.1016, -0.0017, -0.4115, -0.1022, -0.0039, -0.4115, -0.1022, -0.0017, -0.4141, -0.1012, -0.0039, -0.4141, -0.1012, 0.0, -0.4095, -0.1029, 0.0, -0.4065, -0.1039, 0.0, -0.4075, -0.1036, 0.0, -0.4093, -0.1004, 0.0, -0.4093, -0.0982, 0.0, -0.4071, -0.1004, 0.0, -0.4071, -0.0982, -0.0017, -0.4093, -0.103, -0.0039, -0.4093, -0.103, -0.0017, -0.4067, -0.1038, -0.0017, -0.4078, -0.1035,
-0.0039, -0.4067, -0.1038, -0.0039, -0.4078, -0.1035, 0.0, -0.4136, -0.096, 0.0, -0.4136, -0.0939, 0.0, -0.4115, -0.096, 0.0, -0.4115, -0.0939, 0.0, -0.4136, -0.0917, 0.0, -0.4136, -0.0895, 0.0, -0.4115, -0.0917, 0.0, -0.4115, -0.0895, 0.0, -0.4093, -0.096, 0.0, -0.4093, -0.0939, 0.0, -0.4071, -0.096, 0.0, -0.4071, -0.0939, 0.0, -0.4093, -0.0917, 0.0, -0.4093, -0.0895, 0.0, -0.4071, -0.0917, 0.0, -0.4071, -0.0895, 0.0, -0.405, -0.1043, 0.0, -0.4049, -0.1025, 0.0, -0.4029, -0.1048, 0.0, -0.4028, -0.1025, 0.0, -0.4049, -0.1004, 0.0, -0.4049, -0.0982, 0.0, -0.4028, -0.1004, 0.0, -0.4028, -0.0982, -0.0017, -0.4049, -0.1043, -0.0039, -0.4049, -0.1043, -0.0017, -0.4028, -0.1048, -0.0039, -0.4028, -0.1048,
0.0, -0.4008, -0.1053, 0.0, -0.4006, -0.1025, 0.0, -0.3985, -0.1057, 0.0, -0.3984, -0.1025, 0.0, -0.4006, -0.1004, 0.0, -0.4006, -0.0982, 0.0, -0.3984, -0.1004, 0.0, -0.3984, -0.0982, -0.0017, -0.4006, -0.1053, -0.0039, -0.4006, -0.1053, -0.0017, -0.3988, -0.1056, -0.0039, -0.3988, -0.1056, 0.0, -0.4049, -0.096, 0.0, -0.4049, -0.0939, 0.0, -0.4028, -0.096, 0.0, -0.4028, -0.0939, 0.0, -0.4049, -0.0917, 0.0, -0.4049, -0.0895, 0.0, -0.4028, -0.0917, 0.0, -0.4026, -0.0886, -0.0017, -0.4024, -0.0887, -0.0039, -0.4024, -0.0887, 0.0, -0.4006, -0.096, 0.0, -0.4006, -0.0939, 0.0, -0.3984, -0.096, 0.0, -0.3984, -0.0939, 0.0, -0.4006, -0.0917, 0.0, -0.4004, -0.0894, 0.0, -0.3984, -0.0917, 0.0, -0.3984, -0.09,
-0.0017, -0.4006, -0.0893, -0.0039, -0.4006, -0.0893, -0.0017, -0.3984, -0.0899, -0.0039, -0.3984, -0.0899, 0.0, -0.4136, -0.0874, 0.0, -0.4131, -0.0842, 0.0, -0.4115, -0.0874, 0.0, -0.4112, -0.0851, 0.0, -0.4141, -0.0837, -0.0017, -0.4129, -0.0843, -0.0039, -0.4129, -0.0843, -0.0017, -0.4115, -0.085, -0.0039, -0.4115, -0.085, -0.0017, -0.414, -0.0838, -0.0039, -0.414, -0.0838, 0.0, -0.4087, -0.0862, 0.0, -0.4097, -0.0858, 0.0, -0.4068, -0.087, -0.0017, -0.4084, -0.0863, -0.0017, -0.4095, -0.0859, -0.0039, -0.4084, -0.0863, -0.0039, -0.4095, -0.0859, -0.0017, -0.4071, -0.0869, -0.0039, -0.4071, -0.0869, 0.0, -0.4048, -0.0878, 0.0, -0.4036, -0.0883, -0.0017, -0.4049, -0.0877, -0.0039, -0.4049, -0.0877, -0.0017, -0.4035, -0.0883, -0.0039, -0.4035, -0.0883,
0.0, -0.3963, -0.1047, 0.0, -0.3963, -0.1025, 0.0, -0.3941, -0.1047, 0.0, -0.3941, -0.1025, 0.0, -0.3963, -0.1004, 0.0, -0.3963, -0.0982, 0.0, -0.3941, -0.1004, 0.0, -0.3941, -0.0982, 0.0, -0.3919, -0.1047, 0.0, -0.3919, -0.1025, 0.0, -0.3897, -0.1047, 0.0, -0.3897, -0.1025, 0.0, -0.3919, -0.1004, 0.0, -0.3919, -0.0982, 0.0, -0.3897, -0.1004, 0.0, -0.3897, -0.0982, 0.0, -0.3963, -0.096, 0.0, -0.3963, -0.0939, 0.0, -0.3941, -0.096, 0.0, -0.3941, -0.0939, 0.0, -0.3959, -0.0905, 0.0, -0.3967, -0.0903, 0.0, -0.3939, -0.091, -0.0017, -0.3954, -0.0907, -0.0017, -0.3965, -0.0904, -0.0039, -0.3954, -0.0907, -0.0039, -0.3965, -0.0904, -0.0017, -0.3941, -0.091, -0.0039, -0.3941, -0.091, 0.0, -0.3919, -0.096,
0.0, -0.3919, -0.0939, 0.0, -0.3897, -0.096, 0.0, -0.3897, -0.0939, 0.0, -0.3918, -0.0916, 0.0, -0.3897, -0.0921, -0.0017, -0.3919, -0.0915, -0.0039, -0.3919, -0.0915, -0.0017, -0.3898, -0.0921, -0.0039, -0.3898, -0.0921, 0.0, -0.3876, -0.1047, 0.0, -0.3876, -0.1025, 0.0, -0.3854, -0.1047, 0.0, -0.3854, -0.1025, 0.0, -0.3876, -0.1004, 0.0, -0.3876, -0.0982, 0.0, -0.3854, -0.1004, 0.0, -0.3854, -0.0982, 0.0, -0.3832, -0.1047, 0.0, -0.3832, -0.1025, 0.0, -0.3811, -0.1047, 0.0, -0.3811, -0.1025, 0.0, -0.3832, -0.1004, 0.0, -0.3832, -0.0982, 0.0, -0.3811, -0.1004, 0.0, -0.3811, -0.0982, 0.0, -0.3876, -0.096, 0.0, -0.3876, -0.0939, 0.0, -0.3854, -0.096, 0.0, -0.3854, -0.0939, 0.0, -0.3876, -0.0924,
0.0, -0.3854, -0.0925, -0.0017, -0.3876, -0.0924, -0.0039, -0.3876, -0.0924, -0.0017, -0.3854, -0.0925, -0.0039, -0.3854, -0.0925, 0.0, -0.3832, -0.096, 0.0, -0.3831, -0.0928, 0.0, -0.3811, -0.096, 0.0, -0.381, -0.0931, 0.0, -0.3839, -0.0927, -0.0017, -0.3827, -0.0928, -0.0039, -0.3827, -0.0928, -0.0017, -0.3811, -0.0931, -0.0039, -0.3811, -0.0931, -0.0017, -0.3838, -0.0927, -0.0039, -0.3838, -0.0927, 0.0, -0.3789, -0.1072, 0.0, -0.3767, -0.1072, -0.0017, -0.3789, -0.1072, -0.0039, -0.3789, -0.1072, -0.0017, -0.3767, -0.1072, -0.0039, -0.3767, -0.1072, 0.0, -0.3746, -0.1072, 0.0, -0.3724, -0.1072, -0.0017, -0.3746, -0.1072, -0.0039, -0.3746, -0.1072, -0.0017, -0.3724, -0.1072, -0.0039, -0.3724, -0.1072, 0.0, -0.3702, -0.1072, 0.0, -0.368, -0.1071,
-0.0017, -0.3702, -0.1072, -0.0039, -0.3702, -0.1072, -0.0017, -0.368, -0.1072, -0.0039, -0.368, -0.1072, 0.0, -0.3658, -0.107, 0.0, -0.3636, -0.1067, -0.0017, -0.3659, -0.107, -0.0039, -0.3659, -0.107, -0.0017, -0.3637, -0.1067, -0.0039, -0.3637, -0.1067, 0.0, -0.3615, -0.1062, 0.0, -0.3601, -0.1059, -0.0017, -0.3615, -0.1062, -0.0039, -0.3615, -0.1062, -0.0017, -0.36, -0.1059, -0.0039, -0.36, -0.1059, 0.0, -0.3789, -0.1047, 0.0, -0.3789, -0.1025, 0.0, -0.3767, -0.1047, 0.0, -0.3767, -0.1025, 0.0, -0.3789, -0.1004, 0.0, -0.3789, -0.0982, 0.0, -0.3767, -0.1004, 0.0, -0.3767, -0.0982, 0.0, -0.3746, -0.1047, 0.0, -0.3746, -0.1025, 0.0, -0.3724, -0.1047, 0.0, -0.3724, -0.1025, 0.0, -0.3746, -0.1004, 0.0, -0.3746, -0.0982,
0.0, -0.3724, -0.1004, 0.0, -0.3724, -0.0982, 0.0, -0.3789, -0.096, 0.0, -0.3788, -0.0933, 0.0, -0.3767, -0.096, 0.0, -0.3767, -0.0934, -0.0017, -0.3789, -0.0933, -0.0039, -0.3789, -0.0933, -0.0017, -0.3767, -0.0934, -0.0039, -0.3767, -0.0934, 0.0, -0.3746, -0.096, 0.0, -0.3746, -0.0934, 0.0, -0.3724, -0.096, 0.0, -0.3724, -0.0934, -0.0017, -0.3746, -0.0934, -0.0039, -0.3746, -0.0934, -0.0017, -0.3724, -0.0934, -0.0039, -0.3724, -0.0934, 0.0, -0.3702, -0.1047, 0.0, -0.3702, -0.1025, 0.0, -0.368, -0.1047, 0.0, -0.368, -0.1025, 0.0, -0.3702, -0.1004, 0.0, -0.3702, -0.0982, 0.0, -0.368, -0.1004, 0.0, -0.368, -0.0982, 0.0, -0.3659, -0.1047, 0.0, -0.3659, -0.1025, 0.0, -0.3637, -0.1047, 0.0, -0.3637, -0.1025,
0.0, -0.3659, -0.1004, 0.0, -0.3659, -0.0982, 0.0, -0.3637, -0.1004, 0.0, -0.3637, -0.0982, 0.0, -0.3702, -0.096, 0.0, -0.3703, -0.0934, 0.0, -0.368, -0.096, 0.0, -0.3682, -0.0931, -0.0017, -0.3702, -0.0934, -0.0039, -0.3702, -0.0934, -0.0017, -0.368, -0.0931, -0.0039, -0.368, -0.0931, 0.0, -0.3659, -0.096, 0.0, -0.366, -0.0928, 0.0, -0.3637, -0.096, 0.0, -0.3637, -0.0939, 0.0, -0.3651, -0.0927, 0.0, -0.3637, -0.0925, -0.0017, -0.3664, -0.0929, -0.0039, -0.3664, -0.0929, -0.0017, -0.3653, -0.0927, -0.0039, -0.3653, -0.0927, -0.0017, -0.3637, -0.0925, -0.0039, -0.3637, -0.0925, 0.0, -0.3615, -0.1047, 0.0, -0.3615, -0.1025, 0.0, -0.3592, -0.1057, 0.0, -0.3594, -0.1025, 0.0, -0.3615, -0.1004, 0.0, -0.3615, -0.0982,
0.0, -0.3594, -0.1004, 0.0, -0.3594, -0.0982, -0.0017, -0.3589, -0.1057, -0.0039, -0.3589, -0.1057, 0.0, -0.357, -0.1052, 0.0, -0.3572, -0.1025, 0.0, -0.3549, -0.1048, 0.0, -0.355, -0.1025, 0.0, -0.3572, -0.1004, 0.0, -0.3572, -0.0982, 0.0, -0.355, -0.1004, 0.0, -0.355, -0.0982, -0.0017, -0.3572, -0.1053, -0.0039, -0.3572, -0.1053, -0.0017, -0.355, -0.1048, -0.0039, -0.355, -0.1048, 0.0, -0.3615, -0.096, 0.0, -0.3615, -0.0939, 0.0, -0.3594, -0.096, 0.0, -0.3594, -0.0939, 0.0, -0.3616, -0.0924, 0.0, -0.3594, -0.0922, -0.0017, -0.3615, -0.0924, -0.0039, -0.3615, -0.0924, -0.0017, -0.3594, -0.0922, -0.0039, -0.3594, -0.0922, 0.0, -0.3572, -0.096, 0.0, -0.3572, -0.0939, 0.0, -0.355, -0.096, 0.0, -0.355, -0.0939,
0.0, -0.3573, -0.0918, 0.0, -0.3552, -0.0913, -0.0017, -0.3572, -0.0918, -0.0039, -0.3572, -0.0918, -0.0017, -0.355, -0.0912, -0.0039, -0.355, -0.0912, 0.0, -0.3528, -0.1043, 0.0, -0.3529, -0.1025, 0.0, -0.3512, -0.1039, 0.0, -0.3503, -0.1037, 0.0, -0.3529, -0.1004, 0.0, -0.3529, -0.0982, 0.0, -0.3507, -0.1004, 0.0, -0.3507, -0.0982, -0.0017, -0.3529, -0.1043, -0.0039, -0.3529, -0.1043, -0.0017, -0.3509, -0.1039, -0.0017, -0.3498, -0.1036, -0.0039, -0.3509, -0.1039, -0.0039, -0.3498, -0.1036, 0.0, -0.3483, -0.1031, 0.0, -0.3462, -0.1024, 0.0, -0.3485, -0.1004, 0.0, -0.3485, -0.0982, 0.0, -0.3463, -0.1004, 0.0, -0.3463, -0.0982, -0.0017, -0.3485, -0.1032, -0.0039, -0.3485, -0.1032, -0.0017, -0.3463, -0.1025, -0.0039, -0.3463, -0.1025,
0.0, -0.3529, -0.096, 0.0, -0.3529, -0.0939, 0.0, -0.3507, -0.096, 0.0, -0.3507, -0.0939, 0.0, -0.3531, -0.0907, 0.0, -0.3521, -0.0904, 0.0, -0.3507, -0.0917, 0.0, -0.3508, -0.0901, -0.0017, -0.3534, -0.0908, -0.0017, -0.3523, -0.0905, -0.0039, -0.3534, -0.0908, -0.0039, -0.3523, -0.0905, -0.0017, -0.3507, -0.0901, -0.0039, -0.3507, -0.0901, 0.0, -0.3485, -0.096, 0.0, -0.3485, -0.0939, 0.0, -0.3463, -0.096, 0.0, -0.3463, -0.0939, 0.0, -0.3485, -0.0917, 0.0, -0.3487, -0.0895, 0.0, -0.3463, -0.0917, 0.0, -0.3464, -0.0888, -0.0017, -0.3485, -0.0895, -0.0039, -0.3485, -0.0895, -0.0017, -0.3465, -0.0888, -0.0039, -0.3465, -0.0888, 0.0, -0.3454, -0.0884, -0.0017, -0.3454, -0.0884, -0.0039, -0.3454, -0.0884, 0.0, -0.4134, -0.0633,
0.0, -0.412, -0.0627, -0.0017, -0.4136, -0.0634, -0.0039, -0.4136, -0.0634, -0.0017, -0.4119, -0.0627, -0.0039, -0.4119, -0.0627, 0.0, -0.4136, -0.0613, 0.0, -0.4136, -0.0591, 0.0, -0.411, -0.0623, 0.0, -0.4115, -0.0591, 0.0, -0.4136, -0.057, 0.0, -0.4136, -0.0548, 0.0, -0.4115, -0.057, 0.0, -0.4115, -0.0548, -0.0017, -0.4108, -0.0622, -0.0039, -0.4108, -0.0622, 0.0, -0.409, -0.0614, 0.0, -0.4093, -0.0591, 0.0, -0.4076, -0.0607, 0.0, -0.4065, -0.0602, 0.0, -0.4093, -0.057, 0.0, -0.4093, -0.0548, 0.0, -0.4071, -0.057, 0.0, -0.4071, -0.0548, -0.0017, -0.4093, -0.0615, -0.0039, -0.4093, -0.0615, -0.0017, -0.4074, -0.0606, -0.0017, -0.4063, -0.0601, -0.0039, -0.4074, -0.0606, -0.0039, -0.4063, -0.0601, 0.0, -0.4047, -0.0592,
0.0, -0.4033, -0.0585, 0.0, -0.4049, -0.057, 0.0, -0.4049, -0.0548, 0.0, -0.4023, -0.0579, 0.0, -0.4028, -0.0548, -0.0017, -0.4049, -0.0594, -0.0039, -0.4049, -0.0594, -0.0017, -0.4032, -0.0584, -0.0039, -0.4032, -0.0584, -0.0017, -0.4021, -0.0578, -0.0039, -0.4021, -0.0578, 0.0, -0.4004, -0.0568, 0.0, -0.4006, -0.0548, 0.0, -0.3992, -0.0561, 0.0, -0.3981, -0.0555, -0.0017, -0.4006, -0.0569, -0.0039, -0.4006, -0.0569, -0.0017, -0.3991, -0.0561, -0.0017, -0.398, -0.0555, -0.0039, -0.3991, -0.0561, -0.0039, -0.398, -0.0555, 0.0, -0.4136, -0.0526, 0.0, -0.4136, -0.0505, 0.0, -0.4115, -0.0526, 0.0, -0.4115, -0.0505, 0.0, -0.4136, -0.0483, 0.0, -0.4136, -0.0461, 0.0, -0.4115, -0.0483, 0.0, -0.4115, -0.0461, 0.0, -0.4093, -0.0526,
0.0, -0.4093, -0.0505, 0.0, -0.4071, -0.0526, 0.0, -0.4071, -0.0505, 0.0, -0.4093, -0.0483, 0.0, -0.4093, -0.0461, 0.0, -0.4071, -0.0483, 0.0, -0.4071, -0.0461, 0.0, -0.4136, -0.0439, 0.0, -0.4136, -0.0418, 0.0, -0.4115, -0.0439, 0.0, -0.4115, -0.0418, 0.0, -0.4138, -0.0398, 0.0, -0.4118, -0.0388, 0.0, -0.4107, -0.0383, -0.0017, -0.4136, -0.0397, -0.0039, -0.4136, -0.0397, -0.0017, -0.4119, -0.0389, -0.0017, -0.4108, -0.0383, -0.0039, -0.4119, -0.0389, -0.0039, -0.4108, -0.0383, 0.0, -0.4093, -0.0439, 0.0, -0.4093, -0.0418, 0.0, -0.4071, -0.0439, 0.0, -0.4071, -0.0418, 0.0, -0.4093, -0.0396, 0.0, -0.4095, -0.0376, 0.0, -0.4071, -0.0396, 0.0, -0.4076, -0.0366, -0.0017, -0.4093, -0.0375, -0.0039, -0.4093, -0.0375,
-0.0017, -0.4077, -0.0366, -0.0039, -0.4077, -0.0366, 0.0, -0.4049, -0.0526, 0.0, -0.4049, -0.0505, 0.0, -0.4028, -0.0526, 0.0, -0.4028, -0.0505, 0.0, -0.4049, -0.0483, 0.0, -0.4049, -0.0461, 0.0, -0.4028, -0.0483, 0.0, -0.4028, -0.0461, 0.0, -0.4006, -0.0526, 0.0, -0.4006, -0.0505, 0.0, -0.3984, -0.0526, 0.0, -0.3984, -0.0505, 0.0, -0.4006, -0.0483, 0.0, -0.4006, -0.0461, 0.0, -0.3984, -0.0483, 0.0, -0.3984, -0.0461, 0.0, -0.4049, -0.0439, 0.0, -0.4049, -0.0418, 0.0, -0.4028, -0.0439, 0.0, -0.4028, -0.0418, 0.0, -0.4049, -0.0396, 0.0, -0.4049, -0.0374, 0.0, -0.4028, -0.0396, 0.0, -0.4028, -0.0374, 0.0, -0.4006, -0.0439, 0.0, -0.4006, -0.0418, 0.0, -0.3984, -0.0439, 0.0, -0.3984, -0.0418,
0.0, -0.4006, -0.0396, 0.0, -0.4006, -0.0374, 0.0, -0.3984, -0.0396, 0.0, -0.3984, -0.0374, 0.0, -0.3961, -0.0544, 0.0, -0.3951, -0.0538, -0.0017, -0.3963, -0.0545, -0.0039, -0.3963, -0.0545, -0.0017, -0.395, -0.0538, -0.0039, -0.395, -0.0538, 0.0, -0.3963, -0.0526, 0.0, -0.3963, -0.0505, 0.0, -0.3939, -0.0531, 0.0, -0.3941, -0.0505, 0.0, -0.3963, -0.0483, 0.0, -0.3963, -0.0461, 0.0, -0.3941, -0.0483, 0.0, -0.3941, -0.0461, -0.0017, -0.394, -0.0532, -0.0039, -0.394, -0.0532, 0.0, -0.3923, -0.0521, 0.0, -0.3912, -0.0515, 0.0, -0.3895, -0.0504, 0.0, -0.3919, -0.0483, 0.0, -0.3919, -0.0461, 0.0, -0.3897, -0.0483, 0.0, -0.3897, -0.0461, -0.0017, -0.3921, -0.0521, -0.0017, -0.3911, -0.0514, -0.0039, -0.3921, -0.0521,
-0.0039, -0.3911, -0.0514, -0.0017, -0.3897, -0.0506, -0.0039, -0.3897, -0.0506, 0.0, -0.3963, -0.0439, 0.0, -0.3963, -0.0418, 0.0, -0.3941, -0.0439, 0.0, -0.3941, -0.0418, 0.0, -0.3963, -0.0396, 0.0, -0.3963, -0.0374, 0.0, -0.3941, -0.0396, 0.0, -0.3941, -0.0374, 0.0, -0.3919, -0.0439, 0.0, -0.3919, -0.0418, 0.0, -0.3897, -0.0439, 0.0, -0.3897, -0.0418, 0.0, -0.3919, -0.0396, 0.0, -0.3919, -0.0374, 0.0, -0.3897, -0.0396, 0.0, -0.3897, -0.0374, 0.0, -0.3883, -0.0497, 0.0, -0.3872, -0.049, 0.0, -0.3876, -0.0461, 0.0, -0.3857, -0.0479, 0.0, -0.3846, -0.0471, -0.0017, -0.3883, -0.0496, -0.0039, -0.3883, -0.0496, -0.0017, -0.3872, -0.0489, -0.0039, -0.3872, -0.0489, -0.0017, -0.3856, -0.0478, -0.0017, -0.3845, -0.0471,
-0.0039, -0.3856, -0.0478, -0.0039, -0.3845, -0.0471, 0.0, -0.383, -0.0459, 0.0, -0.382, -0.0451, -0.0017, -0.3832, -0.0461, -0.0039, -0.3832, -0.0461, -0.0017, -0.382, -0.0451, -0.0039, -0.382, -0.0451, 0.0, -0.3876, -0.0439, 0.0, -0.3876, -0.0418, 0.0, -0.3854, -0.0439, 0.0, -0.3854, -0.0418, 0.0, -0.3876, -0.0396, 0.0, -0.3876, -0.0374, 0.0, -0.3854, -0.0396, 0.0, -0.3854, -0.0374, 0.0, -0.3832, -0.0439, 0.0, -0.3832, -0.0418, 0.0, -0.3809, -0.0443, 0.0, -0.3811, -0.0418, 0.0, -0.3832, -0.0396, 0.0, -0.3832, -0.0374, 0.0, -0.3811, -0.0396, 0.0, -0.3811, -0.0374, -0.0017, -0.3809, -0.0443, -0.0039, -0.3809, -0.0443, 0.0, -0.4134, -0.0308, 0.0, -0.4136, -0.0288, 0.0, -0.4122, -0.0301, 0.0, -0.4111, -0.0295,
-0.0017, -0.4136, -0.0309, -0.0039, -0.4136, -0.0309, -0.0017, -0.4121, -0.0301, -0.0017, -0.411, -0.0295, -0.0039, -0.4121, -0.0301, -0.0039, -0.411, -0.0295, 0.0, -0.4065, -0.036, 0.0, -0.4091, -0.0283, 0.0, -0.4081, -0.0277, -0.0017, -0.4066, -0.036, -0.0039, -0.4066, -0.036, -0.0017, -0.4093, -0.0284, -0.0039, -0.4093, -0.0284, -0.0017, -0.4081, -0.0277, -0.0039, -0.4081, -0.0277, 0.0, -0.4136, -0.0266, 0.0, -0.4136, -0.0244, 0.0, -0.4115, -0.0266, 0.0, -0.4115, -0.0244, 0.0, -0.4136, -0.0222, 0.0, -0.4136, -0.0201, 0.0, -0.4115, -0.0222, 0.0, -0.4115, -0.0201, 0.0, -0.4093, -0.0266, 0.0, -0.4093, -0.0244, 0.0, -0.407, -0.0271, 0.0, -0.4071, -0.0244, 0.0, -0.4093, -0.0222, 0.0, -0.4093, -0.0201, 0.0, -0.4071, -0.0222,
0.0, -0.4071, -0.0201, -0.0017, -0.407, -0.0271, -0.0039, -0.407, -0.0271, 0.0, -0.4052, -0.0352, 0.0, -0.4035, -0.0342, 0.0, -0.4025, -0.0335, -0.0017, -0.4049, -0.0351, -0.0039, -0.4049, -0.0351, -0.0017, -0.4037, -0.0343, -0.0017, -0.4026, -0.0336, -0.0039, -0.4037, -0.0343, -0.0039, -0.4026, -0.0336, 0.0, -0.4006, -0.0353, 0.0, -0.4008, -0.0325, 0.0, -0.3984, -0.0353, 0.0, -0.3984, -0.0331, 0.0, -0.3997, -0.0318, 0.0, -0.3987, -0.0312, -0.0017, -0.4009, -0.0325, -0.0039, -0.4009, -0.0325, -0.0017, -0.3998, -0.0319, -0.0039, -0.3998, -0.0319, -0.0017, -0.3984, -0.031, -0.0039, -0.3984, -0.031, 0.0, -0.4054, -0.0261, 0.0, -0.4043, -0.0254, 0.0, -0.4025, -0.0242, 0.0, -0.4049, -0.0222, 0.0, -0.4049, -0.0201, 0.0, -0.4028, -0.0222,
0.0, -0.4028, -0.0201, -0.0017, -0.4053, -0.026, -0.0017, -0.4042, -0.0253, -0.0039, -0.4053, -0.026, -0.0039, -0.4042, -0.0253, -0.0017, -0.4028, -0.0244, -0.0039, -0.4028, -0.0244, 0.0, -0.4015, -0.0235, 0.0, -0.4004, -0.0228, 0.0, -0.4006, -0.0201, 0.0, -0.3988, -0.0218, 0.0, -0.3978, -0.0211, -0.0017, -0.4014, -0.0235, -0.0039, -0.4014, -0.0235, -0.0017, -0.4003, -0.0228, -0.0039, -0.4003, -0.0228, -0.0017, -0.3987, -0.0217, -0.0017, -0.3976, -0.021, -0.0039, -0.3987, -0.0217, -0.0039, -0.3976, -0.021, 0.0, -0.4136, -0.0179, 0.0, -0.4136, -0.0157, 0.0, -0.4115, -0.0179, 0.0, -0.4115, -0.0157, 0.0, -0.4136, -0.0136, 0.0, -0.4143, -0.0103, 0.0, -0.4115, -0.0136, 0.0, -0.4115, -0.0114, -0.0017, -0.4145, -0.0104, -0.0039, -0.4145, -0.0104,
0.0, -0.4093, -0.0179, 0.0, -0.4093, -0.0157, 0.0, -0.4071, -0.0179, 0.0, -0.4071, -0.0157, 0.0, -0.4093, -0.0136, 0.0, -0.4093, -0.0114, 0.0, -0.4071, -0.0136, 0.0, -0.4071, -0.0114, 0.0, -0.4133, -0.0097, 0.0, -0.4118, -0.0089, 0.0, -0.4142, -0.002, -0.0017, -0.4134, -0.0098, -0.0039, -0.4134, -0.0098, -0.0017, -0.4115, -0.0087, -0.0039, -0.4115, -0.0087, -0.0017, -0.4141, -0.0019, -0.0039, -0.4141, -0.0019, 0.0, -0.41, -0.008, 0.0, -0.4091, -0.0075, 0.0, -0.4071, -0.0092, 0.0, -0.4073, -0.0065, 0.0, -0.4062, -0.0058, -0.0017, -0.4103, -0.0081, -0.0017, -0.4093, -0.0076, -0.0039, -0.4103, -0.0081, -0.0039, -0.4093, -0.0076, -0.0017, -0.4073, -0.0065, -0.0039, -0.4073, -0.0065, -0.0017, -0.4062, -0.0058, -0.0039, -0.4062, -0.0058,
0.0, -0.4049, -0.0179, 0.0, -0.4049, -0.0157, 0.0, -0.4028, -0.0179, 0.0, -0.4028, -0.0157, 0.0, -0.4049, -0.0136, 0.0, -0.4049, -0.0114, 0.0, -0.4028, -0.0136, 0.0, -0.4028, -0.0114, 0.0, -0.4006, -0.0179, 0.0, -0.4006, -0.0157, 0.0, -0.3984, -0.0179, 0.0, -0.3984, -0.0157, 0.0, -0.4006, -0.0136, 0.0, -0.4006, -0.0114, 0.0, -0.3984, -0.0136, 0.0, -0.3984, -0.0114, 0.0, -0.4049, -0.0092, 0.0, -0.4049, -0.0071, 0.0, -0.4028, -0.0092, 0.0, -0.4028, -0.0071, 0.0, -0.4051, -0.0052, 0.0, -0.4032, -0.0041, 0.0, -0.4021, -0.0035, -0.0017, -0.4049, -0.0051, -0.0039, -0.4049, -0.0051, -0.0017, -0.4032, -0.0041, -0.0017, -0.4022, -0.0035, -0.0039, -0.4032, -0.0041, -0.0039, -0.4022, -0.0035, 0.0, -0.4006, -0.0092,
0.0, -0.4006, -0.0071, 0.0, -0.3984, -0.0092, 0.0, -0.3984, -0.0071, 0.0, -0.4006, -0.0049, 0.0, -0.4009, -0.0028, 0.0, -0.3984, -0.0049, 0.0, -0.399, -0.0017, -0.0017, -0.4006, -0.0026, -0.0039, -0.4006, -0.0026, -0.0017, -0.3992, -0.0018, -0.0039, -0.3992, -0.0018, 0.0, -0.3963, -0.0353, 0.0, -0.3963, -0.0331, 0.0, -0.3941, -0.0353, 0.0, -0.3941, -0.0331, 0.0, -0.3968, -0.03, 0.0, -0.3958, -0.0293, 0.0, -0.3941, -0.0309, 0.0, -0.3942, -0.0284, -0.0017, -0.397, -0.0301, -0.0017, -0.3959, -0.0294, -0.0039, -0.397, -0.0301, -0.0039, -0.3959, -0.0294, -0.0017, -0.3941, -0.0283, -0.0039, -0.3941, -0.0283, 0.0, -0.3919, -0.0353, 0.0, -0.3919, -0.0331, 0.0, -0.3897, -0.0353, 0.0, -0.3897, -0.0331, 0.0, -0.3919, -0.0309,
0.0, -0.3919, -0.0288, 0.0, -0.3897, -0.0309, 0.0, -0.3897, -0.0288, 0.0, -0.393, -0.0276, 0.0, -0.396, -0.0199, 0.0, -0.395, -0.0192, -0.0017, -0.393, -0.0277, -0.0039, -0.393, -0.0277, -0.0017, -0.3963, -0.0201, -0.0039, -0.3963, -0.0201, -0.0017, -0.3949, -0.0192, -0.0039, -0.3949, -0.0192, 0.0, -0.3921, -0.027, 0.0, -0.3902, -0.0258, 0.0, -0.3891, -0.0251, -0.0017, -0.3919, -0.0269, -0.0039, -0.3919, -0.0269, -0.0017, -0.3903, -0.0259, -0.0017, -0.3892, -0.0251, -0.0039, -0.3903, -0.0259, -0.0039, -0.3892, -0.0251, 0.0, -0.3876, -0.0353, 0.0, -0.3876, -0.0331, 0.0, -0.3854, -0.0353, 0.0, -0.3854, -0.0331, 0.0, -0.3876, -0.0309, 0.0, -0.3876, -0.0288, 0.0, -0.3854, -0.0309, 0.0, -0.3854, -0.0288, 0.0, -0.3832, -0.0353,
0.0, -0.3832, -0.0331, 0.0, -0.3811, -0.0353, 0.0, -0.3811, -0.0331, 0.0, -0.3832, -0.0309, 0.0, -0.3832, -0.0288, 0.0, -0.3811, -0.0309, 0.0, -0.3811, -0.0288, 0.0, -0.3876, -0.0266, 0.0, -0.3879, -0.0243, 0.0, -0.3854, -0.0266, 0.0, -0.3863, -0.0232, 0.0, -0.3853, -0.0225, -0.0017, -0.3876, -0.0241, -0.0039, -0.3876, -0.0241, -0.0017, -0.3865, -0.0233, -0.0039, -0.3865, -0.0233, -0.0017, -0.3854, -0.0226, -0.0039, -0.3854, -0.0226, 0.0, -0.3832, -0.0266, 0.0, -0.3832, -0.0244, 0.0, -0.3811, -0.0266, 0.0, -0.3811, -0.0244, 0.0, -0.3837, -0.0215, 0.0, -0.3827, -0.0207, 0.0, -0.3811, -0.0222, 0.0, -0.3812, -0.0197, -0.0017, -0.3838, -0.0215, -0.0017, -0.3827, -0.0207, -0.0039, -0.3838, -0.0215, -0.0039, -0.3827, -0.0207,
-0.0017, -0.3812, -0.0197, -0.0039, -0.3812, -0.0197, 0.0, -0.3963, -0.0179, 0.0, -0.3963, -0.0157, 0.0, -0.3938, -0.0184, 0.0, -0.3941, -0.0157, 0.0, -0.3963, -0.0136, 0.0, -0.3963, -0.0114, 0.0, -0.3941, -0.0136, 0.0, -0.3941, -0.0114, -0.0017, -0.3938, -0.0184, -0.0039, -0.3938, -0.0184, 0.0, -0.3923, -0.0174, 0.0, -0.3913, -0.0167, 0.0, -0.3895, -0.0156, 0.0, -0.3919, -0.0136, 0.0, -0.3919, -0.0114, 0.0, -0.3897, -0.0136, 0.0, -0.3897, -0.0114, -0.0017, -0.3922, -0.0174, -0.0017, -0.3911, -0.0166, -0.0039, -0.3922, -0.0174, -0.0039, -0.3911, -0.0166, -0.0017, -0.3897, -0.0157, -0.0039, -0.3897, -0.0157, 0.0, -0.3963, -0.0092, 0.0, -0.3963, -0.0071, 0.0, -0.3941, -0.0092, 0.0, -0.3941, -0.0071, 0.0, -0.3963, -0.0049,
0.0, -0.3963, -0.0027, 0.0, -0.3941, -0.0049, 0.0, -0.3941, -0.0027, 0.0, -0.3919, -0.0092, 0.0, -0.3919, -0.0071, 0.0, -0.3897, -0.0092, 0.0, -0.3897, -0.0071, 0.0, -0.3919, -0.0049, 0.0, -0.3919, -0.0027, 0.0, -0.3897, -0.0049, 0.0, -0.3897, -0.0027, 0.0, -0.3884, -0.0148, 0.0, -0.3873, -0.0141, 0.0, -0.3876, -0.0114, 0.0, -0.3859, -0.013, 0.0, -0.3848, -0.0123, -0.0017, -0.3884, -0.0148, -0.0039, -0.3884, -0.0148, -0.0017, -0.3873, -0.0141, -0.0039, -0.3873, -0.0141, -0.0017, -0.3858, -0.013, -0.0017, -0.3847, -0.0122, -0.0039, -0.3858, -0.013, -0.0039, -0.3847, -0.0122, 0.0, -0.3801, -0.0189, 0.0, -0.383, -0.011, 0.0, -0.3821, -0.0104, -0.0017, -0.3801, -0.0189, -0.0039, -0.3801, -0.0189, -0.0017, -0.3832, -0.0112,
-0.0039, -0.3832, -0.0112, -0.0017, -0.3821, -0.0104, -0.0039, -0.3821, -0.0104, 0.0, -0.3876, -0.0092, 0.0, -0.3876, -0.0071, 0.0, -0.3854, -0.0092, 0.0, -0.3854, -0.0071, 0.0, -0.3876, -0.0049, 0.0, -0.3876, -0.0027, 0.0, -0.3854, -0.0049, 0.0, -0.3854, -0.0027, 0.0, -0.3832, -0.0092, 0.0, -0.3832, -0.0071, 0.0, -0.381, -0.0096, 0.0, -0.3811, -0.0071, 0.0, -0.3832, -0.0049, 0.0, -0.3832, -0.0027, 0.0, -0.3811, -0.0049, 0.0, -0.3811, -0.0027, -0.0017, -0.381, -0.0096, -0.0039, -0.381, -0.0096, 0.0, -0.3796, -0.0432, 0.0, -0.3785, -0.0423, 0.0, -0.3772, -0.0413, 0.0, -0.3789, -0.0396, 0.0, -0.3789, -0.0374, 0.0, -0.3761, -0.0404, 0.0, -0.3767, -0.0374, -0.0017, -0.3796, -0.0432, -0.0017, -0.3785, -0.0423,
-0.0039, -0.3796, -0.0432, -0.0039, -0.3785, -0.0423, -0.0017, -0.3771, -0.0412, -0.0039, -0.3771, -0.0412, -0.0017, -0.376, -0.0404, -0.0039, -0.376, -0.0404, 0.0, -0.3747, -0.0394, 0.0, -0.3737, -0.0385, 0.0, -0.3721, -0.0373, -0.0017, -0.3746, -0.0393, -0.0017, -0.3736, -0.0385, -0.0039, -0.3746, -0.0393, -0.0039, -0.3736, -0.0385, -0.0017, -0.3724, -0.0376, -0.0039, -0.3724, -0.0376, 0.0, -0.3711, -0.0365, -0.0017, -0.3711, -0.0365, -0.0039, -0.3711, -0.0365, 0.0, -0.3789, -0.0353, 0.0, -0.3789, -0.0331, 0.0, -0.3767, -0.0353, 0.0, -0.3767, -0.0331, 0.0, -0.3789, -0.0309, 0.0, -0.3789, -0.0288, 0.0, -0.3767, -0.0309, 0.0, -0.3767, -0.0288, 0.0, -0.3746, -0.0353, 0.0, -0.3746, -0.0331, 0.0, -0.3724, -0.0353, 0.0, -0.3724, -0.0331,
0.0, -0.3746, -0.0309, 0.0, -0.3746, -0.0288, 0.0, -0.3724, -0.0309, 0.0, -0.3724, -0.0288, 0.0, -0.3789, -0.0266, 0.0, -0.3789, -0.0244, 0.0, -0.3767, -0.0266, 0.0, -0.3767, -0.0244, 0.0, -0.3789, -0.0222, 0.0, -0.3789, -0.0201, 0.0, -0.3767, -0.0222, 0.0, -0.3767, -0.0201, 0.0, -0.3746, -0.0266, 0.0, -0.3746, -0.0244, 0.0, -0.3724, -0.0266, 0.0, -0.3724, -0.0244, 0.0, -0.3746, -0.0222, 0.0, -0.3746, -0.0201, 0.0, -0.3724, -0.0222, 0.0, -0.3724, -0.0201, 0.0, -0.37, -0.0357, 0.0, -0.3702, -0.0331, 0.0, -0.3686, -0.0346, 0.0, -0.3676, -0.0338, 0.0, -0.3702, -0.0309, 0.0, -0.3702, -0.0288, 0.0, -0.368, -0.0309, 0.0, -0.368, -0.0288, -0.0017, -0.37, -0.0357, -0.0039, -0.37, -0.0357,
-0.0017, -0.3686, -0.0346, -0.0017, -0.3675, -0.0337, -0.0039, -0.3686, -0.0346, -0.0039, -0.3675, -0.0337, 0.0, -0.3662, -0.0327, 0.0, -0.3651, -0.0318, 0.0, -0.3659, -0.0288, 0.0, -0.3635, -0.0305, 0.0, -0.3637, -0.0288, -0.0017, -0.3662, -0.0327, -0.0039, -0.3662, -0.0327, -0.0017, -0.3651, -0.0318, -0.0039, -0.3651, -0.0318, -0.0017, -0.3637, -0.0307, -0.0039, -0.3637, -0.0307, 0.0, -0.3702, -0.0266, 0.0, -0.3702, -0.0244, 0.0, -0.368, -0.0266, 0.0, -0.368, -0.0244, 0.0, -0.3702, -0.0222, 0.0, -0.3702, -0.0201, 0.0, -0.368, -0.0222, 0.0, -0.368, -0.0201, 0.0, -0.3659, -0.0266, 0.0, -0.3659, -0.0244, 0.0, -0.3637, -0.0266, 0.0, -0.3637, -0.0244, 0.0, -0.3659, -0.0222, 0.0, -0.3659, -0.0201, 0.0, -0.3637, -0.0222,
0.0, -0.3637, -0.0201, 0.0, -0.3791, -0.0182, 0.0, -0.3774, -0.017, 0.0, -0.3763, -0.0162, -0.0017, -0.3789, -0.018, -0.0039, -0.3789, -0.018, -0.0017, -0.3775, -0.017, -0.0017, -0.3764, -0.0163, -0.0039, -0.3775, -0.017, -0.0039, -0.3764, -0.0163, 0.0, -0.3746, -0.0179, 0.0, -0.3749, -0.0152, 0.0, -0.3724, -0.0179, 0.0, -0.3724, -0.0157, 0.0, -0.3738, -0.0144, 0.0, -0.3727, -0.0136, -0.0017, -0.3749, -0.0152, -0.0039, -0.3749, -0.0152, -0.0017, -0.3738, -0.0144, -0.0039, -0.3738, -0.0144, -0.0017, -0.3724, -0.0134, -0.0039, -0.3724, -0.0134, 0.0, -0.3795, -0.0086, 0.0, -0.3784, -0.0078, 0.0, -0.3769, -0.0068, 0.0, -0.3789, -0.0049, 0.0, -0.3789, -0.0027, 0.0, -0.3759, -0.006, 0.0, -0.3767, -0.0027, -0.0017, -0.3794, -0.0085,
-0.0017, -0.3783, -0.0078, -0.0039, -0.3794, -0.0085, -0.0039, -0.3783, -0.0078, -0.0017, -0.3768, -0.0067, -0.0039, -0.3768, -0.0067, -0.0017, -0.3757, -0.0059, -0.0039, -0.3757, -0.0059, 0.0, -0.3742, -0.0049, 0.0, -0.3746, -0.0027, 0.0, -0.3731, -0.0041, 0.0, -0.372, -0.0033, -0.0017, -0.3746, -0.0051, -0.0039, -0.3746, -0.0051, -0.0017, -0.3731, -0.0041, -0.0017, -0.372, -0.0033, -0.0039, -0.3731, -0.0041, -0.0039, -0.372, -0.0033, 0.0, -0.3702, -0.0179, 0.0, -0.3702, -0.0157, 0.0, -0.368, -0.0179, 0.0, -0.368, -0.0157, 0.0, -0.3711, -0.0124, 0.0, -0.37, -0.0117, 0.0, -0.368, -0.0136, 0.0, -0.3685, -0.0106, -0.0017, -0.3712, -0.0125, -0.0017, -0.3701, -0.0118, -0.0039, -0.3712, -0.0125, -0.0039, -0.3701, -0.0118, -0.0017, -0.3686, -0.0107,
-0.0039, -0.3686, -0.0107, 0.0, -0.3659, -0.0179, 0.0, -0.3659, -0.0157, 0.0, -0.3637, -0.0179, 0.0, -0.3637, -0.0157, 0.0, -0.3659, -0.0136, 0.0, -0.3659, -0.0114, 0.0, -0.3637, -0.0136, 0.0, -0.3637, -0.0114, 0.0, -0.3675, -0.0099, 0.0, -0.3706, -0.0023, -0.0017, -0.3675, -0.0099, -0.0039, -0.3675, -0.0099, -0.0017, -0.3704, -0.0022, -0.0039, -0.3704, -0.0022, 0.0, -0.366, -0.0088, 0.0, -0.3649, -0.008, 0.0, -0.3637, -0.0092, 0.0, -0.3639, -0.0074, -0.0017, -0.366, -0.0088, -0.0017, -0.3649, -0.0081, -0.0039, -0.366, -0.0088, -0.0039, -0.3649, -0.0081, -0.0017, -0.3637, -0.0072, -0.0039, -0.3637, -0.0072, 0.0, -0.3626, -0.0298, 0.0, -0.3615, -0.029, 0.0, -0.3602, -0.0279, -0.0017, -0.3626, -0.0298, -0.0017, -0.3615, -0.029,
-0.0039, -0.3626, -0.0298, -0.0039, -0.3615, -0.029, -0.0017, -0.3601, -0.0279, -0.0039, -0.3601, -0.0279, 0.0, -0.3615, -0.0266, 0.0, -0.3615, -0.0244, 0.0, -0.3591, -0.0271, 0.0, -0.3594, -0.0244, 0.0, -0.3615, -0.0222, 0.0, -0.3615, -0.0201, 0.0, -0.3594, -0.0222, 0.0, -0.3594, -0.0201, -0.0017, -0.3591, -0.0271, -0.0039, -0.3591, -0.0271, 0.0, -0.3577, -0.026, 0.0, -0.3566, -0.0252, 0.0, -0.3553, -0.0241, 0.0, -0.3572, -0.0222, 0.0, -0.3572, -0.0201, 0.0, -0.3543, -0.0231, 0.0, -0.355, -0.0201, -0.0017, -0.3577, -0.026, -0.0017, -0.3566, -0.0251, -0.0039, -0.3577, -0.026, -0.0039, -0.3566, -0.0251, -0.0017, -0.3553, -0.024, -0.0039, -0.3553, -0.024, -0.0017, -0.3542, -0.0231, -0.0039, -0.3542, -0.0231, 0.0, -0.353, -0.0221,
0.0, -0.3519, -0.0212, 0.0, -0.3504, -0.0199, -0.0017, -0.3529, -0.022, -0.0017, -0.3519, -0.0211, -0.0039, -0.3529, -0.022, -0.0039, -0.3519, -0.0211, -0.0017, -0.3507, -0.0201, -0.0039, -0.3507, -0.0201, 0.0, -0.3495, -0.0191, -0.0017, -0.3494, -0.0191, -0.0039, -0.3494, -0.0191, 0.0, -0.3615, -0.0179, 0.0, -0.3615, -0.0157, 0.0, -0.3594, -0.0179, 0.0, -0.3594, -0.0157, 0.0, -0.3615, -0.0136, 0.0, -0.3615, -0.0114, 0.0, -0.3594, -0.0136, 0.0, -0.3594, -0.0114, 0.0, -0.3572, -0.0179, 0.0, -0.3572, -0.0157, 0.0, -0.355, -0.0179, 0.0, -0.355, -0.0157, 0.0, -0.3572, -0.0136, 0.0, -0.3572, -0.0114, 0.0, -0.355, -0.0136, 0.0, -0.355, -0.0114, 0.0, -0.3615, -0.0092, 0.0, -0.3622, -0.0061, 0.0, -0.3594, -0.0092,
0.0, -0.3594, -0.0071, 0.0, -0.3611, -0.0054, 0.0, -0.3596, -0.0044, 0.0, -0.3585, -0.0036, -0.0017, -0.3623, -0.0062, -0.0039, -0.3623, -0.0062, -0.0017, -0.3612, -0.0055, -0.0039, -0.3612, -0.0055, -0.0017, -0.3596, -0.0044, -0.0017, -0.3585, -0.0036, -0.0039, -0.3596, -0.0044, -0.0039, -0.3585, -0.0036, 0.0, -0.3572, -0.0092, 0.0, -0.3572, -0.0071, 0.0, -0.355, -0.0092, 0.0, -0.355, -0.0071, 0.0, -0.3572, -0.0049, 0.0, -0.3574, -0.0029, 0.0, -0.355, -0.0049, 0.0, -0.3557, -0.0017, -0.0017, -0.3572, -0.0027, -0.0039, -0.3572, -0.0027, -0.0017, -0.3559, -0.0018, -0.0039, -0.3559, -0.0018, 0.0, -0.3529, -0.0179, 0.0, -0.3529, -0.0157, 0.0, -0.3507, -0.0179, 0.0, -0.3507, -0.0157, 0.0, -0.3529, -0.0136, 0.0, -0.3529, -0.0114,
0.0, -0.3507, -0.0136, 0.0, -0.3507, -0.0114, 0.0, -0.3483, -0.0182, 0.0, -0.3485, -0.0157, 0.0, -0.347, -0.0172, 0.0, -0.346, -0.0163, 0.0, -0.3485, -0.0136, 0.0, -0.3485, -0.0114, 0.0, -0.3463, -0.0136, 0.0, -0.3463, -0.0114, -0.0017, -0.3484, -0.0182, -0.0039, -0.3484, -0.0182, -0.0017, -0.347, -0.0172, -0.0017, -0.3459, -0.0163, -0.0039, -0.347, -0.0172, -0.0039, -0.3459, -0.0163, 0.0, -0.3529, -0.0092, 0.0, -0.3529, -0.0071, 0.0, -0.3507, -0.0092, 0.0, -0.3507, -0.0071, 0.0, -0.3529, -0.0049, 0.0, -0.3529, -0.0027, 0.0, -0.3507, -0.0049, 0.0, -0.3507, -0.0027, 0.0, -0.3485, -0.0092, 0.0, -0.3485, -0.0071, 0.0, -0.3463, -0.0092, 0.0, -0.3463, -0.0071, 0.0, -0.3485, -0.0049, 0.0, -0.3485, -0.0027,
0.0, -0.3463, -0.0049, 0.0, -0.3463, -0.0027, 0.0, -0.3446, -0.1019, 0.0, -0.3437, -0.1015, 0.0, -0.3442, -0.0982, 0.0, -0.3417, -0.1007, 0.0, -0.342, -0.0982, -0.0017, -0.3444, -0.1018, -0.0039, -0.3444, -0.1018, -0.0017, -0.3433, -0.1014, -0.0039, -0.3433, -0.1014, -0.0017, -0.342, -0.1008, -0.0039, -0.342, -0.1008, 0.0, -0.3397, -0.0999, 0.0, -0.3398, -0.0982, 0.0, -0.3385, -0.0994, 0.0, -0.3375, -0.099, -0.0017, -0.3398, -0.1, -0.0039, -0.3398, -0.1, -0.0017, -0.3385, -0.0994, -0.0017, -0.3374, -0.099, -0.0039, -0.3385, -0.0994, -0.0039, -0.3374, -0.099, 0.0, -0.3442, -0.096, 0.0, -0.3442, -0.0939, 0.0, -0.342, -0.096, 0.0, -0.342, -0.0939, 0.0, -0.3442, -0.0917, 0.0, -0.3442, -0.0895, 0.0, -0.342, -0.0917,
0.0, -0.342, -0.0895, 0.0, -0.3398, -0.096, 0.0, -0.3398, -0.0939, 0.0, -0.3377, -0.096, 0.0, -0.3377, -0.0939, 0.0, -0.3398, -0.0917, 0.0, -0.3398, -0.0895, 0.0, -0.3377, -0.0917, 0.0, -0.3377, -0.0895, 0.0, -0.3353, -0.0981, 0.0, -0.334, -0.0974, -0.0017, -0.3355, -0.0981, -0.0039, -0.3355, -0.0981, -0.0017, -0.3339, -0.0974, -0.0039, -0.3339, -0.0974, 0.0, -0.3355, -0.096, 0.0, -0.3355, -0.0939, 0.0, -0.3329, -0.0969, 0.0, -0.3333, -0.0939, 0.0, -0.3355, -0.0917, 0.0, -0.3355, -0.0895, 0.0, -0.3333, -0.0917, 0.0, -0.3333, -0.0895, -0.0017, -0.3328, -0.0968, -0.0039, -0.3328, -0.0968, 0.0, -0.3309, -0.0958, 0.0, -0.3312, -0.0939, 0.0, -0.3298, -0.0952, 0.0, -0.3287, -0.0946, 0.0, -0.3312, -0.0917,
0.0, -0.3312, -0.0895, 0.0, -0.329, -0.0917, 0.0, -0.329, -0.0895, -0.0017, -0.3311, -0.096, -0.0039, -0.3311, -0.096, -0.0017, -0.3297, -0.0952, -0.0017, -0.3286, -0.0945, -0.0039, -0.3297, -0.0952, -0.0039, -0.3286, -0.0945, 0.0, -0.3443, -0.0879, 0.0, -0.3423, -0.087, -0.0017, -0.3442, -0.0879, -0.0039, -0.3442, -0.0879, -0.0017, -0.342, -0.0869, -0.0039, -0.342, -0.0869, 0.0, -0.3404, -0.0862, 0.0, -0.3394, -0.0858, 0.0, -0.3377, -0.0874, 0.0, -0.3379, -0.0851, -0.0017, -0.3407, -0.0864, -0.0017, -0.3396, -0.0859, -0.0039, -0.3407, -0.0864, -0.0039, -0.3396, -0.0859, -0.0017, -0.3377, -0.085, -0.0039, -0.3377, -0.085, 0.0, -0.3355, -0.0874, 0.0, -0.336, -0.0842, 0.0, -0.3333, -0.0874, 0.0, -0.3333, -0.0852, 0.0, -0.3349, -0.0837,
0.0, -0.3336, -0.083, -0.0017, -0.3362, -0.0843, -0.0039, -0.3362, -0.0843, -0.0017, -0.3351, -0.0838, -0.0039, -0.3351, -0.0838, -0.0017, -0.3333, -0.0829, -0.0039, -0.3333, -0.0829, 0.0, -0.3312, -0.0874, 0.0, -0.3312, -0.0852, 0.0, -0.329, -0.0874, 0.0, -0.329, -0.0852, 0.0, -0.3317, -0.082, 0.0, -0.3307, -0.0814, 0.0, -0.329, -0.083, 0.0, -0.3291, -0.0804, -0.0017, -0.3319, -0.0821, -0.0017, -0.3308, -0.0815, -0.0039, -0.3319, -0.0821, -0.0039, -0.3308, -0.0815, -0.0017, -0.329, -0.0804, -0.0039, -0.329, -0.0804, 0.0, -0.328, -0.0797, -0.0017, -0.328, -0.0797, -0.0039, -0.328, -0.0797, 0.0, -0.3267, -0.0934, 0.0, -0.3257, -0.0928, 0.0, -0.3268, -0.0917, 0.0, -0.3268, -0.0895, 0.0, -0.3246, -0.0921, 0.0, -0.3246, -0.0895,
-0.0017, -0.3268, -0.0935, -0.0039, -0.3268, -0.0935, -0.0017, -0.3257, -0.0928, -0.0039, -0.3257, -0.0928, -0.0017, -0.3246, -0.0921, -0.0039, -0.3246, -0.0921, 0.0, -0.323, -0.0911, 0.0, -0.3219, -0.0903, 0.0, -0.3205, -0.0893, -0.0017, -0.3229, -0.091, -0.0017, -0.3219, -0.0903, -0.0039, -0.3229, -0.091, -0.0039, -0.3219, -0.0903, -0.0017, -0.3203, -0.0892, -0.0039, -0.3203, -0.0892, 0.0, -0.3268, -0.0874, 0.0, -0.3268, -0.0852, 0.0, -0.3246, -0.0874, 0.0, -0.3246, -0.0852, 0.0, -0.3268, -0.083, 0.0, -0.3268, -0.0808, 0.0, -0.3246, -0.083, 0.0, -0.3246, -0.0808, 0.0, -0.3225, -0.0874, 0.0, -0.3225, -0.0852, 0.0, -0.3194, -0.0885, 0.0, -0.3203, -0.0852, 0.0, -0.3225, -0.083, 0.0, -0.3225, -0.0808, 0.0, -0.3203, -0.083,
0.0, -0.3203, -0.0808, -0.0017, -0.3193, -0.0884, -0.0039, -0.3193, -0.0884, 0.0, -0.327, -0.0791, 0.0, -0.3252, -0.0779, 0.0, -0.3241, -0.0771, -0.0017, -0.3268, -0.079, -0.0039, -0.3268, -0.079, -0.0017, -0.3253, -0.0779, -0.0017, -0.3242, -0.0772, -0.0039, -0.3253, -0.0779, -0.0039, -0.3242, -0.0772, 0.0, -0.3225, -0.0787, 0.0, -0.3226, -0.0761, 0.0, -0.3203, -0.0787, 0.0, -0.3203, -0.0765, 0.0, -0.3215, -0.0753, 0.0, -0.3206, -0.0745, -0.0017, -0.3226, -0.0761, -0.0039, -0.3226, -0.0761, -0.0017, -0.3216, -0.0753, -0.0039, -0.3216, -0.0753, -0.0017, -0.3203, -0.0743, -0.0039, -0.3203, -0.0743, 0.0, -0.3178, -0.0873, 0.0, -0.3181, -0.0852, 0.0, -0.3168, -0.0865, 0.0, -0.3157, -0.0856, 0.0, -0.3181, -0.083, 0.0, -0.3181, -0.0808,
0.0, -0.316, -0.083, 0.0, -0.316, -0.0808, -0.0017, -0.3181, -0.0875, -0.0039, -0.3181, -0.0875, -0.0017, -0.3168, -0.0865, -0.0017, -0.3157, -0.0856, -0.0039, -0.3168, -0.0865, -0.0039, -0.3157, -0.0856, 0.0, -0.3144, -0.0846, 0.0, -0.3133, -0.0837, 0.0, -0.3138, -0.0808, 0.0, -0.312, -0.0826, 0.0, -0.3109, -0.0816, -0.0017, -0.3143, -0.0845, -0.0039, -0.3143, -0.0845, -0.0017, -0.3133, -0.0836, -0.0039, -0.3133, -0.0836, -0.0017, -0.312, -0.0826, -0.0017, -0.3109, -0.0816, -0.0039, -0.312, -0.0826, -0.0039, -0.3109, -0.0816, 0.0, -0.3181, -0.0787, 0.0, -0.3181, -0.0765, 0.0, -0.316, -0.0787, 0.0, -0.316, -0.0765, 0.0, -0.319, -0.0733, 0.0, -0.318, -0.0724, 0.0, -0.316, -0.0743, 0.0, -0.3167, -0.0713, -0.0017, -0.3191, -0.0733,
-0.0017, -0.318, -0.0724, -0.0039, -0.3191, -0.0733, -0.0039, -0.318, -0.0724, -0.0017, -0.3167, -0.0713, -0.0039, -0.3167, -0.0713, 0.0, -0.3138, -0.0787, 0.0, -0.3138, -0.0765, 0.0, -0.3116, -0.0787, 0.0, -0.3116, -0.0765, 0.0, -0.3138, -0.0743, 0.0, -0.3138, -0.0722, 0.0, -0.3116, -0.0743, 0.0, -0.3116, -0.0722, 0.0, -0.3097, -0.0806, -0.0017, -0.3097, -0.0805, -0.0039, -0.3097, -0.0805, 0.0, -0.3086, -0.0795, 0.0, -0.3095, -0.0765, 0.0, -0.3075, -0.0784, 0.0, -0.3065, -0.0773, 0.0, -0.3095, -0.0743, 0.0, -0.3095, -0.0722, 0.0, -0.3073, -0.0743, 0.0, -0.3073, -0.0722, -0.0017, -0.3086, -0.0795, -0.0039, -0.3086, -0.0795, -0.0017, -0.3075, -0.0784, -0.0017, -0.3065, -0.0773, -0.0039, -0.3075, -0.0784, -0.0039, -0.3065, -0.0773,
0.0, -0.3054, -0.0762, 0.0, -0.3043, -0.075, 0.0, -0.3051, -0.0722, 0.0, -0.3034, -0.0739, 0.0, -0.3023, -0.0726, -0.0017, -0.3054, -0.0762, -0.0039, -0.3054, -0.0762, -0.0017, -0.3044, -0.075, -0.0039, -0.3044, -0.075, -0.0017, -0.3034, -0.0739, -0.0017, -0.3024, -0.0726, -0.0039, -0.3034, -0.0739, -0.0039, -0.3024, -0.0726, 0.0, -0.3015, -0.0715, -0.0017, -0.3015, -0.0715, -0.0039, -0.3015, -0.0715, 0.0, -0.3156, -0.0704, -0.0017, -0.3156, -0.0704, -0.0039, -0.3156, -0.0704, 0.0, -0.3144, -0.0693, 0.0, -0.3133, -0.0683, 0.0, -0.3116, -0.07, 0.0, -0.3122, -0.0673, 0.0, -0.3112, -0.0661, -0.0017, -0.3144, -0.0693, -0.0017, -0.3133, -0.0683, -0.0039, -0.3144, -0.0693, -0.0039, -0.3133, -0.0683, -0.0017, -0.3122, -0.0672, -0.0039, -0.3122, -0.0672,
-0.0017, -0.3111, -0.0661, -0.0039, -0.3111, -0.0661, 0.0, -0.3446, -0.0152, 0.0, -0.3435, -0.0144, 0.0, -0.3442, -0.0114, 0.0, -0.3422, -0.0133, 0.0, -0.3411, -0.0125, -0.0017, -0.3446, -0.0152, -0.0039, -0.3446, -0.0152, -0.0017, -0.3435, -0.0143, -0.0039, -0.3435, -0.0143, -0.0017, -0.3421, -0.0132, -0.0017, -0.341, -0.0124, -0.0039, -0.3421, -0.0132, -0.0039, -0.341, -0.0124, 0.0, -0.3396, -0.0113, 0.0, -0.3385, -0.0105, -0.0017, -0.3398, -0.0115, -0.0039, -0.3398, -0.0115, -0.0017, -0.3385, -0.0105, -0.0039, -0.3385, -0.0105, 0.0, -0.3442, -0.0092, 0.0, -0.3442, -0.0071, 0.0, -0.342, -0.0092, 0.0, -0.342, -0.0071, 0.0, -0.3442, -0.0049, 0.0, -0.3442, -0.0027, 0.0, -0.342, -0.0049, 0.0, -0.342, -0.0027, 0.0, -0.3398, -0.0092,
0.0, -0.3398, -0.0071, 0.0, -0.3374, -0.0097, 0.0, -0.3377, -0.0071, 0.0, -0.3398, -0.0049, 0.0, -0.3398, -0.0027, 0.0, -0.3377, -0.0049, 0.0, -0.3377, -0.0027, -0.0017, -0.3374, -0.0097, -0.0039, -0.3374, -0.0097, 0.0, -0.3361, -0.0086, 0.0, -0.335, -0.0078, 0.0, -0.3336, -0.0067, 0.0, -0.3355, -0.0049, 0.0, -0.3355, -0.0027, 0.0, -0.3325, -0.0059, 0.0, -0.3333, -0.0027, -0.0017, -0.336, -0.0086, -0.0017, -0.3349, -0.0078, -0.0039, -0.336, -0.0086, -0.0039, -0.3349, -0.0078, -0.0017, -0.3335, -0.0067, -0.0039, -0.3335, -0.0067, -0.0017, -0.3324, -0.0059, -0.0039, -0.3324, -0.0059, 0.0, -0.3309, -0.0047, 0.0, -0.3312, -0.0027, 0.0, -0.3298, -0.004, 0.0, -0.3287, -0.0032, -0.0017, -0.3312, -0.0049, -0.0039, -0.3312, -0.0049,
-0.0017, -0.3298, -0.004, -0.0017, -0.3287, -0.0032, -0.0039, -0.3298, -0.004, -0.0039, -0.3287, -0.0032, 0.0, -0.3273, -0.0022, -0.0017, -0.3272, -0.0021, -0.0039, -0.3272, -0.0021, 0.0, -0.3095, -0.07, 0.0, -0.3095, -0.0678, 0.0, -0.3073, -0.07, 0.0, -0.3073, -0.0678, 0.0, -0.3102, -0.065, 0.0, -0.3092, -0.0637, 0.0, -0.3073, -0.0656, 0.0, -0.3083, -0.0626, -0.0017, -0.3102, -0.065, -0.0017, -0.3091, -0.0636, -0.0039, -0.3102, -0.065, -0.0039, -0.3091, -0.0636, -0.0017, -0.3082, -0.0626, -0.0039, -0.3082, -0.0626, 0.0, -0.3051, -0.07, 0.0, -0.3051, -0.0678, 0.0, -0.3029, -0.07, 0.0, -0.3029, -0.0678, 0.0, -0.3051, -0.0656, 0.0, -0.3051, -0.0635, 0.0, -0.3029, -0.0656, 0.0, -0.3029, -0.0635, 0.0, -0.307, -0.0611,
0.0, -0.3063, -0.0601, -0.0017, -0.3072, -0.0613, -0.0017, -0.3063, -0.0601, -0.0039, -0.3072, -0.0613, -0.0039, -0.3063, -0.0601, 0.0, -0.3051, -0.0613, 0.0, -0.3054, -0.059, 0.0, -0.3029, -0.0613, 0.0, -0.3029, -0.0591, 0.0, -0.3044, -0.0576, 0.0, -0.3036, -0.0565, 0.0, -0.3026, -0.0551, -0.0017, -0.3055, -0.059, -0.0039, -0.3055, -0.059, -0.0017, -0.3044, -0.0576, -0.0039, -0.3044, -0.0576, -0.0017, -0.3036, -0.0565, -0.0017, -0.3025, -0.055, -0.0039, -0.3036, -0.0565, -0.0039, -0.3025, -0.055, 0.0, -0.3004, -0.0702, 0.0, -0.3008, -0.0678, 0.0, -0.2996, -0.069, 0.0, -0.2988, -0.0681, 0.0, -0.3008, -0.0656, 0.0, -0.3008, -0.0635, 0.0, -0.2976, -0.0663, 0.0, -0.2986, -0.0635, -0.0017, -0.3004, -0.0701, -0.0039, -0.3004, -0.0701,
-0.0017, -0.2996, -0.069, -0.0017, -0.2987, -0.0678, -0.0039, -0.2996, -0.069, -0.0039, -0.2987, -0.0678, -0.0017, -0.2977, -0.0665, -0.0039, -0.2977, -0.0665, 0.0, -0.2969, -0.0653, 0.0, -0.2959, -0.0637, 0.0, -0.2952, -0.0626, -0.0017, -0.297, -0.0654, -0.0017, -0.2959, -0.0637, -0.0039, -0.297, -0.0654, -0.0039, -0.2959, -0.0637, -0.0017, -0.2952, -0.0627, -0.0039, -0.2952, -0.0627, 0.0, -0.3008, -0.0613, 0.0, -0.3008, -0.0591, 0.0, -0.2986, -0.0613, 0.0, -0.2986, -0.0591, 0.0, -0.3008, -0.057, 0.0, -0.3019, -0.054, 0.0, -0.2986, -0.057, 0.0, -0.2986, -0.0548, -0.0017, -0.3017, -0.0539, -0.0039, -0.3017, -0.0539, 0.0, -0.2964, -0.0613, 0.0, -0.2964, -0.0591, 0.0, -0.2945, -0.0615, 0.0, -0.2933, -0.0597, 0.0, -0.2964, -0.057,
0.0, -0.2964, -0.0548, 0.0, -0.2943, -0.057, 0.0, -0.2943, -0.0548, -0.0017, -0.2943, -0.0613, -0.0017, -0.2934, -0.0599, -0.0039, -0.2943, -0.0613, -0.0039, -0.2934, -0.0599, 0.0, -0.3007, -0.0524, 0.0, -0.3, -0.0512, 0.0, -0.2986, -0.0526, 0.0, -0.2993, -0.0501, 0.0, -0.2982, -0.0481, 0.0, -0.2976, -0.047, -0.0017, -0.3009, -0.0526, -0.0017, -0.3, -0.0511, -0.0039, -0.3009, -0.0526, -0.0039, -0.3, -0.0511, -0.0017, -0.2993, -0.05, -0.0039, -0.2993, -0.05, -0.0017, -0.2983, -0.0483, -0.0017, -0.2976, -0.047, -0.0039, -0.2983, -0.0483, -0.0039, -0.2976, -0.047, 0.0, -0.2964, -0.0526, 0.0, -0.2964, -0.0505, 0.0, -0.2943, -0.0526, 0.0, -0.2943, -0.0505, 0.0, -0.2964, -0.0483, 0.0, -0.2971, -0.0459, 0.0, -0.2943, -0.0483,
0.0, -0.2943, -0.0461, -0.0017, -0.2971, -0.0459, -0.0039, -0.2971, -0.0459, 0.0, -0.2961, -0.0438, 0.0, -0.2956, -0.0426, 0.0, -0.2943, -0.0439, 0.0, -0.2951, -0.0415, 0.0, -0.2943, -0.0394, 0.0, -0.2936, -0.0378, -0.0017, -0.2962, -0.044, -0.0017, -0.2955, -0.0425, -0.0039, -0.2962, -0.044, -0.0039, -0.2955, -0.0425, -0.0017, -0.2951, -0.0414, -0.0039, -0.2951, -0.0414, -0.0017, -0.2943, -0.0396, -0.0017, -0.2935, -0.0375, -0.0039, -0.2943, -0.0396, -0.0039, -0.2935, -0.0375, 0.0, -0.2927, -0.0587, 0.0, -0.2917, -0.057, 0.0, -0.2921, -0.0548, 0.0, -0.291, -0.0559, 0.0, -0.2904, -0.0549, -0.0017, -0.2927, -0.0588, -0.0039, -0.2927, -0.0588, -0.0017, -0.2916, -0.057, -0.0039, -0.2916, -0.057, -0.0017, -0.291, -0.0559, -0.0017, -0.2904, -0.0548,
-0.0039, -0.291, -0.0559, -0.0039, -0.2904, -0.0548, 0.0, -0.2921, -0.0526, 0.0, -0.2921, -0.0505, 0.0, -0.2893, -0.0527, 0.0, -0.2899, -0.0505, 0.0, -0.2921, -0.0483, 0.0, -0.2921, -0.0461, 0.0, -0.2899, -0.0483, 0.0, -0.2899, -0.0461, -0.0017, -0.2893, -0.0527, -0.0039, -0.2893, -0.0527, 0.0, -0.2888, -0.0516, 0.0, -0.2884, -0.0506, 0.0, -0.2875, -0.0486, 0.0, -0.2867, -0.0466, 0.0, -0.2862, -0.0457, -0.0017, -0.2888, -0.0516, -0.0017, -0.2883, -0.0505, -0.0039, -0.2888, -0.0516, -0.0039, -0.2883, -0.0505, -0.0017, -0.2874, -0.0483, -0.0017, -0.2868, -0.0469, -0.0039, -0.2874, -0.0483, -0.0039, -0.2868, -0.0469, -0.0017, -0.2863, -0.0458, -0.0039, -0.2863, -0.0458, 0.0, -0.2921, -0.0439, 0.0, -0.2921, -0.0418, 0.0, -0.2899, -0.0439,
0.0, -0.2899, -0.0418, 0.0, -0.2921, -0.0396, 0.0, -0.2933, -0.0369, 0.0, -0.2899, -0.0396, 0.0, -0.2899, -0.0374, -0.0017, -0.2931, -0.0364, -0.0039, -0.2931, -0.0364, 0.0, -0.2878, -0.0439, 0.0, -0.2878, -0.0418, 0.0, -0.2856, -0.0441, 0.0, -0.2848, -0.0419, 0.0, -0.2878, -0.0396, 0.0, -0.2878, -0.0374, 0.0, -0.2856, -0.0396, 0.0, -0.2856, -0.0374, -0.0017, -0.2855, -0.044, -0.0017, -0.2848, -0.0419, -0.0039, -0.2855, -0.044, -0.0039, -0.2848, -0.0419, 0.0, -0.2844, -0.0408, 0.0, -0.2841, -0.0397, 0.0, -0.2835, -0.0376, -0.0017, -0.2845, -0.0408, -0.0039, -0.2845, -0.0408, -0.0017, -0.2841, -0.0396, -0.0017, -0.2834, -0.0374, -0.0039, -0.2841, -0.0396, -0.0039, -0.2834, -0.0374, 0.0, -0.2926, -0.035, 0.0, -0.292, -0.033,
0.0, -0.2899, -0.0353, 0.0, -0.2899, -0.0331, 0.0, -0.2914, -0.0309, 0.0, -0.2911, -0.0297, 0.0, -0.2899, -0.0309, 0.0, -0.2908, -0.0287, -0.0017, -0.2927, -0.0353, -0.0017, -0.292, -0.0331, -0.0039, -0.2927, -0.0353, -0.0039, -0.292, -0.0331, -0.0017, -0.2914, -0.0309, -0.0017, -0.2911, -0.0297, -0.0039, -0.2914, -0.0309, -0.0039, -0.2911, -0.0297, -0.0017, -0.2908, -0.0286, -0.0039, -0.2908, -0.0286, 0.0, -0.2878, -0.0353, 0.0, -0.2878, -0.0331, 0.0, -0.2856, -0.0353, 0.0, -0.2856, -0.0331, 0.0, -0.2878, -0.0309, 0.0, -0.2878, -0.0288, 0.0, -0.2856, -0.0309, 0.0, -0.2856, -0.0288, 0.0, -0.2903, -0.0264, 0.0, -0.2899, -0.0243, 0.0, -0.2894, -0.0222, 0.0, -0.2891, -0.0206, -0.0017, -0.2904, -0.0266, -0.0017, -0.2899, -0.0244,
-0.0039, -0.2904, -0.0266, -0.0039, -0.2899, -0.0244, -0.0017, -0.2894, -0.0222, -0.0017, -0.289, -0.0204, -0.0039, -0.2894, -0.0222, -0.0039, -0.289, -0.0204, 0.0, -0.2878, -0.0266, 0.0, -0.2878, -0.0244, 0.0, -0.2856, -0.0266, 0.0, -0.2856, -0.0244, 0.0, -0.2878, -0.0222, 0.0, -0.2889, -0.0198, 0.0, -0.2856, -0.0222, 0.0, -0.2856, -0.0201, -0.0017, -0.2888, -0.0193, -0.0039, -0.2888, -0.0193, 0.0, -0.2829, -0.0355, 0.0, -0.2823, -0.0334, 0.0, -0.2821, -0.0325, 0.0, -0.2834, -0.0309, 0.0, -0.2834, -0.0288, 0.0, -0.2817, -0.031, 0.0, -0.2812, -0.0289, -0.0017, -0.2828, -0.0353, -0.0017, -0.2824, -0.0338, -0.0039, -0.2828, -0.0353, -0.0039, -0.2824, -0.0338, -0.0017, -0.2821, -0.0328, -0.0039, -0.2821, -0.0328, -0.0017, -0.2816, -0.0309,
-0.0017, -0.2812, -0.0288, -0.0039, -0.2816, -0.0309, -0.0039, -0.2812, -0.0288, 0.0, -0.2834, -0.0266, 0.0, -0.2834, -0.0244, 0.0, -0.2808, -0.0267, 0.0, -0.2804, -0.0246, 0.0, -0.2834, -0.0222, 0.0, -0.2834, -0.0201, 0.0, -0.2801, -0.0225, 0.0, -0.2812, -0.0201, -0.0017, -0.2808, -0.0266, -0.0017, -0.2804, -0.0244, -0.0039, -0.2808, -0.0266, -0.0039, -0.2804, -0.0244, -0.0017, -0.2802, -0.0231, -0.0039, -0.2802, -0.0231, 0.0, -0.28, -0.0218, 0.0, -0.2797, -0.0201, -0.0017, -0.28, -0.022, -0.0017, -0.2797, -0.0201, -0.0039, -0.28, -0.022, -0.0039, -0.2797, -0.0201, 0.0, -0.2885, -0.0178, 0.0, -0.2883, -0.0157, 0.0, -0.2856, -0.0179, 0.0, -0.2856, -0.0157, 0.0, -0.2882, -0.0136, 0.0, -0.2882, -0.0114, 0.0, -0.2856, -0.0136,
0.0, -0.2856, -0.0114, -0.0017, -0.2885, -0.0179, -0.0017, -0.2883, -0.0157, -0.0039, -0.2885, -0.0179, -0.0039, -0.2883, -0.0157, -0.0017, -0.2882, -0.0136, -0.0017, -0.2882, -0.0114, -0.0039, -0.2882, -0.0136, -0.0039, -0.2882, -0.0114, 0.0, -0.2882, -0.0092, 0.0, -0.2882, -0.0071, 0.0, -0.2856, -0.0092, 0.0, -0.2856, -0.0071, 0.0, -0.2882, -0.0049, 0.0, -0.2882, -0.0027, 0.0, -0.2856, -0.0049, 0.0, -0.2856, -0.0027, -0.0017, -0.2882, -0.0092, -0.0017, -0.2882, -0.0071, -0.0039, -0.2882, -0.0092, -0.0039, -0.2882, -0.0071, -0.0017, -0.2882, -0.0049, -0.0017, -0.2882, -0.0027, -0.0039, -0.2882, -0.0049, -0.0039, -0.2882, -0.0027, 0.0, -0.2834, -0.0179, 0.0, -0.2834, -0.0157, 0.0, -0.2812, -0.0179, 0.0, -0.2812, -0.0157, 0.0, -0.2834, -0.0136,
0.0, -0.2834, -0.0114, 0.0, -0.2812, -0.0136, 0.0, -0.2812, -0.0114, 0.0, -0.2795, -0.0179, 0.0, -0.2794, -0.0157, 0.0, -0.2794, -0.0136, 0.0, -0.2794, -0.0114, -0.0017, -0.2795, -0.0179, -0.0017, -0.2794, -0.0157, -0.0039, -0.2795, -0.0179, -0.0039, -0.2794, -0.0157, -0.0017, -0.2794, -0.0136, -0.0017, -0.2794, -0.0114, -0.0039, -0.2794, -0.0136, -0.0039, -0.2794, -0.0114, 0.0, -0.2834, -0.0092, 0.0, -0.2834, -0.0071, 0.0, -0.2812, -0.0092, 0.0, -0.2812, -0.0071, 0.0, -0.2834, -0.0049, 0.0, -0.2834, -0.0027, 0.0, -0.2812, -0.0049, 0.0, -0.2812, -0.0027, 0.0, -0.2794, -0.0092, 0.0, -0.2794, -0.0071, 0.0, -0.2794, -0.0049, 0.0, -0.2794, -0.0027, -0.0017, -0.2794, -0.0092, -0.0017, -0.2794, -0.0071, -0.0039, -0.2794, -0.0092,
-0.0039, -0.2794, -0.0071, -0.0017, -0.2794, -0.0049, -0.0017, -0.2794, -0.0027, -0.0039, -0.2794, -0.0049, -0.0039, -0.2794, -0.0027, 0.0, -0.2196, -0.067, 0.0, -0.2195, -0.0666, -0.0017, -0.2196, -0.067, -0.0039, -0.2196, -0.067, -0.0017, -0.2195, -0.0666, -0.0039, -0.2195, -0.0666, 0.0, -0.2183, -0.067, 0.0, -0.2172, -0.0667, 0.0, -0.219, -0.0655, 0.0, -0.218, -0.0633, 0.0, -0.216, -0.0665, 0.0, -0.2161, -0.0635, -0.0017, -0.2183, -0.067, -0.0039, -0.2183, -0.067, -0.0017, -0.2172, -0.0667, -0.0039, -0.2172, -0.0667, -0.0017, -0.219, -0.0655, -0.0017, -0.2181, -0.0635, -0.0039, -0.219, -0.0655, -0.0039, -0.2181, -0.0635, -0.0017, -0.216, -0.0665, -0.0039, -0.216, -0.0665, 0.0, -0.2174, -0.0622, 0.0, -0.2168, -0.0611, 0.0, -0.2158, -0.0593,
-0.0017, -0.2174, -0.0621, -0.0039, -0.2174, -0.0621, -0.0017, -0.2168, -0.061, -0.0017, -0.2157, -0.0592, -0.0039, -0.2168, -0.061, -0.0039, -0.2157, -0.0592, 0.0, -0.2142, -0.0654, 0.0, -0.2131, -0.0641, 0.0, -0.2124, -0.063, -0.0017, -0.2141, -0.0652, -0.0017, -0.2132, -0.0642, -0.0039, -0.2141, -0.0652, -0.0039, -0.2132, -0.0642, -0.0017, -0.2124, -0.0631, -0.0039, -0.2124, -0.0631, 0.0, -0.214, -0.0613, 0.0, -0.2152, -0.0583, 0.0, -0.2114, -0.0614, 0.0, -0.2118, -0.0591, 0.0, -0.2142, -0.0566, 0.0, -0.2134, -0.0553, 0.0, -0.2118, -0.057, 0.0, -0.2127, -0.0542, -0.0017, -0.215, -0.0581, -0.0039, -0.215, -0.0581, -0.0017, -0.2113, -0.0614, -0.0039, -0.2113, -0.0614, -0.0017, -0.2144, -0.057, -0.0017, -0.2133, -0.0552, -0.0039, -0.2144, -0.057,
-0.0039, -0.2133, -0.0552, -0.0017, -0.2126, -0.0541, -0.0039, -0.2126, -0.0541, 0.0, -0.2107, -0.0603, 0.0, -0.2101, -0.0593, 0.0, -0.2089, -0.0573, 0.0, -0.2096, -0.0548, 0.0, -0.2083, -0.0562, 0.0, -0.2075, -0.0551, -0.0017, -0.2107, -0.0603, -0.0017, -0.21, -0.0591, -0.0039, -0.2107, -0.0603, -0.0039, -0.21, -0.0591, -0.0017, -0.2089, -0.0574, -0.0039, -0.2089, -0.0574, -0.0017, -0.2083, -0.0563, -0.0017, -0.2074, -0.0548, -0.0039, -0.2083, -0.0563, -0.0039, -0.2074, -0.0548, 0.0, -0.2115, -0.0524, 0.0, -0.2109, -0.0513, -0.0017, -0.2117, -0.0526, -0.0017, -0.2109, -0.0513, -0.0039, -0.2117, -0.0526, -0.0039, -0.2109, -0.0513, 0.0, -0.2096, -0.0526, 0.0, -0.2102, -0.0502, 0.0, -0.2065, -0.0532, 0.0, -0.2075, -0.0505, 0.0, -0.2092, -0.0486,
0.0, -0.2086, -0.0475, 0.0, -0.2074, -0.0458, -0.0017, -0.2102, -0.0502, -0.0039, -0.2102, -0.0502, -0.0017, -0.2066, -0.0534, -0.0039, -0.2066, -0.0534, -0.0017, -0.2091, -0.0484, -0.0039, -0.2091, -0.0484, -0.0017, -0.2085, -0.0473, -0.0017, -0.2076, -0.0461, -0.0039, -0.2085, -0.0473, -0.0039, -0.2076, -0.0461, 0.0, -0.2066, -0.0448, -0.0017, -0.2066, -0.0448, -0.0039, -0.2066, -0.0448, 0.0, -0.2196, -0.0219, 0.0, -0.2196, -0.0201, -0.0017, -0.2196, -0.0219, -0.0017, -0.2196, -0.0201, -0.0039, -0.2196, -0.0219, -0.0039, -0.2196, -0.0201, 0.0, -0.2183, -0.0219, 0.0, -0.2183, -0.0201, 0.0, -0.2161, -0.0219, 0.0, -0.2161, -0.0201, -0.0017, -0.2183, -0.0219, -0.0039, -0.2183, -0.0219, -0.0017, -0.2161, -0.0219, -0.0039, -0.2161, -0.0219, 0.0, -0.214, -0.0219,
0.0, -0.214, -0.0201, 0.0, -0.2118, -0.0219, 0.0, -0.2118, -0.0201, -0.0017, -0.214, -0.0219, -0.0039, -0.214, -0.0219, -0.0017, -0.2118, -0.0219, -0.0039, -0.2118, -0.0219, 0.0, -0.2088, -0.0219, 0.0, -0.2088, -0.0201, -0.0017, -0.2088, -0.0219, -0.0017, -0.2088, -0.0201, -0.0039, -0.2088, -0.0219, -0.0039, -0.2088, -0.0201, 0.0, -0.2196, -0.0179, 0.0, -0.2196, -0.0157, 0.0, -0.2196, -0.0136, 0.0, -0.2196, -0.0114, -0.0017, -0.2196, -0.0179, -0.0017, -0.2196, -0.0157, -0.0039, -0.2196, -0.0179, -0.0039, -0.2196, -0.0157, -0.0017, -0.2196, -0.0136, -0.0017, -0.2196, -0.0114, -0.0039, -0.2196, -0.0136, -0.0039, -0.2196, -0.0114, 0.0, -0.2183, -0.0179, 0.0, -0.2183, -0.0157, 0.0, -0.2161, -0.0179, 0.0, -0.2161, -0.0157, 0.0, -0.2183, -0.0136,
0.0, -0.2183, -0.0114, 0.0, -0.2161, -0.0136, 0.0, -0.2161, -0.0114, 0.0, -0.2196, -0.0092, 0.0, -0.2196, -0.0071, 0.0, -0.2196, -0.0049, 0.0, -0.2196, -0.0027, -0.0017, -0.2196, -0.0092, -0.0017, -0.2196, -0.0071, -0.0039, -0.2196, -0.0092, -0.0039, -0.2196, -0.0071, -0.0017, -0.2196, -0.0049, -0.0017, -0.2196, -0.0027, -0.0039, -0.2196, -0.0049, -0.0039, -0.2196, -0.0027, 0.0, -0.2183, -0.0092, 0.0, -0.2183, -0.0071, 0.0, -0.2161, -0.0092, 0.0, -0.2161, -0.0071, 0.0, -0.2183, -0.0049, 0.0, -0.2183, -0.0027, 0.0, -0.2161, -0.0049, 0.0, -0.2161, -0.0027, 0.0, -0.214, -0.0179, 0.0, -0.214, -0.0157, 0.0, -0.2118, -0.0179, 0.0, -0.2118, -0.0157, 0.0, -0.214, -0.0136, 0.0, -0.214, -0.0114, 0.0, -0.2118, -0.0136,
0.0, -0.2118, -0.0114, 0.0, -0.2088, -0.0179, 0.0, -0.2088, -0.0157, 0.0, -0.2088, -0.0136, 0.0, -0.2088, -0.0114, -0.0017, -0.2088, -0.0179, -0.0017, -0.2088, -0.0157, -0.0039, -0.2088, -0.0179, -0.0039, -0.2088, -0.0157, -0.0017, -0.2088, -0.0136, -0.0017, -0.2088, -0.0114, -0.0039, -0.2088, -0.0136, -0.0039, -0.2088, -0.0114, 0.0, -0.214, -0.0092, 0.0, -0.214, -0.0071, 0.0, -0.2118, -0.0092, 0.0, -0.2118, -0.0071, 0.0, -0.214, -0.0049, 0.0, -0.214, -0.0027, 0.0, -0.2118, -0.0049, 0.0, -0.2118, -0.0027, 0.0, -0.2088, -0.0092, 0.0, -0.2088, -0.0071, 0.0, -0.2088, -0.0049, 0.0, -0.2088, -0.0027, -0.0017, -0.2088, -0.0092, -0.0017, -0.2088, -0.0071, -0.0039, -0.2088, -0.0092, -0.0039, -0.2088, -0.0071, -0.0017, -0.2088, -0.0049,
-0.0017, -0.2088, -0.0027, -0.0039, -0.2088, -0.0049, -0.0039, -0.2088, -0.0027, 0.0, -0.2069, -0.0591, 0.0, -0.2031, -0.0591, 0.0, -0.206, -0.0573, 0.0, -0.2033, -0.0563, -0.0017, -0.2069, -0.0591, -0.0039, -0.2069, -0.0591, -0.0017, -0.2031, -0.0591, -0.0039, -0.2031, -0.0591, -0.0017, -0.206, -0.0573, -0.0039, -0.206, -0.0573, -0.0017, -0.2032, -0.0563, -0.0039, -0.2032, -0.0563, 0.0, -0.2009, -0.0591, 0.0, -0.1988, -0.0591, 0.0, -0.201, -0.0562, 0.0, -0.2001, -0.054, 0.0, -0.199, -0.0562, 0.0, -0.1995, -0.055, -0.0017, -0.2009, -0.0591, -0.0039, -0.2009, -0.0591, -0.0017, -0.1988, -0.0591, -0.0039, -0.1988, -0.0591, -0.0017, -0.2009, -0.0562, -0.0017, -0.2, -0.054, -0.0039, -0.2009, -0.0562, -0.0039, -0.2, -0.054, -0.0017, -0.199, -0.0562,
-0.0017, -0.1995, -0.0551, -0.0039, -0.199, -0.0562, -0.0039, -0.1995, -0.0551, 0.0, -0.1942, -0.0634, -0.0017, -0.1941, -0.0635, -0.0039, -0.1941, -0.0635, 0.0, -0.1893, -0.0668, 0.0, -0.1923, -0.0656, 0.0, -0.1934, -0.0645, 0.0, -0.1904, -0.0667, 0.0, -0.1891, -0.063, -0.0017, -0.1894, -0.0668, -0.0039, -0.1894, -0.0668, -0.0017, -0.1924, -0.0656, -0.0017, -0.1933, -0.0645, -0.0039, -0.1924, -0.0656, -0.0039, -0.1933, -0.0645, -0.0017, -0.1905, -0.0666, -0.0017, -0.1892, -0.0628, -0.0039, -0.1905, -0.0666, -0.0039, -0.1892, -0.0628, 0.0, -0.1957, -0.0605, 0.0, -0.1961, -0.0591, 0.0, -0.1952, -0.0616, 0.0, -0.1944, -0.0591, 0.0, -0.1966, -0.057, 0.0, -0.1966, -0.0548, 0.0, -0.1944, -0.057, 0.0, -0.1939, -0.0546, -0.0017, -0.1956, -0.0605,
-0.0017, -0.1961, -0.0591, -0.0039, -0.1956, -0.0605, -0.0039, -0.1961, -0.0591, -0.0017, -0.1952, -0.0616, -0.0039, -0.1952, -0.0616, -0.0017, -0.1939, -0.0546, -0.0039, -0.1939, -0.0546, 0.0, -0.1923, -0.0613, 0.0, -0.1915, -0.0588, 0.0, -0.1902, -0.0611, 0.0, -0.1909, -0.0599, 0.0, -0.1927, -0.0568, 0.0, -0.1933, -0.0558, -0.0017, -0.1916, -0.0587, -0.0039, -0.1916, -0.0587, -0.0017, -0.1901, -0.0613, -0.0017, -0.1909, -0.0598, -0.0039, -0.1901, -0.0613, -0.0039, -0.1909, -0.0598, -0.0017, -0.1926, -0.057, -0.0017, -0.1933, -0.0557, -0.0039, -0.1926, -0.057, -0.0039, -0.1933, -0.0557, 0.0, -0.2059, -0.0522, 0.0, -0.205, -0.0508, 0.0, -0.2026, -0.05, 0.0, -0.2041, -0.0489, 0.0, -0.2053, -0.0461, 0.0, -0.2039, -0.0484, 0.0, -0.2031, -0.0461,
-0.0017, -0.2059, -0.0523, -0.0017, -0.2049, -0.0505, -0.0039, -0.2059, -0.0523, -0.0039, -0.2049, -0.0505, -0.0017, -0.2026, -0.0501, -0.0039, -0.2026, -0.0501, -0.0017, -0.2043, -0.0492, -0.0039, -0.2043, -0.0492, -0.0017, -0.2039, -0.0484, -0.0039, -0.2039, -0.0484, 0.0, -0.2007, -0.0528, 0.0, -0.2019, -0.0511, 0.0, -0.1988, -0.0526, 0.0, -0.1988, -0.0505, 0.0, -0.2009, -0.0483, 0.0, -0.2009, -0.0461, 0.0, -0.1978, -0.0477, 0.0, -0.199, -0.0458, -0.0017, -0.2008, -0.0526, -0.0017, -0.2018, -0.0512, -0.0039, -0.2008, -0.0526, -0.0039, -0.2018, -0.0512, -0.0017, -0.1979, -0.0476, -0.0017, -0.1989, -0.046, -0.0039, -0.1979, -0.0476, -0.0039, -0.1989, -0.046, 0.0, -0.2052, -0.0436, 0.0, -0.2026, -0.043, -0.0017, -0.2052, -0.0436, -0.0039, -0.2052, -0.0436,
-0.0017, -0.2026, -0.043, -0.0039, -0.2026, -0.043, 0.0, -0.2006, -0.0441, 0.0, -0.1998, -0.045, -0.0017, -0.2006, -0.0441, -0.0039, -0.2006, -0.0441, -0.0017, -0.1998, -0.045, -0.0039, -0.1998, -0.045, 0.0, -0.1955, -0.0519, 0.0, -0.1965, -0.0502, 0.0, -0.1949, -0.0529, 0.0, -0.1972, -0.0488, -0.0017, -0.1956, -0.0517, -0.0017, -0.1963, -0.0505, -0.0039, -0.1956, -0.0517, -0.0039, -0.1963, -0.0505, -0.0017, -0.195, -0.0528, -0.0039, -0.195, -0.0528, -0.0017, -0.1973, -0.0487, -0.0039, -0.1973, -0.0487, 0.0, -0.1873, -0.067, 0.0, -0.1879, -0.0654, 0.0, -0.1886, -0.064, -0.0017, -0.1873, -0.067, -0.0039, -0.1873, -0.067, -0.0017, -0.1878, -0.0656, -0.0017, -0.1886, -0.0639, -0.0039, -0.1878, -0.0656, -0.0039, -0.1886, -0.0639, 0.0, -0.1686, -0.067,
0.0, -0.1686, -0.0656, 0.0, -0.1686, -0.0635, -0.0017, -0.1686, -0.067, -0.0039, -0.1686, -0.067, -0.0017, -0.1686, -0.0656, -0.0017, -0.1686, -0.0635, -0.0039, -0.1686, -0.0656, -0.0039, -0.1686, -0.0635, 0.0, -0.1662, -0.067, 0.0, -0.1641, -0.067, 0.0, -0.1662, -0.0656, 0.0, -0.1662, -0.0635, 0.0, -0.1641, -0.0656, 0.0, -0.1647, -0.0631, -0.0017, -0.1662, -0.067, -0.0039, -0.1662, -0.067, -0.0017, -0.1641, -0.067, -0.0039, -0.1641, -0.067, -0.0017, -0.1647, -0.0631, -0.0039, -0.1647, -0.0631, 0.0, -0.1686, -0.0613, 0.0, -0.1686, -0.0591, 0.0, -0.1686, -0.057, 0.0, -0.1686, -0.0548, -0.0017, -0.1686, -0.0613, -0.0017, -0.1686, -0.0591, -0.0039, -0.1686, -0.0613, -0.0039, -0.1686, -0.0591, -0.0017, -0.1686, -0.057, -0.0017, -0.1686, -0.0548,
-0.0039, -0.1686, -0.057, -0.0039, -0.1686, -0.0548, 0.0, -0.1662, -0.0613, 0.0, -0.1662, -0.0591, 0.0, -0.1647, -0.0613, 0.0, -0.1647, -0.0591, 0.0, -0.1662, -0.057, 0.0, -0.1662, -0.0548, 0.0, -0.1647, -0.057, 0.0, -0.1647, -0.0548, -0.0017, -0.1647, -0.0613, -0.0017, -0.1647, -0.0591, -0.0039, -0.1647, -0.0613, -0.0039, -0.1647, -0.0591, -0.0017, -0.1647, -0.057, -0.0017, -0.1647, -0.0548, -0.0039, -0.1647, -0.057, -0.0039, -0.1647, -0.0548, 0.0, -0.1619, -0.067, 0.0, -0.1597, -0.067, 0.0, -0.1619, -0.0656, 0.0, -0.1619, -0.0631, 0.0, -0.1597, -0.0656, 0.0, -0.1597, -0.0631, -0.0017, -0.1619, -0.067, -0.0039, -0.1619, -0.067, -0.0017, -0.1597, -0.067, -0.0039, -0.1597, -0.067, -0.0017, -0.1619, -0.0631, -0.0039, -0.1619, -0.0631,
-0.0017, -0.1597, -0.0631, -0.0039, -0.1597, -0.0631, 0.0, -0.1575, -0.067, 0.0, -0.1554, -0.067, 0.0, -0.1575, -0.0656, 0.0, -0.1576, -0.0631, 0.0, -0.1554, -0.0656, 0.0, -0.1554, -0.063, -0.0017, -0.1575, -0.067, -0.0039, -0.1575, -0.067, -0.0017, -0.1554, -0.067, -0.0039, -0.1554, -0.067, -0.0017, -0.1575, -0.0631, -0.0039, -0.1575, -0.0631, -0.0017, -0.1554, -0.063, -0.0039, -0.1554, -0.063, 0.0, -0.1686, -0.0526, 0.0, -0.1686, -0.0505, 0.0, -0.1686, -0.0483, 0.0, -0.1686, -0.0461, -0.0017, -0.1686, -0.0526, -0.0017, -0.1686, -0.0505, -0.0039, -0.1686, -0.0526, -0.0039, -0.1686, -0.0505, -0.0017, -0.1686, -0.0483, -0.0017, -0.1686, -0.0461, -0.0039, -0.1686, -0.0483, -0.0039, -0.1686, -0.0461, 0.0, -0.1662, -0.0526, 0.0, -0.1662, -0.0505,
0.0, -0.1647, -0.0526, 0.0, -0.1647, -0.0505, 0.0, -0.1662, -0.0483, 0.0, -0.1662, -0.0461, 0.0, -0.1647, -0.0483, 0.0, -0.1647, -0.0454, -0.0017, -0.1647, -0.0526, -0.0017, -0.1647, -0.0505, -0.0039, -0.1647, -0.0526, -0.0039, -0.1647, -0.0505, -0.0017, -0.1647, -0.0483, -0.0017, -0.1647, -0.0454, -0.0039, -0.1647, -0.0483, -0.0039, -0.1647, -0.0454, 0.0, -0.1686, -0.0439, 0.0, -0.1686, -0.0425, -0.0017, -0.1686, -0.0439, -0.0017, -0.1686, -0.0425, -0.0039, -0.1686, -0.0439, -0.0039, -0.1686, -0.0425, 0.0, -0.1662, -0.0439, 0.0, -0.1662, -0.0425, 0.0, -0.1641, -0.0439, 0.0, -0.1641, -0.0425, -0.0017, -0.1662, -0.0425, -0.0039, -0.1662, -0.0425, -0.0017, -0.1641, -0.0425, -0.0039, -0.1641, -0.0425, 0.0, -0.1619, -0.0454, 0.0, -0.1597, -0.0454,
-0.0017, -0.1619, -0.0454, -0.0039, -0.1619, -0.0454, -0.0017, -0.1597, -0.0454, -0.0039, -0.1597, -0.0454, 0.0, -0.1576, -0.0455, 0.0, -0.1554, -0.0456, -0.0017, -0.1575, -0.0455, -0.0039, -0.1575, -0.0455, -0.0017, -0.1554, -0.0457, -0.0039, -0.1554, -0.0457, 0.0, -0.1619, -0.0439, 0.0, -0.1619, -0.0425, 0.0, -0.1597, -0.0439, 0.0, -0.1597, -0.0425, -0.0017, -0.1619, -0.0425, -0.0039, -0.1619, -0.0425, -0.0017, -0.1597, -0.0425, -0.0039, -0.1597, -0.0425, 0.0, -0.1575, -0.0439, 0.0, -0.1575, -0.0425, 0.0, -0.1554, -0.0439, 0.0, -0.1554, -0.0425, -0.0017, -0.1575, -0.0425, -0.0039, -0.1575, -0.0425, -0.0017, -0.1554, -0.0425, -0.0039, -0.1554, -0.0425, 0.0, -0.1537, -0.0669, 0.0, -0.1529, -0.0668, 0.0, -0.1533, -0.0629, 0.0, -0.1507, -0.0664,
0.0, -0.1513, -0.0624, -0.0017, -0.1535, -0.0669, -0.0039, -0.1535, -0.0669, -0.0017, -0.1524, -0.0667, -0.0017, -0.1532, -0.0629, -0.0039, -0.1524, -0.0667, -0.0039, -0.1532, -0.0629, -0.0017, -0.151, -0.0665, -0.0017, -0.1516, -0.0625, -0.0039, -0.151, -0.0665, -0.0039, -0.1516, -0.0625, 0.0, -0.1486, -0.0655, 0.0, -0.1489, -0.0635, 0.0, -0.1475, -0.0648, 0.0, -0.1463, -0.064, -0.0017, -0.1488, -0.0656, -0.0039, -0.1488, -0.0656, -0.0017, -0.1475, -0.0648, -0.0017, -0.1463, -0.064, -0.0039, -0.1475, -0.0648, -0.0039, -0.1463, -0.064, 0.0, -0.1504, -0.0621, -0.0017, -0.1504, -0.0621, -0.0039, -0.1504, -0.0621, 0.0, -0.149, -0.0612, 0.0, -0.148, -0.06, 0.0, -0.1467, -0.0613, 0.0, -0.1472, -0.059, 0.0, -0.1464, -0.0569, 0.0, -0.1462, -0.0548,
-0.0017, -0.1489, -0.0611, -0.0017, -0.148, -0.06, -0.0039, -0.1489, -0.0611, -0.0039, -0.148, -0.06, -0.0017, -0.1472, -0.059, -0.0039, -0.1472, -0.059, -0.0017, -0.1465, -0.057, -0.0017, -0.1462, -0.0548, -0.0039, -0.1465, -0.057, -0.0039, -0.1462, -0.0548, 0.0, -0.1452, -0.0628, -0.0017, -0.1452, -0.0628, -0.0039, -0.1452, -0.0628, 0.0, -0.1441, -0.0615, 0.0, -0.1445, -0.0591, 0.0, -0.1434, -0.0603, 0.0, -0.1429, -0.0593, 0.0, -0.1445, -0.057, 0.0, -0.1445, -0.0548, 0.0, -0.1423, -0.0571, 0.0, -0.1422, -0.0548, -0.0017, -0.144, -0.0614, -0.0039, -0.144, -0.0614, -0.0017, -0.1434, -0.0603, -0.0017, -0.1428, -0.0592, -0.0039, -0.1434, -0.0603, -0.0039, -0.1428, -0.0592, -0.0017, -0.1423, -0.057, -0.0017, -0.1422, -0.0548, -0.0039, -0.1423, -0.057,
-0.0039, -0.1422, -0.0548, 0.0, -0.1533, -0.0461, 0.0, -0.1502, -0.0474, 0.0, -0.1512, -0.0469, -0.0017, -0.1532, -0.0461, -0.0039, -0.1532, -0.0461, -0.0017, -0.1503, -0.0474, -0.0017, -0.1514, -0.0468, -0.0039, -0.1503, -0.0474, -0.0039, -0.1514, -0.0468, 0.0, -0.1479, -0.0495, 0.0, -0.1464, -0.0527, 0.0, -0.1471, -0.0505, 0.0, -0.1489, -0.0484, 0.0, -0.1489, -0.0461, 0.0, -0.1467, -0.0483, 0.0, -0.1458, -0.0451, -0.0017, -0.1479, -0.0495, -0.0039, -0.1479, -0.0495, -0.0017, -0.1464, -0.0526, -0.0017, -0.1469, -0.0509, -0.0039, -0.1464, -0.0526, -0.0039, -0.1469, -0.0509, -0.0017, -0.1489, -0.0484, -0.0039, -0.1489, -0.0484, -0.0017, -0.1458, -0.0452, -0.0039, -0.1458, -0.0452, 0.0, -0.1532, -0.0439, 0.0, -0.1532, -0.0425, 0.0, -0.1506, -0.0428,
0.0, -0.1515, -0.0426, -0.0017, -0.1532, -0.0425, -0.0039, -0.1532, -0.0425, -0.0017, -0.1501, -0.0429, -0.0017, -0.1511, -0.0426, -0.0039, -0.1501, -0.0429, -0.0039, -0.1511, -0.0426, 0.0, -0.1485, -0.0433, 0.0, -0.1469, -0.0443, -0.0017, -0.1488, -0.0432, -0.0039, -0.1488, -0.0432, -0.0017, -0.1468, -0.0443, -0.0039, -0.1468, -0.0443, 0.0, -0.1445, -0.0526, 0.0, -0.1445, -0.0505, 0.0, -0.1422, -0.0526, 0.0, -0.1425, -0.0503, 0.0, -0.1436, -0.0478, 0.0, -0.1447, -0.0462, 0.0, -0.143, -0.0489, -0.0017, -0.1422, -0.0526, -0.0017, -0.1425, -0.0504, -0.0039, -0.1422, -0.0526, -0.0039, -0.1425, -0.0504, -0.0017, -0.1437, -0.0477, -0.0017, -0.1447, -0.0462, -0.0039, -0.1437, -0.0477, -0.0039, -0.1447, -0.0462, -0.0017, -0.1431, -0.0487, -0.0039, -0.1431, -0.0487,
0.0, -0.151, -0.0219, 0.0, -0.151, -0.0201, -0.0017, -0.151, -0.0219, -0.0017, -0.151, -0.0201, -0.0039, -0.151, -0.0219, -0.0039, -0.151, -0.0201, 0.0, -0.1489, -0.0219, 0.0, -0.1489, -0.0201, 0.0, -0.1467, -0.0219, 0.0, -0.1467, -0.0201, -0.0017, -0.1489, -0.0219, -0.0039, -0.1489, -0.0219, -0.0017, -0.1467, -0.0219, -0.0039, -0.1467, -0.0219, 0.0, -0.1445, -0.0219, 0.0, -0.1445, -0.0201, 0.0, -0.1424, -0.0219, 0.0, -0.1424, -0.0201, -0.0017, -0.1445, -0.0219, -0.0039, -0.1445, -0.0219, -0.0017, -0.1424, -0.0219, -0.0039, -0.1424, -0.0219, 0.0, -0.1402, -0.0219, 0.0, -0.1402, -0.0201, 0.0, -0.138, -0.0219, 0.0, -0.138, -0.0201, -0.0017, -0.1402, -0.0219, -0.0039, -0.1402, -0.0219, -0.0017, -0.138, -0.0219, -0.0039, -0.138, -0.0219,
0.0, -0.151, -0.0179, 0.0, -0.151, -0.0157, 0.0, -0.151, -0.0136, 0.0, -0.151, -0.0114, -0.0017, -0.151, -0.0179, -0.0017, -0.151, -0.0157, -0.0039, -0.151, -0.0179, -0.0039, -0.151, -0.0157, -0.0017, -0.151, -0.0136, -0.0017, -0.151, -0.0114, -0.0039, -0.151, -0.0136, -0.0039, -0.151, -0.0114, 0.0, -0.1489, -0.0179, 0.0, -0.1489, -0.0157, 0.0, -0.1467, -0.0179, 0.0, -0.1467, -0.0157, 0.0, -0.1489, -0.0136, 0.0, -0.1489, -0.0114, 0.0, -0.1467, -0.0136, 0.0, -0.1467, -0.0114, 0.0, -0.151, -0.0092, 0.0, -0.151, -0.0071, 0.0, -0.151, -0.0049, 0.0, -0.151, -0.0027, -0.0017, -0.151, -0.0092, -0.0017, -0.151, -0.0071, -0.0039, -0.151, -0.0092, -0.0039, -0.151, -0.0071, -0.0017, -0.151, -0.0049, -0.0017, -0.151, -0.0027,
-0.0039, -0.151, -0.0049, -0.0039, -0.151, -0.0027, 0.0, -0.1489, -0.0092, 0.0, -0.1489, -0.0071, 0.0, -0.1467, -0.0092, 0.0, -0.1467, -0.0071, 0.0, -0.1489, -0.0049, 0.0, -0.1489, -0.0027, 0.0, -0.1467, -0.0049, 0.0, -0.1467, -0.0027, 0.0, -0.1445, -0.0179, 0.0, -0.1445, -0.0157, 0.0, -0.1424, -0.0179, 0.0, -0.1424, -0.0157, 0.0, -0.1445, -0.0136, 0.0, -0.1445, -0.0114, 0.0, -0.1424, -0.0136, 0.0, -0.1424, -0.0114, 0.0, -0.1402, -0.0179, 0.0, -0.1402, -0.0157, 0.0, -0.138, -0.0179, 0.0, -0.138, -0.0157, 0.0, -0.1402, -0.0136, 0.0, -0.1402, -0.0114, 0.0, -0.138, -0.0136, 0.0, -0.138, -0.0114, 0.0, -0.1445, -0.0092, 0.0, -0.1445, -0.0071, 0.0, -0.1424, -0.0092, 0.0, -0.1424, -0.0071,
0.0, -0.1445, -0.0049, 0.0, -0.1445, -0.0027, 0.0, -0.1424, -0.0049, 0.0, -0.1424, -0.0027, 0.0, -0.1402, -0.0092, 0.0, -0.1402, -0.0071, 0.0, -0.1375, -0.0087, 0.0, -0.1387, -0.0072, 0.0, -0.1391, -0.0048, 0.0, -0.1392, -0.0027, 0.0, -0.139, -0.0056, -0.0017, -0.1376, -0.0086, -0.0017, -0.1387, -0.0072, -0.0039, -0.1376, -0.0086, -0.0039, -0.1387, -0.0072, -0.0017, -0.1391, -0.0043, -0.0017, -0.1392, -0.0027, -0.0039, -0.1391, -0.0043, -0.0039, -0.1392, -0.0027, -0.0017, -0.139, -0.0054, -0.0039, -0.139, -0.0054, 0.0, -0.1225, -0.067, 0.0, -0.1207, -0.067, 0.0, -0.1225, -0.0656, 0.0, -0.1225, -0.0635, 0.0, -0.1207, -0.0656, 0.0, -0.1207, -0.0635, -0.0017, -0.1225, -0.067, -0.0039, -0.1225, -0.067, -0.0017, -0.1207, -0.067,
-0.0039, -0.1207, -0.067, -0.0017, -0.1225, -0.0656, -0.0017, -0.1225, -0.0635, -0.0039, -0.1225, -0.0656, -0.0039, -0.1225, -0.0635, 0.0, -0.1225, -0.0613, 0.0, -0.1225, -0.0591, 0.0, -0.1207, -0.0613, 0.0, -0.1207, -0.0591, 0.0, -0.1225, -0.057, 0.0, -0.1225, -0.0548, 0.0, -0.1207, -0.057, 0.0, -0.1207, -0.0548, -0.0017, -0.1225, -0.0613, -0.0017, -0.1225, -0.0591, -0.0039, -0.1225, -0.0613, -0.0039, -0.1225, -0.0591, -0.0017, -0.1225, -0.057, -0.0017, -0.1225, -0.0548, -0.0039, -0.1225, -0.057, -0.0039, -0.1225, -0.0548, 0.0, -0.1225, -0.0526, 0.0, -0.1225, -0.0505, 0.0, -0.1207, -0.0526, 0.0, -0.1207, -0.0505, 0.0, -0.1225, -0.0483, 0.0, -0.1225, -0.0461, 0.0, -0.1207, -0.0483, 0.0, -0.1207, -0.0461, -0.0017, -0.1225, -0.0526,
-0.0017, -0.1225, -0.0505, -0.0039, -0.1225, -0.0526, -0.0039, -0.1225, -0.0505, -0.0017, -0.1225, -0.0483, -0.0017, -0.1225, -0.0461, -0.0039, -0.1225, -0.0483, -0.0039, -0.1225, -0.0461, 0.0, -0.1225, -0.0439, 0.0, -0.1225, -0.0425, 0.0, -0.1207, -0.0439, 0.0, -0.1207, -0.0425, -0.0017, -0.1225, -0.0439, -0.0017, -0.1225, -0.0425, -0.0039, -0.1225, -0.0439, -0.0039, -0.1225, -0.0425, -0.0017, -0.1207, -0.0425, -0.0039, -0.1207, -0.0425, 0.0, -0.1185, -0.067, 0.0, -0.1163, -0.067, 0.0, -0.1185, -0.0656, 0.0, -0.1186, -0.0631, 0.0, -0.1163, -0.0656, 0.0, -0.1163, -0.0631, -0.0017, -0.1185, -0.067, -0.0039, -0.1185, -0.067, -0.0017, -0.1163, -0.067, -0.0039, -0.1163, -0.067, -0.0017, -0.1186, -0.0631, -0.0039, -0.1186, -0.0631, -0.0017, -0.1163, -0.0631,
-0.0039, -0.1163, -0.0631, 0.0, -0.1141, -0.067, 0.0, -0.112, -0.067, 0.0, -0.1141, -0.0656, 0.0, -0.1141, -0.0631, 0.0, -0.112, -0.0656, 0.0, -0.112, -0.063, -0.0017, -0.1141, -0.067, -0.0039, -0.1141, -0.067, -0.0017, -0.112, -0.067, -0.0039, -0.112, -0.067, -0.0017, -0.1141, -0.0631, -0.0039, -0.1141, -0.0631, -0.0017, -0.112, -0.063, -0.0039, -0.112, -0.063, 0.0, -0.1186, -0.0613, 0.0, -0.1186, -0.0591, 0.0, -0.1186, -0.057, 0.0, -0.1186, -0.0548, -0.0017, -0.1186, -0.0613, -0.0017, -0.1186, -0.0591, -0.0039, -0.1186, -0.0613, -0.0039, -0.1186, -0.0591, -0.0017, -0.1186, -0.057, -0.0017, -0.1186, -0.0548, -0.0039, -0.1186, -0.057, -0.0039, -0.1186, -0.0548, 0.0, -0.1098, -0.067, 0.0, -0.108, -0.0669, 0.0, -0.1098, -0.0656,
0.0, -0.1098, -0.063, 0.0, -0.1072, -0.0669, 0.0, -0.1078, -0.0629, -0.0017, -0.1098, -0.067, -0.0039, -0.1098, -0.067, -0.0017, -0.1076, -0.0669, -0.0039, -0.1076, -0.0669, -0.0017, -0.1098, -0.063, -0.0039, -0.1098, -0.063, -0.0017, -0.1066, -0.0667, -0.0017, -0.1076, -0.0629, -0.0039, -0.1066, -0.0667, -0.0039, -0.1076, -0.0629, 0.0, -0.1052, -0.0665, 0.0, -0.1058, -0.0624, 0.0, -0.1031, -0.0658, 0.0, -0.1033, -0.0635, -0.0017, -0.1054, -0.0666, -0.0017, -0.1061, -0.0625, -0.0039, -0.1054, -0.0666, -0.0039, -0.1061, -0.0625, -0.0017, -0.1033, -0.0658, -0.0039, -0.1033, -0.0658, 0.0, -0.1048, -0.0622, 0.0, -0.1034, -0.0616, -0.0017, -0.105, -0.0622, -0.0039, -0.105, -0.0622, -0.0017, -0.103, -0.0613, -0.0039, -0.103, -0.0613, 0.0, -0.1186, -0.0526,
0.0, -0.1186, -0.0505, 0.0, -0.1186, -0.0483, 0.0, -0.1186, -0.0454, 0.0, -0.1163, -0.0454, -0.0017, -0.1186, -0.0526, -0.0017, -0.1186, -0.0505, -0.0039, -0.1186, -0.0526, -0.0039, -0.1186, -0.0505, -0.0017, -0.1186, -0.0483, -0.0017, -0.1186, -0.0454, -0.0039, -0.1186, -0.0483, -0.0039, -0.1186, -0.0454, -0.0017, -0.1163, -0.0454, -0.0039, -0.1163, -0.0454, 0.0, -0.1141, -0.0454, 0.0, -0.112, -0.0454, -0.0017, -0.1141, -0.0454, -0.0039, -0.1141, -0.0454, -0.0017, -0.112, -0.0454, -0.0039, -0.112, -0.0454, 0.0, -0.1185, -0.0439, 0.0, -0.1185, -0.0425, 0.0, -0.1163, -0.0439, 0.0, -0.1163, -0.0425, -0.0017, -0.1185, -0.0425, -0.0039, -0.1185, -0.0425, -0.0017, -0.1163, -0.0425, -0.0039, -0.1163, -0.0425, 0.0, -0.1141, -0.0439, 0.0, -0.1141, -0.0425,
0.0, -0.112, -0.0439, 0.0, -0.112, -0.0425, -0.0017, -0.1141, -0.0425, -0.0039, -0.1141, -0.0425, -0.0017, -0.112, -0.0425, -0.0039, -0.112, -0.0425, 0.0, -0.1098, -0.0455, 0.0, -0.1077, -0.0459, -0.0017, -0.1098, -0.0455, -0.0039, -0.1098, -0.0455, -0.0017, -0.1076, -0.0459, -0.0039, -0.1076, -0.0459, 0.0, -0.1057, -0.0464, 0.0, -0.103, -0.0479, 0.0, -0.104, -0.0473, -0.0017, -0.1054, -0.0465, -0.0039, -0.1054, -0.0465, -0.0017, -0.1031, -0.0478, -0.0017, -0.1043, -0.0471, -0.0039, -0.1031, -0.0478, -0.0039, -0.1043, -0.0471, 0.0, -0.1098, -0.0439, 0.0, -0.1098, -0.0425, 0.0, -0.1076, -0.0439, 0.0, -0.1076, -0.0425, -0.0017, -0.1098, -0.0425, -0.0039, -0.1098, -0.0425, -0.0017, -0.1076, -0.0425, -0.0039, -0.1076, -0.0425, 0.0, -0.1055, -0.0439,
0.0, -0.1054, -0.0426, 0.0, -0.1032, -0.0431, 0.0, -0.1043, -0.0428, -0.0017, -0.1054, -0.0426, -0.0039, -0.1054, -0.0426, -0.0017, -0.1031, -0.0431, -0.0017, -0.1042, -0.0428, -0.0039, -0.1031, -0.0431, -0.0039, -0.1042, -0.0428, 0.0, -0.1358, -0.0219, 0.0, -0.1358, -0.0201, 0.0, -0.1337, -0.0219, 0.0, -0.1337, -0.0201, -0.0017, -0.1358, -0.0219, -0.0039, -0.1358, -0.0219, -0.0017, -0.1337, -0.0219, -0.0039, -0.1337, -0.0219, 0.0, -0.1315, -0.0219, 0.0, -0.1315, -0.0201, 0.0, -0.1293, -0.0219, 0.0, -0.1293, -0.0201, -0.0017, -0.1315, -0.0219, -0.0039, -0.1315, -0.0219, -0.0017, -0.1293, -0.0219, -0.0039, -0.1293, -0.0219, 0.0, -0.1272, -0.0219, 0.0, -0.1272, -0.0201, 0.0, -0.125, -0.0219, 0.0, -0.125, -0.0201, -0.0017, -0.1272, -0.0219,
-0.0039, -0.1272, -0.0219, -0.0017, -0.125, -0.0219, -0.0039, -0.125, -0.0219, 0.0, -0.1228, -0.0219, 0.0, -0.1228, -0.0201, 0.0, -0.1207, -0.0219, 0.0, -0.1207, -0.0201, -0.0017, -0.1228, -0.0219, -0.0039, -0.1228, -0.0219, -0.0017, -0.1207, -0.0219, -0.0039, -0.1207, -0.0219, 0.0, -0.1358, -0.0179, 0.0, -0.1358, -0.0157, 0.0, -0.1337, -0.0179, 0.0, -0.1337, -0.0157, 0.0, -0.1358, -0.0136, 0.0, -0.1358, -0.0114, 0.0, -0.1337, -0.0136, 0.0, -0.1337, -0.0114, 0.0, -0.1315, -0.0179, 0.0, -0.1315, -0.0157, 0.0, -0.1293, -0.0179, 0.0, -0.1293, -0.0157, 0.0, -0.1315, -0.0136, 0.0, -0.1315, -0.0114, 0.0, -0.1293, -0.0136, 0.0, -0.1293, -0.0114, 0.0, -0.1358, -0.0097, 0.0, -0.1337, -0.01, -0.0017, -0.1358, -0.0097,
-0.0039, -0.1358, -0.0097, -0.0017, -0.1337, -0.01, -0.0039, -0.1337, -0.01, 0.0, -0.1315, -0.0101, 0.0, -0.1293, -0.0101, -0.0017, -0.1315, -0.0101, -0.0039, -0.1315, -0.0101, -0.0017, -0.1293, -0.0101, -0.0039, -0.1293, -0.0101, 0.0, -0.1272, -0.0179, 0.0, -0.1272, -0.0157, 0.0, -0.125, -0.0179, 0.0, -0.125, -0.0157, 0.0, -0.1272, -0.0136, 0.0, -0.1272, -0.0114, 0.0, -0.125, -0.0136, 0.0, -0.125, -0.0114, 0.0, -0.1228, -0.0179, 0.0, -0.1228, -0.0157, 0.0, -0.1207, -0.0179, 0.0, -0.1207, -0.0157, 0.0, -0.1228, -0.0136, 0.0, -0.1228, -0.0114, 0.0, -0.1207, -0.0136, 0.0, -0.1207, -0.0114, 0.0, -0.1272, -0.0101, 0.0, -0.125, -0.0101, -0.0017, -0.1272, -0.0101, -0.0039, -0.1272, -0.0101, -0.0017, -0.125, -0.0101,
-0.0039, -0.125, -0.0101, 0.0, -0.1228, -0.0101, 0.0, -0.1207, -0.0101, -0.0017, -0.1228, -0.0101, -0.0039, -0.1228, -0.0101, -0.0017, -0.1207, -0.0101, -0.0039, -0.1207, -0.0101, 0.0, -0.1185, -0.0219, 0.0, -0.1185, -0.0201, 0.0, -0.1163, -0.0219, 0.0, -0.1163, -0.0201, -0.0017, -0.1185, -0.0219, -0.0039, -0.1185, -0.0219, -0.0017, -0.1163, -0.0219, -0.0039, -0.1163, -0.0219, 0.0, -0.1141, -0.0219, 0.0, -0.1141, -0.0201, 0.0, -0.112, -0.0219, 0.0, -0.112, -0.0201, -0.0017, -0.1141, -0.0219, -0.0039, -0.1141, -0.0219, -0.0017, -0.112, -0.0219, -0.0039, -0.112, -0.0219, 0.0, -0.1098, -0.0219, 0.0, -0.1098, -0.0201, 0.0, -0.1076, -0.0219, 0.0, -0.1076, -0.0201, -0.0017, -0.1098, -0.0219, -0.0039, -0.1098, -0.0219, -0.0017, -0.1076, -0.0219,
-0.0039, -0.1076, -0.0219, 0.0, -0.1055, -0.0219, 0.0, -0.1055, -0.0201, 0.0, -0.1033, -0.0219, 0.0, -0.1033, -0.0201, -0.0017, -0.1055, -0.0219, -0.0039, -0.1055, -0.0219, -0.0017, -0.1033, -0.0219, -0.0039, -0.1033, -0.0219, 0.0, -0.1185, -0.0179, 0.0, -0.1185, -0.0157, 0.0, -0.1163, -0.0179, 0.0, -0.1163, -0.0157, 0.0, -0.1185, -0.0136, 0.0, -0.1185, -0.0114, 0.0, -0.1163, -0.0136, 0.0, -0.1163, -0.0114, 0.0, -0.1141, -0.0179, 0.0, -0.1141, -0.0157, 0.0, -0.112, -0.0179, 0.0, -0.112, -0.0157, 0.0, -0.1141, -0.0136, 0.0, -0.1141, -0.0114, 0.0, -0.112, -0.0136, 0.0, -0.112, -0.0114, 0.0, -0.1185, -0.0101, 0.0, -0.1163, -0.0101, -0.0017, -0.1185, -0.0101, -0.0039, -0.1185, -0.0101, -0.0017, -0.1163, -0.0101,
-0.0039, -0.1163, -0.0101, 0.0, -0.1141, -0.0101, 0.0, -0.112, -0.0101, -0.0017, -0.1141, -0.0101, -0.0039, -0.1141, -0.0101, -0.0017, -0.112, -0.0101, -0.0039, -0.112, -0.0101, 0.0, -0.1098, -0.0179, 0.0, -0.1098, -0.0157, 0.0, -0.1076, -0.0179, 0.0, -0.1076, -0.0157, 0.0, -0.1098, -0.0136, 0.0, -0.1098, -0.0114, 0.0, -0.1076, -0.0136, 0.0, -0.1076, -0.0114, 0.0, -0.1055, -0.0179, 0.0, -0.1055, -0.0157, 0.0, -0.1033, -0.0179, 0.0, -0.1033, -0.0157, 0.0, -0.1055, -0.0136, 0.0, -0.1055, -0.0114, 0.0, -0.1033, -0.0136, 0.0, -0.1033, -0.0114, 0.0, -0.1098, -0.0101, 0.0, -0.1076, -0.0101, -0.0017, -0.1098, -0.0101, -0.0039, -0.1098, -0.0101, -0.0017, -0.1076, -0.0101, -0.0039, -0.1076, -0.0101, 0.0, -0.1055, -0.0101,
0.0, -0.1033, -0.0101, -0.0017, -0.1055, -0.0101, -0.0039, -0.1055, -0.0101, -0.0017, -0.1033, -0.0101, -0.0039, -0.1033, -0.0101, 0.0, -0.1017, -0.065, 0.0, -0.1006, -0.0642, 0.0, -0.0994, -0.0631, -0.0017, -0.1016, -0.065, -0.0017, -0.1005, -0.0642, -0.0039, -0.1016, -0.065, -0.0039, -0.1005, -0.0642, -0.0017, -0.0994, -0.0631, -0.0039, -0.0994, -0.0631, 0.0, -0.102, -0.0604, 0.0, -0.1007, -0.059, 0.0, -0.0983, -0.0618, 0.0, -0.099, -0.0591, 0.0, -0.1002, -0.0578, 0.0, -0.0998, -0.0569, 0.0, -0.0995, -0.0548, -0.0017, -0.102, -0.0604, -0.0017, -0.1007, -0.0591, -0.0039, -0.102, -0.0604, -0.0039, -0.1007, -0.0591, -0.0017, -0.0983, -0.0618, -0.0039, -0.0983, -0.0618, -0.0017, -0.1002, -0.0577, -0.0039, -0.1002, -0.0577, -0.0017, -0.0997, -0.0566,
-0.0017, -0.0995, -0.0548, -0.0039, -0.0997, -0.0566, -0.0039, -0.0995, -0.0548, 0.0, -0.0975, -0.0606, 0.0, -0.0968, -0.0594, 0.0, -0.0962, -0.0571, 0.0, -0.0961, -0.0548, -0.0017, -0.0976, -0.0607, -0.0017, -0.0968, -0.0592, -0.0039, -0.0976, -0.0607, -0.0039, -0.0968, -0.0592, -0.0017, -0.0962, -0.057, -0.0017, -0.0961, -0.0548, -0.0039, -0.0962, -0.057, -0.0039, -0.0961, -0.0548, 0.0, -0.1002, -0.0518, 0.0, -0.1006, -0.0506, 0.0, -0.0998, -0.0527, 0.0, -0.099, -0.0505, 0.0, -0.1019, -0.049, 0.0, -0.1011, -0.0461, 0.0, -0.0978, -0.0474, 0.0, -0.099, -0.0457, -0.0017, -0.1002, -0.0518, -0.0017, -0.1008, -0.0504, -0.0039, -0.1002, -0.0518, -0.0039, -0.1008, -0.0504, -0.0017, -0.0997, -0.0529, -0.0039, -0.0997, -0.0529, -0.0017, -0.1019, -0.049,
-0.0039, -0.1019, -0.049, -0.0017, -0.0979, -0.0473, -0.0017, -0.0988, -0.046, -0.0039, -0.0979, -0.0473, -0.0039, -0.0988, -0.046, 0.0, -0.0961, -0.0526, 0.0, -0.0965, -0.0503, 0.0, -0.0971, -0.0485, -0.0017, -0.0961, -0.0526, -0.0017, -0.0964, -0.0504, -0.0039, -0.0961, -0.0526, -0.0039, -0.0964, -0.0504, -0.0017, -0.0972, -0.0483, -0.0039, -0.0972, -0.0483, 0.0, -0.1009, -0.0442, 0.0, -0.1, -0.045, -0.0017, -0.101, -0.0441, -0.0039, -0.101, -0.0441, -0.0017, -0.1, -0.045, -0.0039, -0.1, -0.045, 0.0, -0.0766, -0.0587, 0.0, -0.0771, -0.057, 0.0, -0.0774, -0.0548, -0.0017, -0.0766, -0.0589, -0.0039, -0.0766, -0.0589, -0.0017, -0.0771, -0.057, -0.0017, -0.0774, -0.0548, -0.0039, -0.0771, -0.057, -0.0039, -0.0774, -0.0548, 0.0, -0.0742, -0.0626,
0.0, -0.072, -0.0647, 0.0, -0.0732, -0.0638, -0.0017, -0.0742, -0.0627, -0.0039, -0.0742, -0.0627, -0.0017, -0.072, -0.0647, -0.0017, -0.0731, -0.0638, -0.0039, -0.072, -0.0647, -0.0039, -0.0731, -0.0638, 0.0, -0.071, -0.0654, 0.0, -0.0707, -0.0635, 0.0, -0.0688, -0.0663, 0.0, -0.0681, -0.0623, -0.0017, -0.0708, -0.0655, -0.0039, -0.0708, -0.0655, -0.0017, -0.0686, -0.0664, -0.0017, -0.0676, -0.0624, -0.0039, -0.0686, -0.0664, -0.0039, -0.0676, -0.0624, 0.0, -0.0751, -0.0617, 0.0, -0.0762, -0.0597, 0.0, -0.0729, -0.0613, 0.0, -0.0722, -0.0588, 0.0, -0.0751, -0.057, 0.0, -0.0751, -0.0548, 0.0, -0.073, -0.0568, 0.0, -0.0733, -0.0548, -0.0017, -0.0753, -0.0614, -0.0017, -0.0761, -0.06, -0.0039, -0.0753, -0.0614, -0.0039, -0.0761, -0.06,
-0.0017, -0.0723, -0.0587, -0.0039, -0.0723, -0.0587, -0.0017, -0.0729, -0.057, -0.0017, -0.0733, -0.0548, -0.0039, -0.0729, -0.057, -0.0039, -0.0733, -0.0548, 0.0, -0.0705, -0.061, 0.0, -0.0714, -0.0598, 0.0, -0.0689, -0.0619, -0.0017, -0.0707, -0.0608, -0.0017, -0.0715, -0.0598, -0.0039, -0.0707, -0.0608, -0.0039, -0.0715, -0.0598, -0.0017, -0.0687, -0.062, -0.0039, -0.0687, -0.062, 0.0, -0.0773, -0.0526, 0.0, -0.0769, -0.0503, 0.0, -0.0764, -0.0491, -0.0017, -0.0773, -0.0526, -0.0017, -0.0769, -0.0504, -0.0039, -0.0773, -0.0526, -0.0039, -0.0769, -0.0504, -0.0017, -0.0763, -0.049, -0.0039, -0.0763, -0.049, 0.0, -0.0751, -0.0526, 0.0, -0.0751, -0.0505, 0.0, -0.0732, -0.0527, 0.0, -0.0725, -0.0505, 0.0, -0.0758, -0.048, 0.0, -0.0748, -0.0464,
0.0, -0.0729, -0.0483, 0.0, -0.0736, -0.0453, -0.0017, -0.0732, -0.0526, -0.0017, -0.0727, -0.0508, -0.0039, -0.0732, -0.0526, -0.0039, -0.0727, -0.0508, -0.0017, -0.0758, -0.0479, -0.0017, -0.0747, -0.0464, -0.0039, -0.0758, -0.0479, -0.0039, -0.0747, -0.0464, -0.0017, -0.0736, -0.0453, -0.0039, -0.0736, -0.0453, 0.0, -0.0717, -0.0495, 0.0, -0.0708, -0.0483, 0.0, -0.0707, -0.0461, 0.0, -0.0694, -0.0474, 0.0, -0.0684, -0.0468, -0.0017, -0.0717, -0.0495, -0.0039, -0.0717, -0.0495, -0.0017, -0.0707, -0.0482, -0.0039, -0.0707, -0.0482, -0.0017, -0.0694, -0.0474, -0.0017, -0.0683, -0.0468, -0.0039, -0.0694, -0.0474, -0.0039, -0.0683, -0.0468, 0.0, -0.0725, -0.0445, -0.0017, -0.0726, -0.0445, -0.0039, -0.0726, -0.0445, 0.0, -0.0711, -0.0437, 0.0, -0.0689, -0.0429,
0.0, -0.0679, -0.0427, -0.0017, -0.0708, -0.0435, -0.0039, -0.0708, -0.0435, -0.0017, -0.0692, -0.043, -0.0017, -0.0681, -0.0427, -0.0039, -0.0692, -0.043, -0.0039, -0.0681, -0.0427, 0.0, -0.1011, -0.0219, 0.0, -0.1011, -0.0201, 0.0, -0.099, -0.0219, 0.0, -0.099, -0.0201, -0.0017, -0.1011, -0.0219, -0.0039, -0.1011, -0.0219, -0.0017, -0.099, -0.0219, -0.0039, -0.099, -0.0219, 0.0, -0.0968, -0.0219, 0.0, -0.0968, -0.0201, 0.0, -0.0946, -0.0219, 0.0, -0.0946, -0.0201, -0.0017, -0.0968, -0.0219, -0.0039, -0.0968, -0.0219, -0.0017, -0.0946, -0.0219, -0.0039, -0.0946, -0.0219, 0.0, -0.0924, -0.0219, 0.0, -0.0924, -0.0201, 0.0, -0.0902, -0.0219, 0.0, -0.0902, -0.0201, -0.0017, -0.0924, -0.0219, -0.0039, -0.0924, -0.0219, -0.0017, -0.0902, -0.0219,
-0.0017, -0.0902, -0.0201, -0.0039, -0.0902, -0.0219, -0.0039, -0.0902, -0.0201, 0.0, -0.1011, -0.0179, 0.0, -0.1011, -0.0157, 0.0, -0.099, -0.0179, 0.0, -0.099, -0.0157, 0.0, -0.1011, -0.0136, 0.0, -0.1011, -0.0114, 0.0, -0.099, -0.0136, 0.0, -0.099, -0.0114, 0.0, -0.0968, -0.0179, 0.0, -0.0968, -0.0157, 0.0, -0.0946, -0.0179, 0.0, -0.0946, -0.0157, 0.0, -0.0968, -0.0136, 0.0, -0.0968, -0.0114, 0.0, -0.0946, -0.0136, 0.0, -0.0946, -0.0114, 0.0, -0.1011, -0.0101, 0.0, -0.099, -0.0101, -0.0017, -0.1011, -0.0101, -0.0039, -0.1011, -0.0101, -0.0017, -0.099, -0.0101, -0.0039, -0.099, -0.0101, 0.0, -0.0968, -0.0101, 0.0, -0.0946, -0.0101, -0.0017, -0.0968, -0.0101, -0.0039, -0.0968, -0.0101, -0.0017, -0.0946, -0.0101,
-0.0039, -0.0946, -0.0101, 0.0, -0.0924, -0.0179, 0.0, -0.0924, -0.0157, 0.0, -0.0902, -0.0179, 0.0, -0.0902, -0.0157, 0.0, -0.0924, -0.0136, 0.0, -0.0924, -0.0114, 0.0, -0.0902, -0.0136, 0.0, -0.0902, -0.0114, -0.0017, -0.0902, -0.0179, -0.0017, -0.0902, -0.0157, -0.0039, -0.0902, -0.0179, -0.0039, -0.0902, -0.0157, -0.0017, -0.0902, -0.0136, -0.0017, -0.0902, -0.0114, -0.0039, -0.0902, -0.0136, -0.0039, -0.0902, -0.0114, 0.0, -0.0924, -0.0101, 0.0, -0.0902, -0.0101, -0.0017, -0.0924, -0.0101, -0.0039, -0.0924, -0.0101, -0.0017, -0.0902, -0.0101, -0.0039, -0.0902, -0.0101, 0.0, -0.0775, -0.0219, 0.0, -0.0775, -0.0201, -0.0017, -0.0775, -0.0219, -0.0017, -0.0775, -0.0201, -0.0039, -0.0775, -0.0219, -0.0039, -0.0775, -0.0201, 0.0, -0.0751, -0.0219,
0.0, -0.0751, -0.0201, 0.0, -0.0729, -0.0219, 0.0, -0.0729, -0.0201, -0.0017, -0.0751, -0.0219, -0.0039, -0.0751, -0.0219, -0.0017, -0.0729, -0.0219, -0.0039, -0.0729, -0.0219, 0.0, -0.0707, -0.0219, 0.0, -0.0707, -0.0201, 0.0, -0.0686, -0.0219, 0.0, -0.0686, -0.0201, -0.0017, -0.0707, -0.0219, -0.0039, -0.0707, -0.0219, -0.0017, -0.0686, -0.0219, -0.0039, -0.0686, -0.0219, 0.0, -0.0775, -0.0179, 0.0, -0.0775, -0.0157, 0.0, -0.0775, -0.0136, 0.0, -0.0775, -0.0114, -0.0017, -0.0775, -0.0179, -0.0017, -0.0775, -0.0157, -0.0039, -0.0775, -0.0179, -0.0039, -0.0775, -0.0157, -0.0017, -0.0775, -0.0136, -0.0017, -0.0775, -0.0114, -0.0039, -0.0775, -0.0136, -0.0039, -0.0775, -0.0114, 0.0, -0.0775, -0.0092, 0.0, -0.0775, -0.0071, 0.0, -0.0775, -0.0049,
0.0, -0.0775, -0.0027, -0.0017, -0.0775, -0.0092, -0.0017, -0.0775, -0.0071, -0.0039, -0.0775, -0.0092, -0.0039, -0.0775, -0.0071, -0.0017, -0.0775, -0.0049, -0.0017, -0.0775, -0.0027, -0.0039, -0.0775, -0.0049, -0.0039, -0.0775, -0.0027, 0.0, -0.0751, -0.0179, 0.0, -0.0751, -0.0157, 0.0, -0.0729, -0.0179, 0.0, -0.0729, -0.0157, 0.0, -0.0751, -0.0136, 0.0, -0.0751, -0.0114, 0.0, -0.0729, -0.0136, 0.0, -0.0729, -0.0114, 0.0, -0.0707, -0.0179, 0.0, -0.0707, -0.0157, 0.0, -0.0686, -0.0179, 0.0, -0.0686, -0.0157, 0.0, -0.0707, -0.0136, 0.0, -0.0707, -0.0114, 0.0, -0.0686, -0.0136, 0.0, -0.0686, -0.0114, 0.0, -0.0751, -0.0092, 0.0, -0.0751, -0.0071, 0.0, -0.0729, -0.0092, 0.0, -0.0729, -0.0071, 0.0, -0.0751, -0.0049,
0.0, -0.0751, -0.0027, 0.0, -0.0729, -0.0049, 0.0, -0.0729, -0.0027, 0.0, -0.0707, -0.0092, 0.0, -0.0707, -0.0071, 0.0, -0.0686, -0.0092, 0.0, -0.0686, -0.0071, 0.0, -0.0707, -0.0049, 0.0, -0.0707, -0.0027, 0.0, -0.0686, -0.0049, 0.0, -0.0686, -0.0027, 0.0, -0.0659, -0.0669, 0.0, -0.0642, -0.067, 0.0, -0.0667, -0.0668, 0.0, -0.0663, -0.0628, 0.0, -0.0642, -0.0656, 0.0, -0.0642, -0.063, -0.0017, -0.0661, -0.0669, -0.0039, -0.0661, -0.0669, -0.0017, -0.0642, -0.067, -0.0039, -0.0642, -0.067, -0.0017, -0.0672, -0.0667, -0.0017, -0.0664, -0.0627, -0.0039, -0.0672, -0.0667, -0.0039, -0.0664, -0.0627, -0.0017, -0.0642, -0.063, -0.0039, -0.0642, -0.063, 0.0, -0.0621, -0.067, 0.0, -0.0599, -0.067, 0.0, -0.0621, -0.0656,
0.0, -0.062, -0.063, 0.0, -0.0599, -0.0656, 0.0, -0.0599, -0.0631, -0.0017, -0.0621, -0.067, -0.0039, -0.0621, -0.067, -0.0017, -0.0599, -0.067, -0.0039, -0.0599, -0.067, -0.0017, -0.0621, -0.063, -0.0039, -0.0621, -0.063, -0.0017, -0.0599, -0.0631, -0.0039, -0.0599, -0.0631, 0.0, -0.0577, -0.0669, 0.0, -0.0566, -0.0668, 0.0, -0.0577, -0.0656, 0.0, -0.0578, -0.063, 0.0, -0.0555, -0.0666, 0.0, -0.0557, -0.0625, -0.0017, -0.0577, -0.0669, -0.0039, -0.0577, -0.0669, -0.0017, -0.0565, -0.0667, -0.0039, -0.0565, -0.0667, -0.0017, -0.0577, -0.063, -0.0039, -0.0577, -0.063, -0.0017, -0.0554, -0.0666, -0.0017, -0.0559, -0.0626, -0.0039, -0.0554, -0.0666, -0.0039, -0.0559, -0.0626, 0.0, -0.0532, -0.0659, 0.0, -0.0534, -0.0635, 0.0, -0.0515, -0.0652,
0.0, -0.0505, -0.0646, -0.0017, -0.0534, -0.066, -0.0039, -0.0534, -0.066, -0.0017, -0.0514, -0.0651, -0.0017, -0.0503, -0.0645, -0.0039, -0.0514, -0.0651, -0.0039, -0.0503, -0.0645, 0.0, -0.0547, -0.0623, -0.0017, -0.0548, -0.0623, -0.0039, -0.0548, -0.0623, 0.0, -0.0535, -0.0617, 0.0, -0.052, -0.0605, 0.0, -0.0507, -0.059, 0.0, -0.0502, -0.058, -0.0017, -0.0533, -0.0615, -0.0039, -0.0533, -0.0615, -0.0017, -0.052, -0.0605, -0.0017, -0.0509, -0.0592, -0.0039, -0.052, -0.0605, -0.0039, -0.0509, -0.0592, -0.0017, -0.0502, -0.058, -0.0039, -0.0502, -0.058, 0.0, -0.0663, -0.0461, 0.0, -0.0642, -0.0456, -0.0017, -0.0664, -0.0461, -0.0039, -0.0664, -0.0461, -0.0017, -0.0642, -0.0456, -0.0039, -0.0642, -0.0456, 0.0, -0.0621, -0.0454, 0.0, -0.0599, -0.0455,
-0.0017, -0.0621, -0.0454, -0.0039, -0.0621, -0.0454, -0.0017, -0.0599, -0.0455, -0.0039, -0.0599, -0.0455, 0.0, -0.0664, -0.0439, 0.0, -0.0664, -0.0425, 0.0, -0.0642, -0.0439, 0.0, -0.0642, -0.0425, -0.0017, -0.0664, -0.0425, -0.0039, -0.0664, -0.0425, -0.0017, -0.0642, -0.0425, -0.0039, -0.0642, -0.0425, 0.0, -0.0621, -0.0439, 0.0, -0.0621, -0.0425, 0.0, -0.0599, -0.0439, 0.0, -0.0599, -0.0425, -0.0017, -0.0621, -0.0425, -0.0039, -0.0621, -0.0425, -0.0017, -0.0599, -0.0425, -0.0039, -0.0599, -0.0425, 0.0, -0.0578, -0.0457, 0.0, -0.0557, -0.0463, -0.0017, -0.0577, -0.0458, -0.0039, -0.0577, -0.0458, -0.0017, -0.0555, -0.0463, -0.0039, -0.0555, -0.0463, 0.0, -0.0507, -0.0501, 0.0, -0.0528, -0.0477, 0.0, -0.0538, -0.0471, 0.0, -0.0517, -0.0487,
0.0, -0.0512, -0.0461, -0.0017, -0.0506, -0.0502, -0.0039, -0.0506, -0.0502, -0.0017, -0.0529, -0.0476, -0.0017, -0.054, -0.047, -0.0039, -0.0529, -0.0476, -0.0039, -0.054, -0.047, -0.0017, -0.0517, -0.0487, -0.0039, -0.0517, -0.0487, 0.0, -0.0577, -0.0439, 0.0, -0.0577, -0.0425, 0.0, -0.0555, -0.0439, 0.0, -0.0555, -0.0426, -0.0017, -0.0577, -0.0425, -0.0039, -0.0577, -0.0425, -0.0017, -0.0555, -0.0426, -0.0039, -0.0555, -0.0426, 0.0, -0.0533, -0.0431, 0.0, -0.0544, -0.0428, 0.0, -0.0509, -0.0442, -0.0017, -0.0532, -0.0431, -0.0017, -0.0544, -0.0428, -0.0039, -0.0532, -0.0431, -0.0039, -0.0544, -0.0428, -0.0017, -0.0511, -0.044, -0.0039, -0.0511, -0.044, 0.0, -0.049, -0.0636, -0.0017, -0.049, -0.0635, -0.0039, -0.049, -0.0635, 0.0, -0.048, -0.0622,
0.0, -0.049, -0.0591, 0.0, -0.0472, -0.0611, 0.0, -0.0465, -0.0594, 0.0, -0.0497, -0.0569, 0.0, -0.0492, -0.0548, 0.0, -0.0461, -0.0571, 0.0, -0.0461, -0.0548, -0.0017, -0.0481, -0.0623, -0.0039, -0.0481, -0.0623, -0.0017, -0.0473, -0.0613, -0.0017, -0.0465, -0.0596, -0.0039, -0.0473, -0.0613, -0.0039, -0.0465, -0.0596, -0.0017, -0.0497, -0.0569, -0.0017, -0.0492, -0.0547, -0.0039, -0.0497, -0.0569, -0.0039, -0.0492, -0.0547, -0.0017, -0.0461, -0.057, -0.0017, -0.0461, -0.0548, -0.0039, -0.0461, -0.057, -0.0039, -0.0461, -0.0548, 0.0, -0.0495, -0.0528, 0.0, -0.0502, -0.051, 0.0, -0.0461, -0.0526, 0.0, -0.0464, -0.0503, 0.0, -0.049, -0.0483, 0.0, -0.0489, -0.0459, 0.0, -0.0471, -0.048, 0.0, -0.0479, -0.0471, -0.0017, -0.0496, -0.0526,
-0.0017, -0.05, -0.0513, -0.0039, -0.0496, -0.0526, -0.0039, -0.05, -0.0513, -0.0017, -0.0461, -0.0526, -0.0017, -0.0463, -0.0504, -0.0039, -0.0461, -0.0526, -0.0039, -0.0463, -0.0504, -0.0017, -0.0488, -0.0459, -0.0039, -0.0488, -0.0459, -0.0017, -0.047, -0.0482, -0.0017, -0.0479, -0.0471, -0.0039, -0.047, -0.0482, -0.0039, -0.0479, -0.0471, 0.0, -0.05, -0.0449, -0.0017, -0.05, -0.045, -0.0039, -0.05, -0.045, 0.0, -0.0667, -0.0219, 0.0, -0.0667, -0.0201, -0.0017, -0.0667, -0.0219, -0.0017, -0.0667, -0.0201, -0.0039, -0.0667, -0.0219, -0.0039, -0.0667, -0.0201, 0.0, -0.0667, -0.0179, 0.0, -0.0667, -0.0157, 0.0, -0.0667, -0.0136, 0.0, -0.0667, -0.0114, -0.0017, -0.0667, -0.0179, -0.0017, -0.0667, -0.0157, -0.0039, -0.0667, -0.0179, -0.0039, -0.0667, -0.0157,
-0.0017, -0.0667, -0.0136, -0.0017, -0.0667, -0.0114, -0.0039, -0.0667, -0.0136, -0.0039, -0.0667, -0.0114, 0.0, -0.0667, -0.0092, 0.0, -0.0667, -0.0071, 0.0, -0.0667, -0.0049, 0.0, -0.0667, -0.0027, -0.0017, -0.0667, -0.0092, -0.0017, -0.0667, -0.0071, -0.0039, -0.0667, -0.0092, -0.0039, -0.0667, -0.0071, -0.0017, -0.0667, -0.0049, -0.0017, -0.0667, -0.0027, -0.0039, -0.0667, -0.0049, -0.0039, -0.0667, -0.0027, 0.0, -0.05, -0.0219, 0.0, -0.05, -0.0201, 0.0, -0.0469, -0.0219, 0.0, -0.0469, -0.0201, -0.0017, -0.05, -0.0219, -0.0017, -0.05, -0.0201, -0.0039, -0.05, -0.0219, -0.0039, -0.05, -0.0201, -0.0017, -0.0469, -0.0219, -0.0039, -0.0469, -0.0219, 0.0, -0.0447, -0.0219, 0.0, -0.0447, -0.0201, 0.0, -0.0425, -0.0219, 0.0, -0.0425, -0.0201,
-0.0017, -0.0447, -0.0219, -0.0039, -0.0447, -0.0219, -0.0017, -0.0425, -0.0219, -0.0039, -0.0425, -0.0219, 0.0, -0.0404, -0.0219, 0.0, -0.0404, -0.0201, 0.0, -0.0382, -0.0219, 0.0, -0.0382, -0.0201, -0.0017, -0.0404, -0.0219, -0.0039, -0.0404, -0.0219, -0.0017, -0.0382, -0.0219, -0.0017, -0.0382, -0.0201, -0.0039, -0.0382, -0.0219, -0.0039, -0.0382, -0.0201, 0.0, -0.05, -0.0179, 0.0, -0.05, -0.0157, 0.0, -0.0469, -0.0179, 0.0, -0.0469, -0.0157, 0.0, -0.05, -0.0136, 0.0, -0.05, -0.0114, 0.0, -0.0469, -0.0136, 0.0, -0.0469, -0.0114, -0.0017, -0.05, -0.0179, -0.0017, -0.05, -0.0157, -0.0039, -0.05, -0.0179, -0.0039, -0.05, -0.0157, -0.0017, -0.05, -0.0136, -0.0017, -0.05, -0.0114, -0.0039, -0.05, -0.0136, -0.0039, -0.05, -0.0114,
0.0, -0.0447, -0.0179, 0.0, -0.0447, -0.0157, 0.0, -0.0425, -0.0179, 0.0, -0.0425, -0.0157, 0.0, -0.0447, -0.0136, 0.0, -0.0447, -0.0114, 0.0, -0.0425, -0.0136, 0.0, -0.0425, -0.0114, 0.0, -0.05, -0.0092, 0.0, -0.05, -0.0071, 0.0, -0.0469, -0.0092, 0.0, -0.0469, -0.0071, 0.0, -0.05, -0.0049, 0.0, -0.05, -0.0027, 0.0, -0.0469, -0.0049, 0.0, -0.0469, -0.0027, -0.0017, -0.05, -0.0092, -0.0017, -0.05, -0.0071, -0.0039, -0.05, -0.0092, -0.0039, -0.05, -0.0071, -0.0017, -0.05, -0.0049, -0.0017, -0.05, -0.0027, -0.0039, -0.05, -0.0049, -0.0039, -0.05, -0.0027, 0.0, -0.0447, -0.0092, 0.0, -0.0447, -0.0071, 0.0, -0.0425, -0.0092, 0.0, -0.0425, -0.0071, 0.0, -0.0447, -0.0049, 0.0, -0.0447, -0.0027,
0.0, -0.0425, -0.0049, 0.0, -0.0425, -0.0027, 0.0, -0.0404, -0.0179, 0.0, -0.0404, -0.0157, 0.0, -0.0382, -0.0179, 0.0, -0.0382, -0.0157, 0.0, -0.0404, -0.0136, 0.0, -0.0404, -0.0114, 0.0, -0.0382, -0.0136, 0.0, -0.0382, -0.0114, -0.0017, -0.0382, -0.0179, -0.0017, -0.0382, -0.0157, -0.0039, -0.0382, -0.0179, -0.0039, -0.0382, -0.0157, -0.0017, -0.0382, -0.0136, -0.0017, -0.0382, -0.0114, -0.0039, -0.0382, -0.0136, -0.0039, -0.0382, -0.0114, 0.0, -0.0404, -0.0092, 0.0, -0.0404, -0.0071, 0.0, -0.0382, -0.0092, 0.0, -0.0382, -0.0071, 0.0, -0.0404, -0.0049, 0.0, -0.0404, -0.0027, 0.0, -0.0382, -0.0049, 0.0, -0.0382, -0.0027, -0.0017, -0.0382, -0.0092, -0.0017, -0.0382, -0.0071, -0.0039, -0.0382, -0.0092, -0.0039, -0.0382, -0.0071,
-0.0017, -0.0382, -0.0049, -0.0017, -0.0382, -0.0027, -0.0039, -0.0382, -0.0049, -0.0039, -0.0382, -0.0027, 0.0, -0.0265, -0.067, 0.0, -0.0252, -0.067, 0.0, -0.0265, -0.0656, 0.0, -0.0265, -0.0635, 0.0, -0.0252, -0.0656, 0.0, -0.0252, -0.0635, -0.0017, -0.0265, -0.067, -0.0039, -0.0265, -0.067, -0.0017, -0.0252, -0.067, -0.0039, -0.0252, -0.067, -0.0017, -0.0265, -0.0656, -0.0017, -0.0265, -0.0635, -0.0039, -0.0265, -0.0656, -0.0039, -0.0265, -0.0635, 0.0, -0.0265, -0.0613, 0.0, -0.0265, -0.0591, 0.0, -0.0252, -0.0613, 0.0, -0.0252, -0.0591, 0.0, -0.0265, -0.057, 0.0, -0.0265, -0.0548, 0.0, -0.0252, -0.057, 0.0, -0.0252, -0.0548, -0.0017, -0.0265, -0.0613, -0.0017, -0.0265, -0.0591, -0.0039, -0.0265, -0.0613, -0.0039, -0.0265, -0.0591,
-0.0017, -0.0265, -0.057, -0.0017, -0.0265, -0.0548, -0.0039, -0.0265, -0.057, -0.0039, -0.0265, -0.0548, 0.0, -0.0225, -0.067, 0.0, -0.0225, -0.0656, 0.0, -0.0225, -0.0635, -0.0017, -0.0225, -0.067, -0.0039, -0.0225, -0.067, -0.0017, -0.0225, -0.0656, -0.0017, -0.0225, -0.0635, -0.0039, -0.0225, -0.0656, -0.0039, -0.0225, -0.0635, 0.0, -0.0225, -0.0613, 0.0, -0.0225, -0.0591, 0.0, -0.0225, -0.057, 0.0, -0.0225, -0.0548, -0.0017, -0.0225, -0.0613, -0.0017, -0.0225, -0.0591, -0.0039, -0.0225, -0.0613, -0.0039, -0.0225, -0.0591, -0.0017, -0.0225, -0.057, -0.0017, -0.0225, -0.0548, -0.0039, -0.0225, -0.057, -0.0039, -0.0225, -0.0548, 0.0, -0.0179, -0.054, 0.0, -0.0157, -0.0562, 0.0, -0.0168, -0.0551, -0.0017, -0.0179, -0.054, -0.0039, -0.0179, -0.054,
-0.0017, -0.0157, -0.0562, -0.0017, -0.0168, -0.0551, -0.0039, -0.0157, -0.0562, -0.0039, -0.0168, -0.0551, 0.0, -0.0265, -0.0526, 0.0, -0.0265, -0.0505, 0.0, -0.0252, -0.0526, 0.0, -0.0252, -0.0505, 0.0, -0.0265, -0.0483, 0.0, -0.0264, -0.0468, 0.0, -0.0252, -0.0483, 0.0, -0.0262, -0.0459, -0.0017, -0.0265, -0.0526, -0.0017, -0.0265, -0.0505, -0.0039, -0.0265, -0.0526, -0.0039, -0.0265, -0.0505, -0.0017, -0.0265, -0.0483, -0.0017, -0.0264, -0.0466, -0.0039, -0.0265, -0.0483, -0.0039, -0.0264, -0.0466, -0.0017, -0.0262, -0.0455, -0.0039, -0.0262, -0.0455, 0.0, -0.0255, -0.0437, -0.0017, -0.0255, -0.0437, -0.0039, -0.0255, -0.0437, 0.0, -0.0225, -0.0526, 0.0, -0.0225, -0.0503, 0.0, -0.0202, -0.052, 0.0, -0.0213, -0.0511, 0.0, -0.023, -0.0483,
0.0, -0.023, -0.0461, 0.0, -0.0208, -0.0483, 0.0, -0.0201, -0.0452, -0.0017, -0.0225, -0.0526, -0.0017, -0.0225, -0.0503, -0.0039, -0.0225, -0.0526, -0.0039, -0.0225, -0.0503, -0.0017, -0.0202, -0.052, -0.0017, -0.0213, -0.0511, -0.0039, -0.0202, -0.052, -0.0039, -0.0213, -0.0511, -0.0017, -0.02, -0.0452, -0.0039, -0.02, -0.0452, 0.0, -0.019, -0.053, 0.0, -0.0187, -0.0505, 0.0, -0.0165, -0.0526, 0.0, -0.0155, -0.0495, 0.0, -0.0177, -0.0473, 0.0, -0.0188, -0.0463, 0.0, -0.0166, -0.0484, -0.0017, -0.019, -0.053, -0.0039, -0.019, -0.053, -0.0017, -0.0155, -0.0495, -0.0039, -0.0155, -0.0495, -0.0017, -0.0177, -0.0473, -0.0017, -0.0188, -0.0463, -0.0039, -0.0177, -0.0473, -0.0039, -0.0188, -0.0463, -0.0017, -0.0166, -0.0484, -0.0039, -0.0166, -0.0484,
0.0, -0.0231, -0.0433, 0.0, -0.0211, -0.0443, -0.0017, -0.0231, -0.0433, -0.0039, -0.0231, -0.0433, -0.0017, -0.021, -0.0444, -0.0039, -0.021, -0.0444, 0.0, -0.0092, -0.0627, 0.0, -0.0069, -0.0647, 0.0, -0.008, -0.0638, -0.0017, -0.0092, -0.0627, -0.0039, -0.0092, -0.0627, -0.0017, -0.0069, -0.0647, -0.0017, -0.008, -0.0638, -0.0039, -0.0069, -0.0647, -0.0039, -0.008, -0.0638, 0.0, -0.0135, -0.0584, 0.0, -0.0114, -0.0605, 0.0, -0.0124, -0.0594, 0.0, -0.0146, -0.0573, 0.0, -0.0143, -0.0548, 0.0, -0.0121, -0.057, 0.0, -0.0112, -0.0538, -0.0017, -0.0135, -0.0584, -0.0039, -0.0135, -0.0584, -0.0017, -0.0114, -0.0605, -0.0017, -0.0124, -0.0594, -0.0039, -0.0114, -0.0605, -0.0039, -0.0124, -0.0594, -0.0017, -0.0146, -0.0573, -0.0039, -0.0146, -0.0573,
-0.0017, -0.0112, -0.0538, -0.0039, -0.0112, -0.0538, 0.0, -0.0103, -0.0616, 0.0, -0.01, -0.0591, 0.0, -0.0078, -0.0613, 0.0, -0.0068, -0.0582, 0.0, -0.009, -0.056, 0.0, -0.0101, -0.0549, 0.0, -0.0079, -0.0571, -0.0017, -0.0103, -0.0616, -0.0039, -0.0103, -0.0616, -0.0017, -0.0068, -0.0582, -0.0039, -0.0068, -0.0582, -0.0017, -0.009, -0.056, -0.0017, -0.0101, -0.0549, -0.0039, -0.009, -0.056, -0.0039, -0.0101, -0.0549, -0.0017, -0.0079, -0.0571, -0.0039, -0.0079, -0.0571, 0.0, -0.0026, -0.0668, 0.0, -0.0059, -0.0655, 0.0, -0.0056, -0.0635, 0.0, -0.0037, -0.0666, 0.0, -0.0035, -0.0635, -0.0017, -0.0028, -0.0668, -0.0039, -0.0028, -0.0668, -0.0017, -0.0057, -0.0656, -0.0039, -0.0057, -0.0656, -0.0017, -0.0039, -0.0666, -0.0039, -0.0039, -0.0666,
0.0, -0.002, -0.067, 0.0, -0.002, -0.0656, 0.0, -0.002, -0.0635, -0.0017, -0.002, -0.067, -0.0039, -0.002, -0.067, -0.0017, -0.002, -0.0656, -0.0017, -0.002, -0.0635, -0.0039, -0.002, -0.0656, -0.0039, -0.002, -0.0635, 0.0, -0.0056, -0.0613, 0.0, -0.0059, -0.0591, 0.0, -0.0035, -0.0613, 0.0, -0.0035, -0.0591, 0.0, -0.0059, -0.057, 0.0, -0.0059, -0.0548, 0.0, -0.0035, -0.057, 0.0, -0.0035, -0.0548, -0.0017, -0.0059, -0.0591, -0.0039, -0.0059, -0.0591, -0.0017, -0.0059, -0.057, -0.0017, -0.0059, -0.0548, -0.0039, -0.0059, -0.057, -0.0039, -0.0059, -0.0548, 0.0, -0.002, -0.0613, 0.0, -0.002, -0.0591, 0.0, -0.002, -0.057, 0.0, -0.002, -0.0548, -0.0017, -0.002, -0.0613, -0.0017, -0.002, -0.0591, -0.0039, -0.002, -0.0613,
-0.0039, -0.002, -0.0591, -0.0017, -0.002, -0.057, -0.0017, -0.002, -0.0548, -0.0039, -0.002, -0.057, -0.0039, -0.002, -0.0548, 0.0, -0.0134, -0.0517, 0.0, -0.0144, -0.0506, 0.0, -0.0123, -0.0528, -0.0017, -0.0134, -0.0517, -0.0017, -0.0144, -0.0506, -0.0039, -0.0134, -0.0517, -0.0039, -0.0144, -0.0506, -0.0017, -0.0123, -0.0528, -0.0039, -0.0123, -0.0528, 0.0, -0.0059, -0.0526, 0.0, -0.0059, -0.0505, 0.0, -0.0035, -0.0526, 0.0, -0.0035, -0.0505, 0.0, -0.0059, -0.0483, 0.0, -0.0059, -0.0461, 0.0, -0.0035, -0.0483, 0.0, -0.0035, -0.0461, -0.0017, -0.0059, -0.0526, -0.0017, -0.0059, -0.0505, -0.0039, -0.0059, -0.0526, -0.0039, -0.0059, -0.0505, -0.0017, -0.0059, -0.0483, -0.0017, -0.0059, -0.0461, -0.0039, -0.0059, -0.0483, -0.0039, -0.0059, -0.0461,
0.0, -0.002, -0.0526, 0.0, -0.002, -0.0505, 0.0, -0.002, -0.0483, 0.0, -0.002, -0.0461, -0.0017, -0.002, -0.0526, -0.0017, -0.002, -0.0505, -0.0039, -0.002, -0.0526, -0.0039, -0.002, -0.0505, -0.0017, -0.002, -0.0483, -0.0017, -0.002, -0.0461, -0.0039, -0.002, -0.0483, -0.0039, -0.002, -0.0461, 0.0, -0.0059, -0.0439, 0.0, -0.0059, -0.0425, 0.0, -0.0035, -0.0439, 0.0, -0.0035, -0.0425, -0.0017, -0.0059, -0.0439, -0.0017, -0.0059, -0.0425, -0.0039, -0.0059, -0.0439, -0.0039, -0.0059, -0.0425, -0.0017, -0.0035, -0.0425, -0.0039, -0.0035, -0.0425, 0.0, -0.002, -0.0439, 0.0, -0.002, -0.0425, -0.0017, -0.002, -0.0439, -0.0017, -0.002, -0.0425, -0.0039, -0.002, -0.0439, -0.0039, -0.002, -0.0425, 0.0, -0.4807, -0.0004, 0.0, -0.4802, 0.0013,
-0.0017, -0.4807, -0.0005, -0.0017, -0.4801, 0.0016, -0.0039, -0.4807, -0.0005, -0.0039, -0.4801, 0.0016, 0.0, -0.4787, -0.0005, 0.0, -0.48, 0.0021, 0.0, -0.4766, -0.0005, 0.0, -0.4766, 0.0016, 0.0, -0.4795, 0.004, 0.0, -0.4789, 0.0061, 0.0, -0.4766, 0.0038, 0.0, -0.4766, 0.006, -0.0017, -0.4798, 0.0027, -0.0039, -0.4798, 0.0027, -0.0017, -0.4795, 0.0038, -0.0017, -0.4789, 0.006, -0.0039, -0.4795, 0.0038, -0.0039, -0.4789, 0.006, 0.0, -0.4784, 0.0082, 0.0, -0.478, 0.0099, 0.0, -0.4766, 0.0081, 0.0, -0.4777, 0.0107, 0.0, -0.4772, 0.0127, 0.0, -0.4765, 0.0148, -0.0017, -0.4784, 0.0081, -0.0017, -0.4779, 0.0102, -0.0039, -0.4784, 0.0081, -0.0039, -0.4779, 0.0102, -0.0017, -0.4776, 0.0112, -0.0039, -0.4776, 0.0112,
-0.0017, -0.4772, 0.0125, -0.0017, -0.4765, 0.0147, -0.0039, -0.4772, 0.0125, -0.0039, -0.4765, 0.0147, 0.0, -0.4744, -0.0005, 0.0, -0.4744, 0.0016, 0.0, -0.4722, -0.0005, 0.0, -0.4722, 0.0016, 0.0, -0.4744, 0.0038, 0.0, -0.4744, 0.006, 0.0, -0.4722, 0.0038, 0.0, -0.4722, 0.006, 0.0, -0.47, -0.0005, 0.0, -0.47, 0.0016, 0.0, -0.4679, -0.0005, 0.0, -0.4679, 0.0016, 0.0, -0.47, 0.0038, 0.0, -0.47, 0.006, 0.0, -0.4679, 0.0038, 0.0, -0.4679, 0.006, 0.0, -0.4744, 0.0081, 0.0, -0.4744, 0.0103, 0.0, -0.4722, 0.0081, 0.0, -0.4722, 0.0103, 0.0, -0.4744, 0.0125, 0.0, -0.4744, 0.0146, 0.0, -0.4722, 0.0125, 0.0, -0.4722, 0.0146, 0.0, -0.47, 0.0081, 0.0, -0.47, 0.0103,
0.0, -0.4679, 0.0081, 0.0, -0.4679, 0.0103, 0.0, -0.47, 0.0125, 0.0, -0.47, 0.0146, 0.0, -0.4679, 0.0125, 0.0, -0.4679, 0.0146, 0.0, -0.4759, 0.0164, -0.0017, -0.4758, 0.0166, -0.0039, -0.4758, 0.0166, 0.0, -0.4755, 0.0173, 0.0, -0.4749, 0.0192, 0.0, -0.4722, 0.0168, 0.0, -0.4722, 0.019, 0.0, -0.4742, 0.0213, 0.0, -0.4736, 0.0228, 0.0, -0.4722, 0.0212, 0.0, -0.4733, 0.0237, -0.0017, -0.4754, 0.0177, -0.0017, -0.4749, 0.019, -0.0039, -0.4754, 0.0177, -0.0039, -0.4749, 0.019, -0.0017, -0.4742, 0.0212, -0.0017, -0.4736, 0.0229, -0.0039, -0.4742, 0.0212, -0.0039, -0.4736, 0.0229, -0.0017, -0.4732, 0.024, -0.0039, -0.4732, 0.024, 0.0, -0.47, 0.0168, 0.0, -0.47, 0.019, 0.0, -0.4679, 0.0168,
0.0, -0.4679, 0.019, 0.0, -0.47, 0.0212, 0.0, -0.47, 0.0233, 0.0, -0.4679, 0.0212, 0.0, -0.4679, 0.0233, 0.0, -0.4724, 0.0258, 0.0, -0.4716, 0.0278, 0.0, -0.4711, 0.0288, -0.0017, -0.4726, 0.0255, -0.0017, -0.4716, 0.0277, -0.0039, -0.4726, 0.0255, -0.0039, -0.4716, 0.0277, -0.0017, -0.4711, 0.0288, -0.0039, -0.4711, 0.0288, 0.0, -0.47, 0.0255, 0.0, -0.47, 0.0277, 0.0, -0.4679, 0.0255, 0.0, -0.4679, 0.0277, 0.0, -0.4707, 0.0299, 0.0, -0.4697, 0.0322, 0.0, -0.4679, 0.0298, 0.0, -0.4679, 0.032, -0.0017, -0.4707, 0.0299, -0.0017, -0.4698, 0.032, -0.0039, -0.4707, 0.0299, -0.0039, -0.4698, 0.032, 0.0, -0.4657, -0.0005, 0.0, -0.4657, 0.0016, 0.0, -0.4628, -0.0008, 0.0, -0.4623, 0.0012,
0.0, -0.4657, 0.0038, 0.0, -0.4657, 0.006, 0.0, -0.4635, 0.0038, 0.0, -0.4635, 0.006, -0.0017, -0.4628, -0.0005, -0.0017, -0.4625, 0.0006, -0.0039, -0.4628, -0.0005, -0.0039, -0.4625, 0.0006, 0.0, -0.4621, 0.002, 0.0, -0.4616, 0.0037, 0.0, -0.4609, 0.0057, -0.0017, -0.4622, 0.0017, -0.0039, -0.4622, 0.0017, -0.0017, -0.4615, 0.0038, -0.0017, -0.4608, 0.006, -0.0039, -0.4615, 0.0038, -0.0039, -0.4608, 0.006, 0.0, -0.4657, 0.0081, 0.0, -0.4657, 0.0103, 0.0, -0.4635, 0.0081, 0.0, -0.4635, 0.0103, 0.0, -0.4657, 0.0125, 0.0, -0.4657, 0.0146, 0.0, -0.4635, 0.0125, 0.0, -0.4635, 0.0146, 0.0, -0.4603, 0.0078, 0.0, -0.4614, 0.0103, 0.0, -0.46, 0.0087, 0.0, -0.4595, 0.0102, 0.0, -0.4614, 0.0125,
0.0, -0.4614, 0.0146, 0.0, -0.4588, 0.0122, 0.0, -0.4581, 0.0142, -0.0017, -0.4604, 0.0074, -0.0039, -0.4604, 0.0074, -0.0017, -0.4601, 0.0085, -0.0017, -0.4595, 0.0103, -0.0039, -0.4601, 0.0085, -0.0039, -0.4595, 0.0103, -0.0017, -0.4587, 0.0125, -0.0017, -0.4582, 0.0138, -0.0039, -0.4587, 0.0125, -0.0039, -0.4582, 0.0138, 0.0, -0.4577, 0.0151, -0.0017, -0.4578, 0.0149, -0.0039, -0.4578, 0.0149, 0.0, -0.4657, 0.0168, 0.0, -0.4657, 0.019, 0.0, -0.4635, 0.0168, 0.0, -0.4635, 0.019, 0.0, -0.4657, 0.0212, 0.0, -0.4657, 0.0233, 0.0, -0.4635, 0.0212, 0.0, -0.4635, 0.0233, 0.0, -0.4614, 0.0168, 0.0, -0.4614, 0.019, 0.0, -0.4592, 0.0168, 0.0, -0.4592, 0.019, 0.0, -0.4614, 0.0212, 0.0, -0.4614, 0.0233,
0.0, -0.4592, 0.0212, 0.0, -0.4592, 0.0233, 0.0, -0.4657, 0.0255, 0.0, -0.4657, 0.0277, 0.0, -0.4635, 0.0255, 0.0, -0.4635, 0.0277, 0.0, -0.4657, 0.0298, 0.0, -0.4657, 0.032, 0.0, -0.4635, 0.0298, 0.0, -0.4635, 0.032, 0.0, -0.4614, 0.0255, 0.0, -0.4614, 0.0277, 0.0, -0.4592, 0.0255, 0.0, -0.4592, 0.0277, 0.0, -0.4614, 0.0298, 0.0, -0.4614, 0.032, 0.0, -0.4592, 0.0298, 0.0, -0.4592, 0.032, 0.0, -0.4572, 0.0167, 0.0, -0.4564, 0.0187, 0.0, -0.4558, 0.0206, 0.0, -0.457, 0.0233, 0.0, -0.4555, 0.0214, 0.0, -0.4549, 0.0231, -0.0017, -0.4571, 0.0168, -0.0017, -0.4563, 0.019, -0.0039, -0.4571, 0.0168, -0.0039, -0.4563, 0.019, -0.0017, -0.4559, 0.0201, -0.0039, -0.4559, 0.0201,
-0.0017, -0.4556, 0.0212, -0.0017, -0.4548, 0.0233, -0.0039, -0.4556, 0.0212, -0.0039, -0.4548, 0.0233, 0.0, -0.457, 0.0255, 0.0, -0.457, 0.0277, 0.0, -0.454, 0.0252, 0.0, -0.4549, 0.0277, 0.0, -0.457, 0.0298, 0.0, -0.457, 0.032, 0.0, -0.4549, 0.0298, 0.0, -0.4549, 0.032, -0.0017, -0.4541, 0.0251, -0.0039, -0.4541, 0.0251, 0.0, -0.4536, 0.0263, 0.0, -0.453, 0.0275, 0.0, -0.4519, 0.0296, 0.0, -0.4527, 0.032, 0.0, -0.4514, 0.0307, 0.0, -0.4508, 0.0318, -0.0017, -0.4536, 0.0262, -0.0017, -0.4529, 0.0277, -0.0039, -0.4536, 0.0262, -0.0039, -0.4529, 0.0277, -0.0017, -0.452, 0.0295, -0.0039, -0.452, 0.0295, -0.0017, -0.4514, 0.0306, -0.0017, -0.4507, 0.032, -0.0039, -0.4514, 0.0306, -0.0039, -0.4507, 0.032,
0.0, -0.4691, 0.0333, 0.0, -0.4686, 0.0344, 0.0, -0.4676, 0.0365, 0.0, -0.467, 0.0377, -0.0017, -0.4691, 0.0334, -0.0039, -0.4691, 0.0334, -0.0017, -0.4686, 0.0345, -0.0017, -0.4677, 0.0364, -0.0039, -0.4686, 0.0345, -0.0039, -0.4677, 0.0364, -0.0017, -0.467, 0.0378, -0.0039, -0.467, 0.0378, 0.0, -0.4657, 0.0342, 0.0, -0.4657, 0.0363, 0.0, -0.4635, 0.0342, 0.0, -0.4635, 0.0363, 0.0, -0.4664, 0.0388, 0.0, -0.4654, 0.0409, 0.0, -0.4635, 0.0385, 0.0, -0.4635, 0.0407, -0.0017, -0.4664, 0.0389, -0.0017, -0.4655, 0.0407, -0.0039, -0.4664, 0.0389, -0.0039, -0.4655, 0.0407, 0.0, -0.4614, 0.0342, 0.0, -0.4614, 0.0363, 0.0, -0.4592, 0.0342, 0.0, -0.4592, 0.0363, 0.0, -0.4614, 0.0385, 0.0, -0.4614, 0.0407,
0.0, -0.4592, 0.0385, 0.0, -0.4592, 0.0407, 0.0, -0.4648, 0.042, 0.0, -0.4641, 0.0431, 0.0, -0.4631, 0.0448, -0.0017, -0.4647, 0.042, -0.0039, -0.4647, 0.042, -0.0017, -0.4641, 0.0431, -0.0017, -0.463, 0.0449, -0.0039, -0.4641, 0.0431, -0.0039, -0.463, 0.0449, 0.0, -0.4614, 0.0429, 0.0, -0.4625, 0.0458, 0.0, -0.4592, 0.0429, 0.0, -0.4592, 0.045, 0.0, -0.4615, 0.0475, 0.0, -0.4607, 0.0488, 0.0, -0.4592, 0.0472, 0.0, -0.46, 0.0498, -0.0017, -0.4624, 0.046, -0.0039, -0.4624, 0.046, -0.0017, -0.4617, 0.0472, -0.0017, -0.4606, 0.0488, -0.0039, -0.4617, 0.0472, -0.0039, -0.4606, 0.0488, -0.0017, -0.4599, 0.0499, -0.0039, -0.4599, 0.0499, 0.0, -0.457, 0.0342, 0.0, -0.457, 0.0363, 0.0, -0.4549, 0.0342,
0.0, -0.4549, 0.0363, 0.0, -0.457, 0.0385, 0.0, -0.457, 0.0407, 0.0, -0.4549, 0.0385, 0.0, -0.4549, 0.0407, 0.0, -0.4527, 0.0342, 0.0, -0.4527, 0.0363, 0.0, -0.4498, 0.0339, 0.0, -0.4505, 0.0363, 0.0, -0.4527, 0.0385, 0.0, -0.4515, 0.0415, 0.0, -0.4499, 0.0393, 0.0, -0.4505, 0.0408, -0.0017, -0.4498, 0.0339, -0.0039, -0.4498, 0.0339, -0.0017, -0.4517, 0.0417, -0.0039, -0.4517, 0.0417, -0.0017, -0.4498, 0.0392, -0.0017, -0.4505, 0.0408, -0.0039, -0.4498, 0.0392, -0.0039, -0.4505, 0.0408, 0.0, -0.457, 0.0429, 0.0, -0.457, 0.045, 0.0, -0.4549, 0.0429, 0.0, -0.4549, 0.045, 0.0, -0.457, 0.0472, 0.0, -0.457, 0.0494, 0.0, -0.4549, 0.0472, 0.0, -0.4549, 0.0494, 0.0, -0.4525, 0.0431,
0.0, -0.452, 0.0453, 0.0, -0.451, 0.0425, 0.0, -0.4515, 0.045, 0.0, -0.452, 0.0468, 0.0, -0.4523, 0.0492, 0.0, -0.4516, 0.0472, 0.0, -0.4508, 0.0496, -0.0017, -0.4523, 0.0426, -0.0017, -0.452, 0.0452, -0.0039, -0.4523, 0.0426, -0.0039, -0.452, 0.0452, -0.0017, -0.4512, 0.0429, -0.0017, -0.4515, 0.0447, -0.0039, -0.4512, 0.0429, -0.0039, -0.4515, 0.0447, -0.0017, -0.4519, 0.0469, -0.0017, -0.4522, 0.0494, -0.0039, -0.4519, 0.0469, -0.0039, -0.4522, 0.0494, -0.0017, -0.4516, 0.0473, -0.0017, -0.4509, 0.0494, -0.0039, -0.4516, 0.0473, -0.0039, -0.4509, 0.0494, 0.0, -0.459, 0.0514, -0.0017, -0.4589, 0.0515, -0.0039, -0.4589, 0.0515, 0.0, -0.4582, 0.0524, 0.0, -0.457, 0.054, 0.0, -0.4549, 0.0515, 0.0, -0.4549, 0.0537,
0.0, -0.4562, 0.0551, 0.0, -0.4553, 0.0562, 0.0, -0.4543, 0.0575, -0.0017, -0.4581, 0.0526, -0.0017, -0.4573, 0.0537, -0.0039, -0.4581, 0.0526, -0.0039, -0.4573, 0.0537, -0.0017, -0.4562, 0.0551, -0.0039, -0.4562, 0.0551, -0.0017, -0.4553, 0.0562, -0.0017, -0.4543, 0.0576, -0.0039, -0.4553, 0.0562, -0.0039, -0.4543, 0.0576, 0.0, -0.4515, 0.0508, 0.0, -0.4527, 0.0537, 0.0, -0.4504, 0.0514, 0.0, -0.4498, 0.0529, 0.0, -0.4527, 0.0559, 0.0, -0.4534, 0.0586, 0.0, -0.4505, 0.0559, 0.0, -0.4505, 0.058, -0.0017, -0.4517, 0.0505, -0.0039, -0.4517, 0.0505, -0.0017, -0.4499, 0.0521, -0.0017, -0.4497, 0.0529, -0.0039, -0.4499, 0.0521, -0.0039, -0.4497, 0.0529, -0.0017, -0.4534, 0.0586, -0.0039, -0.4534, 0.0586, 0.0, -0.4523, 0.0599,
0.0, -0.4513, 0.061, 0.0, -0.4502, 0.0621, -0.0017, -0.4523, 0.0599, -0.0039, -0.4523, 0.0599, -0.0017, -0.4513, 0.061, -0.0017, -0.4502, 0.0621, -0.0039, -0.4513, 0.061, -0.0039, -0.4502, 0.0621, 0.0, -0.4323, 0.0046, 0.0, -0.4328, 0.0058, -0.0017, -0.4322, 0.0045, -0.0017, -0.4328, 0.0059, -0.0039, -0.4322, 0.0045, -0.0039, -0.4328, 0.0059, 0.0, -0.4332, 0.0081, 0.0, -0.4333, 0.0103, 0.0, -0.4333, 0.0125, 0.0, -0.4331, 0.0147, -0.0017, -0.4332, 0.0081, -0.0017, -0.4333, 0.0103, -0.0039, -0.4332, 0.0081, -0.0039, -0.4333, 0.0103, -0.0017, -0.4333, 0.0125, -0.0017, -0.4331, 0.0147, -0.0039, -0.4333, 0.0125, -0.0039, -0.4331, 0.0147, 0.0, -0.4326, 0.0165, -0.0017, -0.4325, 0.0168, -0.0039, -0.4325, 0.0168, 0.0, -0.4306, 0.002,
0.0, -0.4284, -0.0001, 0.0, -0.4295, 0.0008, 0.0, -0.4317, 0.0035, 0.0, -0.431, 0.006, 0.0, -0.4288, 0.0038, 0.0, -0.4288, 0.006, -0.0017, -0.4306, 0.0019, -0.0039, -0.4306, 0.0019, -0.0017, -0.4285, -0.0001, -0.0017, -0.4296, 0.0009, -0.0039, -0.4285, -0.0001, -0.0039, -0.4296, 0.0009, -0.0017, -0.4317, 0.0034, -0.0039, -0.4317, 0.0034, 0.0, -0.4269, -0.0012, 0.0, -0.4266, 0.0016, 0.0, -0.4245, -0.0005, 0.0, -0.4245, 0.0016, 0.0, -0.4266, 0.0038, 0.0, -0.4266, 0.006, 0.0, -0.4245, 0.0038, 0.0, -0.4245, 0.006, -0.0017, -0.4269, -0.0012, -0.0039, -0.4269, -0.0012, 0.0, -0.431, 0.0081, 0.0, -0.431, 0.0103, 0.0, -0.4288, 0.0081, 0.0, -0.4288, 0.0103, 0.0, -0.431, 0.0125, 0.0, -0.431, 0.0146,
0.0, -0.4288, 0.0125, 0.0, -0.4288, 0.0146, 0.0, -0.4266, 0.0081, 0.0, -0.4266, 0.0103, 0.0, -0.4245, 0.0081, 0.0, -0.4245, 0.0103, 0.0, -0.4266, 0.0125, 0.0, -0.4266, 0.0146, 0.0, -0.4245, 0.0125, 0.0, -0.4245, 0.0146, 0.0, -0.4223, -0.0005, 0.0, -0.4223, 0.0016, 0.0, -0.4201, -0.0005, 0.0, -0.4201, 0.0016, 0.0, -0.4223, 0.0038, 0.0, -0.4223, 0.006, 0.0, -0.4201, 0.0038, 0.0, -0.4201, 0.006, 0.0, -0.418, -0.0005, 0.0, -0.418, 0.0016, 0.0, -0.4158, -0.0005, 0.0, -0.4158, 0.0016, 0.0, -0.418, 0.0038, 0.0, -0.418, 0.006, 0.0, -0.4158, 0.0038, 0.0, -0.4158, 0.006, 0.0, -0.4223, 0.0081, 0.0, -0.4223, 0.0103, 0.0, -0.4201, 0.0081, 0.0, -0.4201, 0.0103,
0.0, -0.4223, 0.0125, 0.0, -0.4223, 0.0146, 0.0, -0.4201, 0.0125, 0.0, -0.4201, 0.0146, 0.0, -0.418, 0.0081, 0.0, -0.418, 0.0103, 0.0, -0.4158, 0.0081, 0.0, -0.4158, 0.0103, 0.0, -0.418, 0.0125, 0.0, -0.418, 0.0146, 0.0, -0.4158, 0.0125, 0.0, -0.4158, 0.0146, 0.0, -0.4322, 0.0175, 0.0, -0.4314, 0.0194, 0.0, -0.4288, 0.0168, 0.0, -0.4288, 0.019, 0.0, -0.4304, 0.0206, 0.0, -0.4293, 0.0217, 0.0, -0.4281, 0.0226, -0.0017, -0.4321, 0.0179, -0.0017, -0.4315, 0.0191, -0.0039, -0.4321, 0.0179, -0.0039, -0.4315, 0.0191, -0.0017, -0.4304, 0.0206, -0.0039, -0.4304, 0.0206, -0.0017, -0.4293, 0.0217, -0.0017, -0.4282, 0.0226, -0.0039, -0.4293, 0.0217, -0.0039, -0.4282, 0.0226, 0.0, -0.4266, 0.0168,
0.0, -0.4266, 0.019, 0.0, -0.4245, 0.0168, 0.0, -0.4245, 0.019, 0.0, -0.4266, 0.0212, 0.0, -0.427, 0.0234, 0.0, -0.4245, 0.0212, 0.0, -0.4247, 0.0243, -0.0017, -0.4267, 0.0236, -0.0039, -0.4267, 0.0236, -0.0017, -0.4249, 0.0242, -0.0039, -0.4249, 0.0242, 0.0, -0.4237, 0.0246, -0.0017, -0.4238, 0.0246, -0.0039, -0.4238, 0.0246, 0.0, -0.4223, 0.0168, 0.0, -0.4223, 0.019, 0.0, -0.4201, 0.0168, 0.0, -0.4201, 0.019, 0.0, -0.4223, 0.0212, 0.0, -0.4223, 0.0233, 0.0, -0.4201, 0.0212, 0.0, -0.4201, 0.0233, 0.0, -0.418, 0.0168, 0.0, -0.418, 0.019, 0.0, -0.4158, 0.0168, 0.0, -0.4158, 0.019, 0.0, -0.418, 0.0212, 0.0, -0.418, 0.0233, 0.0, -0.4158, 0.0212, 0.0, -0.4153, 0.0246,
-0.0017, -0.4148, 0.0244, -0.0039, -0.4148, 0.0244, 0.0, -0.4223, 0.0249, 0.0, -0.4201, 0.0251, -0.0017, -0.4223, 0.0249, -0.0039, -0.4223, 0.0249, -0.0017, -0.4201, 0.0251, -0.0039, -0.4201, 0.0251, 0.0, -0.4179, 0.025, 0.0, -0.4162, 0.0247, -0.0017, -0.418, 0.025, -0.0039, -0.418, 0.025, -0.0017, -0.4159, 0.0247, -0.0039, -0.4159, 0.0247, 0.0, -0.4493, 0.035, 0.0, -0.4487, 0.0362, 0.0, -0.4478, 0.0387, 0.0, -0.4483, 0.0407, 0.0, -0.4463, 0.0385, 0.0, -0.4464, 0.0407, -0.0017, -0.4493, 0.035, -0.0017, -0.4487, 0.0363, -0.0039, -0.4493, 0.035, -0.0039, -0.4487, 0.0363, -0.0017, -0.4478, 0.0387, -0.0017, -0.4483, 0.0407, -0.0039, -0.4478, 0.0387, -0.0039, -0.4483, 0.0407, -0.0017, -0.4462, 0.0385, -0.0017, -0.4471, 0.0414,
-0.0039, -0.4462, 0.0385, -0.0039, -0.4471, 0.0414, 0.0, -0.444, 0.0384, 0.0, -0.444, 0.0407, 0.0, -0.4417, 0.0388, 0.0, -0.4418, 0.0407, -0.0017, -0.444, 0.0384, -0.0039, -0.444, 0.0384, -0.0017, -0.4418, 0.0388, -0.0039, -0.4418, 0.0388, 0.0, -0.4483, 0.0426, 0.0, -0.4483, 0.045, 0.0, -0.4471, 0.0419, 0.0, -0.4462, 0.045, 0.0, -0.4483, 0.0472, 0.0, -0.4483, 0.0494, 0.0, -0.4462, 0.0472, 0.0, -0.4462, 0.0494, -0.0017, -0.4476, 0.0419, -0.0039, -0.4476, 0.0419, -0.0017, -0.4471, 0.0419, -0.0039, -0.4471, 0.0419, 0.0, -0.444, 0.0429, 0.0, -0.444, 0.045, 0.0, -0.4418, 0.0429, 0.0, -0.4418, 0.045, 0.0, -0.444, 0.0472, 0.0, -0.444, 0.0494, 0.0, -0.4418, 0.0472, 0.0, -0.4418, 0.0494,
0.0, -0.4402, 0.0392, 0.0, -0.4392, 0.0397, 0.0, -0.4375, 0.0406, -0.0017, -0.44, 0.0393, -0.0017, -0.439, 0.0398, -0.0039, -0.44, 0.0393, -0.0039, -0.439, 0.0398, -0.0017, -0.438, 0.0402, -0.0039, -0.438, 0.0402, 0.0, -0.4397, 0.0429, 0.0, -0.4397, 0.045, 0.0, -0.4364, 0.0421, 0.0, -0.4375, 0.045, 0.0, -0.4397, 0.0472, 0.0, -0.4397, 0.0494, 0.0, -0.4375, 0.0472, 0.0, -0.4363, 0.0501, -0.0017, -0.4365, 0.0419, -0.0039, -0.4365, 0.0419, -0.0017, -0.4364, 0.0504, -0.0039, -0.4364, 0.0504, 0.0, -0.4359, 0.0433, 0.0, -0.4355, 0.045, 0.0, -0.4354, 0.0473, 0.0, -0.4358, 0.0491, -0.0017, -0.4359, 0.043, -0.0017, -0.4355, 0.045, -0.0039, -0.4359, 0.043, -0.0039, -0.4355, 0.045, -0.0017, -0.4354, 0.0472,
-0.0017, -0.4359, 0.0493, -0.0039, -0.4354, 0.0472, -0.0039, -0.4359, 0.0493, 0.0, -0.4488, 0.0521, 0.0, -0.4481, 0.0535, 0.0, -0.4462, 0.0515, 0.0, -0.4464, 0.0538, 0.0, -0.4475, 0.0547, 0.0, -0.4483, 0.058, 0.0, -0.446, 0.0556, 0.0, -0.4462, 0.058, -0.0017, -0.4489, 0.0521, -0.0017, -0.4481, 0.0535, -0.0039, -0.4489, 0.0521, -0.0039, -0.4481, 0.0535, -0.0017, -0.4462, 0.0538, -0.0039, -0.4462, 0.0538, -0.0017, -0.4473, 0.0548, -0.0039, -0.4473, 0.0548, -0.0017, -0.4461, 0.0556, -0.0039, -0.4461, 0.0556, 0.0, -0.444, 0.0515, 0.0, -0.444, 0.0543, 0.0, -0.4418, 0.0515, 0.0, -0.4417, 0.0541, 0.0, -0.444, 0.0559, 0.0, -0.444, 0.058, 0.0, -0.4419, 0.0558, 0.0, -0.4418, 0.058, -0.0017, -0.444, 0.0543,
-0.0039, -0.444, 0.0543, -0.0017, -0.4418, 0.0541, -0.0039, -0.4418, 0.0541, -0.0017, -0.444, 0.0559, -0.0039, -0.444, 0.0559, -0.0017, -0.4418, 0.0557, -0.0039, -0.4418, 0.0557, 0.0, -0.4483, 0.0602, 0.0, -0.4491, 0.0632, 0.0, -0.4462, 0.0602, 0.0, -0.4462, 0.0624, 0.0, -0.448, 0.0642, 0.0, -0.4468, 0.0653, 0.0, -0.4457, 0.0662, -0.0017, -0.4491, 0.0632, -0.0039, -0.4491, 0.0632, -0.0017, -0.4481, 0.0642, -0.0039, -0.4481, 0.0642, -0.0017, -0.4468, 0.0653, -0.0017, -0.4457, 0.0662, -0.0039, -0.4468, 0.0653, -0.0039, -0.4457, 0.0662, 0.0, -0.444, 0.0602, 0.0, -0.444, 0.0624, 0.0, -0.4418, 0.0602, 0.0, -0.4418, 0.0624, 0.0, -0.444, 0.0646, 0.0, -0.4444, 0.0673, 0.0, -0.4418, 0.0646, 0.0, -0.4418, 0.0667,
-0.0017, -0.4444, 0.0673, -0.0039, -0.4444, 0.0673, 0.0, -0.4397, 0.0515, 0.0, -0.4393, 0.0543, 0.0, -0.4375, 0.052, -0.0003, -0.4378, 0.054, 0.0, -0.4398, 0.055, 0.0, -0.4397, 0.058, 0.0, -0.4375, 0.0559, 0.0, -0.4375, 0.058, -0.0017, -0.4395, 0.0544, -0.0039, -0.4395, 0.0544, -0.0017, -0.4373, 0.0518, -0.0017, -0.4378, 0.0539, -0.0039, -0.4373, 0.0518, -0.0039, -0.4378, 0.0539, -0.0017, -0.44, 0.0551, -0.0039, -0.44, 0.0551, 0.0, -0.436, 0.0545, 0.0, -0.4349, 0.0548, 0.0, -0.4353, 0.058, 0.0, -0.4329, 0.056, 0.0, -0.4332, 0.058, -0.0017, -0.4358, 0.0545, -0.0039, -0.4358, 0.0545, -0.0017, -0.4347, 0.0549, -0.0039, -0.4347, 0.0549, -0.0017, -0.4331, 0.0558, -0.0039, -0.4331, 0.0558, 0.0, -0.4397, 0.0602,
0.0, -0.4397, 0.0624, 0.0, -0.4375, 0.0602, 0.0, -0.4375, 0.0624, 0.0, -0.4397, 0.0646, 0.0, -0.4397, 0.0667, 0.0, -0.4375, 0.0646, 0.0, -0.4375, 0.0667, 0.0, -0.4353, 0.0602, 0.0, -0.4353, 0.0624, 0.0, -0.4332, 0.0602, 0.0, -0.4332, 0.0624, 0.0, -0.4353, 0.0646, 0.0, -0.4353, 0.0667, 0.0, -0.4332, 0.0646, 0.0, -0.4332, 0.0667, 0.0, -0.4319, 0.0568, 0.0, -0.4307, 0.0576, 0.0, -0.4294, 0.0587, -0.0017, -0.4318, 0.0568, -0.0017, -0.4307, 0.0576, -0.0039, -0.4318, 0.0568, -0.0039, -0.4307, 0.0576, -0.0017, -0.4294, 0.0587, -0.0039, -0.4294, 0.0587, 0.0, -0.431, 0.0602, 0.0, -0.431, 0.0624, 0.0, -0.4283, 0.0596, 0.0, -0.4288, 0.0624, 0.0, -0.431, 0.0646, 0.0, -0.431, 0.0667,
0.0, -0.4288, 0.0646, 0.0, -0.4288, 0.0667, -0.0017, -0.4283, 0.0596, -0.0039, -0.4283, 0.0596, 0.0, -0.4271, 0.0607, 0.0, -0.426, 0.0615, 0.0, -0.4243, 0.0628, 0.0, -0.4266, 0.0646, 0.0, -0.4266, 0.0667, 0.0, -0.4245, 0.0646, 0.0, -0.4245, 0.0667, -0.0017, -0.427, 0.0607, -0.0017, -0.4259, 0.0616, -0.0039, -0.427, 0.0607, -0.0039, -0.4259, 0.0616, -0.0017, -0.4245, 0.0627, -0.0039, -0.4245, 0.0627, 0.0, -0.4234, 0.0635, 0.0, -0.4222, 0.0642, 0.0, -0.4223, 0.0667, 0.0, -0.4207, 0.0652, 0.0, -0.4197, 0.0659, -0.0017, -0.4234, 0.0635, -0.0039, -0.4234, 0.0635, -0.0017, -0.4223, 0.0642, -0.0039, -0.4223, 0.0642, -0.0017, -0.4206, 0.0653, -0.0017, -0.4195, 0.066, -0.0039, -0.4206, 0.0653, -0.0039, -0.4195, 0.066,
0.0, -0.4178, 0.067, 0.0, -0.4167, 0.0677, -0.0017, -0.418, 0.0669, -0.0039, -0.418, 0.0669, -0.0017, -0.4167, 0.0677, -0.0039, -0.4167, 0.0677, 0.0, -0.4433, 0.0682, 0.0, -0.4421, 0.0693, 0.0, -0.4409, 0.0702, -0.0017, -0.4434, 0.0682, -0.0039, -0.4434, 0.0682, -0.0017, -0.442, 0.0693, -0.0017, -0.441, 0.0701, -0.0039, -0.442, 0.0693, -0.0039, -0.441, 0.0701, 0.0, -0.4397, 0.0689, 0.0, -0.4399, 0.0709, 0.0, -0.4375, 0.0689, 0.0, -0.4383, 0.0721, 0.0, -0.4373, 0.0729, -0.0017, -0.4397, 0.0711, -0.0039, -0.4397, 0.0711, -0.0017, -0.4384, 0.0721, -0.0039, -0.4384, 0.0721, -0.0017, -0.4374, 0.0728, -0.0039, -0.4374, 0.0728, 0.0, -0.4353, 0.0689, 0.0, -0.4353, 0.0711, 0.0, -0.4332, 0.0689, 0.0, -0.4332, 0.0711,
0.0, -0.4357, 0.0739, 0.0, -0.4346, 0.0747, 0.0, -0.4332, 0.0732, 0.0, -0.4334, 0.0754, -0.0017, -0.4358, 0.0739, -0.0017, -0.4347, 0.0746, -0.0039, -0.4358, 0.0739, -0.0039, -0.4347, 0.0746, -0.0017, -0.4332, 0.0755, -0.0039, -0.4332, 0.0755, 0.0, -0.4321, 0.087, 0.0, -0.4322, 0.0879, -0.0017, -0.4321, 0.0868, -0.0017, -0.4322, 0.0882, -0.0039, -0.4321, 0.0868, -0.0039, -0.4322, 0.0882, 0.0, -0.431, 0.0689, 0.0, -0.431, 0.0711, 0.0, -0.4288, 0.0689, 0.0, -0.4288, 0.0711, 0.0, -0.431, 0.0732, 0.0, -0.4309, 0.0763, 0.0, -0.4288, 0.0732, 0.0, -0.4286, 0.0757, -0.0017, -0.4309, 0.0763, -0.0039, -0.4309, 0.0763, -0.0017, -0.4288, 0.0758, -0.0039, -0.4288, 0.0758, 0.0, -0.4266, 0.0689, 0.0, -0.4266, 0.0711,
0.0, -0.4245, 0.0689, 0.0, -0.4245, 0.0711, 0.0, -0.426, 0.0744, 0.0, -0.4266, 0.0754, 0.0, -0.4239, 0.0741, 0.0, -0.4246, 0.0748, -0.0017, -0.4257, 0.0742, -0.0017, -0.4266, 0.0753, -0.0039, -0.4257, 0.0742, -0.0039, -0.4266, 0.0753, -0.0017, -0.4242, 0.0742, -0.0017, -0.4246, 0.0748, -0.0039, -0.4242, 0.0742, -0.0039, -0.4246, 0.0748, 0.0, -0.4304, 0.0802, 0.0, -0.4286, 0.0777, 0.0, -0.4297, 0.0791, 0.0, -0.4311, 0.0817, 0.0, -0.4317, 0.0839, 0.0, -0.4278, 0.0829, 0.0, -0.4276, 0.0836, -0.0017, -0.4303, 0.0801, -0.0039, -0.4303, 0.0801, -0.0017, -0.4286, 0.0777, -0.0017, -0.4296, 0.079, -0.0039, -0.4286, 0.0777, -0.0039, -0.4296, 0.079, -0.0017, -0.4311, 0.0819, -0.0017, -0.4318, 0.0841, -0.0039, -0.4311, 0.0819,
-0.0039, -0.4318, 0.0841, -0.0017, -0.4278, 0.0829, -0.0017, -0.4277, 0.083, -0.0039, -0.4278, 0.0829, -0.0039, -0.4277, 0.083, 0.0, -0.4275, 0.0766, 0.0, -0.4264, 0.0798, 0.0, -0.4246, 0.0778, 0.0, -0.4248, 0.0794, 0.0, -0.4285, 0.0847, 0.0, -0.4275, 0.084, 0.0, -0.4253, 0.0811, 0.0, -0.4245, 0.0841, -0.0017, -0.4275, 0.0767, -0.0017, -0.4264, 0.0798, -0.0039, -0.4275, 0.0767, -0.0039, -0.4264, 0.0798, -0.0017, -0.4246, 0.0778, -0.0017, -0.4248, 0.0794, -0.0039, -0.4246, 0.0778, -0.0039, -0.4248, 0.0794, -0.0017, -0.4285, 0.0847, -0.0017, -0.4275, 0.084, -0.0039, -0.4285, 0.0847, -0.0039, -0.4275, 0.084, -0.0017, -0.4253, 0.0811, -0.0039, -0.4253, 0.0811, 0.0, -0.4223, 0.0689, 0.0, -0.4223, 0.0711, 0.0, -0.4201, 0.0689,
0.0, -0.4201, 0.0711, 0.0, -0.4222, 0.0738, 0.0, -0.4223, 0.0754, 0.0, -0.4201, 0.0734, 0.0, -0.4201, 0.0754, 0.0, -0.418, 0.0689, 0.0, -0.418, 0.0711, 0.0, -0.4156, 0.0683, 0.0, -0.4158, 0.0711, 0.0, -0.418, 0.0733, 0.0, -0.418, 0.0754, 0.0, -0.4158, 0.0736, 0.0, -0.4158, 0.0754, -0.0017, -0.4156, 0.0683, -0.0039, -0.4156, 0.0683, 0.0, -0.422, 0.0778, 0.0, -0.4229, 0.0789, 0.0, -0.4204, 0.0775, 0.0, -0.4201, 0.0798, 0.0, -0.4223, 0.0819, 0.0, -0.4223, 0.0841, 0.0, -0.4201, 0.0819, 0.0, -0.4201, 0.0841, -0.0017, -0.4224, 0.078, -0.0017, -0.4229, 0.079, -0.0039, -0.4224, 0.078, -0.0039, -0.4229, 0.079, 0.0, -0.418, 0.0776, 0.0, -0.418, 0.0798, 0.0, -0.4158, 0.0776,
0.0, -0.4158, 0.0798, 0.0, -0.418, 0.0819, 0.0, -0.418, 0.0841, 0.0, -0.4158, 0.0819, 0.0, -0.4158, 0.0841, 0.0, -0.432, 0.0862, 0.0, -0.4321, 0.0887, 0.0, -0.4288, 0.0863, 0.0, -0.4288, 0.0884, 0.0, -0.4317, 0.0909, 0.0, -0.4309, 0.093, 0.0, -0.4288, 0.0906, 0.0, -0.4288, 0.0928, -0.0017, -0.432, 0.0858, -0.0017, -0.432, 0.0893, -0.0039, -0.432, 0.0858, -0.0039, -0.432, 0.0893, -0.0017, -0.4318, 0.0906, -0.0017, -0.431, 0.0928, -0.0039, -0.4318, 0.0906, -0.0039, -0.431, 0.0928, 0.0, -0.4266, 0.0863, 0.0, -0.4266, 0.0884, 0.0, -0.4245, 0.0863, 0.0, -0.4245, 0.0884, 0.0, -0.4266, 0.0906, 0.0, -0.4266, 0.0928, 0.0, -0.4245, 0.0906, 0.0, -0.4245, 0.0928, 0.0, -0.4301, 0.0941,
0.0, -0.4293, 0.0953, 0.0, -0.4282, 0.0965, -0.0017, -0.4301, 0.0942, -0.0039, -0.4301, 0.0942, -0.0017, -0.4293, 0.0953, -0.0017, -0.4282, 0.0965, -0.0039, -0.4293, 0.0953, -0.0039, -0.4282, 0.0965, 0.0, -0.4266, 0.0949, 0.0, -0.427, 0.0976, 0.0, -0.4245, 0.0949, 0.0, -0.4245, 0.0971, 0.0, -0.4258, 0.0984, 0.0, -0.4247, 0.0992, -0.0017, -0.427, 0.0976, -0.0039, -0.427, 0.0976, -0.0017, -0.4259, 0.0984, -0.0039, -0.4259, 0.0984, -0.0017, -0.4245, 0.0993, -0.0039, -0.4245, 0.0993, 0.0, -0.4223, 0.0863, 0.0, -0.4223, 0.0884, 0.0, -0.4201, 0.0863, 0.0, -0.4201, 0.0884, 0.0, -0.4223, 0.0906, 0.0, -0.4223, 0.0928, 0.0, -0.4201, 0.0906, 0.0, -0.4201, 0.0928, 0.0, -0.418, 0.0863, 0.0, -0.418, 0.0884,
0.0, -0.4158, 0.0863, 0.0, -0.4158, 0.0884, 0.0, -0.418, 0.0906, 0.0, -0.418, 0.0928, 0.0, -0.4158, 0.0906, 0.0, -0.4158, 0.0928, 0.0, -0.4223, 0.0949, 0.0, -0.4223, 0.0971, 0.0, -0.4201, 0.0949, 0.0, -0.4201, 0.0971, 0.0, -0.4226, 0.1, 0.0, -0.4205, 0.1005, 0.0, -0.4197, 0.1006, -0.0017, -0.4223, 0.1001, -0.0039, -0.4223, 0.1001, -0.0017, -0.4211, 0.1003, -0.0017, -0.42, 0.1005, -0.0039, -0.4211, 0.1003, -0.0039, -0.42, 0.1005, 0.0, -0.418, 0.0949, 0.0, -0.418, 0.0971, 0.0, -0.4158, 0.0949, 0.0, -0.4158, 0.0971, 0.0, -0.418, 0.0993, 0.0, -0.418, 0.1006, 0.0, -0.4157, 0.1003, 0.0, -0.4166, 0.1004, -0.0017, -0.418, 0.1006, -0.0039, -0.418, 0.1006, -0.0017, -0.4154, 0.1003,
-0.0017, -0.4165, 0.1004, -0.0039, -0.4154, 0.1003, -0.0039, -0.4165, 0.1004, 0.0, -0.4131, -0.0014, 0.0, -0.4136, 0.0016, 0.0, -0.4112, -0.0002, 0.0, -0.4115, 0.0016, 0.0, -0.4136, 0.0038, 0.0, -0.4136, 0.006, 0.0, -0.4115, 0.0038, 0.0, -0.4115, 0.006, -0.0017, -0.413, -0.0014, -0.0039, -0.413, -0.0014, -0.0017, -0.4114, -0.0004, -0.0039, -0.4114, -0.0004, 0.0, -0.4104, 0.0005, 0.0, -0.4094, 0.0012, 0.0, -0.4081, 0.003, 0.0, -0.4093, 0.006, 0.0, -0.4074, 0.004, 0.0, -0.4065, 0.0057, -0.0017, -0.4104, 0.0005, -0.0017, -0.4091, 0.0015, -0.0039, -0.4104, 0.0005, -0.0039, -0.4091, 0.0015, -0.0017, -0.4082, 0.0027, -0.0039, -0.4082, 0.0027, -0.0017, -0.4075, 0.0038, -0.0017, -0.4064, 0.0059, -0.0039, -0.4075, 0.0038,
-0.0039, -0.4064, 0.0059, 0.0, -0.4136, 0.0081, 0.0, -0.4136, 0.0103, 0.0, -0.4115, 0.0081, 0.0, -0.4115, 0.0103, 0.0, -0.4136, 0.0125, 0.0, -0.4136, 0.0146, 0.0, -0.4115, 0.0125, 0.0, -0.4115, 0.0146, 0.0, -0.4093, 0.0081, 0.0, -0.4093, 0.0103, 0.0, -0.406, 0.008, 0.0, -0.4071, 0.0103, 0.0, -0.4093, 0.0125, 0.0, -0.4093, 0.0146, 0.0, -0.4071, 0.0125, 0.0, -0.4061, 0.0146, -0.0017, -0.4061, 0.0076, -0.0039, -0.4061, 0.0076, -0.0017, -0.4061, 0.0147, -0.0039, -0.4061, 0.0147, 0.0, -0.398, -0.0011, -0.0017, -0.3981, -0.0012, -0.0039, -0.3981, -0.0012, 0.0, -0.4059, 0.0088, 0.0, -0.4059, 0.0103, 0.0, -0.4059, 0.0125, 0.0, -0.406, 0.0136, -0.0017, -0.406, 0.0087, -0.0017, -0.4059, 0.0103,
-0.0039, -0.406, 0.0087, -0.0039, -0.4059, 0.0103, -0.0017, -0.4059, 0.0125, -0.0017, -0.406, 0.0136, -0.0039, -0.4059, 0.0125, -0.0039, -0.406, 0.0136, 0.0, -0.4136, 0.0168, 0.0, -0.4136, 0.019, 0.0, -0.4115, 0.0168, 0.0, -0.4115, 0.019, 0.0, -0.4136, 0.0212, 0.0, -0.4133, 0.0239, 0.0, -0.4105, 0.0222, 0.0, -0.4116, 0.0231, -0.0017, -0.4136, 0.024, -0.0039, -0.4136, 0.024, -0.0017, -0.4105, 0.0221, -0.0017, -0.4115, 0.023, -0.0039, -0.4105, 0.0221, -0.0039, -0.4115, 0.023, 0.0, -0.4093, 0.0168, 0.0, -0.4083, 0.0198, 0.0, -0.4068, 0.0171, 0.0, -0.4075, 0.0187, 0.0, -0.4093, 0.0211, -0.0017, -0.4083, 0.0199, -0.0039, -0.4083, 0.0199, -0.0017, -0.4067, 0.0168, -0.0017, -0.4076, 0.0189, -0.0039, -0.4067, 0.0168,
-0.0039, -0.4076, 0.0189, -0.0017, -0.4093, 0.0211, -0.0039, -0.4093, 0.0211, 0.0, -0.3966, -0.0003, 0.0, -0.3948, 0.0007, 0.0, -0.3939, 0.0012, -0.0017, -0.3963, -0.0001, -0.0039, -0.3963, -0.0001, -0.0017, -0.3951, 0.0005, -0.0017, -0.394, 0.0011, -0.0039, -0.3951, 0.0005, -0.0039, -0.394, 0.0011, 0.0, -0.3919, -0.0005, 0.0, -0.3921, 0.0022, 0.0, -0.3897, -0.0005, 0.0, -0.3897, 0.0016, 0.0, -0.391, 0.0029, 0.0, -0.3899, 0.0035, -0.0017, -0.3921, 0.0022, -0.0039, -0.3921, 0.0022, -0.0017, -0.3911, 0.0028, -0.0039, -0.3911, 0.0028, -0.0017, -0.3897, 0.0036, -0.0039, -0.3897, 0.0036, 0.0, -0.3876, -0.0005, 0.0, -0.3876, 0.0016, 0.0, -0.3854, -0.0005, 0.0, -0.3854, 0.0016, 0.0, -0.388, 0.0046, 0.0, -0.3869, 0.0052,
0.0, -0.3854, 0.0038, 0.0, -0.3857, 0.0059, -0.0017, -0.3881, 0.0045, -0.0017, -0.387, 0.0052, -0.0039, -0.3881, 0.0045, -0.0039, -0.387, 0.0052, -0.0017, -0.3854, 0.0061, -0.0039, -0.3854, 0.0061, 0.0, -0.3832, -0.0005, 0.0, -0.3832, 0.0016, 0.0, -0.3811, -0.0005, 0.0, -0.3811, 0.0016, 0.0, -0.3832, 0.0038, 0.0, -0.3839, 0.007, 0.0, -0.3811, 0.0038, 0.0, -0.3811, 0.006, -0.0017, -0.3841, 0.0069, -0.0039, -0.3841, 0.0069, 0.0, -0.3828, 0.0076, 0.0, -0.3812, 0.0086, 0.0, -0.38, 0.0093, -0.0017, -0.383, 0.0075, -0.0039, -0.383, 0.0075, -0.0017, -0.3811, 0.0086, -0.0017, -0.38, 0.0093, -0.0039, -0.3811, 0.0086, -0.0039, -0.38, 0.0093, 0.0, -0.4024, 0.0452, 0.0, -0.4026, 0.0472, 0.0, -0.4021, 0.0488,
-0.0017, -0.4023, 0.0451, -0.0039, -0.4023, 0.0451, -0.0017, -0.4026, 0.0472, -0.0017, -0.402, 0.0489, -0.0039, -0.4026, 0.0472, -0.0039, -0.402, 0.0489, 0.0, -0.4011, 0.0429, 0.0, -0.4018, 0.0442, 0.0, -0.3986, 0.042, 0.0, -0.3984, 0.045, 0.0, -0.4006, 0.0472, 0.0, -0.4013, 0.05, 0.0, -0.3984, 0.0472, 0.0, -0.3984, 0.0494, -0.0017, -0.4011, 0.0429, -0.0017, -0.4017, 0.044, -0.0039, -0.4011, 0.0429, -0.0039, -0.4017, 0.044, -0.0017, -0.3987, 0.042, -0.0039, -0.3987, 0.042, -0.0017, -0.4014, 0.05, -0.0039, -0.4014, 0.05, 0.0, -0.4001, 0.0509, 0.0, -0.3986, 0.0514, -0.0017, -0.4002, 0.0508, -0.0039, -0.4002, 0.0508, -0.0017, -0.3986, 0.0514, -0.0039, -0.3986, 0.0514, 0.0, -0.396, 0.0423, 0.0, -0.3963, 0.045,
0.0, -0.3946, 0.0434, 0.0, -0.3935, 0.0444, 0.0, -0.3963, 0.0472, 0.0, -0.3963, 0.0494, 0.0, -0.3928, 0.0472, 0.0, -0.3941, 0.0498, -0.0017, -0.396, 0.0423, -0.0039, -0.396, 0.0423, -0.0017, -0.3946, 0.0435, -0.0017, -0.3935, 0.0444, -0.0039, -0.3946, 0.0435, -0.0039, -0.3935, 0.0444, -0.0017, -0.3928, 0.0472, -0.0017, -0.3938, 0.0494, -0.0039, -0.3928, 0.0472, -0.0039, -0.3938, 0.0494, 0.0, -0.3959, 0.0512, 0.0, -0.3951, 0.0505, -0.0017, -0.3959, 0.0512, -0.0039, -0.3959, 0.0512, -0.0017, -0.3951, 0.0505, -0.0039, -0.3951, 0.0505, 0.0, -0.3789, -0.0005, 0.0, -0.3789, 0.0016, 0.0, -0.3767, -0.0005, 0.0, -0.3767, 0.0016, 0.0, -0.3789, 0.0038, 0.0, -0.3789, 0.006, 0.0, -0.3767, 0.0038, 0.0, -0.3767, 0.006,
0.0, -0.3746, -0.0005, 0.0, -0.3746, 0.0016, 0.0, -0.3724, -0.0005, 0.0, -0.3724, 0.0016, 0.0, -0.3746, 0.0038, 0.0, -0.3746, 0.006, 0.0, -0.3724, 0.0038, 0.0, -0.3724, 0.006, 0.0, -0.3789, 0.0081, 0.0, -0.379, 0.0098, 0.0, -0.3767, 0.0081, 0.0, -0.3769, 0.0109, 0.0, -0.3758, 0.0115, -0.0017, -0.3789, 0.0099, -0.0039, -0.3789, 0.0099, -0.0017, -0.377, 0.0109, -0.0039, -0.377, 0.0109, -0.0017, -0.3759, 0.0115, -0.0039, -0.3759, 0.0115, 0.0, -0.3746, 0.0081, 0.0, -0.3746, 0.0103, 0.0, -0.3724, 0.0081, 0.0, -0.3724, 0.0103, 0.0, -0.3747, 0.0122, 0.0, -0.3728, 0.0133, 0.0, -0.3717, 0.0139, -0.0017, -0.3746, 0.0123, -0.0039, -0.3746, 0.0123, -0.0017, -0.3729, 0.0132, -0.0017, -0.3718, 0.0138,
-0.0039, -0.3729, 0.0132, -0.0039, -0.3718, 0.0138, 0.0, -0.3695, -0.0016, 0.0, -0.3702, 0.0016, 0.0, -0.3678, -0.0004, 0.0, -0.368, 0.0016, 0.0, -0.3702, 0.0038, 0.0, -0.3702, 0.006, 0.0, -0.368, 0.0038, 0.0, -0.368, 0.006, -0.0017, -0.3694, -0.0015, -0.0039, -0.3694, -0.0015, -0.0017, -0.3681, -0.0006, -0.0039, -0.3681, -0.0006, 0.0, -0.3667, 0.0003, 0.0, -0.3656, 0.001, 0.0, -0.364, 0.0021, 0.0, -0.3659, 0.0038, 0.0, -0.3659, 0.006, 0.0, -0.363, 0.0027, 0.0, -0.3637, 0.006, -0.0017, -0.3667, 0.0003, -0.0017, -0.3656, 0.0011, -0.0039, -0.3667, 0.0003, -0.0039, -0.3656, 0.0011, -0.0017, -0.3639, 0.0021, -0.0039, -0.3639, 0.0021, -0.0017, -0.3628, 0.0029, -0.0039, -0.3628, 0.0029, 0.0, -0.3702, 0.0081,
0.0, -0.3702, 0.0103, 0.0, -0.368, 0.0081, 0.0, -0.368, 0.0103, 0.0, -0.3702, 0.0125, 0.0, -0.3705, 0.0146, 0.0, -0.368, 0.0125, 0.0, -0.3686, 0.0156, -0.0017, -0.3702, 0.0147, -0.0039, -0.3702, 0.0147, -0.0017, -0.3687, 0.0155, -0.0039, -0.3687, 0.0155, 0.0, -0.3659, 0.0081, 0.0, -0.3659, 0.0103, 0.0, -0.3637, 0.0081, 0.0, -0.3637, 0.0103, 0.0, -0.3659, 0.0125, 0.0, -0.3659, 0.0146, 0.0, -0.3637, 0.0125, 0.0, -0.3637, 0.0146, 0.0, -0.3675, 0.0162, -0.0017, -0.3677, 0.0161, -0.0039, -0.3677, 0.0161, 0.0, -0.3662, 0.0169, 0.0, -0.3643, 0.0179, 0.0, -0.3633, 0.0184, -0.0017, -0.3659, 0.0171, -0.0039, -0.3659, 0.0171, -0.0017, -0.3646, 0.0178, -0.0017, -0.3635, 0.0184, -0.0039, -0.3646, 0.0178,
-0.0039, -0.3635, 0.0184, 0.0, -0.3613, 0.0039, 0.0, -0.3615, 0.006, 0.0, -0.3601, 0.0046, 0.0, -0.359, 0.0053, -0.0017, -0.3615, 0.0037, -0.0039, -0.3615, 0.0037, -0.0017, -0.36, 0.0046, -0.0017, -0.3589, 0.0053, -0.0039, -0.36, 0.0046, -0.0039, -0.3589, 0.0053, 0.0, -0.3547, -0.001, 0.0, -0.3574, 0.0063, -0.0017, -0.3548, -0.0011, -0.0039, -0.3548, -0.0011, -0.0017, -0.3572, 0.0064, -0.0039, -0.3572, 0.0064, 0.0, -0.3615, 0.0081, 0.0, -0.3615, 0.0103, 0.0, -0.3594, 0.0081, 0.0, -0.3594, 0.0103, 0.0, -0.3615, 0.0125, 0.0, -0.3615, 0.0146, 0.0, -0.3594, 0.0125, 0.0, -0.3594, 0.0146, 0.0, -0.3563, 0.0069, 0.0, -0.3572, 0.0103, 0.0, -0.3547, 0.0079, 0.0, -0.355, 0.0103, 0.0, -0.3572, 0.0125,
0.0, -0.3572, 0.0146, 0.0, -0.355, 0.0125, 0.0, -0.355, 0.0146, -0.0017, -0.3561, 0.0071, -0.0039, -0.3561, 0.0071, -0.0017, -0.355, 0.0077, -0.0039, -0.355, 0.0077, 0.0, -0.3531, 0.0, 0.0, -0.352, 0.0007, 0.0, -0.3507, -0.0005, 0.0, -0.3509, 0.0014, -0.0017, -0.3531, -0.0, -0.0017, -0.352, 0.0007, -0.0039, -0.3531, -0.0, -0.0039, -0.352, 0.0007, -0.0017, -0.3507, 0.0015, -0.0039, -0.3507, 0.0015, 0.0, -0.3485, -0.0005, 0.0, -0.3491, 0.0025, 0.0, -0.3463, -0.0005, 0.0, -0.3463, 0.0016, 0.0, -0.348, 0.0032, 0.0, -0.3467, 0.004, -0.0017, -0.3492, 0.0025, -0.0039, -0.3492, 0.0025, -0.0017, -0.3481, 0.0031, -0.0039, -0.3481, 0.0031, -0.0017, -0.3463, 0.0042, -0.0039, -0.3463, 0.0042, 0.0, -0.3534, 0.0088,
0.0, -0.3523, 0.0094, 0.0, -0.3505, 0.0105, 0.0, -0.3529, 0.0125, 0.0, -0.3529, 0.0146, 0.0, -0.3507, 0.0125, 0.0, -0.3507, 0.0146, -0.0017, -0.3533, 0.0088, -0.0017, -0.3522, 0.0095, -0.0039, -0.3533, 0.0088, -0.0039, -0.3522, 0.0095, -0.0017, -0.3507, 0.0104, -0.0039, -0.3507, 0.0104, 0.0, -0.3494, 0.0112, 0.0, -0.3483, 0.0119, 0.0, -0.3485, 0.0146, 0.0, -0.3465, 0.0129, 0.0, -0.3456, 0.0134, -0.0017, -0.3493, 0.0112, -0.0039, -0.3493, 0.0112, -0.0017, -0.3483, 0.0119, -0.0039, -0.3483, 0.0119, -0.0017, -0.3464, 0.013, -0.0017, -0.3453, 0.0136, -0.0039, -0.3464, 0.013, -0.0039, -0.3453, 0.0136, 0.0, -0.3615, 0.0168, 0.0, -0.3619, 0.0192, 0.0, -0.3594, 0.0168, 0.0, -0.3601, 0.0202, 0.0, -0.3591, 0.0207,
-0.0017, -0.3615, 0.0194, -0.0039, -0.3615, 0.0194, -0.0017, -0.3604, 0.02, -0.0039, -0.3604, 0.02, -0.0017, -0.3593, 0.0206, -0.0039, -0.3593, 0.0206, 0.0, -0.3572, 0.0168, 0.0, -0.3572, 0.019, 0.0, -0.355, 0.0168, 0.0, -0.355, 0.019, 0.0, -0.3573, 0.0217, 0.0, -0.3562, 0.0223, 0.0, -0.355, 0.0212, 0.0, -0.3551, 0.0228, -0.0017, -0.3573, 0.0217, -0.0017, -0.3562, 0.0223, -0.0039, -0.3573, 0.0217, -0.0039, -0.3562, 0.0223, -0.0017, -0.355, 0.0228, -0.0039, -0.355, 0.0228, 0.0, -0.3529, 0.0168, 0.0, -0.3529, 0.019, 0.0, -0.3507, 0.0168, 0.0, -0.3507, 0.019, 0.0, -0.3529, 0.0212, 0.0, -0.3531, 0.0237, 0.0, -0.3507, 0.0212, 0.0, -0.3513, 0.0245, -0.0017, -0.3529, 0.0238, -0.0039, -0.3529, 0.0238,
-0.0017, -0.3516, 0.0243, -0.0039, -0.3516, 0.0243, 0.0, -0.3485, 0.0168, 0.0, -0.3485, 0.019, 0.0, -0.3463, 0.0168, 0.0, -0.3463, 0.019, 0.0, -0.3485, 0.0212, 0.0, -0.3485, 0.0233, 0.0, -0.3463, 0.0212, 0.0, -0.3463, 0.0233, 0.0, -0.3503, 0.0249, -0.0017, -0.3505, 0.0248, -0.0039, -0.3505, 0.0248, 0.0, -0.3488, 0.0256, 0.0, -0.3468, 0.0265, 0.0, -0.3458, 0.0269, -0.0017, -0.3485, 0.0257, -0.0039, -0.3485, 0.0257, -0.0017, -0.347, 0.0264, -0.0017, -0.3459, 0.0269, -0.0039, -0.347, 0.0264, -0.0039, -0.3459, 0.0269, 0.0, -0.3715, 0.0633, 0.0, -0.3738, 0.0653, 0.0, -0.3747, 0.0664, 0.0, -0.3727, 0.0641, 0.0, -0.3724, 0.0667, -0.0017, -0.3715, 0.0633, -0.0039, -0.3715, 0.0633, -0.0017, -0.3738, 0.0653,
-0.0017, -0.3749, 0.0667, -0.0039, -0.3738, 0.0653, -0.0039, -0.3749, 0.0667, -0.0017, -0.3727, 0.0641, -0.0039, -0.3727, 0.0641, 0.0, -0.3704, 0.0626, 0.0, -0.3682, 0.0619, 0.0, -0.3702, 0.0646, 0.0, -0.3702, 0.0667, 0.0, -0.368, 0.0646, 0.0, -0.368, 0.0667, -0.0017, -0.3703, 0.0625, -0.0039, -0.3703, 0.0625, -0.0017, -0.3681, 0.0619, -0.0039, -0.3681, 0.0619, 0.0, -0.3659, 0.0617, 0.0, -0.3635, 0.062, 0.0, -0.3659, 0.0646, 0.0, -0.3659, 0.0667, 0.0, -0.3637, 0.0646, 0.0, -0.3637, 0.0667, -0.0017, -0.3659, 0.0617, -0.0039, -0.3659, 0.0617, -0.0017, -0.3636, 0.062, -0.0039, -0.3636, 0.062, 0.0, -0.3549, 0.0408, -0.0017, -0.3548, 0.0406, -0.0039, -0.3548, 0.0406, 0.0, -0.3561, 0.0439, 0.0, -0.3563, 0.045,
0.0, -0.3558, 0.0428, 0.0, -0.355, 0.045, 0.0, -0.3564, 0.0468, 0.0, -0.3562, 0.0476, 0.0, -0.3558, 0.0497, -0.0017, -0.3561, 0.0439, -0.0017, -0.3563, 0.045, -0.0039, -0.3561, 0.0439, -0.0039, -0.3563, 0.045, -0.0017, -0.3558, 0.0427, -0.0039, -0.3558, 0.0427, -0.0017, -0.3563, 0.0472, -0.0039, -0.3563, 0.0472, -0.0017, -0.3561, 0.0483, -0.0017, -0.3559, 0.0494, -0.0039, -0.3561, 0.0483, -0.0039, -0.3559, 0.0494, 0.0, -0.3501, 0.0372, 0.0, -0.3529, 0.0386, 0.0, -0.354, 0.0397, 0.0, -0.3511, 0.0376, 0.0, -0.3507, 0.0407, -0.0017, -0.3502, 0.0372, -0.0039, -0.3502, 0.0372, -0.0017, -0.3533, 0.0389, -0.0017, -0.3539, 0.0396, -0.0039, -0.3533, 0.0389, -0.0039, -0.3539, 0.0396, -0.0017, -0.3513, 0.0377, -0.0039, -0.3513, 0.0377,
0.0, -0.3486, 0.0367, 0.0, -0.3464, 0.0364, 0.0, -0.3485, 0.0385, 0.0, -0.3485, 0.0407, 0.0, -0.3463, 0.0385, 0.0, -0.3463, 0.0407, -0.0017, -0.3485, 0.0366, -0.0039, -0.3485, 0.0366, -0.0017, -0.3464, 0.0364, -0.0039, -0.3464, 0.0364, 0.0, -0.3529, 0.0429, 0.0, -0.3529, 0.045, 0.0, -0.3507, 0.0429, 0.0, -0.3507, 0.045, 0.0, -0.3529, 0.0472, 0.0, -0.3529, 0.0494, 0.0, -0.3507, 0.0472, 0.0, -0.3507, 0.0494, 0.0, -0.3485, 0.0429, 0.0, -0.3485, 0.045, 0.0, -0.3463, 0.0429, 0.0, -0.3463, 0.045, 0.0, -0.3485, 0.0472, 0.0, -0.3485, 0.0494, 0.0, -0.3463, 0.0472, 0.0, -0.3463, 0.0494, 0.0, -0.3549, 0.0515, -0.0017, -0.3548, 0.0516, -0.0039, -0.3548, 0.0516, 0.0, -0.3618, 0.0628,
0.0, -0.3625, 0.0622, 0.0, -0.3607, 0.0661, 0.0, -0.3593, 0.0645, 0.0, -0.3591, 0.0664, -0.0017, -0.3617, 0.0629, -0.0039, -0.3617, 0.0629, -0.0017, -0.3625, 0.0622, -0.0017, -0.3605, 0.0665, -0.0039, -0.3625, 0.0622, -0.0039, -0.3605, 0.0665, -0.0017, -0.3591, 0.0647, -0.0017, -0.3592, 0.0663, -0.0039, -0.3591, 0.0647, -0.0039, -0.3592, 0.0663, 0.0, -0.3577, 0.0671, -0.0017, -0.3578, 0.0669, -0.0039, -0.3578, 0.0669, 0.0, -0.354, 0.0525, 0.0, -0.353, 0.0538, 0.0, -0.3507, 0.0515, 0.0, -0.3512, 0.0547, 0.0, -0.3501, 0.0551, -0.0017, -0.3539, 0.0526, -0.0017, -0.3532, 0.0536, -0.0039, -0.3539, 0.0526, -0.0039, -0.3532, 0.0536, -0.0017, -0.3514, 0.0546, -0.0039, -0.3514, 0.0546, -0.0017, -0.3503, 0.0551, -0.0039, -0.3503, 0.0551,
0.0, -0.3485, 0.0515, 0.0, -0.3485, 0.0537, 0.0, -0.3463, 0.0515, 0.0, -0.3463, 0.0537, 0.0, -0.3486, 0.0554, 0.0, -0.3463, 0.0555, -0.0017, -0.3485, 0.0554, -0.0039, -0.3485, 0.0554, -0.0017, -0.3463, 0.0555, -0.0039, -0.3463, 0.0555, 0.0, -0.4139, 0.0693, 0.0, -0.4129, 0.0699, 0.0, -0.4111, 0.0708, 0.0, -0.414, 0.0738, 0.0, -0.4133, 0.0745, 0.0, -0.4118, 0.074, 0.0, -0.411, 0.0749, -0.0017, -0.4137, 0.0694, -0.0017, -0.4126, 0.07, -0.0039, -0.4137, 0.0694, -0.0039, -0.4126, 0.07, -0.0017, -0.4115, 0.0706, -0.0039, -0.4115, 0.0706, -0.0017, -0.4136, 0.0739, -0.0017, -0.4131, 0.0746, -0.0039, -0.4136, 0.0739, -0.0039, -0.4131, 0.0746, -0.0017, -0.4119, 0.074, -0.0017, -0.4114, 0.0747, -0.0039, -0.4119, 0.074,
-0.0039, -0.4114, 0.0747, 0.0, -0.4095, 0.0716, 0.0, -0.4086, 0.072, 0.0, -0.4095, 0.0757, 0.0, -0.4068, 0.0728, 0.0, -0.4071, 0.0754, -0.0017, -0.4093, 0.0717, -0.0039, -0.4093, 0.0717, -0.0017, -0.4082, 0.0722, -0.0039, -0.4082, 0.0722, -0.0017, -0.4071, 0.0727, -0.0039, -0.4071, 0.0727, 0.0, -0.4136, 0.0776, 0.0, -0.4136, 0.0798, 0.0, -0.4115, 0.0776, 0.0, -0.4115, 0.0798, 0.0, -0.4136, 0.0819, 0.0, -0.4136, 0.0841, 0.0, -0.4115, 0.0819, 0.0, -0.4115, 0.0841, 0.0, -0.4086, 0.0772, 0.0, -0.4093, 0.0798, 0.0, -0.4081, 0.0784, 0.0, -0.4069, 0.0794, 0.0, -0.4093, 0.0819, 0.0, -0.4093, 0.0841, 0.0, -0.4061, 0.0816, 0.0, -0.4071, 0.0841, -0.0017, -0.4068, 0.0797, -0.0039, -0.4068, 0.0797,
-0.0017, -0.4062, 0.0814, -0.0039, -0.4062, 0.0814, 0.0, -0.4048, 0.0735, 0.0, -0.4049, 0.0754, 0.0, -0.4033, 0.074, 0.0, -0.4024, 0.0743, -0.0017, -0.4049, 0.0735, -0.0039, -0.4049, 0.0735, -0.0017, -0.4032, 0.0741, -0.0017, -0.4021, 0.0745, -0.0039, -0.4032, 0.0741, -0.0039, -0.4021, 0.0745, 0.0, -0.4004, 0.075, 0.0, -0.3983, 0.0757, -0.0017, -0.4006, 0.0749, -0.0039, -0.4006, 0.0749, -0.0017, -0.3984, 0.0757, -0.0039, -0.3984, 0.0757, 0.0, -0.4049, 0.0776, 0.0, -0.4057, 0.0802, 0.0, -0.4028, 0.0776, 0.0, -0.4028, 0.0798, 0.0, -0.4055, 0.0823, 0.0, -0.4054, 0.0842, 0.0, -0.4028, 0.0819, 0.0, -0.4028, 0.0841, -0.0017, -0.4057, 0.0803, -0.0039, -0.4057, 0.0803, -0.0017, -0.4056, 0.0819, -0.0039, -0.4056, 0.0819,
0.0, -0.4006, 0.0776, 0.0, -0.4006, 0.0798, 0.0, -0.3984, 0.0776, 0.0, -0.3984, 0.0798, 0.0, -0.4006, 0.0819, 0.0, -0.4006, 0.0841, 0.0, -0.3984, 0.0819, 0.0, -0.3984, 0.0841, 0.0, -0.4136, 0.0863, 0.0, -0.4136, 0.0884, 0.0, -0.4115, 0.0863, 0.0, -0.4115, 0.0884, 0.0, -0.4136, 0.0906, 0.0, -0.4136, 0.0928, 0.0, -0.4115, 0.0906, 0.0, -0.4115, 0.0928, 0.0, -0.4093, 0.0863, 0.0, -0.4093, 0.0884, 0.0, -0.4071, 0.0863, 0.0, -0.4071, 0.0884, 0.0, -0.4093, 0.0906, 0.0, -0.4093, 0.0928, 0.0, -0.4071, 0.0906, 0.0, -0.4064, 0.0929, -0.0017, -0.4064, 0.0928, -0.0039, -0.4064, 0.0928, 0.0, -0.4136, 0.0949, 0.0, -0.4136, 0.0971, 0.0, -0.4115, 0.0949, 0.0, -0.4107, 0.0981,
0.0, -0.4133, 0.0997, 0.0, -0.4118, 0.0988, -0.0017, -0.4106, 0.098, -0.0039, -0.4106, 0.098, -0.0017, -0.4136, 0.0998, -0.0039, -0.4136, 0.0998, -0.0017, -0.4117, 0.0987, -0.0039, -0.4117, 0.0987, 0.0, -0.4083, 0.0957, 0.0, -0.4094, 0.097, 0.0, -0.4075, 0.0946, -0.0017, -0.4084, 0.0958, -0.0017, -0.4094, 0.097, -0.0039, -0.4084, 0.0958, -0.0039, -0.4094, 0.097, -0.0017, -0.4075, 0.0947, -0.0039, -0.4075, 0.0947, 0.0, -0.4045, 0.0862, 0.0, -0.4045, 0.0887, 0.0, -0.4028, 0.0863, 0.0, -0.4029, 0.089, 0.0, -0.4056, 0.0907, 0.0, -0.406, 0.0917, -0.0017, -0.4045, 0.0887, -0.0039, -0.4045, 0.0887, -0.0017, -0.4028, 0.089, -0.0039, -0.4028, 0.089, -0.0017, -0.4056, 0.0906, -0.0017, -0.406, 0.0917, -0.0039, -0.4056, 0.0906,
-0.0039, -0.406, 0.0917, 0.0, -0.4006, 0.0863, 0.0, -0.4005, 0.0893, 0.0, -0.3984, 0.0863, 0.0, -0.3984, 0.0884, 0.0, -0.3995, 0.0895, 0.0, -0.3984, 0.0897, -0.0017, -0.4006, 0.0893, -0.0039, -0.4006, 0.0893, -0.0017, -0.3995, 0.0895, -0.0039, -0.3995, 0.0895, -0.0017, -0.3984, 0.0897, -0.0039, -0.3984, 0.0897, 0.0, -0.3968, 0.0762, -0.0017, -0.3966, 0.0763, -0.0039, -0.3966, 0.0763, 0.0, -0.3959, 0.0765, 0.0, -0.3963, 0.0798, 0.0, -0.3939, 0.077, 0.0, -0.3941, 0.0798, 0.0, -0.3963, 0.0819, 0.0, -0.3963, 0.0841, 0.0, -0.3941, 0.0819, 0.0, -0.3941, 0.0841, -0.0017, -0.3955, 0.0766, -0.0039, -0.3955, 0.0766, -0.0017, -0.3941, 0.077, -0.0039, -0.3941, 0.077, 0.0, -0.3918, 0.0775, 0.0, -0.3919, 0.0798,
0.0, -0.3897, 0.0779, 0.0, -0.3897, 0.0798, 0.0, -0.3919, 0.0819, 0.0, -0.3919, 0.0841, 0.0, -0.3897, 0.0819, 0.0, -0.3897, 0.0841, -0.0017, -0.3919, 0.0775, -0.0039, -0.3919, 0.0775, -0.0017, -0.3898, 0.0779, -0.0039, -0.3898, 0.0779, 0.0, -0.3875, 0.0783, 0.0, -0.3876, 0.0798, 0.0, -0.3861, 0.0785, 0.0, -0.3852, 0.0787, 0.0, -0.3876, 0.0819, 0.0, -0.3876, 0.0841, 0.0, -0.3854, 0.0819, 0.0, -0.3854, 0.0841, -0.0017, -0.3876, 0.0783, -0.0039, -0.3876, 0.0783, -0.0017, -0.3858, 0.0786, -0.0017, -0.3848, 0.0787, -0.0039, -0.3858, 0.0786, -0.0039, -0.3848, 0.0787, 0.0, -0.3832, 0.0789, 0.0, -0.381, 0.079, 0.0, -0.3832, 0.0819, 0.0, -0.3832, 0.0841, 0.0, -0.3811, 0.0819, 0.0, -0.3811, 0.0841,
-0.0017, -0.3832, 0.0789, -0.0039, -0.3832, 0.0789, -0.0017, -0.3811, 0.079, -0.0039, -0.3811, 0.079, 0.0, -0.3963, 0.0863, 0.0, -0.3963, 0.0884, 0.0, -0.3941, 0.0863, 0.0, -0.3941, 0.0884, 0.0, -0.3963, 0.0901, 0.0, -0.3942, 0.0904, -0.0017, -0.3963, 0.0901, -0.0039, -0.3963, 0.0901, -0.0017, -0.3941, 0.0904, -0.0039, -0.3941, 0.0904, 0.0, -0.3919, 0.0863, 0.0, -0.3919, 0.0884, 0.0, -0.3897, 0.0863, 0.0, -0.3897, 0.0884, 0.0, -0.392, 0.0908, 0.0, -0.3899, 0.0911, -0.0017, -0.3919, 0.0908, -0.0039, -0.3919, 0.0908, -0.0017, -0.3898, 0.0912, -0.0039, -0.3898, 0.0912, 0.0, -0.3876, 0.0863, 0.0, -0.3876, 0.0884, 0.0, -0.3854, 0.0863, 0.0, -0.3854, 0.0884, 0.0, -0.3877, 0.0915, 0.0, -0.3856, 0.0917,
0.0, -0.3848, 0.0918, -0.0017, -0.3876, 0.0915, -0.0039, -0.3876, 0.0915, -0.0017, -0.3861, 0.0917, -0.0017, -0.3851, 0.0917, -0.0039, -0.3861, 0.0917, -0.0039, -0.3851, 0.0917, 0.0, -0.3832, 0.0863, 0.0, -0.3832, 0.0884, 0.0, -0.3811, 0.0863, 0.0, -0.3811, 0.0884, 0.0, -0.3832, 0.0906, 0.0, -0.3832, 0.0918, 0.0, -0.3811, 0.0906, 0.0, -0.3811, 0.0918, -0.0017, -0.3832, 0.0918, -0.0039, -0.3832, 0.0918, -0.0017, -0.3811, 0.0918, -0.0039, -0.3811, 0.0918, 0.0, -0.3849, 0.0954, 0.0, -0.386, 0.0967, 0.0, -0.3867, 0.0991, 0.0, -0.3862, 0.1015, -0.0017, -0.3849, 0.0954, -0.0017, -0.3861, 0.0969, -0.0039, -0.3849, 0.0954, -0.0039, -0.3861, 0.0969, -0.0017, -0.3867, 0.0991, -0.0017, -0.3862, 0.1015, -0.0039, -0.3867, 0.0991,
-0.0039, -0.3862, 0.1015, 0.0, -0.3832, 0.0941, 0.0, -0.3832, 0.0971, 0.0, -0.3808, 0.0943, 0.0, -0.3811, 0.0971, 0.0, -0.3832, 0.0993, 0.0, -0.3832, 0.1015, 0.0, -0.3811, 0.0993, 0.0, -0.3811, 0.1015, -0.0017, -0.3832, 0.0941, -0.0039, -0.3832, 0.0941, -0.0017, -0.3804, 0.0944, -0.0039, -0.3804, 0.0944, 0.0, -0.3847, 0.1029, -0.0017, -0.3848, 0.1029, -0.0039, -0.3848, 0.1029, 0.0, -0.3834, 0.1037, 0.0, -0.381, 0.1039, -0.0017, -0.3832, 0.1037, -0.0039, -0.3832, 0.1037, -0.0017, -0.381, 0.1039, -0.0039, -0.381, 0.1039, 0.0, -0.3759, 0.0695, 0.0, -0.3761, 0.071, 0.0, -0.376, 0.0728, -0.0017, -0.3758, 0.0693, -0.0017, -0.3761, 0.0711, -0.0039, -0.3758, 0.0693, -0.0039, -0.3761, 0.0711, -0.0017, -0.376, 0.0731,
-0.0039, -0.376, 0.0731, 0.0, -0.3757, 0.0685, 0.0, -0.3746, 0.0711, 0.0, -0.3724, 0.0689, 0.0, -0.3724, 0.0711, 0.0, -0.3758, 0.0738, 0.0, -0.3749, 0.0758, 0.0, -0.3724, 0.0732, 0.0, -0.3724, 0.0754, -0.0017, -0.3755, 0.0682, -0.0039, -0.3755, 0.0682, -0.0017, -0.3756, 0.0742, -0.0017, -0.3751, 0.0755, -0.0039, -0.3756, 0.0742, -0.0039, -0.3751, 0.0755, 0.0, -0.3789, 0.0791, 0.0, -0.3766, 0.0793, 0.0, -0.3789, 0.0819, 0.0, -0.3789, 0.0841, 0.0, -0.3767, 0.0819, 0.0, -0.3767, 0.0841, -0.0017, -0.3789, 0.0791, -0.0039, -0.3789, 0.0791, -0.0017, -0.3767, 0.0792, -0.0039, -0.3767, 0.0792, 0.0, -0.3741, 0.0772, 0.0, -0.3744, 0.0799, 0.0, -0.3731, 0.0784, 0.0, -0.3725, 0.0799, 0.0, -0.3746, 0.0819,
0.0, -0.3746, 0.0841, 0.0, -0.372, 0.0809, 0.0, -0.3724, 0.0841, -0.0017, -0.374, 0.0773, -0.0017, -0.3745, 0.0798, -0.0039, -0.374, 0.0773, -0.0039, -0.3745, 0.0798, -0.0017, -0.3732, 0.0783, -0.0017, -0.3724, 0.0799, -0.0039, -0.3732, 0.0783, -0.0039, -0.3724, 0.0799, -0.0017, -0.3717, 0.081, -0.0039, -0.3717, 0.081, 0.0, -0.3702, 0.0689, 0.0, -0.3702, 0.0711, 0.0, -0.368, 0.0689, 0.0, -0.368, 0.0711, 0.0, -0.3702, 0.0732, 0.0, -0.3702, 0.0754, 0.0, -0.368, 0.0732, 0.0, -0.368, 0.0754, 0.0, -0.3659, 0.0689, 0.0, -0.3659, 0.0711, 0.0, -0.3637, 0.0689, 0.0, -0.3637, 0.0711, 0.0, -0.3659, 0.0732, 0.0, -0.3659, 0.0754, 0.0, -0.3637, 0.0732, 0.0, -0.3637, 0.0754, 0.0, -0.3702, 0.0776,
0.0, -0.3705, 0.0802, 0.0, -0.368, 0.0776, 0.0, -0.3682, 0.0808, 0.0, -0.37, 0.0817, 0.0, -0.3702, 0.0841, 0.0, -0.3678, 0.0819, 0.0, -0.368, 0.0841, -0.0017, -0.3703, 0.0803, -0.0039, -0.3703, 0.0803, -0.0017, -0.3686, 0.0808, -0.0039, -0.3686, 0.0808, -0.0017, -0.3702, 0.0816, -0.0039, -0.3702, 0.0816, -0.0017, -0.368, 0.0818, -0.0039, -0.368, 0.0818, 0.0, -0.3659, 0.0776, 0.0, -0.3657, 0.0809, 0.0, -0.3637, 0.0776, 0.0, -0.3634, 0.0807, 0.0, -0.366, 0.082, 0.0, -0.3659, 0.0841, 0.0, -0.3638, 0.0821, 0.0, -0.3637, 0.0841, -0.0017, -0.365, 0.0808, -0.0039, -0.365, 0.0808, -0.0017, -0.3637, 0.0807, -0.0039, -0.3637, 0.0807, -0.0017, -0.367, 0.082, -0.0039, -0.367, 0.082, -0.0017, -0.3637, 0.0821,
-0.0039, -0.3637, 0.0821, 0.0, -0.3789, 0.0863, 0.0, -0.3789, 0.0884, 0.0, -0.3767, 0.0863, 0.0, -0.3767, 0.0884, 0.0, -0.3789, 0.0906, 0.0, -0.3789, 0.0918, 0.0, -0.3767, 0.0906, 0.0, -0.3767, 0.0918, -0.0017, -0.3789, 0.0918, -0.0039, -0.3789, 0.0918, -0.0017, -0.3767, 0.0918, -0.0039, -0.3767, 0.0918, 0.0, -0.3746, 0.0863, 0.0, -0.3746, 0.0884, 0.0, -0.3724, 0.0863, 0.0, -0.3724, 0.0884, 0.0, -0.3746, 0.0906, 0.0, -0.3746, 0.0918, 0.0, -0.3724, 0.0906, 0.0, -0.3724, 0.0918, -0.0017, -0.3746, 0.0918, -0.0039, -0.3746, 0.0918, -0.0017, -0.3724, 0.0918, -0.0039, -0.3724, 0.0918, 0.0, -0.379, 0.0951, 0.0, -0.3779, 0.0963, 0.0, -0.3773, 0.0974, 0.0, -0.3789, 0.0993, 0.0, -0.3782, 0.102,
0.0, -0.3769, 0.0991, 0.0, -0.3774, 0.1009, -0.0017, -0.3789, 0.0952, -0.0017, -0.3779, 0.0962, -0.0039, -0.3789, 0.0952, -0.0039, -0.3779, 0.0962, -0.0017, -0.3773, 0.0972, -0.0039, -0.3773, 0.0972, -0.0017, -0.3782, 0.1021, -0.0039, -0.3782, 0.1021, -0.0017, -0.3769, 0.0991, -0.0017, -0.3776, 0.1014, -0.0039, -0.3769, 0.0991, -0.0039, -0.3776, 0.1014, 0.0, -0.3702, 0.0863, 0.0, -0.3702, 0.0884, 0.0, -0.368, 0.0863, 0.0, -0.368, 0.0884, 0.0, -0.3703, 0.0916, 0.0, -0.3712, 0.0917, 0.0, -0.3679, 0.0913, -0.0017, -0.3701, 0.0916, -0.0017, -0.3712, 0.0917, -0.0039, -0.3701, 0.0916, -0.0039, -0.3712, 0.0917, -0.0017, -0.368, 0.0913, -0.0039, -0.368, 0.0913, 0.0, -0.3659, 0.0863, 0.0, -0.3659, 0.0884, 0.0, -0.3637, 0.0863,
0.0, -0.3637, 0.0884, 0.0, -0.3658, 0.0908, 0.0, -0.3636, 0.0904, -0.0017, -0.3659, 0.0909, -0.0039, -0.3659, 0.0909, -0.0017, -0.3637, 0.0905, -0.0039, -0.3637, 0.0905, 0.0, -0.3603, 0.0682, 0.0, -0.3615, 0.0711, 0.0, -0.3596, 0.0691, 0.0, -0.3587, 0.0703, 0.0, -0.3615, 0.0732, 0.0, -0.3615, 0.0754, 0.0, -0.3594, 0.0732, 0.0, -0.3582, 0.0763, -0.0017, -0.3605, 0.0679, -0.0039, -0.3605, 0.0679, -0.0017, -0.3588, 0.0704, -0.0039, -0.3588, 0.0704, -0.0017, -0.3583, 0.0765, -0.0039, -0.3583, 0.0765, 0.0, -0.3571, 0.0687, 0.0, -0.3569, 0.071, 0.0, -0.3571, 0.0733, 0.0, -0.3575, 0.0752, -0.0017, -0.3571, 0.0689, -0.0017, -0.3569, 0.0711, -0.0039, -0.3571, 0.0689, -0.0039, -0.3569, 0.0711, -0.0017, -0.357, 0.0733,
-0.0017, -0.3577, 0.0757, -0.0039, -0.357, 0.0733, -0.0039, -0.3577, 0.0757, 0.0, -0.3615, 0.0776, 0.0, -0.3611, 0.0803, 0.0, -0.3594, 0.078, 0.0, -0.3599, 0.0795, 0.0, -0.3618, 0.0812, 0.0, -0.3615, 0.0841, 0.0, -0.3594, 0.0819, 0.0, -0.3594, 0.0841, -0.0017, -0.3614, 0.0805, -0.0039, -0.3614, 0.0805, -0.0017, -0.3591, 0.0777, -0.0017, -0.3596, 0.0793, -0.0039, -0.3591, 0.0777, -0.0039, -0.3596, 0.0793, -0.0017, -0.3619, 0.0812, -0.0039, -0.3619, 0.0812, 0.0, -0.3569, 0.0781, 0.0, -0.3579, 0.0787, 0.0, -0.3553, 0.0773, 0.0, -0.355, 0.0798, 0.0, -0.3572, 0.0819, 0.0, -0.3572, 0.0841, 0.0, -0.355, 0.0819, 0.0, -0.355, 0.0841, -0.0017, -0.357, 0.0781, -0.0017, -0.3581, 0.0788, -0.0039, -0.357, 0.0781,
-0.0039, -0.3581, 0.0788, -0.0017, -0.3551, 0.0772, -0.0039, -0.3551, 0.0772, 0.0, -0.352, 0.0764, 0.0, -0.3508, 0.076, -0.0017, -0.3521, 0.0764, -0.0039, -0.3521, 0.0764, -0.0017, -0.3507, 0.076, -0.0039, -0.3507, 0.076, 0.0, -0.3487, 0.0754, 0.0, -0.3454, 0.0742, 0.0, -0.3465, 0.0746, -0.0017, -0.3485, 0.0753, -0.0039, -0.3485, 0.0753, -0.0017, -0.3455, 0.0742, -0.0017, -0.3466, 0.0746, -0.0039, -0.3455, 0.0742, -0.0039, -0.3466, 0.0746, 0.0, -0.353, 0.0767, 0.0, -0.3529, 0.0798, 0.0, -0.3507, 0.0776, 0.0, -0.3507, 0.0798, 0.0, -0.3529, 0.0819, 0.0, -0.3529, 0.0841, 0.0, -0.3507, 0.0819, 0.0, -0.3507, 0.0841, -0.0017, -0.3532, 0.0767, -0.0039, -0.3532, 0.0767, 0.0, -0.3485, 0.0776, 0.0, -0.3485, 0.0798,
0.0, -0.3463, 0.0776, 0.0, -0.3463, 0.0798, 0.0, -0.3485, 0.0819, 0.0, -0.3485, 0.0841, 0.0, -0.3463, 0.0819, 0.0, -0.3463, 0.0841, 0.0, -0.3615, 0.0863, 0.0, -0.3615, 0.0884, 0.0, -0.3594, 0.0863, 0.0, -0.359, 0.0896, 0.0, -0.3615, 0.09, 0.0, -0.3597, 0.0898, -0.0017, -0.3583, 0.0895, -0.0039, -0.3583, 0.0895, -0.0017, -0.3615, 0.0901, -0.0039, -0.3615, 0.0901, -0.0017, -0.3594, 0.0897, -0.0039, -0.3594, 0.0897, 0.0, -0.3572, 0.0863, 0.0, -0.357, 0.0893, 0.0, -0.355, 0.0863, 0.0, -0.3549, 0.0889, -0.0017, -0.3572, 0.0893, -0.0039, -0.3572, 0.0893, -0.0017, -0.355, 0.0889, -0.0039, -0.355, 0.0889, 0.0, -0.3529, 0.0863, 0.0, -0.3527, 0.0884, 0.0, -0.3507, 0.0863, 0.0, -0.3506, 0.0879,
-0.0017, -0.3529, 0.0885, -0.0039, -0.3529, 0.0885, -0.0017, -0.3507, 0.0879, -0.0039, -0.3507, 0.0879, 0.0, -0.3483, 0.0871, 0.0, -0.3494, 0.0875, 0.0, -0.3461, 0.0864, -0.0017, -0.3482, 0.0871, -0.0017, -0.3493, 0.0875, -0.0039, -0.3482, 0.0871, -0.0039, -0.3493, 0.0875, -0.0017, -0.3463, 0.0864, -0.0039, -0.3463, 0.0864, 0.0, -0.3791, 0.1031, -0.0017, -0.3791, 0.1031, -0.0039, -0.3791, 0.1031, 0.0, -0.3442, -0.0005, 0.0, -0.3442, 0.0016, 0.0, -0.342, -0.0005, 0.0, -0.342, 0.0016, 0.0, -0.345, 0.005, 0.0, -0.344, 0.0056, 0.0, -0.342, 0.0038, 0.0, -0.3423, 0.0066, -0.0017, -0.3452, 0.0049, -0.0017, -0.3442, 0.0055, -0.0039, -0.3452, 0.0049, -0.0039, -0.3442, 0.0055, -0.0017, -0.3423, 0.0066, -0.0039, -0.3423, 0.0066,
0.0, -0.3398, -0.0005, 0.0, -0.3398, 0.0016, 0.0, -0.3377, -0.0005, 0.0, -0.3377, 0.0016, 0.0, -0.3398, 0.0038, 0.0, -0.3398, 0.006, 0.0, -0.3377, 0.0038, 0.0, -0.3377, 0.006, 0.0, -0.3412, 0.0072, 0.0, -0.3438, 0.0144, 0.0, -0.3424, 0.0152, -0.0017, -0.3412, 0.0072, -0.0039, -0.3412, 0.0072, -0.0017, -0.3442, 0.0142, -0.0039, -0.3442, 0.0142, -0.0017, -0.3423, 0.0153, -0.0039, -0.3423, 0.0153, 0.0, -0.34, 0.0079, 0.0, -0.338, 0.009, 0.0, -0.337, 0.0095, -0.0017, -0.3398, 0.008, -0.0039, -0.3398, 0.008, -0.0017, -0.3381, 0.0089, -0.0017, -0.3371, 0.0095, -0.0039, -0.3381, 0.0089, -0.0039, -0.3371, 0.0095, 0.0, -0.3355, -0.0005, 0.0, -0.3355, 0.0016, 0.0, -0.3333, -0.0005, 0.0, -0.3333, 0.0016,
0.0, -0.3355, 0.0038, 0.0, -0.3355, 0.006, 0.0, -0.3333, 0.0038, 0.0, -0.3333, 0.006, 0.0, -0.3312, -0.0005, 0.0, -0.3312, 0.0016, 0.0, -0.329, -0.0005, 0.0, -0.329, 0.0016, 0.0, -0.3312, 0.0038, 0.0, -0.3312, 0.006, 0.0, -0.329, 0.0038, 0.0, -0.329, 0.006, 0.0, -0.3355, 0.0081, 0.0, -0.3357, 0.0101, 0.0, -0.3333, 0.0081, 0.0, -0.3336, 0.0111, 0.0, -0.3326, 0.0116, -0.0017, -0.3355, 0.0102, -0.0039, -0.3355, 0.0102, -0.0017, -0.3337, 0.0111, -0.0039, -0.3337, 0.0111, -0.0017, -0.3327, 0.0116, -0.0039, -0.3327, 0.0116, 0.0, -0.3312, 0.0081, 0.0, -0.3312, 0.0103, 0.0, -0.329, 0.0081, 0.0, -0.329, 0.0103, 0.0, -0.3313, 0.0122, 0.0, -0.3292, 0.0132, 0.0, -0.3282, 0.0137,
-0.0017, -0.3312, 0.0123, -0.0039, -0.3312, 0.0123, -0.0017, -0.3293, 0.0132, -0.0017, -0.3282, 0.0137, -0.0039, -0.3293, 0.0132, -0.0039, -0.3282, 0.0137, 0.0, -0.3442, 0.0168, 0.0, -0.3442, 0.019, 0.0, -0.3414, 0.0158, 0.0, -0.342, 0.019, 0.0, -0.3442, 0.0212, 0.0, -0.3442, 0.0233, 0.0, -0.342, 0.0212, 0.0, -0.342, 0.0233, -0.0017, -0.3412, 0.0159, -0.0039, -0.3412, 0.0159, 0.0, -0.3396, 0.0168, 0.0, -0.3398, 0.019, 0.0, -0.3383, 0.0175, 0.0, -0.3372, 0.0181, 0.0, -0.3398, 0.0212, 0.0, -0.3398, 0.0233, 0.0, -0.3377, 0.0212, 0.0, -0.3377, 0.0233, -0.0017, -0.3398, 0.0167, -0.0039, -0.3398, 0.0167, -0.0017, -0.3382, 0.0176, -0.0017, -0.3371, 0.0181, -0.0039, -0.3382, 0.0176, -0.0039, -0.3371, 0.0181,
0.0, -0.3442, 0.0255, 0.0, -0.3444, 0.0276, 0.0, -0.342, 0.0255, 0.0, -0.3422, 0.0285, 0.0, -0.3411, 0.0289, -0.0017, -0.3442, 0.0277, -0.0039, -0.3442, 0.0277, -0.0017, -0.3423, 0.0284, -0.0039, -0.3423, 0.0284, -0.0017, -0.3412, 0.0289, -0.0039, -0.3412, 0.0289, 0.0, -0.3398, 0.0255, 0.0, -0.3398, 0.0277, 0.0, -0.3377, 0.0255, 0.0, -0.3377, 0.0277, 0.0, -0.34, 0.0294, 0.0, -0.3379, 0.0302, -0.0017, -0.3398, 0.0294, -0.0039, -0.3398, 0.0294, -0.0017, -0.3377, 0.0303, -0.0039, -0.3377, 0.0303, 0.0, -0.3353, 0.019, 0.0, -0.3339, 0.0197, 0.0, -0.3355, 0.0212, 0.0, -0.3355, 0.0233, 0.0, -0.3329, 0.0202, 0.0, -0.3333, 0.0233, -0.0017, -0.3355, 0.0189, -0.0039, -0.3355, 0.0189, -0.0017, -0.3337, 0.0198,
-0.0039, -0.3337, 0.0198, -0.0017, -0.3327, 0.0203, -0.0039, -0.3327, 0.0203, 0.0, -0.3309, 0.021, 0.0, -0.3312, 0.0233, 0.0, -0.3289, 0.0218, 0.0, -0.329, 0.0233, -0.0017, -0.3312, 0.0209, -0.0039, -0.3312, 0.0209, -0.0017, -0.329, 0.0218, -0.0039, -0.329, 0.0218, 0.0, -0.3355, 0.0255, 0.0, -0.3355, 0.0277, 0.0, -0.3333, 0.0255, 0.0, -0.3333, 0.0277, 0.0, -0.3359, 0.0309, 0.0, -0.335, 0.0313, 0.0, -0.3333, 0.0298, 0.0, -0.3335, 0.0318, -0.0017, -0.3363, 0.0308, -0.0017, -0.3352, 0.0312, -0.0039, -0.3363, 0.0308, -0.0039, -0.3352, 0.0312, -0.0017, -0.3333, 0.0318, -0.0039, -0.3333, 0.0318, 0.0, -0.3312, 0.0255, 0.0, -0.3312, 0.0277, 0.0, -0.329, 0.0255, 0.0, -0.329, 0.0277, 0.0, -0.3312, 0.0298,
0.0, -0.3314, 0.0324, 0.0, -0.329, 0.0298, 0.0, -0.3293, 0.0331, -0.0017, -0.3312, 0.0325, -0.0039, -0.3312, 0.0325, -0.0017, -0.3296, 0.033, -0.0039, -0.3296, 0.033, 0.0, -0.3262, -0.0015, 0.0, -0.3268, 0.0016, 0.0, -0.3244, -0.0003, 0.0, -0.3246, 0.0016, 0.0, -0.3268, 0.0038, 0.0, -0.3268, 0.006, 0.0, -0.3246, 0.0038, 0.0, -0.3246, 0.006, -0.0017, -0.3261, -0.0014, -0.0039, -0.3261, -0.0014, -0.0017, -0.3246, -0.0004, -0.0039, -0.3246, -0.0004, 0.0, -0.3234, 0.0004, 0.0, -0.3223, 0.0011, 0.0, -0.3207, 0.0021, 0.0, -0.3225, 0.0038, 0.0, -0.3225, 0.006, 0.0, -0.3197, 0.0027, 0.0, -0.3203, 0.006, -0.0017, -0.3234, 0.0004, -0.0017, -0.3223, 0.0011, -0.0039, -0.3234, 0.0004, -0.0039, -0.3223, 0.0011,
-0.0017, -0.3205, 0.0022, -0.0039, -0.3205, 0.0022, -0.0017, -0.3194, 0.0028, -0.0039, -0.3194, 0.0028, 0.0, -0.3268, 0.0081, 0.0, -0.3268, 0.0103, 0.0, -0.3246, 0.0081, 0.0, -0.3246, 0.0103, 0.0, -0.3268, 0.0125, 0.0, -0.327, 0.0143, 0.0, -0.3246, 0.0125, 0.0, -0.3249, 0.0151, -0.0017, -0.3268, 0.0144, -0.0039, -0.3268, 0.0144, -0.0017, -0.3247, 0.0153, -0.0039, -0.3247, 0.0153, 0.0, -0.3225, 0.0081, 0.0, -0.3225, 0.0103, 0.0, -0.3203, 0.0081, 0.0, -0.3203, 0.0103, 0.0, -0.3225, 0.0125, 0.0, -0.323, 0.0159, 0.0, -0.3203, 0.0125, 0.0, -0.3203, 0.0146, -0.0017, -0.3235, 0.0157, -0.0039, -0.3235, 0.0157, 0.0, -0.3179, 0.0037, 0.0, -0.3181, 0.006, 0.0, -0.3165, 0.0044, 0.0, -0.3154, 0.005,
-0.0017, -0.3181, 0.0036, -0.0039, -0.3181, 0.0036, -0.0017, -0.3164, 0.0045, -0.0017, -0.3153, 0.0051, -0.0039, -0.3164, 0.0045, -0.0039, -0.3153, 0.0051, 0.0, -0.3136, 0.0061, 0.0, -0.3123, 0.0067, -0.0017, -0.3138, 0.0059, -0.0039, -0.3138, 0.0059, -0.0017, -0.3122, 0.0068, -0.0039, -0.3122, 0.0068, 0.0, -0.3181, 0.0081, 0.0, -0.3181, 0.0103, 0.0, -0.316, 0.0081, 0.0, -0.316, 0.0103, 0.0, -0.3181, 0.0125, 0.0, -0.3181, 0.0146, 0.0, -0.316, 0.0125, 0.0, -0.316, 0.0146, 0.0, -0.3138, 0.0081, 0.0, -0.3138, 0.0103, 0.0, -0.3113, 0.0073, 0.0, -0.3116, 0.0103, 0.0, -0.3138, 0.0125, 0.0, -0.3138, 0.0146, 0.0, -0.3116, 0.0125, 0.0, -0.3116, 0.0146, -0.0017, -0.3111, 0.0074, -0.0039, -0.3111, 0.0074,
0.0, -0.3279, 0.0222, 0.0, -0.3268, 0.0227, 0.0, -0.3245, 0.0236, -0.0017, -0.3279, 0.0222, -0.0017, -0.3268, 0.0227, -0.0039, -0.3279, 0.0222, -0.0039, -0.3268, 0.0227, -0.0017, -0.3246, 0.0235, -0.0039, -0.3246, 0.0235, 0.0, -0.3221, 0.0162, 0.0, -0.3205, 0.0168, 0.0, -0.3231, 0.0241, -0.0017, -0.3224, 0.0161, -0.0039, -0.3224, 0.0161, -0.0017, -0.3203, 0.0169, -0.0039, -0.3203, 0.0169, -0.0017, -0.3229, 0.0242, -0.0039, -0.3229, 0.0242, 0.0, -0.3268, 0.0255, 0.0, -0.3268, 0.0277, 0.0, -0.3246, 0.0255, 0.0, -0.3246, 0.0277, 0.0, -0.3268, 0.0298, 0.0, -0.3268, 0.032, 0.0, -0.3246, 0.0298, 0.0, -0.3246, 0.032, 0.0, -0.3221, 0.0245, 0.0, -0.3225, 0.0277, 0.0, -0.3201, 0.0251, 0.0, -0.3203, 0.0277,
0.0, -0.3225, 0.0298, 0.0, -0.3225, 0.032, 0.0, -0.3203, 0.0298, 0.0, -0.3203, 0.032, -0.0017, -0.3218, 0.0246, -0.0039, -0.3218, 0.0246, -0.0017, -0.3203, 0.0251, -0.0039, -0.3203, 0.0251, 0.0, -0.3183, 0.0176, 0.0, -0.3173, 0.018, 0.0, -0.316, 0.0168, 0.0, -0.3161, 0.0185, -0.0017, -0.3184, 0.0176, -0.0017, -0.3173, 0.018, -0.0039, -0.3184, 0.0176, -0.0039, -0.3173, 0.018, -0.0017, -0.316, 0.0185, -0.0039, -0.316, 0.0185, 0.0, -0.3138, 0.0168, 0.0, -0.314, 0.0192, 0.0, -0.3116, 0.0168, 0.0, -0.3117, 0.0199, 0.0, -0.3107, 0.0201, -0.0017, -0.3138, 0.0192, -0.0039, -0.3138, 0.0192, -0.0017, -0.3118, 0.0198, -0.0039, -0.3118, 0.0198, -0.0017, -0.3107, 0.0201, -0.0039, -0.3107, 0.0201, 0.0, -0.318, 0.0257,
0.0, -0.3181, 0.0277, 0.0, -0.3164, 0.0262, 0.0, -0.3155, 0.0265, 0.0, -0.3181, 0.0298, 0.0, -0.3181, 0.032, 0.0, -0.316, 0.0298, 0.0, -0.316, 0.032, -0.0017, -0.3181, 0.0257, -0.0039, -0.3181, 0.0257, -0.0017, -0.3162, 0.0263, -0.0017, -0.3151, 0.0267, -0.0039, -0.3162, 0.0263, -0.0039, -0.3151, 0.0267, 0.0, -0.3136, 0.0271, 0.0, -0.3115, 0.0277, 0.0, -0.3138, 0.0298, 0.0, -0.3138, 0.032, 0.0, -0.3116, 0.0298, 0.0, -0.3116, 0.032, -0.0017, -0.3138, 0.0271, -0.0039, -0.3138, 0.0271, -0.0017, -0.3116, 0.0277, -0.0039, -0.3116, 0.0277, 0.0, -0.3441, 0.0367, 0.0, -0.3428, 0.0372, 0.0, -0.3442, 0.0385, 0.0, -0.3442, 0.0407, 0.0, -0.3417, 0.0377, 0.0, -0.342, 0.0407, -0.0017, -0.3441, 0.0367,
-0.0039, -0.3441, 0.0367, -0.0017, -0.3427, 0.0373, -0.0039, -0.3427, 0.0373, -0.0017, -0.3416, 0.0378, -0.0039, -0.3416, 0.0378, 0.0, -0.3401, 0.0388, 0.0, -0.3389, 0.0401, 0.0, -0.3383, 0.0412, -0.0017, -0.3401, 0.0388, -0.0017, -0.339, 0.04, -0.0039, -0.3401, 0.0388, -0.0039, -0.339, 0.04, -0.0017, -0.3384, 0.0411, -0.0039, -0.3384, 0.0411, 0.0, -0.3442, 0.0429, 0.0, -0.3442, 0.045, 0.0, -0.342, 0.0429, 0.0, -0.342, 0.045, 0.0, -0.3442, 0.0472, 0.0, -0.3442, 0.0494, 0.0, -0.342, 0.0472, 0.0, -0.342, 0.0494, 0.0, -0.3398, 0.0429, 0.0, -0.3398, 0.045, 0.0, -0.3376, 0.0426, 0.0, -0.3373, 0.045, 0.0, -0.3398, 0.0472, 0.0, -0.3398, 0.0494, 0.0, -0.3373, 0.0473, 0.0, -0.3378, 0.0496,
-0.0017, -0.3378, 0.0422, -0.0017, -0.3373, 0.045, -0.0039, -0.3378, 0.0422, -0.0039, -0.3373, 0.045, -0.0017, -0.3373, 0.0472, -0.0017, -0.3377, 0.0494, -0.0039, -0.3373, 0.0472, -0.0039, -0.3377, 0.0494, 0.0, -0.3284, 0.0334, -0.0017, -0.3286, 0.0333, -0.0039, -0.3286, 0.0333, 0.0, -0.3442, 0.0515, 0.0, -0.3442, 0.0537, 0.0, -0.342, 0.0515, 0.0, -0.3418, 0.0544, 0.0, -0.3441, 0.0553, 0.0, -0.343, 0.0548, -0.0017, -0.3418, 0.0544, -0.0039, -0.3418, 0.0544, -0.0017, -0.3441, 0.0553, -0.0039, -0.3441, 0.0553, -0.0017, -0.343, 0.0548, -0.0039, -0.343, 0.0548, 0.0, -0.3391, 0.052, 0.0, -0.3403, 0.0532, 0.0, -0.3385, 0.0508, -0.0017, -0.3392, 0.0521, -0.0017, -0.3403, 0.0532, -0.0039, -0.3392, 0.0521, -0.0039, -0.3403, 0.0532,
-0.0017, -0.3385, 0.0509, -0.0039, -0.3385, 0.0509, 0.0, -0.3309, 0.0672, 0.0, -0.3281, 0.0655, 0.0, -0.3292, 0.0662, -0.0017, -0.3311, 0.0673, -0.0039, -0.3311, 0.0673, -0.0017, -0.3282, 0.0655, -0.0017, -0.3292, 0.0662, -0.0039, -0.3282, 0.0655, -0.0039, -0.3292, 0.0662, 0.0, -0.3269, 0.0338, 0.0, -0.3248, 0.0344, -0.0017, -0.3268, 0.0339, -0.0039, -0.3268, 0.0339, -0.0017, -0.3247, 0.0345, -0.0039, -0.3247, 0.0345, 0.0, -0.3227, 0.0349, 0.0, -0.3206, 0.0353, 0.0, -0.3198, 0.0355, -0.0017, -0.3225, 0.035, -0.0039, -0.3225, 0.035, -0.0017, -0.3212, 0.0352, -0.0017, -0.3201, 0.0354, -0.0039, -0.3212, 0.0352, -0.0039, -0.3201, 0.0354, 0.0, -0.3181, 0.0342, 0.0, -0.3182, 0.0358, 0.0, -0.316, 0.0342, 0.0, -0.3161, 0.0363,
-0.0017, -0.3181, 0.0358, -0.0039, -0.3181, 0.0358, -0.0017, -0.316, 0.0363, -0.0039, -0.316, 0.0363, 0.0, -0.3138, 0.0342, 0.0, -0.3139, 0.0366, 0.0, -0.3116, 0.0342, 0.0, -0.3117, 0.0368, -0.0017, -0.3138, 0.0366, -0.0039, -0.3138, 0.0366, -0.0017, -0.3116, 0.0369, -0.0039, -0.3116, 0.0369, 0.0, -0.3106, 0.0504, -0.0017, -0.3106, 0.0504, -0.0039, -0.3106, 0.0504, 0.0, -0.3193, 0.059, -0.0017, -0.3193, 0.059, -0.0039, -0.3193, 0.059, 0.0, -0.3243, 0.0629, 0.0, -0.3271, 0.0648, 0.0, -0.3268, 0.0667, 0.0, -0.3253, 0.0636, 0.0, -0.3246, 0.0667, -0.0017, -0.3243, 0.0629, -0.0039, -0.3243, 0.0629, -0.0017, -0.3268, 0.0646, -0.0039, -0.3268, 0.0646, -0.0017, -0.3254, 0.0637, -0.0039, -0.3254, 0.0637, 0.0, -0.3217, 0.061,
0.0, -0.3228, 0.0619, 0.0, -0.3204, 0.06, 0.0, -0.3203, 0.0624, 0.0, -0.3225, 0.0646, 0.0, -0.3225, 0.0667, 0.0, -0.3203, 0.0646, 0.0, -0.3203, 0.0667, -0.0017, -0.3217, 0.061, -0.0017, -0.3228, 0.0619, -0.0039, -0.3217, 0.061, -0.0039, -0.3228, 0.0619, -0.0017, -0.3204, 0.06, -0.0039, -0.3204, 0.06, 0.0, -0.3149, 0.0548, 0.0, -0.3171, 0.0569, 0.0, -0.3182, 0.058, 0.0, -0.316, 0.0558, 0.0, -0.316, 0.058, -0.0017, -0.3149, 0.0548, -0.0039, -0.3149, 0.0548, -0.0017, -0.3171, 0.0569, -0.0017, -0.3182, 0.058, -0.0039, -0.3171, 0.0569, -0.0039, -0.3182, 0.058, -0.0017, -0.316, 0.0558, -0.0039, -0.316, 0.0558, 0.0, -0.3127, 0.0526, 0.0, -0.3138, 0.0537, 0.0, -0.3117, 0.0515, 0.0, -0.3116, 0.0537,
0.0, -0.3138, 0.0559, 0.0, -0.3138, 0.058, 0.0, -0.3116, 0.0559, 0.0, -0.3116, 0.058, -0.0017, -0.3127, 0.0526, -0.0017, -0.3138, 0.0537, -0.0039, -0.3127, 0.0526, -0.0039, -0.3138, 0.0537, -0.0017, -0.3117, 0.0515, -0.0039, -0.3117, 0.0515, 0.0, -0.3181, 0.0602, 0.0, -0.3181, 0.0624, 0.0, -0.316, 0.0602, 0.0, -0.316, 0.0624, 0.0, -0.3181, 0.0646, 0.0, -0.3181, 0.0667, 0.0, -0.316, 0.0646, 0.0, -0.316, 0.0667, 0.0, -0.3138, 0.0602, 0.0, -0.3138, 0.0624, 0.0, -0.3116, 0.0602, 0.0, -0.3116, 0.0624, 0.0, -0.3138, 0.0646, 0.0, -0.313, 0.0676, 0.0, -0.3107, 0.0655, 0.0, -0.3118, 0.0665, -0.0017, -0.313, 0.0676, -0.0039, -0.313, 0.0676, -0.0017, -0.3107, 0.0655, -0.0017, -0.3118, 0.0665,
-0.0039, -0.3107, 0.0655, -0.0039, -0.3118, 0.0665, 0.0, -0.3093, 0.0083, 0.0, -0.3095, 0.0103, 0.0, -0.3078, 0.0089, 0.0, -0.3069, 0.0093, 0.0, -0.3095, 0.0125, 0.0, -0.3095, 0.0146, 0.0, -0.3073, 0.0125, 0.0, -0.3073, 0.0146, -0.0017, -0.3095, 0.0082, -0.0039, -0.3095, 0.0082, -0.0017, -0.3077, 0.0089, -0.0017, -0.3066, 0.0094, -0.0039, -0.3077, 0.0089, -0.0039, -0.3066, 0.0094, 0.0, -0.3049, 0.0101, 0.0, -0.3029, 0.0109, 0.0, -0.3051, 0.0125, 0.0, -0.3051, 0.0146, 0.0, -0.3029, 0.0125, 0.0, -0.3029, 0.0146, -0.0017, -0.3051, 0.01, -0.0039, -0.3051, 0.01, -0.0017, -0.3029, 0.0108, -0.0039, -0.3029, 0.0108, 0.0, -0.3017, 0.0113, 0.0, -0.3007, 0.0117, 0.0, -0.3008, 0.0146, 0.0, -0.2985, 0.0125,
0.0, -0.2986, 0.0146, -0.0017, -0.3017, 0.0113, -0.0039, -0.3017, 0.0113, -0.0017, -0.3006, 0.0118, -0.0039, -0.3006, 0.0118, -0.0017, -0.2986, 0.0125, -0.0039, -0.2986, 0.0125, 0.0, -0.2964, 0.0131, 0.0, -0.2964, 0.0146, 0.0, -0.295, 0.0134, 0.0, -0.2941, 0.0136, -0.0017, -0.2964, 0.0131, -0.0039, -0.2964, 0.0131, -0.0017, -0.2948, 0.0134, -0.0017, -0.2937, 0.0137, -0.0039, -0.2948, 0.0134, -0.0039, -0.2937, 0.0137, 0.0, -0.3095, 0.0168, 0.0, -0.3095, 0.019, 0.0, -0.3073, 0.0168, 0.0, -0.3073, 0.019, 0.0, -0.3095, 0.0205, 0.0, -0.3074, 0.021, -0.0017, -0.3095, 0.0205, -0.0039, -0.3095, 0.0205, -0.0017, -0.3073, 0.021, -0.0039, -0.3073, 0.021, 0.0, -0.3051, 0.0168, 0.0, -0.3051, 0.019, 0.0, -0.3029, 0.0168,
0.0, -0.3029, 0.019, 0.0, -0.3053, 0.0215, 0.0, -0.3031, 0.0219, -0.0017, -0.3051, 0.0215, -0.0039, -0.3051, 0.0215, -0.0017, -0.3029, 0.022, -0.0039, -0.3029, 0.022, 0.0, -0.3094, 0.0282, 0.0, -0.3077, 0.0285, 0.0, -0.3095, 0.0298, 0.0, -0.3095, 0.032, 0.0, -0.3069, 0.0287, 0.0, -0.3073, 0.032, -0.0017, -0.3095, 0.0282, -0.0039, -0.3095, 0.0282, -0.0017, -0.3074, 0.0286, -0.0039, -0.3074, 0.0286, -0.0017, -0.3063, 0.0288, -0.0039, -0.3063, 0.0288, 0.0, -0.305, 0.029, 0.0, -0.3051, 0.032, 0.0, -0.3028, 0.0293, 0.0, -0.3029, 0.032, -0.0017, -0.3051, 0.029, -0.0039, -0.3051, 0.029, -0.0017, -0.3029, 0.0293, -0.0039, -0.3029, 0.0293, 0.0, -0.3008, 0.0168, 0.0, -0.3008, 0.019, 0.0, -0.2986, 0.0168,
0.0, -0.2986, 0.019, 0.0, -0.301, 0.0223, 0.0, -0.3003, 0.0224, 0.0, -0.2986, 0.0212, 0.0, -0.2986, 0.0227, -0.0017, -0.3016, 0.0222, -0.0017, -0.3005, 0.0224, -0.0039, -0.3016, 0.0222, -0.0039, -0.3005, 0.0224, -0.0017, -0.2986, 0.0227, -0.0039, -0.2986, 0.0227, 0.0, -0.2964, 0.0168, 0.0, -0.2964, 0.019, 0.0, -0.2943, 0.0168, 0.0, -0.2943, 0.019, 0.0, -0.2964, 0.0212, 0.0, -0.2965, 0.023, 0.0, -0.2943, 0.0212, 0.0, -0.2943, 0.0232, -0.0017, -0.2964, 0.023, -0.0039, -0.2964, 0.023, -0.0017, -0.2943, 0.0232, -0.0039, -0.2943, 0.0232, 0.0, -0.3007, 0.0297, 0.0, -0.3008, 0.032, 0.0, -0.2986, 0.03, 0.0, -0.2986, 0.032, -0.0017, -0.3008, 0.0297, -0.0039, -0.3008, 0.0297, -0.0017, -0.2986, 0.0299,
-0.0039, -0.2986, 0.0299, 0.0, -0.2964, 0.03, 0.0, -0.2964, 0.032, 0.0, -0.2943, 0.0301, 0.0, -0.2943, 0.032, -0.0017, -0.2964, 0.03, -0.0039, -0.2964, 0.03, -0.0017, -0.2943, 0.0301, -0.0039, -0.2943, 0.0301, 0.0, -0.289, 0.0066, -0.0017, -0.2889, 0.0064, -0.0039, -0.2889, 0.0064, 0.0, -0.2882, -0.0006, 0.0, -0.2883, 0.0016, 0.0, -0.2856, -0.0005, 0.0, -0.2856, 0.0016, 0.0, -0.2885, 0.0037, 0.0, -0.2888, 0.0058, 0.0, -0.2856, 0.0038, 0.0, -0.2856, 0.006, -0.0017, -0.2882, -0.0005, -0.0017, -0.2883, 0.0016, -0.0039, -0.2882, -0.0005, -0.0039, -0.2883, 0.0016, -0.0017, -0.2885, 0.0038, -0.0017, -0.2888, 0.0053, -0.0039, -0.2885, 0.0038, -0.0039, -0.2888, 0.0053, 0.0, -0.2892, 0.0081, 0.0, -0.2896, 0.0102,
0.0, -0.2919, 0.014, 0.0, -0.29, 0.0124, 0.0, -0.2902, 0.0144, -0.0017, -0.2892, 0.0081, -0.0017, -0.2896, 0.0103, -0.0039, -0.2892, 0.0081, -0.0039, -0.2896, 0.0103, -0.0017, -0.2921, 0.014, -0.0039, -0.2921, 0.014, -0.0017, -0.29, 0.0125, -0.0017, -0.2902, 0.0144, -0.0039, -0.29, 0.0125, -0.0039, -0.2902, 0.0144, 0.0, -0.2878, 0.0081, 0.0, -0.2878, 0.0103, 0.0, -0.2856, 0.0081, 0.0, -0.2856, 0.0103, 0.0, -0.2878, 0.0125, 0.0, -0.2878, 0.0146, 0.0, -0.2856, 0.0125, 0.0, -0.2856, 0.0146, 0.0, -0.2834, -0.0005, 0.0, -0.2834, 0.0016, 0.0, -0.2812, -0.0005, 0.0, -0.2812, 0.0016, 0.0, -0.2834, 0.0038, 0.0, -0.2834, 0.006, 0.0, -0.2812, 0.0038, 0.0, -0.2801, 0.0063, -0.0017, -0.2802, 0.0069,
-0.0039, -0.2802, 0.0069, 0.0, -0.2794, -0.0005, 0.0, -0.2795, 0.0017, 0.0, -0.2797, 0.0038, 0.0, -0.28, 0.0055, -0.0017, -0.2794, -0.0005, -0.0017, -0.2795, 0.0016, -0.0039, -0.2794, -0.0005, -0.0039, -0.2795, 0.0016, -0.0017, -0.2797, 0.0038, -0.0017, -0.28, 0.0058, -0.0039, -0.2797, 0.0038, -0.0039, -0.28, 0.0058, 0.0, -0.2834, 0.0081, 0.0, -0.2834, 0.0103, 0.0, -0.2804, 0.0083, 0.0, -0.2807, 0.0104, 0.0, -0.2834, 0.0125, 0.0, -0.2834, 0.0146, 0.0, -0.281, 0.0126, 0.0, -0.2813, 0.0147, -0.0017, -0.2803, 0.0081, -0.0017, -0.2806, 0.0103, -0.0039, -0.2803, 0.0081, -0.0039, -0.2806, 0.0103, -0.0017, -0.281, 0.0125, -0.0017, -0.2813, 0.0146, -0.0039, -0.281, 0.0125, -0.0039, -0.2813, 0.0146, 0.0, -0.2921, 0.0168,
0.0, -0.2921, 0.019, 0.0, -0.2899, 0.0168, 0.0, -0.2899, 0.019, 0.0, -0.2921, 0.0212, 0.0, -0.2921, 0.0232, 0.0, -0.2899, 0.0212, 0.0, -0.2899, 0.0232, -0.0017, -0.2921, 0.0232, -0.0039, -0.2921, 0.0232, -0.0017, -0.2899, 0.0232, -0.0039, -0.2899, 0.0232, 0.0, -0.2878, 0.0168, 0.0, -0.2878, 0.019, 0.0, -0.2856, 0.0168, 0.0, -0.2856, 0.019, 0.0, -0.2878, 0.0212, 0.0, -0.2878, 0.0232, 0.0, -0.2856, 0.0212, 0.0, -0.2856, 0.0232, -0.0017, -0.2878, 0.0232, -0.0039, -0.2878, 0.0232, -0.0017, -0.2856, 0.0232, -0.0039, -0.2856, 0.0232, 0.0, -0.2921, 0.0301, 0.0, -0.2921, 0.032, 0.0, -0.2899, 0.0301, 0.0, -0.2899, 0.032, -0.0017, -0.2921, 0.0301, -0.0039, -0.2921, 0.0301, -0.0017, -0.2899, 0.0301,
-0.0039, -0.2899, 0.0301, 0.0, -0.2878, 0.0301, 0.0, -0.2878, 0.032, 0.0, -0.2856, 0.03, 0.0, -0.2856, 0.032, -0.0017, -0.2878, 0.0301, -0.0039, -0.2878, 0.0301, -0.0017, -0.2856, 0.03, -0.0039, -0.2856, 0.03, 0.0, -0.2834, 0.0168, 0.0, -0.2834, 0.019, 0.0, -0.2814, 0.0163, 0.0, -0.2812, 0.019, 0.0, -0.2834, 0.0212, 0.0, -0.2834, 0.0232, 0.0, -0.2812, 0.0212, 0.0, -0.2812, 0.0232, -0.0017, -0.2814, 0.0163, -0.0039, -0.2814, 0.0163, -0.0017, -0.2834, 0.0232, -0.0039, -0.2834, 0.0232, -0.0017, -0.2812, 0.0232, -0.0039, -0.2812, 0.0232, 0.0, -0.2791, 0.0163, 0.0, -0.2791, 0.019, 0.0, -0.2769, 0.0163, 0.0, -0.2769, 0.019, 0.0, -0.2791, 0.0212, 0.0, -0.279, 0.0232, 0.0, -0.2769, 0.0212,
0.0, -0.2769, 0.023, -0.0017, -0.2791, 0.0163, -0.0039, -0.2791, 0.0163, -0.0017, -0.2769, 0.0163, -0.0039, -0.2769, 0.0163, -0.0017, -0.2791, 0.0232, -0.0039, -0.2791, 0.0232, -0.0017, -0.2769, 0.023, -0.0039, -0.2769, 0.023, 0.0, -0.2835, 0.0299, 0.0, -0.2834, 0.032, 0.0, -0.2813, 0.0296, 0.0, -0.2812, 0.032, -0.0017, -0.2834, 0.0299, -0.0039, -0.2834, 0.0299, -0.0017, -0.2812, 0.0296, -0.0039, -0.2812, 0.0296, 0.0, -0.2762, 0.0286, 0.0, -0.2792, 0.0292, 0.0, -0.2791, 0.032, 0.0, -0.2771, 0.0288, 0.0, -0.2769, 0.032, -0.0017, -0.2764, 0.0286, -0.0039, -0.2764, 0.0286, -0.0017, -0.2791, 0.0292, -0.0039, -0.2791, 0.0292, -0.0017, -0.2774, 0.0289, -0.0039, -0.2774, 0.0289, 0.0, -0.3095, 0.0342, 0.0, -0.3095, 0.037,
0.0, -0.3073, 0.0342, 0.0, -0.3074, 0.0371, -0.0017, -0.3095, 0.037, -0.0039, -0.3095, 0.037, -0.0017, -0.3073, 0.0371, -0.0039, -0.3073, 0.0371, 0.0, -0.3051, 0.0342, 0.0, -0.3053, 0.0374, 0.0, -0.3029, 0.0342, 0.0, -0.3029, 0.0363, 0.0, -0.3045, 0.0376, 0.0, -0.301, 0.038, 0.0, -0.3027, 0.0405, -0.0017, -0.3058, 0.0374, -0.0039, -0.3058, 0.0374, -0.0017, -0.3047, 0.0375, -0.0039, -0.3047, 0.0375, -0.0017, -0.301, 0.038, -0.0017, -0.3029, 0.0407, -0.0039, -0.301, 0.038, -0.0039, -0.3029, 0.0407, 0.0, -0.3065, 0.0457, 0.0, -0.3085, 0.0481, 0.0, -0.3096, 0.0493, 0.0, -0.3076, 0.047, 0.0, -0.3073, 0.0494, -0.0017, -0.3065, 0.0457, -0.0039, -0.3065, 0.0457, -0.0017, -0.3085, 0.0481, -0.0017, -0.3096, 0.0493,
-0.0039, -0.3085, 0.0481, -0.0039, -0.3096, 0.0493, -0.0017, -0.3076, 0.0471, -0.0039, -0.3076, 0.0471, 0.0, -0.3047, 0.0432, 0.0, -0.3057, 0.0447, 0.0, -0.3039, 0.0422, 0.0, -0.3029, 0.045, 0.0, -0.3051, 0.0472, 0.0, -0.3051, 0.0494, 0.0, -0.3029, 0.0472, 0.0, -0.3029, 0.0494, -0.0017, -0.3046, 0.0432, -0.0017, -0.3057, 0.0446, -0.0039, -0.3046, 0.0432, -0.0039, -0.3057, 0.0446, -0.0017, -0.3038, 0.0421, -0.0039, -0.3038, 0.0421, 0.0, -0.3008, 0.0342, 0.0, -0.3008, 0.0363, 0.0, -0.2986, 0.0342, 0.0, -0.2986, 0.0363, 0.0, -0.3013, 0.0384, 0.0, -0.3008, 0.0407, 0.0, -0.2986, 0.0385, 0.0, -0.2986, 0.0407, 0.0, -0.2964, 0.0342, 0.0, -0.2964, 0.0363, 0.0, -0.2943, 0.0342, 0.0, -0.2943, 0.0363,
0.0, -0.2964, 0.0385, 0.0, -0.2964, 0.0407, 0.0, -0.2943, 0.0385, 0.0, -0.2943, 0.0407, 0.0, -0.3008, 0.0429, 0.0, -0.3008, 0.045, 0.0, -0.2986, 0.0429, 0.0, -0.2986, 0.045, 0.0, -0.3008, 0.0472, 0.0, -0.3008, 0.0494, 0.0, -0.2986, 0.0472, 0.0, -0.2986, 0.0494, 0.0, -0.2964, 0.0429, 0.0, -0.2964, 0.045, 0.0, -0.2931, 0.0436, 0.0, -0.2942, 0.0453, 0.0, -0.2956, 0.0477, 0.0, -0.2969, 0.0496, 0.0, -0.295, 0.0466, -0.0017, -0.2932, 0.0438, -0.0017, -0.294, 0.045, -0.0039, -0.2932, 0.0438, -0.0039, -0.294, 0.045, -0.0017, -0.2957, 0.0477, -0.0017, -0.2968, 0.0494, -0.0039, -0.2957, 0.0477, -0.0039, -0.2968, 0.0494, -0.0017, -0.295, 0.0467, -0.0039, -0.295, 0.0467, 0.0, -0.3095, 0.0515,
0.0, -0.3095, 0.0537, 0.0, -0.3073, 0.0515, 0.0, -0.3073, 0.0537, 0.0, -0.3095, 0.0559, 0.0, -0.3095, 0.058, 0.0, -0.3073, 0.0559, 0.0, -0.3073, 0.058, 0.0, -0.3051, 0.0515, 0.0, -0.3051, 0.0537, 0.0, -0.3029, 0.0515, 0.0, -0.3029, 0.0537, 0.0, -0.3051, 0.0559, 0.0, -0.3042, 0.0588, 0.0, -0.3022, 0.0565, 0.0, -0.3033, 0.0577, -0.0017, -0.3043, 0.0589, -0.0039, -0.3043, 0.0589, -0.0017, -0.3022, 0.0565, -0.0017, -0.3033, 0.0578, -0.0039, -0.3022, 0.0565, -0.0039, -0.3033, 0.0578, 0.0, -0.3095, 0.0602, 0.0, -0.3084, 0.0634, 0.0, -0.3063, 0.0611, 0.0, -0.3073, 0.0623, 0.0, -0.3095, 0.0645, -0.0017, -0.3084, 0.0634, -0.0039, -0.3084, 0.0634, -0.0017, -0.3063, 0.0612, -0.0017, -0.3074, 0.0624,
-0.0039, -0.3063, 0.0612, -0.0039, -0.3074, 0.0624, -0.0017, -0.3095, 0.0644, -0.0039, -0.3095, 0.0644, 0.0, -0.3053, 0.0601, -0.0017, -0.3053, 0.0601, -0.0039, -0.3053, 0.0601, 0.0, -0.3008, 0.0515, 0.0, -0.3002, 0.0541, 0.0, -0.2983, 0.0516, 0.0, -0.2994, 0.053, 0.0, -0.3013, 0.0554, -0.0017, -0.3002, 0.0541, -0.0039, -0.3002, 0.0541, -0.0017, -0.2983, 0.0516, -0.0017, -0.2994, 0.053, -0.0039, -0.2983, 0.0516, -0.0039, -0.2994, 0.053, -0.0017, -0.3013, 0.0554, -0.0039, -0.3013, 0.0554, 0.0, -0.2975, 0.0505, -0.0017, -0.2975, 0.0505, -0.0039, -0.2975, 0.0505, 0.0, -0.2921, 0.0342, 0.0, -0.2921, 0.0363, 0.0, -0.2899, 0.0342, 0.0, -0.2902, 0.0369, 0.0, -0.2908, 0.0391, 0.0, -0.2916, 0.041, 0.0, -0.2905, 0.0382,
-0.0017, -0.2902, 0.0369, -0.0039, -0.2902, 0.0369, -0.0017, -0.291, 0.0396, -0.0017, -0.2915, 0.0407, -0.0039, -0.291, 0.0396, -0.0039, -0.2915, 0.0407, -0.0017, -0.2906, 0.0385, -0.0039, -0.2906, 0.0385, 0.0, -0.2878, 0.0342, 0.0, -0.2877, 0.0369, 0.0, -0.2856, 0.0342, 0.0, -0.2855, 0.0367, -0.0017, -0.2878, 0.0369, -0.0039, -0.2878, 0.0369, -0.0017, -0.2856, 0.0367, -0.0039, -0.2856, 0.0367, 0.0, -0.2925, 0.0426, -0.0017, -0.2926, 0.0428, -0.0039, -0.2926, 0.0428, 0.0, -0.2834, 0.0342, 0.0, -0.2833, 0.0363, 0.0, -0.2812, 0.0342, 0.0, -0.2812, 0.0359, -0.0017, -0.2834, 0.0363, -0.0039, -0.2834, 0.0363, -0.0017, -0.2812, 0.0359, -0.0039, -0.2812, 0.0359, 0.0, -0.2787, 0.0353, 0.0, -0.2796, 0.0355, 0.0, -0.2767, 0.0347,
-0.0017, -0.2783, 0.0352, -0.0017, -0.2794, 0.0355, -0.0039, -0.2783, 0.0352, -0.0039, -0.2794, 0.0355, -0.0017, -0.2769, 0.0348, -0.0039, -0.2769, 0.0348, 0.0, -0.3443, 0.0738, 0.0, -0.3442, 0.0754, 0.0, -0.3423, 0.0729, 0.0, -0.342, 0.0754, -0.0017, -0.3442, 0.0737, -0.0039, -0.3442, 0.0737, -0.0017, -0.342, 0.0728, -0.0039, -0.342, 0.0728, 0.0, -0.3395, 0.0716, 0.0, -0.338, 0.0708, 0.0, -0.3405, 0.0721, 0.0, -0.3398, 0.0754, 0.0, -0.3377, 0.0732, 0.0, -0.3377, 0.0754, -0.0017, -0.3397, 0.0717, -0.0039, -0.3397, 0.0717, -0.0017, -0.3377, 0.0707, -0.0039, -0.3377, 0.0707, -0.0017, -0.3408, 0.0722, -0.0039, -0.3408, 0.0722, 0.0, -0.3442, 0.0776, 0.0, -0.3442, 0.0798, 0.0, -0.342, 0.0776, 0.0, -0.342, 0.0798,
0.0, -0.3442, 0.0819, 0.0, -0.3442, 0.0841, 0.0, -0.342, 0.0819, 0.0, -0.3419, 0.0849, -0.0017, -0.3419, 0.0848, -0.0039, -0.3419, 0.0848, 0.0, -0.3398, 0.0776, 0.0, -0.3398, 0.0798, 0.0, -0.3377, 0.0776, 0.0, -0.3377, 0.0798, 0.0, -0.3398, 0.0819, 0.0, -0.3396, 0.0841, 0.0, -0.3371, 0.0831, 0.0, -0.3381, 0.0835, -0.0017, -0.3398, 0.0841, -0.0039, -0.3398, 0.0841, -0.0017, -0.3368, 0.0829, -0.0017, -0.3379, 0.0834, -0.0039, -0.3368, 0.0829, -0.0039, -0.3379, 0.0834, 0.0, -0.3351, 0.0694, 0.0, -0.3361, 0.0699, 0.0, -0.3336, 0.0687, 0.0, -0.3333, 0.0711, 0.0, -0.3355, 0.0732, 0.0, -0.3355, 0.0754, 0.0, -0.3333, 0.0732, 0.0, -0.3333, 0.0754, -0.0017, -0.3353, 0.0695, -0.0017, -0.3364, 0.0701,
-0.0039, -0.3353, 0.0695, -0.0039, -0.3364, 0.0701, -0.0017, -0.3333, 0.0685, -0.0039, -0.3333, 0.0685, 0.0, -0.3319, 0.0677, 0.0, -0.3312, 0.0711, 0.0, -0.329, 0.0689, 0.0, -0.329, 0.0711, 0.0, -0.3312, 0.0732, 0.0, -0.3312, 0.0754, 0.0, -0.329, 0.0732, 0.0, -0.329, 0.0754, -0.0017, -0.3321, 0.0679, -0.0039, -0.3321, 0.0679, 0.0, -0.3355, 0.0776, 0.0, -0.3355, 0.0798, 0.0, -0.3333, 0.0776, 0.0, -0.3327, 0.0809, 0.0, -0.3352, 0.0822, 0.0, -0.3336, 0.0814, -0.0017, -0.3324, 0.0808, -0.0039, -0.3324, 0.0808, -0.0017, -0.3355, 0.0823, -0.0039, -0.3355, 0.0823, -0.0017, -0.3335, 0.0813, -0.0039, -0.3335, 0.0813, 0.0, -0.3312, 0.0776, 0.0, -0.3308, 0.08, 0.0, -0.3284, 0.0786, 0.0, -0.3294, 0.0792,
-0.0017, -0.3311, 0.0802, -0.0039, -0.3311, 0.0802, -0.0017, -0.3282, 0.0785, -0.0017, -0.3293, 0.0791, -0.0039, -0.3282, 0.0785, -0.0039, -0.3293, 0.0791, 0.0, -0.3441, 0.0856, 0.0, -0.343, 0.0852, -0.0017, -0.3442, 0.0856, -0.0039, -0.3442, 0.0856, -0.0017, -0.343, 0.0852, -0.0039, -0.343, 0.0852, 0.0, -0.3268, 0.0689, 0.0, -0.3268, 0.0711, 0.0, -0.3246, 0.0689, 0.0, -0.3246, 0.0711, 0.0, -0.3268, 0.0732, 0.0, -0.3268, 0.0754, 0.0, -0.3246, 0.0732, 0.0, -0.3243, 0.0761, -0.0017, -0.3242, 0.0761, -0.0039, -0.3242, 0.0761, 0.0, -0.3225, 0.0689, 0.0, -0.3225, 0.0711, 0.0, -0.3203, 0.0689, 0.0, -0.3203, 0.0711, 0.0, -0.3217, 0.0743, 0.0, -0.3227, 0.0751, 0.0, -0.32, 0.0732, -0.0017, -0.3215, 0.0742,
-0.0017, -0.3226, 0.075, -0.0039, -0.3215, 0.0742, -0.0039, -0.3226, 0.075, -0.0017, -0.3203, 0.0734, -0.0039, -0.3203, 0.0734, 0.0, -0.3265, 0.0775, 0.0, -0.3254, 0.0768, -0.0017, -0.3268, 0.0777, -0.0039, -0.3268, 0.0777, -0.0017, -0.3253, 0.0768, -0.0039, -0.3253, 0.0768, 0.0, -0.3181, 0.0689, 0.0, -0.3178, 0.0716, 0.0, -0.3153, 0.0696, 0.0, -0.3164, 0.0706, 0.0, -0.3189, 0.0724, -0.0017, -0.3178, 0.0716, -0.0039, -0.3178, 0.0716, -0.0017, -0.3153, 0.0696, -0.0017, -0.3164, 0.0705, -0.0039, -0.3153, 0.0696, -0.0039, -0.3164, 0.0705, -0.0017, -0.3189, 0.0724, -0.0039, -0.3189, 0.0724, 0.0, -0.3141, 0.0686, -0.0017, -0.3141, 0.0685, -0.0039, -0.3141, 0.0685, 0.0, -0.2652, 0.0157, 0.0, -0.2639, 0.0154, -0.0017, -0.2653, 0.0157,
-0.0039, -0.2653, 0.0157, -0.0017, -0.2639, 0.0154, -0.0039, -0.2639, 0.0154, 0.0, -0.2618, 0.015, 0.0, -0.2596, 0.0146, -0.0017, -0.2617, 0.015, -0.0039, -0.2617, 0.015, -0.0017, -0.2595, 0.0146, -0.0039, -0.2595, 0.0146, 0.0, -0.2747, 0.0163, 0.0, -0.2747, 0.019, 0.0, -0.2726, 0.0163, 0.0, -0.2726, 0.019, 0.0, -0.2747, 0.0212, 0.0, -0.2747, 0.0226, 0.0, -0.2725, 0.0221, 0.0, -0.2735, 0.0223, -0.0017, -0.2747, 0.0163, -0.0039, -0.2747, 0.0163, -0.0017, -0.2726, 0.0163, -0.0039, -0.2726, 0.0163, -0.0017, -0.2747, 0.0226, -0.0039, -0.2747, 0.0226, -0.0017, -0.2723, 0.022, -0.0017, -0.2734, 0.0223, -0.0039, -0.2723, 0.022, -0.0039, -0.2734, 0.0223, 0.0, -0.2704, 0.0163, 0.0, -0.2704, 0.019, 0.0, -0.2683, 0.0161,
0.0, -0.2682, 0.019, 0.0, -0.2702, 0.0216, 0.0, -0.2681, 0.0211, -0.0017, -0.2704, 0.0163, -0.0039, -0.2704, 0.0163, -0.0017, -0.2682, 0.0161, -0.0039, -0.2682, 0.0161, -0.0017, -0.2704, 0.0216, -0.0039, -0.2704, 0.0216, -0.0017, -0.2682, 0.0211, -0.0039, -0.2682, 0.0211, 0.0, -0.2748, 0.0283, 0.0, -0.2727, 0.0277, 0.0, -0.2747, 0.0298, 0.0, -0.2747, 0.032, 0.0, -0.2726, 0.0298, 0.0, -0.2721, 0.0331, -0.0017, -0.2747, 0.0283, -0.0039, -0.2747, 0.0283, -0.0017, -0.2726, 0.0277, -0.0039, -0.2726, 0.0277, -0.0017, -0.2717, 0.033, -0.0039, -0.2717, 0.033, 0.0, -0.2706, 0.0271, 0.0, -0.2678, 0.0262, 0.0, -0.2687, 0.0265, 0.0, -0.2704, 0.0298, 0.0, -0.2701, 0.0324, 0.0, -0.2682, 0.0298, 0.0, -0.2681, 0.0316,
-0.0017, -0.2704, 0.027, -0.0039, -0.2704, 0.027, -0.0017, -0.2681, 0.0263, -0.0017, -0.2692, 0.0266, -0.0039, -0.2681, 0.0263, -0.0039, -0.2692, 0.0266, -0.0017, -0.2704, 0.0325, -0.0039, -0.2704, 0.0325, -0.0017, -0.2682, 0.0317, -0.0039, -0.2682, 0.0317, 0.0, -0.2661, 0.0158, 0.0, -0.2661, 0.019, 0.0, -0.2639, 0.0168, 0.0, -0.2637, 0.0199, 0.0, -0.266, 0.0206, 0.0, -0.2647, 0.0202, -0.0017, -0.2664, 0.0159, -0.0039, -0.2664, 0.0159, -0.0017, -0.2635, 0.0198, -0.0039, -0.2635, 0.0198, -0.0017, -0.266, 0.0206, -0.0039, -0.266, 0.0206, -0.0017, -0.2646, 0.0202, -0.0039, -0.2646, 0.0202, 0.0, -0.2617, 0.0168, 0.0, -0.2615, 0.0192, 0.0, -0.2595, 0.0168, 0.0, -0.2594, 0.0185, 0.0, -0.2614, 0.024, 0.0, -0.2598, 0.0233,
-0.0017, -0.2617, 0.0193, -0.0039, -0.2617, 0.0193, -0.0017, -0.2595, 0.0185, -0.0039, -0.2595, 0.0185, -0.0017, -0.2616, 0.0241, -0.0039, -0.2616, 0.0241, -0.0017, -0.2595, 0.0232, -0.0039, -0.2595, 0.0232, 0.0, -0.2662, 0.0256, 0.0, -0.2661, 0.0277, 0.0, -0.2641, 0.025, 0.0, -0.2639, 0.0277, 0.0, -0.2658, 0.0307, 0.0, -0.2669, 0.0311, 0.0, -0.2637, 0.0297, -0.0017, -0.2661, 0.0256, -0.0039, -0.2661, 0.0256, -0.0017, -0.2639, 0.0249, -0.0039, -0.2639, 0.0249, -0.0017, -0.2657, 0.0306, -0.0017, -0.2668, 0.0311, -0.0039, -0.2657, 0.0306, -0.0039, -0.2668, 0.0311, -0.0017, -0.2639, 0.0298, -0.0039, -0.2639, 0.0298, 0.0, -0.2622, 0.0243, 0.0, -0.2613, 0.0286, 0.0, -0.2595, 0.0255, 0.0, -0.2593, 0.0275, 0.0, -0.2623, 0.0291,
-0.0017, -0.2627, 0.0245, -0.0017, -0.2611, 0.0285, -0.0039, -0.2627, 0.0245, -0.0039, -0.2611, 0.0285, -0.0017, -0.2595, 0.0276, -0.0039, -0.2595, 0.0276, -0.0017, -0.2622, 0.029, -0.0039, -0.2622, 0.029, 0.0, -0.2575, 0.0142, 0.0, -0.2541, 0.0135, 0.0, -0.2551, 0.0138, -0.0017, -0.2574, 0.0142, -0.0039, -0.2574, 0.0142, -0.0017, -0.2542, 0.0136, -0.0017, -0.2552, 0.0138, -0.0039, -0.2542, 0.0136, -0.0039, -0.2552, 0.0138, 0.0, -0.2531, 0.0133, 0.0, -0.2527, 0.0155, 0.0, -0.2509, 0.0128, 0.0, -0.2506, 0.0144, -0.0017, -0.253, 0.0133, -0.0017, -0.2526, 0.0154, -0.0039, -0.253, 0.0133, -0.0039, -0.2526, 0.0154, -0.0017, -0.2509, 0.0128, -0.0017, -0.2508, 0.0145, -0.0039, -0.2509, 0.0128, -0.0039, -0.2508, 0.0145, 0.0, -0.2497, 0.0131,
0.0, -0.2496, 0.0137, -0.0017, -0.2467, 0.0121, -0.0017, -0.2496, 0.0137, -0.0039, -0.2467, 0.0121, -0.0039, -0.2496, 0.0137, 0.0, -0.2572, 0.0176, 0.0, -0.2583, 0.018, 0.0, -0.255, 0.0166, 0.0, -0.2568, 0.0219, 0.0, -0.2578, 0.0224, 0.0, -0.2554, 0.0212, 0.0, -0.2544, 0.0245, -0.0017, -0.2571, 0.0176, -0.0017, -0.2582, 0.018, -0.0039, -0.2571, 0.0176, -0.0039, -0.2582, 0.018, -0.0017, -0.2552, 0.0167, -0.0039, -0.2552, 0.0167, -0.0017, -0.2569, 0.022, -0.0017, -0.258, 0.0225, -0.0039, -0.2569, 0.022, -0.0039, -0.258, 0.0225, -0.0017, -0.2552, 0.0211, -0.0017, -0.2542, 0.0244, -0.0039, -0.2552, 0.0211, -0.0039, -0.2542, 0.0244, 0.0, -0.2538, 0.016, 0.0, -0.2524, 0.0197, 0.0, -0.2511, 0.0191, 0.0, -0.2534, 0.0203,
0.0, -0.2527, 0.0233, 0.0, -0.2505, 0.0217, 0.0, -0.2516, 0.0225, -0.0017, -0.2537, 0.016, -0.0017, -0.2525, 0.0198, -0.0039, -0.2537, 0.016, -0.0039, -0.2525, 0.0198, -0.0017, -0.2509, 0.019, -0.0039, -0.2509, 0.019, -0.0017, -0.2536, 0.0203, -0.0017, -0.253, 0.0235, -0.0039, -0.2536, 0.0203, -0.0039, -0.253, 0.0235, -0.0017, -0.2505, 0.0217, -0.0017, -0.2516, 0.0225, -0.0039, -0.2505, 0.0217, -0.0039, -0.2516, 0.0225, 0.0, -0.257, 0.0262, 0.0, -0.2581, 0.0268, 0.0, -0.2554, 0.0252, -0.0017, -0.257, 0.0262, -0.0017, -0.2581, 0.0268, -0.0039, -0.257, 0.0262, -0.0039, -0.2581, 0.0268, -0.0017, -0.2553, 0.0251, -0.0039, -0.2553, 0.0251, 0.0, -0.248, 0.0176, 0.0, -0.2486, 0.019, 0.0, -0.2449, 0.0163, 0.0, -0.2469, 0.0186,
0.0, -0.2491, 0.0207, -0.0017, -0.2482, 0.0176, -0.0017, -0.2486, 0.0189, -0.0039, -0.2482, 0.0176, -0.0039, -0.2486, 0.0189, -0.0017, -0.2449, 0.0163, -0.0017, -0.2469, 0.0186, -0.0039, -0.2449, 0.0163, -0.0039, -0.2469, 0.0186, -0.0017, -0.2491, 0.0206, -0.0039, -0.2491, 0.0206, 0.0, -0.2746, 0.034, 0.0, -0.273, 0.0335, -0.0017, -0.2747, 0.0341, -0.0039, -0.2747, 0.0341, -0.0017, -0.2728, 0.0334, -0.0039, -0.2728, 0.0334, 0.0, -0.2196, -0.0005, 0.0, -0.2196, 0.0016, 0.0, -0.2196, 0.0038, 0.0, -0.2196, 0.006, -0.0017, -0.2196, -0.0005, -0.0017, -0.2196, 0.0016, -0.0039, -0.2196, -0.0005, -0.0039, -0.2196, 0.0016, -0.0017, -0.2196, 0.0038, -0.0017, -0.2196, 0.006, -0.0039, -0.2196, 0.0038, -0.0039, -0.2196, 0.006, 0.0, -0.2183, -0.0005,
0.0, -0.2183, 0.0016, 0.0, -0.2161, -0.0005, 0.0, -0.2161, 0.0016, 0.0, -0.2183, 0.0038, 0.0, -0.2183, 0.006, 0.0, -0.2161, 0.0038, 0.0, -0.2161, 0.006, 0.0, -0.2196, 0.0081, 0.0, -0.2196, 0.0103, 0.0, -0.2196, 0.0125, 0.0, -0.2196, 0.0146, -0.0017, -0.2196, 0.0081, -0.0017, -0.2196, 0.0103, -0.0039, -0.2196, 0.0081, -0.0039, -0.2196, 0.0103, -0.0017, -0.2196, 0.0125, -0.0017, -0.2196, 0.0146, -0.0039, -0.2196, 0.0125, -0.0039, -0.2196, 0.0146, 0.0, -0.2183, 0.0081, 0.0, -0.2183, 0.0103, 0.0, -0.2161, 0.0081, 0.0, -0.2161, 0.0103, 0.0, -0.2183, 0.0125, 0.0, -0.2183, 0.0146, 0.0, -0.2161, 0.0125, 0.0, -0.2161, 0.0146, 0.0, -0.214, -0.0005, 0.0, -0.214, 0.0016, 0.0, -0.2118, -0.0005,
0.0, -0.2118, 0.0016, 0.0, -0.214, 0.0038, 0.0, -0.214, 0.006, 0.0, -0.2118, 0.0038, 0.0, -0.2118, 0.006, 0.0, -0.2088, -0.0005, 0.0, -0.2088, 0.0016, 0.0, -0.2088, 0.0038, 0.0, -0.2088, 0.006, -0.0017, -0.2088, -0.0005, -0.0017, -0.2088, 0.0016, -0.0039, -0.2088, -0.0005, -0.0039, -0.2088, 0.0016, -0.0017, -0.2088, 0.0038, -0.0017, -0.2088, 0.006, -0.0039, -0.2088, 0.0038, -0.0039, -0.2088, 0.006, 0.0, -0.214, 0.0081, 0.0, -0.214, 0.0103, 0.0, -0.2118, 0.0081, 0.0, -0.2118, 0.0103, 0.0, -0.214, 0.0125, 0.0, -0.214, 0.0146, 0.0, -0.2118, 0.0125, 0.0, -0.2118, 0.0146, 0.0, -0.2088, 0.0085, 0.0, -0.2096, 0.0103, 0.0, -0.2075, 0.0085, 0.0, -0.2075, 0.0103, 0.0, -0.2096, 0.0125,
0.0, -0.2096, 0.0146, 0.0, -0.2075, 0.0125, 0.0, -0.2075, 0.0146, -0.0017, -0.2088, 0.0085, -0.0039, -0.2088, 0.0085, -0.0017, -0.2075, 0.0085, -0.0039, -0.2075, 0.0085, 0.0, -0.2196, 0.0168, 0.0, -0.2196, 0.019, 0.0, -0.2196, 0.0212, 0.0, -0.2196, 0.0233, -0.0017, -0.2196, 0.0168, -0.0017, -0.2196, 0.019, -0.0039, -0.2196, 0.0168, -0.0039, -0.2196, 0.019, -0.0017, -0.2196, 0.0212, -0.0017, -0.2196, 0.0233, -0.0039, -0.2196, 0.0212, -0.0039, -0.2196, 0.0233, 0.0, -0.2183, 0.0168, 0.0, -0.2183, 0.019, 0.0, -0.2161, 0.0168, 0.0, -0.2161, 0.019, 0.0, -0.2183, 0.0212, 0.0, -0.2183, 0.0233, 0.0, -0.2161, 0.0212, 0.0, -0.2161, 0.0233, 0.0, -0.2196, 0.0255, 0.0, -0.2196, 0.0277, 0.0, -0.2196, 0.0298,
0.0, -0.2196, 0.032, -0.0017, -0.2196, 0.0255, -0.0017, -0.2196, 0.0277, -0.0039, -0.2196, 0.0255, -0.0039, -0.2196, 0.0277, -0.0017, -0.2196, 0.0298, -0.0017, -0.2196, 0.032, -0.0039, -0.2196, 0.0298, -0.0039, -0.2196, 0.032, 0.0, -0.2183, 0.0255, 0.0, -0.2183, 0.0277, 0.0, -0.2161, 0.0255, 0.0, -0.2161, 0.0277, 0.0, -0.2183, 0.0298, 0.0, -0.2183, 0.032, 0.0, -0.2161, 0.0298, 0.0, -0.2161, 0.032, 0.0, -0.214, 0.0168, 0.0, -0.214, 0.019, 0.0, -0.2118, 0.0168, 0.0, -0.2118, 0.019, 0.0, -0.214, 0.0212, 0.0, -0.214, 0.0233, 0.0, -0.2118, 0.0212, 0.0, -0.2118, 0.0233, 0.0, -0.2096, 0.0168, 0.0, -0.2096, 0.019, 0.0, -0.2075, 0.0168, 0.0, -0.2075, 0.019, 0.0, -0.2088, 0.0203,
0.0, -0.2088, 0.0233, 0.0, -0.2075, 0.0203, -0.0017, -0.2088, 0.0203, -0.0017, -0.2088, 0.0233, -0.0039, -0.2088, 0.0203, -0.0039, -0.2088, 0.0233, -0.0017, -0.2075, 0.0203, -0.0039, -0.2075, 0.0203, 0.0, -0.214, 0.0255, 0.0, -0.214, 0.0277, 0.0, -0.2118, 0.0255, 0.0, -0.2118, 0.0277, 0.0, -0.214, 0.0298, 0.0, -0.214, 0.032, 0.0, -0.2118, 0.0298, 0.0, -0.2118, 0.032, 0.0, -0.2088, 0.0255, 0.0, -0.2088, 0.0277, 0.0, -0.2088, 0.0298, 0.0, -0.2088, 0.0319, -0.0017, -0.2088, 0.0255, -0.0017, -0.2088, 0.0277, -0.0039, -0.2088, 0.0255, -0.0039, -0.2088, 0.0277, -0.0017, -0.2088, 0.0298, -0.0017, -0.2088, 0.032, -0.0039, -0.2088, 0.0298, -0.0039, -0.2088, 0.032, 0.0, -0.2196, 0.0342, 0.0, -0.2196, 0.0363,
0.0, -0.2196, 0.0385, 0.0, -0.2196, 0.0407, -0.0017, -0.2196, 0.0342, -0.0017, -0.2196, 0.0363, -0.0039, -0.2196, 0.0342, -0.0039, -0.2196, 0.0363, -0.0017, -0.2196, 0.0385, -0.0017, -0.2196, 0.0407, -0.0039, -0.2196, 0.0385, -0.0039, -0.2196, 0.0407, 0.0, -0.2183, 0.0342, 0.0, -0.2183, 0.0363, 0.0, -0.2161, 0.0342, 0.0, -0.2161, 0.0363, 0.0, -0.2183, 0.0385, 0.0, -0.2183, 0.0407, 0.0, -0.2161, 0.0385, 0.0, -0.2161, 0.0407, 0.0, -0.2196, 0.0429, 0.0, -0.2196, 0.045, 0.0, -0.2196, 0.0472, 0.0, -0.2196, 0.0487, -0.0017, -0.2196, 0.0429, -0.0017, -0.2196, 0.045, -0.0039, -0.2196, 0.0429, -0.0039, -0.2196, 0.045, -0.0017, -0.2196, 0.0472, -0.0017, -0.2196, 0.0487, -0.0039, -0.2196, 0.0472, -0.0039, -0.2196, 0.0487,
0.0, -0.2183, 0.0429, 0.0, -0.2183, 0.045, 0.0, -0.2161, 0.0429, 0.0, -0.2161, 0.045, 0.0, -0.2183, 0.0472, 0.0, -0.2183, 0.0487, 0.0, -0.2161, 0.0472, 0.0, -0.2161, 0.0487, -0.0017, -0.2183, 0.0487, -0.0039, -0.2183, 0.0487, -0.0017, -0.2161, 0.0487, -0.0039, -0.2161, 0.0487, 0.0, -0.214, 0.0342, 0.0, -0.214, 0.0363, 0.0, -0.2118, 0.0342, 0.0, -0.2118, 0.0363, 0.0, -0.214, 0.0385, 0.0, -0.214, 0.0407, 0.0, -0.2118, 0.0385, 0.0, -0.2118, 0.0407, 0.0, -0.2085, 0.0339, 0.0, -0.2096, 0.0363, 0.0, -0.2083, 0.0347, 0.0, -0.2076, 0.0361, 0.0, -0.2096, 0.0385, 0.0, -0.2096, 0.0407, 0.0, -0.2075, 0.0385, 0.0, -0.2075, 0.0407, -0.0017, -0.2086, 0.0334, -0.0039, -0.2086, 0.0334,
-0.0017, -0.2083, 0.0345, -0.0017, -0.2076, 0.0361, -0.0039, -0.2083, 0.0345, -0.0039, -0.2076, 0.0361, 0.0, -0.214, 0.0429, 0.0, -0.214, 0.045, 0.0, -0.2118, 0.0429, 0.0, -0.2118, 0.045, 0.0, -0.214, 0.0472, 0.0, -0.214, 0.0487, 0.0, -0.2118, 0.0472, 0.0, -0.2118, 0.0487, -0.0017, -0.214, 0.0487, -0.0039, -0.214, 0.0487, -0.0017, -0.2118, 0.0487, -0.0039, -0.2118, 0.0487, 0.0, -0.2096, 0.0429, 0.0, -0.2096, 0.045, 0.0, -0.2075, 0.0429, 0.0, -0.2075, 0.045, 0.0, -0.2096, 0.0472, 0.0, -0.2096, 0.0487, 0.0, -0.2075, 0.0472, 0.0, -0.2075, 0.0487, -0.0017, -0.2096, 0.0487, -0.0039, -0.2096, 0.0487, -0.0017, -0.2075, 0.0487, -0.0039, -0.2075, 0.0487, 0.0, -0.2053, 0.0085, 0.0, -0.2053, 0.0103,
0.0, -0.2031, 0.0085, 0.0, -0.2031, 0.0103, 0.0, -0.2053, 0.0125, 0.0, -0.2053, 0.0146, 0.0, -0.2031, 0.0125, 0.0, -0.2031, 0.0146, -0.0017, -0.2053, 0.0085, -0.0039, -0.2053, 0.0085, -0.0017, -0.2031, 0.0085, -0.0039, -0.2031, 0.0085, 0.0, -0.2009, 0.0085, 0.0, -0.2009, 0.0103, 0.0, -0.1988, 0.0085, 0.0, -0.1988, 0.0103, 0.0, -0.2009, 0.0125, 0.0, -0.2009, 0.0146, 0.0, -0.1988, 0.0125, 0.0, -0.1988, 0.0146, -0.0017, -0.2009, 0.0085, -0.0039, -0.2009, 0.0085, -0.0017, -0.1988, 0.0085, -0.0039, -0.1988, 0.0085, 0.0, -0.1966, 0.0085, 0.0, -0.1966, 0.0103, 0.0, -0.1944, 0.0085, 0.0, -0.1944, 0.0103, 0.0, -0.1966, 0.0125, 0.0, -0.1966, 0.0146, 0.0, -0.1944, 0.0125, 0.0, -0.1944, 0.0146,
-0.0017, -0.1966, 0.0085, -0.0039, -0.1966, 0.0085, -0.0017, -0.1944, 0.0085, -0.0039, -0.1944, 0.0085, 0.0, -0.1923, 0.0085, 0.0, -0.1923, 0.0103, 0.0, -0.1901, 0.0085, 0.0, -0.1901, 0.0103, 0.0, -0.1923, 0.0125, 0.0, -0.1923, 0.0146, 0.0, -0.1901, 0.0125, 0.0, -0.1901, 0.0146, -0.0017, -0.1923, 0.0085, -0.0039, -0.1923, 0.0085, -0.0017, -0.1901, 0.0085, -0.0039, -0.1901, 0.0085, 0.0, -0.2053, 0.0168, 0.0, -0.2053, 0.019, 0.0, -0.2031, 0.0168, 0.0, -0.2031, 0.019, 0.0, -0.2053, 0.0203, 0.0, -0.2031, 0.0203, -0.0017, -0.2053, 0.0203, -0.0039, -0.2053, 0.0203, -0.0017, -0.2031, 0.0203, -0.0039, -0.2031, 0.0203, 0.0, -0.2009, 0.0168, 0.0, -0.2009, 0.019, 0.0, -0.1988, 0.0168, 0.0, -0.1988, 0.019,
0.0, -0.2009, 0.0203, 0.0, -0.1988, 0.0203, -0.0017, -0.2009, 0.0203, -0.0039, -0.2009, 0.0203, -0.0017, -0.1988, 0.0203, -0.0039, -0.1988, 0.0203, 0.0, -0.1966, 0.0168, 0.0, -0.1966, 0.019, 0.0, -0.1944, 0.0168, 0.0, -0.1944, 0.019, 0.0, -0.1966, 0.0203, 0.0, -0.1944, 0.0203, -0.0017, -0.1966, 0.0203, -0.0039, -0.1966, 0.0203, -0.0017, -0.1944, 0.0203, -0.0039, -0.1944, 0.0203, 0.0, -0.1923, 0.0168, 0.0, -0.1923, 0.019, 0.0, -0.1901, 0.0168, 0.0, -0.1901, 0.019, 0.0, -0.1923, 0.0203, 0.0, -0.1901, 0.0203, -0.0017, -0.1923, 0.0203, -0.0039, -0.1923, 0.0203, -0.0017, -0.1901, 0.0203, -0.0039, -0.1901, 0.0203, 0.0, -0.1879, 0.0085, 0.0, -0.1879, 0.0103, 0.0, -0.1858, 0.0085, 0.0, -0.1858, 0.0103,
0.0, -0.1879, 0.0125, 0.0, -0.1879, 0.0146, 0.0, -0.1858, 0.0125, 0.0, -0.1858, 0.0146, -0.0017, -0.1879, 0.0085, -0.0039, -0.1879, 0.0085, -0.0017, -0.1858, 0.0085, -0.0039, -0.1858, 0.0085, 0.0, -0.1836, 0.0085, 0.0, -0.1836, 0.0103, 0.0, -0.1814, 0.0085, 0.0, -0.1814, 0.0103, 0.0, -0.1836, 0.0125, 0.0, -0.1836, 0.0146, 0.0, -0.1814, 0.0125, 0.0, -0.1814, 0.0146, -0.0017, -0.1836, 0.0085, -0.0039, -0.1836, 0.0085, -0.0017, -0.1814, 0.0085, -0.0039, -0.1814, 0.0085, 0.0, -0.1792, 0.0085, 0.0, -0.1792, 0.0103, 0.0, -0.1771, 0.0085, 0.0, -0.1771, 0.0103, 0.0, -0.1792, 0.0125, 0.0, -0.1792, 0.0146, 0.0, -0.1771, 0.0125, 0.0, -0.1771, 0.0146, -0.0017, -0.1792, 0.0085, -0.0039, -0.1792, 0.0085,
-0.0017, -0.1771, 0.0085, -0.0039, -0.1771, 0.0085, 0.0, -0.1749, 0.0085, 0.0, -0.1749, 0.0103, 0.0, -0.1727, 0.0085, 0.0, -0.1727, 0.0103, 0.0, -0.1749, 0.0125, 0.0, -0.1749, 0.0146, 0.0, -0.1727, 0.0125, 0.0, -0.1727, 0.0146, -0.0017, -0.1749, 0.0085, -0.0039, -0.1749, 0.0085, -0.0017, -0.1727, 0.0085, -0.0039, -0.1727, 0.0085, 0.0, -0.1879, 0.0168, 0.0, -0.1879, 0.019, 0.0, -0.1858, 0.0168, 0.0, -0.1858, 0.019, 0.0, -0.1879, 0.0203, 0.0, -0.1858, 0.0203, -0.0017, -0.1879, 0.0203, -0.0039, -0.1879, 0.0203, -0.0017, -0.1858, 0.0203, -0.0039, -0.1858, 0.0203, 0.0, -0.1836, 0.0168, 0.0, -0.1836, 0.019, 0.0, -0.1814, 0.0168, 0.0, -0.1814, 0.019, 0.0, -0.1836, 0.0203, 0.0, -0.1814, 0.0203,
-0.0017, -0.1836, 0.0203, -0.0039, -0.1836, 0.0203, -0.0017, -0.1814, 0.0203, -0.0039, -0.1814, 0.0203, 0.0, -0.1792, 0.0168, 0.0, -0.1792, 0.019, 0.0, -0.1771, 0.0168, 0.0, -0.1771, 0.019, 0.0, -0.1792, 0.0203, 0.0, -0.1771, 0.0203, -0.0017, -0.1792, 0.0203, -0.0039, -0.1792, 0.0203, -0.0017, -0.1771, 0.0203, -0.0039, -0.1771, 0.0203, 0.0, -0.1749, 0.0168, 0.0, -0.1749, 0.019, 0.0, -0.1727, 0.0168, 0.0, -0.1727, 0.019, 0.0, -0.1749, 0.0203, 0.0, -0.1727, 0.0203, -0.0017, -0.1749, 0.0203, -0.0039, -0.1749, 0.0203, -0.0017, -0.1727, 0.0203, -0.0039, -0.1727, 0.0203, 0.0, -0.206, 0.0372, 0.0, -0.2051, 0.0375, 0.0, -0.2053, 0.0407, 0.0, -0.203, 0.0378, 0.0, -0.2031, 0.0407, -0.0017, -0.2059, 0.0372,
-0.0039, -0.2059, 0.0372, -0.0017, -0.2048, 0.0376, -0.0039, -0.2048, 0.0376, -0.0017, -0.2031, 0.0378, -0.0039, -0.2031, 0.0378, 0.0, -0.2009, 0.0379, 0.0, -0.2009, 0.0407, 0.0, -0.1988, 0.0379, 0.0, -0.1988, 0.0407, -0.0017, -0.2009, 0.0379, -0.0039, -0.2009, 0.0379, -0.0017, -0.1988, 0.0379, -0.0039, -0.1988, 0.0379, 0.0, -0.2053, 0.0429, 0.0, -0.2053, 0.045, 0.0, -0.2031, 0.0429, 0.0, -0.2031, 0.045, 0.0, -0.2053, 0.0472, 0.0, -0.2053, 0.0487, 0.0, -0.2031, 0.0472, 0.0, -0.2031, 0.0487, -0.0017, -0.2053, 0.0487, -0.0039, -0.2053, 0.0487, -0.0017, -0.2031, 0.0487, -0.0039, -0.2031, 0.0487, 0.0, -0.2009, 0.0429, 0.0, -0.2009, 0.045, 0.0, -0.1988, 0.0429, 0.0, -0.1988, 0.045, 0.0, -0.2009, 0.0472,
0.0, -0.2009, 0.0487, 0.0, -0.1988, 0.0472, 0.0, -0.1988, 0.0487, -0.0017, -0.2009, 0.0487, -0.0039, -0.2009, 0.0487, -0.0017, -0.1988, 0.0487, -0.0039, -0.1988, 0.0487, 0.0, -0.1966, 0.0379, 0.0, -0.1966, 0.0407, 0.0, -0.1944, 0.0379, 0.0, -0.1944, 0.0407, -0.0017, -0.1966, 0.0379, -0.0039, -0.1966, 0.0379, -0.0017, -0.1944, 0.0379, -0.0039, -0.1944, 0.0379, 0.0, -0.1923, 0.0379, 0.0, -0.1923, 0.0407, 0.0, -0.1901, 0.0379, 0.0, -0.1901, 0.0407, -0.0017, -0.1923, 0.0379, -0.0039, -0.1923, 0.0379, -0.0017, -0.1901, 0.0379, -0.0039, -0.1901, 0.0379, 0.0, -0.1966, 0.0429, 0.0, -0.1966, 0.045, 0.0, -0.1944, 0.0429, 0.0, -0.1944, 0.045, 0.0, -0.1966, 0.0472, 0.0, -0.1966, 0.0487, 0.0, -0.1944, 0.0472,
0.0, -0.1944, 0.0487, -0.0017, -0.1966, 0.0487, -0.0039, -0.1966, 0.0487, -0.0017, -0.1944, 0.0487, -0.0039, -0.1944, 0.0487, 0.0, -0.1923, 0.0429, 0.0, -0.1923, 0.045, 0.0, -0.1901, 0.0429, 0.0, -0.1901, 0.045, 0.0, -0.1923, 0.0472, 0.0, -0.1923, 0.0487, 0.0, -0.1901, 0.0472, 0.0, -0.1901, 0.0487, -0.0017, -0.1923, 0.0487, -0.0039, -0.1923, 0.0487, -0.0017, -0.1901, 0.0487, -0.0039, -0.1901, 0.0487, 0.0, -0.1879, 0.0379, 0.0, -0.1879, 0.0407, 0.0, -0.1858, 0.0379, 0.0, -0.1858, 0.0407, -0.0017, -0.1879, 0.0379, -0.0039, -0.1879, 0.0379, -0.0017, -0.1858, 0.0379, -0.0039, -0.1858, 0.0379, 0.0, -0.1836, 0.0379, 0.0, -0.1836, 0.0407, 0.0, -0.1814, 0.0379, 0.0, -0.1814, 0.0407, -0.0017, -0.1836, 0.0379,
-0.0039, -0.1836, 0.0379, -0.0017, -0.1814, 0.0379, -0.0039, -0.1814, 0.0379, 0.0, -0.1879, 0.0429, 0.0, -0.1879, 0.045, 0.0, -0.1858, 0.0429, 0.0, -0.1858, 0.045, 0.0, -0.1879, 0.0472, 0.0, -0.1879, 0.0487, 0.0, -0.1858, 0.0472, 0.0, -0.1858, 0.0487, -0.0017, -0.1879, 0.0487, -0.0039, -0.1879, 0.0487, -0.0017, -0.1858, 0.0487, -0.0039, -0.1858, 0.0487, 0.0, -0.1836, 0.0429, 0.0, -0.1836, 0.045, 0.0, -0.1814, 0.0429, 0.0, -0.1814, 0.045, 0.0, -0.1836, 0.0472, 0.0, -0.1836, 0.0487, 0.0, -0.1814, 0.0472, 0.0, -0.1814, 0.0487, -0.0017, -0.1836, 0.0487, -0.0039, -0.1836, 0.0487, -0.0017, -0.1814, 0.0487, -0.0039, -0.1814, 0.0487, 0.0, -0.1792, 0.0379, 0.0, -0.1792, 0.0407, 0.0, -0.1771, 0.0379,
0.0, -0.1771, 0.0407, -0.0017, -0.1792, 0.0379, -0.0039, -0.1792, 0.0379, -0.0017, -0.1771, 0.0379, -0.0039, -0.1771, 0.0379, 0.0, -0.1749, 0.0379, 0.0, -0.1749, 0.0407, 0.0, -0.1727, 0.0379, 0.0, -0.1727, 0.0407, -0.0017, -0.1749, 0.0379, -0.0039, -0.1749, 0.0379, -0.0017, -0.1727, 0.0379, -0.0039, -0.1727, 0.0379, 0.0, -0.1792, 0.0429, 0.0, -0.1792, 0.045, 0.0, -0.1771, 0.0429, 0.0, -0.1771, 0.045, 0.0, -0.1792, 0.0472, 0.0, -0.1792, 0.0487, 0.0, -0.1771, 0.0472, 0.0, -0.1771, 0.0487, -0.0017, -0.1792, 0.0487, -0.0039, -0.1792, 0.0487, -0.0017, -0.1771, 0.0487, -0.0039, -0.1771, 0.0487, 0.0, -0.1749, 0.0429, 0.0, -0.1749, 0.045, 0.0, -0.1727, 0.0429, 0.0, -0.1727, 0.045, 0.0, -0.1749, 0.0472,
0.0, -0.1749, 0.0487, 0.0, -0.1727, 0.0472, 0.0, -0.1727, 0.0487, -0.0017, -0.1749, 0.0487, -0.0039, -0.1749, 0.0487, -0.0017, -0.1727, 0.0487, -0.0039, -0.1727, 0.0487, 0.0, -0.1706, 0.0085, 0.0, -0.1706, 0.0103, 0.0, -0.1684, 0.0085, 0.0, -0.1684, 0.0103, 0.0, -0.1706, 0.0125, 0.0, -0.1706, 0.0146, 0.0, -0.1684, 0.0125, 0.0, -0.1684, 0.0146, -0.0017, -0.1706, 0.0085, -0.0039, -0.1706, 0.0085, -0.0017, -0.1684, 0.0085, -0.0039, -0.1684, 0.0085, 0.0, -0.1662, 0.0085, 0.0, -0.1662, 0.0103, 0.0, -0.1641, 0.0085, 0.0, -0.1641, 0.0103, 0.0, -0.1662, 0.0125, 0.0, -0.1662, 0.0146, 0.0, -0.1641, 0.0125, 0.0, -0.1641, 0.0146, -0.0017, -0.1662, 0.0085, -0.0039, -0.1662, 0.0085, -0.0017, -0.1641, 0.0085,
-0.0039, -0.1641, 0.0085, 0.0, -0.1627, 0.0085, 0.0, -0.1627, 0.0103, 0.0, -0.1627, 0.0125, 0.0, -0.1627, 0.0146, -0.0017, -0.1627, 0.0085, -0.0017, -0.1627, 0.0103, -0.0039, -0.1627, 0.0085, -0.0039, -0.1627, 0.0103, -0.0017, -0.1627, 0.0125, -0.0017, -0.1627, 0.0146, -0.0039, -0.1627, 0.0125, -0.0039, -0.1627, 0.0146, 0.0, -0.1706, 0.0168, 0.0, -0.1706, 0.019, 0.0, -0.1684, 0.0168, 0.0, -0.1684, 0.019, 0.0, -0.1706, 0.0203, 0.0, -0.1684, 0.0203, -0.0017, -0.1706, 0.0203, -0.0039, -0.1706, 0.0203, -0.0017, -0.1684, 0.0203, -0.0039, -0.1684, 0.0203, 0.0, -0.1662, 0.0168, 0.0, -0.1662, 0.019, 0.0, -0.1641, 0.0168, 0.0, -0.1641, 0.019, 0.0, -0.1662, 0.0203, 0.0, -0.1641, 0.0203, -0.0017, -0.1662, 0.0203,
-0.0039, -0.1662, 0.0203, -0.0017, -0.1641, 0.0203, -0.0039, -0.1641, 0.0203, 0.0, -0.1627, 0.0168, 0.0, -0.1627, 0.019, 0.0, -0.1627, 0.0203, -0.0017, -0.1627, 0.0168, -0.0017, -0.1627, 0.019, -0.0039, -0.1627, 0.0168, -0.0039, -0.1627, 0.019, -0.0017, -0.1627, 0.0203, -0.0039, -0.1627, 0.0203, 0.0, -0.151, -0.0005, 0.0, -0.151, 0.0016, 0.0, -0.151, 0.0038, 0.0, -0.151, 0.006, -0.0017, -0.151, -0.0005, -0.0017, -0.151, 0.0016, -0.0039, -0.151, -0.0005, -0.0039, -0.151, 0.0016, -0.0017, -0.151, 0.0038, -0.0017, -0.151, 0.006, -0.0039, -0.151, 0.0038, -0.0039, -0.151, 0.006, 0.0, -0.1489, -0.0005, 0.0, -0.1489, 0.0016, 0.0, -0.1467, -0.0005, 0.0, -0.1467, 0.0016, 0.0, -0.1489, 0.0038, 0.0, -0.1489, 0.006,
0.0, -0.1467, 0.0038, 0.0, -0.1467, 0.006, 0.0, -0.151, 0.0081, 0.0, -0.151, 0.0103, 0.0, -0.151, 0.0125, 0.0, -0.151, 0.0146, -0.0017, -0.151, 0.0081, -0.0017, -0.151, 0.0103, -0.0039, -0.151, 0.0081, -0.0039, -0.151, 0.0103, -0.0017, -0.151, 0.0125, -0.0017, -0.151, 0.0146, -0.0039, -0.151, 0.0125, -0.0039, -0.151, 0.0146, 0.0, -0.1489, 0.0081, 0.0, -0.1489, 0.0103, 0.0, -0.1467, 0.0081, 0.0, -0.1467, 0.0103, 0.0, -0.1489, 0.0125, 0.0, -0.1489, 0.0146, 0.0, -0.1467, 0.0125, 0.0, -0.1467, 0.0146, 0.0, -0.1445, -0.0005, 0.0, -0.1445, 0.0016, 0.0, -0.1424, -0.0005, 0.0, -0.1424, 0.0016, 0.0, -0.1445, 0.0038, 0.0, -0.1445, 0.006, 0.0, -0.1424, 0.0038, 0.0, -0.1424, 0.006,
0.0, -0.1392, -0.0005, 0.0, -0.1392, 0.0016, 0.0, -0.1392, 0.0038, 0.0, -0.1392, 0.006, -0.0017, -0.1392, -0.0005, -0.0017, -0.1392, 0.0016, -0.0039, -0.1392, -0.0005, -0.0039, -0.1392, 0.0016, -0.0017, -0.1392, 0.0038, -0.0017, -0.1392, 0.006, -0.0039, -0.1392, 0.0038, -0.0039, -0.1392, 0.006, 0.0, -0.1445, 0.0081, 0.0, -0.1445, 0.0103, 0.0, -0.1424, 0.0081, 0.0, -0.1424, 0.0103, 0.0, -0.1445, 0.0125, 0.0, -0.1445, 0.0146, 0.0, -0.1424, 0.0125, 0.0, -0.1424, 0.0146, 0.0, -0.1392, 0.0081, 0.0, -0.1392, 0.0103, 0.0, -0.1392, 0.0125, 0.0, -0.1392, 0.0146, -0.0017, -0.1392, 0.0081, -0.0017, -0.1392, 0.0103, -0.0039, -0.1392, 0.0081, -0.0039, -0.1392, 0.0103, -0.0017, -0.1392, 0.0125, -0.0017, -0.1392, 0.0146,
-0.0039, -0.1392, 0.0125, -0.0039, -0.1392, 0.0146, 0.0, -0.151, 0.0168, 0.0, -0.151, 0.019, 0.0, -0.151, 0.0212, 0.0, -0.151, 0.0233, -0.0017, -0.151, 0.0168, -0.0017, -0.151, 0.019, -0.0039, -0.151, 0.0168, -0.0039, -0.151, 0.019, -0.0017, -0.151, 0.0212, -0.0017, -0.151, 0.0233, -0.0039, -0.151, 0.0212, -0.0039, -0.151, 0.0233, 0.0, -0.1489, 0.0168, 0.0, -0.1489, 0.019, 0.0, -0.1467, 0.0168, 0.0, -0.1467, 0.019, 0.0, -0.1489, 0.0212, 0.0, -0.1489, 0.0233, 0.0, -0.1467, 0.0212, 0.0, -0.1467, 0.0233, 0.0, -0.151, 0.0255, 0.0, -0.151, 0.0277, 0.0, -0.151, 0.0298, 0.0, -0.151, 0.032, -0.0017, -0.151, 0.0255, -0.0017, -0.151, 0.0277, -0.0039, -0.151, 0.0255, -0.0039, -0.151, 0.0277,
-0.0017, -0.151, 0.0298, -0.0017, -0.151, 0.032, -0.0039, -0.151, 0.0298, -0.0039, -0.151, 0.032, 0.0, -0.1489, 0.0255, 0.0, -0.1489, 0.0277, 0.0, -0.1467, 0.0255, 0.0, -0.1467, 0.0277, 0.0, -0.1489, 0.0298, 0.0, -0.1489, 0.032, 0.0, -0.1467, 0.0298, 0.0, -0.1467, 0.032, 0.0, -0.1445, 0.0168, 0.0, -0.1445, 0.019, 0.0, -0.1424, 0.0168, 0.0, -0.1424, 0.019, 0.0, -0.1445, 0.0212, 0.0, -0.1445, 0.0233, 0.0, -0.1424, 0.0212, 0.0, -0.1424, 0.0233, 0.0, -0.1392, 0.0168, 0.0, -0.1392, 0.019, 0.0, -0.1392, 0.0212, 0.0, -0.1392, 0.0233, -0.0017, -0.1392, 0.0168, -0.0017, -0.1392, 0.019, -0.0039, -0.1392, 0.0168, -0.0039, -0.1392, 0.019, -0.0017, -0.1392, 0.0212, -0.0017, -0.1392, 0.0233,
-0.0039, -0.1392, 0.0212, -0.0039, -0.1392, 0.0233, 0.0, -0.1445, 0.0255, 0.0, -0.1445, 0.0277, 0.0, -0.1424, 0.0255, 0.0, -0.1424, 0.0277, 0.0, -0.1445, 0.0298, 0.0, -0.1445, 0.032, 0.0, -0.1424, 0.0298, 0.0, -0.1424, 0.032, 0.0, -0.1392, 0.0255, 0.0, -0.1392, 0.0277, 0.0, -0.1392, 0.0298, 0.0, -0.1392, 0.032, -0.0017, -0.1392, 0.0255, -0.0017, -0.1392, 0.0277, -0.0039, -0.1392, 0.0255, -0.0039, -0.1392, 0.0277, -0.0017, -0.1392, 0.0298, -0.0017, -0.1392, 0.032, -0.0039, -0.1392, 0.0298, -0.0039, -0.1392, 0.032, 0.0, -0.1706, 0.0379, 0.0, -0.1706, 0.0407, 0.0, -0.1684, 0.0379, 0.0, -0.1684, 0.0407, -0.0017, -0.1706, 0.0379, -0.0039, -0.1706, 0.0379, -0.0017, -0.1684, 0.0379, -0.0039, -0.1684, 0.0379,
0.0, -0.1662, 0.0379, 0.0, -0.1662, 0.0407, 0.0, -0.1641, 0.0379, 0.0, -0.1641, 0.0407, -0.0017, -0.1662, 0.0379, -0.0039, -0.1662, 0.0379, -0.0017, -0.1641, 0.0379, -0.0039, -0.1641, 0.0379, 0.0, -0.1706, 0.0429, 0.0, -0.1706, 0.045, 0.0, -0.1684, 0.0429, 0.0, -0.1684, 0.045, 0.0, -0.1706, 0.0472, 0.0, -0.1706, 0.0487, 0.0, -0.1684, 0.0472, 0.0, -0.1684, 0.0487, -0.0017, -0.1706, 0.0487, -0.0039, -0.1706, 0.0487, -0.0017, -0.1684, 0.0487, -0.0039, -0.1684, 0.0487, 0.0, -0.1662, 0.0429, 0.0, -0.1662, 0.045, 0.0, -0.1641, 0.0429, 0.0, -0.1641, 0.045, 0.0, -0.1662, 0.0472, 0.0, -0.1662, 0.0487, 0.0, -0.1641, 0.0472, 0.0, -0.1641, 0.0487, -0.0017, -0.1662, 0.0487, -0.0039, -0.1662, 0.0487,
-0.0017, -0.1641, 0.0487, -0.0039, -0.1641, 0.0487, 0.0, -0.1619, 0.0379, 0.0, -0.1619, 0.0407, 0.0, -0.1588, 0.0379, 0.0, -0.1588, 0.0407, -0.0017, -0.1619, 0.0379, -0.0039, -0.1619, 0.0379, -0.0017, -0.1588, 0.0379, -0.0017, -0.1588, 0.0407, -0.0039, -0.1588, 0.0379, -0.0039, -0.1588, 0.0407, 0.0, -0.1619, 0.0429, 0.0, -0.1619, 0.045, 0.0, -0.1588, 0.0429, 0.0, -0.1588, 0.045, 0.0, -0.1619, 0.0472, 0.0, -0.1619, 0.0487, 0.0, -0.1588, 0.0472, 0.0, -0.1588, 0.0487, -0.0017, -0.1588, 0.0429, -0.0017, -0.1588, 0.045, -0.0039, -0.1588, 0.0429, -0.0039, -0.1588, 0.045, -0.0017, -0.1619, 0.0487, -0.0039, -0.1619, 0.0487, -0.0017, -0.1588, 0.0472, -0.0017, -0.1588, 0.0487, -0.0039, -0.1588, 0.0472, -0.0039, -0.1588, 0.0487,
0.0, -0.151, 0.0342, 0.0, -0.151, 0.0363, 0.0, -0.151, 0.0385, 0.0, -0.151, 0.0407, -0.0017, -0.151, 0.0342, -0.0017, -0.151, 0.0363, -0.0039, -0.151, 0.0342, -0.0039, -0.151, 0.0363, -0.0017, -0.151, 0.0385, -0.0017, -0.151, 0.0407, -0.0039, -0.151, 0.0385, -0.0039, -0.151, 0.0407, 0.0, -0.1489, 0.0342, 0.0, -0.1489, 0.0363, 0.0, -0.1467, 0.0342, 0.0, -0.1467, 0.0363, 0.0, -0.1489, 0.0385, 0.0, -0.1489, 0.0407, 0.0, -0.1467, 0.0385, 0.0, -0.1467, 0.0407, 0.0, -0.151, 0.0429, 0.0, -0.151, 0.045, 0.0, -0.151, 0.0472, 0.0, -0.151, 0.0487, -0.0017, -0.151, 0.0429, -0.0017, -0.151, 0.045, -0.0039, -0.151, 0.0429, -0.0039, -0.151, 0.045, -0.0017, -0.151, 0.0472, -0.0017, -0.151, 0.0487,
-0.0039, -0.151, 0.0472, -0.0039, -0.151, 0.0487, 0.0, -0.1489, 0.0429, 0.0, -0.1489, 0.045, 0.0, -0.1467, 0.0429, 0.0, -0.1467, 0.045, 0.0, -0.1489, 0.0472, 0.0, -0.1489, 0.0487, 0.0, -0.1467, 0.0472, 0.0, -0.1467, 0.0487, -0.0017, -0.1489, 0.0487, -0.0039, -0.1489, 0.0487, -0.0017, -0.1467, 0.0487, -0.0039, -0.1467, 0.0487, 0.0, -0.1445, 0.0342, 0.0, -0.1445, 0.0363, 0.0, -0.1424, 0.0342, 0.0, -0.1424, 0.0363, 0.0, -0.1445, 0.0385, 0.0, -0.1445, 0.0407, 0.0, -0.1424, 0.0385, 0.0, -0.1424, 0.0407, 0.0, -0.1392, 0.0342, 0.0, -0.1392, 0.0363, 0.0, -0.1392, 0.0385, 0.0, -0.1392, 0.0407, -0.0017, -0.1392, 0.0342, -0.0017, -0.1392, 0.0363, -0.0039, -0.1392, 0.0342, -0.0039, -0.1392, 0.0363,
-0.0017, -0.1392, 0.0385, -0.0017, -0.1392, 0.0407, -0.0039, -0.1392, 0.0385, -0.0039, -0.1392, 0.0407, 0.0, -0.1445, 0.0429, 0.0, -0.1445, 0.045, 0.0, -0.1424, 0.0429, 0.0, -0.1424, 0.045, 0.0, -0.1445, 0.0472, 0.0, -0.1445, 0.0487, 0.0, -0.1424, 0.0472, 0.0, -0.1424, 0.0487, -0.0017, -0.1445, 0.0487, -0.0039, -0.1445, 0.0487, -0.0017, -0.1424, 0.0487, -0.0039, -0.1424, 0.0487, 0.0, -0.1392, 0.0429, 0.0, -0.1392, 0.045, 0.0, -0.1392, 0.0472, 0.0, -0.1392, 0.0487, -0.0017, -0.1392, 0.0429, -0.0017, -0.1392, 0.045, -0.0039, -0.1392, 0.0429, -0.0039, -0.1392, 0.045, -0.0017, -0.1392, 0.0472, -0.0017, -0.1392, 0.0487, -0.0039, -0.1392, 0.0472, -0.0039, -0.1392, 0.0487, 0.0, -0.0775, -0.0005, 0.0, -0.0775, 0.0016,
0.0, -0.0775, 0.0038, 0.0, -0.0775, 0.006, -0.0017, -0.0775, -0.0005, -0.0017, -0.0775, 0.0016, -0.0039, -0.0775, -0.0005, -0.0039, -0.0775, 0.0016, -0.0017, -0.0775, 0.0038, -0.0017, -0.0775, 0.006, -0.0039, -0.0775, 0.0038, -0.0039, -0.0775, 0.006, 0.0, -0.0775, 0.0081, 0.0, -0.0775, 0.0103, 0.0, -0.0775, 0.0125, 0.0, -0.0775, 0.0146, -0.0017, -0.0775, 0.0081, -0.0017, -0.0775, 0.0103, -0.0039, -0.0775, 0.0081, -0.0039, -0.0775, 0.0103, -0.0017, -0.0775, 0.0125, -0.0017, -0.0775, 0.0146, -0.0039, -0.0775, 0.0125, -0.0039, -0.0775, 0.0146, 0.0, -0.0751, -0.0005, 0.0, -0.0751, 0.0016, 0.0, -0.0729, -0.0005, 0.0, -0.0729, 0.0016, 0.0, -0.0751, 0.0038, 0.0, -0.0751, 0.006, 0.0, -0.0729, 0.0038, 0.0, -0.0729, 0.006,
0.0, -0.0707, -0.0005, 0.0, -0.0707, 0.0016, 0.0, -0.0686, -0.0005, 0.0, -0.0686, 0.0016, 0.0, -0.0707, 0.0038, 0.0, -0.0707, 0.006, 0.0, -0.0686, 0.0038, 0.0, -0.0686, 0.006, 0.0, -0.0751, 0.0081, 0.0, -0.0751, 0.0103, 0.0, -0.0729, 0.0081, 0.0, -0.0729, 0.0103, 0.0, -0.0751, 0.0125, 0.0, -0.0751, 0.0146, 0.0, -0.0729, 0.0125, 0.0, -0.0729, 0.0146, 0.0, -0.0707, 0.0081, 0.0, -0.0707, 0.0103, 0.0, -0.0686, 0.0081, 0.0, -0.0686, 0.0103, 0.0, -0.0707, 0.0125, 0.0, -0.0707, 0.0146, 0.0, -0.0686, 0.0125, 0.0, -0.0686, 0.0146, 0.0, -0.0775, 0.0168, 0.0, -0.0775, 0.019, 0.0, -0.0775, 0.0212, 0.0, -0.0775, 0.0233, -0.0017, -0.0775, 0.0168, -0.0017, -0.0775, 0.019,
-0.0039, -0.0775, 0.0168, -0.0039, -0.0775, 0.019, -0.0017, -0.0775, 0.0212, -0.0017, -0.0775, 0.0233, -0.0039, -0.0775, 0.0212, -0.0039, -0.0775, 0.0233, 0.0, -0.0775, 0.0255, 0.0, -0.0775, 0.0277, 0.0, -0.0775, 0.0298, 0.0, -0.0775, 0.032, -0.0017, -0.0775, 0.0255, -0.0017, -0.0775, 0.0277, -0.0039, -0.0775, 0.0255, -0.0039, -0.0775, 0.0277, -0.0017, -0.0775, 0.0298, -0.0017, -0.0775, 0.032, -0.0039, -0.0775, 0.0298, -0.0039, -0.0775, 0.032, 0.0, -0.0751, 0.0168, 0.0, -0.0751, 0.019, 0.0, -0.0729, 0.0168, 0.0, -0.0729, 0.019, 0.0, -0.0751, 0.0212, 0.0, -0.0751, 0.0233, 0.0, -0.0729, 0.0212, 0.0, -0.0729, 0.0233, 0.0, -0.0707, 0.0168, 0.0, -0.0707, 0.019, 0.0, -0.0686, 0.0168, 0.0, -0.0686, 0.019,
0.0, -0.0707, 0.0212, 0.0, -0.0707, 0.0233, 0.0, -0.0686, 0.0212, 0.0, -0.0686, 0.0233, 0.0, -0.0751, 0.0255, 0.0, -0.0751, 0.0277, 0.0, -0.0729, 0.0255, 0.0, -0.0729, 0.0277, 0.0, -0.0751, 0.0298, 0.0, -0.0751, 0.032, 0.0, -0.0729, 0.0298, 0.0, -0.0729, 0.032, 0.0, -0.0707, 0.0255, 0.0, -0.0707, 0.0277, 0.0, -0.0686, 0.0255, 0.0, -0.0686, 0.0277, 0.0, -0.0707, 0.0298, 0.0, -0.0707, 0.032, 0.0, -0.0686, 0.0298, 0.0, -0.0686, 0.032, 0.0, -0.0775, 0.0342, 0.0, -0.0775, 0.0363, 0.0, -0.0775, 0.0385, 0.0, -0.0775, 0.0407, -0.0017, -0.0775, 0.0342, -0.0017, -0.0775, 0.0363, -0.0039, -0.0775, 0.0342, -0.0039, -0.0775, 0.0363, -0.0017, -0.0775, 0.0385, -0.0017, -0.0775, 0.0407,
-0.0039, -0.0775, 0.0385, -0.0039, -0.0775, 0.0407, 0.0, -0.0775, 0.0429, 0.0, -0.0775, 0.045, 0.0, -0.0775, 0.0472, 0.0, -0.0775, 0.0487, -0.0017, -0.0775, 0.0429, -0.0017, -0.0775, 0.045, -0.0039, -0.0775, 0.0429, -0.0039, -0.0775, 0.045, -0.0017, -0.0775, 0.0472, -0.0017, -0.0775, 0.0487, -0.0039, -0.0775, 0.0472, -0.0039, -0.0775, 0.0487, 0.0, -0.0751, 0.0342, 0.0, -0.0751, 0.0363, 0.0, -0.0729, 0.0342, 0.0, -0.0729, 0.0363, 0.0, -0.0751, 0.0385, 0.0, -0.0751, 0.0407, 0.0, -0.0729, 0.0385, 0.0, -0.0729, 0.0407, 0.0, -0.0707, 0.0342, 0.0, -0.0707, 0.0363, 0.0, -0.0686, 0.0342, 0.0, -0.0686, 0.0363, 0.0, -0.0707, 0.0385, 0.0, -0.0707, 0.0407, 0.0, -0.0686, 0.0385, 0.0, -0.0686, 0.0407,
0.0, -0.0751, 0.0429, 0.0, -0.0751, 0.045, 0.0, -0.0729, 0.0429, 0.0, -0.0729, 0.045, 0.0, -0.0751, 0.0472, 0.0, -0.0751, 0.0487, 0.0, -0.0729, 0.0472, 0.0, -0.0729, 0.0487, -0.0017, -0.0751, 0.0487, -0.0039, -0.0751, 0.0487, -0.0017, -0.0729, 0.0487, -0.0039, -0.0729, 0.0487, 0.0, -0.0707, 0.0429, 0.0, -0.0707, 0.045, 0.0, -0.0686, 0.0429, 0.0, -0.0686, 0.045, 0.0, -0.0707, 0.0472, 0.0, -0.0707, 0.0487, 0.0, -0.0686, 0.0472, 0.0, -0.0686, 0.0487, -0.0017, -0.0707, 0.0487, -0.0039, -0.0707, 0.0487, -0.0017, -0.0686, 0.0487, -0.0039, -0.0686, 0.0487, 0.0, -0.0667, -0.0005, 0.0, -0.0667, 0.0016, 0.0, -0.0667, 0.0038, 0.0, -0.0667, 0.006, -0.0017, -0.0667, -0.0005, -0.0017, -0.0667, 0.0016,
-0.0039, -0.0667, -0.0005, -0.0039, -0.0667, 0.0016, -0.0017, -0.0667, 0.0038, -0.0017, -0.0667, 0.006, -0.0039, -0.0667, 0.0038, -0.0039, -0.0667, 0.006, 0.0, -0.0667, 0.0081, 0.0, -0.0667, 0.0103, 0.0, -0.0667, 0.0125, 0.0, -0.0667, 0.0146, -0.0017, -0.0667, 0.0081, -0.0017, -0.0667, 0.0103, -0.0039, -0.0667, 0.0081, -0.0039, -0.0667, 0.0103, -0.0017, -0.0667, 0.0125, -0.0017, -0.0667, 0.0146, -0.0039, -0.0667, 0.0125, -0.0039, -0.0667, 0.0146, 0.0, -0.0667, 0.0168, 0.0, -0.0667, 0.019, 0.0, -0.0667, 0.0212, 0.0, -0.0667, 0.0233, -0.0017, -0.0667, 0.0168, -0.0017, -0.0667, 0.019, -0.0039, -0.0667, 0.0168, -0.0039, -0.0667, 0.019, -0.0017, -0.0667, 0.0212, -0.0017, -0.0667, 0.0233, -0.0039, -0.0667, 0.0212, -0.0039, -0.0667, 0.0233,
0.0, -0.0667, 0.0255, 0.0, -0.0667, 0.0277, 0.0, -0.0667, 0.0298, 0.0, -0.0667, 0.032, -0.0017, -0.0667, 0.0255, -0.0017, -0.0667, 0.0277, -0.0039, -0.0667, 0.0255, -0.0039, -0.0667, 0.0277, -0.0017, -0.0667, 0.0298, -0.0017, -0.0667, 0.032, -0.0039, -0.0667, 0.0298, -0.0039, -0.0667, 0.032, 0.0, -0.05, -0.0005, 0.0, -0.05, 0.0016, 0.0, -0.0469, -0.0005, 0.0, -0.0469, 0.0016, 0.0, -0.05, 0.0038, 0.0, -0.05, 0.006, 0.0, -0.0469, 0.0038, 0.0, -0.0469, 0.006, -0.0017, -0.05, -0.0005, -0.0017, -0.05, 0.0016, -0.0039, -0.05, -0.0005, -0.0039, -0.05, 0.0016, -0.0017, -0.05, 0.0038, -0.0017, -0.05, 0.006, -0.0039, -0.05, 0.0038, -0.0039, -0.05, 0.006, 0.0, -0.0447, -0.0005, 0.0, -0.0447, 0.0016,
0.0, -0.0425, -0.0005, 0.0, -0.0425, 0.0016, 0.0, -0.0447, 0.0038, 0.0, -0.0447, 0.006, 0.0, -0.0425, 0.0038, 0.0, -0.0425, 0.006, 0.0, -0.05, 0.0081, 0.0, -0.05, 0.0103, 0.0, -0.0469, 0.0081, 0.0, -0.0469, 0.0103, 0.0, -0.05, 0.0125, 0.0, -0.05, 0.0146, 0.0, -0.0469, 0.0125, 0.0, -0.0469, 0.0146, -0.0017, -0.05, 0.0081, -0.0017, -0.05, 0.0103, -0.0039, -0.05, 0.0081, -0.0039, -0.05, 0.0103, -0.0017, -0.05, 0.0125, -0.0017, -0.05, 0.0146, -0.0039, -0.05, 0.0125, -0.0039, -0.05, 0.0146, 0.0, -0.0447, 0.0081, 0.0, -0.0447, 0.0103, 0.0, -0.0425, 0.0081, 0.0, -0.0425, 0.0103, 0.0, -0.0447, 0.0125, 0.0, -0.0447, 0.0146, 0.0, -0.0425, 0.0125, 0.0, -0.0425, 0.0146,
0.0, -0.0404, -0.0005, 0.0, -0.0404, 0.0016, 0.0, -0.0382, -0.0013, 0.0, -0.0382, 0.0016, 0.0, -0.0404, 0.0038, 0.0, -0.0404, 0.006, 0.0, -0.0382, 0.0038, 0.0, -0.0382, 0.006, -0.0017, -0.0382, -0.0013, -0.0039, -0.0382, -0.0013, 0.0, -0.036, -0.0013, 0.0, -0.036, 0.0016, 0.0, -0.0338, -0.0013, 0.0, -0.0338, 0.0016, 0.0, -0.036, 0.0038, 0.0, -0.036, 0.006, 0.0, -0.0338, 0.0038, 0.0, -0.0338, 0.006, -0.0017, -0.036, -0.0013, -0.0039, -0.036, -0.0013, -0.0017, -0.0338, -0.0013, -0.0039, -0.0338, -0.0013, 0.0, -0.0404, 0.0081, 0.0, -0.0404, 0.0103, 0.0, -0.0382, 0.0081, 0.0, -0.0382, 0.0105, 0.0, -0.0404, 0.0125, 0.0, -0.0404, 0.0146, 0.0, -0.0382, 0.0125, 0.0, -0.0382, 0.0146,
-0.0017, -0.0382, 0.0105, -0.0039, -0.0382, 0.0105, -0.0017, -0.0382, 0.0125, -0.0017, -0.0382, 0.0146, -0.0039, -0.0382, 0.0125, -0.0039, -0.0382, 0.0146, 0.0, -0.036, 0.0081, 0.0, -0.036, 0.0105, 0.0, -0.0338, 0.0081, 0.0, -0.0338, 0.0105, -0.0017, -0.036, 0.0105, -0.0039, -0.036, 0.0105, -0.0017, -0.0338, 0.0105, -0.0039, -0.0338, 0.0105, 0.0, -0.05, 0.0168, 0.0, -0.05, 0.019, 0.0, -0.0469, 0.0168, 0.0, -0.0469, 0.019, 0.0, -0.05, 0.0212, 0.0, -0.05, 0.0233, 0.0, -0.0469, 0.0212, 0.0, -0.0469, 0.0233, -0.0017, -0.05, 0.0168, -0.0017, -0.05, 0.019, -0.0039, -0.05, 0.0168, -0.0039, -0.05, 0.019, -0.0017, -0.05, 0.0212, -0.0017, -0.05, 0.0233, -0.0039, -0.05, 0.0212, -0.0039, -0.05, 0.0233,
0.0, -0.0447, 0.0168, 0.0, -0.0447, 0.019, 0.0, -0.0425, 0.0168, 0.0, -0.0425, 0.019, 0.0, -0.0447, 0.0212, 0.0, -0.0447, 0.0233, 0.0, -0.0425, 0.0212, 0.0, -0.0425, 0.0233, 0.0, -0.05, 0.0255, 0.0, -0.05, 0.0277, 0.0, -0.0469, 0.0255, 0.0, -0.0469, 0.0277, 0.0, -0.05, 0.0298, 0.0, -0.05, 0.032, 0.0, -0.0469, 0.0298, 0.0, -0.0469, 0.032, -0.0017, -0.05, 0.0255, -0.0017, -0.05, 0.0277, -0.0039, -0.05, 0.0255, -0.0039, -0.05, 0.0277, -0.0017, -0.05, 0.0298, -0.0017, -0.05, 0.032, -0.0039, -0.05, 0.0298, -0.0039, -0.05, 0.032, 0.0, -0.0447, 0.0255, 0.0, -0.0447, 0.0277, 0.0, -0.0425, 0.0255, 0.0, -0.0425, 0.0277, 0.0, -0.0447, 0.0298, 0.0, -0.0447, 0.032,
0.0, -0.0425, 0.0298, 0.0, -0.0425, 0.032, 0.0, -0.0404, 0.0168, 0.0, -0.0404, 0.019, 0.0, -0.0382, 0.0168, 0.0, -0.0382, 0.019, 0.0, -0.0404, 0.0212, 0.0, -0.0404, 0.0233, 0.0, -0.0382, 0.0212, 0.0, -0.0382, 0.0233, -0.0017, -0.0382, 0.0168, -0.0017, -0.0382, 0.019, -0.0039, -0.0382, 0.0168, -0.0039, -0.0382, 0.019, -0.0017, -0.0382, 0.0212, -0.0017, -0.0382, 0.0233, -0.0039, -0.0382, 0.0212, -0.0039, -0.0382, 0.0233, 0.0, -0.0404, 0.0255, 0.0, -0.0404, 0.0277, 0.0, -0.0382, 0.0255, 0.0, -0.0382, 0.0277, 0.0, -0.0404, 0.0298, 0.0, -0.0404, 0.032, 0.0, -0.0382, 0.0298, 0.0, -0.0382, 0.032, -0.0017, -0.0382, 0.0255, -0.0017, -0.0382, 0.0277, -0.0039, -0.0382, 0.0255, -0.0039, -0.0382, 0.0277,
-0.0017, -0.0382, 0.0298, -0.0017, -0.0382, 0.032, -0.0039, -0.0382, 0.0298, -0.0039, -0.0382, 0.032, 0.0, -0.0667, 0.0342, 0.0, -0.0667, 0.0363, 0.0, -0.0667, 0.0385, 0.0, -0.0667, 0.0407, -0.0017, -0.0667, 0.0342, -0.0017, -0.0667, 0.0363, -0.0039, -0.0667, 0.0342, -0.0039, -0.0667, 0.0363, -0.0017, -0.0667, 0.0385, -0.0017, -0.0667, 0.0407, -0.0039, -0.0667, 0.0385, -0.0039, -0.0667, 0.0407, 0.0, -0.0667, 0.0429, 0.0, -0.0667, 0.045, 0.0, -0.0667, 0.0472, 0.0, -0.0667, 0.0487, -0.0017, -0.0667, 0.0429, -0.0017, -0.0667, 0.045, -0.0039, -0.0667, 0.0429, -0.0039, -0.0667, 0.045, -0.0017, -0.0667, 0.0472, -0.0017, -0.0667, 0.0487, -0.0039, -0.0667, 0.0472, -0.0039, -0.0667, 0.0487, 0.0, -0.05, 0.0342, 0.0, -0.05, 0.0363,
0.0, -0.0469, 0.0342, 0.0, -0.0469, 0.0363, 0.0, -0.05, 0.0385, 0.0, -0.05, 0.0407, 0.0, -0.0469, 0.0385, 0.0, -0.0469, 0.0407, -0.0017, -0.05, 0.0342, -0.0017, -0.05, 0.0363, -0.0039, -0.05, 0.0342, -0.0039, -0.05, 0.0363, -0.0017, -0.05, 0.0385, -0.0017, -0.05, 0.0407, -0.0039, -0.05, 0.0385, -0.0039, -0.05, 0.0407, 0.0, -0.0447, 0.0342, 0.0, -0.0447, 0.0363, 0.0, -0.0425, 0.0342, 0.0, -0.0425, 0.0363, 0.0, -0.0447, 0.0385, 0.0, -0.0447, 0.0407, 0.0, -0.0425, 0.0385, 0.0, -0.0425, 0.0407, 0.0, -0.05, 0.0429, 0.0, -0.05, 0.045, 0.0, -0.0469, 0.0429, 0.0, -0.0469, 0.045, 0.0, -0.05, 0.0472, 0.0, -0.05, 0.0487, 0.0, -0.0469, 0.0472, 0.0, -0.0469, 0.0487,
-0.0017, -0.05, 0.0429, -0.0017, -0.05, 0.045, -0.0039, -0.05, 0.0429, -0.0039, -0.05, 0.045, -0.0017, -0.05, 0.0472, -0.0017, -0.05, 0.0487, -0.0039, -0.05, 0.0472, -0.0039, -0.05, 0.0487, -0.0017, -0.0469, 0.0487, -0.0039, -0.0469, 0.0487, 0.0, -0.0447, 0.0429, 0.0, -0.0447, 0.045, 0.0, -0.0425, 0.0429, 0.0, -0.0425, 0.045, 0.0, -0.0447, 0.0472, 0.0, -0.0447, 0.0487, 0.0, -0.0425, 0.0472, 0.0, -0.0425, 0.0487, -0.0017, -0.0447, 0.0487, -0.0039, -0.0447, 0.0487, -0.0017, -0.0425, 0.0487, -0.0039, -0.0425, 0.0487, 0.0, -0.0404, 0.0342, 0.0, -0.0404, 0.0363, 0.0, -0.0382, 0.0342, 0.0, -0.0382, 0.0363, 0.0, -0.0404, 0.0385, 0.0, -0.0404, 0.0407, 0.0, -0.0382, 0.0379, 0.0, -0.0382, 0.0407,
-0.0017, -0.0382, 0.0342, -0.0017, -0.0382, 0.0363, -0.0039, -0.0382, 0.0342, -0.0039, -0.0382, 0.0363, -0.0017, -0.0382, 0.0379, -0.0039, -0.0382, 0.0379, 0.0, -0.036, 0.0379, 0.0, -0.036, 0.0407, 0.0, -0.0338, 0.0379, 0.0, -0.0338, 0.0407, -0.0017, -0.036, 0.0379, -0.0039, -0.036, 0.0379, -0.0017, -0.0338, 0.0379, -0.0039, -0.0338, 0.0379, 0.0, -0.0404, 0.0429, 0.0, -0.0404, 0.045, 0.0, -0.0382, 0.0429, 0.0, -0.0382, 0.045, 0.0, -0.0404, 0.0472, 0.0, -0.0404, 0.0487, 0.0, -0.0382, 0.0472, 0.0, -0.0382, 0.0487, -0.0017, -0.0404, 0.0487, -0.0039, -0.0404, 0.0487, -0.0017, -0.0382, 0.0487, -0.0039, -0.0382, 0.0487, 0.0, -0.036, 0.0429, 0.0, -0.036, 0.045, 0.0, -0.0338, 0.0429, 0.0, -0.0338, 0.045,
0.0, -0.036, 0.0472, 0.0, -0.036, 0.0487, 0.0, -0.0338, 0.0472, 0.0, -0.0338, 0.0487, -0.0017, -0.036, 0.0487, -0.0039, -0.036, 0.0487, -0.0017, -0.0338, 0.0487, -0.0039, -0.0338, 0.0487, 0.0, -0.0317, -0.0013, 0.0, -0.0317, 0.0016, 0.0, -0.0295, -0.0013, 0.0, -0.0295, 0.0016, 0.0, -0.0317, 0.0038, 0.0, -0.0317, 0.006, 0.0, -0.0295, 0.0038, 0.0, -0.0295, 0.006, -0.0017, -0.0317, -0.0013, -0.0039, -0.0317, -0.0013, -0.0017, -0.0295, -0.0013, -0.0039, -0.0295, -0.0013, 0.0, -0.0273, -0.0013, 0.0, -0.0273, 0.0016, 0.0, -0.0252, -0.0013, 0.0, -0.0252, 0.0016, 0.0, -0.0273, 0.0038, 0.0, -0.0273, 0.006, 0.0, -0.0252, 0.0038, 0.0, -0.0252, 0.006, -0.0017, -0.0273, -0.0013, -0.0039, -0.0273, -0.0013,
-0.0017, -0.0252, -0.0013, -0.0039, -0.0252, -0.0013, 0.0, -0.0317, 0.0081, 0.0, -0.0317, 0.0105, 0.0, -0.0295, 0.0081, 0.0, -0.0295, 0.0105, -0.0017, -0.0317, 0.0105, -0.0039, -0.0317, 0.0105, -0.0017, -0.0295, 0.0105, -0.0039, -0.0295, 0.0105, 0.0, -0.0273, 0.0081, 0.0, -0.0273, 0.0105, 0.0, -0.0252, 0.0081, 0.0, -0.0252, 0.0105, -0.0017, -0.0273, 0.0105, -0.0039, -0.0273, 0.0105, -0.0017, -0.0252, 0.0105, -0.0039, -0.0252, 0.0105, 0.0, -0.023, -0.0013, 0.0, -0.023, 0.0016, 0.0, -0.0208, -0.0013, 0.0, -0.0208, 0.0016, 0.0, -0.023, 0.0038, 0.0, -0.023, 0.006, 0.0, -0.0208, 0.0038, 0.0, -0.0208, 0.006, -0.0017, -0.023, -0.0013, -0.0039, -0.023, -0.0013, -0.0017, -0.0208, -0.0013, -0.0039, -0.0208, -0.0013,
0.0, -0.0187, -0.0013, 0.0, -0.0187, 0.0016, 0.0, -0.0165, -0.0013, 0.0, -0.0165, 0.0016, 0.0, -0.0187, 0.0038, 0.0, -0.0187, 0.006, 0.0, -0.0165, 0.0038, 0.0, -0.0165, 0.006, -0.0017, -0.0187, -0.0013, -0.0039, -0.0187, -0.0013, -0.0017, -0.0165, -0.0013, -0.0039, -0.0165, -0.0013, 0.0, -0.023, 0.0081, 0.0, -0.023, 0.0105, 0.0, -0.0208, 0.0081, 0.0, -0.0208, 0.0105, -0.0017, -0.023, 0.0105, -0.0039, -0.023, 0.0105, -0.0017, -0.0208, 0.0105, -0.0039, -0.0208, 0.0105, 0.0, -0.0187, 0.0081, 0.0, -0.0187, 0.0105, 0.0, -0.0165, 0.0081, 0.0, -0.0165, 0.0105, -0.0017, -0.0187, 0.0105, -0.0039, -0.0187, 0.0105, -0.0017, -0.0165, 0.0105, -0.0039, -0.0165, 0.0105, 0.0, -0.0143, -0.0013, 0.0, -0.0143, 0.0016,
0.0, -0.0121, -0.0013, 0.0, -0.0121, 0.0016, 0.0, -0.0143, 0.0038, 0.0, -0.0143, 0.006, 0.0, -0.0121, 0.0038, 0.0, -0.0121, 0.006, -0.0017, -0.0143, -0.0013, -0.0039, -0.0143, -0.0013, -0.0017, -0.0121, -0.0013, -0.0039, -0.0121, -0.0013, 0.0, -0.01, -0.0013, 0.0, -0.01, 0.0016, 0.0, -0.0078, -0.0013, 0.0, -0.0078, 0.0016, 0.0, -0.01, 0.0038, 0.0, -0.01, 0.006, 0.0, -0.0078, 0.0038, 0.0, -0.0078, 0.006, -0.0017, -0.01, -0.0013, -0.0039, -0.01, -0.0013, -0.0017, -0.0078, -0.0013, -0.0039, -0.0078, -0.0013, 0.0, -0.0143, 0.0081, 0.0, -0.0143, 0.0105, 0.0, -0.0121, 0.0081, 0.0, -0.0121, 0.0105, -0.0017, -0.0143, 0.0105, -0.0039, -0.0143, 0.0105, -0.0017, -0.0121, 0.0105, -0.0039, -0.0121, 0.0105,
0.0, -0.01, 0.0081, 0.0, -0.01, 0.0105, 0.0, -0.0078, 0.0081, 0.0, -0.0078, 0.0105, -0.0017, -0.01, 0.0105, -0.0039, -0.01, 0.0105, -0.0017, -0.0078, 0.0105, -0.0039, -0.0078, 0.0105, 0.0, -0.0056, -0.0013, 0.0, -0.0056, 0.0016, 0.0, -0.0035, -0.0013, 0.0, -0.0035, 0.0016, 0.0, -0.0056, 0.0038, 0.0, -0.0056, 0.006, 0.0, -0.0035, 0.0038, 0.0, -0.0035, 0.006, -0.0017, -0.0056, -0.0013, -0.0039, -0.0056, -0.0013, -0.0017, -0.0035, -0.0013, -0.0039, -0.0035, -0.0013, 0.0, -0.0013, -0.0013, 0.0, -0.0013, 0.0016, 0.0, 0.0009, -0.0013, 0.0, 0.0009, 0.0016, 0.0, -0.0013, 0.0038, 0.0, -0.0013, 0.006, 0.0, 0.0009, 0.0038, 0.0, 0.0009, 0.006, -0.0017, -0.0013, -0.0013, -0.0039, -0.0013, -0.0013,
-0.0017, 0.0009, -0.0013, -0.0039, 0.0009, -0.0013, 0.0, -0.0056, 0.0081, 0.0, -0.0056, 0.0105, 0.0, -0.0035, 0.0081, 0.0, -0.0035, 0.0105, -0.0017, -0.0056, 0.0105, -0.0039, -0.0056, 0.0105, -0.0017, -0.0035, 0.0105, -0.0039, -0.0035, 0.0105, 0.0, -0.0013, 0.0081, 0.0, -0.0013, 0.0105, 0.0, 0.0009, 0.0081, 0.0, 0.0009, 0.0105, -0.0017, -0.0013, 0.0105, -0.0039, -0.0013, 0.0105, -0.0017, 0.0009, 0.0105, -0.0039, 0.0009, 0.0105, 0.0, -0.0317, 0.0379, 0.0, -0.0317, 0.0407, 0.0, -0.0295, 0.0379, 0.0, -0.0295, 0.0407, -0.0017, -0.0317, 0.0379, -0.0039, -0.0317, 0.0379, -0.0017, -0.0295, 0.0379, -0.0039, -0.0295, 0.0379, 0.0, -0.0273, 0.0379, 0.0, -0.0273, 0.0407, 0.0, -0.0252, 0.0379, 0.0, -0.0252, 0.0407,
-0.0017, -0.0273, 0.0379, -0.0039, -0.0273, 0.0379, -0.0017, -0.0252, 0.0379, -0.0039, -0.0252, 0.0379, 0.0, -0.0317, 0.0429, 0.0, -0.0317, 0.045, 0.0, -0.0295, 0.0429, 0.0, -0.0295, 0.045, 0.0, -0.0317, 0.0472, 0.0, -0.0317, 0.0487, 0.0, -0.0295, 0.0472, 0.0, -0.0295, 0.0487, -0.0017, -0.0317, 0.0487, -0.0039, -0.0317, 0.0487, -0.0017, -0.0295, 0.0487, -0.0039, -0.0295, 0.0487, 0.0, -0.0273, 0.0429, 0.0, -0.0273, 0.045, 0.0, -0.0252, 0.0429, 0.0, -0.0252, 0.045, 0.0, -0.0273, 0.0472, 0.0, -0.0273, 0.0487, 0.0, -0.0252, 0.0472, 0.0, -0.0252, 0.0487, -0.0017, -0.0273, 0.0487, -0.0039, -0.0273, 0.0487, -0.0017, -0.0252, 0.0487, -0.0039, -0.0252, 0.0487, 0.0, -0.023, 0.0379, 0.0, -0.023, 0.0407,
0.0, -0.0208, 0.0379, 0.0, -0.0208, 0.0407, -0.0017, -0.023, 0.0379, -0.0039, -0.023, 0.0379, -0.0017, -0.0208, 0.0379, -0.0039, -0.0208, 0.0379, 0.0, -0.0187, 0.0379, 0.0, -0.0187, 0.0407, 0.0, -0.0165, 0.0379, 0.0, -0.0165, 0.0407, -0.0017, -0.0187, 0.0379, -0.0039, -0.0187, 0.0379, -0.0017, -0.0165, 0.0379, -0.0039, -0.0165, 0.0379, 0.0, -0.023, 0.0429, 0.0, -0.023, 0.045, 0.0, -0.0208, 0.0429, 0.0, -0.0208, 0.045, 0.0, -0.023, 0.0472, 0.0, -0.023, 0.0487, 0.0, -0.0208, 0.0472, 0.0, -0.0208, 0.0487, -0.0017, -0.023, 0.0487, -0.0039, -0.023, 0.0487, -0.0017, -0.0208, 0.0487, -0.0039, -0.0208, 0.0487, 0.0, -0.0187, 0.0429, 0.0, -0.0187, 0.045, 0.0, -0.0165, 0.0429, 0.0, -0.0165, 0.045,
0.0, -0.0187, 0.0472, 0.0, -0.0187, 0.0487, 0.0, -0.0165, 0.0472, 0.0, -0.0165, 0.0487, -0.0017, -0.0187, 0.0487, -0.0039, -0.0187, 0.0487, -0.0017, -0.0165, 0.0487, -0.0039, -0.0165, 0.0487, 0.0, -0.0143, 0.0379, 0.0, -0.0143, 0.0407, 0.0, -0.0121, 0.0379, 0.0, -0.0121, 0.0407, -0.0017, -0.0143, 0.0379, -0.0039, -0.0143, 0.0379, -0.0017, -0.0121, 0.0379, -0.0039, -0.0121, 0.0379, 0.0, -0.01, 0.0379, 0.0, -0.01, 0.0407, 0.0, -0.0078, 0.0379, 0.0, -0.0078, 0.0407, -0.0017, -0.01, 0.0379, -0.0039, -0.01, 0.0379, -0.0017, -0.0078, 0.0379, -0.0039, -0.0078, 0.0379, 0.0, -0.0143, 0.0429, 0.0, -0.0143, 0.045, 0.0, -0.0121, 0.0429, 0.0, -0.0121, 0.045, 0.0, -0.0143, 0.0472, 0.0, -0.0143, 0.0487,
0.0, -0.0121, 0.0472, 0.0, -0.0121, 0.0487, -0.0017, -0.0143, 0.0487, -0.0039, -0.0143, 0.0487, -0.0017, -0.0121, 0.0487, -0.0039, -0.0121, 0.0487, 0.0, -0.01, 0.0429, 0.0, -0.01, 0.045, 0.0, -0.0078, 0.0429, 0.0, -0.0078, 0.045, 0.0, -0.01, 0.0472, 0.0, -0.01, 0.0487, 0.0, -0.0078, 0.0472, 0.0, -0.0078, 0.0487, -0.0017, -0.01, 0.0487, -0.0039, -0.01, 0.0487, -0.0017, -0.0078, 0.0487, -0.0039, -0.0078, 0.0487, 0.0, -0.0056, 0.0379, 0.0, -0.0056, 0.0407, 0.0, -0.0035, 0.0379, 0.0, -0.0035, 0.0407, -0.0017, -0.0056, 0.0379, -0.0039, -0.0056, 0.0379, -0.0017, -0.0035, 0.0379, -0.0039, -0.0035, 0.0379, 0.0, -0.0013, 0.0379, 0.0, -0.0013, 0.0407, 0.0, 0.0008, 0.0379, 0.0, 0.0009, 0.0407,
-0.0017, -0.0013, 0.0379, -0.0039, -0.0013, 0.0379, -0.0017, 0.0009, 0.0379, -0.0039, 0.0009, 0.0379, 0.0, -0.0056, 0.0429, 0.0, -0.0056, 0.045, 0.0, -0.0035, 0.0429, 0.0, -0.0035, 0.045, 0.0, -0.0056, 0.0472, 0.0, -0.0056, 0.0487, 0.0, -0.0035, 0.0472, 0.0, -0.0035, 0.0487, -0.0017, -0.0056, 0.0487, -0.0039, -0.0056, 0.0487, -0.0017, -0.0035, 0.0487, -0.0039, -0.0035, 0.0487, 0.0, -0.0013, 0.0429, 0.0, -0.0013, 0.045, 0.0, 0.0009, 0.0429, 0.0, 0.0009, 0.045, 0.0, -0.0013, 0.0472, 0.0, -0.0013, 0.0487, 0.0, 0.0009, 0.0472, 0.0, 0.0009, 0.0487, -0.0017, -0.0013, 0.0487, -0.0039, -0.0013, 0.0487, -0.0017, 0.0009, 0.0487, -0.0039, 0.0009, 0.0487, -0.0061, -0.4928, -0.0364, -0.0083, -0.4928, -0.0364,
-0.01, -0.4928, -0.0364, -0.0061, -0.4868, -0.0457, -0.0083, -0.4868, -0.0457, -0.0061, -0.4849, -0.0482, -0.0061, -0.486, -0.0468, -0.0083, -0.4849, -0.0482, -0.0083, -0.486, -0.0468, -0.01, -0.4868, -0.0456, -0.01, -0.485, -0.0481, -0.01, -0.486, -0.0467, -0.0061, -0.4886, -0.043, -0.0061, -0.4895, -0.0418, -0.0083, -0.4886, -0.043, -0.0083, -0.4895, -0.0418, -0.0061, -0.4911, -0.0393, -0.0061, -0.4922, -0.0375, -0.0083, -0.4911, -0.0393, -0.0083, -0.4922, -0.0375, -0.0061, -0.4904, -0.0403, -0.0083, -0.4904, -0.0403, -0.01, -0.4886, -0.043, -0.01, -0.4893, -0.042, -0.01, -0.4912, -0.0391, -0.01, -0.4922, -0.0375, -0.01, -0.4905, -0.0402, -0.01, -0.4896, -0.0374, -0.0061, -0.4879, -0.0441, -0.0083, -0.4879, -0.0441, -0.0061, -0.4868, -0.0368, -0.0083, -0.4868, -0.0368,
-0.0061, -0.4844, -0.0387, -0.0061, -0.4855, -0.0379, -0.0083, -0.4844, -0.0387, -0.0083, -0.4855, -0.0379, -0.01, -0.4879, -0.0441, -0.01, -0.4874, -0.0418, -0.01, -0.4852, -0.0439, -0.01, -0.4852, -0.0418, -0.01, -0.4874, -0.0396, -0.01, -0.4869, -0.0368, -0.01, -0.4845, -0.0386, -0.01, -0.4856, -0.0378, -0.0061, -0.4951, -0.0321, -0.0083, -0.4951, -0.0321, -0.0061, -0.4935, -0.0353, -0.0061, -0.4945, -0.0332, -0.0083, -0.4935, -0.0353, -0.0083, -0.4945, -0.0332, -0.0061, -0.4961, -0.0297, -0.0083, -0.4961, -0.0297, -0.0061, -0.4938, -0.0306, -0.0083, -0.4938, -0.0306, -0.01, -0.4951, -0.0321, -0.01, -0.4934, -0.0354, -0.01, -0.4945, -0.0332, -0.01, -0.4961, -0.0297, -0.01, -0.4935, -0.0308, -0.0061, -0.4934, -0.0241, -0.0083, -0.4934, -0.0241, -0.0061, -0.4952, -0.0215,
-0.0061, -0.4961, -0.0201, -0.0083, -0.4952, -0.0215, -0.0083, -0.4961, -0.0201, -0.0061, -0.4945, -0.0226, -0.0061, -0.4935, -0.0195, -0.0083, -0.4945, -0.0226, -0.0083, -0.4935, -0.0195, -0.01, -0.4934, -0.024, -0.01, -0.4953, -0.0215, -0.01, -0.496, -0.0203, -0.01, -0.4945, -0.0226, -0.01, -0.4935, -0.0195, -0.0061, -0.4914, -0.0326, -0.0083, -0.4914, -0.0326, -0.0061, -0.4891, -0.0347, -0.0061, -0.4902, -0.0337, -0.0083, -0.4891, -0.0347, -0.0083, -0.4902, -0.0337, -0.0061, -0.4925, -0.0317, -0.0083, -0.4925, -0.0317, -0.0061, -0.4894, -0.0286, -0.0083, -0.4894, -0.0286, -0.01, -0.4917, -0.0353, -0.01, -0.4914, -0.0326, -0.01, -0.4891, -0.0347, -0.01, -0.4902, -0.0337, -0.01, -0.4925, -0.0317, -0.01, -0.4894, -0.0286, -0.0061, -0.4879, -0.0358, -0.0083, -0.4879, -0.0358,
-0.0061, -0.4849, -0.0326, -0.0083, -0.4849, -0.0326, -0.0061, -0.4872, -0.0307, -0.0061, -0.4883, -0.0296, -0.0083, -0.4872, -0.0307, -0.0083, -0.4883, -0.0296, -0.0061, -0.486, -0.0317, -0.0083, -0.486, -0.0317, -0.01, -0.4879, -0.0358, -0.01, -0.4849, -0.0327, -0.01, -0.4872, -0.0307, -0.01, -0.4883, -0.0297, -0.01, -0.486, -0.0318, -0.01, -0.4852, -0.0288, -0.0061, -0.4915, -0.0264, -0.0061, -0.4925, -0.0252, -0.0083, -0.4915, -0.0264, -0.0083, -0.4925, -0.0252, -0.0061, -0.4905, -0.0274, -0.0083, -0.4905, -0.0274, -0.0061, -0.4911, -0.0215, -0.0061, -0.4922, -0.0206, -0.0083, -0.4911, -0.0215, -0.0083, -0.4922, -0.0206, -0.0061, -0.4896, -0.0225, -0.0083, -0.4896, -0.0225, -0.01, -0.4915, -0.0263, -0.01, -0.4926, -0.0251, -0.01, -0.4905, -0.0274, -0.01, -0.4896, -0.0244,
-0.01, -0.4911, -0.0214, -0.01, -0.4922, -0.0206, -0.01, -0.4894, -0.0227, -0.0061, -0.4873, -0.0239, -0.0083, -0.4873, -0.0239, -0.0061, -0.4853, -0.025, -0.0083, -0.4853, -0.025, -0.0061, -0.4884, -0.0233, -0.0083, -0.4884, -0.0233, -0.01, -0.4874, -0.0266, -0.01, -0.4873, -0.0239, -0.01, -0.4852, -0.0266, -0.01, -0.4851, -0.025, -0.01, -0.4884, -0.0233, -0.0061, -0.4975, -0.0175, -0.0061, -0.498, -0.016, -0.0083, -0.4975, -0.0175, -0.0083, -0.498, -0.016, -0.01, -0.4975, -0.0173, -0.01, -0.498, -0.016, -0.0061, -0.4983, -0.0157, -0.0061, -0.497, -0.0167, -0.0083, -0.4983, -0.0157, -0.0083, -0.497, -0.0167, -0.0061, -0.4945, -0.0186, -0.0083, -0.4945, -0.0186, -0.01, -0.4983, -0.0157, -0.01, -0.497, -0.0167, -0.01, -0.4946, -0.0186, -0.0061, -0.4561, -0.0713,
-0.0083, -0.4561, -0.0713, -0.0061, -0.4541, -0.0736, -0.0061, -0.4552, -0.0724, -0.0083, -0.4541, -0.0736, -0.0083, -0.4552, -0.0724, -0.01, -0.4561, -0.0713, -0.01, -0.4541, -0.0736, -0.01, -0.4552, -0.0724, -0.0061, -0.452, -0.0759, -0.0083, -0.452, -0.0759, -0.0061, -0.4499, -0.0781, -0.0061, -0.451, -0.077, -0.0083, -0.4499, -0.0781, -0.0083, -0.451, -0.077, -0.0061, -0.4531, -0.0747, -0.0083, -0.4531, -0.0747, -0.01, -0.452, -0.0759, -0.01, -0.4499, -0.0781, -0.01, -0.451, -0.077, -0.01, -0.4531, -0.0747, -0.01, -0.4527, -0.0722, -0.01, -0.4505, -0.0743, -0.01, -0.4505, -0.0722, -0.0061, -0.4372, -0.089, -0.0083, -0.4372, -0.089, -0.01, -0.4371, -0.0891, -0.0061, -0.4345, -0.0908, -0.0061, -0.4356, -0.0901, -0.0083, -0.4345, -0.0908, -0.0083, -0.4356, -0.0901,
-0.0061, -0.4332, -0.0917, -0.0083, -0.4332, -0.0917, -0.01, -0.4345, -0.0908, -0.01, -0.4356, -0.0901, -0.01, -0.4334, -0.0916, -0.01, -0.4332, -0.0895, -0.0061, -0.4478, -0.0802, -0.0083, -0.4478, -0.0802, -0.0061, -0.4455, -0.0823, -0.0061, -0.4466, -0.0813, -0.0083, -0.4455, -0.0823, -0.0083, -0.4466, -0.0813, -0.01, -0.4478, -0.0802, -0.01, -0.4455, -0.0823, -0.01, -0.4466, -0.0813, -0.0061, -0.4432, -0.0844, -0.0083, -0.4432, -0.0844, -0.0061, -0.4408, -0.0863, -0.0061, -0.4419, -0.0855, -0.0083, -0.4408, -0.0863, -0.0083, -0.4419, -0.0855, -0.0061, -0.4443, -0.0834, -0.0083, -0.4443, -0.0834, -0.01, -0.4432, -0.0844, -0.01, -0.4408, -0.0863, -0.01, -0.442, -0.0855, -0.01, -0.4443, -0.0834, -0.01, -0.444, -0.0808, -0.01, -0.4418, -0.083, -0.01, -0.4418, -0.0808,
-0.0061, -0.4488, -0.0792, -0.0083, -0.4488, -0.0792, -0.01, -0.4488, -0.0792, -0.01, -0.4483, -0.0765, -0.01, -0.4462, -0.0787, -0.01, -0.4462, -0.0765, -0.01, -0.4483, -0.0743, -0.01, -0.4483, -0.0722, -0.01, -0.4462, -0.0743, -0.01, -0.4462, -0.0722, -0.01, -0.444, -0.0787, -0.01, -0.444, -0.0765, -0.01, -0.4418, -0.0787, -0.01, -0.4418, -0.0765, -0.01, -0.444, -0.0743, -0.01, -0.444, -0.0722, -0.01, -0.4418, -0.0743, -0.01, -0.4418, -0.0722, -0.0061, -0.4397, -0.0872, -0.0083, -0.4397, -0.0872, -0.0061, -0.4383, -0.0882, -0.0083, -0.4383, -0.0882, -0.01, -0.4399, -0.087, -0.01, -0.4397, -0.0852, -0.01, -0.4382, -0.0883, -0.01, -0.4375, -0.0852, -0.01, -0.4397, -0.083, -0.01, -0.4397, -0.0808, -0.01, -0.4375, -0.083, -0.01, -0.4375, -0.0808,
-0.01, -0.4353, -0.0874, -0.01, -0.4353, -0.0852, -0.01, -0.4332, -0.0874, -0.01, -0.4332, -0.0852, -0.01, -0.4353, -0.083, -0.01, -0.4353, -0.0808, -0.01, -0.4332, -0.083, -0.01, -0.4332, -0.0808, -0.01, -0.4397, -0.0787, -0.01, -0.4397, -0.0765, -0.01, -0.4375, -0.0787, -0.01, -0.4375, -0.0765, -0.01, -0.4397, -0.0743, -0.01, -0.4397, -0.0722, -0.01, -0.4375, -0.0743, -0.01, -0.4375, -0.0722, -0.01, -0.4353, -0.0787, -0.01, -0.4353, -0.0765, -0.01, -0.4332, -0.0787, -0.01, -0.4332, -0.0765, -0.01, -0.4353, -0.0743, -0.01, -0.4353, -0.0722, -0.01, -0.4332, -0.0743, -0.01, -0.4332, -0.0722, -0.0061, -0.4235, -0.0972, -0.0083, -0.4235, -0.0972, -0.01, -0.4234, -0.0972, -0.0061, -0.4307, -0.0933, -0.0083, -0.4307, -0.0933, -0.0061, -0.4277, -0.0949,
-0.0061, -0.4288, -0.0943, -0.0083, -0.4277, -0.0949, -0.0083, -0.4288, -0.0943, -0.0061, -0.4318, -0.0926, -0.0083, -0.4318, -0.0926, -0.01, -0.4306, -0.0933, -0.01, -0.4277, -0.0949, -0.01, -0.4289, -0.0943, -0.01, -0.4316, -0.0927, -0.01, -0.431, -0.0895, -0.01, -0.4288, -0.0917, -0.01, -0.4288, -0.0895, -0.0061, -0.4266, -0.0955, -0.0083, -0.4266, -0.0955, -0.0061, -0.4245, -0.0966, -0.0083, -0.4245, -0.0966, -0.01, -0.4268, -0.0955, -0.01, -0.4266, -0.0939, -0.01, -0.4246, -0.0966, -0.01, -0.4245, -0.0939, -0.01, -0.4266, -0.0917, -0.01, -0.4266, -0.0895, -0.01, -0.4245, -0.0917, -0.01, -0.4245, -0.0895, -0.0061, -0.4223, -0.0977, -0.0083, -0.4223, -0.0977, -0.0061, -0.4201, -0.0988, -0.0083, -0.4201, -0.0988, -0.01, -0.4224, -0.0977, -0.01, -0.4205, -0.0986,
-0.0061, -0.4179, -0.0997, -0.0061, -0.419, -0.0993, -0.0083, -0.4179, -0.0997, -0.0083, -0.419, -0.0993, -0.0061, -0.4158, -0.1006, -0.0083, -0.4158, -0.1006, -0.01, -0.4177, -0.0998, -0.01, -0.4186, -0.0995, -0.01, -0.416, -0.1005, -0.01, -0.4158, -0.0982, -0.01, -0.4223, -0.096, -0.01, -0.4223, -0.0939, -0.01, -0.4201, -0.096, -0.01, -0.4201, -0.0939, -0.01, -0.4223, -0.0917, -0.01, -0.4223, -0.0895, -0.01, -0.4201, -0.0917, -0.01, -0.4201, -0.0895, -0.01, -0.418, -0.096, -0.01, -0.418, -0.0939, -0.01, -0.4158, -0.096, -0.01, -0.4158, -0.0939, -0.01, -0.418, -0.0917, -0.01, -0.418, -0.0895, -0.01, -0.4158, -0.0917, -0.01, -0.4158, -0.0895, -0.01, -0.431, -0.0874, -0.01, -0.431, -0.0852, -0.01, -0.4288, -0.0874, -0.01, -0.4288, -0.0852,
-0.01, -0.431, -0.083, -0.01, -0.431, -0.0808, -0.01, -0.4288, -0.083, -0.01, -0.4288, -0.0808, -0.01, -0.4266, -0.0874, -0.01, -0.4266, -0.0852, -0.01, -0.4245, -0.0874, -0.01, -0.4245, -0.0852, -0.01, -0.4266, -0.083, -0.01, -0.4266, -0.0808, -0.01, -0.4245, -0.083, -0.01, -0.4245, -0.0808, -0.0061, -0.4307, -0.072, -0.0083, -0.4307, -0.072, -0.0061, -0.4287, -0.0741, -0.0061, -0.4297, -0.0731, -0.0083, -0.4287, -0.0741, -0.0083, -0.4297, -0.0731, -0.01, -0.431, -0.0787, -0.01, -0.431, -0.0765, -0.01, -0.4288, -0.0787, -0.01, -0.4288, -0.0765, -0.01, -0.431, -0.0743, -0.01, -0.4307, -0.072, -0.01, -0.4287, -0.0741, -0.01, -0.4297, -0.0731, -0.0061, -0.4263, -0.076, -0.0083, -0.4263, -0.076, -0.0061, -0.4238, -0.0779, -0.0061, -0.4249, -0.0771,
-0.0083, -0.4238, -0.0779, -0.0083, -0.4249, -0.0771, -0.0061, -0.4274, -0.0752, -0.0083, -0.4274, -0.0752, -0.01, -0.4266, -0.0787, -0.01, -0.4263, -0.076, -0.01, -0.4239, -0.0778, -0.01, -0.4249, -0.077, -0.01, -0.4274, -0.0751, -0.0061, -0.42, -0.0804, -0.0083, -0.42, -0.0804, -0.01, -0.4223, -0.0874, -0.01, -0.4223, -0.0852, -0.01, -0.4201, -0.0874, -0.01, -0.4201, -0.0852, -0.01, -0.4223, -0.083, -0.01, -0.4223, -0.0808, -0.01, -0.4201, -0.083, -0.01, -0.42, -0.0804, -0.0061, -0.4172, -0.0821, -0.0061, -0.4183, -0.0815, -0.0083, -0.4172, -0.0821, -0.0083, -0.4183, -0.0815, -0.0061, -0.4158, -0.0828, -0.0083, -0.4158, -0.0828, -0.01, -0.418, -0.0874, -0.01, -0.418, -0.0852, -0.01, -0.4158, -0.0874, -0.01, -0.4158, -0.0852, -0.01, -0.4174, -0.082,
-0.01, -0.4184, -0.0814, -0.01, -0.4155, -0.083, -0.0061, -0.4223, -0.0789, -0.0083, -0.4223, -0.0789, -0.0061, -0.4211, -0.0797, -0.0083, -0.4211, -0.0797, -0.01, -0.4221, -0.079, -0.01, -0.4211, -0.0797, -0.0061, -0.4799, -0.0538, -0.0083, -0.4799, -0.0538, -0.01, -0.4799, -0.0538, -0.0061, -0.4777, -0.0559, -0.0061, -0.4788, -0.0549, -0.0083, -0.4777, -0.0559, -0.0083, -0.4788, -0.0549, -0.0061, -0.4766, -0.0569, -0.0083, -0.4766, -0.0569, -0.01, -0.4777, -0.0559, -0.01, -0.4788, -0.0549, -0.01, -0.4768, -0.0567, -0.01, -0.4766, -0.0548, -0.0061, -0.469, -0.0624, -0.0083, -0.469, -0.0624, -0.0061, -0.4679, -0.0631, -0.0083, -0.4679, -0.0631, -0.01, -0.469, -0.0624, -0.01, -0.468, -0.063, -0.0061, -0.4743, -0.0588, -0.0083, -0.4743, -0.0588, -0.0061, -0.4717, -0.0606,
-0.0061, -0.4728, -0.0599, -0.0083, -0.4717, -0.0606, -0.0083, -0.4728, -0.0599, -0.0061, -0.4753, -0.058, -0.0083, -0.4753, -0.058, -0.01, -0.4742, -0.0589, -0.01, -0.4717, -0.0607, -0.01, -0.4728, -0.0599, -0.01, -0.4753, -0.058, -0.01, -0.4744, -0.0548, -0.01, -0.4722, -0.057, -0.01, -0.4722, -0.0548, -0.0061, -0.4701, -0.0617, -0.0083, -0.4701, -0.0617, -0.01, -0.4701, -0.0617, -0.01, -0.47, -0.0591, -0.01, -0.4679, -0.0613, -0.01, -0.4679, -0.0591, -0.01, -0.47, -0.057, -0.01, -0.47, -0.0548, -0.01, -0.4679, -0.057, -0.01, -0.4679, -0.0548, -0.0061, -0.482, -0.0516, -0.0061, -0.483, -0.0505, -0.0083, -0.482, -0.0516, -0.0083, -0.483, -0.0505, -0.0061, -0.481, -0.0527, -0.0083, -0.481, -0.0527, -0.0061, -0.4841, -0.0493, -0.0083, -0.4841, -0.0493,
-0.01, -0.482, -0.0516, -0.01, -0.4828, -0.0507, -0.01, -0.481, -0.0527, -0.01, -0.4809, -0.0505, -0.01, -0.4841, -0.0492, -0.01, -0.4831, -0.0461, -0.01, -0.4809, -0.0483, -0.01, -0.4809, -0.0461, -0.01, -0.4787, -0.0526, -0.01, -0.4787, -0.0505, -0.01, -0.4766, -0.0526, -0.01, -0.4766, -0.0505, -0.01, -0.4787, -0.0483, -0.01, -0.4787, -0.0461, -0.01, -0.4766, -0.0483, -0.01, -0.4766, -0.0461, -0.0061, -0.4807, -0.0413, -0.0083, -0.4807, -0.0413, -0.0061, -0.4831, -0.0397, -0.0083, -0.4831, -0.0397, -0.0061, -0.4818, -0.0406, -0.0083, -0.4818, -0.0406, -0.01, -0.4831, -0.0439, -0.01, -0.4831, -0.0418, -0.01, -0.4809, -0.0439, -0.01, -0.4807, -0.0413, -0.01, -0.4828, -0.0398, -0.01, -0.4818, -0.0405, -0.0061, -0.4779, -0.0431, -0.0061, -0.479, -0.0424,
-0.0083, -0.4779, -0.0431, -0.0083, -0.479, -0.0424, -0.0061, -0.4766, -0.0439, -0.0083, -0.4766, -0.0439, -0.0061, -0.4785, -0.0368, -0.0083, -0.4785, -0.0368, -0.0061, -0.4766, -0.0377, -0.0083, -0.4766, -0.0377, -0.01, -0.4781, -0.043, -0.01, -0.4791, -0.0423, -0.01, -0.4763, -0.044, -0.01, -0.4783, -0.0369, -0.01, -0.4768, -0.0376, -0.0061, -0.4738, -0.0453, -0.0083, -0.4738, -0.0453, -0.0061, -0.4722, -0.0459, -0.0083, -0.4722, -0.0459, -0.01, -0.4744, -0.0526, -0.01, -0.4744, -0.0505, -0.01, -0.4722, -0.0526, -0.01, -0.4722, -0.0505, -0.01, -0.4744, -0.0483, -0.01, -0.474, -0.0452, -0.01, -0.4722, -0.0483, -0.01, -0.472, -0.046, -0.0061, -0.4701, -0.0468, -0.0083, -0.4701, -0.0468, -0.0061, -0.4678, -0.0476, -0.0061, -0.4689, -0.0472, -0.0083, -0.4678, -0.0476,
-0.0083, -0.4689, -0.0472, -0.01, -0.47, -0.0526, -0.01, -0.47, -0.0505, -0.01, -0.4679, -0.0526, -0.01, -0.4679, -0.0505, -0.01, -0.47, -0.0483, -0.01, -0.47, -0.0468, -0.01, -0.4679, -0.0476, -0.01, -0.4689, -0.0472, -0.0061, -0.4749, -0.0448, -0.0083, -0.4749, -0.0448, -0.0061, -0.4739, -0.0387, -0.0061, -0.475, -0.0383, -0.0083, -0.4739, -0.0387, -0.0083, -0.475, -0.0383, -0.0061, -0.4722, -0.0393, -0.0083, -0.4722, -0.0393, -0.01, -0.475, -0.0447, -0.01, -0.4737, -0.0388, -0.01, -0.4747, -0.0384, -0.01, -0.4723, -0.0393, -0.01, -0.4722, -0.0374, -0.0061, -0.4671, -0.0408, -0.0083, -0.4671, -0.0408, -0.0061, -0.47, -0.04, -0.0083, -0.47, -0.04, -0.0061, -0.4682, -0.0405, -0.0083, -0.4682, -0.0405, -0.01, -0.467, -0.0408, -0.01, -0.4702, -0.0399,
-0.01, -0.47, -0.0374, -0.01, -0.468, -0.0405, -0.01, -0.4679, -0.0374, -0.0061, -0.4649, -0.0647, -0.0061, -0.466, -0.0641, -0.0083, -0.4649, -0.0647, -0.0083, -0.466, -0.0641, -0.0061, -0.4635, -0.0654, -0.0083, -0.4635, -0.0654, -0.01, -0.4648, -0.0647, -0.01, -0.4659, -0.0641, -0.01, -0.4637, -0.0653, -0.01, -0.4635, -0.0635, -0.0061, -0.4607, -0.067, -0.0083, -0.4607, -0.067, -0.0061, -0.4582, -0.069, -0.0061, -0.4593, -0.068, -0.0083, -0.4582, -0.069, -0.0083, -0.4593, -0.068, -0.0061, -0.4618, -0.0663, -0.0083, -0.4618, -0.0663, -0.01, -0.4606, -0.067, -0.01, -0.4582, -0.069, -0.01, -0.4593, -0.068, -0.01, -0.4617, -0.0664, -0.01, -0.4614, -0.0635, -0.01, -0.4592, -0.0656, -0.01, -0.4592, -0.0635, -0.01, -0.4657, -0.0613, -0.01, -0.4657, -0.0591,
-0.01, -0.4635, -0.0613, -0.01, -0.4635, -0.0591, -0.01, -0.4657, -0.057, -0.01, -0.4657, -0.0548, -0.01, -0.4635, -0.057, -0.01, -0.4635, -0.0548, -0.01, -0.4614, -0.0613, -0.01, -0.4614, -0.0591, -0.01, -0.4592, -0.0613, -0.01, -0.4592, -0.0591, -0.01, -0.4614, -0.057, -0.01, -0.4614, -0.0548, -0.01, -0.4592, -0.057, -0.01, -0.4592, -0.0548, -0.0061, -0.4572, -0.0701, -0.0083, -0.4572, -0.0701, -0.01, -0.4572, -0.0701, -0.01, -0.457, -0.0678, -0.01, -0.4549, -0.07, -0.01, -0.4549, -0.0678, -0.01, -0.457, -0.0656, -0.01, -0.457, -0.0635, -0.01, -0.4549, -0.0656, -0.01, -0.4549, -0.0635, -0.01, -0.4527, -0.07, -0.01, -0.4527, -0.0678, -0.01, -0.4505, -0.07, -0.01, -0.4505, -0.0678, -0.01, -0.4527, -0.0656, -0.01, -0.4527, -0.0635,
-0.01, -0.4505, -0.0656, -0.01, -0.4505, -0.0635, -0.01, -0.457, -0.0613, -0.01, -0.457, -0.0591, -0.01, -0.4549, -0.0613, -0.01, -0.4549, -0.0591, -0.01, -0.457, -0.057, -0.01, -0.457, -0.0548, -0.01, -0.4549, -0.057, -0.01, -0.4549, -0.0548, -0.01, -0.4527, -0.0613, -0.01, -0.4527, -0.0591, -0.01, -0.4505, -0.0613, -0.01, -0.4505, -0.0591, -0.01, -0.4527, -0.057, -0.01, -0.4527, -0.0548, -0.01, -0.4505, -0.057, -0.01, -0.4505, -0.0548, -0.0061, -0.4657, -0.0482, -0.0083, -0.4657, -0.0482, -0.0061, -0.4635, -0.0487, -0.0083, -0.4635, -0.0487, -0.01, -0.4657, -0.0526, -0.01, -0.4657, -0.0505, -0.01, -0.4635, -0.0526, -0.01, -0.4635, -0.0505, -0.01, -0.4656, -0.0483, -0.01, -0.4635, -0.0488, -0.0061, -0.4604, -0.0494, -0.0083, -0.4604, -0.0494,
-0.0061, -0.4592, -0.0497, -0.0083, -0.4592, -0.0497, -0.0061, -0.4615, -0.0492, -0.0083, -0.4615, -0.0492, -0.01, -0.4614, -0.0526, -0.01, -0.461, -0.0493, -0.01, -0.4592, -0.0526, -0.01, -0.459, -0.0497, -0.01, -0.4618, -0.0491, -0.0061, -0.4657, -0.0411, -0.0083, -0.4657, -0.0411, -0.0061, -0.4635, -0.0416, -0.0083, -0.4635, -0.0416, -0.01, -0.4658, -0.0411, -0.01, -0.4636, -0.0416, -0.01, -0.4657, -0.0396, -0.01, -0.4657, -0.0374, -0.01, -0.4635, -0.0396, -0.01, -0.4635, -0.0374, -0.0061, -0.4614, -0.0421, -0.0083, -0.4614, -0.0421, -0.0061, -0.4592, -0.0424, -0.0083, -0.4592, -0.0424, -0.01, -0.4615, -0.0421, -0.01, -0.4593, -0.0424, -0.01, -0.4614, -0.0396, -0.01, -0.4614, -0.0374, -0.01, -0.4592, -0.0396, -0.01, -0.4592, -0.0374, -0.0061, -0.457, -0.05,
-0.0083, -0.457, -0.05, -0.0061, -0.4549, -0.0502, -0.0083, -0.4549, -0.0502, -0.01, -0.457, -0.0526, -0.01, -0.4569, -0.05, -0.01, -0.4549, -0.0526, -0.01, -0.4548, -0.0502, -0.0061, -0.4527, -0.0503, -0.0083, -0.4527, -0.0503, -0.0061, -0.4505, -0.0503, -0.0083, -0.4505, -0.0503, -0.01, -0.4527, -0.0526, -0.01, -0.4527, -0.0503, -0.01, -0.4505, -0.0526, -0.01, -0.4505, -0.0503, -0.0061, -0.457, -0.0425, -0.0083, -0.457, -0.0425, -0.0061, -0.4549, -0.0425, -0.0083, -0.4549, -0.0425, -0.01, -0.457, -0.0425, -0.01, -0.4549, -0.0425, -0.01, -0.457, -0.0396, -0.01, -0.457, -0.0374, -0.01, -0.4549, -0.0396, -0.01, -0.4549, -0.0374, -0.0061, -0.4527, -0.0425, -0.0083, -0.4527, -0.0425, -0.0061, -0.4505, -0.0425, -0.0083, -0.4505, -0.0425, -0.01, -0.4527, -0.0425,
-0.01, -0.4505, -0.0425, -0.01, -0.4527, -0.0396, -0.01, -0.4527, -0.0374, -0.01, -0.4505, -0.0396, -0.01, -0.4505, -0.0374, -0.0061, -0.4824, -0.0345, -0.0061, -0.4835, -0.0337, -0.0083, -0.4824, -0.0345, -0.0083, -0.4835, -0.0337, -0.0061, -0.4809, -0.0355, -0.0083, -0.4809, -0.0355, -0.01, -0.4824, -0.0345, -0.01, -0.4835, -0.0338, -0.01, -0.4812, -0.0353, -0.01, -0.4809, -0.0331, -0.01, -0.4831, -0.0309, -0.01, -0.4831, -0.0288, -0.01, -0.4809, -0.0309, -0.01, -0.4809, -0.0288, -0.0061, -0.4795, -0.0362, -0.0083, -0.4795, -0.0362, -0.0061, -0.4764, -0.0279, -0.0083, -0.4764, -0.0279, -0.01, -0.4793, -0.0363, -0.01, -0.4787, -0.0331, -0.01, -0.4766, -0.0353, -0.01, -0.4766, -0.0331, -0.01, -0.4787, -0.0309, -0.01, -0.4787, -0.0288, -0.01, -0.4766, -0.0309,
-0.01, -0.4765, -0.0279, -0.0061, -0.483, -0.0259, -0.0061, -0.4841, -0.0255, -0.0083, -0.483, -0.0259, -0.0083, -0.4841, -0.0255, -0.0061, -0.4809, -0.0266, -0.0083, -0.4809, -0.0266, -0.01, -0.483, -0.0259, -0.01, -0.4841, -0.0255, -0.01, -0.4807, -0.0267, -0.0061, -0.4787, -0.0273, -0.0083, -0.4787, -0.0273, -0.0061, -0.4775, -0.0276, -0.0083, -0.4775, -0.0276, -0.01, -0.4787, -0.0273, -0.01, -0.4775, -0.0276, -0.0061, -0.4744, -0.0283, -0.0083, -0.4744, -0.0283, -0.0061, -0.4722, -0.0286, -0.0083, -0.4722, -0.0286, -0.01, -0.4744, -0.0353, -0.01, -0.4744, -0.0331, -0.01, -0.4722, -0.0353, -0.01, -0.4722, -0.0331, -0.01, -0.4744, -0.0309, -0.01, -0.4743, -0.0284, -0.01, -0.4722, -0.0309, -0.01, -0.4722, -0.0286, -0.0061, -0.47, -0.0287, -0.0083, -0.47, -0.0287,
-0.0061, -0.4679, -0.0287, -0.0083, -0.4679, -0.0287, -0.01, -0.47, -0.0353, -0.01, -0.47, -0.0331, -0.01, -0.4679, -0.0353, -0.01, -0.4679, -0.0331, -0.01, -0.47, -0.0309, -0.01, -0.47, -0.0287, -0.01, -0.4679, -0.0309, -0.01, -0.4679, -0.0287, -0.0061, -0.4801, -0.0148, -0.0083, -0.4801, -0.0148, -0.0061, -0.4824, -0.014, -0.0061, -0.4823, -0.0114, -0.0083, -0.4824, -0.014, -0.0083, -0.4823, -0.0114, -0.0061, -0.4812, -0.0144, -0.0083, -0.4812, -0.0144, -0.01, -0.4801, -0.0148, -0.01, -0.4824, -0.014, -0.01, -0.4823, -0.0114, -0.01, -0.4811, -0.0145, -0.01, -0.4809, -0.0114, -0.0061, -0.4787, -0.0152, -0.0083, -0.4787, -0.0152, -0.0061, -0.4766, -0.0159, -0.0083, -0.4766, -0.0159, -0.01, -0.4788, -0.0152, -0.01, -0.4767, -0.0158, -0.01, -0.4787, -0.0136,
-0.01, -0.4787, -0.0114, -0.01, -0.4766, -0.0136, -0.01, -0.4766, -0.0114, -0.0061, -0.4821, -0.0094, -0.0083, -0.4821, -0.0094, -0.0061, -0.482, -0.0083, -0.0061, -0.4818, -0.0071, -0.0083, -0.482, -0.0083, -0.0083, -0.4818, -0.0071, -0.0061, -0.4815, -0.0049, -0.0061, -0.4812, -0.0027, -0.0083, -0.4815, -0.0049, -0.0083, -0.4812, -0.0027, -0.01, -0.4821, -0.0097, -0.01, -0.482, -0.0089, -0.01, -0.4817, -0.0069, -0.01, -0.4814, -0.0048, -0.01, -0.4811, -0.0026, -0.01, -0.4787, -0.0092, -0.01, -0.4787, -0.0071, -0.01, -0.4766, -0.0092, -0.01, -0.4766, -0.0071, -0.01, -0.4787, -0.0049, -0.01, -0.4787, -0.0027, -0.01, -0.4766, -0.0049, -0.01, -0.4766, -0.0027, -0.0061, -0.4744, -0.0164, -0.0083, -0.4744, -0.0164, -0.0061, -0.4713, -0.0168, -0.0061, -0.4724, -0.0167,
-0.0083, -0.4713, -0.0168, -0.0083, -0.4724, -0.0167, -0.01, -0.4746, -0.0164, -0.01, -0.4713, -0.0168, -0.01, -0.4722, -0.0168, -0.01, -0.4744, -0.0136, -0.01, -0.4744, -0.0114, -0.01, -0.4722, -0.0136, -0.01, -0.4722, -0.0114, -0.0061, -0.47, -0.0169, -0.0083, -0.47, -0.0169, -0.0061, -0.4679, -0.017, -0.0083, -0.4679, -0.017, -0.01, -0.47, -0.0169, -0.01, -0.47, -0.0157, -0.01, -0.4679, -0.017, -0.01, -0.4679, -0.0157, -0.01, -0.47, -0.0136, -0.01, -0.47, -0.0114, -0.01, -0.4679, -0.0136, -0.01, -0.4679, -0.0114, -0.01, -0.4744, -0.0092, -0.01, -0.4744, -0.0071, -0.01, -0.4722, -0.0092, -0.01, -0.4722, -0.0071, -0.01, -0.4744, -0.0049, -0.01, -0.4744, -0.0027, -0.01, -0.4722, -0.0049, -0.01, -0.4722, -0.0027, -0.01, -0.47, -0.0092,
-0.01, -0.47, -0.0071, -0.01, -0.4679, -0.0092, -0.01, -0.4679, -0.0071, -0.01, -0.47, -0.0049, -0.01, -0.47, -0.0027, -0.01, -0.4679, -0.0049, -0.01, -0.4679, -0.0027, -0.0061, -0.4657, -0.0287, -0.0083, -0.4657, -0.0287, -0.0061, -0.4635, -0.0287, -0.0083, -0.4635, -0.0287, -0.01, -0.4657, -0.0353, -0.01, -0.4657, -0.0331, -0.01, -0.4635, -0.0353, -0.01, -0.4635, -0.0331, -0.01, -0.4657, -0.0309, -0.01, -0.4657, -0.0287, -0.01, -0.4635, -0.0309, -0.01, -0.4636, -0.0287, -0.0061, -0.4614, -0.0284, -0.0083, -0.4614, -0.0284, -0.0061, -0.4592, -0.028, -0.0083, -0.4592, -0.028, -0.01, -0.4614, -0.0353, -0.01, -0.4614, -0.0331, -0.01, -0.4592, -0.0353, -0.01, -0.4592, -0.0331, -0.01, -0.4614, -0.0309, -0.01, -0.4615, -0.0284, -0.01, -0.4592, -0.0309,
-0.01, -0.4594, -0.028, -0.0061, -0.458, -0.0277, -0.0083, -0.458, -0.0277, -0.01, -0.457, -0.0353, -0.01, -0.457, -0.0331, -0.01, -0.4549, -0.0353, -0.01, -0.4549, -0.0331, -0.01, -0.457, -0.0309, -0.01, -0.4574, -0.0276, -0.01, -0.4549, -0.0309, -0.01, -0.4549, -0.0288, -0.01, -0.4527, -0.0353, -0.01, -0.4527, -0.0331, -0.01, -0.4505, -0.0353, -0.01, -0.4505, -0.0331, -0.01, -0.4527, -0.0309, -0.01, -0.4527, -0.0288, -0.01, -0.4505, -0.0309, -0.01, -0.4505, -0.0288, -0.0061, -0.4569, -0.0274, -0.0083, -0.4569, -0.0274, -0.0061, -0.4549, -0.027, -0.0083, -0.4549, -0.027, -0.01, -0.4566, -0.0274, -0.01, -0.4549, -0.027, -0.0061, -0.4527, -0.0265, -0.0083, -0.4527, -0.0265, -0.0061, -0.4505, -0.026, -0.0083, -0.4505, -0.026, -0.01, -0.4528, -0.0265,
-0.01, -0.4507, -0.0261, -0.0061, -0.4657, -0.017, -0.0061, -0.4657, -0.0157, -0.0083, -0.4657, -0.017, -0.0083, -0.4657, -0.0157, -0.0061, -0.4655, -0.0136, -0.0061, -0.4651, -0.0114, -0.0083, -0.4655, -0.0136, -0.0083, -0.4651, -0.0114, -0.01, -0.4657, -0.017, -0.01, -0.4657, -0.0158, -0.01, -0.4655, -0.0136, -0.01, -0.4652, -0.0115, -0.0061, -0.4648, -0.0095, -0.0083, -0.4648, -0.0095, -0.0061, -0.4646, -0.0084, -0.0061, -0.4643, -0.0071, -0.0083, -0.4646, -0.0084, -0.0083, -0.4643, -0.0071, -0.0061, -0.4638, -0.0049, -0.0061, -0.4633, -0.0027, -0.0083, -0.4638, -0.0049, -0.0083, -0.4633, -0.0027, -0.01, -0.4647, -0.0093, -0.01, -0.4657, -0.0071, -0.01, -0.4645, -0.0083, -0.01, -0.4643, -0.0071, -0.01, -0.4657, -0.0049, -0.01, -0.4657, -0.0027, -0.01, -0.4638, -0.005,
-0.01, -0.4633, -0.0029, -0.01, -0.4483, -0.07, -0.01, -0.4483, -0.0678, -0.01, -0.4462, -0.07, -0.01, -0.4462, -0.0678, -0.01, -0.4483, -0.0656, -0.01, -0.4483, -0.0635, -0.01, -0.4462, -0.0656, -0.01, -0.4462, -0.0635, -0.01, -0.444, -0.07, -0.01, -0.444, -0.0678, -0.01, -0.4418, -0.07, -0.01, -0.4418, -0.0678, -0.01, -0.444, -0.0656, -0.01, -0.444, -0.0635, -0.01, -0.4418, -0.0656, -0.01, -0.4418, -0.0635, -0.01, -0.4483, -0.0613, -0.01, -0.4483, -0.0591, -0.01, -0.4462, -0.0613, -0.01, -0.4462, -0.0591, -0.01, -0.4483, -0.057, -0.01, -0.4483, -0.0548, -0.01, -0.4462, -0.057, -0.01, -0.4462, -0.0548, -0.01, -0.444, -0.0613, -0.01, -0.444, -0.0591, -0.01, -0.4418, -0.0613, -0.01, -0.4418, -0.0591, -0.01, -0.444, -0.057,
-0.01, -0.444, -0.0548, -0.01, -0.4418, -0.057, -0.01, -0.4418, -0.0548, -0.01, -0.4397, -0.07, -0.01, -0.4397, -0.0678, -0.01, -0.4375, -0.07, -0.01, -0.4375, -0.0678, -0.01, -0.4397, -0.0656, -0.01, -0.4397, -0.0635, -0.01, -0.4375, -0.0656, -0.01, -0.4375, -0.0635, -0.01, -0.4353, -0.07, -0.01, -0.4353, -0.0678, -0.01, -0.433, -0.069, -0.01, -0.4332, -0.0678, -0.01, -0.4353, -0.0656, -0.01, -0.4353, -0.0635, -0.01, -0.4332, -0.0656, -0.01, -0.4332, -0.0635, -0.01, -0.4397, -0.0613, -0.01, -0.4397, -0.0591, -0.01, -0.4375, -0.0613, -0.01, -0.4375, -0.0591, -0.01, -0.4397, -0.057, -0.01, -0.4397, -0.0548, -0.01, -0.4375, -0.057, -0.01, -0.4375, -0.0548, -0.01, -0.4353, -0.0613, -0.01, -0.4353, -0.0591, -0.01, -0.4332, -0.0613,
-0.01, -0.4332, -0.0591, -0.01, -0.4353, -0.057, -0.01, -0.4353, -0.0548, -0.01, -0.4332, -0.057, -0.01, -0.4332, -0.0548, -0.0061, -0.4483, -0.0503, -0.0083, -0.4483, -0.0503, -0.0061, -0.4462, -0.0501, -0.0083, -0.4462, -0.0501, -0.01, -0.4483, -0.0526, -0.01, -0.4484, -0.0503, -0.01, -0.4462, -0.0526, -0.01, -0.4463, -0.0501, -0.0061, -0.444, -0.0497, -0.0083, -0.444, -0.0497, -0.0061, -0.4424, -0.0495, -0.0083, -0.4424, -0.0495, -0.0061, -0.4413, -0.0493, -0.0083, -0.4413, -0.0493, -0.01, -0.444, -0.0526, -0.01, -0.4441, -0.0498, -0.01, -0.4418, -0.0526, -0.01, -0.442, -0.0494, -0.01, -0.4411, -0.0492, -0.0061, -0.4483, -0.0425, -0.0083, -0.4483, -0.0425, -0.0061, -0.4462, -0.0423, -0.0083, -0.4462, -0.0423, -0.01, -0.4483, -0.0425, -0.01, -0.4461, -0.0423,
-0.01, -0.4483, -0.0396, -0.01, -0.4483, -0.0374, -0.01, -0.4462, -0.0396, -0.01, -0.4462, -0.0374, -0.0061, -0.444, -0.042, -0.0083, -0.444, -0.042, -0.0061, -0.4418, -0.0416, -0.0083, -0.4418, -0.0416, -0.01, -0.4439, -0.042, -0.01, -0.4417, -0.0416, -0.01, -0.444, -0.0396, -0.01, -0.444, -0.0374, -0.01, -0.4418, -0.0396, -0.01, -0.4418, -0.0374, -0.0061, -0.4397, -0.049, -0.0083, -0.4397, -0.049, -0.0061, -0.4375, -0.0486, -0.0083, -0.4375, -0.0486, -0.01, -0.4397, -0.0526, -0.01, -0.4397, -0.0505, -0.01, -0.4375, -0.0526, -0.01, -0.4375, -0.0505, -0.01, -0.4397, -0.049, -0.01, -0.4376, -0.0486, -0.0061, -0.4353, -0.0481, -0.0083, -0.4353, -0.0481, -0.0061, -0.4331, -0.0476, -0.0083, -0.4331, -0.0476, -0.01, -0.4353, -0.0526, -0.01, -0.4353, -0.0505,
-0.01, -0.4332, -0.0526, -0.01, -0.4332, -0.0505, -0.01, -0.4355, -0.0482, -0.01, -0.4334, -0.0476, -0.0061, -0.4397, -0.0412, -0.0083, -0.4397, -0.0412, -0.0061, -0.4381, -0.0408, -0.0083, -0.4381, -0.0408, -0.0061, -0.437, -0.0405, -0.0083, -0.437, -0.0405, -0.01, -0.4396, -0.0412, -0.01, -0.4382, -0.0409, -0.01, -0.4397, -0.0396, -0.01, -0.4397, -0.0374, -0.01, -0.4373, -0.0406, -0.01, -0.4375, -0.0374, -0.0061, -0.4353, -0.0401, -0.0083, -0.4353, -0.0401, -0.0061, -0.4331, -0.0396, -0.0083, -0.4331, -0.0396, -0.01, -0.4351, -0.0401, -0.01, -0.4353, -0.0374, -0.01, -0.433, -0.0396, -0.01, -0.4332, -0.0374, -0.0061, -0.4328, -0.069, -0.0061, -0.4308, -0.0689, -0.0083, -0.4328, -0.069, -0.0083, -0.4308, -0.0689, -0.0061, -0.4288, -0.0686, -0.0083, -0.4288, -0.0686,
-0.01, -0.4328, -0.069, -0.01, -0.4311, -0.0689, -0.01, -0.4287, -0.0686, -0.01, -0.431, -0.0656, -0.01, -0.431, -0.0635, -0.01, -0.4288, -0.0656, -0.01, -0.4288, -0.0635, -0.0061, -0.4266, -0.0681, -0.0083, -0.4266, -0.0681, -0.0061, -0.4245, -0.0675, -0.0083, -0.4245, -0.0675, -0.01, -0.4265, -0.0681, -0.01, -0.4244, -0.0675, -0.01, -0.4266, -0.0656, -0.01, -0.4266, -0.0635, -0.01, -0.4245, -0.0656, -0.01, -0.4245, -0.0635, -0.01, -0.431, -0.0613, -0.01, -0.431, -0.0591, -0.01, -0.4288, -0.0613, -0.01, -0.4288, -0.0591, -0.01, -0.431, -0.057, -0.01, -0.431, -0.0548, -0.01, -0.4288, -0.057, -0.01, -0.4288, -0.0548, -0.01, -0.4266, -0.0613, -0.01, -0.4266, -0.0591, -0.01, -0.4245, -0.0613, -0.01, -0.4245, -0.0591, -0.01, -0.4266, -0.057,
-0.01, -0.4266, -0.0548, -0.01, -0.4245, -0.057, -0.01, -0.4245, -0.0548, -0.0061, -0.4227, -0.067, -0.0083, -0.4227, -0.067, -0.0061, -0.4216, -0.0666, -0.0083, -0.4216, -0.0666, -0.0061, -0.4201, -0.066, -0.0083, -0.4201, -0.066, -0.01, -0.4229, -0.067, -0.01, -0.4219, -0.0667, -0.01, -0.4223, -0.0635, -0.01, -0.4199, -0.0659, -0.01, -0.4201, -0.0635, -0.0061, -0.418, -0.0651, -0.0083, -0.418, -0.0651, -0.0061, -0.4167, -0.0646, -0.0061, -0.4156, -0.0642, -0.0083, -0.4167, -0.0646, -0.0083, -0.4156, -0.0642, -0.01, -0.4179, -0.0651, -0.01, -0.418, -0.0635, -0.01, -0.4167, -0.0647, -0.01, -0.4157, -0.0642, -0.01, -0.4223, -0.0613, -0.01, -0.4223, -0.0591, -0.01, -0.4201, -0.0613, -0.01, -0.4201, -0.0591, -0.01, -0.4223, -0.057, -0.01, -0.4223, -0.0548,
-0.01, -0.4201, -0.057, -0.01, -0.4201, -0.0548, -0.01, -0.418, -0.0613, -0.01, -0.418, -0.0591, -0.01, -0.4158, -0.0613, -0.01, -0.4158, -0.0591, -0.01, -0.418, -0.057, -0.01, -0.418, -0.0548, -0.01, -0.4158, -0.057, -0.01, -0.4158, -0.0548, -0.0061, -0.432, -0.0472, -0.0061, -0.4309, -0.0469, -0.0083, -0.432, -0.0472, -0.0083, -0.4309, -0.0469, -0.0061, -0.4288, -0.0462, -0.0083, -0.4288, -0.0462, -0.01, -0.431, -0.0526, -0.01, -0.431, -0.0505, -0.01, -0.4288, -0.0526, -0.01, -0.4288, -0.0505, -0.01, -0.4315, -0.0471, -0.01, -0.4306, -0.0468, -0.01, -0.4288, -0.0483, -0.01, -0.429, -0.0463, -0.0061, -0.4267, -0.0454, -0.0083, -0.4267, -0.0454, -0.01, -0.4266, -0.0526, -0.01, -0.4266, -0.0505, -0.01, -0.4245, -0.0526, -0.01, -0.4245, -0.0505,
-0.01, -0.4266, -0.0483, -0.01, -0.4267, -0.0454, -0.01, -0.4245, -0.0483, -0.01, -0.4245, -0.0461, -0.0061, -0.431, -0.039, -0.0083, -0.431, -0.039, -0.0061, -0.4297, -0.0386, -0.0061, -0.4286, -0.0382, -0.0083, -0.4297, -0.0386, -0.0083, -0.4286, -0.0382, -0.01, -0.4309, -0.039, -0.01, -0.431, -0.0374, -0.01, -0.4298, -0.0386, -0.01, -0.4287, -0.0382, -0.0061, -0.4256, -0.045, -0.0083, -0.4256, -0.045, -0.0061, -0.4245, -0.0446, -0.0083, -0.4245, -0.0446, -0.0061, -0.4266, -0.0373, -0.0083, -0.4266, -0.0373, -0.0061, -0.425, -0.0366, -0.0083, -0.425, -0.0366, -0.01, -0.4256, -0.045, -0.01, -0.4245, -0.0446, -0.01, -0.4265, -0.0373, -0.01, -0.4251, -0.0367, -0.01, -0.4223, -0.0526, -0.01, -0.4223, -0.0505, -0.01, -0.4201, -0.0526, -0.01, -0.4201, -0.0505,
-0.01, -0.4223, -0.0483, -0.01, -0.4223, -0.0461, -0.01, -0.4201, -0.0483, -0.01, -0.4201, -0.0461, -0.01, -0.418, -0.0526, -0.01, -0.418, -0.0505, -0.01, -0.4158, -0.0526, -0.01, -0.4158, -0.0505, -0.01, -0.418, -0.0483, -0.01, -0.418, -0.0461, -0.01, -0.4158, -0.0483, -0.01, -0.4158, -0.0461, -0.0061, -0.4223, -0.0436, -0.0083, -0.4223, -0.0436, -0.0061, -0.4209, -0.043, -0.0061, -0.4198, -0.0425, -0.0083, -0.4209, -0.043, -0.0083, -0.4198, -0.0425, -0.01, -0.4226, -0.0437, -0.01, -0.4206, -0.0429, -0.01, -0.4196, -0.0425, -0.0061, -0.418, -0.0418, -0.0083, -0.418, -0.0418, -0.0061, -0.4162, -0.041, -0.0083, -0.4162, -0.041, -0.0061, -0.4152, -0.0405, -0.0083, -0.4152, -0.0405, -0.01, -0.418, -0.0439, -0.01, -0.4182, -0.0419, -0.01, -0.4158, -0.0439,
-0.01, -0.4161, -0.0409, -0.01, -0.4151, -0.0404, -0.01, -0.4483, -0.0353, -0.01, -0.4483, -0.0331, -0.01, -0.4462, -0.0353, -0.01, -0.4462, -0.0331, -0.01, -0.4483, -0.0309, -0.01, -0.4483, -0.0288, -0.01, -0.4462, -0.0309, -0.01, -0.4462, -0.0288, -0.01, -0.444, -0.0353, -0.01, -0.444, -0.0331, -0.01, -0.4418, -0.0353, -0.01, -0.4418, -0.0331, -0.01, -0.444, -0.0309, -0.01, -0.444, -0.0288, -0.01, -0.4418, -0.0309, -0.01, -0.4418, -0.0288, -0.0061, -0.4489, -0.0256, -0.0061, -0.4479, -0.0253, -0.0083, -0.4489, -0.0256, -0.0083, -0.4479, -0.0253, -0.0061, -0.4462, -0.0248, -0.0083, -0.4462, -0.0248, -0.01, -0.4486, -0.0255, -0.01, -0.4477, -0.0253, -0.01, -0.4462, -0.0266, -0.01, -0.4463, -0.0249, -0.0061, -0.444, -0.0241, -0.0083, -0.444, -0.0241,
-0.0061, -0.4424, -0.0235, -0.0083, -0.4424, -0.0235, -0.0061, -0.4413, -0.0231, -0.0083, -0.4413, -0.0231, -0.01, -0.444, -0.0266, -0.01, -0.4442, -0.0242, -0.01, -0.4418, -0.0266, -0.01, -0.4422, -0.0234, -0.01, -0.4412, -0.0231, -0.01, -0.4397, -0.0353, -0.01, -0.4397, -0.0331, -0.01, -0.4375, -0.0353, -0.01, -0.4375, -0.0331, -0.01, -0.4397, -0.0309, -0.01, -0.4397, -0.0288, -0.01, -0.4375, -0.0309, -0.01, -0.4375, -0.0288, -0.01, -0.4353, -0.0353, -0.01, -0.4353, -0.0331, -0.01, -0.4332, -0.0353, -0.01, -0.4332, -0.0331, -0.01, -0.4353, -0.0309, -0.01, -0.4353, -0.0288, -0.01, -0.4332, -0.0309, -0.01, -0.4332, -0.0288, -0.0061, -0.4397, -0.0225, -0.0083, -0.4397, -0.0225, -0.0061, -0.4375, -0.0216, -0.0061, -0.4364, -0.0212, -0.0083, -0.4375, -0.0216,
-0.0083, -0.4364, -0.0212, -0.01, -0.4397, -0.0266, -0.01, -0.4397, -0.0244, -0.01, -0.4375, -0.0266, -0.01, -0.4375, -0.0244, -0.01, -0.4398, -0.0225, -0.01, -0.4375, -0.0216, -0.01, -0.4364, -0.0212, -0.0061, -0.4353, -0.0207, -0.0083, -0.4353, -0.0207, -0.0061, -0.4332, -0.0197, -0.0083, -0.4332, -0.0197, -0.01, -0.4353, -0.0266, -0.01, -0.4353, -0.0244, -0.01, -0.4332, -0.0266, -0.01, -0.4332, -0.0244, -0.01, -0.4353, -0.0222, -0.01, -0.4354, -0.0207, -0.01, -0.4332, -0.0222, -0.01, -0.4334, -0.0199, -0.01, -0.431, -0.0353, -0.01, -0.431, -0.0331, -0.01, -0.4288, -0.0353, -0.01, -0.4288, -0.0331, -0.01, -0.431, -0.0309, -0.01, -0.431, -0.0288, -0.01, -0.4288, -0.0309, -0.01, -0.4288, -0.0288, -0.0061, -0.4239, -0.0361, -0.0083, -0.4239, -0.0361,
-0.01, -0.4266, -0.0353, -0.01, -0.4266, -0.0331, -0.01, -0.4241, -0.0362, -0.01, -0.4245, -0.0331, -0.01, -0.4266, -0.0309, -0.01, -0.4266, -0.0288, -0.01, -0.4245, -0.0309, -0.01, -0.4245, -0.0288, -0.0061, -0.4318, -0.0191, -0.0083, -0.4318, -0.0191, -0.01, -0.431, -0.0266, -0.01, -0.431, -0.0244, -0.01, -0.4288, -0.0266, -0.01, -0.4288, -0.0244, -0.01, -0.431, -0.0222, -0.01, -0.4315, -0.019, -0.01, -0.4288, -0.0222, -0.01, -0.4288, -0.0201, -0.01, -0.4266, -0.0266, -0.01, -0.4266, -0.0244, -0.01, -0.4245, -0.0266, -0.01, -0.4245, -0.0244, -0.01, -0.4266, -0.0222, -0.01, -0.4266, -0.0201, -0.01, -0.4245, -0.0222, -0.01, -0.4245, -0.0201, -0.0061, -0.4223, -0.0354, -0.0083, -0.4223, -0.0354, -0.0061, -0.4204, -0.0345, -0.0061, -0.4194, -0.034,
-0.0083, -0.4204, -0.0345, -0.0083, -0.4194, -0.034, -0.01, -0.4221, -0.0353, -0.01, -0.4223, -0.0331, -0.01, -0.4206, -0.0346, -0.01, -0.4196, -0.0342, -0.01, -0.4223, -0.0309, -0.01, -0.4223, -0.0288, -0.01, -0.4201, -0.0309, -0.01, -0.4201, -0.0288, -0.0061, -0.4179, -0.0333, -0.0083, -0.4179, -0.0333, -0.0061, -0.4162, -0.0324, -0.0083, -0.4162, -0.0324, -0.0061, -0.4151, -0.0318, -0.0083, -0.4151, -0.0318, -0.01, -0.4177, -0.0332, -0.01, -0.4163, -0.0325, -0.01, -0.418, -0.0309, -0.01, -0.418, -0.0288, -0.01, -0.4153, -0.0319, -0.01, -0.4158, -0.0288, -0.01, -0.4223, -0.0266, -0.01, -0.4223, -0.0244, -0.01, -0.4201, -0.0266, -0.01, -0.4201, -0.0244, -0.01, -0.4223, -0.0222, -0.01, -0.4223, -0.0201, -0.01, -0.4201, -0.0222, -0.01, -0.4201, -0.0201,
-0.01, -0.418, -0.0266, -0.01, -0.418, -0.0244, -0.01, -0.4158, -0.0266, -0.01, -0.4158, -0.0244, -0.01, -0.418, -0.0222, -0.01, -0.418, -0.0201, -0.01, -0.4158, -0.0222, -0.01, -0.4158, -0.0201, -0.0061, -0.4307, -0.0186, -0.0083, -0.4307, -0.0186, -0.0061, -0.4288, -0.0178, -0.0083, -0.4288, -0.0178, -0.01, -0.4305, -0.0185, -0.01, -0.429, -0.0179, -0.0061, -0.4272, -0.017, -0.0061, -0.4262, -0.0165, -0.0083, -0.4272, -0.017, -0.0083, -0.4262, -0.0165, -0.0061, -0.4245, -0.0157, -0.0083, -0.4245, -0.0157, -0.01, -0.427, -0.017, -0.01, -0.426, -0.0165, -0.01, -0.4245, -0.0179, -0.01, -0.4247, -0.0158, -0.0061, -0.4258, -0.0017, -0.0083, -0.4258, -0.0017, -0.0061, -0.4245, -0.0023, -0.0083, -0.4245, -0.0023, -0.01, -0.4257, -0.0018, -0.01, -0.4246, -0.0023,
-0.0061, -0.423, -0.0149, -0.0083, -0.423, -0.0149, -0.0061, -0.4219, -0.0143, -0.0083, -0.4219, -0.0143, -0.0061, -0.4201, -0.0134, -0.0083, -0.4201, -0.0134, -0.01, -0.4223, -0.0179, -0.01, -0.4228, -0.0148, -0.01, -0.4201, -0.0179, -0.01, -0.4201, -0.0157, -0.01, -0.4217, -0.0142, -0.01, -0.4204, -0.0135, -0.0061, -0.4187, -0.0127, -0.0061, -0.4176, -0.0121, -0.0083, -0.4187, -0.0127, -0.0083, -0.4176, -0.0121, -0.0061, -0.4158, -0.0111, -0.0083, -0.4158, -0.0111, -0.01, -0.418, -0.0179, -0.01, -0.418, -0.0157, -0.01, -0.4158, -0.0179, -0.01, -0.4158, -0.0157, -0.01, -0.4185, -0.0126, -0.01, -0.4175, -0.012, -0.01, -0.4158, -0.0136, -0.01, -0.4161, -0.0113, -0.0061, -0.4223, -0.003, -0.0083, -0.4223, -0.003, -0.0061, -0.4201, -0.0032, -0.0083, -0.4201, -0.0032,
-0.01, -0.4224, -0.0029, -0.01, -0.4201, -0.0032, -0.0061, -0.4179, -0.003, -0.0083, -0.4179, -0.003, -0.0061, -0.4158, -0.0026, -0.0083, -0.4158, -0.0026, -0.01, -0.4179, -0.003, -0.01, -0.4156, -0.0025, -0.0061, -0.3977, -0.1059, -0.0083, -0.3977, -0.1059, -0.01, -0.3976, -0.1059, -0.0061, -0.3963, -0.1061, -0.0083, -0.3963, -0.1061, -0.0061, -0.3941, -0.1066, -0.0083, -0.3941, -0.1066, -0.01, -0.3963, -0.1061, -0.01, -0.3942, -0.1066, -0.0061, -0.3919, -0.1069, -0.0083, -0.3919, -0.1069, -0.0061, -0.3898, -0.1071, -0.0083, -0.3898, -0.1071, -0.01, -0.392, -0.1069, -0.01, -0.3898, -0.1071, -0.0061, -0.3876, -0.1072, -0.0083, -0.3876, -0.1072, -0.0061, -0.3854, -0.1072, -0.0083, -0.3854, -0.1072, -0.01, -0.3876, -0.1072, -0.01, -0.3854, -0.1072, -0.0061, -0.3832, -0.1072,
-0.0083, -0.3832, -0.1072, -0.0061, -0.3811, -0.1072, -0.0083, -0.3811, -0.1072, -0.01, -0.3832, -0.1072, -0.01, -0.3811, -0.1072, -0.0061, -0.413, -0.1016, -0.0083, -0.413, -0.1016, -0.0061, -0.4115, -0.1022, -0.0083, -0.4115, -0.1022, -0.0061, -0.4141, -0.1012, -0.0083, -0.4141, -0.1012, -0.01, -0.4129, -0.1017, -0.01, -0.4116, -0.1022, -0.01, -0.4139, -0.1013, -0.01, -0.4136, -0.0982, -0.01, -0.4115, -0.1004, -0.01, -0.4115, -0.0982, -0.0061, -0.4093, -0.103, -0.0083, -0.4093, -0.103, -0.0061, -0.4067, -0.1038, -0.0061, -0.4078, -0.1035, -0.0083, -0.4067, -0.1038, -0.0083, -0.4078, -0.1035, -0.01, -0.4095, -0.1029, -0.01, -0.4065, -0.1039, -0.01, -0.4075, -0.1036, -0.01, -0.4093, -0.1004, -0.01, -0.4093, -0.0982, -0.01, -0.4071, -0.1004, -0.01, -0.4071, -0.0982,
-0.01, -0.4136, -0.096, -0.01, -0.4136, -0.0939, -0.01, -0.4115, -0.096, -0.01, -0.4115, -0.0939, -0.01, -0.4136, -0.0917, -0.01, -0.4136, -0.0895, -0.01, -0.4115, -0.0917, -0.01, -0.4115, -0.0895, -0.01, -0.4093, -0.096, -0.01, -0.4093, -0.0939, -0.01, -0.4071, -0.096, -0.01, -0.4071, -0.0939, -0.01, -0.4093, -0.0917, -0.01, -0.4093, -0.0895, -0.01, -0.4071, -0.0917, -0.01, -0.4071, -0.0895, -0.0061, -0.4049, -0.1043, -0.0083, -0.4049, -0.1043, -0.0061, -0.4028, -0.1048, -0.0083, -0.4028, -0.1048, -0.01, -0.405, -0.1043, -0.01, -0.4049, -0.1025, -0.01, -0.4029, -0.1048, -0.01, -0.4028, -0.1025, -0.01, -0.4049, -0.1004, -0.01, -0.4049, -0.0982, -0.01, -0.4028, -0.1004, -0.01, -0.4028, -0.0982, -0.0061, -0.4006, -0.1053, -0.0083, -0.4006, -0.1053,
-0.0061, -0.3988, -0.1056, -0.0083, -0.3988, -0.1056, -0.01, -0.4008, -0.1053, -0.01, -0.4006, -0.1025, -0.01, -0.3985, -0.1057, -0.01, -0.3984, -0.1025, -0.01, -0.4006, -0.1004, -0.01, -0.4006, -0.0982, -0.01, -0.3984, -0.1004, -0.01, -0.3984, -0.0982, -0.0061, -0.4024, -0.0887, -0.0083, -0.4024, -0.0887, -0.01, -0.4049, -0.096, -0.01, -0.4049, -0.0939, -0.01, -0.4028, -0.096, -0.01, -0.4028, -0.0939, -0.01, -0.4049, -0.0917, -0.01, -0.4049, -0.0895, -0.01, -0.4028, -0.0917, -0.01, -0.4026, -0.0886, -0.0061, -0.4006, -0.0893, -0.0083, -0.4006, -0.0893, -0.0061, -0.3984, -0.0899, -0.0083, -0.3984, -0.0899, -0.01, -0.4006, -0.096, -0.01, -0.4006, -0.0939, -0.01, -0.3984, -0.096, -0.01, -0.3984, -0.0939, -0.01, -0.4006, -0.0917, -0.01, -0.4004, -0.0894,
-0.01, -0.3984, -0.0917, -0.01, -0.3984, -0.09, -0.0061, -0.4129, -0.0843, -0.0083, -0.4129, -0.0843, -0.0061, -0.4115, -0.085, -0.0083, -0.4115, -0.085, -0.0061, -0.414, -0.0838, -0.0083, -0.414, -0.0838, -0.01, -0.4136, -0.0874, -0.01, -0.4131, -0.0842, -0.01, -0.4115, -0.0874, -0.01, -0.4112, -0.0851, -0.01, -0.4141, -0.0837, -0.0061, -0.4084, -0.0863, -0.0061, -0.4095, -0.0859, -0.0083, -0.4084, -0.0863, -0.0083, -0.4095, -0.0859, -0.0061, -0.4071, -0.0869, -0.0083, -0.4071, -0.0869, -0.01, -0.4087, -0.0862, -0.01, -0.4097, -0.0858, -0.01, -0.4068, -0.087, -0.0061, -0.4049, -0.0877, -0.0083, -0.4049, -0.0877, -0.0061, -0.4035, -0.0883, -0.0083, -0.4035, -0.0883, -0.01, -0.4048, -0.0878, -0.01, -0.4036, -0.0883, -0.01, -0.3963, -0.1047, -0.01, -0.3963, -0.1025,
-0.01, -0.3941, -0.1047, -0.01, -0.3941, -0.1025, -0.01, -0.3963, -0.1004, -0.01, -0.3963, -0.0982, -0.01, -0.3941, -0.1004, -0.01, -0.3941, -0.0982, -0.01, -0.3919, -0.1047, -0.01, -0.3919, -0.1025, -0.01, -0.3897, -0.1047, -0.01, -0.3897, -0.1025, -0.01, -0.3919, -0.1004, -0.01, -0.3919, -0.0982, -0.01, -0.3897, -0.1004, -0.01, -0.3897, -0.0982, -0.0061, -0.3954, -0.0907, -0.0061, -0.3965, -0.0904, -0.0083, -0.3954, -0.0907, -0.0083, -0.3965, -0.0904, -0.0061, -0.3941, -0.091, -0.0083, -0.3941, -0.091, -0.01, -0.3963, -0.096, -0.01, -0.3963, -0.0939, -0.01, -0.3941, -0.096, -0.01, -0.3941, -0.0939, -0.01, -0.3959, -0.0905, -0.01, -0.3967, -0.0903, -0.01, -0.3939, -0.091, -0.0061, -0.3919, -0.0915, -0.0083, -0.3919, -0.0915, -0.0061, -0.3898, -0.0921,
-0.0083, -0.3898, -0.0921, -0.01, -0.3919, -0.096, -0.01, -0.3919, -0.0939, -0.01, -0.3897, -0.096, -0.01, -0.3897, -0.0939, -0.01, -0.3918, -0.0916, -0.01, -0.3897, -0.0921, -0.01, -0.3876, -0.1047, -0.01, -0.3876, -0.1025, -0.01, -0.3854, -0.1047, -0.01, -0.3854, -0.1025, -0.01, -0.3876, -0.1004, -0.01, -0.3876, -0.0982, -0.01, -0.3854, -0.1004, -0.01, -0.3854, -0.0982, -0.01, -0.3832, -0.1047, -0.01, -0.3832, -0.1025, -0.01, -0.3811, -0.1047, -0.01, -0.3811, -0.1025, -0.01, -0.3832, -0.1004, -0.01, -0.3832, -0.0982, -0.01, -0.3811, -0.1004, -0.01, -0.3811, -0.0982, -0.0061, -0.3876, -0.0924, -0.0083, -0.3876, -0.0924, -0.0061, -0.3854, -0.0925, -0.0083, -0.3854, -0.0925, -0.01, -0.3876, -0.096, -0.01, -0.3876, -0.0939, -0.01, -0.3854, -0.096,
-0.01, -0.3854, -0.0939, -0.01, -0.3876, -0.0924, -0.01, -0.3854, -0.0925, -0.0061, -0.3827, -0.0928, -0.0083, -0.3827, -0.0928, -0.0061, -0.3811, -0.0931, -0.0083, -0.3811, -0.0931, -0.0061, -0.3838, -0.0927, -0.0083, -0.3838, -0.0927, -0.01, -0.3832, -0.096, -0.01, -0.3831, -0.0928, -0.01, -0.3811, -0.096, -0.01, -0.381, -0.0931, -0.01, -0.3839, -0.0927, -0.0061, -0.3789, -0.1072, -0.0083, -0.3789, -0.1072, -0.0061, -0.3767, -0.1072, -0.0083, -0.3767, -0.1072, -0.01, -0.3789, -0.1072, -0.01, -0.3767, -0.1072, -0.0061, -0.3746, -0.1072, -0.0083, -0.3746, -0.1072, -0.0061, -0.3724, -0.1072, -0.0083, -0.3724, -0.1072, -0.01, -0.3746, -0.1072, -0.01, -0.3724, -0.1072, -0.0061, -0.3702, -0.1072, -0.0083, -0.3702, -0.1072, -0.0061, -0.368, -0.1072, -0.0083, -0.368, -0.1072,
-0.01, -0.3702, -0.1072, -0.01, -0.368, -0.1071, -0.0061, -0.3659, -0.107, -0.0083, -0.3659, -0.107, -0.0061, -0.3637, -0.1067, -0.0083, -0.3637, -0.1067, -0.01, -0.3658, -0.107, -0.01, -0.3636, -0.1067, -0.0061, -0.3615, -0.1062, -0.0083, -0.3615, -0.1062, -0.0061, -0.36, -0.1059, -0.0083, -0.36, -0.1059, -0.01, -0.3615, -0.1062, -0.01, -0.3601, -0.1059, -0.01, -0.3789, -0.1047, -0.01, -0.3789, -0.1025, -0.01, -0.3767, -0.1047, -0.01, -0.3767, -0.1025, -0.01, -0.3789, -0.1004, -0.01, -0.3789, -0.0982, -0.01, -0.3767, -0.1004, -0.01, -0.3767, -0.0982, -0.01, -0.3746, -0.1047, -0.01, -0.3746, -0.1025, -0.01, -0.3724, -0.1047, -0.01, -0.3724, -0.1025, -0.01, -0.3746, -0.1004, -0.01, -0.3746, -0.0982, -0.01, -0.3724, -0.1004, -0.01, -0.3724, -0.0982,
-0.0061, -0.3789, -0.0933, -0.0083, -0.3789, -0.0933, -0.0061, -0.3767, -0.0934, -0.0083, -0.3767, -0.0934, -0.01, -0.3789, -0.096, -0.01, -0.3788, -0.0933, -0.01, -0.3767, -0.096, -0.01, -0.3767, -0.0934, -0.0061, -0.3746, -0.0934, -0.0083, -0.3746, -0.0934, -0.0061, -0.3724, -0.0934, -0.0083, -0.3724, -0.0934, -0.01, -0.3746, -0.096, -0.01, -0.3746, -0.0934, -0.01, -0.3724, -0.096, -0.01, -0.3724, -0.0934, -0.01, -0.3702, -0.1047, -0.01, -0.3702, -0.1025, -0.01, -0.368, -0.1047, -0.01, -0.368, -0.1025, -0.01, -0.3702, -0.1004, -0.01, -0.3702, -0.0982, -0.01, -0.368, -0.1004, -0.01, -0.368, -0.0982, -0.01, -0.3659, -0.1047, -0.01, -0.3659, -0.1025, -0.01, -0.3637, -0.1047, -0.01, -0.3637, -0.1025, -0.01, -0.3659, -0.1004, -0.01, -0.3659, -0.0982,
-0.01, -0.3637, -0.1004, -0.01, -0.3637, -0.0982, -0.0061, -0.3702, -0.0934, -0.0083, -0.3702, -0.0934, -0.0061, -0.368, -0.0931, -0.0083, -0.368, -0.0931, -0.01, -0.3702, -0.096, -0.01, -0.3703, -0.0934, -0.01, -0.368, -0.096, -0.01, -0.3682, -0.0931, -0.0061, -0.3664, -0.0929, -0.0083, -0.3664, -0.0929, -0.0061, -0.3653, -0.0927, -0.0083, -0.3653, -0.0927, -0.0061, -0.3637, -0.0925, -0.0083, -0.3637, -0.0925, -0.01, -0.3659, -0.096, -0.01, -0.366, -0.0928, -0.01, -0.3637, -0.096, -0.01, -0.3637, -0.0939, -0.01, -0.3651, -0.0927, -0.01, -0.3637, -0.0925, -0.0061, -0.3589, -0.1057, -0.0083, -0.3589, -0.1057, -0.01, -0.3615, -0.1047, -0.01, -0.3615, -0.1025, -0.01, -0.3592, -0.1057, -0.01, -0.3594, -0.1025, -0.01, -0.3615, -0.1004, -0.01, -0.3615, -0.0982,
-0.01, -0.3594, -0.1004, -0.01, -0.3594, -0.0982, -0.0061, -0.3572, -0.1053, -0.0083, -0.3572, -0.1053, -0.0061, -0.355, -0.1048, -0.0083, -0.355, -0.1048, -0.01, -0.357, -0.1052, -0.01, -0.3572, -0.1025, -0.01, -0.3549, -0.1048, -0.01, -0.355, -0.1025, -0.01, -0.3572, -0.1004, -0.01, -0.3572, -0.0982, -0.01, -0.355, -0.1004, -0.01, -0.355, -0.0982, -0.0061, -0.3615, -0.0924, -0.0083, -0.3615, -0.0924, -0.0061, -0.3594, -0.0922, -0.0083, -0.3594, -0.0922, -0.01, -0.3615, -0.096, -0.01, -0.3615, -0.0939, -0.01, -0.3594, -0.096, -0.01, -0.3594, -0.0939, -0.01, -0.3616, -0.0924, -0.01, -0.3594, -0.0922, -0.0061, -0.3572, -0.0918, -0.0083, -0.3572, -0.0918, -0.0061, -0.355, -0.0912, -0.0083, -0.355, -0.0912, -0.01, -0.3572, -0.096, -0.01, -0.3572, -0.0939,
-0.01, -0.355, -0.096, -0.01, -0.355, -0.0939, -0.01, -0.3573, -0.0918, -0.01, -0.3552, -0.0913, -0.0061, -0.3529, -0.1043, -0.0083, -0.3529, -0.1043, -0.0061, -0.3509, -0.1039, -0.0061, -0.3498, -0.1036, -0.0083, -0.3509, -0.1039, -0.0083, -0.3498, -0.1036, -0.01, -0.3528, -0.1043, -0.01, -0.3529, -0.1025, -0.01, -0.3512, -0.1039, -0.01, -0.3503, -0.1037, -0.01, -0.3529, -0.1004, -0.01, -0.3529, -0.0982, -0.01, -0.3507, -0.1004, -0.01, -0.3507, -0.0982, -0.0061, -0.3485, -0.1032, -0.0083, -0.3485, -0.1032, -0.0061, -0.3463, -0.1025, -0.0083, -0.3463, -0.1025, -0.01, -0.3483, -0.1031, -0.01, -0.3462, -0.1024, -0.01, -0.3485, -0.1004, -0.01, -0.3485, -0.0982, -0.01, -0.3463, -0.1004, -0.01, -0.3463, -0.0982, -0.0061, -0.3534, -0.0908, -0.0061, -0.3523, -0.0905,
-0.0083, -0.3534, -0.0908, -0.0083, -0.3523, -0.0905, -0.0061, -0.3507, -0.0901, -0.0083, -0.3507, -0.0901, -0.01, -0.3529, -0.096, -0.01, -0.3529, -0.0939, -0.01, -0.3507, -0.096, -0.01, -0.3507, -0.0939, -0.01, -0.3531, -0.0907, -0.01, -0.3521, -0.0904, -0.01, -0.3507, -0.0917, -0.01, -0.3508, -0.0901, -0.0061, -0.3485, -0.0895, -0.0083, -0.3485, -0.0895, -0.0061, -0.3465, -0.0888, -0.0083, -0.3465, -0.0888, -0.01, -0.3485, -0.096, -0.01, -0.3485, -0.0939, -0.01, -0.3463, -0.096, -0.01, -0.3463, -0.0939, -0.01, -0.3485, -0.0917, -0.01, -0.3487, -0.0895, -0.01, -0.3463, -0.0917, -0.01, -0.3464, -0.0888, -0.0061, -0.3454, -0.0884, -0.0083, -0.3454, -0.0884, -0.01, -0.3454, -0.0884, -0.0061, -0.4136, -0.0634, -0.0083, -0.4136, -0.0634, -0.0061, -0.4119, -0.0627,
-0.0083, -0.4119, -0.0627, -0.01, -0.4134, -0.0633, -0.01, -0.412, -0.0627, -0.0061, -0.4108, -0.0622, -0.0083, -0.4108, -0.0622, -0.01, -0.4136, -0.0613, -0.01, -0.4136, -0.0591, -0.01, -0.411, -0.0623, -0.01, -0.4115, -0.0591, -0.01, -0.4136, -0.057, -0.01, -0.4136, -0.0548, -0.01, -0.4115, -0.057, -0.01, -0.4115, -0.0548, -0.0061, -0.4093, -0.0615, -0.0083, -0.4093, -0.0615, -0.0061, -0.4074, -0.0606, -0.0061, -0.4063, -0.0601, -0.0083, -0.4074, -0.0606, -0.0083, -0.4063, -0.0601, -0.01, -0.409, -0.0614, -0.01, -0.4093, -0.0591, -0.01, -0.4076, -0.0607, -0.01, -0.4065, -0.0602, -0.01, -0.4093, -0.057, -0.01, -0.4093, -0.0548, -0.01, -0.4071, -0.057, -0.01, -0.4071, -0.0548, -0.0061, -0.4049, -0.0594, -0.0083, -0.4049, -0.0594, -0.0061, -0.4032, -0.0584,
-0.0083, -0.4032, -0.0584, -0.0061, -0.4021, -0.0578, -0.0083, -0.4021, -0.0578, -0.01, -0.4047, -0.0592, -0.01, -0.4033, -0.0585, -0.01, -0.4049, -0.057, -0.01, -0.4049, -0.0548, -0.01, -0.4023, -0.0579, -0.01, -0.4028, -0.0548, -0.0061, -0.4006, -0.0569, -0.0083, -0.4006, -0.0569, -0.0061, -0.3991, -0.0561, -0.0061, -0.398, -0.0555, -0.0083, -0.3991, -0.0561, -0.0083, -0.398, -0.0555, -0.01, -0.4004, -0.0568, -0.01, -0.4006, -0.0548, -0.01, -0.3992, -0.0561, -0.01, -0.3981, -0.0555, -0.01, -0.4136, -0.0526, -0.01, -0.4136, -0.0505, -0.01, -0.4115, -0.0526, -0.01, -0.4115, -0.0505, -0.01, -0.4136, -0.0483, -0.01, -0.4136, -0.0461, -0.01, -0.4115, -0.0483, -0.01, -0.4115, -0.0461, -0.01, -0.4093, -0.0526, -0.01, -0.4093, -0.0505, -0.01, -0.4071, -0.0526,
-0.01, -0.4071, -0.0505, -0.01, -0.4093, -0.0483, -0.01, -0.4093, -0.0461, -0.01, -0.4071, -0.0483, -0.01, -0.4071, -0.0461, -0.0061, -0.4136, -0.0397, -0.0083, -0.4136, -0.0397, -0.0061, -0.4119, -0.0389, -0.0061, -0.4108, -0.0383, -0.0083, -0.4119, -0.0389, -0.0083, -0.4108, -0.0383, -0.01, -0.4136, -0.0439, -0.01, -0.4136, -0.0418, -0.01, -0.4115, -0.0439, -0.01, -0.4115, -0.0418, -0.01, -0.4138, -0.0398, -0.01, -0.4118, -0.0388, -0.01, -0.4107, -0.0383, -0.0061, -0.4093, -0.0375, -0.0083, -0.4093, -0.0375, -0.0061, -0.4077, -0.0366, -0.0083, -0.4077, -0.0366, -0.01, -0.4093, -0.0439, -0.01, -0.4093, -0.0418, -0.01, -0.4071, -0.0439, -0.01, -0.4071, -0.0418, -0.01, -0.4093, -0.0396, -0.01, -0.4095, -0.0376, -0.01, -0.4071, -0.0396, -0.01, -0.4076, -0.0366,
-0.01, -0.4049, -0.0526, -0.01, -0.4049, -0.0505, -0.01, -0.4028, -0.0526, -0.01, -0.4028, -0.0505, -0.01, -0.4049, -0.0483, -0.01, -0.4049, -0.0461, -0.01, -0.4028, -0.0483, -0.01, -0.4028, -0.0461, -0.01, -0.4006, -0.0526, -0.01, -0.4006, -0.0505, -0.01, -0.3984, -0.0526, -0.01, -0.3984, -0.0505, -0.01, -0.4006, -0.0483, -0.01, -0.4006, -0.0461, -0.01, -0.3984, -0.0483, -0.01, -0.3984, -0.0461, -0.01, -0.4049, -0.0439, -0.01, -0.4049, -0.0418, -0.01, -0.4028, -0.0439, -0.01, -0.4028, -0.0418, -0.01, -0.4049, -0.0396, -0.01, -0.4049, -0.0374, -0.01, -0.4028, -0.0396, -0.01, -0.4028, -0.0374, -0.01, -0.4006, -0.0439, -0.01, -0.4006, -0.0418, -0.01, -0.3984, -0.0439, -0.01, -0.3984, -0.0418, -0.01, -0.4006, -0.0396, -0.01, -0.4006, -0.0374,
-0.01, -0.3984, -0.0396, -0.01, -0.3984, -0.0374, -0.0061, -0.3963, -0.0545, -0.0083, -0.3963, -0.0545, -0.0061, -0.395, -0.0538, -0.0083, -0.395, -0.0538, -0.01, -0.3961, -0.0544, -0.01, -0.3951, -0.0538, -0.0061, -0.394, -0.0532, -0.0083, -0.394, -0.0532, -0.01, -0.3963, -0.0526, -0.01, -0.3963, -0.0505, -0.01, -0.3939, -0.0531, -0.01, -0.3941, -0.0505, -0.01, -0.3963, -0.0483, -0.01, -0.3963, -0.0461, -0.01, -0.3941, -0.0483, -0.01, -0.3941, -0.0461, -0.0061, -0.3921, -0.0521, -0.0061, -0.3911, -0.0514, -0.0083, -0.3921, -0.0521, -0.0083, -0.3911, -0.0514, -0.0061, -0.3897, -0.0506, -0.0083, -0.3897, -0.0506, -0.01, -0.3923, -0.0521, -0.01, -0.3912, -0.0515, -0.01, -0.3895, -0.0504, -0.01, -0.3919, -0.0483, -0.01, -0.3919, -0.0461, -0.01, -0.3897, -0.0483,
-0.01, -0.3897, -0.0461, -0.01, -0.3963, -0.0439, -0.01, -0.3963, -0.0418, -0.01, -0.3941, -0.0439, -0.01, -0.3941, -0.0418, -0.01, -0.3963, -0.0396, -0.01, -0.3963, -0.0374, -0.01, -0.3941, -0.0396, -0.01, -0.3941, -0.0374, -0.01, -0.3919, -0.0439, -0.01, -0.3919, -0.0418, -0.01, -0.3897, -0.0439, -0.01, -0.3897, -0.0418, -0.01, -0.3919, -0.0396, -0.01, -0.3919, -0.0374, -0.01, -0.3897, -0.0396, -0.01, -0.3897, -0.0374, -0.0061, -0.3883, -0.0496, -0.0083, -0.3883, -0.0496, -0.0061, -0.3872, -0.0489, -0.0083, -0.3872, -0.0489, -0.0061, -0.3856, -0.0478, -0.0061, -0.3845, -0.0471, -0.0083, -0.3856, -0.0478, -0.0083, -0.3845, -0.0471, -0.01, -0.3883, -0.0497, -0.01, -0.3872, -0.049, -0.01, -0.3876, -0.0461, -0.01, -0.3857, -0.0479, -0.01, -0.3846, -0.0471,
-0.0061, -0.3832, -0.0461, -0.0083, -0.3832, -0.0461, -0.0061, -0.382, -0.0451, -0.0083, -0.382, -0.0451, -0.01, -0.383, -0.0459, -0.01, -0.382, -0.0451, -0.01, -0.3876, -0.0439, -0.01, -0.3876, -0.0418, -0.01, -0.3854, -0.0439, -0.01, -0.3854, -0.0418, -0.01, -0.3876, -0.0396, -0.01, -0.3876, -0.0374, -0.01, -0.3854, -0.0396, -0.01, -0.3854, -0.0374, -0.0061, -0.3809, -0.0443, -0.0083, -0.3809, -0.0443, -0.01, -0.3832, -0.0439, -0.01, -0.3832, -0.0418, -0.01, -0.3809, -0.0443, -0.01, -0.3811, -0.0418, -0.01, -0.3832, -0.0396, -0.01, -0.3832, -0.0374, -0.01, -0.3811, -0.0396, -0.01, -0.3811, -0.0374, -0.0061, -0.4136, -0.0309, -0.0083, -0.4136, -0.0309, -0.0061, -0.4121, -0.0301, -0.0061, -0.411, -0.0295, -0.0083, -0.4121, -0.0301, -0.0083, -0.411, -0.0295,
-0.01, -0.4134, -0.0308, -0.01, -0.4136, -0.0288, -0.01, -0.4122, -0.0301, -0.01, -0.4111, -0.0295, -0.0061, -0.4066, -0.036, -0.0083, -0.4066, -0.036, -0.0061, -0.4093, -0.0284, -0.0083, -0.4093, -0.0284, -0.0061, -0.4081, -0.0277, -0.0083, -0.4081, -0.0277, -0.01, -0.4065, -0.036, -0.01, -0.4091, -0.0283, -0.01, -0.4081, -0.0277, -0.01, -0.4136, -0.0266, -0.01, -0.4136, -0.0244, -0.01, -0.4115, -0.0266, -0.01, -0.4115, -0.0244, -0.01, -0.4136, -0.0222, -0.01, -0.4136, -0.0201, -0.01, -0.4115, -0.0222, -0.01, -0.4115, -0.0201, -0.0061, -0.407, -0.0271, -0.0083, -0.407, -0.0271, -0.01, -0.4093, -0.0266, -0.01, -0.4093, -0.0244, -0.01, -0.407, -0.0271, -0.01, -0.4071, -0.0244, -0.01, -0.4093, -0.0222, -0.01, -0.4093, -0.0201, -0.01, -0.4071, -0.0222,
-0.01, -0.4071, -0.0201, -0.0061, -0.4049, -0.0351, -0.0083, -0.4049, -0.0351, -0.0061, -0.4037, -0.0343, -0.0061, -0.4026, -0.0336, -0.0083, -0.4037, -0.0343, -0.0083, -0.4026, -0.0336, -0.01, -0.4052, -0.0352, -0.01, -0.4035, -0.0342, -0.01, -0.4025, -0.0335, -0.0061, -0.4009, -0.0325, -0.0083, -0.4009, -0.0325, -0.0061, -0.3998, -0.0319, -0.0083, -0.3998, -0.0319, -0.0061, -0.3984, -0.031, -0.0083, -0.3984, -0.031, -0.01, -0.4006, -0.0353, -0.01, -0.4008, -0.0325, -0.01, -0.3984, -0.0353, -0.01, -0.3984, -0.0331, -0.01, -0.3997, -0.0318, -0.01, -0.3987, -0.0312, -0.0061, -0.4053, -0.026, -0.0061, -0.4042, -0.0253, -0.0083, -0.4053, -0.026, -0.0083, -0.4042, -0.0253, -0.0061, -0.4028, -0.0244, -0.0083, -0.4028, -0.0244, -0.01, -0.4054, -0.0261, -0.01, -0.4043, -0.0254,
-0.01, -0.4025, -0.0242, -0.01, -0.4049, -0.0222, -0.01, -0.4049, -0.0201, -0.01, -0.4028, -0.0222, -0.01, -0.4028, -0.0201, -0.0061, -0.4014, -0.0235, -0.0083, -0.4014, -0.0235, -0.0061, -0.4003, -0.0228, -0.0083, -0.4003, -0.0228, -0.0061, -0.3987, -0.0217, -0.0061, -0.3976, -0.021, -0.0083, -0.3987, -0.0217, -0.0083, -0.3976, -0.021, -0.01, -0.4015, -0.0235, -0.01, -0.4004, -0.0228, -0.01, -0.4006, -0.0201, -0.01, -0.3988, -0.0218, -0.01, -0.3978, -0.0211, -0.0061, -0.4145, -0.0104, -0.0083, -0.4145, -0.0104, -0.01, -0.4136, -0.0179, -0.01, -0.4136, -0.0157, -0.01, -0.4115, -0.0179, -0.01, -0.4115, -0.0157, -0.01, -0.4136, -0.0136, -0.01, -0.4143, -0.0103, -0.01, -0.4115, -0.0136, -0.01, -0.4115, -0.0114, -0.01, -0.4093, -0.0179, -0.01, -0.4093, -0.0157,
-0.01, -0.4071, -0.0179, -0.01, -0.4071, -0.0157, -0.01, -0.4093, -0.0136, -0.01, -0.4093, -0.0114, -0.01, -0.4071, -0.0136, -0.01, -0.4071, -0.0114, -0.0061, -0.4134, -0.0098, -0.0083, -0.4134, -0.0098, -0.0061, -0.4115, -0.0087, -0.0083, -0.4115, -0.0087, -0.0061, -0.4141, -0.0019, -0.0083, -0.4141, -0.0019, -0.01, -0.4133, -0.0097, -0.01, -0.4118, -0.0089, -0.01, -0.4142, -0.002, -0.0061, -0.4103, -0.0081, -0.0061, -0.4093, -0.0076, -0.0083, -0.4103, -0.0081, -0.0083, -0.4093, -0.0076, -0.0061, -0.4073, -0.0065, -0.0083, -0.4073, -0.0065, -0.0061, -0.4062, -0.0058, -0.0083, -0.4062, -0.0058, -0.01, -0.41, -0.008, -0.01, -0.4091, -0.0075, -0.01, -0.4071, -0.0092, -0.01, -0.4073, -0.0065, -0.01, -0.4062, -0.0058, -0.01, -0.4049, -0.0179, -0.01, -0.4049, -0.0157,
-0.01, -0.4028, -0.0179, -0.01, -0.4028, -0.0157, -0.01, -0.4049, -0.0136, -0.01, -0.4049, -0.0114, -0.01, -0.4028, -0.0136, -0.01, -0.4028, -0.0114, -0.01, -0.4006, -0.0179, -0.01, -0.4006, -0.0157, -0.01, -0.3984, -0.0179, -0.01, -0.3984, -0.0157, -0.01, -0.4006, -0.0136, -0.01, -0.4006, -0.0114, -0.01, -0.3984, -0.0136, -0.01, -0.3984, -0.0114, -0.0061, -0.4049, -0.0051, -0.0083, -0.4049, -0.0051, -0.0061, -0.4032, -0.0041, -0.0061, -0.4022, -0.0035, -0.0083, -0.4032, -0.0041, -0.0083, -0.4022, -0.0035, -0.01, -0.4049, -0.0092, -0.01, -0.4049, -0.0071, -0.01, -0.4028, -0.0092, -0.01, -0.4028, -0.0071, -0.01, -0.4051, -0.0052, -0.01, -0.4032, -0.0041, -0.01, -0.4021, -0.0035, -0.0061, -0.4006, -0.0026, -0.0083, -0.4006, -0.0026, -0.0061, -0.3992, -0.0018,
-0.0083, -0.3992, -0.0018, -0.01, -0.4006, -0.0092, -0.01, -0.4006, -0.0071, -0.01, -0.3984, -0.0092, -0.01, -0.3984, -0.0071, -0.01, -0.4006, -0.0049, -0.01, -0.4009, -0.0028, -0.01, -0.3984, -0.0049, -0.01, -0.399, -0.0017, -0.0061, -0.397, -0.0301, -0.0061, -0.3959, -0.0294, -0.0083, -0.397, -0.0301, -0.0083, -0.3959, -0.0294, -0.0061, -0.3941, -0.0283, -0.0083, -0.3941, -0.0283, -0.01, -0.3963, -0.0353, -0.01, -0.3963, -0.0331, -0.01, -0.3941, -0.0353, -0.01, -0.3941, -0.0331, -0.01, -0.3968, -0.03, -0.01, -0.3958, -0.0293, -0.01, -0.3941, -0.0309, -0.01, -0.3942, -0.0284, -0.01, -0.3919, -0.0353, -0.01, -0.3919, -0.0331, -0.01, -0.3897, -0.0353, -0.01, -0.3897, -0.0331, -0.01, -0.3919, -0.0309, -0.01, -0.3919, -0.0288, -0.01, -0.3897, -0.0309,
-0.01, -0.3897, -0.0288, -0.0061, -0.393, -0.0277, -0.0083, -0.393, -0.0277, -0.0061, -0.3963, -0.0201, -0.0083, -0.3963, -0.0201, -0.0061, -0.3949, -0.0192, -0.0083, -0.3949, -0.0192, -0.01, -0.393, -0.0276, -0.01, -0.396, -0.0199, -0.01, -0.395, -0.0192, -0.0061, -0.3919, -0.0269, -0.0083, -0.3919, -0.0269, -0.0061, -0.3903, -0.0259, -0.0061, -0.3892, -0.0251, -0.0083, -0.3903, -0.0259, -0.0083, -0.3892, -0.0251, -0.01, -0.3921, -0.027, -0.01, -0.3902, -0.0258, -0.01, -0.3891, -0.0251, -0.01, -0.3876, -0.0353, -0.01, -0.3876, -0.0331, -0.01, -0.3854, -0.0353, -0.01, -0.3854, -0.0331, -0.01, -0.3876, -0.0309, -0.01, -0.3876, -0.0288, -0.01, -0.3854, -0.0309, -0.01, -0.3854, -0.0288, -0.01, -0.3832, -0.0353, -0.01, -0.3832, -0.0331, -0.01, -0.3811, -0.0353,
-0.01, -0.3811, -0.0331, -0.01, -0.3832, -0.0309, -0.01, -0.3832, -0.0288, -0.01, -0.3811, -0.0309, -0.01, -0.3811, -0.0288, -0.0061, -0.3876, -0.0241, -0.0083, -0.3876, -0.0241, -0.0061, -0.3865, -0.0233, -0.0083, -0.3865, -0.0233, -0.0061, -0.3854, -0.0226, -0.0083, -0.3854, -0.0226, -0.01, -0.3876, -0.0266, -0.01, -0.3879, -0.0243, -0.01, -0.3854, -0.0266, -0.01, -0.3863, -0.0232, -0.01, -0.3853, -0.0225, -0.0061, -0.3838, -0.0215, -0.0061, -0.3827, -0.0207, -0.0083, -0.3838, -0.0215, -0.0083, -0.3827, -0.0207, -0.0061, -0.3812, -0.0197, -0.0083, -0.3812, -0.0197, -0.01, -0.3832, -0.0266, -0.01, -0.3832, -0.0244, -0.01, -0.3811, -0.0266, -0.01, -0.3811, -0.0244, -0.01, -0.3837, -0.0215, -0.01, -0.3827, -0.0207, -0.01, -0.3811, -0.0222, -0.01, -0.3812, -0.0197,
-0.0061, -0.3938, -0.0184, -0.0083, -0.3938, -0.0184, -0.01, -0.3963, -0.0179, -0.01, -0.3963, -0.0157, -0.01, -0.3938, -0.0184, -0.01, -0.3941, -0.0157, -0.01, -0.3963, -0.0136, -0.01, -0.3963, -0.0114, -0.01, -0.3941, -0.0136, -0.01, -0.3941, -0.0114, -0.0061, -0.3922, -0.0174, -0.0061, -0.3911, -0.0166, -0.0083, -0.3922, -0.0174, -0.0083, -0.3911, -0.0166, -0.0061, -0.3897, -0.0157, -0.0083, -0.3897, -0.0157, -0.01, -0.3923, -0.0174, -0.01, -0.3913, -0.0167, -0.01, -0.3895, -0.0156, -0.01, -0.3919, -0.0136, -0.01, -0.3919, -0.0114, -0.01, -0.3897, -0.0136, -0.01, -0.3897, -0.0114, -0.01, -0.3963, -0.0092, -0.01, -0.3963, -0.0071, -0.01, -0.3941, -0.0092, -0.01, -0.3941, -0.0071, -0.01, -0.3963, -0.0049, -0.01, -0.3963, -0.0027, -0.01, -0.3941, -0.0049,
-0.01, -0.3941, -0.0027, -0.01, -0.3919, -0.0092, -0.01, -0.3919, -0.0071, -0.01, -0.3897, -0.0092, -0.01, -0.3897, -0.0071, -0.01, -0.3919, -0.0049, -0.01, -0.3919, -0.0027, -0.01, -0.3897, -0.0049, -0.01, -0.3897, -0.0027, -0.0061, -0.3884, -0.0148, -0.0083, -0.3884, -0.0148, -0.0061, -0.3873, -0.0141, -0.0083, -0.3873, -0.0141, -0.0061, -0.3858, -0.013, -0.0061, -0.3847, -0.0122, -0.0083, -0.3858, -0.013, -0.0083, -0.3847, -0.0122, -0.01, -0.3884, -0.0148, -0.01, -0.3873, -0.0141, -0.01, -0.3876, -0.0114, -0.01, -0.3859, -0.013, -0.01, -0.3848, -0.0123, -0.0061, -0.3801, -0.0189, -0.0083, -0.3801, -0.0189, -0.0061, -0.3832, -0.0112, -0.0083, -0.3832, -0.0112, -0.0061, -0.3821, -0.0104, -0.0083, -0.3821, -0.0104, -0.01, -0.3801, -0.0189, -0.01, -0.383, -0.011,
-0.01, -0.3821, -0.0104, -0.01, -0.3876, -0.0092, -0.01, -0.3876, -0.0071, -0.01, -0.3854, -0.0092, -0.01, -0.3854, -0.0071, -0.01, -0.3876, -0.0049, -0.01, -0.3876, -0.0027, -0.01, -0.3854, -0.0049, -0.01, -0.3854, -0.0027, -0.0061, -0.381, -0.0096, -0.0083, -0.381, -0.0096, -0.01, -0.3832, -0.0092, -0.01, -0.3832, -0.0071, -0.01, -0.381, -0.0096, -0.01, -0.3811, -0.0071, -0.01, -0.3832, -0.0049, -0.01, -0.3832, -0.0027, -0.01, -0.3811, -0.0049, -0.01, -0.3811, -0.0027, -0.0061, -0.3796, -0.0432, -0.0061, -0.3785, -0.0423, -0.0083, -0.3796, -0.0432, -0.0083, -0.3785, -0.0423, -0.0061, -0.3771, -0.0412, -0.0083, -0.3771, -0.0412, -0.0061, -0.376, -0.0404, -0.0083, -0.376, -0.0404, -0.01, -0.3796, -0.0432, -0.01, -0.3785, -0.0423, -0.01, -0.3772, -0.0413,
-0.01, -0.3789, -0.0396, -0.01, -0.3789, -0.0374, -0.01, -0.3761, -0.0404, -0.01, -0.3767, -0.0374, -0.0061, -0.3746, -0.0393, -0.0061, -0.3736, -0.0385, -0.0083, -0.3746, -0.0393, -0.0083, -0.3736, -0.0385, -0.0061, -0.3724, -0.0376, -0.0083, -0.3724, -0.0376, -0.01, -0.3747, -0.0394, -0.01, -0.3737, -0.0385, -0.01, -0.3721, -0.0373, -0.0061, -0.3711, -0.0365, -0.0083, -0.3711, -0.0365, -0.01, -0.3711, -0.0365, -0.01, -0.3789, -0.0353, -0.01, -0.3789, -0.0331, -0.01, -0.3767, -0.0353, -0.01, -0.3767, -0.0331, -0.01, -0.3789, -0.0309, -0.01, -0.3789, -0.0288, -0.01, -0.3767, -0.0309, -0.01, -0.3767, -0.0288, -0.01, -0.3746, -0.0353, -0.01, -0.3746, -0.0331, -0.01, -0.3724, -0.0353, -0.01, -0.3724, -0.0331, -0.01, -0.3746, -0.0309, -0.01, -0.3746, -0.0288,
-0.01, -0.3724, -0.0309, -0.01, -0.3724, -0.0288, -0.01, -0.3789, -0.0266, -0.01, -0.3789, -0.0244, -0.01, -0.3767, -0.0266, -0.01, -0.3767, -0.0244, -0.01, -0.3789, -0.0222, -0.01, -0.3789, -0.0201, -0.01, -0.3767, -0.0222, -0.01, -0.3767, -0.0201, -0.01, -0.3746, -0.0266, -0.01, -0.3746, -0.0244, -0.01, -0.3724, -0.0266, -0.01, -0.3724, -0.0244, -0.01, -0.3746, -0.0222, -0.01, -0.3746, -0.0201, -0.01, -0.3724, -0.0222, -0.01, -0.3724, -0.0201, -0.0061, -0.37, -0.0357, -0.0083, -0.37, -0.0357, -0.0061, -0.3686, -0.0346, -0.0061, -0.3675, -0.0337, -0.0083, -0.3686, -0.0346, -0.0083, -0.3675, -0.0337, -0.01, -0.37, -0.0357, -0.01, -0.3702, -0.0331, -0.01, -0.3686, -0.0346, -0.01, -0.3676, -0.0338, -0.01, -0.3702, -0.0309, -0.01, -0.3702, -0.0288,
-0.01, -0.368, -0.0309, -0.01, -0.368, -0.0288, -0.0061, -0.3662, -0.0327, -0.0083, -0.3662, -0.0327, -0.0061, -0.3651, -0.0318, -0.0083, -0.3651, -0.0318, -0.0061, -0.3637, -0.0307, -0.0083, -0.3637, -0.0307, -0.01, -0.3662, -0.0327, -0.01, -0.3651, -0.0318, -0.01, -0.3659, -0.0288, -0.01, -0.3635, -0.0305, -0.01, -0.3637, -0.0288, -0.01, -0.3702, -0.0266, -0.01, -0.3702, -0.0244, -0.01, -0.368, -0.0266, -0.01, -0.368, -0.0244, -0.01, -0.3702, -0.0222, -0.01, -0.3702, -0.0201, -0.01, -0.368, -0.0222, -0.01, -0.368, -0.0201, -0.01, -0.3659, -0.0266, -0.01, -0.3659, -0.0244, -0.01, -0.3637, -0.0266, -0.01, -0.3637, -0.0244, -0.01, -0.3659, -0.0222, -0.01, -0.3659, -0.0201, -0.01, -0.3637, -0.0222, -0.01, -0.3637, -0.0201, -0.0061, -0.3789, -0.018,
-0.0083, -0.3789, -0.018, -0.0061, -0.3775, -0.017, -0.0061, -0.3764, -0.0163, -0.0083, -0.3775, -0.017, -0.0083, -0.3764, -0.0163, -0.01, -0.3791, -0.0182, -0.01, -0.3774, -0.017, -0.01, -0.3763, -0.0162, -0.0061, -0.3749, -0.0152, -0.0083, -0.3749, -0.0152, -0.0061, -0.3738, -0.0144, -0.0083, -0.3738, -0.0144, -0.0061, -0.3724, -0.0134, -0.0083, -0.3724, -0.0134, -0.01, -0.3746, -0.0179, -0.01, -0.3749, -0.0152, -0.01, -0.3724, -0.0179, -0.01, -0.3724, -0.0157, -0.01, -0.3738, -0.0144, -0.01, -0.3727, -0.0136, -0.0061, -0.3794, -0.0085, -0.0061, -0.3783, -0.0078, -0.0083, -0.3794, -0.0085, -0.0083, -0.3783, -0.0078, -0.0061, -0.3768, -0.0067, -0.0083, -0.3768, -0.0067, -0.0061, -0.3757, -0.0059, -0.0083, -0.3757, -0.0059, -0.01, -0.3795, -0.0086, -0.01, -0.3784, -0.0078,
-0.01, -0.3769, -0.0068, -0.01, -0.3789, -0.0049, -0.01, -0.3789, -0.0027, -0.01, -0.3759, -0.006, -0.01, -0.3767, -0.0027, -0.0061, -0.3746, -0.0051, -0.0083, -0.3746, -0.0051, -0.0061, -0.3731, -0.0041, -0.0061, -0.372, -0.0033, -0.0083, -0.3731, -0.0041, -0.0083, -0.372, -0.0033, -0.01, -0.3742, -0.0049, -0.01, -0.3746, -0.0027, -0.01, -0.3731, -0.0041, -0.01, -0.372, -0.0033, -0.0061, -0.3712, -0.0125, -0.0061, -0.3701, -0.0118, -0.0083, -0.3712, -0.0125, -0.0083, -0.3701, -0.0118, -0.0061, -0.3686, -0.0107, -0.0083, -0.3686, -0.0107, -0.01, -0.3702, -0.0179, -0.01, -0.3702, -0.0157, -0.01, -0.368, -0.0179, -0.01, -0.368, -0.0157, -0.01, -0.3711, -0.0124, -0.01, -0.37, -0.0117, -0.01, -0.368, -0.0136, -0.01, -0.3685, -0.0106, -0.01, -0.3659, -0.0179,
-0.01, -0.3659, -0.0157, -0.01, -0.3637, -0.0179, -0.01, -0.3637, -0.0157, -0.01, -0.3659, -0.0136, -0.01, -0.3659, -0.0114, -0.01, -0.3637, -0.0136, -0.01, -0.3637, -0.0114, -0.0061, -0.3675, -0.0099, -0.0083, -0.3675, -0.0099, -0.0061, -0.3704, -0.0022, -0.0083, -0.3704, -0.0022, -0.01, -0.3675, -0.0099, -0.01, -0.3706, -0.0023, -0.0061, -0.366, -0.0088, -0.0061, -0.3649, -0.0081, -0.0083, -0.366, -0.0088, -0.0083, -0.3649, -0.0081, -0.0061, -0.3637, -0.0072, -0.0083, -0.3637, -0.0072, -0.01, -0.366, -0.0088, -0.01, -0.3649, -0.008, -0.01, -0.3637, -0.0092, -0.01, -0.3639, -0.0074, -0.0061, -0.3626, -0.0298, -0.0061, -0.3615, -0.029, -0.0083, -0.3626, -0.0298, -0.0083, -0.3615, -0.029, -0.0061, -0.3601, -0.0279, -0.0083, -0.3601, -0.0279, -0.01, -0.3626, -0.0298,
-0.01, -0.3615, -0.029, -0.01, -0.3602, -0.0279, -0.0061, -0.3591, -0.0271, -0.0083, -0.3591, -0.0271, -0.01, -0.3615, -0.0266, -0.01, -0.3615, -0.0244, -0.01, -0.3591, -0.0271, -0.01, -0.3594, -0.0244, -0.01, -0.3615, -0.0222, -0.01, -0.3615, -0.0201, -0.01, -0.3594, -0.0222, -0.01, -0.3594, -0.0201, -0.0061, -0.3577, -0.026, -0.0061, -0.3566, -0.0251, -0.0083, -0.3577, -0.026, -0.0083, -0.3566, -0.0251, -0.0061, -0.3553, -0.024, -0.0083, -0.3553, -0.024, -0.0061, -0.3542, -0.0231, -0.0083, -0.3542, -0.0231, -0.01, -0.3577, -0.026, -0.01, -0.3566, -0.0252, -0.01, -0.3553, -0.0241, -0.01, -0.3572, -0.0222, -0.01, -0.3572, -0.0201, -0.01, -0.3543, -0.0231, -0.01, -0.355, -0.0201, -0.0061, -0.3529, -0.022, -0.0061, -0.3519, -0.0211, -0.0083, -0.3529, -0.022,
-0.0083, -0.3519, -0.0211, -0.0061, -0.3507, -0.0201, -0.0083, -0.3507, -0.0201, -0.01, -0.353, -0.0221, -0.01, -0.3519, -0.0212, -0.01, -0.3504, -0.0199, -0.0061, -0.3494, -0.0191, -0.0083, -0.3494, -0.0191, -0.01, -0.3495, -0.0191, -0.01, -0.3615, -0.0179, -0.01, -0.3615, -0.0157, -0.01, -0.3594, -0.0179, -0.01, -0.3594, -0.0157, -0.01, -0.3615, -0.0136, -0.01, -0.3615, -0.0114, -0.01, -0.3594, -0.0136, -0.01, -0.3594, -0.0114, -0.01, -0.3572, -0.0179, -0.01, -0.3572, -0.0157, -0.01, -0.355, -0.0179, -0.01, -0.355, -0.0157, -0.01, -0.3572, -0.0136, -0.01, -0.3572, -0.0114, -0.01, -0.355, -0.0136, -0.01, -0.355, -0.0114, -0.0061, -0.3623, -0.0062, -0.0083, -0.3623, -0.0062, -0.0061, -0.3612, -0.0055, -0.0083, -0.3612, -0.0055, -0.0061, -0.3596, -0.0044,
-0.0061, -0.3585, -0.0036, -0.0083, -0.3596, -0.0044, -0.0083, -0.3585, -0.0036, -0.01, -0.3615, -0.0092, -0.01, -0.3622, -0.0061, -0.01, -0.3594, -0.0092, -0.01, -0.3594, -0.0071, -0.01, -0.3611, -0.0054, -0.01, -0.3596, -0.0044, -0.01, -0.3585, -0.0036, -0.0061, -0.3572, -0.0027, -0.0083, -0.3572, -0.0027, -0.0061, -0.3559, -0.0018, -0.0083, -0.3559, -0.0018, -0.01, -0.3572, -0.0092, -0.01, -0.3572, -0.0071, -0.01, -0.355, -0.0092, -0.01, -0.355, -0.0071, -0.01, -0.3572, -0.0049, -0.01, -0.3574, -0.0029, -0.01, -0.355, -0.0049, -0.01, -0.3557, -0.0017, -0.01, -0.3529, -0.0179, -0.01, -0.3529, -0.0157, -0.01, -0.3507, -0.0179, -0.01, -0.3507, -0.0157, -0.01, -0.3529, -0.0136, -0.01, -0.3529, -0.0114, -0.01, -0.3507, -0.0136, -0.01, -0.3507, -0.0114,
-0.0061, -0.3484, -0.0182, -0.0083, -0.3484, -0.0182, -0.0061, -0.347, -0.0172, -0.0061, -0.3459, -0.0163, -0.0083, -0.347, -0.0172, -0.0083, -0.3459, -0.0163, -0.01, -0.3483, -0.0182, -0.01, -0.3485, -0.0157, -0.01, -0.347, -0.0172, -0.01, -0.346, -0.0163, -0.01, -0.3485, -0.0136, -0.01, -0.3485, -0.0114, -0.01, -0.3463, -0.0136, -0.01, -0.3463, -0.0114, -0.01, -0.3529, -0.0092, -0.01, -0.3529, -0.0071, -0.01, -0.3507, -0.0092, -0.01, -0.3507, -0.0071, -0.01, -0.3529, -0.0049, -0.01, -0.3529, -0.0027, -0.01, -0.3507, -0.0049, -0.01, -0.3507, -0.0027, -0.01, -0.3485, -0.0092, -0.01, -0.3485, -0.0071, -0.01, -0.3463, -0.0092, -0.01, -0.3463, -0.0071, -0.01, -0.3485, -0.0049, -0.01, -0.3485, -0.0027, -0.01, -0.3463, -0.0049, -0.01, -0.3463, -0.0027,
-0.0061, -0.3444, -0.1018, -0.0083, -0.3444, -0.1018, -0.0061, -0.3433, -0.1014, -0.0083, -0.3433, -0.1014, -0.0061, -0.342, -0.1008, -0.0083, -0.342, -0.1008, -0.01, -0.3446, -0.1019, -0.01, -0.3437, -0.1015, -0.01, -0.3442, -0.0982, -0.01, -0.3417, -0.1007, -0.01, -0.342, -0.0982, -0.0061, -0.3398, -0.1, -0.0083, -0.3398, -0.1, -0.0061, -0.3385, -0.0994, -0.0061, -0.3374, -0.099, -0.0083, -0.3385, -0.0994, -0.0083, -0.3374, -0.099, -0.01, -0.3397, -0.0999, -0.01, -0.3398, -0.0982, -0.01, -0.3385, -0.0994, -0.01, -0.3375, -0.099, -0.01, -0.3442, -0.096, -0.01, -0.3442, -0.0939, -0.01, -0.342, -0.096, -0.01, -0.342, -0.0939, -0.01, -0.3442, -0.0917, -0.01, -0.3442, -0.0895, -0.01, -0.342, -0.0917, -0.01, -0.342, -0.0895, -0.01, -0.3398, -0.096,
-0.01, -0.3398, -0.0939, -0.01, -0.3377, -0.096, -0.01, -0.3377, -0.0939, -0.01, -0.3398, -0.0917, -0.01, -0.3398, -0.0895, -0.01, -0.3377, -0.0917, -0.01, -0.3377, -0.0895, -0.0061, -0.3355, -0.0981, -0.0083, -0.3355, -0.0981, -0.0061, -0.3339, -0.0974, -0.0083, -0.3339, -0.0974, -0.01, -0.3353, -0.0981, -0.01, -0.334, -0.0974, -0.0061, -0.3328, -0.0968, -0.0083, -0.3328, -0.0968, -0.01, -0.3355, -0.096, -0.01, -0.3355, -0.0939, -0.01, -0.3329, -0.0969, -0.01, -0.3333, -0.0939, -0.01, -0.3355, -0.0917, -0.01, -0.3355, -0.0895, -0.01, -0.3333, -0.0917, -0.01, -0.3333, -0.0895, -0.0061, -0.3311, -0.096, -0.0083, -0.3311, -0.096, -0.0061, -0.3297, -0.0952, -0.0061, -0.3286, -0.0945, -0.0083, -0.3297, -0.0952, -0.0083, -0.3286, -0.0945, -0.01, -0.3309, -0.0958,
-0.01, -0.3312, -0.0939, -0.01, -0.3298, -0.0952, -0.01, -0.3287, -0.0946, -0.01, -0.3312, -0.0917, -0.01, -0.3312, -0.0895, -0.01, -0.329, -0.0917, -0.01, -0.329, -0.0895, -0.0061, -0.3442, -0.0879, -0.0083, -0.3442, -0.0879, -0.0061, -0.342, -0.0869, -0.0083, -0.342, -0.0869, -0.01, -0.3443, -0.0879, -0.01, -0.3423, -0.087, -0.0061, -0.3407, -0.0864, -0.0061, -0.3396, -0.0859, -0.0083, -0.3407, -0.0864, -0.0083, -0.3396, -0.0859, -0.0061, -0.3377, -0.085, -0.0083, -0.3377, -0.085, -0.01, -0.3404, -0.0862, -0.01, -0.3394, -0.0858, -0.01, -0.3377, -0.0874, -0.01, -0.3379, -0.0851, -0.0061, -0.3362, -0.0843, -0.0083, -0.3362, -0.0843, -0.0061, -0.3351, -0.0838, -0.0083, -0.3351, -0.0838, -0.0061, -0.3333, -0.0829, -0.0083, -0.3333, -0.0829, -0.01, -0.3355, -0.0874,
-0.01, -0.336, -0.0842, -0.01, -0.3333, -0.0874, -0.01, -0.3333, -0.0852, -0.01, -0.3349, -0.0837, -0.01, -0.3336, -0.083, -0.0061, -0.3319, -0.0821, -0.0061, -0.3308, -0.0815, -0.0083, -0.3319, -0.0821, -0.0083, -0.3308, -0.0815, -0.0061, -0.329, -0.0804, -0.0083, -0.329, -0.0804, -0.01, -0.3312, -0.0874, -0.01, -0.3312, -0.0852, -0.01, -0.329, -0.0874, -0.01, -0.329, -0.0852, -0.01, -0.3317, -0.082, -0.01, -0.3307, -0.0814, -0.01, -0.329, -0.083, -0.01, -0.3291, -0.0804, -0.0061, -0.328, -0.0797, -0.0083, -0.328, -0.0797, -0.01, -0.328, -0.0797, -0.0061, -0.3268, -0.0935, -0.0083, -0.3268, -0.0935, -0.0061, -0.3257, -0.0928, -0.0083, -0.3257, -0.0928, -0.0061, -0.3246, -0.0921, -0.0083, -0.3246, -0.0921, -0.01, -0.3267, -0.0934, -0.01, -0.3257, -0.0928,
-0.01, -0.3268, -0.0917, -0.01, -0.3268, -0.0895, -0.01, -0.3246, -0.0921, -0.01, -0.3246, -0.0895, -0.0061, -0.3229, -0.091, -0.0061, -0.3219, -0.0903, -0.0083, -0.3229, -0.091, -0.0083, -0.3219, -0.0903, -0.0061, -0.3203, -0.0892, -0.0083, -0.3203, -0.0892, -0.01, -0.323, -0.0911, -0.01, -0.3219, -0.0903, -0.01, -0.3205, -0.0893, -0.01, -0.3268, -0.0874, -0.01, -0.3268, -0.0852, -0.01, -0.3246, -0.0874, -0.01, -0.3246, -0.0852, -0.01, -0.3268, -0.083, -0.01, -0.3268, -0.0808, -0.01, -0.3246, -0.083, -0.01, -0.3246, -0.0808, -0.0061, -0.3193, -0.0884, -0.0083, -0.3193, -0.0884, -0.01, -0.3225, -0.0874, -0.01, -0.3225, -0.0852, -0.01, -0.3194, -0.0885, -0.01, -0.3203, -0.0852, -0.01, -0.3225, -0.083, -0.01, -0.3225, -0.0808, -0.01, -0.3203, -0.083,
-0.01, -0.3203, -0.0808, -0.0061, -0.3268, -0.079, -0.0083, -0.3268, -0.079, -0.0061, -0.3253, -0.0779, -0.0061, -0.3242, -0.0772, -0.0083, -0.3253, -0.0779, -0.0083, -0.3242, -0.0772, -0.01, -0.327, -0.0791, -0.01, -0.3252, -0.0779, -0.01, -0.3241, -0.0771, -0.0061, -0.3226, -0.0761, -0.0083, -0.3226, -0.0761, -0.0061, -0.3216, -0.0753, -0.0083, -0.3216, -0.0753, -0.0061, -0.3203, -0.0743, -0.0083, -0.3203, -0.0743, -0.01, -0.3225, -0.0787, -0.01, -0.3226, -0.0761, -0.01, -0.3203, -0.0787, -0.01, -0.3203, -0.0765, -0.01, -0.3215, -0.0753, -0.01, -0.3206, -0.0745, -0.0061, -0.3181, -0.0875, -0.0083, -0.3181, -0.0875, -0.0061, -0.3168, -0.0865, -0.0061, -0.3157, -0.0856, -0.0083, -0.3168, -0.0865, -0.0083, -0.3157, -0.0856, -0.01, -0.3178, -0.0873, -0.01, -0.3181, -0.0852,
-0.01, -0.3168, -0.0865, -0.01, -0.3157, -0.0856, -0.01, -0.3181, -0.083, -0.01, -0.3181, -0.0808, -0.01, -0.316, -0.083, -0.01, -0.316, -0.0808, -0.0061, -0.3143, -0.0845, -0.0083, -0.3143, -0.0845, -0.0061, -0.3133, -0.0836, -0.0083, -0.3133, -0.0836, -0.0061, -0.312, -0.0826, -0.0061, -0.3109, -0.0816, -0.0083, -0.312, -0.0826, -0.0083, -0.3109, -0.0816, -0.01, -0.3144, -0.0846, -0.01, -0.3133, -0.0837, -0.01, -0.3138, -0.0808, -0.01, -0.312, -0.0826, -0.01, -0.3109, -0.0816, -0.0061, -0.3191, -0.0733, -0.0061, -0.318, -0.0724, -0.0083, -0.3191, -0.0733, -0.0083, -0.318, -0.0724, -0.0061, -0.3167, -0.0713, -0.0083, -0.3167, -0.0713, -0.01, -0.3181, -0.0787, -0.01, -0.3181, -0.0765, -0.01, -0.316, -0.0787, -0.01, -0.316, -0.0765, -0.01, -0.319, -0.0733,
-0.01, -0.318, -0.0724, -0.01, -0.316, -0.0743, -0.01, -0.3167, -0.0713, -0.01, -0.3138, -0.0787, -0.01, -0.3138, -0.0765, -0.01, -0.3116, -0.0787, -0.01, -0.3116, -0.0765, -0.01, -0.3138, -0.0743, -0.01, -0.3138, -0.0722, -0.01, -0.3116, -0.0743, -0.01, -0.3116, -0.0722, -0.0061, -0.3097, -0.0805, -0.0083, -0.3097, -0.0805, -0.01, -0.3097, -0.0806, -0.0061, -0.3086, -0.0795, -0.0083, -0.3086, -0.0795, -0.0061, -0.3075, -0.0784, -0.0061, -0.3065, -0.0773, -0.0083, -0.3075, -0.0784, -0.0083, -0.3065, -0.0773, -0.01, -0.3086, -0.0795, -0.01, -0.3095, -0.0765, -0.01, -0.3075, -0.0784, -0.01, -0.3065, -0.0773, -0.01, -0.3095, -0.0743, -0.01, -0.3095, -0.0722, -0.01, -0.3073, -0.0743, -0.01, -0.3073, -0.0722, -0.0061, -0.3054, -0.0762, -0.0083, -0.3054, -0.0762,
-0.0061, -0.3044, -0.075, -0.0083, -0.3044, -0.075, -0.0061, -0.3034, -0.0739, -0.0061, -0.3024, -0.0726, -0.0083, -0.3034, -0.0739, -0.0083, -0.3024, -0.0726, -0.01, -0.3054, -0.0762, -0.01, -0.3043, -0.075, -0.01, -0.3051, -0.0722, -0.01, -0.3034, -0.0739, -0.01, -0.3023, -0.0726, -0.0061, -0.3015, -0.0715, -0.0083, -0.3015, -0.0715, -0.01, -0.3015, -0.0715, -0.0061, -0.3156, -0.0704, -0.0083, -0.3156, -0.0704, -0.01, -0.3156, -0.0704, -0.0061, -0.3144, -0.0693, -0.0061, -0.3133, -0.0683, -0.0083, -0.3144, -0.0693, -0.0083, -0.3133, -0.0683, -0.0061, -0.3122, -0.0672, -0.0083, -0.3122, -0.0672, -0.0061, -0.3111, -0.0661, -0.0083, -0.3111, -0.0661, -0.01, -0.3144, -0.0693, -0.01, -0.3133, -0.0683, -0.01, -0.3116, -0.07, -0.01, -0.3122, -0.0673, -0.01, -0.3112, -0.0661,
-0.0061, -0.3446, -0.0152, -0.0083, -0.3446, -0.0152, -0.0061, -0.3435, -0.0143, -0.0083, -0.3435, -0.0143, -0.0061, -0.3421, -0.0132, -0.0061, -0.341, -0.0124, -0.0083, -0.3421, -0.0132, -0.0083, -0.341, -0.0124, -0.01, -0.3446, -0.0152, -0.01, -0.3435, -0.0144, -0.01, -0.3442, -0.0114, -0.01, -0.3422, -0.0133, -0.01, -0.3411, -0.0125, -0.0061, -0.3398, -0.0115, -0.0083, -0.3398, -0.0115, -0.0061, -0.3385, -0.0105, -0.0083, -0.3385, -0.0105, -0.01, -0.3396, -0.0113, -0.01, -0.3385, -0.0105, -0.01, -0.3442, -0.0092, -0.01, -0.3442, -0.0071, -0.01, -0.342, -0.0092, -0.01, -0.342, -0.0071, -0.01, -0.3442, -0.0049, -0.01, -0.3442, -0.0027, -0.01, -0.342, -0.0049, -0.01, -0.342, -0.0027, -0.0061, -0.3374, -0.0097, -0.0083, -0.3374, -0.0097, -0.01, -0.3398, -0.0092,
-0.01, -0.3398, -0.0071, -0.01, -0.3374, -0.0097, -0.01, -0.3377, -0.0071, -0.01, -0.3398, -0.0049, -0.01, -0.3398, -0.0027, -0.01, -0.3377, -0.0049, -0.01, -0.3377, -0.0027, -0.0061, -0.336, -0.0086, -0.0061, -0.3349, -0.0078, -0.0083, -0.336, -0.0086, -0.0083, -0.3349, -0.0078, -0.0061, -0.3335, -0.0067, -0.0083, -0.3335, -0.0067, -0.0061, -0.3324, -0.0059, -0.0083, -0.3324, -0.0059, -0.01, -0.3361, -0.0086, -0.01, -0.335, -0.0078, -0.01, -0.3336, -0.0067, -0.01, -0.3355, -0.0049, -0.01, -0.3355, -0.0027, -0.01, -0.3325, -0.0059, -0.01, -0.3333, -0.0027, -0.0061, -0.3312, -0.0049, -0.0083, -0.3312, -0.0049, -0.0061, -0.3298, -0.004, -0.0061, -0.3287, -0.0032, -0.0083, -0.3298, -0.004, -0.0083, -0.3287, -0.0032, -0.01, -0.3309, -0.0047, -0.01, -0.3312, -0.0027,
-0.01, -0.3298, -0.004, -0.01, -0.3287, -0.0032, -0.0061, -0.3272, -0.0021, -0.0083, -0.3272, -0.0021, -0.01, -0.3273, -0.0022, -0.0061, -0.3102, -0.065, -0.0061, -0.3091, -0.0636, -0.0083, -0.3102, -0.065, -0.0083, -0.3091, -0.0636, -0.0061, -0.3082, -0.0626, -0.0083, -0.3082, -0.0626, -0.01, -0.3095, -0.07, -0.01, -0.3095, -0.0678, -0.01, -0.3073, -0.07, -0.01, -0.3073, -0.0678, -0.01, -0.3102, -0.065, -0.01, -0.3092, -0.0637, -0.01, -0.3073, -0.0656, -0.01, -0.3083, -0.0626, -0.01, -0.3051, -0.07, -0.01, -0.3051, -0.0678, -0.01, -0.3029, -0.07, -0.01, -0.3029, -0.0678, -0.01, -0.3051, -0.0656, -0.01, -0.3051, -0.0635, -0.01, -0.3029, -0.0656, -0.01, -0.3029, -0.0635, -0.0061, -0.3072, -0.0613, -0.0061, -0.3063, -0.0601, -0.0083, -0.3072, -0.0613,
-0.0083, -0.3063, -0.0601, -0.01, -0.307, -0.0611, -0.01, -0.3063, -0.0601, -0.0061, -0.3055, -0.059, -0.0083, -0.3055, -0.059, -0.0061, -0.3044, -0.0576, -0.0083, -0.3044, -0.0576, -0.0061, -0.3036, -0.0565, -0.0061, -0.3025, -0.055, -0.0083, -0.3036, -0.0565, -0.0083, -0.3025, -0.055, -0.01, -0.3051, -0.0613, -0.01, -0.3054, -0.059, -0.01, -0.3029, -0.0613, -0.01, -0.3029, -0.0591, -0.01, -0.3044, -0.0576, -0.01, -0.3036, -0.0565, -0.01, -0.3026, -0.0551, -0.0061, -0.3004, -0.0701, -0.0083, -0.3004, -0.0701, -0.0061, -0.2996, -0.069, -0.0061, -0.2987, -0.0678, -0.0083, -0.2996, -0.069, -0.0083, -0.2987, -0.0678, -0.0061, -0.2977, -0.0665, -0.0083, -0.2977, -0.0665, -0.01, -0.3004, -0.0702, -0.01, -0.3008, -0.0678, -0.01, -0.2996, -0.069, -0.01, -0.2988, -0.0681,
-0.01, -0.3008, -0.0656, -0.01, -0.3008, -0.0635, -0.01, -0.2976, -0.0663, -0.01, -0.2986, -0.0635, -0.0061, -0.297, -0.0654, -0.0061, -0.2959, -0.0637, -0.0083, -0.297, -0.0654, -0.0083, -0.2959, -0.0637, -0.0061, -0.2952, -0.0627, -0.0083, -0.2952, -0.0627, -0.01, -0.2969, -0.0653, -0.01, -0.2959, -0.0637, -0.01, -0.2952, -0.0626, -0.0061, -0.3017, -0.0539, -0.0083, -0.3017, -0.0539, -0.01, -0.3008, -0.0613, -0.01, -0.3008, -0.0591, -0.01, -0.2986, -0.0613, -0.01, -0.2986, -0.0591, -0.01, -0.3008, -0.057, -0.01, -0.3019, -0.054, -0.01, -0.2986, -0.057, -0.01, -0.2986, -0.0548, -0.0061, -0.2943, -0.0613, -0.0061, -0.2934, -0.0599, -0.0083, -0.2943, -0.0613, -0.0083, -0.2934, -0.0599, -0.01, -0.2964, -0.0613, -0.01, -0.2964, -0.0591, -0.01, -0.2945, -0.0615,
-0.01, -0.2933, -0.0597, -0.01, -0.2964, -0.057, -0.01, -0.2964, -0.0548, -0.01, -0.2943, -0.057, -0.01, -0.2943, -0.0548, -0.0061, -0.3009, -0.0526, -0.0061, -0.3, -0.0511, -0.0083, -0.3009, -0.0526, -0.0083, -0.3, -0.0511, -0.0061, -0.2993, -0.05, -0.0083, -0.2993, -0.05, -0.0061, -0.2983, -0.0483, -0.0061, -0.2976, -0.047, -0.0083, -0.2983, -0.0483, -0.0083, -0.2976, -0.047, -0.01, -0.3007, -0.0524, -0.01, -0.3, -0.0512, -0.01, -0.2986, -0.0526, -0.01, -0.2993, -0.0501, -0.01, -0.2982, -0.0481, -0.01, -0.2976, -0.047, -0.0061, -0.2971, -0.0459, -0.0083, -0.2971, -0.0459, -0.01, -0.2964, -0.0526, -0.01, -0.2964, -0.0505, -0.01, -0.2943, -0.0526, -0.01, -0.2943, -0.0505, -0.01, -0.2964, -0.0483, -0.01, -0.2971, -0.0459, -0.01, -0.2943, -0.0483,
-0.01, -0.2943, -0.0461, -0.0061, -0.2962, -0.044, -0.0061, -0.2955, -0.0425, -0.0083, -0.2962, -0.044, -0.0083, -0.2955, -0.0425, -0.0061, -0.2951, -0.0414, -0.0083, -0.2951, -0.0414, -0.0061, -0.2943, -0.0396, -0.0061, -0.2935, -0.0375, -0.0083, -0.2943, -0.0396, -0.0083, -0.2935, -0.0375, -0.01, -0.2961, -0.0438, -0.01, -0.2956, -0.0426, -0.01, -0.2943, -0.0439, -0.01, -0.2951, -0.0415, -0.01, -0.2943, -0.0394, -0.01, -0.2936, -0.0378, -0.0061, -0.2927, -0.0588, -0.0083, -0.2927, -0.0588, -0.0061, -0.2916, -0.057, -0.0083, -0.2916, -0.057, -0.0061, -0.291, -0.0559, -0.0061, -0.2904, -0.0548, -0.0083, -0.291, -0.0559, -0.0083, -0.2904, -0.0548, -0.01, -0.2927, -0.0587, -0.01, -0.2917, -0.057, -0.01, -0.2921, -0.0548, -0.01, -0.291, -0.0559, -0.01, -0.2904, -0.0549,
-0.0061, -0.2893, -0.0527, -0.0083, -0.2893, -0.0527, -0.01, -0.2921, -0.0526, -0.01, -0.2921, -0.0505, -0.01, -0.2893, -0.0527, -0.01, -0.2899, -0.0505, -0.01, -0.2921, -0.0483, -0.01, -0.2921, -0.0461, -0.01, -0.2899, -0.0483, -0.01, -0.2899, -0.0461, -0.0061, -0.2888, -0.0516, -0.0061, -0.2883, -0.0505, -0.0083, -0.2888, -0.0516, -0.0083, -0.2883, -0.0505, -0.0061, -0.2874, -0.0483, -0.0061, -0.2868, -0.0469, -0.0083, -0.2874, -0.0483, -0.0083, -0.2868, -0.0469, -0.0061, -0.2863, -0.0458, -0.0083, -0.2863, -0.0458, -0.01, -0.2888, -0.0516, -0.01, -0.2884, -0.0506, -0.01, -0.2875, -0.0486, -0.01, -0.2867, -0.0466, -0.01, -0.2862, -0.0457, -0.0061, -0.2931, -0.0364, -0.0083, -0.2931, -0.0364, -0.01, -0.2921, -0.0439, -0.01, -0.2921, -0.0418, -0.01, -0.2899, -0.0439,
-0.01, -0.2899, -0.0418, -0.01, -0.2921, -0.0396, -0.01, -0.2933, -0.0369, -0.01, -0.2899, -0.0396, -0.01, -0.2899, -0.0374, -0.0061, -0.2855, -0.044, -0.0061, -0.2848, -0.0419, -0.0083, -0.2855, -0.044, -0.0083, -0.2848, -0.0419, -0.01, -0.2878, -0.0439, -0.01, -0.2878, -0.0418, -0.01, -0.2856, -0.0441, -0.01, -0.2848, -0.0419, -0.01, -0.2878, -0.0396, -0.01, -0.2878, -0.0374, -0.01, -0.2856, -0.0396, -0.01, -0.2856, -0.0374, -0.0061, -0.2845, -0.0408, -0.0083, -0.2845, -0.0408, -0.0061, -0.2841, -0.0396, -0.0061, -0.2834, -0.0374, -0.0083, -0.2841, -0.0396, -0.0083, -0.2834, -0.0374, -0.01, -0.2844, -0.0408, -0.01, -0.2841, -0.0397, -0.01, -0.2835, -0.0376, -0.0061, -0.2927, -0.0353, -0.0061, -0.292, -0.0331, -0.0083, -0.2927, -0.0353, -0.0083, -0.292, -0.0331,
-0.0061, -0.2914, -0.0309, -0.0061, -0.2911, -0.0297, -0.0083, -0.2914, -0.0309, -0.0083, -0.2911, -0.0297, -0.0061, -0.2908, -0.0286, -0.0083, -0.2908, -0.0286, -0.01, -0.2926, -0.035, -0.01, -0.292, -0.033, -0.01, -0.2899, -0.0353, -0.01, -0.2899, -0.0331, -0.01, -0.2914, -0.0309, -0.01, -0.2911, -0.0297, -0.01, -0.2899, -0.0309, -0.01, -0.2908, -0.0287, -0.01, -0.2878, -0.0353, -0.01, -0.2878, -0.0331, -0.01, -0.2856, -0.0353, -0.01, -0.2856, -0.0331, -0.01, -0.2878, -0.0309, -0.01, -0.2878, -0.0288, -0.01, -0.2856, -0.0309, -0.01, -0.2856, -0.0288, -0.0061, -0.2904, -0.0266, -0.0061, -0.2899, -0.0244, -0.0083, -0.2904, -0.0266, -0.0083, -0.2899, -0.0244, -0.0061, -0.2894, -0.0222, -0.0061, -0.289, -0.0204, -0.0083, -0.2894, -0.0222, -0.0083, -0.289, -0.0204,
-0.01, -0.2903, -0.0264, -0.01, -0.2899, -0.0243, -0.01, -0.2894, -0.0222, -0.01, -0.2891, -0.0206, -0.0061, -0.2888, -0.0193, -0.0083, -0.2888, -0.0193, -0.01, -0.2878, -0.0266, -0.01, -0.2878, -0.0244, -0.01, -0.2856, -0.0266, -0.01, -0.2856, -0.0244, -0.01, -0.2878, -0.0222, -0.01, -0.2889, -0.0198, -0.01, -0.2856, -0.0222, -0.01, -0.2856, -0.0201, -0.0061, -0.2828, -0.0353, -0.0061, -0.2824, -0.0338, -0.0083, -0.2828, -0.0353, -0.0083, -0.2824, -0.0338, -0.0061, -0.2821, -0.0328, -0.0083, -0.2821, -0.0328, -0.0061, -0.2816, -0.0309, -0.0061, -0.2812, -0.0288, -0.0083, -0.2816, -0.0309, -0.0083, -0.2812, -0.0288, -0.01, -0.2829, -0.0355, -0.01, -0.2823, -0.0334, -0.01, -0.2821, -0.0325, -0.01, -0.2834, -0.0309, -0.01, -0.2834, -0.0288, -0.01, -0.2817, -0.031,
-0.01, -0.2812, -0.0289, -0.0061, -0.2808, -0.0266, -0.0061, -0.2804, -0.0244, -0.0083, -0.2808, -0.0266, -0.0083, -0.2804, -0.0244, -0.0061, -0.2802, -0.0231, -0.0083, -0.2802, -0.0231, -0.01, -0.2834, -0.0266, -0.01, -0.2834, -0.0244, -0.01, -0.2808, -0.0267, -0.01, -0.2804, -0.0246, -0.01, -0.2834, -0.0222, -0.01, -0.2834, -0.0201, -0.01, -0.2801, -0.0225, -0.01, -0.2812, -0.0201, -0.0061, -0.28, -0.022, -0.0061, -0.2797, -0.0201, -0.0083, -0.28, -0.022, -0.0083, -0.2797, -0.0201, -0.01, -0.28, -0.0218, -0.01, -0.2797, -0.0201, -0.0061, -0.2885, -0.0179, -0.0061, -0.2883, -0.0157, -0.0083, -0.2885, -0.0179, -0.0083, -0.2883, -0.0157, -0.0061, -0.2882, -0.0136, -0.0061, -0.2882, -0.0114, -0.0083, -0.2882, -0.0136, -0.0083, -0.2882, -0.0114, -0.01, -0.2885, -0.0178,
-0.01, -0.2883, -0.0157, -0.01, -0.2856, -0.0179, -0.01, -0.2856, -0.0157, -0.01, -0.2882, -0.0136, -0.01, -0.2882, -0.0114, -0.01, -0.2856, -0.0136, -0.01, -0.2856, -0.0114, -0.0061, -0.2882, -0.0092, -0.0061, -0.2882, -0.0071, -0.0083, -0.2882, -0.0092, -0.0083, -0.2882, -0.0071, -0.0061, -0.2882, -0.0049, -0.0061, -0.2882, -0.0027, -0.0083, -0.2882, -0.0049, -0.0083, -0.2882, -0.0027, -0.01, -0.2882, -0.0092, -0.01, -0.2882, -0.0071, -0.01, -0.2856, -0.0092, -0.01, -0.2856, -0.0071, -0.01, -0.2882, -0.0049, -0.01, -0.2882, -0.0027, -0.01, -0.2856, -0.0049, -0.01, -0.2856, -0.0027, -0.01, -0.2834, -0.0179, -0.01, -0.2834, -0.0157, -0.01, -0.2812, -0.0179, -0.01, -0.2812, -0.0157, -0.01, -0.2834, -0.0136, -0.01, -0.2834, -0.0114, -0.01, -0.2812, -0.0136,
-0.01, -0.2812, -0.0114, -0.0061, -0.2795, -0.0179, -0.0061, -0.2794, -0.0157, -0.0083, -0.2795, -0.0179, -0.0083, -0.2794, -0.0157, -0.0061, -0.2794, -0.0136, -0.0061, -0.2794, -0.0114, -0.0083, -0.2794, -0.0136, -0.0083, -0.2794, -0.0114, -0.01, -0.2795, -0.0179, -0.01, -0.2794, -0.0157, -0.01, -0.2794, -0.0136, -0.01, -0.2794, -0.0114, -0.01, -0.2834, -0.0092, -0.01, -0.2834, -0.0071, -0.01, -0.2812, -0.0092, -0.01, -0.2812, -0.0071, -0.01, -0.2834, -0.0049, -0.01, -0.2834, -0.0027, -0.01, -0.2812, -0.0049, -0.01, -0.2812, -0.0027, -0.0061, -0.2794, -0.0092, -0.0061, -0.2794, -0.0071, -0.0083, -0.2794, -0.0092, -0.0083, -0.2794, -0.0071, -0.0061, -0.2794, -0.0049, -0.0061, -0.2794, -0.0027, -0.0083, -0.2794, -0.0049, -0.0083, -0.2794, -0.0027, -0.01, -0.2794, -0.0092,
-0.01, -0.2794, -0.0071, -0.01, -0.2794, -0.0049, -0.01, -0.2794, -0.0027, -0.0061, -0.2196, -0.067, -0.0083, -0.2196, -0.067, -0.0061, -0.2195, -0.0666, -0.0083, -0.2195, -0.0666, -0.01, -0.2196, -0.067, -0.01, -0.2195, -0.0666, -0.0061, -0.2183, -0.067, -0.0083, -0.2183, -0.067, -0.0061, -0.2172, -0.0667, -0.0083, -0.2172, -0.0667, -0.0061, -0.219, -0.0655, -0.0061, -0.2181, -0.0635, -0.0083, -0.219, -0.0655, -0.0083, -0.2181, -0.0635, -0.0061, -0.216, -0.0665, -0.0083, -0.216, -0.0665, -0.01, -0.2183, -0.067, -0.01, -0.2172, -0.0667, -0.01, -0.219, -0.0655, -0.01, -0.218, -0.0633, -0.01, -0.216, -0.0665, -0.01, -0.2161, -0.0635, -0.0061, -0.2174, -0.0621, -0.0083, -0.2174, -0.0621, -0.0061, -0.2168, -0.061, -0.0061, -0.2157, -0.0592, -0.0083, -0.2168, -0.061,
-0.0083, -0.2157, -0.0592, -0.01, -0.2174, -0.0622, -0.01, -0.2168, -0.0611, -0.01, -0.2158, -0.0593, -0.0061, -0.2141, -0.0652, -0.0061, -0.2132, -0.0642, -0.0083, -0.2141, -0.0652, -0.0083, -0.2132, -0.0642, -0.0061, -0.2124, -0.0631, -0.0083, -0.2124, -0.0631, -0.01, -0.2142, -0.0654, -0.01, -0.2131, -0.0641, -0.01, -0.2124, -0.063, -0.0061, -0.215, -0.0581, -0.0083, -0.215, -0.0581, -0.0061, -0.2113, -0.0614, -0.0083, -0.2113, -0.0614, -0.0061, -0.2144, -0.057, -0.0061, -0.2133, -0.0552, -0.0083, -0.2144, -0.057, -0.0083, -0.2133, -0.0552, -0.0061, -0.2126, -0.0541, -0.0083, -0.2126, -0.0541, -0.01, -0.214, -0.0613, -0.01, -0.2152, -0.0583, -0.01, -0.2114, -0.0614, -0.01, -0.2118, -0.0591, -0.01, -0.2142, -0.0566, -0.01, -0.2134, -0.0553, -0.01, -0.2118, -0.057,
-0.01, -0.2127, -0.0542, -0.0061, -0.2107, -0.0603, -0.0061, -0.21, -0.0591, -0.0083, -0.2107, -0.0603, -0.0083, -0.21, -0.0591, -0.0061, -0.2089, -0.0574, -0.0083, -0.2089, -0.0574, -0.0061, -0.2083, -0.0563, -0.0061, -0.2074, -0.0548, -0.0083, -0.2083, -0.0563, -0.0083, -0.2074, -0.0548, -0.01, -0.2107, -0.0603, -0.01, -0.2101, -0.0593, -0.01, -0.2089, -0.0573, -0.01, -0.2096, -0.0548, -0.01, -0.2083, -0.0562, -0.01, -0.2075, -0.0551, -0.0061, -0.2117, -0.0526, -0.0061, -0.2109, -0.0513, -0.0083, -0.2117, -0.0526, -0.0083, -0.2109, -0.0513, -0.01, -0.2115, -0.0524, -0.01, -0.2109, -0.0513, -0.0061, -0.2102, -0.0502, -0.0083, -0.2102, -0.0502, -0.0061, -0.2066, -0.0534, -0.0083, -0.2066, -0.0534, -0.0061, -0.2091, -0.0484, -0.0083, -0.2091, -0.0484, -0.0061, -0.2085, -0.0473,
-0.0061, -0.2076, -0.0461, -0.0083, -0.2085, -0.0473, -0.0083, -0.2076, -0.0461, -0.01, -0.2096, -0.0526, -0.01, -0.2102, -0.0502, -0.01, -0.2065, -0.0532, -0.01, -0.2075, -0.0505, -0.01, -0.2092, -0.0486, -0.01, -0.2086, -0.0475, -0.01, -0.2074, -0.0458, -0.0061, -0.2066, -0.0448, -0.0083, -0.2066, -0.0448, -0.01, -0.2066, -0.0448, -0.0061, -0.2196, -0.0219, -0.0061, -0.2196, -0.0201, -0.0083, -0.2196, -0.0219, -0.0083, -0.2196, -0.0201, -0.01, -0.2196, -0.0219, -0.01, -0.2196, -0.0201, -0.0061, -0.2183, -0.0219, -0.0083, -0.2183, -0.0219, -0.0061, -0.2161, -0.0219, -0.0083, -0.2161, -0.0219, -0.01, -0.2183, -0.0219, -0.01, -0.2183, -0.0201, -0.01, -0.2161, -0.0219, -0.01, -0.2161, -0.0201, -0.0061, -0.214, -0.0219, -0.0083, -0.214, -0.0219, -0.0061, -0.2118, -0.0219,
-0.0083, -0.2118, -0.0219, -0.01, -0.214, -0.0219, -0.01, -0.214, -0.0201, -0.01, -0.2118, -0.0219, -0.01, -0.2118, -0.0201, -0.0061, -0.2088, -0.0219, -0.0061, -0.2088, -0.0201, -0.0083, -0.2088, -0.0219, -0.0083, -0.2088, -0.0201, -0.01, -0.2088, -0.0219, -0.01, -0.2088, -0.0201, -0.0061, -0.2196, -0.0179, -0.0061, -0.2196, -0.0157, -0.0083, -0.2196, -0.0179, -0.0083, -0.2196, -0.0157, -0.0061, -0.2196, -0.0136, -0.0061, -0.2196, -0.0114, -0.0083, -0.2196, -0.0136, -0.0083, -0.2196, -0.0114, -0.01, -0.2196, -0.0179, -0.01, -0.2196, -0.0157, -0.01, -0.2196, -0.0136, -0.01, -0.2196, -0.0114, -0.01, -0.2183, -0.0179, -0.01, -0.2183, -0.0157, -0.01, -0.2161, -0.0179, -0.01, -0.2161, -0.0157, -0.01, -0.2183, -0.0136, -0.01, -0.2183, -0.0114, -0.01, -0.2161, -0.0136,
-0.01, -0.2161, -0.0114, -0.0061, -0.2196, -0.0092, -0.0061, -0.2196, -0.0071, -0.0083, -0.2196, -0.0092, -0.0083, -0.2196, -0.0071, -0.0061, -0.2196, -0.0049, -0.0061, -0.2196, -0.0027, -0.0083, -0.2196, -0.0049, -0.0083, -0.2196, -0.0027, -0.01, -0.2196, -0.0092, -0.01, -0.2196, -0.0071, -0.01, -0.2196, -0.0049, -0.01, -0.2196, -0.0027, -0.01, -0.2183, -0.0092, -0.01, -0.2183, -0.0071, -0.01, -0.2161, -0.0092, -0.01, -0.2161, -0.0071, -0.01, -0.2183, -0.0049, -0.01, -0.2183, -0.0027, -0.01, -0.2161, -0.0049, -0.01, -0.2161, -0.0027, -0.01, -0.214, -0.0179, -0.01, -0.214, -0.0157, -0.01, -0.2118, -0.0179, -0.01, -0.2118, -0.0157, -0.01, -0.214, -0.0136, -0.01, -0.214, -0.0114, -0.01, -0.2118, -0.0136, -0.01, -0.2118, -0.0114, -0.0061, -0.2088, -0.0179,
-0.0061, -0.2088, -0.0157, -0.0083, -0.2088, -0.0179, -0.0083, -0.2088, -0.0157, -0.0061, -0.2088, -0.0136, -0.0061, -0.2088, -0.0114, -0.0083, -0.2088, -0.0136, -0.0083, -0.2088, -0.0114, -0.01, -0.2088, -0.0179, -0.01, -0.2088, -0.0157, -0.01, -0.2088, -0.0136, -0.01, -0.2088, -0.0114, -0.01, -0.214, -0.0092, -0.01, -0.214, -0.0071, -0.01, -0.2118, -0.0092, -0.01, -0.2118, -0.0071, -0.01, -0.214, -0.0049, -0.01, -0.214, -0.0027, -0.01, -0.2118, -0.0049, -0.01, -0.2118, -0.0027, -0.0061, -0.2088, -0.0092, -0.0061, -0.2088, -0.0071, -0.0083, -0.2088, -0.0092, -0.0083, -0.2088, -0.0071, -0.0061, -0.2088, -0.0049, -0.0061, -0.2088, -0.0027, -0.0083, -0.2088, -0.0049, -0.0083, -0.2088, -0.0027, -0.01, -0.2088, -0.0092, -0.01, -0.2088, -0.0071, -0.01, -0.2088, -0.0049,
-0.01, -0.2088, -0.0027, -0.0061, -0.2069, -0.0591, -0.0083, -0.2069, -0.0591, -0.0061, -0.2031, -0.0591, -0.0083, -0.2031, -0.0591, -0.0061, -0.206, -0.0573, -0.0083, -0.206, -0.0573, -0.0061, -0.2032, -0.0563, -0.0083, -0.2032, -0.0563, -0.01, -0.2069, -0.0591, -0.01, -0.2031, -0.0591, -0.01, -0.206, -0.0573, -0.01, -0.2033, -0.0563, -0.0061, -0.2009, -0.0591, -0.0083, -0.2009, -0.0591, -0.0061, -0.1988, -0.0591, -0.0083, -0.1988, -0.0591, -0.0061, -0.2009, -0.0562, -0.0061, -0.2, -0.054, -0.0083, -0.2009, -0.0562, -0.0083, -0.2, -0.054, -0.0061, -0.199, -0.0562, -0.0061, -0.1995, -0.0551, -0.0083, -0.199, -0.0562, -0.0083, -0.1995, -0.0551, -0.01, -0.2009, -0.0591, -0.01, -0.1988, -0.0591, -0.01, -0.201, -0.0562, -0.01, -0.2001, -0.054, -0.01, -0.199, -0.0562,
-0.01, -0.1995, -0.055, -0.0061, -0.1941, -0.0635, -0.0083, -0.1941, -0.0635, -0.01, -0.1942, -0.0634, -0.0061, -0.1894, -0.0668, -0.0083, -0.1894, -0.0668, -0.0061, -0.1924, -0.0656, -0.0061, -0.1933, -0.0645, -0.0083, -0.1924, -0.0656, -0.0083, -0.1933, -0.0645, -0.0061, -0.1905, -0.0666, -0.0061, -0.1892, -0.0628, -0.0083, -0.1905, -0.0666, -0.0083, -0.1892, -0.0628, -0.01, -0.1893, -0.0668, -0.01, -0.1923, -0.0656, -0.01, -0.1934, -0.0645, -0.01, -0.1904, -0.0667, -0.01, -0.1891, -0.063, -0.0061, -0.1956, -0.0605, -0.0061, -0.1961, -0.0591, -0.0083, -0.1956, -0.0605, -0.0083, -0.1961, -0.0591, -0.0061, -0.1952, -0.0616, -0.0083, -0.1952, -0.0616, -0.0061, -0.1939, -0.0546, -0.0083, -0.1939, -0.0546, -0.01, -0.1957, -0.0605, -0.01, -0.1961, -0.0591, -0.01, -0.1952, -0.0616,
-0.01, -0.1944, -0.0591, -0.01, -0.1966, -0.057, -0.01, -0.1966, -0.0548, -0.01, -0.1944, -0.057, -0.01, -0.1939, -0.0546, -0.0061, -0.1916, -0.0587, -0.0083, -0.1916, -0.0587, -0.0061, -0.1901, -0.0613, -0.0061, -0.1909, -0.0598, -0.0083, -0.1901, -0.0613, -0.0083, -0.1909, -0.0598, -0.0061, -0.1926, -0.057, -0.0061, -0.1933, -0.0557, -0.0083, -0.1926, -0.057, -0.0083, -0.1933, -0.0557, -0.01, -0.1923, -0.0613, -0.01, -0.1915, -0.0588, -0.01, -0.1902, -0.0611, -0.01, -0.1909, -0.0599, -0.01, -0.1927, -0.0568, -0.01, -0.1933, -0.0558, -0.0061, -0.2059, -0.0523, -0.0061, -0.2049, -0.0505, -0.0083, -0.2059, -0.0523, -0.0083, -0.2049, -0.0505, -0.0061, -0.2026, -0.0501, -0.0083, -0.2026, -0.0501, -0.0061, -0.2043, -0.0492, -0.0083, -0.2043, -0.0492, -0.0061, -0.2039, -0.0484,
-0.0083, -0.2039, -0.0484, -0.01, -0.2059, -0.0522, -0.01, -0.205, -0.0508, -0.01, -0.2026, -0.05, -0.01, -0.2041, -0.0489, -0.01, -0.2053, -0.0461, -0.01, -0.2039, -0.0484, -0.01, -0.2031, -0.0461, -0.0061, -0.2008, -0.0526, -0.0061, -0.2018, -0.0512, -0.0083, -0.2008, -0.0526, -0.0083, -0.2018, -0.0512, -0.0061, -0.1979, -0.0476, -0.0061, -0.1989, -0.046, -0.0083, -0.1979, -0.0476, -0.0083, -0.1989, -0.046, -0.01, -0.2007, -0.0528, -0.01, -0.2019, -0.0511, -0.01, -0.1988, -0.0526, -0.01, -0.1988, -0.0505, -0.01, -0.2009, -0.0483, -0.01, -0.2009, -0.0461, -0.01, -0.1978, -0.0477, -0.01, -0.199, -0.0458, -0.0061, -0.2052, -0.0436, -0.0083, -0.2052, -0.0436, -0.0061, -0.2026, -0.043, -0.0083, -0.2026, -0.043, -0.01, -0.2052, -0.0436, -0.01, -0.2026, -0.043,
-0.0061, -0.2006, -0.0441, -0.0083, -0.2006, -0.0441, -0.0061, -0.1998, -0.045, -0.0083, -0.1998, -0.045, -0.01, -0.2006, -0.0441, -0.01, -0.1998, -0.045, -0.0061, -0.1956, -0.0517, -0.0061, -0.1963, -0.0505, -0.0083, -0.1956, -0.0517, -0.0083, -0.1963, -0.0505, -0.0061, -0.195, -0.0528, -0.0083, -0.195, -0.0528, -0.0061, -0.1973, -0.0487, -0.0083, -0.1973, -0.0487, -0.01, -0.1955, -0.0519, -0.01, -0.1965, -0.0502, -0.01, -0.1949, -0.0529, -0.01, -0.1972, -0.0488, -0.0061, -0.1873, -0.067, -0.0083, -0.1873, -0.067, -0.0061, -0.1878, -0.0656, -0.0061, -0.1886, -0.0639, -0.0083, -0.1878, -0.0656, -0.0083, -0.1886, -0.0639, -0.01, -0.1873, -0.067, -0.01, -0.1879, -0.0654, -0.01, -0.1886, -0.064, -0.0061, -0.1686, -0.067, -0.0083, -0.1686, -0.067, -0.0061, -0.1686, -0.0656,
-0.0061, -0.1686, -0.0635, -0.0083, -0.1686, -0.0656, -0.0083, -0.1686, -0.0635, -0.01, -0.1686, -0.067, -0.01, -0.1686, -0.0656, -0.01, -0.1686, -0.0635, -0.0061, -0.1662, -0.067, -0.0083, -0.1662, -0.067, -0.0061, -0.1641, -0.067, -0.0083, -0.1641, -0.067, -0.0061, -0.1647, -0.0631, -0.0083, -0.1647, -0.0631, -0.01, -0.1662, -0.067, -0.01, -0.1641, -0.067, -0.01, -0.1662, -0.0656, -0.01, -0.1662, -0.0635, -0.01, -0.1641, -0.0656, -0.01, -0.1647, -0.0631, -0.0061, -0.1686, -0.0613, -0.0061, -0.1686, -0.0591, -0.0083, -0.1686, -0.0613, -0.0083, -0.1686, -0.0591, -0.0061, -0.1686, -0.057, -0.0061, -0.1686, -0.0548, -0.0083, -0.1686, -0.057, -0.0083, -0.1686, -0.0548, -0.01, -0.1686, -0.0613, -0.01, -0.1686, -0.0591, -0.01, -0.1686, -0.057, -0.01, -0.1686, -0.0548,
-0.0061, -0.1647, -0.0613, -0.0061, -0.1647, -0.0591, -0.0083, -0.1647, -0.0613, -0.0083, -0.1647, -0.0591, -0.0061, -0.1647, -0.057, -0.0061, -0.1647, -0.0548, -0.0083, -0.1647, -0.057, -0.0083, -0.1647, -0.0548, -0.01, -0.1662, -0.0613, -0.01, -0.1662, -0.0591, -0.01, -0.1647, -0.0613, -0.01, -0.1647, -0.0591, -0.01, -0.1662, -0.057, -0.01, -0.1662, -0.0548, -0.01, -0.1647, -0.057, -0.01, -0.1647, -0.0548, -0.0061, -0.1619, -0.067, -0.0083, -0.1619, -0.067, -0.0061, -0.1597, -0.067, -0.0083, -0.1597, -0.067, -0.0061, -0.1619, -0.0631, -0.0083, -0.1619, -0.0631, -0.0061, -0.1597, -0.0631, -0.0083, -0.1597, -0.0631, -0.01, -0.1619, -0.067, -0.01, -0.1597, -0.067, -0.01, -0.1619, -0.0656, -0.01, -0.1619, -0.0631, -0.01, -0.1597, -0.0656, -0.01, -0.1597, -0.0631,
-0.0061, -0.1575, -0.067, -0.0083, -0.1575, -0.067, -0.0061, -0.1554, -0.067, -0.0083, -0.1554, -0.067, -0.0061, -0.1575, -0.0631, -0.0083, -0.1575, -0.0631, -0.0061, -0.1554, -0.063, -0.0083, -0.1554, -0.063, -0.01, -0.1575, -0.067, -0.01, -0.1554, -0.067, -0.01, -0.1575, -0.0656, -0.01, -0.1576, -0.0631, -0.01, -0.1554, -0.0656, -0.01, -0.1554, -0.063, -0.0061, -0.1686, -0.0526, -0.0061, -0.1686, -0.0505, -0.0083, -0.1686, -0.0526, -0.0083, -0.1686, -0.0505, -0.0061, -0.1686, -0.0483, -0.0061, -0.1686, -0.0461, -0.0083, -0.1686, -0.0483, -0.0083, -0.1686, -0.0461, -0.01, -0.1686, -0.0526, -0.01, -0.1686, -0.0505, -0.01, -0.1686, -0.0483, -0.01, -0.1686, -0.0461, -0.0061, -0.1647, -0.0526, -0.0061, -0.1647, -0.0505, -0.0083, -0.1647, -0.0526, -0.0083, -0.1647, -0.0505,
-0.0061, -0.1647, -0.0483, -0.0061, -0.1647, -0.0454, -0.0083, -0.1647, -0.0483, -0.0083, -0.1647, -0.0454, -0.01, -0.1662, -0.0526, -0.01, -0.1662, -0.0505, -0.01, -0.1647, -0.0526, -0.01, -0.1647, -0.0505, -0.01, -0.1662, -0.0483, -0.01, -0.1662, -0.0461, -0.01, -0.1647, -0.0483, -0.01, -0.1647, -0.0454, -0.0061, -0.1686, -0.0439, -0.0061, -0.1686, -0.0425, -0.0083, -0.1686, -0.0439, -0.0083, -0.1686, -0.0425, -0.01, -0.1686, -0.0439, -0.01, -0.1686, -0.0425, -0.0061, -0.1662, -0.0425, -0.0083, -0.1662, -0.0425, -0.0061, -0.1641, -0.0425, -0.0083, -0.1641, -0.0425, -0.01, -0.1662, -0.0439, -0.01, -0.1662, -0.0425, -0.01, -0.1641, -0.0439, -0.01, -0.1641, -0.0425, -0.0061, -0.1619, -0.0454, -0.0083, -0.1619, -0.0454, -0.0061, -0.1597, -0.0454, -0.0083, -0.1597, -0.0454,
-0.01, -0.1619, -0.0454, -0.01, -0.1597, -0.0454, -0.0061, -0.1575, -0.0455, -0.0083, -0.1575, -0.0455, -0.0061, -0.1554, -0.0457, -0.0083, -0.1554, -0.0457, -0.01, -0.1576, -0.0455, -0.01, -0.1554, -0.0456, -0.0061, -0.1619, -0.0425, -0.0083, -0.1619, -0.0425, -0.0061, -0.1597, -0.0425, -0.0083, -0.1597, -0.0425, -0.01, -0.1619, -0.0439, -0.01, -0.1619, -0.0425, -0.01, -0.1597, -0.0439, -0.01, -0.1597, -0.0425, -0.0061, -0.1575, -0.0425, -0.0083, -0.1575, -0.0425, -0.0061, -0.1554, -0.0425, -0.0083, -0.1554, -0.0425, -0.01, -0.1575, -0.0439, -0.01, -0.1575, -0.0425, -0.01, -0.1554, -0.0439, -0.01, -0.1554, -0.0425, -0.0061, -0.1535, -0.0669, -0.0083, -0.1535, -0.0669, -0.0061, -0.1524, -0.0667, -0.0061, -0.1532, -0.0629, -0.0083, -0.1524, -0.0667, -0.0083, -0.1532, -0.0629,
-0.0061, -0.151, -0.0665, -0.0061, -0.1516, -0.0625, -0.0083, -0.151, -0.0665, -0.0083, -0.1516, -0.0625, -0.01, -0.1537, -0.0669, -0.01, -0.1529, -0.0668, -0.01, -0.1533, -0.0629, -0.01, -0.1507, -0.0664, -0.01, -0.1513, -0.0624, -0.0061, -0.1488, -0.0656, -0.0083, -0.1488, -0.0656, -0.0061, -0.1475, -0.0648, -0.0061, -0.1463, -0.064, -0.0083, -0.1475, -0.0648, -0.0083, -0.1463, -0.064, -0.01, -0.1486, -0.0655, -0.01, -0.1489, -0.0635, -0.01, -0.1475, -0.0648, -0.01, -0.1463, -0.064, -0.0061, -0.1504, -0.0621, -0.0083, -0.1504, -0.0621, -0.01, -0.1504, -0.0621, -0.0061, -0.1489, -0.0611, -0.0061, -0.148, -0.06, -0.0083, -0.1489, -0.0611, -0.0083, -0.148, -0.06, -0.0061, -0.1472, -0.059, -0.0083, -0.1472, -0.059, -0.0061, -0.1465, -0.057, -0.0061, -0.1462, -0.0548,
-0.0083, -0.1465, -0.057, -0.0083, -0.1462, -0.0548, -0.01, -0.149, -0.0612, -0.01, -0.148, -0.06, -0.01, -0.1467, -0.0613, -0.01, -0.1472, -0.059, -0.01, -0.1464, -0.0569, -0.01, -0.1462, -0.0548, -0.0061, -0.1452, -0.0628, -0.0083, -0.1452, -0.0628, -0.01, -0.1452, -0.0628, -0.0061, -0.144, -0.0614, -0.0083, -0.144, -0.0614, -0.0061, -0.1434, -0.0603, -0.0061, -0.1428, -0.0592, -0.0083, -0.1434, -0.0603, -0.0083, -0.1428, -0.0592, -0.0061, -0.1423, -0.057, -0.0061, -0.1422, -0.0548, -0.0083, -0.1423, -0.057, -0.0083, -0.1422, -0.0548, -0.01, -0.1441, -0.0615, -0.01, -0.1445, -0.0591, -0.01, -0.1434, -0.0603, -0.01, -0.1429, -0.0593, -0.01, -0.1445, -0.057, -0.01, -0.1445, -0.0548, -0.01, -0.1423, -0.0571, -0.01, -0.1422, -0.0548, -0.0061, -0.1532, -0.0461,
-0.0083, -0.1532, -0.0461, -0.0061, -0.1503, -0.0474, -0.0061, -0.1514, -0.0468, -0.0083, -0.1503, -0.0474, -0.0083, -0.1514, -0.0468, -0.01, -0.1533, -0.0461, -0.01, -0.1502, -0.0474, -0.01, -0.1512, -0.0469, -0.0061, -0.1479, -0.0495, -0.0083, -0.1479, -0.0495, -0.0061, -0.1464, -0.0526, -0.0061, -0.1469, -0.0509, -0.0083, -0.1464, -0.0526, -0.0083, -0.1469, -0.0509, -0.0061, -0.1489, -0.0484, -0.0083, -0.1489, -0.0484, -0.0061, -0.1458, -0.0452, -0.0083, -0.1458, -0.0452, -0.01, -0.1479, -0.0495, -0.01, -0.1464, -0.0527, -0.01, -0.1471, -0.0505, -0.01, -0.1489, -0.0484, -0.01, -0.1489, -0.0461, -0.01, -0.1467, -0.0483, -0.01, -0.1458, -0.0451, -0.0061, -0.1532, -0.0425, -0.0083, -0.1532, -0.0425, -0.0061, -0.1501, -0.0429, -0.0061, -0.1511, -0.0426, -0.0083, -0.1501, -0.0429,
-0.0083, -0.1511, -0.0426, -0.01, -0.1532, -0.0439, -0.01, -0.1532, -0.0425, -0.01, -0.1506, -0.0428, -0.01, -0.1515, -0.0426, -0.0061, -0.1488, -0.0432, -0.0083, -0.1488, -0.0432, -0.0061, -0.1468, -0.0443, -0.0083, -0.1468, -0.0443, -0.01, -0.1485, -0.0433, -0.01, -0.1469, -0.0443, -0.0061, -0.1422, -0.0526, -0.0061, -0.1425, -0.0504, -0.0083, -0.1422, -0.0526, -0.0083, -0.1425, -0.0504, -0.0061, -0.1437, -0.0477, -0.0061, -0.1447, -0.0462, -0.0083, -0.1437, -0.0477, -0.0083, -0.1447, -0.0462, -0.0061, -0.1431, -0.0487, -0.0083, -0.1431, -0.0487, -0.01, -0.1445, -0.0526, -0.01, -0.1445, -0.0505, -0.01, -0.1422, -0.0526, -0.01, -0.1425, -0.0503, -0.01, -0.1436, -0.0478, -0.01, -0.1447, -0.0462, -0.01, -0.143, -0.0489, -0.0061, -0.151, -0.0219, -0.0061, -0.151, -0.0201,
-0.0083, -0.151, -0.0219, -0.0083, -0.151, -0.0201, -0.01, -0.151, -0.0219, -0.01, -0.151, -0.0201, -0.0061, -0.1489, -0.0219, -0.0083, -0.1489, -0.0219, -0.0061, -0.1467, -0.0219, -0.0083, -0.1467, -0.0219, -0.01, -0.1489, -0.0219, -0.01, -0.1489, -0.0201, -0.01, -0.1467, -0.0219, -0.01, -0.1467, -0.0201, -0.0061, -0.1445, -0.0219, -0.0083, -0.1445, -0.0219, -0.0061, -0.1424, -0.0219, -0.0083, -0.1424, -0.0219, -0.01, -0.1445, -0.0219, -0.01, -0.1445, -0.0201, -0.01, -0.1424, -0.0219, -0.01, -0.1424, -0.0201, -0.0061, -0.1402, -0.0219, -0.0083, -0.1402, -0.0219, -0.0061, -0.138, -0.0219, -0.0083, -0.138, -0.0219, -0.01, -0.1402, -0.0219, -0.01, -0.1402, -0.0201, -0.01, -0.138, -0.0219, -0.01, -0.138, -0.0201, -0.0061, -0.151, -0.0179, -0.0061, -0.151, -0.0157,
-0.0083, -0.151, -0.0179, -0.0083, -0.151, -0.0157, -0.0061, -0.151, -0.0136, -0.0061, -0.151, -0.0114, -0.0083, -0.151, -0.0136, -0.0083, -0.151, -0.0114, -0.01, -0.151, -0.0179, -0.01, -0.151, -0.0157, -0.01, -0.151, -0.0136, -0.01, -0.151, -0.0114, -0.01, -0.1489, -0.0179, -0.01, -0.1489, -0.0157, -0.01, -0.1467, -0.0179, -0.01, -0.1467, -0.0157, -0.01, -0.1489, -0.0136, -0.01, -0.1489, -0.0114, -0.01, -0.1467, -0.0136, -0.01, -0.1467, -0.0114, -0.0061, -0.151, -0.0092, -0.0061, -0.151, -0.0071, -0.0083, -0.151, -0.0092, -0.0083, -0.151, -0.0071, -0.0061, -0.151, -0.0049, -0.0061, -0.151, -0.0027, -0.0083, -0.151, -0.0049, -0.0083, -0.151, -0.0027, -0.01, -0.151, -0.0092, -0.01, -0.151, -0.0071, -0.01, -0.151, -0.0049, -0.01, -0.151, -0.0027,
-0.01, -0.1489, -0.0092, -0.01, -0.1489, -0.0071, -0.01, -0.1467, -0.0092, -0.01, -0.1467, -0.0071, -0.01, -0.1489, -0.0049, -0.01, -0.1489, -0.0027, -0.01, -0.1467, -0.0049, -0.01, -0.1467, -0.0027, -0.01, -0.1445, -0.0179, -0.01, -0.1445, -0.0157, -0.01, -0.1424, -0.0179, -0.01, -0.1424, -0.0157, -0.01, -0.1445, -0.0136, -0.01, -0.1445, -0.0114, -0.01, -0.1424, -0.0136, -0.01, -0.1424, -0.0114, -0.01, -0.1402, -0.0179, -0.01, -0.1402, -0.0157, -0.01, -0.138, -0.0179, -0.01, -0.138, -0.0157, -0.01, -0.1402, -0.0136, -0.01, -0.1402, -0.0114, -0.01, -0.138, -0.0136, -0.01, -0.138, -0.0114, -0.01, -0.1445, -0.0092, -0.01, -0.1445, -0.0071, -0.01, -0.1424, -0.0092, -0.01, -0.1424, -0.0071, -0.01, -0.1445, -0.0049, -0.01, -0.1445, -0.0027,
-0.01, -0.1424, -0.0049, -0.01, -0.1424, -0.0027, -0.0061, -0.1376, -0.0086, -0.0061, -0.1387, -0.0072, -0.0083, -0.1376, -0.0086, -0.0083, -0.1387, -0.0072, -0.0061, -0.1391, -0.0043, -0.0061, -0.1392, -0.0027, -0.0083, -0.1391, -0.0043, -0.0083, -0.1392, -0.0027, -0.0061, -0.139, -0.0054, -0.0083, -0.139, -0.0054, -0.01, -0.1402, -0.0092, -0.01, -0.1402, -0.0071, -0.01, -0.1375, -0.0087, -0.01, -0.1387, -0.0072, -0.01, -0.1391, -0.0048, -0.01, -0.1392, -0.0027, -0.01, -0.139, -0.0056, -0.0061, -0.1225, -0.067, -0.0083, -0.1225, -0.067, -0.0061, -0.1207, -0.067, -0.0083, -0.1207, -0.067, -0.0061, -0.1225, -0.0656, -0.0061, -0.1225, -0.0635, -0.0083, -0.1225, -0.0656, -0.0083, -0.1225, -0.0635, -0.01, -0.1225, -0.067, -0.01, -0.1207, -0.067, -0.01, -0.1225, -0.0656,
-0.01, -0.1225, -0.0635, -0.01, -0.1207, -0.0656, -0.01, -0.1207, -0.0635, -0.0061, -0.1225, -0.0613, -0.0061, -0.1225, -0.0591, -0.0083, -0.1225, -0.0613, -0.0083, -0.1225, -0.0591, -0.0061, -0.1225, -0.057, -0.0061, -0.1225, -0.0548, -0.0083, -0.1225, -0.057, -0.0083, -0.1225, -0.0548, -0.01, -0.1225, -0.0613, -0.01, -0.1225, -0.0591, -0.01, -0.1207, -0.0613, -0.01, -0.1207, -0.0591, -0.01, -0.1225, -0.057, -0.01, -0.1225, -0.0548, -0.01, -0.1207, -0.057, -0.01, -0.1207, -0.0548, -0.0061, -0.1225, -0.0526, -0.0061, -0.1225, -0.0505, -0.0083, -0.1225, -0.0526, -0.0083, -0.1225, -0.0505, -0.0061, -0.1225, -0.0483, -0.0061, -0.1225, -0.0461, -0.0083, -0.1225, -0.0483, -0.0083, -0.1225, -0.0461, -0.01, -0.1225, -0.0526, -0.01, -0.1225, -0.0505, -0.01, -0.1207, -0.0526,
-0.01, -0.1207, -0.0505, -0.01, -0.1225, -0.0483, -0.01, -0.1225, -0.0461, -0.01, -0.1207, -0.0483, -0.01, -0.1207, -0.0461, -0.0061, -0.1225, -0.0439, -0.0061, -0.1225, -0.0425, -0.0083, -0.1225, -0.0439, -0.0083, -0.1225, -0.0425, -0.0061, -0.1207, -0.0425, -0.0083, -0.1207, -0.0425, -0.01, -0.1225, -0.0439, -0.01, -0.1225, -0.0425, -0.01, -0.1207, -0.0439, -0.01, -0.1207, -0.0425, -0.0061, -0.1185, -0.067, -0.0083, -0.1185, -0.067, -0.0061, -0.1163, -0.067, -0.0083, -0.1163, -0.067, -0.0061, -0.1186, -0.0631, -0.0083, -0.1186, -0.0631, -0.0061, -0.1163, -0.0631, -0.0083, -0.1163, -0.0631, -0.01, -0.1185, -0.067, -0.01, -0.1163, -0.067, -0.01, -0.1185, -0.0656, -0.01, -0.1186, -0.0631, -0.01, -0.1163, -0.0656, -0.01, -0.1163, -0.0631, -0.0061, -0.1141, -0.067,
-0.0083, -0.1141, -0.067, -0.0061, -0.112, -0.067, -0.0083, -0.112, -0.067, -0.0061, -0.1141, -0.0631, -0.0083, -0.1141, -0.0631, -0.0061, -0.112, -0.063, -0.0083, -0.112, -0.063, -0.01, -0.1141, -0.067, -0.01, -0.112, -0.067, -0.01, -0.1141, -0.0656, -0.01, -0.1141, -0.0631, -0.01, -0.112, -0.0656, -0.01, -0.112, -0.063, -0.0061, -0.1186, -0.0613, -0.0061, -0.1186, -0.0591, -0.0083, -0.1186, -0.0613, -0.0083, -0.1186, -0.0591, -0.0061, -0.1186, -0.057, -0.0061, -0.1186, -0.0548, -0.0083, -0.1186, -0.057, -0.0083, -0.1186, -0.0548, -0.01, -0.1186, -0.0613, -0.01, -0.1186, -0.0591, -0.01, -0.1186, -0.057, -0.01, -0.1186, -0.0548, -0.0061, -0.1098, -0.067, -0.0083, -0.1098, -0.067, -0.0061, -0.1076, -0.0669, -0.0083, -0.1076, -0.0669, -0.0061, -0.1098, -0.063,
-0.0083, -0.1098, -0.063, -0.0061, -0.1066, -0.0667, -0.0061, -0.1076, -0.0629, -0.0083, -0.1066, -0.0667, -0.0083, -0.1076, -0.0629, -0.01, -0.1098, -0.067, -0.01, -0.108, -0.0669, -0.01, -0.1098, -0.0656, -0.01, -0.1098, -0.063, -0.01, -0.1072, -0.0669, -0.01, -0.1078, -0.0629, -0.0061, -0.1054, -0.0666, -0.0061, -0.1061, -0.0625, -0.0083, -0.1054, -0.0666, -0.0083, -0.1061, -0.0625, -0.0061, -0.1033, -0.0658, -0.0083, -0.1033, -0.0658, -0.01, -0.1052, -0.0665, -0.01, -0.1058, -0.0624, -0.01, -0.1031, -0.0658, -0.01, -0.1033, -0.0635, -0.0061, -0.105, -0.0622, -0.0083, -0.105, -0.0622, -0.0061, -0.103, -0.0613, -0.0083, -0.103, -0.0613, -0.01, -0.1048, -0.0622, -0.01, -0.1034, -0.0616, -0.0061, -0.1186, -0.0526, -0.0061, -0.1186, -0.0505, -0.0083, -0.1186, -0.0526,
-0.0083, -0.1186, -0.0505, -0.0061, -0.1186, -0.0483, -0.0061, -0.1186, -0.0454, -0.0083, -0.1186, -0.0483, -0.0083, -0.1186, -0.0454, -0.0061, -0.1163, -0.0454, -0.0083, -0.1163, -0.0454, -0.01, -0.1186, -0.0526, -0.01, -0.1186, -0.0505, -0.01, -0.1186, -0.0483, -0.01, -0.1186, -0.0454, -0.01, -0.1163, -0.0454, -0.0061, -0.1141, -0.0454, -0.0083, -0.1141, -0.0454, -0.0061, -0.112, -0.0454, -0.0083, -0.112, -0.0454, -0.01, -0.1141, -0.0454, -0.01, -0.112, -0.0454, -0.0061, -0.1185, -0.0425, -0.0083, -0.1185, -0.0425, -0.0061, -0.1163, -0.0425, -0.0083, -0.1163, -0.0425, -0.01, -0.1185, -0.0439, -0.01, -0.1185, -0.0425, -0.01, -0.1163, -0.0439, -0.01, -0.1163, -0.0425, -0.0061, -0.1141, -0.0425, -0.0083, -0.1141, -0.0425, -0.0061, -0.112, -0.0425, -0.0083, -0.112, -0.0425,
-0.01, -0.1141, -0.0439, -0.01, -0.1141, -0.0425, -0.01, -0.112, -0.0439, -0.01, -0.112, -0.0425, -0.0061, -0.1098, -0.0455, -0.0083, -0.1098, -0.0455, -0.0061, -0.1076, -0.0459, -0.0083, -0.1076, -0.0459, -0.01, -0.1098, -0.0455, -0.01, -0.1077, -0.0459, -0.0061, -0.1054, -0.0465, -0.0083, -0.1054, -0.0465, -0.0061, -0.1031, -0.0478, -0.0061, -0.1043, -0.0471, -0.0083, -0.1031, -0.0478, -0.0083, -0.1043, -0.0471, -0.01, -0.1057, -0.0464, -0.01, -0.103, -0.0479, -0.01, -0.104, -0.0473, -0.0061, -0.1098, -0.0425, -0.0083, -0.1098, -0.0425, -0.0061, -0.1076, -0.0425, -0.0083, -0.1076, -0.0425, -0.01, -0.1098, -0.0439, -0.01, -0.1098, -0.0425, -0.01, -0.1076, -0.0439, -0.01, -0.1076, -0.0425, -0.0061, -0.1054, -0.0426, -0.0083, -0.1054, -0.0426, -0.0061, -0.1031, -0.0431,
-0.0061, -0.1042, -0.0428, -0.0083, -0.1031, -0.0431, -0.0083, -0.1042, -0.0428, -0.01, -0.1055, -0.0439, -0.01, -0.1054, -0.0426, -0.01, -0.1032, -0.0431, -0.01, -0.1043, -0.0428, -0.0061, -0.1358, -0.0219, -0.0083, -0.1358, -0.0219, -0.0061, -0.1337, -0.0219, -0.0083, -0.1337, -0.0219, -0.01, -0.1358, -0.0219, -0.01, -0.1358, -0.0201, -0.01, -0.1337, -0.0219, -0.01, -0.1337, -0.0201, -0.0061, -0.1315, -0.0219, -0.0083, -0.1315, -0.0219, -0.0061, -0.1293, -0.0219, -0.0083, -0.1293, -0.0219, -0.01, -0.1315, -0.0219, -0.01, -0.1315, -0.0201, -0.01, -0.1293, -0.0219, -0.01, -0.1293, -0.0201, -0.0061, -0.1272, -0.0219, -0.0083, -0.1272, -0.0219, -0.0061, -0.125, -0.0219, -0.0083, -0.125, -0.0219, -0.01, -0.1272, -0.0219, -0.01, -0.1272, -0.0201, -0.01, -0.125, -0.0219,
-0.01, -0.125, -0.0201, -0.0061, -0.1228, -0.0219, -0.0083, -0.1228, -0.0219, -0.0061, -0.1207, -0.0219, -0.0083, -0.1207, -0.0219, -0.01, -0.1228, -0.0219, -0.01, -0.1228, -0.0201, -0.01, -0.1207, -0.0219, -0.01, -0.1207, -0.0201, -0.01, -0.1358, -0.0179, -0.01, -0.1358, -0.0157, -0.01, -0.1337, -0.0179, -0.01, -0.1337, -0.0157, -0.01, -0.1358, -0.0136, -0.01, -0.1358, -0.0114, -0.01, -0.1337, -0.0136, -0.01, -0.1337, -0.0114, -0.01, -0.1315, -0.0179, -0.01, -0.1315, -0.0157, -0.01, -0.1293, -0.0179, -0.01, -0.1293, -0.0157, -0.01, -0.1315, -0.0136, -0.01, -0.1315, -0.0114, -0.01, -0.1293, -0.0136, -0.01, -0.1293, -0.0114, -0.0061, -0.1358, -0.0097, -0.0083, -0.1358, -0.0097, -0.0061, -0.1337, -0.01, -0.0083, -0.1337, -0.01, -0.01, -0.1358, -0.0097,
-0.01, -0.1337, -0.01, -0.0061, -0.1315, -0.0101, -0.0083, -0.1315, -0.0101, -0.0061, -0.1293, -0.0101, -0.0083, -0.1293, -0.0101, -0.01, -0.1315, -0.0101, -0.01, -0.1293, -0.0101, -0.01, -0.1272, -0.0179, -0.01, -0.1272, -0.0157, -0.01, -0.125, -0.0179, -0.01, -0.125, -0.0157, -0.01, -0.1272, -0.0136, -0.01, -0.1272, -0.0114, -0.01, -0.125, -0.0136, -0.01, -0.125, -0.0114, -0.01, -0.1228, -0.0179, -0.01, -0.1228, -0.0157, -0.01, -0.1207, -0.0179, -0.01, -0.1207, -0.0157, -0.01, -0.1228, -0.0136, -0.01, -0.1228, -0.0114, -0.01, -0.1207, -0.0136, -0.01, -0.1207, -0.0114, -0.0061, -0.1272, -0.0101, -0.0083, -0.1272, -0.0101, -0.0061, -0.125, -0.0101, -0.0083, -0.125, -0.0101, -0.01, -0.1272, -0.0101, -0.01, -0.125, -0.0101, -0.0061, -0.1228, -0.0101,
-0.0083, -0.1228, -0.0101, -0.0061, -0.1207, -0.0101, -0.0083, -0.1207, -0.0101, -0.01, -0.1228, -0.0101, -0.01, -0.1207, -0.0101, -0.0061, -0.1185, -0.0219, -0.0083, -0.1185, -0.0219, -0.0061, -0.1163, -0.0219, -0.0083, -0.1163, -0.0219, -0.01, -0.1185, -0.0219, -0.01, -0.1185, -0.0201, -0.01, -0.1163, -0.0219, -0.01, -0.1163, -0.0201, -0.0061, -0.1141, -0.0219, -0.0083, -0.1141, -0.0219, -0.0061, -0.112, -0.0219, -0.0083, -0.112, -0.0219, -0.01, -0.1141, -0.0219, -0.01, -0.1141, -0.0201, -0.01, -0.112, -0.0219, -0.01, -0.112, -0.0201, -0.0061, -0.1098, -0.0219, -0.0083, -0.1098, -0.0219, -0.0061, -0.1076, -0.0219, -0.0083, -0.1076, -0.0219, -0.01, -0.1098, -0.0219, -0.01, -0.1098, -0.0201, -0.01, -0.1076, -0.0219, -0.01, -0.1076, -0.0201, -0.0061, -0.1055, -0.0219,
-0.0083, -0.1055, -0.0219, -0.0061, -0.1033, -0.0219, -0.0083, -0.1033, -0.0219, -0.01, -0.1055, -0.0219, -0.01, -0.1055, -0.0201, -0.01, -0.1033, -0.0219, -0.01, -0.1033, -0.0201, -0.01, -0.1185, -0.0179, -0.01, -0.1185, -0.0157, -0.01, -0.1163, -0.0179, -0.01, -0.1163, -0.0157, -0.01, -0.1185, -0.0136, -0.01, -0.1185, -0.0114, -0.01, -0.1163, -0.0136, -0.01, -0.1163, -0.0114, -0.01, -0.1141, -0.0179, -0.01, -0.1141, -0.0157, -0.01, -0.112, -0.0179, -0.01, -0.112, -0.0157, -0.01, -0.1141, -0.0136, -0.01, -0.1141, -0.0114, -0.01, -0.112, -0.0136, -0.01, -0.112, -0.0114, -0.0061, -0.1185, -0.0101, -0.0083, -0.1185, -0.0101, -0.0061, -0.1163, -0.0101, -0.0083, -0.1163, -0.0101, -0.01, -0.1185, -0.0101, -0.01, -0.1163, -0.0101, -0.0061, -0.1141, -0.0101,
-0.0083, -0.1141, -0.0101, -0.0061, -0.112, -0.0101, -0.0083, -0.112, -0.0101, -0.01, -0.1141, -0.0101, -0.01, -0.112, -0.0101, -0.01, -0.1098, -0.0179, -0.01, -0.1098, -0.0157, -0.01, -0.1076, -0.0179, -0.01, -0.1076, -0.0157, -0.01, -0.1098, -0.0136, -0.01, -0.1098, -0.0114, -0.01, -0.1076, -0.0136, -0.01, -0.1076, -0.0114, -0.01, -0.1055, -0.0179, -0.01, -0.1055, -0.0157, -0.01, -0.1033, -0.0179, -0.01, -0.1033, -0.0157, -0.01, -0.1055, -0.0136, -0.01, -0.1055, -0.0114, -0.01, -0.1033, -0.0136, -0.01, -0.1033, -0.0114, -0.0061, -0.1098, -0.0101, -0.0083, -0.1098, -0.0101, -0.0061, -0.1076, -0.0101, -0.0083, -0.1076, -0.0101, -0.01, -0.1098, -0.0101, -0.01, -0.1076, -0.0101, -0.0061, -0.1055, -0.0101, -0.0083, -0.1055, -0.0101, -0.0061, -0.1033, -0.0101,
-0.0083, -0.1033, -0.0101, -0.01, -0.1055, -0.0101, -0.01, -0.1033, -0.0101, -0.0061, -0.1016, -0.065, -0.0061, -0.1005, -0.0642, -0.0083, -0.1016, -0.065, -0.0083, -0.1005, -0.0642, -0.0061, -0.0994, -0.0631, -0.0083, -0.0994, -0.0631, -0.01, -0.1017, -0.065, -0.01, -0.1006, -0.0642, -0.01, -0.0994, -0.0631, -0.0061, -0.102, -0.0604, -0.0061, -0.1007, -0.0591, -0.0083, -0.102, -0.0604, -0.0083, -0.1007, -0.0591, -0.0061, -0.0983, -0.0618, -0.0083, -0.0983, -0.0618, -0.0061, -0.1002, -0.0577, -0.0083, -0.1002, -0.0577, -0.0061, -0.0997, -0.0566, -0.0061, -0.0995, -0.0548, -0.0083, -0.0997, -0.0566, -0.0083, -0.0995, -0.0548, -0.01, -0.102, -0.0604, -0.01, -0.1007, -0.059, -0.01, -0.0983, -0.0618, -0.01, -0.099, -0.0591, -0.01, -0.1002, -0.0578, -0.01, -0.0998, -0.0569,
-0.01, -0.0995, -0.0548, -0.0061, -0.0976, -0.0607, -0.0061, -0.0968, -0.0592, -0.0083, -0.0976, -0.0607, -0.0083, -0.0968, -0.0592, -0.0061, -0.0962, -0.057, -0.0061, -0.0961, -0.0548, -0.0083, -0.0962, -0.057, -0.0083, -0.0961, -0.0548, -0.01, -0.0975, -0.0606, -0.01, -0.0968, -0.0594, -0.01, -0.0962, -0.0571, -0.01, -0.0961, -0.0548, -0.0061, -0.1002, -0.0518, -0.0061, -0.1008, -0.0504, -0.0083, -0.1002, -0.0518, -0.0083, -0.1008, -0.0504, -0.0061, -0.0997, -0.0529, -0.0083, -0.0997, -0.0529, -0.0061, -0.1019, -0.049, -0.0083, -0.1019, -0.049, -0.0061, -0.0979, -0.0473, -0.0061, -0.0988, -0.046, -0.0083, -0.0979, -0.0473, -0.0083, -0.0988, -0.046, -0.01, -0.1002, -0.0518, -0.01, -0.1006, -0.0506, -0.01, -0.0998, -0.0527, -0.01, -0.099, -0.0505, -0.01, -0.1019, -0.049,
-0.01, -0.1011, -0.0461, -0.01, -0.0978, -0.0474, -0.01, -0.099, -0.0457, -0.0061, -0.0961, -0.0526, -0.0061, -0.0964, -0.0504, -0.0083, -0.0961, -0.0526, -0.0083, -0.0964, -0.0504, -0.0061, -0.0972, -0.0483, -0.0083, -0.0972, -0.0483, -0.01, -0.0961, -0.0526, -0.01, -0.0965, -0.0503, -0.01, -0.0971, -0.0485, -0.0061, -0.101, -0.0441, -0.0083, -0.101, -0.0441, -0.0061, -0.1, -0.045, -0.0083, -0.1, -0.045, -0.01, -0.1009, -0.0442, -0.01, -0.1, -0.045, -0.0061, -0.0766, -0.0589, -0.0083, -0.0766, -0.0589, -0.0061, -0.0771, -0.057, -0.0061, -0.0774, -0.0548, -0.0083, -0.0771, -0.057, -0.0083, -0.0774, -0.0548, -0.01, -0.0766, -0.0587, -0.01, -0.0771, -0.057, -0.01, -0.0774, -0.0548, -0.0061, -0.0742, -0.0627, -0.0083, -0.0742, -0.0627, -0.0061, -0.072, -0.0647,
-0.0061, -0.0731, -0.0638, -0.0083, -0.072, -0.0647, -0.0083, -0.0731, -0.0638, -0.01, -0.0742, -0.0626, -0.01, -0.072, -0.0647, -0.01, -0.0732, -0.0638, -0.0061, -0.0708, -0.0655, -0.0083, -0.0708, -0.0655, -0.0061, -0.0686, -0.0664, -0.0061, -0.0676, -0.0624, -0.0083, -0.0686, -0.0664, -0.0083, -0.0676, -0.0624, -0.01, -0.071, -0.0654, -0.01, -0.0707, -0.0635, -0.01, -0.0688, -0.0663, -0.01, -0.0681, -0.0623, -0.0061, -0.0753, -0.0614, -0.0061, -0.0761, -0.06, -0.0083, -0.0753, -0.0614, -0.0083, -0.0761, -0.06, -0.0061, -0.0723, -0.0587, -0.0083, -0.0723, -0.0587, -0.0061, -0.0729, -0.057, -0.0061, -0.0733, -0.0548, -0.0083, -0.0729, -0.057, -0.0083, -0.0733, -0.0548, -0.01, -0.0751, -0.0617, -0.01, -0.0762, -0.0597, -0.01, -0.0729, -0.0613, -0.01, -0.0722, -0.0588,
-0.01, -0.0751, -0.057, -0.01, -0.0751, -0.0548, -0.01, -0.073, -0.0568, -0.01, -0.0733, -0.0548, -0.0061, -0.0707, -0.0608, -0.0061, -0.0715, -0.0598, -0.0083, -0.0707, -0.0608, -0.0083, -0.0715, -0.0598, -0.0061, -0.0687, -0.062, -0.0083, -0.0687, -0.062, -0.01, -0.0705, -0.061, -0.01, -0.0714, -0.0598, -0.01, -0.0689, -0.0619, -0.0061, -0.0773, -0.0526, -0.0061, -0.0769, -0.0504, -0.0083, -0.0773, -0.0526, -0.0083, -0.0769, -0.0504, -0.0061, -0.0763, -0.049, -0.0083, -0.0763, -0.049, -0.01, -0.0773, -0.0526, -0.01, -0.0769, -0.0503, -0.01, -0.0764, -0.0491, -0.0061, -0.0732, -0.0526, -0.0061, -0.0727, -0.0508, -0.0083, -0.0732, -0.0526, -0.0083, -0.0727, -0.0508, -0.0061, -0.0758, -0.0479, -0.0061, -0.0747, -0.0464, -0.0083, -0.0758, -0.0479, -0.0083, -0.0747, -0.0464,
-0.0061, -0.0736, -0.0453, -0.0083, -0.0736, -0.0453, -0.01, -0.0751, -0.0526, -0.01, -0.0751, -0.0505, -0.01, -0.0732, -0.0527, -0.01, -0.0725, -0.0505, -0.01, -0.0758, -0.048, -0.01, -0.0748, -0.0464, -0.01, -0.0729, -0.0483, -0.01, -0.0736, -0.0453, -0.0061, -0.0717, -0.0495, -0.0083, -0.0717, -0.0495, -0.0061, -0.0707, -0.0482, -0.0083, -0.0707, -0.0482, -0.0061, -0.0694, -0.0474, -0.0061, -0.0683, -0.0468, -0.0083, -0.0694, -0.0474, -0.0083, -0.0683, -0.0468, -0.01, -0.0717, -0.0495, -0.01, -0.0708, -0.0483, -0.01, -0.0707, -0.0461, -0.01, -0.0694, -0.0474, -0.01, -0.0684, -0.0468, -0.0061, -0.0726, -0.0445, -0.0083, -0.0726, -0.0445, -0.01, -0.0725, -0.0445, -0.0061, -0.0708, -0.0435, -0.0083, -0.0708, -0.0435, -0.0061, -0.0692, -0.043, -0.0061, -0.0681, -0.0427,
-0.0083, -0.0692, -0.043, -0.0083, -0.0681, -0.0427, -0.01, -0.0711, -0.0437, -0.01, -0.0689, -0.0429, -0.01, -0.0679, -0.0427, -0.0061, -0.1011, -0.0219, -0.0083, -0.1011, -0.0219, -0.0061, -0.099, -0.0219, -0.0083, -0.099, -0.0219, -0.01, -0.1011, -0.0219, -0.01, -0.1011, -0.0201, -0.01, -0.099, -0.0219, -0.01, -0.099, -0.0201, -0.0061, -0.0968, -0.0219, -0.0083, -0.0968, -0.0219, -0.0061, -0.0946, -0.0219, -0.0083, -0.0946, -0.0219, -0.01, -0.0968, -0.0219, -0.01, -0.0968, -0.0201, -0.01, -0.0946, -0.0219, -0.01, -0.0946, -0.0201, -0.0061, -0.0924, -0.0219, -0.0083, -0.0924, -0.0219, -0.0061, -0.0902, -0.0219, -0.0061, -0.0902, -0.0201, -0.0083, -0.0902, -0.0219, -0.0083, -0.0902, -0.0201, -0.01, -0.0924, -0.0219, -0.01, -0.0924, -0.0201, -0.01, -0.0902, -0.0219,
-0.01, -0.0902, -0.0201, -0.01, -0.1011, -0.0179, -0.01, -0.1011, -0.0157, -0.01, -0.099, -0.0179, -0.01, -0.099, -0.0157, -0.01, -0.1011, -0.0136, -0.01, -0.1011, -0.0114, -0.01, -0.099, -0.0136, -0.01, -0.099, -0.0114, -0.01, -0.0968, -0.0179, -0.01, -0.0968, -0.0157, -0.01, -0.0946, -0.0179, -0.01, -0.0946, -0.0157, -0.01, -0.0968, -0.0136, -0.01, -0.0968, -0.0114, -0.01, -0.0946, -0.0136, -0.01, -0.0946, -0.0114, -0.0061, -0.1011, -0.0101, -0.0083, -0.1011, -0.0101, -0.0061, -0.099, -0.0101, -0.0083, -0.099, -0.0101, -0.01, -0.1011, -0.0101, -0.01, -0.099, -0.0101, -0.0061, -0.0968, -0.0101, -0.0083, -0.0968, -0.0101, -0.0061, -0.0946, -0.0101, -0.0083, -0.0946, -0.0101, -0.01, -0.0968, -0.0101, -0.01, -0.0946, -0.0101, -0.0061, -0.0902, -0.0179,
-0.0061, -0.0902, -0.0157, -0.0083, -0.0902, -0.0179, -0.0083, -0.0902, -0.0157, -0.0061, -0.0902, -0.0136, -0.0061, -0.0902, -0.0114, -0.0083, -0.0902, -0.0136, -0.0083, -0.0902, -0.0114, -0.01, -0.0924, -0.0179, -0.01, -0.0924, -0.0157, -0.01, -0.0902, -0.0179, -0.01, -0.0902, -0.0157, -0.01, -0.0924, -0.0136, -0.01, -0.0924, -0.0114, -0.01, -0.0902, -0.0136, -0.01, -0.0902, -0.0114, -0.0061, -0.0924, -0.0101, -0.0083, -0.0924, -0.0101, -0.0061, -0.0902, -0.0101, -0.0083, -0.0902, -0.0101, -0.01, -0.0924, -0.0101, -0.01, -0.0902, -0.0101, -0.0061, -0.0775, -0.0219, -0.0061, -0.0775, -0.0201, -0.0083, -0.0775, -0.0219, -0.0083, -0.0775, -0.0201, -0.01, -0.0775, -0.0219, -0.01, -0.0775, -0.0201, -0.0061, -0.0751, -0.0219, -0.0083, -0.0751, -0.0219, -0.0061, -0.0729, -0.0219,
-0.0083, -0.0729, -0.0219, -0.01, -0.0751, -0.0219, -0.01, -0.0751, -0.0201, -0.01, -0.0729, -0.0219, -0.01, -0.0729, -0.0201, -0.0061, -0.0707, -0.0219, -0.0083, -0.0707, -0.0219, -0.0061, -0.0686, -0.0219, -0.0083, -0.0686, -0.0219, -0.01, -0.0707, -0.0219, -0.01, -0.0707, -0.0201, -0.01, -0.0686, -0.0219, -0.01, -0.0686, -0.0201, -0.0061, -0.0775, -0.0179, -0.0061, -0.0775, -0.0157, -0.0083, -0.0775, -0.0179, -0.0083, -0.0775, -0.0157, -0.0061, -0.0775, -0.0136, -0.0061, -0.0775, -0.0114, -0.0083, -0.0775, -0.0136, -0.0083, -0.0775, -0.0114, -0.01, -0.0775, -0.0179, -0.01, -0.0775, -0.0157, -0.01, -0.0775, -0.0136, -0.01, -0.0775, -0.0114, -0.0061, -0.0775, -0.0092, -0.0061, -0.0775, -0.0071, -0.0083, -0.0775, -0.0092, -0.0083, -0.0775, -0.0071, -0.0061, -0.0775, -0.0049,
-0.0061, -0.0775, -0.0027, -0.0083, -0.0775, -0.0049, -0.0083, -0.0775, -0.0027, -0.01, -0.0775, -0.0092, -0.01, -0.0775, -0.0071, -0.01, -0.0775, -0.0049, -0.01, -0.0775, -0.0027, -0.01, -0.0751, -0.0179, -0.01, -0.0751, -0.0157, -0.01, -0.0729, -0.0179, -0.01, -0.0729, -0.0157, -0.01, -0.0751, -0.0136, -0.01, -0.0751, -0.0114, -0.01, -0.0729, -0.0136, -0.01, -0.0729, -0.0114, -0.01, -0.0707, -0.0179, -0.01, -0.0707, -0.0157, -0.01, -0.0686, -0.0179, -0.01, -0.0686, -0.0157, -0.01, -0.0707, -0.0136, -0.01, -0.0707, -0.0114, -0.01, -0.0686, -0.0136, -0.01, -0.0686, -0.0114, -0.01, -0.0751, -0.0092, -0.01, -0.0751, -0.0071, -0.01, -0.0729, -0.0092, -0.01, -0.0729, -0.0071, -0.01, -0.0751, -0.0049, -0.01, -0.0751, -0.0027, -0.01, -0.0729, -0.0049,
-0.01, -0.0729, -0.0027, -0.01, -0.0707, -0.0092, -0.01, -0.0707, -0.0071, -0.01, -0.0686, -0.0092, -0.01, -0.0686, -0.0071, -0.01, -0.0707, -0.0049, -0.01, -0.0707, -0.0027, -0.01, -0.0686, -0.0049, -0.01, -0.0686, -0.0027, -0.0061, -0.0661, -0.0669, -0.0083, -0.0661, -0.0669, -0.0061, -0.0642, -0.067, -0.0083, -0.0642, -0.067, -0.0061, -0.0672, -0.0667, -0.0061, -0.0664, -0.0627, -0.0083, -0.0672, -0.0667, -0.0083, -0.0664, -0.0627, -0.0061, -0.0642, -0.063, -0.0083, -0.0642, -0.063, -0.01, -0.0659, -0.0669, -0.01, -0.0642, -0.067, -0.01, -0.0667, -0.0668, -0.01, -0.0663, -0.0628, -0.01, -0.0642, -0.0656, -0.01, -0.0642, -0.063, -0.0061, -0.0621, -0.067, -0.0083, -0.0621, -0.067, -0.0061, -0.0599, -0.067, -0.0083, -0.0599, -0.067, -0.0061, -0.0621, -0.063,
-0.0083, -0.0621, -0.063, -0.0061, -0.0599, -0.0631, -0.0083, -0.0599, -0.0631, -0.01, -0.0621, -0.067, -0.01, -0.0599, -0.067, -0.01, -0.0621, -0.0656, -0.01, -0.062, -0.063, -0.01, -0.0599, -0.0656, -0.01, -0.0599, -0.0631, -0.0061, -0.0577, -0.0669, -0.0083, -0.0577, -0.0669, -0.0061, -0.0565, -0.0667, -0.0083, -0.0565, -0.0667, -0.0061, -0.0577, -0.063, -0.0083, -0.0577, -0.063, -0.0061, -0.0554, -0.0666, -0.0061, -0.0559, -0.0626, -0.0083, -0.0554, -0.0666, -0.0083, -0.0559, -0.0626, -0.01, -0.0577, -0.0669, -0.01, -0.0566, -0.0668, -0.01, -0.0577, -0.0656, -0.01, -0.0578, -0.063, -0.01, -0.0555, -0.0666, -0.01, -0.0557, -0.0625, -0.0061, -0.0534, -0.066, -0.0083, -0.0534, -0.066, -0.0061, -0.0514, -0.0651, -0.0061, -0.0503, -0.0645, -0.0083, -0.0514, -0.0651,
-0.0083, -0.0503, -0.0645, -0.01, -0.0532, -0.0659, -0.01, -0.0534, -0.0635, -0.01, -0.0515, -0.0652, -0.01, -0.0505, -0.0646, -0.0061, -0.0548, -0.0623, -0.0083, -0.0548, -0.0623, -0.01, -0.0547, -0.0623, -0.0061, -0.0533, -0.0615, -0.0083, -0.0533, -0.0615, -0.0061, -0.052, -0.0605, -0.0061, -0.0509, -0.0592, -0.0083, -0.052, -0.0605, -0.0083, -0.0509, -0.0592, -0.0061, -0.0502, -0.058, -0.0083, -0.0502, -0.058, -0.01, -0.0535, -0.0617, -0.01, -0.052, -0.0605, -0.01, -0.0507, -0.059, -0.01, -0.0502, -0.058, -0.0061, -0.0664, -0.0461, -0.0083, -0.0664, -0.0461, -0.0061, -0.0642, -0.0456, -0.0083, -0.0642, -0.0456, -0.01, -0.0663, -0.0461, -0.01, -0.0642, -0.0456, -0.0061, -0.0621, -0.0454, -0.0083, -0.0621, -0.0454, -0.0061, -0.0599, -0.0455, -0.0083, -0.0599, -0.0455,
-0.01, -0.0621, -0.0454, -0.01, -0.0599, -0.0455, -0.0061, -0.0664, -0.0425, -0.0083, -0.0664, -0.0425, -0.0061, -0.0642, -0.0425, -0.0083, -0.0642, -0.0425, -0.01, -0.0664, -0.0439, -0.01, -0.0664, -0.0425, -0.01, -0.0642, -0.0439, -0.01, -0.0642, -0.0425, -0.0061, -0.0621, -0.0425, -0.0083, -0.0621, -0.0425, -0.0061, -0.0599, -0.0425, -0.0083, -0.0599, -0.0425, -0.01, -0.0621, -0.0439, -0.01, -0.0621, -0.0425, -0.01, -0.0599, -0.0439, -0.01, -0.0599, -0.0425, -0.0061, -0.0577, -0.0458, -0.0083, -0.0577, -0.0458, -0.0061, -0.0555, -0.0463, -0.0083, -0.0555, -0.0463, -0.01, -0.0578, -0.0457, -0.01, -0.0557, -0.0463, -0.0061, -0.0506, -0.0502, -0.0083, -0.0506, -0.0502, -0.0061, -0.0529, -0.0476, -0.0061, -0.054, -0.047, -0.0083, -0.0529, -0.0476, -0.0083, -0.054, -0.047,
-0.0061, -0.0517, -0.0487, -0.0083, -0.0517, -0.0487, -0.01, -0.0507, -0.0501, -0.01, -0.0528, -0.0477, -0.01, -0.0538, -0.0471, -0.01, -0.0517, -0.0487, -0.01, -0.0512, -0.0461, -0.0061, -0.0577, -0.0425, -0.0083, -0.0577, -0.0425, -0.0061, -0.0555, -0.0426, -0.0083, -0.0555, -0.0426, -0.01, -0.0577, -0.0439, -0.01, -0.0577, -0.0425, -0.01, -0.0555, -0.0439, -0.01, -0.0555, -0.0426, -0.0061, -0.0532, -0.0431, -0.0061, -0.0544, -0.0428, -0.0083, -0.0532, -0.0431, -0.0083, -0.0544, -0.0428, -0.0061, -0.0511, -0.044, -0.0083, -0.0511, -0.044, -0.01, -0.0533, -0.0431, -0.01, -0.0544, -0.0428, -0.01, -0.0509, -0.0442, -0.0061, -0.049, -0.0635, -0.0083, -0.049, -0.0635, -0.01, -0.049, -0.0636, -0.0061, -0.0481, -0.0623, -0.0083, -0.0481, -0.0623, -0.0061, -0.0473, -0.0613,
-0.0061, -0.0465, -0.0596, -0.0083, -0.0473, -0.0613, -0.0083, -0.0465, -0.0596, -0.0061, -0.0497, -0.0569, -0.0061, -0.0492, -0.0547, -0.0083, -0.0497, -0.0569, -0.0083, -0.0492, -0.0547, -0.0061, -0.0461, -0.057, -0.0061, -0.0461, -0.0548, -0.0083, -0.0461, -0.057, -0.0083, -0.0461, -0.0548, -0.01, -0.048, -0.0622, -0.01, -0.049, -0.0591, -0.01, -0.0472, -0.0611, -0.01, -0.0465, -0.0594, -0.01, -0.0497, -0.0569, -0.01, -0.0492, -0.0548, -0.01, -0.0461, -0.0571, -0.01, -0.0461, -0.0548, -0.0061, -0.0496, -0.0526, -0.0061, -0.05, -0.0513, -0.0083, -0.0496, -0.0526, -0.0083, -0.05, -0.0513, -0.0061, -0.0461, -0.0526, -0.0061, -0.0463, -0.0504, -0.0083, -0.0461, -0.0526, -0.0083, -0.0463, -0.0504, -0.0061, -0.0488, -0.0459, -0.0083, -0.0488, -0.0459, -0.0061, -0.047, -0.0482,
-0.0061, -0.0479, -0.0471, -0.0083, -0.047, -0.0482, -0.0083, -0.0479, -0.0471, -0.01, -0.0495, -0.0528, -0.01, -0.0502, -0.051, -0.01, -0.0461, -0.0526, -0.01, -0.0464, -0.0503, -0.01, -0.049, -0.0483, -0.01, -0.0489, -0.0459, -0.01, -0.0471, -0.048, -0.01, -0.0479, -0.0471, -0.0061, -0.05, -0.045, -0.0083, -0.05, -0.045, -0.01, -0.05, -0.0449, -0.0061, -0.0667, -0.0219, -0.0061, -0.0667, -0.0201, -0.0083, -0.0667, -0.0219, -0.0083, -0.0667, -0.0201, -0.01, -0.0667, -0.0219, -0.01, -0.0667, -0.0201, -0.0061, -0.0667, -0.0179, -0.0061, -0.0667, -0.0157, -0.0083, -0.0667, -0.0179, -0.0083, -0.0667, -0.0157, -0.0061, -0.0667, -0.0136, -0.0061, -0.0667, -0.0114, -0.0083, -0.0667, -0.0136, -0.0083, -0.0667, -0.0114, -0.01, -0.0667, -0.0179, -0.01, -0.0667, -0.0157,
-0.01, -0.0667, -0.0136, -0.01, -0.0667, -0.0114, -0.0061, -0.0667, -0.0092, -0.0061, -0.0667, -0.0071, -0.0083, -0.0667, -0.0092, -0.0083, -0.0667, -0.0071, -0.0061, -0.0667, -0.0049, -0.0061, -0.0667, -0.0027, -0.0083, -0.0667, -0.0049, -0.0083, -0.0667, -0.0027, -0.01, -0.0667, -0.0092, -0.01, -0.0667, -0.0071, -0.01, -0.0667, -0.0049, -0.01, -0.0667, -0.0027, -0.0061, -0.05, -0.0219, -0.0061, -0.05, -0.0201, -0.0083, -0.05, -0.0219, -0.0083, -0.05, -0.0201, -0.0061, -0.0469, -0.0219, -0.0083, -0.0469, -0.0219, -0.01, -0.05, -0.0219, -0.01, -0.05, -0.0201, -0.01, -0.0469, -0.0219, -0.01, -0.0469, -0.0201, -0.0061, -0.0447, -0.0219, -0.0083, -0.0447, -0.0219, -0.0061, -0.0425, -0.0219, -0.0083, -0.0425, -0.0219, -0.01, -0.0447, -0.0219, -0.01, -0.0447, -0.0201,
-0.01, -0.0425, -0.0219, -0.01, -0.0425, -0.0201, -0.0061, -0.0404, -0.0219, -0.0083, -0.0404, -0.0219, -0.0061, -0.0382, -0.0219, -0.0061, -0.0382, -0.0201, -0.0083, -0.0382, -0.0219, -0.0083, -0.0382, -0.0201, -0.01, -0.0404, -0.0219, -0.01, -0.0404, -0.0201, -0.01, -0.0382, -0.0219, -0.01, -0.0382, -0.0201, -0.0061, -0.05, -0.0179, -0.0061, -0.05, -0.0157, -0.0083, -0.05, -0.0179, -0.0083, -0.05, -0.0157, -0.0061, -0.05, -0.0136, -0.0061, -0.05, -0.0114, -0.0083, -0.05, -0.0136, -0.0083, -0.05, -0.0114, -0.01, -0.05, -0.0179, -0.01, -0.05, -0.0157, -0.01, -0.0469, -0.0179, -0.01, -0.0469, -0.0157, -0.01, -0.05, -0.0136, -0.01, -0.05, -0.0114, -0.01, -0.0469, -0.0136, -0.01, -0.0469, -0.0114, -0.01, -0.0447, -0.0179, -0.01, -0.0447, -0.0157,
-0.01, -0.0425, -0.0179, -0.01, -0.0425, -0.0157, -0.01, -0.0447, -0.0136, -0.01, -0.0447, -0.0114, -0.01, -0.0425, -0.0136, -0.01, -0.0425, -0.0114, -0.0061, -0.05, -0.0092, -0.0061, -0.05, -0.0071, -0.0083, -0.05, -0.0092, -0.0083, -0.05, -0.0071, -0.0061, -0.05, -0.0049, -0.0061, -0.05, -0.0027, -0.0083, -0.05, -0.0049, -0.0083, -0.05, -0.0027, -0.01, -0.05, -0.0092, -0.01, -0.05, -0.0071, -0.01, -0.0469, -0.0092, -0.01, -0.0469, -0.0071, -0.01, -0.05, -0.0049, -0.01, -0.05, -0.0027, -0.01, -0.0469, -0.0049, -0.01, -0.0469, -0.0027, -0.01, -0.0447, -0.0092, -0.01, -0.0447, -0.0071, -0.01, -0.0425, -0.0092, -0.01, -0.0425, -0.0071, -0.01, -0.0447, -0.0049, -0.01, -0.0447, -0.0027, -0.01, -0.0425, -0.0049, -0.01, -0.0425, -0.0027,
-0.0061, -0.0382, -0.0179, -0.0061, -0.0382, -0.0157, -0.0083, -0.0382, -0.0179, -0.0083, -0.0382, -0.0157, -0.0061, -0.0382, -0.0136, -0.0061, -0.0382, -0.0114, -0.0083, -0.0382, -0.0136, -0.0083, -0.0382, -0.0114, -0.01, -0.0404, -0.0179, -0.01, -0.0404, -0.0157, -0.01, -0.0382, -0.0179, -0.01, -0.0382, -0.0157, -0.01, -0.0404, -0.0136, -0.01, -0.0404, -0.0114, -0.01, -0.0382, -0.0136, -0.01, -0.0382, -0.0114, -0.0061, -0.0382, -0.0092, -0.0061, -0.0382, -0.0071, -0.0083, -0.0382, -0.0092, -0.0083, -0.0382, -0.0071, -0.0061, -0.0382, -0.0049, -0.0061, -0.0382, -0.0027, -0.0083, -0.0382, -0.0049, -0.0083, -0.0382, -0.0027, -0.01, -0.0404, -0.0092, -0.01, -0.0404, -0.0071, -0.01, -0.0382, -0.0092, -0.01, -0.0382, -0.0071, -0.01, -0.0404, -0.0049, -0.01, -0.0404, -0.0027,
-0.01, -0.0382, -0.0049, -0.01, -0.0382, -0.0027, -0.0061, -0.0265, -0.067, -0.0083, -0.0265, -0.067, -0.0061, -0.0252, -0.067, -0.0083, -0.0252, -0.067, -0.0061, -0.0265, -0.0656, -0.0061, -0.0265, -0.0635, -0.0083, -0.0265, -0.0656, -0.0083, -0.0265, -0.0635, -0.01, -0.0265, -0.067, -0.01, -0.0252, -0.067, -0.01, -0.0265, -0.0656, -0.01, -0.0265, -0.0635, -0.01, -0.0252, -0.0656, -0.01, -0.0252, -0.0635, -0.0061, -0.0265, -0.0613, -0.0061, -0.0265, -0.0591, -0.0083, -0.0265, -0.0613, -0.0083, -0.0265, -0.0591, -0.0061, -0.0265, -0.057, -0.0061, -0.0265, -0.0548, -0.0083, -0.0265, -0.057, -0.0083, -0.0265, -0.0548, -0.01, -0.0265, -0.0613, -0.01, -0.0265, -0.0591, -0.01, -0.0252, -0.0613, -0.01, -0.0252, -0.0591, -0.01, -0.0265, -0.057, -0.01, -0.0265, -0.0548,
-0.01, -0.0252, -0.057, -0.01, -0.0252, -0.0548, -0.0061, -0.0225, -0.067, -0.0083, -0.0225, -0.067, -0.0061, -0.0225, -0.0656, -0.0061, -0.0225, -0.0635, -0.0083, -0.0225, -0.0656, -0.0083, -0.0225, -0.0635, -0.01, -0.0225, -0.067, -0.01, -0.0225, -0.0656, -0.01, -0.0225, -0.0635, -0.0061, -0.0225, -0.0613, -0.0061, -0.0225, -0.0591, -0.0083, -0.0225, -0.0613, -0.0083, -0.0225, -0.0591, -0.0061, -0.0225, -0.057, -0.0061, -0.0225, -0.0548, -0.0083, -0.0225, -0.057, -0.0083, -0.0225, -0.0548, -0.01, -0.0225, -0.0613, -0.01, -0.0225, -0.0591, -0.01, -0.0225, -0.057, -0.01, -0.0225, -0.0548, -0.0061, -0.0179, -0.054, -0.0083, -0.0179, -0.054, -0.0061, -0.0157, -0.0562, -0.0061, -0.0168, -0.0551, -0.0083, -0.0157, -0.0562, -0.0083, -0.0168, -0.0551, -0.01, -0.0179, -0.054,
-0.01, -0.0157, -0.0562, -0.01, -0.0168, -0.0551, -0.0061, -0.0265, -0.0526, -0.0061, -0.0265, -0.0505, -0.0083, -0.0265, -0.0526, -0.0083, -0.0265, -0.0505, -0.0061, -0.0265, -0.0483, -0.0061, -0.0264, -0.0466, -0.0083, -0.0265, -0.0483, -0.0083, -0.0264, -0.0466, -0.0061, -0.0262, -0.0455, -0.0083, -0.0262, -0.0455, -0.01, -0.0265, -0.0526, -0.01, -0.0265, -0.0505, -0.01, -0.0252, -0.0526, -0.01, -0.0252, -0.0505, -0.01, -0.0265, -0.0483, -0.01, -0.0264, -0.0468, -0.01, -0.0252, -0.0483, -0.01, -0.0262, -0.0459, -0.0061, -0.0255, -0.0437, -0.0083, -0.0255, -0.0437, -0.01, -0.0255, -0.0437, -0.0061, -0.0225, -0.0526, -0.0061, -0.0225, -0.0503, -0.0083, -0.0225, -0.0526, -0.0083, -0.0225, -0.0503, -0.0061, -0.0202, -0.052, -0.0061, -0.0213, -0.0511, -0.0083, -0.0202, -0.052,
-0.0083, -0.0213, -0.0511, -0.0061, -0.02, -0.0452, -0.0083, -0.02, -0.0452, -0.01, -0.0225, -0.0526, -0.01, -0.0225, -0.0503, -0.01, -0.0202, -0.052, -0.01, -0.0213, -0.0511, -0.01, -0.023, -0.0483, -0.01, -0.023, -0.0461, -0.01, -0.0208, -0.0483, -0.01, -0.0201, -0.0452, -0.0061, -0.019, -0.053, -0.0083, -0.019, -0.053, -0.0061, -0.0155, -0.0495, -0.0083, -0.0155, -0.0495, -0.0061, -0.0177, -0.0473, -0.0061, -0.0188, -0.0463, -0.0083, -0.0177, -0.0473, -0.0083, -0.0188, -0.0463, -0.0061, -0.0166, -0.0484, -0.0083, -0.0166, -0.0484, -0.01, -0.019, -0.053, -0.01, -0.0187, -0.0505, -0.01, -0.0165, -0.0526, -0.01, -0.0155, -0.0495, -0.01, -0.0177, -0.0473, -0.01, -0.0188, -0.0463, -0.01, -0.0166, -0.0484, -0.0061, -0.0231, -0.0433, -0.0083, -0.0231, -0.0433,
-0.0061, -0.021, -0.0444, -0.0083, -0.021, -0.0444, -0.01, -0.0231, -0.0433, -0.01, -0.0211, -0.0443, -0.0061, -0.0092, -0.0627, -0.0083, -0.0092, -0.0627, -0.0061, -0.0069, -0.0647, -0.0061, -0.008, -0.0638, -0.0083, -0.0069, -0.0647, -0.0083, -0.008, -0.0638, -0.01, -0.0092, -0.0627, -0.01, -0.0069, -0.0647, -0.01, -0.008, -0.0638, -0.0061, -0.0135, -0.0584, -0.0083, -0.0135, -0.0584, -0.0061, -0.0114, -0.0605, -0.0061, -0.0124, -0.0594, -0.0083, -0.0114, -0.0605, -0.0083, -0.0124, -0.0594, -0.0061, -0.0146, -0.0573, -0.0083, -0.0146, -0.0573, -0.0061, -0.0112, -0.0538, -0.0083, -0.0112, -0.0538, -0.01, -0.0135, -0.0584, -0.01, -0.0114, -0.0605, -0.01, -0.0124, -0.0594, -0.01, -0.0146, -0.0573, -0.01, -0.0143, -0.0548, -0.01, -0.0121, -0.057, -0.01, -0.0112, -0.0538,
-0.0061, -0.0103, -0.0616, -0.0083, -0.0103, -0.0616, -0.0061, -0.0068, -0.0582, -0.0083, -0.0068, -0.0582, -0.0061, -0.009, -0.056, -0.0061, -0.0101, -0.0549, -0.0083, -0.009, -0.056, -0.0083, -0.0101, -0.0549, -0.0061, -0.0079, -0.0571, -0.0083, -0.0079, -0.0571, -0.01, -0.0103, -0.0616, -0.01, -0.01, -0.0591, -0.01, -0.0078, -0.0613, -0.01, -0.0068, -0.0582, -0.01, -0.009, -0.056, -0.01, -0.0101, -0.0549, -0.01, -0.0079, -0.0571, -0.0061, -0.0028, -0.0668, -0.0083, -0.0028, -0.0668, -0.0061, -0.0057, -0.0656, -0.0083, -0.0057, -0.0656, -0.0061, -0.0039, -0.0666, -0.0083, -0.0039, -0.0666, -0.01, -0.0026, -0.0668, -0.01, -0.0059, -0.0655, -0.01, -0.0056, -0.0635, -0.01, -0.0037, -0.0666, -0.01, -0.0035, -0.0635, -0.0061, -0.002, -0.067, -0.0083, -0.002, -0.067,
-0.0061, -0.002, -0.0656, -0.0061, -0.002, -0.0635, -0.0083, -0.002, -0.0656, -0.0083, -0.002, -0.0635, -0.01, -0.002, -0.067, -0.01, -0.002, -0.0656, -0.01, -0.002, -0.0635, -0.0061, -0.0059, -0.0591, -0.0083, -0.0059, -0.0591, -0.0061, -0.0059, -0.057, -0.0061, -0.0059, -0.0548, -0.0083, -0.0059, -0.057, -0.0083, -0.0059, -0.0548, -0.01, -0.0056, -0.0613, -0.01, -0.0059, -0.0591, -0.01, -0.0035, -0.0613, -0.01, -0.0035, -0.0591, -0.01, -0.0059, -0.057, -0.01, -0.0059, -0.0548, -0.01, -0.0035, -0.057, -0.01, -0.0035, -0.0548, -0.0061, -0.002, -0.0613, -0.0061, -0.002, -0.0591, -0.0083, -0.002, -0.0613, -0.0083, -0.002, -0.0591, -0.0061, -0.002, -0.057, -0.0061, -0.002, -0.0548, -0.0083, -0.002, -0.057, -0.0083, -0.002, -0.0548, -0.01, -0.002, -0.0613,
-0.01, -0.002, -0.0591, -0.01, -0.002, -0.057, -0.01, -0.002, -0.0548, -0.0061, -0.0134, -0.0517, -0.0061, -0.0144, -0.0506, -0.0083, -0.0134, -0.0517, -0.0083, -0.0144, -0.0506, -0.0061, -0.0123, -0.0528, -0.0083, -0.0123, -0.0528, -0.01, -0.0134, -0.0517, -0.01, -0.0144, -0.0506, -0.01, -0.0123, -0.0528, -0.0061, -0.0059, -0.0526, -0.0061, -0.0059, -0.0505, -0.0083, -0.0059, -0.0526, -0.0083, -0.0059, -0.0505, -0.0061, -0.0059, -0.0483, -0.0061, -0.0059, -0.0461, -0.0083, -0.0059, -0.0483, -0.0083, -0.0059, -0.0461, -0.01, -0.0059, -0.0526, -0.01, -0.0059, -0.0505, -0.01, -0.0035, -0.0526, -0.01, -0.0035, -0.0505, -0.01, -0.0059, -0.0483, -0.01, -0.0059, -0.0461, -0.01, -0.0035, -0.0483, -0.01, -0.0035, -0.0461, -0.0061, -0.002, -0.0526, -0.0061, -0.002, -0.0505,
-0.0083, -0.002, -0.0526, -0.0083, -0.002, -0.0505, -0.0061, -0.002, -0.0483, -0.0061, -0.002, -0.0461, -0.0083, -0.002, -0.0483, -0.0083, -0.002, -0.0461, -0.01, -0.002, -0.0526, -0.01, -0.002, -0.0505, -0.01, -0.002, -0.0483, -0.01, -0.002, -0.0461, -0.0061, -0.0059, -0.0439, -0.0061, -0.0059, -0.0425, -0.0083, -0.0059, -0.0439, -0.0083, -0.0059, -0.0425, -0.0061, -0.0035, -0.0425, -0.0083, -0.0035, -0.0425, -0.01, -0.0059, -0.0439, -0.01, -0.0059, -0.0425, -0.01, -0.0035, -0.0439, -0.01, -0.0035, -0.0425, -0.0061, -0.002, -0.0439, -0.0061, -0.002, -0.0425, -0.0083, -0.002, -0.0439, -0.0083, -0.002, -0.0425, -0.01, -0.002, -0.0439, -0.01, -0.002, -0.0425, -0.0061, -0.4807, -0.0005, -0.0061, -0.4801, 0.0016, -0.0083, -0.4807, -0.0005, -0.0083, -0.4801, 0.0016,
-0.01, -0.4807, -0.0004, -0.01, -0.4802, 0.0013, -0.0061, -0.4798, 0.0027, -0.0083, -0.4798, 0.0027, -0.0061, -0.4795, 0.0038, -0.0061, -0.4789, 0.006, -0.0083, -0.4795, 0.0038, -0.0083, -0.4789, 0.006, -0.01, -0.4787, -0.0005, -0.01, -0.48, 0.0021, -0.01, -0.4766, -0.0005, -0.01, -0.4766, 0.0016, -0.01, -0.4795, 0.004, -0.01, -0.4789, 0.0061, -0.01, -0.4766, 0.0038, -0.01, -0.4766, 0.006, -0.0061, -0.4784, 0.0081, -0.0061, -0.4779, 0.0102, -0.0083, -0.4784, 0.0081, -0.0083, -0.4779, 0.0102, -0.0061, -0.4776, 0.0112, -0.0083, -0.4776, 0.0112, -0.0061, -0.4772, 0.0125, -0.0061, -0.4765, 0.0147, -0.0083, -0.4772, 0.0125, -0.0083, -0.4765, 0.0147, -0.01, -0.4784, 0.0082, -0.01, -0.478, 0.0099, -0.01, -0.4766, 0.0081, -0.01, -0.4777, 0.0107,
-0.01, -0.4772, 0.0127, -0.01, -0.4765, 0.0148, -0.01, -0.4744, -0.0005, -0.01, -0.4744, 0.0016, -0.01, -0.4722, -0.0005, -0.01, -0.4722, 0.0016, -0.01, -0.4744, 0.0038, -0.01, -0.4744, 0.006, -0.01, -0.4722, 0.0038, -0.01, -0.4722, 0.006, -0.01, -0.47, -0.0005, -0.01, -0.47, 0.0016, -0.01, -0.4679, -0.0005, -0.01, -0.4679, 0.0016, -0.01, -0.47, 0.0038, -0.01, -0.47, 0.006, -0.01, -0.4679, 0.0038, -0.01, -0.4679, 0.006, -0.01, -0.4744, 0.0081, -0.01, -0.4744, 0.0103, -0.01, -0.4722, 0.0081, -0.01, -0.4722, 0.0103, -0.01, -0.4744, 0.0125, -0.01, -0.4744, 0.0146, -0.01, -0.4722, 0.0125, -0.01, -0.4722, 0.0146, -0.01, -0.47, 0.0081, -0.01, -0.47, 0.0103, -0.01, -0.4679, 0.0081, -0.01, -0.4679, 0.0103,
-0.01, -0.47, 0.0125, -0.01, -0.47, 0.0146, -0.01, -0.4679, 0.0125, -0.01, -0.4679, 0.0146, -0.0061, -0.4758, 0.0166, -0.0083, -0.4758, 0.0166, -0.01, -0.4759, 0.0164, -0.0061, -0.4754, 0.0177, -0.0061, -0.4749, 0.019, -0.0083, -0.4754, 0.0177, -0.0083, -0.4749, 0.019, -0.0061, -0.4742, 0.0212, -0.0061, -0.4736, 0.0229, -0.0083, -0.4742, 0.0212, -0.0083, -0.4736, 0.0229, -0.0061, -0.4732, 0.024, -0.0083, -0.4732, 0.024, -0.01, -0.4755, 0.0173, -0.01, -0.4749, 0.0192, -0.01, -0.4722, 0.0168, -0.01, -0.4722, 0.019, -0.01, -0.4742, 0.0213, -0.01, -0.4736, 0.0228, -0.01, -0.4722, 0.0212, -0.01, -0.4733, 0.0237, -0.01, -0.47, 0.0168, -0.01, -0.47, 0.019, -0.01, -0.4679, 0.0168, -0.01, -0.4679, 0.019, -0.01, -0.47, 0.0212,
-0.01, -0.47, 0.0233, -0.01, -0.4679, 0.0212, -0.01, -0.4679, 0.0233, -0.0061, -0.4726, 0.0255, -0.0061, -0.4716, 0.0277, -0.0083, -0.4726, 0.0255, -0.0083, -0.4716, 0.0277, -0.0061, -0.4711, 0.0288, -0.0083, -0.4711, 0.0288, -0.01, -0.4724, 0.0258, -0.01, -0.4716, 0.0278, -0.01, -0.4711, 0.0288, -0.0061, -0.4707, 0.0299, -0.0061, -0.4698, 0.032, -0.0083, -0.4707, 0.0299, -0.0083, -0.4698, 0.032, -0.01, -0.47, 0.0255, -0.01, -0.47, 0.0277, -0.01, -0.4679, 0.0255, -0.01, -0.4679, 0.0277, -0.01, -0.4707, 0.0299, -0.01, -0.4697, 0.0322, -0.01, -0.4679, 0.0298, -0.01, -0.4679, 0.032, -0.0061, -0.4628, -0.0005, -0.0061, -0.4625, 0.0006, -0.0083, -0.4628, -0.0005, -0.0083, -0.4625, 0.0006, -0.01, -0.4657, -0.0005, -0.01, -0.4657, 0.0016,
-0.01, -0.4628, -0.0008, -0.01, -0.4623, 0.0012, -0.01, -0.4657, 0.0038, -0.01, -0.4657, 0.006, -0.01, -0.4635, 0.0038, -0.01, -0.4635, 0.006, -0.0061, -0.4622, 0.0017, -0.0083, -0.4622, 0.0017, -0.0061, -0.4615, 0.0038, -0.0061, -0.4608, 0.006, -0.0083, -0.4615, 0.0038, -0.0083, -0.4608, 0.006, -0.01, -0.4621, 0.002, -0.01, -0.4616, 0.0037, -0.01, -0.4609, 0.0057, -0.01, -0.4657, 0.0081, -0.01, -0.4657, 0.0103, -0.01, -0.4635, 0.0081, -0.01, -0.4635, 0.0103, -0.01, -0.4657, 0.0125, -0.01, -0.4657, 0.0146, -0.01, -0.4635, 0.0125, -0.01, -0.4635, 0.0146, -0.0061, -0.4604, 0.0074, -0.0083, -0.4604, 0.0074, -0.0061, -0.4601, 0.0085, -0.0061, -0.4595, 0.0103, -0.0083, -0.4601, 0.0085, -0.0083, -0.4595, 0.0103, -0.0061, -0.4587, 0.0125,
-0.0061, -0.4582, 0.0138, -0.0083, -0.4587, 0.0125, -0.0083, -0.4582, 0.0138, -0.01, -0.4603, 0.0078, -0.01, -0.4614, 0.0103, -0.01, -0.46, 0.0087, -0.01, -0.4595, 0.0102, -0.01, -0.4614, 0.0125, -0.01, -0.4614, 0.0146, -0.01, -0.4588, 0.0122, -0.01, -0.4581, 0.0142, -0.0061, -0.4578, 0.0149, -0.0083, -0.4578, 0.0149, -0.01, -0.4577, 0.0151, -0.01, -0.4657, 0.0168, -0.01, -0.4657, 0.019, -0.01, -0.4635, 0.0168, -0.01, -0.4635, 0.019, -0.01, -0.4657, 0.0212, -0.01, -0.4657, 0.0233, -0.01, -0.4635, 0.0212, -0.01, -0.4635, 0.0233, -0.01, -0.4614, 0.0168, -0.01, -0.4614, 0.019, -0.01, -0.4592, 0.0168, -0.01, -0.4592, 0.019, -0.01, -0.4614, 0.0212, -0.01, -0.4614, 0.0233, -0.01, -0.4592, 0.0212, -0.01, -0.4592, 0.0233,
-0.01, -0.4657, 0.0255, -0.01, -0.4657, 0.0277, -0.01, -0.4635, 0.0255, -0.01, -0.4635, 0.0277, -0.01, -0.4657, 0.0298, -0.01, -0.4657, 0.032, -0.01, -0.4635, 0.0298, -0.01, -0.4635, 0.032, -0.01, -0.4614, 0.0255, -0.01, -0.4614, 0.0277, -0.01, -0.4592, 0.0255, -0.01, -0.4592, 0.0277, -0.01, -0.4614, 0.0298, -0.01, -0.4614, 0.032, -0.01, -0.4592, 0.0298, -0.01, -0.4592, 0.032, -0.0061, -0.4571, 0.0168, -0.0061, -0.4563, 0.019, -0.0083, -0.4571, 0.0168, -0.0083, -0.4563, 0.019, -0.0061, -0.4559, 0.0201, -0.0083, -0.4559, 0.0201, -0.0061, -0.4556, 0.0212, -0.0061, -0.4548, 0.0233, -0.0083, -0.4556, 0.0212, -0.0083, -0.4548, 0.0233, -0.01, -0.4572, 0.0167, -0.01, -0.4564, 0.0187, -0.01, -0.4558, 0.0206, -0.01, -0.457, 0.0233,
-0.01, -0.4555, 0.0214, -0.01, -0.4549, 0.0231, -0.0061, -0.4541, 0.0251, -0.0083, -0.4541, 0.0251, -0.01, -0.457, 0.0255, -0.01, -0.457, 0.0277, -0.01, -0.454, 0.0252, -0.01, -0.4549, 0.0277, -0.01, -0.457, 0.0298, -0.01, -0.457, 0.032, -0.01, -0.4549, 0.0298, -0.01, -0.4549, 0.032, -0.0061, -0.4536, 0.0262, -0.0061, -0.4529, 0.0277, -0.0083, -0.4536, 0.0262, -0.0083, -0.4529, 0.0277, -0.0061, -0.452, 0.0295, -0.0083, -0.452, 0.0295, -0.0061, -0.4514, 0.0306, -0.0061, -0.4507, 0.032, -0.0083, -0.4514, 0.0306, -0.0083, -0.4507, 0.032, -0.01, -0.4536, 0.0263, -0.01, -0.453, 0.0275, -0.01, -0.4519, 0.0296, -0.01, -0.4527, 0.032, -0.01, -0.4514, 0.0307, -0.01, -0.4508, 0.0318, -0.0061, -0.4691, 0.0334, -0.0083, -0.4691, 0.0334,
-0.0061, -0.4686, 0.0345, -0.0061, -0.4677, 0.0364, -0.0083, -0.4686, 0.0345, -0.0083, -0.4677, 0.0364, -0.0061, -0.467, 0.0378, -0.0083, -0.467, 0.0378, -0.01, -0.4691, 0.0333, -0.01, -0.4686, 0.0344, -0.01, -0.4676, 0.0365, -0.01, -0.467, 0.0377, -0.0061, -0.4664, 0.0389, -0.0061, -0.4655, 0.0407, -0.0083, -0.4664, 0.0389, -0.0083, -0.4655, 0.0407, -0.01, -0.4657, 0.0342, -0.01, -0.4657, 0.0363, -0.01, -0.4635, 0.0342, -0.01, -0.4635, 0.0363, -0.01, -0.4664, 0.0388, -0.01, -0.4654, 0.0409, -0.01, -0.4635, 0.0385, -0.01, -0.4635, 0.0407, -0.01, -0.4614, 0.0342, -0.01, -0.4614, 0.0363, -0.01, -0.4592, 0.0342, -0.01, -0.4592, 0.0363, -0.01, -0.4614, 0.0385, -0.01, -0.4614, 0.0407, -0.01, -0.4592, 0.0385, -0.01, -0.4592, 0.0407,
-0.0061, -0.4647, 0.042, -0.0083, -0.4647, 0.042, -0.0061, -0.4641, 0.0431, -0.0061, -0.463, 0.0449, -0.0083, -0.4641, 0.0431, -0.0083, -0.463, 0.0449, -0.01, -0.4648, 0.042, -0.01, -0.4641, 0.0431, -0.01, -0.4631, 0.0448, -0.0061, -0.4624, 0.046, -0.0083, -0.4624, 0.046, -0.0061, -0.4617, 0.0472, -0.0061, -0.4606, 0.0488, -0.0083, -0.4617, 0.0472, -0.0083, -0.4606, 0.0488, -0.0061, -0.4599, 0.0499, -0.0083, -0.4599, 0.0499, -0.01, -0.4614, 0.0429, -0.01, -0.4625, 0.0458, -0.01, -0.4592, 0.0429, -0.01, -0.4592, 0.045, -0.01, -0.4615, 0.0475, -0.01, -0.4607, 0.0488, -0.01, -0.4592, 0.0472, -0.01, -0.46, 0.0498, -0.01, -0.457, 0.0342, -0.01, -0.457, 0.0363, -0.01, -0.4549, 0.0342, -0.01, -0.4549, 0.0363, -0.01, -0.457, 0.0385,
-0.01, -0.457, 0.0407, -0.01, -0.4549, 0.0385, -0.01, -0.4549, 0.0407, -0.0061, -0.4498, 0.0339, -0.0083, -0.4498, 0.0339, -0.0061, -0.4517, 0.0417, -0.0083, -0.4517, 0.0417, -0.0061, -0.4498, 0.0392, -0.0061, -0.4505, 0.0408, -0.0083, -0.4498, 0.0392, -0.0083, -0.4505, 0.0408, -0.01, -0.4527, 0.0342, -0.01, -0.4527, 0.0363, -0.01, -0.4498, 0.0339, -0.01, -0.4505, 0.0363, -0.01, -0.4527, 0.0385, -0.01, -0.4515, 0.0415, -0.01, -0.4499, 0.0393, -0.01, -0.4505, 0.0408, -0.01, -0.457, 0.0429, -0.01, -0.457, 0.045, -0.01, -0.4549, 0.0429, -0.01, -0.4549, 0.045, -0.01, -0.457, 0.0472, -0.01, -0.457, 0.0494, -0.01, -0.4549, 0.0472, -0.01, -0.4549, 0.0494, -0.0061, -0.4523, 0.0426, -0.0061, -0.452, 0.0452, -0.0083, -0.4523, 0.0426,
-0.0083, -0.452, 0.0452, -0.0061, -0.4512, 0.0429, -0.0061, -0.4515, 0.0447, -0.0083, -0.4512, 0.0429, -0.0083, -0.4515, 0.0447, -0.0061, -0.4519, 0.0469, -0.0061, -0.4522, 0.0494, -0.0083, -0.4519, 0.0469, -0.0083, -0.4522, 0.0494, -0.0061, -0.4516, 0.0473, -0.0061, -0.4509, 0.0494, -0.0083, -0.4516, 0.0473, -0.0083, -0.4509, 0.0494, -0.01, -0.4525, 0.0431, -0.01, -0.452, 0.0453, -0.01, -0.451, 0.0425, -0.01, -0.4515, 0.045, -0.01, -0.452, 0.0468, -0.01, -0.4523, 0.0492, -0.01, -0.4516, 0.0472, -0.01, -0.4508, 0.0496, -0.0061, -0.4589, 0.0515, -0.0083, -0.4589, 0.0515, -0.01, -0.459, 0.0514, -0.0061, -0.4581, 0.0526, -0.0061, -0.4573, 0.0537, -0.0083, -0.4581, 0.0526, -0.0083, -0.4573, 0.0537, -0.0061, -0.4562, 0.0551, -0.0083, -0.4562, 0.0551,
-0.0061, -0.4553, 0.0562, -0.0061, -0.4543, 0.0576, -0.0083, -0.4553, 0.0562, -0.0083, -0.4543, 0.0576, -0.01, -0.4582, 0.0524, -0.01, -0.457, 0.054, -0.01, -0.4549, 0.0515, -0.01, -0.4549, 0.0537, -0.01, -0.4562, 0.0551, -0.01, -0.4553, 0.0562, -0.01, -0.4543, 0.0575, -0.0061, -0.4517, 0.0505, -0.0083, -0.4517, 0.0505, -0.0061, -0.4499, 0.0521, -0.0061, -0.4497, 0.0529, -0.0083, -0.4499, 0.0521, -0.0083, -0.4497, 0.0529, -0.0061, -0.4534, 0.0586, -0.0083, -0.4534, 0.0586, -0.01, -0.4515, 0.0508, -0.01, -0.4527, 0.0537, -0.01, -0.4504, 0.0514, -0.01, -0.4498, 0.0529, -0.01, -0.4527, 0.0559, -0.01, -0.4534, 0.0586, -0.01, -0.4505, 0.0559, -0.01, -0.4505, 0.058, -0.0061, -0.4523, 0.0599, -0.0083, -0.4523, 0.0599, -0.0061, -0.4513, 0.061,
-0.0061, -0.4502, 0.0621, -0.0083, -0.4513, 0.061, -0.0083, -0.4502, 0.0621, -0.01, -0.4523, 0.0599, -0.01, -0.4513, 0.061, -0.01, -0.4502, 0.0621, -0.0061, -0.4322, 0.0045, -0.0061, -0.4328, 0.0059, -0.0083, -0.4322, 0.0045, -0.0083, -0.4328, 0.0059, -0.01, -0.4323, 0.0046, -0.01, -0.4328, 0.0058, -0.0061, -0.4332, 0.0081, -0.0061, -0.4333, 0.0103, -0.0083, -0.4332, 0.0081, -0.0083, -0.4333, 0.0103, -0.0061, -0.4333, 0.0125, -0.0061, -0.4331, 0.0147, -0.0083, -0.4333, 0.0125, -0.0083, -0.4331, 0.0147, -0.01, -0.4332, 0.0081, -0.01, -0.4333, 0.0103, -0.01, -0.4333, 0.0125, -0.01, -0.4331, 0.0147, -0.0061, -0.4325, 0.0168, -0.0083, -0.4325, 0.0168, -0.01, -0.4326, 0.0165, -0.0061, -0.4306, 0.0019, -0.0083, -0.4306, 0.0019, -0.0061, -0.4285, -0.0001,
-0.0061, -0.4296, 0.0009, -0.0083, -0.4285, -0.0001, -0.0083, -0.4296, 0.0009, -0.0061, -0.4317, 0.0034, -0.0083, -0.4317, 0.0034, -0.01, -0.4306, 0.002, -0.01, -0.4284, -0.0001, -0.01, -0.4295, 0.0008, -0.01, -0.4317, 0.0035, -0.01, -0.431, 0.006, -0.01, -0.4288, 0.0038, -0.01, -0.4288, 0.006, -0.0061, -0.4269, -0.0012, -0.0083, -0.4269, -0.0012, -0.01, -0.4269, -0.0012, -0.01, -0.4266, 0.0016, -0.01, -0.4245, -0.0005, -0.01, -0.4245, 0.0016, -0.01, -0.4266, 0.0038, -0.01, -0.4266, 0.006, -0.01, -0.4245, 0.0038, -0.01, -0.4245, 0.006, -0.01, -0.431, 0.0081, -0.01, -0.431, 0.0103, -0.01, -0.4288, 0.0081, -0.01, -0.4288, 0.0103, -0.01, -0.431, 0.0125, -0.01, -0.431, 0.0146, -0.01, -0.4288, 0.0125, -0.01, -0.4288, 0.0146,
-0.01, -0.4266, 0.0081, -0.01, -0.4266, 0.0103, -0.01, -0.4245, 0.0081, -0.01, -0.4245, 0.0103, -0.01, -0.4266, 0.0125, -0.01, -0.4266, 0.0146, -0.01, -0.4245, 0.0125, -0.01, -0.4245, 0.0146, -0.01, -0.4223, -0.0005, -0.01, -0.4223, 0.0016, -0.01, -0.4201, -0.0005, -0.01, -0.4201, 0.0016, -0.01, -0.4223, 0.0038, -0.01, -0.4223, 0.006, -0.01, -0.4201, 0.0038, -0.01, -0.4201, 0.006, -0.01, -0.418, -0.0005, -0.01, -0.418, 0.0016, -0.01, -0.4158, -0.0005, -0.01, -0.4158, 0.0016, -0.01, -0.418, 0.0038, -0.01, -0.418, 0.006, -0.01, -0.4158, 0.0038, -0.01, -0.4158, 0.006, -0.01, -0.4223, 0.0081, -0.01, -0.4223, 0.0103, -0.01, -0.4201, 0.0081, -0.01, -0.4201, 0.0103, -0.01, -0.4223, 0.0125, -0.01, -0.4223, 0.0146,
-0.01, -0.4201, 0.0125, -0.01, -0.4201, 0.0146, -0.01, -0.418, 0.0081, -0.01, -0.418, 0.0103, -0.01, -0.4158, 0.0081, -0.01, -0.4158, 0.0103, -0.01, -0.418, 0.0125, -0.01, -0.418, 0.0146, -0.01, -0.4158, 0.0125, -0.01, -0.4158, 0.0146, -0.0061, -0.4321, 0.0179, -0.0061, -0.4315, 0.0191, -0.0083, -0.4321, 0.0179, -0.0083, -0.4315, 0.0191, -0.0061, -0.4304, 0.0206, -0.0083, -0.4304, 0.0206, -0.0061, -0.4293, 0.0217, -0.0061, -0.4282, 0.0226, -0.0083, -0.4293, 0.0217, -0.0083, -0.4282, 0.0226, -0.01, -0.4322, 0.0175, -0.01, -0.4314, 0.0194, -0.01, -0.4288, 0.0168, -0.01, -0.4288, 0.019, -0.01, -0.4304, 0.0206, -0.01, -0.4293, 0.0217, -0.01, -0.4281, 0.0226, -0.0061, -0.4267, 0.0236, -0.0083, -0.4267, 0.0236, -0.0061, -0.4249, 0.0242,
-0.0083, -0.4249, 0.0242, -0.01, -0.4266, 0.0168, -0.01, -0.4266, 0.019, -0.01, -0.4245, 0.0168, -0.01, -0.4245, 0.019, -0.01, -0.4266, 0.0212, -0.01, -0.427, 0.0234, -0.01, -0.4245, 0.0212, -0.01, -0.4247, 0.0243, -0.0061, -0.4238, 0.0246, -0.0083, -0.4238, 0.0246, -0.01, -0.4237, 0.0246, -0.01, -0.4223, 0.0168, -0.01, -0.4223, 0.019, -0.01, -0.4201, 0.0168, -0.01, -0.4201, 0.019, -0.01, -0.4223, 0.0212, -0.01, -0.4223, 0.0233, -0.01, -0.4201, 0.0212, -0.01, -0.4201, 0.0233, -0.0061, -0.4148, 0.0244, -0.0083, -0.4148, 0.0244, -0.01, -0.418, 0.0168, -0.01, -0.418, 0.019, -0.01, -0.4158, 0.0168, -0.01, -0.4158, 0.019, -0.01, -0.418, 0.0212, -0.01, -0.418, 0.0233, -0.01, -0.4158, 0.0212, -0.01, -0.4153, 0.0246,
-0.0061, -0.4223, 0.0249, -0.0083, -0.4223, 0.0249, -0.0061, -0.4201, 0.0251, -0.0083, -0.4201, 0.0251, -0.01, -0.4223, 0.0249, -0.01, -0.4201, 0.0251, -0.0061, -0.418, 0.025, -0.0083, -0.418, 0.025, -0.0061, -0.4159, 0.0247, -0.0083, -0.4159, 0.0247, -0.01, -0.4179, 0.025, -0.01, -0.4162, 0.0247, -0.0061, -0.4493, 0.035, -0.0061, -0.4487, 0.0363, -0.0083, -0.4493, 0.035, -0.0083, -0.4487, 0.0363, -0.0061, -0.4478, 0.0387, -0.0061, -0.4483, 0.0407, -0.0083, -0.4478, 0.0387, -0.0083, -0.4483, 0.0407, -0.0061, -0.4462, 0.0385, -0.0061, -0.4471, 0.0414, -0.0083, -0.4462, 0.0385, -0.0083, -0.4471, 0.0414, -0.01, -0.4493, 0.035, -0.01, -0.4487, 0.0362, -0.01, -0.4478, 0.0387, -0.01, -0.4483, 0.0407, -0.01, -0.4463, 0.0385, -0.01, -0.4464, 0.0407,
-0.0061, -0.444, 0.0384, -0.0083, -0.444, 0.0384, -0.0061, -0.4418, 0.0388, -0.0083, -0.4418, 0.0388, -0.01, -0.444, 0.0384, -0.01, -0.444, 0.0407, -0.01, -0.4417, 0.0388, -0.01, -0.4418, 0.0407, -0.0061, -0.4476, 0.0419, -0.0083, -0.4476, 0.0419, -0.0061, -0.4471, 0.0419, -0.0083, -0.4471, 0.0419, -0.01, -0.4483, 0.0426, -0.01, -0.4483, 0.045, -0.01, -0.4471, 0.0419, -0.01, -0.4462, 0.045, -0.01, -0.4483, 0.0472, -0.01, -0.4483, 0.0494, -0.01, -0.4462, 0.0472, -0.01, -0.4462, 0.0494, -0.01, -0.444, 0.0429, -0.01, -0.444, 0.045, -0.01, -0.4418, 0.0429, -0.01, -0.4418, 0.045, -0.01, -0.444, 0.0472, -0.01, -0.444, 0.0494, -0.01, -0.4418, 0.0472, -0.01, -0.4418, 0.0494, -0.0061, -0.44, 0.0393, -0.0061, -0.439, 0.0398,
-0.0083, -0.44, 0.0393, -0.0083, -0.439, 0.0398, -0.0061, -0.438, 0.0402, -0.0083, -0.438, 0.0402, -0.01, -0.4402, 0.0392, -0.01, -0.4392, 0.0397, -0.01, -0.4375, 0.0406, -0.0061, -0.4365, 0.0419, -0.0083, -0.4365, 0.0419, -0.0061, -0.4364, 0.0504, -0.0083, -0.4364, 0.0504, -0.01, -0.4397, 0.0429, -0.01, -0.4397, 0.045, -0.01, -0.4364, 0.0421, -0.01, -0.4375, 0.045, -0.01, -0.4397, 0.0472, -0.01, -0.4397, 0.0494, -0.01, -0.4375, 0.0472, -0.01, -0.4363, 0.0501, -0.0061, -0.4359, 0.043, -0.0061, -0.4355, 0.045, -0.0083, -0.4359, 0.043, -0.0083, -0.4355, 0.045, -0.0061, -0.4354, 0.0472, -0.0061, -0.4359, 0.0493, -0.0083, -0.4354, 0.0472, -0.0083, -0.4359, 0.0493, -0.01, -0.4359, 0.0433, -0.01, -0.4355, 0.045, -0.01, -0.4354, 0.0473,
-0.01, -0.4358, 0.0491, -0.0061, -0.4489, 0.0521, -0.0061, -0.4481, 0.0535, -0.0083, -0.4489, 0.0521, -0.0083, -0.4481, 0.0535, -0.0061, -0.4462, 0.0538, -0.0083, -0.4462, 0.0538, -0.0061, -0.4473, 0.0548, -0.0083, -0.4473, 0.0548, -0.0061, -0.4461, 0.0556, -0.0083, -0.4461, 0.0556, -0.01, -0.4488, 0.0521, -0.01, -0.4481, 0.0535, -0.01, -0.4462, 0.0515, -0.01, -0.4464, 0.0538, -0.01, -0.4475, 0.0547, -0.01, -0.4483, 0.058, -0.01, -0.446, 0.0556, -0.01, -0.4462, 0.058, -0.0061, -0.444, 0.0543, -0.0083, -0.444, 0.0543, -0.0061, -0.4418, 0.0541, -0.0083, -0.4418, 0.0541, -0.0061, -0.444, 0.0559, -0.0083, -0.444, 0.0559, -0.0061, -0.4418, 0.0557, -0.0083, -0.4418, 0.0557, -0.01, -0.444, 0.0515, -0.01, -0.444, 0.0543, -0.01, -0.4418, 0.0515,
-0.01, -0.4417, 0.0541, -0.01, -0.444, 0.0559, -0.01, -0.444, 0.058, -0.01, -0.4419, 0.0558, -0.01, -0.4418, 0.058, -0.0061, -0.4491, 0.0632, -0.0083, -0.4491, 0.0632, -0.0061, -0.4481, 0.0642, -0.0083, -0.4481, 0.0642, -0.0061, -0.4468, 0.0653, -0.0061, -0.4457, 0.0662, -0.0083, -0.4468, 0.0653, -0.0083, -0.4457, 0.0662, -0.01, -0.4483, 0.0602, -0.01, -0.4491, 0.0632, -0.01, -0.4462, 0.0602, -0.01, -0.4462, 0.0624, -0.01, -0.448, 0.0642, -0.01, -0.4468, 0.0653, -0.01, -0.4457, 0.0662, -0.0061, -0.4444, 0.0673, -0.0083, -0.4444, 0.0673, -0.01, -0.444, 0.0602, -0.01, -0.444, 0.0624, -0.01, -0.4418, 0.0602, -0.01, -0.4418, 0.0624, -0.01, -0.444, 0.0646, -0.01, -0.4444, 0.0673, -0.01, -0.4418, 0.0646, -0.01, -0.4418, 0.0667,
-0.0061, -0.4395, 0.0544, -0.0083, -0.4395, 0.0544, -0.0061, -0.4373, 0.0518, -0.0061, -0.4378, 0.0539, -0.0083, -0.4373, 0.0518, -0.0083, -0.4378, 0.0539, -0.0061, -0.44, 0.0551, -0.0083, -0.44, 0.0551, -0.01, -0.4397, 0.0515, -0.01, -0.4393, 0.0543, -0.01, -0.4375, 0.052, -0.0097, -0.4378, 0.054, -0.01, -0.4398, 0.055, -0.01, -0.4397, 0.058, -0.01, -0.4375, 0.0559, -0.01, -0.4375, 0.058, -0.0061, -0.4358, 0.0545, -0.0083, -0.4358, 0.0545, -0.0061, -0.4347, 0.0549, -0.0083, -0.4347, 0.0549, -0.0061, -0.4331, 0.0558, -0.0083, -0.4331, 0.0558, -0.01, -0.436, 0.0545, -0.01, -0.4349, 0.0548, -0.01, -0.4353, 0.058, -0.01, -0.4329, 0.056, -0.01, -0.4332, 0.058, -0.01, -0.4397, 0.0602, -0.01, -0.4397, 0.0624, -0.01, -0.4375, 0.0602,
-0.01, -0.4375, 0.0624, -0.01, -0.4397, 0.0646, -0.01, -0.4397, 0.0667, -0.01, -0.4375, 0.0646, -0.01, -0.4375, 0.0667, -0.01, -0.4353, 0.0602, -0.01, -0.4353, 0.0624, -0.01, -0.4332, 0.0602, -0.01, -0.4332, 0.0624, -0.01, -0.4353, 0.0646, -0.01, -0.4353, 0.0667, -0.01, -0.4332, 0.0646, -0.01, -0.4332, 0.0667, -0.0061, -0.4318, 0.0568, -0.0061, -0.4307, 0.0576, -0.0083, -0.4318, 0.0568, -0.0083, -0.4307, 0.0576, -0.0061, -0.4294, 0.0587, -0.0083, -0.4294, 0.0587, -0.01, -0.4319, 0.0568, -0.01, -0.4307, 0.0576, -0.01, -0.4294, 0.0587, -0.0061, -0.4283, 0.0596, -0.0083, -0.4283, 0.0596, -0.01, -0.431, 0.0602, -0.01, -0.431, 0.0624, -0.01, -0.4283, 0.0596, -0.01, -0.4288, 0.0624, -0.01, -0.431, 0.0646, -0.01, -0.431, 0.0667,
-0.01, -0.4288, 0.0646, -0.01, -0.4288, 0.0667, -0.0061, -0.427, 0.0607, -0.0061, -0.4259, 0.0616, -0.0083, -0.427, 0.0607, -0.0083, -0.4259, 0.0616, -0.0061, -0.4245, 0.0627, -0.0083, -0.4245, 0.0627, -0.01, -0.4271, 0.0607, -0.01, -0.426, 0.0615, -0.01, -0.4243, 0.0628, -0.01, -0.4266, 0.0646, -0.01, -0.4266, 0.0667, -0.01, -0.4245, 0.0646, -0.01, -0.4245, 0.0667, -0.0061, -0.4234, 0.0635, -0.0083, -0.4234, 0.0635, -0.0061, -0.4223, 0.0642, -0.0083, -0.4223, 0.0642, -0.0061, -0.4206, 0.0653, -0.0061, -0.4195, 0.066, -0.0083, -0.4206, 0.0653, -0.0083, -0.4195, 0.066, -0.01, -0.4234, 0.0635, -0.01, -0.4222, 0.0642, -0.01, -0.4223, 0.0667, -0.01, -0.4207, 0.0652, -0.01, -0.4197, 0.0659, -0.0061, -0.418, 0.0669, -0.0083, -0.418, 0.0669,
-0.0061, -0.4167, 0.0677, -0.0083, -0.4167, 0.0677, -0.01, -0.4178, 0.067, -0.01, -0.4167, 0.0677, -0.0061, -0.4434, 0.0682, -0.0083, -0.4434, 0.0682, -0.0061, -0.442, 0.0693, -0.0061, -0.441, 0.0701, -0.0083, -0.442, 0.0693, -0.0083, -0.441, 0.0701, -0.01, -0.4433, 0.0682, -0.01, -0.4421, 0.0693, -0.01, -0.4409, 0.0702, -0.0061, -0.4397, 0.0711, -0.0083, -0.4397, 0.0711, -0.0061, -0.4384, 0.0721, -0.0083, -0.4384, 0.0721, -0.0061, -0.4374, 0.0728, -0.0083, -0.4374, 0.0728, -0.01, -0.4397, 0.0689, -0.01, -0.4399, 0.0709, -0.01, -0.4375, 0.0689, -0.01, -0.4383, 0.0721, -0.01, -0.4373, 0.0729, -0.0061, -0.4358, 0.0739, -0.0061, -0.4347, 0.0746, -0.0083, -0.4358, 0.0739, -0.0083, -0.4347, 0.0746, -0.0061, -0.4332, 0.0755, -0.0083, -0.4332, 0.0755,
-0.01, -0.4353, 0.0689, -0.01, -0.4353, 0.0711, -0.01, -0.4332, 0.0689, -0.01, -0.4332, 0.0711, -0.01, -0.4357, 0.0739, -0.01, -0.4346, 0.0747, -0.01, -0.4332, 0.0732, -0.01, -0.4334, 0.0754, -0.0061, -0.4321, 0.0868, -0.0061, -0.4322, 0.0882, -0.0083, -0.4321, 0.0868, -0.0083, -0.4322, 0.0882, -0.01, -0.4321, 0.087, -0.01, -0.4322, 0.0879, -0.0061, -0.4309, 0.0763, -0.0083, -0.4309, 0.0763, -0.0061, -0.4288, 0.0758, -0.0083, -0.4288, 0.0758, -0.01, -0.431, 0.0689, -0.01, -0.431, 0.0711, -0.01, -0.4288, 0.0689, -0.01, -0.4288, 0.0711, -0.01, -0.431, 0.0732, -0.01, -0.4309, 0.0763, -0.01, -0.4288, 0.0732, -0.01, -0.4286, 0.0757, -0.0061, -0.4257, 0.0742, -0.0061, -0.4266, 0.0753, -0.0083, -0.4257, 0.0742, -0.0083, -0.4266, 0.0753,
-0.0061, -0.4242, 0.0742, -0.0061, -0.4246, 0.0748, -0.0083, -0.4242, 0.0742, -0.0083, -0.4246, 0.0748, -0.01, -0.4266, 0.0689, -0.01, -0.4266, 0.0711, -0.01, -0.4245, 0.0689, -0.01, -0.4245, 0.0711, -0.01, -0.426, 0.0744, -0.01, -0.4266, 0.0754, -0.01, -0.4239, 0.0741, -0.01, -0.4246, 0.0748, -0.0061, -0.4303, 0.0801, -0.0083, -0.4303, 0.0801, -0.0061, -0.4286, 0.0777, -0.0061, -0.4296, 0.079, -0.0083, -0.4286, 0.0777, -0.0083, -0.4296, 0.079, -0.0061, -0.4311, 0.0819, -0.0061, -0.4318, 0.0841, -0.0083, -0.4311, 0.0819, -0.0083, -0.4318, 0.0841, -0.0061, -0.4278, 0.0829, -0.0061, -0.4277, 0.083, -0.0083, -0.4278, 0.0829, -0.0083, -0.4277, 0.083, -0.01, -0.4304, 0.0802, -0.01, -0.4286, 0.0777, -0.01, -0.4297, 0.0791, -0.01, -0.4311, 0.0817,
-0.01, -0.4317, 0.0839, -0.01, -0.4278, 0.0829, -0.01, -0.4276, 0.0836, -0.0061, -0.4275, 0.0767, -0.0061, -0.4264, 0.0798, -0.0083, -0.4275, 0.0767, -0.0083, -0.4264, 0.0798, -0.0061, -0.4246, 0.0778, -0.0061, -0.4248, 0.0794, -0.0083, -0.4246, 0.0778, -0.0083, -0.4248, 0.0794, -0.0061, -0.4285, 0.0847, -0.0061, -0.4275, 0.084, -0.0083, -0.4285, 0.0847, -0.0083, -0.4275, 0.084, -0.0061, -0.4253, 0.0811, -0.0083, -0.4253, 0.0811, -0.01, -0.4275, 0.0766, -0.01, -0.4264, 0.0798, -0.01, -0.4246, 0.0778, -0.01, -0.4248, 0.0794, -0.01, -0.4285, 0.0847, -0.01, -0.4275, 0.084, -0.01, -0.4253, 0.0811, -0.01, -0.4245, 0.0841, -0.01, -0.4223, 0.0689, -0.01, -0.4223, 0.0711, -0.01, -0.4201, 0.0689, -0.01, -0.4201, 0.0711, -0.01, -0.4222, 0.0738,
-0.01, -0.4223, 0.0754, -0.01, -0.4201, 0.0734, -0.01, -0.4201, 0.0754, -0.0061, -0.4156, 0.0683, -0.0083, -0.4156, 0.0683, -0.01, -0.418, 0.0689, -0.01, -0.418, 0.0711, -0.01, -0.4156, 0.0683, -0.01, -0.4158, 0.0711, -0.01, -0.418, 0.0733, -0.01, -0.418, 0.0754, -0.01, -0.4158, 0.0736, -0.01, -0.4158, 0.0754, -0.0061, -0.4224, 0.078, -0.0061, -0.4229, 0.079, -0.0083, -0.4224, 0.078, -0.0083, -0.4229, 0.079, -0.01, -0.422, 0.0778, -0.01, -0.4229, 0.0789, -0.01, -0.4204, 0.0775, -0.01, -0.4201, 0.0798, -0.01, -0.4223, 0.0819, -0.01, -0.4223, 0.0841, -0.01, -0.4201, 0.0819, -0.01, -0.4201, 0.0841, -0.01, -0.418, 0.0776, -0.01, -0.418, 0.0798, -0.01, -0.4158, 0.0776, -0.01, -0.4158, 0.0798, -0.01, -0.418, 0.0819,
-0.01, -0.418, 0.0841, -0.01, -0.4158, 0.0819, -0.01, -0.4158, 0.0841, -0.0061, -0.432, 0.0858, -0.0061, -0.432, 0.0893, -0.0083, -0.432, 0.0858, -0.0083, -0.432, 0.0893, -0.0061, -0.4318, 0.0906, -0.0061, -0.431, 0.0928, -0.0083, -0.4318, 0.0906, -0.0083, -0.431, 0.0928, -0.01, -0.432, 0.0862, -0.01, -0.4321, 0.0887, -0.01, -0.4288, 0.0863, -0.01, -0.4288, 0.0884, -0.01, -0.4317, 0.0909, -0.01, -0.4309, 0.093, -0.01, -0.4288, 0.0906, -0.01, -0.4288, 0.0928, -0.01, -0.4266, 0.0863, -0.01, -0.4266, 0.0884, -0.01, -0.4245, 0.0863, -0.01, -0.4245, 0.0884, -0.01, -0.4266, 0.0906, -0.01, -0.4266, 0.0928, -0.01, -0.4245, 0.0906, -0.01, -0.4245, 0.0928, -0.0061, -0.4301, 0.0942, -0.0083, -0.4301, 0.0942, -0.0061, -0.4293, 0.0953,
-0.0061, -0.4282, 0.0965, -0.0083, -0.4293, 0.0953, -0.0083, -0.4282, 0.0965, -0.01, -0.4301, 0.0941, -0.01, -0.4293, 0.0953, -0.01, -0.4282, 0.0965, -0.0061, -0.427, 0.0976, -0.0083, -0.427, 0.0976, -0.0061, -0.4259, 0.0984, -0.0083, -0.4259, 0.0984, -0.0061, -0.4245, 0.0993, -0.0083, -0.4245, 0.0993, -0.01, -0.4266, 0.0949, -0.01, -0.427, 0.0976, -0.01, -0.4245, 0.0949, -0.01, -0.4245, 0.0971, -0.01, -0.4258, 0.0984, -0.01, -0.4247, 0.0992, -0.01, -0.4223, 0.0863, -0.01, -0.4223, 0.0884, -0.01, -0.4201, 0.0863, -0.01, -0.4201, 0.0884, -0.01, -0.4223, 0.0906, -0.01, -0.4223, 0.0928, -0.01, -0.4201, 0.0906, -0.01, -0.4201, 0.0928, -0.01, -0.418, 0.0863, -0.01, -0.418, 0.0884, -0.01, -0.4158, 0.0863, -0.01, -0.4158, 0.0884,
-0.01, -0.418, 0.0906, -0.01, -0.418, 0.0928, -0.01, -0.4158, 0.0906, -0.01, -0.4158, 0.0928, -0.0061, -0.4223, 0.1001, -0.0083, -0.4223, 0.1001, -0.0061, -0.4211, 0.1003, -0.0061, -0.42, 0.1005, -0.0083, -0.4211, 0.1003, -0.0083, -0.42, 0.1005, -0.01, -0.4223, 0.0949, -0.01, -0.4223, 0.0971, -0.01, -0.4201, 0.0949, -0.01, -0.4201, 0.0971, -0.01, -0.4226, 0.1, -0.01, -0.4205, 0.1005, -0.01, -0.4197, 0.1006, -0.0061, -0.418, 0.1006, -0.0083, -0.418, 0.1006, -0.0061, -0.4154, 0.1003, -0.0061, -0.4165, 0.1004, -0.0083, -0.4154, 0.1003, -0.0083, -0.4165, 0.1004, -0.01, -0.418, 0.0949, -0.01, -0.418, 0.0971, -0.01, -0.4158, 0.0949, -0.01, -0.4158, 0.0971, -0.01, -0.418, 0.0993, -0.01, -0.418, 0.1006, -0.01, -0.4157, 0.1003,
-0.01, -0.4166, 0.1004, -0.0061, -0.413, -0.0014, -0.0083, -0.413, -0.0014, -0.0061, -0.4114, -0.0004, -0.0083, -0.4114, -0.0004, -0.01, -0.4131, -0.0014, -0.01, -0.4136, 0.0016, -0.01, -0.4112, -0.0002, -0.01, -0.4115, 0.0016, -0.01, -0.4136, 0.0038, -0.01, -0.4136, 0.006, -0.01, -0.4115, 0.0038, -0.01, -0.4115, 0.006, -0.0061, -0.4104, 0.0005, -0.0061, -0.4091, 0.0015, -0.0083, -0.4104, 0.0005, -0.0083, -0.4091, 0.0015, -0.0061, -0.4082, 0.0027, -0.0083, -0.4082, 0.0027, -0.0061, -0.4075, 0.0038, -0.0061, -0.4064, 0.0059, -0.0083, -0.4075, 0.0038, -0.0083, -0.4064, 0.0059, -0.01, -0.4104, 0.0005, -0.01, -0.4094, 0.0012, -0.01, -0.4081, 0.003, -0.01, -0.4093, 0.006, -0.01, -0.4074, 0.004, -0.01, -0.4065, 0.0057, -0.01, -0.4136, 0.0081,
-0.01, -0.4136, 0.0103, -0.01, -0.4115, 0.0081, -0.01, -0.4115, 0.0103, -0.01, -0.4136, 0.0125, -0.01, -0.4136, 0.0146, -0.01, -0.4115, 0.0125, -0.01, -0.4115, 0.0146, -0.0061, -0.4061, 0.0076, -0.0083, -0.4061, 0.0076, -0.0061, -0.4061, 0.0147, -0.0083, -0.4061, 0.0147, -0.01, -0.4093, 0.0081, -0.01, -0.4093, 0.0103, -0.01, -0.406, 0.008, -0.01, -0.4071, 0.0103, -0.01, -0.4093, 0.0125, -0.01, -0.4093, 0.0146, -0.01, -0.4071, 0.0125, -0.01, -0.4061, 0.0146, -0.0061, -0.3981, -0.0012, -0.0083, -0.3981, -0.0012, -0.01, -0.398, -0.0011, -0.0061, -0.406, 0.0087, -0.0061, -0.4059, 0.0103, -0.0083, -0.406, 0.0087, -0.0083, -0.4059, 0.0103, -0.0061, -0.4059, 0.0125, -0.0061, -0.406, 0.0136, -0.0083, -0.4059, 0.0125, -0.0083, -0.406, 0.0136,
-0.01, -0.4059, 0.0088, -0.01, -0.4059, 0.0103, -0.01, -0.4059, 0.0125, -0.01, -0.406, 0.0136, -0.0061, -0.4136, 0.024, -0.0083, -0.4136, 0.024, -0.0061, -0.4105, 0.0221, -0.0061, -0.4115, 0.023, -0.0083, -0.4105, 0.0221, -0.0083, -0.4115, 0.023, -0.01, -0.4136, 0.0168, -0.01, -0.4136, 0.019, -0.01, -0.4115, 0.0168, -0.01, -0.4115, 0.019, -0.01, -0.4136, 0.0212, -0.01, -0.4133, 0.0239, -0.01, -0.4105, 0.0222, -0.01, -0.4116, 0.0231, -0.0061, -0.4083, 0.0199, -0.0083, -0.4083, 0.0199, -0.0061, -0.4067, 0.0168, -0.0061, -0.4076, 0.0189, -0.0083, -0.4067, 0.0168, -0.0083, -0.4076, 0.0189, -0.0061, -0.4093, 0.0211, -0.0083, -0.4093, 0.0211, -0.01, -0.4093, 0.0168, -0.01, -0.4083, 0.0198, -0.01, -0.4068, 0.0171, -0.01, -0.4075, 0.0187,
-0.01, -0.4093, 0.0211, -0.0061, -0.3963, -0.0001, -0.0083, -0.3963, -0.0001, -0.0061, -0.3951, 0.0005, -0.0061, -0.394, 0.0011, -0.0083, -0.3951, 0.0005, -0.0083, -0.394, 0.0011, -0.01, -0.3966, -0.0003, -0.01, -0.3948, 0.0007, -0.01, -0.3939, 0.0012, -0.0061, -0.3921, 0.0022, -0.0083, -0.3921, 0.0022, -0.0061, -0.3911, 0.0028, -0.0083, -0.3911, 0.0028, -0.0061, -0.3897, 0.0036, -0.0083, -0.3897, 0.0036, -0.01, -0.3919, -0.0005, -0.01, -0.3921, 0.0022, -0.01, -0.3897, -0.0005, -0.01, -0.3897, 0.0016, -0.01, -0.391, 0.0029, -0.01, -0.3899, 0.0035, -0.0061, -0.3881, 0.0045, -0.0061, -0.387, 0.0052, -0.0083, -0.3881, 0.0045, -0.0083, -0.387, 0.0052, -0.0061, -0.3854, 0.0061, -0.0083, -0.3854, 0.0061, -0.01, -0.3876, -0.0005, -0.01, -0.3876, 0.0016,
-0.01, -0.3854, -0.0005, -0.01, -0.3854, 0.0016, -0.01, -0.388, 0.0046, -0.01, -0.3869, 0.0052, -0.01, -0.3854, 0.0038, -0.01, -0.3857, 0.0059, -0.0061, -0.3841, 0.0069, -0.0083, -0.3841, 0.0069, -0.01, -0.3832, -0.0005, -0.01, -0.3832, 0.0016, -0.01, -0.3811, -0.0005, -0.01, -0.3811, 0.0016, -0.01, -0.3832, 0.0038, -0.01, -0.3839, 0.007, -0.01, -0.3811, 0.0038, -0.01, -0.3811, 0.006, -0.0061, -0.383, 0.0075, -0.0083, -0.383, 0.0075, -0.0061, -0.3811, 0.0086, -0.0061, -0.38, 0.0093, -0.0083, -0.3811, 0.0086, -0.0083, -0.38, 0.0093, -0.01, -0.3828, 0.0076, -0.01, -0.3812, 0.0086, -0.01, -0.38, 0.0093, -0.0061, -0.4023, 0.0451, -0.0083, -0.4023, 0.0451, -0.0061, -0.4026, 0.0472, -0.0061, -0.402, 0.0489, -0.0083, -0.4026, 0.0472,
-0.0083, -0.402, 0.0489, -0.01, -0.4024, 0.0452, -0.01, -0.4026, 0.0472, -0.01, -0.4021, 0.0488, -0.0061, -0.4011, 0.0429, -0.0061, -0.4017, 0.044, -0.0083, -0.4011, 0.0429, -0.0083, -0.4017, 0.044, -0.0061, -0.3987, 0.042, -0.0083, -0.3987, 0.042, -0.0061, -0.4014, 0.05, -0.0083, -0.4014, 0.05, -0.01, -0.4011, 0.0429, -0.01, -0.4018, 0.0442, -0.01, -0.3986, 0.042, -0.01, -0.3984, 0.045, -0.01, -0.4006, 0.0472, -0.01, -0.4013, 0.05, -0.01, -0.3984, 0.0472, -0.01, -0.3984, 0.0494, -0.0061, -0.4002, 0.0508, -0.0083, -0.4002, 0.0508, -0.0061, -0.3986, 0.0514, -0.0083, -0.3986, 0.0514, -0.01, -0.4001, 0.0509, -0.01, -0.3986, 0.0514, -0.0061, -0.396, 0.0423, -0.0083, -0.396, 0.0423, -0.0061, -0.3946, 0.0435, -0.0061, -0.3935, 0.0444,
-0.0083, -0.3946, 0.0435, -0.0083, -0.3935, 0.0444, -0.0061, -0.3928, 0.0472, -0.0061, -0.3938, 0.0494, -0.0083, -0.3928, 0.0472, -0.0083, -0.3938, 0.0494, -0.01, -0.396, 0.0423, -0.01, -0.3963, 0.045, -0.01, -0.3946, 0.0434, -0.01, -0.3935, 0.0444, -0.01, -0.3963, 0.0472, -0.01, -0.3963, 0.0494, -0.01, -0.3928, 0.0472, -0.01, -0.3941, 0.0498, -0.0061, -0.3959, 0.0512, -0.0083, -0.3959, 0.0512, -0.0061, -0.3951, 0.0505, -0.0083, -0.3951, 0.0505, -0.01, -0.3959, 0.0512, -0.01, -0.3951, 0.0505, -0.01, -0.3789, -0.0005, -0.01, -0.3789, 0.0016, -0.01, -0.3767, -0.0005, -0.01, -0.3767, 0.0016, -0.01, -0.3789, 0.0038, -0.01, -0.3789, 0.006, -0.01, -0.3767, 0.0038, -0.01, -0.3767, 0.006, -0.01, -0.3746, -0.0005, -0.01, -0.3746, 0.0016,
-0.01, -0.3724, -0.0005, -0.01, -0.3724, 0.0016, -0.01, -0.3746, 0.0038, -0.01, -0.3746, 0.006, -0.01, -0.3724, 0.0038, -0.01, -0.3724, 0.006, -0.0061, -0.3789, 0.0099, -0.0083, -0.3789, 0.0099, -0.0061, -0.377, 0.0109, -0.0083, -0.377, 0.0109, -0.0061, -0.3759, 0.0115, -0.0083, -0.3759, 0.0115, -0.01, -0.3789, 0.0081, -0.01, -0.379, 0.0098, -0.01, -0.3767, 0.0081, -0.01, -0.3769, 0.0109, -0.01, -0.3758, 0.0115, -0.0061, -0.3746, 0.0123, -0.0083, -0.3746, 0.0123, -0.0061, -0.3729, 0.0132, -0.0061, -0.3718, 0.0138, -0.0083, -0.3729, 0.0132, -0.0083, -0.3718, 0.0138, -0.01, -0.3746, 0.0081, -0.01, -0.3746, 0.0103, -0.01, -0.3724, 0.0081, -0.01, -0.3724, 0.0103, -0.01, -0.3747, 0.0122, -0.01, -0.3728, 0.0133, -0.01, -0.3717, 0.0139,
-0.0061, -0.3694, -0.0015, -0.0083, -0.3694, -0.0015, -0.0061, -0.3681, -0.0006, -0.0083, -0.3681, -0.0006, -0.01, -0.3695, -0.0016, -0.01, -0.3702, 0.0016, -0.01, -0.3678, -0.0004, -0.01, -0.368, 0.0016, -0.01, -0.3702, 0.0038, -0.01, -0.3702, 0.006, -0.01, -0.368, 0.0038, -0.01, -0.368, 0.006, -0.0061, -0.3667, 0.0003, -0.0061, -0.3656, 0.0011, -0.0083, -0.3667, 0.0003, -0.0083, -0.3656, 0.0011, -0.0061, -0.3639, 0.0021, -0.0083, -0.3639, 0.0021, -0.0061, -0.3628, 0.0029, -0.0083, -0.3628, 0.0029, -0.01, -0.3667, 0.0003, -0.01, -0.3656, 0.001, -0.01, -0.364, 0.0021, -0.01, -0.3659, 0.0038, -0.01, -0.3659, 0.006, -0.01, -0.363, 0.0027, -0.01, -0.3637, 0.006, -0.0061, -0.3702, 0.0147, -0.0083, -0.3702, 0.0147, -0.0061, -0.3687, 0.0155,
-0.0083, -0.3687, 0.0155, -0.01, -0.3702, 0.0081, -0.01, -0.3702, 0.0103, -0.01, -0.368, 0.0081, -0.01, -0.368, 0.0103, -0.01, -0.3702, 0.0125, -0.01, -0.3705, 0.0146, -0.01, -0.368, 0.0125, -0.01, -0.3686, 0.0156, -0.01, -0.3659, 0.0081, -0.01, -0.3659, 0.0103, -0.01, -0.3637, 0.0081, -0.01, -0.3637, 0.0103, -0.01, -0.3659, 0.0125, -0.01, -0.3659, 0.0146, -0.01, -0.3637, 0.0125, -0.01, -0.3637, 0.0146, -0.0061, -0.3677, 0.0161, -0.0083, -0.3677, 0.0161, -0.01, -0.3675, 0.0162, -0.0061, -0.3659, 0.0171, -0.0083, -0.3659, 0.0171, -0.0061, -0.3646, 0.0178, -0.0061, -0.3635, 0.0184, -0.0083, -0.3646, 0.0178, -0.0083, -0.3635, 0.0184, -0.01, -0.3662, 0.0169, -0.01, -0.3643, 0.0179, -0.01, -0.3633, 0.0184, -0.0061, -0.3615, 0.0037,
-0.0083, -0.3615, 0.0037, -0.0061, -0.36, 0.0046, -0.0061, -0.3589, 0.0053, -0.0083, -0.36, 0.0046, -0.0083, -0.3589, 0.0053, -0.01, -0.3613, 0.0039, -0.01, -0.3615, 0.006, -0.01, -0.3601, 0.0046, -0.01, -0.359, 0.0053, -0.0061, -0.3548, -0.0011, -0.0083, -0.3548, -0.0011, -0.0061, -0.3572, 0.0064, -0.0083, -0.3572, 0.0064, -0.01, -0.3547, -0.001, -0.01, -0.3574, 0.0063, -0.01, -0.3615, 0.0081, -0.01, -0.3615, 0.0103, -0.01, -0.3594, 0.0081, -0.01, -0.3594, 0.0103, -0.01, -0.3615, 0.0125, -0.01, -0.3615, 0.0146, -0.01, -0.3594, 0.0125, -0.01, -0.3594, 0.0146, -0.0061, -0.3561, 0.0071, -0.0083, -0.3561, 0.0071, -0.0061, -0.355, 0.0077, -0.0083, -0.355, 0.0077, -0.01, -0.3563, 0.0069, -0.01, -0.3572, 0.0103, -0.01, -0.3547, 0.0079,
-0.01, -0.355, 0.0103, -0.01, -0.3572, 0.0125, -0.01, -0.3572, 0.0146, -0.01, -0.355, 0.0125, -0.01, -0.355, 0.0146, -0.0061, -0.3531, -0.0, -0.0061, -0.352, 0.0007, -0.0083, -0.3531, -0.0, -0.0083, -0.352, 0.0007, -0.0061, -0.3507, 0.0015, -0.0083, -0.3507, 0.0015, -0.01, -0.3531, 0.0, -0.01, -0.352, 0.0007, -0.01, -0.3507, -0.0005, -0.01, -0.3509, 0.0014, -0.0061, -0.3492, 0.0025, -0.0083, -0.3492, 0.0025, -0.0061, -0.3481, 0.0031, -0.0083, -0.3481, 0.0031, -0.0061, -0.3463, 0.0042, -0.0083, -0.3463, 0.0042, -0.01, -0.3485, -0.0005, -0.01, -0.3491, 0.0025, -0.01, -0.3463, -0.0005, -0.01, -0.3463, 0.0016, -0.01, -0.348, 0.0032, -0.01, -0.3467, 0.004, -0.0061, -0.3533, 0.0088, -0.0061, -0.3522, 0.0095, -0.0083, -0.3533, 0.0088,
-0.0083, -0.3522, 0.0095, -0.0061, -0.3507, 0.0104, -0.0083, -0.3507, 0.0104, -0.01, -0.3534, 0.0088, -0.01, -0.3523, 0.0094, -0.01, -0.3505, 0.0105, -0.01, -0.3529, 0.0125, -0.01, -0.3529, 0.0146, -0.01, -0.3507, 0.0125, -0.01, -0.3507, 0.0146, -0.0061, -0.3493, 0.0112, -0.0083, -0.3493, 0.0112, -0.0061, -0.3483, 0.0119, -0.0083, -0.3483, 0.0119, -0.0061, -0.3464, 0.013, -0.0061, -0.3453, 0.0136, -0.0083, -0.3464, 0.013, -0.0083, -0.3453, 0.0136, -0.01, -0.3494, 0.0112, -0.01, -0.3483, 0.0119, -0.01, -0.3485, 0.0146, -0.01, -0.3465, 0.0129, -0.01, -0.3456, 0.0134, -0.0061, -0.3615, 0.0194, -0.0083, -0.3615, 0.0194, -0.0061, -0.3604, 0.02, -0.0083, -0.3604, 0.02, -0.0061, -0.3593, 0.0206, -0.0083, -0.3593, 0.0206, -0.01, -0.3615, 0.0168,
-0.01, -0.3619, 0.0192, -0.01, -0.3594, 0.0168, -0.01, -0.3601, 0.0202, -0.01, -0.3591, 0.0207, -0.0061, -0.3573, 0.0217, -0.0061, -0.3562, 0.0223, -0.0083, -0.3573, 0.0217, -0.0083, -0.3562, 0.0223, -0.0061, -0.355, 0.0228, -0.0083, -0.355, 0.0228, -0.01, -0.3572, 0.0168, -0.01, -0.3572, 0.019, -0.01, -0.355, 0.0168, -0.01, -0.355, 0.019, -0.01, -0.3573, 0.0217, -0.01, -0.3562, 0.0223, -0.01, -0.355, 0.0212, -0.01, -0.3551, 0.0228, -0.0061, -0.3529, 0.0238, -0.0083, -0.3529, 0.0238, -0.0061, -0.3516, 0.0243, -0.0083, -0.3516, 0.0243, -0.01, -0.3529, 0.0168, -0.01, -0.3529, 0.019, -0.01, -0.3507, 0.0168, -0.01, -0.3507, 0.019, -0.01, -0.3529, 0.0212, -0.01, -0.3531, 0.0237, -0.01, -0.3507, 0.0212, -0.01, -0.3513, 0.0245,
-0.01, -0.3485, 0.0168, -0.01, -0.3485, 0.019, -0.01, -0.3463, 0.0168, -0.01, -0.3463, 0.019, -0.01, -0.3485, 0.0212, -0.01, -0.3485, 0.0233, -0.01, -0.3463, 0.0212, -0.01, -0.3463, 0.0233, -0.0061, -0.3505, 0.0248, -0.0083, -0.3505, 0.0248, -0.01, -0.3503, 0.0249, -0.0061, -0.3485, 0.0257, -0.0083, -0.3485, 0.0257, -0.0061, -0.347, 0.0264, -0.0061, -0.3459, 0.0269, -0.0083, -0.347, 0.0264, -0.0083, -0.3459, 0.0269, -0.01, -0.3488, 0.0256, -0.01, -0.3468, 0.0265, -0.01, -0.3458, 0.0269, -0.0061, -0.3715, 0.0633, -0.0083, -0.3715, 0.0633, -0.0061, -0.3738, 0.0653, -0.0061, -0.3749, 0.0667, -0.0083, -0.3738, 0.0653, -0.0083, -0.3749, 0.0667, -0.0061, -0.3727, 0.0641, -0.0083, -0.3727, 0.0641, -0.01, -0.3715, 0.0633, -0.01, -0.3738, 0.0653,
-0.01, -0.3747, 0.0664, -0.01, -0.3727, 0.0641, -0.01, -0.3724, 0.0667, -0.0061, -0.3703, 0.0625, -0.0083, -0.3703, 0.0625, -0.0061, -0.3681, 0.0619, -0.0083, -0.3681, 0.0619, -0.01, -0.3704, 0.0626, -0.01, -0.3682, 0.0619, -0.01, -0.3702, 0.0646, -0.01, -0.3702, 0.0667, -0.01, -0.368, 0.0646, -0.01, -0.368, 0.0667, -0.0061, -0.3659, 0.0617, -0.0083, -0.3659, 0.0617, -0.0061, -0.3636, 0.062, -0.0083, -0.3636, 0.062, -0.01, -0.3659, 0.0617, -0.01, -0.3635, 0.062, -0.01, -0.3659, 0.0646, -0.01, -0.3659, 0.0667, -0.01, -0.3637, 0.0646, -0.01, -0.3637, 0.0667, -0.0061, -0.3548, 0.0406, -0.0083, -0.3548, 0.0406, -0.01, -0.3549, 0.0408, -0.0061, -0.3561, 0.0439, -0.0061, -0.3563, 0.045, -0.0083, -0.3561, 0.0439, -0.0083, -0.3563, 0.045,
-0.0061, -0.3558, 0.0427, -0.0083, -0.3558, 0.0427, -0.0061, -0.3563, 0.0472, -0.0083, -0.3563, 0.0472, -0.0061, -0.3561, 0.0483, -0.0061, -0.3559, 0.0494, -0.0083, -0.3561, 0.0483, -0.0083, -0.3559, 0.0494, -0.01, -0.3561, 0.0439, -0.01, -0.3563, 0.045, -0.01, -0.3558, 0.0428, -0.01, -0.355, 0.045, -0.01, -0.3564, 0.0468, -0.01, -0.3562, 0.0476, -0.01, -0.3558, 0.0497, -0.0061, -0.3502, 0.0372, -0.0083, -0.3502, 0.0372, -0.0061, -0.3533, 0.0389, -0.0061, -0.3539, 0.0396, -0.0083, -0.3533, 0.0389, -0.0083, -0.3539, 0.0396, -0.0061, -0.3513, 0.0377, -0.0083, -0.3513, 0.0377, -0.01, -0.3501, 0.0372, -0.01, -0.3529, 0.0386, -0.01, -0.354, 0.0397, -0.01, -0.3511, 0.0376, -0.01, -0.3507, 0.0407, -0.0061, -0.3485, 0.0366, -0.0083, -0.3485, 0.0366,
-0.0061, -0.3464, 0.0364, -0.0083, -0.3464, 0.0364, -0.01, -0.3486, 0.0367, -0.01, -0.3464, 0.0364, -0.01, -0.3485, 0.0385, -0.01, -0.3485, 0.0407, -0.01, -0.3463, 0.0385, -0.01, -0.3463, 0.0407, -0.01, -0.3529, 0.0429, -0.01, -0.3529, 0.045, -0.01, -0.3507, 0.0429, -0.01, -0.3507, 0.045, -0.01, -0.3529, 0.0472, -0.01, -0.3529, 0.0494, -0.01, -0.3507, 0.0472, -0.01, -0.3507, 0.0494, -0.01, -0.3485, 0.0429, -0.01, -0.3485, 0.045, -0.01, -0.3463, 0.0429, -0.01, -0.3463, 0.045, -0.01, -0.3485, 0.0472, -0.01, -0.3485, 0.0494, -0.01, -0.3463, 0.0472, -0.01, -0.3463, 0.0494, -0.0061, -0.3548, 0.0516, -0.0083, -0.3548, 0.0516, -0.01, -0.3549, 0.0515, -0.0061, -0.3617, 0.0629, -0.0083, -0.3617, 0.0629, -0.0061, -0.3625, 0.0622,
-0.0061, -0.3605, 0.0665, -0.0083, -0.3625, 0.0622, -0.0083, -0.3605, 0.0665, -0.0061, -0.3591, 0.0647, -0.0061, -0.3592, 0.0663, -0.0083, -0.3591, 0.0647, -0.0083, -0.3592, 0.0663, -0.01, -0.3618, 0.0628, -0.01, -0.3625, 0.0622, -0.01, -0.3607, 0.0661, -0.01, -0.3593, 0.0645, -0.01, -0.3591, 0.0664, -0.0061, -0.3578, 0.0669, -0.0083, -0.3578, 0.0669, -0.01, -0.3577, 0.0671, -0.0061, -0.3539, 0.0526, -0.0061, -0.3532, 0.0536, -0.0083, -0.3539, 0.0526, -0.0083, -0.3532, 0.0536, -0.0061, -0.3514, 0.0546, -0.0083, -0.3514, 0.0546, -0.0061, -0.3503, 0.0551, -0.0083, -0.3503, 0.0551, -0.01, -0.354, 0.0525, -0.01, -0.353, 0.0538, -0.01, -0.3507, 0.0515, -0.01, -0.3512, 0.0547, -0.01, -0.3501, 0.0551, -0.0061, -0.3485, 0.0554, -0.0083, -0.3485, 0.0554,
-0.0061, -0.3463, 0.0555, -0.0083, -0.3463, 0.0555, -0.01, -0.3485, 0.0515, -0.01, -0.3485, 0.0537, -0.01, -0.3463, 0.0515, -0.01, -0.3463, 0.0537, -0.01, -0.3486, 0.0554, -0.01, -0.3463, 0.0555, -0.0061, -0.4137, 0.0694, -0.0061, -0.4126, 0.07, -0.0083, -0.4137, 0.0694, -0.0083, -0.4126, 0.07, -0.0061, -0.4115, 0.0706, -0.0083, -0.4115, 0.0706, -0.0061, -0.4136, 0.0739, -0.0061, -0.4131, 0.0746, -0.0083, -0.4136, 0.0739, -0.0083, -0.4131, 0.0746, -0.0061, -0.4119, 0.074, -0.0061, -0.4114, 0.0747, -0.0083, -0.4119, 0.074, -0.0083, -0.4114, 0.0747, -0.01, -0.4139, 0.0693, -0.01, -0.4129, 0.0699, -0.01, -0.4111, 0.0708, -0.01, -0.414, 0.0738, -0.01, -0.4133, 0.0745, -0.01, -0.4118, 0.074, -0.01, -0.411, 0.0749, -0.0061, -0.4093, 0.0717,
-0.0083, -0.4093, 0.0717, -0.0061, -0.4082, 0.0722, -0.0083, -0.4082, 0.0722, -0.0061, -0.4071, 0.0727, -0.0083, -0.4071, 0.0727, -0.01, -0.4095, 0.0716, -0.01, -0.4086, 0.072, -0.01, -0.4095, 0.0757, -0.01, -0.4068, 0.0728, -0.01, -0.4071, 0.0754, -0.01, -0.4136, 0.0776, -0.01, -0.4136, 0.0798, -0.01, -0.4115, 0.0776, -0.01, -0.4115, 0.0798, -0.01, -0.4136, 0.0819, -0.01, -0.4136, 0.0841, -0.01, -0.4115, 0.0819, -0.01, -0.4115, 0.0841, -0.0061, -0.4068, 0.0797, -0.0083, -0.4068, 0.0797, -0.0061, -0.4062, 0.0814, -0.0083, -0.4062, 0.0814, -0.01, -0.4086, 0.0772, -0.01, -0.4093, 0.0798, -0.01, -0.4081, 0.0784, -0.01, -0.4069, 0.0794, -0.01, -0.4093, 0.0819, -0.01, -0.4093, 0.0841, -0.01, -0.4061, 0.0816, -0.01, -0.4071, 0.0841,
-0.0061, -0.4049, 0.0735, -0.0083, -0.4049, 0.0735, -0.0061, -0.4032, 0.0741, -0.0061, -0.4021, 0.0745, -0.0083, -0.4032, 0.0741, -0.0083, -0.4021, 0.0745, -0.01, -0.4048, 0.0735, -0.01, -0.4049, 0.0754, -0.01, -0.4033, 0.074, -0.01, -0.4024, 0.0743, -0.0061, -0.4006, 0.0749, -0.0083, -0.4006, 0.0749, -0.0061, -0.3984, 0.0757, -0.0083, -0.3984, 0.0757, -0.01, -0.4004, 0.075, -0.01, -0.3983, 0.0757, -0.0061, -0.4057, 0.0803, -0.0083, -0.4057, 0.0803, -0.0061, -0.4056, 0.0819, -0.0083, -0.4056, 0.0819, -0.01, -0.4049, 0.0776, -0.01, -0.4057, 0.0802, -0.01, -0.4028, 0.0776, -0.01, -0.4028, 0.0798, -0.01, -0.4055, 0.0823, -0.01, -0.4054, 0.0842, -0.01, -0.4028, 0.0819, -0.01, -0.4028, 0.0841, -0.01, -0.4006, 0.0776, -0.01, -0.4006, 0.0798,
-0.01, -0.3984, 0.0776, -0.01, -0.3984, 0.0798, -0.01, -0.4006, 0.0819, -0.01, -0.4006, 0.0841, -0.01, -0.3984, 0.0819, -0.01, -0.3984, 0.0841, -0.01, -0.4136, 0.0863, -0.01, -0.4136, 0.0884, -0.01, -0.4115, 0.0863, -0.01, -0.4115, 0.0884, -0.01, -0.4136, 0.0906, -0.01, -0.4136, 0.0928, -0.01, -0.4115, 0.0906, -0.01, -0.4115, 0.0928, -0.0061, -0.4064, 0.0928, -0.0083, -0.4064, 0.0928, -0.01, -0.4093, 0.0863, -0.01, -0.4093, 0.0884, -0.01, -0.4071, 0.0863, -0.01, -0.4071, 0.0884, -0.01, -0.4093, 0.0906, -0.01, -0.4093, 0.0928, -0.01, -0.4071, 0.0906, -0.01, -0.4064, 0.0929, -0.0061, -0.4106, 0.098, -0.0083, -0.4106, 0.098, -0.0061, -0.4136, 0.0998, -0.0083, -0.4136, 0.0998, -0.0061, -0.4117, 0.0987, -0.0083, -0.4117, 0.0987,
-0.01, -0.4136, 0.0949, -0.01, -0.4136, 0.0971, -0.01, -0.4115, 0.0949, -0.01, -0.4107, 0.0981, -0.01, -0.4133, 0.0997, -0.01, -0.4118, 0.0988, -0.0061, -0.4084, 0.0958, -0.0061, -0.4094, 0.097, -0.0083, -0.4084, 0.0958, -0.0083, -0.4094, 0.097, -0.0061, -0.4075, 0.0947, -0.0083, -0.4075, 0.0947, -0.01, -0.4083, 0.0957, -0.01, -0.4094, 0.097, -0.01, -0.4075, 0.0946, -0.0061, -0.4045, 0.0887, -0.0083, -0.4045, 0.0887, -0.0061, -0.4028, 0.089, -0.0083, -0.4028, 0.089, -0.0061, -0.4056, 0.0906, -0.0061, -0.406, 0.0917, -0.0083, -0.4056, 0.0906, -0.0083, -0.406, 0.0917, -0.01, -0.4045, 0.0862, -0.01, -0.4045, 0.0887, -0.01, -0.4028, 0.0863, -0.01, -0.4029, 0.089, -0.01, -0.4056, 0.0907, -0.01, -0.406, 0.0917, -0.0061, -0.4006, 0.0893,
-0.0083, -0.4006, 0.0893, -0.0061, -0.3995, 0.0895, -0.0083, -0.3995, 0.0895, -0.0061, -0.3984, 0.0897, -0.0083, -0.3984, 0.0897, -0.01, -0.4006, 0.0863, -0.01, -0.4005, 0.0893, -0.01, -0.3984, 0.0863, -0.01, -0.3984, 0.0884, -0.01, -0.3995, 0.0895, -0.01, -0.3984, 0.0897, -0.0061, -0.3966, 0.0763, -0.0083, -0.3966, 0.0763, -0.01, -0.3968, 0.0762, -0.0061, -0.3955, 0.0766, -0.0083, -0.3955, 0.0766, -0.0061, -0.3941, 0.077, -0.0083, -0.3941, 0.077, -0.01, -0.3959, 0.0765, -0.01, -0.3963, 0.0798, -0.01, -0.3939, 0.077, -0.01, -0.3941, 0.0798, -0.01, -0.3963, 0.0819, -0.01, -0.3963, 0.0841, -0.01, -0.3941, 0.0819, -0.01, -0.3941, 0.0841, -0.0061, -0.3919, 0.0775, -0.0083, -0.3919, 0.0775, -0.0061, -0.3898, 0.0779, -0.0083, -0.3898, 0.0779,
-0.01, -0.3918, 0.0775, -0.01, -0.3919, 0.0798, -0.01, -0.3897, 0.0779, -0.01, -0.3897, 0.0798, -0.01, -0.3919, 0.0819, -0.01, -0.3919, 0.0841, -0.01, -0.3897, 0.0819, -0.01, -0.3897, 0.0841, -0.0061, -0.3876, 0.0783, -0.0083, -0.3876, 0.0783, -0.0061, -0.3858, 0.0786, -0.0061, -0.3848, 0.0787, -0.0083, -0.3858, 0.0786, -0.0083, -0.3848, 0.0787, -0.01, -0.3875, 0.0783, -0.01, -0.3876, 0.0798, -0.01, -0.3861, 0.0785, -0.01, -0.3852, 0.0787, -0.01, -0.3876, 0.0819, -0.01, -0.3876, 0.0841, -0.01, -0.3854, 0.0819, -0.01, -0.3854, 0.0841, -0.0061, -0.3832, 0.0789, -0.0083, -0.3832, 0.0789, -0.0061, -0.3811, 0.079, -0.0083, -0.3811, 0.079, -0.01, -0.3832, 0.0789, -0.01, -0.381, 0.079, -0.01, -0.3832, 0.0819, -0.01, -0.3832, 0.0841,
-0.01, -0.3811, 0.0819, -0.01, -0.3811, 0.0841, -0.0061, -0.3963, 0.0901, -0.0083, -0.3963, 0.0901, -0.0061, -0.3941, 0.0904, -0.0083, -0.3941, 0.0904, -0.01, -0.3963, 0.0863, -0.01, -0.3963, 0.0884, -0.01, -0.3941, 0.0863, -0.01, -0.3941, 0.0884, -0.01, -0.3963, 0.0901, -0.01, -0.3942, 0.0904, -0.0061, -0.3919, 0.0908, -0.0083, -0.3919, 0.0908, -0.0061, -0.3898, 0.0912, -0.0083, -0.3898, 0.0912, -0.01, -0.3919, 0.0863, -0.01, -0.3919, 0.0884, -0.01, -0.3897, 0.0863, -0.01, -0.3897, 0.0884, -0.01, -0.392, 0.0908, -0.01, -0.3899, 0.0911, -0.0061, -0.3876, 0.0915, -0.0083, -0.3876, 0.0915, -0.0061, -0.3861, 0.0917, -0.0061, -0.3851, 0.0917, -0.0083, -0.3861, 0.0917, -0.0083, -0.3851, 0.0917, -0.01, -0.3876, 0.0863, -0.01, -0.3876, 0.0884,
-0.01, -0.3854, 0.0863, -0.01, -0.3854, 0.0884, -0.01, -0.3877, 0.0915, -0.01, -0.3856, 0.0917, -0.01, -0.3848, 0.0918, -0.0061, -0.3832, 0.0918, -0.0083, -0.3832, 0.0918, -0.0061, -0.3811, 0.0918, -0.0083, -0.3811, 0.0918, -0.01, -0.3832, 0.0863, -0.01, -0.3832, 0.0884, -0.01, -0.3811, 0.0863, -0.01, -0.3811, 0.0884, -0.01, -0.3832, 0.0906, -0.01, -0.3832, 0.0918, -0.01, -0.3811, 0.0906, -0.01, -0.3811, 0.0918, -0.0061, -0.3849, 0.0954, -0.0061, -0.3861, 0.0969, -0.0083, -0.3849, 0.0954, -0.0083, -0.3861, 0.0969, -0.0061, -0.3867, 0.0991, -0.0061, -0.3862, 0.1015, -0.0083, -0.3867, 0.0991, -0.0083, -0.3862, 0.1015, -0.01, -0.3849, 0.0954, -0.01, -0.386, 0.0967, -0.01, -0.3867, 0.0991, -0.01, -0.3862, 0.1015, -0.0061, -0.3832, 0.0941,
-0.0083, -0.3832, 0.0941, -0.0061, -0.3804, 0.0944, -0.0083, -0.3804, 0.0944, -0.01, -0.3832, 0.0941, -0.01, -0.3832, 0.0971, -0.01, -0.3808, 0.0943, -0.01, -0.3811, 0.0971, -0.01, -0.3832, 0.0993, -0.01, -0.3832, 0.1015, -0.01, -0.3811, 0.0993, -0.01, -0.3811, 0.1015, -0.0061, -0.3848, 0.1029, -0.0083, -0.3848, 0.1029, -0.01, -0.3847, 0.1029, -0.0061, -0.3832, 0.1037, -0.0083, -0.3832, 0.1037, -0.0061, -0.381, 0.1039, -0.0083, -0.381, 0.1039, -0.01, -0.3834, 0.1037, -0.01, -0.381, 0.1039, -0.0061, -0.3758, 0.0693, -0.0061, -0.3761, 0.0711, -0.0083, -0.3758, 0.0693, -0.0083, -0.3761, 0.0711, -0.0061, -0.376, 0.0731, -0.0083, -0.376, 0.0731, -0.01, -0.3759, 0.0695, -0.01, -0.3761, 0.071, -0.01, -0.376, 0.0728, -0.0061, -0.3755, 0.0682,
-0.0083, -0.3755, 0.0682, -0.0061, -0.3756, 0.0742, -0.0061, -0.3751, 0.0755, -0.0083, -0.3756, 0.0742, -0.0083, -0.3751, 0.0755, -0.01, -0.3757, 0.0685, -0.01, -0.3746, 0.0711, -0.01, -0.3724, 0.0689, -0.01, -0.3724, 0.0711, -0.01, -0.3758, 0.0738, -0.01, -0.3749, 0.0758, -0.01, -0.3724, 0.0732, -0.01, -0.3724, 0.0754, -0.0061, -0.3789, 0.0791, -0.0083, -0.3789, 0.0791, -0.0061, -0.3767, 0.0792, -0.0083, -0.3767, 0.0792, -0.01, -0.3789, 0.0791, -0.01, -0.3766, 0.0793, -0.01, -0.3789, 0.0819, -0.01, -0.3789, 0.0841, -0.01, -0.3767, 0.0819, -0.01, -0.3767, 0.0841, -0.0061, -0.374, 0.0773, -0.0061, -0.3745, 0.0798, -0.0083, -0.374, 0.0773, -0.0083, -0.3745, 0.0798, -0.0061, -0.3732, 0.0783, -0.0061, -0.3724, 0.0799, -0.0083, -0.3732, 0.0783,
-0.0083, -0.3724, 0.0799, -0.0061, -0.3717, 0.081, -0.0083, -0.3717, 0.081, -0.01, -0.3741, 0.0772, -0.01, -0.3744, 0.0799, -0.01, -0.3731, 0.0784, -0.01, -0.3725, 0.0799, -0.01, -0.3746, 0.0819, -0.01, -0.3746, 0.0841, -0.01, -0.372, 0.0809, -0.01, -0.3724, 0.0841, -0.01, -0.3702, 0.0689, -0.01, -0.3702, 0.0711, -0.01, -0.368, 0.0689, -0.01, -0.368, 0.0711, -0.01, -0.3702, 0.0732, -0.01, -0.3702, 0.0754, -0.01, -0.368, 0.0732, -0.01, -0.368, 0.0754, -0.01, -0.3659, 0.0689, -0.01, -0.3659, 0.0711, -0.01, -0.3637, 0.0689, -0.01, -0.3637, 0.0711, -0.01, -0.3659, 0.0732, -0.01, -0.3659, 0.0754, -0.01, -0.3637, 0.0732, -0.01, -0.3637, 0.0754, -0.0061, -0.3703, 0.0803, -0.0083, -0.3703, 0.0803, -0.0061, -0.3686, 0.0808,
-0.0083, -0.3686, 0.0808, -0.0061, -0.3702, 0.0816, -0.0083, -0.3702, 0.0816, -0.0061, -0.368, 0.0818, -0.0083, -0.368, 0.0818, -0.01, -0.3702, 0.0776, -0.01, -0.3705, 0.0802, -0.01, -0.368, 0.0776, -0.01, -0.3682, 0.0808, -0.01, -0.37, 0.0817, -0.01, -0.3702, 0.0841, -0.01, -0.3678, 0.0819, -0.01, -0.368, 0.0841, -0.0061, -0.365, 0.0808, -0.0083, -0.365, 0.0808, -0.0061, -0.3637, 0.0807, -0.0083, -0.3637, 0.0807, -0.0061, -0.367, 0.082, -0.0083, -0.367, 0.082, -0.0061, -0.3637, 0.0821, -0.0083, -0.3637, 0.0821, -0.01, -0.3659, 0.0776, -0.01, -0.3657, 0.0809, -0.01, -0.3637, 0.0776, -0.01, -0.3634, 0.0807, -0.01, -0.366, 0.082, -0.01, -0.3659, 0.0841, -0.01, -0.3638, 0.0821, -0.01, -0.3637, 0.0841, -0.0061, -0.3789, 0.0918,
-0.0083, -0.3789, 0.0918, -0.0061, -0.3767, 0.0918, -0.0083, -0.3767, 0.0918, -0.01, -0.3789, 0.0863, -0.01, -0.3789, 0.0884, -0.01, -0.3767, 0.0863, -0.01, -0.3767, 0.0884, -0.01, -0.3789, 0.0906, -0.01, -0.3789, 0.0918, -0.01, -0.3767, 0.0906, -0.01, -0.3767, 0.0918, -0.0061, -0.3746, 0.0918, -0.0083, -0.3746, 0.0918, -0.0061, -0.3724, 0.0918, -0.0083, -0.3724, 0.0918, -0.01, -0.3746, 0.0863, -0.01, -0.3746, 0.0884, -0.01, -0.3724, 0.0863, -0.01, -0.3724, 0.0884, -0.01, -0.3746, 0.0906, -0.01, -0.3746, 0.0918, -0.01, -0.3724, 0.0906, -0.01, -0.3724, 0.0918, -0.0061, -0.3789, 0.0952, -0.0061, -0.3779, 0.0962, -0.0083, -0.3789, 0.0952, -0.0083, -0.3779, 0.0962, -0.0061, -0.3773, 0.0972, -0.0083, -0.3773, 0.0972, -0.0061, -0.3782, 0.1021,
-0.0083, -0.3782, 0.1021, -0.0061, -0.3769, 0.0991, -0.0061, -0.3776, 0.1014, -0.0083, -0.3769, 0.0991, -0.0083, -0.3776, 0.1014, -0.01, -0.379, 0.0951, -0.01, -0.3779, 0.0963, -0.01, -0.3773, 0.0974, -0.01, -0.3789, 0.0993, -0.01, -0.3782, 0.102, -0.01, -0.3769, 0.0991, -0.01, -0.3774, 0.1009, -0.0061, -0.3701, 0.0916, -0.0061, -0.3712, 0.0917, -0.0083, -0.3701, 0.0916, -0.0083, -0.3712, 0.0917, -0.0061, -0.368, 0.0913, -0.0083, -0.368, 0.0913, -0.01, -0.3702, 0.0863, -0.01, -0.3702, 0.0884, -0.01, -0.368, 0.0863, -0.01, -0.368, 0.0884, -0.01, -0.3703, 0.0916, -0.01, -0.3712, 0.0917, -0.01, -0.3679, 0.0913, -0.0061, -0.3659, 0.0909, -0.0083, -0.3659, 0.0909, -0.0061, -0.3637, 0.0905, -0.0083, -0.3637, 0.0905, -0.01, -0.3659, 0.0863,
-0.01, -0.3659, 0.0884, -0.01, -0.3637, 0.0863, -0.01, -0.3637, 0.0884, -0.01, -0.3658, 0.0908, -0.01, -0.3636, 0.0904, -0.0061, -0.3605, 0.0679, -0.0083, -0.3605, 0.0679, -0.0061, -0.3588, 0.0704, -0.0083, -0.3588, 0.0704, -0.0061, -0.3583, 0.0765, -0.0083, -0.3583, 0.0765, -0.01, -0.3603, 0.0682, -0.01, -0.3615, 0.0711, -0.01, -0.3596, 0.0691, -0.01, -0.3587, 0.0703, -0.01, -0.3615, 0.0732, -0.01, -0.3615, 0.0754, -0.01, -0.3594, 0.0732, -0.01, -0.3582, 0.0763, -0.0061, -0.3571, 0.0689, -0.0061, -0.3569, 0.0711, -0.0083, -0.3571, 0.0689, -0.0083, -0.3569, 0.0711, -0.0061, -0.357, 0.0733, -0.0061, -0.3577, 0.0757, -0.0083, -0.357, 0.0733, -0.0083, -0.3577, 0.0757, -0.01, -0.3571, 0.0687, -0.01, -0.3569, 0.071, -0.01, -0.3571, 0.0733,
-0.01, -0.3575, 0.0752, -0.0061, -0.3614, 0.0805, -0.0083, -0.3614, 0.0805, -0.0061, -0.3591, 0.0777, -0.0061, -0.3596, 0.0793, -0.0083, -0.3591, 0.0777, -0.0083, -0.3596, 0.0793, -0.0061, -0.3619, 0.0812, -0.0083, -0.3619, 0.0812, -0.01, -0.3615, 0.0776, -0.01, -0.3611, 0.0803, -0.01, -0.3594, 0.078, -0.01, -0.3599, 0.0795, -0.01, -0.3618, 0.0812, -0.01, -0.3615, 0.0841, -0.01, -0.3594, 0.0819, -0.01, -0.3594, 0.0841, -0.0061, -0.357, 0.0781, -0.0061, -0.3581, 0.0788, -0.0083, -0.357, 0.0781, -0.0083, -0.3581, 0.0788, -0.0061, -0.3551, 0.0772, -0.0083, -0.3551, 0.0772, -0.01, -0.3569, 0.0781, -0.01, -0.3579, 0.0787, -0.01, -0.3553, 0.0773, -0.01, -0.355, 0.0798, -0.01, -0.3572, 0.0819, -0.01, -0.3572, 0.0841, -0.01, -0.355, 0.0819,
-0.01, -0.355, 0.0841, -0.0061, -0.3521, 0.0764, -0.0083, -0.3521, 0.0764, -0.0061, -0.3507, 0.076, -0.0083, -0.3507, 0.076, -0.01, -0.352, 0.0764, -0.01, -0.3508, 0.076, -0.0061, -0.3485, 0.0753, -0.0083, -0.3485, 0.0753, -0.0061, -0.3455, 0.0742, -0.0061, -0.3466, 0.0746, -0.0083, -0.3455, 0.0742, -0.0083, -0.3466, 0.0746, -0.01, -0.3487, 0.0754, -0.01, -0.3454, 0.0742, -0.01, -0.3465, 0.0746, -0.0061, -0.3532, 0.0767, -0.0083, -0.3532, 0.0767, -0.01, -0.353, 0.0767, -0.01, -0.3529, 0.0798, -0.01, -0.3507, 0.0776, -0.01, -0.3507, 0.0798, -0.01, -0.3529, 0.0819, -0.01, -0.3529, 0.0841, -0.01, -0.3507, 0.0819, -0.01, -0.3507, 0.0841, -0.01, -0.3485, 0.0776, -0.01, -0.3485, 0.0798, -0.01, -0.3463, 0.0776, -0.01, -0.3463, 0.0798,
-0.01, -0.3485, 0.0819, -0.01, -0.3485, 0.0841, -0.01, -0.3463, 0.0819, -0.01, -0.3463, 0.0841, -0.0061, -0.3583, 0.0895, -0.0083, -0.3583, 0.0895, -0.0061, -0.3615, 0.0901, -0.0083, -0.3615, 0.0901, -0.0061, -0.3594, 0.0897, -0.0083, -0.3594, 0.0897, -0.01, -0.3615, 0.0863, -0.01, -0.3615, 0.0884, -0.01, -0.3594, 0.0863, -0.01, -0.359, 0.0896, -0.01, -0.3615, 0.09, -0.01, -0.3597, 0.0898, -0.0061, -0.3572, 0.0893, -0.0083, -0.3572, 0.0893, -0.0061, -0.355, 0.0889, -0.0083, -0.355, 0.0889, -0.01, -0.3572, 0.0863, -0.01, -0.357, 0.0893, -0.01, -0.355, 0.0863, -0.01, -0.3549, 0.0889, -0.0061, -0.3529, 0.0885, -0.0083, -0.3529, 0.0885, -0.0061, -0.3507, 0.0879, -0.0083, -0.3507, 0.0879, -0.01, -0.3529, 0.0863, -0.01, -0.3527, 0.0884,
-0.01, -0.3507, 0.0863, -0.01, -0.3506, 0.0879, -0.0061, -0.3482, 0.0871, -0.0061, -0.3493, 0.0875, -0.0083, -0.3482, 0.0871, -0.0083, -0.3493, 0.0875, -0.0061, -0.3463, 0.0864, -0.0083, -0.3463, 0.0864, -0.01, -0.3483, 0.0871, -0.01, -0.3494, 0.0875, -0.01, -0.3461, 0.0864, -0.0061, -0.3791, 0.1031, -0.0083, -0.3791, 0.1031, -0.01, -0.3791, 0.1031, -0.0061, -0.3452, 0.0049, -0.0061, -0.3442, 0.0055, -0.0083, -0.3452, 0.0049, -0.0083, -0.3442, 0.0055, -0.0061, -0.3423, 0.0066, -0.0083, -0.3423, 0.0066, -0.01, -0.3442, -0.0005, -0.01, -0.3442, 0.0016, -0.01, -0.342, -0.0005, -0.01, -0.342, 0.0016, -0.01, -0.345, 0.005, -0.01, -0.344, 0.0056, -0.01, -0.342, 0.0038, -0.01, -0.3423, 0.0066, -0.01, -0.3398, -0.0005, -0.01, -0.3398, 0.0016,
-0.01, -0.3377, -0.0005, -0.01, -0.3377, 0.0016, -0.01, -0.3398, 0.0038, -0.01, -0.3398, 0.006, -0.01, -0.3377, 0.0038, -0.01, -0.3377, 0.006, -0.0061, -0.3412, 0.0072, -0.0083, -0.3412, 0.0072, -0.0061, -0.3442, 0.0142, -0.0083, -0.3442, 0.0142, -0.0061, -0.3423, 0.0153, -0.0083, -0.3423, 0.0153, -0.01, -0.3412, 0.0072, -0.01, -0.3438, 0.0144, -0.01, -0.3424, 0.0152, -0.0061, -0.3398, 0.008, -0.0083, -0.3398, 0.008, -0.0061, -0.3381, 0.0089, -0.0061, -0.3371, 0.0095, -0.0083, -0.3381, 0.0089, -0.0083, -0.3371, 0.0095, -0.01, -0.34, 0.0079, -0.01, -0.338, 0.009, -0.01, -0.337, 0.0095, -0.01, -0.3355, -0.0005, -0.01, -0.3355, 0.0016, -0.01, -0.3333, -0.0005, -0.01, -0.3333, 0.0016, -0.01, -0.3355, 0.0038, -0.01, -0.3355, 0.006,
-0.01, -0.3333, 0.0038, -0.01, -0.3333, 0.006, -0.01, -0.3312, -0.0005, -0.01, -0.3312, 0.0016, -0.01, -0.329, -0.0005, -0.01, -0.329, 0.0016, -0.01, -0.3312, 0.0038, -0.01, -0.3312, 0.006, -0.01, -0.329, 0.0038, -0.01, -0.329, 0.006, -0.0061, -0.3355, 0.0102, -0.0083, -0.3355, 0.0102, -0.0061, -0.3337, 0.0111, -0.0083, -0.3337, 0.0111, -0.0061, -0.3327, 0.0116, -0.0083, -0.3327, 0.0116, -0.01, -0.3355, 0.0081, -0.01, -0.3357, 0.0101, -0.01, -0.3333, 0.0081, -0.01, -0.3336, 0.0111, -0.01, -0.3326, 0.0116, -0.0061, -0.3312, 0.0123, -0.0083, -0.3312, 0.0123, -0.0061, -0.3293, 0.0132, -0.0061, -0.3282, 0.0137, -0.0083, -0.3293, 0.0132, -0.0083, -0.3282, 0.0137, -0.01, -0.3312, 0.0081, -0.01, -0.3312, 0.0103, -0.01, -0.329, 0.0081,
-0.01, -0.329, 0.0103, -0.01, -0.3313, 0.0122, -0.01, -0.3292, 0.0132, -0.01, -0.3282, 0.0137, -0.0061, -0.3412, 0.0159, -0.0083, -0.3412, 0.0159, -0.01, -0.3442, 0.0168, -0.01, -0.3442, 0.019, -0.01, -0.3414, 0.0158, -0.01, -0.342, 0.019, -0.01, -0.3442, 0.0212, -0.01, -0.3442, 0.0233, -0.01, -0.342, 0.0212, -0.01, -0.342, 0.0233, -0.0061, -0.3398, 0.0167, -0.0083, -0.3398, 0.0167, -0.0061, -0.3382, 0.0176, -0.0061, -0.3371, 0.0181, -0.0083, -0.3382, 0.0176, -0.0083, -0.3371, 0.0181, -0.01, -0.3396, 0.0168, -0.01, -0.3398, 0.019, -0.01, -0.3383, 0.0175, -0.01, -0.3372, 0.0181, -0.01, -0.3398, 0.0212, -0.01, -0.3398, 0.0233, -0.01, -0.3377, 0.0212, -0.01, -0.3377, 0.0233, -0.0061, -0.3442, 0.0277, -0.0083, -0.3442, 0.0277,
-0.0061, -0.3423, 0.0284, -0.0083, -0.3423, 0.0284, -0.0061, -0.3412, 0.0289, -0.0083, -0.3412, 0.0289, -0.01, -0.3442, 0.0255, -0.01, -0.3444, 0.0276, -0.01, -0.342, 0.0255, -0.01, -0.3422, 0.0285, -0.01, -0.3411, 0.0289, -0.0061, -0.3398, 0.0294, -0.0083, -0.3398, 0.0294, -0.0061, -0.3377, 0.0303, -0.0083, -0.3377, 0.0303, -0.01, -0.3398, 0.0255, -0.01, -0.3398, 0.0277, -0.01, -0.3377, 0.0255, -0.01, -0.3377, 0.0277, -0.01, -0.34, 0.0294, -0.01, -0.3379, 0.0302, -0.0061, -0.3355, 0.0189, -0.0083, -0.3355, 0.0189, -0.0061, -0.3337, 0.0198, -0.0083, -0.3337, 0.0198, -0.0061, -0.3327, 0.0203, -0.0083, -0.3327, 0.0203, -0.01, -0.3353, 0.019, -0.01, -0.3339, 0.0197, -0.01, -0.3355, 0.0212, -0.01, -0.3355, 0.0233, -0.01, -0.3329, 0.0202,
-0.01, -0.3333, 0.0233, -0.0061, -0.3312, 0.0209, -0.0083, -0.3312, 0.0209, -0.0061, -0.329, 0.0218, -0.0083, -0.329, 0.0218, -0.01, -0.3309, 0.021, -0.01, -0.3312, 0.0233, -0.01, -0.3289, 0.0218, -0.01, -0.329, 0.0233, -0.0061, -0.3363, 0.0308, -0.0061, -0.3352, 0.0312, -0.0083, -0.3363, 0.0308, -0.0083, -0.3352, 0.0312, -0.0061, -0.3333, 0.0318, -0.0083, -0.3333, 0.0318, -0.01, -0.3355, 0.0255, -0.01, -0.3355, 0.0277, -0.01, -0.3333, 0.0255, -0.01, -0.3333, 0.0277, -0.01, -0.3359, 0.0309, -0.01, -0.335, 0.0313, -0.01, -0.3333, 0.0298, -0.01, -0.3335, 0.0318, -0.0061, -0.3312, 0.0325, -0.0083, -0.3312, 0.0325, -0.0061, -0.3296, 0.033, -0.0083, -0.3296, 0.033, -0.01, -0.3312, 0.0255, -0.01, -0.3312, 0.0277, -0.01, -0.329, 0.0255,
-0.01, -0.329, 0.0277, -0.01, -0.3312, 0.0298, -0.01, -0.3314, 0.0324, -0.01, -0.329, 0.0298, -0.01, -0.3293, 0.0331, -0.0061, -0.3261, -0.0014, -0.0083, -0.3261, -0.0014, -0.0061, -0.3246, -0.0004, -0.0083, -0.3246, -0.0004, -0.01, -0.3262, -0.0015, -0.01, -0.3268, 0.0016, -0.01, -0.3244, -0.0003, -0.01, -0.3246, 0.0016, -0.01, -0.3268, 0.0038, -0.01, -0.3268, 0.006, -0.01, -0.3246, 0.0038, -0.01, -0.3246, 0.006, -0.0061, -0.3234, 0.0004, -0.0061, -0.3223, 0.0011, -0.0083, -0.3234, 0.0004, -0.0083, -0.3223, 0.0011, -0.0061, -0.3205, 0.0022, -0.0083, -0.3205, 0.0022, -0.0061, -0.3194, 0.0028, -0.0083, -0.3194, 0.0028, -0.01, -0.3234, 0.0004, -0.01, -0.3223, 0.0011, -0.01, -0.3207, 0.0021, -0.01, -0.3225, 0.0038, -0.01, -0.3225, 0.006,
-0.01, -0.3197, 0.0027, -0.01, -0.3203, 0.006, -0.0061, -0.3268, 0.0144, -0.0083, -0.3268, 0.0144, -0.0061, -0.3247, 0.0153, -0.0083, -0.3247, 0.0153, -0.01, -0.3268, 0.0081, -0.01, -0.3268, 0.0103, -0.01, -0.3246, 0.0081, -0.01, -0.3246, 0.0103, -0.01, -0.3268, 0.0125, -0.01, -0.327, 0.0143, -0.01, -0.3246, 0.0125, -0.01, -0.3249, 0.0151, -0.0061, -0.3235, 0.0157, -0.0083, -0.3235, 0.0157, -0.01, -0.3225, 0.0081, -0.01, -0.3225, 0.0103, -0.01, -0.3203, 0.0081, -0.01, -0.3203, 0.0103, -0.01, -0.3225, 0.0125, -0.01, -0.323, 0.0159, -0.01, -0.3203, 0.0125, -0.01, -0.3203, 0.0146, -0.0061, -0.3181, 0.0036, -0.0083, -0.3181, 0.0036, -0.0061, -0.3164, 0.0045, -0.0061, -0.3153, 0.0051, -0.0083, -0.3164, 0.0045, -0.0083, -0.3153, 0.0051,
-0.01, -0.3179, 0.0037, -0.01, -0.3181, 0.006, -0.01, -0.3165, 0.0044, -0.01, -0.3154, 0.005, -0.0061, -0.3138, 0.0059, -0.0083, -0.3138, 0.0059, -0.0061, -0.3122, 0.0068, -0.0083, -0.3122, 0.0068, -0.01, -0.3136, 0.0061, -0.01, -0.3123, 0.0067, -0.01, -0.3181, 0.0081, -0.01, -0.3181, 0.0103, -0.01, -0.316, 0.0081, -0.01, -0.316, 0.0103, -0.01, -0.3181, 0.0125, -0.01, -0.3181, 0.0146, -0.01, -0.316, 0.0125, -0.01, -0.316, 0.0146, -0.0061, -0.3111, 0.0074, -0.0083, -0.3111, 0.0074, -0.01, -0.3138, 0.0081, -0.01, -0.3138, 0.0103, -0.01, -0.3113, 0.0073, -0.01, -0.3116, 0.0103, -0.01, -0.3138, 0.0125, -0.01, -0.3138, 0.0146, -0.01, -0.3116, 0.0125, -0.01, -0.3116, 0.0146, -0.0061, -0.3279, 0.0222, -0.0061, -0.3268, 0.0227,
-0.0083, -0.3279, 0.0222, -0.0083, -0.3268, 0.0227, -0.0061, -0.3246, 0.0235, -0.0083, -0.3246, 0.0235, -0.01, -0.3279, 0.0222, -0.01, -0.3268, 0.0227, -0.01, -0.3245, 0.0236, -0.0061, -0.3224, 0.0161, -0.0083, -0.3224, 0.0161, -0.0061, -0.3203, 0.0169, -0.0083, -0.3203, 0.0169, -0.0061, -0.3229, 0.0242, -0.0083, -0.3229, 0.0242, -0.01, -0.3221, 0.0162, -0.01, -0.3205, 0.0168, -0.01, -0.3231, 0.0241, -0.01, -0.3268, 0.0255, -0.01, -0.3268, 0.0277, -0.01, -0.3246, 0.0255, -0.01, -0.3246, 0.0277, -0.01, -0.3268, 0.0298, -0.01, -0.3268, 0.032, -0.01, -0.3246, 0.0298, -0.01, -0.3246, 0.032, -0.0061, -0.3218, 0.0246, -0.0083, -0.3218, 0.0246, -0.0061, -0.3203, 0.0251, -0.0083, -0.3203, 0.0251, -0.01, -0.3221, 0.0245, -0.01, -0.3225, 0.0277,
-0.01, -0.3201, 0.0251, -0.01, -0.3203, 0.0277, -0.01, -0.3225, 0.0298, -0.01, -0.3225, 0.032, -0.01, -0.3203, 0.0298, -0.01, -0.3203, 0.032, -0.0061, -0.3184, 0.0176, -0.0061, -0.3173, 0.018, -0.0083, -0.3184, 0.0176, -0.0083, -0.3173, 0.018, -0.0061, -0.316, 0.0185, -0.0083, -0.316, 0.0185, -0.01, -0.3183, 0.0176, -0.01, -0.3173, 0.018, -0.01, -0.316, 0.0168, -0.01, -0.3161, 0.0185, -0.0061, -0.3138, 0.0192, -0.0083, -0.3138, 0.0192, -0.0061, -0.3118, 0.0198, -0.0083, -0.3118, 0.0198, -0.0061, -0.3107, 0.0201, -0.0083, -0.3107, 0.0201, -0.01, -0.3138, 0.0168, -0.01, -0.314, 0.0192, -0.01, -0.3116, 0.0168, -0.01, -0.3117, 0.0199, -0.01, -0.3107, 0.0201, -0.0061, -0.3181, 0.0257, -0.0083, -0.3181, 0.0257, -0.0061, -0.3162, 0.0263,
-0.0061, -0.3151, 0.0267, -0.0083, -0.3162, 0.0263, -0.0083, -0.3151, 0.0267, -0.01, -0.318, 0.0257, -0.01, -0.3181, 0.0277, -0.01, -0.3164, 0.0262, -0.01, -0.3155, 0.0265, -0.01, -0.3181, 0.0298, -0.01, -0.3181, 0.032, -0.01, -0.316, 0.0298, -0.01, -0.316, 0.032, -0.0061, -0.3138, 0.0271, -0.0083, -0.3138, 0.0271, -0.0061, -0.3116, 0.0277, -0.0083, -0.3116, 0.0277, -0.01, -0.3136, 0.0271, -0.01, -0.3115, 0.0277, -0.01, -0.3138, 0.0298, -0.01, -0.3138, 0.032, -0.01, -0.3116, 0.0298, -0.01, -0.3116, 0.032, -0.0061, -0.3441, 0.0367, -0.0083, -0.3441, 0.0367, -0.0061, -0.3427, 0.0373, -0.0083, -0.3427, 0.0373, -0.0061, -0.3416, 0.0378, -0.0083, -0.3416, 0.0378, -0.01, -0.3441, 0.0367, -0.01, -0.3428, 0.0372, -0.01, -0.3442, 0.0385,
-0.01, -0.3442, 0.0407, -0.01, -0.3417, 0.0377, -0.01, -0.342, 0.0407, -0.0061, -0.3401, 0.0388, -0.0061, -0.339, 0.04, -0.0083, -0.3401, 0.0388, -0.0083, -0.339, 0.04, -0.0061, -0.3384, 0.0411, -0.0083, -0.3384, 0.0411, -0.01, -0.3401, 0.0388, -0.01, -0.3389, 0.0401, -0.01, -0.3383, 0.0412, -0.01, -0.3442, 0.0429, -0.01, -0.3442, 0.045, -0.01, -0.342, 0.0429, -0.01, -0.342, 0.045, -0.01, -0.3442, 0.0472, -0.01, -0.3442, 0.0494, -0.01, -0.342, 0.0472, -0.01, -0.342, 0.0494, -0.0061, -0.3378, 0.0422, -0.0061, -0.3373, 0.045, -0.0083, -0.3378, 0.0422, -0.0083, -0.3373, 0.045, -0.0061, -0.3373, 0.0472, -0.0061, -0.3377, 0.0494, -0.0083, -0.3373, 0.0472, -0.0083, -0.3377, 0.0494, -0.01, -0.3398, 0.0429, -0.01, -0.3398, 0.045,
-0.01, -0.3376, 0.0426, -0.01, -0.3373, 0.045, -0.01, -0.3398, 0.0472, -0.01, -0.3398, 0.0494, -0.01, -0.3373, 0.0473, -0.01, -0.3378, 0.0496, -0.0061, -0.3286, 0.0333, -0.0083, -0.3286, 0.0333, -0.01, -0.3284, 0.0334, -0.0061, -0.3418, 0.0544, -0.0083, -0.3418, 0.0544, -0.0061, -0.3441, 0.0553, -0.0083, -0.3441, 0.0553, -0.0061, -0.343, 0.0548, -0.0083, -0.343, 0.0548, -0.01, -0.3442, 0.0515, -0.01, -0.3442, 0.0537, -0.01, -0.342, 0.0515, -0.01, -0.3418, 0.0544, -0.01, -0.3441, 0.0553, -0.01, -0.343, 0.0548, -0.0061, -0.3392, 0.0521, -0.0061, -0.3403, 0.0532, -0.0083, -0.3392, 0.0521, -0.0083, -0.3403, 0.0532, -0.0061, -0.3385, 0.0509, -0.0083, -0.3385, 0.0509, -0.01, -0.3391, 0.052, -0.01, -0.3403, 0.0532, -0.01, -0.3385, 0.0508,
-0.0061, -0.3311, 0.0673, -0.0083, -0.3311, 0.0673, -0.0061, -0.3282, 0.0655, -0.0061, -0.3292, 0.0662, -0.0083, -0.3282, 0.0655, -0.0083, -0.3292, 0.0662, -0.01, -0.3309, 0.0672, -0.01, -0.3281, 0.0655, -0.01, -0.3292, 0.0662, -0.0061, -0.3268, 0.0339, -0.0083, -0.3268, 0.0339, -0.0061, -0.3247, 0.0345, -0.0083, -0.3247, 0.0345, -0.01, -0.3269, 0.0338, -0.01, -0.3248, 0.0344, -0.0061, -0.3225, 0.035, -0.0083, -0.3225, 0.035, -0.0061, -0.3212, 0.0352, -0.0061, -0.3201, 0.0354, -0.0083, -0.3212, 0.0352, -0.0083, -0.3201, 0.0354, -0.01, -0.3227, 0.0349, -0.01, -0.3206, 0.0353, -0.01, -0.3198, 0.0355, -0.0061, -0.3181, 0.0358, -0.0083, -0.3181, 0.0358, -0.0061, -0.316, 0.0363, -0.0083, -0.316, 0.0363, -0.01, -0.3181, 0.0342, -0.01, -0.3182, 0.0358,
-0.01, -0.316, 0.0342, -0.01, -0.3161, 0.0363, -0.0061, -0.3138, 0.0366, -0.0083, -0.3138, 0.0366, -0.0061, -0.3116, 0.0369, -0.0083, -0.3116, 0.0369, -0.01, -0.3138, 0.0342, -0.01, -0.3139, 0.0366, -0.01, -0.3116, 0.0342, -0.01, -0.3117, 0.0368, -0.0061, -0.3106, 0.0504, -0.0083, -0.3106, 0.0504, -0.01, -0.3106, 0.0504, -0.0061, -0.3193, 0.059, -0.0083, -0.3193, 0.059, -0.01, -0.3193, 0.059, -0.0061, -0.3243, 0.0629, -0.0083, -0.3243, 0.0629, -0.0061, -0.3268, 0.0646, -0.0083, -0.3268, 0.0646, -0.0061, -0.3254, 0.0637, -0.0083, -0.3254, 0.0637, -0.01, -0.3243, 0.0629, -0.01, -0.3271, 0.0648, -0.01, -0.3268, 0.0667, -0.01, -0.3253, 0.0636, -0.01, -0.3246, 0.0667, -0.0061, -0.3217, 0.061, -0.0061, -0.3228, 0.0619, -0.0083, -0.3217, 0.061,
-0.0083, -0.3228, 0.0619, -0.0061, -0.3204, 0.06, -0.0083, -0.3204, 0.06, -0.01, -0.3217, 0.061, -0.01, -0.3228, 0.0619, -0.01, -0.3204, 0.06, -0.01, -0.3203, 0.0624, -0.01, -0.3225, 0.0646, -0.01, -0.3225, 0.0667, -0.01, -0.3203, 0.0646, -0.01, -0.3203, 0.0667, -0.0061, -0.3149, 0.0548, -0.0083, -0.3149, 0.0548, -0.0061, -0.3171, 0.0569, -0.0061, -0.3182, 0.058, -0.0083, -0.3171, 0.0569, -0.0083, -0.3182, 0.058, -0.0061, -0.316, 0.0558, -0.0083, -0.316, 0.0558, -0.01, -0.3149, 0.0548, -0.01, -0.3171, 0.0569, -0.01, -0.3182, 0.058, -0.01, -0.316, 0.0558, -0.01, -0.316, 0.058, -0.0061, -0.3127, 0.0526, -0.0061, -0.3138, 0.0537, -0.0083, -0.3127, 0.0526, -0.0083, -0.3138, 0.0537, -0.0061, -0.3117, 0.0515, -0.0083, -0.3117, 0.0515,
-0.01, -0.3127, 0.0526, -0.01, -0.3138, 0.0537, -0.01, -0.3117, 0.0515, -0.01, -0.3116, 0.0537, -0.01, -0.3138, 0.0559, -0.01, -0.3138, 0.058, -0.01, -0.3116, 0.0559, -0.01, -0.3116, 0.058, -0.01, -0.3181, 0.0602, -0.01, -0.3181, 0.0624, -0.01, -0.316, 0.0602, -0.01, -0.316, 0.0624, -0.01, -0.3181, 0.0646, -0.01, -0.3181, 0.0667, -0.01, -0.316, 0.0646, -0.01, -0.316, 0.0667, -0.0061, -0.313, 0.0676, -0.0083, -0.313, 0.0676, -0.0061, -0.3107, 0.0655, -0.0061, -0.3118, 0.0665, -0.0083, -0.3107, 0.0655, -0.0083, -0.3118, 0.0665, -0.01, -0.3138, 0.0602, -0.01, -0.3138, 0.0624, -0.01, -0.3116, 0.0602, -0.01, -0.3116, 0.0624, -0.01, -0.3138, 0.0646, -0.01, -0.313, 0.0676, -0.01, -0.3107, 0.0655, -0.01, -0.3118, 0.0665,
-0.0061, -0.3095, 0.0082, -0.0083, -0.3095, 0.0082, -0.0061, -0.3077, 0.0089, -0.0061, -0.3066, 0.0094, -0.0083, -0.3077, 0.0089, -0.0083, -0.3066, 0.0094, -0.01, -0.3093, 0.0083, -0.01, -0.3095, 0.0103, -0.01, -0.3078, 0.0089, -0.01, -0.3069, 0.0093, -0.01, -0.3095, 0.0125, -0.01, -0.3095, 0.0146, -0.01, -0.3073, 0.0125, -0.01, -0.3073, 0.0146, -0.0061, -0.3051, 0.01, -0.0083, -0.3051, 0.01, -0.0061, -0.3029, 0.0108, -0.0083, -0.3029, 0.0108, -0.01, -0.3049, 0.0101, -0.01, -0.3029, 0.0109, -0.01, -0.3051, 0.0125, -0.01, -0.3051, 0.0146, -0.01, -0.3029, 0.0125, -0.01, -0.3029, 0.0146, -0.0061, -0.3017, 0.0113, -0.0083, -0.3017, 0.0113, -0.0061, -0.3006, 0.0118, -0.0083, -0.3006, 0.0118, -0.0061, -0.2986, 0.0125, -0.0083, -0.2986, 0.0125,
-0.01, -0.3017, 0.0113, -0.01, -0.3007, 0.0117, -0.01, -0.3008, 0.0146, -0.01, -0.2985, 0.0125, -0.01, -0.2986, 0.0146, -0.0061, -0.2964, 0.0131, -0.0083, -0.2964, 0.0131, -0.0061, -0.2948, 0.0134, -0.0061, -0.2937, 0.0137, -0.0083, -0.2948, 0.0134, -0.0083, -0.2937, 0.0137, -0.01, -0.2964, 0.0131, -0.01, -0.2964, 0.0146, -0.01, -0.295, 0.0134, -0.01, -0.2941, 0.0136, -0.0061, -0.3095, 0.0205, -0.0083, -0.3095, 0.0205, -0.0061, -0.3073, 0.021, -0.0083, -0.3073, 0.021, -0.01, -0.3095, 0.0168, -0.01, -0.3095, 0.019, -0.01, -0.3073, 0.0168, -0.01, -0.3073, 0.019, -0.01, -0.3095, 0.0205, -0.01, -0.3074, 0.021, -0.0061, -0.3051, 0.0215, -0.0083, -0.3051, 0.0215, -0.0061, -0.3029, 0.022, -0.0083, -0.3029, 0.022, -0.01, -0.3051, 0.0168,
-0.01, -0.3051, 0.019, -0.01, -0.3029, 0.0168, -0.01, -0.3029, 0.019, -0.01, -0.3053, 0.0215, -0.01, -0.3031, 0.0219, -0.0061, -0.3095, 0.0282, -0.0083, -0.3095, 0.0282, -0.0061, -0.3074, 0.0286, -0.0083, -0.3074, 0.0286, -0.0061, -0.3063, 0.0288, -0.0083, -0.3063, 0.0288, -0.01, -0.3094, 0.0282, -0.01, -0.3077, 0.0285, -0.01, -0.3095, 0.0298, -0.01, -0.3095, 0.032, -0.01, -0.3069, 0.0287, -0.01, -0.3073, 0.032, -0.0061, -0.3051, 0.029, -0.0083, -0.3051, 0.029, -0.0061, -0.3029, 0.0293, -0.0083, -0.3029, 0.0293, -0.01, -0.305, 0.029, -0.01, -0.3051, 0.032, -0.01, -0.3028, 0.0293, -0.01, -0.3029, 0.032, -0.0061, -0.3016, 0.0222, -0.0061, -0.3005, 0.0224, -0.0083, -0.3016, 0.0222, -0.0083, -0.3005, 0.0224, -0.0061, -0.2986, 0.0227,
-0.0083, -0.2986, 0.0227, -0.01, -0.3008, 0.0168, -0.01, -0.3008, 0.019, -0.01, -0.2986, 0.0168, -0.01, -0.2986, 0.019, -0.01, -0.301, 0.0223, -0.01, -0.3003, 0.0224, -0.01, -0.2986, 0.0212, -0.01, -0.2986, 0.0227, -0.0061, -0.2964, 0.023, -0.0083, -0.2964, 0.023, -0.0061, -0.2943, 0.0232, -0.0083, -0.2943, 0.0232, -0.01, -0.2964, 0.0168, -0.01, -0.2964, 0.019, -0.01, -0.2943, 0.0168, -0.01, -0.2943, 0.019, -0.01, -0.2964, 0.0212, -0.01, -0.2965, 0.023, -0.01, -0.2943, 0.0212, -0.01, -0.2943, 0.0232, -0.0061, -0.3008, 0.0297, -0.0083, -0.3008, 0.0297, -0.0061, -0.2986, 0.0299, -0.0083, -0.2986, 0.0299, -0.01, -0.3007, 0.0297, -0.01, -0.3008, 0.032, -0.01, -0.2986, 0.03, -0.01, -0.2986, 0.032, -0.0061, -0.2964, 0.03,
-0.0083, -0.2964, 0.03, -0.0061, -0.2943, 0.0301, -0.0083, -0.2943, 0.0301, -0.01, -0.2964, 0.03, -0.01, -0.2964, 0.032, -0.01, -0.2943, 0.0301, -0.01, -0.2943, 0.032, -0.0061, -0.2889, 0.0064, -0.0083, -0.2889, 0.0064, -0.01, -0.289, 0.0066, -0.0061, -0.2882, -0.0005, -0.0061, -0.2883, 0.0016, -0.0083, -0.2882, -0.0005, -0.0083, -0.2883, 0.0016, -0.0061, -0.2885, 0.0038, -0.0061, -0.2888, 0.0053, -0.0083, -0.2885, 0.0038, -0.0083, -0.2888, 0.0053, -0.01, -0.2882, -0.0006, -0.01, -0.2883, 0.0016, -0.01, -0.2856, -0.0005, -0.01, -0.2856, 0.0016, -0.01, -0.2885, 0.0037, -0.01, -0.2888, 0.0058, -0.01, -0.2856, 0.0038, -0.01, -0.2856, 0.006, -0.0061, -0.2892, 0.0081, -0.0061, -0.2896, 0.0103, -0.0083, -0.2892, 0.0081, -0.0083, -0.2896, 0.0103,
-0.0061, -0.2921, 0.014, -0.0083, -0.2921, 0.014, -0.0061, -0.29, 0.0125, -0.0061, -0.2902, 0.0144, -0.0083, -0.29, 0.0125, -0.0083, -0.2902, 0.0144, -0.01, -0.2892, 0.0081, -0.01, -0.2896, 0.0102, -0.01, -0.2919, 0.014, -0.01, -0.29, 0.0124, -0.01, -0.2902, 0.0144, -0.01, -0.2878, 0.0081, -0.01, -0.2878, 0.0103, -0.01, -0.2856, 0.0081, -0.01, -0.2856, 0.0103, -0.01, -0.2878, 0.0125, -0.01, -0.2878, 0.0146, -0.01, -0.2856, 0.0125, -0.01, -0.2856, 0.0146, -0.0061, -0.2802, 0.0069, -0.0083, -0.2802, 0.0069, -0.01, -0.2834, -0.0005, -0.01, -0.2834, 0.0016, -0.01, -0.2812, -0.0005, -0.01, -0.2812, 0.0016, -0.01, -0.2834, 0.0038, -0.01, -0.2834, 0.006, -0.01, -0.2812, 0.0038, -0.01, -0.2801, 0.0063, -0.0061, -0.2794, -0.0005,
-0.0061, -0.2795, 0.0016, -0.0083, -0.2794, -0.0005, -0.0083, -0.2795, 0.0016, -0.0061, -0.2797, 0.0038, -0.0061, -0.28, 0.0058, -0.0083, -0.2797, 0.0038, -0.0083, -0.28, 0.0058, -0.01, -0.2794, -0.0005, -0.01, -0.2795, 0.0017, -0.01, -0.2797, 0.0038, -0.01, -0.28, 0.0055, -0.0061, -0.2803, 0.0081, -0.0061, -0.2806, 0.0103, -0.0083, -0.2803, 0.0081, -0.0083, -0.2806, 0.0103, -0.0061, -0.281, 0.0125, -0.0061, -0.2813, 0.0146, -0.0083, -0.281, 0.0125, -0.0083, -0.2813, 0.0146, -0.01, -0.2834, 0.0081, -0.01, -0.2834, 0.0103, -0.01, -0.2804, 0.0083, -0.01, -0.2807, 0.0104, -0.01, -0.2834, 0.0125, -0.01, -0.2834, 0.0146, -0.01, -0.281, 0.0126, -0.01, -0.2813, 0.0147, -0.0061, -0.2921, 0.0232, -0.0083, -0.2921, 0.0232, -0.0061, -0.2899, 0.0232,
-0.0083, -0.2899, 0.0232, -0.01, -0.2921, 0.0168, -0.01, -0.2921, 0.019, -0.01, -0.2899, 0.0168, -0.01, -0.2899, 0.019, -0.01, -0.2921, 0.0212, -0.01, -0.2921, 0.0232, -0.01, -0.2899, 0.0212, -0.01, -0.2899, 0.0232, -0.0061, -0.2878, 0.0232, -0.0083, -0.2878, 0.0232, -0.0061, -0.2856, 0.0232, -0.0083, -0.2856, 0.0232, -0.01, -0.2878, 0.0168, -0.01, -0.2878, 0.019, -0.01, -0.2856, 0.0168, -0.01, -0.2856, 0.019, -0.01, -0.2878, 0.0212, -0.01, -0.2878, 0.0232, -0.01, -0.2856, 0.0212, -0.01, -0.2856, 0.0232, -0.0061, -0.2921, 0.0301, -0.0083, -0.2921, 0.0301, -0.0061, -0.2899, 0.0301, -0.0083, -0.2899, 0.0301, -0.01, -0.2921, 0.0301, -0.01, -0.2921, 0.032, -0.01, -0.2899, 0.0301, -0.01, -0.2899, 0.032, -0.0061, -0.2878, 0.0301,
-0.0083, -0.2878, 0.0301, -0.0061, -0.2856, 0.03, -0.0083, -0.2856, 0.03, -0.01, -0.2878, 0.0301, -0.01, -0.2878, 0.032, -0.01, -0.2856, 0.03, -0.01, -0.2856, 0.032, -0.0061, -0.2814, 0.0163, -0.0083, -0.2814, 0.0163, -0.0061, -0.2834, 0.0232, -0.0083, -0.2834, 0.0232, -0.0061, -0.2812, 0.0232, -0.0083, -0.2812, 0.0232, -0.01, -0.2834, 0.0168, -0.01, -0.2834, 0.019, -0.01, -0.2814, 0.0163, -0.01, -0.2812, 0.019, -0.01, -0.2834, 0.0212, -0.01, -0.2834, 0.0232, -0.01, -0.2812, 0.0212, -0.01, -0.2812, 0.0232, -0.0061, -0.2791, 0.0163, -0.0083, -0.2791, 0.0163, -0.0061, -0.2769, 0.0163, -0.0083, -0.2769, 0.0163, -0.0061, -0.2791, 0.0232, -0.0083, -0.2791, 0.0232, -0.0061, -0.2769, 0.023, -0.0083, -0.2769, 0.023, -0.01, -0.2791, 0.0163,
-0.01, -0.2791, 0.019, -0.01, -0.2769, 0.0163, -0.01, -0.2769, 0.019, -0.01, -0.2791, 0.0212, -0.01, -0.279, 0.0232, -0.01, -0.2769, 0.0212, -0.01, -0.2769, 0.023, -0.0061, -0.2834, 0.0299, -0.0083, -0.2834, 0.0299, -0.0061, -0.2812, 0.0296, -0.0083, -0.2812, 0.0296, -0.01, -0.2835, 0.0299, -0.01, -0.2834, 0.032, -0.01, -0.2813, 0.0296, -0.01, -0.2812, 0.032, -0.0061, -0.2764, 0.0286, -0.0083, -0.2764, 0.0286, -0.0061, -0.2791, 0.0292, -0.0083, -0.2791, 0.0292, -0.0061, -0.2774, 0.0289, -0.0083, -0.2774, 0.0289, -0.01, -0.2762, 0.0286, -0.01, -0.2792, 0.0292, -0.01, -0.2791, 0.032, -0.01, -0.2771, 0.0288, -0.01, -0.2769, 0.032, -0.0061, -0.3095, 0.037, -0.0083, -0.3095, 0.037, -0.0061, -0.3073, 0.0371, -0.0083, -0.3073, 0.0371,
-0.01, -0.3095, 0.0342, -0.01, -0.3095, 0.037, -0.01, -0.3073, 0.0342, -0.01, -0.3074, 0.0371, -0.0061, -0.3058, 0.0374, -0.0083, -0.3058, 0.0374, -0.0061, -0.3047, 0.0375, -0.0083, -0.3047, 0.0375, -0.0061, -0.301, 0.038, -0.0061, -0.3029, 0.0407, -0.0083, -0.301, 0.038, -0.0083, -0.3029, 0.0407, -0.01, -0.3051, 0.0342, -0.01, -0.3053, 0.0374, -0.01, -0.3029, 0.0342, -0.01, -0.3029, 0.0363, -0.01, -0.3045, 0.0376, -0.01, -0.301, 0.038, -0.01, -0.3027, 0.0405, -0.0061, -0.3065, 0.0457, -0.0083, -0.3065, 0.0457, -0.0061, -0.3085, 0.0481, -0.0061, -0.3096, 0.0493, -0.0083, -0.3085, 0.0481, -0.0083, -0.3096, 0.0493, -0.0061, -0.3076, 0.0471, -0.0083, -0.3076, 0.0471, -0.01, -0.3065, 0.0457, -0.01, -0.3085, 0.0481, -0.01, -0.3096, 0.0493,
-0.01, -0.3076, 0.047, -0.01, -0.3073, 0.0494, -0.0061, -0.3046, 0.0432, -0.0061, -0.3057, 0.0446, -0.0083, -0.3046, 0.0432, -0.0083, -0.3057, 0.0446, -0.0061, -0.3038, 0.0421, -0.0083, -0.3038, 0.0421, -0.01, -0.3047, 0.0432, -0.01, -0.3057, 0.0447, -0.01, -0.3039, 0.0422, -0.01, -0.3029, 0.045, -0.01, -0.3051, 0.0472, -0.01, -0.3051, 0.0494, -0.01, -0.3029, 0.0472, -0.01, -0.3029, 0.0494, -0.01, -0.3008, 0.0342, -0.01, -0.3008, 0.0363, -0.01, -0.2986, 0.0342, -0.01, -0.2986, 0.0363, -0.01, -0.3013, 0.0384, -0.01, -0.3008, 0.0407, -0.01, -0.2986, 0.0385, -0.01, -0.2986, 0.0407, -0.01, -0.2964, 0.0342, -0.01, -0.2964, 0.0363, -0.01, -0.2943, 0.0342, -0.01, -0.2943, 0.0363, -0.01, -0.2964, 0.0385, -0.01, -0.2964, 0.0407,
-0.01, -0.2943, 0.0385, -0.01, -0.2943, 0.0407, -0.01, -0.3008, 0.0429, -0.01, -0.3008, 0.045, -0.01, -0.2986, 0.0429, -0.01, -0.2986, 0.045, -0.01, -0.3008, 0.0472, -0.01, -0.3008, 0.0494, -0.01, -0.2986, 0.0472, -0.01, -0.2986, 0.0494, -0.0061, -0.2932, 0.0438, -0.0061, -0.294, 0.045, -0.0083, -0.2932, 0.0438, -0.0083, -0.294, 0.045, -0.0061, -0.2957, 0.0477, -0.0061, -0.2968, 0.0494, -0.0083, -0.2957, 0.0477, -0.0083, -0.2968, 0.0494, -0.0061, -0.295, 0.0467, -0.0083, -0.295, 0.0467, -0.01, -0.2964, 0.0429, -0.01, -0.2964, 0.045, -0.01, -0.2931, 0.0436, -0.01, -0.2942, 0.0453, -0.01, -0.2956, 0.0477, -0.01, -0.2969, 0.0496, -0.01, -0.295, 0.0466, -0.01, -0.3095, 0.0515, -0.01, -0.3095, 0.0537, -0.01, -0.3073, 0.0515,
-0.01, -0.3073, 0.0537, -0.01, -0.3095, 0.0559, -0.01, -0.3095, 0.058, -0.01, -0.3073, 0.0559, -0.01, -0.3073, 0.058, -0.0061, -0.3043, 0.0589, -0.0083, -0.3043, 0.0589, -0.0061, -0.3022, 0.0565, -0.0061, -0.3033, 0.0578, -0.0083, -0.3022, 0.0565, -0.0083, -0.3033, 0.0578, -0.01, -0.3051, 0.0515, -0.01, -0.3051, 0.0537, -0.01, -0.3029, 0.0515, -0.01, -0.3029, 0.0537, -0.01, -0.3051, 0.0559, -0.01, -0.3042, 0.0588, -0.01, -0.3022, 0.0565, -0.01, -0.3033, 0.0577, -0.0061, -0.3084, 0.0634, -0.0083, -0.3084, 0.0634, -0.0061, -0.3063, 0.0612, -0.0061, -0.3074, 0.0624, -0.0083, -0.3063, 0.0612, -0.0083, -0.3074, 0.0624, -0.0061, -0.3095, 0.0644, -0.0083, -0.3095, 0.0644, -0.01, -0.3095, 0.0602, -0.01, -0.3084, 0.0634, -0.01, -0.3063, 0.0611,
-0.01, -0.3073, 0.0623, -0.01, -0.3095, 0.0645, -0.0061, -0.3053, 0.0601, -0.0083, -0.3053, 0.0601, -0.01, -0.3053, 0.0601, -0.0061, -0.3002, 0.0541, -0.0083, -0.3002, 0.0541, -0.0061, -0.2983, 0.0516, -0.0061, -0.2994, 0.053, -0.0083, -0.2983, 0.0516, -0.0083, -0.2994, 0.053, -0.0061, -0.3013, 0.0554, -0.0083, -0.3013, 0.0554, -0.01, -0.3008, 0.0515, -0.01, -0.3002, 0.0541, -0.01, -0.2983, 0.0516, -0.01, -0.2994, 0.053, -0.01, -0.3013, 0.0554, -0.0061, -0.2975, 0.0505, -0.0083, -0.2975, 0.0505, -0.01, -0.2975, 0.0505, -0.0061, -0.2902, 0.0369, -0.0083, -0.2902, 0.0369, -0.0061, -0.291, 0.0396, -0.0061, -0.2915, 0.0407, -0.0083, -0.291, 0.0396, -0.0083, -0.2915, 0.0407, -0.0061, -0.2906, 0.0385, -0.0083, -0.2906, 0.0385, -0.01, -0.2921, 0.0342,
-0.01, -0.2921, 0.0363, -0.01, -0.2899, 0.0342, -0.01, -0.2902, 0.0369, -0.01, -0.2908, 0.0391, -0.01, -0.2916, 0.041, -0.01, -0.2905, 0.0382, -0.0061, -0.2878, 0.0369, -0.0083, -0.2878, 0.0369, -0.0061, -0.2856, 0.0367, -0.0083, -0.2856, 0.0367, -0.01, -0.2878, 0.0342, -0.01, -0.2877, 0.0369, -0.01, -0.2856, 0.0342, -0.01, -0.2855, 0.0367, -0.0061, -0.2926, 0.0428, -0.0083, -0.2926, 0.0428, -0.01, -0.2925, 0.0426, -0.0061, -0.2834, 0.0363, -0.0083, -0.2834, 0.0363, -0.0061, -0.2812, 0.0359, -0.0083, -0.2812, 0.0359, -0.01, -0.2834, 0.0342, -0.01, -0.2833, 0.0363, -0.01, -0.2812, 0.0342, -0.01, -0.2812, 0.0359, -0.0061, -0.2783, 0.0352, -0.0061, -0.2794, 0.0355, -0.0083, -0.2783, 0.0352, -0.0083, -0.2794, 0.0355, -0.0061, -0.2769, 0.0348,
-0.0083, -0.2769, 0.0348, -0.01, -0.2787, 0.0353, -0.01, -0.2796, 0.0355, -0.01, -0.2767, 0.0347, -0.0061, -0.3442, 0.0737, -0.0083, -0.3442, 0.0737, -0.0061, -0.342, 0.0728, -0.0083, -0.342, 0.0728, -0.01, -0.3443, 0.0738, -0.01, -0.3442, 0.0754, -0.01, -0.3423, 0.0729, -0.01, -0.342, 0.0754, -0.0061, -0.3397, 0.0717, -0.0083, -0.3397, 0.0717, -0.0061, -0.3377, 0.0707, -0.0083, -0.3377, 0.0707, -0.0061, -0.3408, 0.0722, -0.0083, -0.3408, 0.0722, -0.01, -0.3395, 0.0716, -0.01, -0.338, 0.0708, -0.01, -0.3405, 0.0721, -0.01, -0.3398, 0.0754, -0.01, -0.3377, 0.0732, -0.01, -0.3377, 0.0754, -0.0061, -0.3419, 0.0848, -0.0083, -0.3419, 0.0848, -0.01, -0.3442, 0.0776, -0.01, -0.3442, 0.0798, -0.01, -0.342, 0.0776, -0.01, -0.342, 0.0798,
-0.01, -0.3442, 0.0819, -0.01, -0.3442, 0.0841, -0.01, -0.342, 0.0819, -0.01, -0.3419, 0.0849, -0.0061, -0.3398, 0.0841, -0.0083, -0.3398, 0.0841, -0.0061, -0.3368, 0.0829, -0.0061, -0.3379, 0.0834, -0.0083, -0.3368, 0.0829, -0.0083, -0.3379, 0.0834, -0.01, -0.3398, 0.0776, -0.01, -0.3398, 0.0798, -0.01, -0.3377, 0.0776, -0.01, -0.3377, 0.0798, -0.01, -0.3398, 0.0819, -0.01, -0.3396, 0.0841, -0.01, -0.3371, 0.0831, -0.01, -0.3381, 0.0835, -0.0061, -0.3353, 0.0695, -0.0061, -0.3364, 0.0701, -0.0083, -0.3353, 0.0695, -0.0083, -0.3364, 0.0701, -0.0061, -0.3333, 0.0685, -0.0083, -0.3333, 0.0685, -0.01, -0.3351, 0.0694, -0.01, -0.3361, 0.0699, -0.01, -0.3336, 0.0687, -0.01, -0.3333, 0.0711, -0.01, -0.3355, 0.0732, -0.01, -0.3355, 0.0754,
-0.01, -0.3333, 0.0732, -0.01, -0.3333, 0.0754, -0.0061, -0.3321, 0.0679, -0.0083, -0.3321, 0.0679, -0.01, -0.3319, 0.0677, -0.01, -0.3312, 0.0711, -0.01, -0.329, 0.0689, -0.01, -0.329, 0.0711, -0.01, -0.3312, 0.0732, -0.01, -0.3312, 0.0754, -0.01, -0.329, 0.0732, -0.01, -0.329, 0.0754, -0.0061, -0.3324, 0.0808, -0.0083, -0.3324, 0.0808, -0.0061, -0.3355, 0.0823, -0.0083, -0.3355, 0.0823, -0.0061, -0.3335, 0.0813, -0.0083, -0.3335, 0.0813, -0.01, -0.3355, 0.0776, -0.01, -0.3355, 0.0798, -0.01, -0.3333, 0.0776, -0.01, -0.3327, 0.0809, -0.01, -0.3352, 0.0822, -0.01, -0.3336, 0.0814, -0.0061, -0.3311, 0.0802, -0.0083, -0.3311, 0.0802, -0.0061, -0.3282, 0.0785, -0.0061, -0.3293, 0.0791, -0.0083, -0.3282, 0.0785, -0.0083, -0.3293, 0.0791,
-0.01, -0.3312, 0.0776, -0.01, -0.3308, 0.08, -0.01, -0.3284, 0.0786, -0.01, -0.3294, 0.0792, -0.0061, -0.3442, 0.0856, -0.0083, -0.3442, 0.0856, -0.0061, -0.343, 0.0852, -0.0083, -0.343, 0.0852, -0.01, -0.3441, 0.0856, -0.01, -0.343, 0.0852, -0.0061, -0.3242, 0.0761, -0.0083, -0.3242, 0.0761, -0.01, -0.3268, 0.0689, -0.01, -0.3268, 0.0711, -0.01, -0.3246, 0.0689, -0.01, -0.3246, 0.0711, -0.01, -0.3268, 0.0732, -0.01, -0.3268, 0.0754, -0.01, -0.3246, 0.0732, -0.01, -0.3243, 0.0761, -0.0061, -0.3215, 0.0742, -0.0061, -0.3226, 0.075, -0.0083, -0.3215, 0.0742, -0.0083, -0.3226, 0.075, -0.0061, -0.3203, 0.0734, -0.0083, -0.3203, 0.0734, -0.01, -0.3225, 0.0689, -0.01, -0.3225, 0.0711, -0.01, -0.3203, 0.0689, -0.01, -0.3203, 0.0711,
-0.01, -0.3217, 0.0743, -0.01, -0.3227, 0.0751, -0.01, -0.32, 0.0732, -0.0061, -0.3268, 0.0777, -0.0083, -0.3268, 0.0777, -0.0061, -0.3253, 0.0768, -0.0083, -0.3253, 0.0768, -0.01, -0.3265, 0.0775, -0.01, -0.3254, 0.0768, -0.0061, -0.3178, 0.0716, -0.0083, -0.3178, 0.0716, -0.0061, -0.3153, 0.0696, -0.0061, -0.3164, 0.0705, -0.0083, -0.3153, 0.0696, -0.0083, -0.3164, 0.0705, -0.0061, -0.3189, 0.0724, -0.0083, -0.3189, 0.0724, -0.01, -0.3181, 0.0689, -0.01, -0.3178, 0.0716, -0.01, -0.3153, 0.0696, -0.01, -0.3164, 0.0706, -0.01, -0.3189, 0.0724, -0.0061, -0.3141, 0.0685, -0.0083, -0.3141, 0.0685, -0.01, -0.3141, 0.0686, -0.0061, -0.2653, 0.0157, -0.0083, -0.2653, 0.0157, -0.0061, -0.2639, 0.0154, -0.0083, -0.2639, 0.0154, -0.01, -0.2652, 0.0157,
-0.01, -0.2639, 0.0154, -0.0061, -0.2617, 0.015, -0.0083, -0.2617, 0.015, -0.0061, -0.2595, 0.0146, -0.0083, -0.2595, 0.0146, -0.01, -0.2618, 0.015, -0.01, -0.2596, 0.0146, -0.0061, -0.2747, 0.0163, -0.0083, -0.2747, 0.0163, -0.0061, -0.2726, 0.0163, -0.0083, -0.2726, 0.0163, -0.0061, -0.2747, 0.0226, -0.0083, -0.2747, 0.0226, -0.0061, -0.2723, 0.022, -0.0061, -0.2734, 0.0223, -0.0083, -0.2723, 0.022, -0.0083, -0.2734, 0.0223, -0.01, -0.2747, 0.0163, -0.01, -0.2747, 0.019, -0.01, -0.2726, 0.0163, -0.01, -0.2726, 0.019, -0.01, -0.2747, 0.0212, -0.01, -0.2747, 0.0226, -0.01, -0.2725, 0.0221, -0.01, -0.2735, 0.0223, -0.0061, -0.2704, 0.0163, -0.0083, -0.2704, 0.0163, -0.0061, -0.2682, 0.0161, -0.0083, -0.2682, 0.0161, -0.0061, -0.2704, 0.0216,
-0.0083, -0.2704, 0.0216, -0.0061, -0.2682, 0.0211, -0.0083, -0.2682, 0.0211, -0.01, -0.2704, 0.0163, -0.01, -0.2704, 0.019, -0.01, -0.2683, 0.0161, -0.01, -0.2682, 0.019, -0.01, -0.2702, 0.0216, -0.01, -0.2681, 0.0211, -0.0061, -0.2747, 0.0283, -0.0083, -0.2747, 0.0283, -0.0061, -0.2726, 0.0277, -0.0083, -0.2726, 0.0277, -0.0061, -0.2717, 0.033, -0.0083, -0.2717, 0.033, -0.01, -0.2748, 0.0283, -0.01, -0.2727, 0.0277, -0.01, -0.2747, 0.0298, -0.01, -0.2747, 0.032, -0.01, -0.2726, 0.0298, -0.01, -0.2721, 0.0331, -0.0061, -0.2704, 0.027, -0.0083, -0.2704, 0.027, -0.0061, -0.2681, 0.0263, -0.0061, -0.2692, 0.0266, -0.0083, -0.2681, 0.0263, -0.0083, -0.2692, 0.0266, -0.0061, -0.2704, 0.0325, -0.0083, -0.2704, 0.0325, -0.0061, -0.2682, 0.0317,
-0.0083, -0.2682, 0.0317, -0.01, -0.2706, 0.0271, -0.01, -0.2678, 0.0262, -0.01, -0.2687, 0.0265, -0.01, -0.2704, 0.0298, -0.01, -0.2701, 0.0324, -0.01, -0.2682, 0.0298, -0.01, -0.2681, 0.0316, -0.0061, -0.2664, 0.0159, -0.0083, -0.2664, 0.0159, -0.0061, -0.2635, 0.0198, -0.0083, -0.2635, 0.0198, -0.0061, -0.266, 0.0206, -0.0083, -0.266, 0.0206, -0.0061, -0.2646, 0.0202, -0.0083, -0.2646, 0.0202, -0.01, -0.2661, 0.0158, -0.01, -0.2661, 0.019, -0.01, -0.2639, 0.0168, -0.01, -0.2637, 0.0199, -0.01, -0.266, 0.0206, -0.01, -0.2647, 0.0202, -0.0061, -0.2617, 0.0193, -0.0083, -0.2617, 0.0193, -0.0061, -0.2595, 0.0185, -0.0083, -0.2595, 0.0185, -0.0061, -0.2616, 0.0241, -0.0083, -0.2616, 0.0241, -0.0061, -0.2595, 0.0232, -0.0083, -0.2595, 0.0232,
-0.01, -0.2617, 0.0168, -0.01, -0.2615, 0.0192, -0.01, -0.2595, 0.0168, -0.01, -0.2594, 0.0185, -0.01, -0.2614, 0.024, -0.01, -0.2598, 0.0233, -0.0061, -0.2661, 0.0256, -0.0083, -0.2661, 0.0256, -0.0061, -0.2639, 0.0249, -0.0083, -0.2639, 0.0249, -0.0061, -0.2657, 0.0306, -0.0061, -0.2668, 0.0311, -0.0083, -0.2657, 0.0306, -0.0083, -0.2668, 0.0311, -0.0061, -0.2639, 0.0298, -0.0083, -0.2639, 0.0298, -0.01, -0.2662, 0.0256, -0.01, -0.2661, 0.0277, -0.01, -0.2641, 0.025, -0.01, -0.2639, 0.0277, -0.01, -0.2658, 0.0307, -0.01, -0.2669, 0.0311, -0.01, -0.2637, 0.0297, -0.0061, -0.2627, 0.0245, -0.0061, -0.2611, 0.0285, -0.0083, -0.2627, 0.0245, -0.0083, -0.2611, 0.0285, -0.0061, -0.2595, 0.0276, -0.0083, -0.2595, 0.0276, -0.0061, -0.2622, 0.029,
-0.0083, -0.2622, 0.029, -0.01, -0.2622, 0.0243, -0.01, -0.2613, 0.0286, -0.01, -0.2595, 0.0255, -0.01, -0.2593, 0.0275, -0.01, -0.2623, 0.0291, -0.0061, -0.2574, 0.0142, -0.0083, -0.2574, 0.0142, -0.0061, -0.2542, 0.0136, -0.0061, -0.2552, 0.0138, -0.0083, -0.2542, 0.0136, -0.0083, -0.2552, 0.0138, -0.01, -0.2575, 0.0142, -0.01, -0.2541, 0.0135, -0.01, -0.2551, 0.0138, -0.0061, -0.253, 0.0133, -0.0061, -0.2526, 0.0154, -0.0083, -0.253, 0.0133, -0.0083, -0.2526, 0.0154, -0.0061, -0.2509, 0.0128, -0.0061, -0.2508, 0.0145, -0.0083, -0.2509, 0.0128, -0.0083, -0.2508, 0.0145, -0.01, -0.2531, 0.0133, -0.01, -0.2527, 0.0155, -0.01, -0.2509, 0.0128, -0.01, -0.2506, 0.0144, -0.0061, -0.2467, 0.0121, -0.0061, -0.2496, 0.0137, -0.0083, -0.2467, 0.0121,
-0.0083, -0.2496, 0.0137, -0.01, -0.2497, 0.0131, -0.01, -0.2496, 0.0137, -0.0061, -0.2571, 0.0176, -0.0061, -0.2582, 0.018, -0.0083, -0.2571, 0.0176, -0.0083, -0.2582, 0.018, -0.0061, -0.2552, 0.0167, -0.0083, -0.2552, 0.0167, -0.0061, -0.2569, 0.022, -0.0061, -0.258, 0.0225, -0.0083, -0.2569, 0.022, -0.0083, -0.258, 0.0225, -0.0061, -0.2552, 0.0211, -0.0061, -0.2542, 0.0244, -0.0083, -0.2552, 0.0211, -0.0083, -0.2542, 0.0244, -0.01, -0.2572, 0.0176, -0.01, -0.2583, 0.018, -0.01, -0.255, 0.0166, -0.01, -0.2568, 0.0219, -0.01, -0.2578, 0.0224, -0.01, -0.2554, 0.0212, -0.01, -0.2544, 0.0245, -0.0061, -0.2537, 0.016, -0.0061, -0.2525, 0.0198, -0.0083, -0.2537, 0.016, -0.0083, -0.2525, 0.0198, -0.0061, -0.2509, 0.019, -0.0083, -0.2509, 0.019,
-0.0061, -0.2536, 0.0203, -0.0061, -0.253, 0.0235, -0.0083, -0.2536, 0.0203, -0.0083, -0.253, 0.0235, -0.0061, -0.2505, 0.0217, -0.0061, -0.2516, 0.0225, -0.0083, -0.2505, 0.0217, -0.0083, -0.2516, 0.0225, -0.01, -0.2538, 0.016, -0.01, -0.2524, 0.0197, -0.01, -0.2511, 0.0191, -0.01, -0.2534, 0.0203, -0.01, -0.2527, 0.0233, -0.01, -0.2505, 0.0217, -0.01, -0.2516, 0.0225, -0.0061, -0.257, 0.0262, -0.0061, -0.2581, 0.0268, -0.0083, -0.257, 0.0262, -0.0083, -0.2581, 0.0268, -0.0061, -0.2553, 0.0251, -0.0083, -0.2553, 0.0251, -0.01, -0.257, 0.0262, -0.01, -0.2581, 0.0268, -0.01, -0.2554, 0.0252, -0.0061, -0.2482, 0.0176, -0.0061, -0.2486, 0.0189, -0.0083, -0.2482, 0.0176, -0.0083, -0.2486, 0.0189, -0.0061, -0.2449, 0.0163, -0.0061, -0.2469, 0.0186,
-0.0083, -0.2449, 0.0163, -0.0083, -0.2469, 0.0186, -0.0061, -0.2491, 0.0206, -0.0083, -0.2491, 0.0206, -0.01, -0.248, 0.0176, -0.01, -0.2486, 0.019, -0.01, -0.2449, 0.0163, -0.01, -0.2469, 0.0186, -0.01, -0.2491, 0.0207, -0.0061, -0.2747, 0.0341, -0.0083, -0.2747, 0.0341, -0.0061, -0.2728, 0.0334, -0.0083, -0.2728, 0.0334, -0.01, -0.2746, 0.034, -0.01, -0.273, 0.0335, -0.0061, -0.2196, -0.0005, -0.0061, -0.2196, 0.0016, -0.0083, -0.2196, -0.0005, -0.0083, -0.2196, 0.0016, -0.0061, -0.2196, 0.0038, -0.0061, -0.2196, 0.006, -0.0083, -0.2196, 0.0038, -0.0083, -0.2196, 0.006, -0.01, -0.2196, -0.0005, -0.01, -0.2196, 0.0016, -0.01, -0.2196, 0.0038, -0.01, -0.2196, 0.006, -0.01, -0.2183, -0.0005, -0.01, -0.2183, 0.0016, -0.01, -0.2161, -0.0005,
-0.01, -0.2161, 0.0016, -0.01, -0.2183, 0.0038, -0.01, -0.2183, 0.006, -0.01, -0.2161, 0.0038, -0.01, -0.2161, 0.006, -0.0061, -0.2196, 0.0081, -0.0061, -0.2196, 0.0103, -0.0083, -0.2196, 0.0081, -0.0083, -0.2196, 0.0103, -0.0061, -0.2196, 0.0125, -0.0061, -0.2196, 0.0146, -0.0083, -0.2196, 0.0125, -0.0083, -0.2196, 0.0146, -0.01, -0.2196, 0.0081, -0.01, -0.2196, 0.0103, -0.01, -0.2196, 0.0125, -0.01, -0.2196, 0.0146, -0.01, -0.2183, 0.0081, -0.01, -0.2183, 0.0103, -0.01, -0.2161, 0.0081, -0.01, -0.2161, 0.0103, -0.01, -0.2183, 0.0125, -0.01, -0.2183, 0.0146, -0.01, -0.2161, 0.0125, -0.01, -0.2161, 0.0146, -0.01, -0.214, -0.0005, -0.01, -0.214, 0.0016, -0.01, -0.2118, -0.0005, -0.01, -0.2118, 0.0016, -0.01, -0.214, 0.0038,
-0.01, -0.214, 0.006, -0.01, -0.2118, 0.0038, -0.01, -0.2118, 0.006, -0.0061, -0.2088, -0.0005, -0.0061, -0.2088, 0.0016, -0.0083, -0.2088, -0.0005, -0.0083, -0.2088, 0.0016, -0.0061, -0.2088, 0.0038, -0.0061, -0.2088, 0.006, -0.0083, -0.2088, 0.0038, -0.0083, -0.2088, 0.006, -0.01, -0.2088, -0.0005, -0.01, -0.2088, 0.0016, -0.01, -0.2088, 0.0038, -0.01, -0.2088, 0.006, -0.01, -0.214, 0.0081, -0.01, -0.214, 0.0103, -0.01, -0.2118, 0.0081, -0.01, -0.2118, 0.0103, -0.01, -0.214, 0.0125, -0.01, -0.214, 0.0146, -0.01, -0.2118, 0.0125, -0.01, -0.2118, 0.0146, -0.0061, -0.2088, 0.0085, -0.0083, -0.2088, 0.0085, -0.0061, -0.2075, 0.0085, -0.0083, -0.2075, 0.0085, -0.01, -0.2088, 0.0085, -0.01, -0.2096, 0.0103, -0.01, -0.2075, 0.0085,
-0.01, -0.2075, 0.0103, -0.01, -0.2096, 0.0125, -0.01, -0.2096, 0.0146, -0.01, -0.2075, 0.0125, -0.01, -0.2075, 0.0146, -0.0061, -0.2196, 0.0168, -0.0061, -0.2196, 0.019, -0.0083, -0.2196, 0.0168, -0.0083, -0.2196, 0.019, -0.0061, -0.2196, 0.0212, -0.0061, -0.2196, 0.0233, -0.0083, -0.2196, 0.0212, -0.0083, -0.2196, 0.0233, -0.01, -0.2196, 0.0168, -0.01, -0.2196, 0.019, -0.01, -0.2196, 0.0212, -0.01, -0.2196, 0.0233, -0.01, -0.2183, 0.0168, -0.01, -0.2183, 0.019, -0.01, -0.2161, 0.0168, -0.01, -0.2161, 0.019, -0.01, -0.2183, 0.0212, -0.01, -0.2183, 0.0233, -0.01, -0.2161, 0.0212, -0.01, -0.2161, 0.0233, -0.0061, -0.2196, 0.0255, -0.0061, -0.2196, 0.0277, -0.0083, -0.2196, 0.0255, -0.0083, -0.2196, 0.0277, -0.0061, -0.2196, 0.0298,
-0.0061, -0.2196, 0.032, -0.0083, -0.2196, 0.0298, -0.0083, -0.2196, 0.032, -0.01, -0.2196, 0.0255, -0.01, -0.2196, 0.0277, -0.01, -0.2196, 0.0298, -0.01, -0.2196, 0.032, -0.01, -0.2183, 0.0255, -0.01, -0.2183, 0.0277, -0.01, -0.2161, 0.0255, -0.01, -0.2161, 0.0277, -0.01, -0.2183, 0.0298, -0.01, -0.2183, 0.032, -0.01, -0.2161, 0.0298, -0.01, -0.2161, 0.032, -0.01, -0.214, 0.0168, -0.01, -0.214, 0.019, -0.01, -0.2118, 0.0168, -0.01, -0.2118, 0.019, -0.01, -0.214, 0.0212, -0.01, -0.214, 0.0233, -0.01, -0.2118, 0.0212, -0.01, -0.2118, 0.0233, -0.0061, -0.2088, 0.0203, -0.0061, -0.2088, 0.0233, -0.0083, -0.2088, 0.0203, -0.0083, -0.2088, 0.0233, -0.0061, -0.2075, 0.0203, -0.0083, -0.2075, 0.0203, -0.01, -0.2096, 0.0168,
-0.01, -0.2096, 0.019, -0.01, -0.2075, 0.0168, -0.01, -0.2075, 0.019, -0.01, -0.2088, 0.0203, -0.01, -0.2088, 0.0233, -0.01, -0.2075, 0.0203, -0.01, -0.214, 0.0255, -0.01, -0.214, 0.0277, -0.01, -0.2118, 0.0255, -0.01, -0.2118, 0.0277, -0.01, -0.214, 0.0298, -0.01, -0.214, 0.032, -0.01, -0.2118, 0.0298, -0.01, -0.2118, 0.032, -0.0061, -0.2088, 0.0255, -0.0061, -0.2088, 0.0277, -0.0083, -0.2088, 0.0255, -0.0083, -0.2088, 0.0277, -0.0061, -0.2088, 0.0298, -0.0061, -0.2088, 0.032, -0.0083, -0.2088, 0.0298, -0.0083, -0.2088, 0.032, -0.01, -0.2088, 0.0255, -0.01, -0.2088, 0.0277, -0.01, -0.2088, 0.0298, -0.01, -0.2088, 0.0319, -0.0061, -0.2196, 0.0342, -0.0061, -0.2196, 0.0363, -0.0083, -0.2196, 0.0342, -0.0083, -0.2196, 0.0363,
-0.0061, -0.2196, 0.0385, -0.0061, -0.2196, 0.0407, -0.0083, -0.2196, 0.0385, -0.0083, -0.2196, 0.0407, -0.01, -0.2196, 0.0342, -0.01, -0.2196, 0.0363, -0.01, -0.2196, 0.0385, -0.01, -0.2196, 0.0407, -0.01, -0.2183, 0.0342, -0.01, -0.2183, 0.0363, -0.01, -0.2161, 0.0342, -0.01, -0.2161, 0.0363, -0.01, -0.2183, 0.0385, -0.01, -0.2183, 0.0407, -0.01, -0.2161, 0.0385, -0.01, -0.2161, 0.0407, -0.0061, -0.2196, 0.0429, -0.0061, -0.2196, 0.045, -0.0083, -0.2196, 0.0429, -0.0083, -0.2196, 0.045, -0.0061, -0.2196, 0.0472, -0.0061, -0.2196, 0.0487, -0.0083, -0.2196, 0.0472, -0.0083, -0.2196, 0.0487, -0.01, -0.2196, 0.0429, -0.01, -0.2196, 0.045, -0.01, -0.2196, 0.0472, -0.01, -0.2196, 0.0487, -0.0061, -0.2183, 0.0487, -0.0083, -0.2183, 0.0487,
-0.0061, -0.2161, 0.0487, -0.0083, -0.2161, 0.0487, -0.01, -0.2183, 0.0429, -0.01, -0.2183, 0.045, -0.01, -0.2161, 0.0429, -0.01, -0.2161, 0.045, -0.01, -0.2183, 0.0472, -0.01, -0.2183, 0.0487, -0.01, -0.2161, 0.0472, -0.01, -0.2161, 0.0487, -0.01, -0.214, 0.0342, -0.01, -0.214, 0.0363, -0.01, -0.2118, 0.0342, -0.01, -0.2118, 0.0363, -0.01, -0.214, 0.0385, -0.01, -0.214, 0.0407, -0.01, -0.2118, 0.0385, -0.01, -0.2118, 0.0407, -0.0061, -0.2086, 0.0334, -0.0083, -0.2086, 0.0334, -0.0061, -0.2083, 0.0345, -0.0061, -0.2076, 0.0361, -0.0083, -0.2083, 0.0345, -0.0083, -0.2076, 0.0361, -0.01, -0.2085, 0.0339, -0.01, -0.2096, 0.0363, -0.01, -0.2083, 0.0347, -0.01, -0.2076, 0.0361, -0.01, -0.2096, 0.0385, -0.01, -0.2096, 0.0407,
-0.01, -0.2075, 0.0385, -0.01, -0.2075, 0.0407, -0.0061, -0.214, 0.0487, -0.0083, -0.214, 0.0487, -0.0061, -0.2118, 0.0487, -0.0083, -0.2118, 0.0487, -0.01, -0.214, 0.0429, -0.01, -0.214, 0.045, -0.01, -0.2118, 0.0429, -0.01, -0.2118, 0.045, -0.01, -0.214, 0.0472, -0.01, -0.214, 0.0487, -0.01, -0.2118, 0.0472, -0.01, -0.2118, 0.0487, -0.0061, -0.2096, 0.0487, -0.0083, -0.2096, 0.0487, -0.0061, -0.2075, 0.0487, -0.0083, -0.2075, 0.0487, -0.01, -0.2096, 0.0429, -0.01, -0.2096, 0.045, -0.01, -0.2075, 0.0429, -0.01, -0.2075, 0.045, -0.01, -0.2096, 0.0472, -0.01, -0.2096, 0.0487, -0.01, -0.2075, 0.0472, -0.01, -0.2075, 0.0487, -0.0061, -0.2053, 0.0085, -0.0083, -0.2053, 0.0085, -0.0061, -0.2031, 0.0085, -0.0083, -0.2031, 0.0085,
-0.01, -0.2053, 0.0085, -0.01, -0.2053, 0.0103, -0.01, -0.2031, 0.0085, -0.01, -0.2031, 0.0103, -0.01, -0.2053, 0.0125, -0.01, -0.2053, 0.0146, -0.01, -0.2031, 0.0125, -0.01, -0.2031, 0.0146, -0.0061, -0.2009, 0.0085, -0.0083, -0.2009, 0.0085, -0.0061, -0.1988, 0.0085, -0.0083, -0.1988, 0.0085, -0.01, -0.2009, 0.0085, -0.01, -0.2009, 0.0103, -0.01, -0.1988, 0.0085, -0.01, -0.1988, 0.0103, -0.01, -0.2009, 0.0125, -0.01, -0.2009, 0.0146, -0.01, -0.1988, 0.0125, -0.01, -0.1988, 0.0146, -0.0061, -0.1966, 0.0085, -0.0083, -0.1966, 0.0085, -0.0061, -0.1944, 0.0085, -0.0083, -0.1944, 0.0085, -0.01, -0.1966, 0.0085, -0.01, -0.1966, 0.0103, -0.01, -0.1944, 0.0085, -0.01, -0.1944, 0.0103, -0.01, -0.1966, 0.0125, -0.01, -0.1966, 0.0146,
-0.01, -0.1944, 0.0125, -0.01, -0.1944, 0.0146, -0.0061, -0.1923, 0.0085, -0.0083, -0.1923, 0.0085, -0.0061, -0.1901, 0.0085, -0.0083, -0.1901, 0.0085, -0.01, -0.1923, 0.0085, -0.01, -0.1923, 0.0103, -0.01, -0.1901, 0.0085, -0.01, -0.1901, 0.0103, -0.01, -0.1923, 0.0125, -0.01, -0.1923, 0.0146, -0.01, -0.1901, 0.0125, -0.01, -0.1901, 0.0146, -0.0061, -0.2053, 0.0203, -0.0083, -0.2053, 0.0203, -0.0061, -0.2031, 0.0203, -0.0083, -0.2031, 0.0203, -0.01, -0.2053, 0.0168, -0.01, -0.2053, 0.019, -0.01, -0.2031, 0.0168, -0.01, -0.2031, 0.019, -0.01, -0.2053, 0.0203, -0.01, -0.2031, 0.0203, -0.0061, -0.2009, 0.0203, -0.0083, -0.2009, 0.0203, -0.0061, -0.1988, 0.0203, -0.0083, -0.1988, 0.0203, -0.01, -0.2009, 0.0168, -0.01, -0.2009, 0.019,
-0.01, -0.1988, 0.0168, -0.01, -0.1988, 0.019, -0.01, -0.2009, 0.0203, -0.01, -0.1988, 0.0203, -0.0061, -0.1966, 0.0203, -0.0083, -0.1966, 0.0203, -0.0061, -0.1944, 0.0203, -0.0083, -0.1944, 0.0203, -0.01, -0.1966, 0.0168, -0.01, -0.1966, 0.019, -0.01, -0.1944, 0.0168, -0.01, -0.1944, 0.019, -0.01, -0.1966, 0.0203, -0.01, -0.1944, 0.0203, -0.0061, -0.1923, 0.0203, -0.0083, -0.1923, 0.0203, -0.0061, -0.1901, 0.0203, -0.0083, -0.1901, 0.0203, -0.01, -0.1923, 0.0168, -0.01, -0.1923, 0.019, -0.01, -0.1901, 0.0168, -0.01, -0.1901, 0.019, -0.01, -0.1923, 0.0203, -0.01, -0.1901, 0.0203, -0.0061, -0.1879, 0.0085, -0.0083, -0.1879, 0.0085, -0.0061, -0.1858, 0.0085, -0.0083, -0.1858, 0.0085, -0.01, -0.1879, 0.0085, -0.01, -0.1879, 0.0103,
-0.01, -0.1858, 0.0085, -0.01, -0.1858, 0.0103, -0.01, -0.1879, 0.0125, -0.01, -0.1879, 0.0146, -0.01, -0.1858, 0.0125, -0.01, -0.1858, 0.0146, -0.0061, -0.1836, 0.0085, -0.0083, -0.1836, 0.0085, -0.0061, -0.1814, 0.0085, -0.0083, -0.1814, 0.0085, -0.01, -0.1836, 0.0085, -0.01, -0.1836, 0.0103, -0.01, -0.1814, 0.0085, -0.01, -0.1814, 0.0103, -0.01, -0.1836, 0.0125, -0.01, -0.1836, 0.0146, -0.01, -0.1814, 0.0125, -0.01, -0.1814, 0.0146, -0.0061, -0.1792, 0.0085, -0.0083, -0.1792, 0.0085, -0.0061, -0.1771, 0.0085, -0.0083, -0.1771, 0.0085, -0.01, -0.1792, 0.0085, -0.01, -0.1792, 0.0103, -0.01, -0.1771, 0.0085, -0.01, -0.1771, 0.0103, -0.01, -0.1792, 0.0125, -0.01, -0.1792, 0.0146, -0.01, -0.1771, 0.0125, -0.01, -0.1771, 0.0146,
-0.0061, -0.1749, 0.0085, -0.0083, -0.1749, 0.0085, -0.0061, -0.1727, 0.0085, -0.0083, -0.1727, 0.0085, -0.01, -0.1749, 0.0085, -0.01, -0.1749, 0.0103, -0.01, -0.1727, 0.0085, -0.01, -0.1727, 0.0103, -0.01, -0.1749, 0.0125, -0.01, -0.1749, 0.0146, -0.01, -0.1727, 0.0125, -0.01, -0.1727, 0.0146, -0.0061, -0.1879, 0.0203, -0.0083, -0.1879, 0.0203, -0.0061, -0.1858, 0.0203, -0.0083, -0.1858, 0.0203, -0.01, -0.1879, 0.0168, -0.01, -0.1879, 0.019, -0.01, -0.1858, 0.0168, -0.01, -0.1858, 0.019, -0.01, -0.1879, 0.0203, -0.01, -0.1858, 0.0203, -0.0061, -0.1836, 0.0203, -0.0083, -0.1836, 0.0203, -0.0061, -0.1814, 0.0203, -0.0083, -0.1814, 0.0203, -0.01, -0.1836, 0.0168, -0.01, -0.1836, 0.019, -0.01, -0.1814, 0.0168, -0.01, -0.1814, 0.019,
-0.01, -0.1836, 0.0203, -0.01, -0.1814, 0.0203, -0.0061, -0.1792, 0.0203, -0.0083, -0.1792, 0.0203, -0.0061, -0.1771, 0.0203, -0.0083, -0.1771, 0.0203, -0.01, -0.1792, 0.0168, -0.01, -0.1792, 0.019, -0.01, -0.1771, 0.0168, -0.01, -0.1771, 0.019, -0.01, -0.1792, 0.0203, -0.01, -0.1771, 0.0203, -0.0061, -0.1749, 0.0203, -0.0083, -0.1749, 0.0203, -0.0061, -0.1727, 0.0203, -0.0083, -0.1727, 0.0203, -0.01, -0.1749, 0.0168, -0.01, -0.1749, 0.019, -0.01, -0.1727, 0.0168, -0.01, -0.1727, 0.019, -0.01, -0.1749, 0.0203, -0.01, -0.1727, 0.0203, -0.0061, -0.2059, 0.0372, -0.0083, -0.2059, 0.0372, -0.0061, -0.2048, 0.0376, -0.0083, -0.2048, 0.0376, -0.0061, -0.2031, 0.0378, -0.0083, -0.2031, 0.0378, -0.01, -0.206, 0.0372, -0.01, -0.2051, 0.0375,
-0.01, -0.2053, 0.0407, -0.01, -0.203, 0.0378, -0.01, -0.2031, 0.0407, -0.0061, -0.2009, 0.0379, -0.0083, -0.2009, 0.0379, -0.0061, -0.1988, 0.0379, -0.0083, -0.1988, 0.0379, -0.01, -0.2009, 0.0379, -0.01, -0.2009, 0.0407, -0.01, -0.1988, 0.0379, -0.01, -0.1988, 0.0407, -0.0061, -0.2053, 0.0487, -0.0083, -0.2053, 0.0487, -0.0061, -0.2031, 0.0487, -0.0083, -0.2031, 0.0487, -0.01, -0.2053, 0.0429, -0.01, -0.2053, 0.045, -0.01, -0.2031, 0.0429, -0.01, -0.2031, 0.045, -0.01, -0.2053, 0.0472, -0.01, -0.2053, 0.0487, -0.01, -0.2031, 0.0472, -0.01, -0.2031, 0.0487, -0.0061, -0.2009, 0.0487, -0.0083, -0.2009, 0.0487, -0.0061, -0.1988, 0.0487, -0.0083, -0.1988, 0.0487, -0.01, -0.2009, 0.0429, -0.01, -0.2009, 0.045, -0.01, -0.1988, 0.0429,
-0.01, -0.1988, 0.045, -0.01, -0.2009, 0.0472, -0.01, -0.2009, 0.0487, -0.01, -0.1988, 0.0472, -0.01, -0.1988, 0.0487, -0.0061, -0.1966, 0.0379, -0.0083, -0.1966, 0.0379, -0.0061, -0.1944, 0.0379, -0.0083, -0.1944, 0.0379, -0.01, -0.1966, 0.0379, -0.01, -0.1966, 0.0407, -0.01, -0.1944, 0.0379, -0.01, -0.1944, 0.0407, -0.0061, -0.1923, 0.0379, -0.0083, -0.1923, 0.0379, -0.0061, -0.1901, 0.0379, -0.0083, -0.1901, 0.0379, -0.01, -0.1923, 0.0379, -0.01, -0.1923, 0.0407, -0.01, -0.1901, 0.0379, -0.01, -0.1901, 0.0407, -0.0061, -0.1966, 0.0487, -0.0083, -0.1966, 0.0487, -0.0061, -0.1944, 0.0487, -0.0083, -0.1944, 0.0487, -0.01, -0.1966, 0.0429, -0.01, -0.1966, 0.045, -0.01, -0.1944, 0.0429, -0.01, -0.1944, 0.045, -0.01, -0.1966, 0.0472,
-0.01, -0.1966, 0.0487, -0.01, -0.1944, 0.0472, -0.01, -0.1944, 0.0487, -0.0061, -0.1923, 0.0487, -0.0083, -0.1923, 0.0487, -0.0061, -0.1901, 0.0487, -0.0083, -0.1901, 0.0487, -0.01, -0.1923, 0.0429, -0.01, -0.1923, 0.045, -0.01, -0.1901, 0.0429, -0.01, -0.1901, 0.045, -0.01, -0.1923, 0.0472, -0.01, -0.1923, 0.0487, -0.01, -0.1901, 0.0472, -0.01, -0.1901, 0.0487, -0.0061, -0.1879, 0.0379, -0.0083, -0.1879, 0.0379, -0.0061, -0.1858, 0.0379, -0.0083, -0.1858, 0.0379, -0.01, -0.1879, 0.0379, -0.01, -0.1879, 0.0407, -0.01, -0.1858, 0.0379, -0.01, -0.1858, 0.0407, -0.0061, -0.1836, 0.0379, -0.0083, -0.1836, 0.0379, -0.0061, -0.1814, 0.0379, -0.0083, -0.1814, 0.0379, -0.01, -0.1836, 0.0379, -0.01, -0.1836, 0.0407, -0.01, -0.1814, 0.0379,
-0.01, -0.1814, 0.0407, -0.0061, -0.1879, 0.0487, -0.0083, -0.1879, 0.0487, -0.0061, -0.1858, 0.0487, -0.0083, -0.1858, 0.0487, -0.01, -0.1879, 0.0429, -0.01, -0.1879, 0.045, -0.01, -0.1858, 0.0429, -0.01, -0.1858, 0.045, -0.01, -0.1879, 0.0472, -0.01, -0.1879, 0.0487, -0.01, -0.1858, 0.0472, -0.01, -0.1858, 0.0487, -0.0061, -0.1836, 0.0487, -0.0083, -0.1836, 0.0487, -0.0061, -0.1814, 0.0487, -0.0083, -0.1814, 0.0487, -0.01, -0.1836, 0.0429, -0.01, -0.1836, 0.045, -0.01, -0.1814, 0.0429, -0.01, -0.1814, 0.045, -0.01, -0.1836, 0.0472, -0.01, -0.1836, 0.0487, -0.01, -0.1814, 0.0472, -0.01, -0.1814, 0.0487, -0.0061, -0.1792, 0.0379, -0.0083, -0.1792, 0.0379, -0.0061, -0.1771, 0.0379, -0.0083, -0.1771, 0.0379, -0.01, -0.1792, 0.0379,
-0.01, -0.1792, 0.0407, -0.01, -0.1771, 0.0379, -0.01, -0.1771, 0.0407, -0.0061, -0.1749, 0.0379, -0.0083, -0.1749, 0.0379, -0.0061, -0.1727, 0.0379, -0.0083, -0.1727, 0.0379, -0.01, -0.1749, 0.0379, -0.01, -0.1749, 0.0407, -0.01, -0.1727, 0.0379, -0.01, -0.1727, 0.0407, -0.0061, -0.1792, 0.0487, -0.0083, -0.1792, 0.0487, -0.0061, -0.1771, 0.0487, -0.0083, -0.1771, 0.0487, -0.01, -0.1792, 0.0429, -0.01, -0.1792, 0.045, -0.01, -0.1771, 0.0429, -0.01, -0.1771, 0.045, -0.01, -0.1792, 0.0472, -0.01, -0.1792, 0.0487, -0.01, -0.1771, 0.0472, -0.01, -0.1771, 0.0487, -0.0061, -0.1749, 0.0487, -0.0083, -0.1749, 0.0487, -0.0061, -0.1727, 0.0487, -0.0083, -0.1727, 0.0487, -0.01, -0.1749, 0.0429, -0.01, -0.1749, 0.045, -0.01, -0.1727, 0.0429,
-0.01, -0.1727, 0.045, -0.01, -0.1749, 0.0472, -0.01, -0.1749, 0.0487, -0.01, -0.1727, 0.0472, -0.01, -0.1727, 0.0487, -0.0061, -0.1706, 0.0085, -0.0083, -0.1706, 0.0085, -0.0061, -0.1684, 0.0085, -0.0083, -0.1684, 0.0085, -0.01, -0.1706, 0.0085, -0.01, -0.1706, 0.0103, -0.01, -0.1684, 0.0085, -0.01, -0.1684, 0.0103, -0.01, -0.1706, 0.0125, -0.01, -0.1706, 0.0146, -0.01, -0.1684, 0.0125, -0.01, -0.1684, 0.0146, -0.0061, -0.1662, 0.0085, -0.0083, -0.1662, 0.0085, -0.0061, -0.1641, 0.0085, -0.0083, -0.1641, 0.0085, -0.01, -0.1662, 0.0085, -0.01, -0.1662, 0.0103, -0.01, -0.1641, 0.0085, -0.01, -0.1641, 0.0103, -0.01, -0.1662, 0.0125, -0.01, -0.1662, 0.0146, -0.01, -0.1641, 0.0125, -0.01, -0.1641, 0.0146, -0.0061, -0.1627, 0.0085,
-0.0061, -0.1627, 0.0103, -0.0083, -0.1627, 0.0085, -0.0083, -0.1627, 0.0103, -0.0061, -0.1627, 0.0125, -0.0061, -0.1627, 0.0146, -0.0083, -0.1627, 0.0125, -0.0083, -0.1627, 0.0146, -0.01, -0.1627, 0.0085, -0.01, -0.1627, 0.0103, -0.01, -0.1627, 0.0125, -0.01, -0.1627, 0.0146, -0.0061, -0.1706, 0.0203, -0.0083, -0.1706, 0.0203, -0.0061, -0.1684, 0.0203, -0.0083, -0.1684, 0.0203, -0.01, -0.1706, 0.0168, -0.01, -0.1706, 0.019, -0.01, -0.1684, 0.0168, -0.01, -0.1684, 0.019, -0.01, -0.1706, 0.0203, -0.01, -0.1684, 0.0203, -0.0061, -0.1662, 0.0203, -0.0083, -0.1662, 0.0203, -0.0061, -0.1641, 0.0203, -0.0083, -0.1641, 0.0203, -0.01, -0.1662, 0.0168, -0.01, -0.1662, 0.019, -0.01, -0.1641, 0.0168, -0.01, -0.1641, 0.019, -0.01, -0.1662, 0.0203,
-0.01, -0.1641, 0.0203, -0.0061, -0.1627, 0.0168, -0.0061, -0.1627, 0.019, -0.0083, -0.1627, 0.0168, -0.0083, -0.1627, 0.019, -0.0061, -0.1627, 0.0203, -0.0083, -0.1627, 0.0203, -0.01, -0.1627, 0.0168, -0.01, -0.1627, 0.019, -0.01, -0.1627, 0.0203, -0.0061, -0.151, -0.0005, -0.0061, -0.151, 0.0016, -0.0083, -0.151, -0.0005, -0.0083, -0.151, 0.0016, -0.0061, -0.151, 0.0038, -0.0061, -0.151, 0.006, -0.0083, -0.151, 0.0038, -0.0083, -0.151, 0.006, -0.01, -0.151, -0.0005, -0.01, -0.151, 0.0016, -0.01, -0.151, 0.0038, -0.01, -0.151, 0.006, -0.01, -0.1489, -0.0005, -0.01, -0.1489, 0.0016, -0.01, -0.1467, -0.0005, -0.01, -0.1467, 0.0016, -0.01, -0.1489, 0.0038, -0.01, -0.1489, 0.006, -0.01, -0.1467, 0.0038, -0.01, -0.1467, 0.006,
-0.0061, -0.151, 0.0081, -0.0061, -0.151, 0.0103, -0.0083, -0.151, 0.0081, -0.0083, -0.151, 0.0103, -0.0061, -0.151, 0.0125, -0.0061, -0.151, 0.0146, -0.0083, -0.151, 0.0125, -0.0083, -0.151, 0.0146, -0.01, -0.151, 0.0081, -0.01, -0.151, 0.0103, -0.01, -0.151, 0.0125, -0.01, -0.151, 0.0146, -0.01, -0.1489, 0.0081, -0.01, -0.1489, 0.0103, -0.01, -0.1467, 0.0081, -0.01, -0.1467, 0.0103, -0.01, -0.1489, 0.0125, -0.01, -0.1489, 0.0146, -0.01, -0.1467, 0.0125, -0.01, -0.1467, 0.0146, -0.01, -0.1445, -0.0005, -0.01, -0.1445, 0.0016, -0.01, -0.1424, -0.0005, -0.01, -0.1424, 0.0016, -0.01, -0.1445, 0.0038, -0.01, -0.1445, 0.006, -0.01, -0.1424, 0.0038, -0.01, -0.1424, 0.006, -0.0061, -0.1392, -0.0005, -0.0061, -0.1392, 0.0016,
-0.0083, -0.1392, -0.0005, -0.0083, -0.1392, 0.0016, -0.0061, -0.1392, 0.0038, -0.0061, -0.1392, 0.006, -0.0083, -0.1392, 0.0038, -0.0083, -0.1392, 0.006, -0.01, -0.1392, -0.0005, -0.01, -0.1392, 0.0016, -0.01, -0.1392, 0.0038, -0.01, -0.1392, 0.006, -0.01, -0.1445, 0.0081, -0.01, -0.1445, 0.0103, -0.01, -0.1424, 0.0081, -0.01, -0.1424, 0.0103, -0.01, -0.1445, 0.0125, -0.01, -0.1445, 0.0146, -0.01, -0.1424, 0.0125, -0.01, -0.1424, 0.0146, -0.0061, -0.1392, 0.0081, -0.0061, -0.1392, 0.0103, -0.0083, -0.1392, 0.0081, -0.0083, -0.1392, 0.0103, -0.0061, -0.1392, 0.0125, -0.0061, -0.1392, 0.0146, -0.0083, -0.1392, 0.0125, -0.0083, -0.1392, 0.0146, -0.01, -0.1392, 0.0081, -0.01, -0.1392, 0.0103, -0.01, -0.1392, 0.0125, -0.01, -0.1392, 0.0146,
-0.0061, -0.151, 0.0168, -0.0061, -0.151, 0.019, -0.0083, -0.151, 0.0168, -0.0083, -0.151, 0.019, -0.0061, -0.151, 0.0212, -0.0061, -0.151, 0.0233, -0.0083, -0.151, 0.0212, -0.0083, -0.151, 0.0233, -0.01, -0.151, 0.0168, -0.01, -0.151, 0.019, -0.01, -0.151, 0.0212, -0.01, -0.151, 0.0233, -0.01, -0.1489, 0.0168, -0.01, -0.1489, 0.019, -0.01, -0.1467, 0.0168, -0.01, -0.1467, 0.019, -0.01, -0.1489, 0.0212, -0.01, -0.1489, 0.0233, -0.01, -0.1467, 0.0212, -0.01, -0.1467, 0.0233, -0.0061, -0.151, 0.0255, -0.0061, -0.151, 0.0277, -0.0083, -0.151, 0.0255, -0.0083, -0.151, 0.0277, -0.0061, -0.151, 0.0298, -0.0061, -0.151, 0.032, -0.0083, -0.151, 0.0298, -0.0083, -0.151, 0.032, -0.01, -0.151, 0.0255, -0.01, -0.151, 0.0277,
-0.01, -0.151, 0.0298, -0.01, -0.151, 0.032, -0.01, -0.1489, 0.0255, -0.01, -0.1489, 0.0277, -0.01, -0.1467, 0.0255, -0.01, -0.1467, 0.0277, -0.01, -0.1489, 0.0298, -0.01, -0.1489, 0.032, -0.01, -0.1467, 0.0298, -0.01, -0.1467, 0.032, -0.01, -0.1445, 0.0168, -0.01, -0.1445, 0.019, -0.01, -0.1424, 0.0168, -0.01, -0.1424, 0.019, -0.01, -0.1445, 0.0212, -0.01, -0.1445, 0.0233, -0.01, -0.1424, 0.0212, -0.01, -0.1424, 0.0233, -0.0061, -0.1392, 0.0168, -0.0061, -0.1392, 0.019, -0.0083, -0.1392, 0.0168, -0.0083, -0.1392, 0.019, -0.0061, -0.1392, 0.0212, -0.0061, -0.1392, 0.0233, -0.0083, -0.1392, 0.0212, -0.0083, -0.1392, 0.0233, -0.01, -0.1392, 0.0168, -0.01, -0.1392, 0.019, -0.01, -0.1392, 0.0212, -0.01, -0.1392, 0.0233,
-0.01, -0.1445, 0.0255, -0.01, -0.1445, 0.0277, -0.01, -0.1424, 0.0255, -0.01, -0.1424, 0.0277, -0.01, -0.1445, 0.0298, -0.01, -0.1445, 0.032, -0.01, -0.1424, 0.0298, -0.01, -0.1424, 0.032, -0.0061, -0.1392, 0.0255, -0.0061, -0.1392, 0.0277, -0.0083, -0.1392, 0.0255, -0.0083, -0.1392, 0.0277, -0.0061, -0.1392, 0.0298, -0.0061, -0.1392, 0.032, -0.0083, -0.1392, 0.0298, -0.0083, -0.1392, 0.032, -0.01, -0.1392, 0.0255, -0.01, -0.1392, 0.0277, -0.01, -0.1392, 0.0298, -0.01, -0.1392, 0.032, -0.0061, -0.1706, 0.0379, -0.0083, -0.1706, 0.0379, -0.0061, -0.1684, 0.0379, -0.0083, -0.1684, 0.0379, -0.01, -0.1706, 0.0379, -0.01, -0.1706, 0.0407, -0.01, -0.1684, 0.0379, -0.01, -0.1684, 0.0407, -0.0061, -0.1662, 0.0379, -0.0083, -0.1662, 0.0379,
-0.0061, -0.1641, 0.0379, -0.0083, -0.1641, 0.0379, -0.01, -0.1662, 0.0379, -0.01, -0.1662, 0.0407, -0.01, -0.1641, 0.0379, -0.01, -0.1641, 0.0407, -0.0061, -0.1706, 0.0487, -0.0083, -0.1706, 0.0487, -0.0061, -0.1684, 0.0487, -0.0083, -0.1684, 0.0487, -0.01, -0.1706, 0.0429, -0.01, -0.1706, 0.045, -0.01, -0.1684, 0.0429, -0.01, -0.1684, 0.045, -0.01, -0.1706, 0.0472, -0.01, -0.1706, 0.0487, -0.01, -0.1684, 0.0472, -0.01, -0.1684, 0.0487, -0.0061, -0.1662, 0.0487, -0.0083, -0.1662, 0.0487, -0.0061, -0.1641, 0.0487, -0.0083, -0.1641, 0.0487, -0.01, -0.1662, 0.0429, -0.01, -0.1662, 0.045, -0.01, -0.1641, 0.0429, -0.01, -0.1641, 0.045, -0.01, -0.1662, 0.0472, -0.01, -0.1662, 0.0487, -0.01, -0.1641, 0.0472, -0.01, -0.1641, 0.0487,
-0.0061, -0.1619, 0.0379, -0.0083, -0.1619, 0.0379, -0.0061, -0.1588, 0.0379, -0.0061, -0.1588, 0.0407, -0.0083, -0.1588, 0.0379, -0.0083, -0.1588, 0.0407, -0.01, -0.1619, 0.0379, -0.01, -0.1619, 0.0407, -0.01, -0.1588, 0.0379, -0.01, -0.1588, 0.0407, -0.0061, -0.1588, 0.0429, -0.0061, -0.1588, 0.045, -0.0083, -0.1588, 0.0429, -0.0083, -0.1588, 0.045, -0.0061, -0.1619, 0.0487, -0.0083, -0.1619, 0.0487, -0.0061, -0.1588, 0.0472, -0.0061, -0.1588, 0.0487, -0.0083, -0.1588, 0.0472, -0.0083, -0.1588, 0.0487, -0.01, -0.1619, 0.0429, -0.01, -0.1619, 0.045, -0.01, -0.1588, 0.0429, -0.01, -0.1588, 0.045, -0.01, -0.1619, 0.0472, -0.01, -0.1619, 0.0487, -0.01, -0.1588, 0.0472, -0.01, -0.1588, 0.0487, -0.0061, -0.151, 0.0342, -0.0061, -0.151, 0.0363,
-0.0083, -0.151, 0.0342, -0.0083, -0.151, 0.0363, -0.0061, -0.151, 0.0385, -0.0061, -0.151, 0.0407, -0.0083, -0.151, 0.0385, -0.0083, -0.151, 0.0407, -0.01, -0.151, 0.0342, -0.01, -0.151, 0.0363, -0.01, -0.151, 0.0385, -0.01, -0.151, 0.0407, -0.01, -0.1489, 0.0342, -0.01, -0.1489, 0.0363, -0.01, -0.1467, 0.0342, -0.01, -0.1467, 0.0363, -0.01, -0.1489, 0.0385, -0.01, -0.1489, 0.0407, -0.01, -0.1467, 0.0385, -0.01, -0.1467, 0.0407, -0.0061, -0.151, 0.0429, -0.0061, -0.151, 0.045, -0.0083, -0.151, 0.0429, -0.0083, -0.151, 0.045, -0.0061, -0.151, 0.0472, -0.0061, -0.151, 0.0487, -0.0083, -0.151, 0.0472, -0.0083, -0.151, 0.0487, -0.01, -0.151, 0.0429, -0.01, -0.151, 0.045, -0.01, -0.151, 0.0472, -0.01, -0.151, 0.0487,
-0.0061, -0.1489, 0.0487, -0.0083, -0.1489, 0.0487, -0.0061, -0.1467, 0.0487, -0.0083, -0.1467, 0.0487, -0.01, -0.1489, 0.0429, -0.01, -0.1489, 0.045, -0.01, -0.1467, 0.0429, -0.01, -0.1467, 0.045, -0.01, -0.1489, 0.0472, -0.01, -0.1489, 0.0487, -0.01, -0.1467, 0.0472, -0.01, -0.1467, 0.0487, -0.01, -0.1445, 0.0342, -0.01, -0.1445, 0.0363, -0.01, -0.1424, 0.0342, -0.01, -0.1424, 0.0363, -0.01, -0.1445, 0.0385, -0.01, -0.1445, 0.0407, -0.01, -0.1424, 0.0385, -0.01, -0.1424, 0.0407, -0.0061, -0.1392, 0.0342, -0.0061, -0.1392, 0.0363, -0.0083, -0.1392, 0.0342, -0.0083, -0.1392, 0.0363, -0.0061, -0.1392, 0.0385, -0.0061, -0.1392, 0.0407, -0.0083, -0.1392, 0.0385, -0.0083, -0.1392, 0.0407, -0.01, -0.1392, 0.0342, -0.01, -0.1392, 0.0363,
-0.01, -0.1392, 0.0385, -0.01, -0.1392, 0.0407, -0.0061, -0.1445, 0.0487, -0.0083, -0.1445, 0.0487, -0.0061, -0.1424, 0.0487, -0.0083, -0.1424, 0.0487, -0.01, -0.1445, 0.0429, -0.01, -0.1445, 0.045, -0.01, -0.1424, 0.0429, -0.01, -0.1424, 0.045, -0.01, -0.1445, 0.0472, -0.01, -0.1445, 0.0487, -0.01, -0.1424, 0.0472, -0.01, -0.1424, 0.0487, -0.0061, -0.1392, 0.0429, -0.0061, -0.1392, 0.045, -0.0083, -0.1392, 0.0429, -0.0083, -0.1392, 0.045, -0.0061, -0.1392, 0.0472, -0.0061, -0.1392, 0.0487, -0.0083, -0.1392, 0.0472, -0.0083, -0.1392, 0.0487, -0.01, -0.1392, 0.0429, -0.01, -0.1392, 0.045, -0.01, -0.1392, 0.0472, -0.01, -0.1392, 0.0487, -0.0061, -0.0775, -0.0005, -0.0061, -0.0775, 0.0016, -0.0083, -0.0775, -0.0005, -0.0083, -0.0775, 0.0016,
-0.0061, -0.0775, 0.0038, -0.0061, -0.0775, 0.006, -0.0083, -0.0775, 0.0038, -0.0083, -0.0775, 0.006, -0.01, -0.0775, -0.0005, -0.01, -0.0775, 0.0016, -0.01, -0.0775, 0.0038, -0.01, -0.0775, 0.006, -0.0061, -0.0775, 0.0081, -0.0061, -0.0775, 0.0103, -0.0083, -0.0775, 0.0081, -0.0083, -0.0775, 0.0103, -0.0061, -0.0775, 0.0125, -0.0061, -0.0775, 0.0146, -0.0083, -0.0775, 0.0125, -0.0083, -0.0775, 0.0146, -0.01, -0.0775, 0.0081, -0.01, -0.0775, 0.0103, -0.01, -0.0775, 0.0125, -0.01, -0.0775, 0.0146, -0.01, -0.0751, -0.0005, -0.01, -0.0751, 0.0016, -0.01, -0.0729, -0.0005, -0.01, -0.0729, 0.0016, -0.01, -0.0751, 0.0038, -0.01, -0.0751, 0.006, -0.01, -0.0729, 0.0038, -0.01, -0.0729, 0.006, -0.01, -0.0707, -0.0005, -0.01, -0.0707, 0.0016,
-0.01, -0.0686, -0.0005, -0.01, -0.0686, 0.0016, -0.01, -0.0707, 0.0038, -0.01, -0.0707, 0.006, -0.01, -0.0686, 0.0038, -0.01, -0.0686, 0.006, -0.01, -0.0751, 0.0081, -0.01, -0.0751, 0.0103, -0.01, -0.0729, 0.0081, -0.01, -0.0729, 0.0103, -0.01, -0.0751, 0.0125, -0.01, -0.0751, 0.0146, -0.01, -0.0729, 0.0125, -0.01, -0.0729, 0.0146, -0.01, -0.0707, 0.0081, -0.01, -0.0707, 0.0103, -0.01, -0.0686, 0.0081, -0.01, -0.0686, 0.0103, -0.01, -0.0707, 0.0125, -0.01, -0.0707, 0.0146, -0.01, -0.0686, 0.0125, -0.01, -0.0686, 0.0146, -0.0061, -0.0775, 0.0168, -0.0061, -0.0775, 0.019, -0.0083, -0.0775, 0.0168, -0.0083, -0.0775, 0.019, -0.0061, -0.0775, 0.0212, -0.0061, -0.0775, 0.0233, -0.0083, -0.0775, 0.0212, -0.0083, -0.0775, 0.0233,
-0.01, -0.0775, 0.0168, -0.01, -0.0775, 0.019, -0.01, -0.0775, 0.0212, -0.01, -0.0775, 0.0233, -0.0061, -0.0775, 0.0255, -0.0061, -0.0775, 0.0277, -0.0083, -0.0775, 0.0255, -0.0083, -0.0775, 0.0277, -0.0061, -0.0775, 0.0298, -0.0061, -0.0775, 0.032, -0.0083, -0.0775, 0.0298, -0.0083, -0.0775, 0.032, -0.01, -0.0775, 0.0255, -0.01, -0.0775, 0.0277, -0.01, -0.0775, 0.0298, -0.01, -0.0775, 0.032, -0.01, -0.0751, 0.0168, -0.01, -0.0751, 0.019, -0.01, -0.0729, 0.0168, -0.01, -0.0729, 0.019, -0.01, -0.0751, 0.0212, -0.01, -0.0751, 0.0233, -0.01, -0.0729, 0.0212, -0.01, -0.0729, 0.0233, -0.01, -0.0707, 0.0168, -0.01, -0.0707, 0.019, -0.01, -0.0686, 0.0168, -0.01, -0.0686, 0.019, -0.01, -0.0707, 0.0212, -0.01, -0.0707, 0.0233,
-0.01, -0.0686, 0.0212, -0.01, -0.0686, 0.0233, -0.01, -0.0751, 0.0255, -0.01, -0.0751, 0.0277, -0.01, -0.0729, 0.0255, -0.01, -0.0729, 0.0277, -0.01, -0.0751, 0.0298, -0.01, -0.0751, 0.032, -0.01, -0.0729, 0.0298, -0.01, -0.0729, 0.032, -0.01, -0.0707, 0.0255, -0.01, -0.0707, 0.0277, -0.01, -0.0686, 0.0255, -0.01, -0.0686, 0.0277, -0.01, -0.0707, 0.0298, -0.01, -0.0707, 0.032, -0.01, -0.0686, 0.0298, -0.01, -0.0686, 0.032, -0.0061, -0.0775, 0.0342, -0.0061, -0.0775, 0.0363, -0.0083, -0.0775, 0.0342, -0.0083, -0.0775, 0.0363, -0.0061, -0.0775, 0.0385, -0.0061, -0.0775, 0.0407, -0.0083, -0.0775, 0.0385, -0.0083, -0.0775, 0.0407, -0.01, -0.0775, 0.0342, -0.01, -0.0775, 0.0363, -0.01, -0.0775, 0.0385, -0.01, -0.0775, 0.0407,
-0.0061, -0.0775, 0.0429, -0.0061, -0.0775, 0.045, -0.0083, -0.0775, 0.0429, -0.0083, -0.0775, 0.045, -0.0061, -0.0775, 0.0472, -0.0061, -0.0775, 0.0487, -0.0083, -0.0775, 0.0472, -0.0083, -0.0775, 0.0487, -0.01, -0.0775, 0.0429, -0.01, -0.0775, 0.045, -0.01, -0.0775, 0.0472, -0.01, -0.0775, 0.0487, -0.01, -0.0751, 0.0342, -0.01, -0.0751, 0.0363, -0.01, -0.0729, 0.0342, -0.01, -0.0729, 0.0363, -0.01, -0.0751, 0.0385, -0.01, -0.0751, 0.0407, -0.01, -0.0729, 0.0385, -0.01, -0.0729, 0.0407, -0.01, -0.0707, 0.0342, -0.01, -0.0707, 0.0363, -0.01, -0.0686, 0.0342, -0.01, -0.0686, 0.0363, -0.01, -0.0707, 0.0385, -0.01, -0.0707, 0.0407, -0.01, -0.0686, 0.0385, -0.01, -0.0686, 0.0407, -0.0061, -0.0751, 0.0487, -0.0083, -0.0751, 0.0487,
-0.0061, -0.0729, 0.0487, -0.0083, -0.0729, 0.0487, -0.01, -0.0751, 0.0429, -0.01, -0.0751, 0.045, -0.01, -0.0729, 0.0429, -0.01, -0.0729, 0.045, -0.01, -0.0751, 0.0472, -0.01, -0.0751, 0.0487, -0.01, -0.0729, 0.0472, -0.01, -0.0729, 0.0487, -0.0061, -0.0707, 0.0487, -0.0083, -0.0707, 0.0487, -0.0061, -0.0686, 0.0487, -0.0083, -0.0686, 0.0487, -0.01, -0.0707, 0.0429, -0.01, -0.0707, 0.045, -0.01, -0.0686, 0.0429, -0.01, -0.0686, 0.045, -0.01, -0.0707, 0.0472, -0.01, -0.0707, 0.0487, -0.01, -0.0686, 0.0472, -0.01, -0.0686, 0.0487, -0.0061, -0.0667, -0.0005, -0.0061, -0.0667, 0.0016, -0.0083, -0.0667, -0.0005, -0.0083, -0.0667, 0.0016, -0.0061, -0.0667, 0.0038, -0.0061, -0.0667, 0.006, -0.0083, -0.0667, 0.0038, -0.0083, -0.0667, 0.006,
-0.01, -0.0667, -0.0005, -0.01, -0.0667, 0.0016, -0.01, -0.0667, 0.0038, -0.01, -0.0667, 0.006, -0.0061, -0.0667, 0.0081, -0.0061, -0.0667, 0.0103, -0.0083, -0.0667, 0.0081, -0.0083, -0.0667, 0.0103, -0.0061, -0.0667, 0.0125, -0.0061, -0.0667, 0.0146, -0.0083, -0.0667, 0.0125, -0.0083, -0.0667, 0.0146, -0.01, -0.0667, 0.0081, -0.01, -0.0667, 0.0103, -0.01, -0.0667, 0.0125, -0.01, -0.0667, 0.0146, -0.0061, -0.0667, 0.0168, -0.0061, -0.0667, 0.019, -0.0083, -0.0667, 0.0168, -0.0083, -0.0667, 0.019, -0.0061, -0.0667, 0.0212, -0.0061, -0.0667, 0.0233, -0.0083, -0.0667, 0.0212, -0.0083, -0.0667, 0.0233, -0.01, -0.0667, 0.0168, -0.01, -0.0667, 0.019, -0.01, -0.0667, 0.0212, -0.01, -0.0667, 0.0233, -0.0061, -0.0667, 0.0255, -0.0061, -0.0667, 0.0277,
-0.0083, -0.0667, 0.0255, -0.0083, -0.0667, 0.0277, -0.0061, -0.0667, 0.0298, -0.0061, -0.0667, 0.032, -0.0083, -0.0667, 0.0298, -0.0083, -0.0667, 0.032, -0.01, -0.0667, 0.0255, -0.01, -0.0667, 0.0277, -0.01, -0.0667, 0.0298, -0.01, -0.0667, 0.032, -0.0061, -0.05, -0.0005, -0.0061, -0.05, 0.0016, -0.0083, -0.05, -0.0005, -0.0083, -0.05, 0.0016, -0.0061, -0.05, 0.0038, -0.0061, -0.05, 0.006, -0.0083, -0.05, 0.0038, -0.0083, -0.05, 0.006, -0.01, -0.05, -0.0005, -0.01, -0.05, 0.0016, -0.01, -0.0469, -0.0005, -0.01, -0.0469, 0.0016, -0.01, -0.05, 0.0038, -0.01, -0.05, 0.006, -0.01, -0.0469, 0.0038, -0.01, -0.0469, 0.006, -0.01, -0.0447, -0.0005, -0.01, -0.0447, 0.0016, -0.01, -0.0425, -0.0005, -0.01, -0.0425, 0.0016,
-0.01, -0.0447, 0.0038, -0.01, -0.0447, 0.006, -0.01, -0.0425, 0.0038, -0.01, -0.0425, 0.006, -0.0061, -0.05, 0.0081, -0.0061, -0.05, 0.0103, -0.0083, -0.05, 0.0081, -0.0083, -0.05, 0.0103, -0.0061, -0.05, 0.0125, -0.0061, -0.05, 0.0146, -0.0083, -0.05, 0.0125, -0.0083, -0.05, 0.0146, -0.01, -0.05, 0.0081, -0.01, -0.05, 0.0103, -0.01, -0.0469, 0.0081, -0.01, -0.0469, 0.0103, -0.01, -0.05, 0.0125, -0.01, -0.05, 0.0146, -0.01, -0.0469, 0.0125, -0.01, -0.0469, 0.0146, -0.01, -0.0447, 0.0081, -0.01, -0.0447, 0.0103, -0.01, -0.0425, 0.0081, -0.01, -0.0425, 0.0103, -0.01, -0.0447, 0.0125, -0.01, -0.0447, 0.0146, -0.01, -0.0425, 0.0125, -0.01, -0.0425, 0.0146, -0.0061, -0.0382, -0.0013, -0.0083, -0.0382, -0.0013,
-0.01, -0.0404, -0.0005, -0.01, -0.0404, 0.0016, -0.01, -0.0382, -0.0013, -0.01, -0.0382, 0.0016, -0.01, -0.0404, 0.0038, -0.01, -0.0404, 0.006, -0.01, -0.0382, 0.0038, -0.01, -0.0382, 0.006, -0.0061, -0.036, -0.0013, -0.0083, -0.036, -0.0013, -0.0061, -0.0338, -0.0013, -0.0083, -0.0338, -0.0013, -0.01, -0.036, -0.0013, -0.01, -0.036, 0.0016, -0.01, -0.0338, -0.0013, -0.01, -0.0338, 0.0016, -0.01, -0.036, 0.0038, -0.01, -0.036, 0.006, -0.01, -0.0338, 0.0038, -0.01, -0.0338, 0.006, -0.0061, -0.0382, 0.0105, -0.0083, -0.0382, 0.0105, -0.0061, -0.0382, 0.0125, -0.0061, -0.0382, 0.0146, -0.0083, -0.0382, 0.0125, -0.0083, -0.0382, 0.0146, -0.01, -0.0404, 0.0081, -0.01, -0.0404, 0.0103, -0.01, -0.0382, 0.0081, -0.01, -0.0382, 0.0105,
-0.01, -0.0404, 0.0125, -0.01, -0.0404, 0.0146, -0.01, -0.0382, 0.0125, -0.01, -0.0382, 0.0146, -0.0061, -0.036, 0.0105, -0.0083, -0.036, 0.0105, -0.0061, -0.0338, 0.0105, -0.0083, -0.0338, 0.0105, -0.01, -0.036, 0.0081, -0.01, -0.036, 0.0105, -0.01, -0.0338, 0.0081, -0.01, -0.0338, 0.0105, -0.0061, -0.05, 0.0168, -0.0061, -0.05, 0.019, -0.0083, -0.05, 0.0168, -0.0083, -0.05, 0.019, -0.0061, -0.05, 0.0212, -0.0061, -0.05, 0.0233, -0.0083, -0.05, 0.0212, -0.0083, -0.05, 0.0233, -0.01, -0.05, 0.0168, -0.01, -0.05, 0.019, -0.01, -0.0469, 0.0168, -0.01, -0.0469, 0.019, -0.01, -0.05, 0.0212, -0.01, -0.05, 0.0233, -0.01, -0.0469, 0.0212, -0.01, -0.0469, 0.0233, -0.01, -0.0447, 0.0168, -0.01, -0.0447, 0.019,
-0.01, -0.0425, 0.0168, -0.01, -0.0425, 0.019, -0.01, -0.0447, 0.0212, -0.01, -0.0447, 0.0233, -0.01, -0.0425, 0.0212, -0.01, -0.0425, 0.0233, -0.0061, -0.05, 0.0255, -0.0061, -0.05, 0.0277, -0.0083, -0.05, 0.0255, -0.0083, -0.05, 0.0277, -0.0061, -0.05, 0.0298, -0.0061, -0.05, 0.032, -0.0083, -0.05, 0.0298, -0.0083, -0.05, 0.032, -0.01, -0.05, 0.0255, -0.01, -0.05, 0.0277, -0.01, -0.0469, 0.0255, -0.01, -0.0469, 0.0277, -0.01, -0.05, 0.0298, -0.01, -0.05, 0.032, -0.01, -0.0469, 0.0298, -0.01, -0.0469, 0.032, -0.01, -0.0447, 0.0255, -0.01, -0.0447, 0.0277, -0.01, -0.0425, 0.0255, -0.01, -0.0425, 0.0277, -0.01, -0.0447, 0.0298, -0.01, -0.0447, 0.032, -0.01, -0.0425, 0.0298, -0.01, -0.0425, 0.032,
-0.0061, -0.0382, 0.0168, -0.0061, -0.0382, 0.019, -0.0083, -0.0382, 0.0168, -0.0083, -0.0382, 0.019, -0.0061, -0.0382, 0.0212, -0.0061, -0.0382, 0.0233, -0.0083, -0.0382, 0.0212, -0.0083, -0.0382, 0.0233, -0.01, -0.0404, 0.0168, -0.01, -0.0404, 0.019, -0.01, -0.0382, 0.0168, -0.01, -0.0382, 0.019, -0.01, -0.0404, 0.0212, -0.01, -0.0404, 0.0233, -0.01, -0.0382, 0.0212, -0.01, -0.0382, 0.0233, -0.0061, -0.0382, 0.0255, -0.0061, -0.0382, 0.0277, -0.0083, -0.0382, 0.0255, -0.0083, -0.0382, 0.0277, -0.0061, -0.0382, 0.0298, -0.0061, -0.0382, 0.032, -0.0083, -0.0382, 0.0298, -0.0083, -0.0382, 0.032, -0.01, -0.0404, 0.0255, -0.01, -0.0404, 0.0277, -0.01, -0.0382, 0.0255, -0.01, -0.0382, 0.0277, -0.01, -0.0404, 0.0298, -0.01, -0.0404, 0.032,
-0.01, -0.0382, 0.0298, -0.01, -0.0382, 0.032, -0.0061, -0.0667, 0.0342, -0.0061, -0.0667, 0.0363, -0.0083, -0.0667, 0.0342, -0.0083, -0.0667, 0.0363, -0.0061, -0.0667, 0.0385, -0.0061, -0.0667, 0.0407, -0.0083, -0.0667, 0.0385, -0.0083, -0.0667, 0.0407, -0.01, -0.0667, 0.0342, -0.01, -0.0667, 0.0363, -0.01, -0.0667, 0.0385, -0.01, -0.0667, 0.0407, -0.0061, -0.0667, 0.0429, -0.0061, -0.0667, 0.045, -0.0083, -0.0667, 0.0429, -0.0083, -0.0667, 0.045, -0.0061, -0.0667, 0.0472, -0.0061, -0.0667, 0.0487, -0.0083, -0.0667, 0.0472, -0.0083, -0.0667, 0.0487, -0.01, -0.0667, 0.0429, -0.01, -0.0667, 0.045, -0.01, -0.0667, 0.0472, -0.01, -0.0667, 0.0487, -0.0061, -0.05, 0.0342, -0.0061, -0.05, 0.0363, -0.0083, -0.05, 0.0342, -0.0083, -0.05, 0.0363,
-0.0061, -0.05, 0.0385, -0.0061, -0.05, 0.0407, -0.0083, -0.05, 0.0385, -0.0083, -0.05, 0.0407, -0.01, -0.05, 0.0342, -0.01, -0.05, 0.0363, -0.01, -0.0469, 0.0342, -0.01, -0.0469, 0.0363, -0.01, -0.05, 0.0385, -0.01, -0.05, 0.0407, -0.01, -0.0469, 0.0385, -0.01, -0.0469, 0.0407, -0.01, -0.0447, 0.0342, -0.01, -0.0447, 0.0363, -0.01, -0.0425, 0.0342, -0.01, -0.0425, 0.0363, -0.01, -0.0447, 0.0385, -0.01, -0.0447, 0.0407, -0.01, -0.0425, 0.0385, -0.01, -0.0425, 0.0407, -0.0061, -0.05, 0.0429, -0.0061, -0.05, 0.045, -0.0083, -0.05, 0.0429, -0.0083, -0.05, 0.045, -0.0061, -0.05, 0.0472, -0.0061, -0.05, 0.0487, -0.0083, -0.05, 0.0472, -0.0083, -0.05, 0.0487, -0.0061, -0.0469, 0.0487, -0.0083, -0.0469, 0.0487,
-0.01, -0.05, 0.0429, -0.01, -0.05, 0.045, -0.01, -0.0469, 0.0429, -0.01, -0.0469, 0.045, -0.01, -0.05, 0.0472, -0.01, -0.05, 0.0487, -0.01, -0.0469, 0.0472, -0.01, -0.0469, 0.0487, -0.0061, -0.0447, 0.0487, -0.0083, -0.0447, 0.0487, -0.0061, -0.0425, 0.0487, -0.0083, -0.0425, 0.0487, -0.01, -0.0447, 0.0429, -0.01, -0.0447, 0.045, -0.01, -0.0425, 0.0429, -0.01, -0.0425, 0.045, -0.01, -0.0447, 0.0472, -0.01, -0.0447, 0.0487, -0.01, -0.0425, 0.0472, -0.01, -0.0425, 0.0487, -0.0061, -0.0382, 0.0342, -0.0061, -0.0382, 0.0363, -0.0083, -0.0382, 0.0342, -0.0083, -0.0382, 0.0363, -0.0061, -0.0382, 0.0379, -0.0083, -0.0382, 0.0379, -0.01, -0.0404, 0.0342, -0.01, -0.0404, 0.0363, -0.01, -0.0382, 0.0342, -0.01, -0.0382, 0.0363,
-0.01, -0.0404, 0.0385, -0.01, -0.0404, 0.0407, -0.01, -0.0382, 0.0379, -0.01, -0.0382, 0.0407, -0.0061, -0.036, 0.0379, -0.0083, -0.036, 0.0379, -0.0061, -0.0338, 0.0379, -0.0083, -0.0338, 0.0379, -0.01, -0.036, 0.0379, -0.01, -0.036, 0.0407, -0.01, -0.0338, 0.0379, -0.01, -0.0338, 0.0407, -0.0061, -0.0404, 0.0487, -0.0083, -0.0404, 0.0487, -0.0061, -0.0382, 0.0487, -0.0083, -0.0382, 0.0487, -0.01, -0.0404, 0.0429, -0.01, -0.0404, 0.045, -0.01, -0.0382, 0.0429, -0.01, -0.0382, 0.045, -0.01, -0.0404, 0.0472, -0.01, -0.0404, 0.0487, -0.01, -0.0382, 0.0472, -0.01, -0.0382, 0.0487, -0.0061, -0.036, 0.0487, -0.0083, -0.036, 0.0487, -0.0061, -0.0338, 0.0487, -0.0083, -0.0338, 0.0487, -0.01, -0.036, 0.0429, -0.01, -0.036, 0.045,
-0.01, -0.0338, 0.0429, -0.01, -0.0338, 0.045, -0.01, -0.036, 0.0472, -0.01, -0.036, 0.0487, -0.01, -0.0338, 0.0472, -0.01, -0.0338, 0.0487, -0.0061, -0.0317, -0.0013, -0.0083, -0.0317, -0.0013, -0.0061, -0.0295, -0.0013, -0.0083, -0.0295, -0.0013, -0.01, -0.0317, -0.0013, -0.01, -0.0317, 0.0016, -0.01, -0.0295, -0.0013, -0.01, -0.0295, 0.0016, -0.01, -0.0317, 0.0038, -0.01, -0.0317, 0.006, -0.01, -0.0295, 0.0038, -0.01, -0.0295, 0.006, -0.0061, -0.0273, -0.0013, -0.0083, -0.0273, -0.0013, -0.0061, -0.0252, -0.0013, -0.0083, -0.0252, -0.0013, -0.01, -0.0273, -0.0013, -0.01, -0.0273, 0.0016, -0.01, -0.0252, -0.0013, -0.01, -0.0252, 0.0016, -0.01, -0.0273, 0.0038, -0.01, -0.0273, 0.006, -0.01, -0.0252, 0.0038, -0.01, -0.0252, 0.006,
-0.0061, -0.0317, 0.0105, -0.0083, -0.0317, 0.0105, -0.0061, -0.0295, 0.0105, -0.0083, -0.0295, 0.0105, -0.01, -0.0317, 0.0081, -0.01, -0.0317, 0.0105, -0.01, -0.0295, 0.0081, -0.01, -0.0295, 0.0105, -0.0061, -0.0273, 0.0105, -0.0083, -0.0273, 0.0105, -0.0061, -0.0252, 0.0105, -0.0083, -0.0252, 0.0105, -0.01, -0.0273, 0.0081, -0.01, -0.0273, 0.0105, -0.01, -0.0252, 0.0081, -0.01, -0.0252, 0.0105, -0.0061, -0.023, -0.0013, -0.0083, -0.023, -0.0013, -0.0061, -0.0208, -0.0013, -0.0083, -0.0208, -0.0013, -0.01, -0.023, -0.0013, -0.01, -0.023, 0.0016, -0.01, -0.0208, -0.0013, -0.01, -0.0208, 0.0016, -0.01, -0.023, 0.0038, -0.01, -0.023, 0.006, -0.01, -0.0208, 0.0038, -0.01, -0.0208, 0.006, -0.0061, -0.0187, -0.0013, -0.0083, -0.0187, -0.0013,
-0.0061, -0.0165, -0.0013, -0.0083, -0.0165, -0.0013, -0.01, -0.0187, -0.0013, -0.01, -0.0187, 0.0016, -0.01, -0.0165, -0.0013, -0.01, -0.0165, 0.0016, -0.01, -0.0187, 0.0038, -0.01, -0.0187, 0.006, -0.01, -0.0165, 0.0038, -0.01, -0.0165, 0.006, -0.0061, -0.023, 0.0105, -0.0083, -0.023, 0.0105, -0.0061, -0.0208, 0.0105, -0.0083, -0.0208, 0.0105, -0.01, -0.023, 0.0081, -0.01, -0.023, 0.0105, -0.01, -0.0208, 0.0081, -0.01, -0.0208, 0.0105, -0.0061, -0.0187, 0.0105, -0.0083, -0.0187, 0.0105, -0.0061, -0.0165, 0.0105, -0.0083, -0.0165, 0.0105, -0.01, -0.0187, 0.0081, -0.01, -0.0187, 0.0105, -0.01, -0.0165, 0.0081, -0.01, -0.0165, 0.0105, -0.0061, -0.0143, -0.0013, -0.0083, -0.0143, -0.0013, -0.0061, -0.0121, -0.0013, -0.0083, -0.0121, -0.0013,
-0.01, -0.0143, -0.0013, -0.01, -0.0143, 0.0016, -0.01, -0.0121, -0.0013, -0.01, -0.0121, 0.0016, -0.01, -0.0143, 0.0038, -0.01, -0.0143, 0.006, -0.01, -0.0121, 0.0038, -0.01, -0.0121, 0.006, -0.0061, -0.01, -0.0013, -0.0083, -0.01, -0.0013, -0.0061, -0.0078, -0.0013, -0.0083, -0.0078, -0.0013, -0.01, -0.01, -0.0013, -0.01, -0.01, 0.0016, -0.01, -0.0078, -0.0013, -0.01, -0.0078, 0.0016, -0.01, -0.01, 0.0038, -0.01, -0.01, 0.006, -0.01, -0.0078, 0.0038, -0.01, -0.0078, 0.006, -0.0061, -0.0143, 0.0105, -0.0083, -0.0143, 0.0105, -0.0061, -0.0121, 0.0105, -0.0083, -0.0121, 0.0105, -0.01, -0.0143, 0.0081, -0.01, -0.0143, 0.0105, -0.01, -0.0121, 0.0081, -0.01, -0.0121, 0.0105, -0.0061, -0.01, 0.0105, -0.0083, -0.01, 0.0105,
-0.0061, -0.0078, 0.0105, -0.0083, -0.0078, 0.0105, -0.01, -0.01, 0.0081, -0.01, -0.01, 0.0105, -0.01, -0.0078, 0.0081, -0.01, -0.0078, 0.0105, -0.0061, -0.0056, -0.0013, -0.0083, -0.0056, -0.0013, -0.0061, -0.0035, -0.0013, -0.0083, -0.0035, -0.0013, -0.01, -0.0056, -0.0013, -0.01, -0.0056, 0.0016, -0.01, -0.0035, -0.0013, -0.01, -0.0035, 0.0016, -0.01, -0.0056, 0.0038, -0.01, -0.0056, 0.006, -0.01, -0.0035, 0.0038, -0.01, -0.0035, 0.006, -0.0061, -0.0013, -0.0013, -0.0083, -0.0013, -0.0013, -0.0061, 0.0009, -0.0013, -0.0083, 0.0009, -0.0013, -0.01, -0.0013, -0.0013, -0.01, -0.0013, 0.0016, -0.01, 0.0009, -0.0013, -0.01, 0.0009, 0.0016, -0.01, -0.0013, 0.0038, -0.01, -0.0013, 0.006, -0.01, 0.0009, 0.0038, -0.01, 0.0009, 0.006,
-0.0061, -0.0056, 0.0105, -0.0083, -0.0056, 0.0105, -0.0061, -0.0035, 0.0105, -0.0083, -0.0035, 0.0105, -0.01, -0.0056, 0.0081, -0.01, -0.0056, 0.0105, -0.01, -0.0035, 0.0081, -0.01, -0.0035, 0.0105, -0.0061, -0.0013, 0.0105, -0.0083, -0.0013, 0.0105, -0.0061, 0.0009, 0.0105, -0.0083, 0.0009, 0.0105, -0.01, -0.0013, 0.0081, -0.01, -0.0013, 0.0105, -0.01, 0.0009, 0.0081, -0.01, 0.0009, 0.0105, -0.0061, -0.0317, 0.0379, -0.0083, -0.0317, 0.0379, -0.0061, -0.0295, 0.0379, -0.0083, -0.0295, 0.0379, -0.01, -0.0317, 0.0379, -0.01, -0.0317, 0.0407, -0.01, -0.0295, 0.0379, -0.01, -0.0295, 0.0407, -0.0061, -0.0273, 0.0379, -0.0083, -0.0273, 0.0379, -0.0061, -0.0252, 0.0379, -0.0083, -0.0252, 0.0379, -0.01, -0.0273, 0.0379, -0.01, -0.0273, 0.0407,
-0.01, -0.0252, 0.0379, -0.01, -0.0252, 0.0407, -0.0061, -0.0317, 0.0487, -0.0083, -0.0317, 0.0487, -0.0061, -0.0295, 0.0487, -0.0083, -0.0295, 0.0487, -0.01, -0.0317, 0.0429, -0.01, -0.0317, 0.045, -0.01, -0.0295, 0.0429, -0.01, -0.0295, 0.045, -0.01, -0.0317, 0.0472, -0.01, -0.0317, 0.0487, -0.01, -0.0295, 0.0472, -0.01, -0.0295, 0.0487, -0.0061, -0.0273, 0.0487, -0.0083, -0.0273, 0.0487, -0.0061, -0.0252, 0.0487, -0.0083, -0.0252, 0.0487, -0.01, -0.0273, 0.0429, -0.01, -0.0273, 0.045, -0.01, -0.0252, 0.0429, -0.01, -0.0252, 0.045, -0.01, -0.0273, 0.0472, -0.01, -0.0273, 0.0487, -0.01, -0.0252, 0.0472, -0.01, -0.0252, 0.0487, -0.0061, -0.023, 0.0379, -0.0083, -0.023, 0.0379, -0.0061, -0.0208, 0.0379, -0.0083, -0.0208, 0.0379,
-0.01, -0.023, 0.0379, -0.01, -0.023, 0.0407, -0.01, -0.0208, 0.0379, -0.01, -0.0208, 0.0407, -0.0061, -0.0187, 0.0379, -0.0083, -0.0187, 0.0379, -0.0061, -0.0165, 0.0379, -0.0083, -0.0165, 0.0379, -0.01, -0.0187, 0.0379, -0.01, -0.0187, 0.0407, -0.01, -0.0165, 0.0379, -0.01, -0.0165, 0.0407, -0.0061, -0.023, 0.0487, -0.0083, -0.023, 0.0487, -0.0061, -0.0208, 0.0487, -0.0083, -0.0208, 0.0487, -0.01, -0.023, 0.0429, -0.01, -0.023, 0.045, -0.01, -0.0208, 0.0429, -0.01, -0.0208, 0.045, -0.01, -0.023, 0.0472, -0.01, -0.023, 0.0487, -0.01, -0.0208, 0.0472, -0.01, -0.0208, 0.0487, -0.0061, -0.0187, 0.0487, -0.0083, -0.0187, 0.0487, -0.0061, -0.0165, 0.0487, -0.0083, -0.0165, 0.0487, -0.01, -0.0187, 0.0429, -0.01, -0.0187, 0.045,
-0.01, -0.0165, 0.0429, -0.01, -0.0165, 0.045, -0.01, -0.0187, 0.0472, -0.01, -0.0187, 0.0487, -0.01, -0.0165, 0.0472, -0.01, -0.0165, 0.0487, -0.0061, -0.0143, 0.0379, -0.0083, -0.0143, 0.0379, -0.0061, -0.0121, 0.0379, -0.0083, -0.0121, 0.0379, -0.01, -0.0143, 0.0379, -0.01, -0.0143, 0.0407, -0.01, -0.0121, 0.0379, -0.01, -0.0121, 0.0407, -0.0061, -0.01, 0.0379, -0.0083, -0.01, 0.0379, -0.0061, -0.0078, 0.0379, -0.0083, -0.0078, 0.0379, -0.01, -0.01, 0.0379, -0.01, -0.01, 0.0407, -0.01, -0.0078, 0.0379, -0.01, -0.0078, 0.0407, -0.0061, -0.0143, 0.0487, -0.0083, -0.0143, 0.0487, -0.0061, -0.0121, 0.0487, -0.0083, -0.0121, 0.0487, -0.01, -0.0143, 0.0429, -0.01, -0.0143, 0.045, -0.01, -0.0121, 0.0429, -0.01, -0.0121, 0.045,
-0.01, -0.0143, 0.0472, -0.01, -0.0143, 0.0487, -0.01, -0.0121, 0.0472, -0.01, -0.0121, 0.0487, -0.0061, -0.01, 0.0487, -0.0083, -0.01, 0.0487, -0.0061, -0.0078, 0.0487, -0.0083, -0.0078, 0.0487, -0.01, -0.01, 0.0429, -0.01, -0.01, 0.045, -0.01, -0.0078, 0.0429, -0.01, -0.0078, 0.045, -0.01, -0.01, 0.0472, -0.01, -0.01, 0.0487, -0.01, -0.0078, 0.0472, -0.01, -0.0078, 0.0487, -0.0061, -0.0056, 0.0379, -0.0083, -0.0056, 0.0379, -0.0061, -0.0035, 0.0379, -0.0083, -0.0035, 0.0379, -0.01, -0.0056, 0.0379, -0.01, -0.0056, 0.0407, -0.01, -0.0035, 0.0379, -0.01, -0.0035, 0.0407, -0.0061, -0.0013, 0.0379, -0.0083, -0.0013, 0.0379, -0.0061, 0.0009, 0.0379, -0.0083, 0.0009, 0.0379, -0.01, -0.0013, 0.0379, -0.01, -0.0013, 0.0407,
-0.01, 0.0008, 0.0379, -0.01, 0.0009, 0.0407, -0.0061, -0.0056, 0.0487, -0.0083, -0.0056, 0.0487, -0.0061, -0.0035, 0.0487, -0.0083, -0.0035, 0.0487, -0.01, -0.0056, 0.0429, -0.01, -0.0056, 0.045, -0.01, -0.0035, 0.0429, -0.01, -0.0035, 0.045, -0.01, -0.0056, 0.0472, -0.01, -0.0056, 0.0487, -0.01, -0.0035, 0.0472, -0.01, -0.0035, 0.0487, -0.0061, -0.0013, 0.0487, -0.0083, -0.0013, 0.0487, -0.0061, 0.0009, 0.0487, -0.0083, 0.0009, 0.0487, -0.01, -0.0013, 0.0429, -0.01, -0.0013, 0.045, -0.01, 0.0009, 0.0429, -0.01, 0.0009, 0.045, -0.01, -0.0013, 0.0472, -0.01, -0.0013, 0.0487, -0.01, 0.0009, 0.0472, -0.01, 0.0009, 0.0487, 0.0, 0.0461, -0.067, 0.0, 0.0486, -0.067, 0.0, 0.0461, -0.0656, 0.0, 0.0461, -0.0635,
0.0, 0.0486, -0.0656, 0.0, 0.0486, -0.0635, -0.0017, 0.0461, -0.067, -0.0039, 0.0461, -0.067, -0.0017, 0.0486, -0.067, -0.0039, 0.0486, -0.067, -0.0017, 0.0461, -0.0656, -0.0017, 0.0461, -0.0635, -0.0039, 0.0461, -0.0656, -0.0039, 0.0461, -0.0635, 0.0, 0.05, -0.067, 0.0, 0.05, -0.0656, 0.0, 0.05, -0.0635, -0.0017, 0.05, -0.067, -0.0039, 0.05, -0.067, -0.0017, 0.05, -0.0656, -0.0017, 0.05, -0.0635, -0.0039, 0.05, -0.0656, -0.0039, 0.05, -0.0635, 0.0, 0.0461, -0.0613, 0.0, 0.0461, -0.0591, 0.0, 0.0486, -0.0613, 0.0, 0.0486, -0.0591, 0.0, 0.0461, -0.057, 0.0, 0.0461, -0.0548, 0.0, 0.0486, -0.057, 0.0, 0.0486, -0.0548, -0.0017, 0.0461, -0.0613, -0.0017, 0.0461, -0.0591, -0.0039, 0.0461, -0.0613,
-0.0039, 0.0461, -0.0591, -0.0017, 0.0461, -0.057, -0.0017, 0.0461, -0.0548, -0.0039, 0.0461, -0.057, -0.0039, 0.0461, -0.0548, 0.0, 0.05, -0.0613, 0.0, 0.05, -0.0591, 0.0, 0.05, -0.057, 0.0, 0.05, -0.0548, 0.0, 0.0529, -0.0562, 0.0, 0.0529, -0.0548, -0.0017, 0.05, -0.0613, -0.0017, 0.05, -0.0591, -0.0039, 0.05, -0.0613, -0.0039, 0.05, -0.0591, -0.0017, 0.05, -0.057, -0.0017, 0.05, -0.0548, -0.0039, 0.05, -0.057, -0.0039, 0.05, -0.0548, -0.0017, 0.0529, -0.0562, -0.0017, 0.0529, -0.0548, -0.0039, 0.0529, -0.0562, -0.0039, 0.0529, -0.0548, 0.0, 0.0461, -0.0526, 0.0, 0.0461, -0.0505, 0.0, 0.0486, -0.0526, 0.0, 0.0486, -0.0505, 0.0, 0.0461, -0.0483, 0.0, 0.0461, -0.0461, 0.0, 0.0486, -0.0483,
0.0, 0.0486, -0.0461, -0.0017, 0.0461, -0.0526, -0.0017, 0.0461, -0.0505, -0.0039, 0.0461, -0.0526, -0.0039, 0.0461, -0.0505, -0.0017, 0.0461, -0.0483, -0.0017, 0.0461, -0.0461, -0.0039, 0.0461, -0.0483, -0.0039, 0.0461, -0.0461, 0.0, 0.05, -0.0526, 0.0, 0.05, -0.0505, 0.0, 0.0529, -0.0523, 0.0, 0.05, -0.0483, 0.0, 0.05, -0.0454, 0.0, 0.053, -0.0454, -0.0017, 0.05, -0.0526, -0.0017, 0.05, -0.0505, -0.0039, 0.05, -0.0526, -0.0039, 0.05, -0.0505, -0.0017, 0.0529, -0.0523, -0.0039, 0.0529, -0.0523, -0.0017, 0.05, -0.0483, -0.0017, 0.05, -0.0454, -0.0039, 0.05, -0.0483, -0.0039, 0.05, -0.0454, -0.0017, 0.053, -0.0454, -0.0039, 0.053, -0.0454, 0.0, 0.0461, -0.0439, 0.0, 0.0461, -0.0425, 0.0, 0.0486, -0.0439,
0.0, 0.0486, -0.0425, -0.0017, 0.0461, -0.0439, -0.0017, 0.0461, -0.0425, -0.0039, 0.0461, -0.0439, -0.0039, 0.0461, -0.0425, -0.0017, 0.0486, -0.0425, -0.0039, 0.0486, -0.0425, 0.0, 0.0508, -0.0439, 0.0, 0.0508, -0.0425, 0.0, 0.053, -0.0439, 0.0, 0.053, -0.0425, -0.0017, 0.0508, -0.0425, -0.0039, 0.0508, -0.0425, -0.0017, 0.053, -0.0425, -0.0039, 0.053, -0.0425, 0.0, 0.0551, -0.0562, 0.0, 0.0551, -0.0548, 0.0, 0.0573, -0.0562, 0.0, 0.0573, -0.0548, -0.0017, 0.0551, -0.0562, -0.0039, 0.0551, -0.0562, -0.0017, 0.0573, -0.0562, -0.0039, 0.0573, -0.0562, 0.0, 0.0595, -0.0562, 0.0, 0.0595, -0.0548, 0.0, 0.0616, -0.0562, 0.0, 0.0616, -0.0548, -0.0017, 0.0595, -0.0562, -0.0039, 0.0595, -0.0562, -0.0017, 0.0616, -0.0562,
-0.0039, 0.0616, -0.0562, 0.0, 0.0638, -0.0562, 0.0, 0.0638, -0.0548, 0.0, 0.0657, -0.0562, 0.0, 0.0657, -0.0548, -0.0017, 0.0638, -0.0562, -0.0039, 0.0638, -0.0562, -0.0017, 0.0657, -0.0562, -0.0017, 0.0657, -0.0548, -0.0039, 0.0657, -0.0562, -0.0039, 0.0657, -0.0548, 0.0, 0.0551, -0.0523, 0.0, 0.0573, -0.0523, 0.0, 0.0551, -0.0454, 0.0, 0.0573, -0.0454, -0.0017, 0.0551, -0.0523, -0.0039, 0.0551, -0.0523, -0.0017, 0.0573, -0.0523, -0.0039, 0.0573, -0.0523, -0.0017, 0.0551, -0.0454, -0.0039, 0.0551, -0.0454, -0.0017, 0.0573, -0.0454, -0.0039, 0.0573, -0.0454, 0.0, 0.0595, -0.0523, 0.0, 0.0616, -0.0523, 0.0, 0.0595, -0.0454, 0.0, 0.0616, -0.0454, -0.0017, 0.0595, -0.0523, -0.0039, 0.0595, -0.0523, -0.0017, 0.0616, -0.0523,
-0.0039, 0.0616, -0.0523, -0.0017, 0.0595, -0.0454, -0.0039, 0.0595, -0.0454, -0.0017, 0.0616, -0.0454, -0.0039, 0.0616, -0.0454, 0.0, 0.0551, -0.0439, 0.0, 0.0551, -0.0425, 0.0, 0.0573, -0.0439, 0.0, 0.0573, -0.0425, -0.0017, 0.0551, -0.0425, -0.0039, 0.0551, -0.0425, -0.0017, 0.0573, -0.0425, -0.0039, 0.0573, -0.0425, 0.0, 0.0595, -0.0439, 0.0, 0.0595, -0.0425, 0.0, 0.0616, -0.0439, 0.0, 0.0616, -0.0425, -0.0017, 0.0595, -0.0425, -0.0039, 0.0595, -0.0425, -0.0017, 0.0616, -0.0425, -0.0039, 0.0616, -0.0425, 0.0, 0.0638, -0.0523, 0.0, 0.0657, -0.0523, 0.0, 0.0638, -0.0454, 0.0, 0.0667, -0.0454, -0.0017, 0.0638, -0.0523, -0.0039, 0.0638, -0.0523, -0.0017, 0.0657, -0.0523, -0.0039, 0.0657, -0.0523, -0.0017, 0.0638, -0.0454,
-0.0039, 0.0638, -0.0454, -0.0017, 0.0667, -0.0454, -0.0039, 0.0667, -0.0454, 0.0, 0.0638, -0.0439, 0.0, 0.0638, -0.0425, 0.0, 0.0667, -0.0439, 0.0, 0.0667, -0.0425, -0.0017, 0.0638, -0.0425, -0.0039, 0.0638, -0.0425, -0.0017, 0.0667, -0.0439, -0.0017, 0.0667, -0.0425, -0.0039, 0.0667, -0.0439, -0.0039, 0.0667, -0.0425, 0.0, 0.0887, -0.0624, -0.0017, 0.0887, -0.0625, -0.0039, 0.0887, -0.0625, 0.0, 0.088, -0.0616, 0.0, 0.087, -0.0594, 0.0, 0.0864, -0.0565, 0.0, 0.0863, -0.0548, 0.0, 0.0865, -0.0573, 0.0, 0.0877, -0.0548, -0.0017, 0.0879, -0.0614, -0.0017, 0.0869, -0.0592, -0.0039, 0.0879, -0.0614, -0.0039, 0.0869, -0.0592, -0.0017, 0.0864, -0.0568, -0.0017, 0.0863, -0.0548, -0.0039, 0.0864, -0.0568, -0.0039, 0.0863, -0.0548,
-0.0017, 0.0866, -0.0579, -0.0039, 0.0866, -0.0579, 0.0, 0.0863, -0.0526, 0.0, 0.0865, -0.0513, 0.0, 0.0877, -0.0526, 0.0, 0.0867, -0.0503, 0.0, 0.0875, -0.048, 0.0, 0.0884, -0.0468, -0.0017, 0.0863, -0.0526, -0.0017, 0.0865, -0.0511, -0.0039, 0.0863, -0.0526, -0.0039, 0.0865, -0.0511, -0.0017, 0.0867, -0.0501, -0.0039, 0.0867, -0.0501, -0.0017, 0.0874, -0.0482, -0.0017, 0.0884, -0.0468, -0.0039, 0.0874, -0.0482, -0.0039, 0.0884, -0.0468, 0.0, 0.0909, -0.0646, 0.0, 0.0897, -0.0637, 0.0, 0.0918, -0.0652, 0.0, 0.092, -0.0635, -0.0017, 0.0909, -0.0646, -0.0017, 0.0897, -0.0637, -0.0039, 0.0909, -0.0646, -0.0039, 0.0897, -0.0637, -0.0017, 0.092, -0.0653, -0.0039, 0.092, -0.0653, 0.0, 0.0972, -0.0668, 0.0, 0.0939, -0.0661,
0.0, 0.0942, -0.0635, 0.0, 0.0962, -0.0667, 0.0, 0.0963, -0.0627, -0.0017, 0.097, -0.0668, -0.0039, 0.097, -0.0668, -0.0017, 0.0942, -0.0662, -0.0039, 0.0942, -0.0662, -0.0017, 0.0959, -0.0666, -0.0017, 0.0964, -0.0627, -0.0039, 0.0959, -0.0666, -0.0039, 0.0964, -0.0627, 0.0, 0.0899, -0.0613, 0.0, 0.091, -0.0586, 0.0, 0.0926, -0.0608, 0.0, 0.0915, -0.0595, 0.0, 0.0904, -0.0568, 0.0, 0.0902, -0.0548, -0.0017, 0.0908, -0.0583, -0.0039, 0.0908, -0.0583, -0.0017, 0.0926, -0.0608, -0.0017, 0.0914, -0.0594, -0.0039, 0.0926, -0.0608, -0.0039, 0.0914, -0.0594, -0.0017, 0.0904, -0.057, -0.0017, 0.0903, -0.0548, -0.0039, 0.0904, -0.057, -0.0039, 0.0903, -0.0548, 0.0, 0.0942, -0.0621, 0.0, 0.0953, -0.0624, -0.0017, 0.0942, -0.0621,
-0.0039, 0.0942, -0.0621, -0.0017, 0.0953, -0.0624, -0.0039, 0.0953, -0.0624, 0.0, 0.0985, -0.0669, 0.0, 0.1007, -0.067, 0.0, 0.0985, -0.0656, 0.0, 0.0986, -0.063, 0.0, 0.1007, -0.0656, 0.0, 0.1007, -0.0631, -0.0017, 0.0985, -0.0669, -0.0039, 0.0985, -0.0669, -0.0017, 0.1007, -0.067, -0.0039, 0.1007, -0.067, -0.0017, 0.0985, -0.063, -0.0039, 0.0985, -0.063, -0.0017, 0.1007, -0.0631, -0.0039, 0.1007, -0.0631, 0.0, 0.1029, -0.067, 0.0, 0.105, -0.0669, 0.0, 0.1029, -0.0656, 0.0, 0.1029, -0.0631, 0.0, 0.105, -0.0656, 0.0, 0.105, -0.063, -0.0017, 0.1029, -0.067, -0.0039, 0.1029, -0.067, -0.0017, 0.105, -0.067, -0.0039, 0.105, -0.067, -0.0017, 0.1029, -0.0631, -0.0039, 0.1029, -0.0631, -0.0017, 0.105, -0.063,
-0.0039, 0.105, -0.063, 0.0, 0.0904, -0.0528, 0.0, 0.0909, -0.0508, 0.0, 0.0913, -0.0499, 0.0, 0.0899, -0.0483, 0.0, 0.0894, -0.0456, 0.0, 0.0922, -0.0485, 0.0, 0.092, -0.0461, -0.0017, 0.0904, -0.0526, -0.0017, 0.0908, -0.0512, -0.0039, 0.0904, -0.0526, -0.0039, 0.0908, -0.0512, -0.0017, 0.0912, -0.05, -0.0039, 0.0912, -0.05, -0.0017, 0.0894, -0.0456, -0.0039, 0.0894, -0.0456, -0.0017, 0.0923, -0.0485, -0.0039, 0.0923, -0.0485, 0.0, 0.0934, -0.0475, 0.0, 0.0945, -0.0469, 0.0, 0.0965, -0.046, -0.0017, 0.0935, -0.0474, -0.0017, 0.0946, -0.0468, -0.0039, 0.0935, -0.0474, -0.0039, 0.0946, -0.0468, -0.0017, 0.0963, -0.0461, -0.0039, 0.0963, -0.0461, 0.0, 0.0905, -0.0447, 0.0, 0.0917, -0.0438, -0.0017, 0.0905, -0.0447,
-0.0039, 0.0905, -0.0447, -0.0017, 0.092, -0.0437, -0.0039, 0.092, -0.0437, 0.0, 0.0939, -0.0429, 0.0, 0.0949, -0.0427, 0.0, 0.0964, -0.0439, 0.0, 0.0963, -0.0425, -0.0017, 0.0937, -0.043, -0.0017, 0.0948, -0.0427, -0.0039, 0.0937, -0.043, -0.0039, 0.0948, -0.0427, -0.0017, 0.0964, -0.0425, -0.0039, 0.0964, -0.0425, 0.0, 0.0986, -0.0456, 0.0, 0.1007, -0.0454, -0.0017, 0.0985, -0.0456, -0.0039, 0.0985, -0.0456, -0.0017, 0.1007, -0.0454, -0.0039, 0.1007, -0.0454, 0.0, 0.1029, -0.0455, 0.0, 0.105, -0.0457, -0.0017, 0.1029, -0.0455, -0.0039, 0.1029, -0.0455, -0.0017, 0.1051, -0.0457, -0.0039, 0.1051, -0.0457, 0.0, 0.0985, -0.0439, 0.0, 0.0985, -0.0425, 0.0, 0.1007, -0.0439, 0.0, 0.1007, -0.0425, -0.0017, 0.0985, -0.0425,
-0.0039, 0.0985, -0.0425, -0.0017, 0.1007, -0.0425, -0.0039, 0.1007, -0.0425, 0.0, 0.1029, -0.0439, 0.0, 0.1029, -0.0425, 0.0, 0.105, -0.0439, 0.0, 0.105, -0.0425, -0.0017, 0.1029, -0.0425, -0.0039, 0.1029, -0.0425, -0.0017, 0.105, -0.0425, -0.0039, 0.105, -0.0425, 0.0, 0.0716, -0.0219, 0.0, 0.0716, -0.0201, 0.0, 0.0747, -0.0219, 0.0, 0.0747, -0.0201, -0.0017, 0.0716, -0.0219, -0.0017, 0.0716, -0.0201, -0.0039, 0.0716, -0.0219, -0.0039, 0.0716, -0.0201, -0.0017, 0.0747, -0.0219, -0.0039, 0.0747, -0.0219, 0.0, 0.0768, -0.0219, 0.0, 0.0768, -0.0201, 0.0, 0.079, -0.0219, 0.0, 0.079, -0.0201, -0.0017, 0.0768, -0.0219, -0.0039, 0.0768, -0.0219, -0.0017, 0.079, -0.0219, -0.0039, 0.079, -0.0219, 0.0, 0.0812, -0.0219,
0.0, 0.0812, -0.0201, 0.0, 0.0833, -0.0219, 0.0, 0.0833, -0.0201, -0.0017, 0.0812, -0.0219, -0.0039, 0.0812, -0.0219, -0.0017, 0.0833, -0.0219, -0.0017, 0.0833, -0.0201, -0.0039, 0.0833, -0.0219, -0.0039, 0.0833, -0.0201, 0.0, 0.0716, -0.0179, 0.0, 0.0716, -0.0157, 0.0, 0.0747, -0.0179, 0.0, 0.0747, -0.0157, 0.0, 0.0716, -0.0136, 0.0, 0.0716, -0.0114, 0.0, 0.0747, -0.0136, 0.0, 0.0747, -0.0114, -0.0017, 0.0716, -0.0179, -0.0017, 0.0716, -0.0157, -0.0039, 0.0716, -0.0179, -0.0039, 0.0716, -0.0157, -0.0017, 0.0716, -0.0136, -0.0017, 0.0716, -0.0114, -0.0039, 0.0716, -0.0136, -0.0039, 0.0716, -0.0114, 0.0, 0.0768, -0.0179, 0.0, 0.0768, -0.0157, 0.0, 0.079, -0.0179, 0.0, 0.079, -0.0157, 0.0, 0.0768, -0.0136,
0.0, 0.0768, -0.0114, 0.0, 0.079, -0.0136, 0.0, 0.079, -0.0114, 0.0, 0.0716, -0.0092, 0.0, 0.0716, -0.0071, 0.0, 0.0747, -0.0092, 0.0, 0.0747, -0.0071, 0.0, 0.0716, -0.0049, 0.0, 0.0716, -0.0027, 0.0, 0.0747, -0.0049, 0.0, 0.0747, -0.0027, -0.0017, 0.0716, -0.0092, -0.0017, 0.0716, -0.0071, -0.0039, 0.0716, -0.0092, -0.0039, 0.0716, -0.0071, -0.0017, 0.0716, -0.0049, -0.0017, 0.0716, -0.0027, -0.0039, 0.0716, -0.0049, -0.0039, 0.0716, -0.0027, 0.0, 0.0768, -0.0092, 0.0, 0.0768, -0.0071, 0.0, 0.079, -0.0092, 0.0, 0.079, -0.0071, 0.0, 0.0768, -0.0049, 0.0, 0.0768, -0.0027, 0.0, 0.079, -0.0049, 0.0, 0.079, -0.0027, 0.0, 0.0812, -0.0179, 0.0, 0.0812, -0.0157, 0.0, 0.0833, -0.0179,
0.0, 0.0833, -0.0157, 0.0, 0.0812, -0.0136, 0.0, 0.0812, -0.0114, 0.0, 0.0833, -0.0136, 0.0, 0.0833, -0.0114, -0.0017, 0.0833, -0.0179, -0.0017, 0.0833, -0.0157, -0.0039, 0.0833, -0.0179, -0.0039, 0.0833, -0.0157, -0.0017, 0.0833, -0.0136, -0.0017, 0.0833, -0.0114, -0.0039, 0.0833, -0.0136, -0.0039, 0.0833, -0.0114, 0.0, 0.0812, -0.0092, 0.0, 0.0812, -0.0071, 0.0, 0.0833, -0.0092, 0.0, 0.0833, -0.0071, 0.0, 0.0812, -0.0049, 0.0, 0.0812, -0.0027, 0.0, 0.0833, -0.0049, 0.0, 0.0833, -0.0027, -0.0017, 0.0833, -0.0092, -0.0017, 0.0833, -0.0071, -0.0039, 0.0833, -0.0092, -0.0039, 0.0833, -0.0071, -0.0017, 0.0833, -0.0049, -0.0017, 0.0833, -0.0027, -0.0039, 0.0833, -0.0049, -0.0039, 0.0833, -0.0027, 0.0, 0.1065, -0.0668,
0.0, 0.1074, -0.0667, 0.0, 0.1072, -0.0626, 0.0, 0.1097, -0.0661, 0.0, 0.1094, -0.0635, -0.0017, 0.1066, -0.0668, -0.0039, 0.1066, -0.0668, -0.0017, 0.1077, -0.0666, -0.0017, 0.1071, -0.0627, -0.0039, 0.1077, -0.0666, -0.0039, 0.1071, -0.0627, -0.0017, 0.1094, -0.0662, -0.0039, 0.1094, -0.0662, 0.0, 0.1113, -0.0653, 0.0, 0.1123, -0.0646, 0.0, 0.1138, -0.0636, -0.0017, 0.1115, -0.0652, -0.0017, 0.1125, -0.0645, -0.0039, 0.1115, -0.0652, -0.0039, 0.1125, -0.0645, -0.0017, 0.1138, -0.0635, -0.0039, 0.1138, -0.0635, 0.0, 0.1082, -0.0623, 0.0, 0.1093, -0.062, -0.0017, 0.1082, -0.0624, -0.0039, 0.1082, -0.0624, -0.0017, 0.1094, -0.062, -0.0039, 0.1094, -0.062, 0.0, 0.1112, -0.0608, 0.0, 0.1122, -0.0596, 0.0, 0.1148, -0.0622,
0.0, 0.1127, -0.0587, 0.0, 0.1134, -0.0569, 0.0, 0.1136, -0.0548, -0.0017, 0.1111, -0.0608, -0.0017, 0.1122, -0.0596, -0.0039, 0.1111, -0.0608, -0.0039, 0.1122, -0.0596, -0.0017, 0.1148, -0.0623, -0.0017, 0.1129, -0.0584, -0.0039, 0.1148, -0.0623, -0.0039, 0.1129, -0.0584, -0.0017, 0.1134, -0.057, -0.0017, 0.1136, -0.0548, -0.0039, 0.1134, -0.057, -0.0039, 0.1136, -0.0548, 0.0, 0.1156, -0.0612, 0.0, 0.1167, -0.0593, 0.0, 0.117, -0.0581, 0.0, 0.1159, -0.057, 0.0, 0.1159, -0.0548, 0.0, 0.1173, -0.057, 0.0, 0.1176, -0.0548, -0.0017, 0.1155, -0.0613, -0.0017, 0.1167, -0.0593, -0.0039, 0.1155, -0.0613, -0.0039, 0.1167, -0.0593, -0.0017, 0.117, -0.0581, -0.0039, 0.117, -0.0581, -0.0017, 0.1173, -0.057, -0.0017, 0.1176, -0.0548,
-0.0039, 0.1173, -0.057, -0.0039, 0.1176, -0.0548, 0.0, 0.107, -0.0462, 0.0, 0.11, -0.0476, 0.0, 0.109, -0.047, -0.0017, 0.1072, -0.0463, -0.0039, 0.1072, -0.0463, -0.0017, 0.1099, -0.0476, -0.0017, 0.1088, -0.047, -0.0039, 0.1099, -0.0476, -0.0039, 0.1088, -0.047, 0.0, 0.1122, -0.0498, 0.0, 0.1134, -0.0528, 0.0, 0.1128, -0.0508, 0.0, 0.1112, -0.0486, 0.0, 0.1116, -0.0461, 0.0, 0.1137, -0.0483, 0.0, 0.1142, -0.0456, -0.0017, 0.1123, -0.0499, -0.0039, 0.1123, -0.0499, -0.0017, 0.1134, -0.0526, -0.0017, 0.1129, -0.051, -0.0039, 0.1134, -0.0526, -0.0039, 0.1129, -0.051, -0.0017, 0.1112, -0.0486, -0.0039, 0.1112, -0.0486, -0.0017, 0.1142, -0.0456, -0.0039, 0.1142, -0.0456, 0.0, 0.1072, -0.0439, 0.0, 0.1072, -0.0426,
0.0, 0.1096, -0.043, 0.0, 0.1086, -0.0427, -0.0017, 0.1072, -0.0426, -0.0039, 0.1072, -0.0426, -0.0017, 0.1098, -0.043, -0.0017, 0.1088, -0.0427, -0.0039, 0.1098, -0.043, -0.0039, 0.1088, -0.0427, 0.0, 0.1119, -0.0437, 0.0, 0.1131, -0.0447, -0.0017, 0.1116, -0.0436, -0.0039, 0.1116, -0.0436, -0.0017, 0.1131, -0.0447, -0.0039, 0.1131, -0.0447, 0.0, 0.1159, -0.0526, 0.0, 0.1169, -0.0501, 0.0, 0.1175, -0.0526, 0.0, 0.1172, -0.0511, 0.0, 0.1161, -0.048, 0.0, 0.1152, -0.0468, -0.0017, 0.1168, -0.0498, -0.0039, 0.1168, -0.0498, -0.0017, 0.1175, -0.0526, -0.0017, 0.1172, -0.0509, -0.0039, 0.1175, -0.0526, -0.0039, 0.1172, -0.0509, -0.0017, 0.1162, -0.0482, -0.0017, 0.1152, -0.0467, -0.0039, 0.1162, -0.0482, -0.0039, 0.1152, -0.0467,
0.0, 0.1363, -0.067, 0.0, 0.1363, -0.0656, 0.0, 0.1363, -0.0635, -0.0017, 0.1363, -0.067, -0.0039, 0.1363, -0.067, -0.0017, 0.1363, -0.0656, -0.0017, 0.1363, -0.0635, -0.0039, 0.1363, -0.0656, -0.0039, 0.1363, -0.0635, 0.0, 0.1376, -0.067, 0.0, 0.1402, -0.067, 0.0, 0.1376, -0.0656, 0.0, 0.1376, -0.0635, 0.0, 0.1402, -0.0656, 0.0, 0.1402, -0.0635, -0.0017, 0.1376, -0.067, -0.0039, 0.1376, -0.067, -0.0017, 0.1402, -0.067, -0.0039, 0.1402, -0.067, -0.0017, 0.1402, -0.0656, -0.0017, 0.1402, -0.0635, -0.0039, 0.1402, -0.0656, -0.0039, 0.1402, -0.0635, 0.0, 0.1363, -0.0613, 0.0, 0.1363, -0.0591, 0.0, 0.1363, -0.057, 0.0, 0.1363, -0.0548, -0.0017, 0.1363, -0.0613, -0.0017, 0.1363, -0.0591, -0.0039, 0.1363, -0.0613,
-0.0039, 0.1363, -0.0591, -0.0017, 0.1363, -0.057, -0.0017, 0.1363, -0.0548, -0.0039, 0.1363, -0.057, -0.0039, 0.1363, -0.0548, 0.0, 0.1376, -0.0613, 0.0, 0.1376, -0.0591, 0.0, 0.1402, -0.0613, 0.0, 0.1402, -0.0591, 0.0, 0.1376, -0.057, 0.0, 0.1376, -0.0548, 0.0, 0.1402, -0.057, 0.0, 0.1402, -0.0548, -0.0017, 0.1402, -0.0613, -0.0017, 0.1402, -0.0591, -0.0039, 0.1402, -0.0613, -0.0039, 0.1402, -0.0591, -0.0017, 0.1402, -0.057, -0.0017, 0.1402, -0.0548, -0.0039, 0.1402, -0.057, -0.0039, 0.1402, -0.0548, 0.0, 0.1363, -0.0526, 0.0, 0.1363, -0.0505, 0.0, 0.1363, -0.0483, 0.0, 0.1363, -0.0461, -0.0017, 0.1363, -0.0526, -0.0017, 0.1363, -0.0505, -0.0039, 0.1363, -0.0526, -0.0039, 0.1363, -0.0505, -0.0017, 0.1363, -0.0483,
-0.0017, 0.1363, -0.0461, -0.0039, 0.1363, -0.0483, -0.0039, 0.1363, -0.0461, 0.0, 0.1376, -0.0526, 0.0, 0.1376, -0.0505, 0.0, 0.1402, -0.0526, 0.0, 0.1402, -0.0505, 0.0, 0.1376, -0.0483, 0.0, 0.1376, -0.0461, 0.0, 0.1402, -0.0483, 0.0, 0.1402, -0.0454, -0.0017, 0.1402, -0.0526, -0.0017, 0.1402, -0.0505, -0.0039, 0.1402, -0.0526, -0.0039, 0.1402, -0.0505, -0.0017, 0.1402, -0.0483, -0.0017, 0.1402, -0.0454, -0.0039, 0.1402, -0.0483, -0.0039, 0.1402, -0.0454, 0.0, 0.1363, -0.0439, 0.0, 0.1363, -0.0425, -0.0017, 0.1363, -0.0439, -0.0017, 0.1363, -0.0425, -0.0039, 0.1363, -0.0439, -0.0039, 0.1363, -0.0425, 0.0, 0.1376, -0.0439, 0.0, 0.1376, -0.0425, 0.0, 0.1398, -0.0439, 0.0, 0.1398, -0.0425, -0.0017, 0.1376, -0.0425,
-0.0039, 0.1376, -0.0425, -0.0017, 0.1398, -0.0425, -0.0039, 0.1398, -0.0425, 0.0, 0.1431, -0.0591, 0.0, 0.1431, -0.0562, -0.0017, 0.1431, -0.0591, -0.0039, 0.1431, -0.0591, -0.0017, 0.1431, -0.0562, -0.0039, 0.1431, -0.0562, 0.0, 0.1463, -0.0591, 0.0, 0.1484, -0.0592, 0.0, 0.1463, -0.0562, 0.0, 0.1484, -0.0562, -0.0017, 0.1463, -0.0591, -0.0039, 0.1463, -0.0591, -0.0017, 0.1485, -0.0592, -0.0039, 0.1485, -0.0592, -0.0017, 0.1463, -0.0562, -0.0039, 0.1463, -0.0562, -0.0017, 0.1484, -0.0562, -0.0039, 0.1484, -0.0562, 0.0, 0.1553, -0.0632, 0.0, 0.1576, -0.0652, 0.0, 0.1564, -0.0643, -0.0017, 0.1553, -0.0632, -0.0039, 0.1553, -0.0632, -0.0017, 0.1575, -0.0651, -0.0017, 0.1564, -0.0643, -0.0039, 0.1575, -0.0651, -0.0039, 0.1564, -0.0643,
0.0, 0.1504, -0.0596, 0.0, 0.153, -0.061, 0.0, 0.1521, -0.0603, 0.0, 0.1506, -0.0562, 0.0, 0.1527, -0.0562, -0.0017, 0.1506, -0.0596, -0.0039, 0.1506, -0.0596, -0.0017, 0.1529, -0.0609, -0.0017, 0.1518, -0.0602, -0.0039, 0.1529, -0.0609, -0.0039, 0.1518, -0.0602, -0.0017, 0.1506, -0.0562, -0.0039, 0.1506, -0.0562, -0.0017, 0.1528, -0.0562, -0.0039, 0.1528, -0.0562, 0.0, 0.1542, -0.0621, 0.0, 0.155, -0.0591, 0.0, 0.1571, -0.0613, 0.0, 0.1571, -0.0591, 0.0, 0.1549, -0.056, 0.0, 0.1557, -0.0557, 0.0, 0.1571, -0.057, 0.0, 0.1571, -0.0553, -0.0017, 0.1542, -0.0621, -0.0039, 0.1542, -0.0621, -0.0017, 0.1546, -0.0561, -0.0017, 0.1556, -0.0558, -0.0039, 0.1546, -0.0561, -0.0039, 0.1556, -0.0558, -0.0017, 0.1572, -0.0553,
-0.0039, 0.1572, -0.0553, 0.0, 0.1419, -0.0454, 0.0, 0.1441, -0.0454, -0.0017, 0.1419, -0.0454, -0.0039, 0.1419, -0.0454, -0.0017, 0.1441, -0.0454, -0.0039, 0.1441, -0.0454, 0.0, 0.1463, -0.0454, 0.0, 0.1484, -0.0454, -0.0017, 0.1463, -0.0454, -0.0039, 0.1463, -0.0454, -0.0017, 0.1484, -0.0454, -0.0039, 0.1484, -0.0454, 0.0, 0.1419, -0.0439, 0.0, 0.1419, -0.0425, 0.0, 0.1441, -0.0439, 0.0, 0.1441, -0.0425, -0.0017, 0.1419, -0.0425, -0.0039, 0.1419, -0.0425, -0.0017, 0.1441, -0.0425, -0.0039, 0.1441, -0.0425, 0.0, 0.1463, -0.0439, 0.0, 0.1463, -0.0425, 0.0, 0.1484, -0.0439, 0.0, 0.1484, -0.0425, -0.0017, 0.1463, -0.0425, -0.0039, 0.1463, -0.0425, -0.0017, 0.1484, -0.0425, -0.0039, 0.1484, -0.0425, 0.0, 0.1506, -0.0454,
0.0, 0.1528, -0.0456, -0.0017, 0.1506, -0.0454, -0.0039, 0.1506, -0.0454, -0.0017, 0.1528, -0.0456, -0.0039, 0.1528, -0.0456, 0.0, 0.1549, -0.0459, 0.0, 0.1579, -0.0475, 0.0, 0.1569, -0.0468, -0.0017, 0.155, -0.0459, -0.0039, 0.155, -0.0459, -0.0017, 0.1579, -0.0474, -0.0017, 0.1568, -0.0467, -0.0039, 0.1579, -0.0474, -0.0039, 0.1568, -0.0467, 0.0, 0.1506, -0.0439, 0.0, 0.1506, -0.0425, 0.0, 0.1528, -0.0439, 0.0, 0.1528, -0.0425, -0.0017, 0.1506, -0.0425, -0.0039, 0.1506, -0.0425, -0.0017, 0.1528, -0.0425, -0.0039, 0.1528, -0.0425, 0.0, 0.155, -0.0439, 0.0, 0.155, -0.0425, 0.0, 0.1571, -0.0439, 0.0, 0.1571, -0.0426, -0.0017, 0.155, -0.0425, -0.0039, 0.155, -0.0425, -0.0017, 0.1571, -0.0426, -0.0039, 0.1571, -0.0426,
0.0, 0.1621, -0.0669, 0.0, 0.159, -0.0661, 0.0, 0.1593, -0.0635, 0.0, 0.1612, -0.0667, 0.0, 0.1619, -0.0632, -0.0017, 0.1619, -0.0669, -0.0039, 0.1619, -0.0669, -0.0017, 0.1592, -0.0662, -0.0039, 0.1592, -0.0662, -0.0017, 0.1608, -0.0666, -0.0017, 0.1619, -0.0632, -0.0039, 0.1608, -0.0666, -0.0039, 0.1619, -0.0632, 0.0, 0.1648, -0.0671, 0.0, 0.1636, -0.0653, 0.0, 0.1628, -0.0643, -0.0017, 0.1648, -0.0671, -0.0039, 0.1648, -0.0671, -0.0017, 0.1638, -0.0656, -0.0017, 0.1628, -0.0643, -0.0039, 0.1638, -0.0656, -0.0039, 0.1628, -0.0643, 0.0, 0.16, -0.0609, 0.0, 0.1588, -0.0591, 0.0, 0.1608, -0.0619, 0.0, 0.1606, -0.0583, 0.0, 0.1593, -0.057, 0.0, 0.1588, -0.0543, 0.0, 0.1618, -0.0574, 0.0, 0.1615, -0.0548,
-0.0017, 0.1599, -0.0608, -0.0017, 0.1588, -0.0591, -0.0039, 0.1599, -0.0608, -0.0039, 0.1588, -0.0591, -0.0017, 0.1608, -0.0619, -0.0017, 0.1607, -0.0583, -0.0039, 0.1608, -0.0619, -0.0039, 0.1607, -0.0583, -0.0017, 0.1587, -0.0543, -0.0039, 0.1587, -0.0543, -0.0017, 0.1618, -0.0574, -0.0039, 0.1618, -0.0574, 0.0, 0.1629, -0.0563, 0.0, 0.1637, -0.0551, -0.0017, 0.1629, -0.0563, -0.0017, 0.1638, -0.0549, -0.0039, 0.1629, -0.0563, -0.0039, 0.1638, -0.0549, 0.0, 0.1598, -0.0525, 0.0, 0.1603, -0.0503, 0.0, 0.1615, -0.0526, 0.0, 0.1615, -0.0505, 0.0, 0.1591, -0.0484, 0.0, 0.1593, -0.0461, 0.0, 0.1615, -0.0483, 0.0, 0.1615, -0.0461, -0.0017, 0.1596, -0.053, -0.0017, 0.1603, -0.0503, -0.0039, 0.1596, -0.053, -0.0039, 0.1603, -0.0503,
-0.0017, 0.1585, -0.0479, -0.0039, 0.1585, -0.0479, 0.0, 0.1645, -0.0529, 0.0, 0.1647, -0.0505, 0.0, 0.1644, -0.048, 0.0, 0.1633, -0.0459, -0.0017, 0.1646, -0.0527, -0.0017, 0.1647, -0.0505, -0.0039, 0.1646, -0.0527, -0.0039, 0.1647, -0.0505, -0.0017, 0.1645, -0.048, -0.0017, 0.1634, -0.046, -0.0039, 0.1645, -0.048, -0.0039, 0.1634, -0.046, 0.0, 0.1595, -0.043, 0.0, 0.1584, -0.0428, 0.0, 0.1618, -0.0442, -0.0017, 0.1596, -0.0431, -0.0017, 0.1585, -0.0428, -0.0039, 0.1596, -0.0431, -0.0039, 0.1585, -0.0428, -0.0017, 0.162, -0.0444, -0.0039, 0.162, -0.0444, 0.0, 0.1626, -0.045, -0.0017, 0.1626, -0.045, -0.0039, 0.1626, -0.045, 0.0, 0.1412, -0.0219, 0.0, 0.1412, -0.0201, 0.0, 0.1441, -0.0219, 0.0, 0.1441, -0.0201,
-0.0017, 0.1412, -0.0219, -0.0017, 0.1412, -0.0201, -0.0039, 0.1412, -0.0219, -0.0039, 0.1412, -0.0201, -0.0017, 0.1441, -0.0219, -0.0039, 0.1441, -0.0219, 0.0, 0.1463, -0.0219, 0.0, 0.1463, -0.0201, 0.0, 0.1484, -0.0219, 0.0, 0.1484, -0.0201, -0.0017, 0.1463, -0.0219, -0.0039, 0.1463, -0.0219, -0.0017, 0.1484, -0.0219, -0.0039, 0.1484, -0.0219, 0.0, 0.1506, -0.0219, 0.0, 0.1506, -0.0201, 0.0, 0.1528, -0.0219, 0.0, 0.1528, -0.0201, -0.0017, 0.1506, -0.0219, -0.0039, 0.1506, -0.0219, -0.0017, 0.1528, -0.0219, -0.0039, 0.1528, -0.0219, 0.0, 0.155, -0.0219, 0.0, 0.155, -0.0201, 0.0, 0.1571, -0.0219, 0.0, 0.1571, -0.0201, -0.0017, 0.155, -0.0219, -0.0039, 0.155, -0.0219, -0.0017, 0.1571, -0.0219, -0.0039, 0.1571, -0.0219,
0.0, 0.1412, -0.0179, 0.0, 0.1412, -0.0157, 0.0, 0.1441, -0.0179, 0.0, 0.1441, -0.0157, 0.0, 0.1412, -0.0136, 0.0, 0.1412, -0.0114, 0.0, 0.1441, -0.0136, 0.0, 0.1441, -0.0114, -0.0017, 0.1412, -0.0179, -0.0017, 0.1412, -0.0157, -0.0039, 0.1412, -0.0179, -0.0039, 0.1412, -0.0157, -0.0017, 0.1412, -0.0136, -0.0017, 0.1412, -0.0114, -0.0039, 0.1412, -0.0136, -0.0039, 0.1412, -0.0114, 0.0, 0.1463, -0.0179, 0.0, 0.1463, -0.0157, 0.0, 0.1484, -0.0179, 0.0, 0.1484, -0.0157, 0.0, 0.1463, -0.0136, 0.0, 0.1463, -0.0114, 0.0, 0.1484, -0.0136, 0.0, 0.1484, -0.0114, 0.0, 0.1412, -0.0092, 0.0, 0.1412, -0.0071, 0.0, 0.1441, -0.0092, 0.0, 0.1441, -0.0071, 0.0, 0.1412, -0.0049, 0.0, 0.1412, -0.0027,
0.0, 0.1441, -0.0049, 0.0, 0.1441, -0.0027, -0.0017, 0.1412, -0.0092, -0.0017, 0.1412, -0.0071, -0.0039, 0.1412, -0.0092, -0.0039, 0.1412, -0.0071, -0.0017, 0.1412, -0.0049, -0.0017, 0.1412, -0.0027, -0.0039, 0.1412, -0.0049, -0.0039, 0.1412, -0.0027, 0.0, 0.1463, -0.0092, 0.0, 0.1463, -0.0071, 0.0, 0.1484, -0.0092, 0.0, 0.1484, -0.0071, 0.0, 0.1463, -0.0049, 0.0, 0.1463, -0.0027, 0.0, 0.1484, -0.0049, 0.0, 0.1484, -0.0027, 0.0, 0.1506, -0.0179, 0.0, 0.1506, -0.0157, 0.0, 0.1528, -0.0179, 0.0, 0.1528, -0.0157, 0.0, 0.1506, -0.0136, 0.0, 0.1506, -0.0114, 0.0, 0.1528, -0.0136, 0.0, 0.1528, -0.0114, 0.0, 0.155, -0.0179, 0.0, 0.155, -0.0157, 0.0, 0.1571, -0.0179, 0.0, 0.1571, -0.0157,
0.0, 0.155, -0.0136, 0.0, 0.155, -0.0114, 0.0, 0.1571, -0.0136, 0.0, 0.1571, -0.0114, 0.0, 0.1506, -0.0092, 0.0, 0.1506, -0.0071, 0.0, 0.152, -0.0101, 0.0, 0.152, -0.0071, 0.0, 0.1506, -0.0049, 0.0, 0.1506, -0.0027, 0.0, 0.152, -0.0049, 0.0, 0.152, -0.0027, -0.0017, 0.152, -0.0101, -0.0017, 0.152, -0.0071, -0.0039, 0.152, -0.0101, -0.0039, 0.152, -0.0071, -0.0017, 0.152, -0.0049, -0.0017, 0.152, -0.0027, -0.0039, 0.152, -0.0049, -0.0039, 0.152, -0.0027, 0.0, 0.155, -0.0101, 0.0, 0.1571, -0.0101, -0.0017, 0.155, -0.0101, -0.0039, 0.155, -0.0101, -0.0017, 0.1571, -0.0101, -0.0039, 0.1571, -0.0101, 0.0, 0.1593, -0.0219, 0.0, 0.1593, -0.0201, 0.0, 0.1615, -0.0219, 0.0, 0.1615, -0.0201,
-0.0017, 0.1593, -0.0219, -0.0039, 0.1593, -0.0219, -0.0017, 0.1615, -0.0219, -0.0039, 0.1615, -0.0219, 0.0, 0.1636, -0.0219, 0.0, 0.1636, -0.0201, 0.0, 0.1658, -0.0219, 0.0, 0.1658, -0.0201, -0.0017, 0.1636, -0.0219, -0.0039, 0.1636, -0.0219, -0.0017, 0.1658, -0.0219, -0.0039, 0.1658, -0.0219, 0.0, 0.168, -0.0219, 0.0, 0.168, -0.0201, 0.0, 0.1701, -0.0219, 0.0, 0.1701, -0.0201, -0.0017, 0.168, -0.0219, -0.0039, 0.168, -0.0219, -0.0017, 0.1701, -0.0219, -0.0039, 0.1701, -0.0219, 0.0, 0.1723, -0.0219, 0.0, 0.1723, -0.0201, 0.0, 0.1745, -0.0219, 0.0, 0.1745, -0.0201, -0.0017, 0.1723, -0.0219, -0.0039, 0.1723, -0.0219, -0.0017, 0.1745, -0.0219, -0.0039, 0.1745, -0.0219, 0.0, 0.1593, -0.0179, 0.0, 0.1593, -0.0157,
0.0, 0.1615, -0.0179, 0.0, 0.1615, -0.0157, 0.0, 0.1593, -0.0136, 0.0, 0.1593, -0.0114, 0.0, 0.1615, -0.0136, 0.0, 0.1615, -0.0114, 0.0, 0.1636, -0.0179, 0.0, 0.1636, -0.0157, 0.0, 0.1658, -0.0179, 0.0, 0.1658, -0.0157, 0.0, 0.1636, -0.0136, 0.0, 0.1636, -0.0114, 0.0, 0.1658, -0.0136, 0.0, 0.1658, -0.0114, 0.0, 0.1593, -0.0101, 0.0, 0.1615, -0.0101, -0.0017, 0.1593, -0.0101, -0.0039, 0.1593, -0.0101, -0.0017, 0.1615, -0.0101, -0.0039, 0.1615, -0.0101, 0.0, 0.1636, -0.0101, 0.0, 0.1658, -0.0101, -0.0017, 0.1636, -0.0101, -0.0039, 0.1636, -0.0101, -0.0017, 0.1658, -0.0101, -0.0039, 0.1658, -0.0101, 0.0, 0.168, -0.0179, 0.0, 0.168, -0.0157, 0.0, 0.1701, -0.0179, 0.0, 0.1701, -0.0157,
0.0, 0.168, -0.0136, 0.0, 0.168, -0.0114, 0.0, 0.1701, -0.0136, 0.0, 0.1701, -0.0114, 0.0, 0.1723, -0.0179, 0.0, 0.1723, -0.0157, 0.0, 0.1745, -0.0179, 0.0, 0.1745, -0.0157, 0.0, 0.1723, -0.0136, 0.0, 0.1723, -0.0114, 0.0, 0.1745, -0.0136, 0.0, 0.1745, -0.0114, 0.0, 0.168, -0.0101, 0.0, 0.1701, -0.0101, -0.0017, 0.168, -0.0101, -0.0039, 0.168, -0.0101, -0.0017, 0.1701, -0.0101, -0.0039, 0.1701, -0.0101, 0.0, 0.1723, -0.0101, 0.0, 0.1745, -0.0101, -0.0017, 0.1723, -0.0101, -0.0039, 0.1723, -0.0101, -0.0017, 0.1745, -0.0101, -0.0039, 0.1745, -0.0101, 0.0, 0.1767, -0.0219, 0.0, 0.1767, -0.0201, 0.0, 0.1788, -0.0219, 0.0, 0.1788, -0.0201, -0.0017, 0.1767, -0.0219, -0.0039, 0.1767, -0.0219,
-0.0017, 0.1788, -0.0219, -0.0039, 0.1788, -0.0219, 0.0, 0.181, -0.0219, 0.0, 0.181, -0.0201, 0.0, 0.1832, -0.0219, 0.0, 0.1832, -0.0201, -0.0017, 0.181, -0.0219, -0.0039, 0.181, -0.0219, -0.0017, 0.1832, -0.0219, -0.0039, 0.1832, -0.0219, 0.0, 0.1853, -0.0219, 0.0, 0.1853, -0.0201, 0.0, 0.1875, -0.0219, 0.0, 0.1875, -0.0201, -0.0017, 0.1853, -0.0219, -0.0039, 0.1853, -0.0219, -0.0017, 0.1875, -0.0219, -0.0039, 0.1875, -0.0219, 0.0, 0.1897, -0.0219, 0.0, 0.1897, -0.0201, 0.0, 0.1918, -0.0219, 0.0, 0.1918, -0.0201, -0.0017, 0.1897, -0.0219, -0.0039, 0.1897, -0.0219, -0.0017, 0.1918, -0.0219, -0.0039, 0.1918, -0.0219, 0.0, 0.1767, -0.0179, 0.0, 0.1767, -0.0157, 0.0, 0.1788, -0.0179, 0.0, 0.1788, -0.0157,
0.0, 0.1767, -0.0136, 0.0, 0.1767, -0.0114, 0.0, 0.1788, -0.0136, 0.0, 0.1788, -0.0114, 0.0, 0.181, -0.0179, 0.0, 0.181, -0.0157, 0.0, 0.1832, -0.0179, 0.0, 0.1832, -0.0157, 0.0, 0.181, -0.0136, 0.0, 0.181, -0.0114, 0.0, 0.1832, -0.0136, 0.0, 0.1832, -0.0114, 0.0, 0.1767, -0.0101, 0.0, 0.1788, -0.0101, -0.0017, 0.1767, -0.0101, -0.0039, 0.1767, -0.0101, -0.0017, 0.1788, -0.0101, -0.0039, 0.1788, -0.0101, 0.0, 0.181, -0.0101, 0.0, 0.1832, -0.0101, -0.0017, 0.181, -0.0101, -0.0039, 0.181, -0.0101, -0.0017, 0.1832, -0.0101, -0.0039, 0.1832, -0.0101, 0.0, 0.1853, -0.0179, 0.0, 0.1853, -0.0157, 0.0, 0.1875, -0.0179, 0.0, 0.1875, -0.0157, 0.0, 0.1853, -0.0136, 0.0, 0.1853, -0.0114,
0.0, 0.1875, -0.0136, 0.0, 0.1875, -0.0114, 0.0, 0.1897, -0.0179, 0.0, 0.1897, -0.0157, 0.0, 0.1918, -0.0179, 0.0, 0.1918, -0.0157, 0.0, 0.1897, -0.0136, 0.0, 0.1897, -0.0114, 0.0, 0.1918, -0.0136, 0.0, 0.1918, -0.0114, 0.0, 0.1853, -0.0101, 0.0, 0.1875, -0.0101, -0.0017, 0.1853, -0.0101, -0.0039, 0.1853, -0.0101, -0.0017, 0.1875, -0.0101, -0.0039, 0.1875, -0.0101, 0.0, 0.1897, -0.0101, 0.0, 0.1918, -0.0101, -0.0017, 0.1897, -0.0101, -0.0039, 0.1897, -0.0101, -0.0017, 0.1918, -0.0101, -0.0039, 0.1918, -0.0101, 0.0, 0.194, -0.0219, 0.0, 0.194, -0.0201, 0.0, 0.1962, -0.0219, 0.0, 0.1962, -0.0201, -0.0017, 0.194, -0.0219, -0.0039, 0.194, -0.0219, -0.0017, 0.1962, -0.0219, -0.0039, 0.1962, -0.0219,
0.0, 0.1984, -0.0219, 0.0, 0.1984, -0.0201, 0.0, 0.2005, -0.0219, 0.0, 0.2005, -0.0201, -0.0017, 0.1984, -0.0219, -0.0039, 0.1984, -0.0219, -0.0017, 0.2005, -0.0219, -0.0039, 0.2005, -0.0219, 0.0, 0.202, -0.0219, 0.0, 0.202, -0.0201, -0.0017, 0.202, -0.0219, -0.0017, 0.202, -0.0201, -0.0039, 0.202, -0.0219, -0.0039, 0.202, -0.0201, 0.0, 0.194, -0.0179, 0.0, 0.194, -0.0157, 0.0, 0.1962, -0.0179, 0.0, 0.1962, -0.0157, 0.0, 0.194, -0.0136, 0.0, 0.194, -0.0114, 0.0, 0.1962, -0.0136, 0.0, 0.1962, -0.0114, 0.0, 0.1984, -0.0179, 0.0, 0.1984, -0.0157, 0.0, 0.2005, -0.0179, 0.0, 0.2005, -0.0157, 0.0, 0.1984, -0.0136, 0.0, 0.1984, -0.0114, 0.0, 0.2005, -0.0136, 0.0, 0.2005, -0.0114,
0.0, 0.194, -0.0101, 0.0, 0.1962, -0.0101, -0.0017, 0.194, -0.0101, -0.0039, 0.194, -0.0101, -0.0017, 0.1962, -0.0101, -0.0039, 0.1962, -0.0101, 0.0, 0.1984, -0.0101, 0.0, 0.2005, -0.0101, -0.0017, 0.1984, -0.0101, -0.0039, 0.1984, -0.0101, -0.0017, 0.2005, -0.0101, -0.0039, 0.2005, -0.0101, 0.0, 0.202, -0.0179, 0.0, 0.202, -0.0157, 0.0, 0.202, -0.0136, 0.0, 0.202, -0.0114, -0.0017, 0.202, -0.0179, -0.0017, 0.202, -0.0157, -0.0039, 0.202, -0.0179, -0.0039, 0.202, -0.0157, -0.0017, 0.202, -0.0136, -0.0017, 0.202, -0.0114, -0.0039, 0.202, -0.0136, -0.0039, 0.202, -0.0114, 0.0, 0.202, -0.0101, -0.0017, 0.202, -0.0101, -0.0039, 0.202, -0.0101, 0.0, 0.2118, -0.067, 0.0, 0.2135, -0.067, 0.0, 0.2118, -0.0656,
0.0, 0.2118, -0.0635, 0.0, 0.2135, -0.0656, 0.0, 0.2135, -0.0635, -0.0017, 0.2118, -0.067, -0.0039, 0.2118, -0.067, -0.0017, 0.2135, -0.067, -0.0039, 0.2135, -0.067, -0.0017, 0.2118, -0.0656, -0.0017, 0.2118, -0.0635, -0.0039, 0.2118, -0.0656, -0.0039, 0.2118, -0.0635, 0.0, 0.2157, -0.067, 0.0, 0.2179, -0.067, 0.0, 0.2157, -0.0656, 0.0, 0.2157, -0.0631, 0.0, 0.2179, -0.0656, 0.0, 0.2179, -0.0631, -0.0017, 0.2157, -0.067, -0.0039, 0.2157, -0.067, -0.0017, 0.2179, -0.067, -0.0039, 0.2179, -0.067, -0.0017, 0.2157, -0.0631, -0.0039, 0.2157, -0.0631, -0.0017, 0.2179, -0.0631, -0.0039, 0.2179, -0.0631, 0.0, 0.2118, -0.0613, 0.0, 0.2118, -0.0591, 0.0, 0.2135, -0.0613, 0.0, 0.2135, -0.0591, 0.0, 0.2118, -0.057,
0.0, 0.2118, -0.0548, 0.0, 0.2135, -0.057, 0.0, 0.2135, -0.0548, -0.0017, 0.2118, -0.0613, -0.0017, 0.2118, -0.0591, -0.0039, 0.2118, -0.0613, -0.0039, 0.2118, -0.0591, -0.0017, 0.2118, -0.057, -0.0017, 0.2118, -0.0548, -0.0039, 0.2118, -0.057, -0.0039, 0.2118, -0.0548, 0.0, 0.2157, -0.0613, 0.0, 0.2157, -0.0591, 0.0, 0.2157, -0.057, 0.0, 0.2157, -0.0548, -0.0017, 0.2157, -0.0613, -0.0017, 0.2157, -0.0591, -0.0039, 0.2157, -0.0613, -0.0039, 0.2157, -0.0591, -0.0017, 0.2157, -0.057, -0.0017, 0.2157, -0.0548, -0.0039, 0.2157, -0.057, -0.0039, 0.2157, -0.0548, 0.0, 0.2201, -0.067, 0.0, 0.2222, -0.067, 0.0, 0.2201, -0.0656, 0.0, 0.2201, -0.0631, 0.0, 0.2222, -0.0656, 0.0, 0.2222, -0.0631, -0.0017, 0.2201, -0.067,
-0.0039, 0.2201, -0.067, -0.0017, 0.2222, -0.067, -0.0039, 0.2222, -0.067, -0.0017, 0.2201, -0.0631, -0.0039, 0.2201, -0.0631, -0.0017, 0.2222, -0.0631, -0.0039, 0.2222, -0.0631, 0.0, 0.2244, -0.067, 0.0, 0.2266, -0.067, 0.0, 0.2244, -0.0656, 0.0, 0.2244, -0.0631, 0.0, 0.2266, -0.0656, 0.0, 0.2266, -0.0631, -0.0017, 0.2244, -0.067, -0.0039, 0.2244, -0.067, -0.0017, 0.2266, -0.067, -0.0039, 0.2266, -0.067, -0.0017, 0.2244, -0.0631, -0.0039, 0.2244, -0.0631, -0.0017, 0.2266, -0.0631, -0.0039, 0.2266, -0.0631, 0.0, 0.2196, -0.0562, 0.0, 0.2196, -0.0548, 0.0, 0.2222, -0.0562, 0.0, 0.2222, -0.0548, -0.0017, 0.2196, -0.0562, -0.0017, 0.2196, -0.0548, -0.0039, 0.2196, -0.0562, -0.0039, 0.2196, -0.0548, -0.0017, 0.2222, -0.0562,
-0.0039, 0.2222, -0.0562, 0.0, 0.2244, -0.0562, 0.0, 0.2244, -0.0548, 0.0, 0.2266, -0.0562, 0.0, 0.2266, -0.0548, -0.0017, 0.2244, -0.0562, -0.0039, 0.2244, -0.0562, -0.0017, 0.2266, -0.0562, -0.0039, 0.2266, -0.0562, 0.0, 0.2118, -0.0526, 0.0, 0.2118, -0.0505, 0.0, 0.2135, -0.0526, 0.0, 0.2135, -0.0505, 0.0, 0.2118, -0.0483, 0.0, 0.2118, -0.0461, 0.0, 0.2135, -0.0483, 0.0, 0.2135, -0.0461, -0.0017, 0.2118, -0.0526, -0.0017, 0.2118, -0.0505, -0.0039, 0.2118, -0.0526, -0.0039, 0.2118, -0.0505, -0.0017, 0.2118, -0.0483, -0.0017, 0.2118, -0.0461, -0.0039, 0.2118, -0.0483, -0.0039, 0.2118, -0.0461, 0.0, 0.2157, -0.0526, 0.0, 0.2157, -0.0505, 0.0, 0.2157, -0.0483, 0.0, 0.2157, -0.0454, 0.0, 0.2179, -0.0454,
-0.0017, 0.2157, -0.0526, -0.0017, 0.2157, -0.0505, -0.0039, 0.2157, -0.0526, -0.0039, 0.2157, -0.0505, -0.0017, 0.2157, -0.0483, -0.0017, 0.2157, -0.0454, -0.0039, 0.2157, -0.0483, -0.0039, 0.2157, -0.0454, -0.0017, 0.2179, -0.0454, -0.0039, 0.2179, -0.0454, 0.0, 0.2118, -0.0439, 0.0, 0.2118, -0.0425, 0.0, 0.2135, -0.0439, 0.0, 0.2135, -0.0425, -0.0017, 0.2118, -0.0439, -0.0017, 0.2118, -0.0425, -0.0039, 0.2118, -0.0439, -0.0039, 0.2118, -0.0425, -0.0017, 0.2135, -0.0425, -0.0039, 0.2135, -0.0425, 0.0, 0.2157, -0.0439, 0.0, 0.2157, -0.0425, 0.0, 0.2179, -0.0439, 0.0, 0.2179, -0.0425, -0.0017, 0.2157, -0.0425, -0.0039, 0.2157, -0.0425, -0.0017, 0.2179, -0.0425, -0.0039, 0.2179, -0.0425, 0.0, 0.2196, -0.0523, 0.0, 0.2222, -0.0523,
0.0, 0.2201, -0.0454, 0.0, 0.2222, -0.0454, -0.0017, 0.2196, -0.0523, -0.0039, 0.2196, -0.0523, -0.0017, 0.2222, -0.0523, -0.0039, 0.2222, -0.0523, -0.0017, 0.2201, -0.0454, -0.0039, 0.2201, -0.0454, -0.0017, 0.2222, -0.0454, -0.0039, 0.2222, -0.0454, 0.0, 0.2244, -0.0523, 0.0, 0.2266, -0.0523, 0.0, 0.2244, -0.0454, 0.0, 0.2266, -0.0454, -0.0017, 0.2244, -0.0523, -0.0039, 0.2244, -0.0523, -0.0017, 0.2266, -0.0523, -0.0039, 0.2266, -0.0523, -0.0017, 0.2244, -0.0454, -0.0039, 0.2244, -0.0454, -0.0017, 0.2266, -0.0454, -0.0039, 0.2266, -0.0454, 0.0, 0.2201, -0.0439, 0.0, 0.2201, -0.0425, 0.0, 0.2222, -0.0439, 0.0, 0.2222, -0.0425, -0.0017, 0.2201, -0.0425, -0.0039, 0.2201, -0.0425, -0.0017, 0.2222, -0.0425, -0.0039, 0.2222, -0.0425,
0.0, 0.2244, -0.0439, 0.0, 0.2244, -0.0425, 0.0, 0.2266, -0.0439, 0.0, 0.2266, -0.0425, -0.0017, 0.2244, -0.0425, -0.0039, 0.2244, -0.0425, -0.0017, 0.2266, -0.0425, -0.0039, 0.2266, -0.0425, 0.0, 0.2287, -0.067, 0.0, 0.2304, -0.0669, 0.0, 0.2287, -0.0656, 0.0, 0.2287, -0.0631, 0.0, 0.2312, -0.0668, 0.0, 0.2308, -0.0629, -0.0017, 0.2287, -0.067, -0.0039, 0.2287, -0.067, -0.0017, 0.2307, -0.0669, -0.0039, 0.2307, -0.0669, -0.0017, 0.2287, -0.063, -0.0039, 0.2287, -0.063, -0.0017, 0.2318, -0.0667, -0.0017, 0.2309, -0.0629, -0.0039, 0.2318, -0.0667, -0.0039, 0.2309, -0.0629, 0.0, 0.2333, -0.0665, 0.0, 0.2328, -0.0624, 0.0, 0.2356, -0.0657, 0.0, 0.2353, -0.0635, -0.0017, 0.2331, -0.0665, -0.0017, 0.2325, -0.0625,
-0.0039, 0.2331, -0.0665, -0.0039, 0.2325, -0.0625, -0.0017, 0.2355, -0.0657, -0.0039, 0.2355, -0.0657, 0.0, 0.2287, -0.0563, 0.0, 0.2287, -0.0548, 0.0, 0.2309, -0.0565, 0.0, 0.2309, -0.0548, -0.0017, 0.2287, -0.0563, -0.0039, 0.2287, -0.0563, -0.0017, 0.2309, -0.0565, -0.0039, 0.2309, -0.0565, 0.0, 0.2337, -0.0621, 0.0, 0.2349, -0.0613, 0.0, 0.2363, -0.0591, 0.0, 0.2329, -0.0571, 0.0, 0.2331, -0.0548, 0.0, 0.2346, -0.058, 0.0, 0.2353, -0.0548, -0.0017, 0.2336, -0.0621, -0.0039, 0.2336, -0.0621, -0.0017, 0.2349, -0.0613, -0.0017, 0.2363, -0.0591, -0.0039, 0.2349, -0.0613, -0.0039, 0.2363, -0.0591, -0.0017, 0.2331, -0.0571, -0.0039, 0.2331, -0.0571, -0.0017, 0.2345, -0.0579, -0.0039, 0.2345, -0.0579, 0.0, 0.2365, -0.0648,
0.0, 0.238, -0.0634, -0.0017, 0.2366, -0.0648, -0.0017, 0.238, -0.0634, -0.0039, 0.2366, -0.0648, -0.0039, 0.238, -0.0634, 0.0, 0.2386, -0.0618, 0.0, 0.2374, -0.0591, 0.0, 0.2389, -0.0609, 0.0, 0.239, -0.0591, 0.0, 0.2374, -0.057, 0.0, 0.2384, -0.0547, 0.0, 0.2387, -0.057, 0.0, 0.2384, -0.0546, -0.0017, 0.2384, -0.0622, -0.0039, 0.2384, -0.0622, -0.0017, 0.2388, -0.0612, -0.0017, 0.239, -0.0591, -0.0039, 0.2388, -0.0612, -0.0039, 0.239, -0.0591, -0.0017, 0.2387, -0.057, -0.0017, 0.2384, -0.0544, -0.0039, 0.2387, -0.057, -0.0039, 0.2384, -0.0544, 0.0, 0.2287, -0.0522, 0.0, 0.2309, -0.0522, 0.0, 0.2287, -0.0455, 0.0, 0.2308, -0.0458, -0.0017, 0.2287, -0.0522, -0.0039, 0.2287, -0.0522, -0.0017, 0.2309, -0.0522,
-0.0039, 0.2309, -0.0522, -0.0017, 0.2287, -0.0455, -0.0039, 0.2287, -0.0455, -0.0017, 0.2309, -0.0459, -0.0039, 0.2309, -0.0459, 0.0, 0.2329, -0.0518, 0.0, 0.2338, -0.0512, 0.0, 0.2353, -0.0526, 0.0, 0.2345, -0.0504, 0.0, 0.2339, -0.0475, 0.0, 0.2329, -0.0466, 0.0, 0.2348, -0.0484, 0.0, 0.2353, -0.0461, -0.0017, 0.2326, -0.0519, -0.0017, 0.2337, -0.0513, -0.0039, 0.2326, -0.0519, -0.0039, 0.2337, -0.0513, -0.0017, 0.2345, -0.0504, -0.0039, 0.2345, -0.0504, -0.0017, 0.2339, -0.0475, -0.0017, 0.2328, -0.0465, -0.0039, 0.2339, -0.0475, -0.0039, 0.2328, -0.0465, -0.0017, 0.2348, -0.0484, -0.0039, 0.2348, -0.0484, 0.0, 0.2287, -0.0439, 0.0, 0.2287, -0.0425, 0.0, 0.2309, -0.0439, 0.0, 0.2309, -0.0425, -0.0017, 0.2287, -0.0425,
-0.0039, 0.2287, -0.0425, -0.0017, 0.2309, -0.0425, -0.0039, 0.2309, -0.0425, 0.0, 0.2334, -0.0429, 0.0, 0.2324, -0.0427, 0.0, 0.2356, -0.0438, -0.0017, 0.2337, -0.043, -0.0017, 0.2326, -0.0427, -0.0039, 0.2337, -0.043, -0.0039, 0.2326, -0.0427, -0.0017, 0.2353, -0.0437, -0.0039, 0.2353, -0.0437, 0.0, 0.2374, -0.0526, 0.0, 0.2374, -0.0505, 0.0, 0.2386, -0.0526, 0.0, 0.239, -0.0505, 0.0, 0.2387, -0.0475, 0.0, 0.2377, -0.0457, 0.0, 0.239, -0.0485, -0.0017, 0.2386, -0.0526, -0.0017, 0.239, -0.0505, -0.0039, 0.2386, -0.0526, -0.0039, 0.239, -0.0505, -0.0017, 0.2385, -0.0472, -0.0017, 0.2379, -0.046, -0.0039, 0.2385, -0.0472, -0.0039, 0.2379, -0.046, -0.0017, 0.2388, -0.0478, -0.0039, 0.2388, -0.0478, 0.0, 0.2367, -0.0447,
-0.0017, 0.2368, -0.0447, -0.0039, 0.2368, -0.0447, 0.0, 0.2187, -0.0106, -0.0017, 0.2187, -0.0107, -0.0039, 0.2187, -0.0107, 0.0, 0.2143, -0.0022, -0.0017, 0.2144, -0.0024, -0.0039, 0.2144, -0.0024, 0.0, 0.2162, -0.0067, 0.0, 0.2178, -0.0096, 0.0, 0.2168, -0.0077, 0.0, 0.2154, -0.0052, 0.0, 0.2146, -0.0032, 0.0, 0.2179, -0.0049, 0.0, 0.2179, -0.0027, -0.0017, 0.2163, -0.0069, -0.0039, 0.2163, -0.0069, -0.0017, 0.2176, -0.0093, -0.0017, 0.2169, -0.008, -0.0039, 0.2176, -0.0093, -0.0039, 0.2169, -0.008, -0.0017, 0.2153, -0.0049, -0.0017, 0.2147, -0.0035, -0.0039, 0.2153, -0.0049, -0.0039, 0.2147, -0.0035, 0.0, 0.2231, -0.0149, 0.0, 0.2207, -0.0129, 0.0, 0.2196, -0.0118, 0.0, 0.2219, -0.014, 0.0, 0.2222, -0.0114,
-0.0017, 0.223, -0.0149, -0.0039, 0.223, -0.0149, -0.0017, 0.2207, -0.0129, -0.0017, 0.2196, -0.0118, -0.0039, 0.2207, -0.0129, -0.0039, 0.2196, -0.0118, -0.0017, 0.2219, -0.014, -0.0039, 0.2219, -0.014, 0.0, 0.2241, -0.0156, 0.0, 0.227, -0.0173, 0.0, 0.226, -0.0168, 0.0, 0.2244, -0.0136, 0.0, 0.2244, -0.0114, 0.0, 0.2266, -0.0136, 0.0, 0.2266, -0.0114, -0.0017, 0.2244, -0.0158, -0.0039, 0.2244, -0.0158, -0.0017, 0.2269, -0.0172, -0.0017, 0.2258, -0.0167, -0.0039, 0.2269, -0.0172, -0.0039, 0.2258, -0.0167, 0.0, 0.2201, -0.0092, 0.0, 0.2201, -0.0071, 0.0, 0.2222, -0.0092, 0.0, 0.2222, -0.0071, 0.0, 0.2201, -0.0049, 0.0, 0.2201, -0.0027, 0.0, 0.2222, -0.0049, 0.0, 0.2222, -0.0027, 0.0, 0.2244, -0.0092,
0.0, 0.2244, -0.0071, 0.0, 0.2266, -0.0092, 0.0, 0.2266, -0.0071, 0.0, 0.2244, -0.0049, 0.0, 0.2244, -0.0027, 0.0, 0.2266, -0.0049, 0.0, 0.2266, -0.0027, 0.0, 0.2318, -0.0191, -0.0017, 0.2318, -0.0191, -0.0039, 0.2318, -0.0191, 0.0, 0.233, -0.0194, 0.0, 0.2351, -0.02, -0.0017, 0.2331, -0.0194, -0.0039, 0.2331, -0.0194, -0.0017, 0.2352, -0.02, -0.0039, 0.2352, -0.02, 0.0, 0.2373, -0.0205, 0.0, 0.2395, -0.0208, -0.0017, 0.2374, -0.0205, -0.0039, 0.2374, -0.0205, -0.0017, 0.2396, -0.0208, -0.0039, 0.2396, -0.0208, 0.0, 0.2416, -0.021, 0.0, 0.2443, -0.0214, 0.0, 0.2436, -0.0213, -0.0017, 0.2418, -0.021, -0.0039, 0.2418, -0.021, -0.0017, 0.244, -0.0213, -0.0017, 0.2429, -0.0212, -0.0039, 0.244, -0.0213,
-0.0039, 0.2429, -0.0212, 0.0, 0.2285, -0.0179, 0.0, 0.2287, -0.0157, 0.0, 0.2308, -0.0187, 0.0, 0.2309, -0.0157, 0.0, 0.2287, -0.0136, 0.0, 0.2287, -0.0114, 0.0, 0.2309, -0.0136, 0.0, 0.2309, -0.0114, -0.0017, 0.2287, -0.018, -0.0039, 0.2287, -0.018, -0.0017, 0.2307, -0.0187, -0.0039, 0.2307, -0.0187, 0.0, 0.2331, -0.0179, 0.0, 0.2331, -0.0157, 0.0, 0.2353, -0.0179, 0.0, 0.2353, -0.0157, 0.0, 0.2331, -0.0136, 0.0, 0.2331, -0.0114, 0.0, 0.2353, -0.0136, 0.0, 0.2353, -0.0114, 0.0, 0.2287, -0.0092, 0.0, 0.2287, -0.0071, 0.0, 0.2309, -0.0092, 0.0, 0.2309, -0.0071, 0.0, 0.2287, -0.0049, 0.0, 0.2289, -0.0026, 0.0, 0.231, -0.0047, 0.0, 0.23, -0.0036, -0.0017, 0.2289, -0.0026,
-0.0039, 0.2289, -0.0026, -0.0017, 0.231, -0.0047, -0.0017, 0.23, -0.0036, -0.0039, 0.231, -0.0047, -0.0039, 0.23, -0.0036, 0.0, 0.2331, -0.0092, 0.0, 0.2334, -0.0063, 0.0, 0.2353, -0.0092, 0.0, 0.2354, -0.0075, 0.0, 0.2323, -0.0057, -0.0017, 0.2335, -0.0064, -0.0039, 0.2335, -0.0064, -0.0017, 0.2352, -0.0074, -0.0039, 0.2352, -0.0074, -0.0017, 0.2324, -0.0057, -0.0039, 0.2324, -0.0057, 0.0, 0.2374, -0.0179, 0.0, 0.2374, -0.0157, 0.0, 0.2396, -0.0179, 0.0, 0.2396, -0.0157, 0.0, 0.2374, -0.0136, 0.0, 0.2374, -0.0114, 0.0, 0.2396, -0.0136, 0.0, 0.2396, -0.0114, 0.0, 0.2418, -0.0179, 0.0, 0.2418, -0.0157, 0.0, 0.2439, -0.0179, 0.0, 0.2439, -0.0157, 0.0, 0.2418, -0.0136, 0.0, 0.2418, -0.0114,
0.0, 0.2439, -0.0136, 0.0, 0.2439, -0.0114, 0.0, 0.2377, -0.0081, 0.0, 0.2369, -0.0079, 0.0, 0.2398, -0.0086, -0.0017, 0.2381, -0.0082, -0.0017, 0.2371, -0.0079, -0.0039, 0.2381, -0.0082, -0.0039, 0.2371, -0.0079, -0.0017, 0.2396, -0.0086, -0.0039, 0.2396, -0.0086, 0.0, 0.2419, -0.009, 0.0, 0.244, -0.0093, -0.0017, 0.2418, -0.009, -0.0039, 0.2418, -0.009, -0.0017, 0.2439, -0.0093, -0.0039, 0.2439, -0.0093, 0.0, 0.2588, -0.067, 0.0, 0.2613, -0.067, 0.0, 0.2588, -0.0656, 0.0, 0.2588, -0.0635, 0.0, 0.2613, -0.0656, 0.0, 0.2618, -0.0631, -0.0017, 0.2588, -0.067, -0.0039, 0.2588, -0.067, -0.0017, 0.2613, -0.067, -0.0039, 0.2613, -0.067, -0.0017, 0.2588, -0.0656, -0.0017, 0.2588, -0.0635, -0.0039, 0.2588, -0.0656,
-0.0039, 0.2588, -0.0635, -0.0017, 0.2618, -0.0631, -0.0039, 0.2618, -0.0631, 0.0, 0.2588, -0.0613, 0.0, 0.2588, -0.0591, 0.0, 0.2618, -0.0613, 0.0, 0.2618, -0.0591, 0.0, 0.2588, -0.057, 0.0, 0.2588, -0.0548, 0.0, 0.2618, -0.057, 0.0, 0.2618, -0.0548, -0.0017, 0.2588, -0.0613, -0.0017, 0.2588, -0.0591, -0.0039, 0.2588, -0.0613, -0.0039, 0.2588, -0.0591, -0.0017, 0.2618, -0.0613, -0.0017, 0.2618, -0.0591, -0.0039, 0.2618, -0.0613, -0.0039, 0.2618, -0.0591, -0.0017, 0.2588, -0.057, -0.0017, 0.2588, -0.0548, -0.0039, 0.2588, -0.057, -0.0039, 0.2588, -0.0548, -0.0017, 0.2618, -0.057, -0.0017, 0.2618, -0.0548, -0.0039, 0.2618, -0.057, -0.0039, 0.2618, -0.0548, 0.0, 0.2588, -0.0526, 0.0, 0.2588, -0.0505, 0.0, 0.2618, -0.0526,
0.0, 0.2618, -0.0505, 0.0, 0.2588, -0.0483, 0.0, 0.2588, -0.0461, 0.0, 0.2618, -0.0483, 0.0, 0.2618, -0.0461, -0.0017, 0.2588, -0.0526, -0.0017, 0.2588, -0.0505, -0.0039, 0.2588, -0.0526, -0.0039, 0.2588, -0.0505, -0.0017, 0.2618, -0.0526, -0.0017, 0.2618, -0.0505, -0.0039, 0.2618, -0.0526, -0.0039, 0.2618, -0.0505, -0.0017, 0.2588, -0.0483, -0.0017, 0.2588, -0.0461, -0.0039, 0.2588, -0.0483, -0.0039, 0.2588, -0.0461, -0.0017, 0.2618, -0.0483, -0.0017, 0.2618, -0.0461, -0.0039, 0.2618, -0.0483, -0.0039, 0.2618, -0.0461, 0.0, 0.2588, -0.0439, 0.0, 0.2588, -0.0425, 0.0, 0.2618, -0.0439, 0.0, 0.2618, -0.0425, -0.0017, 0.2588, -0.0439, -0.0017, 0.2588, -0.0425, -0.0039, 0.2588, -0.0439, -0.0039, 0.2588, -0.0425, -0.0017, 0.2618, -0.0439,
-0.0017, 0.2618, -0.0425, -0.0039, 0.2618, -0.0439, -0.0039, 0.2618, -0.0425, 0.0, 0.2635, -0.067, 0.0, 0.2656, -0.067, 0.0, 0.2635, -0.0656, 0.0, 0.2635, -0.0631, 0.0, 0.2656, -0.0656, 0.0, 0.2656, -0.0631, -0.0017, 0.2635, -0.067, -0.0039, 0.2635, -0.067, -0.0017, 0.2656, -0.067, -0.0039, 0.2656, -0.067, -0.0017, 0.2635, -0.0631, -0.0039, 0.2635, -0.0631, -0.0017, 0.2656, -0.0631, -0.0039, 0.2656, -0.0631, 0.0, 0.2678, -0.067, 0.0, 0.27, -0.067, 0.0, 0.2678, -0.0656, 0.0, 0.2678, -0.0631, 0.0, 0.27, -0.0656, 0.0, 0.27, -0.0631, -0.0017, 0.2678, -0.067, -0.0039, 0.2678, -0.067, -0.0017, 0.27, -0.067, -0.0039, 0.27, -0.067, -0.0017, 0.2678, -0.0631, -0.0039, 0.2678, -0.0631, -0.0017, 0.27, -0.0631,
-0.0039, 0.27, -0.0631, 0.0, 0.2721, -0.067, 0.0, 0.2743, -0.067, 0.0, 0.2721, -0.0656, 0.0, 0.2721, -0.0631, 0.0, 0.2743, -0.0656, 0.0, 0.2743, -0.0631, -0.0017, 0.2721, -0.067, -0.0039, 0.2721, -0.067, -0.0017, 0.2743, -0.067, -0.0039, 0.2743, -0.067, -0.0017, 0.2721, -0.0631, -0.0039, 0.2721, -0.0631, -0.0017, 0.2743, -0.0631, -0.0039, 0.2743, -0.0631, 0.0, 0.2765, -0.067, 0.0, 0.2794, -0.067, 0.0, 0.2765, -0.0656, 0.0, 0.2765, -0.0631, 0.0, 0.2794, -0.0656, 0.0, 0.2794, -0.0631, -0.0017, 0.2765, -0.067, -0.0039, 0.2765, -0.067, -0.0017, 0.2794, -0.067, -0.0039, 0.2794, -0.067, -0.0017, 0.2765, -0.0631, -0.0039, 0.2765, -0.0631, -0.0017, 0.2794, -0.0656, -0.0017, 0.2794, -0.0631, -0.0039, 0.2794, -0.0656,
-0.0039, 0.2794, -0.0631, 0.0, 0.2461, -0.0216, 0.0, 0.2461, -0.0201, 0.0, 0.2482, -0.0218, 0.0, 0.2483, -0.0201, -0.0017, 0.2461, -0.0216, -0.0039, 0.2461, -0.0216, -0.0017, 0.2483, -0.0218, -0.0039, 0.2483, -0.0218, 0.0, 0.2504, -0.0219, 0.0, 0.2504, -0.0201, 0.0, 0.2526, -0.0219, 0.0, 0.2526, -0.0201, -0.0017, 0.2504, -0.0219, -0.0039, 0.2504, -0.0219, -0.0017, 0.2526, -0.0219, -0.0039, 0.2526, -0.0219, 0.0, 0.2548, -0.0219, 0.0, 0.2548, -0.0201, 0.0, 0.257, -0.0217, 0.0, 0.257, -0.0201, -0.0017, 0.2548, -0.0219, -0.0039, 0.2548, -0.0219, -0.0017, 0.257, -0.0217, -0.0039, 0.257, -0.0217, 0.0, 0.2591, -0.0214, 0.0, 0.2591, -0.0201, 0.0, 0.2605, -0.0212, 0.0, 0.2613, -0.0211, -0.0017, 0.2591, -0.0214,
-0.0039, 0.2591, -0.0214, -0.0017, 0.2606, -0.0212, -0.0017, 0.2617, -0.0211, -0.0039, 0.2606, -0.0212, -0.0039, 0.2617, -0.0211, 0.0, 0.2461, -0.0179, 0.0, 0.2461, -0.0157, 0.0, 0.2483, -0.0179, 0.0, 0.2483, -0.0157, 0.0, 0.2461, -0.0136, 0.0, 0.2461, -0.0114, 0.0, 0.2483, -0.0136, 0.0, 0.2483, -0.0114, 0.0, 0.2504, -0.0179, 0.0, 0.2504, -0.0157, 0.0, 0.2526, -0.0179, 0.0, 0.2526, -0.0157, 0.0, 0.2504, -0.0136, 0.0, 0.2504, -0.0114, 0.0, 0.2526, -0.0136, 0.0, 0.2526, -0.0114, 0.0, 0.2462, -0.0096, 0.0, 0.2483, -0.0099, -0.0017, 0.2461, -0.0096, -0.0039, 0.2461, -0.0096, -0.0017, 0.2483, -0.0099, -0.0039, 0.2483, -0.0099, 0.0, 0.2504, -0.0101, 0.0, 0.2526, -0.0101, -0.0017, 0.2504, -0.0101,
-0.0039, 0.2504, -0.0101, -0.0017, 0.2526, -0.0101, -0.0039, 0.2526, -0.0101, 0.0, 0.2548, -0.0179, 0.0, 0.2548, -0.0157, 0.0, 0.257, -0.0179, 0.0, 0.257, -0.0157, 0.0, 0.2548, -0.0136, 0.0, 0.2548, -0.0114, 0.0, 0.257, -0.0136, 0.0, 0.257, -0.0114, 0.0, 0.2591, -0.0179, 0.0, 0.2591, -0.0157, 0.0, 0.2613, -0.0179, 0.0, 0.2613, -0.0157, 0.0, 0.2591, -0.0136, 0.0, 0.2591, -0.0114, 0.0, 0.2613, -0.0136, 0.0, 0.2613, -0.0114, 0.0, 0.2548, -0.0101, 0.0, 0.2569, -0.01, -0.0017, 0.2548, -0.0101, -0.0039, 0.2548, -0.0101, -0.0017, 0.257, -0.01, -0.0039, 0.257, -0.01, 0.0, 0.2591, -0.0097, 0.0, 0.2612, -0.0093, -0.0017, 0.2591, -0.0097, -0.0039, 0.2591, -0.0097, -0.0017, 0.2613, -0.0093,
-0.0039, 0.2613, -0.0093, 0.0, 0.2636, -0.0209, 0.0, 0.2658, -0.0206, -0.0017, 0.2635, -0.0209, -0.0039, 0.2635, -0.0209, -0.0017, 0.2656, -0.0206, -0.0039, 0.2656, -0.0206, 0.0, 0.2679, -0.0202, 0.0, 0.2701, -0.0197, -0.0017, 0.2678, -0.0202, -0.0039, 0.2678, -0.0202, -0.0017, 0.27, -0.0197, -0.0039, 0.27, -0.0197, 0.0, 0.2715, -0.0192, -0.0017, 0.2717, -0.0192, -0.0039, 0.2717, -0.0192, 0.0, 0.2635, -0.0179, 0.0, 0.2635, -0.0157, 0.0, 0.2656, -0.0179, 0.0, 0.2656, -0.0157, 0.0, 0.2635, -0.0136, 0.0, 0.2635, -0.0114, 0.0, 0.2656, -0.0136, 0.0, 0.2656, -0.0114, 0.0, 0.2678, -0.0179, 0.0, 0.2678, -0.0157, 0.0, 0.27, -0.0179, 0.0, 0.27, -0.0157, 0.0, 0.2678, -0.0136, 0.0, 0.2678, -0.0114,
0.0, 0.27, -0.0136, 0.0, 0.27, -0.0114, 0.0, 0.2633, -0.009, 0.0, 0.2656, -0.0084, 0.0, 0.2666, -0.0081, -0.0017, 0.2635, -0.0089, -0.0039, 0.2635, -0.0089, -0.0017, 0.2655, -0.0085, -0.0017, 0.2666, -0.0081, -0.0039, 0.2655, -0.0085, -0.0039, 0.2666, -0.0081, 0.0, 0.2678, -0.0092, 0.0, 0.2677, -0.0077, 0.0, 0.27, -0.0092, 0.0, 0.2699, -0.0066, 0.0, 0.271, -0.0059, -0.0017, 0.2678, -0.0076, -0.0039, 0.2678, -0.0076, -0.0017, 0.27, -0.0066, -0.0039, 0.27, -0.0066, -0.0017, 0.271, -0.0059, -0.0039, 0.271, -0.0059, 0.0, 0.2725, -0.019, 0.0, 0.2721, -0.0157, 0.0, 0.2746, -0.0182, 0.0, 0.2743, -0.0157, 0.0, 0.2721, -0.0136, 0.0, 0.2721, -0.0114, 0.0, 0.2743, -0.0136, 0.0, 0.2743, -0.0114,
-0.0017, 0.2728, -0.0189, -0.0039, 0.2728, -0.0189, -0.0017, 0.2743, -0.0183, -0.0039, 0.2743, -0.0183, 0.0, 0.2766, -0.0174, 0.0, 0.2765, -0.0157, 0.0, 0.2776, -0.0169, 0.0, 0.2788, -0.0163, 0.0, 0.2765, -0.0136, 0.0, 0.2765, -0.0114, 0.0, 0.2787, -0.0136, 0.0, 0.2787, -0.0114, -0.0017, 0.2765, -0.0175, -0.0039, 0.2765, -0.0175, -0.0017, 0.2776, -0.0169, -0.0017, 0.2788, -0.0163, -0.0039, 0.2776, -0.0169, -0.0039, 0.2788, -0.0163, 0.0, 0.2721, -0.0092, 0.0, 0.2721, -0.0071, 0.0, 0.2743, -0.0092, 0.0, 0.2743, -0.0071, 0.0, 0.272, -0.0053, 0.0, 0.2735, -0.004, 0.0, 0.2746, -0.0029, -0.0017, 0.2722, -0.0051, -0.0039, 0.2722, -0.0051, -0.0017, 0.2735, -0.004, -0.0017, 0.2746, -0.0029, -0.0039, 0.2735, -0.004,
-0.0039, 0.2746, -0.0029, 0.0, 0.2765, -0.0092, 0.0, 0.2765, -0.0071, 0.0, 0.2787, -0.0092, 0.0, 0.2787, -0.0071, 0.0, 0.2765, -0.0049, 0.0, 0.2754, -0.0019, 0.0, 0.2787, -0.0049, 0.0, 0.2787, -0.0027, -0.0017, 0.2755, -0.0017, -0.0039, 0.2755, -0.0017, 0.0, 0.298, -0.067, 0.0, 0.3004, -0.067, 0.0, 0.298, -0.0656, 0.0, 0.298, -0.0635, 0.0, 0.3004, -0.0656, 0.0, 0.3004, -0.0635, -0.0017, 0.298, -0.067, -0.0039, 0.298, -0.067, -0.0017, 0.3004, -0.067, -0.0039, 0.3004, -0.067, -0.0017, 0.298, -0.0656, -0.0017, 0.298, -0.0635, -0.0039, 0.298, -0.0656, -0.0039, 0.298, -0.0635, 0.0, 0.3025, -0.067, 0.0, 0.3047, -0.067, 0.0, 0.3025, -0.0656, 0.0, 0.302, -0.0631, 0.0, 0.3047, -0.0656,
0.0, 0.3047, -0.0631, -0.0017, 0.3025, -0.067, -0.0039, 0.3025, -0.067, -0.0017, 0.3047, -0.067, -0.0039, 0.3047, -0.067, -0.0017, 0.302, -0.0631, -0.0039, 0.302, -0.0631, -0.0017, 0.3047, -0.0631, -0.0039, 0.3047, -0.0631, 0.0, 0.298, -0.0613, 0.0, 0.298, -0.0591, 0.0, 0.3004, -0.0613, 0.0, 0.3004, -0.0591, 0.0, 0.298, -0.057, 0.0, 0.298, -0.0548, 0.0, 0.3004, -0.057, 0.0, 0.3004, -0.0548, -0.0017, 0.298, -0.0613, -0.0017, 0.298, -0.0591, -0.0039, 0.298, -0.0613, -0.0039, 0.298, -0.0591, -0.0017, 0.298, -0.057, -0.0017, 0.298, -0.0548, -0.0039, 0.298, -0.057, -0.0039, 0.298, -0.0548, 0.0, 0.302, -0.0613, 0.0, 0.302, -0.0591, 0.0, 0.302, -0.057, 0.0, 0.302, -0.0548, 0.0, 0.3049, -0.0562,
0.0, 0.3049, -0.0548, -0.0017, 0.302, -0.0613, -0.0017, 0.302, -0.0591, -0.0039, 0.302, -0.0613, -0.0039, 0.302, -0.0591, -0.0017, 0.302, -0.057, -0.0017, 0.302, -0.0548, -0.0039, 0.302, -0.057, -0.0039, 0.302, -0.0548, -0.0017, 0.3049, -0.0562, -0.0017, 0.3049, -0.0548, -0.0039, 0.3049, -0.0562, -0.0039, 0.3049, -0.0548, 0.0, 0.3069, -0.067, 0.0, 0.309, -0.067, 0.0, 0.3069, -0.0656, 0.0, 0.3069, -0.0631, 0.0, 0.309, -0.0656, 0.0, 0.309, -0.0631, -0.0017, 0.3069, -0.067, -0.0039, 0.3069, -0.067, -0.0017, 0.309, -0.067, -0.0039, 0.309, -0.067, -0.0017, 0.3069, -0.0631, -0.0039, 0.3069, -0.0631, -0.0017, 0.309, -0.0631, -0.0039, 0.309, -0.0631, 0.0, 0.3112, -0.067, 0.0, 0.3134, -0.067, 0.0, 0.3112, -0.0656,
0.0, 0.3112, -0.0631, 0.0, 0.3134, -0.0656, 0.0, 0.3134, -0.0631, -0.0017, 0.3112, -0.067, -0.0039, 0.3112, -0.067, -0.0017, 0.3134, -0.067, -0.0039, 0.3134, -0.067, -0.0017, 0.3112, -0.0631, -0.0039, 0.3112, -0.0631, -0.0017, 0.3134, -0.0631, -0.0039, 0.3134, -0.0631, 0.0, 0.3069, -0.0562, 0.0, 0.3069, -0.0548, 0.0, 0.309, -0.0562, 0.0, 0.309, -0.0548, -0.0017, 0.3069, -0.0562, -0.0039, 0.3069, -0.0562, -0.0017, 0.309, -0.0562, -0.0039, 0.309, -0.0562, 0.0, 0.3112, -0.0562, 0.0, 0.3112, -0.0548, 0.0, 0.3134, -0.0562, 0.0, 0.3134, -0.0548, -0.0017, 0.3112, -0.0562, -0.0039, 0.3112, -0.0562, -0.0017, 0.3134, -0.0562, -0.0039, 0.3134, -0.0562, 0.0, 0.298, -0.0526, 0.0, 0.298, -0.0505, 0.0, 0.3004, -0.0526,
0.0, 0.3004, -0.0505, 0.0, 0.298, -0.0483, 0.0, 0.298, -0.0461, 0.0, 0.3004, -0.0483, 0.0, 0.3004, -0.0461, -0.0017, 0.298, -0.0526, -0.0017, 0.298, -0.0505, -0.0039, 0.298, -0.0526, -0.0039, 0.298, -0.0505, -0.0017, 0.298, -0.0483, -0.0017, 0.298, -0.0461, -0.0039, 0.298, -0.0483, -0.0039, 0.298, -0.0461, 0.0, 0.302, -0.0526, 0.0, 0.302, -0.0505, 0.0, 0.3049, -0.0523, 0.0, 0.302, -0.0483, 0.0, 0.302, -0.0454, 0.0, 0.3047, -0.0454, -0.0017, 0.302, -0.0526, -0.0017, 0.302, -0.0505, -0.0039, 0.302, -0.0526, -0.0039, 0.302, -0.0505, -0.0017, 0.3049, -0.0523, -0.0039, 0.3049, -0.0523, -0.0017, 0.302, -0.0483, -0.0017, 0.302, -0.0454, -0.0039, 0.302, -0.0483, -0.0039, 0.302, -0.0454, -0.0017, 0.3047, -0.0454,
-0.0039, 0.3047, -0.0454, 0.0, 0.298, -0.0439, 0.0, 0.298, -0.0425, 0.0, 0.3004, -0.0439, 0.0, 0.3004, -0.0425, -0.0017, 0.298, -0.0439, -0.0017, 0.298, -0.0425, -0.0039, 0.298, -0.0439, -0.0039, 0.298, -0.0425, -0.0017, 0.3004, -0.0425, -0.0039, 0.3004, -0.0425, 0.0, 0.3025, -0.0439, 0.0, 0.3025, -0.0425, 0.0, 0.3047, -0.0439, 0.0, 0.3047, -0.0425, -0.0017, 0.3025, -0.0425, -0.0039, 0.3025, -0.0425, -0.0017, 0.3047, -0.0425, -0.0039, 0.3047, -0.0425, 0.0, 0.3069, -0.0523, 0.0, 0.309, -0.0523, 0.0, 0.3069, -0.0454, 0.0, 0.309, -0.0454, -0.0017, 0.3069, -0.0523, -0.0039, 0.3069, -0.0523, -0.0017, 0.309, -0.0523, -0.0039, 0.309, -0.0523, -0.0017, 0.3069, -0.0454, -0.0039, 0.3069, -0.0454, -0.0017, 0.309, -0.0454,
-0.0039, 0.309, -0.0454, 0.0, 0.3112, -0.0523, 0.0, 0.3134, -0.0523, 0.0, 0.3112, -0.0454, 0.0, 0.3134, -0.0454, -0.0017, 0.3112, -0.0523, -0.0039, 0.3112, -0.0523, -0.0017, 0.3134, -0.0523, -0.0039, 0.3134, -0.0523, -0.0017, 0.3112, -0.0454, -0.0039, 0.3112, -0.0454, -0.0017, 0.3134, -0.0454, -0.0039, 0.3134, -0.0454, 0.0, 0.3069, -0.0439, 0.0, 0.3069, -0.0425, 0.0, 0.309, -0.0439, 0.0, 0.309, -0.0425, -0.0017, 0.3069, -0.0425, -0.0039, 0.3069, -0.0425, -0.0017, 0.309, -0.0425, -0.0039, 0.309, -0.0425, 0.0, 0.3112, -0.0439, 0.0, 0.3112, -0.0425, 0.0, 0.3134, -0.0439, 0.0, 0.3134, -0.0425, -0.0017, 0.3112, -0.0425, -0.0039, 0.3112, -0.0425, -0.0017, 0.3134, -0.0425, -0.0039, 0.3134, -0.0425, 0.0, 0.2804, -0.0152,
0.0, 0.2815, -0.0144, 0.0, 0.2808, -0.0114, 0.0, 0.2827, -0.0133, 0.0, 0.2838, -0.0122, -0.0017, 0.2805, -0.0152, -0.0039, 0.2805, -0.0152, -0.0017, 0.2815, -0.0143, -0.0039, 0.2815, -0.0143, -0.0017, 0.2828, -0.0133, -0.0017, 0.2838, -0.0122, -0.0039, 0.2828, -0.0133, -0.0039, 0.2838, -0.0122, 0.0, 0.2848, -0.0111, -0.0017, 0.2847, -0.0111, -0.0039, 0.2847, -0.0111, 0.0, 0.2808, -0.0092, 0.0, 0.2808, -0.0071, 0.0, 0.283, -0.0092, 0.0, 0.283, -0.0071, 0.0, 0.2808, -0.0049, 0.0, 0.2808, -0.0027, 0.0, 0.283, -0.0049, 0.0, 0.283, -0.0027, 0.0, 0.2858, -0.0096, 0.0, 0.2852, -0.0071, 0.0, 0.2866, -0.0085, 0.0, 0.2873, -0.0073, 0.0, 0.2852, -0.0049, 0.0, 0.2852, -0.0027, 0.0, 0.2883, -0.0052,
0.0, 0.2873, -0.0027, -0.0017, 0.2858, -0.0097, -0.0039, 0.2858, -0.0097, -0.0017, 0.2865, -0.0086, -0.0017, 0.2874, -0.0071, -0.0039, 0.2865, -0.0086, -0.0039, 0.2874, -0.0071, -0.0017, 0.2882, -0.0054, -0.0039, 0.2882, -0.0054, 0.0, 0.2887, -0.0042, 0.0, 0.2891, -0.0028, -0.0017, 0.2886, -0.0043, -0.0017, 0.2892, -0.0027, -0.0039, 0.2886, -0.0043, -0.0039, 0.2892, -0.0027, 0.0, 0.3088, -0.0219, 0.0, 0.3088, -0.0201, -0.0017, 0.3088, -0.0219, -0.0017, 0.3088, -0.0201, -0.0039, 0.3088, -0.0219, -0.0039, 0.3088, -0.0201, 0.0, 0.3112, -0.0219, 0.0, 0.3112, -0.0201, 0.0, 0.3134, -0.0219, 0.0, 0.3134, -0.0201, -0.0017, 0.3112, -0.0219, -0.0039, 0.3112, -0.0219, -0.0017, 0.3134, -0.0219, -0.0039, 0.3134, -0.0219, 0.0, 0.3088, -0.0179,
0.0, 0.3088, -0.0157, 0.0, 0.3088, -0.0136, 0.0, 0.3088, -0.0114, -0.0017, 0.3088, -0.0179, -0.0017, 0.3088, -0.0157, -0.0039, 0.3088, -0.0179, -0.0039, 0.3088, -0.0157, -0.0017, 0.3088, -0.0136, -0.0017, 0.3088, -0.0114, -0.0039, 0.3088, -0.0136, -0.0039, 0.3088, -0.0114, 0.0, 0.3112, -0.0179, 0.0, 0.3112, -0.0157, 0.0, 0.3134, -0.0179, 0.0, 0.3134, -0.0157, 0.0, 0.3112, -0.0136, 0.0, 0.3112, -0.0114, 0.0, 0.3134, -0.0136, 0.0, 0.3134, -0.0114, 0.0, 0.3088, -0.0092, 0.0, 0.3088, -0.0071, 0.0, 0.3088, -0.0049, 0.0, 0.3088, -0.0027, -0.0017, 0.3088, -0.0092, -0.0017, 0.3088, -0.0071, -0.0039, 0.3088, -0.0092, -0.0039, 0.3088, -0.0071, -0.0017, 0.3088, -0.0049, -0.0017, 0.3088, -0.0027, -0.0039, 0.3088, -0.0049,
-0.0039, 0.3088, -0.0027, 0.0, 0.3112, -0.0092, 0.0, 0.3112, -0.0071, 0.0, 0.3134, -0.0092, 0.0, 0.3134, -0.0071, 0.0, 0.3112, -0.0049, 0.0, 0.3112, -0.0027, 0.0, 0.3134, -0.0049, 0.0, 0.3134, -0.0027, 0.0, 0.3155, -0.067, 0.0, 0.3177, -0.067, 0.0, 0.3155, -0.0656, 0.0, 0.3155, -0.0631, 0.0, 0.3177, -0.0656, 0.0, 0.3177, -0.0631, -0.0017, 0.3155, -0.067, -0.0039, 0.3155, -0.067, -0.0017, 0.3177, -0.067, -0.0039, 0.3177, -0.067, -0.0017, 0.3155, -0.0631, -0.0039, 0.3155, -0.0631, -0.0017, 0.3177, -0.0631, -0.0039, 0.3177, -0.0631, 0.0, 0.3196, -0.067, 0.0, 0.3196, -0.0656, 0.0, 0.3196, -0.0631, -0.0017, 0.3196, -0.067, -0.0039, 0.3196, -0.067, -0.0017, 0.3196, -0.0656, -0.0017, 0.3196, -0.0631,
-0.0039, 0.3196, -0.0656, -0.0039, 0.3196, -0.0631, 0.0, 0.3155, -0.0562, 0.0, 0.3155, -0.0548, 0.0, 0.3176, -0.0562, 0.0, 0.3176, -0.0548, -0.0017, 0.3155, -0.0562, -0.0039, 0.3155, -0.0562, -0.0017, 0.3176, -0.0562, -0.0017, 0.3176, -0.0548, -0.0039, 0.3176, -0.0562, -0.0039, 0.3176, -0.0548, 0.0, 0.3155, -0.0523, 0.0, 0.3176, -0.0523, 0.0, 0.3155, -0.0454, 0.0, 0.3177, -0.0454, -0.0017, 0.3155, -0.0523, -0.0039, 0.3155, -0.0523, -0.0017, 0.3176, -0.0523, -0.0039, 0.3176, -0.0523, -0.0017, 0.3155, -0.0454, -0.0039, 0.3155, -0.0454, -0.0017, 0.3177, -0.0454, -0.0039, 0.3177, -0.0454, 0.0, 0.3196, -0.0454, -0.0017, 0.3196, -0.0454, -0.0039, 0.3196, -0.0454, 0.0, 0.3155, -0.0439, 0.0, 0.3155, -0.0425, 0.0, 0.3177, -0.0439,
0.0, 0.3177, -0.0425, -0.0017, 0.3155, -0.0425, -0.0039, 0.3155, -0.0425, -0.0017, 0.3177, -0.0425, -0.0039, 0.3177, -0.0425, 0.0, 0.3196, -0.0439, 0.0, 0.3196, -0.0425, -0.0017, 0.3196, -0.0439, -0.0017, 0.3196, -0.0425, -0.0039, 0.3196, -0.0439, -0.0039, 0.3196, -0.0425, 0.0, 0.3402, -0.067, 0.0, 0.3402, -0.0656, 0.0, 0.3402, -0.0635, -0.0017, 0.3402, -0.067, -0.0039, 0.3402, -0.067, -0.0017, 0.3402, -0.0656, -0.0017, 0.3402, -0.0635, -0.0039, 0.3402, -0.0656, -0.0039, 0.3402, -0.0635, 0.0, 0.3402, -0.0613, 0.0, 0.3402, -0.0591, 0.0, 0.3402, -0.057, 0.0, 0.3402, -0.0548, -0.0017, 0.3402, -0.0613, -0.0017, 0.3402, -0.0591, -0.0039, 0.3402, -0.0613, -0.0039, 0.3402, -0.0591, -0.0017, 0.3402, -0.057, -0.0017, 0.3402, -0.0548,
-0.0039, 0.3402, -0.057, -0.0039, 0.3402, -0.0548, 0.0, 0.3416, -0.067, 0.0, 0.3431, -0.067, 0.0, 0.3416, -0.0656, 0.0, 0.3416, -0.0635, 0.0, 0.3431, -0.0656, 0.0, 0.3431, -0.0635, -0.0017, 0.3416, -0.067, -0.0039, 0.3416, -0.067, -0.0017, 0.3431, -0.067, -0.0039, 0.3431, -0.067, -0.0017, 0.3431, -0.0656, -0.0017, 0.3431, -0.0635, -0.0039, 0.3431, -0.0656, -0.0039, 0.3431, -0.0635, 0.0, 0.3416, -0.0613, 0.0, 0.3416, -0.0591, 0.0, 0.3431, -0.0613, 0.0, 0.3431, -0.0591, 0.0, 0.3416, -0.057, 0.0, 0.3416, -0.0548, 0.0, 0.3431, -0.057, 0.0, 0.3431, -0.0548, -0.0017, 0.3431, -0.0613, -0.0017, 0.3431, -0.0591, -0.0039, 0.3431, -0.0613, -0.0039, 0.3431, -0.0591, -0.0017, 0.3431, -0.057, -0.0017, 0.3431, -0.0548,
-0.0039, 0.3431, -0.057, -0.0039, 0.3431, -0.0548, 0.0, 0.3483, -0.0546, -0.0017, 0.3483, -0.0545, -0.0039, 0.3483, -0.0545, 0.0, 0.3402, -0.0526, 0.0, 0.3402, -0.0505, 0.0, 0.3402, -0.0483, 0.0, 0.3402, -0.0461, -0.0017, 0.3402, -0.0526, -0.0017, 0.3402, -0.0505, -0.0039, 0.3402, -0.0526, -0.0039, 0.3402, -0.0505, -0.0017, 0.3402, -0.0483, -0.0017, 0.3402, -0.0461, -0.0039, 0.3402, -0.0483, -0.0039, 0.3402, -0.0461, 0.0, 0.3402, -0.0439, 0.0, 0.3402, -0.0425, -0.0017, 0.3402, -0.0439, -0.0017, 0.3402, -0.0425, -0.0039, 0.3402, -0.0439, -0.0039, 0.3402, -0.0425, 0.0, 0.3416, -0.0526, 0.0, 0.3416, -0.0505, 0.0, 0.3431, -0.0526, 0.0, 0.3431, -0.0499, 0.0, 0.3416, -0.0483, 0.0, 0.3416, -0.0461, 0.0, 0.3438, -0.0483,
0.0, 0.3438, -0.0461, -0.0017, 0.3431, -0.0526, -0.0017, 0.3431, -0.0499, -0.0039, 0.3431, -0.0526, -0.0039, 0.3431, -0.0499, 0.0, 0.346, -0.0525, 0.0, 0.345, -0.0515, 0.0, 0.3472, -0.0535, 0.0, 0.3481, -0.0505, 0.0, 0.3459, -0.0483, 0.0, 0.3462, -0.0457, 0.0, 0.3485, -0.0479, 0.0, 0.3474, -0.0468, -0.0017, 0.346, -0.0524, -0.0017, 0.3449, -0.0515, -0.0039, 0.346, -0.0524, -0.0039, 0.3449, -0.0515, -0.0017, 0.3472, -0.0535, -0.0039, 0.3472, -0.0535, -0.0017, 0.3462, -0.0457, -0.0039, 0.3462, -0.0457, -0.0017, 0.3485, -0.0479, -0.0017, 0.3474, -0.0468, -0.0039, 0.3485, -0.0479, -0.0039, 0.3474, -0.0468, 0.0, 0.3421, -0.0429, 0.0, 0.3411, -0.0426, 0.0, 0.3441, -0.0439, -0.0017, 0.3423, -0.043, -0.0017, 0.3413, -0.0426,
-0.0039, 0.3423, -0.043, -0.0039, 0.3413, -0.0426, -0.0017, 0.3438, -0.0438, -0.0039, 0.3438, -0.0438, 0.0, 0.3451, -0.0448, -0.0017, 0.3452, -0.0448, -0.0039, 0.3452, -0.0448, 0.0, 0.3155, -0.0219, 0.0, 0.3155, -0.0201, 0.0, 0.3177, -0.0219, 0.0, 0.3177, -0.0201, -0.0017, 0.3155, -0.0219, -0.0039, 0.3155, -0.0219, -0.0017, 0.3177, -0.0219, -0.0039, 0.3177, -0.0219, 0.0, 0.3206, -0.0219, 0.0, 0.3206, -0.0201, -0.0017, 0.3206, -0.0219, -0.0017, 0.3206, -0.0201, -0.0039, 0.3206, -0.0219, -0.0039, 0.3206, -0.0201, 0.0, 0.3155, -0.0179, 0.0, 0.3155, -0.0157, 0.0, 0.3177, -0.0179, 0.0, 0.3177, -0.0157, 0.0, 0.3155, -0.0136, 0.0, 0.3155, -0.0114, 0.0, 0.3177, -0.0136, 0.0, 0.3177, -0.0114, 0.0, 0.3206, -0.0179,
0.0, 0.3206, -0.0157, 0.0, 0.3206, -0.0136, 0.0, 0.3206, -0.0114, -0.0017, 0.3206, -0.0179, -0.0017, 0.3206, -0.0157, -0.0039, 0.3206, -0.0179, -0.0039, 0.3206, -0.0157, -0.0017, 0.3206, -0.0136, -0.0017, 0.3206, -0.0114, -0.0039, 0.3206, -0.0136, -0.0039, 0.3206, -0.0114, 0.0, 0.3155, -0.0092, 0.0, 0.3155, -0.0071, 0.0, 0.3177, -0.0092, 0.0, 0.3177, -0.0071, 0.0, 0.3155, -0.0049, 0.0, 0.3155, -0.0027, 0.0, 0.3177, -0.0049, 0.0, 0.3177, -0.0027, 0.0, 0.3206, -0.0092, 0.0, 0.3206, -0.0071, 0.0, 0.3206, -0.0049, 0.0, 0.3206, -0.0027, -0.0017, 0.3206, -0.0092, -0.0017, 0.3206, -0.0071, -0.0039, 0.3206, -0.0092, -0.0039, 0.3206, -0.0071, -0.0017, 0.3206, -0.0049, -0.0017, 0.3206, -0.0027, -0.0039, 0.3206, -0.0049,
-0.0039, 0.3206, -0.0027, 0.0, 0.3402, -0.0219, 0.0, 0.3402, -0.0201, -0.0017, 0.3402, -0.0219, -0.0017, 0.3402, -0.0201, -0.0039, 0.3402, -0.0219, -0.0039, 0.3402, -0.0201, 0.0, 0.3416, -0.0219, 0.0, 0.3416, -0.0201, 0.0, 0.3438, -0.0219, 0.0, 0.3438, -0.0201, -0.0017, 0.3416, -0.0219, -0.0039, 0.3416, -0.0219, -0.0017, 0.3438, -0.0219, -0.0039, 0.3438, -0.0219, 0.0, 0.3459, -0.0219, 0.0, 0.3459, -0.0201, 0.0, 0.3481, -0.0219, 0.0, 0.3481, -0.0201, -0.0017, 0.3459, -0.0219, -0.0039, 0.3459, -0.0219, -0.0017, 0.3481, -0.0219, -0.0039, 0.3481, -0.0219, 0.0, 0.3402, -0.0179, 0.0, 0.3402, -0.0157, 0.0, 0.3402, -0.0136, 0.0, 0.3402, -0.0114, -0.0017, 0.3402, -0.0179, -0.0017, 0.3402, -0.0157, -0.0039, 0.3402, -0.0179,
-0.0039, 0.3402, -0.0157, -0.0017, 0.3402, -0.0136, -0.0017, 0.3402, -0.0114, -0.0039, 0.3402, -0.0136, -0.0039, 0.3402, -0.0114, 0.0, 0.3402, -0.0092, 0.0, 0.3402, -0.0071, 0.0, 0.3402, -0.0049, 0.0, 0.3402, -0.0027, -0.0017, 0.3402, -0.0092, -0.0017, 0.3402, -0.0071, -0.0039, 0.3402, -0.0092, -0.0039, 0.3402, -0.0071, -0.0017, 0.3402, -0.0049, -0.0017, 0.3402, -0.0027, -0.0039, 0.3402, -0.0049, -0.0039, 0.3402, -0.0027, 0.0, 0.3416, -0.0179, 0.0, 0.3416, -0.0157, 0.0, 0.3438, -0.0179, 0.0, 0.3438, -0.0157, 0.0, 0.3416, -0.0136, 0.0, 0.3416, -0.0114, 0.0, 0.3438, -0.0136, 0.0, 0.3438, -0.0114, 0.0, 0.3459, -0.0179, 0.0, 0.3459, -0.0157, 0.0, 0.3481, -0.0179, 0.0, 0.3481, -0.0157, 0.0, 0.3459, -0.0136,
0.0, 0.3459, -0.0114, 0.0, 0.3481, -0.0136, 0.0, 0.3481, -0.0114, 0.0, 0.3416, -0.0092, 0.0, 0.3416, -0.0071, 0.0, 0.3438, -0.0092, 0.0, 0.3438, -0.0071, 0.0, 0.3416, -0.0049, 0.0, 0.3416, -0.0027, 0.0, 0.3438, -0.0049, 0.0, 0.3438, -0.0027, 0.0, 0.3459, -0.0092, 0.0, 0.3459, -0.0071, 0.0, 0.3481, -0.0092, 0.0, 0.3481, -0.0071, 0.0, 0.3459, -0.0049, 0.0, 0.3459, -0.0027, 0.0, 0.3481, -0.0049, 0.0, 0.3481, -0.0027, 0.0, 0.3571, -0.0631, -0.0017, 0.357, -0.0631, -0.0039, 0.357, -0.0631, 0.0, 0.3527, -0.0589, 0.0, 0.3505, -0.0567, 0.0, 0.3494, -0.0556, 0.0, 0.3516, -0.0578, 0.0, 0.3524, -0.0548, -0.0017, 0.3527, -0.0589, -0.0039, 0.3527, -0.0589, -0.0017, 0.3505, -0.0567,
-0.0017, 0.3494, -0.0556, -0.0039, 0.3505, -0.0567, -0.0039, 0.3494, -0.0556, -0.0017, 0.3516, -0.0578, -0.0039, 0.3516, -0.0578, 0.0, 0.3549, -0.061, 0.0, 0.3538, -0.06, 0.0, 0.356, -0.0621, 0.0, 0.3568, -0.0591, 0.0, 0.3546, -0.057, 0.0, 0.355, -0.0544, 0.0, 0.3572, -0.0565, 0.0, 0.3561, -0.0554, -0.0017, 0.3549, -0.061, -0.0017, 0.3538, -0.06, -0.0039, 0.3549, -0.061, -0.0039, 0.3538, -0.06, -0.0017, 0.356, -0.0621, -0.0039, 0.356, -0.0621, -0.0017, 0.355, -0.0544, -0.0039, 0.355, -0.0544, -0.0017, 0.3572, -0.0565, -0.0017, 0.3562, -0.0554, -0.0039, 0.3572, -0.0565, -0.0039, 0.3562, -0.0554, 0.0, 0.3594, -0.0651, 0.0, 0.3583, -0.0642, 0.0, 0.3608, -0.066, 0.0, 0.3611, -0.0635, -0.0017, 0.3594, -0.065,
-0.0017, 0.3583, -0.0642, -0.0039, 0.3594, -0.065, -0.0039, 0.3583, -0.0642, -0.0017, 0.361, -0.0662, -0.0039, 0.361, -0.0662, 0.0, 0.3637, -0.067, 0.0, 0.3637, -0.0635, -0.0017, 0.3637, -0.067, -0.0017, 0.3637, -0.0635, -0.0039, 0.3637, -0.067, -0.0039, 0.3637, -0.0635, 0.0, 0.3589, -0.0613, 0.0, 0.3595, -0.0583, 0.0, 0.3611, -0.0613, 0.0, 0.3608, -0.0591, 0.0, 0.3585, -0.0575, 0.0, 0.3608, -0.057, 0.0, 0.3608, -0.0548, -0.0017, 0.3596, -0.0584, -0.0039, 0.3596, -0.0584, -0.0017, 0.3608, -0.0591, -0.0039, 0.3608, -0.0591, -0.0017, 0.3585, -0.0575, -0.0039, 0.3585, -0.0575, -0.0017, 0.3608, -0.057, -0.0017, 0.3608, -0.0548, -0.0039, 0.3608, -0.057, -0.0039, 0.3608, -0.0548, 0.0, 0.3637, -0.0613, 0.0, 0.3637, -0.0591,
0.0, 0.3637, -0.057, 0.0, 0.3637, -0.0548, -0.0017, 0.3637, -0.0613, -0.0017, 0.3637, -0.0591, -0.0039, 0.3637, -0.0613, -0.0039, 0.3637, -0.0591, -0.0017, 0.3637, -0.057, -0.0017, 0.3637, -0.0548, -0.0039, 0.3637, -0.057, -0.0039, 0.3637, -0.0548, 0.0, 0.3503, -0.0526, 0.0, 0.3507, -0.05, 0.0, 0.3529, -0.0522, 0.0, 0.3518, -0.0511, 0.0, 0.3496, -0.0489, -0.0017, 0.3507, -0.05, -0.0039, 0.3507, -0.05, -0.0017, 0.3529, -0.0522, -0.0017, 0.3518, -0.0511, -0.0039, 0.3529, -0.0522, -0.0039, 0.3518, -0.0511, -0.0017, 0.3496, -0.049, -0.0039, 0.3496, -0.049, 0.0, 0.354, -0.0533, -0.0017, 0.354, -0.0533, -0.0039, 0.354, -0.0533, 0.0, 0.3608, -0.0526, 0.0, 0.3608, -0.0505, 0.0, 0.3608, -0.0483, 0.0, 0.3608, -0.0461,
-0.0017, 0.3608, -0.0526, -0.0017, 0.3608, -0.0505, -0.0039, 0.3608, -0.0526, -0.0039, 0.3608, -0.0505, -0.0017, 0.3608, -0.0483, -0.0017, 0.3608, -0.0461, -0.0039, 0.3608, -0.0483, -0.0039, 0.3608, -0.0461, 0.0, 0.3637, -0.0526, 0.0, 0.3637, -0.0505, 0.0, 0.3637, -0.0483, 0.0, 0.3637, -0.0461, -0.0017, 0.3637, -0.0526, -0.0017, 0.3637, -0.0505, -0.0039, 0.3637, -0.0526, -0.0039, 0.3637, -0.0505, -0.0017, 0.3637, -0.0483, -0.0017, 0.3637, -0.0461, -0.0039, 0.3637, -0.0483, -0.0039, 0.3637, -0.0461, 0.0, 0.3608, -0.0439, 0.0, 0.3608, -0.0425, -0.0017, 0.3608, -0.0439, -0.0017, 0.3608, -0.0425, -0.0039, 0.3608, -0.0439, -0.0039, 0.3608, -0.0425, 0.0, 0.3637, -0.0439, 0.0, 0.3637, -0.0425, -0.0017, 0.3637, -0.0439, -0.0017, 0.3637, -0.0425,
-0.0039, 0.3637, -0.0439, -0.0039, 0.3637, -0.0425, 0.0, 0.3503, -0.0219, 0.0, 0.3503, -0.0201, 0.0, 0.3524, -0.0219, 0.0, 0.3524, -0.0201, -0.0017, 0.3503, -0.0219, -0.0039, 0.3503, -0.0219, -0.0017, 0.3524, -0.0219, -0.0039, 0.3524, -0.0219, 0.0, 0.3546, -0.0219, 0.0, 0.3546, -0.0201, 0.0, 0.3568, -0.0219, 0.0, 0.3568, -0.0201, -0.0017, 0.3546, -0.0219, -0.0039, 0.3546, -0.0219, -0.0017, 0.3568, -0.0219, -0.0039, 0.3568, -0.0219, 0.0, 0.3589, -0.0219, 0.0, 0.3589, -0.0201, 0.0, 0.3611, -0.0219, 0.0, 0.3611, -0.0201, -0.0017, 0.3589, -0.0219, -0.0039, 0.3589, -0.0219, -0.0017, 0.3611, -0.0219, -0.0039, 0.3611, -0.0219, 0.0, 0.3633, -0.0219, 0.0, 0.3633, -0.0201, 0.0, 0.3655, -0.0219, 0.0, 0.3655, -0.0201,
-0.0017, 0.3633, -0.0219, -0.0039, 0.3633, -0.0219, -0.0017, 0.3655, -0.0219, -0.0039, 0.3655, -0.0219, 0.0, 0.3503, -0.0179, 0.0, 0.3503, -0.0157, 0.0, 0.3524, -0.0179, 0.0, 0.3524, -0.0157, 0.0, 0.3503, -0.0136, 0.0, 0.3503, -0.0114, 0.0, 0.3524, -0.0136, 0.0, 0.3524, -0.0114, 0.0, 0.3546, -0.0179, 0.0, 0.3546, -0.0157, 0.0, 0.3568, -0.0179, 0.0, 0.3568, -0.0157, 0.0, 0.3546, -0.0136, 0.0, 0.3546, -0.0114, 0.0, 0.3568, -0.0136, 0.0, 0.3568, -0.0114, 0.0, 0.3503, -0.0092, 0.0, 0.3503, -0.0071, 0.0, 0.352, -0.0101, 0.0, 0.352, -0.0071, 0.0, 0.3503, -0.0049, 0.0, 0.3503, -0.0027, 0.0, 0.352, -0.0049, 0.0, 0.352, -0.0027, -0.0017, 0.352, -0.0101, -0.0017, 0.352, -0.0071,
-0.0039, 0.352, -0.0101, -0.0039, 0.352, -0.0071, -0.0017, 0.352, -0.0049, -0.0017, 0.352, -0.0027, -0.0039, 0.352, -0.0049, -0.0039, 0.352, -0.0027, 0.0, 0.3546, -0.0101, 0.0, 0.3568, -0.0101, -0.0017, 0.3546, -0.0101, -0.0039, 0.3546, -0.0101, -0.0017, 0.3568, -0.0101, -0.0039, 0.3568, -0.0101, 0.0, 0.3589, -0.0179, 0.0, 0.3589, -0.0157, 0.0, 0.3611, -0.0179, 0.0, 0.3611, -0.0157, 0.0, 0.3589, -0.0136, 0.0, 0.3589, -0.0114, 0.0, 0.3611, -0.0136, 0.0, 0.3611, -0.0114, 0.0, 0.3633, -0.0179, 0.0, 0.3633, -0.0157, 0.0, 0.3655, -0.0179, 0.0, 0.3655, -0.0157, 0.0, 0.3633, -0.0136, 0.0, 0.3633, -0.0114, 0.0, 0.3655, -0.0136, 0.0, 0.3655, -0.0114, 0.0, 0.3589, -0.0101, 0.0, 0.3611, -0.0101,
-0.0017, 0.3589, -0.0101, -0.0039, 0.3589, -0.0101, -0.0017, 0.3611, -0.0101, -0.0039, 0.3611, -0.0101, 0.0, 0.3633, -0.0101, 0.0, 0.3655, -0.0101, -0.0017, 0.3633, -0.0101, -0.0039, 0.3633, -0.0101, -0.0017, 0.3655, -0.0101, -0.0039, 0.3655, -0.0101, 0.0, 0.3676, -0.0219, 0.0, 0.3676, -0.0201, 0.0, 0.3698, -0.0219, 0.0, 0.3698, -0.0201, -0.0017, 0.3676, -0.0219, -0.0039, 0.3676, -0.0219, -0.0017, 0.3698, -0.0219, -0.0039, 0.3698, -0.0219, 0.0, 0.372, -0.0219, 0.0, 0.372, -0.0201, 0.0, 0.3741, -0.0219, 0.0, 0.3741, -0.0201, -0.0017, 0.372, -0.0219, -0.0039, 0.372, -0.0219, -0.0017, 0.3741, -0.0219, -0.0039, 0.3741, -0.0219, 0.0, 0.3763, -0.0219, 0.0, 0.3763, -0.0201, 0.0, 0.3785, -0.0219, 0.0, 0.3785, -0.0201,
-0.0017, 0.3763, -0.0219, -0.0039, 0.3763, -0.0219, -0.0017, 0.3785, -0.0219, -0.0039, 0.3785, -0.0219, 0.0, 0.3807, -0.0218, 0.0, 0.3807, -0.0201, 0.0, 0.3828, -0.0215, 0.0, 0.3828, -0.0201, -0.0017, 0.3807, -0.0218, -0.0039, 0.3807, -0.0218, -0.0017, 0.3828, -0.0215, -0.0039, 0.3828, -0.0215, 0.0, 0.3676, -0.0179, 0.0, 0.3676, -0.0157, 0.0, 0.3698, -0.0179, 0.0, 0.3698, -0.0157, 0.0, 0.3676, -0.0136, 0.0, 0.3676, -0.0114, 0.0, 0.3698, -0.0136, 0.0, 0.3698, -0.0114, 0.0, 0.372, -0.0179, 0.0, 0.372, -0.0157, 0.0, 0.3741, -0.0179, 0.0, 0.3741, -0.0157, 0.0, 0.372, -0.0136, 0.0, 0.372, -0.0114, 0.0, 0.3741, -0.0136, 0.0, 0.3741, -0.0114, 0.0, 0.3676, -0.0101, 0.0, 0.3698, -0.0101,
-0.0017, 0.3676, -0.0101, -0.0039, 0.3676, -0.0101, -0.0017, 0.3698, -0.0101, -0.0039, 0.3698, -0.0101, 0.0, 0.372, -0.0101, 0.0, 0.3741, -0.0101, -0.0017, 0.372, -0.0101, -0.0039, 0.372, -0.0101, -0.0017, 0.3741, -0.0101, -0.0039, 0.3741, -0.0101, 0.0, 0.3763, -0.0179, 0.0, 0.3763, -0.0157, 0.0, 0.3785, -0.0179, 0.0, 0.3785, -0.0157, 0.0, 0.3763, -0.0136, 0.0, 0.3763, -0.0114, 0.0, 0.3785, -0.0136, 0.0, 0.3785, -0.0114, 0.0, 0.3807, -0.0179, 0.0, 0.3807, -0.0157, 0.0, 0.3828, -0.0179, 0.0, 0.3828, -0.0157, 0.0, 0.3807, -0.0136, 0.0, 0.3807, -0.0114, 0.0, 0.3828, -0.0136, 0.0, 0.3828, -0.0114, 0.0, 0.3763, -0.0101, 0.0, 0.3785, -0.0101, -0.0017, 0.3763, -0.0101, -0.0039, 0.3763, -0.0101,
-0.0017, 0.3785, -0.0101, -0.0039, 0.3785, -0.0101, 0.0, 0.3806, -0.0101, 0.0, 0.3828, -0.0099, -0.0017, 0.3807, -0.0101, -0.0039, 0.3807, -0.0101, -0.0017, 0.3828, -0.0099, -0.0039, 0.3828, -0.0099, 0.0, 0.3863, -0.067, 0.0, 0.3863, -0.0656, 0.0, 0.3863, -0.0635, -0.0017, 0.3863, -0.067, -0.0039, 0.3863, -0.067, -0.0017, 0.3863, -0.0656, -0.0017, 0.3863, -0.0635, -0.0039, 0.3863, -0.0656, -0.0039, 0.3863, -0.0635, 0.0, 0.3893, -0.067, 0.0, 0.3915, -0.067, 0.0, 0.3893, -0.0656, 0.0, 0.3902, -0.0631, 0.0, 0.3915, -0.0656, 0.0, 0.3915, -0.0631, -0.0017, 0.3893, -0.067, -0.0039, 0.3893, -0.067, -0.0017, 0.3915, -0.067, -0.0039, 0.3915, -0.067, -0.0017, 0.3902, -0.0631, -0.0039, 0.3902, -0.0631, -0.0017, 0.3915, -0.0631,
-0.0039, 0.3915, -0.0631, 0.0, 0.3863, -0.0613, 0.0, 0.3863, -0.0591, 0.0, 0.3863, -0.057, 0.0, 0.3863, -0.0548, -0.0017, 0.3863, -0.0613, -0.0017, 0.3863, -0.0591, -0.0039, 0.3863, -0.0613, -0.0039, 0.3863, -0.0591, -0.0017, 0.3863, -0.057, -0.0017, 0.3863, -0.0548, -0.0039, 0.3863, -0.057, -0.0039, 0.3863, -0.0548, 0.0, 0.3902, -0.0613, 0.0, 0.3902, -0.0591, 0.0, 0.3902, -0.057, 0.0, 0.3902, -0.0548, -0.0017, 0.3902, -0.0613, -0.0017, 0.3902, -0.0591, -0.0039, 0.3902, -0.0613, -0.0039, 0.3902, -0.0591, -0.0017, 0.3902, -0.057, -0.0017, 0.3902, -0.0548, -0.0039, 0.3902, -0.057, -0.0039, 0.3902, -0.0548, 0.0, 0.3937, -0.067, 0.0, 0.3958, -0.067, 0.0, 0.3937, -0.0656, 0.0, 0.3937, -0.0631, 0.0, 0.3958, -0.0656,
0.0, 0.3958, -0.0631, -0.0017, 0.3937, -0.067, -0.0039, 0.3937, -0.067, -0.0017, 0.3958, -0.067, -0.0039, 0.3958, -0.067, -0.0017, 0.3937, -0.0631, -0.0039, 0.3937, -0.0631, -0.0017, 0.3958, -0.0631, -0.0039, 0.3958, -0.0631, 0.0, 0.398, -0.067, 0.0, 0.4002, -0.0669, 0.0, 0.398, -0.0656, 0.0, 0.398, -0.063, 0.0, 0.4002, -0.0656, 0.0, 0.4001, -0.063, -0.0017, 0.398, -0.067, -0.0039, 0.398, -0.067, -0.0017, 0.4002, -0.0669, -0.0039, 0.4002, -0.0669, -0.0017, 0.398, -0.063, -0.0039, 0.398, -0.063, -0.0017, 0.4002, -0.063, -0.0039, 0.4002, -0.063, 0.0, 0.3863, -0.0526, 0.0, 0.3863, -0.0505, 0.0, 0.3863, -0.0483, 0.0, 0.3863, -0.0461, -0.0017, 0.3863, -0.0526, -0.0017, 0.3863, -0.0505, -0.0039, 0.3863, -0.0526,
-0.0039, 0.3863, -0.0505, -0.0017, 0.3863, -0.0483, -0.0017, 0.3863, -0.0461, -0.0039, 0.3863, -0.0483, -0.0039, 0.3863, -0.0461, 0.0, 0.3902, -0.0526, 0.0, 0.3902, -0.0505, 0.0, 0.3902, -0.0483, 0.0, 0.3902, -0.0454, 0.0, 0.3915, -0.0454, -0.0017, 0.3902, -0.0526, -0.0017, 0.3902, -0.0505, -0.0039, 0.3902, -0.0526, -0.0039, 0.3902, -0.0505, -0.0017, 0.3902, -0.0483, -0.0017, 0.3902, -0.0454, -0.0039, 0.3902, -0.0483, -0.0039, 0.3902, -0.0454, -0.0017, 0.3915, -0.0454, -0.0039, 0.3915, -0.0454, 0.0, 0.3863, -0.0439, 0.0, 0.3863, -0.0425, -0.0017, 0.3863, -0.0439, -0.0017, 0.3863, -0.0425, -0.0039, 0.3863, -0.0439, -0.0039, 0.3863, -0.0425, 0.0, 0.3893, -0.0439, 0.0, 0.3893, -0.0425, 0.0, 0.3915, -0.0439, 0.0, 0.3915, -0.0425,
-0.0017, 0.3893, -0.0425, -0.0039, 0.3893, -0.0425, -0.0017, 0.3915, -0.0425, -0.0039, 0.3915, -0.0425, 0.0, 0.3937, -0.0454, 0.0, 0.3958, -0.0454, -0.0017, 0.3937, -0.0454, -0.0039, 0.3937, -0.0454, -0.0017, 0.3958, -0.0454, -0.0039, 0.3958, -0.0454, 0.0, 0.398, -0.0455, 0.0, 0.4001, -0.0457, -0.0017, 0.398, -0.0455, -0.0039, 0.398, -0.0455, -0.0017, 0.4002, -0.0458, -0.0039, 0.4002, -0.0458, 0.0, 0.3937, -0.0439, 0.0, 0.3937, -0.0425, 0.0, 0.3958, -0.0439, 0.0, 0.3958, -0.0425, -0.0017, 0.3937, -0.0425, -0.0039, 0.3937, -0.0425, -0.0017, 0.3958, -0.0425, -0.0039, 0.3958, -0.0425, 0.0, 0.398, -0.0439, 0.0, 0.398, -0.0425, 0.0, 0.4002, -0.0439, 0.0, 0.4002, -0.0425, -0.0017, 0.398, -0.0425, -0.0039, 0.398, -0.0425,
-0.0017, 0.4002, -0.0425, -0.0039, 0.4002, -0.0425, 0.0, 0.4016, -0.0668, 0.0, 0.4025, -0.0667, 0.0, 0.4023, -0.0627, 0.0, 0.4048, -0.0661, 0.0, 0.4045, -0.0635, -0.0017, 0.4017, -0.0668, -0.0039, 0.4017, -0.0668, -0.0017, 0.4028, -0.0666, -0.0017, 0.4023, -0.0627, -0.0039, 0.4028, -0.0666, -0.0039, 0.4023, -0.0627, -0.0017, 0.4046, -0.0662, -0.0039, 0.4046, -0.0662, 0.0, 0.4064, -0.0653, 0.0, 0.4075, -0.0645, 0.0, 0.4089, -0.0635, -0.0017, 0.4066, -0.0652, -0.0017, 0.4076, -0.0644, -0.0039, 0.4066, -0.0652, -0.0039, 0.4076, -0.0644, -0.0017, 0.409, -0.0635, -0.0039, 0.409, -0.0635, 0.0, 0.4033, -0.0623, 0.0, 0.4045, -0.0619, -0.0017, 0.4033, -0.0623, -0.0039, 0.4033, -0.0623, -0.0017, 0.4046, -0.0619, -0.0039, 0.4046, -0.0619,
0.0, 0.4062, -0.0608, 0.0, 0.4073, -0.0596, 0.0, 0.41, -0.0623, 0.0, 0.4079, -0.0587, 0.0, 0.4085, -0.0569, 0.0, 0.4087, -0.0548, -0.0017, 0.4062, -0.0608, -0.0017, 0.4074, -0.0595, -0.0039, 0.4062, -0.0608, -0.0039, 0.4074, -0.0595, -0.0017, 0.4099, -0.0624, -0.0017, 0.408, -0.0584, -0.0039, 0.4099, -0.0624, -0.0039, 0.408, -0.0584, -0.0017, 0.4085, -0.057, -0.0017, 0.4087, -0.0548, -0.0039, 0.4085, -0.057, -0.0039, 0.4087, -0.0548, 0.0, 0.4108, -0.0612, 0.0, 0.4119, -0.0593, 0.0, 0.4122, -0.0582, 0.0, 0.411, -0.057, 0.0, 0.411, -0.0548, 0.0, 0.4125, -0.057, 0.0, 0.4127, -0.0548, -0.0017, 0.4107, -0.0613, -0.0017, 0.4119, -0.0594, -0.0039, 0.4107, -0.0613, -0.0039, 0.4119, -0.0594, -0.0017, 0.4122, -0.0582,
-0.0039, 0.4122, -0.0582, -0.0017, 0.4125, -0.057, -0.0017, 0.4127, -0.0548, -0.0039, 0.4125, -0.057, -0.0039, 0.4127, -0.0548, 0.0, 0.4022, -0.0462, 0.0, 0.4051, -0.0476, 0.0, 0.4041, -0.0471, -0.0017, 0.4024, -0.0463, -0.0039, 0.4024, -0.0463, -0.0017, 0.405, -0.0476, -0.0017, 0.4039, -0.047, -0.0039, 0.405, -0.0476, -0.0039, 0.4039, -0.047, 0.0, 0.4074, -0.0498, 0.0, 0.4086, -0.0528, 0.0, 0.408, -0.0508, 0.0, 0.4064, -0.0486, 0.0, 0.4067, -0.0461, 0.0, 0.4089, -0.0483, 0.0, 0.4093, -0.0456, -0.0017, 0.4075, -0.0498, -0.0039, 0.4075, -0.0498, -0.0017, 0.4086, -0.0526, -0.0017, 0.4081, -0.0509, -0.0039, 0.4086, -0.0526, -0.0039, 0.4081, -0.0509, -0.0017, 0.4064, -0.0486, -0.0039, 0.4064, -0.0486, -0.0017, 0.4093, -0.0456,
-0.0039, 0.4093, -0.0456, 0.0, 0.4024, -0.0439, 0.0, 0.4024, -0.0426, 0.0, 0.4047, -0.043, 0.0, 0.4036, -0.0428, -0.0017, 0.4024, -0.0426, -0.0039, 0.4024, -0.0426, -0.0017, 0.4048, -0.0431, -0.0017, 0.4037, -0.0428, -0.0039, 0.4048, -0.0431, -0.0039, 0.4037, -0.0428, 0.0, 0.407, -0.044, 0.0, 0.4081, -0.0447, -0.0017, 0.4067, -0.0439, -0.0039, 0.4067, -0.0439, -0.0017, 0.4082, -0.0448, -0.0039, 0.4082, -0.0448, 0.0, 0.411, -0.0526, 0.0, 0.4121, -0.0502, 0.0, 0.4126, -0.0526, 0.0, 0.4123, -0.0511, 0.0, 0.4111, -0.048, 0.0, 0.4103, -0.0468, -0.0017, 0.412, -0.0499, -0.0039, 0.412, -0.0499, -0.0017, 0.4126, -0.0526, -0.0017, 0.4122, -0.051, -0.0039, 0.4126, -0.0526, -0.0039, 0.4122, -0.051, -0.0017, 0.4112, -0.0482,
-0.0017, 0.4103, -0.0468, -0.0039, 0.4112, -0.0482, -0.0039, 0.4103, -0.0468, 0.0, 0.3844, -0.0213, 0.0, 0.3852, -0.0212, 0.0, 0.3873, -0.0209, -0.0017, 0.3846, -0.0213, -0.0017, 0.3857, -0.0211, -0.0039, 0.3846, -0.0213, -0.0039, 0.3857, -0.0211, -0.0017, 0.3872, -0.0209, -0.0039, 0.3872, -0.0209, 0.0, 0.3895, -0.0206, 0.0, 0.3917, -0.02, -0.0017, 0.3893, -0.0206, -0.0039, 0.3893, -0.0206, -0.0017, 0.3915, -0.0201, -0.0039, 0.3915, -0.0201, 0.0, 0.3933, -0.0194, -0.0017, 0.3935, -0.0194, -0.0039, 0.3935, -0.0194, 0.0, 0.385, -0.0179, 0.0, 0.385, -0.0157, 0.0, 0.3872, -0.0179, 0.0, 0.3872, -0.0157, 0.0, 0.385, -0.0136, 0.0, 0.385, -0.0114, 0.0, 0.3872, -0.0136, 0.0, 0.3872, -0.0114, 0.0, 0.3893, -0.0179,
0.0, 0.3893, -0.0157, 0.0, 0.3915, -0.0179, 0.0, 0.3915, -0.0157, 0.0, 0.3893, -0.0136, 0.0, 0.3893, -0.0114, 0.0, 0.3915, -0.0136, 0.0, 0.3915, -0.0114, 0.0, 0.3849, -0.0096, 0.0, 0.387, -0.0091, -0.0017, 0.385, -0.0096, -0.0039, 0.385, -0.0096, -0.0017, 0.3872, -0.009, -0.0039, 0.3872, -0.009, 0.0, 0.3892, -0.0084, 0.0, 0.3902, -0.008, 0.0, 0.3915, -0.0092, 0.0, 0.3914, -0.0076, -0.0017, 0.3891, -0.0084, -0.0017, 0.3902, -0.0081, -0.0039, 0.3891, -0.0084, -0.0039, 0.3902, -0.0081, -0.0017, 0.3915, -0.0075, -0.0039, 0.3915, -0.0075, 0.0, 0.3942, -0.0191, 0.0, 0.3937, -0.0157, 0.0, 0.3961, -0.0183, 0.0, 0.3958, -0.0157, 0.0, 0.3937, -0.0136, 0.0, 0.3937, -0.0114, 0.0, 0.3958, -0.0136,
0.0, 0.3958, -0.0114, -0.0017, 0.3946, -0.0189, -0.0039, 0.3946, -0.0189, -0.0017, 0.3959, -0.0185, -0.0039, 0.3959, -0.0185, 0.0, 0.3982, -0.0174, 0.0, 0.398, -0.0157, 0.0, 0.3991, -0.0169, 0.0, 0.4003, -0.0162, 0.0, 0.398, -0.0136, 0.0, 0.398, -0.0114, 0.0, 0.4002, -0.0136, 0.0, 0.4002, -0.0114, -0.0017, 0.398, -0.0175, -0.0039, 0.398, -0.0175, -0.0017, 0.3992, -0.0169, -0.0017, 0.4003, -0.0162, -0.0039, 0.3992, -0.0169, -0.0039, 0.4003, -0.0162, 0.0, 0.3937, -0.0092, 0.0, 0.3935, -0.0065, 0.0, 0.3958, -0.0092, 0.0, 0.3958, -0.0071, 0.0, 0.3945, -0.0058, 0.0, 0.3956, -0.0051, -0.0017, 0.3935, -0.0065, -0.0039, 0.3935, -0.0065, -0.0017, 0.3945, -0.0058, -0.0039, 0.3945, -0.0058, -0.0017, 0.3959, -0.0049,
-0.0039, 0.3959, -0.0049, 0.0, 0.398, -0.0092, 0.0, 0.398, -0.0071, 0.0, 0.4002, -0.0092, 0.0, 0.4002, -0.0071, 0.0, 0.3971, -0.0038, 0.0, 0.3981, -0.0028, 0.0, 0.4002, -0.0049, 0.0, 0.399, -0.0017, -0.0017, 0.397, -0.0039, -0.0017, 0.3982, -0.0028, -0.0039, 0.397, -0.0039, -0.0039, 0.3982, -0.0028, -0.0017, 0.3991, -0.0016, -0.0039, 0.3991, -0.0016, 0.0, 0.4018, -0.0151, 0.0, 0.4029, -0.0143, 0.0, 0.4024, -0.0114, 0.0, 0.4042, -0.0132, 0.0, 0.4053, -0.0122, -0.0017, 0.4018, -0.0151, -0.0039, 0.4018, -0.0151, -0.0017, 0.4029, -0.0142, -0.0039, 0.4029, -0.0142, -0.0017, 0.4042, -0.0132, -0.0017, 0.4053, -0.0122, -0.0039, 0.4042, -0.0132, -0.0039, 0.4053, -0.0122, 0.0, 0.4064, -0.0111, -0.0017, 0.4064, -0.0111,
-0.0039, 0.4064, -0.0111, 0.0, 0.4024, -0.0092, 0.0, 0.4024, -0.0071, 0.0, 0.4045, -0.0092, 0.0, 0.4045, -0.0071, 0.0, 0.4024, -0.0049, 0.0, 0.4024, -0.0027, 0.0, 0.4045, -0.0049, 0.0, 0.4045, -0.0027, 0.0, 0.4075, -0.0099, 0.0, 0.4067, -0.0071, 0.0, 0.4084, -0.0088, 0.0, 0.4095, -0.0074, 0.0, 0.4067, -0.0049, 0.0, 0.4067, -0.0027, 0.0, 0.4089, -0.0049, 0.0, 0.4089, -0.0027, -0.0017, 0.4075, -0.0099, -0.0039, 0.4075, -0.0099, -0.0017, 0.4084, -0.0089, -0.0017, 0.4094, -0.0074, -0.0039, 0.4084, -0.0089, -0.0039, 0.4094, -0.0074, 0.0, 0.4102, -0.0063, 0.0, 0.4109, -0.0051, 0.0, 0.412, -0.0032, 0.0, 0.4125, -0.0021, -0.0017, 0.4102, -0.0063, -0.0039, 0.4102, -0.0063, -0.0017, 0.4111, -0.0049,
-0.0017, 0.4119, -0.0034, -0.0039, 0.4111, -0.0049, -0.0039, 0.4119, -0.0034, -0.0017, 0.4124, -0.0023, -0.0039, 0.4124, -0.0023, 0.0, 0.4324, -0.067, 0.0, 0.4349, -0.067, 0.0, 0.4324, -0.0656, 0.0, 0.4324, -0.0635, 0.0, 0.4349, -0.0656, 0.0, 0.4353, -0.0631, -0.0017, 0.4324, -0.067, -0.0039, 0.4324, -0.067, -0.0017, 0.4349, -0.067, -0.0039, 0.4349, -0.067, -0.0017, 0.4324, -0.0656, -0.0017, 0.4324, -0.0635, -0.0039, 0.4324, -0.0656, -0.0039, 0.4324, -0.0635, -0.0017, 0.4353, -0.0631, -0.0039, 0.4353, -0.0631, 0.0, 0.4324, -0.0613, 0.0, 0.4324, -0.0591, 0.0, 0.4353, -0.0613, 0.0, 0.4353, -0.0591, 0.0, 0.4324, -0.057, 0.0, 0.4324, -0.0548, 0.0, 0.4353, -0.057, 0.0, 0.4353, -0.0548, -0.0017, 0.4324, -0.0613,
-0.0017, 0.4324, -0.0591, -0.0039, 0.4324, -0.0613, -0.0039, 0.4324, -0.0591, -0.0017, 0.4353, -0.0613, -0.0017, 0.4353, -0.0591, -0.0039, 0.4353, -0.0613, -0.0039, 0.4353, -0.0591, -0.0017, 0.4324, -0.057, -0.0017, 0.4324, -0.0548, -0.0039, 0.4324, -0.057, -0.0039, 0.4324, -0.0548, -0.0017, 0.4353, -0.057, -0.0017, 0.4353, -0.0548, -0.0039, 0.4353, -0.057, -0.0039, 0.4353, -0.0548, 0.0, 0.4324, -0.0526, 0.0, 0.4324, -0.0505, 0.0, 0.4353, -0.0527, 0.0, 0.4353, -0.0506, 0.0, 0.4324, -0.0483, 0.0, 0.4324, -0.0461, 0.0, 0.4359, -0.0487, 0.0, 0.4349, -0.0461, -0.0017, 0.4324, -0.0526, -0.0017, 0.4324, -0.0505, -0.0039, 0.4324, -0.0526, -0.0039, 0.4324, -0.0505, -0.0017, 0.4353, -0.0526, -0.0017, 0.4355, -0.0497, -0.0039, 0.4353, -0.0526,
-0.0039, 0.4355, -0.0497, -0.0017, 0.4324, -0.0483, -0.0017, 0.4324, -0.0461, -0.0039, 0.4324, -0.0483, -0.0039, 0.4324, -0.0461, -0.0017, 0.4358, -0.0489, -0.0039, 0.4358, -0.0489, 0.0, 0.4324, -0.0439, 0.0, 0.4324, -0.0425, 0.0, 0.4349, -0.0439, 0.0, 0.4349, -0.0425, -0.0017, 0.4324, -0.0439, -0.0017, 0.4324, -0.0425, -0.0039, 0.4324, -0.0439, -0.0039, 0.4324, -0.0425, -0.0017, 0.4349, -0.0425, -0.0039, 0.4349, -0.0425, 0.0, 0.4371, -0.067, 0.0, 0.4392, -0.067, 0.0, 0.4371, -0.0656, 0.0, 0.4371, -0.0631, 0.0, 0.4392, -0.0656, 0.0, 0.4392, -0.0631, -0.0017, 0.4371, -0.067, -0.0039, 0.4371, -0.067, -0.0017, 0.4392, -0.067, -0.0039, 0.4392, -0.067, -0.0017, 0.4371, -0.0631, -0.0039, 0.4371, -0.0631, -0.0017, 0.4392, -0.0631,
-0.0039, 0.4392, -0.0631, 0.0, 0.4414, -0.067, 0.0, 0.4436, -0.067, 0.0, 0.4414, -0.0656, 0.0, 0.4414, -0.0631, 0.0, 0.4436, -0.0656, 0.0, 0.4436, -0.0631, -0.0017, 0.4414, -0.067, -0.0039, 0.4414, -0.067, -0.0017, 0.4436, -0.067, -0.0039, 0.4436, -0.067, -0.0017, 0.4414, -0.0631, -0.0039, 0.4414, -0.0631, -0.0017, 0.4436, -0.0631, -0.0039, 0.4436, -0.0631, 0.0, 0.4382, -0.0562, 0.0, 0.4382, -0.0548, -0.0017, 0.4382, -0.0562, -0.0017, 0.4382, -0.0548, -0.0039, 0.4382, -0.0562, -0.0039, 0.4382, -0.0548, 0.0, 0.4414, -0.0562, 0.0, 0.4414, -0.0548, 0.0, 0.4436, -0.0562, 0.0, 0.4436, -0.0548, -0.0017, 0.4414, -0.0562, -0.0039, 0.4414, -0.0562, -0.0017, 0.4436, -0.0562, -0.0039, 0.4436, -0.0562, 0.0, 0.4458, -0.067,
0.0, 0.4479, -0.067, 0.0, 0.4458, -0.0656, 0.0, 0.4458, -0.0631, 0.0, 0.4479, -0.0656, 0.0, 0.4479, -0.0631, -0.0017, 0.4458, -0.067, -0.0039, 0.4458, -0.067, -0.0017, 0.4479, -0.067, -0.0039, 0.4479, -0.067, -0.0017, 0.4458, -0.0631, -0.0039, 0.4458, -0.0631, -0.0017, 0.4479, -0.0631, -0.0039, 0.4479, -0.0631, 0.0, 0.4501, -0.067, 0.0, 0.4529, -0.067, 0.0, 0.4501, -0.0656, 0.0, 0.4501, -0.0631, 0.0, 0.4529, -0.0656, 0.0, 0.4529, -0.0631, -0.0017, 0.4501, -0.067, -0.0039, 0.4501, -0.067, -0.0017, 0.4529, -0.067, -0.0039, 0.4529, -0.067, -0.0017, 0.4501, -0.0631, -0.0039, 0.4501, -0.0631, -0.0017, 0.4529, -0.0656, -0.0017, 0.4529, -0.0631, -0.0039, 0.4529, -0.0656, -0.0039, 0.4529, -0.0631, 0.0, 0.4458, -0.0562,
0.0, 0.4458, -0.0548, 0.0, 0.4479, -0.0562, 0.0, 0.4479, -0.0548, -0.0017, 0.4458, -0.0562, -0.0039, 0.4458, -0.0562, -0.0017, 0.4479, -0.0562, -0.0039, 0.4479, -0.0562, 0.0, 0.4501, -0.0562, 0.0, 0.4501, -0.0548, 0.0, 0.452, -0.0562, 0.0, 0.452, -0.0548, -0.0017, 0.4501, -0.0562, -0.0039, 0.4501, -0.0562, -0.0017, 0.452, -0.0562, -0.0017, 0.452, -0.0548, -0.0039, 0.452, -0.0562, -0.0039, 0.452, -0.0548, 0.0, 0.4382, -0.0523, 0.0, 0.4363, -0.0477, 0.0, 0.4372, -0.0466, 0.0, 0.4393, -0.0457, -0.0017, 0.4382, -0.0523, -0.0039, 0.4382, -0.0523, -0.0017, 0.4362, -0.0478, -0.0017, 0.4372, -0.0466, -0.0039, 0.4362, -0.0478, -0.0039, 0.4372, -0.0466, -0.0017, 0.4396, -0.0456, -0.0039, 0.4396, -0.0456, 0.0, 0.4414, -0.0523,
0.0, 0.4436, -0.0523, 0.0, 0.4414, -0.0455, 0.0, 0.4436, -0.0454, -0.0017, 0.4414, -0.0523, -0.0039, 0.4414, -0.0523, -0.0017, 0.4436, -0.0523, -0.0039, 0.4436, -0.0523, -0.0017, 0.4414, -0.0455, -0.0039, 0.4414, -0.0455, -0.0017, 0.4436, -0.0454, -0.0039, 0.4436, -0.0454, 0.0, 0.4371, -0.0439, 0.0, 0.4371, -0.0425, 0.0, 0.4392, -0.0439, 0.0, 0.4392, -0.0425, -0.0017, 0.4371, -0.0425, -0.0039, 0.4371, -0.0425, -0.0017, 0.4392, -0.0425, -0.0039, 0.4392, -0.0425, 0.0, 0.4414, -0.0439, 0.0, 0.4414, -0.0425, 0.0, 0.4436, -0.0439, 0.0, 0.4436, -0.0425, -0.0017, 0.4414, -0.0425, -0.0039, 0.4414, -0.0425, -0.0017, 0.4436, -0.0425, -0.0039, 0.4436, -0.0425, 0.0, 0.4458, -0.0523, 0.0, 0.4479, -0.0523, 0.0, 0.4458, -0.0454,
0.0, 0.4479, -0.0454, -0.0017, 0.4458, -0.0523, -0.0039, 0.4458, -0.0523, -0.0017, 0.4479, -0.0523, -0.0039, 0.4479, -0.0523, -0.0017, 0.4458, -0.0454, -0.0039, 0.4458, -0.0454, -0.0017, 0.4479, -0.0454, -0.0039, 0.4479, -0.0454, 0.0, 0.4501, -0.0523, 0.0, 0.452, -0.0523, 0.0, 0.4501, -0.0454, 0.0, 0.4529, -0.0454, -0.0017, 0.4501, -0.0523, -0.0039, 0.4501, -0.0523, -0.0017, 0.452, -0.0523, -0.0039, 0.452, -0.0523, -0.0017, 0.4501, -0.0454, -0.0039, 0.4501, -0.0454, -0.0017, 0.4529, -0.0454, -0.0039, 0.4529, -0.0454, 0.0, 0.4458, -0.0439, 0.0, 0.4458, -0.0425, 0.0, 0.4479, -0.0439, 0.0, 0.4479, -0.0425, -0.0017, 0.4458, -0.0425, -0.0039, 0.4458, -0.0425, -0.0017, 0.4479, -0.0425, -0.0039, 0.4479, -0.0425, 0.0, 0.4501, -0.0439,
0.0, 0.4501, -0.0425, 0.0, 0.4529, -0.0439, 0.0, 0.4529, -0.0425, -0.0017, 0.4501, -0.0425, -0.0039, 0.4501, -0.0425, -0.0017, 0.4529, -0.0439, -0.0017, 0.4529, -0.0425, -0.0039, 0.4529, -0.0439, -0.0039, 0.4529, -0.0425, 0.0, 0.4245, -0.0219, 0.0, 0.4245, -0.0201, 0.0, 0.4262, -0.0219, 0.0, 0.4262, -0.0201, -0.0017, 0.4245, -0.0219, -0.0017, 0.4245, -0.0201, -0.0039, 0.4245, -0.0219, -0.0039, 0.4245, -0.0201, -0.0017, 0.4262, -0.0219, -0.0039, 0.4262, -0.0219, 0.0, 0.4284, -0.0219, 0.0, 0.4284, -0.0201, 0.0, 0.4306, -0.0219, 0.0, 0.4306, -0.0201, -0.0017, 0.4284, -0.0219, -0.0039, 0.4284, -0.0219, -0.0017, 0.4306, -0.0219, -0.0039, 0.4306, -0.0219, 0.0, 0.4327, -0.0219, 0.0, 0.4327, -0.0201, 0.0, 0.4349, -0.0219,
0.0, 0.4349, -0.0201, -0.0017, 0.4327, -0.0219, -0.0039, 0.4327, -0.0219, -0.0017, 0.4349, -0.0219, -0.0039, 0.4349, -0.0219, 0.0, 0.4245, -0.0179, 0.0, 0.4245, -0.0157, 0.0, 0.4262, -0.0179, 0.0, 0.4262, -0.0157, 0.0, 0.4245, -0.0136, 0.0, 0.4245, -0.0114, 0.0, 0.4262, -0.0136, 0.0, 0.4262, -0.0114, -0.0017, 0.4245, -0.0179, -0.0017, 0.4245, -0.0157, -0.0039, 0.4245, -0.0179, -0.0039, 0.4245, -0.0157, -0.0017, 0.4245, -0.0136, -0.0017, 0.4245, -0.0114, -0.0039, 0.4245, -0.0136, -0.0039, 0.4245, -0.0114, 0.0, 0.4245, -0.0101, 0.0, 0.4262, -0.0101, -0.0017, 0.4245, -0.0101, -0.0039, 0.4245, -0.0101, -0.0017, 0.4262, -0.0101, -0.0039, 0.4262, -0.0101, 0.0, 0.4284, -0.0179, 0.0, 0.4284, -0.0157, 0.0, 0.4306, -0.0179,
0.0, 0.4306, -0.0157, 0.0, 0.4284, -0.0136, 0.0, 0.4284, -0.0114, 0.0, 0.4306, -0.0136, 0.0, 0.4306, -0.0114, 0.0, 0.4327, -0.0179, 0.0, 0.4327, -0.0157, 0.0, 0.4349, -0.0179, 0.0, 0.4349, -0.0157, 0.0, 0.4327, -0.0136, 0.0, 0.4327, -0.0114, 0.0, 0.4349, -0.0136, 0.0, 0.4349, -0.0114, 0.0, 0.4284, -0.0101, 0.0, 0.4306, -0.0101, -0.0017, 0.4284, -0.0101, -0.0039, 0.4284, -0.0101, -0.0017, 0.4306, -0.0101, -0.0039, 0.4306, -0.0101, 0.0, 0.4327, -0.0101, 0.0, 0.4349, -0.0101, -0.0017, 0.4327, -0.0101, -0.0039, 0.4327, -0.0101, -0.0017, 0.4349, -0.0101, -0.0039, 0.4349, -0.0101, 0.0, 0.4371, -0.0219, 0.0, 0.4371, -0.0201, 0.0, 0.4392, -0.0219, 0.0, 0.4392, -0.0201, -0.0017, 0.4371, -0.0219,
-0.0039, 0.4371, -0.0219, -0.0017, 0.4392, -0.0219, -0.0039, 0.4392, -0.0219, 0.0, 0.4414, -0.0219, 0.0, 0.4414, -0.0201, 0.0, 0.4436, -0.0219, 0.0, 0.4436, -0.0201, -0.0017, 0.4414, -0.0219, -0.0039, 0.4414, -0.0219, -0.0017, 0.4436, -0.0219, -0.0039, 0.4436, -0.0219, 0.0, 0.4458, -0.0219, 0.0, 0.4458, -0.0201, 0.0, 0.4479, -0.0219, 0.0, 0.4479, -0.0201, -0.0017, 0.4458, -0.0219, -0.0039, 0.4458, -0.0219, -0.0017, 0.4479, -0.0219, -0.0039, 0.4479, -0.0219, 0.0, 0.4501, -0.0219, 0.0, 0.4501, -0.0201, 0.0, 0.4523, -0.0219, 0.0, 0.4523, -0.0201, -0.0017, 0.4501, -0.0219, -0.0039, 0.4501, -0.0219, -0.0017, 0.4523, -0.0219, -0.0039, 0.4523, -0.0219, 0.0, 0.4371, -0.0179, 0.0, 0.4371, -0.0157, 0.0, 0.4392, -0.0179,
0.0, 0.4392, -0.0157, 0.0, 0.4371, -0.0136, 0.0, 0.4371, -0.0114, 0.0, 0.4392, -0.0136, 0.0, 0.4392, -0.0114, 0.0, 0.4414, -0.0179, 0.0, 0.4414, -0.0157, 0.0, 0.4436, -0.0179, 0.0, 0.4436, -0.0157, 0.0, 0.4414, -0.0136, 0.0, 0.4414, -0.0114, 0.0, 0.4436, -0.0136, 0.0, 0.4436, -0.0114, 0.0, 0.4371, -0.0101, 0.0, 0.4392, -0.0101, -0.0017, 0.4371, -0.0101, -0.0039, 0.4371, -0.0101, -0.0017, 0.4392, -0.0101, -0.0039, 0.4392, -0.0101, 0.0, 0.4414, -0.0101, 0.0, 0.4436, -0.0101, -0.0017, 0.4414, -0.0101, -0.0039, 0.4414, -0.0101, -0.0017, 0.4436, -0.0101, -0.0039, 0.4436, -0.0101, 0.0, 0.4458, -0.0179, 0.0, 0.4458, -0.0157, 0.0, 0.4479, -0.0179, 0.0, 0.4479, -0.0157, 0.0, 0.4458, -0.0136,
0.0, 0.4458, -0.0114, 0.0, 0.4479, -0.0136, 0.0, 0.4479, -0.0114, 0.0, 0.4501, -0.0179, 0.0, 0.4501, -0.0157, 0.0, 0.4523, -0.0179, 0.0, 0.4523, -0.0157, 0.0, 0.4501, -0.0136, 0.0, 0.4501, -0.0114, 0.0, 0.4523, -0.0136, 0.0, 0.4523, -0.0114, 0.0, 0.4458, -0.0101, 0.0, 0.4479, -0.0101, -0.0017, 0.4458, -0.0101, -0.0039, 0.4458, -0.0101, -0.0017, 0.4479, -0.0101, -0.0039, 0.4479, -0.0101, 0.0, 0.4501, -0.0101, 0.0, 0.4523, -0.0101, -0.0017, 0.4501, -0.0101, -0.0039, 0.4501, -0.0101, -0.0017, 0.4523, -0.0101, -0.0039, 0.4523, -0.0101, 0.0, 0.4725, -0.067, 0.0, 0.474, -0.067, 0.0, 0.4725, -0.0656, 0.0, 0.4725, -0.0635, 0.0, 0.474, -0.0656, 0.0, 0.474, -0.0635, -0.0017, 0.4725, -0.067,
-0.0039, 0.4725, -0.067, -0.0017, 0.474, -0.067, -0.0039, 0.474, -0.067, -0.0017, 0.4725, -0.0656, -0.0017, 0.4725, -0.0635, -0.0039, 0.4725, -0.0656, -0.0039, 0.4725, -0.0635, 0.0, 0.4765, -0.067, 0.0, 0.4765, -0.0656, 0.0, 0.4765, -0.0635, -0.0017, 0.4765, -0.067, -0.0039, 0.4765, -0.067, -0.0017, 0.4765, -0.0656, -0.0017, 0.4765, -0.0635, -0.0039, 0.4765, -0.0656, -0.0039, 0.4765, -0.0635, 0.0, 0.4725, -0.0613, 0.0, 0.4725, -0.0591, 0.0, 0.474, -0.0613, 0.0, 0.474, -0.0591, 0.0, 0.4725, -0.057, 0.0, 0.4725, -0.0548, 0.0, 0.474, -0.057, 0.0, 0.474, -0.0548, -0.0017, 0.4725, -0.0613, -0.0017, 0.4725, -0.0591, -0.0039, 0.4725, -0.0613, -0.0039, 0.4725, -0.0591, -0.0017, 0.4725, -0.057, -0.0017, 0.4725, -0.0548,
-0.0039, 0.4725, -0.057, -0.0039, 0.4725, -0.0548, 0.0, 0.4765, -0.0613, 0.0, 0.4765, -0.0591, 0.0, 0.4765, -0.057, 0.0, 0.4765, -0.0548, -0.0017, 0.4765, -0.0613, -0.0017, 0.4765, -0.0591, -0.0039, 0.4765, -0.0613, -0.0039, 0.4765, -0.0591, -0.0017, 0.4765, -0.057, -0.0017, 0.4765, -0.0548, -0.0039, 0.4765, -0.057, -0.0039, 0.4765, -0.0548, 0.0, 0.4804, -0.0591, 0.0, 0.4826, -0.0591, 0.0, 0.4804, -0.0562, 0.0, 0.4826, -0.0562, -0.0017, 0.4804, -0.0591, -0.0039, 0.4804, -0.0591, -0.0017, 0.4826, -0.0591, -0.0039, 0.4826, -0.0591, -0.0017, 0.4804, -0.0562, -0.0039, 0.4804, -0.0562, -0.0017, 0.4826, -0.0562, -0.0039, 0.4826, -0.0562, 0.0, 0.4848, -0.0592, 0.0, 0.4868, -0.0596, 0.0, 0.4848, -0.0562, 0.0, 0.487, -0.0562,
-0.0017, 0.4848, -0.0592, -0.0039, 0.4848, -0.0592, -0.0017, 0.487, -0.0596, -0.0039, 0.487, -0.0596, -0.0017, 0.4848, -0.0562, -0.0039, 0.4848, -0.0562, -0.0017, 0.487, -0.0562, -0.0039, 0.487, -0.0562, 0.0, 0.4725, -0.0526, 0.0, 0.4725, -0.0505, 0.0, 0.474, -0.0526, 0.0, 0.474, -0.0505, 0.0, 0.4725, -0.0483, 0.0, 0.4725, -0.0461, 0.0, 0.474, -0.0483, 0.0, 0.474, -0.0461, -0.0017, 0.4725, -0.0526, -0.0017, 0.4725, -0.0505, -0.0039, 0.4725, -0.0526, -0.0039, 0.4725, -0.0505, -0.0017, 0.4725, -0.0483, -0.0017, 0.4725, -0.0461, -0.0039, 0.4725, -0.0483, -0.0039, 0.4725, -0.0461, 0.0, 0.4765, -0.0527, 0.0, 0.4766, -0.0506, 0.0, 0.477, -0.0485, 0.0, 0.4761, -0.0461, 0.0, 0.4775, -0.0476, 0.0, 0.4786, -0.0465,
-0.0017, 0.4765, -0.0526, -0.0017, 0.4766, -0.0504, -0.0039, 0.4765, -0.0526, -0.0039, 0.4766, -0.0504, -0.0017, 0.4769, -0.0487, -0.0039, 0.4769, -0.0487, -0.0017, 0.4775, -0.0476, -0.0017, 0.478, -0.047, -0.0039, 0.4775, -0.0476, -0.0039, 0.478, -0.047, 0.0, 0.4725, -0.0439, 0.0, 0.4725, -0.0425, 0.0, 0.474, -0.0439, 0.0, 0.474, -0.0425, -0.0017, 0.4725, -0.0439, -0.0017, 0.4725, -0.0425, -0.0039, 0.4725, -0.0439, -0.0039, 0.4725, -0.0425, -0.0017, 0.474, -0.0425, -0.0039, 0.474, -0.0425, 0.0, 0.4761, -0.0439, 0.0, 0.4761, -0.0425, 0.0, 0.4783, -0.0439, 0.0, 0.4783, -0.0425, -0.0017, 0.4761, -0.0425, -0.0039, 0.4761, -0.0425, -0.0017, 0.4783, -0.0425, -0.0039, 0.4783, -0.0425, 0.0, 0.4801, -0.0456, 0.0, 0.4827, -0.0455,
-0.0017, 0.4801, -0.0456, -0.0039, 0.4801, -0.0456, -0.0017, 0.4826, -0.0455, -0.0039, 0.4826, -0.0455, 0.0, 0.4848, -0.0454, 0.0, 0.487, -0.0454, -0.0017, 0.4848, -0.0454, -0.0039, 0.4848, -0.0454, -0.0017, 0.487, -0.0454, -0.0039, 0.487, -0.0454, 0.0, 0.4805, -0.0439, 0.0, 0.4805, -0.0425, 0.0, 0.4826, -0.0439, 0.0, 0.4826, -0.0425, -0.0017, 0.4805, -0.0425, -0.0039, 0.4805, -0.0425, -0.0017, 0.4826, -0.0425, -0.0039, 0.4826, -0.0425, 0.0, 0.4848, -0.0439, 0.0, 0.4848, -0.0425, 0.0, 0.487, -0.0439, 0.0, 0.487, -0.0425, -0.0017, 0.4848, -0.0425, -0.0039, 0.4848, -0.0425, -0.0017, 0.487, -0.0425, -0.0039, 0.487, -0.0425, 0.0, 0.4544, -0.0219, 0.0, 0.4544, -0.0201, 0.0, 0.4566, -0.0219, 0.0, 0.4566, -0.0201,
-0.0017, 0.4544, -0.0219, -0.0039, 0.4544, -0.0219, -0.0017, 0.4566, -0.0219, -0.0039, 0.4566, -0.0219, 0.0, 0.4588, -0.0219, 0.0, 0.4588, -0.0201, 0.0, 0.4609, -0.0219, 0.0, 0.4609, -0.0201, -0.0017, 0.4588, -0.0219, -0.0039, 0.4588, -0.0219, -0.0017, 0.4609, -0.0219, -0.0039, 0.4609, -0.0219, 0.0, 0.4631, -0.0219, 0.0, 0.4631, -0.0201, 0.0, 0.4653, -0.0219, 0.0, 0.4653, -0.0201, -0.0017, 0.4631, -0.0219, -0.0039, 0.4631, -0.0219, -0.0017, 0.4653, -0.0219, -0.0039, 0.4653, -0.0219, 0.0, 0.4675, -0.0219, 0.0, 0.4675, -0.0201, 0.0, 0.4696, -0.0219, 0.0, 0.4696, -0.0201, -0.0017, 0.4675, -0.0219, -0.0039, 0.4675, -0.0219, -0.0017, 0.4696, -0.0219, -0.0039, 0.4696, -0.0219, 0.0, 0.4544, -0.0179, 0.0, 0.4544, -0.0157,
0.0, 0.4566, -0.0179, 0.0, 0.4566, -0.0157, 0.0, 0.4544, -0.0136, 0.0, 0.4544, -0.0114, 0.0, 0.4566, -0.0136, 0.0, 0.4566, -0.0114, 0.0, 0.4588, -0.0179, 0.0, 0.4588, -0.0157, 0.0, 0.4609, -0.0179, 0.0, 0.4609, -0.0157, 0.0, 0.4588, -0.0136, 0.0, 0.4588, -0.0114, 0.0, 0.4609, -0.0136, 0.0, 0.4609, -0.0114, 0.0, 0.4544, -0.0101, 0.0, 0.4566, -0.0101, -0.0017, 0.4544, -0.0101, -0.0039, 0.4544, -0.0101, -0.0017, 0.4566, -0.0101, -0.0039, 0.4566, -0.0101, 0.0, 0.4588, -0.0101, 0.0, 0.4609, -0.0101, -0.0017, 0.4588, -0.0101, -0.0039, 0.4588, -0.0101, -0.0017, 0.4609, -0.0101, -0.0039, 0.4609, -0.0101, 0.0, 0.4631, -0.0179, 0.0, 0.4631, -0.0157, 0.0, 0.4653, -0.0179, 0.0, 0.4653, -0.0157,
0.0, 0.4631, -0.0136, 0.0, 0.4631, -0.0114, 0.0, 0.4653, -0.0136, 0.0, 0.4653, -0.0114, 0.0, 0.4675, -0.0179, 0.0, 0.4675, -0.0157, 0.0, 0.4696, -0.0179, 0.0, 0.4696, -0.0157, 0.0, 0.4675, -0.0136, 0.0, 0.4675, -0.0114, 0.0, 0.4696, -0.0136, 0.0, 0.4696, -0.0114, 0.0, 0.4631, -0.0101, 0.0, 0.4653, -0.0101, -0.0017, 0.4631, -0.0101, -0.0039, 0.4631, -0.0101, -0.0017, 0.4653, -0.0101, -0.0039, 0.4653, -0.0101, 0.0, 0.4675, -0.0101, 0.0, 0.4696, -0.0101, -0.0017, 0.4675, -0.0101, -0.0039, 0.4675, -0.0101, -0.0017, 0.4696, -0.0101, -0.0039, 0.4696, -0.0101, 0.0, 0.4718, -0.0219, 0.0, 0.4718, -0.0201, 0.0, 0.474, -0.0219, 0.0, 0.474, -0.0201, -0.0017, 0.4718, -0.0219, -0.0039, 0.4718, -0.0219,
-0.0017, 0.474, -0.0219, -0.0039, 0.474, -0.0219, 0.0, 0.4761, -0.0219, 0.0, 0.4761, -0.0201, 0.0, 0.4783, -0.0218, 0.0, 0.4783, -0.0201, -0.0017, 0.4761, -0.0219, -0.0039, 0.4761, -0.0219, -0.0017, 0.4783, -0.0218, -0.0039, 0.4783, -0.0218, 0.0, 0.4805, -0.0216, 0.0, 0.4805, -0.0201, 0.0, 0.4821, -0.0213, 0.0, 0.4829, -0.0212, -0.0017, 0.4805, -0.0216, -0.0039, 0.4805, -0.0216, -0.0017, 0.4823, -0.0213, -0.0017, 0.4834, -0.0211, -0.0039, 0.4823, -0.0213, -0.0039, 0.4834, -0.0211, 0.0, 0.485, -0.0207, 0.0, 0.4872, -0.0202, -0.0017, 0.4848, -0.0208, -0.0039, 0.4848, -0.0208, -0.0017, 0.487, -0.0203, -0.0039, 0.487, -0.0203, 0.0, 0.4718, -0.0179, 0.0, 0.4718, -0.0157, 0.0, 0.474, -0.0179, 0.0, 0.474, -0.0157,
0.0, 0.4718, -0.0136, 0.0, 0.4718, -0.0114, 0.0, 0.474, -0.0136, 0.0, 0.474, -0.0114, 0.0, 0.4761, -0.0179, 0.0, 0.4761, -0.0157, 0.0, 0.4783, -0.0179, 0.0, 0.4783, -0.0157, 0.0, 0.4761, -0.0136, 0.0, 0.4761, -0.0114, 0.0, 0.4783, -0.0136, 0.0, 0.4783, -0.0114, 0.0, 0.4718, -0.0101, 0.0, 0.474, -0.0101, -0.0017, 0.4718, -0.0101, -0.0039, 0.4718, -0.0101, -0.0017, 0.474, -0.0101, -0.0039, 0.474, -0.0101, 0.0, 0.4761, -0.0101, 0.0, 0.4783, -0.01, -0.0017, 0.4761, -0.0101, -0.0039, 0.4761, -0.0101, -0.0017, 0.4783, -0.01, -0.0039, 0.4783, -0.01, 0.0, 0.4805, -0.0179, 0.0, 0.4805, -0.0157, 0.0, 0.4826, -0.0179, 0.0, 0.4826, -0.0157, 0.0, 0.4805, -0.0136, 0.0, 0.4805, -0.0114,
0.0, 0.4826, -0.0136, 0.0, 0.4826, -0.0114, 0.0, 0.4848, -0.0179, 0.0, 0.4848, -0.0157, 0.0, 0.487, -0.0179, 0.0, 0.487, -0.0157, 0.0, 0.4848, -0.0136, 0.0, 0.4848, -0.0114, 0.0, 0.487, -0.0136, 0.0, 0.487, -0.0114, 0.0, 0.4804, -0.0096, 0.0, 0.4824, -0.0091, -0.0017, 0.4805, -0.0096, -0.0039, 0.4805, -0.0096, -0.0017, 0.4827, -0.009, -0.0039, 0.4827, -0.009, 0.0, 0.4843, -0.0082, 0.0, 0.4853, -0.0077, 0.0, 0.487, -0.0092, 0.0, 0.4869, -0.0068, 0.0, 0.4879, -0.0058, -0.0017, 0.4841, -0.0083, -0.0017, 0.4852, -0.0078, -0.0039, 0.4841, -0.0083, -0.0039, 0.4852, -0.0078, -0.0017, 0.4869, -0.0067, -0.0039, 0.4869, -0.0067, -0.0017, 0.4879, -0.0058, -0.0039, 0.4879, -0.0058, 0.0, 0.4919, -0.0629,
-0.0017, 0.4919, -0.0628, -0.0039, 0.4919, -0.0628, 0.0, 0.4943, -0.0648, 0.0, 0.4932, -0.064, 0.0, 0.4954, -0.0657, 0.0, 0.4957, -0.0635, -0.0017, 0.4943, -0.0648, -0.0017, 0.4932, -0.0639, -0.0039, 0.4943, -0.0648, -0.0039, 0.4932, -0.0639, -0.0017, 0.4956, -0.0658, -0.0039, 0.4956, -0.0658, 0.0, 0.4895, -0.0608, 0.0, 0.4886, -0.0603, 0.0, 0.4909, -0.0619, 0.0, 0.4913, -0.0591, 0.0, 0.4891, -0.0562, 0.0, 0.4912, -0.056, 0.0, 0.4921, -0.0558, -0.0017, 0.4894, -0.0608, -0.0017, 0.4882, -0.0601, -0.0039, 0.4894, -0.0608, -0.0039, 0.4882, -0.0601, -0.0017, 0.4908, -0.0619, -0.0039, 0.4908, -0.0619, -0.0017, 0.4892, -0.0562, -0.0039, 0.4892, -0.0562, -0.0017, 0.4909, -0.056, -0.0017, 0.492, -0.0558, -0.0039, 0.4909, -0.056,
-0.0039, 0.492, -0.0558, 0.0, 0.4935, -0.0613, 0.0, 0.4935, -0.0591, 0.0, 0.4963, -0.0609, 0.0, 0.4953, -0.0594, 0.0, 0.4935, -0.057, 0.0, 0.4934, -0.0553, 0.0, 0.4957, -0.057, 0.0, 0.4952, -0.0543, -0.0017, 0.4963, -0.0608, -0.0017, 0.4953, -0.0594, -0.0039, 0.4963, -0.0608, -0.0039, 0.4953, -0.0594, -0.0017, 0.4935, -0.0553, -0.0039, 0.4935, -0.0553, -0.0017, 0.4953, -0.0543, -0.0039, 0.4953, -0.0543, 0.0, 0.4987, -0.0668, 0.0, 0.5, -0.0669, 0.0, 0.4977, -0.0666, 0.0, 0.4983, -0.063, 0.0, 0.5005, -0.0652, 0.0, 0.4994, -0.0641, -0.0017, 0.4986, -0.0668, -0.0039, 0.4986, -0.0668, -0.0017, 0.5, -0.0669, -0.0039, 0.5, -0.0669, -0.0017, 0.4975, -0.0665, -0.0017, 0.4983, -0.063, -0.0039, 0.4975, -0.0665,
-0.0039, 0.4983, -0.063, -0.0017, 0.5005, -0.0652, -0.0017, 0.4994, -0.0641, -0.0039, 0.5005, -0.0652, -0.0039, 0.4994, -0.0641, 0.0, 0.502, -0.067, 0.0, 0.5014, -0.0664, -0.0017, 0.502, -0.067, -0.0039, 0.502, -0.067, -0.0017, 0.5014, -0.0663, -0.0039, 0.5014, -0.0663, 0.0, 0.4972, -0.0619, 0.0, 0.497, -0.0582, 0.0, 0.4981, -0.0574, 0.0, 0.4978, -0.0548, 0.0, 0.4994, -0.0563, 0.0, 0.5006, -0.0549, -0.0017, 0.4972, -0.0619, -0.0017, 0.497, -0.0582, -0.0039, 0.4972, -0.0619, -0.0039, 0.497, -0.0582, -0.0017, 0.4981, -0.0574, -0.0039, 0.4981, -0.0574, -0.0017, 0.4994, -0.0563, -0.0017, 0.5006, -0.0549, -0.0039, 0.4994, -0.0563, -0.0039, 0.5006, -0.0549, 0.0, 0.5011, -0.0537, -0.0017, 0.5011, -0.0538, -0.0039, 0.5011, -0.0538,
0.0, 0.4891, -0.0455, 0.0, 0.4912, -0.0459, -0.0017, 0.4892, -0.0455, -0.0039, 0.4892, -0.0455, -0.0017, 0.4914, -0.046, -0.0039, 0.4914, -0.046, 0.0, 0.4964, -0.0526, 0.0, 0.497, -0.0503, 0.0, 0.4943, -0.0474, 0.0, 0.4933, -0.0468, 0.0, 0.4955, -0.0484, 0.0, 0.4957, -0.0461, -0.0017, 0.4962, -0.053, -0.0017, 0.497, -0.0503, -0.0039, 0.4962, -0.053, -0.0039, 0.497, -0.0503, -0.0017, 0.4943, -0.0474, -0.0017, 0.4932, -0.0468, -0.0039, 0.4943, -0.0474, -0.0039, 0.4932, -0.0468, -0.0017, 0.4955, -0.0484, -0.0039, 0.4955, -0.0484, 0.0, 0.4892, -0.0439, 0.0, 0.4892, -0.0425, 0.0, 0.4913, -0.0439, 0.0, 0.4913, -0.0425, -0.0017, 0.4892, -0.0425, -0.0039, 0.4892, -0.0425, -0.0017, 0.4913, -0.0425, -0.0039, 0.4913, -0.0425,
0.0, 0.4935, -0.0439, 0.0, 0.4935, -0.0426, 0.0, 0.4958, -0.0431, 0.0, 0.4946, -0.0428, -0.0017, 0.4935, -0.0426, -0.0039, 0.4935, -0.0426, -0.0017, 0.4958, -0.0431, -0.0017, 0.4947, -0.0428, -0.0039, 0.4958, -0.0431, -0.0039, 0.4947, -0.0428, 0.0, 0.4978, -0.0526, 0.0, 0.4978, -0.0505, 0.0, 0.5, -0.0526, 0.0, 0.5, -0.0505, 0.0, 0.4978, -0.0483, 0.0, 0.4978, -0.0461, 0.0, 0.5011, -0.0477, 0.0, 0.4998, -0.0458, -0.0017, 0.501, -0.0475, -0.0017, 0.5, -0.046, -0.0039, 0.501, -0.0475, -0.0039, 0.5, -0.046, 0.0, 0.5015, -0.0527, 0.0, 0.5018, -0.0505, 0.0, 0.5015, -0.0488, -0.0017, 0.5015, -0.0527, -0.0017, 0.5018, -0.0505, -0.0039, 0.5015, -0.0527, -0.0039, 0.5018, -0.0505, -0.0017, 0.5015, -0.0486,
-0.0039, 0.5015, -0.0486, 0.0, 0.4981, -0.0442, 0.0, 0.499, -0.045, -0.0017, 0.4979, -0.0441, -0.0039, 0.4979, -0.0441, -0.0017, 0.499, -0.045, -0.0039, 0.499, -0.045, 0.0, 0.4889, -0.0196, -0.0017, 0.4892, -0.0195, -0.0039, 0.4892, -0.0195, 0.0, 0.4898, -0.0192, 0.0, 0.4892, -0.0157, 0.0, 0.4917, -0.0183, 0.0, 0.4913, -0.0157, 0.0, 0.4892, -0.0136, 0.0, 0.4892, -0.0114, 0.0, 0.4913, -0.0136, 0.0, 0.4913, -0.0114, -0.0017, 0.4902, -0.019, -0.0039, 0.4902, -0.019, -0.0017, 0.4914, -0.0185, -0.0039, 0.4914, -0.0185, 0.0, 0.4931, -0.0174, 0.0, 0.4942, -0.0166, 0.0, 0.4955, -0.0156, 0.0, 0.4935, -0.0136, 0.0, 0.4935, -0.0114, 0.0, 0.4966, -0.0146, 0.0, 0.4957, -0.0114, -0.0017, 0.4932, -0.0174,
-0.0017, 0.4943, -0.0166, -0.0039, 0.4932, -0.0174, -0.0039, 0.4943, -0.0166, -0.0017, 0.4956, -0.0155, -0.0039, 0.4956, -0.0155, -0.0017, 0.4966, -0.0145, -0.0039, 0.4966, -0.0145, 0.0, 0.4892, -0.0092, 0.0, 0.4892, -0.0071, 0.0, 0.4913, -0.0092, 0.0, 0.4913, -0.0071, 0.0, 0.489, -0.0048, 0.0, 0.4899, -0.0034, 0.0, 0.4913, -0.0049, 0.0, 0.4904, -0.0025, -0.0017, 0.489, -0.0048, -0.0017, 0.49, -0.0034, -0.0039, 0.489, -0.0048, -0.0039, 0.49, -0.0034, -0.0017, 0.4905, -0.0023, -0.0039, 0.4905, -0.0023, 0.0, 0.4935, -0.0092, 0.0, 0.4935, -0.0071, 0.0, 0.4957, -0.0092, 0.0, 0.4957, -0.0071, 0.0, 0.4935, -0.0049, 0.0, 0.4935, -0.0027, 0.0, 0.4957, -0.0049, 0.0, 0.4957, -0.0027, 0.0, 0.4977, -0.0135,
0.0, 0.4988, -0.0121, 0.0, 0.4995, -0.011, -0.0017, 0.4977, -0.0135, -0.0017, 0.4987, -0.0122, -0.0039, 0.4977, -0.0135, -0.0039, 0.4987, -0.0122, -0.0017, 0.4994, -0.0111, -0.0039, 0.4994, -0.0111, 0.0, 0.4978, -0.0092, 0.0, 0.4978, -0.0071, 0.0, 0.5004, -0.0096, 0.0, 0.5012, -0.0076, 0.0, 0.4978, -0.0049, 0.0, 0.4978, -0.0027, 0.0, 0.5, -0.0049, 0.0, 0.5, -0.0027, -0.0017, 0.5005, -0.0093, -0.0017, 0.501, -0.008, -0.0039, 0.5005, -0.0093, -0.0039, 0.501, -0.008, 0.0, 0.5015, -0.0067, 0.0, 0.5018, -0.0049, 0.0, 0.5019, -0.0027, -0.0017, 0.5014, -0.0069, -0.0039, 0.5014, -0.0069, -0.0017, 0.5018, -0.0049, -0.0017, 0.5019, -0.0027, -0.0039, 0.5018, -0.0049, -0.0039, 0.5019, -0.0027, 0.0, 0.0031, -0.0013,
0.0, 0.003, 0.0016, 0.0, 0.0053, -0.0012, 0.0, 0.0052, 0.0016, 0.0, 0.003, 0.0038, 0.0, 0.003, 0.006, 0.0, 0.0052, 0.0038, 0.0, 0.0052, 0.006, -0.0017, 0.003, -0.0013, -0.0039, 0.003, -0.0013, -0.0017, 0.0052, -0.0012, -0.0039, 0.0052, -0.0012, 0.0, 0.0075, -0.0009, 0.0, 0.0074, 0.0016, 0.0, 0.0097, -0.0004, 0.0, 0.0096, 0.0016, 0.0, 0.0074, 0.0038, 0.0, 0.0074, 0.006, 0.0, 0.0096, 0.0038, 0.0, 0.0096, 0.006, -0.0017, 0.0074, -0.0009, -0.0039, 0.0074, -0.0009, -0.0017, 0.0096, -0.0005, -0.0039, 0.0096, -0.0005, 0.0, 0.003, 0.0081, 0.0, 0.003, 0.0105, 0.0, 0.0052, 0.0081, 0.0, 0.0051, 0.0105, -0.0017, 0.003, 0.0105, -0.0039, 0.003, 0.0105, -0.0017, 0.0052, 0.0106,
-0.0039, 0.0052, 0.0106, 0.0, 0.0074, 0.0081, 0.0, 0.0072, 0.0109, 0.0, 0.0096, 0.0081, 0.0, 0.0089, 0.0115, 0.0, 0.0099, 0.0119, -0.0017, 0.0074, 0.0109, -0.0039, 0.0074, 0.0109, -0.0017, 0.0085, 0.0114, -0.0039, 0.0085, 0.0114, -0.0017, 0.0096, 0.0118, -0.0039, 0.0096, 0.0118, 0.0, 0.0114, 0.0001, 0.0, 0.0123, 0.0004, 0.0, 0.0142, 0.0012, 0.0, 0.0117, 0.0038, 0.0, 0.0117, 0.006, 0.0, 0.0139, 0.0038, 0.0, 0.0139, 0.006, -0.0017, 0.0116, 0.0002, -0.0017, 0.0127, 0.0006, -0.0039, 0.0116, 0.0002, -0.0039, 0.0127, 0.0006, -0.0017, 0.0139, 0.0011, -0.0039, 0.0139, 0.0011, 0.0, 0.0162, 0.0022, 0.0, 0.0172, 0.0027, 0.0, 0.0161, 0.0038, 0.0, 0.0161, 0.006, 0.0, 0.0183, 0.0034,
0.0, 0.0182, 0.006, -0.0017, 0.0161, 0.0021, -0.0039, 0.0161, 0.0021, -0.0017, 0.0172, 0.0027, -0.0039, 0.0172, 0.0027, -0.0017, 0.0183, 0.0033, -0.0039, 0.0183, 0.0033, 0.0, 0.0117, 0.0081, 0.0, 0.0117, 0.0103, 0.0, 0.0139, 0.0081, 0.0, 0.0139, 0.0103, 0.0, 0.0115, 0.0129, 0.0, 0.0126, 0.0137, 0.0, 0.0139, 0.0125, 0.0, 0.0138, 0.0147, -0.0017, 0.0116, 0.0129, -0.0017, 0.0126, 0.0137, -0.0039, 0.0116, 0.0129, -0.0039, 0.0126, 0.0137, -0.0017, 0.0138, 0.0147, -0.0039, 0.0138, 0.0147, 0.0, 0.0161, 0.0081, 0.0, 0.0161, 0.0103, 0.0, 0.0182, 0.0081, 0.0, 0.0182, 0.0103, 0.0, 0.0161, 0.0125, 0.0, 0.0161, 0.0146, 0.0, 0.0182, 0.0125, 0.0, 0.0182, 0.0146, 0.0, 0.0148, 0.0159,
-0.0017, 0.0148, 0.016, -0.0039, 0.0148, 0.016, 0.0, 0.0156, 0.017, 0.0, 0.0167, 0.019, 0.0, 0.0182, 0.0168, 0.0, 0.0182, 0.019, 0.0, 0.0171, 0.0203, 0.0, 0.0173, 0.0211, 0.0, 0.0176, 0.0233, -0.0017, 0.0156, 0.017, -0.0017, 0.0167, 0.0189, -0.0039, 0.0156, 0.017, -0.0039, 0.0167, 0.0189, -0.0017, 0.0171, 0.0203, -0.0039, 0.0171, 0.0203, -0.0017, 0.0174, 0.0214, -0.0017, 0.0176, 0.0233, -0.0039, 0.0174, 0.0214, -0.0039, 0.0176, 0.0233, 0.0, 0.0144, 0.0325, -0.0017, 0.0144, 0.0324, -0.0039, 0.0144, 0.0324, 0.0, 0.0169, 0.0282, 0.0, 0.0175, 0.0254, 0.0, 0.0171, 0.0274, 0.0, 0.0163, 0.0297, 0.0, 0.0152, 0.0315, 0.0, 0.0182, 0.0298, 0.0, 0.0182, 0.032, -0.0017, 0.017, 0.028,
-0.0039, 0.017, 0.028, -0.0017, 0.0175, 0.0255, -0.0017, 0.0172, 0.0269, -0.0039, 0.0175, 0.0255, -0.0039, 0.0172, 0.0269, -0.0017, 0.0162, 0.0299, -0.0017, 0.0153, 0.0313, -0.0039, 0.0162, 0.0299, -0.0039, 0.0153, 0.0313, 0.0, 0.0198, 0.0044, 0.0, 0.021, 0.0053, 0.0, 0.0221, 0.0064, -0.0017, 0.0199, 0.0044, -0.0017, 0.021, 0.0053, -0.0039, 0.0199, 0.0044, -0.0039, 0.021, 0.0053, -0.0017, 0.0221, 0.0064, -0.0039, 0.0221, 0.0064, 0.0, 0.0204, 0.0081, 0.0, 0.0204, 0.0103, 0.0, 0.0232, 0.0076, 0.0, 0.0226, 0.0103, 0.0, 0.0204, 0.0125, 0.0, 0.0204, 0.0146, 0.0, 0.0226, 0.0125, 0.0, 0.0226, 0.0146, -0.0017, 0.0232, 0.0076, -0.0039, 0.0232, 0.0076, 0.0, 0.0241, 0.0087, 0.0, 0.0252, 0.0102,
0.0, 0.0259, 0.0113, 0.0, 0.0247, 0.0125, 0.0, 0.0247, 0.0146, 0.0, 0.0265, 0.0123, 0.0, 0.0273, 0.0144, -0.0017, 0.0241, 0.0087, -0.0017, 0.0252, 0.0102, -0.0039, 0.0241, 0.0087, -0.0039, 0.0252, 0.0102, -0.0017, 0.0259, 0.0113, -0.0039, 0.0259, 0.0113, -0.0017, 0.0265, 0.0124, -0.0017, 0.0274, 0.0146, -0.0039, 0.0265, 0.0124, -0.0039, 0.0274, 0.0146, 0.0, 0.0204, 0.0168, 0.0, 0.0204, 0.019, 0.0, 0.0226, 0.0168, 0.0, 0.0226, 0.019, 0.0, 0.0204, 0.0212, 0.0, 0.0204, 0.0233, 0.0, 0.0226, 0.0212, 0.0, 0.0226, 0.0233, 0.0, 0.0247, 0.0168, 0.0, 0.0247, 0.019, 0.0, 0.028, 0.0166, 0.0, 0.0269, 0.019, 0.0, 0.0247, 0.0212, 0.0, 0.0247, 0.0233, 0.0, 0.0269, 0.0212,
0.0, 0.0269, 0.0233, -0.0017, 0.0279, 0.0163, -0.0039, 0.0279, 0.0163, 0.0, 0.0204, 0.0255, 0.0, 0.0204, 0.0277, 0.0, 0.0226, 0.0255, 0.0, 0.0226, 0.0277, 0.0, 0.0204, 0.0298, 0.0, 0.0204, 0.032, 0.0, 0.0226, 0.0298, 0.0, 0.0226, 0.032, 0.0, 0.0247, 0.0255, 0.0, 0.0247, 0.0277, 0.0, 0.0269, 0.0255, 0.0, 0.0269, 0.0277, 0.0, 0.0247, 0.0298, 0.0, 0.0247, 0.032, 0.0, 0.0281, 0.0302, 0.0, 0.0276, 0.0323, -0.0017, 0.028, 0.0308, -0.0017, 0.0277, 0.032, -0.0039, 0.028, 0.0308, -0.0039, 0.0277, 0.032, 0.0, 0.0281, 0.0175, 0.0, 0.0283, 0.019, 0.0, 0.0284, 0.0212, 0.0, 0.0284, 0.0233, -0.0017, 0.0281, 0.0174, -0.0017, 0.0283, 0.019, -0.0039, 0.0281, 0.0174,
-0.0039, 0.0283, 0.019, -0.0017, 0.0284, 0.0212, -0.0017, 0.0284, 0.0233, -0.0039, 0.0284, 0.0212, -0.0039, 0.0284, 0.0233, 0.0, 0.0284, 0.0255, 0.0, 0.0284, 0.0277, 0.0, 0.0282, 0.0294, -0.0017, 0.0284, 0.0255, -0.0017, 0.0284, 0.0277, -0.0039, 0.0284, 0.0255, -0.0039, 0.0284, 0.0277, -0.0017, 0.0282, 0.0297, -0.0039, 0.0282, 0.0297, 0.0, 0.0061, 0.0373, 0.0, 0.003, 0.0378, 0.0, 0.003, 0.0407, 0.0, 0.0052, 0.0376, 0.0, 0.0052, 0.0407, -0.0017, 0.006, 0.0374, -0.0039, 0.006, 0.0374, -0.0017, 0.003, 0.0378, -0.0039, 0.003, 0.0378, -0.0017, 0.0049, 0.0376, -0.0039, 0.0049, 0.0376, 0.0, 0.0073, 0.037, 0.0, 0.0093, 0.0362, 0.0, 0.0074, 0.0385, 0.0, 0.0074, 0.0407, 0.0, 0.0096, 0.0385,
0.0, 0.0096, 0.0407, -0.0017, 0.0074, 0.037, -0.0039, 0.0074, 0.037, -0.0017, 0.0096, 0.0361, -0.0039, 0.0096, 0.0361, 0.0, 0.003, 0.0429, 0.0, 0.003, 0.045, 0.0, 0.0052, 0.0429, 0.0, 0.0052, 0.045, 0.0, 0.003, 0.0472, 0.0, 0.003, 0.0487, 0.0, 0.0052, 0.0472, 0.0, 0.0052, 0.0487, -0.0017, 0.003, 0.0487, -0.0039, 0.003, 0.0487, -0.0017, 0.0052, 0.0487, -0.0039, 0.0052, 0.0487, 0.0, 0.0074, 0.0429, 0.0, 0.0074, 0.045, 0.0, 0.0096, 0.0429, 0.0, 0.0096, 0.045, 0.0, 0.0078, 0.0484, 0.0, 0.007, 0.0485, 0.0, 0.0098, 0.048, -0.0017, 0.0084, 0.0483, -0.0017, 0.0073, 0.0485, -0.0039, 0.0084, 0.0483, -0.0039, 0.0073, 0.0485, -0.0017, 0.0096, 0.048, -0.0039, 0.0096, 0.048,
0.0, 0.012, 0.0346, 0.0, 0.011, 0.0352, 0.0, 0.0133, 0.0335, 0.0, 0.0139, 0.0363, 0.0, 0.0117, 0.0385, 0.0, 0.0117, 0.0407, 0.0, 0.0139, 0.0385, 0.0, 0.0139, 0.0407, -0.0017, 0.012, 0.0347, -0.0017, 0.0108, 0.0354, -0.0039, 0.012, 0.0347, -0.0039, 0.0108, 0.0354, -0.0017, 0.0133, 0.0335, -0.0039, 0.0133, 0.0335, 0.0, 0.0161, 0.0342, 0.0, 0.0161, 0.0363, 0.0, 0.0182, 0.0342, 0.0, 0.0182, 0.0363, 0.0, 0.0161, 0.0385, 0.0, 0.0161, 0.0407, 0.0, 0.0182, 0.0385, 0.0, 0.0182, 0.0407, 0.0, 0.0117, 0.0429, 0.0, 0.0117, 0.045, 0.0, 0.0139, 0.0429, 0.0, 0.0146, 0.0463, 0.0, 0.0119, 0.0473, 0.0, 0.0136, 0.0467, -0.0017, 0.0149, 0.0461, -0.0039, 0.0149, 0.0461,
-0.0017, 0.0117, 0.0474, -0.0039, 0.0117, 0.0474, -0.0017, 0.0139, 0.0466, -0.0039, 0.0139, 0.0466, 0.0, 0.0161, 0.0429, 0.0, 0.0164, 0.0454, 0.0, 0.0189, 0.0438, 0.0, 0.0178, 0.0445, -0.0017, 0.0161, 0.0455, -0.0039, 0.0161, 0.0455, -0.0017, 0.019, 0.0437, -0.0017, 0.0179, 0.0445, -0.0039, 0.019, 0.0437, -0.0039, 0.0179, 0.0445, 0.0, 0.0204, 0.0342, 0.0, 0.0204, 0.0363, 0.0, 0.0226, 0.0342, 0.0, 0.0226, 0.0363, 0.0, 0.0204, 0.0385, 0.0, 0.0214, 0.0417, 0.0, 0.0236, 0.0394, 0.0, 0.0225, 0.0406, -0.0017, 0.0214, 0.0417, -0.0039, 0.0214, 0.0417, -0.0017, 0.0235, 0.0394, -0.0017, 0.0225, 0.0407, -0.0039, 0.0235, 0.0394, -0.0039, 0.0225, 0.0407, 0.0, 0.0247, 0.0342, 0.0, 0.0255, 0.0367,
0.0, 0.0268, 0.0344, 0.0, 0.0261, 0.0356, 0.0, 0.0244, 0.0383, -0.0017, 0.0254, 0.0368, -0.0039, 0.0254, 0.0368, -0.0017, 0.0269, 0.0342, -0.0017, 0.0261, 0.0357, -0.0039, 0.0269, 0.0342, -0.0039, 0.0261, 0.0357, -0.0017, 0.0244, 0.0384, -0.0039, 0.0244, 0.0384, 0.0, 0.0203, 0.0427, -0.0017, 0.0204, 0.0427, -0.0039, 0.0204, 0.0427, 0.0, 0.0716, -0.0005, 0.0, 0.0716, 0.0016, 0.0, 0.0747, -0.0005, 0.0, 0.0747, 0.0016, 0.0, 0.0716, 0.0038, 0.0, 0.0716, 0.006, 0.0, 0.0747, 0.0038, 0.0, 0.0747, 0.006, -0.0017, 0.0716, -0.0005, -0.0017, 0.0716, 0.0016, -0.0039, 0.0716, -0.0005, -0.0039, 0.0716, 0.0016, -0.0017, 0.0716, 0.0038, -0.0017, 0.0716, 0.006, -0.0039, 0.0716, 0.0038, -0.0039, 0.0716, 0.006,
0.0, 0.0768, -0.0005, 0.0, 0.0768, 0.0016, 0.0, 0.079, -0.0005, 0.0, 0.079, 0.0016, 0.0, 0.0768, 0.0038, 0.0, 0.0768, 0.006, 0.0, 0.079, 0.0038, 0.0, 0.079, 0.006, 0.0, 0.0716, 0.0081, 0.0, 0.0716, 0.0103, 0.0, 0.0747, 0.0081, 0.0, 0.0747, 0.0103, 0.0, 0.0716, 0.0125, 0.0, 0.0716, 0.0146, 0.0, 0.0747, 0.0125, 0.0, 0.0747, 0.0146, -0.0017, 0.0716, 0.0081, -0.0017, 0.0716, 0.0103, -0.0039, 0.0716, 0.0081, -0.0039, 0.0716, 0.0103, -0.0017, 0.0716, 0.0125, -0.0017, 0.0716, 0.0146, -0.0039, 0.0716, 0.0125, -0.0039, 0.0716, 0.0146, 0.0, 0.0768, 0.0081, 0.0, 0.0768, 0.0103, 0.0, 0.079, 0.0081, 0.0, 0.079, 0.0103, 0.0, 0.0768, 0.0125, 0.0, 0.0768, 0.0146,
0.0, 0.079, 0.0125, 0.0, 0.079, 0.0146, 0.0, 0.0812, -0.0005, 0.0, 0.0812, 0.0016, 0.0, 0.0833, -0.0005, 0.0, 0.0833, 0.0016, 0.0, 0.0812, 0.0038, 0.0, 0.0812, 0.006, 0.0, 0.0833, 0.0038, 0.0, 0.0833, 0.006, -0.0017, 0.0833, -0.0005, -0.0017, 0.0833, 0.0016, -0.0039, 0.0833, -0.0005, -0.0039, 0.0833, 0.0016, -0.0017, 0.0833, 0.0038, -0.0017, 0.0833, 0.006, -0.0039, 0.0833, 0.0038, -0.0039, 0.0833, 0.006, 0.0, 0.0812, 0.0081, 0.0, 0.0812, 0.0103, 0.0, 0.0833, 0.0085, 0.0, 0.0833, 0.0103, 0.0, 0.0812, 0.0125, 0.0, 0.0812, 0.0146, 0.0, 0.0833, 0.0125, 0.0, 0.0833, 0.0146, -0.0017, 0.0833, 0.0085, -0.0039, 0.0833, 0.0085, 0.0, 0.0855, 0.0085, 0.0, 0.0855, 0.0103,
0.0, 0.0877, 0.0085, 0.0, 0.0877, 0.0103, 0.0, 0.0855, 0.0125, 0.0, 0.0855, 0.0146, 0.0, 0.0877, 0.0125, 0.0, 0.0877, 0.0146, -0.0017, 0.0855, 0.0085, -0.0039, 0.0855, 0.0085, -0.0017, 0.0877, 0.0085, -0.0039, 0.0877, 0.0085, 0.0, 0.0716, 0.0168, 0.0, 0.0716, 0.019, 0.0, 0.0747, 0.0168, 0.0, 0.0747, 0.019, 0.0, 0.0716, 0.0212, 0.0, 0.0716, 0.0233, 0.0, 0.0747, 0.0212, 0.0, 0.0747, 0.0233, -0.0017, 0.0716, 0.0168, -0.0017, 0.0716, 0.019, -0.0039, 0.0716, 0.0168, -0.0039, 0.0716, 0.019, -0.0017, 0.0716, 0.0212, -0.0017, 0.0716, 0.0233, -0.0039, 0.0716, 0.0212, -0.0039, 0.0716, 0.0233, 0.0, 0.0768, 0.0168, 0.0, 0.0768, 0.019, 0.0, 0.079, 0.0168, 0.0, 0.079, 0.019,
0.0, 0.0768, 0.0212, 0.0, 0.0768, 0.0233, 0.0, 0.079, 0.0212, 0.0, 0.079, 0.0233, 0.0, 0.0716, 0.0255, 0.0, 0.0716, 0.0277, 0.0, 0.0747, 0.0255, 0.0, 0.0747, 0.0277, 0.0, 0.0716, 0.0298, 0.0, 0.0716, 0.032, 0.0, 0.0747, 0.0298, 0.0, 0.0747, 0.032, -0.0017, 0.0716, 0.0255, -0.0017, 0.0716, 0.0277, -0.0039, 0.0716, 0.0255, -0.0039, 0.0716, 0.0277, -0.0017, 0.0716, 0.0298, -0.0017, 0.0716, 0.032, -0.0039, 0.0716, 0.0298, -0.0039, 0.0716, 0.032, 0.0, 0.0768, 0.0255, 0.0, 0.0768, 0.0277, 0.0, 0.079, 0.0255, 0.0, 0.079, 0.0277, 0.0, 0.0768, 0.0298, 0.0, 0.0768, 0.032, 0.0, 0.079, 0.0298, 0.0, 0.079, 0.032, 0.0, 0.0812, 0.0168, 0.0, 0.0812, 0.019,
0.0, 0.0833, 0.0168, 0.0, 0.0833, 0.019, 0.0, 0.0812, 0.0212, 0.0, 0.0812, 0.0233, 0.0, 0.0833, 0.0203, 0.0, 0.0833, 0.0233, -0.0017, 0.0833, 0.0203, -0.0017, 0.0833, 0.0233, -0.0039, 0.0833, 0.0203, -0.0039, 0.0833, 0.0233, 0.0, 0.0855, 0.0168, 0.0, 0.0855, 0.019, 0.0, 0.0877, 0.0168, 0.0, 0.0877, 0.019, 0.0, 0.0855, 0.0203, 0.0, 0.0877, 0.0203, -0.0017, 0.0855, 0.0203, -0.0039, 0.0855, 0.0203, -0.0017, 0.0877, 0.0203, -0.0039, 0.0877, 0.0203, 0.0, 0.0812, 0.0255, 0.0, 0.0812, 0.0277, 0.0, 0.0833, 0.0255, 0.0, 0.0833, 0.0277, 0.0, 0.0812, 0.0298, 0.0, 0.0812, 0.032, 0.0, 0.0833, 0.0298, 0.0, 0.0833, 0.032, -0.0017, 0.0833, 0.0255, -0.0017, 0.0833, 0.0277,
-0.0039, 0.0833, 0.0255, -0.0039, 0.0833, 0.0277, -0.0017, 0.0833, 0.0298, -0.0017, 0.0833, 0.032, -0.0039, 0.0833, 0.0298, -0.0039, 0.0833, 0.032, 0.0, 0.0899, 0.0085, 0.0, 0.0899, 0.0103, 0.0, 0.092, 0.0085, 0.0, 0.092, 0.0103, 0.0, 0.0899, 0.0125, 0.0, 0.0899, 0.0146, 0.0, 0.092, 0.0125, 0.0, 0.092, 0.0146, -0.0017, 0.0899, 0.0085, -0.0039, 0.0899, 0.0085, -0.0017, 0.092, 0.0085, -0.0039, 0.092, 0.0085, 0.0, 0.0942, 0.0085, 0.0, 0.0942, 0.0103, 0.0, 0.0964, 0.0085, 0.0, 0.0964, 0.0103, 0.0, 0.0942, 0.0125, 0.0, 0.0942, 0.0146, 0.0, 0.0964, 0.0125, 0.0, 0.0964, 0.0146, -0.0017, 0.0942, 0.0085, -0.0039, 0.0942, 0.0085, -0.0017, 0.0964, 0.0085, -0.0039, 0.0964, 0.0085,
0.0, 0.0985, 0.0085, 0.0, 0.0985, 0.0103, 0.0, 0.1007, 0.0085, 0.0, 0.1007, 0.0103, 0.0, 0.0985, 0.0125, 0.0, 0.0985, 0.0146, 0.0, 0.1007, 0.0125, 0.0, 0.1007, 0.0146, -0.0017, 0.0985, 0.0085, -0.0039, 0.0985, 0.0085, -0.0017, 0.1007, 0.0085, -0.0039, 0.1007, 0.0085, 0.0, 0.1029, 0.0085, 0.0, 0.1029, 0.0103, 0.0, 0.105, 0.0085, 0.0, 0.105, 0.0103, 0.0, 0.1029, 0.0125, 0.0, 0.1029, 0.0146, 0.0, 0.105, 0.0125, 0.0, 0.105, 0.0146, -0.0017, 0.1029, 0.0085, -0.0039, 0.1029, 0.0085, -0.0017, 0.105, 0.0085, -0.0039, 0.105, 0.0085, 0.0, 0.0899, 0.0168, 0.0, 0.0899, 0.019, 0.0, 0.092, 0.0168, 0.0, 0.092, 0.019, 0.0, 0.0899, 0.0203, 0.0, 0.092, 0.0203,
-0.0017, 0.0899, 0.0203, -0.0039, 0.0899, 0.0203, -0.0017, 0.092, 0.0203, -0.0039, 0.092, 0.0203, 0.0, 0.0942, 0.0168, 0.0, 0.0942, 0.019, 0.0, 0.0964, 0.0168, 0.0, 0.0964, 0.019, 0.0, 0.0942, 0.0203, 0.0, 0.0964, 0.0203, -0.0017, 0.0942, 0.0203, -0.0039, 0.0942, 0.0203, -0.0017, 0.0964, 0.0203, -0.0039, 0.0964, 0.0203, 0.0, 0.0985, 0.0168, 0.0, 0.0985, 0.019, 0.0, 0.1007, 0.0168, 0.0, 0.1007, 0.019, 0.0, 0.0985, 0.0203, 0.0, 0.1007, 0.0203, -0.0017, 0.0985, 0.0203, -0.0039, 0.0985, 0.0203, -0.0017, 0.1007, 0.0203, -0.0039, 0.1007, 0.0203, 0.0, 0.1029, 0.0168, 0.0, 0.1029, 0.019, 0.0, 0.105, 0.0168, 0.0, 0.105, 0.019, 0.0, 0.1029, 0.0203, 0.0, 0.105, 0.0203,
-0.0017, 0.1029, 0.0203, -0.0039, 0.1029, 0.0203, -0.0017, 0.105, 0.0203, -0.0039, 0.105, 0.0203, 0.0, 0.0716, 0.0342, 0.0, 0.0716, 0.0363, 0.0, 0.0747, 0.0342, 0.0, 0.0747, 0.0363, 0.0, 0.0716, 0.0385, 0.0, 0.0716, 0.0407, 0.0, 0.0747, 0.0385, 0.0, 0.0747, 0.0407, -0.0017, 0.0716, 0.0342, -0.0017, 0.0716, 0.0363, -0.0039, 0.0716, 0.0342, -0.0039, 0.0716, 0.0363, -0.0017, 0.0716, 0.0385, -0.0017, 0.0716, 0.0407, -0.0039, 0.0716, 0.0385, -0.0039, 0.0716, 0.0407, 0.0, 0.0768, 0.0342, 0.0, 0.0768, 0.0363, 0.0, 0.079, 0.0342, 0.0, 0.079, 0.0363, 0.0, 0.0768, 0.0385, 0.0, 0.0768, 0.0407, 0.0, 0.079, 0.0385, 0.0, 0.079, 0.0407, 0.0, 0.0716, 0.0429, 0.0, 0.0716, 0.045,
0.0, 0.0747, 0.0429, 0.0, 0.0747, 0.045, 0.0, 0.0716, 0.0472, 0.0, 0.0716, 0.0487, 0.0, 0.0747, 0.0472, 0.0, 0.0747, 0.0487, -0.0017, 0.0716, 0.0429, -0.0017, 0.0716, 0.045, -0.0039, 0.0716, 0.0429, -0.0039, 0.0716, 0.045, -0.0017, 0.0716, 0.0472, -0.0017, 0.0716, 0.0487, -0.0039, 0.0716, 0.0472, -0.0039, 0.0716, 0.0487, -0.0017, 0.0747, 0.0487, -0.0039, 0.0747, 0.0487, 0.0, 0.0768, 0.0429, 0.0, 0.0768, 0.045, 0.0, 0.079, 0.0429, 0.0, 0.079, 0.045, 0.0, 0.0768, 0.0472, 0.0, 0.0768, 0.0487, 0.0, 0.079, 0.0472, 0.0, 0.079, 0.0487, -0.0017, 0.0768, 0.0487, -0.0039, 0.0768, 0.0487, -0.0017, 0.079, 0.0487, -0.0039, 0.079, 0.0487, 0.0, 0.0812, 0.0342, 0.0, 0.0812, 0.0363,
0.0, 0.0833, 0.0342, 0.0, 0.0833, 0.0363, 0.0, 0.0812, 0.0385, 0.0, 0.0812, 0.0407, 0.0, 0.0833, 0.0379, 0.0, 0.0833, 0.0407, -0.0017, 0.0833, 0.0342, -0.0017, 0.0833, 0.0363, -0.0039, 0.0833, 0.0342, -0.0039, 0.0833, 0.0363, -0.0017, 0.0833, 0.0379, -0.0039, 0.0833, 0.0379, 0.0, 0.0855, 0.0379, 0.0, 0.0855, 0.0407, 0.0, 0.0877, 0.0379, 0.0, 0.0877, 0.0407, -0.0017, 0.0855, 0.0379, -0.0039, 0.0855, 0.0379, -0.0017, 0.0877, 0.0379, -0.0039, 0.0877, 0.0379, 0.0, 0.0812, 0.0429, 0.0, 0.0812, 0.045, 0.0, 0.0833, 0.0429, 0.0, 0.0833, 0.045, 0.0, 0.0812, 0.0472, 0.0, 0.0812, 0.0487, 0.0, 0.0833, 0.0472, 0.0, 0.0833, 0.0487, -0.0017, 0.0812, 0.0487, -0.0039, 0.0812, 0.0487,
-0.0017, 0.0833, 0.0487, -0.0039, 0.0833, 0.0487, 0.0, 0.0855, 0.0429, 0.0, 0.0855, 0.045, 0.0, 0.0877, 0.0429, 0.0, 0.0877, 0.045, 0.0, 0.0855, 0.0472, 0.0, 0.0855, 0.0487, 0.0, 0.0877, 0.0472, 0.0, 0.0877, 0.0487, -0.0017, 0.0855, 0.0487, -0.0039, 0.0855, 0.0487, -0.0017, 0.0877, 0.0487, -0.0039, 0.0877, 0.0487, 0.0, 0.0899, 0.0379, 0.0, 0.0899, 0.0407, 0.0, 0.092, 0.0379, 0.0, 0.092, 0.0407, -0.0017, 0.0899, 0.0379, -0.0039, 0.0899, 0.0379, -0.0017, 0.092, 0.0379, -0.0039, 0.092, 0.0379, 0.0, 0.0942, 0.0379, 0.0, 0.0942, 0.0407, 0.0, 0.0964, 0.0379, 0.0, 0.0964, 0.0407, -0.0017, 0.0942, 0.0379, -0.0039, 0.0942, 0.0379, -0.0017, 0.0964, 0.0379, -0.0039, 0.0964, 0.0379,
0.0, 0.0899, 0.0429, 0.0, 0.0899, 0.045, 0.0, 0.092, 0.0429, 0.0, 0.092, 0.045, 0.0, 0.0899, 0.0472, 0.0, 0.0899, 0.0487, 0.0, 0.092, 0.0472, 0.0, 0.092, 0.0487, -0.0017, 0.0899, 0.0487, -0.0039, 0.0899, 0.0487, -0.0017, 0.092, 0.0487, -0.0039, 0.092, 0.0487, 0.0, 0.0942, 0.0429, 0.0, 0.0942, 0.045, 0.0, 0.0964, 0.0429, 0.0, 0.0964, 0.045, 0.0, 0.0942, 0.0472, 0.0, 0.0942, 0.0487, 0.0, 0.0964, 0.0472, 0.0, 0.0964, 0.0487, -0.0017, 0.0942, 0.0487, -0.0039, 0.0942, 0.0487, -0.0017, 0.0964, 0.0487, -0.0039, 0.0964, 0.0487, 0.0, 0.0985, 0.0379, 0.0, 0.0985, 0.0407, 0.0, 0.1007, 0.0379, 0.0, 0.1007, 0.0407, -0.0017, 0.0985, 0.0379, -0.0039, 0.0985, 0.0379,
-0.0017, 0.1007, 0.0379, -0.0039, 0.1007, 0.0379, 0.0, 0.1029, 0.0379, 0.0, 0.1029, 0.0407, 0.0, 0.105, 0.0379, 0.0, 0.105, 0.0407, -0.0017, 0.1029, 0.0379, -0.0039, 0.1029, 0.0379, -0.0017, 0.105, 0.0379, -0.0039, 0.105, 0.0379, 0.0, 0.0985, 0.0429, 0.0, 0.0985, 0.045, 0.0, 0.1007, 0.0429, 0.0, 0.1007, 0.045, 0.0, 0.0985, 0.0472, 0.0, 0.0985, 0.0487, 0.0, 0.1007, 0.0472, 0.0, 0.1007, 0.0487, -0.0017, 0.0985, 0.0487, -0.0039, 0.0985, 0.0487, -0.0017, 0.1007, 0.0487, -0.0039, 0.1007, 0.0487, 0.0, 0.1029, 0.0429, 0.0, 0.1029, 0.045, 0.0, 0.105, 0.0429, 0.0, 0.105, 0.045, 0.0, 0.1029, 0.0472, 0.0, 0.1029, 0.0487, 0.0, 0.105, 0.0472, 0.0, 0.105, 0.0487,
-0.0017, 0.1029, 0.0487, -0.0039, 0.1029, 0.0487, -0.0017, 0.105, 0.0487, -0.0039, 0.105, 0.0487, 0.0, 0.1072, 0.0085, 0.0, 0.1072, 0.0103, 0.0, 0.1094, 0.0085, 0.0, 0.1094, 0.0103, 0.0, 0.1072, 0.0125, 0.0, 0.1072, 0.0146, 0.0, 0.1094, 0.0125, 0.0, 0.1094, 0.0146, -0.0017, 0.1072, 0.0085, -0.0039, 0.1072, 0.0085, -0.0017, 0.1094, 0.0085, -0.0039, 0.1094, 0.0085, 0.0, 0.1116, 0.0085, 0.0, 0.1116, 0.0103, 0.0, 0.1137, 0.0085, 0.0, 0.1137, 0.0103, 0.0, 0.1116, 0.0125, 0.0, 0.1116, 0.0146, 0.0, 0.1137, 0.0125, 0.0, 0.1137, 0.0146, -0.0017, 0.1116, 0.0085, -0.0039, 0.1116, 0.0085, -0.0017, 0.1137, 0.0085, -0.0039, 0.1137, 0.0085, 0.0, 0.1159, 0.0085, 0.0, 0.1159, 0.0103,
0.0, 0.1181, 0.0085, 0.0, 0.1181, 0.0103, 0.0, 0.1159, 0.0125, 0.0, 0.1159, 0.0146, 0.0, 0.1181, 0.0125, 0.0, 0.1181, 0.0146, -0.0017, 0.1159, 0.0085, -0.0039, 0.1159, 0.0085, -0.0017, 0.1181, 0.0085, -0.0039, 0.1181, 0.0085, 0.0, 0.1202, 0.0085, 0.0, 0.1202, 0.0103, 0.0, 0.1224, 0.0085, 0.0, 0.1224, 0.0103, 0.0, 0.1202, 0.0125, 0.0, 0.1202, 0.0146, 0.0, 0.1224, 0.0125, 0.0, 0.1224, 0.0146, -0.0017, 0.1202, 0.0085, -0.0039, 0.1202, 0.0085, -0.0017, 0.1224, 0.0085, -0.0039, 0.1224, 0.0085, 0.0, 0.1072, 0.0168, 0.0, 0.1072, 0.019, 0.0, 0.1094, 0.0168, 0.0, 0.1094, 0.019, 0.0, 0.1072, 0.0203, 0.0, 0.1094, 0.0203, -0.0017, 0.1072, 0.0203, -0.0039, 0.1072, 0.0203,
-0.0017, 0.1094, 0.0203, -0.0039, 0.1094, 0.0203, 0.0, 0.1116, 0.0168, 0.0, 0.1116, 0.019, 0.0, 0.1137, 0.0168, 0.0, 0.1137, 0.019, 0.0, 0.1116, 0.0203, 0.0, 0.1137, 0.0203, -0.0017, 0.1116, 0.0203, -0.0039, 0.1116, 0.0203, -0.0017, 0.1137, 0.0203, -0.0039, 0.1137, 0.0203, 0.0, 0.1159, 0.0168, 0.0, 0.1159, 0.019, 0.0, 0.1181, 0.0168, 0.0, 0.1181, 0.019, 0.0, 0.1159, 0.0203, 0.0, 0.1181, 0.0203, -0.0017, 0.1159, 0.0203, -0.0039, 0.1159, 0.0203, -0.0017, 0.1181, 0.0203, -0.0039, 0.1181, 0.0203, 0.0, 0.1202, 0.0168, 0.0, 0.1202, 0.019, 0.0, 0.1224, 0.0168, 0.0, 0.1224, 0.019, 0.0, 0.1202, 0.0203, 0.0, 0.1224, 0.0203, -0.0017, 0.1202, 0.0203, -0.0039, 0.1202, 0.0203,
-0.0017, 0.1224, 0.0203, -0.0039, 0.1224, 0.0203, 0.0, 0.1246, 0.0085, 0.0, 0.1246, 0.0103, 0.0, 0.1267, 0.0085, 0.0, 0.1267, 0.0103, 0.0, 0.1246, 0.0125, 0.0, 0.1246, 0.0146, 0.0, 0.1267, 0.0125, 0.0, 0.1267, 0.0146, -0.0017, 0.1246, 0.0085, -0.0039, 0.1246, 0.0085, -0.0017, 0.1267, 0.0085, -0.0039, 0.1267, 0.0085, 0.0, 0.1294, 0.0085, 0.0, 0.1294, 0.0103, 0.0, 0.1294, 0.0125, 0.0, 0.1294, 0.0146, -0.0017, 0.1294, 0.0085, -0.0017, 0.1294, 0.0103, -0.0039, 0.1294, 0.0085, -0.0039, 0.1294, 0.0103, -0.0017, 0.1294, 0.0125, -0.0017, 0.1294, 0.0146, -0.0039, 0.1294, 0.0125, -0.0039, 0.1294, 0.0146, 0.0, 0.1246, 0.0168, 0.0, 0.1246, 0.019, 0.0, 0.1267, 0.0168, 0.0, 0.1267, 0.019,
0.0, 0.1246, 0.0203, 0.0, 0.1267, 0.0203, -0.0017, 0.1246, 0.0203, -0.0039, 0.1246, 0.0203, -0.0017, 0.1267, 0.0203, -0.0039, 0.1267, 0.0203, 0.0, 0.1294, 0.0168, 0.0, 0.1294, 0.019, 0.0, 0.1294, 0.0203, -0.0017, 0.1294, 0.0168, -0.0017, 0.1294, 0.019, -0.0039, 0.1294, 0.0168, -0.0039, 0.1294, 0.019, -0.0017, 0.1294, 0.0203, -0.0039, 0.1294, 0.0203, 0.0, 0.1072, 0.0379, 0.0, 0.1072, 0.0407, 0.0, 0.1094, 0.0379, 0.0, 0.1094, 0.0407, -0.0017, 0.1072, 0.0379, -0.0039, 0.1072, 0.0379, -0.0017, 0.1094, 0.0379, -0.0039, 0.1094, 0.0379, 0.0, 0.1116, 0.0379, 0.0, 0.1116, 0.0407, 0.0, 0.1137, 0.0379, 0.0, 0.1137, 0.0407, -0.0017, 0.1116, 0.0379, -0.0039, 0.1116, 0.0379, -0.0017, 0.1137, 0.0379,
-0.0039, 0.1137, 0.0379, 0.0, 0.1072, 0.0429, 0.0, 0.1072, 0.045, 0.0, 0.1094, 0.0429, 0.0, 0.1094, 0.045, 0.0, 0.1072, 0.0472, 0.0, 0.1072, 0.0487, 0.0, 0.1094, 0.0472, 0.0, 0.1094, 0.0487, -0.0017, 0.1072, 0.0487, -0.0039, 0.1072, 0.0487, -0.0017, 0.1094, 0.0487, -0.0039, 0.1094, 0.0487, 0.0, 0.1116, 0.0429, 0.0, 0.1116, 0.045, 0.0, 0.1137, 0.0429, 0.0, 0.1137, 0.045, 0.0, 0.1116, 0.0472, 0.0, 0.1116, 0.0487, 0.0, 0.1137, 0.0472, 0.0, 0.1137, 0.0487, -0.0017, 0.1116, 0.0487, -0.0039, 0.1116, 0.0487, -0.0017, 0.1137, 0.0487, -0.0039, 0.1137, 0.0487, 0.0, 0.1159, 0.0379, 0.0, 0.1159, 0.0407, 0.0, 0.1181, 0.0379, 0.0, 0.1181, 0.0407, -0.0017, 0.1159, 0.0379,
-0.0039, 0.1159, 0.0379, -0.0017, 0.1181, 0.0379, -0.0039, 0.1181, 0.0379, 0.0, 0.1202, 0.0379, 0.0, 0.1202, 0.0407, 0.0, 0.1224, 0.0379, 0.0, 0.1224, 0.0407, -0.0017, 0.1202, 0.0379, -0.0039, 0.1202, 0.0379, -0.0017, 0.1224, 0.0379, -0.0039, 0.1224, 0.0379, 0.0, 0.1159, 0.0429, 0.0, 0.1159, 0.045, 0.0, 0.1181, 0.0429, 0.0, 0.1181, 0.045, 0.0, 0.1159, 0.0472, 0.0, 0.1159, 0.0487, 0.0, 0.1181, 0.0472, 0.0, 0.1181, 0.0487, -0.0017, 0.1159, 0.0487, -0.0039, 0.1159, 0.0487, -0.0017, 0.1181, 0.0487, -0.0039, 0.1181, 0.0487, 0.0, 0.1202, 0.0429, 0.0, 0.1202, 0.045, 0.0, 0.1224, 0.0429, 0.0, 0.1224, 0.045, 0.0, 0.1202, 0.0472, 0.0, 0.1202, 0.0487, 0.0, 0.1224, 0.0472,
0.0, 0.1224, 0.0487, -0.0017, 0.1202, 0.0487, -0.0039, 0.1202, 0.0487, -0.0017, 0.1224, 0.0487, -0.0039, 0.1224, 0.0487, 0.0, 0.1246, 0.0379, 0.0, 0.1246, 0.0407, 0.0, 0.1267, 0.0379, 0.0, 0.1267, 0.0407, -0.0017, 0.1246, 0.0379, -0.0039, 0.1246, 0.0379, -0.0017, 0.1267, 0.0379, -0.0039, 0.1267, 0.0379, 0.0, 0.1289, 0.0379, 0.0, 0.1289, 0.0407, 0.0, 0.1311, 0.0379, 0.0, 0.1311, 0.0407, -0.0017, 0.1289, 0.0379, -0.0039, 0.1289, 0.0379, -0.0017, 0.1311, 0.0379, -0.0039, 0.1311, 0.0379, 0.0, 0.1246, 0.0429, 0.0, 0.1246, 0.045, 0.0, 0.1267, 0.0429, 0.0, 0.1267, 0.045, 0.0, 0.1246, 0.0472, 0.0, 0.1246, 0.0487, 0.0, 0.1267, 0.0472, 0.0, 0.1267, 0.0487, -0.0017, 0.1246, 0.0487,
-0.0039, 0.1246, 0.0487, -0.0017, 0.1267, 0.0487, -0.0039, 0.1267, 0.0487, 0.0, 0.1289, 0.0429, 0.0, 0.1289, 0.045, 0.0, 0.1311, 0.0429, 0.0, 0.1311, 0.045, 0.0, 0.1289, 0.0472, 0.0, 0.1289, 0.0487, 0.0, 0.1311, 0.0472, 0.0, 0.1311, 0.0487, -0.0017, 0.1289, 0.0487, -0.0039, 0.1289, 0.0487, -0.0017, 0.1311, 0.0487, -0.0039, 0.1311, 0.0487, 0.0, 0.1324, 0.0379, 0.0, 0.1324, 0.0407, -0.0017, 0.1324, 0.0379, -0.0017, 0.1324, 0.0407, -0.0039, 0.1324, 0.0379, -0.0039, 0.1324, 0.0407, 0.0, 0.1324, 0.0429, 0.0, 0.1324, 0.045, 0.0, 0.1324, 0.0472, 0.0, 0.1324, 0.0487, -0.0017, 0.1324, 0.0429, -0.0017, 0.1324, 0.045, -0.0039, 0.1324, 0.0429, -0.0039, 0.1324, 0.045, -0.0017, 0.1324, 0.0472,
-0.0017, 0.1324, 0.0487, -0.0039, 0.1324, 0.0472, -0.0039, 0.1324, 0.0487, 0.0, 0.1412, -0.0005, 0.0, 0.1412, 0.0016, 0.0, 0.1441, -0.0005, 0.0, 0.1441, 0.0016, 0.0, 0.1412, 0.0038, 0.0, 0.1412, 0.006, 0.0, 0.1441, 0.0038, 0.0, 0.1441, 0.006, -0.0017, 0.1412, -0.0005, -0.0017, 0.1412, 0.0016, -0.0039, 0.1412, -0.0005, -0.0039, 0.1412, 0.0016, -0.0017, 0.1412, 0.0038, -0.0017, 0.1412, 0.006, -0.0039, 0.1412, 0.0038, -0.0039, 0.1412, 0.006, 0.0, 0.1463, -0.0005, 0.0, 0.1463, 0.0016, 0.0, 0.1484, -0.0005, 0.0, 0.1484, 0.0016, 0.0, 0.1463, 0.0038, 0.0, 0.1463, 0.006, 0.0, 0.1484, 0.0038, 0.0, 0.1484, 0.006, 0.0, 0.1412, 0.0081, 0.0, 0.1412, 0.0103, 0.0, 0.1441, 0.0081,
0.0, 0.1441, 0.0103, 0.0, 0.1412, 0.0125, 0.0, 0.1412, 0.0146, 0.0, 0.1441, 0.0125, 0.0, 0.1441, 0.0146, -0.0017, 0.1412, 0.0081, -0.0017, 0.1412, 0.0103, -0.0039, 0.1412, 0.0081, -0.0039, 0.1412, 0.0103, -0.0017, 0.1412, 0.0125, -0.0017, 0.1412, 0.0146, -0.0039, 0.1412, 0.0125, -0.0039, 0.1412, 0.0146, 0.0, 0.1463, 0.0081, 0.0, 0.1463, 0.0103, 0.0, 0.1484, 0.0081, 0.0, 0.1484, 0.0103, 0.0, 0.1463, 0.0125, 0.0, 0.1463, 0.0146, 0.0, 0.1484, 0.0125, 0.0, 0.1484, 0.0146, 0.0, 0.1506, -0.0005, 0.0, 0.1506, 0.0016, 0.0, 0.152, -0.0005, 0.0, 0.152, 0.0016, 0.0, 0.1506, 0.0038, 0.0, 0.1506, 0.006, 0.0, 0.152, 0.0038, 0.0, 0.152, 0.006, -0.0017, 0.152, -0.0005,
-0.0017, 0.152, 0.0016, -0.0039, 0.152, -0.0005, -0.0039, 0.152, 0.0016, -0.0017, 0.152, 0.0038, -0.0017, 0.152, 0.006, -0.0039, 0.152, 0.0038, -0.0039, 0.152, 0.006, 0.0, 0.1506, 0.0081, 0.0, 0.1506, 0.0103, 0.0, 0.152, 0.0081, 0.0, 0.152, 0.0103, 0.0, 0.1506, 0.0125, 0.0, 0.1506, 0.0146, 0.0, 0.152, 0.0125, 0.0, 0.152, 0.0146, -0.0017, 0.152, 0.0081, -0.0017, 0.152, 0.0103, -0.0039, 0.152, 0.0081, -0.0039, 0.152, 0.0103, -0.0017, 0.152, 0.0125, -0.0017, 0.152, 0.0146, -0.0039, 0.152, 0.0125, -0.0039, 0.152, 0.0146, 0.0, 0.1412, 0.0168, 0.0, 0.1412, 0.019, 0.0, 0.1441, 0.0168, 0.0, 0.1441, 0.019, 0.0, 0.1412, 0.0212, 0.0, 0.1412, 0.0233, 0.0, 0.1441, 0.0212,
0.0, 0.1441, 0.0233, -0.0017, 0.1412, 0.0168, -0.0017, 0.1412, 0.019, -0.0039, 0.1412, 0.0168, -0.0039, 0.1412, 0.019, -0.0017, 0.1412, 0.0212, -0.0017, 0.1412, 0.0233, -0.0039, 0.1412, 0.0212, -0.0039, 0.1412, 0.0233, 0.0, 0.1463, 0.0168, 0.0, 0.1463, 0.019, 0.0, 0.1484, 0.0168, 0.0, 0.1484, 0.019, 0.0, 0.1463, 0.0212, 0.0, 0.1463, 0.0233, 0.0, 0.1484, 0.0212, 0.0, 0.1484, 0.0233, 0.0, 0.1412, 0.0255, 0.0, 0.1412, 0.0277, 0.0, 0.1441, 0.0255, 0.0, 0.1441, 0.0277, 0.0, 0.1412, 0.0298, 0.0, 0.1412, 0.032, 0.0, 0.1441, 0.0298, 0.0, 0.1441, 0.032, -0.0017, 0.1412, 0.0255, -0.0017, 0.1412, 0.0277, -0.0039, 0.1412, 0.0255, -0.0039, 0.1412, 0.0277, -0.0017, 0.1412, 0.0298,
-0.0017, 0.1412, 0.032, -0.0039, 0.1412, 0.0298, -0.0039, 0.1412, 0.032, 0.0, 0.1463, 0.0255, 0.0, 0.1463, 0.0277, 0.0, 0.1484, 0.0255, 0.0, 0.1484, 0.0277, 0.0, 0.1463, 0.0298, 0.0, 0.1463, 0.032, 0.0, 0.1484, 0.0298, 0.0, 0.1484, 0.032, 0.0, 0.1506, 0.0168, 0.0, 0.1506, 0.019, 0.0, 0.152, 0.0168, 0.0, 0.152, 0.019, 0.0, 0.1506, 0.0212, 0.0, 0.1506, 0.0233, 0.0, 0.152, 0.0212, 0.0, 0.152, 0.0233, -0.0017, 0.152, 0.0168, -0.0017, 0.152, 0.019, -0.0039, 0.152, 0.0168, -0.0039, 0.152, 0.019, -0.0017, 0.152, 0.0212, -0.0017, 0.152, 0.0233, -0.0039, 0.152, 0.0212, -0.0039, 0.152, 0.0233, 0.0, 0.1506, 0.0255, 0.0, 0.1506, 0.0277, 0.0, 0.152, 0.0255,
0.0, 0.152, 0.0277, 0.0, 0.1506, 0.0298, 0.0, 0.1506, 0.032, 0.0, 0.152, 0.0298, 0.0, 0.152, 0.032, -0.0017, 0.152, 0.0255, -0.0017, 0.152, 0.0277, -0.0039, 0.152, 0.0255, -0.0039, 0.152, 0.0277, -0.0017, 0.152, 0.0298, -0.0017, 0.152, 0.032, -0.0039, 0.152, 0.0298, -0.0039, 0.152, 0.032, 0.0, 0.1412, 0.0342, 0.0, 0.1412, 0.0363, 0.0, 0.1441, 0.0342, 0.0, 0.1441, 0.0363, 0.0, 0.1412, 0.0385, 0.0, 0.1412, 0.0407, 0.0, 0.1441, 0.0385, 0.0, 0.1441, 0.0407, -0.0017, 0.1412, 0.0342, -0.0017, 0.1412, 0.0363, -0.0039, 0.1412, 0.0342, -0.0039, 0.1412, 0.0363, -0.0017, 0.1412, 0.0385, -0.0017, 0.1412, 0.0407, -0.0039, 0.1412, 0.0385, -0.0039, 0.1412, 0.0407, 0.0, 0.1463, 0.0342,
0.0, 0.1463, 0.0363, 0.0, 0.1484, 0.0342, 0.0, 0.1484, 0.0363, 0.0, 0.1463, 0.0385, 0.0, 0.1463, 0.0407, 0.0, 0.1484, 0.0385, 0.0, 0.1484, 0.0407, 0.0, 0.1412, 0.0429, 0.0, 0.1412, 0.045, 0.0, 0.1441, 0.0429, 0.0, 0.1441, 0.045, 0.0, 0.1412, 0.0472, 0.0, 0.1412, 0.0487, 0.0, 0.1441, 0.0472, 0.0, 0.1441, 0.0487, -0.0017, 0.1412, 0.0429, -0.0017, 0.1412, 0.045, -0.0039, 0.1412, 0.0429, -0.0039, 0.1412, 0.045, -0.0017, 0.1412, 0.0472, -0.0017, 0.1412, 0.0487, -0.0039, 0.1412, 0.0472, -0.0039, 0.1412, 0.0487, -0.0017, 0.1441, 0.0487, -0.0039, 0.1441, 0.0487, 0.0, 0.1463, 0.0429, 0.0, 0.1463, 0.045, 0.0, 0.1484, 0.0429, 0.0, 0.1484, 0.045, 0.0, 0.1463, 0.0472,
0.0, 0.1463, 0.0487, 0.0, 0.1484, 0.0472, 0.0, 0.1484, 0.0487, -0.0017, 0.1463, 0.0487, -0.0039, 0.1463, 0.0487, -0.0017, 0.1484, 0.0487, -0.0039, 0.1484, 0.0487, 0.0, 0.1506, 0.0342, 0.0, 0.1506, 0.0363, 0.0, 0.152, 0.0342, 0.0, 0.152, 0.0363, 0.0, 0.1506, 0.0385, 0.0, 0.1506, 0.0407, 0.0, 0.152, 0.0385, 0.0, 0.152, 0.0407, -0.0017, 0.152, 0.0342, -0.0017, 0.152, 0.0363, -0.0039, 0.152, 0.0342, -0.0039, 0.152, 0.0363, -0.0017, 0.152, 0.0385, -0.0017, 0.152, 0.0407, -0.0039, 0.152, 0.0385, -0.0039, 0.152, 0.0407, 0.0, 0.1506, 0.0429, 0.0, 0.1506, 0.045, 0.0, 0.152, 0.0429, 0.0, 0.152, 0.045, 0.0, 0.1506, 0.0472, 0.0, 0.1506, 0.0487, 0.0, 0.152, 0.0472,
0.0, 0.152, 0.0487, -0.0017, 0.152, 0.0429, -0.0017, 0.152, 0.045, -0.0039, 0.152, 0.0429, -0.0039, 0.152, 0.045, -0.0017, 0.1506, 0.0487, -0.0039, 0.1506, 0.0487, -0.0017, 0.152, 0.0472, -0.0017, 0.152, 0.0487, -0.0039, 0.152, 0.0472, -0.0039, 0.152, 0.0487, 0.0, 0.2139, -0.0006, 0.0, 0.2136, 0.0015, 0.0, 0.2133, 0.0037, 0.0, 0.213, 0.0059, -0.0017, 0.2139, -0.0006, -0.0017, 0.2136, 0.0016, -0.0039, 0.2139, -0.0006, -0.0039, 0.2136, 0.0016, -0.0017, 0.2133, 0.0038, -0.0017, 0.213, 0.006, -0.0039, 0.2133, 0.0038, -0.0039, 0.213, 0.006, 0.0, 0.2157, -0.0005, 0.0, 0.2157, 0.0016, 0.0, 0.2179, -0.0005, 0.0, 0.2179, 0.0016, 0.0, 0.2157, 0.0038, 0.0, 0.2157, 0.006, 0.0, 0.2179, 0.0038,
0.0, 0.2179, 0.006, 0.0, 0.2128, 0.0081, 0.0, 0.2127, 0.0103, 0.0, 0.2127, 0.0125, 0.0, 0.2127, 0.0146, -0.0017, 0.2128, 0.0081, -0.0017, 0.2127, 0.0103, -0.0039, 0.2128, 0.0081, -0.0039, 0.2127, 0.0103, -0.0017, 0.2127, 0.0125, -0.0017, 0.2127, 0.0146, -0.0039, 0.2127, 0.0125, -0.0039, 0.2127, 0.0146, 0.0, 0.2157, 0.0081, 0.0, 0.2157, 0.0103, 0.0, 0.2179, 0.0081, 0.0, 0.2179, 0.0103, 0.0, 0.2157, 0.0125, 0.0, 0.2157, 0.0146, 0.0, 0.2179, 0.0125, 0.0, 0.2179, 0.0146, 0.0, 0.2201, -0.0005, 0.0, 0.2201, 0.0016, 0.0, 0.2222, -0.0005, 0.0, 0.2222, 0.0016, 0.0, 0.2201, 0.0038, 0.0, 0.2201, 0.006, 0.0, 0.2222, 0.0038, 0.0, 0.2222, 0.006, 0.0, 0.2244, -0.0005,
0.0, 0.2244, 0.0016, 0.0, 0.2272, -0.0002, 0.0, 0.2259, 0.0017, 0.0, 0.2252, 0.0038, 0.0, 0.2247, 0.0061, 0.0, 0.2256, 0.0028, -0.0017, 0.2271, -0.0002, -0.0017, 0.226, 0.0016, -0.0039, 0.2271, -0.0002, -0.0039, 0.226, 0.0016, -0.0017, 0.2252, 0.0039, -0.0017, 0.2248, 0.006, -0.0039, 0.2252, 0.0039, -0.0039, 0.2248, 0.006, -0.0017, 0.2255, 0.0029, -0.0039, 0.2255, 0.0029, 0.0, 0.2201, 0.0081, 0.0, 0.2201, 0.0103, 0.0, 0.2222, 0.0081, 0.0, 0.2222, 0.0103, 0.0, 0.2201, 0.0125, 0.0, 0.2201, 0.0146, 0.0, 0.2222, 0.0125, 0.0, 0.2222, 0.0146, 0.0, 0.2245, 0.0082, 0.0, 0.2245, 0.0103, 0.0, 0.2245, 0.0125, 0.0, 0.2245, 0.0146, -0.0017, 0.2245, 0.0081, -0.0017, 0.2245, 0.0103,
-0.0039, 0.2245, 0.0081, -0.0039, 0.2245, 0.0103, -0.0017, 0.2245, 0.0125, -0.0017, 0.2245, 0.0146, -0.0039, 0.2245, 0.0125, -0.0039, 0.2245, 0.0146, 0.0, 0.2127, 0.0168, 0.0, 0.2127, 0.019, 0.0, 0.2127, 0.0212, 0.0, 0.2127, 0.0233, -0.0017, 0.2127, 0.0168, -0.0017, 0.2127, 0.019, -0.0039, 0.2127, 0.0168, -0.0039, 0.2127, 0.019, -0.0017, 0.2127, 0.0212, -0.0017, 0.2127, 0.0233, -0.0039, 0.2127, 0.0212, -0.0039, 0.2127, 0.0233, 0.0, 0.2157, 0.0168, 0.0, 0.2157, 0.019, 0.0, 0.2179, 0.0168, 0.0, 0.2179, 0.019, 0.0, 0.2157, 0.0212, 0.0, 0.2157, 0.0233, 0.0, 0.2179, 0.0212, 0.0, 0.2179, 0.0233, 0.0, 0.2127, 0.0255, 0.0, 0.2127, 0.0277, 0.0, 0.2127, 0.0298, 0.0, 0.2127, 0.032,
-0.0017, 0.2127, 0.0255, -0.0017, 0.2127, 0.0277, -0.0039, 0.2127, 0.0255, -0.0039, 0.2127, 0.0277, -0.0017, 0.2127, 0.0298, -0.0017, 0.2127, 0.032, -0.0039, 0.2127, 0.0298, -0.0039, 0.2127, 0.032, 0.0, 0.2157, 0.0255, 0.0, 0.2157, 0.0277, 0.0, 0.2179, 0.0255, 0.0, 0.2179, 0.0277, 0.0, 0.2157, 0.0298, 0.0, 0.2157, 0.032, 0.0, 0.2179, 0.0298, 0.0, 0.2179, 0.032, 0.0, 0.2201, 0.0168, 0.0, 0.2201, 0.019, 0.0, 0.2222, 0.0168, 0.0, 0.2222, 0.019, 0.0, 0.2201, 0.0212, 0.0, 0.2201, 0.0233, 0.0, 0.2222, 0.0212, 0.0, 0.2222, 0.0233, 0.0, 0.2245, 0.0168, 0.0, 0.2245, 0.019, 0.0, 0.2245, 0.0212, 0.0, 0.2245, 0.0233, -0.0017, 0.2245, 0.0168, -0.0017, 0.2245, 0.019,
-0.0039, 0.2245, 0.0168, -0.0039, 0.2245, 0.019, -0.0017, 0.2245, 0.0212, -0.0017, 0.2245, 0.0233, -0.0039, 0.2245, 0.0212, -0.0039, 0.2245, 0.0233, 0.0, 0.2201, 0.0255, 0.0, 0.2201, 0.0277, 0.0, 0.2222, 0.0255, 0.0, 0.2222, 0.0277, 0.0, 0.2201, 0.0298, 0.0, 0.2201, 0.032, 0.0, 0.2222, 0.0298, 0.0, 0.2222, 0.032, 0.0, 0.2245, 0.0255, 0.0, 0.2245, 0.0277, 0.0, 0.2245, 0.0298, 0.0, 0.2245, 0.032, -0.0017, 0.2245, 0.0255, -0.0017, 0.2245, 0.0277, -0.0039, 0.2245, 0.0255, -0.0039, 0.2245, 0.0277, -0.0017, 0.2245, 0.0298, -0.0017, 0.2245, 0.032, -0.0039, 0.2245, 0.0298, -0.0039, 0.2245, 0.032, 0.0, 0.2279, -0.0014, -0.0017, 0.2279, -0.0013, -0.0039, 0.2279, -0.0013, 0.0, 0.2127, 0.0342,
0.0, 0.2127, 0.0363, 0.0, 0.2127, 0.0385, 0.0, 0.2127, 0.0407, -0.0017, 0.2127, 0.0342, -0.0017, 0.2127, 0.0363, -0.0039, 0.2127, 0.0342, -0.0039, 0.2127, 0.0363, -0.0017, 0.2127, 0.0385, -0.0017, 0.2127, 0.0407, -0.0039, 0.2127, 0.0385, -0.0039, 0.2127, 0.0407, 0.0, 0.2157, 0.0342, 0.0, 0.2157, 0.0363, 0.0, 0.2179, 0.0342, 0.0, 0.2179, 0.0363, 0.0, 0.2157, 0.0385, 0.0, 0.2157, 0.0407, 0.0, 0.2179, 0.0385, 0.0, 0.2179, 0.0407, 0.0, 0.2127, 0.0429, 0.0, 0.2127, 0.045, 0.0, 0.2127, 0.0472, 0.0, 0.2127, 0.0487, -0.0017, 0.2127, 0.0429, -0.0017, 0.2127, 0.045, -0.0039, 0.2127, 0.0429, -0.0039, 0.2127, 0.045, -0.0017, 0.2127, 0.0472, -0.0017, 0.2127, 0.0487, -0.0039, 0.2127, 0.0472,
-0.0039, 0.2127, 0.0487, 0.0, 0.2157, 0.0429, 0.0, 0.2157, 0.045, 0.0, 0.2179, 0.0429, 0.0, 0.2179, 0.045, 0.0, 0.2157, 0.0472, 0.0, 0.2157, 0.0487, 0.0, 0.2179, 0.0472, 0.0, 0.2179, 0.0487, -0.0017, 0.2157, 0.0487, -0.0039, 0.2157, 0.0487, -0.0017, 0.2179, 0.0487, -0.0039, 0.2179, 0.0487, 0.0, 0.2201, 0.0342, 0.0, 0.2201, 0.0363, 0.0, 0.2222, 0.0342, 0.0, 0.2222, 0.0363, 0.0, 0.2201, 0.0385, 0.0, 0.2201, 0.0407, 0.0, 0.2222, 0.0385, 0.0, 0.2222, 0.0407, 0.0, 0.2245, 0.0342, 0.0, 0.2245, 0.0363, 0.0, 0.2245, 0.0385, 0.0, 0.2245, 0.0407, -0.0017, 0.2245, 0.0342, -0.0017, 0.2245, 0.0363, -0.0039, 0.2245, 0.0342, -0.0039, 0.2245, 0.0363, -0.0017, 0.2245, 0.0385,
-0.0017, 0.2245, 0.0407, -0.0039, 0.2245, 0.0385, -0.0039, 0.2245, 0.0407, 0.0, 0.2201, 0.0429, 0.0, 0.2201, 0.045, 0.0, 0.2222, 0.0429, 0.0, 0.2222, 0.045, 0.0, 0.2201, 0.0472, 0.0, 0.2201, 0.0487, 0.0, 0.2222, 0.0472, 0.0, 0.2222, 0.0487, -0.0017, 0.2201, 0.0487, -0.0039, 0.2201, 0.0487, -0.0017, 0.2222, 0.0487, -0.0039, 0.2222, 0.0487, 0.0, 0.2245, 0.0429, 0.0, 0.2245, 0.045, 0.0, 0.2245, 0.0472, 0.0, 0.2245, 0.0487, -0.0017, 0.2245, 0.0429, -0.0017, 0.2245, 0.045, -0.0039, 0.2245, 0.0429, -0.0039, 0.2245, 0.045, -0.0017, 0.2245, 0.0472, -0.0017, 0.2245, 0.0487, -0.0039, 0.2245, 0.0472, -0.0039, 0.2245, 0.0487, 0.0, 0.2765, -0.0003, 0.0, 0.2772, 0.0011, 0.0, 0.2787, -0.0005,
0.0, 0.2776, 0.002, 0.0, 0.2783, 0.004, 0.0, 0.2787, 0.0061, -0.0017, 0.2763, -0.0006, -0.0017, 0.2773, 0.0012, -0.0039, 0.2763, -0.0006, -0.0039, 0.2773, 0.0012, -0.0017, 0.2777, 0.0023, -0.0039, 0.2777, 0.0023, -0.0017, 0.2782, 0.0038, -0.0017, 0.2787, 0.006, -0.0039, 0.2782, 0.0038, -0.0039, 0.2787, 0.006, 0.0, 0.2791, 0.0082, 0.0, 0.2793, 0.0103, 0.0, 0.2794, 0.0125, 0.0, 0.2794, 0.0146, -0.0017, 0.2791, 0.0081, -0.0017, 0.2793, 0.0103, -0.0039, 0.2791, 0.0081, -0.0039, 0.2793, 0.0103, -0.0017, 0.2794, 0.0125, -0.0017, 0.2794, 0.0146, -0.0039, 0.2794, 0.0125, -0.0039, 0.2794, 0.0146, 0.0, 0.2794, 0.0168, 0.0, 0.2794, 0.019, 0.0, 0.2794, 0.0212, 0.0, 0.2794, 0.0233, -0.0017, 0.2794, 0.0168,
-0.0017, 0.2794, 0.019, -0.0039, 0.2794, 0.0168, -0.0039, 0.2794, 0.019, -0.0017, 0.2794, 0.0212, -0.0017, 0.2794, 0.0233, -0.0039, 0.2794, 0.0212, -0.0039, 0.2794, 0.0233, 0.0, 0.2794, 0.0255, 0.0, 0.2794, 0.0277, 0.0, 0.2794, 0.0298, 0.0, 0.2794, 0.032, -0.0017, 0.2794, 0.0255, -0.0017, 0.2794, 0.0277, -0.0039, 0.2794, 0.0255, -0.0039, 0.2794, 0.0277, -0.0017, 0.2794, 0.0298, -0.0017, 0.2794, 0.032, -0.0039, 0.2794, 0.0298, -0.0039, 0.2794, 0.032, 0.0, 0.2794, 0.0342, 0.0, 0.2794, 0.0363, 0.0, 0.2794, 0.0385, 0.0, 0.2794, 0.0407, -0.0017, 0.2794, 0.0342, -0.0017, 0.2794, 0.0363, -0.0039, 0.2794, 0.0342, -0.0039, 0.2794, 0.0363, -0.0017, 0.2794, 0.0385, -0.0017, 0.2794, 0.0407, -0.0039, 0.2794, 0.0385,
-0.0039, 0.2794, 0.0407, 0.0, 0.2794, 0.0429, 0.0, 0.2794, 0.045, 0.0, 0.2794, 0.0472, 0.0, 0.2794, 0.0487, -0.0017, 0.2794, 0.0429, -0.0017, 0.2794, 0.045, -0.0039, 0.2794, 0.0429, -0.0039, 0.2794, 0.045, -0.0017, 0.2794, 0.0472, -0.0017, 0.2794, 0.0487, -0.0039, 0.2794, 0.0472, -0.0039, 0.2794, 0.0487, 0.0, 0.2808, -0.0005, 0.0, 0.2808, 0.0016, 0.0, 0.283, -0.0005, 0.0, 0.283, 0.0016, 0.0, 0.2808, 0.0038, 0.0, 0.2808, 0.006, 0.0, 0.283, 0.0038, 0.0, 0.283, 0.006, 0.0, 0.2852, -0.0005, 0.0, 0.2852, 0.0016, 0.0, 0.2873, -0.0005, 0.0, 0.2873, 0.0016, 0.0, 0.2852, 0.0038, 0.0, 0.2852, 0.006, 0.0, 0.2873, 0.0038, 0.0, 0.2873, 0.006, 0.0, 0.2808, 0.0081,
0.0, 0.2808, 0.0103, 0.0, 0.283, 0.0081, 0.0, 0.283, 0.0103, 0.0, 0.2808, 0.0125, 0.0, 0.2808, 0.0146, 0.0, 0.283, 0.0125, 0.0, 0.283, 0.0146, 0.0, 0.2852, 0.0081, 0.0, 0.2852, 0.0103, 0.0, 0.2873, 0.0081, 0.0, 0.2873, 0.0103, 0.0, 0.2852, 0.0125, 0.0, 0.2852, 0.0146, 0.0, 0.2873, 0.0125, 0.0, 0.2873, 0.0146, 0.0, 0.2897, -0.0007, 0.0, 0.2901, 0.0015, 0.0, 0.2903, 0.0037, 0.0, 0.2906, 0.0057, 0.0, 0.2908, 0.0065, -0.0017, 0.2897, -0.0006, -0.0017, 0.2901, 0.0016, -0.0039, 0.2897, -0.0006, -0.0039, 0.2901, 0.0016, -0.0017, 0.2903, 0.0038, -0.0017, 0.2906, 0.0051, -0.0039, 0.2903, 0.0038, -0.0039, 0.2906, 0.0051, -0.0017, 0.2907, 0.0062, -0.0039, 0.2907, 0.0062,
0.0, 0.2895, 0.0081, 0.0, 0.2895, 0.0103, 0.0, 0.291, 0.0081, 0.0, 0.2911, 0.0103, 0.0, 0.2895, 0.0125, 0.0, 0.2895, 0.0146, 0.0, 0.2912, 0.0125, 0.0, 0.2912, 0.0146, -0.0017, 0.291, 0.0081, -0.0017, 0.2911, 0.0103, -0.0039, 0.291, 0.0081, -0.0039, 0.2911, 0.0103, -0.0017, 0.2912, 0.0125, -0.0017, 0.2912, 0.0146, -0.0039, 0.2912, 0.0125, -0.0039, 0.2912, 0.0146, 0.0, 0.2808, 0.0168, 0.0, 0.2808, 0.019, 0.0, 0.283, 0.0168, 0.0, 0.283, 0.019, 0.0, 0.2808, 0.0212, 0.0, 0.2808, 0.0233, 0.0, 0.283, 0.0212, 0.0, 0.283, 0.0233, 0.0, 0.2852, 0.0168, 0.0, 0.2852, 0.019, 0.0, 0.2873, 0.0168, 0.0, 0.2873, 0.019, 0.0, 0.2852, 0.0212, 0.0, 0.2852, 0.0233,
0.0, 0.2873, 0.0212, 0.0, 0.2873, 0.0233, 0.0, 0.2808, 0.0255, 0.0, 0.2808, 0.0277, 0.0, 0.283, 0.0255, 0.0, 0.283, 0.0277, 0.0, 0.2808, 0.0298, 0.0, 0.2808, 0.032, 0.0, 0.283, 0.0298, 0.0, 0.283, 0.032, 0.0, 0.2852, 0.0255, 0.0, 0.2852, 0.0277, 0.0, 0.2873, 0.0255, 0.0, 0.2873, 0.0277, 0.0, 0.2852, 0.0298, 0.0, 0.2852, 0.032, 0.0, 0.2873, 0.0298, 0.0, 0.2873, 0.032, 0.0, 0.2895, 0.0168, 0.0, 0.2895, 0.019, 0.0, 0.2912, 0.0168, 0.0, 0.2912, 0.019, 0.0, 0.2895, 0.0212, 0.0, 0.2895, 0.0233, 0.0, 0.2912, 0.0212, 0.0, 0.2912, 0.0233, -0.0017, 0.2912, 0.0168, -0.0017, 0.2912, 0.019, -0.0039, 0.2912, 0.0168, -0.0039, 0.2912, 0.019,
-0.0017, 0.2912, 0.0212, -0.0017, 0.2912, 0.0233, -0.0039, 0.2912, 0.0212, -0.0039, 0.2912, 0.0233, 0.0, 0.2895, 0.0255, 0.0, 0.2895, 0.0277, 0.0, 0.2912, 0.0255, 0.0, 0.2912, 0.0277, 0.0, 0.2895, 0.0298, 0.0, 0.2895, 0.032, 0.0, 0.2912, 0.0298, 0.0, 0.2912, 0.032, -0.0017, 0.2912, 0.0255, -0.0017, 0.2912, 0.0277, -0.0039, 0.2912, 0.0255, -0.0039, 0.2912, 0.0277, -0.0017, 0.2912, 0.0298, -0.0017, 0.2912, 0.032, -0.0039, 0.2912, 0.0298, -0.0039, 0.2912, 0.032, 0.0, 0.3088, -0.0005, 0.0, 0.3088, 0.0016, 0.0, 0.3088, 0.0038, 0.0, 0.3088, 0.006, -0.0017, 0.3088, -0.0005, -0.0017, 0.3088, 0.0016, -0.0039, 0.3088, -0.0005, -0.0039, 0.3088, 0.0016, -0.0017, 0.3088, 0.0038, -0.0017, 0.3088, 0.006,
-0.0039, 0.3088, 0.0038, -0.0039, 0.3088, 0.006, 0.0, 0.3112, -0.0005, 0.0, 0.3112, 0.0016, 0.0, 0.3134, -0.0005, 0.0, 0.3134, 0.0016, 0.0, 0.3112, 0.0038, 0.0, 0.3112, 0.006, 0.0, 0.3134, 0.0038, 0.0, 0.3134, 0.006, 0.0, 0.3088, 0.0081, 0.0, 0.3088, 0.0103, 0.0, 0.3088, 0.0125, 0.0, 0.3088, 0.0146, -0.0017, 0.3088, 0.0081, -0.0017, 0.3088, 0.0103, -0.0039, 0.3088, 0.0081, -0.0039, 0.3088, 0.0103, -0.0017, 0.3088, 0.0125, -0.0017, 0.3088, 0.0146, -0.0039, 0.3088, 0.0125, -0.0039, 0.3088, 0.0146, 0.0, 0.3112, 0.0081, 0.0, 0.3112, 0.0103, 0.0, 0.3134, 0.0081, 0.0, 0.3134, 0.0103, 0.0, 0.3112, 0.0125, 0.0, 0.3112, 0.0146, 0.0, 0.3134, 0.0125, 0.0, 0.3134, 0.0146,
0.0, 0.3088, 0.0168, 0.0, 0.3088, 0.019, 0.0, 0.3088, 0.0212, 0.0, 0.3088, 0.0233, -0.0017, 0.3088, 0.0168, -0.0017, 0.3088, 0.019, -0.0039, 0.3088, 0.0168, -0.0039, 0.3088, 0.019, -0.0017, 0.3088, 0.0212, -0.0017, 0.3088, 0.0233, -0.0039, 0.3088, 0.0212, -0.0039, 0.3088, 0.0233, 0.0, 0.3112, 0.0168, 0.0, 0.3112, 0.019, 0.0, 0.3134, 0.0168, 0.0, 0.3134, 0.019, 0.0, 0.3112, 0.0212, 0.0, 0.3112, 0.0233, 0.0, 0.3134, 0.0212, 0.0, 0.3134, 0.0233, 0.0, 0.3088, 0.0255, 0.0, 0.3088, 0.0277, 0.0, 0.3088, 0.0298, 0.0, 0.3088, 0.032, -0.0017, 0.3088, 0.0255, -0.0017, 0.3088, 0.0277, -0.0039, 0.3088, 0.0255, -0.0039, 0.3088, 0.0277, -0.0017, 0.3088, 0.0298, -0.0017, 0.3088, 0.032,
-0.0039, 0.3088, 0.0298, -0.0039, 0.3088, 0.032, 0.0, 0.3112, 0.0255, 0.0, 0.3112, 0.0277, 0.0, 0.3134, 0.0255, 0.0, 0.3134, 0.0277, 0.0, 0.3112, 0.0298, 0.0, 0.3112, 0.032, 0.0, 0.3134, 0.0298, 0.0, 0.3134, 0.032, 0.0, 0.2808, 0.0342, 0.0, 0.2808, 0.0363, 0.0, 0.283, 0.0342, 0.0, 0.283, 0.0363, 0.0, 0.2808, 0.0385, 0.0, 0.2808, 0.0407, 0.0, 0.283, 0.0385, 0.0, 0.283, 0.0407, 0.0, 0.2852, 0.0342, 0.0, 0.2852, 0.0363, 0.0, 0.2873, 0.0342, 0.0, 0.2873, 0.0363, 0.0, 0.2852, 0.0385, 0.0, 0.2852, 0.0407, 0.0, 0.2873, 0.0385, 0.0, 0.2873, 0.0407, 0.0, 0.2808, 0.0429, 0.0, 0.2808, 0.045, 0.0, 0.283, 0.0429, 0.0, 0.283, 0.045,
0.0, 0.2808, 0.0472, 0.0, 0.2808, 0.0487, 0.0, 0.283, 0.0472, 0.0, 0.283, 0.0487, -0.0017, 0.2808, 0.0487, -0.0039, 0.2808, 0.0487, -0.0017, 0.283, 0.0487, -0.0039, 0.283, 0.0487, 0.0, 0.2852, 0.0429, 0.0, 0.2852, 0.045, 0.0, 0.2873, 0.0429, 0.0, 0.2873, 0.045, 0.0, 0.2852, 0.0472, 0.0, 0.2852, 0.0487, 0.0, 0.2873, 0.0472, 0.0, 0.2873, 0.0487, -0.0017, 0.2852, 0.0487, -0.0039, 0.2852, 0.0487, -0.0017, 0.2873, 0.0487, -0.0039, 0.2873, 0.0487, 0.0, 0.2895, 0.0342, 0.0, 0.2895, 0.0363, 0.0, 0.2912, 0.0342, 0.0, 0.2912, 0.0363, 0.0, 0.2895, 0.0385, 0.0, 0.2895, 0.0407, 0.0, 0.2912, 0.0385, 0.0, 0.2912, 0.0407, -0.0017, 0.2912, 0.0342, -0.0017, 0.2912, 0.0363,
-0.0039, 0.2912, 0.0342, -0.0039, 0.2912, 0.0363, -0.0017, 0.2912, 0.0385, -0.0017, 0.2912, 0.0407, -0.0039, 0.2912, 0.0385, -0.0039, 0.2912, 0.0407, 0.0, 0.2895, 0.0429, 0.0, 0.2895, 0.045, 0.0, 0.2912, 0.0429, 0.0, 0.2912, 0.045, 0.0, 0.2895, 0.0472, 0.0, 0.2895, 0.0487, 0.0, 0.2912, 0.0472, 0.0, 0.2912, 0.0487, -0.0017, 0.2912, 0.0429, -0.0017, 0.2912, 0.045, -0.0039, 0.2912, 0.0429, -0.0039, 0.2912, 0.045, -0.0017, 0.2895, 0.0487, -0.0039, 0.2895, 0.0487, -0.0017, 0.2912, 0.0472, -0.0017, 0.2912, 0.0487, -0.0039, 0.2912, 0.0472, -0.0039, 0.2912, 0.0487, 0.0, 0.3088, 0.0342, 0.0, 0.3088, 0.0363, 0.0, 0.3088, 0.0385, 0.0, 0.3088, 0.0407, -0.0017, 0.3088, 0.0342, -0.0017, 0.3088, 0.0363,
-0.0039, 0.3088, 0.0342, -0.0039, 0.3088, 0.0363, -0.0017, 0.3088, 0.0385, -0.0017, 0.3088, 0.0407, -0.0039, 0.3088, 0.0385, -0.0039, 0.3088, 0.0407, 0.0, 0.3112, 0.0342, 0.0, 0.3112, 0.0363, 0.0, 0.3134, 0.0342, 0.0, 0.3134, 0.0363, 0.0, 0.3112, 0.0385, 0.0, 0.3112, 0.0407, 0.0, 0.3134, 0.0385, 0.0, 0.3134, 0.0407, 0.0, 0.3088, 0.0429, 0.0, 0.3088, 0.045, 0.0, 0.3088, 0.0472, 0.0, 0.3088, 0.0487, -0.0017, 0.3088, 0.0429, -0.0017, 0.3088, 0.045, -0.0039, 0.3088, 0.0429, -0.0039, 0.3088, 0.045, -0.0017, 0.3088, 0.0472, -0.0017, 0.3088, 0.0487, -0.0039, 0.3088, 0.0472, -0.0039, 0.3088, 0.0487, 0.0, 0.3112, 0.0429, 0.0, 0.3112, 0.045, 0.0, 0.3134, 0.0429, 0.0, 0.3134, 0.045,
0.0, 0.3112, 0.0472, 0.0, 0.3112, 0.0487, 0.0, 0.3134, 0.0472, 0.0, 0.3134, 0.0487, -0.0017, 0.3112, 0.0487, -0.0039, 0.3112, 0.0487, -0.0017, 0.3134, 0.0487, -0.0039, 0.3134, 0.0487, 0.0, 0.3155, -0.0005, 0.0, 0.3155, 0.0016, 0.0, 0.3177, -0.0005, 0.0, 0.3177, 0.0016, 0.0, 0.3155, 0.0038, 0.0, 0.3155, 0.006, 0.0, 0.3177, 0.0038, 0.0, 0.3177, 0.006, 0.0, 0.3206, -0.0005, 0.0, 0.3206, 0.0016, 0.0, 0.3206, 0.0038, 0.0, 0.3206, 0.006, -0.0017, 0.3206, -0.0005, -0.0017, 0.3206, 0.0016, -0.0039, 0.3206, -0.0005, -0.0039, 0.3206, 0.0016, -0.0017, 0.3206, 0.0038, -0.0017, 0.3206, 0.006, -0.0039, 0.3206, 0.0038, -0.0039, 0.3206, 0.006, 0.0, 0.3155, 0.0081, 0.0, 0.3155, 0.0103,
0.0, 0.3177, 0.0081, 0.0, 0.3177, 0.0103, 0.0, 0.3155, 0.0125, 0.0, 0.3155, 0.0146, 0.0, 0.3177, 0.0125, 0.0, 0.3177, 0.0146, 0.0, 0.3206, 0.0081, 0.0, 0.3206, 0.0103, 0.0, 0.3206, 0.0125, 0.0, 0.3206, 0.0146, -0.0017, 0.3206, 0.0081, -0.0017, 0.3206, 0.0103, -0.0039, 0.3206, 0.0081, -0.0039, 0.3206, 0.0103, -0.0017, 0.3206, 0.0125, -0.0017, 0.3206, 0.0146, -0.0039, 0.3206, 0.0125, -0.0039, 0.3206, 0.0146, 0.0, 0.3155, 0.0168, 0.0, 0.3155, 0.019, 0.0, 0.3177, 0.0168, 0.0, 0.3177, 0.019, 0.0, 0.3155, 0.0212, 0.0, 0.3155, 0.0233, 0.0, 0.3177, 0.0212, 0.0, 0.3177, 0.0233, 0.0, 0.3206, 0.0168, 0.0, 0.3206, 0.019, 0.0, 0.3206, 0.0212, 0.0, 0.3206, 0.0233,
-0.0017, 0.3206, 0.0168, -0.0017, 0.3206, 0.019, -0.0039, 0.3206, 0.0168, -0.0039, 0.3206, 0.019, -0.0017, 0.3206, 0.0212, -0.0017, 0.3206, 0.0233, -0.0039, 0.3206, 0.0212, -0.0039, 0.3206, 0.0233, 0.0, 0.3155, 0.0255, 0.0, 0.3155, 0.0277, 0.0, 0.3177, 0.0255, 0.0, 0.3177, 0.0277, 0.0, 0.3155, 0.0298, 0.0, 0.3155, 0.032, 0.0, 0.3177, 0.0298, 0.0, 0.3177, 0.032, 0.0, 0.3206, 0.0255, 0.0, 0.3206, 0.0277, 0.0, 0.3206, 0.0298, 0.0, 0.3206, 0.032, -0.0017, 0.3206, 0.0255, -0.0017, 0.3206, 0.0277, -0.0039, 0.3206, 0.0255, -0.0039, 0.3206, 0.0277, -0.0017, 0.3206, 0.0298, -0.0017, 0.3206, 0.032, -0.0039, 0.3206, 0.0298, -0.0039, 0.3206, 0.032, 0.0, 0.3402, -0.0005, 0.0, 0.3402, 0.0016,
0.0, 0.3402, 0.0038, 0.0, 0.3402, 0.006, -0.0017, 0.3402, -0.0005, -0.0017, 0.3402, 0.0016, -0.0039, 0.3402, -0.0005, -0.0039, 0.3402, 0.0016, -0.0017, 0.3402, 0.0038, -0.0017, 0.3402, 0.006, -0.0039, 0.3402, 0.0038, -0.0039, 0.3402, 0.006, 0.0, 0.3402, 0.0081, 0.0, 0.3402, 0.0103, 0.0, 0.3402, 0.0125, 0.0, 0.3402, 0.0146, -0.0017, 0.3402, 0.0081, -0.0017, 0.3402, 0.0103, -0.0039, 0.3402, 0.0081, -0.0039, 0.3402, 0.0103, -0.0017, 0.3402, 0.0125, -0.0017, 0.3402, 0.0146, -0.0039, 0.3402, 0.0125, -0.0039, 0.3402, 0.0146, 0.0, 0.3416, -0.0005, 0.0, 0.3416, 0.0016, 0.0, 0.3438, -0.0005, 0.0, 0.3438, 0.0016, 0.0, 0.3416, 0.0038, 0.0, 0.3416, 0.006, 0.0, 0.3438, 0.0038, 0.0, 0.3438, 0.006,
0.0, 0.3459, -0.0005, 0.0, 0.3459, 0.0016, 0.0, 0.3481, -0.0005, 0.0, 0.3481, 0.0016, 0.0, 0.3459, 0.0038, 0.0, 0.3459, 0.006, 0.0, 0.3481, 0.0038, 0.0, 0.3481, 0.006, 0.0, 0.3416, 0.0081, 0.0, 0.3416, 0.0103, 0.0, 0.3438, 0.0081, 0.0, 0.3438, 0.0103, 0.0, 0.3416, 0.0125, 0.0, 0.3416, 0.0146, 0.0, 0.3438, 0.0125, 0.0, 0.3438, 0.0146, 0.0, 0.3459, 0.0081, 0.0, 0.3459, 0.0103, 0.0, 0.3481, 0.0081, 0.0, 0.3481, 0.0103, 0.0, 0.3459, 0.0125, 0.0, 0.3459, 0.0146, 0.0, 0.3481, 0.0125, 0.0, 0.3481, 0.0146, 0.0, 0.3402, 0.0168, 0.0, 0.3402, 0.019, 0.0, 0.3402, 0.0212, 0.0, 0.3402, 0.0233, -0.0017, 0.3402, 0.0168, -0.0017, 0.3402, 0.019,
-0.0039, 0.3402, 0.0168, -0.0039, 0.3402, 0.019, -0.0017, 0.3402, 0.0212, -0.0017, 0.3402, 0.0233, -0.0039, 0.3402, 0.0212, -0.0039, 0.3402, 0.0233, 0.0, 0.3402, 0.0255, 0.0, 0.3402, 0.0277, 0.0, 0.3402, 0.0298, 0.0, 0.3402, 0.032, -0.0017, 0.3402, 0.0255, -0.0017, 0.3402, 0.0277, -0.0039, 0.3402, 0.0255, -0.0039, 0.3402, 0.0277, -0.0017, 0.3402, 0.0298, -0.0017, 0.3402, 0.032, -0.0039, 0.3402, 0.0298, -0.0039, 0.3402, 0.032, 0.0, 0.3416, 0.0168, 0.0, 0.3416, 0.019, 0.0, 0.3438, 0.0168, 0.0, 0.3438, 0.019, 0.0, 0.3416, 0.0212, 0.0, 0.3416, 0.0233, 0.0, 0.3438, 0.0212, 0.0, 0.3438, 0.0233, 0.0, 0.3459, 0.0168, 0.0, 0.3459, 0.019, 0.0, 0.3481, 0.0168, 0.0, 0.3481, 0.019,
0.0, 0.3459, 0.0212, 0.0, 0.3459, 0.0233, 0.0, 0.3481, 0.0212, 0.0, 0.3481, 0.0233, 0.0, 0.3416, 0.0255, 0.0, 0.3416, 0.0277, 0.0, 0.3438, 0.0255, 0.0, 0.3438, 0.0277, 0.0, 0.3416, 0.0298, 0.0, 0.3416, 0.032, 0.0, 0.3438, 0.0298, 0.0, 0.3438, 0.032, 0.0, 0.3459, 0.0255, 0.0, 0.3459, 0.0277, 0.0, 0.3481, 0.0255, 0.0, 0.3481, 0.0277, 0.0, 0.3459, 0.0298, 0.0, 0.3459, 0.032, 0.0, 0.3481, 0.0298, 0.0, 0.3481, 0.032, 0.0, 0.3155, 0.0342, 0.0, 0.3155, 0.0363, 0.0, 0.3177, 0.0342, 0.0, 0.3177, 0.0363, 0.0, 0.3155, 0.0385, 0.0, 0.3155, 0.0407, 0.0, 0.3177, 0.0385, 0.0, 0.3177, 0.0407, 0.0, 0.3206, 0.0342, 0.0, 0.3206, 0.0363,
0.0, 0.3206, 0.0385, 0.0, 0.3206, 0.0407, -0.0017, 0.3206, 0.0342, -0.0017, 0.3206, 0.0363, -0.0039, 0.3206, 0.0342, -0.0039, 0.3206, 0.0363, -0.0017, 0.3206, 0.0385, -0.0017, 0.3206, 0.0407, -0.0039, 0.3206, 0.0385, -0.0039, 0.3206, 0.0407, 0.0, 0.3155, 0.0429, 0.0, 0.3155, 0.045, 0.0, 0.3177, 0.0429, 0.0, 0.3177, 0.045, 0.0, 0.3155, 0.0472, 0.0, 0.3155, 0.0487, 0.0, 0.3177, 0.0472, 0.0, 0.3177, 0.0487, -0.0017, 0.3155, 0.0487, -0.0039, 0.3155, 0.0487, -0.0017, 0.3177, 0.0487, -0.0039, 0.3177, 0.0487, 0.0, 0.3206, 0.0429, 0.0, 0.3206, 0.045, 0.0, 0.3206, 0.0472, 0.0, 0.3206, 0.0487, -0.0017, 0.3206, 0.0429, -0.0017, 0.3206, 0.045, -0.0039, 0.3206, 0.0429, -0.0039, 0.3206, 0.045,
-0.0017, 0.3206, 0.0472, -0.0017, 0.3206, 0.0487, -0.0039, 0.3206, 0.0472, -0.0039, 0.3206, 0.0487, 0.0, 0.3402, 0.0342, 0.0, 0.3402, 0.0363, 0.0, 0.3402, 0.0385, 0.0, 0.3402, 0.0407, -0.0017, 0.3402, 0.0342, -0.0017, 0.3402, 0.0363, -0.0039, 0.3402, 0.0342, -0.0039, 0.3402, 0.0363, -0.0017, 0.3402, 0.0385, -0.0017, 0.3402, 0.0407, -0.0039, 0.3402, 0.0385, -0.0039, 0.3402, 0.0407, 0.0, 0.3402, 0.0429, 0.0, 0.3402, 0.045, 0.0, 0.3402, 0.0472, 0.0, 0.3402, 0.0487, -0.0017, 0.3402, 0.0429, -0.0017, 0.3402, 0.045, -0.0039, 0.3402, 0.0429, -0.0039, 0.3402, 0.045, -0.0017, 0.3402, 0.0472, -0.0017, 0.3402, 0.0487, -0.0039, 0.3402, 0.0472, -0.0039, 0.3402, 0.0487, 0.0, 0.3416, 0.0342, 0.0, 0.3416, 0.0363,
0.0, 0.3438, 0.0342, 0.0, 0.3438, 0.0363, 0.0, 0.3416, 0.0385, 0.0, 0.3416, 0.0407, 0.0, 0.3438, 0.0385, 0.0, 0.3438, 0.0407, 0.0, 0.3459, 0.0342, 0.0, 0.3459, 0.0363, 0.0, 0.3481, 0.0342, 0.0, 0.3481, 0.0363, 0.0, 0.3459, 0.0385, 0.0, 0.3459, 0.0407, 0.0, 0.3481, 0.0385, 0.0, 0.3481, 0.0407, 0.0, 0.3416, 0.0429, 0.0, 0.3416, 0.045, 0.0, 0.3438, 0.0429, 0.0, 0.3438, 0.045, 0.0, 0.3416, 0.0472, 0.0, 0.3416, 0.0487, 0.0, 0.3438, 0.0472, 0.0, 0.3438, 0.0487, -0.0017, 0.3416, 0.0487, -0.0039, 0.3416, 0.0487, -0.0017, 0.3438, 0.0487, -0.0039, 0.3438, 0.0487, 0.0, 0.3459, 0.0429, 0.0, 0.3459, 0.045, 0.0, 0.3481, 0.0429, 0.0, 0.3481, 0.045,
0.0, 0.3459, 0.0472, 0.0, 0.3459, 0.0487, 0.0, 0.3481, 0.0472, 0.0, 0.3481, 0.0487, -0.0017, 0.3459, 0.0487, -0.0039, 0.3459, 0.0487, -0.0017, 0.3481, 0.0487, -0.0039, 0.3481, 0.0487, 0.0, 0.3503, -0.0005, 0.0, 0.3503, 0.0016, 0.0, 0.352, -0.0005, 0.0, 0.352, 0.0016, 0.0, 0.3503, 0.0038, 0.0, 0.3503, 0.006, 0.0, 0.352, 0.0038, 0.0, 0.352, 0.006, -0.0017, 0.352, -0.0005, -0.0017, 0.352, 0.0016, -0.0039, 0.352, -0.0005, -0.0039, 0.352, 0.0016, -0.0017, 0.352, 0.0038, -0.0017, 0.352, 0.006, -0.0039, 0.352, 0.0038, -0.0039, 0.352, 0.006, 0.0, 0.3503, 0.0081, 0.0, 0.3503, 0.0103, 0.0, 0.352, 0.0081, 0.0, 0.352, 0.0103, 0.0, 0.3503, 0.0125, 0.0, 0.3503, 0.0146,
0.0, 0.352, 0.0125, 0.0, 0.352, 0.0146, -0.0017, 0.352, 0.0081, -0.0017, 0.352, 0.0103, -0.0039, 0.352, 0.0081, -0.0039, 0.352, 0.0103, -0.0017, 0.352, 0.0125, -0.0017, 0.352, 0.0146, -0.0039, 0.352, 0.0125, -0.0039, 0.352, 0.0146, 0.0, 0.3503, 0.0168, 0.0, 0.3503, 0.019, 0.0, 0.352, 0.0168, 0.0, 0.352, 0.019, 0.0, 0.3503, 0.0212, 0.0, 0.3503, 0.0233, 0.0, 0.352, 0.0212, 0.0, 0.352, 0.0233, -0.0017, 0.352, 0.0168, -0.0017, 0.352, 0.019, -0.0039, 0.352, 0.0168, -0.0039, 0.352, 0.019, -0.0017, 0.352, 0.0212, -0.0017, 0.352, 0.0233, -0.0039, 0.352, 0.0212, -0.0039, 0.352, 0.0233, 0.0, 0.3503, 0.0255, 0.0, 0.3503, 0.0277, 0.0, 0.352, 0.0255, 0.0, 0.352, 0.0277,
0.0, 0.3503, 0.0298, 0.0, 0.3503, 0.032, 0.0, 0.352, 0.0298, 0.0, 0.352, 0.032, -0.0017, 0.352, 0.0255, -0.0017, 0.352, 0.0277, -0.0039, 0.352, 0.0255, -0.0039, 0.352, 0.0277, -0.0017, 0.352, 0.0298, -0.0017, 0.352, 0.032, -0.0039, 0.352, 0.0298, -0.0039, 0.352, 0.032, 0.0, 0.3503, 0.0342, 0.0, 0.3503, 0.0363, 0.0, 0.352, 0.0342, 0.0, 0.352, 0.0363, 0.0, 0.3503, 0.0385, 0.0, 0.3503, 0.0407, 0.0, 0.352, 0.0379, 0.0, 0.3524, 0.0407, -0.0017, 0.352, 0.0342, -0.0017, 0.352, 0.0363, -0.0039, 0.352, 0.0342, -0.0039, 0.352, 0.0363, -0.0017, 0.352, 0.0379, -0.0039, 0.352, 0.0379, 0.0, 0.3546, 0.0379, 0.0, 0.3546, 0.0407, 0.0, 0.3568, 0.0379, 0.0, 0.3568, 0.0407,
-0.0017, 0.3546, 0.0379, -0.0039, 0.3546, 0.0379, -0.0017, 0.3568, 0.0379, -0.0039, 0.3568, 0.0379, 0.0, 0.3503, 0.0429, 0.0, 0.3503, 0.045, 0.0, 0.3524, 0.0429, 0.0, 0.3524, 0.045, 0.0, 0.3503, 0.0472, 0.0, 0.3503, 0.0487, 0.0, 0.3524, 0.0472, 0.0, 0.3524, 0.0487, -0.0017, 0.3503, 0.0487, -0.0039, 0.3503, 0.0487, -0.0017, 0.3524, 0.0487, -0.0039, 0.3524, 0.0487, 0.0, 0.3546, 0.0429, 0.0, 0.3546, 0.045, 0.0, 0.3568, 0.0429, 0.0, 0.3568, 0.045, 0.0, 0.3546, 0.0472, 0.0, 0.3546, 0.0487, 0.0, 0.3568, 0.0472, 0.0, 0.3568, 0.0487, -0.0017, 0.3546, 0.0487, -0.0039, 0.3546, 0.0487, -0.0017, 0.3568, 0.0487, -0.0039, 0.3568, 0.0487, 0.0, 0.3589, 0.0379, 0.0, 0.3589, 0.0407,
0.0, 0.3611, 0.0379, 0.0, 0.3611, 0.0407, -0.0017, 0.3589, 0.0379, -0.0039, 0.3589, 0.0379, -0.0017, 0.3611, 0.0379, -0.0039, 0.3611, 0.0379, 0.0, 0.3633, 0.0379, 0.0, 0.3633, 0.0407, 0.0, 0.3655, 0.0379, 0.0, 0.3655, 0.0407, -0.0017, 0.3633, 0.0379, -0.0039, 0.3633, 0.0379, -0.0017, 0.3655, 0.0379, -0.0039, 0.3655, 0.0379, 0.0, 0.3589, 0.0429, 0.0, 0.3589, 0.045, 0.0, 0.3611, 0.0429, 0.0, 0.3611, 0.045, 0.0, 0.3589, 0.0472, 0.0, 0.3589, 0.0487, 0.0, 0.3611, 0.0472, 0.0, 0.3611, 0.0487, -0.0017, 0.3589, 0.0487, -0.0039, 0.3589, 0.0487, -0.0017, 0.3611, 0.0487, -0.0039, 0.3611, 0.0487, 0.0, 0.3633, 0.0429, 0.0, 0.3633, 0.045, 0.0, 0.3655, 0.0429, 0.0, 0.3655, 0.045,
0.0, 0.3633, 0.0472, 0.0, 0.3633, 0.0487, 0.0, 0.3655, 0.0472, 0.0, 0.3655, 0.0487, -0.0017, 0.3633, 0.0487, -0.0039, 0.3633, 0.0487, -0.0017, 0.3655, 0.0487, -0.0039, 0.3655, 0.0487, 0.0, 0.3676, 0.0379, 0.0, 0.3676, 0.0407, 0.0, 0.3698, 0.0379, 0.0, 0.3698, 0.0407, -0.0017, 0.3676, 0.0379, -0.0039, 0.3676, 0.0379, -0.0017, 0.3698, 0.0379, -0.0039, 0.3698, 0.0379, 0.0, 0.372, 0.0379, 0.0, 0.372, 0.0407, 0.0, 0.3741, 0.0379, 0.0, 0.3741, 0.0407, -0.0017, 0.372, 0.0379, -0.0039, 0.372, 0.0379, -0.0017, 0.3741, 0.0379, -0.0039, 0.3741, 0.0379, 0.0, 0.3676, 0.0429, 0.0, 0.3676, 0.045, 0.0, 0.3698, 0.0429, 0.0, 0.3698, 0.045, 0.0, 0.3676, 0.0472, 0.0, 0.3676, 0.0487,
0.0, 0.3698, 0.0472, 0.0, 0.3698, 0.0487, -0.0017, 0.3676, 0.0487, -0.0039, 0.3676, 0.0487, -0.0017, 0.3698, 0.0487, -0.0039, 0.3698, 0.0487, 0.0, 0.372, 0.0429, 0.0, 0.372, 0.045, 0.0, 0.3741, 0.0429, 0.0, 0.3741, 0.045, 0.0, 0.372, 0.0472, 0.0, 0.372, 0.0487, 0.0, 0.3741, 0.0472, 0.0, 0.3741, 0.0487, -0.0017, 0.372, 0.0487, -0.0039, 0.372, 0.0487, -0.0017, 0.3741, 0.0487, -0.0039, 0.3741, 0.0487, 0.0, 0.3763, 0.0379, 0.0, 0.3763, 0.0407, 0.0, 0.3784, 0.0379, 0.0, 0.3785, 0.0407, -0.0017, 0.3763, 0.0379, -0.0039, 0.3763, 0.0379, -0.0017, 0.3785, 0.0379, -0.0039, 0.3785, 0.0379, 0.0, 0.3835, 0.0373, 0.0, 0.3806, 0.0377, 0.0, 0.3807, 0.0407, 0.0, 0.3827, 0.0374,
0.0, 0.3828, 0.0407, -0.0017, 0.3833, 0.0373, -0.0039, 0.3833, 0.0373, -0.0017, 0.3807, 0.0377, -0.0039, 0.3807, 0.0377, -0.0017, 0.3822, 0.0375, -0.0039, 0.3822, 0.0375, 0.0, 0.3763, 0.0429, 0.0, 0.3763, 0.045, 0.0, 0.3785, 0.0429, 0.0, 0.3785, 0.045, 0.0, 0.3763, 0.0472, 0.0, 0.3763, 0.0487, 0.0, 0.3785, 0.0472, 0.0, 0.3785, 0.0487, -0.0017, 0.3763, 0.0487, -0.0039, 0.3763, 0.0487, -0.0017, 0.3785, 0.0487, -0.0039, 0.3785, 0.0487, 0.0, 0.3807, 0.0429, 0.0, 0.3807, 0.045, 0.0, 0.3828, 0.0429, 0.0, 0.3828, 0.045, 0.0, 0.3807, 0.0472, 0.0, 0.3807, 0.0487, 0.0, 0.3828, 0.0472, 0.0, 0.3828, 0.0487, -0.0017, 0.3807, 0.0487, -0.0039, 0.3807, 0.0487, -0.0017, 0.3828, 0.0487,
-0.0039, 0.3828, 0.0487, 0.0, 0.4002, -0.0003, 0.0, 0.4009, 0.001, -0.0017, 0.3999, -0.0006, -0.0017, 0.401, 0.0011, -0.0039, 0.3999, -0.0006, -0.0039, 0.401, 0.0011, 0.0, 0.3947, 0.033, 0.0, 0.3956, 0.0324, -0.0017, 0.3947, 0.033, -0.0039, 0.3947, 0.033, -0.0017, 0.3959, 0.0321, -0.0039, 0.3959, 0.0321, 0.0, 0.4009, 0.0261, 0.0, 0.4001, 0.0274, 0.0, 0.3982, 0.03, 0.0, 0.3971, 0.031, 0.0, 0.399, 0.0289, 0.0, 0.4002, 0.032, -0.0017, 0.4009, 0.026, -0.0017, 0.3999, 0.0277, -0.0039, 0.4009, 0.026, -0.0039, 0.3999, 0.0277, -0.0017, 0.3982, 0.0299, -0.0017, 0.3971, 0.031, -0.0039, 0.3982, 0.0299, -0.0039, 0.3971, 0.031, -0.0017, 0.3991, 0.0288, -0.0039, 0.3991, 0.0288, 0.0, 0.4024, -0.0005,
0.0, 0.4015, 0.002, 0.0, 0.4045, -0.0005, 0.0, 0.4045, 0.0016, 0.0, 0.4025, 0.004, 0.0, 0.4031, 0.0056, 0.0, 0.4045, 0.0038, 0.0, 0.4033, 0.0064, -0.0017, 0.4016, 0.0021, -0.0039, 0.4016, 0.0021, -0.0017, 0.4024, 0.0038, -0.0017, 0.4032, 0.0058, -0.0039, 0.4024, 0.0038, -0.0039, 0.4032, 0.0058, -0.0017, 0.4035, 0.0069, -0.0039, 0.4035, 0.0069, 0.0, 0.4067, -0.0005, 0.0, 0.4067, 0.0016, 0.0, 0.4089, -0.0005, 0.0, 0.4089, 0.0016, 0.0, 0.4067, 0.0038, 0.0, 0.4067, 0.006, 0.0, 0.4089, 0.0038, 0.0, 0.4089, 0.006, 0.0, 0.4038, 0.0083, 0.0, 0.4039, 0.0104, 0.0, 0.4039, 0.0125, 0.0, 0.4039, 0.0146, -0.0017, 0.4037, 0.0081, -0.0017, 0.4039, 0.0103, -0.0039, 0.4037, 0.0081,
-0.0039, 0.4039, 0.0103, -0.0017, 0.4039, 0.0125, -0.0017, 0.4039, 0.0146, -0.0039, 0.4039, 0.0125, -0.0039, 0.4039, 0.0146, 0.0, 0.4067, 0.0081, 0.0, 0.4067, 0.0103, 0.0, 0.4089, 0.0081, 0.0, 0.4089, 0.0103, 0.0, 0.4067, 0.0125, 0.0, 0.4067, 0.0146, 0.0, 0.4089, 0.0125, 0.0, 0.4089, 0.0146, 0.0, 0.411, -0.0005, 0.0, 0.411, 0.0016, 0.0, 0.4131, -0.0007, 0.0, 0.414, 0.0015, 0.0, 0.411, 0.0038, 0.0, 0.411, 0.006, 0.0, 0.4132, 0.0038, 0.0, 0.4132, 0.006, -0.0017, 0.4132, -0.0006, -0.0017, 0.414, 0.0015, -0.0039, 0.4132, -0.0006, -0.0039, 0.414, 0.0015, 0.0, 0.4144, 0.0026, 0.0, 0.4147, 0.0037, 0.0, 0.4152, 0.0059, -0.0017, 0.4143, 0.0026, -0.0039, 0.4143, 0.0026,
-0.0017, 0.4147, 0.0038, -0.0017, 0.4153, 0.006, -0.0039, 0.4147, 0.0038, -0.0039, 0.4153, 0.006, 0.0, 0.411, 0.0081, 0.0, 0.411, 0.0103, 0.0, 0.4132, 0.0081, 0.0, 0.4132, 0.0103, 0.0, 0.411, 0.0125, 0.0, 0.411, 0.0146, 0.0, 0.4132, 0.0125, 0.0, 0.4132, 0.0146, 0.0, 0.4156, 0.0081, 0.0, 0.4157, 0.0103, 0.0, 0.4157, 0.0125, 0.0, 0.4157, 0.0146, -0.0017, 0.4156, 0.0081, -0.0017, 0.4157, 0.0103, -0.0039, 0.4156, 0.0081, -0.0039, 0.4157, 0.0103, -0.0017, 0.4157, 0.0125, -0.0017, 0.4157, 0.0146, -0.0039, 0.4157, 0.0125, -0.0039, 0.4157, 0.0146, 0.0, 0.4039, 0.0168, 0.0, 0.4037, 0.0188, 0.0, 0.4031, 0.0215, 0.0, 0.4025, 0.0231, 0.0, 0.4033, 0.0207, 0.0, 0.4045, 0.0233,
-0.0017, 0.4039, 0.0168, -0.0017, 0.4037, 0.019, -0.0039, 0.4039, 0.0168, -0.0039, 0.4037, 0.019, -0.0017, 0.4032, 0.0212, -0.0017, 0.4024, 0.0234, -0.0039, 0.4032, 0.0212, -0.0039, 0.4024, 0.0234, -0.0017, 0.4034, 0.0201, -0.0039, 0.4034, 0.0201, 0.0, 0.4067, 0.0168, 0.0, 0.4067, 0.019, 0.0, 0.4089, 0.0168, 0.0, 0.4089, 0.019, 0.0, 0.4067, 0.0212, 0.0, 0.4067, 0.0233, 0.0, 0.4089, 0.0212, 0.0, 0.4089, 0.0233, 0.0, 0.4015, 0.0251, 0.0, 0.4024, 0.0277, 0.0, 0.4045, 0.0255, 0.0, 0.4045, 0.0277, 0.0, 0.4024, 0.0298, 0.0, 0.4024, 0.032, 0.0, 0.4045, 0.0298, 0.0, 0.4045, 0.032, -0.0017, 0.4016, 0.025, -0.0039, 0.4016, 0.025, 0.0, 0.4067, 0.0255, 0.0, 0.4067, 0.0277,
0.0, 0.4089, 0.0255, 0.0, 0.4089, 0.0277, 0.0, 0.4067, 0.0298, 0.0, 0.4067, 0.032, 0.0, 0.4089, 0.0298, 0.0, 0.4089, 0.032, 0.0, 0.411, 0.0168, 0.0, 0.411, 0.019, 0.0, 0.4132, 0.0168, 0.0, 0.4132, 0.019, 0.0, 0.411, 0.0212, 0.0, 0.411, 0.0233, 0.0, 0.4132, 0.0212, 0.0, 0.4132, 0.0233, 0.0, 0.4157, 0.0168, 0.0, 0.4157, 0.019, 0.0, 0.4155, 0.0212, 0.0, 0.4152, 0.0234, -0.0017, 0.4157, 0.0168, -0.0017, 0.4157, 0.019, -0.0039, 0.4157, 0.0168, -0.0039, 0.4157, 0.019, -0.0017, 0.4155, 0.0212, -0.0017, 0.4152, 0.0233, -0.0039, 0.4155, 0.0212, -0.0039, 0.4152, 0.0233, 0.0, 0.411, 0.0255, 0.0, 0.411, 0.0277, 0.0, 0.4144, 0.026, 0.0, 0.4137, 0.0279,
0.0, 0.411, 0.0298, 0.0, 0.4117, 0.0323, 0.0, 0.4129, 0.03, 0.0, 0.4123, 0.0312, -0.0017, 0.4142, 0.0265, -0.0017, 0.4138, 0.0277, -0.0039, 0.4142, 0.0265, -0.0039, 0.4138, 0.0277, -0.0017, 0.4117, 0.0323, -0.0039, 0.4117, 0.0323, -0.0017, 0.4129, 0.0299, -0.0017, 0.4123, 0.0312, -0.0039, 0.4129, 0.0299, -0.0039, 0.4123, 0.0312, 0.0, 0.4147, 0.0251, -0.0017, 0.4146, 0.0254, -0.0039, 0.4146, 0.0254, 0.0, 0.385, 0.0371, 0.0, 0.3871, 0.0367, 0.0, 0.385, 0.0385, 0.0, 0.385, 0.0407, 0.0, 0.3872, 0.0385, 0.0, 0.3872, 0.0407, -0.0017, 0.385, 0.0371, -0.0039, 0.385, 0.0371, -0.0017, 0.3872, 0.0366, -0.0039, 0.3872, 0.0366, 0.0, 0.3891, 0.0361, 0.0, 0.3919, 0.0348, 0.0, 0.3909, 0.0352,
0.0, 0.3893, 0.0385, 0.0, 0.3893, 0.0407, 0.0, 0.3915, 0.0385, 0.0, 0.3915, 0.0407, -0.0017, 0.3894, 0.036, -0.0039, 0.3894, 0.036, -0.0017, 0.3918, 0.0348, -0.0017, 0.3907, 0.0354, -0.0039, 0.3918, 0.0348, -0.0039, 0.3907, 0.0354, 0.0, 0.385, 0.0429, 0.0, 0.385, 0.045, 0.0, 0.3872, 0.0429, 0.0, 0.3872, 0.045, 0.0, 0.385, 0.0472, 0.0, 0.385, 0.0487, 0.0, 0.3872, 0.0472, 0.0, 0.3872, 0.0486, -0.0017, 0.385, 0.0487, -0.0039, 0.385, 0.0487, -0.0017, 0.3872, 0.0486, -0.0039, 0.3872, 0.0486, 0.0, 0.3893, 0.0429, 0.0, 0.3893, 0.045, 0.0, 0.3915, 0.0429, 0.0, 0.3915, 0.045, 0.0, 0.3896, 0.0483, 0.0, 0.3887, 0.0485, 0.0, 0.3917, 0.0478, -0.0017, 0.39, 0.0482,
-0.0017, 0.389, 0.0484, -0.0039, 0.39, 0.0482, -0.0039, 0.389, 0.0484, -0.0017, 0.3915, 0.0479, -0.0039, 0.3915, 0.0479, 0.0, 0.3936, 0.0338, 0.0, 0.3937, 0.0363, 0.0, 0.3958, 0.0342, 0.0, 0.3958, 0.0363, 0.0, 0.3937, 0.0385, 0.0, 0.3937, 0.0407, 0.0, 0.3958, 0.0385, 0.0, 0.3958, 0.0407, -0.0017, 0.3936, 0.0338, -0.0039, 0.3936, 0.0338, 0.0, 0.398, 0.0342, 0.0, 0.398, 0.0363, 0.0, 0.4002, 0.0342, 0.0, 0.4002, 0.0363, 0.0, 0.398, 0.0385, 0.0, 0.398, 0.0407, 0.0, 0.4002, 0.0385, 0.0, 0.4002, 0.0407, 0.0, 0.3937, 0.0429, 0.0, 0.3937, 0.045, 0.0, 0.3958, 0.0429, 0.0, 0.3964, 0.0461, 0.0, 0.3938, 0.0471, 0.0, 0.3954, 0.0465, -0.0017, 0.3967, 0.046,
-0.0039, 0.3967, 0.046, -0.0017, 0.3937, 0.0472, -0.0039, 0.3937, 0.0472, -0.0017, 0.3956, 0.0465, -0.0039, 0.3956, 0.0465, 0.0, 0.398, 0.0429, 0.0, 0.3983, 0.0453, 0.0, 0.4008, 0.0438, 0.0, 0.3997, 0.0445, -0.0017, 0.398, 0.0454, -0.0039, 0.398, 0.0454, -0.0017, 0.4009, 0.0437, -0.0017, 0.3998, 0.0444, -0.0039, 0.4009, 0.0437, -0.0039, 0.3998, 0.0444, 0.0, 0.4024, 0.0342, 0.0, 0.4024, 0.0363, 0.0, 0.4045, 0.0342, 0.0, 0.4045, 0.0363, 0.0, 0.4024, 0.0385, 0.0, 0.4024, 0.0407, 0.0, 0.4045, 0.0385, 0.0, 0.4046, 0.0409, -0.0017, 0.4046, 0.0409, -0.0039, 0.4046, 0.0409, 0.0, 0.4067, 0.0342, 0.0, 0.4067, 0.0363, 0.0, 0.41, 0.035, 0.0, 0.4088, 0.0367, 0.0, 0.4069, 0.0387,
0.0, 0.4058, 0.0398, 0.0, 0.4079, 0.0376, -0.0017, 0.4099, 0.0352, -0.0017, 0.409, 0.0364, -0.0039, 0.4099, 0.0352, -0.0039, 0.409, 0.0364, -0.0017, 0.4069, 0.0387, -0.0017, 0.4058, 0.0398, -0.0039, 0.4069, 0.0387, -0.0039, 0.4058, 0.0398, -0.0017, 0.4079, 0.0376, -0.0039, 0.4079, 0.0376, 0.0, 0.4026, 0.0426, 0.0, 0.4035, 0.0418, -0.0017, 0.4024, 0.0427, -0.0039, 0.4024, 0.0427, -0.0017, 0.4035, 0.0418, -0.0039, 0.4035, 0.0418, 0.0, 0.4107, 0.034, -0.0017, 0.4106, 0.0341, -0.0039, 0.4106, 0.0341, 0.0, 0.4291, 0.0153, 0.0, 0.4316, 0.0135, 0.0, 0.4304, 0.0142, -0.0017, 0.429, 0.0153, -0.0039, 0.429, 0.0153, -0.0017, 0.4316, 0.0135, -0.0017, 0.4305, 0.0142, -0.0039, 0.4316, 0.0135, -0.0039, 0.4305, 0.0142,
0.0, 0.4326, 0.013, 0.0, 0.4327, 0.0146, 0.0, 0.4346, 0.012, 0.0, 0.4349, 0.0146, -0.0017, 0.4327, 0.0129, -0.0039, 0.4327, 0.0129, -0.0017, 0.4349, 0.0119, -0.0039, 0.4349, 0.0119, 0.0, 0.427, 0.0175, 0.0, 0.4261, 0.0186, 0.0, 0.4244, 0.0214, 0.0, 0.4238, 0.0231, 0.0, 0.425, 0.0204, 0.0, 0.4262, 0.0233, -0.0017, 0.427, 0.0175, -0.0017, 0.4258, 0.0189, -0.0039, 0.427, 0.0175, -0.0039, 0.4258, 0.0189, -0.0017, 0.4245, 0.0211, -0.0017, 0.4237, 0.0233, -0.0039, 0.4245, 0.0211, -0.0039, 0.4237, 0.0233, -0.0017, 0.4252, 0.0201, -0.0039, 0.4252, 0.0201, 0.0, 0.4229, 0.0265, 0.0, 0.4228, 0.0277, 0.0, 0.4226, 0.0298, 0.0, 0.4228, 0.0316, -0.0017, 0.423, 0.0265, -0.0017, 0.4228, 0.0277,
-0.0039, 0.423, 0.0265, -0.0039, 0.4228, 0.0277, -0.0017, 0.4226, 0.0298, -0.0017, 0.4228, 0.0319, -0.0039, 0.4226, 0.0298, -0.0039, 0.4228, 0.0319, 0.0, 0.4231, 0.0255, 0.0, 0.4241, 0.0277, 0.0, 0.4262, 0.0255, 0.0, 0.4262, 0.0277, 0.0, 0.4241, 0.0298, 0.0, 0.4229, 0.0324, 0.0, 0.4262, 0.0298, 0.0, 0.4262, 0.032, -0.0017, 0.4231, 0.0254, -0.0039, 0.4231, 0.0254, -0.0017, 0.423, 0.033, -0.0039, 0.423, 0.033, 0.0, 0.4279, 0.0164, 0.0, 0.4284, 0.019, 0.0, 0.4306, 0.0168, 0.0, 0.4306, 0.019, 0.0, 0.4284, 0.0212, 0.0, 0.4284, 0.0233, 0.0, 0.4306, 0.0212, 0.0, 0.4306, 0.0233, -0.0017, 0.4279, 0.0164, -0.0039, 0.4279, 0.0164, 0.0, 0.4327, 0.0168, 0.0, 0.4327, 0.019,
0.0, 0.4349, 0.0168, 0.0, 0.4349, 0.019, 0.0, 0.4327, 0.0212, 0.0, 0.4327, 0.0233, 0.0, 0.4349, 0.0212, 0.0, 0.4349, 0.0233, 0.0, 0.4284, 0.0255, 0.0, 0.4284, 0.0277, 0.0, 0.4306, 0.0255, 0.0, 0.4306, 0.0277, 0.0, 0.4284, 0.0298, 0.0, 0.4284, 0.032, 0.0, 0.4306, 0.0298, 0.0, 0.4306, 0.032, 0.0, 0.4327, 0.0255, 0.0, 0.4327, 0.0277, 0.0, 0.4349, 0.0255, 0.0, 0.4349, 0.0277, 0.0, 0.4327, 0.0298, 0.0, 0.4327, 0.032, 0.0, 0.4349, 0.0298, 0.0, 0.4349, 0.032, 0.0, 0.4377, 0.0112, 0.0, 0.4392, 0.0109, 0.0, 0.4368, 0.0114, 0.0, 0.4371, 0.0146, 0.0, 0.4392, 0.0125, 0.0, 0.4392, 0.0146, -0.0017, 0.4375, 0.0112, -0.0039, 0.4375, 0.0112,
-0.0017, 0.4392, 0.0109, -0.0039, 0.4392, 0.0109, -0.0017, 0.4364, 0.0115, -0.0039, 0.4364, 0.0115, 0.0, 0.4414, 0.0106, 0.0, 0.4436, 0.0105, 0.0, 0.4414, 0.0125, 0.0, 0.4414, 0.0146, 0.0, 0.4436, 0.0125, 0.0, 0.4436, 0.0146, -0.0017, 0.4414, 0.0106, -0.0039, 0.4414, 0.0106, -0.0017, 0.4436, 0.0105, -0.0039, 0.4436, 0.0105, 0.0, 0.4458, 0.0105, 0.0, 0.4479, 0.0105, 0.0, 0.4458, 0.0125, 0.0, 0.4458, 0.0146, 0.0, 0.4479, 0.0125, 0.0, 0.4479, 0.0146, -0.0017, 0.4458, 0.0105, -0.0039, 0.4458, 0.0105, -0.0017, 0.4479, 0.0105, -0.0039, 0.4479, 0.0105, 0.0, 0.4501, 0.0105, 0.0, 0.4523, 0.0105, 0.0, 0.4501, 0.0125, 0.0, 0.4501, 0.0146, 0.0, 0.4523, 0.0125, 0.0, 0.4523, 0.0146,
-0.0017, 0.4501, 0.0105, -0.0039, 0.4501, 0.0105, -0.0017, 0.4523, 0.0105, -0.0039, 0.4523, 0.0105, 0.0, 0.4371, 0.0168, 0.0, 0.4371, 0.019, 0.0, 0.4392, 0.0168, 0.0, 0.4392, 0.019, 0.0, 0.4371, 0.0212, 0.0, 0.4371, 0.0233, 0.0, 0.4392, 0.0212, 0.0, 0.4398, 0.0242, -0.0017, 0.4399, 0.0241, -0.0039, 0.4399, 0.0241, 0.0, 0.4414, 0.0168, 0.0, 0.4414, 0.019, 0.0, 0.4436, 0.0168, 0.0, 0.4436, 0.019, 0.0, 0.4414, 0.0212, 0.0, 0.4416, 0.0232, 0.0, 0.4436, 0.0212, 0.0, 0.4436, 0.0225, -0.0017, 0.4414, 0.0232, -0.0039, 0.4414, 0.0232, -0.0017, 0.4436, 0.0225, -0.0039, 0.4436, 0.0225, 0.0, 0.4378, 0.026, 0.0, 0.4366, 0.0278, 0.0, 0.4388, 0.025, 0.0, 0.436, 0.0297,
0.0, 0.4366, 0.0319, -0.0017, 0.4377, 0.0261, -0.0017, 0.4367, 0.0276, -0.0039, 0.4377, 0.0261, -0.0039, 0.4367, 0.0276, -0.0017, 0.4388, 0.025, -0.0039, 0.4388, 0.025, -0.0017, 0.436, 0.0297, -0.0017, 0.4367, 0.0321, -0.0039, 0.436, 0.0297, -0.0039, 0.4367, 0.0321, 0.0, 0.4458, 0.0168, 0.0, 0.4458, 0.019, 0.0, 0.4479, 0.0168, 0.0, 0.4479, 0.019, 0.0, 0.4459, 0.0223, 0.0, 0.4453, 0.0223, 0.0, 0.4479, 0.0222, -0.0017, 0.4467, 0.0222, -0.0017, 0.4456, 0.0223, -0.0039, 0.4467, 0.0222, -0.0039, 0.4456, 0.0223, -0.0017, 0.4479, 0.0222, -0.0039, 0.4479, 0.0222, 0.0, 0.4501, 0.0168, 0.0, 0.4501, 0.019, 0.0, 0.4523, 0.0168, 0.0, 0.4523, 0.019, 0.0, 0.4501, 0.0222, 0.0, 0.4523, 0.0222,
-0.0017, 0.4501, 0.0222, -0.0039, 0.4501, 0.0222, -0.0017, 0.4523, 0.0222, -0.0039, 0.4523, 0.0222, 0.0, 0.4232, 0.0344, 0.0, 0.4238, 0.0366, 0.0, 0.4262, 0.0342, 0.0, 0.4262, 0.0363, 0.0, 0.4244, 0.0383, 0.0, 0.4251, 0.0393, 0.0, 0.4262, 0.041, -0.0017, 0.4232, 0.0342, -0.0017, 0.4237, 0.0364, -0.0039, 0.4232, 0.0342, -0.0039, 0.4237, 0.0364, -0.0017, 0.4245, 0.0385, -0.0039, 0.4245, 0.0385, -0.0017, 0.4252, 0.0395, -0.0017, 0.426, 0.0407, -0.0039, 0.4252, 0.0395, -0.0039, 0.426, 0.0407, 0.0, 0.427, 0.042, -0.0017, 0.427, 0.0421, -0.0039, 0.427, 0.0421, 0.0, 0.4284, 0.0342, 0.0, 0.4284, 0.0363, 0.0, 0.4306, 0.0342, 0.0, 0.4306, 0.0363, 0.0, 0.4284, 0.0385, 0.0, 0.4284, 0.0407,
0.0, 0.4306, 0.0385, 0.0, 0.4306, 0.0407, 0.0, 0.4327, 0.0342, 0.0, 0.4327, 0.0363, 0.0, 0.4349, 0.0342, 0.0, 0.4349, 0.0363, 0.0, 0.4327, 0.0385, 0.0, 0.4327, 0.0407, 0.0, 0.4349, 0.0385, 0.0, 0.4349, 0.0407, 0.0, 0.4281, 0.0432, 0.0, 0.4292, 0.0442, 0.0, 0.4306, 0.0429, 0.0, 0.4302, 0.045, -0.0017, 0.4281, 0.0432, -0.0017, 0.4292, 0.0442, -0.0039, 0.4281, 0.0432, -0.0039, 0.4292, 0.0442, -0.0017, 0.4305, 0.0452, -0.0039, 0.4305, 0.0452, 0.0, 0.4327, 0.0429, 0.0, 0.432, 0.0461, 0.0, 0.4349, 0.0429, 0.0, 0.4349, 0.045, 0.0, 0.433, 0.0468, 0.0, 0.4346, 0.0475, -0.0017, 0.4318, 0.046, -0.0039, 0.4318, 0.046, -0.0017, 0.4329, 0.0467, -0.0039, 0.4329, 0.0467,
-0.0017, 0.4349, 0.0476, -0.0039, 0.4349, 0.0476, 0.0, 0.4377, 0.0337, 0.0, 0.4371, 0.0363, 0.0, 0.4388, 0.0347, 0.0, 0.4398, 0.0356, 0.0, 0.4371, 0.0385, 0.0, 0.4371, 0.0407, 0.0, 0.4392, 0.0385, 0.0, 0.4392, 0.0407, -0.0017, 0.4377, 0.0336, -0.0039, 0.4377, 0.0336, -0.0017, 0.4388, 0.0347, -0.0017, 0.4398, 0.0356, -0.0039, 0.4388, 0.0347, -0.0039, 0.4398, 0.0356, 0.0, 0.4416, 0.0367, 0.0, 0.443, 0.0372, 0.0, 0.4414, 0.0385, 0.0, 0.4414, 0.0407, 0.0, 0.4439, 0.0374, 0.0, 0.4436, 0.0407, -0.0017, 0.4414, 0.0366, -0.0039, 0.4414, 0.0366, -0.0017, 0.4432, 0.0373, -0.0039, 0.4432, 0.0373, -0.0017, 0.4443, 0.0375, -0.0039, 0.4443, 0.0375, 0.0, 0.4371, 0.0429, 0.0, 0.4371, 0.045,
0.0, 0.4392, 0.0429, 0.0, 0.4392, 0.045, 0.0, 0.4368, 0.0482, 0.0, 0.4378, 0.0484, 0.0, 0.4392, 0.0472, 0.0, 0.4392, 0.0486, -0.0017, 0.4366, 0.0481, -0.0017, 0.4376, 0.0484, -0.0039, 0.4366, 0.0481, -0.0039, 0.4376, 0.0484, -0.0017, 0.4392, 0.0486, -0.0039, 0.4392, 0.0486, 0.0, 0.4414, 0.0429, 0.0, 0.4414, 0.045, 0.0, 0.4436, 0.0429, 0.0, 0.4436, 0.045, 0.0, 0.4414, 0.0472, 0.0, 0.4414, 0.0487, 0.0, 0.4436, 0.0472, 0.0, 0.4436, 0.0487, -0.0017, 0.4414, 0.0487, -0.0039, 0.4414, 0.0487, -0.0017, 0.4436, 0.0487, -0.0039, 0.4436, 0.0487, 0.0, 0.4459, 0.0378, 0.0, 0.4458, 0.0407, 0.0, 0.4479, 0.0379, 0.0, 0.4479, 0.0407, -0.0017, 0.4457, 0.0378, -0.0039, 0.4457, 0.0378,
-0.0017, 0.4479, 0.0379, -0.0039, 0.4479, 0.0379, 0.0, 0.4501, 0.0379, 0.0, 0.4501, 0.0407, 0.0, 0.4523, 0.0379, 0.0, 0.4523, 0.0407, -0.0017, 0.4501, 0.0379, -0.0039, 0.4501, 0.0379, -0.0017, 0.4523, 0.0379, -0.0039, 0.4523, 0.0379, 0.0, 0.4458, 0.0429, 0.0, 0.4458, 0.045, 0.0, 0.4479, 0.0429, 0.0, 0.4479, 0.045, 0.0, 0.4458, 0.0472, 0.0, 0.4458, 0.0487, 0.0, 0.4479, 0.0472, 0.0, 0.4479, 0.0487, -0.0017, 0.4458, 0.0487, -0.0039, 0.4458, 0.0487, -0.0017, 0.4479, 0.0487, -0.0039, 0.4479, 0.0487, 0.0, 0.4501, 0.0429, 0.0, 0.4501, 0.045, 0.0, 0.4523, 0.0429, 0.0, 0.4523, 0.045, 0.0, 0.4501, 0.0472, 0.0, 0.4501, 0.0487, 0.0, 0.4523, 0.0472, 0.0, 0.4523, 0.0487,
-0.0017, 0.4501, 0.0487, -0.0039, 0.4501, 0.0487, -0.0017, 0.4523, 0.0487, -0.0039, 0.4523, 0.0487, 0.0, 0.4544, 0.0105, 0.0, 0.4566, 0.0105, 0.0, 0.4544, 0.0125, 0.0, 0.4544, 0.0146, 0.0, 0.4566, 0.0125, 0.0, 0.4566, 0.0146, -0.0017, 0.4544, 0.0105, -0.0039, 0.4544, 0.0105, -0.0017, 0.4566, 0.0105, -0.0039, 0.4566, 0.0105, 0.0, 0.4588, 0.0105, 0.0, 0.4609, 0.0105, 0.0, 0.4588, 0.0125, 0.0, 0.4588, 0.0146, 0.0, 0.4609, 0.0125, 0.0, 0.4609, 0.0146, -0.0017, 0.4588, 0.0105, -0.0039, 0.4588, 0.0105, -0.0017, 0.4609, 0.0105, -0.0039, 0.4609, 0.0105, 0.0, 0.4631, 0.0105, 0.0, 0.4653, 0.0105, 0.0, 0.4631, 0.0125, 0.0, 0.4631, 0.0146, 0.0, 0.4653, 0.0125, 0.0, 0.4653, 0.0146,
-0.0017, 0.4631, 0.0105, -0.0039, 0.4631, 0.0105, -0.0017, 0.4653, 0.0105, -0.0039, 0.4653, 0.0105, 0.0, 0.4675, 0.0105, 0.0, 0.4696, 0.0105, 0.0, 0.4675, 0.0125, 0.0, 0.4675, 0.0146, 0.0, 0.4696, 0.0125, 0.0, 0.4696, 0.0146, -0.0017, 0.4675, 0.0105, -0.0039, 0.4675, 0.0105, -0.0017, 0.4696, 0.0105, -0.0039, 0.4696, 0.0105, 0.0, 0.4544, 0.0168, 0.0, 0.4544, 0.019, 0.0, 0.4566, 0.0168, 0.0, 0.4566, 0.019, 0.0, 0.4544, 0.0222, 0.0, 0.4566, 0.0222, -0.0017, 0.4544, 0.0222, -0.0039, 0.4544, 0.0222, -0.0017, 0.4566, 0.0222, -0.0039, 0.4566, 0.0222, 0.0, 0.4588, 0.0168, 0.0, 0.4588, 0.019, 0.0, 0.4609, 0.0168, 0.0, 0.4609, 0.019, 0.0, 0.4588, 0.0222, 0.0, 0.4609, 0.0222,
-0.0017, 0.4588, 0.0222, -0.0039, 0.4588, 0.0222, -0.0017, 0.4609, 0.0222, -0.0039, 0.4609, 0.0222, 0.0, 0.4631, 0.0168, 0.0, 0.4631, 0.019, 0.0, 0.4653, 0.0168, 0.0, 0.4653, 0.019, 0.0, 0.4631, 0.0222, 0.0, 0.4653, 0.0222, -0.0017, 0.4631, 0.0222, -0.0039, 0.4631, 0.0222, -0.0017, 0.4653, 0.0222, -0.0039, 0.4653, 0.0222, 0.0, 0.4675, 0.0168, 0.0, 0.4675, 0.019, 0.0, 0.4696, 0.0168, 0.0, 0.4696, 0.019, 0.0, 0.4675, 0.0222, 0.0, 0.4696, 0.0222, -0.0017, 0.4675, 0.0222, -0.0039, 0.4675, 0.0222, -0.0017, 0.4696, 0.0222, -0.0039, 0.4696, 0.0222, 0.0, 0.4718, 0.0105, 0.0, 0.474, 0.0105, 0.0, 0.4718, 0.0125, 0.0, 0.4718, 0.0146, 0.0, 0.474, 0.0125, 0.0, 0.474, 0.0146,
-0.0017, 0.4718, 0.0105, -0.0039, 0.4718, 0.0105, -0.0017, 0.474, 0.0105, -0.0039, 0.474, 0.0105, 0.0, 0.4761, 0.0104, 0.0, 0.4783, 0.0104, 0.0, 0.4761, 0.0125, 0.0, 0.4761, 0.0146, 0.0, 0.4783, 0.0125, 0.0, 0.4783, 0.0146, -0.0017, 0.4761, 0.0104, -0.0039, 0.4761, 0.0104, -0.0017, 0.4783, 0.0104, -0.0039, 0.4783, 0.0104, 0.0, 0.4877, 0.0067, -0.0017, 0.4877, 0.0067, -0.0039, 0.4877, 0.0067, 0.0, 0.4804, 0.0102, 0.0, 0.4837, 0.0092, 0.0, 0.4826, 0.0096, 0.0, 0.4805, 0.0125, 0.0, 0.4805, 0.0146, 0.0, 0.4826, 0.0125, 0.0, 0.4826, 0.0146, -0.0017, 0.4805, 0.0102, -0.0039, 0.4805, 0.0102, -0.0017, 0.4836, 0.0092, -0.0017, 0.4826, 0.0096, -0.0039, 0.4836, 0.0092, -0.0039, 0.4826, 0.0096,
0.0, 0.4847, 0.0087, 0.0, 0.4848, 0.0103, 0.0, 0.4866, 0.0076, 0.0, 0.487, 0.0103, 0.0, 0.4848, 0.0125, 0.0, 0.4848, 0.0146, 0.0, 0.487, 0.0125, 0.0, 0.487, 0.0146, -0.0017, 0.4849, 0.0087, -0.0039, 0.4849, 0.0087, -0.0017, 0.4866, 0.0076, -0.0039, 0.4866, 0.0076, 0.0, 0.4718, 0.0168, 0.0, 0.4718, 0.019, 0.0, 0.474, 0.0168, 0.0, 0.474, 0.019, 0.0, 0.4718, 0.0222, 0.0, 0.474, 0.0222, -0.0017, 0.4718, 0.0222, -0.0039, 0.4718, 0.0222, -0.0017, 0.474, 0.0222, -0.0039, 0.474, 0.0222, 0.0, 0.4761, 0.0168, 0.0, 0.4761, 0.019, 0.0, 0.4783, 0.0168, 0.0, 0.4783, 0.019, 0.0, 0.4761, 0.0222, 0.0, 0.4783, 0.0222, -0.0017, 0.4761, 0.0222, -0.0039, 0.4761, 0.0222,
-0.0017, 0.4783, 0.0222, -0.0039, 0.4783, 0.0222, 0.0, 0.4805, 0.0168, 0.0, 0.4805, 0.019, 0.0, 0.4826, 0.0168, 0.0, 0.4826, 0.019, 0.0, 0.4806, 0.0221, 0.0, 0.4828, 0.0219, -0.0017, 0.4805, 0.0222, -0.0039, 0.4805, 0.0222, -0.0017, 0.4827, 0.0219, -0.0039, 0.4827, 0.0219, 0.0, 0.4848, 0.0168, 0.0, 0.4848, 0.019, 0.0, 0.487, 0.0168, 0.0, 0.487, 0.019, 0.0, 0.485, 0.0214, 0.0, 0.4871, 0.0208, -0.0017, 0.4848, 0.0214, -0.0039, 0.4848, 0.0214, -0.0017, 0.487, 0.0208, -0.0039, 0.487, 0.0208, 0.0, 0.4544, 0.0379, 0.0, 0.4544, 0.0407, 0.0, 0.4566, 0.0379, 0.0, 0.4566, 0.0407, -0.0017, 0.4544, 0.0379, -0.0039, 0.4544, 0.0379, -0.0017, 0.4566, 0.0379, -0.0039, 0.4566, 0.0379,
0.0, 0.4588, 0.0379, 0.0, 0.4588, 0.0407, 0.0, 0.4609, 0.0379, 0.0, 0.4609, 0.0407, -0.0017, 0.4588, 0.0379, -0.0039, 0.4588, 0.0379, -0.0017, 0.4609, 0.0379, -0.0039, 0.4609, 0.0379, 0.0, 0.4544, 0.0429, 0.0, 0.4544, 0.045, 0.0, 0.4566, 0.0429, 0.0, 0.4566, 0.045, 0.0, 0.4544, 0.0472, 0.0, 0.4544, 0.0487, 0.0, 0.4566, 0.0472, 0.0, 0.4566, 0.0487, -0.0017, 0.4544, 0.0487, -0.0039, 0.4544, 0.0487, -0.0017, 0.4566, 0.0487, -0.0039, 0.4566, 0.0487, 0.0, 0.4588, 0.0429, 0.0, 0.4588, 0.045, 0.0, 0.4609, 0.0429, 0.0, 0.4609, 0.045, 0.0, 0.4588, 0.0472, 0.0, 0.4588, 0.0487, 0.0, 0.4609, 0.0472, 0.0, 0.4609, 0.0487, -0.0017, 0.4588, 0.0487, -0.0039, 0.4588, 0.0487,
-0.0017, 0.4609, 0.0487, -0.0039, 0.4609, 0.0487, 0.0, 0.4631, 0.0379, 0.0, 0.4631, 0.0407, 0.0, 0.4653, 0.0379, 0.0, 0.4653, 0.0407, -0.0017, 0.4631, 0.0379, -0.0039, 0.4631, 0.0379, -0.0017, 0.4653, 0.0379, -0.0039, 0.4653, 0.0379, 0.0, 0.4675, 0.0379, 0.0, 0.4675, 0.0407, 0.0, 0.4696, 0.0379, 0.0, 0.4696, 0.0407, -0.0017, 0.4675, 0.0379, -0.0039, 0.4675, 0.0379, -0.0017, 0.4696, 0.0379, -0.0039, 0.4696, 0.0379, 0.0, 0.4631, 0.0429, 0.0, 0.4631, 0.045, 0.0, 0.4653, 0.0429, 0.0, 0.4653, 0.045, 0.0, 0.4631, 0.0472, 0.0, 0.4631, 0.0487, 0.0, 0.4653, 0.0472, 0.0, 0.4653, 0.0487, -0.0017, 0.4631, 0.0487, -0.0039, 0.4631, 0.0487, -0.0017, 0.4653, 0.0487, -0.0039, 0.4653, 0.0487,
0.0, 0.4675, 0.0429, 0.0, 0.4675, 0.045, 0.0, 0.4696, 0.0429, 0.0, 0.4696, 0.045, 0.0, 0.4675, 0.0472, 0.0, 0.4675, 0.0487, 0.0, 0.4696, 0.0472, 0.0, 0.4696, 0.0487, -0.0017, 0.4675, 0.0487, -0.0039, 0.4675, 0.0487, -0.0017, 0.4696, 0.0487, -0.0039, 0.4696, 0.0487, 0.0, 0.4718, 0.0379, 0.0, 0.4718, 0.0407, 0.0, 0.474, 0.0379, 0.0, 0.474, 0.0407, -0.0017, 0.4718, 0.0379, -0.0039, 0.4718, 0.0379, -0.0017, 0.474, 0.0379, -0.0039, 0.474, 0.0379, 0.0, 0.4761, 0.0379, 0.0, 0.4761, 0.0407, 0.0, 0.4783, 0.0379, 0.0, 0.4783, 0.0407, -0.0017, 0.4761, 0.0379, -0.0039, 0.4761, 0.0379, -0.0017, 0.4783, 0.0379, -0.0039, 0.4783, 0.0379, 0.0, 0.4718, 0.0429, 0.0, 0.4718, 0.045,
0.0, 0.474, 0.0429, 0.0, 0.474, 0.045, 0.0, 0.4718, 0.0472, 0.0, 0.4718, 0.0487, 0.0, 0.474, 0.0472, 0.0, 0.474, 0.0487, -0.0017, 0.4718, 0.0487, -0.0039, 0.4718, 0.0487, -0.0017, 0.474, 0.0487, -0.0039, 0.474, 0.0487, 0.0, 0.4761, 0.0429, 0.0, 0.4761, 0.045, 0.0, 0.4783, 0.0429, 0.0, 0.4783, 0.045, 0.0, 0.4761, 0.0472, 0.0, 0.4761, 0.0487, 0.0, 0.4783, 0.0472, 0.0, 0.4783, 0.0487, -0.0017, 0.4761, 0.0487, -0.0039, 0.4761, 0.0487, -0.0017, 0.4783, 0.0487, -0.0039, 0.4783, 0.0487, 0.0, 0.4805, 0.0379, 0.0, 0.4805, 0.0407, 0.0, 0.4826, 0.0379, 0.0, 0.4826, 0.0407, -0.0017, 0.4805, 0.0379, -0.0039, 0.4805, 0.0379, -0.0017, 0.4826, 0.0379, -0.0039, 0.4826, 0.0379,
0.0, 0.4848, 0.0379, 0.0, 0.4848, 0.0407, 0.0, 0.487, 0.0379, 0.0, 0.487, 0.0407, -0.0017, 0.4848, 0.0379, -0.0039, 0.4848, 0.0379, -0.0017, 0.487, 0.0379, -0.0039, 0.487, 0.0379, 0.0, 0.4805, 0.0429, 0.0, 0.4805, 0.045, 0.0, 0.4826, 0.0429, 0.0, 0.4826, 0.045, 0.0, 0.4805, 0.0472, 0.0, 0.4805, 0.0487, 0.0, 0.4826, 0.0472, 0.0, 0.4826, 0.0487, -0.0017, 0.4805, 0.0487, -0.0039, 0.4805, 0.0487, -0.0017, 0.4826, 0.0487, -0.0039, 0.4826, 0.0487, 0.0, 0.4848, 0.0429, 0.0, 0.4848, 0.045, 0.0, 0.487, 0.0429, 0.0, 0.487, 0.045, 0.0, 0.4848, 0.0472, 0.0, 0.4848, 0.0487, 0.0, 0.487, 0.0472, 0.0, 0.487, 0.0487, -0.0017, 0.4848, 0.0487, -0.0039, 0.4848, 0.0487,
-0.0017, 0.487, 0.0487, -0.0039, 0.487, 0.0487, 0.0, 0.491, -0.0004, 0.0, 0.4911, 0.0017, 0.0, 0.4897, 0.0043, 0.0, 0.4888, 0.0057, 0.0, 0.4904, 0.0033, 0.0, 0.4913, 0.006, -0.0017, 0.491, -0.0006, -0.0017, 0.4911, 0.0017, -0.0039, 0.491, -0.0006, -0.0039, 0.4911, 0.0017, -0.0017, 0.4898, 0.0042, -0.0017, 0.4888, 0.0057, -0.0039, 0.4898, 0.0042, -0.0039, 0.4888, 0.0057, -0.0017, 0.4905, 0.0032, -0.0039, 0.4905, 0.0032, 0.0, 0.4935, -0.0005, 0.0, 0.4935, 0.0016, 0.0, 0.4957, -0.0005, 0.0, 0.4957, 0.0016, 0.0, 0.4935, 0.0038, 0.0, 0.4935, 0.006, 0.0, 0.4957, 0.0038, 0.0, 0.4957, 0.006, 0.0, 0.4892, 0.0081, 0.0, 0.4892, 0.0103, 0.0, 0.4913, 0.0081, 0.0, 0.4913, 0.0103,
0.0, 0.4892, 0.0125, 0.0, 0.4892, 0.0146, 0.0, 0.4913, 0.0125, 0.0, 0.4913, 0.0146, 0.0, 0.4935, 0.0081, 0.0, 0.4935, 0.0103, 0.0, 0.4957, 0.0081, 0.0, 0.4957, 0.0103, 0.0, 0.4935, 0.0125, 0.0, 0.4935, 0.0146, 0.0, 0.4957, 0.0125, 0.0, 0.4965, 0.0154, -0.0017, 0.4965, 0.0155, -0.0039, 0.4965, 0.0155, 0.0, 0.4978, -0.0005, 0.0, 0.4978, 0.0016, 0.0, 0.5, -0.0005, 0.0, 0.5, 0.0016, 0.0, 0.4978, 0.0038, 0.0, 0.4978, 0.006, 0.0, 0.5, 0.0038, 0.0, 0.5, 0.006, 0.0, 0.502, -0.0005, 0.0, 0.502, 0.0016, 0.0, 0.5019, 0.0038, 0.0, 0.5017, 0.006, -0.0017, 0.502, -0.0005, -0.0017, 0.502, 0.0016, -0.0039, 0.502, -0.0005, -0.0039, 0.502, 0.0016,
-0.0017, 0.5019, 0.0038, -0.0017, 0.5017, 0.006, -0.0039, 0.5019, 0.0038, -0.0039, 0.5017, 0.006, 0.0, 0.4978, 0.0081, 0.0, 0.4978, 0.0103, 0.0, 0.501, 0.0085, 0.0, 0.5001, 0.0106, 0.0, 0.4986, 0.013, 0.0, 0.4975, 0.0143, 0.0, 0.4993, 0.0118, -0.0017, 0.5009, 0.0087, -0.0017, 0.5002, 0.0103, -0.0039, 0.5009, 0.0087, -0.0039, 0.5002, 0.0103, -0.0017, 0.4985, 0.013, -0.0017, 0.4974, 0.0144, -0.0039, 0.4985, 0.013, -0.0039, 0.4974, 0.0144, -0.0017, 0.4992, 0.0119, -0.0039, 0.4992, 0.0119, 0.0, 0.5013, 0.0075, -0.0017, 0.5013, 0.0076, -0.0039, 0.5013, 0.0076, 0.0, 0.4892, 0.0168, 0.0, 0.4895, 0.0199, 0.0, 0.4913, 0.0168, 0.0, 0.4916, 0.019, 0.0, 0.4884, 0.0203, -0.0017, 0.4896, 0.0198,
-0.0039, 0.4896, 0.0198, -0.0017, 0.4914, 0.0191, -0.0039, 0.4914, 0.0191, -0.0017, 0.4886, 0.0203, -0.0039, 0.4886, 0.0203, 0.0, 0.4939, 0.0176, 0.0, 0.4928, 0.0182, 0.0, 0.4954, 0.0165, -0.0017, 0.494, 0.0175, -0.0017, 0.4929, 0.0182, -0.0039, 0.494, 0.0175, -0.0039, 0.4929, 0.0182, -0.0017, 0.4954, 0.0165, -0.0039, 0.4954, 0.0165, 0.0, 0.4892, 0.0379, 0.0, 0.4892, 0.0407, 0.0, 0.4913, 0.0379, 0.0, 0.4913, 0.0407, -0.0017, 0.4892, 0.0379, -0.0039, 0.4892, 0.0379, -0.0017, 0.4913, 0.0379, -0.0039, 0.4913, 0.0379, 0.0, 0.4935, 0.0379, 0.0, 0.4935, 0.0407, 0.0, 0.4957, 0.0379, 0.0, 0.4957, 0.0407, -0.0017, 0.4935, 0.0379, -0.0039, 0.4935, 0.0379, -0.0017, 0.4957, 0.0379, -0.0039, 0.4957, 0.0379,
0.0, 0.4892, 0.0429, 0.0, 0.4892, 0.045, 0.0, 0.4913, 0.0429, 0.0, 0.4913, 0.045, 0.0, 0.4892, 0.0472, 0.0, 0.4892, 0.0487, 0.0, 0.4913, 0.0472, 0.0, 0.4913, 0.0487, -0.0017, 0.4892, 0.0487, -0.0039, 0.4892, 0.0487, -0.0017, 0.4913, 0.0487, -0.0039, 0.4913, 0.0487, 0.0, 0.4935, 0.0429, 0.0, 0.4935, 0.045, 0.0, 0.4957, 0.0429, 0.0, 0.4957, 0.045, 0.0, 0.4935, 0.0472, 0.0, 0.4935, 0.0487, 0.0, 0.4957, 0.0472, 0.0, 0.4957, 0.0487, -0.0017, 0.4935, 0.0487, -0.0039, 0.4935, 0.0487, -0.0017, 0.4957, 0.0487, -0.0039, 0.4957, 0.0487, 0.0, 0.498, 0.0379, 0.0, 0.498, 0.0407, -0.0017, 0.498, 0.0379, -0.0017, 0.498, 0.0407, -0.0039, 0.498, 0.0379, -0.0039, 0.498, 0.0407,
0.0, 0.498, 0.0429, 0.0, 0.498, 0.045, 0.0, 0.498, 0.0472, 0.0, 0.498, 0.0487, -0.0017, 0.498, 0.0429, -0.0017, 0.498, 0.045, -0.0039, 0.498, 0.0429, -0.0039, 0.498, 0.045, -0.0017, 0.498, 0.0472, -0.0017, 0.498, 0.0487, -0.0039, 0.498, 0.0472, -0.0039, 0.498, 0.0487, -0.0061, 0.0461, -0.067, -0.0083, 0.0461, -0.067, -0.0061, 0.0486, -0.067, -0.0083, 0.0486, -0.067, -0.0061, 0.0461, -0.0656, -0.0061, 0.0461, -0.0635, -0.0083, 0.0461, -0.0656, -0.0083, 0.0461, -0.0635, -0.01, 0.0461, -0.067, -0.01, 0.0486, -0.067, -0.01, 0.0461, -0.0656, -0.01, 0.0461, -0.0635, -0.01, 0.0486, -0.0656, -0.01, 0.0486, -0.0635, -0.0061, 0.05, -0.067, -0.0083, 0.05, -0.067, -0.0061, 0.05, -0.0656, -0.0061, 0.05, -0.0635,
-0.0083, 0.05, -0.0656, -0.0083, 0.05, -0.0635, -0.01, 0.05, -0.067, -0.01, 0.05, -0.0656, -0.01, 0.05, -0.0635, -0.0061, 0.0461, -0.0613, -0.0061, 0.0461, -0.0591, -0.0083, 0.0461, -0.0613, -0.0083, 0.0461, -0.0591, -0.0061, 0.0461, -0.057, -0.0061, 0.0461, -0.0548, -0.0083, 0.0461, -0.057, -0.0083, 0.0461, -0.0548, -0.01, 0.0461, -0.0613, -0.01, 0.0461, -0.0591, -0.01, 0.0486, -0.0613, -0.01, 0.0486, -0.0591, -0.01, 0.0461, -0.057, -0.01, 0.0461, -0.0548, -0.01, 0.0486, -0.057, -0.01, 0.0486, -0.0548, -0.0061, 0.05, -0.0613, -0.0061, 0.05, -0.0591, -0.0083, 0.05, -0.0613, -0.0083, 0.05, -0.0591, -0.0061, 0.05, -0.057, -0.0061, 0.05, -0.0548, -0.0083, 0.05, -0.057, -0.0083, 0.05, -0.0548, -0.0061, 0.0529, -0.0562,
-0.0061, 0.0529, -0.0548, -0.0083, 0.0529, -0.0562, -0.0083, 0.0529, -0.0548, -0.01, 0.05, -0.0613, -0.01, 0.05, -0.0591, -0.01, 0.05, -0.057, -0.01, 0.05, -0.0548, -0.01, 0.0529, -0.0562, -0.01, 0.0529, -0.0548, -0.0061, 0.0461, -0.0526, -0.0061, 0.0461, -0.0505, -0.0083, 0.0461, -0.0526, -0.0083, 0.0461, -0.0505, -0.0061, 0.0461, -0.0483, -0.0061, 0.0461, -0.0461, -0.0083, 0.0461, -0.0483, -0.0083, 0.0461, -0.0461, -0.01, 0.0461, -0.0526, -0.01, 0.0461, -0.0505, -0.01, 0.0486, -0.0526, -0.01, 0.0486, -0.0505, -0.01, 0.0461, -0.0483, -0.01, 0.0461, -0.0461, -0.01, 0.0486, -0.0483, -0.01, 0.0486, -0.0461, -0.0061, 0.05, -0.0526, -0.0061, 0.05, -0.0505, -0.0083, 0.05, -0.0526, -0.0083, 0.05, -0.0505, -0.0061, 0.0529, -0.0523,
-0.0083, 0.0529, -0.0523, -0.0061, 0.05, -0.0483, -0.0061, 0.05, -0.0454, -0.0083, 0.05, -0.0483, -0.0083, 0.05, -0.0454, -0.0061, 0.053, -0.0454, -0.0083, 0.053, -0.0454, -0.01, 0.05, -0.0526, -0.01, 0.05, -0.0505, -0.01, 0.0529, -0.0523, -0.01, 0.05, -0.0483, -0.01, 0.05, -0.0454, -0.01, 0.053, -0.0454, -0.0061, 0.0461, -0.0439, -0.0061, 0.0461, -0.0425, -0.0083, 0.0461, -0.0439, -0.0083, 0.0461, -0.0425, -0.0061, 0.0486, -0.0425, -0.0083, 0.0486, -0.0425, -0.01, 0.0461, -0.0439, -0.01, 0.0461, -0.0425, -0.01, 0.0486, -0.0439, -0.01, 0.0486, -0.0425, -0.0061, 0.0508, -0.0425, -0.0083, 0.0508, -0.0425, -0.0061, 0.053, -0.0425, -0.0083, 0.053, -0.0425, -0.01, 0.0508, -0.0439, -0.01, 0.0508, -0.0425, -0.01, 0.053, -0.0439,
-0.01, 0.053, -0.0425, -0.0061, 0.0551, -0.0562, -0.0083, 0.0551, -0.0562, -0.0061, 0.0573, -0.0562, -0.0083, 0.0573, -0.0562, -0.01, 0.0551, -0.0562, -0.01, 0.0551, -0.0548, -0.01, 0.0573, -0.0562, -0.01, 0.0573, -0.0548, -0.0061, 0.0595, -0.0562, -0.0083, 0.0595, -0.0562, -0.0061, 0.0616, -0.0562, -0.0083, 0.0616, -0.0562, -0.01, 0.0595, -0.0562, -0.01, 0.0595, -0.0548, -0.01, 0.0616, -0.0562, -0.01, 0.0616, -0.0548, -0.0061, 0.0638, -0.0562, -0.0083, 0.0638, -0.0562, -0.0061, 0.0657, -0.0562, -0.0061, 0.0657, -0.0548, -0.0083, 0.0657, -0.0562, -0.0083, 0.0657, -0.0548, -0.01, 0.0638, -0.0562, -0.01, 0.0638, -0.0548, -0.01, 0.0657, -0.0562, -0.01, 0.0657, -0.0548, -0.0061, 0.0551, -0.0523, -0.0083, 0.0551, -0.0523, -0.0061, 0.0573, -0.0523,
-0.0083, 0.0573, -0.0523, -0.0061, 0.0551, -0.0454, -0.0083, 0.0551, -0.0454, -0.0061, 0.0573, -0.0454, -0.0083, 0.0573, -0.0454, -0.01, 0.0551, -0.0523, -0.01, 0.0573, -0.0523, -0.01, 0.0551, -0.0454, -0.01, 0.0573, -0.0454, -0.0061, 0.0595, -0.0523, -0.0083, 0.0595, -0.0523, -0.0061, 0.0616, -0.0523, -0.0083, 0.0616, -0.0523, -0.0061, 0.0595, -0.0454, -0.0083, 0.0595, -0.0454, -0.0061, 0.0616, -0.0454, -0.0083, 0.0616, -0.0454, -0.01, 0.0595, -0.0523, -0.01, 0.0616, -0.0523, -0.01, 0.0595, -0.0454, -0.01, 0.0616, -0.0454, -0.0061, 0.0551, -0.0425, -0.0083, 0.0551, -0.0425, -0.0061, 0.0573, -0.0425, -0.0083, 0.0573, -0.0425, -0.01, 0.0551, -0.0439, -0.01, 0.0551, -0.0425, -0.01, 0.0573, -0.0439, -0.01, 0.0573, -0.0425, -0.0061, 0.0595, -0.0425,
-0.0083, 0.0595, -0.0425, -0.0061, 0.0616, -0.0425, -0.0083, 0.0616, -0.0425, -0.01, 0.0595, -0.0439, -0.01, 0.0595, -0.0425, -0.01, 0.0616, -0.0439, -0.01, 0.0616, -0.0425, -0.0061, 0.0638, -0.0523, -0.0083, 0.0638, -0.0523, -0.0061, 0.0657, -0.0523, -0.0083, 0.0657, -0.0523, -0.0061, 0.0638, -0.0454, -0.0083, 0.0638, -0.0454, -0.0061, 0.0667, -0.0454, -0.0083, 0.0667, -0.0454, -0.01, 0.0638, -0.0523, -0.01, 0.0657, -0.0523, -0.01, 0.0638, -0.0454, -0.01, 0.0667, -0.0454, -0.0061, 0.0638, -0.0425, -0.0083, 0.0638, -0.0425, -0.0061, 0.0667, -0.0439, -0.0061, 0.0667, -0.0425, -0.0083, 0.0667, -0.0439, -0.0083, 0.0667, -0.0425, -0.01, 0.0638, -0.0439, -0.01, 0.0638, -0.0425, -0.01, 0.0667, -0.0439, -0.01, 0.0667, -0.0425, -0.0061, 0.0887, -0.0625,
-0.0083, 0.0887, -0.0625, -0.01, 0.0887, -0.0624, -0.0061, 0.0879, -0.0614, -0.0061, 0.0869, -0.0592, -0.0083, 0.0879, -0.0614, -0.0083, 0.0869, -0.0592, -0.0061, 0.0864, -0.0568, -0.0061, 0.0863, -0.0548, -0.0083, 0.0864, -0.0568, -0.0083, 0.0863, -0.0548, -0.0061, 0.0866, -0.0579, -0.0083, 0.0866, -0.0579, -0.01, 0.088, -0.0616, -0.01, 0.087, -0.0594, -0.01, 0.0864, -0.0565, -0.01, 0.0863, -0.0548, -0.01, 0.0865, -0.0573, -0.01, 0.0877, -0.0548, -0.0061, 0.0863, -0.0526, -0.0061, 0.0865, -0.0511, -0.0083, 0.0863, -0.0526, -0.0083, 0.0865, -0.0511, -0.0061, 0.0867, -0.0501, -0.0083, 0.0867, -0.0501, -0.0061, 0.0874, -0.0482, -0.0061, 0.0884, -0.0468, -0.0083, 0.0874, -0.0482, -0.0083, 0.0884, -0.0468, -0.01, 0.0863, -0.0526, -0.01, 0.0865, -0.0513,
-0.01, 0.0877, -0.0526, -0.01, 0.0867, -0.0503, -0.01, 0.0875, -0.048, -0.01, 0.0884, -0.0468, -0.0061, 0.0909, -0.0646, -0.0061, 0.0897, -0.0637, -0.0083, 0.0909, -0.0646, -0.0083, 0.0897, -0.0637, -0.0061, 0.092, -0.0653, -0.0083, 0.092, -0.0653, -0.01, 0.0909, -0.0646, -0.01, 0.0897, -0.0637, -0.01, 0.0918, -0.0652, -0.01, 0.092, -0.0635, -0.0061, 0.097, -0.0668, -0.0083, 0.097, -0.0668, -0.0061, 0.0942, -0.0662, -0.0083, 0.0942, -0.0662, -0.0061, 0.0959, -0.0666, -0.0061, 0.0964, -0.0627, -0.0083, 0.0959, -0.0666, -0.0083, 0.0964, -0.0627, -0.01, 0.0972, -0.0668, -0.01, 0.0939, -0.0661, -0.01, 0.0942, -0.0635, -0.01, 0.0962, -0.0667, -0.01, 0.0963, -0.0627, -0.0061, 0.0908, -0.0583, -0.0083, 0.0908, -0.0583, -0.0061, 0.0926, -0.0608,
-0.0061, 0.0914, -0.0594, -0.0083, 0.0926, -0.0608, -0.0083, 0.0914, -0.0594, -0.0061, 0.0904, -0.057, -0.0061, 0.0903, -0.0548, -0.0083, 0.0904, -0.057, -0.0083, 0.0903, -0.0548, -0.01, 0.0899, -0.0613, -0.01, 0.091, -0.0586, -0.01, 0.0926, -0.0608, -0.01, 0.0915, -0.0595, -0.01, 0.0904, -0.0568, -0.01, 0.0902, -0.0548, -0.0061, 0.0942, -0.0621, -0.0083, 0.0942, -0.0621, -0.0061, 0.0953, -0.0624, -0.0083, 0.0953, -0.0624, -0.01, 0.0942, -0.0621, -0.01, 0.0953, -0.0624, -0.0061, 0.0985, -0.0669, -0.0083, 0.0985, -0.0669, -0.0061, 0.1007, -0.067, -0.0083, 0.1007, -0.067, -0.0061, 0.0985, -0.063, -0.0083, 0.0985, -0.063, -0.0061, 0.1007, -0.0631, -0.0083, 0.1007, -0.0631, -0.01, 0.0985, -0.0669, -0.01, 0.1007, -0.067, -0.01, 0.0985, -0.0656,
-0.01, 0.0986, -0.063, -0.01, 0.1007, -0.0656, -0.01, 0.1007, -0.0631, -0.0061, 0.1029, -0.067, -0.0083, 0.1029, -0.067, -0.0061, 0.105, -0.067, -0.0083, 0.105, -0.067, -0.0061, 0.1029, -0.0631, -0.0083, 0.1029, -0.0631, -0.0061, 0.105, -0.063, -0.0083, 0.105, -0.063, -0.01, 0.1029, -0.067, -0.01, 0.105, -0.0669, -0.01, 0.1029, -0.0656, -0.01, 0.1029, -0.0631, -0.01, 0.105, -0.0656, -0.01, 0.105, -0.063, -0.0061, 0.0904, -0.0526, -0.0061, 0.0908, -0.0512, -0.0083, 0.0904, -0.0526, -0.0083, 0.0908, -0.0512, -0.0061, 0.0912, -0.05, -0.0083, 0.0912, -0.05, -0.0061, 0.0894, -0.0456, -0.0083, 0.0894, -0.0456, -0.0061, 0.0923, -0.0485, -0.0083, 0.0923, -0.0485, -0.01, 0.0904, -0.0528, -0.01, 0.0909, -0.0508, -0.01, 0.0913, -0.0499,
-0.01, 0.0899, -0.0483, -0.01, 0.0894, -0.0456, -0.01, 0.0922, -0.0485, -0.01, 0.092, -0.0461, -0.0061, 0.0935, -0.0474, -0.0061, 0.0946, -0.0468, -0.0083, 0.0935, -0.0474, -0.0083, 0.0946, -0.0468, -0.0061, 0.0963, -0.0461, -0.0083, 0.0963, -0.0461, -0.01, 0.0934, -0.0475, -0.01, 0.0945, -0.0469, -0.01, 0.0965, -0.046, -0.0061, 0.0905, -0.0447, -0.0083, 0.0905, -0.0447, -0.0061, 0.092, -0.0437, -0.0083, 0.092, -0.0437, -0.01, 0.0905, -0.0447, -0.01, 0.0917, -0.0438, -0.0061, 0.0937, -0.043, -0.0061, 0.0948, -0.0427, -0.0083, 0.0937, -0.043, -0.0083, 0.0948, -0.0427, -0.0061, 0.0964, -0.0425, -0.0083, 0.0964, -0.0425, -0.01, 0.0939, -0.0429, -0.01, 0.0949, -0.0427, -0.01, 0.0964, -0.0439, -0.01, 0.0963, -0.0425, -0.0061, 0.0985, -0.0456,
-0.0083, 0.0985, -0.0456, -0.0061, 0.1007, -0.0454, -0.0083, 0.1007, -0.0454, -0.01, 0.0986, -0.0456, -0.01, 0.1007, -0.0454, -0.0061, 0.1029, -0.0455, -0.0083, 0.1029, -0.0455, -0.0061, 0.1051, -0.0457, -0.0083, 0.1051, -0.0457, -0.01, 0.1029, -0.0455, -0.01, 0.105, -0.0457, -0.0061, 0.0985, -0.0425, -0.0083, 0.0985, -0.0425, -0.0061, 0.1007, -0.0425, -0.0083, 0.1007, -0.0425, -0.01, 0.0985, -0.0439, -0.01, 0.0985, -0.0425, -0.01, 0.1007, -0.0439, -0.01, 0.1007, -0.0425, -0.0061, 0.1029, -0.0425, -0.0083, 0.1029, -0.0425, -0.0061, 0.105, -0.0425, -0.0083, 0.105, -0.0425, -0.01, 0.1029, -0.0439, -0.01, 0.1029, -0.0425, -0.01, 0.105, -0.0439, -0.01, 0.105, -0.0425, -0.0061, 0.0716, -0.0219, -0.0061, 0.0716, -0.0201, -0.0083, 0.0716, -0.0219,
-0.0083, 0.0716, -0.0201, -0.0061, 0.0747, -0.0219, -0.0083, 0.0747, -0.0219, -0.01, 0.0716, -0.0219, -0.01, 0.0716, -0.0201, -0.01, 0.0747, -0.0219, -0.01, 0.0747, -0.0201, -0.0061, 0.0768, -0.0219, -0.0083, 0.0768, -0.0219, -0.0061, 0.079, -0.0219, -0.0083, 0.079, -0.0219, -0.01, 0.0768, -0.0219, -0.01, 0.0768, -0.0201, -0.01, 0.079, -0.0219, -0.01, 0.079, -0.0201, -0.0061, 0.0812, -0.0219, -0.0083, 0.0812, -0.0219, -0.0061, 0.0833, -0.0219, -0.0061, 0.0833, -0.0201, -0.0083, 0.0833, -0.0219, -0.0083, 0.0833, -0.0201, -0.01, 0.0812, -0.0219, -0.01, 0.0812, -0.0201, -0.01, 0.0833, -0.0219, -0.01, 0.0833, -0.0201, -0.0061, 0.0716, -0.0179, -0.0061, 0.0716, -0.0157, -0.0083, 0.0716, -0.0179, -0.0083, 0.0716, -0.0157, -0.0061, 0.0716, -0.0136,
-0.0061, 0.0716, -0.0114, -0.0083, 0.0716, -0.0136, -0.0083, 0.0716, -0.0114, -0.01, 0.0716, -0.0179, -0.01, 0.0716, -0.0157, -0.01, 0.0747, -0.0179, -0.01, 0.0747, -0.0157, -0.01, 0.0716, -0.0136, -0.01, 0.0716, -0.0114, -0.01, 0.0747, -0.0136, -0.01, 0.0747, -0.0114, -0.01, 0.0768, -0.0179, -0.01, 0.0768, -0.0157, -0.01, 0.079, -0.0179, -0.01, 0.079, -0.0157, -0.01, 0.0768, -0.0136, -0.01, 0.0768, -0.0114, -0.01, 0.079, -0.0136, -0.01, 0.079, -0.0114, -0.0061, 0.0716, -0.0092, -0.0061, 0.0716, -0.0071, -0.0083, 0.0716, -0.0092, -0.0083, 0.0716, -0.0071, -0.0061, 0.0716, -0.0049, -0.0061, 0.0716, -0.0027, -0.0083, 0.0716, -0.0049, -0.0083, 0.0716, -0.0027, -0.01, 0.0716, -0.0092, -0.01, 0.0716, -0.0071, -0.01, 0.0747, -0.0092,
-0.01, 0.0747, -0.0071, -0.01, 0.0716, -0.0049, -0.01, 0.0716, -0.0027, -0.01, 0.0747, -0.0049, -0.01, 0.0747, -0.0027, -0.01, 0.0768, -0.0092, -0.01, 0.0768, -0.0071, -0.01, 0.079, -0.0092, -0.01, 0.079, -0.0071, -0.01, 0.0768, -0.0049, -0.01, 0.0768, -0.0027, -0.01, 0.079, -0.0049, -0.01, 0.079, -0.0027, -0.0061, 0.0833, -0.0179, -0.0061, 0.0833, -0.0157, -0.0083, 0.0833, -0.0179, -0.0083, 0.0833, -0.0157, -0.0061, 0.0833, -0.0136, -0.0061, 0.0833, -0.0114, -0.0083, 0.0833, -0.0136, -0.0083, 0.0833, -0.0114, -0.01, 0.0812, -0.0179, -0.01, 0.0812, -0.0157, -0.01, 0.0833, -0.0179, -0.01, 0.0833, -0.0157, -0.01, 0.0812, -0.0136, -0.01, 0.0812, -0.0114, -0.01, 0.0833, -0.0136, -0.01, 0.0833, -0.0114, -0.0061, 0.0833, -0.0092,
-0.0061, 0.0833, -0.0071, -0.0083, 0.0833, -0.0092, -0.0083, 0.0833, -0.0071, -0.0061, 0.0833, -0.0049, -0.0061, 0.0833, -0.0027, -0.0083, 0.0833, -0.0049, -0.0083, 0.0833, -0.0027, -0.01, 0.0812, -0.0092, -0.01, 0.0812, -0.0071, -0.01, 0.0833, -0.0092, -0.01, 0.0833, -0.0071, -0.01, 0.0812, -0.0049, -0.01, 0.0812, -0.0027, -0.01, 0.0833, -0.0049, -0.01, 0.0833, -0.0027, -0.0061, 0.1066, -0.0668, -0.0083, 0.1066, -0.0668, -0.0061, 0.1077, -0.0666, -0.0061, 0.1071, -0.0627, -0.0083, 0.1077, -0.0666, -0.0083, 0.1071, -0.0627, -0.0061, 0.1094, -0.0662, -0.0083, 0.1094, -0.0662, -0.01, 0.1065, -0.0668, -0.01, 0.1074, -0.0667, -0.01, 0.1072, -0.0626, -0.01, 0.1097, -0.0661, -0.01, 0.1094, -0.0635, -0.0061, 0.1115, -0.0652, -0.0061, 0.1125, -0.0645,
-0.0083, 0.1115, -0.0652, -0.0083, 0.1125, -0.0645, -0.0061, 0.1138, -0.0635, -0.0083, 0.1138, -0.0635, -0.01, 0.1113, -0.0653, -0.01, 0.1123, -0.0646, -0.01, 0.1138, -0.0636, -0.0061, 0.1082, -0.0624, -0.0083, 0.1082, -0.0624, -0.0061, 0.1094, -0.062, -0.0083, 0.1094, -0.062, -0.01, 0.1082, -0.0623, -0.01, 0.1093, -0.062, -0.0061, 0.1111, -0.0608, -0.0061, 0.1122, -0.0596, -0.0083, 0.1111, -0.0608, -0.0083, 0.1122, -0.0596, -0.0061, 0.1148, -0.0623, -0.0061, 0.1129, -0.0584, -0.0083, 0.1148, -0.0623, -0.0083, 0.1129, -0.0584, -0.0061, 0.1134, -0.057, -0.0061, 0.1136, -0.0548, -0.0083, 0.1134, -0.057, -0.0083, 0.1136, -0.0548, -0.01, 0.1112, -0.0608, -0.01, 0.1122, -0.0596, -0.01, 0.1148, -0.0622, -0.01, 0.1127, -0.0587, -0.01, 0.1134, -0.0569,
-0.01, 0.1136, -0.0548, -0.0061, 0.1155, -0.0613, -0.0061, 0.1167, -0.0593, -0.0083, 0.1155, -0.0613, -0.0083, 0.1167, -0.0593, -0.0061, 0.117, -0.0581, -0.0083, 0.117, -0.0581, -0.0061, 0.1173, -0.057, -0.0061, 0.1176, -0.0548, -0.0083, 0.1173, -0.057, -0.0083, 0.1176, -0.0548, -0.01, 0.1156, -0.0612, -0.01, 0.1167, -0.0593, -0.01, 0.117, -0.0581, -0.01, 0.1159, -0.057, -0.01, 0.1159, -0.0548, -0.01, 0.1173, -0.057, -0.01, 0.1176, -0.0548, -0.0061, 0.1072, -0.0463, -0.0083, 0.1072, -0.0463, -0.0061, 0.1099, -0.0476, -0.0061, 0.1088, -0.047, -0.0083, 0.1099, -0.0476, -0.0083, 0.1088, -0.047, -0.01, 0.107, -0.0462, -0.01, 0.11, -0.0476, -0.01, 0.109, -0.047, -0.0061, 0.1123, -0.0499, -0.0083, 0.1123, -0.0499, -0.0061, 0.1134, -0.0526,
-0.0061, 0.1129, -0.051, -0.0083, 0.1134, -0.0526, -0.0083, 0.1129, -0.051, -0.0061, 0.1112, -0.0486, -0.0083, 0.1112, -0.0486, -0.0061, 0.1142, -0.0456, -0.0083, 0.1142, -0.0456, -0.01, 0.1122, -0.0498, -0.01, 0.1134, -0.0528, -0.01, 0.1128, -0.0508, -0.01, 0.1112, -0.0486, -0.01, 0.1116, -0.0461, -0.01, 0.1137, -0.0483, -0.01, 0.1142, -0.0456, -0.0061, 0.1072, -0.0426, -0.0083, 0.1072, -0.0426, -0.0061, 0.1098, -0.043, -0.0061, 0.1088, -0.0427, -0.0083, 0.1098, -0.043, -0.0083, 0.1088, -0.0427, -0.01, 0.1072, -0.0439, -0.01, 0.1072, -0.0426, -0.01, 0.1096, -0.043, -0.01, 0.1086, -0.0427, -0.0061, 0.1116, -0.0436, -0.0083, 0.1116, -0.0436, -0.0061, 0.1131, -0.0447, -0.0083, 0.1131, -0.0447, -0.01, 0.1119, -0.0437, -0.01, 0.1131, -0.0447,
-0.0061, 0.1168, -0.0498, -0.0083, 0.1168, -0.0498, -0.0061, 0.1175, -0.0526, -0.0061, 0.1172, -0.0509, -0.0083, 0.1175, -0.0526, -0.0083, 0.1172, -0.0509, -0.0061, 0.1162, -0.0482, -0.0061, 0.1152, -0.0467, -0.0083, 0.1162, -0.0482, -0.0083, 0.1152, -0.0467, -0.01, 0.1159, -0.0526, -0.01, 0.1169, -0.0501, -0.01, 0.1175, -0.0526, -0.01, 0.1172, -0.0511, -0.01, 0.1161, -0.048, -0.01, 0.1152, -0.0468, -0.0061, 0.1363, -0.067, -0.0083, 0.1363, -0.067, -0.0061, 0.1363, -0.0656, -0.0061, 0.1363, -0.0635, -0.0083, 0.1363, -0.0656, -0.0083, 0.1363, -0.0635, -0.01, 0.1363, -0.067, -0.01, 0.1363, -0.0656, -0.01, 0.1363, -0.0635, -0.0061, 0.1376, -0.067, -0.0083, 0.1376, -0.067, -0.0061, 0.1402, -0.067, -0.0083, 0.1402, -0.067, -0.0061, 0.1402, -0.0656,
-0.0061, 0.1402, -0.0635, -0.0083, 0.1402, -0.0656, -0.0083, 0.1402, -0.0635, -0.01, 0.1376, -0.067, -0.01, 0.1402, -0.067, -0.01, 0.1376, -0.0656, -0.01, 0.1376, -0.0635, -0.01, 0.1402, -0.0656, -0.01, 0.1402, -0.0635, -0.0061, 0.1363, -0.0613, -0.0061, 0.1363, -0.0591, -0.0083, 0.1363, -0.0613, -0.0083, 0.1363, -0.0591, -0.0061, 0.1363, -0.057, -0.0061, 0.1363, -0.0548, -0.0083, 0.1363, -0.057, -0.0083, 0.1363, -0.0548, -0.01, 0.1363, -0.0613, -0.01, 0.1363, -0.0591, -0.01, 0.1363, -0.057, -0.01, 0.1363, -0.0548, -0.0061, 0.1402, -0.0613, -0.0061, 0.1402, -0.0591, -0.0083, 0.1402, -0.0613, -0.0083, 0.1402, -0.0591, -0.0061, 0.1402, -0.057, -0.0061, 0.1402, -0.0548, -0.0083, 0.1402, -0.057, -0.0083, 0.1402, -0.0548, -0.01, 0.1376, -0.0613,
-0.01, 0.1376, -0.0591, -0.01, 0.1402, -0.0613, -0.01, 0.1402, -0.0591, -0.01, 0.1376, -0.057, -0.01, 0.1376, -0.0548, -0.01, 0.1402, -0.057, -0.01, 0.1402, -0.0548, -0.0061, 0.1363, -0.0526, -0.0061, 0.1363, -0.0505, -0.0083, 0.1363, -0.0526, -0.0083, 0.1363, -0.0505, -0.0061, 0.1363, -0.0483, -0.0061, 0.1363, -0.0461, -0.0083, 0.1363, -0.0483, -0.0083, 0.1363, -0.0461, -0.01, 0.1363, -0.0526, -0.01, 0.1363, -0.0505, -0.01, 0.1363, -0.0483, -0.01, 0.1363, -0.0461, -0.0061, 0.1402, -0.0526, -0.0061, 0.1402, -0.0505, -0.0083, 0.1402, -0.0526, -0.0083, 0.1402, -0.0505, -0.0061, 0.1402, -0.0483, -0.0061, 0.1402, -0.0454, -0.0083, 0.1402, -0.0483, -0.0083, 0.1402, -0.0454, -0.01, 0.1376, -0.0526, -0.01, 0.1376, -0.0505, -0.01, 0.1402, -0.0526,
-0.01, 0.1402, -0.0505, -0.01, 0.1376, -0.0483, -0.01, 0.1376, -0.0461, -0.01, 0.1402, -0.0483, -0.01, 0.1402, -0.0454, -0.0061, 0.1363, -0.0439, -0.0061, 0.1363, -0.0425, -0.0083, 0.1363, -0.0439, -0.0083, 0.1363, -0.0425, -0.01, 0.1363, -0.0439, -0.01, 0.1363, -0.0425, -0.0061, 0.1376, -0.0425, -0.0083, 0.1376, -0.0425, -0.0061, 0.1398, -0.0425, -0.0083, 0.1398, -0.0425, -0.01, 0.1376, -0.0439, -0.01, 0.1376, -0.0425, -0.01, 0.1398, -0.0439, -0.01, 0.1398, -0.0425, -0.0061, 0.1431, -0.0591, -0.0083, 0.1431, -0.0591, -0.0061, 0.1431, -0.0562, -0.0083, 0.1431, -0.0562, -0.01, 0.1431, -0.0591, -0.01, 0.1431, -0.0562, -0.0061, 0.1463, -0.0591, -0.0083, 0.1463, -0.0591, -0.0061, 0.1485, -0.0592, -0.0083, 0.1485, -0.0592, -0.0061, 0.1463, -0.0562,
-0.0083, 0.1463, -0.0562, -0.0061, 0.1484, -0.0562, -0.0083, 0.1484, -0.0562, -0.01, 0.1463, -0.0591, -0.01, 0.1484, -0.0592, -0.01, 0.1463, -0.0562, -0.01, 0.1484, -0.0562, -0.0061, 0.1553, -0.0632, -0.0083, 0.1553, -0.0632, -0.0061, 0.1575, -0.0651, -0.0061, 0.1564, -0.0643, -0.0083, 0.1575, -0.0651, -0.0083, 0.1564, -0.0643, -0.01, 0.1553, -0.0632, -0.01, 0.1576, -0.0652, -0.01, 0.1564, -0.0643, -0.0061, 0.1506, -0.0596, -0.0083, 0.1506, -0.0596, -0.0061, 0.1529, -0.0609, -0.0061, 0.1518, -0.0602, -0.0083, 0.1529, -0.0609, -0.0083, 0.1518, -0.0602, -0.0061, 0.1506, -0.0562, -0.0083, 0.1506, -0.0562, -0.0061, 0.1528, -0.0562, -0.0083, 0.1528, -0.0562, -0.01, 0.1504, -0.0596, -0.01, 0.153, -0.061, -0.01, 0.1521, -0.0603, -0.01, 0.1506, -0.0562,
-0.01, 0.1527, -0.0562, -0.0061, 0.1542, -0.0621, -0.0083, 0.1542, -0.0621, -0.0061, 0.1546, -0.0561, -0.0061, 0.1556, -0.0558, -0.0083, 0.1546, -0.0561, -0.0083, 0.1556, -0.0558, -0.0061, 0.1572, -0.0553, -0.0083, 0.1572, -0.0553, -0.01, 0.1542, -0.0621, -0.01, 0.155, -0.0591, -0.01, 0.1571, -0.0613, -0.01, 0.1571, -0.0591, -0.01, 0.1549, -0.056, -0.01, 0.1557, -0.0557, -0.01, 0.1571, -0.057, -0.01, 0.1571, -0.0553, -0.0061, 0.1419, -0.0454, -0.0083, 0.1419, -0.0454, -0.0061, 0.1441, -0.0454, -0.0083, 0.1441, -0.0454, -0.01, 0.1419, -0.0454, -0.01, 0.1441, -0.0454, -0.0061, 0.1463, -0.0454, -0.0083, 0.1463, -0.0454, -0.0061, 0.1484, -0.0454, -0.0083, 0.1484, -0.0454, -0.01, 0.1463, -0.0454, -0.01, 0.1484, -0.0454, -0.0061, 0.1419, -0.0425,
-0.0083, 0.1419, -0.0425, -0.0061, 0.1441, -0.0425, -0.0083, 0.1441, -0.0425, -0.01, 0.1419, -0.0439, -0.01, 0.1419, -0.0425, -0.01, 0.1441, -0.0439, -0.01, 0.1441, -0.0425, -0.0061, 0.1463, -0.0425, -0.0083, 0.1463, -0.0425, -0.0061, 0.1484, -0.0425, -0.0083, 0.1484, -0.0425, -0.01, 0.1463, -0.0439, -0.01, 0.1463, -0.0425, -0.01, 0.1484, -0.0439, -0.01, 0.1484, -0.0425, -0.0061, 0.1506, -0.0454, -0.0083, 0.1506, -0.0454, -0.0061, 0.1528, -0.0456, -0.0083, 0.1528, -0.0456, -0.01, 0.1506, -0.0454, -0.01, 0.1528, -0.0456, -0.0061, 0.155, -0.0459, -0.0083, 0.155, -0.0459, -0.0061, 0.1579, -0.0474, -0.0061, 0.1568, -0.0467, -0.0083, 0.1579, -0.0474, -0.0083, 0.1568, -0.0467, -0.01, 0.1549, -0.0459, -0.01, 0.1579, -0.0475, -0.01, 0.1569, -0.0468,
-0.0061, 0.1506, -0.0425, -0.0083, 0.1506, -0.0425, -0.0061, 0.1528, -0.0425, -0.0083, 0.1528, -0.0425, -0.01, 0.1506, -0.0439, -0.01, 0.1506, -0.0425, -0.01, 0.1528, -0.0439, -0.01, 0.1528, -0.0425, -0.0061, 0.155, -0.0425, -0.0083, 0.155, -0.0425, -0.0061, 0.1571, -0.0426, -0.0083, 0.1571, -0.0426, -0.01, 0.155, -0.0439, -0.01, 0.155, -0.0425, -0.01, 0.1571, -0.0439, -0.01, 0.1571, -0.0426, -0.0061, 0.1619, -0.0669, -0.0083, 0.1619, -0.0669, -0.0061, 0.1592, -0.0662, -0.0083, 0.1592, -0.0662, -0.0061, 0.1608, -0.0666, -0.0061, 0.1619, -0.0632, -0.0083, 0.1608, -0.0666, -0.0083, 0.1619, -0.0632, -0.01, 0.1621, -0.0669, -0.01, 0.159, -0.0661, -0.01, 0.1593, -0.0635, -0.01, 0.1612, -0.0667, -0.01, 0.1619, -0.0632, -0.0061, 0.1648, -0.0671,
-0.0083, 0.1648, -0.0671, -0.0061, 0.1638, -0.0656, -0.0061, 0.1628, -0.0643, -0.0083, 0.1638, -0.0656, -0.0083, 0.1628, -0.0643, -0.01, 0.1648, -0.0671, -0.01, 0.1636, -0.0653, -0.01, 0.1628, -0.0643, -0.0061, 0.1599, -0.0608, -0.0061, 0.1588, -0.0591, -0.0083, 0.1599, -0.0608, -0.0083, 0.1588, -0.0591, -0.0061, 0.1608, -0.0619, -0.0061, 0.1607, -0.0583, -0.0083, 0.1608, -0.0619, -0.0083, 0.1607, -0.0583, -0.0061, 0.1587, -0.0543, -0.0083, 0.1587, -0.0543, -0.0061, 0.1618, -0.0574, -0.0083, 0.1618, -0.0574, -0.01, 0.16, -0.0609, -0.01, 0.1588, -0.0591, -0.01, 0.1608, -0.0619, -0.01, 0.1606, -0.0583, -0.01, 0.1593, -0.057, -0.01, 0.1588, -0.0543, -0.01, 0.1618, -0.0574, -0.01, 0.1615, -0.0548, -0.0061, 0.1629, -0.0563, -0.0061, 0.1638, -0.0549,
-0.0083, 0.1629, -0.0563, -0.0083, 0.1638, -0.0549, -0.01, 0.1629, -0.0563, -0.01, 0.1637, -0.0551, -0.0061, 0.1596, -0.053, -0.0061, 0.1603, -0.0503, -0.0083, 0.1596, -0.053, -0.0083, 0.1603, -0.0503, -0.0061, 0.1585, -0.0479, -0.0083, 0.1585, -0.0479, -0.01, 0.1598, -0.0525, -0.01, 0.1603, -0.0503, -0.01, 0.1615, -0.0526, -0.01, 0.1615, -0.0505, -0.01, 0.1591, -0.0484, -0.01, 0.1593, -0.0461, -0.01, 0.1615, -0.0483, -0.01, 0.1615, -0.0461, -0.0061, 0.1646, -0.0527, -0.0061, 0.1647, -0.0505, -0.0083, 0.1646, -0.0527, -0.0083, 0.1647, -0.0505, -0.0061, 0.1645, -0.048, -0.0061, 0.1634, -0.046, -0.0083, 0.1645, -0.048, -0.0083, 0.1634, -0.046, -0.01, 0.1645, -0.0529, -0.01, 0.1647, -0.0505, -0.01, 0.1644, -0.048, -0.01, 0.1633, -0.0459,
-0.0061, 0.1596, -0.0431, -0.0061, 0.1585, -0.0428, -0.0083, 0.1596, -0.0431, -0.0083, 0.1585, -0.0428, -0.0061, 0.162, -0.0444, -0.0083, 0.162, -0.0444, -0.01, 0.1595, -0.043, -0.01, 0.1584, -0.0428, -0.01, 0.1618, -0.0442, -0.0061, 0.1626, -0.045, -0.0083, 0.1626, -0.045, -0.01, 0.1626, -0.045, -0.0061, 0.1412, -0.0219, -0.0061, 0.1412, -0.0201, -0.0083, 0.1412, -0.0219, -0.0083, 0.1412, -0.0201, -0.0061, 0.1441, -0.0219, -0.0083, 0.1441, -0.0219, -0.01, 0.1412, -0.0219, -0.01, 0.1412, -0.0201, -0.01, 0.1441, -0.0219, -0.01, 0.1441, -0.0201, -0.0061, 0.1463, -0.0219, -0.0083, 0.1463, -0.0219, -0.0061, 0.1484, -0.0219, -0.0083, 0.1484, -0.0219, -0.01, 0.1463, -0.0219, -0.01, 0.1463, -0.0201, -0.01, 0.1484, -0.0219, -0.01, 0.1484, -0.0201,
-0.0061, 0.1506, -0.0219, -0.0083, 0.1506, -0.0219, -0.0061, 0.1528, -0.0219, -0.0083, 0.1528, -0.0219, -0.01, 0.1506, -0.0219, -0.01, 0.1506, -0.0201, -0.01, 0.1528, -0.0219, -0.01, 0.1528, -0.0201, -0.0061, 0.155, -0.0219, -0.0083, 0.155, -0.0219, -0.0061, 0.1571, -0.0219, -0.0083, 0.1571, -0.0219, -0.01, 0.155, -0.0219, -0.01, 0.155, -0.0201, -0.01, 0.1571, -0.0219, -0.01, 0.1571, -0.0201, -0.0061, 0.1412, -0.0179, -0.0061, 0.1412, -0.0157, -0.0083, 0.1412, -0.0179, -0.0083, 0.1412, -0.0157, -0.0061, 0.1412, -0.0136, -0.0061, 0.1412, -0.0114, -0.0083, 0.1412, -0.0136, -0.0083, 0.1412, -0.0114, -0.01, 0.1412, -0.0179, -0.01, 0.1412, -0.0157, -0.01, 0.1441, -0.0179, -0.01, 0.1441, -0.0157, -0.01, 0.1412, -0.0136, -0.01, 0.1412, -0.0114,
-0.01, 0.1441, -0.0136, -0.01, 0.1441, -0.0114, -0.01, 0.1463, -0.0179, -0.01, 0.1463, -0.0157, -0.01, 0.1484, -0.0179, -0.01, 0.1484, -0.0157, -0.01, 0.1463, -0.0136, -0.01, 0.1463, -0.0114, -0.01, 0.1484, -0.0136, -0.01, 0.1484, -0.0114, -0.0061, 0.1412, -0.0092, -0.0061, 0.1412, -0.0071, -0.0083, 0.1412, -0.0092, -0.0083, 0.1412, -0.0071, -0.0061, 0.1412, -0.0049, -0.0061, 0.1412, -0.0027, -0.0083, 0.1412, -0.0049, -0.0083, 0.1412, -0.0027, -0.01, 0.1412, -0.0092, -0.01, 0.1412, -0.0071, -0.01, 0.1441, -0.0092, -0.01, 0.1441, -0.0071, -0.01, 0.1412, -0.0049, -0.01, 0.1412, -0.0027, -0.01, 0.1441, -0.0049, -0.01, 0.1441, -0.0027, -0.01, 0.1463, -0.0092, -0.01, 0.1463, -0.0071, -0.01, 0.1484, -0.0092, -0.01, 0.1484, -0.0071,
-0.01, 0.1463, -0.0049, -0.01, 0.1463, -0.0027, -0.01, 0.1484, -0.0049, -0.01, 0.1484, -0.0027, -0.01, 0.1506, -0.0179, -0.01, 0.1506, -0.0157, -0.01, 0.1528, -0.0179, -0.01, 0.1528, -0.0157, -0.01, 0.1506, -0.0136, -0.01, 0.1506, -0.0114, -0.01, 0.1528, -0.0136, -0.01, 0.1528, -0.0114, -0.01, 0.155, -0.0179, -0.01, 0.155, -0.0157, -0.01, 0.1571, -0.0179, -0.01, 0.1571, -0.0157, -0.01, 0.155, -0.0136, -0.01, 0.155, -0.0114, -0.01, 0.1571, -0.0136, -0.01, 0.1571, -0.0114, -0.0061, 0.152, -0.0101, -0.0061, 0.152, -0.0071, -0.0083, 0.152, -0.0101, -0.0083, 0.152, -0.0071, -0.0061, 0.152, -0.0049, -0.0061, 0.152, -0.0027, -0.0083, 0.152, -0.0049, -0.0083, 0.152, -0.0027, -0.01, 0.1506, -0.0092, -0.01, 0.1506, -0.0071,
-0.01, 0.152, -0.0101, -0.01, 0.152, -0.0071, -0.01, 0.1506, -0.0049, -0.01, 0.1506, -0.0027, -0.01, 0.152, -0.0049, -0.01, 0.152, -0.0027, -0.0061, 0.155, -0.0101, -0.0083, 0.155, -0.0101, -0.0061, 0.1571, -0.0101, -0.0083, 0.1571, -0.0101, -0.01, 0.155, -0.0101, -0.01, 0.1571, -0.0101, -0.0061, 0.1593, -0.0219, -0.0083, 0.1593, -0.0219, -0.0061, 0.1615, -0.0219, -0.0083, 0.1615, -0.0219, -0.01, 0.1593, -0.0219, -0.01, 0.1593, -0.0201, -0.01, 0.1615, -0.0219, -0.01, 0.1615, -0.0201, -0.0061, 0.1636, -0.0219, -0.0083, 0.1636, -0.0219, -0.0061, 0.1658, -0.0219, -0.0083, 0.1658, -0.0219, -0.01, 0.1636, -0.0219, -0.01, 0.1636, -0.0201, -0.01, 0.1658, -0.0219, -0.01, 0.1658, -0.0201, -0.0061, 0.168, -0.0219, -0.0083, 0.168, -0.0219,
-0.0061, 0.1701, -0.0219, -0.0083, 0.1701, -0.0219, -0.01, 0.168, -0.0219, -0.01, 0.168, -0.0201, -0.01, 0.1701, -0.0219, -0.01, 0.1701, -0.0201, -0.0061, 0.1723, -0.0219, -0.0083, 0.1723, -0.0219, -0.0061, 0.1745, -0.0219, -0.0083, 0.1745, -0.0219, -0.01, 0.1723, -0.0219, -0.01, 0.1723, -0.0201, -0.01, 0.1745, -0.0219, -0.01, 0.1745, -0.0201, -0.01, 0.1593, -0.0179, -0.01, 0.1593, -0.0157, -0.01, 0.1615, -0.0179, -0.01, 0.1615, -0.0157, -0.01, 0.1593, -0.0136, -0.01, 0.1593, -0.0114, -0.01, 0.1615, -0.0136, -0.01, 0.1615, -0.0114, -0.01, 0.1636, -0.0179, -0.01, 0.1636, -0.0157, -0.01, 0.1658, -0.0179, -0.01, 0.1658, -0.0157, -0.01, 0.1636, -0.0136, -0.01, 0.1636, -0.0114, -0.01, 0.1658, -0.0136, -0.01, 0.1658, -0.0114,
-0.0061, 0.1593, -0.0101, -0.0083, 0.1593, -0.0101, -0.0061, 0.1615, -0.0101, -0.0083, 0.1615, -0.0101, -0.01, 0.1593, -0.0101, -0.01, 0.1615, -0.0101, -0.0061, 0.1636, -0.0101, -0.0083, 0.1636, -0.0101, -0.0061, 0.1658, -0.0101, -0.0083, 0.1658, -0.0101, -0.01, 0.1636, -0.0101, -0.01, 0.1658, -0.0101, -0.01, 0.168, -0.0179, -0.01, 0.168, -0.0157, -0.01, 0.1701, -0.0179, -0.01, 0.1701, -0.0157, -0.01, 0.168, -0.0136, -0.01, 0.168, -0.0114, -0.01, 0.1701, -0.0136, -0.01, 0.1701, -0.0114, -0.01, 0.1723, -0.0179, -0.01, 0.1723, -0.0157, -0.01, 0.1745, -0.0179, -0.01, 0.1745, -0.0157, -0.01, 0.1723, -0.0136, -0.01, 0.1723, -0.0114, -0.01, 0.1745, -0.0136, -0.01, 0.1745, -0.0114, -0.0061, 0.168, -0.0101, -0.0083, 0.168, -0.0101,
-0.0061, 0.1701, -0.0101, -0.0083, 0.1701, -0.0101, -0.01, 0.168, -0.0101, -0.01, 0.1701, -0.0101, -0.0061, 0.1723, -0.0101, -0.0083, 0.1723, -0.0101, -0.0061, 0.1745, -0.0101, -0.0083, 0.1745, -0.0101, -0.01, 0.1723, -0.0101, -0.01, 0.1745, -0.0101, -0.0061, 0.1767, -0.0219, -0.0083, 0.1767, -0.0219, -0.0061, 0.1788, -0.0219, -0.0083, 0.1788, -0.0219, -0.01, 0.1767, -0.0219, -0.01, 0.1767, -0.0201, -0.01, 0.1788, -0.0219, -0.01, 0.1788, -0.0201, -0.0061, 0.181, -0.0219, -0.0083, 0.181, -0.0219, -0.0061, 0.1832, -0.0219, -0.0083, 0.1832, -0.0219, -0.01, 0.181, -0.0219, -0.01, 0.181, -0.0201, -0.01, 0.1832, -0.0219, -0.01, 0.1832, -0.0201, -0.0061, 0.1853, -0.0219, -0.0083, 0.1853, -0.0219, -0.0061, 0.1875, -0.0219, -0.0083, 0.1875, -0.0219,
-0.01, 0.1853, -0.0219, -0.01, 0.1853, -0.0201, -0.01, 0.1875, -0.0219, -0.01, 0.1875, -0.0201, -0.0061, 0.1897, -0.0219, -0.0083, 0.1897, -0.0219, -0.0061, 0.1918, -0.0219, -0.0083, 0.1918, -0.0219, -0.01, 0.1897, -0.0219, -0.01, 0.1897, -0.0201, -0.01, 0.1918, -0.0219, -0.01, 0.1918, -0.0201, -0.01, 0.1767, -0.0179, -0.01, 0.1767, -0.0157, -0.01, 0.1788, -0.0179, -0.01, 0.1788, -0.0157, -0.01, 0.1767, -0.0136, -0.01, 0.1767, -0.0114, -0.01, 0.1788, -0.0136, -0.01, 0.1788, -0.0114, -0.01, 0.181, -0.0179, -0.01, 0.181, -0.0157, -0.01, 0.1832, -0.0179, -0.01, 0.1832, -0.0157, -0.01, 0.181, -0.0136, -0.01, 0.181, -0.0114, -0.01, 0.1832, -0.0136, -0.01, 0.1832, -0.0114, -0.0061, 0.1767, -0.0101, -0.0083, 0.1767, -0.0101,
-0.0061, 0.1788, -0.0101, -0.0083, 0.1788, -0.0101, -0.01, 0.1767, -0.0101, -0.01, 0.1788, -0.0101, -0.0061, 0.181, -0.0101, -0.0083, 0.181, -0.0101, -0.0061, 0.1832, -0.0101, -0.0083, 0.1832, -0.0101, -0.01, 0.181, -0.0101, -0.01, 0.1832, -0.0101, -0.01, 0.1853, -0.0179, -0.01, 0.1853, -0.0157, -0.01, 0.1875, -0.0179, -0.01, 0.1875, -0.0157, -0.01, 0.1853, -0.0136, -0.01, 0.1853, -0.0114, -0.01, 0.1875, -0.0136, -0.01, 0.1875, -0.0114, -0.01, 0.1897, -0.0179, -0.01, 0.1897, -0.0157, -0.01, 0.1918, -0.0179, -0.01, 0.1918, -0.0157, -0.01, 0.1897, -0.0136, -0.01, 0.1897, -0.0114, -0.01, 0.1918, -0.0136, -0.01, 0.1918, -0.0114, -0.0061, 0.1853, -0.0101, -0.0083, 0.1853, -0.0101, -0.0061, 0.1875, -0.0101, -0.0083, 0.1875, -0.0101,
-0.01, 0.1853, -0.0101, -0.01, 0.1875, -0.0101, -0.0061, 0.1897, -0.0101, -0.0083, 0.1897, -0.0101, -0.0061, 0.1918, -0.0101, -0.0083, 0.1918, -0.0101, -0.01, 0.1897, -0.0101, -0.01, 0.1918, -0.0101, -0.0061, 0.194, -0.0219, -0.0083, 0.194, -0.0219, -0.0061, 0.1962, -0.0219, -0.0083, 0.1962, -0.0219, -0.01, 0.194, -0.0219, -0.01, 0.194, -0.0201, -0.01, 0.1962, -0.0219, -0.01, 0.1962, -0.0201, -0.0061, 0.1984, -0.0219, -0.0083, 0.1984, -0.0219, -0.0061, 0.2005, -0.0219, -0.0083, 0.2005, -0.0219, -0.01, 0.1984, -0.0219, -0.01, 0.1984, -0.0201, -0.01, 0.2005, -0.0219, -0.01, 0.2005, -0.0201, -0.0061, 0.202, -0.0219, -0.0061, 0.202, -0.0201, -0.0083, 0.202, -0.0219, -0.0083, 0.202, -0.0201, -0.01, 0.202, -0.0219, -0.01, 0.202, -0.0201,
-0.01, 0.194, -0.0179, -0.01, 0.194, -0.0157, -0.01, 0.1962, -0.0179, -0.01, 0.1962, -0.0157, -0.01, 0.194, -0.0136, -0.01, 0.194, -0.0114, -0.01, 0.1962, -0.0136, -0.01, 0.1962, -0.0114, -0.01, 0.1984, -0.0179, -0.01, 0.1984, -0.0157, -0.01, 0.2005, -0.0179, -0.01, 0.2005, -0.0157, -0.01, 0.1984, -0.0136, -0.01, 0.1984, -0.0114, -0.01, 0.2005, -0.0136, -0.01, 0.2005, -0.0114, -0.0061, 0.194, -0.0101, -0.0083, 0.194, -0.0101, -0.0061, 0.1962, -0.0101, -0.0083, 0.1962, -0.0101, -0.01, 0.194, -0.0101, -0.01, 0.1962, -0.0101, -0.0061, 0.1984, -0.0101, -0.0083, 0.1984, -0.0101, -0.0061, 0.2005, -0.0101, -0.0083, 0.2005, -0.0101, -0.01, 0.1984, -0.0101, -0.01, 0.2005, -0.0101, -0.0061, 0.202, -0.0179, -0.0061, 0.202, -0.0157,
-0.0083, 0.202, -0.0179, -0.0083, 0.202, -0.0157, -0.0061, 0.202, -0.0136, -0.0061, 0.202, -0.0114, -0.0083, 0.202, -0.0136, -0.0083, 0.202, -0.0114, -0.01, 0.202, -0.0179, -0.01, 0.202, -0.0157, -0.01, 0.202, -0.0136, -0.01, 0.202, -0.0114, -0.0061, 0.202, -0.0101, -0.0083, 0.202, -0.0101, -0.01, 0.202, -0.0101, -0.0061, 0.2118, -0.067, -0.0083, 0.2118, -0.067, -0.0061, 0.2135, -0.067, -0.0083, 0.2135, -0.067, -0.0061, 0.2118, -0.0656, -0.0061, 0.2118, -0.0635, -0.0083, 0.2118, -0.0656, -0.0083, 0.2118, -0.0635, -0.01, 0.2118, -0.067, -0.01, 0.2135, -0.067, -0.01, 0.2118, -0.0656, -0.01, 0.2118, -0.0635, -0.01, 0.2135, -0.0656, -0.01, 0.2135, -0.0635, -0.0061, 0.2157, -0.067, -0.0083, 0.2157, -0.067, -0.0061, 0.2179, -0.067,
-0.0083, 0.2179, -0.067, -0.0061, 0.2157, -0.0631, -0.0083, 0.2157, -0.0631, -0.0061, 0.2179, -0.0631, -0.0083, 0.2179, -0.0631, -0.01, 0.2157, -0.067, -0.01, 0.2179, -0.067, -0.01, 0.2157, -0.0656, -0.01, 0.2157, -0.0631, -0.01, 0.2179, -0.0656, -0.01, 0.2179, -0.0631, -0.0061, 0.2118, -0.0613, -0.0061, 0.2118, -0.0591, -0.0083, 0.2118, -0.0613, -0.0083, 0.2118, -0.0591, -0.0061, 0.2118, -0.057, -0.0061, 0.2118, -0.0548, -0.0083, 0.2118, -0.057, -0.0083, 0.2118, -0.0548, -0.01, 0.2118, -0.0613, -0.01, 0.2118, -0.0591, -0.01, 0.2135, -0.0613, -0.01, 0.2135, -0.0591, -0.01, 0.2118, -0.057, -0.01, 0.2118, -0.0548, -0.01, 0.2135, -0.057, -0.01, 0.2135, -0.0548, -0.0061, 0.2157, -0.0613, -0.0061, 0.2157, -0.0591, -0.0083, 0.2157, -0.0613,
-0.0083, 0.2157, -0.0591, -0.0061, 0.2157, -0.057, -0.0061, 0.2157, -0.0548, -0.0083, 0.2157, -0.057, -0.0083, 0.2157, -0.0548, -0.01, 0.2157, -0.0613, -0.01, 0.2157, -0.0591, -0.01, 0.2157, -0.057, -0.01, 0.2157, -0.0548, -0.0061, 0.2201, -0.067, -0.0083, 0.2201, -0.067, -0.0061, 0.2222, -0.067, -0.0083, 0.2222, -0.067, -0.0061, 0.2201, -0.0631, -0.0083, 0.2201, -0.0631, -0.0061, 0.2222, -0.0631, -0.0083, 0.2222, -0.0631, -0.01, 0.2201, -0.067, -0.01, 0.2222, -0.067, -0.01, 0.2201, -0.0656, -0.01, 0.2201, -0.0631, -0.01, 0.2222, -0.0656, -0.01, 0.2222, -0.0631, -0.0061, 0.2244, -0.067, -0.0083, 0.2244, -0.067, -0.0061, 0.2266, -0.067, -0.0083, 0.2266, -0.067, -0.0061, 0.2244, -0.0631, -0.0083, 0.2244, -0.0631, -0.0061, 0.2266, -0.0631,
-0.0083, 0.2266, -0.0631, -0.01, 0.2244, -0.067, -0.01, 0.2266, -0.067, -0.01, 0.2244, -0.0656, -0.01, 0.2244, -0.0631, -0.01, 0.2266, -0.0656, -0.01, 0.2266, -0.0631, -0.0061, 0.2196, -0.0562, -0.0061, 0.2196, -0.0548, -0.0083, 0.2196, -0.0562, -0.0083, 0.2196, -0.0548, -0.0061, 0.2222, -0.0562, -0.0083, 0.2222, -0.0562, -0.01, 0.2196, -0.0562, -0.01, 0.2196, -0.0548, -0.01, 0.2222, -0.0562, -0.01, 0.2222, -0.0548, -0.0061, 0.2244, -0.0562, -0.0083, 0.2244, -0.0562, -0.0061, 0.2266, -0.0562, -0.0083, 0.2266, -0.0562, -0.01, 0.2244, -0.0562, -0.01, 0.2244, -0.0548, -0.01, 0.2266, -0.0562, -0.01, 0.2266, -0.0548, -0.0061, 0.2118, -0.0526, -0.0061, 0.2118, -0.0505, -0.0083, 0.2118, -0.0526, -0.0083, 0.2118, -0.0505, -0.0061, 0.2118, -0.0483,
-0.0061, 0.2118, -0.0461, -0.0083, 0.2118, -0.0483, -0.0083, 0.2118, -0.0461, -0.01, 0.2118, -0.0526, -0.01, 0.2118, -0.0505, -0.01, 0.2135, -0.0526, -0.01, 0.2135, -0.0505, -0.01, 0.2118, -0.0483, -0.01, 0.2118, -0.0461, -0.01, 0.2135, -0.0483, -0.01, 0.2135, -0.0461, -0.0061, 0.2157, -0.0526, -0.0061, 0.2157, -0.0505, -0.0083, 0.2157, -0.0526, -0.0083, 0.2157, -0.0505, -0.0061, 0.2157, -0.0483, -0.0061, 0.2157, -0.0454, -0.0083, 0.2157, -0.0483, -0.0083, 0.2157, -0.0454, -0.0061, 0.2179, -0.0454, -0.0083, 0.2179, -0.0454, -0.01, 0.2157, -0.0526, -0.01, 0.2157, -0.0505, -0.01, 0.2157, -0.0483, -0.01, 0.2157, -0.0454, -0.01, 0.2179, -0.0454, -0.0061, 0.2118, -0.0439, -0.0061, 0.2118, -0.0425, -0.0083, 0.2118, -0.0439, -0.0083, 0.2118, -0.0425,
-0.0061, 0.2135, -0.0425, -0.0083, 0.2135, -0.0425, -0.01, 0.2118, -0.0439, -0.01, 0.2118, -0.0425, -0.01, 0.2135, -0.0439, -0.01, 0.2135, -0.0425, -0.0061, 0.2157, -0.0425, -0.0083, 0.2157, -0.0425, -0.0061, 0.2179, -0.0425, -0.0083, 0.2179, -0.0425, -0.01, 0.2157, -0.0439, -0.01, 0.2157, -0.0425, -0.01, 0.2179, -0.0439, -0.01, 0.2179, -0.0425, -0.0061, 0.2196, -0.0523, -0.0083, 0.2196, -0.0523, -0.0061, 0.2222, -0.0523, -0.0083, 0.2222, -0.0523, -0.0061, 0.2201, -0.0454, -0.0083, 0.2201, -0.0454, -0.0061, 0.2222, -0.0454, -0.0083, 0.2222, -0.0454, -0.01, 0.2196, -0.0523, -0.01, 0.2222, -0.0523, -0.01, 0.2201, -0.0454, -0.01, 0.2222, -0.0454, -0.0061, 0.2244, -0.0523, -0.0083, 0.2244, -0.0523, -0.0061, 0.2266, -0.0523, -0.0083, 0.2266, -0.0523,
-0.0061, 0.2244, -0.0454, -0.0083, 0.2244, -0.0454, -0.0061, 0.2266, -0.0454, -0.0083, 0.2266, -0.0454, -0.01, 0.2244, -0.0523, -0.01, 0.2266, -0.0523, -0.01, 0.2244, -0.0454, -0.01, 0.2266, -0.0454, -0.0061, 0.2201, -0.0425, -0.0083, 0.2201, -0.0425, -0.0061, 0.2222, -0.0425, -0.0083, 0.2222, -0.0425, -0.01, 0.2201, -0.0439, -0.01, 0.2201, -0.0425, -0.01, 0.2222, -0.0439, -0.01, 0.2222, -0.0425, -0.0061, 0.2244, -0.0425, -0.0083, 0.2244, -0.0425, -0.0061, 0.2266, -0.0425, -0.0083, 0.2266, -0.0425, -0.01, 0.2244, -0.0439, -0.01, 0.2244, -0.0425, -0.01, 0.2266, -0.0439, -0.01, 0.2266, -0.0425, -0.0061, 0.2287, -0.067, -0.0083, 0.2287, -0.067, -0.0061, 0.2307, -0.0669, -0.0083, 0.2307, -0.0669, -0.0061, 0.2287, -0.063, -0.0083, 0.2287, -0.063,
-0.0061, 0.2318, -0.0667, -0.0061, 0.2309, -0.0629, -0.0083, 0.2318, -0.0667, -0.0083, 0.2309, -0.0629, -0.01, 0.2287, -0.067, -0.01, 0.2304, -0.0669, -0.01, 0.2287, -0.0656, -0.01, 0.2287, -0.0631, -0.01, 0.2312, -0.0668, -0.01, 0.2308, -0.0629, -0.0061, 0.2331, -0.0665, -0.0061, 0.2325, -0.0625, -0.0083, 0.2331, -0.0665, -0.0083, 0.2325, -0.0625, -0.0061, 0.2355, -0.0657, -0.0083, 0.2355, -0.0657, -0.01, 0.2333, -0.0665, -0.01, 0.2328, -0.0624, -0.01, 0.2356, -0.0657, -0.01, 0.2353, -0.0635, -0.0061, 0.2287, -0.0563, -0.0083, 0.2287, -0.0563, -0.0061, 0.2309, -0.0565, -0.0083, 0.2309, -0.0565, -0.01, 0.2287, -0.0563, -0.01, 0.2287, -0.0548, -0.01, 0.2309, -0.0565, -0.01, 0.2309, -0.0548, -0.0061, 0.2336, -0.0621, -0.0083, 0.2336, -0.0621,
-0.0061, 0.2349, -0.0613, -0.0061, 0.2363, -0.0591, -0.0083, 0.2349, -0.0613, -0.0083, 0.2363, -0.0591, -0.0061, 0.2331, -0.0571, -0.0083, 0.2331, -0.0571, -0.0061, 0.2345, -0.0579, -0.0083, 0.2345, -0.0579, -0.01, 0.2337, -0.0621, -0.01, 0.2349, -0.0613, -0.01, 0.2363, -0.0591, -0.01, 0.2329, -0.0571, -0.01, 0.2331, -0.0548, -0.01, 0.2346, -0.058, -0.01, 0.2353, -0.0548, -0.0061, 0.2366, -0.0648, -0.0061, 0.238, -0.0634, -0.0083, 0.2366, -0.0648, -0.0083, 0.238, -0.0634, -0.01, 0.2365, -0.0648, -0.01, 0.238, -0.0634, -0.0061, 0.2384, -0.0622, -0.0083, 0.2384, -0.0622, -0.0061, 0.2388, -0.0612, -0.0061, 0.239, -0.0591, -0.0083, 0.2388, -0.0612, -0.0083, 0.239, -0.0591, -0.0061, 0.2387, -0.057, -0.0061, 0.2384, -0.0544, -0.0083, 0.2387, -0.057,
-0.0083, 0.2384, -0.0544, -0.01, 0.2386, -0.0618, -0.01, 0.2374, -0.0591, -0.01, 0.2389, -0.0609, -0.01, 0.239, -0.0591, -0.01, 0.2374, -0.057, -0.01, 0.2384, -0.0547, -0.01, 0.2387, -0.057, -0.01, 0.2384, -0.0546, -0.0061, 0.2287, -0.0522, -0.0083, 0.2287, -0.0522, -0.0061, 0.2309, -0.0522, -0.0083, 0.2309, -0.0522, -0.0061, 0.2287, -0.0455, -0.0083, 0.2287, -0.0455, -0.0061, 0.2309, -0.0459, -0.0083, 0.2309, -0.0459, -0.01, 0.2287, -0.0522, -0.01, 0.2309, -0.0522, -0.01, 0.2287, -0.0455, -0.01, 0.2308, -0.0458, -0.0061, 0.2326, -0.0519, -0.0061, 0.2337, -0.0513, -0.0083, 0.2326, -0.0519, -0.0083, 0.2337, -0.0513, -0.0061, 0.2345, -0.0504, -0.0083, 0.2345, -0.0504, -0.0061, 0.2339, -0.0475, -0.0061, 0.2328, -0.0465, -0.0083, 0.2339, -0.0475,
-0.0083, 0.2328, -0.0465, -0.0061, 0.2348, -0.0484, -0.0083, 0.2348, -0.0484, -0.01, 0.2329, -0.0518, -0.01, 0.2338, -0.0512, -0.01, 0.2353, -0.0526, -0.01, 0.2345, -0.0504, -0.01, 0.2339, -0.0475, -0.01, 0.2329, -0.0466, -0.01, 0.2348, -0.0484, -0.01, 0.2353, -0.0461, -0.0061, 0.2287, -0.0425, -0.0083, 0.2287, -0.0425, -0.0061, 0.2309, -0.0425, -0.0083, 0.2309, -0.0425, -0.01, 0.2287, -0.0439, -0.01, 0.2287, -0.0425, -0.01, 0.2309, -0.0439, -0.01, 0.2309, -0.0425, -0.0061, 0.2337, -0.043, -0.0061, 0.2326, -0.0427, -0.0083, 0.2337, -0.043, -0.0083, 0.2326, -0.0427, -0.0061, 0.2353, -0.0437, -0.0083, 0.2353, -0.0437, -0.01, 0.2334, -0.0429, -0.01, 0.2324, -0.0427, -0.01, 0.2356, -0.0438, -0.0061, 0.2386, -0.0526, -0.0061, 0.239, -0.0505,
-0.0083, 0.2386, -0.0526, -0.0083, 0.239, -0.0505, -0.0061, 0.2385, -0.0472, -0.0061, 0.2379, -0.046, -0.0083, 0.2385, -0.0472, -0.0083, 0.2379, -0.046, -0.0061, 0.2388, -0.0478, -0.0083, 0.2388, -0.0478, -0.01, 0.2374, -0.0526, -0.01, 0.2374, -0.0505, -0.01, 0.2386, -0.0526, -0.01, 0.239, -0.0505, -0.01, 0.2387, -0.0475, -0.01, 0.2377, -0.0457, -0.01, 0.239, -0.0485, -0.0061, 0.2368, -0.0447, -0.0083, 0.2368, -0.0447, -0.01, 0.2367, -0.0447, -0.0061, 0.2187, -0.0107, -0.0083, 0.2187, -0.0107, -0.01, 0.2187, -0.0106, -0.0061, 0.2144, -0.0024, -0.0083, 0.2144, -0.0024, -0.01, 0.2143, -0.0022, -0.0061, 0.2163, -0.0069, -0.0083, 0.2163, -0.0069, -0.0061, 0.2176, -0.0093, -0.0061, 0.2169, -0.008, -0.0083, 0.2176, -0.0093, -0.0083, 0.2169, -0.008,
-0.0061, 0.2153, -0.0049, -0.0061, 0.2147, -0.0035, -0.0083, 0.2153, -0.0049, -0.0083, 0.2147, -0.0035, -0.01, 0.2162, -0.0067, -0.01, 0.2178, -0.0096, -0.01, 0.2168, -0.0077, -0.01, 0.2154, -0.0052, -0.01, 0.2146, -0.0032, -0.01, 0.2179, -0.0049, -0.01, 0.2179, -0.0027, -0.0061, 0.223, -0.0149, -0.0083, 0.223, -0.0149, -0.0061, 0.2207, -0.0129, -0.0061, 0.2196, -0.0118, -0.0083, 0.2207, -0.0129, -0.0083, 0.2196, -0.0118, -0.0061, 0.2219, -0.014, -0.0083, 0.2219, -0.014, -0.01, 0.2231, -0.0149, -0.01, 0.2207, -0.0129, -0.01, 0.2196, -0.0118, -0.01, 0.2219, -0.014, -0.01, 0.2222, -0.0114, -0.0061, 0.2244, -0.0158, -0.0083, 0.2244, -0.0158, -0.0061, 0.2269, -0.0172, -0.0061, 0.2258, -0.0167, -0.0083, 0.2269, -0.0172, -0.0083, 0.2258, -0.0167,
-0.01, 0.2241, -0.0156, -0.01, 0.227, -0.0173, -0.01, 0.226, -0.0168, -0.01, 0.2244, -0.0136, -0.01, 0.2244, -0.0114, -0.01, 0.2266, -0.0136, -0.01, 0.2266, -0.0114, -0.01, 0.2201, -0.0092, -0.01, 0.2201, -0.0071, -0.01, 0.2222, -0.0092, -0.01, 0.2222, -0.0071, -0.01, 0.2201, -0.0049, -0.01, 0.2201, -0.0027, -0.01, 0.2222, -0.0049, -0.01, 0.2222, -0.0027, -0.01, 0.2244, -0.0092, -0.01, 0.2244, -0.0071, -0.01, 0.2266, -0.0092, -0.01, 0.2266, -0.0071, -0.01, 0.2244, -0.0049, -0.01, 0.2244, -0.0027, -0.01, 0.2266, -0.0049, -0.01, 0.2266, -0.0027, -0.0061, 0.2318, -0.0191, -0.0083, 0.2318, -0.0191, -0.01, 0.2318, -0.0191, -0.0061, 0.2331, -0.0194, -0.0083, 0.2331, -0.0194, -0.0061, 0.2352, -0.02, -0.0083, 0.2352, -0.02,
-0.01, 0.233, -0.0194, -0.01, 0.2351, -0.02, -0.0061, 0.2374, -0.0205, -0.0083, 0.2374, -0.0205, -0.0061, 0.2396, -0.0208, -0.0083, 0.2396, -0.0208, -0.01, 0.2373, -0.0205, -0.01, 0.2395, -0.0208, -0.0061, 0.2418, -0.021, -0.0083, 0.2418, -0.021, -0.0061, 0.244, -0.0213, -0.0061, 0.2429, -0.0212, -0.0083, 0.244, -0.0213, -0.0083, 0.2429, -0.0212, -0.01, 0.2416, -0.021, -0.01, 0.2443, -0.0214, -0.01, 0.2436, -0.0213, -0.0061, 0.2287, -0.018, -0.0083, 0.2287, -0.018, -0.0061, 0.2307, -0.0187, -0.0083, 0.2307, -0.0187, -0.01, 0.2285, -0.0179, -0.01, 0.2287, -0.0157, -0.01, 0.2308, -0.0187, -0.01, 0.2309, -0.0157, -0.01, 0.2287, -0.0136, -0.01, 0.2287, -0.0114, -0.01, 0.2309, -0.0136, -0.01, 0.2309, -0.0114, -0.01, 0.2331, -0.0179,
-0.01, 0.2331, -0.0157, -0.01, 0.2353, -0.0179, -0.01, 0.2353, -0.0157, -0.01, 0.2331, -0.0136, -0.01, 0.2331, -0.0114, -0.01, 0.2353, -0.0136, -0.01, 0.2353, -0.0114, -0.0061, 0.2289, -0.0026, -0.0083, 0.2289, -0.0026, -0.0061, 0.231, -0.0047, -0.0061, 0.23, -0.0036, -0.0083, 0.231, -0.0047, -0.0083, 0.23, -0.0036, -0.01, 0.2287, -0.0092, -0.01, 0.2287, -0.0071, -0.01, 0.2309, -0.0092, -0.01, 0.2309, -0.0071, -0.01, 0.2287, -0.0049, -0.01, 0.2289, -0.0026, -0.01, 0.231, -0.0047, -0.01, 0.23, -0.0036, -0.0061, 0.2335, -0.0064, -0.0083, 0.2335, -0.0064, -0.0061, 0.2352, -0.0074, -0.0083, 0.2352, -0.0074, -0.0061, 0.2324, -0.0057, -0.0083, 0.2324, -0.0057, -0.01, 0.2331, -0.0092, -0.01, 0.2334, -0.0063, -0.01, 0.2353, -0.0092,
-0.01, 0.2354, -0.0075, -0.01, 0.2323, -0.0057, -0.01, 0.2374, -0.0179, -0.01, 0.2374, -0.0157, -0.01, 0.2396, -0.0179, -0.01, 0.2396, -0.0157, -0.01, 0.2374, -0.0136, -0.01, 0.2374, -0.0114, -0.01, 0.2396, -0.0136, -0.01, 0.2396, -0.0114, -0.01, 0.2418, -0.0179, -0.01, 0.2418, -0.0157, -0.01, 0.2439, -0.0179, -0.01, 0.2439, -0.0157, -0.01, 0.2418, -0.0136, -0.01, 0.2418, -0.0114, -0.01, 0.2439, -0.0136, -0.01, 0.2439, -0.0114, -0.0061, 0.2381, -0.0082, -0.0061, 0.2371, -0.0079, -0.0083, 0.2381, -0.0082, -0.0083, 0.2371, -0.0079, -0.0061, 0.2396, -0.0086, -0.0083, 0.2396, -0.0086, -0.01, 0.2377, -0.0081, -0.01, 0.2369, -0.0079, -0.01, 0.2398, -0.0086, -0.0061, 0.2418, -0.009, -0.0083, 0.2418, -0.009, -0.0061, 0.2439, -0.0093,
-0.0083, 0.2439, -0.0093, -0.01, 0.2419, -0.009, -0.01, 0.244, -0.0093, -0.0061, 0.2588, -0.067, -0.0083, 0.2588, -0.067, -0.0061, 0.2613, -0.067, -0.0083, 0.2613, -0.067, -0.0061, 0.2588, -0.0656, -0.0061, 0.2588, -0.0635, -0.0083, 0.2588, -0.0656, -0.0083, 0.2588, -0.0635, -0.0061, 0.2618, -0.0631, -0.0083, 0.2618, -0.0631, -0.01, 0.2588, -0.067, -0.01, 0.2613, -0.067, -0.01, 0.2588, -0.0656, -0.01, 0.2588, -0.0635, -0.01, 0.2613, -0.0656, -0.01, 0.2618, -0.0631, -0.0061, 0.2588, -0.0613, -0.0061, 0.2588, -0.0591, -0.0083, 0.2588, -0.0613, -0.0083, 0.2588, -0.0591, -0.0061, 0.2618, -0.0613, -0.0061, 0.2618, -0.0591, -0.0083, 0.2618, -0.0613, -0.0083, 0.2618, -0.0591, -0.0061, 0.2588, -0.057, -0.0061, 0.2588, -0.0548, -0.0083, 0.2588, -0.057,
-0.0083, 0.2588, -0.0548, -0.0061, 0.2618, -0.057, -0.0061, 0.2618, -0.0548, -0.0083, 0.2618, -0.057, -0.0083, 0.2618, -0.0548, -0.01, 0.2588, -0.0613, -0.01, 0.2588, -0.0591, -0.01, 0.2618, -0.0613, -0.01, 0.2618, -0.0591, -0.01, 0.2588, -0.057, -0.01, 0.2588, -0.0548, -0.01, 0.2618, -0.057, -0.01, 0.2618, -0.0548, -0.0061, 0.2588, -0.0526, -0.0061, 0.2588, -0.0505, -0.0083, 0.2588, -0.0526, -0.0083, 0.2588, -0.0505, -0.0061, 0.2618, -0.0526, -0.0061, 0.2618, -0.0505, -0.0083, 0.2618, -0.0526, -0.0083, 0.2618, -0.0505, -0.0061, 0.2588, -0.0483, -0.0061, 0.2588, -0.0461, -0.0083, 0.2588, -0.0483, -0.0083, 0.2588, -0.0461, -0.0061, 0.2618, -0.0483, -0.0061, 0.2618, -0.0461, -0.0083, 0.2618, -0.0483, -0.0083, 0.2618, -0.0461, -0.01, 0.2588, -0.0526,
-0.01, 0.2588, -0.0505, -0.01, 0.2618, -0.0526, -0.01, 0.2618, -0.0505, -0.01, 0.2588, -0.0483, -0.01, 0.2588, -0.0461, -0.01, 0.2618, -0.0483, -0.01, 0.2618, -0.0461, -0.0061, 0.2588, -0.0439, -0.0061, 0.2588, -0.0425, -0.0083, 0.2588, -0.0439, -0.0083, 0.2588, -0.0425, -0.0061, 0.2618, -0.0439, -0.0061, 0.2618, -0.0425, -0.0083, 0.2618, -0.0439, -0.0083, 0.2618, -0.0425, -0.01, 0.2588, -0.0439, -0.01, 0.2588, -0.0425, -0.01, 0.2618, -0.0439, -0.01, 0.2618, -0.0425, -0.0061, 0.2635, -0.067, -0.0083, 0.2635, -0.067, -0.0061, 0.2656, -0.067, -0.0083, 0.2656, -0.067, -0.0061, 0.2635, -0.0631, -0.0083, 0.2635, -0.0631, -0.0061, 0.2656, -0.0631, -0.0083, 0.2656, -0.0631, -0.01, 0.2635, -0.067, -0.01, 0.2656, -0.067, -0.01, 0.2635, -0.0656,
-0.01, 0.2635, -0.0631, -0.01, 0.2656, -0.0656, -0.01, 0.2656, -0.0631, -0.0061, 0.2678, -0.067, -0.0083, 0.2678, -0.067, -0.0061, 0.27, -0.067, -0.0083, 0.27, -0.067, -0.0061, 0.2678, -0.0631, -0.0083, 0.2678, -0.0631, -0.0061, 0.27, -0.0631, -0.0083, 0.27, -0.0631, -0.01, 0.2678, -0.067, -0.01, 0.27, -0.067, -0.01, 0.2678, -0.0656, -0.01, 0.2678, -0.0631, -0.01, 0.27, -0.0656, -0.01, 0.27, -0.0631, -0.0061, 0.2721, -0.067, -0.0083, 0.2721, -0.067, -0.0061, 0.2743, -0.067, -0.0083, 0.2743, -0.067, -0.0061, 0.2721, -0.0631, -0.0083, 0.2721, -0.0631, -0.0061, 0.2743, -0.0631, -0.0083, 0.2743, -0.0631, -0.01, 0.2721, -0.067, -0.01, 0.2743, -0.067, -0.01, 0.2721, -0.0656, -0.01, 0.2721, -0.0631, -0.01, 0.2743, -0.0656,
-0.01, 0.2743, -0.0631, -0.0061, 0.2765, -0.067, -0.0083, 0.2765, -0.067, -0.0061, 0.2794, -0.067, -0.0083, 0.2794, -0.067, -0.0061, 0.2765, -0.0631, -0.0083, 0.2765, -0.0631, -0.0061, 0.2794, -0.0656, -0.0061, 0.2794, -0.0631, -0.0083, 0.2794, -0.0656, -0.0083, 0.2794, -0.0631, -0.01, 0.2765, -0.067, -0.01, 0.2794, -0.067, -0.01, 0.2765, -0.0656, -0.01, 0.2765, -0.0631, -0.01, 0.2794, -0.0656, -0.01, 0.2794, -0.0631, -0.0061, 0.2461, -0.0216, -0.0083, 0.2461, -0.0216, -0.0061, 0.2483, -0.0218, -0.0083, 0.2483, -0.0218, -0.01, 0.2461, -0.0216, -0.01, 0.2461, -0.0201, -0.01, 0.2482, -0.0218, -0.01, 0.2483, -0.0201, -0.0061, 0.2504, -0.0219, -0.0083, 0.2504, -0.0219, -0.0061, 0.2526, -0.0219, -0.0083, 0.2526, -0.0219, -0.01, 0.2504, -0.0219,
-0.01, 0.2504, -0.0201, -0.01, 0.2526, -0.0219, -0.01, 0.2526, -0.0201, -0.0061, 0.2548, -0.0219, -0.0083, 0.2548, -0.0219, -0.0061, 0.257, -0.0217, -0.0083, 0.257, -0.0217, -0.01, 0.2548, -0.0219, -0.01, 0.2548, -0.0201, -0.01, 0.257, -0.0217, -0.01, 0.257, -0.0201, -0.0061, 0.2591, -0.0214, -0.0083, 0.2591, -0.0214, -0.0061, 0.2606, -0.0212, -0.0061, 0.2617, -0.0211, -0.0083, 0.2606, -0.0212, -0.0083, 0.2617, -0.0211, -0.01, 0.2591, -0.0214, -0.01, 0.2591, -0.0201, -0.01, 0.2605, -0.0212, -0.01, 0.2613, -0.0211, -0.01, 0.2461, -0.0179, -0.01, 0.2461, -0.0157, -0.01, 0.2483, -0.0179, -0.01, 0.2483, -0.0157, -0.01, 0.2461, -0.0136, -0.01, 0.2461, -0.0114, -0.01, 0.2483, -0.0136, -0.01, 0.2483, -0.0114, -0.01, 0.2504, -0.0179,
-0.01, 0.2504, -0.0157, -0.01, 0.2526, -0.0179, -0.01, 0.2526, -0.0157, -0.01, 0.2504, -0.0136, -0.01, 0.2504, -0.0114, -0.01, 0.2526, -0.0136, -0.01, 0.2526, -0.0114, -0.0061, 0.2461, -0.0096, -0.0083, 0.2461, -0.0096, -0.0061, 0.2483, -0.0099, -0.0083, 0.2483, -0.0099, -0.01, 0.2462, -0.0096, -0.01, 0.2483, -0.0099, -0.0061, 0.2504, -0.0101, -0.0083, 0.2504, -0.0101, -0.0061, 0.2526, -0.0101, -0.0083, 0.2526, -0.0101, -0.01, 0.2504, -0.0101, -0.01, 0.2526, -0.0101, -0.01, 0.2548, -0.0179, -0.01, 0.2548, -0.0157, -0.01, 0.257, -0.0179, -0.01, 0.257, -0.0157, -0.01, 0.2548, -0.0136, -0.01, 0.2548, -0.0114, -0.01, 0.257, -0.0136, -0.01, 0.257, -0.0114, -0.01, 0.2591, -0.0179, -0.01, 0.2591, -0.0157, -0.01, 0.2613, -0.0179,
-0.01, 0.2613, -0.0157, -0.01, 0.2591, -0.0136, -0.01, 0.2591, -0.0114, -0.01, 0.2613, -0.0136, -0.01, 0.2613, -0.0114, -0.0061, 0.2548, -0.0101, -0.0083, 0.2548, -0.0101, -0.0061, 0.257, -0.01, -0.0083, 0.257, -0.01, -0.01, 0.2548, -0.0101, -0.01, 0.2569, -0.01, -0.0061, 0.2591, -0.0097, -0.0083, 0.2591, -0.0097, -0.0061, 0.2613, -0.0093, -0.0083, 0.2613, -0.0093, -0.01, 0.2591, -0.0097, -0.01, 0.2612, -0.0093, -0.0061, 0.2635, -0.0209, -0.0083, 0.2635, -0.0209, -0.0061, 0.2656, -0.0206, -0.0083, 0.2656, -0.0206, -0.01, 0.2636, -0.0209, -0.01, 0.2658, -0.0206, -0.0061, 0.2678, -0.0202, -0.0083, 0.2678, -0.0202, -0.0061, 0.27, -0.0197, -0.0083, 0.27, -0.0197, -0.01, 0.2679, -0.0202, -0.01, 0.2701, -0.0197, -0.0061, 0.2717, -0.0192,
-0.0083, 0.2717, -0.0192, -0.01, 0.2715, -0.0192, -0.01, 0.2635, -0.0179, -0.01, 0.2635, -0.0157, -0.01, 0.2656, -0.0179, -0.01, 0.2656, -0.0157, -0.01, 0.2635, -0.0136, -0.01, 0.2635, -0.0114, -0.01, 0.2656, -0.0136, -0.01, 0.2656, -0.0114, -0.01, 0.2678, -0.0179, -0.01, 0.2678, -0.0157, -0.01, 0.27, -0.0179, -0.01, 0.27, -0.0157, -0.01, 0.2678, -0.0136, -0.01, 0.2678, -0.0114, -0.01, 0.27, -0.0136, -0.01, 0.27, -0.0114, -0.0061, 0.2635, -0.0089, -0.0083, 0.2635, -0.0089, -0.0061, 0.2655, -0.0085, -0.0061, 0.2666, -0.0081, -0.0083, 0.2655, -0.0085, -0.0083, 0.2666, -0.0081, -0.01, 0.2633, -0.009, -0.01, 0.2656, -0.0084, -0.01, 0.2666, -0.0081, -0.0061, 0.2678, -0.0076, -0.0083, 0.2678, -0.0076, -0.0061, 0.27, -0.0066,
-0.0083, 0.27, -0.0066, -0.0061, 0.271, -0.0059, -0.0083, 0.271, -0.0059, -0.01, 0.2678, -0.0092, -0.01, 0.2677, -0.0077, -0.01, 0.27, -0.0092, -0.01, 0.2699, -0.0066, -0.01, 0.271, -0.0059, -0.0061, 0.2728, -0.0189, -0.0083, 0.2728, -0.0189, -0.0061, 0.2743, -0.0183, -0.0083, 0.2743, -0.0183, -0.01, 0.2725, -0.019, -0.01, 0.2721, -0.0157, -0.01, 0.2746, -0.0182, -0.01, 0.2743, -0.0157, -0.01, 0.2721, -0.0136, -0.01, 0.2721, -0.0114, -0.01, 0.2743, -0.0136, -0.01, 0.2743, -0.0114, -0.0061, 0.2765, -0.0175, -0.0083, 0.2765, -0.0175, -0.0061, 0.2776, -0.0169, -0.0061, 0.2788, -0.0163, -0.0083, 0.2776, -0.0169, -0.0083, 0.2788, -0.0163, -0.01, 0.2766, -0.0174, -0.01, 0.2765, -0.0157, -0.01, 0.2776, -0.0169, -0.01, 0.2788, -0.0163,
-0.01, 0.2765, -0.0136, -0.01, 0.2765, -0.0114, -0.01, 0.2787, -0.0136, -0.01, 0.2787, -0.0114, -0.0061, 0.2722, -0.0051, -0.0083, 0.2722, -0.0051, -0.0061, 0.2735, -0.004, -0.0061, 0.2746, -0.0029, -0.0083, 0.2735, -0.004, -0.0083, 0.2746, -0.0029, -0.01, 0.2721, -0.0092, -0.01, 0.2721, -0.0071, -0.01, 0.2743, -0.0092, -0.01, 0.2743, -0.0071, -0.01, 0.272, -0.0053, -0.01, 0.2735, -0.004, -0.01, 0.2746, -0.0029, -0.0061, 0.2755, -0.0017, -0.0083, 0.2755, -0.0017, -0.01, 0.2765, -0.0092, -0.01, 0.2765, -0.0071, -0.01, 0.2787, -0.0092, -0.01, 0.2787, -0.0071, -0.01, 0.2765, -0.0049, -0.01, 0.2754, -0.0019, -0.01, 0.2787, -0.0049, -0.01, 0.2787, -0.0027, -0.0061, 0.298, -0.067, -0.0083, 0.298, -0.067, -0.0061, 0.3004, -0.067,
-0.0083, 0.3004, -0.067, -0.0061, 0.298, -0.0656, -0.0061, 0.298, -0.0635, -0.0083, 0.298, -0.0656, -0.0083, 0.298, -0.0635, -0.01, 0.298, -0.067, -0.01, 0.3004, -0.067, -0.01, 0.298, -0.0656, -0.01, 0.298, -0.0635, -0.01, 0.3004, -0.0656, -0.01, 0.3004, -0.0635, -0.0061, 0.3025, -0.067, -0.0083, 0.3025, -0.067, -0.0061, 0.3047, -0.067, -0.0083, 0.3047, -0.067, -0.0061, 0.302, -0.0631, -0.0083, 0.302, -0.0631, -0.0061, 0.3047, -0.0631, -0.0083, 0.3047, -0.0631, -0.01, 0.3025, -0.067, -0.01, 0.3047, -0.067, -0.01, 0.3025, -0.0656, -0.01, 0.302, -0.0631, -0.01, 0.3047, -0.0656, -0.01, 0.3047, -0.0631, -0.0061, 0.298, -0.0613, -0.0061, 0.298, -0.0591, -0.0083, 0.298, -0.0613, -0.0083, 0.298, -0.0591, -0.0061, 0.298, -0.057,
-0.0061, 0.298, -0.0548, -0.0083, 0.298, -0.057, -0.0083, 0.298, -0.0548, -0.01, 0.298, -0.0613, -0.01, 0.298, -0.0591, -0.01, 0.3004, -0.0613, -0.01, 0.3004, -0.0591, -0.01, 0.298, -0.057, -0.01, 0.298, -0.0548, -0.01, 0.3004, -0.057, -0.01, 0.3004, -0.0548, -0.0061, 0.302, -0.0613, -0.0061, 0.302, -0.0591, -0.0083, 0.302, -0.0613, -0.0083, 0.302, -0.0591, -0.0061, 0.302, -0.057, -0.0061, 0.302, -0.0548, -0.0083, 0.302, -0.057, -0.0083, 0.302, -0.0548, -0.0061, 0.3049, -0.0562, -0.0061, 0.3049, -0.0548, -0.0083, 0.3049, -0.0562, -0.0083, 0.3049, -0.0548, -0.01, 0.302, -0.0613, -0.01, 0.302, -0.0591, -0.01, 0.302, -0.057, -0.01, 0.302, -0.0548, -0.01, 0.3049, -0.0562, -0.01, 0.3049, -0.0548, -0.0061, 0.3069, -0.067,
-0.0083, 0.3069, -0.067, -0.0061, 0.309, -0.067, -0.0083, 0.309, -0.067, -0.0061, 0.3069, -0.0631, -0.0083, 0.3069, -0.0631, -0.0061, 0.309, -0.0631, -0.0083, 0.309, -0.0631, -0.01, 0.3069, -0.067, -0.01, 0.309, -0.067, -0.01, 0.3069, -0.0656, -0.01, 0.3069, -0.0631, -0.01, 0.309, -0.0656, -0.01, 0.309, -0.0631, -0.0061, 0.3112, -0.067, -0.0083, 0.3112, -0.067, -0.0061, 0.3134, -0.067, -0.0083, 0.3134, -0.067, -0.0061, 0.3112, -0.0631, -0.0083, 0.3112, -0.0631, -0.0061, 0.3134, -0.0631, -0.0083, 0.3134, -0.0631, -0.01, 0.3112, -0.067, -0.01, 0.3134, -0.067, -0.01, 0.3112, -0.0656, -0.01, 0.3112, -0.0631, -0.01, 0.3134, -0.0656, -0.01, 0.3134, -0.0631, -0.0061, 0.3069, -0.0562, -0.0083, 0.3069, -0.0562, -0.0061, 0.309, -0.0562,
-0.0083, 0.309, -0.0562, -0.01, 0.3069, -0.0562, -0.01, 0.3069, -0.0548, -0.01, 0.309, -0.0562, -0.01, 0.309, -0.0548, -0.0061, 0.3112, -0.0562, -0.0083, 0.3112, -0.0562, -0.0061, 0.3134, -0.0562, -0.0083, 0.3134, -0.0562, -0.01, 0.3112, -0.0562, -0.01, 0.3112, -0.0548, -0.01, 0.3134, -0.0562, -0.01, 0.3134, -0.0548, -0.0061, 0.298, -0.0526, -0.0061, 0.298, -0.0505, -0.0083, 0.298, -0.0526, -0.0083, 0.298, -0.0505, -0.0061, 0.298, -0.0483, -0.0061, 0.298, -0.0461, -0.0083, 0.298, -0.0483, -0.0083, 0.298, -0.0461, -0.01, 0.298, -0.0526, -0.01, 0.298, -0.0505, -0.01, 0.3004, -0.0526, -0.01, 0.3004, -0.0505, -0.01, 0.298, -0.0483, -0.01, 0.298, -0.0461, -0.01, 0.3004, -0.0483, -0.01, 0.3004, -0.0461, -0.0061, 0.302, -0.0526,
-0.0061, 0.302, -0.0505, -0.0083, 0.302, -0.0526, -0.0083, 0.302, -0.0505, -0.0061, 0.3049, -0.0523, -0.0083, 0.3049, -0.0523, -0.0061, 0.302, -0.0483, -0.0061, 0.302, -0.0454, -0.0083, 0.302, -0.0483, -0.0083, 0.302, -0.0454, -0.0061, 0.3047, -0.0454, -0.0083, 0.3047, -0.0454, -0.01, 0.302, -0.0526, -0.01, 0.302, -0.0505, -0.01, 0.3049, -0.0523, -0.01, 0.302, -0.0483, -0.01, 0.302, -0.0454, -0.01, 0.3047, -0.0454, -0.0061, 0.298, -0.0439, -0.0061, 0.298, -0.0425, -0.0083, 0.298, -0.0439, -0.0083, 0.298, -0.0425, -0.0061, 0.3004, -0.0425, -0.0083, 0.3004, -0.0425, -0.01, 0.298, -0.0439, -0.01, 0.298, -0.0425, -0.01, 0.3004, -0.0439, -0.01, 0.3004, -0.0425, -0.0061, 0.3025, -0.0425, -0.0083, 0.3025, -0.0425, -0.0061, 0.3047, -0.0425,
-0.0083, 0.3047, -0.0425, -0.01, 0.3025, -0.0439, -0.01, 0.3025, -0.0425, -0.01, 0.3047, -0.0439, -0.01, 0.3047, -0.0425, -0.0061, 0.3069, -0.0523, -0.0083, 0.3069, -0.0523, -0.0061, 0.309, -0.0523, -0.0083, 0.309, -0.0523, -0.0061, 0.3069, -0.0454, -0.0083, 0.3069, -0.0454, -0.0061, 0.309, -0.0454, -0.0083, 0.309, -0.0454, -0.01, 0.3069, -0.0523, -0.01, 0.309, -0.0523, -0.01, 0.3069, -0.0454, -0.01, 0.309, -0.0454, -0.0061, 0.3112, -0.0523, -0.0083, 0.3112, -0.0523, -0.0061, 0.3134, -0.0523, -0.0083, 0.3134, -0.0523, -0.0061, 0.3112, -0.0454, -0.0083, 0.3112, -0.0454, -0.0061, 0.3134, -0.0454, -0.0083, 0.3134, -0.0454, -0.01, 0.3112, -0.0523, -0.01, 0.3134, -0.0523, -0.01, 0.3112, -0.0454, -0.01, 0.3134, -0.0454, -0.0061, 0.3069, -0.0425,
-0.0083, 0.3069, -0.0425, -0.0061, 0.309, -0.0425, -0.0083, 0.309, -0.0425, -0.01, 0.3069, -0.0439, -0.01, 0.3069, -0.0425, -0.01, 0.309, -0.0439, -0.01, 0.309, -0.0425, -0.0061, 0.3112, -0.0425, -0.0083, 0.3112, -0.0425, -0.0061, 0.3134, -0.0425, -0.0083, 0.3134, -0.0425, -0.01, 0.3112, -0.0439, -0.01, 0.3112, -0.0425, -0.01, 0.3134, -0.0439, -0.01, 0.3134, -0.0425, -0.0061, 0.2805, -0.0152, -0.0083, 0.2805, -0.0152, -0.0061, 0.2815, -0.0143, -0.0083, 0.2815, -0.0143, -0.0061, 0.2828, -0.0133, -0.0061, 0.2838, -0.0122, -0.0083, 0.2828, -0.0133, -0.0083, 0.2838, -0.0122, -0.01, 0.2804, -0.0152, -0.01, 0.2815, -0.0144, -0.01, 0.2808, -0.0114, -0.01, 0.2827, -0.0133, -0.01, 0.2838, -0.0122, -0.0061, 0.2847, -0.0111, -0.0083, 0.2847, -0.0111,
-0.01, 0.2848, -0.0111, -0.01, 0.2808, -0.0092, -0.01, 0.2808, -0.0071, -0.01, 0.283, -0.0092, -0.01, 0.283, -0.0071, -0.01, 0.2808, -0.0049, -0.01, 0.2808, -0.0027, -0.01, 0.283, -0.0049, -0.01, 0.283, -0.0027, -0.0061, 0.2858, -0.0097, -0.0083, 0.2858, -0.0097, -0.0061, 0.2865, -0.0086, -0.0061, 0.2874, -0.0071, -0.0083, 0.2865, -0.0086, -0.0083, 0.2874, -0.0071, -0.0061, 0.2882, -0.0054, -0.0083, 0.2882, -0.0054, -0.01, 0.2858, -0.0096, -0.01, 0.2852, -0.0071, -0.01, 0.2866, -0.0085, -0.01, 0.2873, -0.0073, -0.01, 0.2852, -0.0049, -0.01, 0.2852, -0.0027, -0.01, 0.2883, -0.0052, -0.01, 0.2873, -0.0027, -0.0061, 0.2886, -0.0043, -0.0061, 0.2892, -0.0027, -0.0083, 0.2886, -0.0043, -0.0083, 0.2892, -0.0027, -0.01, 0.2887, -0.0042,
-0.01, 0.2891, -0.0028, -0.0061, 0.3088, -0.0219, -0.0061, 0.3088, -0.0201, -0.0083, 0.3088, -0.0219, -0.0083, 0.3088, -0.0201, -0.01, 0.3088, -0.0219, -0.01, 0.3088, -0.0201, -0.0061, 0.3112, -0.0219, -0.0083, 0.3112, -0.0219, -0.0061, 0.3134, -0.0219, -0.0083, 0.3134, -0.0219, -0.01, 0.3112, -0.0219, -0.01, 0.3112, -0.0201, -0.01, 0.3134, -0.0219, -0.01, 0.3134, -0.0201, -0.0061, 0.3088, -0.0179, -0.0061, 0.3088, -0.0157, -0.0083, 0.3088, -0.0179, -0.0083, 0.3088, -0.0157, -0.0061, 0.3088, -0.0136, -0.0061, 0.3088, -0.0114, -0.0083, 0.3088, -0.0136, -0.0083, 0.3088, -0.0114, -0.01, 0.3088, -0.0179, -0.01, 0.3088, -0.0157, -0.01, 0.3088, -0.0136, -0.01, 0.3088, -0.0114, -0.01, 0.3112, -0.0179, -0.01, 0.3112, -0.0157, -0.01, 0.3134, -0.0179,
-0.01, 0.3134, -0.0157, -0.01, 0.3112, -0.0136, -0.01, 0.3112, -0.0114, -0.01, 0.3134, -0.0136, -0.01, 0.3134, -0.0114, -0.0061, 0.3088, -0.0092, -0.0061, 0.3088, -0.0071, -0.0083, 0.3088, -0.0092, -0.0083, 0.3088, -0.0071, -0.0061, 0.3088, -0.0049, -0.0061, 0.3088, -0.0027, -0.0083, 0.3088, -0.0049, -0.0083, 0.3088, -0.0027, -0.01, 0.3088, -0.0092, -0.01, 0.3088, -0.0071, -0.01, 0.3088, -0.0049, -0.01, 0.3088, -0.0027, -0.01, 0.3112, -0.0092, -0.01, 0.3112, -0.0071, -0.01, 0.3134, -0.0092, -0.01, 0.3134, -0.0071, -0.01, 0.3112, -0.0049, -0.01, 0.3112, -0.0027, -0.01, 0.3134, -0.0049, -0.01, 0.3134, -0.0027, -0.0061, 0.3155, -0.067, -0.0083, 0.3155, -0.067, -0.0061, 0.3177, -0.067, -0.0083, 0.3177, -0.067, -0.0061, 0.3155, -0.0631,
-0.0083, 0.3155, -0.0631, -0.0061, 0.3177, -0.0631, -0.0083, 0.3177, -0.0631, -0.01, 0.3155, -0.067, -0.01, 0.3177, -0.067, -0.01, 0.3155, -0.0656, -0.01, 0.3155, -0.0631, -0.01, 0.3177, -0.0656, -0.01, 0.3177, -0.0631, -0.0061, 0.3196, -0.067, -0.0083, 0.3196, -0.067, -0.0061, 0.3196, -0.0656, -0.0061, 0.3196, -0.0631, -0.0083, 0.3196, -0.0656, -0.0083, 0.3196, -0.0631, -0.01, 0.3196, -0.067, -0.01, 0.3196, -0.0656, -0.01, 0.3196, -0.0631, -0.0061, 0.3155, -0.0562, -0.0083, 0.3155, -0.0562, -0.0061, 0.3176, -0.0562, -0.0061, 0.3176, -0.0548, -0.0083, 0.3176, -0.0562, -0.0083, 0.3176, -0.0548, -0.01, 0.3155, -0.0562, -0.01, 0.3155, -0.0548, -0.01, 0.3176, -0.0562, -0.01, 0.3176, -0.0548, -0.0061, 0.3155, -0.0523, -0.0083, 0.3155, -0.0523,
-0.0061, 0.3176, -0.0523, -0.0083, 0.3176, -0.0523, -0.0061, 0.3155, -0.0454, -0.0083, 0.3155, -0.0454, -0.0061, 0.3177, -0.0454, -0.0083, 0.3177, -0.0454, -0.01, 0.3155, -0.0523, -0.01, 0.3176, -0.0523, -0.01, 0.3155, -0.0454, -0.01, 0.3177, -0.0454, -0.0061, 0.3196, -0.0454, -0.0083, 0.3196, -0.0454, -0.01, 0.3196, -0.0454, -0.0061, 0.3155, -0.0425, -0.0083, 0.3155, -0.0425, -0.0061, 0.3177, -0.0425, -0.0083, 0.3177, -0.0425, -0.01, 0.3155, -0.0439, -0.01, 0.3155, -0.0425, -0.01, 0.3177, -0.0439, -0.01, 0.3177, -0.0425, -0.0061, 0.3196, -0.0439, -0.0061, 0.3196, -0.0425, -0.0083, 0.3196, -0.0439, -0.0083, 0.3196, -0.0425, -0.01, 0.3196, -0.0439, -0.01, 0.3196, -0.0425, -0.0061, 0.3402, -0.067, -0.0083, 0.3402, -0.067, -0.0061, 0.3402, -0.0656,
-0.0061, 0.3402, -0.0635, -0.0083, 0.3402, -0.0656, -0.0083, 0.3402, -0.0635, -0.01, 0.3402, -0.067, -0.01, 0.3402, -0.0656, -0.01, 0.3402, -0.0635, -0.0061, 0.3402, -0.0613, -0.0061, 0.3402, -0.0591, -0.0083, 0.3402, -0.0613, -0.0083, 0.3402, -0.0591, -0.0061, 0.3402, -0.057, -0.0061, 0.3402, -0.0548, -0.0083, 0.3402, -0.057, -0.0083, 0.3402, -0.0548, -0.01, 0.3402, -0.0613, -0.01, 0.3402, -0.0591, -0.01, 0.3402, -0.057, -0.01, 0.3402, -0.0548, -0.0061, 0.3416, -0.067, -0.0083, 0.3416, -0.067, -0.0061, 0.3431, -0.067, -0.0083, 0.3431, -0.067, -0.0061, 0.3431, -0.0656, -0.0061, 0.3431, -0.0635, -0.0083, 0.3431, -0.0656, -0.0083, 0.3431, -0.0635, -0.01, 0.3416, -0.067, -0.01, 0.3431, -0.067, -0.01, 0.3416, -0.0656, -0.01, 0.3416, -0.0635,
-0.01, 0.3431, -0.0656, -0.01, 0.3431, -0.0635, -0.0061, 0.3431, -0.0613, -0.0061, 0.3431, -0.0591, -0.0083, 0.3431, -0.0613, -0.0083, 0.3431, -0.0591, -0.0061, 0.3431, -0.057, -0.0061, 0.3431, -0.0548, -0.0083, 0.3431, -0.057, -0.0083, 0.3431, -0.0548, -0.01, 0.3416, -0.0613, -0.01, 0.3416, -0.0591, -0.01, 0.3431, -0.0613, -0.01, 0.3431, -0.0591, -0.01, 0.3416, -0.057, -0.01, 0.3416, -0.0548, -0.01, 0.3431, -0.057, -0.01, 0.3431, -0.0548, -0.0061, 0.3483, -0.0545, -0.0083, 0.3483, -0.0545, -0.01, 0.3483, -0.0546, -0.0061, 0.3402, -0.0526, -0.0061, 0.3402, -0.0505, -0.0083, 0.3402, -0.0526, -0.0083, 0.3402, -0.0505, -0.0061, 0.3402, -0.0483, -0.0061, 0.3402, -0.0461, -0.0083, 0.3402, -0.0483, -0.0083, 0.3402, -0.0461, -0.01, 0.3402, -0.0526,
-0.01, 0.3402, -0.0505, -0.01, 0.3402, -0.0483, -0.01, 0.3402, -0.0461, -0.0061, 0.3402, -0.0439, -0.0061, 0.3402, -0.0425, -0.0083, 0.3402, -0.0439, -0.0083, 0.3402, -0.0425, -0.01, 0.3402, -0.0439, -0.01, 0.3402, -0.0425, -0.0061, 0.3431, -0.0526, -0.0061, 0.3431, -0.0499, -0.0083, 0.3431, -0.0526, -0.0083, 0.3431, -0.0499, -0.01, 0.3416, -0.0526, -0.01, 0.3416, -0.0505, -0.01, 0.3431, -0.0526, -0.01, 0.3431, -0.0499, -0.01, 0.3416, -0.0483, -0.01, 0.3416, -0.0461, -0.01, 0.3438, -0.0483, -0.01, 0.3438, -0.0461, -0.0061, 0.346, -0.0524, -0.0061, 0.3449, -0.0515, -0.0083, 0.346, -0.0524, -0.0083, 0.3449, -0.0515, -0.0061, 0.3472, -0.0535, -0.0083, 0.3472, -0.0535, -0.0061, 0.3462, -0.0457, -0.0083, 0.3462, -0.0457, -0.0061, 0.3485, -0.0479,
-0.0061, 0.3474, -0.0468, -0.0083, 0.3485, -0.0479, -0.0083, 0.3474, -0.0468, -0.01, 0.346, -0.0525, -0.01, 0.345, -0.0515, -0.01, 0.3472, -0.0535, -0.01, 0.3481, -0.0505, -0.01, 0.3459, -0.0483, -0.01, 0.3462, -0.0457, -0.01, 0.3485, -0.0479, -0.01, 0.3474, -0.0468, -0.0061, 0.3423, -0.043, -0.0061, 0.3413, -0.0426, -0.0083, 0.3423, -0.043, -0.0083, 0.3413, -0.0426, -0.0061, 0.3438, -0.0438, -0.0083, 0.3438, -0.0438, -0.01, 0.3421, -0.0429, -0.01, 0.3411, -0.0426, -0.01, 0.3441, -0.0439, -0.0061, 0.3452, -0.0448, -0.0083, 0.3452, -0.0448, -0.01, 0.3451, -0.0448, -0.0061, 0.3155, -0.0219, -0.0083, 0.3155, -0.0219, -0.0061, 0.3177, -0.0219, -0.0083, 0.3177, -0.0219, -0.01, 0.3155, -0.0219, -0.01, 0.3155, -0.0201, -0.01, 0.3177, -0.0219,
-0.01, 0.3177, -0.0201, -0.0061, 0.3206, -0.0219, -0.0061, 0.3206, -0.0201, -0.0083, 0.3206, -0.0219, -0.0083, 0.3206, -0.0201, -0.01, 0.3206, -0.0219, -0.01, 0.3206, -0.0201, -0.01, 0.3155, -0.0179, -0.01, 0.3155, -0.0157, -0.01, 0.3177, -0.0179, -0.01, 0.3177, -0.0157, -0.01, 0.3155, -0.0136, -0.01, 0.3155, -0.0114, -0.01, 0.3177, -0.0136, -0.01, 0.3177, -0.0114, -0.0061, 0.3206, -0.0179, -0.0061, 0.3206, -0.0157, -0.0083, 0.3206, -0.0179, -0.0083, 0.3206, -0.0157, -0.0061, 0.3206, -0.0136, -0.0061, 0.3206, -0.0114, -0.0083, 0.3206, -0.0136, -0.0083, 0.3206, -0.0114, -0.01, 0.3206, -0.0179, -0.01, 0.3206, -0.0157, -0.01, 0.3206, -0.0136, -0.01, 0.3206, -0.0114, -0.01, 0.3155, -0.0092, -0.01, 0.3155, -0.0071, -0.01, 0.3177, -0.0092,
-0.01, 0.3177, -0.0071, -0.01, 0.3155, -0.0049, -0.01, 0.3155, -0.0027, -0.01, 0.3177, -0.0049, -0.01, 0.3177, -0.0027, -0.0061, 0.3206, -0.0092, -0.0061, 0.3206, -0.0071, -0.0083, 0.3206, -0.0092, -0.0083, 0.3206, -0.0071, -0.0061, 0.3206, -0.0049, -0.0061, 0.3206, -0.0027, -0.0083, 0.3206, -0.0049, -0.0083, 0.3206, -0.0027, -0.01, 0.3206, -0.0092, -0.01, 0.3206, -0.0071, -0.01, 0.3206, -0.0049, -0.01, 0.3206, -0.0027, -0.0061, 0.3402, -0.0219, -0.0061, 0.3402, -0.0201, -0.0083, 0.3402, -0.0219, -0.0083, 0.3402, -0.0201, -0.01, 0.3402, -0.0219, -0.01, 0.3402, -0.0201, -0.0061, 0.3416, -0.0219, -0.0083, 0.3416, -0.0219, -0.0061, 0.3438, -0.0219, -0.0083, 0.3438, -0.0219, -0.01, 0.3416, -0.0219, -0.01, 0.3416, -0.0201, -0.01, 0.3438, -0.0219,
-0.01, 0.3438, -0.0201, -0.0061, 0.3459, -0.0219, -0.0083, 0.3459, -0.0219, -0.0061, 0.3481, -0.0219, -0.0083, 0.3481, -0.0219, -0.01, 0.3459, -0.0219, -0.01, 0.3459, -0.0201, -0.01, 0.3481, -0.0219, -0.01, 0.3481, -0.0201, -0.0061, 0.3402, -0.0179, -0.0061, 0.3402, -0.0157, -0.0083, 0.3402, -0.0179, -0.0083, 0.3402, -0.0157, -0.0061, 0.3402, -0.0136, -0.0061, 0.3402, -0.0114, -0.0083, 0.3402, -0.0136, -0.0083, 0.3402, -0.0114, -0.01, 0.3402, -0.0179, -0.01, 0.3402, -0.0157, -0.01, 0.3402, -0.0136, -0.01, 0.3402, -0.0114, -0.0061, 0.3402, -0.0092, -0.0061, 0.3402, -0.0071, -0.0083, 0.3402, -0.0092, -0.0083, 0.3402, -0.0071, -0.0061, 0.3402, -0.0049, -0.0061, 0.3402, -0.0027, -0.0083, 0.3402, -0.0049, -0.0083, 0.3402, -0.0027, -0.01, 0.3402, -0.0092,
-0.01, 0.3402, -0.0071, -0.01, 0.3402, -0.0049, -0.01, 0.3402, -0.0027, -0.01, 0.3416, -0.0179, -0.01, 0.3416, -0.0157, -0.01, 0.3438, -0.0179, -0.01, 0.3438, -0.0157, -0.01, 0.3416, -0.0136, -0.01, 0.3416, -0.0114, -0.01, 0.3438, -0.0136, -0.01, 0.3438, -0.0114, -0.01, 0.3459, -0.0179, -0.01, 0.3459, -0.0157, -0.01, 0.3481, -0.0179, -0.01, 0.3481, -0.0157, -0.01, 0.3459, -0.0136, -0.01, 0.3459, -0.0114, -0.01, 0.3481, -0.0136, -0.01, 0.3481, -0.0114, -0.01, 0.3416, -0.0092, -0.01, 0.3416, -0.0071, -0.01, 0.3438, -0.0092, -0.01, 0.3438, -0.0071, -0.01, 0.3416, -0.0049, -0.01, 0.3416, -0.0027, -0.01, 0.3438, -0.0049, -0.01, 0.3438, -0.0027, -0.01, 0.3459, -0.0092, -0.01, 0.3459, -0.0071, -0.01, 0.3481, -0.0092,
-0.01, 0.3481, -0.0071, -0.01, 0.3459, -0.0049, -0.01, 0.3459, -0.0027, -0.01, 0.3481, -0.0049, -0.01, 0.3481, -0.0027, -0.0061, 0.357, -0.0631, -0.0083, 0.357, -0.0631, -0.01, 0.3571, -0.0631, -0.0061, 0.3527, -0.0589, -0.0083, 0.3527, -0.0589, -0.0061, 0.3505, -0.0567, -0.0061, 0.3494, -0.0556, -0.0083, 0.3505, -0.0567, -0.0083, 0.3494, -0.0556, -0.0061, 0.3516, -0.0578, -0.0083, 0.3516, -0.0578, -0.01, 0.3527, -0.0589, -0.01, 0.3505, -0.0567, -0.01, 0.3494, -0.0556, -0.01, 0.3516, -0.0578, -0.01, 0.3524, -0.0548, -0.0061, 0.3549, -0.061, -0.0061, 0.3538, -0.06, -0.0083, 0.3549, -0.061, -0.0083, 0.3538, -0.06, -0.0061, 0.356, -0.0621, -0.0083, 0.356, -0.0621, -0.0061, 0.355, -0.0544, -0.0083, 0.355, -0.0544, -0.0061, 0.3572, -0.0565,
-0.0061, 0.3562, -0.0554, -0.0083, 0.3572, -0.0565, -0.0083, 0.3562, -0.0554, -0.01, 0.3549, -0.061, -0.01, 0.3538, -0.06, -0.01, 0.356, -0.0621, -0.01, 0.3568, -0.0591, -0.01, 0.3546, -0.057, -0.01, 0.355, -0.0544, -0.01, 0.3572, -0.0565, -0.01, 0.3561, -0.0554, -0.0061, 0.3594, -0.065, -0.0061, 0.3583, -0.0642, -0.0083, 0.3594, -0.065, -0.0083, 0.3583, -0.0642, -0.0061, 0.361, -0.0662, -0.0083, 0.361, -0.0662, -0.01, 0.3594, -0.0651, -0.01, 0.3583, -0.0642, -0.01, 0.3608, -0.066, -0.01, 0.3611, -0.0635, -0.0061, 0.3637, -0.067, -0.0061, 0.3637, -0.0635, -0.0083, 0.3637, -0.067, -0.0083, 0.3637, -0.0635, -0.01, 0.3637, -0.067, -0.01, 0.3637, -0.0635, -0.0061, 0.3596, -0.0584, -0.0083, 0.3596, -0.0584, -0.0061, 0.3608, -0.0591,
-0.0083, 0.3608, -0.0591, -0.0061, 0.3585, -0.0575, -0.0083, 0.3585, -0.0575, -0.0061, 0.3608, -0.057, -0.0061, 0.3608, -0.0548, -0.0083, 0.3608, -0.057, -0.0083, 0.3608, -0.0548, -0.01, 0.3589, -0.0613, -0.01, 0.3595, -0.0583, -0.01, 0.3611, -0.0613, -0.01, 0.3608, -0.0591, -0.01, 0.3585, -0.0575, -0.01, 0.3608, -0.057, -0.01, 0.3608, -0.0548, -0.0061, 0.3637, -0.0613, -0.0061, 0.3637, -0.0591, -0.0083, 0.3637, -0.0613, -0.0083, 0.3637, -0.0591, -0.0061, 0.3637, -0.057, -0.0061, 0.3637, -0.0548, -0.0083, 0.3637, -0.057, -0.0083, 0.3637, -0.0548, -0.01, 0.3637, -0.0613, -0.01, 0.3637, -0.0591, -0.01, 0.3637, -0.057, -0.01, 0.3637, -0.0548, -0.0061, 0.3507, -0.05, -0.0083, 0.3507, -0.05, -0.0061, 0.3529, -0.0522, -0.0061, 0.3518, -0.0511,
-0.0083, 0.3529, -0.0522, -0.0083, 0.3518, -0.0511, -0.0061, 0.3496, -0.049, -0.0083, 0.3496, -0.049, -0.01, 0.3503, -0.0526, -0.01, 0.3507, -0.05, -0.01, 0.3529, -0.0522, -0.01, 0.3518, -0.0511, -0.01, 0.3496, -0.0489, -0.0061, 0.354, -0.0533, -0.0083, 0.354, -0.0533, -0.01, 0.354, -0.0533, -0.0061, 0.3608, -0.0526, -0.0061, 0.3608, -0.0505, -0.0083, 0.3608, -0.0526, -0.0083, 0.3608, -0.0505, -0.0061, 0.3608, -0.0483, -0.0061, 0.3608, -0.0461, -0.0083, 0.3608, -0.0483, -0.0083, 0.3608, -0.0461, -0.01, 0.3608, -0.0526, -0.01, 0.3608, -0.0505, -0.01, 0.3608, -0.0483, -0.01, 0.3608, -0.0461, -0.0061, 0.3637, -0.0526, -0.0061, 0.3637, -0.0505, -0.0083, 0.3637, -0.0526, -0.0083, 0.3637, -0.0505, -0.0061, 0.3637, -0.0483, -0.0061, 0.3637, -0.0461,
-0.0083, 0.3637, -0.0483, -0.0083, 0.3637, -0.0461, -0.01, 0.3637, -0.0526, -0.01, 0.3637, -0.0505, -0.01, 0.3637, -0.0483, -0.01, 0.3637, -0.0461, -0.0061, 0.3608, -0.0439, -0.0061, 0.3608, -0.0425, -0.0083, 0.3608, -0.0439, -0.0083, 0.3608, -0.0425, -0.01, 0.3608, -0.0439, -0.01, 0.3608, -0.0425, -0.0061, 0.3637, -0.0439, -0.0061, 0.3637, -0.0425, -0.0083, 0.3637, -0.0439, -0.0083, 0.3637, -0.0425, -0.01, 0.3637, -0.0439, -0.01, 0.3637, -0.0425, -0.0061, 0.3503, -0.0219, -0.0083, 0.3503, -0.0219, -0.0061, 0.3524, -0.0219, -0.0083, 0.3524, -0.0219, -0.01, 0.3503, -0.0219, -0.01, 0.3503, -0.0201, -0.01, 0.3524, -0.0219, -0.01, 0.3524, -0.0201, -0.0061, 0.3546, -0.0219, -0.0083, 0.3546, -0.0219, -0.0061, 0.3568, -0.0219, -0.0083, 0.3568, -0.0219,
-0.01, 0.3546, -0.0219, -0.01, 0.3546, -0.0201, -0.01, 0.3568, -0.0219, -0.01, 0.3568, -0.0201, -0.0061, 0.3589, -0.0219, -0.0083, 0.3589, -0.0219, -0.0061, 0.3611, -0.0219, -0.0083, 0.3611, -0.0219, -0.01, 0.3589, -0.0219, -0.01, 0.3589, -0.0201, -0.01, 0.3611, -0.0219, -0.01, 0.3611, -0.0201, -0.0061, 0.3633, -0.0219, -0.0083, 0.3633, -0.0219, -0.0061, 0.3655, -0.0219, -0.0083, 0.3655, -0.0219, -0.01, 0.3633, -0.0219, -0.01, 0.3633, -0.0201, -0.01, 0.3655, -0.0219, -0.01, 0.3655, -0.0201, -0.01, 0.3503, -0.0179, -0.01, 0.3503, -0.0157, -0.01, 0.3524, -0.0179, -0.01, 0.3524, -0.0157, -0.01, 0.3503, -0.0136, -0.01, 0.3503, -0.0114, -0.01, 0.3524, -0.0136, -0.01, 0.3524, -0.0114, -0.01, 0.3546, -0.0179, -0.01, 0.3546, -0.0157,
-0.01, 0.3568, -0.0179, -0.01, 0.3568, -0.0157, -0.01, 0.3546, -0.0136, -0.01, 0.3546, -0.0114, -0.01, 0.3568, -0.0136, -0.01, 0.3568, -0.0114, -0.0061, 0.352, -0.0101, -0.0061, 0.352, -0.0071, -0.0083, 0.352, -0.0101, -0.0083, 0.352, -0.0071, -0.0061, 0.352, -0.0049, -0.0061, 0.352, -0.0027, -0.0083, 0.352, -0.0049, -0.0083, 0.352, -0.0027, -0.01, 0.3503, -0.0092, -0.01, 0.3503, -0.0071, -0.01, 0.352, -0.0101, -0.01, 0.352, -0.0071, -0.01, 0.3503, -0.0049, -0.01, 0.3503, -0.0027, -0.01, 0.352, -0.0049, -0.01, 0.352, -0.0027, -0.0061, 0.3546, -0.0101, -0.0083, 0.3546, -0.0101, -0.0061, 0.3568, -0.0101, -0.0083, 0.3568, -0.0101, -0.01, 0.3546, -0.0101, -0.01, 0.3568, -0.0101, -0.01, 0.3589, -0.0179, -0.01, 0.3589, -0.0157,
-0.01, 0.3611, -0.0179, -0.01, 0.3611, -0.0157, -0.01, 0.3589, -0.0136, -0.01, 0.3589, -0.0114, -0.01, 0.3611, -0.0136, -0.01, 0.3611, -0.0114, -0.01, 0.3633, -0.0179, -0.01, 0.3633, -0.0157, -0.01, 0.3655, -0.0179, -0.01, 0.3655, -0.0157, -0.01, 0.3633, -0.0136, -0.01, 0.3633, -0.0114, -0.01, 0.3655, -0.0136, -0.01, 0.3655, -0.0114, -0.0061, 0.3589, -0.0101, -0.0083, 0.3589, -0.0101, -0.0061, 0.3611, -0.0101, -0.0083, 0.3611, -0.0101, -0.01, 0.3589, -0.0101, -0.01, 0.3611, -0.0101, -0.0061, 0.3633, -0.0101, -0.0083, 0.3633, -0.0101, -0.0061, 0.3655, -0.0101, -0.0083, 0.3655, -0.0101, -0.01, 0.3633, -0.0101, -0.01, 0.3655, -0.0101, -0.0061, 0.3676, -0.0219, -0.0083, 0.3676, -0.0219, -0.0061, 0.3698, -0.0219, -0.0083, 0.3698, -0.0219,
-0.01, 0.3676, -0.0219, -0.01, 0.3676, -0.0201, -0.01, 0.3698, -0.0219, -0.01, 0.3698, -0.0201, -0.0061, 0.372, -0.0219, -0.0083, 0.372, -0.0219, -0.0061, 0.3741, -0.0219, -0.0083, 0.3741, -0.0219, -0.01, 0.372, -0.0219, -0.01, 0.372, -0.0201, -0.01, 0.3741, -0.0219, -0.01, 0.3741, -0.0201, -0.0061, 0.3763, -0.0219, -0.0083, 0.3763, -0.0219, -0.0061, 0.3785, -0.0219, -0.0083, 0.3785, -0.0219, -0.01, 0.3763, -0.0219, -0.01, 0.3763, -0.0201, -0.01, 0.3785, -0.0219, -0.01, 0.3785, -0.0201, -0.0061, 0.3807, -0.0218, -0.0083, 0.3807, -0.0218, -0.0061, 0.3828, -0.0215, -0.0083, 0.3828, -0.0215, -0.01, 0.3807, -0.0218, -0.01, 0.3807, -0.0201, -0.01, 0.3828, -0.0215, -0.01, 0.3828, -0.0201, -0.01, 0.3676, -0.0179, -0.01, 0.3676, -0.0157,
-0.01, 0.3698, -0.0179, -0.01, 0.3698, -0.0157, -0.01, 0.3676, -0.0136, -0.01, 0.3676, -0.0114, -0.01, 0.3698, -0.0136, -0.01, 0.3698, -0.0114, -0.01, 0.372, -0.0179, -0.01, 0.372, -0.0157, -0.01, 0.3741, -0.0179, -0.01, 0.3741, -0.0157, -0.01, 0.372, -0.0136, -0.01, 0.372, -0.0114, -0.01, 0.3741, -0.0136, -0.01, 0.3741, -0.0114, -0.0061, 0.3676, -0.0101, -0.0083, 0.3676, -0.0101, -0.0061, 0.3698, -0.0101, -0.0083, 0.3698, -0.0101, -0.01, 0.3676, -0.0101, -0.01, 0.3698, -0.0101, -0.0061, 0.372, -0.0101, -0.0083, 0.372, -0.0101, -0.0061, 0.3741, -0.0101, -0.0083, 0.3741, -0.0101, -0.01, 0.372, -0.0101, -0.01, 0.3741, -0.0101, -0.01, 0.3763, -0.0179, -0.01, 0.3763, -0.0157, -0.01, 0.3785, -0.0179, -0.01, 0.3785, -0.0157,
-0.01, 0.3763, -0.0136, -0.01, 0.3763, -0.0114, -0.01, 0.3785, -0.0136, -0.01, 0.3785, -0.0114, -0.01, 0.3807, -0.0179, -0.01, 0.3807, -0.0157, -0.01, 0.3828, -0.0179, -0.01, 0.3828, -0.0157, -0.01, 0.3807, -0.0136, -0.01, 0.3807, -0.0114, -0.01, 0.3828, -0.0136, -0.01, 0.3828, -0.0114, -0.0061, 0.3763, -0.0101, -0.0083, 0.3763, -0.0101, -0.0061, 0.3785, -0.0101, -0.0083, 0.3785, -0.0101, -0.01, 0.3763, -0.0101, -0.01, 0.3785, -0.0101, -0.0061, 0.3807, -0.0101, -0.0083, 0.3807, -0.0101, -0.0061, 0.3828, -0.0099, -0.0083, 0.3828, -0.0099, -0.01, 0.3806, -0.0101, -0.01, 0.3828, -0.0099, -0.0061, 0.3863, -0.067, -0.0083, 0.3863, -0.067, -0.0061, 0.3863, -0.0656, -0.0061, 0.3863, -0.0635, -0.0083, 0.3863, -0.0656, -0.0083, 0.3863, -0.0635,
-0.01, 0.3863, -0.067, -0.01, 0.3863, -0.0656, -0.01, 0.3863, -0.0635, -0.0061, 0.3893, -0.067, -0.0083, 0.3893, -0.067, -0.0061, 0.3915, -0.067, -0.0083, 0.3915, -0.067, -0.0061, 0.3902, -0.0631, -0.0083, 0.3902, -0.0631, -0.0061, 0.3915, -0.0631, -0.0083, 0.3915, -0.0631, -0.01, 0.3893, -0.067, -0.01, 0.3915, -0.067, -0.01, 0.3893, -0.0656, -0.01, 0.3902, -0.0631, -0.01, 0.3915, -0.0656, -0.01, 0.3915, -0.0631, -0.0061, 0.3863, -0.0613, -0.0061, 0.3863, -0.0591, -0.0083, 0.3863, -0.0613, -0.0083, 0.3863, -0.0591, -0.0061, 0.3863, -0.057, -0.0061, 0.3863, -0.0548, -0.0083, 0.3863, -0.057, -0.0083, 0.3863, -0.0548, -0.01, 0.3863, -0.0613, -0.01, 0.3863, -0.0591, -0.01, 0.3863, -0.057, -0.01, 0.3863, -0.0548, -0.0061, 0.3902, -0.0613,
-0.0061, 0.3902, -0.0591, -0.0083, 0.3902, -0.0613, -0.0083, 0.3902, -0.0591, -0.0061, 0.3902, -0.057, -0.0061, 0.3902, -0.0548, -0.0083, 0.3902, -0.057, -0.0083, 0.3902, -0.0548, -0.01, 0.3902, -0.0613, -0.01, 0.3902, -0.0591, -0.01, 0.3902, -0.057, -0.01, 0.3902, -0.0548, -0.0061, 0.3937, -0.067, -0.0083, 0.3937, -0.067, -0.0061, 0.3958, -0.067, -0.0083, 0.3958, -0.067, -0.0061, 0.3937, -0.0631, -0.0083, 0.3937, -0.0631, -0.0061, 0.3958, -0.0631, -0.0083, 0.3958, -0.0631, -0.01, 0.3937, -0.067, -0.01, 0.3958, -0.067, -0.01, 0.3937, -0.0656, -0.01, 0.3937, -0.0631, -0.01, 0.3958, -0.0656, -0.01, 0.3958, -0.0631, -0.0061, 0.398, -0.067, -0.0083, 0.398, -0.067, -0.0061, 0.4002, -0.0669, -0.0083, 0.4002, -0.0669, -0.0061, 0.398, -0.063,
-0.0083, 0.398, -0.063, -0.0061, 0.4002, -0.063, -0.0083, 0.4002, -0.063, -0.01, 0.398, -0.067, -0.01, 0.4002, -0.0669, -0.01, 0.398, -0.0656, -0.01, 0.398, -0.063, -0.01, 0.4002, -0.0656, -0.01, 0.4001, -0.063, -0.0061, 0.3863, -0.0526, -0.0061, 0.3863, -0.0505, -0.0083, 0.3863, -0.0526, -0.0083, 0.3863, -0.0505, -0.0061, 0.3863, -0.0483, -0.0061, 0.3863, -0.0461, -0.0083, 0.3863, -0.0483, -0.0083, 0.3863, -0.0461, -0.01, 0.3863, -0.0526, -0.01, 0.3863, -0.0505, -0.01, 0.3863, -0.0483, -0.01, 0.3863, -0.0461, -0.0061, 0.3902, -0.0526, -0.0061, 0.3902, -0.0505, -0.0083, 0.3902, -0.0526, -0.0083, 0.3902, -0.0505, -0.0061, 0.3902, -0.0483, -0.0061, 0.3902, -0.0454, -0.0083, 0.3902, -0.0483, -0.0083, 0.3902, -0.0454, -0.0061, 0.3915, -0.0454,
-0.0083, 0.3915, -0.0454, -0.01, 0.3902, -0.0526, -0.01, 0.3902, -0.0505, -0.01, 0.3902, -0.0483, -0.01, 0.3902, -0.0454, -0.01, 0.3915, -0.0454, -0.0061, 0.3863, -0.0439, -0.0061, 0.3863, -0.0425, -0.0083, 0.3863, -0.0439, -0.0083, 0.3863, -0.0425, -0.01, 0.3863, -0.0439, -0.01, 0.3863, -0.0425, -0.0061, 0.3893, -0.0425, -0.0083, 0.3893, -0.0425, -0.0061, 0.3915, -0.0425, -0.0083, 0.3915, -0.0425, -0.01, 0.3893, -0.0439, -0.01, 0.3893, -0.0425, -0.01, 0.3915, -0.0439, -0.01, 0.3915, -0.0425, -0.0061, 0.3937, -0.0454, -0.0083, 0.3937, -0.0454, -0.0061, 0.3958, -0.0454, -0.0083, 0.3958, -0.0454, -0.01, 0.3937, -0.0454, -0.01, 0.3958, -0.0454, -0.0061, 0.398, -0.0455, -0.0083, 0.398, -0.0455, -0.0061, 0.4002, -0.0458, -0.0083, 0.4002, -0.0458,
-0.01, 0.398, -0.0455, -0.01, 0.4001, -0.0457, -0.0061, 0.3937, -0.0425, -0.0083, 0.3937, -0.0425, -0.0061, 0.3958, -0.0425, -0.0083, 0.3958, -0.0425, -0.01, 0.3937, -0.0439, -0.01, 0.3937, -0.0425, -0.01, 0.3958, -0.0439, -0.01, 0.3958, -0.0425, -0.0061, 0.398, -0.0425, -0.0083, 0.398, -0.0425, -0.0061, 0.4002, -0.0425, -0.0083, 0.4002, -0.0425, -0.01, 0.398, -0.0439, -0.01, 0.398, -0.0425, -0.01, 0.4002, -0.0439, -0.01, 0.4002, -0.0425, -0.0061, 0.4017, -0.0668, -0.0083, 0.4017, -0.0668, -0.0061, 0.4028, -0.0666, -0.0061, 0.4023, -0.0627, -0.0083, 0.4028, -0.0666, -0.0083, 0.4023, -0.0627, -0.0061, 0.4046, -0.0662, -0.0083, 0.4046, -0.0662, -0.01, 0.4016, -0.0668, -0.01, 0.4025, -0.0667, -0.01, 0.4023, -0.0627, -0.01, 0.4048, -0.0661,
-0.01, 0.4045, -0.0635, -0.0061, 0.4066, -0.0652, -0.0061, 0.4076, -0.0644, -0.0083, 0.4066, -0.0652, -0.0083, 0.4076, -0.0644, -0.0061, 0.409, -0.0635, -0.0083, 0.409, -0.0635, -0.01, 0.4064, -0.0653, -0.01, 0.4075, -0.0645, -0.01, 0.4089, -0.0635, -0.0061, 0.4033, -0.0623, -0.0083, 0.4033, -0.0623, -0.0061, 0.4046, -0.0619, -0.0083, 0.4046, -0.0619, -0.01, 0.4033, -0.0623, -0.01, 0.4045, -0.0619, -0.0061, 0.4062, -0.0608, -0.0061, 0.4074, -0.0595, -0.0083, 0.4062, -0.0608, -0.0083, 0.4074, -0.0595, -0.0061, 0.4099, -0.0624, -0.0061, 0.408, -0.0584, -0.0083, 0.4099, -0.0624, -0.0083, 0.408, -0.0584, -0.0061, 0.4085, -0.057, -0.0061, 0.4087, -0.0548, -0.0083, 0.4085, -0.057, -0.0083, 0.4087, -0.0548, -0.01, 0.4062, -0.0608, -0.01, 0.4073, -0.0596,
-0.01, 0.41, -0.0623, -0.01, 0.4079, -0.0587, -0.01, 0.4085, -0.0569, -0.01, 0.4087, -0.0548, -0.0061, 0.4107, -0.0613, -0.0061, 0.4119, -0.0594, -0.0083, 0.4107, -0.0613, -0.0083, 0.4119, -0.0594, -0.0061, 0.4122, -0.0582, -0.0083, 0.4122, -0.0582, -0.0061, 0.4125, -0.057, -0.0061, 0.4127, -0.0548, -0.0083, 0.4125, -0.057, -0.0083, 0.4127, -0.0548, -0.01, 0.4108, -0.0612, -0.01, 0.4119, -0.0593, -0.01, 0.4122, -0.0582, -0.01, 0.411, -0.057, -0.01, 0.411, -0.0548, -0.01, 0.4125, -0.057, -0.01, 0.4127, -0.0548, -0.0061, 0.4024, -0.0463, -0.0083, 0.4024, -0.0463, -0.0061, 0.405, -0.0476, -0.0061, 0.4039, -0.047, -0.0083, 0.405, -0.0476, -0.0083, 0.4039, -0.047, -0.01, 0.4022, -0.0462, -0.01, 0.4051, -0.0476, -0.01, 0.4041, -0.0471,
-0.0061, 0.4075, -0.0498, -0.0083, 0.4075, -0.0498, -0.0061, 0.4086, -0.0526, -0.0061, 0.4081, -0.0509, -0.0083, 0.4086, -0.0526, -0.0083, 0.4081, -0.0509, -0.0061, 0.4064, -0.0486, -0.0083, 0.4064, -0.0486, -0.0061, 0.4093, -0.0456, -0.0083, 0.4093, -0.0456, -0.01, 0.4074, -0.0498, -0.01, 0.4086, -0.0528, -0.01, 0.408, -0.0508, -0.01, 0.4064, -0.0486, -0.01, 0.4067, -0.0461, -0.01, 0.4089, -0.0483, -0.01, 0.4093, -0.0456, -0.0061, 0.4024, -0.0426, -0.0083, 0.4024, -0.0426, -0.0061, 0.4048, -0.0431, -0.0061, 0.4037, -0.0428, -0.0083, 0.4048, -0.0431, -0.0083, 0.4037, -0.0428, -0.01, 0.4024, -0.0439, -0.01, 0.4024, -0.0426, -0.01, 0.4047, -0.043, -0.01, 0.4036, -0.0428, -0.0061, 0.4067, -0.0439, -0.0083, 0.4067, -0.0439, -0.0061, 0.4082, -0.0448,
-0.0083, 0.4082, -0.0448, -0.01, 0.407, -0.044, -0.01, 0.4081, -0.0447, -0.0061, 0.412, -0.0499, -0.0083, 0.412, -0.0499, -0.0061, 0.4126, -0.0526, -0.0061, 0.4122, -0.051, -0.0083, 0.4126, -0.0526, -0.0083, 0.4122, -0.051, -0.0061, 0.4112, -0.0482, -0.0061, 0.4103, -0.0468, -0.0083, 0.4112, -0.0482, -0.0083, 0.4103, -0.0468, -0.01, 0.411, -0.0526, -0.01, 0.4121, -0.0502, -0.01, 0.4126, -0.0526, -0.01, 0.4123, -0.0511, -0.01, 0.4111, -0.048, -0.01, 0.4103, -0.0468, -0.0061, 0.3846, -0.0213, -0.0061, 0.3857, -0.0211, -0.0083, 0.3846, -0.0213, -0.0083, 0.3857, -0.0211, -0.0061, 0.3872, -0.0209, -0.0083, 0.3872, -0.0209, -0.01, 0.3844, -0.0213, -0.01, 0.3852, -0.0212, -0.01, 0.3873, -0.0209, -0.0061, 0.3893, -0.0206, -0.0083, 0.3893, -0.0206,
-0.0061, 0.3915, -0.0201, -0.0083, 0.3915, -0.0201, -0.01, 0.3895, -0.0206, -0.01, 0.3917, -0.02, -0.0061, 0.3935, -0.0194, -0.0083, 0.3935, -0.0194, -0.01, 0.3933, -0.0194, -0.01, 0.385, -0.0179, -0.01, 0.385, -0.0157, -0.01, 0.3872, -0.0179, -0.01, 0.3872, -0.0157, -0.01, 0.385, -0.0136, -0.01, 0.385, -0.0114, -0.01, 0.3872, -0.0136, -0.01, 0.3872, -0.0114, -0.01, 0.3893, -0.0179, -0.01, 0.3893, -0.0157, -0.01, 0.3915, -0.0179, -0.01, 0.3915, -0.0157, -0.01, 0.3893, -0.0136, -0.01, 0.3893, -0.0114, -0.01, 0.3915, -0.0136, -0.01, 0.3915, -0.0114, -0.0061, 0.385, -0.0096, -0.0083, 0.385, -0.0096, -0.0061, 0.3872, -0.009, -0.0083, 0.3872, -0.009, -0.01, 0.3849, -0.0096, -0.01, 0.387, -0.0091, -0.0061, 0.3891, -0.0084,
-0.0061, 0.3902, -0.0081, -0.0083, 0.3891, -0.0084, -0.0083, 0.3902, -0.0081, -0.0061, 0.3915, -0.0075, -0.0083, 0.3915, -0.0075, -0.01, 0.3892, -0.0084, -0.01, 0.3902, -0.008, -0.01, 0.3915, -0.0092, -0.01, 0.3914, -0.0076, -0.0061, 0.3946, -0.0189, -0.0083, 0.3946, -0.0189, -0.0061, 0.3959, -0.0185, -0.0083, 0.3959, -0.0185, -0.01, 0.3942, -0.0191, -0.01, 0.3937, -0.0157, -0.01, 0.3961, -0.0183, -0.01, 0.3958, -0.0157, -0.01, 0.3937, -0.0136, -0.01, 0.3937, -0.0114, -0.01, 0.3958, -0.0136, -0.01, 0.3958, -0.0114, -0.0061, 0.398, -0.0175, -0.0083, 0.398, -0.0175, -0.0061, 0.3992, -0.0169, -0.0061, 0.4003, -0.0162, -0.0083, 0.3992, -0.0169, -0.0083, 0.4003, -0.0162, -0.01, 0.3982, -0.0174, -0.01, 0.398, -0.0157, -0.01, 0.3991, -0.0169,
-0.01, 0.4003, -0.0162, -0.01, 0.398, -0.0136, -0.01, 0.398, -0.0114, -0.01, 0.4002, -0.0136, -0.01, 0.4002, -0.0114, -0.0061, 0.3935, -0.0065, -0.0083, 0.3935, -0.0065, -0.0061, 0.3945, -0.0058, -0.0083, 0.3945, -0.0058, -0.0061, 0.3959, -0.0049, -0.0083, 0.3959, -0.0049, -0.01, 0.3937, -0.0092, -0.01, 0.3935, -0.0065, -0.01, 0.3958, -0.0092, -0.01, 0.3958, -0.0071, -0.01, 0.3945, -0.0058, -0.01, 0.3956, -0.0051, -0.0061, 0.397, -0.0039, -0.0061, 0.3982, -0.0028, -0.0083, 0.397, -0.0039, -0.0083, 0.3982, -0.0028, -0.0061, 0.3991, -0.0016, -0.0083, 0.3991, -0.0016, -0.01, 0.398, -0.0092, -0.01, 0.398, -0.0071, -0.01, 0.4002, -0.0092, -0.01, 0.4002, -0.0071, -0.01, 0.3971, -0.0038, -0.01, 0.3981, -0.0028, -0.01, 0.4002, -0.0049,
-0.01, 0.399, -0.0017, -0.0061, 0.4018, -0.0151, -0.0083, 0.4018, -0.0151, -0.0061, 0.4029, -0.0142, -0.0083, 0.4029, -0.0142, -0.0061, 0.4042, -0.0132, -0.0061, 0.4053, -0.0122, -0.0083, 0.4042, -0.0132, -0.0083, 0.4053, -0.0122, -0.01, 0.4018, -0.0151, -0.01, 0.4029, -0.0143, -0.01, 0.4024, -0.0114, -0.01, 0.4042, -0.0132, -0.01, 0.4053, -0.0122, -0.0061, 0.4064, -0.0111, -0.0083, 0.4064, -0.0111, -0.01, 0.4064, -0.0111, -0.01, 0.4024, -0.0092, -0.01, 0.4024, -0.0071, -0.01, 0.4045, -0.0092, -0.01, 0.4045, -0.0071, -0.01, 0.4024, -0.0049, -0.01, 0.4024, -0.0027, -0.01, 0.4045, -0.0049, -0.01, 0.4045, -0.0027, -0.0061, 0.4075, -0.0099, -0.0083, 0.4075, -0.0099, -0.0061, 0.4084, -0.0089, -0.0061, 0.4094, -0.0074, -0.0083, 0.4084, -0.0089,
-0.0083, 0.4094, -0.0074, -0.01, 0.4075, -0.0099, -0.01, 0.4067, -0.0071, -0.01, 0.4084, -0.0088, -0.01, 0.4095, -0.0074, -0.01, 0.4067, -0.0049, -0.01, 0.4067, -0.0027, -0.01, 0.4089, -0.0049, -0.01, 0.4089, -0.0027, -0.0061, 0.4102, -0.0063, -0.0083, 0.4102, -0.0063, -0.0061, 0.4111, -0.0049, -0.0061, 0.4119, -0.0034, -0.0083, 0.4111, -0.0049, -0.0083, 0.4119, -0.0034, -0.0061, 0.4124, -0.0023, -0.0083, 0.4124, -0.0023, -0.01, 0.4102, -0.0063, -0.01, 0.4109, -0.0051, -0.01, 0.412, -0.0032, -0.01, 0.4125, -0.0021, -0.0061, 0.4324, -0.067, -0.0083, 0.4324, -0.067, -0.0061, 0.4349, -0.067, -0.0083, 0.4349, -0.067, -0.0061, 0.4324, -0.0656, -0.0061, 0.4324, -0.0635, -0.0083, 0.4324, -0.0656, -0.0083, 0.4324, -0.0635, -0.0061, 0.4353, -0.0631,
-0.0083, 0.4353, -0.0631, -0.01, 0.4324, -0.067, -0.01, 0.4349, -0.067, -0.01, 0.4324, -0.0656, -0.01, 0.4324, -0.0635, -0.01, 0.4349, -0.0656, -0.01, 0.4353, -0.0631, -0.0061, 0.4324, -0.0613, -0.0061, 0.4324, -0.0591, -0.0083, 0.4324, -0.0613, -0.0083, 0.4324, -0.0591, -0.0061, 0.4353, -0.0613, -0.0061, 0.4353, -0.0591, -0.0083, 0.4353, -0.0613, -0.0083, 0.4353, -0.0591, -0.0061, 0.4324, -0.057, -0.0061, 0.4324, -0.0548, -0.0083, 0.4324, -0.057, -0.0083, 0.4324, -0.0548, -0.0061, 0.4353, -0.057, -0.0061, 0.4353, -0.0548, -0.0083, 0.4353, -0.057, -0.0083, 0.4353, -0.0548, -0.01, 0.4324, -0.0613, -0.01, 0.4324, -0.0591, -0.01, 0.4353, -0.0613, -0.01, 0.4353, -0.0591, -0.01, 0.4324, -0.057, -0.01, 0.4324, -0.0548, -0.01, 0.4353, -0.057,
-0.01, 0.4353, -0.0548, -0.0061, 0.4324, -0.0526, -0.0061, 0.4324, -0.0505, -0.0083, 0.4324, -0.0526, -0.0083, 0.4324, -0.0505, -0.0061, 0.4353, -0.0526, -0.0061, 0.4355, -0.0497, -0.0083, 0.4353, -0.0526, -0.0083, 0.4355, -0.0497, -0.0061, 0.4324, -0.0483, -0.0061, 0.4324, -0.0461, -0.0083, 0.4324, -0.0483, -0.0083, 0.4324, -0.0461, -0.0061, 0.4358, -0.0489, -0.0083, 0.4358, -0.0489, -0.01, 0.4324, -0.0526, -0.01, 0.4324, -0.0505, -0.01, 0.4353, -0.0527, -0.01, 0.4353, -0.0506, -0.01, 0.4324, -0.0483, -0.01, 0.4324, -0.0461, -0.01, 0.4359, -0.0487, -0.01, 0.4349, -0.0461, -0.0061, 0.4324, -0.0439, -0.0061, 0.4324, -0.0425, -0.0083, 0.4324, -0.0439, -0.0083, 0.4324, -0.0425, -0.0061, 0.4349, -0.0425, -0.0083, 0.4349, -0.0425, -0.01, 0.4324, -0.0439,
-0.01, 0.4324, -0.0425, -0.01, 0.4349, -0.0439, -0.01, 0.4349, -0.0425, -0.0061, 0.4371, -0.067, -0.0083, 0.4371, -0.067, -0.0061, 0.4392, -0.067, -0.0083, 0.4392, -0.067, -0.0061, 0.4371, -0.0631, -0.0083, 0.4371, -0.0631, -0.0061, 0.4392, -0.0631, -0.0083, 0.4392, -0.0631, -0.01, 0.4371, -0.067, -0.01, 0.4392, -0.067, -0.01, 0.4371, -0.0656, -0.01, 0.4371, -0.0631, -0.01, 0.4392, -0.0656, -0.01, 0.4392, -0.0631, -0.0061, 0.4414, -0.067, -0.0083, 0.4414, -0.067, -0.0061, 0.4436, -0.067, -0.0083, 0.4436, -0.067, -0.0061, 0.4414, -0.0631, -0.0083, 0.4414, -0.0631, -0.0061, 0.4436, -0.0631, -0.0083, 0.4436, -0.0631, -0.01, 0.4414, -0.067, -0.01, 0.4436, -0.067, -0.01, 0.4414, -0.0656, -0.01, 0.4414, -0.0631, -0.01, 0.4436, -0.0656,
-0.01, 0.4436, -0.0631, -0.0061, 0.4382, -0.0562, -0.0061, 0.4382, -0.0548, -0.0083, 0.4382, -0.0562, -0.0083, 0.4382, -0.0548, -0.01, 0.4382, -0.0562, -0.01, 0.4382, -0.0548, -0.0061, 0.4414, -0.0562, -0.0083, 0.4414, -0.0562, -0.0061, 0.4436, -0.0562, -0.0083, 0.4436, -0.0562, -0.01, 0.4414, -0.0562, -0.01, 0.4414, -0.0548, -0.01, 0.4436, -0.0562, -0.01, 0.4436, -0.0548, -0.0061, 0.4458, -0.067, -0.0083, 0.4458, -0.067, -0.0061, 0.4479, -0.067, -0.0083, 0.4479, -0.067, -0.0061, 0.4458, -0.0631, -0.0083, 0.4458, -0.0631, -0.0061, 0.4479, -0.0631, -0.0083, 0.4479, -0.0631, -0.01, 0.4458, -0.067, -0.01, 0.4479, -0.067, -0.01, 0.4458, -0.0656, -0.01, 0.4458, -0.0631, -0.01, 0.4479, -0.0656, -0.01, 0.4479, -0.0631, -0.0061, 0.4501, -0.067,
-0.0083, 0.4501, -0.067, -0.0061, 0.4529, -0.067, -0.0083, 0.4529, -0.067, -0.0061, 0.4501, -0.0631, -0.0083, 0.4501, -0.0631, -0.0061, 0.4529, -0.0656, -0.0061, 0.4529, -0.0631, -0.0083, 0.4529, -0.0656, -0.0083, 0.4529, -0.0631, -0.01, 0.4501, -0.067, -0.01, 0.4529, -0.067, -0.01, 0.4501, -0.0656, -0.01, 0.4501, -0.0631, -0.01, 0.4529, -0.0656, -0.01, 0.4529, -0.0631, -0.0061, 0.4458, -0.0562, -0.0083, 0.4458, -0.0562, -0.0061, 0.4479, -0.0562, -0.0083, 0.4479, -0.0562, -0.01, 0.4458, -0.0562, -0.01, 0.4458, -0.0548, -0.01, 0.4479, -0.0562, -0.01, 0.4479, -0.0548, -0.0061, 0.4501, -0.0562, -0.0083, 0.4501, -0.0562, -0.0061, 0.452, -0.0562, -0.0061, 0.452, -0.0548, -0.0083, 0.452, -0.0562, -0.0083, 0.452, -0.0548, -0.01, 0.4501, -0.0562,
-0.01, 0.4501, -0.0548, -0.01, 0.452, -0.0562, -0.01, 0.452, -0.0548, -0.0061, 0.4382, -0.0523, -0.0083, 0.4382, -0.0523, -0.0061, 0.4362, -0.0478, -0.0061, 0.4372, -0.0466, -0.0083, 0.4362, -0.0478, -0.0083, 0.4372, -0.0466, -0.0061, 0.4396, -0.0456, -0.0083, 0.4396, -0.0456, -0.01, 0.4382, -0.0523, -0.01, 0.4363, -0.0477, -0.01, 0.4372, -0.0466, -0.01, 0.4393, -0.0457, -0.0061, 0.4414, -0.0523, -0.0083, 0.4414, -0.0523, -0.0061, 0.4436, -0.0523, -0.0083, 0.4436, -0.0523, -0.0061, 0.4414, -0.0455, -0.0083, 0.4414, -0.0455, -0.0061, 0.4436, -0.0454, -0.0083, 0.4436, -0.0454, -0.01, 0.4414, -0.0523, -0.01, 0.4436, -0.0523, -0.01, 0.4414, -0.0455, -0.01, 0.4436, -0.0454, -0.0061, 0.4371, -0.0425, -0.0083, 0.4371, -0.0425, -0.0061, 0.4392, -0.0425,
-0.0083, 0.4392, -0.0425, -0.01, 0.4371, -0.0439, -0.01, 0.4371, -0.0425, -0.01, 0.4392, -0.0439, -0.01, 0.4392, -0.0425, -0.0061, 0.4414, -0.0425, -0.0083, 0.4414, -0.0425, -0.0061, 0.4436, -0.0425, -0.0083, 0.4436, -0.0425, -0.01, 0.4414, -0.0439, -0.01, 0.4414, -0.0425, -0.01, 0.4436, -0.0439, -0.01, 0.4436, -0.0425, -0.0061, 0.4458, -0.0523, -0.0083, 0.4458, -0.0523, -0.0061, 0.4479, -0.0523, -0.0083, 0.4479, -0.0523, -0.0061, 0.4458, -0.0454, -0.0083, 0.4458, -0.0454, -0.0061, 0.4479, -0.0454, -0.0083, 0.4479, -0.0454, -0.01, 0.4458, -0.0523, -0.01, 0.4479, -0.0523, -0.01, 0.4458, -0.0454, -0.01, 0.4479, -0.0454, -0.0061, 0.4501, -0.0523, -0.0083, 0.4501, -0.0523, -0.0061, 0.452, -0.0523, -0.0083, 0.452, -0.0523, -0.0061, 0.4501, -0.0454,
-0.0083, 0.4501, -0.0454, -0.0061, 0.4529, -0.0454, -0.0083, 0.4529, -0.0454, -0.01, 0.4501, -0.0523, -0.01, 0.452, -0.0523, -0.01, 0.4501, -0.0454, -0.01, 0.4529, -0.0454, -0.0061, 0.4458, -0.0425, -0.0083, 0.4458, -0.0425, -0.0061, 0.4479, -0.0425, -0.0083, 0.4479, -0.0425, -0.01, 0.4458, -0.0439, -0.01, 0.4458, -0.0425, -0.01, 0.4479, -0.0439, -0.01, 0.4479, -0.0425, -0.0061, 0.4501, -0.0425, -0.0083, 0.4501, -0.0425, -0.0061, 0.4529, -0.0439, -0.0061, 0.4529, -0.0425, -0.0083, 0.4529, -0.0439, -0.0083, 0.4529, -0.0425, -0.01, 0.4501, -0.0439, -0.01, 0.4501, -0.0425, -0.01, 0.4529, -0.0439, -0.01, 0.4529, -0.0425, -0.0061, 0.4245, -0.0219, -0.0061, 0.4245, -0.0201, -0.0083, 0.4245, -0.0219, -0.0083, 0.4245, -0.0201, -0.0061, 0.4262, -0.0219,
-0.0083, 0.4262, -0.0219, -0.01, 0.4245, -0.0219, -0.01, 0.4245, -0.0201, -0.01, 0.4262, -0.0219, -0.01, 0.4262, -0.0201, -0.0061, 0.4284, -0.0219, -0.0083, 0.4284, -0.0219, -0.0061, 0.4306, -0.0219, -0.0083, 0.4306, -0.0219, -0.01, 0.4284, -0.0219, -0.01, 0.4284, -0.0201, -0.01, 0.4306, -0.0219, -0.01, 0.4306, -0.0201, -0.0061, 0.4327, -0.0219, -0.0083, 0.4327, -0.0219, -0.0061, 0.4349, -0.0219, -0.0083, 0.4349, -0.0219, -0.01, 0.4327, -0.0219, -0.01, 0.4327, -0.0201, -0.01, 0.4349, -0.0219, -0.01, 0.4349, -0.0201, -0.0061, 0.4245, -0.0179, -0.0061, 0.4245, -0.0157, -0.0083, 0.4245, -0.0179, -0.0083, 0.4245, -0.0157, -0.0061, 0.4245, -0.0136, -0.0061, 0.4245, -0.0114, -0.0083, 0.4245, -0.0136, -0.0083, 0.4245, -0.0114, -0.01, 0.4245, -0.0179,
-0.01, 0.4245, -0.0157, -0.01, 0.4262, -0.0179, -0.01, 0.4262, -0.0157, -0.01, 0.4245, -0.0136, -0.01, 0.4245, -0.0114, -0.01, 0.4262, -0.0136, -0.01, 0.4262, -0.0114, -0.0061, 0.4245, -0.0101, -0.0083, 0.4245, -0.0101, -0.0061, 0.4262, -0.0101, -0.0083, 0.4262, -0.0101, -0.01, 0.4245, -0.0101, -0.01, 0.4262, -0.0101, -0.01, 0.4284, -0.0179, -0.01, 0.4284, -0.0157, -0.01, 0.4306, -0.0179, -0.01, 0.4306, -0.0157, -0.01, 0.4284, -0.0136, -0.01, 0.4284, -0.0114, -0.01, 0.4306, -0.0136, -0.01, 0.4306, -0.0114, -0.01, 0.4327, -0.0179, -0.01, 0.4327, -0.0157, -0.01, 0.4349, -0.0179, -0.01, 0.4349, -0.0157, -0.01, 0.4327, -0.0136, -0.01, 0.4327, -0.0114, -0.01, 0.4349, -0.0136, -0.01, 0.4349, -0.0114, -0.0061, 0.4284, -0.0101,
-0.0083, 0.4284, -0.0101, -0.0061, 0.4306, -0.0101, -0.0083, 0.4306, -0.0101, -0.01, 0.4284, -0.0101, -0.01, 0.4306, -0.0101, -0.0061, 0.4327, -0.0101, -0.0083, 0.4327, -0.0101, -0.0061, 0.4349, -0.0101, -0.0083, 0.4349, -0.0101, -0.01, 0.4327, -0.0101, -0.01, 0.4349, -0.0101, -0.0061, 0.4371, -0.0219, -0.0083, 0.4371, -0.0219, -0.0061, 0.4392, -0.0219, -0.0083, 0.4392, -0.0219, -0.01, 0.4371, -0.0219, -0.01, 0.4371, -0.0201, -0.01, 0.4392, -0.0219, -0.01, 0.4392, -0.0201, -0.0061, 0.4414, -0.0219, -0.0083, 0.4414, -0.0219, -0.0061, 0.4436, -0.0219, -0.0083, 0.4436, -0.0219, -0.01, 0.4414, -0.0219, -0.01, 0.4414, -0.0201, -0.01, 0.4436, -0.0219, -0.01, 0.4436, -0.0201, -0.0061, 0.4458, -0.0219, -0.0083, 0.4458, -0.0219, -0.0061, 0.4479, -0.0219,
-0.0083, 0.4479, -0.0219, -0.01, 0.4458, -0.0219, -0.01, 0.4458, -0.0201, -0.01, 0.4479, -0.0219, -0.01, 0.4479, -0.0201, -0.0061, 0.4501, -0.0219, -0.0083, 0.4501, -0.0219, -0.0061, 0.4523, -0.0219, -0.0083, 0.4523, -0.0219, -0.01, 0.4501, -0.0219, -0.01, 0.4501, -0.0201, -0.01, 0.4523, -0.0219, -0.01, 0.4523, -0.0201, -0.01, 0.4371, -0.0179, -0.01, 0.4371, -0.0157, -0.01, 0.4392, -0.0179, -0.01, 0.4392, -0.0157, -0.01, 0.4371, -0.0136, -0.01, 0.4371, -0.0114, -0.01, 0.4392, -0.0136, -0.01, 0.4392, -0.0114, -0.01, 0.4414, -0.0179, -0.01, 0.4414, -0.0157, -0.01, 0.4436, -0.0179, -0.01, 0.4436, -0.0157, -0.01, 0.4414, -0.0136, -0.01, 0.4414, -0.0114, -0.01, 0.4436, -0.0136, -0.01, 0.4436, -0.0114, -0.0061, 0.4371, -0.0101,
-0.0083, 0.4371, -0.0101, -0.0061, 0.4392, -0.0101, -0.0083, 0.4392, -0.0101, -0.01, 0.4371, -0.0101, -0.01, 0.4392, -0.0101, -0.0061, 0.4414, -0.0101, -0.0083, 0.4414, -0.0101, -0.0061, 0.4436, -0.0101, -0.0083, 0.4436, -0.0101, -0.01, 0.4414, -0.0101, -0.01, 0.4436, -0.0101, -0.01, 0.4458, -0.0179, -0.01, 0.4458, -0.0157, -0.01, 0.4479, -0.0179, -0.01, 0.4479, -0.0157, -0.01, 0.4458, -0.0136, -0.01, 0.4458, -0.0114, -0.01, 0.4479, -0.0136, -0.01, 0.4479, -0.0114, -0.01, 0.4501, -0.0179, -0.01, 0.4501, -0.0157, -0.01, 0.4523, -0.0179, -0.01, 0.4523, -0.0157, -0.01, 0.4501, -0.0136, -0.01, 0.4501, -0.0114, -0.01, 0.4523, -0.0136, -0.01, 0.4523, -0.0114, -0.0061, 0.4458, -0.0101, -0.0083, 0.4458, -0.0101, -0.0061, 0.4479, -0.0101,
-0.0083, 0.4479, -0.0101, -0.01, 0.4458, -0.0101, -0.01, 0.4479, -0.0101, -0.0061, 0.4501, -0.0101, -0.0083, 0.4501, -0.0101, -0.0061, 0.4523, -0.0101, -0.0083, 0.4523, -0.0101, -0.01, 0.4501, -0.0101, -0.01, 0.4523, -0.0101, -0.0061, 0.4725, -0.067, -0.0083, 0.4725, -0.067, -0.0061, 0.474, -0.067, -0.0083, 0.474, -0.067, -0.0061, 0.4725, -0.0656, -0.0061, 0.4725, -0.0635, -0.0083, 0.4725, -0.0656, -0.0083, 0.4725, -0.0635, -0.01, 0.4725, -0.067, -0.01, 0.474, -0.067, -0.01, 0.4725, -0.0656, -0.01, 0.4725, -0.0635, -0.01, 0.474, -0.0656, -0.01, 0.474, -0.0635, -0.0061, 0.4765, -0.067, -0.0083, 0.4765, -0.067, -0.0061, 0.4765, -0.0656, -0.0061, 0.4765, -0.0635, -0.0083, 0.4765, -0.0656, -0.0083, 0.4765, -0.0635, -0.01, 0.4765, -0.067,
-0.01, 0.4765, -0.0656, -0.01, 0.4765, -0.0635, -0.0061, 0.4725, -0.0613, -0.0061, 0.4725, -0.0591, -0.0083, 0.4725, -0.0613, -0.0083, 0.4725, -0.0591, -0.0061, 0.4725, -0.057, -0.0061, 0.4725, -0.0548, -0.0083, 0.4725, -0.057, -0.0083, 0.4725, -0.0548, -0.01, 0.4725, -0.0613, -0.01, 0.4725, -0.0591, -0.01, 0.474, -0.0613, -0.01, 0.474, -0.0591, -0.01, 0.4725, -0.057, -0.01, 0.4725, -0.0548, -0.01, 0.474, -0.057, -0.01, 0.474, -0.0548, -0.0061, 0.4765, -0.0613, -0.0061, 0.4765, -0.0591, -0.0083, 0.4765, -0.0613, -0.0083, 0.4765, -0.0591, -0.0061, 0.4765, -0.057, -0.0061, 0.4765, -0.0548, -0.0083, 0.4765, -0.057, -0.0083, 0.4765, -0.0548, -0.01, 0.4765, -0.0613, -0.01, 0.4765, -0.0591, -0.01, 0.4765, -0.057, -0.01, 0.4765, -0.0548,
-0.0061, 0.4804, -0.0591, -0.0083, 0.4804, -0.0591, -0.0061, 0.4826, -0.0591, -0.0083, 0.4826, -0.0591, -0.0061, 0.4804, -0.0562, -0.0083, 0.4804, -0.0562, -0.0061, 0.4826, -0.0562, -0.0083, 0.4826, -0.0562, -0.01, 0.4804, -0.0591, -0.01, 0.4826, -0.0591, -0.01, 0.4804, -0.0562, -0.01, 0.4826, -0.0562, -0.0061, 0.4848, -0.0592, -0.0083, 0.4848, -0.0592, -0.0061, 0.487, -0.0596, -0.0083, 0.487, -0.0596, -0.0061, 0.4848, -0.0562, -0.0083, 0.4848, -0.0562, -0.0061, 0.487, -0.0562, -0.0083, 0.487, -0.0562, -0.01, 0.4848, -0.0592, -0.01, 0.4868, -0.0596, -0.01, 0.4848, -0.0562, -0.01, 0.487, -0.0562, -0.0061, 0.4725, -0.0526, -0.0061, 0.4725, -0.0505, -0.0083, 0.4725, -0.0526, -0.0083, 0.4725, -0.0505, -0.0061, 0.4725, -0.0483, -0.0061, 0.4725, -0.0461,
-0.0083, 0.4725, -0.0483, -0.0083, 0.4725, -0.0461, -0.01, 0.4725, -0.0526, -0.01, 0.4725, -0.0505, -0.01, 0.474, -0.0526, -0.01, 0.474, -0.0505, -0.01, 0.4725, -0.0483, -0.01, 0.4725, -0.0461, -0.01, 0.474, -0.0483, -0.01, 0.474, -0.0461, -0.0061, 0.4765, -0.0526, -0.0061, 0.4766, -0.0504, -0.0083, 0.4765, -0.0526, -0.0083, 0.4766, -0.0504, -0.0061, 0.4769, -0.0487, -0.0083, 0.4769, -0.0487, -0.0061, 0.4775, -0.0476, -0.0061, 0.478, -0.047, -0.0083, 0.4775, -0.0476, -0.0083, 0.478, -0.047, -0.01, 0.4765, -0.0527, -0.01, 0.4766, -0.0506, -0.01, 0.477, -0.0485, -0.01, 0.4761, -0.0461, -0.01, 0.4775, -0.0476, -0.01, 0.4786, -0.0465, -0.0061, 0.4725, -0.0439, -0.0061, 0.4725, -0.0425, -0.0083, 0.4725, -0.0439, -0.0083, 0.4725, -0.0425,
-0.0061, 0.474, -0.0425, -0.0083, 0.474, -0.0425, -0.01, 0.4725, -0.0439, -0.01, 0.4725, -0.0425, -0.01, 0.474, -0.0439, -0.01, 0.474, -0.0425, -0.0061, 0.4761, -0.0425, -0.0083, 0.4761, -0.0425, -0.0061, 0.4783, -0.0425, -0.0083, 0.4783, -0.0425, -0.01, 0.4761, -0.0439, -0.01, 0.4761, -0.0425, -0.01, 0.4783, -0.0439, -0.01, 0.4783, -0.0425, -0.0061, 0.4801, -0.0456, -0.0083, 0.4801, -0.0456, -0.0061, 0.4826, -0.0455, -0.0083, 0.4826, -0.0455, -0.01, 0.4801, -0.0456, -0.01, 0.4827, -0.0455, -0.0061, 0.4848, -0.0454, -0.0083, 0.4848, -0.0454, -0.0061, 0.487, -0.0454, -0.0083, 0.487, -0.0454, -0.01, 0.4848, -0.0454, -0.01, 0.487, -0.0454, -0.0061, 0.4805, -0.0425, -0.0083, 0.4805, -0.0425, -0.0061, 0.4826, -0.0425, -0.0083, 0.4826, -0.0425,
-0.01, 0.4805, -0.0439, -0.01, 0.4805, -0.0425, -0.01, 0.4826, -0.0439, -0.01, 0.4826, -0.0425, -0.0061, 0.4848, -0.0425, -0.0083, 0.4848, -0.0425, -0.0061, 0.487, -0.0425, -0.0083, 0.487, -0.0425, -0.01, 0.4848, -0.0439, -0.01, 0.4848, -0.0425, -0.01, 0.487, -0.0439, -0.01, 0.487, -0.0425, -0.0061, 0.4544, -0.0219, -0.0083, 0.4544, -0.0219, -0.0061, 0.4566, -0.0219, -0.0083, 0.4566, -0.0219, -0.01, 0.4544, -0.0219, -0.01, 0.4544, -0.0201, -0.01, 0.4566, -0.0219, -0.01, 0.4566, -0.0201, -0.0061, 0.4588, -0.0219, -0.0083, 0.4588, -0.0219, -0.0061, 0.4609, -0.0219, -0.0083, 0.4609, -0.0219, -0.01, 0.4588, -0.0219, -0.01, 0.4588, -0.0201, -0.01, 0.4609, -0.0219, -0.01, 0.4609, -0.0201, -0.0061, 0.4631, -0.0219, -0.0083, 0.4631, -0.0219,
-0.0061, 0.4653, -0.0219, -0.0083, 0.4653, -0.0219, -0.01, 0.4631, -0.0219, -0.01, 0.4631, -0.0201, -0.01, 0.4653, -0.0219, -0.01, 0.4653, -0.0201, -0.0061, 0.4675, -0.0219, -0.0083, 0.4675, -0.0219, -0.0061, 0.4696, -0.0219, -0.0083, 0.4696, -0.0219, -0.01, 0.4675, -0.0219, -0.01, 0.4675, -0.0201, -0.01, 0.4696, -0.0219, -0.01, 0.4696, -0.0201, -0.01, 0.4544, -0.0179, -0.01, 0.4544, -0.0157, -0.01, 0.4566, -0.0179, -0.01, 0.4566, -0.0157, -0.01, 0.4544, -0.0136, -0.01, 0.4544, -0.0114, -0.01, 0.4566, -0.0136, -0.01, 0.4566, -0.0114, -0.01, 0.4588, -0.0179, -0.01, 0.4588, -0.0157, -0.01, 0.4609, -0.0179, -0.01, 0.4609, -0.0157, -0.01, 0.4588, -0.0136, -0.01, 0.4588, -0.0114, -0.01, 0.4609, -0.0136, -0.01, 0.4609, -0.0114,
-0.0061, 0.4544, -0.0101, -0.0083, 0.4544, -0.0101, -0.0061, 0.4566, -0.0101, -0.0083, 0.4566, -0.0101, -0.01, 0.4544, -0.0101, -0.01, 0.4566, -0.0101, -0.0061, 0.4588, -0.0101, -0.0083, 0.4588, -0.0101, -0.0061, 0.4609, -0.0101, -0.0083, 0.4609, -0.0101, -0.01, 0.4588, -0.0101, -0.01, 0.4609, -0.0101, -0.01, 0.4631, -0.0179, -0.01, 0.4631, -0.0157, -0.01, 0.4653, -0.0179, -0.01, 0.4653, -0.0157, -0.01, 0.4631, -0.0136, -0.01, 0.4631, -0.0114, -0.01, 0.4653, -0.0136, -0.01, 0.4653, -0.0114, -0.01, 0.4675, -0.0179, -0.01, 0.4675, -0.0157, -0.01, 0.4696, -0.0179, -0.01, 0.4696, -0.0157, -0.01, 0.4675, -0.0136, -0.01, 0.4675, -0.0114, -0.01, 0.4696, -0.0136, -0.01, 0.4696, -0.0114, -0.0061, 0.4631, -0.0101, -0.0083, 0.4631, -0.0101,
-0.0061, 0.4653, -0.0101, -0.0083, 0.4653, -0.0101, -0.01, 0.4631, -0.0101, -0.01, 0.4653, -0.0101, -0.0061, 0.4675, -0.0101, -0.0083, 0.4675, -0.0101, -0.0061, 0.4696, -0.0101, -0.0083, 0.4696, -0.0101, -0.01, 0.4675, -0.0101, -0.01, 0.4696, -0.0101, -0.0061, 0.4718, -0.0219, -0.0083, 0.4718, -0.0219, -0.0061, 0.474, -0.0219, -0.0083, 0.474, -0.0219, -0.01, 0.4718, -0.0219, -0.01, 0.4718, -0.0201, -0.01, 0.474, -0.0219, -0.01, 0.474, -0.0201, -0.0061, 0.4761, -0.0219, -0.0083, 0.4761, -0.0219, -0.0061, 0.4783, -0.0218, -0.0083, 0.4783, -0.0218, -0.01, 0.4761, -0.0219, -0.01, 0.4761, -0.0201, -0.01, 0.4783, -0.0218, -0.01, 0.4783, -0.0201, -0.0061, 0.4805, -0.0216, -0.0083, 0.4805, -0.0216, -0.0061, 0.4823, -0.0213, -0.0061, 0.4834, -0.0211,
-0.0083, 0.4823, -0.0213, -0.0083, 0.4834, -0.0211, -0.01, 0.4805, -0.0216, -0.01, 0.4805, -0.0201, -0.01, 0.4821, -0.0213, -0.01, 0.4829, -0.0212, -0.0061, 0.4848, -0.0208, -0.0083, 0.4848, -0.0208, -0.0061, 0.487, -0.0203, -0.0083, 0.487, -0.0203, -0.01, 0.485, -0.0207, -0.01, 0.4872, -0.0202, -0.01, 0.4718, -0.0179, -0.01, 0.4718, -0.0157, -0.01, 0.474, -0.0179, -0.01, 0.474, -0.0157, -0.01, 0.4718, -0.0136, -0.01, 0.4718, -0.0114, -0.01, 0.474, -0.0136, -0.01, 0.474, -0.0114, -0.01, 0.4761, -0.0179, -0.01, 0.4761, -0.0157, -0.01, 0.4783, -0.0179, -0.01, 0.4783, -0.0157, -0.01, 0.4761, -0.0136, -0.01, 0.4761, -0.0114, -0.01, 0.4783, -0.0136, -0.01, 0.4783, -0.0114, -0.0061, 0.4718, -0.0101, -0.0083, 0.4718, -0.0101,
-0.0061, 0.474, -0.0101, -0.0083, 0.474, -0.0101, -0.01, 0.4718, -0.0101, -0.01, 0.474, -0.0101, -0.0061, 0.4761, -0.0101, -0.0083, 0.4761, -0.0101, -0.0061, 0.4783, -0.01, -0.0083, 0.4783, -0.01, -0.01, 0.4761, -0.0101, -0.01, 0.4783, -0.01, -0.01, 0.4805, -0.0179, -0.01, 0.4805, -0.0157, -0.01, 0.4826, -0.0179, -0.01, 0.4826, -0.0157, -0.01, 0.4805, -0.0136, -0.01, 0.4805, -0.0114, -0.01, 0.4826, -0.0136, -0.01, 0.4826, -0.0114, -0.01, 0.4848, -0.0179, -0.01, 0.4848, -0.0157, -0.01, 0.487, -0.0179, -0.01, 0.487, -0.0157, -0.01, 0.4848, -0.0136, -0.01, 0.4848, -0.0114, -0.01, 0.487, -0.0136, -0.01, 0.487, -0.0114, -0.0061, 0.4805, -0.0096, -0.0083, 0.4805, -0.0096, -0.0061, 0.4827, -0.009, -0.0083, 0.4827, -0.009,
-0.01, 0.4804, -0.0096, -0.01, 0.4824, -0.0091, -0.0061, 0.4841, -0.0083, -0.0061, 0.4852, -0.0078, -0.0083, 0.4841, -0.0083, -0.0083, 0.4852, -0.0078, -0.0061, 0.4869, -0.0067, -0.0083, 0.4869, -0.0067, -0.0061, 0.4879, -0.0058, -0.0083, 0.4879, -0.0058, -0.01, 0.4843, -0.0082, -0.01, 0.4853, -0.0077, -0.01, 0.487, -0.0092, -0.01, 0.4869, -0.0068, -0.01, 0.4879, -0.0058, -0.0061, 0.4919, -0.0628, -0.0083, 0.4919, -0.0628, -0.01, 0.4919, -0.0629, -0.0061, 0.4943, -0.0648, -0.0061, 0.4932, -0.0639, -0.0083, 0.4943, -0.0648, -0.0083, 0.4932, -0.0639, -0.0061, 0.4956, -0.0658, -0.0083, 0.4956, -0.0658, -0.01, 0.4943, -0.0648, -0.01, 0.4932, -0.064, -0.01, 0.4954, -0.0657, -0.01, 0.4957, -0.0635, -0.0061, 0.4894, -0.0608, -0.0061, 0.4882, -0.0601,
-0.0083, 0.4894, -0.0608, -0.0083, 0.4882, -0.0601, -0.0061, 0.4908, -0.0619, -0.0083, 0.4908, -0.0619, -0.0061, 0.4892, -0.0562, -0.0083, 0.4892, -0.0562, -0.0061, 0.4909, -0.056, -0.0061, 0.492, -0.0558, -0.0083, 0.4909, -0.056, -0.0083, 0.492, -0.0558, -0.01, 0.4895, -0.0608, -0.01, 0.4886, -0.0603, -0.01, 0.4909, -0.0619, -0.01, 0.4913, -0.0591, -0.01, 0.4891, -0.0562, -0.01, 0.4912, -0.056, -0.01, 0.4921, -0.0558, -0.0061, 0.4963, -0.0608, -0.0061, 0.4953, -0.0594, -0.0083, 0.4963, -0.0608, -0.0083, 0.4953, -0.0594, -0.0061, 0.4935, -0.0553, -0.0083, 0.4935, -0.0553, -0.0061, 0.4953, -0.0543, -0.0083, 0.4953, -0.0543, -0.01, 0.4935, -0.0613, -0.01, 0.4935, -0.0591, -0.01, 0.4963, -0.0609, -0.01, 0.4953, -0.0594, -0.01, 0.4935, -0.057,
-0.01, 0.4934, -0.0553, -0.01, 0.4957, -0.057, -0.01, 0.4952, -0.0543, -0.0061, 0.4986, -0.0668, -0.0083, 0.4986, -0.0668, -0.0061, 0.5, -0.0669, -0.0083, 0.5, -0.0669, -0.0061, 0.4975, -0.0665, -0.0061, 0.4983, -0.063, -0.0083, 0.4975, -0.0665, -0.0083, 0.4983, -0.063, -0.0061, 0.5005, -0.0652, -0.0061, 0.4994, -0.0641, -0.0083, 0.5005, -0.0652, -0.0083, 0.4994, -0.0641, -0.01, 0.4987, -0.0668, -0.01, 0.5, -0.0669, -0.01, 0.4977, -0.0666, -0.01, 0.4983, -0.063, -0.01, 0.5005, -0.0652, -0.01, 0.4994, -0.0641, -0.0061, 0.502, -0.067, -0.0083, 0.502, -0.067, -0.0061, 0.5014, -0.0663, -0.0083, 0.5014, -0.0663, -0.01, 0.502, -0.067, -0.01, 0.5014, -0.0664, -0.0061, 0.4972, -0.0619, -0.0061, 0.497, -0.0582, -0.0083, 0.4972, -0.0619,
-0.0083, 0.497, -0.0582, -0.0061, 0.4981, -0.0574, -0.0083, 0.4981, -0.0574, -0.0061, 0.4994, -0.0563, -0.0061, 0.5006, -0.0549, -0.0083, 0.4994, -0.0563, -0.0083, 0.5006, -0.0549, -0.01, 0.4972, -0.0619, -0.01, 0.497, -0.0582, -0.01, 0.4981, -0.0574, -0.01, 0.4978, -0.0548, -0.01, 0.4994, -0.0563, -0.01, 0.5006, -0.0549, -0.0061, 0.5011, -0.0538, -0.0083, 0.5011, -0.0538, -0.01, 0.5011, -0.0537, -0.0061, 0.4892, -0.0455, -0.0083, 0.4892, -0.0455, -0.0061, 0.4914, -0.046, -0.0083, 0.4914, -0.046, -0.01, 0.4891, -0.0455, -0.01, 0.4912, -0.0459, -0.0061, 0.4962, -0.053, -0.0061, 0.497, -0.0503, -0.0083, 0.4962, -0.053, -0.0083, 0.497, -0.0503, -0.0061, 0.4943, -0.0474, -0.0061, 0.4932, -0.0468, -0.0083, 0.4943, -0.0474, -0.0083, 0.4932, -0.0468,
-0.0061, 0.4955, -0.0484, -0.0083, 0.4955, -0.0484, -0.01, 0.4964, -0.0526, -0.01, 0.497, -0.0503, -0.01, 0.4943, -0.0474, -0.01, 0.4933, -0.0468, -0.01, 0.4955, -0.0484, -0.01, 0.4957, -0.0461, -0.0061, 0.4892, -0.0425, -0.0083, 0.4892, -0.0425, -0.0061, 0.4913, -0.0425, -0.0083, 0.4913, -0.0425, -0.01, 0.4892, -0.0439, -0.01, 0.4892, -0.0425, -0.01, 0.4913, -0.0439, -0.01, 0.4913, -0.0425, -0.0061, 0.4935, -0.0426, -0.0083, 0.4935, -0.0426, -0.0061, 0.4958, -0.0431, -0.0061, 0.4947, -0.0428, -0.0083, 0.4958, -0.0431, -0.0083, 0.4947, -0.0428, -0.01, 0.4935, -0.0439, -0.01, 0.4935, -0.0426, -0.01, 0.4958, -0.0431, -0.01, 0.4946, -0.0428, -0.0061, 0.501, -0.0475, -0.0061, 0.5, -0.046, -0.0083, 0.501, -0.0475, -0.0083, 0.5, -0.046,
-0.01, 0.4978, -0.0526, -0.01, 0.4978, -0.0505, -0.01, 0.5, -0.0526, -0.01, 0.5, -0.0505, -0.01, 0.4978, -0.0483, -0.01, 0.4978, -0.0461, -0.01, 0.5011, -0.0477, -0.01, 0.4998, -0.0458, -0.0061, 0.5015, -0.0527, -0.0061, 0.5018, -0.0505, -0.0083, 0.5015, -0.0527, -0.0083, 0.5018, -0.0505, -0.0061, 0.5015, -0.0486, -0.0083, 0.5015, -0.0486, -0.01, 0.5015, -0.0527, -0.01, 0.5018, -0.0505, -0.01, 0.5015, -0.0488, -0.0061, 0.4979, -0.0441, -0.0083, 0.4979, -0.0441, -0.0061, 0.499, -0.045, -0.0083, 0.499, -0.045, -0.01, 0.4981, -0.0442, -0.01, 0.499, -0.045, -0.0061, 0.4892, -0.0195, -0.0083, 0.4892, -0.0195, -0.01, 0.4889, -0.0196, -0.0061, 0.4902, -0.019, -0.0083, 0.4902, -0.019, -0.0061, 0.4914, -0.0185, -0.0083, 0.4914, -0.0185,
-0.01, 0.4898, -0.0192, -0.01, 0.4892, -0.0157, -0.01, 0.4917, -0.0183, -0.01, 0.4913, -0.0157, -0.01, 0.4892, -0.0136, -0.01, 0.4892, -0.0114, -0.01, 0.4913, -0.0136, -0.01, 0.4913, -0.0114, -0.0061, 0.4932, -0.0174, -0.0061, 0.4943, -0.0166, -0.0083, 0.4932, -0.0174, -0.0083, 0.4943, -0.0166, -0.0061, 0.4956, -0.0155, -0.0083, 0.4956, -0.0155, -0.0061, 0.4966, -0.0145, -0.0083, 0.4966, -0.0145, -0.01, 0.4931, -0.0174, -0.01, 0.4942, -0.0166, -0.01, 0.4955, -0.0156, -0.01, 0.4935, -0.0136, -0.01, 0.4935, -0.0114, -0.01, 0.4966, -0.0146, -0.01, 0.4957, -0.0114, -0.0061, 0.489, -0.0048, -0.0061, 0.49, -0.0034, -0.0083, 0.489, -0.0048, -0.0083, 0.49, -0.0034, -0.0061, 0.4905, -0.0023, -0.0083, 0.4905, -0.0023, -0.01, 0.4892, -0.0092,
-0.01, 0.4892, -0.0071, -0.01, 0.4913, -0.0092, -0.01, 0.4913, -0.0071, -0.01, 0.489, -0.0048, -0.01, 0.4899, -0.0034, -0.01, 0.4913, -0.0049, -0.01, 0.4904, -0.0025, -0.01, 0.4935, -0.0092, -0.01, 0.4935, -0.0071, -0.01, 0.4957, -0.0092, -0.01, 0.4957, -0.0071, -0.01, 0.4935, -0.0049, -0.01, 0.4935, -0.0027, -0.01, 0.4957, -0.0049, -0.01, 0.4957, -0.0027, -0.0061, 0.4977, -0.0135, -0.0061, 0.4987, -0.0122, -0.0083, 0.4977, -0.0135, -0.0083, 0.4987, -0.0122, -0.0061, 0.4994, -0.0111, -0.0083, 0.4994, -0.0111, -0.01, 0.4977, -0.0135, -0.01, 0.4988, -0.0121, -0.01, 0.4995, -0.011, -0.0061, 0.5005, -0.0093, -0.0061, 0.501, -0.008, -0.0083, 0.5005, -0.0093, -0.0083, 0.501, -0.008, -0.01, 0.4978, -0.0092, -0.01, 0.4978, -0.0071,
-0.01, 0.5004, -0.0096, -0.01, 0.5012, -0.0076, -0.01, 0.4978, -0.0049, -0.01, 0.4978, -0.0027, -0.01, 0.5, -0.0049, -0.01, 0.5, -0.0027, -0.0061, 0.5014, -0.0069, -0.0083, 0.5014, -0.0069, -0.0061, 0.5018, -0.0049, -0.0061, 0.5019, -0.0027, -0.0083, 0.5018, -0.0049, -0.0083, 0.5019, -0.0027, -0.01, 0.5015, -0.0067, -0.01, 0.5018, -0.0049, -0.01, 0.5019, -0.0027, -0.0061, 0.003, -0.0013, -0.0083, 0.003, -0.0013, -0.0061, 0.0052, -0.0012, -0.0083, 0.0052, -0.0012, -0.01, 0.0031, -0.0013, -0.01, 0.003, 0.0016, -0.01, 0.0053, -0.0012, -0.01, 0.0052, 0.0016, -0.01, 0.003, 0.0038, -0.01, 0.003, 0.006, -0.01, 0.0052, 0.0038, -0.01, 0.0052, 0.006, -0.0061, 0.0074, -0.0009, -0.0083, 0.0074, -0.0009, -0.0061, 0.0096, -0.0005,
-0.0083, 0.0096, -0.0005, -0.01, 0.0075, -0.0009, -0.01, 0.0074, 0.0016, -0.01, 0.0097, -0.0004, -0.01, 0.0096, 0.0016, -0.01, 0.0074, 0.0038, -0.01, 0.0074, 0.006, -0.01, 0.0096, 0.0038, -0.01, 0.0096, 0.006, -0.0061, 0.003, 0.0105, -0.0083, 0.003, 0.0105, -0.0061, 0.0052, 0.0106, -0.0083, 0.0052, 0.0106, -0.01, 0.003, 0.0081, -0.01, 0.003, 0.0105, -0.01, 0.0052, 0.0081, -0.01, 0.0051, 0.0105, -0.0061, 0.0074, 0.0109, -0.0083, 0.0074, 0.0109, -0.0061, 0.0085, 0.0114, -0.0083, 0.0085, 0.0114, -0.0061, 0.0096, 0.0118, -0.0083, 0.0096, 0.0118, -0.01, 0.0074, 0.0081, -0.01, 0.0072, 0.0109, -0.01, 0.0096, 0.0081, -0.01, 0.0089, 0.0115, -0.01, 0.0099, 0.0119, -0.0061, 0.0116, 0.0002, -0.0061, 0.0127, 0.0006,
-0.0083, 0.0116, 0.0002, -0.0083, 0.0127, 0.0006, -0.0061, 0.0139, 0.0011, -0.0083, 0.0139, 0.0011, -0.01, 0.0114, 0.0001, -0.01, 0.0123, 0.0004, -0.01, 0.0142, 0.0012, -0.01, 0.0117, 0.0038, -0.01, 0.0117, 0.006, -0.01, 0.0139, 0.0038, -0.01, 0.0139, 0.006, -0.0061, 0.0161, 0.0021, -0.0083, 0.0161, 0.0021, -0.0061, 0.0172, 0.0027, -0.0083, 0.0172, 0.0027, -0.0061, 0.0183, 0.0033, -0.0083, 0.0183, 0.0033, -0.01, 0.0162, 0.0022, -0.01, 0.0172, 0.0027, -0.01, 0.0161, 0.0038, -0.01, 0.0161, 0.006, -0.01, 0.0183, 0.0034, -0.01, 0.0182, 0.006, -0.0061, 0.0116, 0.0129, -0.0061, 0.0126, 0.0137, -0.0083, 0.0116, 0.0129, -0.0083, 0.0126, 0.0137, -0.0061, 0.0138, 0.0147, -0.0083, 0.0138, 0.0147, -0.01, 0.0117, 0.0081,
-0.01, 0.0117, 0.0103, -0.01, 0.0139, 0.0081, -0.01, 0.0139, 0.0103, -0.01, 0.0115, 0.0129, -0.01, 0.0126, 0.0137, -0.01, 0.0139, 0.0125, -0.01, 0.0138, 0.0147, -0.01, 0.0161, 0.0081, -0.01, 0.0161, 0.0103, -0.01, 0.0182, 0.0081, -0.01, 0.0182, 0.0103, -0.01, 0.0161, 0.0125, -0.01, 0.0161, 0.0146, -0.01, 0.0182, 0.0125, -0.01, 0.0182, 0.0146, -0.0061, 0.0148, 0.016, -0.0083, 0.0148, 0.016, -0.01, 0.0148, 0.0159, -0.0061, 0.0156, 0.017, -0.0061, 0.0167, 0.0189, -0.0083, 0.0156, 0.017, -0.0083, 0.0167, 0.0189, -0.0061, 0.0171, 0.0203, -0.0083, 0.0171, 0.0203, -0.0061, 0.0174, 0.0214, -0.0061, 0.0176, 0.0233, -0.0083, 0.0174, 0.0214, -0.0083, 0.0176, 0.0233, -0.01, 0.0156, 0.017, -0.01, 0.0167, 0.019,
-0.01, 0.0182, 0.0168, -0.01, 0.0182, 0.019, -0.01, 0.0171, 0.0203, -0.01, 0.0173, 0.0211, -0.01, 0.0176, 0.0233, -0.0061, 0.0144, 0.0324, -0.0083, 0.0144, 0.0324, -0.01, 0.0144, 0.0325, -0.0061, 0.017, 0.028, -0.0083, 0.017, 0.028, -0.0061, 0.0175, 0.0255, -0.0061, 0.0172, 0.0269, -0.0083, 0.0175, 0.0255, -0.0083, 0.0172, 0.0269, -0.0061, 0.0162, 0.0299, -0.0061, 0.0153, 0.0313, -0.0083, 0.0162, 0.0299, -0.0083, 0.0153, 0.0313, -0.01, 0.0169, 0.0282, -0.01, 0.0175, 0.0254, -0.01, 0.0171, 0.0274, -0.01, 0.0163, 0.0297, -0.01, 0.0152, 0.0315, -0.01, 0.0182, 0.0298, -0.01, 0.0182, 0.032, -0.0061, 0.0199, 0.0044, -0.0061, 0.021, 0.0053, -0.0083, 0.0199, 0.0044, -0.0083, 0.021, 0.0053, -0.0061, 0.0221, 0.0064,
-0.0083, 0.0221, 0.0064, -0.01, 0.0198, 0.0044, -0.01, 0.021, 0.0053, -0.01, 0.0221, 0.0064, -0.0061, 0.0232, 0.0076, -0.0083, 0.0232, 0.0076, -0.01, 0.0204, 0.0081, -0.01, 0.0204, 0.0103, -0.01, 0.0232, 0.0076, -0.01, 0.0226, 0.0103, -0.01, 0.0204, 0.0125, -0.01, 0.0204, 0.0146, -0.01, 0.0226, 0.0125, -0.01, 0.0226, 0.0146, -0.0061, 0.0241, 0.0087, -0.0061, 0.0252, 0.0102, -0.0083, 0.0241, 0.0087, -0.0083, 0.0252, 0.0102, -0.0061, 0.0259, 0.0113, -0.0083, 0.0259, 0.0113, -0.0061, 0.0265, 0.0124, -0.0061, 0.0274, 0.0146, -0.0083, 0.0265, 0.0124, -0.0083, 0.0274, 0.0146, -0.01, 0.0241, 0.0087, -0.01, 0.0252, 0.0102, -0.01, 0.0259, 0.0113, -0.01, 0.0247, 0.0125, -0.01, 0.0247, 0.0146, -0.01, 0.0265, 0.0123,
-0.01, 0.0273, 0.0144, -0.01, 0.0204, 0.0168, -0.01, 0.0204, 0.019, -0.01, 0.0226, 0.0168, -0.01, 0.0226, 0.019, -0.01, 0.0204, 0.0212, -0.01, 0.0204, 0.0233, -0.01, 0.0226, 0.0212, -0.01, 0.0226, 0.0233, -0.0061, 0.0279, 0.0163, -0.0083, 0.0279, 0.0163, -0.01, 0.0247, 0.0168, -0.01, 0.0247, 0.019, -0.01, 0.028, 0.0166, -0.01, 0.0269, 0.019, -0.01, 0.0247, 0.0212, -0.01, 0.0247, 0.0233, -0.01, 0.0269, 0.0212, -0.01, 0.0269, 0.0233, -0.01, 0.0204, 0.0255, -0.01, 0.0204, 0.0277, -0.01, 0.0226, 0.0255, -0.01, 0.0226, 0.0277, -0.01, 0.0204, 0.0298, -0.01, 0.0204, 0.032, -0.01, 0.0226, 0.0298, -0.01, 0.0226, 0.032, -0.0061, 0.028, 0.0308, -0.0061, 0.0277, 0.032, -0.0083, 0.028, 0.0308,
-0.0083, 0.0277, 0.032, -0.01, 0.0247, 0.0255, -0.01, 0.0247, 0.0277, -0.01, 0.0269, 0.0255, -0.01, 0.0269, 0.0277, -0.01, 0.0247, 0.0298, -0.01, 0.0247, 0.032, -0.01, 0.0281, 0.0302, -0.01, 0.0276, 0.0323, -0.0061, 0.0281, 0.0174, -0.0061, 0.0283, 0.019, -0.0083, 0.0281, 0.0174, -0.0083, 0.0283, 0.019, -0.0061, 0.0284, 0.0212, -0.0061, 0.0284, 0.0233, -0.0083, 0.0284, 0.0212, -0.0083, 0.0284, 0.0233, -0.01, 0.0281, 0.0175, -0.01, 0.0283, 0.019, -0.01, 0.0284, 0.0212, -0.01, 0.0284, 0.0233, -0.0061, 0.0284, 0.0255, -0.0061, 0.0284, 0.0277, -0.0083, 0.0284, 0.0255, -0.0083, 0.0284, 0.0277, -0.0061, 0.0282, 0.0297, -0.0083, 0.0282, 0.0297, -0.01, 0.0284, 0.0255, -0.01, 0.0284, 0.0277, -0.01, 0.0282, 0.0294,
-0.0061, 0.006, 0.0374, -0.0083, 0.006, 0.0374, -0.0061, 0.003, 0.0378, -0.0083, 0.003, 0.0378, -0.0061, 0.0049, 0.0376, -0.0083, 0.0049, 0.0376, -0.01, 0.0061, 0.0373, -0.01, 0.003, 0.0378, -0.01, 0.003, 0.0407, -0.01, 0.0052, 0.0376, -0.01, 0.0052, 0.0407, -0.0061, 0.0074, 0.037, -0.0083, 0.0074, 0.037, -0.0061, 0.0096, 0.0361, -0.0083, 0.0096, 0.0361, -0.01, 0.0073, 0.037, -0.01, 0.0093, 0.0362, -0.01, 0.0074, 0.0385, -0.01, 0.0074, 0.0407, -0.01, 0.0096, 0.0385, -0.01, 0.0096, 0.0407, -0.0061, 0.003, 0.0487, -0.0083, 0.003, 0.0487, -0.0061, 0.0052, 0.0487, -0.0083, 0.0052, 0.0487, -0.01, 0.003, 0.0429, -0.01, 0.003, 0.045, -0.01, 0.0052, 0.0429, -0.01, 0.0052, 0.045, -0.01, 0.003, 0.0472,
-0.01, 0.003, 0.0487, -0.01, 0.0052, 0.0472, -0.01, 0.0052, 0.0487, -0.0061, 0.0084, 0.0483, -0.0061, 0.0073, 0.0485, -0.0083, 0.0084, 0.0483, -0.0083, 0.0073, 0.0485, -0.0061, 0.0096, 0.048, -0.0083, 0.0096, 0.048, -0.01, 0.0074, 0.0429, -0.01, 0.0074, 0.045, -0.01, 0.0096, 0.0429, -0.01, 0.0096, 0.045, -0.01, 0.0078, 0.0484, -0.01, 0.007, 0.0485, -0.01, 0.0098, 0.048, -0.0061, 0.012, 0.0347, -0.0061, 0.0108, 0.0354, -0.0083, 0.012, 0.0347, -0.0083, 0.0108, 0.0354, -0.0061, 0.0133, 0.0335, -0.0083, 0.0133, 0.0335, -0.01, 0.012, 0.0346, -0.01, 0.011, 0.0352, -0.01, 0.0133, 0.0335, -0.01, 0.0139, 0.0363, -0.01, 0.0117, 0.0385, -0.01, 0.0117, 0.0407, -0.01, 0.0139, 0.0385, -0.01, 0.0139, 0.0407,
-0.01, 0.0161, 0.0342, -0.01, 0.0161, 0.0363, -0.01, 0.0182, 0.0342, -0.01, 0.0182, 0.0363, -0.01, 0.0161, 0.0385, -0.01, 0.0161, 0.0407, -0.01, 0.0182, 0.0385, -0.01, 0.0182, 0.0407, -0.0061, 0.0149, 0.0461, -0.0083, 0.0149, 0.0461, -0.0061, 0.0117, 0.0474, -0.0083, 0.0117, 0.0474, -0.0061, 0.0139, 0.0466, -0.0083, 0.0139, 0.0466, -0.01, 0.0117, 0.0429, -0.01, 0.0117, 0.045, -0.01, 0.0139, 0.0429, -0.01, 0.0146, 0.0463, -0.01, 0.0119, 0.0473, -0.01, 0.0136, 0.0467, -0.0061, 0.0161, 0.0455, -0.0083, 0.0161, 0.0455, -0.0061, 0.019, 0.0437, -0.0061, 0.0179, 0.0445, -0.0083, 0.019, 0.0437, -0.0083, 0.0179, 0.0445, -0.01, 0.0161, 0.0429, -0.01, 0.0164, 0.0454, -0.01, 0.0189, 0.0438, -0.01, 0.0178, 0.0445,
-0.0061, 0.0214, 0.0417, -0.0083, 0.0214, 0.0417, -0.0061, 0.0235, 0.0394, -0.0061, 0.0225, 0.0407, -0.0083, 0.0235, 0.0394, -0.0083, 0.0225, 0.0407, -0.01, 0.0204, 0.0342, -0.01, 0.0204, 0.0363, -0.01, 0.0226, 0.0342, -0.01, 0.0226, 0.0363, -0.01, 0.0204, 0.0385, -0.01, 0.0214, 0.0417, -0.01, 0.0236, 0.0394, -0.01, 0.0225, 0.0406, -0.0061, 0.0254, 0.0368, -0.0083, 0.0254, 0.0368, -0.0061, 0.0269, 0.0342, -0.0061, 0.0261, 0.0357, -0.0083, 0.0269, 0.0342, -0.0083, 0.0261, 0.0357, -0.0061, 0.0244, 0.0384, -0.0083, 0.0244, 0.0384, -0.01, 0.0247, 0.0342, -0.01, 0.0255, 0.0367, -0.01, 0.0268, 0.0344, -0.01, 0.0261, 0.0356, -0.01, 0.0244, 0.0383, -0.0061, 0.0204, 0.0427, -0.0083, 0.0204, 0.0427, -0.01, 0.0203, 0.0427,
-0.0061, 0.0716, -0.0005, -0.0061, 0.0716, 0.0016, -0.0083, 0.0716, -0.0005, -0.0083, 0.0716, 0.0016, -0.0061, 0.0716, 0.0038, -0.0061, 0.0716, 0.006, -0.0083, 0.0716, 0.0038, -0.0083, 0.0716, 0.006, -0.01, 0.0716, -0.0005, -0.01, 0.0716, 0.0016, -0.01, 0.0747, -0.0005, -0.01, 0.0747, 0.0016, -0.01, 0.0716, 0.0038, -0.01, 0.0716, 0.006, -0.01, 0.0747, 0.0038, -0.01, 0.0747, 0.006, -0.01, 0.0768, -0.0005, -0.01, 0.0768, 0.0016, -0.01, 0.079, -0.0005, -0.01, 0.079, 0.0016, -0.01, 0.0768, 0.0038, -0.01, 0.0768, 0.006, -0.01, 0.079, 0.0038, -0.01, 0.079, 0.006, -0.0061, 0.0716, 0.0081, -0.0061, 0.0716, 0.0103, -0.0083, 0.0716, 0.0081, -0.0083, 0.0716, 0.0103, -0.0061, 0.0716, 0.0125, -0.0061, 0.0716, 0.0146,
-0.0083, 0.0716, 0.0125, -0.0083, 0.0716, 0.0146, -0.01, 0.0716, 0.0081, -0.01, 0.0716, 0.0103, -0.01, 0.0747, 0.0081, -0.01, 0.0747, 0.0103, -0.01, 0.0716, 0.0125, -0.01, 0.0716, 0.0146, -0.01, 0.0747, 0.0125, -0.01, 0.0747, 0.0146, -0.01, 0.0768, 0.0081, -0.01, 0.0768, 0.0103, -0.01, 0.079, 0.0081, -0.01, 0.079, 0.0103, -0.01, 0.0768, 0.0125, -0.01, 0.0768, 0.0146, -0.01, 0.079, 0.0125, -0.01, 0.079, 0.0146, -0.0061, 0.0833, -0.0005, -0.0061, 0.0833, 0.0016, -0.0083, 0.0833, -0.0005, -0.0083, 0.0833, 0.0016, -0.0061, 0.0833, 0.0038, -0.0061, 0.0833, 0.006, -0.0083, 0.0833, 0.0038, -0.0083, 0.0833, 0.006, -0.01, 0.0812, -0.0005, -0.01, 0.0812, 0.0016, -0.01, 0.0833, -0.0005, -0.01, 0.0833, 0.0016,
-0.01, 0.0812, 0.0038, -0.01, 0.0812, 0.006, -0.01, 0.0833, 0.0038, -0.01, 0.0833, 0.006, -0.0061, 0.0833, 0.0085, -0.0083, 0.0833, 0.0085, -0.01, 0.0812, 0.0081, -0.01, 0.0812, 0.0103, -0.01, 0.0833, 0.0085, -0.01, 0.0833, 0.0103, -0.01, 0.0812, 0.0125, -0.01, 0.0812, 0.0146, -0.01, 0.0833, 0.0125, -0.01, 0.0833, 0.0146, -0.0061, 0.0855, 0.0085, -0.0083, 0.0855, 0.0085, -0.0061, 0.0877, 0.0085, -0.0083, 0.0877, 0.0085, -0.01, 0.0855, 0.0085, -0.01, 0.0855, 0.0103, -0.01, 0.0877, 0.0085, -0.01, 0.0877, 0.0103, -0.01, 0.0855, 0.0125, -0.01, 0.0855, 0.0146, -0.01, 0.0877, 0.0125, -0.01, 0.0877, 0.0146, -0.0061, 0.0716, 0.0168, -0.0061, 0.0716, 0.019, -0.0083, 0.0716, 0.0168, -0.0083, 0.0716, 0.019,
-0.0061, 0.0716, 0.0212, -0.0061, 0.0716, 0.0233, -0.0083, 0.0716, 0.0212, -0.0083, 0.0716, 0.0233, -0.01, 0.0716, 0.0168, -0.01, 0.0716, 0.019, -0.01, 0.0747, 0.0168, -0.01, 0.0747, 0.019, -0.01, 0.0716, 0.0212, -0.01, 0.0716, 0.0233, -0.01, 0.0747, 0.0212, -0.01, 0.0747, 0.0233, -0.01, 0.0768, 0.0168, -0.01, 0.0768, 0.019, -0.01, 0.079, 0.0168, -0.01, 0.079, 0.019, -0.01, 0.0768, 0.0212, -0.01, 0.0768, 0.0233, -0.01, 0.079, 0.0212, -0.01, 0.079, 0.0233, -0.0061, 0.0716, 0.0255, -0.0061, 0.0716, 0.0277, -0.0083, 0.0716, 0.0255, -0.0083, 0.0716, 0.0277, -0.0061, 0.0716, 0.0298, -0.0061, 0.0716, 0.032, -0.0083, 0.0716, 0.0298, -0.0083, 0.0716, 0.032, -0.01, 0.0716, 0.0255, -0.01, 0.0716, 0.0277,
-0.01, 0.0747, 0.0255, -0.01, 0.0747, 0.0277, -0.01, 0.0716, 0.0298, -0.01, 0.0716, 0.032, -0.01, 0.0747, 0.0298, -0.01, 0.0747, 0.032, -0.01, 0.0768, 0.0255, -0.01, 0.0768, 0.0277, -0.01, 0.079, 0.0255, -0.01, 0.079, 0.0277, -0.01, 0.0768, 0.0298, -0.01, 0.0768, 0.032, -0.01, 0.079, 0.0298, -0.01, 0.079, 0.032, -0.0061, 0.0833, 0.0203, -0.0061, 0.0833, 0.0233, -0.0083, 0.0833, 0.0203, -0.0083, 0.0833, 0.0233, -0.01, 0.0812, 0.0168, -0.01, 0.0812, 0.019, -0.01, 0.0833, 0.0168, -0.01, 0.0833, 0.019, -0.01, 0.0812, 0.0212, -0.01, 0.0812, 0.0233, -0.01, 0.0833, 0.0203, -0.01, 0.0833, 0.0233, -0.0061, 0.0855, 0.0203, -0.0083, 0.0855, 0.0203, -0.0061, 0.0877, 0.0203, -0.0083, 0.0877, 0.0203,
-0.01, 0.0855, 0.0168, -0.01, 0.0855, 0.019, -0.01, 0.0877, 0.0168, -0.01, 0.0877, 0.019, -0.01, 0.0855, 0.0203, -0.01, 0.0877, 0.0203, -0.0061, 0.0833, 0.0255, -0.0061, 0.0833, 0.0277, -0.0083, 0.0833, 0.0255, -0.0083, 0.0833, 0.0277, -0.0061, 0.0833, 0.0298, -0.0061, 0.0833, 0.032, -0.0083, 0.0833, 0.0298, -0.0083, 0.0833, 0.032, -0.01, 0.0812, 0.0255, -0.01, 0.0812, 0.0277, -0.01, 0.0833, 0.0255, -0.01, 0.0833, 0.0277, -0.01, 0.0812, 0.0298, -0.01, 0.0812, 0.032, -0.01, 0.0833, 0.0298, -0.01, 0.0833, 0.032, -0.0061, 0.0899, 0.0085, -0.0083, 0.0899, 0.0085, -0.0061, 0.092, 0.0085, -0.0083, 0.092, 0.0085, -0.01, 0.0899, 0.0085, -0.01, 0.0899, 0.0103, -0.01, 0.092, 0.0085, -0.01, 0.092, 0.0103,
-0.01, 0.0899, 0.0125, -0.01, 0.0899, 0.0146, -0.01, 0.092, 0.0125, -0.01, 0.092, 0.0146, -0.0061, 0.0942, 0.0085, -0.0083, 0.0942, 0.0085, -0.0061, 0.0964, 0.0085, -0.0083, 0.0964, 0.0085, -0.01, 0.0942, 0.0085, -0.01, 0.0942, 0.0103, -0.01, 0.0964, 0.0085, -0.01, 0.0964, 0.0103, -0.01, 0.0942, 0.0125, -0.01, 0.0942, 0.0146, -0.01, 0.0964, 0.0125, -0.01, 0.0964, 0.0146, -0.0061, 0.0985, 0.0085, -0.0083, 0.0985, 0.0085, -0.0061, 0.1007, 0.0085, -0.0083, 0.1007, 0.0085, -0.01, 0.0985, 0.0085, -0.01, 0.0985, 0.0103, -0.01, 0.1007, 0.0085, -0.01, 0.1007, 0.0103, -0.01, 0.0985, 0.0125, -0.01, 0.0985, 0.0146, -0.01, 0.1007, 0.0125, -0.01, 0.1007, 0.0146, -0.0061, 0.1029, 0.0085, -0.0083, 0.1029, 0.0085,
-0.0061, 0.105, 0.0085, -0.0083, 0.105, 0.0085, -0.01, 0.1029, 0.0085, -0.01, 0.1029, 0.0103, -0.01, 0.105, 0.0085, -0.01, 0.105, 0.0103, -0.01, 0.1029, 0.0125, -0.01, 0.1029, 0.0146, -0.01, 0.105, 0.0125, -0.01, 0.105, 0.0146, -0.0061, 0.0899, 0.0203, -0.0083, 0.0899, 0.0203, -0.0061, 0.092, 0.0203, -0.0083, 0.092, 0.0203, -0.01, 0.0899, 0.0168, -0.01, 0.0899, 0.019, -0.01, 0.092, 0.0168, -0.01, 0.092, 0.019, -0.01, 0.0899, 0.0203, -0.01, 0.092, 0.0203, -0.0061, 0.0942, 0.0203, -0.0083, 0.0942, 0.0203, -0.0061, 0.0964, 0.0203, -0.0083, 0.0964, 0.0203, -0.01, 0.0942, 0.0168, -0.01, 0.0942, 0.019, -0.01, 0.0964, 0.0168, -0.01, 0.0964, 0.019, -0.01, 0.0942, 0.0203, -0.01, 0.0964, 0.0203,
-0.0061, 0.0985, 0.0203, -0.0083, 0.0985, 0.0203, -0.0061, 0.1007, 0.0203, -0.0083, 0.1007, 0.0203, -0.01, 0.0985, 0.0168, -0.01, 0.0985, 0.019, -0.01, 0.1007, 0.0168, -0.01, 0.1007, 0.019, -0.01, 0.0985, 0.0203, -0.01, 0.1007, 0.0203, -0.0061, 0.1029, 0.0203, -0.0083, 0.1029, 0.0203, -0.0061, 0.105, 0.0203, -0.0083, 0.105, 0.0203, -0.01, 0.1029, 0.0168, -0.01, 0.1029, 0.019, -0.01, 0.105, 0.0168, -0.01, 0.105, 0.019, -0.01, 0.1029, 0.0203, -0.01, 0.105, 0.0203, -0.0061, 0.0716, 0.0342, -0.0061, 0.0716, 0.0363, -0.0083, 0.0716, 0.0342, -0.0083, 0.0716, 0.0363, -0.0061, 0.0716, 0.0385, -0.0061, 0.0716, 0.0407, -0.0083, 0.0716, 0.0385, -0.0083, 0.0716, 0.0407, -0.01, 0.0716, 0.0342, -0.01, 0.0716, 0.0363,
-0.01, 0.0747, 0.0342, -0.01, 0.0747, 0.0363, -0.01, 0.0716, 0.0385, -0.01, 0.0716, 0.0407, -0.01, 0.0747, 0.0385, -0.01, 0.0747, 0.0407, -0.01, 0.0768, 0.0342, -0.01, 0.0768, 0.0363, -0.01, 0.079, 0.0342, -0.01, 0.079, 0.0363, -0.01, 0.0768, 0.0385, -0.01, 0.0768, 0.0407, -0.01, 0.079, 0.0385, -0.01, 0.079, 0.0407, -0.0061, 0.0716, 0.0429, -0.0061, 0.0716, 0.045, -0.0083, 0.0716, 0.0429, -0.0083, 0.0716, 0.045, -0.0061, 0.0716, 0.0472, -0.0061, 0.0716, 0.0487, -0.0083, 0.0716, 0.0472, -0.0083, 0.0716, 0.0487, -0.0061, 0.0747, 0.0487, -0.0083, 0.0747, 0.0487, -0.01, 0.0716, 0.0429, -0.01, 0.0716, 0.045, -0.01, 0.0747, 0.0429, -0.01, 0.0747, 0.045, -0.01, 0.0716, 0.0472, -0.01, 0.0716, 0.0487,
-0.01, 0.0747, 0.0472, -0.01, 0.0747, 0.0487, -0.0061, 0.0768, 0.0487, -0.0083, 0.0768, 0.0487, -0.0061, 0.079, 0.0487, -0.0083, 0.079, 0.0487, -0.01, 0.0768, 0.0429, -0.01, 0.0768, 0.045, -0.01, 0.079, 0.0429, -0.01, 0.079, 0.045, -0.01, 0.0768, 0.0472, -0.01, 0.0768, 0.0487, -0.01, 0.079, 0.0472, -0.01, 0.079, 0.0487, -0.0061, 0.0833, 0.0342, -0.0061, 0.0833, 0.0363, -0.0083, 0.0833, 0.0342, -0.0083, 0.0833, 0.0363, -0.0061, 0.0833, 0.0379, -0.0083, 0.0833, 0.0379, -0.01, 0.0812, 0.0342, -0.01, 0.0812, 0.0363, -0.01, 0.0833, 0.0342, -0.01, 0.0833, 0.0363, -0.01, 0.0812, 0.0385, -0.01, 0.0812, 0.0407, -0.01, 0.0833, 0.0379, -0.01, 0.0833, 0.0407, -0.0061, 0.0855, 0.0379, -0.0083, 0.0855, 0.0379,
-0.0061, 0.0877, 0.0379, -0.0083, 0.0877, 0.0379, -0.01, 0.0855, 0.0379, -0.01, 0.0855, 0.0407, -0.01, 0.0877, 0.0379, -0.01, 0.0877, 0.0407, -0.0061, 0.0812, 0.0487, -0.0083, 0.0812, 0.0487, -0.0061, 0.0833, 0.0487, -0.0083, 0.0833, 0.0487, -0.01, 0.0812, 0.0429, -0.01, 0.0812, 0.045, -0.01, 0.0833, 0.0429, -0.01, 0.0833, 0.045, -0.01, 0.0812, 0.0472, -0.01, 0.0812, 0.0487, -0.01, 0.0833, 0.0472, -0.01, 0.0833, 0.0487, -0.0061, 0.0855, 0.0487, -0.0083, 0.0855, 0.0487, -0.0061, 0.0877, 0.0487, -0.0083, 0.0877, 0.0487, -0.01, 0.0855, 0.0429, -0.01, 0.0855, 0.045, -0.01, 0.0877, 0.0429, -0.01, 0.0877, 0.045, -0.01, 0.0855, 0.0472, -0.01, 0.0855, 0.0487, -0.01, 0.0877, 0.0472, -0.01, 0.0877, 0.0487,
-0.0061, 0.0899, 0.0379, -0.0083, 0.0899, 0.0379, -0.0061, 0.092, 0.0379, -0.0083, 0.092, 0.0379, -0.01, 0.0899, 0.0379, -0.01, 0.0899, 0.0407, -0.01, 0.092, 0.0379, -0.01, 0.092, 0.0407, -0.0061, 0.0942, 0.0379, -0.0083, 0.0942, 0.0379, -0.0061, 0.0964, 0.0379, -0.0083, 0.0964, 0.0379, -0.01, 0.0942, 0.0379, -0.01, 0.0942, 0.0407, -0.01, 0.0964, 0.0379, -0.01, 0.0964, 0.0407, -0.0061, 0.0899, 0.0487, -0.0083, 0.0899, 0.0487, -0.0061, 0.092, 0.0487, -0.0083, 0.092, 0.0487, -0.01, 0.0899, 0.0429, -0.01, 0.0899, 0.045, -0.01, 0.092, 0.0429, -0.01, 0.092, 0.045, -0.01, 0.0899, 0.0472, -0.01, 0.0899, 0.0487, -0.01, 0.092, 0.0472, -0.01, 0.092, 0.0487, -0.0061, 0.0942, 0.0487, -0.0083, 0.0942, 0.0487,
-0.0061, 0.0964, 0.0487, -0.0083, 0.0964, 0.0487, -0.01, 0.0942, 0.0429, -0.01, 0.0942, 0.045, -0.01, 0.0964, 0.0429, -0.01, 0.0964, 0.045, -0.01, 0.0942, 0.0472, -0.01, 0.0942, 0.0487, -0.01, 0.0964, 0.0472, -0.01, 0.0964, 0.0487, -0.0061, 0.0985, 0.0379, -0.0083, 0.0985, 0.0379, -0.0061, 0.1007, 0.0379, -0.0083, 0.1007, 0.0379, -0.01, 0.0985, 0.0379, -0.01, 0.0985, 0.0407, -0.01, 0.1007, 0.0379, -0.01, 0.1007, 0.0407, -0.0061, 0.1029, 0.0379, -0.0083, 0.1029, 0.0379, -0.0061, 0.105, 0.0379, -0.0083, 0.105, 0.0379, -0.01, 0.1029, 0.0379, -0.01, 0.1029, 0.0407, -0.01, 0.105, 0.0379, -0.01, 0.105, 0.0407, -0.0061, 0.0985, 0.0487, -0.0083, 0.0985, 0.0487, -0.0061, 0.1007, 0.0487, -0.0083, 0.1007, 0.0487,
-0.01, 0.0985, 0.0429, -0.01, 0.0985, 0.045, -0.01, 0.1007, 0.0429, -0.01, 0.1007, 0.045, -0.01, 0.0985, 0.0472, -0.01, 0.0985, 0.0487, -0.01, 0.1007, 0.0472, -0.01, 0.1007, 0.0487, -0.0061, 0.1029, 0.0487, -0.0083, 0.1029, 0.0487, -0.0061, 0.105, 0.0487, -0.0083, 0.105, 0.0487, -0.01, 0.1029, 0.0429, -0.01, 0.1029, 0.045, -0.01, 0.105, 0.0429, -0.01, 0.105, 0.045, -0.01, 0.1029, 0.0472, -0.01, 0.1029, 0.0487, -0.01, 0.105, 0.0472, -0.01, 0.105, 0.0487, -0.0061, 0.1072, 0.0085, -0.0083, 0.1072, 0.0085, -0.0061, 0.1094, 0.0085, -0.0083, 0.1094, 0.0085, -0.01, 0.1072, 0.0085, -0.01, 0.1072, 0.0103, -0.01, 0.1094, 0.0085, -0.01, 0.1094, 0.0103, -0.01, 0.1072, 0.0125, -0.01, 0.1072, 0.0146,
-0.01, 0.1094, 0.0125, -0.01, 0.1094, 0.0146, -0.0061, 0.1116, 0.0085, -0.0083, 0.1116, 0.0085, -0.0061, 0.1137, 0.0085, -0.0083, 0.1137, 0.0085, -0.01, 0.1116, 0.0085, -0.01, 0.1116, 0.0103, -0.01, 0.1137, 0.0085, -0.01, 0.1137, 0.0103, -0.01, 0.1116, 0.0125, -0.01, 0.1116, 0.0146, -0.01, 0.1137, 0.0125, -0.01, 0.1137, 0.0146, -0.0061, 0.1159, 0.0085, -0.0083, 0.1159, 0.0085, -0.0061, 0.1181, 0.0085, -0.0083, 0.1181, 0.0085, -0.01, 0.1159, 0.0085, -0.01, 0.1159, 0.0103, -0.01, 0.1181, 0.0085, -0.01, 0.1181, 0.0103, -0.01, 0.1159, 0.0125, -0.01, 0.1159, 0.0146, -0.01, 0.1181, 0.0125, -0.01, 0.1181, 0.0146, -0.0061, 0.1202, 0.0085, -0.0083, 0.1202, 0.0085, -0.0061, 0.1224, 0.0085, -0.0083, 0.1224, 0.0085,
-0.01, 0.1202, 0.0085, -0.01, 0.1202, 0.0103, -0.01, 0.1224, 0.0085, -0.01, 0.1224, 0.0103, -0.01, 0.1202, 0.0125, -0.01, 0.1202, 0.0146, -0.01, 0.1224, 0.0125, -0.01, 0.1224, 0.0146, -0.0061, 0.1072, 0.0203, -0.0083, 0.1072, 0.0203, -0.0061, 0.1094, 0.0203, -0.0083, 0.1094, 0.0203, -0.01, 0.1072, 0.0168, -0.01, 0.1072, 0.019, -0.01, 0.1094, 0.0168, -0.01, 0.1094, 0.019, -0.01, 0.1072, 0.0203, -0.01, 0.1094, 0.0203, -0.0061, 0.1116, 0.0203, -0.0083, 0.1116, 0.0203, -0.0061, 0.1137, 0.0203, -0.0083, 0.1137, 0.0203, -0.01, 0.1116, 0.0168, -0.01, 0.1116, 0.019, -0.01, 0.1137, 0.0168, -0.01, 0.1137, 0.019, -0.01, 0.1116, 0.0203, -0.01, 0.1137, 0.0203, -0.0061, 0.1159, 0.0203, -0.0083, 0.1159, 0.0203,
-0.0061, 0.1181, 0.0203, -0.0083, 0.1181, 0.0203, -0.01, 0.1159, 0.0168, -0.01, 0.1159, 0.019, -0.01, 0.1181, 0.0168, -0.01, 0.1181, 0.019, -0.01, 0.1159, 0.0203, -0.01, 0.1181, 0.0203, -0.0061, 0.1202, 0.0203, -0.0083, 0.1202, 0.0203, -0.0061, 0.1224, 0.0203, -0.0083, 0.1224, 0.0203, -0.01, 0.1202, 0.0168, -0.01, 0.1202, 0.019, -0.01, 0.1224, 0.0168, -0.01, 0.1224, 0.019, -0.01, 0.1202, 0.0203, -0.01, 0.1224, 0.0203, -0.0061, 0.1246, 0.0085, -0.0083, 0.1246, 0.0085, -0.0061, 0.1267, 0.0085, -0.0083, 0.1267, 0.0085, -0.01, 0.1246, 0.0085, -0.01, 0.1246, 0.0103, -0.01, 0.1267, 0.0085, -0.01, 0.1267, 0.0103, -0.01, 0.1246, 0.0125, -0.01, 0.1246, 0.0146, -0.01, 0.1267, 0.0125, -0.01, 0.1267, 0.0146,
-0.0061, 0.1294, 0.0085, -0.0061, 0.1294, 0.0103, -0.0083, 0.1294, 0.0085, -0.0083, 0.1294, 0.0103, -0.0061, 0.1294, 0.0125, -0.0061, 0.1294, 0.0146, -0.0083, 0.1294, 0.0125, -0.0083, 0.1294, 0.0146, -0.01, 0.1294, 0.0085, -0.01, 0.1294, 0.0103, -0.01, 0.1294, 0.0125, -0.01, 0.1294, 0.0146, -0.0061, 0.1246, 0.0203, -0.0083, 0.1246, 0.0203, -0.0061, 0.1267, 0.0203, -0.0083, 0.1267, 0.0203, -0.01, 0.1246, 0.0168, -0.01, 0.1246, 0.019, -0.01, 0.1267, 0.0168, -0.01, 0.1267, 0.019, -0.01, 0.1246, 0.0203, -0.01, 0.1267, 0.0203, -0.0061, 0.1294, 0.0168, -0.0061, 0.1294, 0.019, -0.0083, 0.1294, 0.0168, -0.0083, 0.1294, 0.019, -0.0061, 0.1294, 0.0203, -0.0083, 0.1294, 0.0203, -0.01, 0.1294, 0.0168, -0.01, 0.1294, 0.019,
-0.01, 0.1294, 0.0203, -0.0061, 0.1072, 0.0379, -0.0083, 0.1072, 0.0379, -0.0061, 0.1094, 0.0379, -0.0083, 0.1094, 0.0379, -0.01, 0.1072, 0.0379, -0.01, 0.1072, 0.0407, -0.01, 0.1094, 0.0379, -0.01, 0.1094, 0.0407, -0.0061, 0.1116, 0.0379, -0.0083, 0.1116, 0.0379, -0.0061, 0.1137, 0.0379, -0.0083, 0.1137, 0.0379, -0.01, 0.1116, 0.0379, -0.01, 0.1116, 0.0407, -0.01, 0.1137, 0.0379, -0.01, 0.1137, 0.0407, -0.0061, 0.1072, 0.0487, -0.0083, 0.1072, 0.0487, -0.0061, 0.1094, 0.0487, -0.0083, 0.1094, 0.0487, -0.01, 0.1072, 0.0429, -0.01, 0.1072, 0.045, -0.01, 0.1094, 0.0429, -0.01, 0.1094, 0.045, -0.01, 0.1072, 0.0472, -0.01, 0.1072, 0.0487, -0.01, 0.1094, 0.0472, -0.01, 0.1094, 0.0487, -0.0061, 0.1116, 0.0487,
-0.0083, 0.1116, 0.0487, -0.0061, 0.1137, 0.0487, -0.0083, 0.1137, 0.0487, -0.01, 0.1116, 0.0429, -0.01, 0.1116, 0.045, -0.01, 0.1137, 0.0429, -0.01, 0.1137, 0.045, -0.01, 0.1116, 0.0472, -0.01, 0.1116, 0.0487, -0.01, 0.1137, 0.0472, -0.01, 0.1137, 0.0487, -0.0061, 0.1159, 0.0379, -0.0083, 0.1159, 0.0379, -0.0061, 0.1181, 0.0379, -0.0083, 0.1181, 0.0379, -0.01, 0.1159, 0.0379, -0.01, 0.1159, 0.0407, -0.01, 0.1181, 0.0379, -0.01, 0.1181, 0.0407, -0.0061, 0.1202, 0.0379, -0.0083, 0.1202, 0.0379, -0.0061, 0.1224, 0.0379, -0.0083, 0.1224, 0.0379, -0.01, 0.1202, 0.0379, -0.01, 0.1202, 0.0407, -0.01, 0.1224, 0.0379, -0.01, 0.1224, 0.0407, -0.0061, 0.1159, 0.0487, -0.0083, 0.1159, 0.0487, -0.0061, 0.1181, 0.0487,
-0.0083, 0.1181, 0.0487, -0.01, 0.1159, 0.0429, -0.01, 0.1159, 0.045, -0.01, 0.1181, 0.0429, -0.01, 0.1181, 0.045, -0.01, 0.1159, 0.0472, -0.01, 0.1159, 0.0487, -0.01, 0.1181, 0.0472, -0.01, 0.1181, 0.0487, -0.0061, 0.1202, 0.0487, -0.0083, 0.1202, 0.0487, -0.0061, 0.1224, 0.0487, -0.0083, 0.1224, 0.0487, -0.01, 0.1202, 0.0429, -0.01, 0.1202, 0.045, -0.01, 0.1224, 0.0429, -0.01, 0.1224, 0.045, -0.01, 0.1202, 0.0472, -0.01, 0.1202, 0.0487, -0.01, 0.1224, 0.0472, -0.01, 0.1224, 0.0487, -0.0061, 0.1246, 0.0379, -0.0083, 0.1246, 0.0379, -0.0061, 0.1267, 0.0379, -0.0083, 0.1267, 0.0379, -0.01, 0.1246, 0.0379, -0.01, 0.1246, 0.0407, -0.01, 0.1267, 0.0379, -0.01, 0.1267, 0.0407, -0.0061, 0.1289, 0.0379,
-0.0083, 0.1289, 0.0379, -0.0061, 0.1311, 0.0379, -0.0083, 0.1311, 0.0379, -0.01, 0.1289, 0.0379, -0.01, 0.1289, 0.0407, -0.01, 0.1311, 0.0379, -0.01, 0.1311, 0.0407, -0.0061, 0.1246, 0.0487, -0.0083, 0.1246, 0.0487, -0.0061, 0.1267, 0.0487, -0.0083, 0.1267, 0.0487, -0.01, 0.1246, 0.0429, -0.01, 0.1246, 0.045, -0.01, 0.1267, 0.0429, -0.01, 0.1267, 0.045, -0.01, 0.1246, 0.0472, -0.01, 0.1246, 0.0487, -0.01, 0.1267, 0.0472, -0.01, 0.1267, 0.0487, -0.0061, 0.1289, 0.0487, -0.0083, 0.1289, 0.0487, -0.0061, 0.1311, 0.0487, -0.0083, 0.1311, 0.0487, -0.01, 0.1289, 0.0429, -0.01, 0.1289, 0.045, -0.01, 0.1311, 0.0429, -0.01, 0.1311, 0.045, -0.01, 0.1289, 0.0472, -0.01, 0.1289, 0.0487, -0.01, 0.1311, 0.0472,
-0.01, 0.1311, 0.0487, -0.0061, 0.1324, 0.0379, -0.0061, 0.1324, 0.0407, -0.0083, 0.1324, 0.0379, -0.0083, 0.1324, 0.0407, -0.01, 0.1324, 0.0379, -0.01, 0.1324, 0.0407, -0.0061, 0.1324, 0.0429, -0.0061, 0.1324, 0.045, -0.0083, 0.1324, 0.0429, -0.0083, 0.1324, 0.045, -0.0061, 0.1324, 0.0472, -0.0061, 0.1324, 0.0487, -0.0083, 0.1324, 0.0472, -0.0083, 0.1324, 0.0487, -0.01, 0.1324, 0.0429, -0.01, 0.1324, 0.045, -0.01, 0.1324, 0.0472, -0.01, 0.1324, 0.0487, -0.0061, 0.1412, -0.0005, -0.0061, 0.1412, 0.0016, -0.0083, 0.1412, -0.0005, -0.0083, 0.1412, 0.0016, -0.0061, 0.1412, 0.0038, -0.0061, 0.1412, 0.006, -0.0083, 0.1412, 0.0038, -0.0083, 0.1412, 0.006, -0.01, 0.1412, -0.0005, -0.01, 0.1412, 0.0016, -0.01, 0.1441, -0.0005,
-0.01, 0.1441, 0.0016, -0.01, 0.1412, 0.0038, -0.01, 0.1412, 0.006, -0.01, 0.1441, 0.0038, -0.01, 0.1441, 0.006, -0.01, 0.1463, -0.0005, -0.01, 0.1463, 0.0016, -0.01, 0.1484, -0.0005, -0.01, 0.1484, 0.0016, -0.01, 0.1463, 0.0038, -0.01, 0.1463, 0.006, -0.01, 0.1484, 0.0038, -0.01, 0.1484, 0.006, -0.0061, 0.1412, 0.0081, -0.0061, 0.1412, 0.0103, -0.0083, 0.1412, 0.0081, -0.0083, 0.1412, 0.0103, -0.0061, 0.1412, 0.0125, -0.0061, 0.1412, 0.0146, -0.0083, 0.1412, 0.0125, -0.0083, 0.1412, 0.0146, -0.01, 0.1412, 0.0081, -0.01, 0.1412, 0.0103, -0.01, 0.1441, 0.0081, -0.01, 0.1441, 0.0103, -0.01, 0.1412, 0.0125, -0.01, 0.1412, 0.0146, -0.01, 0.1441, 0.0125, -0.01, 0.1441, 0.0146, -0.01, 0.1463, 0.0081,
-0.01, 0.1463, 0.0103, -0.01, 0.1484, 0.0081, -0.01, 0.1484, 0.0103, -0.01, 0.1463, 0.0125, -0.01, 0.1463, 0.0146, -0.01, 0.1484, 0.0125, -0.01, 0.1484, 0.0146, -0.0061, 0.152, -0.0005, -0.0061, 0.152, 0.0016, -0.0083, 0.152, -0.0005, -0.0083, 0.152, 0.0016, -0.0061, 0.152, 0.0038, -0.0061, 0.152, 0.006, -0.0083, 0.152, 0.0038, -0.0083, 0.152, 0.006, -0.01, 0.1506, -0.0005, -0.01, 0.1506, 0.0016, -0.01, 0.152, -0.0005, -0.01, 0.152, 0.0016, -0.01, 0.1506, 0.0038, -0.01, 0.1506, 0.006, -0.01, 0.152, 0.0038, -0.01, 0.152, 0.006, -0.0061, 0.152, 0.0081, -0.0061, 0.152, 0.0103, -0.0083, 0.152, 0.0081, -0.0083, 0.152, 0.0103, -0.0061, 0.152, 0.0125, -0.0061, 0.152, 0.0146, -0.0083, 0.152, 0.0125,
-0.0083, 0.152, 0.0146, -0.01, 0.1506, 0.0081, -0.01, 0.1506, 0.0103, -0.01, 0.152, 0.0081, -0.01, 0.152, 0.0103, -0.01, 0.1506, 0.0125, -0.01, 0.1506, 0.0146, -0.01, 0.152, 0.0125, -0.01, 0.152, 0.0146, -0.0061, 0.1412, 0.0168, -0.0061, 0.1412, 0.019, -0.0083, 0.1412, 0.0168, -0.0083, 0.1412, 0.019, -0.0061, 0.1412, 0.0212, -0.0061, 0.1412, 0.0233, -0.0083, 0.1412, 0.0212, -0.0083, 0.1412, 0.0233, -0.01, 0.1412, 0.0168, -0.01, 0.1412, 0.019, -0.01, 0.1441, 0.0168, -0.01, 0.1441, 0.019, -0.01, 0.1412, 0.0212, -0.01, 0.1412, 0.0233, -0.01, 0.1441, 0.0212, -0.01, 0.1441, 0.0233, -0.01, 0.1463, 0.0168, -0.01, 0.1463, 0.019, -0.01, 0.1484, 0.0168, -0.01, 0.1484, 0.019, -0.01, 0.1463, 0.0212,
-0.01, 0.1463, 0.0233, -0.01, 0.1484, 0.0212, -0.01, 0.1484, 0.0233, -0.0061, 0.1412, 0.0255, -0.0061, 0.1412, 0.0277, -0.0083, 0.1412, 0.0255, -0.0083, 0.1412, 0.0277, -0.0061, 0.1412, 0.0298, -0.0061, 0.1412, 0.032, -0.0083, 0.1412, 0.0298, -0.0083, 0.1412, 0.032, -0.01, 0.1412, 0.0255, -0.01, 0.1412, 0.0277, -0.01, 0.1441, 0.0255, -0.01, 0.1441, 0.0277, -0.01, 0.1412, 0.0298, -0.01, 0.1412, 0.032, -0.01, 0.1441, 0.0298, -0.01, 0.1441, 0.032, -0.01, 0.1463, 0.0255, -0.01, 0.1463, 0.0277, -0.01, 0.1484, 0.0255, -0.01, 0.1484, 0.0277, -0.01, 0.1463, 0.0298, -0.01, 0.1463, 0.032, -0.01, 0.1484, 0.0298, -0.01, 0.1484, 0.032, -0.0061, 0.152, 0.0168, -0.0061, 0.152, 0.019, -0.0083, 0.152, 0.0168,
-0.0083, 0.152, 0.019, -0.0061, 0.152, 0.0212, -0.0061, 0.152, 0.0233, -0.0083, 0.152, 0.0212, -0.0083, 0.152, 0.0233, -0.01, 0.1506, 0.0168, -0.01, 0.1506, 0.019, -0.01, 0.152, 0.0168, -0.01, 0.152, 0.019, -0.01, 0.1506, 0.0212, -0.01, 0.1506, 0.0233, -0.01, 0.152, 0.0212, -0.01, 0.152, 0.0233, -0.0061, 0.152, 0.0255, -0.0061, 0.152, 0.0277, -0.0083, 0.152, 0.0255, -0.0083, 0.152, 0.0277, -0.0061, 0.152, 0.0298, -0.0061, 0.152, 0.032, -0.0083, 0.152, 0.0298, -0.0083, 0.152, 0.032, -0.01, 0.1506, 0.0255, -0.01, 0.1506, 0.0277, -0.01, 0.152, 0.0255, -0.01, 0.152, 0.0277, -0.01, 0.1506, 0.0298, -0.01, 0.1506, 0.032, -0.01, 0.152, 0.0298, -0.01, 0.152, 0.032, -0.0061, 0.1412, 0.0342,
-0.0061, 0.1412, 0.0363, -0.0083, 0.1412, 0.0342, -0.0083, 0.1412, 0.0363, -0.0061, 0.1412, 0.0385, -0.0061, 0.1412, 0.0407, -0.0083, 0.1412, 0.0385, -0.0083, 0.1412, 0.0407, -0.01, 0.1412, 0.0342, -0.01, 0.1412, 0.0363, -0.01, 0.1441, 0.0342, -0.01, 0.1441, 0.0363, -0.01, 0.1412, 0.0385, -0.01, 0.1412, 0.0407, -0.01, 0.1441, 0.0385, -0.01, 0.1441, 0.0407, -0.01, 0.1463, 0.0342, -0.01, 0.1463, 0.0363, -0.01, 0.1484, 0.0342, -0.01, 0.1484, 0.0363, -0.01, 0.1463, 0.0385, -0.01, 0.1463, 0.0407, -0.01, 0.1484, 0.0385, -0.01, 0.1484, 0.0407, -0.0061, 0.1412, 0.0429, -0.0061, 0.1412, 0.045, -0.0083, 0.1412, 0.0429, -0.0083, 0.1412, 0.045, -0.0061, 0.1412, 0.0472, -0.0061, 0.1412, 0.0487, -0.0083, 0.1412, 0.0472,
-0.0083, 0.1412, 0.0487, -0.0061, 0.1441, 0.0487, -0.0083, 0.1441, 0.0487, -0.01, 0.1412, 0.0429, -0.01, 0.1412, 0.045, -0.01, 0.1441, 0.0429, -0.01, 0.1441, 0.045, -0.01, 0.1412, 0.0472, -0.01, 0.1412, 0.0487, -0.01, 0.1441, 0.0472, -0.01, 0.1441, 0.0487, -0.0061, 0.1463, 0.0487, -0.0083, 0.1463, 0.0487, -0.0061, 0.1484, 0.0487, -0.0083, 0.1484, 0.0487, -0.01, 0.1463, 0.0429, -0.01, 0.1463, 0.045, -0.01, 0.1484, 0.0429, -0.01, 0.1484, 0.045, -0.01, 0.1463, 0.0472, -0.01, 0.1463, 0.0487, -0.01, 0.1484, 0.0472, -0.01, 0.1484, 0.0487, -0.0061, 0.152, 0.0342, -0.0061, 0.152, 0.0363, -0.0083, 0.152, 0.0342, -0.0083, 0.152, 0.0363, -0.0061, 0.152, 0.0385, -0.0061, 0.152, 0.0407, -0.0083, 0.152, 0.0385,
-0.0083, 0.152, 0.0407, -0.01, 0.1506, 0.0342, -0.01, 0.1506, 0.0363, -0.01, 0.152, 0.0342, -0.01, 0.152, 0.0363, -0.01, 0.1506, 0.0385, -0.01, 0.1506, 0.0407, -0.01, 0.152, 0.0385, -0.01, 0.152, 0.0407, -0.0061, 0.152, 0.0429, -0.0061, 0.152, 0.045, -0.0083, 0.152, 0.0429, -0.0083, 0.152, 0.045, -0.0061, 0.1506, 0.0487, -0.0083, 0.1506, 0.0487, -0.0061, 0.152, 0.0472, -0.0061, 0.152, 0.0487, -0.0083, 0.152, 0.0472, -0.0083, 0.152, 0.0487, -0.01, 0.1506, 0.0429, -0.01, 0.1506, 0.045, -0.01, 0.152, 0.0429, -0.01, 0.152, 0.045, -0.01, 0.1506, 0.0472, -0.01, 0.1506, 0.0487, -0.01, 0.152, 0.0472, -0.01, 0.152, 0.0487, -0.0061, 0.2139, -0.0006, -0.0061, 0.2136, 0.0016, -0.0083, 0.2139, -0.0006,
-0.0083, 0.2136, 0.0016, -0.0061, 0.2133, 0.0038, -0.0061, 0.213, 0.006, -0.0083, 0.2133, 0.0038, -0.0083, 0.213, 0.006, -0.01, 0.2139, -0.0006, -0.01, 0.2136, 0.0015, -0.01, 0.2133, 0.0037, -0.01, 0.213, 0.0059, -0.01, 0.2157, -0.0005, -0.01, 0.2157, 0.0016, -0.01, 0.2179, -0.0005, -0.01, 0.2179, 0.0016, -0.01, 0.2157, 0.0038, -0.01, 0.2157, 0.006, -0.01, 0.2179, 0.0038, -0.01, 0.2179, 0.006, -0.0061, 0.2128, 0.0081, -0.0061, 0.2127, 0.0103, -0.0083, 0.2128, 0.0081, -0.0083, 0.2127, 0.0103, -0.0061, 0.2127, 0.0125, -0.0061, 0.2127, 0.0146, -0.0083, 0.2127, 0.0125, -0.0083, 0.2127, 0.0146, -0.01, 0.2128, 0.0081, -0.01, 0.2127, 0.0103, -0.01, 0.2127, 0.0125, -0.01, 0.2127, 0.0146, -0.01, 0.2157, 0.0081,
-0.01, 0.2157, 0.0103, -0.01, 0.2179, 0.0081, -0.01, 0.2179, 0.0103, -0.01, 0.2157, 0.0125, -0.01, 0.2157, 0.0146, -0.01, 0.2179, 0.0125, -0.01, 0.2179, 0.0146, -0.01, 0.2201, -0.0005, -0.01, 0.2201, 0.0016, -0.01, 0.2222, -0.0005, -0.01, 0.2222, 0.0016, -0.01, 0.2201, 0.0038, -0.01, 0.2201, 0.006, -0.01, 0.2222, 0.0038, -0.01, 0.2222, 0.006, -0.0061, 0.2271, -0.0002, -0.0061, 0.226, 0.0016, -0.0083, 0.2271, -0.0002, -0.0083, 0.226, 0.0016, -0.0061, 0.2252, 0.0039, -0.0061, 0.2248, 0.006, -0.0083, 0.2252, 0.0039, -0.0083, 0.2248, 0.006, -0.0061, 0.2255, 0.0029, -0.0083, 0.2255, 0.0029, -0.01, 0.2244, -0.0005, -0.01, 0.2244, 0.0016, -0.01, 0.2272, -0.0002, -0.01, 0.2259, 0.0017, -0.01, 0.2252, 0.0038,
-0.01, 0.2247, 0.0061, -0.01, 0.2256, 0.0028, -0.01, 0.2201, 0.0081, -0.01, 0.2201, 0.0103, -0.01, 0.2222, 0.0081, -0.01, 0.2222, 0.0103, -0.01, 0.2201, 0.0125, -0.01, 0.2201, 0.0146, -0.01, 0.2222, 0.0125, -0.01, 0.2222, 0.0146, -0.0061, 0.2245, 0.0081, -0.0061, 0.2245, 0.0103, -0.0083, 0.2245, 0.0081, -0.0083, 0.2245, 0.0103, -0.0061, 0.2245, 0.0125, -0.0061, 0.2245, 0.0146, -0.0083, 0.2245, 0.0125, -0.0083, 0.2245, 0.0146, -0.01, 0.2245, 0.0082, -0.01, 0.2245, 0.0103, -0.01, 0.2245, 0.0125, -0.01, 0.2245, 0.0146, -0.0061, 0.2127, 0.0168, -0.0061, 0.2127, 0.019, -0.0083, 0.2127, 0.0168, -0.0083, 0.2127, 0.019, -0.0061, 0.2127, 0.0212, -0.0061, 0.2127, 0.0233, -0.0083, 0.2127, 0.0212, -0.0083, 0.2127, 0.0233,
-0.01, 0.2127, 0.0168, -0.01, 0.2127, 0.019, -0.01, 0.2127, 0.0212, -0.01, 0.2127, 0.0233, -0.01, 0.2157, 0.0168, -0.01, 0.2157, 0.019, -0.01, 0.2179, 0.0168, -0.01, 0.2179, 0.019, -0.01, 0.2157, 0.0212, -0.01, 0.2157, 0.0233, -0.01, 0.2179, 0.0212, -0.01, 0.2179, 0.0233, -0.0061, 0.2127, 0.0255, -0.0061, 0.2127, 0.0277, -0.0083, 0.2127, 0.0255, -0.0083, 0.2127, 0.0277, -0.0061, 0.2127, 0.0298, -0.0061, 0.2127, 0.032, -0.0083, 0.2127, 0.0298, -0.0083, 0.2127, 0.032, -0.01, 0.2127, 0.0255, -0.01, 0.2127, 0.0277, -0.01, 0.2127, 0.0298, -0.01, 0.2127, 0.032, -0.01, 0.2157, 0.0255, -0.01, 0.2157, 0.0277, -0.01, 0.2179, 0.0255, -0.01, 0.2179, 0.0277, -0.01, 0.2157, 0.0298, -0.01, 0.2157, 0.032,
-0.01, 0.2179, 0.0298, -0.01, 0.2179, 0.032, -0.01, 0.2201, 0.0168, -0.01, 0.2201, 0.019, -0.01, 0.2222, 0.0168, -0.01, 0.2222, 0.019, -0.01, 0.2201, 0.0212, -0.01, 0.2201, 0.0233, -0.01, 0.2222, 0.0212, -0.01, 0.2222, 0.0233, -0.0061, 0.2245, 0.0168, -0.0061, 0.2245, 0.019, -0.0083, 0.2245, 0.0168, -0.0083, 0.2245, 0.019, -0.0061, 0.2245, 0.0212, -0.0061, 0.2245, 0.0233, -0.0083, 0.2245, 0.0212, -0.0083, 0.2245, 0.0233, -0.01, 0.2245, 0.0168, -0.01, 0.2245, 0.019, -0.01, 0.2245, 0.0212, -0.01, 0.2245, 0.0233, -0.01, 0.2201, 0.0255, -0.01, 0.2201, 0.0277, -0.01, 0.2222, 0.0255, -0.01, 0.2222, 0.0277, -0.01, 0.2201, 0.0298, -0.01, 0.2201, 0.032, -0.01, 0.2222, 0.0298, -0.01, 0.2222, 0.032,
-0.0061, 0.2245, 0.0255, -0.0061, 0.2245, 0.0277, -0.0083, 0.2245, 0.0255, -0.0083, 0.2245, 0.0277, -0.0061, 0.2245, 0.0298, -0.0061, 0.2245, 0.032, -0.0083, 0.2245, 0.0298, -0.0083, 0.2245, 0.032, -0.01, 0.2245, 0.0255, -0.01, 0.2245, 0.0277, -0.01, 0.2245, 0.0298, -0.01, 0.2245, 0.032, -0.0061, 0.2279, -0.0013, -0.0083, 0.2279, -0.0013, -0.01, 0.2279, -0.0014, -0.0061, 0.2127, 0.0342, -0.0061, 0.2127, 0.0363, -0.0083, 0.2127, 0.0342, -0.0083, 0.2127, 0.0363, -0.0061, 0.2127, 0.0385, -0.0061, 0.2127, 0.0407, -0.0083, 0.2127, 0.0385, -0.0083, 0.2127, 0.0407, -0.01, 0.2127, 0.0342, -0.01, 0.2127, 0.0363, -0.01, 0.2127, 0.0385, -0.01, 0.2127, 0.0407, -0.01, 0.2157, 0.0342, -0.01, 0.2157, 0.0363, -0.01, 0.2179, 0.0342,
-0.01, 0.2179, 0.0363, -0.01, 0.2157, 0.0385, -0.01, 0.2157, 0.0407, -0.01, 0.2179, 0.0385, -0.01, 0.2179, 0.0407, -0.0061, 0.2127, 0.0429, -0.0061, 0.2127, 0.045, -0.0083, 0.2127, 0.0429, -0.0083, 0.2127, 0.045, -0.0061, 0.2127, 0.0472, -0.0061, 0.2127, 0.0487, -0.0083, 0.2127, 0.0472, -0.0083, 0.2127, 0.0487, -0.01, 0.2127, 0.0429, -0.01, 0.2127, 0.045, -0.01, 0.2127, 0.0472, -0.01, 0.2127, 0.0487, -0.0061, 0.2157, 0.0487, -0.0083, 0.2157, 0.0487, -0.0061, 0.2179, 0.0487, -0.0083, 0.2179, 0.0487, -0.01, 0.2157, 0.0429, -0.01, 0.2157, 0.045, -0.01, 0.2179, 0.0429, -0.01, 0.2179, 0.045, -0.01, 0.2157, 0.0472, -0.01, 0.2157, 0.0487, -0.01, 0.2179, 0.0472, -0.01, 0.2179, 0.0487, -0.01, 0.2201, 0.0342,
-0.01, 0.2201, 0.0363, -0.01, 0.2222, 0.0342, -0.01, 0.2222, 0.0363, -0.01, 0.2201, 0.0385, -0.01, 0.2201, 0.0407, -0.01, 0.2222, 0.0385, -0.01, 0.2222, 0.0407, -0.0061, 0.2245, 0.0342, -0.0061, 0.2245, 0.0363, -0.0083, 0.2245, 0.0342, -0.0083, 0.2245, 0.0363, -0.0061, 0.2245, 0.0385, -0.0061, 0.2245, 0.0407, -0.0083, 0.2245, 0.0385, -0.0083, 0.2245, 0.0407, -0.01, 0.2245, 0.0342, -0.01, 0.2245, 0.0363, -0.01, 0.2245, 0.0385, -0.01, 0.2245, 0.0407, -0.0061, 0.2201, 0.0487, -0.0083, 0.2201, 0.0487, -0.0061, 0.2222, 0.0487, -0.0083, 0.2222, 0.0487, -0.01, 0.2201, 0.0429, -0.01, 0.2201, 0.045, -0.01, 0.2222, 0.0429, -0.01, 0.2222, 0.045, -0.01, 0.2201, 0.0472, -0.01, 0.2201, 0.0487, -0.01, 0.2222, 0.0472,
-0.01, 0.2222, 0.0487, -0.0061, 0.2245, 0.0429, -0.0061, 0.2245, 0.045, -0.0083, 0.2245, 0.0429, -0.0083, 0.2245, 0.045, -0.0061, 0.2245, 0.0472, -0.0061, 0.2245, 0.0487, -0.0083, 0.2245, 0.0472, -0.0083, 0.2245, 0.0487, -0.01, 0.2245, 0.0429, -0.01, 0.2245, 0.045, -0.01, 0.2245, 0.0472, -0.01, 0.2245, 0.0487, -0.0061, 0.2763, -0.0006, -0.0061, 0.2773, 0.0012, -0.0083, 0.2763, -0.0006, -0.0083, 0.2773, 0.0012, -0.0061, 0.2777, 0.0023, -0.0083, 0.2777, 0.0023, -0.0061, 0.2782, 0.0038, -0.0061, 0.2787, 0.006, -0.0083, 0.2782, 0.0038, -0.0083, 0.2787, 0.006, -0.01, 0.2765, -0.0003, -0.01, 0.2772, 0.0011, -0.01, 0.2787, -0.0005, -0.01, 0.2776, 0.002, -0.01, 0.2783, 0.004, -0.01, 0.2787, 0.0061, -0.0061, 0.2791, 0.0081,
-0.0061, 0.2793, 0.0103, -0.0083, 0.2791, 0.0081, -0.0083, 0.2793, 0.0103, -0.0061, 0.2794, 0.0125, -0.0061, 0.2794, 0.0146, -0.0083, 0.2794, 0.0125, -0.0083, 0.2794, 0.0146, -0.01, 0.2791, 0.0082, -0.01, 0.2793, 0.0103, -0.01, 0.2794, 0.0125, -0.01, 0.2794, 0.0146, -0.0061, 0.2794, 0.0168, -0.0061, 0.2794, 0.019, -0.0083, 0.2794, 0.0168, -0.0083, 0.2794, 0.019, -0.0061, 0.2794, 0.0212, -0.0061, 0.2794, 0.0233, -0.0083, 0.2794, 0.0212, -0.0083, 0.2794, 0.0233, -0.01, 0.2794, 0.0168, -0.01, 0.2794, 0.019, -0.01, 0.2794, 0.0212, -0.01, 0.2794, 0.0233, -0.0061, 0.2794, 0.0255, -0.0061, 0.2794, 0.0277, -0.0083, 0.2794, 0.0255, -0.0083, 0.2794, 0.0277, -0.0061, 0.2794, 0.0298, -0.0061, 0.2794, 0.032, -0.0083, 0.2794, 0.0298,
-0.0083, 0.2794, 0.032, -0.01, 0.2794, 0.0255, -0.01, 0.2794, 0.0277, -0.01, 0.2794, 0.0298, -0.01, 0.2794, 0.032, -0.0061, 0.2794, 0.0342, -0.0061, 0.2794, 0.0363, -0.0083, 0.2794, 0.0342, -0.0083, 0.2794, 0.0363, -0.0061, 0.2794, 0.0385, -0.0061, 0.2794, 0.0407, -0.0083, 0.2794, 0.0385, -0.0083, 0.2794, 0.0407, -0.01, 0.2794, 0.0342, -0.01, 0.2794, 0.0363, -0.01, 0.2794, 0.0385, -0.01, 0.2794, 0.0407, -0.0061, 0.2794, 0.0429, -0.0061, 0.2794, 0.045, -0.0083, 0.2794, 0.0429, -0.0083, 0.2794, 0.045, -0.0061, 0.2794, 0.0472, -0.0061, 0.2794, 0.0487, -0.0083, 0.2794, 0.0472, -0.0083, 0.2794, 0.0487, -0.01, 0.2794, 0.0429, -0.01, 0.2794, 0.045, -0.01, 0.2794, 0.0472, -0.01, 0.2794, 0.0487, -0.01, 0.2808, -0.0005,
-0.01, 0.2808, 0.0016, -0.01, 0.283, -0.0005, -0.01, 0.283, 0.0016, -0.01, 0.2808, 0.0038, -0.01, 0.2808, 0.006, -0.01, 0.283, 0.0038, -0.01, 0.283, 0.006, -0.01, 0.2852, -0.0005, -0.01, 0.2852, 0.0016, -0.01, 0.2873, -0.0005, -0.01, 0.2873, 0.0016, -0.01, 0.2852, 0.0038, -0.01, 0.2852, 0.006, -0.01, 0.2873, 0.0038, -0.01, 0.2873, 0.006, -0.01, 0.2808, 0.0081, -0.01, 0.2808, 0.0103, -0.01, 0.283, 0.0081, -0.01, 0.283, 0.0103, -0.01, 0.2808, 0.0125, -0.01, 0.2808, 0.0146, -0.01, 0.283, 0.0125, -0.01, 0.283, 0.0146, -0.01, 0.2852, 0.0081, -0.01, 0.2852, 0.0103, -0.01, 0.2873, 0.0081, -0.01, 0.2873, 0.0103, -0.01, 0.2852, 0.0125, -0.01, 0.2852, 0.0146, -0.01, 0.2873, 0.0125,
-0.01, 0.2873, 0.0146, -0.0061, 0.2897, -0.0006, -0.0061, 0.2901, 0.0016, -0.0083, 0.2897, -0.0006, -0.0083, 0.2901, 0.0016, -0.0061, 0.2903, 0.0038, -0.0061, 0.2906, 0.0051, -0.0083, 0.2903, 0.0038, -0.0083, 0.2906, 0.0051, -0.0061, 0.2907, 0.0062, -0.0083, 0.2907, 0.0062, -0.01, 0.2897, -0.0007, -0.01, 0.2901, 0.0015, -0.01, 0.2903, 0.0037, -0.01, 0.2906, 0.0057, -0.01, 0.2908, 0.0065, -0.0061, 0.291, 0.0081, -0.0061, 0.2911, 0.0103, -0.0083, 0.291, 0.0081, -0.0083, 0.2911, 0.0103, -0.0061, 0.2912, 0.0125, -0.0061, 0.2912, 0.0146, -0.0083, 0.2912, 0.0125, -0.0083, 0.2912, 0.0146, -0.01, 0.2895, 0.0081, -0.01, 0.2895, 0.0103, -0.01, 0.291, 0.0081, -0.01, 0.2911, 0.0103, -0.01, 0.2895, 0.0125, -0.01, 0.2895, 0.0146,
-0.01, 0.2912, 0.0125, -0.01, 0.2912, 0.0146, -0.01, 0.2808, 0.0168, -0.01, 0.2808, 0.019, -0.01, 0.283, 0.0168, -0.01, 0.283, 0.019, -0.01, 0.2808, 0.0212, -0.01, 0.2808, 0.0233, -0.01, 0.283, 0.0212, -0.01, 0.283, 0.0233, -0.01, 0.2852, 0.0168, -0.01, 0.2852, 0.019, -0.01, 0.2873, 0.0168, -0.01, 0.2873, 0.019, -0.01, 0.2852, 0.0212, -0.01, 0.2852, 0.0233, -0.01, 0.2873, 0.0212, -0.01, 0.2873, 0.0233, -0.01, 0.2808, 0.0255, -0.01, 0.2808, 0.0277, -0.01, 0.283, 0.0255, -0.01, 0.283, 0.0277, -0.01, 0.2808, 0.0298, -0.01, 0.2808, 0.032, -0.01, 0.283, 0.0298, -0.01, 0.283, 0.032, -0.01, 0.2852, 0.0255, -0.01, 0.2852, 0.0277, -0.01, 0.2873, 0.0255, -0.01, 0.2873, 0.0277,
-0.01, 0.2852, 0.0298, -0.01, 0.2852, 0.032, -0.01, 0.2873, 0.0298, -0.01, 0.2873, 0.032, -0.0061, 0.2912, 0.0168, -0.0061, 0.2912, 0.019, -0.0083, 0.2912, 0.0168, -0.0083, 0.2912, 0.019, -0.0061, 0.2912, 0.0212, -0.0061, 0.2912, 0.0233, -0.0083, 0.2912, 0.0212, -0.0083, 0.2912, 0.0233, -0.01, 0.2895, 0.0168, -0.01, 0.2895, 0.019, -0.01, 0.2912, 0.0168, -0.01, 0.2912, 0.019, -0.01, 0.2895, 0.0212, -0.01, 0.2895, 0.0233, -0.01, 0.2912, 0.0212, -0.01, 0.2912, 0.0233, -0.0061, 0.2912, 0.0255, -0.0061, 0.2912, 0.0277, -0.0083, 0.2912, 0.0255, -0.0083, 0.2912, 0.0277, -0.0061, 0.2912, 0.0298, -0.0061, 0.2912, 0.032, -0.0083, 0.2912, 0.0298, -0.0083, 0.2912, 0.032, -0.01, 0.2895, 0.0255, -0.01, 0.2895, 0.0277,
-0.01, 0.2912, 0.0255, -0.01, 0.2912, 0.0277, -0.01, 0.2895, 0.0298, -0.01, 0.2895, 0.032, -0.01, 0.2912, 0.0298, -0.01, 0.2912, 0.032, -0.0061, 0.3088, -0.0005, -0.0061, 0.3088, 0.0016, -0.0083, 0.3088, -0.0005, -0.0083, 0.3088, 0.0016, -0.0061, 0.3088, 0.0038, -0.0061, 0.3088, 0.006, -0.0083, 0.3088, 0.0038, -0.0083, 0.3088, 0.006, -0.01, 0.3088, -0.0005, -0.01, 0.3088, 0.0016, -0.01, 0.3088, 0.0038, -0.01, 0.3088, 0.006, -0.01, 0.3112, -0.0005, -0.01, 0.3112, 0.0016, -0.01, 0.3134, -0.0005, -0.01, 0.3134, 0.0016, -0.01, 0.3112, 0.0038, -0.01, 0.3112, 0.006, -0.01, 0.3134, 0.0038, -0.01, 0.3134, 0.006, -0.0061, 0.3088, 0.0081, -0.0061, 0.3088, 0.0103, -0.0083, 0.3088, 0.0081, -0.0083, 0.3088, 0.0103,
-0.0061, 0.3088, 0.0125, -0.0061, 0.3088, 0.0146, -0.0083, 0.3088, 0.0125, -0.0083, 0.3088, 0.0146, -0.01, 0.3088, 0.0081, -0.01, 0.3088, 0.0103, -0.01, 0.3088, 0.0125, -0.01, 0.3088, 0.0146, -0.01, 0.3112, 0.0081, -0.01, 0.3112, 0.0103, -0.01, 0.3134, 0.0081, -0.01, 0.3134, 0.0103, -0.01, 0.3112, 0.0125, -0.01, 0.3112, 0.0146, -0.01, 0.3134, 0.0125, -0.01, 0.3134, 0.0146, -0.0061, 0.3088, 0.0168, -0.0061, 0.3088, 0.019, -0.0083, 0.3088, 0.0168, -0.0083, 0.3088, 0.019, -0.0061, 0.3088, 0.0212, -0.0061, 0.3088, 0.0233, -0.0083, 0.3088, 0.0212, -0.0083, 0.3088, 0.0233, -0.01, 0.3088, 0.0168, -0.01, 0.3088, 0.019, -0.01, 0.3088, 0.0212, -0.01, 0.3088, 0.0233, -0.01, 0.3112, 0.0168, -0.01, 0.3112, 0.019,
-0.01, 0.3134, 0.0168, -0.01, 0.3134, 0.019, -0.01, 0.3112, 0.0212, -0.01, 0.3112, 0.0233, -0.01, 0.3134, 0.0212, -0.01, 0.3134, 0.0233, -0.0061, 0.3088, 0.0255, -0.0061, 0.3088, 0.0277, -0.0083, 0.3088, 0.0255, -0.0083, 0.3088, 0.0277, -0.0061, 0.3088, 0.0298, -0.0061, 0.3088, 0.032, -0.0083, 0.3088, 0.0298, -0.0083, 0.3088, 0.032, -0.01, 0.3088, 0.0255, -0.01, 0.3088, 0.0277, -0.01, 0.3088, 0.0298, -0.01, 0.3088, 0.032, -0.01, 0.3112, 0.0255, -0.01, 0.3112, 0.0277, -0.01, 0.3134, 0.0255, -0.01, 0.3134, 0.0277, -0.01, 0.3112, 0.0298, -0.01, 0.3112, 0.032, -0.01, 0.3134, 0.0298, -0.01, 0.3134, 0.032, -0.01, 0.2808, 0.0342, -0.01, 0.2808, 0.0363, -0.01, 0.283, 0.0342, -0.01, 0.283, 0.0363,
-0.01, 0.2808, 0.0385, -0.01, 0.2808, 0.0407, -0.01, 0.283, 0.0385, -0.01, 0.283, 0.0407, -0.01, 0.2852, 0.0342, -0.01, 0.2852, 0.0363, -0.01, 0.2873, 0.0342, -0.01, 0.2873, 0.0363, -0.01, 0.2852, 0.0385, -0.01, 0.2852, 0.0407, -0.01, 0.2873, 0.0385, -0.01, 0.2873, 0.0407, -0.0061, 0.2808, 0.0487, -0.0083, 0.2808, 0.0487, -0.0061, 0.283, 0.0487, -0.0083, 0.283, 0.0487, -0.01, 0.2808, 0.0429, -0.01, 0.2808, 0.045, -0.01, 0.283, 0.0429, -0.01, 0.283, 0.045, -0.01, 0.2808, 0.0472, -0.01, 0.2808, 0.0487, -0.01, 0.283, 0.0472, -0.01, 0.283, 0.0487, -0.0061, 0.2852, 0.0487, -0.0083, 0.2852, 0.0487, -0.0061, 0.2873, 0.0487, -0.0083, 0.2873, 0.0487, -0.01, 0.2852, 0.0429, -0.01, 0.2852, 0.045,
-0.01, 0.2873, 0.0429, -0.01, 0.2873, 0.045, -0.01, 0.2852, 0.0472, -0.01, 0.2852, 0.0487, -0.01, 0.2873, 0.0472, -0.01, 0.2873, 0.0487, -0.0061, 0.2912, 0.0342, -0.0061, 0.2912, 0.0363, -0.0083, 0.2912, 0.0342, -0.0083, 0.2912, 0.0363, -0.0061, 0.2912, 0.0385, -0.0061, 0.2912, 0.0407, -0.0083, 0.2912, 0.0385, -0.0083, 0.2912, 0.0407, -0.01, 0.2895, 0.0342, -0.01, 0.2895, 0.0363, -0.01, 0.2912, 0.0342, -0.01, 0.2912, 0.0363, -0.01, 0.2895, 0.0385, -0.01, 0.2895, 0.0407, -0.01, 0.2912, 0.0385, -0.01, 0.2912, 0.0407, -0.0061, 0.2912, 0.0429, -0.0061, 0.2912, 0.045, -0.0083, 0.2912, 0.0429, -0.0083, 0.2912, 0.045, -0.0061, 0.2895, 0.0487, -0.0083, 0.2895, 0.0487, -0.0061, 0.2912, 0.0472, -0.0061, 0.2912, 0.0487,
-0.0083, 0.2912, 0.0472, -0.0083, 0.2912, 0.0487, -0.01, 0.2895, 0.0429, -0.01, 0.2895, 0.045, -0.01, 0.2912, 0.0429, -0.01, 0.2912, 0.045, -0.01, 0.2895, 0.0472, -0.01, 0.2895, 0.0487, -0.01, 0.2912, 0.0472, -0.01, 0.2912, 0.0487, -0.0061, 0.3088, 0.0342, -0.0061, 0.3088, 0.0363, -0.0083, 0.3088, 0.0342, -0.0083, 0.3088, 0.0363, -0.0061, 0.3088, 0.0385, -0.0061, 0.3088, 0.0407, -0.0083, 0.3088, 0.0385, -0.0083, 0.3088, 0.0407, -0.01, 0.3088, 0.0342, -0.01, 0.3088, 0.0363, -0.01, 0.3088, 0.0385, -0.01, 0.3088, 0.0407, -0.01, 0.3112, 0.0342, -0.01, 0.3112, 0.0363, -0.01, 0.3134, 0.0342, -0.01, 0.3134, 0.0363, -0.01, 0.3112, 0.0385, -0.01, 0.3112, 0.0407, -0.01, 0.3134, 0.0385, -0.01, 0.3134, 0.0407,
-0.0061, 0.3088, 0.0429, -0.0061, 0.3088, 0.045, -0.0083, 0.3088, 0.0429, -0.0083, 0.3088, 0.045, -0.0061, 0.3088, 0.0472, -0.0061, 0.3088, 0.0487, -0.0083, 0.3088, 0.0472, -0.0083, 0.3088, 0.0487, -0.01, 0.3088, 0.0429, -0.01, 0.3088, 0.045, -0.01, 0.3088, 0.0472, -0.01, 0.3088, 0.0487, -0.0061, 0.3112, 0.0487, -0.0083, 0.3112, 0.0487, -0.0061, 0.3134, 0.0487, -0.0083, 0.3134, 0.0487, -0.01, 0.3112, 0.0429, -0.01, 0.3112, 0.045, -0.01, 0.3134, 0.0429, -0.01, 0.3134, 0.045, -0.01, 0.3112, 0.0472, -0.01, 0.3112, 0.0487, -0.01, 0.3134, 0.0472, -0.01, 0.3134, 0.0487, -0.01, 0.3155, -0.0005, -0.01, 0.3155, 0.0016, -0.01, 0.3177, -0.0005, -0.01, 0.3177, 0.0016, -0.01, 0.3155, 0.0038, -0.01, 0.3155, 0.006,
-0.01, 0.3177, 0.0038, -0.01, 0.3177, 0.006, -0.0061, 0.3206, -0.0005, -0.0061, 0.3206, 0.0016, -0.0083, 0.3206, -0.0005, -0.0083, 0.3206, 0.0016, -0.0061, 0.3206, 0.0038, -0.0061, 0.3206, 0.006, -0.0083, 0.3206, 0.0038, -0.0083, 0.3206, 0.006, -0.01, 0.3206, -0.0005, -0.01, 0.3206, 0.0016, -0.01, 0.3206, 0.0038, -0.01, 0.3206, 0.006, -0.01, 0.3155, 0.0081, -0.01, 0.3155, 0.0103, -0.01, 0.3177, 0.0081, -0.01, 0.3177, 0.0103, -0.01, 0.3155, 0.0125, -0.01, 0.3155, 0.0146, -0.01, 0.3177, 0.0125, -0.01, 0.3177, 0.0146, -0.0061, 0.3206, 0.0081, -0.0061, 0.3206, 0.0103, -0.0083, 0.3206, 0.0081, -0.0083, 0.3206, 0.0103, -0.0061, 0.3206, 0.0125, -0.0061, 0.3206, 0.0146, -0.0083, 0.3206, 0.0125, -0.0083, 0.3206, 0.0146,
-0.01, 0.3206, 0.0081, -0.01, 0.3206, 0.0103, -0.01, 0.3206, 0.0125, -0.01, 0.3206, 0.0146, -0.01, 0.3155, 0.0168, -0.01, 0.3155, 0.019, -0.01, 0.3177, 0.0168, -0.01, 0.3177, 0.019, -0.01, 0.3155, 0.0212, -0.01, 0.3155, 0.0233, -0.01, 0.3177, 0.0212, -0.01, 0.3177, 0.0233, -0.0061, 0.3206, 0.0168, -0.0061, 0.3206, 0.019, -0.0083, 0.3206, 0.0168, -0.0083, 0.3206, 0.019, -0.0061, 0.3206, 0.0212, -0.0061, 0.3206, 0.0233, -0.0083, 0.3206, 0.0212, -0.0083, 0.3206, 0.0233, -0.01, 0.3206, 0.0168, -0.01, 0.3206, 0.019, -0.01, 0.3206, 0.0212, -0.01, 0.3206, 0.0233, -0.01, 0.3155, 0.0255, -0.01, 0.3155, 0.0277, -0.01, 0.3177, 0.0255, -0.01, 0.3177, 0.0277, -0.01, 0.3155, 0.0298, -0.01, 0.3155, 0.032,
-0.01, 0.3177, 0.0298, -0.01, 0.3177, 0.032, -0.0061, 0.3206, 0.0255, -0.0061, 0.3206, 0.0277, -0.0083, 0.3206, 0.0255, -0.0083, 0.3206, 0.0277, -0.0061, 0.3206, 0.0298, -0.0061, 0.3206, 0.032, -0.0083, 0.3206, 0.0298, -0.0083, 0.3206, 0.032, -0.01, 0.3206, 0.0255, -0.01, 0.3206, 0.0277, -0.01, 0.3206, 0.0298, -0.01, 0.3206, 0.032, -0.0061, 0.3402, -0.0005, -0.0061, 0.3402, 0.0016, -0.0083, 0.3402, -0.0005, -0.0083, 0.3402, 0.0016, -0.0061, 0.3402, 0.0038, -0.0061, 0.3402, 0.006, -0.0083, 0.3402, 0.0038, -0.0083, 0.3402, 0.006, -0.01, 0.3402, -0.0005, -0.01, 0.3402, 0.0016, -0.01, 0.3402, 0.0038, -0.01, 0.3402, 0.006, -0.0061, 0.3402, 0.0081, -0.0061, 0.3402, 0.0103, -0.0083, 0.3402, 0.0081, -0.0083, 0.3402, 0.0103,
-0.0061, 0.3402, 0.0125, -0.0061, 0.3402, 0.0146, -0.0083, 0.3402, 0.0125, -0.0083, 0.3402, 0.0146, -0.01, 0.3402, 0.0081, -0.01, 0.3402, 0.0103, -0.01, 0.3402, 0.0125, -0.01, 0.3402, 0.0146, -0.01, 0.3416, -0.0005, -0.01, 0.3416, 0.0016, -0.01, 0.3438, -0.0005, -0.01, 0.3438, 0.0016, -0.01, 0.3416, 0.0038, -0.01, 0.3416, 0.006, -0.01, 0.3438, 0.0038, -0.01, 0.3438, 0.006, -0.01, 0.3459, -0.0005, -0.01, 0.3459, 0.0016, -0.01, 0.3481, -0.0005, -0.01, 0.3481, 0.0016, -0.01, 0.3459, 0.0038, -0.01, 0.3459, 0.006, -0.01, 0.3481, 0.0038, -0.01, 0.3481, 0.006, -0.01, 0.3416, 0.0081, -0.01, 0.3416, 0.0103, -0.01, 0.3438, 0.0081, -0.01, 0.3438, 0.0103, -0.01, 0.3416, 0.0125, -0.01, 0.3416, 0.0146,
-0.01, 0.3438, 0.0125, -0.01, 0.3438, 0.0146, -0.01, 0.3459, 0.0081, -0.01, 0.3459, 0.0103, -0.01, 0.3481, 0.0081, -0.01, 0.3481, 0.0103, -0.01, 0.3459, 0.0125, -0.01, 0.3459, 0.0146, -0.01, 0.3481, 0.0125, -0.01, 0.3481, 0.0146, -0.0061, 0.3402, 0.0168, -0.0061, 0.3402, 0.019, -0.0083, 0.3402, 0.0168, -0.0083, 0.3402, 0.019, -0.0061, 0.3402, 0.0212, -0.0061, 0.3402, 0.0233, -0.0083, 0.3402, 0.0212, -0.0083, 0.3402, 0.0233, -0.01, 0.3402, 0.0168, -0.01, 0.3402, 0.019, -0.01, 0.3402, 0.0212, -0.01, 0.3402, 0.0233, -0.0061, 0.3402, 0.0255, -0.0061, 0.3402, 0.0277, -0.0083, 0.3402, 0.0255, -0.0083, 0.3402, 0.0277, -0.0061, 0.3402, 0.0298, -0.0061, 0.3402, 0.032, -0.0083, 0.3402, 0.0298, -0.0083, 0.3402, 0.032,
-0.01, 0.3402, 0.0255, -0.01, 0.3402, 0.0277, -0.01, 0.3402, 0.0298, -0.01, 0.3402, 0.032, -0.01, 0.3416, 0.0168, -0.01, 0.3416, 0.019, -0.01, 0.3438, 0.0168, -0.01, 0.3438, 0.019, -0.01, 0.3416, 0.0212, -0.01, 0.3416, 0.0233, -0.01, 0.3438, 0.0212, -0.01, 0.3438, 0.0233, -0.01, 0.3459, 0.0168, -0.01, 0.3459, 0.019, -0.01, 0.3481, 0.0168, -0.01, 0.3481, 0.019, -0.01, 0.3459, 0.0212, -0.01, 0.3459, 0.0233, -0.01, 0.3481, 0.0212, -0.01, 0.3481, 0.0233, -0.01, 0.3416, 0.0255, -0.01, 0.3416, 0.0277, -0.01, 0.3438, 0.0255, -0.01, 0.3438, 0.0277, -0.01, 0.3416, 0.0298, -0.01, 0.3416, 0.032, -0.01, 0.3438, 0.0298, -0.01, 0.3438, 0.032, -0.01, 0.3459, 0.0255, -0.01, 0.3459, 0.0277,
-0.01, 0.3481, 0.0255, -0.01, 0.3481, 0.0277, -0.01, 0.3459, 0.0298, -0.01, 0.3459, 0.032, -0.01, 0.3481, 0.0298, -0.01, 0.3481, 0.032, -0.01, 0.3155, 0.0342, -0.01, 0.3155, 0.0363, -0.01, 0.3177, 0.0342, -0.01, 0.3177, 0.0363, -0.01, 0.3155, 0.0385, -0.01, 0.3155, 0.0407, -0.01, 0.3177, 0.0385, -0.01, 0.3177, 0.0407, -0.0061, 0.3206, 0.0342, -0.0061, 0.3206, 0.0363, -0.0083, 0.3206, 0.0342, -0.0083, 0.3206, 0.0363, -0.0061, 0.3206, 0.0385, -0.0061, 0.3206, 0.0407, -0.0083, 0.3206, 0.0385, -0.0083, 0.3206, 0.0407, -0.01, 0.3206, 0.0342, -0.01, 0.3206, 0.0363, -0.01, 0.3206, 0.0385, -0.01, 0.3206, 0.0407, -0.0061, 0.3155, 0.0487, -0.0083, 0.3155, 0.0487, -0.0061, 0.3177, 0.0487, -0.0083, 0.3177, 0.0487,
-0.01, 0.3155, 0.0429, -0.01, 0.3155, 0.045, -0.01, 0.3177, 0.0429, -0.01, 0.3177, 0.045, -0.01, 0.3155, 0.0472, -0.01, 0.3155, 0.0487, -0.01, 0.3177, 0.0472, -0.01, 0.3177, 0.0487, -0.0061, 0.3206, 0.0429, -0.0061, 0.3206, 0.045, -0.0083, 0.3206, 0.0429, -0.0083, 0.3206, 0.045, -0.0061, 0.3206, 0.0472, -0.0061, 0.3206, 0.0487, -0.0083, 0.3206, 0.0472, -0.0083, 0.3206, 0.0487, -0.01, 0.3206, 0.0429, -0.01, 0.3206, 0.045, -0.01, 0.3206, 0.0472, -0.01, 0.3206, 0.0487, -0.0061, 0.3402, 0.0342, -0.0061, 0.3402, 0.0363, -0.0083, 0.3402, 0.0342, -0.0083, 0.3402, 0.0363, -0.0061, 0.3402, 0.0385, -0.0061, 0.3402, 0.0407, -0.0083, 0.3402, 0.0385, -0.0083, 0.3402, 0.0407, -0.01, 0.3402, 0.0342, -0.01, 0.3402, 0.0363,
-0.01, 0.3402, 0.0385, -0.01, 0.3402, 0.0407, -0.0061, 0.3402, 0.0429, -0.0061, 0.3402, 0.045, -0.0083, 0.3402, 0.0429, -0.0083, 0.3402, 0.045, -0.0061, 0.3402, 0.0472, -0.0061, 0.3402, 0.0487, -0.0083, 0.3402, 0.0472, -0.0083, 0.3402, 0.0487, -0.01, 0.3402, 0.0429, -0.01, 0.3402, 0.045, -0.01, 0.3402, 0.0472, -0.01, 0.3402, 0.0487, -0.01, 0.3416, 0.0342, -0.01, 0.3416, 0.0363, -0.01, 0.3438, 0.0342, -0.01, 0.3438, 0.0363, -0.01, 0.3416, 0.0385, -0.01, 0.3416, 0.0407, -0.01, 0.3438, 0.0385, -0.01, 0.3438, 0.0407, -0.01, 0.3459, 0.0342, -0.01, 0.3459, 0.0363, -0.01, 0.3481, 0.0342, -0.01, 0.3481, 0.0363, -0.01, 0.3459, 0.0385, -0.01, 0.3459, 0.0407, -0.01, 0.3481, 0.0385, -0.01, 0.3481, 0.0407,
-0.0061, 0.3416, 0.0487, -0.0083, 0.3416, 0.0487, -0.0061, 0.3438, 0.0487, -0.0083, 0.3438, 0.0487, -0.01, 0.3416, 0.0429, -0.01, 0.3416, 0.045, -0.01, 0.3438, 0.0429, -0.01, 0.3438, 0.045, -0.01, 0.3416, 0.0472, -0.01, 0.3416, 0.0487, -0.01, 0.3438, 0.0472, -0.01, 0.3438, 0.0487, -0.0061, 0.3459, 0.0487, -0.0083, 0.3459, 0.0487, -0.0061, 0.3481, 0.0487, -0.0083, 0.3481, 0.0487, -0.01, 0.3459, 0.0429, -0.01, 0.3459, 0.045, -0.01, 0.3481, 0.0429, -0.01, 0.3481, 0.045, -0.01, 0.3459, 0.0472, -0.01, 0.3459, 0.0487, -0.01, 0.3481, 0.0472, -0.01, 0.3481, 0.0487, -0.0061, 0.352, -0.0005, -0.0061, 0.352, 0.0016, -0.0083, 0.352, -0.0005, -0.0083, 0.352, 0.0016, -0.0061, 0.352, 0.0038, -0.0061, 0.352, 0.006,
-0.0083, 0.352, 0.0038, -0.0083, 0.352, 0.006, -0.01, 0.3503, -0.0005, -0.01, 0.3503, 0.0016, -0.01, 0.352, -0.0005, -0.01, 0.352, 0.0016, -0.01, 0.3503, 0.0038, -0.01, 0.3503, 0.006, -0.01, 0.352, 0.0038, -0.01, 0.352, 0.006, -0.0061, 0.352, 0.0081, -0.0061, 0.352, 0.0103, -0.0083, 0.352, 0.0081, -0.0083, 0.352, 0.0103, -0.0061, 0.352, 0.0125, -0.0061, 0.352, 0.0146, -0.0083, 0.352, 0.0125, -0.0083, 0.352, 0.0146, -0.01, 0.3503, 0.0081, -0.01, 0.3503, 0.0103, -0.01, 0.352, 0.0081, -0.01, 0.352, 0.0103, -0.01, 0.3503, 0.0125, -0.01, 0.3503, 0.0146, -0.01, 0.352, 0.0125, -0.01, 0.352, 0.0146, -0.0061, 0.352, 0.0168, -0.0061, 0.352, 0.019, -0.0083, 0.352, 0.0168, -0.0083, 0.352, 0.019,
-0.0061, 0.352, 0.0212, -0.0061, 0.352, 0.0233, -0.0083, 0.352, 0.0212, -0.0083, 0.352, 0.0233, -0.01, 0.3503, 0.0168, -0.01, 0.3503, 0.019, -0.01, 0.352, 0.0168, -0.01, 0.352, 0.019, -0.01, 0.3503, 0.0212, -0.01, 0.3503, 0.0233, -0.01, 0.352, 0.0212, -0.01, 0.352, 0.0233, -0.0061, 0.352, 0.0255, -0.0061, 0.352, 0.0277, -0.0083, 0.352, 0.0255, -0.0083, 0.352, 0.0277, -0.0061, 0.352, 0.0298, -0.0061, 0.352, 0.032, -0.0083, 0.352, 0.0298, -0.0083, 0.352, 0.032, -0.01, 0.3503, 0.0255, -0.01, 0.3503, 0.0277, -0.01, 0.352, 0.0255, -0.01, 0.352, 0.0277, -0.01, 0.3503, 0.0298, -0.01, 0.3503, 0.032, -0.01, 0.352, 0.0298, -0.01, 0.352, 0.032, -0.0061, 0.352, 0.0342, -0.0061, 0.352, 0.0363,
-0.0083, 0.352, 0.0342, -0.0083, 0.352, 0.0363, -0.0061, 0.352, 0.0379, -0.0083, 0.352, 0.0379, -0.01, 0.3503, 0.0342, -0.01, 0.3503, 0.0363, -0.01, 0.352, 0.0342, -0.01, 0.352, 0.0363, -0.01, 0.3503, 0.0385, -0.01, 0.3503, 0.0407, -0.01, 0.352, 0.0379, -0.01, 0.3524, 0.0407, -0.0061, 0.3546, 0.0379, -0.0083, 0.3546, 0.0379, -0.0061, 0.3568, 0.0379, -0.0083, 0.3568, 0.0379, -0.01, 0.3546, 0.0379, -0.01, 0.3546, 0.0407, -0.01, 0.3568, 0.0379, -0.01, 0.3568, 0.0407, -0.0061, 0.3503, 0.0487, -0.0083, 0.3503, 0.0487, -0.0061, 0.3524, 0.0487, -0.0083, 0.3524, 0.0487, -0.01, 0.3503, 0.0429, -0.01, 0.3503, 0.045, -0.01, 0.3524, 0.0429, -0.01, 0.3524, 0.045, -0.01, 0.3503, 0.0472, -0.01, 0.3503, 0.0487,
-0.01, 0.3524, 0.0472, -0.01, 0.3524, 0.0487, -0.0061, 0.3546, 0.0487, -0.0083, 0.3546, 0.0487, -0.0061, 0.3568, 0.0487, -0.0083, 0.3568, 0.0487, -0.01, 0.3546, 0.0429, -0.01, 0.3546, 0.045, -0.01, 0.3568, 0.0429, -0.01, 0.3568, 0.045, -0.01, 0.3546, 0.0472, -0.01, 0.3546, 0.0487, -0.01, 0.3568, 0.0472, -0.01, 0.3568, 0.0487, -0.0061, 0.3589, 0.0379, -0.0083, 0.3589, 0.0379, -0.0061, 0.3611, 0.0379, -0.0083, 0.3611, 0.0379, -0.01, 0.3589, 0.0379, -0.01, 0.3589, 0.0407, -0.01, 0.3611, 0.0379, -0.01, 0.3611, 0.0407, -0.0061, 0.3633, 0.0379, -0.0083, 0.3633, 0.0379, -0.0061, 0.3655, 0.0379, -0.0083, 0.3655, 0.0379, -0.01, 0.3633, 0.0379, -0.01, 0.3633, 0.0407, -0.01, 0.3655, 0.0379, -0.01, 0.3655, 0.0407,
-0.0061, 0.3589, 0.0487, -0.0083, 0.3589, 0.0487, -0.0061, 0.3611, 0.0487, -0.0083, 0.3611, 0.0487, -0.01, 0.3589, 0.0429, -0.01, 0.3589, 0.045, -0.01, 0.3611, 0.0429, -0.01, 0.3611, 0.045, -0.01, 0.3589, 0.0472, -0.01, 0.3589, 0.0487, -0.01, 0.3611, 0.0472, -0.01, 0.3611, 0.0487, -0.0061, 0.3633, 0.0487, -0.0083, 0.3633, 0.0487, -0.0061, 0.3655, 0.0487, -0.0083, 0.3655, 0.0487, -0.01, 0.3633, 0.0429, -0.01, 0.3633, 0.045, -0.01, 0.3655, 0.0429, -0.01, 0.3655, 0.045, -0.01, 0.3633, 0.0472, -0.01, 0.3633, 0.0487, -0.01, 0.3655, 0.0472, -0.01, 0.3655, 0.0487, -0.0061, 0.3676, 0.0379, -0.0083, 0.3676, 0.0379, -0.0061, 0.3698, 0.0379, -0.0083, 0.3698, 0.0379, -0.01, 0.3676, 0.0379, -0.01, 0.3676, 0.0407,
-0.01, 0.3698, 0.0379, -0.01, 0.3698, 0.0407, -0.0061, 0.372, 0.0379, -0.0083, 0.372, 0.0379, -0.0061, 0.3741, 0.0379, -0.0083, 0.3741, 0.0379, -0.01, 0.372, 0.0379, -0.01, 0.372, 0.0407, -0.01, 0.3741, 0.0379, -0.01, 0.3741, 0.0407, -0.0061, 0.3676, 0.0487, -0.0083, 0.3676, 0.0487, -0.0061, 0.3698, 0.0487, -0.0083, 0.3698, 0.0487, -0.01, 0.3676, 0.0429, -0.01, 0.3676, 0.045, -0.01, 0.3698, 0.0429, -0.01, 0.3698, 0.045, -0.01, 0.3676, 0.0472, -0.01, 0.3676, 0.0487, -0.01, 0.3698, 0.0472, -0.01, 0.3698, 0.0487, -0.0061, 0.372, 0.0487, -0.0083, 0.372, 0.0487, -0.0061, 0.3741, 0.0487, -0.0083, 0.3741, 0.0487, -0.01, 0.372, 0.0429, -0.01, 0.372, 0.045, -0.01, 0.3741, 0.0429, -0.01, 0.3741, 0.045,
-0.01, 0.372, 0.0472, -0.01, 0.372, 0.0487, -0.01, 0.3741, 0.0472, -0.01, 0.3741, 0.0487, -0.0061, 0.3763, 0.0379, -0.0083, 0.3763, 0.0379, -0.0061, 0.3785, 0.0379, -0.0083, 0.3785, 0.0379, -0.01, 0.3763, 0.0379, -0.01, 0.3763, 0.0407, -0.01, 0.3784, 0.0379, -0.01, 0.3785, 0.0407, -0.0061, 0.3833, 0.0373, -0.0083, 0.3833, 0.0373, -0.0061, 0.3807, 0.0377, -0.0083, 0.3807, 0.0377, -0.0061, 0.3822, 0.0375, -0.0083, 0.3822, 0.0375, -0.01, 0.3835, 0.0373, -0.01, 0.3806, 0.0377, -0.01, 0.3807, 0.0407, -0.01, 0.3827, 0.0374, -0.01, 0.3828, 0.0407, -0.0061, 0.3763, 0.0487, -0.0083, 0.3763, 0.0487, -0.0061, 0.3785, 0.0487, -0.0083, 0.3785, 0.0487, -0.01, 0.3763, 0.0429, -0.01, 0.3763, 0.045, -0.01, 0.3785, 0.0429,
-0.01, 0.3785, 0.045, -0.01, 0.3763, 0.0472, -0.01, 0.3763, 0.0487, -0.01, 0.3785, 0.0472, -0.01, 0.3785, 0.0487, -0.0061, 0.3807, 0.0487, -0.0083, 0.3807, 0.0487, -0.0061, 0.3828, 0.0487, -0.0083, 0.3828, 0.0487, -0.01, 0.3807, 0.0429, -0.01, 0.3807, 0.045, -0.01, 0.3828, 0.0429, -0.01, 0.3828, 0.045, -0.01, 0.3807, 0.0472, -0.01, 0.3807, 0.0487, -0.01, 0.3828, 0.0472, -0.01, 0.3828, 0.0487, -0.0061, 0.3999, -0.0006, -0.0061, 0.401, 0.0011, -0.0083, 0.3999, -0.0006, -0.0083, 0.401, 0.0011, -0.01, 0.4002, -0.0003, -0.01, 0.4009, 0.001, -0.0061, 0.3947, 0.033, -0.0083, 0.3947, 0.033, -0.0061, 0.3959, 0.0321, -0.0083, 0.3959, 0.0321, -0.01, 0.3947, 0.033, -0.01, 0.3956, 0.0324, -0.0061, 0.4009, 0.026,
-0.0061, 0.3999, 0.0277, -0.0083, 0.4009, 0.026, -0.0083, 0.3999, 0.0277, -0.0061, 0.3982, 0.0299, -0.0061, 0.3971, 0.031, -0.0083, 0.3982, 0.0299, -0.0083, 0.3971, 0.031, -0.0061, 0.3991, 0.0288, -0.0083, 0.3991, 0.0288, -0.01, 0.4009, 0.0261, -0.01, 0.4001, 0.0274, -0.01, 0.3982, 0.03, -0.01, 0.3971, 0.031, -0.01, 0.399, 0.0289, -0.01, 0.4002, 0.032, -0.0061, 0.4016, 0.0021, -0.0083, 0.4016, 0.0021, -0.0061, 0.4024, 0.0038, -0.0061, 0.4032, 0.0058, -0.0083, 0.4024, 0.0038, -0.0083, 0.4032, 0.0058, -0.0061, 0.4035, 0.0069, -0.0083, 0.4035, 0.0069, -0.01, 0.4024, -0.0005, -0.01, 0.4015, 0.002, -0.01, 0.4045, -0.0005, -0.01, 0.4045, 0.0016, -0.01, 0.4025, 0.004, -0.01, 0.4031, 0.0056, -0.01, 0.4045, 0.0038,
-0.01, 0.4033, 0.0064, -0.01, 0.4067, -0.0005, -0.01, 0.4067, 0.0016, -0.01, 0.4089, -0.0005, -0.01, 0.4089, 0.0016, -0.01, 0.4067, 0.0038, -0.01, 0.4067, 0.006, -0.01, 0.4089, 0.0038, -0.01, 0.4089, 0.006, -0.0061, 0.4037, 0.0081, -0.0061, 0.4039, 0.0103, -0.0083, 0.4037, 0.0081, -0.0083, 0.4039, 0.0103, -0.0061, 0.4039, 0.0125, -0.0061, 0.4039, 0.0146, -0.0083, 0.4039, 0.0125, -0.0083, 0.4039, 0.0146, -0.01, 0.4038, 0.0083, -0.01, 0.4039, 0.0104, -0.01, 0.4039, 0.0125, -0.01, 0.4039, 0.0146, -0.01, 0.4067, 0.0081, -0.01, 0.4067, 0.0103, -0.01, 0.4089, 0.0081, -0.01, 0.4089, 0.0103, -0.01, 0.4067, 0.0125, -0.01, 0.4067, 0.0146, -0.01, 0.4089, 0.0125, -0.01, 0.4089, 0.0146, -0.0061, 0.4132, -0.0006,
-0.0061, 0.414, 0.0015, -0.0083, 0.4132, -0.0006, -0.0083, 0.414, 0.0015, -0.01, 0.411, -0.0005, -0.01, 0.411, 0.0016, -0.01, 0.4131, -0.0007, -0.01, 0.414, 0.0015, -0.01, 0.411, 0.0038, -0.01, 0.411, 0.006, -0.01, 0.4132, 0.0038, -0.01, 0.4132, 0.006, -0.0061, 0.4143, 0.0026, -0.0083, 0.4143, 0.0026, -0.0061, 0.4147, 0.0038, -0.0061, 0.4153, 0.006, -0.0083, 0.4147, 0.0038, -0.0083, 0.4153, 0.006, -0.01, 0.4144, 0.0026, -0.01, 0.4147, 0.0037, -0.01, 0.4152, 0.0059, -0.01, 0.411, 0.0081, -0.01, 0.411, 0.0103, -0.01, 0.4132, 0.0081, -0.01, 0.4132, 0.0103, -0.01, 0.411, 0.0125, -0.01, 0.411, 0.0146, -0.01, 0.4132, 0.0125, -0.01, 0.4132, 0.0146, -0.0061, 0.4156, 0.0081, -0.0061, 0.4157, 0.0103,
-0.0083, 0.4156, 0.0081, -0.0083, 0.4157, 0.0103, -0.0061, 0.4157, 0.0125, -0.0061, 0.4157, 0.0146, -0.0083, 0.4157, 0.0125, -0.0083, 0.4157, 0.0146, -0.01, 0.4156, 0.0081, -0.01, 0.4157, 0.0103, -0.01, 0.4157, 0.0125, -0.01, 0.4157, 0.0146, -0.0061, 0.4039, 0.0168, -0.0061, 0.4037, 0.019, -0.0083, 0.4039, 0.0168, -0.0083, 0.4037, 0.019, -0.0061, 0.4032, 0.0212, -0.0061, 0.4024, 0.0234, -0.0083, 0.4032, 0.0212, -0.0083, 0.4024, 0.0234, -0.0061, 0.4034, 0.0201, -0.0083, 0.4034, 0.0201, -0.01, 0.4039, 0.0168, -0.01, 0.4037, 0.0188, -0.01, 0.4031, 0.0215, -0.01, 0.4025, 0.0231, -0.01, 0.4033, 0.0207, -0.01, 0.4045, 0.0233, -0.01, 0.4067, 0.0168, -0.01, 0.4067, 0.019, -0.01, 0.4089, 0.0168, -0.01, 0.4089, 0.019,
-0.01, 0.4067, 0.0212, -0.01, 0.4067, 0.0233, -0.01, 0.4089, 0.0212, -0.01, 0.4089, 0.0233, -0.0061, 0.4016, 0.025, -0.0083, 0.4016, 0.025, -0.01, 0.4015, 0.0251, -0.01, 0.4024, 0.0277, -0.01, 0.4045, 0.0255, -0.01, 0.4045, 0.0277, -0.01, 0.4024, 0.0298, -0.01, 0.4024, 0.032, -0.01, 0.4045, 0.0298, -0.01, 0.4045, 0.032, -0.01, 0.4067, 0.0255, -0.01, 0.4067, 0.0277, -0.01, 0.4089, 0.0255, -0.01, 0.4089, 0.0277, -0.01, 0.4067, 0.0298, -0.01, 0.4067, 0.032, -0.01, 0.4089, 0.0298, -0.01, 0.4089, 0.032, -0.01, 0.411, 0.0168, -0.01, 0.411, 0.019, -0.01, 0.4132, 0.0168, -0.01, 0.4132, 0.019, -0.01, 0.411, 0.0212, -0.01, 0.411, 0.0233, -0.01, 0.4132, 0.0212, -0.01, 0.4132, 0.0233,
-0.0061, 0.4157, 0.0168, -0.0061, 0.4157, 0.019, -0.0083, 0.4157, 0.0168, -0.0083, 0.4157, 0.019, -0.0061, 0.4155, 0.0212, -0.0061, 0.4152, 0.0233, -0.0083, 0.4155, 0.0212, -0.0083, 0.4152, 0.0233, -0.01, 0.4157, 0.0168, -0.01, 0.4157, 0.019, -0.01, 0.4155, 0.0212, -0.01, 0.4152, 0.0234, -0.0061, 0.4142, 0.0265, -0.0061, 0.4138, 0.0277, -0.0083, 0.4142, 0.0265, -0.0083, 0.4138, 0.0277, -0.0061, 0.4117, 0.0323, -0.0083, 0.4117, 0.0323, -0.0061, 0.4129, 0.0299, -0.0061, 0.4123, 0.0312, -0.0083, 0.4129, 0.0299, -0.0083, 0.4123, 0.0312, -0.01, 0.411, 0.0255, -0.01, 0.411, 0.0277, -0.01, 0.4144, 0.026, -0.01, 0.4137, 0.0279, -0.01, 0.411, 0.0298, -0.01, 0.4117, 0.0323, -0.01, 0.4129, 0.03, -0.01, 0.4123, 0.0312,
-0.0061, 0.4146, 0.0254, -0.0083, 0.4146, 0.0254, -0.01, 0.4147, 0.0251, -0.0061, 0.385, 0.0371, -0.0083, 0.385, 0.0371, -0.0061, 0.3872, 0.0366, -0.0083, 0.3872, 0.0366, -0.01, 0.385, 0.0371, -0.01, 0.3871, 0.0367, -0.01, 0.385, 0.0385, -0.01, 0.385, 0.0407, -0.01, 0.3872, 0.0385, -0.01, 0.3872, 0.0407, -0.0061, 0.3894, 0.036, -0.0083, 0.3894, 0.036, -0.0061, 0.3918, 0.0348, -0.0061, 0.3907, 0.0354, -0.0083, 0.3918, 0.0348, -0.0083, 0.3907, 0.0354, -0.01, 0.3891, 0.0361, -0.01, 0.3919, 0.0348, -0.01, 0.3909, 0.0352, -0.01, 0.3893, 0.0385, -0.01, 0.3893, 0.0407, -0.01, 0.3915, 0.0385, -0.01, 0.3915, 0.0407, -0.0061, 0.385, 0.0487, -0.0083, 0.385, 0.0487, -0.0061, 0.3872, 0.0486, -0.0083, 0.3872, 0.0486,
-0.01, 0.385, 0.0429, -0.01, 0.385, 0.045, -0.01, 0.3872, 0.0429, -0.01, 0.3872, 0.045, -0.01, 0.385, 0.0472, -0.01, 0.385, 0.0487, -0.01, 0.3872, 0.0472, -0.01, 0.3872, 0.0486, -0.0061, 0.39, 0.0482, -0.0061, 0.389, 0.0484, -0.0083, 0.39, 0.0482, -0.0083, 0.389, 0.0484, -0.0061, 0.3915, 0.0479, -0.0083, 0.3915, 0.0479, -0.01, 0.3893, 0.0429, -0.01, 0.3893, 0.045, -0.01, 0.3915, 0.0429, -0.01, 0.3915, 0.045, -0.01, 0.3896, 0.0483, -0.01, 0.3887, 0.0485, -0.01, 0.3917, 0.0478, -0.0061, 0.3936, 0.0338, -0.0083, 0.3936, 0.0338, -0.01, 0.3936, 0.0338, -0.01, 0.3937, 0.0363, -0.01, 0.3958, 0.0342, -0.01, 0.3958, 0.0363, -0.01, 0.3937, 0.0385, -0.01, 0.3937, 0.0407, -0.01, 0.3958, 0.0385,
-0.01, 0.3958, 0.0407, -0.01, 0.398, 0.0342, -0.01, 0.398, 0.0363, -0.01, 0.4002, 0.0342, -0.01, 0.4002, 0.0363, -0.01, 0.398, 0.0385, -0.01, 0.398, 0.0407, -0.01, 0.4002, 0.0385, -0.01, 0.4002, 0.0407, -0.0061, 0.3967, 0.046, -0.0083, 0.3967, 0.046, -0.0061, 0.3937, 0.0472, -0.0083, 0.3937, 0.0472, -0.0061, 0.3956, 0.0465, -0.0083, 0.3956, 0.0465, -0.01, 0.3937, 0.0429, -0.01, 0.3937, 0.045, -0.01, 0.3958, 0.0429, -0.01, 0.3964, 0.0461, -0.01, 0.3938, 0.0471, -0.01, 0.3954, 0.0465, -0.0061, 0.398, 0.0454, -0.0083, 0.398, 0.0454, -0.0061, 0.4009, 0.0437, -0.0061, 0.3998, 0.0444, -0.0083, 0.4009, 0.0437, -0.0083, 0.3998, 0.0444, -0.01, 0.398, 0.0429, -0.01, 0.3983, 0.0453, -0.01, 0.4008, 0.0438,
-0.01, 0.3997, 0.0445, -0.0061, 0.4046, 0.0409, -0.0083, 0.4046, 0.0409, -0.01, 0.4024, 0.0342, -0.01, 0.4024, 0.0363, -0.01, 0.4045, 0.0342, -0.01, 0.4045, 0.0363, -0.01, 0.4024, 0.0385, -0.01, 0.4024, 0.0407, -0.01, 0.4045, 0.0385, -0.01, 0.4046, 0.0409, -0.0061, 0.4099, 0.0352, -0.0061, 0.409, 0.0364, -0.0083, 0.4099, 0.0352, -0.0083, 0.409, 0.0364, -0.0061, 0.4069, 0.0387, -0.0061, 0.4058, 0.0398, -0.0083, 0.4069, 0.0387, -0.0083, 0.4058, 0.0398, -0.0061, 0.4079, 0.0376, -0.0083, 0.4079, 0.0376, -0.01, 0.4067, 0.0342, -0.01, 0.4067, 0.0363, -0.01, 0.41, 0.035, -0.01, 0.4088, 0.0367, -0.01, 0.4069, 0.0387, -0.01, 0.4058, 0.0398, -0.01, 0.4079, 0.0376, -0.0061, 0.4024, 0.0427, -0.0083, 0.4024, 0.0427,
-0.0061, 0.4035, 0.0418, -0.0083, 0.4035, 0.0418, -0.01, 0.4026, 0.0426, -0.01, 0.4035, 0.0418, -0.0061, 0.4106, 0.0341, -0.0083, 0.4106, 0.0341, -0.01, 0.4107, 0.034, -0.0061, 0.429, 0.0153, -0.0083, 0.429, 0.0153, -0.0061, 0.4316, 0.0135, -0.0061, 0.4305, 0.0142, -0.0083, 0.4316, 0.0135, -0.0083, 0.4305, 0.0142, -0.01, 0.4291, 0.0153, -0.01, 0.4316, 0.0135, -0.01, 0.4304, 0.0142, -0.0061, 0.4327, 0.0129, -0.0083, 0.4327, 0.0129, -0.0061, 0.4349, 0.0119, -0.0083, 0.4349, 0.0119, -0.01, 0.4326, 0.013, -0.01, 0.4327, 0.0146, -0.01, 0.4346, 0.012, -0.01, 0.4349, 0.0146, -0.0061, 0.427, 0.0175, -0.0061, 0.4258, 0.0189, -0.0083, 0.427, 0.0175, -0.0083, 0.4258, 0.0189, -0.0061, 0.4245, 0.0211, -0.0061, 0.4237, 0.0233,
-0.0083, 0.4245, 0.0211, -0.0083, 0.4237, 0.0233, -0.0061, 0.4252, 0.0201, -0.0083, 0.4252, 0.0201, -0.01, 0.427, 0.0175, -0.01, 0.4261, 0.0186, -0.01, 0.4244, 0.0214, -0.01, 0.4238, 0.0231, -0.01, 0.425, 0.0204, -0.01, 0.4262, 0.0233, -0.0061, 0.423, 0.0265, -0.0061, 0.4228, 0.0277, -0.0083, 0.423, 0.0265, -0.0083, 0.4228, 0.0277, -0.0061, 0.4226, 0.0298, -0.0061, 0.4228, 0.0319, -0.0083, 0.4226, 0.0298, -0.0083, 0.4228, 0.0319, -0.01, 0.4229, 0.0265, -0.01, 0.4228, 0.0277, -0.01, 0.4226, 0.0298, -0.01, 0.4228, 0.0316, -0.0061, 0.4231, 0.0254, -0.0083, 0.4231, 0.0254, -0.0061, 0.423, 0.033, -0.0083, 0.423, 0.033, -0.01, 0.4231, 0.0255, -0.01, 0.4241, 0.0277, -0.01, 0.4262, 0.0255, -0.01, 0.4262, 0.0277,
-0.01, 0.4241, 0.0298, -0.01, 0.4229, 0.0324, -0.01, 0.4262, 0.0298, -0.01, 0.4262, 0.032, -0.0061, 0.4279, 0.0164, -0.0083, 0.4279, 0.0164, -0.01, 0.4279, 0.0164, -0.01, 0.4284, 0.019, -0.01, 0.4306, 0.0168, -0.01, 0.4306, 0.019, -0.01, 0.4284, 0.0212, -0.01, 0.4284, 0.0233, -0.01, 0.4306, 0.0212, -0.01, 0.4306, 0.0233, -0.01, 0.4327, 0.0168, -0.01, 0.4327, 0.019, -0.01, 0.4349, 0.0168, -0.01, 0.4349, 0.019, -0.01, 0.4327, 0.0212, -0.01, 0.4327, 0.0233, -0.01, 0.4349, 0.0212, -0.01, 0.4349, 0.0233, -0.01, 0.4284, 0.0255, -0.01, 0.4284, 0.0277, -0.01, 0.4306, 0.0255, -0.01, 0.4306, 0.0277, -0.01, 0.4284, 0.0298, -0.01, 0.4284, 0.032, -0.01, 0.4306, 0.0298, -0.01, 0.4306, 0.032,
-0.01, 0.4327, 0.0255, -0.01, 0.4327, 0.0277, -0.01, 0.4349, 0.0255, -0.01, 0.4349, 0.0277, -0.01, 0.4327, 0.0298, -0.01, 0.4327, 0.032, -0.01, 0.4349, 0.0298, -0.01, 0.4349, 0.032, -0.0061, 0.4375, 0.0112, -0.0083, 0.4375, 0.0112, -0.0061, 0.4392, 0.0109, -0.0083, 0.4392, 0.0109, -0.0061, 0.4364, 0.0115, -0.0083, 0.4364, 0.0115, -0.01, 0.4377, 0.0112, -0.01, 0.4392, 0.0109, -0.01, 0.4368, 0.0114, -0.01, 0.4371, 0.0146, -0.01, 0.4392, 0.0125, -0.01, 0.4392, 0.0146, -0.0061, 0.4414, 0.0106, -0.0083, 0.4414, 0.0106, -0.0061, 0.4436, 0.0105, -0.0083, 0.4436, 0.0105, -0.01, 0.4414, 0.0106, -0.01, 0.4436, 0.0105, -0.01, 0.4414, 0.0125, -0.01, 0.4414, 0.0146, -0.01, 0.4436, 0.0125, -0.01, 0.4436, 0.0146,
-0.0061, 0.4458, 0.0105, -0.0083, 0.4458, 0.0105, -0.0061, 0.4479, 0.0105, -0.0083, 0.4479, 0.0105, -0.01, 0.4458, 0.0105, -0.01, 0.4479, 0.0105, -0.01, 0.4458, 0.0125, -0.01, 0.4458, 0.0146, -0.01, 0.4479, 0.0125, -0.01, 0.4479, 0.0146, -0.0061, 0.4501, 0.0105, -0.0083, 0.4501, 0.0105, -0.0061, 0.4523, 0.0105, -0.0083, 0.4523, 0.0105, -0.01, 0.4501, 0.0105, -0.01, 0.4523, 0.0105, -0.01, 0.4501, 0.0125, -0.01, 0.4501, 0.0146, -0.01, 0.4523, 0.0125, -0.01, 0.4523, 0.0146, -0.0061, 0.4399, 0.0241, -0.0083, 0.4399, 0.0241, -0.01, 0.4371, 0.0168, -0.01, 0.4371, 0.019, -0.01, 0.4392, 0.0168, -0.01, 0.4392, 0.019, -0.01, 0.4371, 0.0212, -0.01, 0.4371, 0.0233, -0.01, 0.4392, 0.0212, -0.01, 0.4398, 0.0242,
-0.0061, 0.4414, 0.0232, -0.0083, 0.4414, 0.0232, -0.0061, 0.4436, 0.0225, -0.0083, 0.4436, 0.0225, -0.01, 0.4414, 0.0168, -0.01, 0.4414, 0.019, -0.01, 0.4436, 0.0168, -0.01, 0.4436, 0.019, -0.01, 0.4414, 0.0212, -0.01, 0.4416, 0.0232, -0.01, 0.4436, 0.0212, -0.01, 0.4436, 0.0225, -0.0061, 0.4377, 0.0261, -0.0061, 0.4367, 0.0276, -0.0083, 0.4377, 0.0261, -0.0083, 0.4367, 0.0276, -0.0061, 0.4388, 0.025, -0.0083, 0.4388, 0.025, -0.0061, 0.436, 0.0297, -0.0061, 0.4367, 0.0321, -0.0083, 0.436, 0.0297, -0.0083, 0.4367, 0.0321, -0.01, 0.4378, 0.026, -0.01, 0.4366, 0.0278, -0.01, 0.4388, 0.025, -0.01, 0.436, 0.0297, -0.01, 0.4366, 0.0319, -0.0061, 0.4467, 0.0222, -0.0061, 0.4456, 0.0223, -0.0083, 0.4467, 0.0222,
-0.0083, 0.4456, 0.0223, -0.0061, 0.4479, 0.0222, -0.0083, 0.4479, 0.0222, -0.01, 0.4458, 0.0168, -0.01, 0.4458, 0.019, -0.01, 0.4479, 0.0168, -0.01, 0.4479, 0.019, -0.01, 0.4459, 0.0223, -0.01, 0.4453, 0.0223, -0.01, 0.4479, 0.0222, -0.0061, 0.4501, 0.0222, -0.0083, 0.4501, 0.0222, -0.0061, 0.4523, 0.0222, -0.0083, 0.4523, 0.0222, -0.01, 0.4501, 0.0168, -0.01, 0.4501, 0.019, -0.01, 0.4523, 0.0168, -0.01, 0.4523, 0.019, -0.01, 0.4501, 0.0222, -0.01, 0.4523, 0.0222, -0.0061, 0.4232, 0.0342, -0.0061, 0.4237, 0.0364, -0.0083, 0.4232, 0.0342, -0.0083, 0.4237, 0.0364, -0.0061, 0.4245, 0.0385, -0.0083, 0.4245, 0.0385, -0.0061, 0.4252, 0.0395, -0.0061, 0.426, 0.0407, -0.0083, 0.4252, 0.0395, -0.0083, 0.426, 0.0407,
-0.01, 0.4232, 0.0344, -0.01, 0.4238, 0.0366, -0.01, 0.4262, 0.0342, -0.01, 0.4262, 0.0363, -0.01, 0.4244, 0.0383, -0.01, 0.4251, 0.0393, -0.01, 0.4262, 0.041, -0.0061, 0.427, 0.0421, -0.0083, 0.427, 0.0421, -0.01, 0.427, 0.042, -0.01, 0.4284, 0.0342, -0.01, 0.4284, 0.0363, -0.01, 0.4306, 0.0342, -0.01, 0.4306, 0.0363, -0.01, 0.4284, 0.0385, -0.01, 0.4284, 0.0407, -0.01, 0.4306, 0.0385, -0.01, 0.4306, 0.0407, -0.01, 0.4327, 0.0342, -0.01, 0.4327, 0.0363, -0.01, 0.4349, 0.0342, -0.01, 0.4349, 0.0363, -0.01, 0.4327, 0.0385, -0.01, 0.4327, 0.0407, -0.01, 0.4349, 0.0385, -0.01, 0.4349, 0.0407, -0.0061, 0.4281, 0.0432, -0.0061, 0.4292, 0.0442, -0.0083, 0.4281, 0.0432, -0.0083, 0.4292, 0.0442,
-0.0061, 0.4305, 0.0452, -0.0083, 0.4305, 0.0452, -0.01, 0.4281, 0.0432, -0.01, 0.4292, 0.0442, -0.01, 0.4306, 0.0429, -0.01, 0.4302, 0.045, -0.0061, 0.4318, 0.046, -0.0083, 0.4318, 0.046, -0.0061, 0.4329, 0.0467, -0.0083, 0.4329, 0.0467, -0.0061, 0.4349, 0.0476, -0.0083, 0.4349, 0.0476, -0.01, 0.4327, 0.0429, -0.01, 0.432, 0.0461, -0.01, 0.4349, 0.0429, -0.01, 0.4349, 0.045, -0.01, 0.433, 0.0468, -0.01, 0.4346, 0.0475, -0.0061, 0.4377, 0.0336, -0.0083, 0.4377, 0.0336, -0.0061, 0.4388, 0.0347, -0.0061, 0.4398, 0.0356, -0.0083, 0.4388, 0.0347, -0.0083, 0.4398, 0.0356, -0.01, 0.4377, 0.0337, -0.01, 0.4371, 0.0363, -0.01, 0.4388, 0.0347, -0.01, 0.4398, 0.0356, -0.01, 0.4371, 0.0385, -0.01, 0.4371, 0.0407,
-0.01, 0.4392, 0.0385, -0.01, 0.4392, 0.0407, -0.0061, 0.4414, 0.0366, -0.0083, 0.4414, 0.0366, -0.0061, 0.4432, 0.0373, -0.0083, 0.4432, 0.0373, -0.0061, 0.4443, 0.0375, -0.0083, 0.4443, 0.0375, -0.01, 0.4416, 0.0367, -0.01, 0.443, 0.0372, -0.01, 0.4414, 0.0385, -0.01, 0.4414, 0.0407, -0.01, 0.4439, 0.0374, -0.01, 0.4436, 0.0407, -0.0061, 0.4366, 0.0481, -0.0061, 0.4376, 0.0484, -0.0083, 0.4366, 0.0481, -0.0083, 0.4376, 0.0484, -0.0061, 0.4392, 0.0486, -0.0083, 0.4392, 0.0486, -0.01, 0.4371, 0.0429, -0.01, 0.4371, 0.045, -0.01, 0.4392, 0.0429, -0.01, 0.4392, 0.045, -0.01, 0.4368, 0.0482, -0.01, 0.4378, 0.0484, -0.01, 0.4392, 0.0472, -0.01, 0.4392, 0.0486, -0.0061, 0.4414, 0.0487, -0.0083, 0.4414, 0.0487,
-0.0061, 0.4436, 0.0487, -0.0083, 0.4436, 0.0487, -0.01, 0.4414, 0.0429, -0.01, 0.4414, 0.045, -0.01, 0.4436, 0.0429, -0.01, 0.4436, 0.045, -0.01, 0.4414, 0.0472, -0.01, 0.4414, 0.0487, -0.01, 0.4436, 0.0472, -0.01, 0.4436, 0.0487, -0.0061, 0.4457, 0.0378, -0.0083, 0.4457, 0.0378, -0.0061, 0.4479, 0.0379, -0.0083, 0.4479, 0.0379, -0.01, 0.4459, 0.0378, -0.01, 0.4458, 0.0407, -0.01, 0.4479, 0.0379, -0.01, 0.4479, 0.0407, -0.0061, 0.4501, 0.0379, -0.0083, 0.4501, 0.0379, -0.0061, 0.4523, 0.0379, -0.0083, 0.4523, 0.0379, -0.01, 0.4501, 0.0379, -0.01, 0.4501, 0.0407, -0.01, 0.4523, 0.0379, -0.01, 0.4523, 0.0407, -0.0061, 0.4458, 0.0487, -0.0083, 0.4458, 0.0487, -0.0061, 0.4479, 0.0487, -0.0083, 0.4479, 0.0487,
-0.01, 0.4458, 0.0429, -0.01, 0.4458, 0.045, -0.01, 0.4479, 0.0429, -0.01, 0.4479, 0.045, -0.01, 0.4458, 0.0472, -0.01, 0.4458, 0.0487, -0.01, 0.4479, 0.0472, -0.01, 0.4479, 0.0487, -0.0061, 0.4501, 0.0487, -0.0083, 0.4501, 0.0487, -0.0061, 0.4523, 0.0487, -0.0083, 0.4523, 0.0487, -0.01, 0.4501, 0.0429, -0.01, 0.4501, 0.045, -0.01, 0.4523, 0.0429, -0.01, 0.4523, 0.045, -0.01, 0.4501, 0.0472, -0.01, 0.4501, 0.0487, -0.01, 0.4523, 0.0472, -0.01, 0.4523, 0.0487, -0.0061, 0.4544, 0.0105, -0.0083, 0.4544, 0.0105, -0.0061, 0.4566, 0.0105, -0.0083, 0.4566, 0.0105, -0.01, 0.4544, 0.0105, -0.01, 0.4566, 0.0105, -0.01, 0.4544, 0.0125, -0.01, 0.4544, 0.0146, -0.01, 0.4566, 0.0125, -0.01, 0.4566, 0.0146,
-0.0061, 0.4588, 0.0105, -0.0083, 0.4588, 0.0105, -0.0061, 0.4609, 0.0105, -0.0083, 0.4609, 0.0105, -0.01, 0.4588, 0.0105, -0.01, 0.4609, 0.0105, -0.01, 0.4588, 0.0125, -0.01, 0.4588, 0.0146, -0.01, 0.4609, 0.0125, -0.01, 0.4609, 0.0146, -0.0061, 0.4631, 0.0105, -0.0083, 0.4631, 0.0105, -0.0061, 0.4653, 0.0105, -0.0083, 0.4653, 0.0105, -0.01, 0.4631, 0.0105, -0.01, 0.4653, 0.0105, -0.01, 0.4631, 0.0125, -0.01, 0.4631, 0.0146, -0.01, 0.4653, 0.0125, -0.01, 0.4653, 0.0146, -0.0061, 0.4675, 0.0105, -0.0083, 0.4675, 0.0105, -0.0061, 0.4696, 0.0105, -0.0083, 0.4696, 0.0105, -0.01, 0.4675, 0.0105, -0.01, 0.4696, 0.0105, -0.01, 0.4675, 0.0125, -0.01, 0.4675, 0.0146, -0.01, 0.4696, 0.0125, -0.01, 0.4696, 0.0146,
-0.0061, 0.4544, 0.0222, -0.0083, 0.4544, 0.0222, -0.0061, 0.4566, 0.0222, -0.0083, 0.4566, 0.0222, -0.01, 0.4544, 0.0168, -0.01, 0.4544, 0.019, -0.01, 0.4566, 0.0168, -0.01, 0.4566, 0.019, -0.01, 0.4544, 0.0222, -0.01, 0.4566, 0.0222, -0.0061, 0.4588, 0.0222, -0.0083, 0.4588, 0.0222, -0.0061, 0.4609, 0.0222, -0.0083, 0.4609, 0.0222, -0.01, 0.4588, 0.0168, -0.01, 0.4588, 0.019, -0.01, 0.4609, 0.0168, -0.01, 0.4609, 0.019, -0.01, 0.4588, 0.0222, -0.01, 0.4609, 0.0222, -0.0061, 0.4631, 0.0222, -0.0083, 0.4631, 0.0222, -0.0061, 0.4653, 0.0222, -0.0083, 0.4653, 0.0222, -0.01, 0.4631, 0.0168, -0.01, 0.4631, 0.019, -0.01, 0.4653, 0.0168, -0.01, 0.4653, 0.019, -0.01, 0.4631, 0.0222, -0.01, 0.4653, 0.0222,
-0.0061, 0.4675, 0.0222, -0.0083, 0.4675, 0.0222, -0.0061, 0.4696, 0.0222, -0.0083, 0.4696, 0.0222, -0.01, 0.4675, 0.0168, -0.01, 0.4675, 0.019, -0.01, 0.4696, 0.0168, -0.01, 0.4696, 0.019, -0.01, 0.4675, 0.0222, -0.01, 0.4696, 0.0222, -0.0061, 0.4718, 0.0105, -0.0083, 0.4718, 0.0105, -0.0061, 0.474, 0.0105, -0.0083, 0.474, 0.0105, -0.01, 0.4718, 0.0105, -0.01, 0.474, 0.0105, -0.01, 0.4718, 0.0125, -0.01, 0.4718, 0.0146, -0.01, 0.474, 0.0125, -0.01, 0.474, 0.0146, -0.0061, 0.4761, 0.0104, -0.0083, 0.4761, 0.0104, -0.0061, 0.4783, 0.0104, -0.0083, 0.4783, 0.0104, -0.01, 0.4761, 0.0104, -0.01, 0.4783, 0.0104, -0.01, 0.4761, 0.0125, -0.01, 0.4761, 0.0146, -0.01, 0.4783, 0.0125, -0.01, 0.4783, 0.0146,
-0.0061, 0.4877, 0.0067, -0.0083, 0.4877, 0.0067, -0.01, 0.4877, 0.0067, -0.0061, 0.4805, 0.0102, -0.0083, 0.4805, 0.0102, -0.0061, 0.4836, 0.0092, -0.0061, 0.4826, 0.0096, -0.0083, 0.4836, 0.0092, -0.0083, 0.4826, 0.0096, -0.01, 0.4804, 0.0102, -0.01, 0.4837, 0.0092, -0.01, 0.4826, 0.0096, -0.01, 0.4805, 0.0125, -0.01, 0.4805, 0.0146, -0.01, 0.4826, 0.0125, -0.01, 0.4826, 0.0146, -0.0061, 0.4849, 0.0087, -0.0083, 0.4849, 0.0087, -0.0061, 0.4866, 0.0076, -0.0083, 0.4866, 0.0076, -0.01, 0.4847, 0.0087, -0.01, 0.4848, 0.0103, -0.01, 0.4866, 0.0076, -0.01, 0.487, 0.0103, -0.01, 0.4848, 0.0125, -0.01, 0.4848, 0.0146, -0.01, 0.487, 0.0125, -0.01, 0.487, 0.0146, -0.0061, 0.4718, 0.0222, -0.0083, 0.4718, 0.0222,
-0.0061, 0.474, 0.0222, -0.0083, 0.474, 0.0222, -0.01, 0.4718, 0.0168, -0.01, 0.4718, 0.019, -0.01, 0.474, 0.0168, -0.01, 0.474, 0.019, -0.01, 0.4718, 0.0222, -0.01, 0.474, 0.0222, -0.0061, 0.4761, 0.0222, -0.0083, 0.4761, 0.0222, -0.0061, 0.4783, 0.0222, -0.0083, 0.4783, 0.0222, -0.01, 0.4761, 0.0168, -0.01, 0.4761, 0.019, -0.01, 0.4783, 0.0168, -0.01, 0.4783, 0.019, -0.01, 0.4761, 0.0222, -0.01, 0.4783, 0.0222, -0.0061, 0.4805, 0.0222, -0.0083, 0.4805, 0.0222, -0.0061, 0.4827, 0.0219, -0.0083, 0.4827, 0.0219, -0.01, 0.4805, 0.0168, -0.01, 0.4805, 0.019, -0.01, 0.4826, 0.0168, -0.01, 0.4826, 0.019, -0.01, 0.4806, 0.0221, -0.01, 0.4828, 0.0219, -0.0061, 0.4848, 0.0214, -0.0083, 0.4848, 0.0214,
-0.0061, 0.487, 0.0208, -0.0083, 0.487, 0.0208, -0.01, 0.4848, 0.0168, -0.01, 0.4848, 0.019, -0.01, 0.487, 0.0168, -0.01, 0.487, 0.019, -0.01, 0.485, 0.0214, -0.01, 0.4871, 0.0208, -0.0061, 0.4544, 0.0379, -0.0083, 0.4544, 0.0379, -0.0061, 0.4566, 0.0379, -0.0083, 0.4566, 0.0379, -0.01, 0.4544, 0.0379, -0.01, 0.4544, 0.0407, -0.01, 0.4566, 0.0379, -0.01, 0.4566, 0.0407, -0.0061, 0.4588, 0.0379, -0.0083, 0.4588, 0.0379, -0.0061, 0.4609, 0.0379, -0.0083, 0.4609, 0.0379, -0.01, 0.4588, 0.0379, -0.01, 0.4588, 0.0407, -0.01, 0.4609, 0.0379, -0.01, 0.4609, 0.0407, -0.0061, 0.4544, 0.0487, -0.0083, 0.4544, 0.0487, -0.0061, 0.4566, 0.0487, -0.0083, 0.4566, 0.0487, -0.01, 0.4544, 0.0429, -0.01, 0.4544, 0.045,
-0.01, 0.4566, 0.0429, -0.01, 0.4566, 0.045, -0.01, 0.4544, 0.0472, -0.01, 0.4544, 0.0487, -0.01, 0.4566, 0.0472, -0.01, 0.4566, 0.0487, -0.0061, 0.4588, 0.0487, -0.0083, 0.4588, 0.0487, -0.0061, 0.4609, 0.0487, -0.0083, 0.4609, 0.0487, -0.01, 0.4588, 0.0429, -0.01, 0.4588, 0.045, -0.01, 0.4609, 0.0429, -0.01, 0.4609, 0.045, -0.01, 0.4588, 0.0472, -0.01, 0.4588, 0.0487, -0.01, 0.4609, 0.0472, -0.01, 0.4609, 0.0487, -0.0061, 0.4631, 0.0379, -0.0083, 0.4631, 0.0379, -0.0061, 0.4653, 0.0379, -0.0083, 0.4653, 0.0379, -0.01, 0.4631, 0.0379, -0.01, 0.4631, 0.0407, -0.01, 0.4653, 0.0379, -0.01, 0.4653, 0.0407, -0.0061, 0.4675, 0.0379, -0.0083, 0.4675, 0.0379, -0.0061, 0.4696, 0.0379, -0.0083, 0.4696, 0.0379,
-0.01, 0.4675, 0.0379, -0.01, 0.4675, 0.0407, -0.01, 0.4696, 0.0379, -0.01, 0.4696, 0.0407, -0.0061, 0.4631, 0.0487, -0.0083, 0.4631, 0.0487, -0.0061, 0.4653, 0.0487, -0.0083, 0.4653, 0.0487, -0.01, 0.4631, 0.0429, -0.01, 0.4631, 0.045, -0.01, 0.4653, 0.0429, -0.01, 0.4653, 0.045, -0.01, 0.4631, 0.0472, -0.01, 0.4631, 0.0487, -0.01, 0.4653, 0.0472, -0.01, 0.4653, 0.0487, -0.0061, 0.4675, 0.0487, -0.0083, 0.4675, 0.0487, -0.0061, 0.4696, 0.0487, -0.0083, 0.4696, 0.0487, -0.01, 0.4675, 0.0429, -0.01, 0.4675, 0.045, -0.01, 0.4696, 0.0429, -0.01, 0.4696, 0.045, -0.01, 0.4675, 0.0472, -0.01, 0.4675, 0.0487, -0.01, 0.4696, 0.0472, -0.01, 0.4696, 0.0487, -0.0061, 0.4718, 0.0379, -0.0083, 0.4718, 0.0379,
-0.0061, 0.474, 0.0379, -0.0083, 0.474, 0.0379, -0.01, 0.4718, 0.0379, -0.01, 0.4718, 0.0407, -0.01, 0.474, 0.0379, -0.01, 0.474, 0.0407, -0.0061, 0.4761, 0.0379, -0.0083, 0.4761, 0.0379, -0.0061, 0.4783, 0.0379, -0.0083, 0.4783, 0.0379, -0.01, 0.4761, 0.0379, -0.01, 0.4761, 0.0407, -0.01, 0.4783, 0.0379, -0.01, 0.4783, 0.0407, -0.0061, 0.4718, 0.0487, -0.0083, 0.4718, 0.0487, -0.0061, 0.474, 0.0487, -0.0083, 0.474, 0.0487, -0.01, 0.4718, 0.0429, -0.01, 0.4718, 0.045, -0.01, 0.474, 0.0429, -0.01, 0.474, 0.045, -0.01, 0.4718, 0.0472, -0.01, 0.4718, 0.0487, -0.01, 0.474, 0.0472, -0.01, 0.474, 0.0487, -0.0061, 0.4761, 0.0487, -0.0083, 0.4761, 0.0487, -0.0061, 0.4783, 0.0487, -0.0083, 0.4783, 0.0487,
-0.01, 0.4761, 0.0429, -0.01, 0.4761, 0.045, -0.01, 0.4783, 0.0429, -0.01, 0.4783, 0.045, -0.01, 0.4761, 0.0472, -0.01, 0.4761, 0.0487, -0.01, 0.4783, 0.0472, -0.01, 0.4783, 0.0487, -0.0061, 0.4805, 0.0379, -0.0083, 0.4805, 0.0379, -0.0061, 0.4826, 0.0379, -0.0083, 0.4826, 0.0379, -0.01, 0.4805, 0.0379, -0.01, 0.4805, 0.0407, -0.01, 0.4826, 0.0379, -0.01, 0.4826, 0.0407, -0.0061, 0.4848, 0.0379, -0.0083, 0.4848, 0.0379, -0.0061, 0.487, 0.0379, -0.0083, 0.487, 0.0379, -0.01, 0.4848, 0.0379, -0.01, 0.4848, 0.0407, -0.01, 0.487, 0.0379, -0.01, 0.487, 0.0407, -0.0061, 0.4805, 0.0487, -0.0083, 0.4805, 0.0487, -0.0061, 0.4826, 0.0487, -0.0083, 0.4826, 0.0487, -0.01, 0.4805, 0.0429, -0.01, 0.4805, 0.045,
-0.01, 0.4826, 0.0429, -0.01, 0.4826, 0.045, -0.01, 0.4805, 0.0472, -0.01, 0.4805, 0.0487, -0.01, 0.4826, 0.0472, -0.01, 0.4826, 0.0487, -0.0061, 0.4848, 0.0487, -0.0083, 0.4848, 0.0487, -0.0061, 0.487, 0.0487, -0.0083, 0.487, 0.0487, -0.01, 0.4848, 0.0429, -0.01, 0.4848, 0.045, -0.01, 0.487, 0.0429, -0.01, 0.487, 0.045, -0.01, 0.4848, 0.0472, -0.01, 0.4848, 0.0487, -0.01, 0.487, 0.0472, -0.01, 0.487, 0.0487, -0.0061, 0.491, -0.0006, -0.0061, 0.4911, 0.0017, -0.0083, 0.491, -0.0006, -0.0083, 0.4911, 0.0017, -0.0061, 0.4898, 0.0042, -0.0061, 0.4888, 0.0057, -0.0083, 0.4898, 0.0042, -0.0083, 0.4888, 0.0057, -0.0061, 0.4905, 0.0032, -0.0083, 0.4905, 0.0032, -0.01, 0.491, -0.0004, -0.01, 0.4911, 0.0017,
-0.01, 0.4897, 0.0043, -0.01, 0.4888, 0.0057, -0.01, 0.4904, 0.0033, -0.01, 0.4913, 0.006, -0.01, 0.4935, -0.0005, -0.01, 0.4935, 0.0016, -0.01, 0.4957, -0.0005, -0.01, 0.4957, 0.0016, -0.01, 0.4935, 0.0038, -0.01, 0.4935, 0.006, -0.01, 0.4957, 0.0038, -0.01, 0.4957, 0.006, -0.01, 0.4892, 0.0081, -0.01, 0.4892, 0.0103, -0.01, 0.4913, 0.0081, -0.01, 0.4913, 0.0103, -0.01, 0.4892, 0.0125, -0.01, 0.4892, 0.0146, -0.01, 0.4913, 0.0125, -0.01, 0.4913, 0.0146, -0.0061, 0.4965, 0.0155, -0.0083, 0.4965, 0.0155, -0.01, 0.4935, 0.0081, -0.01, 0.4935, 0.0103, -0.01, 0.4957, 0.0081, -0.01, 0.4957, 0.0103, -0.01, 0.4935, 0.0125, -0.01, 0.4935, 0.0146, -0.01, 0.4957, 0.0125, -0.01, 0.4965, 0.0154,
-0.01, 0.4978, -0.0005, -0.01, 0.4978, 0.0016, -0.01, 0.5, -0.0005, -0.01, 0.5, 0.0016, -0.01, 0.4978, 0.0038, -0.01, 0.4978, 0.006, -0.01, 0.5, 0.0038, -0.01, 0.5, 0.006, -0.0061, 0.502, -0.0005, -0.0061, 0.502, 0.0016, -0.0083, 0.502, -0.0005, -0.0083, 0.502, 0.0016, -0.0061, 0.5019, 0.0038, -0.0061, 0.5017, 0.006, -0.0083, 0.5019, 0.0038, -0.0083, 0.5017, 0.006, -0.01, 0.502, -0.0005, -0.01, 0.502, 0.0016, -0.01, 0.5019, 0.0038, -0.01, 0.5017, 0.006, -0.0061, 0.5009, 0.0087, -0.0061, 0.5002, 0.0103, -0.0083, 0.5009, 0.0087, -0.0083, 0.5002, 0.0103, -0.0061, 0.4985, 0.013, -0.0061, 0.4974, 0.0144, -0.0083, 0.4985, 0.013, -0.0083, 0.4974, 0.0144, -0.0061, 0.4992, 0.0119, -0.0083, 0.4992, 0.0119,
-0.01, 0.4978, 0.0081, -0.01, 0.4978, 0.0103, -0.01, 0.501, 0.0085, -0.01, 0.5001, 0.0106, -0.01, 0.4986, 0.013, -0.01, 0.4975, 0.0143, -0.01, 0.4993, 0.0118, -0.0061, 0.5013, 0.0076, -0.0083, 0.5013, 0.0076, -0.01, 0.5013, 0.0075, -0.0061, 0.4896, 0.0198, -0.0083, 0.4896, 0.0198, -0.0061, 0.4914, 0.0191, -0.0083, 0.4914, 0.0191, -0.0061, 0.4886, 0.0203, -0.0083, 0.4886, 0.0203, -0.01, 0.4892, 0.0168, -0.01, 0.4895, 0.0199, -0.01, 0.4913, 0.0168, -0.01, 0.4916, 0.019, -0.01, 0.4884, 0.0203, -0.0061, 0.494, 0.0175, -0.0061, 0.4929, 0.0182, -0.0083, 0.494, 0.0175, -0.0083, 0.4929, 0.0182, -0.0061, 0.4954, 0.0165, -0.0083, 0.4954, 0.0165, -0.01, 0.4939, 0.0176, -0.01, 0.4928, 0.0182, -0.01, 0.4954, 0.0165,
-0.0061, 0.4892, 0.0379, -0.0083, 0.4892, 0.0379, -0.0061, 0.4913, 0.0379, -0.0083, 0.4913, 0.0379, -0.01, 0.4892, 0.0379, -0.01, 0.4892, 0.0407, -0.01, 0.4913, 0.0379, -0.01, 0.4913, 0.0407, -0.0061, 0.4935, 0.0379, -0.0083, 0.4935, 0.0379, -0.0061, 0.4957, 0.0379, -0.0083, 0.4957, 0.0379, -0.01, 0.4935, 0.0379, -0.01, 0.4935, 0.0407, -0.01, 0.4957, 0.0379, -0.01, 0.4957, 0.0407, -0.0061, 0.4892, 0.0487, -0.0083, 0.4892, 0.0487, -0.0061, 0.4913, 0.0487, -0.0083, 0.4913, 0.0487, -0.01, 0.4892, 0.0429, -0.01, 0.4892, 0.045, -0.01, 0.4913, 0.0429, -0.01, 0.4913, 0.045, -0.01, 0.4892, 0.0472, -0.01, 0.4892, 0.0487, -0.01, 0.4913, 0.0472, -0.01, 0.4913, 0.0487, -0.0061, 0.4935, 0.0487, -0.0083, 0.4935, 0.0487,
-0.0061, 0.4957, 0.0487, -0.0083, 0.4957, 0.0487, -0.01, 0.4935, 0.0429, -0.01, 0.4935, 0.045, -0.01, 0.4957, 0.0429, -0.01, 0.4957, 0.045, -0.01, 0.4935, 0.0472, -0.01, 0.4935, 0.0487, -0.01, 0.4957, 0.0472, -0.01, 0.4957, 0.0487, -0.0061, 0.498, 0.0379, -0.0061, 0.498, 0.0407, -0.0083, 0.498, 0.0379, -0.0083, 0.498, 0.0407, -0.01, 0.498, 0.0379, -0.01, 0.498, 0.0407, -0.0061, 0.498, 0.0429, -0.0061, 0.498, 0.045, -0.0083, 0.498, 0.0429, -0.0083, 0.498, 0.045, -0.0061, 0.498, 0.0472, -0.0061, 0.498, 0.0487, -0.0083, 0.498, 0.0472, -0.0083, 0.498, 0.0487, -0.01, 0.498, 0.0429, -0.01, 0.498, 0.045, -0.01, 0.498, 0.0472, -0.01, 0.498, 0.0487});
    _triy = _trix;

    _vertz = std::vector<float>({0.0364, -0.0, -0.4928, 0.0364, 0.0017, -0.4928, 0.0364, 0.0039, -0.4928, 0.0456, -0.0, -0.4868, 0.0481, -0.0, -0.485, 0.0467, -0.0, -0.486, 0.0457, 0.0017, -0.4868, 0.0457, 0.0039, -0.4868, 0.0482, 0.0017, -0.4849, 0.0468, 0.0017, -0.486, 0.0482, 0.0039, -0.4849, 0.0468, 0.0039, -0.486, 0.043, -0.0, -0.4886, 0.042, -0.0, -0.4893, 0.0391, -0.0, -0.4912, 0.0375, -0.0, -0.4922, 0.0402, -0.0, -0.4905, 0.0374, -0.0, -0.4896, 0.043, 0.0017, -0.4886, 0.0418, 0.0017, -0.4895, 0.043, 0.0039, -0.4886, 0.0418, 0.0039, -0.4895, 0.0393, 0.0017, -0.4911, 0.0375, 0.0017, -0.4922, 0.0393, 0.0039, -0.4911, 0.0375, 0.0039, -0.4922, 0.0403, 0.0017, -0.4904, 0.0403, 0.0039, -0.4904, 0.0441, -0.0, -0.4879, 0.0418, -0.0, -0.4874,
0.0439, -0.0, -0.4852, 0.0418, -0.0, -0.4852, 0.0396, -0.0, -0.4874, 0.0368, -0.0, -0.4869, 0.0386, -0.0, -0.4845, 0.0378, -0.0, -0.4856, 0.0441, 0.0017, -0.4879, 0.0441, 0.0039, -0.4879, 0.0368, 0.0017, -0.4868, 0.0368, 0.0039, -0.4868, 0.0387, 0.0017, -0.4844, 0.0379, 0.0017, -0.4855, 0.0387, 0.0039, -0.4844, 0.0379, 0.0039, -0.4855, 0.0321, -0.0, -0.4951, 0.0354, -0.0, -0.4934, 0.0332, -0.0, -0.4945, 0.0297, -0.0, -0.4961, 0.0308, -0.0, -0.4935, 0.0321, 0.0017, -0.4951, 0.0321, 0.0039, -0.4951, 0.0353, 0.0017, -0.4935, 0.0332, 0.0017, -0.4945, 0.0353, 0.0039, -0.4935, 0.0332, 0.0039, -0.4945, 0.0297, 0.0017, -0.4961, 0.0297, 0.0039, -0.4961, 0.0306, 0.0017, -0.4938, 0.0306, 0.0039, -0.4938, 0.024, -0.0, -0.4934,
0.0215, -0.0, -0.4953, 0.0203, -0.0, -0.496, 0.0226, -0.0, -0.4945, 0.0195, -0.0, -0.4935, 0.0241, 0.0017, -0.4934, 0.0241, 0.0039, -0.4934, 0.0215, 0.0017, -0.4952, 0.0201, 0.0017, -0.4961, 0.0215, 0.0039, -0.4952, 0.0201, 0.0039, -0.4961, 0.0226, 0.0017, -0.4945, 0.0195, 0.0017, -0.4935, 0.0226, 0.0039, -0.4945, 0.0195, 0.0039, -0.4935, 0.0353, -0.0, -0.4917, 0.0326, -0.0, -0.4914, 0.0347, -0.0, -0.4891, 0.0337, -0.0, -0.4902, 0.0317, -0.0, -0.4925, 0.0286, -0.0, -0.4894, 0.0326, 0.0017, -0.4914, 0.0326, 0.0039, -0.4914, 0.0347, 0.0017, -0.4891, 0.0337, 0.0017, -0.4902, 0.0347, 0.0039, -0.4891, 0.0337, 0.0039, -0.4902, 0.0317, 0.0017, -0.4925, 0.0317, 0.0039, -0.4925, 0.0286, 0.0017, -0.4894, 0.0286, 0.0039, -0.4894,
0.0358, -0.0, -0.4879, 0.0327, -0.0, -0.4849, 0.0307, -0.0, -0.4872, 0.0297, -0.0, -0.4883, 0.0318, -0.0, -0.486, 0.0288, -0.0, -0.4852, 0.0358, 0.0017, -0.4879, 0.0358, 0.0039, -0.4879, 0.0326, 0.0017, -0.4849, 0.0326, 0.0039, -0.4849, 0.0307, 0.0017, -0.4872, 0.0296, 0.0017, -0.4883, 0.0307, 0.0039, -0.4872, 0.0296, 0.0039, -0.4883, 0.0317, 0.0017, -0.486, 0.0317, 0.0039, -0.486, 0.0263, -0.0, -0.4915, 0.0251, -0.0, -0.4926, 0.0274, -0.0, -0.4905, 0.0244, -0.0, -0.4896, 0.0214, -0.0, -0.4911, 0.0206, -0.0, -0.4922, 0.0227, -0.0, -0.4894, 0.0264, 0.0017, -0.4915, 0.0252, 0.0017, -0.4925, 0.0264, 0.0039, -0.4915, 0.0252, 0.0039, -0.4925, 0.0274, 0.0017, -0.4905, 0.0274, 0.0039, -0.4905, 0.0215, 0.0017, -0.4911,
0.0206, 0.0017, -0.4922, 0.0215, 0.0039, -0.4911, 0.0206, 0.0039, -0.4922, 0.0225, 0.0017, -0.4896, 0.0225, 0.0039, -0.4896, 0.0266, -0.0, -0.4874, 0.0239, -0.0, -0.4873, 0.0266, -0.0, -0.4852, 0.025, -0.0, -0.4851, 0.0233, -0.0, -0.4884, 0.0239, 0.0017, -0.4873, 0.0239, 0.0039, -0.4873, 0.025, 0.0017, -0.4853, 0.025, 0.0039, -0.4853, 0.0233, 0.0017, -0.4884, 0.0233, 0.0039, -0.4884, 0.0173, -0.0, -0.4975, 0.016, -0.0, -0.498, 0.0175, 0.0017, -0.4975, 0.016, 0.0017, -0.498, 0.0175, 0.0039, -0.4975, 0.016, 0.0039, -0.498, 0.0157, -0.0, -0.4983, 0.0167, -0.0, -0.497, 0.0186, -0.0, -0.4946, 0.0157, 0.0017, -0.4983, 0.0167, 0.0017, -0.497, 0.0157, 0.0039, -0.4983, 0.0167, 0.0039, -0.497, 0.0186, 0.0017, -0.4945,
0.0186, 0.0039, -0.4945, 0.0713, -0.0, -0.4561, 0.0736, -0.0, -0.4541, 0.0724, -0.0, -0.4552, 0.0713, 0.0017, -0.4561, 0.0713, 0.0039, -0.4561, 0.0736, 0.0017, -0.4541, 0.0724, 0.0017, -0.4552, 0.0736, 0.0039, -0.4541, 0.0724, 0.0039, -0.4552, 0.0759, -0.0, -0.452, 0.0781, -0.0, -0.4499, 0.077, -0.0, -0.451, 0.0747, -0.0, -0.4531, 0.0722, -0.0, -0.4527, 0.0743, -0.0, -0.4505, 0.0722, -0.0, -0.4505, 0.0759, 0.0017, -0.452, 0.0759, 0.0039, -0.452, 0.0781, 0.0017, -0.4499, 0.077, 0.0017, -0.451, 0.0781, 0.0039, -0.4499, 0.077, 0.0039, -0.451, 0.0747, 0.0017, -0.4531, 0.0747, 0.0039, -0.4531, 0.0891, -0.0, -0.4371, 0.089, 0.0017, -0.4372, 0.089, 0.0039, -0.4372, 0.0908, -0.0, -0.4345, 0.0901, -0.0, -0.4356,
0.0916, -0.0, -0.4334, 0.0895, -0.0, -0.4332, 0.0908, 0.0017, -0.4345, 0.0901, 0.0017, -0.4356, 0.0908, 0.0039, -0.4345, 0.0901, 0.0039, -0.4356, 0.0917, 0.0017, -0.4332, 0.0917, 0.0039, -0.4332, 0.0802, -0.0, -0.4478, 0.0823, -0.0, -0.4455, 0.0813, -0.0, -0.4466, 0.0802, 0.0017, -0.4478, 0.0802, 0.0039, -0.4478, 0.0823, 0.0017, -0.4455, 0.0813, 0.0017, -0.4466, 0.0823, 0.0039, -0.4455, 0.0813, 0.0039, -0.4466, 0.0844, -0.0, -0.4432, 0.0863, -0.0, -0.4408, 0.0855, -0.0, -0.442, 0.0834, -0.0, -0.4443, 0.0808, -0.0, -0.444, 0.083, -0.0, -0.4418, 0.0808, -0.0, -0.4418, 0.0844, 0.0017, -0.4432, 0.0844, 0.0039, -0.4432, 0.0863, 0.0017, -0.4408, 0.0855, 0.0017, -0.4419, 0.0863, 0.0039, -0.4408, 0.0855, 0.0039, -0.4419,
0.0834, 0.0017, -0.4443, 0.0834, 0.0039, -0.4443, 0.0792, -0.0, -0.4488, 0.0765, -0.0, -0.4483, 0.0787, -0.0, -0.4462, 0.0765, -0.0, -0.4462, 0.0743, -0.0, -0.4483, 0.0722, -0.0, -0.4483, 0.0743, -0.0, -0.4462, 0.0722, -0.0, -0.4462, 0.0792, 0.0017, -0.4488, 0.0792, 0.0039, -0.4488, 0.0787, -0.0, -0.444, 0.0765, -0.0, -0.444, 0.0787, -0.0, -0.4418, 0.0765, -0.0, -0.4418, 0.0743, -0.0, -0.444, 0.0722, -0.0, -0.444, 0.0743, -0.0, -0.4418, 0.0722, -0.0, -0.4418, 0.087, -0.0, -0.4399, 0.0852, -0.0, -0.4397, 0.0883, -0.0, -0.4382, 0.0852, -0.0, -0.4375, 0.083, -0.0, -0.4397, 0.0808, -0.0, -0.4397, 0.083, -0.0, -0.4375, 0.0808, -0.0, -0.4375, 0.0872, 0.0017, -0.4397, 0.0872, 0.0039, -0.4397,
0.0882, 0.0017, -0.4383, 0.0882, 0.0039, -0.4383, 0.0874, -0.0, -0.4353, 0.0852, -0.0, -0.4353, 0.0874, -0.0, -0.4332, 0.0852, -0.0, -0.4332, 0.083, -0.0, -0.4353, 0.0808, -0.0, -0.4353, 0.083, -0.0, -0.4332, 0.0808, -0.0, -0.4332, 0.0787, -0.0, -0.4397, 0.0765, -0.0, -0.4397, 0.0787, -0.0, -0.4375, 0.0765, -0.0, -0.4375, 0.0743, -0.0, -0.4397, 0.0722, -0.0, -0.4397, 0.0743, -0.0, -0.4375, 0.0722, -0.0, -0.4375, 0.0787, -0.0, -0.4353, 0.0765, -0.0, -0.4353, 0.0787, -0.0, -0.4332, 0.0765, -0.0, -0.4332, 0.0743, -0.0, -0.4353, 0.0722, -0.0, -0.4353, 0.0743, -0.0, -0.4332, 0.0722, -0.0, -0.4332, 0.0972, -0.0, -0.4234, 0.0972, 0.0017, -0.4235, 0.0972, 0.0039, -0.4235, 0.0933, -0.0, -0.4306,
0.0949, -0.0, -0.4277, 0.0943, -0.0, -0.4289, 0.0927, -0.0, -0.4316, 0.0895, -0.0, -0.431, 0.0917, -0.0, -0.4288, 0.0895, -0.0, -0.4288, 0.0933, 0.0017, -0.4307, 0.0933, 0.0039, -0.4307, 0.0949, 0.0017, -0.4277, 0.0943, 0.0017, -0.4288, 0.0949, 0.0039, -0.4277, 0.0943, 0.0039, -0.4288, 0.0926, 0.0017, -0.4318, 0.0926, 0.0039, -0.4318, 0.0955, -0.0, -0.4268, 0.0939, -0.0, -0.4266, 0.0966, -0.0, -0.4246, 0.0939, -0.0, -0.4245, 0.0917, -0.0, -0.4266, 0.0895, -0.0, -0.4266, 0.0917, -0.0, -0.4245, 0.0895, -0.0, -0.4245, 0.0955, 0.0017, -0.4266, 0.0955, 0.0039, -0.4266, 0.0966, 0.0017, -0.4245, 0.0966, 0.0039, -0.4245, 0.0977, -0.0, -0.4224, 0.0986, -0.0, -0.4205, 0.0977, 0.0017, -0.4223, 0.0977, 0.0039, -0.4223,
0.0988, 0.0017, -0.4201, 0.0988, 0.0039, -0.4201, 0.0998, -0.0, -0.4177, 0.0995, -0.0, -0.4186, 0.1005, -0.0, -0.416, 0.0982, -0.0, -0.4158, 0.0997, 0.0017, -0.4179, 0.0993, 0.0017, -0.419, 0.0997, 0.0039, -0.4179, 0.0993, 0.0039, -0.419, 0.1006, 0.0017, -0.4158, 0.1006, 0.0039, -0.4158, 0.096, -0.0, -0.4223, 0.0939, -0.0, -0.4223, 0.096, -0.0, -0.4201, 0.0939, -0.0, -0.4201, 0.0917, -0.0, -0.4223, 0.0895, -0.0, -0.4223, 0.0917, -0.0, -0.4201, 0.0895, -0.0, -0.4201, 0.096, -0.0, -0.418, 0.0939, -0.0, -0.418, 0.096, -0.0, -0.4158, 0.0939, -0.0, -0.4158, 0.0917, -0.0, -0.418, 0.0895, -0.0, -0.418, 0.0917, -0.0, -0.4158, 0.0895, -0.0, -0.4158, 0.0874, -0.0, -0.431, 0.0852, -0.0, -0.431,
0.0874, -0.0, -0.4288, 0.0852, -0.0, -0.4288, 0.083, -0.0, -0.431, 0.0808, -0.0, -0.431, 0.083, -0.0, -0.4288, 0.0808, -0.0, -0.4288, 0.0874, -0.0, -0.4266, 0.0852, -0.0, -0.4266, 0.0874, -0.0, -0.4245, 0.0852, -0.0, -0.4245, 0.083, -0.0, -0.4266, 0.0808, -0.0, -0.4266, 0.083, -0.0, -0.4245, 0.0808, -0.0, -0.4245, 0.0787, -0.0, -0.431, 0.0765, -0.0, -0.431, 0.0787, -0.0, -0.4288, 0.0765, -0.0, -0.4288, 0.0743, -0.0, -0.431, 0.072, -0.0, -0.4307, 0.0741, -0.0, -0.4287, 0.0731, -0.0, -0.4297, 0.072, 0.0017, -0.4307, 0.072, 0.0039, -0.4307, 0.0741, 0.0017, -0.4287, 0.0731, 0.0017, -0.4297, 0.0741, 0.0039, -0.4287, 0.0731, 0.0039, -0.4297, 0.0787, -0.0, -0.4266, 0.076, -0.0, -0.4263,
0.0778, -0.0, -0.4239, 0.077, -0.0, -0.4249, 0.0751, -0.0, -0.4274, 0.076, 0.0017, -0.4263, 0.076, 0.0039, -0.4263, 0.0779, 0.0017, -0.4238, 0.0771, 0.0017, -0.4249, 0.0779, 0.0039, -0.4238, 0.0771, 0.0039, -0.4249, 0.0752, 0.0017, -0.4274, 0.0752, 0.0039, -0.4274, 0.0874, -0.0, -0.4223, 0.0852, -0.0, -0.4223, 0.0874, -0.0, -0.4201, 0.0852, -0.0, -0.4201, 0.083, -0.0, -0.4223, 0.0808, -0.0, -0.4223, 0.083, -0.0, -0.4201, 0.0804, -0.0, -0.42, 0.0804, 0.0017, -0.42, 0.0804, 0.0039, -0.42, 0.0874, -0.0, -0.418, 0.0852, -0.0, -0.418, 0.0874, -0.0, -0.4158, 0.0852, -0.0, -0.4158, 0.082, -0.0, -0.4174, 0.0814, -0.0, -0.4184, 0.083, -0.0, -0.4155, 0.0821, 0.0017, -0.4172, 0.0815, 0.0017, -0.4183,
0.0821, 0.0039, -0.4172, 0.0815, 0.0039, -0.4183, 0.0828, 0.0017, -0.4158, 0.0828, 0.0039, -0.4158, 0.079, -0.0, -0.4221, 0.0797, -0.0, -0.4211, 0.0789, 0.0017, -0.4223, 0.0789, 0.0039, -0.4223, 0.0797, 0.0017, -0.4211, 0.0797, 0.0039, -0.4211, 0.0538, -0.0, -0.4799, 0.0538, 0.0017, -0.4799, 0.0538, 0.0039, -0.4799, 0.0559, -0.0, -0.4777, 0.0549, -0.0, -0.4788, 0.0567, -0.0, -0.4768, 0.0548, -0.0, -0.4766, 0.0559, 0.0017, -0.4777, 0.0549, 0.0017, -0.4788, 0.0559, 0.0039, -0.4777, 0.0549, 0.0039, -0.4788, 0.0569, 0.0017, -0.4766, 0.0569, 0.0039, -0.4766, 0.0624, -0.0, -0.469, 0.063, -0.0, -0.468, 0.0624, 0.0017, -0.469, 0.0624, 0.0039, -0.469, 0.0631, 0.0017, -0.4679, 0.0631, 0.0039, -0.4679, 0.0589, -0.0, -0.4742,
0.0607, -0.0, -0.4717, 0.0599, -0.0, -0.4728, 0.058, -0.0, -0.4753, 0.0548, -0.0, -0.4744, 0.057, -0.0, -0.4722, 0.0548, -0.0, -0.4722, 0.0588, 0.0017, -0.4743, 0.0588, 0.0039, -0.4743, 0.0606, 0.0017, -0.4717, 0.0599, 0.0017, -0.4728, 0.0606, 0.0039, -0.4717, 0.0599, 0.0039, -0.4728, 0.058, 0.0017, -0.4753, 0.058, 0.0039, -0.4753, 0.0617, -0.0, -0.4701, 0.0591, -0.0, -0.47, 0.0613, -0.0, -0.4679, 0.0591, -0.0, -0.4679, 0.057, -0.0, -0.47, 0.0548, -0.0, -0.47, 0.057, -0.0, -0.4679, 0.0548, -0.0, -0.4679, 0.0617, 0.0017, -0.4701, 0.0617, 0.0039, -0.4701, 0.0516, -0.0, -0.482, 0.0507, -0.0, -0.4828, 0.0527, -0.0, -0.481, 0.0505, -0.0, -0.4809, 0.0492, -0.0, -0.4841, 0.0461, -0.0, -0.4831,
0.0483, -0.0, -0.4809, 0.0461, -0.0, -0.4809, 0.0516, 0.0017, -0.482, 0.0505, 0.0017, -0.483, 0.0516, 0.0039, -0.482, 0.0505, 0.0039, -0.483, 0.0527, 0.0017, -0.481, 0.0527, 0.0039, -0.481, 0.0493, 0.0017, -0.4841, 0.0493, 0.0039, -0.4841, 0.0526, -0.0, -0.4787, 0.0505, -0.0, -0.4787, 0.0526, -0.0, -0.4766, 0.0505, -0.0, -0.4766, 0.0483, -0.0, -0.4787, 0.0461, -0.0, -0.4787, 0.0483, -0.0, -0.4766, 0.0461, -0.0, -0.4766, 0.0439, -0.0, -0.4831, 0.0418, -0.0, -0.4831, 0.0439, -0.0, -0.4809, 0.0413, -0.0, -0.4807, 0.0398, -0.0, -0.4828, 0.0405, -0.0, -0.4818, 0.0413, 0.0017, -0.4807, 0.0413, 0.0039, -0.4807, 0.0397, 0.0017, -0.4831, 0.0397, 0.0039, -0.4831, 0.0406, 0.0017, -0.4818, 0.0406, 0.0039, -0.4818,
0.043, -0.0, -0.4781, 0.0423, -0.0, -0.4791, 0.044, -0.0, -0.4763, 0.0369, -0.0, -0.4783, 0.0376, -0.0, -0.4768, 0.0431, 0.0017, -0.4779, 0.0424, 0.0017, -0.479, 0.0431, 0.0039, -0.4779, 0.0424, 0.0039, -0.479, 0.0439, 0.0017, -0.4766, 0.0439, 0.0039, -0.4766, 0.0368, 0.0017, -0.4785, 0.0368, 0.0039, -0.4785, 0.0377, 0.0017, -0.4766, 0.0377, 0.0039, -0.4766, 0.0526, -0.0, -0.4744, 0.0505, -0.0, -0.4744, 0.0526, -0.0, -0.4722, 0.0505, -0.0, -0.4722, 0.0483, -0.0, -0.4744, 0.0452, -0.0, -0.474, 0.0483, -0.0, -0.4722, 0.046, -0.0, -0.472, 0.0453, 0.0017, -0.4738, 0.0453, 0.0039, -0.4738, 0.0459, 0.0017, -0.4722, 0.0459, 0.0039, -0.4722, 0.0526, -0.0, -0.47, 0.0505, -0.0, -0.47, 0.0526, -0.0, -0.4679,
0.0505, -0.0, -0.4679, 0.0483, -0.0, -0.47, 0.0468, -0.0, -0.47, 0.0476, -0.0, -0.4679, 0.0472, -0.0, -0.4689, 0.0468, 0.0017, -0.4701, 0.0468, 0.0039, -0.4701, 0.0476, 0.0017, -0.4678, 0.0472, 0.0017, -0.4689, 0.0476, 0.0039, -0.4678, 0.0472, 0.0039, -0.4689, 0.0447, -0.0, -0.475, 0.0388, -0.0, -0.4737, 0.0384, -0.0, -0.4747, 0.0393, -0.0, -0.4723, 0.0374, -0.0, -0.4722, 0.0448, 0.0017, -0.4749, 0.0448, 0.0039, -0.4749, 0.0387, 0.0017, -0.4739, 0.0383, 0.0017, -0.475, 0.0387, 0.0039, -0.4739, 0.0383, 0.0039, -0.475, 0.0393, 0.0017, -0.4722, 0.0393, 0.0039, -0.4722, 0.0408, -0.0, -0.467, 0.0399, -0.0, -0.4702, 0.0374, -0.0, -0.47, 0.0405, -0.0, -0.468, 0.0374, -0.0, -0.4679, 0.0408, 0.0017, -0.4671,
0.0408, 0.0039, -0.4671, 0.04, 0.0017, -0.47, 0.04, 0.0039, -0.47, 0.0405, 0.0017, -0.4682, 0.0405, 0.0039, -0.4682, 0.0647, -0.0, -0.4648, 0.0641, -0.0, -0.4659, 0.0653, -0.0, -0.4637, 0.0635, -0.0, -0.4635, 0.0647, 0.0017, -0.4649, 0.0641, 0.0017, -0.466, 0.0647, 0.0039, -0.4649, 0.0641, 0.0039, -0.466, 0.0654, 0.0017, -0.4635, 0.0654, 0.0039, -0.4635, 0.067, -0.0, -0.4606, 0.069, -0.0, -0.4582, 0.068, -0.0, -0.4593, 0.0664, -0.0, -0.4617, 0.0635, -0.0, -0.4614, 0.0656, -0.0, -0.4592, 0.0635, -0.0, -0.4592, 0.067, 0.0017, -0.4607, 0.067, 0.0039, -0.4607, 0.069, 0.0017, -0.4582, 0.068, 0.0017, -0.4593, 0.069, 0.0039, -0.4582, 0.068, 0.0039, -0.4593, 0.0663, 0.0017, -0.4618, 0.0663, 0.0039, -0.4618,
0.0613, -0.0, -0.4657, 0.0591, -0.0, -0.4657, 0.0613, -0.0, -0.4635, 0.0591, -0.0, -0.4635, 0.057, -0.0, -0.4657, 0.0548, -0.0, -0.4657, 0.057, -0.0, -0.4635, 0.0548, -0.0, -0.4635, 0.0613, -0.0, -0.4614, 0.0591, -0.0, -0.4614, 0.0613, -0.0, -0.4592, 0.0591, -0.0, -0.4592, 0.057, -0.0, -0.4614, 0.0548, -0.0, -0.4614, 0.057, -0.0, -0.4592, 0.0548, -0.0, -0.4592, 0.0701, -0.0, -0.4572, 0.0678, -0.0, -0.457, 0.07, -0.0, -0.4549, 0.0678, -0.0, -0.4549, 0.0656, -0.0, -0.457, 0.0635, -0.0, -0.457, 0.0656, -0.0, -0.4549, 0.0635, -0.0, -0.4549, 0.0701, 0.0017, -0.4572, 0.0701, 0.0039, -0.4572, 0.07, -0.0, -0.4527, 0.0678, -0.0, -0.4527, 0.07, -0.0, -0.4505, 0.0678, -0.0, -0.4505,
0.0656, -0.0, -0.4527, 0.0635, -0.0, -0.4527, 0.0656, -0.0, -0.4505, 0.0635, -0.0, -0.4505, 0.0613, -0.0, -0.457, 0.0591, -0.0, -0.457, 0.0613, -0.0, -0.4549, 0.0591, -0.0, -0.4549, 0.057, -0.0, -0.457, 0.0548, -0.0, -0.457, 0.057, -0.0, -0.4549, 0.0548, -0.0, -0.4549, 0.0613, -0.0, -0.4527, 0.0591, -0.0, -0.4527, 0.0613, -0.0, -0.4505, 0.0591, -0.0, -0.4505, 0.057, -0.0, -0.4527, 0.0548, -0.0, -0.4527, 0.057, -0.0, -0.4505, 0.0548, -0.0, -0.4505, 0.0526, -0.0, -0.4657, 0.0505, -0.0, -0.4657, 0.0526, -0.0, -0.4635, 0.0505, -0.0, -0.4635, 0.0483, -0.0, -0.4656, 0.0488, -0.0, -0.4635, 0.0482, 0.0017, -0.4657, 0.0482, 0.0039, -0.4657, 0.0487, 0.0017, -0.4635, 0.0487, 0.0039, -0.4635,
0.0526, -0.0, -0.4614, 0.0493, -0.0, -0.461, 0.0526, -0.0, -0.4592, 0.0497, -0.0, -0.459, 0.0491, -0.0, -0.4618, 0.0494, 0.0017, -0.4604, 0.0494, 0.0039, -0.4604, 0.0497, 0.0017, -0.4592, 0.0497, 0.0039, -0.4592, 0.0492, 0.0017, -0.4615, 0.0492, 0.0039, -0.4615, 0.0411, -0.0, -0.4658, 0.0416, -0.0, -0.4636, 0.0396, -0.0, -0.4657, 0.0374, -0.0, -0.4657, 0.0396, -0.0, -0.4635, 0.0374, -0.0, -0.4635, 0.0411, 0.0017, -0.4657, 0.0411, 0.0039, -0.4657, 0.0416, 0.0017, -0.4635, 0.0416, 0.0039, -0.4635, 0.0421, -0.0, -0.4615, 0.0424, -0.0, -0.4593, 0.0396, -0.0, -0.4614, 0.0374, -0.0, -0.4614, 0.0396, -0.0, -0.4592, 0.0374, -0.0, -0.4592, 0.0421, 0.0017, -0.4614, 0.0421, 0.0039, -0.4614, 0.0424, 0.0017, -0.4592,
0.0424, 0.0039, -0.4592, 0.0526, -0.0, -0.457, 0.05, -0.0, -0.4569, 0.0526, -0.0, -0.4549, 0.0502, -0.0, -0.4548, 0.05, 0.0017, -0.457, 0.05, 0.0039, -0.457, 0.0502, 0.0017, -0.4549, 0.0502, 0.0039, -0.4549, 0.0526, -0.0, -0.4527, 0.0503, -0.0, -0.4527, 0.0526, -0.0, -0.4505, 0.0503, -0.0, -0.4505, 0.0503, 0.0017, -0.4527, 0.0503, 0.0039, -0.4527, 0.0503, 0.0017, -0.4505, 0.0503, 0.0039, -0.4505, 0.0425, -0.0, -0.457, 0.0425, -0.0, -0.4549, 0.0396, -0.0, -0.457, 0.0374, -0.0, -0.457, 0.0396, -0.0, -0.4549, 0.0374, -0.0, -0.4549, 0.0425, 0.0017, -0.457, 0.0425, 0.0039, -0.457, 0.0425, 0.0017, -0.4549, 0.0425, 0.0039, -0.4549, 0.0425, -0.0, -0.4527, 0.0425, -0.0, -0.4505, 0.0396, -0.0, -0.4527,
0.0374, -0.0, -0.4527, 0.0396, -0.0, -0.4505, 0.0374, -0.0, -0.4505, 0.0425, 0.0017, -0.4527, 0.0425, 0.0039, -0.4527, 0.0425, 0.0017, -0.4505, 0.0425, 0.0039, -0.4505, 0.0345, -0.0, -0.4824, 0.0338, -0.0, -0.4835, 0.0353, -0.0, -0.4812, 0.0331, -0.0, -0.4809, 0.0309, -0.0, -0.4831, 0.0288, -0.0, -0.4831, 0.0309, -0.0, -0.4809, 0.0288, -0.0, -0.4809, 0.0345, 0.0017, -0.4824, 0.0337, 0.0017, -0.4835, 0.0345, 0.0039, -0.4824, 0.0337, 0.0039, -0.4835, 0.0355, 0.0017, -0.4809, 0.0355, 0.0039, -0.4809, 0.0363, -0.0, -0.4793, 0.0331, -0.0, -0.4787, 0.0353, -0.0, -0.4766, 0.0331, -0.0, -0.4766, 0.0309, -0.0, -0.4787, 0.0288, -0.0, -0.4787, 0.0309, -0.0, -0.4766, 0.0279, -0.0, -0.4765, 0.0362, 0.0017, -0.4795,
0.0362, 0.0039, -0.4795, 0.0279, 0.0017, -0.4764, 0.0279, 0.0039, -0.4764, 0.0259, -0.0, -0.483, 0.0255, -0.0, -0.4841, 0.0267, -0.0, -0.4807, 0.0259, 0.0017, -0.483, 0.0255, 0.0017, -0.4841, 0.0259, 0.0039, -0.483, 0.0255, 0.0039, -0.4841, 0.0266, 0.0017, -0.4809, 0.0266, 0.0039, -0.4809, 0.0273, -0.0, -0.4787, 0.0276, -0.0, -0.4775, 0.0273, 0.0017, -0.4787, 0.0273, 0.0039, -0.4787, 0.0276, 0.0017, -0.4775, 0.0276, 0.0039, -0.4775, 0.0353, -0.0, -0.4744, 0.0331, -0.0, -0.4744, 0.0353, -0.0, -0.4722, 0.0331, -0.0, -0.4722, 0.0309, -0.0, -0.4744, 0.0284, -0.0, -0.4743, 0.0309, -0.0, -0.4722, 0.0286, -0.0, -0.4722, 0.0283, 0.0017, -0.4744, 0.0283, 0.0039, -0.4744, 0.0286, 0.0017, -0.4722, 0.0286, 0.0039, -0.4722,
0.0353, -0.0, -0.47, 0.0331, -0.0, -0.47, 0.0353, -0.0, -0.4679, 0.0331, -0.0, -0.4679, 0.0309, -0.0, -0.47, 0.0287, -0.0, -0.47, 0.0309, -0.0, -0.4679, 0.0287, -0.0, -0.4679, 0.0287, 0.0017, -0.47, 0.0287, 0.0039, -0.47, 0.0287, 0.0017, -0.4679, 0.0287, 0.0039, -0.4679, 0.0148, -0.0, -0.4801, 0.014, -0.0, -0.4824, 0.0114, -0.0, -0.4823, 0.0145, -0.0, -0.4811, 0.0114, -0.0, -0.4809, 0.0148, 0.0017, -0.4801, 0.0148, 0.0039, -0.4801, 0.014, 0.0017, -0.4824, 0.0114, 0.0017, -0.4823, 0.014, 0.0039, -0.4824, 0.0114, 0.0039, -0.4823, 0.0144, 0.0017, -0.4812, 0.0144, 0.0039, -0.4812, 0.0152, -0.0, -0.4788, 0.0158, -0.0, -0.4767, 0.0136, -0.0, -0.4787, 0.0114, -0.0, -0.4787, 0.0136, -0.0, -0.4766,
0.0114, -0.0, -0.4766, 0.0152, 0.0017, -0.4787, 0.0152, 0.0039, -0.4787, 0.0159, 0.0017, -0.4766, 0.0159, 0.0039, -0.4766, 0.0097, -0.0, -0.4821, 0.0089, -0.0, -0.482, 0.0069, -0.0, -0.4817, 0.0048, -0.0, -0.4814, 0.0026, -0.0, -0.4811, 0.0094, 0.0017, -0.4821, 0.0094, 0.0039, -0.4821, 0.0083, 0.0017, -0.482, 0.0071, 0.0017, -0.4818, 0.0083, 0.0039, -0.482, 0.0071, 0.0039, -0.4818, 0.0049, 0.0017, -0.4815, 0.0027, 0.0017, -0.4812, 0.0049, 0.0039, -0.4815, 0.0027, 0.0039, -0.4812, 0.0092, -0.0, -0.4787, 0.0071, -0.0, -0.4787, 0.0092, -0.0, -0.4766, 0.0071, -0.0, -0.4766, 0.0049, -0.0, -0.4787, 0.0027, -0.0, -0.4787, 0.0049, -0.0, -0.4766, 0.0027, -0.0, -0.4766, 0.0164, -0.0, -0.4746, 0.0168, -0.0, -0.4713,
0.0168, -0.0, -0.4722, 0.0136, -0.0, -0.4744, 0.0114, -0.0, -0.4744, 0.0136, -0.0, -0.4722, 0.0114, -0.0, -0.4722, 0.0164, 0.0017, -0.4744, 0.0164, 0.0039, -0.4744, 0.0168, 0.0017, -0.4713, 0.0167, 0.0017, -0.4724, 0.0168, 0.0039, -0.4713, 0.0167, 0.0039, -0.4724, 0.0169, -0.0, -0.47, 0.0157, -0.0, -0.47, 0.017, -0.0, -0.4679, 0.0157, -0.0, -0.4679, 0.0136, -0.0, -0.47, 0.0114, -0.0, -0.47, 0.0136, -0.0, -0.4679, 0.0114, -0.0, -0.4679, 0.0169, 0.0017, -0.47, 0.0169, 0.0039, -0.47, 0.017, 0.0017, -0.4679, 0.017, 0.0039, -0.4679, 0.0092, -0.0, -0.4744, 0.0071, -0.0, -0.4744, 0.0092, -0.0, -0.4722, 0.0071, -0.0, -0.4722, 0.0049, -0.0, -0.4744, 0.0027, -0.0, -0.4744, 0.0049, -0.0, -0.4722,
0.0027, -0.0, -0.4722, 0.0092, -0.0, -0.47, 0.0071, -0.0, -0.47, 0.0092, -0.0, -0.4679, 0.0071, -0.0, -0.4679, 0.0049, -0.0, -0.47, 0.0027, -0.0, -0.47, 0.0049, -0.0, -0.4679, 0.0027, -0.0, -0.4679, 0.0353, -0.0, -0.4657, 0.0331, -0.0, -0.4657, 0.0353, -0.0, -0.4635, 0.0331, -0.0, -0.4635, 0.0309, -0.0, -0.4657, 0.0287, -0.0, -0.4657, 0.0309, -0.0, -0.4635, 0.0287, -0.0, -0.4636, 0.0287, 0.0017, -0.4657, 0.0287, 0.0039, -0.4657, 0.0287, 0.0017, -0.4635, 0.0287, 0.0039, -0.4635, 0.0353, -0.0, -0.4614, 0.0331, -0.0, -0.4614, 0.0353, -0.0, -0.4592, 0.0331, -0.0, -0.4592, 0.0309, -0.0, -0.4614, 0.0284, -0.0, -0.4615, 0.0309, -0.0, -0.4592, 0.028, -0.0, -0.4594, 0.0284, 0.0017, -0.4614,
0.0284, 0.0039, -0.4614, 0.028, 0.0017, -0.4592, 0.028, 0.0039, -0.4592, 0.0353, -0.0, -0.457, 0.0331, -0.0, -0.457, 0.0353, -0.0, -0.4549, 0.0331, -0.0, -0.4549, 0.0309, -0.0, -0.457, 0.0276, -0.0, -0.4574, 0.0309, -0.0, -0.4549, 0.0288, -0.0, -0.4549, 0.0277, 0.0017, -0.458, 0.0277, 0.0039, -0.458, 0.0353, -0.0, -0.4527, 0.0331, -0.0, -0.4527, 0.0353, -0.0, -0.4505, 0.0331, -0.0, -0.4505, 0.0309, -0.0, -0.4527, 0.0288, -0.0, -0.4527, 0.0309, -0.0, -0.4505, 0.0288, -0.0, -0.4505, 0.0274, -0.0, -0.4566, 0.027, -0.0, -0.4549, 0.0274, 0.0017, -0.4569, 0.0274, 0.0039, -0.4569, 0.027, 0.0017, -0.4549, 0.027, 0.0039, -0.4549, 0.0265, -0.0, -0.4528, 0.0261, -0.0, -0.4507, 0.0265, 0.0017, -0.4527,
0.0265, 0.0039, -0.4527, 0.026, 0.0017, -0.4505, 0.026, 0.0039, -0.4505, 0.017, -0.0, -0.4657, 0.0158, -0.0, -0.4657, 0.0136, -0.0, -0.4655, 0.0115, -0.0, -0.4652, 0.017, 0.0017, -0.4657, 0.0157, 0.0017, -0.4657, 0.017, 0.0039, -0.4657, 0.0157, 0.0039, -0.4657, 0.0136, 0.0017, -0.4655, 0.0114, 0.0017, -0.4651, 0.0136, 0.0039, -0.4655, 0.0114, 0.0039, -0.4651, 0.0093, -0.0, -0.4647, 0.0071, -0.0, -0.4657, 0.0083, -0.0, -0.4645, 0.0071, -0.0, -0.4643, 0.0049, -0.0, -0.4657, 0.0027, -0.0, -0.4657, 0.005, -0.0, -0.4638, 0.0029, -0.0, -0.4633, 0.0095, 0.0017, -0.4648, 0.0095, 0.0039, -0.4648, 0.0084, 0.0017, -0.4646, 0.0071, 0.0017, -0.4643, 0.0084, 0.0039, -0.4646, 0.0071, 0.0039, -0.4643, 0.0049, 0.0017, -0.4638,
0.0027, 0.0017, -0.4633, 0.0049, 0.0039, -0.4638, 0.0027, 0.0039, -0.4633, 0.07, -0.0, -0.4483, 0.0678, -0.0, -0.4483, 0.07, -0.0, -0.4462, 0.0678, -0.0, -0.4462, 0.0656, -0.0, -0.4483, 0.0635, -0.0, -0.4483, 0.0656, -0.0, -0.4462, 0.0635, -0.0, -0.4462, 0.07, -0.0, -0.444, 0.0678, -0.0, -0.444, 0.07, -0.0, -0.4418, 0.0678, -0.0, -0.4418, 0.0656, -0.0, -0.444, 0.0635, -0.0, -0.444, 0.0656, -0.0, -0.4418, 0.0635, -0.0, -0.4418, 0.0613, -0.0, -0.4483, 0.0591, -0.0, -0.4483, 0.0613, -0.0, -0.4462, 0.0591, -0.0, -0.4462, 0.057, -0.0, -0.4483, 0.0548, -0.0, -0.4483, 0.057, -0.0, -0.4462, 0.0548, -0.0, -0.4462, 0.0613, -0.0, -0.444, 0.0591, -0.0, -0.444, 0.0613, -0.0, -0.4418,
0.0591, -0.0, -0.4418, 0.057, -0.0, -0.444, 0.0548, -0.0, -0.444, 0.057, -0.0, -0.4418, 0.0548, -0.0, -0.4418, 0.07, -0.0, -0.4397, 0.0678, -0.0, -0.4397, 0.07, -0.0, -0.4375, 0.0678, -0.0, -0.4375, 0.0656, -0.0, -0.4397, 0.0635, -0.0, -0.4397, 0.0656, -0.0, -0.4375, 0.0635, -0.0, -0.4375, 0.07, -0.0, -0.4353, 0.0678, -0.0, -0.4353, 0.069, -0.0, -0.433, 0.0678, -0.0, -0.4332, 0.0656, -0.0, -0.4353, 0.0635, -0.0, -0.4353, 0.0656, -0.0, -0.4332, 0.0635, -0.0, -0.4332, 0.0613, -0.0, -0.4397, 0.0591, -0.0, -0.4397, 0.0613, -0.0, -0.4375, 0.0591, -0.0, -0.4375, 0.057, -0.0, -0.4397, 0.0548, -0.0, -0.4397, 0.057, -0.0, -0.4375, 0.0548, -0.0, -0.4375, 0.0613, -0.0, -0.4353,
0.0591, -0.0, -0.4353, 0.0613, -0.0, -0.4332, 0.0591, -0.0, -0.4332, 0.057, -0.0, -0.4353, 0.0548, -0.0, -0.4353, 0.057, -0.0, -0.4332, 0.0548, -0.0, -0.4332, 0.0526, -0.0, -0.4483, 0.0503, -0.0, -0.4484, 0.0526, -0.0, -0.4462, 0.0501, -0.0, -0.4463, 0.0503, 0.0017, -0.4483, 0.0503, 0.0039, -0.4483, 0.0501, 0.0017, -0.4462, 0.0501, 0.0039, -0.4462, 0.0526, -0.0, -0.444, 0.0498, -0.0, -0.4441, 0.0526, -0.0, -0.4418, 0.0494, -0.0, -0.442, 0.0492, -0.0, -0.4411, 0.0497, 0.0017, -0.444, 0.0497, 0.0039, -0.444, 0.0495, 0.0017, -0.4424, 0.0495, 0.0039, -0.4424, 0.0493, 0.0017, -0.4413, 0.0493, 0.0039, -0.4413, 0.0425, -0.0, -0.4483, 0.0423, -0.0, -0.4461, 0.0396, -0.0, -0.4483, 0.0374, -0.0, -0.4483,
0.0396, -0.0, -0.4462, 0.0374, -0.0, -0.4462, 0.0425, 0.0017, -0.4483, 0.0425, 0.0039, -0.4483, 0.0423, 0.0017, -0.4462, 0.0423, 0.0039, -0.4462, 0.042, -0.0, -0.4439, 0.0416, -0.0, -0.4417, 0.0396, -0.0, -0.444, 0.0374, -0.0, -0.444, 0.0396, -0.0, -0.4418, 0.0374, -0.0, -0.4418, 0.042, 0.0017, -0.444, 0.042, 0.0039, -0.444, 0.0416, 0.0017, -0.4418, 0.0416, 0.0039, -0.4418, 0.0526, -0.0, -0.4397, 0.0505, -0.0, -0.4397, 0.0526, -0.0, -0.4375, 0.0505, -0.0, -0.4375, 0.049, -0.0, -0.4397, 0.0486, -0.0, -0.4376, 0.049, 0.0017, -0.4397, 0.049, 0.0039, -0.4397, 0.0486, 0.0017, -0.4375, 0.0486, 0.0039, -0.4375, 0.0526, -0.0, -0.4353, 0.0505, -0.0, -0.4353, 0.0526, -0.0, -0.4332, 0.0505, -0.0, -0.4332,
0.0482, -0.0, -0.4355, 0.0476, -0.0, -0.4334, 0.0481, 0.0017, -0.4353, 0.0481, 0.0039, -0.4353, 0.0476, 0.0017, -0.4331, 0.0476, 0.0039, -0.4331, 0.0412, -0.0, -0.4396, 0.0409, -0.0, -0.4382, 0.0396, -0.0, -0.4397, 0.0374, -0.0, -0.4397, 0.0406, -0.0, -0.4373, 0.0374, -0.0, -0.4375, 0.0412, 0.0017, -0.4397, 0.0412, 0.0039, -0.4397, 0.0408, 0.0017, -0.4381, 0.0408, 0.0039, -0.4381, 0.0406, 0.0017, -0.437, 0.0406, 0.0039, -0.437, 0.0401, -0.0, -0.4351, 0.0374, -0.0, -0.4353, 0.0396, -0.0, -0.433, 0.0374, -0.0, -0.4332, 0.0401, 0.0017, -0.4353, 0.0401, 0.0039, -0.4353, 0.0396, 0.0017, -0.4331, 0.0396, 0.0039, -0.4331, 0.069, -0.0, -0.4328, 0.0689, -0.0, -0.4311, 0.0686, -0.0, -0.4286, 0.0656, -0.0, -0.431,
0.0635, -0.0, -0.431, 0.0656, -0.0, -0.4288, 0.0635, -0.0, -0.4288, 0.069, 0.0017, -0.4328, 0.0689, 0.0017, -0.4308, 0.069, 0.0039, -0.4328, 0.0689, 0.0039, -0.4308, 0.0686, 0.0017, -0.4288, 0.0686, 0.0039, -0.4288, 0.0681, -0.0, -0.4265, 0.0675, -0.0, -0.4244, 0.0656, -0.0, -0.4266, 0.0635, -0.0, -0.4266, 0.0656, -0.0, -0.4245, 0.0635, -0.0, -0.4245, 0.0681, 0.0017, -0.4266, 0.0681, 0.0039, -0.4266, 0.0675, 0.0017, -0.4245, 0.0675, 0.0039, -0.4245, 0.0613, -0.0, -0.431, 0.0591, -0.0, -0.431, 0.0613, -0.0, -0.4288, 0.0591, -0.0, -0.4288, 0.057, -0.0, -0.431, 0.0548, -0.0, -0.431, 0.057, -0.0, -0.4288, 0.0548, -0.0, -0.4288, 0.0613, -0.0, -0.4266, 0.0591, -0.0, -0.4266, 0.0613, -0.0, -0.4245,
0.0591, -0.0, -0.4245, 0.057, -0.0, -0.4266, 0.0548, -0.0, -0.4266, 0.057, -0.0, -0.4245, 0.0548, -0.0, -0.4245, 0.067, -0.0, -0.4229, 0.0667, -0.0, -0.4219, 0.0635, -0.0, -0.4223, 0.0659, -0.0, -0.4199, 0.0635, -0.0, -0.4201, 0.067, 0.0017, -0.4227, 0.067, 0.0039, -0.4227, 0.0666, 0.0017, -0.4216, 0.0666, 0.0039, -0.4216, 0.066, 0.0017, -0.4201, 0.066, 0.0039, -0.4201, 0.0651, -0.0, -0.4179, 0.0635, -0.0, -0.418, 0.0647, -0.0, -0.4167, 0.0642, -0.0, -0.4157, 0.0651, 0.0017, -0.418, 0.0651, 0.0039, -0.418, 0.0646, 0.0017, -0.4167, 0.0642, 0.0017, -0.4156, 0.0646, 0.0039, -0.4167, 0.0642, 0.0039, -0.4156, 0.0613, -0.0, -0.4223, 0.0591, -0.0, -0.4223, 0.0613, -0.0, -0.4201, 0.0591, -0.0, -0.4201,
0.057, -0.0, -0.4223, 0.0548, -0.0, -0.4223, 0.057, -0.0, -0.4201, 0.0548, -0.0, -0.4201, 0.0613, -0.0, -0.418, 0.0591, -0.0, -0.418, 0.0613, -0.0, -0.4158, 0.0591, -0.0, -0.4158, 0.057, -0.0, -0.418, 0.0548, -0.0, -0.418, 0.057, -0.0, -0.4158, 0.0548, -0.0, -0.4158, 0.0526, -0.0, -0.431, 0.0505, -0.0, -0.431, 0.0526, -0.0, -0.4288, 0.0505, -0.0, -0.4288, 0.0471, -0.0, -0.4315, 0.0468, -0.0, -0.4306, 0.0483, -0.0, -0.4288, 0.0463, -0.0, -0.429, 0.0472, 0.0017, -0.432, 0.0469, 0.0017, -0.4309, 0.0472, 0.0039, -0.432, 0.0469, 0.0039, -0.4309, 0.0462, 0.0017, -0.4288, 0.0462, 0.0039, -0.4288, 0.0526, -0.0, -0.4266, 0.0505, -0.0, -0.4266, 0.0526, -0.0, -0.4245, 0.0505, -0.0, -0.4245,
0.0483, -0.0, -0.4266, 0.0454, -0.0, -0.4267, 0.0483, -0.0, -0.4245, 0.0461, -0.0, -0.4245, 0.0454, 0.0017, -0.4267, 0.0454, 0.0039, -0.4267, 0.039, -0.0, -0.4309, 0.0374, -0.0, -0.431, 0.0386, -0.0, -0.4298, 0.0382, -0.0, -0.4287, 0.039, 0.0017, -0.431, 0.039, 0.0039, -0.431, 0.0386, 0.0017, -0.4297, 0.0382, 0.0017, -0.4286, 0.0386, 0.0039, -0.4297, 0.0382, 0.0039, -0.4286, 0.045, -0.0, -0.4256, 0.0446, -0.0, -0.4245, 0.0373, -0.0, -0.4265, 0.0367, -0.0, -0.4251, 0.045, 0.0017, -0.4256, 0.045, 0.0039, -0.4256, 0.0446, 0.0017, -0.4245, 0.0446, 0.0039, -0.4245, 0.0373, 0.0017, -0.4266, 0.0373, 0.0039, -0.4266, 0.0366, 0.0017, -0.425, 0.0366, 0.0039, -0.425, 0.0526, -0.0, -0.4223, 0.0505, -0.0, -0.4223,
0.0526, -0.0, -0.4201, 0.0505, -0.0, -0.4201, 0.0483, -0.0, -0.4223, 0.0461, -0.0, -0.4223, 0.0483, -0.0, -0.4201, 0.0461, -0.0, -0.4201, 0.0526, -0.0, -0.418, 0.0505, -0.0, -0.418, 0.0526, -0.0, -0.4158, 0.0505, -0.0, -0.4158, 0.0483, -0.0, -0.418, 0.0461, -0.0, -0.418, 0.0483, -0.0, -0.4158, 0.0461, -0.0, -0.4158, 0.0437, -0.0, -0.4226, 0.0429, -0.0, -0.4206, 0.0425, -0.0, -0.4196, 0.0436, 0.0017, -0.4223, 0.0436, 0.0039, -0.4223, 0.043, 0.0017, -0.4209, 0.0425, 0.0017, -0.4198, 0.043, 0.0039, -0.4209, 0.0425, 0.0039, -0.4198, 0.0439, -0.0, -0.418, 0.0419, -0.0, -0.4182, 0.0439, -0.0, -0.4158, 0.0409, -0.0, -0.4161, 0.0404, -0.0, -0.4151, 0.0418, 0.0017, -0.418, 0.0418, 0.0039, -0.418,
0.041, 0.0017, -0.4162, 0.041, 0.0039, -0.4162, 0.0405, 0.0017, -0.4152, 0.0405, 0.0039, -0.4152, 0.0353, -0.0, -0.4483, 0.0331, -0.0, -0.4483, 0.0353, -0.0, -0.4462, 0.0331, -0.0, -0.4462, 0.0309, -0.0, -0.4483, 0.0288, -0.0, -0.4483, 0.0309, -0.0, -0.4462, 0.0288, -0.0, -0.4462, 0.0353, -0.0, -0.444, 0.0331, -0.0, -0.444, 0.0353, -0.0, -0.4418, 0.0331, -0.0, -0.4418, 0.0309, -0.0, -0.444, 0.0288, -0.0, -0.444, 0.0309, -0.0, -0.4418, 0.0288, -0.0, -0.4418, 0.0255, -0.0, -0.4486, 0.0253, -0.0, -0.4477, 0.0266, -0.0, -0.4462, 0.0249, -0.0, -0.4463, 0.0256, 0.0017, -0.4489, 0.0253, 0.0017, -0.4479, 0.0256, 0.0039, -0.4489, 0.0253, 0.0039, -0.4479, 0.0248, 0.0017, -0.4462, 0.0248, 0.0039, -0.4462,
0.0266, -0.0, -0.444, 0.0242, -0.0, -0.4442, 0.0266, -0.0, -0.4418, 0.0234, -0.0, -0.4422, 0.0231, -0.0, -0.4412, 0.0241, 0.0017, -0.444, 0.0241, 0.0039, -0.444, 0.0235, 0.0017, -0.4424, 0.0235, 0.0039, -0.4424, 0.0231, 0.0017, -0.4413, 0.0231, 0.0039, -0.4413, 0.0353, -0.0, -0.4397, 0.0331, -0.0, -0.4397, 0.0353, -0.0, -0.4375, 0.0331, -0.0, -0.4375, 0.0309, -0.0, -0.4397, 0.0288, -0.0, -0.4397, 0.0309, -0.0, -0.4375, 0.0288, -0.0, -0.4375, 0.0353, -0.0, -0.4353, 0.0331, -0.0, -0.4353, 0.0353, -0.0, -0.4332, 0.0331, -0.0, -0.4332, 0.0309, -0.0, -0.4353, 0.0288, -0.0, -0.4353, 0.0309, -0.0, -0.4332, 0.0288, -0.0, -0.4332, 0.0266, -0.0, -0.4397, 0.0244, -0.0, -0.4397, 0.0266, -0.0, -0.4375,
0.0244, -0.0, -0.4375, 0.0225, -0.0, -0.4398, 0.0216, -0.0, -0.4375, 0.0212, -0.0, -0.4364, 0.0225, 0.0017, -0.4397, 0.0225, 0.0039, -0.4397, 0.0216, 0.0017, -0.4375, 0.0212, 0.0017, -0.4364, 0.0216, 0.0039, -0.4375, 0.0212, 0.0039, -0.4364, 0.0266, -0.0, -0.4353, 0.0244, -0.0, -0.4353, 0.0266, -0.0, -0.4332, 0.0244, -0.0, -0.4332, 0.0222, -0.0, -0.4353, 0.0207, -0.0, -0.4354, 0.0222, -0.0, -0.4332, 0.0199, -0.0, -0.4334, 0.0207, 0.0017, -0.4353, 0.0207, 0.0039, -0.4353, 0.0197, 0.0017, -0.4332, 0.0197, 0.0039, -0.4332, 0.0353, -0.0, -0.431, 0.0331, -0.0, -0.431, 0.0353, -0.0, -0.4288, 0.0331, -0.0, -0.4288, 0.0309, -0.0, -0.431, 0.0288, -0.0, -0.431, 0.0309, -0.0, -0.4288, 0.0288, -0.0, -0.4288,
0.0353, -0.0, -0.4266, 0.0331, -0.0, -0.4266, 0.0362, -0.0, -0.4241, 0.0331, -0.0, -0.4245, 0.0309, -0.0, -0.4266, 0.0288, -0.0, -0.4266, 0.0309, -0.0, -0.4245, 0.0288, -0.0, -0.4245, 0.0361, 0.0017, -0.4239, 0.0361, 0.0039, -0.4239, 0.0266, -0.0, -0.431, 0.0244, -0.0, -0.431, 0.0266, -0.0, -0.4288, 0.0244, -0.0, -0.4288, 0.0222, -0.0, -0.431, 0.019, -0.0, -0.4315, 0.0222, -0.0, -0.4288, 0.0201, -0.0, -0.4288, 0.0191, 0.0017, -0.4318, 0.0191, 0.0039, -0.4318, 0.0266, -0.0, -0.4266, 0.0244, -0.0, -0.4266, 0.0266, -0.0, -0.4245, 0.0244, -0.0, -0.4245, 0.0222, -0.0, -0.4266, 0.0201, -0.0, -0.4266, 0.0222, -0.0, -0.4245, 0.0201, -0.0, -0.4245, 0.0353, -0.0, -0.4221, 0.0331, -0.0, -0.4223,
0.0346, -0.0, -0.4206, 0.0342, -0.0, -0.4196, 0.0309, -0.0, -0.4223, 0.0288, -0.0, -0.4223, 0.0309, -0.0, -0.4201, 0.0288, -0.0, -0.4201, 0.0354, 0.0017, -0.4223, 0.0354, 0.0039, -0.4223, 0.0345, 0.0017, -0.4204, 0.034, 0.0017, -0.4194, 0.0345, 0.0039, -0.4204, 0.034, 0.0039, -0.4194, 0.0332, -0.0, -0.4177, 0.0325, -0.0, -0.4163, 0.0309, -0.0, -0.418, 0.0288, -0.0, -0.418, 0.0319, -0.0, -0.4153, 0.0288, -0.0, -0.4158, 0.0333, 0.0017, -0.4179, 0.0333, 0.0039, -0.4179, 0.0324, 0.0017, -0.4162, 0.0324, 0.0039, -0.4162, 0.0318, 0.0017, -0.4151, 0.0318, 0.0039, -0.4151, 0.0266, -0.0, -0.4223, 0.0244, -0.0, -0.4223, 0.0266, -0.0, -0.4201, 0.0244, -0.0, -0.4201, 0.0222, -0.0, -0.4223, 0.0201, -0.0, -0.4223,
0.0222, -0.0, -0.4201, 0.0201, -0.0, -0.4201, 0.0266, -0.0, -0.418, 0.0244, -0.0, -0.418, 0.0266, -0.0, -0.4158, 0.0244, -0.0, -0.4158, 0.0222, -0.0, -0.418, 0.0201, -0.0, -0.418, 0.0222, -0.0, -0.4158, 0.0201, -0.0, -0.4158, 0.0185, -0.0, -0.4305, 0.0179, -0.0, -0.429, 0.0186, 0.0017, -0.4307, 0.0186, 0.0039, -0.4307, 0.0178, 0.0017, -0.4288, 0.0178, 0.0039, -0.4288, 0.017, -0.0, -0.427, 0.0165, -0.0, -0.426, 0.0179, -0.0, -0.4245, 0.0158, -0.0, -0.4247, 0.017, 0.0017, -0.4272, 0.0165, 0.0017, -0.4262, 0.017, 0.0039, -0.4272, 0.0165, 0.0039, -0.4262, 0.0157, 0.0017, -0.4245, 0.0157, 0.0039, -0.4245, 0.0018, -0.0, -0.4257, 0.0023, -0.0, -0.4246, 0.0017, 0.0017, -0.4258, 0.0017, 0.0039, -0.4258,
0.0023, 0.0017, -0.4245, 0.0023, 0.0039, -0.4245, 0.0179, -0.0, -0.4223, 0.0148, -0.0, -0.4228, 0.0179, -0.0, -0.4201, 0.0157, -0.0, -0.4201, 0.0142, -0.0, -0.4217, 0.0135, -0.0, -0.4204, 0.0149, 0.0017, -0.423, 0.0149, 0.0039, -0.423, 0.0143, 0.0017, -0.4219, 0.0143, 0.0039, -0.4219, 0.0134, 0.0017, -0.4201, 0.0134, 0.0039, -0.4201, 0.0179, -0.0, -0.418, 0.0157, -0.0, -0.418, 0.0179, -0.0, -0.4158, 0.0157, -0.0, -0.4158, 0.0126, -0.0, -0.4185, 0.012, -0.0, -0.4175, 0.0136, -0.0, -0.4158, 0.0113, -0.0, -0.4161, 0.0127, 0.0017, -0.4187, 0.0121, 0.0017, -0.4176, 0.0127, 0.0039, -0.4187, 0.0121, 0.0039, -0.4176, 0.0111, 0.0017, -0.4158, 0.0111, 0.0039, -0.4158, 0.0029, -0.0, -0.4224, 0.0032, -0.0, -0.4201,
0.003, 0.0017, -0.4223, 0.003, 0.0039, -0.4223, 0.0032, 0.0017, -0.4201, 0.0032, 0.0039, -0.4201, 0.003, -0.0, -0.4179, 0.0025, -0.0, -0.4156, 0.003, 0.0017, -0.4179, 0.003, 0.0039, -0.4179, 0.0026, 0.0017, -0.4158, 0.0026, 0.0039, -0.4158, 0.1059, -0.0, -0.3976, 0.1059, 0.0017, -0.3977, 0.1059, 0.0039, -0.3977, 0.1061, -0.0, -0.3963, 0.1066, -0.0, -0.3942, 0.1061, 0.0017, -0.3963, 0.1061, 0.0039, -0.3963, 0.1066, 0.0017, -0.3941, 0.1066, 0.0039, -0.3941, 0.1069, -0.0, -0.392, 0.1071, -0.0, -0.3898, 0.1069, 0.0017, -0.3919, 0.1069, 0.0039, -0.3919, 0.1071, 0.0017, -0.3898, 0.1071, 0.0039, -0.3898, 0.1072, -0.0, -0.3876, 0.1072, -0.0, -0.3854, 0.1072, 0.0017, -0.3876, 0.1072, 0.0039, -0.3876, 0.1072, 0.0017, -0.3854,
0.1072, 0.0039, -0.3854, 0.1072, -0.0, -0.3832, 0.1072, -0.0, -0.3811, 0.1072, 0.0017, -0.3832, 0.1072, 0.0039, -0.3832, 0.1072, 0.0017, -0.3811, 0.1072, 0.0039, -0.3811, 0.1017, -0.0, -0.4129, 0.1022, -0.0, -0.4116, 0.1013, -0.0, -0.4139, 0.0982, -0.0, -0.4136, 0.1004, -0.0, -0.4115, 0.0982, -0.0, -0.4115, 0.1016, 0.0017, -0.413, 0.1016, 0.0039, -0.413, 0.1022, 0.0017, -0.4115, 0.1022, 0.0039, -0.4115, 0.1012, 0.0017, -0.4141, 0.1012, 0.0039, -0.4141, 0.1029, -0.0, -0.4095, 0.1039, -0.0, -0.4065, 0.1036, -0.0, -0.4075, 0.1004, -0.0, -0.4093, 0.0982, -0.0, -0.4093, 0.1004, -0.0, -0.4071, 0.0982, -0.0, -0.4071, 0.103, 0.0017, -0.4093, 0.103, 0.0039, -0.4093, 0.1038, 0.0017, -0.4067, 0.1035, 0.0017, -0.4078,
0.1038, 0.0039, -0.4067, 0.1035, 0.0039, -0.4078, 0.096, -0.0, -0.4136, 0.0939, -0.0, -0.4136, 0.096, -0.0, -0.4115, 0.0939, -0.0, -0.4115, 0.0917, -0.0, -0.4136, 0.0895, -0.0, -0.4136, 0.0917, -0.0, -0.4115, 0.0895, -0.0, -0.4115, 0.096, -0.0, -0.4093, 0.0939, -0.0, -0.4093, 0.096, -0.0, -0.4071, 0.0939, -0.0, -0.4071, 0.0917, -0.0, -0.4093, 0.0895, -0.0, -0.4093, 0.0917, -0.0, -0.4071, 0.0895, -0.0, -0.4071, 0.1043, -0.0, -0.405, 0.1025, -0.0, -0.4049, 0.1048, -0.0, -0.4029, 0.1025, -0.0, -0.4028, 0.1004, -0.0, -0.4049, 0.0982, -0.0, -0.4049, 0.1004, -0.0, -0.4028, 0.0982, -0.0, -0.4028, 0.1043, 0.0017, -0.4049, 0.1043, 0.0039, -0.4049, 0.1048, 0.0017, -0.4028, 0.1048, 0.0039, -0.4028,
0.1053, -0.0, -0.4008, 0.1025, -0.0, -0.4006, 0.1057, -0.0, -0.3985, 0.1025, -0.0, -0.3984, 0.1004, -0.0, -0.4006, 0.0982, -0.0, -0.4006, 0.1004, -0.0, -0.3984, 0.0982, -0.0, -0.3984, 0.1053, 0.0017, -0.4006, 0.1053, 0.0039, -0.4006, 0.1056, 0.0017, -0.3988, 0.1056, 0.0039, -0.3988, 0.096, -0.0, -0.4049, 0.0939, -0.0, -0.4049, 0.096, -0.0, -0.4028, 0.0939, -0.0, -0.4028, 0.0917, -0.0, -0.4049, 0.0895, -0.0, -0.4049, 0.0917, -0.0, -0.4028, 0.0886, -0.0, -0.4026, 0.0887, 0.0017, -0.4024, 0.0887, 0.0039, -0.4024, 0.096, -0.0, -0.4006, 0.0939, -0.0, -0.4006, 0.096, -0.0, -0.3984, 0.0939, -0.0, -0.3984, 0.0917, -0.0, -0.4006, 0.0894, -0.0, -0.4004, 0.0917, -0.0, -0.3984, 0.09, -0.0, -0.3984,
0.0893, 0.0017, -0.4006, 0.0893, 0.0039, -0.4006, 0.0899, 0.0017, -0.3984, 0.0899, 0.0039, -0.3984, 0.0874, -0.0, -0.4136, 0.0842, -0.0, -0.4131, 0.0874, -0.0, -0.4115, 0.0851, -0.0, -0.4112, 0.0837, -0.0, -0.4141, 0.0843, 0.0017, -0.4129, 0.0843, 0.0039, -0.4129, 0.085, 0.0017, -0.4115, 0.085, 0.0039, -0.4115, 0.0838, 0.0017, -0.414, 0.0838, 0.0039, -0.414, 0.0862, -0.0, -0.4087, 0.0858, -0.0, -0.4097, 0.087, -0.0, -0.4068, 0.0863, 0.0017, -0.4084, 0.0859, 0.0017, -0.4095, 0.0863, 0.0039, -0.4084, 0.0859, 0.0039, -0.4095, 0.0869, 0.0017, -0.4071, 0.0869, 0.0039, -0.4071, 0.0878, -0.0, -0.4048, 0.0883, -0.0, -0.4036, 0.0877, 0.0017, -0.4049, 0.0877, 0.0039, -0.4049, 0.0883, 0.0017, -0.4035, 0.0883, 0.0039, -0.4035,
0.1047, -0.0, -0.3963, 0.1025, -0.0, -0.3963, 0.1047, -0.0, -0.3941, 0.1025, -0.0, -0.3941, 0.1004, -0.0, -0.3963, 0.0982, -0.0, -0.3963, 0.1004, -0.0, -0.3941, 0.0982, -0.0, -0.3941, 0.1047, -0.0, -0.3919, 0.1025, -0.0, -0.3919, 0.1047, -0.0, -0.3897, 0.1025, -0.0, -0.3897, 0.1004, -0.0, -0.3919, 0.0982, -0.0, -0.3919, 0.1004, -0.0, -0.3897, 0.0982, -0.0, -0.3897, 0.096, -0.0, -0.3963, 0.0939, -0.0, -0.3963, 0.096, -0.0, -0.3941, 0.0939, -0.0, -0.3941, 0.0905, -0.0, -0.3959, 0.0903, -0.0, -0.3967, 0.091, -0.0, -0.3939, 0.0907, 0.0017, -0.3954, 0.0904, 0.0017, -0.3965, 0.0907, 0.0039, -0.3954, 0.0904, 0.0039, -0.3965, 0.091, 0.0017, -0.3941, 0.091, 0.0039, -0.3941, 0.096, -0.0, -0.3919,
0.0939, -0.0, -0.3919, 0.096, -0.0, -0.3897, 0.0939, -0.0, -0.3897, 0.0916, -0.0, -0.3918, 0.0921, -0.0, -0.3897, 0.0915, 0.0017, -0.3919, 0.0915, 0.0039, -0.3919, 0.0921, 0.0017, -0.3898, 0.0921, 0.0039, -0.3898, 0.1047, -0.0, -0.3876, 0.1025, -0.0, -0.3876, 0.1047, -0.0, -0.3854, 0.1025, -0.0, -0.3854, 0.1004, -0.0, -0.3876, 0.0982, -0.0, -0.3876, 0.1004, -0.0, -0.3854, 0.0982, -0.0, -0.3854, 0.1047, -0.0, -0.3832, 0.1025, -0.0, -0.3832, 0.1047, -0.0, -0.3811, 0.1025, -0.0, -0.3811, 0.1004, -0.0, -0.3832, 0.0982, -0.0, -0.3832, 0.1004, -0.0, -0.3811, 0.0982, -0.0, -0.3811, 0.096, -0.0, -0.3876, 0.0939, -0.0, -0.3876, 0.096, -0.0, -0.3854, 0.0939, -0.0, -0.3854, 0.0924, -0.0, -0.3876,
0.0925, -0.0, -0.3854, 0.0924, 0.0017, -0.3876, 0.0924, 0.0039, -0.3876, 0.0925, 0.0017, -0.3854, 0.0925, 0.0039, -0.3854, 0.096, -0.0, -0.3832, 0.0928, -0.0, -0.3831, 0.096, -0.0, -0.3811, 0.0931, -0.0, -0.381, 0.0927, -0.0, -0.3839, 0.0928, 0.0017, -0.3827, 0.0928, 0.0039, -0.3827, 0.0931, 0.0017, -0.3811, 0.0931, 0.0039, -0.3811, 0.0927, 0.0017, -0.3838, 0.0927, 0.0039, -0.3838, 0.1072, -0.0, -0.3789, 0.1072, -0.0, -0.3767, 0.1072, 0.0017, -0.3789, 0.1072, 0.0039, -0.3789, 0.1072, 0.0017, -0.3767, 0.1072, 0.0039, -0.3767, 0.1072, -0.0, -0.3746, 0.1072, -0.0, -0.3724, 0.1072, 0.0017, -0.3746, 0.1072, 0.0039, -0.3746, 0.1072, 0.0017, -0.3724, 0.1072, 0.0039, -0.3724, 0.1072, -0.0, -0.3702, 0.1071, -0.0, -0.368,
0.1072, 0.0017, -0.3702, 0.1072, 0.0039, -0.3702, 0.1072, 0.0017, -0.368, 0.1072, 0.0039, -0.368, 0.107, -0.0, -0.3658, 0.1067, -0.0, -0.3636, 0.107, 0.0017, -0.3659, 0.107, 0.0039, -0.3659, 0.1067, 0.0017, -0.3637, 0.1067, 0.0039, -0.3637, 0.1062, -0.0, -0.3615, 0.1059, -0.0, -0.3601, 0.1062, 0.0017, -0.3615, 0.1062, 0.0039, -0.3615, 0.1059, 0.0017, -0.36, 0.1059, 0.0039, -0.36, 0.1047, -0.0, -0.3789, 0.1025, -0.0, -0.3789, 0.1047, -0.0, -0.3767, 0.1025, -0.0, -0.3767, 0.1004, -0.0, -0.3789, 0.0982, -0.0, -0.3789, 0.1004, -0.0, -0.3767, 0.0982, -0.0, -0.3767, 0.1047, -0.0, -0.3746, 0.1025, -0.0, -0.3746, 0.1047, -0.0, -0.3724, 0.1025, -0.0, -0.3724, 0.1004, -0.0, -0.3746, 0.0982, -0.0, -0.3746,
0.1004, -0.0, -0.3724, 0.0982, -0.0, -0.3724, 0.096, -0.0, -0.3789, 0.0933, -0.0, -0.3788, 0.096, -0.0, -0.3767, 0.0934, -0.0, -0.3767, 0.0933, 0.0017, -0.3789, 0.0933, 0.0039, -0.3789, 0.0934, 0.0017, -0.3767, 0.0934, 0.0039, -0.3767, 0.096, -0.0, -0.3746, 0.0934, -0.0, -0.3746, 0.096, -0.0, -0.3724, 0.0934, -0.0, -0.3724, 0.0934, 0.0017, -0.3746, 0.0934, 0.0039, -0.3746, 0.0934, 0.0017, -0.3724, 0.0934, 0.0039, -0.3724, 0.1047, -0.0, -0.3702, 0.1025, -0.0, -0.3702, 0.1047, -0.0, -0.368, 0.1025, -0.0, -0.368, 0.1004, -0.0, -0.3702, 0.0982, -0.0, -0.3702, 0.1004, -0.0, -0.368, 0.0982, -0.0, -0.368, 0.1047, -0.0, -0.3659, 0.1025, -0.0, -0.3659, 0.1047, -0.0, -0.3637, 0.1025, -0.0, -0.3637,
0.1004, -0.0, -0.3659, 0.0982, -0.0, -0.3659, 0.1004, -0.0, -0.3637, 0.0982, -0.0, -0.3637, 0.096, -0.0, -0.3702, 0.0934, -0.0, -0.3703, 0.096, -0.0, -0.368, 0.0931, -0.0, -0.3682, 0.0934, 0.0017, -0.3702, 0.0934, 0.0039, -0.3702, 0.0931, 0.0017, -0.368, 0.0931, 0.0039, -0.368, 0.096, -0.0, -0.3659, 0.0928, -0.0, -0.366, 0.096, -0.0, -0.3637, 0.0939, -0.0, -0.3637, 0.0927, -0.0, -0.3651, 0.0925, -0.0, -0.3637, 0.0929, 0.0017, -0.3664, 0.0929, 0.0039, -0.3664, 0.0927, 0.0017, -0.3653, 0.0927, 0.0039, -0.3653, 0.0925, 0.0017, -0.3637, 0.0925, 0.0039, -0.3637, 0.1047, -0.0, -0.3615, 0.1025, -0.0, -0.3615, 0.1057, -0.0, -0.3592, 0.1025, -0.0, -0.3594, 0.1004, -0.0, -0.3615, 0.0982, -0.0, -0.3615,
0.1004, -0.0, -0.3594, 0.0982, -0.0, -0.3594, 0.1057, 0.0017, -0.3589, 0.1057, 0.0039, -0.3589, 0.1052, -0.0, -0.357, 0.1025, -0.0, -0.3572, 0.1048, -0.0, -0.3549, 0.1025, -0.0, -0.355, 0.1004, -0.0, -0.3572, 0.0982, -0.0, -0.3572, 0.1004, -0.0, -0.355, 0.0982, -0.0, -0.355, 0.1053, 0.0017, -0.3572, 0.1053, 0.0039, -0.3572, 0.1048, 0.0017, -0.355, 0.1048, 0.0039, -0.355, 0.096, -0.0, -0.3615, 0.0939, -0.0, -0.3615, 0.096, -0.0, -0.3594, 0.0939, -0.0, -0.3594, 0.0924, -0.0, -0.3616, 0.0922, -0.0, -0.3594, 0.0924, 0.0017, -0.3615, 0.0924, 0.0039, -0.3615, 0.0922, 0.0017, -0.3594, 0.0922, 0.0039, -0.3594, 0.096, -0.0, -0.3572, 0.0939, -0.0, -0.3572, 0.096, -0.0, -0.355, 0.0939, -0.0, -0.355,
0.0918, -0.0, -0.3573, 0.0913, -0.0, -0.3552, 0.0918, 0.0017, -0.3572, 0.0918, 0.0039, -0.3572, 0.0912, 0.0017, -0.355, 0.0912, 0.0039, -0.355, 0.1043, -0.0, -0.3528, 0.1025, -0.0, -0.3529, 0.1039, -0.0, -0.3512, 0.1037, -0.0, -0.3503, 0.1004, -0.0, -0.3529, 0.0982, -0.0, -0.3529, 0.1004, -0.0, -0.3507, 0.0982, -0.0, -0.3507, 0.1043, 0.0017, -0.3529, 0.1043, 0.0039, -0.3529, 0.1039, 0.0017, -0.3509, 0.1036, 0.0017, -0.3498, 0.1039, 0.0039, -0.3509, 0.1036, 0.0039, -0.3498, 0.1031, -0.0, -0.3483, 0.1024, -0.0, -0.3462, 0.1004, -0.0, -0.3485, 0.0982, -0.0, -0.3485, 0.1004, -0.0, -0.3463, 0.0982, -0.0, -0.3463, 0.1032, 0.0017, -0.3485, 0.1032, 0.0039, -0.3485, 0.1025, 0.0017, -0.3463, 0.1025, 0.0039, -0.3463,
0.096, -0.0, -0.3529, 0.0939, -0.0, -0.3529, 0.096, -0.0, -0.3507, 0.0939, -0.0, -0.3507, 0.0907, -0.0, -0.3531, 0.0904, -0.0, -0.3521, 0.0917, -0.0, -0.3507, 0.0901, -0.0, -0.3508, 0.0908, 0.0017, -0.3534, 0.0905, 0.0017, -0.3523, 0.0908, 0.0039, -0.3534, 0.0905, 0.0039, -0.3523, 0.0901, 0.0017, -0.3507, 0.0901, 0.0039, -0.3507, 0.096, -0.0, -0.3485, 0.0939, -0.0, -0.3485, 0.096, -0.0, -0.3463, 0.0939, -0.0, -0.3463, 0.0917, -0.0, -0.3485, 0.0895, -0.0, -0.3487, 0.0917, -0.0, -0.3463, 0.0888, -0.0, -0.3464, 0.0895, 0.0017, -0.3485, 0.0895, 0.0039, -0.3485, 0.0888, 0.0017, -0.3465, 0.0888, 0.0039, -0.3465, 0.0884, -0.0, -0.3454, 0.0884, 0.0017, -0.3454, 0.0884, 0.0039, -0.3454, 0.0633, -0.0, -0.4134,
0.0627, -0.0, -0.412, 0.0634, 0.0017, -0.4136, 0.0634, 0.0039, -0.4136, 0.0627, 0.0017, -0.4119, 0.0627, 0.0039, -0.4119, 0.0613, -0.0, -0.4136, 0.0591, -0.0, -0.4136, 0.0623, -0.0, -0.411, 0.0591, -0.0, -0.4115, 0.057, -0.0, -0.4136, 0.0548, -0.0, -0.4136, 0.057, -0.0, -0.4115, 0.0548, -0.0, -0.4115, 0.0622, 0.0017, -0.4108, 0.0622, 0.0039, -0.4108, 0.0614, -0.0, -0.409, 0.0591, -0.0, -0.4093, 0.0607, -0.0, -0.4076, 0.0602, -0.0, -0.4065, 0.057, -0.0, -0.4093, 0.0548, -0.0, -0.4093, 0.057, -0.0, -0.4071, 0.0548, -0.0, -0.4071, 0.0615, 0.0017, -0.4093, 0.0615, 0.0039, -0.4093, 0.0606, 0.0017, -0.4074, 0.0601, 0.0017, -0.4063, 0.0606, 0.0039, -0.4074, 0.0601, 0.0039, -0.4063, 0.0592, -0.0, -0.4047,
0.0585, -0.0, -0.4033, 0.057, -0.0, -0.4049, 0.0548, -0.0, -0.4049, 0.0579, -0.0, -0.4023, 0.0548, -0.0, -0.4028, 0.0594, 0.0017, -0.4049, 0.0594, 0.0039, -0.4049, 0.0584, 0.0017, -0.4032, 0.0584, 0.0039, -0.4032, 0.0578, 0.0017, -0.4021, 0.0578, 0.0039, -0.4021, 0.0568, -0.0, -0.4004, 0.0548, -0.0, -0.4006, 0.0561, -0.0, -0.3992, 0.0555, -0.0, -0.3981, 0.0569, 0.0017, -0.4006, 0.0569, 0.0039, -0.4006, 0.0561, 0.0017, -0.3991, 0.0555, 0.0017, -0.398, 0.0561, 0.0039, -0.3991, 0.0555, 0.0039, -0.398, 0.0526, -0.0, -0.4136, 0.0505, -0.0, -0.4136, 0.0526, -0.0, -0.4115, 0.0505, -0.0, -0.4115, 0.0483, -0.0, -0.4136, 0.0461, -0.0, -0.4136, 0.0483, -0.0, -0.4115, 0.0461, -0.0, -0.4115, 0.0526, -0.0, -0.4093,
0.0505, -0.0, -0.4093, 0.0526, -0.0, -0.4071, 0.0505, -0.0, -0.4071, 0.0483, -0.0, -0.4093, 0.0461, -0.0, -0.4093, 0.0483, -0.0, -0.4071, 0.0461, -0.0, -0.4071, 0.0439, -0.0, -0.4136, 0.0418, -0.0, -0.4136, 0.0439, -0.0, -0.4115, 0.0418, -0.0, -0.4115, 0.0398, -0.0, -0.4138, 0.0388, -0.0, -0.4118, 0.0383, -0.0, -0.4107, 0.0397, 0.0017, -0.4136, 0.0397, 0.0039, -0.4136, 0.0389, 0.0017, -0.4119, 0.0383, 0.0017, -0.4108, 0.0389, 0.0039, -0.4119, 0.0383, 0.0039, -0.4108, 0.0439, -0.0, -0.4093, 0.0418, -0.0, -0.4093, 0.0439, -0.0, -0.4071, 0.0418, -0.0, -0.4071, 0.0396, -0.0, -0.4093, 0.0376, -0.0, -0.4095, 0.0396, -0.0, -0.4071, 0.0366, -0.0, -0.4076, 0.0375, 0.0017, -0.4093, 0.0375, 0.0039, -0.4093,
0.0366, 0.0017, -0.4077, 0.0366, 0.0039, -0.4077, 0.0526, -0.0, -0.4049, 0.0505, -0.0, -0.4049, 0.0526, -0.0, -0.4028, 0.0505, -0.0, -0.4028, 0.0483, -0.0, -0.4049, 0.0461, -0.0, -0.4049, 0.0483, -0.0, -0.4028, 0.0461, -0.0, -0.4028, 0.0526, -0.0, -0.4006, 0.0505, -0.0, -0.4006, 0.0526, -0.0, -0.3984, 0.0505, -0.0, -0.3984, 0.0483, -0.0, -0.4006, 0.0461, -0.0, -0.4006, 0.0483, -0.0, -0.3984, 0.0461, -0.0, -0.3984, 0.0439, -0.0, -0.4049, 0.0418, -0.0, -0.4049, 0.0439, -0.0, -0.4028, 0.0418, -0.0, -0.4028, 0.0396, -0.0, -0.4049, 0.0374, -0.0, -0.4049, 0.0396, -0.0, -0.4028, 0.0374, -0.0, -0.4028, 0.0439, -0.0, -0.4006, 0.0418, -0.0, -0.4006, 0.0439, -0.0, -0.3984, 0.0418, -0.0, -0.3984,
0.0396, -0.0, -0.4006, 0.0374, -0.0, -0.4006, 0.0396, -0.0, -0.3984, 0.0374, -0.0, -0.3984, 0.0544, -0.0, -0.3961, 0.0538, -0.0, -0.3951, 0.0545, 0.0017, -0.3963, 0.0545, 0.0039, -0.3963, 0.0538, 0.0017, -0.395, 0.0538, 0.0039, -0.395, 0.0526, -0.0, -0.3963, 0.0505, -0.0, -0.3963, 0.0531, -0.0, -0.3939, 0.0505, -0.0, -0.3941, 0.0483, -0.0, -0.3963, 0.0461, -0.0, -0.3963, 0.0483, -0.0, -0.3941, 0.0461, -0.0, -0.3941, 0.0532, 0.0017, -0.394, 0.0532, 0.0039, -0.394, 0.0521, -0.0, -0.3923, 0.0515, -0.0, -0.3912, 0.0504, -0.0, -0.3895, 0.0483, -0.0, -0.3919, 0.0461, -0.0, -0.3919, 0.0483, -0.0, -0.3897, 0.0461, -0.0, -0.3897, 0.0521, 0.0017, -0.3921, 0.0514, 0.0017, -0.3911, 0.0521, 0.0039, -0.3921,
0.0514, 0.0039, -0.3911, 0.0506, 0.0017, -0.3897, 0.0506, 0.0039, -0.3897, 0.0439, -0.0, -0.3963, 0.0418, -0.0, -0.3963, 0.0439, -0.0, -0.3941, 0.0418, -0.0, -0.3941, 0.0396, -0.0, -0.3963, 0.0374, -0.0, -0.3963, 0.0396, -0.0, -0.3941, 0.0374, -0.0, -0.3941, 0.0439, -0.0, -0.3919, 0.0418, -0.0, -0.3919, 0.0439, -0.0, -0.3897, 0.0418, -0.0, -0.3897, 0.0396, -0.0, -0.3919, 0.0374, -0.0, -0.3919, 0.0396, -0.0, -0.3897, 0.0374, -0.0, -0.3897, 0.0497, -0.0, -0.3883, 0.049, -0.0, -0.3872, 0.0461, -0.0, -0.3876, 0.0479, -0.0, -0.3857, 0.0471, -0.0, -0.3846, 0.0496, 0.0017, -0.3883, 0.0496, 0.0039, -0.3883, 0.0489, 0.0017, -0.3872, 0.0489, 0.0039, -0.3872, 0.0478, 0.0017, -0.3856, 0.0471, 0.0017, -0.3845,
0.0478, 0.0039, -0.3856, 0.0471, 0.0039, -0.3845, 0.0459, -0.0, -0.383, 0.0451, -0.0, -0.382, 0.0461, 0.0017, -0.3832, 0.0461, 0.0039, -0.3832, 0.0451, 0.0017, -0.382, 0.0451, 0.0039, -0.382, 0.0439, -0.0, -0.3876, 0.0418, -0.0, -0.3876, 0.0439, -0.0, -0.3854, 0.0418, -0.0, -0.3854, 0.0396, -0.0, -0.3876, 0.0374, -0.0, -0.3876, 0.0396, -0.0, -0.3854, 0.0374, -0.0, -0.3854, 0.0439, -0.0, -0.3832, 0.0418, -0.0, -0.3832, 0.0443, -0.0, -0.3809, 0.0418, -0.0, -0.3811, 0.0396, -0.0, -0.3832, 0.0374, -0.0, -0.3832, 0.0396, -0.0, -0.3811, 0.0374, -0.0, -0.3811, 0.0443, 0.0017, -0.3809, 0.0443, 0.0039, -0.3809, 0.0308, -0.0, -0.4134, 0.0288, -0.0, -0.4136, 0.0301, -0.0, -0.4122, 0.0295, -0.0, -0.4111,
0.0309, 0.0017, -0.4136, 0.0309, 0.0039, -0.4136, 0.0301, 0.0017, -0.4121, 0.0295, 0.0017, -0.411, 0.0301, 0.0039, -0.4121, 0.0295, 0.0039, -0.411, 0.036, -0.0, -0.4065, 0.0283, -0.0, -0.4091, 0.0277, -0.0, -0.4081, 0.036, 0.0017, -0.4066, 0.036, 0.0039, -0.4066, 0.0284, 0.0017, -0.4093, 0.0284, 0.0039, -0.4093, 0.0277, 0.0017, -0.4081, 0.0277, 0.0039, -0.4081, 0.0266, -0.0, -0.4136, 0.0244, -0.0, -0.4136, 0.0266, -0.0, -0.4115, 0.0244, -0.0, -0.4115, 0.0222, -0.0, -0.4136, 0.0201, -0.0, -0.4136, 0.0222, -0.0, -0.4115, 0.0201, -0.0, -0.4115, 0.0266, -0.0, -0.4093, 0.0244, -0.0, -0.4093, 0.0271, -0.0, -0.407, 0.0244, -0.0, -0.4071, 0.0222, -0.0, -0.4093, 0.0201, -0.0, -0.4093, 0.0222, -0.0, -0.4071,
0.0201, -0.0, -0.4071, 0.0271, 0.0017, -0.407, 0.0271, 0.0039, -0.407, 0.0352, -0.0, -0.4052, 0.0342, -0.0, -0.4035, 0.0335, -0.0, -0.4025, 0.0351, 0.0017, -0.4049, 0.0351, 0.0039, -0.4049, 0.0343, 0.0017, -0.4037, 0.0336, 0.0017, -0.4026, 0.0343, 0.0039, -0.4037, 0.0336, 0.0039, -0.4026, 0.0353, -0.0, -0.4006, 0.0325, -0.0, -0.4008, 0.0353, -0.0, -0.3984, 0.0331, -0.0, -0.3984, 0.0318, -0.0, -0.3997, 0.0312, -0.0, -0.3987, 0.0325, 0.0017, -0.4009, 0.0325, 0.0039, -0.4009, 0.0319, 0.0017, -0.3998, 0.0319, 0.0039, -0.3998, 0.031, 0.0017, -0.3984, 0.031, 0.0039, -0.3984, 0.0261, -0.0, -0.4054, 0.0254, -0.0, -0.4043, 0.0242, -0.0, -0.4025, 0.0222, -0.0, -0.4049, 0.0201, -0.0, -0.4049, 0.0222, -0.0, -0.4028,
0.0201, -0.0, -0.4028, 0.026, 0.0017, -0.4053, 0.0253, 0.0017, -0.4042, 0.026, 0.0039, -0.4053, 0.0253, 0.0039, -0.4042, 0.0244, 0.0017, -0.4028, 0.0244, 0.0039, -0.4028, 0.0235, -0.0, -0.4015, 0.0228, -0.0, -0.4004, 0.0201, -0.0, -0.4006, 0.0218, -0.0, -0.3988, 0.0211, -0.0, -0.3978, 0.0235, 0.0017, -0.4014, 0.0235, 0.0039, -0.4014, 0.0228, 0.0017, -0.4003, 0.0228, 0.0039, -0.4003, 0.0217, 0.0017, -0.3987, 0.021, 0.0017, -0.3976, 0.0217, 0.0039, -0.3987, 0.021, 0.0039, -0.3976, 0.0179, -0.0, -0.4136, 0.0157, -0.0, -0.4136, 0.0179, -0.0, -0.4115, 0.0157, -0.0, -0.4115, 0.0136, -0.0, -0.4136, 0.0103, -0.0, -0.4143, 0.0136, -0.0, -0.4115, 0.0114, -0.0, -0.4115, 0.0104, 0.0017, -0.4145, 0.0104, 0.0039, -0.4145,
0.0179, -0.0, -0.4093, 0.0157, -0.0, -0.4093, 0.0179, -0.0, -0.4071, 0.0157, -0.0, -0.4071, 0.0136, -0.0, -0.4093, 0.0114, -0.0, -0.4093, 0.0136, -0.0, -0.4071, 0.0114, -0.0, -0.4071, 0.0097, -0.0, -0.4133, 0.0089, -0.0, -0.4118, 0.002, -0.0, -0.4142, 0.0098, 0.0017, -0.4134, 0.0098, 0.0039, -0.4134, 0.0087, 0.0017, -0.4115, 0.0087, 0.0039, -0.4115, 0.0019, 0.0017, -0.4141, 0.0019, 0.0039, -0.4141, 0.008, -0.0, -0.41, 0.0075, -0.0, -0.4091, 0.0092, -0.0, -0.4071, 0.0065, -0.0, -0.4073, 0.0058, -0.0, -0.4062, 0.0081, 0.0017, -0.4103, 0.0076, 0.0017, -0.4093, 0.0081, 0.0039, -0.4103, 0.0076, 0.0039, -0.4093, 0.0065, 0.0017, -0.4073, 0.0065, 0.0039, -0.4073, 0.0058, 0.0017, -0.4062, 0.0058, 0.0039, -0.4062,
0.0179, -0.0, -0.4049, 0.0157, -0.0, -0.4049, 0.0179, -0.0, -0.4028, 0.0157, -0.0, -0.4028, 0.0136, -0.0, -0.4049, 0.0114, -0.0, -0.4049, 0.0136, -0.0, -0.4028, 0.0114, -0.0, -0.4028, 0.0179, -0.0, -0.4006, 0.0157, -0.0, -0.4006, 0.0179, -0.0, -0.3984, 0.0157, -0.0, -0.3984, 0.0136, -0.0, -0.4006, 0.0114, -0.0, -0.4006, 0.0136, -0.0, -0.3984, 0.0114, -0.0, -0.3984, 0.0092, -0.0, -0.4049, 0.0071, -0.0, -0.4049, 0.0092, -0.0, -0.4028, 0.0071, -0.0, -0.4028, 0.0052, -0.0, -0.4051, 0.0041, -0.0, -0.4032, 0.0035, -0.0, -0.4021, 0.0051, 0.0017, -0.4049, 0.0051, 0.0039, -0.4049, 0.0041, 0.0017, -0.4032, 0.0035, 0.0017, -0.4022, 0.0041, 0.0039, -0.4032, 0.0035, 0.0039, -0.4022, 0.0092, -0.0, -0.4006,
0.0071, -0.0, -0.4006, 0.0092, -0.0, -0.3984, 0.0071, -0.0, -0.3984, 0.0049, -0.0, -0.4006, 0.0028, -0.0, -0.4009, 0.0049, -0.0, -0.3984, 0.0017, -0.0, -0.399, 0.0026, 0.0017, -0.4006, 0.0026, 0.0039, -0.4006, 0.0018, 0.0017, -0.3992, 0.0018, 0.0039, -0.3992, 0.0353, -0.0, -0.3963, 0.0331, -0.0, -0.3963, 0.0353, -0.0, -0.3941, 0.0331, -0.0, -0.3941, 0.03, -0.0, -0.3968, 0.0293, -0.0, -0.3958, 0.0309, -0.0, -0.3941, 0.0284, -0.0, -0.3942, 0.0301, 0.0017, -0.397, 0.0294, 0.0017, -0.3959, 0.0301, 0.0039, -0.397, 0.0294, 0.0039, -0.3959, 0.0283, 0.0017, -0.3941, 0.0283, 0.0039, -0.3941, 0.0353, -0.0, -0.3919, 0.0331, -0.0, -0.3919, 0.0353, -0.0, -0.3897, 0.0331, -0.0, -0.3897, 0.0309, -0.0, -0.3919,
0.0288, -0.0, -0.3919, 0.0309, -0.0, -0.3897, 0.0288, -0.0, -0.3897, 0.0276, -0.0, -0.393, 0.0199, -0.0, -0.396, 0.0192, -0.0, -0.395, 0.0277, 0.0017, -0.393, 0.0277, 0.0039, -0.393, 0.0201, 0.0017, -0.3963, 0.0201, 0.0039, -0.3963, 0.0192, 0.0017, -0.3949, 0.0192, 0.0039, -0.3949, 0.027, -0.0, -0.3921, 0.0258, -0.0, -0.3902, 0.0251, -0.0, -0.3891, 0.0269, 0.0017, -0.3919, 0.0269, 0.0039, -0.3919, 0.0259, 0.0017, -0.3903, 0.0251, 0.0017, -0.3892, 0.0259, 0.0039, -0.3903, 0.0251, 0.0039, -0.3892, 0.0353, -0.0, -0.3876, 0.0331, -0.0, -0.3876, 0.0353, -0.0, -0.3854, 0.0331, -0.0, -0.3854, 0.0309, -0.0, -0.3876, 0.0288, -0.0, -0.3876, 0.0309, -0.0, -0.3854, 0.0288, -0.0, -0.3854, 0.0353, -0.0, -0.3832,
0.0331, -0.0, -0.3832, 0.0353, -0.0, -0.3811, 0.0331, -0.0, -0.3811, 0.0309, -0.0, -0.3832, 0.0288, -0.0, -0.3832, 0.0309, -0.0, -0.3811, 0.0288, -0.0, -0.3811, 0.0266, -0.0, -0.3876, 0.0243, -0.0, -0.3879, 0.0266, -0.0, -0.3854, 0.0232, -0.0, -0.3863, 0.0225, -0.0, -0.3853, 0.0241, 0.0017, -0.3876, 0.0241, 0.0039, -0.3876, 0.0233, 0.0017, -0.3865, 0.0233, 0.0039, -0.3865, 0.0226, 0.0017, -0.3854, 0.0226, 0.0039, -0.3854, 0.0266, -0.0, -0.3832, 0.0244, -0.0, -0.3832, 0.0266, -0.0, -0.3811, 0.0244, -0.0, -0.3811, 0.0215, -0.0, -0.3837, 0.0207, -0.0, -0.3827, 0.0222, -0.0, -0.3811, 0.0197, -0.0, -0.3812, 0.0215, 0.0017, -0.3838, 0.0207, 0.0017, -0.3827, 0.0215, 0.0039, -0.3838, 0.0207, 0.0039, -0.3827,
0.0197, 0.0017, -0.3812, 0.0197, 0.0039, -0.3812, 0.0179, -0.0, -0.3963, 0.0157, -0.0, -0.3963, 0.0184, -0.0, -0.3938, 0.0157, -0.0, -0.3941, 0.0136, -0.0, -0.3963, 0.0114, -0.0, -0.3963, 0.0136, -0.0, -0.3941, 0.0114, -0.0, -0.3941, 0.0184, 0.0017, -0.3938, 0.0184, 0.0039, -0.3938, 0.0174, -0.0, -0.3923, 0.0167, -0.0, -0.3913, 0.0156, -0.0, -0.3895, 0.0136, -0.0, -0.3919, 0.0114, -0.0, -0.3919, 0.0136, -0.0, -0.3897, 0.0114, -0.0, -0.3897, 0.0174, 0.0017, -0.3922, 0.0166, 0.0017, -0.3911, 0.0174, 0.0039, -0.3922, 0.0166, 0.0039, -0.3911, 0.0157, 0.0017, -0.3897, 0.0157, 0.0039, -0.3897, 0.0092, -0.0, -0.3963, 0.0071, -0.0, -0.3963, 0.0092, -0.0, -0.3941, 0.0071, -0.0, -0.3941, 0.0049, -0.0, -0.3963,
0.0027, -0.0, -0.3963, 0.0049, -0.0, -0.3941, 0.0027, -0.0, -0.3941, 0.0092, -0.0, -0.3919, 0.0071, -0.0, -0.3919, 0.0092, -0.0, -0.3897, 0.0071, -0.0, -0.3897, 0.0049, -0.0, -0.3919, 0.0027, -0.0, -0.3919, 0.0049, -0.0, -0.3897, 0.0027, -0.0, -0.3897, 0.0148, -0.0, -0.3884, 0.0141, -0.0, -0.3873, 0.0114, -0.0, -0.3876, 0.013, -0.0, -0.3859, 0.0123, -0.0, -0.3848, 0.0148, 0.0017, -0.3884, 0.0148, 0.0039, -0.3884, 0.0141, 0.0017, -0.3873, 0.0141, 0.0039, -0.3873, 0.013, 0.0017, -0.3858, 0.0122, 0.0017, -0.3847, 0.013, 0.0039, -0.3858, 0.0122, 0.0039, -0.3847, 0.0189, -0.0, -0.3801, 0.011, -0.0, -0.383, 0.0104, -0.0, -0.3821, 0.0189, 0.0017, -0.3801, 0.0189, 0.0039, -0.3801, 0.0112, 0.0017, -0.3832,
0.0112, 0.0039, -0.3832, 0.0104, 0.0017, -0.3821, 0.0104, 0.0039, -0.3821, 0.0092, -0.0, -0.3876, 0.0071, -0.0, -0.3876, 0.0092, -0.0, -0.3854, 0.0071, -0.0, -0.3854, 0.0049, -0.0, -0.3876, 0.0027, -0.0, -0.3876, 0.0049, -0.0, -0.3854, 0.0027, -0.0, -0.3854, 0.0092, -0.0, -0.3832, 0.0071, -0.0, -0.3832, 0.0096, -0.0, -0.381, 0.0071, -0.0, -0.3811, 0.0049, -0.0, -0.3832, 0.0027, -0.0, -0.3832, 0.0049, -0.0, -0.3811, 0.0027, -0.0, -0.3811, 0.0096, 0.0017, -0.381, 0.0096, 0.0039, -0.381, 0.0432, -0.0, -0.3796, 0.0423, -0.0, -0.3785, 0.0413, -0.0, -0.3772, 0.0396, -0.0, -0.3789, 0.0374, -0.0, -0.3789, 0.0404, -0.0, -0.3761, 0.0374, -0.0, -0.3767, 0.0432, 0.0017, -0.3796, 0.0423, 0.0017, -0.3785,
0.0432, 0.0039, -0.3796, 0.0423, 0.0039, -0.3785, 0.0412, 0.0017, -0.3771, 0.0412, 0.0039, -0.3771, 0.0404, 0.0017, -0.376, 0.0404, 0.0039, -0.376, 0.0394, -0.0, -0.3747, 0.0385, -0.0, -0.3737, 0.0373, -0.0, -0.3721, 0.0393, 0.0017, -0.3746, 0.0385, 0.0017, -0.3736, 0.0393, 0.0039, -0.3746, 0.0385, 0.0039, -0.3736, 0.0376, 0.0017, -0.3724, 0.0376, 0.0039, -0.3724, 0.0365, -0.0, -0.3711, 0.0365, 0.0017, -0.3711, 0.0365, 0.0039, -0.3711, 0.0353, -0.0, -0.3789, 0.0331, -0.0, -0.3789, 0.0353, -0.0, -0.3767, 0.0331, -0.0, -0.3767, 0.0309, -0.0, -0.3789, 0.0288, -0.0, -0.3789, 0.0309, -0.0, -0.3767, 0.0288, -0.0, -0.3767, 0.0353, -0.0, -0.3746, 0.0331, -0.0, -0.3746, 0.0353, -0.0, -0.3724, 0.0331, -0.0, -0.3724,
0.0309, -0.0, -0.3746, 0.0288, -0.0, -0.3746, 0.0309, -0.0, -0.3724, 0.0288, -0.0, -0.3724, 0.0266, -0.0, -0.3789, 0.0244, -0.0, -0.3789, 0.0266, -0.0, -0.3767, 0.0244, -0.0, -0.3767, 0.0222, -0.0, -0.3789, 0.0201, -0.0, -0.3789, 0.0222, -0.0, -0.3767, 0.0201, -0.0, -0.3767, 0.0266, -0.0, -0.3746, 0.0244, -0.0, -0.3746, 0.0266, -0.0, -0.3724, 0.0244, -0.0, -0.3724, 0.0222, -0.0, -0.3746, 0.0201, -0.0, -0.3746, 0.0222, -0.0, -0.3724, 0.0201, -0.0, -0.3724, 0.0357, -0.0, -0.37, 0.0331, -0.0, -0.3702, 0.0346, -0.0, -0.3686, 0.0338, -0.0, -0.3676, 0.0309, -0.0, -0.3702, 0.0288, -0.0, -0.3702, 0.0309, -0.0, -0.368, 0.0288, -0.0, -0.368, 0.0357, 0.0017, -0.37, 0.0357, 0.0039, -0.37,
0.0346, 0.0017, -0.3686, 0.0337, 0.0017, -0.3675, 0.0346, 0.0039, -0.3686, 0.0337, 0.0039, -0.3675, 0.0327, -0.0, -0.3662, 0.0318, -0.0, -0.3651, 0.0288, -0.0, -0.3659, 0.0305, -0.0, -0.3635, 0.0288, -0.0, -0.3637, 0.0327, 0.0017, -0.3662, 0.0327, 0.0039, -0.3662, 0.0318, 0.0017, -0.3651, 0.0318, 0.0039, -0.3651, 0.0307, 0.0017, -0.3637, 0.0307, 0.0039, -0.3637, 0.0266, -0.0, -0.3702, 0.0244, -0.0, -0.3702, 0.0266, -0.0, -0.368, 0.0244, -0.0, -0.368, 0.0222, -0.0, -0.3702, 0.0201, -0.0, -0.3702, 0.0222, -0.0, -0.368, 0.0201, -0.0, -0.368, 0.0266, -0.0, -0.3659, 0.0244, -0.0, -0.3659, 0.0266, -0.0, -0.3637, 0.0244, -0.0, -0.3637, 0.0222, -0.0, -0.3659, 0.0201, -0.0, -0.3659, 0.0222, -0.0, -0.3637,
0.0201, -0.0, -0.3637, 0.0182, -0.0, -0.3791, 0.017, -0.0, -0.3774, 0.0162, -0.0, -0.3763, 0.018, 0.0017, -0.3789, 0.018, 0.0039, -0.3789, 0.017, 0.0017, -0.3775, 0.0163, 0.0017, -0.3764, 0.017, 0.0039, -0.3775, 0.0163, 0.0039, -0.3764, 0.0179, -0.0, -0.3746, 0.0152, -0.0, -0.3749, 0.0179, -0.0, -0.3724, 0.0157, -0.0, -0.3724, 0.0144, -0.0, -0.3738, 0.0136, -0.0, -0.3727, 0.0152, 0.0017, -0.3749, 0.0152, 0.0039, -0.3749, 0.0144, 0.0017, -0.3738, 0.0144, 0.0039, -0.3738, 0.0134, 0.0017, -0.3724, 0.0134, 0.0039, -0.3724, 0.0086, -0.0, -0.3795, 0.0078, -0.0, -0.3784, 0.0068, -0.0, -0.3769, 0.0049, -0.0, -0.3789, 0.0027, -0.0, -0.3789, 0.006, -0.0, -0.3759, 0.0027, -0.0, -0.3767, 0.0085, 0.0017, -0.3794,
0.0078, 0.0017, -0.3783, 0.0085, 0.0039, -0.3794, 0.0078, 0.0039, -0.3783, 0.0067, 0.0017, -0.3768, 0.0067, 0.0039, -0.3768, 0.0059, 0.0017, -0.3757, 0.0059, 0.0039, -0.3757, 0.0049, -0.0, -0.3742, 0.0027, -0.0, -0.3746, 0.0041, -0.0, -0.3731, 0.0033, -0.0, -0.372, 0.0051, 0.0017, -0.3746, 0.0051, 0.0039, -0.3746, 0.0041, 0.0017, -0.3731, 0.0033, 0.0017, -0.372, 0.0041, 0.0039, -0.3731, 0.0033, 0.0039, -0.372, 0.0179, -0.0, -0.3702, 0.0157, -0.0, -0.3702, 0.0179, -0.0, -0.368, 0.0157, -0.0, -0.368, 0.0124, -0.0, -0.3711, 0.0117, -0.0, -0.37, 0.0136, -0.0, -0.368, 0.0106, -0.0, -0.3685, 0.0125, 0.0017, -0.3712, 0.0118, 0.0017, -0.3701, 0.0125, 0.0039, -0.3712, 0.0118, 0.0039, -0.3701, 0.0107, 0.0017, -0.3686,
0.0107, 0.0039, -0.3686, 0.0179, -0.0, -0.3659, 0.0157, -0.0, -0.3659, 0.0179, -0.0, -0.3637, 0.0157, -0.0, -0.3637, 0.0136, -0.0, -0.3659, 0.0114, -0.0, -0.3659, 0.0136, -0.0, -0.3637, 0.0114, -0.0, -0.3637, 0.0099, -0.0, -0.3675, 0.0023, -0.0, -0.3706, 0.0099, 0.0017, -0.3675, 0.0099, 0.0039, -0.3675, 0.0022, 0.0017, -0.3704, 0.0022, 0.0039, -0.3704, 0.0088, -0.0, -0.366, 0.008, -0.0, -0.3649, 0.0092, -0.0, -0.3637, 0.0074, -0.0, -0.3639, 0.0088, 0.0017, -0.366, 0.0081, 0.0017, -0.3649, 0.0088, 0.0039, -0.366, 0.0081, 0.0039, -0.3649, 0.0072, 0.0017, -0.3637, 0.0072, 0.0039, -0.3637, 0.0298, -0.0, -0.3626, 0.029, -0.0, -0.3615, 0.0279, -0.0, -0.3602, 0.0298, 0.0017, -0.3626, 0.029, 0.0017, -0.3615,
0.0298, 0.0039, -0.3626, 0.029, 0.0039, -0.3615, 0.0279, 0.0017, -0.3601, 0.0279, 0.0039, -0.3601, 0.0266, -0.0, -0.3615, 0.0244, -0.0, -0.3615, 0.0271, -0.0, -0.3591, 0.0244, -0.0, -0.3594, 0.0222, -0.0, -0.3615, 0.0201, -0.0, -0.3615, 0.0222, -0.0, -0.3594, 0.0201, -0.0, -0.3594, 0.0271, 0.0017, -0.3591, 0.0271, 0.0039, -0.3591, 0.026, -0.0, -0.3577, 0.0252, -0.0, -0.3566, 0.0241, -0.0, -0.3553, 0.0222, -0.0, -0.3572, 0.0201, -0.0, -0.3572, 0.0231, -0.0, -0.3543, 0.0201, -0.0, -0.355, 0.026, 0.0017, -0.3577, 0.0251, 0.0017, -0.3566, 0.026, 0.0039, -0.3577, 0.0251, 0.0039, -0.3566, 0.024, 0.0017, -0.3553, 0.024, 0.0039, -0.3553, 0.0231, 0.0017, -0.3542, 0.0231, 0.0039, -0.3542, 0.0221, -0.0, -0.353,
0.0212, -0.0, -0.3519, 0.0199, -0.0, -0.3504, 0.022, 0.0017, -0.3529, 0.0211, 0.0017, -0.3519, 0.022, 0.0039, -0.3529, 0.0211, 0.0039, -0.3519, 0.0201, 0.0017, -0.3507, 0.0201, 0.0039, -0.3507, 0.0191, -0.0, -0.3495, 0.0191, 0.0017, -0.3494, 0.0191, 0.0039, -0.3494, 0.0179, -0.0, -0.3615, 0.0157, -0.0, -0.3615, 0.0179, -0.0, -0.3594, 0.0157, -0.0, -0.3594, 0.0136, -0.0, -0.3615, 0.0114, -0.0, -0.3615, 0.0136, -0.0, -0.3594, 0.0114, -0.0, -0.3594, 0.0179, -0.0, -0.3572, 0.0157, -0.0, -0.3572, 0.0179, -0.0, -0.355, 0.0157, -0.0, -0.355, 0.0136, -0.0, -0.3572, 0.0114, -0.0, -0.3572, 0.0136, -0.0, -0.355, 0.0114, -0.0, -0.355, 0.0092, -0.0, -0.3615, 0.0061, -0.0, -0.3622, 0.0092, -0.0, -0.3594,
0.0071, -0.0, -0.3594, 0.0054, -0.0, -0.3611, 0.0044, -0.0, -0.3596, 0.0036, -0.0, -0.3585, 0.0062, 0.0017, -0.3623, 0.0062, 0.0039, -0.3623, 0.0055, 0.0017, -0.3612, 0.0055, 0.0039, -0.3612, 0.0044, 0.0017, -0.3596, 0.0036, 0.0017, -0.3585, 0.0044, 0.0039, -0.3596, 0.0036, 0.0039, -0.3585, 0.0092, -0.0, -0.3572, 0.0071, -0.0, -0.3572, 0.0092, -0.0, -0.355, 0.0071, -0.0, -0.355, 0.0049, -0.0, -0.3572, 0.0029, -0.0, -0.3574, 0.0049, -0.0, -0.355, 0.0017, -0.0, -0.3557, 0.0027, 0.0017, -0.3572, 0.0027, 0.0039, -0.3572, 0.0018, 0.0017, -0.3559, 0.0018, 0.0039, -0.3559, 0.0179, -0.0, -0.3529, 0.0157, -0.0, -0.3529, 0.0179, -0.0, -0.3507, 0.0157, -0.0, -0.3507, 0.0136, -0.0, -0.3529, 0.0114, -0.0, -0.3529,
0.0136, -0.0, -0.3507, 0.0114, -0.0, -0.3507, 0.0182, -0.0, -0.3483, 0.0157, -0.0, -0.3485, 0.0172, -0.0, -0.347, 0.0163, -0.0, -0.3459, 0.0136, -0.0, -0.3485, 0.0114, -0.0, -0.3485, 0.0136, -0.0, -0.3463, 0.0114, -0.0, -0.3463, 0.0182, 0.0017, -0.3484, 0.0182, 0.0039, -0.3484, 0.0172, 0.0017, -0.347, 0.0163, 0.0017, -0.3459, 0.0172, 0.0039, -0.347, 0.0163, 0.0039, -0.3459, 0.0092, -0.0, -0.3529, 0.0071, -0.0, -0.3529, 0.0092, -0.0, -0.3507, 0.0071, -0.0, -0.3507, 0.0049, -0.0, -0.3529, 0.0027, -0.0, -0.3529, 0.0049, -0.0, -0.3507, 0.0027, -0.0, -0.3507, 0.0092, -0.0, -0.3485, 0.0071, -0.0, -0.3485, 0.0092, -0.0, -0.3463, 0.0071, -0.0, -0.3463, 0.0049, -0.0, -0.3485, 0.0027, -0.0, -0.3485,
0.0049, -0.0, -0.3463, 0.0027, -0.0, -0.3463, 0.1019, -0.0, -0.3446, 0.1015, -0.0, -0.3437, 0.0982, -0.0, -0.3442, 0.1007, -0.0, -0.3417, 0.0982, -0.0, -0.342, 0.1018, 0.0017, -0.3444, 0.1018, 0.0039, -0.3444, 0.1014, 0.0017, -0.3433, 0.1014, 0.0039, -0.3433, 0.1008, 0.0017, -0.342, 0.1008, 0.0039, -0.342, 0.0999, -0.0, -0.3397, 0.0982, -0.0, -0.3398, 0.0994, -0.0, -0.3385, 0.099, -0.0, -0.3375, 0.1, 0.0017, -0.3398, 0.1, 0.0039, -0.3398, 0.0994, 0.0017, -0.3385, 0.099, 0.0017, -0.3374, 0.0994, 0.0039, -0.3385, 0.099, 0.0039, -0.3374, 0.096, -0.0, -0.3442, 0.0939, -0.0, -0.3442, 0.096, -0.0, -0.342, 0.0939, -0.0, -0.342, 0.0917, -0.0, -0.3442, 0.0895, -0.0, -0.3442, 0.0917, -0.0, -0.342,
0.0895, -0.0, -0.342, 0.096, -0.0, -0.3398, 0.0939, -0.0, -0.3398, 0.096, -0.0, -0.3377, 0.0939, -0.0, -0.3377, 0.0917, -0.0, -0.3398, 0.0895, -0.0, -0.3398, 0.0917, -0.0, -0.3377, 0.0895, -0.0, -0.3377, 0.0981, -0.0, -0.3353, 0.0974, -0.0, -0.334, 0.0981, 0.0017, -0.3355, 0.0981, 0.0039, -0.3355, 0.0974, 0.0017, -0.3339, 0.0974, 0.0039, -0.3339, 0.096, -0.0, -0.3355, 0.0939, -0.0, -0.3355, 0.0969, -0.0, -0.3329, 0.0939, -0.0, -0.3333, 0.0917, -0.0, -0.3355, 0.0895, -0.0, -0.3355, 0.0917, -0.0, -0.3333, 0.0895, -0.0, -0.3333, 0.0968, 0.0017, -0.3328, 0.0968, 0.0039, -0.3328, 0.0958, -0.0, -0.3309, 0.0939, -0.0, -0.3312, 0.0952, -0.0, -0.3298, 0.0946, -0.0, -0.3287, 0.0917, -0.0, -0.3312,
0.0895, -0.0, -0.3312, 0.0917, -0.0, -0.329, 0.0895, -0.0, -0.329, 0.096, 0.0017, -0.3311, 0.096, 0.0039, -0.3311, 0.0952, 0.0017, -0.3297, 0.0945, 0.0017, -0.3286, 0.0952, 0.0039, -0.3297, 0.0945, 0.0039, -0.3286, 0.0879, -0.0, -0.3443, 0.087, -0.0, -0.3423, 0.0879, 0.0017, -0.3442, 0.0879, 0.0039, -0.3442, 0.0869, 0.0017, -0.342, 0.0869, 0.0039, -0.342, 0.0862, -0.0, -0.3404, 0.0858, -0.0, -0.3394, 0.0874, -0.0, -0.3377, 0.0851, -0.0, -0.3379, 0.0864, 0.0017, -0.3407, 0.0859, 0.0017, -0.3396, 0.0864, 0.0039, -0.3407, 0.0859, 0.0039, -0.3396, 0.085, 0.0017, -0.3377, 0.085, 0.0039, -0.3377, 0.0874, -0.0, -0.3355, 0.0842, -0.0, -0.336, 0.0874, -0.0, -0.3333, 0.0852, -0.0, -0.3333, 0.0837, -0.0, -0.3349,
0.083, -0.0, -0.3336, 0.0843, 0.0017, -0.3362, 0.0843, 0.0039, -0.3362, 0.0838, 0.0017, -0.3351, 0.0838, 0.0039, -0.3351, 0.0829, 0.0017, -0.3333, 0.0829, 0.0039, -0.3333, 0.0874, -0.0, -0.3312, 0.0852, -0.0, -0.3312, 0.0874, -0.0, -0.329, 0.0852, -0.0, -0.329, 0.082, -0.0, -0.3317, 0.0814, -0.0, -0.3307, 0.083, -0.0, -0.329, 0.0804, -0.0, -0.3291, 0.0821, 0.0017, -0.3319, 0.0815, 0.0017, -0.3308, 0.0821, 0.0039, -0.3319, 0.0815, 0.0039, -0.3308, 0.0804, 0.0017, -0.329, 0.0804, 0.0039, -0.329, 0.0797, -0.0, -0.328, 0.0797, 0.0017, -0.328, 0.0797, 0.0039, -0.328, 0.0934, -0.0, -0.3267, 0.0928, -0.0, -0.3257, 0.0917, -0.0, -0.3268, 0.0895, -0.0, -0.3268, 0.0921, -0.0, -0.3246, 0.0895, -0.0, -0.3246,
0.0935, 0.0017, -0.3268, 0.0935, 0.0039, -0.3268, 0.0928, 0.0017, -0.3257, 0.0928, 0.0039, -0.3257, 0.0921, 0.0017, -0.3246, 0.0921, 0.0039, -0.3246, 0.0911, -0.0, -0.323, 0.0903, -0.0, -0.3219, 0.0893, -0.0, -0.3205, 0.091, 0.0017, -0.3229, 0.0903, 0.0017, -0.3219, 0.091, 0.0039, -0.3229, 0.0903, 0.0039, -0.3219, 0.0892, 0.0017, -0.3203, 0.0892, 0.0039, -0.3203, 0.0874, -0.0, -0.3268, 0.0852, -0.0, -0.3268, 0.0874, -0.0, -0.3246, 0.0852, -0.0, -0.3246, 0.083, -0.0, -0.3268, 0.0808, -0.0, -0.3268, 0.083, -0.0, -0.3246, 0.0808, -0.0, -0.3246, 0.0874, -0.0, -0.3225, 0.0852, -0.0, -0.3225, 0.0885, -0.0, -0.3194, 0.0852, -0.0, -0.3203, 0.083, -0.0, -0.3225, 0.0808, -0.0, -0.3225, 0.083, -0.0, -0.3203,
0.0808, -0.0, -0.3203, 0.0884, 0.0017, -0.3193, 0.0884, 0.0039, -0.3193, 0.0791, -0.0, -0.327, 0.0779, -0.0, -0.3252, 0.0771, -0.0, -0.3241, 0.079, 0.0017, -0.3268, 0.079, 0.0039, -0.3268, 0.0779, 0.0017, -0.3253, 0.0772, 0.0017, -0.3242, 0.0779, 0.0039, -0.3253, 0.0772, 0.0039, -0.3242, 0.0787, -0.0, -0.3225, 0.0761, -0.0, -0.3226, 0.0787, -0.0, -0.3203, 0.0765, -0.0, -0.3203, 0.0753, -0.0, -0.3215, 0.0745, -0.0, -0.3206, 0.0761, 0.0017, -0.3226, 0.0761, 0.0039, -0.3226, 0.0753, 0.0017, -0.3216, 0.0753, 0.0039, -0.3216, 0.0743, 0.0017, -0.3203, 0.0743, 0.0039, -0.3203, 0.0873, -0.0, -0.3178, 0.0852, -0.0, -0.3181, 0.0865, -0.0, -0.3168, 0.0856, -0.0, -0.3157, 0.083, -0.0, -0.3181, 0.0808, -0.0, -0.3181,
0.083, -0.0, -0.316, 0.0808, -0.0, -0.316, 0.0875, 0.0017, -0.3181, 0.0875, 0.0039, -0.3181, 0.0865, 0.0017, -0.3168, 0.0856, 0.0017, -0.3157, 0.0865, 0.0039, -0.3168, 0.0856, 0.0039, -0.3157, 0.0846, -0.0, -0.3144, 0.0837, -0.0, -0.3133, 0.0808, -0.0, -0.3138, 0.0826, -0.0, -0.312, 0.0816, -0.0, -0.3109, 0.0845, 0.0017, -0.3143, 0.0845, 0.0039, -0.3143, 0.0836, 0.0017, -0.3133, 0.0836, 0.0039, -0.3133, 0.0826, 0.0017, -0.312, 0.0816, 0.0017, -0.3109, 0.0826, 0.0039, -0.312, 0.0816, 0.0039, -0.3109, 0.0787, -0.0, -0.3181, 0.0765, -0.0, -0.3181, 0.0787, -0.0, -0.316, 0.0765, -0.0, -0.316, 0.0733, -0.0, -0.319, 0.0724, -0.0, -0.318, 0.0743, -0.0, -0.316, 0.0713, -0.0, -0.3167, 0.0733, 0.0017, -0.3191,
0.0724, 0.0017, -0.318, 0.0733, 0.0039, -0.3191, 0.0724, 0.0039, -0.318, 0.0713, 0.0017, -0.3167, 0.0713, 0.0039, -0.3167, 0.0787, -0.0, -0.3138, 0.0765, -0.0, -0.3138, 0.0787, -0.0, -0.3116, 0.0765, -0.0, -0.3116, 0.0743, -0.0, -0.3138, 0.0722, -0.0, -0.3138, 0.0743, -0.0, -0.3116, 0.0722, -0.0, -0.3116, 0.0806, -0.0, -0.3097, 0.0805, 0.0017, -0.3097, 0.0805, 0.0039, -0.3097, 0.0795, -0.0, -0.3086, 0.0765, -0.0, -0.3095, 0.0784, -0.0, -0.3075, 0.0773, -0.0, -0.3065, 0.0743, -0.0, -0.3095, 0.0722, -0.0, -0.3095, 0.0743, -0.0, -0.3073, 0.0722, -0.0, -0.3073, 0.0795, 0.0017, -0.3086, 0.0795, 0.0039, -0.3086, 0.0784, 0.0017, -0.3075, 0.0773, 0.0017, -0.3065, 0.0784, 0.0039, -0.3075, 0.0773, 0.0039, -0.3065,
0.0762, -0.0, -0.3054, 0.075, -0.0, -0.3043, 0.0722, -0.0, -0.3051, 0.0739, -0.0, -0.3034, 0.0726, -0.0, -0.3023, 0.0762, 0.0017, -0.3054, 0.0762, 0.0039, -0.3054, 0.075, 0.0017, -0.3044, 0.075, 0.0039, -0.3044, 0.0739, 0.0017, -0.3034, 0.0726, 0.0017, -0.3024, 0.0739, 0.0039, -0.3034, 0.0726, 0.0039, -0.3024, 0.0715, -0.0, -0.3015, 0.0715, 0.0017, -0.3015, 0.0715, 0.0039, -0.3015, 0.0704, -0.0, -0.3156, 0.0704, 0.0017, -0.3156, 0.0704, 0.0039, -0.3156, 0.0693, -0.0, -0.3144, 0.0683, -0.0, -0.3133, 0.07, -0.0, -0.3116, 0.0673, -0.0, -0.3122, 0.0661, -0.0, -0.3112, 0.0693, 0.0017, -0.3144, 0.0683, 0.0017, -0.3133, 0.0693, 0.0039, -0.3144, 0.0683, 0.0039, -0.3133, 0.0672, 0.0017, -0.3122, 0.0672, 0.0039, -0.3122,
0.0661, 0.0017, -0.3111, 0.0661, 0.0039, -0.3111, 0.0152, -0.0, -0.3446, 0.0144, -0.0, -0.3435, 0.0114, -0.0, -0.3442, 0.0133, -0.0, -0.3422, 0.0125, -0.0, -0.3411, 0.0152, 0.0017, -0.3446, 0.0152, 0.0039, -0.3446, 0.0143, 0.0017, -0.3435, 0.0143, 0.0039, -0.3435, 0.0132, 0.0017, -0.3421, 0.0124, 0.0017, -0.341, 0.0132, 0.0039, -0.3421, 0.0124, 0.0039, -0.341, 0.0113, -0.0, -0.3396, 0.0105, -0.0, -0.3385, 0.0115, 0.0017, -0.3398, 0.0115, 0.0039, -0.3398, 0.0105, 0.0017, -0.3385, 0.0105, 0.0039, -0.3385, 0.0092, -0.0, -0.3442, 0.0071, -0.0, -0.3442, 0.0092, -0.0, -0.342, 0.0071, -0.0, -0.342, 0.0049, -0.0, -0.3442, 0.0027, -0.0, -0.3442, 0.0049, -0.0, -0.342, 0.0027, -0.0, -0.342, 0.0092, -0.0, -0.3398,
0.0071, -0.0, -0.3398, 0.0097, -0.0, -0.3374, 0.0071, -0.0, -0.3377, 0.0049, -0.0, -0.3398, 0.0027, -0.0, -0.3398, 0.0049, -0.0, -0.3377, 0.0027, -0.0, -0.3377, 0.0097, 0.0017, -0.3374, 0.0097, 0.0039, -0.3374, 0.0086, -0.0, -0.3361, 0.0078, -0.0, -0.335, 0.0067, -0.0, -0.3336, 0.0049, -0.0, -0.3355, 0.0027, -0.0, -0.3355, 0.0059, -0.0, -0.3325, 0.0027, -0.0, -0.3333, 0.0086, 0.0017, -0.336, 0.0078, 0.0017, -0.3349, 0.0086, 0.0039, -0.336, 0.0078, 0.0039, -0.3349, 0.0067, 0.0017, -0.3335, 0.0067, 0.0039, -0.3335, 0.0059, 0.0017, -0.3324, 0.0059, 0.0039, -0.3324, 0.0047, -0.0, -0.3309, 0.0027, -0.0, -0.3312, 0.004, -0.0, -0.3298, 0.0032, -0.0, -0.3287, 0.0049, 0.0017, -0.3312, 0.0049, 0.0039, -0.3312,
0.004, 0.0017, -0.3298, 0.0032, 0.0017, -0.3287, 0.004, 0.0039, -0.3298, 0.0032, 0.0039, -0.3287, 0.0022, -0.0, -0.3273, 0.0021, 0.0017, -0.3272, 0.0021, 0.0039, -0.3272, 0.07, -0.0, -0.3095, 0.0678, -0.0, -0.3095, 0.07, -0.0, -0.3073, 0.0678, -0.0, -0.3073, 0.065, -0.0, -0.3102, 0.0637, -0.0, -0.3092, 0.0656, -0.0, -0.3073, 0.0626, -0.0, -0.3083, 0.065, 0.0017, -0.3102, 0.0636, 0.0017, -0.3091, 0.065, 0.0039, -0.3102, 0.0636, 0.0039, -0.3091, 0.0626, 0.0017, -0.3082, 0.0626, 0.0039, -0.3082, 0.07, -0.0, -0.3051, 0.0678, -0.0, -0.3051, 0.07, -0.0, -0.3029, 0.0678, -0.0, -0.3029, 0.0656, -0.0, -0.3051, 0.0635, -0.0, -0.3051, 0.0656, -0.0, -0.3029, 0.0635, -0.0, -0.3029, 0.0611, -0.0, -0.307,
0.0601, -0.0, -0.3063, 0.0613, 0.0017, -0.3072, 0.0601, 0.0017, -0.3063, 0.0613, 0.0039, -0.3072, 0.0601, 0.0039, -0.3063, 0.0613, -0.0, -0.3051, 0.059, -0.0, -0.3054, 0.0613, -0.0, -0.3029, 0.0591, -0.0, -0.3029, 0.0576, -0.0, -0.3044, 0.0565, -0.0, -0.3036, 0.0551, -0.0, -0.3026, 0.059, 0.0017, -0.3055, 0.059, 0.0039, -0.3055, 0.0576, 0.0017, -0.3044, 0.0576, 0.0039, -0.3044, 0.0565, 0.0017, -0.3036, 0.055, 0.0017, -0.3025, 0.0565, 0.0039, -0.3036, 0.055, 0.0039, -0.3025, 0.0702, -0.0, -0.3004, 0.0678, -0.0, -0.3008, 0.069, -0.0, -0.2996, 0.0681, -0.0, -0.2988, 0.0656, -0.0, -0.3008, 0.0635, -0.0, -0.3008, 0.0663, -0.0, -0.2976, 0.0635, -0.0, -0.2986, 0.0701, 0.0017, -0.3004, 0.0701, 0.0039, -0.3004,
0.069, 0.0017, -0.2996, 0.0678, 0.0017, -0.2987, 0.069, 0.0039, -0.2996, 0.0678, 0.0039, -0.2987, 0.0665, 0.0017, -0.2977, 0.0665, 0.0039, -0.2977, 0.0653, -0.0, -0.2969, 0.0637, -0.0, -0.2959, 0.0626, -0.0, -0.2952, 0.0654, 0.0017, -0.297, 0.0637, 0.0017, -0.2959, 0.0654, 0.0039, -0.297, 0.0637, 0.0039, -0.2959, 0.0627, 0.0017, -0.2952, 0.0627, 0.0039, -0.2952, 0.0613, -0.0, -0.3008, 0.0591, -0.0, -0.3008, 0.0613, -0.0, -0.2986, 0.0591, -0.0, -0.2986, 0.057, -0.0, -0.3008, 0.054, -0.0, -0.3019, 0.057, -0.0, -0.2986, 0.0548, -0.0, -0.2986, 0.0539, 0.0017, -0.3017, 0.0539, 0.0039, -0.3017, 0.0613, -0.0, -0.2964, 0.0591, -0.0, -0.2964, 0.0615, -0.0, -0.2945, 0.0597, -0.0, -0.2933, 0.057, -0.0, -0.2964,
0.0548, -0.0, -0.2964, 0.057, -0.0, -0.2943, 0.0548, -0.0, -0.2943, 0.0613, 0.0017, -0.2943, 0.0599, 0.0017, -0.2934, 0.0613, 0.0039, -0.2943, 0.0599, 0.0039, -0.2934, 0.0524, -0.0, -0.3007, 0.0512, -0.0, -0.3, 0.0526, -0.0, -0.2986, 0.0501, -0.0, -0.2993, 0.0481, -0.0, -0.2982, 0.047, -0.0, -0.2976, 0.0526, 0.0017, -0.3009, 0.0511, 0.0017, -0.3, 0.0526, 0.0039, -0.3009, 0.0511, 0.0039, -0.3, 0.05, 0.0017, -0.2993, 0.05, 0.0039, -0.2993, 0.0483, 0.0017, -0.2983, 0.047, 0.0017, -0.2976, 0.0483, 0.0039, -0.2983, 0.047, 0.0039, -0.2976, 0.0526, -0.0, -0.2964, 0.0505, -0.0, -0.2964, 0.0526, -0.0, -0.2943, 0.0505, -0.0, -0.2943, 0.0483, -0.0, -0.2964, 0.0459, -0.0, -0.2971, 0.0483, -0.0, -0.2943,
0.0461, -0.0, -0.2943, 0.0459, 0.0017, -0.2971, 0.0459, 0.0039, -0.2971, 0.0438, -0.0, -0.2961, 0.0426, -0.0, -0.2956, 0.0439, -0.0, -0.2943, 0.0415, -0.0, -0.2951, 0.0394, -0.0, -0.2943, 0.0378, -0.0, -0.2936, 0.044, 0.0017, -0.2962, 0.0425, 0.0017, -0.2955, 0.044, 0.0039, -0.2962, 0.0425, 0.0039, -0.2955, 0.0414, 0.0017, -0.2951, 0.0414, 0.0039, -0.2951, 0.0396, 0.0017, -0.2943, 0.0375, 0.0017, -0.2935, 0.0396, 0.0039, -0.2943, 0.0375, 0.0039, -0.2935, 0.0587, -0.0, -0.2927, 0.057, -0.0, -0.2917, 0.0548, -0.0, -0.2921, 0.0559, -0.0, -0.291, 0.0549, -0.0, -0.2904, 0.0588, 0.0017, -0.2927, 0.0588, 0.0039, -0.2927, 0.057, 0.0017, -0.2916, 0.057, 0.0039, -0.2916, 0.0559, 0.0017, -0.291, 0.0548, 0.0017, -0.2904,
0.0559, 0.0039, -0.291, 0.0548, 0.0039, -0.2904, 0.0526, -0.0, -0.2921, 0.0505, -0.0, -0.2921, 0.0527, -0.0, -0.2893, 0.0505, -0.0, -0.2899, 0.0483, -0.0, -0.2921, 0.0461, -0.0, -0.2921, 0.0483, -0.0, -0.2899, 0.0461, -0.0, -0.2899, 0.0527, 0.0017, -0.2893, 0.0527, 0.0039, -0.2893, 0.0516, -0.0, -0.2888, 0.0506, -0.0, -0.2884, 0.0486, -0.0, -0.2875, 0.0466, -0.0, -0.2867, 0.0457, -0.0, -0.2862, 0.0516, 0.0017, -0.2888, 0.0505, 0.0017, -0.2883, 0.0516, 0.0039, -0.2888, 0.0505, 0.0039, -0.2883, 0.0483, 0.0017, -0.2874, 0.0469, 0.0017, -0.2868, 0.0483, 0.0039, -0.2874, 0.0469, 0.0039, -0.2868, 0.0458, 0.0017, -0.2863, 0.0458, 0.0039, -0.2863, 0.0439, -0.0, -0.2921, 0.0418, -0.0, -0.2921, 0.0439, -0.0, -0.2899,
0.0418, -0.0, -0.2899, 0.0396, -0.0, -0.2921, 0.0369, -0.0, -0.2933, 0.0396, -0.0, -0.2899, 0.0374, -0.0, -0.2899, 0.0364, 0.0017, -0.2931, 0.0364, 0.0039, -0.2931, 0.0439, -0.0, -0.2878, 0.0418, -0.0, -0.2878, 0.0441, -0.0, -0.2856, 0.0419, -0.0, -0.2848, 0.0396, -0.0, -0.2878, 0.0374, -0.0, -0.2878, 0.0396, -0.0, -0.2856, 0.0374, -0.0, -0.2856, 0.044, 0.0017, -0.2855, 0.0419, 0.0017, -0.2848, 0.044, 0.0039, -0.2855, 0.0419, 0.0039, -0.2848, 0.0408, -0.0, -0.2844, 0.0397, -0.0, -0.2841, 0.0376, -0.0, -0.2835, 0.0408, 0.0017, -0.2845, 0.0408, 0.0039, -0.2845, 0.0396, 0.0017, -0.2841, 0.0374, 0.0017, -0.2834, 0.0396, 0.0039, -0.2841, 0.0374, 0.0039, -0.2834, 0.035, -0.0, -0.2926, 0.033, -0.0, -0.292,
0.0353, -0.0, -0.2899, 0.0331, -0.0, -0.2899, 0.0309, -0.0, -0.2914, 0.0297, -0.0, -0.2911, 0.0309, -0.0, -0.2899, 0.0287, -0.0, -0.2908, 0.0353, 0.0017, -0.2927, 0.0331, 0.0017, -0.292, 0.0353, 0.0039, -0.2927, 0.0331, 0.0039, -0.292, 0.0309, 0.0017, -0.2914, 0.0297, 0.0017, -0.2911, 0.0309, 0.0039, -0.2914, 0.0297, 0.0039, -0.2911, 0.0286, 0.0017, -0.2908, 0.0286, 0.0039, -0.2908, 0.0353, -0.0, -0.2878, 0.0331, -0.0, -0.2878, 0.0353, -0.0, -0.2856, 0.0331, -0.0, -0.2856, 0.0309, -0.0, -0.2878, 0.0288, -0.0, -0.2878, 0.0309, -0.0, -0.2856, 0.0288, -0.0, -0.2856, 0.0264, -0.0, -0.2903, 0.0243, -0.0, -0.2899, 0.0222, -0.0, -0.2894, 0.0206, -0.0, -0.2891, 0.0266, 0.0017, -0.2904, 0.0244, 0.0017, -0.2899,
0.0266, 0.0039, -0.2904, 0.0244, 0.0039, -0.2899, 0.0222, 0.0017, -0.2894, 0.0204, 0.0017, -0.289, 0.0222, 0.0039, -0.2894, 0.0204, 0.0039, -0.289, 0.0266, -0.0, -0.2878, 0.0244, -0.0, -0.2878, 0.0266, -0.0, -0.2856, 0.0244, -0.0, -0.2856, 0.0222, -0.0, -0.2878, 0.0198, -0.0, -0.2889, 0.0222, -0.0, -0.2856, 0.0201, -0.0, -0.2856, 0.0193, 0.0017, -0.2888, 0.0193, 0.0039, -0.2888, 0.0355, -0.0, -0.2829, 0.0334, -0.0, -0.2823, 0.0325, -0.0, -0.2821, 0.0309, -0.0, -0.2834, 0.0288, -0.0, -0.2834, 0.031, -0.0, -0.2817, 0.0289, -0.0, -0.2812, 0.0353, 0.0017, -0.2828, 0.0338, 0.0017, -0.2824, 0.0353, 0.0039, -0.2828, 0.0338, 0.0039, -0.2824, 0.0328, 0.0017, -0.2821, 0.0328, 0.0039, -0.2821, 0.0309, 0.0017, -0.2816,
0.0288, 0.0017, -0.2812, 0.0309, 0.0039, -0.2816, 0.0288, 0.0039, -0.2812, 0.0266, -0.0, -0.2834, 0.0244, -0.0, -0.2834, 0.0267, -0.0, -0.2808, 0.0246, -0.0, -0.2804, 0.0222, -0.0, -0.2834, 0.0201, -0.0, -0.2834, 0.0225, -0.0, -0.2801, 0.0201, -0.0, -0.2812, 0.0266, 0.0017, -0.2808, 0.0244, 0.0017, -0.2804, 0.0266, 0.0039, -0.2808, 0.0244, 0.0039, -0.2804, 0.0231, 0.0017, -0.2802, 0.0231, 0.0039, -0.2802, 0.0218, -0.0, -0.28, 0.0201, -0.0, -0.2797, 0.022, 0.0017, -0.28, 0.0201, 0.0017, -0.2797, 0.022, 0.0039, -0.28, 0.0201, 0.0039, -0.2797, 0.0178, -0.0, -0.2885, 0.0157, -0.0, -0.2883, 0.0179, -0.0, -0.2856, 0.0157, -0.0, -0.2856, 0.0136, -0.0, -0.2882, 0.0114, -0.0, -0.2882, 0.0136, -0.0, -0.2856,
0.0114, -0.0, -0.2856, 0.0179, 0.0017, -0.2885, 0.0157, 0.0017, -0.2883, 0.0179, 0.0039, -0.2885, 0.0157, 0.0039, -0.2883, 0.0136, 0.0017, -0.2882, 0.0114, 0.0017, -0.2882, 0.0136, 0.0039, -0.2882, 0.0114, 0.0039, -0.2882, 0.0092, -0.0, -0.2882, 0.0071, -0.0, -0.2882, 0.0092, -0.0, -0.2856, 0.0071, -0.0, -0.2856, 0.0049, -0.0, -0.2882, 0.0027, -0.0, -0.2882, 0.0049, -0.0, -0.2856, 0.0027, -0.0, -0.2856, 0.0092, 0.0017, -0.2882, 0.0071, 0.0017, -0.2882, 0.0092, 0.0039, -0.2882, 0.0071, 0.0039, -0.2882, 0.0049, 0.0017, -0.2882, 0.0027, 0.0017, -0.2882, 0.0049, 0.0039, -0.2882, 0.0027, 0.0039, -0.2882, 0.0179, -0.0, -0.2834, 0.0157, -0.0, -0.2834, 0.0179, -0.0, -0.2812, 0.0157, -0.0, -0.2812, 0.0136, -0.0, -0.2834,
0.0114, -0.0, -0.2834, 0.0136, -0.0, -0.2812, 0.0114, -0.0, -0.2812, 0.0179, -0.0, -0.2795, 0.0157, -0.0, -0.2794, 0.0136, -0.0, -0.2794, 0.0114, -0.0, -0.2794, 0.0179, 0.0017, -0.2795, 0.0157, 0.0017, -0.2794, 0.0179, 0.0039, -0.2795, 0.0157, 0.0039, -0.2794, 0.0136, 0.0017, -0.2794, 0.0114, 0.0017, -0.2794, 0.0136, 0.0039, -0.2794, 0.0114, 0.0039, -0.2794, 0.0092, -0.0, -0.2834, 0.0071, -0.0, -0.2834, 0.0092, -0.0, -0.2812, 0.0071, -0.0, -0.2812, 0.0049, -0.0, -0.2834, 0.0027, -0.0, -0.2834, 0.0049, -0.0, -0.2812, 0.0027, -0.0, -0.2812, 0.0092, -0.0, -0.2794, 0.0071, -0.0, -0.2794, 0.0049, -0.0, -0.2794, 0.0027, -0.0, -0.2794, 0.0092, 0.0017, -0.2794, 0.0071, 0.0017, -0.2794, 0.0092, 0.0039, -0.2794,
0.0071, 0.0039, -0.2794, 0.0049, 0.0017, -0.2794, 0.0027, 0.0017, -0.2794, 0.0049, 0.0039, -0.2794, 0.0027, 0.0039, -0.2794, 0.067, -0.0, -0.2196, 0.0666, -0.0, -0.2195, 0.067, 0.0017, -0.2196, 0.067, 0.0039, -0.2196, 0.0666, 0.0017, -0.2195, 0.0666, 0.0039, -0.2195, 0.067, -0.0, -0.2183, 0.0667, -0.0, -0.2172, 0.0655, -0.0, -0.219, 0.0633, -0.0, -0.218, 0.0665, -0.0, -0.216, 0.0635, -0.0, -0.2161, 0.067, 0.0017, -0.2183, 0.067, 0.0039, -0.2183, 0.0667, 0.0017, -0.2172, 0.0667, 0.0039, -0.2172, 0.0655, 0.0017, -0.219, 0.0635, 0.0017, -0.2181, 0.0655, 0.0039, -0.219, 0.0635, 0.0039, -0.2181, 0.0665, 0.0017, -0.216, 0.0665, 0.0039, -0.216, 0.0622, -0.0, -0.2174, 0.0611, -0.0, -0.2168, 0.0593, -0.0, -0.2158,
0.0621, 0.0017, -0.2174, 0.0621, 0.0039, -0.2174, 0.061, 0.0017, -0.2168, 0.0592, 0.0017, -0.2157, 0.061, 0.0039, -0.2168, 0.0592, 0.0039, -0.2157, 0.0654, -0.0, -0.2142, 0.0641, -0.0, -0.2131, 0.063, -0.0, -0.2124, 0.0652, 0.0017, -0.2141, 0.0642, 0.0017, -0.2132, 0.0652, 0.0039, -0.2141, 0.0642, 0.0039, -0.2132, 0.0631, 0.0017, -0.2124, 0.0631, 0.0039, -0.2124, 0.0613, -0.0, -0.214, 0.0583, -0.0, -0.2152, 0.0614, -0.0, -0.2114, 0.0591, -0.0, -0.2118, 0.0566, -0.0, -0.2142, 0.0553, -0.0, -0.2134, 0.057, -0.0, -0.2118, 0.0542, -0.0, -0.2127, 0.0581, 0.0017, -0.215, 0.0581, 0.0039, -0.215, 0.0614, 0.0017, -0.2113, 0.0614, 0.0039, -0.2113, 0.057, 0.0017, -0.2144, 0.0552, 0.0017, -0.2133, 0.057, 0.0039, -0.2144,
0.0552, 0.0039, -0.2133, 0.0541, 0.0017, -0.2126, 0.0541, 0.0039, -0.2126, 0.0603, -0.0, -0.2107, 0.0593, -0.0, -0.2101, 0.0573, -0.0, -0.2089, 0.0548, -0.0, -0.2096, 0.0562, -0.0, -0.2083, 0.0551, -0.0, -0.2075, 0.0603, 0.0017, -0.2107, 0.0591, 0.0017, -0.21, 0.0603, 0.0039, -0.2107, 0.0591, 0.0039, -0.21, 0.0574, 0.0017, -0.2089, 0.0574, 0.0039, -0.2089, 0.0563, 0.0017, -0.2083, 0.0548, 0.0017, -0.2074, 0.0563, 0.0039, -0.2083, 0.0548, 0.0039, -0.2074, 0.0524, -0.0, -0.2115, 0.0513, -0.0, -0.2109, 0.0526, 0.0017, -0.2117, 0.0513, 0.0017, -0.2109, 0.0526, 0.0039, -0.2117, 0.0513, 0.0039, -0.2109, 0.0526, -0.0, -0.2096, 0.0502, -0.0, -0.2102, 0.0532, -0.0, -0.2065, 0.0505, -0.0, -0.2075, 0.0486, -0.0, -0.2092,
0.0475, -0.0, -0.2086, 0.0458, -0.0, -0.2074, 0.0502, 0.0017, -0.2102, 0.0502, 0.0039, -0.2102, 0.0534, 0.0017, -0.2066, 0.0534, 0.0039, -0.2066, 0.0484, 0.0017, -0.2091, 0.0484, 0.0039, -0.2091, 0.0473, 0.0017, -0.2085, 0.0461, 0.0017, -0.2076, 0.0473, 0.0039, -0.2085, 0.0461, 0.0039, -0.2076, 0.0448, -0.0, -0.2066, 0.0448, 0.0017, -0.2066, 0.0448, 0.0039, -0.2066, 0.0219, -0.0, -0.2196, 0.0201, -0.0, -0.2196, 0.0219, 0.0017, -0.2196, 0.0201, 0.0017, -0.2196, 0.0219, 0.0039, -0.2196, 0.0201, 0.0039, -0.2196, 0.0219, -0.0, -0.2183, 0.0201, -0.0, -0.2183, 0.0219, -0.0, -0.2161, 0.0201, -0.0, -0.2161, 0.0219, 0.0017, -0.2183, 0.0219, 0.0039, -0.2183, 0.0219, 0.0017, -0.2161, 0.0219, 0.0039, -0.2161, 0.0219, -0.0, -0.214,
0.0201, -0.0, -0.214, 0.0219, -0.0, -0.2118, 0.0201, -0.0, -0.2118, 0.0219, 0.0017, -0.214, 0.0219, 0.0039, -0.214, 0.0219, 0.0017, -0.2118, 0.0219, 0.0039, -0.2118, 0.0219, -0.0, -0.2088, 0.0201, -0.0, -0.2088, 0.0219, 0.0017, -0.2088, 0.0201, 0.0017, -0.2088, 0.0219, 0.0039, -0.2088, 0.0201, 0.0039, -0.2088, 0.0179, -0.0, -0.2196, 0.0157, -0.0, -0.2196, 0.0136, -0.0, -0.2196, 0.0114, -0.0, -0.2196, 0.0179, 0.0017, -0.2196, 0.0157, 0.0017, -0.2196, 0.0179, 0.0039, -0.2196, 0.0157, 0.0039, -0.2196, 0.0136, 0.0017, -0.2196, 0.0114, 0.0017, -0.2196, 0.0136, 0.0039, -0.2196, 0.0114, 0.0039, -0.2196, 0.0179, -0.0, -0.2183, 0.0157, -0.0, -0.2183, 0.0179, -0.0, -0.2161, 0.0157, -0.0, -0.2161, 0.0136, -0.0, -0.2183,
0.0114, -0.0, -0.2183, 0.0136, -0.0, -0.2161, 0.0114, -0.0, -0.2161, 0.0092, -0.0, -0.2196, 0.0071, -0.0, -0.2196, 0.0049, -0.0, -0.2196, 0.0027, -0.0, -0.2196, 0.0092, 0.0017, -0.2196, 0.0071, 0.0017, -0.2196, 0.0092, 0.0039, -0.2196, 0.0071, 0.0039, -0.2196, 0.0049, 0.0017, -0.2196, 0.0027, 0.0017, -0.2196, 0.0049, 0.0039, -0.2196, 0.0027, 0.0039, -0.2196, 0.0092, -0.0, -0.2183, 0.0071, -0.0, -0.2183, 0.0092, -0.0, -0.2161, 0.0071, -0.0, -0.2161, 0.0049, -0.0, -0.2183, 0.0027, -0.0, -0.2183, 0.0049, -0.0, -0.2161, 0.0027, -0.0, -0.2161, 0.0179, -0.0, -0.214, 0.0157, -0.0, -0.214, 0.0179, -0.0, -0.2118, 0.0157, -0.0, -0.2118, 0.0136, -0.0, -0.214, 0.0114, -0.0, -0.214, 0.0136, -0.0, -0.2118,
0.0114, -0.0, -0.2118, 0.0179, -0.0, -0.2088, 0.0157, -0.0, -0.2088, 0.0136, -0.0, -0.2088, 0.0114, -0.0, -0.2088, 0.0179, 0.0017, -0.2088, 0.0157, 0.0017, -0.2088, 0.0179, 0.0039, -0.2088, 0.0157, 0.0039, -0.2088, 0.0136, 0.0017, -0.2088, 0.0114, 0.0017, -0.2088, 0.0136, 0.0039, -0.2088, 0.0114, 0.0039, -0.2088, 0.0092, -0.0, -0.214, 0.0071, -0.0, -0.214, 0.0092, -0.0, -0.2118, 0.0071, -0.0, -0.2118, 0.0049, -0.0, -0.214, 0.0027, -0.0, -0.214, 0.0049, -0.0, -0.2118, 0.0027, -0.0, -0.2118, 0.0092, -0.0, -0.2088, 0.0071, -0.0, -0.2088, 0.0049, -0.0, -0.2088, 0.0027, -0.0, -0.2088, 0.0092, 0.0017, -0.2088, 0.0071, 0.0017, -0.2088, 0.0092, 0.0039, -0.2088, 0.0071, 0.0039, -0.2088, 0.0049, 0.0017, -0.2088,
0.0027, 0.0017, -0.2088, 0.0049, 0.0039, -0.2088, 0.0027, 0.0039, -0.2088, 0.0591, -0.0, -0.2069, 0.0591, -0.0, -0.2031, 0.0573, -0.0, -0.206, 0.0563, -0.0, -0.2033, 0.0591, 0.0017, -0.2069, 0.0591, 0.0039, -0.2069, 0.0591, 0.0017, -0.2031, 0.0591, 0.0039, -0.2031, 0.0573, 0.0017, -0.206, 0.0573, 0.0039, -0.206, 0.0563, 0.0017, -0.2032, 0.0563, 0.0039, -0.2032, 0.0591, -0.0, -0.2009, 0.0591, -0.0, -0.1988, 0.0562, -0.0, -0.201, 0.054, -0.0, -0.2001, 0.0562, -0.0, -0.199, 0.055, -0.0, -0.1995, 0.0591, 0.0017, -0.2009, 0.0591, 0.0039, -0.2009, 0.0591, 0.0017, -0.1988, 0.0591, 0.0039, -0.1988, 0.0562, 0.0017, -0.2009, 0.054, 0.0017, -0.2, 0.0562, 0.0039, -0.2009, 0.054, 0.0039, -0.2, 0.0562, 0.0017, -0.199,
0.0551, 0.0017, -0.1995, 0.0562, 0.0039, -0.199, 0.0551, 0.0039, -0.1995, 0.0634, -0.0, -0.1942, 0.0635, 0.0017, -0.1941, 0.0635, 0.0039, -0.1941, 0.0668, -0.0, -0.1893, 0.0656, -0.0, -0.1923, 0.0645, -0.0, -0.1934, 0.0667, -0.0, -0.1904, 0.063, -0.0, -0.1891, 0.0668, 0.0017, -0.1894, 0.0668, 0.0039, -0.1894, 0.0656, 0.0017, -0.1924, 0.0645, 0.0017, -0.1933, 0.0656, 0.0039, -0.1924, 0.0645, 0.0039, -0.1933, 0.0666, 0.0017, -0.1905, 0.0628, 0.0017, -0.1892, 0.0666, 0.0039, -0.1905, 0.0628, 0.0039, -0.1892, 0.0605, -0.0, -0.1957, 0.0591, -0.0, -0.1961, 0.0616, -0.0, -0.1952, 0.0591, -0.0, -0.1944, 0.057, -0.0, -0.1966, 0.0548, -0.0, -0.1966, 0.057, -0.0, -0.1944, 0.0546, -0.0, -0.1939, 0.0605, 0.0017, -0.1956,
0.0591, 0.0017, -0.1961, 0.0605, 0.0039, -0.1956, 0.0591, 0.0039, -0.1961, 0.0616, 0.0017, -0.1952, 0.0616, 0.0039, -0.1952, 0.0546, 0.0017, -0.1939, 0.0546, 0.0039, -0.1939, 0.0613, -0.0, -0.1923, 0.0588, -0.0, -0.1915, 0.0611, -0.0, -0.1902, 0.0599, -0.0, -0.1909, 0.0568, -0.0, -0.1927, 0.0558, -0.0, -0.1933, 0.0587, 0.0017, -0.1916, 0.0587, 0.0039, -0.1916, 0.0613, 0.0017, -0.1901, 0.0598, 0.0017, -0.1909, 0.0613, 0.0039, -0.1901, 0.0598, 0.0039, -0.1909, 0.057, 0.0017, -0.1926, 0.0557, 0.0017, -0.1933, 0.057, 0.0039, -0.1926, 0.0557, 0.0039, -0.1933, 0.0522, -0.0, -0.2059, 0.0508, -0.0, -0.205, 0.05, -0.0, -0.2026, 0.0489, -0.0, -0.2041, 0.0461, -0.0, -0.2053, 0.0484, -0.0, -0.2039, 0.0461, -0.0, -0.2031,
0.0523, 0.0017, -0.2059, 0.0505, 0.0017, -0.2049, 0.0523, 0.0039, -0.2059, 0.0505, 0.0039, -0.2049, 0.0501, 0.0017, -0.2026, 0.0501, 0.0039, -0.2026, 0.0492, 0.0017, -0.2043, 0.0492, 0.0039, -0.2043, 0.0484, 0.0017, -0.2039, 0.0484, 0.0039, -0.2039, 0.0528, -0.0, -0.2007, 0.0511, -0.0, -0.2019, 0.0526, -0.0, -0.1988, 0.0505, -0.0, -0.1988, 0.0483, -0.0, -0.2009, 0.0461, -0.0, -0.2009, 0.0477, -0.0, -0.1978, 0.0458, -0.0, -0.199, 0.0526, 0.0017, -0.2008, 0.0512, 0.0017, -0.2018, 0.0526, 0.0039, -0.2008, 0.0512, 0.0039, -0.2018, 0.0476, 0.0017, -0.1979, 0.046, 0.0017, -0.1989, 0.0476, 0.0039, -0.1979, 0.046, 0.0039, -0.1989, 0.0436, -0.0, -0.2052, 0.043, -0.0, -0.2026, 0.0436, 0.0017, -0.2052, 0.0436, 0.0039, -0.2052,
0.043, 0.0017, -0.2026, 0.043, 0.0039, -0.2026, 0.0441, -0.0, -0.2006, 0.045, -0.0, -0.1998, 0.0441, 0.0017, -0.2006, 0.0441, 0.0039, -0.2006, 0.045, 0.0017, -0.1998, 0.045, 0.0039, -0.1998, 0.0519, -0.0, -0.1955, 0.0502, -0.0, -0.1965, 0.0529, -0.0, -0.1949, 0.0488, -0.0, -0.1972, 0.0517, 0.0017, -0.1956, 0.0505, 0.0017, -0.1963, 0.0517, 0.0039, -0.1956, 0.0505, 0.0039, -0.1963, 0.0528, 0.0017, -0.195, 0.0528, 0.0039, -0.195, 0.0487, 0.0017, -0.1973, 0.0487, 0.0039, -0.1973, 0.067, -0.0, -0.1873, 0.0654, -0.0, -0.1879, 0.064, -0.0, -0.1886, 0.067, 0.0017, -0.1873, 0.067, 0.0039, -0.1873, 0.0656, 0.0017, -0.1878, 0.0639, 0.0017, -0.1886, 0.0656, 0.0039, -0.1878, 0.0639, 0.0039, -0.1886, 0.067, -0.0, -0.1686,
0.0656, -0.0, -0.1686, 0.0635, -0.0, -0.1686, 0.067, 0.0017, -0.1686, 0.067, 0.0039, -0.1686, 0.0656, 0.0017, -0.1686, 0.0635, 0.0017, -0.1686, 0.0656, 0.0039, -0.1686, 0.0635, 0.0039, -0.1686, 0.067, -0.0, -0.1662, 0.067, -0.0, -0.1641, 0.0656, -0.0, -0.1662, 0.0635, -0.0, -0.1662, 0.0656, -0.0, -0.1641, 0.0631, -0.0, -0.1647, 0.067, 0.0017, -0.1662, 0.067, 0.0039, -0.1662, 0.067, 0.0017, -0.1641, 0.067, 0.0039, -0.1641, 0.0631, 0.0017, -0.1647, 0.0631, 0.0039, -0.1647, 0.0613, -0.0, -0.1686, 0.0591, -0.0, -0.1686, 0.057, -0.0, -0.1686, 0.0548, -0.0, -0.1686, 0.0613, 0.0017, -0.1686, 0.0591, 0.0017, -0.1686, 0.0613, 0.0039, -0.1686, 0.0591, 0.0039, -0.1686, 0.057, 0.0017, -0.1686, 0.0548, 0.0017, -0.1686,
0.057, 0.0039, -0.1686, 0.0548, 0.0039, -0.1686, 0.0613, -0.0, -0.1662, 0.0591, -0.0, -0.1662, 0.0613, -0.0, -0.1647, 0.0591, -0.0, -0.1647, 0.057, -0.0, -0.1662, 0.0548, -0.0, -0.1662, 0.057, -0.0, -0.1647, 0.0548, -0.0, -0.1647, 0.0613, 0.0017, -0.1647, 0.0591, 0.0017, -0.1647, 0.0613, 0.0039, -0.1647, 0.0591, 0.0039, -0.1647, 0.057, 0.0017, -0.1647, 0.0548, 0.0017, -0.1647, 0.057, 0.0039, -0.1647, 0.0548, 0.0039, -0.1647, 0.067, -0.0, -0.1619, 0.067, -0.0, -0.1597, 0.0656, -0.0, -0.1619, 0.0631, -0.0, -0.1619, 0.0656, -0.0, -0.1597, 0.0631, -0.0, -0.1597, 0.067, 0.0017, -0.1619, 0.067, 0.0039, -0.1619, 0.067, 0.0017, -0.1597, 0.067, 0.0039, -0.1597, 0.0631, 0.0017, -0.1619, 0.0631, 0.0039, -0.1619,
0.0631, 0.0017, -0.1597, 0.0631, 0.0039, -0.1597, 0.067, -0.0, -0.1575, 0.067, -0.0, -0.1554, 0.0656, -0.0, -0.1575, 0.0631, -0.0, -0.1576, 0.0656, -0.0, -0.1554, 0.063, -0.0, -0.1554, 0.067, 0.0017, -0.1575, 0.067, 0.0039, -0.1575, 0.067, 0.0017, -0.1554, 0.067, 0.0039, -0.1554, 0.0631, 0.0017, -0.1575, 0.0631, 0.0039, -0.1575, 0.063, 0.0017, -0.1554, 0.063, 0.0039, -0.1554, 0.0526, -0.0, -0.1686, 0.0505, -0.0, -0.1686, 0.0483, -0.0, -0.1686, 0.0461, -0.0, -0.1686, 0.0526, 0.0017, -0.1686, 0.0505, 0.0017, -0.1686, 0.0526, 0.0039, -0.1686, 0.0505, 0.0039, -0.1686, 0.0483, 0.0017, -0.1686, 0.0461, 0.0017, -0.1686, 0.0483, 0.0039, -0.1686, 0.0461, 0.0039, -0.1686, 0.0526, -0.0, -0.1662, 0.0505, -0.0, -0.1662,
0.0526, -0.0, -0.1647, 0.0505, -0.0, -0.1647, 0.0483, -0.0, -0.1662, 0.0461, -0.0, -0.1662, 0.0483, -0.0, -0.1647, 0.0454, -0.0, -0.1647, 0.0526, 0.0017, -0.1647, 0.0505, 0.0017, -0.1647, 0.0526, 0.0039, -0.1647, 0.0505, 0.0039, -0.1647, 0.0483, 0.0017, -0.1647, 0.0454, 0.0017, -0.1647, 0.0483, 0.0039, -0.1647, 0.0454, 0.0039, -0.1647, 0.0439, -0.0, -0.1686, 0.0425, -0.0, -0.1686, 0.0439, 0.0017, -0.1686, 0.0425, 0.0017, -0.1686, 0.0439, 0.0039, -0.1686, 0.0425, 0.0039, -0.1686, 0.0439, -0.0, -0.1662, 0.0425, -0.0, -0.1662, 0.0439, -0.0, -0.1641, 0.0425, -0.0, -0.1641, 0.0425, 0.0017, -0.1662, 0.0425, 0.0039, -0.1662, 0.0425, 0.0017, -0.1641, 0.0425, 0.0039, -0.1641, 0.0454, -0.0, -0.1619, 0.0454, -0.0, -0.1597,
0.0454, 0.0017, -0.1619, 0.0454, 0.0039, -0.1619, 0.0454, 0.0017, -0.1597, 0.0454, 0.0039, -0.1597, 0.0455, -0.0, -0.1576, 0.0456, -0.0, -0.1554, 0.0455, 0.0017, -0.1575, 0.0455, 0.0039, -0.1575, 0.0457, 0.0017, -0.1554, 0.0457, 0.0039, -0.1554, 0.0439, -0.0, -0.1619, 0.0425, -0.0, -0.1619, 0.0439, -0.0, -0.1597, 0.0425, -0.0, -0.1597, 0.0425, 0.0017, -0.1619, 0.0425, 0.0039, -0.1619, 0.0425, 0.0017, -0.1597, 0.0425, 0.0039, -0.1597, 0.0439, -0.0, -0.1575, 0.0425, -0.0, -0.1575, 0.0439, -0.0, -0.1554, 0.0425, -0.0, -0.1554, 0.0425, 0.0017, -0.1575, 0.0425, 0.0039, -0.1575, 0.0425, 0.0017, -0.1554, 0.0425, 0.0039, -0.1554, 0.0669, -0.0, -0.1537, 0.0668, -0.0, -0.1529, 0.0629, -0.0, -0.1533, 0.0664, -0.0, -0.1507,
0.0624, -0.0, -0.1513, 0.0669, 0.0017, -0.1535, 0.0669, 0.0039, -0.1535, 0.0667, 0.0017, -0.1524, 0.0629, 0.0017, -0.1532, 0.0667, 0.0039, -0.1524, 0.0629, 0.0039, -0.1532, 0.0665, 0.0017, -0.151, 0.0625, 0.0017, -0.1516, 0.0665, 0.0039, -0.151, 0.0625, 0.0039, -0.1516, 0.0655, -0.0, -0.1486, 0.0635, -0.0, -0.1489, 0.0648, -0.0, -0.1475, 0.064, -0.0, -0.1463, 0.0656, 0.0017, -0.1488, 0.0656, 0.0039, -0.1488, 0.0648, 0.0017, -0.1475, 0.064, 0.0017, -0.1463, 0.0648, 0.0039, -0.1475, 0.064, 0.0039, -0.1463, 0.0621, -0.0, -0.1504, 0.0621, 0.0017, -0.1504, 0.0621, 0.0039, -0.1504, 0.0612, -0.0, -0.149, 0.06, -0.0, -0.148, 0.0613, -0.0, -0.1467, 0.059, -0.0, -0.1472, 0.0569, -0.0, -0.1464, 0.0548, -0.0, -0.1462,
0.0611, 0.0017, -0.1489, 0.06, 0.0017, -0.148, 0.0611, 0.0039, -0.1489, 0.06, 0.0039, -0.148, 0.059, 0.0017, -0.1472, 0.059, 0.0039, -0.1472, 0.057, 0.0017, -0.1465, 0.0548, 0.0017, -0.1462, 0.057, 0.0039, -0.1465, 0.0548, 0.0039, -0.1462, 0.0628, -0.0, -0.1452, 0.0628, 0.0017, -0.1452, 0.0628, 0.0039, -0.1452, 0.0615, -0.0, -0.1441, 0.0591, -0.0, -0.1445, 0.0603, -0.0, -0.1434, 0.0593, -0.0, -0.1429, 0.057, -0.0, -0.1445, 0.0548, -0.0, -0.1445, 0.0571, -0.0, -0.1423, 0.0548, -0.0, -0.1422, 0.0614, 0.0017, -0.144, 0.0614, 0.0039, -0.144, 0.0603, 0.0017, -0.1434, 0.0592, 0.0017, -0.1428, 0.0603, 0.0039, -0.1434, 0.0592, 0.0039, -0.1428, 0.057, 0.0017, -0.1423, 0.0548, 0.0017, -0.1422, 0.057, 0.0039, -0.1423,
0.0548, 0.0039, -0.1422, 0.0461, -0.0, -0.1533, 0.0474, -0.0, -0.1502, 0.0469, -0.0, -0.1512, 0.0461, 0.0017, -0.1532, 0.0461, 0.0039, -0.1532, 0.0474, 0.0017, -0.1503, 0.0468, 0.0017, -0.1514, 0.0474, 0.0039, -0.1503, 0.0468, 0.0039, -0.1514, 0.0495, -0.0, -0.1479, 0.0527, -0.0, -0.1464, 0.0505, -0.0, -0.1471, 0.0484, -0.0, -0.1489, 0.0461, -0.0, -0.1489, 0.0483, -0.0, -0.1467, 0.0451, -0.0, -0.1458, 0.0495, 0.0017, -0.1479, 0.0495, 0.0039, -0.1479, 0.0526, 0.0017, -0.1464, 0.0509, 0.0017, -0.1469, 0.0526, 0.0039, -0.1464, 0.0509, 0.0039, -0.1469, 0.0484, 0.0017, -0.1489, 0.0484, 0.0039, -0.1489, 0.0452, 0.0017, -0.1458, 0.0452, 0.0039, -0.1458, 0.0439, -0.0, -0.1532, 0.0425, -0.0, -0.1532, 0.0428, -0.0, -0.1506,
0.0426, -0.0, -0.1515, 0.0425, 0.0017, -0.1532, 0.0425, 0.0039, -0.1532, 0.0429, 0.0017, -0.1501, 0.0426, 0.0017, -0.1511, 0.0429, 0.0039, -0.1501, 0.0426, 0.0039, -0.1511, 0.0433, -0.0, -0.1485, 0.0443, -0.0, -0.1469, 0.0432, 0.0017, -0.1488, 0.0432, 0.0039, -0.1488, 0.0443, 0.0017, -0.1468, 0.0443, 0.0039, -0.1468, 0.0526, -0.0, -0.1445, 0.0505, -0.0, -0.1445, 0.0526, -0.0, -0.1422, 0.0503, -0.0, -0.1425, 0.0478, -0.0, -0.1436, 0.0462, -0.0, -0.1447, 0.0489, -0.0, -0.143, 0.0526, 0.0017, -0.1422, 0.0504, 0.0017, -0.1425, 0.0526, 0.0039, -0.1422, 0.0504, 0.0039, -0.1425, 0.0477, 0.0017, -0.1437, 0.0462, 0.0017, -0.1447, 0.0477, 0.0039, -0.1437, 0.0462, 0.0039, -0.1447, 0.0487, 0.0017, -0.1431, 0.0487, 0.0039, -0.1431,
0.0219, -0.0, -0.151, 0.0201, -0.0, -0.151, 0.0219, 0.0017, -0.151, 0.0201, 0.0017, -0.151, 0.0219, 0.0039, -0.151, 0.0201, 0.0039, -0.151, 0.0219, -0.0, -0.1489, 0.0201, -0.0, -0.1489, 0.0219, -0.0, -0.1467, 0.0201, -0.0, -0.1467, 0.0219, 0.0017, -0.1489, 0.0219, 0.0039, -0.1489, 0.0219, 0.0017, -0.1467, 0.0219, 0.0039, -0.1467, 0.0219, -0.0, -0.1445, 0.0201, -0.0, -0.1445, 0.0219, -0.0, -0.1424, 0.0201, -0.0, -0.1424, 0.0219, 0.0017, -0.1445, 0.0219, 0.0039, -0.1445, 0.0219, 0.0017, -0.1424, 0.0219, 0.0039, -0.1424, 0.0219, -0.0, -0.1402, 0.0201, -0.0, -0.1402, 0.0219, -0.0, -0.138, 0.0201, -0.0, -0.138, 0.0219, 0.0017, -0.1402, 0.0219, 0.0039, -0.1402, 0.0219, 0.0017, -0.138, 0.0219, 0.0039, -0.138,
0.0179, -0.0, -0.151, 0.0157, -0.0, -0.151, 0.0136, -0.0, -0.151, 0.0114, -0.0, -0.151, 0.0179, 0.0017, -0.151, 0.0157, 0.0017, -0.151, 0.0179, 0.0039, -0.151, 0.0157, 0.0039, -0.151, 0.0136, 0.0017, -0.151, 0.0114, 0.0017, -0.151, 0.0136, 0.0039, -0.151, 0.0114, 0.0039, -0.151, 0.0179, -0.0, -0.1489, 0.0157, -0.0, -0.1489, 0.0179, -0.0, -0.1467, 0.0157, -0.0, -0.1467, 0.0136, -0.0, -0.1489, 0.0114, -0.0, -0.1489, 0.0136, -0.0, -0.1467, 0.0114, -0.0, -0.1467, 0.0092, -0.0, -0.151, 0.0071, -0.0, -0.151, 0.0049, -0.0, -0.151, 0.0027, -0.0, -0.151, 0.0092, 0.0017, -0.151, 0.0071, 0.0017, -0.151, 0.0092, 0.0039, -0.151, 0.0071, 0.0039, -0.151, 0.0049, 0.0017, -0.151, 0.0027, 0.0017, -0.151,
0.0049, 0.0039, -0.151, 0.0027, 0.0039, -0.151, 0.0092, -0.0, -0.1489, 0.0071, -0.0, -0.1489, 0.0092, -0.0, -0.1467, 0.0071, -0.0, -0.1467, 0.0049, -0.0, -0.1489, 0.0027, -0.0, -0.1489, 0.0049, -0.0, -0.1467, 0.0027, -0.0, -0.1467, 0.0179, -0.0, -0.1445, 0.0157, -0.0, -0.1445, 0.0179, -0.0, -0.1424, 0.0157, -0.0, -0.1424, 0.0136, -0.0, -0.1445, 0.0114, -0.0, -0.1445, 0.0136, -0.0, -0.1424, 0.0114, -0.0, -0.1424, 0.0179, -0.0, -0.1402, 0.0157, -0.0, -0.1402, 0.0179, -0.0, -0.138, 0.0157, -0.0, -0.138, 0.0136, -0.0, -0.1402, 0.0114, -0.0, -0.1402, 0.0136, -0.0, -0.138, 0.0114, -0.0, -0.138, 0.0092, -0.0, -0.1445, 0.0071, -0.0, -0.1445, 0.0092, -0.0, -0.1424, 0.0071, -0.0, -0.1424,
0.0049, -0.0, -0.1445, 0.0027, -0.0, -0.1445, 0.0049, -0.0, -0.1424, 0.0027, -0.0, -0.1424, 0.0092, -0.0, -0.1402, 0.0071, -0.0, -0.1402, 0.0087, -0.0, -0.1375, 0.0072, -0.0, -0.1387, 0.0048, -0.0, -0.1391, 0.0027, -0.0, -0.1392, 0.0056, -0.0, -0.139, 0.0086, 0.0017, -0.1376, 0.0072, 0.0017, -0.1387, 0.0086, 0.0039, -0.1376, 0.0072, 0.0039, -0.1387, 0.0043, 0.0017, -0.1391, 0.0027, 0.0017, -0.1392, 0.0043, 0.0039, -0.1391, 0.0027, 0.0039, -0.1392, 0.0054, 0.0017, -0.139, 0.0054, 0.0039, -0.139, 0.067, -0.0, -0.1225, 0.067, -0.0, -0.1207, 0.0656, -0.0, -0.1225, 0.0635, -0.0, -0.1225, 0.0656, -0.0, -0.1207, 0.0635, -0.0, -0.1207, 0.067, 0.0017, -0.1225, 0.067, 0.0039, -0.1225, 0.067, 0.0017, -0.1207,
0.067, 0.0039, -0.1207, 0.0656, 0.0017, -0.1225, 0.0635, 0.0017, -0.1225, 0.0656, 0.0039, -0.1225, 0.0635, 0.0039, -0.1225, 0.0613, -0.0, -0.1225, 0.0591, -0.0, -0.1225, 0.0613, -0.0, -0.1207, 0.0591, -0.0, -0.1207, 0.057, -0.0, -0.1225, 0.0548, -0.0, -0.1225, 0.057, -0.0, -0.1207, 0.0548, -0.0, -0.1207, 0.0613, 0.0017, -0.1225, 0.0591, 0.0017, -0.1225, 0.0613, 0.0039, -0.1225, 0.0591, 0.0039, -0.1225, 0.057, 0.0017, -0.1225, 0.0548, 0.0017, -0.1225, 0.057, 0.0039, -0.1225, 0.0548, 0.0039, -0.1225, 0.0526, -0.0, -0.1225, 0.0505, -0.0, -0.1225, 0.0526, -0.0, -0.1207, 0.0505, -0.0, -0.1207, 0.0483, -0.0, -0.1225, 0.0461, -0.0, -0.1225, 0.0483, -0.0, -0.1207, 0.0461, -0.0, -0.1207, 0.0526, 0.0017, -0.1225,
0.0505, 0.0017, -0.1225, 0.0526, 0.0039, -0.1225, 0.0505, 0.0039, -0.1225, 0.0483, 0.0017, -0.1225, 0.0461, 0.0017, -0.1225, 0.0483, 0.0039, -0.1225, 0.0461, 0.0039, -0.1225, 0.0439, -0.0, -0.1225, 0.0425, -0.0, -0.1225, 0.0439, -0.0, -0.1207, 0.0425, -0.0, -0.1207, 0.0439, 0.0017, -0.1225, 0.0425, 0.0017, -0.1225, 0.0439, 0.0039, -0.1225, 0.0425, 0.0039, -0.1225, 0.0425, 0.0017, -0.1207, 0.0425, 0.0039, -0.1207, 0.067, -0.0, -0.1185, 0.067, -0.0, -0.1163, 0.0656, -0.0, -0.1185, 0.0631, -0.0, -0.1186, 0.0656, -0.0, -0.1163, 0.0631, -0.0, -0.1163, 0.067, 0.0017, -0.1185, 0.067, 0.0039, -0.1185, 0.067, 0.0017, -0.1163, 0.067, 0.0039, -0.1163, 0.0631, 0.0017, -0.1186, 0.0631, 0.0039, -0.1186, 0.0631, 0.0017, -0.1163,
0.0631, 0.0039, -0.1163, 0.067, -0.0, -0.1141, 0.067, -0.0, -0.112, 0.0656, -0.0, -0.1141, 0.0631, -0.0, -0.1141, 0.0656, -0.0, -0.112, 0.063, -0.0, -0.112, 0.067, 0.0017, -0.1141, 0.067, 0.0039, -0.1141, 0.067, 0.0017, -0.112, 0.067, 0.0039, -0.112, 0.0631, 0.0017, -0.1141, 0.0631, 0.0039, -0.1141, 0.063, 0.0017, -0.112, 0.063, 0.0039, -0.112, 0.0613, -0.0, -0.1186, 0.0591, -0.0, -0.1186, 0.057, -0.0, -0.1186, 0.0548, -0.0, -0.1186, 0.0613, 0.0017, -0.1186, 0.0591, 0.0017, -0.1186, 0.0613, 0.0039, -0.1186, 0.0591, 0.0039, -0.1186, 0.057, 0.0017, -0.1186, 0.0548, 0.0017, -0.1186, 0.057, 0.0039, -0.1186, 0.0548, 0.0039, -0.1186, 0.067, -0.0, -0.1098, 0.0669, -0.0, -0.108, 0.0656, -0.0, -0.1098,
0.063, -0.0, -0.1098, 0.0669, -0.0, -0.1072, 0.0629, -0.0, -0.1078, 0.067, 0.0017, -0.1098, 0.067, 0.0039, -0.1098, 0.0669, 0.0017, -0.1076, 0.0669, 0.0039, -0.1076, 0.063, 0.0017, -0.1098, 0.063, 0.0039, -0.1098, 0.0667, 0.0017, -0.1066, 0.0629, 0.0017, -0.1076, 0.0667, 0.0039, -0.1066, 0.0629, 0.0039, -0.1076, 0.0665, -0.0, -0.1052, 0.0624, -0.0, -0.1058, 0.0658, -0.0, -0.1031, 0.0635, -0.0, -0.1033, 0.0666, 0.0017, -0.1054, 0.0625, 0.0017, -0.1061, 0.0666, 0.0039, -0.1054, 0.0625, 0.0039, -0.1061, 0.0658, 0.0017, -0.1033, 0.0658, 0.0039, -0.1033, 0.0622, -0.0, -0.1048, 0.0616, -0.0, -0.1034, 0.0622, 0.0017, -0.105, 0.0622, 0.0039, -0.105, 0.0613, 0.0017, -0.103, 0.0613, 0.0039, -0.103, 0.0526, -0.0, -0.1186,
0.0505, -0.0, -0.1186, 0.0483, -0.0, -0.1186, 0.0454, -0.0, -0.1186, 0.0454, -0.0, -0.1163, 0.0526, 0.0017, -0.1186, 0.0505, 0.0017, -0.1186, 0.0526, 0.0039, -0.1186, 0.0505, 0.0039, -0.1186, 0.0483, 0.0017, -0.1186, 0.0454, 0.0017, -0.1186, 0.0483, 0.0039, -0.1186, 0.0454, 0.0039, -0.1186, 0.0454, 0.0017, -0.1163, 0.0454, 0.0039, -0.1163, 0.0454, -0.0, -0.1141, 0.0454, -0.0, -0.112, 0.0454, 0.0017, -0.1141, 0.0454, 0.0039, -0.1141, 0.0454, 0.0017, -0.112, 0.0454, 0.0039, -0.112, 0.0439, -0.0, -0.1185, 0.0425, -0.0, -0.1185, 0.0439, -0.0, -0.1163, 0.0425, -0.0, -0.1163, 0.0425, 0.0017, -0.1185, 0.0425, 0.0039, -0.1185, 0.0425, 0.0017, -0.1163, 0.0425, 0.0039, -0.1163, 0.0439, -0.0, -0.1141, 0.0425, -0.0, -0.1141,
0.0439, -0.0, -0.112, 0.0425, -0.0, -0.112, 0.0425, 0.0017, -0.1141, 0.0425, 0.0039, -0.1141, 0.0425, 0.0017, -0.112, 0.0425, 0.0039, -0.112, 0.0455, -0.0, -0.1098, 0.0459, -0.0, -0.1077, 0.0455, 0.0017, -0.1098, 0.0455, 0.0039, -0.1098, 0.0459, 0.0017, -0.1076, 0.0459, 0.0039, -0.1076, 0.0464, -0.0, -0.1057, 0.0479, -0.0, -0.103, 0.0473, -0.0, -0.104, 0.0465, 0.0017, -0.1054, 0.0465, 0.0039, -0.1054, 0.0478, 0.0017, -0.1031, 0.0471, 0.0017, -0.1043, 0.0478, 0.0039, -0.1031, 0.0471, 0.0039, -0.1043, 0.0439, -0.0, -0.1098, 0.0425, -0.0, -0.1098, 0.0439, -0.0, -0.1076, 0.0425, -0.0, -0.1076, 0.0425, 0.0017, -0.1098, 0.0425, 0.0039, -0.1098, 0.0425, 0.0017, -0.1076, 0.0425, 0.0039, -0.1076, 0.0439, -0.0, -0.1055,
0.0426, -0.0, -0.1054, 0.0431, -0.0, -0.1032, 0.0428, -0.0, -0.1043, 0.0426, 0.0017, -0.1054, 0.0426, 0.0039, -0.1054, 0.0431, 0.0017, -0.1031, 0.0428, 0.0017, -0.1042, 0.0431, 0.0039, -0.1031, 0.0428, 0.0039, -0.1042, 0.0219, -0.0, -0.1358, 0.0201, -0.0, -0.1358, 0.0219, -0.0, -0.1337, 0.0201, -0.0, -0.1337, 0.0219, 0.0017, -0.1358, 0.0219, 0.0039, -0.1358, 0.0219, 0.0017, -0.1337, 0.0219, 0.0039, -0.1337, 0.0219, -0.0, -0.1315, 0.0201, -0.0, -0.1315, 0.0219, -0.0, -0.1293, 0.0201, -0.0, -0.1293, 0.0219, 0.0017, -0.1315, 0.0219, 0.0039, -0.1315, 0.0219, 0.0017, -0.1293, 0.0219, 0.0039, -0.1293, 0.0219, -0.0, -0.1272, 0.0201, -0.0, -0.1272, 0.0219, -0.0, -0.125, 0.0201, -0.0, -0.125, 0.0219, 0.0017, -0.1272,
0.0219, 0.0039, -0.1272, 0.0219, 0.0017, -0.125, 0.0219, 0.0039, -0.125, 0.0219, -0.0, -0.1228, 0.0201, -0.0, -0.1228, 0.0219, -0.0, -0.1207, 0.0201, -0.0, -0.1207, 0.0219, 0.0017, -0.1228, 0.0219, 0.0039, -0.1228, 0.0219, 0.0017, -0.1207, 0.0219, 0.0039, -0.1207, 0.0179, -0.0, -0.1358, 0.0157, -0.0, -0.1358, 0.0179, -0.0, -0.1337, 0.0157, -0.0, -0.1337, 0.0136, -0.0, -0.1358, 0.0114, -0.0, -0.1358, 0.0136, -0.0, -0.1337, 0.0114, -0.0, -0.1337, 0.0179, -0.0, -0.1315, 0.0157, -0.0, -0.1315, 0.0179, -0.0, -0.1293, 0.0157, -0.0, -0.1293, 0.0136, -0.0, -0.1315, 0.0114, -0.0, -0.1315, 0.0136, -0.0, -0.1293, 0.0114, -0.0, -0.1293, 0.0097, -0.0, -0.1358, 0.01, -0.0, -0.1337, 0.0097, 0.0017, -0.1358,
0.0097, 0.0039, -0.1358, 0.01, 0.0017, -0.1337, 0.01, 0.0039, -0.1337, 0.0101, -0.0, -0.1315, 0.0101, -0.0, -0.1293, 0.0101, 0.0017, -0.1315, 0.0101, 0.0039, -0.1315, 0.0101, 0.0017, -0.1293, 0.0101, 0.0039, -0.1293, 0.0179, -0.0, -0.1272, 0.0157, -0.0, -0.1272, 0.0179, -0.0, -0.125, 0.0157, -0.0, -0.125, 0.0136, -0.0, -0.1272, 0.0114, -0.0, -0.1272, 0.0136, -0.0, -0.125, 0.0114, -0.0, -0.125, 0.0179, -0.0, -0.1228, 0.0157, -0.0, -0.1228, 0.0179, -0.0, -0.1207, 0.0157, -0.0, -0.1207, 0.0136, -0.0, -0.1228, 0.0114, -0.0, -0.1228, 0.0136, -0.0, -0.1207, 0.0114, -0.0, -0.1207, 0.0101, -0.0, -0.1272, 0.0101, -0.0, -0.125, 0.0101, 0.0017, -0.1272, 0.0101, 0.0039, -0.1272, 0.0101, 0.0017, -0.125,
0.0101, 0.0039, -0.125, 0.0101, -0.0, -0.1228, 0.0101, -0.0, -0.1207, 0.0101, 0.0017, -0.1228, 0.0101, 0.0039, -0.1228, 0.0101, 0.0017, -0.1207, 0.0101, 0.0039, -0.1207, 0.0219, -0.0, -0.1185, 0.0201, -0.0, -0.1185, 0.0219, -0.0, -0.1163, 0.0201, -0.0, -0.1163, 0.0219, 0.0017, -0.1185, 0.0219, 0.0039, -0.1185, 0.0219, 0.0017, -0.1163, 0.0219, 0.0039, -0.1163, 0.0219, -0.0, -0.1141, 0.0201, -0.0, -0.1141, 0.0219, -0.0, -0.112, 0.0201, -0.0, -0.112, 0.0219, 0.0017, -0.1141, 0.0219, 0.0039, -0.1141, 0.0219, 0.0017, -0.112, 0.0219, 0.0039, -0.112, 0.0219, -0.0, -0.1098, 0.0201, -0.0, -0.1098, 0.0219, -0.0, -0.1076, 0.0201, -0.0, -0.1076, 0.0219, 0.0017, -0.1098, 0.0219, 0.0039, -0.1098, 0.0219, 0.0017, -0.1076,
0.0219, 0.0039, -0.1076, 0.0219, -0.0, -0.1055, 0.0201, -0.0, -0.1055, 0.0219, -0.0, -0.1033, 0.0201, -0.0, -0.1033, 0.0219, 0.0017, -0.1055, 0.0219, 0.0039, -0.1055, 0.0219, 0.0017, -0.1033, 0.0219, 0.0039, -0.1033, 0.0179, -0.0, -0.1185, 0.0157, -0.0, -0.1185, 0.0179, -0.0, -0.1163, 0.0157, -0.0, -0.1163, 0.0136, -0.0, -0.1185, 0.0114, -0.0, -0.1185, 0.0136, -0.0, -0.1163, 0.0114, -0.0, -0.1163, 0.0179, -0.0, -0.1141, 0.0157, -0.0, -0.1141, 0.0179, -0.0, -0.112, 0.0157, -0.0, -0.112, 0.0136, -0.0, -0.1141, 0.0114, -0.0, -0.1141, 0.0136, -0.0, -0.112, 0.0114, -0.0, -0.112, 0.0101, -0.0, -0.1185, 0.0101, -0.0, -0.1163, 0.0101, 0.0017, -0.1185, 0.0101, 0.0039, -0.1185, 0.0101, 0.0017, -0.1163,
0.0101, 0.0039, -0.1163, 0.0101, -0.0, -0.1141, 0.0101, -0.0, -0.112, 0.0101, 0.0017, -0.1141, 0.0101, 0.0039, -0.1141, 0.0101, 0.0017, -0.112, 0.0101, 0.0039, -0.112, 0.0179, -0.0, -0.1098, 0.0157, -0.0, -0.1098, 0.0179, -0.0, -0.1076, 0.0157, -0.0, -0.1076, 0.0136, -0.0, -0.1098, 0.0114, -0.0, -0.1098, 0.0136, -0.0, -0.1076, 0.0114, -0.0, -0.1076, 0.0179, -0.0, -0.1055, 0.0157, -0.0, -0.1055, 0.0179, -0.0, -0.1033, 0.0157, -0.0, -0.1033, 0.0136, -0.0, -0.1055, 0.0114, -0.0, -0.1055, 0.0136, -0.0, -0.1033, 0.0114, -0.0, -0.1033, 0.0101, -0.0, -0.1098, 0.0101, -0.0, -0.1076, 0.0101, 0.0017, -0.1098, 0.0101, 0.0039, -0.1098, 0.0101, 0.0017, -0.1076, 0.0101, 0.0039, -0.1076, 0.0101, -0.0, -0.1055,
0.0101, -0.0, -0.1033, 0.0101, 0.0017, -0.1055, 0.0101, 0.0039, -0.1055, 0.0101, 0.0017, -0.1033, 0.0101, 0.0039, -0.1033, 0.065, -0.0, -0.1017, 0.0642, -0.0, -0.1006, 0.0631, -0.0, -0.0994, 0.065, 0.0017, -0.1016, 0.0642, 0.0017, -0.1005, 0.065, 0.0039, -0.1016, 0.0642, 0.0039, -0.1005, 0.0631, 0.0017, -0.0994, 0.0631, 0.0039, -0.0994, 0.0604, -0.0, -0.102, 0.059, -0.0, -0.1007, 0.0618, -0.0, -0.0983, 0.0591, -0.0, -0.099, 0.0578, -0.0, -0.1002, 0.0569, -0.0, -0.0998, 0.0548, -0.0, -0.0995, 0.0604, 0.0017, -0.102, 0.0591, 0.0017, -0.1007, 0.0604, 0.0039, -0.102, 0.0591, 0.0039, -0.1007, 0.0618, 0.0017, -0.0983, 0.0618, 0.0039, -0.0983, 0.0577, 0.0017, -0.1002, 0.0577, 0.0039, -0.1002, 0.0566, 0.0017, -0.0997,
0.0548, 0.0017, -0.0995, 0.0566, 0.0039, -0.0997, 0.0548, 0.0039, -0.0995, 0.0606, -0.0, -0.0975, 0.0594, -0.0, -0.0968, 0.0571, -0.0, -0.0962, 0.0548, -0.0, -0.0961, 0.0607, 0.0017, -0.0976, 0.0592, 0.0017, -0.0968, 0.0607, 0.0039, -0.0976, 0.0592, 0.0039, -0.0968, 0.057, 0.0017, -0.0962, 0.0548, 0.0017, -0.0961, 0.057, 0.0039, -0.0962, 0.0548, 0.0039, -0.0961, 0.0518, -0.0, -0.1002, 0.0506, -0.0, -0.1006, 0.0527, -0.0, -0.0998, 0.0505, -0.0, -0.099, 0.049, -0.0, -0.1019, 0.0461, -0.0, -0.1011, 0.0474, -0.0, -0.0978, 0.0457, -0.0, -0.099, 0.0518, 0.0017, -0.1002, 0.0504, 0.0017, -0.1008, 0.0518, 0.0039, -0.1002, 0.0504, 0.0039, -0.1008, 0.0529, 0.0017, -0.0997, 0.0529, 0.0039, -0.0997, 0.049, 0.0017, -0.1019,
0.049, 0.0039, -0.1019, 0.0473, 0.0017, -0.0979, 0.046, 0.0017, -0.0988, 0.0473, 0.0039, -0.0979, 0.046, 0.0039, -0.0988, 0.0526, -0.0, -0.0961, 0.0503, -0.0, -0.0965, 0.0485, -0.0, -0.0971, 0.0526, 0.0017, -0.0961, 0.0504, 0.0017, -0.0964, 0.0526, 0.0039, -0.0961, 0.0504, 0.0039, -0.0964, 0.0483, 0.0017, -0.0972, 0.0483, 0.0039, -0.0972, 0.0442, -0.0, -0.1009, 0.045, -0.0, -0.1, 0.0441, 0.0017, -0.101, 0.0441, 0.0039, -0.101, 0.045, 0.0017, -0.1, 0.045, 0.0039, -0.1, 0.0587, -0.0, -0.0766, 0.057, -0.0, -0.0771, 0.0548, -0.0, -0.0774, 0.0589, 0.0017, -0.0766, 0.0589, 0.0039, -0.0766, 0.057, 0.0017, -0.0771, 0.0548, 0.0017, -0.0774, 0.057, 0.0039, -0.0771, 0.0548, 0.0039, -0.0774, 0.0626, -0.0, -0.0742,
0.0647, -0.0, -0.072, 0.0638, -0.0, -0.0732, 0.0627, 0.0017, -0.0742, 0.0627, 0.0039, -0.0742, 0.0647, 0.0017, -0.072, 0.0638, 0.0017, -0.0731, 0.0647, 0.0039, -0.072, 0.0638, 0.0039, -0.0731, 0.0654, -0.0, -0.071, 0.0635, -0.0, -0.0707, 0.0663, -0.0, -0.0688, 0.0623, -0.0, -0.0681, 0.0655, 0.0017, -0.0708, 0.0655, 0.0039, -0.0708, 0.0664, 0.0017, -0.0686, 0.0624, 0.0017, -0.0676, 0.0664, 0.0039, -0.0686, 0.0624, 0.0039, -0.0676, 0.0617, -0.0, -0.0751, 0.0597, -0.0, -0.0762, 0.0613, -0.0, -0.0729, 0.0588, -0.0, -0.0722, 0.057, -0.0, -0.0751, 0.0548, -0.0, -0.0751, 0.0568, -0.0, -0.073, 0.0548, -0.0, -0.0733, 0.0614, 0.0017, -0.0753, 0.06, 0.0017, -0.0761, 0.0614, 0.0039, -0.0753, 0.06, 0.0039, -0.0761,
0.0587, 0.0017, -0.0723, 0.0587, 0.0039, -0.0723, 0.057, 0.0017, -0.0729, 0.0548, 0.0017, -0.0733, 0.057, 0.0039, -0.0729, 0.0548, 0.0039, -0.0733, 0.061, -0.0, -0.0705, 0.0598, -0.0, -0.0714, 0.0619, -0.0, -0.0689, 0.0608, 0.0017, -0.0707, 0.0598, 0.0017, -0.0715, 0.0608, 0.0039, -0.0707, 0.0598, 0.0039, -0.0715, 0.062, 0.0017, -0.0687, 0.062, 0.0039, -0.0687, 0.0526, -0.0, -0.0773, 0.0503, -0.0, -0.0769, 0.0491, -0.0, -0.0764, 0.0526, 0.0017, -0.0773, 0.0504, 0.0017, -0.0769, 0.0526, 0.0039, -0.0773, 0.0504, 0.0039, -0.0769, 0.049, 0.0017, -0.0763, 0.049, 0.0039, -0.0763, 0.0526, -0.0, -0.0751, 0.0505, -0.0, -0.0751, 0.0527, -0.0, -0.0732, 0.0505, -0.0, -0.0725, 0.048, -0.0, -0.0758, 0.0464, -0.0, -0.0748,
0.0483, -0.0, -0.0729, 0.0453, -0.0, -0.0736, 0.0526, 0.0017, -0.0732, 0.0508, 0.0017, -0.0727, 0.0526, 0.0039, -0.0732, 0.0508, 0.0039, -0.0727, 0.0479, 0.0017, -0.0758, 0.0464, 0.0017, -0.0747, 0.0479, 0.0039, -0.0758, 0.0464, 0.0039, -0.0747, 0.0453, 0.0017, -0.0736, 0.0453, 0.0039, -0.0736, 0.0495, -0.0, -0.0717, 0.0483, -0.0, -0.0708, 0.0461, -0.0, -0.0707, 0.0474, -0.0, -0.0694, 0.0468, -0.0, -0.0684, 0.0495, 0.0017, -0.0717, 0.0495, 0.0039, -0.0717, 0.0482, 0.0017, -0.0707, 0.0482, 0.0039, -0.0707, 0.0474, 0.0017, -0.0694, 0.0468, 0.0017, -0.0683, 0.0474, 0.0039, -0.0694, 0.0468, 0.0039, -0.0683, 0.0445, -0.0, -0.0725, 0.0445, 0.0017, -0.0726, 0.0445, 0.0039, -0.0726, 0.0437, -0.0, -0.0711, 0.0429, -0.0, -0.0689,
0.0427, -0.0, -0.0679, 0.0435, 0.0017, -0.0708, 0.0435, 0.0039, -0.0708, 0.043, 0.0017, -0.0692, 0.0427, 0.0017, -0.0681, 0.043, 0.0039, -0.0692, 0.0427, 0.0039, -0.0681, 0.0219, -0.0, -0.1011, 0.0201, -0.0, -0.1011, 0.0219, -0.0, -0.099, 0.0201, -0.0, -0.099, 0.0219, 0.0017, -0.1011, 0.0219, 0.0039, -0.1011, 0.0219, 0.0017, -0.099, 0.0219, 0.0039, -0.099, 0.0219, -0.0, -0.0968, 0.0201, -0.0, -0.0968, 0.0219, -0.0, -0.0946, 0.0201, -0.0, -0.0946, 0.0219, 0.0017, -0.0968, 0.0219, 0.0039, -0.0968, 0.0219, 0.0017, -0.0946, 0.0219, 0.0039, -0.0946, 0.0219, -0.0, -0.0924, 0.0201, -0.0, -0.0924, 0.0219, -0.0, -0.0902, 0.0201, -0.0, -0.0902, 0.0219, 0.0017, -0.0924, 0.0219, 0.0039, -0.0924, 0.0219, 0.0017, -0.0902,
0.0201, 0.0017, -0.0902, 0.0219, 0.0039, -0.0902, 0.0201, 0.0039, -0.0902, 0.0179, -0.0, -0.1011, 0.0157, -0.0, -0.1011, 0.0179, -0.0, -0.099, 0.0157, -0.0, -0.099, 0.0136, -0.0, -0.1011, 0.0114, -0.0, -0.1011, 0.0136, -0.0, -0.099, 0.0114, -0.0, -0.099, 0.0179, -0.0, -0.0968, 0.0157, -0.0, -0.0968, 0.0179, -0.0, -0.0946, 0.0157, -0.0, -0.0946, 0.0136, -0.0, -0.0968, 0.0114, -0.0, -0.0968, 0.0136, -0.0, -0.0946, 0.0114, -0.0, -0.0946, 0.0101, -0.0, -0.1011, 0.0101, -0.0, -0.099, 0.0101, 0.0017, -0.1011, 0.0101, 0.0039, -0.1011, 0.0101, 0.0017, -0.099, 0.0101, 0.0039, -0.099, 0.0101, -0.0, -0.0968, 0.0101, -0.0, -0.0946, 0.0101, 0.0017, -0.0968, 0.0101, 0.0039, -0.0968, 0.0101, 0.0017, -0.0946,
0.0101, 0.0039, -0.0946, 0.0179, -0.0, -0.0924, 0.0157, -0.0, -0.0924, 0.0179, -0.0, -0.0902, 0.0157, -0.0, -0.0902, 0.0136, -0.0, -0.0924, 0.0114, -0.0, -0.0924, 0.0136, -0.0, -0.0902, 0.0114, -0.0, -0.0902, 0.0179, 0.0017, -0.0902, 0.0157, 0.0017, -0.0902, 0.0179, 0.0039, -0.0902, 0.0157, 0.0039, -0.0902, 0.0136, 0.0017, -0.0902, 0.0114, 0.0017, -0.0902, 0.0136, 0.0039, -0.0902, 0.0114, 0.0039, -0.0902, 0.0101, -0.0, -0.0924, 0.0101, -0.0, -0.0902, 0.0101, 0.0017, -0.0924, 0.0101, 0.0039, -0.0924, 0.0101, 0.0017, -0.0902, 0.0101, 0.0039, -0.0902, 0.0219, -0.0, -0.0775, 0.0201, -0.0, -0.0775, 0.0219, 0.0017, -0.0775, 0.0201, 0.0017, -0.0775, 0.0219, 0.0039, -0.0775, 0.0201, 0.0039, -0.0775, 0.0219, -0.0, -0.0751,
0.0201, -0.0, -0.0751, 0.0219, -0.0, -0.0729, 0.0201, -0.0, -0.0729, 0.0219, 0.0017, -0.0751, 0.0219, 0.0039, -0.0751, 0.0219, 0.0017, -0.0729, 0.0219, 0.0039, -0.0729, 0.0219, -0.0, -0.0707, 0.0201, -0.0, -0.0707, 0.0219, -0.0, -0.0686, 0.0201, -0.0, -0.0686, 0.0219, 0.0017, -0.0707, 0.0219, 0.0039, -0.0707, 0.0219, 0.0017, -0.0686, 0.0219, 0.0039, -0.0686, 0.0179, -0.0, -0.0775, 0.0157, -0.0, -0.0775, 0.0136, -0.0, -0.0775, 0.0114, -0.0, -0.0775, 0.0179, 0.0017, -0.0775, 0.0157, 0.0017, -0.0775, 0.0179, 0.0039, -0.0775, 0.0157, 0.0039, -0.0775, 0.0136, 0.0017, -0.0775, 0.0114, 0.0017, -0.0775, 0.0136, 0.0039, -0.0775, 0.0114, 0.0039, -0.0775, 0.0092, -0.0, -0.0775, 0.0071, -0.0, -0.0775, 0.0049, -0.0, -0.0775,
0.0027, -0.0, -0.0775, 0.0092, 0.0017, -0.0775, 0.0071, 0.0017, -0.0775, 0.0092, 0.0039, -0.0775, 0.0071, 0.0039, -0.0775, 0.0049, 0.0017, -0.0775, 0.0027, 0.0017, -0.0775, 0.0049, 0.0039, -0.0775, 0.0027, 0.0039, -0.0775, 0.0179, -0.0, -0.0751, 0.0157, -0.0, -0.0751, 0.0179, -0.0, -0.0729, 0.0157, -0.0, -0.0729, 0.0136, -0.0, -0.0751, 0.0114, -0.0, -0.0751, 0.0136, -0.0, -0.0729, 0.0114, -0.0, -0.0729, 0.0179, -0.0, -0.0707, 0.0157, -0.0, -0.0707, 0.0179, -0.0, -0.0686, 0.0157, -0.0, -0.0686, 0.0136, -0.0, -0.0707, 0.0114, -0.0, -0.0707, 0.0136, -0.0, -0.0686, 0.0114, -0.0, -0.0686, 0.0092, -0.0, -0.0751, 0.0071, -0.0, -0.0751, 0.0092, -0.0, -0.0729, 0.0071, -0.0, -0.0729, 0.0049, -0.0, -0.0751,
0.0027, -0.0, -0.0751, 0.0049, -0.0, -0.0729, 0.0027, -0.0, -0.0729, 0.0092, -0.0, -0.0707, 0.0071, -0.0, -0.0707, 0.0092, -0.0, -0.0686, 0.0071, -0.0, -0.0686, 0.0049, -0.0, -0.0707, 0.0027, -0.0, -0.0707, 0.0049, -0.0, -0.0686, 0.0027, -0.0, -0.0686, 0.0669, -0.0, -0.0659, 0.067, -0.0, -0.0642, 0.0668, -0.0, -0.0667, 0.0628, -0.0, -0.0663, 0.0656, -0.0, -0.0642, 0.063, -0.0, -0.0642, 0.0669, 0.0017, -0.0661, 0.0669, 0.0039, -0.0661, 0.067, 0.0017, -0.0642, 0.067, 0.0039, -0.0642, 0.0667, 0.0017, -0.0672, 0.0627, 0.0017, -0.0664, 0.0667, 0.0039, -0.0672, 0.0627, 0.0039, -0.0664, 0.063, 0.0017, -0.0642, 0.063, 0.0039, -0.0642, 0.067, -0.0, -0.0621, 0.067, -0.0, -0.0599, 0.0656, -0.0, -0.0621,
0.063, -0.0, -0.062, 0.0656, -0.0, -0.0599, 0.0631, -0.0, -0.0599, 0.067, 0.0017, -0.0621, 0.067, 0.0039, -0.0621, 0.067, 0.0017, -0.0599, 0.067, 0.0039, -0.0599, 0.063, 0.0017, -0.0621, 0.063, 0.0039, -0.0621, 0.0631, 0.0017, -0.0599, 0.0631, 0.0039, -0.0599, 0.0669, -0.0, -0.0577, 0.0668, -0.0, -0.0566, 0.0656, -0.0, -0.0577, 0.063, -0.0, -0.0578, 0.0666, -0.0, -0.0555, 0.0625, -0.0, -0.0557, 0.0669, 0.0017, -0.0577, 0.0669, 0.0039, -0.0577, 0.0667, 0.0017, -0.0565, 0.0667, 0.0039, -0.0565, 0.063, 0.0017, -0.0577, 0.063, 0.0039, -0.0577, 0.0666, 0.0017, -0.0554, 0.0626, 0.0017, -0.0559, 0.0666, 0.0039, -0.0554, 0.0626, 0.0039, -0.0559, 0.0659, -0.0, -0.0531, 0.0635, -0.0, -0.0534, 0.0652, -0.0, -0.0515,
0.0646, -0.0, -0.0505, 0.066, 0.0017, -0.0534, 0.066, 0.0039, -0.0534, 0.0651, 0.0017, -0.0514, 0.0645, 0.0017, -0.0503, 0.0651, 0.0039, -0.0514, 0.0645, 0.0039, -0.0503, 0.0623, -0.0, -0.0547, 0.0623, 0.0017, -0.0548, 0.0623, 0.0039, -0.0548, 0.0617, -0.0, -0.0535, 0.0605, -0.0, -0.052, 0.059, -0.0, -0.0507, 0.058, -0.0, -0.0502, 0.0615, 0.0017, -0.0533, 0.0615, 0.0039, -0.0533, 0.0605, 0.0017, -0.052, 0.0592, 0.0017, -0.0509, 0.0605, 0.0039, -0.052, 0.0592, 0.0039, -0.0509, 0.058, 0.0017, -0.0502, 0.058, 0.0039, -0.0502, 0.0461, -0.0, -0.0663, 0.0456, -0.0, -0.0642, 0.0461, 0.0017, -0.0664, 0.0461, 0.0039, -0.0664, 0.0456, 0.0017, -0.0642, 0.0456, 0.0039, -0.0642, 0.0454, -0.0, -0.0621, 0.0455, -0.0, -0.0599,
0.0454, 0.0017, -0.0621, 0.0454, 0.0039, -0.0621, 0.0455, 0.0017, -0.0599, 0.0455, 0.0039, -0.0599, 0.0439, -0.0, -0.0664, 0.0425, -0.0, -0.0664, 0.0439, -0.0, -0.0642, 0.0425, -0.0, -0.0642, 0.0425, 0.0017, -0.0664, 0.0425, 0.0039, -0.0664, 0.0425, 0.0017, -0.0642, 0.0425, 0.0039, -0.0642, 0.0439, -0.0, -0.0621, 0.0425, -0.0, -0.0621, 0.0439, -0.0, -0.0599, 0.0425, -0.0, -0.0599, 0.0425, 0.0017, -0.0621, 0.0425, 0.0039, -0.0621, 0.0425, 0.0017, -0.0599, 0.0425, 0.0039, -0.0599, 0.0457, -0.0, -0.0578, 0.0463, -0.0, -0.0557, 0.0458, 0.0017, -0.0577, 0.0458, 0.0039, -0.0577, 0.0463, 0.0017, -0.0555, 0.0463, 0.0039, -0.0555, 0.0501, -0.0, -0.0507, 0.0477, -0.0, -0.0528, 0.0471, -0.0, -0.0538, 0.0487, -0.0, -0.0517,
0.0461, -0.0, -0.0512, 0.0502, 0.0017, -0.0506, 0.0502, 0.0039, -0.0506, 0.0476, 0.0017, -0.0529, 0.047, 0.0017, -0.054, 0.0476, 0.0039, -0.0529, 0.047, 0.0039, -0.054, 0.0487, 0.0017, -0.0517, 0.0487, 0.0039, -0.0517, 0.0439, -0.0, -0.0577, 0.0425, -0.0, -0.0577, 0.0439, -0.0, -0.0555, 0.0426, -0.0, -0.0555, 0.0425, 0.0017, -0.0577, 0.0425, 0.0039, -0.0577, 0.0426, 0.0017, -0.0555, 0.0426, 0.0039, -0.0555, 0.0431, -0.0, -0.0533, 0.0428, -0.0, -0.0544, 0.0442, -0.0, -0.0509, 0.0431, 0.0017, -0.0532, 0.0428, 0.0017, -0.0544, 0.0431, 0.0039, -0.0532, 0.0428, 0.0039, -0.0544, 0.044, 0.0017, -0.0511, 0.044, 0.0039, -0.0511, 0.0636, -0.0, -0.049, 0.0635, 0.0017, -0.049, 0.0635, 0.0039, -0.049, 0.0622, -0.0, -0.048,
0.0591, -0.0, -0.049, 0.0611, -0.0, -0.0472, 0.0594, -0.0, -0.0465, 0.0569, -0.0, -0.0497, 0.0548, -0.0, -0.0492, 0.0571, -0.0, -0.0461, 0.0548, -0.0, -0.0461, 0.0623, 0.0017, -0.0481, 0.0623, 0.0039, -0.0481, 0.0613, 0.0017, -0.0473, 0.0596, 0.0017, -0.0465, 0.0613, 0.0039, -0.0473, 0.0596, 0.0039, -0.0465, 0.0569, 0.0017, -0.0497, 0.0547, 0.0017, -0.0492, 0.0569, 0.0039, -0.0497, 0.0547, 0.0039, -0.0492, 0.057, 0.0017, -0.0461, 0.0548, 0.0017, -0.0461, 0.057, 0.0039, -0.0461, 0.0548, 0.0039, -0.0461, 0.0528, -0.0, -0.0495, 0.051, -0.0, -0.0502, 0.0526, -0.0, -0.0461, 0.0503, -0.0, -0.0464, 0.0483, -0.0, -0.049, 0.0459, -0.0, -0.0489, 0.048, -0.0, -0.0471, 0.0471, -0.0, -0.0479, 0.0526, 0.0017, -0.0496,
0.0513, 0.0017, -0.05, 0.0526, 0.0039, -0.0496, 0.0513, 0.0039, -0.05, 0.0526, 0.0017, -0.0461, 0.0504, 0.0017, -0.0463, 0.0526, 0.0039, -0.0461, 0.0504, 0.0039, -0.0463, 0.0459, 0.0017, -0.0488, 0.0459, 0.0039, -0.0488, 0.0482, 0.0017, -0.047, 0.0471, 0.0017, -0.0479, 0.0482, 0.0039, -0.047, 0.0471, 0.0039, -0.0479, 0.0449, -0.0, -0.05, 0.045, 0.0017, -0.05, 0.045, 0.0039, -0.05, 0.0219, -0.0, -0.0667, 0.0201, -0.0, -0.0667, 0.0219, 0.0017, -0.0667, 0.0201, 0.0017, -0.0667, 0.0219, 0.0039, -0.0667, 0.0201, 0.0039, -0.0667, 0.0179, -0.0, -0.0667, 0.0157, -0.0, -0.0667, 0.0136, -0.0, -0.0667, 0.0114, -0.0, -0.0667, 0.0179, 0.0017, -0.0667, 0.0157, 0.0017, -0.0667, 0.0179, 0.0039, -0.0667, 0.0157, 0.0039, -0.0667,
0.0136, 0.0017, -0.0667, 0.0114, 0.0017, -0.0667, 0.0136, 0.0039, -0.0667, 0.0114, 0.0039, -0.0667, 0.0092, -0.0, -0.0667, 0.0071, -0.0, -0.0667, 0.0049, -0.0, -0.0667, 0.0027, -0.0, -0.0667, 0.0092, 0.0017, -0.0667, 0.0071, 0.0017, -0.0667, 0.0092, 0.0039, -0.0667, 0.0071, 0.0039, -0.0667, 0.0049, 0.0017, -0.0667, 0.0027, 0.0017, -0.0667, 0.0049, 0.0039, -0.0667, 0.0027, 0.0039, -0.0667, 0.0219, -0.0, -0.05, 0.0201, -0.0, -0.05, 0.0219, -0.0, -0.0469, 0.0201, -0.0, -0.0469, 0.0219, 0.0017, -0.05, 0.0201, 0.0017, -0.05, 0.0219, 0.0039, -0.05, 0.0201, 0.0039, -0.05, 0.0219, 0.0017, -0.0469, 0.0219, 0.0039, -0.0469, 0.0219, -0.0, -0.0447, 0.0201, -0.0, -0.0447, 0.0219, -0.0, -0.0425, 0.0201, -0.0, -0.0425,
0.0219, 0.0017, -0.0447, 0.0219, 0.0039, -0.0447, 0.0219, 0.0017, -0.0425, 0.0219, 0.0039, -0.0425, 0.0219, -0.0, -0.0404, 0.0201, -0.0, -0.0404, 0.0219, -0.0, -0.0382, 0.0201, -0.0, -0.0382, 0.0219, 0.0017, -0.0404, 0.0219, 0.0039, -0.0404, 0.0219, 0.0017, -0.0382, 0.0201, 0.0017, -0.0382, 0.0219, 0.0039, -0.0382, 0.0201, 0.0039, -0.0382, 0.0179, -0.0, -0.05, 0.0157, -0.0, -0.05, 0.0179, -0.0, -0.0469, 0.0157, -0.0, -0.0469, 0.0136, -0.0, -0.05, 0.0114, -0.0, -0.05, 0.0136, -0.0, -0.0469, 0.0114, -0.0, -0.0469, 0.0179, 0.0017, -0.05, 0.0157, 0.0017, -0.05, 0.0179, 0.0039, -0.05, 0.0157, 0.0039, -0.05, 0.0136, 0.0017, -0.05, 0.0114, 0.0017, -0.05, 0.0136, 0.0039, -0.05, 0.0114, 0.0039, -0.05,
0.0179, -0.0, -0.0447, 0.0157, -0.0, -0.0447, 0.0179, -0.0, -0.0425, 0.0157, -0.0, -0.0425, 0.0136, -0.0, -0.0447, 0.0114, -0.0, -0.0447, 0.0136, -0.0, -0.0425, 0.0114, -0.0, -0.0425, 0.0092, -0.0, -0.05, 0.0071, -0.0, -0.05, 0.0092, -0.0, -0.0469, 0.0071, -0.0, -0.0469, 0.0049, -0.0, -0.05, 0.0027, -0.0, -0.05, 0.0049, -0.0, -0.0469, 0.0027, -0.0, -0.0469, 0.0092, 0.0017, -0.05, 0.0071, 0.0017, -0.05, 0.0092, 0.0039, -0.05, 0.0071, 0.0039, -0.05, 0.0049, 0.0017, -0.05, 0.0027, 0.0017, -0.05, 0.0049, 0.0039, -0.05, 0.0027, 0.0039, -0.05, 0.0092, -0.0, -0.0447, 0.0071, -0.0, -0.0447, 0.0092, -0.0, -0.0425, 0.0071, -0.0, -0.0425, 0.0049, -0.0, -0.0447, 0.0027, -0.0, -0.0447,
0.0049, -0.0, -0.0425, 0.0027, -0.0, -0.0425, 0.0179, -0.0, -0.0404, 0.0157, -0.0, -0.0404, 0.0179, -0.0, -0.0382, 0.0157, -0.0, -0.0382, 0.0136, -0.0, -0.0404, 0.0114, -0.0, -0.0404, 0.0136, -0.0, -0.0382, 0.0114, -0.0, -0.0382, 0.0179, 0.0017, -0.0382, 0.0157, 0.0017, -0.0382, 0.0179, 0.0039, -0.0382, 0.0157, 0.0039, -0.0382, 0.0136, 0.0017, -0.0382, 0.0114, 0.0017, -0.0382, 0.0136, 0.0039, -0.0382, 0.0114, 0.0039, -0.0382, 0.0092, -0.0, -0.0404, 0.0071, -0.0, -0.0404, 0.0092, -0.0, -0.0382, 0.0071, -0.0, -0.0382, 0.0049, -0.0, -0.0404, 0.0027, -0.0, -0.0404, 0.0049, -0.0, -0.0382, 0.0027, -0.0, -0.0382, 0.0092, 0.0017, -0.0382, 0.0071, 0.0017, -0.0382, 0.0092, 0.0039, -0.0382, 0.0071, 0.0039, -0.0382,
0.0049, 0.0017, -0.0382, 0.0027, 0.0017, -0.0382, 0.0049, 0.0039, -0.0382, 0.0027, 0.0039, -0.0382, 0.067, -0.0, -0.0265, 0.067, -0.0, -0.0252, 0.0656, -0.0, -0.0265, 0.0635, -0.0, -0.0265, 0.0656, -0.0, -0.0252, 0.0635, -0.0, -0.0252, 0.067, 0.0017, -0.0265, 0.067, 0.0039, -0.0265, 0.067, 0.0017, -0.0252, 0.067, 0.0039, -0.0252, 0.0656, 0.0017, -0.0265, 0.0635, 0.0017, -0.0265, 0.0656, 0.0039, -0.0265, 0.0635, 0.0039, -0.0265, 0.0613, -0.0, -0.0265, 0.0591, -0.0, -0.0265, 0.0613, -0.0, -0.0252, 0.0591, -0.0, -0.0252, 0.057, -0.0, -0.0265, 0.0548, -0.0, -0.0265, 0.057, -0.0, -0.0252, 0.0548, -0.0, -0.0252, 0.0613, 0.0017, -0.0265, 0.0591, 0.0017, -0.0265, 0.0613, 0.0039, -0.0265, 0.0591, 0.0039, -0.0265,
0.057, 0.0017, -0.0265, 0.0548, 0.0017, -0.0265, 0.057, 0.0039, -0.0265, 0.0548, 0.0039, -0.0265, 0.067, -0.0, -0.0225, 0.0656, -0.0, -0.0225, 0.0635, -0.0, -0.0225, 0.067, 0.0017, -0.0225, 0.067, 0.0039, -0.0225, 0.0656, 0.0017, -0.0225, 0.0635, 0.0017, -0.0225, 0.0656, 0.0039, -0.0225, 0.0635, 0.0039, -0.0225, 0.0613, -0.0, -0.0225, 0.0591, -0.0, -0.0225, 0.057, -0.0, -0.0225, 0.0548, -0.0, -0.0225, 0.0613, 0.0017, -0.0225, 0.0591, 0.0017, -0.0225, 0.0613, 0.0039, -0.0225, 0.0591, 0.0039, -0.0225, 0.057, 0.0017, -0.0225, 0.0548, 0.0017, -0.0225, 0.057, 0.0039, -0.0225, 0.0548, 0.0039, -0.0225, 0.054, -0.0, -0.0179, 0.0562, -0.0, -0.0157, 0.0551, -0.0, -0.0168, 0.054, 0.0017, -0.0179, 0.054, 0.0039, -0.0179,
0.0562, 0.0017, -0.0157, 0.0551, 0.0017, -0.0168, 0.0562, 0.0039, -0.0157, 0.0551, 0.0039, -0.0168, 0.0526, -0.0, -0.0265, 0.0505, -0.0, -0.0265, 0.0526, -0.0, -0.0252, 0.0505, -0.0, -0.0252, 0.0483, -0.0, -0.0265, 0.0468, -0.0, -0.0264, 0.0483, -0.0, -0.0252, 0.0459, -0.0, -0.0262, 0.0526, 0.0017, -0.0265, 0.0505, 0.0017, -0.0265, 0.0526, 0.0039, -0.0265, 0.0505, 0.0039, -0.0265, 0.0483, 0.0017, -0.0265, 0.0466, 0.0017, -0.0264, 0.0483, 0.0039, -0.0265, 0.0466, 0.0039, -0.0264, 0.0455, 0.0017, -0.0262, 0.0455, 0.0039, -0.0262, 0.0437, -0.0, -0.0255, 0.0437, 0.0017, -0.0255, 0.0437, 0.0039, -0.0255, 0.0526, -0.0, -0.0225, 0.0503, -0.0, -0.0225, 0.052, -0.0, -0.0202, 0.0511, -0.0, -0.0213, 0.0483, -0.0, -0.023,
0.0461, -0.0, -0.023, 0.0483, -0.0, -0.0208, 0.0452, -0.0, -0.0201, 0.0526, 0.0017, -0.0225, 0.0503, 0.0017, -0.0225, 0.0526, 0.0039, -0.0225, 0.0503, 0.0039, -0.0225, 0.052, 0.0017, -0.0202, 0.0511, 0.0017, -0.0213, 0.052, 0.0039, -0.0202, 0.0511, 0.0039, -0.0213, 0.0452, 0.0017, -0.02, 0.0452, 0.0039, -0.02, 0.053, -0.0, -0.019, 0.0505, -0.0, -0.0187, 0.0526, -0.0, -0.0165, 0.0495, -0.0, -0.0155, 0.0473, -0.0, -0.0177, 0.0463, -0.0, -0.0188, 0.0484, -0.0, -0.0166, 0.053, 0.0017, -0.019, 0.053, 0.0039, -0.019, 0.0495, 0.0017, -0.0155, 0.0495, 0.0039, -0.0155, 0.0473, 0.0017, -0.0177, 0.0463, 0.0017, -0.0188, 0.0473, 0.0039, -0.0177, 0.0463, 0.0039, -0.0188, 0.0484, 0.0017, -0.0166, 0.0484, 0.0039, -0.0166,
0.0433, -0.0, -0.0231, 0.0443, -0.0, -0.0211, 0.0433, 0.0017, -0.0231, 0.0433, 0.0039, -0.0231, 0.0444, 0.0017, -0.021, 0.0444, 0.0039, -0.021, 0.0627, -0.0, -0.0092, 0.0647, -0.0, -0.0069, 0.0638, -0.0, -0.008, 0.0627, 0.0017, -0.0092, 0.0627, 0.0039, -0.0092, 0.0647, 0.0017, -0.0069, 0.0638, 0.0017, -0.008, 0.0647, 0.0039, -0.0069, 0.0638, 0.0039, -0.008, 0.0584, -0.0, -0.0135, 0.0605, -0.0, -0.0114, 0.0594, -0.0, -0.0124, 0.0573, -0.0, -0.0146, 0.0548, -0.0, -0.0143, 0.057, -0.0, -0.0121, 0.0538, -0.0, -0.0112, 0.0584, 0.0017, -0.0135, 0.0584, 0.0039, -0.0135, 0.0605, 0.0017, -0.0114, 0.0594, 0.0017, -0.0124, 0.0605, 0.0039, -0.0114, 0.0594, 0.0039, -0.0124, 0.0573, 0.0017, -0.0146, 0.0573, 0.0039, -0.0146,
0.0538, 0.0017, -0.0112, 0.0538, 0.0039, -0.0112, 0.0616, -0.0, -0.0103, 0.0591, -0.0, -0.01, 0.0613, -0.0, -0.0078, 0.0582, -0.0, -0.0068, 0.056, -0.0, -0.009, 0.0549, -0.0, -0.0101, 0.0571, -0.0, -0.0079, 0.0616, 0.0017, -0.0103, 0.0616, 0.0039, -0.0103, 0.0582, 0.0017, -0.0068, 0.0582, 0.0039, -0.0068, 0.056, 0.0017, -0.009, 0.0549, 0.0017, -0.0101, 0.056, 0.0039, -0.009, 0.0549, 0.0039, -0.0101, 0.0571, 0.0017, -0.0079, 0.0571, 0.0039, -0.0079, 0.0668, -0.0, -0.0026, 0.0655, -0.0, -0.0059, 0.0635, -0.0, -0.0056, 0.0666, -0.0, -0.0037, 0.0635, -0.0, -0.0035, 0.0668, 0.0017, -0.0028, 0.0668, 0.0039, -0.0028, 0.0656, 0.0017, -0.0057, 0.0656, 0.0039, -0.0057, 0.0666, 0.0017, -0.0039, 0.0666, 0.0039, -0.0039,
0.067, -0.0, -0.002, 0.0656, -0.0, -0.002, 0.0635, -0.0, -0.002, 0.067, 0.0017, -0.002, 0.067, 0.0039, -0.002, 0.0656, 0.0017, -0.002, 0.0635, 0.0017, -0.002, 0.0656, 0.0039, -0.002, 0.0635, 0.0039, -0.002, 0.0613, -0.0, -0.0056, 0.0591, -0.0, -0.0059, 0.0613, -0.0, -0.0035, 0.0591, -0.0, -0.0035, 0.057, -0.0, -0.0059, 0.0548, -0.0, -0.0059, 0.057, -0.0, -0.0035, 0.0548, -0.0, -0.0035, 0.0591, 0.0017, -0.0059, 0.0591, 0.0039, -0.0059, 0.057, 0.0017, -0.0059, 0.0548, 0.0017, -0.0059, 0.057, 0.0039, -0.0059, 0.0548, 0.0039, -0.0059, 0.0613, -0.0, -0.002, 0.0591, -0.0, -0.002, 0.057, -0.0, -0.002, 0.0548, -0.0, -0.002, 0.0613, 0.0017, -0.002, 0.0591, 0.0017, -0.002, 0.0613, 0.0039, -0.002,
0.0591, 0.0039, -0.002, 0.057, 0.0017, -0.002, 0.0548, 0.0017, -0.002, 0.057, 0.0039, -0.002, 0.0548, 0.0039, -0.002, 0.0517, -0.0, -0.0134, 0.0506, -0.0, -0.0144, 0.0528, -0.0, -0.0123, 0.0517, 0.0017, -0.0134, 0.0506, 0.0017, -0.0144, 0.0517, 0.0039, -0.0134, 0.0506, 0.0039, -0.0144, 0.0528, 0.0017, -0.0123, 0.0528, 0.0039, -0.0123, 0.0526, -0.0, -0.0059, 0.0505, -0.0, -0.0059, 0.0526, -0.0, -0.0035, 0.0505, -0.0, -0.0035, 0.0483, -0.0, -0.0059, 0.0461, -0.0, -0.0059, 0.0483, -0.0, -0.0035, 0.0461, -0.0, -0.0035, 0.0526, 0.0017, -0.0059, 0.0505, 0.0017, -0.0059, 0.0526, 0.0039, -0.0059, 0.0505, 0.0039, -0.0059, 0.0483, 0.0017, -0.0059, 0.0461, 0.0017, -0.0059, 0.0483, 0.0039, -0.0059, 0.0461, 0.0039, -0.0059,
0.0526, -0.0, -0.002, 0.0505, -0.0, -0.002, 0.0483, -0.0, -0.002, 0.0461, -0.0, -0.002, 0.0526, 0.0017, -0.002, 0.0505, 0.0017, -0.002, 0.0526, 0.0039, -0.002, 0.0505, 0.0039, -0.002, 0.0483, 0.0017, -0.002, 0.0461, 0.0017, -0.002, 0.0483, 0.0039, -0.002, 0.0461, 0.0039, -0.002, 0.0439, -0.0, -0.0059, 0.0425, -0.0, -0.0059, 0.0439, -0.0, -0.0035, 0.0425, -0.0, -0.0035, 0.0439, 0.0017, -0.0059, 0.0425, 0.0017, -0.0059, 0.0439, 0.0039, -0.0059, 0.0425, 0.0039, -0.0059, 0.0425, 0.0017, -0.0035, 0.0425, 0.0039, -0.0035, 0.0439, -0.0, -0.002, 0.0425, -0.0, -0.002, 0.0439, 0.0017, -0.002, 0.0425, 0.0017, -0.002, 0.0439, 0.0039, -0.002, 0.0425, 0.0039, -0.002, 0.0004, -0.0, -0.4807, -0.0013, -0.0, -0.4802,
0.0005, 0.0017, -0.4807, -0.0016, 0.0017, -0.4801, 0.0005, 0.0039, -0.4807, -0.0016, 0.0039, -0.4801, 0.0005, -0.0, -0.4787, -0.0021, -0.0, -0.48, 0.0005, -0.0, -0.4766, -0.0016, -0.0, -0.4766, -0.004, -0.0, -0.4795, -0.0061, -0.0, -0.4789, -0.0038, -0.0, -0.4766, -0.006, -0.0, -0.4766, -0.0027, 0.0017, -0.4798, -0.0027, 0.0039, -0.4798, -0.0038, 0.0017, -0.4795, -0.006, 0.0017, -0.4789, -0.0038, 0.0039, -0.4795, -0.006, 0.0039, -0.4789, -0.0082, -0.0, -0.4784, -0.0099, -0.0, -0.478, -0.0081, -0.0, -0.4766, -0.0107, -0.0, -0.4777, -0.0127, -0.0, -0.4772, -0.0148, -0.0, -0.4765, -0.0081, 0.0017, -0.4784, -0.0102, 0.0017, -0.4779, -0.0081, 0.0039, -0.4784, -0.0102, 0.0039, -0.4779, -0.0112, 0.0017, -0.4776, -0.0112, 0.0039, -0.4776,
-0.0125, 0.0017, -0.4772, -0.0147, 0.0017, -0.4765, -0.0125, 0.0039, -0.4772, -0.0147, 0.0039, -0.4765, 0.0005, -0.0, -0.4744, -0.0016, -0.0, -0.4744, 0.0005, -0.0, -0.4722, -0.0016, -0.0, -0.4722, -0.0038, -0.0, -0.4744, -0.006, -0.0, -0.4744, -0.0038, -0.0, -0.4722, -0.006, -0.0, -0.4722, 0.0005, -0.0, -0.47, -0.0016, -0.0, -0.47, 0.0005, -0.0, -0.4679, -0.0016, -0.0, -0.4679, -0.0038, -0.0, -0.47, -0.006, -0.0, -0.47, -0.0038, -0.0, -0.4679, -0.006, -0.0, -0.4679, -0.0081, -0.0, -0.4744, -0.0103, -0.0, -0.4744, -0.0081, -0.0, -0.4722, -0.0103, -0.0, -0.4722, -0.0125, -0.0, -0.4744, -0.0146, -0.0, -0.4744, -0.0125, -0.0, -0.4722, -0.0146, -0.0, -0.4722, -0.0081, -0.0, -0.47, -0.0103, -0.0, -0.47,
-0.0081, -0.0, -0.4679, -0.0103, -0.0, -0.4679, -0.0125, -0.0, -0.47, -0.0146, -0.0, -0.47, -0.0125, -0.0, -0.4679, -0.0146, -0.0, -0.4679, -0.0164, -0.0, -0.4759, -0.0166, 0.0017, -0.4758, -0.0166, 0.0039, -0.4758, -0.0173, -0.0, -0.4755, -0.0192, -0.0, -0.4749, -0.0168, -0.0, -0.4722, -0.019, -0.0, -0.4722, -0.0213, -0.0, -0.4742, -0.0228, -0.0, -0.4736, -0.0212, -0.0, -0.4722, -0.0237, -0.0, -0.4733, -0.0177, 0.0017, -0.4754, -0.019, 0.0017, -0.4749, -0.0177, 0.0039, -0.4754, -0.019, 0.0039, -0.4749, -0.0212, 0.0017, -0.4742, -0.0229, 0.0017, -0.4736, -0.0212, 0.0039, -0.4742, -0.0229, 0.0039, -0.4736, -0.024, 0.0017, -0.4732, -0.024, 0.0039, -0.4732, -0.0168, -0.0, -0.47, -0.019, -0.0, -0.47, -0.0168, -0.0, -0.4679,
-0.019, -0.0, -0.4679, -0.0212, -0.0, -0.47, -0.0233, -0.0, -0.47, -0.0212, -0.0, -0.4679, -0.0233, -0.0, -0.4679, -0.0258, -0.0, -0.4724, -0.0278, -0.0, -0.4716, -0.0288, -0.0, -0.4711, -0.0255, 0.0017, -0.4726, -0.0277, 0.0017, -0.4716, -0.0255, 0.0039, -0.4726, -0.0277, 0.0039, -0.4716, -0.0288, 0.0017, -0.4711, -0.0288, 0.0039, -0.4711, -0.0255, -0.0, -0.47, -0.0277, -0.0, -0.47, -0.0255, -0.0, -0.4679, -0.0277, -0.0, -0.4679, -0.0299, -0.0, -0.4707, -0.0322, -0.0, -0.4697, -0.0298, -0.0, -0.4679, -0.032, -0.0, -0.4679, -0.0299, 0.0017, -0.4707, -0.032, 0.0017, -0.4698, -0.0299, 0.0039, -0.4707, -0.032, 0.0039, -0.4698, 0.0005, -0.0, -0.4657, -0.0016, -0.0, -0.4657, 0.0008, -0.0, -0.4628, -0.0012, -0.0, -0.4623,
-0.0038, -0.0, -0.4657, -0.006, -0.0, -0.4657, -0.0038, -0.0, -0.4635, -0.006, -0.0, -0.4635, 0.0005, 0.0017, -0.4628, -0.0006, 0.0017, -0.4625, 0.0005, 0.0039, -0.4628, -0.0006, 0.0039, -0.4625, -0.002, -0.0, -0.4621, -0.0037, -0.0, -0.4616, -0.0057, -0.0, -0.4609, -0.0017, 0.0017, -0.4622, -0.0017, 0.0039, -0.4622, -0.0038, 0.0017, -0.4615, -0.006, 0.0017, -0.4608, -0.0038, 0.0039, -0.4615, -0.006, 0.0039, -0.4608, -0.0081, -0.0, -0.4657, -0.0103, -0.0, -0.4657, -0.0081, -0.0, -0.4635, -0.0103, -0.0, -0.4635, -0.0125, -0.0, -0.4657, -0.0146, -0.0, -0.4657, -0.0125, -0.0, -0.4635, -0.0146, -0.0, -0.4635, -0.0078, -0.0, -0.4603, -0.0103, -0.0, -0.4614, -0.0087, -0.0, -0.46, -0.0102, -0.0, -0.4595, -0.0125, -0.0, -0.4614,
-0.0146, -0.0, -0.4614, -0.0122, -0.0, -0.4588, -0.0142, -0.0, -0.4581, -0.0074, 0.0017, -0.4604, -0.0074, 0.0039, -0.4604, -0.0085, 0.0017, -0.4601, -0.0103, 0.0017, -0.4595, -0.0085, 0.0039, -0.4601, -0.0103, 0.0039, -0.4595, -0.0125, 0.0017, -0.4587, -0.0138, 0.0017, -0.4582, -0.0125, 0.0039, -0.4587, -0.0138, 0.0039, -0.4582, -0.0151, -0.0, -0.4577, -0.0149, 0.0017, -0.4578, -0.0149, 0.0039, -0.4578, -0.0168, -0.0, -0.4657, -0.019, -0.0, -0.4657, -0.0168, -0.0, -0.4635, -0.019, -0.0, -0.4635, -0.0212, -0.0, -0.4657, -0.0233, -0.0, -0.4657, -0.0212, -0.0, -0.4635, -0.0233, -0.0, -0.4635, -0.0168, -0.0, -0.4614, -0.019, -0.0, -0.4614, -0.0168, -0.0, -0.4592, -0.019, -0.0, -0.4592, -0.0212, -0.0, -0.4614, -0.0233, -0.0, -0.4614,
-0.0212, -0.0, -0.4592, -0.0233, -0.0, -0.4592, -0.0255, -0.0, -0.4657, -0.0277, -0.0, -0.4657, -0.0255, -0.0, -0.4635, -0.0277, -0.0, -0.4635, -0.0298, -0.0, -0.4657, -0.032, -0.0, -0.4657, -0.0298, -0.0, -0.4635, -0.032, -0.0, -0.4635, -0.0255, -0.0, -0.4614, -0.0277, -0.0, -0.4614, -0.0255, -0.0, -0.4592, -0.0277, -0.0, -0.4592, -0.0298, -0.0, -0.4614, -0.032, -0.0, -0.4614, -0.0298, -0.0, -0.4592, -0.032, -0.0, -0.4592, -0.0167, -0.0, -0.4572, -0.0187, -0.0, -0.4564, -0.0206, -0.0, -0.4558, -0.0233, -0.0, -0.457, -0.0214, -0.0, -0.4555, -0.0231, -0.0, -0.4549, -0.0168, 0.0017, -0.4571, -0.019, 0.0017, -0.4563, -0.0168, 0.0039, -0.4571, -0.019, 0.0039, -0.4563, -0.0201, 0.0017, -0.4559, -0.0201, 0.0039, -0.4559,
-0.0212, 0.0017, -0.4556, -0.0233, 0.0017, -0.4548, -0.0212, 0.0039, -0.4556, -0.0233, 0.0039, -0.4548, -0.0255, -0.0, -0.457, -0.0277, -0.0, -0.457, -0.0252, -0.0, -0.454, -0.0277, -0.0, -0.4549, -0.0298, -0.0, -0.457, -0.032, -0.0, -0.457, -0.0298, -0.0, -0.4549, -0.032, -0.0, -0.4549, -0.0251, 0.0017, -0.4541, -0.0251, 0.0039, -0.4541, -0.0263, -0.0, -0.4536, -0.0275, -0.0, -0.453, -0.0296, -0.0, -0.4519, -0.032, -0.0, -0.4527, -0.0307, -0.0, -0.4514, -0.0318, -0.0, -0.4508, -0.0262, 0.0017, -0.4536, -0.0277, 0.0017, -0.4529, -0.0262, 0.0039, -0.4536, -0.0277, 0.0039, -0.4529, -0.0295, 0.0017, -0.452, -0.0295, 0.0039, -0.452, -0.0306, 0.0017, -0.4514, -0.032, 0.0017, -0.4507, -0.0306, 0.0039, -0.4514, -0.032, 0.0039, -0.4507,
-0.0333, -0.0, -0.4691, -0.0344, -0.0, -0.4686, -0.0365, -0.0, -0.4676, -0.0377, -0.0, -0.467, -0.0334, 0.0017, -0.4691, -0.0334, 0.0039, -0.4691, -0.0345, 0.0017, -0.4686, -0.0364, 0.0017, -0.4677, -0.0345, 0.0039, -0.4686, -0.0364, 0.0039, -0.4677, -0.0378, 0.0017, -0.467, -0.0378, 0.0039, -0.467, -0.0342, -0.0, -0.4657, -0.0363, -0.0, -0.4657, -0.0342, -0.0, -0.4635, -0.0363, -0.0, -0.4635, -0.0388, -0.0, -0.4664, -0.0409, -0.0, -0.4654, -0.0385, -0.0, -0.4635, -0.0407, -0.0, -0.4635, -0.0389, 0.0017, -0.4664, -0.0407, 0.0017, -0.4655, -0.0389, 0.0039, -0.4664, -0.0407, 0.0039, -0.4655, -0.0342, -0.0, -0.4614, -0.0363, -0.0, -0.4614, -0.0342, -0.0, -0.4592, -0.0363, -0.0, -0.4592, -0.0385, -0.0, -0.4614, -0.0407, -0.0, -0.4614,
-0.0385, -0.0, -0.4592, -0.0407, -0.0, -0.4592, -0.042, -0.0, -0.4648, -0.0431, -0.0, -0.4641, -0.0448, -0.0, -0.4631, -0.042, 0.0017, -0.4647, -0.042, 0.0039, -0.4647, -0.0431, 0.0017, -0.4641, -0.0449, 0.0017, -0.463, -0.0431, 0.0039, -0.4641, -0.0449, 0.0039, -0.463, -0.0429, -0.0, -0.4614, -0.0458, -0.0, -0.4625, -0.0429, -0.0, -0.4592, -0.045, -0.0, -0.4592, -0.0475, -0.0, -0.4615, -0.0488, -0.0, -0.4607, -0.0472, -0.0, -0.4592, -0.0498, -0.0, -0.46, -0.046, 0.0017, -0.4624, -0.046, 0.0039, -0.4624, -0.0472, 0.0017, -0.4617, -0.0488, 0.0017, -0.4606, -0.0472, 0.0039, -0.4617, -0.0488, 0.0039, -0.4606, -0.0499, 0.0017, -0.4599, -0.0499, 0.0039, -0.4599, -0.0342, -0.0, -0.457, -0.0363, -0.0, -0.457, -0.0342, -0.0, -0.4549,
-0.0363, -0.0, -0.4549, -0.0385, -0.0, -0.457, -0.0407, -0.0, -0.457, -0.0385, -0.0, -0.4549, -0.0407, -0.0, -0.4549, -0.0342, -0.0, -0.4527, -0.0363, -0.0, -0.4527, -0.0339, -0.0, -0.4498, -0.0363, -0.0, -0.4505, -0.0385, -0.0, -0.4527, -0.0415, -0.0, -0.4515, -0.0393, -0.0, -0.4499, -0.0408, -0.0, -0.4505, -0.0339, 0.0017, -0.4498, -0.0339, 0.0039, -0.4498, -0.0417, 0.0017, -0.4517, -0.0417, 0.0039, -0.4517, -0.0392, 0.0017, -0.4498, -0.0408, 0.0017, -0.4505, -0.0392, 0.0039, -0.4498, -0.0408, 0.0039, -0.4505, -0.0429, -0.0, -0.457, -0.045, -0.0, -0.457, -0.0429, -0.0, -0.4549, -0.045, -0.0, -0.4549, -0.0472, -0.0, -0.457, -0.0494, -0.0, -0.457, -0.0472, -0.0, -0.4549, -0.0494, -0.0, -0.4549, -0.0431, -0.0, -0.4525,
-0.0453, -0.0, -0.452, -0.0425, -0.0, -0.451, -0.045, -0.0, -0.4515, -0.0468, -0.0, -0.452, -0.0492, -0.0, -0.4523, -0.0472, -0.0, -0.4516, -0.0496, -0.0, -0.4508, -0.0426, 0.0017, -0.4523, -0.0452, 0.0017, -0.452, -0.0426, 0.0039, -0.4523, -0.0452, 0.0039, -0.452, -0.0429, 0.0017, -0.4512, -0.0447, 0.0017, -0.4515, -0.0429, 0.0039, -0.4512, -0.0447, 0.0039, -0.4515, -0.0469, 0.0017, -0.4519, -0.0494, 0.0017, -0.4522, -0.0469, 0.0039, -0.4519, -0.0494, 0.0039, -0.4522, -0.0473, 0.0017, -0.4516, -0.0494, 0.0017, -0.4509, -0.0473, 0.0039, -0.4516, -0.0494, 0.0039, -0.4509, -0.0514, -0.0, -0.459, -0.0515, 0.0017, -0.4589, -0.0515, 0.0039, -0.4589, -0.0524, -0.0, -0.4582, -0.054, -0.0, -0.457, -0.0515, -0.0, -0.4549, -0.0537, -0.0, -0.4549,
-0.0551, -0.0, -0.4562, -0.0562, -0.0, -0.4553, -0.0575, -0.0, -0.4543, -0.0526, 0.0017, -0.4581, -0.0537, 0.0017, -0.4573, -0.0526, 0.0039, -0.4581, -0.0537, 0.0039, -0.4573, -0.0551, 0.0017, -0.4562, -0.0551, 0.0039, -0.4562, -0.0562, 0.0017, -0.4553, -0.0576, 0.0017, -0.4543, -0.0562, 0.0039, -0.4553, -0.0576, 0.0039, -0.4543, -0.0508, -0.0, -0.4515, -0.0537, -0.0, -0.4527, -0.0514, -0.0, -0.4504, -0.0529, -0.0, -0.4498, -0.0559, -0.0, -0.4527, -0.0586, -0.0, -0.4534, -0.0559, -0.0, -0.4505, -0.058, -0.0, -0.4505, -0.0505, 0.0017, -0.4517, -0.0505, 0.0039, -0.4517, -0.0521, 0.0017, -0.4499, -0.0529, 0.0017, -0.4497, -0.0521, 0.0039, -0.4499, -0.0529, 0.0039, -0.4497, -0.0586, 0.0017, -0.4534, -0.0586, 0.0039, -0.4534, -0.0599, -0.0, -0.4523,
-0.061, -0.0, -0.4513, -0.0621, -0.0, -0.4502, -0.0599, 0.0017, -0.4523, -0.0599, 0.0039, -0.4523, -0.061, 0.0017, -0.4513, -0.0621, 0.0017, -0.4502, -0.061, 0.0039, -0.4513, -0.0621, 0.0039, -0.4502, -0.0046, -0.0, -0.4323, -0.0058, -0.0, -0.4328, -0.0045, 0.0017, -0.4322, -0.0059, 0.0017, -0.4328, -0.0045, 0.0039, -0.4322, -0.0059, 0.0039, -0.4328, -0.0081, -0.0, -0.4332, -0.0103, -0.0, -0.4333, -0.0125, -0.0, -0.4333, -0.0147, -0.0, -0.4331, -0.0081, 0.0017, -0.4332, -0.0103, 0.0017, -0.4333, -0.0081, 0.0039, -0.4332, -0.0103, 0.0039, -0.4333, -0.0125, 0.0017, -0.4333, -0.0147, 0.0017, -0.4331, -0.0125, 0.0039, -0.4333, -0.0147, 0.0039, -0.4331, -0.0165, -0.0, -0.4326, -0.0168, 0.0017, -0.4325, -0.0168, 0.0039, -0.4325, -0.002, -0.0, -0.4306,
0.0001, -0.0, -0.4284, -0.0008, -0.0, -0.4295, -0.0035, -0.0, -0.4317, -0.006, -0.0, -0.431, -0.0038, -0.0, -0.4288, -0.006, -0.0, -0.4288, -0.0019, 0.0017, -0.4306, -0.0019, 0.0039, -0.4306, 0.0001, 0.0017, -0.4285, -0.0009, 0.0017, -0.4296, 0.0001, 0.0039, -0.4285, -0.0009, 0.0039, -0.4296, -0.0034, 0.0017, -0.4317, -0.0034, 0.0039, -0.4317, 0.0012, -0.0, -0.4269, -0.0016, -0.0, -0.4266, 0.0005, -0.0, -0.4245, -0.0016, -0.0, -0.4245, -0.0038, -0.0, -0.4266, -0.006, -0.0, -0.4266, -0.0038, -0.0, -0.4245, -0.006, -0.0, -0.4245, 0.0012, 0.0017, -0.4269, 0.0012, 0.0039, -0.4269, -0.0081, -0.0, -0.431, -0.0103, -0.0, -0.431, -0.0081, -0.0, -0.4288, -0.0103, -0.0, -0.4288, -0.0125, -0.0, -0.431, -0.0146, -0.0, -0.431,
-0.0125, -0.0, -0.4288, -0.0146, -0.0, -0.4288, -0.0081, -0.0, -0.4266, -0.0103, -0.0, -0.4266, -0.0081, -0.0, -0.4245, -0.0103, -0.0, -0.4245, -0.0125, -0.0, -0.4266, -0.0146, -0.0, -0.4266, -0.0125, -0.0, -0.4245, -0.0146, -0.0, -0.4245, 0.0005, -0.0, -0.4223, -0.0016, -0.0, -0.4223, 0.0005, -0.0, -0.4201, -0.0016, -0.0, -0.4201, -0.0038, -0.0, -0.4223, -0.006, -0.0, -0.4223, -0.0038, -0.0, -0.4201, -0.006, -0.0, -0.4201, 0.0005, -0.0, -0.418, -0.0016, -0.0, -0.418, 0.0005, -0.0, -0.4158, -0.0016, -0.0, -0.4158, -0.0038, -0.0, -0.418, -0.006, -0.0, -0.418, -0.0038, -0.0, -0.4158, -0.006, -0.0, -0.4158, -0.0081, -0.0, -0.4223, -0.0103, -0.0, -0.4223, -0.0081, -0.0, -0.4201, -0.0103, -0.0, -0.4201,
-0.0125, -0.0, -0.4223, -0.0146, -0.0, -0.4223, -0.0125, -0.0, -0.4201, -0.0146, -0.0, -0.4201, -0.0081, -0.0, -0.418, -0.0103, -0.0, -0.418, -0.0081, -0.0, -0.4158, -0.0103, -0.0, -0.4158, -0.0125, -0.0, -0.418, -0.0146, -0.0, -0.418, -0.0125, -0.0, -0.4158, -0.0146, -0.0, -0.4158, -0.0175, -0.0, -0.4322, -0.0194, -0.0, -0.4314, -0.0168, -0.0, -0.4288, -0.019, -0.0, -0.4288, -0.0206, -0.0, -0.4304, -0.0217, -0.0, -0.4293, -0.0226, -0.0, -0.4281, -0.0179, 0.0017, -0.4321, -0.0191, 0.0017, -0.4315, -0.0179, 0.0039, -0.4321, -0.0191, 0.0039, -0.4315, -0.0206, 0.0017, -0.4304, -0.0206, 0.0039, -0.4304, -0.0217, 0.0017, -0.4293, -0.0226, 0.0017, -0.4282, -0.0217, 0.0039, -0.4293, -0.0226, 0.0039, -0.4282, -0.0168, -0.0, -0.4266,
-0.019, -0.0, -0.4266, -0.0168, -0.0, -0.4245, -0.019, -0.0, -0.4245, -0.0212, -0.0, -0.4266, -0.0234, -0.0, -0.427, -0.0212, -0.0, -0.4245, -0.0243, -0.0, -0.4247, -0.0236, 0.0017, -0.4267, -0.0236, 0.0039, -0.4267, -0.0242, 0.0017, -0.4249, -0.0242, 0.0039, -0.4249, -0.0246, -0.0, -0.4237, -0.0246, 0.0017, -0.4238, -0.0246, 0.0039, -0.4238, -0.0168, -0.0, -0.4223, -0.019, -0.0, -0.4223, -0.0168, -0.0, -0.4201, -0.019, -0.0, -0.4201, -0.0212, -0.0, -0.4223, -0.0233, -0.0, -0.4223, -0.0212, -0.0, -0.4201, -0.0233, -0.0, -0.4201, -0.0168, -0.0, -0.418, -0.019, -0.0, -0.418, -0.0168, -0.0, -0.4158, -0.019, -0.0, -0.4158, -0.0212, -0.0, -0.418, -0.0233, -0.0, -0.418, -0.0212, -0.0, -0.4158, -0.0246, -0.0, -0.4153,
-0.0244, 0.0017, -0.4148, -0.0244, 0.0039, -0.4148, -0.0249, -0.0, -0.4223, -0.0251, -0.0, -0.4201, -0.0249, 0.0017, -0.4223, -0.0249, 0.0039, -0.4223, -0.0251, 0.0017, -0.4201, -0.0251, 0.0039, -0.4201, -0.025, -0.0, -0.4179, -0.0247, -0.0, -0.4162, -0.025, 0.0017, -0.418, -0.025, 0.0039, -0.418, -0.0247, 0.0017, -0.4159, -0.0247, 0.0039, -0.4159, -0.035, -0.0, -0.4493, -0.0362, -0.0, -0.4487, -0.0387, -0.0, -0.4478, -0.0407, -0.0, -0.4483, -0.0385, -0.0, -0.4463, -0.0407, -0.0, -0.4464, -0.035, 0.0017, -0.4493, -0.0363, 0.0017, -0.4487, -0.035, 0.0039, -0.4493, -0.0363, 0.0039, -0.4487, -0.0387, 0.0017, -0.4478, -0.0407, 0.0017, -0.4483, -0.0387, 0.0039, -0.4478, -0.0407, 0.0039, -0.4483, -0.0385, 0.0017, -0.4462, -0.0414, 0.0017, -0.4471,
-0.0385, 0.0039, -0.4462, -0.0414, 0.0039, -0.4471, -0.0384, -0.0, -0.444, -0.0407, -0.0, -0.444, -0.0388, -0.0, -0.4417, -0.0407, -0.0, -0.4418, -0.0384, 0.0017, -0.444, -0.0384, 0.0039, -0.444, -0.0388, 0.0017, -0.4418, -0.0388, 0.0039, -0.4418, -0.0426, -0.0, -0.4483, -0.045, -0.0, -0.4483, -0.0419, -0.0, -0.4471, -0.045, -0.0, -0.4462, -0.0472, -0.0, -0.4483, -0.0494, -0.0, -0.4483, -0.0472, -0.0, -0.4462, -0.0494, -0.0, -0.4462, -0.0419, 0.0017, -0.4476, -0.0419, 0.0039, -0.4476, -0.0419, 0.0017, -0.4471, -0.0419, 0.0039, -0.4471, -0.0429, -0.0, -0.444, -0.045, -0.0, -0.444, -0.0429, -0.0, -0.4418, -0.045, -0.0, -0.4418, -0.0472, -0.0, -0.444, -0.0494, -0.0, -0.444, -0.0472, -0.0, -0.4418, -0.0494, -0.0, -0.4418,
-0.0392, -0.0, -0.4402, -0.0397, -0.0, -0.4392, -0.0406, -0.0, -0.4375, -0.0393, 0.0017, -0.44, -0.0398, 0.0017, -0.439, -0.0393, 0.0039, -0.44, -0.0398, 0.0039, -0.439, -0.0402, 0.0017, -0.438, -0.0402, 0.0039, -0.438, -0.0429, -0.0, -0.4397, -0.045, -0.0, -0.4397, -0.0421, -0.0, -0.4364, -0.045, -0.0, -0.4375, -0.0472, -0.0, -0.4397, -0.0494, -0.0, -0.4397, -0.0472, -0.0, -0.4375, -0.0501, -0.0, -0.4363, -0.0419, 0.0017, -0.4365, -0.0419, 0.0039, -0.4365, -0.0504, 0.0017, -0.4364, -0.0504, 0.0039, -0.4364, -0.0433, -0.0, -0.4359, -0.045, -0.0, -0.4355, -0.0473, -0.0, -0.4354, -0.0491, -0.0, -0.4358, -0.043, 0.0017, -0.4359, -0.045, 0.0017, -0.4355, -0.043, 0.0039, -0.4359, -0.045, 0.0039, -0.4355, -0.0472, 0.0017, -0.4354,
-0.0493, 0.0017, -0.4359, -0.0472, 0.0039, -0.4354, -0.0493, 0.0039, -0.4359, -0.0521, -0.0, -0.4488, -0.0535, -0.0, -0.4481, -0.0515, -0.0, -0.4462, -0.0538, -0.0, -0.4464, -0.0547, -0.0, -0.4475, -0.058, -0.0, -0.4483, -0.0556, -0.0, -0.446, -0.058, -0.0, -0.4462, -0.0521, 0.0017, -0.4489, -0.0535, 0.0017, -0.4481, -0.0521, 0.0039, -0.4489, -0.0535, 0.0039, -0.4481, -0.0538, 0.0017, -0.4462, -0.0538, 0.0039, -0.4462, -0.0548, 0.0017, -0.4473, -0.0548, 0.0039, -0.4473, -0.0556, 0.0017, -0.4461, -0.0556, 0.0039, -0.4461, -0.0515, -0.0, -0.444, -0.0543, -0.0, -0.444, -0.0515, -0.0, -0.4418, -0.0541, -0.0, -0.4417, -0.0559, -0.0, -0.444, -0.058, -0.0, -0.444, -0.0558, -0.0, -0.4419, -0.058, -0.0, -0.4418, -0.0543, 0.0017, -0.444,
-0.0543, 0.0039, -0.444, -0.0541, 0.0017, -0.4418, -0.0541, 0.0039, -0.4418, -0.0559, 0.0017, -0.444, -0.0559, 0.0039, -0.444, -0.0557, 0.0017, -0.4418, -0.0557, 0.0039, -0.4418, -0.0602, -0.0, -0.4483, -0.0632, -0.0, -0.4491, -0.0602, -0.0, -0.4462, -0.0624, -0.0, -0.4462, -0.0642, -0.0, -0.448, -0.0653, -0.0, -0.4468, -0.0662, -0.0, -0.4457, -0.0632, 0.0017, -0.4491, -0.0632, 0.0039, -0.4491, -0.0642, 0.0017, -0.4481, -0.0642, 0.0039, -0.4481, -0.0653, 0.0017, -0.4468, -0.0662, 0.0017, -0.4457, -0.0653, 0.0039, -0.4468, -0.0662, 0.0039, -0.4457, -0.0602, -0.0, -0.444, -0.0624, -0.0, -0.444, -0.0602, -0.0, -0.4418, -0.0624, -0.0, -0.4418, -0.0646, -0.0, -0.444, -0.0673, -0.0, -0.4444, -0.0646, -0.0, -0.4418, -0.0667, -0.0, -0.4418,
-0.0673, 0.0017, -0.4444, -0.0673, 0.0039, -0.4444, -0.0515, -0.0, -0.4397, -0.0543, -0.0, -0.4393, -0.052, -0.0, -0.4375, -0.054, 0.0003, -0.4378, -0.055, -0.0, -0.4398, -0.058, -0.0, -0.4397, -0.0559, -0.0, -0.4375, -0.058, -0.0, -0.4375, -0.0544, 0.0017, -0.4395, -0.0544, 0.0039, -0.4395, -0.0518, 0.0017, -0.4373, -0.0539, 0.0017, -0.4378, -0.0518, 0.0039, -0.4373, -0.0539, 0.0039, -0.4378, -0.0551, 0.0017, -0.44, -0.0551, 0.0039, -0.44, -0.0545, -0.0, -0.436, -0.0548, -0.0, -0.4349, -0.058, -0.0, -0.4353, -0.056, -0.0, -0.4329, -0.058, -0.0, -0.4332, -0.0545, 0.0017, -0.4358, -0.0545, 0.0039, -0.4358, -0.0549, 0.0017, -0.4347, -0.0549, 0.0039, -0.4347, -0.0558, 0.0017, -0.4331, -0.0558, 0.0039, -0.4331, -0.0602, -0.0, -0.4397,
-0.0624, -0.0, -0.4397, -0.0602, -0.0, -0.4375, -0.0624, -0.0, -0.4375, -0.0646, -0.0, -0.4397, -0.0667, -0.0, -0.4397, -0.0646, -0.0, -0.4375, -0.0667, -0.0, -0.4375, -0.0602, -0.0, -0.4353, -0.0624, -0.0, -0.4353, -0.0602, -0.0, -0.4332, -0.0624, -0.0, -0.4332, -0.0646, -0.0, -0.4353, -0.0667, -0.0, -0.4353, -0.0646, -0.0, -0.4332, -0.0667, -0.0, -0.4332, -0.0568, -0.0, -0.4319, -0.0576, -0.0, -0.4307, -0.0587, -0.0, -0.4294, -0.0568, 0.0017, -0.4318, -0.0576, 0.0017, -0.4307, -0.0568, 0.0039, -0.4318, -0.0576, 0.0039, -0.4307, -0.0587, 0.0017, -0.4294, -0.0587, 0.0039, -0.4294, -0.0602, -0.0, -0.431, -0.0624, -0.0, -0.431, -0.0596, -0.0, -0.4283, -0.0624, -0.0, -0.4288, -0.0646, -0.0, -0.431, -0.0667, -0.0, -0.431,
-0.0646, -0.0, -0.4288, -0.0667, -0.0, -0.4288, -0.0596, 0.0017, -0.4283, -0.0596, 0.0039, -0.4283, -0.0607, -0.0, -0.4271, -0.0615, -0.0, -0.426, -0.0628, -0.0, -0.4243, -0.0646, -0.0, -0.4266, -0.0667, -0.0, -0.4266, -0.0646, -0.0, -0.4245, -0.0667, -0.0, -0.4245, -0.0607, 0.0017, -0.427, -0.0616, 0.0017, -0.4259, -0.0607, 0.0039, -0.427, -0.0616, 0.0039, -0.4259, -0.0627, 0.0017, -0.4245, -0.0627, 0.0039, -0.4245, -0.0635, -0.0, -0.4234, -0.0642, -0.0, -0.4222, -0.0667, -0.0, -0.4223, -0.0652, -0.0, -0.4207, -0.0659, -0.0, -0.4197, -0.0635, 0.0017, -0.4234, -0.0635, 0.0039, -0.4234, -0.0642, 0.0017, -0.4223, -0.0642, 0.0039, -0.4223, -0.0653, 0.0017, -0.4206, -0.066, 0.0017, -0.4195, -0.0653, 0.0039, -0.4206, -0.066, 0.0039, -0.4195,
-0.067, -0.0, -0.4178, -0.0677, -0.0, -0.4167, -0.0669, 0.0017, -0.418, -0.0669, 0.0039, -0.418, -0.0677, 0.0017, -0.4167, -0.0677, 0.0039, -0.4167, -0.0682, -0.0, -0.4433, -0.0693, -0.0, -0.4421, -0.0702, -0.0, -0.4409, -0.0682, 0.0017, -0.4434, -0.0682, 0.0039, -0.4434, -0.0693, 0.0017, -0.442, -0.0701, 0.0017, -0.441, -0.0693, 0.0039, -0.442, -0.0701, 0.0039, -0.441, -0.0689, -0.0, -0.4397, -0.0709, -0.0, -0.4399, -0.0689, -0.0, -0.4375, -0.0721, -0.0, -0.4383, -0.0729, -0.0, -0.4373, -0.0711, 0.0017, -0.4397, -0.0711, 0.0039, -0.4397, -0.0721, 0.0017, -0.4384, -0.0721, 0.0039, -0.4384, -0.0728, 0.0017, -0.4374, -0.0728, 0.0039, -0.4374, -0.0689, -0.0, -0.4353, -0.0711, -0.0, -0.4353, -0.0689, -0.0, -0.4332, -0.0711, -0.0, -0.4332,
-0.0739, -0.0, -0.4357, -0.0747, -0.0, -0.4346, -0.0732, -0.0, -0.4332, -0.0754, -0.0, -0.4334, -0.0739, 0.0017, -0.4358, -0.0746, 0.0017, -0.4347, -0.0739, 0.0039, -0.4358, -0.0746, 0.0039, -0.4347, -0.0755, 0.0017, -0.4332, -0.0755, 0.0039, -0.4332, -0.087, -0.0, -0.4321, -0.0879, -0.0, -0.4322, -0.0868, 0.0017, -0.4321, -0.0882, 0.0017, -0.4322, -0.0868, 0.0039, -0.4321, -0.0882, 0.0039, -0.4322, -0.0689, -0.0, -0.431, -0.0711, -0.0, -0.431, -0.0689, -0.0, -0.4288, -0.0711, -0.0, -0.4288, -0.0732, -0.0, -0.431, -0.0763, -0.0, -0.4309, -0.0732, -0.0, -0.4288, -0.0757, -0.0, -0.4286, -0.0763, 0.0017, -0.4309, -0.0763, 0.0039, -0.4309, -0.0758, 0.0017, -0.4288, -0.0758, 0.0039, -0.4288, -0.0689, -0.0, -0.4266, -0.0711, -0.0, -0.4266,
-0.0689, -0.0, -0.4245, -0.0711, -0.0, -0.4245, -0.0744, -0.0, -0.426, -0.0754, -0.0, -0.4266, -0.0741, -0.0, -0.4239, -0.0748, -0.0, -0.4246, -0.0742, 0.0017, -0.4257, -0.0753, 0.0017, -0.4266, -0.0742, 0.0039, -0.4257, -0.0753, 0.0039, -0.4266, -0.0742, 0.0017, -0.4242, -0.0748, 0.0017, -0.4246, -0.0742, 0.0039, -0.4242, -0.0748, 0.0039, -0.4246, -0.0802, -0.0, -0.4304, -0.0777, -0.0, -0.4286, -0.0791, -0.0, -0.4297, -0.0817, -0.0, -0.4311, -0.0839, -0.0, -0.4317, -0.0829, -0.0, -0.4278, -0.0836, -0.0, -0.4276, -0.0801, 0.0017, -0.4303, -0.0801, 0.0039, -0.4303, -0.0777, 0.0017, -0.4286, -0.079, 0.0017, -0.4296, -0.0777, 0.0039, -0.4286, -0.079, 0.0039, -0.4296, -0.0819, 0.0017, -0.4311, -0.0841, 0.0017, -0.4318, -0.0819, 0.0039, -0.4311,
-0.0841, 0.0039, -0.4318, -0.0829, 0.0017, -0.4278, -0.083, 0.0017, -0.4277, -0.0829, 0.0039, -0.4278, -0.083, 0.0039, -0.4277, -0.0766, -0.0, -0.4275, -0.0798, -0.0, -0.4264, -0.0778, -0.0, -0.4246, -0.0794, -0.0, -0.4248, -0.0847, -0.0, -0.4285, -0.084, -0.0, -0.4275, -0.0811, -0.0, -0.4253, -0.0841, -0.0, -0.4245, -0.0767, 0.0017, -0.4275, -0.0798, 0.0017, -0.4264, -0.0767, 0.0039, -0.4275, -0.0798, 0.0039, -0.4264, -0.0778, 0.0017, -0.4246, -0.0794, 0.0017, -0.4248, -0.0778, 0.0039, -0.4246, -0.0794, 0.0039, -0.4248, -0.0847, 0.0017, -0.4285, -0.084, 0.0017, -0.4275, -0.0847, 0.0039, -0.4285, -0.084, 0.0039, -0.4275, -0.0811, 0.0017, -0.4253, -0.0811, 0.0039, -0.4253, -0.0689, -0.0, -0.4223, -0.0711, -0.0, -0.4223, -0.0689, -0.0, -0.4201,
-0.0711, -0.0, -0.4201, -0.0738, -0.0, -0.4222, -0.0754, -0.0, -0.4223, -0.0734, -0.0, -0.4201, -0.0754, -0.0, -0.4201, -0.0689, -0.0, -0.418, -0.0711, -0.0, -0.418, -0.0683, -0.0, -0.4156, -0.0711, -0.0, -0.4158, -0.0733, -0.0, -0.418, -0.0754, -0.0, -0.418, -0.0736, -0.0, -0.4158, -0.0754, -0.0, -0.4158, -0.0683, 0.0017, -0.4156, -0.0683, 0.0039, -0.4156, -0.0778, -0.0, -0.422, -0.0789, -0.0, -0.4229, -0.0775, -0.0, -0.4204, -0.0798, -0.0, -0.4201, -0.0819, -0.0, -0.4223, -0.0841, -0.0, -0.4223, -0.0819, -0.0, -0.4201, -0.0841, -0.0, -0.4201, -0.078, 0.0017, -0.4224, -0.079, 0.0017, -0.4229, -0.078, 0.0039, -0.4224, -0.079, 0.0039, -0.4229, -0.0776, -0.0, -0.418, -0.0798, -0.0, -0.418, -0.0776, -0.0, -0.4158,
-0.0798, -0.0, -0.4158, -0.0819, -0.0, -0.418, -0.0841, -0.0, -0.418, -0.0819, -0.0, -0.4158, -0.0841, -0.0, -0.4158, -0.0862, -0.0, -0.432, -0.0887, -0.0, -0.4321, -0.0863, -0.0, -0.4288, -0.0884, -0.0, -0.4288, -0.0909, -0.0, -0.4317, -0.093, -0.0, -0.4309, -0.0906, -0.0, -0.4288, -0.0928, -0.0, -0.4288, -0.0858, 0.0017, -0.432, -0.0893, 0.0017, -0.432, -0.0858, 0.0039, -0.432, -0.0893, 0.0039, -0.432, -0.0906, 0.0017, -0.4318, -0.0928, 0.0017, -0.431, -0.0906, 0.0039, -0.4318, -0.0928, 0.0039, -0.431, -0.0863, -0.0, -0.4266, -0.0884, -0.0, -0.4266, -0.0863, -0.0, -0.4245, -0.0884, -0.0, -0.4245, -0.0906, -0.0, -0.4266, -0.0928, -0.0, -0.4266, -0.0906, -0.0, -0.4245, -0.0928, -0.0, -0.4245, -0.0941, -0.0, -0.4301,
-0.0953, -0.0, -0.4293, -0.0965, -0.0, -0.4282, -0.0942, 0.0017, -0.4301, -0.0942, 0.0039, -0.4301, -0.0953, 0.0017, -0.4293, -0.0965, 0.0017, -0.4282, -0.0953, 0.0039, -0.4293, -0.0965, 0.0039, -0.4282, -0.0949, -0.0, -0.4266, -0.0976, -0.0, -0.427, -0.0949, -0.0, -0.4245, -0.0971, -0.0, -0.4245, -0.0984, -0.0, -0.4258, -0.0992, -0.0, -0.4247, -0.0976, 0.0017, -0.427, -0.0976, 0.0039, -0.427, -0.0984, 0.0017, -0.4259, -0.0984, 0.0039, -0.4259, -0.0993, 0.0017, -0.4245, -0.0993, 0.0039, -0.4245, -0.0863, -0.0, -0.4223, -0.0884, -0.0, -0.4223, -0.0863, -0.0, -0.4201, -0.0884, -0.0, -0.4201, -0.0906, -0.0, -0.4223, -0.0928, -0.0, -0.4223, -0.0906, -0.0, -0.4201, -0.0928, -0.0, -0.4201, -0.0863, -0.0, -0.418, -0.0884, -0.0, -0.418,
-0.0863, -0.0, -0.4158, -0.0884, -0.0, -0.4158, -0.0906, -0.0, -0.418, -0.0928, -0.0, -0.418, -0.0906, -0.0, -0.4158, -0.0928, -0.0, -0.4158, -0.0949, -0.0, -0.4223, -0.0971, -0.0, -0.4223, -0.0949, -0.0, -0.4201, -0.0971, -0.0, -0.4201, -0.1, -0.0, -0.4226, -0.1005, -0.0, -0.4205, -0.1006, -0.0, -0.4197, -0.1001, 0.0017, -0.4223, -0.1001, 0.0039, -0.4223, -0.1003, 0.0017, -0.4211, -0.1005, 0.0017, -0.42, -0.1003, 0.0039, -0.4211, -0.1005, 0.0039, -0.42, -0.0949, -0.0, -0.418, -0.0971, -0.0, -0.418, -0.0949, -0.0, -0.4158, -0.0971, -0.0, -0.4158, -0.0993, -0.0, -0.418, -0.1006, -0.0, -0.418, -0.1003, -0.0, -0.4157, -0.1004, -0.0, -0.4166, -0.1006, 0.0017, -0.418, -0.1006, 0.0039, -0.418, -0.1003, 0.0017, -0.4154,
-0.1004, 0.0017, -0.4165, -0.1003, 0.0039, -0.4154, -0.1004, 0.0039, -0.4165, 0.0014, -0.0, -0.4131, -0.0016, -0.0, -0.4136, 0.0002, -0.0, -0.4112, -0.0016, -0.0, -0.4115, -0.0038, -0.0, -0.4136, -0.006, -0.0, -0.4136, -0.0038, -0.0, -0.4115, -0.006, -0.0, -0.4115, 0.0014, 0.0017, -0.413, 0.0014, 0.0039, -0.413, 0.0004, 0.0017, -0.4114, 0.0004, 0.0039, -0.4114, -0.0005, -0.0, -0.4104, -0.0012, -0.0, -0.4094, -0.003, -0.0, -0.4081, -0.006, -0.0, -0.4093, -0.004, -0.0, -0.4074, -0.0057, -0.0, -0.4065, -0.0005, 0.0017, -0.4104, -0.0015, 0.0017, -0.4091, -0.0005, 0.0039, -0.4104, -0.0015, 0.0039, -0.4091, -0.0027, 0.0017, -0.4082, -0.0027, 0.0039, -0.4082, -0.0038, 0.0017, -0.4075, -0.0059, 0.0017, -0.4064, -0.0038, 0.0039, -0.4075,
-0.0059, 0.0039, -0.4064, -0.0081, -0.0, -0.4136, -0.0103, -0.0, -0.4136, -0.0081, -0.0, -0.4115, -0.0103, -0.0, -0.4115, -0.0125, -0.0, -0.4136, -0.0146, -0.0, -0.4136, -0.0125, -0.0, -0.4115, -0.0146, -0.0, -0.4115, -0.0081, -0.0, -0.4093, -0.0103, -0.0, -0.4093, -0.008, -0.0, -0.406, -0.0103, -0.0, -0.4071, -0.0125, -0.0, -0.4093, -0.0146, -0.0, -0.4093, -0.0125, -0.0, -0.4071, -0.0146, -0.0, -0.4061, -0.0076, 0.0017, -0.4061, -0.0076, 0.0039, -0.4061, -0.0147, 0.0017, -0.4061, -0.0147, 0.0039, -0.4061, 0.0011, -0.0, -0.398, 0.0012, 0.0017, -0.3981, 0.0012, 0.0039, -0.3981, -0.0088, -0.0, -0.4059, -0.0103, -0.0, -0.4059, -0.0125, -0.0, -0.4059, -0.0136, -0.0, -0.406, -0.0087, 0.0017, -0.406, -0.0103, 0.0017, -0.4059,
-0.0087, 0.0039, -0.406, -0.0103, 0.0039, -0.4059, -0.0125, 0.0017, -0.4059, -0.0136, 0.0017, -0.406, -0.0125, 0.0039, -0.4059, -0.0136, 0.0039, -0.406, -0.0168, -0.0, -0.4136, -0.019, -0.0, -0.4136, -0.0168, -0.0, -0.4115, -0.019, -0.0, -0.4115, -0.0212, -0.0, -0.4136, -0.0239, -0.0, -0.4133, -0.0222, -0.0, -0.4105, -0.0231, -0.0, -0.4116, -0.024, 0.0017, -0.4136, -0.024, 0.0039, -0.4136, -0.0221, 0.0017, -0.4105, -0.023, 0.0017, -0.4115, -0.0221, 0.0039, -0.4105, -0.023, 0.0039, -0.4115, -0.0168, -0.0, -0.4093, -0.0198, -0.0, -0.4083, -0.0171, -0.0, -0.4068, -0.0187, -0.0, -0.4075, -0.0211, -0.0, -0.4093, -0.0199, 0.0017, -0.4083, -0.0199, 0.0039, -0.4083, -0.0168, 0.0017, -0.4067, -0.0189, 0.0017, -0.4076, -0.0168, 0.0039, -0.4067,
-0.0189, 0.0039, -0.4076, -0.0211, 0.0017, -0.4093, -0.0211, 0.0039, -0.4093, 0.0003, -0.0, -0.3966, -0.0007, -0.0, -0.3948, -0.0012, -0.0, -0.3939, 0.0001, 0.0017, -0.3963, 0.0001, 0.0039, -0.3963, -0.0005, 0.0017, -0.3951, -0.0011, 0.0017, -0.394, -0.0005, 0.0039, -0.3951, -0.0011, 0.0039, -0.394, 0.0005, -0.0, -0.3919, -0.0022, -0.0, -0.3921, 0.0005, -0.0, -0.3897, -0.0016, -0.0, -0.3897, -0.0029, -0.0, -0.391, -0.0035, -0.0, -0.3899, -0.0022, 0.0017, -0.3921, -0.0022, 0.0039, -0.3921, -0.0028, 0.0017, -0.3911, -0.0028, 0.0039, -0.3911, -0.0036, 0.0017, -0.3897, -0.0036, 0.0039, -0.3897, 0.0005, -0.0, -0.3876, -0.0016, -0.0, -0.3876, 0.0005, -0.0, -0.3854, -0.0016, -0.0, -0.3854, -0.0046, -0.0, -0.388, -0.0052, -0.0, -0.3869,
-0.0038, -0.0, -0.3854, -0.0059, -0.0, -0.3857, -0.0045, 0.0017, -0.3881, -0.0052, 0.0017, -0.387, -0.0045, 0.0039, -0.3881, -0.0052, 0.0039, -0.387, -0.0061, 0.0017, -0.3854, -0.0061, 0.0039, -0.3854, 0.0005, -0.0, -0.3832, -0.0016, -0.0, -0.3832, 0.0005, -0.0, -0.3811, -0.0016, -0.0, -0.3811, -0.0038, -0.0, -0.3832, -0.007, -0.0, -0.3839, -0.0038, -0.0, -0.3811, -0.006, -0.0, -0.3811, -0.0069, 0.0017, -0.3841, -0.0069, 0.0039, -0.3841, -0.0076, -0.0, -0.3828, -0.0086, -0.0, -0.3812, -0.0093, -0.0, -0.38, -0.0075, 0.0017, -0.383, -0.0075, 0.0039, -0.383, -0.0086, 0.0017, -0.3811, -0.0093, 0.0017, -0.38, -0.0086, 0.0039, -0.3811, -0.0093, 0.0039, -0.38, -0.0452, -0.0, -0.4024, -0.0472, -0.0, -0.4026, -0.0488, -0.0, -0.4021,
-0.0451, 0.0017, -0.4023, -0.0451, 0.0039, -0.4023, -0.0472, 0.0017, -0.4026, -0.0489, 0.0017, -0.402, -0.0472, 0.0039, -0.4026, -0.0489, 0.0039, -0.402, -0.0429, -0.0, -0.4011, -0.0442, -0.0, -0.4018, -0.042, -0.0, -0.3986, -0.045, -0.0, -0.3984, -0.0472, -0.0, -0.4006, -0.05, -0.0, -0.4013, -0.0472, -0.0, -0.3984, -0.0494, -0.0, -0.3984, -0.0429, 0.0017, -0.4011, -0.044, 0.0017, -0.4017, -0.0429, 0.0039, -0.4011, -0.044, 0.0039, -0.4017, -0.042, 0.0017, -0.3987, -0.042, 0.0039, -0.3987, -0.05, 0.0017, -0.4014, -0.05, 0.0039, -0.4014, -0.0509, -0.0, -0.4001, -0.0514, -0.0, -0.3986, -0.0508, 0.0017, -0.4002, -0.0508, 0.0039, -0.4002, -0.0514, 0.0017, -0.3986, -0.0514, 0.0039, -0.3986, -0.0423, -0.0, -0.396, -0.045, -0.0, -0.3963,
-0.0434, -0.0, -0.3946, -0.0444, -0.0, -0.3935, -0.0472, -0.0, -0.3963, -0.0494, -0.0, -0.3963, -0.0472, -0.0, -0.3928, -0.0498, -0.0, -0.3941, -0.0423, 0.0017, -0.396, -0.0423, 0.0039, -0.396, -0.0435, 0.0017, -0.3946, -0.0444, 0.0017, -0.3935, -0.0435, 0.0039, -0.3946, -0.0444, 0.0039, -0.3935, -0.0472, 0.0017, -0.3928, -0.0494, 0.0017, -0.3938, -0.0472, 0.0039, -0.3928, -0.0494, 0.0039, -0.3938, -0.0512, -0.0, -0.3959, -0.0505, -0.0, -0.3951, -0.0512, 0.0017, -0.3959, -0.0512, 0.0039, -0.3959, -0.0505, 0.0017, -0.3951, -0.0505, 0.0039, -0.3951, 0.0005, -0.0, -0.3789, -0.0016, -0.0, -0.3789, 0.0005, -0.0, -0.3767, -0.0016, -0.0, -0.3767, -0.0038, -0.0, -0.3789, -0.006, -0.0, -0.3789, -0.0038, -0.0, -0.3767, -0.006, -0.0, -0.3767,
0.0005, -0.0, -0.3746, -0.0016, -0.0, -0.3746, 0.0005, -0.0, -0.3724, -0.0016, -0.0, -0.3724, -0.0038, -0.0, -0.3746, -0.006, -0.0, -0.3746, -0.0038, -0.0, -0.3724, -0.006, -0.0, -0.3724, -0.0081, -0.0, -0.3789, -0.0098, -0.0, -0.379, -0.0081, -0.0, -0.3767, -0.0109, -0.0, -0.3769, -0.0115, -0.0, -0.3758, -0.0099, 0.0017, -0.3789, -0.0099, 0.0039, -0.3789, -0.0109, 0.0017, -0.377, -0.0109, 0.0039, -0.377, -0.0115, 0.0017, -0.3759, -0.0115, 0.0039, -0.3759, -0.0081, -0.0, -0.3746, -0.0103, -0.0, -0.3746, -0.0081, -0.0, -0.3724, -0.0103, -0.0, -0.3724, -0.0122, -0.0, -0.3747, -0.0133, -0.0, -0.3728, -0.0139, -0.0, -0.3717, -0.0123, 0.0017, -0.3746, -0.0123, 0.0039, -0.3746, -0.0132, 0.0017, -0.3729, -0.0138, 0.0017, -0.3718,
-0.0132, 0.0039, -0.3729, -0.0138, 0.0039, -0.3718, 0.0016, -0.0, -0.3695, -0.0016, -0.0, -0.3702, 0.0004, -0.0, -0.3678, -0.0016, -0.0, -0.368, -0.0038, -0.0, -0.3702, -0.006, -0.0, -0.3702, -0.0038, -0.0, -0.368, -0.006, -0.0, -0.368, 0.0015, 0.0017, -0.3694, 0.0015, 0.0039, -0.3694, 0.0006, 0.0017, -0.3681, 0.0006, 0.0039, -0.3681, -0.0003, -0.0, -0.3667, -0.001, -0.0, -0.3656, -0.0021, -0.0, -0.364, -0.0038, -0.0, -0.3659, -0.006, -0.0, -0.3659, -0.0027, -0.0, -0.363, -0.006, -0.0, -0.3637, -0.0003, 0.0017, -0.3667, -0.0011, 0.0017, -0.3656, -0.0003, 0.0039, -0.3667, -0.0011, 0.0039, -0.3656, -0.0021, 0.0017, -0.3639, -0.0021, 0.0039, -0.3639, -0.0029, 0.0017, -0.3628, -0.0029, 0.0039, -0.3628, -0.0081, -0.0, -0.3702,
-0.0103, -0.0, -0.3702, -0.0081, -0.0, -0.368, -0.0103, -0.0, -0.368, -0.0125, -0.0, -0.3702, -0.0146, -0.0, -0.3705, -0.0125, -0.0, -0.368, -0.0156, -0.0, -0.3686, -0.0147, 0.0017, -0.3702, -0.0147, 0.0039, -0.3702, -0.0155, 0.0017, -0.3687, -0.0155, 0.0039, -0.3687, -0.0081, -0.0, -0.3659, -0.0103, -0.0, -0.3659, -0.0081, -0.0, -0.3637, -0.0103, -0.0, -0.3637, -0.0125, -0.0, -0.3659, -0.0146, -0.0, -0.3659, -0.0125, -0.0, -0.3637, -0.0146, -0.0, -0.3637, -0.0162, -0.0, -0.3675, -0.0161, 0.0017, -0.3677, -0.0161, 0.0039, -0.3677, -0.0169, -0.0, -0.3662, -0.0179, -0.0, -0.3643, -0.0184, -0.0, -0.3633, -0.0171, 0.0017, -0.3659, -0.0171, 0.0039, -0.3659, -0.0178, 0.0017, -0.3646, -0.0184, 0.0017, -0.3635, -0.0178, 0.0039, -0.3646,
-0.0184, 0.0039, -0.3635, -0.0039, -0.0, -0.3613, -0.006, -0.0, -0.3615, -0.0046, -0.0, -0.3601, -0.0053, -0.0, -0.359, -0.0037, 0.0017, -0.3615, -0.0037, 0.0039, -0.3615, -0.0046, 0.0017, -0.36, -0.0053, 0.0017, -0.3589, -0.0046, 0.0039, -0.36, -0.0053, 0.0039, -0.3589, 0.001, -0.0, -0.3547, -0.0063, -0.0, -0.3574, 0.0011, 0.0017, -0.3548, 0.0011, 0.0039, -0.3548, -0.0064, 0.0017, -0.3572, -0.0064, 0.0039, -0.3572, -0.0081, -0.0, -0.3615, -0.0103, -0.0, -0.3615, -0.0081, -0.0, -0.3594, -0.0103, -0.0, -0.3594, -0.0125, -0.0, -0.3615, -0.0146, -0.0, -0.3615, -0.0125, -0.0, -0.3594, -0.0146, -0.0, -0.3594, -0.0069, -0.0, -0.3563, -0.0103, -0.0, -0.3572, -0.0079, -0.0, -0.3547, -0.0103, -0.0, -0.355, -0.0125, -0.0, -0.3572,
-0.0146, -0.0, -0.3572, -0.0125, -0.0, -0.355, -0.0146, -0.0, -0.355, -0.0071, 0.0017, -0.3561, -0.0071, 0.0039, -0.3561, -0.0077, 0.0017, -0.355, -0.0077, 0.0039, -0.355, -0.0, -0.0, -0.3531, -0.0007, -0.0, -0.352, 0.0005, -0.0, -0.3507, -0.0014, -0.0, -0.3509, 0.0, 0.0017, -0.3531, -0.0007, 0.0017, -0.352, 0.0, 0.0039, -0.3531, -0.0007, 0.0039, -0.352, -0.0015, 0.0017, -0.3507, -0.0015, 0.0039, -0.3507, 0.0005, -0.0, -0.3485, -0.0025, -0.0, -0.3491, 0.0005, -0.0, -0.3463, -0.0016, -0.0, -0.3463, -0.0032, -0.0, -0.348, -0.004, -0.0, -0.3467, -0.0025, 0.0017, -0.3492, -0.0025, 0.0039, -0.3492, -0.0031, 0.0017, -0.3481, -0.0031, 0.0039, -0.3481, -0.0042, 0.0017, -0.3463, -0.0042, 0.0039, -0.3463, -0.0088, -0.0, -0.3534,
-0.0094, -0.0, -0.3523, -0.0105, -0.0, -0.3505, -0.0125, -0.0, -0.3529, -0.0146, -0.0, -0.3529, -0.0125, -0.0, -0.3507, -0.0146, -0.0, -0.3507, -0.0088, 0.0017, -0.3533, -0.0095, 0.0017, -0.3522, -0.0088, 0.0039, -0.3533, -0.0095, 0.0039, -0.3522, -0.0104, 0.0017, -0.3507, -0.0104, 0.0039, -0.3507, -0.0112, -0.0, -0.3494, -0.0119, -0.0, -0.3483, -0.0146, -0.0, -0.3485, -0.0129, -0.0, -0.3465, -0.0134, -0.0, -0.3456, -0.0112, 0.0017, -0.3493, -0.0112, 0.0039, -0.3493, -0.0119, 0.0017, -0.3483, -0.0119, 0.0039, -0.3483, -0.013, 0.0017, -0.3464, -0.0136, 0.0017, -0.3453, -0.013, 0.0039, -0.3464, -0.0136, 0.0039, -0.3453, -0.0168, -0.0, -0.3615, -0.0192, -0.0, -0.3619, -0.0168, -0.0, -0.3594, -0.0202, -0.0, -0.3601, -0.0207, -0.0, -0.3591,
-0.0194, 0.0017, -0.3615, -0.0194, 0.0039, -0.3615, -0.02, 0.0017, -0.3604, -0.02, 0.0039, -0.3604, -0.0206, 0.0017, -0.3593, -0.0206, 0.0039, -0.3593, -0.0168, -0.0, -0.3572, -0.019, -0.0, -0.3572, -0.0168, -0.0, -0.355, -0.019, -0.0, -0.355, -0.0217, -0.0, -0.3573, -0.0223, -0.0, -0.3562, -0.0212, -0.0, -0.355, -0.0228, -0.0, -0.3551, -0.0217, 0.0017, -0.3573, -0.0223, 0.0017, -0.3562, -0.0217, 0.0039, -0.3573, -0.0223, 0.0039, -0.3562, -0.0228, 0.0017, -0.355, -0.0228, 0.0039, -0.355, -0.0168, -0.0, -0.3529, -0.019, -0.0, -0.3529, -0.0168, -0.0, -0.3507, -0.019, -0.0, -0.3507, -0.0212, -0.0, -0.3529, -0.0237, -0.0, -0.3531, -0.0212, -0.0, -0.3507, -0.0245, -0.0, -0.3513, -0.0238, 0.0017, -0.3529, -0.0238, 0.0039, -0.3529,
-0.0243, 0.0017, -0.3516, -0.0243, 0.0039, -0.3516, -0.0168, -0.0, -0.3485, -0.019, -0.0, -0.3485, -0.0168, -0.0, -0.3463, -0.019, -0.0, -0.3463, -0.0212, -0.0, -0.3485, -0.0233, -0.0, -0.3485, -0.0212, -0.0, -0.3463, -0.0233, -0.0, -0.3463, -0.0249, -0.0, -0.3503, -0.0248, 0.0017, -0.3505, -0.0248, 0.0039, -0.3505, -0.0256, -0.0, -0.3488, -0.0265, -0.0, -0.3468, -0.0269, -0.0, -0.3458, -0.0257, 0.0017, -0.3485, -0.0257, 0.0039, -0.3485, -0.0264, 0.0017, -0.347, -0.0269, 0.0017, -0.3459, -0.0264, 0.0039, -0.347, -0.0269, 0.0039, -0.3459, -0.0633, -0.0, -0.3715, -0.0653, -0.0, -0.3738, -0.0664, -0.0, -0.3747, -0.0641, -0.0, -0.3727, -0.0667, -0.0, -0.3724, -0.0633, 0.0017, -0.3715, -0.0633, 0.0039, -0.3715, -0.0653, 0.0017, -0.3738,
-0.0667, 0.0017, -0.3749, -0.0653, 0.0039, -0.3738, -0.0667, 0.0039, -0.3749, -0.0641, 0.0017, -0.3727, -0.0641, 0.0039, -0.3727, -0.0626, -0.0, -0.3704, -0.0619, -0.0, -0.3682, -0.0646, -0.0, -0.3702, -0.0667, -0.0, -0.3702, -0.0646, -0.0, -0.368, -0.0667, -0.0, -0.368, -0.0625, 0.0017, -0.3703, -0.0625, 0.0039, -0.3703, -0.0619, 0.0017, -0.3681, -0.0619, 0.0039, -0.3681, -0.0617, -0.0, -0.3659, -0.062, -0.0, -0.3635, -0.0646, -0.0, -0.3659, -0.0667, -0.0, -0.3659, -0.0646, -0.0, -0.3637, -0.0667, -0.0, -0.3637, -0.0617, 0.0017, -0.3659, -0.0617, 0.0039, -0.3659, -0.062, 0.0017, -0.3636, -0.062, 0.0039, -0.3636, -0.0408, -0.0, -0.3549, -0.0406, 0.0017, -0.3548, -0.0406, 0.0039, -0.3548, -0.0439, -0.0, -0.3561, -0.045, -0.0, -0.3563,
-0.0428, -0.0, -0.3558, -0.045, -0.0, -0.355, -0.0468, -0.0, -0.3564, -0.0476, -0.0, -0.3562, -0.0497, -0.0, -0.3558, -0.0439, 0.0017, -0.3561, -0.045, 0.0017, -0.3563, -0.0439, 0.0039, -0.3561, -0.045, 0.0039, -0.3563, -0.0427, 0.0017, -0.3558, -0.0427, 0.0039, -0.3558, -0.0472, 0.0017, -0.3563, -0.0472, 0.0039, -0.3563, -0.0483, 0.0017, -0.3561, -0.0494, 0.0017, -0.3559, -0.0483, 0.0039, -0.3561, -0.0494, 0.0039, -0.3559, -0.0372, -0.0, -0.3501, -0.0386, -0.0, -0.3529, -0.0397, -0.0, -0.354, -0.0376, -0.0, -0.3511, -0.0407, -0.0, -0.3507, -0.0372, 0.0017, -0.3502, -0.0372, 0.0039, -0.3502, -0.0389, 0.0017, -0.3533, -0.0396, 0.0017, -0.3539, -0.0389, 0.0039, -0.3533, -0.0396, 0.0039, -0.3539, -0.0377, 0.0017, -0.3513, -0.0377, 0.0039, -0.3513,
-0.0367, -0.0, -0.3486, -0.0364, -0.0, -0.3464, -0.0385, -0.0, -0.3485, -0.0407, -0.0, -0.3485, -0.0385, -0.0, -0.3463, -0.0407, -0.0, -0.3463, -0.0366, 0.0017, -0.3485, -0.0366, 0.0039, -0.3485, -0.0364, 0.0017, -0.3464, -0.0364, 0.0039, -0.3464, -0.0429, -0.0, -0.3529, -0.045, -0.0, -0.3529, -0.0429, -0.0, -0.3507, -0.045, -0.0, -0.3507, -0.0472, -0.0, -0.3529, -0.0494, -0.0, -0.3529, -0.0472, -0.0, -0.3507, -0.0494, -0.0, -0.3507, -0.0429, -0.0, -0.3485, -0.045, -0.0, -0.3485, -0.0429, -0.0, -0.3463, -0.045, -0.0, -0.3463, -0.0472, -0.0, -0.3485, -0.0494, -0.0, -0.3485, -0.0472, -0.0, -0.3463, -0.0494, -0.0, -0.3463, -0.0515, -0.0, -0.3549, -0.0516, 0.0017, -0.3548, -0.0516, 0.0039, -0.3548, -0.0628, -0.0, -0.3618,
-0.0622, -0.0, -0.3625, -0.0661, -0.0, -0.3607, -0.0645, -0.0, -0.3593, -0.0664, -0.0, -0.3591, -0.0629, 0.0017, -0.3617, -0.0629, 0.0039, -0.3617, -0.0622, 0.0017, -0.3625, -0.0665, 0.0017, -0.3605, -0.0622, 0.0039, -0.3625, -0.0665, 0.0039, -0.3605, -0.0647, 0.0017, -0.3591, -0.0663, 0.0017, -0.3592, -0.0647, 0.0039, -0.3591, -0.0663, 0.0039, -0.3592, -0.0671, -0.0, -0.3577, -0.0669, 0.0017, -0.3578, -0.0669, 0.0039, -0.3578, -0.0525, -0.0, -0.354, -0.0538, -0.0, -0.353, -0.0515, -0.0, -0.3507, -0.0547, -0.0, -0.3512, -0.0551, -0.0, -0.3501, -0.0526, 0.0017, -0.3539, -0.0536, 0.0017, -0.3532, -0.0526, 0.0039, -0.3539, -0.0536, 0.0039, -0.3532, -0.0546, 0.0017, -0.3514, -0.0546, 0.0039, -0.3514, -0.0551, 0.0017, -0.3503, -0.0551, 0.0039, -0.3503,
-0.0515, -0.0, -0.3485, -0.0537, -0.0, -0.3485, -0.0515, -0.0, -0.3463, -0.0537, -0.0, -0.3463, -0.0554, -0.0, -0.3486, -0.0555, -0.0, -0.3463, -0.0554, 0.0017, -0.3485, -0.0554, 0.0039, -0.3485, -0.0555, 0.0017, -0.3463, -0.0555, 0.0039, -0.3463, -0.0693, -0.0, -0.4139, -0.0699, -0.0, -0.4129, -0.0708, -0.0, -0.4111, -0.0738, -0.0, -0.414, -0.0745, -0.0, -0.4133, -0.074, -0.0, -0.4118, -0.0749, -0.0, -0.411, -0.0694, 0.0017, -0.4137, -0.07, 0.0017, -0.4126, -0.0694, 0.0039, -0.4137, -0.07, 0.0039, -0.4126, -0.0706, 0.0017, -0.4115, -0.0706, 0.0039, -0.4115, -0.0739, 0.0017, -0.4136, -0.0746, 0.0017, -0.4131, -0.0739, 0.0039, -0.4136, -0.0746, 0.0039, -0.4131, -0.074, 0.0017, -0.4119, -0.0747, 0.0017, -0.4114, -0.074, 0.0039, -0.4119,
-0.0747, 0.0039, -0.4114, -0.0716, -0.0, -0.4095, -0.072, -0.0, -0.4086, -0.0757, -0.0, -0.4095, -0.0728, -0.0, -0.4068, -0.0754, -0.0, -0.4071, -0.0717, 0.0017, -0.4093, -0.0717, 0.0039, -0.4093, -0.0722, 0.0017, -0.4082, -0.0722, 0.0039, -0.4082, -0.0727, 0.0017, -0.4071, -0.0727, 0.0039, -0.4071, -0.0776, -0.0, -0.4136, -0.0798, -0.0, -0.4136, -0.0776, -0.0, -0.4115, -0.0798, -0.0, -0.4115, -0.0819, -0.0, -0.4136, -0.0841, -0.0, -0.4136, -0.0819, -0.0, -0.4115, -0.0841, -0.0, -0.4115, -0.0772, -0.0, -0.4086, -0.0798, -0.0, -0.4093, -0.0784, -0.0, -0.4081, -0.0794, -0.0, -0.4069, -0.0819, -0.0, -0.4093, -0.0841, -0.0, -0.4093, -0.0816, -0.0, -0.4061, -0.0841, -0.0, -0.4071, -0.0797, 0.0017, -0.4068, -0.0797, 0.0039, -0.4068,
-0.0814, 0.0017, -0.4062, -0.0814, 0.0039, -0.4062, -0.0735, -0.0, -0.4048, -0.0754, -0.0, -0.4049, -0.074, -0.0, -0.4033, -0.0743, -0.0, -0.4024, -0.0735, 0.0017, -0.4049, -0.0735, 0.0039, -0.4049, -0.0741, 0.0017, -0.4032, -0.0745, 0.0017, -0.4021, -0.0741, 0.0039, -0.4032, -0.0745, 0.0039, -0.4021, -0.075, -0.0, -0.4004, -0.0757, -0.0, -0.3983, -0.0749, 0.0017, -0.4006, -0.0749, 0.0039, -0.4006, -0.0757, 0.0017, -0.3984, -0.0757, 0.0039, -0.3984, -0.0776, -0.0, -0.4049, -0.0802, -0.0, -0.4057, -0.0776, -0.0, -0.4028, -0.0798, -0.0, -0.4028, -0.0823, -0.0, -0.4055, -0.0842, -0.0, -0.4054, -0.0819, -0.0, -0.4028, -0.0841, -0.0, -0.4028, -0.0803, 0.0017, -0.4057, -0.0803, 0.0039, -0.4057, -0.0819, 0.0017, -0.4056, -0.0819, 0.0039, -0.4056,
-0.0776, -0.0, -0.4006, -0.0798, -0.0, -0.4006, -0.0776, -0.0, -0.3984, -0.0798, -0.0, -0.3984, -0.0819, -0.0, -0.4006, -0.0841, -0.0, -0.4006, -0.0819, -0.0, -0.3984, -0.0841, -0.0, -0.3984, -0.0863, -0.0, -0.4136, -0.0884, -0.0, -0.4136, -0.0863, -0.0, -0.4115, -0.0884, -0.0, -0.4115, -0.0906, -0.0, -0.4136, -0.0928, -0.0, -0.4136, -0.0906, -0.0, -0.4115, -0.0928, -0.0, -0.4115, -0.0863, -0.0, -0.4093, -0.0884, -0.0, -0.4093, -0.0863, -0.0, -0.4071, -0.0884, -0.0, -0.4071, -0.0906, -0.0, -0.4093, -0.0928, -0.0, -0.4093, -0.0906, -0.0, -0.4071, -0.0929, -0.0, -0.4064, -0.0928, 0.0017, -0.4064, -0.0928, 0.0039, -0.4064, -0.0949, -0.0, -0.4136, -0.0971, -0.0, -0.4136, -0.0949, -0.0, -0.4115, -0.0981, -0.0, -0.4107,
-0.0997, -0.0, -0.4133, -0.0988, -0.0, -0.4118, -0.098, 0.0017, -0.4106, -0.098, 0.0039, -0.4106, -0.0998, 0.0017, -0.4136, -0.0998, 0.0039, -0.4136, -0.0987, 0.0017, -0.4117, -0.0987, 0.0039, -0.4117, -0.0957, -0.0, -0.4083, -0.097, -0.0, -0.4094, -0.0946, -0.0, -0.4075, -0.0958, 0.0017, -0.4084, -0.097, 0.0017, -0.4094, -0.0958, 0.0039, -0.4084, -0.097, 0.0039, -0.4094, -0.0947, 0.0017, -0.4075, -0.0947, 0.0039, -0.4075, -0.0862, -0.0, -0.4045, -0.0887, -0.0, -0.4045, -0.0863, -0.0, -0.4028, -0.089, -0.0, -0.4029, -0.0907, -0.0, -0.4056, -0.0917, -0.0, -0.406, -0.0887, 0.0017, -0.4045, -0.0887, 0.0039, -0.4045, -0.089, 0.0017, -0.4028, -0.089, 0.0039, -0.4028, -0.0906, 0.0017, -0.4056, -0.0917, 0.0017, -0.406, -0.0906, 0.0039, -0.4056,
-0.0917, 0.0039, -0.406, -0.0863, -0.0, -0.4006, -0.0893, -0.0, -0.4005, -0.0863, -0.0, -0.3984, -0.0884, -0.0, -0.3984, -0.0895, -0.0, -0.3995, -0.0897, -0.0, -0.3984, -0.0893, 0.0017, -0.4006, -0.0893, 0.0039, -0.4006, -0.0895, 0.0017, -0.3995, -0.0895, 0.0039, -0.3995, -0.0897, 0.0017, -0.3984, -0.0897, 0.0039, -0.3984, -0.0762, -0.0, -0.3968, -0.0763, 0.0017, -0.3966, -0.0763, 0.0039, -0.3966, -0.0765, -0.0, -0.3959, -0.0798, -0.0, -0.3963, -0.077, -0.0, -0.3939, -0.0798, -0.0, -0.3941, -0.0819, -0.0, -0.3963, -0.0841, -0.0, -0.3963, -0.0819, -0.0, -0.3941, -0.0841, -0.0, -0.3941, -0.0766, 0.0017, -0.3955, -0.0766, 0.0039, -0.3955, -0.077, 0.0017, -0.3941, -0.077, 0.0039, -0.3941, -0.0775, -0.0, -0.3918, -0.0798, -0.0, -0.3919,
-0.0779, -0.0, -0.3897, -0.0798, -0.0, -0.3897, -0.0819, -0.0, -0.3919, -0.0841, -0.0, -0.3919, -0.0819, -0.0, -0.3897, -0.0841, -0.0, -0.3897, -0.0775, 0.0017, -0.3919, -0.0775, 0.0039, -0.3919, -0.0779, 0.0017, -0.3898, -0.0779, 0.0039, -0.3898, -0.0783, -0.0, -0.3875, -0.0798, -0.0, -0.3876, -0.0785, -0.0, -0.3861, -0.0787, -0.0, -0.3852, -0.0819, -0.0, -0.3876, -0.0841, -0.0, -0.3876, -0.0819, -0.0, -0.3854, -0.0841, -0.0, -0.3854, -0.0783, 0.0017, -0.3876, -0.0783, 0.0039, -0.3876, -0.0786, 0.0017, -0.3858, -0.0787, 0.0017, -0.3848, -0.0786, 0.0039, -0.3858, -0.0787, 0.0039, -0.3848, -0.0789, -0.0, -0.3832, -0.079, -0.0, -0.381, -0.0819, -0.0, -0.3832, -0.0841, -0.0, -0.3832, -0.0819, -0.0, -0.3811, -0.0841, -0.0, -0.3811,
-0.0789, 0.0017, -0.3832, -0.0789, 0.0039, -0.3832, -0.079, 0.0017, -0.3811, -0.079, 0.0039, -0.3811, -0.0863, -0.0, -0.3963, -0.0884, -0.0, -0.3963, -0.0863, -0.0, -0.3941, -0.0884, -0.0, -0.3941, -0.0901, -0.0, -0.3963, -0.0904, -0.0, -0.3942, -0.0901, 0.0017, -0.3963, -0.0901, 0.0039, -0.3963, -0.0904, 0.0017, -0.3941, -0.0904, 0.0039, -0.3941, -0.0863, -0.0, -0.3919, -0.0884, -0.0, -0.3919, -0.0863, -0.0, -0.3897, -0.0884, -0.0, -0.3897, -0.0908, -0.0, -0.392, -0.0911, -0.0, -0.3899, -0.0908, 0.0017, -0.3919, -0.0908, 0.0039, -0.3919, -0.0912, 0.0017, -0.3898, -0.0912, 0.0039, -0.3898, -0.0863, -0.0, -0.3876, -0.0884, -0.0, -0.3876, -0.0863, -0.0, -0.3854, -0.0884, -0.0, -0.3854, -0.0915, -0.0, -0.3877, -0.0917, -0.0, -0.3856,
-0.0918, -0.0, -0.3848, -0.0915, 0.0017, -0.3876, -0.0915, 0.0039, -0.3876, -0.0917, 0.0017, -0.3861, -0.0917, 0.0017, -0.3851, -0.0917, 0.0039, -0.3861, -0.0917, 0.0039, -0.3851, -0.0863, -0.0, -0.3832, -0.0884, -0.0, -0.3832, -0.0863, -0.0, -0.3811, -0.0884, -0.0, -0.3811, -0.0906, -0.0, -0.3832, -0.0918, -0.0, -0.3832, -0.0906, -0.0, -0.3811, -0.0918, -0.0, -0.3811, -0.0918, 0.0017, -0.3832, -0.0918, 0.0039, -0.3832, -0.0918, 0.0017, -0.3811, -0.0918, 0.0039, -0.3811, -0.0954, -0.0, -0.3849, -0.0967, -0.0, -0.386, -0.0991, -0.0, -0.3867, -0.1015, -0.0, -0.3862, -0.0954, 0.0017, -0.3849, -0.0969, 0.0017, -0.3861, -0.0954, 0.0039, -0.3849, -0.0969, 0.0039, -0.3861, -0.0991, 0.0017, -0.3867, -0.1015, 0.0017, -0.3862, -0.0991, 0.0039, -0.3867,
-0.1015, 0.0039, -0.3862, -0.0941, -0.0, -0.3832, -0.0971, -0.0, -0.3832, -0.0943, -0.0, -0.3808, -0.0971, -0.0, -0.3811, -0.0993, -0.0, -0.3832, -0.1015, -0.0, -0.3832, -0.0993, -0.0, -0.3811, -0.1015, -0.0, -0.3811, -0.0941, 0.0017, -0.3832, -0.0941, 0.0039, -0.3832, -0.0944, 0.0017, -0.3804, -0.0944, 0.0039, -0.3804, -0.1029, -0.0, -0.3847, -0.1029, 0.0017, -0.3848, -0.1029, 0.0039, -0.3848, -0.1037, -0.0, -0.3834, -0.1039, -0.0, -0.381, -0.1037, 0.0017, -0.3832, -0.1037, 0.0039, -0.3832, -0.1039, 0.0017, -0.381, -0.1039, 0.0039, -0.381, -0.0695, -0.0, -0.3759, -0.071, -0.0, -0.3761, -0.0728, -0.0, -0.376, -0.0693, 0.0017, -0.3758, -0.0711, 0.0017, -0.3761, -0.0693, 0.0039, -0.3758, -0.0711, 0.0039, -0.3761, -0.0731, 0.0017, -0.376,
-0.0731, 0.0039, -0.376, -0.0685, -0.0, -0.3757, -0.0711, -0.0, -0.3746, -0.0689, -0.0, -0.3724, -0.0711, -0.0, -0.3724, -0.0738, -0.0, -0.3758, -0.0758, -0.0, -0.3749, -0.0732, -0.0, -0.3724, -0.0754, -0.0, -0.3724, -0.0682, 0.0017, -0.3755, -0.0682, 0.0039, -0.3755, -0.0742, 0.0017, -0.3756, -0.0755, 0.0017, -0.3751, -0.0742, 0.0039, -0.3756, -0.0755, 0.0039, -0.3751, -0.0791, -0.0, -0.3788, -0.0793, -0.0, -0.3766, -0.0819, -0.0, -0.3789, -0.0841, -0.0, -0.3789, -0.0819, -0.0, -0.3767, -0.0841, -0.0, -0.3767, -0.0791, 0.0017, -0.3789, -0.0791, 0.0039, -0.3789, -0.0792, 0.0017, -0.3767, -0.0792, 0.0039, -0.3767, -0.0772, -0.0, -0.3741, -0.0799, -0.0, -0.3744, -0.0784, -0.0, -0.3731, -0.0799, -0.0, -0.3725, -0.0819, -0.0, -0.3746,
-0.0841, -0.0, -0.3746, -0.0809, -0.0, -0.372, -0.0841, -0.0, -0.3724, -0.0773, 0.0017, -0.374, -0.0798, 0.0017, -0.3745, -0.0773, 0.0039, -0.374, -0.0798, 0.0039, -0.3745, -0.0783, 0.0017, -0.3732, -0.0799, 0.0017, -0.3724, -0.0783, 0.0039, -0.3732, -0.0799, 0.0039, -0.3724, -0.081, 0.0017, -0.3717, -0.081, 0.0039, -0.3717, -0.0689, -0.0, -0.3702, -0.0711, -0.0, -0.3702, -0.0689, -0.0, -0.368, -0.0711, -0.0, -0.368, -0.0732, -0.0, -0.3702, -0.0754, -0.0, -0.3702, -0.0732, -0.0, -0.368, -0.0754, -0.0, -0.368, -0.0689, -0.0, -0.3659, -0.0711, -0.0, -0.3659, -0.0689, -0.0, -0.3637, -0.0711, -0.0, -0.3637, -0.0732, -0.0, -0.3659, -0.0754, -0.0, -0.3659, -0.0732, -0.0, -0.3637, -0.0754, -0.0, -0.3637, -0.0776, -0.0, -0.3702,
-0.0802, -0.0, -0.3705, -0.0776, -0.0, -0.368, -0.0808, -0.0, -0.3682, -0.0817, -0.0, -0.37, -0.0841, -0.0, -0.3702, -0.0819, -0.0, -0.3678, -0.0841, -0.0, -0.368, -0.0803, 0.0017, -0.3703, -0.0803, 0.0039, -0.3703, -0.0808, 0.0017, -0.3686, -0.0808, 0.0039, -0.3686, -0.0816, 0.0017, -0.3702, -0.0816, 0.0039, -0.3702, -0.0818, 0.0017, -0.368, -0.0818, 0.0039, -0.368, -0.0776, -0.0, -0.3659, -0.0809, -0.0, -0.3657, -0.0776, -0.0, -0.3637, -0.0807, -0.0, -0.3634, -0.082, -0.0, -0.366, -0.0841, -0.0, -0.3659, -0.0821, -0.0, -0.3638, -0.0841, -0.0, -0.3637, -0.0808, 0.0017, -0.365, -0.0808, 0.0039, -0.365, -0.0807, 0.0017, -0.3637, -0.0807, 0.0039, -0.3637, -0.082, 0.0017, -0.367, -0.082, 0.0039, -0.367, -0.0821, 0.0017, -0.3637,
-0.0821, 0.0039, -0.3637, -0.0863, -0.0, -0.3789, -0.0884, -0.0, -0.3789, -0.0863, -0.0, -0.3767, -0.0884, -0.0, -0.3767, -0.0906, -0.0, -0.3789, -0.0918, -0.0, -0.3789, -0.0906, -0.0, -0.3767, -0.0918, -0.0, -0.3767, -0.0918, 0.0017, -0.3789, -0.0918, 0.0039, -0.3789, -0.0918, 0.0017, -0.3767, -0.0918, 0.0039, -0.3767, -0.0863, -0.0, -0.3746, -0.0884, -0.0, -0.3746, -0.0863, -0.0, -0.3724, -0.0884, -0.0, -0.3724, -0.0906, -0.0, -0.3746, -0.0918, -0.0, -0.3746, -0.0906, -0.0, -0.3724, -0.0918, -0.0, -0.3724, -0.0918, 0.0017, -0.3746, -0.0918, 0.0039, -0.3746, -0.0918, 0.0017, -0.3724, -0.0918, 0.0039, -0.3724, -0.0951, -0.0, -0.379, -0.0963, -0.0, -0.3779, -0.0974, -0.0, -0.3773, -0.0993, -0.0, -0.3789, -0.102, -0.0, -0.3782,
-0.0991, -0.0, -0.3769, -0.1009, -0.0, -0.3774, -0.0952, 0.0017, -0.3789, -0.0962, 0.0017, -0.3779, -0.0952, 0.0039, -0.3789, -0.0962, 0.0039, -0.3779, -0.0972, 0.0017, -0.3773, -0.0972, 0.0039, -0.3773, -0.1021, 0.0017, -0.3782, -0.1021, 0.0039, -0.3782, -0.0991, 0.0017, -0.3769, -0.1014, 0.0017, -0.3776, -0.0991, 0.0039, -0.3769, -0.1014, 0.0039, -0.3776, -0.0863, -0.0, -0.3702, -0.0884, -0.0, -0.3702, -0.0863, -0.0, -0.368, -0.0884, -0.0, -0.368, -0.0916, -0.0, -0.3703, -0.0917, -0.0, -0.3712, -0.0913, -0.0, -0.3679, -0.0916, 0.0017, -0.3701, -0.0917, 0.0017, -0.3712, -0.0916, 0.0039, -0.3701, -0.0917, 0.0039, -0.3712, -0.0913, 0.0017, -0.368, -0.0913, 0.0039, -0.368, -0.0863, -0.0, -0.3659, -0.0884, -0.0, -0.3659, -0.0863, -0.0, -0.3637,
-0.0884, -0.0, -0.3637, -0.0908, -0.0, -0.3658, -0.0904, -0.0, -0.3636, -0.0909, 0.0017, -0.3659, -0.0909, 0.0039, -0.3659, -0.0905, 0.0017, -0.3637, -0.0905, 0.0039, -0.3637, -0.0682, -0.0, -0.3603, -0.0711, -0.0, -0.3615, -0.0691, -0.0, -0.3596, -0.0703, -0.0, -0.3587, -0.0732, -0.0, -0.3615, -0.0754, -0.0, -0.3615, -0.0732, -0.0, -0.3594, -0.0763, -0.0, -0.3582, -0.0679, 0.0017, -0.3605, -0.0679, 0.0039, -0.3605, -0.0704, 0.0017, -0.3588, -0.0704, 0.0039, -0.3588, -0.0765, 0.0017, -0.3583, -0.0765, 0.0039, -0.3583, -0.0687, -0.0, -0.3571, -0.071, -0.0, -0.3569, -0.0733, -0.0, -0.3571, -0.0752, -0.0, -0.3575, -0.0689, 0.0017, -0.3571, -0.0711, 0.0017, -0.3569, -0.0689, 0.0039, -0.3571, -0.0711, 0.0039, -0.3569, -0.0733, 0.0017, -0.357,
-0.0757, 0.0017, -0.3577, -0.0733, 0.0039, -0.357, -0.0757, 0.0039, -0.3577, -0.0776, -0.0, -0.3615, -0.0803, -0.0, -0.3611, -0.078, -0.0, -0.3594, -0.0795, -0.0, -0.3599, -0.0812, -0.0, -0.3618, -0.0841, -0.0, -0.3615, -0.0819, -0.0, -0.3594, -0.0841, -0.0, -0.3594, -0.0805, 0.0017, -0.3614, -0.0805, 0.0039, -0.3614, -0.0777, 0.0017, -0.3591, -0.0793, 0.0017, -0.3596, -0.0777, 0.0039, -0.3591, -0.0793, 0.0039, -0.3596, -0.0812, 0.0017, -0.3619, -0.0812, 0.0039, -0.3619, -0.0781, -0.0, -0.3569, -0.0787, -0.0, -0.3579, -0.0773, -0.0, -0.3553, -0.0798, -0.0, -0.355, -0.0819, -0.0, -0.3572, -0.0841, -0.0, -0.3572, -0.0819, -0.0, -0.355, -0.0841, -0.0, -0.355, -0.0781, 0.0017, -0.357, -0.0788, 0.0017, -0.3581, -0.0781, 0.0039, -0.357,
-0.0788, 0.0039, -0.3581, -0.0772, 0.0017, -0.3551, -0.0772, 0.0039, -0.3551, -0.0764, -0.0, -0.352, -0.076, -0.0, -0.3508, -0.0764, 0.0017, -0.3521, -0.0764, 0.0039, -0.3521, -0.076, 0.0017, -0.3507, -0.076, 0.0039, -0.3507, -0.0754, -0.0, -0.3487, -0.0742, -0.0, -0.3454, -0.0746, -0.0, -0.3465, -0.0753, 0.0017, -0.3485, -0.0753, 0.0039, -0.3485, -0.0742, 0.0017, -0.3455, -0.0746, 0.0017, -0.3466, -0.0742, 0.0039, -0.3455, -0.0746, 0.0039, -0.3466, -0.0767, -0.0, -0.353, -0.0798, -0.0, -0.3529, -0.0776, -0.0, -0.3507, -0.0798, -0.0, -0.3507, -0.0819, -0.0, -0.3529, -0.0841, -0.0, -0.3529, -0.0819, -0.0, -0.3507, -0.0841, -0.0, -0.3507, -0.0767, 0.0017, -0.3532, -0.0767, 0.0039, -0.3532, -0.0776, -0.0, -0.3485, -0.0798, -0.0, -0.3485,
-0.0776, -0.0, -0.3463, -0.0798, -0.0, -0.3463, -0.0819, -0.0, -0.3485, -0.0841, -0.0, -0.3485, -0.0819, -0.0, -0.3463, -0.0841, -0.0, -0.3463, -0.0863, -0.0, -0.3615, -0.0884, -0.0, -0.3615, -0.0863, -0.0, -0.3594, -0.0896, -0.0, -0.359, -0.09, -0.0, -0.3615, -0.0898, -0.0, -0.3597, -0.0895, 0.0017, -0.3583, -0.0895, 0.0039, -0.3583, -0.0901, 0.0017, -0.3615, -0.0901, 0.0039, -0.3615, -0.0897, 0.0017, -0.3594, -0.0897, 0.0039, -0.3594, -0.0863, -0.0, -0.3572, -0.0893, -0.0, -0.357, -0.0863, -0.0, -0.355, -0.0889, -0.0, -0.3549, -0.0893, 0.0017, -0.3572, -0.0893, 0.0039, -0.3572, -0.0889, 0.0017, -0.355, -0.0889, 0.0039, -0.355, -0.0863, -0.0, -0.3529, -0.0884, -0.0, -0.3527, -0.0863, -0.0, -0.3507, -0.0879, -0.0, -0.3506,
-0.0885, 0.0017, -0.3529, -0.0885, 0.0039, -0.3529, -0.0879, 0.0017, -0.3507, -0.0879, 0.0039, -0.3507, -0.0871, -0.0, -0.3483, -0.0875, -0.0, -0.3494, -0.0864, -0.0, -0.3461, -0.0871, 0.0017, -0.3482, -0.0875, 0.0017, -0.3493, -0.0871, 0.0039, -0.3482, -0.0875, 0.0039, -0.3493, -0.0864, 0.0017, -0.3463, -0.0864, 0.0039, -0.3463, -0.1031, -0.0, -0.3791, -0.1031, 0.0017, -0.3791, -0.1031, 0.0039, -0.3791, 0.0005, -0.0, -0.3442, -0.0016, -0.0, -0.3442, 0.0005, -0.0, -0.342, -0.0016, -0.0, -0.342, -0.005, -0.0, -0.345, -0.0056, -0.0, -0.344, -0.0038, -0.0, -0.342, -0.0066, -0.0, -0.3423, -0.0049, 0.0017, -0.3452, -0.0055, 0.0017, -0.3442, -0.0049, 0.0039, -0.3452, -0.0055, 0.0039, -0.3442, -0.0066, 0.0017, -0.3423, -0.0066, 0.0039, -0.3423,
0.0005, -0.0, -0.3398, -0.0016, -0.0, -0.3398, 0.0005, -0.0, -0.3377, -0.0016, -0.0, -0.3377, -0.0038, -0.0, -0.3398, -0.006, -0.0, -0.3398, -0.0038, -0.0, -0.3377, -0.006, -0.0, -0.3377, -0.0072, -0.0, -0.3412, -0.0144, -0.0, -0.3438, -0.0152, -0.0, -0.3424, -0.0072, 0.0017, -0.3412, -0.0072, 0.0039, -0.3412, -0.0142, 0.0017, -0.3442, -0.0142, 0.0039, -0.3442, -0.0153, 0.0017, -0.3423, -0.0153, 0.0039, -0.3423, -0.0079, -0.0, -0.34, -0.009, -0.0, -0.338, -0.0095, -0.0, -0.337, -0.008, 0.0017, -0.3398, -0.008, 0.0039, -0.3398, -0.0089, 0.0017, -0.3381, -0.0095, 0.0017, -0.3371, -0.0089, 0.0039, -0.3381, -0.0095, 0.0039, -0.3371, 0.0005, -0.0, -0.3355, -0.0016, -0.0, -0.3355, 0.0005, -0.0, -0.3333, -0.0016, -0.0, -0.3333,
-0.0038, -0.0, -0.3355, -0.006, -0.0, -0.3355, -0.0038, -0.0, -0.3333, -0.006, -0.0, -0.3333, 0.0005, -0.0, -0.3312, -0.0016, -0.0, -0.3312, 0.0005, -0.0, -0.329, -0.0016, -0.0, -0.329, -0.0038, -0.0, -0.3312, -0.006, -0.0, -0.3312, -0.0038, -0.0, -0.329, -0.006, -0.0, -0.329, -0.0081, -0.0, -0.3355, -0.0101, -0.0, -0.3357, -0.0081, -0.0, -0.3333, -0.0111, -0.0, -0.3336, -0.0116, -0.0, -0.3326, -0.0102, 0.0017, -0.3355, -0.0102, 0.0039, -0.3355, -0.0111, 0.0017, -0.3337, -0.0111, 0.0039, -0.3337, -0.0116, 0.0017, -0.3327, -0.0116, 0.0039, -0.3327, -0.0081, -0.0, -0.3312, -0.0103, -0.0, -0.3312, -0.0081, -0.0, -0.329, -0.0103, -0.0, -0.329, -0.0122, -0.0, -0.3313, -0.0132, -0.0, -0.3292, -0.0137, -0.0, -0.3282,
-0.0123, 0.0017, -0.3312, -0.0123, 0.0039, -0.3312, -0.0132, 0.0017, -0.3293, -0.0137, 0.0017, -0.3282, -0.0132, 0.0039, -0.3293, -0.0137, 0.0039, -0.3282, -0.0168, -0.0, -0.3442, -0.019, -0.0, -0.3442, -0.0158, -0.0, -0.3414, -0.019, -0.0, -0.342, -0.0212, -0.0, -0.3442, -0.0233, -0.0, -0.3442, -0.0212, -0.0, -0.342, -0.0233, -0.0, -0.342, -0.0159, 0.0017, -0.3412, -0.0159, 0.0039, -0.3412, -0.0168, -0.0, -0.3396, -0.019, -0.0, -0.3398, -0.0175, -0.0, -0.3383, -0.0181, -0.0, -0.3372, -0.0212, -0.0, -0.3398, -0.0233, -0.0, -0.3398, -0.0212, -0.0, -0.3377, -0.0233, -0.0, -0.3377, -0.0167, 0.0017, -0.3398, -0.0167, 0.0039, -0.3398, -0.0176, 0.0017, -0.3382, -0.0181, 0.0017, -0.3371, -0.0176, 0.0039, -0.3382, -0.0181, 0.0039, -0.3371,
-0.0255, -0.0, -0.3442, -0.0276, -0.0, -0.3444, -0.0255, -0.0, -0.342, -0.0285, -0.0, -0.3422, -0.0289, -0.0, -0.3411, -0.0277, 0.0017, -0.3442, -0.0277, 0.0039, -0.3442, -0.0284, 0.0017, -0.3423, -0.0284, 0.0039, -0.3423, -0.0289, 0.0017, -0.3412, -0.0289, 0.0039, -0.3412, -0.0255, -0.0, -0.3398, -0.0277, -0.0, -0.3398, -0.0255, -0.0, -0.3377, -0.0277, -0.0, -0.3377, -0.0294, -0.0, -0.34, -0.0302, -0.0, -0.3379, -0.0294, 0.0017, -0.3398, -0.0294, 0.0039, -0.3398, -0.0303, 0.0017, -0.3377, -0.0303, 0.0039, -0.3377, -0.019, -0.0, -0.3353, -0.0197, -0.0, -0.3339, -0.0212, -0.0, -0.3355, -0.0233, -0.0, -0.3355, -0.0202, -0.0, -0.3329, -0.0233, -0.0, -0.3333, -0.0189, 0.0017, -0.3355, -0.0189, 0.0039, -0.3355, -0.0198, 0.0017, -0.3337,
-0.0198, 0.0039, -0.3337, -0.0203, 0.0017, -0.3327, -0.0203, 0.0039, -0.3327, -0.021, -0.0, -0.3309, -0.0233, -0.0, -0.3312, -0.0218, -0.0, -0.3289, -0.0233, -0.0, -0.329, -0.0209, 0.0017, -0.3312, -0.0209, 0.0039, -0.3312, -0.0218, 0.0017, -0.329, -0.0218, 0.0039, -0.329, -0.0255, -0.0, -0.3355, -0.0277, -0.0, -0.3355, -0.0255, -0.0, -0.3333, -0.0277, -0.0, -0.3333, -0.0309, -0.0, -0.3359, -0.0313, -0.0, -0.335, -0.0298, -0.0, -0.3333, -0.0318, -0.0, -0.3335, -0.0308, 0.0017, -0.3363, -0.0312, 0.0017, -0.3352, -0.0308, 0.0039, -0.3363, -0.0312, 0.0039, -0.3352, -0.0318, 0.0017, -0.3333, -0.0318, 0.0039, -0.3333, -0.0255, -0.0, -0.3312, -0.0277, -0.0, -0.3312, -0.0255, -0.0, -0.329, -0.0277, -0.0, -0.329, -0.0298, -0.0, -0.3312,
-0.0324, -0.0, -0.3314, -0.0298, -0.0, -0.329, -0.0331, -0.0, -0.3293, -0.0325, 0.0017, -0.3312, -0.0325, 0.0039, -0.3312, -0.033, 0.0017, -0.3296, -0.033, 0.0039, -0.3296, 0.0015, -0.0, -0.3262, -0.0016, -0.0, -0.3268, 0.0003, -0.0, -0.3244, -0.0016, -0.0, -0.3246, -0.0038, -0.0, -0.3268, -0.006, -0.0, -0.3268, -0.0038, -0.0, -0.3246, -0.006, -0.0, -0.3246, 0.0014, 0.0017, -0.3261, 0.0014, 0.0039, -0.3261, 0.0004, 0.0017, -0.3246, 0.0004, 0.0039, -0.3246, -0.0004, -0.0, -0.3234, -0.0011, -0.0, -0.3223, -0.0021, -0.0, -0.3207, -0.0038, -0.0, -0.3225, -0.006, -0.0, -0.3225, -0.0027, -0.0, -0.3197, -0.006, -0.0, -0.3203, -0.0004, 0.0017, -0.3234, -0.0011, 0.0017, -0.3223, -0.0004, 0.0039, -0.3234, -0.0011, 0.0039, -0.3223,
-0.0022, 0.0017, -0.3205, -0.0022, 0.0039, -0.3205, -0.0028, 0.0017, -0.3194, -0.0028, 0.0039, -0.3194, -0.0081, -0.0, -0.3268, -0.0103, -0.0, -0.3268, -0.0081, -0.0, -0.3246, -0.0103, -0.0, -0.3246, -0.0125, -0.0, -0.3268, -0.0143, -0.0, -0.327, -0.0125, -0.0, -0.3246, -0.0151, -0.0, -0.3249, -0.0144, 0.0017, -0.3268, -0.0144, 0.0039, -0.3268, -0.0153, 0.0017, -0.3247, -0.0153, 0.0039, -0.3247, -0.0081, -0.0, -0.3225, -0.0103, -0.0, -0.3225, -0.0081, -0.0, -0.3203, -0.0103, -0.0, -0.3203, -0.0125, -0.0, -0.3225, -0.0159, -0.0, -0.323, -0.0125, -0.0, -0.3203, -0.0146, -0.0, -0.3203, -0.0157, 0.0017, -0.3235, -0.0157, 0.0039, -0.3235, -0.0037, -0.0, -0.3179, -0.006, -0.0, -0.3181, -0.0044, -0.0, -0.3165, -0.005, -0.0, -0.3154,
-0.0036, 0.0017, -0.3181, -0.0036, 0.0039, -0.3181, -0.0045, 0.0017, -0.3164, -0.0051, 0.0017, -0.3153, -0.0045, 0.0039, -0.3164, -0.0051, 0.0039, -0.3153, -0.0061, -0.0, -0.3136, -0.0067, -0.0, -0.3123, -0.0059, 0.0017, -0.3138, -0.0059, 0.0039, -0.3138, -0.0068, 0.0017, -0.3122, -0.0068, 0.0039, -0.3122, -0.0081, -0.0, -0.3181, -0.0103, -0.0, -0.3181, -0.0081, -0.0, -0.316, -0.0103, -0.0, -0.316, -0.0125, -0.0, -0.3181, -0.0146, -0.0, -0.3181, -0.0125, -0.0, -0.316, -0.0146, -0.0, -0.316, -0.0081, -0.0, -0.3138, -0.0103, -0.0, -0.3138, -0.0073, -0.0, -0.3113, -0.0103, -0.0, -0.3116, -0.0125, -0.0, -0.3138, -0.0146, -0.0, -0.3138, -0.0125, -0.0, -0.3116, -0.0146, -0.0, -0.3116, -0.0074, 0.0017, -0.3111, -0.0074, 0.0039, -0.3111,
-0.0222, -0.0, -0.3279, -0.0227, -0.0, -0.3268, -0.0236, -0.0, -0.3245, -0.0222, 0.0017, -0.3279, -0.0227, 0.0017, -0.3268, -0.0222, 0.0039, -0.3279, -0.0227, 0.0039, -0.3268, -0.0235, 0.0017, -0.3246, -0.0235, 0.0039, -0.3246, -0.0162, -0.0, -0.3221, -0.0168, -0.0, -0.3205, -0.0241, -0.0, -0.3231, -0.0161, 0.0017, -0.3224, -0.0161, 0.0039, -0.3224, -0.0169, 0.0017, -0.3203, -0.0169, 0.0039, -0.3203, -0.0242, 0.0017, -0.3229, -0.0242, 0.0039, -0.3229, -0.0255, -0.0, -0.3268, -0.0277, -0.0, -0.3268, -0.0255, -0.0, -0.3246, -0.0277, -0.0, -0.3246, -0.0298, -0.0, -0.3268, -0.032, -0.0, -0.3268, -0.0298, -0.0, -0.3246, -0.032, -0.0, -0.3246, -0.0245, -0.0, -0.3221, -0.0277, -0.0, -0.3225, -0.0251, -0.0, -0.3201, -0.0277, -0.0, -0.3203,
-0.0298, -0.0, -0.3225, -0.032, -0.0, -0.3225, -0.0298, -0.0, -0.3203, -0.032, -0.0, -0.3203, -0.0246, 0.0017, -0.3218, -0.0246, 0.0039, -0.3218, -0.0251, 0.0017, -0.3203, -0.0251, 0.0039, -0.3203, -0.0176, -0.0, -0.3183, -0.018, -0.0, -0.3173, -0.0168, -0.0, -0.316, -0.0185, -0.0, -0.3161, -0.0176, 0.0017, -0.3184, -0.018, 0.0017, -0.3173, -0.0176, 0.0039, -0.3184, -0.018, 0.0039, -0.3173, -0.0185, 0.0017, -0.316, -0.0185, 0.0039, -0.316, -0.0168, -0.0, -0.3138, -0.0192, -0.0, -0.314, -0.0168, -0.0, -0.3116, -0.0199, -0.0, -0.3117, -0.0201, -0.0, -0.3107, -0.0192, 0.0017, -0.3138, -0.0192, 0.0039, -0.3138, -0.0198, 0.0017, -0.3118, -0.0198, 0.0039, -0.3118, -0.0201, 0.0017, -0.3107, -0.0201, 0.0039, -0.3107, -0.0257, -0.0, -0.318,
-0.0277, -0.0, -0.3181, -0.0262, -0.0, -0.3164, -0.0265, -0.0, -0.3155, -0.0298, -0.0, -0.3181, -0.032, -0.0, -0.3181, -0.0298, -0.0, -0.316, -0.032, -0.0, -0.316, -0.0257, 0.0017, -0.3181, -0.0257, 0.0039, -0.3181, -0.0263, 0.0017, -0.3162, -0.0267, 0.0017, -0.3151, -0.0263, 0.0039, -0.3162, -0.0267, 0.0039, -0.3151, -0.0271, -0.0, -0.3136, -0.0277, -0.0, -0.3115, -0.0298, -0.0, -0.3138, -0.032, -0.0, -0.3138, -0.0298, -0.0, -0.3116, -0.032, -0.0, -0.3116, -0.0271, 0.0017, -0.3138, -0.0271, 0.0039, -0.3138, -0.0277, 0.0017, -0.3116, -0.0277, 0.0039, -0.3116, -0.0367, -0.0, -0.3441, -0.0372, -0.0, -0.3428, -0.0385, -0.0, -0.3442, -0.0407, -0.0, -0.3442, -0.0377, -0.0, -0.3417, -0.0407, -0.0, -0.342, -0.0367, 0.0017, -0.3441,
-0.0367, 0.0039, -0.3441, -0.0373, 0.0017, -0.3427, -0.0373, 0.0039, -0.3427, -0.0378, 0.0017, -0.3416, -0.0378, 0.0039, -0.3416, -0.0388, -0.0, -0.3401, -0.0401, -0.0, -0.3389, -0.0412, -0.0, -0.3383, -0.0388, 0.0017, -0.3401, -0.04, 0.0017, -0.339, -0.0388, 0.0039, -0.3401, -0.04, 0.0039, -0.339, -0.0411, 0.0017, -0.3384, -0.0411, 0.0039, -0.3384, -0.0429, -0.0, -0.3442, -0.045, -0.0, -0.3442, -0.0429, -0.0, -0.342, -0.045, -0.0, -0.342, -0.0472, -0.0, -0.3442, -0.0494, -0.0, -0.3442, -0.0472, -0.0, -0.342, -0.0494, -0.0, -0.342, -0.0429, -0.0, -0.3398, -0.045, -0.0, -0.3398, -0.0426, -0.0, -0.3376, -0.045, -0.0, -0.3373, -0.0472, -0.0, -0.3398, -0.0494, -0.0, -0.3398, -0.0473, -0.0, -0.3373, -0.0496, -0.0, -0.3378,
-0.0422, 0.0017, -0.3378, -0.045, 0.0017, -0.3373, -0.0422, 0.0039, -0.3378, -0.045, 0.0039, -0.3373, -0.0472, 0.0017, -0.3373, -0.0494, 0.0017, -0.3377, -0.0472, 0.0039, -0.3373, -0.0494, 0.0039, -0.3377, -0.0334, -0.0, -0.3284, -0.0333, 0.0017, -0.3286, -0.0333, 0.0039, -0.3286, -0.0515, -0.0, -0.3442, -0.0537, -0.0, -0.3442, -0.0515, -0.0, -0.342, -0.0544, -0.0, -0.3418, -0.0553, -0.0, -0.3441, -0.0548, -0.0, -0.343, -0.0544, 0.0017, -0.3418, -0.0544, 0.0039, -0.3418, -0.0553, 0.0017, -0.3441, -0.0553, 0.0039, -0.3441, -0.0548, 0.0017, -0.343, -0.0548, 0.0039, -0.343, -0.052, -0.0, -0.3391, -0.0532, -0.0, -0.3403, -0.0508, -0.0, -0.3385, -0.0521, 0.0017, -0.3392, -0.0532, 0.0017, -0.3403, -0.0521, 0.0039, -0.3392, -0.0532, 0.0039, -0.3403,
-0.0509, 0.0017, -0.3385, -0.0509, 0.0039, -0.3385, -0.0672, -0.0, -0.3309, -0.0654, -0.0, -0.3281, -0.0662, -0.0, -0.3292, -0.0673, 0.0017, -0.3311, -0.0673, 0.0039, -0.3311, -0.0655, 0.0017, -0.3282, -0.0662, 0.0017, -0.3292, -0.0655, 0.0039, -0.3282, -0.0662, 0.0039, -0.3292, -0.0338, -0.0, -0.3269, -0.0344, -0.0, -0.3248, -0.0339, 0.0017, -0.3268, -0.0339, 0.0039, -0.3268, -0.0345, 0.0017, -0.3247, -0.0345, 0.0039, -0.3247, -0.0349, -0.0, -0.3227, -0.0353, -0.0, -0.3206, -0.0355, -0.0, -0.3198, -0.035, 0.0017, -0.3225, -0.035, 0.0039, -0.3225, -0.0352, 0.0017, -0.3212, -0.0354, 0.0017, -0.3201, -0.0352, 0.0039, -0.3212, -0.0354, 0.0039, -0.3201, -0.0342, -0.0, -0.3181, -0.0358, -0.0, -0.3182, -0.0342, -0.0, -0.316, -0.0363, -0.0, -0.3161,
-0.0358, 0.0017, -0.3181, -0.0358, 0.0039, -0.3181, -0.0363, 0.0017, -0.316, -0.0363, 0.0039, -0.316, -0.0342, -0.0, -0.3138, -0.0366, -0.0, -0.3139, -0.0342, -0.0, -0.3116, -0.0368, -0.0, -0.3117, -0.0366, 0.0017, -0.3138, -0.0366, 0.0039, -0.3138, -0.0369, 0.0017, -0.3116, -0.0369, 0.0039, -0.3116, -0.0504, -0.0, -0.3106, -0.0504, 0.0017, -0.3106, -0.0504, 0.0039, -0.3106, -0.059, -0.0, -0.3193, -0.059, 0.0017, -0.3193, -0.059, 0.0039, -0.3193, -0.0629, -0.0, -0.3243, -0.0648, -0.0, -0.3271, -0.0667, -0.0, -0.3268, -0.0636, -0.0, -0.3253, -0.0667, -0.0, -0.3246, -0.0629, 0.0017, -0.3243, -0.0629, 0.0039, -0.3243, -0.0646, 0.0017, -0.3268, -0.0646, 0.0039, -0.3268, -0.0637, 0.0017, -0.3254, -0.0637, 0.0039, -0.3254, -0.061, -0.0, -0.3217,
-0.0619, -0.0, -0.3228, -0.06, -0.0, -0.3204, -0.0624, -0.0, -0.3203, -0.0646, -0.0, -0.3225, -0.0667, -0.0, -0.3225, -0.0646, -0.0, -0.3203, -0.0667, -0.0, -0.3203, -0.061, 0.0017, -0.3217, -0.0619, 0.0017, -0.3228, -0.061, 0.0039, -0.3217, -0.0619, 0.0039, -0.3228, -0.06, 0.0017, -0.3204, -0.06, 0.0039, -0.3204, -0.0548, -0.0, -0.3149, -0.0569, -0.0, -0.3171, -0.058, -0.0, -0.3182, -0.0558, -0.0, -0.316, -0.058, -0.0, -0.316, -0.0548, 0.0017, -0.3149, -0.0548, 0.0039, -0.3149, -0.0569, 0.0017, -0.3171, -0.058, 0.0017, -0.3182, -0.0569, 0.0039, -0.3171, -0.058, 0.0039, -0.3182, -0.0558, 0.0017, -0.316, -0.0558, 0.0039, -0.316, -0.0526, -0.0, -0.3127, -0.0537, -0.0, -0.3138, -0.0515, -0.0, -0.3117, -0.0537, -0.0, -0.3116,
-0.0559, -0.0, -0.3138, -0.058, -0.0, -0.3138, -0.0559, -0.0, -0.3116, -0.058, -0.0, -0.3116, -0.0526, 0.0017, -0.3127, -0.0537, 0.0017, -0.3138, -0.0526, 0.0039, -0.3127, -0.0537, 0.0039, -0.3138, -0.0515, 0.0017, -0.3117, -0.0515, 0.0039, -0.3117, -0.0602, -0.0, -0.3181, -0.0624, -0.0, -0.3181, -0.0602, -0.0, -0.316, -0.0624, -0.0, -0.316, -0.0646, -0.0, -0.3181, -0.0667, -0.0, -0.3181, -0.0646, -0.0, -0.316, -0.0667, -0.0, -0.316, -0.0602, -0.0, -0.3138, -0.0624, -0.0, -0.3138, -0.0602, -0.0, -0.3116, -0.0624, -0.0, -0.3116, -0.0646, -0.0, -0.3138, -0.0676, -0.0, -0.313, -0.0655, -0.0, -0.3107, -0.0665, -0.0, -0.3118, -0.0676, 0.0017, -0.313, -0.0676, 0.0039, -0.313, -0.0655, 0.0017, -0.3107, -0.0665, 0.0017, -0.3118,
-0.0655, 0.0039, -0.3107, -0.0665, 0.0039, -0.3118, -0.0083, -0.0, -0.3093, -0.0103, -0.0, -0.3095, -0.0089, -0.0, -0.3078, -0.0093, -0.0, -0.3069, -0.0125, -0.0, -0.3095, -0.0146, -0.0, -0.3095, -0.0125, -0.0, -0.3073, -0.0146, -0.0, -0.3073, -0.0082, 0.0017, -0.3095, -0.0082, 0.0039, -0.3095, -0.0089, 0.0017, -0.3077, -0.0094, 0.0017, -0.3066, -0.0089, 0.0039, -0.3077, -0.0094, 0.0039, -0.3066, -0.0101, -0.0, -0.3049, -0.0109, -0.0, -0.3029, -0.0125, -0.0, -0.3051, -0.0146, -0.0, -0.3051, -0.0125, -0.0, -0.3029, -0.0146, -0.0, -0.3029, -0.01, 0.0017, -0.3051, -0.01, 0.0039, -0.3051, -0.0108, 0.0017, -0.3029, -0.0108, 0.0039, -0.3029, -0.0113, -0.0, -0.3017, -0.0117, -0.0, -0.3007, -0.0146, -0.0, -0.3008, -0.0125, -0.0, -0.2985,
-0.0146, -0.0, -0.2986, -0.0113, 0.0017, -0.3017, -0.0113, 0.0039, -0.3017, -0.0118, 0.0017, -0.3006, -0.0118, 0.0039, -0.3006, -0.0125, 0.0017, -0.2986, -0.0125, 0.0039, -0.2986, -0.0131, -0.0, -0.2964, -0.0146, -0.0, -0.2964, -0.0134, -0.0, -0.295, -0.0136, -0.0, -0.2941, -0.0131, 0.0017, -0.2964, -0.0131, 0.0039, -0.2964, -0.0134, 0.0017, -0.2948, -0.0137, 0.0017, -0.2937, -0.0134, 0.0039, -0.2948, -0.0137, 0.0039, -0.2937, -0.0168, -0.0, -0.3095, -0.019, -0.0, -0.3095, -0.0168, -0.0, -0.3073, -0.019, -0.0, -0.3073, -0.0205, -0.0, -0.3095, -0.021, -0.0, -0.3074, -0.0205, 0.0017, -0.3095, -0.0205, 0.0039, -0.3095, -0.021, 0.0017, -0.3073, -0.021, 0.0039, -0.3073, -0.0168, -0.0, -0.3051, -0.019, -0.0, -0.3051, -0.0168, -0.0, -0.3029,
-0.019, -0.0, -0.3029, -0.0215, -0.0, -0.3053, -0.0219, -0.0, -0.3031, -0.0215, 0.0017, -0.3051, -0.0215, 0.0039, -0.3051, -0.022, 0.0017, -0.3029, -0.022, 0.0039, -0.3029, -0.0282, -0.0, -0.3094, -0.0285, -0.0, -0.3077, -0.0298, -0.0, -0.3095, -0.032, -0.0, -0.3095, -0.0287, -0.0, -0.3069, -0.032, -0.0, -0.3073, -0.0282, 0.0017, -0.3095, -0.0282, 0.0039, -0.3095, -0.0286, 0.0017, -0.3074, -0.0286, 0.0039, -0.3074, -0.0288, 0.0017, -0.3063, -0.0288, 0.0039, -0.3063, -0.029, -0.0, -0.305, -0.032, -0.0, -0.3051, -0.0293, -0.0, -0.3028, -0.032, -0.0, -0.3029, -0.029, 0.0017, -0.3051, -0.029, 0.0039, -0.3051, -0.0293, 0.0017, -0.3029, -0.0293, 0.0039, -0.3029, -0.0168, -0.0, -0.3008, -0.019, -0.0, -0.3008, -0.0168, -0.0, -0.2986,
-0.019, -0.0, -0.2986, -0.0223, -0.0, -0.301, -0.0224, -0.0, -0.3003, -0.0212, -0.0, -0.2986, -0.0227, -0.0, -0.2986, -0.0222, 0.0017, -0.3016, -0.0224, 0.0017, -0.3005, -0.0222, 0.0039, -0.3016, -0.0224, 0.0039, -0.3005, -0.0227, 0.0017, -0.2986, -0.0227, 0.0039, -0.2986, -0.0168, -0.0, -0.2964, -0.019, -0.0, -0.2964, -0.0168, -0.0, -0.2943, -0.019, -0.0, -0.2943, -0.0212, -0.0, -0.2964, -0.023, -0.0, -0.2965, -0.0212, -0.0, -0.2943, -0.0232, -0.0, -0.2943, -0.023, 0.0017, -0.2964, -0.023, 0.0039, -0.2964, -0.0232, 0.0017, -0.2943, -0.0232, 0.0039, -0.2943, -0.0297, -0.0, -0.3007, -0.032, -0.0, -0.3008, -0.03, -0.0, -0.2986, -0.032, -0.0, -0.2986, -0.0297, 0.0017, -0.3008, -0.0297, 0.0039, -0.3008, -0.0299, 0.0017, -0.2986,
-0.0299, 0.0039, -0.2986, -0.03, -0.0, -0.2964, -0.032, -0.0, -0.2964, -0.0301, -0.0, -0.2943, -0.032, -0.0, -0.2943, -0.03, 0.0017, -0.2964, -0.03, 0.0039, -0.2964, -0.0301, 0.0017, -0.2943, -0.0301, 0.0039, -0.2943, -0.0066, -0.0, -0.289, -0.0064, 0.0017, -0.2889, -0.0064, 0.0039, -0.2889, 0.0006, -0.0, -0.2882, -0.0016, -0.0, -0.2883, 0.0005, -0.0, -0.2856, -0.0016, -0.0, -0.2856, -0.0037, -0.0, -0.2885, -0.0058, -0.0, -0.2888, -0.0038, -0.0, -0.2856, -0.006, -0.0, -0.2856, 0.0005, 0.0017, -0.2882, -0.0016, 0.0017, -0.2883, 0.0005, 0.0039, -0.2882, -0.0016, 0.0039, -0.2883, -0.0038, 0.0017, -0.2885, -0.0053, 0.0017, -0.2888, -0.0038, 0.0039, -0.2885, -0.0053, 0.0039, -0.2888, -0.0081, -0.0, -0.2892, -0.0102, -0.0, -0.2896,
-0.014, -0.0, -0.2919, -0.0124, -0.0, -0.29, -0.0144, -0.0, -0.2902, -0.0081, 0.0017, -0.2892, -0.0103, 0.0017, -0.2896, -0.0081, 0.0039, -0.2892, -0.0103, 0.0039, -0.2896, -0.014, 0.0017, -0.2921, -0.014, 0.0039, -0.2921, -0.0125, 0.0017, -0.29, -0.0144, 0.0017, -0.2902, -0.0125, 0.0039, -0.29, -0.0144, 0.0039, -0.2902, -0.0081, -0.0, -0.2878, -0.0103, -0.0, -0.2878, -0.0081, -0.0, -0.2856, -0.0103, -0.0, -0.2856, -0.0125, -0.0, -0.2878, -0.0146, -0.0, -0.2878, -0.0125, -0.0, -0.2856, -0.0146, -0.0, -0.2856, 0.0005, -0.0, -0.2834, -0.0016, -0.0, -0.2834, 0.0005, -0.0, -0.2812, -0.0016, -0.0, -0.2812, -0.0038, -0.0, -0.2834, -0.006, -0.0, -0.2834, -0.0038, -0.0, -0.2812, -0.0063, -0.0, -0.2801, -0.0069, 0.0017, -0.2802,
-0.0069, 0.0039, -0.2802, 0.0005, -0.0, -0.2794, -0.0017, -0.0, -0.2795, -0.0038, -0.0, -0.2797, -0.0055, -0.0, -0.28, 0.0005, 0.0017, -0.2794, -0.0016, 0.0017, -0.2795, 0.0005, 0.0039, -0.2794, -0.0016, 0.0039, -0.2795, -0.0038, 0.0017, -0.2797, -0.0058, 0.0017, -0.28, -0.0038, 0.0039, -0.2797, -0.0058, 0.0039, -0.28, -0.0081, -0.0, -0.2834, -0.0103, -0.0, -0.2834, -0.0083, -0.0, -0.2804, -0.0104, -0.0, -0.2807, -0.0125, -0.0, -0.2834, -0.0146, -0.0, -0.2834, -0.0126, -0.0, -0.281, -0.0147, -0.0, -0.2813, -0.0081, 0.0017, -0.2803, -0.0103, 0.0017, -0.2806, -0.0081, 0.0039, -0.2803, -0.0103, 0.0039, -0.2806, -0.0125, 0.0017, -0.281, -0.0146, 0.0017, -0.2813, -0.0125, 0.0039, -0.281, -0.0146, 0.0039, -0.2813, -0.0168, -0.0, -0.2921,
-0.019, -0.0, -0.2921, -0.0168, -0.0, -0.2899, -0.019, -0.0, -0.2899, -0.0212, -0.0, -0.2921, -0.0232, -0.0, -0.2921, -0.0212, -0.0, -0.2899, -0.0232, -0.0, -0.2899, -0.0232, 0.0017, -0.2921, -0.0232, 0.0039, -0.2921, -0.0232, 0.0017, -0.2899, -0.0232, 0.0039, -0.2899, -0.0168, -0.0, -0.2878, -0.019, -0.0, -0.2878, -0.0168, -0.0, -0.2856, -0.019, -0.0, -0.2856, -0.0212, -0.0, -0.2878, -0.0232, -0.0, -0.2878, -0.0212, -0.0, -0.2856, -0.0232, -0.0, -0.2856, -0.0232, 0.0017, -0.2878, -0.0232, 0.0039, -0.2878, -0.0232, 0.0017, -0.2856, -0.0232, 0.0039, -0.2856, -0.0301, -0.0, -0.2921, -0.032, -0.0, -0.2921, -0.0301, -0.0, -0.2899, -0.032, -0.0, -0.2899, -0.0301, 0.0017, -0.2921, -0.0301, 0.0039, -0.2921, -0.0301, 0.0017, -0.2899,
-0.0301, 0.0039, -0.2899, -0.0301, -0.0, -0.2878, -0.032, -0.0, -0.2878, -0.03, -0.0, -0.2856, -0.032, -0.0, -0.2856, -0.0301, 0.0017, -0.2878, -0.0301, 0.0039, -0.2878, -0.03, 0.0017, -0.2856, -0.03, 0.0039, -0.2856, -0.0168, -0.0, -0.2834, -0.019, -0.0, -0.2834, -0.0163, -0.0, -0.2814, -0.019, -0.0, -0.2812, -0.0212, -0.0, -0.2834, -0.0232, -0.0, -0.2834, -0.0212, -0.0, -0.2812, -0.0232, -0.0, -0.2812, -0.0163, 0.0017, -0.2814, -0.0163, 0.0039, -0.2814, -0.0232, 0.0017, -0.2834, -0.0232, 0.0039, -0.2834, -0.0232, 0.0017, -0.2812, -0.0232, 0.0039, -0.2812, -0.0163, -0.0, -0.2791, -0.019, -0.0, -0.2791, -0.0163, -0.0, -0.2769, -0.019, -0.0, -0.2769, -0.0212, -0.0, -0.2791, -0.0232, -0.0, -0.279, -0.0212, -0.0, -0.2769,
-0.023, -0.0, -0.2769, -0.0163, 0.0017, -0.2791, -0.0163, 0.0039, -0.2791, -0.0163, 0.0017, -0.2769, -0.0163, 0.0039, -0.2769, -0.0232, 0.0017, -0.2791, -0.0232, 0.0039, -0.2791, -0.023, 0.0017, -0.2769, -0.023, 0.0039, -0.2769, -0.0299, -0.0, -0.2835, -0.032, -0.0, -0.2834, -0.0296, -0.0, -0.2813, -0.032, -0.0, -0.2812, -0.0299, 0.0017, -0.2834, -0.0299, 0.0039, -0.2834, -0.0296, 0.0017, -0.2812, -0.0296, 0.0039, -0.2812, -0.0286, -0.0, -0.2762, -0.0292, -0.0, -0.2792, -0.032, -0.0, -0.2791, -0.0288, -0.0, -0.2771, -0.032, -0.0, -0.2769, -0.0286, 0.0017, -0.2764, -0.0286, 0.0039, -0.2764, -0.0292, 0.0017, -0.2791, -0.0292, 0.0039, -0.2791, -0.0289, 0.0017, -0.2774, -0.0289, 0.0039, -0.2774, -0.0342, -0.0, -0.3095, -0.037, -0.0, -0.3095,
-0.0342, -0.0, -0.3073, -0.0371, -0.0, -0.3074, -0.037, 0.0017, -0.3095, -0.037, 0.0039, -0.3095, -0.0371, 0.0017, -0.3073, -0.0371, 0.0039, -0.3073, -0.0342, -0.0, -0.3051, -0.0374, -0.0, -0.3053, -0.0342, -0.0, -0.3029, -0.0363, -0.0, -0.3029, -0.0376, -0.0, -0.3045, -0.038, -0.0, -0.301, -0.0405, -0.0, -0.3027, -0.0374, 0.0017, -0.3058, -0.0374, 0.0039, -0.3058, -0.0375, 0.0017, -0.3047, -0.0375, 0.0039, -0.3047, -0.038, 0.0017, -0.301, -0.0407, 0.0017, -0.3029, -0.038, 0.0039, -0.301, -0.0407, 0.0039, -0.3029, -0.0457, -0.0, -0.3065, -0.0481, -0.0, -0.3085, -0.0493, -0.0, -0.3096, -0.047, -0.0, -0.3076, -0.0494, -0.0, -0.3073, -0.0457, 0.0017, -0.3065, -0.0457, 0.0039, -0.3065, -0.0481, 0.0017, -0.3085, -0.0493, 0.0017, -0.3096,
-0.0481, 0.0039, -0.3085, -0.0493, 0.0039, -0.3096, -0.0471, 0.0017, -0.3076, -0.0471, 0.0039, -0.3076, -0.0432, -0.0, -0.3047, -0.0447, -0.0, -0.3057, -0.0422, -0.0, -0.3039, -0.045, -0.0, -0.3029, -0.0472, -0.0, -0.3051, -0.0494, -0.0, -0.3051, -0.0472, -0.0, -0.3029, -0.0494, -0.0, -0.3029, -0.0432, 0.0017, -0.3046, -0.0446, 0.0017, -0.3057, -0.0432, 0.0039, -0.3046, -0.0446, 0.0039, -0.3057, -0.0421, 0.0017, -0.3038, -0.0421, 0.0039, -0.3038, -0.0342, -0.0, -0.3008, -0.0363, -0.0, -0.3008, -0.0342, -0.0, -0.2986, -0.0363, -0.0, -0.2986, -0.0384, -0.0, -0.3013, -0.0407, -0.0, -0.3008, -0.0385, -0.0, -0.2986, -0.0407, -0.0, -0.2986, -0.0342, -0.0, -0.2964, -0.0363, -0.0, -0.2964, -0.0342, -0.0, -0.2943, -0.0363, -0.0, -0.2943,
-0.0385, -0.0, -0.2964, -0.0407, -0.0, -0.2964, -0.0385, -0.0, -0.2943, -0.0407, -0.0, -0.2943, -0.0429, -0.0, -0.3008, -0.045, -0.0, -0.3008, -0.0429, -0.0, -0.2986, -0.045, -0.0, -0.2986, -0.0472, -0.0, -0.3008, -0.0494, -0.0, -0.3008, -0.0472, -0.0, -0.2986, -0.0494, -0.0, -0.2986, -0.0429, -0.0, -0.2964, -0.045, -0.0, -0.2964, -0.0436, -0.0, -0.2931, -0.0453, -0.0, -0.2942, -0.0477, -0.0, -0.2956, -0.0496, -0.0, -0.2969, -0.0466, -0.0, -0.295, -0.0438, 0.0017, -0.2932, -0.045, 0.0017, -0.294, -0.0438, 0.0039, -0.2932, -0.045, 0.0039, -0.294, -0.0477, 0.0017, -0.2957, -0.0494, 0.0017, -0.2968, -0.0477, 0.0039, -0.2957, -0.0494, 0.0039, -0.2968, -0.0467, 0.0017, -0.295, -0.0467, 0.0039, -0.295, -0.0515, -0.0, -0.3095,
-0.0537, -0.0, -0.3095, -0.0515, -0.0, -0.3073, -0.0537, -0.0, -0.3073, -0.0559, -0.0, -0.3095, -0.058, -0.0, -0.3095, -0.0559, -0.0, -0.3073, -0.058, -0.0, -0.3073, -0.0515, -0.0, -0.3051, -0.0537, -0.0, -0.3051, -0.0515, -0.0, -0.3029, -0.0537, -0.0, -0.3029, -0.0559, -0.0, -0.3051, -0.0588, -0.0, -0.3042, -0.0565, -0.0, -0.3022, -0.0577, -0.0, -0.3033, -0.0589, 0.0017, -0.3043, -0.0589, 0.0039, -0.3043, -0.0565, 0.0017, -0.3022, -0.0578, 0.0017, -0.3033, -0.0565, 0.0039, -0.3022, -0.0578, 0.0039, -0.3033, -0.0602, -0.0, -0.3095, -0.0634, -0.0, -0.3084, -0.0611, -0.0, -0.3063, -0.0623, -0.0, -0.3073, -0.0645, -0.0, -0.3095, -0.0634, 0.0017, -0.3084, -0.0634, 0.0039, -0.3084, -0.0612, 0.0017, -0.3063, -0.0624, 0.0017, -0.3074,
-0.0612, 0.0039, -0.3063, -0.0624, 0.0039, -0.3074, -0.0644, 0.0017, -0.3095, -0.0644, 0.0039, -0.3095, -0.0601, -0.0, -0.3053, -0.0601, 0.0017, -0.3053, -0.0601, 0.0039, -0.3053, -0.0515, -0.0, -0.3008, -0.0541, -0.0, -0.3002, -0.0516, -0.0, -0.2983, -0.053, -0.0, -0.2994, -0.0554, -0.0, -0.3013, -0.0541, 0.0017, -0.3002, -0.0541, 0.0039, -0.3002, -0.0516, 0.0017, -0.2983, -0.053, 0.0017, -0.2994, -0.0516, 0.0039, -0.2983, -0.053, 0.0039, -0.2994, -0.0554, 0.0017, -0.3013, -0.0554, 0.0039, -0.3013, -0.0505, -0.0, -0.2975, -0.0505, 0.0017, -0.2975, -0.0505, 0.0039, -0.2975, -0.0342, -0.0, -0.2921, -0.0363, -0.0, -0.2921, -0.0342, -0.0, -0.2899, -0.0369, -0.0, -0.2902, -0.0391, -0.0, -0.2908, -0.041, -0.0, -0.2916, -0.0382, -0.0, -0.2905,
-0.0369, 0.0017, -0.2902, -0.0369, 0.0039, -0.2902, -0.0396, 0.0017, -0.291, -0.0407, 0.0017, -0.2915, -0.0396, 0.0039, -0.291, -0.0407, 0.0039, -0.2915, -0.0385, 0.0017, -0.2906, -0.0385, 0.0039, -0.2906, -0.0342, -0.0, -0.2878, -0.0369, -0.0, -0.2877, -0.0342, -0.0, -0.2856, -0.0367, -0.0, -0.2855, -0.0369, 0.0017, -0.2878, -0.0369, 0.0039, -0.2878, -0.0367, 0.0017, -0.2856, -0.0367, 0.0039, -0.2856, -0.0426, -0.0, -0.2925, -0.0428, 0.0017, -0.2926, -0.0428, 0.0039, -0.2926, -0.0342, -0.0, -0.2834, -0.0363, -0.0, -0.2833, -0.0342, -0.0, -0.2812, -0.0359, -0.0, -0.2812, -0.0363, 0.0017, -0.2834, -0.0363, 0.0039, -0.2834, -0.0359, 0.0017, -0.2812, -0.0359, 0.0039, -0.2812, -0.0353, -0.0, -0.2787, -0.0355, -0.0, -0.2796, -0.0347, -0.0, -0.2767,
-0.0352, 0.0017, -0.2783, -0.0355, 0.0017, -0.2794, -0.0352, 0.0039, -0.2783, -0.0355, 0.0039, -0.2794, -0.0348, 0.0017, -0.2769, -0.0348, 0.0039, -0.2769, -0.0738, -0.0, -0.3443, -0.0754, -0.0, -0.3442, -0.0729, -0.0, -0.3423, -0.0754, -0.0, -0.342, -0.0737, 0.0017, -0.3442, -0.0737, 0.0039, -0.3442, -0.0728, 0.0017, -0.342, -0.0728, 0.0039, -0.342, -0.0716, -0.0, -0.3395, -0.0708, -0.0, -0.338, -0.0721, -0.0, -0.3405, -0.0754, -0.0, -0.3398, -0.0732, -0.0, -0.3377, -0.0754, -0.0, -0.3377, -0.0717, 0.0017, -0.3397, -0.0717, 0.0039, -0.3397, -0.0707, 0.0017, -0.3377, -0.0707, 0.0039, -0.3377, -0.0722, 0.0017, -0.3408, -0.0722, 0.0039, -0.3408, -0.0776, -0.0, -0.3442, -0.0798, -0.0, -0.3442, -0.0776, -0.0, -0.342, -0.0798, -0.0, -0.342,
-0.0819, -0.0, -0.3442, -0.0841, -0.0, -0.3442, -0.0819, -0.0, -0.342, -0.0849, -0.0, -0.3419, -0.0848, 0.0017, -0.3419, -0.0848, 0.0039, -0.3419, -0.0776, -0.0, -0.3398, -0.0798, -0.0, -0.3398, -0.0776, -0.0, -0.3377, -0.0798, -0.0, -0.3377, -0.0819, -0.0, -0.3398, -0.0841, -0.0, -0.3396, -0.0831, -0.0, -0.3371, -0.0835, -0.0, -0.3381, -0.0841, 0.0017, -0.3398, -0.0841, 0.0039, -0.3398, -0.0829, 0.0017, -0.3368, -0.0834, 0.0017, -0.3379, -0.0829, 0.0039, -0.3368, -0.0834, 0.0039, -0.3379, -0.0694, -0.0, -0.3351, -0.0699, -0.0, -0.3361, -0.0687, -0.0, -0.3336, -0.0711, -0.0, -0.3333, -0.0732, -0.0, -0.3355, -0.0754, -0.0, -0.3355, -0.0732, -0.0, -0.3333, -0.0754, -0.0, -0.3333, -0.0695, 0.0017, -0.3353, -0.0701, 0.0017, -0.3364,
-0.0695, 0.0039, -0.3353, -0.0701, 0.0039, -0.3364, -0.0685, 0.0017, -0.3333, -0.0685, 0.0039, -0.3333, -0.0677, -0.0, -0.3319, -0.0711, -0.0, -0.3312, -0.0689, -0.0, -0.329, -0.0711, -0.0, -0.329, -0.0732, -0.0, -0.3312, -0.0754, -0.0, -0.3312, -0.0732, -0.0, -0.329, -0.0754, -0.0, -0.329, -0.0679, 0.0017, -0.3321, -0.0679, 0.0039, -0.3321, -0.0776, -0.0, -0.3355, -0.0798, -0.0, -0.3355, -0.0776, -0.0, -0.3333, -0.0809, -0.0, -0.3327, -0.0822, -0.0, -0.3352, -0.0814, -0.0, -0.3336, -0.0808, 0.0017, -0.3324, -0.0808, 0.0039, -0.3324, -0.0823, 0.0017, -0.3355, -0.0823, 0.0039, -0.3355, -0.0813, 0.0017, -0.3335, -0.0813, 0.0039, -0.3335, -0.0776, -0.0, -0.3312, -0.08, -0.0, -0.3308, -0.0786, -0.0, -0.3284, -0.0792, -0.0, -0.3294,
-0.0802, 0.0017, -0.3311, -0.0802, 0.0039, -0.3311, -0.0785, 0.0017, -0.3282, -0.0791, 0.0017, -0.3293, -0.0785, 0.0039, -0.3282, -0.0791, 0.0039, -0.3293, -0.0856, -0.0, -0.3441, -0.0852, -0.0, -0.343, -0.0856, 0.0017, -0.3442, -0.0856, 0.0039, -0.3442, -0.0852, 0.0017, -0.343, -0.0852, 0.0039, -0.343, -0.0689, -0.0, -0.3268, -0.0711, -0.0, -0.3268, -0.0689, -0.0, -0.3246, -0.0711, -0.0, -0.3246, -0.0732, -0.0, -0.3268, -0.0754, -0.0, -0.3268, -0.0732, -0.0, -0.3246, -0.0761, -0.0, -0.3243, -0.0761, 0.0017, -0.3242, -0.0761, 0.0039, -0.3242, -0.0689, -0.0, -0.3225, -0.0711, -0.0, -0.3225, -0.0689, -0.0, -0.3203, -0.0711, -0.0, -0.3203, -0.0743, -0.0, -0.3217, -0.0751, -0.0, -0.3227, -0.0732, -0.0, -0.32, -0.0742, 0.0017, -0.3215,
-0.075, 0.0017, -0.3226, -0.0742, 0.0039, -0.3215, -0.075, 0.0039, -0.3226, -0.0734, 0.0017, -0.3203, -0.0734, 0.0039, -0.3203, -0.0775, -0.0, -0.3265, -0.0768, -0.0, -0.3254, -0.0777, 0.0017, -0.3268, -0.0777, 0.0039, -0.3268, -0.0768, 0.0017, -0.3253, -0.0768, 0.0039, -0.3253, -0.0689, -0.0, -0.3181, -0.0716, -0.0, -0.3178, -0.0696, -0.0, -0.3153, -0.0706, -0.0, -0.3164, -0.0724, -0.0, -0.3189, -0.0716, 0.0017, -0.3178, -0.0716, 0.0039, -0.3178, -0.0696, 0.0017, -0.3153, -0.0705, 0.0017, -0.3164, -0.0696, 0.0039, -0.3153, -0.0705, 0.0039, -0.3164, -0.0724, 0.0017, -0.3189, -0.0724, 0.0039, -0.3189, -0.0686, -0.0, -0.3141, -0.0685, 0.0017, -0.3141, -0.0685, 0.0039, -0.3141, -0.0157, -0.0, -0.2652, -0.0154, -0.0, -0.2639, -0.0157, 0.0017, -0.2653,
-0.0157, 0.0039, -0.2653, -0.0154, 0.0017, -0.2639, -0.0154, 0.0039, -0.2639, -0.015, -0.0, -0.2618, -0.0146, -0.0, -0.2596, -0.015, 0.0017, -0.2617, -0.015, 0.0039, -0.2617, -0.0146, 0.0017, -0.2595, -0.0146, 0.0039, -0.2595, -0.0163, -0.0, -0.2747, -0.019, -0.0, -0.2747, -0.0163, -0.0, -0.2726, -0.019, -0.0, -0.2726, -0.0212, -0.0, -0.2747, -0.0226, -0.0, -0.2747, -0.0221, -0.0, -0.2725, -0.0223, -0.0, -0.2735, -0.0163, 0.0017, -0.2747, -0.0163, 0.0039, -0.2747, -0.0163, 0.0017, -0.2726, -0.0163, 0.0039, -0.2726, -0.0226, 0.0017, -0.2747, -0.0226, 0.0039, -0.2747, -0.022, 0.0017, -0.2723, -0.0223, 0.0017, -0.2734, -0.022, 0.0039, -0.2723, -0.0223, 0.0039, -0.2734, -0.0163, -0.0, -0.2704, -0.019, -0.0, -0.2704, -0.0161, -0.0, -0.2683,
-0.019, -0.0, -0.2682, -0.0216, -0.0, -0.2702, -0.0211, -0.0, -0.2681, -0.0163, 0.0017, -0.2704, -0.0163, 0.0039, -0.2704, -0.0161, 0.0017, -0.2682, -0.0161, 0.0039, -0.2682, -0.0216, 0.0017, -0.2704, -0.0216, 0.0039, -0.2704, -0.0211, 0.0017, -0.2682, -0.0211, 0.0039, -0.2682, -0.0283, -0.0, -0.2748, -0.0277, -0.0, -0.2727, -0.0298, -0.0, -0.2747, -0.032, -0.0, -0.2747, -0.0298, -0.0, -0.2726, -0.0331, -0.0, -0.2721, -0.0283, 0.0017, -0.2747, -0.0283, 0.0039, -0.2747, -0.0277, 0.0017, -0.2726, -0.0277, 0.0039, -0.2726, -0.033, 0.0017, -0.2717, -0.033, 0.0039, -0.2717, -0.0271, -0.0, -0.2706, -0.0262, -0.0, -0.2678, -0.0265, -0.0, -0.2687, -0.0298, -0.0, -0.2704, -0.0324, -0.0, -0.2701, -0.0298, -0.0, -0.2682, -0.0316, -0.0, -0.2681,
-0.027, 0.0017, -0.2704, -0.027, 0.0039, -0.2704, -0.0263, 0.0017, -0.2681, -0.0266, 0.0017, -0.2692, -0.0263, 0.0039, -0.2681, -0.0266, 0.0039, -0.2692, -0.0325, 0.0017, -0.2704, -0.0325, 0.0039, -0.2704, -0.0317, 0.0017, -0.2682, -0.0317, 0.0039, -0.2682, -0.0158, -0.0, -0.2661, -0.019, -0.0, -0.2661, -0.0168, -0.0, -0.2639, -0.0199, -0.0, -0.2637, -0.0206, -0.0, -0.266, -0.0202, -0.0, -0.2647, -0.0159, 0.0017, -0.2664, -0.0159, 0.0039, -0.2664, -0.0198, 0.0017, -0.2635, -0.0198, 0.0039, -0.2635, -0.0206, 0.0017, -0.266, -0.0206, 0.0039, -0.266, -0.0202, 0.0017, -0.2646, -0.0202, 0.0039, -0.2646, -0.0168, -0.0, -0.2617, -0.0192, -0.0, -0.2615, -0.0168, -0.0, -0.2595, -0.0185, -0.0, -0.2594, -0.024, -0.0, -0.2614, -0.0233, -0.0, -0.2598,
-0.0193, 0.0017, -0.2617, -0.0193, 0.0039, -0.2617, -0.0185, 0.0017, -0.2595, -0.0185, 0.0039, -0.2595, -0.0241, 0.0017, -0.2616, -0.0241, 0.0039, -0.2616, -0.0232, 0.0017, -0.2595, -0.0232, 0.0039, -0.2595, -0.0256, -0.0, -0.2662, -0.0277, -0.0, -0.2661, -0.025, -0.0, -0.2641, -0.0277, -0.0, -0.2639, -0.0307, -0.0, -0.2658, -0.0311, -0.0, -0.2669, -0.0297, -0.0, -0.2637, -0.0256, 0.0017, -0.2661, -0.0256, 0.0039, -0.2661, -0.0249, 0.0017, -0.2639, -0.0249, 0.0039, -0.2639, -0.0306, 0.0017, -0.2657, -0.0311, 0.0017, -0.2668, -0.0306, 0.0039, -0.2657, -0.0311, 0.0039, -0.2668, -0.0298, 0.0017, -0.2639, -0.0298, 0.0039, -0.2639, -0.0243, -0.0, -0.2622, -0.0286, -0.0, -0.2613, -0.0255, -0.0, -0.2595, -0.0275, -0.0, -0.2593, -0.0291, -0.0, -0.2623,
-0.0245, 0.0017, -0.2627, -0.0285, 0.0017, -0.2611, -0.0245, 0.0039, -0.2627, -0.0285, 0.0039, -0.2611, -0.0276, 0.0017, -0.2595, -0.0276, 0.0039, -0.2595, -0.029, 0.0017, -0.2622, -0.029, 0.0039, -0.2622, -0.0142, -0.0, -0.2575, -0.0135, -0.0, -0.2541, -0.0138, -0.0, -0.2551, -0.0142, 0.0017, -0.2574, -0.0142, 0.0039, -0.2574, -0.0136, 0.0017, -0.2542, -0.0138, 0.0017, -0.2552, -0.0136, 0.0039, -0.2542, -0.0138, 0.0039, -0.2552, -0.0133, -0.0, -0.2531, -0.0155, -0.0, -0.2527, -0.0128, -0.0, -0.2509, -0.0144, -0.0, -0.2506, -0.0133, 0.0017, -0.253, -0.0154, 0.0017, -0.2526, -0.0133, 0.0039, -0.253, -0.0154, 0.0039, -0.2526, -0.0128, 0.0017, -0.2509, -0.0145, 0.0017, -0.2508, -0.0128, 0.0039, -0.2509, -0.0145, 0.0039, -0.2508, -0.0131, -0.0, -0.2497,
-0.0137, -0.0, -0.2496, -0.0121, 0.0017, -0.2467, -0.0137, 0.0017, -0.2496, -0.0121, 0.0039, -0.2467, -0.0137, 0.0039, -0.2496, -0.0176, -0.0, -0.2572, -0.018, -0.0, -0.2583, -0.0166, -0.0, -0.255, -0.0219, -0.0, -0.2568, -0.0224, -0.0, -0.2578, -0.0212, -0.0, -0.2554, -0.0245, -0.0, -0.2544, -0.0176, 0.0017, -0.2571, -0.018, 0.0017, -0.2582, -0.0176, 0.0039, -0.2571, -0.018, 0.0039, -0.2582, -0.0167, 0.0017, -0.2552, -0.0167, 0.0039, -0.2552, -0.022, 0.0017, -0.2569, -0.0225, 0.0017, -0.258, -0.022, 0.0039, -0.2569, -0.0225, 0.0039, -0.258, -0.0211, 0.0017, -0.2552, -0.0244, 0.0017, -0.2542, -0.0211, 0.0039, -0.2552, -0.0244, 0.0039, -0.2542, -0.016, -0.0, -0.2538, -0.0197, -0.0, -0.2524, -0.0191, -0.0, -0.2511, -0.0203, -0.0, -0.2534,
-0.0233, -0.0, -0.2527, -0.0217, -0.0, -0.2505, -0.0225, -0.0, -0.2516, -0.016, 0.0017, -0.2537, -0.0198, 0.0017, -0.2525, -0.016, 0.0039, -0.2537, -0.0198, 0.0039, -0.2525, -0.019, 0.0017, -0.2509, -0.019, 0.0039, -0.2509, -0.0203, 0.0017, -0.2536, -0.0235, 0.0017, -0.253, -0.0203, 0.0039, -0.2536, -0.0235, 0.0039, -0.253, -0.0217, 0.0017, -0.2505, -0.0225, 0.0017, -0.2516, -0.0217, 0.0039, -0.2505, -0.0225, 0.0039, -0.2516, -0.0262, -0.0, -0.257, -0.0268, -0.0, -0.2581, -0.0252, -0.0, -0.2554, -0.0262, 0.0017, -0.257, -0.0268, 0.0017, -0.2581, -0.0262, 0.0039, -0.257, -0.0268, 0.0039, -0.2581, -0.0251, 0.0017, -0.2553, -0.0251, 0.0039, -0.2553, -0.0176, -0.0, -0.248, -0.019, -0.0, -0.2486, -0.0163, -0.0, -0.2449, -0.0186, -0.0, -0.2469,
-0.0207, -0.0, -0.2491, -0.0176, 0.0017, -0.2481, -0.0189, 0.0017, -0.2486, -0.0176, 0.0039, -0.2481, -0.0189, 0.0039, -0.2486, -0.0163, 0.0017, -0.2449, -0.0186, 0.0017, -0.2469, -0.0163, 0.0039, -0.2449, -0.0186, 0.0039, -0.2469, -0.0206, 0.0017, -0.2491, -0.0206, 0.0039, -0.2491, -0.034, -0.0, -0.2746, -0.0335, -0.0, -0.273, -0.0341, 0.0017, -0.2747, -0.0341, 0.0039, -0.2747, -0.0334, 0.0017, -0.2728, -0.0334, 0.0039, -0.2728, 0.0005, -0.0, -0.2196, -0.0016, -0.0, -0.2196, -0.0038, -0.0, -0.2196, -0.006, -0.0, -0.2196, 0.0005, 0.0017, -0.2196, -0.0016, 0.0017, -0.2196, 0.0005, 0.0039, -0.2196, -0.0016, 0.0039, -0.2196, -0.0038, 0.0017, -0.2196, -0.006, 0.0017, -0.2196, -0.0038, 0.0039, -0.2196, -0.006, 0.0039, -0.2196, 0.0005, -0.0, -0.2183,
-0.0016, -0.0, -0.2183, 0.0005, -0.0, -0.2161, -0.0016, -0.0, -0.2161, -0.0038, -0.0, -0.2183, -0.006, -0.0, -0.2183, -0.0038, -0.0, -0.2161, -0.006, -0.0, -0.2161, -0.0081, -0.0, -0.2196, -0.0103, -0.0, -0.2196, -0.0125, -0.0, -0.2196, -0.0146, -0.0, -0.2196, -0.0081, 0.0017, -0.2196, -0.0103, 0.0017, -0.2196, -0.0081, 0.0039, -0.2196, -0.0103, 0.0039, -0.2196, -0.0125, 0.0017, -0.2196, -0.0146, 0.0017, -0.2196, -0.0125, 0.0039, -0.2196, -0.0146, 0.0039, -0.2196, -0.0081, -0.0, -0.2183, -0.0103, -0.0, -0.2183, -0.0081, -0.0, -0.2161, -0.0103, -0.0, -0.2161, -0.0125, -0.0, -0.2183, -0.0146, -0.0, -0.2183, -0.0125, -0.0, -0.2161, -0.0146, -0.0, -0.2161, 0.0005, -0.0, -0.214, -0.0016, -0.0, -0.214, 0.0005, -0.0, -0.2118,
-0.0016, -0.0, -0.2118, -0.0038, -0.0, -0.214, -0.006, -0.0, -0.214, -0.0038, -0.0, -0.2118, -0.006, -0.0, -0.2118, 0.0005, -0.0, -0.2088, -0.0016, -0.0, -0.2088, -0.0038, -0.0, -0.2088, -0.006, -0.0, -0.2088, 0.0005, 0.0017, -0.2088, -0.0016, 0.0017, -0.2088, 0.0005, 0.0039, -0.2088, -0.0016, 0.0039, -0.2088, -0.0038, 0.0017, -0.2088, -0.006, 0.0017, -0.2088, -0.0038, 0.0039, -0.2088, -0.006, 0.0039, -0.2088, -0.0081, -0.0, -0.214, -0.0103, -0.0, -0.214, -0.0081, -0.0, -0.2118, -0.0103, -0.0, -0.2118, -0.0125, -0.0, -0.214, -0.0146, -0.0, -0.214, -0.0125, -0.0, -0.2118, -0.0146, -0.0, -0.2118, -0.0085, -0.0, -0.2088, -0.0103, -0.0, -0.2096, -0.0085, -0.0, -0.2075, -0.0103, -0.0, -0.2075, -0.0125, -0.0, -0.2096,
-0.0146, -0.0, -0.2096, -0.0125, -0.0, -0.2075, -0.0146, -0.0, -0.2075, -0.0085, 0.0017, -0.2088, -0.0085, 0.0039, -0.2088, -0.0085, 0.0017, -0.2075, -0.0085, 0.0039, -0.2075, -0.0168, -0.0, -0.2196, -0.019, -0.0, -0.2196, -0.0212, -0.0, -0.2196, -0.0233, -0.0, -0.2196, -0.0168, 0.0017, -0.2196, -0.019, 0.0017, -0.2196, -0.0168, 0.0039, -0.2196, -0.019, 0.0039, -0.2196, -0.0212, 0.0017, -0.2196, -0.0233, 0.0017, -0.2196, -0.0212, 0.0039, -0.2196, -0.0233, 0.0039, -0.2196, -0.0168, -0.0, -0.2183, -0.019, -0.0, -0.2183, -0.0168, -0.0, -0.2161, -0.019, -0.0, -0.2161, -0.0212, -0.0, -0.2183, -0.0233, -0.0, -0.2183, -0.0212, -0.0, -0.2161, -0.0233, -0.0, -0.2161, -0.0255, -0.0, -0.2196, -0.0277, -0.0, -0.2196, -0.0298, -0.0, -0.2196,
-0.032, -0.0, -0.2196, -0.0255, 0.0017, -0.2196, -0.0277, 0.0017, -0.2196, -0.0255, 0.0039, -0.2196, -0.0277, 0.0039, -0.2196, -0.0298, 0.0017, -0.2196, -0.032, 0.0017, -0.2196, -0.0298, 0.0039, -0.2196, -0.032, 0.0039, -0.2196, -0.0255, -0.0, -0.2183, -0.0277, -0.0, -0.2183, -0.0255, -0.0, -0.2161, -0.0277, -0.0, -0.2161, -0.0298, -0.0, -0.2183, -0.032, -0.0, -0.2183, -0.0298, -0.0, -0.2161, -0.032, -0.0, -0.2161, -0.0168, -0.0, -0.214, -0.019, -0.0, -0.214, -0.0168, -0.0, -0.2118, -0.019, -0.0, -0.2118, -0.0212, -0.0, -0.214, -0.0233, -0.0, -0.214, -0.0212, -0.0, -0.2118, -0.0233, -0.0, -0.2118, -0.0168, -0.0, -0.2096, -0.019, -0.0, -0.2096, -0.0168, -0.0, -0.2075, -0.019, -0.0, -0.2075, -0.0203, -0.0, -0.2088,
-0.0233, -0.0, -0.2088, -0.0203, -0.0, -0.2075, -0.0203, 0.0017, -0.2088, -0.0233, 0.0017, -0.2088, -0.0203, 0.0039, -0.2088, -0.0233, 0.0039, -0.2088, -0.0203, 0.0017, -0.2075, -0.0203, 0.0039, -0.2075, -0.0255, -0.0, -0.214, -0.0277, -0.0, -0.214, -0.0255, -0.0, -0.2118, -0.0277, -0.0, -0.2118, -0.0298, -0.0, -0.214, -0.032, -0.0, -0.214, -0.0298, -0.0, -0.2118, -0.032, -0.0, -0.2118, -0.0255, -0.0, -0.2088, -0.0277, -0.0, -0.2088, -0.0298, -0.0, -0.2088, -0.0319, -0.0, -0.2088, -0.0255, 0.0017, -0.2088, -0.0277, 0.0017, -0.2088, -0.0255, 0.0039, -0.2088, -0.0277, 0.0039, -0.2088, -0.0298, 0.0017, -0.2088, -0.032, 0.0017, -0.2088, -0.0298, 0.0039, -0.2088, -0.032, 0.0039, -0.2088, -0.0342, -0.0, -0.2196, -0.0363, -0.0, -0.2196,
-0.0385, -0.0, -0.2196, -0.0407, -0.0, -0.2196, -0.0342, 0.0017, -0.2196, -0.0363, 0.0017, -0.2196, -0.0342, 0.0039, -0.2196, -0.0363, 0.0039, -0.2196, -0.0385, 0.0017, -0.2196, -0.0407, 0.0017, -0.2196, -0.0385, 0.0039, -0.2196, -0.0407, 0.0039, -0.2196, -0.0342, -0.0, -0.2183, -0.0363, -0.0, -0.2183, -0.0342, -0.0, -0.2161, -0.0363, -0.0, -0.2161, -0.0385, -0.0, -0.2183, -0.0407, -0.0, -0.2183, -0.0385, -0.0, -0.2161, -0.0407, -0.0, -0.2161, -0.0429, -0.0, -0.2196, -0.045, -0.0, -0.2196, -0.0472, -0.0, -0.2196, -0.0487, -0.0, -0.2196, -0.0429, 0.0017, -0.2196, -0.045, 0.0017, -0.2196, -0.0429, 0.0039, -0.2196, -0.045, 0.0039, -0.2196, -0.0472, 0.0017, -0.2196, -0.0487, 0.0017, -0.2196, -0.0472, 0.0039, -0.2196, -0.0487, 0.0039, -0.2196,
-0.0429, -0.0, -0.2183, -0.045, -0.0, -0.2183, -0.0429, -0.0, -0.2161, -0.045, -0.0, -0.2161, -0.0472, -0.0, -0.2183, -0.0487, -0.0, -0.2183, -0.0472, -0.0, -0.2161, -0.0487, -0.0, -0.2161, -0.0487, 0.0017, -0.2183, -0.0487, 0.0039, -0.2183, -0.0487, 0.0017, -0.2161, -0.0487, 0.0039, -0.2161, -0.0342, -0.0, -0.214, -0.0363, -0.0, -0.214, -0.0342, -0.0, -0.2118, -0.0363, -0.0, -0.2118, -0.0385, -0.0, -0.214, -0.0407, -0.0, -0.214, -0.0385, -0.0, -0.2118, -0.0407, -0.0, -0.2118, -0.0339, -0.0, -0.2085, -0.0363, -0.0, -0.2096, -0.0347, -0.0, -0.2083, -0.0361, -0.0, -0.2076, -0.0385, -0.0, -0.2096, -0.0407, -0.0, -0.2096, -0.0385, -0.0, -0.2075, -0.0407, -0.0, -0.2075, -0.0334, 0.0017, -0.2086, -0.0334, 0.0039, -0.2086,
-0.0345, 0.0017, -0.2083, -0.0361, 0.0017, -0.2076, -0.0345, 0.0039, -0.2083, -0.0361, 0.0039, -0.2076, -0.0429, -0.0, -0.214, -0.045, -0.0, -0.214, -0.0429, -0.0, -0.2118, -0.045, -0.0, -0.2118, -0.0472, -0.0, -0.214, -0.0487, -0.0, -0.214, -0.0472, -0.0, -0.2118, -0.0487, -0.0, -0.2118, -0.0487, 0.0017, -0.214, -0.0487, 0.0039, -0.214, -0.0487, 0.0017, -0.2118, -0.0487, 0.0039, -0.2118, -0.0429, -0.0, -0.2096, -0.045, -0.0, -0.2096, -0.0429, -0.0, -0.2075, -0.045, -0.0, -0.2075, -0.0472, -0.0, -0.2096, -0.0487, -0.0, -0.2096, -0.0472, -0.0, -0.2075, -0.0487, -0.0, -0.2075, -0.0487, 0.0017, -0.2096, -0.0487, 0.0039, -0.2096, -0.0487, 0.0017, -0.2075, -0.0487, 0.0039, -0.2075, -0.0085, -0.0, -0.2053, -0.0103, -0.0, -0.2053,
-0.0085, -0.0, -0.2031, -0.0103, -0.0, -0.2031, -0.0125, -0.0, -0.2053, -0.0146, -0.0, -0.2053, -0.0125, -0.0, -0.2031, -0.0146, -0.0, -0.2031, -0.0085, 0.0017, -0.2053, -0.0085, 0.0039, -0.2053, -0.0085, 0.0017, -0.2031, -0.0085, 0.0039, -0.2031, -0.0085, -0.0, -0.2009, -0.0103, -0.0, -0.2009, -0.0085, -0.0, -0.1988, -0.0103, -0.0, -0.1988, -0.0125, -0.0, -0.2009, -0.0146, -0.0, -0.2009, -0.0125, -0.0, -0.1988, -0.0146, -0.0, -0.1988, -0.0085, 0.0017, -0.2009, -0.0085, 0.0039, -0.2009, -0.0085, 0.0017, -0.1988, -0.0085, 0.0039, -0.1988, -0.0085, -0.0, -0.1966, -0.0103, -0.0, -0.1966, -0.0085, -0.0, -0.1944, -0.0103, -0.0, -0.1944, -0.0125, -0.0, -0.1966, -0.0146, -0.0, -0.1966, -0.0125, -0.0, -0.1944, -0.0146, -0.0, -0.1944,
-0.0085, 0.0017, -0.1966, -0.0085, 0.0039, -0.1966, -0.0085, 0.0017, -0.1944, -0.0085, 0.0039, -0.1944, -0.0085, -0.0, -0.1923, -0.0103, -0.0, -0.1923, -0.0085, -0.0, -0.1901, -0.0103, -0.0, -0.1901, -0.0125, -0.0, -0.1923, -0.0146, -0.0, -0.1923, -0.0125, -0.0, -0.1901, -0.0146, -0.0, -0.1901, -0.0085, 0.0017, -0.1923, -0.0085, 0.0039, -0.1923, -0.0085, 0.0017, -0.1901, -0.0085, 0.0039, -0.1901, -0.0168, -0.0, -0.2053, -0.019, -0.0, -0.2053, -0.0168, -0.0, -0.2031, -0.019, -0.0, -0.2031, -0.0203, -0.0, -0.2053, -0.0203, -0.0, -0.2031, -0.0203, 0.0017, -0.2053, -0.0203, 0.0039, -0.2053, -0.0203, 0.0017, -0.2031, -0.0203, 0.0039, -0.2031, -0.0168, -0.0, -0.2009, -0.019, -0.0, -0.2009, -0.0168, -0.0, -0.1988, -0.019, -0.0, -0.1988,
-0.0203, -0.0, -0.2009, -0.0203, -0.0, -0.1988, -0.0203, 0.0017, -0.2009, -0.0203, 0.0039, -0.2009, -0.0203, 0.0017, -0.1988, -0.0203, 0.0039, -0.1988, -0.0168, -0.0, -0.1966, -0.019, -0.0, -0.1966, -0.0168, -0.0, -0.1944, -0.019, -0.0, -0.1944, -0.0203, -0.0, -0.1966, -0.0203, -0.0, -0.1944, -0.0203, 0.0017, -0.1966, -0.0203, 0.0039, -0.1966, -0.0203, 0.0017, -0.1944, -0.0203, 0.0039, -0.1944, -0.0168, -0.0, -0.1923, -0.019, -0.0, -0.1923, -0.0168, -0.0, -0.1901, -0.019, -0.0, -0.1901, -0.0203, -0.0, -0.1923, -0.0203, -0.0, -0.1901, -0.0203, 0.0017, -0.1923, -0.0203, 0.0039, -0.1923, -0.0203, 0.0017, -0.1901, -0.0203, 0.0039, -0.1901, -0.0085, -0.0, -0.1879, -0.0103, -0.0, -0.1879, -0.0085, -0.0, -0.1858, -0.0103, -0.0, -0.1858,
-0.0125, -0.0, -0.1879, -0.0146, -0.0, -0.1879, -0.0125, -0.0, -0.1858, -0.0146, -0.0, -0.1858, -0.0085, 0.0017, -0.1879, -0.0085, 0.0039, -0.1879, -0.0085, 0.0017, -0.1858, -0.0085, 0.0039, -0.1858, -0.0085, -0.0, -0.1836, -0.0103, -0.0, -0.1836, -0.0085, -0.0, -0.1814, -0.0103, -0.0, -0.1814, -0.0125, -0.0, -0.1836, -0.0146, -0.0, -0.1836, -0.0125, -0.0, -0.1814, -0.0146, -0.0, -0.1814, -0.0085, 0.0017, -0.1836, -0.0085, 0.0039, -0.1836, -0.0085, 0.0017, -0.1814, -0.0085, 0.0039, -0.1814, -0.0085, -0.0, -0.1792, -0.0103, -0.0, -0.1792, -0.0085, -0.0, -0.1771, -0.0103, -0.0, -0.1771, -0.0125, -0.0, -0.1792, -0.0146, -0.0, -0.1792, -0.0125, -0.0, -0.1771, -0.0146, -0.0, -0.1771, -0.0085, 0.0017, -0.1792, -0.0085, 0.0039, -0.1792,
-0.0085, 0.0017, -0.1771, -0.0085, 0.0039, -0.1771, -0.0085, -0.0, -0.1749, -0.0103, -0.0, -0.1749, -0.0085, -0.0, -0.1727, -0.0103, -0.0, -0.1727, -0.0125, -0.0, -0.1749, -0.0146, -0.0, -0.1749, -0.0125, -0.0, -0.1727, -0.0146, -0.0, -0.1727, -0.0085, 0.0017, -0.1749, -0.0085, 0.0039, -0.1749, -0.0085, 0.0017, -0.1727, -0.0085, 0.0039, -0.1727, -0.0168, -0.0, -0.1879, -0.019, -0.0, -0.1879, -0.0168, -0.0, -0.1858, -0.019, -0.0, -0.1858, -0.0203, -0.0, -0.1879, -0.0203, -0.0, -0.1858, -0.0203, 0.0017, -0.1879, -0.0203, 0.0039, -0.1879, -0.0203, 0.0017, -0.1858, -0.0203, 0.0039, -0.1858, -0.0168, -0.0, -0.1836, -0.019, -0.0, -0.1836, -0.0168, -0.0, -0.1814, -0.019, -0.0, -0.1814, -0.0203, -0.0, -0.1836, -0.0203, -0.0, -0.1814,
-0.0203, 0.0017, -0.1836, -0.0203, 0.0039, -0.1836, -0.0203, 0.0017, -0.1814, -0.0203, 0.0039, -0.1814, -0.0168, -0.0, -0.1792, -0.019, -0.0, -0.1792, -0.0168, -0.0, -0.1771, -0.019, -0.0, -0.1771, -0.0203, -0.0, -0.1792, -0.0203, -0.0, -0.1771, -0.0203, 0.0017, -0.1792, -0.0203, 0.0039, -0.1792, -0.0203, 0.0017, -0.1771, -0.0203, 0.0039, -0.1771, -0.0168, -0.0, -0.1749, -0.019, -0.0, -0.1749, -0.0168, -0.0, -0.1727, -0.019, -0.0, -0.1727, -0.0203, -0.0, -0.1749, -0.0203, -0.0, -0.1727, -0.0203, 0.0017, -0.1749, -0.0203, 0.0039, -0.1749, -0.0203, 0.0017, -0.1727, -0.0203, 0.0039, -0.1727, -0.0372, -0.0, -0.206, -0.0375, -0.0, -0.2051, -0.0407, -0.0, -0.2053, -0.0378, -0.0, -0.203, -0.0407, -0.0, -0.2031, -0.0372, 0.0017, -0.2059,
-0.0372, 0.0039, -0.2059, -0.0376, 0.0017, -0.2048, -0.0376, 0.0039, -0.2048, -0.0378, 0.0017, -0.2031, -0.0378, 0.0039, -0.2031, -0.0379, -0.0, -0.2009, -0.0407, -0.0, -0.2009, -0.0379, -0.0, -0.1988, -0.0407, -0.0, -0.1988, -0.0379, 0.0017, -0.2009, -0.0379, 0.0039, -0.2009, -0.0379, 0.0017, -0.1988, -0.0379, 0.0039, -0.1988, -0.0429, -0.0, -0.2053, -0.045, -0.0, -0.2053, -0.0429, -0.0, -0.2031, -0.045, -0.0, -0.2031, -0.0472, -0.0, -0.2053, -0.0487, -0.0, -0.2053, -0.0472, -0.0, -0.2031, -0.0487, -0.0, -0.2031, -0.0487, 0.0017, -0.2053, -0.0487, 0.0039, -0.2053, -0.0487, 0.0017, -0.2031, -0.0487, 0.0039, -0.2031, -0.0429, -0.0, -0.2009, -0.045, -0.0, -0.2009, -0.0429, -0.0, -0.1988, -0.045, -0.0, -0.1988, -0.0472, -0.0, -0.2009,
-0.0487, -0.0, -0.2009, -0.0472, -0.0, -0.1988, -0.0487, -0.0, -0.1988, -0.0487, 0.0017, -0.2009, -0.0487, 0.0039, -0.2009, -0.0487, 0.0017, -0.1988, -0.0487, 0.0039, -0.1988, -0.0379, -0.0, -0.1966, -0.0407, -0.0, -0.1966, -0.0379, -0.0, -0.1944, -0.0407, -0.0, -0.1944, -0.0379, 0.0017, -0.1966, -0.0379, 0.0039, -0.1966, -0.0379, 0.0017, -0.1944, -0.0379, 0.0039, -0.1944, -0.0379, -0.0, -0.1923, -0.0407, -0.0, -0.1923, -0.0379, -0.0, -0.1901, -0.0407, -0.0, -0.1901, -0.0379, 0.0017, -0.1923, -0.0379, 0.0039, -0.1923, -0.0379, 0.0017, -0.1901, -0.0379, 0.0039, -0.1901, -0.0429, -0.0, -0.1966, -0.045, -0.0, -0.1966, -0.0429, -0.0, -0.1944, -0.045, -0.0, -0.1944, -0.0472, -0.0, -0.1966, -0.0487, -0.0, -0.1966, -0.0472, -0.0, -0.1944,
-0.0487, -0.0, -0.1944, -0.0487, 0.0017, -0.1966, -0.0487, 0.0039, -0.1966, -0.0487, 0.0017, -0.1944, -0.0487, 0.0039, -0.1944, -0.0429, -0.0, -0.1923, -0.045, -0.0, -0.1923, -0.0429, -0.0, -0.1901, -0.045, -0.0, -0.1901, -0.0472, -0.0, -0.1923, -0.0487, -0.0, -0.1923, -0.0472, -0.0, -0.1901, -0.0487, -0.0, -0.1901, -0.0487, 0.0017, -0.1923, -0.0487, 0.0039, -0.1923, -0.0487, 0.0017, -0.1901, -0.0487, 0.0039, -0.1901, -0.0379, -0.0, -0.1879, -0.0407, -0.0, -0.1879, -0.0379, -0.0, -0.1858, -0.0407, -0.0, -0.1858, -0.0379, 0.0017, -0.1879, -0.0379, 0.0039, -0.1879, -0.0379, 0.0017, -0.1858, -0.0379, 0.0039, -0.1858, -0.0379, -0.0, -0.1836, -0.0407, -0.0, -0.1836, -0.0379, -0.0, -0.1814, -0.0407, -0.0, -0.1814, -0.0379, 0.0017, -0.1836,
-0.0379, 0.0039, -0.1836, -0.0379, 0.0017, -0.1814, -0.0379, 0.0039, -0.1814, -0.0429, -0.0, -0.1879, -0.045, -0.0, -0.1879, -0.0429, -0.0, -0.1858, -0.045, -0.0, -0.1858, -0.0472, -0.0, -0.1879, -0.0487, -0.0, -0.1879, -0.0472, -0.0, -0.1858, -0.0487, -0.0, -0.1858, -0.0487, 0.0017, -0.1879, -0.0487, 0.0039, -0.1879, -0.0487, 0.0017, -0.1858, -0.0487, 0.0039, -0.1858, -0.0429, -0.0, -0.1836, -0.045, -0.0, -0.1836, -0.0429, -0.0, -0.1814, -0.045, -0.0, -0.1814, -0.0472, -0.0, -0.1836, -0.0487, -0.0, -0.1836, -0.0472, -0.0, -0.1814, -0.0487, -0.0, -0.1814, -0.0487, 0.0017, -0.1836, -0.0487, 0.0039, -0.1836, -0.0487, 0.0017, -0.1814, -0.0487, 0.0039, -0.1814, -0.0379, -0.0, -0.1792, -0.0407, -0.0, -0.1792, -0.0379, -0.0, -0.1771,
-0.0407, -0.0, -0.1771, -0.0379, 0.0017, -0.1792, -0.0379, 0.0039, -0.1792, -0.0379, 0.0017, -0.1771, -0.0379, 0.0039, -0.1771, -0.0379, -0.0, -0.1749, -0.0407, -0.0, -0.1749, -0.0379, -0.0, -0.1727, -0.0407, -0.0, -0.1727, -0.0379, 0.0017, -0.1749, -0.0379, 0.0039, -0.1749, -0.0379, 0.0017, -0.1727, -0.0379, 0.0039, -0.1727, -0.0429, -0.0, -0.1792, -0.045, -0.0, -0.1792, -0.0429, -0.0, -0.1771, -0.045, -0.0, -0.1771, -0.0472, -0.0, -0.1792, -0.0487, -0.0, -0.1792, -0.0472, -0.0, -0.1771, -0.0487, -0.0, -0.1771, -0.0487, 0.0017, -0.1792, -0.0487, 0.0039, -0.1792, -0.0487, 0.0017, -0.1771, -0.0487, 0.0039, -0.1771, -0.0429, -0.0, -0.1749, -0.045, -0.0, -0.1749, -0.0429, -0.0, -0.1727, -0.045, -0.0, -0.1727, -0.0472, -0.0, -0.1749,
-0.0487, -0.0, -0.1749, -0.0472, -0.0, -0.1727, -0.0487, -0.0, -0.1727, -0.0487, 0.0017, -0.1749, -0.0487, 0.0039, -0.1749, -0.0487, 0.0017, -0.1727, -0.0487, 0.0039, -0.1727, -0.0085, -0.0, -0.1706, -0.0103, -0.0, -0.1706, -0.0085, -0.0, -0.1684, -0.0103, -0.0, -0.1684, -0.0125, -0.0, -0.1706, -0.0146, -0.0, -0.1706, -0.0125, -0.0, -0.1684, -0.0146, -0.0, -0.1684, -0.0085, 0.0017, -0.1706, -0.0085, 0.0039, -0.1706, -0.0085, 0.0017, -0.1684, -0.0085, 0.0039, -0.1684, -0.0085, -0.0, -0.1662, -0.0103, -0.0, -0.1662, -0.0085, -0.0, -0.1641, -0.0103, -0.0, -0.1641, -0.0125, -0.0, -0.1662, -0.0146, -0.0, -0.1662, -0.0125, -0.0, -0.1641, -0.0146, -0.0, -0.1641, -0.0085, 0.0017, -0.1662, -0.0085, 0.0039, -0.1662, -0.0085, 0.0017, -0.1641,
-0.0085, 0.0039, -0.1641, -0.0085, -0.0, -0.1627, -0.0103, -0.0, -0.1627, -0.0125, -0.0, -0.1627, -0.0146, -0.0, -0.1627, -0.0085, 0.0017, -0.1627, -0.0103, 0.0017, -0.1627, -0.0085, 0.0039, -0.1627, -0.0103, 0.0039, -0.1627, -0.0125, 0.0017, -0.1627, -0.0146, 0.0017, -0.1627, -0.0125, 0.0039, -0.1627, -0.0146, 0.0039, -0.1627, -0.0168, -0.0, -0.1706, -0.019, -0.0, -0.1706, -0.0168, -0.0, -0.1684, -0.019, -0.0, -0.1684, -0.0203, -0.0, -0.1706, -0.0203, -0.0, -0.1684, -0.0203, 0.0017, -0.1706, -0.0203, 0.0039, -0.1706, -0.0203, 0.0017, -0.1684, -0.0203, 0.0039, -0.1684, -0.0168, -0.0, -0.1662, -0.019, -0.0, -0.1662, -0.0168, -0.0, -0.1641, -0.019, -0.0, -0.1641, -0.0203, -0.0, -0.1662, -0.0203, -0.0, -0.1641, -0.0203, 0.0017, -0.1662,
-0.0203, 0.0039, -0.1662, -0.0203, 0.0017, -0.1641, -0.0203, 0.0039, -0.1641, -0.0168, -0.0, -0.1627, -0.019, -0.0, -0.1627, -0.0203, -0.0, -0.1627, -0.0168, 0.0017, -0.1627, -0.019, 0.0017, -0.1627, -0.0168, 0.0039, -0.1627, -0.019, 0.0039, -0.1627, -0.0203, 0.0017, -0.1627, -0.0203, 0.0039, -0.1627, 0.0005, -0.0, -0.151, -0.0016, -0.0, -0.151, -0.0038, -0.0, -0.151, -0.006, -0.0, -0.151, 0.0005, 0.0017, -0.151, -0.0016, 0.0017, -0.151, 0.0005, 0.0039, -0.151, -0.0016, 0.0039, -0.151, -0.0038, 0.0017, -0.151, -0.006, 0.0017, -0.151, -0.0038, 0.0039, -0.151, -0.006, 0.0039, -0.151, 0.0005, -0.0, -0.1489, -0.0016, -0.0, -0.1489, 0.0005, -0.0, -0.1467, -0.0016, -0.0, -0.1467, -0.0038, -0.0, -0.1489, -0.006, -0.0, -0.1489,
-0.0038, -0.0, -0.1467, -0.006, -0.0, -0.1467, -0.0081, -0.0, -0.151, -0.0103, -0.0, -0.151, -0.0125, -0.0, -0.151, -0.0146, -0.0, -0.151, -0.0081, 0.0017, -0.151, -0.0103, 0.0017, -0.151, -0.0081, 0.0039, -0.151, -0.0103, 0.0039, -0.151, -0.0125, 0.0017, -0.151, -0.0146, 0.0017, -0.151, -0.0125, 0.0039, -0.151, -0.0146, 0.0039, -0.151, -0.0081, -0.0, -0.1489, -0.0103, -0.0, -0.1489, -0.0081, -0.0, -0.1467, -0.0103, -0.0, -0.1467, -0.0125, -0.0, -0.1489, -0.0146, -0.0, -0.1489, -0.0125, -0.0, -0.1467, -0.0146, -0.0, -0.1467, 0.0005, -0.0, -0.1445, -0.0016, -0.0, -0.1445, 0.0005, -0.0, -0.1424, -0.0016, -0.0, -0.1424, -0.0038, -0.0, -0.1445, -0.006, -0.0, -0.1445, -0.0038, -0.0, -0.1424, -0.006, -0.0, -0.1424,
0.0005, -0.0, -0.1392, -0.0016, -0.0, -0.1392, -0.0038, -0.0, -0.1392, -0.006, -0.0, -0.1392, 0.0005, 0.0017, -0.1392, -0.0016, 0.0017, -0.1392, 0.0005, 0.0039, -0.1392, -0.0016, 0.0039, -0.1392, -0.0038, 0.0017, -0.1392, -0.006, 0.0017, -0.1392, -0.0038, 0.0039, -0.1392, -0.006, 0.0039, -0.1392, -0.0081, -0.0, -0.1445, -0.0103, -0.0, -0.1445, -0.0081, -0.0, -0.1424, -0.0103, -0.0, -0.1424, -0.0125, -0.0, -0.1445, -0.0146, -0.0, -0.1445, -0.0125, -0.0, -0.1424, -0.0146, -0.0, -0.1424, -0.0081, -0.0, -0.1392, -0.0103, -0.0, -0.1392, -0.0125, -0.0, -0.1392, -0.0146, -0.0, -0.1392, -0.0081, 0.0017, -0.1392, -0.0103, 0.0017, -0.1392, -0.0081, 0.0039, -0.1392, -0.0103, 0.0039, -0.1392, -0.0125, 0.0017, -0.1392, -0.0146, 0.0017, -0.1392,
-0.0125, 0.0039, -0.1392, -0.0146, 0.0039, -0.1392, -0.0168, -0.0, -0.151, -0.019, -0.0, -0.151, -0.0212, -0.0, -0.151, -0.0233, -0.0, -0.151, -0.0168, 0.0017, -0.151, -0.019, 0.0017, -0.151, -0.0168, 0.0039, -0.151, -0.019, 0.0039, -0.151, -0.0212, 0.0017, -0.151, -0.0233, 0.0017, -0.151, -0.0212, 0.0039, -0.151, -0.0233, 0.0039, -0.151, -0.0168, -0.0, -0.1489, -0.019, -0.0, -0.1489, -0.0168, -0.0, -0.1467, -0.019, -0.0, -0.1467, -0.0212, -0.0, -0.1489, -0.0233, -0.0, -0.1489, -0.0212, -0.0, -0.1467, -0.0233, -0.0, -0.1467, -0.0255, -0.0, -0.151, -0.0277, -0.0, -0.151, -0.0298, -0.0, -0.151, -0.032, -0.0, -0.151, -0.0255, 0.0017, -0.151, -0.0277, 0.0017, -0.151, -0.0255, 0.0039, -0.151, -0.0277, 0.0039, -0.151,
-0.0298, 0.0017, -0.151, -0.032, 0.0017, -0.151, -0.0298, 0.0039, -0.151, -0.032, 0.0039, -0.151, -0.0255, -0.0, -0.1489, -0.0277, -0.0, -0.1489, -0.0255, -0.0, -0.1467, -0.0277, -0.0, -0.1467, -0.0298, -0.0, -0.1489, -0.032, -0.0, -0.1489, -0.0298, -0.0, -0.1467, -0.032, -0.0, -0.1467, -0.0168, -0.0, -0.1445, -0.019, -0.0, -0.1445, -0.0168, -0.0, -0.1424, -0.019, -0.0, -0.1424, -0.0212, -0.0, -0.1445, -0.0233, -0.0, -0.1445, -0.0212, -0.0, -0.1424, -0.0233, -0.0, -0.1424, -0.0168, -0.0, -0.1392, -0.019, -0.0, -0.1392, -0.0212, -0.0, -0.1392, -0.0233, -0.0, -0.1392, -0.0168, 0.0017, -0.1392, -0.019, 0.0017, -0.1392, -0.0168, 0.0039, -0.1392, -0.019, 0.0039, -0.1392, -0.0212, 0.0017, -0.1392, -0.0233, 0.0017, -0.1392,
-0.0212, 0.0039, -0.1392, -0.0233, 0.0039, -0.1392, -0.0255, -0.0, -0.1445, -0.0277, -0.0, -0.1445, -0.0255, -0.0, -0.1424, -0.0277, -0.0, -0.1424, -0.0298, -0.0, -0.1445, -0.032, -0.0, -0.1445, -0.0298, -0.0, -0.1424, -0.032, -0.0, -0.1424, -0.0255, -0.0, -0.1392, -0.0277, -0.0, -0.1392, -0.0298, -0.0, -0.1392, -0.032, -0.0, -0.1392, -0.0255, 0.0017, -0.1392, -0.0277, 0.0017, -0.1392, -0.0255, 0.0039, -0.1392, -0.0277, 0.0039, -0.1392, -0.0298, 0.0017, -0.1392, -0.032, 0.0017, -0.1392, -0.0298, 0.0039, -0.1392, -0.032, 0.0039, -0.1392, -0.0379, -0.0, -0.1706, -0.0407, -0.0, -0.1706, -0.0379, -0.0, -0.1684, -0.0407, -0.0, -0.1684, -0.0379, 0.0017, -0.1706, -0.0379, 0.0039, -0.1706, -0.0379, 0.0017, -0.1684, -0.0379, 0.0039, -0.1684,
-0.0379, -0.0, -0.1662, -0.0407, -0.0, -0.1662, -0.0379, -0.0, -0.1641, -0.0407, -0.0, -0.1641, -0.0379, 0.0017, -0.1662, -0.0379, 0.0039, -0.1662, -0.0379, 0.0017, -0.1641, -0.0379, 0.0039, -0.1641, -0.0429, -0.0, -0.1706, -0.045, -0.0, -0.1706, -0.0429, -0.0, -0.1684, -0.045, -0.0, -0.1684, -0.0472, -0.0, -0.1706, -0.0487, -0.0, -0.1706, -0.0472, -0.0, -0.1684, -0.0487, -0.0, -0.1684, -0.0487, 0.0017, -0.1706, -0.0487, 0.0039, -0.1706, -0.0487, 0.0017, -0.1684, -0.0487, 0.0039, -0.1684, -0.0429, -0.0, -0.1662, -0.045, -0.0, -0.1662, -0.0429, -0.0, -0.1641, -0.045, -0.0, -0.1641, -0.0472, -0.0, -0.1662, -0.0487, -0.0, -0.1662, -0.0472, -0.0, -0.1641, -0.0487, -0.0, -0.1641, -0.0487, 0.0017, -0.1662, -0.0487, 0.0039, -0.1662,
-0.0487, 0.0017, -0.1641, -0.0487, 0.0039, -0.1641, -0.0379, -0.0, -0.1619, -0.0407, -0.0, -0.1619, -0.0379, -0.0, -0.1588, -0.0407, -0.0, -0.1588, -0.0379, 0.0017, -0.1619, -0.0379, 0.0039, -0.1619, -0.0379, 0.0017, -0.1588, -0.0407, 0.0017, -0.1588, -0.0379, 0.0039, -0.1588, -0.0407, 0.0039, -0.1588, -0.0429, -0.0, -0.1619, -0.045, -0.0, -0.1619, -0.0429, -0.0, -0.1588, -0.045, -0.0, -0.1588, -0.0472, -0.0, -0.1619, -0.0487, -0.0, -0.1619, -0.0472, -0.0, -0.1588, -0.0487, -0.0, -0.1588, -0.0429, 0.0017, -0.1588, -0.045, 0.0017, -0.1588, -0.0429, 0.0039, -0.1588, -0.045, 0.0039, -0.1588, -0.0487, 0.0017, -0.1619, -0.0487, 0.0039, -0.1619, -0.0472, 0.0017, -0.1588, -0.0487, 0.0017, -0.1588, -0.0472, 0.0039, -0.1588, -0.0487, 0.0039, -0.1588,
-0.0342, -0.0, -0.151, -0.0363, -0.0, -0.151, -0.0385, -0.0, -0.151, -0.0407, -0.0, -0.151, -0.0342, 0.0017, -0.151, -0.0363, 0.0017, -0.151, -0.0342, 0.0039, -0.151, -0.0363, 0.0039, -0.151, -0.0385, 0.0017, -0.151, -0.0407, 0.0017, -0.151, -0.0385, 0.0039, -0.151, -0.0407, 0.0039, -0.151, -0.0342, -0.0, -0.1489, -0.0363, -0.0, -0.1489, -0.0342, -0.0, -0.1467, -0.0363, -0.0, -0.1467, -0.0385, -0.0, -0.1489, -0.0407, -0.0, -0.1489, -0.0385, -0.0, -0.1467, -0.0407, -0.0, -0.1467, -0.0429, -0.0, -0.151, -0.045, -0.0, -0.151, -0.0472, -0.0, -0.151, -0.0487, -0.0, -0.151, -0.0429, 0.0017, -0.151, -0.045, 0.0017, -0.151, -0.0429, 0.0039, -0.151, -0.045, 0.0039, -0.151, -0.0472, 0.0017, -0.151, -0.0487, 0.0017, -0.151,
-0.0472, 0.0039, -0.151, -0.0487, 0.0039, -0.151, -0.0429, -0.0, -0.1489, -0.045, -0.0, -0.1489, -0.0429, -0.0, -0.1467, -0.045, -0.0, -0.1467, -0.0472, -0.0, -0.1489, -0.0487, -0.0, -0.1489, -0.0472, -0.0, -0.1467, -0.0487, -0.0, -0.1467, -0.0487, 0.0017, -0.1489, -0.0487, 0.0039, -0.1489, -0.0487, 0.0017, -0.1467, -0.0487, 0.0039, -0.1467, -0.0342, -0.0, -0.1445, -0.0363, -0.0, -0.1445, -0.0342, -0.0, -0.1424, -0.0363, -0.0, -0.1424, -0.0385, -0.0, -0.1445, -0.0407, -0.0, -0.1445, -0.0385, -0.0, -0.1424, -0.0407, -0.0, -0.1424, -0.0342, -0.0, -0.1392, -0.0363, -0.0, -0.1392, -0.0385, -0.0, -0.1392, -0.0407, -0.0, -0.1392, -0.0342, 0.0017, -0.1392, -0.0363, 0.0017, -0.1392, -0.0342, 0.0039, -0.1392, -0.0363, 0.0039, -0.1392,
-0.0385, 0.0017, -0.1392, -0.0407, 0.0017, -0.1392, -0.0385, 0.0039, -0.1392, -0.0407, 0.0039, -0.1392, -0.0429, -0.0, -0.1445, -0.045, -0.0, -0.1445, -0.0429, -0.0, -0.1424, -0.045, -0.0, -0.1424, -0.0472, -0.0, -0.1445, -0.0487, -0.0, -0.1445, -0.0472, -0.0, -0.1424, -0.0487, -0.0, -0.1424, -0.0487, 0.0017, -0.1445, -0.0487, 0.0039, -0.1445, -0.0487, 0.0017, -0.1424, -0.0487, 0.0039, -0.1424, -0.0429, -0.0, -0.1392, -0.045, -0.0, -0.1392, -0.0472, -0.0, -0.1392, -0.0487, -0.0, -0.1392, -0.0429, 0.0017, -0.1392, -0.045, 0.0017, -0.1392, -0.0429, 0.0039, -0.1392, -0.045, 0.0039, -0.1392, -0.0472, 0.0017, -0.1392, -0.0487, 0.0017, -0.1392, -0.0472, 0.0039, -0.1392, -0.0487, 0.0039, -0.1392, 0.0005, -0.0, -0.0775, -0.0016, -0.0, -0.0775,
-0.0038, -0.0, -0.0775, -0.006, -0.0, -0.0775, 0.0005, 0.0017, -0.0775, -0.0016, 0.0017, -0.0775, 0.0005, 0.0039, -0.0775, -0.0016, 0.0039, -0.0775, -0.0038, 0.0017, -0.0775, -0.006, 0.0017, -0.0775, -0.0038, 0.0039, -0.0775, -0.006, 0.0039, -0.0775, -0.0081, -0.0, -0.0775, -0.0103, -0.0, -0.0775, -0.0125, -0.0, -0.0775, -0.0146, -0.0, -0.0775, -0.0081, 0.0017, -0.0775, -0.0103, 0.0017, -0.0775, -0.0081, 0.0039, -0.0775, -0.0103, 0.0039, -0.0775, -0.0125, 0.0017, -0.0775, -0.0146, 0.0017, -0.0775, -0.0125, 0.0039, -0.0775, -0.0146, 0.0039, -0.0775, 0.0005, -0.0, -0.0751, -0.0016, -0.0, -0.0751, 0.0005, -0.0, -0.0729, -0.0016, -0.0, -0.0729, -0.0038, -0.0, -0.0751, -0.006, -0.0, -0.0751, -0.0038, -0.0, -0.0729, -0.006, -0.0, -0.0729,
0.0005, -0.0, -0.0707, -0.0016, -0.0, -0.0707, 0.0005, -0.0, -0.0686, -0.0016, -0.0, -0.0686, -0.0038, -0.0, -0.0707, -0.006, -0.0, -0.0707, -0.0038, -0.0, -0.0686, -0.006, -0.0, -0.0686, -0.0081, -0.0, -0.0751, -0.0103, -0.0, -0.0751, -0.0081, -0.0, -0.0729, -0.0103, -0.0, -0.0729, -0.0125, -0.0, -0.0751, -0.0146, -0.0, -0.0751, -0.0125, -0.0, -0.0729, -0.0146, -0.0, -0.0729, -0.0081, -0.0, -0.0707, -0.0103, -0.0, -0.0707, -0.0081, -0.0, -0.0686, -0.0103, -0.0, -0.0686, -0.0125, -0.0, -0.0707, -0.0146, -0.0, -0.0707, -0.0125, -0.0, -0.0686, -0.0146, -0.0, -0.0686, -0.0168, -0.0, -0.0775, -0.019, -0.0, -0.0775, -0.0212, -0.0, -0.0775, -0.0233, -0.0, -0.0775, -0.0168, 0.0017, -0.0775, -0.019, 0.0017, -0.0775,
-0.0168, 0.0039, -0.0775, -0.019, 0.0039, -0.0775, -0.0212, 0.0017, -0.0775, -0.0233, 0.0017, -0.0775, -0.0212, 0.0039, -0.0775, -0.0233, 0.0039, -0.0775, -0.0255, -0.0, -0.0775, -0.0277, -0.0, -0.0775, -0.0298, -0.0, -0.0775, -0.032, -0.0, -0.0775, -0.0255, 0.0017, -0.0775, -0.0277, 0.0017, -0.0775, -0.0255, 0.0039, -0.0775, -0.0277, 0.0039, -0.0775, -0.0298, 0.0017, -0.0775, -0.032, 0.0017, -0.0775, -0.0298, 0.0039, -0.0775, -0.032, 0.0039, -0.0775, -0.0168, -0.0, -0.0751, -0.019, -0.0, -0.0751, -0.0168, -0.0, -0.0729, -0.019, -0.0, -0.0729, -0.0212, -0.0, -0.0751, -0.0233, -0.0, -0.0751, -0.0212, -0.0, -0.0729, -0.0233, -0.0, -0.0729, -0.0168, -0.0, -0.0707, -0.019, -0.0, -0.0707, -0.0168, -0.0, -0.0686, -0.019, -0.0, -0.0686,
-0.0212, -0.0, -0.0707, -0.0233, -0.0, -0.0707, -0.0212, -0.0, -0.0686, -0.0233, -0.0, -0.0686, -0.0255, -0.0, -0.0751, -0.0277, -0.0, -0.0751, -0.0255, -0.0, -0.0729, -0.0277, -0.0, -0.0729, -0.0298, -0.0, -0.0751, -0.032, -0.0, -0.0751, -0.0298, -0.0, -0.0729, -0.032, -0.0, -0.0729, -0.0255, -0.0, -0.0707, -0.0277, -0.0, -0.0707, -0.0255, -0.0, -0.0686, -0.0277, -0.0, -0.0686, -0.0298, -0.0, -0.0707, -0.032, -0.0, -0.0707, -0.0298, -0.0, -0.0686, -0.032, -0.0, -0.0686, -0.0342, -0.0, -0.0775, -0.0363, -0.0, -0.0775, -0.0385, -0.0, -0.0775, -0.0407, -0.0, -0.0775, -0.0342, 0.0017, -0.0775, -0.0363, 0.0017, -0.0775, -0.0342, 0.0039, -0.0775, -0.0363, 0.0039, -0.0775, -0.0385, 0.0017, -0.0775, -0.0407, 0.0017, -0.0775,
-0.0385, 0.0039, -0.0775, -0.0407, 0.0039, -0.0775, -0.0429, -0.0, -0.0775, -0.045, -0.0, -0.0775, -0.0472, -0.0, -0.0775, -0.0487, -0.0, -0.0775, -0.0429, 0.0017, -0.0775, -0.045, 0.0017, -0.0775, -0.0429, 0.0039, -0.0775, -0.045, 0.0039, -0.0775, -0.0472, 0.0017, -0.0775, -0.0487, 0.0017, -0.0775, -0.0472, 0.0039, -0.0775, -0.0487, 0.0039, -0.0775, -0.0342, -0.0, -0.0751, -0.0363, -0.0, -0.0751, -0.0342, -0.0, -0.0729, -0.0363, -0.0, -0.0729, -0.0385, -0.0, -0.0751, -0.0407, -0.0, -0.0751, -0.0385, -0.0, -0.0729, -0.0407, -0.0, -0.0729, -0.0342, -0.0, -0.0707, -0.0363, -0.0, -0.0707, -0.0342, -0.0, -0.0686, -0.0363, -0.0, -0.0686, -0.0385, -0.0, -0.0707, -0.0407, -0.0, -0.0707, -0.0385, -0.0, -0.0686, -0.0407, -0.0, -0.0686,
-0.0429, -0.0, -0.0751, -0.045, -0.0, -0.0751, -0.0429, -0.0, -0.0729, -0.045, -0.0, -0.0729, -0.0472, -0.0, -0.0751, -0.0487, -0.0, -0.0751, -0.0472, -0.0, -0.0729, -0.0487, -0.0, -0.0729, -0.0487, 0.0017, -0.0751, -0.0487, 0.0039, -0.0751, -0.0487, 0.0017, -0.0729, -0.0487, 0.0039, -0.0729, -0.0429, -0.0, -0.0707, -0.045, -0.0, -0.0707, -0.0429, -0.0, -0.0686, -0.045, -0.0, -0.0686, -0.0472, -0.0, -0.0707, -0.0487, -0.0, -0.0707, -0.0472, -0.0, -0.0686, -0.0487, -0.0, -0.0686, -0.0487, 0.0017, -0.0707, -0.0487, 0.0039, -0.0707, -0.0487, 0.0017, -0.0686, -0.0487, 0.0039, -0.0686, 0.0005, -0.0, -0.0667, -0.0016, -0.0, -0.0667, -0.0038, -0.0, -0.0667, -0.006, -0.0, -0.0667, 0.0005, 0.0017, -0.0667, -0.0016, 0.0017, -0.0667,
0.0005, 0.0039, -0.0667, -0.0016, 0.0039, -0.0667, -0.0038, 0.0017, -0.0667, -0.006, 0.0017, -0.0667, -0.0038, 0.0039, -0.0667, -0.006, 0.0039, -0.0667, -0.0081, -0.0, -0.0667, -0.0103, -0.0, -0.0667, -0.0125, -0.0, -0.0667, -0.0146, -0.0, -0.0667, -0.0081, 0.0017, -0.0667, -0.0103, 0.0017, -0.0667, -0.0081, 0.0039, -0.0667, -0.0103, 0.0039, -0.0667, -0.0125, 0.0017, -0.0667, -0.0146, 0.0017, -0.0667, -0.0125, 0.0039, -0.0667, -0.0146, 0.0039, -0.0667, -0.0168, -0.0, -0.0667, -0.019, -0.0, -0.0667, -0.0212, -0.0, -0.0667, -0.0233, -0.0, -0.0667, -0.0168, 0.0017, -0.0667, -0.019, 0.0017, -0.0667, -0.0168, 0.0039, -0.0667, -0.019, 0.0039, -0.0667, -0.0212, 0.0017, -0.0667, -0.0233, 0.0017, -0.0667, -0.0212, 0.0039, -0.0667, -0.0233, 0.0039, -0.0667,
-0.0255, -0.0, -0.0667, -0.0277, -0.0, -0.0667, -0.0298, -0.0, -0.0667, -0.032, -0.0, -0.0667, -0.0255, 0.0017, -0.0667, -0.0277, 0.0017, -0.0667, -0.0255, 0.0039, -0.0667, -0.0277, 0.0039, -0.0667, -0.0298, 0.0017, -0.0667, -0.032, 0.0017, -0.0667, -0.0298, 0.0039, -0.0667, -0.032, 0.0039, -0.0667, 0.0005, -0.0, -0.05, -0.0016, -0.0, -0.05, 0.0005, -0.0, -0.0469, -0.0016, -0.0, -0.0469, -0.0038, -0.0, -0.05, -0.006, -0.0, -0.05, -0.0038, -0.0, -0.0469, -0.006, -0.0, -0.0469, 0.0005, 0.0017, -0.05, -0.0016, 0.0017, -0.05, 0.0005, 0.0039, -0.05, -0.0016, 0.0039, -0.05, -0.0038, 0.0017, -0.05, -0.006, 0.0017, -0.05, -0.0038, 0.0039, -0.05, -0.006, 0.0039, -0.05, 0.0005, -0.0, -0.0447, -0.0016, -0.0, -0.0447,
0.0005, -0.0, -0.0425, -0.0016, -0.0, -0.0425, -0.0038, -0.0, -0.0447, -0.006, -0.0, -0.0447, -0.0038, -0.0, -0.0425, -0.006, -0.0, -0.0425, -0.0081, -0.0, -0.05, -0.0103, -0.0, -0.05, -0.0081, -0.0, -0.0469, -0.0103, -0.0, -0.0469, -0.0125, -0.0, -0.05, -0.0146, -0.0, -0.05, -0.0125, -0.0, -0.0469, -0.0146, -0.0, -0.0469, -0.0081, 0.0017, -0.05, -0.0103, 0.0017, -0.05, -0.0081, 0.0039, -0.05, -0.0103, 0.0039, -0.05, -0.0125, 0.0017, -0.05, -0.0146, 0.0017, -0.05, -0.0125, 0.0039, -0.05, -0.0146, 0.0039, -0.05, -0.0081, -0.0, -0.0447, -0.0103, -0.0, -0.0447, -0.0081, -0.0, -0.0425, -0.0103, -0.0, -0.0425, -0.0125, -0.0, -0.0447, -0.0146, -0.0, -0.0447, -0.0125, -0.0, -0.0425, -0.0146, -0.0, -0.0425,
0.0005, -0.0, -0.0404, -0.0016, -0.0, -0.0404, 0.0013, -0.0, -0.0382, -0.0016, -0.0, -0.0382, -0.0038, -0.0, -0.0404, -0.006, -0.0, -0.0404, -0.0038, -0.0, -0.0382, -0.006, -0.0, -0.0382, 0.0013, 0.0017, -0.0382, 0.0013, 0.0039, -0.0382, 0.0013, -0.0, -0.036, -0.0016, -0.0, -0.036, 0.0013, -0.0, -0.0338, -0.0016, -0.0, -0.0338, -0.0038, -0.0, -0.036, -0.006, -0.0, -0.036, -0.0038, -0.0, -0.0338, -0.006, -0.0, -0.0338, 0.0013, 0.0017, -0.036, 0.0013, 0.0039, -0.036, 0.0013, 0.0017, -0.0338, 0.0013, 0.0039, -0.0338, -0.0081, -0.0, -0.0404, -0.0103, -0.0, -0.0404, -0.0081, -0.0, -0.0382, -0.0105, -0.0, -0.0382, -0.0125, -0.0, -0.0404, -0.0146, -0.0, -0.0404, -0.0125, -0.0, -0.0382, -0.0146, -0.0, -0.0382,
-0.0105, 0.0017, -0.0382, -0.0105, 0.0039, -0.0382, -0.0125, 0.0017, -0.0382, -0.0146, 0.0017, -0.0382, -0.0125, 0.0039, -0.0382, -0.0146, 0.0039, -0.0382, -0.0081, -0.0, -0.036, -0.0105, -0.0, -0.036, -0.0081, -0.0, -0.0338, -0.0105, -0.0, -0.0338, -0.0105, 0.0017, -0.036, -0.0105, 0.0039, -0.036, -0.0105, 0.0017, -0.0338, -0.0105, 0.0039, -0.0338, -0.0168, -0.0, -0.05, -0.019, -0.0, -0.05, -0.0168, -0.0, -0.0469, -0.019, -0.0, -0.0469, -0.0212, -0.0, -0.05, -0.0233, -0.0, -0.05, -0.0212, -0.0, -0.0469, -0.0233, -0.0, -0.0469, -0.0168, 0.0017, -0.05, -0.019, 0.0017, -0.05, -0.0168, 0.0039, -0.05, -0.019, 0.0039, -0.05, -0.0212, 0.0017, -0.05, -0.0233, 0.0017, -0.05, -0.0212, 0.0039, -0.05, -0.0233, 0.0039, -0.05,
-0.0168, -0.0, -0.0447, -0.019, -0.0, -0.0447, -0.0168, -0.0, -0.0425, -0.019, -0.0, -0.0425, -0.0212, -0.0, -0.0447, -0.0233, -0.0, -0.0447, -0.0212, -0.0, -0.0425, -0.0233, -0.0, -0.0425, -0.0255, -0.0, -0.05, -0.0277, -0.0, -0.05, -0.0255, -0.0, -0.0469, -0.0277, -0.0, -0.0469, -0.0298, -0.0, -0.05, -0.032, -0.0, -0.05, -0.0298, -0.0, -0.0469, -0.032, -0.0, -0.0469, -0.0255, 0.0017, -0.05, -0.0277, 0.0017, -0.05, -0.0255, 0.0039, -0.05, -0.0277, 0.0039, -0.05, -0.0298, 0.0017, -0.05, -0.032, 0.0017, -0.05, -0.0298, 0.0039, -0.05, -0.032, 0.0039, -0.05, -0.0255, -0.0, -0.0447, -0.0277, -0.0, -0.0447, -0.0255, -0.0, -0.0425, -0.0277, -0.0, -0.0425, -0.0298, -0.0, -0.0447, -0.032, -0.0, -0.0447,
-0.0298, -0.0, -0.0425, -0.032, -0.0, -0.0425, -0.0168, -0.0, -0.0404, -0.019, -0.0, -0.0404, -0.0168, -0.0, -0.0382, -0.019, -0.0, -0.0382, -0.0212, -0.0, -0.0404, -0.0233, -0.0, -0.0404, -0.0212, -0.0, -0.0382, -0.0233, -0.0, -0.0382, -0.0168, 0.0017, -0.0382, -0.019, 0.0017, -0.0382, -0.0168, 0.0039, -0.0382, -0.019, 0.0039, -0.0382, -0.0212, 0.0017, -0.0382, -0.0233, 0.0017, -0.0382, -0.0212, 0.0039, -0.0382, -0.0233, 0.0039, -0.0382, -0.0255, -0.0, -0.0404, -0.0277, -0.0, -0.0404, -0.0255, -0.0, -0.0382, -0.0277, -0.0, -0.0382, -0.0298, -0.0, -0.0404, -0.032, -0.0, -0.0404, -0.0298, -0.0, -0.0382, -0.032, -0.0, -0.0382, -0.0255, 0.0017, -0.0382, -0.0277, 0.0017, -0.0382, -0.0255, 0.0039, -0.0382, -0.0277, 0.0039, -0.0382,
-0.0298, 0.0017, -0.0382, -0.032, 0.0017, -0.0382, -0.0298, 0.0039, -0.0382, -0.032, 0.0039, -0.0382, -0.0342, -0.0, -0.0667, -0.0363, -0.0, -0.0667, -0.0385, -0.0, -0.0667, -0.0407, -0.0, -0.0667, -0.0342, 0.0017, -0.0667, -0.0363, 0.0017, -0.0667, -0.0342, 0.0039, -0.0667, -0.0363, 0.0039, -0.0667, -0.0385, 0.0017, -0.0667, -0.0407, 0.0017, -0.0667, -0.0385, 0.0039, -0.0667, -0.0407, 0.0039, -0.0667, -0.0429, -0.0, -0.0667, -0.045, -0.0, -0.0667, -0.0472, -0.0, -0.0667, -0.0487, -0.0, -0.0667, -0.0429, 0.0017, -0.0667, -0.045, 0.0017, -0.0667, -0.0429, 0.0039, -0.0667, -0.045, 0.0039, -0.0667, -0.0472, 0.0017, -0.0667, -0.0487, 0.0017, -0.0667, -0.0472, 0.0039, -0.0667, -0.0487, 0.0039, -0.0667, -0.0342, -0.0, -0.05, -0.0363, -0.0, -0.05,
-0.0342, -0.0, -0.0469, -0.0363, -0.0, -0.0469, -0.0385, -0.0, -0.05, -0.0407, -0.0, -0.05, -0.0385, -0.0, -0.0469, -0.0407, -0.0, -0.0469, -0.0342, 0.0017, -0.05, -0.0363, 0.0017, -0.05, -0.0342, 0.0039, -0.05, -0.0363, 0.0039, -0.05, -0.0385, 0.0017, -0.05, -0.0407, 0.0017, -0.05, -0.0385, 0.0039, -0.05, -0.0407, 0.0039, -0.05, -0.0342, -0.0, -0.0447, -0.0363, -0.0, -0.0447, -0.0342, -0.0, -0.0425, -0.0363, -0.0, -0.0425, -0.0385, -0.0, -0.0447, -0.0407, -0.0, -0.0447, -0.0385, -0.0, -0.0425, -0.0407, -0.0, -0.0425, -0.0429, -0.0, -0.05, -0.045, -0.0, -0.05, -0.0429, -0.0, -0.0469, -0.045, -0.0, -0.0469, -0.0472, -0.0, -0.05, -0.0487, -0.0, -0.05, -0.0472, -0.0, -0.0469, -0.0487, -0.0, -0.0469,
-0.0429, 0.0017, -0.05, -0.045, 0.0017, -0.05, -0.0429, 0.0039, -0.05, -0.045, 0.0039, -0.05, -0.0472, 0.0017, -0.05, -0.0487, 0.0017, -0.05, -0.0472, 0.0039, -0.05, -0.0487, 0.0039, -0.05, -0.0487, 0.0017, -0.0469, -0.0487, 0.0039, -0.0469, -0.0429, -0.0, -0.0447, -0.045, -0.0, -0.0447, -0.0429, -0.0, -0.0425, -0.045, -0.0, -0.0425, -0.0472, -0.0, -0.0447, -0.0487, -0.0, -0.0447, -0.0472, -0.0, -0.0425, -0.0487, -0.0, -0.0425, -0.0487, 0.0017, -0.0447, -0.0487, 0.0039, -0.0447, -0.0487, 0.0017, -0.0425, -0.0487, 0.0039, -0.0425, -0.0342, -0.0, -0.0404, -0.0363, -0.0, -0.0404, -0.0342, -0.0, -0.0382, -0.0363, -0.0, -0.0382, -0.0385, -0.0, -0.0404, -0.0407, -0.0, -0.0404, -0.0379, -0.0, -0.0382, -0.0407, -0.0, -0.0382,
-0.0342, 0.0017, -0.0382, -0.0363, 0.0017, -0.0382, -0.0342, 0.0039, -0.0382, -0.0363, 0.0039, -0.0382, -0.0379, 0.0017, -0.0382, -0.0379, 0.0039, -0.0382, -0.0379, -0.0, -0.036, -0.0407, -0.0, -0.036, -0.0379, -0.0, -0.0338, -0.0407, -0.0, -0.0338, -0.0379, 0.0017, -0.036, -0.0379, 0.0039, -0.036, -0.0379, 0.0017, -0.0338, -0.0379, 0.0039, -0.0338, -0.0429, -0.0, -0.0404, -0.045, -0.0, -0.0404, -0.0429, -0.0, -0.0382, -0.045, -0.0, -0.0382, -0.0472, -0.0, -0.0404, -0.0487, -0.0, -0.0404, -0.0472, -0.0, -0.0382, -0.0487, -0.0, -0.0382, -0.0487, 0.0017, -0.0404, -0.0487, 0.0039, -0.0404, -0.0487, 0.0017, -0.0382, -0.0487, 0.0039, -0.0382, -0.0429, -0.0, -0.036, -0.045, -0.0, -0.036, -0.0429, -0.0, -0.0338, -0.045, -0.0, -0.0338,
-0.0472, -0.0, -0.036, -0.0487, -0.0, -0.036, -0.0472, -0.0, -0.0338, -0.0487, -0.0, -0.0338, -0.0487, 0.0017, -0.036, -0.0487, 0.0039, -0.036, -0.0487, 0.0017, -0.0338, -0.0487, 0.0039, -0.0338, 0.0013, -0.0, -0.0317, -0.0016, -0.0, -0.0317, 0.0013, -0.0, -0.0295, -0.0016, -0.0, -0.0295, -0.0038, -0.0, -0.0317, -0.006, -0.0, -0.0317, -0.0038, -0.0, -0.0295, -0.006, -0.0, -0.0295, 0.0013, 0.0017, -0.0317, 0.0013, 0.0039, -0.0317, 0.0013, 0.0017, -0.0295, 0.0013, 0.0039, -0.0295, 0.0013, -0.0, -0.0273, -0.0016, -0.0, -0.0273, 0.0013, -0.0, -0.0252, -0.0016, -0.0, -0.0252, -0.0038, -0.0, -0.0273, -0.006, -0.0, -0.0273, -0.0038, -0.0, -0.0252, -0.006, -0.0, -0.0252, 0.0013, 0.0017, -0.0273, 0.0013, 0.0039, -0.0273,
0.0013, 0.0017, -0.0252, 0.0013, 0.0039, -0.0252, -0.0081, -0.0, -0.0317, -0.0105, -0.0, -0.0317, -0.0081, -0.0, -0.0295, -0.0105, -0.0, -0.0295, -0.0105, 0.0017, -0.0317, -0.0105, 0.0039, -0.0317, -0.0105, 0.0017, -0.0295, -0.0105, 0.0039, -0.0295, -0.0081, -0.0, -0.0273, -0.0105, -0.0, -0.0273, -0.0081, -0.0, -0.0252, -0.0105, -0.0, -0.0252, -0.0105, 0.0017, -0.0273, -0.0105, 0.0039, -0.0273, -0.0105, 0.0017, -0.0252, -0.0105, 0.0039, -0.0252, 0.0013, -0.0, -0.023, -0.0016, -0.0, -0.023, 0.0013, -0.0, -0.0208, -0.0016, -0.0, -0.0208, -0.0038, -0.0, -0.023, -0.006, -0.0, -0.023, -0.0038, -0.0, -0.0208, -0.006, -0.0, -0.0208, 0.0013, 0.0017, -0.023, 0.0013, 0.0039, -0.023, 0.0013, 0.0017, -0.0208, 0.0013, 0.0039, -0.0208,
0.0013, -0.0, -0.0187, -0.0016, -0.0, -0.0187, 0.0013, -0.0, -0.0165, -0.0016, -0.0, -0.0165, -0.0038, -0.0, -0.0187, -0.006, -0.0, -0.0187, -0.0038, -0.0, -0.0165, -0.006, -0.0, -0.0165, 0.0013, 0.0017, -0.0187, 0.0013, 0.0039, -0.0187, 0.0013, 0.0017, -0.0165, 0.0013, 0.0039, -0.0165, -0.0081, -0.0, -0.023, -0.0105, -0.0, -0.023, -0.0081, -0.0, -0.0208, -0.0105, -0.0, -0.0208, -0.0105, 0.0017, -0.023, -0.0105, 0.0039, -0.023, -0.0105, 0.0017, -0.0208, -0.0105, 0.0039, -0.0208, -0.0081, -0.0, -0.0187, -0.0105, -0.0, -0.0187, -0.0081, -0.0, -0.0165, -0.0105, -0.0, -0.0165, -0.0105, 0.0017, -0.0187, -0.0105, 0.0039, -0.0187, -0.0105, 0.0017, -0.0165, -0.0105, 0.0039, -0.0165, 0.0013, -0.0, -0.0143, -0.0016, -0.0, -0.0143,
0.0013, -0.0, -0.0121, -0.0016, -0.0, -0.0121, -0.0038, -0.0, -0.0143, -0.006, -0.0, -0.0143, -0.0038, -0.0, -0.0121, -0.006, -0.0, -0.0121, 0.0013, 0.0017, -0.0143, 0.0013, 0.0039, -0.0143, 0.0013, 0.0017, -0.0121, 0.0013, 0.0039, -0.0121, 0.0013, -0.0, -0.01, -0.0016, -0.0, -0.01, 0.0013, -0.0, -0.0078, -0.0016, -0.0, -0.0078, -0.0038, -0.0, -0.01, -0.006, -0.0, -0.01, -0.0038, -0.0, -0.0078, -0.006, -0.0, -0.0078, 0.0013, 0.0017, -0.01, 0.0013, 0.0039, -0.01, 0.0013, 0.0017, -0.0078, 0.0013, 0.0039, -0.0078, -0.0081, -0.0, -0.0143, -0.0105, -0.0, -0.0143, -0.0081, -0.0, -0.0121, -0.0105, -0.0, -0.0121, -0.0105, 0.0017, -0.0143, -0.0105, 0.0039, -0.0143, -0.0105, 0.0017, -0.0121, -0.0105, 0.0039, -0.0121,
-0.0081, -0.0, -0.01, -0.0105, -0.0, -0.01, -0.0081, -0.0, -0.0078, -0.0105, -0.0, -0.0078, -0.0105, 0.0017, -0.01, -0.0105, 0.0039, -0.01, -0.0105, 0.0017, -0.0078, -0.0105, 0.0039, -0.0078, 0.0013, -0.0, -0.0056, -0.0016, -0.0, -0.0056, 0.0013, -0.0, -0.0035, -0.0016, -0.0, -0.0035, -0.0038, -0.0, -0.0056, -0.006, -0.0, -0.0056, -0.0038, -0.0, -0.0035, -0.006, -0.0, -0.0035, 0.0013, 0.0017, -0.0056, 0.0013, 0.0039, -0.0056, 0.0013, 0.0017, -0.0035, 0.0013, 0.0039, -0.0035, 0.0013, -0.0, -0.0013, -0.0016, -0.0, -0.0013, 0.0013, -0.0, 0.0009, -0.0016, -0.0, 0.0009, -0.0038, -0.0, -0.0013, -0.006, -0.0, -0.0013, -0.0038, -0.0, 0.0009, -0.006, -0.0, 0.0009, 0.0013, 0.0017, -0.0013, 0.0013, 0.0039, -0.0013,
0.0013, 0.0017, 0.0009, 0.0013, 0.0039, 0.0009, -0.0081, -0.0, -0.0056, -0.0105, -0.0, -0.0056, -0.0081, -0.0, -0.0035, -0.0105, -0.0, -0.0035, -0.0105, 0.0017, -0.0056, -0.0105, 0.0039, -0.0056, -0.0105, 0.0017, -0.0035, -0.0105, 0.0039, -0.0035, -0.0081, -0.0, -0.0013, -0.0105, -0.0, -0.0013, -0.0081, -0.0, 0.0009, -0.0105, -0.0, 0.0009, -0.0105, 0.0017, -0.0013, -0.0105, 0.0039, -0.0013, -0.0105, 0.0017, 0.0009, -0.0105, 0.0039, 0.0009, -0.0379, -0.0, -0.0317, -0.0407, -0.0, -0.0317, -0.0379, -0.0, -0.0295, -0.0407, -0.0, -0.0295, -0.0379, 0.0017, -0.0317, -0.0379, 0.0039, -0.0317, -0.0379, 0.0017, -0.0295, -0.0379, 0.0039, -0.0295, -0.0379, -0.0, -0.0273, -0.0407, -0.0, -0.0273, -0.0379, -0.0, -0.0252, -0.0407, -0.0, -0.0252,
-0.0379, 0.0017, -0.0273, -0.0379, 0.0039, -0.0273, -0.0379, 0.0017, -0.0252, -0.0379, 0.0039, -0.0252, -0.0429, -0.0, -0.0317, -0.045, -0.0, -0.0317, -0.0429, -0.0, -0.0295, -0.045, -0.0, -0.0295, -0.0472, -0.0, -0.0317, -0.0487, -0.0, -0.0317, -0.0472, -0.0, -0.0295, -0.0487, -0.0, -0.0295, -0.0487, 0.0017, -0.0317, -0.0487, 0.0039, -0.0317, -0.0487, 0.0017, -0.0295, -0.0487, 0.0039, -0.0295, -0.0429, -0.0, -0.0273, -0.045, -0.0, -0.0273, -0.0429, -0.0, -0.0252, -0.045, -0.0, -0.0252, -0.0472, -0.0, -0.0273, -0.0487, -0.0, -0.0273, -0.0472, -0.0, -0.0252, -0.0487, -0.0, -0.0252, -0.0487, 0.0017, -0.0273, -0.0487, 0.0039, -0.0273, -0.0487, 0.0017, -0.0252, -0.0487, 0.0039, -0.0252, -0.0379, -0.0, -0.023, -0.0407, -0.0, -0.023,
-0.0379, -0.0, -0.0208, -0.0407, -0.0, -0.0208, -0.0379, 0.0017, -0.023, -0.0379, 0.0039, -0.023, -0.0379, 0.0017, -0.0208, -0.0379, 0.0039, -0.0208, -0.0379, -0.0, -0.0187, -0.0407, -0.0, -0.0187, -0.0379, -0.0, -0.0165, -0.0407, -0.0, -0.0165, -0.0379, 0.0017, -0.0187, -0.0379, 0.0039, -0.0187, -0.0379, 0.0017, -0.0165, -0.0379, 0.0039, -0.0165, -0.0429, -0.0, -0.023, -0.045, -0.0, -0.023, -0.0429, -0.0, -0.0208, -0.045, -0.0, -0.0208, -0.0472, -0.0, -0.023, -0.0487, -0.0, -0.023, -0.0472, -0.0, -0.0208, -0.0487, -0.0, -0.0208, -0.0487, 0.0017, -0.023, -0.0487, 0.0039, -0.023, -0.0487, 0.0017, -0.0208, -0.0487, 0.0039, -0.0208, -0.0429, -0.0, -0.0187, -0.045, -0.0, -0.0187, -0.0429, -0.0, -0.0165, -0.045, -0.0, -0.0165,
-0.0472, -0.0, -0.0187, -0.0487, -0.0, -0.0187, -0.0472, -0.0, -0.0165, -0.0487, -0.0, -0.0165, -0.0487, 0.0017, -0.0187, -0.0487, 0.0039, -0.0187, -0.0487, 0.0017, -0.0165, -0.0487, 0.0039, -0.0165, -0.0379, -0.0, -0.0143, -0.0407, -0.0, -0.0143, -0.0379, -0.0, -0.0121, -0.0407, -0.0, -0.0121, -0.0379, 0.0017, -0.0143, -0.0379, 0.0039, -0.0143, -0.0379, 0.0017, -0.0121, -0.0379, 0.0039, -0.0121, -0.0379, -0.0, -0.01, -0.0407, -0.0, -0.01, -0.0379, -0.0, -0.0078, -0.0407, -0.0, -0.0078, -0.0379, 0.0017, -0.01, -0.0379, 0.0039, -0.01, -0.0379, 0.0017, -0.0078, -0.0379, 0.0039, -0.0078, -0.0429, -0.0, -0.0143, -0.045, -0.0, -0.0143, -0.0429, -0.0, -0.0121, -0.045, -0.0, -0.0121, -0.0472, -0.0, -0.0143, -0.0487, -0.0, -0.0143,
-0.0472, -0.0, -0.0121, -0.0487, -0.0, -0.0121, -0.0487, 0.0017, -0.0143, -0.0487, 0.0039, -0.0143, -0.0487, 0.0017, -0.0121, -0.0487, 0.0039, -0.0121, -0.0429, -0.0, -0.01, -0.045, -0.0, -0.01, -0.0429, -0.0, -0.0078, -0.045, -0.0, -0.0078, -0.0472, -0.0, -0.01, -0.0487, -0.0, -0.01, -0.0472, -0.0, -0.0078, -0.0487, -0.0, -0.0078, -0.0487, 0.0017, -0.01, -0.0487, 0.0039, -0.01, -0.0487, 0.0017, -0.0078, -0.0487, 0.0039, -0.0078, -0.0379, -0.0, -0.0056, -0.0407, -0.0, -0.0056, -0.0379, -0.0, -0.0035, -0.0407, -0.0, -0.0035, -0.0379, 0.0017, -0.0056, -0.0379, 0.0039, -0.0056, -0.0379, 0.0017, -0.0035, -0.0379, 0.0039, -0.0035, -0.0379, -0.0, -0.0013, -0.0407, -0.0, -0.0013, -0.0379, -0.0, 0.0008, -0.0407, -0.0, 0.0009,
-0.0379, 0.0017, -0.0013, -0.0379, 0.0039, -0.0013, -0.0379, 0.0017, 0.0009, -0.0379, 0.0039, 0.0009, -0.0429, -0.0, -0.0056, -0.045, -0.0, -0.0056, -0.0429, -0.0, -0.0035, -0.045, -0.0, -0.0035, -0.0472, -0.0, -0.0056, -0.0487, -0.0, -0.0056, -0.0472, -0.0, -0.0035, -0.0487, -0.0, -0.0035, -0.0487, 0.0017, -0.0056, -0.0487, 0.0039, -0.0056, -0.0487, 0.0017, -0.0035, -0.0487, 0.0039, -0.0035, -0.0429, -0.0, -0.0013, -0.045, -0.0, -0.0013, -0.0429, -0.0, 0.0009, -0.045, -0.0, 0.0009, -0.0472, -0.0, -0.0013, -0.0487, -0.0, -0.0013, -0.0472, -0.0, 0.0009, -0.0487, -0.0, 0.0009, -0.0487, 0.0017, -0.0013, -0.0487, 0.0039, -0.0013, -0.0487, 0.0017, 0.0009, -0.0487, 0.0039, 0.0009, 0.0364, 0.0061, -0.4928, 0.0364, 0.0083, -0.4928,
0.0364, 0.01, -0.4928, 0.0457, 0.0061, -0.4868, 0.0457, 0.0083, -0.4868, 0.0482, 0.0061, -0.4849, 0.0468, 0.0061, -0.486, 0.0482, 0.0083, -0.4849, 0.0468, 0.0083, -0.486, 0.0456, 0.01, -0.4868, 0.0481, 0.01, -0.485, 0.0467, 0.01, -0.486, 0.043, 0.0061, -0.4886, 0.0418, 0.0061, -0.4895, 0.043, 0.0083, -0.4886, 0.0418, 0.0083, -0.4895, 0.0393, 0.0061, -0.4911, 0.0375, 0.0061, -0.4922, 0.0393, 0.0083, -0.4911, 0.0375, 0.0083, -0.4922, 0.0403, 0.0061, -0.4904, 0.0403, 0.0083, -0.4904, 0.043, 0.01, -0.4886, 0.042, 0.01, -0.4893, 0.0391, 0.01, -0.4912, 0.0375, 0.01, -0.4922, 0.0402, 0.01, -0.4905, 0.0374, 0.01, -0.4896, 0.0441, 0.0061, -0.4879, 0.0441, 0.0083, -0.4879, 0.0368, 0.0061, -0.4868, 0.0368, 0.0083, -0.4868,
0.0387, 0.0061, -0.4844, 0.0379, 0.0061, -0.4855, 0.0387, 0.0083, -0.4844, 0.0379, 0.0083, -0.4855, 0.0441, 0.01, -0.4879, 0.0418, 0.01, -0.4874, 0.0439, 0.01, -0.4852, 0.0418, 0.01, -0.4852, 0.0396, 0.01, -0.4874, 0.0368, 0.01, -0.4869, 0.0386, 0.01, -0.4845, 0.0378, 0.01, -0.4856, 0.0321, 0.0061, -0.4951, 0.0321, 0.0083, -0.4951, 0.0353, 0.0061, -0.4935, 0.0332, 0.0061, -0.4945, 0.0353, 0.0083, -0.4935, 0.0332, 0.0083, -0.4945, 0.0297, 0.0061, -0.4961, 0.0297, 0.0083, -0.4961, 0.0306, 0.0061, -0.4938, 0.0306, 0.0083, -0.4938, 0.0321, 0.01, -0.4951, 0.0354, 0.01, -0.4934, 0.0332, 0.01, -0.4945, 0.0297, 0.01, -0.4961, 0.0308, 0.01, -0.4935, 0.0241, 0.0061, -0.4934, 0.0241, 0.0083, -0.4934, 0.0215, 0.0061, -0.4952,
0.0201, 0.0061, -0.4961, 0.0215, 0.0083, -0.4952, 0.0201, 0.0083, -0.4961, 0.0226, 0.0061, -0.4945, 0.0195, 0.0061, -0.4935, 0.0226, 0.0083, -0.4945, 0.0195, 0.0083, -0.4935, 0.024, 0.01, -0.4934, 0.0215, 0.01, -0.4953, 0.0203, 0.01, -0.496, 0.0226, 0.01, -0.4945, 0.0195, 0.01, -0.4935, 0.0326, 0.0061, -0.4914, 0.0326, 0.0083, -0.4914, 0.0347, 0.0061, -0.4891, 0.0337, 0.0061, -0.4902, 0.0347, 0.0083, -0.4891, 0.0337, 0.0083, -0.4902, 0.0317, 0.0061, -0.4925, 0.0317, 0.0083, -0.4925, 0.0286, 0.0061, -0.4894, 0.0286, 0.0083, -0.4894, 0.0353, 0.01, -0.4917, 0.0326, 0.01, -0.4914, 0.0347, 0.01, -0.4891, 0.0337, 0.01, -0.4902, 0.0317, 0.01, -0.4925, 0.0286, 0.01, -0.4894, 0.0358, 0.0061, -0.4879, 0.0358, 0.0083, -0.4879,
0.0326, 0.0061, -0.4849, 0.0326, 0.0083, -0.4849, 0.0307, 0.0061, -0.4872, 0.0296, 0.0061, -0.4883, 0.0307, 0.0083, -0.4872, 0.0296, 0.0083, -0.4883, 0.0317, 0.0061, -0.486, 0.0317, 0.0083, -0.486, 0.0358, 0.01, -0.4879, 0.0327, 0.01, -0.4849, 0.0307, 0.01, -0.4872, 0.0297, 0.01, -0.4883, 0.0318, 0.01, -0.486, 0.0288, 0.01, -0.4852, 0.0264, 0.0061, -0.4915, 0.0252, 0.0061, -0.4925, 0.0264, 0.0083, -0.4915, 0.0252, 0.0083, -0.4925, 0.0274, 0.0061, -0.4905, 0.0274, 0.0083, -0.4905, 0.0215, 0.0061, -0.4911, 0.0206, 0.0061, -0.4922, 0.0215, 0.0083, -0.4911, 0.0206, 0.0083, -0.4922, 0.0225, 0.0061, -0.4896, 0.0225, 0.0083, -0.4896, 0.0263, 0.01, -0.4915, 0.0251, 0.01, -0.4926, 0.0274, 0.01, -0.4905, 0.0244, 0.01, -0.4896,
0.0214, 0.01, -0.4911, 0.0206, 0.01, -0.4922, 0.0227, 0.01, -0.4894, 0.0239, 0.0061, -0.4873, 0.0239, 0.0083, -0.4873, 0.025, 0.0061, -0.4853, 0.025, 0.0083, -0.4853, 0.0233, 0.0061, -0.4884, 0.0233, 0.0083, -0.4884, 0.0266, 0.01, -0.4874, 0.0239, 0.01, -0.4873, 0.0266, 0.01, -0.4852, 0.025, 0.01, -0.4851, 0.0233, 0.01, -0.4884, 0.0175, 0.0061, -0.4975, 0.016, 0.0061, -0.498, 0.0175, 0.0083, -0.4975, 0.016, 0.0083, -0.498, 0.0173, 0.01, -0.4975, 0.016, 0.01, -0.498, 0.0157, 0.0061, -0.4983, 0.0167, 0.0061, -0.497, 0.0157, 0.0083, -0.4983, 0.0167, 0.0083, -0.497, 0.0186, 0.0061, -0.4945, 0.0186, 0.0083, -0.4945, 0.0157, 0.01, -0.4983, 0.0167, 0.01, -0.497, 0.0186, 0.01, -0.4946, 0.0713, 0.0061, -0.4561,
0.0713, 0.0083, -0.4561, 0.0736, 0.0061, -0.4541, 0.0724, 0.0061, -0.4552, 0.0736, 0.0083, -0.4541, 0.0724, 0.0083, -0.4552, 0.0713, 0.01, -0.4561, 0.0736, 0.01, -0.4541, 0.0724, 0.01, -0.4552, 0.0759, 0.0061, -0.452, 0.0759, 0.0083, -0.452, 0.0781, 0.0061, -0.4499, 0.077, 0.0061, -0.451, 0.0781, 0.0083, -0.4499, 0.077, 0.0083, -0.451, 0.0747, 0.0061, -0.4531, 0.0747, 0.0083, -0.4531, 0.0759, 0.01, -0.452, 0.0781, 0.01, -0.4499, 0.077, 0.01, -0.451, 0.0747, 0.01, -0.4531, 0.0722, 0.01, -0.4527, 0.0743, 0.01, -0.4505, 0.0722, 0.01, -0.4505, 0.089, 0.0061, -0.4372, 0.089, 0.0083, -0.4372, 0.0891, 0.01, -0.4371, 0.0908, 0.0061, -0.4345, 0.0901, 0.0061, -0.4356, 0.0908, 0.0083, -0.4345, 0.0901, 0.0083, -0.4356,
0.0917, 0.0061, -0.4332, 0.0917, 0.0083, -0.4332, 0.0908, 0.01, -0.4345, 0.0901, 0.01, -0.4356, 0.0916, 0.01, -0.4334, 0.0895, 0.01, -0.4332, 0.0802, 0.0061, -0.4478, 0.0802, 0.0083, -0.4478, 0.0823, 0.0061, -0.4455, 0.0813, 0.0061, -0.4466, 0.0823, 0.0083, -0.4455, 0.0813, 0.0083, -0.4466, 0.0802, 0.01, -0.4478, 0.0823, 0.01, -0.4455, 0.0813, 0.01, -0.4466, 0.0844, 0.0061, -0.4432, 0.0844, 0.0083, -0.4432, 0.0863, 0.0061, -0.4408, 0.0855, 0.0061, -0.4419, 0.0863, 0.0083, -0.4408, 0.0855, 0.0083, -0.4419, 0.0834, 0.0061, -0.4443, 0.0834, 0.0083, -0.4443, 0.0844, 0.01, -0.4432, 0.0863, 0.01, -0.4408, 0.0855, 0.01, -0.442, 0.0834, 0.01, -0.4443, 0.0808, 0.01, -0.444, 0.083, 0.01, -0.4418, 0.0808, 0.01, -0.4418,
0.0792, 0.0061, -0.4488, 0.0792, 0.0083, -0.4488, 0.0792, 0.01, -0.4488, 0.0765, 0.01, -0.4483, 0.0787, 0.01, -0.4462, 0.0765, 0.01, -0.4462, 0.0743, 0.01, -0.4483, 0.0722, 0.01, -0.4483, 0.0743, 0.01, -0.4462, 0.0722, 0.01, -0.4462, 0.0787, 0.01, -0.444, 0.0765, 0.01, -0.444, 0.0787, 0.01, -0.4418, 0.0765, 0.01, -0.4418, 0.0743, 0.01, -0.444, 0.0722, 0.01, -0.444, 0.0743, 0.01, -0.4418, 0.0722, 0.01, -0.4418, 0.0872, 0.0061, -0.4397, 0.0872, 0.0083, -0.4397, 0.0882, 0.0061, -0.4383, 0.0882, 0.0083, -0.4383, 0.087, 0.01, -0.4399, 0.0852, 0.01, -0.4397, 0.0883, 0.01, -0.4382, 0.0852, 0.01, -0.4375, 0.083, 0.01, -0.4397, 0.0808, 0.01, -0.4397, 0.083, 0.01, -0.4375, 0.0808, 0.01, -0.4375,
0.0874, 0.01, -0.4353, 0.0852, 0.01, -0.4353, 0.0874, 0.01, -0.4332, 0.0852, 0.01, -0.4332, 0.083, 0.01, -0.4353, 0.0808, 0.01, -0.4353, 0.083, 0.01, -0.4332, 0.0808, 0.01, -0.4332, 0.0787, 0.01, -0.4397, 0.0765, 0.01, -0.4397, 0.0787, 0.01, -0.4375, 0.0765, 0.01, -0.4375, 0.0743, 0.01, -0.4397, 0.0722, 0.01, -0.4397, 0.0743, 0.01, -0.4375, 0.0722, 0.01, -0.4375, 0.0787, 0.01, -0.4353, 0.0765, 0.01, -0.4353, 0.0787, 0.01, -0.4332, 0.0765, 0.01, -0.4332, 0.0743, 0.01, -0.4353, 0.0722, 0.01, -0.4353, 0.0743, 0.01, -0.4332, 0.0722, 0.01, -0.4332, 0.0972, 0.0061, -0.4235, 0.0972, 0.0083, -0.4235, 0.0972, 0.01, -0.4234, 0.0933, 0.0061, -0.4307, 0.0933, 0.0083, -0.4307, 0.0949, 0.0061, -0.4277,
0.0943, 0.0061, -0.4288, 0.0949, 0.0083, -0.4277, 0.0943, 0.0083, -0.4288, 0.0926, 0.0061, -0.4318, 0.0926, 0.0083, -0.4318, 0.0933, 0.01, -0.4306, 0.0949, 0.01, -0.4277, 0.0943, 0.01, -0.4289, 0.0927, 0.01, -0.4316, 0.0895, 0.01, -0.431, 0.0917, 0.01, -0.4288, 0.0895, 0.01, -0.4288, 0.0955, 0.0061, -0.4266, 0.0955, 0.0083, -0.4266, 0.0966, 0.0061, -0.4245, 0.0966, 0.0083, -0.4245, 0.0955, 0.01, -0.4268, 0.0939, 0.01, -0.4266, 0.0966, 0.01, -0.4246, 0.0939, 0.01, -0.4245, 0.0917, 0.01, -0.4266, 0.0895, 0.01, -0.4266, 0.0917, 0.01, -0.4245, 0.0895, 0.01, -0.4245, 0.0977, 0.0061, -0.4223, 0.0977, 0.0083, -0.4223, 0.0988, 0.0061, -0.4201, 0.0988, 0.0083, -0.4201, 0.0977, 0.01, -0.4224, 0.0986, 0.01, -0.4205,
0.0997, 0.0061, -0.4179, 0.0993, 0.0061, -0.419, 0.0997, 0.0083, -0.4179, 0.0993, 0.0083, -0.419, 0.1006, 0.0061, -0.4158, 0.1006, 0.0083, -0.4158, 0.0998, 0.01, -0.4177, 0.0995, 0.01, -0.4186, 0.1005, 0.01, -0.416, 0.0982, 0.01, -0.4158, 0.096, 0.01, -0.4223, 0.0939, 0.01, -0.4223, 0.096, 0.01, -0.4201, 0.0939, 0.01, -0.4201, 0.0917, 0.01, -0.4223, 0.0895, 0.01, -0.4223, 0.0917, 0.01, -0.4201, 0.0895, 0.01, -0.4201, 0.096, 0.01, -0.418, 0.0939, 0.01, -0.418, 0.096, 0.01, -0.4158, 0.0939, 0.01, -0.4158, 0.0917, 0.01, -0.418, 0.0895, 0.01, -0.418, 0.0917, 0.01, -0.4158, 0.0895, 0.01, -0.4158, 0.0874, 0.01, -0.431, 0.0852, 0.01, -0.431, 0.0874, 0.01, -0.4288, 0.0852, 0.01, -0.4288,
0.083, 0.01, -0.431, 0.0808, 0.01, -0.431, 0.083, 0.01, -0.4288, 0.0808, 0.01, -0.4288, 0.0874, 0.01, -0.4266, 0.0852, 0.01, -0.4266, 0.0874, 0.01, -0.4245, 0.0852, 0.01, -0.4245, 0.083, 0.01, -0.4266, 0.0808, 0.01, -0.4266, 0.083, 0.01, -0.4245, 0.0808, 0.01, -0.4245, 0.072, 0.0061, -0.4307, 0.072, 0.0083, -0.4307, 0.0741, 0.0061, -0.4287, 0.0731, 0.0061, -0.4297, 0.0741, 0.0083, -0.4287, 0.0731, 0.0083, -0.4297, 0.0787, 0.01, -0.431, 0.0765, 0.01, -0.431, 0.0787, 0.01, -0.4288, 0.0765, 0.01, -0.4288, 0.0743, 0.01, -0.431, 0.072, 0.01, -0.4307, 0.0741, 0.01, -0.4287, 0.0731, 0.01, -0.4297, 0.076, 0.0061, -0.4263, 0.076, 0.0083, -0.4263, 0.0779, 0.0061, -0.4238, 0.0771, 0.0061, -0.4249,
0.0779, 0.0083, -0.4238, 0.0771, 0.0083, -0.4249, 0.0752, 0.0061, -0.4274, 0.0752, 0.0083, -0.4274, 0.0787, 0.01, -0.4266, 0.076, 0.01, -0.4263, 0.0778, 0.01, -0.4239, 0.077, 0.01, -0.4249, 0.0751, 0.01, -0.4274, 0.0804, 0.0061, -0.42, 0.0804, 0.0083, -0.42, 0.0874, 0.01, -0.4223, 0.0852, 0.01, -0.4223, 0.0874, 0.01, -0.4201, 0.0852, 0.01, -0.4201, 0.083, 0.01, -0.4223, 0.0808, 0.01, -0.4223, 0.083, 0.01, -0.4201, 0.0804, 0.01, -0.42, 0.0821, 0.0061, -0.4172, 0.0815, 0.0061, -0.4183, 0.0821, 0.0083, -0.4172, 0.0815, 0.0083, -0.4183, 0.0828, 0.0061, -0.4158, 0.0828, 0.0083, -0.4158, 0.0874, 0.01, -0.418, 0.0852, 0.01, -0.418, 0.0874, 0.01, -0.4158, 0.0852, 0.01, -0.4158, 0.082, 0.01, -0.4174,
0.0814, 0.01, -0.4184, 0.083, 0.01, -0.4155, 0.0789, 0.0061, -0.4223, 0.0789, 0.0083, -0.4223, 0.0797, 0.0061, -0.4211, 0.0797, 0.0083, -0.4211, 0.079, 0.01, -0.4221, 0.0797, 0.01, -0.4211, 0.0538, 0.0061, -0.4799, 0.0538, 0.0083, -0.4799, 0.0538, 0.01, -0.4799, 0.0559, 0.0061, -0.4777, 0.0549, 0.0061, -0.4788, 0.0559, 0.0083, -0.4777, 0.0549, 0.0083, -0.4788, 0.0569, 0.0061, -0.4766, 0.0569, 0.0083, -0.4766, 0.0559, 0.01, -0.4777, 0.0549, 0.01, -0.4788, 0.0567, 0.01, -0.4768, 0.0548, 0.01, -0.4766, 0.0624, 0.0061, -0.469, 0.0624, 0.0083, -0.469, 0.0631, 0.0061, -0.4679, 0.0631, 0.0083, -0.4679, 0.0624, 0.01, -0.469, 0.063, 0.01, -0.468, 0.0588, 0.0061, -0.4743, 0.0588, 0.0083, -0.4743, 0.0606, 0.0061, -0.4717,
0.0599, 0.0061, -0.4728, 0.0606, 0.0083, -0.4717, 0.0599, 0.0083, -0.4728, 0.058, 0.0061, -0.4753, 0.058, 0.0083, -0.4753, 0.0589, 0.01, -0.4742, 0.0607, 0.01, -0.4717, 0.0599, 0.01, -0.4728, 0.058, 0.01, -0.4753, 0.0548, 0.01, -0.4744, 0.057, 0.01, -0.4722, 0.0548, 0.01, -0.4722, 0.0617, 0.0061, -0.4701, 0.0617, 0.0083, -0.4701, 0.0617, 0.01, -0.4701, 0.0591, 0.01, -0.47, 0.0613, 0.01, -0.4679, 0.0591, 0.01, -0.4679, 0.057, 0.01, -0.47, 0.0548, 0.01, -0.47, 0.057, 0.01, -0.4679, 0.0548, 0.01, -0.4679, 0.0516, 0.0061, -0.482, 0.0505, 0.0061, -0.483, 0.0516, 0.0083, -0.482, 0.0505, 0.0083, -0.483, 0.0527, 0.0061, -0.481, 0.0527, 0.0083, -0.481, 0.0493, 0.0061, -0.4841, 0.0493, 0.0083, -0.4841,
0.0516, 0.01, -0.482, 0.0507, 0.01, -0.4828, 0.0527, 0.01, -0.481, 0.0505, 0.01, -0.4809, 0.0492, 0.01, -0.4841, 0.0461, 0.01, -0.4831, 0.0483, 0.01, -0.4809, 0.0461, 0.01, -0.4809, 0.0526, 0.01, -0.4787, 0.0505, 0.01, -0.4787, 0.0526, 0.01, -0.4766, 0.0505, 0.01, -0.4766, 0.0483, 0.01, -0.4787, 0.0461, 0.01, -0.4787, 0.0483, 0.01, -0.4766, 0.0461, 0.01, -0.4766, 0.0413, 0.0061, -0.4807, 0.0413, 0.0083, -0.4807, 0.0397, 0.0061, -0.4831, 0.0397, 0.0083, -0.4831, 0.0406, 0.0061, -0.4818, 0.0406, 0.0083, -0.4818, 0.0439, 0.01, -0.4831, 0.0418, 0.01, -0.4831, 0.0439, 0.01, -0.4809, 0.0413, 0.01, -0.4807, 0.0398, 0.01, -0.4828, 0.0405, 0.01, -0.4818, 0.0431, 0.0061, -0.4779, 0.0424, 0.0061, -0.479,
0.0431, 0.0083, -0.4779, 0.0424, 0.0083, -0.479, 0.0439, 0.0061, -0.4766, 0.0439, 0.0083, -0.4766, 0.0368, 0.0061, -0.4785, 0.0368, 0.0083, -0.4785, 0.0377, 0.0061, -0.4766, 0.0377, 0.0083, -0.4766, 0.043, 0.01, -0.4781, 0.0423, 0.01, -0.4791, 0.044, 0.01, -0.4763, 0.0369, 0.01, -0.4783, 0.0376, 0.01, -0.4768, 0.0453, 0.0061, -0.4738, 0.0453, 0.0083, -0.4738, 0.0459, 0.0061, -0.4722, 0.0459, 0.0083, -0.4722, 0.0526, 0.01, -0.4744, 0.0505, 0.01, -0.4744, 0.0526, 0.01, -0.4722, 0.0505, 0.01, -0.4722, 0.0483, 0.01, -0.4744, 0.0452, 0.01, -0.474, 0.0483, 0.01, -0.4722, 0.046, 0.01, -0.472, 0.0468, 0.0061, -0.4701, 0.0468, 0.0083, -0.4701, 0.0476, 0.0061, -0.4678, 0.0472, 0.0061, -0.4689, 0.0476, 0.0083, -0.4678,
0.0472, 0.0083, -0.4689, 0.0526, 0.01, -0.47, 0.0505, 0.01, -0.47, 0.0526, 0.01, -0.4679, 0.0505, 0.01, -0.4679, 0.0483, 0.01, -0.47, 0.0468, 0.01, -0.47, 0.0476, 0.01, -0.4679, 0.0472, 0.01, -0.4689, 0.0448, 0.0061, -0.4749, 0.0448, 0.0083, -0.4749, 0.0387, 0.0061, -0.4739, 0.0383, 0.0061, -0.475, 0.0387, 0.0083, -0.4739, 0.0383, 0.0083, -0.475, 0.0393, 0.0061, -0.4722, 0.0393, 0.0083, -0.4722, 0.0447, 0.01, -0.475, 0.0388, 0.01, -0.4737, 0.0384, 0.01, -0.4747, 0.0393, 0.01, -0.4723, 0.0374, 0.01, -0.4722, 0.0408, 0.0061, -0.4671, 0.0408, 0.0083, -0.4671, 0.04, 0.0061, -0.47, 0.04, 0.0083, -0.47, 0.0405, 0.0061, -0.4682, 0.0405, 0.0083, -0.4682, 0.0408, 0.01, -0.467, 0.0399, 0.01, -0.4702,
0.0374, 0.01, -0.47, 0.0405, 0.01, -0.468, 0.0374, 0.01, -0.4679, 0.0647, 0.0061, -0.4649, 0.0641, 0.0061, -0.466, 0.0647, 0.0083, -0.4649, 0.0641, 0.0083, -0.466, 0.0654, 0.0061, -0.4635, 0.0654, 0.0083, -0.4635, 0.0647, 0.01, -0.4648, 0.0641, 0.01, -0.4659, 0.0653, 0.01, -0.4637, 0.0635, 0.01, -0.4635, 0.067, 0.0061, -0.4607, 0.067, 0.0083, -0.4607, 0.069, 0.0061, -0.4582, 0.068, 0.0061, -0.4593, 0.069, 0.0083, -0.4582, 0.068, 0.0083, -0.4593, 0.0663, 0.0061, -0.4618, 0.0663, 0.0083, -0.4618, 0.067, 0.01, -0.4606, 0.069, 0.01, -0.4582, 0.068, 0.01, -0.4593, 0.0664, 0.01, -0.4617, 0.0635, 0.01, -0.4614, 0.0656, 0.01, -0.4592, 0.0635, 0.01, -0.4592, 0.0613, 0.01, -0.4657, 0.0591, 0.01, -0.4657,
0.0613, 0.01, -0.4635, 0.0591, 0.01, -0.4635, 0.057, 0.01, -0.4657, 0.0548, 0.01, -0.4657, 0.057, 0.01, -0.4635, 0.0548, 0.01, -0.4635, 0.0613, 0.01, -0.4614, 0.0591, 0.01, -0.4614, 0.0613, 0.01, -0.4592, 0.0591, 0.01, -0.4592, 0.057, 0.01, -0.4614, 0.0548, 0.01, -0.4614, 0.057, 0.01, -0.4592, 0.0548, 0.01, -0.4592, 0.0701, 0.0061, -0.4572, 0.0701, 0.0083, -0.4572, 0.0701, 0.01, -0.4572, 0.0678, 0.01, -0.457, 0.07, 0.01, -0.4549, 0.0678, 0.01, -0.4549, 0.0656, 0.01, -0.457, 0.0635, 0.01, -0.457, 0.0656, 0.01, -0.4549, 0.0635, 0.01, -0.4549, 0.07, 0.01, -0.4527, 0.0678, 0.01, -0.4527, 0.07, 0.01, -0.4505, 0.0678, 0.01, -0.4505, 0.0656, 0.01, -0.4527, 0.0635, 0.01, -0.4527,
0.0656, 0.01, -0.4505, 0.0635, 0.01, -0.4505, 0.0613, 0.01, -0.457, 0.0591, 0.01, -0.457, 0.0613, 0.01, -0.4549, 0.0591, 0.01, -0.4549, 0.057, 0.01, -0.457, 0.0548, 0.01, -0.457, 0.057, 0.01, -0.4549, 0.0548, 0.01, -0.4549, 0.0613, 0.01, -0.4527, 0.0591, 0.01, -0.4527, 0.0613, 0.01, -0.4505, 0.0591, 0.01, -0.4505, 0.057, 0.01, -0.4527, 0.0548, 0.01, -0.4527, 0.057, 0.01, -0.4505, 0.0548, 0.01, -0.4505, 0.0482, 0.0061, -0.4657, 0.0482, 0.0083, -0.4657, 0.0487, 0.0061, -0.4635, 0.0487, 0.0083, -0.4635, 0.0526, 0.01, -0.4657, 0.0505, 0.01, -0.4657, 0.0526, 0.01, -0.4635, 0.0505, 0.01, -0.4635, 0.0483, 0.01, -0.4656, 0.0488, 0.01, -0.4635, 0.0494, 0.0061, -0.4604, 0.0494, 0.0083, -0.4604,
0.0497, 0.0061, -0.4592, 0.0497, 0.0083, -0.4592, 0.0492, 0.0061, -0.4615, 0.0492, 0.0083, -0.4615, 0.0526, 0.01, -0.4614, 0.0493, 0.01, -0.461, 0.0526, 0.01, -0.4592, 0.0497, 0.01, -0.459, 0.0491, 0.01, -0.4618, 0.0411, 0.0061, -0.4657, 0.0411, 0.0083, -0.4657, 0.0416, 0.0061, -0.4635, 0.0416, 0.0083, -0.4635, 0.0411, 0.01, -0.4658, 0.0416, 0.01, -0.4636, 0.0396, 0.01, -0.4657, 0.0374, 0.01, -0.4657, 0.0396, 0.01, -0.4635, 0.0374, 0.01, -0.4635, 0.0421, 0.0061, -0.4614, 0.0421, 0.0083, -0.4614, 0.0424, 0.0061, -0.4592, 0.0424, 0.0083, -0.4592, 0.0421, 0.01, -0.4615, 0.0424, 0.01, -0.4593, 0.0396, 0.01, -0.4614, 0.0374, 0.01, -0.4614, 0.0396, 0.01, -0.4592, 0.0374, 0.01, -0.4592, 0.05, 0.0061, -0.457,
0.05, 0.0083, -0.457, 0.0502, 0.0061, -0.4549, 0.0502, 0.0083, -0.4549, 0.0526, 0.01, -0.457, 0.05, 0.01, -0.4569, 0.0526, 0.01, -0.4549, 0.0502, 0.01, -0.4548, 0.0503, 0.0061, -0.4527, 0.0503, 0.0083, -0.4527, 0.0503, 0.0061, -0.4505, 0.0503, 0.0083, -0.4505, 0.0526, 0.01, -0.4527, 0.0503, 0.01, -0.4527, 0.0526, 0.01, -0.4505, 0.0503, 0.01, -0.4505, 0.0425, 0.0061, -0.457, 0.0425, 0.0083, -0.457, 0.0425, 0.0061, -0.4549, 0.0425, 0.0083, -0.4549, 0.0425, 0.01, -0.457, 0.0425, 0.01, -0.4549, 0.0396, 0.01, -0.457, 0.0374, 0.01, -0.457, 0.0396, 0.01, -0.4549, 0.0374, 0.01, -0.4549, 0.0425, 0.0061, -0.4527, 0.0425, 0.0083, -0.4527, 0.0425, 0.0061, -0.4505, 0.0425, 0.0083, -0.4505, 0.0425, 0.01, -0.4527,
0.0425, 0.01, -0.4505, 0.0396, 0.01, -0.4527, 0.0374, 0.01, -0.4527, 0.0396, 0.01, -0.4505, 0.0374, 0.01, -0.4505, 0.0345, 0.0061, -0.4824, 0.0337, 0.0061, -0.4835, 0.0345, 0.0083, -0.4824, 0.0337, 0.0083, -0.4835, 0.0355, 0.0061, -0.4809, 0.0355, 0.0083, -0.4809, 0.0345, 0.01, -0.4824, 0.0338, 0.01, -0.4835, 0.0353, 0.01, -0.4812, 0.0331, 0.01, -0.4809, 0.0309, 0.01, -0.4831, 0.0288, 0.01, -0.4831, 0.0309, 0.01, -0.4809, 0.0288, 0.01, -0.4809, 0.0362, 0.0061, -0.4795, 0.0362, 0.0083, -0.4795, 0.0279, 0.0061, -0.4764, 0.0279, 0.0083, -0.4764, 0.0363, 0.01, -0.4793, 0.0331, 0.01, -0.4787, 0.0353, 0.01, -0.4766, 0.0331, 0.01, -0.4766, 0.0309, 0.01, -0.4787, 0.0288, 0.01, -0.4787, 0.0309, 0.01, -0.4766,
0.0279, 0.01, -0.4765, 0.0259, 0.0061, -0.483, 0.0255, 0.0061, -0.4841, 0.0259, 0.0083, -0.483, 0.0255, 0.0083, -0.4841, 0.0266, 0.0061, -0.4809, 0.0266, 0.0083, -0.4809, 0.0259, 0.01, -0.483, 0.0255, 0.01, -0.4841, 0.0267, 0.01, -0.4807, 0.0273, 0.0061, -0.4787, 0.0273, 0.0083, -0.4787, 0.0276, 0.0061, -0.4775, 0.0276, 0.0083, -0.4775, 0.0273, 0.01, -0.4787, 0.0276, 0.01, -0.4775, 0.0283, 0.0061, -0.4744, 0.0283, 0.0083, -0.4744, 0.0286, 0.0061, -0.4722, 0.0286, 0.0083, -0.4722, 0.0353, 0.01, -0.4744, 0.0331, 0.01, -0.4744, 0.0353, 0.01, -0.4722, 0.0331, 0.01, -0.4722, 0.0309, 0.01, -0.4744, 0.0284, 0.01, -0.4743, 0.0309, 0.01, -0.4722, 0.0286, 0.01, -0.4722, 0.0287, 0.0061, -0.47, 0.0287, 0.0083, -0.47,
0.0287, 0.0061, -0.4679, 0.0287, 0.0083, -0.4679, 0.0353, 0.01, -0.47, 0.0331, 0.01, -0.47, 0.0353, 0.01, -0.4679, 0.0331, 0.01, -0.4679, 0.0309, 0.01, -0.47, 0.0287, 0.01, -0.47, 0.0309, 0.01, -0.4679, 0.0287, 0.01, -0.4679, 0.0148, 0.0061, -0.4801, 0.0148, 0.0083, -0.4801, 0.014, 0.0061, -0.4824, 0.0114, 0.0061, -0.4823, 0.014, 0.0083, -0.4824, 0.0114, 0.0083, -0.4823, 0.0144, 0.0061, -0.4812, 0.0144, 0.0083, -0.4812, 0.0148, 0.01, -0.4801, 0.014, 0.01, -0.4824, 0.0114, 0.01, -0.4823, 0.0145, 0.01, -0.4811, 0.0114, 0.01, -0.4809, 0.0152, 0.0061, -0.4787, 0.0152, 0.0083, -0.4787, 0.0159, 0.0061, -0.4766, 0.0159, 0.0083, -0.4766, 0.0152, 0.01, -0.4788, 0.0158, 0.01, -0.4767, 0.0136, 0.01, -0.4787,
0.0114, 0.01, -0.4787, 0.0136, 0.01, -0.4766, 0.0114, 0.01, -0.4766, 0.0094, 0.0061, -0.4821, 0.0094, 0.0083, -0.4821, 0.0083, 0.0061, -0.482, 0.0071, 0.0061, -0.4818, 0.0083, 0.0083, -0.482, 0.0071, 0.0083, -0.4818, 0.0049, 0.0061, -0.4815, 0.0027, 0.0061, -0.4812, 0.0049, 0.0083, -0.4815, 0.0027, 0.0083, -0.4812, 0.0097, 0.01, -0.4821, 0.0089, 0.01, -0.482, 0.0069, 0.01, -0.4817, 0.0048, 0.01, -0.4814, 0.0026, 0.01, -0.4811, 0.0092, 0.01, -0.4787, 0.0071, 0.01, -0.4787, 0.0092, 0.01, -0.4766, 0.0071, 0.01, -0.4766, 0.0049, 0.01, -0.4787, 0.0027, 0.01, -0.4787, 0.0049, 0.01, -0.4766, 0.0027, 0.01, -0.4766, 0.0164, 0.0061, -0.4744, 0.0164, 0.0083, -0.4744, 0.0168, 0.0061, -0.4713, 0.0167, 0.0061, -0.4724,
0.0168, 0.0083, -0.4713, 0.0167, 0.0083, -0.4724, 0.0164, 0.01, -0.4746, 0.0168, 0.01, -0.4713, 0.0168, 0.01, -0.4722, 0.0136, 0.01, -0.4744, 0.0114, 0.01, -0.4744, 0.0136, 0.01, -0.4722, 0.0114, 0.01, -0.4722, 0.0169, 0.0061, -0.47, 0.0169, 0.0083, -0.47, 0.017, 0.0061, -0.4679, 0.017, 0.0083, -0.4679, 0.0169, 0.01, -0.47, 0.0157, 0.01, -0.47, 0.017, 0.01, -0.4679, 0.0157, 0.01, -0.4679, 0.0136, 0.01, -0.47, 0.0114, 0.01, -0.47, 0.0136, 0.01, -0.4679, 0.0114, 0.01, -0.4679, 0.0092, 0.01, -0.4744, 0.0071, 0.01, -0.4744, 0.0092, 0.01, -0.4722, 0.0071, 0.01, -0.4722, 0.0049, 0.01, -0.4744, 0.0027, 0.01, -0.4744, 0.0049, 0.01, -0.4722, 0.0027, 0.01, -0.4722, 0.0092, 0.01, -0.47,
0.0071, 0.01, -0.47, 0.0092, 0.01, -0.4679, 0.0071, 0.01, -0.4679, 0.0049, 0.01, -0.47, 0.0027, 0.01, -0.47, 0.0049, 0.01, -0.4679, 0.0027, 0.01, -0.4679, 0.0287, 0.0061, -0.4657, 0.0287, 0.0083, -0.4657, 0.0287, 0.0061, -0.4635, 0.0287, 0.0083, -0.4635, 0.0353, 0.01, -0.4657, 0.0331, 0.01, -0.4657, 0.0353, 0.01, -0.4635, 0.0331, 0.01, -0.4635, 0.0309, 0.01, -0.4657, 0.0287, 0.01, -0.4657, 0.0309, 0.01, -0.4635, 0.0287, 0.01, -0.4636, 0.0284, 0.0061, -0.4614, 0.0284, 0.0083, -0.4614, 0.028, 0.0061, -0.4592, 0.028, 0.0083, -0.4592, 0.0353, 0.01, -0.4614, 0.0331, 0.01, -0.4614, 0.0353, 0.01, -0.4592, 0.0331, 0.01, -0.4592, 0.0309, 0.01, -0.4614, 0.0284, 0.01, -0.4615, 0.0309, 0.01, -0.4592,
0.028, 0.01, -0.4594, 0.0277, 0.0061, -0.458, 0.0277, 0.0083, -0.458, 0.0353, 0.01, -0.457, 0.0331, 0.01, -0.457, 0.0353, 0.01, -0.4549, 0.0331, 0.01, -0.4549, 0.0309, 0.01, -0.457, 0.0276, 0.01, -0.4574, 0.0309, 0.01, -0.4549, 0.0288, 0.01, -0.4549, 0.0353, 0.01, -0.4527, 0.0331, 0.01, -0.4527, 0.0353, 0.01, -0.4505, 0.0331, 0.01, -0.4505, 0.0309, 0.01, -0.4527, 0.0288, 0.01, -0.4527, 0.0309, 0.01, -0.4505, 0.0288, 0.01, -0.4505, 0.0274, 0.0061, -0.4569, 0.0274, 0.0083, -0.4569, 0.027, 0.0061, -0.4549, 0.027, 0.0083, -0.4549, 0.0274, 0.01, -0.4566, 0.027, 0.01, -0.4549, 0.0265, 0.0061, -0.4527, 0.0265, 0.0083, -0.4527, 0.026, 0.0061, -0.4505, 0.026, 0.0083, -0.4505, 0.0265, 0.01, -0.4528,
0.0261, 0.01, -0.4507, 0.017, 0.0061, -0.4657, 0.0157, 0.0061, -0.4657, 0.017, 0.0083, -0.4657, 0.0157, 0.0083, -0.4657, 0.0136, 0.0061, -0.4655, 0.0114, 0.0061, -0.4651, 0.0136, 0.0083, -0.4655, 0.0114, 0.0083, -0.4651, 0.017, 0.01, -0.4657, 0.0158, 0.01, -0.4657, 0.0136, 0.01, -0.4655, 0.0115, 0.01, -0.4652, 0.0095, 0.0061, -0.4648, 0.0095, 0.0083, -0.4648, 0.0084, 0.0061, -0.4646, 0.0071, 0.0061, -0.4643, 0.0084, 0.0083, -0.4646, 0.0071, 0.0083, -0.4643, 0.0049, 0.0061, -0.4638, 0.0027, 0.0061, -0.4633, 0.0049, 0.0083, -0.4638, 0.0027, 0.0083, -0.4633, 0.0093, 0.01, -0.4647, 0.0071, 0.01, -0.4657, 0.0083, 0.01, -0.4645, 0.0071, 0.01, -0.4643, 0.0049, 0.01, -0.4657, 0.0027, 0.01, -0.4657, 0.005, 0.01, -0.4638,
0.0029, 0.01, -0.4633, 0.07, 0.01, -0.4483, 0.0678, 0.01, -0.4483, 0.07, 0.01, -0.4462, 0.0678, 0.01, -0.4462, 0.0656, 0.01, -0.4483, 0.0635, 0.01, -0.4483, 0.0656, 0.01, -0.4462, 0.0635, 0.01, -0.4462, 0.07, 0.01, -0.444, 0.0678, 0.01, -0.444, 0.07, 0.01, -0.4418, 0.0678, 0.01, -0.4418, 0.0656, 0.01, -0.444, 0.0635, 0.01, -0.444, 0.0656, 0.01, -0.4418, 0.0635, 0.01, -0.4418, 0.0613, 0.01, -0.4483, 0.0591, 0.01, -0.4483, 0.0613, 0.01, -0.4462, 0.0591, 0.01, -0.4462, 0.057, 0.01, -0.4483, 0.0548, 0.01, -0.4483, 0.057, 0.01, -0.4462, 0.0548, 0.01, -0.4462, 0.0613, 0.01, -0.444, 0.0591, 0.01, -0.444, 0.0613, 0.01, -0.4418, 0.0591, 0.01, -0.4418, 0.057, 0.01, -0.444,
0.0548, 0.01, -0.444, 0.057, 0.01, -0.4418, 0.0548, 0.01, -0.4418, 0.07, 0.01, -0.4397, 0.0678, 0.01, -0.4397, 0.07, 0.01, -0.4375, 0.0678, 0.01, -0.4375, 0.0656, 0.01, -0.4397, 0.0635, 0.01, -0.4397, 0.0656, 0.01, -0.4375, 0.0635, 0.01, -0.4375, 0.07, 0.01, -0.4353, 0.0678, 0.01, -0.4353, 0.069, 0.01, -0.433, 0.0678, 0.01, -0.4332, 0.0656, 0.01, -0.4353, 0.0635, 0.01, -0.4353, 0.0656, 0.01, -0.4332, 0.0635, 0.01, -0.4332, 0.0613, 0.01, -0.4397, 0.0591, 0.01, -0.4397, 0.0613, 0.01, -0.4375, 0.0591, 0.01, -0.4375, 0.057, 0.01, -0.4397, 0.0548, 0.01, -0.4397, 0.057, 0.01, -0.4375, 0.0548, 0.01, -0.4375, 0.0613, 0.01, -0.4353, 0.0591, 0.01, -0.4353, 0.0613, 0.01, -0.4332,
0.0591, 0.01, -0.4332, 0.057, 0.01, -0.4353, 0.0548, 0.01, -0.4353, 0.057, 0.01, -0.4332, 0.0548, 0.01, -0.4332, 0.0503, 0.0061, -0.4483, 0.0503, 0.0083, -0.4483, 0.0501, 0.0061, -0.4462, 0.0501, 0.0083, -0.4462, 0.0526, 0.01, -0.4483, 0.0503, 0.01, -0.4484, 0.0526, 0.01, -0.4462, 0.0501, 0.01, -0.4463, 0.0497, 0.0061, -0.444, 0.0497, 0.0083, -0.444, 0.0495, 0.0061, -0.4424, 0.0495, 0.0083, -0.4424, 0.0493, 0.0061, -0.4413, 0.0493, 0.0083, -0.4413, 0.0526, 0.01, -0.444, 0.0498, 0.01, -0.4441, 0.0526, 0.01, -0.4418, 0.0494, 0.01, -0.442, 0.0492, 0.01, -0.4411, 0.0425, 0.0061, -0.4483, 0.0425, 0.0083, -0.4483, 0.0423, 0.0061, -0.4462, 0.0423, 0.0083, -0.4462, 0.0425, 0.01, -0.4483, 0.0423, 0.01, -0.4461,
0.0396, 0.01, -0.4483, 0.0374, 0.01, -0.4483, 0.0396, 0.01, -0.4462, 0.0374, 0.01, -0.4462, 0.042, 0.0061, -0.444, 0.042, 0.0083, -0.444, 0.0416, 0.0061, -0.4418, 0.0416, 0.0083, -0.4418, 0.042, 0.01, -0.4439, 0.0416, 0.01, -0.4417, 0.0396, 0.01, -0.444, 0.0374, 0.01, -0.444, 0.0396, 0.01, -0.4418, 0.0374, 0.01, -0.4418, 0.049, 0.0061, -0.4397, 0.049, 0.0083, -0.4397, 0.0486, 0.0061, -0.4375, 0.0486, 0.0083, -0.4375, 0.0526, 0.01, -0.4397, 0.0505, 0.01, -0.4397, 0.0526, 0.01, -0.4375, 0.0505, 0.01, -0.4375, 0.049, 0.01, -0.4397, 0.0486, 0.01, -0.4376, 0.0481, 0.0061, -0.4353, 0.0481, 0.0083, -0.4353, 0.0476, 0.0061, -0.4331, 0.0476, 0.0083, -0.4331, 0.0526, 0.01, -0.4353, 0.0505, 0.01, -0.4353,
0.0526, 0.01, -0.4332, 0.0505, 0.01, -0.4332, 0.0482, 0.01, -0.4355, 0.0476, 0.01, -0.4334, 0.0412, 0.0061, -0.4397, 0.0412, 0.0083, -0.4397, 0.0408, 0.0061, -0.4381, 0.0408, 0.0083, -0.4381, 0.0406, 0.0061, -0.437, 0.0406, 0.0083, -0.437, 0.0412, 0.01, -0.4396, 0.0409, 0.01, -0.4382, 0.0396, 0.01, -0.4397, 0.0374, 0.01, -0.4397, 0.0406, 0.01, -0.4373, 0.0374, 0.01, -0.4375, 0.0401, 0.0061, -0.4353, 0.0401, 0.0083, -0.4353, 0.0396, 0.0061, -0.4331, 0.0396, 0.0083, -0.4331, 0.0401, 0.01, -0.4351, 0.0374, 0.01, -0.4353, 0.0396, 0.01, -0.433, 0.0374, 0.01, -0.4332, 0.069, 0.0061, -0.4328, 0.0689, 0.0061, -0.4308, 0.069, 0.0083, -0.4328, 0.0689, 0.0083, -0.4308, 0.0686, 0.0061, -0.4288, 0.0686, 0.0083, -0.4288,
0.069, 0.01, -0.4328, 0.0689, 0.01, -0.4311, 0.0686, 0.01, -0.4286, 0.0656, 0.01, -0.431, 0.0635, 0.01, -0.431, 0.0656, 0.01, -0.4288, 0.0635, 0.01, -0.4288, 0.0681, 0.0061, -0.4266, 0.0681, 0.0083, -0.4266, 0.0675, 0.0061, -0.4245, 0.0675, 0.0083, -0.4245, 0.0681, 0.01, -0.4265, 0.0675, 0.01, -0.4244, 0.0656, 0.01, -0.4266, 0.0635, 0.01, -0.4266, 0.0656, 0.01, -0.4245, 0.0635, 0.01, -0.4245, 0.0613, 0.01, -0.431, 0.0591, 0.01, -0.431, 0.0613, 0.01, -0.4288, 0.0591, 0.01, -0.4288, 0.057, 0.01, -0.431, 0.0548, 0.01, -0.431, 0.057, 0.01, -0.4288, 0.0548, 0.01, -0.4288, 0.0613, 0.01, -0.4266, 0.0591, 0.01, -0.4266, 0.0613, 0.01, -0.4245, 0.0591, 0.01, -0.4245, 0.057, 0.01, -0.4266,
0.0548, 0.01, -0.4266, 0.057, 0.01, -0.4245, 0.0548, 0.01, -0.4245, 0.067, 0.0061, -0.4227, 0.067, 0.0083, -0.4227, 0.0666, 0.0061, -0.4216, 0.0666, 0.0083, -0.4216, 0.066, 0.0061, -0.4201, 0.066, 0.0083, -0.4201, 0.067, 0.01, -0.4229, 0.0667, 0.01, -0.4219, 0.0635, 0.01, -0.4223, 0.0659, 0.01, -0.4199, 0.0635, 0.01, -0.4201, 0.0651, 0.0061, -0.418, 0.0651, 0.0083, -0.418, 0.0646, 0.0061, -0.4167, 0.0642, 0.0061, -0.4156, 0.0646, 0.0083, -0.4167, 0.0642, 0.0083, -0.4156, 0.0651, 0.01, -0.4179, 0.0635, 0.01, -0.418, 0.0647, 0.01, -0.4167, 0.0642, 0.01, -0.4157, 0.0613, 0.01, -0.4223, 0.0591, 0.01, -0.4223, 0.0613, 0.01, -0.4201, 0.0591, 0.01, -0.4201, 0.057, 0.01, -0.4223, 0.0548, 0.01, -0.4223,
0.057, 0.01, -0.4201, 0.0548, 0.01, -0.4201, 0.0613, 0.01, -0.418, 0.0591, 0.01, -0.418, 0.0613, 0.01, -0.4158, 0.0591, 0.01, -0.4158, 0.057, 0.01, -0.418, 0.0548, 0.01, -0.418, 0.057, 0.01, -0.4158, 0.0548, 0.01, -0.4158, 0.0472, 0.0061, -0.432, 0.0469, 0.0061, -0.4309, 0.0472, 0.0083, -0.432, 0.0469, 0.0083, -0.4309, 0.0462, 0.0061, -0.4288, 0.0462, 0.0083, -0.4288, 0.0526, 0.01, -0.431, 0.0505, 0.01, -0.431, 0.0526, 0.01, -0.4288, 0.0505, 0.01, -0.4288, 0.0471, 0.01, -0.4315, 0.0468, 0.01, -0.4306, 0.0483, 0.01, -0.4288, 0.0463, 0.01, -0.429, 0.0454, 0.0061, -0.4267, 0.0454, 0.0083, -0.4267, 0.0526, 0.01, -0.4266, 0.0505, 0.01, -0.4266, 0.0526, 0.01, -0.4245, 0.0505, 0.01, -0.4245,
0.0483, 0.01, -0.4266, 0.0454, 0.01, -0.4267, 0.0483, 0.01, -0.4245, 0.0461, 0.01, -0.4245, 0.039, 0.0061, -0.431, 0.039, 0.0083, -0.431, 0.0386, 0.0061, -0.4297, 0.0382, 0.0061, -0.4286, 0.0386, 0.0083, -0.4297, 0.0382, 0.0083, -0.4286, 0.039, 0.01, -0.4309, 0.0374, 0.01, -0.431, 0.0386, 0.01, -0.4298, 0.0382, 0.01, -0.4287, 0.045, 0.0061, -0.4256, 0.045, 0.0083, -0.4256, 0.0446, 0.0061, -0.4245, 0.0446, 0.0083, -0.4245, 0.0373, 0.0061, -0.4266, 0.0373, 0.0083, -0.4266, 0.0366, 0.0061, -0.425, 0.0366, 0.0083, -0.425, 0.045, 0.01, -0.4256, 0.0446, 0.01, -0.4245, 0.0373, 0.01, -0.4265, 0.0367, 0.01, -0.4251, 0.0526, 0.01, -0.4223, 0.0505, 0.01, -0.4223, 0.0526, 0.01, -0.4201, 0.0505, 0.01, -0.4201,
0.0483, 0.01, -0.4223, 0.0461, 0.01, -0.4223, 0.0483, 0.01, -0.4201, 0.0461, 0.01, -0.4201, 0.0526, 0.01, -0.418, 0.0505, 0.01, -0.418, 0.0526, 0.01, -0.4158, 0.0505, 0.01, -0.4158, 0.0483, 0.01, -0.418, 0.0461, 0.01, -0.418, 0.0483, 0.01, -0.4158, 0.0461, 0.01, -0.4158, 0.0436, 0.0061, -0.4223, 0.0436, 0.0083, -0.4223, 0.043, 0.0061, -0.4209, 0.0425, 0.0061, -0.4198, 0.043, 0.0083, -0.4209, 0.0425, 0.0083, -0.4198, 0.0437, 0.01, -0.4226, 0.0429, 0.01, -0.4206, 0.0425, 0.01, -0.4196, 0.0418, 0.0061, -0.418, 0.0418, 0.0083, -0.418, 0.041, 0.0061, -0.4162, 0.041, 0.0083, -0.4162, 0.0405, 0.0061, -0.4152, 0.0405, 0.0083, -0.4152, 0.0439, 0.01, -0.418, 0.0419, 0.01, -0.4182, 0.0439, 0.01, -0.4158,
0.0409, 0.01, -0.4161, 0.0404, 0.01, -0.4151, 0.0353, 0.01, -0.4483, 0.0331, 0.01, -0.4483, 0.0353, 0.01, -0.4462, 0.0331, 0.01, -0.4462, 0.0309, 0.01, -0.4483, 0.0288, 0.01, -0.4483, 0.0309, 0.01, -0.4462, 0.0288, 0.01, -0.4462, 0.0353, 0.01, -0.444, 0.0331, 0.01, -0.444, 0.0353, 0.01, -0.4418, 0.0331, 0.01, -0.4418, 0.0309, 0.01, -0.444, 0.0288, 0.01, -0.444, 0.0309, 0.01, -0.4418, 0.0288, 0.01, -0.4418, 0.0256, 0.0061, -0.4489, 0.0253, 0.0061, -0.4479, 0.0256, 0.0083, -0.4489, 0.0253, 0.0083, -0.4479, 0.0248, 0.0061, -0.4462, 0.0248, 0.0083, -0.4462, 0.0255, 0.01, -0.4486, 0.0253, 0.01, -0.4477, 0.0266, 0.01, -0.4462, 0.0249, 0.01, -0.4463, 0.0241, 0.0061, -0.444, 0.0241, 0.0083, -0.444,
0.0235, 0.0061, -0.4424, 0.0235, 0.0083, -0.4424, 0.0231, 0.0061, -0.4413, 0.0231, 0.0083, -0.4413, 0.0266, 0.01, -0.444, 0.0242, 0.01, -0.4442, 0.0266, 0.01, -0.4418, 0.0234, 0.01, -0.4422, 0.0231, 0.01, -0.4412, 0.0353, 0.01, -0.4397, 0.0331, 0.01, -0.4397, 0.0353, 0.01, -0.4375, 0.0331, 0.01, -0.4375, 0.0309, 0.01, -0.4397, 0.0288, 0.01, -0.4397, 0.0309, 0.01, -0.4375, 0.0288, 0.01, -0.4375, 0.0353, 0.01, -0.4353, 0.0331, 0.01, -0.4353, 0.0353, 0.01, -0.4332, 0.0331, 0.01, -0.4332, 0.0309, 0.01, -0.4353, 0.0288, 0.01, -0.4353, 0.0309, 0.01, -0.4332, 0.0288, 0.01, -0.4332, 0.0225, 0.0061, -0.4397, 0.0225, 0.0083, -0.4397, 0.0216, 0.0061, -0.4375, 0.0212, 0.0061, -0.4364, 0.0216, 0.0083, -0.4375,
0.0212, 0.0083, -0.4364, 0.0266, 0.01, -0.4397, 0.0244, 0.01, -0.4397, 0.0266, 0.01, -0.4375, 0.0244, 0.01, -0.4375, 0.0225, 0.01, -0.4398, 0.0216, 0.01, -0.4375, 0.0212, 0.01, -0.4364, 0.0207, 0.0061, -0.4353, 0.0207, 0.0083, -0.4353, 0.0197, 0.0061, -0.4332, 0.0197, 0.0083, -0.4332, 0.0266, 0.01, -0.4353, 0.0244, 0.01, -0.4353, 0.0266, 0.01, -0.4332, 0.0244, 0.01, -0.4332, 0.0222, 0.01, -0.4353, 0.0207, 0.01, -0.4354, 0.0222, 0.01, -0.4332, 0.0199, 0.01, -0.4334, 0.0353, 0.01, -0.431, 0.0331, 0.01, -0.431, 0.0353, 0.01, -0.4288, 0.0331, 0.01, -0.4288, 0.0309, 0.01, -0.431, 0.0288, 0.01, -0.431, 0.0309, 0.01, -0.4288, 0.0288, 0.01, -0.4288, 0.0361, 0.0061, -0.4239, 0.0361, 0.0083, -0.4239,
0.0353, 0.01, -0.4266, 0.0331, 0.01, -0.4266, 0.0362, 0.01, -0.4241, 0.0331, 0.01, -0.4245, 0.0309, 0.01, -0.4266, 0.0288, 0.01, -0.4266, 0.0309, 0.01, -0.4245, 0.0288, 0.01, -0.4245, 0.0191, 0.0061, -0.4318, 0.0191, 0.0083, -0.4318, 0.0266, 0.01, -0.431, 0.0244, 0.01, -0.431, 0.0266, 0.01, -0.4288, 0.0244, 0.01, -0.4288, 0.0222, 0.01, -0.431, 0.019, 0.01, -0.4315, 0.0222, 0.01, -0.4288, 0.0201, 0.01, -0.4288, 0.0266, 0.01, -0.4266, 0.0244, 0.01, -0.4266, 0.0266, 0.01, -0.4245, 0.0244, 0.01, -0.4245, 0.0222, 0.01, -0.4266, 0.0201, 0.01, -0.4266, 0.0222, 0.01, -0.4245, 0.0201, 0.01, -0.4245, 0.0354, 0.0061, -0.4223, 0.0354, 0.0083, -0.4223, 0.0345, 0.0061, -0.4204, 0.034, 0.0061, -0.4194,
0.0345, 0.0083, -0.4204, 0.034, 0.0083, -0.4194, 0.0353, 0.01, -0.4221, 0.0331, 0.01, -0.4223, 0.0346, 0.01, -0.4206, 0.0342, 0.01, -0.4196, 0.0309, 0.01, -0.4223, 0.0288, 0.01, -0.4223, 0.0309, 0.01, -0.4201, 0.0288, 0.01, -0.4201, 0.0333, 0.0061, -0.4179, 0.0333, 0.0083, -0.4179, 0.0324, 0.0061, -0.4162, 0.0324, 0.0083, -0.4162, 0.0318, 0.0061, -0.4151, 0.0318, 0.0083, -0.4151, 0.0332, 0.01, -0.4177, 0.0325, 0.01, -0.4163, 0.0309, 0.01, -0.418, 0.0288, 0.01, -0.418, 0.0319, 0.01, -0.4153, 0.0288, 0.01, -0.4158, 0.0266, 0.01, -0.4223, 0.0244, 0.01, -0.4223, 0.0266, 0.01, -0.4201, 0.0244, 0.01, -0.4201, 0.0222, 0.01, -0.4223, 0.0201, 0.01, -0.4223, 0.0222, 0.01, -0.4201, 0.0201, 0.01, -0.4201,
0.0266, 0.01, -0.418, 0.0244, 0.01, -0.418, 0.0266, 0.01, -0.4158, 0.0244, 0.01, -0.4158, 0.0222, 0.01, -0.418, 0.0201, 0.01, -0.418, 0.0222, 0.01, -0.4158, 0.0201, 0.01, -0.4158, 0.0186, 0.0061, -0.4307, 0.0186, 0.0083, -0.4307, 0.0178, 0.0061, -0.4288, 0.0178, 0.0083, -0.4288, 0.0185, 0.01, -0.4305, 0.0179, 0.01, -0.429, 0.017, 0.0061, -0.4272, 0.0165, 0.0061, -0.4262, 0.017, 0.0083, -0.4272, 0.0165, 0.0083, -0.4262, 0.0157, 0.0061, -0.4245, 0.0157, 0.0083, -0.4245, 0.017, 0.01, -0.427, 0.0165, 0.01, -0.426, 0.0179, 0.01, -0.4245, 0.0158, 0.01, -0.4247, 0.0017, 0.0061, -0.4258, 0.0017, 0.0083, -0.4258, 0.0023, 0.0061, -0.4245, 0.0023, 0.0083, -0.4245, 0.0018, 0.01, -0.4257, 0.0023, 0.01, -0.4246,
0.0149, 0.0061, -0.423, 0.0149, 0.0083, -0.423, 0.0143, 0.0061, -0.4219, 0.0143, 0.0083, -0.4219, 0.0134, 0.0061, -0.4201, 0.0134, 0.0083, -0.4201, 0.0179, 0.01, -0.4223, 0.0148, 0.01, -0.4228, 0.0179, 0.01, -0.4201, 0.0157, 0.01, -0.4201, 0.0142, 0.01, -0.4217, 0.0135, 0.01, -0.4204, 0.0127, 0.0061, -0.4187, 0.0121, 0.0061, -0.4176, 0.0127, 0.0083, -0.4187, 0.0121, 0.0083, -0.4176, 0.0111, 0.0061, -0.4158, 0.0111, 0.0083, -0.4158, 0.0179, 0.01, -0.418, 0.0157, 0.01, -0.418, 0.0179, 0.01, -0.4158, 0.0157, 0.01, -0.4158, 0.0126, 0.01, -0.4185, 0.012, 0.01, -0.4175, 0.0136, 0.01, -0.4158, 0.0113, 0.01, -0.4161, 0.003, 0.0061, -0.4223, 0.003, 0.0083, -0.4223, 0.0032, 0.0061, -0.4201, 0.0032, 0.0083, -0.4201,
0.0029, 0.01, -0.4224, 0.0032, 0.01, -0.4201, 0.003, 0.0061, -0.4179, 0.003, 0.0083, -0.4179, 0.0026, 0.0061, -0.4158, 0.0026, 0.0083, -0.4158, 0.003, 0.01, -0.4179, 0.0025, 0.01, -0.4156, 0.1059, 0.0061, -0.3977, 0.1059, 0.0083, -0.3977, 0.1059, 0.01, -0.3976, 0.1061, 0.0061, -0.3963, 0.1061, 0.0083, -0.3963, 0.1066, 0.0061, -0.3941, 0.1066, 0.0083, -0.3941, 0.1061, 0.01, -0.3963, 0.1066, 0.01, -0.3942, 0.1069, 0.0061, -0.3919, 0.1069, 0.0083, -0.3919, 0.1071, 0.0061, -0.3898, 0.1071, 0.0083, -0.3898, 0.1069, 0.01, -0.392, 0.1071, 0.01, -0.3898, 0.1072, 0.0061, -0.3876, 0.1072, 0.0083, -0.3876, 0.1072, 0.0061, -0.3854, 0.1072, 0.0083, -0.3854, 0.1072, 0.01, -0.3876, 0.1072, 0.01, -0.3854, 0.1072, 0.0061, -0.3832,
0.1072, 0.0083, -0.3832, 0.1072, 0.0061, -0.3811, 0.1072, 0.0083, -0.3811, 0.1072, 0.01, -0.3832, 0.1072, 0.01, -0.3811, 0.1016, 0.0061, -0.413, 0.1016, 0.0083, -0.413, 0.1022, 0.0061, -0.4115, 0.1022, 0.0083, -0.4115, 0.1012, 0.0061, -0.4141, 0.1012, 0.0083, -0.4141, 0.1017, 0.01, -0.4129, 0.1022, 0.01, -0.4116, 0.1013, 0.01, -0.4139, 0.0982, 0.01, -0.4136, 0.1004, 0.01, -0.4115, 0.0982, 0.01, -0.4115, 0.103, 0.0061, -0.4093, 0.103, 0.0083, -0.4093, 0.1038, 0.0061, -0.4067, 0.1035, 0.0061, -0.4078, 0.1038, 0.0083, -0.4067, 0.1035, 0.0083, -0.4078, 0.1029, 0.01, -0.4095, 0.1039, 0.01, -0.4065, 0.1036, 0.01, -0.4075, 0.1004, 0.01, -0.4093, 0.0982, 0.01, -0.4093, 0.1004, 0.01, -0.4071, 0.0982, 0.01, -0.4071,
0.096, 0.01, -0.4136, 0.0939, 0.01, -0.4136, 0.096, 0.01, -0.4115, 0.0939, 0.01, -0.4115, 0.0917, 0.01, -0.4136, 0.0895, 0.01, -0.4136, 0.0917, 0.01, -0.4115, 0.0895, 0.01, -0.4115, 0.096, 0.01, -0.4093, 0.0939, 0.01, -0.4093, 0.096, 0.01, -0.4071, 0.0939, 0.01, -0.4071, 0.0917, 0.01, -0.4093, 0.0895, 0.01, -0.4093, 0.0917, 0.01, -0.4071, 0.0895, 0.01, -0.4071, 0.1043, 0.0061, -0.4049, 0.1043, 0.0083, -0.4049, 0.1048, 0.0061, -0.4028, 0.1048, 0.0083, -0.4028, 0.1043, 0.01, -0.405, 0.1025, 0.01, -0.4049, 0.1048, 0.01, -0.4029, 0.1025, 0.01, -0.4028, 0.1004, 0.01, -0.4049, 0.0982, 0.01, -0.4049, 0.1004, 0.01, -0.4028, 0.0982, 0.01, -0.4028, 0.1053, 0.0061, -0.4006, 0.1053, 0.0083, -0.4006,
0.1056, 0.0061, -0.3988, 0.1056, 0.0083, -0.3988, 0.1053, 0.01, -0.4008, 0.1025, 0.01, -0.4006, 0.1057, 0.01, -0.3985, 0.1025, 0.01, -0.3984, 0.1004, 0.01, -0.4006, 0.0982, 0.01, -0.4006, 0.1004, 0.01, -0.3984, 0.0982, 0.01, -0.3984, 0.0887, 0.0061, -0.4024, 0.0887, 0.0083, -0.4024, 0.096, 0.01, -0.4049, 0.0939, 0.01, -0.4049, 0.096, 0.01, -0.4028, 0.0939, 0.01, -0.4028, 0.0917, 0.01, -0.4049, 0.0895, 0.01, -0.4049, 0.0917, 0.01, -0.4028, 0.0886, 0.01, -0.4026, 0.0893, 0.0061, -0.4006, 0.0893, 0.0083, -0.4006, 0.0899, 0.0061, -0.3984, 0.0899, 0.0083, -0.3984, 0.096, 0.01, -0.4006, 0.0939, 0.01, -0.4006, 0.096, 0.01, -0.3984, 0.0939, 0.01, -0.3984, 0.0917, 0.01, -0.4006, 0.0894, 0.01, -0.4004,
0.0917, 0.01, -0.3984, 0.09, 0.01, -0.3984, 0.0843, 0.0061, -0.4129, 0.0843, 0.0083, -0.4129, 0.085, 0.0061, -0.4115, 0.085, 0.0083, -0.4115, 0.0838, 0.0061, -0.414, 0.0838, 0.0083, -0.414, 0.0874, 0.01, -0.4136, 0.0842, 0.01, -0.4131, 0.0874, 0.01, -0.4115, 0.0851, 0.01, -0.4112, 0.0837, 0.01, -0.4141, 0.0863, 0.0061, -0.4084, 0.0859, 0.0061, -0.4095, 0.0863, 0.0083, -0.4084, 0.0859, 0.0083, -0.4095, 0.0869, 0.0061, -0.4071, 0.0869, 0.0083, -0.4071, 0.0862, 0.01, -0.4087, 0.0858, 0.01, -0.4097, 0.087, 0.01, -0.4068, 0.0877, 0.0061, -0.4049, 0.0877, 0.0083, -0.4049, 0.0883, 0.0061, -0.4035, 0.0883, 0.0083, -0.4035, 0.0878, 0.01, -0.4048, 0.0883, 0.01, -0.4036, 0.1047, 0.01, -0.3963, 0.1025, 0.01, -0.3963,
0.1047, 0.01, -0.3941, 0.1025, 0.01, -0.3941, 0.1004, 0.01, -0.3963, 0.0982, 0.01, -0.3963, 0.1004, 0.01, -0.3941, 0.0982, 0.01, -0.3941, 0.1047, 0.01, -0.3919, 0.1025, 0.01, -0.3919, 0.1047, 0.01, -0.3897, 0.1025, 0.01, -0.3897, 0.1004, 0.01, -0.3919, 0.0982, 0.01, -0.3919, 0.1004, 0.01, -0.3897, 0.0982, 0.01, -0.3897, 0.0907, 0.0061, -0.3954, 0.0904, 0.0061, -0.3965, 0.0907, 0.0083, -0.3954, 0.0904, 0.0083, -0.3965, 0.091, 0.0061, -0.3941, 0.091, 0.0083, -0.3941, 0.096, 0.01, -0.3963, 0.0939, 0.01, -0.3963, 0.096, 0.01, -0.3941, 0.0939, 0.01, -0.3941, 0.0905, 0.01, -0.3959, 0.0903, 0.01, -0.3967, 0.091, 0.01, -0.3939, 0.0915, 0.0061, -0.3919, 0.0915, 0.0083, -0.3919, 0.0921, 0.0061, -0.3898,
0.0921, 0.0083, -0.3898, 0.096, 0.01, -0.3919, 0.0939, 0.01, -0.3919, 0.096, 0.01, -0.3897, 0.0939, 0.01, -0.3897, 0.0916, 0.01, -0.3918, 0.0921, 0.01, -0.3897, 0.1047, 0.01, -0.3876, 0.1025, 0.01, -0.3876, 0.1047, 0.01, -0.3854, 0.1025, 0.01, -0.3854, 0.1004, 0.01, -0.3876, 0.0982, 0.01, -0.3876, 0.1004, 0.01, -0.3854, 0.0982, 0.01, -0.3854, 0.1047, 0.01, -0.3832, 0.1025, 0.01, -0.3832, 0.1047, 0.01, -0.3811, 0.1025, 0.01, -0.3811, 0.1004, 0.01, -0.3832, 0.0982, 0.01, -0.3832, 0.1004, 0.01, -0.3811, 0.0982, 0.01, -0.3811, 0.0924, 0.0061, -0.3876, 0.0924, 0.0083, -0.3876, 0.0925, 0.0061, -0.3854, 0.0925, 0.0083, -0.3854, 0.096, 0.01, -0.3876, 0.0939, 0.01, -0.3876, 0.096, 0.01, -0.3854,
0.0939, 0.01, -0.3854, 0.0924, 0.01, -0.3876, 0.0925, 0.01, -0.3854, 0.0928, 0.0061, -0.3827, 0.0928, 0.0083, -0.3827, 0.0931, 0.0061, -0.3811, 0.0931, 0.0083, -0.3811, 0.0927, 0.0061, -0.3838, 0.0927, 0.0083, -0.3838, 0.096, 0.01, -0.3832, 0.0928, 0.01, -0.3831, 0.096, 0.01, -0.3811, 0.0931, 0.01, -0.381, 0.0927, 0.01, -0.3839, 0.1072, 0.0061, -0.3789, 0.1072, 0.0083, -0.3789, 0.1072, 0.0061, -0.3767, 0.1072, 0.0083, -0.3767, 0.1072, 0.01, -0.3789, 0.1072, 0.01, -0.3767, 0.1072, 0.0061, -0.3746, 0.1072, 0.0083, -0.3746, 0.1072, 0.0061, -0.3724, 0.1072, 0.0083, -0.3724, 0.1072, 0.01, -0.3746, 0.1072, 0.01, -0.3724, 0.1072, 0.0061, -0.3702, 0.1072, 0.0083, -0.3702, 0.1072, 0.0061, -0.368, 0.1072, 0.0083, -0.368,
0.1072, 0.01, -0.3702, 0.1071, 0.01, -0.368, 0.107, 0.0061, -0.3659, 0.107, 0.0083, -0.3659, 0.1067, 0.0061, -0.3637, 0.1067, 0.0083, -0.3637, 0.107, 0.01, -0.3658, 0.1067, 0.01, -0.3636, 0.1062, 0.0061, -0.3615, 0.1062, 0.0083, -0.3615, 0.1059, 0.0061, -0.36, 0.1059, 0.0083, -0.36, 0.1062, 0.01, -0.3615, 0.1059, 0.01, -0.3601, 0.1047, 0.01, -0.3789, 0.1025, 0.01, -0.3789, 0.1047, 0.01, -0.3767, 0.1025, 0.01, -0.3767, 0.1004, 0.01, -0.3789, 0.0982, 0.01, -0.3789, 0.1004, 0.01, -0.3767, 0.0982, 0.01, -0.3767, 0.1047, 0.01, -0.3746, 0.1025, 0.01, -0.3746, 0.1047, 0.01, -0.3724, 0.1025, 0.01, -0.3724, 0.1004, 0.01, -0.3746, 0.0982, 0.01, -0.3746, 0.1004, 0.01, -0.3724, 0.0982, 0.01, -0.3724,
0.0933, 0.0061, -0.3789, 0.0933, 0.0083, -0.3789, 0.0934, 0.0061, -0.3767, 0.0934, 0.0083, -0.3767, 0.096, 0.01, -0.3789, 0.0933, 0.01, -0.3788, 0.096, 0.01, -0.3767, 0.0934, 0.01, -0.3767, 0.0934, 0.0061, -0.3746, 0.0934, 0.0083, -0.3746, 0.0934, 0.0061, -0.3724, 0.0934, 0.0083, -0.3724, 0.096, 0.01, -0.3746, 0.0934, 0.01, -0.3746, 0.096, 0.01, -0.3724, 0.0934, 0.01, -0.3724, 0.1047, 0.01, -0.3702, 0.1025, 0.01, -0.3702, 0.1047, 0.01, -0.368, 0.1025, 0.01, -0.368, 0.1004, 0.01, -0.3702, 0.0982, 0.01, -0.3702, 0.1004, 0.01, -0.368, 0.0982, 0.01, -0.368, 0.1047, 0.01, -0.3659, 0.1025, 0.01, -0.3659, 0.1047, 0.01, -0.3637, 0.1025, 0.01, -0.3637, 0.1004, 0.01, -0.3659, 0.0982, 0.01, -0.3659,
0.1004, 0.01, -0.3637, 0.0982, 0.01, -0.3637, 0.0934, 0.0061, -0.3702, 0.0934, 0.0083, -0.3702, 0.0931, 0.0061, -0.368, 0.0931, 0.0083, -0.368, 0.096, 0.01, -0.3702, 0.0934, 0.01, -0.3703, 0.096, 0.01, -0.368, 0.0931, 0.01, -0.3682, 0.0929, 0.0061, -0.3664, 0.0929, 0.0083, -0.3664, 0.0927, 0.0061, -0.3653, 0.0927, 0.0083, -0.3653, 0.0925, 0.0061, -0.3637, 0.0925, 0.0083, -0.3637, 0.096, 0.01, -0.3659, 0.0928, 0.01, -0.366, 0.096, 0.01, -0.3637, 0.0939, 0.01, -0.3637, 0.0927, 0.01, -0.3651, 0.0925, 0.01, -0.3637, 0.1057, 0.0061, -0.3589, 0.1057, 0.0083, -0.3589, 0.1047, 0.01, -0.3615, 0.1025, 0.01, -0.3615, 0.1057, 0.01, -0.3592, 0.1025, 0.01, -0.3594, 0.1004, 0.01, -0.3615, 0.0982, 0.01, -0.3615,
0.1004, 0.01, -0.3594, 0.0982, 0.01, -0.3594, 0.1053, 0.0061, -0.3572, 0.1053, 0.0083, -0.3572, 0.1048, 0.0061, -0.355, 0.1048, 0.0083, -0.355, 0.1052, 0.01, -0.357, 0.1025, 0.01, -0.3572, 0.1048, 0.01, -0.3549, 0.1025, 0.01, -0.355, 0.1004, 0.01, -0.3572, 0.0982, 0.01, -0.3572, 0.1004, 0.01, -0.355, 0.0982, 0.01, -0.355, 0.0924, 0.0061, -0.3615, 0.0924, 0.0083, -0.3615, 0.0922, 0.0061, -0.3594, 0.0922, 0.0083, -0.3594, 0.096, 0.01, -0.3615, 0.0939, 0.01, -0.3615, 0.096, 0.01, -0.3594, 0.0939, 0.01, -0.3594, 0.0924, 0.01, -0.3616, 0.0922, 0.01, -0.3594, 0.0918, 0.0061, -0.3572, 0.0918, 0.0083, -0.3572, 0.0912, 0.0061, -0.355, 0.0912, 0.0083, -0.355, 0.096, 0.01, -0.3572, 0.0939, 0.01, -0.3572,
0.096, 0.01, -0.355, 0.0939, 0.01, -0.355, 0.0918, 0.01, -0.3573, 0.0913, 0.01, -0.3552, 0.1043, 0.0061, -0.3529, 0.1043, 0.0083, -0.3529, 0.1039, 0.0061, -0.3509, 0.1036, 0.0061, -0.3498, 0.1039, 0.0083, -0.3509, 0.1036, 0.0083, -0.3498, 0.1043, 0.01, -0.3528, 0.1025, 0.01, -0.3529, 0.1039, 0.01, -0.3512, 0.1037, 0.01, -0.3503, 0.1004, 0.01, -0.3529, 0.0982, 0.01, -0.3529, 0.1004, 0.01, -0.3507, 0.0982, 0.01, -0.3507, 0.1032, 0.0061, -0.3485, 0.1032, 0.0083, -0.3485, 0.1025, 0.0061, -0.3463, 0.1025, 0.0083, -0.3463, 0.1031, 0.01, -0.3483, 0.1024, 0.01, -0.3462, 0.1004, 0.01, -0.3485, 0.0982, 0.01, -0.3485, 0.1004, 0.01, -0.3463, 0.0982, 0.01, -0.3463, 0.0908, 0.0061, -0.3534, 0.0905, 0.0061, -0.3523,
0.0908, 0.0083, -0.3534, 0.0905, 0.0083, -0.3523, 0.0901, 0.0061, -0.3507, 0.0901, 0.0083, -0.3507, 0.096, 0.01, -0.3529, 0.0939, 0.01, -0.3529, 0.096, 0.01, -0.3507, 0.0939, 0.01, -0.3507, 0.0907, 0.01, -0.3531, 0.0904, 0.01, -0.3521, 0.0917, 0.01, -0.3507, 0.0901, 0.01, -0.3508, 0.0895, 0.0061, -0.3485, 0.0895, 0.0083, -0.3485, 0.0888, 0.0061, -0.3465, 0.0888, 0.0083, -0.3465, 0.096, 0.01, -0.3485, 0.0939, 0.01, -0.3485, 0.096, 0.01, -0.3463, 0.0939, 0.01, -0.3463, 0.0917, 0.01, -0.3485, 0.0895, 0.01, -0.3487, 0.0917, 0.01, -0.3463, 0.0888, 0.01, -0.3464, 0.0884, 0.0061, -0.3454, 0.0884, 0.0083, -0.3454, 0.0884, 0.01, -0.3454, 0.0634, 0.0061, -0.4136, 0.0634, 0.0083, -0.4136, 0.0627, 0.0061, -0.4119,
0.0627, 0.0083, -0.4119, 0.0633, 0.01, -0.4134, 0.0627, 0.01, -0.412, 0.0622, 0.0061, -0.4108, 0.0622, 0.0083, -0.4108, 0.0613, 0.01, -0.4136, 0.0591, 0.01, -0.4136, 0.0623, 0.01, -0.411, 0.0591, 0.01, -0.4115, 0.057, 0.01, -0.4136, 0.0548, 0.01, -0.4136, 0.057, 0.01, -0.4115, 0.0548, 0.01, -0.4115, 0.0615, 0.0061, -0.4093, 0.0615, 0.0083, -0.4093, 0.0606, 0.0061, -0.4074, 0.0601, 0.0061, -0.4063, 0.0606, 0.0083, -0.4074, 0.0601, 0.0083, -0.4063, 0.0614, 0.01, -0.409, 0.0591, 0.01, -0.4093, 0.0607, 0.01, -0.4076, 0.0602, 0.01, -0.4065, 0.057, 0.01, -0.4093, 0.0548, 0.01, -0.4093, 0.057, 0.01, -0.4071, 0.0548, 0.01, -0.4071, 0.0594, 0.0061, -0.4049, 0.0594, 0.0083, -0.4049, 0.0584, 0.0061, -0.4032,
0.0584, 0.0083, -0.4032, 0.0578, 0.0061, -0.4021, 0.0578, 0.0083, -0.4021, 0.0592, 0.01, -0.4047, 0.0585, 0.01, -0.4033, 0.057, 0.01, -0.4049, 0.0548, 0.01, -0.4049, 0.0579, 0.01, -0.4023, 0.0548, 0.01, -0.4028, 0.0569, 0.0061, -0.4006, 0.0569, 0.0083, -0.4006, 0.0561, 0.0061, -0.3991, 0.0555, 0.0061, -0.398, 0.0561, 0.0083, -0.3991, 0.0555, 0.0083, -0.398, 0.0568, 0.01, -0.4004, 0.0548, 0.01, -0.4006, 0.0561, 0.01, -0.3992, 0.0555, 0.01, -0.3981, 0.0526, 0.01, -0.4136, 0.0505, 0.01, -0.4136, 0.0526, 0.01, -0.4115, 0.0505, 0.01, -0.4115, 0.0483, 0.01, -0.4136, 0.0461, 0.01, -0.4136, 0.0483, 0.01, -0.4115, 0.0461, 0.01, -0.4115, 0.0526, 0.01, -0.4093, 0.0505, 0.01, -0.4093, 0.0526, 0.01, -0.4071,
0.0505, 0.01, -0.4071, 0.0483, 0.01, -0.4093, 0.0461, 0.01, -0.4093, 0.0483, 0.01, -0.4071, 0.0461, 0.01, -0.4071, 0.0397, 0.0061, -0.4136, 0.0397, 0.0083, -0.4136, 0.0389, 0.0061, -0.4119, 0.0383, 0.0061, -0.4108, 0.0389, 0.0083, -0.4119, 0.0383, 0.0083, -0.4108, 0.0439, 0.01, -0.4136, 0.0418, 0.01, -0.4136, 0.0439, 0.01, -0.4115, 0.0418, 0.01, -0.4115, 0.0398, 0.01, -0.4138, 0.0388, 0.01, -0.4118, 0.0383, 0.01, -0.4107, 0.0375, 0.0061, -0.4093, 0.0375, 0.0083, -0.4093, 0.0366, 0.0061, -0.4077, 0.0366, 0.0083, -0.4077, 0.0439, 0.01, -0.4093, 0.0418, 0.01, -0.4093, 0.0439, 0.01, -0.4071, 0.0418, 0.01, -0.4071, 0.0396, 0.01, -0.4093, 0.0376, 0.01, -0.4095, 0.0396, 0.01, -0.4071, 0.0366, 0.01, -0.4076,
0.0526, 0.01, -0.4049, 0.0505, 0.01, -0.4049, 0.0526, 0.01, -0.4028, 0.0505, 0.01, -0.4028, 0.0483, 0.01, -0.4049, 0.0461, 0.01, -0.4049, 0.0483, 0.01, -0.4028, 0.0461, 0.01, -0.4028, 0.0526, 0.01, -0.4006, 0.0505, 0.01, -0.4006, 0.0526, 0.01, -0.3984, 0.0505, 0.01, -0.3984, 0.0483, 0.01, -0.4006, 0.0461, 0.01, -0.4006, 0.0483, 0.01, -0.3984, 0.0461, 0.01, -0.3984, 0.0439, 0.01, -0.4049, 0.0418, 0.01, -0.4049, 0.0439, 0.01, -0.4028, 0.0418, 0.01, -0.4028, 0.0396, 0.01, -0.4049, 0.0374, 0.01, -0.4049, 0.0396, 0.01, -0.4028, 0.0374, 0.01, -0.4028, 0.0439, 0.01, -0.4006, 0.0418, 0.01, -0.4006, 0.0439, 0.01, -0.3984, 0.0418, 0.01, -0.3984, 0.0396, 0.01, -0.4006, 0.0374, 0.01, -0.4006,
0.0396, 0.01, -0.3984, 0.0374, 0.01, -0.3984, 0.0545, 0.0061, -0.3963, 0.0545, 0.0083, -0.3963, 0.0538, 0.0061, -0.395, 0.0538, 0.0083, -0.395, 0.0544, 0.01, -0.3961, 0.0538, 0.01, -0.3951, 0.0532, 0.0061, -0.394, 0.0532, 0.0083, -0.394, 0.0526, 0.01, -0.3963, 0.0505, 0.01, -0.3963, 0.0531, 0.01, -0.3939, 0.0505, 0.01, -0.3941, 0.0483, 0.01, -0.3963, 0.0461, 0.01, -0.3963, 0.0483, 0.01, -0.3941, 0.0461, 0.01, -0.3941, 0.0521, 0.0061, -0.3921, 0.0514, 0.0061, -0.3911, 0.0521, 0.0083, -0.3921, 0.0514, 0.0083, -0.3911, 0.0506, 0.0061, -0.3897, 0.0506, 0.0083, -0.3897, 0.0521, 0.01, -0.3923, 0.0515, 0.01, -0.3912, 0.0504, 0.01, -0.3895, 0.0483, 0.01, -0.3919, 0.0461, 0.01, -0.3919, 0.0483, 0.01, -0.3897,
0.0461, 0.01, -0.3897, 0.0439, 0.01, -0.3963, 0.0418, 0.01, -0.3963, 0.0439, 0.01, -0.3941, 0.0418, 0.01, -0.3941, 0.0396, 0.01, -0.3963, 0.0374, 0.01, -0.3963, 0.0396, 0.01, -0.3941, 0.0374, 0.01, -0.3941, 0.0439, 0.01, -0.3919, 0.0418, 0.01, -0.3919, 0.0439, 0.01, -0.3897, 0.0418, 0.01, -0.3897, 0.0396, 0.01, -0.3919, 0.0374, 0.01, -0.3919, 0.0396, 0.01, -0.3897, 0.0374, 0.01, -0.3897, 0.0496, 0.0061, -0.3883, 0.0496, 0.0083, -0.3883, 0.0489, 0.0061, -0.3872, 0.0489, 0.0083, -0.3872, 0.0478, 0.0061, -0.3856, 0.0471, 0.0061, -0.3845, 0.0478, 0.0083, -0.3856, 0.0471, 0.0083, -0.3845, 0.0497, 0.01, -0.3883, 0.049, 0.01, -0.3872, 0.0461, 0.01, -0.3876, 0.0479, 0.01, -0.3857, 0.0471, 0.01, -0.3846,
0.0461, 0.0061, -0.3832, 0.0461, 0.0083, -0.3832, 0.0451, 0.0061, -0.382, 0.0451, 0.0083, -0.382, 0.0459, 0.01, -0.383, 0.0451, 0.01, -0.382, 0.0439, 0.01, -0.3876, 0.0418, 0.01, -0.3876, 0.0439, 0.01, -0.3854, 0.0418, 0.01, -0.3854, 0.0396, 0.01, -0.3876, 0.0374, 0.01, -0.3876, 0.0396, 0.01, -0.3854, 0.0374, 0.01, -0.3854, 0.0443, 0.0061, -0.3809, 0.0443, 0.0083, -0.3809, 0.0439, 0.01, -0.3832, 0.0418, 0.01, -0.3832, 0.0443, 0.01, -0.3809, 0.0418, 0.01, -0.3811, 0.0396, 0.01, -0.3832, 0.0374, 0.01, -0.3832, 0.0396, 0.01, -0.3811, 0.0374, 0.01, -0.3811, 0.0309, 0.0061, -0.4136, 0.0309, 0.0083, -0.4136, 0.0301, 0.0061, -0.4121, 0.0295, 0.0061, -0.411, 0.0301, 0.0083, -0.4121, 0.0295, 0.0083, -0.411,
0.0308, 0.01, -0.4134, 0.0288, 0.01, -0.4136, 0.0301, 0.01, -0.4122, 0.0295, 0.01, -0.4111, 0.036, 0.0061, -0.4066, 0.036, 0.0083, -0.4066, 0.0284, 0.0061, -0.4093, 0.0284, 0.0083, -0.4093, 0.0277, 0.0061, -0.4081, 0.0277, 0.0083, -0.4081, 0.036, 0.01, -0.4065, 0.0283, 0.01, -0.4091, 0.0277, 0.01, -0.4081, 0.0266, 0.01, -0.4136, 0.0244, 0.01, -0.4136, 0.0266, 0.01, -0.4115, 0.0244, 0.01, -0.4115, 0.0222, 0.01, -0.4136, 0.0201, 0.01, -0.4136, 0.0222, 0.01, -0.4115, 0.0201, 0.01, -0.4115, 0.0271, 0.0061, -0.407, 0.0271, 0.0083, -0.407, 0.0266, 0.01, -0.4093, 0.0244, 0.01, -0.4093, 0.0271, 0.01, -0.407, 0.0244, 0.01, -0.4071, 0.0222, 0.01, -0.4093, 0.0201, 0.01, -0.4093, 0.0222, 0.01, -0.4071,
0.0201, 0.01, -0.4071, 0.0351, 0.0061, -0.4049, 0.0351, 0.0083, -0.4049, 0.0343, 0.0061, -0.4037, 0.0336, 0.0061, -0.4026, 0.0343, 0.0083, -0.4037, 0.0336, 0.0083, -0.4026, 0.0352, 0.01, -0.4052, 0.0342, 0.01, -0.4035, 0.0335, 0.01, -0.4025, 0.0325, 0.0061, -0.4009, 0.0325, 0.0083, -0.4009, 0.0319, 0.0061, -0.3998, 0.0319, 0.0083, -0.3998, 0.031, 0.0061, -0.3984, 0.031, 0.0083, -0.3984, 0.0353, 0.01, -0.4006, 0.0325, 0.01, -0.4008, 0.0353, 0.01, -0.3984, 0.0331, 0.01, -0.3984, 0.0318, 0.01, -0.3997, 0.0312, 0.01, -0.3987, 0.026, 0.0061, -0.4053, 0.0253, 0.0061, -0.4042, 0.026, 0.0083, -0.4053, 0.0253, 0.0083, -0.4042, 0.0244, 0.0061, -0.4028, 0.0244, 0.0083, -0.4028, 0.0261, 0.01, -0.4054, 0.0254, 0.01, -0.4043,
0.0242, 0.01, -0.4025, 0.0222, 0.01, -0.4049, 0.0201, 0.01, -0.4049, 0.0222, 0.01, -0.4028, 0.0201, 0.01, -0.4028, 0.0235, 0.0061, -0.4014, 0.0235, 0.0083, -0.4014, 0.0228, 0.0061, -0.4003, 0.0228, 0.0083, -0.4003, 0.0217, 0.0061, -0.3987, 0.021, 0.0061, -0.3976, 0.0217, 0.0083, -0.3987, 0.021, 0.0083, -0.3976, 0.0235, 0.01, -0.4015, 0.0228, 0.01, -0.4004, 0.0201, 0.01, -0.4006, 0.0218, 0.01, -0.3988, 0.0211, 0.01, -0.3978, 0.0104, 0.0061, -0.4145, 0.0104, 0.0083, -0.4145, 0.0179, 0.01, -0.4136, 0.0157, 0.01, -0.4136, 0.0179, 0.01, -0.4115, 0.0157, 0.01, -0.4115, 0.0136, 0.01, -0.4136, 0.0103, 0.01, -0.4143, 0.0136, 0.01, -0.4115, 0.0114, 0.01, -0.4115, 0.0179, 0.01, -0.4093, 0.0157, 0.01, -0.4093,
0.0179, 0.01, -0.4071, 0.0157, 0.01, -0.4071, 0.0136, 0.01, -0.4093, 0.0114, 0.01, -0.4093, 0.0136, 0.01, -0.4071, 0.0114, 0.01, -0.4071, 0.0098, 0.0061, -0.4134, 0.0098, 0.0083, -0.4134, 0.0087, 0.0061, -0.4115, 0.0087, 0.0083, -0.4115, 0.0019, 0.0061, -0.4141, 0.0019, 0.0083, -0.4141, 0.0097, 0.01, -0.4133, 0.0089, 0.01, -0.4118, 0.002, 0.01, -0.4142, 0.0081, 0.0061, -0.4103, 0.0076, 0.0061, -0.4093, 0.0081, 0.0083, -0.4103, 0.0076, 0.0083, -0.4093, 0.0065, 0.0061, -0.4073, 0.0065, 0.0083, -0.4073, 0.0058, 0.0061, -0.4062, 0.0058, 0.0083, -0.4062, 0.008, 0.01, -0.41, 0.0075, 0.01, -0.4091, 0.0092, 0.01, -0.4071, 0.0065, 0.01, -0.4073, 0.0058, 0.01, -0.4062, 0.0179, 0.01, -0.4049, 0.0157, 0.01, -0.4049,
0.0179, 0.01, -0.4028, 0.0157, 0.01, -0.4028, 0.0136, 0.01, -0.4049, 0.0114, 0.01, -0.4049, 0.0136, 0.01, -0.4028, 0.0114, 0.01, -0.4028, 0.0179, 0.01, -0.4006, 0.0157, 0.01, -0.4006, 0.0179, 0.01, -0.3984, 0.0157, 0.01, -0.3984, 0.0136, 0.01, -0.4006, 0.0114, 0.01, -0.4006, 0.0136, 0.01, -0.3984, 0.0114, 0.01, -0.3984, 0.0051, 0.0061, -0.4049, 0.0051, 0.0083, -0.4049, 0.0041, 0.0061, -0.4032, 0.0035, 0.0061, -0.4022, 0.0041, 0.0083, -0.4032, 0.0035, 0.0083, -0.4022, 0.0092, 0.01, -0.4049, 0.0071, 0.01, -0.4049, 0.0092, 0.01, -0.4028, 0.0071, 0.01, -0.4028, 0.0052, 0.01, -0.4051, 0.0041, 0.01, -0.4032, 0.0035, 0.01, -0.4021, 0.0026, 0.0061, -0.4006, 0.0026, 0.0083, -0.4006, 0.0018, 0.0061, -0.3992,
0.0018, 0.0083, -0.3992, 0.0092, 0.01, -0.4006, 0.0071, 0.01, -0.4006, 0.0092, 0.01, -0.3984, 0.0071, 0.01, -0.3984, 0.0049, 0.01, -0.4006, 0.0028, 0.01, -0.4009, 0.0049, 0.01, -0.3984, 0.0017, 0.01, -0.399, 0.0301, 0.0061, -0.397, 0.0294, 0.0061, -0.3959, 0.0301, 0.0083, -0.397, 0.0294, 0.0083, -0.3959, 0.0283, 0.0061, -0.3941, 0.0283, 0.0083, -0.3941, 0.0353, 0.01, -0.3963, 0.0331, 0.01, -0.3963, 0.0353, 0.01, -0.3941, 0.0331, 0.01, -0.3941, 0.03, 0.01, -0.3968, 0.0293, 0.01, -0.3958, 0.0309, 0.01, -0.3941, 0.0284, 0.01, -0.3942, 0.0353, 0.01, -0.3919, 0.0331, 0.01, -0.3919, 0.0353, 0.01, -0.3897, 0.0331, 0.01, -0.3897, 0.0309, 0.01, -0.3919, 0.0288, 0.01, -0.3919, 0.0309, 0.01, -0.3897,
0.0288, 0.01, -0.3897, 0.0277, 0.0061, -0.393, 0.0277, 0.0083, -0.393, 0.0201, 0.0061, -0.3963, 0.0201, 0.0083, -0.3963, 0.0192, 0.0061, -0.3949, 0.0192, 0.0083, -0.3949, 0.0276, 0.01, -0.393, 0.0199, 0.01, -0.396, 0.0192, 0.01, -0.395, 0.0269, 0.0061, -0.3919, 0.0269, 0.0083, -0.3919, 0.0259, 0.0061, -0.3903, 0.0251, 0.0061, -0.3892, 0.0259, 0.0083, -0.3903, 0.0251, 0.0083, -0.3892, 0.027, 0.01, -0.3921, 0.0258, 0.01, -0.3902, 0.0251, 0.01, -0.3891, 0.0353, 0.01, -0.3876, 0.0331, 0.01, -0.3876, 0.0353, 0.01, -0.3854, 0.0331, 0.01, -0.3854, 0.0309, 0.01, -0.3876, 0.0288, 0.01, -0.3876, 0.0309, 0.01, -0.3854, 0.0288, 0.01, -0.3854, 0.0353, 0.01, -0.3832, 0.0331, 0.01, -0.3832, 0.0353, 0.01, -0.3811,
0.0331, 0.01, -0.3811, 0.0309, 0.01, -0.3832, 0.0288, 0.01, -0.3832, 0.0309, 0.01, -0.3811, 0.0288, 0.01, -0.3811, 0.0241, 0.0061, -0.3876, 0.0241, 0.0083, -0.3876, 0.0233, 0.0061, -0.3865, 0.0233, 0.0083, -0.3865, 0.0226, 0.0061, -0.3854, 0.0226, 0.0083, -0.3854, 0.0266, 0.01, -0.3876, 0.0243, 0.01, -0.3879, 0.0266, 0.01, -0.3854, 0.0232, 0.01, -0.3863, 0.0225, 0.01, -0.3853, 0.0215, 0.0061, -0.3838, 0.0207, 0.0061, -0.3827, 0.0215, 0.0083, -0.3838, 0.0207, 0.0083, -0.3827, 0.0197, 0.0061, -0.3812, 0.0197, 0.0083, -0.3812, 0.0266, 0.01, -0.3832, 0.0244, 0.01, -0.3832, 0.0266, 0.01, -0.3811, 0.0244, 0.01, -0.3811, 0.0215, 0.01, -0.3837, 0.0207, 0.01, -0.3827, 0.0222, 0.01, -0.3811, 0.0197, 0.01, -0.3812,
0.0184, 0.0061, -0.3938, 0.0184, 0.0083, -0.3938, 0.0179, 0.01, -0.3963, 0.0157, 0.01, -0.3963, 0.0184, 0.01, -0.3938, 0.0157, 0.01, -0.3941, 0.0136, 0.01, -0.3963, 0.0114, 0.01, -0.3963, 0.0136, 0.01, -0.3941, 0.0114, 0.01, -0.3941, 0.0174, 0.0061, -0.3922, 0.0166, 0.0061, -0.3911, 0.0174, 0.0083, -0.3922, 0.0166, 0.0083, -0.3911, 0.0157, 0.0061, -0.3897, 0.0157, 0.0083, -0.3897, 0.0174, 0.01, -0.3923, 0.0167, 0.01, -0.3913, 0.0156, 0.01, -0.3895, 0.0136, 0.01, -0.3919, 0.0114, 0.01, -0.3919, 0.0136, 0.01, -0.3897, 0.0114, 0.01, -0.3897, 0.0092, 0.01, -0.3963, 0.0071, 0.01, -0.3963, 0.0092, 0.01, -0.3941, 0.0071, 0.01, -0.3941, 0.0049, 0.01, -0.3963, 0.0027, 0.01, -0.3963, 0.0049, 0.01, -0.3941,
0.0027, 0.01, -0.3941, 0.0092, 0.01, -0.3919, 0.0071, 0.01, -0.3919, 0.0092, 0.01, -0.3897, 0.0071, 0.01, -0.3897, 0.0049, 0.01, -0.3919, 0.0027, 0.01, -0.3919, 0.0049, 0.01, -0.3897, 0.0027, 0.01, -0.3897, 0.0148, 0.0061, -0.3884, 0.0148, 0.0083, -0.3884, 0.0141, 0.0061, -0.3873, 0.0141, 0.0083, -0.3873, 0.013, 0.0061, -0.3858, 0.0122, 0.0061, -0.3847, 0.013, 0.0083, -0.3858, 0.0122, 0.0083, -0.3847, 0.0148, 0.01, -0.3884, 0.0141, 0.01, -0.3873, 0.0114, 0.01, -0.3876, 0.013, 0.01, -0.3859, 0.0123, 0.01, -0.3848, 0.0189, 0.0061, -0.3801, 0.0189, 0.0083, -0.3801, 0.0112, 0.0061, -0.3832, 0.0112, 0.0083, -0.3832, 0.0104, 0.0061, -0.3821, 0.0104, 0.0083, -0.3821, 0.0189, 0.01, -0.3801, 0.011, 0.01, -0.383,
0.0104, 0.01, -0.3821, 0.0092, 0.01, -0.3876, 0.0071, 0.01, -0.3876, 0.0092, 0.01, -0.3854, 0.0071, 0.01, -0.3854, 0.0049, 0.01, -0.3876, 0.0027, 0.01, -0.3876, 0.0049, 0.01, -0.3854, 0.0027, 0.01, -0.3854, 0.0096, 0.0061, -0.381, 0.0096, 0.0083, -0.381, 0.0092, 0.01, -0.3832, 0.0071, 0.01, -0.3832, 0.0096, 0.01, -0.381, 0.0071, 0.01, -0.3811, 0.0049, 0.01, -0.3832, 0.0027, 0.01, -0.3832, 0.0049, 0.01, -0.3811, 0.0027, 0.01, -0.3811, 0.0432, 0.0061, -0.3796, 0.0423, 0.0061, -0.3785, 0.0432, 0.0083, -0.3796, 0.0423, 0.0083, -0.3785, 0.0412, 0.0061, -0.3771, 0.0412, 0.0083, -0.3771, 0.0404, 0.0061, -0.376, 0.0404, 0.0083, -0.376, 0.0432, 0.01, -0.3796, 0.0423, 0.01, -0.3785, 0.0413, 0.01, -0.3772,
0.0396, 0.01, -0.3789, 0.0374, 0.01, -0.3789, 0.0404, 0.01, -0.3761, 0.0374, 0.01, -0.3767, 0.0393, 0.0061, -0.3746, 0.0385, 0.0061, -0.3736, 0.0393, 0.0083, -0.3746, 0.0385, 0.0083, -0.3736, 0.0376, 0.0061, -0.3724, 0.0376, 0.0083, -0.3724, 0.0394, 0.01, -0.3747, 0.0385, 0.01, -0.3737, 0.0373, 0.01, -0.3721, 0.0365, 0.0061, -0.3711, 0.0365, 0.0083, -0.3711, 0.0365, 0.01, -0.3711, 0.0353, 0.01, -0.3789, 0.0331, 0.01, -0.3789, 0.0353, 0.01, -0.3767, 0.0331, 0.01, -0.3767, 0.0309, 0.01, -0.3789, 0.0288, 0.01, -0.3789, 0.0309, 0.01, -0.3767, 0.0288, 0.01, -0.3767, 0.0353, 0.01, -0.3746, 0.0331, 0.01, -0.3746, 0.0353, 0.01, -0.3724, 0.0331, 0.01, -0.3724, 0.0309, 0.01, -0.3746, 0.0288, 0.01, -0.3746,
0.0309, 0.01, -0.3724, 0.0288, 0.01, -0.3724, 0.0266, 0.01, -0.3789, 0.0244, 0.01, -0.3789, 0.0266, 0.01, -0.3767, 0.0244, 0.01, -0.3767, 0.0222, 0.01, -0.3789, 0.0201, 0.01, -0.3789, 0.0222, 0.01, -0.3767, 0.0201, 0.01, -0.3767, 0.0266, 0.01, -0.3746, 0.0244, 0.01, -0.3746, 0.0266, 0.01, -0.3724, 0.0244, 0.01, -0.3724, 0.0222, 0.01, -0.3746, 0.0201, 0.01, -0.3746, 0.0222, 0.01, -0.3724, 0.0201, 0.01, -0.3724, 0.0357, 0.0061, -0.37, 0.0357, 0.0083, -0.37, 0.0346, 0.0061, -0.3686, 0.0337, 0.0061, -0.3675, 0.0346, 0.0083, -0.3686, 0.0337, 0.0083, -0.3675, 0.0357, 0.01, -0.37, 0.0331, 0.01, -0.3702, 0.0346, 0.01, -0.3686, 0.0338, 0.01, -0.3676, 0.0309, 0.01, -0.3702, 0.0288, 0.01, -0.3702,
0.0309, 0.01, -0.368, 0.0288, 0.01, -0.368, 0.0327, 0.0061, -0.3662, 0.0327, 0.0083, -0.3662, 0.0318, 0.0061, -0.3651, 0.0318, 0.0083, -0.3651, 0.0307, 0.0061, -0.3637, 0.0307, 0.0083, -0.3637, 0.0327, 0.01, -0.3662, 0.0318, 0.01, -0.3651, 0.0288, 0.01, -0.3659, 0.0305, 0.01, -0.3635, 0.0288, 0.01, -0.3637, 0.0266, 0.01, -0.3702, 0.0244, 0.01, -0.3702, 0.0266, 0.01, -0.368, 0.0244, 0.01, -0.368, 0.0222, 0.01, -0.3702, 0.0201, 0.01, -0.3702, 0.0222, 0.01, -0.368, 0.0201, 0.01, -0.368, 0.0266, 0.01, -0.3659, 0.0244, 0.01, -0.3659, 0.0266, 0.01, -0.3637, 0.0244, 0.01, -0.3637, 0.0222, 0.01, -0.3659, 0.0201, 0.01, -0.3659, 0.0222, 0.01, -0.3637, 0.0201, 0.01, -0.3637, 0.018, 0.0061, -0.3789,
0.018, 0.0083, -0.3789, 0.017, 0.0061, -0.3775, 0.0163, 0.0061, -0.3764, 0.017, 0.0083, -0.3775, 0.0163, 0.0083, -0.3764, 0.0182, 0.01, -0.3791, 0.017, 0.01, -0.3774, 0.0162, 0.01, -0.3763, 0.0152, 0.0061, -0.3749, 0.0152, 0.0083, -0.3749, 0.0144, 0.0061, -0.3738, 0.0144, 0.0083, -0.3738, 0.0134, 0.0061, -0.3724, 0.0134, 0.0083, -0.3724, 0.0179, 0.01, -0.3746, 0.0152, 0.01, -0.3749, 0.0179, 0.01, -0.3724, 0.0157, 0.01, -0.3724, 0.0144, 0.01, -0.3738, 0.0136, 0.01, -0.3727, 0.0085, 0.0061, -0.3794, 0.0078, 0.0061, -0.3783, 0.0085, 0.0083, -0.3794, 0.0078, 0.0083, -0.3783, 0.0067, 0.0061, -0.3768, 0.0067, 0.0083, -0.3768, 0.0059, 0.0061, -0.3757, 0.0059, 0.0083, -0.3757, 0.0086, 0.01, -0.3795, 0.0078, 0.01, -0.3784,
0.0068, 0.01, -0.3769, 0.0049, 0.01, -0.3789, 0.0027, 0.01, -0.3789, 0.006, 0.01, -0.3759, 0.0027, 0.01, -0.3767, 0.0051, 0.0061, -0.3746, 0.0051, 0.0083, -0.3746, 0.0041, 0.0061, -0.3731, 0.0033, 0.0061, -0.372, 0.0041, 0.0083, -0.3731, 0.0033, 0.0083, -0.372, 0.0049, 0.01, -0.3742, 0.0027, 0.01, -0.3746, 0.0041, 0.01, -0.3731, 0.0033, 0.01, -0.372, 0.0125, 0.0061, -0.3712, 0.0118, 0.0061, -0.3701, 0.0125, 0.0083, -0.3712, 0.0118, 0.0083, -0.3701, 0.0107, 0.0061, -0.3686, 0.0107, 0.0083, -0.3686, 0.0179, 0.01, -0.3702, 0.0157, 0.01, -0.3702, 0.0179, 0.01, -0.368, 0.0157, 0.01, -0.368, 0.0124, 0.01, -0.3711, 0.0117, 0.01, -0.37, 0.0136, 0.01, -0.368, 0.0106, 0.01, -0.3685, 0.0179, 0.01, -0.3659,
0.0157, 0.01, -0.3659, 0.0179, 0.01, -0.3637, 0.0157, 0.01, -0.3637, 0.0136, 0.01, -0.3659, 0.0114, 0.01, -0.3659, 0.0136, 0.01, -0.3637, 0.0114, 0.01, -0.3637, 0.0099, 0.0061, -0.3675, 0.0099, 0.0083, -0.3675, 0.0022, 0.0061, -0.3704, 0.0022, 0.0083, -0.3704, 0.0099, 0.01, -0.3675, 0.0023, 0.01, -0.3706, 0.0088, 0.0061, -0.366, 0.0081, 0.0061, -0.3649, 0.0088, 0.0083, -0.366, 0.0081, 0.0083, -0.3649, 0.0072, 0.0061, -0.3637, 0.0072, 0.0083, -0.3637, 0.0088, 0.01, -0.366, 0.008, 0.01, -0.3649, 0.0092, 0.01, -0.3637, 0.0074, 0.01, -0.3639, 0.0298, 0.0061, -0.3626, 0.029, 0.0061, -0.3615, 0.0298, 0.0083, -0.3626, 0.029, 0.0083, -0.3615, 0.0279, 0.0061, -0.3601, 0.0279, 0.0083, -0.3601, 0.0298, 0.01, -0.3626,
0.029, 0.01, -0.3615, 0.0279, 0.01, -0.3602, 0.0271, 0.0061, -0.3591, 0.0271, 0.0083, -0.3591, 0.0266, 0.01, -0.3615, 0.0244, 0.01, -0.3615, 0.0271, 0.01, -0.3591, 0.0244, 0.01, -0.3594, 0.0222, 0.01, -0.3615, 0.0201, 0.01, -0.3615, 0.0222, 0.01, -0.3594, 0.0201, 0.01, -0.3594, 0.026, 0.0061, -0.3577, 0.0251, 0.0061, -0.3566, 0.026, 0.0083, -0.3577, 0.0251, 0.0083, -0.3566, 0.024, 0.0061, -0.3553, 0.024, 0.0083, -0.3553, 0.0231, 0.0061, -0.3542, 0.0231, 0.0083, -0.3542, 0.026, 0.01, -0.3577, 0.0252, 0.01, -0.3566, 0.0241, 0.01, -0.3553, 0.0222, 0.01, -0.3572, 0.0201, 0.01, -0.3572, 0.0231, 0.01, -0.3543, 0.0201, 0.01, -0.355, 0.022, 0.0061, -0.3529, 0.0211, 0.0061, -0.3519, 0.022, 0.0083, -0.3529,
0.0211, 0.0083, -0.3519, 0.0201, 0.0061, -0.3507, 0.0201, 0.0083, -0.3507, 0.0221, 0.01, -0.353, 0.0212, 0.01, -0.3519, 0.0199, 0.01, -0.3504, 0.0191, 0.0061, -0.3494, 0.0191, 0.0083, -0.3494, 0.0191, 0.01, -0.3495, 0.0179, 0.01, -0.3615, 0.0157, 0.01, -0.3615, 0.0179, 0.01, -0.3594, 0.0157, 0.01, -0.3594, 0.0136, 0.01, -0.3615, 0.0114, 0.01, -0.3615, 0.0136, 0.01, -0.3594, 0.0114, 0.01, -0.3594, 0.0179, 0.01, -0.3572, 0.0157, 0.01, -0.3572, 0.0179, 0.01, -0.355, 0.0157, 0.01, -0.355, 0.0136, 0.01, -0.3572, 0.0114, 0.01, -0.3572, 0.0136, 0.01, -0.355, 0.0114, 0.01, -0.355, 0.0062, 0.0061, -0.3623, 0.0062, 0.0083, -0.3623, 0.0055, 0.0061, -0.3612, 0.0055, 0.0083, -0.3612, 0.0044, 0.0061, -0.3596,
0.0036, 0.0061, -0.3585, 0.0044, 0.0083, -0.3596, 0.0036, 0.0083, -0.3585, 0.0092, 0.01, -0.3615, 0.0061, 0.01, -0.3622, 0.0092, 0.01, -0.3594, 0.0071, 0.01, -0.3594, 0.0054, 0.01, -0.3611, 0.0044, 0.01, -0.3596, 0.0036, 0.01, -0.3585, 0.0027, 0.0061, -0.3572, 0.0027, 0.0083, -0.3572, 0.0018, 0.0061, -0.3559, 0.0018, 0.0083, -0.3559, 0.0092, 0.01, -0.3572, 0.0071, 0.01, -0.3572, 0.0092, 0.01, -0.355, 0.0071, 0.01, -0.355, 0.0049, 0.01, -0.3572, 0.0029, 0.01, -0.3574, 0.0049, 0.01, -0.355, 0.0017, 0.01, -0.3557, 0.0179, 0.01, -0.3529, 0.0157, 0.01, -0.3529, 0.0179, 0.01, -0.3507, 0.0157, 0.01, -0.3507, 0.0136, 0.01, -0.3529, 0.0114, 0.01, -0.3529, 0.0136, 0.01, -0.3507, 0.0114, 0.01, -0.3507,
0.0182, 0.0061, -0.3484, 0.0182, 0.0083, -0.3484, 0.0172, 0.0061, -0.347, 0.0163, 0.0061, -0.3459, 0.0172, 0.0083, -0.347, 0.0163, 0.0083, -0.3459, 0.0182, 0.01, -0.3483, 0.0157, 0.01, -0.3485, 0.0172, 0.01, -0.347, 0.0163, 0.01, -0.3459, 0.0136, 0.01, -0.3485, 0.0114, 0.01, -0.3485, 0.0136, 0.01, -0.3463, 0.0114, 0.01, -0.3463, 0.0092, 0.01, -0.3529, 0.0071, 0.01, -0.3529, 0.0092, 0.01, -0.3507, 0.0071, 0.01, -0.3507, 0.0049, 0.01, -0.3529, 0.0027, 0.01, -0.3529, 0.0049, 0.01, -0.3507, 0.0027, 0.01, -0.3507, 0.0092, 0.01, -0.3485, 0.0071, 0.01, -0.3485, 0.0092, 0.01, -0.3463, 0.0071, 0.01, -0.3463, 0.0049, 0.01, -0.3485, 0.0027, 0.01, -0.3485, 0.0049, 0.01, -0.3463, 0.0027, 0.01, -0.3463,
0.1018, 0.0061, -0.3444, 0.1018, 0.0083, -0.3444, 0.1014, 0.0061, -0.3433, 0.1014, 0.0083, -0.3433, 0.1008, 0.0061, -0.342, 0.1008, 0.0083, -0.342, 0.1019, 0.01, -0.3446, 0.1015, 0.01, -0.3437, 0.0982, 0.01, -0.3442, 0.1007, 0.01, -0.3417, 0.0982, 0.01, -0.342, 0.1, 0.0061, -0.3398, 0.1, 0.0083, -0.3398, 0.0994, 0.0061, -0.3385, 0.099, 0.0061, -0.3374, 0.0994, 0.0083, -0.3385, 0.099, 0.0083, -0.3374, 0.0999, 0.01, -0.3397, 0.0982, 0.01, -0.3398, 0.0994, 0.01, -0.3385, 0.099, 0.01, -0.3375, 0.096, 0.01, -0.3442, 0.0939, 0.01, -0.3442, 0.096, 0.01, -0.342, 0.0939, 0.01, -0.342, 0.0917, 0.01, -0.3442, 0.0895, 0.01, -0.3442, 0.0917, 0.01, -0.342, 0.0895, 0.01, -0.342, 0.096, 0.01, -0.3398,
0.0939, 0.01, -0.3398, 0.096, 0.01, -0.3377, 0.0939, 0.01, -0.3377, 0.0917, 0.01, -0.3398, 0.0895, 0.01, -0.3398, 0.0917, 0.01, -0.3377, 0.0895, 0.01, -0.3377, 0.0981, 0.0061, -0.3355, 0.0981, 0.0083, -0.3355, 0.0974, 0.0061, -0.3339, 0.0974, 0.0083, -0.3339, 0.0981, 0.01, -0.3353, 0.0974, 0.01, -0.334, 0.0968, 0.0061, -0.3328, 0.0968, 0.0083, -0.3328, 0.096, 0.01, -0.3355, 0.0939, 0.01, -0.3355, 0.0969, 0.01, -0.3329, 0.0939, 0.01, -0.3333, 0.0917, 0.01, -0.3355, 0.0895, 0.01, -0.3355, 0.0917, 0.01, -0.3333, 0.0895, 0.01, -0.3333, 0.096, 0.0061, -0.3311, 0.096, 0.0083, -0.3311, 0.0952, 0.0061, -0.3297, 0.0945, 0.0061, -0.3286, 0.0952, 0.0083, -0.3297, 0.0945, 0.0083, -0.3286, 0.0958, 0.01, -0.3309,
0.0939, 0.01, -0.3312, 0.0952, 0.01, -0.3298, 0.0946, 0.01, -0.3287, 0.0917, 0.01, -0.3312, 0.0895, 0.01, -0.3312, 0.0917, 0.01, -0.329, 0.0895, 0.01, -0.329, 0.0879, 0.0061, -0.3442, 0.0879, 0.0083, -0.3442, 0.0869, 0.0061, -0.342, 0.0869, 0.0083, -0.342, 0.0879, 0.01, -0.3443, 0.087, 0.01, -0.3423, 0.0864, 0.0061, -0.3407, 0.0859, 0.0061, -0.3396, 0.0864, 0.0083, -0.3407, 0.0859, 0.0083, -0.3396, 0.085, 0.0061, -0.3377, 0.085, 0.0083, -0.3377, 0.0862, 0.01, -0.3404, 0.0858, 0.01, -0.3394, 0.0874, 0.01, -0.3377, 0.0851, 0.01, -0.3379, 0.0843, 0.0061, -0.3362, 0.0843, 0.0083, -0.3362, 0.0838, 0.0061, -0.3351, 0.0838, 0.0083, -0.3351, 0.0829, 0.0061, -0.3333, 0.0829, 0.0083, -0.3333, 0.0874, 0.01, -0.3355,
0.0842, 0.01, -0.336, 0.0874, 0.01, -0.3333, 0.0852, 0.01, -0.3333, 0.0837, 0.01, -0.3349, 0.083, 0.01, -0.3336, 0.0821, 0.0061, -0.3319, 0.0815, 0.0061, -0.3308, 0.0821, 0.0083, -0.3319, 0.0815, 0.0083, -0.3308, 0.0804, 0.0061, -0.329, 0.0804, 0.0083, -0.329, 0.0874, 0.01, -0.3312, 0.0852, 0.01, -0.3312, 0.0874, 0.01, -0.329, 0.0852, 0.01, -0.329, 0.082, 0.01, -0.3317, 0.0814, 0.01, -0.3307, 0.083, 0.01, -0.329, 0.0804, 0.01, -0.3291, 0.0797, 0.0061, -0.328, 0.0797, 0.0083, -0.328, 0.0797, 0.01, -0.328, 0.0935, 0.0061, -0.3268, 0.0935, 0.0083, -0.3268, 0.0928, 0.0061, -0.3257, 0.0928, 0.0083, -0.3257, 0.0921, 0.0061, -0.3246, 0.0921, 0.0083, -0.3246, 0.0934, 0.01, -0.3267, 0.0928, 0.01, -0.3257,
0.0917, 0.01, -0.3268, 0.0895, 0.01, -0.3268, 0.0921, 0.01, -0.3246, 0.0895, 0.01, -0.3246, 0.091, 0.0061, -0.3229, 0.0903, 0.0061, -0.3219, 0.091, 0.0083, -0.3229, 0.0903, 0.0083, -0.3219, 0.0892, 0.0061, -0.3203, 0.0892, 0.0083, -0.3203, 0.0911, 0.01, -0.323, 0.0903, 0.01, -0.3219, 0.0893, 0.01, -0.3205, 0.0874, 0.01, -0.3268, 0.0852, 0.01, -0.3268, 0.0874, 0.01, -0.3246, 0.0852, 0.01, -0.3246, 0.083, 0.01, -0.3268, 0.0808, 0.01, -0.3268, 0.083, 0.01, -0.3246, 0.0808, 0.01, -0.3246, 0.0884, 0.0061, -0.3193, 0.0884, 0.0083, -0.3193, 0.0874, 0.01, -0.3225, 0.0852, 0.01, -0.3225, 0.0885, 0.01, -0.3194, 0.0852, 0.01, -0.3203, 0.083, 0.01, -0.3225, 0.0808, 0.01, -0.3225, 0.083, 0.01, -0.3203,
0.0808, 0.01, -0.3203, 0.079, 0.0061, -0.3268, 0.079, 0.0083, -0.3268, 0.0779, 0.0061, -0.3253, 0.0772, 0.0061, -0.3242, 0.0779, 0.0083, -0.3253, 0.0772, 0.0083, -0.3242, 0.0791, 0.01, -0.327, 0.0779, 0.01, -0.3252, 0.0771, 0.01, -0.3241, 0.0761, 0.0061, -0.3226, 0.0761, 0.0083, -0.3226, 0.0753, 0.0061, -0.3216, 0.0753, 0.0083, -0.3216, 0.0743, 0.0061, -0.3203, 0.0743, 0.0083, -0.3203, 0.0787, 0.01, -0.3225, 0.0761, 0.01, -0.3226, 0.0787, 0.01, -0.3203, 0.0765, 0.01, -0.3203, 0.0753, 0.01, -0.3215, 0.0745, 0.01, -0.3206, 0.0875, 0.0061, -0.3181, 0.0875, 0.0083, -0.3181, 0.0865, 0.0061, -0.3168, 0.0856, 0.0061, -0.3157, 0.0865, 0.0083, -0.3168, 0.0856, 0.0083, -0.3157, 0.0873, 0.01, -0.3178, 0.0852, 0.01, -0.3181,
0.0865, 0.01, -0.3168, 0.0856, 0.01, -0.3157, 0.083, 0.01, -0.3181, 0.0808, 0.01, -0.3181, 0.083, 0.01, -0.316, 0.0808, 0.01, -0.316, 0.0845, 0.0061, -0.3143, 0.0845, 0.0083, -0.3143, 0.0836, 0.0061, -0.3133, 0.0836, 0.0083, -0.3133, 0.0826, 0.0061, -0.312, 0.0816, 0.0061, -0.3109, 0.0826, 0.0083, -0.312, 0.0816, 0.0083, -0.3109, 0.0846, 0.01, -0.3144, 0.0837, 0.01, -0.3133, 0.0808, 0.01, -0.3138, 0.0826, 0.01, -0.312, 0.0816, 0.01, -0.3109, 0.0733, 0.0061, -0.3191, 0.0724, 0.0061, -0.318, 0.0733, 0.0083, -0.3191, 0.0724, 0.0083, -0.318, 0.0713, 0.0061, -0.3167, 0.0713, 0.0083, -0.3167, 0.0787, 0.01, -0.3181, 0.0765, 0.01, -0.3181, 0.0787, 0.01, -0.316, 0.0765, 0.01, -0.316, 0.0733, 0.01, -0.319,
0.0724, 0.01, -0.318, 0.0743, 0.01, -0.316, 0.0713, 0.01, -0.3167, 0.0787, 0.01, -0.3138, 0.0765, 0.01, -0.3138, 0.0787, 0.01, -0.3116, 0.0765, 0.01, -0.3116, 0.0743, 0.01, -0.3138, 0.0722, 0.01, -0.3138, 0.0743, 0.01, -0.3116, 0.0722, 0.01, -0.3116, 0.0805, 0.0061, -0.3097, 0.0805, 0.0083, -0.3097, 0.0806, 0.01, -0.3097, 0.0795, 0.0061, -0.3086, 0.0795, 0.0083, -0.3086, 0.0784, 0.0061, -0.3075, 0.0773, 0.0061, -0.3065, 0.0784, 0.0083, -0.3075, 0.0773, 0.0083, -0.3065, 0.0795, 0.01, -0.3086, 0.0765, 0.01, -0.3095, 0.0784, 0.01, -0.3075, 0.0773, 0.01, -0.3065, 0.0743, 0.01, -0.3095, 0.0722, 0.01, -0.3095, 0.0743, 0.01, -0.3073, 0.0722, 0.01, -0.3073, 0.0762, 0.0061, -0.3054, 0.0762, 0.0083, -0.3054,
0.075, 0.0061, -0.3044, 0.075, 0.0083, -0.3044, 0.0739, 0.0061, -0.3034, 0.0726, 0.0061, -0.3024, 0.0739, 0.0083, -0.3034, 0.0726, 0.0083, -0.3024, 0.0762, 0.01, -0.3054, 0.075, 0.01, -0.3043, 0.0722, 0.01, -0.3051, 0.0739, 0.01, -0.3034, 0.0726, 0.01, -0.3023, 0.0715, 0.0061, -0.3015, 0.0715, 0.0083, -0.3015, 0.0715, 0.01, -0.3015, 0.0704, 0.0061, -0.3156, 0.0704, 0.0083, -0.3156, 0.0704, 0.01, -0.3156, 0.0693, 0.0061, -0.3144, 0.0683, 0.0061, -0.3133, 0.0693, 0.0083, -0.3144, 0.0683, 0.0083, -0.3133, 0.0672, 0.0061, -0.3122, 0.0672, 0.0083, -0.3122, 0.0661, 0.0061, -0.3111, 0.0661, 0.0083, -0.3111, 0.0693, 0.01, -0.3144, 0.0683, 0.01, -0.3133, 0.07, 0.01, -0.3116, 0.0673, 0.01, -0.3122, 0.0661, 0.01, -0.3112,
0.0152, 0.0061, -0.3446, 0.0152, 0.0083, -0.3446, 0.0143, 0.0061, -0.3435, 0.0143, 0.0083, -0.3435, 0.0132, 0.0061, -0.3421, 0.0124, 0.0061, -0.341, 0.0132, 0.0083, -0.3421, 0.0124, 0.0083, -0.341, 0.0152, 0.01, -0.3446, 0.0144, 0.01, -0.3435, 0.0114, 0.01, -0.3442, 0.0133, 0.01, -0.3422, 0.0125, 0.01, -0.3411, 0.0115, 0.0061, -0.3398, 0.0115, 0.0083, -0.3398, 0.0105, 0.0061, -0.3385, 0.0105, 0.0083, -0.3385, 0.0113, 0.01, -0.3396, 0.0105, 0.01, -0.3385, 0.0092, 0.01, -0.3442, 0.0071, 0.01, -0.3442, 0.0092, 0.01, -0.342, 0.0071, 0.01, -0.342, 0.0049, 0.01, -0.3442, 0.0027, 0.01, -0.3442, 0.0049, 0.01, -0.342, 0.0027, 0.01, -0.342, 0.0097, 0.0061, -0.3374, 0.0097, 0.0083, -0.3374, 0.0092, 0.01, -0.3398,
0.0071, 0.01, -0.3398, 0.0097, 0.01, -0.3374, 0.0071, 0.01, -0.3377, 0.0049, 0.01, -0.3398, 0.0027, 0.01, -0.3398, 0.0049, 0.01, -0.3377, 0.0027, 0.01, -0.3377, 0.0086, 0.0061, -0.336, 0.0078, 0.0061, -0.3349, 0.0086, 0.0083, -0.336, 0.0078, 0.0083, -0.3349, 0.0067, 0.0061, -0.3335, 0.0067, 0.0083, -0.3335, 0.0059, 0.0061, -0.3324, 0.0059, 0.0083, -0.3324, 0.0086, 0.01, -0.3361, 0.0078, 0.01, -0.335, 0.0067, 0.01, -0.3336, 0.0049, 0.01, -0.3355, 0.0027, 0.01, -0.3355, 0.0059, 0.01, -0.3325, 0.0027, 0.01, -0.3333, 0.0049, 0.0061, -0.3312, 0.0049, 0.0083, -0.3312, 0.004, 0.0061, -0.3298, 0.0032, 0.0061, -0.3287, 0.004, 0.0083, -0.3298, 0.0032, 0.0083, -0.3287, 0.0047, 0.01, -0.3309, 0.0027, 0.01, -0.3312,
0.004, 0.01, -0.3298, 0.0032, 0.01, -0.3287, 0.0021, 0.0061, -0.3272, 0.0021, 0.0083, -0.3272, 0.0022, 0.01, -0.3273, 0.065, 0.0061, -0.3102, 0.0636, 0.0061, -0.3091, 0.065, 0.0083, -0.3102, 0.0636, 0.0083, -0.3091, 0.0626, 0.0061, -0.3082, 0.0626, 0.0083, -0.3082, 0.07, 0.01, -0.3095, 0.0678, 0.01, -0.3095, 0.07, 0.01, -0.3073, 0.0678, 0.01, -0.3073, 0.065, 0.01, -0.3102, 0.0637, 0.01, -0.3092, 0.0656, 0.01, -0.3073, 0.0626, 0.01, -0.3083, 0.07, 0.01, -0.3051, 0.0678, 0.01, -0.3051, 0.07, 0.01, -0.3029, 0.0678, 0.01, -0.3029, 0.0656, 0.01, -0.3051, 0.0635, 0.01, -0.3051, 0.0656, 0.01, -0.3029, 0.0635, 0.01, -0.3029, 0.0613, 0.0061, -0.3072, 0.0601, 0.0061, -0.3063, 0.0613, 0.0083, -0.3072,
0.0601, 0.0083, -0.3063, 0.0611, 0.01, -0.307, 0.0601, 0.01, -0.3063, 0.059, 0.0061, -0.3055, 0.059, 0.0083, -0.3055, 0.0576, 0.0061, -0.3044, 0.0576, 0.0083, -0.3044, 0.0565, 0.0061, -0.3036, 0.055, 0.0061, -0.3025, 0.0565, 0.0083, -0.3036, 0.055, 0.0083, -0.3025, 0.0613, 0.01, -0.3051, 0.059, 0.01, -0.3054, 0.0613, 0.01, -0.3029, 0.0591, 0.01, -0.3029, 0.0576, 0.01, -0.3044, 0.0565, 0.01, -0.3036, 0.0551, 0.01, -0.3026, 0.0701, 0.0061, -0.3004, 0.0701, 0.0083, -0.3004, 0.069, 0.0061, -0.2996, 0.0678, 0.0061, -0.2987, 0.069, 0.0083, -0.2996, 0.0678, 0.0083, -0.2987, 0.0665, 0.0061, -0.2977, 0.0665, 0.0083, -0.2977, 0.0702, 0.01, -0.3004, 0.0678, 0.01, -0.3008, 0.069, 0.01, -0.2996, 0.0681, 0.01, -0.2988,
0.0656, 0.01, -0.3008, 0.0635, 0.01, -0.3008, 0.0663, 0.01, -0.2976, 0.0635, 0.01, -0.2986, 0.0654, 0.0061, -0.297, 0.0637, 0.0061, -0.2959, 0.0654, 0.0083, -0.297, 0.0637, 0.0083, -0.2959, 0.0627, 0.0061, -0.2952, 0.0627, 0.0083, -0.2952, 0.0653, 0.01, -0.2969, 0.0637, 0.01, -0.2959, 0.0626, 0.01, -0.2952, 0.0539, 0.0061, -0.3017, 0.0539, 0.0083, -0.3017, 0.0613, 0.01, -0.3008, 0.0591, 0.01, -0.3008, 0.0613, 0.01, -0.2986, 0.0591, 0.01, -0.2986, 0.057, 0.01, -0.3008, 0.054, 0.01, -0.3019, 0.057, 0.01, -0.2986, 0.0548, 0.01, -0.2986, 0.0613, 0.0061, -0.2943, 0.0599, 0.0061, -0.2934, 0.0613, 0.0083, -0.2943, 0.0599, 0.0083, -0.2934, 0.0613, 0.01, -0.2964, 0.0591, 0.01, -0.2964, 0.0615, 0.01, -0.2945,
0.0597, 0.01, -0.2933, 0.057, 0.01, -0.2964, 0.0548, 0.01, -0.2964, 0.057, 0.01, -0.2943, 0.0548, 0.01, -0.2943, 0.0526, 0.0061, -0.3009, 0.0511, 0.0061, -0.3, 0.0526, 0.0083, -0.3009, 0.0511, 0.0083, -0.3, 0.05, 0.0061, -0.2993, 0.05, 0.0083, -0.2993, 0.0483, 0.0061, -0.2983, 0.047, 0.0061, -0.2976, 0.0483, 0.0083, -0.2983, 0.047, 0.0083, -0.2976, 0.0524, 0.01, -0.3007, 0.0512, 0.01, -0.3, 0.0526, 0.01, -0.2986, 0.0501, 0.01, -0.2993, 0.0481, 0.01, -0.2982, 0.047, 0.01, -0.2976, 0.0459, 0.0061, -0.2971, 0.0459, 0.0083, -0.2971, 0.0526, 0.01, -0.2964, 0.0505, 0.01, -0.2964, 0.0526, 0.01, -0.2943, 0.0505, 0.01, -0.2943, 0.0483, 0.01, -0.2964, 0.0459, 0.01, -0.2971, 0.0483, 0.01, -0.2943,
0.0461, 0.01, -0.2943, 0.044, 0.0061, -0.2962, 0.0425, 0.0061, -0.2955, 0.044, 0.0083, -0.2962, 0.0425, 0.0083, -0.2955, 0.0414, 0.0061, -0.2951, 0.0414, 0.0083, -0.2951, 0.0396, 0.0061, -0.2943, 0.0375, 0.0061, -0.2935, 0.0396, 0.0083, -0.2943, 0.0375, 0.0083, -0.2935, 0.0438, 0.01, -0.2961, 0.0426, 0.01, -0.2956, 0.0439, 0.01, -0.2943, 0.0415, 0.01, -0.2951, 0.0394, 0.01, -0.2943, 0.0378, 0.01, -0.2936, 0.0588, 0.0061, -0.2927, 0.0588, 0.0083, -0.2927, 0.057, 0.0061, -0.2916, 0.057, 0.0083, -0.2916, 0.0559, 0.0061, -0.291, 0.0548, 0.0061, -0.2904, 0.0559, 0.0083, -0.291, 0.0548, 0.0083, -0.2904, 0.0587, 0.01, -0.2927, 0.057, 0.01, -0.2917, 0.0548, 0.01, -0.2921, 0.0559, 0.01, -0.291, 0.0549, 0.01, -0.2904,
0.0527, 0.0061, -0.2893, 0.0527, 0.0083, -0.2893, 0.0526, 0.01, -0.2921, 0.0505, 0.01, -0.2921, 0.0527, 0.01, -0.2893, 0.0505, 0.01, -0.2899, 0.0483, 0.01, -0.2921, 0.0461, 0.01, -0.2921, 0.0483, 0.01, -0.2899, 0.0461, 0.01, -0.2899, 0.0516, 0.0061, -0.2888, 0.0505, 0.0061, -0.2883, 0.0516, 0.0083, -0.2888, 0.0505, 0.0083, -0.2883, 0.0483, 0.0061, -0.2874, 0.0469, 0.0061, -0.2868, 0.0483, 0.0083, -0.2874, 0.0469, 0.0083, -0.2868, 0.0458, 0.0061, -0.2863, 0.0458, 0.0083, -0.2863, 0.0516, 0.01, -0.2888, 0.0506, 0.01, -0.2884, 0.0486, 0.01, -0.2875, 0.0466, 0.01, -0.2867, 0.0457, 0.01, -0.2862, 0.0364, 0.0061, -0.2931, 0.0364, 0.0083, -0.2931, 0.0439, 0.01, -0.2921, 0.0418, 0.01, -0.2921, 0.0439, 0.01, -0.2899,
0.0418, 0.01, -0.2899, 0.0396, 0.01, -0.2921, 0.0369, 0.01, -0.2933, 0.0396, 0.01, -0.2899, 0.0374, 0.01, -0.2899, 0.044, 0.0061, -0.2855, 0.0419, 0.0061, -0.2848, 0.044, 0.0083, -0.2855, 0.0419, 0.0083, -0.2848, 0.0439, 0.01, -0.2878, 0.0418, 0.01, -0.2878, 0.0441, 0.01, -0.2856, 0.0419, 0.01, -0.2848, 0.0396, 0.01, -0.2878, 0.0374, 0.01, -0.2878, 0.0396, 0.01, -0.2856, 0.0374, 0.01, -0.2856, 0.0408, 0.0061, -0.2845, 0.0408, 0.0083, -0.2845, 0.0396, 0.0061, -0.2841, 0.0374, 0.0061, -0.2834, 0.0396, 0.0083, -0.2841, 0.0374, 0.0083, -0.2834, 0.0408, 0.01, -0.2844, 0.0397, 0.01, -0.2841, 0.0376, 0.01, -0.2835, 0.0353, 0.0061, -0.2927, 0.0331, 0.0061, -0.292, 0.0353, 0.0083, -0.2927, 0.0331, 0.0083, -0.292,
0.0309, 0.0061, -0.2914, 0.0297, 0.0061, -0.2911, 0.0309, 0.0083, -0.2914, 0.0297, 0.0083, -0.2911, 0.0286, 0.0061, -0.2908, 0.0286, 0.0083, -0.2908, 0.035, 0.01, -0.2926, 0.033, 0.01, -0.292, 0.0353, 0.01, -0.2899, 0.0331, 0.01, -0.2899, 0.0309, 0.01, -0.2914, 0.0297, 0.01, -0.2911, 0.0309, 0.01, -0.2899, 0.0287, 0.01, -0.2908, 0.0353, 0.01, -0.2878, 0.0331, 0.01, -0.2878, 0.0353, 0.01, -0.2856, 0.0331, 0.01, -0.2856, 0.0309, 0.01, -0.2878, 0.0288, 0.01, -0.2878, 0.0309, 0.01, -0.2856, 0.0288, 0.01, -0.2856, 0.0266, 0.0061, -0.2904, 0.0244, 0.0061, -0.2899, 0.0266, 0.0083, -0.2904, 0.0244, 0.0083, -0.2899, 0.0222, 0.0061, -0.2894, 0.0204, 0.0061, -0.289, 0.0222, 0.0083, -0.2894, 0.0204, 0.0083, -0.289,
0.0264, 0.01, -0.2903, 0.0243, 0.01, -0.2899, 0.0222, 0.01, -0.2894, 0.0206, 0.01, -0.2891, 0.0193, 0.0061, -0.2888, 0.0193, 0.0083, -0.2888, 0.0266, 0.01, -0.2878, 0.0244, 0.01, -0.2878, 0.0266, 0.01, -0.2856, 0.0244, 0.01, -0.2856, 0.0222, 0.01, -0.2878, 0.0198, 0.01, -0.2889, 0.0222, 0.01, -0.2856, 0.0201, 0.01, -0.2856, 0.0353, 0.0061, -0.2828, 0.0338, 0.0061, -0.2824, 0.0353, 0.0083, -0.2828, 0.0338, 0.0083, -0.2824, 0.0328, 0.0061, -0.2821, 0.0328, 0.0083, -0.2821, 0.0309, 0.0061, -0.2816, 0.0288, 0.0061, -0.2812, 0.0309, 0.0083, -0.2816, 0.0288, 0.0083, -0.2812, 0.0355, 0.01, -0.2829, 0.0334, 0.01, -0.2823, 0.0325, 0.01, -0.2821, 0.0309, 0.01, -0.2834, 0.0288, 0.01, -0.2834, 0.031, 0.01, -0.2817,
0.0289, 0.01, -0.2812, 0.0266, 0.0061, -0.2808, 0.0244, 0.0061, -0.2804, 0.0266, 0.0083, -0.2808, 0.0244, 0.0083, -0.2804, 0.0231, 0.0061, -0.2802, 0.0231, 0.0083, -0.2802, 0.0266, 0.01, -0.2834, 0.0244, 0.01, -0.2834, 0.0267, 0.01, -0.2808, 0.0246, 0.01, -0.2804, 0.0222, 0.01, -0.2834, 0.0201, 0.01, -0.2834, 0.0225, 0.01, -0.2801, 0.0201, 0.01, -0.2812, 0.022, 0.0061, -0.28, 0.0201, 0.0061, -0.2797, 0.022, 0.0083, -0.28, 0.0201, 0.0083, -0.2797, 0.0218, 0.01, -0.28, 0.0201, 0.01, -0.2797, 0.0179, 0.0061, -0.2885, 0.0157, 0.0061, -0.2883, 0.0179, 0.0083, -0.2885, 0.0157, 0.0083, -0.2883, 0.0136, 0.0061, -0.2882, 0.0114, 0.0061, -0.2882, 0.0136, 0.0083, -0.2882, 0.0114, 0.0083, -0.2882, 0.0178, 0.01, -0.2885,
0.0157, 0.01, -0.2883, 0.0179, 0.01, -0.2856, 0.0157, 0.01, -0.2856, 0.0136, 0.01, -0.2882, 0.0114, 0.01, -0.2882, 0.0136, 0.01, -0.2856, 0.0114, 0.01, -0.2856, 0.0092, 0.0061, -0.2882, 0.0071, 0.0061, -0.2882, 0.0092, 0.0083, -0.2882, 0.0071, 0.0083, -0.2882, 0.0049, 0.0061, -0.2882, 0.0027, 0.0061, -0.2882, 0.0049, 0.0083, -0.2882, 0.0027, 0.0083, -0.2882, 0.0092, 0.01, -0.2882, 0.0071, 0.01, -0.2882, 0.0092, 0.01, -0.2856, 0.0071, 0.01, -0.2856, 0.0049, 0.01, -0.2882, 0.0027, 0.01, -0.2882, 0.0049, 0.01, -0.2856, 0.0027, 0.01, -0.2856, 0.0179, 0.01, -0.2834, 0.0157, 0.01, -0.2834, 0.0179, 0.01, -0.2812, 0.0157, 0.01, -0.2812, 0.0136, 0.01, -0.2834, 0.0114, 0.01, -0.2834, 0.0136, 0.01, -0.2812,
0.0114, 0.01, -0.2812, 0.0179, 0.0061, -0.2795, 0.0157, 0.0061, -0.2794, 0.0179, 0.0083, -0.2795, 0.0157, 0.0083, -0.2794, 0.0136, 0.0061, -0.2794, 0.0114, 0.0061, -0.2794, 0.0136, 0.0083, -0.2794, 0.0114, 0.0083, -0.2794, 0.0179, 0.01, -0.2795, 0.0157, 0.01, -0.2794, 0.0136, 0.01, -0.2794, 0.0114, 0.01, -0.2794, 0.0092, 0.01, -0.2834, 0.0071, 0.01, -0.2834, 0.0092, 0.01, -0.2812, 0.0071, 0.01, -0.2812, 0.0049, 0.01, -0.2834, 0.0027, 0.01, -0.2834, 0.0049, 0.01, -0.2812, 0.0027, 0.01, -0.2812, 0.0092, 0.0061, -0.2794, 0.0071, 0.0061, -0.2794, 0.0092, 0.0083, -0.2794, 0.0071, 0.0083, -0.2794, 0.0049, 0.0061, -0.2794, 0.0027, 0.0061, -0.2794, 0.0049, 0.0083, -0.2794, 0.0027, 0.0083, -0.2794, 0.0092, 0.01, -0.2794,
0.0071, 0.01, -0.2794, 0.0049, 0.01, -0.2794, 0.0027, 0.01, -0.2794, 0.067, 0.0061, -0.2196, 0.067, 0.0083, -0.2196, 0.0666, 0.0061, -0.2195, 0.0666, 0.0083, -0.2195, 0.067, 0.01, -0.2196, 0.0666, 0.01, -0.2195, 0.067, 0.0061, -0.2183, 0.067, 0.0083, -0.2183, 0.0667, 0.0061, -0.2172, 0.0667, 0.0083, -0.2172, 0.0655, 0.0061, -0.219, 0.0635, 0.0061, -0.2181, 0.0655, 0.0083, -0.219, 0.0635, 0.0083, -0.2181, 0.0665, 0.0061, -0.216, 0.0665, 0.0083, -0.216, 0.067, 0.01, -0.2183, 0.0667, 0.01, -0.2172, 0.0655, 0.01, -0.219, 0.0633, 0.01, -0.218, 0.0665, 0.01, -0.216, 0.0635, 0.01, -0.2161, 0.0621, 0.0061, -0.2174, 0.0621, 0.0083, -0.2174, 0.061, 0.0061, -0.2168, 0.0592, 0.0061, -0.2157, 0.061, 0.0083, -0.2168,
0.0592, 0.0083, -0.2157, 0.0622, 0.01, -0.2174, 0.0611, 0.01, -0.2168, 0.0593, 0.01, -0.2158, 0.0652, 0.0061, -0.2141, 0.0642, 0.0061, -0.2132, 0.0652, 0.0083, -0.2141, 0.0642, 0.0083, -0.2132, 0.0631, 0.0061, -0.2124, 0.0631, 0.0083, -0.2124, 0.0654, 0.01, -0.2142, 0.0641, 0.01, -0.2131, 0.063, 0.01, -0.2124, 0.0581, 0.0061, -0.215, 0.0581, 0.0083, -0.215, 0.0614, 0.0061, -0.2113, 0.0614, 0.0083, -0.2113, 0.057, 0.0061, -0.2144, 0.0552, 0.0061, -0.2133, 0.057, 0.0083, -0.2144, 0.0552, 0.0083, -0.2133, 0.0541, 0.0061, -0.2126, 0.0541, 0.0083, -0.2126, 0.0613, 0.01, -0.214, 0.0583, 0.01, -0.2152, 0.0614, 0.01, -0.2114, 0.0591, 0.01, -0.2118, 0.0566, 0.01, -0.2142, 0.0553, 0.01, -0.2134, 0.057, 0.01, -0.2118,
0.0542, 0.01, -0.2127, 0.0603, 0.0061, -0.2107, 0.0591, 0.0061, -0.21, 0.0603, 0.0083, -0.2107, 0.0591, 0.0083, -0.21, 0.0574, 0.0061, -0.2089, 0.0574, 0.0083, -0.2089, 0.0563, 0.0061, -0.2083, 0.0548, 0.0061, -0.2074, 0.0563, 0.0083, -0.2083, 0.0548, 0.0083, -0.2074, 0.0603, 0.01, -0.2107, 0.0593, 0.01, -0.2101, 0.0573, 0.01, -0.2089, 0.0548, 0.01, -0.2096, 0.0562, 0.01, -0.2083, 0.0551, 0.01, -0.2075, 0.0526, 0.0061, -0.2117, 0.0513, 0.0061, -0.2109, 0.0526, 0.0083, -0.2117, 0.0513, 0.0083, -0.2109, 0.0524, 0.01, -0.2115, 0.0513, 0.01, -0.2109, 0.0502, 0.0061, -0.2102, 0.0502, 0.0083, -0.2102, 0.0534, 0.0061, -0.2066, 0.0534, 0.0083, -0.2066, 0.0484, 0.0061, -0.2091, 0.0484, 0.0083, -0.2091, 0.0473, 0.0061, -0.2085,
0.0461, 0.0061, -0.2076, 0.0473, 0.0083, -0.2085, 0.0461, 0.0083, -0.2076, 0.0526, 0.01, -0.2096, 0.0502, 0.01, -0.2102, 0.0532, 0.01, -0.2065, 0.0505, 0.01, -0.2075, 0.0486, 0.01, -0.2092, 0.0475, 0.01, -0.2086, 0.0458, 0.01, -0.2074, 0.0448, 0.0061, -0.2066, 0.0448, 0.0083, -0.2066, 0.0448, 0.01, -0.2066, 0.0219, 0.0061, -0.2196, 0.0201, 0.0061, -0.2196, 0.0219, 0.0083, -0.2196, 0.0201, 0.0083, -0.2196, 0.0219, 0.01, -0.2196, 0.0201, 0.01, -0.2196, 0.0219, 0.0061, -0.2183, 0.0219, 0.0083, -0.2183, 0.0219, 0.0061, -0.2161, 0.0219, 0.0083, -0.2161, 0.0219, 0.01, -0.2183, 0.0201, 0.01, -0.2183, 0.0219, 0.01, -0.2161, 0.0201, 0.01, -0.2161, 0.0219, 0.0061, -0.214, 0.0219, 0.0083, -0.214, 0.0219, 0.0061, -0.2118,
0.0219, 0.0083, -0.2118, 0.0219, 0.01, -0.214, 0.0201, 0.01, -0.214, 0.0219, 0.01, -0.2118, 0.0201, 0.01, -0.2118, 0.0219, 0.0061, -0.2088, 0.0201, 0.0061, -0.2088, 0.0219, 0.0083, -0.2088, 0.0201, 0.0083, -0.2088, 0.0219, 0.01, -0.2088, 0.0201, 0.01, -0.2088, 0.0179, 0.0061, -0.2196, 0.0157, 0.0061, -0.2196, 0.0179, 0.0083, -0.2196, 0.0157, 0.0083, -0.2196, 0.0136, 0.0061, -0.2196, 0.0114, 0.0061, -0.2196, 0.0136, 0.0083, -0.2196, 0.0114, 0.0083, -0.2196, 0.0179, 0.01, -0.2196, 0.0157, 0.01, -0.2196, 0.0136, 0.01, -0.2196, 0.0114, 0.01, -0.2196, 0.0179, 0.01, -0.2183, 0.0157, 0.01, -0.2183, 0.0179, 0.01, -0.2161, 0.0157, 0.01, -0.2161, 0.0136, 0.01, -0.2183, 0.0114, 0.01, -0.2183, 0.0136, 0.01, -0.2161,
0.0114, 0.01, -0.2161, 0.0092, 0.0061, -0.2196, 0.0071, 0.0061, -0.2196, 0.0092, 0.0083, -0.2196, 0.0071, 0.0083, -0.2196, 0.0049, 0.0061, -0.2196, 0.0027, 0.0061, -0.2196, 0.0049, 0.0083, -0.2196, 0.0027, 0.0083, -0.2196, 0.0092, 0.01, -0.2196, 0.0071, 0.01, -0.2196, 0.0049, 0.01, -0.2196, 0.0027, 0.01, -0.2196, 0.0092, 0.01, -0.2183, 0.0071, 0.01, -0.2183, 0.0092, 0.01, -0.2161, 0.0071, 0.01, -0.2161, 0.0049, 0.01, -0.2183, 0.0027, 0.01, -0.2183, 0.0049, 0.01, -0.2161, 0.0027, 0.01, -0.2161, 0.0179, 0.01, -0.214, 0.0157, 0.01, -0.214, 0.0179, 0.01, -0.2118, 0.0157, 0.01, -0.2118, 0.0136, 0.01, -0.214, 0.0114, 0.01, -0.214, 0.0136, 0.01, -0.2118, 0.0114, 0.01, -0.2118, 0.0179, 0.0061, -0.2088,
0.0157, 0.0061, -0.2088, 0.0179, 0.0083, -0.2088, 0.0157, 0.0083, -0.2088, 0.0136, 0.0061, -0.2088, 0.0114, 0.0061, -0.2088, 0.0136, 0.0083, -0.2088, 0.0114, 0.0083, -0.2088, 0.0179, 0.01, -0.2088, 0.0157, 0.01, -0.2088, 0.0136, 0.01, -0.2088, 0.0114, 0.01, -0.2088, 0.0092, 0.01, -0.214, 0.0071, 0.01, -0.214, 0.0092, 0.01, -0.2118, 0.0071, 0.01, -0.2118, 0.0049, 0.01, -0.214, 0.0027, 0.01, -0.214, 0.0049, 0.01, -0.2118, 0.0027, 0.01, -0.2118, 0.0092, 0.0061, -0.2088, 0.0071, 0.0061, -0.2088, 0.0092, 0.0083, -0.2088, 0.0071, 0.0083, -0.2088, 0.0049, 0.0061, -0.2088, 0.0027, 0.0061, -0.2088, 0.0049, 0.0083, -0.2088, 0.0027, 0.0083, -0.2088, 0.0092, 0.01, -0.2088, 0.0071, 0.01, -0.2088, 0.0049, 0.01, -0.2088,
0.0027, 0.01, -0.2088, 0.0591, 0.0061, -0.2069, 0.0591, 0.0083, -0.2069, 0.0591, 0.0061, -0.2031, 0.0591, 0.0083, -0.2031, 0.0573, 0.0061, -0.206, 0.0573, 0.0083, -0.206, 0.0563, 0.0061, -0.2032, 0.0563, 0.0083, -0.2032, 0.0591, 0.01, -0.2069, 0.0591, 0.01, -0.2031, 0.0573, 0.01, -0.206, 0.0563, 0.01, -0.2033, 0.0591, 0.0061, -0.2009, 0.0591, 0.0083, -0.2009, 0.0591, 0.0061, -0.1988, 0.0591, 0.0083, -0.1988, 0.0562, 0.0061, -0.2009, 0.054, 0.0061, -0.2, 0.0562, 0.0083, -0.2009, 0.054, 0.0083, -0.2, 0.0562, 0.0061, -0.199, 0.0551, 0.0061, -0.1995, 0.0562, 0.0083, -0.199, 0.0551, 0.0083, -0.1995, 0.0591, 0.01, -0.2009, 0.0591, 0.01, -0.1988, 0.0562, 0.01, -0.201, 0.054, 0.01, -0.2001, 0.0562, 0.01, -0.199,
0.055, 0.01, -0.1995, 0.0635, 0.0061, -0.1941, 0.0635, 0.0083, -0.1941, 0.0634, 0.01, -0.1942, 0.0668, 0.0061, -0.1894, 0.0668, 0.0083, -0.1894, 0.0656, 0.0061, -0.1924, 0.0645, 0.0061, -0.1933, 0.0656, 0.0083, -0.1924, 0.0645, 0.0083, -0.1933, 0.0666, 0.0061, -0.1905, 0.0628, 0.0061, -0.1892, 0.0666, 0.0083, -0.1905, 0.0628, 0.0083, -0.1892, 0.0668, 0.01, -0.1893, 0.0656, 0.01, -0.1923, 0.0645, 0.01, -0.1934, 0.0667, 0.01, -0.1904, 0.063, 0.01, -0.1891, 0.0605, 0.0061, -0.1956, 0.0591, 0.0061, -0.1961, 0.0605, 0.0083, -0.1956, 0.0591, 0.0083, -0.1961, 0.0616, 0.0061, -0.1952, 0.0616, 0.0083, -0.1952, 0.0546, 0.0061, -0.1939, 0.0546, 0.0083, -0.1939, 0.0605, 0.01, -0.1957, 0.0591, 0.01, -0.1961, 0.0616, 0.01, -0.1952,
0.0591, 0.01, -0.1944, 0.057, 0.01, -0.1966, 0.0548, 0.01, -0.1966, 0.057, 0.01, -0.1944, 0.0546, 0.01, -0.1939, 0.0587, 0.0061, -0.1916, 0.0587, 0.0083, -0.1916, 0.0613, 0.0061, -0.1901, 0.0598, 0.0061, -0.1909, 0.0613, 0.0083, -0.1901, 0.0598, 0.0083, -0.1909, 0.057, 0.0061, -0.1926, 0.0557, 0.0061, -0.1933, 0.057, 0.0083, -0.1926, 0.0557, 0.0083, -0.1933, 0.0613, 0.01, -0.1923, 0.0588, 0.01, -0.1915, 0.0611, 0.01, -0.1902, 0.0599, 0.01, -0.1909, 0.0568, 0.01, -0.1927, 0.0558, 0.01, -0.1933, 0.0523, 0.0061, -0.2059, 0.0505, 0.0061, -0.2049, 0.0523, 0.0083, -0.2059, 0.0505, 0.0083, -0.2049, 0.0501, 0.0061, -0.2026, 0.0501, 0.0083, -0.2026, 0.0492, 0.0061, -0.2043, 0.0492, 0.0083, -0.2043, 0.0484, 0.0061, -0.2039,
0.0484, 0.0083, -0.2039, 0.0522, 0.01, -0.2059, 0.0508, 0.01, -0.205, 0.05, 0.01, -0.2026, 0.0489, 0.01, -0.2041, 0.0461, 0.01, -0.2053, 0.0484, 0.01, -0.2039, 0.0461, 0.01, -0.2031, 0.0526, 0.0061, -0.2008, 0.0512, 0.0061, -0.2018, 0.0526, 0.0083, -0.2008, 0.0512, 0.0083, -0.2018, 0.0476, 0.0061, -0.1979, 0.046, 0.0061, -0.1989, 0.0476, 0.0083, -0.1979, 0.046, 0.0083, -0.1989, 0.0528, 0.01, -0.2007, 0.0511, 0.01, -0.2019, 0.0526, 0.01, -0.1988, 0.0505, 0.01, -0.1988, 0.0483, 0.01, -0.2009, 0.0461, 0.01, -0.2009, 0.0477, 0.01, -0.1978, 0.0458, 0.01, -0.199, 0.0436, 0.0061, -0.2052, 0.0436, 0.0083, -0.2052, 0.043, 0.0061, -0.2026, 0.043, 0.0083, -0.2026, 0.0436, 0.01, -0.2052, 0.043, 0.01, -0.2026,
0.0441, 0.0061, -0.2006, 0.0441, 0.0083, -0.2006, 0.045, 0.0061, -0.1998, 0.045, 0.0083, -0.1998, 0.0441, 0.01, -0.2006, 0.045, 0.01, -0.1998, 0.0517, 0.0061, -0.1956, 0.0505, 0.0061, -0.1963, 0.0517, 0.0083, -0.1956, 0.0505, 0.0083, -0.1963, 0.0528, 0.0061, -0.195, 0.0528, 0.0083, -0.195, 0.0487, 0.0061, -0.1973, 0.0487, 0.0083, -0.1973, 0.0519, 0.01, -0.1955, 0.0502, 0.01, -0.1965, 0.0529, 0.01, -0.1949, 0.0488, 0.01, -0.1972, 0.067, 0.0061, -0.1873, 0.067, 0.0083, -0.1873, 0.0656, 0.0061, -0.1878, 0.0639, 0.0061, -0.1886, 0.0656, 0.0083, -0.1878, 0.0639, 0.0083, -0.1886, 0.067, 0.01, -0.1873, 0.0654, 0.01, -0.1879, 0.064, 0.01, -0.1886, 0.067, 0.0061, -0.1686, 0.067, 0.0083, -0.1686, 0.0656, 0.0061, -0.1686,
0.0635, 0.0061, -0.1686, 0.0656, 0.0083, -0.1686, 0.0635, 0.0083, -0.1686, 0.067, 0.01, -0.1686, 0.0656, 0.01, -0.1686, 0.0635, 0.01, -0.1686, 0.067, 0.0061, -0.1662, 0.067, 0.0083, -0.1662, 0.067, 0.0061, -0.1641, 0.067, 0.0083, -0.1641, 0.0631, 0.0061, -0.1647, 0.0631, 0.0083, -0.1647, 0.067, 0.01, -0.1662, 0.067, 0.01, -0.1641, 0.0656, 0.01, -0.1662, 0.0635, 0.01, -0.1662, 0.0656, 0.01, -0.1641, 0.0631, 0.01, -0.1647, 0.0613, 0.0061, -0.1686, 0.0591, 0.0061, -0.1686, 0.0613, 0.0083, -0.1686, 0.0591, 0.0083, -0.1686, 0.057, 0.0061, -0.1686, 0.0548, 0.0061, -0.1686, 0.057, 0.0083, -0.1686, 0.0548, 0.0083, -0.1686, 0.0613, 0.01, -0.1686, 0.0591, 0.01, -0.1686, 0.057, 0.01, -0.1686, 0.0548, 0.01, -0.1686,
0.0613, 0.0061, -0.1647, 0.0591, 0.0061, -0.1647, 0.0613, 0.0083, -0.1647, 0.0591, 0.0083, -0.1647, 0.057, 0.0061, -0.1647, 0.0548, 0.0061, -0.1647, 0.057, 0.0083, -0.1647, 0.0548, 0.0083, -0.1647, 0.0613, 0.01, -0.1662, 0.0591, 0.01, -0.1662, 0.0613, 0.01, -0.1647, 0.0591, 0.01, -0.1647, 0.057, 0.01, -0.1662, 0.0548, 0.01, -0.1662, 0.057, 0.01, -0.1647, 0.0548, 0.01, -0.1647, 0.067, 0.0061, -0.1619, 0.067, 0.0083, -0.1619, 0.067, 0.0061, -0.1597, 0.067, 0.0083, -0.1597, 0.0631, 0.0061, -0.1619, 0.0631, 0.0083, -0.1619, 0.0631, 0.0061, -0.1597, 0.0631, 0.0083, -0.1597, 0.067, 0.01, -0.1619, 0.067, 0.01, -0.1597, 0.0656, 0.01, -0.1619, 0.0631, 0.01, -0.1619, 0.0656, 0.01, -0.1597, 0.0631, 0.01, -0.1597,
0.067, 0.0061, -0.1575, 0.067, 0.0083, -0.1575, 0.067, 0.0061, -0.1554, 0.067, 0.0083, -0.1554, 0.0631, 0.0061, -0.1575, 0.0631, 0.0083, -0.1575, 0.063, 0.0061, -0.1554, 0.063, 0.0083, -0.1554, 0.067, 0.01, -0.1575, 0.067, 0.01, -0.1554, 0.0656, 0.01, -0.1575, 0.0631, 0.01, -0.1576, 0.0656, 0.01, -0.1554, 0.063, 0.01, -0.1554, 0.0526, 0.0061, -0.1686, 0.0505, 0.0061, -0.1686, 0.0526, 0.0083, -0.1686, 0.0505, 0.0083, -0.1686, 0.0483, 0.0061, -0.1686, 0.0461, 0.0061, -0.1686, 0.0483, 0.0083, -0.1686, 0.0461, 0.0083, -0.1686, 0.0526, 0.01, -0.1686, 0.0505, 0.01, -0.1686, 0.0483, 0.01, -0.1686, 0.0461, 0.01, -0.1686, 0.0526, 0.0061, -0.1647, 0.0505, 0.0061, -0.1647, 0.0526, 0.0083, -0.1647, 0.0505, 0.0083, -0.1647,
0.0483, 0.0061, -0.1647, 0.0454, 0.0061, -0.1647, 0.0483, 0.0083, -0.1647, 0.0454, 0.0083, -0.1647, 0.0526, 0.01, -0.1662, 0.0505, 0.01, -0.1662, 0.0526, 0.01, -0.1647, 0.0505, 0.01, -0.1647, 0.0483, 0.01, -0.1662, 0.0461, 0.01, -0.1662, 0.0483, 0.01, -0.1647, 0.0454, 0.01, -0.1647, 0.0439, 0.0061, -0.1686, 0.0425, 0.0061, -0.1686, 0.0439, 0.0083, -0.1686, 0.0425, 0.0083, -0.1686, 0.0439, 0.01, -0.1686, 0.0425, 0.01, -0.1686, 0.0425, 0.0061, -0.1662, 0.0425, 0.0083, -0.1662, 0.0425, 0.0061, -0.1641, 0.0425, 0.0083, -0.1641, 0.0439, 0.01, -0.1662, 0.0425, 0.01, -0.1662, 0.0439, 0.01, -0.1641, 0.0425, 0.01, -0.1641, 0.0454, 0.0061, -0.1619, 0.0454, 0.0083, -0.1619, 0.0454, 0.0061, -0.1597, 0.0454, 0.0083, -0.1597,
0.0454, 0.01, -0.1619, 0.0454, 0.01, -0.1597, 0.0455, 0.0061, -0.1575, 0.0455, 0.0083, -0.1575, 0.0457, 0.0061, -0.1554, 0.0457, 0.0083, -0.1554, 0.0455, 0.01, -0.1576, 0.0456, 0.01, -0.1554, 0.0425, 0.0061, -0.1619, 0.0425, 0.0083, -0.1619, 0.0425, 0.0061, -0.1597, 0.0425, 0.0083, -0.1597, 0.0439, 0.01, -0.1619, 0.0425, 0.01, -0.1619, 0.0439, 0.01, -0.1597, 0.0425, 0.01, -0.1597, 0.0425, 0.0061, -0.1575, 0.0425, 0.0083, -0.1575, 0.0425, 0.0061, -0.1554, 0.0425, 0.0083, -0.1554, 0.0439, 0.01, -0.1575, 0.0425, 0.01, -0.1575, 0.0439, 0.01, -0.1554, 0.0425, 0.01, -0.1554, 0.0669, 0.0061, -0.1535, 0.0669, 0.0083, -0.1535, 0.0667, 0.0061, -0.1524, 0.0629, 0.0061, -0.1532, 0.0667, 0.0083, -0.1524, 0.0629, 0.0083, -0.1532,
0.0665, 0.0061, -0.151, 0.0625, 0.0061, -0.1516, 0.0665, 0.0083, -0.151, 0.0625, 0.0083, -0.1516, 0.0669, 0.01, -0.1537, 0.0668, 0.01, -0.1529, 0.0629, 0.01, -0.1533, 0.0664, 0.01, -0.1507, 0.0624, 0.01, -0.1513, 0.0656, 0.0061, -0.1488, 0.0656, 0.0083, -0.1488, 0.0648, 0.0061, -0.1475, 0.064, 0.0061, -0.1463, 0.0648, 0.0083, -0.1475, 0.064, 0.0083, -0.1463, 0.0655, 0.01, -0.1486, 0.0635, 0.01, -0.1489, 0.0648, 0.01, -0.1475, 0.064, 0.01, -0.1463, 0.0621, 0.0061, -0.1504, 0.0621, 0.0083, -0.1504, 0.0621, 0.01, -0.1504, 0.0611, 0.0061, -0.1489, 0.06, 0.0061, -0.148, 0.0611, 0.0083, -0.1489, 0.06, 0.0083, -0.148, 0.059, 0.0061, -0.1472, 0.059, 0.0083, -0.1472, 0.057, 0.0061, -0.1465, 0.0548, 0.0061, -0.1462,
0.057, 0.0083, -0.1465, 0.0548, 0.0083, -0.1462, 0.0612, 0.01, -0.149, 0.06, 0.01, -0.148, 0.0613, 0.01, -0.1467, 0.059, 0.01, -0.1472, 0.0569, 0.01, -0.1464, 0.0548, 0.01, -0.1462, 0.0628, 0.0061, -0.1452, 0.0628, 0.0083, -0.1452, 0.0628, 0.01, -0.1452, 0.0614, 0.0061, -0.144, 0.0614, 0.0083, -0.144, 0.0603, 0.0061, -0.1434, 0.0592, 0.0061, -0.1428, 0.0603, 0.0083, -0.1434, 0.0592, 0.0083, -0.1428, 0.057, 0.0061, -0.1423, 0.0548, 0.0061, -0.1422, 0.057, 0.0083, -0.1423, 0.0548, 0.0083, -0.1422, 0.0615, 0.01, -0.1441, 0.0591, 0.01, -0.1445, 0.0603, 0.01, -0.1434, 0.0593, 0.01, -0.1429, 0.057, 0.01, -0.1445, 0.0548, 0.01, -0.1445, 0.0571, 0.01, -0.1423, 0.0548, 0.01, -0.1422, 0.0461, 0.0061, -0.1532,
0.0461, 0.0083, -0.1532, 0.0474, 0.0061, -0.1503, 0.0468, 0.0061, -0.1514, 0.0474, 0.0083, -0.1503, 0.0468, 0.0083, -0.1514, 0.0461, 0.01, -0.1533, 0.0474, 0.01, -0.1502, 0.0469, 0.01, -0.1512, 0.0495, 0.0061, -0.1479, 0.0495, 0.0083, -0.1479, 0.0526, 0.0061, -0.1464, 0.0509, 0.0061, -0.1469, 0.0526, 0.0083, -0.1464, 0.0509, 0.0083, -0.1469, 0.0484, 0.0061, -0.1489, 0.0484, 0.0083, -0.1489, 0.0452, 0.0061, -0.1458, 0.0452, 0.0083, -0.1458, 0.0495, 0.01, -0.1479, 0.0527, 0.01, -0.1464, 0.0505, 0.01, -0.1471, 0.0484, 0.01, -0.1489, 0.0461, 0.01, -0.1489, 0.0483, 0.01, -0.1467, 0.0451, 0.01, -0.1458, 0.0425, 0.0061, -0.1532, 0.0425, 0.0083, -0.1532, 0.0429, 0.0061, -0.1501, 0.0426, 0.0061, -0.1511, 0.0429, 0.0083, -0.1501,
0.0426, 0.0083, -0.1511, 0.0439, 0.01, -0.1532, 0.0425, 0.01, -0.1532, 0.0428, 0.01, -0.1506, 0.0426, 0.01, -0.1515, 0.0432, 0.0061, -0.1488, 0.0432, 0.0083, -0.1488, 0.0443, 0.0061, -0.1468, 0.0443, 0.0083, -0.1468, 0.0433, 0.01, -0.1485, 0.0443, 0.01, -0.1469, 0.0526, 0.0061, -0.1422, 0.0504, 0.0061, -0.1425, 0.0526, 0.0083, -0.1422, 0.0504, 0.0083, -0.1425, 0.0477, 0.0061, -0.1437, 0.0462, 0.0061, -0.1447, 0.0477, 0.0083, -0.1437, 0.0462, 0.0083, -0.1447, 0.0487, 0.0061, -0.1431, 0.0487, 0.0083, -0.1431, 0.0526, 0.01, -0.1445, 0.0505, 0.01, -0.1445, 0.0526, 0.01, -0.1422, 0.0503, 0.01, -0.1425, 0.0478, 0.01, -0.1436, 0.0462, 0.01, -0.1447, 0.0489, 0.01, -0.143, 0.0219, 0.0061, -0.151, 0.0201, 0.0061, -0.151,
0.0219, 0.0083, -0.151, 0.0201, 0.0083, -0.151, 0.0219, 0.01, -0.151, 0.0201, 0.01, -0.151, 0.0219, 0.0061, -0.1489, 0.0219, 0.0083, -0.1489, 0.0219, 0.0061, -0.1467, 0.0219, 0.0083, -0.1467, 0.0219, 0.01, -0.1489, 0.0201, 0.01, -0.1489, 0.0219, 0.01, -0.1467, 0.0201, 0.01, -0.1467, 0.0219, 0.0061, -0.1445, 0.0219, 0.0083, -0.1445, 0.0219, 0.0061, -0.1424, 0.0219, 0.0083, -0.1424, 0.0219, 0.01, -0.1445, 0.0201, 0.01, -0.1445, 0.0219, 0.01, -0.1424, 0.0201, 0.01, -0.1424, 0.0219, 0.0061, -0.1402, 0.0219, 0.0083, -0.1402, 0.0219, 0.0061, -0.138, 0.0219, 0.0083, -0.138, 0.0219, 0.01, -0.1402, 0.0201, 0.01, -0.1402, 0.0219, 0.01, -0.138, 0.0201, 0.01, -0.138, 0.0179, 0.0061, -0.151, 0.0157, 0.0061, -0.151,
0.0179, 0.0083, -0.151, 0.0157, 0.0083, -0.151, 0.0136, 0.0061, -0.151, 0.0114, 0.0061, -0.151, 0.0136, 0.0083, -0.151, 0.0114, 0.0083, -0.151, 0.0179, 0.01, -0.151, 0.0157, 0.01, -0.151, 0.0136, 0.01, -0.151, 0.0114, 0.01, -0.151, 0.0179, 0.01, -0.1489, 0.0157, 0.01, -0.1489, 0.0179, 0.01, -0.1467, 0.0157, 0.01, -0.1467, 0.0136, 0.01, -0.1489, 0.0114, 0.01, -0.1489, 0.0136, 0.01, -0.1467, 0.0114, 0.01, -0.1467, 0.0092, 0.0061, -0.151, 0.0071, 0.0061, -0.151, 0.0092, 0.0083, -0.151, 0.0071, 0.0083, -0.151, 0.0049, 0.0061, -0.151, 0.0027, 0.0061, -0.151, 0.0049, 0.0083, -0.151, 0.0027, 0.0083, -0.151, 0.0092, 0.01, -0.151, 0.0071, 0.01, -0.151, 0.0049, 0.01, -0.151, 0.0027, 0.01, -0.151,
0.0092, 0.01, -0.1489, 0.0071, 0.01, -0.1489, 0.0092, 0.01, -0.1467, 0.0071, 0.01, -0.1467, 0.0049, 0.01, -0.1489, 0.0027, 0.01, -0.1489, 0.0049, 0.01, -0.1467, 0.0027, 0.01, -0.1467, 0.0179, 0.01, -0.1445, 0.0157, 0.01, -0.1445, 0.0179, 0.01, -0.1424, 0.0157, 0.01, -0.1424, 0.0136, 0.01, -0.1445, 0.0114, 0.01, -0.1445, 0.0136, 0.01, -0.1424, 0.0114, 0.01, -0.1424, 0.0179, 0.01, -0.1402, 0.0157, 0.01, -0.1402, 0.0179, 0.01, -0.138, 0.0157, 0.01, -0.138, 0.0136, 0.01, -0.1402, 0.0114, 0.01, -0.1402, 0.0136, 0.01, -0.138, 0.0114, 0.01, -0.138, 0.0092, 0.01, -0.1445, 0.0071, 0.01, -0.1445, 0.0092, 0.01, -0.1424, 0.0071, 0.01, -0.1424, 0.0049, 0.01, -0.1445, 0.0027, 0.01, -0.1445,
0.0049, 0.01, -0.1424, 0.0027, 0.01, -0.1424, 0.0086, 0.0061, -0.1376, 0.0072, 0.0061, -0.1387, 0.0086, 0.0083, -0.1376, 0.0072, 0.0083, -0.1387, 0.0043, 0.0061, -0.1391, 0.0027, 0.0061, -0.1392, 0.0043, 0.0083, -0.1391, 0.0027, 0.0083, -0.1392, 0.0054, 0.0061, -0.139, 0.0054, 0.0083, -0.139, 0.0092, 0.01, -0.1402, 0.0071, 0.01, -0.1402, 0.0087, 0.01, -0.1375, 0.0072, 0.01, -0.1387, 0.0048, 0.01, -0.1391, 0.0027, 0.01, -0.1392, 0.0056, 0.01, -0.139, 0.067, 0.0061, -0.1225, 0.067, 0.0083, -0.1225, 0.067, 0.0061, -0.1207, 0.067, 0.0083, -0.1207, 0.0656, 0.0061, -0.1225, 0.0635, 0.0061, -0.1225, 0.0656, 0.0083, -0.1225, 0.0635, 0.0083, -0.1225, 0.067, 0.01, -0.1225, 0.067, 0.01, -0.1207, 0.0656, 0.01, -0.1225,
0.0635, 0.01, -0.1225, 0.0656, 0.01, -0.1207, 0.0635, 0.01, -0.1207, 0.0613, 0.0061, -0.1225, 0.0591, 0.0061, -0.1225, 0.0613, 0.0083, -0.1225, 0.0591, 0.0083, -0.1225, 0.057, 0.0061, -0.1225, 0.0548, 0.0061, -0.1225, 0.057, 0.0083, -0.1225, 0.0548, 0.0083, -0.1225, 0.0613, 0.01, -0.1225, 0.0591, 0.01, -0.1225, 0.0613, 0.01, -0.1207, 0.0591, 0.01, -0.1207, 0.057, 0.01, -0.1225, 0.0548, 0.01, -0.1225, 0.057, 0.01, -0.1207, 0.0548, 0.01, -0.1207, 0.0526, 0.0061, -0.1225, 0.0505, 0.0061, -0.1225, 0.0526, 0.0083, -0.1225, 0.0505, 0.0083, -0.1225, 0.0483, 0.0061, -0.1225, 0.0461, 0.0061, -0.1225, 0.0483, 0.0083, -0.1225, 0.0461, 0.0083, -0.1225, 0.0526, 0.01, -0.1225, 0.0505, 0.01, -0.1225, 0.0526, 0.01, -0.1207,
0.0505, 0.01, -0.1207, 0.0483, 0.01, -0.1225, 0.0461, 0.01, -0.1225, 0.0483, 0.01, -0.1207, 0.0461, 0.01, -0.1207, 0.0439, 0.0061, -0.1225, 0.0425, 0.0061, -0.1225, 0.0439, 0.0083, -0.1225, 0.0425, 0.0083, -0.1225, 0.0425, 0.0061, -0.1207, 0.0425, 0.0083, -0.1207, 0.0439, 0.01, -0.1225, 0.0425, 0.01, -0.1225, 0.0439, 0.01, -0.1207, 0.0425, 0.01, -0.1207, 0.067, 0.0061, -0.1185, 0.067, 0.0083, -0.1185, 0.067, 0.0061, -0.1163, 0.067, 0.0083, -0.1163, 0.0631, 0.0061, -0.1186, 0.0631, 0.0083, -0.1186, 0.0631, 0.0061, -0.1163, 0.0631, 0.0083, -0.1163, 0.067, 0.01, -0.1185, 0.067, 0.01, -0.1163, 0.0656, 0.01, -0.1185, 0.0631, 0.01, -0.1186, 0.0656, 0.01, -0.1163, 0.0631, 0.01, -0.1163, 0.067, 0.0061, -0.1141,
0.067, 0.0083, -0.1141, 0.067, 0.0061, -0.112, 0.067, 0.0083, -0.112, 0.0631, 0.0061, -0.1141, 0.0631, 0.0083, -0.1141, 0.063, 0.0061, -0.112, 0.063, 0.0083, -0.112, 0.067, 0.01, -0.1141, 0.067, 0.01, -0.112, 0.0656, 0.01, -0.1141, 0.0631, 0.01, -0.1141, 0.0656, 0.01, -0.112, 0.063, 0.01, -0.112, 0.0613, 0.0061, -0.1186, 0.0591, 0.0061, -0.1186, 0.0613, 0.0083, -0.1186, 0.0591, 0.0083, -0.1186, 0.057, 0.0061, -0.1186, 0.0548, 0.0061, -0.1186, 0.057, 0.0083, -0.1186, 0.0548, 0.0083, -0.1186, 0.0613, 0.01, -0.1186, 0.0591, 0.01, -0.1186, 0.057, 0.01, -0.1186, 0.0548, 0.01, -0.1186, 0.067, 0.0061, -0.1098, 0.067, 0.0083, -0.1098, 0.0669, 0.0061, -0.1076, 0.0669, 0.0083, -0.1076, 0.063, 0.0061, -0.1098,
0.063, 0.0083, -0.1098, 0.0667, 0.0061, -0.1066, 0.0629, 0.0061, -0.1076, 0.0667, 0.0083, -0.1066, 0.0629, 0.0083, -0.1076, 0.067, 0.01, -0.1098, 0.0669, 0.01, -0.108, 0.0656, 0.01, -0.1098, 0.063, 0.01, -0.1098, 0.0669, 0.01, -0.1072, 0.0629, 0.01, -0.1078, 0.0666, 0.0061, -0.1054, 0.0625, 0.0061, -0.1061, 0.0666, 0.0083, -0.1054, 0.0625, 0.0083, -0.1061, 0.0658, 0.0061, -0.1033, 0.0658, 0.0083, -0.1033, 0.0665, 0.01, -0.1052, 0.0624, 0.01, -0.1058, 0.0658, 0.01, -0.1031, 0.0635, 0.01, -0.1033, 0.0622, 0.0061, -0.105, 0.0622, 0.0083, -0.105, 0.0613, 0.0061, -0.103, 0.0613, 0.0083, -0.103, 0.0622, 0.01, -0.1048, 0.0616, 0.01, -0.1034, 0.0526, 0.0061, -0.1186, 0.0505, 0.0061, -0.1186, 0.0526, 0.0083, -0.1186,
0.0505, 0.0083, -0.1186, 0.0483, 0.0061, -0.1186, 0.0454, 0.0061, -0.1186, 0.0483, 0.0083, -0.1186, 0.0454, 0.0083, -0.1186, 0.0454, 0.0061, -0.1163, 0.0454, 0.0083, -0.1163, 0.0526, 0.01, -0.1186, 0.0505, 0.01, -0.1186, 0.0483, 0.01, -0.1186, 0.0454, 0.01, -0.1186, 0.0454, 0.01, -0.1163, 0.0454, 0.0061, -0.1141, 0.0454, 0.0083, -0.1141, 0.0454, 0.0061, -0.112, 0.0454, 0.0083, -0.112, 0.0454, 0.01, -0.1141, 0.0454, 0.01, -0.112, 0.0425, 0.0061, -0.1185, 0.0425, 0.0083, -0.1185, 0.0425, 0.0061, -0.1163, 0.0425, 0.0083, -0.1163, 0.0439, 0.01, -0.1185, 0.0425, 0.01, -0.1185, 0.0439, 0.01, -0.1163, 0.0425, 0.01, -0.1163, 0.0425, 0.0061, -0.1141, 0.0425, 0.0083, -0.1141, 0.0425, 0.0061, -0.112, 0.0425, 0.0083, -0.112,
0.0439, 0.01, -0.1141, 0.0425, 0.01, -0.1141, 0.0439, 0.01, -0.112, 0.0425, 0.01, -0.112, 0.0455, 0.0061, -0.1098, 0.0455, 0.0083, -0.1098, 0.0459, 0.0061, -0.1076, 0.0459, 0.0083, -0.1076, 0.0455, 0.01, -0.1098, 0.0459, 0.01, -0.1077, 0.0465, 0.0061, -0.1054, 0.0465, 0.0083, -0.1054, 0.0478, 0.0061, -0.1031, 0.0471, 0.0061, -0.1043, 0.0478, 0.0083, -0.1031, 0.0471, 0.0083, -0.1043, 0.0464, 0.01, -0.1057, 0.0479, 0.01, -0.103, 0.0473, 0.01, -0.104, 0.0425, 0.0061, -0.1098, 0.0425, 0.0083, -0.1098, 0.0425, 0.0061, -0.1076, 0.0425, 0.0083, -0.1076, 0.0439, 0.01, -0.1098, 0.0425, 0.01, -0.1098, 0.0439, 0.01, -0.1076, 0.0425, 0.01, -0.1076, 0.0426, 0.0061, -0.1054, 0.0426, 0.0083, -0.1054, 0.0431, 0.0061, -0.1031,
0.0428, 0.0061, -0.1042, 0.0431, 0.0083, -0.1031, 0.0428, 0.0083, -0.1042, 0.0439, 0.01, -0.1055, 0.0426, 0.01, -0.1054, 0.0431, 0.01, -0.1032, 0.0428, 0.01, -0.1043, 0.0219, 0.0061, -0.1358, 0.0219, 0.0083, -0.1358, 0.0219, 0.0061, -0.1337, 0.0219, 0.0083, -0.1337, 0.0219, 0.01, -0.1358, 0.0201, 0.01, -0.1358, 0.0219, 0.01, -0.1337, 0.0201, 0.01, -0.1337, 0.0219, 0.0061, -0.1315, 0.0219, 0.0083, -0.1315, 0.0219, 0.0061, -0.1293, 0.0219, 0.0083, -0.1293, 0.0219, 0.01, -0.1315, 0.0201, 0.01, -0.1315, 0.0219, 0.01, -0.1293, 0.0201, 0.01, -0.1293, 0.0219, 0.0061, -0.1272, 0.0219, 0.0083, -0.1272, 0.0219, 0.0061, -0.125, 0.0219, 0.0083, -0.125, 0.0219, 0.01, -0.1272, 0.0201, 0.01, -0.1272, 0.0219, 0.01, -0.125,
0.0201, 0.01, -0.125, 0.0219, 0.0061, -0.1228, 0.0219, 0.0083, -0.1228, 0.0219, 0.0061, -0.1207, 0.0219, 0.0083, -0.1207, 0.0219, 0.01, -0.1228, 0.0201, 0.01, -0.1228, 0.0219, 0.01, -0.1207, 0.0201, 0.01, -0.1207, 0.0179, 0.01, -0.1358, 0.0157, 0.01, -0.1358, 0.0179, 0.01, -0.1337, 0.0157, 0.01, -0.1337, 0.0136, 0.01, -0.1358, 0.0114, 0.01, -0.1358, 0.0136, 0.01, -0.1337, 0.0114, 0.01, -0.1337, 0.0179, 0.01, -0.1315, 0.0157, 0.01, -0.1315, 0.0179, 0.01, -0.1293, 0.0157, 0.01, -0.1293, 0.0136, 0.01, -0.1315, 0.0114, 0.01, -0.1315, 0.0136, 0.01, -0.1293, 0.0114, 0.01, -0.1293, 0.0097, 0.0061, -0.1358, 0.0097, 0.0083, -0.1358, 0.01, 0.0061, -0.1337, 0.01, 0.0083, -0.1337, 0.0097, 0.01, -0.1358,
0.01, 0.01, -0.1337, 0.0101, 0.0061, -0.1315, 0.0101, 0.0083, -0.1315, 0.0101, 0.0061, -0.1293, 0.0101, 0.0083, -0.1293, 0.0101, 0.01, -0.1315, 0.0101, 0.01, -0.1293, 0.0179, 0.01, -0.1272, 0.0157, 0.01, -0.1272, 0.0179, 0.01, -0.125, 0.0157, 0.01, -0.125, 0.0136, 0.01, -0.1272, 0.0114, 0.01, -0.1272, 0.0136, 0.01, -0.125, 0.0114, 0.01, -0.125, 0.0179, 0.01, -0.1228, 0.0157, 0.01, -0.1228, 0.0179, 0.01, -0.1207, 0.0157, 0.01, -0.1207, 0.0136, 0.01, -0.1228, 0.0114, 0.01, -0.1228, 0.0136, 0.01, -0.1207, 0.0114, 0.01, -0.1207, 0.0101, 0.0061, -0.1272, 0.0101, 0.0083, -0.1272, 0.0101, 0.0061, -0.125, 0.0101, 0.0083, -0.125, 0.0101, 0.01, -0.1272, 0.0101, 0.01, -0.125, 0.0101, 0.0061, -0.1228,
0.0101, 0.0083, -0.1228, 0.0101, 0.0061, -0.1207, 0.0101, 0.0083, -0.1207, 0.0101, 0.01, -0.1228, 0.0101, 0.01, -0.1207, 0.0219, 0.0061, -0.1185, 0.0219, 0.0083, -0.1185, 0.0219, 0.0061, -0.1163, 0.0219, 0.0083, -0.1163, 0.0219, 0.01, -0.1185, 0.0201, 0.01, -0.1185, 0.0219, 0.01, -0.1163, 0.0201, 0.01, -0.1163, 0.0219, 0.0061, -0.1141, 0.0219, 0.0083, -0.1141, 0.0219, 0.0061, -0.112, 0.0219, 0.0083, -0.112, 0.0219, 0.01, -0.1141, 0.0201, 0.01, -0.1141, 0.0219, 0.01, -0.112, 0.0201, 0.01, -0.112, 0.0219, 0.0061, -0.1098, 0.0219, 0.0083, -0.1098, 0.0219, 0.0061, -0.1076, 0.0219, 0.0083, -0.1076, 0.0219, 0.01, -0.1098, 0.0201, 0.01, -0.1098, 0.0219, 0.01, -0.1076, 0.0201, 0.01, -0.1076, 0.0219, 0.0061, -0.1055,
0.0219, 0.0083, -0.1055, 0.0219, 0.0061, -0.1033, 0.0219, 0.0083, -0.1033, 0.0219, 0.01, -0.1055, 0.0201, 0.01, -0.1055, 0.0219, 0.01, -0.1033, 0.0201, 0.01, -0.1033, 0.0179, 0.01, -0.1185, 0.0157, 0.01, -0.1185, 0.0179, 0.01, -0.1163, 0.0157, 0.01, -0.1163, 0.0136, 0.01, -0.1185, 0.0114, 0.01, -0.1185, 0.0136, 0.01, -0.1163, 0.0114, 0.01, -0.1163, 0.0179, 0.01, -0.1141, 0.0157, 0.01, -0.1141, 0.0179, 0.01, -0.112, 0.0157, 0.01, -0.112, 0.0136, 0.01, -0.1141, 0.0114, 0.01, -0.1141, 0.0136, 0.01, -0.112, 0.0114, 0.01, -0.112, 0.0101, 0.0061, -0.1185, 0.0101, 0.0083, -0.1185, 0.0101, 0.0061, -0.1163, 0.0101, 0.0083, -0.1163, 0.0101, 0.01, -0.1185, 0.0101, 0.01, -0.1163, 0.0101, 0.0061, -0.1141,
0.0101, 0.0083, -0.1141, 0.0101, 0.0061, -0.112, 0.0101, 0.0083, -0.112, 0.0101, 0.01, -0.1141, 0.0101, 0.01, -0.112, 0.0179, 0.01, -0.1098, 0.0157, 0.01, -0.1098, 0.0179, 0.01, -0.1076, 0.0157, 0.01, -0.1076, 0.0136, 0.01, -0.1098, 0.0114, 0.01, -0.1098, 0.0136, 0.01, -0.1076, 0.0114, 0.01, -0.1076, 0.0179, 0.01, -0.1055, 0.0157, 0.01, -0.1055, 0.0179, 0.01, -0.1033, 0.0157, 0.01, -0.1033, 0.0136, 0.01, -0.1055, 0.0114, 0.01, -0.1055, 0.0136, 0.01, -0.1033, 0.0114, 0.01, -0.1033, 0.0101, 0.0061, -0.1098, 0.0101, 0.0083, -0.1098, 0.0101, 0.0061, -0.1076, 0.0101, 0.0083, -0.1076, 0.0101, 0.01, -0.1098, 0.0101, 0.01, -0.1076, 0.0101, 0.0061, -0.1055, 0.0101, 0.0083, -0.1055, 0.0101, 0.0061, -0.1033,
0.0101, 0.0083, -0.1033, 0.0101, 0.01, -0.1055, 0.0101, 0.01, -0.1033, 0.065, 0.0061, -0.1016, 0.0642, 0.0061, -0.1005, 0.065, 0.0083, -0.1016, 0.0642, 0.0083, -0.1005, 0.0631, 0.0061, -0.0994, 0.0631, 0.0083, -0.0994, 0.065, 0.01, -0.1017, 0.0642, 0.01, -0.1006, 0.0631, 0.01, -0.0994, 0.0604, 0.0061, -0.102, 0.0591, 0.0061, -0.1007, 0.0604, 0.0083, -0.102, 0.0591, 0.0083, -0.1007, 0.0618, 0.0061, -0.0983, 0.0618, 0.0083, -0.0983, 0.0577, 0.0061, -0.1002, 0.0577, 0.0083, -0.1002, 0.0566, 0.0061, -0.0997, 0.0548, 0.0061, -0.0995, 0.0566, 0.0083, -0.0997, 0.0548, 0.0083, -0.0995, 0.0604, 0.01, -0.102, 0.059, 0.01, -0.1007, 0.0618, 0.01, -0.0983, 0.0591, 0.01, -0.099, 0.0578, 0.01, -0.1002, 0.0569, 0.01, -0.0998,
0.0548, 0.01, -0.0995, 0.0607, 0.0061, -0.0976, 0.0592, 0.0061, -0.0968, 0.0607, 0.0083, -0.0976, 0.0592, 0.0083, -0.0968, 0.057, 0.0061, -0.0962, 0.0548, 0.0061, -0.0961, 0.057, 0.0083, -0.0962, 0.0548, 0.0083, -0.0961, 0.0606, 0.01, -0.0975, 0.0594, 0.01, -0.0968, 0.0571, 0.01, -0.0962, 0.0548, 0.01, -0.0961, 0.0518, 0.0061, -0.1002, 0.0504, 0.0061, -0.1008, 0.0518, 0.0083, -0.1002, 0.0504, 0.0083, -0.1008, 0.0529, 0.0061, -0.0997, 0.0529, 0.0083, -0.0997, 0.049, 0.0061, -0.1019, 0.049, 0.0083, -0.1019, 0.0473, 0.0061, -0.0979, 0.046, 0.0061, -0.0988, 0.0473, 0.0083, -0.0979, 0.046, 0.0083, -0.0988, 0.0518, 0.01, -0.1002, 0.0506, 0.01, -0.1006, 0.0527, 0.01, -0.0998, 0.0505, 0.01, -0.099, 0.049, 0.01, -0.1019,
0.0461, 0.01, -0.1011, 0.0474, 0.01, -0.0978, 0.0457, 0.01, -0.099, 0.0526, 0.0061, -0.0961, 0.0504, 0.0061, -0.0964, 0.0526, 0.0083, -0.0961, 0.0504, 0.0083, -0.0964, 0.0483, 0.0061, -0.0972, 0.0483, 0.0083, -0.0972, 0.0526, 0.01, -0.0961, 0.0503, 0.01, -0.0965, 0.0485, 0.01, -0.0971, 0.0441, 0.0061, -0.101, 0.0441, 0.0083, -0.101, 0.045, 0.0061, -0.1, 0.045, 0.0083, -0.1, 0.0442, 0.01, -0.1009, 0.045, 0.01, -0.1, 0.0589, 0.0061, -0.0766, 0.0589, 0.0083, -0.0766, 0.057, 0.0061, -0.0771, 0.0548, 0.0061, -0.0774, 0.057, 0.0083, -0.0771, 0.0548, 0.0083, -0.0774, 0.0587, 0.01, -0.0766, 0.057, 0.01, -0.0771, 0.0548, 0.01, -0.0774, 0.0627, 0.0061, -0.0742, 0.0627, 0.0083, -0.0742, 0.0647, 0.0061, -0.072,
0.0638, 0.0061, -0.0731, 0.0647, 0.0083, -0.072, 0.0638, 0.0083, -0.0731, 0.0626, 0.01, -0.0742, 0.0647, 0.01, -0.072, 0.0638, 0.01, -0.0732, 0.0655, 0.0061, -0.0708, 0.0655, 0.0083, -0.0708, 0.0664, 0.0061, -0.0686, 0.0624, 0.0061, -0.0676, 0.0664, 0.0083, -0.0686, 0.0624, 0.0083, -0.0676, 0.0654, 0.01, -0.071, 0.0635, 0.01, -0.0707, 0.0663, 0.01, -0.0688, 0.0623, 0.01, -0.0681, 0.0614, 0.0061, -0.0753, 0.06, 0.0061, -0.0761, 0.0614, 0.0083, -0.0753, 0.06, 0.0083, -0.0761, 0.0587, 0.0061, -0.0723, 0.0587, 0.0083, -0.0723, 0.057, 0.0061, -0.0729, 0.0548, 0.0061, -0.0733, 0.057, 0.0083, -0.0729, 0.0548, 0.0083, -0.0733, 0.0617, 0.01, -0.0751, 0.0597, 0.01, -0.0762, 0.0613, 0.01, -0.0729, 0.0588, 0.01, -0.0722,
0.057, 0.01, -0.0751, 0.0548, 0.01, -0.0751, 0.0568, 0.01, -0.073, 0.0548, 0.01, -0.0733, 0.0608, 0.0061, -0.0707, 0.0598, 0.0061, -0.0715, 0.0608, 0.0083, -0.0707, 0.0598, 0.0083, -0.0715, 0.062, 0.0061, -0.0687, 0.062, 0.0083, -0.0687, 0.061, 0.01, -0.0705, 0.0598, 0.01, -0.0714, 0.0619, 0.01, -0.0689, 0.0526, 0.0061, -0.0773, 0.0504, 0.0061, -0.0769, 0.0526, 0.0083, -0.0773, 0.0504, 0.0083, -0.0769, 0.049, 0.0061, -0.0763, 0.049, 0.0083, -0.0763, 0.0526, 0.01, -0.0773, 0.0503, 0.01, -0.0769, 0.0491, 0.01, -0.0764, 0.0526, 0.0061, -0.0732, 0.0508, 0.0061, -0.0727, 0.0526, 0.0083, -0.0732, 0.0508, 0.0083, -0.0727, 0.0479, 0.0061, -0.0758, 0.0464, 0.0061, -0.0747, 0.0479, 0.0083, -0.0758, 0.0464, 0.0083, -0.0747,
0.0453, 0.0061, -0.0736, 0.0453, 0.0083, -0.0736, 0.0526, 0.01, -0.0751, 0.0505, 0.01, -0.0751, 0.0527, 0.01, -0.0732, 0.0505, 0.01, -0.0725, 0.048, 0.01, -0.0758, 0.0464, 0.01, -0.0748, 0.0483, 0.01, -0.0729, 0.0453, 0.01, -0.0736, 0.0495, 0.0061, -0.0717, 0.0495, 0.0083, -0.0717, 0.0482, 0.0061, -0.0707, 0.0482, 0.0083, -0.0707, 0.0474, 0.0061, -0.0694, 0.0468, 0.0061, -0.0683, 0.0474, 0.0083, -0.0694, 0.0468, 0.0083, -0.0683, 0.0495, 0.01, -0.0717, 0.0483, 0.01, -0.0708, 0.0461, 0.01, -0.0707, 0.0474, 0.01, -0.0694, 0.0468, 0.01, -0.0684, 0.0445, 0.0061, -0.0726, 0.0445, 0.0083, -0.0726, 0.0445, 0.01, -0.0725, 0.0435, 0.0061, -0.0708, 0.0435, 0.0083, -0.0708, 0.043, 0.0061, -0.0692, 0.0427, 0.0061, -0.0681,
0.043, 0.0083, -0.0692, 0.0427, 0.0083, -0.0681, 0.0437, 0.01, -0.0711, 0.0429, 0.01, -0.0689, 0.0427, 0.01, -0.0679, 0.0219, 0.0061, -0.1011, 0.0219, 0.0083, -0.1011, 0.0219, 0.0061, -0.099, 0.0219, 0.0083, -0.099, 0.0219, 0.01, -0.1011, 0.0201, 0.01, -0.1011, 0.0219, 0.01, -0.099, 0.0201, 0.01, -0.099, 0.0219, 0.0061, -0.0968, 0.0219, 0.0083, -0.0968, 0.0219, 0.0061, -0.0946, 0.0219, 0.0083, -0.0946, 0.0219, 0.01, -0.0968, 0.0201, 0.01, -0.0968, 0.0219, 0.01, -0.0946, 0.0201, 0.01, -0.0946, 0.0219, 0.0061, -0.0924, 0.0219, 0.0083, -0.0924, 0.0219, 0.0061, -0.0902, 0.0201, 0.0061, -0.0902, 0.0219, 0.0083, -0.0902, 0.0201, 0.0083, -0.0902, 0.0219, 0.01, -0.0924, 0.0201, 0.01, -0.0924, 0.0219, 0.01, -0.0902,
0.0201, 0.01, -0.0902, 0.0179, 0.01, -0.1011, 0.0157, 0.01, -0.1011, 0.0179, 0.01, -0.099, 0.0157, 0.01, -0.099, 0.0136, 0.01, -0.1011, 0.0114, 0.01, -0.1011, 0.0136, 0.01, -0.099, 0.0114, 0.01, -0.099, 0.0179, 0.01, -0.0968, 0.0157, 0.01, -0.0968, 0.0179, 0.01, -0.0946, 0.0157, 0.01, -0.0946, 0.0136, 0.01, -0.0968, 0.0114, 0.01, -0.0968, 0.0136, 0.01, -0.0946, 0.0114, 0.01, -0.0946, 0.0101, 0.0061, -0.1011, 0.0101, 0.0083, -0.1011, 0.0101, 0.0061, -0.099, 0.0101, 0.0083, -0.099, 0.0101, 0.01, -0.1011, 0.0101, 0.01, -0.099, 0.0101, 0.0061, -0.0968, 0.0101, 0.0083, -0.0968, 0.0101, 0.0061, -0.0946, 0.0101, 0.0083, -0.0946, 0.0101, 0.01, -0.0968, 0.0101, 0.01, -0.0946, 0.0179, 0.0061, -0.0902,
0.0157, 0.0061, -0.0902, 0.0179, 0.0083, -0.0902, 0.0157, 0.0083, -0.0902, 0.0136, 0.0061, -0.0902, 0.0114, 0.0061, -0.0902, 0.0136, 0.0083, -0.0902, 0.0114, 0.0083, -0.0902, 0.0179, 0.01, -0.0924, 0.0157, 0.01, -0.0924, 0.0179, 0.01, -0.0902, 0.0157, 0.01, -0.0902, 0.0136, 0.01, -0.0924, 0.0114, 0.01, -0.0924, 0.0136, 0.01, -0.0902, 0.0114, 0.01, -0.0902, 0.0101, 0.0061, -0.0924, 0.0101, 0.0083, -0.0924, 0.0101, 0.0061, -0.0902, 0.0101, 0.0083, -0.0902, 0.0101, 0.01, -0.0924, 0.0101, 0.01, -0.0902, 0.0219, 0.0061, -0.0775, 0.0201, 0.0061, -0.0775, 0.0219, 0.0083, -0.0775, 0.0201, 0.0083, -0.0775, 0.0219, 0.01, -0.0775, 0.0201, 0.01, -0.0775, 0.0219, 0.0061, -0.0751, 0.0219, 0.0083, -0.0751, 0.0219, 0.0061, -0.0729,
0.0219, 0.0083, -0.0729, 0.0219, 0.01, -0.0751, 0.0201, 0.01, -0.0751, 0.0219, 0.01, -0.0729, 0.0201, 0.01, -0.0729, 0.0219, 0.0061, -0.0707, 0.0219, 0.0083, -0.0707, 0.0219, 0.0061, -0.0686, 0.0219, 0.0083, -0.0686, 0.0219, 0.01, -0.0707, 0.0201, 0.01, -0.0707, 0.0219, 0.01, -0.0686, 0.0201, 0.01, -0.0686, 0.0179, 0.0061, -0.0775, 0.0157, 0.0061, -0.0775, 0.0179, 0.0083, -0.0775, 0.0157, 0.0083, -0.0775, 0.0136, 0.0061, -0.0775, 0.0114, 0.0061, -0.0775, 0.0136, 0.0083, -0.0775, 0.0114, 0.0083, -0.0775, 0.0179, 0.01, -0.0775, 0.0157, 0.01, -0.0775, 0.0136, 0.01, -0.0775, 0.0114, 0.01, -0.0775, 0.0092, 0.0061, -0.0775, 0.0071, 0.0061, -0.0775, 0.0092, 0.0083, -0.0775, 0.0071, 0.0083, -0.0775, 0.0049, 0.0061, -0.0775,
0.0027, 0.0061, -0.0775, 0.0049, 0.0083, -0.0775, 0.0027, 0.0083, -0.0775, 0.0092, 0.01, -0.0775, 0.0071, 0.01, -0.0775, 0.0049, 0.01, -0.0775, 0.0027, 0.01, -0.0775, 0.0179, 0.01, -0.0751, 0.0157, 0.01, -0.0751, 0.0179, 0.01, -0.0729, 0.0157, 0.01, -0.0729, 0.0136, 0.01, -0.0751, 0.0114, 0.01, -0.0751, 0.0136, 0.01, -0.0729, 0.0114, 0.01, -0.0729, 0.0179, 0.01, -0.0707, 0.0157, 0.01, -0.0707, 0.0179, 0.01, -0.0686, 0.0157, 0.01, -0.0686, 0.0136, 0.01, -0.0707, 0.0114, 0.01, -0.0707, 0.0136, 0.01, -0.0686, 0.0114, 0.01, -0.0686, 0.0092, 0.01, -0.0751, 0.0071, 0.01, -0.0751, 0.0092, 0.01, -0.0729, 0.0071, 0.01, -0.0729, 0.0049, 0.01, -0.0751, 0.0027, 0.01, -0.0751, 0.0049, 0.01, -0.0729,
0.0027, 0.01, -0.0729, 0.0092, 0.01, -0.0707, 0.0071, 0.01, -0.0707, 0.0092, 0.01, -0.0686, 0.0071, 0.01, -0.0686, 0.0049, 0.01, -0.0707, 0.0027, 0.01, -0.0707, 0.0049, 0.01, -0.0686, 0.0027, 0.01, -0.0686, 0.0669, 0.0061, -0.0661, 0.0669, 0.0083, -0.0661, 0.067, 0.0061, -0.0642, 0.067, 0.0083, -0.0642, 0.0667, 0.0061, -0.0672, 0.0627, 0.0061, -0.0664, 0.0667, 0.0083, -0.0672, 0.0627, 0.0083, -0.0664, 0.063, 0.0061, -0.0642, 0.063, 0.0083, -0.0642, 0.0669, 0.01, -0.0659, 0.067, 0.01, -0.0642, 0.0668, 0.01, -0.0667, 0.0628, 0.01, -0.0663, 0.0656, 0.01, -0.0642, 0.063, 0.01, -0.0642, 0.067, 0.0061, -0.0621, 0.067, 0.0083, -0.0621, 0.067, 0.0061, -0.0599, 0.067, 0.0083, -0.0599, 0.063, 0.0061, -0.0621,
0.063, 0.0083, -0.0621, 0.0631, 0.0061, -0.0599, 0.0631, 0.0083, -0.0599, 0.067, 0.01, -0.0621, 0.067, 0.01, -0.0599, 0.0656, 0.01, -0.0621, 0.063, 0.01, -0.062, 0.0656, 0.01, -0.0599, 0.0631, 0.01, -0.0599, 0.0669, 0.0061, -0.0577, 0.0669, 0.0083, -0.0577, 0.0667, 0.0061, -0.0565, 0.0667, 0.0083, -0.0565, 0.063, 0.0061, -0.0577, 0.063, 0.0083, -0.0577, 0.0666, 0.0061, -0.0554, 0.0626, 0.0061, -0.0559, 0.0666, 0.0083, -0.0554, 0.0626, 0.0083, -0.0559, 0.0669, 0.01, -0.0577, 0.0668, 0.01, -0.0566, 0.0656, 0.01, -0.0577, 0.063, 0.01, -0.0578, 0.0666, 0.01, -0.0555, 0.0625, 0.01, -0.0557, 0.066, 0.0061, -0.0534, 0.066, 0.0083, -0.0534, 0.0651, 0.0061, -0.0514, 0.0645, 0.0061, -0.0503, 0.0651, 0.0083, -0.0514,
0.0645, 0.0083, -0.0503, 0.0659, 0.01, -0.0531, 0.0635, 0.01, -0.0534, 0.0652, 0.01, -0.0515, 0.0646, 0.01, -0.0505, 0.0623, 0.0061, -0.0548, 0.0623, 0.0083, -0.0548, 0.0623, 0.01, -0.0547, 0.0615, 0.0061, -0.0533, 0.0615, 0.0083, -0.0533, 0.0605, 0.0061, -0.052, 0.0592, 0.0061, -0.0509, 0.0605, 0.0083, -0.052, 0.0592, 0.0083, -0.0509, 0.058, 0.0061, -0.0502, 0.058, 0.0083, -0.0502, 0.0617, 0.01, -0.0535, 0.0605, 0.01, -0.052, 0.059, 0.01, -0.0507, 0.058, 0.01, -0.0502, 0.0461, 0.0061, -0.0664, 0.0461, 0.0083, -0.0664, 0.0456, 0.0061, -0.0642, 0.0456, 0.0083, -0.0642, 0.0461, 0.01, -0.0663, 0.0456, 0.01, -0.0642, 0.0454, 0.0061, -0.0621, 0.0454, 0.0083, -0.0621, 0.0455, 0.0061, -0.0599, 0.0455, 0.0083, -0.0599,
0.0454, 0.01, -0.0621, 0.0455, 0.01, -0.0599, 0.0425, 0.0061, -0.0664, 0.0425, 0.0083, -0.0664, 0.0425, 0.0061, -0.0642, 0.0425, 0.0083, -0.0642, 0.0439, 0.01, -0.0664, 0.0425, 0.01, -0.0664, 0.0439, 0.01, -0.0642, 0.0425, 0.01, -0.0642, 0.0425, 0.0061, -0.0621, 0.0425, 0.0083, -0.0621, 0.0425, 0.0061, -0.0599, 0.0425, 0.0083, -0.0599, 0.0439, 0.01, -0.0621, 0.0425, 0.01, -0.0621, 0.0439, 0.01, -0.0599, 0.0425, 0.01, -0.0599, 0.0458, 0.0061, -0.0577, 0.0458, 0.0083, -0.0577, 0.0463, 0.0061, -0.0555, 0.0463, 0.0083, -0.0555, 0.0457, 0.01, -0.0578, 0.0463, 0.01, -0.0557, 0.0502, 0.0061, -0.0506, 0.0502, 0.0083, -0.0506, 0.0476, 0.0061, -0.0529, 0.047, 0.0061, -0.054, 0.0476, 0.0083, -0.0529, 0.047, 0.0083, -0.054,
0.0487, 0.0061, -0.0517, 0.0487, 0.0083, -0.0517, 0.0501, 0.01, -0.0507, 0.0477, 0.01, -0.0528, 0.0471, 0.01, -0.0538, 0.0487, 0.01, -0.0517, 0.0461, 0.01, -0.0512, 0.0425, 0.0061, -0.0577, 0.0425, 0.0083, -0.0577, 0.0426, 0.0061, -0.0555, 0.0426, 0.0083, -0.0555, 0.0439, 0.01, -0.0577, 0.0425, 0.01, -0.0577, 0.0439, 0.01, -0.0555, 0.0426, 0.01, -0.0555, 0.0431, 0.0061, -0.0532, 0.0428, 0.0061, -0.0544, 0.0431, 0.0083, -0.0532, 0.0428, 0.0083, -0.0544, 0.044, 0.0061, -0.0511, 0.044, 0.0083, -0.0511, 0.0431, 0.01, -0.0533, 0.0428, 0.01, -0.0544, 0.0442, 0.01, -0.0509, 0.0635, 0.0061, -0.049, 0.0635, 0.0083, -0.049, 0.0636, 0.01, -0.049, 0.0623, 0.0061, -0.0481, 0.0623, 0.0083, -0.0481, 0.0613, 0.0061, -0.0473,
0.0596, 0.0061, -0.0465, 0.0613, 0.0083, -0.0473, 0.0596, 0.0083, -0.0465, 0.0569, 0.0061, -0.0497, 0.0547, 0.0061, -0.0492, 0.0569, 0.0083, -0.0497, 0.0547, 0.0083, -0.0492, 0.057, 0.0061, -0.0461, 0.0548, 0.0061, -0.0461, 0.057, 0.0083, -0.0461, 0.0548, 0.0083, -0.0461, 0.0622, 0.01, -0.048, 0.0591, 0.01, -0.049, 0.0611, 0.01, -0.0472, 0.0594, 0.01, -0.0465, 0.0569, 0.01, -0.0497, 0.0548, 0.01, -0.0492, 0.0571, 0.01, -0.0461, 0.0548, 0.01, -0.0461, 0.0526, 0.0061, -0.0496, 0.0513, 0.0061, -0.05, 0.0526, 0.0083, -0.0496, 0.0513, 0.0083, -0.05, 0.0526, 0.0061, -0.0461, 0.0504, 0.0061, -0.0463, 0.0526, 0.0083, -0.0461, 0.0504, 0.0083, -0.0463, 0.0459, 0.0061, -0.0488, 0.0459, 0.0083, -0.0488, 0.0482, 0.0061, -0.047,
0.0471, 0.0061, -0.0479, 0.0482, 0.0083, -0.047, 0.0471, 0.0083, -0.0479, 0.0528, 0.01, -0.0495, 0.051, 0.01, -0.0502, 0.0526, 0.01, -0.0461, 0.0503, 0.01, -0.0464, 0.0483, 0.01, -0.049, 0.0459, 0.01, -0.0489, 0.048, 0.01, -0.0471, 0.0471, 0.01, -0.0479, 0.045, 0.0061, -0.05, 0.045, 0.0083, -0.05, 0.0449, 0.01, -0.05, 0.0219, 0.0061, -0.0667, 0.0201, 0.0061, -0.0667, 0.0219, 0.0083, -0.0667, 0.0201, 0.0083, -0.0667, 0.0219, 0.01, -0.0667, 0.0201, 0.01, -0.0667, 0.0179, 0.0061, -0.0667, 0.0157, 0.0061, -0.0667, 0.0179, 0.0083, -0.0667, 0.0157, 0.0083, -0.0667, 0.0136, 0.0061, -0.0667, 0.0114, 0.0061, -0.0667, 0.0136, 0.0083, -0.0667, 0.0114, 0.0083, -0.0667, 0.0179, 0.01, -0.0667, 0.0157, 0.01, -0.0667,
0.0136, 0.01, -0.0667, 0.0114, 0.01, -0.0667, 0.0092, 0.0061, -0.0667, 0.0071, 0.0061, -0.0667, 0.0092, 0.0083, -0.0667, 0.0071, 0.0083, -0.0667, 0.0049, 0.0061, -0.0667, 0.0027, 0.0061, -0.0667, 0.0049, 0.0083, -0.0667, 0.0027, 0.0083, -0.0667, 0.0092, 0.01, -0.0667, 0.0071, 0.01, -0.0667, 0.0049, 0.01, -0.0667, 0.0027, 0.01, -0.0667, 0.0219, 0.0061, -0.05, 0.0201, 0.0061, -0.05, 0.0219, 0.0083, -0.05, 0.0201, 0.0083, -0.05, 0.0219, 0.0061, -0.0469, 0.0219, 0.0083, -0.0469, 0.0219, 0.01, -0.05, 0.0201, 0.01, -0.05, 0.0219, 0.01, -0.0469, 0.0201, 0.01, -0.0469, 0.0219, 0.0061, -0.0447, 0.0219, 0.0083, -0.0447, 0.0219, 0.0061, -0.0425, 0.0219, 0.0083, -0.0425, 0.0219, 0.01, -0.0447, 0.0201, 0.01, -0.0447,
0.0219, 0.01, -0.0425, 0.0201, 0.01, -0.0425, 0.0219, 0.0061, -0.0404, 0.0219, 0.0083, -0.0404, 0.0219, 0.0061, -0.0382, 0.0201, 0.0061, -0.0382, 0.0219, 0.0083, -0.0382, 0.0201, 0.0083, -0.0382, 0.0219, 0.01, -0.0404, 0.0201, 0.01, -0.0404, 0.0219, 0.01, -0.0382, 0.0201, 0.01, -0.0382, 0.0179, 0.0061, -0.05, 0.0157, 0.0061, -0.05, 0.0179, 0.0083, -0.05, 0.0157, 0.0083, -0.05, 0.0136, 0.0061, -0.05, 0.0114, 0.0061, -0.05, 0.0136, 0.0083, -0.05, 0.0114, 0.0083, -0.05, 0.0179, 0.01, -0.05, 0.0157, 0.01, -0.05, 0.0179, 0.01, -0.0469, 0.0157, 0.01, -0.0469, 0.0136, 0.01, -0.05, 0.0114, 0.01, -0.05, 0.0136, 0.01, -0.0469, 0.0114, 0.01, -0.0469, 0.0179, 0.01, -0.0447, 0.0157, 0.01, -0.0447,
0.0179, 0.01, -0.0425, 0.0157, 0.01, -0.0425, 0.0136, 0.01, -0.0447, 0.0114, 0.01, -0.0447, 0.0136, 0.01, -0.0425, 0.0114, 0.01, -0.0425, 0.0092, 0.0061, -0.05, 0.0071, 0.0061, -0.05, 0.0092, 0.0083, -0.05, 0.0071, 0.0083, -0.05, 0.0049, 0.0061, -0.05, 0.0027, 0.0061, -0.05, 0.0049, 0.0083, -0.05, 0.0027, 0.0083, -0.05, 0.0092, 0.01, -0.05, 0.0071, 0.01, -0.05, 0.0092, 0.01, -0.0469, 0.0071, 0.01, -0.0469, 0.0049, 0.01, -0.05, 0.0027, 0.01, -0.05, 0.0049, 0.01, -0.0469, 0.0027, 0.01, -0.0469, 0.0092, 0.01, -0.0447, 0.0071, 0.01, -0.0447, 0.0092, 0.01, -0.0425, 0.0071, 0.01, -0.0425, 0.0049, 0.01, -0.0447, 0.0027, 0.01, -0.0447, 0.0049, 0.01, -0.0425, 0.0027, 0.01, -0.0425,
0.0179, 0.0061, -0.0382, 0.0157, 0.0061, -0.0382, 0.0179, 0.0083, -0.0382, 0.0157, 0.0083, -0.0382, 0.0136, 0.0061, -0.0382, 0.0114, 0.0061, -0.0382, 0.0136, 0.0083, -0.0382, 0.0114, 0.0083, -0.0382, 0.0179, 0.01, -0.0404, 0.0157, 0.01, -0.0404, 0.0179, 0.01, -0.0382, 0.0157, 0.01, -0.0382, 0.0136, 0.01, -0.0404, 0.0114, 0.01, -0.0404, 0.0136, 0.01, -0.0382, 0.0114, 0.01, -0.0382, 0.0092, 0.0061, -0.0382, 0.0071, 0.0061, -0.0382, 0.0092, 0.0083, -0.0382, 0.0071, 0.0083, -0.0382, 0.0049, 0.0061, -0.0382, 0.0027, 0.0061, -0.0382, 0.0049, 0.0083, -0.0382, 0.0027, 0.0083, -0.0382, 0.0092, 0.01, -0.0404, 0.0071, 0.01, -0.0404, 0.0092, 0.01, -0.0382, 0.0071, 0.01, -0.0382, 0.0049, 0.01, -0.0404, 0.0027, 0.01, -0.0404,
0.0049, 0.01, -0.0382, 0.0027, 0.01, -0.0382, 0.067, 0.0061, -0.0265, 0.067, 0.0083, -0.0265, 0.067, 0.0061, -0.0252, 0.067, 0.0083, -0.0252, 0.0656, 0.0061, -0.0265, 0.0635, 0.0061, -0.0265, 0.0656, 0.0083, -0.0265, 0.0635, 0.0083, -0.0265, 0.067, 0.01, -0.0265, 0.067, 0.01, -0.0252, 0.0656, 0.01, -0.0265, 0.0635, 0.01, -0.0265, 0.0656, 0.01, -0.0252, 0.0635, 0.01, -0.0252, 0.0613, 0.0061, -0.0265, 0.0591, 0.0061, -0.0265, 0.0613, 0.0083, -0.0265, 0.0591, 0.0083, -0.0265, 0.057, 0.0061, -0.0265, 0.0548, 0.0061, -0.0265, 0.057, 0.0083, -0.0265, 0.0548, 0.0083, -0.0265, 0.0613, 0.01, -0.0265, 0.0591, 0.01, -0.0265, 0.0613, 0.01, -0.0252, 0.0591, 0.01, -0.0252, 0.057, 0.01, -0.0265, 0.0548, 0.01, -0.0265,
0.057, 0.01, -0.0252, 0.0548, 0.01, -0.0252, 0.067, 0.0061, -0.0225, 0.067, 0.0083, -0.0225, 0.0656, 0.0061, -0.0225, 0.0635, 0.0061, -0.0225, 0.0656, 0.0083, -0.0225, 0.0635, 0.0083, -0.0225, 0.067, 0.01, -0.0225, 0.0656, 0.01, -0.0225, 0.0635, 0.01, -0.0225, 0.0613, 0.0061, -0.0225, 0.0591, 0.0061, -0.0225, 0.0613, 0.0083, -0.0225, 0.0591, 0.0083, -0.0225, 0.057, 0.0061, -0.0225, 0.0548, 0.0061, -0.0225, 0.057, 0.0083, -0.0225, 0.0548, 0.0083, -0.0225, 0.0613, 0.01, -0.0225, 0.0591, 0.01, -0.0225, 0.057, 0.01, -0.0225, 0.0548, 0.01, -0.0225, 0.054, 0.0061, -0.0179, 0.054, 0.0083, -0.0179, 0.0562, 0.0061, -0.0157, 0.0551, 0.0061, -0.0168, 0.0562, 0.0083, -0.0157, 0.0551, 0.0083, -0.0168, 0.054, 0.01, -0.0179,
0.0562, 0.01, -0.0157, 0.0551, 0.01, -0.0168, 0.0526, 0.0061, -0.0265, 0.0505, 0.0061, -0.0265, 0.0526, 0.0083, -0.0265, 0.0505, 0.0083, -0.0265, 0.0483, 0.0061, -0.0265, 0.0466, 0.0061, -0.0264, 0.0483, 0.0083, -0.0265, 0.0466, 0.0083, -0.0264, 0.0455, 0.0061, -0.0262, 0.0455, 0.0083, -0.0262, 0.0526, 0.01, -0.0265, 0.0505, 0.01, -0.0265, 0.0526, 0.01, -0.0252, 0.0505, 0.01, -0.0252, 0.0483, 0.01, -0.0265, 0.0468, 0.01, -0.0264, 0.0483, 0.01, -0.0252, 0.0459, 0.01, -0.0262, 0.0437, 0.0061, -0.0255, 0.0437, 0.0083, -0.0255, 0.0437, 0.01, -0.0255, 0.0526, 0.0061, -0.0225, 0.0503, 0.0061, -0.0225, 0.0526, 0.0083, -0.0225, 0.0503, 0.0083, -0.0225, 0.052, 0.0061, -0.0202, 0.0511, 0.0061, -0.0213, 0.052, 0.0083, -0.0202,
0.0511, 0.0083, -0.0213, 0.0452, 0.0061, -0.02, 0.0452, 0.0083, -0.02, 0.0526, 0.01, -0.0225, 0.0503, 0.01, -0.0225, 0.052, 0.01, -0.0202, 0.0511, 0.01, -0.0213, 0.0483, 0.01, -0.023, 0.0461, 0.01, -0.023, 0.0483, 0.01, -0.0208, 0.0452, 0.01, -0.0201, 0.053, 0.0061, -0.019, 0.053, 0.0083, -0.019, 0.0495, 0.0061, -0.0155, 0.0495, 0.0083, -0.0155, 0.0473, 0.0061, -0.0177, 0.0463, 0.0061, -0.0188, 0.0473, 0.0083, -0.0177, 0.0463, 0.0083, -0.0188, 0.0484, 0.0061, -0.0166, 0.0484, 0.0083, -0.0166, 0.053, 0.01, -0.019, 0.0505, 0.01, -0.0187, 0.0526, 0.01, -0.0165, 0.0495, 0.01, -0.0155, 0.0473, 0.01, -0.0177, 0.0463, 0.01, -0.0188, 0.0484, 0.01, -0.0166, 0.0433, 0.0061, -0.0231, 0.0433, 0.0083, -0.0231,
0.0444, 0.0061, -0.021, 0.0444, 0.0083, -0.021, 0.0433, 0.01, -0.0231, 0.0443, 0.01, -0.0211, 0.0627, 0.0061, -0.0092, 0.0627, 0.0083, -0.0092, 0.0647, 0.0061, -0.0069, 0.0638, 0.0061, -0.008, 0.0647, 0.0083, -0.0069, 0.0638, 0.0083, -0.008, 0.0627, 0.01, -0.0092, 0.0647, 0.01, -0.0069, 0.0638, 0.01, -0.008, 0.0584, 0.0061, -0.0135, 0.0584, 0.0083, -0.0135, 0.0605, 0.0061, -0.0114, 0.0594, 0.0061, -0.0124, 0.0605, 0.0083, -0.0114, 0.0594, 0.0083, -0.0124, 0.0573, 0.0061, -0.0146, 0.0573, 0.0083, -0.0146, 0.0538, 0.0061, -0.0112, 0.0538, 0.0083, -0.0112, 0.0584, 0.01, -0.0135, 0.0605, 0.01, -0.0114, 0.0594, 0.01, -0.0124, 0.0573, 0.01, -0.0146, 0.0548, 0.01, -0.0143, 0.057, 0.01, -0.0121, 0.0538, 0.01, -0.0112,
0.0616, 0.0061, -0.0103, 0.0616, 0.0083, -0.0103, 0.0582, 0.0061, -0.0068, 0.0582, 0.0083, -0.0068, 0.056, 0.0061, -0.009, 0.0549, 0.0061, -0.0101, 0.056, 0.0083, -0.009, 0.0549, 0.0083, -0.0101, 0.0571, 0.0061, -0.0079, 0.0571, 0.0083, -0.0079, 0.0616, 0.01, -0.0103, 0.0591, 0.01, -0.01, 0.0613, 0.01, -0.0078, 0.0582, 0.01, -0.0068, 0.056, 0.01, -0.009, 0.0549, 0.01, -0.0101, 0.0571, 0.01, -0.0079, 0.0668, 0.0061, -0.0028, 0.0668, 0.0083, -0.0028, 0.0656, 0.0061, -0.0057, 0.0656, 0.0083, -0.0057, 0.0666, 0.0061, -0.0039, 0.0666, 0.0083, -0.0039, 0.0668, 0.01, -0.0026, 0.0655, 0.01, -0.0059, 0.0635, 0.01, -0.0056, 0.0666, 0.01, -0.0037, 0.0635, 0.01, -0.0035, 0.067, 0.0061, -0.002, 0.067, 0.0083, -0.002,
0.0656, 0.0061, -0.002, 0.0635, 0.0061, -0.002, 0.0656, 0.0083, -0.002, 0.0635, 0.0083, -0.002, 0.067, 0.01, -0.002, 0.0656, 0.01, -0.002, 0.0635, 0.01, -0.002, 0.0591, 0.0061, -0.0059, 0.0591, 0.0083, -0.0059, 0.057, 0.0061, -0.0059, 0.0548, 0.0061, -0.0059, 0.057, 0.0083, -0.0059, 0.0548, 0.0083, -0.0059, 0.0613, 0.01, -0.0056, 0.0591, 0.01, -0.0059, 0.0613, 0.01, -0.0035, 0.0591, 0.01, -0.0035, 0.057, 0.01, -0.0059, 0.0548, 0.01, -0.0059, 0.057, 0.01, -0.0035, 0.0548, 0.01, -0.0035, 0.0613, 0.0061, -0.002, 0.0591, 0.0061, -0.002, 0.0613, 0.0083, -0.002, 0.0591, 0.0083, -0.002, 0.057, 0.0061, -0.002, 0.0548, 0.0061, -0.002, 0.057, 0.0083, -0.002, 0.0548, 0.0083, -0.002, 0.0613, 0.01, -0.002,
0.0591, 0.01, -0.002, 0.057, 0.01, -0.002, 0.0548, 0.01, -0.002, 0.0517, 0.0061, -0.0134, 0.0506, 0.0061, -0.0144, 0.0517, 0.0083, -0.0134, 0.0506, 0.0083, -0.0144, 0.0528, 0.0061, -0.0123, 0.0528, 0.0083, -0.0123, 0.0517, 0.01, -0.0134, 0.0506, 0.01, -0.0144, 0.0528, 0.01, -0.0123, 0.0526, 0.0061, -0.0059, 0.0505, 0.0061, -0.0059, 0.0526, 0.0083, -0.0059, 0.0505, 0.0083, -0.0059, 0.0483, 0.0061, -0.0059, 0.0461, 0.0061, -0.0059, 0.0483, 0.0083, -0.0059, 0.0461, 0.0083, -0.0059, 0.0526, 0.01, -0.0059, 0.0505, 0.01, -0.0059, 0.0526, 0.01, -0.0035, 0.0505, 0.01, -0.0035, 0.0483, 0.01, -0.0059, 0.0461, 0.01, -0.0059, 0.0483, 0.01, -0.0035, 0.0461, 0.01, -0.0035, 0.0526, 0.0061, -0.002, 0.0505, 0.0061, -0.002,
0.0526, 0.0083, -0.002, 0.0505, 0.0083, -0.002, 0.0483, 0.0061, -0.002, 0.0461, 0.0061, -0.002, 0.0483, 0.0083, -0.002, 0.0461, 0.0083, -0.002, 0.0526, 0.01, -0.002, 0.0505, 0.01, -0.002, 0.0483, 0.01, -0.002, 0.0461, 0.01, -0.002, 0.0439, 0.0061, -0.0059, 0.0425, 0.0061, -0.0059, 0.0439, 0.0083, -0.0059, 0.0425, 0.0083, -0.0059, 0.0425, 0.0061, -0.0035, 0.0425, 0.0083, -0.0035, 0.0439, 0.01, -0.0059, 0.0425, 0.01, -0.0059, 0.0439, 0.01, -0.0035, 0.0425, 0.01, -0.0035, 0.0439, 0.0061, -0.002, 0.0425, 0.0061, -0.002, 0.0439, 0.0083, -0.002, 0.0425, 0.0083, -0.002, 0.0439, 0.01, -0.002, 0.0425, 0.01, -0.002, 0.0005, 0.0061, -0.4807, -0.0016, 0.0061, -0.4801, 0.0005, 0.0083, -0.4807, -0.0016, 0.0083, -0.4801,
0.0004, 0.01, -0.4807, -0.0013, 0.01, -0.4802, -0.0027, 0.0061, -0.4798, -0.0027, 0.0083, -0.4798, -0.0038, 0.0061, -0.4795, -0.006, 0.0061, -0.4789, -0.0038, 0.0083, -0.4795, -0.006, 0.0083, -0.4789, 0.0005, 0.01, -0.4787, -0.0021, 0.01, -0.48, 0.0005, 0.01, -0.4766, -0.0016, 0.01, -0.4766, -0.004, 0.01, -0.4795, -0.0061, 0.01, -0.4789, -0.0038, 0.01, -0.4766, -0.006, 0.01, -0.4766, -0.0081, 0.0061, -0.4784, -0.0102, 0.0061, -0.4779, -0.0081, 0.0083, -0.4784, -0.0102, 0.0083, -0.4779, -0.0112, 0.0061, -0.4776, -0.0112, 0.0083, -0.4776, -0.0125, 0.0061, -0.4772, -0.0147, 0.0061, -0.4765, -0.0125, 0.0083, -0.4772, -0.0147, 0.0083, -0.4765, -0.0082, 0.01, -0.4784, -0.0099, 0.01, -0.478, -0.0081, 0.01, -0.4766, -0.0107, 0.01, -0.4777,
-0.0127, 0.01, -0.4772, -0.0148, 0.01, -0.4765, 0.0005, 0.01, -0.4744, -0.0016, 0.01, -0.4744, 0.0005, 0.01, -0.4722, -0.0016, 0.01, -0.4722, -0.0038, 0.01, -0.4744, -0.006, 0.01, -0.4744, -0.0038, 0.01, -0.4722, -0.006, 0.01, -0.4722, 0.0005, 0.01, -0.47, -0.0016, 0.01, -0.47, 0.0005, 0.01, -0.4679, -0.0016, 0.01, -0.4679, -0.0038, 0.01, -0.47, -0.006, 0.01, -0.47, -0.0038, 0.01, -0.4679, -0.006, 0.01, -0.4679, -0.0081, 0.01, -0.4744, -0.0103, 0.01, -0.4744, -0.0081, 0.01, -0.4722, -0.0103, 0.01, -0.4722, -0.0125, 0.01, -0.4744, -0.0146, 0.01, -0.4744, -0.0125, 0.01, -0.4722, -0.0146, 0.01, -0.4722, -0.0081, 0.01, -0.47, -0.0103, 0.01, -0.47, -0.0081, 0.01, -0.4679, -0.0103, 0.01, -0.4679,
-0.0125, 0.01, -0.47, -0.0146, 0.01, -0.47, -0.0125, 0.01, -0.4679, -0.0146, 0.01, -0.4679, -0.0166, 0.0061, -0.4758, -0.0166, 0.0083, -0.4758, -0.0164, 0.01, -0.4759, -0.0177, 0.0061, -0.4754, -0.019, 0.0061, -0.4749, -0.0177, 0.0083, -0.4754, -0.019, 0.0083, -0.4749, -0.0212, 0.0061, -0.4742, -0.0229, 0.0061, -0.4736, -0.0212, 0.0083, -0.4742, -0.0229, 0.0083, -0.4736, -0.024, 0.0061, -0.4732, -0.024, 0.0083, -0.4732, -0.0173, 0.01, -0.4755, -0.0192, 0.01, -0.4749, -0.0168, 0.01, -0.4722, -0.019, 0.01, -0.4722, -0.0213, 0.01, -0.4742, -0.0228, 0.01, -0.4736, -0.0212, 0.01, -0.4722, -0.0237, 0.01, -0.4733, -0.0168, 0.01, -0.47, -0.019, 0.01, -0.47, -0.0168, 0.01, -0.4679, -0.019, 0.01, -0.4679, -0.0212, 0.01, -0.47,
-0.0233, 0.01, -0.47, -0.0212, 0.01, -0.4679, -0.0233, 0.01, -0.4679, -0.0255, 0.0061, -0.4726, -0.0277, 0.0061, -0.4716, -0.0255, 0.0083, -0.4726, -0.0277, 0.0083, -0.4716, -0.0288, 0.0061, -0.4711, -0.0288, 0.0083, -0.4711, -0.0258, 0.01, -0.4724, -0.0278, 0.01, -0.4716, -0.0288, 0.01, -0.4711, -0.0299, 0.0061, -0.4707, -0.032, 0.0061, -0.4698, -0.0299, 0.0083, -0.4707, -0.032, 0.0083, -0.4698, -0.0255, 0.01, -0.47, -0.0277, 0.01, -0.47, -0.0255, 0.01, -0.4679, -0.0277, 0.01, -0.4679, -0.0299, 0.01, -0.4707, -0.0322, 0.01, -0.4697, -0.0298, 0.01, -0.4679, -0.032, 0.01, -0.4679, 0.0005, 0.0061, -0.4628, -0.0006, 0.0061, -0.4625, 0.0005, 0.0083, -0.4628, -0.0006, 0.0083, -0.4625, 0.0005, 0.01, -0.4657, -0.0016, 0.01, -0.4657,
0.0008, 0.01, -0.4628, -0.0012, 0.01, -0.4623, -0.0038, 0.01, -0.4657, -0.006, 0.01, -0.4657, -0.0038, 0.01, -0.4635, -0.006, 0.01, -0.4635, -0.0017, 0.0061, -0.4622, -0.0017, 0.0083, -0.4622, -0.0038, 0.0061, -0.4615, -0.006, 0.0061, -0.4608, -0.0038, 0.0083, -0.4615, -0.006, 0.0083, -0.4608, -0.002, 0.01, -0.4621, -0.0037, 0.01, -0.4616, -0.0057, 0.01, -0.4609, -0.0081, 0.01, -0.4657, -0.0103, 0.01, -0.4657, -0.0081, 0.01, -0.4635, -0.0103, 0.01, -0.4635, -0.0125, 0.01, -0.4657, -0.0146, 0.01, -0.4657, -0.0125, 0.01, -0.4635, -0.0146, 0.01, -0.4635, -0.0074, 0.0061, -0.4604, -0.0074, 0.0083, -0.4604, -0.0085, 0.0061, -0.4601, -0.0103, 0.0061, -0.4595, -0.0085, 0.0083, -0.4601, -0.0103, 0.0083, -0.4595, -0.0125, 0.0061, -0.4587,
-0.0138, 0.0061, -0.4582, -0.0125, 0.0083, -0.4587, -0.0138, 0.0083, -0.4582, -0.0078, 0.01, -0.4603, -0.0103, 0.01, -0.4614, -0.0087, 0.01, -0.46, -0.0102, 0.01, -0.4595, -0.0125, 0.01, -0.4614, -0.0146, 0.01, -0.4614, -0.0122, 0.01, -0.4588, -0.0142, 0.01, -0.4581, -0.0149, 0.0061, -0.4578, -0.0149, 0.0083, -0.4578, -0.0151, 0.01, -0.4577, -0.0168, 0.01, -0.4657, -0.019, 0.01, -0.4657, -0.0168, 0.01, -0.4635, -0.019, 0.01, -0.4635, -0.0212, 0.01, -0.4657, -0.0233, 0.01, -0.4657, -0.0212, 0.01, -0.4635, -0.0233, 0.01, -0.4635, -0.0168, 0.01, -0.4614, -0.019, 0.01, -0.4614, -0.0168, 0.01, -0.4592, -0.019, 0.01, -0.4592, -0.0212, 0.01, -0.4614, -0.0233, 0.01, -0.4614, -0.0212, 0.01, -0.4592, -0.0233, 0.01, -0.4592,
-0.0255, 0.01, -0.4657, -0.0277, 0.01, -0.4657, -0.0255, 0.01, -0.4635, -0.0277, 0.01, -0.4635, -0.0298, 0.01, -0.4657, -0.032, 0.01, -0.4657, -0.0298, 0.01, -0.4635, -0.032, 0.01, -0.4635, -0.0255, 0.01, -0.4614, -0.0277, 0.01, -0.4614, -0.0255, 0.01, -0.4592, -0.0277, 0.01, -0.4592, -0.0298, 0.01, -0.4614, -0.032, 0.01, -0.4614, -0.0298, 0.01, -0.4592, -0.032, 0.01, -0.4592, -0.0168, 0.0061, -0.4571, -0.019, 0.0061, -0.4563, -0.0168, 0.0083, -0.4571, -0.019, 0.0083, -0.4563, -0.0201, 0.0061, -0.4559, -0.0201, 0.0083, -0.4559, -0.0212, 0.0061, -0.4556, -0.0233, 0.0061, -0.4548, -0.0212, 0.0083, -0.4556, -0.0233, 0.0083, -0.4548, -0.0167, 0.01, -0.4572, -0.0187, 0.01, -0.4564, -0.0206, 0.01, -0.4558, -0.0233, 0.01, -0.457,
-0.0214, 0.01, -0.4555, -0.0231, 0.01, -0.4549, -0.0251, 0.0061, -0.4541, -0.0251, 0.0083, -0.4541, -0.0255, 0.01, -0.457, -0.0277, 0.01, -0.457, -0.0252, 0.01, -0.454, -0.0277, 0.01, -0.4549, -0.0298, 0.01, -0.457, -0.032, 0.01, -0.457, -0.0298, 0.01, -0.4549, -0.032, 0.01, -0.4549, -0.0262, 0.0061, -0.4536, -0.0277, 0.0061, -0.4529, -0.0262, 0.0083, -0.4536, -0.0277, 0.0083, -0.4529, -0.0295, 0.0061, -0.452, -0.0295, 0.0083, -0.452, -0.0306, 0.0061, -0.4514, -0.032, 0.0061, -0.4507, -0.0306, 0.0083, -0.4514, -0.032, 0.0083, -0.4507, -0.0263, 0.01, -0.4536, -0.0275, 0.01, -0.453, -0.0296, 0.01, -0.4519, -0.032, 0.01, -0.4527, -0.0307, 0.01, -0.4514, -0.0318, 0.01, -0.4508, -0.0334, 0.0061, -0.4691, -0.0334, 0.0083, -0.4691,
-0.0345, 0.0061, -0.4686, -0.0364, 0.0061, -0.4677, -0.0345, 0.0083, -0.4686, -0.0364, 0.0083, -0.4677, -0.0378, 0.0061, -0.467, -0.0378, 0.0083, -0.467, -0.0333, 0.01, -0.4691, -0.0344, 0.01, -0.4686, -0.0365, 0.01, -0.4676, -0.0377, 0.01, -0.467, -0.0389, 0.0061, -0.4664, -0.0407, 0.0061, -0.4655, -0.0389, 0.0083, -0.4664, -0.0407, 0.0083, -0.4655, -0.0342, 0.01, -0.4657, -0.0363, 0.01, -0.4657, -0.0342, 0.01, -0.4635, -0.0363, 0.01, -0.4635, -0.0388, 0.01, -0.4664, -0.0409, 0.01, -0.4654, -0.0385, 0.01, -0.4635, -0.0407, 0.01, -0.4635, -0.0342, 0.01, -0.4614, -0.0363, 0.01, -0.4614, -0.0342, 0.01, -0.4592, -0.0363, 0.01, -0.4592, -0.0385, 0.01, -0.4614, -0.0407, 0.01, -0.4614, -0.0385, 0.01, -0.4592, -0.0407, 0.01, -0.4592,
-0.042, 0.0061, -0.4647, -0.042, 0.0083, -0.4647, -0.0431, 0.0061, -0.4641, -0.0449, 0.0061, -0.463, -0.0431, 0.0083, -0.4641, -0.0449, 0.0083, -0.463, -0.042, 0.01, -0.4648, -0.0431, 0.01, -0.4641, -0.0448, 0.01, -0.4631, -0.046, 0.0061, -0.4624, -0.046, 0.0083, -0.4624, -0.0472, 0.0061, -0.4617, -0.0488, 0.0061, -0.4606, -0.0472, 0.0083, -0.4617, -0.0488, 0.0083, -0.4606, -0.0499, 0.0061, -0.4599, -0.0499, 0.0083, -0.4599, -0.0429, 0.01, -0.4614, -0.0458, 0.01, -0.4625, -0.0429, 0.01, -0.4592, -0.045, 0.01, -0.4592, -0.0475, 0.01, -0.4615, -0.0488, 0.01, -0.4607, -0.0472, 0.01, -0.4592, -0.0498, 0.01, -0.46, -0.0342, 0.01, -0.457, -0.0363, 0.01, -0.457, -0.0342, 0.01, -0.4549, -0.0363, 0.01, -0.4549, -0.0385, 0.01, -0.457,
-0.0407, 0.01, -0.457, -0.0385, 0.01, -0.4549, -0.0407, 0.01, -0.4549, -0.0339, 0.0061, -0.4498, -0.0339, 0.0083, -0.4498, -0.0417, 0.0061, -0.4517, -0.0417, 0.0083, -0.4517, -0.0392, 0.0061, -0.4498, -0.0408, 0.0061, -0.4505, -0.0392, 0.0083, -0.4498, -0.0408, 0.0083, -0.4505, -0.0342, 0.01, -0.4527, -0.0363, 0.01, -0.4527, -0.0339, 0.01, -0.4498, -0.0363, 0.01, -0.4505, -0.0385, 0.01, -0.4527, -0.0415, 0.01, -0.4515, -0.0393, 0.01, -0.4499, -0.0408, 0.01, -0.4505, -0.0429, 0.01, -0.457, -0.045, 0.01, -0.457, -0.0429, 0.01, -0.4549, -0.045, 0.01, -0.4549, -0.0472, 0.01, -0.457, -0.0494, 0.01, -0.457, -0.0472, 0.01, -0.4549, -0.0494, 0.01, -0.4549, -0.0426, 0.0061, -0.4523, -0.0452, 0.0061, -0.452, -0.0426, 0.0083, -0.4523,
-0.0452, 0.0083, -0.452, -0.0429, 0.0061, -0.4512, -0.0447, 0.0061, -0.4515, -0.0429, 0.0083, -0.4512, -0.0447, 0.0083, -0.4515, -0.0469, 0.0061, -0.4519, -0.0494, 0.0061, -0.4522, -0.0469, 0.0083, -0.4519, -0.0494, 0.0083, -0.4522, -0.0473, 0.0061, -0.4516, -0.0494, 0.0061, -0.4509, -0.0473, 0.0083, -0.4516, -0.0494, 0.0083, -0.4509, -0.0431, 0.01, -0.4525, -0.0453, 0.01, -0.452, -0.0425, 0.01, -0.451, -0.045, 0.01, -0.4515, -0.0468, 0.01, -0.452, -0.0492, 0.01, -0.4523, -0.0472, 0.01, -0.4516, -0.0496, 0.01, -0.4508, -0.0515, 0.0061, -0.4589, -0.0515, 0.0083, -0.4589, -0.0514, 0.01, -0.459, -0.0526, 0.0061, -0.4581, -0.0537, 0.0061, -0.4573, -0.0526, 0.0083, -0.4581, -0.0537, 0.0083, -0.4573, -0.0551, 0.0061, -0.4562, -0.0551, 0.0083, -0.4562,
-0.0562, 0.0061, -0.4553, -0.0576, 0.0061, -0.4543, -0.0562, 0.0083, -0.4553, -0.0576, 0.0083, -0.4543, -0.0524, 0.01, -0.4582, -0.054, 0.01, -0.457, -0.0515, 0.01, -0.4549, -0.0537, 0.01, -0.4549, -0.0551, 0.01, -0.4562, -0.0562, 0.01, -0.4553, -0.0575, 0.01, -0.4543, -0.0505, 0.0061, -0.4517, -0.0505, 0.0083, -0.4517, -0.0521, 0.0061, -0.4499, -0.0529, 0.0061, -0.4497, -0.0521, 0.0083, -0.4499, -0.0529, 0.0083, -0.4497, -0.0586, 0.0061, -0.4534, -0.0586, 0.0083, -0.4534, -0.0508, 0.01, -0.4515, -0.0537, 0.01, -0.4527, -0.0514, 0.01, -0.4504, -0.0529, 0.01, -0.4498, -0.0559, 0.01, -0.4527, -0.0586, 0.01, -0.4534, -0.0559, 0.01, -0.4505, -0.058, 0.01, -0.4505, -0.0599, 0.0061, -0.4523, -0.0599, 0.0083, -0.4523, -0.061, 0.0061, -0.4513,
-0.0621, 0.0061, -0.4502, -0.061, 0.0083, -0.4513, -0.0621, 0.0083, -0.4502, -0.0599, 0.01, -0.4523, -0.061, 0.01, -0.4513, -0.0621, 0.01, -0.4502, -0.0045, 0.0061, -0.4322, -0.0059, 0.0061, -0.4328, -0.0045, 0.0083, -0.4322, -0.0059, 0.0083, -0.4328, -0.0046, 0.01, -0.4323, -0.0058, 0.01, -0.4328, -0.0081, 0.0061, -0.4332, -0.0103, 0.0061, -0.4333, -0.0081, 0.0083, -0.4332, -0.0103, 0.0083, -0.4333, -0.0125, 0.0061, -0.4333, -0.0147, 0.0061, -0.4331, -0.0125, 0.0083, -0.4333, -0.0147, 0.0083, -0.4331, -0.0081, 0.01, -0.4332, -0.0103, 0.01, -0.4333, -0.0125, 0.01, -0.4333, -0.0147, 0.01, -0.4331, -0.0168, 0.0061, -0.4325, -0.0168, 0.0083, -0.4325, -0.0165, 0.01, -0.4326, -0.0019, 0.0061, -0.4306, -0.0019, 0.0083, -0.4306, 0.0001, 0.0061, -0.4285,
-0.0009, 0.0061, -0.4296, 0.0001, 0.0083, -0.4285, -0.0009, 0.0083, -0.4296, -0.0034, 0.0061, -0.4317, -0.0034, 0.0083, -0.4317, -0.002, 0.01, -0.4306, 0.0001, 0.01, -0.4284, -0.0008, 0.01, -0.4295, -0.0035, 0.01, -0.4317, -0.006, 0.01, -0.431, -0.0038, 0.01, -0.4288, -0.006, 0.01, -0.4288, 0.0012, 0.0061, -0.4269, 0.0012, 0.0083, -0.4269, 0.0012, 0.01, -0.4269, -0.0016, 0.01, -0.4266, 0.0005, 0.01, -0.4245, -0.0016, 0.01, -0.4245, -0.0038, 0.01, -0.4266, -0.006, 0.01, -0.4266, -0.0038, 0.01, -0.4245, -0.006, 0.01, -0.4245, -0.0081, 0.01, -0.431, -0.0103, 0.01, -0.431, -0.0081, 0.01, -0.4288, -0.0103, 0.01, -0.4288, -0.0125, 0.01, -0.431, -0.0146, 0.01, -0.431, -0.0125, 0.01, -0.4288, -0.0146, 0.01, -0.4288,
-0.0081, 0.01, -0.4266, -0.0103, 0.01, -0.4266, -0.0081, 0.01, -0.4245, -0.0103, 0.01, -0.4245, -0.0125, 0.01, -0.4266, -0.0146, 0.01, -0.4266, -0.0125, 0.01, -0.4245, -0.0146, 0.01, -0.4245, 0.0005, 0.01, -0.4223, -0.0016, 0.01, -0.4223, 0.0005, 0.01, -0.4201, -0.0016, 0.01, -0.4201, -0.0038, 0.01, -0.4223, -0.006, 0.01, -0.4223, -0.0038, 0.01, -0.4201, -0.006, 0.01, -0.4201, 0.0005, 0.01, -0.418, -0.0016, 0.01, -0.418, 0.0005, 0.01, -0.4158, -0.0016, 0.01, -0.4158, -0.0038, 0.01, -0.418, -0.006, 0.01, -0.418, -0.0038, 0.01, -0.4158, -0.006, 0.01, -0.4158, -0.0081, 0.01, -0.4223, -0.0103, 0.01, -0.4223, -0.0081, 0.01, -0.4201, -0.0103, 0.01, -0.4201, -0.0125, 0.01, -0.4223, -0.0146, 0.01, -0.4223,
-0.0125, 0.01, -0.4201, -0.0146, 0.01, -0.4201, -0.0081, 0.01, -0.418, -0.0103, 0.01, -0.418, -0.0081, 0.01, -0.4158, -0.0103, 0.01, -0.4158, -0.0125, 0.01, -0.418, -0.0146, 0.01, -0.418, -0.0125, 0.01, -0.4158, -0.0146, 0.01, -0.4158, -0.0179, 0.0061, -0.4321, -0.0191, 0.0061, -0.4315, -0.0179, 0.0083, -0.4321, -0.0191, 0.0083, -0.4315, -0.0206, 0.0061, -0.4304, -0.0206, 0.0083, -0.4304, -0.0217, 0.0061, -0.4293, -0.0226, 0.0061, -0.4282, -0.0217, 0.0083, -0.4293, -0.0226, 0.0083, -0.4282, -0.0175, 0.01, -0.4322, -0.0194, 0.01, -0.4314, -0.0168, 0.01, -0.4288, -0.019, 0.01, -0.4288, -0.0206, 0.01, -0.4304, -0.0217, 0.01, -0.4293, -0.0226, 0.01, -0.4281, -0.0236, 0.0061, -0.4267, -0.0236, 0.0083, -0.4267, -0.0242, 0.0061, -0.4249,
-0.0242, 0.0083, -0.4249, -0.0168, 0.01, -0.4266, -0.019, 0.01, -0.4266, -0.0168, 0.01, -0.4245, -0.019, 0.01, -0.4245, -0.0212, 0.01, -0.4266, -0.0234, 0.01, -0.427, -0.0212, 0.01, -0.4245, -0.0243, 0.01, -0.4247, -0.0246, 0.0061, -0.4238, -0.0246, 0.0083, -0.4238, -0.0246, 0.01, -0.4237, -0.0168, 0.01, -0.4223, -0.019, 0.01, -0.4223, -0.0168, 0.01, -0.4201, -0.019, 0.01, -0.4201, -0.0212, 0.01, -0.4223, -0.0233, 0.01, -0.4223, -0.0212, 0.01, -0.4201, -0.0233, 0.01, -0.4201, -0.0244, 0.0061, -0.4148, -0.0244, 0.0083, -0.4148, -0.0168, 0.01, -0.418, -0.019, 0.01, -0.418, -0.0168, 0.01, -0.4158, -0.019, 0.01, -0.4158, -0.0212, 0.01, -0.418, -0.0233, 0.01, -0.418, -0.0212, 0.01, -0.4158, -0.0246, 0.01, -0.4153,
-0.0249, 0.0061, -0.4223, -0.0249, 0.0083, -0.4223, -0.0251, 0.0061, -0.4201, -0.0251, 0.0083, -0.4201, -0.0249, 0.01, -0.4223, -0.0251, 0.01, -0.4201, -0.025, 0.0061, -0.418, -0.025, 0.0083, -0.418, -0.0247, 0.0061, -0.4159, -0.0247, 0.0083, -0.4159, -0.025, 0.01, -0.4179, -0.0247, 0.01, -0.4162, -0.035, 0.0061, -0.4493, -0.0363, 0.0061, -0.4487, -0.035, 0.0083, -0.4493, -0.0363, 0.0083, -0.4487, -0.0387, 0.0061, -0.4478, -0.0407, 0.0061, -0.4483, -0.0387, 0.0083, -0.4478, -0.0407, 0.0083, -0.4483, -0.0385, 0.0061, -0.4462, -0.0414, 0.0061, -0.4471, -0.0385, 0.0083, -0.4462, -0.0414, 0.0083, -0.4471, -0.035, 0.01, -0.4493, -0.0362, 0.01, -0.4487, -0.0387, 0.01, -0.4478, -0.0407, 0.01, -0.4483, -0.0385, 0.01, -0.4463, -0.0407, 0.01, -0.4464,
-0.0384, 0.0061, -0.444, -0.0384, 0.0083, -0.444, -0.0388, 0.0061, -0.4418, -0.0388, 0.0083, -0.4418, -0.0384, 0.01, -0.444, -0.0407, 0.01, -0.444, -0.0388, 0.01, -0.4417, -0.0407, 0.01, -0.4418, -0.0419, 0.0061, -0.4476, -0.0419, 0.0083, -0.4476, -0.0419, 0.0061, -0.4471, -0.0419, 0.0083, -0.4471, -0.0426, 0.01, -0.4483, -0.045, 0.01, -0.4483, -0.0419, 0.01, -0.4471, -0.045, 0.01, -0.4462, -0.0472, 0.01, -0.4483, -0.0494, 0.01, -0.4483, -0.0472, 0.01, -0.4462, -0.0494, 0.01, -0.4462, -0.0429, 0.01, -0.444, -0.045, 0.01, -0.444, -0.0429, 0.01, -0.4418, -0.045, 0.01, -0.4418, -0.0472, 0.01, -0.444, -0.0494, 0.01, -0.444, -0.0472, 0.01, -0.4418, -0.0494, 0.01, -0.4418, -0.0393, 0.0061, -0.44, -0.0398, 0.0061, -0.439,
-0.0393, 0.0083, -0.44, -0.0398, 0.0083, -0.439, -0.0402, 0.0061, -0.438, -0.0402, 0.0083, -0.438, -0.0392, 0.01, -0.4402, -0.0397, 0.01, -0.4392, -0.0406, 0.01, -0.4375, -0.0419, 0.0061, -0.4365, -0.0419, 0.0083, -0.4365, -0.0504, 0.0061, -0.4364, -0.0504, 0.0083, -0.4364, -0.0429, 0.01, -0.4397, -0.045, 0.01, -0.4397, -0.0421, 0.01, -0.4364, -0.045, 0.01, -0.4375, -0.0472, 0.01, -0.4397, -0.0494, 0.01, -0.4397, -0.0472, 0.01, -0.4375, -0.0501, 0.01, -0.4363, -0.043, 0.0061, -0.4359, -0.045, 0.0061, -0.4355, -0.043, 0.0083, -0.4359, -0.045, 0.0083, -0.4355, -0.0472, 0.0061, -0.4354, -0.0493, 0.0061, -0.4359, -0.0472, 0.0083, -0.4354, -0.0493, 0.0083, -0.4359, -0.0433, 0.01, -0.4359, -0.045, 0.01, -0.4355, -0.0473, 0.01, -0.4354,
-0.0491, 0.01, -0.4358, -0.0521, 0.0061, -0.4489, -0.0535, 0.0061, -0.4481, -0.0521, 0.0083, -0.4489, -0.0535, 0.0083, -0.4481, -0.0538, 0.0061, -0.4462, -0.0538, 0.0083, -0.4462, -0.0548, 0.0061, -0.4473, -0.0548, 0.0083, -0.4473, -0.0556, 0.0061, -0.4461, -0.0556, 0.0083, -0.4461, -0.0521, 0.01, -0.4488, -0.0535, 0.01, -0.4481, -0.0515, 0.01, -0.4462, -0.0538, 0.01, -0.4464, -0.0547, 0.01, -0.4475, -0.058, 0.01, -0.4483, -0.0556, 0.01, -0.446, -0.058, 0.01, -0.4462, -0.0543, 0.0061, -0.444, -0.0543, 0.0083, -0.444, -0.0541, 0.0061, -0.4418, -0.0541, 0.0083, -0.4418, -0.0559, 0.0061, -0.444, -0.0559, 0.0083, -0.444, -0.0557, 0.0061, -0.4418, -0.0557, 0.0083, -0.4418, -0.0515, 0.01, -0.444, -0.0543, 0.01, -0.444, -0.0515, 0.01, -0.4418,
-0.0541, 0.01, -0.4417, -0.0559, 0.01, -0.444, -0.058, 0.01, -0.444, -0.0558, 0.01, -0.4419, -0.058, 0.01, -0.4418, -0.0632, 0.0061, -0.4491, -0.0632, 0.0083, -0.4491, -0.0642, 0.0061, -0.4481, -0.0642, 0.0083, -0.4481, -0.0653, 0.0061, -0.4468, -0.0662, 0.0061, -0.4457, -0.0653, 0.0083, -0.4468, -0.0662, 0.0083, -0.4457, -0.0602, 0.01, -0.4483, -0.0632, 0.01, -0.4491, -0.0602, 0.01, -0.4462, -0.0624, 0.01, -0.4462, -0.0642, 0.01, -0.448, -0.0653, 0.01, -0.4468, -0.0662, 0.01, -0.4457, -0.0673, 0.0061, -0.4444, -0.0673, 0.0083, -0.4444, -0.0602, 0.01, -0.444, -0.0624, 0.01, -0.444, -0.0602, 0.01, -0.4418, -0.0624, 0.01, -0.4418, -0.0646, 0.01, -0.444, -0.0673, 0.01, -0.4444, -0.0646, 0.01, -0.4418, -0.0667, 0.01, -0.4418,
-0.0544, 0.0061, -0.4395, -0.0544, 0.0083, -0.4395, -0.0518, 0.0061, -0.4373, -0.0539, 0.0061, -0.4378, -0.0518, 0.0083, -0.4373, -0.0539, 0.0083, -0.4378, -0.0551, 0.0061, -0.44, -0.0551, 0.0083, -0.44, -0.0515, 0.01, -0.4397, -0.0543, 0.01, -0.4393, -0.052, 0.01, -0.4375, -0.054, 0.0097, -0.4378, -0.055, 0.01, -0.4398, -0.058, 0.01, -0.4397, -0.0559, 0.01, -0.4375, -0.058, 0.01, -0.4375, -0.0545, 0.0061, -0.4358, -0.0545, 0.0083, -0.4358, -0.0549, 0.0061, -0.4347, -0.0549, 0.0083, -0.4347, -0.0558, 0.0061, -0.4331, -0.0558, 0.0083, -0.4331, -0.0545, 0.01, -0.436, -0.0548, 0.01, -0.4349, -0.058, 0.01, -0.4353, -0.056, 0.01, -0.4329, -0.058, 0.01, -0.4332, -0.0602, 0.01, -0.4397, -0.0624, 0.01, -0.4397, -0.0602, 0.01, -0.4375,
-0.0624, 0.01, -0.4375, -0.0646, 0.01, -0.4397, -0.0667, 0.01, -0.4397, -0.0646, 0.01, -0.4375, -0.0667, 0.01, -0.4375, -0.0602, 0.01, -0.4353, -0.0624, 0.01, -0.4353, -0.0602, 0.01, -0.4332, -0.0624, 0.01, -0.4332, -0.0646, 0.01, -0.4353, -0.0667, 0.01, -0.4353, -0.0646, 0.01, -0.4332, -0.0667, 0.01, -0.4332, -0.0568, 0.0061, -0.4318, -0.0576, 0.0061, -0.4307, -0.0568, 0.0083, -0.4318, -0.0576, 0.0083, -0.4307, -0.0587, 0.0061, -0.4294, -0.0587, 0.0083, -0.4294, -0.0568, 0.01, -0.4319, -0.0576, 0.01, -0.4307, -0.0587, 0.01, -0.4294, -0.0596, 0.0061, -0.4283, -0.0596, 0.0083, -0.4283, -0.0602, 0.01, -0.431, -0.0624, 0.01, -0.431, -0.0596, 0.01, -0.4283, -0.0624, 0.01, -0.4288, -0.0646, 0.01, -0.431, -0.0667, 0.01, -0.431,
-0.0646, 0.01, -0.4288, -0.0667, 0.01, -0.4288, -0.0607, 0.0061, -0.427, -0.0616, 0.0061, -0.4259, -0.0607, 0.0083, -0.427, -0.0616, 0.0083, -0.4259, -0.0627, 0.0061, -0.4245, -0.0627, 0.0083, -0.4245, -0.0607, 0.01, -0.4271, -0.0615, 0.01, -0.426, -0.0628, 0.01, -0.4243, -0.0646, 0.01, -0.4266, -0.0667, 0.01, -0.4266, -0.0646, 0.01, -0.4245, -0.0667, 0.01, -0.4245, -0.0635, 0.0061, -0.4234, -0.0635, 0.0083, -0.4234, -0.0642, 0.0061, -0.4223, -0.0642, 0.0083, -0.4223, -0.0653, 0.0061, -0.4206, -0.066, 0.0061, -0.4195, -0.0653, 0.0083, -0.4206, -0.066, 0.0083, -0.4195, -0.0635, 0.01, -0.4234, -0.0642, 0.01, -0.4222, -0.0667, 0.01, -0.4223, -0.0652, 0.01, -0.4207, -0.0659, 0.01, -0.4197, -0.0669, 0.0061, -0.418, -0.0669, 0.0083, -0.418,
-0.0677, 0.0061, -0.4167, -0.0677, 0.0083, -0.4167, -0.067, 0.01, -0.4178, -0.0677, 0.01, -0.4167, -0.0682, 0.0061, -0.4434, -0.0682, 0.0083, -0.4434, -0.0693, 0.0061, -0.442, -0.0701, 0.0061, -0.441, -0.0693, 0.0083, -0.442, -0.0701, 0.0083, -0.441, -0.0682, 0.01, -0.4433, -0.0693, 0.01, -0.4421, -0.0702, 0.01, -0.4409, -0.0711, 0.0061, -0.4397, -0.0711, 0.0083, -0.4397, -0.0721, 0.0061, -0.4384, -0.0721, 0.0083, -0.4384, -0.0728, 0.0061, -0.4374, -0.0728, 0.0083, -0.4374, -0.0689, 0.01, -0.4397, -0.0709, 0.01, -0.4399, -0.0689, 0.01, -0.4375, -0.0721, 0.01, -0.4383, -0.0729, 0.01, -0.4373, -0.0739, 0.0061, -0.4358, -0.0746, 0.0061, -0.4347, -0.0739, 0.0083, -0.4358, -0.0746, 0.0083, -0.4347, -0.0755, 0.0061, -0.4332, -0.0755, 0.0083, -0.4332,
-0.0689, 0.01, -0.4353, -0.0711, 0.01, -0.4353, -0.0689, 0.01, -0.4332, -0.0711, 0.01, -0.4332, -0.0739, 0.01, -0.4357, -0.0747, 0.01, -0.4346, -0.0732, 0.01, -0.4332, -0.0754, 0.01, -0.4334, -0.0868, 0.0061, -0.4321, -0.0882, 0.0061, -0.4322, -0.0868, 0.0083, -0.4321, -0.0882, 0.0083, -0.4322, -0.087, 0.01, -0.4321, -0.0879, 0.01, -0.4322, -0.0763, 0.0061, -0.4309, -0.0763, 0.0083, -0.4309, -0.0758, 0.0061, -0.4288, -0.0758, 0.0083, -0.4288, -0.0689, 0.01, -0.431, -0.0711, 0.01, -0.431, -0.0689, 0.01, -0.4288, -0.0711, 0.01, -0.4288, -0.0732, 0.01, -0.431, -0.0763, 0.01, -0.4309, -0.0732, 0.01, -0.4288, -0.0757, 0.01, -0.4286, -0.0742, 0.0061, -0.4257, -0.0753, 0.0061, -0.4266, -0.0742, 0.0083, -0.4257, -0.0753, 0.0083, -0.4266,
-0.0742, 0.0061, -0.4242, -0.0748, 0.0061, -0.4246, -0.0742, 0.0083, -0.4242, -0.0748, 0.0083, -0.4246, -0.0689, 0.01, -0.4266, -0.0711, 0.01, -0.4266, -0.0689, 0.01, -0.4245, -0.0711, 0.01, -0.4245, -0.0744, 0.01, -0.426, -0.0754, 0.01, -0.4266, -0.0741, 0.01, -0.4239, -0.0748, 0.01, -0.4246, -0.0801, 0.0061, -0.4303, -0.0801, 0.0083, -0.4303, -0.0777, 0.0061, -0.4286, -0.079, 0.0061, -0.4296, -0.0777, 0.0083, -0.4286, -0.079, 0.0083, -0.4296, -0.0819, 0.0061, -0.4311, -0.0841, 0.0061, -0.4318, -0.0819, 0.0083, -0.4311, -0.0841, 0.0083, -0.4318, -0.0829, 0.0061, -0.4278, -0.083, 0.0061, -0.4277, -0.0829, 0.0083, -0.4278, -0.083, 0.0083, -0.4277, -0.0802, 0.01, -0.4304, -0.0777, 0.01, -0.4286, -0.0791, 0.01, -0.4297, -0.0817, 0.01, -0.4311,
-0.0839, 0.01, -0.4317, -0.0829, 0.01, -0.4278, -0.0836, 0.01, -0.4276, -0.0767, 0.0061, -0.4275, -0.0798, 0.0061, -0.4264, -0.0767, 0.0083, -0.4275, -0.0798, 0.0083, -0.4264, -0.0778, 0.0061, -0.4246, -0.0794, 0.0061, -0.4248, -0.0778, 0.0083, -0.4246, -0.0794, 0.0083, -0.4248, -0.0847, 0.0061, -0.4285, -0.084, 0.0061, -0.4275, -0.0847, 0.0083, -0.4285, -0.084, 0.0083, -0.4275, -0.0811, 0.0061, -0.4253, -0.0811, 0.0083, -0.4253, -0.0766, 0.01, -0.4275, -0.0798, 0.01, -0.4264, -0.0778, 0.01, -0.4246, -0.0794, 0.01, -0.4248, -0.0847, 0.01, -0.4285, -0.084, 0.01, -0.4275, -0.0811, 0.01, -0.4253, -0.0841, 0.01, -0.4245, -0.0689, 0.01, -0.4223, -0.0711, 0.01, -0.4223, -0.0689, 0.01, -0.4201, -0.0711, 0.01, -0.4201, -0.0738, 0.01, -0.4222,
-0.0754, 0.01, -0.4223, -0.0734, 0.01, -0.4201, -0.0754, 0.01, -0.4201, -0.0683, 0.0061, -0.4156, -0.0683, 0.0083, -0.4156, -0.0689, 0.01, -0.418, -0.0711, 0.01, -0.418, -0.0683, 0.01, -0.4156, -0.0711, 0.01, -0.4158, -0.0733, 0.01, -0.418, -0.0754, 0.01, -0.418, -0.0736, 0.01, -0.4158, -0.0754, 0.01, -0.4158, -0.078, 0.0061, -0.4224, -0.079, 0.0061, -0.4229, -0.078, 0.0083, -0.4224, -0.079, 0.0083, -0.4229, -0.0778, 0.01, -0.422, -0.0789, 0.01, -0.4229, -0.0775, 0.01, -0.4204, -0.0798, 0.01, -0.4201, -0.0819, 0.01, -0.4223, -0.0841, 0.01, -0.4223, -0.0819, 0.01, -0.4201, -0.0841, 0.01, -0.4201, -0.0776, 0.01, -0.418, -0.0798, 0.01, -0.418, -0.0776, 0.01, -0.4158, -0.0798, 0.01, -0.4158, -0.0819, 0.01, -0.418,
-0.0841, 0.01, -0.418, -0.0819, 0.01, -0.4158, -0.0841, 0.01, -0.4158, -0.0858, 0.0061, -0.432, -0.0893, 0.0061, -0.432, -0.0858, 0.0083, -0.432, -0.0893, 0.0083, -0.432, -0.0906, 0.0061, -0.4318, -0.0928, 0.0061, -0.431, -0.0906, 0.0083, -0.4318, -0.0928, 0.0083, -0.431, -0.0862, 0.01, -0.432, -0.0887, 0.01, -0.4321, -0.0863, 0.01, -0.4288, -0.0884, 0.01, -0.4288, -0.0909, 0.01, -0.4317, -0.093, 0.01, -0.4309, -0.0906, 0.01, -0.4288, -0.0928, 0.01, -0.4288, -0.0863, 0.01, -0.4266, -0.0884, 0.01, -0.4266, -0.0863, 0.01, -0.4245, -0.0884, 0.01, -0.4245, -0.0906, 0.01, -0.4266, -0.0928, 0.01, -0.4266, -0.0906, 0.01, -0.4245, -0.0928, 0.01, -0.4245, -0.0942, 0.0061, -0.4301, -0.0942, 0.0083, -0.4301, -0.0953, 0.0061, -0.4293,
-0.0965, 0.0061, -0.4282, -0.0953, 0.0083, -0.4293, -0.0965, 0.0083, -0.4282, -0.0941, 0.01, -0.4301, -0.0953, 0.01, -0.4293, -0.0965, 0.01, -0.4282, -0.0976, 0.0061, -0.427, -0.0976, 0.0083, -0.427, -0.0984, 0.0061, -0.4259, -0.0984, 0.0083, -0.4259, -0.0993, 0.0061, -0.4245, -0.0993, 0.0083, -0.4245, -0.0949, 0.01, -0.4266, -0.0976, 0.01, -0.427, -0.0949, 0.01, -0.4245, -0.0971, 0.01, -0.4245, -0.0984, 0.01, -0.4258, -0.0992, 0.01, -0.4247, -0.0863, 0.01, -0.4223, -0.0884, 0.01, -0.4223, -0.0863, 0.01, -0.4201, -0.0884, 0.01, -0.4201, -0.0906, 0.01, -0.4223, -0.0928, 0.01, -0.4223, -0.0906, 0.01, -0.4201, -0.0928, 0.01, -0.4201, -0.0863, 0.01, -0.418, -0.0884, 0.01, -0.418, -0.0863, 0.01, -0.4158, -0.0884, 0.01, -0.4158,
-0.0906, 0.01, -0.418, -0.0928, 0.01, -0.418, -0.0906, 0.01, -0.4158, -0.0928, 0.01, -0.4158, -0.1001, 0.0061, -0.4223, -0.1001, 0.0083, -0.4223, -0.1003, 0.0061, -0.4211, -0.1005, 0.0061, -0.42, -0.1003, 0.0083, -0.4211, -0.1005, 0.0083, -0.42, -0.0949, 0.01, -0.4223, -0.0971, 0.01, -0.4223, -0.0949, 0.01, -0.4201, -0.0971, 0.01, -0.4201, -0.1, 0.01, -0.4226, -0.1005, 0.01, -0.4205, -0.1006, 0.01, -0.4197, -0.1006, 0.0061, -0.418, -0.1006, 0.0083, -0.418, -0.1003, 0.0061, -0.4154, -0.1004, 0.0061, -0.4165, -0.1003, 0.0083, -0.4154, -0.1004, 0.0083, -0.4165, -0.0949, 0.01, -0.418, -0.0971, 0.01, -0.418, -0.0949, 0.01, -0.4158, -0.0971, 0.01, -0.4158, -0.0993, 0.01, -0.418, -0.1006, 0.01, -0.418, -0.1003, 0.01, -0.4157,
-0.1004, 0.01, -0.4166, 0.0014, 0.0061, -0.413, 0.0014, 0.0083, -0.413, 0.0004, 0.0061, -0.4114, 0.0004, 0.0083, -0.4114, 0.0014, 0.01, -0.4131, -0.0016, 0.01, -0.4136, 0.0002, 0.01, -0.4112, -0.0016, 0.01, -0.4115, -0.0038, 0.01, -0.4136, -0.006, 0.01, -0.4136, -0.0038, 0.01, -0.4115, -0.006, 0.01, -0.4115, -0.0005, 0.0061, -0.4104, -0.0015, 0.0061, -0.4091, -0.0005, 0.0083, -0.4104, -0.0015, 0.0083, -0.4091, -0.0027, 0.0061, -0.4082, -0.0027, 0.0083, -0.4082, -0.0038, 0.0061, -0.4075, -0.0059, 0.0061, -0.4064, -0.0038, 0.0083, -0.4075, -0.0059, 0.0083, -0.4064, -0.0005, 0.01, -0.4104, -0.0012, 0.01, -0.4094, -0.003, 0.01, -0.4081, -0.006, 0.01, -0.4093, -0.004, 0.01, -0.4074, -0.0057, 0.01, -0.4065, -0.0081, 0.01, -0.4136,
-0.0103, 0.01, -0.4136, -0.0081, 0.01, -0.4115, -0.0103, 0.01, -0.4115, -0.0125, 0.01, -0.4136, -0.0146, 0.01, -0.4136, -0.0125, 0.01, -0.4115, -0.0146, 0.01, -0.4115, -0.0076, 0.0061, -0.4061, -0.0076, 0.0083, -0.4061, -0.0147, 0.0061, -0.4061, -0.0147, 0.0083, -0.4061, -0.0081, 0.01, -0.4093, -0.0103, 0.01, -0.4093, -0.008, 0.01, -0.406, -0.0103, 0.01, -0.4071, -0.0125, 0.01, -0.4093, -0.0146, 0.01, -0.4093, -0.0125, 0.01, -0.4071, -0.0146, 0.01, -0.4061, 0.0012, 0.0061, -0.3981, 0.0012, 0.0083, -0.3981, 0.0011, 0.01, -0.398, -0.0087, 0.0061, -0.406, -0.0103, 0.0061, -0.4059, -0.0087, 0.0083, -0.406, -0.0103, 0.0083, -0.4059, -0.0125, 0.0061, -0.4059, -0.0136, 0.0061, -0.406, -0.0125, 0.0083, -0.4059, -0.0136, 0.0083, -0.406,
-0.0088, 0.01, -0.4059, -0.0103, 0.01, -0.4059, -0.0125, 0.01, -0.4059, -0.0136, 0.01, -0.406, -0.024, 0.0061, -0.4136, -0.024, 0.0083, -0.4136, -0.0221, 0.0061, -0.4105, -0.023, 0.0061, -0.4115, -0.0221, 0.0083, -0.4105, -0.023, 0.0083, -0.4115, -0.0168, 0.01, -0.4136, -0.019, 0.01, -0.4136, -0.0168, 0.01, -0.4115, -0.019, 0.01, -0.4115, -0.0212, 0.01, -0.4136, -0.0239, 0.01, -0.4133, -0.0222, 0.01, -0.4105, -0.0231, 0.01, -0.4116, -0.0199, 0.0061, -0.4083, -0.0199, 0.0083, -0.4083, -0.0168, 0.0061, -0.4067, -0.0189, 0.0061, -0.4076, -0.0168, 0.0083, -0.4067, -0.0189, 0.0083, -0.4076, -0.0211, 0.0061, -0.4093, -0.0211, 0.0083, -0.4093, -0.0168, 0.01, -0.4093, -0.0198, 0.01, -0.4083, -0.0171, 0.01, -0.4068, -0.0187, 0.01, -0.4075,
-0.0211, 0.01, -0.4093, 0.0001, 0.0061, -0.3963, 0.0001, 0.0083, -0.3963, -0.0005, 0.0061, -0.3951, -0.0011, 0.0061, -0.394, -0.0005, 0.0083, -0.3951, -0.0011, 0.0083, -0.394, 0.0003, 0.01, -0.3966, -0.0007, 0.01, -0.3948, -0.0012, 0.01, -0.3939, -0.0022, 0.0061, -0.3921, -0.0022, 0.0083, -0.3921, -0.0028, 0.0061, -0.3911, -0.0028, 0.0083, -0.3911, -0.0036, 0.0061, -0.3897, -0.0036, 0.0083, -0.3897, 0.0005, 0.01, -0.3919, -0.0022, 0.01, -0.3921, 0.0005, 0.01, -0.3897, -0.0016, 0.01, -0.3897, -0.0029, 0.01, -0.391, -0.0035, 0.01, -0.3899, -0.0045, 0.0061, -0.3881, -0.0052, 0.0061, -0.387, -0.0045, 0.0083, -0.3881, -0.0052, 0.0083, -0.387, -0.0061, 0.0061, -0.3854, -0.0061, 0.0083, -0.3854, 0.0005, 0.01, -0.3876, -0.0016, 0.01, -0.3876,
0.0005, 0.01, -0.3854, -0.0016, 0.01, -0.3854, -0.0046, 0.01, -0.388, -0.0052, 0.01, -0.3869, -0.0038, 0.01, -0.3854, -0.0059, 0.01, -0.3857, -0.0069, 0.0061, -0.3841, -0.0069, 0.0083, -0.3841, 0.0005, 0.01, -0.3832, -0.0016, 0.01, -0.3832, 0.0005, 0.01, -0.3811, -0.0016, 0.01, -0.3811, -0.0038, 0.01, -0.3832, -0.007, 0.01, -0.3839, -0.0038, 0.01, -0.3811, -0.006, 0.01, -0.3811, -0.0075, 0.0061, -0.383, -0.0075, 0.0083, -0.383, -0.0086, 0.0061, -0.3811, -0.0093, 0.0061, -0.38, -0.0086, 0.0083, -0.3811, -0.0093, 0.0083, -0.38, -0.0076, 0.01, -0.3828, -0.0086, 0.01, -0.3812, -0.0093, 0.01, -0.38, -0.0451, 0.0061, -0.4023, -0.0451, 0.0083, -0.4023, -0.0472, 0.0061, -0.4026, -0.0489, 0.0061, -0.402, -0.0472, 0.0083, -0.4026,
-0.0489, 0.0083, -0.402, -0.0452, 0.01, -0.4024, -0.0472, 0.01, -0.4026, -0.0488, 0.01, -0.4021, -0.0429, 0.0061, -0.4011, -0.044, 0.0061, -0.4017, -0.0429, 0.0083, -0.4011, -0.044, 0.0083, -0.4017, -0.042, 0.0061, -0.3987, -0.042, 0.0083, -0.3987, -0.05, 0.0061, -0.4014, -0.05, 0.0083, -0.4014, -0.0429, 0.01, -0.4011, -0.0442, 0.01, -0.4018, -0.042, 0.01, -0.3986, -0.045, 0.01, -0.3984, -0.0472, 0.01, -0.4006, -0.05, 0.01, -0.4013, -0.0472, 0.01, -0.3984, -0.0494, 0.01, -0.3984, -0.0508, 0.0061, -0.4002, -0.0508, 0.0083, -0.4002, -0.0514, 0.0061, -0.3986, -0.0514, 0.0083, -0.3986, -0.0509, 0.01, -0.4001, -0.0514, 0.01, -0.3986, -0.0423, 0.0061, -0.396, -0.0423, 0.0083, -0.396, -0.0435, 0.0061, -0.3946, -0.0444, 0.0061, -0.3935,
-0.0435, 0.0083, -0.3946, -0.0444, 0.0083, -0.3935, -0.0472, 0.0061, -0.3928, -0.0494, 0.0061, -0.3938, -0.0472, 0.0083, -0.3928, -0.0494, 0.0083, -0.3938, -0.0423, 0.01, -0.396, -0.045, 0.01, -0.3963, -0.0434, 0.01, -0.3946, -0.0444, 0.01, -0.3935, -0.0472, 0.01, -0.3963, -0.0494, 0.01, -0.3963, -0.0472, 0.01, -0.3928, -0.0498, 0.01, -0.3941, -0.0512, 0.0061, -0.3959, -0.0512, 0.0083, -0.3959, -0.0505, 0.0061, -0.3951, -0.0505, 0.0083, -0.3951, -0.0512, 0.01, -0.3959, -0.0505, 0.01, -0.3951, 0.0005, 0.01, -0.3789, -0.0016, 0.01, -0.3789, 0.0005, 0.01, -0.3767, -0.0016, 0.01, -0.3767, -0.0038, 0.01, -0.3789, -0.006, 0.01, -0.3789, -0.0038, 0.01, -0.3767, -0.006, 0.01, -0.3767, 0.0005, 0.01, -0.3746, -0.0016, 0.01, -0.3746,
0.0005, 0.01, -0.3724, -0.0016, 0.01, -0.3724, -0.0038, 0.01, -0.3746, -0.006, 0.01, -0.3746, -0.0038, 0.01, -0.3724, -0.006, 0.01, -0.3724, -0.0099, 0.0061, -0.3789, -0.0099, 0.0083, -0.3789, -0.0109, 0.0061, -0.377, -0.0109, 0.0083, -0.377, -0.0115, 0.0061, -0.3759, -0.0115, 0.0083, -0.3759, -0.0081, 0.01, -0.3789, -0.0098, 0.01, -0.379, -0.0081, 0.01, -0.3767, -0.0109, 0.01, -0.3769, -0.0115, 0.01, -0.3758, -0.0123, 0.0061, -0.3746, -0.0123, 0.0083, -0.3746, -0.0132, 0.0061, -0.3729, -0.0138, 0.0061, -0.3718, -0.0132, 0.0083, -0.3729, -0.0138, 0.0083, -0.3718, -0.0081, 0.01, -0.3746, -0.0103, 0.01, -0.3746, -0.0081, 0.01, -0.3724, -0.0103, 0.01, -0.3724, -0.0122, 0.01, -0.3747, -0.0133, 0.01, -0.3728, -0.0139, 0.01, -0.3717,
0.0015, 0.0061, -0.3694, 0.0015, 0.0083, -0.3694, 0.0006, 0.0061, -0.3681, 0.0006, 0.0083, -0.3681, 0.0016, 0.01, -0.3695, -0.0016, 0.01, -0.3702, 0.0004, 0.01, -0.3678, -0.0016, 0.01, -0.368, -0.0038, 0.01, -0.3702, -0.006, 0.01, -0.3702, -0.0038, 0.01, -0.368, -0.006, 0.01, -0.368, -0.0003, 0.0061, -0.3667, -0.0011, 0.0061, -0.3656, -0.0003, 0.0083, -0.3667, -0.0011, 0.0083, -0.3656, -0.0021, 0.0061, -0.3639, -0.0021, 0.0083, -0.3639, -0.0029, 0.0061, -0.3628, -0.0029, 0.0083, -0.3628, -0.0003, 0.01, -0.3667, -0.001, 0.01, -0.3656, -0.0021, 0.01, -0.364, -0.0038, 0.01, -0.3659, -0.006, 0.01, -0.3659, -0.0027, 0.01, -0.363, -0.006, 0.01, -0.3637, -0.0147, 0.0061, -0.3702, -0.0147, 0.0083, -0.3702, -0.0155, 0.0061, -0.3687,
-0.0155, 0.0083, -0.3687, -0.0081, 0.01, -0.3702, -0.0103, 0.01, -0.3702, -0.0081, 0.01, -0.368, -0.0103, 0.01, -0.368, -0.0125, 0.01, -0.3702, -0.0146, 0.01, -0.3705, -0.0125, 0.01, -0.368, -0.0156, 0.01, -0.3686, -0.0081, 0.01, -0.3659, -0.0103, 0.01, -0.3659, -0.0081, 0.01, -0.3637, -0.0103, 0.01, -0.3637, -0.0125, 0.01, -0.3659, -0.0146, 0.01, -0.3659, -0.0125, 0.01, -0.3637, -0.0146, 0.01, -0.3637, -0.0161, 0.0061, -0.3677, -0.0161, 0.0083, -0.3677, -0.0162, 0.01, -0.3675, -0.0171, 0.0061, -0.3659, -0.0171, 0.0083, -0.3659, -0.0178, 0.0061, -0.3646, -0.0184, 0.0061, -0.3635, -0.0178, 0.0083, -0.3646, -0.0184, 0.0083, -0.3635, -0.0169, 0.01, -0.3662, -0.0179, 0.01, -0.3643, -0.0184, 0.01, -0.3633, -0.0037, 0.0061, -0.3615,
-0.0037, 0.0083, -0.3615, -0.0046, 0.0061, -0.36, -0.0053, 0.0061, -0.3589, -0.0046, 0.0083, -0.36, -0.0053, 0.0083, -0.3589, -0.0039, 0.01, -0.3613, -0.006, 0.01, -0.3615, -0.0046, 0.01, -0.3601, -0.0053, 0.01, -0.359, 0.0011, 0.0061, -0.3548, 0.0011, 0.0083, -0.3548, -0.0064, 0.0061, -0.3572, -0.0064, 0.0083, -0.3572, 0.001, 0.01, -0.3547, -0.0063, 0.01, -0.3574, -0.0081, 0.01, -0.3615, -0.0103, 0.01, -0.3615, -0.0081, 0.01, -0.3594, -0.0103, 0.01, -0.3594, -0.0125, 0.01, -0.3615, -0.0146, 0.01, -0.3615, -0.0125, 0.01, -0.3594, -0.0146, 0.01, -0.3594, -0.0071, 0.0061, -0.3561, -0.0071, 0.0083, -0.3561, -0.0077, 0.0061, -0.355, -0.0077, 0.0083, -0.355, -0.0069, 0.01, -0.3563, -0.0103, 0.01, -0.3572, -0.0079, 0.01, -0.3547,
-0.0103, 0.01, -0.355, -0.0125, 0.01, -0.3572, -0.0146, 0.01, -0.3572, -0.0125, 0.01, -0.355, -0.0146, 0.01, -0.355, 0.0, 0.0061, -0.3531, -0.0007, 0.0061, -0.352, 0.0, 0.0083, -0.3531, -0.0007, 0.0083, -0.352, -0.0015, 0.0061, -0.3507, -0.0015, 0.0083, -0.3507, -0.0, 0.01, -0.3531, -0.0007, 0.01, -0.352, 0.0005, 0.01, -0.3507, -0.0014, 0.01, -0.3509, -0.0025, 0.0061, -0.3492, -0.0025, 0.0083, -0.3492, -0.0031, 0.0061, -0.3481, -0.0031, 0.0083, -0.3481, -0.0042, 0.0061, -0.3463, -0.0042, 0.0083, -0.3463, 0.0005, 0.01, -0.3485, -0.0025, 0.01, -0.3491, 0.0005, 0.01, -0.3463, -0.0016, 0.01, -0.3463, -0.0032, 0.01, -0.348, -0.004, 0.01, -0.3467, -0.0088, 0.0061, -0.3533, -0.0095, 0.0061, -0.3522, -0.0088, 0.0083, -0.3533,
-0.0095, 0.0083, -0.3522, -0.0104, 0.0061, -0.3507, -0.0104, 0.0083, -0.3507, -0.0088, 0.01, -0.3534, -0.0094, 0.01, -0.3523, -0.0105, 0.01, -0.3505, -0.0125, 0.01, -0.3529, -0.0146, 0.01, -0.3529, -0.0125, 0.01, -0.3507, -0.0146, 0.01, -0.3507, -0.0112, 0.0061, -0.3493, -0.0112, 0.0083, -0.3493, -0.0119, 0.0061, -0.3483, -0.0119, 0.0083, -0.3483, -0.013, 0.0061, -0.3464, -0.0136, 0.0061, -0.3453, -0.013, 0.0083, -0.3464, -0.0136, 0.0083, -0.3453, -0.0112, 0.01, -0.3494, -0.0119, 0.01, -0.3483, -0.0146, 0.01, -0.3485, -0.0129, 0.01, -0.3465, -0.0134, 0.01, -0.3456, -0.0194, 0.0061, -0.3615, -0.0194, 0.0083, -0.3615, -0.02, 0.0061, -0.3604, -0.02, 0.0083, -0.3604, -0.0206, 0.0061, -0.3593, -0.0206, 0.0083, -0.3593, -0.0168, 0.01, -0.3615,
-0.0192, 0.01, -0.3619, -0.0168, 0.01, -0.3594, -0.0202, 0.01, -0.3601, -0.0207, 0.01, -0.3591, -0.0217, 0.0061, -0.3573, -0.0223, 0.0061, -0.3562, -0.0217, 0.0083, -0.3573, -0.0223, 0.0083, -0.3562, -0.0228, 0.0061, -0.355, -0.0228, 0.0083, -0.355, -0.0168, 0.01, -0.3572, -0.019, 0.01, -0.3572, -0.0168, 0.01, -0.355, -0.019, 0.01, -0.355, -0.0217, 0.01, -0.3573, -0.0223, 0.01, -0.3562, -0.0212, 0.01, -0.355, -0.0228, 0.01, -0.3551, -0.0238, 0.0061, -0.3529, -0.0238, 0.0083, -0.3529, -0.0243, 0.0061, -0.3516, -0.0243, 0.0083, -0.3516, -0.0168, 0.01, -0.3529, -0.019, 0.01, -0.3529, -0.0168, 0.01, -0.3507, -0.019, 0.01, -0.3507, -0.0212, 0.01, -0.3529, -0.0237, 0.01, -0.3531, -0.0212, 0.01, -0.3507, -0.0245, 0.01, -0.3513,
-0.0168, 0.01, -0.3485, -0.019, 0.01, -0.3485, -0.0168, 0.01, -0.3463, -0.019, 0.01, -0.3463, -0.0212, 0.01, -0.3485, -0.0233, 0.01, -0.3485, -0.0212, 0.01, -0.3463, -0.0233, 0.01, -0.3463, -0.0248, 0.0061, -0.3505, -0.0248, 0.0083, -0.3505, -0.0249, 0.01, -0.3503, -0.0257, 0.0061, -0.3485, -0.0257, 0.0083, -0.3485, -0.0264, 0.0061, -0.347, -0.0269, 0.0061, -0.3459, -0.0264, 0.0083, -0.347, -0.0269, 0.0083, -0.3459, -0.0256, 0.01, -0.3488, -0.0265, 0.01, -0.3468, -0.0269, 0.01, -0.3458, -0.0633, 0.0061, -0.3715, -0.0633, 0.0083, -0.3715, -0.0653, 0.0061, -0.3738, -0.0667, 0.0061, -0.3749, -0.0653, 0.0083, -0.3738, -0.0667, 0.0083, -0.3749, -0.0641, 0.0061, -0.3727, -0.0641, 0.0083, -0.3727, -0.0633, 0.01, -0.3715, -0.0653, 0.01, -0.3738,
-0.0664, 0.01, -0.3747, -0.0641, 0.01, -0.3727, -0.0667, 0.01, -0.3724, -0.0625, 0.0061, -0.3703, -0.0625, 0.0083, -0.3703, -0.0619, 0.0061, -0.3681, -0.0619, 0.0083, -0.3681, -0.0626, 0.01, -0.3704, -0.0619, 0.01, -0.3682, -0.0646, 0.01, -0.3702, -0.0667, 0.01, -0.3702, -0.0646, 0.01, -0.368, -0.0667, 0.01, -0.368, -0.0617, 0.0061, -0.3659, -0.0617, 0.0083, -0.3659, -0.062, 0.0061, -0.3636, -0.062, 0.0083, -0.3636, -0.0617, 0.01, -0.3659, -0.062, 0.01, -0.3635, -0.0646, 0.01, -0.3659, -0.0667, 0.01, -0.3659, -0.0646, 0.01, -0.3637, -0.0667, 0.01, -0.3637, -0.0406, 0.0061, -0.3548, -0.0406, 0.0083, -0.3548, -0.0408, 0.01, -0.3549, -0.0439, 0.0061, -0.3561, -0.045, 0.0061, -0.3563, -0.0439, 0.0083, -0.3561, -0.045, 0.0083, -0.3563,
-0.0427, 0.0061, -0.3558, -0.0427, 0.0083, -0.3558, -0.0472, 0.0061, -0.3563, -0.0472, 0.0083, -0.3563, -0.0483, 0.0061, -0.3561, -0.0494, 0.0061, -0.3559, -0.0483, 0.0083, -0.3561, -0.0494, 0.0083, -0.3559, -0.0439, 0.01, -0.3561, -0.045, 0.01, -0.3563, -0.0428, 0.01, -0.3558, -0.045, 0.01, -0.355, -0.0468, 0.01, -0.3564, -0.0476, 0.01, -0.3562, -0.0497, 0.01, -0.3558, -0.0372, 0.0061, -0.3502, -0.0372, 0.0083, -0.3502, -0.0389, 0.0061, -0.3533, -0.0396, 0.0061, -0.3539, -0.0389, 0.0083, -0.3533, -0.0396, 0.0083, -0.3539, -0.0377, 0.0061, -0.3513, -0.0377, 0.0083, -0.3513, -0.0372, 0.01, -0.3501, -0.0386, 0.01, -0.3529, -0.0397, 0.01, -0.354, -0.0376, 0.01, -0.3511, -0.0407, 0.01, -0.3507, -0.0366, 0.0061, -0.3485, -0.0366, 0.0083, -0.3485,
-0.0364, 0.0061, -0.3464, -0.0364, 0.0083, -0.3464, -0.0367, 0.01, -0.3486, -0.0364, 0.01, -0.3464, -0.0385, 0.01, -0.3485, -0.0407, 0.01, -0.3485, -0.0385, 0.01, -0.3463, -0.0407, 0.01, -0.3463, -0.0429, 0.01, -0.3529, -0.045, 0.01, -0.3529, -0.0429, 0.01, -0.3507, -0.045, 0.01, -0.3507, -0.0472, 0.01, -0.3529, -0.0494, 0.01, -0.3529, -0.0472, 0.01, -0.3507, -0.0494, 0.01, -0.3507, -0.0429, 0.01, -0.3485, -0.045, 0.01, -0.3485, -0.0429, 0.01, -0.3463, -0.045, 0.01, -0.3463, -0.0472, 0.01, -0.3485, -0.0494, 0.01, -0.3485, -0.0472, 0.01, -0.3463, -0.0494, 0.01, -0.3463, -0.0516, 0.0061, -0.3548, -0.0516, 0.0083, -0.3548, -0.0515, 0.01, -0.3549, -0.0629, 0.0061, -0.3617, -0.0629, 0.0083, -0.3617, -0.0622, 0.0061, -0.3625,
-0.0665, 0.0061, -0.3605, -0.0622, 0.0083, -0.3625, -0.0665, 0.0083, -0.3605, -0.0647, 0.0061, -0.3591, -0.0663, 0.0061, -0.3592, -0.0647, 0.0083, -0.3591, -0.0663, 0.0083, -0.3592, -0.0628, 0.01, -0.3618, -0.0622, 0.01, -0.3625, -0.0661, 0.01, -0.3607, -0.0645, 0.01, -0.3593, -0.0664, 0.01, -0.3591, -0.0669, 0.0061, -0.3578, -0.0669, 0.0083, -0.3578, -0.0671, 0.01, -0.3577, -0.0526, 0.0061, -0.3539, -0.0536, 0.0061, -0.3532, -0.0526, 0.0083, -0.3539, -0.0536, 0.0083, -0.3532, -0.0546, 0.0061, -0.3514, -0.0546, 0.0083, -0.3514, -0.0551, 0.0061, -0.3503, -0.0551, 0.0083, -0.3503, -0.0525, 0.01, -0.354, -0.0538, 0.01, -0.353, -0.0515, 0.01, -0.3507, -0.0547, 0.01, -0.3512, -0.0551, 0.01, -0.3501, -0.0554, 0.0061, -0.3485, -0.0554, 0.0083, -0.3485,
-0.0555, 0.0061, -0.3463, -0.0555, 0.0083, -0.3463, -0.0515, 0.01, -0.3485, -0.0537, 0.01, -0.3485, -0.0515, 0.01, -0.3463, -0.0537, 0.01, -0.3463, -0.0554, 0.01, -0.3486, -0.0555, 0.01, -0.3463, -0.0694, 0.0061, -0.4137, -0.07, 0.0061, -0.4126, -0.0694, 0.0083, -0.4137, -0.07, 0.0083, -0.4126, -0.0706, 0.0061, -0.4115, -0.0706, 0.0083, -0.4115, -0.0739, 0.0061, -0.4136, -0.0746, 0.0061, -0.4131, -0.0739, 0.0083, -0.4136, -0.0746, 0.0083, -0.4131, -0.074, 0.0061, -0.4119, -0.0747, 0.0061, -0.4114, -0.074, 0.0083, -0.4119, -0.0747, 0.0083, -0.4114, -0.0693, 0.01, -0.4139, -0.0699, 0.01, -0.4129, -0.0708, 0.01, -0.4111, -0.0738, 0.01, -0.414, -0.0745, 0.01, -0.4133, -0.074, 0.01, -0.4118, -0.0749, 0.01, -0.411, -0.0717, 0.0061, -0.4093,
-0.0717, 0.0083, -0.4093, -0.0722, 0.0061, -0.4082, -0.0722, 0.0083, -0.4082, -0.0727, 0.0061, -0.4071, -0.0727, 0.0083, -0.4071, -0.0716, 0.01, -0.4095, -0.072, 0.01, -0.4086, -0.0757, 0.01, -0.4095, -0.0728, 0.01, -0.4068, -0.0754, 0.01, -0.4071, -0.0776, 0.01, -0.4136, -0.0798, 0.01, -0.4136, -0.0776, 0.01, -0.4115, -0.0798, 0.01, -0.4115, -0.0819, 0.01, -0.4136, -0.0841, 0.01, -0.4136, -0.0819, 0.01, -0.4115, -0.0841, 0.01, -0.4115, -0.0797, 0.0061, -0.4068, -0.0797, 0.0083, -0.4068, -0.0814, 0.0061, -0.4062, -0.0814, 0.0083, -0.4062, -0.0772, 0.01, -0.4086, -0.0798, 0.01, -0.4093, -0.0784, 0.01, -0.4081, -0.0794, 0.01, -0.4069, -0.0819, 0.01, -0.4093, -0.0841, 0.01, -0.4093, -0.0816, 0.01, -0.4061, -0.0841, 0.01, -0.4071,
-0.0735, 0.0061, -0.4049, -0.0735, 0.0083, -0.4049, -0.0741, 0.0061, -0.4032, -0.0745, 0.0061, -0.4021, -0.0741, 0.0083, -0.4032, -0.0745, 0.0083, -0.4021, -0.0735, 0.01, -0.4048, -0.0754, 0.01, -0.4049, -0.074, 0.01, -0.4033, -0.0743, 0.01, -0.4024, -0.0749, 0.0061, -0.4006, -0.0749, 0.0083, -0.4006, -0.0757, 0.0061, -0.3984, -0.0757, 0.0083, -0.3984, -0.075, 0.01, -0.4004, -0.0757, 0.01, -0.3983, -0.0803, 0.0061, -0.4057, -0.0803, 0.0083, -0.4057, -0.0819, 0.0061, -0.4056, -0.0819, 0.0083, -0.4056, -0.0776, 0.01, -0.4049, -0.0802, 0.01, -0.4057, -0.0776, 0.01, -0.4028, -0.0798, 0.01, -0.4028, -0.0823, 0.01, -0.4055, -0.0842, 0.01, -0.4054, -0.0819, 0.01, -0.4028, -0.0841, 0.01, -0.4028, -0.0776, 0.01, -0.4006, -0.0798, 0.01, -0.4006,
-0.0776, 0.01, -0.3984, -0.0798, 0.01, -0.3984, -0.0819, 0.01, -0.4006, -0.0841, 0.01, -0.4006, -0.0819, 0.01, -0.3984, -0.0841, 0.01, -0.3984, -0.0863, 0.01, -0.4136, -0.0884, 0.01, -0.4136, -0.0863, 0.01, -0.4115, -0.0884, 0.01, -0.4115, -0.0906, 0.01, -0.4136, -0.0928, 0.01, -0.4136, -0.0906, 0.01, -0.4115, -0.0928, 0.01, -0.4115, -0.0928, 0.0061, -0.4064, -0.0928, 0.0083, -0.4064, -0.0863, 0.01, -0.4093, -0.0884, 0.01, -0.4093, -0.0863, 0.01, -0.4071, -0.0884, 0.01, -0.4071, -0.0906, 0.01, -0.4093, -0.0928, 0.01, -0.4093, -0.0906, 0.01, -0.4071, -0.0929, 0.01, -0.4064, -0.098, 0.0061, -0.4106, -0.098, 0.0083, -0.4106, -0.0998, 0.0061, -0.4136, -0.0998, 0.0083, -0.4136, -0.0987, 0.0061, -0.4117, -0.0987, 0.0083, -0.4117,
-0.0949, 0.01, -0.4136, -0.0971, 0.01, -0.4136, -0.0949, 0.01, -0.4115, -0.0981, 0.01, -0.4107, -0.0997, 0.01, -0.4133, -0.0988, 0.01, -0.4118, -0.0958, 0.0061, -0.4084, -0.097, 0.0061, -0.4094, -0.0958, 0.0083, -0.4084, -0.097, 0.0083, -0.4094, -0.0947, 0.0061, -0.4075, -0.0947, 0.0083, -0.4075, -0.0957, 0.01, -0.4083, -0.097, 0.01, -0.4094, -0.0946, 0.01, -0.4075, -0.0887, 0.0061, -0.4045, -0.0887, 0.0083, -0.4045, -0.089, 0.0061, -0.4028, -0.089, 0.0083, -0.4028, -0.0906, 0.0061, -0.4056, -0.0917, 0.0061, -0.406, -0.0906, 0.0083, -0.4056, -0.0917, 0.0083, -0.406, -0.0862, 0.01, -0.4045, -0.0887, 0.01, -0.4045, -0.0863, 0.01, -0.4028, -0.089, 0.01, -0.4029, -0.0907, 0.01, -0.4056, -0.0917, 0.01, -0.406, -0.0893, 0.0061, -0.4006,
-0.0893, 0.0083, -0.4006, -0.0895, 0.0061, -0.3995, -0.0895, 0.0083, -0.3995, -0.0897, 0.0061, -0.3984, -0.0897, 0.0083, -0.3984, -0.0863, 0.01, -0.4006, -0.0893, 0.01, -0.4005, -0.0863, 0.01, -0.3984, -0.0884, 0.01, -0.3984, -0.0895, 0.01, -0.3995, -0.0897, 0.01, -0.3984, -0.0763, 0.0061, -0.3966, -0.0763, 0.0083, -0.3966, -0.0762, 0.01, -0.3968, -0.0766, 0.0061, -0.3955, -0.0766, 0.0083, -0.3955, -0.077, 0.0061, -0.3941, -0.077, 0.0083, -0.3941, -0.0765, 0.01, -0.3959, -0.0798, 0.01, -0.3963, -0.077, 0.01, -0.3939, -0.0798, 0.01, -0.3941, -0.0819, 0.01, -0.3963, -0.0841, 0.01, -0.3963, -0.0819, 0.01, -0.3941, -0.0841, 0.01, -0.3941, -0.0775, 0.0061, -0.3919, -0.0775, 0.0083, -0.3919, -0.0779, 0.0061, -0.3898, -0.0779, 0.0083, -0.3898,
-0.0775, 0.01, -0.3918, -0.0798, 0.01, -0.3919, -0.0779, 0.01, -0.3897, -0.0798, 0.01, -0.3897, -0.0819, 0.01, -0.3919, -0.0841, 0.01, -0.3919, -0.0819, 0.01, -0.3897, -0.0841, 0.01, -0.3897, -0.0783, 0.0061, -0.3876, -0.0783, 0.0083, -0.3876, -0.0786, 0.0061, -0.3858, -0.0787, 0.0061, -0.3848, -0.0786, 0.0083, -0.3858, -0.0787, 0.0083, -0.3848, -0.0783, 0.01, -0.3875, -0.0798, 0.01, -0.3876, -0.0785, 0.01, -0.3861, -0.0787, 0.01, -0.3852, -0.0819, 0.01, -0.3876, -0.0841, 0.01, -0.3876, -0.0819, 0.01, -0.3854, -0.0841, 0.01, -0.3854, -0.0789, 0.0061, -0.3832, -0.0789, 0.0083, -0.3832, -0.079, 0.0061, -0.3811, -0.079, 0.0083, -0.3811, -0.0789, 0.01, -0.3832, -0.079, 0.01, -0.381, -0.0819, 0.01, -0.3832, -0.0841, 0.01, -0.3832,
-0.0819, 0.01, -0.3811, -0.0841, 0.01, -0.3811, -0.0901, 0.0061, -0.3963, -0.0901, 0.0083, -0.3963, -0.0904, 0.0061, -0.3941, -0.0904, 0.0083, -0.3941, -0.0863, 0.01, -0.3963, -0.0884, 0.01, -0.3963, -0.0863, 0.01, -0.3941, -0.0884, 0.01, -0.3941, -0.0901, 0.01, -0.3963, -0.0904, 0.01, -0.3942, -0.0908, 0.0061, -0.3919, -0.0908, 0.0083, -0.3919, -0.0912, 0.0061, -0.3898, -0.0912, 0.0083, -0.3898, -0.0863, 0.01, -0.3919, -0.0884, 0.01, -0.3919, -0.0863, 0.01, -0.3897, -0.0884, 0.01, -0.3897, -0.0908, 0.01, -0.392, -0.0911, 0.01, -0.3899, -0.0915, 0.0061, -0.3876, -0.0915, 0.0083, -0.3876, -0.0917, 0.0061, -0.3861, -0.0917, 0.0061, -0.3851, -0.0917, 0.0083, -0.3861, -0.0917, 0.0083, -0.3851, -0.0863, 0.01, -0.3876, -0.0884, 0.01, -0.3876,
-0.0863, 0.01, -0.3854, -0.0884, 0.01, -0.3854, -0.0915, 0.01, -0.3877, -0.0917, 0.01, -0.3856, -0.0918, 0.01, -0.3848, -0.0918, 0.0061, -0.3832, -0.0918, 0.0083, -0.3832, -0.0918, 0.0061, -0.3811, -0.0918, 0.0083, -0.3811, -0.0863, 0.01, -0.3832, -0.0884, 0.01, -0.3832, -0.0863, 0.01, -0.3811, -0.0884, 0.01, -0.3811, -0.0906, 0.01, -0.3832, -0.0918, 0.01, -0.3832, -0.0906, 0.01, -0.3811, -0.0918, 0.01, -0.3811, -0.0954, 0.0061, -0.3849, -0.0969, 0.0061, -0.3861, -0.0954, 0.0083, -0.3849, -0.0969, 0.0083, -0.3861, -0.0991, 0.0061, -0.3867, -0.1015, 0.0061, -0.3862, -0.0991, 0.0083, -0.3867, -0.1015, 0.0083, -0.3862, -0.0954, 0.01, -0.3849, -0.0967, 0.01, -0.386, -0.0991, 0.01, -0.3867, -0.1015, 0.01, -0.3862, -0.0941, 0.0061, -0.3832,
-0.0941, 0.0083, -0.3832, -0.0944, 0.0061, -0.3804, -0.0944, 0.0083, -0.3804, -0.0941, 0.01, -0.3832, -0.0971, 0.01, -0.3832, -0.0943, 0.01, -0.3808, -0.0971, 0.01, -0.3811, -0.0993, 0.01, -0.3832, -0.1015, 0.01, -0.3832, -0.0993, 0.01, -0.3811, -0.1015, 0.01, -0.3811, -0.1029, 0.0061, -0.3848, -0.1029, 0.0083, -0.3848, -0.1029, 0.01, -0.3847, -0.1037, 0.0061, -0.3832, -0.1037, 0.0083, -0.3832, -0.1039, 0.0061, -0.381, -0.1039, 0.0083, -0.381, -0.1037, 0.01, -0.3834, -0.1039, 0.01, -0.381, -0.0693, 0.0061, -0.3758, -0.0711, 0.0061, -0.3761, -0.0693, 0.0083, -0.3758, -0.0711, 0.0083, -0.3761, -0.0731, 0.0061, -0.376, -0.0731, 0.0083, -0.376, -0.0695, 0.01, -0.3759, -0.071, 0.01, -0.3761, -0.0728, 0.01, -0.376, -0.0682, 0.0061, -0.3755,
-0.0682, 0.0083, -0.3755, -0.0742, 0.0061, -0.3756, -0.0755, 0.0061, -0.3751, -0.0742, 0.0083, -0.3756, -0.0755, 0.0083, -0.3751, -0.0685, 0.01, -0.3757, -0.0711, 0.01, -0.3746, -0.0689, 0.01, -0.3724, -0.0711, 0.01, -0.3724, -0.0738, 0.01, -0.3758, -0.0758, 0.01, -0.3749, -0.0732, 0.01, -0.3724, -0.0754, 0.01, -0.3724, -0.0791, 0.0061, -0.3789, -0.0791, 0.0083, -0.3789, -0.0792, 0.0061, -0.3767, -0.0792, 0.0083, -0.3767, -0.0791, 0.01, -0.3788, -0.0793, 0.01, -0.3766, -0.0819, 0.01, -0.3789, -0.0841, 0.01, -0.3789, -0.0819, 0.01, -0.3767, -0.0841, 0.01, -0.3767, -0.0773, 0.0061, -0.374, -0.0798, 0.0061, -0.3745, -0.0773, 0.0083, -0.374, -0.0798, 0.0083, -0.3745, -0.0783, 0.0061, -0.3732, -0.0799, 0.0061, -0.3724, -0.0783, 0.0083, -0.3732,
-0.0799, 0.0083, -0.3724, -0.081, 0.0061, -0.3717, -0.081, 0.0083, -0.3717, -0.0772, 0.01, -0.3741, -0.0799, 0.01, -0.3744, -0.0784, 0.01, -0.3731, -0.0799, 0.01, -0.3725, -0.0819, 0.01, -0.3746, -0.0841, 0.01, -0.3746, -0.0809, 0.01, -0.372, -0.0841, 0.01, -0.3724, -0.0689, 0.01, -0.3702, -0.0711, 0.01, -0.3702, -0.0689, 0.01, -0.368, -0.0711, 0.01, -0.368, -0.0732, 0.01, -0.3702, -0.0754, 0.01, -0.3702, -0.0732, 0.01, -0.368, -0.0754, 0.01, -0.368, -0.0689, 0.01, -0.3659, -0.0711, 0.01, -0.3659, -0.0689, 0.01, -0.3637, -0.0711, 0.01, -0.3637, -0.0732, 0.01, -0.3659, -0.0754, 0.01, -0.3659, -0.0732, 0.01, -0.3637, -0.0754, 0.01, -0.3637, -0.0803, 0.0061, -0.3703, -0.0803, 0.0083, -0.3703, -0.0808, 0.0061, -0.3686,
-0.0808, 0.0083, -0.3686, -0.0816, 0.0061, -0.3702, -0.0816, 0.0083, -0.3702, -0.0818, 0.0061, -0.368, -0.0818, 0.0083, -0.368, -0.0776, 0.01, -0.3702, -0.0802, 0.01, -0.3705, -0.0776, 0.01, -0.368, -0.0808, 0.01, -0.3682, -0.0817, 0.01, -0.37, -0.0841, 0.01, -0.3702, -0.0819, 0.01, -0.3678, -0.0841, 0.01, -0.368, -0.0808, 0.0061, -0.365, -0.0808, 0.0083, -0.365, -0.0807, 0.0061, -0.3637, -0.0807, 0.0083, -0.3637, -0.082, 0.0061, -0.367, -0.082, 0.0083, -0.367, -0.0821, 0.0061, -0.3637, -0.0821, 0.0083, -0.3637, -0.0776, 0.01, -0.3659, -0.0809, 0.01, -0.3657, -0.0776, 0.01, -0.3637, -0.0807, 0.01, -0.3634, -0.082, 0.01, -0.366, -0.0841, 0.01, -0.3659, -0.0821, 0.01, -0.3638, -0.0841, 0.01, -0.3637, -0.0918, 0.0061, -0.3789,
-0.0918, 0.0083, -0.3789, -0.0918, 0.0061, -0.3767, -0.0918, 0.0083, -0.3767, -0.0863, 0.01, -0.3789, -0.0884, 0.01, -0.3789, -0.0863, 0.01, -0.3767, -0.0884, 0.01, -0.3767, -0.0906, 0.01, -0.3789, -0.0918, 0.01, -0.3789, -0.0906, 0.01, -0.3767, -0.0918, 0.01, -0.3767, -0.0918, 0.0061, -0.3746, -0.0918, 0.0083, -0.3746, -0.0918, 0.0061, -0.3724, -0.0918, 0.0083, -0.3724, -0.0863, 0.01, -0.3746, -0.0884, 0.01, -0.3746, -0.0863, 0.01, -0.3724, -0.0884, 0.01, -0.3724, -0.0906, 0.01, -0.3746, -0.0918, 0.01, -0.3746, -0.0906, 0.01, -0.3724, -0.0918, 0.01, -0.3724, -0.0952, 0.0061, -0.3789, -0.0962, 0.0061, -0.3779, -0.0952, 0.0083, -0.3789, -0.0962, 0.0083, -0.3779, -0.0972, 0.0061, -0.3773, -0.0972, 0.0083, -0.3773, -0.1021, 0.0061, -0.3782,
-0.1021, 0.0083, -0.3782, -0.0991, 0.0061, -0.3769, -0.1014, 0.0061, -0.3776, -0.0991, 0.0083, -0.3769, -0.1014, 0.0083, -0.3776, -0.0951, 0.01, -0.379, -0.0963, 0.01, -0.3779, -0.0974, 0.01, -0.3773, -0.0993, 0.01, -0.3789, -0.102, 0.01, -0.3782, -0.0991, 0.01, -0.3769, -0.1009, 0.01, -0.3774, -0.0916, 0.0061, -0.3701, -0.0917, 0.0061, -0.3712, -0.0916, 0.0083, -0.3701, -0.0917, 0.0083, -0.3712, -0.0913, 0.0061, -0.368, -0.0913, 0.0083, -0.368, -0.0863, 0.01, -0.3702, -0.0884, 0.01, -0.3702, -0.0863, 0.01, -0.368, -0.0884, 0.01, -0.368, -0.0916, 0.01, -0.3703, -0.0917, 0.01, -0.3712, -0.0913, 0.01, -0.3679, -0.0909, 0.0061, -0.3659, -0.0909, 0.0083, -0.3659, -0.0905, 0.0061, -0.3637, -0.0905, 0.0083, -0.3637, -0.0863, 0.01, -0.3659,
-0.0884, 0.01, -0.3659, -0.0863, 0.01, -0.3637, -0.0884, 0.01, -0.3637, -0.0908, 0.01, -0.3658, -0.0904, 0.01, -0.3636, -0.0679, 0.0061, -0.3605, -0.0679, 0.0083, -0.3605, -0.0704, 0.0061, -0.3588, -0.0704, 0.0083, -0.3588, -0.0765, 0.0061, -0.3583, -0.0765, 0.0083, -0.3583, -0.0682, 0.01, -0.3603, -0.0711, 0.01, -0.3615, -0.0691, 0.01, -0.3596, -0.0703, 0.01, -0.3587, -0.0732, 0.01, -0.3615, -0.0754, 0.01, -0.3615, -0.0732, 0.01, -0.3594, -0.0763, 0.01, -0.3582, -0.0689, 0.0061, -0.3571, -0.0711, 0.0061, -0.3569, -0.0689, 0.0083, -0.3571, -0.0711, 0.0083, -0.3569, -0.0733, 0.0061, -0.357, -0.0757, 0.0061, -0.3577, -0.0733, 0.0083, -0.357, -0.0757, 0.0083, -0.3577, -0.0687, 0.01, -0.3571, -0.071, 0.01, -0.3569, -0.0733, 0.01, -0.3571,
-0.0752, 0.01, -0.3575, -0.0805, 0.0061, -0.3614, -0.0805, 0.0083, -0.3614, -0.0777, 0.0061, -0.3591, -0.0793, 0.0061, -0.3596, -0.0777, 0.0083, -0.3591, -0.0793, 0.0083, -0.3596, -0.0812, 0.0061, -0.3619, -0.0812, 0.0083, -0.3619, -0.0776, 0.01, -0.3615, -0.0803, 0.01, -0.3611, -0.078, 0.01, -0.3594, -0.0795, 0.01, -0.3599, -0.0812, 0.01, -0.3618, -0.0841, 0.01, -0.3615, -0.0819, 0.01, -0.3594, -0.0841, 0.01, -0.3594, -0.0781, 0.0061, -0.357, -0.0788, 0.0061, -0.3581, -0.0781, 0.0083, -0.357, -0.0788, 0.0083, -0.3581, -0.0772, 0.0061, -0.3551, -0.0772, 0.0083, -0.3551, -0.0781, 0.01, -0.3569, -0.0787, 0.01, -0.3579, -0.0773, 0.01, -0.3553, -0.0798, 0.01, -0.355, -0.0819, 0.01, -0.3572, -0.0841, 0.01, -0.3572, -0.0819, 0.01, -0.355,
-0.0841, 0.01, -0.355, -0.0764, 0.0061, -0.3521, -0.0764, 0.0083, -0.3521, -0.076, 0.0061, -0.3507, -0.076, 0.0083, -0.3507, -0.0764, 0.01, -0.352, -0.076, 0.01, -0.3508, -0.0753, 0.0061, -0.3485, -0.0753, 0.0083, -0.3485, -0.0742, 0.0061, -0.3455, -0.0746, 0.0061, -0.3466, -0.0742, 0.0083, -0.3455, -0.0746, 0.0083, -0.3466, -0.0754, 0.01, -0.3487, -0.0742, 0.01, -0.3454, -0.0746, 0.01, -0.3465, -0.0767, 0.0061, -0.3532, -0.0767, 0.0083, -0.3532, -0.0767, 0.01, -0.353, -0.0798, 0.01, -0.3529, -0.0776, 0.01, -0.3507, -0.0798, 0.01, -0.3507, -0.0819, 0.01, -0.3529, -0.0841, 0.01, -0.3529, -0.0819, 0.01, -0.3507, -0.0841, 0.01, -0.3507, -0.0776, 0.01, -0.3485, -0.0798, 0.01, -0.3485, -0.0776, 0.01, -0.3463, -0.0798, 0.01, -0.3463,
-0.0819, 0.01, -0.3485, -0.0841, 0.01, -0.3485, -0.0819, 0.01, -0.3463, -0.0841, 0.01, -0.3463, -0.0895, 0.0061, -0.3583, -0.0895, 0.0083, -0.3583, -0.0901, 0.0061, -0.3615, -0.0901, 0.0083, -0.3615, -0.0897, 0.0061, -0.3594, -0.0897, 0.0083, -0.3594, -0.0863, 0.01, -0.3615, -0.0884, 0.01, -0.3615, -0.0863, 0.01, -0.3594, -0.0896, 0.01, -0.359, -0.09, 0.01, -0.3615, -0.0898, 0.01, -0.3597, -0.0893, 0.0061, -0.3572, -0.0893, 0.0083, -0.3572, -0.0889, 0.0061, -0.355, -0.0889, 0.0083, -0.355, -0.0863, 0.01, -0.3572, -0.0893, 0.01, -0.357, -0.0863, 0.01, -0.355, -0.0889, 0.01, -0.3549, -0.0885, 0.0061, -0.3529, -0.0885, 0.0083, -0.3529, -0.0879, 0.0061, -0.3507, -0.0879, 0.0083, -0.3507, -0.0863, 0.01, -0.3529, -0.0884, 0.01, -0.3527,
-0.0863, 0.01, -0.3507, -0.0879, 0.01, -0.3506, -0.0871, 0.0061, -0.3482, -0.0875, 0.0061, -0.3493, -0.0871, 0.0083, -0.3482, -0.0875, 0.0083, -0.3493, -0.0864, 0.0061, -0.3463, -0.0864, 0.0083, -0.3463, -0.0871, 0.01, -0.3483, -0.0875, 0.01, -0.3494, -0.0864, 0.01, -0.3461, -0.1031, 0.0061, -0.3791, -0.1031, 0.0083, -0.3791, -0.1031, 0.01, -0.3791, -0.0049, 0.0061, -0.3452, -0.0055, 0.0061, -0.3442, -0.0049, 0.0083, -0.3452, -0.0055, 0.0083, -0.3442, -0.0066, 0.0061, -0.3423, -0.0066, 0.0083, -0.3423, 0.0005, 0.01, -0.3442, -0.0016, 0.01, -0.3442, 0.0005, 0.01, -0.342, -0.0016, 0.01, -0.342, -0.005, 0.01, -0.345, -0.0056, 0.01, -0.344, -0.0038, 0.01, -0.342, -0.0066, 0.01, -0.3423, 0.0005, 0.01, -0.3398, -0.0016, 0.01, -0.3398,
0.0005, 0.01, -0.3377, -0.0016, 0.01, -0.3377, -0.0038, 0.01, -0.3398, -0.006, 0.01, -0.3398, -0.0038, 0.01, -0.3377, -0.006, 0.01, -0.3377, -0.0072, 0.0061, -0.3412, -0.0072, 0.0083, -0.3412, -0.0142, 0.0061, -0.3442, -0.0142, 0.0083, -0.3442, -0.0153, 0.0061, -0.3423, -0.0153, 0.0083, -0.3423, -0.0072, 0.01, -0.3412, -0.0144, 0.01, -0.3438, -0.0152, 0.01, -0.3424, -0.008, 0.0061, -0.3398, -0.008, 0.0083, -0.3398, -0.0089, 0.0061, -0.3381, -0.0095, 0.0061, -0.3371, -0.0089, 0.0083, -0.3381, -0.0095, 0.0083, -0.3371, -0.0079, 0.01, -0.34, -0.009, 0.01, -0.338, -0.0095, 0.01, -0.337, 0.0005, 0.01, -0.3355, -0.0016, 0.01, -0.3355, 0.0005, 0.01, -0.3333, -0.0016, 0.01, -0.3333, -0.0038, 0.01, -0.3355, -0.006, 0.01, -0.3355,
-0.0038, 0.01, -0.3333, -0.006, 0.01, -0.3333, 0.0005, 0.01, -0.3312, -0.0016, 0.01, -0.3312, 0.0005, 0.01, -0.329, -0.0016, 0.01, -0.329, -0.0038, 0.01, -0.3312, -0.006, 0.01, -0.3312, -0.0038, 0.01, -0.329, -0.006, 0.01, -0.329, -0.0102, 0.0061, -0.3355, -0.0102, 0.0083, -0.3355, -0.0111, 0.0061, -0.3337, -0.0111, 0.0083, -0.3337, -0.0116, 0.0061, -0.3327, -0.0116, 0.0083, -0.3327, -0.0081, 0.01, -0.3355, -0.0101, 0.01, -0.3357, -0.0081, 0.01, -0.3333, -0.0111, 0.01, -0.3336, -0.0116, 0.01, -0.3326, -0.0123, 0.0061, -0.3312, -0.0123, 0.0083, -0.3312, -0.0132, 0.0061, -0.3293, -0.0137, 0.0061, -0.3282, -0.0132, 0.0083, -0.3293, -0.0137, 0.0083, -0.3282, -0.0081, 0.01, -0.3312, -0.0103, 0.01, -0.3312, -0.0081, 0.01, -0.329,
-0.0103, 0.01, -0.329, -0.0122, 0.01, -0.3313, -0.0132, 0.01, -0.3292, -0.0137, 0.01, -0.3282, -0.0159, 0.0061, -0.3412, -0.0159, 0.0083, -0.3412, -0.0168, 0.01, -0.3442, -0.019, 0.01, -0.3442, -0.0158, 0.01, -0.3414, -0.019, 0.01, -0.342, -0.0212, 0.01, -0.3442, -0.0233, 0.01, -0.3442, -0.0212, 0.01, -0.342, -0.0233, 0.01, -0.342, -0.0167, 0.0061, -0.3398, -0.0167, 0.0083, -0.3398, -0.0176, 0.0061, -0.3382, -0.0181, 0.0061, -0.3371, -0.0176, 0.0083, -0.3382, -0.0181, 0.0083, -0.3371, -0.0168, 0.01, -0.3396, -0.019, 0.01, -0.3398, -0.0175, 0.01, -0.3383, -0.0181, 0.01, -0.3372, -0.0212, 0.01, -0.3398, -0.0233, 0.01, -0.3398, -0.0212, 0.01, -0.3377, -0.0233, 0.01, -0.3377, -0.0277, 0.0061, -0.3442, -0.0277, 0.0083, -0.3442,
-0.0284, 0.0061, -0.3423, -0.0284, 0.0083, -0.3423, -0.0289, 0.0061, -0.3412, -0.0289, 0.0083, -0.3412, -0.0255, 0.01, -0.3442, -0.0276, 0.01, -0.3444, -0.0255, 0.01, -0.342, -0.0285, 0.01, -0.3422, -0.0289, 0.01, -0.3411, -0.0294, 0.0061, -0.3398, -0.0294, 0.0083, -0.3398, -0.0303, 0.0061, -0.3377, -0.0303, 0.0083, -0.3377, -0.0255, 0.01, -0.3398, -0.0277, 0.01, -0.3398, -0.0255, 0.01, -0.3377, -0.0277, 0.01, -0.3377, -0.0294, 0.01, -0.34, -0.0302, 0.01, -0.3379, -0.0189, 0.0061, -0.3355, -0.0189, 0.0083, -0.3355, -0.0198, 0.0061, -0.3337, -0.0198, 0.0083, -0.3337, -0.0203, 0.0061, -0.3327, -0.0203, 0.0083, -0.3327, -0.019, 0.01, -0.3353, -0.0197, 0.01, -0.3339, -0.0212, 0.01, -0.3355, -0.0233, 0.01, -0.3355, -0.0202, 0.01, -0.3329,
-0.0233, 0.01, -0.3333, -0.0209, 0.0061, -0.3312, -0.0209, 0.0083, -0.3312, -0.0218, 0.0061, -0.329, -0.0218, 0.0083, -0.329, -0.021, 0.01, -0.3309, -0.0233, 0.01, -0.3312, -0.0218, 0.01, -0.3289, -0.0233, 0.01, -0.329, -0.0308, 0.0061, -0.3363, -0.0312, 0.0061, -0.3352, -0.0308, 0.0083, -0.3363, -0.0312, 0.0083, -0.3352, -0.0318, 0.0061, -0.3333, -0.0318, 0.0083, -0.3333, -0.0255, 0.01, -0.3355, -0.0277, 0.01, -0.3355, -0.0255, 0.01, -0.3333, -0.0277, 0.01, -0.3333, -0.0309, 0.01, -0.3359, -0.0313, 0.01, -0.335, -0.0298, 0.01, -0.3333, -0.0318, 0.01, -0.3335, -0.0325, 0.0061, -0.3312, -0.0325, 0.0083, -0.3312, -0.033, 0.0061, -0.3296, -0.033, 0.0083, -0.3296, -0.0255, 0.01, -0.3312, -0.0277, 0.01, -0.3312, -0.0255, 0.01, -0.329,
-0.0277, 0.01, -0.329, -0.0298, 0.01, -0.3312, -0.0324, 0.01, -0.3314, -0.0298, 0.01, -0.329, -0.0331, 0.01, -0.3293, 0.0014, 0.0061, -0.3261, 0.0014, 0.0083, -0.3261, 0.0004, 0.0061, -0.3246, 0.0004, 0.0083, -0.3246, 0.0015, 0.01, -0.3262, -0.0016, 0.01, -0.3268, 0.0003, 0.01, -0.3244, -0.0016, 0.01, -0.3246, -0.0038, 0.01, -0.3268, -0.006, 0.01, -0.3268, -0.0038, 0.01, -0.3246, -0.006, 0.01, -0.3246, -0.0004, 0.0061, -0.3234, -0.0011, 0.0061, -0.3223, -0.0004, 0.0083, -0.3234, -0.0011, 0.0083, -0.3223, -0.0022, 0.0061, -0.3205, -0.0022, 0.0083, -0.3205, -0.0028, 0.0061, -0.3194, -0.0028, 0.0083, -0.3194, -0.0004, 0.01, -0.3234, -0.0011, 0.01, -0.3223, -0.0021, 0.01, -0.3207, -0.0038, 0.01, -0.3225, -0.006, 0.01, -0.3225,
-0.0027, 0.01, -0.3197, -0.006, 0.01, -0.3203, -0.0144, 0.0061, -0.3268, -0.0144, 0.0083, -0.3268, -0.0153, 0.0061, -0.3247, -0.0153, 0.0083, -0.3247, -0.0081, 0.01, -0.3268, -0.0103, 0.01, -0.3268, -0.0081, 0.01, -0.3246, -0.0103, 0.01, -0.3246, -0.0125, 0.01, -0.3268, -0.0143, 0.01, -0.327, -0.0125, 0.01, -0.3246, -0.0151, 0.01, -0.3249, -0.0157, 0.0061, -0.3235, -0.0157, 0.0083, -0.3235, -0.0081, 0.01, -0.3225, -0.0103, 0.01, -0.3225, -0.0081, 0.01, -0.3203, -0.0103, 0.01, -0.3203, -0.0125, 0.01, -0.3225, -0.0159, 0.01, -0.323, -0.0125, 0.01, -0.3203, -0.0146, 0.01, -0.3203, -0.0036, 0.0061, -0.3181, -0.0036, 0.0083, -0.3181, -0.0045, 0.0061, -0.3164, -0.0051, 0.0061, -0.3153, -0.0045, 0.0083, -0.3164, -0.0051, 0.0083, -0.3153,
-0.0037, 0.01, -0.3179, -0.006, 0.01, -0.3181, -0.0044, 0.01, -0.3165, -0.005, 0.01, -0.3154, -0.0059, 0.0061, -0.3138, -0.0059, 0.0083, -0.3138, -0.0068, 0.0061, -0.3122, -0.0068, 0.0083, -0.3122, -0.0061, 0.01, -0.3136, -0.0067, 0.01, -0.3123, -0.0081, 0.01, -0.3181, -0.0103, 0.01, -0.3181, -0.0081, 0.01, -0.316, -0.0103, 0.01, -0.316, -0.0125, 0.01, -0.3181, -0.0146, 0.01, -0.3181, -0.0125, 0.01, -0.316, -0.0146, 0.01, -0.316, -0.0074, 0.0061, -0.3111, -0.0074, 0.0083, -0.3111, -0.0081, 0.01, -0.3138, -0.0103, 0.01, -0.3138, -0.0073, 0.01, -0.3113, -0.0103, 0.01, -0.3116, -0.0125, 0.01, -0.3138, -0.0146, 0.01, -0.3138, -0.0125, 0.01, -0.3116, -0.0146, 0.01, -0.3116, -0.0222, 0.0061, -0.3279, -0.0227, 0.0061, -0.3268,
-0.0222, 0.0083, -0.3279, -0.0227, 0.0083, -0.3268, -0.0235, 0.0061, -0.3246, -0.0235, 0.0083, -0.3246, -0.0222, 0.01, -0.3279, -0.0227, 0.01, -0.3268, -0.0236, 0.01, -0.3245, -0.0161, 0.0061, -0.3224, -0.0161, 0.0083, -0.3224, -0.0169, 0.0061, -0.3203, -0.0169, 0.0083, -0.3203, -0.0242, 0.0061, -0.3229, -0.0242, 0.0083, -0.3229, -0.0162, 0.01, -0.3221, -0.0168, 0.01, -0.3205, -0.0241, 0.01, -0.3231, -0.0255, 0.01, -0.3268, -0.0277, 0.01, -0.3268, -0.0255, 0.01, -0.3246, -0.0277, 0.01, -0.3246, -0.0298, 0.01, -0.3268, -0.032, 0.01, -0.3268, -0.0298, 0.01, -0.3246, -0.032, 0.01, -0.3246, -0.0246, 0.0061, -0.3218, -0.0246, 0.0083, -0.3218, -0.0251, 0.0061, -0.3203, -0.0251, 0.0083, -0.3203, -0.0245, 0.01, -0.3221, -0.0277, 0.01, -0.3225,
-0.0251, 0.01, -0.3201, -0.0277, 0.01, -0.3203, -0.0298, 0.01, -0.3225, -0.032, 0.01, -0.3225, -0.0298, 0.01, -0.3203, -0.032, 0.01, -0.3203, -0.0176, 0.0061, -0.3184, -0.018, 0.0061, -0.3173, -0.0176, 0.0083, -0.3184, -0.018, 0.0083, -0.3173, -0.0185, 0.0061, -0.316, -0.0185, 0.0083, -0.316, -0.0176, 0.01, -0.3183, -0.018, 0.01, -0.3173, -0.0168, 0.01, -0.316, -0.0185, 0.01, -0.3161, -0.0192, 0.0061, -0.3138, -0.0192, 0.0083, -0.3138, -0.0198, 0.0061, -0.3118, -0.0198, 0.0083, -0.3118, -0.0201, 0.0061, -0.3107, -0.0201, 0.0083, -0.3107, -0.0168, 0.01, -0.3138, -0.0192, 0.01, -0.314, -0.0168, 0.01, -0.3116, -0.0199, 0.01, -0.3117, -0.0201, 0.01, -0.3107, -0.0257, 0.0061, -0.3181, -0.0257, 0.0083, -0.3181, -0.0263, 0.0061, -0.3162,
-0.0267, 0.0061, -0.3151, -0.0263, 0.0083, -0.3162, -0.0267, 0.0083, -0.3151, -0.0257, 0.01, -0.318, -0.0277, 0.01, -0.3181, -0.0262, 0.01, -0.3164, -0.0265, 0.01, -0.3155, -0.0298, 0.01, -0.3181, -0.032, 0.01, -0.3181, -0.0298, 0.01, -0.316, -0.032, 0.01, -0.316, -0.0271, 0.0061, -0.3138, -0.0271, 0.0083, -0.3138, -0.0277, 0.0061, -0.3116, -0.0277, 0.0083, -0.3116, -0.0271, 0.01, -0.3136, -0.0277, 0.01, -0.3115, -0.0298, 0.01, -0.3138, -0.032, 0.01, -0.3138, -0.0298, 0.01, -0.3116, -0.032, 0.01, -0.3116, -0.0367, 0.0061, -0.3441, -0.0367, 0.0083, -0.3441, -0.0373, 0.0061, -0.3427, -0.0373, 0.0083, -0.3427, -0.0378, 0.0061, -0.3416, -0.0378, 0.0083, -0.3416, -0.0367, 0.01, -0.3441, -0.0372, 0.01, -0.3428, -0.0385, 0.01, -0.3442,
-0.0407, 0.01, -0.3442, -0.0377, 0.01, -0.3417, -0.0407, 0.01, -0.342, -0.0388, 0.0061, -0.3401, -0.04, 0.0061, -0.339, -0.0388, 0.0083, -0.3401, -0.04, 0.0083, -0.339, -0.0411, 0.0061, -0.3384, -0.0411, 0.0083, -0.3384, -0.0388, 0.01, -0.3401, -0.0401, 0.01, -0.3389, -0.0412, 0.01, -0.3383, -0.0429, 0.01, -0.3442, -0.045, 0.01, -0.3442, -0.0429, 0.01, -0.342, -0.045, 0.01, -0.342, -0.0472, 0.01, -0.3442, -0.0494, 0.01, -0.3442, -0.0472, 0.01, -0.342, -0.0494, 0.01, -0.342, -0.0422, 0.0061, -0.3378, -0.045, 0.0061, -0.3373, -0.0422, 0.0083, -0.3378, -0.045, 0.0083, -0.3373, -0.0472, 0.0061, -0.3373, -0.0494, 0.0061, -0.3377, -0.0472, 0.0083, -0.3373, -0.0494, 0.0083, -0.3377, -0.0429, 0.01, -0.3398, -0.045, 0.01, -0.3398,
-0.0426, 0.01, -0.3376, -0.045, 0.01, -0.3373, -0.0472, 0.01, -0.3398, -0.0494, 0.01, -0.3398, -0.0473, 0.01, -0.3373, -0.0496, 0.01, -0.3378, -0.0333, 0.0061, -0.3286, -0.0333, 0.0083, -0.3286, -0.0334, 0.01, -0.3284, -0.0544, 0.0061, -0.3418, -0.0544, 0.0083, -0.3418, -0.0553, 0.0061, -0.3441, -0.0553, 0.0083, -0.3441, -0.0548, 0.0061, -0.343, -0.0548, 0.0083, -0.343, -0.0515, 0.01, -0.3442, -0.0537, 0.01, -0.3442, -0.0515, 0.01, -0.342, -0.0544, 0.01, -0.3418, -0.0553, 0.01, -0.3441, -0.0548, 0.01, -0.343, -0.0521, 0.0061, -0.3392, -0.0532, 0.0061, -0.3403, -0.0521, 0.0083, -0.3392, -0.0532, 0.0083, -0.3403, -0.0509, 0.0061, -0.3385, -0.0509, 0.0083, -0.3385, -0.052, 0.01, -0.3391, -0.0532, 0.01, -0.3403, -0.0508, 0.01, -0.3385,
-0.0673, 0.0061, -0.3311, -0.0673, 0.0083, -0.3311, -0.0655, 0.0061, -0.3282, -0.0662, 0.0061, -0.3292, -0.0655, 0.0083, -0.3282, -0.0662, 0.0083, -0.3292, -0.0672, 0.01, -0.3309, -0.0654, 0.01, -0.3281, -0.0662, 0.01, -0.3292, -0.0339, 0.0061, -0.3268, -0.0339, 0.0083, -0.3268, -0.0345, 0.0061, -0.3247, -0.0345, 0.0083, -0.3247, -0.0338, 0.01, -0.3269, -0.0344, 0.01, -0.3248, -0.035, 0.0061, -0.3225, -0.035, 0.0083, -0.3225, -0.0352, 0.0061, -0.3212, -0.0354, 0.0061, -0.3201, -0.0352, 0.0083, -0.3212, -0.0354, 0.0083, -0.3201, -0.0349, 0.01, -0.3227, -0.0353, 0.01, -0.3206, -0.0355, 0.01, -0.3198, -0.0358, 0.0061, -0.3181, -0.0358, 0.0083, -0.3181, -0.0363, 0.0061, -0.316, -0.0363, 0.0083, -0.316, -0.0342, 0.01, -0.3181, -0.0358, 0.01, -0.3182,
-0.0342, 0.01, -0.316, -0.0363, 0.01, -0.3161, -0.0366, 0.0061, -0.3138, -0.0366, 0.0083, -0.3138, -0.0369, 0.0061, -0.3116, -0.0369, 0.0083, -0.3116, -0.0342, 0.01, -0.3138, -0.0366, 0.01, -0.3139, -0.0342, 0.01, -0.3116, -0.0368, 0.01, -0.3117, -0.0504, 0.0061, -0.3106, -0.0504, 0.0083, -0.3106, -0.0504, 0.01, -0.3106, -0.059, 0.0061, -0.3193, -0.059, 0.0083, -0.3193, -0.059, 0.01, -0.3193, -0.0629, 0.0061, -0.3243, -0.0629, 0.0083, -0.3243, -0.0646, 0.0061, -0.3268, -0.0646, 0.0083, -0.3268, -0.0637, 0.0061, -0.3254, -0.0637, 0.0083, -0.3254, -0.0629, 0.01, -0.3243, -0.0648, 0.01, -0.3271, -0.0667, 0.01, -0.3268, -0.0636, 0.01, -0.3253, -0.0667, 0.01, -0.3246, -0.061, 0.0061, -0.3217, -0.0619, 0.0061, -0.3228, -0.061, 0.0083, -0.3217,
-0.0619, 0.0083, -0.3228, -0.06, 0.0061, -0.3204, -0.06, 0.0083, -0.3204, -0.061, 0.01, -0.3217, -0.0619, 0.01, -0.3228, -0.06, 0.01, -0.3204, -0.0624, 0.01, -0.3203, -0.0646, 0.01, -0.3225, -0.0667, 0.01, -0.3225, -0.0646, 0.01, -0.3203, -0.0667, 0.01, -0.3203, -0.0548, 0.0061, -0.3149, -0.0548, 0.0083, -0.3149, -0.0569, 0.0061, -0.3171, -0.058, 0.0061, -0.3182, -0.0569, 0.0083, -0.3171, -0.058, 0.0083, -0.3182, -0.0558, 0.0061, -0.316, -0.0558, 0.0083, -0.316, -0.0548, 0.01, -0.3149, -0.0569, 0.01, -0.3171, -0.058, 0.01, -0.3182, -0.0558, 0.01, -0.316, -0.058, 0.01, -0.316, -0.0526, 0.0061, -0.3127, -0.0537, 0.0061, -0.3138, -0.0526, 0.0083, -0.3127, -0.0537, 0.0083, -0.3138, -0.0515, 0.0061, -0.3117, -0.0515, 0.0083, -0.3117,
-0.0526, 0.01, -0.3127, -0.0537, 0.01, -0.3138, -0.0515, 0.01, -0.3117, -0.0537, 0.01, -0.3116, -0.0559, 0.01, -0.3138, -0.058, 0.01, -0.3138, -0.0559, 0.01, -0.3116, -0.058, 0.01, -0.3116, -0.0602, 0.01, -0.3181, -0.0624, 0.01, -0.3181, -0.0602, 0.01, -0.316, -0.0624, 0.01, -0.316, -0.0646, 0.01, -0.3181, -0.0667, 0.01, -0.3181, -0.0646, 0.01, -0.316, -0.0667, 0.01, -0.316, -0.0676, 0.0061, -0.313, -0.0676, 0.0083, -0.313, -0.0655, 0.0061, -0.3107, -0.0665, 0.0061, -0.3118, -0.0655, 0.0083, -0.3107, -0.0665, 0.0083, -0.3118, -0.0602, 0.01, -0.3138, -0.0624, 0.01, -0.3138, -0.0602, 0.01, -0.3116, -0.0624, 0.01, -0.3116, -0.0646, 0.01, -0.3138, -0.0676, 0.01, -0.313, -0.0655, 0.01, -0.3107, -0.0665, 0.01, -0.3118,
-0.0082, 0.0061, -0.3095, -0.0082, 0.0083, -0.3095, -0.0089, 0.0061, -0.3077, -0.0094, 0.0061, -0.3066, -0.0089, 0.0083, -0.3077, -0.0094, 0.0083, -0.3066, -0.0083, 0.01, -0.3093, -0.0103, 0.01, -0.3095, -0.0089, 0.01, -0.3078, -0.0093, 0.01, -0.3069, -0.0125, 0.01, -0.3095, -0.0146, 0.01, -0.3095, -0.0125, 0.01, -0.3073, -0.0146, 0.01, -0.3073, -0.01, 0.0061, -0.3051, -0.01, 0.0083, -0.3051, -0.0108, 0.0061, -0.3029, -0.0108, 0.0083, -0.3029, -0.0101, 0.01, -0.3049, -0.0109, 0.01, -0.3029, -0.0125, 0.01, -0.3051, -0.0146, 0.01, -0.3051, -0.0125, 0.01, -0.3029, -0.0146, 0.01, -0.3029, -0.0113, 0.0061, -0.3017, -0.0113, 0.0083, -0.3017, -0.0118, 0.0061, -0.3006, -0.0118, 0.0083, -0.3006, -0.0125, 0.0061, -0.2986, -0.0125, 0.0083, -0.2986,
-0.0113, 0.01, -0.3017, -0.0117, 0.01, -0.3007, -0.0146, 0.01, -0.3008, -0.0125, 0.01, -0.2985, -0.0146, 0.01, -0.2986, -0.0131, 0.0061, -0.2964, -0.0131, 0.0083, -0.2964, -0.0134, 0.0061, -0.2948, -0.0137, 0.0061, -0.2937, -0.0134, 0.0083, -0.2948, -0.0137, 0.0083, -0.2937, -0.0131, 0.01, -0.2964, -0.0146, 0.01, -0.2964, -0.0134, 0.01, -0.295, -0.0136, 0.01, -0.2941, -0.0205, 0.0061, -0.3095, -0.0205, 0.0083, -0.3095, -0.021, 0.0061, -0.3073, -0.021, 0.0083, -0.3073, -0.0168, 0.01, -0.3095, -0.019, 0.01, -0.3095, -0.0168, 0.01, -0.3073, -0.019, 0.01, -0.3073, -0.0205, 0.01, -0.3095, -0.021, 0.01, -0.3074, -0.0215, 0.0061, -0.3051, -0.0215, 0.0083, -0.3051, -0.022, 0.0061, -0.3029, -0.022, 0.0083, -0.3029, -0.0168, 0.01, -0.3051,
-0.019, 0.01, -0.3051, -0.0168, 0.01, -0.3029, -0.019, 0.01, -0.3029, -0.0215, 0.01, -0.3053, -0.0219, 0.01, -0.3031, -0.0282, 0.0061, -0.3095, -0.0282, 0.0083, -0.3095, -0.0286, 0.0061, -0.3074, -0.0286, 0.0083, -0.3074, -0.0288, 0.0061, -0.3063, -0.0288, 0.0083, -0.3063, -0.0282, 0.01, -0.3094, -0.0285, 0.01, -0.3077, -0.0298, 0.01, -0.3095, -0.032, 0.01, -0.3095, -0.0287, 0.01, -0.3069, -0.032, 0.01, -0.3073, -0.029, 0.0061, -0.3051, -0.029, 0.0083, -0.3051, -0.0293, 0.0061, -0.3029, -0.0293, 0.0083, -0.3029, -0.029, 0.01, -0.305, -0.032, 0.01, -0.3051, -0.0293, 0.01, -0.3028, -0.032, 0.01, -0.3029, -0.0222, 0.0061, -0.3016, -0.0224, 0.0061, -0.3005, -0.0222, 0.0083, -0.3016, -0.0224, 0.0083, -0.3005, -0.0227, 0.0061, -0.2986,
-0.0227, 0.0083, -0.2986, -0.0168, 0.01, -0.3008, -0.019, 0.01, -0.3008, -0.0168, 0.01, -0.2986, -0.019, 0.01, -0.2986, -0.0223, 0.01, -0.301, -0.0224, 0.01, -0.3003, -0.0212, 0.01, -0.2986, -0.0227, 0.01, -0.2986, -0.023, 0.0061, -0.2964, -0.023, 0.0083, -0.2964, -0.0232, 0.0061, -0.2943, -0.0232, 0.0083, -0.2943, -0.0168, 0.01, -0.2964, -0.019, 0.01, -0.2964, -0.0168, 0.01, -0.2943, -0.019, 0.01, -0.2943, -0.0212, 0.01, -0.2964, -0.023, 0.01, -0.2965, -0.0212, 0.01, -0.2943, -0.0232, 0.01, -0.2943, -0.0297, 0.0061, -0.3008, -0.0297, 0.0083, -0.3008, -0.0299, 0.0061, -0.2986, -0.0299, 0.0083, -0.2986, -0.0297, 0.01, -0.3007, -0.032, 0.01, -0.3008, -0.03, 0.01, -0.2986, -0.032, 0.01, -0.2986, -0.03, 0.0061, -0.2964,
-0.03, 0.0083, -0.2964, -0.0301, 0.0061, -0.2943, -0.0301, 0.0083, -0.2943, -0.03, 0.01, -0.2964, -0.032, 0.01, -0.2964, -0.0301, 0.01, -0.2943, -0.032, 0.01, -0.2943, -0.0064, 0.0061, -0.2889, -0.0064, 0.0083, -0.2889, -0.0066, 0.01, -0.289, 0.0005, 0.0061, -0.2882, -0.0016, 0.0061, -0.2883, 0.0005, 0.0083, -0.2882, -0.0016, 0.0083, -0.2883, -0.0038, 0.0061, -0.2885, -0.0053, 0.0061, -0.2888, -0.0038, 0.0083, -0.2885, -0.0053, 0.0083, -0.2888, 0.0006, 0.01, -0.2882, -0.0016, 0.01, -0.2883, 0.0005, 0.01, -0.2856, -0.0016, 0.01, -0.2856, -0.0037, 0.01, -0.2885, -0.0058, 0.01, -0.2888, -0.0038, 0.01, -0.2856, -0.006, 0.01, -0.2856, -0.0081, 0.0061, -0.2892, -0.0103, 0.0061, -0.2896, -0.0081, 0.0083, -0.2892, -0.0103, 0.0083, -0.2896,
-0.014, 0.0061, -0.2921, -0.014, 0.0083, -0.2921, -0.0125, 0.0061, -0.29, -0.0144, 0.0061, -0.2902, -0.0125, 0.0083, -0.29, -0.0144, 0.0083, -0.2902, -0.0081, 0.01, -0.2892, -0.0102, 0.01, -0.2896, -0.014, 0.01, -0.2919, -0.0124, 0.01, -0.29, -0.0144, 0.01, -0.2902, -0.0081, 0.01, -0.2878, -0.0103, 0.01, -0.2878, -0.0081, 0.01, -0.2856, -0.0103, 0.01, -0.2856, -0.0125, 0.01, -0.2878, -0.0146, 0.01, -0.2878, -0.0125, 0.01, -0.2856, -0.0146, 0.01, -0.2856, -0.0069, 0.0061, -0.2802, -0.0069, 0.0083, -0.2802, 0.0005, 0.01, -0.2834, -0.0016, 0.01, -0.2834, 0.0005, 0.01, -0.2812, -0.0016, 0.01, -0.2812, -0.0038, 0.01, -0.2834, -0.006, 0.01, -0.2834, -0.0038, 0.01, -0.2812, -0.0063, 0.01, -0.2801, 0.0005, 0.0061, -0.2794,
-0.0016, 0.0061, -0.2795, 0.0005, 0.0083, -0.2794, -0.0016, 0.0083, -0.2795, -0.0038, 0.0061, -0.2797, -0.0058, 0.0061, -0.28, -0.0038, 0.0083, -0.2797, -0.0058, 0.0083, -0.28, 0.0005, 0.01, -0.2794, -0.0017, 0.01, -0.2795, -0.0038, 0.01, -0.2797, -0.0055, 0.01, -0.28, -0.0081, 0.0061, -0.2803, -0.0103, 0.0061, -0.2806, -0.0081, 0.0083, -0.2803, -0.0103, 0.0083, -0.2806, -0.0125, 0.0061, -0.281, -0.0146, 0.0061, -0.2813, -0.0125, 0.0083, -0.281, -0.0146, 0.0083, -0.2813, -0.0081, 0.01, -0.2834, -0.0103, 0.01, -0.2834, -0.0083, 0.01, -0.2804, -0.0104, 0.01, -0.2807, -0.0125, 0.01, -0.2834, -0.0146, 0.01, -0.2834, -0.0126, 0.01, -0.281, -0.0147, 0.01, -0.2813, -0.0232, 0.0061, -0.2921, -0.0232, 0.0083, -0.2921, -0.0232, 0.0061, -0.2899,
-0.0232, 0.0083, -0.2899, -0.0168, 0.01, -0.2921, -0.019, 0.01, -0.2921, -0.0168, 0.01, -0.2899, -0.019, 0.01, -0.2899, -0.0212, 0.01, -0.2921, -0.0232, 0.01, -0.2921, -0.0212, 0.01, -0.2899, -0.0232, 0.01, -0.2899, -0.0232, 0.0061, -0.2878, -0.0232, 0.0083, -0.2878, -0.0232, 0.0061, -0.2856, -0.0232, 0.0083, -0.2856, -0.0168, 0.01, -0.2878, -0.019, 0.01, -0.2878, -0.0168, 0.01, -0.2856, -0.019, 0.01, -0.2856, -0.0212, 0.01, -0.2878, -0.0232, 0.01, -0.2878, -0.0212, 0.01, -0.2856, -0.0232, 0.01, -0.2856, -0.0301, 0.0061, -0.2921, -0.0301, 0.0083, -0.2921, -0.0301, 0.0061, -0.2899, -0.0301, 0.0083, -0.2899, -0.0301, 0.01, -0.2921, -0.032, 0.01, -0.2921, -0.0301, 0.01, -0.2899, -0.032, 0.01, -0.2899, -0.0301, 0.0061, -0.2878,
-0.0301, 0.0083, -0.2878, -0.03, 0.0061, -0.2856, -0.03, 0.0083, -0.2856, -0.0301, 0.01, -0.2878, -0.032, 0.01, -0.2878, -0.03, 0.01, -0.2856, -0.032, 0.01, -0.2856, -0.0163, 0.0061, -0.2814, -0.0163, 0.0083, -0.2814, -0.0232, 0.0061, -0.2834, -0.0232, 0.0083, -0.2834, -0.0232, 0.0061, -0.2812, -0.0232, 0.0083, -0.2812, -0.0168, 0.01, -0.2834, -0.019, 0.01, -0.2834, -0.0163, 0.01, -0.2814, -0.019, 0.01, -0.2812, -0.0212, 0.01, -0.2834, -0.0232, 0.01, -0.2834, -0.0212, 0.01, -0.2812, -0.0232, 0.01, -0.2812, -0.0163, 0.0061, -0.2791, -0.0163, 0.0083, -0.2791, -0.0163, 0.0061, -0.2769, -0.0163, 0.0083, -0.2769, -0.0232, 0.0061, -0.2791, -0.0232, 0.0083, -0.2791, -0.023, 0.0061, -0.2769, -0.023, 0.0083, -0.2769, -0.0163, 0.01, -0.2791,
-0.019, 0.01, -0.2791, -0.0163, 0.01, -0.2769, -0.019, 0.01, -0.2769, -0.0212, 0.01, -0.2791, -0.0232, 0.01, -0.279, -0.0212, 0.01, -0.2769, -0.023, 0.01, -0.2769, -0.0299, 0.0061, -0.2834, -0.0299, 0.0083, -0.2834, -0.0296, 0.0061, -0.2812, -0.0296, 0.0083, -0.2812, -0.0299, 0.01, -0.2835, -0.032, 0.01, -0.2834, -0.0296, 0.01, -0.2813, -0.032, 0.01, -0.2812, -0.0286, 0.0061, -0.2764, -0.0286, 0.0083, -0.2764, -0.0292, 0.0061, -0.2791, -0.0292, 0.0083, -0.2791, -0.0289, 0.0061, -0.2774, -0.0289, 0.0083, -0.2774, -0.0286, 0.01, -0.2762, -0.0292, 0.01, -0.2792, -0.032, 0.01, -0.2791, -0.0288, 0.01, -0.2771, -0.032, 0.01, -0.2769, -0.037, 0.0061, -0.3095, -0.037, 0.0083, -0.3095, -0.0371, 0.0061, -0.3073, -0.0371, 0.0083, -0.3073,
-0.0342, 0.01, -0.3095, -0.037, 0.01, -0.3095, -0.0342, 0.01, -0.3073, -0.0371, 0.01, -0.3074, -0.0374, 0.0061, -0.3058, -0.0374, 0.0083, -0.3058, -0.0375, 0.0061, -0.3047, -0.0375, 0.0083, -0.3047, -0.038, 0.0061, -0.301, -0.0407, 0.0061, -0.3029, -0.038, 0.0083, -0.301, -0.0407, 0.0083, -0.3029, -0.0342, 0.01, -0.3051, -0.0374, 0.01, -0.3053, -0.0342, 0.01, -0.3029, -0.0363, 0.01, -0.3029, -0.0376, 0.01, -0.3045, -0.038, 0.01, -0.301, -0.0405, 0.01, -0.3027, -0.0457, 0.0061, -0.3065, -0.0457, 0.0083, -0.3065, -0.0481, 0.0061, -0.3085, -0.0493, 0.0061, -0.3096, -0.0481, 0.0083, -0.3085, -0.0493, 0.0083, -0.3096, -0.0471, 0.0061, -0.3076, -0.0471, 0.0083, -0.3076, -0.0457, 0.01, -0.3065, -0.0481, 0.01, -0.3085, -0.0493, 0.01, -0.3096,
-0.047, 0.01, -0.3076, -0.0494, 0.01, -0.3073, -0.0432, 0.0061, -0.3046, -0.0446, 0.0061, -0.3057, -0.0432, 0.0083, -0.3046, -0.0446, 0.0083, -0.3057, -0.0421, 0.0061, -0.3038, -0.0421, 0.0083, -0.3038, -0.0432, 0.01, -0.3047, -0.0447, 0.01, -0.3057, -0.0422, 0.01, -0.3039, -0.045, 0.01, -0.3029, -0.0472, 0.01, -0.3051, -0.0494, 0.01, -0.3051, -0.0472, 0.01, -0.3029, -0.0494, 0.01, -0.3029, -0.0342, 0.01, -0.3008, -0.0363, 0.01, -0.3008, -0.0342, 0.01, -0.2986, -0.0363, 0.01, -0.2986, -0.0384, 0.01, -0.3013, -0.0407, 0.01, -0.3008, -0.0385, 0.01, -0.2986, -0.0407, 0.01, -0.2986, -0.0342, 0.01, -0.2964, -0.0363, 0.01, -0.2964, -0.0342, 0.01, -0.2943, -0.0363, 0.01, -0.2943, -0.0385, 0.01, -0.2964, -0.0407, 0.01, -0.2964,
-0.0385, 0.01, -0.2943, -0.0407, 0.01, -0.2943, -0.0429, 0.01, -0.3008, -0.045, 0.01, -0.3008, -0.0429, 0.01, -0.2986, -0.045, 0.01, -0.2986, -0.0472, 0.01, -0.3008, -0.0494, 0.01, -0.3008, -0.0472, 0.01, -0.2986, -0.0494, 0.01, -0.2986, -0.0438, 0.0061, -0.2932, -0.045, 0.0061, -0.294, -0.0438, 0.0083, -0.2932, -0.045, 0.0083, -0.294, -0.0477, 0.0061, -0.2957, -0.0494, 0.0061, -0.2968, -0.0477, 0.0083, -0.2957, -0.0494, 0.0083, -0.2968, -0.0467, 0.0061, -0.295, -0.0467, 0.0083, -0.295, -0.0429, 0.01, -0.2964, -0.045, 0.01, -0.2964, -0.0436, 0.01, -0.2931, -0.0453, 0.01, -0.2942, -0.0477, 0.01, -0.2956, -0.0496, 0.01, -0.2969, -0.0466, 0.01, -0.295, -0.0515, 0.01, -0.3095, -0.0537, 0.01, -0.3095, -0.0515, 0.01, -0.3073,
-0.0537, 0.01, -0.3073, -0.0559, 0.01, -0.3095, -0.058, 0.01, -0.3095, -0.0559, 0.01, -0.3073, -0.058, 0.01, -0.3073, -0.0589, 0.0061, -0.3043, -0.0589, 0.0083, -0.3043, -0.0565, 0.0061, -0.3022, -0.0578, 0.0061, -0.3033, -0.0565, 0.0083, -0.3022, -0.0578, 0.0083, -0.3033, -0.0515, 0.01, -0.3051, -0.0537, 0.01, -0.3051, -0.0515, 0.01, -0.3029, -0.0537, 0.01, -0.3029, -0.0559, 0.01, -0.3051, -0.0588, 0.01, -0.3042, -0.0565, 0.01, -0.3022, -0.0577, 0.01, -0.3033, -0.0634, 0.0061, -0.3084, -0.0634, 0.0083, -0.3084, -0.0612, 0.0061, -0.3063, -0.0624, 0.0061, -0.3074, -0.0612, 0.0083, -0.3063, -0.0624, 0.0083, -0.3074, -0.0644, 0.0061, -0.3095, -0.0644, 0.0083, -0.3095, -0.0602, 0.01, -0.3095, -0.0634, 0.01, -0.3084, -0.0611, 0.01, -0.3063,
-0.0623, 0.01, -0.3073, -0.0645, 0.01, -0.3095, -0.0601, 0.0061, -0.3053, -0.0601, 0.0083, -0.3053, -0.0601, 0.01, -0.3053, -0.0541, 0.0061, -0.3002, -0.0541, 0.0083, -0.3002, -0.0516, 0.0061, -0.2983, -0.053, 0.0061, -0.2994, -0.0516, 0.0083, -0.2983, -0.053, 0.0083, -0.2994, -0.0554, 0.0061, -0.3013, -0.0554, 0.0083, -0.3013, -0.0515, 0.01, -0.3008, -0.0541, 0.01, -0.3002, -0.0516, 0.01, -0.2983, -0.053, 0.01, -0.2994, -0.0554, 0.01, -0.3013, -0.0505, 0.0061, -0.2975, -0.0505, 0.0083, -0.2975, -0.0505, 0.01, -0.2975, -0.0369, 0.0061, -0.2902, -0.0369, 0.0083, -0.2902, -0.0396, 0.0061, -0.291, -0.0407, 0.0061, -0.2915, -0.0396, 0.0083, -0.291, -0.0407, 0.0083, -0.2915, -0.0385, 0.0061, -0.2906, -0.0385, 0.0083, -0.2906, -0.0342, 0.01, -0.2921,
-0.0363, 0.01, -0.2921, -0.0342, 0.01, -0.2899, -0.0369, 0.01, -0.2902, -0.0391, 0.01, -0.2908, -0.041, 0.01, -0.2916, -0.0382, 0.01, -0.2905, -0.0369, 0.0061, -0.2878, -0.0369, 0.0083, -0.2878, -0.0367, 0.0061, -0.2856, -0.0367, 0.0083, -0.2856, -0.0342, 0.01, -0.2878, -0.0369, 0.01, -0.2877, -0.0342, 0.01, -0.2856, -0.0367, 0.01, -0.2855, -0.0428, 0.0061, -0.2926, -0.0428, 0.0083, -0.2926, -0.0426, 0.01, -0.2925, -0.0363, 0.0061, -0.2834, -0.0363, 0.0083, -0.2834, -0.0359, 0.0061, -0.2812, -0.0359, 0.0083, -0.2812, -0.0342, 0.01, -0.2834, -0.0363, 0.01, -0.2833, -0.0342, 0.01, -0.2812, -0.0359, 0.01, -0.2812, -0.0352, 0.0061, -0.2783, -0.0355, 0.0061, -0.2794, -0.0352, 0.0083, -0.2783, -0.0355, 0.0083, -0.2794, -0.0348, 0.0061, -0.2769,
-0.0348, 0.0083, -0.2769, -0.0353, 0.01, -0.2787, -0.0355, 0.01, -0.2796, -0.0347, 0.01, -0.2767, -0.0737, 0.0061, -0.3442, -0.0737, 0.0083, -0.3442, -0.0728, 0.0061, -0.342, -0.0728, 0.0083, -0.342, -0.0738, 0.01, -0.3443, -0.0754, 0.01, -0.3442, -0.0729, 0.01, -0.3423, -0.0754, 0.01, -0.342, -0.0717, 0.0061, -0.3397, -0.0717, 0.0083, -0.3397, -0.0707, 0.0061, -0.3377, -0.0707, 0.0083, -0.3377, -0.0722, 0.0061, -0.3408, -0.0722, 0.0083, -0.3408, -0.0716, 0.01, -0.3395, -0.0708, 0.01, -0.338, -0.0721, 0.01, -0.3405, -0.0754, 0.01, -0.3398, -0.0732, 0.01, -0.3377, -0.0754, 0.01, -0.3377, -0.0848, 0.0061, -0.3419, -0.0848, 0.0083, -0.3419, -0.0776, 0.01, -0.3442, -0.0798, 0.01, -0.3442, -0.0776, 0.01, -0.342, -0.0798, 0.01, -0.342,
-0.0819, 0.01, -0.3442, -0.0841, 0.01, -0.3442, -0.0819, 0.01, -0.342, -0.0849, 0.01, -0.3419, -0.0841, 0.0061, -0.3398, -0.0841, 0.0083, -0.3398, -0.0829, 0.0061, -0.3368, -0.0834, 0.0061, -0.3379, -0.0829, 0.0083, -0.3368, -0.0834, 0.0083, -0.3379, -0.0776, 0.01, -0.3398, -0.0798, 0.01, -0.3398, -0.0776, 0.01, -0.3377, -0.0798, 0.01, -0.3377, -0.0819, 0.01, -0.3398, -0.0841, 0.01, -0.3396, -0.0831, 0.01, -0.3371, -0.0835, 0.01, -0.3381, -0.0695, 0.0061, -0.3353, -0.0701, 0.0061, -0.3364, -0.0695, 0.0083, -0.3353, -0.0701, 0.0083, -0.3364, -0.0685, 0.0061, -0.3333, -0.0685, 0.0083, -0.3333, -0.0694, 0.01, -0.3351, -0.0699, 0.01, -0.3361, -0.0687, 0.01, -0.3336, -0.0711, 0.01, -0.3333, -0.0732, 0.01, -0.3355, -0.0754, 0.01, -0.3355,
-0.0732, 0.01, -0.3333, -0.0754, 0.01, -0.3333, -0.0679, 0.0061, -0.3321, -0.0679, 0.0083, -0.3321, -0.0677, 0.01, -0.3319, -0.0711, 0.01, -0.3312, -0.0689, 0.01, -0.329, -0.0711, 0.01, -0.329, -0.0732, 0.01, -0.3312, -0.0754, 0.01, -0.3312, -0.0732, 0.01, -0.329, -0.0754, 0.01, -0.329, -0.0808, 0.0061, -0.3324, -0.0808, 0.0083, -0.3324, -0.0823, 0.0061, -0.3355, -0.0823, 0.0083, -0.3355, -0.0813, 0.0061, -0.3335, -0.0813, 0.0083, -0.3335, -0.0776, 0.01, -0.3355, -0.0798, 0.01, -0.3355, -0.0776, 0.01, -0.3333, -0.0809, 0.01, -0.3327, -0.0822, 0.01, -0.3352, -0.0814, 0.01, -0.3336, -0.0802, 0.0061, -0.3311, -0.0802, 0.0083, -0.3311, -0.0785, 0.0061, -0.3282, -0.0791, 0.0061, -0.3293, -0.0785, 0.0083, -0.3282, -0.0791, 0.0083, -0.3293,
-0.0776, 0.01, -0.3312, -0.08, 0.01, -0.3308, -0.0786, 0.01, -0.3284, -0.0792, 0.01, -0.3294, -0.0856, 0.0061, -0.3442, -0.0856, 0.0083, -0.3442, -0.0852, 0.0061, -0.343, -0.0852, 0.0083, -0.343, -0.0856, 0.01, -0.3441, -0.0852, 0.01, -0.343, -0.0761, 0.0061, -0.3242, -0.0761, 0.0083, -0.3242, -0.0689, 0.01, -0.3268, -0.0711, 0.01, -0.3268, -0.0689, 0.01, -0.3246, -0.0711, 0.01, -0.3246, -0.0732, 0.01, -0.3268, -0.0754, 0.01, -0.3268, -0.0732, 0.01, -0.3246, -0.0761, 0.01, -0.3243, -0.0742, 0.0061, -0.3215, -0.075, 0.0061, -0.3226, -0.0742, 0.0083, -0.3215, -0.075, 0.0083, -0.3226, -0.0734, 0.0061, -0.3203, -0.0734, 0.0083, -0.3203, -0.0689, 0.01, -0.3225, -0.0711, 0.01, -0.3225, -0.0689, 0.01, -0.3203, -0.0711, 0.01, -0.3203,
-0.0743, 0.01, -0.3217, -0.0751, 0.01, -0.3227, -0.0732, 0.01, -0.32, -0.0777, 0.0061, -0.3268, -0.0777, 0.0083, -0.3268, -0.0768, 0.0061, -0.3253, -0.0768, 0.0083, -0.3253, -0.0775, 0.01, -0.3265, -0.0768, 0.01, -0.3254, -0.0716, 0.0061, -0.3178, -0.0716, 0.0083, -0.3178, -0.0696, 0.0061, -0.3153, -0.0705, 0.0061, -0.3164, -0.0696, 0.0083, -0.3153, -0.0705, 0.0083, -0.3164, -0.0724, 0.0061, -0.3189, -0.0724, 0.0083, -0.3189, -0.0689, 0.01, -0.3181, -0.0716, 0.01, -0.3178, -0.0696, 0.01, -0.3153, -0.0706, 0.01, -0.3164, -0.0724, 0.01, -0.3189, -0.0685, 0.0061, -0.3141, -0.0685, 0.0083, -0.3141, -0.0686, 0.01, -0.3141, -0.0157, 0.0061, -0.2653, -0.0157, 0.0083, -0.2653, -0.0154, 0.0061, -0.2639, -0.0154, 0.0083, -0.2639, -0.0157, 0.01, -0.2652,
-0.0154, 0.01, -0.2639, -0.015, 0.0061, -0.2617, -0.015, 0.0083, -0.2617, -0.0146, 0.0061, -0.2595, -0.0146, 0.0083, -0.2595, -0.015, 0.01, -0.2618, -0.0146, 0.01, -0.2596, -0.0163, 0.0061, -0.2747, -0.0163, 0.0083, -0.2747, -0.0163, 0.0061, -0.2726, -0.0163, 0.0083, -0.2726, -0.0226, 0.0061, -0.2747, -0.0226, 0.0083, -0.2747, -0.022, 0.0061, -0.2723, -0.0223, 0.0061, -0.2734, -0.022, 0.0083, -0.2723, -0.0223, 0.0083, -0.2734, -0.0163, 0.01, -0.2747, -0.019, 0.01, -0.2747, -0.0163, 0.01, -0.2726, -0.019, 0.01, -0.2726, -0.0212, 0.01, -0.2747, -0.0226, 0.01, -0.2747, -0.0221, 0.01, -0.2725, -0.0223, 0.01, -0.2735, -0.0163, 0.0061, -0.2704, -0.0163, 0.0083, -0.2704, -0.0161, 0.0061, -0.2682, -0.0161, 0.0083, -0.2682, -0.0216, 0.0061, -0.2704,
-0.0216, 0.0083, -0.2704, -0.0211, 0.0061, -0.2682, -0.0211, 0.0083, -0.2682, -0.0163, 0.01, -0.2704, -0.019, 0.01, -0.2704, -0.0161, 0.01, -0.2683, -0.019, 0.01, -0.2682, -0.0216, 0.01, -0.2702, -0.0211, 0.01, -0.2681, -0.0283, 0.0061, -0.2747, -0.0283, 0.0083, -0.2747, -0.0277, 0.0061, -0.2726, -0.0277, 0.0083, -0.2726, -0.033, 0.0061, -0.2717, -0.033, 0.0083, -0.2717, -0.0283, 0.01, -0.2748, -0.0277, 0.01, -0.2727, -0.0298, 0.01, -0.2747, -0.032, 0.01, -0.2747, -0.0298, 0.01, -0.2726, -0.0331, 0.01, -0.2721, -0.027, 0.0061, -0.2704, -0.027, 0.0083, -0.2704, -0.0263, 0.0061, -0.2681, -0.0266, 0.0061, -0.2692, -0.0263, 0.0083, -0.2681, -0.0266, 0.0083, -0.2692, -0.0325, 0.0061, -0.2704, -0.0325, 0.0083, -0.2704, -0.0317, 0.0061, -0.2682,
-0.0317, 0.0083, -0.2682, -0.0271, 0.01, -0.2706, -0.0262, 0.01, -0.2678, -0.0265, 0.01, -0.2687, -0.0298, 0.01, -0.2704, -0.0324, 0.01, -0.2701, -0.0298, 0.01, -0.2682, -0.0316, 0.01, -0.2681, -0.0159, 0.0061, -0.2664, -0.0159, 0.0083, -0.2664, -0.0198, 0.0061, -0.2635, -0.0198, 0.0083, -0.2635, -0.0206, 0.0061, -0.266, -0.0206, 0.0083, -0.266, -0.0202, 0.0061, -0.2646, -0.0202, 0.0083, -0.2646, -0.0158, 0.01, -0.2661, -0.019, 0.01, -0.2661, -0.0168, 0.01, -0.2639, -0.0199, 0.01, -0.2637, -0.0206, 0.01, -0.266, -0.0202, 0.01, -0.2647, -0.0193, 0.0061, -0.2617, -0.0193, 0.0083, -0.2617, -0.0185, 0.0061, -0.2595, -0.0185, 0.0083, -0.2595, -0.0241, 0.0061, -0.2616, -0.0241, 0.0083, -0.2616, -0.0232, 0.0061, -0.2595, -0.0232, 0.0083, -0.2595,
-0.0168, 0.01, -0.2617, -0.0192, 0.01, -0.2615, -0.0168, 0.01, -0.2595, -0.0185, 0.01, -0.2594, -0.024, 0.01, -0.2614, -0.0233, 0.01, -0.2598, -0.0256, 0.0061, -0.2661, -0.0256, 0.0083, -0.2661, -0.0249, 0.0061, -0.2639, -0.0249, 0.0083, -0.2639, -0.0306, 0.0061, -0.2657, -0.0311, 0.0061, -0.2668, -0.0306, 0.0083, -0.2657, -0.0311, 0.0083, -0.2668, -0.0298, 0.0061, -0.2639, -0.0298, 0.0083, -0.2639, -0.0256, 0.01, -0.2662, -0.0277, 0.01, -0.2661, -0.025, 0.01, -0.2641, -0.0277, 0.01, -0.2639, -0.0307, 0.01, -0.2658, -0.0311, 0.01, -0.2669, -0.0297, 0.01, -0.2637, -0.0245, 0.0061, -0.2627, -0.0285, 0.0061, -0.2611, -0.0245, 0.0083, -0.2627, -0.0285, 0.0083, -0.2611, -0.0276, 0.0061, -0.2595, -0.0276, 0.0083, -0.2595, -0.029, 0.0061, -0.2622,
-0.029, 0.0083, -0.2622, -0.0243, 0.01, -0.2622, -0.0286, 0.01, -0.2613, -0.0255, 0.01, -0.2595, -0.0275, 0.01, -0.2593, -0.0291, 0.01, -0.2623, -0.0142, 0.0061, -0.2574, -0.0142, 0.0083, -0.2574, -0.0136, 0.0061, -0.2542, -0.0138, 0.0061, -0.2552, -0.0136, 0.0083, -0.2542, -0.0138, 0.0083, -0.2552, -0.0142, 0.01, -0.2575, -0.0135, 0.01, -0.2541, -0.0138, 0.01, -0.2551, -0.0133, 0.0061, -0.253, -0.0154, 0.0061, -0.2526, -0.0133, 0.0083, -0.253, -0.0154, 0.0083, -0.2526, -0.0128, 0.0061, -0.2509, -0.0145, 0.0061, -0.2508, -0.0128, 0.0083, -0.2509, -0.0145, 0.0083, -0.2508, -0.0133, 0.01, -0.2531, -0.0155, 0.01, -0.2527, -0.0128, 0.01, -0.2509, -0.0144, 0.01, -0.2506, -0.0121, 0.0061, -0.2467, -0.0137, 0.0061, -0.2496, -0.0121, 0.0083, -0.2467,
-0.0137, 0.0083, -0.2496, -0.0131, 0.01, -0.2497, -0.0137, 0.01, -0.2496, -0.0176, 0.0061, -0.2571, -0.018, 0.0061, -0.2582, -0.0176, 0.0083, -0.2571, -0.018, 0.0083, -0.2582, -0.0167, 0.0061, -0.2552, -0.0167, 0.0083, -0.2552, -0.022, 0.0061, -0.2569, -0.0225, 0.0061, -0.258, -0.022, 0.0083, -0.2569, -0.0225, 0.0083, -0.258, -0.0211, 0.0061, -0.2552, -0.0244, 0.0061, -0.2542, -0.0211, 0.0083, -0.2552, -0.0244, 0.0083, -0.2542, -0.0176, 0.01, -0.2572, -0.018, 0.01, -0.2583, -0.0166, 0.01, -0.255, -0.0219, 0.01, -0.2568, -0.0224, 0.01, -0.2578, -0.0212, 0.01, -0.2554, -0.0245, 0.01, -0.2544, -0.016, 0.0061, -0.2537, -0.0198, 0.0061, -0.2525, -0.016, 0.0083, -0.2537, -0.0198, 0.0083, -0.2525, -0.019, 0.0061, -0.2509, -0.019, 0.0083, -0.2509,
-0.0203, 0.0061, -0.2536, -0.0235, 0.0061, -0.253, -0.0203, 0.0083, -0.2536, -0.0235, 0.0083, -0.253, -0.0217, 0.0061, -0.2505, -0.0225, 0.0061, -0.2516, -0.0217, 0.0083, -0.2505, -0.0225, 0.0083, -0.2516, -0.016, 0.01, -0.2538, -0.0197, 0.01, -0.2524, -0.0191, 0.01, -0.2511, -0.0203, 0.01, -0.2534, -0.0233, 0.01, -0.2527, -0.0217, 0.01, -0.2505, -0.0225, 0.01, -0.2516, -0.0262, 0.0061, -0.257, -0.0268, 0.0061, -0.2581, -0.0262, 0.0083, -0.257, -0.0268, 0.0083, -0.2581, -0.0251, 0.0061, -0.2553, -0.0251, 0.0083, -0.2553, -0.0262, 0.01, -0.257, -0.0268, 0.01, -0.2581, -0.0252, 0.01, -0.2554, -0.0176, 0.0061, -0.2481, -0.0189, 0.0061, -0.2486, -0.0176, 0.0083, -0.2481, -0.0189, 0.0083, -0.2486, -0.0163, 0.0061, -0.2449, -0.0186, 0.0061, -0.2469,
-0.0163, 0.0083, -0.2449, -0.0186, 0.0083, -0.2469, -0.0206, 0.0061, -0.2491, -0.0206, 0.0083, -0.2491, -0.0176, 0.01, -0.248, -0.019, 0.01, -0.2486, -0.0163, 0.01, -0.2449, -0.0186, 0.01, -0.2469, -0.0207, 0.01, -0.2491, -0.0341, 0.0061, -0.2747, -0.0341, 0.0083, -0.2747, -0.0334, 0.0061, -0.2728, -0.0334, 0.0083, -0.2728, -0.034, 0.01, -0.2746, -0.0335, 0.01, -0.273, 0.0005, 0.0061, -0.2196, -0.0016, 0.0061, -0.2196, 0.0005, 0.0083, -0.2196, -0.0016, 0.0083, -0.2196, -0.0038, 0.0061, -0.2196, -0.006, 0.0061, -0.2196, -0.0038, 0.0083, -0.2196, -0.006, 0.0083, -0.2196, 0.0005, 0.01, -0.2196, -0.0016, 0.01, -0.2196, -0.0038, 0.01, -0.2196, -0.006, 0.01, -0.2196, 0.0005, 0.01, -0.2183, -0.0016, 0.01, -0.2183, 0.0005, 0.01, -0.2161,
-0.0016, 0.01, -0.2161, -0.0038, 0.01, -0.2183, -0.006, 0.01, -0.2183, -0.0038, 0.01, -0.2161, -0.006, 0.01, -0.2161, -0.0081, 0.0061, -0.2196, -0.0103, 0.0061, -0.2196, -0.0081, 0.0083, -0.2196, -0.0103, 0.0083, -0.2196, -0.0125, 0.0061, -0.2196, -0.0146, 0.0061, -0.2196, -0.0125, 0.0083, -0.2196, -0.0146, 0.0083, -0.2196, -0.0081, 0.01, -0.2196, -0.0103, 0.01, -0.2196, -0.0125, 0.01, -0.2196, -0.0146, 0.01, -0.2196, -0.0081, 0.01, -0.2183, -0.0103, 0.01, -0.2183, -0.0081, 0.01, -0.2161, -0.0103, 0.01, -0.2161, -0.0125, 0.01, -0.2183, -0.0146, 0.01, -0.2183, -0.0125, 0.01, -0.2161, -0.0146, 0.01, -0.2161, 0.0005, 0.01, -0.214, -0.0016, 0.01, -0.214, 0.0005, 0.01, -0.2118, -0.0016, 0.01, -0.2118, -0.0038, 0.01, -0.214,
-0.006, 0.01, -0.214, -0.0038, 0.01, -0.2118, -0.006, 0.01, -0.2118, 0.0005, 0.0061, -0.2088, -0.0016, 0.0061, -0.2088, 0.0005, 0.0083, -0.2088, -0.0016, 0.0083, -0.2088, -0.0038, 0.0061, -0.2088, -0.006, 0.0061, -0.2088, -0.0038, 0.0083, -0.2088, -0.006, 0.0083, -0.2088, 0.0005, 0.01, -0.2088, -0.0016, 0.01, -0.2088, -0.0038, 0.01, -0.2088, -0.006, 0.01, -0.2088, -0.0081, 0.01, -0.214, -0.0103, 0.01, -0.214, -0.0081, 0.01, -0.2118, -0.0103, 0.01, -0.2118, -0.0125, 0.01, -0.214, -0.0146, 0.01, -0.214, -0.0125, 0.01, -0.2118, -0.0146, 0.01, -0.2118, -0.0085, 0.0061, -0.2088, -0.0085, 0.0083, -0.2088, -0.0085, 0.0061, -0.2075, -0.0085, 0.0083, -0.2075, -0.0085, 0.01, -0.2088, -0.0103, 0.01, -0.2096, -0.0085, 0.01, -0.2075,
-0.0103, 0.01, -0.2075, -0.0125, 0.01, -0.2096, -0.0146, 0.01, -0.2096, -0.0125, 0.01, -0.2075, -0.0146, 0.01, -0.2075, -0.0168, 0.0061, -0.2196, -0.019, 0.0061, -0.2196, -0.0168, 0.0083, -0.2196, -0.019, 0.0083, -0.2196, -0.0212, 0.0061, -0.2196, -0.0233, 0.0061, -0.2196, -0.0212, 0.0083, -0.2196, -0.0233, 0.0083, -0.2196, -0.0168, 0.01, -0.2196, -0.019, 0.01, -0.2196, -0.0212, 0.01, -0.2196, -0.0233, 0.01, -0.2196, -0.0168, 0.01, -0.2183, -0.019, 0.01, -0.2183, -0.0168, 0.01, -0.2161, -0.019, 0.01, -0.2161, -0.0212, 0.01, -0.2183, -0.0233, 0.01, -0.2183, -0.0212, 0.01, -0.2161, -0.0233, 0.01, -0.2161, -0.0255, 0.0061, -0.2196, -0.0277, 0.0061, -0.2196, -0.0255, 0.0083, -0.2196, -0.0277, 0.0083, -0.2196, -0.0298, 0.0061, -0.2196,
-0.032, 0.0061, -0.2196, -0.0298, 0.0083, -0.2196, -0.032, 0.0083, -0.2196, -0.0255, 0.01, -0.2196, -0.0277, 0.01, -0.2196, -0.0298, 0.01, -0.2196, -0.032, 0.01, -0.2196, -0.0255, 0.01, -0.2183, -0.0277, 0.01, -0.2183, -0.0255, 0.01, -0.2161, -0.0277, 0.01, -0.2161, -0.0298, 0.01, -0.2183, -0.032, 0.01, -0.2183, -0.0298, 0.01, -0.2161, -0.032, 0.01, -0.2161, -0.0168, 0.01, -0.214, -0.019, 0.01, -0.214, -0.0168, 0.01, -0.2118, -0.019, 0.01, -0.2118, -0.0212, 0.01, -0.214, -0.0233, 0.01, -0.214, -0.0212, 0.01, -0.2118, -0.0233, 0.01, -0.2118, -0.0203, 0.0061, -0.2088, -0.0233, 0.0061, -0.2088, -0.0203, 0.0083, -0.2088, -0.0233, 0.0083, -0.2088, -0.0203, 0.0061, -0.2075, -0.0203, 0.0083, -0.2075, -0.0168, 0.01, -0.2096,
-0.019, 0.01, -0.2096, -0.0168, 0.01, -0.2075, -0.019, 0.01, -0.2075, -0.0203, 0.01, -0.2088, -0.0233, 0.01, -0.2088, -0.0203, 0.01, -0.2075, -0.0255, 0.01, -0.214, -0.0277, 0.01, -0.214, -0.0255, 0.01, -0.2118, -0.0277, 0.01, -0.2118, -0.0298, 0.01, -0.214, -0.032, 0.01, -0.214, -0.0298, 0.01, -0.2118, -0.032, 0.01, -0.2118, -0.0255, 0.0061, -0.2088, -0.0277, 0.0061, -0.2088, -0.0255, 0.0083, -0.2088, -0.0277, 0.0083, -0.2088, -0.0298, 0.0061, -0.2088, -0.032, 0.0061, -0.2088, -0.0298, 0.0083, -0.2088, -0.032, 0.0083, -0.2088, -0.0255, 0.01, -0.2088, -0.0277, 0.01, -0.2088, -0.0298, 0.01, -0.2088, -0.0319, 0.01, -0.2088, -0.0342, 0.0061, -0.2196, -0.0363, 0.0061, -0.2196, -0.0342, 0.0083, -0.2196, -0.0363, 0.0083, -0.2196,
-0.0385, 0.0061, -0.2196, -0.0407, 0.0061, -0.2196, -0.0385, 0.0083, -0.2196, -0.0407, 0.0083, -0.2196, -0.0342, 0.01, -0.2196, -0.0363, 0.01, -0.2196, -0.0385, 0.01, -0.2196, -0.0407, 0.01, -0.2196, -0.0342, 0.01, -0.2183, -0.0363, 0.01, -0.2183, -0.0342, 0.01, -0.2161, -0.0363, 0.01, -0.2161, -0.0385, 0.01, -0.2183, -0.0407, 0.01, -0.2183, -0.0385, 0.01, -0.2161, -0.0407, 0.01, -0.2161, -0.0429, 0.0061, -0.2196, -0.045, 0.0061, -0.2196, -0.0429, 0.0083, -0.2196, -0.045, 0.0083, -0.2196, -0.0472, 0.0061, -0.2196, -0.0487, 0.0061, -0.2196, -0.0472, 0.0083, -0.2196, -0.0487, 0.0083, -0.2196, -0.0429, 0.01, -0.2196, -0.045, 0.01, -0.2196, -0.0472, 0.01, -0.2196, -0.0487, 0.01, -0.2196, -0.0487, 0.0061, -0.2183, -0.0487, 0.0083, -0.2183,
-0.0487, 0.0061, -0.2161, -0.0487, 0.0083, -0.2161, -0.0429, 0.01, -0.2183, -0.045, 0.01, -0.2183, -0.0429, 0.01, -0.2161, -0.045, 0.01, -0.2161, -0.0472, 0.01, -0.2183, -0.0487, 0.01, -0.2183, -0.0472, 0.01, -0.2161, -0.0487, 0.01, -0.2161, -0.0342, 0.01, -0.214, -0.0363, 0.01, -0.214, -0.0342, 0.01, -0.2118, -0.0363, 0.01, -0.2118, -0.0385, 0.01, -0.214, -0.0407, 0.01, -0.214, -0.0385, 0.01, -0.2118, -0.0407, 0.01, -0.2118, -0.0334, 0.0061, -0.2086, -0.0334, 0.0083, -0.2086, -0.0345, 0.0061, -0.2083, -0.0361, 0.0061, -0.2076, -0.0345, 0.0083, -0.2083, -0.0361, 0.0083, -0.2076, -0.0339, 0.01, -0.2085, -0.0363, 0.01, -0.2096, -0.0347, 0.01, -0.2083, -0.0361, 0.01, -0.2076, -0.0385, 0.01, -0.2096, -0.0407, 0.01, -0.2096,
-0.0385, 0.01, -0.2075, -0.0407, 0.01, -0.2075, -0.0487, 0.0061, -0.214, -0.0487, 0.0083, -0.214, -0.0487, 0.0061, -0.2118, -0.0487, 0.0083, -0.2118, -0.0429, 0.01, -0.214, -0.045, 0.01, -0.214, -0.0429, 0.01, -0.2118, -0.045, 0.01, -0.2118, -0.0472, 0.01, -0.214, -0.0487, 0.01, -0.214, -0.0472, 0.01, -0.2118, -0.0487, 0.01, -0.2118, -0.0487, 0.0061, -0.2096, -0.0487, 0.0083, -0.2096, -0.0487, 0.0061, -0.2075, -0.0487, 0.0083, -0.2075, -0.0429, 0.01, -0.2096, -0.045, 0.01, -0.2096, -0.0429, 0.01, -0.2075, -0.045, 0.01, -0.2075, -0.0472, 0.01, -0.2096, -0.0487, 0.01, -0.2096, -0.0472, 0.01, -0.2075, -0.0487, 0.01, -0.2075, -0.0085, 0.0061, -0.2053, -0.0085, 0.0083, -0.2053, -0.0085, 0.0061, -0.2031, -0.0085, 0.0083, -0.2031,
-0.0085, 0.01, -0.2053, -0.0103, 0.01, -0.2053, -0.0085, 0.01, -0.2031, -0.0103, 0.01, -0.2031, -0.0125, 0.01, -0.2053, -0.0146, 0.01, -0.2053, -0.0125, 0.01, -0.2031, -0.0146, 0.01, -0.2031, -0.0085, 0.0061, -0.2009, -0.0085, 0.0083, -0.2009, -0.0085, 0.0061, -0.1988, -0.0085, 0.0083, -0.1988, -0.0085, 0.01, -0.2009, -0.0103, 0.01, -0.2009, -0.0085, 0.01, -0.1988, -0.0103, 0.01, -0.1988, -0.0125, 0.01, -0.2009, -0.0146, 0.01, -0.2009, -0.0125, 0.01, -0.1988, -0.0146, 0.01, -0.1988, -0.0085, 0.0061, -0.1966, -0.0085, 0.0083, -0.1966, -0.0085, 0.0061, -0.1944, -0.0085, 0.0083, -0.1944, -0.0085, 0.01, -0.1966, -0.0103, 0.01, -0.1966, -0.0085, 0.01, -0.1944, -0.0103, 0.01, -0.1944, -0.0125, 0.01, -0.1966, -0.0146, 0.01, -0.1966,
-0.0125, 0.01, -0.1944, -0.0146, 0.01, -0.1944, -0.0085, 0.0061, -0.1923, -0.0085, 0.0083, -0.1923, -0.0085, 0.0061, -0.1901, -0.0085, 0.0083, -0.1901, -0.0085, 0.01, -0.1923, -0.0103, 0.01, -0.1923, -0.0085, 0.01, -0.1901, -0.0103, 0.01, -0.1901, -0.0125, 0.01, -0.1923, -0.0146, 0.01, -0.1923, -0.0125, 0.01, -0.1901, -0.0146, 0.01, -0.1901, -0.0203, 0.0061, -0.2053, -0.0203, 0.0083, -0.2053, -0.0203, 0.0061, -0.2031, -0.0203, 0.0083, -0.2031, -0.0168, 0.01, -0.2053, -0.019, 0.01, -0.2053, -0.0168, 0.01, -0.2031, -0.019, 0.01, -0.2031, -0.0203, 0.01, -0.2053, -0.0203, 0.01, -0.2031, -0.0203, 0.0061, -0.2009, -0.0203, 0.0083, -0.2009, -0.0203, 0.0061, -0.1988, -0.0203, 0.0083, -0.1988, -0.0168, 0.01, -0.2009, -0.019, 0.01, -0.2009,
-0.0168, 0.01, -0.1988, -0.019, 0.01, -0.1988, -0.0203, 0.01, -0.2009, -0.0203, 0.01, -0.1988, -0.0203, 0.0061, -0.1966, -0.0203, 0.0083, -0.1966, -0.0203, 0.0061, -0.1944, -0.0203, 0.0083, -0.1944, -0.0168, 0.01, -0.1966, -0.019, 0.01, -0.1966, -0.0168, 0.01, -0.1944, -0.019, 0.01, -0.1944, -0.0203, 0.01, -0.1966, -0.0203, 0.01, -0.1944, -0.0203, 0.0061, -0.1923, -0.0203, 0.0083, -0.1923, -0.0203, 0.0061, -0.1901, -0.0203, 0.0083, -0.1901, -0.0168, 0.01, -0.1923, -0.019, 0.01, -0.1923, -0.0168, 0.01, -0.1901, -0.019, 0.01, -0.1901, -0.0203, 0.01, -0.1923, -0.0203, 0.01, -0.1901, -0.0085, 0.0061, -0.1879, -0.0085, 0.0083, -0.1879, -0.0085, 0.0061, -0.1858, -0.0085, 0.0083, -0.1858, -0.0085, 0.01, -0.1879, -0.0103, 0.01, -0.1879,
-0.0085, 0.01, -0.1858, -0.0103, 0.01, -0.1858, -0.0125, 0.01, -0.1879, -0.0146, 0.01, -0.1879, -0.0125, 0.01, -0.1858, -0.0146, 0.01, -0.1858, -0.0085, 0.0061, -0.1836, -0.0085, 0.0083, -0.1836, -0.0085, 0.0061, -0.1814, -0.0085, 0.0083, -0.1814, -0.0085, 0.01, -0.1836, -0.0103, 0.01, -0.1836, -0.0085, 0.01, -0.1814, -0.0103, 0.01, -0.1814, -0.0125, 0.01, -0.1836, -0.0146, 0.01, -0.1836, -0.0125, 0.01, -0.1814, -0.0146, 0.01, -0.1814, -0.0085, 0.0061, -0.1792, -0.0085, 0.0083, -0.1792, -0.0085, 0.0061, -0.1771, -0.0085, 0.0083, -0.1771, -0.0085, 0.01, -0.1792, -0.0103, 0.01, -0.1792, -0.0085, 0.01, -0.1771, -0.0103, 0.01, -0.1771, -0.0125, 0.01, -0.1792, -0.0146, 0.01, -0.1792, -0.0125, 0.01, -0.1771, -0.0146, 0.01, -0.1771,
-0.0085, 0.0061, -0.1749, -0.0085, 0.0083, -0.1749, -0.0085, 0.0061, -0.1727, -0.0085, 0.0083, -0.1727, -0.0085, 0.01, -0.1749, -0.0103, 0.01, -0.1749, -0.0085, 0.01, -0.1727, -0.0103, 0.01, -0.1727, -0.0125, 0.01, -0.1749, -0.0146, 0.01, -0.1749, -0.0125, 0.01, -0.1727, -0.0146, 0.01, -0.1727, -0.0203, 0.0061, -0.1879, -0.0203, 0.0083, -0.1879, -0.0203, 0.0061, -0.1858, -0.0203, 0.0083, -0.1858, -0.0168, 0.01, -0.1879, -0.019, 0.01, -0.1879, -0.0168, 0.01, -0.1858, -0.019, 0.01, -0.1858, -0.0203, 0.01, -0.1879, -0.0203, 0.01, -0.1858, -0.0203, 0.0061, -0.1836, -0.0203, 0.0083, -0.1836, -0.0203, 0.0061, -0.1814, -0.0203, 0.0083, -0.1814, -0.0168, 0.01, -0.1836, -0.019, 0.01, -0.1836, -0.0168, 0.01, -0.1814, -0.019, 0.01, -0.1814,
-0.0203, 0.01, -0.1836, -0.0203, 0.01, -0.1814, -0.0203, 0.0061, -0.1792, -0.0203, 0.0083, -0.1792, -0.0203, 0.0061, -0.1771, -0.0203, 0.0083, -0.1771, -0.0168, 0.01, -0.1792, -0.019, 0.01, -0.1792, -0.0168, 0.01, -0.1771, -0.019, 0.01, -0.1771, -0.0203, 0.01, -0.1792, -0.0203, 0.01, -0.1771, -0.0203, 0.0061, -0.1749, -0.0203, 0.0083, -0.1749, -0.0203, 0.0061, -0.1727, -0.0203, 0.0083, -0.1727, -0.0168, 0.01, -0.1749, -0.019, 0.01, -0.1749, -0.0168, 0.01, -0.1727, -0.019, 0.01, -0.1727, -0.0203, 0.01, -0.1749, -0.0203, 0.01, -0.1727, -0.0372, 0.0061, -0.2059, -0.0372, 0.0083, -0.2059, -0.0376, 0.0061, -0.2048, -0.0376, 0.0083, -0.2048, -0.0378, 0.0061, -0.2031, -0.0378, 0.0083, -0.2031, -0.0372, 0.01, -0.206, -0.0375, 0.01, -0.2051,
-0.0407, 0.01, -0.2053, -0.0378, 0.01, -0.203, -0.0407, 0.01, -0.2031, -0.0379, 0.0061, -0.2009, -0.0379, 0.0083, -0.2009, -0.0379, 0.0061, -0.1988, -0.0379, 0.0083, -0.1988, -0.0379, 0.01, -0.2009, -0.0407, 0.01, -0.2009, -0.0379, 0.01, -0.1988, -0.0407, 0.01, -0.1988, -0.0487, 0.0061, -0.2053, -0.0487, 0.0083, -0.2053, -0.0487, 0.0061, -0.2031, -0.0487, 0.0083, -0.2031, -0.0429, 0.01, -0.2053, -0.045, 0.01, -0.2053, -0.0429, 0.01, -0.2031, -0.045, 0.01, -0.2031, -0.0472, 0.01, -0.2053, -0.0487, 0.01, -0.2053, -0.0472, 0.01, -0.2031, -0.0487, 0.01, -0.2031, -0.0487, 0.0061, -0.2009, -0.0487, 0.0083, -0.2009, -0.0487, 0.0061, -0.1988, -0.0487, 0.0083, -0.1988, -0.0429, 0.01, -0.2009, -0.045, 0.01, -0.2009, -0.0429, 0.01, -0.1988,
-0.045, 0.01, -0.1988, -0.0472, 0.01, -0.2009, -0.0487, 0.01, -0.2009, -0.0472, 0.01, -0.1988, -0.0487, 0.01, -0.1988, -0.0379, 0.0061, -0.1966, -0.0379, 0.0083, -0.1966, -0.0379, 0.0061, -0.1944, -0.0379, 0.0083, -0.1944, -0.0379, 0.01, -0.1966, -0.0407, 0.01, -0.1966, -0.0379, 0.01, -0.1944, -0.0407, 0.01, -0.1944, -0.0379, 0.0061, -0.1923, -0.0379, 0.0083, -0.1923, -0.0379, 0.0061, -0.1901, -0.0379, 0.0083, -0.1901, -0.0379, 0.01, -0.1923, -0.0407, 0.01, -0.1923, -0.0379, 0.01, -0.1901, -0.0407, 0.01, -0.1901, -0.0487, 0.0061, -0.1966, -0.0487, 0.0083, -0.1966, -0.0487, 0.0061, -0.1944, -0.0487, 0.0083, -0.1944, -0.0429, 0.01, -0.1966, -0.045, 0.01, -0.1966, -0.0429, 0.01, -0.1944, -0.045, 0.01, -0.1944, -0.0472, 0.01, -0.1966,
-0.0487, 0.01, -0.1966, -0.0472, 0.01, -0.1944, -0.0487, 0.01, -0.1944, -0.0487, 0.0061, -0.1923, -0.0487, 0.0083, -0.1923, -0.0487, 0.0061, -0.1901, -0.0487, 0.0083, -0.1901, -0.0429, 0.01, -0.1923, -0.045, 0.01, -0.1923, -0.0429, 0.01, -0.1901, -0.045, 0.01, -0.1901, -0.0472, 0.01, -0.1923, -0.0487, 0.01, -0.1923, -0.0472, 0.01, -0.1901, -0.0487, 0.01, -0.1901, -0.0379, 0.0061, -0.1879, -0.0379, 0.0083, -0.1879, -0.0379, 0.0061, -0.1858, -0.0379, 0.0083, -0.1858, -0.0379, 0.01, -0.1879, -0.0407, 0.01, -0.1879, -0.0379, 0.01, -0.1858, -0.0407, 0.01, -0.1858, -0.0379, 0.0061, -0.1836, -0.0379, 0.0083, -0.1836, -0.0379, 0.0061, -0.1814, -0.0379, 0.0083, -0.1814, -0.0379, 0.01, -0.1836, -0.0407, 0.01, -0.1836, -0.0379, 0.01, -0.1814,
-0.0407, 0.01, -0.1814, -0.0487, 0.0061, -0.1879, -0.0487, 0.0083, -0.1879, -0.0487, 0.0061, -0.1858, -0.0487, 0.0083, -0.1858, -0.0429, 0.01, -0.1879, -0.045, 0.01, -0.1879, -0.0429, 0.01, -0.1858, -0.045, 0.01, -0.1858, -0.0472, 0.01, -0.1879, -0.0487, 0.01, -0.1879, -0.0472, 0.01, -0.1858, -0.0487, 0.01, -0.1858, -0.0487, 0.0061, -0.1836, -0.0487, 0.0083, -0.1836, -0.0487, 0.0061, -0.1814, -0.0487, 0.0083, -0.1814, -0.0429, 0.01, -0.1836, -0.045, 0.01, -0.1836, -0.0429, 0.01, -0.1814, -0.045, 0.01, -0.1814, -0.0472, 0.01, -0.1836, -0.0487, 0.01, -0.1836, -0.0472, 0.01, -0.1814, -0.0487, 0.01, -0.1814, -0.0379, 0.0061, -0.1792, -0.0379, 0.0083, -0.1792, -0.0379, 0.0061, -0.1771, -0.0379, 0.0083, -0.1771, -0.0379, 0.01, -0.1792,
-0.0407, 0.01, -0.1792, -0.0379, 0.01, -0.1771, -0.0407, 0.01, -0.1771, -0.0379, 0.0061, -0.1749, -0.0379, 0.0083, -0.1749, -0.0379, 0.0061, -0.1727, -0.0379, 0.0083, -0.1727, -0.0379, 0.01, -0.1749, -0.0407, 0.01, -0.1749, -0.0379, 0.01, -0.1727, -0.0407, 0.01, -0.1727, -0.0487, 0.0061, -0.1792, -0.0487, 0.0083, -0.1792, -0.0487, 0.0061, -0.1771, -0.0487, 0.0083, -0.1771, -0.0429, 0.01, -0.1792, -0.045, 0.01, -0.1792, -0.0429, 0.01, -0.1771, -0.045, 0.01, -0.1771, -0.0472, 0.01, -0.1792, -0.0487, 0.01, -0.1792, -0.0472, 0.01, -0.1771, -0.0487, 0.01, -0.1771, -0.0487, 0.0061, -0.1749, -0.0487, 0.0083, -0.1749, -0.0487, 0.0061, -0.1727, -0.0487, 0.0083, -0.1727, -0.0429, 0.01, -0.1749, -0.045, 0.01, -0.1749, -0.0429, 0.01, -0.1727,
-0.045, 0.01, -0.1727, -0.0472, 0.01, -0.1749, -0.0487, 0.01, -0.1749, -0.0472, 0.01, -0.1727, -0.0487, 0.01, -0.1727, -0.0085, 0.0061, -0.1706, -0.0085, 0.0083, -0.1706, -0.0085, 0.0061, -0.1684, -0.0085, 0.0083, -0.1684, -0.0085, 0.01, -0.1706, -0.0103, 0.01, -0.1706, -0.0085, 0.01, -0.1684, -0.0103, 0.01, -0.1684, -0.0125, 0.01, -0.1706, -0.0146, 0.01, -0.1706, -0.0125, 0.01, -0.1684, -0.0146, 0.01, -0.1684, -0.0085, 0.0061, -0.1662, -0.0085, 0.0083, -0.1662, -0.0085, 0.0061, -0.1641, -0.0085, 0.0083, -0.1641, -0.0085, 0.01, -0.1662, -0.0103, 0.01, -0.1662, -0.0085, 0.01, -0.1641, -0.0103, 0.01, -0.1641, -0.0125, 0.01, -0.1662, -0.0146, 0.01, -0.1662, -0.0125, 0.01, -0.1641, -0.0146, 0.01, -0.1641, -0.0085, 0.0061, -0.1627,
-0.0103, 0.0061, -0.1627, -0.0085, 0.0083, -0.1627, -0.0103, 0.0083, -0.1627, -0.0125, 0.0061, -0.1627, -0.0146, 0.0061, -0.1627, -0.0125, 0.0083, -0.1627, -0.0146, 0.0083, -0.1627, -0.0085, 0.01, -0.1627, -0.0103, 0.01, -0.1627, -0.0125, 0.01, -0.1627, -0.0146, 0.01, -0.1627, -0.0203, 0.0061, -0.1706, -0.0203, 0.0083, -0.1706, -0.0203, 0.0061, -0.1684, -0.0203, 0.0083, -0.1684, -0.0168, 0.01, -0.1706, -0.019, 0.01, -0.1706, -0.0168, 0.01, -0.1684, -0.019, 0.01, -0.1684, -0.0203, 0.01, -0.1706, -0.0203, 0.01, -0.1684, -0.0203, 0.0061, -0.1662, -0.0203, 0.0083, -0.1662, -0.0203, 0.0061, -0.1641, -0.0203, 0.0083, -0.1641, -0.0168, 0.01, -0.1662, -0.019, 0.01, -0.1662, -0.0168, 0.01, -0.1641, -0.019, 0.01, -0.1641, -0.0203, 0.01, -0.1662,
-0.0203, 0.01, -0.1641, -0.0168, 0.0061, -0.1627, -0.019, 0.0061, -0.1627, -0.0168, 0.0083, -0.1627, -0.019, 0.0083, -0.1627, -0.0203, 0.0061, -0.1627, -0.0203, 0.0083, -0.1627, -0.0168, 0.01, -0.1627, -0.019, 0.01, -0.1627, -0.0203, 0.01, -0.1627, 0.0005, 0.0061, -0.151, -0.0016, 0.0061, -0.151, 0.0005, 0.0083, -0.151, -0.0016, 0.0083, -0.151, -0.0038, 0.0061, -0.151, -0.006, 0.0061, -0.151, -0.0038, 0.0083, -0.151, -0.006, 0.0083, -0.151, 0.0005, 0.01, -0.151, -0.0016, 0.01, -0.151, -0.0038, 0.01, -0.151, -0.006, 0.01, -0.151, 0.0005, 0.01, -0.1489, -0.0016, 0.01, -0.1489, 0.0005, 0.01, -0.1467, -0.0016, 0.01, -0.1467, -0.0038, 0.01, -0.1489, -0.006, 0.01, -0.1489, -0.0038, 0.01, -0.1467, -0.006, 0.01, -0.1467,
-0.0081, 0.0061, -0.151, -0.0103, 0.0061, -0.151, -0.0081, 0.0083, -0.151, -0.0103, 0.0083, -0.151, -0.0125, 0.0061, -0.151, -0.0146, 0.0061, -0.151, -0.0125, 0.0083, -0.151, -0.0146, 0.0083, -0.151, -0.0081, 0.01, -0.151, -0.0103, 0.01, -0.151, -0.0125, 0.01, -0.151, -0.0146, 0.01, -0.151, -0.0081, 0.01, -0.1489, -0.0103, 0.01, -0.1489, -0.0081, 0.01, -0.1467, -0.0103, 0.01, -0.1467, -0.0125, 0.01, -0.1489, -0.0146, 0.01, -0.1489, -0.0125, 0.01, -0.1467, -0.0146, 0.01, -0.1467, 0.0005, 0.01, -0.1445, -0.0016, 0.01, -0.1445, 0.0005, 0.01, -0.1424, -0.0016, 0.01, -0.1424, -0.0038, 0.01, -0.1445, -0.006, 0.01, -0.1445, -0.0038, 0.01, -0.1424, -0.006, 0.01, -0.1424, 0.0005, 0.0061, -0.1392, -0.0016, 0.0061, -0.1392,
0.0005, 0.0083, -0.1392, -0.0016, 0.0083, -0.1392, -0.0038, 0.0061, -0.1392, -0.006, 0.0061, -0.1392, -0.0038, 0.0083, -0.1392, -0.006, 0.0083, -0.1392, 0.0005, 0.01, -0.1392, -0.0016, 0.01, -0.1392, -0.0038, 0.01, -0.1392, -0.006, 0.01, -0.1392, -0.0081, 0.01, -0.1445, -0.0103, 0.01, -0.1445, -0.0081, 0.01, -0.1424, -0.0103, 0.01, -0.1424, -0.0125, 0.01, -0.1445, -0.0146, 0.01, -0.1445, -0.0125, 0.01, -0.1424, -0.0146, 0.01, -0.1424, -0.0081, 0.0061, -0.1392, -0.0103, 0.0061, -0.1392, -0.0081, 0.0083, -0.1392, -0.0103, 0.0083, -0.1392, -0.0125, 0.0061, -0.1392, -0.0146, 0.0061, -0.1392, -0.0125, 0.0083, -0.1392, -0.0146, 0.0083, -0.1392, -0.0081, 0.01, -0.1392, -0.0103, 0.01, -0.1392, -0.0125, 0.01, -0.1392, -0.0146, 0.01, -0.1392,
-0.0168, 0.0061, -0.151, -0.019, 0.0061, -0.151, -0.0168, 0.0083, -0.151, -0.019, 0.0083, -0.151, -0.0212, 0.0061, -0.151, -0.0233, 0.0061, -0.151, -0.0212, 0.0083, -0.151, -0.0233, 0.0083, -0.151, -0.0168, 0.01, -0.151, -0.019, 0.01, -0.151, -0.0212, 0.01, -0.151, -0.0233, 0.01, -0.151, -0.0168, 0.01, -0.1489, -0.019, 0.01, -0.1489, -0.0168, 0.01, -0.1467, -0.019, 0.01, -0.1467, -0.0212, 0.01, -0.1489, -0.0233, 0.01, -0.1489, -0.0212, 0.01, -0.1467, -0.0233, 0.01, -0.1467, -0.0255, 0.0061, -0.151, -0.0277, 0.0061, -0.151, -0.0255, 0.0083, -0.151, -0.0277, 0.0083, -0.151, -0.0298, 0.0061, -0.151, -0.032, 0.0061, -0.151, -0.0298, 0.0083, -0.151, -0.032, 0.0083, -0.151, -0.0255, 0.01, -0.151, -0.0277, 0.01, -0.151,
-0.0298, 0.01, -0.151, -0.032, 0.01, -0.151, -0.0255, 0.01, -0.1489, -0.0277, 0.01, -0.1489, -0.0255, 0.01, -0.1467, -0.0277, 0.01, -0.1467, -0.0298, 0.01, -0.1489, -0.032, 0.01, -0.1489, -0.0298, 0.01, -0.1467, -0.032, 0.01, -0.1467, -0.0168, 0.01, -0.1445, -0.019, 0.01, -0.1445, -0.0168, 0.01, -0.1424, -0.019, 0.01, -0.1424, -0.0212, 0.01, -0.1445, -0.0233, 0.01, -0.1445, -0.0212, 0.01, -0.1424, -0.0233, 0.01, -0.1424, -0.0168, 0.0061, -0.1392, -0.019, 0.0061, -0.1392, -0.0168, 0.0083, -0.1392, -0.019, 0.0083, -0.1392, -0.0212, 0.0061, -0.1392, -0.0233, 0.0061, -0.1392, -0.0212, 0.0083, -0.1392, -0.0233, 0.0083, -0.1392, -0.0168, 0.01, -0.1392, -0.019, 0.01, -0.1392, -0.0212, 0.01, -0.1392, -0.0233, 0.01, -0.1392,
-0.0255, 0.01, -0.1445, -0.0277, 0.01, -0.1445, -0.0255, 0.01, -0.1424, -0.0277, 0.01, -0.1424, -0.0298, 0.01, -0.1445, -0.032, 0.01, -0.1445, -0.0298, 0.01, -0.1424, -0.032, 0.01, -0.1424, -0.0255, 0.0061, -0.1392, -0.0277, 0.0061, -0.1392, -0.0255, 0.0083, -0.1392, -0.0277, 0.0083, -0.1392, -0.0298, 0.0061, -0.1392, -0.032, 0.0061, -0.1392, -0.0298, 0.0083, -0.1392, -0.032, 0.0083, -0.1392, -0.0255, 0.01, -0.1392, -0.0277, 0.01, -0.1392, -0.0298, 0.01, -0.1392, -0.032, 0.01, -0.1392, -0.0379, 0.0061, -0.1706, -0.0379, 0.0083, -0.1706, -0.0379, 0.0061, -0.1684, -0.0379, 0.0083, -0.1684, -0.0379, 0.01, -0.1706, -0.0407, 0.01, -0.1706, -0.0379, 0.01, -0.1684, -0.0407, 0.01, -0.1684, -0.0379, 0.0061, -0.1662, -0.0379, 0.0083, -0.1662,
-0.0379, 0.0061, -0.1641, -0.0379, 0.0083, -0.1641, -0.0379, 0.01, -0.1662, -0.0407, 0.01, -0.1662, -0.0379, 0.01, -0.1641, -0.0407, 0.01, -0.1641, -0.0487, 0.0061, -0.1706, -0.0487, 0.0083, -0.1706, -0.0487, 0.0061, -0.1684, -0.0487, 0.0083, -0.1684, -0.0429, 0.01, -0.1706, -0.045, 0.01, -0.1706, -0.0429, 0.01, -0.1684, -0.045, 0.01, -0.1684, -0.0472, 0.01, -0.1706, -0.0487, 0.01, -0.1706, -0.0472, 0.01, -0.1684, -0.0487, 0.01, -0.1684, -0.0487, 0.0061, -0.1662, -0.0487, 0.0083, -0.1662, -0.0487, 0.0061, -0.1641, -0.0487, 0.0083, -0.1641, -0.0429, 0.01, -0.1662, -0.045, 0.01, -0.1662, -0.0429, 0.01, -0.1641, -0.045, 0.01, -0.1641, -0.0472, 0.01, -0.1662, -0.0487, 0.01, -0.1662, -0.0472, 0.01, -0.1641, -0.0487, 0.01, -0.1641,
-0.0379, 0.0061, -0.1619, -0.0379, 0.0083, -0.1619, -0.0379, 0.0061, -0.1588, -0.0407, 0.0061, -0.1588, -0.0379, 0.0083, -0.1588, -0.0407, 0.0083, -0.1588, -0.0379, 0.01, -0.1619, -0.0407, 0.01, -0.1619, -0.0379, 0.01, -0.1588, -0.0407, 0.01, -0.1588, -0.0429, 0.0061, -0.1588, -0.045, 0.0061, -0.1588, -0.0429, 0.0083, -0.1588, -0.045, 0.0083, -0.1588, -0.0487, 0.0061, -0.1619, -0.0487, 0.0083, -0.1619, -0.0472, 0.0061, -0.1588, -0.0487, 0.0061, -0.1588, -0.0472, 0.0083, -0.1588, -0.0487, 0.0083, -0.1588, -0.0429, 0.01, -0.1619, -0.045, 0.01, -0.1619, -0.0429, 0.01, -0.1588, -0.045, 0.01, -0.1588, -0.0472, 0.01, -0.1619, -0.0487, 0.01, -0.1619, -0.0472, 0.01, -0.1588, -0.0487, 0.01, -0.1588, -0.0342, 0.0061, -0.151, -0.0363, 0.0061, -0.151,
-0.0342, 0.0083, -0.151, -0.0363, 0.0083, -0.151, -0.0385, 0.0061, -0.151, -0.0407, 0.0061, -0.151, -0.0385, 0.0083, -0.151, -0.0407, 0.0083, -0.151, -0.0342, 0.01, -0.151, -0.0363, 0.01, -0.151, -0.0385, 0.01, -0.151, -0.0407, 0.01, -0.151, -0.0342, 0.01, -0.1489, -0.0363, 0.01, -0.1489, -0.0342, 0.01, -0.1467, -0.0363, 0.01, -0.1467, -0.0385, 0.01, -0.1489, -0.0407, 0.01, -0.1489, -0.0385, 0.01, -0.1467, -0.0407, 0.01, -0.1467, -0.0429, 0.0061, -0.151, -0.045, 0.0061, -0.151, -0.0429, 0.0083, -0.151, -0.045, 0.0083, -0.151, -0.0472, 0.0061, -0.151, -0.0487, 0.0061, -0.151, -0.0472, 0.0083, -0.151, -0.0487, 0.0083, -0.151, -0.0429, 0.01, -0.151, -0.045, 0.01, -0.151, -0.0472, 0.01, -0.151, -0.0487, 0.01, -0.151,
-0.0487, 0.0061, -0.1489, -0.0487, 0.0083, -0.1489, -0.0487, 0.0061, -0.1467, -0.0487, 0.0083, -0.1467, -0.0429, 0.01, -0.1489, -0.045, 0.01, -0.1489, -0.0429, 0.01, -0.1467, -0.045, 0.01, -0.1467, -0.0472, 0.01, -0.1489, -0.0487, 0.01, -0.1489, -0.0472, 0.01, -0.1467, -0.0487, 0.01, -0.1467, -0.0342, 0.01, -0.1445, -0.0363, 0.01, -0.1445, -0.0342, 0.01, -0.1424, -0.0363, 0.01, -0.1424, -0.0385, 0.01, -0.1445, -0.0407, 0.01, -0.1445, -0.0385, 0.01, -0.1424, -0.0407, 0.01, -0.1424, -0.0342, 0.0061, -0.1392, -0.0363, 0.0061, -0.1392, -0.0342, 0.0083, -0.1392, -0.0363, 0.0083, -0.1392, -0.0385, 0.0061, -0.1392, -0.0407, 0.0061, -0.1392, -0.0385, 0.0083, -0.1392, -0.0407, 0.0083, -0.1392, -0.0342, 0.01, -0.1392, -0.0363, 0.01, -0.1392,
-0.0385, 0.01, -0.1392, -0.0407, 0.01, -0.1392, -0.0487, 0.0061, -0.1445, -0.0487, 0.0083, -0.1445, -0.0487, 0.0061, -0.1424, -0.0487, 0.0083, -0.1424, -0.0429, 0.01, -0.1445, -0.045, 0.01, -0.1445, -0.0429, 0.01, -0.1424, -0.045, 0.01, -0.1424, -0.0472, 0.01, -0.1445, -0.0487, 0.01, -0.1445, -0.0472, 0.01, -0.1424, -0.0487, 0.01, -0.1424, -0.0429, 0.0061, -0.1392, -0.045, 0.0061, -0.1392, -0.0429, 0.0083, -0.1392, -0.045, 0.0083, -0.1392, -0.0472, 0.0061, -0.1392, -0.0487, 0.0061, -0.1392, -0.0472, 0.0083, -0.1392, -0.0487, 0.0083, -0.1392, -0.0429, 0.01, -0.1392, -0.045, 0.01, -0.1392, -0.0472, 0.01, -0.1392, -0.0487, 0.01, -0.1392, 0.0005, 0.0061, -0.0775, -0.0016, 0.0061, -0.0775, 0.0005, 0.0083, -0.0775, -0.0016, 0.0083, -0.0775,
-0.0038, 0.0061, -0.0775, -0.006, 0.0061, -0.0775, -0.0038, 0.0083, -0.0775, -0.006, 0.0083, -0.0775, 0.0005, 0.01, -0.0775, -0.0016, 0.01, -0.0775, -0.0038, 0.01, -0.0775, -0.006, 0.01, -0.0775, -0.0081, 0.0061, -0.0775, -0.0103, 0.0061, -0.0775, -0.0081, 0.0083, -0.0775, -0.0103, 0.0083, -0.0775, -0.0125, 0.0061, -0.0775, -0.0146, 0.0061, -0.0775, -0.0125, 0.0083, -0.0775, -0.0146, 0.0083, -0.0775, -0.0081, 0.01, -0.0775, -0.0103, 0.01, -0.0775, -0.0125, 0.01, -0.0775, -0.0146, 0.01, -0.0775, 0.0005, 0.01, -0.0751, -0.0016, 0.01, -0.0751, 0.0005, 0.01, -0.0729, -0.0016, 0.01, -0.0729, -0.0038, 0.01, -0.0751, -0.006, 0.01, -0.0751, -0.0038, 0.01, -0.0729, -0.006, 0.01, -0.0729, 0.0005, 0.01, -0.0707, -0.0016, 0.01, -0.0707,
0.0005, 0.01, -0.0686, -0.0016, 0.01, -0.0686, -0.0038, 0.01, -0.0707, -0.006, 0.01, -0.0707, -0.0038, 0.01, -0.0686, -0.006, 0.01, -0.0686, -0.0081, 0.01, -0.0751, -0.0103, 0.01, -0.0751, -0.0081, 0.01, -0.0729, -0.0103, 0.01, -0.0729, -0.0125, 0.01, -0.0751, -0.0146, 0.01, -0.0751, -0.0125, 0.01, -0.0729, -0.0146, 0.01, -0.0729, -0.0081, 0.01, -0.0707, -0.0103, 0.01, -0.0707, -0.0081, 0.01, -0.0686, -0.0103, 0.01, -0.0686, -0.0125, 0.01, -0.0707, -0.0146, 0.01, -0.0707, -0.0125, 0.01, -0.0686, -0.0146, 0.01, -0.0686, -0.0168, 0.0061, -0.0775, -0.019, 0.0061, -0.0775, -0.0168, 0.0083, -0.0775, -0.019, 0.0083, -0.0775, -0.0212, 0.0061, -0.0775, -0.0233, 0.0061, -0.0775, -0.0212, 0.0083, -0.0775, -0.0233, 0.0083, -0.0775,
-0.0168, 0.01, -0.0775, -0.019, 0.01, -0.0775, -0.0212, 0.01, -0.0775, -0.0233, 0.01, -0.0775, -0.0255, 0.0061, -0.0775, -0.0277, 0.0061, -0.0775, -0.0255, 0.0083, -0.0775, -0.0277, 0.0083, -0.0775, -0.0298, 0.0061, -0.0775, -0.032, 0.0061, -0.0775, -0.0298, 0.0083, -0.0775, -0.032, 0.0083, -0.0775, -0.0255, 0.01, -0.0775, -0.0277, 0.01, -0.0775, -0.0298, 0.01, -0.0775, -0.032, 0.01, -0.0775, -0.0168, 0.01, -0.0751, -0.019, 0.01, -0.0751, -0.0168, 0.01, -0.0729, -0.019, 0.01, -0.0729, -0.0212, 0.01, -0.0751, -0.0233, 0.01, -0.0751, -0.0212, 0.01, -0.0729, -0.0233, 0.01, -0.0729, -0.0168, 0.01, -0.0707, -0.019, 0.01, -0.0707, -0.0168, 0.01, -0.0686, -0.019, 0.01, -0.0686, -0.0212, 0.01, -0.0707, -0.0233, 0.01, -0.0707,
-0.0212, 0.01, -0.0686, -0.0233, 0.01, -0.0686, -0.0255, 0.01, -0.0751, -0.0277, 0.01, -0.0751, -0.0255, 0.01, -0.0729, -0.0277, 0.01, -0.0729, -0.0298, 0.01, -0.0751, -0.032, 0.01, -0.0751, -0.0298, 0.01, -0.0729, -0.032, 0.01, -0.0729, -0.0255, 0.01, -0.0707, -0.0277, 0.01, -0.0707, -0.0255, 0.01, -0.0686, -0.0277, 0.01, -0.0686, -0.0298, 0.01, -0.0707, -0.032, 0.01, -0.0707, -0.0298, 0.01, -0.0686, -0.032, 0.01, -0.0686, -0.0342, 0.0061, -0.0775, -0.0363, 0.0061, -0.0775, -0.0342, 0.0083, -0.0775, -0.0363, 0.0083, -0.0775, -0.0385, 0.0061, -0.0775, -0.0407, 0.0061, -0.0775, -0.0385, 0.0083, -0.0775, -0.0407, 0.0083, -0.0775, -0.0342, 0.01, -0.0775, -0.0363, 0.01, -0.0775, -0.0385, 0.01, -0.0775, -0.0407, 0.01, -0.0775,
-0.0429, 0.0061, -0.0775, -0.045, 0.0061, -0.0775, -0.0429, 0.0083, -0.0775, -0.045, 0.0083, -0.0775, -0.0472, 0.0061, -0.0775, -0.0487, 0.0061, -0.0775, -0.0472, 0.0083, -0.0775, -0.0487, 0.0083, -0.0775, -0.0429, 0.01, -0.0775, -0.045, 0.01, -0.0775, -0.0472, 0.01, -0.0775, -0.0487, 0.01, -0.0775, -0.0342, 0.01, -0.0751, -0.0363, 0.01, -0.0751, -0.0342, 0.01, -0.0729, -0.0363, 0.01, -0.0729, -0.0385, 0.01, -0.0751, -0.0407, 0.01, -0.0751, -0.0385, 0.01, -0.0729, -0.0407, 0.01, -0.0729, -0.0342, 0.01, -0.0707, -0.0363, 0.01, -0.0707, -0.0342, 0.01, -0.0686, -0.0363, 0.01, -0.0686, -0.0385, 0.01, -0.0707, -0.0407, 0.01, -0.0707, -0.0385, 0.01, -0.0686, -0.0407, 0.01, -0.0686, -0.0487, 0.0061, -0.0751, -0.0487, 0.0083, -0.0751,
-0.0487, 0.0061, -0.0729, -0.0487, 0.0083, -0.0729, -0.0429, 0.01, -0.0751, -0.045, 0.01, -0.0751, -0.0429, 0.01, -0.0729, -0.045, 0.01, -0.0729, -0.0472, 0.01, -0.0751, -0.0487, 0.01, -0.0751, -0.0472, 0.01, -0.0729, -0.0487, 0.01, -0.0729, -0.0487, 0.0061, -0.0707, -0.0487, 0.0083, -0.0707, -0.0487, 0.0061, -0.0686, -0.0487, 0.0083, -0.0686, -0.0429, 0.01, -0.0707, -0.045, 0.01, -0.0707, -0.0429, 0.01, -0.0686, -0.045, 0.01, -0.0686, -0.0472, 0.01, -0.0707, -0.0487, 0.01, -0.0707, -0.0472, 0.01, -0.0686, -0.0487, 0.01, -0.0686, 0.0005, 0.0061, -0.0667, -0.0016, 0.0061, -0.0667, 0.0005, 0.0083, -0.0667, -0.0016, 0.0083, -0.0667, -0.0038, 0.0061, -0.0667, -0.006, 0.0061, -0.0667, -0.0038, 0.0083, -0.0667, -0.006, 0.0083, -0.0667,
0.0005, 0.01, -0.0667, -0.0016, 0.01, -0.0667, -0.0038, 0.01, -0.0667, -0.006, 0.01, -0.0667, -0.0081, 0.0061, -0.0667, -0.0103, 0.0061, -0.0667, -0.0081, 0.0083, -0.0667, -0.0103, 0.0083, -0.0667, -0.0125, 0.0061, -0.0667, -0.0146, 0.0061, -0.0667, -0.0125, 0.0083, -0.0667, -0.0146, 0.0083, -0.0667, -0.0081, 0.01, -0.0667, -0.0103, 0.01, -0.0667, -0.0125, 0.01, -0.0667, -0.0146, 0.01, -0.0667, -0.0168, 0.0061, -0.0667, -0.019, 0.0061, -0.0667, -0.0168, 0.0083, -0.0667, -0.019, 0.0083, -0.0667, -0.0212, 0.0061, -0.0667, -0.0233, 0.0061, -0.0667, -0.0212, 0.0083, -0.0667, -0.0233, 0.0083, -0.0667, -0.0168, 0.01, -0.0667, -0.019, 0.01, -0.0667, -0.0212, 0.01, -0.0667, -0.0233, 0.01, -0.0667, -0.0255, 0.0061, -0.0667, -0.0277, 0.0061, -0.0667,
-0.0255, 0.0083, -0.0667, -0.0277, 0.0083, -0.0667, -0.0298, 0.0061, -0.0667, -0.032, 0.0061, -0.0667, -0.0298, 0.0083, -0.0667, -0.032, 0.0083, -0.0667, -0.0255, 0.01, -0.0667, -0.0277, 0.01, -0.0667, -0.0298, 0.01, -0.0667, -0.032, 0.01, -0.0667, 0.0005, 0.0061, -0.05, -0.0016, 0.0061, -0.05, 0.0005, 0.0083, -0.05, -0.0016, 0.0083, -0.05, -0.0038, 0.0061, -0.05, -0.006, 0.0061, -0.05, -0.0038, 0.0083, -0.05, -0.006, 0.0083, -0.05, 0.0005, 0.01, -0.05, -0.0016, 0.01, -0.05, 0.0005, 0.01, -0.0469, -0.0016, 0.01, -0.0469, -0.0038, 0.01, -0.05, -0.006, 0.01, -0.05, -0.0038, 0.01, -0.0469, -0.006, 0.01, -0.0469, 0.0005, 0.01, -0.0447, -0.0016, 0.01, -0.0447, 0.0005, 0.01, -0.0425, -0.0016, 0.01, -0.0425,
-0.0038, 0.01, -0.0447, -0.006, 0.01, -0.0447, -0.0038, 0.01, -0.0425, -0.006, 0.01, -0.0425, -0.0081, 0.0061, -0.05, -0.0103, 0.0061, -0.05, -0.0081, 0.0083, -0.05, -0.0103, 0.0083, -0.05, -0.0125, 0.0061, -0.05, -0.0146, 0.0061, -0.05, -0.0125, 0.0083, -0.05, -0.0146, 0.0083, -0.05, -0.0081, 0.01, -0.05, -0.0103, 0.01, -0.05, -0.0081, 0.01, -0.0469, -0.0103, 0.01, -0.0469, -0.0125, 0.01, -0.05, -0.0146, 0.01, -0.05, -0.0125, 0.01, -0.0469, -0.0146, 0.01, -0.0469, -0.0081, 0.01, -0.0447, -0.0103, 0.01, -0.0447, -0.0081, 0.01, -0.0425, -0.0103, 0.01, -0.0425, -0.0125, 0.01, -0.0447, -0.0146, 0.01, -0.0447, -0.0125, 0.01, -0.0425, -0.0146, 0.01, -0.0425, 0.0013, 0.0061, -0.0382, 0.0013, 0.0083, -0.0382,
0.0005, 0.01, -0.0404, -0.0016, 0.01, -0.0404, 0.0013, 0.01, -0.0382, -0.0016, 0.01, -0.0382, -0.0038, 0.01, -0.0404, -0.006, 0.01, -0.0404, -0.0038, 0.01, -0.0382, -0.006, 0.01, -0.0382, 0.0013, 0.0061, -0.036, 0.0013, 0.0083, -0.036, 0.0013, 0.0061, -0.0338, 0.0013, 0.0083, -0.0338, 0.0013, 0.01, -0.036, -0.0016, 0.01, -0.036, 0.0013, 0.01, -0.0338, -0.0016, 0.01, -0.0338, -0.0038, 0.01, -0.036, -0.006, 0.01, -0.036, -0.0038, 0.01, -0.0338, -0.006, 0.01, -0.0338, -0.0105, 0.0061, -0.0382, -0.0105, 0.0083, -0.0382, -0.0125, 0.0061, -0.0382, -0.0146, 0.0061, -0.0382, -0.0125, 0.0083, -0.0382, -0.0146, 0.0083, -0.0382, -0.0081, 0.01, -0.0404, -0.0103, 0.01, -0.0404, -0.0081, 0.01, -0.0382, -0.0105, 0.01, -0.0382,
-0.0125, 0.01, -0.0404, -0.0146, 0.01, -0.0404, -0.0125, 0.01, -0.0382, -0.0146, 0.01, -0.0382, -0.0105, 0.0061, -0.036, -0.0105, 0.0083, -0.036, -0.0105, 0.0061, -0.0338, -0.0105, 0.0083, -0.0338, -0.0081, 0.01, -0.036, -0.0105, 0.01, -0.036, -0.0081, 0.01, -0.0338, -0.0105, 0.01, -0.0338, -0.0168, 0.0061, -0.05, -0.019, 0.0061, -0.05, -0.0168, 0.0083, -0.05, -0.019, 0.0083, -0.05, -0.0212, 0.0061, -0.05, -0.0233, 0.0061, -0.05, -0.0212, 0.0083, -0.05, -0.0233, 0.0083, -0.05, -0.0168, 0.01, -0.05, -0.019, 0.01, -0.05, -0.0168, 0.01, -0.0469, -0.019, 0.01, -0.0469, -0.0212, 0.01, -0.05, -0.0233, 0.01, -0.05, -0.0212, 0.01, -0.0469, -0.0233, 0.01, -0.0469, -0.0168, 0.01, -0.0447, -0.019, 0.01, -0.0447,
-0.0168, 0.01, -0.0425, -0.019, 0.01, -0.0425, -0.0212, 0.01, -0.0447, -0.0233, 0.01, -0.0447, -0.0212, 0.01, -0.0425, -0.0233, 0.01, -0.0425, -0.0255, 0.0061, -0.05, -0.0277, 0.0061, -0.05, -0.0255, 0.0083, -0.05, -0.0277, 0.0083, -0.05, -0.0298, 0.0061, -0.05, -0.032, 0.0061, -0.05, -0.0298, 0.0083, -0.05, -0.032, 0.0083, -0.05, -0.0255, 0.01, -0.05, -0.0277, 0.01, -0.05, -0.0255, 0.01, -0.0469, -0.0277, 0.01, -0.0469, -0.0298, 0.01, -0.05, -0.032, 0.01, -0.05, -0.0298, 0.01, -0.0469, -0.032, 0.01, -0.0469, -0.0255, 0.01, -0.0447, -0.0277, 0.01, -0.0447, -0.0255, 0.01, -0.0425, -0.0277, 0.01, -0.0425, -0.0298, 0.01, -0.0447, -0.032, 0.01, -0.0447, -0.0298, 0.01, -0.0425, -0.032, 0.01, -0.0425,
-0.0168, 0.0061, -0.0382, -0.019, 0.0061, -0.0382, -0.0168, 0.0083, -0.0382, -0.019, 0.0083, -0.0382, -0.0212, 0.0061, -0.0382, -0.0233, 0.0061, -0.0382, -0.0212, 0.0083, -0.0382, -0.0233, 0.0083, -0.0382, -0.0168, 0.01, -0.0404, -0.019, 0.01, -0.0404, -0.0168, 0.01, -0.0382, -0.019, 0.01, -0.0382, -0.0212, 0.01, -0.0404, -0.0233, 0.01, -0.0404, -0.0212, 0.01, -0.0382, -0.0233, 0.01, -0.0382, -0.0255, 0.0061, -0.0382, -0.0277, 0.0061, -0.0382, -0.0255, 0.0083, -0.0382, -0.0277, 0.0083, -0.0382, -0.0298, 0.0061, -0.0382, -0.032, 0.0061, -0.0382, -0.0298, 0.0083, -0.0382, -0.032, 0.0083, -0.0382, -0.0255, 0.01, -0.0404, -0.0277, 0.01, -0.0404, -0.0255, 0.01, -0.0382, -0.0277, 0.01, -0.0382, -0.0298, 0.01, -0.0404, -0.032, 0.01, -0.0404,
-0.0298, 0.01, -0.0382, -0.032, 0.01, -0.0382, -0.0342, 0.0061, -0.0667, -0.0363, 0.0061, -0.0667, -0.0342, 0.0083, -0.0667, -0.0363, 0.0083, -0.0667, -0.0385, 0.0061, -0.0667, -0.0407, 0.0061, -0.0667, -0.0385, 0.0083, -0.0667, -0.0407, 0.0083, -0.0667, -0.0342, 0.01, -0.0667, -0.0363, 0.01, -0.0667, -0.0385, 0.01, -0.0667, -0.0407, 0.01, -0.0667, -0.0429, 0.0061, -0.0667, -0.045, 0.0061, -0.0667, -0.0429, 0.0083, -0.0667, -0.045, 0.0083, -0.0667, -0.0472, 0.0061, -0.0667, -0.0487, 0.0061, -0.0667, -0.0472, 0.0083, -0.0667, -0.0487, 0.0083, -0.0667, -0.0429, 0.01, -0.0667, -0.045, 0.01, -0.0667, -0.0472, 0.01, -0.0667, -0.0487, 0.01, -0.0667, -0.0342, 0.0061, -0.05, -0.0363, 0.0061, -0.05, -0.0342, 0.0083, -0.05, -0.0363, 0.0083, -0.05,
-0.0385, 0.0061, -0.05, -0.0407, 0.0061, -0.05, -0.0385, 0.0083, -0.05, -0.0407, 0.0083, -0.05, -0.0342, 0.01, -0.05, -0.0363, 0.01, -0.05, -0.0342, 0.01, -0.0469, -0.0363, 0.01, -0.0469, -0.0385, 0.01, -0.05, -0.0407, 0.01, -0.05, -0.0385, 0.01, -0.0469, -0.0407, 0.01, -0.0469, -0.0342, 0.01, -0.0447, -0.0363, 0.01, -0.0447, -0.0342, 0.01, -0.0425, -0.0363, 0.01, -0.0425, -0.0385, 0.01, -0.0447, -0.0407, 0.01, -0.0447, -0.0385, 0.01, -0.0425, -0.0407, 0.01, -0.0425, -0.0429, 0.0061, -0.05, -0.045, 0.0061, -0.05, -0.0429, 0.0083, -0.05, -0.045, 0.0083, -0.05, -0.0472, 0.0061, -0.05, -0.0487, 0.0061, -0.05, -0.0472, 0.0083, -0.05, -0.0487, 0.0083, -0.05, -0.0487, 0.0061, -0.0469, -0.0487, 0.0083, -0.0469,
-0.0429, 0.01, -0.05, -0.045, 0.01, -0.05, -0.0429, 0.01, -0.0469, -0.045, 0.01, -0.0469, -0.0472, 0.01, -0.05, -0.0487, 0.01, -0.05, -0.0472, 0.01, -0.0469, -0.0487, 0.01, -0.0469, -0.0487, 0.0061, -0.0447, -0.0487, 0.0083, -0.0447, -0.0487, 0.0061, -0.0425, -0.0487, 0.0083, -0.0425, -0.0429, 0.01, -0.0447, -0.045, 0.01, -0.0447, -0.0429, 0.01, -0.0425, -0.045, 0.01, -0.0425, -0.0472, 0.01, -0.0447, -0.0487, 0.01, -0.0447, -0.0472, 0.01, -0.0425, -0.0487, 0.01, -0.0425, -0.0342, 0.0061, -0.0382, -0.0363, 0.0061, -0.0382, -0.0342, 0.0083, -0.0382, -0.0363, 0.0083, -0.0382, -0.0379, 0.0061, -0.0382, -0.0379, 0.0083, -0.0382, -0.0342, 0.01, -0.0404, -0.0363, 0.01, -0.0404, -0.0342, 0.01, -0.0382, -0.0363, 0.01, -0.0382,
-0.0385, 0.01, -0.0404, -0.0407, 0.01, -0.0404, -0.0379, 0.01, -0.0382, -0.0407, 0.01, -0.0382, -0.0379, 0.0061, -0.036, -0.0379, 0.0083, -0.036, -0.0379, 0.0061, -0.0338, -0.0379, 0.0083, -0.0338, -0.0379, 0.01, -0.036, -0.0407, 0.01, -0.036, -0.0379, 0.01, -0.0338, -0.0407, 0.01, -0.0338, -0.0487, 0.0061, -0.0404, -0.0487, 0.0083, -0.0404, -0.0487, 0.0061, -0.0382, -0.0487, 0.0083, -0.0382, -0.0429, 0.01, -0.0404, -0.045, 0.01, -0.0404, -0.0429, 0.01, -0.0382, -0.045, 0.01, -0.0382, -0.0472, 0.01, -0.0404, -0.0487, 0.01, -0.0404, -0.0472, 0.01, -0.0382, -0.0487, 0.01, -0.0382, -0.0487, 0.0061, -0.036, -0.0487, 0.0083, -0.036, -0.0487, 0.0061, -0.0338, -0.0487, 0.0083, -0.0338, -0.0429, 0.01, -0.036, -0.045, 0.01, -0.036,
-0.0429, 0.01, -0.0338, -0.045, 0.01, -0.0338, -0.0472, 0.01, -0.036, -0.0487, 0.01, -0.036, -0.0472, 0.01, -0.0338, -0.0487, 0.01, -0.0338, 0.0013, 0.0061, -0.0317, 0.0013, 0.0083, -0.0317, 0.0013, 0.0061, -0.0295, 0.0013, 0.0083, -0.0295, 0.0013, 0.01, -0.0317, -0.0016, 0.01, -0.0317, 0.0013, 0.01, -0.0295, -0.0016, 0.01, -0.0295, -0.0038, 0.01, -0.0317, -0.006, 0.01, -0.0317, -0.0038, 0.01, -0.0295, -0.006, 0.01, -0.0295, 0.0013, 0.0061, -0.0273, 0.0013, 0.0083, -0.0273, 0.0013, 0.0061, -0.0252, 0.0013, 0.0083, -0.0252, 0.0013, 0.01, -0.0273, -0.0016, 0.01, -0.0273, 0.0013, 0.01, -0.0252, -0.0016, 0.01, -0.0252, -0.0038, 0.01, -0.0273, -0.006, 0.01, -0.0273, -0.0038, 0.01, -0.0252, -0.006, 0.01, -0.0252,
-0.0105, 0.0061, -0.0317, -0.0105, 0.0083, -0.0317, -0.0105, 0.0061, -0.0295, -0.0105, 0.0083, -0.0295, -0.0081, 0.01, -0.0317, -0.0105, 0.01, -0.0317, -0.0081, 0.01, -0.0295, -0.0105, 0.01, -0.0295, -0.0105, 0.0061, -0.0273, -0.0105, 0.0083, -0.0273, -0.0105, 0.0061, -0.0252, -0.0105, 0.0083, -0.0252, -0.0081, 0.01, -0.0273, -0.0105, 0.01, -0.0273, -0.0081, 0.01, -0.0252, -0.0105, 0.01, -0.0252, 0.0013, 0.0061, -0.023, 0.0013, 0.0083, -0.023, 0.0013, 0.0061, -0.0208, 0.0013, 0.0083, -0.0208, 0.0013, 0.01, -0.023, -0.0016, 0.01, -0.023, 0.0013, 0.01, -0.0208, -0.0016, 0.01, -0.0208, -0.0038, 0.01, -0.023, -0.006, 0.01, -0.023, -0.0038, 0.01, -0.0208, -0.006, 0.01, -0.0208, 0.0013, 0.0061, -0.0187, 0.0013, 0.0083, -0.0187,
0.0013, 0.0061, -0.0165, 0.0013, 0.0083, -0.0165, 0.0013, 0.01, -0.0187, -0.0016, 0.01, -0.0187, 0.0013, 0.01, -0.0165, -0.0016, 0.01, -0.0165, -0.0038, 0.01, -0.0187, -0.006, 0.01, -0.0187, -0.0038, 0.01, -0.0165, -0.006, 0.01, -0.0165, -0.0105, 0.0061, -0.023, -0.0105, 0.0083, -0.023, -0.0105, 0.0061, -0.0208, -0.0105, 0.0083, -0.0208, -0.0081, 0.01, -0.023, -0.0105, 0.01, -0.023, -0.0081, 0.01, -0.0208, -0.0105, 0.01, -0.0208, -0.0105, 0.0061, -0.0187, -0.0105, 0.0083, -0.0187, -0.0105, 0.0061, -0.0165, -0.0105, 0.0083, -0.0165, -0.0081, 0.01, -0.0187, -0.0105, 0.01, -0.0187, -0.0081, 0.01, -0.0165, -0.0105, 0.01, -0.0165, 0.0013, 0.0061, -0.0143, 0.0013, 0.0083, -0.0143, 0.0013, 0.0061, -0.0121, 0.0013, 0.0083, -0.0121,
0.0013, 0.01, -0.0143, -0.0016, 0.01, -0.0143, 0.0013, 0.01, -0.0121, -0.0016, 0.01, -0.0121, -0.0038, 0.01, -0.0143, -0.006, 0.01, -0.0143, -0.0038, 0.01, -0.0121, -0.006, 0.01, -0.0121, 0.0013, 0.0061, -0.01, 0.0013, 0.0083, -0.01, 0.0013, 0.0061, -0.0078, 0.0013, 0.0083, -0.0078, 0.0013, 0.01, -0.01, -0.0016, 0.01, -0.01, 0.0013, 0.01, -0.0078, -0.0016, 0.01, -0.0078, -0.0038, 0.01, -0.01, -0.006, 0.01, -0.01, -0.0038, 0.01, -0.0078, -0.006, 0.01, -0.0078, -0.0105, 0.0061, -0.0143, -0.0105, 0.0083, -0.0143, -0.0105, 0.0061, -0.0121, -0.0105, 0.0083, -0.0121, -0.0081, 0.01, -0.0143, -0.0105, 0.01, -0.0143, -0.0081, 0.01, -0.0121, -0.0105, 0.01, -0.0121, -0.0105, 0.0061, -0.01, -0.0105, 0.0083, -0.01,
-0.0105, 0.0061, -0.0078, -0.0105, 0.0083, -0.0078, -0.0081, 0.01, -0.01, -0.0105, 0.01, -0.01, -0.0081, 0.01, -0.0078, -0.0105, 0.01, -0.0078, 0.0013, 0.0061, -0.0056, 0.0013, 0.0083, -0.0056, 0.0013, 0.0061, -0.0035, 0.0013, 0.0083, -0.0035, 0.0013, 0.01, -0.0056, -0.0016, 0.01, -0.0056, 0.0013, 0.01, -0.0035, -0.0016, 0.01, -0.0035, -0.0038, 0.01, -0.0056, -0.006, 0.01, -0.0056, -0.0038, 0.01, -0.0035, -0.006, 0.01, -0.0035, 0.0013, 0.0061, -0.0013, 0.0013, 0.0083, -0.0013, 0.0013, 0.0061, 0.0009, 0.0013, 0.0083, 0.0009, 0.0013, 0.01, -0.0013, -0.0016, 0.01, -0.0013, 0.0013, 0.01, 0.0009, -0.0016, 0.01, 0.0009, -0.0038, 0.01, -0.0013, -0.006, 0.01, -0.0013, -0.0038, 0.01, 0.0009, -0.006, 0.01, 0.0009,
-0.0105, 0.0061, -0.0056, -0.0105, 0.0083, -0.0056, -0.0105, 0.0061, -0.0035, -0.0105, 0.0083, -0.0035, -0.0081, 0.01, -0.0056, -0.0105, 0.01, -0.0056, -0.0081, 0.01, -0.0035, -0.0105, 0.01, -0.0035, -0.0105, 0.0061, -0.0013, -0.0105, 0.0083, -0.0013, -0.0105, 0.0061, 0.0009, -0.0105, 0.0083, 0.0009, -0.0081, 0.01, -0.0013, -0.0105, 0.01, -0.0013, -0.0081, 0.01, 0.0009, -0.0105, 0.01, 0.0009, -0.0379, 0.0061, -0.0317, -0.0379, 0.0083, -0.0317, -0.0379, 0.0061, -0.0295, -0.0379, 0.0083, -0.0295, -0.0379, 0.01, -0.0317, -0.0407, 0.01, -0.0317, -0.0379, 0.01, -0.0295, -0.0407, 0.01, -0.0295, -0.0379, 0.0061, -0.0273, -0.0379, 0.0083, -0.0273, -0.0379, 0.0061, -0.0252, -0.0379, 0.0083, -0.0252, -0.0379, 0.01, -0.0273, -0.0407, 0.01, -0.0273,
-0.0379, 0.01, -0.0252, -0.0407, 0.01, -0.0252, -0.0487, 0.0061, -0.0317, -0.0487, 0.0083, -0.0317, -0.0487, 0.0061, -0.0295, -0.0487, 0.0083, -0.0295, -0.0429, 0.01, -0.0317, -0.045, 0.01, -0.0317, -0.0429, 0.01, -0.0295, -0.045, 0.01, -0.0295, -0.0472, 0.01, -0.0317, -0.0487, 0.01, -0.0317, -0.0472, 0.01, -0.0295, -0.0487, 0.01, -0.0295, -0.0487, 0.0061, -0.0273, -0.0487, 0.0083, -0.0273, -0.0487, 0.0061, -0.0252, -0.0487, 0.0083, -0.0252, -0.0429, 0.01, -0.0273, -0.045, 0.01, -0.0273, -0.0429, 0.01, -0.0252, -0.045, 0.01, -0.0252, -0.0472, 0.01, -0.0273, -0.0487, 0.01, -0.0273, -0.0472, 0.01, -0.0252, -0.0487, 0.01, -0.0252, -0.0379, 0.0061, -0.023, -0.0379, 0.0083, -0.023, -0.0379, 0.0061, -0.0208, -0.0379, 0.0083, -0.0208,
-0.0379, 0.01, -0.023, -0.0407, 0.01, -0.023, -0.0379, 0.01, -0.0208, -0.0407, 0.01, -0.0208, -0.0379, 0.0061, -0.0187, -0.0379, 0.0083, -0.0187, -0.0379, 0.0061, -0.0165, -0.0379, 0.0083, -0.0165, -0.0379, 0.01, -0.0187, -0.0407, 0.01, -0.0187, -0.0379, 0.01, -0.0165, -0.0407, 0.01, -0.0165, -0.0487, 0.0061, -0.023, -0.0487, 0.0083, -0.023, -0.0487, 0.0061, -0.0208, -0.0487, 0.0083, -0.0208, -0.0429, 0.01, -0.023, -0.045, 0.01, -0.023, -0.0429, 0.01, -0.0208, -0.045, 0.01, -0.0208, -0.0472, 0.01, -0.023, -0.0487, 0.01, -0.023, -0.0472, 0.01, -0.0208, -0.0487, 0.01, -0.0208, -0.0487, 0.0061, -0.0187, -0.0487, 0.0083, -0.0187, -0.0487, 0.0061, -0.0165, -0.0487, 0.0083, -0.0165, -0.0429, 0.01, -0.0187, -0.045, 0.01, -0.0187,
-0.0429, 0.01, -0.0165, -0.045, 0.01, -0.0165, -0.0472, 0.01, -0.0187, -0.0487, 0.01, -0.0187, -0.0472, 0.01, -0.0165, -0.0487, 0.01, -0.0165, -0.0379, 0.0061, -0.0143, -0.0379, 0.0083, -0.0143, -0.0379, 0.0061, -0.0121, -0.0379, 0.0083, -0.0121, -0.0379, 0.01, -0.0143, -0.0407, 0.01, -0.0143, -0.0379, 0.01, -0.0121, -0.0407, 0.01, -0.0121, -0.0379, 0.0061, -0.01, -0.0379, 0.0083, -0.01, -0.0379, 0.0061, -0.0078, -0.0379, 0.0083, -0.0078, -0.0379, 0.01, -0.01, -0.0407, 0.01, -0.01, -0.0379, 0.01, -0.0078, -0.0407, 0.01, -0.0078, -0.0487, 0.0061, -0.0143, -0.0487, 0.0083, -0.0143, -0.0487, 0.0061, -0.0121, -0.0487, 0.0083, -0.0121, -0.0429, 0.01, -0.0143, -0.045, 0.01, -0.0143, -0.0429, 0.01, -0.0121, -0.045, 0.01, -0.0121,
-0.0472, 0.01, -0.0143, -0.0487, 0.01, -0.0143, -0.0472, 0.01, -0.0121, -0.0487, 0.01, -0.0121, -0.0487, 0.0061, -0.01, -0.0487, 0.0083, -0.01, -0.0487, 0.0061, -0.0078, -0.0487, 0.0083, -0.0078, -0.0429, 0.01, -0.01, -0.045, 0.01, -0.01, -0.0429, 0.01, -0.0078, -0.045, 0.01, -0.0078, -0.0472, 0.01, -0.01, -0.0487, 0.01, -0.01, -0.0472, 0.01, -0.0078, -0.0487, 0.01, -0.0078, -0.0379, 0.0061, -0.0056, -0.0379, 0.0083, -0.0056, -0.0379, 0.0061, -0.0035, -0.0379, 0.0083, -0.0035, -0.0379, 0.01, -0.0056, -0.0407, 0.01, -0.0056, -0.0379, 0.01, -0.0035, -0.0407, 0.01, -0.0035, -0.0379, 0.0061, -0.0013, -0.0379, 0.0083, -0.0013, -0.0379, 0.0061, 0.0009, -0.0379, 0.0083, 0.0009, -0.0379, 0.01, -0.0013, -0.0407, 0.01, -0.0013,
-0.0379, 0.01, 0.0008, -0.0407, 0.01, 0.0009, -0.0487, 0.0061, -0.0056, -0.0487, 0.0083, -0.0056, -0.0487, 0.0061, -0.0035, -0.0487, 0.0083, -0.0035, -0.0429, 0.01, -0.0056, -0.045, 0.01, -0.0056, -0.0429, 0.01, -0.0035, -0.045, 0.01, -0.0035, -0.0472, 0.01, -0.0056, -0.0487, 0.01, -0.0056, -0.0472, 0.01, -0.0035, -0.0487, 0.01, -0.0035, -0.0487, 0.0061, -0.0013, -0.0487, 0.0083, -0.0013, -0.0487, 0.0061, 0.0009, -0.0487, 0.0083, 0.0009, -0.0429, 0.01, -0.0013, -0.045, 0.01, -0.0013, -0.0429, 0.01, 0.0009, -0.045, 0.01, 0.0009, -0.0472, 0.01, -0.0013, -0.0487, 0.01, -0.0013, -0.0472, 0.01, 0.0009, -0.0487, 0.01, 0.0009, 0.067, -0.0, 0.0461, 0.067, -0.0, 0.0486, 0.0656, -0.0, 0.0461, 0.0635, -0.0, 0.0461,
0.0656, -0.0, 0.0486, 0.0635, -0.0, 0.0486, 0.067, 0.0017, 0.0461, 0.067, 0.0039, 0.0461, 0.067, 0.0017, 0.0486, 0.067, 0.0039, 0.0486, 0.0656, 0.0017, 0.0461, 0.0635, 0.0017, 0.0461, 0.0656, 0.0039, 0.0461, 0.0635, 0.0039, 0.0461, 0.067, -0.0, 0.05, 0.0656, -0.0, 0.05, 0.0635, -0.0, 0.05, 0.067, 0.0017, 0.05, 0.067, 0.0039, 0.05, 0.0656, 0.0017, 0.05, 0.0635, 0.0017, 0.05, 0.0656, 0.0039, 0.05, 0.0635, 0.0039, 0.05, 0.0613, -0.0, 0.0461, 0.0591, -0.0, 0.0461, 0.0613, -0.0, 0.0486, 0.0591, -0.0, 0.0486, 0.057, -0.0, 0.0461, 0.0548, -0.0, 0.0461, 0.057, -0.0, 0.0486, 0.0548, -0.0, 0.0486, 0.0613, 0.0017, 0.0461, 0.0591, 0.0017, 0.0461, 0.0613, 0.0039, 0.0461,
0.0591, 0.0039, 0.0461, 0.057, 0.0017, 0.0461, 0.0548, 0.0017, 0.0461, 0.057, 0.0039, 0.0461, 0.0548, 0.0039, 0.0461, 0.0613, -0.0, 0.05, 0.0591, -0.0, 0.05, 0.057, -0.0, 0.05, 0.0548, -0.0, 0.05, 0.0562, -0.0, 0.0529, 0.0548, -0.0, 0.0529, 0.0613, 0.0017, 0.05, 0.0591, 0.0017, 0.05, 0.0613, 0.0039, 0.05, 0.0591, 0.0039, 0.05, 0.057, 0.0017, 0.05, 0.0548, 0.0017, 0.05, 0.057, 0.0039, 0.05, 0.0548, 0.0039, 0.05, 0.0562, 0.0017, 0.0529, 0.0548, 0.0017, 0.0529, 0.0562, 0.0039, 0.0529, 0.0548, 0.0039, 0.0529, 0.0526, -0.0, 0.0461, 0.0505, -0.0, 0.0461, 0.0526, -0.0, 0.0486, 0.0505, -0.0, 0.0486, 0.0483, -0.0, 0.0461, 0.0461, -0.0, 0.0461, 0.0483, -0.0, 0.0486,
0.0461, -0.0, 0.0486, 0.0526, 0.0017, 0.0461, 0.0505, 0.0017, 0.0461, 0.0526, 0.0039, 0.0461, 0.0505, 0.0039, 0.0461, 0.0483, 0.0017, 0.0461, 0.0461, 0.0017, 0.0461, 0.0483, 0.0039, 0.0461, 0.0461, 0.0039, 0.0461, 0.0526, -0.0, 0.05, 0.0505, -0.0, 0.05, 0.0523, -0.0, 0.0529, 0.0483, -0.0, 0.05, 0.0454, -0.0, 0.05, 0.0454, -0.0, 0.053, 0.0526, 0.0017, 0.05, 0.0505, 0.0017, 0.05, 0.0526, 0.0039, 0.05, 0.0505, 0.0039, 0.05, 0.0523, 0.0017, 0.0529, 0.0523, 0.0039, 0.0529, 0.0483, 0.0017, 0.05, 0.0454, 0.0017, 0.05, 0.0483, 0.0039, 0.05, 0.0454, 0.0039, 0.05, 0.0454, 0.0017, 0.053, 0.0454, 0.0039, 0.053, 0.0439, -0.0, 0.0461, 0.0425, -0.0, 0.0461, 0.0439, -0.0, 0.0486,
0.0425, -0.0, 0.0486, 0.0439, 0.0017, 0.0461, 0.0425, 0.0017, 0.0461, 0.0439, 0.0039, 0.0461, 0.0425, 0.0039, 0.0461, 0.0425, 0.0017, 0.0486, 0.0425, 0.0039, 0.0486, 0.0439, -0.0, 0.0508, 0.0425, -0.0, 0.0508, 0.0439, -0.0, 0.053, 0.0425, -0.0, 0.053, 0.0425, 0.0017, 0.0508, 0.0425, 0.0039, 0.0508, 0.0425, 0.0017, 0.053, 0.0425, 0.0039, 0.053, 0.0562, -0.0, 0.0551, 0.0548, -0.0, 0.0551, 0.0562, -0.0, 0.0573, 0.0548, -0.0, 0.0573, 0.0562, 0.0017, 0.0551, 0.0562, 0.0039, 0.0551, 0.0562, 0.0017, 0.0573, 0.0562, 0.0039, 0.0573, 0.0562, -0.0, 0.0595, 0.0548, -0.0, 0.0595, 0.0562, -0.0, 0.0616, 0.0548, -0.0, 0.0616, 0.0562, 0.0017, 0.0595, 0.0562, 0.0039, 0.0595, 0.0562, 0.0017, 0.0616,
0.0562, 0.0039, 0.0616, 0.0562, -0.0, 0.0638, 0.0548, -0.0, 0.0638, 0.0562, -0.0, 0.0657, 0.0548, -0.0, 0.0657, 0.0562, 0.0017, 0.0638, 0.0562, 0.0039, 0.0638, 0.0562, 0.0017, 0.0657, 0.0548, 0.0017, 0.0657, 0.0562, 0.0039, 0.0657, 0.0548, 0.0039, 0.0657, 0.0523, -0.0, 0.0551, 0.0523, -0.0, 0.0573, 0.0454, -0.0, 0.0551, 0.0454, -0.0, 0.0573, 0.0523, 0.0017, 0.0551, 0.0523, 0.0039, 0.0551, 0.0523, 0.0017, 0.0573, 0.0523, 0.0039, 0.0573, 0.0454, 0.0017, 0.0551, 0.0454, 0.0039, 0.0551, 0.0454, 0.0017, 0.0573, 0.0454, 0.0039, 0.0573, 0.0523, -0.0, 0.0595, 0.0523, -0.0, 0.0616, 0.0454, -0.0, 0.0595, 0.0454, -0.0, 0.0616, 0.0523, 0.0017, 0.0595, 0.0523, 0.0039, 0.0595, 0.0523, 0.0017, 0.0616,
0.0523, 0.0039, 0.0616, 0.0454, 0.0017, 0.0595, 0.0454, 0.0039, 0.0595, 0.0454, 0.0017, 0.0616, 0.0454, 0.0039, 0.0616, 0.0439, -0.0, 0.0551, 0.0425, -0.0, 0.0551, 0.0439, -0.0, 0.0573, 0.0425, -0.0, 0.0573, 0.0425, 0.0017, 0.0551, 0.0425, 0.0039, 0.0551, 0.0425, 0.0017, 0.0573, 0.0425, 0.0039, 0.0573, 0.0439, -0.0, 0.0595, 0.0425, -0.0, 0.0595, 0.0439, -0.0, 0.0616, 0.0425, -0.0, 0.0616, 0.0425, 0.0017, 0.0595, 0.0425, 0.0039, 0.0595, 0.0425, 0.0017, 0.0616, 0.0425, 0.0039, 0.0616, 0.0523, -0.0, 0.0638, 0.0523, -0.0, 0.0657, 0.0454, -0.0, 0.0638, 0.0454, -0.0, 0.0667, 0.0523, 0.0017, 0.0638, 0.0523, 0.0039, 0.0638, 0.0523, 0.0017, 0.0657, 0.0523, 0.0039, 0.0657, 0.0454, 0.0017, 0.0638,
0.0454, 0.0039, 0.0638, 0.0454, 0.0017, 0.0667, 0.0454, 0.0039, 0.0667, 0.0439, -0.0, 0.0638, 0.0425, -0.0, 0.0638, 0.0439, -0.0, 0.0667, 0.0425, -0.0, 0.0667, 0.0425, 0.0017, 0.0638, 0.0425, 0.0039, 0.0638, 0.0439, 0.0017, 0.0667, 0.0425, 0.0017, 0.0667, 0.0439, 0.0039, 0.0667, 0.0425, 0.0039, 0.0667, 0.0624, -0.0, 0.0887, 0.0625, 0.0017, 0.0887, 0.0625, 0.0039, 0.0887, 0.0616, -0.0, 0.088, 0.0594, -0.0, 0.087, 0.0565, -0.0, 0.0864, 0.0548, -0.0, 0.0863, 0.0573, -0.0, 0.0865, 0.0548, -0.0, 0.0877, 0.0614, 0.0017, 0.0879, 0.0592, 0.0017, 0.0869, 0.0614, 0.0039, 0.0879, 0.0592, 0.0039, 0.0869, 0.0568, 0.0017, 0.0864, 0.0548, 0.0017, 0.0863, 0.0568, 0.0039, 0.0864, 0.0548, 0.0039, 0.0863,
0.0579, 0.0017, 0.0866, 0.0579, 0.0039, 0.0866, 0.0526, -0.0, 0.0863, 0.0513, -0.0, 0.0865, 0.0526, -0.0, 0.0877, 0.0503, -0.0, 0.0867, 0.048, -0.0, 0.0875, 0.0468, -0.0, 0.0884, 0.0526, 0.0017, 0.0863, 0.0511, 0.0017, 0.0865, 0.0526, 0.0039, 0.0863, 0.0511, 0.0039, 0.0865, 0.0501, 0.0017, 0.0867, 0.0501, 0.0039, 0.0867, 0.0482, 0.0017, 0.0874, 0.0468, 0.0017, 0.0884, 0.0482, 0.0039, 0.0874, 0.0468, 0.0039, 0.0884, 0.0646, -0.0, 0.0909, 0.0637, -0.0, 0.0897, 0.0652, -0.0, 0.0918, 0.0635, -0.0, 0.092, 0.0646, 0.0017, 0.0909, 0.0637, 0.0017, 0.0897, 0.0646, 0.0039, 0.0909, 0.0637, 0.0039, 0.0897, 0.0653, 0.0017, 0.092, 0.0653, 0.0039, 0.092, 0.0668, -0.0, 0.0972, 0.0661, -0.0, 0.0939,
0.0635, -0.0, 0.0942, 0.0667, -0.0, 0.0962, 0.0627, -0.0, 0.0963, 0.0668, 0.0017, 0.097, 0.0668, 0.0039, 0.097, 0.0662, 0.0017, 0.0942, 0.0662, 0.0039, 0.0942, 0.0666, 0.0017, 0.0959, 0.0627, 0.0017, 0.0964, 0.0666, 0.0039, 0.0959, 0.0627, 0.0039, 0.0964, 0.0613, -0.0, 0.0899, 0.0586, -0.0, 0.091, 0.0608, -0.0, 0.0926, 0.0595, -0.0, 0.0915, 0.0568, -0.0, 0.0904, 0.0548, -0.0, 0.0902, 0.0583, 0.0017, 0.0908, 0.0583, 0.0039, 0.0908, 0.0608, 0.0017, 0.0926, 0.0594, 0.0017, 0.0914, 0.0608, 0.0039, 0.0926, 0.0594, 0.0039, 0.0914, 0.057, 0.0017, 0.0904, 0.0548, 0.0017, 0.0903, 0.057, 0.0039, 0.0904, 0.0548, 0.0039, 0.0903, 0.0621, -0.0, 0.0942, 0.0624, -0.0, 0.0953, 0.0621, 0.0017, 0.0942,
0.0621, 0.0039, 0.0942, 0.0624, 0.0017, 0.0953, 0.0624, 0.0039, 0.0953, 0.0669, -0.0, 0.0985, 0.067, -0.0, 0.1007, 0.0656, -0.0, 0.0985, 0.063, -0.0, 0.0986, 0.0656, -0.0, 0.1007, 0.0631, -0.0, 0.1007, 0.0669, 0.0017, 0.0985, 0.0669, 0.0039, 0.0985, 0.067, 0.0017, 0.1007, 0.067, 0.0039, 0.1007, 0.063, 0.0017, 0.0985, 0.063, 0.0039, 0.0985, 0.0631, 0.0017, 0.1007, 0.0631, 0.0039, 0.1007, 0.067, -0.0, 0.1029, 0.0669, -0.0, 0.105, 0.0656, -0.0, 0.1029, 0.0631, -0.0, 0.1029, 0.0656, -0.0, 0.105, 0.063, -0.0, 0.105, 0.067, 0.0017, 0.1029, 0.067, 0.0039, 0.1029, 0.067, 0.0017, 0.105, 0.067, 0.0039, 0.105, 0.0631, 0.0017, 0.1029, 0.0631, 0.0039, 0.1029, 0.063, 0.0017, 0.105,
0.063, 0.0039, 0.105, 0.0528, -0.0, 0.0904, 0.0508, -0.0, 0.0909, 0.0499, -0.0, 0.0913, 0.0483, -0.0, 0.0899, 0.0456, -0.0, 0.0894, 0.0485, -0.0, 0.0922, 0.0461, -0.0, 0.092, 0.0526, 0.0017, 0.0904, 0.0512, 0.0017, 0.0908, 0.0526, 0.0039, 0.0904, 0.0512, 0.0039, 0.0908, 0.05, 0.0017, 0.0912, 0.05, 0.0039, 0.0912, 0.0456, 0.0017, 0.0894, 0.0456, 0.0039, 0.0894, 0.0485, 0.0017, 0.0923, 0.0485, 0.0039, 0.0923, 0.0475, -0.0, 0.0934, 0.0469, -0.0, 0.0945, 0.046, -0.0, 0.0965, 0.0474, 0.0017, 0.0935, 0.0468, 0.0017, 0.0946, 0.0474, 0.0039, 0.0935, 0.0468, 0.0039, 0.0946, 0.0461, 0.0017, 0.0963, 0.0461, 0.0039, 0.0963, 0.0447, -0.0, 0.0905, 0.0438, -0.0, 0.0917, 0.0447, 0.0017, 0.0905,
0.0447, 0.0039, 0.0905, 0.0437, 0.0017, 0.092, 0.0437, 0.0039, 0.092, 0.0429, -0.0, 0.0939, 0.0427, -0.0, 0.0949, 0.0439, -0.0, 0.0964, 0.0425, -0.0, 0.0963, 0.043, 0.0017, 0.0937, 0.0427, 0.0017, 0.0948, 0.043, 0.0039, 0.0937, 0.0427, 0.0039, 0.0948, 0.0425, 0.0017, 0.0964, 0.0425, 0.0039, 0.0964, 0.0456, -0.0, 0.0986, 0.0454, -0.0, 0.1007, 0.0456, 0.0017, 0.0985, 0.0456, 0.0039, 0.0985, 0.0454, 0.0017, 0.1007, 0.0454, 0.0039, 0.1007, 0.0455, -0.0, 0.1029, 0.0457, -0.0, 0.105, 0.0455, 0.0017, 0.1029, 0.0455, 0.0039, 0.1029, 0.0457, 0.0017, 0.1051, 0.0457, 0.0039, 0.1051, 0.0439, -0.0, 0.0985, 0.0425, -0.0, 0.0985, 0.0439, -0.0, 0.1007, 0.0425, -0.0, 0.1007, 0.0425, 0.0017, 0.0985,
0.0425, 0.0039, 0.0985, 0.0425, 0.0017, 0.1007, 0.0425, 0.0039, 0.1007, 0.0439, -0.0, 0.1029, 0.0425, -0.0, 0.1029, 0.0439, -0.0, 0.105, 0.0425, -0.0, 0.105, 0.0425, 0.0017, 0.1029, 0.0425, 0.0039, 0.1029, 0.0425, 0.0017, 0.105, 0.0425, 0.0039, 0.105, 0.0219, -0.0, 0.0716, 0.0201, -0.0, 0.0716, 0.0219, -0.0, 0.0747, 0.0201, -0.0, 0.0747, 0.0219, 0.0017, 0.0716, 0.0201, 0.0017, 0.0716, 0.0219, 0.0039, 0.0716, 0.0201, 0.0039, 0.0716, 0.0219, 0.0017, 0.0747, 0.0219, 0.0039, 0.0747, 0.0219, -0.0, 0.0768, 0.0201, -0.0, 0.0768, 0.0219, -0.0, 0.079, 0.0201, -0.0, 0.079, 0.0219, 0.0017, 0.0768, 0.0219, 0.0039, 0.0768, 0.0219, 0.0017, 0.079, 0.0219, 0.0039, 0.079, 0.0219, -0.0, 0.0812,
0.0201, -0.0, 0.0812, 0.0219, -0.0, 0.0833, 0.0201, -0.0, 0.0833, 0.0219, 0.0017, 0.0812, 0.0219, 0.0039, 0.0812, 0.0219, 0.0017, 0.0833, 0.0201, 0.0017, 0.0833, 0.0219, 0.0039, 0.0833, 0.0201, 0.0039, 0.0833, 0.0179, -0.0, 0.0716, 0.0157, -0.0, 0.0716, 0.0179, -0.0, 0.0747, 0.0157, -0.0, 0.0747, 0.0136, -0.0, 0.0716, 0.0114, -0.0, 0.0716, 0.0136, -0.0, 0.0747, 0.0114, -0.0, 0.0747, 0.0179, 0.0017, 0.0716, 0.0157, 0.0017, 0.0716, 0.0179, 0.0039, 0.0716, 0.0157, 0.0039, 0.0716, 0.0136, 0.0017, 0.0716, 0.0114, 0.0017, 0.0716, 0.0136, 0.0039, 0.0716, 0.0114, 0.0039, 0.0716, 0.0179, -0.0, 0.0768, 0.0157, -0.0, 0.0768, 0.0179, -0.0, 0.079, 0.0157, -0.0, 0.079, 0.0136, -0.0, 0.0768,
0.0114, -0.0, 0.0768, 0.0136, -0.0, 0.079, 0.0114, -0.0, 0.079, 0.0092, -0.0, 0.0716, 0.0071, -0.0, 0.0716, 0.0092, -0.0, 0.0747, 0.0071, -0.0, 0.0747, 0.0049, -0.0, 0.0716, 0.0027, -0.0, 0.0716, 0.0049, -0.0, 0.0747, 0.0027, -0.0, 0.0747, 0.0092, 0.0017, 0.0716, 0.0071, 0.0017, 0.0716, 0.0092, 0.0039, 0.0716, 0.0071, 0.0039, 0.0716, 0.0049, 0.0017, 0.0716, 0.0027, 0.0017, 0.0716, 0.0049, 0.0039, 0.0716, 0.0027, 0.0039, 0.0716, 0.0092, -0.0, 0.0768, 0.0071, -0.0, 0.0768, 0.0092, -0.0, 0.079, 0.0071, -0.0, 0.079, 0.0049, -0.0, 0.0768, 0.0027, -0.0, 0.0768, 0.0049, -0.0, 0.079, 0.0027, -0.0, 0.079, 0.0179, -0.0, 0.0812, 0.0157, -0.0, 0.0812, 0.0179, -0.0, 0.0833,
0.0157, -0.0, 0.0833, 0.0136, -0.0, 0.0812, 0.0114, -0.0, 0.0812, 0.0136, -0.0, 0.0833, 0.0114, -0.0, 0.0833, 0.0179, 0.0017, 0.0833, 0.0157, 0.0017, 0.0833, 0.0179, 0.0039, 0.0833, 0.0157, 0.0039, 0.0833, 0.0136, 0.0017, 0.0833, 0.0114, 0.0017, 0.0833, 0.0136, 0.0039, 0.0833, 0.0114, 0.0039, 0.0833, 0.0092, -0.0, 0.0812, 0.0071, -0.0, 0.0812, 0.0092, -0.0, 0.0833, 0.0071, -0.0, 0.0833, 0.0049, -0.0, 0.0812, 0.0027, -0.0, 0.0812, 0.0049, -0.0, 0.0833, 0.0027, -0.0, 0.0833, 0.0092, 0.0017, 0.0833, 0.0071, 0.0017, 0.0833, 0.0092, 0.0039, 0.0833, 0.0071, 0.0039, 0.0833, 0.0049, 0.0017, 0.0833, 0.0027, 0.0017, 0.0833, 0.0049, 0.0039, 0.0833, 0.0027, 0.0039, 0.0833, 0.0668, -0.0, 0.1065,
0.0667, -0.0, 0.1074, 0.0626, -0.0, 0.1072, 0.0661, -0.0, 0.1097, 0.0635, -0.0, 0.1094, 0.0668, 0.0017, 0.1066, 0.0668, 0.0039, 0.1066, 0.0666, 0.0017, 0.1077, 0.0627, 0.0017, 0.1071, 0.0666, 0.0039, 0.1077, 0.0627, 0.0039, 0.1071, 0.0662, 0.0017, 0.1094, 0.0662, 0.0039, 0.1094, 0.0653, -0.0, 0.1113, 0.0646, -0.0, 0.1123, 0.0636, -0.0, 0.1138, 0.0652, 0.0017, 0.1115, 0.0645, 0.0017, 0.1125, 0.0652, 0.0039, 0.1115, 0.0645, 0.0039, 0.1125, 0.0635, 0.0017, 0.1138, 0.0635, 0.0039, 0.1138, 0.0623, -0.0, 0.1082, 0.062, -0.0, 0.1093, 0.0624, 0.0017, 0.1082, 0.0624, 0.0039, 0.1082, 0.062, 0.0017, 0.1094, 0.062, 0.0039, 0.1094, 0.0608, -0.0, 0.1112, 0.0596, -0.0, 0.1122, 0.0622, -0.0, 0.1148,
0.0587, -0.0, 0.1127, 0.0569, -0.0, 0.1134, 0.0548, -0.0, 0.1136, 0.0608, 0.0017, 0.1111, 0.0596, 0.0017, 0.1122, 0.0608, 0.0039, 0.1111, 0.0596, 0.0039, 0.1122, 0.0623, 0.0017, 0.1148, 0.0584, 0.0017, 0.1129, 0.0623, 0.0039, 0.1148, 0.0584, 0.0039, 0.1129, 0.057, 0.0017, 0.1134, 0.0548, 0.0017, 0.1136, 0.057, 0.0039, 0.1134, 0.0548, 0.0039, 0.1136, 0.0612, -0.0, 0.1156, 0.0593, -0.0, 0.1167, 0.0581, -0.0, 0.117, 0.057, -0.0, 0.1159, 0.0548, -0.0, 0.1159, 0.057, -0.0, 0.1173, 0.0548, -0.0, 0.1176, 0.0613, 0.0017, 0.1155, 0.0593, 0.0017, 0.1167, 0.0613, 0.0039, 0.1155, 0.0593, 0.0039, 0.1167, 0.0581, 0.0017, 0.117, 0.0581, 0.0039, 0.117, 0.057, 0.0017, 0.1173, 0.0548, 0.0017, 0.1176,
0.057, 0.0039, 0.1173, 0.0548, 0.0039, 0.1176, 0.0462, -0.0, 0.107, 0.0476, -0.0, 0.11, 0.047, -0.0, 0.109, 0.0463, 0.0017, 0.1072, 0.0463, 0.0039, 0.1072, 0.0476, 0.0017, 0.1099, 0.047, 0.0017, 0.1088, 0.0476, 0.0039, 0.1099, 0.047, 0.0039, 0.1088, 0.0498, -0.0, 0.1122, 0.0528, -0.0, 0.1134, 0.0508, -0.0, 0.1128, 0.0486, -0.0, 0.1112, 0.0461, -0.0, 0.1116, 0.0483, -0.0, 0.1137, 0.0456, -0.0, 0.1142, 0.0499, 0.0017, 0.1123, 0.0499, 0.0039, 0.1123, 0.0526, 0.0017, 0.1134, 0.051, 0.0017, 0.1129, 0.0526, 0.0039, 0.1134, 0.051, 0.0039, 0.1129, 0.0486, 0.0017, 0.1112, 0.0486, 0.0039, 0.1112, 0.0456, 0.0017, 0.1142, 0.0456, 0.0039, 0.1142, 0.0439, -0.0, 0.1072, 0.0426, -0.0, 0.1072,
0.043, -0.0, 0.1096, 0.0427, -0.0, 0.1086, 0.0426, 0.0017, 0.1072, 0.0426, 0.0039, 0.1072, 0.043, 0.0017, 0.1098, 0.0427, 0.0017, 0.1088, 0.043, 0.0039, 0.1098, 0.0427, 0.0039, 0.1088, 0.0437, -0.0, 0.1119, 0.0447, -0.0, 0.1131, 0.0436, 0.0017, 0.1116, 0.0436, 0.0039, 0.1116, 0.0447, 0.0017, 0.1131, 0.0447, 0.0039, 0.1131, 0.0526, -0.0, 0.1159, 0.0501, -0.0, 0.1169, 0.0526, -0.0, 0.1175, 0.0511, -0.0, 0.1172, 0.048, -0.0, 0.1161, 0.0468, -0.0, 0.1152, 0.0498, 0.0017, 0.1168, 0.0498, 0.0039, 0.1168, 0.0526, 0.0017, 0.1175, 0.0509, 0.0017, 0.1172, 0.0526, 0.0039, 0.1175, 0.0509, 0.0039, 0.1172, 0.0482, 0.0017, 0.1162, 0.0467, 0.0017, 0.1152, 0.0482, 0.0039, 0.1162, 0.0467, 0.0039, 0.1152,
0.067, -0.0, 0.1363, 0.0656, -0.0, 0.1363, 0.0635, -0.0, 0.1363, 0.067, 0.0017, 0.1363, 0.067, 0.0039, 0.1363, 0.0656, 0.0017, 0.1363, 0.0635, 0.0017, 0.1363, 0.0656, 0.0039, 0.1363, 0.0635, 0.0039, 0.1363, 0.067, -0.0, 0.1376, 0.067, -0.0, 0.1402, 0.0656, -0.0, 0.1376, 0.0635, -0.0, 0.1376, 0.0656, -0.0, 0.1402, 0.0635, -0.0, 0.1402, 0.067, 0.0017, 0.1376, 0.067, 0.0039, 0.1376, 0.067, 0.0017, 0.1402, 0.067, 0.0039, 0.1402, 0.0656, 0.0017, 0.1402, 0.0635, 0.0017, 0.1402, 0.0656, 0.0039, 0.1402, 0.0635, 0.0039, 0.1402, 0.0613, -0.0, 0.1363, 0.0591, -0.0, 0.1363, 0.057, -0.0, 0.1363, 0.0548, -0.0, 0.1363, 0.0613, 0.0017, 0.1363, 0.0591, 0.0017, 0.1363, 0.0613, 0.0039, 0.1363,
0.0591, 0.0039, 0.1363, 0.057, 0.0017, 0.1363, 0.0548, 0.0017, 0.1363, 0.057, 0.0039, 0.1363, 0.0548, 0.0039, 0.1363, 0.0613, -0.0, 0.1376, 0.0591, -0.0, 0.1376, 0.0613, -0.0, 0.1402, 0.0591, -0.0, 0.1402, 0.057, -0.0, 0.1376, 0.0548, -0.0, 0.1376, 0.057, -0.0, 0.1402, 0.0548, -0.0, 0.1402, 0.0613, 0.0017, 0.1402, 0.0591, 0.0017, 0.1402, 0.0613, 0.0039, 0.1402, 0.0591, 0.0039, 0.1402, 0.057, 0.0017, 0.1402, 0.0548, 0.0017, 0.1402, 0.057, 0.0039, 0.1402, 0.0548, 0.0039, 0.1402, 0.0526, -0.0, 0.1363, 0.0505, -0.0, 0.1363, 0.0483, -0.0, 0.1363, 0.0461, -0.0, 0.1363, 0.0526, 0.0017, 0.1363, 0.0505, 0.0017, 0.1363, 0.0526, 0.0039, 0.1363, 0.0505, 0.0039, 0.1363, 0.0483, 0.0017, 0.1363,
0.0461, 0.0017, 0.1363, 0.0483, 0.0039, 0.1363, 0.0461, 0.0039, 0.1363, 0.0526, -0.0, 0.1376, 0.0505, -0.0, 0.1376, 0.0526, -0.0, 0.1402, 0.0505, -0.0, 0.1402, 0.0483, -0.0, 0.1376, 0.0461, -0.0, 0.1376, 0.0483, -0.0, 0.1402, 0.0454, -0.0, 0.1402, 0.0526, 0.0017, 0.1402, 0.0505, 0.0017, 0.1402, 0.0526, 0.0039, 0.1402, 0.0505, 0.0039, 0.1402, 0.0483, 0.0017, 0.1402, 0.0454, 0.0017, 0.1402, 0.0483, 0.0039, 0.1402, 0.0454, 0.0039, 0.1402, 0.0439, -0.0, 0.1363, 0.0425, -0.0, 0.1363, 0.0439, 0.0017, 0.1363, 0.0425, 0.0017, 0.1363, 0.0439, 0.0039, 0.1363, 0.0425, 0.0039, 0.1363, 0.0439, -0.0, 0.1376, 0.0425, -0.0, 0.1376, 0.0439, -0.0, 0.1398, 0.0425, -0.0, 0.1398, 0.0425, 0.0017, 0.1376,
0.0425, 0.0039, 0.1376, 0.0425, 0.0017, 0.1398, 0.0425, 0.0039, 0.1398, 0.0591, -0.0, 0.1431, 0.0562, -0.0, 0.1431, 0.0591, 0.0017, 0.1431, 0.0591, 0.0039, 0.1431, 0.0562, 0.0017, 0.1431, 0.0562, 0.0039, 0.1431, 0.0591, -0.0, 0.1463, 0.0592, -0.0, 0.1484, 0.0562, -0.0, 0.1463, 0.0562, -0.0, 0.1484, 0.0591, 0.0017, 0.1463, 0.0591, 0.0039, 0.1463, 0.0592, 0.0017, 0.1485, 0.0592, 0.0039, 0.1485, 0.0562, 0.0017, 0.1463, 0.0562, 0.0039, 0.1463, 0.0562, 0.0017, 0.1484, 0.0562, 0.0039, 0.1484, 0.0632, -0.0, 0.1553, 0.0652, -0.0, 0.1576, 0.0643, -0.0, 0.1564, 0.0632, 0.0017, 0.1553, 0.0632, 0.0039, 0.1553, 0.0651, 0.0017, 0.1575, 0.0643, 0.0017, 0.1564, 0.0651, 0.0039, 0.1575, 0.0643, 0.0039, 0.1564,
0.0596, -0.0, 0.1504, 0.061, -0.0, 0.153, 0.0603, -0.0, 0.1521, 0.0562, -0.0, 0.1506, 0.0562, -0.0, 0.1527, 0.0596, 0.0017, 0.1506, 0.0596, 0.0039, 0.1506, 0.0609, 0.0017, 0.1529, 0.0602, 0.0017, 0.1518, 0.0609, 0.0039, 0.1529, 0.0602, 0.0039, 0.1518, 0.0562, 0.0017, 0.1506, 0.0562, 0.0039, 0.1506, 0.0562, 0.0017, 0.1528, 0.0562, 0.0039, 0.1528, 0.0621, -0.0, 0.1542, 0.0591, -0.0, 0.155, 0.0613, -0.0, 0.1571, 0.0591, -0.0, 0.1571, 0.056, -0.0, 0.1549, 0.0557, -0.0, 0.1557, 0.057, -0.0, 0.1571, 0.0553, -0.0, 0.1571, 0.0621, 0.0017, 0.1542, 0.0621, 0.0039, 0.1542, 0.0561, 0.0017, 0.1546, 0.0558, 0.0017, 0.1556, 0.0561, 0.0039, 0.1546, 0.0558, 0.0039, 0.1556, 0.0553, 0.0017, 0.1572,
0.0553, 0.0039, 0.1572, 0.0454, -0.0, 0.1419, 0.0454, -0.0, 0.1441, 0.0454, 0.0017, 0.1419, 0.0454, 0.0039, 0.1419, 0.0454, 0.0017, 0.1441, 0.0454, 0.0039, 0.1441, 0.0454, -0.0, 0.1463, 0.0454, -0.0, 0.1484, 0.0454, 0.0017, 0.1463, 0.0454, 0.0039, 0.1463, 0.0454, 0.0017, 0.1484, 0.0454, 0.0039, 0.1484, 0.0439, -0.0, 0.1419, 0.0425, -0.0, 0.1419, 0.0439, -0.0, 0.1441, 0.0425, -0.0, 0.1441, 0.0425, 0.0017, 0.1419, 0.0425, 0.0039, 0.1419, 0.0425, 0.0017, 0.1441, 0.0425, 0.0039, 0.1441, 0.0439, -0.0, 0.1463, 0.0425, -0.0, 0.1463, 0.0439, -0.0, 0.1484, 0.0425, -0.0, 0.1484, 0.0425, 0.0017, 0.1463, 0.0425, 0.0039, 0.1463, 0.0425, 0.0017, 0.1484, 0.0425, 0.0039, 0.1484, 0.0454, -0.0, 0.1506,
0.0456, -0.0, 0.1528, 0.0454, 0.0017, 0.1506, 0.0454, 0.0039, 0.1506, 0.0456, 0.0017, 0.1528, 0.0456, 0.0039, 0.1528, 0.0459, -0.0, 0.1549, 0.0475, -0.0, 0.1579, 0.0468, -0.0, 0.1569, 0.0459, 0.0017, 0.155, 0.0459, 0.0039, 0.155, 0.0474, 0.0017, 0.1579, 0.0467, 0.0017, 0.1568, 0.0474, 0.0039, 0.1579, 0.0467, 0.0039, 0.1568, 0.0439, -0.0, 0.1506, 0.0425, -0.0, 0.1506, 0.0439, -0.0, 0.1528, 0.0425, -0.0, 0.1528, 0.0425, 0.0017, 0.1506, 0.0425, 0.0039, 0.1506, 0.0425, 0.0017, 0.1528, 0.0425, 0.0039, 0.1528, 0.0439, -0.0, 0.155, 0.0425, -0.0, 0.155, 0.0439, -0.0, 0.1571, 0.0426, -0.0, 0.1571, 0.0425, 0.0017, 0.155, 0.0425, 0.0039, 0.155, 0.0426, 0.0017, 0.1571, 0.0426, 0.0039, 0.1571,
0.0669, -0.0, 0.1621, 0.0661, -0.0, 0.159, 0.0635, -0.0, 0.1593, 0.0667, -0.0, 0.1612, 0.0632, -0.0, 0.1619, 0.0669, 0.0017, 0.1619, 0.0669, 0.0039, 0.1619, 0.0662, 0.0017, 0.1592, 0.0662, 0.0039, 0.1592, 0.0666, 0.0017, 0.1608, 0.0632, 0.0017, 0.1619, 0.0666, 0.0039, 0.1608, 0.0632, 0.0039, 0.1619, 0.0671, -0.0, 0.1648, 0.0653, -0.0, 0.1636, 0.0643, -0.0, 0.1628, 0.0671, 0.0017, 0.1648, 0.0671, 0.0039, 0.1648, 0.0656, 0.0017, 0.1638, 0.0643, 0.0017, 0.1628, 0.0656, 0.0039, 0.1638, 0.0643, 0.0039, 0.1628, 0.0609, -0.0, 0.16, 0.0591, -0.0, 0.1588, 0.0619, -0.0, 0.1608, 0.0583, -0.0, 0.1606, 0.057, -0.0, 0.1593, 0.0543, -0.0, 0.1588, 0.0574, -0.0, 0.1618, 0.0548, -0.0, 0.1615,
0.0608, 0.0017, 0.1599, 0.0591, 0.0017, 0.1588, 0.0608, 0.0039, 0.1599, 0.0591, 0.0039, 0.1588, 0.0619, 0.0017, 0.1608, 0.0583, 0.0017, 0.1607, 0.0619, 0.0039, 0.1608, 0.0583, 0.0039, 0.1607, 0.0543, 0.0017, 0.1587, 0.0543, 0.0039, 0.1587, 0.0574, 0.0017, 0.1618, 0.0574, 0.0039, 0.1618, 0.0563, -0.0, 0.1629, 0.0551, -0.0, 0.1637, 0.0563, 0.0017, 0.1629, 0.0549, 0.0017, 0.1638, 0.0563, 0.0039, 0.1629, 0.0549, 0.0039, 0.1638, 0.0525, -0.0, 0.1598, 0.0503, -0.0, 0.1603, 0.0526, -0.0, 0.1615, 0.0505, -0.0, 0.1615, 0.0484, -0.0, 0.1591, 0.0461, -0.0, 0.1593, 0.0483, -0.0, 0.1615, 0.0461, -0.0, 0.1615, 0.053, 0.0017, 0.1596, 0.0503, 0.0017, 0.1603, 0.053, 0.0039, 0.1596, 0.0503, 0.0039, 0.1603,
0.0479, 0.0017, 0.1585, 0.0479, 0.0039, 0.1585, 0.0529, -0.0, 0.1645, 0.0505, -0.0, 0.1647, 0.048, -0.0, 0.1644, 0.0459, -0.0, 0.1633, 0.0527, 0.0017, 0.1646, 0.0505, 0.0017, 0.1647, 0.0527, 0.0039, 0.1646, 0.0505, 0.0039, 0.1647, 0.048, 0.0017, 0.1645, 0.046, 0.0017, 0.1634, 0.048, 0.0039, 0.1645, 0.046, 0.0039, 0.1634, 0.043, -0.0, 0.1595, 0.0428, -0.0, 0.1584, 0.0442, -0.0, 0.1618, 0.0431, 0.0017, 0.1596, 0.0428, 0.0017, 0.1585, 0.0431, 0.0039, 0.1596, 0.0428, 0.0039, 0.1585, 0.0444, 0.0017, 0.162, 0.0444, 0.0039, 0.162, 0.045, -0.0, 0.1626, 0.045, 0.0017, 0.1626, 0.045, 0.0039, 0.1626, 0.0219, -0.0, 0.1412, 0.0201, -0.0, 0.1412, 0.0219, -0.0, 0.1441, 0.0201, -0.0, 0.1441,
0.0219, 0.0017, 0.1412, 0.0201, 0.0017, 0.1412, 0.0219, 0.0039, 0.1412, 0.0201, 0.0039, 0.1412, 0.0219, 0.0017, 0.1441, 0.0219, 0.0039, 0.1441, 0.0219, -0.0, 0.1463, 0.0201, -0.0, 0.1463, 0.0219, -0.0, 0.1484, 0.0201, -0.0, 0.1484, 0.0219, 0.0017, 0.1463, 0.0219, 0.0039, 0.1463, 0.0219, 0.0017, 0.1484, 0.0219, 0.0039, 0.1484, 0.0219, -0.0, 0.1506, 0.0201, -0.0, 0.1506, 0.0219, -0.0, 0.1528, 0.0201, -0.0, 0.1528, 0.0219, 0.0017, 0.1506, 0.0219, 0.0039, 0.1506, 0.0219, 0.0017, 0.1528, 0.0219, 0.0039, 0.1528, 0.0219, -0.0, 0.155, 0.0201, -0.0, 0.155, 0.0219, -0.0, 0.1571, 0.0201, -0.0, 0.1571, 0.0219, 0.0017, 0.155, 0.0219, 0.0039, 0.155, 0.0219, 0.0017, 0.1571, 0.0219, 0.0039, 0.1571,
0.0179, -0.0, 0.1412, 0.0157, -0.0, 0.1412, 0.0179, -0.0, 0.1441, 0.0157, -0.0, 0.1441, 0.0136, -0.0, 0.1412, 0.0114, -0.0, 0.1412, 0.0136, -0.0, 0.1441, 0.0114, -0.0, 0.1441, 0.0179, 0.0017, 0.1412, 0.0157, 0.0017, 0.1412, 0.0179, 0.0039, 0.1412, 0.0157, 0.0039, 0.1412, 0.0136, 0.0017, 0.1412, 0.0114, 0.0017, 0.1412, 0.0136, 0.0039, 0.1412, 0.0114, 0.0039, 0.1412, 0.0179, -0.0, 0.1463, 0.0157, -0.0, 0.1463, 0.0179, -0.0, 0.1484, 0.0157, -0.0, 0.1484, 0.0136, -0.0, 0.1463, 0.0114, -0.0, 0.1463, 0.0136, -0.0, 0.1484, 0.0114, -0.0, 0.1484, 0.0092, -0.0, 0.1412, 0.0071, -0.0, 0.1412, 0.0092, -0.0, 0.1441, 0.0071, -0.0, 0.1441, 0.0049, -0.0, 0.1412, 0.0027, -0.0, 0.1412,
0.0049, -0.0, 0.1441, 0.0027, -0.0, 0.1441, 0.0092, 0.0017, 0.1412, 0.0071, 0.0017, 0.1412, 0.0092, 0.0039, 0.1412, 0.0071, 0.0039, 0.1412, 0.0049, 0.0017, 0.1412, 0.0027, 0.0017, 0.1412, 0.0049, 0.0039, 0.1412, 0.0027, 0.0039, 0.1412, 0.0092, -0.0, 0.1463, 0.0071, -0.0, 0.1463, 0.0092, -0.0, 0.1484, 0.0071, -0.0, 0.1484, 0.0049, -0.0, 0.1463, 0.0027, -0.0, 0.1463, 0.0049, -0.0, 0.1484, 0.0027, -0.0, 0.1484, 0.0179, -0.0, 0.1506, 0.0157, -0.0, 0.1506, 0.0179, -0.0, 0.1528, 0.0157, -0.0, 0.1528, 0.0136, -0.0, 0.1506, 0.0114, -0.0, 0.1506, 0.0136, -0.0, 0.1528, 0.0114, -0.0, 0.1528, 0.0179, -0.0, 0.155, 0.0157, -0.0, 0.155, 0.0179, -0.0, 0.1571, 0.0157, -0.0, 0.1571,
0.0136, -0.0, 0.155, 0.0114, -0.0, 0.155, 0.0136, -0.0, 0.1571, 0.0114, -0.0, 0.1571, 0.0092, -0.0, 0.1506, 0.0071, -0.0, 0.1506, 0.0101, -0.0, 0.152, 0.0071, -0.0, 0.152, 0.0049, -0.0, 0.1506, 0.0027, -0.0, 0.1506, 0.0049, -0.0, 0.152, 0.0027, -0.0, 0.152, 0.0101, 0.0017, 0.152, 0.0071, 0.0017, 0.152, 0.0101, 0.0039, 0.152, 0.0071, 0.0039, 0.152, 0.0049, 0.0017, 0.152, 0.0027, 0.0017, 0.152, 0.0049, 0.0039, 0.152, 0.0027, 0.0039, 0.152, 0.0101, -0.0, 0.155, 0.0101, -0.0, 0.1571, 0.0101, 0.0017, 0.155, 0.0101, 0.0039, 0.155, 0.0101, 0.0017, 0.1571, 0.0101, 0.0039, 0.1571, 0.0219, -0.0, 0.1593, 0.0201, -0.0, 0.1593, 0.0219, -0.0, 0.1615, 0.0201, -0.0, 0.1615,
0.0219, 0.0017, 0.1593, 0.0219, 0.0039, 0.1593, 0.0219, 0.0017, 0.1615, 0.0219, 0.0039, 0.1615, 0.0219, -0.0, 0.1636, 0.0201, -0.0, 0.1636, 0.0219, -0.0, 0.1658, 0.0201, -0.0, 0.1658, 0.0219, 0.0017, 0.1636, 0.0219, 0.0039, 0.1636, 0.0219, 0.0017, 0.1658, 0.0219, 0.0039, 0.1658, 0.0219, -0.0, 0.168, 0.0201, -0.0, 0.168, 0.0219, -0.0, 0.1701, 0.0201, -0.0, 0.1701, 0.0219, 0.0017, 0.168, 0.0219, 0.0039, 0.168, 0.0219, 0.0017, 0.1701, 0.0219, 0.0039, 0.1701, 0.0219, -0.0, 0.1723, 0.0201, -0.0, 0.1723, 0.0219, -0.0, 0.1745, 0.0201, -0.0, 0.1745, 0.0219, 0.0017, 0.1723, 0.0219, 0.0039, 0.1723, 0.0219, 0.0017, 0.1745, 0.0219, 0.0039, 0.1745, 0.0179, -0.0, 0.1593, 0.0157, -0.0, 0.1593,
0.0179, -0.0, 0.1615, 0.0157, -0.0, 0.1615, 0.0136, -0.0, 0.1593, 0.0114, -0.0, 0.1593, 0.0136, -0.0, 0.1615, 0.0114, -0.0, 0.1615, 0.0179, -0.0, 0.1636, 0.0157, -0.0, 0.1636, 0.0179, -0.0, 0.1658, 0.0157, -0.0, 0.1658, 0.0136, -0.0, 0.1636, 0.0114, -0.0, 0.1636, 0.0136, -0.0, 0.1658, 0.0114, -0.0, 0.1658, 0.0101, -0.0, 0.1593, 0.0101, -0.0, 0.1615, 0.0101, 0.0017, 0.1593, 0.0101, 0.0039, 0.1593, 0.0101, 0.0017, 0.1615, 0.0101, 0.0039, 0.1615, 0.0101, -0.0, 0.1636, 0.0101, -0.0, 0.1658, 0.0101, 0.0017, 0.1636, 0.0101, 0.0039, 0.1636, 0.0101, 0.0017, 0.1658, 0.0101, 0.0039, 0.1658, 0.0179, -0.0, 0.168, 0.0157, -0.0, 0.168, 0.0179, -0.0, 0.1701, 0.0157, -0.0, 0.1701,
0.0136, -0.0, 0.168, 0.0114, -0.0, 0.168, 0.0136, -0.0, 0.1701, 0.0114, -0.0, 0.1701, 0.0179, -0.0, 0.1723, 0.0157, -0.0, 0.1723, 0.0179, -0.0, 0.1745, 0.0157, -0.0, 0.1745, 0.0136, -0.0, 0.1723, 0.0114, -0.0, 0.1723, 0.0136, -0.0, 0.1745, 0.0114, -0.0, 0.1745, 0.0101, -0.0, 0.168, 0.0101, -0.0, 0.1701, 0.0101, 0.0017, 0.168, 0.0101, 0.0039, 0.168, 0.0101, 0.0017, 0.1701, 0.0101, 0.0039, 0.1701, 0.0101, -0.0, 0.1723, 0.0101, -0.0, 0.1745, 0.0101, 0.0017, 0.1723, 0.0101, 0.0039, 0.1723, 0.0101, 0.0017, 0.1745, 0.0101, 0.0039, 0.1745, 0.0219, -0.0, 0.1767, 0.0201, -0.0, 0.1767, 0.0219, -0.0, 0.1788, 0.0201, -0.0, 0.1788, 0.0219, 0.0017, 0.1767, 0.0219, 0.0039, 0.1767,
0.0219, 0.0017, 0.1788, 0.0219, 0.0039, 0.1788, 0.0219, -0.0, 0.181, 0.0201, -0.0, 0.181, 0.0219, -0.0, 0.1832, 0.0201, -0.0, 0.1832, 0.0219, 0.0017, 0.181, 0.0219, 0.0039, 0.181, 0.0219, 0.0017, 0.1832, 0.0219, 0.0039, 0.1832, 0.0219, -0.0, 0.1853, 0.0201, -0.0, 0.1853, 0.0219, -0.0, 0.1875, 0.0201, -0.0, 0.1875, 0.0219, 0.0017, 0.1853, 0.0219, 0.0039, 0.1853, 0.0219, 0.0017, 0.1875, 0.0219, 0.0039, 0.1875, 0.0219, -0.0, 0.1897, 0.0201, -0.0, 0.1897, 0.0219, -0.0, 0.1918, 0.0201, -0.0, 0.1918, 0.0219, 0.0017, 0.1897, 0.0219, 0.0039, 0.1897, 0.0219, 0.0017, 0.1918, 0.0219, 0.0039, 0.1918, 0.0179, -0.0, 0.1767, 0.0157, -0.0, 0.1767, 0.0179, -0.0, 0.1788, 0.0157, -0.0, 0.1788,
0.0136, -0.0, 0.1767, 0.0114, -0.0, 0.1767, 0.0136, -0.0, 0.1788, 0.0114, -0.0, 0.1788, 0.0179, -0.0, 0.181, 0.0157, -0.0, 0.181, 0.0179, -0.0, 0.1832, 0.0157, -0.0, 0.1832, 0.0136, -0.0, 0.181, 0.0114, -0.0, 0.181, 0.0136, -0.0, 0.1832, 0.0114, -0.0, 0.1832, 0.0101, -0.0, 0.1767, 0.0101, -0.0, 0.1788, 0.0101, 0.0017, 0.1767, 0.0101, 0.0039, 0.1767, 0.0101, 0.0017, 0.1788, 0.0101, 0.0039, 0.1788, 0.0101, -0.0, 0.181, 0.0101, -0.0, 0.1832, 0.0101, 0.0017, 0.181, 0.0101, 0.0039, 0.181, 0.0101, 0.0017, 0.1832, 0.0101, 0.0039, 0.1832, 0.0179, -0.0, 0.1853, 0.0157, -0.0, 0.1853, 0.0179, -0.0, 0.1875, 0.0157, -0.0, 0.1875, 0.0136, -0.0, 0.1853, 0.0114, -0.0, 0.1853,
0.0136, -0.0, 0.1875, 0.0114, -0.0, 0.1875, 0.0179, -0.0, 0.1897, 0.0157, -0.0, 0.1897, 0.0179, -0.0, 0.1918, 0.0157, -0.0, 0.1918, 0.0136, -0.0, 0.1897, 0.0114, -0.0, 0.1897, 0.0136, -0.0, 0.1918, 0.0114, -0.0, 0.1918, 0.0101, -0.0, 0.1853, 0.0101, -0.0, 0.1875, 0.0101, 0.0017, 0.1853, 0.0101, 0.0039, 0.1853, 0.0101, 0.0017, 0.1875, 0.0101, 0.0039, 0.1875, 0.0101, -0.0, 0.1897, 0.0101, -0.0, 0.1918, 0.0101, 0.0017, 0.1897, 0.0101, 0.0039, 0.1897, 0.0101, 0.0017, 0.1918, 0.0101, 0.0039, 0.1918, 0.0219, -0.0, 0.194, 0.0201, -0.0, 0.194, 0.0219, -0.0, 0.1962, 0.0201, -0.0, 0.1962, 0.0219, 0.0017, 0.194, 0.0219, 0.0039, 0.194, 0.0219, 0.0017, 0.1962, 0.0219, 0.0039, 0.1962,
0.0219, -0.0, 0.1984, 0.0201, -0.0, 0.1984, 0.0219, -0.0, 0.2005, 0.0201, -0.0, 0.2005, 0.0219, 0.0017, 0.1984, 0.0219, 0.0039, 0.1984, 0.0219, 0.0017, 0.2005, 0.0219, 0.0039, 0.2005, 0.0219, -0.0, 0.202, 0.0201, -0.0, 0.202, 0.0219, 0.0017, 0.202, 0.0201, 0.0017, 0.202, 0.0219, 0.0039, 0.202, 0.0201, 0.0039, 0.202, 0.0179, -0.0, 0.194, 0.0157, -0.0, 0.194, 0.0179, -0.0, 0.1962, 0.0157, -0.0, 0.1962, 0.0136, -0.0, 0.194, 0.0114, -0.0, 0.194, 0.0136, -0.0, 0.1962, 0.0114, -0.0, 0.1962, 0.0179, -0.0, 0.1984, 0.0157, -0.0, 0.1984, 0.0179, -0.0, 0.2005, 0.0157, -0.0, 0.2005, 0.0136, -0.0, 0.1984, 0.0114, -0.0, 0.1984, 0.0136, -0.0, 0.2005, 0.0114, -0.0, 0.2005,
0.0101, -0.0, 0.194, 0.0101, -0.0, 0.1962, 0.0101, 0.0017, 0.194, 0.0101, 0.0039, 0.194, 0.0101, 0.0017, 0.1962, 0.0101, 0.0039, 0.1962, 0.0101, -0.0, 0.1984, 0.0101, -0.0, 0.2005, 0.0101, 0.0017, 0.1984, 0.0101, 0.0039, 0.1984, 0.0101, 0.0017, 0.2005, 0.0101, 0.0039, 0.2005, 0.0179, -0.0, 0.202, 0.0157, -0.0, 0.202, 0.0136, -0.0, 0.202, 0.0114, -0.0, 0.202, 0.0179, 0.0017, 0.202, 0.0157, 0.0017, 0.202, 0.0179, 0.0039, 0.202, 0.0157, 0.0039, 0.202, 0.0136, 0.0017, 0.202, 0.0114, 0.0017, 0.202, 0.0136, 0.0039, 0.202, 0.0114, 0.0039, 0.202, 0.0101, -0.0, 0.202, 0.0101, 0.0017, 0.202, 0.0101, 0.0039, 0.202, 0.067, -0.0, 0.2118, 0.067, -0.0, 0.2135, 0.0656, -0.0, 0.2118,
0.0635, -0.0, 0.2118, 0.0656, -0.0, 0.2135, 0.0635, -0.0, 0.2135, 0.067, 0.0017, 0.2118, 0.067, 0.0039, 0.2118, 0.067, 0.0017, 0.2135, 0.067, 0.0039, 0.2135, 0.0656, 0.0017, 0.2118, 0.0635, 0.0017, 0.2118, 0.0656, 0.0039, 0.2118, 0.0635, 0.0039, 0.2118, 0.067, -0.0, 0.2157, 0.067, -0.0, 0.2179, 0.0656, -0.0, 0.2157, 0.0631, -0.0, 0.2157, 0.0656, -0.0, 0.2179, 0.0631, -0.0, 0.2179, 0.067, 0.0017, 0.2157, 0.067, 0.0039, 0.2157, 0.067, 0.0017, 0.2179, 0.067, 0.0039, 0.2179, 0.0631, 0.0017, 0.2157, 0.0631, 0.0039, 0.2157, 0.0631, 0.0017, 0.2179, 0.0631, 0.0039, 0.2179, 0.0613, -0.0, 0.2118, 0.0591, -0.0, 0.2118, 0.0613, -0.0, 0.2135, 0.0591, -0.0, 0.2135, 0.057, -0.0, 0.2118,
0.0548, -0.0, 0.2118, 0.057, -0.0, 0.2135, 0.0548, -0.0, 0.2135, 0.0613, 0.0017, 0.2118, 0.0591, 0.0017, 0.2118, 0.0613, 0.0039, 0.2118, 0.0591, 0.0039, 0.2118, 0.057, 0.0017, 0.2118, 0.0548, 0.0017, 0.2118, 0.057, 0.0039, 0.2118, 0.0548, 0.0039, 0.2118, 0.0613, -0.0, 0.2157, 0.0591, -0.0, 0.2157, 0.057, -0.0, 0.2157, 0.0548, -0.0, 0.2157, 0.0613, 0.0017, 0.2157, 0.0591, 0.0017, 0.2157, 0.0613, 0.0039, 0.2157, 0.0591, 0.0039, 0.2157, 0.057, 0.0017, 0.2157, 0.0548, 0.0017, 0.2157, 0.057, 0.0039, 0.2157, 0.0548, 0.0039, 0.2157, 0.067, -0.0, 0.2201, 0.067, -0.0, 0.2222, 0.0656, -0.0, 0.2201, 0.0631, -0.0, 0.2201, 0.0656, -0.0, 0.2222, 0.0631, -0.0, 0.2222, 0.067, 0.0017, 0.2201,
0.067, 0.0039, 0.2201, 0.067, 0.0017, 0.2222, 0.067, 0.0039, 0.2222, 0.0631, 0.0017, 0.2201, 0.0631, 0.0039, 0.2201, 0.0631, 0.0017, 0.2222, 0.0631, 0.0039, 0.2222, 0.067, -0.0, 0.2244, 0.067, -0.0, 0.2266, 0.0656, -0.0, 0.2244, 0.0631, -0.0, 0.2244, 0.0656, -0.0, 0.2266, 0.0631, -0.0, 0.2266, 0.067, 0.0017, 0.2244, 0.067, 0.0039, 0.2244, 0.067, 0.0017, 0.2266, 0.067, 0.0039, 0.2266, 0.0631, 0.0017, 0.2244, 0.0631, 0.0039, 0.2244, 0.0631, 0.0017, 0.2266, 0.0631, 0.0039, 0.2266, 0.0562, -0.0, 0.2196, 0.0548, -0.0, 0.2196, 0.0562, -0.0, 0.2222, 0.0548, -0.0, 0.2222, 0.0562, 0.0017, 0.2196, 0.0548, 0.0017, 0.2196, 0.0562, 0.0039, 0.2196, 0.0548, 0.0039, 0.2196, 0.0562, 0.0017, 0.2222,
0.0562, 0.0039, 0.2222, 0.0562, -0.0, 0.2244, 0.0548, -0.0, 0.2244, 0.0562, -0.0, 0.2266, 0.0548, -0.0, 0.2266, 0.0562, 0.0017, 0.2244, 0.0562, 0.0039, 0.2244, 0.0562, 0.0017, 0.2266, 0.0562, 0.0039, 0.2266, 0.0526, -0.0, 0.2118, 0.0505, -0.0, 0.2118, 0.0526, -0.0, 0.2135, 0.0505, -0.0, 0.2135, 0.0483, -0.0, 0.2118, 0.0461, -0.0, 0.2118, 0.0483, -0.0, 0.2135, 0.0461, -0.0, 0.2135, 0.0526, 0.0017, 0.2118, 0.0505, 0.0017, 0.2118, 0.0526, 0.0039, 0.2118, 0.0505, 0.0039, 0.2118, 0.0483, 0.0017, 0.2118, 0.0461, 0.0017, 0.2118, 0.0483, 0.0039, 0.2118, 0.0461, 0.0039, 0.2118, 0.0526, -0.0, 0.2157, 0.0505, -0.0, 0.2157, 0.0483, -0.0, 0.2157, 0.0454, -0.0, 0.2157, 0.0454, -0.0, 0.2179,
0.0526, 0.0017, 0.2157, 0.0505, 0.0017, 0.2157, 0.0526, 0.0039, 0.2157, 0.0505, 0.0039, 0.2157, 0.0483, 0.0017, 0.2157, 0.0454, 0.0017, 0.2157, 0.0483, 0.0039, 0.2157, 0.0454, 0.0039, 0.2157, 0.0454, 0.0017, 0.2179, 0.0454, 0.0039, 0.2179, 0.0439, -0.0, 0.2118, 0.0425, -0.0, 0.2118, 0.0439, -0.0, 0.2135, 0.0425, -0.0, 0.2135, 0.0439, 0.0017, 0.2118, 0.0425, 0.0017, 0.2118, 0.0439, 0.0039, 0.2118, 0.0425, 0.0039, 0.2118, 0.0425, 0.0017, 0.2135, 0.0425, 0.0039, 0.2135, 0.0439, -0.0, 0.2157, 0.0425, -0.0, 0.2157, 0.0439, -0.0, 0.2179, 0.0425, -0.0, 0.2179, 0.0425, 0.0017, 0.2157, 0.0425, 0.0039, 0.2157, 0.0425, 0.0017, 0.2179, 0.0425, 0.0039, 0.2179, 0.0523, -0.0, 0.2196, 0.0523, -0.0, 0.2222,
0.0454, -0.0, 0.2201, 0.0454, -0.0, 0.2222, 0.0523, 0.0017, 0.2196, 0.0523, 0.0039, 0.2196, 0.0523, 0.0017, 0.2222, 0.0523, 0.0039, 0.2222, 0.0454, 0.0017, 0.2201, 0.0454, 0.0039, 0.2201, 0.0454, 0.0017, 0.2222, 0.0454, 0.0039, 0.2222, 0.0523, -0.0, 0.2244, 0.0523, -0.0, 0.2266, 0.0454, -0.0, 0.2244, 0.0454, -0.0, 0.2266, 0.0523, 0.0017, 0.2244, 0.0523, 0.0039, 0.2244, 0.0523, 0.0017, 0.2266, 0.0523, 0.0039, 0.2266, 0.0454, 0.0017, 0.2244, 0.0454, 0.0039, 0.2244, 0.0454, 0.0017, 0.2266, 0.0454, 0.0039, 0.2266, 0.0439, -0.0, 0.2201, 0.0425, -0.0, 0.2201, 0.0439, -0.0, 0.2222, 0.0425, -0.0, 0.2222, 0.0425, 0.0017, 0.2201, 0.0425, 0.0039, 0.2201, 0.0425, 0.0017, 0.2222, 0.0425, 0.0039, 0.2222,
0.0439, -0.0, 0.2244, 0.0425, -0.0, 0.2244, 0.0439, -0.0, 0.2266, 0.0425, -0.0, 0.2266, 0.0425, 0.0017, 0.2244, 0.0425, 0.0039, 0.2244, 0.0425, 0.0017, 0.2266, 0.0425, 0.0039, 0.2266, 0.067, -0.0, 0.2287, 0.0669, -0.0, 0.2304, 0.0656, -0.0, 0.2287, 0.0631, -0.0, 0.2287, 0.0668, -0.0, 0.2312, 0.0629, -0.0, 0.2308, 0.067, 0.0017, 0.2287, 0.067, 0.0039, 0.2287, 0.0669, 0.0017, 0.2307, 0.0669, 0.0039, 0.2307, 0.063, 0.0017, 0.2287, 0.063, 0.0039, 0.2287, 0.0667, 0.0017, 0.2318, 0.0629, 0.0017, 0.2309, 0.0667, 0.0039, 0.2318, 0.0629, 0.0039, 0.2309, 0.0665, -0.0, 0.2333, 0.0624, -0.0, 0.2328, 0.0657, -0.0, 0.2356, 0.0635, -0.0, 0.2353, 0.0665, 0.0017, 0.2331, 0.0625, 0.0017, 0.2325,
0.0665, 0.0039, 0.2331, 0.0625, 0.0039, 0.2325, 0.0657, 0.0017, 0.2355, 0.0657, 0.0039, 0.2355, 0.0563, -0.0, 0.2287, 0.0548, -0.0, 0.2287, 0.0565, -0.0, 0.2309, 0.0548, -0.0, 0.2309, 0.0563, 0.0017, 0.2287, 0.0563, 0.0039, 0.2287, 0.0565, 0.0017, 0.2309, 0.0565, 0.0039, 0.2309, 0.0621, -0.0, 0.2337, 0.0613, -0.0, 0.2349, 0.0591, -0.0, 0.2363, 0.0571, -0.0, 0.2329, 0.0548, -0.0, 0.2331, 0.058, -0.0, 0.2346, 0.0548, -0.0, 0.2353, 0.0621, 0.0017, 0.2336, 0.0621, 0.0039, 0.2336, 0.0613, 0.0017, 0.2349, 0.0591, 0.0017, 0.2363, 0.0613, 0.0039, 0.2349, 0.0591, 0.0039, 0.2363, 0.0571, 0.0017, 0.2331, 0.0571, 0.0039, 0.2331, 0.0579, 0.0017, 0.2345, 0.0579, 0.0039, 0.2345, 0.0648, -0.0, 0.2365,
0.0634, -0.0, 0.238, 0.0648, 0.0017, 0.2366, 0.0634, 0.0017, 0.238, 0.0648, 0.0039, 0.2366, 0.0634, 0.0039, 0.238, 0.0618, -0.0, 0.2386, 0.0591, -0.0, 0.2374, 0.0609, -0.0, 0.2389, 0.0591, -0.0, 0.239, 0.057, -0.0, 0.2374, 0.0547, -0.0, 0.2384, 0.057, -0.0, 0.2387, 0.0546, -0.0, 0.2384, 0.0622, 0.0017, 0.2384, 0.0622, 0.0039, 0.2384, 0.0612, 0.0017, 0.2388, 0.0591, 0.0017, 0.239, 0.0612, 0.0039, 0.2388, 0.0591, 0.0039, 0.239, 0.057, 0.0017, 0.2387, 0.0544, 0.0017, 0.2384, 0.057, 0.0039, 0.2387, 0.0544, 0.0039, 0.2384, 0.0522, -0.0, 0.2287, 0.0522, -0.0, 0.2309, 0.0455, -0.0, 0.2287, 0.0458, -0.0, 0.2308, 0.0522, 0.0017, 0.2287, 0.0522, 0.0039, 0.2287, 0.0522, 0.0017, 0.2309,
0.0522, 0.0039, 0.2309, 0.0455, 0.0017, 0.2287, 0.0455, 0.0039, 0.2287, 0.0459, 0.0017, 0.2309, 0.0459, 0.0039, 0.2309, 0.0518, -0.0, 0.2329, 0.0512, -0.0, 0.2338, 0.0526, -0.0, 0.2353, 0.0504, -0.0, 0.2345, 0.0475, -0.0, 0.2339, 0.0466, -0.0, 0.2329, 0.0484, -0.0, 0.2348, 0.0461, -0.0, 0.2353, 0.0519, 0.0017, 0.2326, 0.0513, 0.0017, 0.2337, 0.0519, 0.0039, 0.2326, 0.0513, 0.0039, 0.2337, 0.0504, 0.0017, 0.2345, 0.0504, 0.0039, 0.2345, 0.0475, 0.0017, 0.2339, 0.0465, 0.0017, 0.2328, 0.0475, 0.0039, 0.2339, 0.0465, 0.0039, 0.2328, 0.0484, 0.0017, 0.2348, 0.0484, 0.0039, 0.2348, 0.0439, -0.0, 0.2287, 0.0425, -0.0, 0.2287, 0.0439, -0.0, 0.2309, 0.0425, -0.0, 0.2309, 0.0425, 0.0017, 0.2287,
0.0425, 0.0039, 0.2287, 0.0425, 0.0017, 0.2309, 0.0425, 0.0039, 0.2309, 0.0429, -0.0, 0.2334, 0.0427, -0.0, 0.2324, 0.0438, -0.0, 0.2356, 0.043, 0.0017, 0.2337, 0.0427, 0.0017, 0.2326, 0.043, 0.0039, 0.2337, 0.0427, 0.0039, 0.2326, 0.0437, 0.0017, 0.2353, 0.0437, 0.0039, 0.2353, 0.0526, -0.0, 0.2374, 0.0505, -0.0, 0.2374, 0.0526, -0.0, 0.2386, 0.0505, -0.0, 0.239, 0.0475, -0.0, 0.2387, 0.0457, -0.0, 0.2377, 0.0485, -0.0, 0.239, 0.0526, 0.0017, 0.2386, 0.0505, 0.0017, 0.239, 0.0526, 0.0039, 0.2386, 0.0505, 0.0039, 0.239, 0.0472, 0.0017, 0.2385, 0.046, 0.0017, 0.2379, 0.0472, 0.0039, 0.2385, 0.046, 0.0039, 0.2379, 0.0478, 0.0017, 0.2388, 0.0478, 0.0039, 0.2388, 0.0447, -0.0, 0.2367,
0.0447, 0.0017, 0.2368, 0.0447, 0.0039, 0.2368, 0.0106, -0.0, 0.2187, 0.0107, 0.0017, 0.2187, 0.0107, 0.0039, 0.2187, 0.0022, -0.0, 0.2143, 0.0024, 0.0017, 0.2144, 0.0024, 0.0039, 0.2144, 0.0067, -0.0, 0.2162, 0.0096, -0.0, 0.2178, 0.0077, -0.0, 0.2168, 0.0052, -0.0, 0.2154, 0.0032, -0.0, 0.2146, 0.0049, -0.0, 0.2179, 0.0027, -0.0, 0.2179, 0.0069, 0.0017, 0.2163, 0.0069, 0.0039, 0.2163, 0.0093, 0.0017, 0.2176, 0.008, 0.0017, 0.2169, 0.0093, 0.0039, 0.2176, 0.008, 0.0039, 0.2169, 0.0049, 0.0017, 0.2153, 0.0035, 0.0017, 0.2147, 0.0049, 0.0039, 0.2153, 0.0035, 0.0039, 0.2147, 0.0149, -0.0, 0.2231, 0.0129, -0.0, 0.2207, 0.0118, -0.0, 0.2196, 0.014, -0.0, 0.2219, 0.0114, -0.0, 0.2222,
0.0149, 0.0017, 0.223, 0.0149, 0.0039, 0.223, 0.0129, 0.0017, 0.2207, 0.0118, 0.0017, 0.2196, 0.0129, 0.0039, 0.2207, 0.0118, 0.0039, 0.2196, 0.014, 0.0017, 0.2219, 0.014, 0.0039, 0.2219, 0.0156, -0.0, 0.2241, 0.0173, -0.0, 0.227, 0.0168, -0.0, 0.226, 0.0136, -0.0, 0.2244, 0.0114, -0.0, 0.2244, 0.0136, -0.0, 0.2266, 0.0114, -0.0, 0.2266, 0.0158, 0.0017, 0.2244, 0.0158, 0.0039, 0.2244, 0.0172, 0.0017, 0.2269, 0.0167, 0.0017, 0.2258, 0.0172, 0.0039, 0.2269, 0.0167, 0.0039, 0.2258, 0.0092, -0.0, 0.2201, 0.0071, -0.0, 0.2201, 0.0092, -0.0, 0.2222, 0.0071, -0.0, 0.2222, 0.0049, -0.0, 0.2201, 0.0027, -0.0, 0.2201, 0.0049, -0.0, 0.2222, 0.0027, -0.0, 0.2222, 0.0092, -0.0, 0.2244,
0.0071, -0.0, 0.2244, 0.0092, -0.0, 0.2266, 0.0071, -0.0, 0.2266, 0.0049, -0.0, 0.2244, 0.0027, -0.0, 0.2244, 0.0049, -0.0, 0.2266, 0.0027, -0.0, 0.2266, 0.0191, -0.0, 0.2318, 0.0191, 0.0017, 0.2318, 0.0191, 0.0039, 0.2318, 0.0194, -0.0, 0.233, 0.02, -0.0, 0.2351, 0.0194, 0.0017, 0.2331, 0.0194, 0.0039, 0.2331, 0.02, 0.0017, 0.2352, 0.02, 0.0039, 0.2352, 0.0205, -0.0, 0.2373, 0.0208, -0.0, 0.2395, 0.0205, 0.0017, 0.2374, 0.0205, 0.0039, 0.2374, 0.0208, 0.0017, 0.2396, 0.0208, 0.0039, 0.2396, 0.021, -0.0, 0.2416, 0.0214, -0.0, 0.2443, 0.0213, -0.0, 0.2436, 0.021, 0.0017, 0.2418, 0.021, 0.0039, 0.2418, 0.0213, 0.0017, 0.244, 0.0212, 0.0017, 0.2429, 0.0213, 0.0039, 0.244,
0.0212, 0.0039, 0.2429, 0.0179, -0.0, 0.2285, 0.0157, -0.0, 0.2287, 0.0187, -0.0, 0.2308, 0.0157, -0.0, 0.2309, 0.0136, -0.0, 0.2287, 0.0114, -0.0, 0.2287, 0.0136, -0.0, 0.2309, 0.0114, -0.0, 0.2309, 0.018, 0.0017, 0.2287, 0.018, 0.0039, 0.2287, 0.0187, 0.0017, 0.2307, 0.0187, 0.0039, 0.2307, 0.0179, -0.0, 0.2331, 0.0157, -0.0, 0.2331, 0.0179, -0.0, 0.2353, 0.0157, -0.0, 0.2353, 0.0136, -0.0, 0.2331, 0.0114, -0.0, 0.2331, 0.0136, -0.0, 0.2353, 0.0114, -0.0, 0.2353, 0.0092, -0.0, 0.2287, 0.0071, -0.0, 0.2287, 0.0092, -0.0, 0.2309, 0.0071, -0.0, 0.2309, 0.0049, -0.0, 0.2287, 0.0026, -0.0, 0.2289, 0.0047, -0.0, 0.231, 0.0036, -0.0, 0.23, 0.0026, 0.0017, 0.2289,
0.0026, 0.0039, 0.2289, 0.0047, 0.0017, 0.231, 0.0036, 0.0017, 0.23, 0.0047, 0.0039, 0.231, 0.0036, 0.0039, 0.23, 0.0092, -0.0, 0.2331, 0.0063, -0.0, 0.2334, 0.0092, -0.0, 0.2353, 0.0075, -0.0, 0.2354, 0.0057, -0.0, 0.2323, 0.0064, 0.0017, 0.2335, 0.0064, 0.0039, 0.2335, 0.0074, 0.0017, 0.2352, 0.0074, 0.0039, 0.2352, 0.0057, 0.0017, 0.2324, 0.0057, 0.0039, 0.2324, 0.0179, -0.0, 0.2374, 0.0157, -0.0, 0.2374, 0.0179, -0.0, 0.2396, 0.0157, -0.0, 0.2396, 0.0136, -0.0, 0.2374, 0.0114, -0.0, 0.2374, 0.0136, -0.0, 0.2396, 0.0114, -0.0, 0.2396, 0.0179, -0.0, 0.2418, 0.0157, -0.0, 0.2418, 0.0179, -0.0, 0.2439, 0.0157, -0.0, 0.2439, 0.0136, -0.0, 0.2418, 0.0114, -0.0, 0.2418,
0.0136, -0.0, 0.2439, 0.0114, -0.0, 0.2439, 0.0081, -0.0, 0.2377, 0.0079, -0.0, 0.2369, 0.0086, -0.0, 0.2398, 0.0082, 0.0017, 0.2381, 0.0079, 0.0017, 0.2371, 0.0082, 0.0039, 0.2381, 0.0079, 0.0039, 0.2371, 0.0086, 0.0017, 0.2396, 0.0086, 0.0039, 0.2396, 0.009, -0.0, 0.2419, 0.0093, -0.0, 0.244, 0.009, 0.0017, 0.2418, 0.009, 0.0039, 0.2418, 0.0093, 0.0017, 0.2439, 0.0093, 0.0039, 0.2439, 0.067, -0.0, 0.2588, 0.067, -0.0, 0.2613, 0.0656, -0.0, 0.2588, 0.0635, -0.0, 0.2588, 0.0656, -0.0, 0.2613, 0.0631, -0.0, 0.2618, 0.067, 0.0017, 0.2588, 0.067, 0.0039, 0.2588, 0.067, 0.0017, 0.2613, 0.067, 0.0039, 0.2613, 0.0656, 0.0017, 0.2588, 0.0635, 0.0017, 0.2588, 0.0656, 0.0039, 0.2588,
0.0635, 0.0039, 0.2588, 0.0631, 0.0017, 0.2618, 0.0631, 0.0039, 0.2618, 0.0613, -0.0, 0.2588, 0.0591, -0.0, 0.2588, 0.0613, -0.0, 0.2618, 0.0591, -0.0, 0.2618, 0.057, -0.0, 0.2588, 0.0548, -0.0, 0.2588, 0.057, -0.0, 0.2618, 0.0548, -0.0, 0.2618, 0.0613, 0.0017, 0.2588, 0.0591, 0.0017, 0.2588, 0.0613, 0.0039, 0.2588, 0.0591, 0.0039, 0.2588, 0.0613, 0.0017, 0.2618, 0.0591, 0.0017, 0.2618, 0.0613, 0.0039, 0.2618, 0.0591, 0.0039, 0.2618, 0.057, 0.0017, 0.2588, 0.0548, 0.0017, 0.2588, 0.057, 0.0039, 0.2588, 0.0548, 0.0039, 0.2588, 0.057, 0.0017, 0.2618, 0.0548, 0.0017, 0.2618, 0.057, 0.0039, 0.2618, 0.0548, 0.0039, 0.2618, 0.0526, -0.0, 0.2588, 0.0505, -0.0, 0.2588, 0.0526, -0.0, 0.2618,
0.0505, -0.0, 0.2618, 0.0483, -0.0, 0.2588, 0.0461, -0.0, 0.2588, 0.0483, -0.0, 0.2618, 0.0461, -0.0, 0.2618, 0.0526, 0.0017, 0.2588, 0.0505, 0.0017, 0.2588, 0.0526, 0.0039, 0.2588, 0.0505, 0.0039, 0.2588, 0.0526, 0.0017, 0.2618, 0.0505, 0.0017, 0.2618, 0.0526, 0.0039, 0.2618, 0.0505, 0.0039, 0.2618, 0.0483, 0.0017, 0.2588, 0.0461, 0.0017, 0.2588, 0.0483, 0.0039, 0.2588, 0.0461, 0.0039, 0.2588, 0.0483, 0.0017, 0.2618, 0.0461, 0.0017, 0.2618, 0.0483, 0.0039, 0.2618, 0.0461, 0.0039, 0.2618, 0.0439, -0.0, 0.2588, 0.0425, -0.0, 0.2588, 0.0439, -0.0, 0.2618, 0.0425, -0.0, 0.2618, 0.0439, 0.0017, 0.2588, 0.0425, 0.0017, 0.2588, 0.0439, 0.0039, 0.2588, 0.0425, 0.0039, 0.2588, 0.0439, 0.0017, 0.2618,
0.0425, 0.0017, 0.2618, 0.0439, 0.0039, 0.2618, 0.0425, 0.0039, 0.2618, 0.067, -0.0, 0.2635, 0.067, -0.0, 0.2656, 0.0656, -0.0, 0.2635, 0.0631, -0.0, 0.2635, 0.0656, -0.0, 0.2656, 0.0631, -0.0, 0.2656, 0.067, 0.0017, 0.2635, 0.067, 0.0039, 0.2635, 0.067, 0.0017, 0.2656, 0.067, 0.0039, 0.2656, 0.0631, 0.0017, 0.2635, 0.0631, 0.0039, 0.2635, 0.0631, 0.0017, 0.2656, 0.0631, 0.0039, 0.2656, 0.067, -0.0, 0.2678, 0.067, -0.0, 0.27, 0.0656, -0.0, 0.2678, 0.0631, -0.0, 0.2678, 0.0656, -0.0, 0.27, 0.0631, -0.0, 0.27, 0.067, 0.0017, 0.2678, 0.067, 0.0039, 0.2678, 0.067, 0.0017, 0.27, 0.067, 0.0039, 0.27, 0.0631, 0.0017, 0.2678, 0.0631, 0.0039, 0.2678, 0.0631, 0.0017, 0.27,
0.0631, 0.0039, 0.27, 0.067, -0.0, 0.2721, 0.067, -0.0, 0.2743, 0.0656, -0.0, 0.2721, 0.0631, -0.0, 0.2721, 0.0656, -0.0, 0.2743, 0.0631, -0.0, 0.2743, 0.067, 0.0017, 0.2721, 0.067, 0.0039, 0.2721, 0.067, 0.0017, 0.2743, 0.067, 0.0039, 0.2743, 0.0631, 0.0017, 0.2721, 0.0631, 0.0039, 0.2721, 0.0631, 0.0017, 0.2743, 0.0631, 0.0039, 0.2743, 0.067, -0.0, 0.2765, 0.067, -0.0, 0.2794, 0.0656, -0.0, 0.2765, 0.0631, -0.0, 0.2765, 0.0656, -0.0, 0.2794, 0.0631, -0.0, 0.2794, 0.067, 0.0017, 0.2765, 0.067, 0.0039, 0.2765, 0.067, 0.0017, 0.2794, 0.067, 0.0039, 0.2794, 0.0631, 0.0017, 0.2765, 0.0631, 0.0039, 0.2765, 0.0656, 0.0017, 0.2794, 0.0631, 0.0017, 0.2794, 0.0656, 0.0039, 0.2794,
0.0631, 0.0039, 0.2794, 0.0216, -0.0, 0.2461, 0.0201, -0.0, 0.2461, 0.0218, -0.0, 0.2482, 0.0201, -0.0, 0.2483, 0.0216, 0.0017, 0.2461, 0.0216, 0.0039, 0.2461, 0.0218, 0.0017, 0.2483, 0.0218, 0.0039, 0.2483, 0.0219, -0.0, 0.2504, 0.0201, -0.0, 0.2504, 0.0219, -0.0, 0.2526, 0.0201, -0.0, 0.2526, 0.0219, 0.0017, 0.2504, 0.0219, 0.0039, 0.2504, 0.0219, 0.0017, 0.2526, 0.0219, 0.0039, 0.2526, 0.0219, -0.0, 0.2548, 0.0201, -0.0, 0.2548, 0.0217, -0.0, 0.257, 0.0201, -0.0, 0.257, 0.0219, 0.0017, 0.2548, 0.0219, 0.0039, 0.2548, 0.0217, 0.0017, 0.257, 0.0217, 0.0039, 0.257, 0.0214, -0.0, 0.2591, 0.0201, -0.0, 0.2591, 0.0212, -0.0, 0.2605, 0.0211, -0.0, 0.2613, 0.0214, 0.0017, 0.2591,
0.0214, 0.0039, 0.2591, 0.0212, 0.0017, 0.2606, 0.0211, 0.0017, 0.2617, 0.0212, 0.0039, 0.2606, 0.0211, 0.0039, 0.2617, 0.0179, -0.0, 0.2461, 0.0157, -0.0, 0.2461, 0.0179, -0.0, 0.2483, 0.0157, -0.0, 0.2483, 0.0136, -0.0, 0.2461, 0.0114, -0.0, 0.2461, 0.0136, -0.0, 0.2483, 0.0114, -0.0, 0.2483, 0.0179, -0.0, 0.2504, 0.0157, -0.0, 0.2504, 0.0179, -0.0, 0.2526, 0.0157, -0.0, 0.2526, 0.0136, -0.0, 0.2504, 0.0114, -0.0, 0.2504, 0.0136, -0.0, 0.2526, 0.0114, -0.0, 0.2526, 0.0096, -0.0, 0.2462, 0.0099, -0.0, 0.2483, 0.0096, 0.0017, 0.2461, 0.0096, 0.0039, 0.2461, 0.0099, 0.0017, 0.2483, 0.0099, 0.0039, 0.2483, 0.0101, -0.0, 0.2504, 0.0101, -0.0, 0.2526, 0.0101, 0.0017, 0.2504,
0.0101, 0.0039, 0.2504, 0.0101, 0.0017, 0.2526, 0.0101, 0.0039, 0.2526, 0.0179, -0.0, 0.2548, 0.0157, -0.0, 0.2548, 0.0179, -0.0, 0.257, 0.0157, -0.0, 0.257, 0.0136, -0.0, 0.2548, 0.0114, -0.0, 0.2548, 0.0136, -0.0, 0.257, 0.0114, -0.0, 0.257, 0.0179, -0.0, 0.2591, 0.0157, -0.0, 0.2591, 0.0179, -0.0, 0.2613, 0.0157, -0.0, 0.2613, 0.0136, -0.0, 0.2591, 0.0114, -0.0, 0.2591, 0.0136, -0.0, 0.2613, 0.0114, -0.0, 0.2613, 0.0101, -0.0, 0.2548, 0.01, -0.0, 0.2569, 0.0101, 0.0017, 0.2548, 0.0101, 0.0039, 0.2548, 0.01, 0.0017, 0.257, 0.01, 0.0039, 0.257, 0.0097, -0.0, 0.2591, 0.0093, -0.0, 0.2612, 0.0097, 0.0017, 0.2591, 0.0097, 0.0039, 0.2591, 0.0093, 0.0017, 0.2613,
0.0093, 0.0039, 0.2613, 0.0209, -0.0, 0.2636, 0.0206, -0.0, 0.2658, 0.0209, 0.0017, 0.2635, 0.0209, 0.0039, 0.2635, 0.0206, 0.0017, 0.2656, 0.0206, 0.0039, 0.2656, 0.0202, -0.0, 0.2679, 0.0197, -0.0, 0.2701, 0.0202, 0.0017, 0.2678, 0.0202, 0.0039, 0.2678, 0.0197, 0.0017, 0.27, 0.0197, 0.0039, 0.27, 0.0192, -0.0, 0.2715, 0.0192, 0.0017, 0.2717, 0.0192, 0.0039, 0.2717, 0.0179, -0.0, 0.2635, 0.0157, -0.0, 0.2635, 0.0179, -0.0, 0.2656, 0.0157, -0.0, 0.2656, 0.0136, -0.0, 0.2635, 0.0114, -0.0, 0.2635, 0.0136, -0.0, 0.2656, 0.0114, -0.0, 0.2656, 0.0179, -0.0, 0.2678, 0.0157, -0.0, 0.2678, 0.0179, -0.0, 0.27, 0.0157, -0.0, 0.27, 0.0136, -0.0, 0.2678, 0.0114, -0.0, 0.2678,
0.0136, -0.0, 0.27, 0.0114, -0.0, 0.27, 0.009, -0.0, 0.2633, 0.0084, -0.0, 0.2656, 0.0081, -0.0, 0.2666, 0.0089, 0.0017, 0.2635, 0.0089, 0.0039, 0.2635, 0.0085, 0.0017, 0.2655, 0.0081, 0.0017, 0.2666, 0.0085, 0.0039, 0.2655, 0.0081, 0.0039, 0.2666, 0.0092, -0.0, 0.2678, 0.0077, -0.0, 0.2677, 0.0092, -0.0, 0.27, 0.0066, -0.0, 0.2699, 0.0059, -0.0, 0.271, 0.0076, 0.0017, 0.2678, 0.0076, 0.0039, 0.2678, 0.0066, 0.0017, 0.27, 0.0066, 0.0039, 0.27, 0.0059, 0.0017, 0.271, 0.0059, 0.0039, 0.271, 0.019, -0.0, 0.2725, 0.0157, -0.0, 0.2721, 0.0182, -0.0, 0.2746, 0.0157, -0.0, 0.2743, 0.0136, -0.0, 0.2721, 0.0114, -0.0, 0.2721, 0.0136, -0.0, 0.2743, 0.0114, -0.0, 0.2743,
0.0189, 0.0017, 0.2728, 0.0189, 0.0039, 0.2728, 0.0183, 0.0017, 0.2743, 0.0183, 0.0039, 0.2743, 0.0174, -0.0, 0.2766, 0.0157, -0.0, 0.2765, 0.0169, -0.0, 0.2776, 0.0163, -0.0, 0.2788, 0.0136, -0.0, 0.2765, 0.0114, -0.0, 0.2765, 0.0136, -0.0, 0.2787, 0.0114, -0.0, 0.2787, 0.0175, 0.0017, 0.2765, 0.0175, 0.0039, 0.2765, 0.0169, 0.0017, 0.2776, 0.0163, 0.0017, 0.2788, 0.0169, 0.0039, 0.2776, 0.0163, 0.0039, 0.2788, 0.0092, -0.0, 0.2721, 0.0071, -0.0, 0.2721, 0.0092, -0.0, 0.2743, 0.0071, -0.0, 0.2743, 0.0053, -0.0, 0.272, 0.004, -0.0, 0.2735, 0.0029, -0.0, 0.2746, 0.0051, 0.0017, 0.2722, 0.0051, 0.0039, 0.2722, 0.004, 0.0017, 0.2735, 0.0029, 0.0017, 0.2746, 0.004, 0.0039, 0.2735,
0.0029, 0.0039, 0.2746, 0.0092, -0.0, 0.2765, 0.0071, -0.0, 0.2765, 0.0092, -0.0, 0.2787, 0.0071, -0.0, 0.2787, 0.0049, -0.0, 0.2765, 0.0019, -0.0, 0.2754, 0.0049, -0.0, 0.2787, 0.0027, -0.0, 0.2787, 0.0017, 0.0017, 0.2755, 0.0017, 0.0039, 0.2755, 0.067, -0.0, 0.298, 0.067, -0.0, 0.3004, 0.0656, -0.0, 0.298, 0.0635, -0.0, 0.298, 0.0656, -0.0, 0.3004, 0.0635, -0.0, 0.3004, 0.067, 0.0017, 0.298, 0.067, 0.0039, 0.298, 0.067, 0.0017, 0.3004, 0.067, 0.0039, 0.3004, 0.0656, 0.0017, 0.298, 0.0635, 0.0017, 0.298, 0.0656, 0.0039, 0.298, 0.0635, 0.0039, 0.298, 0.067, -0.0, 0.3025, 0.067, -0.0, 0.3047, 0.0656, -0.0, 0.3025, 0.0631, -0.0, 0.302, 0.0656, -0.0, 0.3047,
0.0631, -0.0, 0.3047, 0.067, 0.0017, 0.3025, 0.067, 0.0039, 0.3025, 0.067, 0.0017, 0.3047, 0.067, 0.0039, 0.3047, 0.0631, 0.0017, 0.302, 0.0631, 0.0039, 0.302, 0.0631, 0.0017, 0.3047, 0.0631, 0.0039, 0.3047, 0.0613, -0.0, 0.298, 0.0591, -0.0, 0.298, 0.0613, -0.0, 0.3004, 0.0591, -0.0, 0.3004, 0.057, -0.0, 0.298, 0.0548, -0.0, 0.298, 0.057, -0.0, 0.3004, 0.0548, -0.0, 0.3004, 0.0613, 0.0017, 0.298, 0.0591, 0.0017, 0.298, 0.0613, 0.0039, 0.298, 0.0591, 0.0039, 0.298, 0.057, 0.0017, 0.298, 0.0548, 0.0017, 0.298, 0.057, 0.0039, 0.298, 0.0548, 0.0039, 0.298, 0.0613, -0.0, 0.302, 0.0591, -0.0, 0.302, 0.057, -0.0, 0.302, 0.0548, -0.0, 0.302, 0.0562, -0.0, 0.3049,
0.0548, -0.0, 0.3049, 0.0613, 0.0017, 0.302, 0.0591, 0.0017, 0.302, 0.0613, 0.0039, 0.302, 0.0591, 0.0039, 0.302, 0.057, 0.0017, 0.302, 0.0548, 0.0017, 0.302, 0.057, 0.0039, 0.302, 0.0548, 0.0039, 0.302, 0.0562, 0.0017, 0.3049, 0.0548, 0.0017, 0.3049, 0.0562, 0.0039, 0.3049, 0.0548, 0.0039, 0.3049, 0.067, -0.0, 0.3069, 0.067, -0.0, 0.309, 0.0656, -0.0, 0.3069, 0.0631, -0.0, 0.3069, 0.0656, -0.0, 0.309, 0.0631, -0.0, 0.309, 0.067, 0.0017, 0.3069, 0.067, 0.0039, 0.3069, 0.067, 0.0017, 0.309, 0.067, 0.0039, 0.309, 0.0631, 0.0017, 0.3069, 0.0631, 0.0039, 0.3069, 0.0631, 0.0017, 0.309, 0.0631, 0.0039, 0.309, 0.067, -0.0, 0.3112, 0.067, -0.0, 0.3134, 0.0656, -0.0, 0.3112,
0.0631, -0.0, 0.3112, 0.0656, -0.0, 0.3134, 0.0631, -0.0, 0.3134, 0.067, 0.0017, 0.3112, 0.067, 0.0039, 0.3112, 0.067, 0.0017, 0.3134, 0.067, 0.0039, 0.3134, 0.0631, 0.0017, 0.3112, 0.0631, 0.0039, 0.3112, 0.0631, 0.0017, 0.3134, 0.0631, 0.0039, 0.3134, 0.0562, -0.0, 0.3069, 0.0548, -0.0, 0.3069, 0.0562, -0.0, 0.309, 0.0548, -0.0, 0.309, 0.0562, 0.0017, 0.3069, 0.0562, 0.0039, 0.3069, 0.0562, 0.0017, 0.309, 0.0562, 0.0039, 0.309, 0.0562, -0.0, 0.3112, 0.0548, -0.0, 0.3112, 0.0562, -0.0, 0.3134, 0.0548, -0.0, 0.3134, 0.0562, 0.0017, 0.3112, 0.0562, 0.0039, 0.3112, 0.0562, 0.0017, 0.3134, 0.0562, 0.0039, 0.3134, 0.0526, -0.0, 0.298, 0.0505, -0.0, 0.298, 0.0526, -0.0, 0.3004,
0.0505, -0.0, 0.3004, 0.0483, -0.0, 0.298, 0.0461, -0.0, 0.298, 0.0483, -0.0, 0.3004, 0.0461, -0.0, 0.3004, 0.0526, 0.0017, 0.298, 0.0505, 0.0017, 0.298, 0.0526, 0.0039, 0.298, 0.0505, 0.0039, 0.298, 0.0483, 0.0017, 0.298, 0.0461, 0.0017, 0.298, 0.0483, 0.0039, 0.298, 0.0461, 0.0039, 0.298, 0.0526, -0.0, 0.302, 0.0505, -0.0, 0.302, 0.0523, -0.0, 0.3049, 0.0483, -0.0, 0.302, 0.0454, -0.0, 0.302, 0.0454, -0.0, 0.3047, 0.0526, 0.0017, 0.302, 0.0505, 0.0017, 0.302, 0.0526, 0.0039, 0.302, 0.0505, 0.0039, 0.302, 0.0523, 0.0017, 0.3049, 0.0523, 0.0039, 0.3049, 0.0483, 0.0017, 0.302, 0.0454, 0.0017, 0.302, 0.0483, 0.0039, 0.302, 0.0454, 0.0039, 0.302, 0.0454, 0.0017, 0.3047,
0.0454, 0.0039, 0.3047, 0.0439, -0.0, 0.298, 0.0425, -0.0, 0.298, 0.0439, -0.0, 0.3004, 0.0425, -0.0, 0.3004, 0.0439, 0.0017, 0.298, 0.0425, 0.0017, 0.298, 0.0439, 0.0039, 0.298, 0.0425, 0.0039, 0.298, 0.0425, 0.0017, 0.3004, 0.0425, 0.0039, 0.3004, 0.0439, -0.0, 0.3025, 0.0425, -0.0, 0.3025, 0.0439, -0.0, 0.3047, 0.0425, -0.0, 0.3047, 0.0425, 0.0017, 0.3025, 0.0425, 0.0039, 0.3025, 0.0425, 0.0017, 0.3047, 0.0425, 0.0039, 0.3047, 0.0523, -0.0, 0.3069, 0.0523, -0.0, 0.309, 0.0454, -0.0, 0.3069, 0.0454, -0.0, 0.309, 0.0523, 0.0017, 0.3069, 0.0523, 0.0039, 0.3069, 0.0523, 0.0017, 0.309, 0.0523, 0.0039, 0.309, 0.0454, 0.0017, 0.3069, 0.0454, 0.0039, 0.3069, 0.0454, 0.0017, 0.309,
0.0454, 0.0039, 0.309, 0.0523, -0.0, 0.3112, 0.0523, -0.0, 0.3134, 0.0454, -0.0, 0.3112, 0.0454, -0.0, 0.3134, 0.0523, 0.0017, 0.3112, 0.0523, 0.0039, 0.3112, 0.0523, 0.0017, 0.3134, 0.0523, 0.0039, 0.3134, 0.0454, 0.0017, 0.3112, 0.0454, 0.0039, 0.3112, 0.0454, 0.0017, 0.3134, 0.0454, 0.0039, 0.3134, 0.0439, -0.0, 0.3069, 0.0425, -0.0, 0.3069, 0.0439, -0.0, 0.309, 0.0425, -0.0, 0.309, 0.0425, 0.0017, 0.3069, 0.0425, 0.0039, 0.3069, 0.0425, 0.0017, 0.309, 0.0425, 0.0039, 0.309, 0.0439, -0.0, 0.3112, 0.0425, -0.0, 0.3112, 0.0439, -0.0, 0.3134, 0.0425, -0.0, 0.3134, 0.0425, 0.0017, 0.3112, 0.0425, 0.0039, 0.3112, 0.0425, 0.0017, 0.3134, 0.0425, 0.0039, 0.3134, 0.0152, -0.0, 0.2804,
0.0144, -0.0, 0.2815, 0.0114, -0.0, 0.2808, 0.0133, -0.0, 0.2827, 0.0122, -0.0, 0.2838, 0.0152, 0.0017, 0.2805, 0.0152, 0.0039, 0.2805, 0.0143, 0.0017, 0.2815, 0.0143, 0.0039, 0.2815, 0.0133, 0.0017, 0.2828, 0.0122, 0.0017, 0.2838, 0.0133, 0.0039, 0.2828, 0.0122, 0.0039, 0.2838, 0.0111, -0.0, 0.2848, 0.0111, 0.0017, 0.2847, 0.0111, 0.0039, 0.2847, 0.0092, -0.0, 0.2808, 0.0071, -0.0, 0.2808, 0.0092, -0.0, 0.283, 0.0071, -0.0, 0.283, 0.0049, -0.0, 0.2808, 0.0027, -0.0, 0.2808, 0.0049, -0.0, 0.283, 0.0027, -0.0, 0.283, 0.0096, -0.0, 0.2858, 0.0071, -0.0, 0.2852, 0.0085, -0.0, 0.2866, 0.0073, -0.0, 0.2873, 0.0049, -0.0, 0.2852, 0.0027, -0.0, 0.2852, 0.0052, -0.0, 0.2883,
0.0027, -0.0, 0.2873, 0.0097, 0.0017, 0.2858, 0.0097, 0.0039, 0.2858, 0.0086, 0.0017, 0.2865, 0.0071, 0.0017, 0.2874, 0.0086, 0.0039, 0.2865, 0.0071, 0.0039, 0.2874, 0.0054, 0.0017, 0.2882, 0.0054, 0.0039, 0.2882, 0.0042, -0.0, 0.2887, 0.0028, -0.0, 0.2891, 0.0043, 0.0017, 0.2886, 0.0027, 0.0017, 0.2892, 0.0043, 0.0039, 0.2886, 0.0027, 0.0039, 0.2892, 0.0219, -0.0, 0.3088, 0.0201, -0.0, 0.3088, 0.0219, 0.0017, 0.3088, 0.0201, 0.0017, 0.3088, 0.0219, 0.0039, 0.3088, 0.0201, 0.0039, 0.3088, 0.0219, -0.0, 0.3112, 0.0201, -0.0, 0.3112, 0.0219, -0.0, 0.3134, 0.0201, -0.0, 0.3134, 0.0219, 0.0017, 0.3112, 0.0219, 0.0039, 0.3112, 0.0219, 0.0017, 0.3134, 0.0219, 0.0039, 0.3134, 0.0179, -0.0, 0.3088,
0.0157, -0.0, 0.3088, 0.0136, -0.0, 0.3088, 0.0114, -0.0, 0.3088, 0.0179, 0.0017, 0.3088, 0.0157, 0.0017, 0.3088, 0.0179, 0.0039, 0.3088, 0.0157, 0.0039, 0.3088, 0.0136, 0.0017, 0.3088, 0.0114, 0.0017, 0.3088, 0.0136, 0.0039, 0.3088, 0.0114, 0.0039, 0.3088, 0.0179, -0.0, 0.3112, 0.0157, -0.0, 0.3112, 0.0179, -0.0, 0.3134, 0.0157, -0.0, 0.3134, 0.0136, -0.0, 0.3112, 0.0114, -0.0, 0.3112, 0.0136, -0.0, 0.3134, 0.0114, -0.0, 0.3134, 0.0092, -0.0, 0.3088, 0.0071, -0.0, 0.3088, 0.0049, -0.0, 0.3088, 0.0027, -0.0, 0.3088, 0.0092, 0.0017, 0.3088, 0.0071, 0.0017, 0.3088, 0.0092, 0.0039, 0.3088, 0.0071, 0.0039, 0.3088, 0.0049, 0.0017, 0.3088, 0.0027, 0.0017, 0.3088, 0.0049, 0.0039, 0.3088,
0.0027, 0.0039, 0.3088, 0.0092, -0.0, 0.3112, 0.0071, -0.0, 0.3112, 0.0092, -0.0, 0.3134, 0.0071, -0.0, 0.3134, 0.0049, -0.0, 0.3112, 0.0027, -0.0, 0.3112, 0.0049, -0.0, 0.3134, 0.0027, -0.0, 0.3134, 0.067, -0.0, 0.3155, 0.067, -0.0, 0.3177, 0.0656, -0.0, 0.3155, 0.0631, -0.0, 0.3155, 0.0656, -0.0, 0.3177, 0.0631, -0.0, 0.3177, 0.067, 0.0017, 0.3155, 0.067, 0.0039, 0.3155, 0.067, 0.0017, 0.3177, 0.067, 0.0039, 0.3177, 0.0631, 0.0017, 0.3155, 0.0631, 0.0039, 0.3155, 0.0631, 0.0017, 0.3177, 0.0631, 0.0039, 0.3177, 0.067, -0.0, 0.3196, 0.0656, -0.0, 0.3196, 0.0631, -0.0, 0.3196, 0.067, 0.0017, 0.3196, 0.067, 0.0039, 0.3196, 0.0656, 0.0017, 0.3196, 0.0631, 0.0017, 0.3196,
0.0656, 0.0039, 0.3196, 0.0631, 0.0039, 0.3196, 0.0562, -0.0, 0.3155, 0.0548, -0.0, 0.3155, 0.0562, -0.0, 0.3176, 0.0548, -0.0, 0.3176, 0.0562, 0.0017, 0.3155, 0.0562, 0.0039, 0.3155, 0.0562, 0.0017, 0.3176, 0.0548, 0.0017, 0.3176, 0.0562, 0.0039, 0.3176, 0.0548, 0.0039, 0.3176, 0.0523, -0.0, 0.3155, 0.0523, -0.0, 0.3176, 0.0454, -0.0, 0.3155, 0.0454, -0.0, 0.3177, 0.0523, 0.0017, 0.3155, 0.0523, 0.0039, 0.3155, 0.0523, 0.0017, 0.3176, 0.0523, 0.0039, 0.3176, 0.0454, 0.0017, 0.3155, 0.0454, 0.0039, 0.3155, 0.0454, 0.0017, 0.3177, 0.0454, 0.0039, 0.3177, 0.0454, -0.0, 0.3196, 0.0454, 0.0017, 0.3196, 0.0454, 0.0039, 0.3196, 0.0439, -0.0, 0.3155, 0.0425, -0.0, 0.3155, 0.0439, -0.0, 0.3177,
0.0425, -0.0, 0.3177, 0.0425, 0.0017, 0.3155, 0.0425, 0.0039, 0.3155, 0.0425, 0.0017, 0.3177, 0.0425, 0.0039, 0.3177, 0.0439, -0.0, 0.3196, 0.0425, -0.0, 0.3196, 0.0439, 0.0017, 0.3196, 0.0425, 0.0017, 0.3196, 0.0439, 0.0039, 0.3196, 0.0425, 0.0039, 0.3196, 0.067, -0.0, 0.3402, 0.0656, -0.0, 0.3402, 0.0635, -0.0, 0.3402, 0.067, 0.0017, 0.3402, 0.067, 0.0039, 0.3402, 0.0656, 0.0017, 0.3402, 0.0635, 0.0017, 0.3402, 0.0656, 0.0039, 0.3402, 0.0635, 0.0039, 0.3402, 0.0613, -0.0, 0.3402, 0.0591, -0.0, 0.3402, 0.057, -0.0, 0.3402, 0.0548, -0.0, 0.3402, 0.0613, 0.0017, 0.3402, 0.0591, 0.0017, 0.3402, 0.0613, 0.0039, 0.3402, 0.0591, 0.0039, 0.3402, 0.057, 0.0017, 0.3402, 0.0548, 0.0017, 0.3402,
0.057, 0.0039, 0.3402, 0.0548, 0.0039, 0.3402, 0.067, -0.0, 0.3416, 0.067, -0.0, 0.3431, 0.0656, -0.0, 0.3416, 0.0635, -0.0, 0.3416, 0.0656, -0.0, 0.3431, 0.0635, -0.0, 0.3431, 0.067, 0.0017, 0.3416, 0.067, 0.0039, 0.3416, 0.067, 0.0017, 0.3431, 0.067, 0.0039, 0.3431, 0.0656, 0.0017, 0.3431, 0.0635, 0.0017, 0.3431, 0.0656, 0.0039, 0.3431, 0.0635, 0.0039, 0.3431, 0.0613, -0.0, 0.3416, 0.0591, -0.0, 0.3416, 0.0613, -0.0, 0.3431, 0.0591, -0.0, 0.3431, 0.057, -0.0, 0.3416, 0.0548, -0.0, 0.3416, 0.057, -0.0, 0.3431, 0.0548, -0.0, 0.3431, 0.0613, 0.0017, 0.3431, 0.0591, 0.0017, 0.3431, 0.0613, 0.0039, 0.3431, 0.0591, 0.0039, 0.3431, 0.057, 0.0017, 0.3431, 0.0548, 0.0017, 0.3431,
0.057, 0.0039, 0.3431, 0.0548, 0.0039, 0.3431, 0.0546, -0.0, 0.3483, 0.0545, 0.0017, 0.3483, 0.0545, 0.0039, 0.3483, 0.0526, -0.0, 0.3402, 0.0505, -0.0, 0.3402, 0.0483, -0.0, 0.3402, 0.0461, -0.0, 0.3402, 0.0526, 0.0017, 0.3402, 0.0505, 0.0017, 0.3402, 0.0526, 0.0039, 0.3402, 0.0505, 0.0039, 0.3402, 0.0483, 0.0017, 0.3402, 0.0461, 0.0017, 0.3402, 0.0483, 0.0039, 0.3402, 0.0461, 0.0039, 0.3402, 0.0439, -0.0, 0.3402, 0.0425, -0.0, 0.3402, 0.0439, 0.0017, 0.3402, 0.0425, 0.0017, 0.3402, 0.0439, 0.0039, 0.3402, 0.0425, 0.0039, 0.3402, 0.0526, -0.0, 0.3416, 0.0505, -0.0, 0.3416, 0.0526, -0.0, 0.3431, 0.0499, -0.0, 0.3431, 0.0483, -0.0, 0.3416, 0.0461, -0.0, 0.3416, 0.0483, -0.0, 0.3438,
0.0461, -0.0, 0.3438, 0.0526, 0.0017, 0.3431, 0.0499, 0.0017, 0.3431, 0.0526, 0.0039, 0.3431, 0.0499, 0.0039, 0.3431, 0.0525, -0.0, 0.346, 0.0515, -0.0, 0.345, 0.0535, -0.0, 0.3472, 0.0505, -0.0, 0.3481, 0.0483, -0.0, 0.3459, 0.0457, -0.0, 0.3462, 0.0479, -0.0, 0.3485, 0.0468, -0.0, 0.3474, 0.0524, 0.0017, 0.346, 0.0515, 0.0017, 0.3449, 0.0524, 0.0039, 0.346, 0.0515, 0.0039, 0.3449, 0.0535, 0.0017, 0.3472, 0.0535, 0.0039, 0.3472, 0.0457, 0.0017, 0.3462, 0.0457, 0.0039, 0.3462, 0.0479, 0.0017, 0.3485, 0.0468, 0.0017, 0.3474, 0.0479, 0.0039, 0.3485, 0.0468, 0.0039, 0.3474, 0.0429, -0.0, 0.3421, 0.0426, -0.0, 0.3411, 0.0439, -0.0, 0.3441, 0.043, 0.0017, 0.3423, 0.0426, 0.0017, 0.3413,
0.043, 0.0039, 0.3423, 0.0426, 0.0039, 0.3413, 0.0438, 0.0017, 0.3438, 0.0438, 0.0039, 0.3438, 0.0448, -0.0, 0.3451, 0.0448, 0.0017, 0.3452, 0.0448, 0.0039, 0.3452, 0.0219, -0.0, 0.3155, 0.0201, -0.0, 0.3155, 0.0219, -0.0, 0.3177, 0.0201, -0.0, 0.3177, 0.0219, 0.0017, 0.3155, 0.0219, 0.0039, 0.3155, 0.0219, 0.0017, 0.3177, 0.0219, 0.0039, 0.3177, 0.0219, -0.0, 0.3206, 0.0201, -0.0, 0.3206, 0.0219, 0.0017, 0.3206, 0.0201, 0.0017, 0.3206, 0.0219, 0.0039, 0.3206, 0.0201, 0.0039, 0.3206, 0.0179, -0.0, 0.3155, 0.0157, -0.0, 0.3155, 0.0179, -0.0, 0.3177, 0.0157, -0.0, 0.3177, 0.0136, -0.0, 0.3155, 0.0114, -0.0, 0.3155, 0.0136, -0.0, 0.3177, 0.0114, -0.0, 0.3177, 0.0179, -0.0, 0.3206,
0.0157, -0.0, 0.3206, 0.0136, -0.0, 0.3206, 0.0114, -0.0, 0.3206, 0.0179, 0.0017, 0.3206, 0.0157, 0.0017, 0.3206, 0.0179, 0.0039, 0.3206, 0.0157, 0.0039, 0.3206, 0.0136, 0.0017, 0.3206, 0.0114, 0.0017, 0.3206, 0.0136, 0.0039, 0.3206, 0.0114, 0.0039, 0.3206, 0.0092, -0.0, 0.3155, 0.0071, -0.0, 0.3155, 0.0092, -0.0, 0.3177, 0.0071, -0.0, 0.3177, 0.0049, -0.0, 0.3155, 0.0027, -0.0, 0.3155, 0.0049, -0.0, 0.3177, 0.0027, -0.0, 0.3177, 0.0092, -0.0, 0.3206, 0.0071, -0.0, 0.3206, 0.0049, -0.0, 0.3206, 0.0027, -0.0, 0.3206, 0.0092, 0.0017, 0.3206, 0.0071, 0.0017, 0.3206, 0.0092, 0.0039, 0.3206, 0.0071, 0.0039, 0.3206, 0.0049, 0.0017, 0.3206, 0.0027, 0.0017, 0.3206, 0.0049, 0.0039, 0.3206,
0.0027, 0.0039, 0.3206, 0.0219, -0.0, 0.3402, 0.0201, -0.0, 0.3402, 0.0219, 0.0017, 0.3402, 0.0201, 0.0017, 0.3402, 0.0219, 0.0039, 0.3402, 0.0201, 0.0039, 0.3402, 0.0219, -0.0, 0.3416, 0.0201, -0.0, 0.3416, 0.0219, -0.0, 0.3438, 0.0201, -0.0, 0.3438, 0.0219, 0.0017, 0.3416, 0.0219, 0.0039, 0.3416, 0.0219, 0.0017, 0.3438, 0.0219, 0.0039, 0.3438, 0.0219, -0.0, 0.3459, 0.0201, -0.0, 0.3459, 0.0219, -0.0, 0.3481, 0.0201, -0.0, 0.3481, 0.0219, 0.0017, 0.3459, 0.0219, 0.0039, 0.3459, 0.0219, 0.0017, 0.3481, 0.0219, 0.0039, 0.3481, 0.0179, -0.0, 0.3402, 0.0157, -0.0, 0.3402, 0.0136, -0.0, 0.3402, 0.0114, -0.0, 0.3402, 0.0179, 0.0017, 0.3402, 0.0157, 0.0017, 0.3402, 0.0179, 0.0039, 0.3402,
0.0157, 0.0039, 0.3402, 0.0136, 0.0017, 0.3402, 0.0114, 0.0017, 0.3402, 0.0136, 0.0039, 0.3402, 0.0114, 0.0039, 0.3402, 0.0092, -0.0, 0.3402, 0.0071, -0.0, 0.3402, 0.0049, -0.0, 0.3402, 0.0027, -0.0, 0.3402, 0.0092, 0.0017, 0.3402, 0.0071, 0.0017, 0.3402, 0.0092, 0.0039, 0.3402, 0.0071, 0.0039, 0.3402, 0.0049, 0.0017, 0.3402, 0.0027, 0.0017, 0.3402, 0.0049, 0.0039, 0.3402, 0.0027, 0.0039, 0.3402, 0.0179, -0.0, 0.3416, 0.0157, -0.0, 0.3416, 0.0179, -0.0, 0.3438, 0.0157, -0.0, 0.3438, 0.0136, -0.0, 0.3416, 0.0114, -0.0, 0.3416, 0.0136, -0.0, 0.3438, 0.0114, -0.0, 0.3438, 0.0179, -0.0, 0.3459, 0.0157, -0.0, 0.3459, 0.0179, -0.0, 0.3481, 0.0157, -0.0, 0.3481, 0.0136, -0.0, 0.3459,
0.0114, -0.0, 0.3459, 0.0136, -0.0, 0.3481, 0.0114, -0.0, 0.3481, 0.0092, -0.0, 0.3416, 0.0071, -0.0, 0.3416, 0.0092, -0.0, 0.3438, 0.0071, -0.0, 0.3438, 0.0049, -0.0, 0.3416, 0.0027, -0.0, 0.3416, 0.0049, -0.0, 0.3438, 0.0027, -0.0, 0.3438, 0.0092, -0.0, 0.3459, 0.0071, -0.0, 0.3459, 0.0092, -0.0, 0.3481, 0.0071, -0.0, 0.3481, 0.0049, -0.0, 0.3459, 0.0027, -0.0, 0.3459, 0.0049, -0.0, 0.3481, 0.0027, -0.0, 0.3481, 0.0631, -0.0, 0.3571, 0.0631, 0.0017, 0.357, 0.0631, 0.0039, 0.357, 0.0589, -0.0, 0.3527, 0.0567, -0.0, 0.3505, 0.0556, -0.0, 0.3494, 0.0578, -0.0, 0.3516, 0.0548, -0.0, 0.3524, 0.0589, 0.0017, 0.3527, 0.0589, 0.0039, 0.3527, 0.0567, 0.0017, 0.3505,
0.0556, 0.0017, 0.3494, 0.0567, 0.0039, 0.3505, 0.0556, 0.0039, 0.3494, 0.0578, 0.0017, 0.3516, 0.0578, 0.0039, 0.3516, 0.061, -0.0, 0.3549, 0.06, -0.0, 0.3538, 0.0621, -0.0, 0.356, 0.0591, -0.0, 0.3568, 0.057, -0.0, 0.3546, 0.0544, -0.0, 0.355, 0.0565, -0.0, 0.3572, 0.0554, -0.0, 0.3561, 0.061, 0.0017, 0.3549, 0.06, 0.0017, 0.3538, 0.061, 0.0039, 0.3549, 0.06, 0.0039, 0.3538, 0.0621, 0.0017, 0.356, 0.0621, 0.0039, 0.356, 0.0544, 0.0017, 0.355, 0.0544, 0.0039, 0.355, 0.0565, 0.0017, 0.3572, 0.0554, 0.0017, 0.3562, 0.0565, 0.0039, 0.3572, 0.0554, 0.0039, 0.3562, 0.0651, -0.0, 0.3594, 0.0642, -0.0, 0.3583, 0.066, -0.0, 0.3608, 0.0635, -0.0, 0.3611, 0.065, 0.0017, 0.3594,
0.0642, 0.0017, 0.3583, 0.065, 0.0039, 0.3594, 0.0642, 0.0039, 0.3583, 0.0662, 0.0017, 0.361, 0.0662, 0.0039, 0.361, 0.067, -0.0, 0.3637, 0.0635, -0.0, 0.3637, 0.067, 0.0017, 0.3637, 0.0635, 0.0017, 0.3637, 0.067, 0.0039, 0.3637, 0.0635, 0.0039, 0.3637, 0.0613, -0.0, 0.3589, 0.0583, -0.0, 0.3595, 0.0613, -0.0, 0.3611, 0.0591, -0.0, 0.3608, 0.0575, -0.0, 0.3585, 0.057, -0.0, 0.3608, 0.0548, -0.0, 0.3608, 0.0584, 0.0017, 0.3596, 0.0584, 0.0039, 0.3596, 0.0591, 0.0017, 0.3608, 0.0591, 0.0039, 0.3608, 0.0575, 0.0017, 0.3585, 0.0575, 0.0039, 0.3585, 0.057, 0.0017, 0.3608, 0.0548, 0.0017, 0.3608, 0.057, 0.0039, 0.3608, 0.0548, 0.0039, 0.3608, 0.0613, -0.0, 0.3637, 0.0591, -0.0, 0.3637,
0.057, -0.0, 0.3637, 0.0548, -0.0, 0.3637, 0.0613, 0.0017, 0.3637, 0.0591, 0.0017, 0.3637, 0.0613, 0.0039, 0.3637, 0.0591, 0.0039, 0.3637, 0.057, 0.0017, 0.3637, 0.0548, 0.0017, 0.3637, 0.057, 0.0039, 0.3637, 0.0548, 0.0039, 0.3637, 0.0526, -0.0, 0.3503, 0.05, -0.0, 0.3507, 0.0522, -0.0, 0.3529, 0.0511, -0.0, 0.3518, 0.0489, -0.0, 0.3496, 0.05, 0.0017, 0.3507, 0.05, 0.0039, 0.3507, 0.0522, 0.0017, 0.3529, 0.0511, 0.0017, 0.3518, 0.0522, 0.0039, 0.3529, 0.0511, 0.0039, 0.3518, 0.049, 0.0017, 0.3496, 0.049, 0.0039, 0.3496, 0.0533, -0.0, 0.354, 0.0533, 0.0017, 0.354, 0.0533, 0.0039, 0.354, 0.0526, -0.0, 0.3608, 0.0505, -0.0, 0.3608, 0.0483, -0.0, 0.3608, 0.0461, -0.0, 0.3608,
0.0526, 0.0017, 0.3608, 0.0505, 0.0017, 0.3608, 0.0526, 0.0039, 0.3608, 0.0505, 0.0039, 0.3608, 0.0483, 0.0017, 0.3608, 0.0461, 0.0017, 0.3608, 0.0483, 0.0039, 0.3608, 0.0461, 0.0039, 0.3608, 0.0526, -0.0, 0.3637, 0.0505, -0.0, 0.3637, 0.0483, -0.0, 0.3637, 0.0461, -0.0, 0.3637, 0.0526, 0.0017, 0.3637, 0.0505, 0.0017, 0.3637, 0.0526, 0.0039, 0.3637, 0.0505, 0.0039, 0.3637, 0.0483, 0.0017, 0.3637, 0.0461, 0.0017, 0.3637, 0.0483, 0.0039, 0.3637, 0.0461, 0.0039, 0.3637, 0.0439, -0.0, 0.3608, 0.0425, -0.0, 0.3608, 0.0439, 0.0017, 0.3608, 0.0425, 0.0017, 0.3608, 0.0439, 0.0039, 0.3608, 0.0425, 0.0039, 0.3608, 0.0439, -0.0, 0.3637, 0.0425, -0.0, 0.3637, 0.0439, 0.0017, 0.3637, 0.0425, 0.0017, 0.3637,
0.0439, 0.0039, 0.3637, 0.0425, 0.0039, 0.3637, 0.0219, -0.0, 0.3503, 0.0201, -0.0, 0.3503, 0.0219, -0.0, 0.3524, 0.0201, -0.0, 0.3524, 0.0219, 0.0017, 0.3503, 0.0219, 0.0039, 0.3503, 0.0219, 0.0017, 0.3524, 0.0219, 0.0039, 0.3524, 0.0219, -0.0, 0.3546, 0.0201, -0.0, 0.3546, 0.0219, -0.0, 0.3568, 0.0201, -0.0, 0.3568, 0.0219, 0.0017, 0.3546, 0.0219, 0.0039, 0.3546, 0.0219, 0.0017, 0.3568, 0.0219, 0.0039, 0.3568, 0.0219, -0.0, 0.3589, 0.0201, -0.0, 0.3589, 0.0219, -0.0, 0.3611, 0.0201, -0.0, 0.3611, 0.0219, 0.0017, 0.3589, 0.0219, 0.0039, 0.3589, 0.0219, 0.0017, 0.3611, 0.0219, 0.0039, 0.3611, 0.0219, -0.0, 0.3633, 0.0201, -0.0, 0.3633, 0.0219, -0.0, 0.3655, 0.0201, -0.0, 0.3655,
0.0219, 0.0017, 0.3633, 0.0219, 0.0039, 0.3633, 0.0219, 0.0017, 0.3655, 0.0219, 0.0039, 0.3655, 0.0179, -0.0, 0.3503, 0.0157, -0.0, 0.3503, 0.0179, -0.0, 0.3524, 0.0157, -0.0, 0.3524, 0.0136, -0.0, 0.3503, 0.0114, -0.0, 0.3503, 0.0136, -0.0, 0.3524, 0.0114, -0.0, 0.3524, 0.0179, -0.0, 0.3546, 0.0157, -0.0, 0.3546, 0.0179, -0.0, 0.3568, 0.0157, -0.0, 0.3568, 0.0136, -0.0, 0.3546, 0.0114, -0.0, 0.3546, 0.0136, -0.0, 0.3568, 0.0114, -0.0, 0.3568, 0.0092, -0.0, 0.3503, 0.0071, -0.0, 0.3503, 0.0101, -0.0, 0.352, 0.0071, -0.0, 0.352, 0.0049, -0.0, 0.3503, 0.0027, -0.0, 0.3503, 0.0049, -0.0, 0.352, 0.0027, -0.0, 0.352, 0.0101, 0.0017, 0.352, 0.0071, 0.0017, 0.352,
0.0101, 0.0039, 0.352, 0.0071, 0.0039, 0.352, 0.0049, 0.0017, 0.352, 0.0027, 0.0017, 0.352, 0.0049, 0.0039, 0.352, 0.0027, 0.0039, 0.352, 0.0101, -0.0, 0.3546, 0.0101, -0.0, 0.3568, 0.0101, 0.0017, 0.3546, 0.0101, 0.0039, 0.3546, 0.0101, 0.0017, 0.3568, 0.0101, 0.0039, 0.3568, 0.0179, -0.0, 0.3589, 0.0157, -0.0, 0.3589, 0.0179, -0.0, 0.3611, 0.0157, -0.0, 0.3611, 0.0136, -0.0, 0.3589, 0.0114, -0.0, 0.3589, 0.0136, -0.0, 0.3611, 0.0114, -0.0, 0.3611, 0.0179, -0.0, 0.3633, 0.0157, -0.0, 0.3633, 0.0179, -0.0, 0.3655, 0.0157, -0.0, 0.3655, 0.0136, -0.0, 0.3633, 0.0114, -0.0, 0.3633, 0.0136, -0.0, 0.3655, 0.0114, -0.0, 0.3655, 0.0101, -0.0, 0.3589, 0.0101, -0.0, 0.3611,
0.0101, 0.0017, 0.3589, 0.0101, 0.0039, 0.3589, 0.0101, 0.0017, 0.3611, 0.0101, 0.0039, 0.3611, 0.0101, -0.0, 0.3633, 0.0101, -0.0, 0.3655, 0.0101, 0.0017, 0.3633, 0.0101, 0.0039, 0.3633, 0.0101, 0.0017, 0.3655, 0.0101, 0.0039, 0.3655, 0.0219, -0.0, 0.3676, 0.0201, -0.0, 0.3676, 0.0219, -0.0, 0.3698, 0.0201, -0.0, 0.3698, 0.0219, 0.0017, 0.3676, 0.0219, 0.0039, 0.3676, 0.0219, 0.0017, 0.3698, 0.0219, 0.0039, 0.3698, 0.0219, -0.0, 0.372, 0.0201, -0.0, 0.372, 0.0219, -0.0, 0.3741, 0.0201, -0.0, 0.3741, 0.0219, 0.0017, 0.372, 0.0219, 0.0039, 0.372, 0.0219, 0.0017, 0.3741, 0.0219, 0.0039, 0.3741, 0.0219, -0.0, 0.3763, 0.0201, -0.0, 0.3763, 0.0219, -0.0, 0.3785, 0.0201, -0.0, 0.3785,
0.0219, 0.0017, 0.3763, 0.0219, 0.0039, 0.3763, 0.0219, 0.0017, 0.3785, 0.0219, 0.0039, 0.3785, 0.0218, -0.0, 0.3807, 0.0201, -0.0, 0.3806, 0.0215, -0.0, 0.3828, 0.0201, -0.0, 0.3828, 0.0218, 0.0017, 0.3807, 0.0218, 0.0039, 0.3807, 0.0215, 0.0017, 0.3828, 0.0215, 0.0039, 0.3828, 0.0179, -0.0, 0.3676, 0.0157, -0.0, 0.3676, 0.0179, -0.0, 0.3698, 0.0157, -0.0, 0.3698, 0.0136, -0.0, 0.3676, 0.0114, -0.0, 0.3676, 0.0136, -0.0, 0.3698, 0.0114, -0.0, 0.3698, 0.0179, -0.0, 0.372, 0.0157, -0.0, 0.372, 0.0179, -0.0, 0.3741, 0.0157, -0.0, 0.3741, 0.0136, -0.0, 0.372, 0.0114, -0.0, 0.372, 0.0136, -0.0, 0.3741, 0.0114, -0.0, 0.3741, 0.0101, -0.0, 0.3676, 0.0101, -0.0, 0.3698,
0.0101, 0.0017, 0.3676, 0.0101, 0.0039, 0.3676, 0.0101, 0.0017, 0.3698, 0.0101, 0.0039, 0.3698, 0.0101, -0.0, 0.372, 0.0101, -0.0, 0.3741, 0.0101, 0.0017, 0.372, 0.0101, 0.0039, 0.372, 0.0101, 0.0017, 0.3741, 0.0101, 0.0039, 0.3741, 0.0179, -0.0, 0.3763, 0.0157, -0.0, 0.3763, 0.0179, -0.0, 0.3785, 0.0157, -0.0, 0.3785, 0.0136, -0.0, 0.3763, 0.0114, -0.0, 0.3763, 0.0136, -0.0, 0.3785, 0.0114, -0.0, 0.3785, 0.0179, -0.0, 0.3806, 0.0157, -0.0, 0.3806, 0.0179, -0.0, 0.3828, 0.0157, -0.0, 0.3828, 0.0136, -0.0, 0.3806, 0.0114, -0.0, 0.3806, 0.0136, -0.0, 0.3828, 0.0114, -0.0, 0.3828, 0.0101, -0.0, 0.3763, 0.0101, -0.0, 0.3785, 0.0101, 0.0017, 0.3763, 0.0101, 0.0039, 0.3763,
0.0101, 0.0017, 0.3785, 0.0101, 0.0039, 0.3785, 0.0101, -0.0, 0.3806, 0.0099, -0.0, 0.3828, 0.0101, 0.0017, 0.3807, 0.0101, 0.0039, 0.3807, 0.0099, 0.0017, 0.3828, 0.0099, 0.0039, 0.3828, 0.067, -0.0, 0.3863, 0.0656, -0.0, 0.3863, 0.0635, -0.0, 0.3863, 0.067, 0.0017, 0.3863, 0.067, 0.0039, 0.3863, 0.0656, 0.0017, 0.3863, 0.0635, 0.0017, 0.3863, 0.0656, 0.0039, 0.3863, 0.0635, 0.0039, 0.3863, 0.067, -0.0, 0.3893, 0.067, -0.0, 0.3915, 0.0656, -0.0, 0.3893, 0.0631, -0.0, 0.3902, 0.0656, -0.0, 0.3915, 0.0631, -0.0, 0.3915, 0.067, 0.0017, 0.3893, 0.067, 0.0039, 0.3893, 0.067, 0.0017, 0.3915, 0.067, 0.0039, 0.3915, 0.0631, 0.0017, 0.3902, 0.0631, 0.0039, 0.3902, 0.0631, 0.0017, 0.3915,
0.0631, 0.0039, 0.3915, 0.0613, -0.0, 0.3863, 0.0591, -0.0, 0.3863, 0.057, -0.0, 0.3863, 0.0548, -0.0, 0.3863, 0.0613, 0.0017, 0.3863, 0.0591, 0.0017, 0.3863, 0.0613, 0.0039, 0.3863, 0.0591, 0.0039, 0.3863, 0.057, 0.0017, 0.3863, 0.0548, 0.0017, 0.3863, 0.057, 0.0039, 0.3863, 0.0548, 0.0039, 0.3863, 0.0613, -0.0, 0.3902, 0.0591, -0.0, 0.3902, 0.057, -0.0, 0.3902, 0.0548, -0.0, 0.3902, 0.0613, 0.0017, 0.3902, 0.0591, 0.0017, 0.3902, 0.0613, 0.0039, 0.3902, 0.0591, 0.0039, 0.3902, 0.057, 0.0017, 0.3902, 0.0548, 0.0017, 0.3902, 0.057, 0.0039, 0.3902, 0.0548, 0.0039, 0.3902, 0.067, -0.0, 0.3937, 0.067, -0.0, 0.3958, 0.0656, -0.0, 0.3937, 0.0631, -0.0, 0.3937, 0.0656, -0.0, 0.3958,
0.0631, -0.0, 0.3958, 0.067, 0.0017, 0.3937, 0.067, 0.0039, 0.3937, 0.067, 0.0017, 0.3958, 0.067, 0.0039, 0.3958, 0.0631, 0.0017, 0.3937, 0.0631, 0.0039, 0.3937, 0.0631, 0.0017, 0.3958, 0.0631, 0.0039, 0.3958, 0.067, -0.0, 0.398, 0.0669, -0.0, 0.4002, 0.0656, -0.0, 0.398, 0.063, -0.0, 0.398, 0.0656, -0.0, 0.4002, 0.063, -0.0, 0.4001, 0.067, 0.0017, 0.398, 0.067, 0.0039, 0.398, 0.0669, 0.0017, 0.4002, 0.0669, 0.0039, 0.4002, 0.063, 0.0017, 0.398, 0.063, 0.0039, 0.398, 0.063, 0.0017, 0.4002, 0.063, 0.0039, 0.4002, 0.0526, -0.0, 0.3863, 0.0505, -0.0, 0.3863, 0.0483, -0.0, 0.3863, 0.0461, -0.0, 0.3863, 0.0526, 0.0017, 0.3863, 0.0505, 0.0017, 0.3863, 0.0526, 0.0039, 0.3863,
0.0505, 0.0039, 0.3863, 0.0483, 0.0017, 0.3863, 0.0461, 0.0017, 0.3863, 0.0483, 0.0039, 0.3863, 0.0461, 0.0039, 0.3863, 0.0526, -0.0, 0.3902, 0.0505, -0.0, 0.3902, 0.0483, -0.0, 0.3902, 0.0454, -0.0, 0.3902, 0.0454, -0.0, 0.3915, 0.0526, 0.0017, 0.3902, 0.0505, 0.0017, 0.3902, 0.0526, 0.0039, 0.3902, 0.0505, 0.0039, 0.3902, 0.0483, 0.0017, 0.3902, 0.0454, 0.0017, 0.3902, 0.0483, 0.0039, 0.3902, 0.0454, 0.0039, 0.3902, 0.0454, 0.0017, 0.3915, 0.0454, 0.0039, 0.3915, 0.0439, -0.0, 0.3863, 0.0425, -0.0, 0.3863, 0.0439, 0.0017, 0.3863, 0.0425, 0.0017, 0.3863, 0.0439, 0.0039, 0.3863, 0.0425, 0.0039, 0.3863, 0.0439, -0.0, 0.3893, 0.0425, -0.0, 0.3893, 0.0439, -0.0, 0.3915, 0.0425, -0.0, 0.3915,
0.0425, 0.0017, 0.3893, 0.0425, 0.0039, 0.3893, 0.0425, 0.0017, 0.3915, 0.0425, 0.0039, 0.3915, 0.0454, -0.0, 0.3937, 0.0454, -0.0, 0.3958, 0.0454, 0.0017, 0.3937, 0.0454, 0.0039, 0.3937, 0.0454, 0.0017, 0.3958, 0.0454, 0.0039, 0.3958, 0.0455, -0.0, 0.398, 0.0457, -0.0, 0.4001, 0.0455, 0.0017, 0.398, 0.0455, 0.0039, 0.398, 0.0458, 0.0017, 0.4002, 0.0458, 0.0039, 0.4002, 0.0439, -0.0, 0.3937, 0.0425, -0.0, 0.3937, 0.0439, -0.0, 0.3958, 0.0425, -0.0, 0.3958, 0.0425, 0.0017, 0.3937, 0.0425, 0.0039, 0.3937, 0.0425, 0.0017, 0.3958, 0.0425, 0.0039, 0.3958, 0.0439, -0.0, 0.398, 0.0425, -0.0, 0.398, 0.0439, -0.0, 0.4002, 0.0425, -0.0, 0.4002, 0.0425, 0.0017, 0.398, 0.0425, 0.0039, 0.398,
0.0425, 0.0017, 0.4002, 0.0425, 0.0039, 0.4002, 0.0668, -0.0, 0.4016, 0.0667, -0.0, 0.4025, 0.0627, -0.0, 0.4023, 0.0661, -0.0, 0.4048, 0.0635, -0.0, 0.4045, 0.0668, 0.0017, 0.4017, 0.0668, 0.0039, 0.4017, 0.0666, 0.0017, 0.4028, 0.0627, 0.0017, 0.4023, 0.0666, 0.0039, 0.4028, 0.0627, 0.0039, 0.4023, 0.0662, 0.0017, 0.4046, 0.0662, 0.0039, 0.4046, 0.0653, -0.0, 0.4064, 0.0645, -0.0, 0.4075, 0.0635, -0.0, 0.4089, 0.0652, 0.0017, 0.4066, 0.0644, 0.0017, 0.4076, 0.0652, 0.0039, 0.4066, 0.0644, 0.0039, 0.4076, 0.0635, 0.0017, 0.409, 0.0635, 0.0039, 0.409, 0.0623, -0.0, 0.4033, 0.0619, -0.0, 0.4045, 0.0623, 0.0017, 0.4033, 0.0623, 0.0039, 0.4033, 0.0619, 0.0017, 0.4046, 0.0619, 0.0039, 0.4046,
0.0608, -0.0, 0.4062, 0.0596, -0.0, 0.4073, 0.0623, -0.0, 0.41, 0.0587, -0.0, 0.4079, 0.0569, -0.0, 0.4085, 0.0548, -0.0, 0.4087, 0.0608, 0.0017, 0.4062, 0.0595, 0.0017, 0.4074, 0.0608, 0.0039, 0.4062, 0.0595, 0.0039, 0.4074, 0.0624, 0.0017, 0.4099, 0.0584, 0.0017, 0.408, 0.0624, 0.0039, 0.4099, 0.0584, 0.0039, 0.408, 0.057, 0.0017, 0.4085, 0.0548, 0.0017, 0.4087, 0.057, 0.0039, 0.4085, 0.0548, 0.0039, 0.4087, 0.0612, -0.0, 0.4108, 0.0593, -0.0, 0.4119, 0.0582, -0.0, 0.4122, 0.057, -0.0, 0.411, 0.0548, -0.0, 0.411, 0.057, -0.0, 0.4125, 0.0548, -0.0, 0.4127, 0.0613, 0.0017, 0.4107, 0.0594, 0.0017, 0.4119, 0.0613, 0.0039, 0.4107, 0.0594, 0.0039, 0.4119, 0.0582, 0.0017, 0.4122,
0.0582, 0.0039, 0.4122, 0.057, 0.0017, 0.4125, 0.0548, 0.0017, 0.4127, 0.057, 0.0039, 0.4125, 0.0548, 0.0039, 0.4127, 0.0462, -0.0, 0.4022, 0.0476, -0.0, 0.4051, 0.0471, -0.0, 0.4041, 0.0463, 0.0017, 0.4024, 0.0463, 0.0039, 0.4024, 0.0476, 0.0017, 0.405, 0.047, 0.0017, 0.4039, 0.0476, 0.0039, 0.405, 0.047, 0.0039, 0.4039, 0.0498, -0.0, 0.4074, 0.0528, -0.0, 0.4086, 0.0508, -0.0, 0.408, 0.0486, -0.0, 0.4064, 0.0461, -0.0, 0.4067, 0.0483, -0.0, 0.4089, 0.0456, -0.0, 0.4093, 0.0498, 0.0017, 0.4075, 0.0498, 0.0039, 0.4075, 0.0526, 0.0017, 0.4086, 0.0509, 0.0017, 0.4081, 0.0526, 0.0039, 0.4086, 0.0509, 0.0039, 0.4081, 0.0486, 0.0017, 0.4064, 0.0486, 0.0039, 0.4064, 0.0456, 0.0017, 0.4093,
0.0456, 0.0039, 0.4093, 0.0439, -0.0, 0.4024, 0.0426, -0.0, 0.4024, 0.043, -0.0, 0.4047, 0.0428, -0.0, 0.4036, 0.0426, 0.0017, 0.4024, 0.0426, 0.0039, 0.4024, 0.0431, 0.0017, 0.4048, 0.0428, 0.0017, 0.4037, 0.0431, 0.0039, 0.4048, 0.0428, 0.0039, 0.4037, 0.044, -0.0, 0.407, 0.0447, -0.0, 0.4081, 0.0439, 0.0017, 0.4067, 0.0439, 0.0039, 0.4067, 0.0448, 0.0017, 0.4082, 0.0448, 0.0039, 0.4082, 0.0526, -0.0, 0.411, 0.0502, -0.0, 0.4121, 0.0526, -0.0, 0.4126, 0.0511, -0.0, 0.4123, 0.048, -0.0, 0.4111, 0.0468, -0.0, 0.4103, 0.0499, 0.0017, 0.412, 0.0499, 0.0039, 0.412, 0.0526, 0.0017, 0.4126, 0.051, 0.0017, 0.4122, 0.0526, 0.0039, 0.4126, 0.051, 0.0039, 0.4122, 0.0482, 0.0017, 0.4112,
0.0468, 0.0017, 0.4103, 0.0482, 0.0039, 0.4112, 0.0468, 0.0039, 0.4103, 0.0213, -0.0, 0.3844, 0.0212, -0.0, 0.3852, 0.0209, -0.0, 0.3873, 0.0213, 0.0017, 0.3846, 0.0211, 0.0017, 0.3857, 0.0213, 0.0039, 0.3846, 0.0211, 0.0039, 0.3857, 0.0209, 0.0017, 0.3872, 0.0209, 0.0039, 0.3872, 0.0206, -0.0, 0.3895, 0.02, -0.0, 0.3917, 0.0206, 0.0017, 0.3893, 0.0206, 0.0039, 0.3893, 0.0201, 0.0017, 0.3915, 0.0201, 0.0039, 0.3915, 0.0194, -0.0, 0.3933, 0.0194, 0.0017, 0.3935, 0.0194, 0.0039, 0.3935, 0.0179, -0.0, 0.385, 0.0157, -0.0, 0.385, 0.0179, -0.0, 0.3872, 0.0157, -0.0, 0.3872, 0.0136, -0.0, 0.385, 0.0114, -0.0, 0.385, 0.0136, -0.0, 0.3872, 0.0114, -0.0, 0.3872, 0.0179, -0.0, 0.3893,
0.0157, -0.0, 0.3893, 0.0179, -0.0, 0.3915, 0.0157, -0.0, 0.3915, 0.0136, -0.0, 0.3893, 0.0114, -0.0, 0.3893, 0.0136, -0.0, 0.3915, 0.0114, -0.0, 0.3915, 0.0096, -0.0, 0.3849, 0.0091, -0.0, 0.387, 0.0096, 0.0017, 0.385, 0.0096, 0.0039, 0.385, 0.009, 0.0017, 0.3872, 0.009, 0.0039, 0.3872, 0.0084, -0.0, 0.3892, 0.008, -0.0, 0.3902, 0.0092, -0.0, 0.3915, 0.0076, -0.0, 0.3914, 0.0084, 0.0017, 0.3891, 0.0081, 0.0017, 0.3902, 0.0084, 0.0039, 0.3891, 0.0081, 0.0039, 0.3902, 0.0075, 0.0017, 0.3915, 0.0075, 0.0039, 0.3915, 0.0191, -0.0, 0.3942, 0.0157, -0.0, 0.3937, 0.0183, -0.0, 0.3961, 0.0157, -0.0, 0.3958, 0.0136, -0.0, 0.3937, 0.0114, -0.0, 0.3937, 0.0136, -0.0, 0.3958,
0.0114, -0.0, 0.3958, 0.0189, 0.0017, 0.3946, 0.0189, 0.0039, 0.3946, 0.0185, 0.0017, 0.3959, 0.0185, 0.0039, 0.3959, 0.0174, -0.0, 0.3982, 0.0157, -0.0, 0.398, 0.0169, -0.0, 0.3991, 0.0162, -0.0, 0.4003, 0.0136, -0.0, 0.398, 0.0114, -0.0, 0.398, 0.0136, -0.0, 0.4002, 0.0114, -0.0, 0.4002, 0.0175, 0.0017, 0.398, 0.0175, 0.0039, 0.398, 0.0169, 0.0017, 0.3992, 0.0162, 0.0017, 0.4003, 0.0169, 0.0039, 0.3992, 0.0162, 0.0039, 0.4003, 0.0092, -0.0, 0.3937, 0.0065, -0.0, 0.3935, 0.0092, -0.0, 0.3958, 0.0071, -0.0, 0.3958, 0.0058, -0.0, 0.3945, 0.0051, -0.0, 0.3956, 0.0065, 0.0017, 0.3935, 0.0065, 0.0039, 0.3935, 0.0058, 0.0017, 0.3945, 0.0058, 0.0039, 0.3945, 0.0049, 0.0017, 0.3959,
0.0049, 0.0039, 0.3959, 0.0092, -0.0, 0.398, 0.0071, -0.0, 0.398, 0.0092, -0.0, 0.4002, 0.0071, -0.0, 0.4002, 0.0038, -0.0, 0.3971, 0.0028, -0.0, 0.3981, 0.0049, -0.0, 0.4002, 0.0017, -0.0, 0.399, 0.0039, 0.0017, 0.397, 0.0028, 0.0017, 0.3982, 0.0039, 0.0039, 0.397, 0.0028, 0.0039, 0.3982, 0.0016, 0.0017, 0.3991, 0.0016, 0.0039, 0.3991, 0.0151, -0.0, 0.4018, 0.0143, -0.0, 0.4029, 0.0114, -0.0, 0.4024, 0.0132, -0.0, 0.4042, 0.0122, -0.0, 0.4053, 0.0151, 0.0017, 0.4018, 0.0151, 0.0039, 0.4018, 0.0142, 0.0017, 0.4029, 0.0142, 0.0039, 0.4029, 0.0132, 0.0017, 0.4042, 0.0122, 0.0017, 0.4053, 0.0132, 0.0039, 0.4042, 0.0122, 0.0039, 0.4053, 0.0111, -0.0, 0.4064, 0.0111, 0.0017, 0.4064,
0.0111, 0.0039, 0.4064, 0.0092, -0.0, 0.4024, 0.0071, -0.0, 0.4024, 0.0092, -0.0, 0.4045, 0.0071, -0.0, 0.4045, 0.0049, -0.0, 0.4024, 0.0027, -0.0, 0.4024, 0.0049, -0.0, 0.4045, 0.0027, -0.0, 0.4045, 0.0099, -0.0, 0.4075, 0.0071, -0.0, 0.4067, 0.0088, -0.0, 0.4084, 0.0074, -0.0, 0.4095, 0.0049, -0.0, 0.4067, 0.0027, -0.0, 0.4067, 0.0049, -0.0, 0.4089, 0.0027, -0.0, 0.4089, 0.0099, 0.0017, 0.4075, 0.0099, 0.0039, 0.4075, 0.0089, 0.0017, 0.4084, 0.0074, 0.0017, 0.4094, 0.0089, 0.0039, 0.4084, 0.0074, 0.0039, 0.4094, 0.0063, -0.0, 0.4102, 0.0051, -0.0, 0.4109, 0.0032, -0.0, 0.412, 0.0021, -0.0, 0.4125, 0.0063, 0.0017, 0.4102, 0.0063, 0.0039, 0.4102, 0.0049, 0.0017, 0.4111,
0.0034, 0.0017, 0.4119, 0.0049, 0.0039, 0.4111, 0.0034, 0.0039, 0.4119, 0.0023, 0.0017, 0.4124, 0.0023, 0.0039, 0.4124, 0.067, -0.0, 0.4324, 0.067, -0.0, 0.4349, 0.0656, -0.0, 0.4324, 0.0635, -0.0, 0.4324, 0.0656, -0.0, 0.4349, 0.0631, -0.0, 0.4353, 0.067, 0.0017, 0.4324, 0.067, 0.0039, 0.4324, 0.067, 0.0017, 0.4349, 0.067, 0.0039, 0.4349, 0.0656, 0.0017, 0.4324, 0.0635, 0.0017, 0.4324, 0.0656, 0.0039, 0.4324, 0.0635, 0.0039, 0.4324, 0.0631, 0.0017, 0.4353, 0.0631, 0.0039, 0.4353, 0.0613, -0.0, 0.4324, 0.0591, -0.0, 0.4324, 0.0613, -0.0, 0.4353, 0.0591, -0.0, 0.4353, 0.057, -0.0, 0.4324, 0.0548, -0.0, 0.4324, 0.057, -0.0, 0.4353, 0.0548, -0.0, 0.4353, 0.0613, 0.0017, 0.4324,
0.0591, 0.0017, 0.4324, 0.0613, 0.0039, 0.4324, 0.0591, 0.0039, 0.4324, 0.0613, 0.0017, 0.4353, 0.0591, 0.0017, 0.4353, 0.0613, 0.0039, 0.4353, 0.0591, 0.0039, 0.4353, 0.057, 0.0017, 0.4324, 0.0548, 0.0017, 0.4324, 0.057, 0.0039, 0.4324, 0.0548, 0.0039, 0.4324, 0.057, 0.0017, 0.4353, 0.0548, 0.0017, 0.4353, 0.057, 0.0039, 0.4353, 0.0548, 0.0039, 0.4353, 0.0526, -0.0, 0.4324, 0.0505, -0.0, 0.4324, 0.0527, -0.0, 0.4353, 0.0506, -0.0, 0.4353, 0.0483, -0.0, 0.4324, 0.0461, -0.0, 0.4324, 0.0487, -0.0, 0.4359, 0.0461, -0.0, 0.4349, 0.0526, 0.0017, 0.4324, 0.0505, 0.0017, 0.4324, 0.0526, 0.0039, 0.4324, 0.0505, 0.0039, 0.4324, 0.0526, 0.0017, 0.4353, 0.0497, 0.0017, 0.4355, 0.0526, 0.0039, 0.4353,
0.0497, 0.0039, 0.4355, 0.0483, 0.0017, 0.4324, 0.0461, 0.0017, 0.4324, 0.0483, 0.0039, 0.4324, 0.0461, 0.0039, 0.4324, 0.0489, 0.0017, 0.4358, 0.0489, 0.0039, 0.4358, 0.0439, -0.0, 0.4324, 0.0425, -0.0, 0.4324, 0.0439, -0.0, 0.4349, 0.0425, -0.0, 0.4349, 0.0439, 0.0017, 0.4324, 0.0425, 0.0017, 0.4324, 0.0439, 0.0039, 0.4324, 0.0425, 0.0039, 0.4324, 0.0425, 0.0017, 0.4349, 0.0425, 0.0039, 0.4349, 0.067, -0.0, 0.4371, 0.067, -0.0, 0.4392, 0.0656, -0.0, 0.4371, 0.0631, -0.0, 0.4371, 0.0656, -0.0, 0.4392, 0.0631, -0.0, 0.4392, 0.067, 0.0017, 0.4371, 0.067, 0.0039, 0.4371, 0.067, 0.0017, 0.4392, 0.067, 0.0039, 0.4392, 0.0631, 0.0017, 0.4371, 0.0631, 0.0039, 0.4371, 0.0631, 0.0017, 0.4392,
0.0631, 0.0039, 0.4392, 0.067, -0.0, 0.4414, 0.067, -0.0, 0.4436, 0.0656, -0.0, 0.4414, 0.0631, -0.0, 0.4414, 0.0656, -0.0, 0.4436, 0.0631, -0.0, 0.4436, 0.067, 0.0017, 0.4414, 0.067, 0.0039, 0.4414, 0.067, 0.0017, 0.4436, 0.067, 0.0039, 0.4436, 0.0631, 0.0017, 0.4414, 0.0631, 0.0039, 0.4414, 0.0631, 0.0017, 0.4436, 0.0631, 0.0039, 0.4436, 0.0562, -0.0, 0.4382, 0.0548, -0.0, 0.4382, 0.0562, 0.0017, 0.4382, 0.0548, 0.0017, 0.4382, 0.0562, 0.0039, 0.4382, 0.0548, 0.0039, 0.4382, 0.0562, -0.0, 0.4414, 0.0548, -0.0, 0.4414, 0.0562, -0.0, 0.4436, 0.0548, -0.0, 0.4436, 0.0562, 0.0017, 0.4414, 0.0562, 0.0039, 0.4414, 0.0562, 0.0017, 0.4436, 0.0562, 0.0039, 0.4436, 0.067, -0.0, 0.4458,
0.067, -0.0, 0.4479, 0.0656, -0.0, 0.4458, 0.0631, -0.0, 0.4458, 0.0656, -0.0, 0.4479, 0.0631, -0.0, 0.4479, 0.067, 0.0017, 0.4458, 0.067, 0.0039, 0.4458, 0.067, 0.0017, 0.4479, 0.067, 0.0039, 0.4479, 0.0631, 0.0017, 0.4458, 0.0631, 0.0039, 0.4458, 0.0631, 0.0017, 0.4479, 0.0631, 0.0039, 0.4479, 0.067, -0.0, 0.4501, 0.067, -0.0, 0.4529, 0.0656, -0.0, 0.4501, 0.0631, -0.0, 0.4501, 0.0656, -0.0, 0.4529, 0.0631, -0.0, 0.4529, 0.067, 0.0017, 0.4501, 0.067, 0.0039, 0.4501, 0.067, 0.0017, 0.4529, 0.067, 0.0039, 0.4529, 0.0631, 0.0017, 0.4501, 0.0631, 0.0039, 0.4501, 0.0656, 0.0017, 0.4529, 0.0631, 0.0017, 0.4529, 0.0656, 0.0039, 0.4529, 0.0631, 0.0039, 0.4529, 0.0562, -0.0, 0.4458,
0.0548, -0.0, 0.4458, 0.0562, -0.0, 0.4479, 0.0548, -0.0, 0.4479, 0.0562, 0.0017, 0.4458, 0.0562, 0.0039, 0.4458, 0.0562, 0.0017, 0.4479, 0.0562, 0.0039, 0.4479, 0.0562, -0.0, 0.4501, 0.0548, -0.0, 0.4501, 0.0562, -0.0, 0.452, 0.0548, -0.0, 0.452, 0.0562, 0.0017, 0.4501, 0.0562, 0.0039, 0.4501, 0.0562, 0.0017, 0.452, 0.0548, 0.0017, 0.452, 0.0562, 0.0039, 0.452, 0.0548, 0.0039, 0.452, 0.0523, -0.0, 0.4382, 0.0477, -0.0, 0.4363, 0.0466, -0.0, 0.4372, 0.0457, -0.0, 0.4393, 0.0523, 0.0017, 0.4382, 0.0523, 0.0039, 0.4382, 0.0478, 0.0017, 0.4362, 0.0466, 0.0017, 0.4372, 0.0478, 0.0039, 0.4362, 0.0466, 0.0039, 0.4372, 0.0456, 0.0017, 0.4396, 0.0456, 0.0039, 0.4396, 0.0523, -0.0, 0.4414,
0.0523, -0.0, 0.4436, 0.0455, -0.0, 0.4414, 0.0454, -0.0, 0.4436, 0.0523, 0.0017, 0.4414, 0.0523, 0.0039, 0.4414, 0.0523, 0.0017, 0.4436, 0.0523, 0.0039, 0.4436, 0.0455, 0.0017, 0.4414, 0.0455, 0.0039, 0.4414, 0.0454, 0.0017, 0.4436, 0.0454, 0.0039, 0.4436, 0.0439, -0.0, 0.4371, 0.0425, -0.0, 0.4371, 0.0439, -0.0, 0.4392, 0.0425, -0.0, 0.4392, 0.0425, 0.0017, 0.4371, 0.0425, 0.0039, 0.4371, 0.0425, 0.0017, 0.4392, 0.0425, 0.0039, 0.4392, 0.0439, -0.0, 0.4414, 0.0425, -0.0, 0.4414, 0.0439, -0.0, 0.4436, 0.0425, -0.0, 0.4436, 0.0425, 0.0017, 0.4414, 0.0425, 0.0039, 0.4414, 0.0425, 0.0017, 0.4436, 0.0425, 0.0039, 0.4436, 0.0523, -0.0, 0.4458, 0.0523, -0.0, 0.4479, 0.0454, -0.0, 0.4458,
0.0454, -0.0, 0.4479, 0.0523, 0.0017, 0.4458, 0.0523, 0.0039, 0.4458, 0.0523, 0.0017, 0.4479, 0.0523, 0.0039, 0.4479, 0.0454, 0.0017, 0.4458, 0.0454, 0.0039, 0.4458, 0.0454, 0.0017, 0.4479, 0.0454, 0.0039, 0.4479, 0.0523, -0.0, 0.4501, 0.0523, -0.0, 0.452, 0.0454, -0.0, 0.4501, 0.0454, -0.0, 0.4529, 0.0523, 0.0017, 0.4501, 0.0523, 0.0039, 0.4501, 0.0523, 0.0017, 0.452, 0.0523, 0.0039, 0.452, 0.0454, 0.0017, 0.4501, 0.0454, 0.0039, 0.4501, 0.0454, 0.0017, 0.4529, 0.0454, 0.0039, 0.4529, 0.0439, -0.0, 0.4458, 0.0425, -0.0, 0.4458, 0.0439, -0.0, 0.4479, 0.0425, -0.0, 0.4479, 0.0425, 0.0017, 0.4458, 0.0425, 0.0039, 0.4458, 0.0425, 0.0017, 0.4479, 0.0425, 0.0039, 0.4479, 0.0439, -0.0, 0.4501,
0.0425, -0.0, 0.4501, 0.0439, -0.0, 0.4529, 0.0425, -0.0, 0.4529, 0.0425, 0.0017, 0.4501, 0.0425, 0.0039, 0.4501, 0.0439, 0.0017, 0.4529, 0.0425, 0.0017, 0.4529, 0.0439, 0.0039, 0.4529, 0.0425, 0.0039, 0.4529, 0.0219, -0.0, 0.4245, 0.0201, -0.0, 0.4245, 0.0219, -0.0, 0.4262, 0.0201, -0.0, 0.4262, 0.0219, 0.0017, 0.4245, 0.0201, 0.0017, 0.4245, 0.0219, 0.0039, 0.4245, 0.0201, 0.0039, 0.4245, 0.0219, 0.0017, 0.4262, 0.0219, 0.0039, 0.4262, 0.0219, -0.0, 0.4284, 0.0201, -0.0, 0.4284, 0.0219, -0.0, 0.4306, 0.0201, -0.0, 0.4306, 0.0219, 0.0017, 0.4284, 0.0219, 0.0039, 0.4284, 0.0219, 0.0017, 0.4306, 0.0219, 0.0039, 0.4306, 0.0219, -0.0, 0.4327, 0.0201, -0.0, 0.4327, 0.0219, -0.0, 0.4349,
0.0201, -0.0, 0.4349, 0.0219, 0.0017, 0.4327, 0.0219, 0.0039, 0.4327, 0.0219, 0.0017, 0.4349, 0.0219, 0.0039, 0.4349, 0.0179, -0.0, 0.4245, 0.0157, -0.0, 0.4245, 0.0179, -0.0, 0.4262, 0.0157, -0.0, 0.4262, 0.0136, -0.0, 0.4245, 0.0114, -0.0, 0.4245, 0.0136, -0.0, 0.4262, 0.0114, -0.0, 0.4262, 0.0179, 0.0017, 0.4245, 0.0157, 0.0017, 0.4245, 0.0179, 0.0039, 0.4245, 0.0157, 0.0039, 0.4245, 0.0136, 0.0017, 0.4245, 0.0114, 0.0017, 0.4245, 0.0136, 0.0039, 0.4245, 0.0114, 0.0039, 0.4245, 0.0101, -0.0, 0.4245, 0.0101, -0.0, 0.4262, 0.0101, 0.0017, 0.4245, 0.0101, 0.0039, 0.4245, 0.0101, 0.0017, 0.4262, 0.0101, 0.0039, 0.4262, 0.0179, -0.0, 0.4284, 0.0157, -0.0, 0.4284, 0.0179, -0.0, 0.4306,
0.0157, -0.0, 0.4306, 0.0136, -0.0, 0.4284, 0.0114, -0.0, 0.4284, 0.0136, -0.0, 0.4306, 0.0114, -0.0, 0.4306, 0.0179, -0.0, 0.4327, 0.0157, -0.0, 0.4327, 0.0179, -0.0, 0.4349, 0.0157, -0.0, 0.4349, 0.0136, -0.0, 0.4327, 0.0114, -0.0, 0.4327, 0.0136, -0.0, 0.4349, 0.0114, -0.0, 0.4349, 0.0101, -0.0, 0.4284, 0.0101, -0.0, 0.4306, 0.0101, 0.0017, 0.4284, 0.0101, 0.0039, 0.4284, 0.0101, 0.0017, 0.4306, 0.0101, 0.0039, 0.4306, 0.0101, -0.0, 0.4327, 0.0101, -0.0, 0.4349, 0.0101, 0.0017, 0.4327, 0.0101, 0.0039, 0.4327, 0.0101, 0.0017, 0.4349, 0.0101, 0.0039, 0.4349, 0.0219, -0.0, 0.4371, 0.0201, -0.0, 0.4371, 0.0219, -0.0, 0.4392, 0.0201, -0.0, 0.4392, 0.0219, 0.0017, 0.4371,
0.0219, 0.0039, 0.4371, 0.0219, 0.0017, 0.4392, 0.0219, 0.0039, 0.4392, 0.0219, -0.0, 0.4414, 0.0201, -0.0, 0.4414, 0.0219, -0.0, 0.4436, 0.0201, -0.0, 0.4436, 0.0219, 0.0017, 0.4414, 0.0219, 0.0039, 0.4414, 0.0219, 0.0017, 0.4436, 0.0219, 0.0039, 0.4436, 0.0219, -0.0, 0.4458, 0.0201, -0.0, 0.4458, 0.0219, -0.0, 0.4479, 0.0201, -0.0, 0.4479, 0.0219, 0.0017, 0.4458, 0.0219, 0.0039, 0.4458, 0.0219, 0.0017, 0.4479, 0.0219, 0.0039, 0.4479, 0.0219, -0.0, 0.4501, 0.0201, -0.0, 0.4501, 0.0219, -0.0, 0.4523, 0.0201, -0.0, 0.4523, 0.0219, 0.0017, 0.4501, 0.0219, 0.0039, 0.4501, 0.0219, 0.0017, 0.4523, 0.0219, 0.0039, 0.4523, 0.0179, -0.0, 0.4371, 0.0157, -0.0, 0.4371, 0.0179, -0.0, 0.4392,
0.0157, -0.0, 0.4392, 0.0136, -0.0, 0.4371, 0.0114, -0.0, 0.4371, 0.0136, -0.0, 0.4392, 0.0114, -0.0, 0.4392, 0.0179, -0.0, 0.4414, 0.0157, -0.0, 0.4414, 0.0179, -0.0, 0.4436, 0.0157, -0.0, 0.4436, 0.0136, -0.0, 0.4414, 0.0114, -0.0, 0.4414, 0.0136, -0.0, 0.4436, 0.0114, -0.0, 0.4436, 0.0101, -0.0, 0.4371, 0.0101, -0.0, 0.4392, 0.0101, 0.0017, 0.4371, 0.0101, 0.0039, 0.4371, 0.0101, 0.0017, 0.4392, 0.0101, 0.0039, 0.4392, 0.0101, -0.0, 0.4414, 0.0101, -0.0, 0.4436, 0.0101, 0.0017, 0.4414, 0.0101, 0.0039, 0.4414, 0.0101, 0.0017, 0.4436, 0.0101, 0.0039, 0.4436, 0.0179, -0.0, 0.4458, 0.0157, -0.0, 0.4458, 0.0179, -0.0, 0.4479, 0.0157, -0.0, 0.4479, 0.0136, -0.0, 0.4458,
0.0114, -0.0, 0.4458, 0.0136, -0.0, 0.4479, 0.0114, -0.0, 0.4479, 0.0179, -0.0, 0.4501, 0.0157, -0.0, 0.4501, 0.0179, -0.0, 0.4523, 0.0157, -0.0, 0.4523, 0.0136, -0.0, 0.4501, 0.0114, -0.0, 0.4501, 0.0136, -0.0, 0.4523, 0.0114, -0.0, 0.4523, 0.0101, -0.0, 0.4458, 0.0101, -0.0, 0.4479, 0.0101, 0.0017, 0.4458, 0.0101, 0.0039, 0.4458, 0.0101, 0.0017, 0.4479, 0.0101, 0.0039, 0.4479, 0.0101, -0.0, 0.4501, 0.0101, -0.0, 0.4523, 0.0101, 0.0017, 0.4501, 0.0101, 0.0039, 0.4501, 0.0101, 0.0017, 0.4523, 0.0101, 0.0039, 0.4523, 0.067, -0.0, 0.4725, 0.067, -0.0, 0.474, 0.0656, -0.0, 0.4725, 0.0635, -0.0, 0.4725, 0.0656, -0.0, 0.474, 0.0635, -0.0, 0.474, 0.067, 0.0017, 0.4725,
0.067, 0.0039, 0.4725, 0.067, 0.0017, 0.474, 0.067, 0.0039, 0.474, 0.0656, 0.0017, 0.4725, 0.0635, 0.0017, 0.4725, 0.0656, 0.0039, 0.4725, 0.0635, 0.0039, 0.4725, 0.067, -0.0, 0.4765, 0.0656, -0.0, 0.4765, 0.0635, -0.0, 0.4765, 0.067, 0.0017, 0.4765, 0.067, 0.0039, 0.4765, 0.0656, 0.0017, 0.4765, 0.0635, 0.0017, 0.4765, 0.0656, 0.0039, 0.4765, 0.0635, 0.0039, 0.4765, 0.0613, -0.0, 0.4725, 0.0591, -0.0, 0.4725, 0.0613, -0.0, 0.474, 0.0591, -0.0, 0.474, 0.057, -0.0, 0.4725, 0.0548, -0.0, 0.4725, 0.057, -0.0, 0.474, 0.0548, -0.0, 0.474, 0.0613, 0.0017, 0.4725, 0.0591, 0.0017, 0.4725, 0.0613, 0.0039, 0.4725, 0.0591, 0.0039, 0.4725, 0.057, 0.0017, 0.4725, 0.0548, 0.0017, 0.4725,
0.057, 0.0039, 0.4725, 0.0548, 0.0039, 0.4725, 0.0613, -0.0, 0.4765, 0.0591, -0.0, 0.4765, 0.057, -0.0, 0.4765, 0.0548, -0.0, 0.4765, 0.0613, 0.0017, 0.4765, 0.0591, 0.0017, 0.4765, 0.0613, 0.0039, 0.4765, 0.0591, 0.0039, 0.4765, 0.057, 0.0017, 0.4765, 0.0548, 0.0017, 0.4765, 0.057, 0.0039, 0.4765, 0.0548, 0.0039, 0.4765, 0.0591, -0.0, 0.4804, 0.0591, -0.0, 0.4826, 0.0562, -0.0, 0.4804, 0.0562, -0.0, 0.4826, 0.0591, 0.0017, 0.4804, 0.0591, 0.0039, 0.4804, 0.0591, 0.0017, 0.4826, 0.0591, 0.0039, 0.4826, 0.0562, 0.0017, 0.4804, 0.0562, 0.0039, 0.4804, 0.0562, 0.0017, 0.4826, 0.0562, 0.0039, 0.4826, 0.0592, -0.0, 0.4848, 0.0596, -0.0, 0.4868, 0.0562, -0.0, 0.4848, 0.0562, -0.0, 0.487,
0.0592, 0.0017, 0.4848, 0.0592, 0.0039, 0.4848, 0.0596, 0.0017, 0.487, 0.0596, 0.0039, 0.487, 0.0562, 0.0017, 0.4848, 0.0562, 0.0039, 0.4848, 0.0562, 0.0017, 0.487, 0.0562, 0.0039, 0.487, 0.0526, -0.0, 0.4725, 0.0505, -0.0, 0.4725, 0.0526, -0.0, 0.474, 0.0505, -0.0, 0.474, 0.0483, -0.0, 0.4725, 0.0461, -0.0, 0.4725, 0.0483, -0.0, 0.474, 0.0461, -0.0, 0.474, 0.0526, 0.0017, 0.4725, 0.0505, 0.0017, 0.4725, 0.0526, 0.0039, 0.4725, 0.0505, 0.0039, 0.4725, 0.0483, 0.0017, 0.4725, 0.0461, 0.0017, 0.4725, 0.0483, 0.0039, 0.4725, 0.0461, 0.0039, 0.4725, 0.0527, -0.0, 0.4765, 0.0506, -0.0, 0.4766, 0.0485, -0.0, 0.477, 0.0461, -0.0, 0.4761, 0.0476, -0.0, 0.4775, 0.0465, -0.0, 0.4786,
0.0526, 0.0017, 0.4765, 0.0504, 0.0017, 0.4766, 0.0526, 0.0039, 0.4765, 0.0504, 0.0039, 0.4766, 0.0487, 0.0017, 0.4769, 0.0487, 0.0039, 0.4769, 0.0476, 0.0017, 0.4775, 0.047, 0.0017, 0.478, 0.0476, 0.0039, 0.4775, 0.047, 0.0039, 0.478, 0.0439, -0.0, 0.4725, 0.0425, -0.0, 0.4725, 0.0439, -0.0, 0.474, 0.0425, -0.0, 0.474, 0.0439, 0.0017, 0.4725, 0.0425, 0.0017, 0.4725, 0.0439, 0.0039, 0.4725, 0.0425, 0.0039, 0.4725, 0.0425, 0.0017, 0.474, 0.0425, 0.0039, 0.474, 0.0439, -0.0, 0.4761, 0.0425, -0.0, 0.4761, 0.0439, -0.0, 0.4783, 0.0425, -0.0, 0.4783, 0.0425, 0.0017, 0.4761, 0.0425, 0.0039, 0.4761, 0.0425, 0.0017, 0.4783, 0.0425, 0.0039, 0.4783, 0.0456, -0.0, 0.4801, 0.0455, -0.0, 0.4827,
0.0456, 0.0017, 0.4801, 0.0456, 0.0039, 0.4801, 0.0455, 0.0017, 0.4826, 0.0455, 0.0039, 0.4826, 0.0454, -0.0, 0.4848, 0.0454, -0.0, 0.487, 0.0454, 0.0017, 0.4848, 0.0454, 0.0039, 0.4848, 0.0454, 0.0017, 0.487, 0.0454, 0.0039, 0.487, 0.0439, -0.0, 0.4805, 0.0425, -0.0, 0.4805, 0.0439, -0.0, 0.4826, 0.0425, -0.0, 0.4826, 0.0425, 0.0017, 0.4805, 0.0425, 0.0039, 0.4805, 0.0425, 0.0017, 0.4826, 0.0425, 0.0039, 0.4826, 0.0439, -0.0, 0.4848, 0.0425, -0.0, 0.4848, 0.0439, -0.0, 0.487, 0.0425, -0.0, 0.487, 0.0425, 0.0017, 0.4848, 0.0425, 0.0039, 0.4848, 0.0425, 0.0017, 0.487, 0.0425, 0.0039, 0.487, 0.0219, -0.0, 0.4544, 0.0201, -0.0, 0.4544, 0.0219, -0.0, 0.4566, 0.0201, -0.0, 0.4566,
0.0219, 0.0017, 0.4544, 0.0219, 0.0039, 0.4544, 0.0219, 0.0017, 0.4566, 0.0219, 0.0039, 0.4566, 0.0219, -0.0, 0.4588, 0.0201, -0.0, 0.4588, 0.0219, -0.0, 0.4609, 0.0201, -0.0, 0.4609, 0.0219, 0.0017, 0.4588, 0.0219, 0.0039, 0.4588, 0.0219, 0.0017, 0.4609, 0.0219, 0.0039, 0.4609, 0.0219, -0.0, 0.4631, 0.0201, -0.0, 0.4631, 0.0219, -0.0, 0.4653, 0.0201, -0.0, 0.4653, 0.0219, 0.0017, 0.4631, 0.0219, 0.0039, 0.4631, 0.0219, 0.0017, 0.4653, 0.0219, 0.0039, 0.4653, 0.0219, -0.0, 0.4675, 0.0201, -0.0, 0.4675, 0.0219, -0.0, 0.4696, 0.0201, -0.0, 0.4696, 0.0219, 0.0017, 0.4675, 0.0219, 0.0039, 0.4675, 0.0219, 0.0017, 0.4696, 0.0219, 0.0039, 0.4696, 0.0179, -0.0, 0.4544, 0.0157, -0.0, 0.4544,
0.0179, -0.0, 0.4566, 0.0157, -0.0, 0.4566, 0.0136, -0.0, 0.4544, 0.0114, -0.0, 0.4544, 0.0136, -0.0, 0.4566, 0.0114, -0.0, 0.4566, 0.0179, -0.0, 0.4588, 0.0157, -0.0, 0.4588, 0.0179, -0.0, 0.4609, 0.0157, -0.0, 0.4609, 0.0136, -0.0, 0.4588, 0.0114, -0.0, 0.4588, 0.0136, -0.0, 0.4609, 0.0114, -0.0, 0.4609, 0.0101, -0.0, 0.4544, 0.0101, -0.0, 0.4566, 0.0101, 0.0017, 0.4544, 0.0101, 0.0039, 0.4544, 0.0101, 0.0017, 0.4566, 0.0101, 0.0039, 0.4566, 0.0101, -0.0, 0.4588, 0.0101, -0.0, 0.4609, 0.0101, 0.0017, 0.4588, 0.0101, 0.0039, 0.4588, 0.0101, 0.0017, 0.4609, 0.0101, 0.0039, 0.4609, 0.0179, -0.0, 0.4631, 0.0157, -0.0, 0.4631, 0.0179, -0.0, 0.4653, 0.0157, -0.0, 0.4653,
0.0136, -0.0, 0.4631, 0.0114, -0.0, 0.4631, 0.0136, -0.0, 0.4653, 0.0114, -0.0, 0.4653, 0.0179, -0.0, 0.4675, 0.0157, -0.0, 0.4675, 0.0179, -0.0, 0.4696, 0.0157, -0.0, 0.4696, 0.0136, -0.0, 0.4675, 0.0114, -0.0, 0.4675, 0.0136, -0.0, 0.4696, 0.0114, -0.0, 0.4696, 0.0101, -0.0, 0.4631, 0.0101, -0.0, 0.4653, 0.0101, 0.0017, 0.4631, 0.0101, 0.0039, 0.4631, 0.0101, 0.0017, 0.4653, 0.0101, 0.0039, 0.4653, 0.0101, -0.0, 0.4675, 0.0101, -0.0, 0.4696, 0.0101, 0.0017, 0.4675, 0.0101, 0.0039, 0.4675, 0.0101, 0.0017, 0.4696, 0.0101, 0.0039, 0.4696, 0.0219, -0.0, 0.4718, 0.0201, -0.0, 0.4718, 0.0219, -0.0, 0.474, 0.0201, -0.0, 0.474, 0.0219, 0.0017, 0.4718, 0.0219, 0.0039, 0.4718,
0.0219, 0.0017, 0.474, 0.0219, 0.0039, 0.474, 0.0219, -0.0, 0.4761, 0.0201, -0.0, 0.4761, 0.0218, -0.0, 0.4783, 0.0201, -0.0, 0.4783, 0.0219, 0.0017, 0.4761, 0.0219, 0.0039, 0.4761, 0.0218, 0.0017, 0.4783, 0.0218, 0.0039, 0.4783, 0.0216, -0.0, 0.4805, 0.0201, -0.0, 0.4805, 0.0213, -0.0, 0.4821, 0.0212, -0.0, 0.4829, 0.0216, 0.0017, 0.4805, 0.0216, 0.0039, 0.4805, 0.0213, 0.0017, 0.4823, 0.0211, 0.0017, 0.4834, 0.0213, 0.0039, 0.4823, 0.0211, 0.0039, 0.4834, 0.0207, -0.0, 0.485, 0.0202, -0.0, 0.4872, 0.0208, 0.0017, 0.4848, 0.0208, 0.0039, 0.4848, 0.0203, 0.0017, 0.487, 0.0203, 0.0039, 0.487, 0.0179, -0.0, 0.4718, 0.0157, -0.0, 0.4718, 0.0179, -0.0, 0.474, 0.0157, -0.0, 0.474,
0.0136, -0.0, 0.4718, 0.0114, -0.0, 0.4718, 0.0136, -0.0, 0.474, 0.0114, -0.0, 0.474, 0.0179, -0.0, 0.4761, 0.0157, -0.0, 0.4761, 0.0179, -0.0, 0.4783, 0.0157, -0.0, 0.4783, 0.0136, -0.0, 0.4761, 0.0114, -0.0, 0.4761, 0.0136, -0.0, 0.4783, 0.0114, -0.0, 0.4783, 0.0101, -0.0, 0.4718, 0.0101, -0.0, 0.474, 0.0101, 0.0017, 0.4718, 0.0101, 0.0039, 0.4718, 0.0101, 0.0017, 0.474, 0.0101, 0.0039, 0.474, 0.0101, -0.0, 0.4761, 0.01, -0.0, 0.4783, 0.0101, 0.0017, 0.4761, 0.0101, 0.0039, 0.4761, 0.01, 0.0017, 0.4783, 0.01, 0.0039, 0.4783, 0.0179, -0.0, 0.4805, 0.0157, -0.0, 0.4805, 0.0179, -0.0, 0.4826, 0.0157, -0.0, 0.4826, 0.0136, -0.0, 0.4805, 0.0114, -0.0, 0.4805,
0.0136, -0.0, 0.4826, 0.0114, -0.0, 0.4826, 0.0179, -0.0, 0.4848, 0.0157, -0.0, 0.4848, 0.0179, -0.0, 0.487, 0.0157, -0.0, 0.487, 0.0136, -0.0, 0.4848, 0.0114, -0.0, 0.4848, 0.0136, -0.0, 0.487, 0.0114, -0.0, 0.487, 0.0096, -0.0, 0.4804, 0.0091, -0.0, 0.4824, 0.0096, 0.0017, 0.4805, 0.0096, 0.0039, 0.4805, 0.009, 0.0017, 0.4827, 0.009, 0.0039, 0.4827, 0.0082, -0.0, 0.4843, 0.0077, -0.0, 0.4853, 0.0092, -0.0, 0.487, 0.0068, -0.0, 0.4869, 0.0058, -0.0, 0.4879, 0.0083, 0.0017, 0.4841, 0.0078, 0.0017, 0.4852, 0.0083, 0.0039, 0.4841, 0.0078, 0.0039, 0.4852, 0.0067, 0.0017, 0.4869, 0.0067, 0.0039, 0.4869, 0.0058, 0.0017, 0.4879, 0.0058, 0.0039, 0.4879, 0.0629, -0.0, 0.4919,
0.0628, 0.0017, 0.4919, 0.0628, 0.0039, 0.4919, 0.0648, -0.0, 0.4943, 0.064, -0.0, 0.4932, 0.0657, -0.0, 0.4954, 0.0635, -0.0, 0.4957, 0.0648, 0.0017, 0.4943, 0.0639, 0.0017, 0.4932, 0.0648, 0.0039, 0.4943, 0.0639, 0.0039, 0.4932, 0.0658, 0.0017, 0.4956, 0.0658, 0.0039, 0.4956, 0.0608, -0.0, 0.4895, 0.0603, -0.0, 0.4886, 0.0619, -0.0, 0.4909, 0.0591, -0.0, 0.4913, 0.0562, -0.0, 0.4891, 0.056, -0.0, 0.4912, 0.0558, -0.0, 0.4921, 0.0608, 0.0017, 0.4894, 0.0601, 0.0017, 0.4882, 0.0608, 0.0039, 0.4894, 0.0601, 0.0039, 0.4882, 0.0619, 0.0017, 0.4908, 0.0619, 0.0039, 0.4908, 0.0562, 0.0017, 0.4892, 0.0562, 0.0039, 0.4892, 0.056, 0.0017, 0.4909, 0.0558, 0.0017, 0.492, 0.056, 0.0039, 0.4909,
0.0558, 0.0039, 0.492, 0.0613, -0.0, 0.4935, 0.0591, -0.0, 0.4935, 0.0609, -0.0, 0.4963, 0.0594, -0.0, 0.4953, 0.057, -0.0, 0.4935, 0.0553, -0.0, 0.4934, 0.057, -0.0, 0.4957, 0.0543, -0.0, 0.4952, 0.0608, 0.0017, 0.4963, 0.0594, 0.0017, 0.4953, 0.0608, 0.0039, 0.4963, 0.0594, 0.0039, 0.4953, 0.0553, 0.0017, 0.4935, 0.0553, 0.0039, 0.4935, 0.0543, 0.0017, 0.4953, 0.0543, 0.0039, 0.4953, 0.0668, -0.0, 0.4987, 0.0669, -0.0, 0.5, 0.0666, -0.0, 0.4977, 0.063, -0.0, 0.4983, 0.0652, -0.0, 0.5005, 0.0641, -0.0, 0.4994, 0.0668, 0.0017, 0.4986, 0.0668, 0.0039, 0.4986, 0.0669, 0.0017, 0.5, 0.0669, 0.0039, 0.5, 0.0665, 0.0017, 0.4975, 0.063, 0.0017, 0.4983, 0.0665, 0.0039, 0.4975,
0.063, 0.0039, 0.4983, 0.0652, 0.0017, 0.5005, 0.0641, 0.0017, 0.4994, 0.0652, 0.0039, 0.5005, 0.0641, 0.0039, 0.4994, 0.067, -0.0, 0.502, 0.0664, -0.0, 0.5014, 0.067, 0.0017, 0.502, 0.067, 0.0039, 0.502, 0.0663, 0.0017, 0.5014, 0.0663, 0.0039, 0.5014, 0.0619, -0.0, 0.4972, 0.0582, -0.0, 0.497, 0.0574, -0.0, 0.4981, 0.0548, -0.0, 0.4978, 0.0563, -0.0, 0.4994, 0.0549, -0.0, 0.5006, 0.0619, 0.0017, 0.4972, 0.0582, 0.0017, 0.497, 0.0619, 0.0039, 0.4972, 0.0582, 0.0039, 0.497, 0.0574, 0.0017, 0.4981, 0.0574, 0.0039, 0.4981, 0.0563, 0.0017, 0.4994, 0.0549, 0.0017, 0.5006, 0.0563, 0.0039, 0.4994, 0.0549, 0.0039, 0.5006, 0.0537, -0.0, 0.5011, 0.0538, 0.0017, 0.5011, 0.0538, 0.0039, 0.5011,
0.0455, -0.0, 0.4891, 0.0459, -0.0, 0.4912, 0.0455, 0.0017, 0.4892, 0.0455, 0.0039, 0.4892, 0.046, 0.0017, 0.4914, 0.046, 0.0039, 0.4914, 0.0526, -0.0, 0.4964, 0.0503, -0.0, 0.497, 0.0474, -0.0, 0.4943, 0.0468, -0.0, 0.4933, 0.0484, -0.0, 0.4955, 0.0461, -0.0, 0.4957, 0.053, 0.0017, 0.4962, 0.0503, 0.0017, 0.497, 0.053, 0.0039, 0.4962, 0.0503, 0.0039, 0.497, 0.0474, 0.0017, 0.4943, 0.0468, 0.0017, 0.4932, 0.0474, 0.0039, 0.4943, 0.0468, 0.0039, 0.4932, 0.0484, 0.0017, 0.4955, 0.0484, 0.0039, 0.4955, 0.0439, -0.0, 0.4892, 0.0425, -0.0, 0.4892, 0.0439, -0.0, 0.4913, 0.0425, -0.0, 0.4913, 0.0425, 0.0017, 0.4892, 0.0425, 0.0039, 0.4892, 0.0425, 0.0017, 0.4913, 0.0425, 0.0039, 0.4913,
0.0439, -0.0, 0.4935, 0.0426, -0.0, 0.4935, 0.0431, -0.0, 0.4958, 0.0428, -0.0, 0.4946, 0.0426, 0.0017, 0.4935, 0.0426, 0.0039, 0.4935, 0.0431, 0.0017, 0.4958, 0.0428, 0.0017, 0.4947, 0.0431, 0.0039, 0.4958, 0.0428, 0.0039, 0.4947, 0.0526, -0.0, 0.4978, 0.0505, -0.0, 0.4978, 0.0526, -0.0, 0.5, 0.0505, -0.0, 0.5, 0.0483, -0.0, 0.4978, 0.0461, -0.0, 0.4978, 0.0477, -0.0, 0.5011, 0.0458, -0.0, 0.4998, 0.0475, 0.0017, 0.501, 0.046, 0.0017, 0.5, 0.0475, 0.0039, 0.501, 0.046, 0.0039, 0.5, 0.0527, -0.0, 0.5015, 0.0505, -0.0, 0.5018, 0.0488, -0.0, 0.5015, 0.0527, 0.0017, 0.5015, 0.0505, 0.0017, 0.5018, 0.0527, 0.0039, 0.5015, 0.0505, 0.0039, 0.5018, 0.0486, 0.0017, 0.5015,
0.0486, 0.0039, 0.5015, 0.0442, -0.0, 0.4981, 0.045, -0.0, 0.499, 0.0441, 0.0017, 0.4979, 0.0441, 0.0039, 0.4979, 0.045, 0.0017, 0.499, 0.045, 0.0039, 0.499, 0.0196, -0.0, 0.4889, 0.0195, 0.0017, 0.4892, 0.0195, 0.0039, 0.4892, 0.0192, -0.0, 0.4898, 0.0157, -0.0, 0.4892, 0.0183, -0.0, 0.4917, 0.0157, -0.0, 0.4913, 0.0136, -0.0, 0.4892, 0.0114, -0.0, 0.4892, 0.0136, -0.0, 0.4913, 0.0114, -0.0, 0.4913, 0.019, 0.0017, 0.4902, 0.019, 0.0039, 0.4902, 0.0185, 0.0017, 0.4914, 0.0185, 0.0039, 0.4914, 0.0174, -0.0, 0.4931, 0.0166, -0.0, 0.4942, 0.0156, -0.0, 0.4955, 0.0136, -0.0, 0.4935, 0.0114, -0.0, 0.4935, 0.0146, -0.0, 0.4966, 0.0114, -0.0, 0.4957, 0.0174, 0.0017, 0.4932,
0.0166, 0.0017, 0.4942, 0.0174, 0.0039, 0.4932, 0.0166, 0.0039, 0.4942, 0.0155, 0.0017, 0.4956, 0.0155, 0.0039, 0.4956, 0.0145, 0.0017, 0.4966, 0.0145, 0.0039, 0.4966, 0.0092, -0.0, 0.4892, 0.0071, -0.0, 0.4892, 0.0092, -0.0, 0.4913, 0.0071, -0.0, 0.4913, 0.0048, -0.0, 0.489, 0.0034, -0.0, 0.4899, 0.0049, -0.0, 0.4913, 0.0025, -0.0, 0.4904, 0.0048, 0.0017, 0.489, 0.0034, 0.0017, 0.49, 0.0048, 0.0039, 0.489, 0.0034, 0.0039, 0.49, 0.0023, 0.0017, 0.4905, 0.0023, 0.0039, 0.4905, 0.0092, -0.0, 0.4935, 0.0071, -0.0, 0.4935, 0.0092, -0.0, 0.4957, 0.0071, -0.0, 0.4957, 0.0049, -0.0, 0.4935, 0.0027, -0.0, 0.4935, 0.0049, -0.0, 0.4957, 0.0027, -0.0, 0.4957, 0.0135, -0.0, 0.4977,
0.0121, -0.0, 0.4988, 0.011, -0.0, 0.4995, 0.0135, 0.0017, 0.4977, 0.0122, 0.0017, 0.4987, 0.0135, 0.0039, 0.4977, 0.0122, 0.0039, 0.4987, 0.0111, 0.0017, 0.4994, 0.0111, 0.0039, 0.4994, 0.0092, -0.0, 0.4978, 0.0071, -0.0, 0.4978, 0.0096, -0.0, 0.5004, 0.0076, -0.0, 0.5012, 0.0049, -0.0, 0.4978, 0.0027, -0.0, 0.4978, 0.0049, -0.0, 0.5, 0.0027, -0.0, 0.5, 0.0093, 0.0017, 0.5005, 0.008, 0.0017, 0.501, 0.0093, 0.0039, 0.5005, 0.008, 0.0039, 0.501, 0.0067, -0.0, 0.5015, 0.0049, -0.0, 0.5018, 0.0027, -0.0, 0.5019, 0.0069, 0.0017, 0.5014, 0.0069, 0.0039, 0.5014, 0.0049, 0.0017, 0.5018, 0.0027, 0.0017, 0.5019, 0.0049, 0.0039, 0.5018, 0.0027, 0.0039, 0.5019, 0.0013, -0.0, 0.0031,
-0.0016, -0.0, 0.003, 0.0012, -0.0, 0.0053, -0.0016, -0.0, 0.0052, -0.0038, -0.0, 0.003, -0.006, -0.0, 0.003, -0.0038, -0.0, 0.0052, -0.006, -0.0, 0.0052, 0.0013, 0.0017, 0.003, 0.0013, 0.0039, 0.003, 0.0012, 0.0017, 0.0052, 0.0012, 0.0039, 0.0052, 0.0009, -0.0, 0.0075, -0.0016, -0.0, 0.0074, 0.0004, -0.0, 0.0097, -0.0016, -0.0, 0.0096, -0.0038, -0.0, 0.0074, -0.006, -0.0, 0.0074, -0.0038, -0.0, 0.0096, -0.006, -0.0, 0.0096, 0.0009, 0.0017, 0.0074, 0.0009, 0.0039, 0.0074, 0.0005, 0.0017, 0.0096, 0.0005, 0.0039, 0.0096, -0.0081, -0.0, 0.003, -0.0105, -0.0, 0.003, -0.0081, -0.0, 0.0052, -0.0105, -0.0, 0.0051, -0.0105, 0.0017, 0.003, -0.0105, 0.0039, 0.003, -0.0106, 0.0017, 0.0052,
-0.0106, 0.0039, 0.0052, -0.0081, -0.0, 0.0074, -0.0109, -0.0, 0.0072, -0.0081, -0.0, 0.0096, -0.0115, -0.0, 0.0089, -0.0119, -0.0, 0.0099, -0.0109, 0.0017, 0.0074, -0.0109, 0.0039, 0.0074, -0.0114, 0.0017, 0.0085, -0.0114, 0.0039, 0.0085, -0.0118, 0.0017, 0.0096, -0.0118, 0.0039, 0.0096, -0.0001, -0.0, 0.0114, -0.0004, -0.0, 0.0123, -0.0012, -0.0, 0.0142, -0.0038, -0.0, 0.0117, -0.006, -0.0, 0.0117, -0.0038, -0.0, 0.0139, -0.006, -0.0, 0.0139, -0.0002, 0.0017, 0.0116, -0.0006, 0.0017, 0.0127, -0.0002, 0.0039, 0.0116, -0.0006, 0.0039, 0.0127, -0.0011, 0.0017, 0.0139, -0.0011, 0.0039, 0.0139, -0.0022, -0.0, 0.0162, -0.0027, -0.0, 0.0172, -0.0038, -0.0, 0.0161, -0.006, -0.0, 0.0161, -0.0034, -0.0, 0.0183,
-0.006, -0.0, 0.0182, -0.0021, 0.0017, 0.0161, -0.0021, 0.0039, 0.0161, -0.0027, 0.0017, 0.0172, -0.0027, 0.0039, 0.0172, -0.0033, 0.0017, 0.0183, -0.0033, 0.0039, 0.0183, -0.0081, -0.0, 0.0117, -0.0103, -0.0, 0.0117, -0.0081, -0.0, 0.0139, -0.0103, -0.0, 0.0139, -0.0129, -0.0, 0.0115, -0.0137, -0.0, 0.0126, -0.0125, -0.0, 0.0139, -0.0147, -0.0, 0.0138, -0.0129, 0.0017, 0.0116, -0.0137, 0.0017, 0.0126, -0.0129, 0.0039, 0.0116, -0.0137, 0.0039, 0.0126, -0.0147, 0.0017, 0.0138, -0.0147, 0.0039, 0.0138, -0.0081, -0.0, 0.0161, -0.0103, -0.0, 0.0161, -0.0081, -0.0, 0.0182, -0.0103, -0.0, 0.0182, -0.0125, -0.0, 0.0161, -0.0146, -0.0, 0.0161, -0.0125, -0.0, 0.0182, -0.0146, -0.0, 0.0182, -0.0159, -0.0, 0.0148,
-0.016, 0.0017, 0.0148, -0.016, 0.0039, 0.0148, -0.017, -0.0, 0.0156, -0.019, -0.0, 0.0167, -0.0168, -0.0, 0.0182, -0.019, -0.0, 0.0182, -0.0203, -0.0, 0.0171, -0.0211, -0.0, 0.0173, -0.0233, -0.0, 0.0176, -0.017, 0.0017, 0.0156, -0.0189, 0.0017, 0.0167, -0.017, 0.0039, 0.0156, -0.0189, 0.0039, 0.0167, -0.0203, 0.0017, 0.0171, -0.0203, 0.0039, 0.0171, -0.0214, 0.0017, 0.0174, -0.0233, 0.0017, 0.0176, -0.0214, 0.0039, 0.0174, -0.0233, 0.0039, 0.0176, -0.0325, -0.0, 0.0144, -0.0324, 0.0017, 0.0144, -0.0324, 0.0039, 0.0144, -0.0282, -0.0, 0.0169, -0.0254, -0.0, 0.0175, -0.0274, -0.0, 0.0171, -0.0297, -0.0, 0.0163, -0.0315, -0.0, 0.0152, -0.0298, -0.0, 0.0182, -0.032, -0.0, 0.0182, -0.028, 0.0017, 0.017,
-0.028, 0.0039, 0.017, -0.0255, 0.0017, 0.0175, -0.0269, 0.0017, 0.0172, -0.0255, 0.0039, 0.0175, -0.0269, 0.0039, 0.0172, -0.0299, 0.0017, 0.0162, -0.0313, 0.0017, 0.0153, -0.0299, 0.0039, 0.0162, -0.0313, 0.0039, 0.0153, -0.0044, -0.0, 0.0198, -0.0053, -0.0, 0.021, -0.0064, -0.0, 0.0221, -0.0044, 0.0017, 0.0199, -0.0053, 0.0017, 0.021, -0.0044, 0.0039, 0.0199, -0.0053, 0.0039, 0.021, -0.0064, 0.0017, 0.0221, -0.0064, 0.0039, 0.0221, -0.0081, -0.0, 0.0204, -0.0103, -0.0, 0.0204, -0.0076, -0.0, 0.0232, -0.0103, -0.0, 0.0226, -0.0125, -0.0, 0.0204, -0.0146, -0.0, 0.0204, -0.0125, -0.0, 0.0226, -0.0146, -0.0, 0.0226, -0.0076, 0.0017, 0.0232, -0.0076, 0.0039, 0.0232, -0.0087, -0.0, 0.0241, -0.0102, -0.0, 0.0252,
-0.0113, -0.0, 0.0259, -0.0125, -0.0, 0.0247, -0.0146, -0.0, 0.0247, -0.0123, -0.0, 0.0265, -0.0144, -0.0, 0.0273, -0.0087, 0.0017, 0.0241, -0.0102, 0.0017, 0.0252, -0.0087, 0.0039, 0.0241, -0.0102, 0.0039, 0.0252, -0.0113, 0.0017, 0.0259, -0.0113, 0.0039, 0.0259, -0.0124, 0.0017, 0.0265, -0.0146, 0.0017, 0.0274, -0.0124, 0.0039, 0.0265, -0.0146, 0.0039, 0.0274, -0.0168, -0.0, 0.0204, -0.019, -0.0, 0.0204, -0.0168, -0.0, 0.0226, -0.019, -0.0, 0.0226, -0.0212, -0.0, 0.0204, -0.0233, -0.0, 0.0204, -0.0212, -0.0, 0.0226, -0.0233, -0.0, 0.0226, -0.0168, -0.0, 0.0247, -0.019, -0.0, 0.0247, -0.0166, -0.0, 0.028, -0.019, -0.0, 0.0269, -0.0212, -0.0, 0.0247, -0.0233, -0.0, 0.0247, -0.0212, -0.0, 0.0269,
-0.0233, -0.0, 0.0269, -0.0163, 0.0017, 0.0279, -0.0163, 0.0039, 0.0279, -0.0255, -0.0, 0.0204, -0.0277, -0.0, 0.0204, -0.0255, -0.0, 0.0226, -0.0277, -0.0, 0.0226, -0.0298, -0.0, 0.0204, -0.032, -0.0, 0.0204, -0.0298, -0.0, 0.0226, -0.032, -0.0, 0.0226, -0.0255, -0.0, 0.0247, -0.0277, -0.0, 0.0247, -0.0255, -0.0, 0.0269, -0.0277, -0.0, 0.0269, -0.0298, -0.0, 0.0247, -0.032, -0.0, 0.0247, -0.0302, -0.0, 0.0281, -0.0323, -0.0, 0.0276, -0.0308, 0.0017, 0.028, -0.032, 0.0017, 0.0277, -0.0308, 0.0039, 0.028, -0.032, 0.0039, 0.0277, -0.0175, -0.0, 0.0281, -0.019, -0.0, 0.0283, -0.0212, -0.0, 0.0284, -0.0233, -0.0, 0.0284, -0.0174, 0.0017, 0.0281, -0.019, 0.0017, 0.0283, -0.0174, 0.0039, 0.0281,
-0.019, 0.0039, 0.0283, -0.0212, 0.0017, 0.0284, -0.0233, 0.0017, 0.0284, -0.0212, 0.0039, 0.0284, -0.0233, 0.0039, 0.0284, -0.0255, -0.0, 0.0284, -0.0277, -0.0, 0.0284, -0.0294, -0.0, 0.0282, -0.0255, 0.0017, 0.0284, -0.0277, 0.0017, 0.0284, -0.0255, 0.0039, 0.0284, -0.0277, 0.0039, 0.0284, -0.0297, 0.0017, 0.0282, -0.0297, 0.0039, 0.0282, -0.0373, -0.0, 0.0061, -0.0378, -0.0, 0.003, -0.0407, -0.0, 0.003, -0.0376, -0.0, 0.0052, -0.0407, -0.0, 0.0052, -0.0374, 0.0017, 0.006, -0.0374, 0.0039, 0.006, -0.0378, 0.0017, 0.003, -0.0378, 0.0039, 0.003, -0.0376, 0.0017, 0.0049, -0.0376, 0.0039, 0.0049, -0.037, -0.0, 0.0073, -0.0362, -0.0, 0.0093, -0.0385, -0.0, 0.0074, -0.0407, -0.0, 0.0074, -0.0385, -0.0, 0.0096,
-0.0407, -0.0, 0.0096, -0.037, 0.0017, 0.0074, -0.037, 0.0039, 0.0074, -0.0361, 0.0017, 0.0096, -0.0361, 0.0039, 0.0096, -0.0429, -0.0, 0.003, -0.045, -0.0, 0.003, -0.0429, -0.0, 0.0052, -0.045, -0.0, 0.0052, -0.0472, -0.0, 0.003, -0.0487, -0.0, 0.003, -0.0472, -0.0, 0.0052, -0.0487, -0.0, 0.0052, -0.0487, 0.0017, 0.003, -0.0487, 0.0039, 0.003, -0.0487, 0.0017, 0.0052, -0.0487, 0.0039, 0.0052, -0.0429, -0.0, 0.0074, -0.045, -0.0, 0.0074, -0.0429, -0.0, 0.0096, -0.045, -0.0, 0.0096, -0.0484, -0.0, 0.0078, -0.0485, -0.0, 0.007, -0.048, -0.0, 0.0098, -0.0483, 0.0017, 0.0084, -0.0485, 0.0017, 0.0073, -0.0483, 0.0039, 0.0084, -0.0485, 0.0039, 0.0073, -0.048, 0.0017, 0.0096, -0.048, 0.0039, 0.0096,
-0.0346, -0.0, 0.012, -0.0352, -0.0, 0.011, -0.0335, -0.0, 0.0133, -0.0363, -0.0, 0.0139, -0.0385, -0.0, 0.0117, -0.0407, -0.0, 0.0117, -0.0385, -0.0, 0.0139, -0.0407, -0.0, 0.0139, -0.0347, 0.0017, 0.012, -0.0354, 0.0017, 0.0108, -0.0347, 0.0039, 0.012, -0.0354, 0.0039, 0.0108, -0.0335, 0.0017, 0.0133, -0.0335, 0.0039, 0.0133, -0.0342, -0.0, 0.0161, -0.0363, -0.0, 0.0161, -0.0342, -0.0, 0.0182, -0.0363, -0.0, 0.0182, -0.0385, -0.0, 0.0161, -0.0407, -0.0, 0.0161, -0.0385, -0.0, 0.0182, -0.0407, -0.0, 0.0182, -0.0429, -0.0, 0.0117, -0.045, -0.0, 0.0117, -0.0429, -0.0, 0.0139, -0.0463, -0.0, 0.0146, -0.0473, -0.0, 0.0119, -0.0467, -0.0, 0.0136, -0.0461, 0.0017, 0.0149, -0.0461, 0.0039, 0.0149,
-0.0474, 0.0017, 0.0117, -0.0474, 0.0039, 0.0117, -0.0466, 0.0017, 0.0139, -0.0466, 0.0039, 0.0139, -0.0429, -0.0, 0.0161, -0.0454, -0.0, 0.0164, -0.0438, -0.0, 0.0189, -0.0445, -0.0, 0.0178, -0.0455, 0.0017, 0.0161, -0.0455, 0.0039, 0.0161, -0.0437, 0.0017, 0.019, -0.0445, 0.0017, 0.0179, -0.0437, 0.0039, 0.019, -0.0445, 0.0039, 0.0179, -0.0342, -0.0, 0.0204, -0.0363, -0.0, 0.0204, -0.0342, -0.0, 0.0226, -0.0363, -0.0, 0.0226, -0.0385, -0.0, 0.0204, -0.0417, -0.0, 0.0214, -0.0394, -0.0, 0.0236, -0.0406, -0.0, 0.0225, -0.0417, 0.0017, 0.0214, -0.0417, 0.0039, 0.0214, -0.0394, 0.0017, 0.0235, -0.0407, 0.0017, 0.0225, -0.0394, 0.0039, 0.0235, -0.0407, 0.0039, 0.0225, -0.0342, -0.0, 0.0247, -0.0367, -0.0, 0.0255,
-0.0344, -0.0, 0.0268, -0.0356, -0.0, 0.0261, -0.0383, -0.0, 0.0244, -0.0368, 0.0017, 0.0254, -0.0368, 0.0039, 0.0254, -0.0342, 0.0017, 0.0269, -0.0357, 0.0017, 0.0261, -0.0342, 0.0039, 0.0269, -0.0357, 0.0039, 0.0261, -0.0384, 0.0017, 0.0244, -0.0384, 0.0039, 0.0244, -0.0427, -0.0, 0.0203, -0.0427, 0.0017, 0.0204, -0.0427, 0.0039, 0.0204, 0.0005, -0.0, 0.0716, -0.0016, -0.0, 0.0716, 0.0005, -0.0, 0.0747, -0.0016, -0.0, 0.0747, -0.0038, -0.0, 0.0716, -0.006, -0.0, 0.0716, -0.0038, -0.0, 0.0747, -0.006, -0.0, 0.0747, 0.0005, 0.0017, 0.0716, -0.0016, 0.0017, 0.0716, 0.0005, 0.0039, 0.0716, -0.0016, 0.0039, 0.0716, -0.0038, 0.0017, 0.0716, -0.006, 0.0017, 0.0716, -0.0038, 0.0039, 0.0716, -0.006, 0.0039, 0.0716,
0.0005, -0.0, 0.0768, -0.0016, -0.0, 0.0768, 0.0005, -0.0, 0.079, -0.0016, -0.0, 0.079, -0.0038, -0.0, 0.0768, -0.006, -0.0, 0.0768, -0.0038, -0.0, 0.079, -0.006, -0.0, 0.079, -0.0081, -0.0, 0.0716, -0.0103, -0.0, 0.0716, -0.0081, -0.0, 0.0747, -0.0103, -0.0, 0.0747, -0.0125, -0.0, 0.0716, -0.0146, -0.0, 0.0716, -0.0125, -0.0, 0.0747, -0.0146, -0.0, 0.0747, -0.0081, 0.0017, 0.0716, -0.0103, 0.0017, 0.0716, -0.0081, 0.0039, 0.0716, -0.0103, 0.0039, 0.0716, -0.0125, 0.0017, 0.0716, -0.0146, 0.0017, 0.0716, -0.0125, 0.0039, 0.0716, -0.0146, 0.0039, 0.0716, -0.0081, -0.0, 0.0768, -0.0103, -0.0, 0.0768, -0.0081, -0.0, 0.079, -0.0103, -0.0, 0.079, -0.0125, -0.0, 0.0768, -0.0146, -0.0, 0.0768,
-0.0125, -0.0, 0.079, -0.0146, -0.0, 0.079, 0.0005, -0.0, 0.0812, -0.0016, -0.0, 0.0812, 0.0005, -0.0, 0.0833, -0.0016, -0.0, 0.0833, -0.0038, -0.0, 0.0812, -0.006, -0.0, 0.0812, -0.0038, -0.0, 0.0833, -0.006, -0.0, 0.0833, 0.0005, 0.0017, 0.0833, -0.0016, 0.0017, 0.0833, 0.0005, 0.0039, 0.0833, -0.0016, 0.0039, 0.0833, -0.0038, 0.0017, 0.0833, -0.006, 0.0017, 0.0833, -0.0038, 0.0039, 0.0833, -0.006, 0.0039, 0.0833, -0.0081, -0.0, 0.0812, -0.0103, -0.0, 0.0812, -0.0085, -0.0, 0.0833, -0.0103, -0.0, 0.0833, -0.0125, -0.0, 0.0812, -0.0146, -0.0, 0.0812, -0.0125, -0.0, 0.0833, -0.0146, -0.0, 0.0833, -0.0085, 0.0017, 0.0833, -0.0085, 0.0039, 0.0833, -0.0085, -0.0, 0.0855, -0.0103, -0.0, 0.0855,
-0.0085, -0.0, 0.0877, -0.0103, -0.0, 0.0877, -0.0125, -0.0, 0.0855, -0.0146, -0.0, 0.0855, -0.0125, -0.0, 0.0877, -0.0146, -0.0, 0.0877, -0.0085, 0.0017, 0.0855, -0.0085, 0.0039, 0.0855, -0.0085, 0.0017, 0.0877, -0.0085, 0.0039, 0.0877, -0.0168, -0.0, 0.0716, -0.019, -0.0, 0.0716, -0.0168, -0.0, 0.0747, -0.019, -0.0, 0.0747, -0.0212, -0.0, 0.0716, -0.0233, -0.0, 0.0716, -0.0212, -0.0, 0.0747, -0.0233, -0.0, 0.0747, -0.0168, 0.0017, 0.0716, -0.019, 0.0017, 0.0716, -0.0168, 0.0039, 0.0716, -0.019, 0.0039, 0.0716, -0.0212, 0.0017, 0.0716, -0.0233, 0.0017, 0.0716, -0.0212, 0.0039, 0.0716, -0.0233, 0.0039, 0.0716, -0.0168, -0.0, 0.0768, -0.019, -0.0, 0.0768, -0.0168, -0.0, 0.079, -0.019, -0.0, 0.079,
-0.0212, -0.0, 0.0768, -0.0233, -0.0, 0.0768, -0.0212, -0.0, 0.079, -0.0233, -0.0, 0.079, -0.0255, -0.0, 0.0716, -0.0277, -0.0, 0.0716, -0.0255, -0.0, 0.0747, -0.0277, -0.0, 0.0747, -0.0298, -0.0, 0.0716, -0.032, -0.0, 0.0716, -0.0298, -0.0, 0.0747, -0.032, -0.0, 0.0747, -0.0255, 0.0017, 0.0716, -0.0277, 0.0017, 0.0716, -0.0255, 0.0039, 0.0716, -0.0277, 0.0039, 0.0716, -0.0298, 0.0017, 0.0716, -0.032, 0.0017, 0.0716, -0.0298, 0.0039, 0.0716, -0.032, 0.0039, 0.0716, -0.0255, -0.0, 0.0768, -0.0277, -0.0, 0.0768, -0.0255, -0.0, 0.079, -0.0277, -0.0, 0.079, -0.0298, -0.0, 0.0768, -0.032, -0.0, 0.0768, -0.0298, -0.0, 0.079, -0.032, -0.0, 0.079, -0.0168, -0.0, 0.0812, -0.019, -0.0, 0.0812,
-0.0168, -0.0, 0.0833, -0.019, -0.0, 0.0833, -0.0212, -0.0, 0.0812, -0.0233, -0.0, 0.0812, -0.0203, -0.0, 0.0833, -0.0233, -0.0, 0.0833, -0.0203, 0.0017, 0.0833, -0.0233, 0.0017, 0.0833, -0.0203, 0.0039, 0.0833, -0.0233, 0.0039, 0.0833, -0.0168, -0.0, 0.0855, -0.019, -0.0, 0.0855, -0.0168, -0.0, 0.0877, -0.019, -0.0, 0.0877, -0.0203, -0.0, 0.0855, -0.0203, -0.0, 0.0877, -0.0203, 0.0017, 0.0855, -0.0203, 0.0039, 0.0855, -0.0203, 0.0017, 0.0877, -0.0203, 0.0039, 0.0877, -0.0255, -0.0, 0.0812, -0.0277, -0.0, 0.0812, -0.0255, -0.0, 0.0833, -0.0277, -0.0, 0.0833, -0.0298, -0.0, 0.0812, -0.032, -0.0, 0.0812, -0.0298, -0.0, 0.0833, -0.032, -0.0, 0.0833, -0.0255, 0.0017, 0.0833, -0.0277, 0.0017, 0.0833,
-0.0255, 0.0039, 0.0833, -0.0277, 0.0039, 0.0833, -0.0298, 0.0017, 0.0833, -0.032, 0.0017, 0.0833, -0.0298, 0.0039, 0.0833, -0.032, 0.0039, 0.0833, -0.0085, -0.0, 0.0899, -0.0103, -0.0, 0.0899, -0.0085, -0.0, 0.092, -0.0103, -0.0, 0.092, -0.0125, -0.0, 0.0899, -0.0146, -0.0, 0.0899, -0.0125, -0.0, 0.092, -0.0146, -0.0, 0.092, -0.0085, 0.0017, 0.0899, -0.0085, 0.0039, 0.0899, -0.0085, 0.0017, 0.092, -0.0085, 0.0039, 0.092, -0.0085, -0.0, 0.0942, -0.0103, -0.0, 0.0942, -0.0085, -0.0, 0.0964, -0.0103, -0.0, 0.0964, -0.0125, -0.0, 0.0942, -0.0146, -0.0, 0.0942, -0.0125, -0.0, 0.0964, -0.0146, -0.0, 0.0964, -0.0085, 0.0017, 0.0942, -0.0085, 0.0039, 0.0942, -0.0085, 0.0017, 0.0964, -0.0085, 0.0039, 0.0964,
-0.0085, -0.0, 0.0985, -0.0103, -0.0, 0.0985, -0.0085, -0.0, 0.1007, -0.0103, -0.0, 0.1007, -0.0125, -0.0, 0.0985, -0.0146, -0.0, 0.0985, -0.0125, -0.0, 0.1007, -0.0146, -0.0, 0.1007, -0.0085, 0.0017, 0.0985, -0.0085, 0.0039, 0.0985, -0.0085, 0.0017, 0.1007, -0.0085, 0.0039, 0.1007, -0.0085, -0.0, 0.1029, -0.0103, -0.0, 0.1029, -0.0085, -0.0, 0.105, -0.0103, -0.0, 0.105, -0.0125, -0.0, 0.1029, -0.0146, -0.0, 0.1029, -0.0125, -0.0, 0.105, -0.0146, -0.0, 0.105, -0.0085, 0.0017, 0.1029, -0.0085, 0.0039, 0.1029, -0.0085, 0.0017, 0.105, -0.0085, 0.0039, 0.105, -0.0168, -0.0, 0.0899, -0.019, -0.0, 0.0899, -0.0168, -0.0, 0.092, -0.019, -0.0, 0.092, -0.0203, -0.0, 0.0899, -0.0203, -0.0, 0.092,
-0.0203, 0.0017, 0.0899, -0.0203, 0.0039, 0.0899, -0.0203, 0.0017, 0.092, -0.0203, 0.0039, 0.092, -0.0168, -0.0, 0.0942, -0.019, -0.0, 0.0942, -0.0168, -0.0, 0.0964, -0.019, -0.0, 0.0964, -0.0203, -0.0, 0.0942, -0.0203, -0.0, 0.0964, -0.0203, 0.0017, 0.0942, -0.0203, 0.0039, 0.0942, -0.0203, 0.0017, 0.0964, -0.0203, 0.0039, 0.0964, -0.0168, -0.0, 0.0985, -0.019, -0.0, 0.0985, -0.0168, -0.0, 0.1007, -0.019, -0.0, 0.1007, -0.0203, -0.0, 0.0985, -0.0203, -0.0, 0.1007, -0.0203, 0.0017, 0.0985, -0.0203, 0.0039, 0.0985, -0.0203, 0.0017, 0.1007, -0.0203, 0.0039, 0.1007, -0.0168, -0.0, 0.1029, -0.019, -0.0, 0.1029, -0.0168, -0.0, 0.105, -0.019, -0.0, 0.105, -0.0203, -0.0, 0.1029, -0.0203, -0.0, 0.105,
-0.0203, 0.0017, 0.1029, -0.0203, 0.0039, 0.1029, -0.0203, 0.0017, 0.105, -0.0203, 0.0039, 0.105, -0.0342, -0.0, 0.0716, -0.0363, -0.0, 0.0716, -0.0342, -0.0, 0.0747, -0.0363, -0.0, 0.0747, -0.0385, -0.0, 0.0716, -0.0407, -0.0, 0.0716, -0.0385, -0.0, 0.0747, -0.0407, -0.0, 0.0747, -0.0342, 0.0017, 0.0716, -0.0363, 0.0017, 0.0716, -0.0342, 0.0039, 0.0716, -0.0363, 0.0039, 0.0716, -0.0385, 0.0017, 0.0716, -0.0407, 0.0017, 0.0716, -0.0385, 0.0039, 0.0716, -0.0407, 0.0039, 0.0716, -0.0342, -0.0, 0.0768, -0.0363, -0.0, 0.0768, -0.0342, -0.0, 0.079, -0.0363, -0.0, 0.079, -0.0385, -0.0, 0.0768, -0.0407, -0.0, 0.0768, -0.0385, -0.0, 0.079, -0.0407, -0.0, 0.079, -0.0429, -0.0, 0.0716, -0.045, -0.0, 0.0716,
-0.0429, -0.0, 0.0747, -0.045, -0.0, 0.0747, -0.0472, -0.0, 0.0716, -0.0487, -0.0, 0.0716, -0.0472, -0.0, 0.0747, -0.0487, -0.0, 0.0747, -0.0429, 0.0017, 0.0716, -0.045, 0.0017, 0.0716, -0.0429, 0.0039, 0.0716, -0.045, 0.0039, 0.0716, -0.0472, 0.0017, 0.0716, -0.0487, 0.0017, 0.0716, -0.0472, 0.0039, 0.0716, -0.0487, 0.0039, 0.0716, -0.0487, 0.0017, 0.0747, -0.0487, 0.0039, 0.0747, -0.0429, -0.0, 0.0768, -0.045, -0.0, 0.0768, -0.0429, -0.0, 0.079, -0.045, -0.0, 0.079, -0.0472, -0.0, 0.0768, -0.0487, -0.0, 0.0768, -0.0472, -0.0, 0.079, -0.0487, -0.0, 0.079, -0.0487, 0.0017, 0.0768, -0.0487, 0.0039, 0.0768, -0.0487, 0.0017, 0.079, -0.0487, 0.0039, 0.079, -0.0342, -0.0, 0.0812, -0.0363, -0.0, 0.0812,
-0.0342, -0.0, 0.0833, -0.0363, -0.0, 0.0833, -0.0385, -0.0, 0.0812, -0.0407, -0.0, 0.0812, -0.0379, -0.0, 0.0833, -0.0407, -0.0, 0.0833, -0.0342, 0.0017, 0.0833, -0.0363, 0.0017, 0.0833, -0.0342, 0.0039, 0.0833, -0.0363, 0.0039, 0.0833, -0.0379, 0.0017, 0.0833, -0.0379, 0.0039, 0.0833, -0.0379, -0.0, 0.0855, -0.0407, -0.0, 0.0855, -0.0379, -0.0, 0.0877, -0.0407, -0.0, 0.0877, -0.0379, 0.0017, 0.0855, -0.0379, 0.0039, 0.0855, -0.0379, 0.0017, 0.0877, -0.0379, 0.0039, 0.0877, -0.0429, -0.0, 0.0812, -0.045, -0.0, 0.0812, -0.0429, -0.0, 0.0833, -0.045, -0.0, 0.0833, -0.0472, -0.0, 0.0812, -0.0487, -0.0, 0.0812, -0.0472, -0.0, 0.0833, -0.0487, -0.0, 0.0833, -0.0487, 0.0017, 0.0812, -0.0487, 0.0039, 0.0812,
-0.0487, 0.0017, 0.0833, -0.0487, 0.0039, 0.0833, -0.0429, -0.0, 0.0855, -0.045, -0.0, 0.0855, -0.0429, -0.0, 0.0877, -0.045, -0.0, 0.0877, -0.0472, -0.0, 0.0855, -0.0487, -0.0, 0.0855, -0.0472, -0.0, 0.0877, -0.0487, -0.0, 0.0877, -0.0487, 0.0017, 0.0855, -0.0487, 0.0039, 0.0855, -0.0487, 0.0017, 0.0877, -0.0487, 0.0039, 0.0877, -0.0379, -0.0, 0.0899, -0.0407, -0.0, 0.0899, -0.0379, -0.0, 0.092, -0.0407, -0.0, 0.092, -0.0379, 0.0017, 0.0899, -0.0379, 0.0039, 0.0899, -0.0379, 0.0017, 0.092, -0.0379, 0.0039, 0.092, -0.0379, -0.0, 0.0942, -0.0407, -0.0, 0.0942, -0.0379, -0.0, 0.0964, -0.0407, -0.0, 0.0964, -0.0379, 0.0017, 0.0942, -0.0379, 0.0039, 0.0942, -0.0379, 0.0017, 0.0964, -0.0379, 0.0039, 0.0964,
-0.0429, -0.0, 0.0899, -0.045, -0.0, 0.0899, -0.0429, -0.0, 0.092, -0.045, -0.0, 0.092, -0.0472, -0.0, 0.0899, -0.0487, -0.0, 0.0899, -0.0472, -0.0, 0.092, -0.0487, -0.0, 0.092, -0.0487, 0.0017, 0.0899, -0.0487, 0.0039, 0.0899, -0.0487, 0.0017, 0.092, -0.0487, 0.0039, 0.092, -0.0429, -0.0, 0.0942, -0.045, -0.0, 0.0942, -0.0429, -0.0, 0.0964, -0.045, -0.0, 0.0964, -0.0472, -0.0, 0.0942, -0.0487, -0.0, 0.0942, -0.0472, -0.0, 0.0964, -0.0487, -0.0, 0.0964, -0.0487, 0.0017, 0.0942, -0.0487, 0.0039, 0.0942, -0.0487, 0.0017, 0.0964, -0.0487, 0.0039, 0.0964, -0.0379, -0.0, 0.0985, -0.0407, -0.0, 0.0985, -0.0379, -0.0, 0.1007, -0.0407, -0.0, 0.1007, -0.0379, 0.0017, 0.0985, -0.0379, 0.0039, 0.0985,
-0.0379, 0.0017, 0.1007, -0.0379, 0.0039, 0.1007, -0.0379, -0.0, 0.1029, -0.0407, -0.0, 0.1029, -0.0379, -0.0, 0.105, -0.0407, -0.0, 0.105, -0.0379, 0.0017, 0.1029, -0.0379, 0.0039, 0.1029, -0.0379, 0.0017, 0.105, -0.0379, 0.0039, 0.105, -0.0429, -0.0, 0.0985, -0.045, -0.0, 0.0985, -0.0429, -0.0, 0.1007, -0.045, -0.0, 0.1007, -0.0472, -0.0, 0.0985, -0.0487, -0.0, 0.0985, -0.0472, -0.0, 0.1007, -0.0487, -0.0, 0.1007, -0.0487, 0.0017, 0.0985, -0.0487, 0.0039, 0.0985, -0.0487, 0.0017, 0.1007, -0.0487, 0.0039, 0.1007, -0.0429, -0.0, 0.1029, -0.045, -0.0, 0.1029, -0.0429, -0.0, 0.105, -0.045, -0.0, 0.105, -0.0472, -0.0, 0.1029, -0.0487, -0.0, 0.1029, -0.0472, -0.0, 0.105, -0.0487, -0.0, 0.105,
-0.0487, 0.0017, 0.1029, -0.0487, 0.0039, 0.1029, -0.0487, 0.0017, 0.105, -0.0487, 0.0039, 0.105, -0.0085, -0.0, 0.1072, -0.0103, -0.0, 0.1072, -0.0085, -0.0, 0.1094, -0.0103, -0.0, 0.1094, -0.0125, -0.0, 0.1072, -0.0146, -0.0, 0.1072, -0.0125, -0.0, 0.1094, -0.0146, -0.0, 0.1094, -0.0085, 0.0017, 0.1072, -0.0085, 0.0039, 0.1072, -0.0085, 0.0017, 0.1094, -0.0085, 0.0039, 0.1094, -0.0085, -0.0, 0.1116, -0.0103, -0.0, 0.1116, -0.0085, -0.0, 0.1137, -0.0103, -0.0, 0.1137, -0.0125, -0.0, 0.1116, -0.0146, -0.0, 0.1116, -0.0125, -0.0, 0.1137, -0.0146, -0.0, 0.1137, -0.0085, 0.0017, 0.1116, -0.0085, 0.0039, 0.1116, -0.0085, 0.0017, 0.1137, -0.0085, 0.0039, 0.1137, -0.0085, -0.0, 0.1159, -0.0103, -0.0, 0.1159,
-0.0085, -0.0, 0.1181, -0.0103, -0.0, 0.1181, -0.0125, -0.0, 0.1159, -0.0146, -0.0, 0.1159, -0.0125, -0.0, 0.1181, -0.0146, -0.0, 0.1181, -0.0085, 0.0017, 0.1159, -0.0085, 0.0039, 0.1159, -0.0085, 0.0017, 0.1181, -0.0085, 0.0039, 0.1181, -0.0085, -0.0, 0.1202, -0.0103, -0.0, 0.1202, -0.0085, -0.0, 0.1224, -0.0103, -0.0, 0.1224, -0.0125, -0.0, 0.1202, -0.0146, -0.0, 0.1202, -0.0125, -0.0, 0.1224, -0.0146, -0.0, 0.1224, -0.0085, 0.0017, 0.1202, -0.0085, 0.0039, 0.1202, -0.0085, 0.0017, 0.1224, -0.0085, 0.0039, 0.1224, -0.0168, -0.0, 0.1072, -0.019, -0.0, 0.1072, -0.0168, -0.0, 0.1094, -0.019, -0.0, 0.1094, -0.0203, -0.0, 0.1072, -0.0203, -0.0, 0.1094, -0.0203, 0.0017, 0.1072, -0.0203, 0.0039, 0.1072,
-0.0203, 0.0017, 0.1094, -0.0203, 0.0039, 0.1094, -0.0168, -0.0, 0.1116, -0.019, -0.0, 0.1116, -0.0168, -0.0, 0.1137, -0.019, -0.0, 0.1137, -0.0203, -0.0, 0.1116, -0.0203, -0.0, 0.1137, -0.0203, 0.0017, 0.1116, -0.0203, 0.0039, 0.1116, -0.0203, 0.0017, 0.1137, -0.0203, 0.0039, 0.1137, -0.0168, -0.0, 0.1159, -0.019, -0.0, 0.1159, -0.0168, -0.0, 0.1181, -0.019, -0.0, 0.1181, -0.0203, -0.0, 0.1159, -0.0203, -0.0, 0.1181, -0.0203, 0.0017, 0.1159, -0.0203, 0.0039, 0.1159, -0.0203, 0.0017, 0.1181, -0.0203, 0.0039, 0.1181, -0.0168, -0.0, 0.1202, -0.019, -0.0, 0.1202, -0.0168, -0.0, 0.1224, -0.019, -0.0, 0.1224, -0.0203, -0.0, 0.1202, -0.0203, -0.0, 0.1224, -0.0203, 0.0017, 0.1202, -0.0203, 0.0039, 0.1202,
-0.0203, 0.0017, 0.1224, -0.0203, 0.0039, 0.1224, -0.0085, -0.0, 0.1246, -0.0103, -0.0, 0.1246, -0.0085, -0.0, 0.1267, -0.0103, -0.0, 0.1267, -0.0125, -0.0, 0.1246, -0.0146, -0.0, 0.1246, -0.0125, -0.0, 0.1267, -0.0146, -0.0, 0.1267, -0.0085, 0.0017, 0.1246, -0.0085, 0.0039, 0.1246, -0.0085, 0.0017, 0.1267, -0.0085, 0.0039, 0.1267, -0.0085, -0.0, 0.1294, -0.0103, -0.0, 0.1294, -0.0125, -0.0, 0.1294, -0.0146, -0.0, 0.1294, -0.0085, 0.0017, 0.1294, -0.0103, 0.0017, 0.1294, -0.0085, 0.0039, 0.1294, -0.0103, 0.0039, 0.1294, -0.0125, 0.0017, 0.1294, -0.0146, 0.0017, 0.1294, -0.0125, 0.0039, 0.1294, -0.0146, 0.0039, 0.1294, -0.0168, -0.0, 0.1246, -0.019, -0.0, 0.1246, -0.0168, -0.0, 0.1267, -0.019, -0.0, 0.1267,
-0.0203, -0.0, 0.1246, -0.0203, -0.0, 0.1267, -0.0203, 0.0017, 0.1246, -0.0203, 0.0039, 0.1246, -0.0203, 0.0017, 0.1267, -0.0203, 0.0039, 0.1267, -0.0168, -0.0, 0.1294, -0.019, -0.0, 0.1294, -0.0203, -0.0, 0.1294, -0.0168, 0.0017, 0.1294, -0.019, 0.0017, 0.1294, -0.0168, 0.0039, 0.1294, -0.019, 0.0039, 0.1294, -0.0203, 0.0017, 0.1294, -0.0203, 0.0039, 0.1294, -0.0379, -0.0, 0.1072, -0.0407, -0.0, 0.1072, -0.0379, -0.0, 0.1094, -0.0407, -0.0, 0.1094, -0.0379, 0.0017, 0.1072, -0.0379, 0.0039, 0.1072, -0.0379, 0.0017, 0.1094, -0.0379, 0.0039, 0.1094, -0.0379, -0.0, 0.1116, -0.0407, -0.0, 0.1116, -0.0379, -0.0, 0.1137, -0.0407, -0.0, 0.1137, -0.0379, 0.0017, 0.1116, -0.0379, 0.0039, 0.1116, -0.0379, 0.0017, 0.1137,
-0.0379, 0.0039, 0.1137, -0.0429, -0.0, 0.1072, -0.045, -0.0, 0.1072, -0.0429, -0.0, 0.1094, -0.045, -0.0, 0.1094, -0.0472, -0.0, 0.1072, -0.0487, -0.0, 0.1072, -0.0472, -0.0, 0.1094, -0.0487, -0.0, 0.1094, -0.0487, 0.0017, 0.1072, -0.0487, 0.0039, 0.1072, -0.0487, 0.0017, 0.1094, -0.0487, 0.0039, 0.1094, -0.0429, -0.0, 0.1116, -0.045, -0.0, 0.1116, -0.0429, -0.0, 0.1137, -0.045, -0.0, 0.1137, -0.0472, -0.0, 0.1116, -0.0487, -0.0, 0.1116, -0.0472, -0.0, 0.1137, -0.0487, -0.0, 0.1137, -0.0487, 0.0017, 0.1116, -0.0487, 0.0039, 0.1116, -0.0487, 0.0017, 0.1137, -0.0487, 0.0039, 0.1137, -0.0379, -0.0, 0.1159, -0.0407, -0.0, 0.1159, -0.0379, -0.0, 0.1181, -0.0407, -0.0, 0.1181, -0.0379, 0.0017, 0.1159,
-0.0379, 0.0039, 0.1159, -0.0379, 0.0017, 0.1181, -0.0379, 0.0039, 0.1181, -0.0379, -0.0, 0.1202, -0.0407, -0.0, 0.1202, -0.0379, -0.0, 0.1224, -0.0407, -0.0, 0.1224, -0.0379, 0.0017, 0.1202, -0.0379, 0.0039, 0.1202, -0.0379, 0.0017, 0.1224, -0.0379, 0.0039, 0.1224, -0.0429, -0.0, 0.1159, -0.045, -0.0, 0.1159, -0.0429, -0.0, 0.1181, -0.045, -0.0, 0.1181, -0.0472, -0.0, 0.1159, -0.0487, -0.0, 0.1159, -0.0472, -0.0, 0.1181, -0.0487, -0.0, 0.1181, -0.0487, 0.0017, 0.1159, -0.0487, 0.0039, 0.1159, -0.0487, 0.0017, 0.1181, -0.0487, 0.0039, 0.1181, -0.0429, -0.0, 0.1202, -0.045, -0.0, 0.1202, -0.0429, -0.0, 0.1224, -0.045, -0.0, 0.1224, -0.0472, -0.0, 0.1202, -0.0487, -0.0, 0.1202, -0.0472, -0.0, 0.1224,
-0.0487, -0.0, 0.1224, -0.0487, 0.0017, 0.1202, -0.0487, 0.0039, 0.1202, -0.0487, 0.0017, 0.1224, -0.0487, 0.0039, 0.1224, -0.0379, -0.0, 0.1246, -0.0407, -0.0, 0.1246, -0.0379, -0.0, 0.1267, -0.0407, -0.0, 0.1267, -0.0379, 0.0017, 0.1246, -0.0379, 0.0039, 0.1246, -0.0379, 0.0017, 0.1267, -0.0379, 0.0039, 0.1267, -0.0379, -0.0, 0.1289, -0.0407, -0.0, 0.1289, -0.0379, -0.0, 0.1311, -0.0407, -0.0, 0.1311, -0.0379, 0.0017, 0.1289, -0.0379, 0.0039, 0.1289, -0.0379, 0.0017, 0.1311, -0.0379, 0.0039, 0.1311, -0.0429, -0.0, 0.1246, -0.045, -0.0, 0.1246, -0.0429, -0.0, 0.1267, -0.045, -0.0, 0.1267, -0.0472, -0.0, 0.1246, -0.0487, -0.0, 0.1246, -0.0472, -0.0, 0.1267, -0.0487, -0.0, 0.1267, -0.0487, 0.0017, 0.1246,
-0.0487, 0.0039, 0.1246, -0.0487, 0.0017, 0.1267, -0.0487, 0.0039, 0.1267, -0.0429, -0.0, 0.1289, -0.045, -0.0, 0.1289, -0.0429, -0.0, 0.1311, -0.045, -0.0, 0.1311, -0.0472, -0.0, 0.1289, -0.0487, -0.0, 0.1289, -0.0472, -0.0, 0.1311, -0.0487, -0.0, 0.1311, -0.0487, 0.0017, 0.1289, -0.0487, 0.0039, 0.1289, -0.0487, 0.0017, 0.1311, -0.0487, 0.0039, 0.1311, -0.0379, -0.0, 0.1324, -0.0407, -0.0, 0.1324, -0.0379, 0.0017, 0.1324, -0.0407, 0.0017, 0.1324, -0.0379, 0.0039, 0.1324, -0.0407, 0.0039, 0.1324, -0.0429, -0.0, 0.1324, -0.045, -0.0, 0.1324, -0.0472, -0.0, 0.1324, -0.0487, -0.0, 0.1324, -0.0429, 0.0017, 0.1324, -0.045, 0.0017, 0.1324, -0.0429, 0.0039, 0.1324, -0.045, 0.0039, 0.1324, -0.0472, 0.0017, 0.1324,
-0.0487, 0.0017, 0.1324, -0.0472, 0.0039, 0.1324, -0.0487, 0.0039, 0.1324, 0.0005, -0.0, 0.1412, -0.0016, -0.0, 0.1412, 0.0005, -0.0, 0.1441, -0.0016, -0.0, 0.1441, -0.0038, -0.0, 0.1412, -0.006, -0.0, 0.1412, -0.0038, -0.0, 0.1441, -0.006, -0.0, 0.1441, 0.0005, 0.0017, 0.1412, -0.0016, 0.0017, 0.1412, 0.0005, 0.0039, 0.1412, -0.0016, 0.0039, 0.1412, -0.0038, 0.0017, 0.1412, -0.006, 0.0017, 0.1412, -0.0038, 0.0039, 0.1412, -0.006, 0.0039, 0.1412, 0.0005, -0.0, 0.1463, -0.0016, -0.0, 0.1463, 0.0005, -0.0, 0.1484, -0.0016, -0.0, 0.1484, -0.0038, -0.0, 0.1463, -0.006, -0.0, 0.1463, -0.0038, -0.0, 0.1484, -0.006, -0.0, 0.1484, -0.0081, -0.0, 0.1412, -0.0103, -0.0, 0.1412, -0.0081, -0.0, 0.1441,
-0.0103, -0.0, 0.1441, -0.0125, -0.0, 0.1412, -0.0146, -0.0, 0.1412, -0.0125, -0.0, 0.1441, -0.0146, -0.0, 0.1441, -0.0081, 0.0017, 0.1412, -0.0103, 0.0017, 0.1412, -0.0081, 0.0039, 0.1412, -0.0103, 0.0039, 0.1412, -0.0125, 0.0017, 0.1412, -0.0146, 0.0017, 0.1412, -0.0125, 0.0039, 0.1412, -0.0146, 0.0039, 0.1412, -0.0081, -0.0, 0.1463, -0.0103, -0.0, 0.1463, -0.0081, -0.0, 0.1484, -0.0103, -0.0, 0.1484, -0.0125, -0.0, 0.1463, -0.0146, -0.0, 0.1463, -0.0125, -0.0, 0.1484, -0.0146, -0.0, 0.1484, 0.0005, -0.0, 0.1506, -0.0016, -0.0, 0.1506, 0.0005, -0.0, 0.152, -0.0016, -0.0, 0.152, -0.0038, -0.0, 0.1506, -0.006, -0.0, 0.1506, -0.0038, -0.0, 0.152, -0.006, -0.0, 0.152, 0.0005, 0.0017, 0.152,
-0.0016, 0.0017, 0.152, 0.0005, 0.0039, 0.152, -0.0016, 0.0039, 0.152, -0.0038, 0.0017, 0.152, -0.006, 0.0017, 0.152, -0.0038, 0.0039, 0.152, -0.006, 0.0039, 0.152, -0.0081, -0.0, 0.1506, -0.0103, -0.0, 0.1506, -0.0081, -0.0, 0.152, -0.0103, -0.0, 0.152, -0.0125, -0.0, 0.1506, -0.0146, -0.0, 0.1506, -0.0125, -0.0, 0.152, -0.0146, -0.0, 0.152, -0.0081, 0.0017, 0.152, -0.0103, 0.0017, 0.152, -0.0081, 0.0039, 0.152, -0.0103, 0.0039, 0.152, -0.0125, 0.0017, 0.152, -0.0146, 0.0017, 0.152, -0.0125, 0.0039, 0.152, -0.0146, 0.0039, 0.152, -0.0168, -0.0, 0.1412, -0.019, -0.0, 0.1412, -0.0168, -0.0, 0.1441, -0.019, -0.0, 0.1441, -0.0212, -0.0, 0.1412, -0.0233, -0.0, 0.1412, -0.0212, -0.0, 0.1441,
-0.0233, -0.0, 0.1441, -0.0168, 0.0017, 0.1412, -0.019, 0.0017, 0.1412, -0.0168, 0.0039, 0.1412, -0.019, 0.0039, 0.1412, -0.0212, 0.0017, 0.1412, -0.0233, 0.0017, 0.1412, -0.0212, 0.0039, 0.1412, -0.0233, 0.0039, 0.1412, -0.0168, -0.0, 0.1463, -0.019, -0.0, 0.1463, -0.0168, -0.0, 0.1484, -0.019, -0.0, 0.1484, -0.0212, -0.0, 0.1463, -0.0233, -0.0, 0.1463, -0.0212, -0.0, 0.1484, -0.0233, -0.0, 0.1484, -0.0255, -0.0, 0.1412, -0.0277, -0.0, 0.1412, -0.0255, -0.0, 0.1441, -0.0277, -0.0, 0.1441, -0.0298, -0.0, 0.1412, -0.032, -0.0, 0.1412, -0.0298, -0.0, 0.1441, -0.032, -0.0, 0.1441, -0.0255, 0.0017, 0.1412, -0.0277, 0.0017, 0.1412, -0.0255, 0.0039, 0.1412, -0.0277, 0.0039, 0.1412, -0.0298, 0.0017, 0.1412,
-0.032, 0.0017, 0.1412, -0.0298, 0.0039, 0.1412, -0.032, 0.0039, 0.1412, -0.0255, -0.0, 0.1463, -0.0277, -0.0, 0.1463, -0.0255, -0.0, 0.1484, -0.0277, -0.0, 0.1484, -0.0298, -0.0, 0.1463, -0.032, -0.0, 0.1463, -0.0298, -0.0, 0.1484, -0.032, -0.0, 0.1484, -0.0168, -0.0, 0.1506, -0.019, -0.0, 0.1506, -0.0168, -0.0, 0.152, -0.019, -0.0, 0.152, -0.0212, -0.0, 0.1506, -0.0233, -0.0, 0.1506, -0.0212, -0.0, 0.152, -0.0233, -0.0, 0.152, -0.0168, 0.0017, 0.152, -0.019, 0.0017, 0.152, -0.0168, 0.0039, 0.152, -0.019, 0.0039, 0.152, -0.0212, 0.0017, 0.152, -0.0233, 0.0017, 0.152, -0.0212, 0.0039, 0.152, -0.0233, 0.0039, 0.152, -0.0255, -0.0, 0.1506, -0.0277, -0.0, 0.1506, -0.0255, -0.0, 0.152,
-0.0277, -0.0, 0.152, -0.0298, -0.0, 0.1506, -0.032, -0.0, 0.1506, -0.0298, -0.0, 0.152, -0.032, -0.0, 0.152, -0.0255, 0.0017, 0.152, -0.0277, 0.0017, 0.152, -0.0255, 0.0039, 0.152, -0.0277, 0.0039, 0.152, -0.0298, 0.0017, 0.152, -0.032, 0.0017, 0.152, -0.0298, 0.0039, 0.152, -0.032, 0.0039, 0.152, -0.0342, -0.0, 0.1412, -0.0363, -0.0, 0.1412, -0.0342, -0.0, 0.1441, -0.0363, -0.0, 0.1441, -0.0385, -0.0, 0.1412, -0.0407, -0.0, 0.1412, -0.0385, -0.0, 0.1441, -0.0407, -0.0, 0.1441, -0.0342, 0.0017, 0.1412, -0.0363, 0.0017, 0.1412, -0.0342, 0.0039, 0.1412, -0.0363, 0.0039, 0.1412, -0.0385, 0.0017, 0.1412, -0.0407, 0.0017, 0.1412, -0.0385, 0.0039, 0.1412, -0.0407, 0.0039, 0.1412, -0.0342, -0.0, 0.1463,
-0.0363, -0.0, 0.1463, -0.0342, -0.0, 0.1484, -0.0363, -0.0, 0.1484, -0.0385, -0.0, 0.1463, -0.0407, -0.0, 0.1463, -0.0385, -0.0, 0.1484, -0.0407, -0.0, 0.1484, -0.0429, -0.0, 0.1412, -0.045, -0.0, 0.1412, -0.0429, -0.0, 0.1441, -0.045, -0.0, 0.1441, -0.0472, -0.0, 0.1412, -0.0487, -0.0, 0.1412, -0.0472, -0.0, 0.1441, -0.0487, -0.0, 0.1441, -0.0429, 0.0017, 0.1412, -0.045, 0.0017, 0.1412, -0.0429, 0.0039, 0.1412, -0.045, 0.0039, 0.1412, -0.0472, 0.0017, 0.1412, -0.0487, 0.0017, 0.1412, -0.0472, 0.0039, 0.1412, -0.0487, 0.0039, 0.1412, -0.0487, 0.0017, 0.1441, -0.0487, 0.0039, 0.1441, -0.0429, -0.0, 0.1463, -0.045, -0.0, 0.1463, -0.0429, -0.0, 0.1484, -0.045, -0.0, 0.1484, -0.0472, -0.0, 0.1463,
-0.0487, -0.0, 0.1463, -0.0472, -0.0, 0.1484, -0.0487, -0.0, 0.1484, -0.0487, 0.0017, 0.1463, -0.0487, 0.0039, 0.1463, -0.0487, 0.0017, 0.1484, -0.0487, 0.0039, 0.1484, -0.0342, -0.0, 0.1506, -0.0363, -0.0, 0.1506, -0.0342, -0.0, 0.152, -0.0363, -0.0, 0.152, -0.0385, -0.0, 0.1506, -0.0407, -0.0, 0.1506, -0.0385, -0.0, 0.152, -0.0407, -0.0, 0.152, -0.0342, 0.0017, 0.152, -0.0363, 0.0017, 0.152, -0.0342, 0.0039, 0.152, -0.0363, 0.0039, 0.152, -0.0385, 0.0017, 0.152, -0.0407, 0.0017, 0.152, -0.0385, 0.0039, 0.152, -0.0407, 0.0039, 0.152, -0.0429, -0.0, 0.1506, -0.045, -0.0, 0.1506, -0.0429, -0.0, 0.152, -0.045, -0.0, 0.152, -0.0472, -0.0, 0.1506, -0.0487, -0.0, 0.1506, -0.0472, -0.0, 0.152,
-0.0487, -0.0, 0.152, -0.0429, 0.0017, 0.152, -0.045, 0.0017, 0.152, -0.0429, 0.0039, 0.152, -0.045, 0.0039, 0.152, -0.0487, 0.0017, 0.1506, -0.0487, 0.0039, 0.1506, -0.0472, 0.0017, 0.152, -0.0487, 0.0017, 0.152, -0.0472, 0.0039, 0.152, -0.0487, 0.0039, 0.152, 0.0006, -0.0, 0.2139, -0.0015, -0.0, 0.2136, -0.0037, -0.0, 0.2133, -0.0059, -0.0, 0.213, 0.0006, 0.0017, 0.2139, -0.0016, 0.0017, 0.2136, 0.0006, 0.0039, 0.2139, -0.0016, 0.0039, 0.2136, -0.0038, 0.0017, 0.2133, -0.006, 0.0017, 0.213, -0.0038, 0.0039, 0.2133, -0.006, 0.0039, 0.213, 0.0005, -0.0, 0.2157, -0.0016, -0.0, 0.2157, 0.0005, -0.0, 0.2179, -0.0016, -0.0, 0.2179, -0.0038, -0.0, 0.2157, -0.006, -0.0, 0.2157, -0.0038, -0.0, 0.2179,
-0.006, -0.0, 0.2179, -0.0081, -0.0, 0.2128, -0.0103, -0.0, 0.2127, -0.0125, -0.0, 0.2127, -0.0146, -0.0, 0.2127, -0.0081, 0.0017, 0.2128, -0.0103, 0.0017, 0.2127, -0.0081, 0.0039, 0.2128, -0.0103, 0.0039, 0.2127, -0.0125, 0.0017, 0.2127, -0.0146, 0.0017, 0.2127, -0.0125, 0.0039, 0.2127, -0.0146, 0.0039, 0.2127, -0.0081, -0.0, 0.2157, -0.0103, -0.0, 0.2157, -0.0081, -0.0, 0.2179, -0.0103, -0.0, 0.2179, -0.0125, -0.0, 0.2157, -0.0146, -0.0, 0.2157, -0.0125, -0.0, 0.2179, -0.0146, -0.0, 0.2179, 0.0005, -0.0, 0.2201, -0.0016, -0.0, 0.2201, 0.0005, -0.0, 0.2222, -0.0016, -0.0, 0.2222, -0.0038, -0.0, 0.2201, -0.006, -0.0, 0.2201, -0.0038, -0.0, 0.2222, -0.006, -0.0, 0.2222, 0.0005, -0.0, 0.2244,
-0.0016, -0.0, 0.2244, 0.0002, -0.0, 0.2272, -0.0017, -0.0, 0.2259, -0.0038, -0.0, 0.2252, -0.0061, -0.0, 0.2247, -0.0028, -0.0, 0.2256, 0.0002, 0.0017, 0.2271, -0.0016, 0.0017, 0.226, 0.0002, 0.0039, 0.2271, -0.0016, 0.0039, 0.226, -0.0039, 0.0017, 0.2252, -0.006, 0.0017, 0.2248, -0.0039, 0.0039, 0.2252, -0.006, 0.0039, 0.2248, -0.0029, 0.0017, 0.2255, -0.0029, 0.0039, 0.2255, -0.0081, -0.0, 0.2201, -0.0103, -0.0, 0.2201, -0.0081, -0.0, 0.2222, -0.0103, -0.0, 0.2222, -0.0125, -0.0, 0.2201, -0.0146, -0.0, 0.2201, -0.0125, -0.0, 0.2222, -0.0146, -0.0, 0.2222, -0.0082, -0.0, 0.2245, -0.0103, -0.0, 0.2245, -0.0125, -0.0, 0.2245, -0.0146, -0.0, 0.2245, -0.0081, 0.0017, 0.2245, -0.0103, 0.0017, 0.2245,
-0.0081, 0.0039, 0.2245, -0.0103, 0.0039, 0.2245, -0.0125, 0.0017, 0.2245, -0.0146, 0.0017, 0.2245, -0.0125, 0.0039, 0.2245, -0.0146, 0.0039, 0.2245, -0.0168, -0.0, 0.2127, -0.019, -0.0, 0.2127, -0.0212, -0.0, 0.2127, -0.0233, -0.0, 0.2127, -0.0168, 0.0017, 0.2127, -0.019, 0.0017, 0.2127, -0.0168, 0.0039, 0.2127, -0.019, 0.0039, 0.2127, -0.0212, 0.0017, 0.2127, -0.0233, 0.0017, 0.2127, -0.0212, 0.0039, 0.2127, -0.0233, 0.0039, 0.2127, -0.0168, -0.0, 0.2157, -0.019, -0.0, 0.2157, -0.0168, -0.0, 0.2179, -0.019, -0.0, 0.2179, -0.0212, -0.0, 0.2157, -0.0233, -0.0, 0.2157, -0.0212, -0.0, 0.2179, -0.0233, -0.0, 0.2179, -0.0255, -0.0, 0.2127, -0.0277, -0.0, 0.2127, -0.0298, -0.0, 0.2127, -0.032, -0.0, 0.2127,
-0.0255, 0.0017, 0.2127, -0.0277, 0.0017, 0.2127, -0.0255, 0.0039, 0.2127, -0.0277, 0.0039, 0.2127, -0.0298, 0.0017, 0.2127, -0.032, 0.0017, 0.2127, -0.0298, 0.0039, 0.2127, -0.032, 0.0039, 0.2127, -0.0255, -0.0, 0.2157, -0.0277, -0.0, 0.2157, -0.0255, -0.0, 0.2179, -0.0277, -0.0, 0.2179, -0.0298, -0.0, 0.2157, -0.032, -0.0, 0.2157, -0.0298, -0.0, 0.2179, -0.032, -0.0, 0.2179, -0.0168, -0.0, 0.2201, -0.019, -0.0, 0.2201, -0.0168, -0.0, 0.2222, -0.019, -0.0, 0.2222, -0.0212, -0.0, 0.2201, -0.0233, -0.0, 0.2201, -0.0212, -0.0, 0.2222, -0.0233, -0.0, 0.2222, -0.0168, -0.0, 0.2245, -0.019, -0.0, 0.2245, -0.0212, -0.0, 0.2245, -0.0233, -0.0, 0.2245, -0.0168, 0.0017, 0.2245, -0.019, 0.0017, 0.2245,
-0.0168, 0.0039, 0.2245, -0.019, 0.0039, 0.2245, -0.0212, 0.0017, 0.2245, -0.0233, 0.0017, 0.2245, -0.0212, 0.0039, 0.2245, -0.0233, 0.0039, 0.2245, -0.0255, -0.0, 0.2201, -0.0277, -0.0, 0.2201, -0.0255, -0.0, 0.2222, -0.0277, -0.0, 0.2222, -0.0298, -0.0, 0.2201, -0.032, -0.0, 0.2201, -0.0298, -0.0, 0.2222, -0.032, -0.0, 0.2222, -0.0255, -0.0, 0.2245, -0.0277, -0.0, 0.2245, -0.0298, -0.0, 0.2245, -0.032, -0.0, 0.2245, -0.0255, 0.0017, 0.2245, -0.0277, 0.0017, 0.2245, -0.0255, 0.0039, 0.2245, -0.0277, 0.0039, 0.2245, -0.0298, 0.0017, 0.2245, -0.032, 0.0017, 0.2245, -0.0298, 0.0039, 0.2245, -0.032, 0.0039, 0.2245, 0.0014, -0.0, 0.2279, 0.0013, 0.0017, 0.2279, 0.0013, 0.0039, 0.2279, -0.0342, -0.0, 0.2127,
-0.0363, -0.0, 0.2127, -0.0385, -0.0, 0.2127, -0.0407, -0.0, 0.2127, -0.0342, 0.0017, 0.2127, -0.0363, 0.0017, 0.2127, -0.0342, 0.0039, 0.2127, -0.0363, 0.0039, 0.2127, -0.0385, 0.0017, 0.2127, -0.0407, 0.0017, 0.2127, -0.0385, 0.0039, 0.2127, -0.0407, 0.0039, 0.2127, -0.0342, -0.0, 0.2157, -0.0363, -0.0, 0.2157, -0.0342, -0.0, 0.2179, -0.0363, -0.0, 0.2179, -0.0385, -0.0, 0.2157, -0.0407, -0.0, 0.2157, -0.0385, -0.0, 0.2179, -0.0407, -0.0, 0.2179, -0.0429, -0.0, 0.2127, -0.045, -0.0, 0.2127, -0.0472, -0.0, 0.2127, -0.0487, -0.0, 0.2127, -0.0429, 0.0017, 0.2127, -0.045, 0.0017, 0.2127, -0.0429, 0.0039, 0.2127, -0.045, 0.0039, 0.2127, -0.0472, 0.0017, 0.2127, -0.0487, 0.0017, 0.2127, -0.0472, 0.0039, 0.2127,
-0.0487, 0.0039, 0.2127, -0.0429, -0.0, 0.2157, -0.045, -0.0, 0.2157, -0.0429, -0.0, 0.2179, -0.045, -0.0, 0.2179, -0.0472, -0.0, 0.2157, -0.0487, -0.0, 0.2157, -0.0472, -0.0, 0.2179, -0.0487, -0.0, 0.2179, -0.0487, 0.0017, 0.2157, -0.0487, 0.0039, 0.2157, -0.0487, 0.0017, 0.2179, -0.0487, 0.0039, 0.2179, -0.0342, -0.0, 0.2201, -0.0363, -0.0, 0.2201, -0.0342, -0.0, 0.2222, -0.0363, -0.0, 0.2222, -0.0385, -0.0, 0.2201, -0.0407, -0.0, 0.2201, -0.0385, -0.0, 0.2222, -0.0407, -0.0, 0.2222, -0.0342, -0.0, 0.2245, -0.0363, -0.0, 0.2245, -0.0385, -0.0, 0.2245, -0.0407, -0.0, 0.2245, -0.0342, 0.0017, 0.2245, -0.0363, 0.0017, 0.2245, -0.0342, 0.0039, 0.2245, -0.0363, 0.0039, 0.2245, -0.0385, 0.0017, 0.2245,
-0.0407, 0.0017, 0.2245, -0.0385, 0.0039, 0.2245, -0.0407, 0.0039, 0.2245, -0.0429, -0.0, 0.2201, -0.045, -0.0, 0.2201, -0.0429, -0.0, 0.2222, -0.045, -0.0, 0.2222, -0.0472, -0.0, 0.2201, -0.0487, -0.0, 0.2201, -0.0472, -0.0, 0.2222, -0.0487, -0.0, 0.2222, -0.0487, 0.0017, 0.2201, -0.0487, 0.0039, 0.2201, -0.0487, 0.0017, 0.2222, -0.0487, 0.0039, 0.2222, -0.0429, -0.0, 0.2245, -0.045, -0.0, 0.2245, -0.0472, -0.0, 0.2245, -0.0487, -0.0, 0.2245, -0.0429, 0.0017, 0.2245, -0.045, 0.0017, 0.2245, -0.0429, 0.0039, 0.2245, -0.045, 0.0039, 0.2245, -0.0472, 0.0017, 0.2245, -0.0487, 0.0017, 0.2245, -0.0472, 0.0039, 0.2245, -0.0487, 0.0039, 0.2245, 0.0003, -0.0, 0.2765, -0.0011, -0.0, 0.2772, 0.0005, -0.0, 0.2787,
-0.002, -0.0, 0.2776, -0.004, -0.0, 0.2783, -0.0061, -0.0, 0.2787, 0.0006, 0.0017, 0.2763, -0.0012, 0.0017, 0.2773, 0.0006, 0.0039, 0.2763, -0.0012, 0.0039, 0.2773, -0.0023, 0.0017, 0.2777, -0.0023, 0.0039, 0.2777, -0.0038, 0.0017, 0.2782, -0.006, 0.0017, 0.2787, -0.0038, 0.0039, 0.2782, -0.006, 0.0039, 0.2787, -0.0082, -0.0, 0.2791, -0.0103, -0.0, 0.2793, -0.0125, -0.0, 0.2794, -0.0146, -0.0, 0.2794, -0.0081, 0.0017, 0.2791, -0.0103, 0.0017, 0.2793, -0.0081, 0.0039, 0.2791, -0.0103, 0.0039, 0.2793, -0.0125, 0.0017, 0.2794, -0.0146, 0.0017, 0.2794, -0.0125, 0.0039, 0.2794, -0.0146, 0.0039, 0.2794, -0.0168, -0.0, 0.2794, -0.019, -0.0, 0.2794, -0.0212, -0.0, 0.2794, -0.0233, -0.0, 0.2794, -0.0168, 0.0017, 0.2794,
-0.019, 0.0017, 0.2794, -0.0168, 0.0039, 0.2794, -0.019, 0.0039, 0.2794, -0.0212, 0.0017, 0.2794, -0.0233, 0.0017, 0.2794, -0.0212, 0.0039, 0.2794, -0.0233, 0.0039, 0.2794, -0.0255, -0.0, 0.2794, -0.0277, -0.0, 0.2794, -0.0298, -0.0, 0.2794, -0.032, -0.0, 0.2794, -0.0255, 0.0017, 0.2794, -0.0277, 0.0017, 0.2794, -0.0255, 0.0039, 0.2794, -0.0277, 0.0039, 0.2794, -0.0298, 0.0017, 0.2794, -0.032, 0.0017, 0.2794, -0.0298, 0.0039, 0.2794, -0.032, 0.0039, 0.2794, -0.0342, -0.0, 0.2794, -0.0363, -0.0, 0.2794, -0.0385, -0.0, 0.2794, -0.0407, -0.0, 0.2794, -0.0342, 0.0017, 0.2794, -0.0363, 0.0017, 0.2794, -0.0342, 0.0039, 0.2794, -0.0363, 0.0039, 0.2794, -0.0385, 0.0017, 0.2794, -0.0407, 0.0017, 0.2794, -0.0385, 0.0039, 0.2794,
-0.0407, 0.0039, 0.2794, -0.0429, -0.0, 0.2794, -0.045, -0.0, 0.2794, -0.0472, -0.0, 0.2794, -0.0487, -0.0, 0.2794, -0.0429, 0.0017, 0.2794, -0.045, 0.0017, 0.2794, -0.0429, 0.0039, 0.2794, -0.045, 0.0039, 0.2794, -0.0472, 0.0017, 0.2794, -0.0487, 0.0017, 0.2794, -0.0472, 0.0039, 0.2794, -0.0487, 0.0039, 0.2794, 0.0005, -0.0, 0.2808, -0.0016, -0.0, 0.2808, 0.0005, -0.0, 0.283, -0.0016, -0.0, 0.283, -0.0038, -0.0, 0.2808, -0.006, -0.0, 0.2808, -0.0038, -0.0, 0.283, -0.006, -0.0, 0.283, 0.0005, -0.0, 0.2852, -0.0016, -0.0, 0.2852, 0.0005, -0.0, 0.2873, -0.0016, -0.0, 0.2873, -0.0038, -0.0, 0.2852, -0.006, -0.0, 0.2852, -0.0038, -0.0, 0.2873, -0.006, -0.0, 0.2873, -0.0081, -0.0, 0.2808,
-0.0103, -0.0, 0.2808, -0.0081, -0.0, 0.283, -0.0103, -0.0, 0.283, -0.0125, -0.0, 0.2808, -0.0146, -0.0, 0.2808, -0.0125, -0.0, 0.283, -0.0146, -0.0, 0.283, -0.0081, -0.0, 0.2852, -0.0103, -0.0, 0.2852, -0.0081, -0.0, 0.2873, -0.0103, -0.0, 0.2873, -0.0125, -0.0, 0.2852, -0.0146, -0.0, 0.2852, -0.0125, -0.0, 0.2873, -0.0146, -0.0, 0.2873, 0.0007, -0.0, 0.2897, -0.0015, -0.0, 0.2901, -0.0037, -0.0, 0.2903, -0.0057, -0.0, 0.2906, -0.0065, -0.0, 0.2908, 0.0006, 0.0017, 0.2897, -0.0016, 0.0017, 0.2901, 0.0006, 0.0039, 0.2897, -0.0016, 0.0039, 0.2901, -0.0038, 0.0017, 0.2903, -0.0051, 0.0017, 0.2906, -0.0038, 0.0039, 0.2903, -0.0051, 0.0039, 0.2906, -0.0062, 0.0017, 0.2907, -0.0062, 0.0039, 0.2907,
-0.0081, -0.0, 0.2895, -0.0103, -0.0, 0.2895, -0.0081, -0.0, 0.291, -0.0103, -0.0, 0.2911, -0.0125, -0.0, 0.2895, -0.0146, -0.0, 0.2895, -0.0125, -0.0, 0.2912, -0.0146, -0.0, 0.2912, -0.0081, 0.0017, 0.291, -0.0103, 0.0017, 0.2911, -0.0081, 0.0039, 0.291, -0.0103, 0.0039, 0.2911, -0.0125, 0.0017, 0.2912, -0.0146, 0.0017, 0.2912, -0.0125, 0.0039, 0.2912, -0.0146, 0.0039, 0.2912, -0.0168, -0.0, 0.2808, -0.019, -0.0, 0.2808, -0.0168, -0.0, 0.283, -0.019, -0.0, 0.283, -0.0212, -0.0, 0.2808, -0.0233, -0.0, 0.2808, -0.0212, -0.0, 0.283, -0.0233, -0.0, 0.283, -0.0168, -0.0, 0.2852, -0.019, -0.0, 0.2852, -0.0168, -0.0, 0.2873, -0.019, -0.0, 0.2873, -0.0212, -0.0, 0.2852, -0.0233, -0.0, 0.2852,
-0.0212, -0.0, 0.2873, -0.0233, -0.0, 0.2873, -0.0255, -0.0, 0.2808, -0.0277, -0.0, 0.2808, -0.0255, -0.0, 0.283, -0.0277, -0.0, 0.283, -0.0298, -0.0, 0.2808, -0.032, -0.0, 0.2808, -0.0298, -0.0, 0.283, -0.032, -0.0, 0.283, -0.0255, -0.0, 0.2852, -0.0277, -0.0, 0.2852, -0.0255, -0.0, 0.2873, -0.0277, -0.0, 0.2873, -0.0298, -0.0, 0.2852, -0.032, -0.0, 0.2852, -0.0298, -0.0, 0.2873, -0.032, -0.0, 0.2873, -0.0168, -0.0, 0.2895, -0.019, -0.0, 0.2895, -0.0168, -0.0, 0.2912, -0.019, -0.0, 0.2912, -0.0212, -0.0, 0.2895, -0.0233, -0.0, 0.2895, -0.0212, -0.0, 0.2912, -0.0233, -0.0, 0.2912, -0.0168, 0.0017, 0.2912, -0.019, 0.0017, 0.2912, -0.0168, 0.0039, 0.2912, -0.019, 0.0039, 0.2912,
-0.0212, 0.0017, 0.2912, -0.0233, 0.0017, 0.2912, -0.0212, 0.0039, 0.2912, -0.0233, 0.0039, 0.2912, -0.0255, -0.0, 0.2895, -0.0277, -0.0, 0.2895, -0.0255, -0.0, 0.2912, -0.0277, -0.0, 0.2912, -0.0298, -0.0, 0.2895, -0.032, -0.0, 0.2895, -0.0298, -0.0, 0.2912, -0.032, -0.0, 0.2912, -0.0255, 0.0017, 0.2912, -0.0277, 0.0017, 0.2912, -0.0255, 0.0039, 0.2912, -0.0277, 0.0039, 0.2912, -0.0298, 0.0017, 0.2912, -0.032, 0.0017, 0.2912, -0.0298, 0.0039, 0.2912, -0.032, 0.0039, 0.2912, 0.0005, -0.0, 0.3088, -0.0016, -0.0, 0.3088, -0.0038, -0.0, 0.3088, -0.006, -0.0, 0.3088, 0.0005, 0.0017, 0.3088, -0.0016, 0.0017, 0.3088, 0.0005, 0.0039, 0.3088, -0.0016, 0.0039, 0.3088, -0.0038, 0.0017, 0.3088, -0.006, 0.0017, 0.3088,
-0.0038, 0.0039, 0.3088, -0.006, 0.0039, 0.3088, 0.0005, -0.0, 0.3112, -0.0016, -0.0, 0.3112, 0.0005, -0.0, 0.3134, -0.0016, -0.0, 0.3134, -0.0038, -0.0, 0.3112, -0.006, -0.0, 0.3112, -0.0038, -0.0, 0.3134, -0.006, -0.0, 0.3134, -0.0081, -0.0, 0.3088, -0.0103, -0.0, 0.3088, -0.0125, -0.0, 0.3088, -0.0146, -0.0, 0.3088, -0.0081, 0.0017, 0.3088, -0.0103, 0.0017, 0.3088, -0.0081, 0.0039, 0.3088, -0.0103, 0.0039, 0.3088, -0.0125, 0.0017, 0.3088, -0.0146, 0.0017, 0.3088, -0.0125, 0.0039, 0.3088, -0.0146, 0.0039, 0.3088, -0.0081, -0.0, 0.3112, -0.0103, -0.0, 0.3112, -0.0081, -0.0, 0.3134, -0.0103, -0.0, 0.3134, -0.0125, -0.0, 0.3112, -0.0146, -0.0, 0.3112, -0.0125, -0.0, 0.3134, -0.0146, -0.0, 0.3134,
-0.0168, -0.0, 0.3088, -0.019, -0.0, 0.3088, -0.0212, -0.0, 0.3088, -0.0233, -0.0, 0.3088, -0.0168, 0.0017, 0.3088, -0.019, 0.0017, 0.3088, -0.0168, 0.0039, 0.3088, -0.019, 0.0039, 0.3088, -0.0212, 0.0017, 0.3088, -0.0233, 0.0017, 0.3088, -0.0212, 0.0039, 0.3088, -0.0233, 0.0039, 0.3088, -0.0168, -0.0, 0.3112, -0.019, -0.0, 0.3112, -0.0168, -0.0, 0.3134, -0.019, -0.0, 0.3134, -0.0212, -0.0, 0.3112, -0.0233, -0.0, 0.3112, -0.0212, -0.0, 0.3134, -0.0233, -0.0, 0.3134, -0.0255, -0.0, 0.3088, -0.0277, -0.0, 0.3088, -0.0298, -0.0, 0.3088, -0.032, -0.0, 0.3088, -0.0255, 0.0017, 0.3088, -0.0277, 0.0017, 0.3088, -0.0255, 0.0039, 0.3088, -0.0277, 0.0039, 0.3088, -0.0298, 0.0017, 0.3088, -0.032, 0.0017, 0.3088,
-0.0298, 0.0039, 0.3088, -0.032, 0.0039, 0.3088, -0.0255, -0.0, 0.3112, -0.0277, -0.0, 0.3112, -0.0255, -0.0, 0.3134, -0.0277, -0.0, 0.3134, -0.0298, -0.0, 0.3112, -0.032, -0.0, 0.3112, -0.0298, -0.0, 0.3134, -0.032, -0.0, 0.3134, -0.0342, -0.0, 0.2808, -0.0363, -0.0, 0.2808, -0.0342, -0.0, 0.283, -0.0363, -0.0, 0.283, -0.0385, -0.0, 0.2808, -0.0407, -0.0, 0.2808, -0.0385, -0.0, 0.283, -0.0407, -0.0, 0.283, -0.0342, -0.0, 0.2852, -0.0363, -0.0, 0.2852, -0.0342, -0.0, 0.2873, -0.0363, -0.0, 0.2873, -0.0385, -0.0, 0.2852, -0.0407, -0.0, 0.2852, -0.0385, -0.0, 0.2873, -0.0407, -0.0, 0.2873, -0.0429, -0.0, 0.2808, -0.045, -0.0, 0.2808, -0.0429, -0.0, 0.283, -0.045, -0.0, 0.283,
-0.0472, -0.0, 0.2808, -0.0487, -0.0, 0.2808, -0.0472, -0.0, 0.283, -0.0487, -0.0, 0.283, -0.0487, 0.0017, 0.2808, -0.0487, 0.0039, 0.2808, -0.0487, 0.0017, 0.283, -0.0487, 0.0039, 0.283, -0.0429, -0.0, 0.2852, -0.045, -0.0, 0.2852, -0.0429, -0.0, 0.2873, -0.045, -0.0, 0.2873, -0.0472, -0.0, 0.2852, -0.0487, -0.0, 0.2852, -0.0472, -0.0, 0.2873, -0.0487, -0.0, 0.2873, -0.0487, 0.0017, 0.2852, -0.0487, 0.0039, 0.2852, -0.0487, 0.0017, 0.2873, -0.0487, 0.0039, 0.2873, -0.0342, -0.0, 0.2895, -0.0363, -0.0, 0.2895, -0.0342, -0.0, 0.2912, -0.0363, -0.0, 0.2912, -0.0385, -0.0, 0.2895, -0.0407, -0.0, 0.2895, -0.0385, -0.0, 0.2912, -0.0407, -0.0, 0.2912, -0.0342, 0.0017, 0.2912, -0.0363, 0.0017, 0.2912,
-0.0342, 0.0039, 0.2912, -0.0363, 0.0039, 0.2912, -0.0385, 0.0017, 0.2912, -0.0407, 0.0017, 0.2912, -0.0385, 0.0039, 0.2912, -0.0407, 0.0039, 0.2912, -0.0429, -0.0, 0.2895, -0.045, -0.0, 0.2895, -0.0429, -0.0, 0.2912, -0.045, -0.0, 0.2912, -0.0472, -0.0, 0.2895, -0.0487, -0.0, 0.2895, -0.0472, -0.0, 0.2912, -0.0487, -0.0, 0.2912, -0.0429, 0.0017, 0.2912, -0.045, 0.0017, 0.2912, -0.0429, 0.0039, 0.2912, -0.045, 0.0039, 0.2912, -0.0487, 0.0017, 0.2895, -0.0487, 0.0039, 0.2895, -0.0472, 0.0017, 0.2912, -0.0487, 0.0017, 0.2912, -0.0472, 0.0039, 0.2912, -0.0487, 0.0039, 0.2912, -0.0342, -0.0, 0.3088, -0.0363, -0.0, 0.3088, -0.0385, -0.0, 0.3088, -0.0407, -0.0, 0.3088, -0.0342, 0.0017, 0.3088, -0.0363, 0.0017, 0.3088,
-0.0342, 0.0039, 0.3088, -0.0363, 0.0039, 0.3088, -0.0385, 0.0017, 0.3088, -0.0407, 0.0017, 0.3088, -0.0385, 0.0039, 0.3088, -0.0407, 0.0039, 0.3088, -0.0342, -0.0, 0.3112, -0.0363, -0.0, 0.3112, -0.0342, -0.0, 0.3134, -0.0363, -0.0, 0.3134, -0.0385, -0.0, 0.3112, -0.0407, -0.0, 0.3112, -0.0385, -0.0, 0.3134, -0.0407, -0.0, 0.3134, -0.0429, -0.0, 0.3088, -0.045, -0.0, 0.3088, -0.0472, -0.0, 0.3088, -0.0487, -0.0, 0.3088, -0.0429, 0.0017, 0.3088, -0.045, 0.0017, 0.3088, -0.0429, 0.0039, 0.3088, -0.045, 0.0039, 0.3088, -0.0472, 0.0017, 0.3088, -0.0487, 0.0017, 0.3088, -0.0472, 0.0039, 0.3088, -0.0487, 0.0039, 0.3088, -0.0429, -0.0, 0.3112, -0.045, -0.0, 0.3112, -0.0429, -0.0, 0.3134, -0.045, -0.0, 0.3134,
-0.0472, -0.0, 0.3112, -0.0487, -0.0, 0.3112, -0.0472, -0.0, 0.3134, -0.0487, -0.0, 0.3134, -0.0487, 0.0017, 0.3112, -0.0487, 0.0039, 0.3112, -0.0487, 0.0017, 0.3134, -0.0487, 0.0039, 0.3134, 0.0005, -0.0, 0.3155, -0.0016, -0.0, 0.3155, 0.0005, -0.0, 0.3177, -0.0016, -0.0, 0.3177, -0.0038, -0.0, 0.3155, -0.006, -0.0, 0.3155, -0.0038, -0.0, 0.3177, -0.006, -0.0, 0.3177, 0.0005, -0.0, 0.3206, -0.0016, -0.0, 0.3206, -0.0038, -0.0, 0.3206, -0.006, -0.0, 0.3206, 0.0005, 0.0017, 0.3206, -0.0016, 0.0017, 0.3206, 0.0005, 0.0039, 0.3206, -0.0016, 0.0039, 0.3206, -0.0038, 0.0017, 0.3206, -0.006, 0.0017, 0.3206, -0.0038, 0.0039, 0.3206, -0.006, 0.0039, 0.3206, -0.0081, -0.0, 0.3155, -0.0103, -0.0, 0.3155,
-0.0081, -0.0, 0.3177, -0.0103, -0.0, 0.3177, -0.0125, -0.0, 0.3155, -0.0146, -0.0, 0.3155, -0.0125, -0.0, 0.3177, -0.0146, -0.0, 0.3177, -0.0081, -0.0, 0.3206, -0.0103, -0.0, 0.3206, -0.0125, -0.0, 0.3206, -0.0146, -0.0, 0.3206, -0.0081, 0.0017, 0.3206, -0.0103, 0.0017, 0.3206, -0.0081, 0.0039, 0.3206, -0.0103, 0.0039, 0.3206, -0.0125, 0.0017, 0.3206, -0.0146, 0.0017, 0.3206, -0.0125, 0.0039, 0.3206, -0.0146, 0.0039, 0.3206, -0.0168, -0.0, 0.3155, -0.019, -0.0, 0.3155, -0.0168, -0.0, 0.3177, -0.019, -0.0, 0.3177, -0.0212, -0.0, 0.3155, -0.0233, -0.0, 0.3155, -0.0212, -0.0, 0.3177, -0.0233, -0.0, 0.3177, -0.0168, -0.0, 0.3206, -0.019, -0.0, 0.3206, -0.0212, -0.0, 0.3206, -0.0233, -0.0, 0.3206,
-0.0168, 0.0017, 0.3206, -0.019, 0.0017, 0.3206, -0.0168, 0.0039, 0.3206, -0.019, 0.0039, 0.3206, -0.0212, 0.0017, 0.3206, -0.0233, 0.0017, 0.3206, -0.0212, 0.0039, 0.3206, -0.0233, 0.0039, 0.3206, -0.0255, -0.0, 0.3155, -0.0277, -0.0, 0.3155, -0.0255, -0.0, 0.3177, -0.0277, -0.0, 0.3177, -0.0298, -0.0, 0.3155, -0.032, -0.0, 0.3155, -0.0298, -0.0, 0.3177, -0.032, -0.0, 0.3177, -0.0255, -0.0, 0.3206, -0.0277, -0.0, 0.3206, -0.0298, -0.0, 0.3206, -0.032, -0.0, 0.3206, -0.0255, 0.0017, 0.3206, -0.0277, 0.0017, 0.3206, -0.0255, 0.0039, 0.3206, -0.0277, 0.0039, 0.3206, -0.0298, 0.0017, 0.3206, -0.032, 0.0017, 0.3206, -0.0298, 0.0039, 0.3206, -0.032, 0.0039, 0.3206, 0.0005, -0.0, 0.3402, -0.0016, -0.0, 0.3402,
-0.0038, -0.0, 0.3402, -0.006, -0.0, 0.3402, 0.0005, 0.0017, 0.3402, -0.0016, 0.0017, 0.3402, 0.0005, 0.0039, 0.3402, -0.0016, 0.0039, 0.3402, -0.0038, 0.0017, 0.3402, -0.006, 0.0017, 0.3402, -0.0038, 0.0039, 0.3402, -0.006, 0.0039, 0.3402, -0.0081, -0.0, 0.3402, -0.0103, -0.0, 0.3402, -0.0125, -0.0, 0.3402, -0.0146, -0.0, 0.3402, -0.0081, 0.0017, 0.3402, -0.0103, 0.0017, 0.3402, -0.0081, 0.0039, 0.3402, -0.0103, 0.0039, 0.3402, -0.0125, 0.0017, 0.3402, -0.0146, 0.0017, 0.3402, -0.0125, 0.0039, 0.3402, -0.0146, 0.0039, 0.3402, 0.0005, -0.0, 0.3416, -0.0016, -0.0, 0.3416, 0.0005, -0.0, 0.3438, -0.0016, -0.0, 0.3438, -0.0038, -0.0, 0.3416, -0.006, -0.0, 0.3416, -0.0038, -0.0, 0.3438, -0.006, -0.0, 0.3438,
0.0005, -0.0, 0.3459, -0.0016, -0.0, 0.3459, 0.0005, -0.0, 0.3481, -0.0016, -0.0, 0.3481, -0.0038, -0.0, 0.3459, -0.006, -0.0, 0.3459, -0.0038, -0.0, 0.3481, -0.006, -0.0, 0.3481, -0.0081, -0.0, 0.3416, -0.0103, -0.0, 0.3416, -0.0081, -0.0, 0.3438, -0.0103, -0.0, 0.3438, -0.0125, -0.0, 0.3416, -0.0146, -0.0, 0.3416, -0.0125, -0.0, 0.3438, -0.0146, -0.0, 0.3438, -0.0081, -0.0, 0.3459, -0.0103, -0.0, 0.3459, -0.0081, -0.0, 0.3481, -0.0103, -0.0, 0.3481, -0.0125, -0.0, 0.3459, -0.0146, -0.0, 0.3459, -0.0125, -0.0, 0.3481, -0.0146, -0.0, 0.3481, -0.0168, -0.0, 0.3402, -0.019, -0.0, 0.3402, -0.0212, -0.0, 0.3402, -0.0233, -0.0, 0.3402, -0.0168, 0.0017, 0.3402, -0.019, 0.0017, 0.3402,
-0.0168, 0.0039, 0.3402, -0.019, 0.0039, 0.3402, -0.0212, 0.0017, 0.3402, -0.0233, 0.0017, 0.3402, -0.0212, 0.0039, 0.3402, -0.0233, 0.0039, 0.3402, -0.0255, -0.0, 0.3402, -0.0277, -0.0, 0.3402, -0.0298, -0.0, 0.3402, -0.032, -0.0, 0.3402, -0.0255, 0.0017, 0.3402, -0.0277, 0.0017, 0.3402, -0.0255, 0.0039, 0.3402, -0.0277, 0.0039, 0.3402, -0.0298, 0.0017, 0.3402, -0.032, 0.0017, 0.3402, -0.0298, 0.0039, 0.3402, -0.032, 0.0039, 0.3402, -0.0168, -0.0, 0.3416, -0.019, -0.0, 0.3416, -0.0168, -0.0, 0.3438, -0.019, -0.0, 0.3438, -0.0212, -0.0, 0.3416, -0.0233, -0.0, 0.3416, -0.0212, -0.0, 0.3438, -0.0233, -0.0, 0.3438, -0.0168, -0.0, 0.3459, -0.019, -0.0, 0.3459, -0.0168, -0.0, 0.3481, -0.019, -0.0, 0.3481,
-0.0212, -0.0, 0.3459, -0.0233, -0.0, 0.3459, -0.0212, -0.0, 0.3481, -0.0233, -0.0, 0.3481, -0.0255, -0.0, 0.3416, -0.0277, -0.0, 0.3416, -0.0255, -0.0, 0.3438, -0.0277, -0.0, 0.3438, -0.0298, -0.0, 0.3416, -0.032, -0.0, 0.3416, -0.0298, -0.0, 0.3438, -0.032, -0.0, 0.3438, -0.0255, -0.0, 0.3459, -0.0277, -0.0, 0.3459, -0.0255, -0.0, 0.3481, -0.0277, -0.0, 0.3481, -0.0298, -0.0, 0.3459, -0.032, -0.0, 0.3459, -0.0298, -0.0, 0.3481, -0.032, -0.0, 0.3481, -0.0342, -0.0, 0.3155, -0.0363, -0.0, 0.3155, -0.0342, -0.0, 0.3177, -0.0363, -0.0, 0.3177, -0.0385, -0.0, 0.3155, -0.0407, -0.0, 0.3155, -0.0385, -0.0, 0.3177, -0.0407, -0.0, 0.3177, -0.0342, -0.0, 0.3206, -0.0363, -0.0, 0.3206,
-0.0385, -0.0, 0.3206, -0.0407, -0.0, 0.3206, -0.0342, 0.0017, 0.3206, -0.0363, 0.0017, 0.3206, -0.0342, 0.0039, 0.3206, -0.0363, 0.0039, 0.3206, -0.0385, 0.0017, 0.3206, -0.0407, 0.0017, 0.3206, -0.0385, 0.0039, 0.3206, -0.0407, 0.0039, 0.3206, -0.0429, -0.0, 0.3155, -0.045, -0.0, 0.3155, -0.0429, -0.0, 0.3177, -0.045, -0.0, 0.3177, -0.0472, -0.0, 0.3155, -0.0487, -0.0, 0.3155, -0.0472, -0.0, 0.3177, -0.0487, -0.0, 0.3177, -0.0487, 0.0017, 0.3155, -0.0487, 0.0039, 0.3155, -0.0487, 0.0017, 0.3177, -0.0487, 0.0039, 0.3177, -0.0429, -0.0, 0.3206, -0.045, -0.0, 0.3206, -0.0472, -0.0, 0.3206, -0.0487, -0.0, 0.3206, -0.0429, 0.0017, 0.3206, -0.045, 0.0017, 0.3206, -0.0429, 0.0039, 0.3206, -0.045, 0.0039, 0.3206,
-0.0472, 0.0017, 0.3206, -0.0487, 0.0017, 0.3206, -0.0472, 0.0039, 0.3206, -0.0487, 0.0039, 0.3206, -0.0342, -0.0, 0.3402, -0.0363, -0.0, 0.3402, -0.0385, -0.0, 0.3402, -0.0407, -0.0, 0.3402, -0.0342, 0.0017, 0.3402, -0.0363, 0.0017, 0.3402, -0.0342, 0.0039, 0.3402, -0.0363, 0.0039, 0.3402, -0.0385, 0.0017, 0.3402, -0.0407, 0.0017, 0.3402, -0.0385, 0.0039, 0.3402, -0.0407, 0.0039, 0.3402, -0.0429, -0.0, 0.3402, -0.045, -0.0, 0.3402, -0.0472, -0.0, 0.3402, -0.0487, -0.0, 0.3402, -0.0429, 0.0017, 0.3402, -0.045, 0.0017, 0.3402, -0.0429, 0.0039, 0.3402, -0.045, 0.0039, 0.3402, -0.0472, 0.0017, 0.3402, -0.0487, 0.0017, 0.3402, -0.0472, 0.0039, 0.3402, -0.0487, 0.0039, 0.3402, -0.0342, -0.0, 0.3416, -0.0363, -0.0, 0.3416,
-0.0342, -0.0, 0.3438, -0.0363, -0.0, 0.3438, -0.0385, -0.0, 0.3416, -0.0407, -0.0, 0.3416, -0.0385, -0.0, 0.3438, -0.0407, -0.0, 0.3438, -0.0342, -0.0, 0.3459, -0.0363, -0.0, 0.3459, -0.0342, -0.0, 0.3481, -0.0363, -0.0, 0.3481, -0.0385, -0.0, 0.3459, -0.0407, -0.0, 0.3459, -0.0385, -0.0, 0.3481, -0.0407, -0.0, 0.3481, -0.0429, -0.0, 0.3416, -0.045, -0.0, 0.3416, -0.0429, -0.0, 0.3438, -0.045, -0.0, 0.3438, -0.0472, -0.0, 0.3416, -0.0487, -0.0, 0.3416, -0.0472, -0.0, 0.3438, -0.0487, -0.0, 0.3438, -0.0487, 0.0017, 0.3416, -0.0487, 0.0039, 0.3416, -0.0487, 0.0017, 0.3438, -0.0487, 0.0039, 0.3438, -0.0429, -0.0, 0.3459, -0.045, -0.0, 0.3459, -0.0429, -0.0, 0.3481, -0.045, -0.0, 0.3481,
-0.0472, -0.0, 0.3459, -0.0487, -0.0, 0.3459, -0.0472, -0.0, 0.3481, -0.0487, -0.0, 0.3481, -0.0487, 0.0017, 0.3459, -0.0487, 0.0039, 0.3459, -0.0487, 0.0017, 0.3481, -0.0487, 0.0039, 0.3481, 0.0005, -0.0, 0.3503, -0.0016, -0.0, 0.3503, 0.0005, -0.0, 0.352, -0.0016, -0.0, 0.352, -0.0038, -0.0, 0.3503, -0.006, -0.0, 0.3503, -0.0038, -0.0, 0.352, -0.006, -0.0, 0.352, 0.0005, 0.0017, 0.352, -0.0016, 0.0017, 0.352, 0.0005, 0.0039, 0.352, -0.0016, 0.0039, 0.352, -0.0038, 0.0017, 0.352, -0.006, 0.0017, 0.352, -0.0038, 0.0039, 0.352, -0.006, 0.0039, 0.352, -0.0081, -0.0, 0.3503, -0.0103, -0.0, 0.3503, -0.0081, -0.0, 0.352, -0.0103, -0.0, 0.352, -0.0125, -0.0, 0.3503, -0.0146, -0.0, 0.3503,
-0.0125, -0.0, 0.352, -0.0146, -0.0, 0.352, -0.0081, 0.0017, 0.352, -0.0103, 0.0017, 0.352, -0.0081, 0.0039, 0.352, -0.0103, 0.0039, 0.352, -0.0125, 0.0017, 0.352, -0.0146, 0.0017, 0.352, -0.0125, 0.0039, 0.352, -0.0146, 0.0039, 0.352, -0.0168, -0.0, 0.3503, -0.019, -0.0, 0.3503, -0.0168, -0.0, 0.352, -0.019, -0.0, 0.352, -0.0212, -0.0, 0.3503, -0.0233, -0.0, 0.3503, -0.0212, -0.0, 0.352, -0.0233, -0.0, 0.352, -0.0168, 0.0017, 0.352, -0.019, 0.0017, 0.352, -0.0168, 0.0039, 0.352, -0.019, 0.0039, 0.352, -0.0212, 0.0017, 0.352, -0.0233, 0.0017, 0.352, -0.0212, 0.0039, 0.352, -0.0233, 0.0039, 0.352, -0.0255, -0.0, 0.3503, -0.0277, -0.0, 0.3503, -0.0255, -0.0, 0.352, -0.0277, -0.0, 0.352,
-0.0298, -0.0, 0.3503, -0.032, -0.0, 0.3503, -0.0298, -0.0, 0.352, -0.032, -0.0, 0.352, -0.0255, 0.0017, 0.352, -0.0277, 0.0017, 0.352, -0.0255, 0.0039, 0.352, -0.0277, 0.0039, 0.352, -0.0298, 0.0017, 0.352, -0.032, 0.0017, 0.352, -0.0298, 0.0039, 0.352, -0.032, 0.0039, 0.352, -0.0342, -0.0, 0.3503, -0.0363, -0.0, 0.3503, -0.0342, -0.0, 0.352, -0.0363, -0.0, 0.352, -0.0385, -0.0, 0.3503, -0.0407, -0.0, 0.3503, -0.0379, -0.0, 0.352, -0.0407, -0.0, 0.3524, -0.0342, 0.0017, 0.352, -0.0363, 0.0017, 0.352, -0.0342, 0.0039, 0.352, -0.0363, 0.0039, 0.352, -0.0379, 0.0017, 0.352, -0.0379, 0.0039, 0.352, -0.0379, -0.0, 0.3546, -0.0407, -0.0, 0.3546, -0.0379, -0.0, 0.3568, -0.0407, -0.0, 0.3568,
-0.0379, 0.0017, 0.3546, -0.0379, 0.0039, 0.3546, -0.0379, 0.0017, 0.3568, -0.0379, 0.0039, 0.3568, -0.0429, -0.0, 0.3503, -0.045, -0.0, 0.3503, -0.0429, -0.0, 0.3524, -0.045, -0.0, 0.3524, -0.0472, -0.0, 0.3503, -0.0487, -0.0, 0.3503, -0.0472, -0.0, 0.3524, -0.0487, -0.0, 0.3524, -0.0487, 0.0017, 0.3503, -0.0487, 0.0039, 0.3503, -0.0487, 0.0017, 0.3524, -0.0487, 0.0039, 0.3524, -0.0429, -0.0, 0.3546, -0.045, -0.0, 0.3546, -0.0429, -0.0, 0.3568, -0.045, -0.0, 0.3568, -0.0472, -0.0, 0.3546, -0.0487, -0.0, 0.3546, -0.0472, -0.0, 0.3568, -0.0487, -0.0, 0.3568, -0.0487, 0.0017, 0.3546, -0.0487, 0.0039, 0.3546, -0.0487, 0.0017, 0.3568, -0.0487, 0.0039, 0.3568, -0.0379, -0.0, 0.3589, -0.0407, -0.0, 0.3589,
-0.0379, -0.0, 0.3611, -0.0407, -0.0, 0.3611, -0.0379, 0.0017, 0.3589, -0.0379, 0.0039, 0.3589, -0.0379, 0.0017, 0.3611, -0.0379, 0.0039, 0.3611, -0.0379, -0.0, 0.3633, -0.0407, -0.0, 0.3633, -0.0379, -0.0, 0.3655, -0.0407, -0.0, 0.3655, -0.0379, 0.0017, 0.3633, -0.0379, 0.0039, 0.3633, -0.0379, 0.0017, 0.3655, -0.0379, 0.0039, 0.3655, -0.0429, -0.0, 0.3589, -0.045, -0.0, 0.3589, -0.0429, -0.0, 0.3611, -0.045, -0.0, 0.3611, -0.0472, -0.0, 0.3589, -0.0487, -0.0, 0.3589, -0.0472, -0.0, 0.3611, -0.0487, -0.0, 0.3611, -0.0487, 0.0017, 0.3589, -0.0487, 0.0039, 0.3589, -0.0487, 0.0017, 0.3611, -0.0487, 0.0039, 0.3611, -0.0429, -0.0, 0.3633, -0.045, -0.0, 0.3633, -0.0429, -0.0, 0.3655, -0.045, -0.0, 0.3655,
-0.0472, -0.0, 0.3633, -0.0487, -0.0, 0.3633, -0.0472, -0.0, 0.3655, -0.0487, -0.0, 0.3655, -0.0487, 0.0017, 0.3633, -0.0487, 0.0039, 0.3633, -0.0487, 0.0017, 0.3655, -0.0487, 0.0039, 0.3655, -0.0379, -0.0, 0.3676, -0.0407, -0.0, 0.3676, -0.0379, -0.0, 0.3698, -0.0407, -0.0, 0.3698, -0.0379, 0.0017, 0.3676, -0.0379, 0.0039, 0.3676, -0.0379, 0.0017, 0.3698, -0.0379, 0.0039, 0.3698, -0.0379, -0.0, 0.372, -0.0407, -0.0, 0.372, -0.0379, -0.0, 0.3741, -0.0407, -0.0, 0.3741, -0.0379, 0.0017, 0.372, -0.0379, 0.0039, 0.372, -0.0379, 0.0017, 0.3741, -0.0379, 0.0039, 0.3741, -0.0429, -0.0, 0.3676, -0.045, -0.0, 0.3676, -0.0429, -0.0, 0.3698, -0.045, -0.0, 0.3698, -0.0472, -0.0, 0.3676, -0.0487, -0.0, 0.3676,
-0.0472, -0.0, 0.3698, -0.0487, -0.0, 0.3698, -0.0487, 0.0017, 0.3676, -0.0487, 0.0039, 0.3676, -0.0487, 0.0017, 0.3698, -0.0487, 0.0039, 0.3698, -0.0429, -0.0, 0.372, -0.045, -0.0, 0.372, -0.0429, -0.0, 0.3741, -0.045, -0.0, 0.3741, -0.0472, -0.0, 0.372, -0.0487, -0.0, 0.372, -0.0472, -0.0, 0.3741, -0.0487, -0.0, 0.3741, -0.0487, 0.0017, 0.372, -0.0487, 0.0039, 0.372, -0.0487, 0.0017, 0.3741, -0.0487, 0.0039, 0.3741, -0.0379, -0.0, 0.3763, -0.0407, -0.0, 0.3763, -0.0379, -0.0, 0.3784, -0.0407, -0.0, 0.3785, -0.0379, 0.0017, 0.3763, -0.0379, 0.0039, 0.3763, -0.0379, 0.0017, 0.3785, -0.0379, 0.0039, 0.3785, -0.0373, -0.0, 0.3835, -0.0377, -0.0, 0.3806, -0.0407, -0.0, 0.3806, -0.0374, -0.0, 0.3827,
-0.0407, -0.0, 0.3828, -0.0373, 0.0017, 0.3833, -0.0373, 0.0039, 0.3833, -0.0377, 0.0017, 0.3807, -0.0377, 0.0039, 0.3807, -0.0375, 0.0017, 0.3822, -0.0375, 0.0039, 0.3822, -0.0429, -0.0, 0.3763, -0.045, -0.0, 0.3763, -0.0429, -0.0, 0.3785, -0.045, -0.0, 0.3785, -0.0472, -0.0, 0.3763, -0.0487, -0.0, 0.3763, -0.0472, -0.0, 0.3785, -0.0487, -0.0, 0.3785, -0.0487, 0.0017, 0.3763, -0.0487, 0.0039, 0.3763, -0.0487, 0.0017, 0.3785, -0.0487, 0.0039, 0.3785, -0.0429, -0.0, 0.3806, -0.045, -0.0, 0.3806, -0.0429, -0.0, 0.3828, -0.045, -0.0, 0.3828, -0.0472, -0.0, 0.3806, -0.0487, -0.0, 0.3806, -0.0472, -0.0, 0.3828, -0.0487, -0.0, 0.3828, -0.0487, 0.0017, 0.3806, -0.0487, 0.0039, 0.3806, -0.0487, 0.0017, 0.3828,
-0.0487, 0.0039, 0.3828, 0.0003, -0.0, 0.4002, -0.001, -0.0, 0.4009, 0.0006, 0.0017, 0.3999, -0.0011, 0.0017, 0.401, 0.0006, 0.0039, 0.3999, -0.0011, 0.0039, 0.401, -0.033, -0.0, 0.3947, -0.0324, -0.0, 0.3956, -0.033, 0.0017, 0.3947, -0.033, 0.0039, 0.3947, -0.0321, 0.0017, 0.3959, -0.0321, 0.0039, 0.3959, -0.0261, -0.0, 0.4009, -0.0274, -0.0, 0.4001, -0.03, -0.0, 0.3982, -0.031, -0.0, 0.3971, -0.0289, -0.0, 0.399, -0.032, -0.0, 0.4002, -0.026, 0.0017, 0.4009, -0.0277, 0.0017, 0.3999, -0.026, 0.0039, 0.4009, -0.0277, 0.0039, 0.3999, -0.0299, 0.0017, 0.3982, -0.031, 0.0017, 0.3971, -0.0299, 0.0039, 0.3982, -0.031, 0.0039, 0.3971, -0.0288, 0.0017, 0.3991, -0.0288, 0.0039, 0.3991, 0.0005, -0.0, 0.4024,
-0.002, -0.0, 0.4015, 0.0005, -0.0, 0.4045, -0.0016, -0.0, 0.4045, -0.004, -0.0, 0.4025, -0.0056, -0.0, 0.4031, -0.0038, -0.0, 0.4045, -0.0064, -0.0, 0.4033, -0.0021, 0.0017, 0.4016, -0.0021, 0.0039, 0.4016, -0.0038, 0.0017, 0.4024, -0.0058, 0.0017, 0.4032, -0.0038, 0.0039, 0.4024, -0.0058, 0.0039, 0.4032, -0.0069, 0.0017, 0.4035, -0.0069, 0.0039, 0.4035, 0.0005, -0.0, 0.4067, -0.0016, -0.0, 0.4067, 0.0005, -0.0, 0.4089, -0.0016, -0.0, 0.4089, -0.0038, -0.0, 0.4067, -0.006, -0.0, 0.4067, -0.0038, -0.0, 0.4089, -0.006, -0.0, 0.4089, -0.0083, -0.0, 0.4038, -0.0104, -0.0, 0.4039, -0.0125, -0.0, 0.4039, -0.0146, -0.0, 0.4039, -0.0081, 0.0017, 0.4037, -0.0103, 0.0017, 0.4039, -0.0081, 0.0039, 0.4037,
-0.0103, 0.0039, 0.4039, -0.0125, 0.0017, 0.4039, -0.0146, 0.0017, 0.4039, -0.0125, 0.0039, 0.4039, -0.0146, 0.0039, 0.4039, -0.0081, -0.0, 0.4067, -0.0103, -0.0, 0.4067, -0.0081, -0.0, 0.4089, -0.0103, -0.0, 0.4089, -0.0125, -0.0, 0.4067, -0.0146, -0.0, 0.4067, -0.0125, -0.0, 0.4089, -0.0146, -0.0, 0.4089, 0.0005, -0.0, 0.411, -0.0016, -0.0, 0.411, 0.0007, -0.0, 0.4131, -0.0015, -0.0, 0.414, -0.0038, -0.0, 0.411, -0.006, -0.0, 0.411, -0.0038, -0.0, 0.4132, -0.006, -0.0, 0.4132, 0.0006, 0.0017, 0.4132, -0.0015, 0.0017, 0.414, 0.0006, 0.0039, 0.4132, -0.0015, 0.0039, 0.414, -0.0026, -0.0, 0.4144, -0.0037, -0.0, 0.4147, -0.0059, -0.0, 0.4152, -0.0026, 0.0017, 0.4143, -0.0026, 0.0039, 0.4143,
-0.0038, 0.0017, 0.4147, -0.006, 0.0017, 0.4153, -0.0038, 0.0039, 0.4147, -0.006, 0.0039, 0.4153, -0.0081, -0.0, 0.411, -0.0103, -0.0, 0.411, -0.0081, -0.0, 0.4132, -0.0103, -0.0, 0.4132, -0.0125, -0.0, 0.411, -0.0146, -0.0, 0.411, -0.0125, -0.0, 0.4132, -0.0146, -0.0, 0.4132, -0.0081, -0.0, 0.4156, -0.0103, -0.0, 0.4157, -0.0125, -0.0, 0.4157, -0.0146, -0.0, 0.4157, -0.0081, 0.0017, 0.4156, -0.0103, 0.0017, 0.4157, -0.0081, 0.0039, 0.4156, -0.0103, 0.0039, 0.4157, -0.0125, 0.0017, 0.4157, -0.0146, 0.0017, 0.4157, -0.0125, 0.0039, 0.4157, -0.0146, 0.0039, 0.4157, -0.0168, -0.0, 0.4039, -0.0188, -0.0, 0.4037, -0.0215, -0.0, 0.4031, -0.0231, -0.0, 0.4025, -0.0207, -0.0, 0.4033, -0.0233, -0.0, 0.4045,
-0.0168, 0.0017, 0.4039, -0.019, 0.0017, 0.4037, -0.0168, 0.0039, 0.4039, -0.019, 0.0039, 0.4037, -0.0212, 0.0017, 0.4032, -0.0234, 0.0017, 0.4024, -0.0212, 0.0039, 0.4032, -0.0234, 0.0039, 0.4024, -0.0201, 0.0017, 0.4034, -0.0201, 0.0039, 0.4034, -0.0168, -0.0, 0.4067, -0.019, -0.0, 0.4067, -0.0168, -0.0, 0.4089, -0.019, -0.0, 0.4089, -0.0212, -0.0, 0.4067, -0.0233, -0.0, 0.4067, -0.0212, -0.0, 0.4089, -0.0233, -0.0, 0.4089, -0.0251, -0.0, 0.4015, -0.0277, -0.0, 0.4024, -0.0255, -0.0, 0.4045, -0.0277, -0.0, 0.4045, -0.0298, -0.0, 0.4024, -0.032, -0.0, 0.4024, -0.0298, -0.0, 0.4045, -0.032, -0.0, 0.4045, -0.025, 0.0017, 0.4016, -0.025, 0.0039, 0.4016, -0.0255, -0.0, 0.4067, -0.0277, -0.0, 0.4067,
-0.0255, -0.0, 0.4089, -0.0277, -0.0, 0.4089, -0.0298, -0.0, 0.4067, -0.032, -0.0, 0.4067, -0.0298, -0.0, 0.4089, -0.032, -0.0, 0.4089, -0.0168, -0.0, 0.411, -0.019, -0.0, 0.411, -0.0168, -0.0, 0.4132, -0.019, -0.0, 0.4132, -0.0212, -0.0, 0.411, -0.0233, -0.0, 0.411, -0.0212, -0.0, 0.4132, -0.0233, -0.0, 0.4132, -0.0168, -0.0, 0.4157, -0.019, -0.0, 0.4157, -0.0212, -0.0, 0.4155, -0.0234, -0.0, 0.4152, -0.0168, 0.0017, 0.4157, -0.019, 0.0017, 0.4157, -0.0168, 0.0039, 0.4157, -0.019, 0.0039, 0.4157, -0.0212, 0.0017, 0.4155, -0.0233, 0.0017, 0.4152, -0.0212, 0.0039, 0.4155, -0.0233, 0.0039, 0.4152, -0.0255, -0.0, 0.411, -0.0277, -0.0, 0.411, -0.026, -0.0, 0.4144, -0.0279, -0.0, 0.4137,
-0.0298, -0.0, 0.411, -0.0323, -0.0, 0.4117, -0.03, -0.0, 0.4129, -0.0312, -0.0, 0.4123, -0.0265, 0.0017, 0.4142, -0.0277, 0.0017, 0.4138, -0.0265, 0.0039, 0.4142, -0.0277, 0.0039, 0.4138, -0.0323, 0.0017, 0.4117, -0.0323, 0.0039, 0.4117, -0.0299, 0.0017, 0.4129, -0.0312, 0.0017, 0.4123, -0.0299, 0.0039, 0.4129, -0.0312, 0.0039, 0.4123, -0.0251, -0.0, 0.4147, -0.0254, 0.0017, 0.4146, -0.0254, 0.0039, 0.4146, -0.0371, -0.0, 0.385, -0.0367, -0.0, 0.3871, -0.0385, -0.0, 0.385, -0.0407, -0.0, 0.385, -0.0385, -0.0, 0.3872, -0.0407, -0.0, 0.3872, -0.0371, 0.0017, 0.385, -0.0371, 0.0039, 0.385, -0.0366, 0.0017, 0.3872, -0.0366, 0.0039, 0.3872, -0.0361, -0.0, 0.3891, -0.0348, -0.0, 0.3919, -0.0352, -0.0, 0.3909,
-0.0385, -0.0, 0.3893, -0.0407, -0.0, 0.3893, -0.0385, -0.0, 0.3915, -0.0407, -0.0, 0.3915, -0.036, 0.0017, 0.3894, -0.036, 0.0039, 0.3894, -0.0348, 0.0017, 0.3918, -0.0354, 0.0017, 0.3907, -0.0348, 0.0039, 0.3918, -0.0354, 0.0039, 0.3907, -0.0429, -0.0, 0.385, -0.045, -0.0, 0.385, -0.0429, -0.0, 0.3872, -0.045, -0.0, 0.3872, -0.0472, -0.0, 0.385, -0.0487, -0.0, 0.385, -0.0472, -0.0, 0.3872, -0.0486, -0.0, 0.3872, -0.0487, 0.0017, 0.385, -0.0487, 0.0039, 0.385, -0.0486, 0.0017, 0.3872, -0.0486, 0.0039, 0.3872, -0.0429, -0.0, 0.3893, -0.045, -0.0, 0.3893, -0.0429, -0.0, 0.3915, -0.045, -0.0, 0.3915, -0.0483, -0.0, 0.3896, -0.0485, -0.0, 0.3887, -0.0478, -0.0, 0.3917, -0.0482, 0.0017, 0.39,
-0.0484, 0.0017, 0.389, -0.0482, 0.0039, 0.39, -0.0484, 0.0039, 0.389, -0.0479, 0.0017, 0.3915, -0.0479, 0.0039, 0.3915, -0.0338, -0.0, 0.3936, -0.0363, -0.0, 0.3937, -0.0342, -0.0, 0.3958, -0.0363, -0.0, 0.3958, -0.0385, -0.0, 0.3937, -0.0407, -0.0, 0.3937, -0.0385, -0.0, 0.3958, -0.0407, -0.0, 0.3958, -0.0338, 0.0017, 0.3936, -0.0338, 0.0039, 0.3936, -0.0342, -0.0, 0.398, -0.0363, -0.0, 0.398, -0.0342, -0.0, 0.4002, -0.0363, -0.0, 0.4002, -0.0385, -0.0, 0.398, -0.0407, -0.0, 0.398, -0.0385, -0.0, 0.4002, -0.0407, -0.0, 0.4002, -0.0429, -0.0, 0.3937, -0.045, -0.0, 0.3937, -0.0429, -0.0, 0.3958, -0.0461, -0.0, 0.3964, -0.0471, -0.0, 0.3938, -0.0465, -0.0, 0.3954, -0.046, 0.0017, 0.3967,
-0.046, 0.0039, 0.3967, -0.0472, 0.0017, 0.3937, -0.0472, 0.0039, 0.3937, -0.0465, 0.0017, 0.3956, -0.0465, 0.0039, 0.3956, -0.0429, -0.0, 0.398, -0.0453, -0.0, 0.3983, -0.0438, -0.0, 0.4008, -0.0445, -0.0, 0.3997, -0.0454, 0.0017, 0.398, -0.0454, 0.0039, 0.398, -0.0437, 0.0017, 0.4009, -0.0444, 0.0017, 0.3998, -0.0437, 0.0039, 0.4009, -0.0444, 0.0039, 0.3998, -0.0342, -0.0, 0.4024, -0.0363, -0.0, 0.4024, -0.0342, -0.0, 0.4045, -0.0363, -0.0, 0.4045, -0.0385, -0.0, 0.4024, -0.0407, -0.0, 0.4024, -0.0385, -0.0, 0.4045, -0.0409, -0.0, 0.4046, -0.0409, 0.0017, 0.4046, -0.0409, 0.0039, 0.4046, -0.0342, -0.0, 0.4067, -0.0363, -0.0, 0.4067, -0.035, -0.0, 0.41, -0.0367, -0.0, 0.4088, -0.0387, -0.0, 0.4069,
-0.0398, -0.0, 0.4058, -0.0376, -0.0, 0.4079, -0.0352, 0.0017, 0.4099, -0.0364, 0.0017, 0.409, -0.0352, 0.0039, 0.4099, -0.0364, 0.0039, 0.409, -0.0387, 0.0017, 0.4069, -0.0398, 0.0017, 0.4058, -0.0387, 0.0039, 0.4069, -0.0398, 0.0039, 0.4058, -0.0376, 0.0017, 0.4079, -0.0376, 0.0039, 0.4079, -0.0426, -0.0, 0.4026, -0.0418, -0.0, 0.4035, -0.0427, 0.0017, 0.4024, -0.0427, 0.0039, 0.4024, -0.0418, 0.0017, 0.4035, -0.0418, 0.0039, 0.4035, -0.034, -0.0, 0.4107, -0.0341, 0.0017, 0.4106, -0.0341, 0.0039, 0.4106, -0.0153, -0.0, 0.4291, -0.0135, -0.0, 0.4316, -0.0142, -0.0, 0.4304, -0.0153, 0.0017, 0.429, -0.0153, 0.0039, 0.429, -0.0135, 0.0017, 0.4316, -0.0142, 0.0017, 0.4305, -0.0135, 0.0039, 0.4316, -0.0142, 0.0039, 0.4305,
-0.013, -0.0, 0.4326, -0.0146, -0.0, 0.4327, -0.012, -0.0, 0.4346, -0.0146, -0.0, 0.4349, -0.0129, 0.0017, 0.4327, -0.0129, 0.0039, 0.4327, -0.0119, 0.0017, 0.4349, -0.0119, 0.0039, 0.4349, -0.0175, -0.0, 0.427, -0.0186, -0.0, 0.4261, -0.0214, -0.0, 0.4244, -0.0231, -0.0, 0.4238, -0.0204, -0.0, 0.425, -0.0233, -0.0, 0.4262, -0.0175, 0.0017, 0.427, -0.0189, 0.0017, 0.4258, -0.0175, 0.0039, 0.427, -0.0189, 0.0039, 0.4258, -0.0211, 0.0017, 0.4245, -0.0233, 0.0017, 0.4237, -0.0211, 0.0039, 0.4245, -0.0233, 0.0039, 0.4237, -0.0201, 0.0017, 0.4252, -0.0201, 0.0039, 0.4252, -0.0265, -0.0, 0.4229, -0.0277, -0.0, 0.4228, -0.0298, -0.0, 0.4226, -0.0316, -0.0, 0.4228, -0.0265, 0.0017, 0.423, -0.0277, 0.0017, 0.4228,
-0.0265, 0.0039, 0.423, -0.0277, 0.0039, 0.4228, -0.0298, 0.0017, 0.4226, -0.0319, 0.0017, 0.4228, -0.0298, 0.0039, 0.4226, -0.0319, 0.0039, 0.4228, -0.0255, -0.0, 0.4231, -0.0277, -0.0, 0.4241, -0.0255, -0.0, 0.4262, -0.0277, -0.0, 0.4262, -0.0298, -0.0, 0.4241, -0.0324, -0.0, 0.4229, -0.0298, -0.0, 0.4262, -0.032, -0.0, 0.4262, -0.0254, 0.0017, 0.4231, -0.0254, 0.0039, 0.4231, -0.033, 0.0017, 0.423, -0.033, 0.0039, 0.423, -0.0164, -0.0, 0.4279, -0.019, -0.0, 0.4284, -0.0168, -0.0, 0.4306, -0.019, -0.0, 0.4306, -0.0212, -0.0, 0.4284, -0.0233, -0.0, 0.4284, -0.0212, -0.0, 0.4306, -0.0233, -0.0, 0.4306, -0.0164, 0.0017, 0.4279, -0.0164, 0.0039, 0.4279, -0.0168, -0.0, 0.4327, -0.019, -0.0, 0.4327,
-0.0168, -0.0, 0.4349, -0.019, -0.0, 0.4349, -0.0212, -0.0, 0.4327, -0.0233, -0.0, 0.4327, -0.0212, -0.0, 0.4349, -0.0233, -0.0, 0.4349, -0.0255, -0.0, 0.4284, -0.0277, -0.0, 0.4284, -0.0255, -0.0, 0.4306, -0.0277, -0.0, 0.4306, -0.0298, -0.0, 0.4284, -0.032, -0.0, 0.4284, -0.0298, -0.0, 0.4306, -0.032, -0.0, 0.4306, -0.0255, -0.0, 0.4327, -0.0277, -0.0, 0.4327, -0.0255, -0.0, 0.4349, -0.0277, -0.0, 0.4349, -0.0298, -0.0, 0.4327, -0.032, -0.0, 0.4327, -0.0298, -0.0, 0.4349, -0.032, -0.0, 0.4349, -0.0112, -0.0, 0.4377, -0.0109, -0.0, 0.4392, -0.0114, -0.0, 0.4368, -0.0146, -0.0, 0.4371, -0.0125, -0.0, 0.4392, -0.0146, -0.0, 0.4392, -0.0112, 0.0017, 0.4375, -0.0112, 0.0039, 0.4375,
-0.0109, 0.0017, 0.4392, -0.0109, 0.0039, 0.4392, -0.0115, 0.0017, 0.4364, -0.0115, 0.0039, 0.4364, -0.0106, -0.0, 0.4414, -0.0105, -0.0, 0.4436, -0.0125, -0.0, 0.4414, -0.0146, -0.0, 0.4414, -0.0125, -0.0, 0.4436, -0.0146, -0.0, 0.4436, -0.0106, 0.0017, 0.4414, -0.0106, 0.0039, 0.4414, -0.0105, 0.0017, 0.4436, -0.0105, 0.0039, 0.4436, -0.0105, -0.0, 0.4458, -0.0105, -0.0, 0.4479, -0.0125, -0.0, 0.4458, -0.0146, -0.0, 0.4458, -0.0125, -0.0, 0.4479, -0.0146, -0.0, 0.4479, -0.0105, 0.0017, 0.4458, -0.0105, 0.0039, 0.4458, -0.0105, 0.0017, 0.4479, -0.0105, 0.0039, 0.4479, -0.0105, -0.0, 0.4501, -0.0105, -0.0, 0.4523, -0.0125, -0.0, 0.4501, -0.0146, -0.0, 0.4501, -0.0125, -0.0, 0.4523, -0.0146, -0.0, 0.4523,
-0.0105, 0.0017, 0.4501, -0.0105, 0.0039, 0.4501, -0.0105, 0.0017, 0.4523, -0.0105, 0.0039, 0.4523, -0.0168, -0.0, 0.4371, -0.019, -0.0, 0.4371, -0.0168, -0.0, 0.4392, -0.019, -0.0, 0.4392, -0.0212, -0.0, 0.4371, -0.0233, -0.0, 0.4371, -0.0212, -0.0, 0.4392, -0.0242, -0.0, 0.4398, -0.0241, 0.0017, 0.4399, -0.0241, 0.0039, 0.4399, -0.0168, -0.0, 0.4414, -0.019, -0.0, 0.4414, -0.0168, -0.0, 0.4436, -0.019, -0.0, 0.4436, -0.0212, -0.0, 0.4414, -0.0232, -0.0, 0.4416, -0.0212, -0.0, 0.4436, -0.0225, -0.0, 0.4436, -0.0232, 0.0017, 0.4414, -0.0232, 0.0039, 0.4414, -0.0225, 0.0017, 0.4436, -0.0225, 0.0039, 0.4436, -0.026, -0.0, 0.4378, -0.0278, -0.0, 0.4366, -0.025, -0.0, 0.4388, -0.0297, -0.0, 0.436,
-0.0319, -0.0, 0.4366, -0.0261, 0.0017, 0.4377, -0.0276, 0.0017, 0.4367, -0.0261, 0.0039, 0.4377, -0.0276, 0.0039, 0.4367, -0.025, 0.0017, 0.4388, -0.025, 0.0039, 0.4388, -0.0297, 0.0017, 0.436, -0.0321, 0.0017, 0.4367, -0.0297, 0.0039, 0.436, -0.0321, 0.0039, 0.4367, -0.0168, -0.0, 0.4458, -0.019, -0.0, 0.4458, -0.0168, -0.0, 0.4479, -0.019, -0.0, 0.4479, -0.0223, -0.0, 0.4459, -0.0223, -0.0, 0.4453, -0.0222, -0.0, 0.4479, -0.0222, 0.0017, 0.4467, -0.0223, 0.0017, 0.4456, -0.0222, 0.0039, 0.4467, -0.0223, 0.0039, 0.4456, -0.0222, 0.0017, 0.4479, -0.0222, 0.0039, 0.4479, -0.0168, -0.0, 0.4501, -0.019, -0.0, 0.4501, -0.0168, -0.0, 0.4523, -0.019, -0.0, 0.4523, -0.0222, -0.0, 0.4501, -0.0222, -0.0, 0.4523,
-0.0222, 0.0017, 0.4501, -0.0222, 0.0039, 0.4501, -0.0222, 0.0017, 0.4523, -0.0222, 0.0039, 0.4523, -0.0344, -0.0, 0.4232, -0.0366, -0.0, 0.4238, -0.0342, -0.0, 0.4262, -0.0363, -0.0, 0.4262, -0.0383, -0.0, 0.4244, -0.0393, -0.0, 0.4251, -0.041, -0.0, 0.4262, -0.0342, 0.0017, 0.4232, -0.0364, 0.0017, 0.4237, -0.0342, 0.0039, 0.4232, -0.0364, 0.0039, 0.4237, -0.0385, 0.0017, 0.4245, -0.0385, 0.0039, 0.4245, -0.0395, 0.0017, 0.4252, -0.0407, 0.0017, 0.426, -0.0395, 0.0039, 0.4252, -0.0407, 0.0039, 0.426, -0.042, -0.0, 0.427, -0.0421, 0.0017, 0.427, -0.0421, 0.0039, 0.427, -0.0342, -0.0, 0.4284, -0.0363, -0.0, 0.4284, -0.0342, -0.0, 0.4306, -0.0363, -0.0, 0.4306, -0.0385, -0.0, 0.4284, -0.0407, -0.0, 0.4284,
-0.0385, -0.0, 0.4306, -0.0407, -0.0, 0.4306, -0.0342, -0.0, 0.4327, -0.0363, -0.0, 0.4327, -0.0342, -0.0, 0.4349, -0.0363, -0.0, 0.4349, -0.0385, -0.0, 0.4327, -0.0407, -0.0, 0.4327, -0.0385, -0.0, 0.4349, -0.0407, -0.0, 0.4349, -0.0432, -0.0, 0.4281, -0.0442, -0.0, 0.4292, -0.0429, -0.0, 0.4306, -0.045, -0.0, 0.4302, -0.0432, 0.0017, 0.4281, -0.0442, 0.0017, 0.4292, -0.0432, 0.0039, 0.4281, -0.0442, 0.0039, 0.4292, -0.0452, 0.0017, 0.4305, -0.0452, 0.0039, 0.4305, -0.0429, -0.0, 0.4327, -0.0461, -0.0, 0.432, -0.0429, -0.0, 0.4349, -0.045, -0.0, 0.4349, -0.0468, -0.0, 0.433, -0.0475, -0.0, 0.4346, -0.046, 0.0017, 0.4318, -0.046, 0.0039, 0.4318, -0.0467, 0.0017, 0.4329, -0.0467, 0.0039, 0.4329,
-0.0476, 0.0017, 0.4349, -0.0476, 0.0039, 0.4349, -0.0337, -0.0, 0.4377, -0.0363, -0.0, 0.4371, -0.0347, -0.0, 0.4388, -0.0356, -0.0, 0.4398, -0.0385, -0.0, 0.4371, -0.0407, -0.0, 0.4371, -0.0385, -0.0, 0.4392, -0.0407, -0.0, 0.4392, -0.0336, 0.0017, 0.4377, -0.0336, 0.0039, 0.4377, -0.0347, 0.0017, 0.4388, -0.0356, 0.0017, 0.4398, -0.0347, 0.0039, 0.4388, -0.0356, 0.0039, 0.4398, -0.0367, -0.0, 0.4416, -0.0372, -0.0, 0.443, -0.0385, -0.0, 0.4414, -0.0407, -0.0, 0.4414, -0.0374, -0.0, 0.4439, -0.0407, -0.0, 0.4436, -0.0366, 0.0017, 0.4414, -0.0366, 0.0039, 0.4414, -0.0373, 0.0017, 0.4432, -0.0373, 0.0039, 0.4432, -0.0375, 0.0017, 0.4443, -0.0375, 0.0039, 0.4443, -0.0429, -0.0, 0.4371, -0.045, -0.0, 0.4371,
-0.0429, -0.0, 0.4392, -0.045, -0.0, 0.4392, -0.0482, -0.0, 0.4368, -0.0484, -0.0, 0.4378, -0.0472, -0.0, 0.4392, -0.0486, -0.0, 0.4392, -0.0481, 0.0017, 0.4366, -0.0484, 0.0017, 0.4376, -0.0481, 0.0039, 0.4366, -0.0484, 0.0039, 0.4376, -0.0486, 0.0017, 0.4392, -0.0486, 0.0039, 0.4392, -0.0429, -0.0, 0.4414, -0.045, -0.0, 0.4414, -0.0429, -0.0, 0.4436, -0.045, -0.0, 0.4436, -0.0472, -0.0, 0.4414, -0.0487, -0.0, 0.4414, -0.0472, -0.0, 0.4436, -0.0487, -0.0, 0.4436, -0.0487, 0.0017, 0.4414, -0.0487, 0.0039, 0.4414, -0.0487, 0.0017, 0.4436, -0.0487, 0.0039, 0.4436, -0.0378, -0.0, 0.4459, -0.0407, -0.0, 0.4458, -0.0379, -0.0, 0.4479, -0.0407, -0.0, 0.4479, -0.0378, 0.0017, 0.4457, -0.0378, 0.0039, 0.4457,
-0.0379, 0.0017, 0.4479, -0.0379, 0.0039, 0.4479, -0.0379, -0.0, 0.4501, -0.0407, -0.0, 0.4501, -0.0379, -0.0, 0.4523, -0.0407, -0.0, 0.4523, -0.0379, 0.0017, 0.4501, -0.0379, 0.0039, 0.4501, -0.0379, 0.0017, 0.4523, -0.0379, 0.0039, 0.4523, -0.0429, -0.0, 0.4458, -0.045, -0.0, 0.4458, -0.0429, -0.0, 0.4479, -0.045, -0.0, 0.4479, -0.0472, -0.0, 0.4458, -0.0487, -0.0, 0.4458, -0.0472, -0.0, 0.4479, -0.0487, -0.0, 0.4479, -0.0487, 0.0017, 0.4458, -0.0487, 0.0039, 0.4458, -0.0487, 0.0017, 0.4479, -0.0487, 0.0039, 0.4479, -0.0429, -0.0, 0.4501, -0.045, -0.0, 0.4501, -0.0429, -0.0, 0.4523, -0.045, -0.0, 0.4523, -0.0472, -0.0, 0.4501, -0.0487, -0.0, 0.4501, -0.0472, -0.0, 0.4523, -0.0487, -0.0, 0.4523,
-0.0487, 0.0017, 0.4501, -0.0487, 0.0039, 0.4501, -0.0487, 0.0017, 0.4523, -0.0487, 0.0039, 0.4523, -0.0105, -0.0, 0.4544, -0.0105, -0.0, 0.4566, -0.0125, -0.0, 0.4544, -0.0146, -0.0, 0.4544, -0.0125, -0.0, 0.4566, -0.0146, -0.0, 0.4566, -0.0105, 0.0017, 0.4544, -0.0105, 0.0039, 0.4544, -0.0105, 0.0017, 0.4566, -0.0105, 0.0039, 0.4566, -0.0105, -0.0, 0.4588, -0.0105, -0.0, 0.4609, -0.0125, -0.0, 0.4588, -0.0146, -0.0, 0.4588, -0.0125, -0.0, 0.4609, -0.0146, -0.0, 0.4609, -0.0105, 0.0017, 0.4588, -0.0105, 0.0039, 0.4588, -0.0105, 0.0017, 0.4609, -0.0105, 0.0039, 0.4609, -0.0105, -0.0, 0.4631, -0.0105, -0.0, 0.4653, -0.0125, -0.0, 0.4631, -0.0146, -0.0, 0.4631, -0.0125, -0.0, 0.4653, -0.0146, -0.0, 0.4653,
-0.0105, 0.0017, 0.4631, -0.0105, 0.0039, 0.4631, -0.0105, 0.0017, 0.4653, -0.0105, 0.0039, 0.4653, -0.0105, -0.0, 0.4675, -0.0105, -0.0, 0.4696, -0.0125, -0.0, 0.4675, -0.0146, -0.0, 0.4675, -0.0125, -0.0, 0.4696, -0.0146, -0.0, 0.4696, -0.0105, 0.0017, 0.4675, -0.0105, 0.0039, 0.4675, -0.0105, 0.0017, 0.4696, -0.0105, 0.0039, 0.4696, -0.0168, -0.0, 0.4544, -0.019, -0.0, 0.4544, -0.0168, -0.0, 0.4566, -0.019, -0.0, 0.4566, -0.0222, -0.0, 0.4544, -0.0222, -0.0, 0.4566, -0.0222, 0.0017, 0.4544, -0.0222, 0.0039, 0.4544, -0.0222, 0.0017, 0.4566, -0.0222, 0.0039, 0.4566, -0.0168, -0.0, 0.4588, -0.019, -0.0, 0.4588, -0.0168, -0.0, 0.4609, -0.019, -0.0, 0.4609, -0.0222, -0.0, 0.4588, -0.0222, -0.0, 0.4609,
-0.0222, 0.0017, 0.4588, -0.0222, 0.0039, 0.4588, -0.0222, 0.0017, 0.4609, -0.0222, 0.0039, 0.4609, -0.0168, -0.0, 0.4631, -0.019, -0.0, 0.4631, -0.0168, -0.0, 0.4653, -0.019, -0.0, 0.4653, -0.0222, -0.0, 0.4631, -0.0222, -0.0, 0.4653, -0.0222, 0.0017, 0.4631, -0.0222, 0.0039, 0.4631, -0.0222, 0.0017, 0.4653, -0.0222, 0.0039, 0.4653, -0.0168, -0.0, 0.4675, -0.019, -0.0, 0.4675, -0.0168, -0.0, 0.4696, -0.019, -0.0, 0.4696, -0.0222, -0.0, 0.4675, -0.0222, -0.0, 0.4696, -0.0222, 0.0017, 0.4675, -0.0222, 0.0039, 0.4675, -0.0222, 0.0017, 0.4696, -0.0222, 0.0039, 0.4696, -0.0105, -0.0, 0.4718, -0.0105, -0.0, 0.474, -0.0125, -0.0, 0.4718, -0.0146, -0.0, 0.4718, -0.0125, -0.0, 0.474, -0.0146, -0.0, 0.474,
-0.0105, 0.0017, 0.4718, -0.0105, 0.0039, 0.4718, -0.0105, 0.0017, 0.474, -0.0105, 0.0039, 0.474, -0.0104, -0.0, 0.4761, -0.0104, -0.0, 0.4783, -0.0125, -0.0, 0.4761, -0.0146, -0.0, 0.4761, -0.0125, -0.0, 0.4783, -0.0146, -0.0, 0.4783, -0.0104, 0.0017, 0.4761, -0.0104, 0.0039, 0.4761, -0.0104, 0.0017, 0.4783, -0.0104, 0.0039, 0.4783, -0.0067, -0.0, 0.4877, -0.0067, 0.0017, 0.4877, -0.0067, 0.0039, 0.4877, -0.0102, -0.0, 0.4804, -0.0092, -0.0, 0.4837, -0.0096, -0.0, 0.4826, -0.0125, -0.0, 0.4805, -0.0146, -0.0, 0.4805, -0.0125, -0.0, 0.4826, -0.0146, -0.0, 0.4826, -0.0102, 0.0017, 0.4805, -0.0102, 0.0039, 0.4805, -0.0092, 0.0017, 0.4836, -0.0096, 0.0017, 0.4826, -0.0092, 0.0039, 0.4836, -0.0096, 0.0039, 0.4826,
-0.0087, -0.0, 0.4847, -0.0103, -0.0, 0.4848, -0.0076, -0.0, 0.4866, -0.0103, -0.0, 0.487, -0.0125, -0.0, 0.4848, -0.0146, -0.0, 0.4848, -0.0125, -0.0, 0.487, -0.0146, -0.0, 0.487, -0.0087, 0.0017, 0.4849, -0.0087, 0.0039, 0.4849, -0.0076, 0.0017, 0.4866, -0.0076, 0.0039, 0.4866, -0.0168, -0.0, 0.4718, -0.019, -0.0, 0.4718, -0.0168, -0.0, 0.474, -0.019, -0.0, 0.474, -0.0222, -0.0, 0.4718, -0.0222, -0.0, 0.474, -0.0222, 0.0017, 0.4718, -0.0222, 0.0039, 0.4718, -0.0222, 0.0017, 0.474, -0.0222, 0.0039, 0.474, -0.0168, -0.0, 0.4761, -0.019, -0.0, 0.4761, -0.0168, -0.0, 0.4783, -0.019, -0.0, 0.4783, -0.0222, -0.0, 0.4761, -0.0222, -0.0, 0.4783, -0.0222, 0.0017, 0.4761, -0.0222, 0.0039, 0.4761,
-0.0222, 0.0017, 0.4783, -0.0222, 0.0039, 0.4783, -0.0168, -0.0, 0.4805, -0.019, -0.0, 0.4805, -0.0168, -0.0, 0.4826, -0.019, -0.0, 0.4826, -0.0221, -0.0, 0.4806, -0.0219, -0.0, 0.4828, -0.0222, 0.0017, 0.4805, -0.0222, 0.0039, 0.4805, -0.0219, 0.0017, 0.4827, -0.0219, 0.0039, 0.4827, -0.0168, -0.0, 0.4848, -0.019, -0.0, 0.4848, -0.0168, -0.0, 0.487, -0.019, -0.0, 0.487, -0.0214, -0.0, 0.485, -0.0208, -0.0, 0.4871, -0.0214, 0.0017, 0.4848, -0.0214, 0.0039, 0.4848, -0.0208, 0.0017, 0.487, -0.0208, 0.0039, 0.487, -0.0379, -0.0, 0.4544, -0.0407, -0.0, 0.4544, -0.0379, -0.0, 0.4566, -0.0407, -0.0, 0.4566, -0.0379, 0.0017, 0.4544, -0.0379, 0.0039, 0.4544, -0.0379, 0.0017, 0.4566, -0.0379, 0.0039, 0.4566,
-0.0379, -0.0, 0.4588, -0.0407, -0.0, 0.4588, -0.0379, -0.0, 0.4609, -0.0407, -0.0, 0.4609, -0.0379, 0.0017, 0.4588, -0.0379, 0.0039, 0.4588, -0.0379, 0.0017, 0.4609, -0.0379, 0.0039, 0.4609, -0.0429, -0.0, 0.4544, -0.045, -0.0, 0.4544, -0.0429, -0.0, 0.4566, -0.045, -0.0, 0.4566, -0.0472, -0.0, 0.4544, -0.0487, -0.0, 0.4544, -0.0472, -0.0, 0.4566, -0.0487, -0.0, 0.4566, -0.0487, 0.0017, 0.4544, -0.0487, 0.0039, 0.4544, -0.0487, 0.0017, 0.4566, -0.0487, 0.0039, 0.4566, -0.0429, -0.0, 0.4588, -0.045, -0.0, 0.4588, -0.0429, -0.0, 0.4609, -0.045, -0.0, 0.4609, -0.0472, -0.0, 0.4588, -0.0487, -0.0, 0.4588, -0.0472, -0.0, 0.4609, -0.0487, -0.0, 0.4609, -0.0487, 0.0017, 0.4588, -0.0487, 0.0039, 0.4588,
-0.0487, 0.0017, 0.4609, -0.0487, 0.0039, 0.4609, -0.0379, -0.0, 0.4631, -0.0407, -0.0, 0.4631, -0.0379, -0.0, 0.4653, -0.0407, -0.0, 0.4653, -0.0379, 0.0017, 0.4631, -0.0379, 0.0039, 0.4631, -0.0379, 0.0017, 0.4653, -0.0379, 0.0039, 0.4653, -0.0379, -0.0, 0.4675, -0.0407, -0.0, 0.4675, -0.0379, -0.0, 0.4696, -0.0407, -0.0, 0.4696, -0.0379, 0.0017, 0.4675, -0.0379, 0.0039, 0.4675, -0.0379, 0.0017, 0.4696, -0.0379, 0.0039, 0.4696, -0.0429, -0.0, 0.4631, -0.045, -0.0, 0.4631, -0.0429, -0.0, 0.4653, -0.045, -0.0, 0.4653, -0.0472, -0.0, 0.4631, -0.0487, -0.0, 0.4631, -0.0472, -0.0, 0.4653, -0.0487, -0.0, 0.4653, -0.0487, 0.0017, 0.4631, -0.0487, 0.0039, 0.4631, -0.0487, 0.0017, 0.4653, -0.0487, 0.0039, 0.4653,
-0.0429, -0.0, 0.4675, -0.045, -0.0, 0.4675, -0.0429, -0.0, 0.4696, -0.045, -0.0, 0.4696, -0.0472, -0.0, 0.4675, -0.0487, -0.0, 0.4675, -0.0472, -0.0, 0.4696, -0.0487, -0.0, 0.4696, -0.0487, 0.0017, 0.4675, -0.0487, 0.0039, 0.4675, -0.0487, 0.0017, 0.4696, -0.0487, 0.0039, 0.4696, -0.0379, -0.0, 0.4718, -0.0407, -0.0, 0.4718, -0.0379, -0.0, 0.474, -0.0407, -0.0, 0.474, -0.0379, 0.0017, 0.4718, -0.0379, 0.0039, 0.4718, -0.0379, 0.0017, 0.474, -0.0379, 0.0039, 0.474, -0.0379, -0.0, 0.4761, -0.0407, -0.0, 0.4761, -0.0379, -0.0, 0.4783, -0.0407, -0.0, 0.4783, -0.0379, 0.0017, 0.4761, -0.0379, 0.0039, 0.4761, -0.0379, 0.0017, 0.4783, -0.0379, 0.0039, 0.4783, -0.0429, -0.0, 0.4718, -0.045, -0.0, 0.4718,
-0.0429, -0.0, 0.474, -0.045, -0.0, 0.474, -0.0472, -0.0, 0.4718, -0.0487, -0.0, 0.4718, -0.0472, -0.0, 0.474, -0.0487, -0.0, 0.474, -0.0487, 0.0017, 0.4718, -0.0487, 0.0039, 0.4718, -0.0487, 0.0017, 0.474, -0.0487, 0.0039, 0.474, -0.0429, -0.0, 0.4761, -0.045, -0.0, 0.4761, -0.0429, -0.0, 0.4783, -0.045, -0.0, 0.4783, -0.0472, -0.0, 0.4761, -0.0487, -0.0, 0.4761, -0.0472, -0.0, 0.4783, -0.0487, -0.0, 0.4783, -0.0487, 0.0017, 0.4761, -0.0487, 0.0039, 0.4761, -0.0487, 0.0017, 0.4783, -0.0487, 0.0039, 0.4783, -0.0379, -0.0, 0.4805, -0.0407, -0.0, 0.4805, -0.0379, -0.0, 0.4826, -0.0407, -0.0, 0.4826, -0.0379, 0.0017, 0.4805, -0.0379, 0.0039, 0.4805, -0.0379, 0.0017, 0.4826, -0.0379, 0.0039, 0.4826,
-0.0379, -0.0, 0.4848, -0.0407, -0.0, 0.4848, -0.0379, -0.0, 0.487, -0.0407, -0.0, 0.487, -0.0379, 0.0017, 0.4848, -0.0379, 0.0039, 0.4848, -0.0379, 0.0017, 0.487, -0.0379, 0.0039, 0.487, -0.0429, -0.0, 0.4805, -0.045, -0.0, 0.4805, -0.0429, -0.0, 0.4826, -0.045, -0.0, 0.4826, -0.0472, -0.0, 0.4805, -0.0487, -0.0, 0.4805, -0.0472, -0.0, 0.4826, -0.0487, -0.0, 0.4826, -0.0487, 0.0017, 0.4805, -0.0487, 0.0039, 0.4805, -0.0487, 0.0017, 0.4826, -0.0487, 0.0039, 0.4826, -0.0429, -0.0, 0.4848, -0.045, -0.0, 0.4848, -0.0429, -0.0, 0.487, -0.045, -0.0, 0.487, -0.0472, -0.0, 0.4848, -0.0487, -0.0, 0.4848, -0.0472, -0.0, 0.487, -0.0487, -0.0, 0.487, -0.0487, 0.0017, 0.4848, -0.0487, 0.0039, 0.4848,
-0.0487, 0.0017, 0.487, -0.0487, 0.0039, 0.487, 0.0004, -0.0, 0.491, -0.0017, -0.0, 0.4911, -0.0043, -0.0, 0.4897, -0.0057, -0.0, 0.4888, -0.0033, -0.0, 0.4904, -0.006, -0.0, 0.4913, 0.0006, 0.0017, 0.491, -0.0017, 0.0017, 0.4911, 0.0006, 0.0039, 0.491, -0.0017, 0.0039, 0.4911, -0.0042, 0.0017, 0.4898, -0.0057, 0.0017, 0.4888, -0.0042, 0.0039, 0.4898, -0.0057, 0.0039, 0.4888, -0.0032, 0.0017, 0.4905, -0.0032, 0.0039, 0.4905, 0.0005, -0.0, 0.4935, -0.0016, -0.0, 0.4935, 0.0005, -0.0, 0.4957, -0.0016, -0.0, 0.4957, -0.0038, -0.0, 0.4935, -0.006, -0.0, 0.4935, -0.0038, -0.0, 0.4957, -0.006, -0.0, 0.4957, -0.0081, -0.0, 0.4892, -0.0103, -0.0, 0.4892, -0.0081, -0.0, 0.4913, -0.0103, -0.0, 0.4913,
-0.0125, -0.0, 0.4892, -0.0146, -0.0, 0.4892, -0.0125, -0.0, 0.4913, -0.0146, -0.0, 0.4913, -0.0081, -0.0, 0.4935, -0.0103, -0.0, 0.4935, -0.0081, -0.0, 0.4957, -0.0103, -0.0, 0.4957, -0.0125, -0.0, 0.4935, -0.0146, -0.0, 0.4935, -0.0125, -0.0, 0.4957, -0.0154, -0.0, 0.4965, -0.0155, 0.0017, 0.4965, -0.0155, 0.0039, 0.4965, 0.0005, -0.0, 0.4978, -0.0016, -0.0, 0.4978, 0.0005, -0.0, 0.5, -0.0016, -0.0, 0.5, -0.0038, -0.0, 0.4978, -0.006, -0.0, 0.4978, -0.0038, -0.0, 0.5, -0.006, -0.0, 0.5, 0.0005, -0.0, 0.502, -0.0016, -0.0, 0.502, -0.0038, -0.0, 0.5019, -0.006, -0.0, 0.5017, 0.0005, 0.0017, 0.502, -0.0016, 0.0017, 0.502, 0.0005, 0.0039, 0.502, -0.0016, 0.0039, 0.502,
-0.0038, 0.0017, 0.5019, -0.006, 0.0017, 0.5017, -0.0038, 0.0039, 0.5019, -0.006, 0.0039, 0.5017, -0.0081, -0.0, 0.4978, -0.0103, -0.0, 0.4978, -0.0085, -0.0, 0.501, -0.0106, -0.0, 0.5001, -0.013, -0.0, 0.4986, -0.0143, -0.0, 0.4975, -0.0118, -0.0, 0.4993, -0.0087, 0.0017, 0.5009, -0.0103, 0.0017, 0.5002, -0.0087, 0.0039, 0.5009, -0.0103, 0.0039, 0.5002, -0.013, 0.0017, 0.4985, -0.0144, 0.0017, 0.4974, -0.013, 0.0039, 0.4985, -0.0144, 0.0039, 0.4974, -0.0119, 0.0017, 0.4992, -0.0119, 0.0039, 0.4992, -0.0075, -0.0, 0.5013, -0.0076, 0.0017, 0.5013, -0.0076, 0.0039, 0.5013, -0.0168, -0.0, 0.4892, -0.0199, -0.0, 0.4895, -0.0168, -0.0, 0.4913, -0.019, -0.0, 0.4916, -0.0203, -0.0, 0.4884, -0.0198, 0.0017, 0.4896,
-0.0198, 0.0039, 0.4896, -0.0191, 0.0017, 0.4914, -0.0191, 0.0039, 0.4914, -0.0203, 0.0017, 0.4886, -0.0203, 0.0039, 0.4886, -0.0176, -0.0, 0.4939, -0.0182, -0.0, 0.4928, -0.0165, -0.0, 0.4954, -0.0175, 0.0017, 0.494, -0.0182, 0.0017, 0.4929, -0.0175, 0.0039, 0.494, -0.0182, 0.0039, 0.4929, -0.0165, 0.0017, 0.4954, -0.0165, 0.0039, 0.4954, -0.0379, -0.0, 0.4892, -0.0407, -0.0, 0.4892, -0.0379, -0.0, 0.4913, -0.0407, -0.0, 0.4913, -0.0379, 0.0017, 0.4892, -0.0379, 0.0039, 0.4892, -0.0379, 0.0017, 0.4913, -0.0379, 0.0039, 0.4913, -0.0379, -0.0, 0.4935, -0.0407, -0.0, 0.4935, -0.0379, -0.0, 0.4957, -0.0407, -0.0, 0.4957, -0.0379, 0.0017, 0.4935, -0.0379, 0.0039, 0.4935, -0.0379, 0.0017, 0.4957, -0.0379, 0.0039, 0.4957,
-0.0429, -0.0, 0.4892, -0.045, -0.0, 0.4892, -0.0429, -0.0, 0.4913, -0.045, -0.0, 0.4913, -0.0472, -0.0, 0.4892, -0.0487, -0.0, 0.4892, -0.0472, -0.0, 0.4913, -0.0487, -0.0, 0.4913, -0.0487, 0.0017, 0.4892, -0.0487, 0.0039, 0.4892, -0.0487, 0.0017, 0.4913, -0.0487, 0.0039, 0.4913, -0.0429, -0.0, 0.4935, -0.045, -0.0, 0.4935, -0.0429, -0.0, 0.4957, -0.045, -0.0, 0.4957, -0.0472, -0.0, 0.4935, -0.0487, -0.0, 0.4935, -0.0472, -0.0, 0.4957, -0.0487, -0.0, 0.4957, -0.0487, 0.0017, 0.4935, -0.0487, 0.0039, 0.4935, -0.0487, 0.0017, 0.4957, -0.0487, 0.0039, 0.4957, -0.0379, -0.0, 0.498, -0.0407, -0.0, 0.498, -0.0379, 0.0017, 0.498, -0.0407, 0.0017, 0.498, -0.0379, 0.0039, 0.498, -0.0407, 0.0039, 0.498,
-0.0429, -0.0, 0.498, -0.045, -0.0, 0.498, -0.0472, -0.0, 0.498, -0.0487, -0.0, 0.498, -0.0429, 0.0017, 0.498, -0.045, 0.0017, 0.498, -0.0429, 0.0039, 0.498, -0.045, 0.0039, 0.498, -0.0472, 0.0017, 0.498, -0.0487, 0.0017, 0.498, -0.0472, 0.0039, 0.498, -0.0487, 0.0039, 0.498, 0.067, 0.0061, 0.0461, 0.067, 0.0083, 0.0461, 0.067, 0.0061, 0.0486, 0.067, 0.0083, 0.0486, 0.0656, 0.0061, 0.0461, 0.0635, 0.0061, 0.0461, 0.0656, 0.0083, 0.0461, 0.0635, 0.0083, 0.0461, 0.067, 0.01, 0.0461, 0.067, 0.01, 0.0486, 0.0656, 0.01, 0.0461, 0.0635, 0.01, 0.0461, 0.0656, 0.01, 0.0486, 0.0635, 0.01, 0.0486, 0.067, 0.0061, 0.05, 0.067, 0.0083, 0.05, 0.0656, 0.0061, 0.05, 0.0635, 0.0061, 0.05,
0.0656, 0.0083, 0.05, 0.0635, 0.0083, 0.05, 0.067, 0.01, 0.05, 0.0656, 0.01, 0.05, 0.0635, 0.01, 0.05, 0.0613, 0.0061, 0.0461, 0.0591, 0.0061, 0.0461, 0.0613, 0.0083, 0.0461, 0.0591, 0.0083, 0.0461, 0.057, 0.0061, 0.0461, 0.0548, 0.0061, 0.0461, 0.057, 0.0083, 0.0461, 0.0548, 0.0083, 0.0461, 0.0613, 0.01, 0.0461, 0.0591, 0.01, 0.0461, 0.0613, 0.01, 0.0486, 0.0591, 0.01, 0.0486, 0.057, 0.01, 0.0461, 0.0548, 0.01, 0.0461, 0.057, 0.01, 0.0486, 0.0548, 0.01, 0.0486, 0.0613, 0.0061, 0.05, 0.0591, 0.0061, 0.05, 0.0613, 0.0083, 0.05, 0.0591, 0.0083, 0.05, 0.057, 0.0061, 0.05, 0.0548, 0.0061, 0.05, 0.057, 0.0083, 0.05, 0.0548, 0.0083, 0.05, 0.0562, 0.0061, 0.0529,
0.0548, 0.0061, 0.0529, 0.0562, 0.0083, 0.0529, 0.0548, 0.0083, 0.0529, 0.0613, 0.01, 0.05, 0.0591, 0.01, 0.05, 0.057, 0.01, 0.05, 0.0548, 0.01, 0.05, 0.0562, 0.01, 0.0529, 0.0548, 0.01, 0.0529, 0.0526, 0.0061, 0.0461, 0.0505, 0.0061, 0.0461, 0.0526, 0.0083, 0.0461, 0.0505, 0.0083, 0.0461, 0.0483, 0.0061, 0.0461, 0.0461, 0.0061, 0.0461, 0.0483, 0.0083, 0.0461, 0.0461, 0.0083, 0.0461, 0.0526, 0.01, 0.0461, 0.0505, 0.01, 0.0461, 0.0526, 0.01, 0.0486, 0.0505, 0.01, 0.0486, 0.0483, 0.01, 0.0461, 0.0461, 0.01, 0.0461, 0.0483, 0.01, 0.0486, 0.0461, 0.01, 0.0486, 0.0526, 0.0061, 0.05, 0.0505, 0.0061, 0.05, 0.0526, 0.0083, 0.05, 0.0505, 0.0083, 0.05, 0.0523, 0.0061, 0.0529,
0.0523, 0.0083, 0.0529, 0.0483, 0.0061, 0.05, 0.0454, 0.0061, 0.05, 0.0483, 0.0083, 0.05, 0.0454, 0.0083, 0.05, 0.0454, 0.0061, 0.053, 0.0454, 0.0083, 0.053, 0.0526, 0.01, 0.05, 0.0505, 0.01, 0.05, 0.0523, 0.01, 0.0529, 0.0483, 0.01, 0.05, 0.0454, 0.01, 0.05, 0.0454, 0.01, 0.053, 0.0439, 0.0061, 0.0461, 0.0425, 0.0061, 0.0461, 0.0439, 0.0083, 0.0461, 0.0425, 0.0083, 0.0461, 0.0425, 0.0061, 0.0486, 0.0425, 0.0083, 0.0486, 0.0439, 0.01, 0.0461, 0.0425, 0.01, 0.0461, 0.0439, 0.01, 0.0486, 0.0425, 0.01, 0.0486, 0.0425, 0.0061, 0.0508, 0.0425, 0.0083, 0.0508, 0.0425, 0.0061, 0.053, 0.0425, 0.0083, 0.053, 0.0439, 0.01, 0.0508, 0.0425, 0.01, 0.0508, 0.0439, 0.01, 0.053,
0.0425, 0.01, 0.053, 0.0562, 0.0061, 0.0551, 0.0562, 0.0083, 0.0551, 0.0562, 0.0061, 0.0573, 0.0562, 0.0083, 0.0573, 0.0562, 0.01, 0.0551, 0.0548, 0.01, 0.0551, 0.0562, 0.01, 0.0573, 0.0548, 0.01, 0.0573, 0.0562, 0.0061, 0.0595, 0.0562, 0.0083, 0.0595, 0.0562, 0.0061, 0.0616, 0.0562, 0.0083, 0.0616, 0.0562, 0.01, 0.0595, 0.0548, 0.01, 0.0595, 0.0562, 0.01, 0.0616, 0.0548, 0.01, 0.0616, 0.0562, 0.0061, 0.0638, 0.0562, 0.0083, 0.0638, 0.0562, 0.0061, 0.0657, 0.0548, 0.0061, 0.0657, 0.0562, 0.0083, 0.0657, 0.0548, 0.0083, 0.0657, 0.0562, 0.01, 0.0638, 0.0548, 0.01, 0.0638, 0.0562, 0.01, 0.0657, 0.0548, 0.01, 0.0657, 0.0523, 0.0061, 0.0551, 0.0523, 0.0083, 0.0551, 0.0523, 0.0061, 0.0573,
0.0523, 0.0083, 0.0573, 0.0454, 0.0061, 0.0551, 0.0454, 0.0083, 0.0551, 0.0454, 0.0061, 0.0573, 0.0454, 0.0083, 0.0573, 0.0523, 0.01, 0.0551, 0.0523, 0.01, 0.0573, 0.0454, 0.01, 0.0551, 0.0454, 0.01, 0.0573, 0.0523, 0.0061, 0.0595, 0.0523, 0.0083, 0.0595, 0.0523, 0.0061, 0.0616, 0.0523, 0.0083, 0.0616, 0.0454, 0.0061, 0.0595, 0.0454, 0.0083, 0.0595, 0.0454, 0.0061, 0.0616, 0.0454, 0.0083, 0.0616, 0.0523, 0.01, 0.0595, 0.0523, 0.01, 0.0616, 0.0454, 0.01, 0.0595, 0.0454, 0.01, 0.0616, 0.0425, 0.0061, 0.0551, 0.0425, 0.0083, 0.0551, 0.0425, 0.0061, 0.0573, 0.0425, 0.0083, 0.0573, 0.0439, 0.01, 0.0551, 0.0425, 0.01, 0.0551, 0.0439, 0.01, 0.0573, 0.0425, 0.01, 0.0573, 0.0425, 0.0061, 0.0595,
0.0425, 0.0083, 0.0595, 0.0425, 0.0061, 0.0616, 0.0425, 0.0083, 0.0616, 0.0439, 0.01, 0.0595, 0.0425, 0.01, 0.0595, 0.0439, 0.01, 0.0616, 0.0425, 0.01, 0.0616, 0.0523, 0.0061, 0.0638, 0.0523, 0.0083, 0.0638, 0.0523, 0.0061, 0.0657, 0.0523, 0.0083, 0.0657, 0.0454, 0.0061, 0.0638, 0.0454, 0.0083, 0.0638, 0.0454, 0.0061, 0.0667, 0.0454, 0.0083, 0.0667, 0.0523, 0.01, 0.0638, 0.0523, 0.01, 0.0657, 0.0454, 0.01, 0.0638, 0.0454, 0.01, 0.0667, 0.0425, 0.0061, 0.0638, 0.0425, 0.0083, 0.0638, 0.0439, 0.0061, 0.0667, 0.0425, 0.0061, 0.0667, 0.0439, 0.0083, 0.0667, 0.0425, 0.0083, 0.0667, 0.0439, 0.01, 0.0638, 0.0425, 0.01, 0.0638, 0.0439, 0.01, 0.0667, 0.0425, 0.01, 0.0667, 0.0625, 0.0061, 0.0887,
0.0625, 0.0083, 0.0887, 0.0624, 0.01, 0.0887, 0.0614, 0.0061, 0.0879, 0.0592, 0.0061, 0.0869, 0.0614, 0.0083, 0.0879, 0.0592, 0.0083, 0.0869, 0.0568, 0.0061, 0.0864, 0.0548, 0.0061, 0.0863, 0.0568, 0.0083, 0.0864, 0.0548, 0.0083, 0.0863, 0.0579, 0.0061, 0.0866, 0.0579, 0.0083, 0.0866, 0.0616, 0.01, 0.088, 0.0594, 0.01, 0.087, 0.0565, 0.01, 0.0864, 0.0548, 0.01, 0.0863, 0.0573, 0.01, 0.0865, 0.0548, 0.01, 0.0877, 0.0526, 0.0061, 0.0863, 0.0511, 0.0061, 0.0865, 0.0526, 0.0083, 0.0863, 0.0511, 0.0083, 0.0865, 0.0501, 0.0061, 0.0867, 0.0501, 0.0083, 0.0867, 0.0482, 0.0061, 0.0874, 0.0468, 0.0061, 0.0884, 0.0482, 0.0083, 0.0874, 0.0468, 0.0083, 0.0884, 0.0526, 0.01, 0.0863, 0.0513, 0.01, 0.0865,
0.0526, 0.01, 0.0877, 0.0503, 0.01, 0.0867, 0.048, 0.01, 0.0875, 0.0468, 0.01, 0.0884, 0.0646, 0.0061, 0.0909, 0.0637, 0.0061, 0.0897, 0.0646, 0.0083, 0.0909, 0.0637, 0.0083, 0.0897, 0.0653, 0.0061, 0.092, 0.0653, 0.0083, 0.092, 0.0646, 0.01, 0.0909, 0.0637, 0.01, 0.0897, 0.0652, 0.01, 0.0918, 0.0635, 0.01, 0.092, 0.0668, 0.0061, 0.097, 0.0668, 0.0083, 0.097, 0.0662, 0.0061, 0.0942, 0.0662, 0.0083, 0.0942, 0.0666, 0.0061, 0.0959, 0.0627, 0.0061, 0.0964, 0.0666, 0.0083, 0.0959, 0.0627, 0.0083, 0.0964, 0.0668, 0.01, 0.0972, 0.0661, 0.01, 0.0939, 0.0635, 0.01, 0.0942, 0.0667, 0.01, 0.0962, 0.0627, 0.01, 0.0963, 0.0583, 0.0061, 0.0908, 0.0583, 0.0083, 0.0908, 0.0608, 0.0061, 0.0926,
0.0594, 0.0061, 0.0914, 0.0608, 0.0083, 0.0926, 0.0594, 0.0083, 0.0914, 0.057, 0.0061, 0.0904, 0.0548, 0.0061, 0.0903, 0.057, 0.0083, 0.0904, 0.0548, 0.0083, 0.0903, 0.0613, 0.01, 0.0899, 0.0586, 0.01, 0.091, 0.0608, 0.01, 0.0926, 0.0595, 0.01, 0.0915, 0.0568, 0.01, 0.0904, 0.0548, 0.01, 0.0902, 0.0621, 0.0061, 0.0942, 0.0621, 0.0083, 0.0942, 0.0624, 0.0061, 0.0953, 0.0624, 0.0083, 0.0953, 0.0621, 0.01, 0.0942, 0.0624, 0.01, 0.0953, 0.0669, 0.0061, 0.0985, 0.0669, 0.0083, 0.0985, 0.067, 0.0061, 0.1007, 0.067, 0.0083, 0.1007, 0.063, 0.0061, 0.0985, 0.063, 0.0083, 0.0985, 0.0631, 0.0061, 0.1007, 0.0631, 0.0083, 0.1007, 0.0669, 0.01, 0.0985, 0.067, 0.01, 0.1007, 0.0656, 0.01, 0.0985,
0.063, 0.01, 0.0986, 0.0656, 0.01, 0.1007, 0.0631, 0.01, 0.1007, 0.067, 0.0061, 0.1029, 0.067, 0.0083, 0.1029, 0.067, 0.0061, 0.105, 0.067, 0.0083, 0.105, 0.0631, 0.0061, 0.1029, 0.0631, 0.0083, 0.1029, 0.063, 0.0061, 0.105, 0.063, 0.0083, 0.105, 0.067, 0.01, 0.1029, 0.0669, 0.01, 0.105, 0.0656, 0.01, 0.1029, 0.0631, 0.01, 0.1029, 0.0656, 0.01, 0.105, 0.063, 0.01, 0.105, 0.0526, 0.0061, 0.0904, 0.0512, 0.0061, 0.0908, 0.0526, 0.0083, 0.0904, 0.0512, 0.0083, 0.0908, 0.05, 0.0061, 0.0912, 0.05, 0.0083, 0.0912, 0.0456, 0.0061, 0.0894, 0.0456, 0.0083, 0.0894, 0.0485, 0.0061, 0.0923, 0.0485, 0.0083, 0.0923, 0.0528, 0.01, 0.0904, 0.0508, 0.01, 0.0909, 0.0499, 0.01, 0.0913,
0.0483, 0.01, 0.0899, 0.0456, 0.01, 0.0894, 0.0485, 0.01, 0.0922, 0.0461, 0.01, 0.092, 0.0474, 0.0061, 0.0935, 0.0468, 0.0061, 0.0946, 0.0474, 0.0083, 0.0935, 0.0468, 0.0083, 0.0946, 0.0461, 0.0061, 0.0963, 0.0461, 0.0083, 0.0963, 0.0475, 0.01, 0.0934, 0.0469, 0.01, 0.0945, 0.046, 0.01, 0.0965, 0.0447, 0.0061, 0.0905, 0.0447, 0.0083, 0.0905, 0.0437, 0.0061, 0.092, 0.0437, 0.0083, 0.092, 0.0447, 0.01, 0.0905, 0.0438, 0.01, 0.0917, 0.043, 0.0061, 0.0937, 0.0427, 0.0061, 0.0948, 0.043, 0.0083, 0.0937, 0.0427, 0.0083, 0.0948, 0.0425, 0.0061, 0.0964, 0.0425, 0.0083, 0.0964, 0.0429, 0.01, 0.0939, 0.0427, 0.01, 0.0949, 0.0439, 0.01, 0.0964, 0.0425, 0.01, 0.0963, 0.0456, 0.0061, 0.0985,
0.0456, 0.0083, 0.0985, 0.0454, 0.0061, 0.1007, 0.0454, 0.0083, 0.1007, 0.0456, 0.01, 0.0986, 0.0454, 0.01, 0.1007, 0.0455, 0.0061, 0.1029, 0.0455, 0.0083, 0.1029, 0.0457, 0.0061, 0.1051, 0.0457, 0.0083, 0.1051, 0.0455, 0.01, 0.1029, 0.0457, 0.01, 0.105, 0.0425, 0.0061, 0.0985, 0.0425, 0.0083, 0.0985, 0.0425, 0.0061, 0.1007, 0.0425, 0.0083, 0.1007, 0.0439, 0.01, 0.0985, 0.0425, 0.01, 0.0985, 0.0439, 0.01, 0.1007, 0.0425, 0.01, 0.1007, 0.0425, 0.0061, 0.1029, 0.0425, 0.0083, 0.1029, 0.0425, 0.0061, 0.105, 0.0425, 0.0083, 0.105, 0.0439, 0.01, 0.1029, 0.0425, 0.01, 0.1029, 0.0439, 0.01, 0.105, 0.0425, 0.01, 0.105, 0.0219, 0.0061, 0.0716, 0.0201, 0.0061, 0.0716, 0.0219, 0.0083, 0.0716,
0.0201, 0.0083, 0.0716, 0.0219, 0.0061, 0.0747, 0.0219, 0.0083, 0.0747, 0.0219, 0.01, 0.0716, 0.0201, 0.01, 0.0716, 0.0219, 0.01, 0.0747, 0.0201, 0.01, 0.0747, 0.0219, 0.0061, 0.0768, 0.0219, 0.0083, 0.0768, 0.0219, 0.0061, 0.079, 0.0219, 0.0083, 0.079, 0.0219, 0.01, 0.0768, 0.0201, 0.01, 0.0768, 0.0219, 0.01, 0.079, 0.0201, 0.01, 0.079, 0.0219, 0.0061, 0.0812, 0.0219, 0.0083, 0.0812, 0.0219, 0.0061, 0.0833, 0.0201, 0.0061, 0.0833, 0.0219, 0.0083, 0.0833, 0.0201, 0.0083, 0.0833, 0.0219, 0.01, 0.0812, 0.0201, 0.01, 0.0812, 0.0219, 0.01, 0.0833, 0.0201, 0.01, 0.0833, 0.0179, 0.0061, 0.0716, 0.0157, 0.0061, 0.0716, 0.0179, 0.0083, 0.0716, 0.0157, 0.0083, 0.0716, 0.0136, 0.0061, 0.0716,
0.0114, 0.0061, 0.0716, 0.0136, 0.0083, 0.0716, 0.0114, 0.0083, 0.0716, 0.0179, 0.01, 0.0716, 0.0157, 0.01, 0.0716, 0.0179, 0.01, 0.0747, 0.0157, 0.01, 0.0747, 0.0136, 0.01, 0.0716, 0.0114, 0.01, 0.0716, 0.0136, 0.01, 0.0747, 0.0114, 0.01, 0.0747, 0.0179, 0.01, 0.0768, 0.0157, 0.01, 0.0768, 0.0179, 0.01, 0.079, 0.0157, 0.01, 0.079, 0.0136, 0.01, 0.0768, 0.0114, 0.01, 0.0768, 0.0136, 0.01, 0.079, 0.0114, 0.01, 0.079, 0.0092, 0.0061, 0.0716, 0.0071, 0.0061, 0.0716, 0.0092, 0.0083, 0.0716, 0.0071, 0.0083, 0.0716, 0.0049, 0.0061, 0.0716, 0.0027, 0.0061, 0.0716, 0.0049, 0.0083, 0.0716, 0.0027, 0.0083, 0.0716, 0.0092, 0.01, 0.0716, 0.0071, 0.01, 0.0716, 0.0092, 0.01, 0.0747,
0.0071, 0.01, 0.0747, 0.0049, 0.01, 0.0716, 0.0027, 0.01, 0.0716, 0.0049, 0.01, 0.0747, 0.0027, 0.01, 0.0747, 0.0092, 0.01, 0.0768, 0.0071, 0.01, 0.0768, 0.0092, 0.01, 0.079, 0.0071, 0.01, 0.079, 0.0049, 0.01, 0.0768, 0.0027, 0.01, 0.0768, 0.0049, 0.01, 0.079, 0.0027, 0.01, 0.079, 0.0179, 0.0061, 0.0833, 0.0157, 0.0061, 0.0833, 0.0179, 0.0083, 0.0833, 0.0157, 0.0083, 0.0833, 0.0136, 0.0061, 0.0833, 0.0114, 0.0061, 0.0833, 0.0136, 0.0083, 0.0833, 0.0114, 0.0083, 0.0833, 0.0179, 0.01, 0.0812, 0.0157, 0.01, 0.0812, 0.0179, 0.01, 0.0833, 0.0157, 0.01, 0.0833, 0.0136, 0.01, 0.0812, 0.0114, 0.01, 0.0812, 0.0136, 0.01, 0.0833, 0.0114, 0.01, 0.0833, 0.0092, 0.0061, 0.0833,
0.0071, 0.0061, 0.0833, 0.0092, 0.0083, 0.0833, 0.0071, 0.0083, 0.0833, 0.0049, 0.0061, 0.0833, 0.0027, 0.0061, 0.0833, 0.0049, 0.0083, 0.0833, 0.0027, 0.0083, 0.0833, 0.0092, 0.01, 0.0812, 0.0071, 0.01, 0.0812, 0.0092, 0.01, 0.0833, 0.0071, 0.01, 0.0833, 0.0049, 0.01, 0.0812, 0.0027, 0.01, 0.0812, 0.0049, 0.01, 0.0833, 0.0027, 0.01, 0.0833, 0.0668, 0.0061, 0.1066, 0.0668, 0.0083, 0.1066, 0.0666, 0.0061, 0.1077, 0.0627, 0.0061, 0.1071, 0.0666, 0.0083, 0.1077, 0.0627, 0.0083, 0.1071, 0.0662, 0.0061, 0.1094, 0.0662, 0.0083, 0.1094, 0.0668, 0.01, 0.1065, 0.0667, 0.01, 0.1074, 0.0626, 0.01, 0.1072, 0.0661, 0.01, 0.1097, 0.0635, 0.01, 0.1094, 0.0652, 0.0061, 0.1115, 0.0645, 0.0061, 0.1125,
0.0652, 0.0083, 0.1115, 0.0645, 0.0083, 0.1125, 0.0635, 0.0061, 0.1138, 0.0635, 0.0083, 0.1138, 0.0653, 0.01, 0.1113, 0.0646, 0.01, 0.1123, 0.0636, 0.01, 0.1138, 0.0624, 0.0061, 0.1082, 0.0624, 0.0083, 0.1082, 0.062, 0.0061, 0.1094, 0.062, 0.0083, 0.1094, 0.0623, 0.01, 0.1082, 0.062, 0.01, 0.1093, 0.0608, 0.0061, 0.1111, 0.0596, 0.0061, 0.1122, 0.0608, 0.0083, 0.1111, 0.0596, 0.0083, 0.1122, 0.0623, 0.0061, 0.1148, 0.0584, 0.0061, 0.1129, 0.0623, 0.0083, 0.1148, 0.0584, 0.0083, 0.1129, 0.057, 0.0061, 0.1134, 0.0548, 0.0061, 0.1136, 0.057, 0.0083, 0.1134, 0.0548, 0.0083, 0.1136, 0.0608, 0.01, 0.1112, 0.0596, 0.01, 0.1122, 0.0622, 0.01, 0.1148, 0.0587, 0.01, 0.1127, 0.0569, 0.01, 0.1134,
0.0548, 0.01, 0.1136, 0.0613, 0.0061, 0.1155, 0.0593, 0.0061, 0.1167, 0.0613, 0.0083, 0.1155, 0.0593, 0.0083, 0.1167, 0.0581, 0.0061, 0.117, 0.0581, 0.0083, 0.117, 0.057, 0.0061, 0.1173, 0.0548, 0.0061, 0.1176, 0.057, 0.0083, 0.1173, 0.0548, 0.0083, 0.1176, 0.0612, 0.01, 0.1156, 0.0593, 0.01, 0.1167, 0.0581, 0.01, 0.117, 0.057, 0.01, 0.1159, 0.0548, 0.01, 0.1159, 0.057, 0.01, 0.1173, 0.0548, 0.01, 0.1176, 0.0463, 0.0061, 0.1072, 0.0463, 0.0083, 0.1072, 0.0476, 0.0061, 0.1099, 0.047, 0.0061, 0.1088, 0.0476, 0.0083, 0.1099, 0.047, 0.0083, 0.1088, 0.0462, 0.01, 0.107, 0.0476, 0.01, 0.11, 0.047, 0.01, 0.109, 0.0499, 0.0061, 0.1123, 0.0499, 0.0083, 0.1123, 0.0526, 0.0061, 0.1134,
0.051, 0.0061, 0.1129, 0.0526, 0.0083, 0.1134, 0.051, 0.0083, 0.1129, 0.0486, 0.0061, 0.1112, 0.0486, 0.0083, 0.1112, 0.0456, 0.0061, 0.1142, 0.0456, 0.0083, 0.1142, 0.0498, 0.01, 0.1122, 0.0528, 0.01, 0.1134, 0.0508, 0.01, 0.1128, 0.0486, 0.01, 0.1112, 0.0461, 0.01, 0.1116, 0.0483, 0.01, 0.1137, 0.0456, 0.01, 0.1142, 0.0426, 0.0061, 0.1072, 0.0426, 0.0083, 0.1072, 0.043, 0.0061, 0.1098, 0.0427, 0.0061, 0.1088, 0.043, 0.0083, 0.1098, 0.0427, 0.0083, 0.1088, 0.0439, 0.01, 0.1072, 0.0426, 0.01, 0.1072, 0.043, 0.01, 0.1096, 0.0427, 0.01, 0.1086, 0.0436, 0.0061, 0.1116, 0.0436, 0.0083, 0.1116, 0.0447, 0.0061, 0.1131, 0.0447, 0.0083, 0.1131, 0.0437, 0.01, 0.1119, 0.0447, 0.01, 0.1131,
0.0498, 0.0061, 0.1168, 0.0498, 0.0083, 0.1168, 0.0526, 0.0061, 0.1175, 0.0509, 0.0061, 0.1172, 0.0526, 0.0083, 0.1175, 0.0509, 0.0083, 0.1172, 0.0482, 0.0061, 0.1162, 0.0467, 0.0061, 0.1152, 0.0482, 0.0083, 0.1162, 0.0467, 0.0083, 0.1152, 0.0526, 0.01, 0.1159, 0.0501, 0.01, 0.1169, 0.0526, 0.01, 0.1175, 0.0511, 0.01, 0.1172, 0.048, 0.01, 0.1161, 0.0468, 0.01, 0.1152, 0.067, 0.0061, 0.1363, 0.067, 0.0083, 0.1363, 0.0656, 0.0061, 0.1363, 0.0635, 0.0061, 0.1363, 0.0656, 0.0083, 0.1363, 0.0635, 0.0083, 0.1363, 0.067, 0.01, 0.1363, 0.0656, 0.01, 0.1363, 0.0635, 0.01, 0.1363, 0.067, 0.0061, 0.1376, 0.067, 0.0083, 0.1376, 0.067, 0.0061, 0.1402, 0.067, 0.0083, 0.1402, 0.0656, 0.0061, 0.1402,
0.0635, 0.0061, 0.1402, 0.0656, 0.0083, 0.1402, 0.0635, 0.0083, 0.1402, 0.067, 0.01, 0.1376, 0.067, 0.01, 0.1402, 0.0656, 0.01, 0.1376, 0.0635, 0.01, 0.1376, 0.0656, 0.01, 0.1402, 0.0635, 0.01, 0.1402, 0.0613, 0.0061, 0.1363, 0.0591, 0.0061, 0.1363, 0.0613, 0.0083, 0.1363, 0.0591, 0.0083, 0.1363, 0.057, 0.0061, 0.1363, 0.0548, 0.0061, 0.1363, 0.057, 0.0083, 0.1363, 0.0548, 0.0083, 0.1363, 0.0613, 0.01, 0.1363, 0.0591, 0.01, 0.1363, 0.057, 0.01, 0.1363, 0.0548, 0.01, 0.1363, 0.0613, 0.0061, 0.1402, 0.0591, 0.0061, 0.1402, 0.0613, 0.0083, 0.1402, 0.0591, 0.0083, 0.1402, 0.057, 0.0061, 0.1402, 0.0548, 0.0061, 0.1402, 0.057, 0.0083, 0.1402, 0.0548, 0.0083, 0.1402, 0.0613, 0.01, 0.1376,
0.0591, 0.01, 0.1376, 0.0613, 0.01, 0.1402, 0.0591, 0.01, 0.1402, 0.057, 0.01, 0.1376, 0.0548, 0.01, 0.1376, 0.057, 0.01, 0.1402, 0.0548, 0.01, 0.1402, 0.0526, 0.0061, 0.1363, 0.0505, 0.0061, 0.1363, 0.0526, 0.0083, 0.1363, 0.0505, 0.0083, 0.1363, 0.0483, 0.0061, 0.1363, 0.0461, 0.0061, 0.1363, 0.0483, 0.0083, 0.1363, 0.0461, 0.0083, 0.1363, 0.0526, 0.01, 0.1363, 0.0505, 0.01, 0.1363, 0.0483, 0.01, 0.1363, 0.0461, 0.01, 0.1363, 0.0526, 0.0061, 0.1402, 0.0505, 0.0061, 0.1402, 0.0526, 0.0083, 0.1402, 0.0505, 0.0083, 0.1402, 0.0483, 0.0061, 0.1402, 0.0454, 0.0061, 0.1402, 0.0483, 0.0083, 0.1402, 0.0454, 0.0083, 0.1402, 0.0526, 0.01, 0.1376, 0.0505, 0.01, 0.1376, 0.0526, 0.01, 0.1402,
0.0505, 0.01, 0.1402, 0.0483, 0.01, 0.1376, 0.0461, 0.01, 0.1376, 0.0483, 0.01, 0.1402, 0.0454, 0.01, 0.1402, 0.0439, 0.0061, 0.1363, 0.0425, 0.0061, 0.1363, 0.0439, 0.0083, 0.1363, 0.0425, 0.0083, 0.1363, 0.0439, 0.01, 0.1363, 0.0425, 0.01, 0.1363, 0.0425, 0.0061, 0.1376, 0.0425, 0.0083, 0.1376, 0.0425, 0.0061, 0.1398, 0.0425, 0.0083, 0.1398, 0.0439, 0.01, 0.1376, 0.0425, 0.01, 0.1376, 0.0439, 0.01, 0.1398, 0.0425, 0.01, 0.1398, 0.0591, 0.0061, 0.1431, 0.0591, 0.0083, 0.1431, 0.0562, 0.0061, 0.1431, 0.0562, 0.0083, 0.1431, 0.0591, 0.01, 0.1431, 0.0562, 0.01, 0.1431, 0.0591, 0.0061, 0.1463, 0.0591, 0.0083, 0.1463, 0.0592, 0.0061, 0.1485, 0.0592, 0.0083, 0.1485, 0.0562, 0.0061, 0.1463,
0.0562, 0.0083, 0.1463, 0.0562, 0.0061, 0.1484, 0.0562, 0.0083, 0.1484, 0.0591, 0.01, 0.1463, 0.0592, 0.01, 0.1484, 0.0562, 0.01, 0.1463, 0.0562, 0.01, 0.1484, 0.0632, 0.0061, 0.1553, 0.0632, 0.0083, 0.1553, 0.0651, 0.0061, 0.1575, 0.0643, 0.0061, 0.1564, 0.0651, 0.0083, 0.1575, 0.0643, 0.0083, 0.1564, 0.0632, 0.01, 0.1553, 0.0652, 0.01, 0.1576, 0.0643, 0.01, 0.1564, 0.0596, 0.0061, 0.1506, 0.0596, 0.0083, 0.1506, 0.0609, 0.0061, 0.1529, 0.0602, 0.0061, 0.1518, 0.0609, 0.0083, 0.1529, 0.0602, 0.0083, 0.1518, 0.0562, 0.0061, 0.1506, 0.0562, 0.0083, 0.1506, 0.0562, 0.0061, 0.1528, 0.0562, 0.0083, 0.1528, 0.0596, 0.01, 0.1504, 0.061, 0.01, 0.153, 0.0603, 0.01, 0.1521, 0.0562, 0.01, 0.1506,
0.0562, 0.01, 0.1527, 0.0621, 0.0061, 0.1542, 0.0621, 0.0083, 0.1542, 0.0561, 0.0061, 0.1546, 0.0558, 0.0061, 0.1556, 0.0561, 0.0083, 0.1546, 0.0558, 0.0083, 0.1556, 0.0553, 0.0061, 0.1572, 0.0553, 0.0083, 0.1572, 0.0621, 0.01, 0.1542, 0.0591, 0.01, 0.155, 0.0613, 0.01, 0.1571, 0.0591, 0.01, 0.1571, 0.056, 0.01, 0.1549, 0.0557, 0.01, 0.1557, 0.057, 0.01, 0.1571, 0.0553, 0.01, 0.1571, 0.0454, 0.0061, 0.1419, 0.0454, 0.0083, 0.1419, 0.0454, 0.0061, 0.1441, 0.0454, 0.0083, 0.1441, 0.0454, 0.01, 0.1419, 0.0454, 0.01, 0.1441, 0.0454, 0.0061, 0.1463, 0.0454, 0.0083, 0.1463, 0.0454, 0.0061, 0.1484, 0.0454, 0.0083, 0.1484, 0.0454, 0.01, 0.1463, 0.0454, 0.01, 0.1484, 0.0425, 0.0061, 0.1419,
0.0425, 0.0083, 0.1419, 0.0425, 0.0061, 0.1441, 0.0425, 0.0083, 0.1441, 0.0439, 0.01, 0.1419, 0.0425, 0.01, 0.1419, 0.0439, 0.01, 0.1441, 0.0425, 0.01, 0.1441, 0.0425, 0.0061, 0.1463, 0.0425, 0.0083, 0.1463, 0.0425, 0.0061, 0.1484, 0.0425, 0.0083, 0.1484, 0.0439, 0.01, 0.1463, 0.0425, 0.01, 0.1463, 0.0439, 0.01, 0.1484, 0.0425, 0.01, 0.1484, 0.0454, 0.0061, 0.1506, 0.0454, 0.0083, 0.1506, 0.0456, 0.0061, 0.1528, 0.0456, 0.0083, 0.1528, 0.0454, 0.01, 0.1506, 0.0456, 0.01, 0.1528, 0.0459, 0.0061, 0.155, 0.0459, 0.0083, 0.155, 0.0474, 0.0061, 0.1579, 0.0467, 0.0061, 0.1568, 0.0474, 0.0083, 0.1579, 0.0467, 0.0083, 0.1568, 0.0459, 0.01, 0.1549, 0.0475, 0.01, 0.1579, 0.0468, 0.01, 0.1569,
0.0425, 0.0061, 0.1506, 0.0425, 0.0083, 0.1506, 0.0425, 0.0061, 0.1528, 0.0425, 0.0083, 0.1528, 0.0439, 0.01, 0.1506, 0.0425, 0.01, 0.1506, 0.0439, 0.01, 0.1528, 0.0425, 0.01, 0.1528, 0.0425, 0.0061, 0.155, 0.0425, 0.0083, 0.155, 0.0426, 0.0061, 0.1571, 0.0426, 0.0083, 0.1571, 0.0439, 0.01, 0.155, 0.0425, 0.01, 0.155, 0.0439, 0.01, 0.1571, 0.0426, 0.01, 0.1571, 0.0669, 0.0061, 0.1619, 0.0669, 0.0083, 0.1619, 0.0662, 0.0061, 0.1592, 0.0662, 0.0083, 0.1592, 0.0666, 0.0061, 0.1608, 0.0632, 0.0061, 0.1619, 0.0666, 0.0083, 0.1608, 0.0632, 0.0083, 0.1619, 0.0669, 0.01, 0.1621, 0.0661, 0.01, 0.159, 0.0635, 0.01, 0.1593, 0.0667, 0.01, 0.1612, 0.0632, 0.01, 0.1619, 0.0671, 0.0061, 0.1648,
0.0671, 0.0083, 0.1648, 0.0656, 0.0061, 0.1638, 0.0643, 0.0061, 0.1628, 0.0656, 0.0083, 0.1638, 0.0643, 0.0083, 0.1628, 0.0671, 0.01, 0.1648, 0.0653, 0.01, 0.1636, 0.0643, 0.01, 0.1628, 0.0608, 0.0061, 0.1599, 0.0591, 0.0061, 0.1588, 0.0608, 0.0083, 0.1599, 0.0591, 0.0083, 0.1588, 0.0619, 0.0061, 0.1608, 0.0583, 0.0061, 0.1607, 0.0619, 0.0083, 0.1608, 0.0583, 0.0083, 0.1607, 0.0543, 0.0061, 0.1587, 0.0543, 0.0083, 0.1587, 0.0574, 0.0061, 0.1618, 0.0574, 0.0083, 0.1618, 0.0609, 0.01, 0.16, 0.0591, 0.01, 0.1588, 0.0619, 0.01, 0.1608, 0.0583, 0.01, 0.1606, 0.057, 0.01, 0.1593, 0.0543, 0.01, 0.1588, 0.0574, 0.01, 0.1618, 0.0548, 0.01, 0.1615, 0.0563, 0.0061, 0.1629, 0.0549, 0.0061, 0.1638,
0.0563, 0.0083, 0.1629, 0.0549, 0.0083, 0.1638, 0.0563, 0.01, 0.1629, 0.0551, 0.01, 0.1637, 0.053, 0.0061, 0.1596, 0.0503, 0.0061, 0.1603, 0.053, 0.0083, 0.1596, 0.0503, 0.0083, 0.1603, 0.0479, 0.0061, 0.1585, 0.0479, 0.0083, 0.1585, 0.0525, 0.01, 0.1598, 0.0503, 0.01, 0.1603, 0.0526, 0.01, 0.1615, 0.0505, 0.01, 0.1615, 0.0484, 0.01, 0.1591, 0.0461, 0.01, 0.1593, 0.0483, 0.01, 0.1615, 0.0461, 0.01, 0.1615, 0.0527, 0.0061, 0.1646, 0.0505, 0.0061, 0.1647, 0.0527, 0.0083, 0.1646, 0.0505, 0.0083, 0.1647, 0.048, 0.0061, 0.1645, 0.046, 0.0061, 0.1634, 0.048, 0.0083, 0.1645, 0.046, 0.0083, 0.1634, 0.0529, 0.01, 0.1645, 0.0505, 0.01, 0.1647, 0.048, 0.01, 0.1644, 0.0459, 0.01, 0.1633,
0.0431, 0.0061, 0.1596, 0.0428, 0.0061, 0.1585, 0.0431, 0.0083, 0.1596, 0.0428, 0.0083, 0.1585, 0.0444, 0.0061, 0.162, 0.0444, 0.0083, 0.162, 0.043, 0.01, 0.1595, 0.0428, 0.01, 0.1584, 0.0442, 0.01, 0.1618, 0.045, 0.0061, 0.1626, 0.045, 0.0083, 0.1626, 0.045, 0.01, 0.1626, 0.0219, 0.0061, 0.1412, 0.0201, 0.0061, 0.1412, 0.0219, 0.0083, 0.1412, 0.0201, 0.0083, 0.1412, 0.0219, 0.0061, 0.1441, 0.0219, 0.0083, 0.1441, 0.0219, 0.01, 0.1412, 0.0201, 0.01, 0.1412, 0.0219, 0.01, 0.1441, 0.0201, 0.01, 0.1441, 0.0219, 0.0061, 0.1463, 0.0219, 0.0083, 0.1463, 0.0219, 0.0061, 0.1484, 0.0219, 0.0083, 0.1484, 0.0219, 0.01, 0.1463, 0.0201, 0.01, 0.1463, 0.0219, 0.01, 0.1484, 0.0201, 0.01, 0.1484,
0.0219, 0.0061, 0.1506, 0.0219, 0.0083, 0.1506, 0.0219, 0.0061, 0.1528, 0.0219, 0.0083, 0.1528, 0.0219, 0.01, 0.1506, 0.0201, 0.01, 0.1506, 0.0219, 0.01, 0.1528, 0.0201, 0.01, 0.1528, 0.0219, 0.0061, 0.155, 0.0219, 0.0083, 0.155, 0.0219, 0.0061, 0.1571, 0.0219, 0.0083, 0.1571, 0.0219, 0.01, 0.155, 0.0201, 0.01, 0.155, 0.0219, 0.01, 0.1571, 0.0201, 0.01, 0.1571, 0.0179, 0.0061, 0.1412, 0.0157, 0.0061, 0.1412, 0.0179, 0.0083, 0.1412, 0.0157, 0.0083, 0.1412, 0.0136, 0.0061, 0.1412, 0.0114, 0.0061, 0.1412, 0.0136, 0.0083, 0.1412, 0.0114, 0.0083, 0.1412, 0.0179, 0.01, 0.1412, 0.0157, 0.01, 0.1412, 0.0179, 0.01, 0.1441, 0.0157, 0.01, 0.1441, 0.0136, 0.01, 0.1412, 0.0114, 0.01, 0.1412,
0.0136, 0.01, 0.1441, 0.0114, 0.01, 0.1441, 0.0179, 0.01, 0.1463, 0.0157, 0.01, 0.1463, 0.0179, 0.01, 0.1484, 0.0157, 0.01, 0.1484, 0.0136, 0.01, 0.1463, 0.0114, 0.01, 0.1463, 0.0136, 0.01, 0.1484, 0.0114, 0.01, 0.1484, 0.0092, 0.0061, 0.1412, 0.0071, 0.0061, 0.1412, 0.0092, 0.0083, 0.1412, 0.0071, 0.0083, 0.1412, 0.0049, 0.0061, 0.1412, 0.0027, 0.0061, 0.1412, 0.0049, 0.0083, 0.1412, 0.0027, 0.0083, 0.1412, 0.0092, 0.01, 0.1412, 0.0071, 0.01, 0.1412, 0.0092, 0.01, 0.1441, 0.0071, 0.01, 0.1441, 0.0049, 0.01, 0.1412, 0.0027, 0.01, 0.1412, 0.0049, 0.01, 0.1441, 0.0027, 0.01, 0.1441, 0.0092, 0.01, 0.1463, 0.0071, 0.01, 0.1463, 0.0092, 0.01, 0.1484, 0.0071, 0.01, 0.1484,
0.0049, 0.01, 0.1463, 0.0027, 0.01, 0.1463, 0.0049, 0.01, 0.1484, 0.0027, 0.01, 0.1484, 0.0179, 0.01, 0.1506, 0.0157, 0.01, 0.1506, 0.0179, 0.01, 0.1528, 0.0157, 0.01, 0.1528, 0.0136, 0.01, 0.1506, 0.0114, 0.01, 0.1506, 0.0136, 0.01, 0.1528, 0.0114, 0.01, 0.1528, 0.0179, 0.01, 0.155, 0.0157, 0.01, 0.155, 0.0179, 0.01, 0.1571, 0.0157, 0.01, 0.1571, 0.0136, 0.01, 0.155, 0.0114, 0.01, 0.155, 0.0136, 0.01, 0.1571, 0.0114, 0.01, 0.1571, 0.0101, 0.0061, 0.152, 0.0071, 0.0061, 0.152, 0.0101, 0.0083, 0.152, 0.0071, 0.0083, 0.152, 0.0049, 0.0061, 0.152, 0.0027, 0.0061, 0.152, 0.0049, 0.0083, 0.152, 0.0027, 0.0083, 0.152, 0.0092, 0.01, 0.1506, 0.0071, 0.01, 0.1506,
0.0101, 0.01, 0.152, 0.0071, 0.01, 0.152, 0.0049, 0.01, 0.1506, 0.0027, 0.01, 0.1506, 0.0049, 0.01, 0.152, 0.0027, 0.01, 0.152, 0.0101, 0.0061, 0.155, 0.0101, 0.0083, 0.155, 0.0101, 0.0061, 0.1571, 0.0101, 0.0083, 0.1571, 0.0101, 0.01, 0.155, 0.0101, 0.01, 0.1571, 0.0219, 0.0061, 0.1593, 0.0219, 0.0083, 0.1593, 0.0219, 0.0061, 0.1615, 0.0219, 0.0083, 0.1615, 0.0219, 0.01, 0.1593, 0.0201, 0.01, 0.1593, 0.0219, 0.01, 0.1615, 0.0201, 0.01, 0.1615, 0.0219, 0.0061, 0.1636, 0.0219, 0.0083, 0.1636, 0.0219, 0.0061, 0.1658, 0.0219, 0.0083, 0.1658, 0.0219, 0.01, 0.1636, 0.0201, 0.01, 0.1636, 0.0219, 0.01, 0.1658, 0.0201, 0.01, 0.1658, 0.0219, 0.0061, 0.168, 0.0219, 0.0083, 0.168,
0.0219, 0.0061, 0.1701, 0.0219, 0.0083, 0.1701, 0.0219, 0.01, 0.168, 0.0201, 0.01, 0.168, 0.0219, 0.01, 0.1701, 0.0201, 0.01, 0.1701, 0.0219, 0.0061, 0.1723, 0.0219, 0.0083, 0.1723, 0.0219, 0.0061, 0.1745, 0.0219, 0.0083, 0.1745, 0.0219, 0.01, 0.1723, 0.0201, 0.01, 0.1723, 0.0219, 0.01, 0.1745, 0.0201, 0.01, 0.1745, 0.0179, 0.01, 0.1593, 0.0157, 0.01, 0.1593, 0.0179, 0.01, 0.1615, 0.0157, 0.01, 0.1615, 0.0136, 0.01, 0.1593, 0.0114, 0.01, 0.1593, 0.0136, 0.01, 0.1615, 0.0114, 0.01, 0.1615, 0.0179, 0.01, 0.1636, 0.0157, 0.01, 0.1636, 0.0179, 0.01, 0.1658, 0.0157, 0.01, 0.1658, 0.0136, 0.01, 0.1636, 0.0114, 0.01, 0.1636, 0.0136, 0.01, 0.1658, 0.0114, 0.01, 0.1658,
0.0101, 0.0061, 0.1593, 0.0101, 0.0083, 0.1593, 0.0101, 0.0061, 0.1615, 0.0101, 0.0083, 0.1615, 0.0101, 0.01, 0.1593, 0.0101, 0.01, 0.1615, 0.0101, 0.0061, 0.1636, 0.0101, 0.0083, 0.1636, 0.0101, 0.0061, 0.1658, 0.0101, 0.0083, 0.1658, 0.0101, 0.01, 0.1636, 0.0101, 0.01, 0.1658, 0.0179, 0.01, 0.168, 0.0157, 0.01, 0.168, 0.0179, 0.01, 0.1701, 0.0157, 0.01, 0.1701, 0.0136, 0.01, 0.168, 0.0114, 0.01, 0.168, 0.0136, 0.01, 0.1701, 0.0114, 0.01, 0.1701, 0.0179, 0.01, 0.1723, 0.0157, 0.01, 0.1723, 0.0179, 0.01, 0.1745, 0.0157, 0.01, 0.1745, 0.0136, 0.01, 0.1723, 0.0114, 0.01, 0.1723, 0.0136, 0.01, 0.1745, 0.0114, 0.01, 0.1745, 0.0101, 0.0061, 0.168, 0.0101, 0.0083, 0.168,
0.0101, 0.0061, 0.1701, 0.0101, 0.0083, 0.1701, 0.0101, 0.01, 0.168, 0.0101, 0.01, 0.1701, 0.0101, 0.0061, 0.1723, 0.0101, 0.0083, 0.1723, 0.0101, 0.0061, 0.1745, 0.0101, 0.0083, 0.1745, 0.0101, 0.01, 0.1723, 0.0101, 0.01, 0.1745, 0.0219, 0.0061, 0.1767, 0.0219, 0.0083, 0.1767, 0.0219, 0.0061, 0.1788, 0.0219, 0.0083, 0.1788, 0.0219, 0.01, 0.1767, 0.0201, 0.01, 0.1767, 0.0219, 0.01, 0.1788, 0.0201, 0.01, 0.1788, 0.0219, 0.0061, 0.181, 0.0219, 0.0083, 0.181, 0.0219, 0.0061, 0.1832, 0.0219, 0.0083, 0.1832, 0.0219, 0.01, 0.181, 0.0201, 0.01, 0.181, 0.0219, 0.01, 0.1832, 0.0201, 0.01, 0.1832, 0.0219, 0.0061, 0.1853, 0.0219, 0.0083, 0.1853, 0.0219, 0.0061, 0.1875, 0.0219, 0.0083, 0.1875,
0.0219, 0.01, 0.1853, 0.0201, 0.01, 0.1853, 0.0219, 0.01, 0.1875, 0.0201, 0.01, 0.1875, 0.0219, 0.0061, 0.1897, 0.0219, 0.0083, 0.1897, 0.0219, 0.0061, 0.1918, 0.0219, 0.0083, 0.1918, 0.0219, 0.01, 0.1897, 0.0201, 0.01, 0.1897, 0.0219, 0.01, 0.1918, 0.0201, 0.01, 0.1918, 0.0179, 0.01, 0.1767, 0.0157, 0.01, 0.1767, 0.0179, 0.01, 0.1788, 0.0157, 0.01, 0.1788, 0.0136, 0.01, 0.1767, 0.0114, 0.01, 0.1767, 0.0136, 0.01, 0.1788, 0.0114, 0.01, 0.1788, 0.0179, 0.01, 0.181, 0.0157, 0.01, 0.181, 0.0179, 0.01, 0.1832, 0.0157, 0.01, 0.1832, 0.0136, 0.01, 0.181, 0.0114, 0.01, 0.181, 0.0136, 0.01, 0.1832, 0.0114, 0.01, 0.1832, 0.0101, 0.0061, 0.1767, 0.0101, 0.0083, 0.1767,
0.0101, 0.0061, 0.1788, 0.0101, 0.0083, 0.1788, 0.0101, 0.01, 0.1767, 0.0101, 0.01, 0.1788, 0.0101, 0.0061, 0.181, 0.0101, 0.0083, 0.181, 0.0101, 0.0061, 0.1832, 0.0101, 0.0083, 0.1832, 0.0101, 0.01, 0.181, 0.0101, 0.01, 0.1832, 0.0179, 0.01, 0.1853, 0.0157, 0.01, 0.1853, 0.0179, 0.01, 0.1875, 0.0157, 0.01, 0.1875, 0.0136, 0.01, 0.1853, 0.0114, 0.01, 0.1853, 0.0136, 0.01, 0.1875, 0.0114, 0.01, 0.1875, 0.0179, 0.01, 0.1897, 0.0157, 0.01, 0.1897, 0.0179, 0.01, 0.1918, 0.0157, 0.01, 0.1918, 0.0136, 0.01, 0.1897, 0.0114, 0.01, 0.1897, 0.0136, 0.01, 0.1918, 0.0114, 0.01, 0.1918, 0.0101, 0.0061, 0.1853, 0.0101, 0.0083, 0.1853, 0.0101, 0.0061, 0.1875, 0.0101, 0.0083, 0.1875,
0.0101, 0.01, 0.1853, 0.0101, 0.01, 0.1875, 0.0101, 0.0061, 0.1897, 0.0101, 0.0083, 0.1897, 0.0101, 0.0061, 0.1918, 0.0101, 0.0083, 0.1918, 0.0101, 0.01, 0.1897, 0.0101, 0.01, 0.1918, 0.0219, 0.0061, 0.194, 0.0219, 0.0083, 0.194, 0.0219, 0.0061, 0.1962, 0.0219, 0.0083, 0.1962, 0.0219, 0.01, 0.194, 0.0201, 0.01, 0.194, 0.0219, 0.01, 0.1962, 0.0201, 0.01, 0.1962, 0.0219, 0.0061, 0.1984, 0.0219, 0.0083, 0.1984, 0.0219, 0.0061, 0.2005, 0.0219, 0.0083, 0.2005, 0.0219, 0.01, 0.1984, 0.0201, 0.01, 0.1984, 0.0219, 0.01, 0.2005, 0.0201, 0.01, 0.2005, 0.0219, 0.0061, 0.202, 0.0201, 0.0061, 0.202, 0.0219, 0.0083, 0.202, 0.0201, 0.0083, 0.202, 0.0219, 0.01, 0.202, 0.0201, 0.01, 0.202,
0.0179, 0.01, 0.194, 0.0157, 0.01, 0.194, 0.0179, 0.01, 0.1962, 0.0157, 0.01, 0.1962, 0.0136, 0.01, 0.194, 0.0114, 0.01, 0.194, 0.0136, 0.01, 0.1962, 0.0114, 0.01, 0.1962, 0.0179, 0.01, 0.1984, 0.0157, 0.01, 0.1984, 0.0179, 0.01, 0.2005, 0.0157, 0.01, 0.2005, 0.0136, 0.01, 0.1984, 0.0114, 0.01, 0.1984, 0.0136, 0.01, 0.2005, 0.0114, 0.01, 0.2005, 0.0101, 0.0061, 0.194, 0.0101, 0.0083, 0.194, 0.0101, 0.0061, 0.1962, 0.0101, 0.0083, 0.1962, 0.0101, 0.01, 0.194, 0.0101, 0.01, 0.1962, 0.0101, 0.0061, 0.1984, 0.0101, 0.0083, 0.1984, 0.0101, 0.0061, 0.2005, 0.0101, 0.0083, 0.2005, 0.0101, 0.01, 0.1984, 0.0101, 0.01, 0.2005, 0.0179, 0.0061, 0.202, 0.0157, 0.0061, 0.202,
0.0179, 0.0083, 0.202, 0.0157, 0.0083, 0.202, 0.0136, 0.0061, 0.202, 0.0114, 0.0061, 0.202, 0.0136, 0.0083, 0.202, 0.0114, 0.0083, 0.202, 0.0179, 0.01, 0.202, 0.0157, 0.01, 0.202, 0.0136, 0.01, 0.202, 0.0114, 0.01, 0.202, 0.0101, 0.0061, 0.202, 0.0101, 0.0083, 0.202, 0.0101, 0.01, 0.202, 0.067, 0.0061, 0.2118, 0.067, 0.0083, 0.2118, 0.067, 0.0061, 0.2135, 0.067, 0.0083, 0.2135, 0.0656, 0.0061, 0.2118, 0.0635, 0.0061, 0.2118, 0.0656, 0.0083, 0.2118, 0.0635, 0.0083, 0.2118, 0.067, 0.01, 0.2118, 0.067, 0.01, 0.2135, 0.0656, 0.01, 0.2118, 0.0635, 0.01, 0.2118, 0.0656, 0.01, 0.2135, 0.0635, 0.01, 0.2135, 0.067, 0.0061, 0.2157, 0.067, 0.0083, 0.2157, 0.067, 0.0061, 0.2179,
0.067, 0.0083, 0.2179, 0.0631, 0.0061, 0.2157, 0.0631, 0.0083, 0.2157, 0.0631, 0.0061, 0.2179, 0.0631, 0.0083, 0.2179, 0.067, 0.01, 0.2157, 0.067, 0.01, 0.2179, 0.0656, 0.01, 0.2157, 0.0631, 0.01, 0.2157, 0.0656, 0.01, 0.2179, 0.0631, 0.01, 0.2179, 0.0613, 0.0061, 0.2118, 0.0591, 0.0061, 0.2118, 0.0613, 0.0083, 0.2118, 0.0591, 0.0083, 0.2118, 0.057, 0.0061, 0.2118, 0.0548, 0.0061, 0.2118, 0.057, 0.0083, 0.2118, 0.0548, 0.0083, 0.2118, 0.0613, 0.01, 0.2118, 0.0591, 0.01, 0.2118, 0.0613, 0.01, 0.2135, 0.0591, 0.01, 0.2135, 0.057, 0.01, 0.2118, 0.0548, 0.01, 0.2118, 0.057, 0.01, 0.2135, 0.0548, 0.01, 0.2135, 0.0613, 0.0061, 0.2157, 0.0591, 0.0061, 0.2157, 0.0613, 0.0083, 0.2157,
0.0591, 0.0083, 0.2157, 0.057, 0.0061, 0.2157, 0.0548, 0.0061, 0.2157, 0.057, 0.0083, 0.2157, 0.0548, 0.0083, 0.2157, 0.0613, 0.01, 0.2157, 0.0591, 0.01, 0.2157, 0.057, 0.01, 0.2157, 0.0548, 0.01, 0.2157, 0.067, 0.0061, 0.2201, 0.067, 0.0083, 0.2201, 0.067, 0.0061, 0.2222, 0.067, 0.0083, 0.2222, 0.0631, 0.0061, 0.2201, 0.0631, 0.0083, 0.2201, 0.0631, 0.0061, 0.2222, 0.0631, 0.0083, 0.2222, 0.067, 0.01, 0.2201, 0.067, 0.01, 0.2222, 0.0656, 0.01, 0.2201, 0.0631, 0.01, 0.2201, 0.0656, 0.01, 0.2222, 0.0631, 0.01, 0.2222, 0.067, 0.0061, 0.2244, 0.067, 0.0083, 0.2244, 0.067, 0.0061, 0.2266, 0.067, 0.0083, 0.2266, 0.0631, 0.0061, 0.2244, 0.0631, 0.0083, 0.2244, 0.0631, 0.0061, 0.2266,
0.0631, 0.0083, 0.2266, 0.067, 0.01, 0.2244, 0.067, 0.01, 0.2266, 0.0656, 0.01, 0.2244, 0.0631, 0.01, 0.2244, 0.0656, 0.01, 0.2266, 0.0631, 0.01, 0.2266, 0.0562, 0.0061, 0.2196, 0.0548, 0.0061, 0.2196, 0.0562, 0.0083, 0.2196, 0.0548, 0.0083, 0.2196, 0.0562, 0.0061, 0.2222, 0.0562, 0.0083, 0.2222, 0.0562, 0.01, 0.2196, 0.0548, 0.01, 0.2196, 0.0562, 0.01, 0.2222, 0.0548, 0.01, 0.2222, 0.0562, 0.0061, 0.2244, 0.0562, 0.0083, 0.2244, 0.0562, 0.0061, 0.2266, 0.0562, 0.0083, 0.2266, 0.0562, 0.01, 0.2244, 0.0548, 0.01, 0.2244, 0.0562, 0.01, 0.2266, 0.0548, 0.01, 0.2266, 0.0526, 0.0061, 0.2118, 0.0505, 0.0061, 0.2118, 0.0526, 0.0083, 0.2118, 0.0505, 0.0083, 0.2118, 0.0483, 0.0061, 0.2118,
0.0461, 0.0061, 0.2118, 0.0483, 0.0083, 0.2118, 0.0461, 0.0083, 0.2118, 0.0526, 0.01, 0.2118, 0.0505, 0.01, 0.2118, 0.0526, 0.01, 0.2135, 0.0505, 0.01, 0.2135, 0.0483, 0.01, 0.2118, 0.0461, 0.01, 0.2118, 0.0483, 0.01, 0.2135, 0.0461, 0.01, 0.2135, 0.0526, 0.0061, 0.2157, 0.0505, 0.0061, 0.2157, 0.0526, 0.0083, 0.2157, 0.0505, 0.0083, 0.2157, 0.0483, 0.0061, 0.2157, 0.0454, 0.0061, 0.2157, 0.0483, 0.0083, 0.2157, 0.0454, 0.0083, 0.2157, 0.0454, 0.0061, 0.2179, 0.0454, 0.0083, 0.2179, 0.0526, 0.01, 0.2157, 0.0505, 0.01, 0.2157, 0.0483, 0.01, 0.2157, 0.0454, 0.01, 0.2157, 0.0454, 0.01, 0.2179, 0.0439, 0.0061, 0.2118, 0.0425, 0.0061, 0.2118, 0.0439, 0.0083, 0.2118, 0.0425, 0.0083, 0.2118,
0.0425, 0.0061, 0.2135, 0.0425, 0.0083, 0.2135, 0.0439, 0.01, 0.2118, 0.0425, 0.01, 0.2118, 0.0439, 0.01, 0.2135, 0.0425, 0.01, 0.2135, 0.0425, 0.0061, 0.2157, 0.0425, 0.0083, 0.2157, 0.0425, 0.0061, 0.2179, 0.0425, 0.0083, 0.2179, 0.0439, 0.01, 0.2157, 0.0425, 0.01, 0.2157, 0.0439, 0.01, 0.2179, 0.0425, 0.01, 0.2179, 0.0523, 0.0061, 0.2196, 0.0523, 0.0083, 0.2196, 0.0523, 0.0061, 0.2222, 0.0523, 0.0083, 0.2222, 0.0454, 0.0061, 0.2201, 0.0454, 0.0083, 0.2201, 0.0454, 0.0061, 0.2222, 0.0454, 0.0083, 0.2222, 0.0523, 0.01, 0.2196, 0.0523, 0.01, 0.2222, 0.0454, 0.01, 0.2201, 0.0454, 0.01, 0.2222, 0.0523, 0.0061, 0.2244, 0.0523, 0.0083, 0.2244, 0.0523, 0.0061, 0.2266, 0.0523, 0.0083, 0.2266,
0.0454, 0.0061, 0.2244, 0.0454, 0.0083, 0.2244, 0.0454, 0.0061, 0.2266, 0.0454, 0.0083, 0.2266, 0.0523, 0.01, 0.2244, 0.0523, 0.01, 0.2266, 0.0454, 0.01, 0.2244, 0.0454, 0.01, 0.2266, 0.0425, 0.0061, 0.2201, 0.0425, 0.0083, 0.2201, 0.0425, 0.0061, 0.2222, 0.0425, 0.0083, 0.2222, 0.0439, 0.01, 0.2201, 0.0425, 0.01, 0.2201, 0.0439, 0.01, 0.2222, 0.0425, 0.01, 0.2222, 0.0425, 0.0061, 0.2244, 0.0425, 0.0083, 0.2244, 0.0425, 0.0061, 0.2266, 0.0425, 0.0083, 0.2266, 0.0439, 0.01, 0.2244, 0.0425, 0.01, 0.2244, 0.0439, 0.01, 0.2266, 0.0425, 0.01, 0.2266, 0.067, 0.0061, 0.2287, 0.067, 0.0083, 0.2287, 0.0669, 0.0061, 0.2307, 0.0669, 0.0083, 0.2307, 0.063, 0.0061, 0.2287, 0.063, 0.0083, 0.2287,
0.0667, 0.0061, 0.2318, 0.0629, 0.0061, 0.2309, 0.0667, 0.0083, 0.2318, 0.0629, 0.0083, 0.2309, 0.067, 0.01, 0.2287, 0.0669, 0.01, 0.2304, 0.0656, 0.01, 0.2287, 0.0631, 0.01, 0.2287, 0.0668, 0.01, 0.2312, 0.0629, 0.01, 0.2308, 0.0665, 0.0061, 0.2331, 0.0625, 0.0061, 0.2325, 0.0665, 0.0083, 0.2331, 0.0625, 0.0083, 0.2325, 0.0657, 0.0061, 0.2355, 0.0657, 0.0083, 0.2355, 0.0665, 0.01, 0.2333, 0.0624, 0.01, 0.2328, 0.0657, 0.01, 0.2356, 0.0635, 0.01, 0.2353, 0.0563, 0.0061, 0.2287, 0.0563, 0.0083, 0.2287, 0.0565, 0.0061, 0.2309, 0.0565, 0.0083, 0.2309, 0.0563, 0.01, 0.2287, 0.0548, 0.01, 0.2287, 0.0565, 0.01, 0.2309, 0.0548, 0.01, 0.2309, 0.0621, 0.0061, 0.2336, 0.0621, 0.0083, 0.2336,
0.0613, 0.0061, 0.2349, 0.0591, 0.0061, 0.2363, 0.0613, 0.0083, 0.2349, 0.0591, 0.0083, 0.2363, 0.0571, 0.0061, 0.2331, 0.0571, 0.0083, 0.2331, 0.0579, 0.0061, 0.2345, 0.0579, 0.0083, 0.2345, 0.0621, 0.01, 0.2337, 0.0613, 0.01, 0.2349, 0.0591, 0.01, 0.2363, 0.0571, 0.01, 0.2329, 0.0548, 0.01, 0.2331, 0.058, 0.01, 0.2346, 0.0548, 0.01, 0.2353, 0.0648, 0.0061, 0.2366, 0.0634, 0.0061, 0.238, 0.0648, 0.0083, 0.2366, 0.0634, 0.0083, 0.238, 0.0648, 0.01, 0.2365, 0.0634, 0.01, 0.238, 0.0622, 0.0061, 0.2384, 0.0622, 0.0083, 0.2384, 0.0612, 0.0061, 0.2388, 0.0591, 0.0061, 0.239, 0.0612, 0.0083, 0.2388, 0.0591, 0.0083, 0.239, 0.057, 0.0061, 0.2387, 0.0544, 0.0061, 0.2384, 0.057, 0.0083, 0.2387,
0.0544, 0.0083, 0.2384, 0.0618, 0.01, 0.2386, 0.0591, 0.01, 0.2374, 0.0609, 0.01, 0.2389, 0.0591, 0.01, 0.239, 0.057, 0.01, 0.2374, 0.0547, 0.01, 0.2384, 0.057, 0.01, 0.2387, 0.0546, 0.01, 0.2384, 0.0522, 0.0061, 0.2287, 0.0522, 0.0083, 0.2287, 0.0522, 0.0061, 0.2309, 0.0522, 0.0083, 0.2309, 0.0455, 0.0061, 0.2287, 0.0455, 0.0083, 0.2287, 0.0459, 0.0061, 0.2309, 0.0459, 0.0083, 0.2309, 0.0522, 0.01, 0.2287, 0.0522, 0.01, 0.2309, 0.0455, 0.01, 0.2287, 0.0458, 0.01, 0.2308, 0.0519, 0.0061, 0.2326, 0.0513, 0.0061, 0.2337, 0.0519, 0.0083, 0.2326, 0.0513, 0.0083, 0.2337, 0.0504, 0.0061, 0.2345, 0.0504, 0.0083, 0.2345, 0.0475, 0.0061, 0.2339, 0.0465, 0.0061, 0.2328, 0.0475, 0.0083, 0.2339,
0.0465, 0.0083, 0.2328, 0.0484, 0.0061, 0.2348, 0.0484, 0.0083, 0.2348, 0.0518, 0.01, 0.2329, 0.0512, 0.01, 0.2338, 0.0526, 0.01, 0.2353, 0.0504, 0.01, 0.2345, 0.0475, 0.01, 0.2339, 0.0466, 0.01, 0.2329, 0.0484, 0.01, 0.2348, 0.0461, 0.01, 0.2353, 0.0425, 0.0061, 0.2287, 0.0425, 0.0083, 0.2287, 0.0425, 0.0061, 0.2309, 0.0425, 0.0083, 0.2309, 0.0439, 0.01, 0.2287, 0.0425, 0.01, 0.2287, 0.0439, 0.01, 0.2309, 0.0425, 0.01, 0.2309, 0.043, 0.0061, 0.2337, 0.0427, 0.0061, 0.2326, 0.043, 0.0083, 0.2337, 0.0427, 0.0083, 0.2326, 0.0437, 0.0061, 0.2353, 0.0437, 0.0083, 0.2353, 0.0429, 0.01, 0.2334, 0.0427, 0.01, 0.2324, 0.0438, 0.01, 0.2356, 0.0526, 0.0061, 0.2386, 0.0505, 0.0061, 0.239,
0.0526, 0.0083, 0.2386, 0.0505, 0.0083, 0.239, 0.0472, 0.0061, 0.2385, 0.046, 0.0061, 0.2379, 0.0472, 0.0083, 0.2385, 0.046, 0.0083, 0.2379, 0.0478, 0.0061, 0.2388, 0.0478, 0.0083, 0.2388, 0.0526, 0.01, 0.2374, 0.0505, 0.01, 0.2374, 0.0526, 0.01, 0.2386, 0.0505, 0.01, 0.239, 0.0475, 0.01, 0.2387, 0.0457, 0.01, 0.2377, 0.0485, 0.01, 0.239, 0.0447, 0.0061, 0.2368, 0.0447, 0.0083, 0.2368, 0.0447, 0.01, 0.2367, 0.0107, 0.0061, 0.2187, 0.0107, 0.0083, 0.2187, 0.0106, 0.01, 0.2187, 0.0024, 0.0061, 0.2144, 0.0024, 0.0083, 0.2144, 0.0022, 0.01, 0.2143, 0.0069, 0.0061, 0.2163, 0.0069, 0.0083, 0.2163, 0.0093, 0.0061, 0.2176, 0.008, 0.0061, 0.2169, 0.0093, 0.0083, 0.2176, 0.008, 0.0083, 0.2169,
0.0049, 0.0061, 0.2153, 0.0035, 0.0061, 0.2147, 0.0049, 0.0083, 0.2153, 0.0035, 0.0083, 0.2147, 0.0067, 0.01, 0.2162, 0.0096, 0.01, 0.2178, 0.0077, 0.01, 0.2168, 0.0052, 0.01, 0.2154, 0.0032, 0.01, 0.2146, 0.0049, 0.01, 0.2179, 0.0027, 0.01, 0.2179, 0.0149, 0.0061, 0.223, 0.0149, 0.0083, 0.223, 0.0129, 0.0061, 0.2207, 0.0118, 0.0061, 0.2196, 0.0129, 0.0083, 0.2207, 0.0118, 0.0083, 0.2196, 0.014, 0.0061, 0.2219, 0.014, 0.0083, 0.2219, 0.0149, 0.01, 0.2231, 0.0129, 0.01, 0.2207, 0.0118, 0.01, 0.2196, 0.014, 0.01, 0.2219, 0.0114, 0.01, 0.2222, 0.0158, 0.0061, 0.2244, 0.0158, 0.0083, 0.2244, 0.0172, 0.0061, 0.2269, 0.0167, 0.0061, 0.2258, 0.0172, 0.0083, 0.2269, 0.0167, 0.0083, 0.2258,
0.0156, 0.01, 0.2241, 0.0173, 0.01, 0.227, 0.0168, 0.01, 0.226, 0.0136, 0.01, 0.2244, 0.0114, 0.01, 0.2244, 0.0136, 0.01, 0.2266, 0.0114, 0.01, 0.2266, 0.0092, 0.01, 0.2201, 0.0071, 0.01, 0.2201, 0.0092, 0.01, 0.2222, 0.0071, 0.01, 0.2222, 0.0049, 0.01, 0.2201, 0.0027, 0.01, 0.2201, 0.0049, 0.01, 0.2222, 0.0027, 0.01, 0.2222, 0.0092, 0.01, 0.2244, 0.0071, 0.01, 0.2244, 0.0092, 0.01, 0.2266, 0.0071, 0.01, 0.2266, 0.0049, 0.01, 0.2244, 0.0027, 0.01, 0.2244, 0.0049, 0.01, 0.2266, 0.0027, 0.01, 0.2266, 0.0191, 0.0061, 0.2318, 0.0191, 0.0083, 0.2318, 0.0191, 0.01, 0.2318, 0.0194, 0.0061, 0.2331, 0.0194, 0.0083, 0.2331, 0.02, 0.0061, 0.2352, 0.02, 0.0083, 0.2352,
0.0194, 0.01, 0.233, 0.02, 0.01, 0.2351, 0.0205, 0.0061, 0.2374, 0.0205, 0.0083, 0.2374, 0.0208, 0.0061, 0.2396, 0.0208, 0.0083, 0.2396, 0.0205, 0.01, 0.2373, 0.0208, 0.01, 0.2395, 0.021, 0.0061, 0.2418, 0.021, 0.0083, 0.2418, 0.0213, 0.0061, 0.244, 0.0212, 0.0061, 0.2429, 0.0213, 0.0083, 0.244, 0.0212, 0.0083, 0.2429, 0.021, 0.01, 0.2416, 0.0214, 0.01, 0.2443, 0.0213, 0.01, 0.2436, 0.018, 0.0061, 0.2287, 0.018, 0.0083, 0.2287, 0.0187, 0.0061, 0.2307, 0.0187, 0.0083, 0.2307, 0.0179, 0.01, 0.2285, 0.0157, 0.01, 0.2287, 0.0187, 0.01, 0.2308, 0.0157, 0.01, 0.2309, 0.0136, 0.01, 0.2287, 0.0114, 0.01, 0.2287, 0.0136, 0.01, 0.2309, 0.0114, 0.01, 0.2309, 0.0179, 0.01, 0.2331,
0.0157, 0.01, 0.2331, 0.0179, 0.01, 0.2353, 0.0157, 0.01, 0.2353, 0.0136, 0.01, 0.2331, 0.0114, 0.01, 0.2331, 0.0136, 0.01, 0.2353, 0.0114, 0.01, 0.2353, 0.0026, 0.0061, 0.2289, 0.0026, 0.0083, 0.2289, 0.0047, 0.0061, 0.231, 0.0036, 0.0061, 0.23, 0.0047, 0.0083, 0.231, 0.0036, 0.0083, 0.23, 0.0092, 0.01, 0.2287, 0.0071, 0.01, 0.2287, 0.0092, 0.01, 0.2309, 0.0071, 0.01, 0.2309, 0.0049, 0.01, 0.2287, 0.0026, 0.01, 0.2289, 0.0047, 0.01, 0.231, 0.0036, 0.01, 0.23, 0.0064, 0.0061, 0.2335, 0.0064, 0.0083, 0.2335, 0.0074, 0.0061, 0.2352, 0.0074, 0.0083, 0.2352, 0.0057, 0.0061, 0.2324, 0.0057, 0.0083, 0.2324, 0.0092, 0.01, 0.2331, 0.0063, 0.01, 0.2334, 0.0092, 0.01, 0.2353,
0.0075, 0.01, 0.2354, 0.0057, 0.01, 0.2323, 0.0179, 0.01, 0.2374, 0.0157, 0.01, 0.2374, 0.0179, 0.01, 0.2396, 0.0157, 0.01, 0.2396, 0.0136, 0.01, 0.2374, 0.0114, 0.01, 0.2374, 0.0136, 0.01, 0.2396, 0.0114, 0.01, 0.2396, 0.0179, 0.01, 0.2418, 0.0157, 0.01, 0.2418, 0.0179, 0.01, 0.2439, 0.0157, 0.01, 0.2439, 0.0136, 0.01, 0.2418, 0.0114, 0.01, 0.2418, 0.0136, 0.01, 0.2439, 0.0114, 0.01, 0.2439, 0.0082, 0.0061, 0.2381, 0.0079, 0.0061, 0.2371, 0.0082, 0.0083, 0.2381, 0.0079, 0.0083, 0.2371, 0.0086, 0.0061, 0.2396, 0.0086, 0.0083, 0.2396, 0.0081, 0.01, 0.2377, 0.0079, 0.01, 0.2369, 0.0086, 0.01, 0.2398, 0.009, 0.0061, 0.2418, 0.009, 0.0083, 0.2418, 0.0093, 0.0061, 0.2439,
0.0093, 0.0083, 0.2439, 0.009, 0.01, 0.2419, 0.0093, 0.01, 0.244, 0.067, 0.0061, 0.2588, 0.067, 0.0083, 0.2588, 0.067, 0.0061, 0.2613, 0.067, 0.0083, 0.2613, 0.0656, 0.0061, 0.2588, 0.0635, 0.0061, 0.2588, 0.0656, 0.0083, 0.2588, 0.0635, 0.0083, 0.2588, 0.0631, 0.0061, 0.2618, 0.0631, 0.0083, 0.2618, 0.067, 0.01, 0.2588, 0.067, 0.01, 0.2613, 0.0656, 0.01, 0.2588, 0.0635, 0.01, 0.2588, 0.0656, 0.01, 0.2613, 0.0631, 0.01, 0.2618, 0.0613, 0.0061, 0.2588, 0.0591, 0.0061, 0.2588, 0.0613, 0.0083, 0.2588, 0.0591, 0.0083, 0.2588, 0.0613, 0.0061, 0.2618, 0.0591, 0.0061, 0.2618, 0.0613, 0.0083, 0.2618, 0.0591, 0.0083, 0.2618, 0.057, 0.0061, 0.2588, 0.0548, 0.0061, 0.2588, 0.057, 0.0083, 0.2588,
0.0548, 0.0083, 0.2588, 0.057, 0.0061, 0.2618, 0.0548, 0.0061, 0.2618, 0.057, 0.0083, 0.2618, 0.0548, 0.0083, 0.2618, 0.0613, 0.01, 0.2588, 0.0591, 0.01, 0.2588, 0.0613, 0.01, 0.2618, 0.0591, 0.01, 0.2618, 0.057, 0.01, 0.2588, 0.0548, 0.01, 0.2588, 0.057, 0.01, 0.2618, 0.0548, 0.01, 0.2618, 0.0526, 0.0061, 0.2588, 0.0505, 0.0061, 0.2588, 0.0526, 0.0083, 0.2588, 0.0505, 0.0083, 0.2588, 0.0526, 0.0061, 0.2618, 0.0505, 0.0061, 0.2618, 0.0526, 0.0083, 0.2618, 0.0505, 0.0083, 0.2618, 0.0483, 0.0061, 0.2588, 0.0461, 0.0061, 0.2588, 0.0483, 0.0083, 0.2588, 0.0461, 0.0083, 0.2588, 0.0483, 0.0061, 0.2618, 0.0461, 0.0061, 0.2618, 0.0483, 0.0083, 0.2618, 0.0461, 0.0083, 0.2618, 0.0526, 0.01, 0.2588,
0.0505, 0.01, 0.2588, 0.0526, 0.01, 0.2618, 0.0505, 0.01, 0.2618, 0.0483, 0.01, 0.2588, 0.0461, 0.01, 0.2588, 0.0483, 0.01, 0.2618, 0.0461, 0.01, 0.2618, 0.0439, 0.0061, 0.2588, 0.0425, 0.0061, 0.2588, 0.0439, 0.0083, 0.2588, 0.0425, 0.0083, 0.2588, 0.0439, 0.0061, 0.2618, 0.0425, 0.0061, 0.2618, 0.0439, 0.0083, 0.2618, 0.0425, 0.0083, 0.2618, 0.0439, 0.01, 0.2588, 0.0425, 0.01, 0.2588, 0.0439, 0.01, 0.2618, 0.0425, 0.01, 0.2618, 0.067, 0.0061, 0.2635, 0.067, 0.0083, 0.2635, 0.067, 0.0061, 0.2656, 0.067, 0.0083, 0.2656, 0.0631, 0.0061, 0.2635, 0.0631, 0.0083, 0.2635, 0.0631, 0.0061, 0.2656, 0.0631, 0.0083, 0.2656, 0.067, 0.01, 0.2635, 0.067, 0.01, 0.2656, 0.0656, 0.01, 0.2635,
0.0631, 0.01, 0.2635, 0.0656, 0.01, 0.2656, 0.0631, 0.01, 0.2656, 0.067, 0.0061, 0.2678, 0.067, 0.0083, 0.2678, 0.067, 0.0061, 0.27, 0.067, 0.0083, 0.27, 0.0631, 0.0061, 0.2678, 0.0631, 0.0083, 0.2678, 0.0631, 0.0061, 0.27, 0.0631, 0.0083, 0.27, 0.067, 0.01, 0.2678, 0.067, 0.01, 0.27, 0.0656, 0.01, 0.2678, 0.0631, 0.01, 0.2678, 0.0656, 0.01, 0.27, 0.0631, 0.01, 0.27, 0.067, 0.0061, 0.2721, 0.067, 0.0083, 0.2721, 0.067, 0.0061, 0.2743, 0.067, 0.0083, 0.2743, 0.0631, 0.0061, 0.2721, 0.0631, 0.0083, 0.2721, 0.0631, 0.0061, 0.2743, 0.0631, 0.0083, 0.2743, 0.067, 0.01, 0.2721, 0.067, 0.01, 0.2743, 0.0656, 0.01, 0.2721, 0.0631, 0.01, 0.2721, 0.0656, 0.01, 0.2743,
0.0631, 0.01, 0.2743, 0.067, 0.0061, 0.2765, 0.067, 0.0083, 0.2765, 0.067, 0.0061, 0.2794, 0.067, 0.0083, 0.2794, 0.0631, 0.0061, 0.2765, 0.0631, 0.0083, 0.2765, 0.0656, 0.0061, 0.2794, 0.0631, 0.0061, 0.2794, 0.0656, 0.0083, 0.2794, 0.0631, 0.0083, 0.2794, 0.067, 0.01, 0.2765, 0.067, 0.01, 0.2794, 0.0656, 0.01, 0.2765, 0.0631, 0.01, 0.2765, 0.0656, 0.01, 0.2794, 0.0631, 0.01, 0.2794, 0.0216, 0.0061, 0.2461, 0.0216, 0.0083, 0.2461, 0.0218, 0.0061, 0.2483, 0.0218, 0.0083, 0.2483, 0.0216, 0.01, 0.2461, 0.0201, 0.01, 0.2461, 0.0218, 0.01, 0.2482, 0.0201, 0.01, 0.2483, 0.0219, 0.0061, 0.2504, 0.0219, 0.0083, 0.2504, 0.0219, 0.0061, 0.2526, 0.0219, 0.0083, 0.2526, 0.0219, 0.01, 0.2504,
0.0201, 0.01, 0.2504, 0.0219, 0.01, 0.2526, 0.0201, 0.01, 0.2526, 0.0219, 0.0061, 0.2548, 0.0219, 0.0083, 0.2548, 0.0217, 0.0061, 0.257, 0.0217, 0.0083, 0.257, 0.0219, 0.01, 0.2548, 0.0201, 0.01, 0.2548, 0.0217, 0.01, 0.257, 0.0201, 0.01, 0.257, 0.0214, 0.0061, 0.2591, 0.0214, 0.0083, 0.2591, 0.0212, 0.0061, 0.2606, 0.0211, 0.0061, 0.2617, 0.0212, 0.0083, 0.2606, 0.0211, 0.0083, 0.2617, 0.0214, 0.01, 0.2591, 0.0201, 0.01, 0.2591, 0.0212, 0.01, 0.2605, 0.0211, 0.01, 0.2613, 0.0179, 0.01, 0.2461, 0.0157, 0.01, 0.2461, 0.0179, 0.01, 0.2483, 0.0157, 0.01, 0.2483, 0.0136, 0.01, 0.2461, 0.0114, 0.01, 0.2461, 0.0136, 0.01, 0.2483, 0.0114, 0.01, 0.2483, 0.0179, 0.01, 0.2504,
0.0157, 0.01, 0.2504, 0.0179, 0.01, 0.2526, 0.0157, 0.01, 0.2526, 0.0136, 0.01, 0.2504, 0.0114, 0.01, 0.2504, 0.0136, 0.01, 0.2526, 0.0114, 0.01, 0.2526, 0.0096, 0.0061, 0.2461, 0.0096, 0.0083, 0.2461, 0.0099, 0.0061, 0.2483, 0.0099, 0.0083, 0.2483, 0.0096, 0.01, 0.2462, 0.0099, 0.01, 0.2483, 0.0101, 0.0061, 0.2504, 0.0101, 0.0083, 0.2504, 0.0101, 0.0061, 0.2526, 0.0101, 0.0083, 0.2526, 0.0101, 0.01, 0.2504, 0.0101, 0.01, 0.2526, 0.0179, 0.01, 0.2548, 0.0157, 0.01, 0.2548, 0.0179, 0.01, 0.257, 0.0157, 0.01, 0.257, 0.0136, 0.01, 0.2548, 0.0114, 0.01, 0.2548, 0.0136, 0.01, 0.257, 0.0114, 0.01, 0.257, 0.0179, 0.01, 0.2591, 0.0157, 0.01, 0.2591, 0.0179, 0.01, 0.2613,
0.0157, 0.01, 0.2613, 0.0136, 0.01, 0.2591, 0.0114, 0.01, 0.2591, 0.0136, 0.01, 0.2613, 0.0114, 0.01, 0.2613, 0.0101, 0.0061, 0.2548, 0.0101, 0.0083, 0.2548, 0.01, 0.0061, 0.257, 0.01, 0.0083, 0.257, 0.0101, 0.01, 0.2548, 0.01, 0.01, 0.2569, 0.0097, 0.0061, 0.2591, 0.0097, 0.0083, 0.2591, 0.0093, 0.0061, 0.2613, 0.0093, 0.0083, 0.2613, 0.0097, 0.01, 0.2591, 0.0093, 0.01, 0.2612, 0.0209, 0.0061, 0.2635, 0.0209, 0.0083, 0.2635, 0.0206, 0.0061, 0.2656, 0.0206, 0.0083, 0.2656, 0.0209, 0.01, 0.2636, 0.0206, 0.01, 0.2658, 0.0202, 0.0061, 0.2678, 0.0202, 0.0083, 0.2678, 0.0197, 0.0061, 0.27, 0.0197, 0.0083, 0.27, 0.0202, 0.01, 0.2679, 0.0197, 0.01, 0.2701, 0.0192, 0.0061, 0.2717,
0.0192, 0.0083, 0.2717, 0.0192, 0.01, 0.2715, 0.0179, 0.01, 0.2635, 0.0157, 0.01, 0.2635, 0.0179, 0.01, 0.2656, 0.0157, 0.01, 0.2656, 0.0136, 0.01, 0.2635, 0.0114, 0.01, 0.2635, 0.0136, 0.01, 0.2656, 0.0114, 0.01, 0.2656, 0.0179, 0.01, 0.2678, 0.0157, 0.01, 0.2678, 0.0179, 0.01, 0.27, 0.0157, 0.01, 0.27, 0.0136, 0.01, 0.2678, 0.0114, 0.01, 0.2678, 0.0136, 0.01, 0.27, 0.0114, 0.01, 0.27, 0.0089, 0.0061, 0.2635, 0.0089, 0.0083, 0.2635, 0.0085, 0.0061, 0.2655, 0.0081, 0.0061, 0.2666, 0.0085, 0.0083, 0.2655, 0.0081, 0.0083, 0.2666, 0.009, 0.01, 0.2633, 0.0084, 0.01, 0.2656, 0.0081, 0.01, 0.2666, 0.0076, 0.0061, 0.2678, 0.0076, 0.0083, 0.2678, 0.0066, 0.0061, 0.27,
0.0066, 0.0083, 0.27, 0.0059, 0.0061, 0.271, 0.0059, 0.0083, 0.271, 0.0092, 0.01, 0.2678, 0.0077, 0.01, 0.2677, 0.0092, 0.01, 0.27, 0.0066, 0.01, 0.2699, 0.0059, 0.01, 0.271, 0.0189, 0.0061, 0.2728, 0.0189, 0.0083, 0.2728, 0.0183, 0.0061, 0.2743, 0.0183, 0.0083, 0.2743, 0.019, 0.01, 0.2725, 0.0157, 0.01, 0.2721, 0.0182, 0.01, 0.2746, 0.0157, 0.01, 0.2743, 0.0136, 0.01, 0.2721, 0.0114, 0.01, 0.2721, 0.0136, 0.01, 0.2743, 0.0114, 0.01, 0.2743, 0.0175, 0.0061, 0.2765, 0.0175, 0.0083, 0.2765, 0.0169, 0.0061, 0.2776, 0.0163, 0.0061, 0.2788, 0.0169, 0.0083, 0.2776, 0.0163, 0.0083, 0.2788, 0.0174, 0.01, 0.2766, 0.0157, 0.01, 0.2765, 0.0169, 0.01, 0.2776, 0.0163, 0.01, 0.2788,
0.0136, 0.01, 0.2765, 0.0114, 0.01, 0.2765, 0.0136, 0.01, 0.2787, 0.0114, 0.01, 0.2787, 0.0051, 0.0061, 0.2722, 0.0051, 0.0083, 0.2722, 0.004, 0.0061, 0.2735, 0.0029, 0.0061, 0.2746, 0.004, 0.0083, 0.2735, 0.0029, 0.0083, 0.2746, 0.0092, 0.01, 0.2721, 0.0071, 0.01, 0.2721, 0.0092, 0.01, 0.2743, 0.0071, 0.01, 0.2743, 0.0053, 0.01, 0.272, 0.004, 0.01, 0.2735, 0.0029, 0.01, 0.2746, 0.0017, 0.0061, 0.2755, 0.0017, 0.0083, 0.2755, 0.0092, 0.01, 0.2765, 0.0071, 0.01, 0.2765, 0.0092, 0.01, 0.2787, 0.0071, 0.01, 0.2787, 0.0049, 0.01, 0.2765, 0.0019, 0.01, 0.2754, 0.0049, 0.01, 0.2787, 0.0027, 0.01, 0.2787, 0.067, 0.0061, 0.298, 0.067, 0.0083, 0.298, 0.067, 0.0061, 0.3004,
0.067, 0.0083, 0.3004, 0.0656, 0.0061, 0.298, 0.0635, 0.0061, 0.298, 0.0656, 0.0083, 0.298, 0.0635, 0.0083, 0.298, 0.067, 0.01, 0.298, 0.067, 0.01, 0.3004, 0.0656, 0.01, 0.298, 0.0635, 0.01, 0.298, 0.0656, 0.01, 0.3004, 0.0635, 0.01, 0.3004, 0.067, 0.0061, 0.3025, 0.067, 0.0083, 0.3025, 0.067, 0.0061, 0.3047, 0.067, 0.0083, 0.3047, 0.0631, 0.0061, 0.302, 0.0631, 0.0083, 0.302, 0.0631, 0.0061, 0.3047, 0.0631, 0.0083, 0.3047, 0.067, 0.01, 0.3025, 0.067, 0.01, 0.3047, 0.0656, 0.01, 0.3025, 0.0631, 0.01, 0.302, 0.0656, 0.01, 0.3047, 0.0631, 0.01, 0.3047, 0.0613, 0.0061, 0.298, 0.0591, 0.0061, 0.298, 0.0613, 0.0083, 0.298, 0.0591, 0.0083, 0.298, 0.057, 0.0061, 0.298,
0.0548, 0.0061, 0.298, 0.057, 0.0083, 0.298, 0.0548, 0.0083, 0.298, 0.0613, 0.01, 0.298, 0.0591, 0.01, 0.298, 0.0613, 0.01, 0.3004, 0.0591, 0.01, 0.3004, 0.057, 0.01, 0.298, 0.0548, 0.01, 0.298, 0.057, 0.01, 0.3004, 0.0548, 0.01, 0.3004, 0.0613, 0.0061, 0.302, 0.0591, 0.0061, 0.302, 0.0613, 0.0083, 0.302, 0.0591, 0.0083, 0.302, 0.057, 0.0061, 0.302, 0.0548, 0.0061, 0.302, 0.057, 0.0083, 0.302, 0.0548, 0.0083, 0.302, 0.0562, 0.0061, 0.3049, 0.0548, 0.0061, 0.3049, 0.0562, 0.0083, 0.3049, 0.0548, 0.0083, 0.3049, 0.0613, 0.01, 0.302, 0.0591, 0.01, 0.302, 0.057, 0.01, 0.302, 0.0548, 0.01, 0.302, 0.0562, 0.01, 0.3049, 0.0548, 0.01, 0.3049, 0.067, 0.0061, 0.3069,
0.067, 0.0083, 0.3069, 0.067, 0.0061, 0.309, 0.067, 0.0083, 0.309, 0.0631, 0.0061, 0.3069, 0.0631, 0.0083, 0.3069, 0.0631, 0.0061, 0.309, 0.0631, 0.0083, 0.309, 0.067, 0.01, 0.3069, 0.067, 0.01, 0.309, 0.0656, 0.01, 0.3069, 0.0631, 0.01, 0.3069, 0.0656, 0.01, 0.309, 0.0631, 0.01, 0.309, 0.067, 0.0061, 0.3112, 0.067, 0.0083, 0.3112, 0.067, 0.0061, 0.3134, 0.067, 0.0083, 0.3134, 0.0631, 0.0061, 0.3112, 0.0631, 0.0083, 0.3112, 0.0631, 0.0061, 0.3134, 0.0631, 0.0083, 0.3134, 0.067, 0.01, 0.3112, 0.067, 0.01, 0.3134, 0.0656, 0.01, 0.3112, 0.0631, 0.01, 0.3112, 0.0656, 0.01, 0.3134, 0.0631, 0.01, 0.3134, 0.0562, 0.0061, 0.3069, 0.0562, 0.0083, 0.3069, 0.0562, 0.0061, 0.309,
0.0562, 0.0083, 0.309, 0.0562, 0.01, 0.3069, 0.0548, 0.01, 0.3069, 0.0562, 0.01, 0.309, 0.0548, 0.01, 0.309, 0.0562, 0.0061, 0.3112, 0.0562, 0.0083, 0.3112, 0.0562, 0.0061, 0.3134, 0.0562, 0.0083, 0.3134, 0.0562, 0.01, 0.3112, 0.0548, 0.01, 0.3112, 0.0562, 0.01, 0.3134, 0.0548, 0.01, 0.3134, 0.0526, 0.0061, 0.298, 0.0505, 0.0061, 0.298, 0.0526, 0.0083, 0.298, 0.0505, 0.0083, 0.298, 0.0483, 0.0061, 0.298, 0.0461, 0.0061, 0.298, 0.0483, 0.0083, 0.298, 0.0461, 0.0083, 0.298, 0.0526, 0.01, 0.298, 0.0505, 0.01, 0.298, 0.0526, 0.01, 0.3004, 0.0505, 0.01, 0.3004, 0.0483, 0.01, 0.298, 0.0461, 0.01, 0.298, 0.0483, 0.01, 0.3004, 0.0461, 0.01, 0.3004, 0.0526, 0.0061, 0.302,
0.0505, 0.0061, 0.302, 0.0526, 0.0083, 0.302, 0.0505, 0.0083, 0.302, 0.0523, 0.0061, 0.3049, 0.0523, 0.0083, 0.3049, 0.0483, 0.0061, 0.302, 0.0454, 0.0061, 0.302, 0.0483, 0.0083, 0.302, 0.0454, 0.0083, 0.302, 0.0454, 0.0061, 0.3047, 0.0454, 0.0083, 0.3047, 0.0526, 0.01, 0.302, 0.0505, 0.01, 0.302, 0.0523, 0.01, 0.3049, 0.0483, 0.01, 0.302, 0.0454, 0.01, 0.302, 0.0454, 0.01, 0.3047, 0.0439, 0.0061, 0.298, 0.0425, 0.0061, 0.298, 0.0439, 0.0083, 0.298, 0.0425, 0.0083, 0.298, 0.0425, 0.0061, 0.3004, 0.0425, 0.0083, 0.3004, 0.0439, 0.01, 0.298, 0.0425, 0.01, 0.298, 0.0439, 0.01, 0.3004, 0.0425, 0.01, 0.3004, 0.0425, 0.0061, 0.3025, 0.0425, 0.0083, 0.3025, 0.0425, 0.0061, 0.3047,
0.0425, 0.0083, 0.3047, 0.0439, 0.01, 0.3025, 0.0425, 0.01, 0.3025, 0.0439, 0.01, 0.3047, 0.0425, 0.01, 0.3047, 0.0523, 0.0061, 0.3069, 0.0523, 0.0083, 0.3069, 0.0523, 0.0061, 0.309, 0.0523, 0.0083, 0.309, 0.0454, 0.0061, 0.3069, 0.0454, 0.0083, 0.3069, 0.0454, 0.0061, 0.309, 0.0454, 0.0083, 0.309, 0.0523, 0.01, 0.3069, 0.0523, 0.01, 0.309, 0.0454, 0.01, 0.3069, 0.0454, 0.01, 0.309, 0.0523, 0.0061, 0.3112, 0.0523, 0.0083, 0.3112, 0.0523, 0.0061, 0.3134, 0.0523, 0.0083, 0.3134, 0.0454, 0.0061, 0.3112, 0.0454, 0.0083, 0.3112, 0.0454, 0.0061, 0.3134, 0.0454, 0.0083, 0.3134, 0.0523, 0.01, 0.3112, 0.0523, 0.01, 0.3134, 0.0454, 0.01, 0.3112, 0.0454, 0.01, 0.3134, 0.0425, 0.0061, 0.3069,
0.0425, 0.0083, 0.3069, 0.0425, 0.0061, 0.309, 0.0425, 0.0083, 0.309, 0.0439, 0.01, 0.3069, 0.0425, 0.01, 0.3069, 0.0439, 0.01, 0.309, 0.0425, 0.01, 0.309, 0.0425, 0.0061, 0.3112, 0.0425, 0.0083, 0.3112, 0.0425, 0.0061, 0.3134, 0.0425, 0.0083, 0.3134, 0.0439, 0.01, 0.3112, 0.0425, 0.01, 0.3112, 0.0439, 0.01, 0.3134, 0.0425, 0.01, 0.3134, 0.0152, 0.0061, 0.2805, 0.0152, 0.0083, 0.2805, 0.0143, 0.0061, 0.2815, 0.0143, 0.0083, 0.2815, 0.0133, 0.0061, 0.2828, 0.0122, 0.0061, 0.2838, 0.0133, 0.0083, 0.2828, 0.0122, 0.0083, 0.2838, 0.0152, 0.01, 0.2804, 0.0144, 0.01, 0.2815, 0.0114, 0.01, 0.2808, 0.0133, 0.01, 0.2827, 0.0122, 0.01, 0.2838, 0.0111, 0.0061, 0.2847, 0.0111, 0.0083, 0.2847,
0.0111, 0.01, 0.2848, 0.0092, 0.01, 0.2808, 0.0071, 0.01, 0.2808, 0.0092, 0.01, 0.283, 0.0071, 0.01, 0.283, 0.0049, 0.01, 0.2808, 0.0027, 0.01, 0.2808, 0.0049, 0.01, 0.283, 0.0027, 0.01, 0.283, 0.0097, 0.0061, 0.2858, 0.0097, 0.0083, 0.2858, 0.0086, 0.0061, 0.2865, 0.0071, 0.0061, 0.2874, 0.0086, 0.0083, 0.2865, 0.0071, 0.0083, 0.2874, 0.0054, 0.0061, 0.2882, 0.0054, 0.0083, 0.2882, 0.0096, 0.01, 0.2858, 0.0071, 0.01, 0.2852, 0.0085, 0.01, 0.2866, 0.0073, 0.01, 0.2873, 0.0049, 0.01, 0.2852, 0.0027, 0.01, 0.2852, 0.0052, 0.01, 0.2883, 0.0027, 0.01, 0.2873, 0.0043, 0.0061, 0.2886, 0.0027, 0.0061, 0.2892, 0.0043, 0.0083, 0.2886, 0.0027, 0.0083, 0.2892, 0.0042, 0.01, 0.2887,
0.0028, 0.01, 0.2891, 0.0219, 0.0061, 0.3088, 0.0201, 0.0061, 0.3088, 0.0219, 0.0083, 0.3088, 0.0201, 0.0083, 0.3088, 0.0219, 0.01, 0.3088, 0.0201, 0.01, 0.3088, 0.0219, 0.0061, 0.3112, 0.0219, 0.0083, 0.3112, 0.0219, 0.0061, 0.3134, 0.0219, 0.0083, 0.3134, 0.0219, 0.01, 0.3112, 0.0201, 0.01, 0.3112, 0.0219, 0.01, 0.3134, 0.0201, 0.01, 0.3134, 0.0179, 0.0061, 0.3088, 0.0157, 0.0061, 0.3088, 0.0179, 0.0083, 0.3088, 0.0157, 0.0083, 0.3088, 0.0136, 0.0061, 0.3088, 0.0114, 0.0061, 0.3088, 0.0136, 0.0083, 0.3088, 0.0114, 0.0083, 0.3088, 0.0179, 0.01, 0.3088, 0.0157, 0.01, 0.3088, 0.0136, 0.01, 0.3088, 0.0114, 0.01, 0.3088, 0.0179, 0.01, 0.3112, 0.0157, 0.01, 0.3112, 0.0179, 0.01, 0.3134,
0.0157, 0.01, 0.3134, 0.0136, 0.01, 0.3112, 0.0114, 0.01, 0.3112, 0.0136, 0.01, 0.3134, 0.0114, 0.01, 0.3134, 0.0092, 0.0061, 0.3088, 0.0071, 0.0061, 0.3088, 0.0092, 0.0083, 0.3088, 0.0071, 0.0083, 0.3088, 0.0049, 0.0061, 0.3088, 0.0027, 0.0061, 0.3088, 0.0049, 0.0083, 0.3088, 0.0027, 0.0083, 0.3088, 0.0092, 0.01, 0.3088, 0.0071, 0.01, 0.3088, 0.0049, 0.01, 0.3088, 0.0027, 0.01, 0.3088, 0.0092, 0.01, 0.3112, 0.0071, 0.01, 0.3112, 0.0092, 0.01, 0.3134, 0.0071, 0.01, 0.3134, 0.0049, 0.01, 0.3112, 0.0027, 0.01, 0.3112, 0.0049, 0.01, 0.3134, 0.0027, 0.01, 0.3134, 0.067, 0.0061, 0.3155, 0.067, 0.0083, 0.3155, 0.067, 0.0061, 0.3177, 0.067, 0.0083, 0.3177, 0.0631, 0.0061, 0.3155,
0.0631, 0.0083, 0.3155, 0.0631, 0.0061, 0.3177, 0.0631, 0.0083, 0.3177, 0.067, 0.01, 0.3155, 0.067, 0.01, 0.3177, 0.0656, 0.01, 0.3155, 0.0631, 0.01, 0.3155, 0.0656, 0.01, 0.3177, 0.0631, 0.01, 0.3177, 0.067, 0.0061, 0.3196, 0.067, 0.0083, 0.3196, 0.0656, 0.0061, 0.3196, 0.0631, 0.0061, 0.3196, 0.0656, 0.0083, 0.3196, 0.0631, 0.0083, 0.3196, 0.067, 0.01, 0.3196, 0.0656, 0.01, 0.3196, 0.0631, 0.01, 0.3196, 0.0562, 0.0061, 0.3155, 0.0562, 0.0083, 0.3155, 0.0562, 0.0061, 0.3176, 0.0548, 0.0061, 0.3176, 0.0562, 0.0083, 0.3176, 0.0548, 0.0083, 0.3176, 0.0562, 0.01, 0.3155, 0.0548, 0.01, 0.3155, 0.0562, 0.01, 0.3176, 0.0548, 0.01, 0.3176, 0.0523, 0.0061, 0.3155, 0.0523, 0.0083, 0.3155,
0.0523, 0.0061, 0.3176, 0.0523, 0.0083, 0.3176, 0.0454, 0.0061, 0.3155, 0.0454, 0.0083, 0.3155, 0.0454, 0.0061, 0.3177, 0.0454, 0.0083, 0.3177, 0.0523, 0.01, 0.3155, 0.0523, 0.01, 0.3176, 0.0454, 0.01, 0.3155, 0.0454, 0.01, 0.3177, 0.0454, 0.0061, 0.3196, 0.0454, 0.0083, 0.3196, 0.0454, 0.01, 0.3196, 0.0425, 0.0061, 0.3155, 0.0425, 0.0083, 0.3155, 0.0425, 0.0061, 0.3177, 0.0425, 0.0083, 0.3177, 0.0439, 0.01, 0.3155, 0.0425, 0.01, 0.3155, 0.0439, 0.01, 0.3177, 0.0425, 0.01, 0.3177, 0.0439, 0.0061, 0.3196, 0.0425, 0.0061, 0.3196, 0.0439, 0.0083, 0.3196, 0.0425, 0.0083, 0.3196, 0.0439, 0.01, 0.3196, 0.0425, 0.01, 0.3196, 0.067, 0.0061, 0.3402, 0.067, 0.0083, 0.3402, 0.0656, 0.0061, 0.3402,
0.0635, 0.0061, 0.3402, 0.0656, 0.0083, 0.3402, 0.0635, 0.0083, 0.3402, 0.067, 0.01, 0.3402, 0.0656, 0.01, 0.3402, 0.0635, 0.01, 0.3402, 0.0613, 0.0061, 0.3402, 0.0591, 0.0061, 0.3402, 0.0613, 0.0083, 0.3402, 0.0591, 0.0083, 0.3402, 0.057, 0.0061, 0.3402, 0.0548, 0.0061, 0.3402, 0.057, 0.0083, 0.3402, 0.0548, 0.0083, 0.3402, 0.0613, 0.01, 0.3402, 0.0591, 0.01, 0.3402, 0.057, 0.01, 0.3402, 0.0548, 0.01, 0.3402, 0.067, 0.0061, 0.3416, 0.067, 0.0083, 0.3416, 0.067, 0.0061, 0.3431, 0.067, 0.0083, 0.3431, 0.0656, 0.0061, 0.3431, 0.0635, 0.0061, 0.3431, 0.0656, 0.0083, 0.3431, 0.0635, 0.0083, 0.3431, 0.067, 0.01, 0.3416, 0.067, 0.01, 0.3431, 0.0656, 0.01, 0.3416, 0.0635, 0.01, 0.3416,
0.0656, 0.01, 0.3431, 0.0635, 0.01, 0.3431, 0.0613, 0.0061, 0.3431, 0.0591, 0.0061, 0.3431, 0.0613, 0.0083, 0.3431, 0.0591, 0.0083, 0.3431, 0.057, 0.0061, 0.3431, 0.0548, 0.0061, 0.3431, 0.057, 0.0083, 0.3431, 0.0548, 0.0083, 0.3431, 0.0613, 0.01, 0.3416, 0.0591, 0.01, 0.3416, 0.0613, 0.01, 0.3431, 0.0591, 0.01, 0.3431, 0.057, 0.01, 0.3416, 0.0548, 0.01, 0.3416, 0.057, 0.01, 0.3431, 0.0548, 0.01, 0.3431, 0.0545, 0.0061, 0.3483, 0.0545, 0.0083, 0.3483, 0.0546, 0.01, 0.3483, 0.0526, 0.0061, 0.3402, 0.0505, 0.0061, 0.3402, 0.0526, 0.0083, 0.3402, 0.0505, 0.0083, 0.3402, 0.0483, 0.0061, 0.3402, 0.0461, 0.0061, 0.3402, 0.0483, 0.0083, 0.3402, 0.0461, 0.0083, 0.3402, 0.0526, 0.01, 0.3402,
0.0505, 0.01, 0.3402, 0.0483, 0.01, 0.3402, 0.0461, 0.01, 0.3402, 0.0439, 0.0061, 0.3402, 0.0425, 0.0061, 0.3402, 0.0439, 0.0083, 0.3402, 0.0425, 0.0083, 0.3402, 0.0439, 0.01, 0.3402, 0.0425, 0.01, 0.3402, 0.0526, 0.0061, 0.3431, 0.0499, 0.0061, 0.3431, 0.0526, 0.0083, 0.3431, 0.0499, 0.0083, 0.3431, 0.0526, 0.01, 0.3416, 0.0505, 0.01, 0.3416, 0.0526, 0.01, 0.3431, 0.0499, 0.01, 0.3431, 0.0483, 0.01, 0.3416, 0.0461, 0.01, 0.3416, 0.0483, 0.01, 0.3438, 0.0461, 0.01, 0.3438, 0.0524, 0.0061, 0.346, 0.0515, 0.0061, 0.3449, 0.0524, 0.0083, 0.346, 0.0515, 0.0083, 0.3449, 0.0535, 0.0061, 0.3472, 0.0535, 0.0083, 0.3472, 0.0457, 0.0061, 0.3462, 0.0457, 0.0083, 0.3462, 0.0479, 0.0061, 0.3485,
0.0468, 0.0061, 0.3474, 0.0479, 0.0083, 0.3485, 0.0468, 0.0083, 0.3474, 0.0525, 0.01, 0.346, 0.0515, 0.01, 0.345, 0.0535, 0.01, 0.3472, 0.0505, 0.01, 0.3481, 0.0483, 0.01, 0.3459, 0.0457, 0.01, 0.3462, 0.0479, 0.01, 0.3485, 0.0468, 0.01, 0.3474, 0.043, 0.0061, 0.3423, 0.0426, 0.0061, 0.3413, 0.043, 0.0083, 0.3423, 0.0426, 0.0083, 0.3413, 0.0438, 0.0061, 0.3438, 0.0438, 0.0083, 0.3438, 0.0429, 0.01, 0.3421, 0.0426, 0.01, 0.3411, 0.0439, 0.01, 0.3441, 0.0448, 0.0061, 0.3452, 0.0448, 0.0083, 0.3452, 0.0448, 0.01, 0.3451, 0.0219, 0.0061, 0.3155, 0.0219, 0.0083, 0.3155, 0.0219, 0.0061, 0.3177, 0.0219, 0.0083, 0.3177, 0.0219, 0.01, 0.3155, 0.0201, 0.01, 0.3155, 0.0219, 0.01, 0.3177,
0.0201, 0.01, 0.3177, 0.0219, 0.0061, 0.3206, 0.0201, 0.0061, 0.3206, 0.0219, 0.0083, 0.3206, 0.0201, 0.0083, 0.3206, 0.0219, 0.01, 0.3206, 0.0201, 0.01, 0.3206, 0.0179, 0.01, 0.3155, 0.0157, 0.01, 0.3155, 0.0179, 0.01, 0.3177, 0.0157, 0.01, 0.3177, 0.0136, 0.01, 0.3155, 0.0114, 0.01, 0.3155, 0.0136, 0.01, 0.3177, 0.0114, 0.01, 0.3177, 0.0179, 0.0061, 0.3206, 0.0157, 0.0061, 0.3206, 0.0179, 0.0083, 0.3206, 0.0157, 0.0083, 0.3206, 0.0136, 0.0061, 0.3206, 0.0114, 0.0061, 0.3206, 0.0136, 0.0083, 0.3206, 0.0114, 0.0083, 0.3206, 0.0179, 0.01, 0.3206, 0.0157, 0.01, 0.3206, 0.0136, 0.01, 0.3206, 0.0114, 0.01, 0.3206, 0.0092, 0.01, 0.3155, 0.0071, 0.01, 0.3155, 0.0092, 0.01, 0.3177,
0.0071, 0.01, 0.3177, 0.0049, 0.01, 0.3155, 0.0027, 0.01, 0.3155, 0.0049, 0.01, 0.3177, 0.0027, 0.01, 0.3177, 0.0092, 0.0061, 0.3206, 0.0071, 0.0061, 0.3206, 0.0092, 0.0083, 0.3206, 0.0071, 0.0083, 0.3206, 0.0049, 0.0061, 0.3206, 0.0027, 0.0061, 0.3206, 0.0049, 0.0083, 0.3206, 0.0027, 0.0083, 0.3206, 0.0092, 0.01, 0.3206, 0.0071, 0.01, 0.3206, 0.0049, 0.01, 0.3206, 0.0027, 0.01, 0.3206, 0.0219, 0.0061, 0.3402, 0.0201, 0.0061, 0.3402, 0.0219, 0.0083, 0.3402, 0.0201, 0.0083, 0.3402, 0.0219, 0.01, 0.3402, 0.0201, 0.01, 0.3402, 0.0219, 0.0061, 0.3416, 0.0219, 0.0083, 0.3416, 0.0219, 0.0061, 0.3438, 0.0219, 0.0083, 0.3438, 0.0219, 0.01, 0.3416, 0.0201, 0.01, 0.3416, 0.0219, 0.01, 0.3438,
0.0201, 0.01, 0.3438, 0.0219, 0.0061, 0.3459, 0.0219, 0.0083, 0.3459, 0.0219, 0.0061, 0.3481, 0.0219, 0.0083, 0.3481, 0.0219, 0.01, 0.3459, 0.0201, 0.01, 0.3459, 0.0219, 0.01, 0.3481, 0.0201, 0.01, 0.3481, 0.0179, 0.0061, 0.3402, 0.0157, 0.0061, 0.3402, 0.0179, 0.0083, 0.3402, 0.0157, 0.0083, 0.3402, 0.0136, 0.0061, 0.3402, 0.0114, 0.0061, 0.3402, 0.0136, 0.0083, 0.3402, 0.0114, 0.0083, 0.3402, 0.0179, 0.01, 0.3402, 0.0157, 0.01, 0.3402, 0.0136, 0.01, 0.3402, 0.0114, 0.01, 0.3402, 0.0092, 0.0061, 0.3402, 0.0071, 0.0061, 0.3402, 0.0092, 0.0083, 0.3402, 0.0071, 0.0083, 0.3402, 0.0049, 0.0061, 0.3402, 0.0027, 0.0061, 0.3402, 0.0049, 0.0083, 0.3402, 0.0027, 0.0083, 0.3402, 0.0092, 0.01, 0.3402,
0.0071, 0.01, 0.3402, 0.0049, 0.01, 0.3402, 0.0027, 0.01, 0.3402, 0.0179, 0.01, 0.3416, 0.0157, 0.01, 0.3416, 0.0179, 0.01, 0.3438, 0.0157, 0.01, 0.3438, 0.0136, 0.01, 0.3416, 0.0114, 0.01, 0.3416, 0.0136, 0.01, 0.3438, 0.0114, 0.01, 0.3438, 0.0179, 0.01, 0.3459, 0.0157, 0.01, 0.3459, 0.0179, 0.01, 0.3481, 0.0157, 0.01, 0.3481, 0.0136, 0.01, 0.3459, 0.0114, 0.01, 0.3459, 0.0136, 0.01, 0.3481, 0.0114, 0.01, 0.3481, 0.0092, 0.01, 0.3416, 0.0071, 0.01, 0.3416, 0.0092, 0.01, 0.3438, 0.0071, 0.01, 0.3438, 0.0049, 0.01, 0.3416, 0.0027, 0.01, 0.3416, 0.0049, 0.01, 0.3438, 0.0027, 0.01, 0.3438, 0.0092, 0.01, 0.3459, 0.0071, 0.01, 0.3459, 0.0092, 0.01, 0.3481,
0.0071, 0.01, 0.3481, 0.0049, 0.01, 0.3459, 0.0027, 0.01, 0.3459, 0.0049, 0.01, 0.3481, 0.0027, 0.01, 0.3481, 0.0631, 0.0061, 0.357, 0.0631, 0.0083, 0.357, 0.0631, 0.01, 0.3571, 0.0589, 0.0061, 0.3527, 0.0589, 0.0083, 0.3527, 0.0567, 0.0061, 0.3505, 0.0556, 0.0061, 0.3494, 0.0567, 0.0083, 0.3505, 0.0556, 0.0083, 0.3494, 0.0578, 0.0061, 0.3516, 0.0578, 0.0083, 0.3516, 0.0589, 0.01, 0.3527, 0.0567, 0.01, 0.3505, 0.0556, 0.01, 0.3494, 0.0578, 0.01, 0.3516, 0.0548, 0.01, 0.3524, 0.061, 0.0061, 0.3549, 0.06, 0.0061, 0.3538, 0.061, 0.0083, 0.3549, 0.06, 0.0083, 0.3538, 0.0621, 0.0061, 0.356, 0.0621, 0.0083, 0.356, 0.0544, 0.0061, 0.355, 0.0544, 0.0083, 0.355, 0.0565, 0.0061, 0.3572,
0.0554, 0.0061, 0.3562, 0.0565, 0.0083, 0.3572, 0.0554, 0.0083, 0.3562, 0.061, 0.01, 0.3549, 0.06, 0.01, 0.3538, 0.0621, 0.01, 0.356, 0.0591, 0.01, 0.3568, 0.057, 0.01, 0.3546, 0.0544, 0.01, 0.355, 0.0565, 0.01, 0.3572, 0.0554, 0.01, 0.3561, 0.065, 0.0061, 0.3594, 0.0642, 0.0061, 0.3583, 0.065, 0.0083, 0.3594, 0.0642, 0.0083, 0.3583, 0.0662, 0.0061, 0.361, 0.0662, 0.0083, 0.361, 0.0651, 0.01, 0.3594, 0.0642, 0.01, 0.3583, 0.066, 0.01, 0.3608, 0.0635, 0.01, 0.3611, 0.067, 0.0061, 0.3637, 0.0635, 0.0061, 0.3637, 0.067, 0.0083, 0.3637, 0.0635, 0.0083, 0.3637, 0.067, 0.01, 0.3637, 0.0635, 0.01, 0.3637, 0.0584, 0.0061, 0.3596, 0.0584, 0.0083, 0.3596, 0.0591, 0.0061, 0.3608,
0.0591, 0.0083, 0.3608, 0.0575, 0.0061, 0.3585, 0.0575, 0.0083, 0.3585, 0.057, 0.0061, 0.3608, 0.0548, 0.0061, 0.3608, 0.057, 0.0083, 0.3608, 0.0548, 0.0083, 0.3608, 0.0613, 0.01, 0.3589, 0.0583, 0.01, 0.3595, 0.0613, 0.01, 0.3611, 0.0591, 0.01, 0.3608, 0.0575, 0.01, 0.3585, 0.057, 0.01, 0.3608, 0.0548, 0.01, 0.3608, 0.0613, 0.0061, 0.3637, 0.0591, 0.0061, 0.3637, 0.0613, 0.0083, 0.3637, 0.0591, 0.0083, 0.3637, 0.057, 0.0061, 0.3637, 0.0548, 0.0061, 0.3637, 0.057, 0.0083, 0.3637, 0.0548, 0.0083, 0.3637, 0.0613, 0.01, 0.3637, 0.0591, 0.01, 0.3637, 0.057, 0.01, 0.3637, 0.0548, 0.01, 0.3637, 0.05, 0.0061, 0.3507, 0.05, 0.0083, 0.3507, 0.0522, 0.0061, 0.3529, 0.0511, 0.0061, 0.3518,
0.0522, 0.0083, 0.3529, 0.0511, 0.0083, 0.3518, 0.049, 0.0061, 0.3496, 0.049, 0.0083, 0.3496, 0.0526, 0.01, 0.3503, 0.05, 0.01, 0.3507, 0.0522, 0.01, 0.3529, 0.0511, 0.01, 0.3518, 0.0489, 0.01, 0.3496, 0.0533, 0.0061, 0.354, 0.0533, 0.0083, 0.354, 0.0533, 0.01, 0.354, 0.0526, 0.0061, 0.3608, 0.0505, 0.0061, 0.3608, 0.0526, 0.0083, 0.3608, 0.0505, 0.0083, 0.3608, 0.0483, 0.0061, 0.3608, 0.0461, 0.0061, 0.3608, 0.0483, 0.0083, 0.3608, 0.0461, 0.0083, 0.3608, 0.0526, 0.01, 0.3608, 0.0505, 0.01, 0.3608, 0.0483, 0.01, 0.3608, 0.0461, 0.01, 0.3608, 0.0526, 0.0061, 0.3637, 0.0505, 0.0061, 0.3637, 0.0526, 0.0083, 0.3637, 0.0505, 0.0083, 0.3637, 0.0483, 0.0061, 0.3637, 0.0461, 0.0061, 0.3637,
0.0483, 0.0083, 0.3637, 0.0461, 0.0083, 0.3637, 0.0526, 0.01, 0.3637, 0.0505, 0.01, 0.3637, 0.0483, 0.01, 0.3637, 0.0461, 0.01, 0.3637, 0.0439, 0.0061, 0.3608, 0.0425, 0.0061, 0.3608, 0.0439, 0.0083, 0.3608, 0.0425, 0.0083, 0.3608, 0.0439, 0.01, 0.3608, 0.0425, 0.01, 0.3608, 0.0439, 0.0061, 0.3637, 0.0425, 0.0061, 0.3637, 0.0439, 0.0083, 0.3637, 0.0425, 0.0083, 0.3637, 0.0439, 0.01, 0.3637, 0.0425, 0.01, 0.3637, 0.0219, 0.0061, 0.3503, 0.0219, 0.0083, 0.3503, 0.0219, 0.0061, 0.3524, 0.0219, 0.0083, 0.3524, 0.0219, 0.01, 0.3503, 0.0201, 0.01, 0.3503, 0.0219, 0.01, 0.3524, 0.0201, 0.01, 0.3524, 0.0219, 0.0061, 0.3546, 0.0219, 0.0083, 0.3546, 0.0219, 0.0061, 0.3568, 0.0219, 0.0083, 0.3568,
0.0219, 0.01, 0.3546, 0.0201, 0.01, 0.3546, 0.0219, 0.01, 0.3568, 0.0201, 0.01, 0.3568, 0.0219, 0.0061, 0.3589, 0.0219, 0.0083, 0.3589, 0.0219, 0.0061, 0.3611, 0.0219, 0.0083, 0.3611, 0.0219, 0.01, 0.3589, 0.0201, 0.01, 0.3589, 0.0219, 0.01, 0.3611, 0.0201, 0.01, 0.3611, 0.0219, 0.0061, 0.3633, 0.0219, 0.0083, 0.3633, 0.0219, 0.0061, 0.3655, 0.0219, 0.0083, 0.3655, 0.0219, 0.01, 0.3633, 0.0201, 0.01, 0.3633, 0.0219, 0.01, 0.3655, 0.0201, 0.01, 0.3655, 0.0179, 0.01, 0.3503, 0.0157, 0.01, 0.3503, 0.0179, 0.01, 0.3524, 0.0157, 0.01, 0.3524, 0.0136, 0.01, 0.3503, 0.0114, 0.01, 0.3503, 0.0136, 0.01, 0.3524, 0.0114, 0.01, 0.3524, 0.0179, 0.01, 0.3546, 0.0157, 0.01, 0.3546,
0.0179, 0.01, 0.3568, 0.0157, 0.01, 0.3568, 0.0136, 0.01, 0.3546, 0.0114, 0.01, 0.3546, 0.0136, 0.01, 0.3568, 0.0114, 0.01, 0.3568, 0.0101, 0.0061, 0.352, 0.0071, 0.0061, 0.352, 0.0101, 0.0083, 0.352, 0.0071, 0.0083, 0.352, 0.0049, 0.0061, 0.352, 0.0027, 0.0061, 0.352, 0.0049, 0.0083, 0.352, 0.0027, 0.0083, 0.352, 0.0092, 0.01, 0.3503, 0.0071, 0.01, 0.3503, 0.0101, 0.01, 0.352, 0.0071, 0.01, 0.352, 0.0049, 0.01, 0.3503, 0.0027, 0.01, 0.3503, 0.0049, 0.01, 0.352, 0.0027, 0.01, 0.352, 0.0101, 0.0061, 0.3546, 0.0101, 0.0083, 0.3546, 0.0101, 0.0061, 0.3568, 0.0101, 0.0083, 0.3568, 0.0101, 0.01, 0.3546, 0.0101, 0.01, 0.3568, 0.0179, 0.01, 0.3589, 0.0157, 0.01, 0.3589,
0.0179, 0.01, 0.3611, 0.0157, 0.01, 0.3611, 0.0136, 0.01, 0.3589, 0.0114, 0.01, 0.3589, 0.0136, 0.01, 0.3611, 0.0114, 0.01, 0.3611, 0.0179, 0.01, 0.3633, 0.0157, 0.01, 0.3633, 0.0179, 0.01, 0.3655, 0.0157, 0.01, 0.3655, 0.0136, 0.01, 0.3633, 0.0114, 0.01, 0.3633, 0.0136, 0.01, 0.3655, 0.0114, 0.01, 0.3655, 0.0101, 0.0061, 0.3589, 0.0101, 0.0083, 0.3589, 0.0101, 0.0061, 0.3611, 0.0101, 0.0083, 0.3611, 0.0101, 0.01, 0.3589, 0.0101, 0.01, 0.3611, 0.0101, 0.0061, 0.3633, 0.0101, 0.0083, 0.3633, 0.0101, 0.0061, 0.3655, 0.0101, 0.0083, 0.3655, 0.0101, 0.01, 0.3633, 0.0101, 0.01, 0.3655, 0.0219, 0.0061, 0.3676, 0.0219, 0.0083, 0.3676, 0.0219, 0.0061, 0.3698, 0.0219, 0.0083, 0.3698,
0.0219, 0.01, 0.3676, 0.0201, 0.01, 0.3676, 0.0219, 0.01, 0.3698, 0.0201, 0.01, 0.3698, 0.0219, 0.0061, 0.372, 0.0219, 0.0083, 0.372, 0.0219, 0.0061, 0.3741, 0.0219, 0.0083, 0.3741, 0.0219, 0.01, 0.372, 0.0201, 0.01, 0.372, 0.0219, 0.01, 0.3741, 0.0201, 0.01, 0.3741, 0.0219, 0.0061, 0.3763, 0.0219, 0.0083, 0.3763, 0.0219, 0.0061, 0.3785, 0.0219, 0.0083, 0.3785, 0.0219, 0.01, 0.3763, 0.0201, 0.01, 0.3763, 0.0219, 0.01, 0.3785, 0.0201, 0.01, 0.3785, 0.0218, 0.0061, 0.3807, 0.0218, 0.0083, 0.3807, 0.0215, 0.0061, 0.3828, 0.0215, 0.0083, 0.3828, 0.0218, 0.01, 0.3807, 0.0201, 0.01, 0.3806, 0.0215, 0.01, 0.3828, 0.0201, 0.01, 0.3828, 0.0179, 0.01, 0.3676, 0.0157, 0.01, 0.3676,
0.0179, 0.01, 0.3698, 0.0157, 0.01, 0.3698, 0.0136, 0.01, 0.3676, 0.0114, 0.01, 0.3676, 0.0136, 0.01, 0.3698, 0.0114, 0.01, 0.3698, 0.0179, 0.01, 0.372, 0.0157, 0.01, 0.372, 0.0179, 0.01, 0.3741, 0.0157, 0.01, 0.3741, 0.0136, 0.01, 0.372, 0.0114, 0.01, 0.372, 0.0136, 0.01, 0.3741, 0.0114, 0.01, 0.3741, 0.0101, 0.0061, 0.3676, 0.0101, 0.0083, 0.3676, 0.0101, 0.0061, 0.3698, 0.0101, 0.0083, 0.3698, 0.0101, 0.01, 0.3676, 0.0101, 0.01, 0.3698, 0.0101, 0.0061, 0.372, 0.0101, 0.0083, 0.372, 0.0101, 0.0061, 0.3741, 0.0101, 0.0083, 0.3741, 0.0101, 0.01, 0.372, 0.0101, 0.01, 0.3741, 0.0179, 0.01, 0.3763, 0.0157, 0.01, 0.3763, 0.0179, 0.01, 0.3785, 0.0157, 0.01, 0.3785,
0.0136, 0.01, 0.3763, 0.0114, 0.01, 0.3763, 0.0136, 0.01, 0.3785, 0.0114, 0.01, 0.3785, 0.0179, 0.01, 0.3806, 0.0157, 0.01, 0.3806, 0.0179, 0.01, 0.3828, 0.0157, 0.01, 0.3828, 0.0136, 0.01, 0.3806, 0.0114, 0.01, 0.3806, 0.0136, 0.01, 0.3828, 0.0114, 0.01, 0.3828, 0.0101, 0.0061, 0.3763, 0.0101, 0.0083, 0.3763, 0.0101, 0.0061, 0.3785, 0.0101, 0.0083, 0.3785, 0.0101, 0.01, 0.3763, 0.0101, 0.01, 0.3785, 0.0101, 0.0061, 0.3807, 0.0101, 0.0083, 0.3807, 0.0099, 0.0061, 0.3828, 0.0099, 0.0083, 0.3828, 0.0101, 0.01, 0.3806, 0.0099, 0.01, 0.3828, 0.067, 0.0061, 0.3863, 0.067, 0.0083, 0.3863, 0.0656, 0.0061, 0.3863, 0.0635, 0.0061, 0.3863, 0.0656, 0.0083, 0.3863, 0.0635, 0.0083, 0.3863,
0.067, 0.01, 0.3863, 0.0656, 0.01, 0.3863, 0.0635, 0.01, 0.3863, 0.067, 0.0061, 0.3893, 0.067, 0.0083, 0.3893, 0.067, 0.0061, 0.3915, 0.067, 0.0083, 0.3915, 0.0631, 0.0061, 0.3902, 0.0631, 0.0083, 0.3902, 0.0631, 0.0061, 0.3915, 0.0631, 0.0083, 0.3915, 0.067, 0.01, 0.3893, 0.067, 0.01, 0.3915, 0.0656, 0.01, 0.3893, 0.0631, 0.01, 0.3902, 0.0656, 0.01, 0.3915, 0.0631, 0.01, 0.3915, 0.0613, 0.0061, 0.3863, 0.0591, 0.0061, 0.3863, 0.0613, 0.0083, 0.3863, 0.0591, 0.0083, 0.3863, 0.057, 0.0061, 0.3863, 0.0548, 0.0061, 0.3863, 0.057, 0.0083, 0.3863, 0.0548, 0.0083, 0.3863, 0.0613, 0.01, 0.3863, 0.0591, 0.01, 0.3863, 0.057, 0.01, 0.3863, 0.0548, 0.01, 0.3863, 0.0613, 0.0061, 0.3902,
0.0591, 0.0061, 0.3902, 0.0613, 0.0083, 0.3902, 0.0591, 0.0083, 0.3902, 0.057, 0.0061, 0.3902, 0.0548, 0.0061, 0.3902, 0.057, 0.0083, 0.3902, 0.0548, 0.0083, 0.3902, 0.0613, 0.01, 0.3902, 0.0591, 0.01, 0.3902, 0.057, 0.01, 0.3902, 0.0548, 0.01, 0.3902, 0.067, 0.0061, 0.3937, 0.067, 0.0083, 0.3937, 0.067, 0.0061, 0.3958, 0.067, 0.0083, 0.3958, 0.0631, 0.0061, 0.3937, 0.0631, 0.0083, 0.3937, 0.0631, 0.0061, 0.3958, 0.0631, 0.0083, 0.3958, 0.067, 0.01, 0.3937, 0.067, 0.01, 0.3958, 0.0656, 0.01, 0.3937, 0.0631, 0.01, 0.3937, 0.0656, 0.01, 0.3958, 0.0631, 0.01, 0.3958, 0.067, 0.0061, 0.398, 0.067, 0.0083, 0.398, 0.0669, 0.0061, 0.4002, 0.0669, 0.0083, 0.4002, 0.063, 0.0061, 0.398,
0.063, 0.0083, 0.398, 0.063, 0.0061, 0.4002, 0.063, 0.0083, 0.4002, 0.067, 0.01, 0.398, 0.0669, 0.01, 0.4002, 0.0656, 0.01, 0.398, 0.063, 0.01, 0.398, 0.0656, 0.01, 0.4002, 0.063, 0.01, 0.4001, 0.0526, 0.0061, 0.3863, 0.0505, 0.0061, 0.3863, 0.0526, 0.0083, 0.3863, 0.0505, 0.0083, 0.3863, 0.0483, 0.0061, 0.3863, 0.0461, 0.0061, 0.3863, 0.0483, 0.0083, 0.3863, 0.0461, 0.0083, 0.3863, 0.0526, 0.01, 0.3863, 0.0505, 0.01, 0.3863, 0.0483, 0.01, 0.3863, 0.0461, 0.01, 0.3863, 0.0526, 0.0061, 0.3902, 0.0505, 0.0061, 0.3902, 0.0526, 0.0083, 0.3902, 0.0505, 0.0083, 0.3902, 0.0483, 0.0061, 0.3902, 0.0454, 0.0061, 0.3902, 0.0483, 0.0083, 0.3902, 0.0454, 0.0083, 0.3902, 0.0454, 0.0061, 0.3915,
0.0454, 0.0083, 0.3915, 0.0526, 0.01, 0.3902, 0.0505, 0.01, 0.3902, 0.0483, 0.01, 0.3902, 0.0454, 0.01, 0.3902, 0.0454, 0.01, 0.3915, 0.0439, 0.0061, 0.3863, 0.0425, 0.0061, 0.3863, 0.0439, 0.0083, 0.3863, 0.0425, 0.0083, 0.3863, 0.0439, 0.01, 0.3863, 0.0425, 0.01, 0.3863, 0.0425, 0.0061, 0.3893, 0.0425, 0.0083, 0.3893, 0.0425, 0.0061, 0.3915, 0.0425, 0.0083, 0.3915, 0.0439, 0.01, 0.3893, 0.0425, 0.01, 0.3893, 0.0439, 0.01, 0.3915, 0.0425, 0.01, 0.3915, 0.0454, 0.0061, 0.3937, 0.0454, 0.0083, 0.3937, 0.0454, 0.0061, 0.3958, 0.0454, 0.0083, 0.3958, 0.0454, 0.01, 0.3937, 0.0454, 0.01, 0.3958, 0.0455, 0.0061, 0.398, 0.0455, 0.0083, 0.398, 0.0458, 0.0061, 0.4002, 0.0458, 0.0083, 0.4002,
0.0455, 0.01, 0.398, 0.0457, 0.01, 0.4001, 0.0425, 0.0061, 0.3937, 0.0425, 0.0083, 0.3937, 0.0425, 0.0061, 0.3958, 0.0425, 0.0083, 0.3958, 0.0439, 0.01, 0.3937, 0.0425, 0.01, 0.3937, 0.0439, 0.01, 0.3958, 0.0425, 0.01, 0.3958, 0.0425, 0.0061, 0.398, 0.0425, 0.0083, 0.398, 0.0425, 0.0061, 0.4002, 0.0425, 0.0083, 0.4002, 0.0439, 0.01, 0.398, 0.0425, 0.01, 0.398, 0.0439, 0.01, 0.4002, 0.0425, 0.01, 0.4002, 0.0668, 0.0061, 0.4017, 0.0668, 0.0083, 0.4017, 0.0666, 0.0061, 0.4028, 0.0627, 0.0061, 0.4023, 0.0666, 0.0083, 0.4028, 0.0627, 0.0083, 0.4023, 0.0662, 0.0061, 0.4046, 0.0662, 0.0083, 0.4046, 0.0668, 0.01, 0.4016, 0.0667, 0.01, 0.4025, 0.0627, 0.01, 0.4023, 0.0661, 0.01, 0.4048,
0.0635, 0.01, 0.4045, 0.0652, 0.0061, 0.4066, 0.0644, 0.0061, 0.4076, 0.0652, 0.0083, 0.4066, 0.0644, 0.0083, 0.4076, 0.0635, 0.0061, 0.409, 0.0635, 0.0083, 0.409, 0.0653, 0.01, 0.4064, 0.0645, 0.01, 0.4075, 0.0635, 0.01, 0.4089, 0.0623, 0.0061, 0.4033, 0.0623, 0.0083, 0.4033, 0.0619, 0.0061, 0.4046, 0.0619, 0.0083, 0.4046, 0.0623, 0.01, 0.4033, 0.0619, 0.01, 0.4045, 0.0608, 0.0061, 0.4062, 0.0595, 0.0061, 0.4074, 0.0608, 0.0083, 0.4062, 0.0595, 0.0083, 0.4074, 0.0624, 0.0061, 0.4099, 0.0584, 0.0061, 0.408, 0.0624, 0.0083, 0.4099, 0.0584, 0.0083, 0.408, 0.057, 0.0061, 0.4085, 0.0548, 0.0061, 0.4087, 0.057, 0.0083, 0.4085, 0.0548, 0.0083, 0.4087, 0.0608, 0.01, 0.4062, 0.0596, 0.01, 0.4073,
0.0623, 0.01, 0.41, 0.0587, 0.01, 0.4079, 0.0569, 0.01, 0.4085, 0.0548, 0.01, 0.4087, 0.0613, 0.0061, 0.4107, 0.0594, 0.0061, 0.4119, 0.0613, 0.0083, 0.4107, 0.0594, 0.0083, 0.4119, 0.0582, 0.0061, 0.4122, 0.0582, 0.0083, 0.4122, 0.057, 0.0061, 0.4125, 0.0548, 0.0061, 0.4127, 0.057, 0.0083, 0.4125, 0.0548, 0.0083, 0.4127, 0.0612, 0.01, 0.4108, 0.0593, 0.01, 0.4119, 0.0582, 0.01, 0.4122, 0.057, 0.01, 0.411, 0.0548, 0.01, 0.411, 0.057, 0.01, 0.4125, 0.0548, 0.01, 0.4127, 0.0463, 0.0061, 0.4024, 0.0463, 0.0083, 0.4024, 0.0476, 0.0061, 0.405, 0.047, 0.0061, 0.4039, 0.0476, 0.0083, 0.405, 0.047, 0.0083, 0.4039, 0.0462, 0.01, 0.4022, 0.0476, 0.01, 0.4051, 0.0471, 0.01, 0.4041,
0.0498, 0.0061, 0.4075, 0.0498, 0.0083, 0.4075, 0.0526, 0.0061, 0.4086, 0.0509, 0.0061, 0.4081, 0.0526, 0.0083, 0.4086, 0.0509, 0.0083, 0.4081, 0.0486, 0.0061, 0.4064, 0.0486, 0.0083, 0.4064, 0.0456, 0.0061, 0.4093, 0.0456, 0.0083, 0.4093, 0.0498, 0.01, 0.4074, 0.0528, 0.01, 0.4086, 0.0508, 0.01, 0.408, 0.0486, 0.01, 0.4064, 0.0461, 0.01, 0.4067, 0.0483, 0.01, 0.4089, 0.0456, 0.01, 0.4093, 0.0426, 0.0061, 0.4024, 0.0426, 0.0083, 0.4024, 0.0431, 0.0061, 0.4048, 0.0428, 0.0061, 0.4037, 0.0431, 0.0083, 0.4048, 0.0428, 0.0083, 0.4037, 0.0439, 0.01, 0.4024, 0.0426, 0.01, 0.4024, 0.043, 0.01, 0.4047, 0.0428, 0.01, 0.4036, 0.0439, 0.0061, 0.4067, 0.0439, 0.0083, 0.4067, 0.0448, 0.0061, 0.4082,
0.0448, 0.0083, 0.4082, 0.044, 0.01, 0.407, 0.0447, 0.01, 0.4081, 0.0499, 0.0061, 0.412, 0.0499, 0.0083, 0.412, 0.0526, 0.0061, 0.4126, 0.051, 0.0061, 0.4122, 0.0526, 0.0083, 0.4126, 0.051, 0.0083, 0.4122, 0.0482, 0.0061, 0.4112, 0.0468, 0.0061, 0.4103, 0.0482, 0.0083, 0.4112, 0.0468, 0.0083, 0.4103, 0.0526, 0.01, 0.411, 0.0502, 0.01, 0.4121, 0.0526, 0.01, 0.4126, 0.0511, 0.01, 0.4123, 0.048, 0.01, 0.4111, 0.0468, 0.01, 0.4103, 0.0213, 0.0061, 0.3846, 0.0211, 0.0061, 0.3857, 0.0213, 0.0083, 0.3846, 0.0211, 0.0083, 0.3857, 0.0209, 0.0061, 0.3872, 0.0209, 0.0083, 0.3872, 0.0213, 0.01, 0.3844, 0.0212, 0.01, 0.3852, 0.0209, 0.01, 0.3873, 0.0206, 0.0061, 0.3893, 0.0206, 0.0083, 0.3893,
0.0201, 0.0061, 0.3915, 0.0201, 0.0083, 0.3915, 0.0206, 0.01, 0.3895, 0.02, 0.01, 0.3917, 0.0194, 0.0061, 0.3935, 0.0194, 0.0083, 0.3935, 0.0194, 0.01, 0.3933, 0.0179, 0.01, 0.385, 0.0157, 0.01, 0.385, 0.0179, 0.01, 0.3872, 0.0157, 0.01, 0.3872, 0.0136, 0.01, 0.385, 0.0114, 0.01, 0.385, 0.0136, 0.01, 0.3872, 0.0114, 0.01, 0.3872, 0.0179, 0.01, 0.3893, 0.0157, 0.01, 0.3893, 0.0179, 0.01, 0.3915, 0.0157, 0.01, 0.3915, 0.0136, 0.01, 0.3893, 0.0114, 0.01, 0.3893, 0.0136, 0.01, 0.3915, 0.0114, 0.01, 0.3915, 0.0096, 0.0061, 0.385, 0.0096, 0.0083, 0.385, 0.009, 0.0061, 0.3872, 0.009, 0.0083, 0.3872, 0.0096, 0.01, 0.3849, 0.0091, 0.01, 0.387, 0.0084, 0.0061, 0.3891,
0.0081, 0.0061, 0.3902, 0.0084, 0.0083, 0.3891, 0.0081, 0.0083, 0.3902, 0.0075, 0.0061, 0.3915, 0.0075, 0.0083, 0.3915, 0.0084, 0.01, 0.3892, 0.008, 0.01, 0.3902, 0.0092, 0.01, 0.3915, 0.0076, 0.01, 0.3914, 0.0189, 0.0061, 0.3946, 0.0189, 0.0083, 0.3946, 0.0185, 0.0061, 0.3959, 0.0185, 0.0083, 0.3959, 0.0191, 0.01, 0.3942, 0.0157, 0.01, 0.3937, 0.0183, 0.01, 0.3961, 0.0157, 0.01, 0.3958, 0.0136, 0.01, 0.3937, 0.0114, 0.01, 0.3937, 0.0136, 0.01, 0.3958, 0.0114, 0.01, 0.3958, 0.0175, 0.0061, 0.398, 0.0175, 0.0083, 0.398, 0.0169, 0.0061, 0.3992, 0.0162, 0.0061, 0.4003, 0.0169, 0.0083, 0.3992, 0.0162, 0.0083, 0.4003, 0.0174, 0.01, 0.3982, 0.0157, 0.01, 0.398, 0.0169, 0.01, 0.3991,
0.0162, 0.01, 0.4003, 0.0136, 0.01, 0.398, 0.0114, 0.01, 0.398, 0.0136, 0.01, 0.4002, 0.0114, 0.01, 0.4002, 0.0065, 0.0061, 0.3935, 0.0065, 0.0083, 0.3935, 0.0058, 0.0061, 0.3945, 0.0058, 0.0083, 0.3945, 0.0049, 0.0061, 0.3959, 0.0049, 0.0083, 0.3959, 0.0092, 0.01, 0.3937, 0.0065, 0.01, 0.3935, 0.0092, 0.01, 0.3958, 0.0071, 0.01, 0.3958, 0.0058, 0.01, 0.3945, 0.0051, 0.01, 0.3956, 0.0039, 0.0061, 0.397, 0.0028, 0.0061, 0.3982, 0.0039, 0.0083, 0.397, 0.0028, 0.0083, 0.3982, 0.0016, 0.0061, 0.3991, 0.0016, 0.0083, 0.3991, 0.0092, 0.01, 0.398, 0.0071, 0.01, 0.398, 0.0092, 0.01, 0.4002, 0.0071, 0.01, 0.4002, 0.0038, 0.01, 0.3971, 0.0028, 0.01, 0.3981, 0.0049, 0.01, 0.4002,
0.0017, 0.01, 0.399, 0.0151, 0.0061, 0.4018, 0.0151, 0.0083, 0.4018, 0.0142, 0.0061, 0.4029, 0.0142, 0.0083, 0.4029, 0.0132, 0.0061, 0.4042, 0.0122, 0.0061, 0.4053, 0.0132, 0.0083, 0.4042, 0.0122, 0.0083, 0.4053, 0.0151, 0.01, 0.4018, 0.0143, 0.01, 0.4029, 0.0114, 0.01, 0.4024, 0.0132, 0.01, 0.4042, 0.0122, 0.01, 0.4053, 0.0111, 0.0061, 0.4064, 0.0111, 0.0083, 0.4064, 0.0111, 0.01, 0.4064, 0.0092, 0.01, 0.4024, 0.0071, 0.01, 0.4024, 0.0092, 0.01, 0.4045, 0.0071, 0.01, 0.4045, 0.0049, 0.01, 0.4024, 0.0027, 0.01, 0.4024, 0.0049, 0.01, 0.4045, 0.0027, 0.01, 0.4045, 0.0099, 0.0061, 0.4075, 0.0099, 0.0083, 0.4075, 0.0089, 0.0061, 0.4084, 0.0074, 0.0061, 0.4094, 0.0089, 0.0083, 0.4084,
0.0074, 0.0083, 0.4094, 0.0099, 0.01, 0.4075, 0.0071, 0.01, 0.4067, 0.0088, 0.01, 0.4084, 0.0074, 0.01, 0.4095, 0.0049, 0.01, 0.4067, 0.0027, 0.01, 0.4067, 0.0049, 0.01, 0.4089, 0.0027, 0.01, 0.4089, 0.0063, 0.0061, 0.4102, 0.0063, 0.0083, 0.4102, 0.0049, 0.0061, 0.4111, 0.0034, 0.0061, 0.4119, 0.0049, 0.0083, 0.4111, 0.0034, 0.0083, 0.4119, 0.0023, 0.0061, 0.4124, 0.0023, 0.0083, 0.4124, 0.0063, 0.01, 0.4102, 0.0051, 0.01, 0.4109, 0.0032, 0.01, 0.412, 0.0021, 0.01, 0.4125, 0.067, 0.0061, 0.4324, 0.067, 0.0083, 0.4324, 0.067, 0.0061, 0.4349, 0.067, 0.0083, 0.4349, 0.0656, 0.0061, 0.4324, 0.0635, 0.0061, 0.4324, 0.0656, 0.0083, 0.4324, 0.0635, 0.0083, 0.4324, 0.0631, 0.0061, 0.4353,
0.0631, 0.0083, 0.4353, 0.067, 0.01, 0.4324, 0.067, 0.01, 0.4349, 0.0656, 0.01, 0.4324, 0.0635, 0.01, 0.4324, 0.0656, 0.01, 0.4349, 0.0631, 0.01, 0.4353, 0.0613, 0.0061, 0.4324, 0.0591, 0.0061, 0.4324, 0.0613, 0.0083, 0.4324, 0.0591, 0.0083, 0.4324, 0.0613, 0.0061, 0.4353, 0.0591, 0.0061, 0.4353, 0.0613, 0.0083, 0.4353, 0.0591, 0.0083, 0.4353, 0.057, 0.0061, 0.4324, 0.0548, 0.0061, 0.4324, 0.057, 0.0083, 0.4324, 0.0548, 0.0083, 0.4324, 0.057, 0.0061, 0.4353, 0.0548, 0.0061, 0.4353, 0.057, 0.0083, 0.4353, 0.0548, 0.0083, 0.4353, 0.0613, 0.01, 0.4324, 0.0591, 0.01, 0.4324, 0.0613, 0.01, 0.4353, 0.0591, 0.01, 0.4353, 0.057, 0.01, 0.4324, 0.0548, 0.01, 0.4324, 0.057, 0.01, 0.4353,
0.0548, 0.01, 0.4353, 0.0526, 0.0061, 0.4324, 0.0505, 0.0061, 0.4324, 0.0526, 0.0083, 0.4324, 0.0505, 0.0083, 0.4324, 0.0526, 0.0061, 0.4353, 0.0497, 0.0061, 0.4355, 0.0526, 0.0083, 0.4353, 0.0497, 0.0083, 0.4355, 0.0483, 0.0061, 0.4324, 0.0461, 0.0061, 0.4324, 0.0483, 0.0083, 0.4324, 0.0461, 0.0083, 0.4324, 0.0489, 0.0061, 0.4358, 0.0489, 0.0083, 0.4358, 0.0526, 0.01, 0.4324, 0.0505, 0.01, 0.4324, 0.0527, 0.01, 0.4353, 0.0506, 0.01, 0.4353, 0.0483, 0.01, 0.4324, 0.0461, 0.01, 0.4324, 0.0487, 0.01, 0.4359, 0.0461, 0.01, 0.4349, 0.0439, 0.0061, 0.4324, 0.0425, 0.0061, 0.4324, 0.0439, 0.0083, 0.4324, 0.0425, 0.0083, 0.4324, 0.0425, 0.0061, 0.4349, 0.0425, 0.0083, 0.4349, 0.0439, 0.01, 0.4324,
0.0425, 0.01, 0.4324, 0.0439, 0.01, 0.4349, 0.0425, 0.01, 0.4349, 0.067, 0.0061, 0.4371, 0.067, 0.0083, 0.4371, 0.067, 0.0061, 0.4392, 0.067, 0.0083, 0.4392, 0.0631, 0.0061, 0.4371, 0.0631, 0.0083, 0.4371, 0.0631, 0.0061, 0.4392, 0.0631, 0.0083, 0.4392, 0.067, 0.01, 0.4371, 0.067, 0.01, 0.4392, 0.0656, 0.01, 0.4371, 0.0631, 0.01, 0.4371, 0.0656, 0.01, 0.4392, 0.0631, 0.01, 0.4392, 0.067, 0.0061, 0.4414, 0.067, 0.0083, 0.4414, 0.067, 0.0061, 0.4436, 0.067, 0.0083, 0.4436, 0.0631, 0.0061, 0.4414, 0.0631, 0.0083, 0.4414, 0.0631, 0.0061, 0.4436, 0.0631, 0.0083, 0.4436, 0.067, 0.01, 0.4414, 0.067, 0.01, 0.4436, 0.0656, 0.01, 0.4414, 0.0631, 0.01, 0.4414, 0.0656, 0.01, 0.4436,
0.0631, 0.01, 0.4436, 0.0562, 0.0061, 0.4382, 0.0548, 0.0061, 0.4382, 0.0562, 0.0083, 0.4382, 0.0548, 0.0083, 0.4382, 0.0562, 0.01, 0.4382, 0.0548, 0.01, 0.4382, 0.0562, 0.0061, 0.4414, 0.0562, 0.0083, 0.4414, 0.0562, 0.0061, 0.4436, 0.0562, 0.0083, 0.4436, 0.0562, 0.01, 0.4414, 0.0548, 0.01, 0.4414, 0.0562, 0.01, 0.4436, 0.0548, 0.01, 0.4436, 0.067, 0.0061, 0.4458, 0.067, 0.0083, 0.4458, 0.067, 0.0061, 0.4479, 0.067, 0.0083, 0.4479, 0.0631, 0.0061, 0.4458, 0.0631, 0.0083, 0.4458, 0.0631, 0.0061, 0.4479, 0.0631, 0.0083, 0.4479, 0.067, 0.01, 0.4458, 0.067, 0.01, 0.4479, 0.0656, 0.01, 0.4458, 0.0631, 0.01, 0.4458, 0.0656, 0.01, 0.4479, 0.0631, 0.01, 0.4479, 0.067, 0.0061, 0.4501,
0.067, 0.0083, 0.4501, 0.067, 0.0061, 0.4529, 0.067, 0.0083, 0.4529, 0.0631, 0.0061, 0.4501, 0.0631, 0.0083, 0.4501, 0.0656, 0.0061, 0.4529, 0.0631, 0.0061, 0.4529, 0.0656, 0.0083, 0.4529, 0.0631, 0.0083, 0.4529, 0.067, 0.01, 0.4501, 0.067, 0.01, 0.4529, 0.0656, 0.01, 0.4501, 0.0631, 0.01, 0.4501, 0.0656, 0.01, 0.4529, 0.0631, 0.01, 0.4529, 0.0562, 0.0061, 0.4458, 0.0562, 0.0083, 0.4458, 0.0562, 0.0061, 0.4479, 0.0562, 0.0083, 0.4479, 0.0562, 0.01, 0.4458, 0.0548, 0.01, 0.4458, 0.0562, 0.01, 0.4479, 0.0548, 0.01, 0.4479, 0.0562, 0.0061, 0.4501, 0.0562, 0.0083, 0.4501, 0.0562, 0.0061, 0.452, 0.0548, 0.0061, 0.452, 0.0562, 0.0083, 0.452, 0.0548, 0.0083, 0.452, 0.0562, 0.01, 0.4501,
0.0548, 0.01, 0.4501, 0.0562, 0.01, 0.452, 0.0548, 0.01, 0.452, 0.0523, 0.0061, 0.4382, 0.0523, 0.0083, 0.4382, 0.0478, 0.0061, 0.4362, 0.0466, 0.0061, 0.4372, 0.0478, 0.0083, 0.4362, 0.0466, 0.0083, 0.4372, 0.0456, 0.0061, 0.4396, 0.0456, 0.0083, 0.4396, 0.0523, 0.01, 0.4382, 0.0477, 0.01, 0.4363, 0.0466, 0.01, 0.4372, 0.0457, 0.01, 0.4393, 0.0523, 0.0061, 0.4414, 0.0523, 0.0083, 0.4414, 0.0523, 0.0061, 0.4436, 0.0523, 0.0083, 0.4436, 0.0455, 0.0061, 0.4414, 0.0455, 0.0083, 0.4414, 0.0454, 0.0061, 0.4436, 0.0454, 0.0083, 0.4436, 0.0523, 0.01, 0.4414, 0.0523, 0.01, 0.4436, 0.0455, 0.01, 0.4414, 0.0454, 0.01, 0.4436, 0.0425, 0.0061, 0.4371, 0.0425, 0.0083, 0.4371, 0.0425, 0.0061, 0.4392,
0.0425, 0.0083, 0.4392, 0.0439, 0.01, 0.4371, 0.0425, 0.01, 0.4371, 0.0439, 0.01, 0.4392, 0.0425, 0.01, 0.4392, 0.0425, 0.0061, 0.4414, 0.0425, 0.0083, 0.4414, 0.0425, 0.0061, 0.4436, 0.0425, 0.0083, 0.4436, 0.0439, 0.01, 0.4414, 0.0425, 0.01, 0.4414, 0.0439, 0.01, 0.4436, 0.0425, 0.01, 0.4436, 0.0523, 0.0061, 0.4458, 0.0523, 0.0083, 0.4458, 0.0523, 0.0061, 0.4479, 0.0523, 0.0083, 0.4479, 0.0454, 0.0061, 0.4458, 0.0454, 0.0083, 0.4458, 0.0454, 0.0061, 0.4479, 0.0454, 0.0083, 0.4479, 0.0523, 0.01, 0.4458, 0.0523, 0.01, 0.4479, 0.0454, 0.01, 0.4458, 0.0454, 0.01, 0.4479, 0.0523, 0.0061, 0.4501, 0.0523, 0.0083, 0.4501, 0.0523, 0.0061, 0.452, 0.0523, 0.0083, 0.452, 0.0454, 0.0061, 0.4501,
0.0454, 0.0083, 0.4501, 0.0454, 0.0061, 0.4529, 0.0454, 0.0083, 0.4529, 0.0523, 0.01, 0.4501, 0.0523, 0.01, 0.452, 0.0454, 0.01, 0.4501, 0.0454, 0.01, 0.4529, 0.0425, 0.0061, 0.4458, 0.0425, 0.0083, 0.4458, 0.0425, 0.0061, 0.4479, 0.0425, 0.0083, 0.4479, 0.0439, 0.01, 0.4458, 0.0425, 0.01, 0.4458, 0.0439, 0.01, 0.4479, 0.0425, 0.01, 0.4479, 0.0425, 0.0061, 0.4501, 0.0425, 0.0083, 0.4501, 0.0439, 0.0061, 0.4529, 0.0425, 0.0061, 0.4529, 0.0439, 0.0083, 0.4529, 0.0425, 0.0083, 0.4529, 0.0439, 0.01, 0.4501, 0.0425, 0.01, 0.4501, 0.0439, 0.01, 0.4529, 0.0425, 0.01, 0.4529, 0.0219, 0.0061, 0.4245, 0.0201, 0.0061, 0.4245, 0.0219, 0.0083, 0.4245, 0.0201, 0.0083, 0.4245, 0.0219, 0.0061, 0.4262,
0.0219, 0.0083, 0.4262, 0.0219, 0.01, 0.4245, 0.0201, 0.01, 0.4245, 0.0219, 0.01, 0.4262, 0.0201, 0.01, 0.4262, 0.0219, 0.0061, 0.4284, 0.0219, 0.0083, 0.4284, 0.0219, 0.0061, 0.4306, 0.0219, 0.0083, 0.4306, 0.0219, 0.01, 0.4284, 0.0201, 0.01, 0.4284, 0.0219, 0.01, 0.4306, 0.0201, 0.01, 0.4306, 0.0219, 0.0061, 0.4327, 0.0219, 0.0083, 0.4327, 0.0219, 0.0061, 0.4349, 0.0219, 0.0083, 0.4349, 0.0219, 0.01, 0.4327, 0.0201, 0.01, 0.4327, 0.0219, 0.01, 0.4349, 0.0201, 0.01, 0.4349, 0.0179, 0.0061, 0.4245, 0.0157, 0.0061, 0.4245, 0.0179, 0.0083, 0.4245, 0.0157, 0.0083, 0.4245, 0.0136, 0.0061, 0.4245, 0.0114, 0.0061, 0.4245, 0.0136, 0.0083, 0.4245, 0.0114, 0.0083, 0.4245, 0.0179, 0.01, 0.4245,
0.0157, 0.01, 0.4245, 0.0179, 0.01, 0.4262, 0.0157, 0.01, 0.4262, 0.0136, 0.01, 0.4245, 0.0114, 0.01, 0.4245, 0.0136, 0.01, 0.4262, 0.0114, 0.01, 0.4262, 0.0101, 0.0061, 0.4245, 0.0101, 0.0083, 0.4245, 0.0101, 0.0061, 0.4262, 0.0101, 0.0083, 0.4262, 0.0101, 0.01, 0.4245, 0.0101, 0.01, 0.4262, 0.0179, 0.01, 0.4284, 0.0157, 0.01, 0.4284, 0.0179, 0.01, 0.4306, 0.0157, 0.01, 0.4306, 0.0136, 0.01, 0.4284, 0.0114, 0.01, 0.4284, 0.0136, 0.01, 0.4306, 0.0114, 0.01, 0.4306, 0.0179, 0.01, 0.4327, 0.0157, 0.01, 0.4327, 0.0179, 0.01, 0.4349, 0.0157, 0.01, 0.4349, 0.0136, 0.01, 0.4327, 0.0114, 0.01, 0.4327, 0.0136, 0.01, 0.4349, 0.0114, 0.01, 0.4349, 0.0101, 0.0061, 0.4284,
0.0101, 0.0083, 0.4284, 0.0101, 0.0061, 0.4306, 0.0101, 0.0083, 0.4306, 0.0101, 0.01, 0.4284, 0.0101, 0.01, 0.4306, 0.0101, 0.0061, 0.4327, 0.0101, 0.0083, 0.4327, 0.0101, 0.0061, 0.4349, 0.0101, 0.0083, 0.4349, 0.0101, 0.01, 0.4327, 0.0101, 0.01, 0.4349, 0.0219, 0.0061, 0.4371, 0.0219, 0.0083, 0.4371, 0.0219, 0.0061, 0.4392, 0.0219, 0.0083, 0.4392, 0.0219, 0.01, 0.4371, 0.0201, 0.01, 0.4371, 0.0219, 0.01, 0.4392, 0.0201, 0.01, 0.4392, 0.0219, 0.0061, 0.4414, 0.0219, 0.0083, 0.4414, 0.0219, 0.0061, 0.4436, 0.0219, 0.0083, 0.4436, 0.0219, 0.01, 0.4414, 0.0201, 0.01, 0.4414, 0.0219, 0.01, 0.4436, 0.0201, 0.01, 0.4436, 0.0219, 0.0061, 0.4458, 0.0219, 0.0083, 0.4458, 0.0219, 0.0061, 0.4479,
0.0219, 0.0083, 0.4479, 0.0219, 0.01, 0.4458, 0.0201, 0.01, 0.4458, 0.0219, 0.01, 0.4479, 0.0201, 0.01, 0.4479, 0.0219, 0.0061, 0.4501, 0.0219, 0.0083, 0.4501, 0.0219, 0.0061, 0.4523, 0.0219, 0.0083, 0.4523, 0.0219, 0.01, 0.4501, 0.0201, 0.01, 0.4501, 0.0219, 0.01, 0.4523, 0.0201, 0.01, 0.4523, 0.0179, 0.01, 0.4371, 0.0157, 0.01, 0.4371, 0.0179, 0.01, 0.4392, 0.0157, 0.01, 0.4392, 0.0136, 0.01, 0.4371, 0.0114, 0.01, 0.4371, 0.0136, 0.01, 0.4392, 0.0114, 0.01, 0.4392, 0.0179, 0.01, 0.4414, 0.0157, 0.01, 0.4414, 0.0179, 0.01, 0.4436, 0.0157, 0.01, 0.4436, 0.0136, 0.01, 0.4414, 0.0114, 0.01, 0.4414, 0.0136, 0.01, 0.4436, 0.0114, 0.01, 0.4436, 0.0101, 0.0061, 0.4371,
0.0101, 0.0083, 0.4371, 0.0101, 0.0061, 0.4392, 0.0101, 0.0083, 0.4392, 0.0101, 0.01, 0.4371, 0.0101, 0.01, 0.4392, 0.0101, 0.0061, 0.4414, 0.0101, 0.0083, 0.4414, 0.0101, 0.0061, 0.4436, 0.0101, 0.0083, 0.4436, 0.0101, 0.01, 0.4414, 0.0101, 0.01, 0.4436, 0.0179, 0.01, 0.4458, 0.0157, 0.01, 0.4458, 0.0179, 0.01, 0.4479, 0.0157, 0.01, 0.4479, 0.0136, 0.01, 0.4458, 0.0114, 0.01, 0.4458, 0.0136, 0.01, 0.4479, 0.0114, 0.01, 0.4479, 0.0179, 0.01, 0.4501, 0.0157, 0.01, 0.4501, 0.0179, 0.01, 0.4523, 0.0157, 0.01, 0.4523, 0.0136, 0.01, 0.4501, 0.0114, 0.01, 0.4501, 0.0136, 0.01, 0.4523, 0.0114, 0.01, 0.4523, 0.0101, 0.0061, 0.4458, 0.0101, 0.0083, 0.4458, 0.0101, 0.0061, 0.4479,
0.0101, 0.0083, 0.4479, 0.0101, 0.01, 0.4458, 0.0101, 0.01, 0.4479, 0.0101, 0.0061, 0.4501, 0.0101, 0.0083, 0.4501, 0.0101, 0.0061, 0.4523, 0.0101, 0.0083, 0.4523, 0.0101, 0.01, 0.4501, 0.0101, 0.01, 0.4523, 0.067, 0.0061, 0.4725, 0.067, 0.0083, 0.4725, 0.067, 0.0061, 0.474, 0.067, 0.0083, 0.474, 0.0656, 0.0061, 0.4725, 0.0635, 0.0061, 0.4725, 0.0656, 0.0083, 0.4725, 0.0635, 0.0083, 0.4725, 0.067, 0.01, 0.4725, 0.067, 0.01, 0.474, 0.0656, 0.01, 0.4725, 0.0635, 0.01, 0.4725, 0.0656, 0.01, 0.474, 0.0635, 0.01, 0.474, 0.067, 0.0061, 0.4765, 0.067, 0.0083, 0.4765, 0.0656, 0.0061, 0.4765, 0.0635, 0.0061, 0.4765, 0.0656, 0.0083, 0.4765, 0.0635, 0.0083, 0.4765, 0.067, 0.01, 0.4765,
0.0656, 0.01, 0.4765, 0.0635, 0.01, 0.4765, 0.0613, 0.0061, 0.4725, 0.0591, 0.0061, 0.4725, 0.0613, 0.0083, 0.4725, 0.0591, 0.0083, 0.4725, 0.057, 0.0061, 0.4725, 0.0548, 0.0061, 0.4725, 0.057, 0.0083, 0.4725, 0.0548, 0.0083, 0.4725, 0.0613, 0.01, 0.4725, 0.0591, 0.01, 0.4725, 0.0613, 0.01, 0.474, 0.0591, 0.01, 0.474, 0.057, 0.01, 0.4725, 0.0548, 0.01, 0.4725, 0.057, 0.01, 0.474, 0.0548, 0.01, 0.474, 0.0613, 0.0061, 0.4765, 0.0591, 0.0061, 0.4765, 0.0613, 0.0083, 0.4765, 0.0591, 0.0083, 0.4765, 0.057, 0.0061, 0.4765, 0.0548, 0.0061, 0.4765, 0.057, 0.0083, 0.4765, 0.0548, 0.0083, 0.4765, 0.0613, 0.01, 0.4765, 0.0591, 0.01, 0.4765, 0.057, 0.01, 0.4765, 0.0548, 0.01, 0.4765,
0.0591, 0.0061, 0.4804, 0.0591, 0.0083, 0.4804, 0.0591, 0.0061, 0.4826, 0.0591, 0.0083, 0.4826, 0.0562, 0.0061, 0.4804, 0.0562, 0.0083, 0.4804, 0.0562, 0.0061, 0.4826, 0.0562, 0.0083, 0.4826, 0.0591, 0.01, 0.4804, 0.0591, 0.01, 0.4826, 0.0562, 0.01, 0.4804, 0.0562, 0.01, 0.4826, 0.0592, 0.0061, 0.4848, 0.0592, 0.0083, 0.4848, 0.0596, 0.0061, 0.487, 0.0596, 0.0083, 0.487, 0.0562, 0.0061, 0.4848, 0.0562, 0.0083, 0.4848, 0.0562, 0.0061, 0.487, 0.0562, 0.0083, 0.487, 0.0592, 0.01, 0.4848, 0.0596, 0.01, 0.4868, 0.0562, 0.01, 0.4848, 0.0562, 0.01, 0.487, 0.0526, 0.0061, 0.4725, 0.0505, 0.0061, 0.4725, 0.0526, 0.0083, 0.4725, 0.0505, 0.0083, 0.4725, 0.0483, 0.0061, 0.4725, 0.0461, 0.0061, 0.4725,
0.0483, 0.0083, 0.4725, 0.0461, 0.0083, 0.4725, 0.0526, 0.01, 0.4725, 0.0505, 0.01, 0.4725, 0.0526, 0.01, 0.474, 0.0505, 0.01, 0.474, 0.0483, 0.01, 0.4725, 0.0461, 0.01, 0.4725, 0.0483, 0.01, 0.474, 0.0461, 0.01, 0.474, 0.0526, 0.0061, 0.4765, 0.0504, 0.0061, 0.4766, 0.0526, 0.0083, 0.4765, 0.0504, 0.0083, 0.4766, 0.0487, 0.0061, 0.4769, 0.0487, 0.0083, 0.4769, 0.0476, 0.0061, 0.4775, 0.047, 0.0061, 0.478, 0.0476, 0.0083, 0.4775, 0.047, 0.0083, 0.478, 0.0527, 0.01, 0.4765, 0.0506, 0.01, 0.4766, 0.0485, 0.01, 0.477, 0.0461, 0.01, 0.4761, 0.0476, 0.01, 0.4775, 0.0465, 0.01, 0.4786, 0.0439, 0.0061, 0.4725, 0.0425, 0.0061, 0.4725, 0.0439, 0.0083, 0.4725, 0.0425, 0.0083, 0.4725,
0.0425, 0.0061, 0.474, 0.0425, 0.0083, 0.474, 0.0439, 0.01, 0.4725, 0.0425, 0.01, 0.4725, 0.0439, 0.01, 0.474, 0.0425, 0.01, 0.474, 0.0425, 0.0061, 0.4761, 0.0425, 0.0083, 0.4761, 0.0425, 0.0061, 0.4783, 0.0425, 0.0083, 0.4783, 0.0439, 0.01, 0.4761, 0.0425, 0.01, 0.4761, 0.0439, 0.01, 0.4783, 0.0425, 0.01, 0.4783, 0.0456, 0.0061, 0.4801, 0.0456, 0.0083, 0.4801, 0.0455, 0.0061, 0.4826, 0.0455, 0.0083, 0.4826, 0.0456, 0.01, 0.4801, 0.0455, 0.01, 0.4827, 0.0454, 0.0061, 0.4848, 0.0454, 0.0083, 0.4848, 0.0454, 0.0061, 0.487, 0.0454, 0.0083, 0.487, 0.0454, 0.01, 0.4848, 0.0454, 0.01, 0.487, 0.0425, 0.0061, 0.4805, 0.0425, 0.0083, 0.4805, 0.0425, 0.0061, 0.4826, 0.0425, 0.0083, 0.4826,
0.0439, 0.01, 0.4805, 0.0425, 0.01, 0.4805, 0.0439, 0.01, 0.4826, 0.0425, 0.01, 0.4826, 0.0425, 0.0061, 0.4848, 0.0425, 0.0083, 0.4848, 0.0425, 0.0061, 0.487, 0.0425, 0.0083, 0.487, 0.0439, 0.01, 0.4848, 0.0425, 0.01, 0.4848, 0.0439, 0.01, 0.487, 0.0425, 0.01, 0.487, 0.0219, 0.0061, 0.4544, 0.0219, 0.0083, 0.4544, 0.0219, 0.0061, 0.4566, 0.0219, 0.0083, 0.4566, 0.0219, 0.01, 0.4544, 0.0201, 0.01, 0.4544, 0.0219, 0.01, 0.4566, 0.0201, 0.01, 0.4566, 0.0219, 0.0061, 0.4588, 0.0219, 0.0083, 0.4588, 0.0219, 0.0061, 0.4609, 0.0219, 0.0083, 0.4609, 0.0219, 0.01, 0.4588, 0.0201, 0.01, 0.4588, 0.0219, 0.01, 0.4609, 0.0201, 0.01, 0.4609, 0.0219, 0.0061, 0.4631, 0.0219, 0.0083, 0.4631,
0.0219, 0.0061, 0.4653, 0.0219, 0.0083, 0.4653, 0.0219, 0.01, 0.4631, 0.0201, 0.01, 0.4631, 0.0219, 0.01, 0.4653, 0.0201, 0.01, 0.4653, 0.0219, 0.0061, 0.4675, 0.0219, 0.0083, 0.4675, 0.0219, 0.0061, 0.4696, 0.0219, 0.0083, 0.4696, 0.0219, 0.01, 0.4675, 0.0201, 0.01, 0.4675, 0.0219, 0.01, 0.4696, 0.0201, 0.01, 0.4696, 0.0179, 0.01, 0.4544, 0.0157, 0.01, 0.4544, 0.0179, 0.01, 0.4566, 0.0157, 0.01, 0.4566, 0.0136, 0.01, 0.4544, 0.0114, 0.01, 0.4544, 0.0136, 0.01, 0.4566, 0.0114, 0.01, 0.4566, 0.0179, 0.01, 0.4588, 0.0157, 0.01, 0.4588, 0.0179, 0.01, 0.4609, 0.0157, 0.01, 0.4609, 0.0136, 0.01, 0.4588, 0.0114, 0.01, 0.4588, 0.0136, 0.01, 0.4609, 0.0114, 0.01, 0.4609,
0.0101, 0.0061, 0.4544, 0.0101, 0.0083, 0.4544, 0.0101, 0.0061, 0.4566, 0.0101, 0.0083, 0.4566, 0.0101, 0.01, 0.4544, 0.0101, 0.01, 0.4566, 0.0101, 0.0061, 0.4588, 0.0101, 0.0083, 0.4588, 0.0101, 0.0061, 0.4609, 0.0101, 0.0083, 0.4609, 0.0101, 0.01, 0.4588, 0.0101, 0.01, 0.4609, 0.0179, 0.01, 0.4631, 0.0157, 0.01, 0.4631, 0.0179, 0.01, 0.4653, 0.0157, 0.01, 0.4653, 0.0136, 0.01, 0.4631, 0.0114, 0.01, 0.4631, 0.0136, 0.01, 0.4653, 0.0114, 0.01, 0.4653, 0.0179, 0.01, 0.4675, 0.0157, 0.01, 0.4675, 0.0179, 0.01, 0.4696, 0.0157, 0.01, 0.4696, 0.0136, 0.01, 0.4675, 0.0114, 0.01, 0.4675, 0.0136, 0.01, 0.4696, 0.0114, 0.01, 0.4696, 0.0101, 0.0061, 0.4631, 0.0101, 0.0083, 0.4631,
0.0101, 0.0061, 0.4653, 0.0101, 0.0083, 0.4653, 0.0101, 0.01, 0.4631, 0.0101, 0.01, 0.4653, 0.0101, 0.0061, 0.4675, 0.0101, 0.0083, 0.4675, 0.0101, 0.0061, 0.4696, 0.0101, 0.0083, 0.4696, 0.0101, 0.01, 0.4675, 0.0101, 0.01, 0.4696, 0.0219, 0.0061, 0.4718, 0.0219, 0.0083, 0.4718, 0.0219, 0.0061, 0.474, 0.0219, 0.0083, 0.474, 0.0219, 0.01, 0.4718, 0.0201, 0.01, 0.4718, 0.0219, 0.01, 0.474, 0.0201, 0.01, 0.474, 0.0219, 0.0061, 0.4761, 0.0219, 0.0083, 0.4761, 0.0218, 0.0061, 0.4783, 0.0218, 0.0083, 0.4783, 0.0219, 0.01, 0.4761, 0.0201, 0.01, 0.4761, 0.0218, 0.01, 0.4783, 0.0201, 0.01, 0.4783, 0.0216, 0.0061, 0.4805, 0.0216, 0.0083, 0.4805, 0.0213, 0.0061, 0.4823, 0.0211, 0.0061, 0.4834,
0.0213, 0.0083, 0.4823, 0.0211, 0.0083, 0.4834, 0.0216, 0.01, 0.4805, 0.0201, 0.01, 0.4805, 0.0213, 0.01, 0.4821, 0.0212, 0.01, 0.4829, 0.0208, 0.0061, 0.4848, 0.0208, 0.0083, 0.4848, 0.0203, 0.0061, 0.487, 0.0203, 0.0083, 0.487, 0.0207, 0.01, 0.485, 0.0202, 0.01, 0.4872, 0.0179, 0.01, 0.4718, 0.0157, 0.01, 0.4718, 0.0179, 0.01, 0.474, 0.0157, 0.01, 0.474, 0.0136, 0.01, 0.4718, 0.0114, 0.01, 0.4718, 0.0136, 0.01, 0.474, 0.0114, 0.01, 0.474, 0.0179, 0.01, 0.4761, 0.0157, 0.01, 0.4761, 0.0179, 0.01, 0.4783, 0.0157, 0.01, 0.4783, 0.0136, 0.01, 0.4761, 0.0114, 0.01, 0.4761, 0.0136, 0.01, 0.4783, 0.0114, 0.01, 0.4783, 0.0101, 0.0061, 0.4718, 0.0101, 0.0083, 0.4718,
0.0101, 0.0061, 0.474, 0.0101, 0.0083, 0.474, 0.0101, 0.01, 0.4718, 0.0101, 0.01, 0.474, 0.0101, 0.0061, 0.4761, 0.0101, 0.0083, 0.4761, 0.01, 0.0061, 0.4783, 0.01, 0.0083, 0.4783, 0.0101, 0.01, 0.4761, 0.01, 0.01, 0.4783, 0.0179, 0.01, 0.4805, 0.0157, 0.01, 0.4805, 0.0179, 0.01, 0.4826, 0.0157, 0.01, 0.4826, 0.0136, 0.01, 0.4805, 0.0114, 0.01, 0.4805, 0.0136, 0.01, 0.4826, 0.0114, 0.01, 0.4826, 0.0179, 0.01, 0.4848, 0.0157, 0.01, 0.4848, 0.0179, 0.01, 0.487, 0.0157, 0.01, 0.487, 0.0136, 0.01, 0.4848, 0.0114, 0.01, 0.4848, 0.0136, 0.01, 0.487, 0.0114, 0.01, 0.487, 0.0096, 0.0061, 0.4805, 0.0096, 0.0083, 0.4805, 0.009, 0.0061, 0.4827, 0.009, 0.0083, 0.4827,
0.0096, 0.01, 0.4804, 0.0091, 0.01, 0.4824, 0.0083, 0.0061, 0.4841, 0.0078, 0.0061, 0.4852, 0.0083, 0.0083, 0.4841, 0.0078, 0.0083, 0.4852, 0.0067, 0.0061, 0.4869, 0.0067, 0.0083, 0.4869, 0.0058, 0.0061, 0.4879, 0.0058, 0.0083, 0.4879, 0.0082, 0.01, 0.4843, 0.0077, 0.01, 0.4853, 0.0092, 0.01, 0.487, 0.0068, 0.01, 0.4869, 0.0058, 0.01, 0.4879, 0.0628, 0.0061, 0.4919, 0.0628, 0.0083, 0.4919, 0.0629, 0.01, 0.4919, 0.0648, 0.0061, 0.4943, 0.0639, 0.0061, 0.4932, 0.0648, 0.0083, 0.4943, 0.0639, 0.0083, 0.4932, 0.0658, 0.0061, 0.4956, 0.0658, 0.0083, 0.4956, 0.0648, 0.01, 0.4943, 0.064, 0.01, 0.4932, 0.0657, 0.01, 0.4954, 0.0635, 0.01, 0.4957, 0.0608, 0.0061, 0.4894, 0.0601, 0.0061, 0.4882,
0.0608, 0.0083, 0.4894, 0.0601, 0.0083, 0.4882, 0.0619, 0.0061, 0.4908, 0.0619, 0.0083, 0.4908, 0.0562, 0.0061, 0.4892, 0.0562, 0.0083, 0.4892, 0.056, 0.0061, 0.4909, 0.0558, 0.0061, 0.492, 0.056, 0.0083, 0.4909, 0.0558, 0.0083, 0.492, 0.0608, 0.01, 0.4895, 0.0603, 0.01, 0.4886, 0.0619, 0.01, 0.4909, 0.0591, 0.01, 0.4913, 0.0562, 0.01, 0.4891, 0.056, 0.01, 0.4912, 0.0558, 0.01, 0.4921, 0.0608, 0.0061, 0.4963, 0.0594, 0.0061, 0.4953, 0.0608, 0.0083, 0.4963, 0.0594, 0.0083, 0.4953, 0.0553, 0.0061, 0.4935, 0.0553, 0.0083, 0.4935, 0.0543, 0.0061, 0.4953, 0.0543, 0.0083, 0.4953, 0.0613, 0.01, 0.4935, 0.0591, 0.01, 0.4935, 0.0609, 0.01, 0.4963, 0.0594, 0.01, 0.4953, 0.057, 0.01, 0.4935,
0.0553, 0.01, 0.4934, 0.057, 0.01, 0.4957, 0.0543, 0.01, 0.4952, 0.0668, 0.0061, 0.4986, 0.0668, 0.0083, 0.4986, 0.0669, 0.0061, 0.5, 0.0669, 0.0083, 0.5, 0.0665, 0.0061, 0.4975, 0.063, 0.0061, 0.4983, 0.0665, 0.0083, 0.4975, 0.063, 0.0083, 0.4983, 0.0652, 0.0061, 0.5005, 0.0641, 0.0061, 0.4994, 0.0652, 0.0083, 0.5005, 0.0641, 0.0083, 0.4994, 0.0668, 0.01, 0.4987, 0.0669, 0.01, 0.5, 0.0666, 0.01, 0.4977, 0.063, 0.01, 0.4983, 0.0652, 0.01, 0.5005, 0.0641, 0.01, 0.4994, 0.067, 0.0061, 0.502, 0.067, 0.0083, 0.502, 0.0663, 0.0061, 0.5014, 0.0663, 0.0083, 0.5014, 0.067, 0.01, 0.502, 0.0664, 0.01, 0.5014, 0.0619, 0.0061, 0.4972, 0.0582, 0.0061, 0.497, 0.0619, 0.0083, 0.4972,
0.0582, 0.0083, 0.497, 0.0574, 0.0061, 0.4981, 0.0574, 0.0083, 0.4981, 0.0563, 0.0061, 0.4994, 0.0549, 0.0061, 0.5006, 0.0563, 0.0083, 0.4994, 0.0549, 0.0083, 0.5006, 0.0619, 0.01, 0.4972, 0.0582, 0.01, 0.497, 0.0574, 0.01, 0.4981, 0.0548, 0.01, 0.4978, 0.0563, 0.01, 0.4994, 0.0549, 0.01, 0.5006, 0.0538, 0.0061, 0.5011, 0.0538, 0.0083, 0.5011, 0.0537, 0.01, 0.5011, 0.0455, 0.0061, 0.4892, 0.0455, 0.0083, 0.4892, 0.046, 0.0061, 0.4914, 0.046, 0.0083, 0.4914, 0.0455, 0.01, 0.4891, 0.0459, 0.01, 0.4912, 0.053, 0.0061, 0.4962, 0.0503, 0.0061, 0.497, 0.053, 0.0083, 0.4962, 0.0503, 0.0083, 0.497, 0.0474, 0.0061, 0.4943, 0.0468, 0.0061, 0.4932, 0.0474, 0.0083, 0.4943, 0.0468, 0.0083, 0.4932,
0.0484, 0.0061, 0.4955, 0.0484, 0.0083, 0.4955, 0.0526, 0.01, 0.4964, 0.0503, 0.01, 0.497, 0.0474, 0.01, 0.4943, 0.0468, 0.01, 0.4933, 0.0484, 0.01, 0.4955, 0.0461, 0.01, 0.4957, 0.0425, 0.0061, 0.4892, 0.0425, 0.0083, 0.4892, 0.0425, 0.0061, 0.4913, 0.0425, 0.0083, 0.4913, 0.0439, 0.01, 0.4892, 0.0425, 0.01, 0.4892, 0.0439, 0.01, 0.4913, 0.0425, 0.01, 0.4913, 0.0426, 0.0061, 0.4935, 0.0426, 0.0083, 0.4935, 0.0431, 0.0061, 0.4958, 0.0428, 0.0061, 0.4947, 0.0431, 0.0083, 0.4958, 0.0428, 0.0083, 0.4947, 0.0439, 0.01, 0.4935, 0.0426, 0.01, 0.4935, 0.0431, 0.01, 0.4958, 0.0428, 0.01, 0.4946, 0.0475, 0.0061, 0.501, 0.046, 0.0061, 0.5, 0.0475, 0.0083, 0.501, 0.046, 0.0083, 0.5,
0.0526, 0.01, 0.4978, 0.0505, 0.01, 0.4978, 0.0526, 0.01, 0.5, 0.0505, 0.01, 0.5, 0.0483, 0.01, 0.4978, 0.0461, 0.01, 0.4978, 0.0477, 0.01, 0.5011, 0.0458, 0.01, 0.4998, 0.0527, 0.0061, 0.5015, 0.0505, 0.0061, 0.5018, 0.0527, 0.0083, 0.5015, 0.0505, 0.0083, 0.5018, 0.0486, 0.0061, 0.5015, 0.0486, 0.0083, 0.5015, 0.0527, 0.01, 0.5015, 0.0505, 0.01, 0.5018, 0.0488, 0.01, 0.5015, 0.0441, 0.0061, 0.4979, 0.0441, 0.0083, 0.4979, 0.045, 0.0061, 0.499, 0.045, 0.0083, 0.499, 0.0442, 0.01, 0.4981, 0.045, 0.01, 0.499, 0.0195, 0.0061, 0.4892, 0.0195, 0.0083, 0.4892, 0.0196, 0.01, 0.4889, 0.019, 0.0061, 0.4902, 0.019, 0.0083, 0.4902, 0.0185, 0.0061, 0.4914, 0.0185, 0.0083, 0.4914,
0.0192, 0.01, 0.4898, 0.0157, 0.01, 0.4892, 0.0183, 0.01, 0.4917, 0.0157, 0.01, 0.4913, 0.0136, 0.01, 0.4892, 0.0114, 0.01, 0.4892, 0.0136, 0.01, 0.4913, 0.0114, 0.01, 0.4913, 0.0174, 0.0061, 0.4932, 0.0166, 0.0061, 0.4942, 0.0174, 0.0083, 0.4932, 0.0166, 0.0083, 0.4942, 0.0155, 0.0061, 0.4956, 0.0155, 0.0083, 0.4956, 0.0145, 0.0061, 0.4966, 0.0145, 0.0083, 0.4966, 0.0174, 0.01, 0.4931, 0.0166, 0.01, 0.4942, 0.0156, 0.01, 0.4955, 0.0136, 0.01, 0.4935, 0.0114, 0.01, 0.4935, 0.0146, 0.01, 0.4966, 0.0114, 0.01, 0.4957, 0.0048, 0.0061, 0.489, 0.0034, 0.0061, 0.49, 0.0048, 0.0083, 0.489, 0.0034, 0.0083, 0.49, 0.0023, 0.0061, 0.4905, 0.0023, 0.0083, 0.4905, 0.0092, 0.01, 0.4892,
0.0071, 0.01, 0.4892, 0.0092, 0.01, 0.4913, 0.0071, 0.01, 0.4913, 0.0048, 0.01, 0.489, 0.0034, 0.01, 0.4899, 0.0049, 0.01, 0.4913, 0.0025, 0.01, 0.4904, 0.0092, 0.01, 0.4935, 0.0071, 0.01, 0.4935, 0.0092, 0.01, 0.4957, 0.0071, 0.01, 0.4957, 0.0049, 0.01, 0.4935, 0.0027, 0.01, 0.4935, 0.0049, 0.01, 0.4957, 0.0027, 0.01, 0.4957, 0.0135, 0.0061, 0.4977, 0.0122, 0.0061, 0.4987, 0.0135, 0.0083, 0.4977, 0.0122, 0.0083, 0.4987, 0.0111, 0.0061, 0.4994, 0.0111, 0.0083, 0.4994, 0.0135, 0.01, 0.4977, 0.0121, 0.01, 0.4988, 0.011, 0.01, 0.4995, 0.0093, 0.0061, 0.5005, 0.008, 0.0061, 0.501, 0.0093, 0.0083, 0.5005, 0.008, 0.0083, 0.501, 0.0092, 0.01, 0.4978, 0.0071, 0.01, 0.4978,
0.0096, 0.01, 0.5004, 0.0076, 0.01, 0.5012, 0.0049, 0.01, 0.4978, 0.0027, 0.01, 0.4978, 0.0049, 0.01, 0.5, 0.0027, 0.01, 0.5, 0.0069, 0.0061, 0.5014, 0.0069, 0.0083, 0.5014, 0.0049, 0.0061, 0.5018, 0.0027, 0.0061, 0.5019, 0.0049, 0.0083, 0.5018, 0.0027, 0.0083, 0.5019, 0.0067, 0.01, 0.5015, 0.0049, 0.01, 0.5018, 0.0027, 0.01, 0.5019, 0.0013, 0.0061, 0.003, 0.0013, 0.0083, 0.003, 0.0012, 0.0061, 0.0052, 0.0012, 0.0083, 0.0052, 0.0013, 0.01, 0.0031, -0.0016, 0.01, 0.003, 0.0012, 0.01, 0.0053, -0.0016, 0.01, 0.0052, -0.0038, 0.01, 0.003, -0.006, 0.01, 0.003, -0.0038, 0.01, 0.0052, -0.006, 0.01, 0.0052, 0.0009, 0.0061, 0.0074, 0.0009, 0.0083, 0.0074, 0.0005, 0.0061, 0.0096,
0.0005, 0.0083, 0.0096, 0.0009, 0.01, 0.0075, -0.0016, 0.01, 0.0074, 0.0004, 0.01, 0.0097, -0.0016, 0.01, 0.0096, -0.0038, 0.01, 0.0074, -0.006, 0.01, 0.0074, -0.0038, 0.01, 0.0096, -0.006, 0.01, 0.0096, -0.0105, 0.0061, 0.003, -0.0105, 0.0083, 0.003, -0.0106, 0.0061, 0.0052, -0.0106, 0.0083, 0.0052, -0.0081, 0.01, 0.003, -0.0105, 0.01, 0.003, -0.0081, 0.01, 0.0052, -0.0105, 0.01, 0.0051, -0.0109, 0.0061, 0.0074, -0.0109, 0.0083, 0.0074, -0.0114, 0.0061, 0.0085, -0.0114, 0.0083, 0.0085, -0.0118, 0.0061, 0.0096, -0.0118, 0.0083, 0.0096, -0.0081, 0.01, 0.0074, -0.0109, 0.01, 0.0072, -0.0081, 0.01, 0.0096, -0.0115, 0.01, 0.0089, -0.0119, 0.01, 0.0099, -0.0002, 0.0061, 0.0116, -0.0006, 0.0061, 0.0127,
-0.0002, 0.0083, 0.0116, -0.0006, 0.0083, 0.0127, -0.0011, 0.0061, 0.0139, -0.0011, 0.0083, 0.0139, -0.0001, 0.01, 0.0114, -0.0004, 0.01, 0.0123, -0.0012, 0.01, 0.0142, -0.0038, 0.01, 0.0117, -0.006, 0.01, 0.0117, -0.0038, 0.01, 0.0139, -0.006, 0.01, 0.0139, -0.0021, 0.0061, 0.0161, -0.0021, 0.0083, 0.0161, -0.0027, 0.0061, 0.0172, -0.0027, 0.0083, 0.0172, -0.0033, 0.0061, 0.0183, -0.0033, 0.0083, 0.0183, -0.0022, 0.01, 0.0162, -0.0027, 0.01, 0.0172, -0.0038, 0.01, 0.0161, -0.006, 0.01, 0.0161, -0.0034, 0.01, 0.0183, -0.006, 0.01, 0.0182, -0.0129, 0.0061, 0.0116, -0.0137, 0.0061, 0.0126, -0.0129, 0.0083, 0.0116, -0.0137, 0.0083, 0.0126, -0.0147, 0.0061, 0.0138, -0.0147, 0.0083, 0.0138, -0.0081, 0.01, 0.0117,
-0.0103, 0.01, 0.0117, -0.0081, 0.01, 0.0139, -0.0103, 0.01, 0.0139, -0.0129, 0.01, 0.0115, -0.0137, 0.01, 0.0126, -0.0125, 0.01, 0.0139, -0.0147, 0.01, 0.0138, -0.0081, 0.01, 0.0161, -0.0103, 0.01, 0.0161, -0.0081, 0.01, 0.0182, -0.0103, 0.01, 0.0182, -0.0125, 0.01, 0.0161, -0.0146, 0.01, 0.0161, -0.0125, 0.01, 0.0182, -0.0146, 0.01, 0.0182, -0.016, 0.0061, 0.0148, -0.016, 0.0083, 0.0148, -0.0159, 0.01, 0.0148, -0.017, 0.0061, 0.0156, -0.0189, 0.0061, 0.0167, -0.017, 0.0083, 0.0156, -0.0189, 0.0083, 0.0167, -0.0203, 0.0061, 0.0171, -0.0203, 0.0083, 0.0171, -0.0214, 0.0061, 0.0174, -0.0233, 0.0061, 0.0176, -0.0214, 0.0083, 0.0174, -0.0233, 0.0083, 0.0176, -0.017, 0.01, 0.0156, -0.019, 0.01, 0.0167,
-0.0168, 0.01, 0.0182, -0.019, 0.01, 0.0182, -0.0203, 0.01, 0.0171, -0.0211, 0.01, 0.0173, -0.0233, 0.01, 0.0176, -0.0324, 0.0061, 0.0144, -0.0324, 0.0083, 0.0144, -0.0325, 0.01, 0.0144, -0.028, 0.0061, 0.017, -0.028, 0.0083, 0.017, -0.0255, 0.0061, 0.0175, -0.0269, 0.0061, 0.0172, -0.0255, 0.0083, 0.0175, -0.0269, 0.0083, 0.0172, -0.0299, 0.0061, 0.0162, -0.0313, 0.0061, 0.0153, -0.0299, 0.0083, 0.0162, -0.0313, 0.0083, 0.0153, -0.0282, 0.01, 0.0169, -0.0254, 0.01, 0.0175, -0.0274, 0.01, 0.0171, -0.0297, 0.01, 0.0163, -0.0315, 0.01, 0.0152, -0.0298, 0.01, 0.0182, -0.032, 0.01, 0.0182, -0.0044, 0.0061, 0.0199, -0.0053, 0.0061, 0.021, -0.0044, 0.0083, 0.0199, -0.0053, 0.0083, 0.021, -0.0064, 0.0061, 0.0221,
-0.0064, 0.0083, 0.0221, -0.0044, 0.01, 0.0198, -0.0053, 0.01, 0.021, -0.0064, 0.01, 0.0221, -0.0076, 0.0061, 0.0232, -0.0076, 0.0083, 0.0232, -0.0081, 0.01, 0.0204, -0.0103, 0.01, 0.0204, -0.0076, 0.01, 0.0232, -0.0103, 0.01, 0.0226, -0.0125, 0.01, 0.0204, -0.0146, 0.01, 0.0204, -0.0125, 0.01, 0.0226, -0.0146, 0.01, 0.0226, -0.0087, 0.0061, 0.0241, -0.0102, 0.0061, 0.0252, -0.0087, 0.0083, 0.0241, -0.0102, 0.0083, 0.0252, -0.0113, 0.0061, 0.0259, -0.0113, 0.0083, 0.0259, -0.0124, 0.0061, 0.0265, -0.0146, 0.0061, 0.0274, -0.0124, 0.0083, 0.0265, -0.0146, 0.0083, 0.0274, -0.0087, 0.01, 0.0241, -0.0102, 0.01, 0.0252, -0.0113, 0.01, 0.0259, -0.0125, 0.01, 0.0247, -0.0146, 0.01, 0.0247, -0.0123, 0.01, 0.0265,
-0.0144, 0.01, 0.0273, -0.0168, 0.01, 0.0204, -0.019, 0.01, 0.0204, -0.0168, 0.01, 0.0226, -0.019, 0.01, 0.0226, -0.0212, 0.01, 0.0204, -0.0233, 0.01, 0.0204, -0.0212, 0.01, 0.0226, -0.0233, 0.01, 0.0226, -0.0163, 0.0061, 0.0279, -0.0163, 0.0083, 0.0279, -0.0168, 0.01, 0.0247, -0.019, 0.01, 0.0247, -0.0166, 0.01, 0.028, -0.019, 0.01, 0.0269, -0.0212, 0.01, 0.0247, -0.0233, 0.01, 0.0247, -0.0212, 0.01, 0.0269, -0.0233, 0.01, 0.0269, -0.0255, 0.01, 0.0204, -0.0277, 0.01, 0.0204, -0.0255, 0.01, 0.0226, -0.0277, 0.01, 0.0226, -0.0298, 0.01, 0.0204, -0.032, 0.01, 0.0204, -0.0298, 0.01, 0.0226, -0.032, 0.01, 0.0226, -0.0308, 0.0061, 0.028, -0.032, 0.0061, 0.0277, -0.0308, 0.0083, 0.028,
-0.032, 0.0083, 0.0277, -0.0255, 0.01, 0.0247, -0.0277, 0.01, 0.0247, -0.0255, 0.01, 0.0269, -0.0277, 0.01, 0.0269, -0.0298, 0.01, 0.0247, -0.032, 0.01, 0.0247, -0.0302, 0.01, 0.0281, -0.0323, 0.01, 0.0276, -0.0174, 0.0061, 0.0281, -0.019, 0.0061, 0.0283, -0.0174, 0.0083, 0.0281, -0.019, 0.0083, 0.0283, -0.0212, 0.0061, 0.0284, -0.0233, 0.0061, 0.0284, -0.0212, 0.0083, 0.0284, -0.0233, 0.0083, 0.0284, -0.0175, 0.01, 0.0281, -0.019, 0.01, 0.0283, -0.0212, 0.01, 0.0284, -0.0233, 0.01, 0.0284, -0.0255, 0.0061, 0.0284, -0.0277, 0.0061, 0.0284, -0.0255, 0.0083, 0.0284, -0.0277, 0.0083, 0.0284, -0.0297, 0.0061, 0.0282, -0.0297, 0.0083, 0.0282, -0.0255, 0.01, 0.0284, -0.0277, 0.01, 0.0284, -0.0294, 0.01, 0.0282,
-0.0374, 0.0061, 0.006, -0.0374, 0.0083, 0.006, -0.0378, 0.0061, 0.003, -0.0378, 0.0083, 0.003, -0.0376, 0.0061, 0.0049, -0.0376, 0.0083, 0.0049, -0.0373, 0.01, 0.0061, -0.0378, 0.01, 0.003, -0.0407, 0.01, 0.003, -0.0376, 0.01, 0.0052, -0.0407, 0.01, 0.0052, -0.037, 0.0061, 0.0074, -0.037, 0.0083, 0.0074, -0.0361, 0.0061, 0.0096, -0.0361, 0.0083, 0.0096, -0.037, 0.01, 0.0073, -0.0362, 0.01, 0.0093, -0.0385, 0.01, 0.0074, -0.0407, 0.01, 0.0074, -0.0385, 0.01, 0.0096, -0.0407, 0.01, 0.0096, -0.0487, 0.0061, 0.003, -0.0487, 0.0083, 0.003, -0.0487, 0.0061, 0.0052, -0.0487, 0.0083, 0.0052, -0.0429, 0.01, 0.003, -0.045, 0.01, 0.003, -0.0429, 0.01, 0.0052, -0.045, 0.01, 0.0052, -0.0472, 0.01, 0.003,
-0.0487, 0.01, 0.003, -0.0472, 0.01, 0.0052, -0.0487, 0.01, 0.0052, -0.0483, 0.0061, 0.0084, -0.0485, 0.0061, 0.0073, -0.0483, 0.0083, 0.0084, -0.0485, 0.0083, 0.0073, -0.048, 0.0061, 0.0096, -0.048, 0.0083, 0.0096, -0.0429, 0.01, 0.0074, -0.045, 0.01, 0.0074, -0.0429, 0.01, 0.0096, -0.045, 0.01, 0.0096, -0.0484, 0.01, 0.0078, -0.0485, 0.01, 0.007, -0.048, 0.01, 0.0098, -0.0347, 0.0061, 0.012, -0.0354, 0.0061, 0.0108, -0.0347, 0.0083, 0.012, -0.0354, 0.0083, 0.0108, -0.0335, 0.0061, 0.0133, -0.0335, 0.0083, 0.0133, -0.0346, 0.01, 0.012, -0.0352, 0.01, 0.011, -0.0335, 0.01, 0.0133, -0.0363, 0.01, 0.0139, -0.0385, 0.01, 0.0117, -0.0407, 0.01, 0.0117, -0.0385, 0.01, 0.0139, -0.0407, 0.01, 0.0139,
-0.0342, 0.01, 0.0161, -0.0363, 0.01, 0.0161, -0.0342, 0.01, 0.0182, -0.0363, 0.01, 0.0182, -0.0385, 0.01, 0.0161, -0.0407, 0.01, 0.0161, -0.0385, 0.01, 0.0182, -0.0407, 0.01, 0.0182, -0.0461, 0.0061, 0.0149, -0.0461, 0.0083, 0.0149, -0.0474, 0.0061, 0.0117, -0.0474, 0.0083, 0.0117, -0.0466, 0.0061, 0.0139, -0.0466, 0.0083, 0.0139, -0.0429, 0.01, 0.0117, -0.045, 0.01, 0.0117, -0.0429, 0.01, 0.0139, -0.0463, 0.01, 0.0146, -0.0473, 0.01, 0.0119, -0.0467, 0.01, 0.0136, -0.0455, 0.0061, 0.0161, -0.0455, 0.0083, 0.0161, -0.0437, 0.0061, 0.019, -0.0445, 0.0061, 0.0179, -0.0437, 0.0083, 0.019, -0.0445, 0.0083, 0.0179, -0.0429, 0.01, 0.0161, -0.0454, 0.01, 0.0164, -0.0438, 0.01, 0.0189, -0.0445, 0.01, 0.0178,
-0.0417, 0.0061, 0.0214, -0.0417, 0.0083, 0.0214, -0.0394, 0.0061, 0.0235, -0.0407, 0.0061, 0.0225, -0.0394, 0.0083, 0.0235, -0.0407, 0.0083, 0.0225, -0.0342, 0.01, 0.0204, -0.0363, 0.01, 0.0204, -0.0342, 0.01, 0.0226, -0.0363, 0.01, 0.0226, -0.0385, 0.01, 0.0204, -0.0417, 0.01, 0.0214, -0.0394, 0.01, 0.0236, -0.0406, 0.01, 0.0225, -0.0368, 0.0061, 0.0254, -0.0368, 0.0083, 0.0254, -0.0342, 0.0061, 0.0269, -0.0357, 0.0061, 0.0261, -0.0342, 0.0083, 0.0269, -0.0357, 0.0083, 0.0261, -0.0384, 0.0061, 0.0244, -0.0384, 0.0083, 0.0244, -0.0342, 0.01, 0.0247, -0.0367, 0.01, 0.0255, -0.0344, 0.01, 0.0268, -0.0356, 0.01, 0.0261, -0.0383, 0.01, 0.0244, -0.0427, 0.0061, 0.0204, -0.0427, 0.0083, 0.0204, -0.0427, 0.01, 0.0203,
0.0005, 0.0061, 0.0716, -0.0016, 0.0061, 0.0716, 0.0005, 0.0083, 0.0716, -0.0016, 0.0083, 0.0716, -0.0038, 0.0061, 0.0716, -0.006, 0.0061, 0.0716, -0.0038, 0.0083, 0.0716, -0.006, 0.0083, 0.0716, 0.0005, 0.01, 0.0716, -0.0016, 0.01, 0.0716, 0.0005, 0.01, 0.0747, -0.0016, 0.01, 0.0747, -0.0038, 0.01, 0.0716, -0.006, 0.01, 0.0716, -0.0038, 0.01, 0.0747, -0.006, 0.01, 0.0747, 0.0005, 0.01, 0.0768, -0.0016, 0.01, 0.0768, 0.0005, 0.01, 0.079, -0.0016, 0.01, 0.079, -0.0038, 0.01, 0.0768, -0.006, 0.01, 0.0768, -0.0038, 0.01, 0.079, -0.006, 0.01, 0.079, -0.0081, 0.0061, 0.0716, -0.0103, 0.0061, 0.0716, -0.0081, 0.0083, 0.0716, -0.0103, 0.0083, 0.0716, -0.0125, 0.0061, 0.0716, -0.0146, 0.0061, 0.0716,
-0.0125, 0.0083, 0.0716, -0.0146, 0.0083, 0.0716, -0.0081, 0.01, 0.0716, -0.0103, 0.01, 0.0716, -0.0081, 0.01, 0.0747, -0.0103, 0.01, 0.0747, -0.0125, 0.01, 0.0716, -0.0146, 0.01, 0.0716, -0.0125, 0.01, 0.0747, -0.0146, 0.01, 0.0747, -0.0081, 0.01, 0.0768, -0.0103, 0.01, 0.0768, -0.0081, 0.01, 0.079, -0.0103, 0.01, 0.079, -0.0125, 0.01, 0.0768, -0.0146, 0.01, 0.0768, -0.0125, 0.01, 0.079, -0.0146, 0.01, 0.079, 0.0005, 0.0061, 0.0833, -0.0016, 0.0061, 0.0833, 0.0005, 0.0083, 0.0833, -0.0016, 0.0083, 0.0833, -0.0038, 0.0061, 0.0833, -0.006, 0.0061, 0.0833, -0.0038, 0.0083, 0.0833, -0.006, 0.0083, 0.0833, 0.0005, 0.01, 0.0812, -0.0016, 0.01, 0.0812, 0.0005, 0.01, 0.0833, -0.0016, 0.01, 0.0833,
-0.0038, 0.01, 0.0812, -0.006, 0.01, 0.0812, -0.0038, 0.01, 0.0833, -0.006, 0.01, 0.0833, -0.0085, 0.0061, 0.0833, -0.0085, 0.0083, 0.0833, -0.0081, 0.01, 0.0812, -0.0103, 0.01, 0.0812, -0.0085, 0.01, 0.0833, -0.0103, 0.01, 0.0833, -0.0125, 0.01, 0.0812, -0.0146, 0.01, 0.0812, -0.0125, 0.01, 0.0833, -0.0146, 0.01, 0.0833, -0.0085, 0.0061, 0.0855, -0.0085, 0.0083, 0.0855, -0.0085, 0.0061, 0.0877, -0.0085, 0.0083, 0.0877, -0.0085, 0.01, 0.0855, -0.0103, 0.01, 0.0855, -0.0085, 0.01, 0.0877, -0.0103, 0.01, 0.0877, -0.0125, 0.01, 0.0855, -0.0146, 0.01, 0.0855, -0.0125, 0.01, 0.0877, -0.0146, 0.01, 0.0877, -0.0168, 0.0061, 0.0716, -0.019, 0.0061, 0.0716, -0.0168, 0.0083, 0.0716, -0.019, 0.0083, 0.0716,
-0.0212, 0.0061, 0.0716, -0.0233, 0.0061, 0.0716, -0.0212, 0.0083, 0.0716, -0.0233, 0.0083, 0.0716, -0.0168, 0.01, 0.0716, -0.019, 0.01, 0.0716, -0.0168, 0.01, 0.0747, -0.019, 0.01, 0.0747, -0.0212, 0.01, 0.0716, -0.0233, 0.01, 0.0716, -0.0212, 0.01, 0.0747, -0.0233, 0.01, 0.0747, -0.0168, 0.01, 0.0768, -0.019, 0.01, 0.0768, -0.0168, 0.01, 0.079, -0.019, 0.01, 0.079, -0.0212, 0.01, 0.0768, -0.0233, 0.01, 0.0768, -0.0212, 0.01, 0.079, -0.0233, 0.01, 0.079, -0.0255, 0.0061, 0.0716, -0.0277, 0.0061, 0.0716, -0.0255, 0.0083, 0.0716, -0.0277, 0.0083, 0.0716, -0.0298, 0.0061, 0.0716, -0.032, 0.0061, 0.0716, -0.0298, 0.0083, 0.0716, -0.032, 0.0083, 0.0716, -0.0255, 0.01, 0.0716, -0.0277, 0.01, 0.0716,
-0.0255, 0.01, 0.0747, -0.0277, 0.01, 0.0747, -0.0298, 0.01, 0.0716, -0.032, 0.01, 0.0716, -0.0298, 0.01, 0.0747, -0.032, 0.01, 0.0747, -0.0255, 0.01, 0.0768, -0.0277, 0.01, 0.0768, -0.0255, 0.01, 0.079, -0.0277, 0.01, 0.079, -0.0298, 0.01, 0.0768, -0.032, 0.01, 0.0768, -0.0298, 0.01, 0.079, -0.032, 0.01, 0.079, -0.0203, 0.0061, 0.0833, -0.0233, 0.0061, 0.0833, -0.0203, 0.0083, 0.0833, -0.0233, 0.0083, 0.0833, -0.0168, 0.01, 0.0812, -0.019, 0.01, 0.0812, -0.0168, 0.01, 0.0833, -0.019, 0.01, 0.0833, -0.0212, 0.01, 0.0812, -0.0233, 0.01, 0.0812, -0.0203, 0.01, 0.0833, -0.0233, 0.01, 0.0833, -0.0203, 0.0061, 0.0855, -0.0203, 0.0083, 0.0855, -0.0203, 0.0061, 0.0877, -0.0203, 0.0083, 0.0877,
-0.0168, 0.01, 0.0855, -0.019, 0.01, 0.0855, -0.0168, 0.01, 0.0877, -0.019, 0.01, 0.0877, -0.0203, 0.01, 0.0855, -0.0203, 0.01, 0.0877, -0.0255, 0.0061, 0.0833, -0.0277, 0.0061, 0.0833, -0.0255, 0.0083, 0.0833, -0.0277, 0.0083, 0.0833, -0.0298, 0.0061, 0.0833, -0.032, 0.0061, 0.0833, -0.0298, 0.0083, 0.0833, -0.032, 0.0083, 0.0833, -0.0255, 0.01, 0.0812, -0.0277, 0.01, 0.0812, -0.0255, 0.01, 0.0833, -0.0277, 0.01, 0.0833, -0.0298, 0.01, 0.0812, -0.032, 0.01, 0.0812, -0.0298, 0.01, 0.0833, -0.032, 0.01, 0.0833, -0.0085, 0.0061, 0.0899, -0.0085, 0.0083, 0.0899, -0.0085, 0.0061, 0.092, -0.0085, 0.0083, 0.092, -0.0085, 0.01, 0.0899, -0.0103, 0.01, 0.0899, -0.0085, 0.01, 0.092, -0.0103, 0.01, 0.092,
-0.0125, 0.01, 0.0899, -0.0146, 0.01, 0.0899, -0.0125, 0.01, 0.092, -0.0146, 0.01, 0.092, -0.0085, 0.0061, 0.0942, -0.0085, 0.0083, 0.0942, -0.0085, 0.0061, 0.0964, -0.0085, 0.0083, 0.0964, -0.0085, 0.01, 0.0942, -0.0103, 0.01, 0.0942, -0.0085, 0.01, 0.0964, -0.0103, 0.01, 0.0964, -0.0125, 0.01, 0.0942, -0.0146, 0.01, 0.0942, -0.0125, 0.01, 0.0964, -0.0146, 0.01, 0.0964, -0.0085, 0.0061, 0.0985, -0.0085, 0.0083, 0.0985, -0.0085, 0.0061, 0.1007, -0.0085, 0.0083, 0.1007, -0.0085, 0.01, 0.0985, -0.0103, 0.01, 0.0985, -0.0085, 0.01, 0.1007, -0.0103, 0.01, 0.1007, -0.0125, 0.01, 0.0985, -0.0146, 0.01, 0.0985, -0.0125, 0.01, 0.1007, -0.0146, 0.01, 0.1007, -0.0085, 0.0061, 0.1029, -0.0085, 0.0083, 0.1029,
-0.0085, 0.0061, 0.105, -0.0085, 0.0083, 0.105, -0.0085, 0.01, 0.1029, -0.0103, 0.01, 0.1029, -0.0085, 0.01, 0.105, -0.0103, 0.01, 0.105, -0.0125, 0.01, 0.1029, -0.0146, 0.01, 0.1029, -0.0125, 0.01, 0.105, -0.0146, 0.01, 0.105, -0.0203, 0.0061, 0.0899, -0.0203, 0.0083, 0.0899, -0.0203, 0.0061, 0.092, -0.0203, 0.0083, 0.092, -0.0168, 0.01, 0.0899, -0.019, 0.01, 0.0899, -0.0168, 0.01, 0.092, -0.019, 0.01, 0.092, -0.0203, 0.01, 0.0899, -0.0203, 0.01, 0.092, -0.0203, 0.0061, 0.0942, -0.0203, 0.0083, 0.0942, -0.0203, 0.0061, 0.0964, -0.0203, 0.0083, 0.0964, -0.0168, 0.01, 0.0942, -0.019, 0.01, 0.0942, -0.0168, 0.01, 0.0964, -0.019, 0.01, 0.0964, -0.0203, 0.01, 0.0942, -0.0203, 0.01, 0.0964,
-0.0203, 0.0061, 0.0985, -0.0203, 0.0083, 0.0985, -0.0203, 0.0061, 0.1007, -0.0203, 0.0083, 0.1007, -0.0168, 0.01, 0.0985, -0.019, 0.01, 0.0985, -0.0168, 0.01, 0.1007, -0.019, 0.01, 0.1007, -0.0203, 0.01, 0.0985, -0.0203, 0.01, 0.1007, -0.0203, 0.0061, 0.1029, -0.0203, 0.0083, 0.1029, -0.0203, 0.0061, 0.105, -0.0203, 0.0083, 0.105, -0.0168, 0.01, 0.1029, -0.019, 0.01, 0.1029, -0.0168, 0.01, 0.105, -0.019, 0.01, 0.105, -0.0203, 0.01, 0.1029, -0.0203, 0.01, 0.105, -0.0342, 0.0061, 0.0716, -0.0363, 0.0061, 0.0716, -0.0342, 0.0083, 0.0716, -0.0363, 0.0083, 0.0716, -0.0385, 0.0061, 0.0716, -0.0407, 0.0061, 0.0716, -0.0385, 0.0083, 0.0716, -0.0407, 0.0083, 0.0716, -0.0342, 0.01, 0.0716, -0.0363, 0.01, 0.0716,
-0.0342, 0.01, 0.0747, -0.0363, 0.01, 0.0747, -0.0385, 0.01, 0.0716, -0.0407, 0.01, 0.0716, -0.0385, 0.01, 0.0747, -0.0407, 0.01, 0.0747, -0.0342, 0.01, 0.0768, -0.0363, 0.01, 0.0768, -0.0342, 0.01, 0.079, -0.0363, 0.01, 0.079, -0.0385, 0.01, 0.0768, -0.0407, 0.01, 0.0768, -0.0385, 0.01, 0.079, -0.0407, 0.01, 0.079, -0.0429, 0.0061, 0.0716, -0.045, 0.0061, 0.0716, -0.0429, 0.0083, 0.0716, -0.045, 0.0083, 0.0716, -0.0472, 0.0061, 0.0716, -0.0487, 0.0061, 0.0716, -0.0472, 0.0083, 0.0716, -0.0487, 0.0083, 0.0716, -0.0487, 0.0061, 0.0747, -0.0487, 0.0083, 0.0747, -0.0429, 0.01, 0.0716, -0.045, 0.01, 0.0716, -0.0429, 0.01, 0.0747, -0.045, 0.01, 0.0747, -0.0472, 0.01, 0.0716, -0.0487, 0.01, 0.0716,
-0.0472, 0.01, 0.0747, -0.0487, 0.01, 0.0747, -0.0487, 0.0061, 0.0768, -0.0487, 0.0083, 0.0768, -0.0487, 0.0061, 0.079, -0.0487, 0.0083, 0.079, -0.0429, 0.01, 0.0768, -0.045, 0.01, 0.0768, -0.0429, 0.01, 0.079, -0.045, 0.01, 0.079, -0.0472, 0.01, 0.0768, -0.0487, 0.01, 0.0768, -0.0472, 0.01, 0.079, -0.0487, 0.01, 0.079, -0.0342, 0.0061, 0.0833, -0.0363, 0.0061, 0.0833, -0.0342, 0.0083, 0.0833, -0.0363, 0.0083, 0.0833, -0.0379, 0.0061, 0.0833, -0.0379, 0.0083, 0.0833, -0.0342, 0.01, 0.0812, -0.0363, 0.01, 0.0812, -0.0342, 0.01, 0.0833, -0.0363, 0.01, 0.0833, -0.0385, 0.01, 0.0812, -0.0407, 0.01, 0.0812, -0.0379, 0.01, 0.0833, -0.0407, 0.01, 0.0833, -0.0379, 0.0061, 0.0855, -0.0379, 0.0083, 0.0855,
-0.0379, 0.0061, 0.0877, -0.0379, 0.0083, 0.0877, -0.0379, 0.01, 0.0855, -0.0407, 0.01, 0.0855, -0.0379, 0.01, 0.0877, -0.0407, 0.01, 0.0877, -0.0487, 0.0061, 0.0812, -0.0487, 0.0083, 0.0812, -0.0487, 0.0061, 0.0833, -0.0487, 0.0083, 0.0833, -0.0429, 0.01, 0.0812, -0.045, 0.01, 0.0812, -0.0429, 0.01, 0.0833, -0.045, 0.01, 0.0833, -0.0472, 0.01, 0.0812, -0.0487, 0.01, 0.0812, -0.0472, 0.01, 0.0833, -0.0487, 0.01, 0.0833, -0.0487, 0.0061, 0.0855, -0.0487, 0.0083, 0.0855, -0.0487, 0.0061, 0.0877, -0.0487, 0.0083, 0.0877, -0.0429, 0.01, 0.0855, -0.045, 0.01, 0.0855, -0.0429, 0.01, 0.0877, -0.045, 0.01, 0.0877, -0.0472, 0.01, 0.0855, -0.0487, 0.01, 0.0855, -0.0472, 0.01, 0.0877, -0.0487, 0.01, 0.0877,
-0.0379, 0.0061, 0.0899, -0.0379, 0.0083, 0.0899, -0.0379, 0.0061, 0.092, -0.0379, 0.0083, 0.092, -0.0379, 0.01, 0.0899, -0.0407, 0.01, 0.0899, -0.0379, 0.01, 0.092, -0.0407, 0.01, 0.092, -0.0379, 0.0061, 0.0942, -0.0379, 0.0083, 0.0942, -0.0379, 0.0061, 0.0964, -0.0379, 0.0083, 0.0964, -0.0379, 0.01, 0.0942, -0.0407, 0.01, 0.0942, -0.0379, 0.01, 0.0964, -0.0407, 0.01, 0.0964, -0.0487, 0.0061, 0.0899, -0.0487, 0.0083, 0.0899, -0.0487, 0.0061, 0.092, -0.0487, 0.0083, 0.092, -0.0429, 0.01, 0.0899, -0.045, 0.01, 0.0899, -0.0429, 0.01, 0.092, -0.045, 0.01, 0.092, -0.0472, 0.01, 0.0899, -0.0487, 0.01, 0.0899, -0.0472, 0.01, 0.092, -0.0487, 0.01, 0.092, -0.0487, 0.0061, 0.0942, -0.0487, 0.0083, 0.0942,
-0.0487, 0.0061, 0.0964, -0.0487, 0.0083, 0.0964, -0.0429, 0.01, 0.0942, -0.045, 0.01, 0.0942, -0.0429, 0.01, 0.0964, -0.045, 0.01, 0.0964, -0.0472, 0.01, 0.0942, -0.0487, 0.01, 0.0942, -0.0472, 0.01, 0.0964, -0.0487, 0.01, 0.0964, -0.0379, 0.0061, 0.0985, -0.0379, 0.0083, 0.0985, -0.0379, 0.0061, 0.1007, -0.0379, 0.0083, 0.1007, -0.0379, 0.01, 0.0985, -0.0407, 0.01, 0.0985, -0.0379, 0.01, 0.1007, -0.0407, 0.01, 0.1007, -0.0379, 0.0061, 0.1029, -0.0379, 0.0083, 0.1029, -0.0379, 0.0061, 0.105, -0.0379, 0.0083, 0.105, -0.0379, 0.01, 0.1029, -0.0407, 0.01, 0.1029, -0.0379, 0.01, 0.105, -0.0407, 0.01, 0.105, -0.0487, 0.0061, 0.0985, -0.0487, 0.0083, 0.0985, -0.0487, 0.0061, 0.1007, -0.0487, 0.0083, 0.1007,
-0.0429, 0.01, 0.0985, -0.045, 0.01, 0.0985, -0.0429, 0.01, 0.1007, -0.045, 0.01, 0.1007, -0.0472, 0.01, 0.0985, -0.0487, 0.01, 0.0985, -0.0472, 0.01, 0.1007, -0.0487, 0.01, 0.1007, -0.0487, 0.0061, 0.1029, -0.0487, 0.0083, 0.1029, -0.0487, 0.0061, 0.105, -0.0487, 0.0083, 0.105, -0.0429, 0.01, 0.1029, -0.045, 0.01, 0.1029, -0.0429, 0.01, 0.105, -0.045, 0.01, 0.105, -0.0472, 0.01, 0.1029, -0.0487, 0.01, 0.1029, -0.0472, 0.01, 0.105, -0.0487, 0.01, 0.105, -0.0085, 0.0061, 0.1072, -0.0085, 0.0083, 0.1072, -0.0085, 0.0061, 0.1094, -0.0085, 0.0083, 0.1094, -0.0085, 0.01, 0.1072, -0.0103, 0.01, 0.1072, -0.0085, 0.01, 0.1094, -0.0103, 0.01, 0.1094, -0.0125, 0.01, 0.1072, -0.0146, 0.01, 0.1072,
-0.0125, 0.01, 0.1094, -0.0146, 0.01, 0.1094, -0.0085, 0.0061, 0.1116, -0.0085, 0.0083, 0.1116, -0.0085, 0.0061, 0.1137, -0.0085, 0.0083, 0.1137, -0.0085, 0.01, 0.1116, -0.0103, 0.01, 0.1116, -0.0085, 0.01, 0.1137, -0.0103, 0.01, 0.1137, -0.0125, 0.01, 0.1116, -0.0146, 0.01, 0.1116, -0.0125, 0.01, 0.1137, -0.0146, 0.01, 0.1137, -0.0085, 0.0061, 0.1159, -0.0085, 0.0083, 0.1159, -0.0085, 0.0061, 0.1181, -0.0085, 0.0083, 0.1181, -0.0085, 0.01, 0.1159, -0.0103, 0.01, 0.1159, -0.0085, 0.01, 0.1181, -0.0103, 0.01, 0.1181, -0.0125, 0.01, 0.1159, -0.0146, 0.01, 0.1159, -0.0125, 0.01, 0.1181, -0.0146, 0.01, 0.1181, -0.0085, 0.0061, 0.1202, -0.0085, 0.0083, 0.1202, -0.0085, 0.0061, 0.1224, -0.0085, 0.0083, 0.1224,
-0.0085, 0.01, 0.1202, -0.0103, 0.01, 0.1202, -0.0085, 0.01, 0.1224, -0.0103, 0.01, 0.1224, -0.0125, 0.01, 0.1202, -0.0146, 0.01, 0.1202, -0.0125, 0.01, 0.1224, -0.0146, 0.01, 0.1224, -0.0203, 0.0061, 0.1072, -0.0203, 0.0083, 0.1072, -0.0203, 0.0061, 0.1094, -0.0203, 0.0083, 0.1094, -0.0168, 0.01, 0.1072, -0.019, 0.01, 0.1072, -0.0168, 0.01, 0.1094, -0.019, 0.01, 0.1094, -0.0203, 0.01, 0.1072, -0.0203, 0.01, 0.1094, -0.0203, 0.0061, 0.1116, -0.0203, 0.0083, 0.1116, -0.0203, 0.0061, 0.1137, -0.0203, 0.0083, 0.1137, -0.0168, 0.01, 0.1116, -0.019, 0.01, 0.1116, -0.0168, 0.01, 0.1137, -0.019, 0.01, 0.1137, -0.0203, 0.01, 0.1116, -0.0203, 0.01, 0.1137, -0.0203, 0.0061, 0.1159, -0.0203, 0.0083, 0.1159,
-0.0203, 0.0061, 0.1181, -0.0203, 0.0083, 0.1181, -0.0168, 0.01, 0.1159, -0.019, 0.01, 0.1159, -0.0168, 0.01, 0.1181, -0.019, 0.01, 0.1181, -0.0203, 0.01, 0.1159, -0.0203, 0.01, 0.1181, -0.0203, 0.0061, 0.1202, -0.0203, 0.0083, 0.1202, -0.0203, 0.0061, 0.1224, -0.0203, 0.0083, 0.1224, -0.0168, 0.01, 0.1202, -0.019, 0.01, 0.1202, -0.0168, 0.01, 0.1224, -0.019, 0.01, 0.1224, -0.0203, 0.01, 0.1202, -0.0203, 0.01, 0.1224, -0.0085, 0.0061, 0.1246, -0.0085, 0.0083, 0.1246, -0.0085, 0.0061, 0.1267, -0.0085, 0.0083, 0.1267, -0.0085, 0.01, 0.1246, -0.0103, 0.01, 0.1246, -0.0085, 0.01, 0.1267, -0.0103, 0.01, 0.1267, -0.0125, 0.01, 0.1246, -0.0146, 0.01, 0.1246, -0.0125, 0.01, 0.1267, -0.0146, 0.01, 0.1267,
-0.0085, 0.0061, 0.1294, -0.0103, 0.0061, 0.1294, -0.0085, 0.0083, 0.1294, -0.0103, 0.0083, 0.1294, -0.0125, 0.0061, 0.1294, -0.0146, 0.0061, 0.1294, -0.0125, 0.0083, 0.1294, -0.0146, 0.0083, 0.1294, -0.0085, 0.01, 0.1294, -0.0103, 0.01, 0.1294, -0.0125, 0.01, 0.1294, -0.0146, 0.01, 0.1294, -0.0203, 0.0061, 0.1246, -0.0203, 0.0083, 0.1246, -0.0203, 0.0061, 0.1267, -0.0203, 0.0083, 0.1267, -0.0168, 0.01, 0.1246, -0.019, 0.01, 0.1246, -0.0168, 0.01, 0.1267, -0.019, 0.01, 0.1267, -0.0203, 0.01, 0.1246, -0.0203, 0.01, 0.1267, -0.0168, 0.0061, 0.1294, -0.019, 0.0061, 0.1294, -0.0168, 0.0083, 0.1294, -0.019, 0.0083, 0.1294, -0.0203, 0.0061, 0.1294, -0.0203, 0.0083, 0.1294, -0.0168, 0.01, 0.1294, -0.019, 0.01, 0.1294,
-0.0203, 0.01, 0.1294, -0.0379, 0.0061, 0.1072, -0.0379, 0.0083, 0.1072, -0.0379, 0.0061, 0.1094, -0.0379, 0.0083, 0.1094, -0.0379, 0.01, 0.1072, -0.0407, 0.01, 0.1072, -0.0379, 0.01, 0.1094, -0.0407, 0.01, 0.1094, -0.0379, 0.0061, 0.1116, -0.0379, 0.0083, 0.1116, -0.0379, 0.0061, 0.1137, -0.0379, 0.0083, 0.1137, -0.0379, 0.01, 0.1116, -0.0407, 0.01, 0.1116, -0.0379, 0.01, 0.1137, -0.0407, 0.01, 0.1137, -0.0487, 0.0061, 0.1072, -0.0487, 0.0083, 0.1072, -0.0487, 0.0061, 0.1094, -0.0487, 0.0083, 0.1094, -0.0429, 0.01, 0.1072, -0.045, 0.01, 0.1072, -0.0429, 0.01, 0.1094, -0.045, 0.01, 0.1094, -0.0472, 0.01, 0.1072, -0.0487, 0.01, 0.1072, -0.0472, 0.01, 0.1094, -0.0487, 0.01, 0.1094, -0.0487, 0.0061, 0.1116,
-0.0487, 0.0083, 0.1116, -0.0487, 0.0061, 0.1137, -0.0487, 0.0083, 0.1137, -0.0429, 0.01, 0.1116, -0.045, 0.01, 0.1116, -0.0429, 0.01, 0.1137, -0.045, 0.01, 0.1137, -0.0472, 0.01, 0.1116, -0.0487, 0.01, 0.1116, -0.0472, 0.01, 0.1137, -0.0487, 0.01, 0.1137, -0.0379, 0.0061, 0.1159, -0.0379, 0.0083, 0.1159, -0.0379, 0.0061, 0.1181, -0.0379, 0.0083, 0.1181, -0.0379, 0.01, 0.1159, -0.0407, 0.01, 0.1159, -0.0379, 0.01, 0.1181, -0.0407, 0.01, 0.1181, -0.0379, 0.0061, 0.1202, -0.0379, 0.0083, 0.1202, -0.0379, 0.0061, 0.1224, -0.0379, 0.0083, 0.1224, -0.0379, 0.01, 0.1202, -0.0407, 0.01, 0.1202, -0.0379, 0.01, 0.1224, -0.0407, 0.01, 0.1224, -0.0487, 0.0061, 0.1159, -0.0487, 0.0083, 0.1159, -0.0487, 0.0061, 0.1181,
-0.0487, 0.0083, 0.1181, -0.0429, 0.01, 0.1159, -0.045, 0.01, 0.1159, -0.0429, 0.01, 0.1181, -0.045, 0.01, 0.1181, -0.0472, 0.01, 0.1159, -0.0487, 0.01, 0.1159, -0.0472, 0.01, 0.1181, -0.0487, 0.01, 0.1181, -0.0487, 0.0061, 0.1202, -0.0487, 0.0083, 0.1202, -0.0487, 0.0061, 0.1224, -0.0487, 0.0083, 0.1224, -0.0429, 0.01, 0.1202, -0.045, 0.01, 0.1202, -0.0429, 0.01, 0.1224, -0.045, 0.01, 0.1224, -0.0472, 0.01, 0.1202, -0.0487, 0.01, 0.1202, -0.0472, 0.01, 0.1224, -0.0487, 0.01, 0.1224, -0.0379, 0.0061, 0.1246, -0.0379, 0.0083, 0.1246, -0.0379, 0.0061, 0.1267, -0.0379, 0.0083, 0.1267, -0.0379, 0.01, 0.1246, -0.0407, 0.01, 0.1246, -0.0379, 0.01, 0.1267, -0.0407, 0.01, 0.1267, -0.0379, 0.0061, 0.1289,
-0.0379, 0.0083, 0.1289, -0.0379, 0.0061, 0.1311, -0.0379, 0.0083, 0.1311, -0.0379, 0.01, 0.1289, -0.0407, 0.01, 0.1289, -0.0379, 0.01, 0.1311, -0.0407, 0.01, 0.1311, -0.0487, 0.0061, 0.1246, -0.0487, 0.0083, 0.1246, -0.0487, 0.0061, 0.1267, -0.0487, 0.0083, 0.1267, -0.0429, 0.01, 0.1246, -0.045, 0.01, 0.1246, -0.0429, 0.01, 0.1267, -0.045, 0.01, 0.1267, -0.0472, 0.01, 0.1246, -0.0487, 0.01, 0.1246, -0.0472, 0.01, 0.1267, -0.0487, 0.01, 0.1267, -0.0487, 0.0061, 0.1289, -0.0487, 0.0083, 0.1289, -0.0487, 0.0061, 0.1311, -0.0487, 0.0083, 0.1311, -0.0429, 0.01, 0.1289, -0.045, 0.01, 0.1289, -0.0429, 0.01, 0.1311, -0.045, 0.01, 0.1311, -0.0472, 0.01, 0.1289, -0.0487, 0.01, 0.1289, -0.0472, 0.01, 0.1311,
-0.0487, 0.01, 0.1311, -0.0379, 0.0061, 0.1324, -0.0407, 0.0061, 0.1324, -0.0379, 0.0083, 0.1324, -0.0407, 0.0083, 0.1324, -0.0379, 0.01, 0.1324, -0.0407, 0.01, 0.1324, -0.0429, 0.0061, 0.1324, -0.045, 0.0061, 0.1324, -0.0429, 0.0083, 0.1324, -0.045, 0.0083, 0.1324, -0.0472, 0.0061, 0.1324, -0.0487, 0.0061, 0.1324, -0.0472, 0.0083, 0.1324, -0.0487, 0.0083, 0.1324, -0.0429, 0.01, 0.1324, -0.045, 0.01, 0.1324, -0.0472, 0.01, 0.1324, -0.0487, 0.01, 0.1324, 0.0005, 0.0061, 0.1412, -0.0016, 0.0061, 0.1412, 0.0005, 0.0083, 0.1412, -0.0016, 0.0083, 0.1412, -0.0038, 0.0061, 0.1412, -0.006, 0.0061, 0.1412, -0.0038, 0.0083, 0.1412, -0.006, 0.0083, 0.1412, 0.0005, 0.01, 0.1412, -0.0016, 0.01, 0.1412, 0.0005, 0.01, 0.1441,
-0.0016, 0.01, 0.1441, -0.0038, 0.01, 0.1412, -0.006, 0.01, 0.1412, -0.0038, 0.01, 0.1441, -0.006, 0.01, 0.1441, 0.0005, 0.01, 0.1463, -0.0016, 0.01, 0.1463, 0.0005, 0.01, 0.1484, -0.0016, 0.01, 0.1484, -0.0038, 0.01, 0.1463, -0.006, 0.01, 0.1463, -0.0038, 0.01, 0.1484, -0.006, 0.01, 0.1484, -0.0081, 0.0061, 0.1412, -0.0103, 0.0061, 0.1412, -0.0081, 0.0083, 0.1412, -0.0103, 0.0083, 0.1412, -0.0125, 0.0061, 0.1412, -0.0146, 0.0061, 0.1412, -0.0125, 0.0083, 0.1412, -0.0146, 0.0083, 0.1412, -0.0081, 0.01, 0.1412, -0.0103, 0.01, 0.1412, -0.0081, 0.01, 0.1441, -0.0103, 0.01, 0.1441, -0.0125, 0.01, 0.1412, -0.0146, 0.01, 0.1412, -0.0125, 0.01, 0.1441, -0.0146, 0.01, 0.1441, -0.0081, 0.01, 0.1463,
-0.0103, 0.01, 0.1463, -0.0081, 0.01, 0.1484, -0.0103, 0.01, 0.1484, -0.0125, 0.01, 0.1463, -0.0146, 0.01, 0.1463, -0.0125, 0.01, 0.1484, -0.0146, 0.01, 0.1484, 0.0005, 0.0061, 0.152, -0.0016, 0.0061, 0.152, 0.0005, 0.0083, 0.152, -0.0016, 0.0083, 0.152, -0.0038, 0.0061, 0.152, -0.006, 0.0061, 0.152, -0.0038, 0.0083, 0.152, -0.006, 0.0083, 0.152, 0.0005, 0.01, 0.1506, -0.0016, 0.01, 0.1506, 0.0005, 0.01, 0.152, -0.0016, 0.01, 0.152, -0.0038, 0.01, 0.1506, -0.006, 0.01, 0.1506, -0.0038, 0.01, 0.152, -0.006, 0.01, 0.152, -0.0081, 0.0061, 0.152, -0.0103, 0.0061, 0.152, -0.0081, 0.0083, 0.152, -0.0103, 0.0083, 0.152, -0.0125, 0.0061, 0.152, -0.0146, 0.0061, 0.152, -0.0125, 0.0083, 0.152,
-0.0146, 0.0083, 0.152, -0.0081, 0.01, 0.1506, -0.0103, 0.01, 0.1506, -0.0081, 0.01, 0.152, -0.0103, 0.01, 0.152, -0.0125, 0.01, 0.1506, -0.0146, 0.01, 0.1506, -0.0125, 0.01, 0.152, -0.0146, 0.01, 0.152, -0.0168, 0.0061, 0.1412, -0.019, 0.0061, 0.1412, -0.0168, 0.0083, 0.1412, -0.019, 0.0083, 0.1412, -0.0212, 0.0061, 0.1412, -0.0233, 0.0061, 0.1412, -0.0212, 0.0083, 0.1412, -0.0233, 0.0083, 0.1412, -0.0168, 0.01, 0.1412, -0.019, 0.01, 0.1412, -0.0168, 0.01, 0.1441, -0.019, 0.01, 0.1441, -0.0212, 0.01, 0.1412, -0.0233, 0.01, 0.1412, -0.0212, 0.01, 0.1441, -0.0233, 0.01, 0.1441, -0.0168, 0.01, 0.1463, -0.019, 0.01, 0.1463, -0.0168, 0.01, 0.1484, -0.019, 0.01, 0.1484, -0.0212, 0.01, 0.1463,
-0.0233, 0.01, 0.1463, -0.0212, 0.01, 0.1484, -0.0233, 0.01, 0.1484, -0.0255, 0.0061, 0.1412, -0.0277, 0.0061, 0.1412, -0.0255, 0.0083, 0.1412, -0.0277, 0.0083, 0.1412, -0.0298, 0.0061, 0.1412, -0.032, 0.0061, 0.1412, -0.0298, 0.0083, 0.1412, -0.032, 0.0083, 0.1412, -0.0255, 0.01, 0.1412, -0.0277, 0.01, 0.1412, -0.0255, 0.01, 0.1441, -0.0277, 0.01, 0.1441, -0.0298, 0.01, 0.1412, -0.032, 0.01, 0.1412, -0.0298, 0.01, 0.1441, -0.032, 0.01, 0.1441, -0.0255, 0.01, 0.1463, -0.0277, 0.01, 0.1463, -0.0255, 0.01, 0.1484, -0.0277, 0.01, 0.1484, -0.0298, 0.01, 0.1463, -0.032, 0.01, 0.1463, -0.0298, 0.01, 0.1484, -0.032, 0.01, 0.1484, -0.0168, 0.0061, 0.152, -0.019, 0.0061, 0.152, -0.0168, 0.0083, 0.152,
-0.019, 0.0083, 0.152, -0.0212, 0.0061, 0.152, -0.0233, 0.0061, 0.152, -0.0212, 0.0083, 0.152, -0.0233, 0.0083, 0.152, -0.0168, 0.01, 0.1506, -0.019, 0.01, 0.1506, -0.0168, 0.01, 0.152, -0.019, 0.01, 0.152, -0.0212, 0.01, 0.1506, -0.0233, 0.01, 0.1506, -0.0212, 0.01, 0.152, -0.0233, 0.01, 0.152, -0.0255, 0.0061, 0.152, -0.0277, 0.0061, 0.152, -0.0255, 0.0083, 0.152, -0.0277, 0.0083, 0.152, -0.0298, 0.0061, 0.152, -0.032, 0.0061, 0.152, -0.0298, 0.0083, 0.152, -0.032, 0.0083, 0.152, -0.0255, 0.01, 0.1506, -0.0277, 0.01, 0.1506, -0.0255, 0.01, 0.152, -0.0277, 0.01, 0.152, -0.0298, 0.01, 0.1506, -0.032, 0.01, 0.1506, -0.0298, 0.01, 0.152, -0.032, 0.01, 0.152, -0.0342, 0.0061, 0.1412,
-0.0363, 0.0061, 0.1412, -0.0342, 0.0083, 0.1412, -0.0363, 0.0083, 0.1412, -0.0385, 0.0061, 0.1412, -0.0407, 0.0061, 0.1412, -0.0385, 0.0083, 0.1412, -0.0407, 0.0083, 0.1412, -0.0342, 0.01, 0.1412, -0.0363, 0.01, 0.1412, -0.0342, 0.01, 0.1441, -0.0363, 0.01, 0.1441, -0.0385, 0.01, 0.1412, -0.0407, 0.01, 0.1412, -0.0385, 0.01, 0.1441, -0.0407, 0.01, 0.1441, -0.0342, 0.01, 0.1463, -0.0363, 0.01, 0.1463, -0.0342, 0.01, 0.1484, -0.0363, 0.01, 0.1484, -0.0385, 0.01, 0.1463, -0.0407, 0.01, 0.1463, -0.0385, 0.01, 0.1484, -0.0407, 0.01, 0.1484, -0.0429, 0.0061, 0.1412, -0.045, 0.0061, 0.1412, -0.0429, 0.0083, 0.1412, -0.045, 0.0083, 0.1412, -0.0472, 0.0061, 0.1412, -0.0487, 0.0061, 0.1412, -0.0472, 0.0083, 0.1412,
-0.0487, 0.0083, 0.1412, -0.0487, 0.0061, 0.1441, -0.0487, 0.0083, 0.1441, -0.0429, 0.01, 0.1412, -0.045, 0.01, 0.1412, -0.0429, 0.01, 0.1441, -0.045, 0.01, 0.1441, -0.0472, 0.01, 0.1412, -0.0487, 0.01, 0.1412, -0.0472, 0.01, 0.1441, -0.0487, 0.01, 0.1441, -0.0487, 0.0061, 0.1463, -0.0487, 0.0083, 0.1463, -0.0487, 0.0061, 0.1484, -0.0487, 0.0083, 0.1484, -0.0429, 0.01, 0.1463, -0.045, 0.01, 0.1463, -0.0429, 0.01, 0.1484, -0.045, 0.01, 0.1484, -0.0472, 0.01, 0.1463, -0.0487, 0.01, 0.1463, -0.0472, 0.01, 0.1484, -0.0487, 0.01, 0.1484, -0.0342, 0.0061, 0.152, -0.0363, 0.0061, 0.152, -0.0342, 0.0083, 0.152, -0.0363, 0.0083, 0.152, -0.0385, 0.0061, 0.152, -0.0407, 0.0061, 0.152, -0.0385, 0.0083, 0.152,
-0.0407, 0.0083, 0.152, -0.0342, 0.01, 0.1506, -0.0363, 0.01, 0.1506, -0.0342, 0.01, 0.152, -0.0363, 0.01, 0.152, -0.0385, 0.01, 0.1506, -0.0407, 0.01, 0.1506, -0.0385, 0.01, 0.152, -0.0407, 0.01, 0.152, -0.0429, 0.0061, 0.152, -0.045, 0.0061, 0.152, -0.0429, 0.0083, 0.152, -0.045, 0.0083, 0.152, -0.0487, 0.0061, 0.1506, -0.0487, 0.0083, 0.1506, -0.0472, 0.0061, 0.152, -0.0487, 0.0061, 0.152, -0.0472, 0.0083, 0.152, -0.0487, 0.0083, 0.152, -0.0429, 0.01, 0.1506, -0.045, 0.01, 0.1506, -0.0429, 0.01, 0.152, -0.045, 0.01, 0.152, -0.0472, 0.01, 0.1506, -0.0487, 0.01, 0.1506, -0.0472, 0.01, 0.152, -0.0487, 0.01, 0.152, 0.0006, 0.0061, 0.2139, -0.0016, 0.0061, 0.2136, 0.0006, 0.0083, 0.2139,
-0.0016, 0.0083, 0.2136, -0.0038, 0.0061, 0.2133, -0.006, 0.0061, 0.213, -0.0038, 0.0083, 0.2133, -0.006, 0.0083, 0.213, 0.0006, 0.01, 0.2139, -0.0015, 0.01, 0.2136, -0.0037, 0.01, 0.2133, -0.0059, 0.01, 0.213, 0.0005, 0.01, 0.2157, -0.0016, 0.01, 0.2157, 0.0005, 0.01, 0.2179, -0.0016, 0.01, 0.2179, -0.0038, 0.01, 0.2157, -0.006, 0.01, 0.2157, -0.0038, 0.01, 0.2179, -0.006, 0.01, 0.2179, -0.0081, 0.0061, 0.2128, -0.0103, 0.0061, 0.2127, -0.0081, 0.0083, 0.2128, -0.0103, 0.0083, 0.2127, -0.0125, 0.0061, 0.2127, -0.0146, 0.0061, 0.2127, -0.0125, 0.0083, 0.2127, -0.0146, 0.0083, 0.2127, -0.0081, 0.01, 0.2128, -0.0103, 0.01, 0.2127, -0.0125, 0.01, 0.2127, -0.0146, 0.01, 0.2127, -0.0081, 0.01, 0.2157,
-0.0103, 0.01, 0.2157, -0.0081, 0.01, 0.2179, -0.0103, 0.01, 0.2179, -0.0125, 0.01, 0.2157, -0.0146, 0.01, 0.2157, -0.0125, 0.01, 0.2179, -0.0146, 0.01, 0.2179, 0.0005, 0.01, 0.2201, -0.0016, 0.01, 0.2201, 0.0005, 0.01, 0.2222, -0.0016, 0.01, 0.2222, -0.0038, 0.01, 0.2201, -0.006, 0.01, 0.2201, -0.0038, 0.01, 0.2222, -0.006, 0.01, 0.2222, 0.0002, 0.0061, 0.2271, -0.0016, 0.0061, 0.226, 0.0002, 0.0083, 0.2271, -0.0016, 0.0083, 0.226, -0.0039, 0.0061, 0.2252, -0.006, 0.0061, 0.2248, -0.0039, 0.0083, 0.2252, -0.006, 0.0083, 0.2248, -0.0029, 0.0061, 0.2255, -0.0029, 0.0083, 0.2255, 0.0005, 0.01, 0.2244, -0.0016, 0.01, 0.2244, 0.0002, 0.01, 0.2272, -0.0017, 0.01, 0.2259, -0.0038, 0.01, 0.2252,
-0.0061, 0.01, 0.2247, -0.0028, 0.01, 0.2256, -0.0081, 0.01, 0.2201, -0.0103, 0.01, 0.2201, -0.0081, 0.01, 0.2222, -0.0103, 0.01, 0.2222, -0.0125, 0.01, 0.2201, -0.0146, 0.01, 0.2201, -0.0125, 0.01, 0.2222, -0.0146, 0.01, 0.2222, -0.0081, 0.0061, 0.2245, -0.0103, 0.0061, 0.2245, -0.0081, 0.0083, 0.2245, -0.0103, 0.0083, 0.2245, -0.0125, 0.0061, 0.2245, -0.0146, 0.0061, 0.2245, -0.0125, 0.0083, 0.2245, -0.0146, 0.0083, 0.2245, -0.0082, 0.01, 0.2245, -0.0103, 0.01, 0.2245, -0.0125, 0.01, 0.2245, -0.0146, 0.01, 0.2245, -0.0168, 0.0061, 0.2127, -0.019, 0.0061, 0.2127, -0.0168, 0.0083, 0.2127, -0.019, 0.0083, 0.2127, -0.0212, 0.0061, 0.2127, -0.0233, 0.0061, 0.2127, -0.0212, 0.0083, 0.2127, -0.0233, 0.0083, 0.2127,
-0.0168, 0.01, 0.2127, -0.019, 0.01, 0.2127, -0.0212, 0.01, 0.2127, -0.0233, 0.01, 0.2127, -0.0168, 0.01, 0.2157, -0.019, 0.01, 0.2157, -0.0168, 0.01, 0.2179, -0.019, 0.01, 0.2179, -0.0212, 0.01, 0.2157, -0.0233, 0.01, 0.2157, -0.0212, 0.01, 0.2179, -0.0233, 0.01, 0.2179, -0.0255, 0.0061, 0.2127, -0.0277, 0.0061, 0.2127, -0.0255, 0.0083, 0.2127, -0.0277, 0.0083, 0.2127, -0.0298, 0.0061, 0.2127, -0.032, 0.0061, 0.2127, -0.0298, 0.0083, 0.2127, -0.032, 0.0083, 0.2127, -0.0255, 0.01, 0.2127, -0.0277, 0.01, 0.2127, -0.0298, 0.01, 0.2127, -0.032, 0.01, 0.2127, -0.0255, 0.01, 0.2157, -0.0277, 0.01, 0.2157, -0.0255, 0.01, 0.2179, -0.0277, 0.01, 0.2179, -0.0298, 0.01, 0.2157, -0.032, 0.01, 0.2157,
-0.0298, 0.01, 0.2179, -0.032, 0.01, 0.2179, -0.0168, 0.01, 0.2201, -0.019, 0.01, 0.2201, -0.0168, 0.01, 0.2222, -0.019, 0.01, 0.2222, -0.0212, 0.01, 0.2201, -0.0233, 0.01, 0.2201, -0.0212, 0.01, 0.2222, -0.0233, 0.01, 0.2222, -0.0168, 0.0061, 0.2245, -0.019, 0.0061, 0.2245, -0.0168, 0.0083, 0.2245, -0.019, 0.0083, 0.2245, -0.0212, 0.0061, 0.2245, -0.0233, 0.0061, 0.2245, -0.0212, 0.0083, 0.2245, -0.0233, 0.0083, 0.2245, -0.0168, 0.01, 0.2245, -0.019, 0.01, 0.2245, -0.0212, 0.01, 0.2245, -0.0233, 0.01, 0.2245, -0.0255, 0.01, 0.2201, -0.0277, 0.01, 0.2201, -0.0255, 0.01, 0.2222, -0.0277, 0.01, 0.2222, -0.0298, 0.01, 0.2201, -0.032, 0.01, 0.2201, -0.0298, 0.01, 0.2222, -0.032, 0.01, 0.2222,
-0.0255, 0.0061, 0.2245, -0.0277, 0.0061, 0.2245, -0.0255, 0.0083, 0.2245, -0.0277, 0.0083, 0.2245, -0.0298, 0.0061, 0.2245, -0.032, 0.0061, 0.2245, -0.0298, 0.0083, 0.2245, -0.032, 0.0083, 0.2245, -0.0255, 0.01, 0.2245, -0.0277, 0.01, 0.2245, -0.0298, 0.01, 0.2245, -0.032, 0.01, 0.2245, 0.0013, 0.0061, 0.2279, 0.0013, 0.0083, 0.2279, 0.0014, 0.01, 0.2279, -0.0342, 0.0061, 0.2127, -0.0363, 0.0061, 0.2127, -0.0342, 0.0083, 0.2127, -0.0363, 0.0083, 0.2127, -0.0385, 0.0061, 0.2127, -0.0407, 0.0061, 0.2127, -0.0385, 0.0083, 0.2127, -0.0407, 0.0083, 0.2127, -0.0342, 0.01, 0.2127, -0.0363, 0.01, 0.2127, -0.0385, 0.01, 0.2127, -0.0407, 0.01, 0.2127, -0.0342, 0.01, 0.2157, -0.0363, 0.01, 0.2157, -0.0342, 0.01, 0.2179,
-0.0363, 0.01, 0.2179, -0.0385, 0.01, 0.2157, -0.0407, 0.01, 0.2157, -0.0385, 0.01, 0.2179, -0.0407, 0.01, 0.2179, -0.0429, 0.0061, 0.2127, -0.045, 0.0061, 0.2127, -0.0429, 0.0083, 0.2127, -0.045, 0.0083, 0.2127, -0.0472, 0.0061, 0.2127, -0.0487, 0.0061, 0.2127, -0.0472, 0.0083, 0.2127, -0.0487, 0.0083, 0.2127, -0.0429, 0.01, 0.2127, -0.045, 0.01, 0.2127, -0.0472, 0.01, 0.2127, -0.0487, 0.01, 0.2127, -0.0487, 0.0061, 0.2157, -0.0487, 0.0083, 0.2157, -0.0487, 0.0061, 0.2179, -0.0487, 0.0083, 0.2179, -0.0429, 0.01, 0.2157, -0.045, 0.01, 0.2157, -0.0429, 0.01, 0.2179, -0.045, 0.01, 0.2179, -0.0472, 0.01, 0.2157, -0.0487, 0.01, 0.2157, -0.0472, 0.01, 0.2179, -0.0487, 0.01, 0.2179, -0.0342, 0.01, 0.2201,
-0.0363, 0.01, 0.2201, -0.0342, 0.01, 0.2222, -0.0363, 0.01, 0.2222, -0.0385, 0.01, 0.2201, -0.0407, 0.01, 0.2201, -0.0385, 0.01, 0.2222, -0.0407, 0.01, 0.2222, -0.0342, 0.0061, 0.2245, -0.0363, 0.0061, 0.2245, -0.0342, 0.0083, 0.2245, -0.0363, 0.0083, 0.2245, -0.0385, 0.0061, 0.2245, -0.0407, 0.0061, 0.2245, -0.0385, 0.0083, 0.2245, -0.0407, 0.0083, 0.2245, -0.0342, 0.01, 0.2245, -0.0363, 0.01, 0.2245, -0.0385, 0.01, 0.2245, -0.0407, 0.01, 0.2245, -0.0487, 0.0061, 0.2201, -0.0487, 0.0083, 0.2201, -0.0487, 0.0061, 0.2222, -0.0487, 0.0083, 0.2222, -0.0429, 0.01, 0.2201, -0.045, 0.01, 0.2201, -0.0429, 0.01, 0.2222, -0.045, 0.01, 0.2222, -0.0472, 0.01, 0.2201, -0.0487, 0.01, 0.2201, -0.0472, 0.01, 0.2222,
-0.0487, 0.01, 0.2222, -0.0429, 0.0061, 0.2245, -0.045, 0.0061, 0.2245, -0.0429, 0.0083, 0.2245, -0.045, 0.0083, 0.2245, -0.0472, 0.0061, 0.2245, -0.0487, 0.0061, 0.2245, -0.0472, 0.0083, 0.2245, -0.0487, 0.0083, 0.2245, -0.0429, 0.01, 0.2245, -0.045, 0.01, 0.2245, -0.0472, 0.01, 0.2245, -0.0487, 0.01, 0.2245, 0.0006, 0.0061, 0.2763, -0.0012, 0.0061, 0.2773, 0.0006, 0.0083, 0.2763, -0.0012, 0.0083, 0.2773, -0.0023, 0.0061, 0.2777, -0.0023, 0.0083, 0.2777, -0.0038, 0.0061, 0.2782, -0.006, 0.0061, 0.2787, -0.0038, 0.0083, 0.2782, -0.006, 0.0083, 0.2787, 0.0003, 0.01, 0.2765, -0.0011, 0.01, 0.2772, 0.0005, 0.01, 0.2787, -0.002, 0.01, 0.2776, -0.004, 0.01, 0.2783, -0.0061, 0.01, 0.2787, -0.0081, 0.0061, 0.2791,
-0.0103, 0.0061, 0.2793, -0.0081, 0.0083, 0.2791, -0.0103, 0.0083, 0.2793, -0.0125, 0.0061, 0.2794, -0.0146, 0.0061, 0.2794, -0.0125, 0.0083, 0.2794, -0.0146, 0.0083, 0.2794, -0.0082, 0.01, 0.2791, -0.0103, 0.01, 0.2793, -0.0125, 0.01, 0.2794, -0.0146, 0.01, 0.2794, -0.0168, 0.0061, 0.2794, -0.019, 0.0061, 0.2794, -0.0168, 0.0083, 0.2794, -0.019, 0.0083, 0.2794, -0.0212, 0.0061, 0.2794, -0.0233, 0.0061, 0.2794, -0.0212, 0.0083, 0.2794, -0.0233, 0.0083, 0.2794, -0.0168, 0.01, 0.2794, -0.019, 0.01, 0.2794, -0.0212, 0.01, 0.2794, -0.0233, 0.01, 0.2794, -0.0255, 0.0061, 0.2794, -0.0277, 0.0061, 0.2794, -0.0255, 0.0083, 0.2794, -0.0277, 0.0083, 0.2794, -0.0298, 0.0061, 0.2794, -0.032, 0.0061, 0.2794, -0.0298, 0.0083, 0.2794,
-0.032, 0.0083, 0.2794, -0.0255, 0.01, 0.2794, -0.0277, 0.01, 0.2794, -0.0298, 0.01, 0.2794, -0.032, 0.01, 0.2794, -0.0342, 0.0061, 0.2794, -0.0363, 0.0061, 0.2794, -0.0342, 0.0083, 0.2794, -0.0363, 0.0083, 0.2794, -0.0385, 0.0061, 0.2794, -0.0407, 0.0061, 0.2794, -0.0385, 0.0083, 0.2794, -0.0407, 0.0083, 0.2794, -0.0342, 0.01, 0.2794, -0.0363, 0.01, 0.2794, -0.0385, 0.01, 0.2794, -0.0407, 0.01, 0.2794, -0.0429, 0.0061, 0.2794, -0.045, 0.0061, 0.2794, -0.0429, 0.0083, 0.2794, -0.045, 0.0083, 0.2794, -0.0472, 0.0061, 0.2794, -0.0487, 0.0061, 0.2794, -0.0472, 0.0083, 0.2794, -0.0487, 0.0083, 0.2794, -0.0429, 0.01, 0.2794, -0.045, 0.01, 0.2794, -0.0472, 0.01, 0.2794, -0.0487, 0.01, 0.2794, 0.0005, 0.01, 0.2808,
-0.0016, 0.01, 0.2808, 0.0005, 0.01, 0.283, -0.0016, 0.01, 0.283, -0.0038, 0.01, 0.2808, -0.006, 0.01, 0.2808, -0.0038, 0.01, 0.283, -0.006, 0.01, 0.283, 0.0005, 0.01, 0.2852, -0.0016, 0.01, 0.2852, 0.0005, 0.01, 0.2873, -0.0016, 0.01, 0.2873, -0.0038, 0.01, 0.2852, -0.006, 0.01, 0.2852, -0.0038, 0.01, 0.2873, -0.006, 0.01, 0.2873, -0.0081, 0.01, 0.2808, -0.0103, 0.01, 0.2808, -0.0081, 0.01, 0.283, -0.0103, 0.01, 0.283, -0.0125, 0.01, 0.2808, -0.0146, 0.01, 0.2808, -0.0125, 0.01, 0.283, -0.0146, 0.01, 0.283, -0.0081, 0.01, 0.2852, -0.0103, 0.01, 0.2852, -0.0081, 0.01, 0.2873, -0.0103, 0.01, 0.2873, -0.0125, 0.01, 0.2852, -0.0146, 0.01, 0.2852, -0.0125, 0.01, 0.2873,
-0.0146, 0.01, 0.2873, 0.0006, 0.0061, 0.2897, -0.0016, 0.0061, 0.2901, 0.0006, 0.0083, 0.2897, -0.0016, 0.0083, 0.2901, -0.0038, 0.0061, 0.2903, -0.0051, 0.0061, 0.2906, -0.0038, 0.0083, 0.2903, -0.0051, 0.0083, 0.2906, -0.0062, 0.0061, 0.2907, -0.0062, 0.0083, 0.2907, 0.0007, 0.01, 0.2897, -0.0015, 0.01, 0.2901, -0.0037, 0.01, 0.2903, -0.0057, 0.01, 0.2906, -0.0065, 0.01, 0.2908, -0.0081, 0.0061, 0.291, -0.0103, 0.0061, 0.2911, -0.0081, 0.0083, 0.291, -0.0103, 0.0083, 0.2911, -0.0125, 0.0061, 0.2912, -0.0146, 0.0061, 0.2912, -0.0125, 0.0083, 0.2912, -0.0146, 0.0083, 0.2912, -0.0081, 0.01, 0.2895, -0.0103, 0.01, 0.2895, -0.0081, 0.01, 0.291, -0.0103, 0.01, 0.2911, -0.0125, 0.01, 0.2895, -0.0146, 0.01, 0.2895,
-0.0125, 0.01, 0.2912, -0.0146, 0.01, 0.2912, -0.0168, 0.01, 0.2808, -0.019, 0.01, 0.2808, -0.0168, 0.01, 0.283, -0.019, 0.01, 0.283, -0.0212, 0.01, 0.2808, -0.0233, 0.01, 0.2808, -0.0212, 0.01, 0.283, -0.0233, 0.01, 0.283, -0.0168, 0.01, 0.2852, -0.019, 0.01, 0.2852, -0.0168, 0.01, 0.2873, -0.019, 0.01, 0.2873, -0.0212, 0.01, 0.2852, -0.0233, 0.01, 0.2852, -0.0212, 0.01, 0.2873, -0.0233, 0.01, 0.2873, -0.0255, 0.01, 0.2808, -0.0277, 0.01, 0.2808, -0.0255, 0.01, 0.283, -0.0277, 0.01, 0.283, -0.0298, 0.01, 0.2808, -0.032, 0.01, 0.2808, -0.0298, 0.01, 0.283, -0.032, 0.01, 0.283, -0.0255, 0.01, 0.2852, -0.0277, 0.01, 0.2852, -0.0255, 0.01, 0.2873, -0.0277, 0.01, 0.2873,
-0.0298, 0.01, 0.2852, -0.032, 0.01, 0.2852, -0.0298, 0.01, 0.2873, -0.032, 0.01, 0.2873, -0.0168, 0.0061, 0.2912, -0.019, 0.0061, 0.2912, -0.0168, 0.0083, 0.2912, -0.019, 0.0083, 0.2912, -0.0212, 0.0061, 0.2912, -0.0233, 0.0061, 0.2912, -0.0212, 0.0083, 0.2912, -0.0233, 0.0083, 0.2912, -0.0168, 0.01, 0.2895, -0.019, 0.01, 0.2895, -0.0168, 0.01, 0.2912, -0.019, 0.01, 0.2912, -0.0212, 0.01, 0.2895, -0.0233, 0.01, 0.2895, -0.0212, 0.01, 0.2912, -0.0233, 0.01, 0.2912, -0.0255, 0.0061, 0.2912, -0.0277, 0.0061, 0.2912, -0.0255, 0.0083, 0.2912, -0.0277, 0.0083, 0.2912, -0.0298, 0.0061, 0.2912, -0.032, 0.0061, 0.2912, -0.0298, 0.0083, 0.2912, -0.032, 0.0083, 0.2912, -0.0255, 0.01, 0.2895, -0.0277, 0.01, 0.2895,
-0.0255, 0.01, 0.2912, -0.0277, 0.01, 0.2912, -0.0298, 0.01, 0.2895, -0.032, 0.01, 0.2895, -0.0298, 0.01, 0.2912, -0.032, 0.01, 0.2912, 0.0005, 0.0061, 0.3088, -0.0016, 0.0061, 0.3088, 0.0005, 0.0083, 0.3088, -0.0016, 0.0083, 0.3088, -0.0038, 0.0061, 0.3088, -0.006, 0.0061, 0.3088, -0.0038, 0.0083, 0.3088, -0.006, 0.0083, 0.3088, 0.0005, 0.01, 0.3088, -0.0016, 0.01, 0.3088, -0.0038, 0.01, 0.3088, -0.006, 0.01, 0.3088, 0.0005, 0.01, 0.3112, -0.0016, 0.01, 0.3112, 0.0005, 0.01, 0.3134, -0.0016, 0.01, 0.3134, -0.0038, 0.01, 0.3112, -0.006, 0.01, 0.3112, -0.0038, 0.01, 0.3134, -0.006, 0.01, 0.3134, -0.0081, 0.0061, 0.3088, -0.0103, 0.0061, 0.3088, -0.0081, 0.0083, 0.3088, -0.0103, 0.0083, 0.3088,
-0.0125, 0.0061, 0.3088, -0.0146, 0.0061, 0.3088, -0.0125, 0.0083, 0.3088, -0.0146, 0.0083, 0.3088, -0.0081, 0.01, 0.3088, -0.0103, 0.01, 0.3088, -0.0125, 0.01, 0.3088, -0.0146, 0.01, 0.3088, -0.0081, 0.01, 0.3112, -0.0103, 0.01, 0.3112, -0.0081, 0.01, 0.3134, -0.0103, 0.01, 0.3134, -0.0125, 0.01, 0.3112, -0.0146, 0.01, 0.3112, -0.0125, 0.01, 0.3134, -0.0146, 0.01, 0.3134, -0.0168, 0.0061, 0.3088, -0.019, 0.0061, 0.3088, -0.0168, 0.0083, 0.3088, -0.019, 0.0083, 0.3088, -0.0212, 0.0061, 0.3088, -0.0233, 0.0061, 0.3088, -0.0212, 0.0083, 0.3088, -0.0233, 0.0083, 0.3088, -0.0168, 0.01, 0.3088, -0.019, 0.01, 0.3088, -0.0212, 0.01, 0.3088, -0.0233, 0.01, 0.3088, -0.0168, 0.01, 0.3112, -0.019, 0.01, 0.3112,
-0.0168, 0.01, 0.3134, -0.019, 0.01, 0.3134, -0.0212, 0.01, 0.3112, -0.0233, 0.01, 0.3112, -0.0212, 0.01, 0.3134, -0.0233, 0.01, 0.3134, -0.0255, 0.0061, 0.3088, -0.0277, 0.0061, 0.3088, -0.0255, 0.0083, 0.3088, -0.0277, 0.0083, 0.3088, -0.0298, 0.0061, 0.3088, -0.032, 0.0061, 0.3088, -0.0298, 0.0083, 0.3088, -0.032, 0.0083, 0.3088, -0.0255, 0.01, 0.3088, -0.0277, 0.01, 0.3088, -0.0298, 0.01, 0.3088, -0.032, 0.01, 0.3088, -0.0255, 0.01, 0.3112, -0.0277, 0.01, 0.3112, -0.0255, 0.01, 0.3134, -0.0277, 0.01, 0.3134, -0.0298, 0.01, 0.3112, -0.032, 0.01, 0.3112, -0.0298, 0.01, 0.3134, -0.032, 0.01, 0.3134, -0.0342, 0.01, 0.2808, -0.0363, 0.01, 0.2808, -0.0342, 0.01, 0.283, -0.0363, 0.01, 0.283,
-0.0385, 0.01, 0.2808, -0.0407, 0.01, 0.2808, -0.0385, 0.01, 0.283, -0.0407, 0.01, 0.283, -0.0342, 0.01, 0.2852, -0.0363, 0.01, 0.2852, -0.0342, 0.01, 0.2873, -0.0363, 0.01, 0.2873, -0.0385, 0.01, 0.2852, -0.0407, 0.01, 0.2852, -0.0385, 0.01, 0.2873, -0.0407, 0.01, 0.2873, -0.0487, 0.0061, 0.2808, -0.0487, 0.0083, 0.2808, -0.0487, 0.0061, 0.283, -0.0487, 0.0083, 0.283, -0.0429, 0.01, 0.2808, -0.045, 0.01, 0.2808, -0.0429, 0.01, 0.283, -0.045, 0.01, 0.283, -0.0472, 0.01, 0.2808, -0.0487, 0.01, 0.2808, -0.0472, 0.01, 0.283, -0.0487, 0.01, 0.283, -0.0487, 0.0061, 0.2852, -0.0487, 0.0083, 0.2852, -0.0487, 0.0061, 0.2873, -0.0487, 0.0083, 0.2873, -0.0429, 0.01, 0.2852, -0.045, 0.01, 0.2852,
-0.0429, 0.01, 0.2873, -0.045, 0.01, 0.2873, -0.0472, 0.01, 0.2852, -0.0487, 0.01, 0.2852, -0.0472, 0.01, 0.2873, -0.0487, 0.01, 0.2873, -0.0342, 0.0061, 0.2912, -0.0363, 0.0061, 0.2912, -0.0342, 0.0083, 0.2912, -0.0363, 0.0083, 0.2912, -0.0385, 0.0061, 0.2912, -0.0407, 0.0061, 0.2912, -0.0385, 0.0083, 0.2912, -0.0407, 0.0083, 0.2912, -0.0342, 0.01, 0.2895, -0.0363, 0.01, 0.2895, -0.0342, 0.01, 0.2912, -0.0363, 0.01, 0.2912, -0.0385, 0.01, 0.2895, -0.0407, 0.01, 0.2895, -0.0385, 0.01, 0.2912, -0.0407, 0.01, 0.2912, -0.0429, 0.0061, 0.2912, -0.045, 0.0061, 0.2912, -0.0429, 0.0083, 0.2912, -0.045, 0.0083, 0.2912, -0.0487, 0.0061, 0.2895, -0.0487, 0.0083, 0.2895, -0.0472, 0.0061, 0.2912, -0.0487, 0.0061, 0.2912,
-0.0472, 0.0083, 0.2912, -0.0487, 0.0083, 0.2912, -0.0429, 0.01, 0.2895, -0.045, 0.01, 0.2895, -0.0429, 0.01, 0.2912, -0.045, 0.01, 0.2912, -0.0472, 0.01, 0.2895, -0.0487, 0.01, 0.2895, -0.0472, 0.01, 0.2912, -0.0487, 0.01, 0.2912, -0.0342, 0.0061, 0.3088, -0.0363, 0.0061, 0.3088, -0.0342, 0.0083, 0.3088, -0.0363, 0.0083, 0.3088, -0.0385, 0.0061, 0.3088, -0.0407, 0.0061, 0.3088, -0.0385, 0.0083, 0.3088, -0.0407, 0.0083, 0.3088, -0.0342, 0.01, 0.3088, -0.0363, 0.01, 0.3088, -0.0385, 0.01, 0.3088, -0.0407, 0.01, 0.3088, -0.0342, 0.01, 0.3112, -0.0363, 0.01, 0.3112, -0.0342, 0.01, 0.3134, -0.0363, 0.01, 0.3134, -0.0385, 0.01, 0.3112, -0.0407, 0.01, 0.3112, -0.0385, 0.01, 0.3134, -0.0407, 0.01, 0.3134,
-0.0429, 0.0061, 0.3088, -0.045, 0.0061, 0.3088, -0.0429, 0.0083, 0.3088, -0.045, 0.0083, 0.3088, -0.0472, 0.0061, 0.3088, -0.0487, 0.0061, 0.3088, -0.0472, 0.0083, 0.3088, -0.0487, 0.0083, 0.3088, -0.0429, 0.01, 0.3088, -0.045, 0.01, 0.3088, -0.0472, 0.01, 0.3088, -0.0487, 0.01, 0.3088, -0.0487, 0.0061, 0.3112, -0.0487, 0.0083, 0.3112, -0.0487, 0.0061, 0.3134, -0.0487, 0.0083, 0.3134, -0.0429, 0.01, 0.3112, -0.045, 0.01, 0.3112, -0.0429, 0.01, 0.3134, -0.045, 0.01, 0.3134, -0.0472, 0.01, 0.3112, -0.0487, 0.01, 0.3112, -0.0472, 0.01, 0.3134, -0.0487, 0.01, 0.3134, 0.0005, 0.01, 0.3155, -0.0016, 0.01, 0.3155, 0.0005, 0.01, 0.3177, -0.0016, 0.01, 0.3177, -0.0038, 0.01, 0.3155, -0.006, 0.01, 0.3155,
-0.0038, 0.01, 0.3177, -0.006, 0.01, 0.3177, 0.0005, 0.0061, 0.3206, -0.0016, 0.0061, 0.3206, 0.0005, 0.0083, 0.3206, -0.0016, 0.0083, 0.3206, -0.0038, 0.0061, 0.3206, -0.006, 0.0061, 0.3206, -0.0038, 0.0083, 0.3206, -0.006, 0.0083, 0.3206, 0.0005, 0.01, 0.3206, -0.0016, 0.01, 0.3206, -0.0038, 0.01, 0.3206, -0.006, 0.01, 0.3206, -0.0081, 0.01, 0.3155, -0.0103, 0.01, 0.3155, -0.0081, 0.01, 0.3177, -0.0103, 0.01, 0.3177, -0.0125, 0.01, 0.3155, -0.0146, 0.01, 0.3155, -0.0125, 0.01, 0.3177, -0.0146, 0.01, 0.3177, -0.0081, 0.0061, 0.3206, -0.0103, 0.0061, 0.3206, -0.0081, 0.0083, 0.3206, -0.0103, 0.0083, 0.3206, -0.0125, 0.0061, 0.3206, -0.0146, 0.0061, 0.3206, -0.0125, 0.0083, 0.3206, -0.0146, 0.0083, 0.3206,
-0.0081, 0.01, 0.3206, -0.0103, 0.01, 0.3206, -0.0125, 0.01, 0.3206, -0.0146, 0.01, 0.3206, -0.0168, 0.01, 0.3155, -0.019, 0.01, 0.3155, -0.0168, 0.01, 0.3177, -0.019, 0.01, 0.3177, -0.0212, 0.01, 0.3155, -0.0233, 0.01, 0.3155, -0.0212, 0.01, 0.3177, -0.0233, 0.01, 0.3177, -0.0168, 0.0061, 0.3206, -0.019, 0.0061, 0.3206, -0.0168, 0.0083, 0.3206, -0.019, 0.0083, 0.3206, -0.0212, 0.0061, 0.3206, -0.0233, 0.0061, 0.3206, -0.0212, 0.0083, 0.3206, -0.0233, 0.0083, 0.3206, -0.0168, 0.01, 0.3206, -0.019, 0.01, 0.3206, -0.0212, 0.01, 0.3206, -0.0233, 0.01, 0.3206, -0.0255, 0.01, 0.3155, -0.0277, 0.01, 0.3155, -0.0255, 0.01, 0.3177, -0.0277, 0.01, 0.3177, -0.0298, 0.01, 0.3155, -0.032, 0.01, 0.3155,
-0.0298, 0.01, 0.3177, -0.032, 0.01, 0.3177, -0.0255, 0.0061, 0.3206, -0.0277, 0.0061, 0.3206, -0.0255, 0.0083, 0.3206, -0.0277, 0.0083, 0.3206, -0.0298, 0.0061, 0.3206, -0.032, 0.0061, 0.3206, -0.0298, 0.0083, 0.3206, -0.032, 0.0083, 0.3206, -0.0255, 0.01, 0.3206, -0.0277, 0.01, 0.3206, -0.0298, 0.01, 0.3206, -0.032, 0.01, 0.3206, 0.0005, 0.0061, 0.3402, -0.0016, 0.0061, 0.3402, 0.0005, 0.0083, 0.3402, -0.0016, 0.0083, 0.3402, -0.0038, 0.0061, 0.3402, -0.006, 0.0061, 0.3402, -0.0038, 0.0083, 0.3402, -0.006, 0.0083, 0.3402, 0.0005, 0.01, 0.3402, -0.0016, 0.01, 0.3402, -0.0038, 0.01, 0.3402, -0.006, 0.01, 0.3402, -0.0081, 0.0061, 0.3402, -0.0103, 0.0061, 0.3402, -0.0081, 0.0083, 0.3402, -0.0103, 0.0083, 0.3402,
-0.0125, 0.0061, 0.3402, -0.0146, 0.0061, 0.3402, -0.0125, 0.0083, 0.3402, -0.0146, 0.0083, 0.3402, -0.0081, 0.01, 0.3402, -0.0103, 0.01, 0.3402, -0.0125, 0.01, 0.3402, -0.0146, 0.01, 0.3402, 0.0005, 0.01, 0.3416, -0.0016, 0.01, 0.3416, 0.0005, 0.01, 0.3438, -0.0016, 0.01, 0.3438, -0.0038, 0.01, 0.3416, -0.006, 0.01, 0.3416, -0.0038, 0.01, 0.3438, -0.006, 0.01, 0.3438, 0.0005, 0.01, 0.3459, -0.0016, 0.01, 0.3459, 0.0005, 0.01, 0.3481, -0.0016, 0.01, 0.3481, -0.0038, 0.01, 0.3459, -0.006, 0.01, 0.3459, -0.0038, 0.01, 0.3481, -0.006, 0.01, 0.3481, -0.0081, 0.01, 0.3416, -0.0103, 0.01, 0.3416, -0.0081, 0.01, 0.3438, -0.0103, 0.01, 0.3438, -0.0125, 0.01, 0.3416, -0.0146, 0.01, 0.3416,
-0.0125, 0.01, 0.3438, -0.0146, 0.01, 0.3438, -0.0081, 0.01, 0.3459, -0.0103, 0.01, 0.3459, -0.0081, 0.01, 0.3481, -0.0103, 0.01, 0.3481, -0.0125, 0.01, 0.3459, -0.0146, 0.01, 0.3459, -0.0125, 0.01, 0.3481, -0.0146, 0.01, 0.3481, -0.0168, 0.0061, 0.3402, -0.019, 0.0061, 0.3402, -0.0168, 0.0083, 0.3402, -0.019, 0.0083, 0.3402, -0.0212, 0.0061, 0.3402, -0.0233, 0.0061, 0.3402, -0.0212, 0.0083, 0.3402, -0.0233, 0.0083, 0.3402, -0.0168, 0.01, 0.3402, -0.019, 0.01, 0.3402, -0.0212, 0.01, 0.3402, -0.0233, 0.01, 0.3402, -0.0255, 0.0061, 0.3402, -0.0277, 0.0061, 0.3402, -0.0255, 0.0083, 0.3402, -0.0277, 0.0083, 0.3402, -0.0298, 0.0061, 0.3402, -0.032, 0.0061, 0.3402, -0.0298, 0.0083, 0.3402, -0.032, 0.0083, 0.3402,
-0.0255, 0.01, 0.3402, -0.0277, 0.01, 0.3402, -0.0298, 0.01, 0.3402, -0.032, 0.01, 0.3402, -0.0168, 0.01, 0.3416, -0.019, 0.01, 0.3416, -0.0168, 0.01, 0.3438, -0.019, 0.01, 0.3438, -0.0212, 0.01, 0.3416, -0.0233, 0.01, 0.3416, -0.0212, 0.01, 0.3438, -0.0233, 0.01, 0.3438, -0.0168, 0.01, 0.3459, -0.019, 0.01, 0.3459, -0.0168, 0.01, 0.3481, -0.019, 0.01, 0.3481, -0.0212, 0.01, 0.3459, -0.0233, 0.01, 0.3459, -0.0212, 0.01, 0.3481, -0.0233, 0.01, 0.3481, -0.0255, 0.01, 0.3416, -0.0277, 0.01, 0.3416, -0.0255, 0.01, 0.3438, -0.0277, 0.01, 0.3438, -0.0298, 0.01, 0.3416, -0.032, 0.01, 0.3416, -0.0298, 0.01, 0.3438, -0.032, 0.01, 0.3438, -0.0255, 0.01, 0.3459, -0.0277, 0.01, 0.3459,
-0.0255, 0.01, 0.3481, -0.0277, 0.01, 0.3481, -0.0298, 0.01, 0.3459, -0.032, 0.01, 0.3459, -0.0298, 0.01, 0.3481, -0.032, 0.01, 0.3481, -0.0342, 0.01, 0.3155, -0.0363, 0.01, 0.3155, -0.0342, 0.01, 0.3177, -0.0363, 0.01, 0.3177, -0.0385, 0.01, 0.3155, -0.0407, 0.01, 0.3155, -0.0385, 0.01, 0.3177, -0.0407, 0.01, 0.3177, -0.0342, 0.0061, 0.3206, -0.0363, 0.0061, 0.3206, -0.0342, 0.0083, 0.3206, -0.0363, 0.0083, 0.3206, -0.0385, 0.0061, 0.3206, -0.0407, 0.0061, 0.3206, -0.0385, 0.0083, 0.3206, -0.0407, 0.0083, 0.3206, -0.0342, 0.01, 0.3206, -0.0363, 0.01, 0.3206, -0.0385, 0.01, 0.3206, -0.0407, 0.01, 0.3206, -0.0487, 0.0061, 0.3155, -0.0487, 0.0083, 0.3155, -0.0487, 0.0061, 0.3177, -0.0487, 0.0083, 0.3177,
-0.0429, 0.01, 0.3155, -0.045, 0.01, 0.3155, -0.0429, 0.01, 0.3177, -0.045, 0.01, 0.3177, -0.0472, 0.01, 0.3155, -0.0487, 0.01, 0.3155, -0.0472, 0.01, 0.3177, -0.0487, 0.01, 0.3177, -0.0429, 0.0061, 0.3206, -0.045, 0.0061, 0.3206, -0.0429, 0.0083, 0.3206, -0.045, 0.0083, 0.3206, -0.0472, 0.0061, 0.3206, -0.0487, 0.0061, 0.3206, -0.0472, 0.0083, 0.3206, -0.0487, 0.0083, 0.3206, -0.0429, 0.01, 0.3206, -0.045, 0.01, 0.3206, -0.0472, 0.01, 0.3206, -0.0487, 0.01, 0.3206, -0.0342, 0.0061, 0.3402, -0.0363, 0.0061, 0.3402, -0.0342, 0.0083, 0.3402, -0.0363, 0.0083, 0.3402, -0.0385, 0.0061, 0.3402, -0.0407, 0.0061, 0.3402, -0.0385, 0.0083, 0.3402, -0.0407, 0.0083, 0.3402, -0.0342, 0.01, 0.3402, -0.0363, 0.01, 0.3402,
-0.0385, 0.01, 0.3402, -0.0407, 0.01, 0.3402, -0.0429, 0.0061, 0.3402, -0.045, 0.0061, 0.3402, -0.0429, 0.0083, 0.3402, -0.045, 0.0083, 0.3402, -0.0472, 0.0061, 0.3402, -0.0487, 0.0061, 0.3402, -0.0472, 0.0083, 0.3402, -0.0487, 0.0083, 0.3402, -0.0429, 0.01, 0.3402, -0.045, 0.01, 0.3402, -0.0472, 0.01, 0.3402, -0.0487, 0.01, 0.3402, -0.0342, 0.01, 0.3416, -0.0363, 0.01, 0.3416, -0.0342, 0.01, 0.3438, -0.0363, 0.01, 0.3438, -0.0385, 0.01, 0.3416, -0.0407, 0.01, 0.3416, -0.0385, 0.01, 0.3438, -0.0407, 0.01, 0.3438, -0.0342, 0.01, 0.3459, -0.0363, 0.01, 0.3459, -0.0342, 0.01, 0.3481, -0.0363, 0.01, 0.3481, -0.0385, 0.01, 0.3459, -0.0407, 0.01, 0.3459, -0.0385, 0.01, 0.3481, -0.0407, 0.01, 0.3481,
-0.0487, 0.0061, 0.3416, -0.0487, 0.0083, 0.3416, -0.0487, 0.0061, 0.3438, -0.0487, 0.0083, 0.3438, -0.0429, 0.01, 0.3416, -0.045, 0.01, 0.3416, -0.0429, 0.01, 0.3438, -0.045, 0.01, 0.3438, -0.0472, 0.01, 0.3416, -0.0487, 0.01, 0.3416, -0.0472, 0.01, 0.3438, -0.0487, 0.01, 0.3438, -0.0487, 0.0061, 0.3459, -0.0487, 0.0083, 0.3459, -0.0487, 0.0061, 0.3481, -0.0487, 0.0083, 0.3481, -0.0429, 0.01, 0.3459, -0.045, 0.01, 0.3459, -0.0429, 0.01, 0.3481, -0.045, 0.01, 0.3481, -0.0472, 0.01, 0.3459, -0.0487, 0.01, 0.3459, -0.0472, 0.01, 0.3481, -0.0487, 0.01, 0.3481, 0.0005, 0.0061, 0.352, -0.0016, 0.0061, 0.352, 0.0005, 0.0083, 0.352, -0.0016, 0.0083, 0.352, -0.0038, 0.0061, 0.352, -0.006, 0.0061, 0.352,
-0.0038, 0.0083, 0.352, -0.006, 0.0083, 0.352, 0.0005, 0.01, 0.3503, -0.0016, 0.01, 0.3503, 0.0005, 0.01, 0.352, -0.0016, 0.01, 0.352, -0.0038, 0.01, 0.3503, -0.006, 0.01, 0.3503, -0.0038, 0.01, 0.352, -0.006, 0.01, 0.352, -0.0081, 0.0061, 0.352, -0.0103, 0.0061, 0.352, -0.0081, 0.0083, 0.352, -0.0103, 0.0083, 0.352, -0.0125, 0.0061, 0.352, -0.0146, 0.0061, 0.352, -0.0125, 0.0083, 0.352, -0.0146, 0.0083, 0.352, -0.0081, 0.01, 0.3503, -0.0103, 0.01, 0.3503, -0.0081, 0.01, 0.352, -0.0103, 0.01, 0.352, -0.0125, 0.01, 0.3503, -0.0146, 0.01, 0.3503, -0.0125, 0.01, 0.352, -0.0146, 0.01, 0.352, -0.0168, 0.0061, 0.352, -0.019, 0.0061, 0.352, -0.0168, 0.0083, 0.352, -0.019, 0.0083, 0.352,
-0.0212, 0.0061, 0.352, -0.0233, 0.0061, 0.352, -0.0212, 0.0083, 0.352, -0.0233, 0.0083, 0.352, -0.0168, 0.01, 0.3503, -0.019, 0.01, 0.3503, -0.0168, 0.01, 0.352, -0.019, 0.01, 0.352, -0.0212, 0.01, 0.3503, -0.0233, 0.01, 0.3503, -0.0212, 0.01, 0.352, -0.0233, 0.01, 0.352, -0.0255, 0.0061, 0.352, -0.0277, 0.0061, 0.352, -0.0255, 0.0083, 0.352, -0.0277, 0.0083, 0.352, -0.0298, 0.0061, 0.352, -0.032, 0.0061, 0.352, -0.0298, 0.0083, 0.352, -0.032, 0.0083, 0.352, -0.0255, 0.01, 0.3503, -0.0277, 0.01, 0.3503, -0.0255, 0.01, 0.352, -0.0277, 0.01, 0.352, -0.0298, 0.01, 0.3503, -0.032, 0.01, 0.3503, -0.0298, 0.01, 0.352, -0.032, 0.01, 0.352, -0.0342, 0.0061, 0.352, -0.0363, 0.0061, 0.352,
-0.0342, 0.0083, 0.352, -0.0363, 0.0083, 0.352, -0.0379, 0.0061, 0.352, -0.0379, 0.0083, 0.352, -0.0342, 0.01, 0.3503, -0.0363, 0.01, 0.3503, -0.0342, 0.01, 0.352, -0.0363, 0.01, 0.352, -0.0385, 0.01, 0.3503, -0.0407, 0.01, 0.3503, -0.0379, 0.01, 0.352, -0.0407, 0.01, 0.3524, -0.0379, 0.0061, 0.3546, -0.0379, 0.0083, 0.3546, -0.0379, 0.0061, 0.3568, -0.0379, 0.0083, 0.3568, -0.0379, 0.01, 0.3546, -0.0407, 0.01, 0.3546, -0.0379, 0.01, 0.3568, -0.0407, 0.01, 0.3568, -0.0487, 0.0061, 0.3503, -0.0487, 0.0083, 0.3503, -0.0487, 0.0061, 0.3524, -0.0487, 0.0083, 0.3524, -0.0429, 0.01, 0.3503, -0.045, 0.01, 0.3503, -0.0429, 0.01, 0.3524, -0.045, 0.01, 0.3524, -0.0472, 0.01, 0.3503, -0.0487, 0.01, 0.3503,
-0.0472, 0.01, 0.3524, -0.0487, 0.01, 0.3524, -0.0487, 0.0061, 0.3546, -0.0487, 0.0083, 0.3546, -0.0487, 0.0061, 0.3568, -0.0487, 0.0083, 0.3568, -0.0429, 0.01, 0.3546, -0.045, 0.01, 0.3546, -0.0429, 0.01, 0.3568, -0.045, 0.01, 0.3568, -0.0472, 0.01, 0.3546, -0.0487, 0.01, 0.3546, -0.0472, 0.01, 0.3568, -0.0487, 0.01, 0.3568, -0.0379, 0.0061, 0.3589, -0.0379, 0.0083, 0.3589, -0.0379, 0.0061, 0.3611, -0.0379, 0.0083, 0.3611, -0.0379, 0.01, 0.3589, -0.0407, 0.01, 0.3589, -0.0379, 0.01, 0.3611, -0.0407, 0.01, 0.3611, -0.0379, 0.0061, 0.3633, -0.0379, 0.0083, 0.3633, -0.0379, 0.0061, 0.3655, -0.0379, 0.0083, 0.3655, -0.0379, 0.01, 0.3633, -0.0407, 0.01, 0.3633, -0.0379, 0.01, 0.3655, -0.0407, 0.01, 0.3655,
-0.0487, 0.0061, 0.3589, -0.0487, 0.0083, 0.3589, -0.0487, 0.0061, 0.3611, -0.0487, 0.0083, 0.3611, -0.0429, 0.01, 0.3589, -0.045, 0.01, 0.3589, -0.0429, 0.01, 0.3611, -0.045, 0.01, 0.3611, -0.0472, 0.01, 0.3589, -0.0487, 0.01, 0.3589, -0.0472, 0.01, 0.3611, -0.0487, 0.01, 0.3611, -0.0487, 0.0061, 0.3633, -0.0487, 0.0083, 0.3633, -0.0487, 0.0061, 0.3655, -0.0487, 0.0083, 0.3655, -0.0429, 0.01, 0.3633, -0.045, 0.01, 0.3633, -0.0429, 0.01, 0.3655, -0.045, 0.01, 0.3655, -0.0472, 0.01, 0.3633, -0.0487, 0.01, 0.3633, -0.0472, 0.01, 0.3655, -0.0487, 0.01, 0.3655, -0.0379, 0.0061, 0.3676, -0.0379, 0.0083, 0.3676, -0.0379, 0.0061, 0.3698, -0.0379, 0.0083, 0.3698, -0.0379, 0.01, 0.3676, -0.0407, 0.01, 0.3676,
-0.0379, 0.01, 0.3698, -0.0407, 0.01, 0.3698, -0.0379, 0.0061, 0.372, -0.0379, 0.0083, 0.372, -0.0379, 0.0061, 0.3741, -0.0379, 0.0083, 0.3741, -0.0379, 0.01, 0.372, -0.0407, 0.01, 0.372, -0.0379, 0.01, 0.3741, -0.0407, 0.01, 0.3741, -0.0487, 0.0061, 0.3676, -0.0487, 0.0083, 0.3676, -0.0487, 0.0061, 0.3698, -0.0487, 0.0083, 0.3698, -0.0429, 0.01, 0.3676, -0.045, 0.01, 0.3676, -0.0429, 0.01, 0.3698, -0.045, 0.01, 0.3698, -0.0472, 0.01, 0.3676, -0.0487, 0.01, 0.3676, -0.0472, 0.01, 0.3698, -0.0487, 0.01, 0.3698, -0.0487, 0.0061, 0.372, -0.0487, 0.0083, 0.372, -0.0487, 0.0061, 0.3741, -0.0487, 0.0083, 0.3741, -0.0429, 0.01, 0.372, -0.045, 0.01, 0.372, -0.0429, 0.01, 0.3741, -0.045, 0.01, 0.3741,
-0.0472, 0.01, 0.372, -0.0487, 0.01, 0.372, -0.0472, 0.01, 0.3741, -0.0487, 0.01, 0.3741, -0.0379, 0.0061, 0.3763, -0.0379, 0.0083, 0.3763, -0.0379, 0.0061, 0.3785, -0.0379, 0.0083, 0.3785, -0.0379, 0.01, 0.3763, -0.0407, 0.01, 0.3763, -0.0379, 0.01, 0.3784, -0.0407, 0.01, 0.3785, -0.0373, 0.0061, 0.3833, -0.0373, 0.0083, 0.3833, -0.0377, 0.0061, 0.3807, -0.0377, 0.0083, 0.3807, -0.0375, 0.0061, 0.3822, -0.0375, 0.0083, 0.3822, -0.0373, 0.01, 0.3835, -0.0377, 0.01, 0.3806, -0.0407, 0.01, 0.3806, -0.0374, 0.01, 0.3827, -0.0407, 0.01, 0.3828, -0.0487, 0.0061, 0.3763, -0.0487, 0.0083, 0.3763, -0.0487, 0.0061, 0.3785, -0.0487, 0.0083, 0.3785, -0.0429, 0.01, 0.3763, -0.045, 0.01, 0.3763, -0.0429, 0.01, 0.3785,
-0.045, 0.01, 0.3785, -0.0472, 0.01, 0.3763, -0.0487, 0.01, 0.3763, -0.0472, 0.01, 0.3785, -0.0487, 0.01, 0.3785, -0.0487, 0.0061, 0.3806, -0.0487, 0.0083, 0.3806, -0.0487, 0.0061, 0.3828, -0.0487, 0.0083, 0.3828, -0.0429, 0.01, 0.3806, -0.045, 0.01, 0.3806, -0.0429, 0.01, 0.3828, -0.045, 0.01, 0.3828, -0.0472, 0.01, 0.3806, -0.0487, 0.01, 0.3806, -0.0472, 0.01, 0.3828, -0.0487, 0.01, 0.3828, 0.0006, 0.0061, 0.3999, -0.0011, 0.0061, 0.401, 0.0006, 0.0083, 0.3999, -0.0011, 0.0083, 0.401, 0.0003, 0.01, 0.4002, -0.001, 0.01, 0.4009, -0.033, 0.0061, 0.3947, -0.033, 0.0083, 0.3947, -0.0321, 0.0061, 0.3959, -0.0321, 0.0083, 0.3959, -0.033, 0.01, 0.3947, -0.0324, 0.01, 0.3956, -0.026, 0.0061, 0.4009,
-0.0277, 0.0061, 0.3999, -0.026, 0.0083, 0.4009, -0.0277, 0.0083, 0.3999, -0.0299, 0.0061, 0.3982, -0.031, 0.0061, 0.3971, -0.0299, 0.0083, 0.3982, -0.031, 0.0083, 0.3971, -0.0288, 0.0061, 0.3991, -0.0288, 0.0083, 0.3991, -0.0261, 0.01, 0.4009, -0.0274, 0.01, 0.4001, -0.03, 0.01, 0.3982, -0.031, 0.01, 0.3971, -0.0289, 0.01, 0.399, -0.032, 0.01, 0.4002, -0.0021, 0.0061, 0.4016, -0.0021, 0.0083, 0.4016, -0.0038, 0.0061, 0.4024, -0.0058, 0.0061, 0.4032, -0.0038, 0.0083, 0.4024, -0.0058, 0.0083, 0.4032, -0.0069, 0.0061, 0.4035, -0.0069, 0.0083, 0.4035, 0.0005, 0.01, 0.4024, -0.002, 0.01, 0.4015, 0.0005, 0.01, 0.4045, -0.0016, 0.01, 0.4045, -0.004, 0.01, 0.4025, -0.0056, 0.01, 0.4031, -0.0038, 0.01, 0.4045,
-0.0064, 0.01, 0.4033, 0.0005, 0.01, 0.4067, -0.0016, 0.01, 0.4067, 0.0005, 0.01, 0.4089, -0.0016, 0.01, 0.4089, -0.0038, 0.01, 0.4067, -0.006, 0.01, 0.4067, -0.0038, 0.01, 0.4089, -0.006, 0.01, 0.4089, -0.0081, 0.0061, 0.4037, -0.0103, 0.0061, 0.4039, -0.0081, 0.0083, 0.4037, -0.0103, 0.0083, 0.4039, -0.0125, 0.0061, 0.4039, -0.0146, 0.0061, 0.4039, -0.0125, 0.0083, 0.4039, -0.0146, 0.0083, 0.4039, -0.0083, 0.01, 0.4038, -0.0104, 0.01, 0.4039, -0.0125, 0.01, 0.4039, -0.0146, 0.01, 0.4039, -0.0081, 0.01, 0.4067, -0.0103, 0.01, 0.4067, -0.0081, 0.01, 0.4089, -0.0103, 0.01, 0.4089, -0.0125, 0.01, 0.4067, -0.0146, 0.01, 0.4067, -0.0125, 0.01, 0.4089, -0.0146, 0.01, 0.4089, 0.0006, 0.0061, 0.4132,
-0.0015, 0.0061, 0.414, 0.0006, 0.0083, 0.4132, -0.0015, 0.0083, 0.414, 0.0005, 0.01, 0.411, -0.0016, 0.01, 0.411, 0.0007, 0.01, 0.4131, -0.0015, 0.01, 0.414, -0.0038, 0.01, 0.411, -0.006, 0.01, 0.411, -0.0038, 0.01, 0.4132, -0.006, 0.01, 0.4132, -0.0026, 0.0061, 0.4143, -0.0026, 0.0083, 0.4143, -0.0038, 0.0061, 0.4147, -0.006, 0.0061, 0.4153, -0.0038, 0.0083, 0.4147, -0.006, 0.0083, 0.4153, -0.0026, 0.01, 0.4144, -0.0037, 0.01, 0.4147, -0.0059, 0.01, 0.4152, -0.0081, 0.01, 0.411, -0.0103, 0.01, 0.411, -0.0081, 0.01, 0.4132, -0.0103, 0.01, 0.4132, -0.0125, 0.01, 0.411, -0.0146, 0.01, 0.411, -0.0125, 0.01, 0.4132, -0.0146, 0.01, 0.4132, -0.0081, 0.0061, 0.4156, -0.0103, 0.0061, 0.4157,
-0.0081, 0.0083, 0.4156, -0.0103, 0.0083, 0.4157, -0.0125, 0.0061, 0.4157, -0.0146, 0.0061, 0.4157, -0.0125, 0.0083, 0.4157, -0.0146, 0.0083, 0.4157, -0.0081, 0.01, 0.4156, -0.0103, 0.01, 0.4157, -0.0125, 0.01, 0.4157, -0.0146, 0.01, 0.4157, -0.0168, 0.0061, 0.4039, -0.019, 0.0061, 0.4037, -0.0168, 0.0083, 0.4039, -0.019, 0.0083, 0.4037, -0.0212, 0.0061, 0.4032, -0.0234, 0.0061, 0.4024, -0.0212, 0.0083, 0.4032, -0.0234, 0.0083, 0.4024, -0.0201, 0.0061, 0.4034, -0.0201, 0.0083, 0.4034, -0.0168, 0.01, 0.4039, -0.0188, 0.01, 0.4037, -0.0215, 0.01, 0.4031, -0.0231, 0.01, 0.4025, -0.0207, 0.01, 0.4033, -0.0233, 0.01, 0.4045, -0.0168, 0.01, 0.4067, -0.019, 0.01, 0.4067, -0.0168, 0.01, 0.4089, -0.019, 0.01, 0.4089,
-0.0212, 0.01, 0.4067, -0.0233, 0.01, 0.4067, -0.0212, 0.01, 0.4089, -0.0233, 0.01, 0.4089, -0.025, 0.0061, 0.4016, -0.025, 0.0083, 0.4016, -0.0251, 0.01, 0.4015, -0.0277, 0.01, 0.4024, -0.0255, 0.01, 0.4045, -0.0277, 0.01, 0.4045, -0.0298, 0.01, 0.4024, -0.032, 0.01, 0.4024, -0.0298, 0.01, 0.4045, -0.032, 0.01, 0.4045, -0.0255, 0.01, 0.4067, -0.0277, 0.01, 0.4067, -0.0255, 0.01, 0.4089, -0.0277, 0.01, 0.4089, -0.0298, 0.01, 0.4067, -0.032, 0.01, 0.4067, -0.0298, 0.01, 0.4089, -0.032, 0.01, 0.4089, -0.0168, 0.01, 0.411, -0.019, 0.01, 0.411, -0.0168, 0.01, 0.4132, -0.019, 0.01, 0.4132, -0.0212, 0.01, 0.411, -0.0233, 0.01, 0.411, -0.0212, 0.01, 0.4132, -0.0233, 0.01, 0.4132,
-0.0168, 0.0061, 0.4157, -0.019, 0.0061, 0.4157, -0.0168, 0.0083, 0.4157, -0.019, 0.0083, 0.4157, -0.0212, 0.0061, 0.4155, -0.0233, 0.0061, 0.4152, -0.0212, 0.0083, 0.4155, -0.0233, 0.0083, 0.4152, -0.0168, 0.01, 0.4157, -0.019, 0.01, 0.4157, -0.0212, 0.01, 0.4155, -0.0234, 0.01, 0.4152, -0.0265, 0.0061, 0.4142, -0.0277, 0.0061, 0.4138, -0.0265, 0.0083, 0.4142, -0.0277, 0.0083, 0.4138, -0.0323, 0.0061, 0.4117, -0.0323, 0.0083, 0.4117, -0.0299, 0.0061, 0.4129, -0.0312, 0.0061, 0.4123, -0.0299, 0.0083, 0.4129, -0.0312, 0.0083, 0.4123, -0.0255, 0.01, 0.411, -0.0277, 0.01, 0.411, -0.026, 0.01, 0.4144, -0.0279, 0.01, 0.4137, -0.0298, 0.01, 0.411, -0.0323, 0.01, 0.4117, -0.03, 0.01, 0.4129, -0.0312, 0.01, 0.4123,
-0.0254, 0.0061, 0.4146, -0.0254, 0.0083, 0.4146, -0.0251, 0.01, 0.4147, -0.0371, 0.0061, 0.385, -0.0371, 0.0083, 0.385, -0.0366, 0.0061, 0.3872, -0.0366, 0.0083, 0.3872, -0.0371, 0.01, 0.385, -0.0367, 0.01, 0.3871, -0.0385, 0.01, 0.385, -0.0407, 0.01, 0.385, -0.0385, 0.01, 0.3872, -0.0407, 0.01, 0.3872, -0.036, 0.0061, 0.3894, -0.036, 0.0083, 0.3894, -0.0348, 0.0061, 0.3918, -0.0354, 0.0061, 0.3907, -0.0348, 0.0083, 0.3918, -0.0354, 0.0083, 0.3907, -0.0361, 0.01, 0.3891, -0.0348, 0.01, 0.3919, -0.0352, 0.01, 0.3909, -0.0385, 0.01, 0.3893, -0.0407, 0.01, 0.3893, -0.0385, 0.01, 0.3915, -0.0407, 0.01, 0.3915, -0.0487, 0.0061, 0.385, -0.0487, 0.0083, 0.385, -0.0486, 0.0061, 0.3872, -0.0486, 0.0083, 0.3872,
-0.0429, 0.01, 0.385, -0.045, 0.01, 0.385, -0.0429, 0.01, 0.3872, -0.045, 0.01, 0.3872, -0.0472, 0.01, 0.385, -0.0487, 0.01, 0.385, -0.0472, 0.01, 0.3872, -0.0486, 0.01, 0.3872, -0.0482, 0.0061, 0.39, -0.0484, 0.0061, 0.389, -0.0482, 0.0083, 0.39, -0.0484, 0.0083, 0.389, -0.0479, 0.0061, 0.3915, -0.0479, 0.0083, 0.3915, -0.0429, 0.01, 0.3893, -0.045, 0.01, 0.3893, -0.0429, 0.01, 0.3915, -0.045, 0.01, 0.3915, -0.0483, 0.01, 0.3896, -0.0485, 0.01, 0.3887, -0.0478, 0.01, 0.3917, -0.0338, 0.0061, 0.3936, -0.0338, 0.0083, 0.3936, -0.0338, 0.01, 0.3936, -0.0363, 0.01, 0.3937, -0.0342, 0.01, 0.3958, -0.0363, 0.01, 0.3958, -0.0385, 0.01, 0.3937, -0.0407, 0.01, 0.3937, -0.0385, 0.01, 0.3958,
-0.0407, 0.01, 0.3958, -0.0342, 0.01, 0.398, -0.0363, 0.01, 0.398, -0.0342, 0.01, 0.4002, -0.0363, 0.01, 0.4002, -0.0385, 0.01, 0.398, -0.0407, 0.01, 0.398, -0.0385, 0.01, 0.4002, -0.0407, 0.01, 0.4002, -0.046, 0.0061, 0.3967, -0.046, 0.0083, 0.3967, -0.0472, 0.0061, 0.3937, -0.0472, 0.0083, 0.3937, -0.0465, 0.0061, 0.3956, -0.0465, 0.0083, 0.3956, -0.0429, 0.01, 0.3937, -0.045, 0.01, 0.3937, -0.0429, 0.01, 0.3958, -0.0461, 0.01, 0.3964, -0.0471, 0.01, 0.3938, -0.0465, 0.01, 0.3954, -0.0454, 0.0061, 0.398, -0.0454, 0.0083, 0.398, -0.0437, 0.0061, 0.4009, -0.0444, 0.0061, 0.3998, -0.0437, 0.0083, 0.4009, -0.0444, 0.0083, 0.3998, -0.0429, 0.01, 0.398, -0.0453, 0.01, 0.3983, -0.0438, 0.01, 0.4008,
-0.0445, 0.01, 0.3997, -0.0409, 0.0061, 0.4046, -0.0409, 0.0083, 0.4046, -0.0342, 0.01, 0.4024, -0.0363, 0.01, 0.4024, -0.0342, 0.01, 0.4045, -0.0363, 0.01, 0.4045, -0.0385, 0.01, 0.4024, -0.0407, 0.01, 0.4024, -0.0385, 0.01, 0.4045, -0.0409, 0.01, 0.4046, -0.0352, 0.0061, 0.4099, -0.0364, 0.0061, 0.409, -0.0352, 0.0083, 0.4099, -0.0364, 0.0083, 0.409, -0.0387, 0.0061, 0.4069, -0.0398, 0.0061, 0.4058, -0.0387, 0.0083, 0.4069, -0.0398, 0.0083, 0.4058, -0.0376, 0.0061, 0.4079, -0.0376, 0.0083, 0.4079, -0.0342, 0.01, 0.4067, -0.0363, 0.01, 0.4067, -0.035, 0.01, 0.41, -0.0367, 0.01, 0.4088, -0.0387, 0.01, 0.4069, -0.0398, 0.01, 0.4058, -0.0376, 0.01, 0.4079, -0.0427, 0.0061, 0.4024, -0.0427, 0.0083, 0.4024,
-0.0418, 0.0061, 0.4035, -0.0418, 0.0083, 0.4035, -0.0426, 0.01, 0.4026, -0.0418, 0.01, 0.4035, -0.0341, 0.0061, 0.4106, -0.0341, 0.0083, 0.4106, -0.034, 0.01, 0.4107, -0.0153, 0.0061, 0.429, -0.0153, 0.0083, 0.429, -0.0135, 0.0061, 0.4316, -0.0142, 0.0061, 0.4305, -0.0135, 0.0083, 0.4316, -0.0142, 0.0083, 0.4305, -0.0153, 0.01, 0.4291, -0.0135, 0.01, 0.4316, -0.0142, 0.01, 0.4304, -0.0129, 0.0061, 0.4327, -0.0129, 0.0083, 0.4327, -0.0119, 0.0061, 0.4349, -0.0119, 0.0083, 0.4349, -0.013, 0.01, 0.4326, -0.0146, 0.01, 0.4327, -0.012, 0.01, 0.4346, -0.0146, 0.01, 0.4349, -0.0175, 0.0061, 0.427, -0.0189, 0.0061, 0.4258, -0.0175, 0.0083, 0.427, -0.0189, 0.0083, 0.4258, -0.0211, 0.0061, 0.4245, -0.0233, 0.0061, 0.4237,
-0.0211, 0.0083, 0.4245, -0.0233, 0.0083, 0.4237, -0.0201, 0.0061, 0.4252, -0.0201, 0.0083, 0.4252, -0.0175, 0.01, 0.427, -0.0186, 0.01, 0.4261, -0.0214, 0.01, 0.4244, -0.0231, 0.01, 0.4238, -0.0204, 0.01, 0.425, -0.0233, 0.01, 0.4262, -0.0265, 0.0061, 0.423, -0.0277, 0.0061, 0.4228, -0.0265, 0.0083, 0.423, -0.0277, 0.0083, 0.4228, -0.0298, 0.0061, 0.4226, -0.0319, 0.0061, 0.4228, -0.0298, 0.0083, 0.4226, -0.0319, 0.0083, 0.4228, -0.0265, 0.01, 0.4229, -0.0277, 0.01, 0.4228, -0.0298, 0.01, 0.4226, -0.0316, 0.01, 0.4228, -0.0254, 0.0061, 0.4231, -0.0254, 0.0083, 0.4231, -0.033, 0.0061, 0.423, -0.033, 0.0083, 0.423, -0.0255, 0.01, 0.4231, -0.0277, 0.01, 0.4241, -0.0255, 0.01, 0.4262, -0.0277, 0.01, 0.4262,
-0.0298, 0.01, 0.4241, -0.0324, 0.01, 0.4229, -0.0298, 0.01, 0.4262, -0.032, 0.01, 0.4262, -0.0164, 0.0061, 0.4279, -0.0164, 0.0083, 0.4279, -0.0164, 0.01, 0.4279, -0.019, 0.01, 0.4284, -0.0168, 0.01, 0.4306, -0.019, 0.01, 0.4306, -0.0212, 0.01, 0.4284, -0.0233, 0.01, 0.4284, -0.0212, 0.01, 0.4306, -0.0233, 0.01, 0.4306, -0.0168, 0.01, 0.4327, -0.019, 0.01, 0.4327, -0.0168, 0.01, 0.4349, -0.019, 0.01, 0.4349, -0.0212, 0.01, 0.4327, -0.0233, 0.01, 0.4327, -0.0212, 0.01, 0.4349, -0.0233, 0.01, 0.4349, -0.0255, 0.01, 0.4284, -0.0277, 0.01, 0.4284, -0.0255, 0.01, 0.4306, -0.0277, 0.01, 0.4306, -0.0298, 0.01, 0.4284, -0.032, 0.01, 0.4284, -0.0298, 0.01, 0.4306, -0.032, 0.01, 0.4306,
-0.0255, 0.01, 0.4327, -0.0277, 0.01, 0.4327, -0.0255, 0.01, 0.4349, -0.0277, 0.01, 0.4349, -0.0298, 0.01, 0.4327, -0.032, 0.01, 0.4327, -0.0298, 0.01, 0.4349, -0.032, 0.01, 0.4349, -0.0112, 0.0061, 0.4375, -0.0112, 0.0083, 0.4375, -0.0109, 0.0061, 0.4392, -0.0109, 0.0083, 0.4392, -0.0115, 0.0061, 0.4364, -0.0115, 0.0083, 0.4364, -0.0112, 0.01, 0.4377, -0.0109, 0.01, 0.4392, -0.0114, 0.01, 0.4368, -0.0146, 0.01, 0.4371, -0.0125, 0.01, 0.4392, -0.0146, 0.01, 0.4392, -0.0106, 0.0061, 0.4414, -0.0106, 0.0083, 0.4414, -0.0105, 0.0061, 0.4436, -0.0105, 0.0083, 0.4436, -0.0106, 0.01, 0.4414, -0.0105, 0.01, 0.4436, -0.0125, 0.01, 0.4414, -0.0146, 0.01, 0.4414, -0.0125, 0.01, 0.4436, -0.0146, 0.01, 0.4436,
-0.0105, 0.0061, 0.4458, -0.0105, 0.0083, 0.4458, -0.0105, 0.0061, 0.4479, -0.0105, 0.0083, 0.4479, -0.0105, 0.01, 0.4458, -0.0105, 0.01, 0.4479, -0.0125, 0.01, 0.4458, -0.0146, 0.01, 0.4458, -0.0125, 0.01, 0.4479, -0.0146, 0.01, 0.4479, -0.0105, 0.0061, 0.4501, -0.0105, 0.0083, 0.4501, -0.0105, 0.0061, 0.4523, -0.0105, 0.0083, 0.4523, -0.0105, 0.01, 0.4501, -0.0105, 0.01, 0.4523, -0.0125, 0.01, 0.4501, -0.0146, 0.01, 0.4501, -0.0125, 0.01, 0.4523, -0.0146, 0.01, 0.4523, -0.0241, 0.0061, 0.4399, -0.0241, 0.0083, 0.4399, -0.0168, 0.01, 0.4371, -0.019, 0.01, 0.4371, -0.0168, 0.01, 0.4392, -0.019, 0.01, 0.4392, -0.0212, 0.01, 0.4371, -0.0233, 0.01, 0.4371, -0.0212, 0.01, 0.4392, -0.0242, 0.01, 0.4398,
-0.0232, 0.0061, 0.4414, -0.0232, 0.0083, 0.4414, -0.0225, 0.0061, 0.4436, -0.0225, 0.0083, 0.4436, -0.0168, 0.01, 0.4414, -0.019, 0.01, 0.4414, -0.0168, 0.01, 0.4436, -0.019, 0.01, 0.4436, -0.0212, 0.01, 0.4414, -0.0232, 0.01, 0.4416, -0.0212, 0.01, 0.4436, -0.0225, 0.01, 0.4436, -0.0261, 0.0061, 0.4377, -0.0276, 0.0061, 0.4367, -0.0261, 0.0083, 0.4377, -0.0276, 0.0083, 0.4367, -0.025, 0.0061, 0.4388, -0.025, 0.0083, 0.4388, -0.0297, 0.0061, 0.436, -0.0321, 0.0061, 0.4367, -0.0297, 0.0083, 0.436, -0.0321, 0.0083, 0.4367, -0.026, 0.01, 0.4378, -0.0278, 0.01, 0.4366, -0.025, 0.01, 0.4388, -0.0297, 0.01, 0.436, -0.0319, 0.01, 0.4366, -0.0222, 0.0061, 0.4467, -0.0223, 0.0061, 0.4456, -0.0222, 0.0083, 0.4467,
-0.0223, 0.0083, 0.4456, -0.0222, 0.0061, 0.4479, -0.0222, 0.0083, 0.4479, -0.0168, 0.01, 0.4458, -0.019, 0.01, 0.4458, -0.0168, 0.01, 0.4479, -0.019, 0.01, 0.4479, -0.0223, 0.01, 0.4459, -0.0223, 0.01, 0.4453, -0.0222, 0.01, 0.4479, -0.0222, 0.0061, 0.4501, -0.0222, 0.0083, 0.4501, -0.0222, 0.0061, 0.4523, -0.0222, 0.0083, 0.4523, -0.0168, 0.01, 0.4501, -0.019, 0.01, 0.4501, -0.0168, 0.01, 0.4523, -0.019, 0.01, 0.4523, -0.0222, 0.01, 0.4501, -0.0222, 0.01, 0.4523, -0.0342, 0.0061, 0.4232, -0.0364, 0.0061, 0.4237, -0.0342, 0.0083, 0.4232, -0.0364, 0.0083, 0.4237, -0.0385, 0.0061, 0.4245, -0.0385, 0.0083, 0.4245, -0.0395, 0.0061, 0.4252, -0.0407, 0.0061, 0.426, -0.0395, 0.0083, 0.4252, -0.0407, 0.0083, 0.426,
-0.0344, 0.01, 0.4232, -0.0366, 0.01, 0.4238, -0.0342, 0.01, 0.4262, -0.0363, 0.01, 0.4262, -0.0383, 0.01, 0.4244, -0.0393, 0.01, 0.4251, -0.041, 0.01, 0.4262, -0.0421, 0.0061, 0.427, -0.0421, 0.0083, 0.427, -0.042, 0.01, 0.427, -0.0342, 0.01, 0.4284, -0.0363, 0.01, 0.4284, -0.0342, 0.01, 0.4306, -0.0363, 0.01, 0.4306, -0.0385, 0.01, 0.4284, -0.0407, 0.01, 0.4284, -0.0385, 0.01, 0.4306, -0.0407, 0.01, 0.4306, -0.0342, 0.01, 0.4327, -0.0363, 0.01, 0.4327, -0.0342, 0.01, 0.4349, -0.0363, 0.01, 0.4349, -0.0385, 0.01, 0.4327, -0.0407, 0.01, 0.4327, -0.0385, 0.01, 0.4349, -0.0407, 0.01, 0.4349, -0.0432, 0.0061, 0.4281, -0.0442, 0.0061, 0.4292, -0.0432, 0.0083, 0.4281, -0.0442, 0.0083, 0.4292,
-0.0452, 0.0061, 0.4305, -0.0452, 0.0083, 0.4305, -0.0432, 0.01, 0.4281, -0.0442, 0.01, 0.4292, -0.0429, 0.01, 0.4306, -0.045, 0.01, 0.4302, -0.046, 0.0061, 0.4318, -0.046, 0.0083, 0.4318, -0.0467, 0.0061, 0.4329, -0.0467, 0.0083, 0.4329, -0.0476, 0.0061, 0.4349, -0.0476, 0.0083, 0.4349, -0.0429, 0.01, 0.4327, -0.0461, 0.01, 0.432, -0.0429, 0.01, 0.4349, -0.045, 0.01, 0.4349, -0.0468, 0.01, 0.433, -0.0475, 0.01, 0.4346, -0.0336, 0.0061, 0.4377, -0.0336, 0.0083, 0.4377, -0.0347, 0.0061, 0.4388, -0.0356, 0.0061, 0.4398, -0.0347, 0.0083, 0.4388, -0.0356, 0.0083, 0.4398, -0.0337, 0.01, 0.4377, -0.0363, 0.01, 0.4371, -0.0347, 0.01, 0.4388, -0.0356, 0.01, 0.4398, -0.0385, 0.01, 0.4371, -0.0407, 0.01, 0.4371,
-0.0385, 0.01, 0.4392, -0.0407, 0.01, 0.4392, -0.0366, 0.0061, 0.4414, -0.0366, 0.0083, 0.4414, -0.0373, 0.0061, 0.4432, -0.0373, 0.0083, 0.4432, -0.0375, 0.0061, 0.4443, -0.0375, 0.0083, 0.4443, -0.0367, 0.01, 0.4416, -0.0372, 0.01, 0.443, -0.0385, 0.01, 0.4414, -0.0407, 0.01, 0.4414, -0.0374, 0.01, 0.4439, -0.0407, 0.01, 0.4436, -0.0481, 0.0061, 0.4366, -0.0484, 0.0061, 0.4376, -0.0481, 0.0083, 0.4366, -0.0484, 0.0083, 0.4376, -0.0486, 0.0061, 0.4392, -0.0486, 0.0083, 0.4392, -0.0429, 0.01, 0.4371, -0.045, 0.01, 0.4371, -0.0429, 0.01, 0.4392, -0.045, 0.01, 0.4392, -0.0482, 0.01, 0.4368, -0.0484, 0.01, 0.4378, -0.0472, 0.01, 0.4392, -0.0486, 0.01, 0.4392, -0.0487, 0.0061, 0.4414, -0.0487, 0.0083, 0.4414,
-0.0487, 0.0061, 0.4436, -0.0487, 0.0083, 0.4436, -0.0429, 0.01, 0.4414, -0.045, 0.01, 0.4414, -0.0429, 0.01, 0.4436, -0.045, 0.01, 0.4436, -0.0472, 0.01, 0.4414, -0.0487, 0.01, 0.4414, -0.0472, 0.01, 0.4436, -0.0487, 0.01, 0.4436, -0.0378, 0.0061, 0.4457, -0.0378, 0.0083, 0.4457, -0.0379, 0.0061, 0.4479, -0.0379, 0.0083, 0.4479, -0.0378, 0.01, 0.4459, -0.0407, 0.01, 0.4458, -0.0379, 0.01, 0.4479, -0.0407, 0.01, 0.4479, -0.0379, 0.0061, 0.4501, -0.0379, 0.0083, 0.4501, -0.0379, 0.0061, 0.4523, -0.0379, 0.0083, 0.4523, -0.0379, 0.01, 0.4501, -0.0407, 0.01, 0.4501, -0.0379, 0.01, 0.4523, -0.0407, 0.01, 0.4523, -0.0487, 0.0061, 0.4458, -0.0487, 0.0083, 0.4458, -0.0487, 0.0061, 0.4479, -0.0487, 0.0083, 0.4479,
-0.0429, 0.01, 0.4458, -0.045, 0.01, 0.4458, -0.0429, 0.01, 0.4479, -0.045, 0.01, 0.4479, -0.0472, 0.01, 0.4458, -0.0487, 0.01, 0.4458, -0.0472, 0.01, 0.4479, -0.0487, 0.01, 0.4479, -0.0487, 0.0061, 0.4501, -0.0487, 0.0083, 0.4501, -0.0487, 0.0061, 0.4523, -0.0487, 0.0083, 0.4523, -0.0429, 0.01, 0.4501, -0.045, 0.01, 0.4501, -0.0429, 0.01, 0.4523, -0.045, 0.01, 0.4523, -0.0472, 0.01, 0.4501, -0.0487, 0.01, 0.4501, -0.0472, 0.01, 0.4523, -0.0487, 0.01, 0.4523, -0.0105, 0.0061, 0.4544, -0.0105, 0.0083, 0.4544, -0.0105, 0.0061, 0.4566, -0.0105, 0.0083, 0.4566, -0.0105, 0.01, 0.4544, -0.0105, 0.01, 0.4566, -0.0125, 0.01, 0.4544, -0.0146, 0.01, 0.4544, -0.0125, 0.01, 0.4566, -0.0146, 0.01, 0.4566,
-0.0105, 0.0061, 0.4588, -0.0105, 0.0083, 0.4588, -0.0105, 0.0061, 0.4609, -0.0105, 0.0083, 0.4609, -0.0105, 0.01, 0.4588, -0.0105, 0.01, 0.4609, -0.0125, 0.01, 0.4588, -0.0146, 0.01, 0.4588, -0.0125, 0.01, 0.4609, -0.0146, 0.01, 0.4609, -0.0105, 0.0061, 0.4631, -0.0105, 0.0083, 0.4631, -0.0105, 0.0061, 0.4653, -0.0105, 0.0083, 0.4653, -0.0105, 0.01, 0.4631, -0.0105, 0.01, 0.4653, -0.0125, 0.01, 0.4631, -0.0146, 0.01, 0.4631, -0.0125, 0.01, 0.4653, -0.0146, 0.01, 0.4653, -0.0105, 0.0061, 0.4675, -0.0105, 0.0083, 0.4675, -0.0105, 0.0061, 0.4696, -0.0105, 0.0083, 0.4696, -0.0105, 0.01, 0.4675, -0.0105, 0.01, 0.4696, -0.0125, 0.01, 0.4675, -0.0146, 0.01, 0.4675, -0.0125, 0.01, 0.4696, -0.0146, 0.01, 0.4696,
-0.0222, 0.0061, 0.4544, -0.0222, 0.0083, 0.4544, -0.0222, 0.0061, 0.4566, -0.0222, 0.0083, 0.4566, -0.0168, 0.01, 0.4544, -0.019, 0.01, 0.4544, -0.0168, 0.01, 0.4566, -0.019, 0.01, 0.4566, -0.0222, 0.01, 0.4544, -0.0222, 0.01, 0.4566, -0.0222, 0.0061, 0.4588, -0.0222, 0.0083, 0.4588, -0.0222, 0.0061, 0.4609, -0.0222, 0.0083, 0.4609, -0.0168, 0.01, 0.4588, -0.019, 0.01, 0.4588, -0.0168, 0.01, 0.4609, -0.019, 0.01, 0.4609, -0.0222, 0.01, 0.4588, -0.0222, 0.01, 0.4609, -0.0222, 0.0061, 0.4631, -0.0222, 0.0083, 0.4631, -0.0222, 0.0061, 0.4653, -0.0222, 0.0083, 0.4653, -0.0168, 0.01, 0.4631, -0.019, 0.01, 0.4631, -0.0168, 0.01, 0.4653, -0.019, 0.01, 0.4653, -0.0222, 0.01, 0.4631, -0.0222, 0.01, 0.4653,
-0.0222, 0.0061, 0.4675, -0.0222, 0.0083, 0.4675, -0.0222, 0.0061, 0.4696, -0.0222, 0.0083, 0.4696, -0.0168, 0.01, 0.4675, -0.019, 0.01, 0.4675, -0.0168, 0.01, 0.4696, -0.019, 0.01, 0.4696, -0.0222, 0.01, 0.4675, -0.0222, 0.01, 0.4696, -0.0105, 0.0061, 0.4718, -0.0105, 0.0083, 0.4718, -0.0105, 0.0061, 0.474, -0.0105, 0.0083, 0.474, -0.0105, 0.01, 0.4718, -0.0105, 0.01, 0.474, -0.0125, 0.01, 0.4718, -0.0146, 0.01, 0.4718, -0.0125, 0.01, 0.474, -0.0146, 0.01, 0.474, -0.0104, 0.0061, 0.4761, -0.0104, 0.0083, 0.4761, -0.0104, 0.0061, 0.4783, -0.0104, 0.0083, 0.4783, -0.0104, 0.01, 0.4761, -0.0104, 0.01, 0.4783, -0.0125, 0.01, 0.4761, -0.0146, 0.01, 0.4761, -0.0125, 0.01, 0.4783, -0.0146, 0.01, 0.4783,
-0.0067, 0.0061, 0.4877, -0.0067, 0.0083, 0.4877, -0.0067, 0.01, 0.4877, -0.0102, 0.0061, 0.4805, -0.0102, 0.0083, 0.4805, -0.0092, 0.0061, 0.4836, -0.0096, 0.0061, 0.4826, -0.0092, 0.0083, 0.4836, -0.0096, 0.0083, 0.4826, -0.0102, 0.01, 0.4804, -0.0092, 0.01, 0.4837, -0.0096, 0.01, 0.4826, -0.0125, 0.01, 0.4805, -0.0146, 0.01, 0.4805, -0.0125, 0.01, 0.4826, -0.0146, 0.01, 0.4826, -0.0087, 0.0061, 0.4849, -0.0087, 0.0083, 0.4849, -0.0076, 0.0061, 0.4866, -0.0076, 0.0083, 0.4866, -0.0087, 0.01, 0.4847, -0.0103, 0.01, 0.4848, -0.0076, 0.01, 0.4866, -0.0103, 0.01, 0.487, -0.0125, 0.01, 0.4848, -0.0146, 0.01, 0.4848, -0.0125, 0.01, 0.487, -0.0146, 0.01, 0.487, -0.0222, 0.0061, 0.4718, -0.0222, 0.0083, 0.4718,
-0.0222, 0.0061, 0.474, -0.0222, 0.0083, 0.474, -0.0168, 0.01, 0.4718, -0.019, 0.01, 0.4718, -0.0168, 0.01, 0.474, -0.019, 0.01, 0.474, -0.0222, 0.01, 0.4718, -0.0222, 0.01, 0.474, -0.0222, 0.0061, 0.4761, -0.0222, 0.0083, 0.4761, -0.0222, 0.0061, 0.4783, -0.0222, 0.0083, 0.4783, -0.0168, 0.01, 0.4761, -0.019, 0.01, 0.4761, -0.0168, 0.01, 0.4783, -0.019, 0.01, 0.4783, -0.0222, 0.01, 0.4761, -0.0222, 0.01, 0.4783, -0.0222, 0.0061, 0.4805, -0.0222, 0.0083, 0.4805, -0.0219, 0.0061, 0.4827, -0.0219, 0.0083, 0.4827, -0.0168, 0.01, 0.4805, -0.019, 0.01, 0.4805, -0.0168, 0.01, 0.4826, -0.019, 0.01, 0.4826, -0.0221, 0.01, 0.4806, -0.0219, 0.01, 0.4828, -0.0214, 0.0061, 0.4848, -0.0214, 0.0083, 0.4848,
-0.0208, 0.0061, 0.487, -0.0208, 0.0083, 0.487, -0.0168, 0.01, 0.4848, -0.019, 0.01, 0.4848, -0.0168, 0.01, 0.487, -0.019, 0.01, 0.487, -0.0214, 0.01, 0.485, -0.0208, 0.01, 0.4871, -0.0379, 0.0061, 0.4544, -0.0379, 0.0083, 0.4544, -0.0379, 0.0061, 0.4566, -0.0379, 0.0083, 0.4566, -0.0379, 0.01, 0.4544, -0.0407, 0.01, 0.4544, -0.0379, 0.01, 0.4566, -0.0407, 0.01, 0.4566, -0.0379, 0.0061, 0.4588, -0.0379, 0.0083, 0.4588, -0.0379, 0.0061, 0.4609, -0.0379, 0.0083, 0.4609, -0.0379, 0.01, 0.4588, -0.0407, 0.01, 0.4588, -0.0379, 0.01, 0.4609, -0.0407, 0.01, 0.4609, -0.0487, 0.0061, 0.4544, -0.0487, 0.0083, 0.4544, -0.0487, 0.0061, 0.4566, -0.0487, 0.0083, 0.4566, -0.0429, 0.01, 0.4544, -0.045, 0.01, 0.4544,
-0.0429, 0.01, 0.4566, -0.045, 0.01, 0.4566, -0.0472, 0.01, 0.4544, -0.0487, 0.01, 0.4544, -0.0472, 0.01, 0.4566, -0.0487, 0.01, 0.4566, -0.0487, 0.0061, 0.4588, -0.0487, 0.0083, 0.4588, -0.0487, 0.0061, 0.4609, -0.0487, 0.0083, 0.4609, -0.0429, 0.01, 0.4588, -0.045, 0.01, 0.4588, -0.0429, 0.01, 0.4609, -0.045, 0.01, 0.4609, -0.0472, 0.01, 0.4588, -0.0487, 0.01, 0.4588, -0.0472, 0.01, 0.4609, -0.0487, 0.01, 0.4609, -0.0379, 0.0061, 0.4631, -0.0379, 0.0083, 0.4631, -0.0379, 0.0061, 0.4653, -0.0379, 0.0083, 0.4653, -0.0379, 0.01, 0.4631, -0.0407, 0.01, 0.4631, -0.0379, 0.01, 0.4653, -0.0407, 0.01, 0.4653, -0.0379, 0.0061, 0.4675, -0.0379, 0.0083, 0.4675, -0.0379, 0.0061, 0.4696, -0.0379, 0.0083, 0.4696,
-0.0379, 0.01, 0.4675, -0.0407, 0.01, 0.4675, -0.0379, 0.01, 0.4696, -0.0407, 0.01, 0.4696, -0.0487, 0.0061, 0.4631, -0.0487, 0.0083, 0.4631, -0.0487, 0.0061, 0.4653, -0.0487, 0.0083, 0.4653, -0.0429, 0.01, 0.4631, -0.045, 0.01, 0.4631, -0.0429, 0.01, 0.4653, -0.045, 0.01, 0.4653, -0.0472, 0.01, 0.4631, -0.0487, 0.01, 0.4631, -0.0472, 0.01, 0.4653, -0.0487, 0.01, 0.4653, -0.0487, 0.0061, 0.4675, -0.0487, 0.0083, 0.4675, -0.0487, 0.0061, 0.4696, -0.0487, 0.0083, 0.4696, -0.0429, 0.01, 0.4675, -0.045, 0.01, 0.4675, -0.0429, 0.01, 0.4696, -0.045, 0.01, 0.4696, -0.0472, 0.01, 0.4675, -0.0487, 0.01, 0.4675, -0.0472, 0.01, 0.4696, -0.0487, 0.01, 0.4696, -0.0379, 0.0061, 0.4718, -0.0379, 0.0083, 0.4718,
-0.0379, 0.0061, 0.474, -0.0379, 0.0083, 0.474, -0.0379, 0.01, 0.4718, -0.0407, 0.01, 0.4718, -0.0379, 0.01, 0.474, -0.0407, 0.01, 0.474, -0.0379, 0.0061, 0.4761, -0.0379, 0.0083, 0.4761, -0.0379, 0.0061, 0.4783, -0.0379, 0.0083, 0.4783, -0.0379, 0.01, 0.4761, -0.0407, 0.01, 0.4761, -0.0379, 0.01, 0.4783, -0.0407, 0.01, 0.4783, -0.0487, 0.0061, 0.4718, -0.0487, 0.0083, 0.4718, -0.0487, 0.0061, 0.474, -0.0487, 0.0083, 0.474, -0.0429, 0.01, 0.4718, -0.045, 0.01, 0.4718, -0.0429, 0.01, 0.474, -0.045, 0.01, 0.474, -0.0472, 0.01, 0.4718, -0.0487, 0.01, 0.4718, -0.0472, 0.01, 0.474, -0.0487, 0.01, 0.474, -0.0487, 0.0061, 0.4761, -0.0487, 0.0083, 0.4761, -0.0487, 0.0061, 0.4783, -0.0487, 0.0083, 0.4783,
-0.0429, 0.01, 0.4761, -0.045, 0.01, 0.4761, -0.0429, 0.01, 0.4783, -0.045, 0.01, 0.4783, -0.0472, 0.01, 0.4761, -0.0487, 0.01, 0.4761, -0.0472, 0.01, 0.4783, -0.0487, 0.01, 0.4783, -0.0379, 0.0061, 0.4805, -0.0379, 0.0083, 0.4805, -0.0379, 0.0061, 0.4826, -0.0379, 0.0083, 0.4826, -0.0379, 0.01, 0.4805, -0.0407, 0.01, 0.4805, -0.0379, 0.01, 0.4826, -0.0407, 0.01, 0.4826, -0.0379, 0.0061, 0.4848, -0.0379, 0.0083, 0.4848, -0.0379, 0.0061, 0.487, -0.0379, 0.0083, 0.487, -0.0379, 0.01, 0.4848, -0.0407, 0.01, 0.4848, -0.0379, 0.01, 0.487, -0.0407, 0.01, 0.487, -0.0487, 0.0061, 0.4805, -0.0487, 0.0083, 0.4805, -0.0487, 0.0061, 0.4826, -0.0487, 0.0083, 0.4826, -0.0429, 0.01, 0.4805, -0.045, 0.01, 0.4805,
-0.0429, 0.01, 0.4826, -0.045, 0.01, 0.4826, -0.0472, 0.01, 0.4805, -0.0487, 0.01, 0.4805, -0.0472, 0.01, 0.4826, -0.0487, 0.01, 0.4826, -0.0487, 0.0061, 0.4848, -0.0487, 0.0083, 0.4848, -0.0487, 0.0061, 0.487, -0.0487, 0.0083, 0.487, -0.0429, 0.01, 0.4848, -0.045, 0.01, 0.4848, -0.0429, 0.01, 0.487, -0.045, 0.01, 0.487, -0.0472, 0.01, 0.4848, -0.0487, 0.01, 0.4848, -0.0472, 0.01, 0.487, -0.0487, 0.01, 0.487, 0.0006, 0.0061, 0.491, -0.0017, 0.0061, 0.4911, 0.0006, 0.0083, 0.491, -0.0017, 0.0083, 0.4911, -0.0042, 0.0061, 0.4898, -0.0057, 0.0061, 0.4888, -0.0042, 0.0083, 0.4898, -0.0057, 0.0083, 0.4888, -0.0032, 0.0061, 0.4905, -0.0032, 0.0083, 0.4905, 0.0004, 0.01, 0.491, -0.0017, 0.01, 0.4911,
-0.0043, 0.01, 0.4897, -0.0057, 0.01, 0.4888, -0.0033, 0.01, 0.4904, -0.006, 0.01, 0.4913, 0.0005, 0.01, 0.4935, -0.0016, 0.01, 0.4935, 0.0005, 0.01, 0.4957, -0.0016, 0.01, 0.4957, -0.0038, 0.01, 0.4935, -0.006, 0.01, 0.4935, -0.0038, 0.01, 0.4957, -0.006, 0.01, 0.4957, -0.0081, 0.01, 0.4892, -0.0103, 0.01, 0.4892, -0.0081, 0.01, 0.4913, -0.0103, 0.01, 0.4913, -0.0125, 0.01, 0.4892, -0.0146, 0.01, 0.4892, -0.0125, 0.01, 0.4913, -0.0146, 0.01, 0.4913, -0.0155, 0.0061, 0.4965, -0.0155, 0.0083, 0.4965, -0.0081, 0.01, 0.4935, -0.0103, 0.01, 0.4935, -0.0081, 0.01, 0.4957, -0.0103, 0.01, 0.4957, -0.0125, 0.01, 0.4935, -0.0146, 0.01, 0.4935, -0.0125, 0.01, 0.4957, -0.0154, 0.01, 0.4965,
0.0005, 0.01, 0.4978, -0.0016, 0.01, 0.4978, 0.0005, 0.01, 0.5, -0.0016, 0.01, 0.5, -0.0038, 0.01, 0.4978, -0.006, 0.01, 0.4978, -0.0038, 0.01, 0.5, -0.006, 0.01, 0.5, 0.0005, 0.0061, 0.502, -0.0016, 0.0061, 0.502, 0.0005, 0.0083, 0.502, -0.0016, 0.0083, 0.502, -0.0038, 0.0061, 0.5019, -0.006, 0.0061, 0.5017, -0.0038, 0.0083, 0.5019, -0.006, 0.0083, 0.5017, 0.0005, 0.01, 0.502, -0.0016, 0.01, 0.502, -0.0038, 0.01, 0.5019, -0.006, 0.01, 0.5017, -0.0087, 0.0061, 0.5009, -0.0103, 0.0061, 0.5002, -0.0087, 0.0083, 0.5009, -0.0103, 0.0083, 0.5002, -0.013, 0.0061, 0.4985, -0.0144, 0.0061, 0.4974, -0.013, 0.0083, 0.4985, -0.0144, 0.0083, 0.4974, -0.0119, 0.0061, 0.4992, -0.0119, 0.0083, 0.4992,
-0.0081, 0.01, 0.4978, -0.0103, 0.01, 0.4978, -0.0085, 0.01, 0.501, -0.0106, 0.01, 0.5001, -0.013, 0.01, 0.4986, -0.0143, 0.01, 0.4975, -0.0118, 0.01, 0.4993, -0.0076, 0.0061, 0.5013, -0.0076, 0.0083, 0.5013, -0.0075, 0.01, 0.5013, -0.0198, 0.0061, 0.4896, -0.0198, 0.0083, 0.4896, -0.0191, 0.0061, 0.4914, -0.0191, 0.0083, 0.4914, -0.0203, 0.0061, 0.4886, -0.0203, 0.0083, 0.4886, -0.0168, 0.01, 0.4892, -0.0199, 0.01, 0.4895, -0.0168, 0.01, 0.4913, -0.019, 0.01, 0.4916, -0.0203, 0.01, 0.4884, -0.0175, 0.0061, 0.494, -0.0182, 0.0061, 0.4929, -0.0175, 0.0083, 0.494, -0.0182, 0.0083, 0.4929, -0.0165, 0.0061, 0.4954, -0.0165, 0.0083, 0.4954, -0.0176, 0.01, 0.4939, -0.0182, 0.01, 0.4928, -0.0165, 0.01, 0.4954,
-0.0379, 0.0061, 0.4892, -0.0379, 0.0083, 0.4892, -0.0379, 0.0061, 0.4913, -0.0379, 0.0083, 0.4913, -0.0379, 0.01, 0.4892, -0.0407, 0.01, 0.4892, -0.0379, 0.01, 0.4913, -0.0407, 0.01, 0.4913, -0.0379, 0.0061, 0.4935, -0.0379, 0.0083, 0.4935, -0.0379, 0.0061, 0.4957, -0.0379, 0.0083, 0.4957, -0.0379, 0.01, 0.4935, -0.0407, 0.01, 0.4935, -0.0379, 0.01, 0.4957, -0.0407, 0.01, 0.4957, -0.0487, 0.0061, 0.4892, -0.0487, 0.0083, 0.4892, -0.0487, 0.0061, 0.4913, -0.0487, 0.0083, 0.4913, -0.0429, 0.01, 0.4892, -0.045, 0.01, 0.4892, -0.0429, 0.01, 0.4913, -0.045, 0.01, 0.4913, -0.0472, 0.01, 0.4892, -0.0487, 0.01, 0.4892, -0.0472, 0.01, 0.4913, -0.0487, 0.01, 0.4913, -0.0487, 0.0061, 0.4935, -0.0487, 0.0083, 0.4935,
-0.0487, 0.0061, 0.4957, -0.0487, 0.0083, 0.4957, -0.0429, 0.01, 0.4935, -0.045, 0.01, 0.4935, -0.0429, 0.01, 0.4957, -0.045, 0.01, 0.4957, -0.0472, 0.01, 0.4935, -0.0487, 0.01, 0.4935, -0.0472, 0.01, 0.4957, -0.0487, 0.01, 0.4957, -0.0379, 0.0061, 0.498, -0.0407, 0.0061, 0.498, -0.0379, 0.0083, 0.498, -0.0407, 0.0083, 0.498, -0.0379, 0.01, 0.498, -0.0407, 0.01, 0.498, -0.0429, 0.0061, 0.498, -0.045, 0.0061, 0.498, -0.0429, 0.0083, 0.498, -0.045, 0.0083, 0.498, -0.0472, 0.0061, 0.498, -0.0487, 0.0061, 0.498, -0.0472, 0.0083, 0.498, -0.0487, 0.0083, 0.498, -0.0429, 0.01, 0.498, -0.045, 0.01, 0.498, -0.0472, 0.01, 0.498, -0.0487, 0.01, 0.498});
    _triz = _trix;

    t.reset();
    t.start();
    _initializeAttributeGrids(isize, jsize, ksize);
    t.stop();

    _logfile.log("Constructing Attribute Grids:      \t", t.getTime(), 4, 1);
}

void FluidSimulation::_initializeParticleSystems() {
    _markerParticles.addAttributeVector3("POSITION");
    _markerParticles.addAttributeVector3("VELOCITY");

    if (_velocityTransferMethod == VelocityTransferMethod::APIC) {
        _markerParticles.addAttributeVector3("AFFINEX");
        _markerParticles.addAttributeVector3("AFFINEY");
        _markerParticles.addAttributeVector3("AFFINEZ");
    }

    if (_isSurfaceAgeAttributeEnabled) {
        _markerParticles.addAttributeFloat("AGE");
    }

    if (_isSurfaceSourceColorAttributeEnabled) {
        _markerParticles.addAttributeVector3("COLOR");
    }

    if (_isSurfaceSourceIDAttributeEnabled) {
        _markerParticles.addAttributeInt("SOURCEID");
    }

    if (_isSurfaceSourceViscosityAttributeEnabled) {
        _markerParticles.addAttributeInt("VISCOSITY");
    }
}

void FluidSimulation::_initializeForceFieldGrid(int isize, int jsize, int ksize, double dx) {
    int reduction = _forceFieldReductionLevel;
    int isizeff = (int)std::ceil((double)isize / (double)reduction);
    int jsizeff = (int)std::ceil((double)jsize / (double)reduction); ;
    int ksizeff = (int)std::ceil((double)ksize / (double)reduction); ;
    double dxff = dx * reduction;
    _forceFieldGrid.initialize(isizeff, jsizeff, ksizeff, dxff);
}

void FluidSimulation::_initializeAttributeGrids(int isize, int jsize, int ksize) {
    if (_isSurfaceVelocityAttributeEnabled && _isSurfaceVelocityAttributeAgainstObstaclesEnabled) {
        _velocityAttributeGrid = MACVelocityField(isize, jsize, ksize, _dx);
        _velocityAttributeValidGrid = ValidVelocityComponentGrid(isize, jsize, ksize);
    }

    if (_isSurfaceVorticityAttributeEnabled) {
        _vorticityAttributeGrid = Array3d<vmath::vec3>(isize, jsize, ksize);
    }

    if (_isSurfaceAgeAttributeEnabled) {
        _ageAttributeGrid = Array3d<float>(isize, jsize, ksize, 0.0f);
        _ageAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);
    }

    if (_isSurfaceSourceViscosityAttributeEnabled) {
        _viscosityAttributeGrid = Array3d<float>(isize, jsize, ksize, 0.0f);
        _viscosityAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);
    }

    if (_isSurfaceSourceColorAttributeEnabled) {
        _colorAttributeGridR = Array3d<float>(isize, jsize, ksize, 0.0f);
        _colorAttributeGridG = Array3d<float>(isize, jsize, ksize, 0.0f);
        _colorAttributeGridB = Array3d<float>(isize, jsize, ksize, 0.0f);
        _colorAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);
    }
}

double FluidSimulation::_getMarkerParticleJitter() {
    double eps = 1e-3;
    return 0.25 * (_markerParticleJitterFactor - eps) * _dx;
}

vmath::vec3 FluidSimulation::_jitterMarkerParticlePosition(vmath::vec3 p, 
                                                           double jitter) {
    p.x += _randomDouble(-jitter, jitter);
    p.y += _randomDouble(-jitter, jitter);
    p.z += _randomDouble(-jitter, jitter);

    return p;
}

void FluidSimulation::_addMarkerParticles(std::vector<MarkerParticle> &particles, MarkerParticleAttributes attributes) {
    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    std::vector<int> *sourceids = nullptr;
    if (_isSurfaceSourceIDAttributeEnabled) {
        _markerParticles.getAttributeValues("SOURCEID", sourceids);
    }

    std::vector<float> *sourceviscosities = nullptr;
    if (_isSurfaceSourceViscosityAttributeEnabled) {
        _markerParticles.getAttributeValues("VISCOSITY", sourceviscosities);
    }

    std::vector<vmath::vec3> *sourcecolors = nullptr;
    if (_isSurfaceSourceColorAttributeEnabled) {
        _markerParticles.getAttributeValues("COLOR", sourcecolors);
    }

    for (size_t i = 0; i < particles.size(); i++) {
        MarkerParticle mp = particles[i];
        GridIndex g = Grid3d::positionToGridIndex(mp.position, _dx);
        if (Grid3d::isGridIndexInRange(g, _isize, _jsize, _ksize)) {
            positions->push_back(mp.position);
            velocities->push_back(mp.velocity);

            if (_isSurfaceSourceIDAttributeEnabled) {
                sourceids->push_back(attributes.sourceID);
            }

            if (_isSurfaceSourceViscosityAttributeEnabled) {
                sourceviscosities->push_back(attributes.sourceViscosity);
            }

            if (_isSurfaceSourceColorAttributeEnabled) {
                sourcecolors->push_back(attributes.sourceColor);
            }
        }
    }

    _markerParticles.update();
}

void FluidSimulation::_initializeParticleRadii() {
    double volume = _dx*_dx*_dx / 8.0;
    double pi = 3.141592653;
    _markerParticleRadius = pow(3*volume / (4*pi), 1.0/3.0);

    _liquidSDFParticleRadius = 0.5 * _liquidSDFParticleScale * _dx * sqrt(3.0); 
}

void FluidSimulation::_initializeRandomGenerator() {
    _randomSeed = std::mt19937(_randomDevice());
    _random = std::uniform_real_distribution<>(0, 1); 
}

void FluidSimulation::_initializeSimulation() {
    _logfile.newline();
    _logfile.log(std::ostringstream().flush() << 
                 "Initializing Simulation:" << std::endl);

    _initializeSimulationGrids(_isize, _jsize, _ksize, _dx);
    _initializeParticleSystems();

    _initializeParticleRadii();
    _initializeRandomGenerator();

    if (_upscalingPreviousCellSize) {
        StopWatch upscaleTimer;
        upscaleTimer.start();
        _upscaleParticleData();
        upscaleTimer.stop();
        _logfile.log("Upscaling Particle Data:     \t", upscaleTimer.getTime(), 4, 1);
    }

    if (_isMarkerParticleLoadPending || _isDiffuseParticleLoadPending) {
        StopWatch loadTimer;
        loadTimer.start();
        _loadParticles();
        loadTimer.stop();
        _logfile.log("Loading Particle Data:       \t", loadTimer.getTime(), 4, 1);
    }

    _isSimulationInitialized = true;
}

void FluidSimulation::_upscaleParticleData() {
    int isize = _upscalingPreviousIsize;
    int jsize = _upscalingPreviousJsize;
    int ksize = _upscalingPreviousKsize;
    double dx = _upscalingPreviousCellSize;
    double particleRadius = 0.5 * _liquidSDFParticleScale * dx * sqrt(3.0);

    // Load particle data into ParticleSystem
    ParticleSystem markerParticles;
    markerParticles.addAttributeVector3("POSITION");
    markerParticles.addAttributeVector3("VELOCITY");

    std::vector<vmath::vec3> *positions, *velocities;
    markerParticles.getAttributeValues("POSITION", positions);
    markerParticles.getAttributeValues("VELOCITY", velocities);

    bool isAgeDataAvailable = false;
    std::vector<float> *ages;
    if (_isSurfaceAgeAttributeEnabled) {
        if (_markerParticleAgeLoadQueue.size() == _markerParticleLoadQueue.size()) {
            isAgeDataAvailable = true;
            for (size_t i = 0; i < _markerParticleLoadQueue.size(); i++) {
                if (_markerParticleAgeLoadQueue[i].particles.size() != _markerParticleLoadQueue[i].particles.size()) {
                    isAgeDataAvailable = false;
                    break;
                }
            }
        }

        if (isAgeDataAvailable) {
            markerParticles.addAttributeFloat("AGE");
            markerParticles.getAttributeValues("AGE", ages);
        }
    }

    bool isViscosityDataAvailable = false;
    std::vector<float> *viscosities;
    if (_isSurfaceSourceViscosityAttributeEnabled) {
        if (_markerParticleViscosityLoadQueue.size() == _markerParticleLoadQueue.size()) {
            isViscosityDataAvailable = true;
            for (size_t i = 0; i < _markerParticleLoadQueue.size(); i++) {
                if (_markerParticleViscosityLoadQueue[i].particles.size() != _markerParticleLoadQueue[i].particles.size()) {
                    isViscosityDataAvailable = false;
                    break;
                }
            }
        }

        if (isViscosityDataAvailable) {
            markerParticles.addAttributeFloat("VISCOSITY");
            markerParticles.getAttributeValues("VISCOSITY", viscosities);
        }
    }

    bool isColorDataAvailable = false;
    std::vector<vmath::vec3> *colors;
    if (_isSurfaceSourceColorAttributeEnabled) {
        if (_markerParticleColorLoadQueue.size() == _markerParticleLoadQueue.size()) {
            isColorDataAvailable = true;
            for (size_t i = 0; i < _markerParticleLoadQueue.size(); i++) {
                if (_markerParticleColorLoadQueue[i].particles.size() != _markerParticleLoadQueue[i].particles.size()) {
                    isColorDataAvailable = false;
                    break;
                }
            }
        }

        if (isColorDataAvailable) {
            markerParticles.addAttributeVector3("COLOR");
            markerParticles.getAttributeValues("COLOR", colors);
        }
    }

    AABB bounds(0.0, 0.0, 0.0, isize * dx, jsize * dx, ksize * dx);
    for (size_t j = 0; j < _markerParticleLoadQueue.size(); j++) {
        for (size_t i = 0; i < _markerParticleLoadQueue[j].particles.size(); i++) {
            MarkerParticle mp = _markerParticleLoadQueue[j].particles[i];
            mp.position = (mp.position - _domainOffset) / _domainScale;
            if (bounds.isPointInside(mp.position)) {
                positions->push_back(mp.position);
                velocities->push_back(mp.velocity);

                if (isAgeDataAvailable) {
                    MarkerParticleAge ma = _markerParticleAgeLoadQueue[j].particles[i];
                    ages->push_back(ma.age);
                }

                if (isViscosityDataAvailable) {
                    MarkerParticleViscosity mvisc = _markerParticleViscosityLoadQueue[j].particles[i];
                    viscosities->push_back(mvisc.viscosity);
                }

                if (isColorDataAvailable) {
                    MarkerParticleColor mc = _markerParticleColorLoadQueue[j].particles[i];
                    colors->push_back(mc.color);
                }
            }
        }
    }

    markerParticles.update();

    if (markerParticles.empty()) {
        return;
    }

    // Compute Velocity Grids
    MACVelocityField vfield(isize, jsize, ksize, dx);
    ValidVelocityComponentGrid validVelocities(isize, jsize, ksize);
    VelocityAdvector velocityAdvector;

    VelocityAdvectorParameters params;
    params.particles = &markerParticles;
    params.vfield = &vfield;
    params.validVelocities = &validVelocities;
    params.particleRadius = particleRadius;

    velocityAdvector.advect(params);
    int extrapolationLayers = (int)std::ceil(std::sqrt(3) * _CFLConditionNumber) + 3;
    vfield.extrapolateVelocityField(validVelocities, extrapolationLayers);

    // Compute Age Grids
    Array3d<float> ageAttributeGrid;
    Array3d<bool> ageAttributeValidGrid;
    if (isAgeDataAvailable) {
        ageAttributeGrid = Array3d<float>(isize, jsize, ksize, 0.0f);
        ageAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);

        markerParticles.getAttributeValues("POSITION", positions);
        markerParticles.getAttributeValues("AGE", ages);

        AttributeTransferParameters<float> params;
        params.positions = positions;
        params.attributes = ages;
        params.attributeGrid = &ageAttributeGrid;
        params.validGrid = &ageAttributeValidGrid;
        params.particleRadius = _ageAttributeRadius * dx;
        params.dx = dx;

        AttributeToGridTransfer<float> attributeTransfer;
        attributeTransfer.transfer(params);

        GridUtils::extrapolateGrid(&ageAttributeGrid, &ageAttributeValidGrid, _CFLConditionNumber);
    }

    // Compute Viscosity Grids
    Array3d<float> viscosityAttributeGrid;
    Array3d<bool> viscosityAttributeValidGrid;
    if (isViscosityDataAvailable) {
        viscosityAttributeGrid = Array3d<float>(isize, jsize, ksize, 0.0f);
        viscosityAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);

        markerParticles.getAttributeValues("POSITION", positions);
        markerParticles.getAttributeValues("VISCOSITY", viscosities);

        AttributeTransferParameters<float> params;
        params.positions = positions;
        params.attributes = viscosities;
        params.attributeGrid = &viscosityAttributeGrid;
        params.validGrid = &viscosityAttributeValidGrid;
        params.particleRadius = _viscosityAttributeRadius * dx;
        params.dx = dx;

        AttributeToGridTransfer<float> attributeTransfer;
        attributeTransfer.transfer(params);

        GridUtils::extrapolateGrid(&viscosityAttributeGrid, &viscosityAttributeValidGrid, _CFLConditionNumber);
    }

    // Compute Color Grids
    Array3d<float> colorAttributeGridR;
    Array3d<float> colorAttributeGridG;
    Array3d<float> colorAttributeGridB;
    Array3d<bool> colorAttributeValidGrid;
    if (isColorDataAvailable) {
        colorAttributeGridR = Array3d<float>(isize, jsize, ksize, 0.0f);
        colorAttributeGridG = Array3d<float>(isize, jsize, ksize, 0.0f);
        colorAttributeGridB = Array3d<float>(isize, jsize, ksize, 0.0f);
        colorAttributeValidGrid = Array3d<bool>(isize, jsize, ksize, false);
        Array3d<vmath::vec3> colorAttributeGrid(isize, jsize, ksize, vmath::vec3());

        markerParticles.getAttributeValues("POSITION", positions);
        markerParticles.getAttributeValues("COLOR", colors);

        AttributeTransferParameters<vmath::vec3> params;
        params.positions = positions;
        params.attributes = colors;
        params.attributeGrid = &colorAttributeGrid;
        params.validGrid = &colorAttributeValidGrid;
        params.particleRadius = _colorAttributeRadius * dx;
        params.dx = dx;

        AttributeToGridTransfer<vmath::vec3> attributeTransfer;
        attributeTransfer.transfer(params);

        for (int k = 0; k < colorAttributeGrid.depth; k++) {
            for (int j = 0; j < colorAttributeGrid.height; j++) {
                for (int i = 0; i < colorAttributeGrid.width; i++) {
                    vmath::vec3 color = colorAttributeGrid(i, j, k);
                    float rval = _clamp(color.x, 0.0f, 1.0f);
                    float gval = _clamp(color.y, 0.0f, 1.0f);
                    float bval = _clamp(color.z, 0.0f, 1.0f);
                    colorAttributeGridR.set(i, j, k, rval);
                    colorAttributeGridG.set(i, j, k, gval);
                    colorAttributeGridB.set(i, j, k, bval);
                }
            }
        }

        GridUtils::extrapolateGrid(&colorAttributeGridR, &colorAttributeValidGrid, _CFLConditionNumber);
        GridUtils::extrapolateGrid(&colorAttributeGridG, &colorAttributeValidGrid, _CFLConditionNumber);
        GridUtils::extrapolateGrid(&colorAttributeGridB, &colorAttributeValidGrid, _CFLConditionNumber);
    }

    // Initialize New Particles
    ParticleMaskGrid maskgrid(_isize, _jsize, _ksize, _dx);
    for (unsigned int i = 0; i < positions->size(); i++) {
        maskgrid.addParticle(positions->at(i));
    }

    ParticleLevelSet liquidSDF(isize, jsize, ksize, dx);
    liquidSDF.calculateSignedDistanceField(markerParticles, particleRadius);

    double q = 0.25 * _dx;
    vmath::vec3 particleOffsets[8] = {
        vmath::vec3(-q, -q, -q),
        vmath::vec3( q, -q, -q),
        vmath::vec3(-q,  q, -q),
        vmath::vec3( q,  q, -q),
        vmath::vec3(-q, -q,  q),
        vmath::vec3( q, -q,  q),
        vmath::vec3(-q,  q,  q),
        vmath::vec3( q,  q,  q)
    };

    double jitter = _getMarkerParticleJitter();
    double currentParticleRadius = 0.5 * _liquidSDFParticleScale * _dx * sqrt(3.0); 
    vmath::vec3 goffset(0.5f * dx, 0.5f * dx, 0.5f * dx);
    MarkerParticleLoadData loadData;
    MarkerParticleAgeLoadData loadAgeData;
    MarkerParticleViscosityLoadData loadViscosityData;
    MarkerParticleColorLoadData loadColorData;
    for (int k = 0; k < _ksize; k++) {
        for (int j = 0; j < _jsize; j++) {
            for (int i = 0; i < _isize; i++) {
                vmath::vec3 c = Grid3d::GridIndexToCellCenter(i, j, k, _dx);
                if (liquidSDF.trilinearInterpolate(c) > 0.0f) {
                    continue;
                }

                for (int oidx = 0; oidx < 8; oidx++) {
                    vmath::vec3 p = c + particleOffsets[oidx];
                    if (maskgrid.isSubCellSet(p)) {
                        continue;
                    }

                    if (_isJitterSurfaceMarkerParticlesEnabled) {
                        p = _jitterMarkerParticlePosition(p, jitter);
                    }

                    if (liquidSDF.trilinearInterpolate(p) < -currentParticleRadius) {
                        if (isAgeDataAvailable) {
                            float age = Interpolation::trilinearInterpolate(p - goffset, dx, ageAttributeGrid);
                            MarkerParticleAge ma(age);
                            loadAgeData.particles.push_back(ma);
                        }

                        if (isViscosityDataAvailable) {
                            float viscosity = Interpolation::trilinearInterpolate(p - goffset, dx, viscosityAttributeGrid);
                            MarkerParticleViscosity mvisc(viscosity);
                            loadViscosityData.particles.push_back(mvisc);
                        }

                        if (isColorDataAvailable) {
                            float r = Interpolation::trilinearInterpolate(p - goffset, dx, colorAttributeGridR);
                            float g = Interpolation::trilinearInterpolate(p - goffset, dx, colorAttributeGridG);
                            float b = Interpolation::trilinearInterpolate(p - goffset, dx, colorAttributeGridB);
                            vmath::vec3 color(r, g, b);
                            MarkerParticleColor mc(color);
                            loadColorData.particles.push_back(mc);
                        }

                        vmath::vec3 v = vfield.evaluateVelocityAtPosition(p);
                        vmath::vec3 newp = p * _domainScale + _domainOffset;
                        MarkerParticle mp(newp, v);
                        loadData.particles.push_back(mp);
                    }
                }
            }
        }
    }

    _markerParticleLoadQueue.push_back(loadData);

    if (isAgeDataAvailable) {
        _markerParticleAgeLoadQueue.push_back(loadAgeData);
    }

    if (isViscosityDataAvailable) {
        _markerParticleViscosityLoadQueue.push_back(loadViscosityData);
    }

    if (isColorDataAvailable) {
        _markerParticleColorLoadQueue.push_back(loadColorData);
    }

    _isMarkerParticleLoadPending = true;
}

void FluidSimulation::_loadMarkerParticles(MarkerParticleLoadData &particleData,
                                           MarkerParticleAffineLoadData &affineData,
                                           MarkerParticleAgeLoadData &ageData,
                                           MarkerParticleColorLoadData &colorData,
                                           MarkerParticleSourceIDLoadData &sourceIDData,
                                           MarkerParticleViscosityLoadData &viscosityData) {

    if (particleData.particles.empty()) {
        return;
    }

    bool loadAffineData = _velocityTransferMethod == VelocityTransferMethod::APIC &&
                          affineData.particles.size() == particleData.particles.size();

    bool loadAgeData = _isSurfaceAgeAttributeEnabled && 
                       ageData.particles.size() == particleData.particles.size();

    bool loadColorData = _isSurfaceSourceColorAttributeEnabled && 
                         colorData.particles.size() == particleData.particles.size();

    bool loadSourceIDData = _isSurfaceSourceIDAttributeEnabled && 
                            sourceIDData.particles.size() == particleData.particles.size();

    bool loadViscosityData = _isSurfaceSourceViscosityAttributeEnabled && 
                             viscosityData.particles.size() == particleData.particles.size();

    _markerParticles.reserve(_markerParticles.size() + particleData.particles.size());

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    std::vector<vmath::vec3> *affineX = nullptr;
    std::vector<vmath::vec3> *affineY = nullptr;
    std::vector<vmath::vec3> *affineZ = nullptr;
    if (loadAffineData) {
        _markerParticles.getAttributeValues("AFFINEX", affineX);
        _markerParticles.getAttributeValues("AFFINEY", affineY);
        _markerParticles.getAttributeValues("AFFINEZ", affineZ);
    }

    std::vector<float> *age = nullptr;
    if (loadAgeData) {
        _markerParticles.getAttributeValues("AGE", age);
    }

    std::vector<vmath::vec3> *color = nullptr;
    if (loadColorData) {
        _markerParticles.getAttributeValues("COLOR", color);
    }

    std::vector<int> *sourceid = nullptr;
    if (loadSourceIDData) {
        _markerParticles.getAttributeValues("SOURCEID", sourceid);
    }

    std::vector<float> *viscosity = nullptr;
    if (loadViscosityData) {
        _markerParticles.getAttributeValues("VISCOSITY", viscosity);
    }

    AABB bounds(0.0, 0.0, 0.0, _isize * _dx, _jsize * _dx, _ksize * _dx);
    for (size_t i = 0; i < particleData.particles.size(); i++) {
        MarkerParticle mp = particleData.particles[i];
        mp.position = (mp.position - _domainOffset) / _domainScale;
        if (bounds.isPointInside(mp.position)) {
            positions->push_back(mp.position);
            velocities->push_back(mp.velocity);

            if (loadAffineData) {
                MarkerParticleAffine ap = affineData.particles[i];
                affineX->push_back(ap.affineX);
                affineY->push_back(ap.affineY);
                affineZ->push_back(ap.affineZ);
            }

            if (loadAgeData) {
                MarkerParticleAge ap = ageData.particles[i];
                age->push_back(ap.age);
            }

            if (loadColorData) {
                MarkerParticleColor c = colorData.particles[i];
                color->push_back(c.color);
            }

            if (loadSourceIDData) {
                MarkerParticleSourceID sid = sourceIDData.particles[i];
                sourceid->push_back(sid.sourceid);
            }

            if (loadViscosityData) {
                MarkerParticleViscosity vd = viscosityData.particles[i];
                viscosity->push_back(vd.viscosity);
            }
        }
    }

    _markerParticles.update();
}

void FluidSimulation::_loadDiffuseParticles(DiffuseParticleLoadData &data) {
    _diffuseMaterial.loadDiffuseParticles(data.particles);
}

void FluidSimulation::_loadParticles() {
    bool isAffineDataAvailable = _markerParticleAffineLoadQueue.size() == _markerParticleLoadQueue.size();
    bool isAgeDataAvailable = _markerParticleAgeLoadQueue.size() == _markerParticleLoadQueue.size();
    bool isColorDataAvailable = _markerParticleColorLoadQueue.size() == _markerParticleLoadQueue.size();
    bool isSourceIDDataAvailable = _markerParticleSourceIDLoadQueue.size() == _markerParticleLoadQueue.size();
    bool isViscosityDataAvailable = _markerParticleViscosityLoadQueue.size() == _markerParticleLoadQueue.size();

    MarkerParticleAffineLoadData emptyAffineData;
    MarkerParticleAgeLoadData emptyAgeData;
    MarkerParticleColorLoadData emptyColorData;
    MarkerParticleSourceIDLoadData emptySourceIDData;
    MarkerParticleViscosityLoadData emptyViscosityData;
    for (size_t i = 0; i < _markerParticleLoadQueue.size(); i++) {
        MarkerParticleAffineLoadData affineData = isAffineDataAvailable ? _markerParticleAffineLoadQueue[i] : emptyAffineData;
        MarkerParticleAgeLoadData ageData = isAgeDataAvailable ? _markerParticleAgeLoadQueue[i] : emptyAgeData;
        MarkerParticleColorLoadData colorData = isColorDataAvailable ? _markerParticleColorLoadQueue[i] : emptyColorData;
        MarkerParticleSourceIDLoadData sourceIDData = isSourceIDDataAvailable ? _markerParticleSourceIDLoadQueue[i] : emptySourceIDData;
        MarkerParticleViscosityLoadData viscosityData = isViscosityDataAvailable ? _markerParticleViscosityLoadQueue[i] : emptyViscosityData;
        _loadMarkerParticles(_markerParticleLoadQueue[i], affineData, ageData, colorData, sourceIDData, viscosityData);
    }
    _markerParticleLoadQueue.clear();
    _markerParticleAffineLoadQueue.clear();
    _markerParticleAgeLoadQueue.clear();
    _markerParticleColorLoadQueue.clear();
    _markerParticleSourceIDLoadQueue.clear();
    _markerParticleViscosityLoadQueue.clear();
    _isMarkerParticleLoadPending = false;
    
    for (size_t i = 0; i < _diffuseParticleLoadQueue.size(); i++) {
        _loadDiffuseParticles(_diffuseParticleLoadQueue[i]);
    }
    _diffuseParticleLoadQueue.clear();
    _isDiffuseParticleLoadPending = false;
}

/********************************************************************************
    #. Update Solid Material
********************************************************************************/
TriangleMesh FluidSimulation::_getTriangleMeshFromAABB(AABB bbox) {
    vmath::vec3 p = bbox.position;
    std::vector<vmath::vec3> verts{
        vmath::vec3(p.x, p.y, p.z),
        vmath::vec3(p.x + bbox.width, p.y, p.z),
        vmath::vec3(p.x + bbox.width, p.y, p.z + bbox.depth),
        vmath::vec3(p.x, p.y, p.z + bbox.depth),
        vmath::vec3(p.x, p.y + bbox.height, p.z),
        vmath::vec3(p.x + bbox.width, p.y + bbox.height, p.z),
        vmath::vec3(p.x + bbox.width, p.y + bbox.height, p.z + bbox.depth),
        vmath::vec3(p.x, p.y + bbox.height, p.z + bbox.depth)
    };

    std::vector<Triangle> tris{
        Triangle(0, 1, 2), Triangle(0, 2, 3), Triangle(4, 7, 6), Triangle(4, 6, 5),
        Triangle(0, 3, 7), Triangle(0, 7, 4), Triangle(1, 5, 6), Triangle(1, 6, 2),
        Triangle(0, 4, 5), Triangle(0, 5, 1), Triangle(3, 2, 6), Triangle(3, 6, 7)
    };

    TriangleMesh m;
    m.vertices = verts;
    m.triangles = tris;

    return m;
}

AABB FluidSimulation::_getBoundaryAABB() {
    double eps = 1e-4;
    AABB domainAABB(0.0, 0.0, 0.0, _isize * _dx, _jsize * _dx, _ksize * _dx);
    domainAABB.expand(-3 * _dx - eps);
    return domainAABB;
}

TriangleMesh FluidSimulation::_getBoundaryTriangleMesh() {
    AABB boundaryAABB = _getBoundaryAABB();
    TriangleMesh boundaryMesh = _getTriangleMeshFromAABB(boundaryAABB);
    return boundaryMesh;
}

void FluidSimulation::_updatePrecomputedSolidLevelSet(double dt, 
                                                      std::vector<MeshObjectStatus> &objectStatus) {

    if (!_isStaticSolidLevelSetPrecomputed) {
        int pi, pj, pk;
        _staticSolidSDF.getGridDimensions(&pi, &pj, &pk);
        if (pi > 0 || pj  > 0 || pk > 0) {
            _staticSolidSDF = MeshLevelSet();
        }
        return;
    }

    if (_isStaticSolidStateChanged(objectStatus)) {
        _isPrecomputedSolidLevelSetUpToDate = false;
    }

    if (_isPrecomputedSolidLevelSetUpToDate) {
        return;
    }

    int pi, pj, pk;
    _staticSolidSDF.getGridDimensions(&pi, &pj, &pk);
    if (pi != _isize || pj != _jsize || pk != _ksize) {
        _staticSolidSDF = MeshLevelSet(_isize, _jsize, _ksize, _dx);
    }

    _addStaticObjectsToSDF(dt, _staticSolidSDF);

    _isPrecomputedSolidLevelSetUpToDate = true;
}

void FluidSimulation::_addAnimatedObjectsToSolidSDF(double dt) {
    std::vector<MeshObject*> inversedObstacles;
    std::vector<MeshObject*> normalObstacles;
    for (size_t i = 0; i < _obstacles.size(); i++) {
        if (_obstacles[i]->isEnabled() && _obstacles[i]->isAnimated()) {
            if (_obstacles[i]->isInversed()) {
                inversedObstacles.push_back(_obstacles[i]);
            } else {
                normalObstacles.push_back(_obstacles[i]);
            }
        }
    }

    float frameTime = (float)(_currentFrameDeltaTimeRemaining + _currentFrameTimeStep);
    float frameProgress = 1.0f - frameTime / (float)_currentFrameDeltaTime;

    if (!_isTempSolidLevelSetEnabled && (!normalObstacles.empty() || !inversedObstacles.empty())) {
        _tempSolidSDF = MeshLevelSet(_isize, _jsize, _ksize, _dx);
    }

    for (size_t i = 0; i < normalObstacles.size(); i++) {
        _tempSolidSDF.reset();
        normalObstacles[i]->getMeshLevelSet(dt, frameProgress, _solidLevelSetExactBand, _tempSolidSDF);
        _solidSDF.calculateUnion(_tempSolidSDF);
    }

    if (!inversedObstacles.empty()) {
        MeshLevelSet tempSolidInversedSDF(_isize, _jsize, _ksize, _dx);
        tempSolidInversedSDF.disableVelocityData();

        for (size_t i = 0; i < inversedObstacles.size(); i++) {
            _tempSolidSDF.reset();
            _tempSolidSDF.disableVelocityData();
            inversedObstacles[i]->getMeshLevelSet(dt, frameProgress, _solidLevelSetExactBand, _tempSolidSDF);
            tempSolidInversedSDF.calculateUnion(_tempSolidSDF);
        }

        tempSolidInversedSDF.enableVelocityData();
        tempSolidInversedSDF.negate();
        _solidSDF.calculateUnion(tempSolidInversedSDF);

        _tempSolidSDF.enableVelocityData();
    }

    if (!_isTempSolidLevelSetEnabled) {
        _tempSolidSDF = MeshLevelSet();
    }
}

void FluidSimulation::_addStaticObjectsToSDF(double dt, MeshLevelSet &sdf){
    TriangleMesh boundaryMesh = _domainMeshObject.getMesh();
    sdf.reset();
    sdf.pushMeshObject(&_domainMeshObject);
    sdf.disableVelocityData();    // Stops velocity data from being calculated 
                                  // twice (once during sdf calculations, and
                                  // once when sdf is negated
    sdf.fastCalculateSignedDistanceField(boundaryMesh, _solidLevelSetExactBand);
    sdf.enableVelocityData();
    sdf.negate();

    std::vector<MeshObject*> inversedObstacles;
    std::vector<MeshObject*> normalObstacles;
    for (size_t i = 0; i < _obstacles.size(); i++) {
        if (_obstacles[i]->isEnabled() && !_obstacles[i]->isAnimated()) {
            if (_obstacles[i]->isInversed()) {
                inversedObstacles.push_back(_obstacles[i]);
            } else {
                normalObstacles.push_back(_obstacles[i]);
            }
        }
    }

    float frameTime = (float)(_currentFrameDeltaTimeRemaining + _currentFrameTimeStep);
    float frameProgress = 1.0f - frameTime / (float)_currentFrameDeltaTime;

    if (!_isTempSolidLevelSetEnabled && (!normalObstacles.empty() || !inversedObstacles.empty())) {
        _tempSolidSDF = MeshLevelSet(_isize, _jsize, _ksize, _dx);
    }

    for (size_t i = 0; i < normalObstacles.size(); i++) {
        _tempSolidSDF.reset();
        normalObstacles[i]->getMeshLevelSet(dt, frameProgress, _solidLevelSetExactBand, _tempSolidSDF);
        sdf.calculateUnion(_tempSolidSDF);
    }

    if (!inversedObstacles.empty()) {
        MeshLevelSet tempSolidInversedSDF(_isize, _jsize, _ksize, _dx);
        tempSolidInversedSDF.disableVelocityData();

        for (size_t i = 0; i < inversedObstacles.size(); i++) {
            _tempSolidSDF.reset();
            _tempSolidSDF.disableVelocityData();
            inversedObstacles[i]->getMeshLevelSet(dt, frameProgress, _solidLevelSetExactBand, _tempSolidSDF);
            tempSolidInversedSDF.calculateUnion(_tempSolidSDF);
        }

        tempSolidInversedSDF.enableVelocityData();
        tempSolidInversedSDF.negate();
        sdf.calculateUnion(tempSolidInversedSDF);

        _tempSolidSDF.enableVelocityData();
    }
}

void FluidSimulation::_addStaticObjectsToSolidSDF(double dt, std::vector<MeshObjectStatus> &objectStatus) {
    StopWatch t;
    t.start();

    _updatePrecomputedSolidLevelSet(dt, objectStatus);

    if (_isStaticSolidLevelSetPrecomputed) {
        if (_isPrecomputedSolidLevelSetUpToDate) {
            _solidSDF.calculateUnion(_staticSolidSDF);
        }

        return;
    }

    _addStaticObjectsToSDF(dt, _solidSDF);
}

bool FluidSimulation::_isSolidStateChanged(std::vector<MeshObjectStatus> &objectStatus) {
    for (size_t i = 0; i < objectStatus.size(); i++) {
        MeshObjectStatus s = objectStatus[i];
        if (s.isStateChanged || (s.isEnabled && s.isAnimated && s.isMeshChanged)) {
            return true;
        }
    }
    return false;
}

bool FluidSimulation::_isStaticSolidStateChanged(std::vector<MeshObjectStatus> &objectStatus) {
    for (size_t i = 0; i < objectStatus.size(); i++) {
        MeshObjectStatus s = objectStatus[i];
        if (!s.isAnimated && s.isStateChanged) {
            return true;
        }
    }
    return false;
}

std::vector<MeshObjectStatus> FluidSimulation::_getSolidObjectStatus() {
    std::vector<MeshObjectStatus> objectData;
    for (size_t i = 0; i < _obstacles.size(); i++) {
        objectData.push_back(_obstacles[i]->getStatus());
        _obstacles[i]->clearObjectStatus();
    }
    return objectData;
}

void FluidSimulation::_updateSolidLevelSet(double dt) {
    std::vector<MeshObjectStatus> objectStatus = _getSolidObjectStatus();
    if (_isSolidStateChanged(objectStatus)) {
        _isSolidLevelSetUpToDate = false;
    }

    if (_isSolidLevelSetUpToDate) {
        return;
    }

    if (_markerParticles.empty() && 
            _addedFluidMeshObjectQueue.empty() && 
            _meshFluidSources.empty() && 
            !_isInternalObstacleMeshOutputEnabled) {
        return;
    }

    _solidSDF.reset();

    StopWatch t;
    t.start();

    int pi, pj, pk;
    _tempSolidSDF.getGridDimensions(&pi, &pj, &pk);
    if (_isTempSolidLevelSetEnabled) {
        if (pi != _isize || pj != _jsize || pk != _ksize) {
            _tempSolidSDF = MeshLevelSet(_isize, _jsize, _ksize, _dx);
        }
    } else {
        if (pi > 0 || pj  > 0 || pk > 0) {
            _tempSolidSDF = MeshLevelSet();
        }
    }

    _addStaticObjectsToSolidSDF(dt, objectStatus);
    _addAnimatedObjectsToSolidSDF(dt);

    _solidSDF.normalizeVelocityGrid();

    _resolveSolidLevelSetUpdateCollisions();

    _isSolidLevelSetUpToDate = true;
    _isWeightGridUpToDate = false;

}

void FluidSimulation::_updateObstacles(double) {
    // Currently, nothing needs to be updated
    /*
    for (size_t i = 0; i < _obstacles.size(); i++) {
        
    }
    */
}

void FluidSimulation::_initializeNearSolidGridThread(int startidx, int endidx) {
    float maxd = _solidLevelSetExactBand * _dx;
    int gridfactor = _nearSolidGridCellSizeFactor;
    for (int idx = startidx; idx < endidx; idx++) {
        GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize);
        if (std::abs(_solidSDF(g)) < maxd) {
            _nearSolidGrid.set(g.i / gridfactor, g.j / gridfactor, g.k / gridfactor, true);
        }
    }
}

void FluidSimulation::_updateNearSolidGrid() {
    _nearSolidGridCellSize = _nearSolidGridCellSizeFactor * _dx;
    int gridi = (int)std::ceil(getSimulationWidth() / _nearSolidGridCellSize);
    int gridj = (int)std::ceil(getSimulationHeight() / _nearSolidGridCellSize);
    int gridk = (int)std::ceil(getSimulationDepth() / _nearSolidGridCellSize);

    if (_nearSolidGrid.width != gridi || 
            _nearSolidGrid.height != gridj ||
            _nearSolidGrid.depth != gridk) {
        _nearSolidGrid = Array3d<bool>(gridi, gridj, gridk, false);
    } else {
        _nearSolidGrid.fill(false);
    }

    int numCPU = ThreadUtils::getMaxThreadCount();
    int gridsize = _isize * _jsize * _ksize;
    int numthreads = (int)fmin(numCPU, gridsize);
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, gridsize, numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_initializeNearSolidGridThread, this,
                                 intervals[i], intervals[i + 1]);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }

    int numlayers = (int)std::ceil((float)_CFLConditionNumber / (float)_nearSolidGridCellSizeFactor);
    for (int i = 0; i < numlayers; i++) {
        GridUtils::featherGrid6(&_nearSolidGrid, ThreadUtils::getMaxThreadCount());
    }
}

void FluidSimulation::_resolveSolidLevelSetUpdateCollisionsThread(int startidx, int endidx) {
    /*
    AABB boundary = _getBoundaryAABB();
    float maxResolvedDistance = _CFLConditionNumber * _dx;
    for (int i = startidx; i < endidx; i++) {
        vmath::vec3 p = _markerParticles[i].position;

        GridIndex n = Grid3d::positionToGridIndex(p, _nearSolidGridCellSize);
        if (!_nearSolidGrid(n)) {
            continue;
        }

        float phi = _solidSDF.trilinearInterpolate(p);
        if (phi >= 0.0f) {
            continue;
        }

        float eps = 1e-5;
        vmath::vec3 grad = _solidSDF.trilinearInterpolateGradient(p);
        if (vmath::length(grad) < eps) {
            continue;
        }

        grad = vmath::normalize(grad);
        vmath::vec3 resolvedPosition = p - (phi - _solidBufferWidth * _dx) * grad;
        float resolvedPhi = _solidSDF.trilinearInterpolate(resolvedPosition);
        float resolvedDistance = vmath::length(resolvedPosition - p);
        if (resolvedPhi < 0 || resolvedDistance > maxResolvedDistance) {
            continue;
        }

        if (!boundary.isPointInside(resolvedPosition)) {
            continue;
        }

        _markerParticles[i].position = resolvedPosition;
    }
    */
}

void FluidSimulation::_resolveSolidLevelSetUpdateCollisions() {
    _nearSolidGridCellSize = _nearSolidGridCellSizeFactor * _dx;
    int gridi = (int)std::ceil(getSimulationWidth() / _nearSolidGridCellSize);
    int gridj = (int)std::ceil(getSimulationHeight() / _nearSolidGridCellSize);
    int gridk = (int)std::ceil(getSimulationDepth() / _nearSolidGridCellSize);

    if (_nearSolidGrid.width != gridi || 
            _nearSolidGrid.height != gridj ||
            _nearSolidGrid.depth != gridk) {
        _nearSolidGrid = Array3d<bool>(gridi, gridj, gridk, false);
    } else {
        _nearSolidGrid.fill(false);
    }
    
    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, _markerParticles.size());
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, _markerParticles.size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_resolveSolidLevelSetUpdateCollisionsThread, this,
                                 intervals[i], intervals[i + 1]);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }
}

void FluidSimulation::_updateObstacleObjects(double) {
    _logfile.logString(_logfile.getTime() + " BEGIN       Update Obstacle Objects");

    StopWatch t;
    t.start();
    _updateObstacles(_currentFrameDeltaTime);
    _updateSolidLevelSet(_currentFrameDeltaTime);
    _updateNearSolidGrid();
    _updateMeshingVolumeSDF();
    t.stop();

    _timingData.updateObstacleObjects += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Obstacle Objects");
}

void FluidSimulation::_launchUpdateObstacleObjectsThread(double dt) {
    _updateObstacleObjectsThread = std::thread(&FluidSimulation::_updateObstacleObjects, 
                                               this, dt);
}
    
void FluidSimulation::_joinUpdateObstacleObjectsThread() {
    _updateObstacleObjectsThread.join();
}

/********************************************************************************
    #. Update Fluid Material
********************************************************************************/

void FluidSimulation::_updateLiquidLevelSet() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Update Liquid Level Set");

    StopWatch t;
    t.start();

    double radius = _liquidSDFParticleRadius;
    if (_isSurfaceTensionEnabled && _isSmoothSurfaceTensionKernelEnabled) {
        radius = _liquidSDFSurfaceTensionParticleScale * _liquidSDFParticleRadius;
    }

    _liquidSDF.calculateSignedDistanceField(_markerParticles, radius);

    t.stop();

    _timingData.updateLiquidLevelSet += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Liquid Level Set");
}

void FluidSimulation::_launchUpdateLiquidLevelSetThread() {
    _updateLiquidLevelSetThread = std::thread(&FluidSimulation::_updateLiquidLevelSet, this);
}

void FluidSimulation::_joinUpdateLiquidLevelSetThread() {
    _updateLiquidLevelSetThread.join();
    _liquidSDF.postProcessSignedDistanceField(_solidSDF);
}

/********************************************************************************
    #.  Advect Velocity Field
********************************************************************************/

void FluidSimulation::_advectVelocityField() {
_logfile.logString(_logfile.getTime() + " BEGIN       Advect Velocity Field");

    StopWatch t;
    t.start();

    _validVelocities.reset();
    _MACVelocity.clear();
    if (!_markerParticles.empty()) {
        double radius = _liquidSDFParticleRadius;
        if (_isSurfaceTensionEnabled && _isSmoothSurfaceTensionKernelEnabled) {
            radius = _liquidSDFSurfaceTensionParticleScale * _liquidSDFParticleRadius;
        }

        VelocityAdvectorParameters params;
        params.particles = &_markerParticles;
        params.vfield = &_MACVelocity;
        params.validVelocities = &_validVelocities;
        params.particleRadius = radius;

        if (_velocityTransferMethod == VelocityTransferMethod::FLIP) {
            params.velocityTransferMethod = VelocityAdvectorTransferMethod::FLIP;
        } else if (_velocityTransferMethod == VelocityTransferMethod::APIC) {
            params.velocityTransferMethod = VelocityAdvectorTransferMethod::APIC;
        }
        
        _velocityAdvector.advect(params);

        _extrapolateFluidVelocities(_MACVelocity, _validVelocities);
    }

    t.stop();
    _timingData.advectVelocityField += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Advect Velocity Field");
}

void FluidSimulation::_launchAdvectVelocityFieldThread() {
    _advectVelocityFieldThread = std::thread(&FluidSimulation::_advectVelocityField, this);
}

void FluidSimulation::_joinAdvectVelocityFieldThread() {
    _advectVelocityFieldThread.join();
}

void FluidSimulation::_saveVelocityField() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Save Velocity Field");

    StopWatch t;
    t.start();
    _savedVelocityField = _MACVelocity;
    t.stop();
    _timingData.saveVelocityField += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Save Velocity Field");
}

void FluidSimulation::_deleteSavedVelocityField() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Delete Saved Velocity Field");

    StopWatch t;
    t.start();
    _savedVelocityField = MACVelocityField();
    t.stop();
    _timingData.deleteSavedVelocityField += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Delete Saved Velocity Field");
}


/********************************************************************************
    #.  Calculate Fluid Curvature
********************************************************************************/

void FluidSimulation::_calculateFluidCurvatureGridThread() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Calculate Surface Curvature");

    StopWatch t;
    t.start();

    if (_fluidCurvatureGrid.width == _isize && 
            _fluidCurvatureGrid.height == _jsize && 
            _fluidCurvatureGrid.depth == _ksize) {
        _fluidSurfaceLevelSet.fill(0.0f);
        _fluidCurvatureGrid.fill(0.0f);
    } else {
        _fluidSurfaceLevelSet = Array3d<float>(_isize, _jsize, _ksize, 0.0f);
        _fluidCurvatureGrid = Array3d<float>(_isize, _jsize, _ksize, 0.0f);
    }

    _liquidSDF.calculateCurvatureGrid(_fluidSurfaceLevelSet, _fluidCurvatureGrid);

    t.stop();
    _timingData.calculateFluidCurvatureGrid += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Calculate Surface Curvature");
}

void FluidSimulation::_launchCalculateFluidCurvatureGridThread() {
    if (!_isSurfaceTensionEnabled && !_isSheetSeedingEnabled && !_isDiffuseMaterialOutputEnabled) {
        return;
    }

    _fluidCurvatureThread = std::thread(&FluidSimulation::_calculateFluidCurvatureGridThread, 
                                        this);
    _isCalculateFluidCurvatureGridThreadRunning = true;
}

void FluidSimulation::_joinCalculateFluidCurvatureGridThread() {
    if (!_isCalculateFluidCurvatureGridThreadRunning) {
        return;
    }

    _fluidCurvatureThread.join();
    _isCalculateFluidCurvatureGridThreadRunning = false;
}

/********************************************************************************
    #. Apply Body Forces
********************************************************************************/

vmath::vec3 FluidSimulation::_getConstantBodyForce() {
    vmath::vec3 bf;
    for (unsigned int i = 0; i < _constantBodyForces.size(); i++) {
        bf += _constantBodyForces[i];
    }

    return bf;
}

void FluidSimulation::_getInflowConstrainedVelocityComponents(ValidVelocityComponentGrid &ex) {
    for (size_t sidx = 0; sidx < _meshFluidSources.size(); sidx++) {
        MeshFluidSource *inflow = _meshFluidSources[sidx];
        if (!inflow->isEnabled() || !inflow->isInflow() || 
                !inflow->isConstrainedFluidVelocityEnabled()) {
            continue;
        }

        float frameProgress = _getFrameInterpolation();
        int numSubsteps = inflow->getSubstepEmissions();

        if (numSubsteps == 0) {
            numSubsteps = 1;
        }

        float substepFactor = (_currentFrameTimeStep / _currentFrameDeltaTime) / (float)numSubsteps;

        for (int subidx = 0; subidx < numSubsteps; subidx++) {
            float frameInterpolation = frameProgress + (float)subidx * substepFactor;
            inflow->setFrame(_currentFrame, frameInterpolation);
            inflow->update(_currentFrameDeltaTime);
            
            for (int k = 0; k < _ksize; k++) {
                for (int j = 0; j < _jsize; j++) {
                    for (int i = 0; i < _isize + 1; i++) {
                        if (!_validVelocities.validU(i, j, k)) {
                            continue;
                        }
                        vmath::vec3 p = Grid3d::FaceIndexToPositionU(i, j, k, _dx);
                        if (inflow->trilinearInterpolate(p) < 0.0f) {
                            ex.validU.set(i, j, k, true);
                        }
                    }
                }
            }

            for (int k = 0; k < _ksize; k++) {
                for (int j = 0; j < _jsize + 1; j++) {
                    for (int i = 0; i < _isize; i++) {
                        if (!_validVelocities.validV(i, j, k)) {
                            continue;
                        }
                        vmath::vec3 p = Grid3d::FaceIndexToPositionV(i, j, k, _dx);
                        if (inflow->trilinearInterpolate(p) < 0.0f) {
                            ex.validV.set(i, j, k, true);
                        }
                    }
                }
            }

            for (int k = 0; k < _ksize + 1; k++) {
                for (int j = 0; j < _jsize; j++) {
                    for (int i = 0; i < _isize; i++) {
                        if (!_validVelocities.validW(i, j, k)) {
                            continue;
                        }
                        vmath::vec3 p = Grid3d::FaceIndexToPositionW(i, j, k, _dx);
                        if (inflow->trilinearInterpolate(p) < 0.0f) {
                            ex.validW.set(i, j, k, true);
                        }
                    }
                }
            }
        }
    }
}

void FluidSimulation::_updateForceFieldGrid(double dt) {
    if (!_isAdaptiveForceFieldTimeSteppingEnabled && _currentFrameTimeStepNumber != 0) {
        return;
    }

    float frameProgress = _getFrameInterpolation();
    _forceFieldGrid.setGravityVector(_getConstantBodyForce());
    _forceFieldGrid.update(dt, frameProgress);
}

void FluidSimulation::_applyConstantBodyForces(ValidVelocityComponentGrid &ex, double dt) {
    vmath::vec3 bodyForce = _getConstantBodyForce();
    float eps = 1e-6;

    if (fabs(bodyForce.x) > eps) {
        for (int k = 0; k < _ksize; k++) {
            for (int j = 0; j < _jsize; j++) {
                for (int i = 0; i < _isize + 1; i++) {
                    if (!ex.validU(i, j, k)) {
                        _MACVelocity.addU(i, j, k, bodyForce.x * dt);
                    }
                }
            }
        }
    }

    if (fabs(bodyForce.y) > eps) {
        for (int k = 0; k < _ksize; k++) {
            for (int j = 0; j < _jsize + 1; j++) {
                for (int i = 0; i < _isize; i++) {
                    if (!ex.validV(i, j, k)) {
                        _MACVelocity.addV(i, j, k, bodyForce.y * dt);
                    }
                }
            }
        }
    }

    if (fabs(bodyForce.z) > eps) {
        for (int k = 0; k < _ksize + 1; k++) {
            for (int j = 0; j < _jsize; j++) {
                for (int i = 0; i < _isize; i++) {
                    if (!ex.validW(i, j, k)) {
                        _MACVelocity.addW(i, j, k, bodyForce.z * dt);
                    }
                }
            }
        }
    }
}

void FluidSimulation::_applyForceFieldGridForces(ValidVelocityComponentGrid &ex, double dt) {
    int U = 0; int V = 1; int W = 2;
    _applyForceFieldGridForcesMT(ex, dt, U);
    _applyForceFieldGridForcesMT(ex, dt, V);
    _applyForceFieldGridForcesMT(ex, dt, W);

}

void FluidSimulation::_applyForceFieldGridForcesMT(ValidVelocityComponentGrid &ex, double dt, int dir) {

    int U = 0; int V = 1; int W = 2;

    int gridsize = 0;
    if (dir == U) {
        gridsize = (_isize + 1) * _jsize * _ksize;
    } else if (dir == V) {
        gridsize = _isize * (_jsize + 1) * _ksize;
    } else if (dir == W) {
        gridsize = _isize * _jsize * (_ksize + 1);
    }

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, gridsize);
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, gridsize, numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_applyForceFieldGridForcesThread, this,
                                 intervals[i], intervals[i + 1], &ex, dt, dir);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }
}

void FluidSimulation::_applyForceFieldGridForcesThread(int startidx, int endidx, 
                                                       ValidVelocityComponentGrid *ex, double dt, int dir) {
    int U = 0; int V = 1; int W = 2;

    if (dir == U) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize + 1, _jsize);
            if (!ex->validU(g)) {
                vmath::vec3 p = Grid3d::FaceIndexToPositionU(g, _dx);
                float xvel = _forceFieldGrid.evaluateForceAtPositionU(p);
                _MACVelocity.addU(g, xvel * dt);
            }
        }

    } else if (dir == V) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize + 1);
            if (!ex->validV(g)) {
                vmath::vec3 p = Grid3d::FaceIndexToPositionV(g, _dx);
                float yvel = _forceFieldGrid.evaluateForceAtPositionV(p);
                _MACVelocity.addV(g, yvel * dt);
            }
        }

    } else if (dir == W) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize);
            if (!ex->validW(g)) {
                vmath::vec3 p = Grid3d::FaceIndexToPositionW(g, _dx);
                float zvel = _forceFieldGrid.evaluateForceAtPositionW(p);
                _MACVelocity.addW(g, zvel * dt);
            }
        }

    }
}

void FluidSimulation::_applyBodyForcesToVelocityField(double dt) {
    _logfile.logString(_logfile.getTime() + " BEGIN       Apply Force Fields");

    StopWatch t;
    t.start();

    ValidVelocityComponentGrid ex(_isize, _jsize, _ksize);
    _getInflowConstrainedVelocityComponents(ex);

    if (_isForceFieldsEnabled) {
        _updateForceFieldGrid(dt);
        _applyForceFieldGridForces(ex, dt);
    } else {
        _applyConstantBodyForces(ex, dt);
    }

    t.stop();
    _timingData.applyBodyForcesToVelocityField += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Apply Force Fields");
}

/********************************************************************************
    #. Viscosity Solve
********************************************************************************/

void FluidSimulation::_applyViscosityToVelocityField(double dt) {
    _viscositySolverStatus = "";

    if (!_isViscosityEnabled || _markerParticles.empty()) {
        return;
    }

    bool isVariableViscosityEnabled = _isSurfaceSourceViscosityAttributeEnabled;
    if (isVariableViscosityEnabled) {
        // Override constant viscosity values
        _viscosity.fill(0.0f);
        Array3d<bool> viscosityValidGrid(_viscosity.width, _viscosity.height, _viscosity.depth, false);

        std::vector<vmath::vec3> *positions;
        std::vector<float> *viscosities;
        _markerParticles.getAttributeValues("POSITION", positions);
        _markerParticles.getAttributeValues("VISCOSITY", viscosities);
        float radius = _viscositySolverAttributeRadius * _dx;

        AttributeTransferParameters<float> atparams;
        atparams.positions = positions;
        atparams.attributes = viscosities;
        atparams.attributeGrid = &_viscosity;
        atparams.validGrid = &viscosityValidGrid;
        atparams.particleRadius = radius;
        atparams.dx = _dx;

        AttributeToGridTransfer<float> attributeTransfer;
        attributeTransfer.transfer(atparams);
        GridUtils::extrapolateGrid(&_viscosity, &viscosityValidGrid, _CFLConditionNumber);
    }

    if (_viscosity.isZero()) {
        return;
    }

    _logfile.logString(_logfile.getTime() + " BEGIN       Apply Viscosity");

    StopWatch t;
    t.start();

    _constrainVelocityField(_MACVelocity);

    ViscositySolverParameters params;
    params.cellwidth = _dx;
    params.deltaTime = dt;
    params.velocityField = &_MACVelocity;
    params.liquidSDF = &_liquidSDF;
    params.solidSDF = &_solidSDF;
    params.viscosity = &_viscosity;
    params.errorTolerance = _viscositySolverErrorTolerance;
    params.maxIterations = _maxViscositySolveIterations;

    _viscositySolver = ViscositySolver();
    bool success = _viscositySolver.applyViscosityToVelocityField(params);
    _viscositySolverStatus = _viscositySolver.getSolverStatus();

    if (_currentFrameTimeStepNumber == 0) {
        _viscositySolverSuccess = success;
        _viscositySolverIterations = _viscositySolver.getIterations();
        _viscositySolverError = _viscositySolver.getError();
    } else {
        // Solver status should capture the first failure state
        // Otherwise, it should capture the substep with max iterations
        int numIterations = _viscositySolver.getIterations();
        float error = _viscositySolver.getError();
        if (_viscositySolverSuccess && (!success || (numIterations > _pressureSolverIterations))) {
            _viscositySolverSuccess = success;
            _viscositySolverIterations = numIterations;
            _viscositySolverError = error;
        }
    }

    t.stop();
    _timingData.applyViscosityToVelocityField += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Apply Viscosity");
}

/********************************************************************************
    #. Pressure Solve
********************************************************************************/

void FluidSimulation::_updateWeightGrid() {
    if (_isWeightGridUpToDate) {
        return;
    }

    int U = 0; int V = 1; int W = 2; int CENTER = 3;
    
    _updateWeightGridMT(U);
    _updateWeightGridMT(V);
    _updateWeightGridMT(W);
    _updateWeightGridMT(CENTER);

    _isWeightGridUpToDate = true;
}

void FluidSimulation::_updateWeightGridMT(int dir) {

    int U = 0; int V = 1; int W = 2; int CENTER = 3;

    int gridsize = 0;
    if (dir == U) {
        gridsize = (_isize + 1) * _jsize * _ksize;
    } else if (dir == V) {
        gridsize = _isize * (_jsize + 1) * _ksize;
    } else if (dir == W) {
        gridsize = _isize * _jsize * (_ksize + 1);
    } else if (dir == CENTER) {
        gridsize = _isize * _jsize * _ksize;
    }

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, gridsize);
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, gridsize, numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_updateWeightGridThread, this,
                                 intervals[i], intervals[i + 1], dir);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }
}

void FluidSimulation::_updateWeightGridThread(int startidx, int endidx, int dir) {
    int U = 0; int V = 1; int W = 2; int CENTER = 3;

    if (dir == U) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize + 1, _jsize);
            float weight = 1.0f - _solidSDF.getFaceWeightU(g);
            weight = _clamp(weight, 0.0f, 1.0f);
            _weightGrid.U.set(g, weight);
        }

    } else if (dir == V) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize + 1);
            float weight = 1.0f - _solidSDF.getFaceWeightV(g);
            weight = _clamp(weight, 0.0f, 1.0f);
            _weightGrid.V.set(g, weight);
        }

    } else if (dir == W) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize);
            float weight = 1.0f - _solidSDF.getFaceWeightW(g);
            weight = _clamp(weight, 0.0f, 1.0f);
            _weightGrid.W.set(g, weight);
        }

    } else if (dir == CENTER) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize);
            float weight = 1.0f - _solidSDF.getCellWeight(g);
            weight = _clamp(weight, 0.0f, 1.0f);
            _weightGrid.center.set(g, weight);
        }

    }
}

void FluidSimulation::_pressureSolve(double dt) {
    _logfile.logString(_logfile.getTime() + " BEGIN       Solve Pressure System");

    StopWatch t;
    t.start();

    _updateWeightGrid();

    PressureSolverParameters params;
    params.cellwidth = _dx;
    params.deltaTime = dt;
    params.tolerance = _pressureSolveTolerance;
    params.acceptableTolerance = _pressureSolveAcceptableTolerance;
    params.maxIterations = _maxPressureSolveIterations;

    params.velocityField = &_MACVelocity;
    params.validVelocities = &_validVelocities;
    params.liquidSDF = &_liquidSDF;
    params.solidSDF = &_solidSDF;
    params.weightGrid = &_weightGrid;

    params.isSurfaceTensionEnabled = _isSurfaceTensionEnabled;
    if (_isSurfaceTensionEnabled) {
        params.surfaceTensionConstant = _surfaceTensionConstant;
        params.curvatureGrid = &_fluidCurvatureGrid;
    }

    PressureSolver psolver;
    bool success = psolver.solve(params);
    _pressureSolverStatus = psolver.getSolverStatus();

    if (_currentFrameTimeStepNumber == 0) {
        _pressureSolverSuccess = success;
        _pressureSolverIterations = psolver.getIterations();
        _pressureSolverError = psolver.getError();
    } else {
        // Solver status should capture the first failure state
        // Otherwise, it should capture the substep with max iterations
        int numIterations = psolver.getIterations();
        float error = psolver.getError();
        if (_pressureSolverSuccess && (!success || (numIterations > _pressureSolverIterations))) {
            _pressureSolverSuccess = success;
            _pressureSolverIterations = numIterations;
            _pressureSolverError = error;
        }
    }

    _extrapolateFluidVelocities(_MACVelocity, _validVelocities);
    
    t.stop();
    _timingData.pressureSolve += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Solve Pressure System");
}

/********************************************************************************
    #. Extrapolate Velocity Field
********************************************************************************/

void FluidSimulation::_extrapolateFluidVelocities(MACVelocityField &MACGrid, 
                                                  ValidVelocityComponentGrid &validVelocities) {
    int numLayers = (int)std::ceil(std::sqrt(3) * _CFLConditionNumber) + 3;
    MACGrid.extrapolateVelocityField(validVelocities, numLayers);
}

/********************************************************************************
    #. Constrain Velocity Field
********************************************************************************/

float FluidSimulation::_getFaceFrictionU(GridIndex g) {
    int i = g.i;
    int j = g.j;
    int k = g.k;

    float friction = 0.0f;
    MeshObject *m = _solidSDF.getClosestMeshObject(i, j, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i, j + 1, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i, j, k + 1);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i, j + 1, k + 1);
    friction += m == nullptr ? 0.0f : m->getFriction();

    return 0.25f * friction;
}

float FluidSimulation::_getFaceFrictionV(GridIndex g) {
    int i = g.i;
    int j = g.j;
    int k = g.k;

    float friction = 0.0f;
    MeshObject *m = _solidSDF.getClosestMeshObject(i, j, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i + 1, j, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i, j, k + 1);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i + 1, j, k + 1);
    friction += m == nullptr ? 0.0f : m->getFriction();

    return 0.25f * friction;
}

float FluidSimulation::_getFaceFrictionW(GridIndex g) {
    int i = g.i;
    int j = g.j;
    int k = g.k;

    float friction = 0.0f;
    MeshObject *m = _solidSDF.getClosestMeshObject(i, j, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i + 1, j, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i, j + 1, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    m = _solidSDF.getClosestMeshObject(i + 1, j + 1, k);
    friction += m == nullptr ? 0.0f : m->getFriction();

    return 0.25f * friction;
}

void FluidSimulation::_constrainVelocityField(MACVelocityField &MACGrid) {
    _updateWeightGrid();

    int U = 0; int V = 1; int W = 2;
    _constrainVelocityFieldMT(MACGrid, U);
    _constrainVelocityFieldMT(MACGrid, V);
    _constrainVelocityFieldMT(MACGrid, W);
}

void FluidSimulation::_constrainVelocityFieldMT(MACVelocityField &MACGrid, int dir) {

    int U = 0; int V = 1; int W = 2;

    int gridsize = 0;
    if (dir == U) {
        gridsize = (_isize + 1) * _jsize * _ksize;
    } else if (dir == V) {
        gridsize = _isize * (_jsize + 1) * _ksize;
    } else if (dir == W) {
        gridsize = _isize * _jsize * (_ksize + 1);
    }

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, gridsize);
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, gridsize, numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_constrainVelocityFieldThread, this,
                                 intervals[i], intervals[i + 1], &MACGrid, dir);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }
}

void FluidSimulation::_constrainVelocityFieldThread(int startidx, int endidx, 
                                                    MACVelocityField *vfield, int dir) {

    int U = 0; int V = 1; int W = 2;

    if (dir == U) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize + 1, _jsize);
            if(_weightGrid.U(g) == 0) {
                vfield->setU(g, _solidSDF.getFaceVelocityU(g));
            } else if (_weightGrid.U(g) < 1.0f) {
                float f = _getFaceFrictionU(g);
                float uface = _solidSDF.getFaceVelocityU(g);
                float umac = vfield->U(g);
                float uf = f * uface + (1.0f - f) * umac;
                vfield->setU(g, uf);
            }
        }

    } else if (dir == V) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize + 1);
            if(_weightGrid.V(g) == 0) {
                vfield->setV(g, _solidSDF.getFaceVelocityV(g));
            } else if (_weightGrid.V(g) < 1.0f) {
                float f = _getFaceFrictionV(g);
                float vface = _solidSDF.getFaceVelocityV(g);
                float vmac = vfield->V(g);
                float vf = f * vface + (1.0f - f) * vmac;
                vfield->setV(g, vf);
            }
        }

    } else if (dir == W) {

        for (int idx = startidx; idx < endidx; idx++) {
            GridIndex g = Grid3d::getUnflattenedIndex(idx, _isize, _jsize);
            if(_weightGrid.W(g) == 0) {
                vfield->setW(g, _solidSDF.getFaceVelocityW(g));
            } else if (_weightGrid.W(g) < 1.0f) {
                float f = _getFaceFrictionW(g);
                float wface = _solidSDF.getFaceVelocityW(g);
                float wmac = vfield->W(g);
                float wf = f * wface + (1.0f - f) * wmac;
                vfield->setW(g, wf);
            }
        }

    }
}

void FluidSimulation::_constrainVelocityFields() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Constrain Velocity Field");

    StopWatch t;
    t.start();
    _constrainVelocityField(_savedVelocityField);
    _constrainVelocityField(_MACVelocity);
    t.stop();
    _timingData.constrainVelocityFields += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Constrain Velocity Field");
}

/********************************************************************************
    #. Update Diffuse Particle Simulation
********************************************************************************/

void FluidSimulation::_updateDiffuseMaterial(double dt) {
    if (!_isDiffuseMaterialOutputEnabled) {
        return;
    }

    _logfile.logString(_logfile.getTime() + " BEGIN       Simulate Diffuse Material");

    StopWatch t;
    t.start();

    _updateDiffuseInfluenceGrid(dt);

    DiffuseParticleSimulationParameters params;
    params.isize = _isize;
    params.jsize = _jsize;
    params.ksize = _ksize;
    params.dx = _dx;
    params.deltaTime = dt;
    params.CFLConditionNumber = _CFLConditionNumber;
    params.markerParticleRadius = _markerParticleRadius;

    params.markerParticles = &_markerParticles;
    params.vfield = &_MACVelocity;
    params.liquidSDF = &_liquidSDF;
    params.solidSDF = &_solidSDF;

    params.meshingVolumeSDF = NULL;
    params.isMeshingVolumeSet = false;
    if (_isMeshingVolumeSet) {
        params.meshingVolumeSDF = &_meshingVolumeSDF;
        params.isMeshingVolumeSet = true;
    }

    params.surfaceSDF = &_fluidSurfaceLevelSet;
    params.curvatureGrid = &_fluidCurvatureGrid;
    params.influenceGrid = _obstacleInfluenceGrid.getInfluenceGrid();
    params.nearSolidGrid = &_nearSolidGrid;
    params.nearSolidGridCellSize = _nearSolidGridCellSize;

    params.bodyForce = _getConstantBodyForce();
    if (_isForceFieldsEnabled) {
        params.forceFieldGrid = &_forceFieldGrid;
        params.isForceFieldGridSet = true;
    }

    _diffuseMaterial.update(params);

    t.stop();
    _timingData.updateDiffuseMaterial += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Simulate Diffuse Material");
}

void FluidSimulation::_updateDiffuseInfluenceGrid(double dt) {
    int infi, infj, infk;
    _obstacleInfluenceGrid.getGridDimensions(&infi, &infj, &infk);
    if (infi != _isize + 1 || infj != _jsize + 1 || infk != _ksize + 1) {
        _obstacleInfluenceGrid = InfluenceGrid(_isize + 1, _jsize + 1, _ksize + 1, _dx, 
                                               _diffuseObstacleInfluenceBaseLevel);
    }

    _domainMeshObject.setWhitewaterInfluence(_diffuseObstacleInfluenceBaseLevel);
    _obstacleInfluenceGrid.setBaseLevel(_diffuseObstacleInfluenceBaseLevel);
    _obstacleInfluenceGrid.setDecayRate(_diffuseObstacleInfluenceDecayRate);
    _obstacleInfluenceGrid.update(&_solidSDF, dt);
}

/********************************************************************************
    #. Update Sheeting
********************************************************************************/

void FluidSimulation::_updateSheetSeeding() {
    if (!_isSheetSeedingEnabled) {
        return;
    }

    _logfile.logString(_logfile.getTime() + " BEGIN       Update Sheet Seeding");

    StopWatch t;
    t.start();


    ParticleSheeterParameters params;
    params.particles = &_markerParticles;
    params.fluidSurfaceLevelSet = &_fluidSurfaceLevelSet;
    params.isize = _isize;
    params.jsize = _jsize;
    params.ksize = _ksize;
    params.dx = _dx;
    params.sheetFillThreshold = _sheetFillThreshold;

    std::vector<vmath::vec3> sheetParticles;
    ParticleSheeter sheeter;
    sheeter.generateSheetParticles(params, sheetParticles);

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    std::vector<float> *ages;
    if (_isSurfaceAgeAttributeEnabled) {
        _markerParticles.getAttributeValues("AGE", ages);
        _updateMarkerParticleAgeAttributeGrid();
    }

    std::vector<float> *viscosities;
    if (_isSurfaceSourceViscosityAttributeEnabled) {
        _markerParticles.getAttributeValues("VISCOSITY", viscosities);
        _updateMarkerParticleViscosityAttributeGrid();
    }

    std::vector<vmath::vec3> *colors;
    if (_isSurfaceSourceColorAttributeEnabled) {
        _markerParticles.getAttributeValues("COLOR", colors);
        _updateMarkerParticleColorAttributeGrid();
    }

    vmath::vec3 goffset(0.5f * _dx, 0.5f * _dx, 0.5f * _dx);
    float solidSheetingWidth = 2.0f * _dx;
    for (size_t i = 0; i < sheetParticles.size(); i++) {
        vmath::vec3 p = sheetParticles[i];
        float sheetFillRate = _sheetFillRate;
        if (_solidSDF.trilinearInterpolate(p) < solidSheetingWidth) {
            GridIndex g = Grid3d::positionToGridIndex(p, _dx);
            if (Grid3d::isGridIndexInRange(g, _isize, _jsize, _ksize)) {
                MeshObject *obj = _solidSDF.getClosestMeshObject(g);
                if (obj != nullptr) {
                    sheetFillRate = _clamp(sheetFillRate * obj->getSheetingStrength(), 0.0f, 1.0f);
                }
            }
        }

        if (_randomDouble(0.0, 1.0) > sheetFillRate) {
            continue;
        }

        vmath::vec3 v = _savedVelocityField.evaluateVelocityAtPositionLinear(p);
        positions->push_back(p);
        velocities->push_back(v);

        if (_isSurfaceAgeAttributeEnabled) {
            float age = Interpolation::trilinearInterpolate(p - goffset, _dx, _ageAttributeGrid);
            ages->push_back(age);
        }

        if (_isSurfaceSourceViscosityAttributeEnabled) {
            float viscosity = Interpolation::trilinearInterpolate(p - goffset, _dx, _viscosityAttributeGrid);
            viscosities->push_back(viscosity);
        }

        if (_isSurfaceSourceColorAttributeEnabled) {
            float r = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridR);
            float g = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridG);
            float b = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridB);
            vmath::vec3 color(r, g, b);
            colors->push_back(color);
        }
    }

    _markerParticles.update();

    t.stop();

    _timingData.updateSheetSeeding += t.getTime();
    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Sheet Seeding");
}

/********************************************************************************
    #. Update MarkerParticle Velocities
********************************************************************************/

void FluidSimulation::_getIndicesAndGradientWeights(vmath::vec3 p, GridIndex indices[8], vmath::vec3 weights[8], int dir) {
    int U = 0; int V = 1; int W = 2;

    vmath::vec3 offset;
    float h = 0.5f * _dx;
    if (dir == U) {
        offset = vmath::vec3(0.0f, h, h);
    } else if (dir == V) {
        offset = vmath::vec3(h, 0.0f, h);
    } else if (dir == W) {
        offset = vmath::vec3(h, h, 0.0f);
    }

    p -= offset;
    GridIndex g = Grid3d::positionToGridIndex(p, _dx);
    vmath::vec3 gpos = Grid3d::GridIndexToPosition(g, _dx);
    vmath::vec3 ipos = (p - gpos) / _dx;

    indices[0] = GridIndex(g.i,     g.j,     g.k);
    indices[1] = GridIndex(g.i + 1, g.j,     g.k);
    indices[2] = GridIndex(g.i,     g.j + 1, g.k);
    indices[3] = GridIndex(g.i + 1, g.j + 1, g.k);
    indices[4] = GridIndex(g.i,     g.j,     g.k + 1);
    indices[5] = GridIndex(g.i + 1, g.j,     g.k + 1);
    indices[6] = GridIndex(g.i,     g.j + 1, g.k + 1);
    indices[7] = GridIndex(g.i + 1, g.j + 1, g.k + 1);

    float invdx = 1.0f / _dx;
    weights[0] = vmath::vec3(
        -invdx * (1.0f - ipos.y) * (1.0f - ipos.z),
        -invdx * (1.0f - ipos.x) * (1.0f - ipos.z),
        -invdx * (1.0f - ipos.x) * (1.0f - ipos.y));
    weights[1] = vmath::vec3(
        invdx * (1.0f - ipos.y) * (1.0f - ipos.z),
        ipos.x * (-invdx) * (1.0f - ipos.z),
        ipos.x * (1.0f - ipos.y) * (-invdx));
    weights[2] = vmath::vec3(
        (-invdx) * ipos.y * (1.0f - ipos.z),
        (1.0f - ipos.x) * invdx * (1.0f - ipos.z),
        (1.0f - ipos.x) * ipos.y * (-invdx));
    weights[3] = vmath::vec3(
        invdx * ipos.y * (1.0f - ipos.z),
        ipos.x * invdx * (1.0f - ipos.z),
        ipos.x * ipos.y * (-invdx));
    weights[4] = vmath::vec3(
        (-invdx) * (1.0f - ipos.y) * ipos.z,
        (1.0f - ipos.x) * (-invdx) * ipos.z,
        (1.0f - ipos.x) * (1.0f - ipos.y) * invdx);
    weights[5] = vmath::vec3(
        invdx * (1.0f - ipos.y) * ipos.z,
        ipos.x * (-invdx) * ipos.z,
        ipos.x * (1.0f - ipos.y) * invdx);
    weights[6] = vmath::vec3(
        (-invdx) * ipos.y * ipos.z,
        (1.0f - ipos.x) * invdx * ipos.z,
        (1.0f - ipos.x) * ipos.y * invdx);
    weights[7] = vmath::vec3(
        invdx * ipos.y * ipos.z,
        ipos.x * invdx * ipos.z,
        ipos.x * ipos.y * invdx);
}

void FluidSimulation::_updatePICFLIPMarkerParticleVelocitiesThread(int startidx, int endidx) {
    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    for (int i = startidx; i < endidx; i++) {
        vmath::vec3 pos = positions->at(i);
        vmath::vec3 vel = velocities->at(i);
        vmath::vec3 vPIC = _MACVelocity.evaluateVelocityAtPositionLinear(pos);
        vmath::vec3 vFLIP = vel + vPIC - _savedVelocityField.evaluateVelocityAtPositionLinear(pos);
        vmath::vec3 v = (float)_ratioPICFLIP * vPIC + (float)(1 - _ratioPICFLIP) * vFLIP;
        velocities->at(i) = v;
    }
}

/*
    The APIC (Affine Particle-In-Cell) velocity transfer method was adapted from
    Doyub Kim's 'Fluid Engine Dev' repository:
        https://github.com/doyubkim/fluid-engine-dev
*/
void FluidSimulation::_updatePICAPICMarkerParticleVelocitiesThread(int startidx, int endidx) {
    std::vector<vmath::vec3> *positions, *velocities;
    std::vector<vmath::vec3> *affineValuesX, *affineValuesY, *affineValuesZ;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);
    _markerParticles.getAttributeValues("AFFINEX", affineValuesX);
    _markerParticles.getAttributeValues("AFFINEY", affineValuesY);
    _markerParticles.getAttributeValues("AFFINEZ", affineValuesZ);

    int U = 0; int V = 1; int W = 2; 

    GridIndex indices[8];
    vmath::vec3 weights[8];
    for (int i = startidx; i < endidx; i++) {
        vmath::vec3 pos = positions->at(i);

        vmath::vec3 affineX;
        vmath::vec3 affineY;
        vmath::vec3 affineZ;
        
        _getIndicesAndGradientWeights(pos, indices, weights, U);
        for (int gidx = 0; gidx < 8; gidx++) {
            GridIndex g = indices[gidx];
            if (!_MACVelocity.isIndexInRangeU(g)) {
                continue;
            }
            affineX += weights[gidx] * _MACVelocity.U(g);
        }

        _getIndicesAndGradientWeights(pos, indices, weights, V);
        for (int gidx = 0; gidx < 8; gidx++) {
            GridIndex g = indices[gidx];
            if (!_MACVelocity.isIndexInRangeV(g)) {
                continue;
            }
            affineY += weights[gidx] * _MACVelocity.V(g);
        }

        _getIndicesAndGradientWeights(pos, indices, weights, W);
        for (int gidx = 0; gidx < 8; gidx++) {
            GridIndex g = indices[gidx];
            if (!_MACVelocity.isIndexInRangeW(g)) {
                continue;
            }
            affineZ += weights[gidx] * _MACVelocity.W(g);
        }

        velocities->at(i) = _MACVelocity.evaluateVelocityAtPositionLinear(pos);
        affineValuesX->at(i) = affineX;
        affineValuesY->at(i) = affineY;
        affineValuesZ->at(i) = affineZ;
    }
}

void FluidSimulation::_updateMarkerParticleVelocitiesThread() {
    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, _markerParticles.size());
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, _markerParticles.size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        if (_velocityTransferMethod == VelocityTransferMethod::FLIP) {
            threads[i] = std::thread(&FluidSimulation::_updatePICFLIPMarkerParticleVelocitiesThread, this,
                                     intervals[i], intervals[i + 1]);
        } else if (_velocityTransferMethod == VelocityTransferMethod::APIC) {
            threads[i] = std::thread(&FluidSimulation::_updatePICAPICMarkerParticleVelocitiesThread, this,
                                     intervals[i], intervals[i + 1]);
        }
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }
}

void FluidSimulation::_constrainMarkerParticleVelocities(MeshFluidSource *inflow) {
    float frameProgress = _getFrameInterpolation();
    int numSubsteps = inflow->getSubstepEmissions();

    if (numSubsteps == 0) {
        numSubsteps = 1;
    }

    float substepFactor = (_currentFrameTimeStep / _currentFrameDeltaTime) / (float)numSubsteps;

    Array3d<bool> isInflowCell(_isize, _jsize, _ksize, false);
    for (int subidx = 0; subidx < numSubsteps; subidx++) {
        float frameInterpolation = frameProgress + (float)subidx * substepFactor;
        inflow->setFrame(_currentFrame, frameInterpolation);
        inflow->update(_currentFrameDeltaTime);

        std::vector<GridIndex> inflowCells;
        inflow->getCells(frameInterpolation, inflowCells);
        isInflowCell.fill(false);
        isInflowCell.set(inflowCells, true);

        std::vector<vmath::vec3> *positions, *velocities;
        _markerParticles.getAttributeValues("POSITION", positions);
        _markerParticles.getAttributeValues("VELOCITY", velocities);

        MeshLevelSet *inflowSDF = inflow->getMeshLevelSet();
        vmath::vec3 v = inflow->getVelocity();
        RigidBodyVelocity rv = inflow->getRigidBodyVelocity(_currentFrameDeltaTime);
        VelocityFieldData *vdata = inflow->getVelocityFieldData();
        for (size_t i = 0; i < positions->size(); i++) {
            vmath::vec3 p = positions->at(i);
            GridIndex g = Grid3d::positionToGridIndex(p, _dx);
            if (!isInflowCell(g)) {
                continue;
            }

            if (inflowSDF->trilinearInterpolate(p) > 0.0f) {
                continue;
            }

            if (inflow->isAppendObjectVelocityEnabled()) {
                if (inflow->isRigidBody()) {
                    vmath::vec3 tv = vmath::cross(rv.angular * rv.axis, p - rv.centroid);
                    velocities->at(i) = v + rv.linear + tv;
                } else {
                    vmath::vec3 datap = p - vdata->offset;
                    vmath::vec3 fv = vdata->vfield.evaluateVelocityAtPositionLinear(datap);
                    velocities->at(i) = v + fv;
                }
            } else {
                velocities->at(i) = v;
            }
        }
    }
}

void FluidSimulation::_constrainMarkerParticleVelocities() {
    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        MeshFluidSource *source = _meshFluidSources[i];
        if (!source->isEnabled() || !source->isInflow() || 
                !source->isConstrainedFluidVelocityEnabled()) {
            continue;
        }
        _constrainMarkerParticleVelocities(source);
    }
}

void FluidSimulation::_updateMarkerParticleVelocities() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Update Marker Particle Velocities");

    StopWatch t;
    t.start();

    _updateMarkerParticleVelocitiesThread();
    _constrainMarkerParticleVelocities();

    t.stop();
    _timingData.updateMarkerParticleVelocities += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Marker Particle Velocities");
}

/********************************************************************************
    #. Update Marker Particle Attributes
********************************************************************************/

void FluidSimulation::_updateMarkerParticleVelocityAttributeGrid() {
    _velocityAttributeGrid.clear();
    _velocityAttributeValidGrid.reset();
    
    if (_markerParticles.empty()) {
        return;
    }

    VelocityAdvectorParameters params;
    params.particles = &_markerParticles;
    params.vfield = &_velocityAttributeGrid;
    params.validVelocities = &_velocityAttributeValidGrid;
    params.particleRadius = _liquidSDFParticleRadius;

    if (_velocityTransferMethod == VelocityTransferMethod::FLIP) {
        params.velocityTransferMethod = VelocityAdvectorTransferMethod::FLIP;
    } else if (_velocityTransferMethod == VelocityTransferMethod::APIC) {
        params.velocityTransferMethod = VelocityAdvectorTransferMethod::APIC;
    }
    
    _velocityAdvector.advect(params);

    _extrapolateFluidVelocities(_velocityAttributeGrid, _velocityAttributeValidGrid);

    // Contraining the velocity field can result in visual banding artifacts
    //_constrainVelocityField(_velocityAttributeGrid);
}

void FluidSimulation::_updateMarkerParticleVorticityAttributeGrid() {
    if (_isSurfaceVelocityAttributeAgainstObstaclesEnabled) {
        _velocityAttributeGrid.generateCurlAtCellCenter(_vorticityAttributeGrid);
    } else {
        _MACVelocity.generateCurlAtCellCenter(_vorticityAttributeGrid);
    }
}

void FluidSimulation::_updateMarkerParticleAgeAttributeGrid() {
    _ageAttributeGrid.fill(0.0f);
    _ageAttributeValidGrid.fill(false);

    std::vector<vmath::vec3> *positions;
    std::vector<float> *ages;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("AGE", ages);
    float radius = _ageAttributeRadius * _dx;

    AttributeTransferParameters<float> params;
    params.positions = positions;
    params.attributes = ages;
    params.attributeGrid = &_ageAttributeGrid;
    params.validGrid = &_ageAttributeValidGrid;
    params.particleRadius = radius;
    params.dx = _dx;

    AttributeToGridTransfer<float> attributeTransfer;
    attributeTransfer.transfer(params);

    GridUtils::extrapolateGrid(&_ageAttributeGrid, &_ageAttributeValidGrid, _CFLConditionNumber);
}

void FluidSimulation::_updateMarkerParticleViscosityAttributeGrid() {
    _ageAttributeGrid.fill(0.0f);
    _ageAttributeValidGrid.fill(false);

    std::vector<vmath::vec3> *positions;
    std::vector<float> *viscosities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VISCOSITY", viscosities);
    float radius = _viscosityAttributeRadius * _dx;

    AttributeTransferParameters<float> params;
    params.positions = positions;
    params.attributes = viscosities;
    params.attributeGrid = &_viscosityAttributeGrid;
    params.validGrid = &_viscosityAttributeValidGrid;
    params.particleRadius = radius;
    params.dx = _dx;

    AttributeToGridTransfer<float> attributeTransfer;
    attributeTransfer.transfer(params);

    GridUtils::extrapolateGrid(&_viscosityAttributeGrid, &_viscosityAttributeValidGrid, _CFLConditionNumber);
}

void FluidSimulation::_updateMarkerParticleColorAttributeGrid() {
    _colorAttributeGridR.fill(0.0f);
    _colorAttributeGridG.fill(0.0f);
    _colorAttributeGridB.fill(0.0f);
    _colorAttributeValidGrid.fill(false);

    Array3d<vmath::vec3> colorAttributeGrid(_isize, _jsize, _ksize, vmath::vec3());

    std::vector<vmath::vec3> *positions;
    std::vector<vmath::vec3> *colors;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("COLOR", colors);
    float radius = _colorAttributeRadius * _dx;

    AttributeTransferParameters<vmath::vec3> params;
    params.positions = positions;
    params.attributes = colors;
    params.attributeGrid = &colorAttributeGrid;
    params.validGrid = &_colorAttributeValidGrid;
    params.particleRadius = radius;
    params.dx = _dx;

    AttributeToGridTransfer<vmath::vec3> attributeTransfer;
    attributeTransfer.transfer(params);

    for (int k = 0; k < _ksize; k++) {
        for (int j = 0; j < _jsize; j++) {
            for (int i = 0; i < _isize; i++) {
                vmath::vec3 color = colorAttributeGrid(i, j, k);
                float rval = _clamp(color.x, 0.0f, 1.0f);
                float gval = _clamp(color.y, 0.0f, 1.0f);
                float bval = _clamp(color.z, 0.0f, 1.0f);
                _colorAttributeGridR.set(i, j, k, rval);
                _colorAttributeGridG.set(i, j, k, gval);
                _colorAttributeGridB.set(i, j, k, bval);
            }
        }
    }

    GridUtils::extrapolateGrid(&_colorAttributeGridR, &_colorAttributeValidGrid, _CFLConditionNumber);
    GridUtils::extrapolateGrid(&_colorAttributeGridG, &_colorAttributeValidGrid, _CFLConditionNumber);
    GridUtils::extrapolateGrid(&_colorAttributeGridB, &_colorAttributeValidGrid, _CFLConditionNumber);
}

void FluidSimulation::_updateMarkerParticleColorAttributeMixing(double dt) {
    if (!_isSurfaceSourceColorAttributeMixingEnabled) {
        return;
    }

    float eps = 1e-6;
    if (_colorAttributeMixingRate < eps || _colorAttributeMixingRadius < eps) {
        return;
    }

    std::vector<vmath::vec3> *positions;
    std::vector<vmath::vec3> *colors;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("COLOR", colors);

    SpatialPointGrid pointGrid(_isize, _jsize, _ksize, _dx);
    pointGrid.insert(*(positions));

    std::vector<vmath::vec3> colorsNew(colors->size(), vmath::vec3(0.0f, 0.0f, 0.0f));
    std::vector<bool> colorsNewValid(colors->size(), false);

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, positions->size());
    std::vector<std::thread> threads(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, positions->size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_updateMarkerParticleColorAttributeMixingThread, this,
                                 intervals[i], intervals[i + 1], dt, &pointGrid,
                                 colors, &colorsNew, &colorsNewValid);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }

    for (size_t i = 0; i < colors->size(); i++) {
        if (colorsNewValid[i]) {
            colors->at(i) = colorsNew[i];
        }
    }
}

// method adapted from https://stackoverflow.com/a/6930407
vmath::vec3 FluidSimulation::_RGBToHSV(vmath::vec3 in) {
    vmath::vec3 out;
    float min, max, delta;

    min = in.x < in.y ? in.x : in.y;
    min = min  < in.z ? min  : in.z;

    max = in.x > in.y ? in.x : in.y;
    max = max  > in.z ? max  : in.z;

    out.z = max;
    delta = max - min;
    if (delta < 0.00001) {
        out.y = 0.0f;
        out.x = 0.0f; // undefined, maybe nan?
        return out;
    }
    if (max > 0.0) { // NOTE: if Max is == 0, this divide would cause a crash
        out.y = (delta / max);                  // s
    } else {
        // if max is 0, then r = g = b = 0              
        // s = 0, h is undefined
        out.y = 0.0f;
        out.x = 0.0f;                            // its now undefined
        return out;
    }
    if (in.x >= max) {                         // > is bogus, just keeps compilor happy
        out.x = (in.y - in.z) / delta;        // between yellow & magenta
    } else {
        if( in.y >= max ) {
            out.x = 2.0 + ( in.z - in.x ) / delta;  // between cyan & yellow
        } else {
            out.x = 4.0 + ( in.x - in.y ) / delta;  // between magenta & cyan
        }
    }

    out.x *= 60.0;                              // degrees

    if( out.x < 0.0 ) {
        out.x += 360.0;
    }

    return out;
}

// method adapted from https://stackoverflow.com/a/6930407
vmath::vec3 FluidSimulation::_HSVToRGB(vmath::vec3 in) {
    float hh, p, q, t, ff;
    int i;

    if (in.y <= 0.0) {       // < is bogus, just shuts up warnings
        vmath::vec3 out(in.z, in.z, in.z);
        return out;
    }
    hh = in.x;
    if(hh >= 360.0) { 
        hh = 0.0f;
    }
    hh /= 60.0f;
    i = (int)hh;
    ff = hh - i;
    p = in.z * (1.0f - in.y);
    q = in.z * (1.0f - (in.y * ff));
    t = in.z * (1.0f - (in.y * (1.0f - ff)));

    vmath::vec3 out;
    switch(i) {
        case 0:
            out.x = in.z;
            out.y = t;
            out.z = p;
            break;
        case 1:
            out.x = q;
            out.y = in.z;
            out.z = p;
            break;
        case 2:
            out.x = p;
            out.y = in.z;
            out.z = t;
            break;

        case 3:
            out.x = p;
            out.y = q;
            out.z = in.z;
            break;
        case 4:
            out.x = t;
            out.y = p;
            out.z = in.z;
            break;
        case 5:
        default:
            out.x = in.z;
            out.y = p;
            out.z = q;
            break;
    }

    return out;     
}

void FluidSimulation::_updateMarkerParticleColorAttributeMixingThread(int startidx, int endidx, double dt,
                                                                      SpatialPointGrid *pointGrid,
                                                                      std::vector<vmath::vec3> *colors,
                                                                      std::vector<vmath::vec3> *colorsNew,
                                                                      std::vector<bool> *colorsNewValid) {
    float mixRate = std::min(_colorAttributeMixingRate * dt, 1.0);
    float searchRadius = _colorAttributeMixingRadius * _dx;

    std::vector<GridPointReference> refs;
    if (_isMixboxEnabled) {
        // Mixbox Blending
        for (int i = startidx; i < endidx; i++) {
            GridPointReference ref(i);
        
            refs.clear();
            pointGrid->queryPointReferencesInsideSphere(ref, searchRadius, refs);

            if (refs.empty()) {
                continue;
            }

            vmath::vec3 colorOld = colors->at(i);
            vmath::vec3 c0 = colors->at(refs[0].id);
            float r = c0.x;
            float g = c0.y;
            float b = c0.z;
            for (size_t ridx = 1; ridx < refs.size(); ridx++) {
                vmath::vec3 ci = colors->at(refs[ridx].id);
                float t = 1.0f / ((float)(ridx + 1));
                mixbox_lerp_srgb32f(r, g, b, ci.x, ci.y, ci.z, t, &r,&g,&b);
            }

            if (refs.size() > 0) {
                mixbox_lerp_srgb32f(colorOld.x, colorOld.y, colorOld.z, r, g, b, mixRate, &r,&g,&b);
                vmath::vec3 colorNew(r, g, b);
                colorsNew->at(i) = colorNew;
                colorsNewValid->at(i) = true;
            }
        }

    } else {
        // RGB Additive Blending
        for (int i = startidx; i < endidx; i++) {
            GridPointReference ref(i);
            
            refs.clear();
            pointGrid->queryPointReferencesInsideSphere(ref, searchRadius, refs);

            vmath::vec3 colorNew;
            for (size_t ridx = 0; ridx < refs.size(); ridx++) {
                colorNew += colors->at(refs[ridx].id);
            }

            if (refs.size() > 0) {
                colorNew /= refs.size();
                vmath::vec3 colorOld = colors->at(i);
                colorsNew->at(i) = (1.0f - mixRate) * colorOld + mixRate * colorNew;
                colorsNewValid->at(i) = true;
            }
        }

    }

}

void FluidSimulation::_updateMarkerParticleVelocityBasedAttributes() {
    if (!_isSurfaceVelocityAttributeEnabled && !_isSurfaceSpeedAttributeEnabled && !_isSurfaceVorticityAttributeEnabled) {
        return;
    }

    if (_currentFrameTimeStepNumber == 0 && _isSurfaceVelocityAttributeAgainstObstaclesEnabled) {
        _updateMarkerParticleVelocityAttributeGrid();
    }

    if (_currentFrameTimeStepNumber == 0 && _isSurfaceVorticityAttributeEnabled) {
        _updateMarkerParticleVorticityAttributeGrid();
    }
}

void FluidSimulation::_updateMarkerParticleAgeAttribute(double dt) {
    if (!_isSurfaceAgeAttributeEnabled) {
        return;
    }

    if (_currentFrameTimeStepNumber == 0) {
        _updateMarkerParticleAgeAttributeGrid();
    }

    std::vector<float> *ages;
    _markerParticles.getAttributeValues("AGE", ages);
    for (size_t i = 0; i < ages->size(); i++) {
        ages->at(i) += dt;
    }
}

void FluidSimulation::_updateMarkerParticleViscosityAttribute() {
    if (!_isSurfaceSourceViscosityAttributeEnabled) {
        return;
    }

    if (_currentFrameTimeStepNumber == 0) {
        _updateMarkerParticleViscosityAttributeGrid();
    }
}

void FluidSimulation::_updateMarkerParticleColorAttribute(double dt) {
    if (!_isSurfaceSourceColorAttributeEnabled) {
        return;
    }

    if (_currentFrameTimeStepNumber == 0) {
        _updateMarkerParticleColorAttributeGrid();
    }

    _updateMarkerParticleColorAttributeMixing(dt);
}

void FluidSimulation::_updateMarkerParticleAttributes(double dt) {
    _logfile.logString(_logfile.getTime() + " BEGIN       Update Marker Particle Attributes");

    StopWatch t;
    t.start();

    _updateMarkerParticleVelocityBasedAttributes();
    _updateMarkerParticleAgeAttribute(dt);
    _updateMarkerParticleViscosityAttribute();
    _updateMarkerParticleColorAttribute(dt);

    t.stop();
    _timingData.updateMarkerParticleVelocities += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Marker Particle Attributes");
}

/********************************************************************************
    #. Advance MarkerParticles
********************************************************************************/

vmath::vec3 FluidSimulation::_RK3(vmath::vec3 p0, double dt) {
    vmath::vec3 k1 = _MACVelocity.evaluateVelocityAtPositionLinear(p0);
    vmath::vec3 k2 = _MACVelocity.evaluateVelocityAtPositionLinear(p0 + (float)(0.5*dt)*k1);
    vmath::vec3 k3 = _MACVelocity.evaluateVelocityAtPositionLinear(p0 + (float)(0.75*dt)*k2);
    vmath::vec3 p1 = p0 + (float)(dt/9.0f)*(2.0f*k1 + 3.0f*k2 + 4.0f*k3);

    return p1;
}

void FluidSimulation::_advanceMarkerParticlesThread(double dt, int startidx, int endidx,
                                                    std::vector<vmath::vec3> *positions,
                                                    std::vector<vmath::vec3> *output) {
    for (int i = startidx; i < endidx; i++) {
        (*output)[i] = _RK3(positions->at(i), dt);
    }

    _resolveMarkerParticleCollisions(startidx, endidx, *positions, *output);
}


void FluidSimulation::_resolveMarkerParticleCollisions(int startidx, int endidx,
                                                       std::vector<vmath::vec3> &positionsOld, 
                                                       std::vector<vmath::vec3> &positionsNew) {
    AABB boundary = _getBoundaryAABB();
    boundary.expand(-_solidBufferWidth * _dx);
    for (int i = startidx; i < endidx; i++) {
        positionsNew[i] = _resolveCollision(positionsOld[i], positionsNew[i], boundary);
    }
}

vmath::vec3 FluidSimulation::_resolveCollision(vmath::vec3 oldp, vmath::vec3 newp,
                                               AABB &boundary) {

    GridIndex gridg = Grid3d::positionToGridIndex(newp, _dx);
    if (!Grid3d::isGridIndexInRange(gridg, _isize, _jsize, _ksize)) {
        newp = boundary.getNearestPointInsideAABB(newp);
    }

    GridIndex oldg = Grid3d::positionToGridIndex(oldp, _nearSolidGridCellSize);
    GridIndex newg = Grid3d::positionToGridIndex(newp, _nearSolidGridCellSize);
    if (!_nearSolidGrid(oldg) && !_nearSolidGrid(newg)) {
        return newp;
    }

    float eps = 1e-6;
    float stepDistance = _markerParticleStepDistanceFactor * (float)_dx;
    float travelDistance = (newp - oldp).length();
    if (travelDistance < eps) {
        return newp;
    }

    int numSteps = (int)std::ceil(travelDistance / stepDistance);
    vmath::vec3 stepdir = (newp - oldp).normalize();

    vmath::vec3 lastPosition = oldp;
    vmath::vec3 currentPosition;
    bool foundCollision = false;
    float collisionPhi = 0.0f;
    for (int stepidx = 0; stepidx < numSteps; stepidx++) {
        if (stepidx == numSteps - 1) {
            currentPosition = newp;
        } else {
            currentPosition = oldp + (float)(stepidx + 1) * stepDistance * stepdir;
        }

        float phi = _solidSDF.trilinearInterpolate(currentPosition);
        if (phi < 0.0f || !boundary.isPointInside(currentPosition)) {
            collisionPhi = phi;
            foundCollision = true;
            break;
        }

        lastPosition = currentPosition;
    }

    if (!foundCollision) {
        return newp;
    }

    vmath::vec3 resolvedPosition;
    float maxResolvedDistance = _CFLConditionNumber * _dx;
    vmath::vec3 grad = _solidSDF.trilinearInterpolateGradient(currentPosition);
    if (vmath::length(grad) > eps) {
        grad = vmath::normalize(grad);
        resolvedPosition = currentPosition - (collisionPhi - _solidBufferWidth * _dx) * grad;
        float resolvedPhi = _solidSDF.trilinearInterpolate(resolvedPosition);
        float resolvedDistance = vmath::length(resolvedPosition - currentPosition);
        if (resolvedPhi < 0 || resolvedDistance > maxResolvedDistance) {
            resolvedPosition = lastPosition;
        }
    } else {
        resolvedPosition = lastPosition;
    }

    if (!boundary.isPointInside(resolvedPosition)) {
        vmath::vec3 origPosition = resolvedPosition;
        resolvedPosition = boundary.getNearestPointInsideAABB(resolvedPosition);
        float resolvedPhi = _solidSDF.trilinearInterpolate(resolvedPosition);
        float resolvedDistance = vmath::length(resolvedPosition - origPosition);
        if (resolvedPhi < 0.0f || resolvedDistance > maxResolvedDistance) {
            resolvedPosition = lastPosition;
        }
    }

    return resolvedPosition;
}

float FluidSimulation::_getMarkerParticleSpeedLimit(double dt) {
    std::vector<vmath::vec3> *velocities;
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    double speedLimitStep = _CFLConditionNumber * _dx / dt;
    std::vector<int> speedLimitCounts(_maxFrameTimeSteps, 0);
    for (unsigned int i = 0; i < velocities->size(); i++) {
        double speed = (double)velocities->at(i).length();
        int speedLimitIndex = fmin(floor(speed / speedLimitStep), _maxFrameTimeSteps - 1);
        speedLimitCounts[speedLimitIndex]++;
    }

    double maxpct = _maxExtremeVelocityRemovalPercent;
    int maxabs = _maxExtremeVelocityRemovalAbsolute;
    int maxRemovalCount = fmin((int)((double)_markerParticles.size() * maxpct), maxabs);
    double maxspeed = _maxFrameTimeSteps * speedLimitStep;
    int currentRemovalCount = 0;
    for (int i = (int)speedLimitCounts.size() - 1; i > 0; i--) {
        if (currentRemovalCount + speedLimitCounts[i] > maxRemovalCount) {
            break;
        }

        currentRemovalCount += speedLimitCounts[i];
        maxspeed = std::max(i + _minTimeStepIncreaseForRemoval, _maxFrameTimeSteps) * speedLimitStep;
    }

    return maxspeed;
}

void FluidSimulation::_removeMarkerParticles(double dt) {
    Array3d<int> countGrid = Array3d<int>(_isize, _jsize, _ksize, 0);

    float maxspeed = _getMarkerParticleSpeedLimit(dt);
    double maxspeedsq = maxspeed * maxspeed;

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    std::vector<bool> isRemoved;
    _solidSDF.trilinearInterpolateSolidPoints(*positions, isRemoved);
    for (unsigned int i = 0; i < _markerParticles.size(); i++) {
        if (isRemoved[i]) {
            continue;
        }

        vmath::vec3 position = positions->at(i);
        vmath::vec3 velocity = velocities->at(i);

        GridIndex g = Grid3d::positionToGridIndex(position, _dx);
        if (countGrid(g) >= _maxMarkerParticlesPerCell) {
            isRemoved[i] = true;
            continue;
        }
        countGrid.add(g, 1);

        if (_isExtremeVelocityRemovalEnabled && 
                vmath::dot(velocity, velocity) > maxspeedsq) {
            isRemoved[i] = true;
            continue;
        }
    }

    _markerParticles.removeParticles(isRemoved);
}

void FluidSimulation::_advanceMarkerParticles(double dt) {
    _logfile.logString(_logfile.getTime() + " BEGIN       Advect Marker Particles");

    StopWatch t;
    t.start();

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);
    
    std::vector<vmath::vec3> positionsCopy = *positions;

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, positionsCopy.size());
    std::vector<std::thread> threads(numthreads);
    std::vector<vmath::vec3> output(positionsCopy.size());
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, positionsCopy.size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_advanceMarkerParticlesThread, this,
                                 dt, intervals[i], intervals[i + 1], &positionsCopy, &output);
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }

    for (size_t i = 0; i < _markerParticles.size(); i++) {
        float distanceTravelled = vmath::length(positions->at(i) - output[i]);
        if (distanceTravelled < 1e-6) {
            // In the rare case that a particle did not move, it could be
            // that this particle is stuck. Velocity should be set to 0.0
            // which helps the particle 'reset' and become unstuck.
            //velocities->at(i) = vmath::vec3();
        }
        positions->at(i) = output[i];
    }

    _removeMarkerParticles(_currentFrameDeltaTime);

    t.stop();
    _timingData.advanceMarkerParticles += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Advect Marker Particles");
}

/********************************************************************************
    #. Update Fluid Objects
********************************************************************************/

void FluidSimulation::_addNewFluidCells(std::vector<GridIndex> &cells, 
                                        vmath::vec3 velocity,
                                        MeshLevelSet &meshSDF,
                                        vmath::vec3 sdfoffset,
                                        MarkerParticleAttributes attributes,
                                        ParticleMaskGrid &maskgrid) {

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, cells.size());
    std::vector<std::thread> threads(numthreads);
    std::vector<std::vector<vmath::vec3> > particleVectors(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, cells.size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_addNewFluidCellsThread, this,
                                 intervals[i], intervals[i + 1], 
                                 &cells, &meshSDF, sdfoffset, 
                                 &(particleVectors[i]));
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }

    std::vector<MarkerParticle> newParticles;
    for (size_t vidx = 0; vidx < particleVectors.size(); vidx++) {
        for (size_t i = 0; i < particleVectors[vidx].size(); i++) {
            vmath::vec3 p = particleVectors[vidx][i];
            if (maskgrid.isSubCellSet(p)) {
                continue;
            }

            newParticles.push_back(MarkerParticle(p, velocity));
            maskgrid.addParticle(p);
        }
    }

    _addMarkerParticles(newParticles, attributes);
}

void FluidSimulation::_addNewFluidCells(std::vector<GridIndex> &cells, 
                                        vmath::vec3 velocity,
                                        RigidBodyVelocity rvelocity,
                                        MeshLevelSet &meshSDF,
                                        vmath::vec3 sdfoffset,
                                        MarkerParticleAttributes attributes,
                                        ParticleMaskGrid &maskgrid) {

    int numCPU = ThreadUtils::getMaxThreadCount();
    int numthreads = (int)fmin(numCPU, cells.size());
    std::vector<std::thread> threads(numthreads);
    std::vector<std::vector<vmath::vec3> > particleVectors(numthreads);
    std::vector<int> intervals = ThreadUtils::splitRangeIntoIntervals(0, cells.size(), numthreads);
    for (int i = 0; i < numthreads; i++) {
        threads[i] = std::thread(&FluidSimulation::_addNewFluidCellsThread, this,
                                 intervals[i], intervals[i + 1], 
                                 &cells, &meshSDF, sdfoffset, 
                                 &(particleVectors[i]));
    }

    for (int i = 0; i < numthreads; i++) {
        threads[i].join();
    }

    std::vector<MarkerParticle> newParticles;
    for (size_t vidx = 0; vidx < particleVectors.size(); vidx++) {
        for (size_t i = 0; i < particleVectors[vidx].size(); i++) {
            vmath::vec3 p = particleVectors[vidx][i];
            if (maskgrid.isSubCellSet(p)) {
                continue;
            }

            vmath::vec3 rotv = vmath::cross(rvelocity.angular * rvelocity.axis, p - rvelocity.centroid);
            vmath::vec3 totv = velocity + rvelocity.linear + rotv;
            newParticles.push_back(MarkerParticle(p, totv));
            maskgrid.addParticle(p);
        }
    }

    _addMarkerParticles(newParticles, attributes);
}

void FluidSimulation::_addNewFluidCells(std::vector<GridIndex> &cells, 
                                        vmath::vec3 velocity,
                                        VelocityFieldData *vdata,
                                        MeshLevelSet &meshSDF,
                                        vmath::vec3 sdfoffset,
                                        MarkerParticleAttributes attributes,
                                        ParticleMaskGrid &maskgrid) {
    double q = 0.25 * _dx;
    vmath::vec3 particleOffsets[8] = {
        vmath::vec3(-q, -q, -q),
        vmath::vec3( q, -q, -q),
        vmath::vec3(-q,  q, -q),
        vmath::vec3( q,  q, -q),
        vmath::vec3(-q, -q,  q),
        vmath::vec3( q, -q,  q),
        vmath::vec3(-q,  q,  q),
        vmath::vec3( q,  q,  q)
    };

    double jitter = _getMarkerParticleJitter();
    std::vector<MarkerParticle> newParticles;
    for (size_t i = 0; i < cells.size(); i++) {
        GridIndex g = cells[i];
        vmath::vec3 c = Grid3d::GridIndexToCellCenter(g, _dx);

        for (unsigned int oidx = 0; oidx < 8; oidx++) {
            vmath::vec3 p = c + particleOffsets[oidx];
            if (maskgrid.isSubCellSet(p)) {
                continue;
            }

            double d = meshSDF.trilinearInterpolate(p - sdfoffset);
            if (d > 0) {
                continue;
            }

            if (_isJitterSurfaceMarkerParticlesEnabled || d < -_dx) {
                p = _jitterMarkerParticlePosition(p, jitter);
            }

            if (_solidSDF.trilinearInterpolate(p) > 0) {
                vmath::vec3 datap = p - vdata->offset;
                vmath::vec3 fv = vdata->vfield.evaluateVelocityAtPositionLinear(datap);
                vmath::vec3 v = velocity + fv;

                newParticles.push_back(MarkerParticle(p, v));
                maskgrid.addParticle(p);
            }
        }
    }

    _addMarkerParticles(newParticles, attributes);
}

void FluidSimulation::_addNewFluidCellsAABB(AABB bbox, 
                                            vmath::vec3 velocity,
                                            MarkerParticleAttributes attributes,
                                            ParticleMaskGrid &maskgrid) {
    double q = 0.25 * _dx;
    vmath::vec3 particleOffsets[8] = {
        vmath::vec3(-q, -q, -q),
        vmath::vec3( q, -q, -q),
        vmath::vec3(-q,  q, -q),
        vmath::vec3( q,  q, -q),
        vmath::vec3(-q, -q,  q),
        vmath::vec3( q, -q,  q),
        vmath::vec3(-q,  q,  q),
        vmath::vec3( q,  q,  q)
    };

    double jitter = _getMarkerParticleJitter();

    vmath::vec3 p1 = bbox.getMinPoint();
    vmath::vec3 p2 = bbox.getMaxPoint();
    GridIndex g1 = Grid3d::positionToGridIndex(p1, _dx);
    GridIndex g2 = Grid3d::positionToGridIndex(p2, _dx);

    g1.i = std::max(g1.i, 1);
    g1.j = std::max(g1.j, 1);
    g1.k = std::max(g1.k, 1);
    g2.i = std::min(g2.i, _isize - 2);
    g2.j = std::min(g2.j, _jsize - 2);
    g2.k = std::min(g2.k, _ksize - 2);

    std::vector<MarkerParticle> newParticles;
    for (int k = g1.k; k <= g2.k; k++) {
        for (int j = g1.j; j <= g2.j; j++) {
            for (int i = g1.i; i <= g2.i; i++) {
                GridIndex g(i, j, k);
                vmath::vec3 c = Grid3d::GridIndexToCellCenter(g, _dx);

                bool isSurfaceParticle = i == g1.i || j == g1.j || k == g1.k || 
                                         i == g2.i || j == g2.j || k == g2.k;
                for (unsigned int oidx = 0; oidx < 8; oidx++) {
                    vmath::vec3 p = c + particleOffsets[oidx];
                    if (maskgrid.isSubCellSet(p)) {
                        continue;
                    }

                    if (_isJitterSurfaceMarkerParticlesEnabled || !isSurfaceParticle) {
                        p = _jitterMarkerParticlePosition(p, jitter);
                    }

                    if (_solidSDF.trilinearInterpolate(p) > 0) {
                        newParticles.push_back(MarkerParticle(p, velocity));
                        maskgrid.addParticle(p);
                    }
                }

            }
        }
    }

    _addMarkerParticles(newParticles, attributes);
}

void FluidSimulation::_addNewFluidCellsThread(int startidx, int endidx,
                                              std::vector<GridIndex> *cells, 
                                              MeshLevelSet *meshSDF,
                                              vmath::vec3 sdfoffset,
                                              std::vector<vmath::vec3> *particles) {

    double q = 0.25 * _dx;
    double jitter = _getMarkerParticleJitter();
    vmath::vec3 particleOffsets[8] = {
        vmath::vec3(-q, -q, -q),
        vmath::vec3( q, -q, -q),
        vmath::vec3(-q,  q, -q),
        vmath::vec3( q,  q, -q),
        vmath::vec3(-q, -q,  q),
        vmath::vec3( q, -q,  q),
        vmath::vec3(-q,  q,  q),
        vmath::vec3( q,  q,  q)
    };

    for (int i = startidx; i < endidx; i++) {
        GridIndex g = cells->at(i);
        vmath::vec3 c = Grid3d::GridIndexToCellCenter(g, _dx);

        for (int oidx = 0; oidx < 8; oidx++) {
            vmath::vec3 p = c + particleOffsets[oidx];

            double d = meshSDF->trilinearInterpolate(p - sdfoffset);
            if (d > 0) {
                continue;
            }

            if (_isJitterSurfaceMarkerParticlesEnabled || d < -_dx) {
                p = _jitterMarkerParticlePosition(p, jitter);
            }

            if (_solidSDF.trilinearInterpolate(p) > 0) {
                particles->push_back(p);
            }
        }
    }
}

void FluidSimulation::_updateInflowMeshFluidSource(MeshFluidSource *source,
                                                   ParticleMaskGrid &maskgrid) {
    if (!source->isEnabled()) {
        return;
    }

    float frameProgress = _getFrameInterpolation();
    int numSubsteps = source->getSubstepEmissions();

    if (numSubsteps == 0) {
        numSubsteps = 1;
        if (_currentFrameTimeStepNumber != 0) {
            return;
        }
    }

    float substepFactor = (_currentFrameTimeStep / _currentFrameDeltaTime) / (float)numSubsteps;

    MarkerParticleAttributes attributes;
    attributes.sourceID = source->getSourceID();
    attributes.sourceViscosity = source->getViscosity();
    attributes.sourceColor = source->getSourceColor();

    for (int i = 0; i < numSubsteps; i++) {
        float frameInterpolation = frameProgress + (float)i * substepFactor;
        source->setFrame(_currentFrame, frameInterpolation);
        source->update(_currentFrameDeltaTime);

        std::vector<GridIndex> sourceCells;
        source->getCells(frameInterpolation, sourceCells);

        MeshLevelSet *sourceSDF = source->getMeshLevelSet();
        vmath::vec3 sourceSDFOffset = source->getMeshLevelSetOffset(); 
        vmath::vec3 velocity = source->getVelocity();

        if (source->isAppendObjectVelocityEnabled()) {
            if (source->isRigidBody()) {
                RigidBodyVelocity rv = source->getRigidBodyVelocity(_currentFrameDeltaTime);
                _addNewFluidCells(sourceCells, velocity, rv, *sourceSDF, sourceSDFOffset, attributes,  maskgrid);
            } else {
                VelocityFieldData *vdata = source->getVelocityFieldData();
                _addNewFluidCells(sourceCells, velocity, vdata, *sourceSDF, sourceSDFOffset, attributes, maskgrid);
            }
        } else {
            _addNewFluidCells(sourceCells, velocity, *sourceSDF, sourceSDFOffset, attributes, maskgrid);
        }
    }
}

void FluidSimulation::_updateOutflowMeshFluidSource(MeshFluidSource *source) {
    if (!source->isEnabled()) {
        return;
    }

    if (!source->isFluidOutflowEnabled() && !source->isDiffuseOutflowEnabled()) {
        return;
    }

    float frameProgress = _getFrameInterpolation();

    source->setFrame(_currentFrame, frameProgress);
    source->update(_currentFrameDeltaTime);

    std::vector<GridIndex> sourceCells;
    source->getCells(frameProgress, sourceCells);
    MeshLevelSet *sourceSDF = source->getMeshLevelSet();
    vmath::vec3 offset = source->getMeshLevelSetOffset();

    Array3d<bool> isOutflowCell(_isize, _jsize, _ksize);
    if (source->isOutflowInversed()) {
        isOutflowCell.fill(true);
        isOutflowCell.set(sourceCells, false);
    } else {
        isOutflowCell.fill(false);
        isOutflowCell.set(sourceCells, true);
    }

    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    if (source->isFluidOutflowEnabled()) {
        std::vector<bool> isRemoved(_markerParticles.size(), false);
        for (size_t i = 0; i < _markerParticles.size(); i++) {
            vmath::vec3 p = positions->at(i);
            GridIndex g = Grid3d::positionToGridIndex(p, _dx);
            if (isOutflowCell(g)) {
                float d = sourceSDF->trilinearInterpolate(p - offset);
                if (source->isOutflowInversed() && d >= 0.0f) {
                    isRemoved[i] = true;
                } else if (!source->isOutflowInversed() && d < 0.0f) {
                    isRemoved[i] = true;
                }
            }
        }
        _markerParticles.removeParticles(isRemoved);
    }
    
    if (source->isDiffuseOutflowEnabled()) {
        std::vector<vmath::vec3> *positions;
        ParticleSystem* dps = _diffuseMaterial.getDiffuseParticles();
        dps->getAttributeValues("POSITION", positions);

        std::vector<bool> isRemoved(dps->size(), false);
        for (size_t i = 0; i < dps->size(); i++) {
            vmath::vec3 p = positions->at(i);
            GridIndex g = Grid3d::positionToGridIndex(p, _dx);
            if (!isOutflowCell.isIndexInRange(g)) {
                continue;
            }

            if (isOutflowCell(g)) {
                float d = sourceSDF->trilinearInterpolate(p - offset);
                if (source->isOutflowInversed() && d >= 0.0f) {
                    isRemoved[i] = true;
                } else if (!source->isOutflowInversed() && d < 0.0f) {
                    isRemoved[i] = true;
                }
            }
        }
        dps->removeParticles(isRemoved);
    }
}

void FluidSimulation::_updateInflowMeshFluidSources() {
    int numInflowSources = 0;
    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (_meshFluidSources[i]->isInflow() && _meshFluidSources[i]->isEnabled()) {
            numInflowSources++;
        }
    }

    if (numInflowSources == 0) {
        return;
    }

    // Sort fluid sources according to priority level (higher priority is added first)
    std::sort(_meshFluidSources.begin(), _meshFluidSources.end(), compareMeshFluidSourcePointerDescending);

    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    ParticleMaskGrid maskgrid(_isize, _jsize, _ksize, _dx);
    for (unsigned int i = 0; i < _markerParticles.size(); i++) {
        maskgrid.addParticle(positions->at(i));
    }

    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (_meshFluidSources[i]->isInflow()) {
            _updateInflowMeshFluidSource(_meshFluidSources[i], maskgrid);
        }
    }
}

void FluidSimulation::_updateOutflowMeshFluidSources() {
    int numOutflowSources = 0;
    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (_meshFluidSources[i]->isOutflow() && _meshFluidSources[i]->isEnabled()) {
            numOutflowSources++;
        }
    }

    if (numOutflowSources == 0) {
        return;
    }

    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        if (_meshFluidSources[i]->isOutflow()) {
            _updateOutflowMeshFluidSource(_meshFluidSources[i]);
        }
    }
}

void FluidSimulation::_updateMeshFluidSources() {
    _updateInflowMeshFluidSources();
    _updateOutflowMeshFluidSources();
}

void FluidSimulation::_updateAddedFluidMeshObjectQueue() {
    if (_addedFluidMeshObjectQueue.empty()) {
        return;
    }

    // Sort according to object priority level (higher priority is added first)
    std::sort(_addedFluidMeshObjectQueue.begin(), _addedFluidMeshObjectQueue.end(), std::greater<FluidMeshObject>());

    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    ParticleMaskGrid maskgrid(_isize, _jsize, _ksize, _dx);
    for (unsigned int i = 0; i < positions->size(); i++) {
        maskgrid.addParticle(positions->at(i));
    }

    MeshLevelSet meshSDF(_isize, _jsize, _ksize, _dx);
    meshSDF.disableVelocityData();

    std::vector<GridIndex> objectCells;
    for (size_t i = 0; i < _addedFluidMeshObjectQueue.size(); i++) {
        MeshObject object = _addedFluidMeshObjectQueue[i].object;
        vmath::vec3 velocity = _addedFluidMeshObjectQueue[i].velocity;

        MarkerParticleAttributes attributes;
        attributes.sourceID = object.getSourceID();
        attributes.sourceViscosity = object.getViscosity();
        attributes.sourceColor = object.getSourceColor();

        bool isAABB = object.isGeometryAABB();
        if (isAABB && !object.isAnimated()) {
            // Optimization for static AABB shaped geometry
            TriangleMesh m = object.getMesh();
            AABB bbox(m.vertices);
            _addNewFluidCellsAABB(bbox, velocity, attributes, maskgrid);
        } else {
            objectCells.clear();
            object.getCells(objectCells);

            TriangleMesh mesh = object.getMesh();
            meshSDF.reset();
            meshSDF.fastCalculateSignedDistanceField(mesh, _liquidLevelSetExactBand);
            vmath::vec3 offset(0.0f, 0.0f, 0.0f);

            if (object.isAppendObjectVelocityEnabled()) {
                RigidBodyVelocity rv = object.getRigidBodyVelocity(_currentFrameDeltaTime);
                _addNewFluidCells(objectCells, velocity, rv, meshSDF, offset, attributes, maskgrid);
            } else {
                _addNewFluidCells(objectCells, velocity, meshSDF, offset, attributes, maskgrid);
            }
        }
    }

    _addedFluidMeshObjectQueue.clear();
}

int FluidSimulation::_getNumFluidCells() {
    int count = 0;
    for (int k = 1; k < _ksize - 1; k++) {
        for (int j = 1; j < _jsize - 1; j++) {
            for (int i = 1; i < _isize - 1; i++) {
                if (_liquidSDF.get(i, j, k) < 0.0f) {
                   count++;
                }
            }
        }
    }

    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    if (count == 0 && !_markerParticles.empty()) {
        Array3d<bool> isFluidCell(_isize, _jsize, _ksize, false);
        for (unsigned int i = 0; i < positions->size(); i++) {
            GridIndex g = Grid3d::positionToGridIndex(positions->at(i), _dx);
            isFluidCell.set(g, true);
        }

        for (int k = 0; k < _ksize; k++) {
            for (int j = 0; j < _jsize; j++) {
                for (int i = 0; i < _isize; i++) {
                    if (isFluidCell(i, j, k)) {
                       count++;
                    }
                }
            }
        }
    }

    return count;
}

void FluidSimulation::_updateFluidObjects() {
    _logfile.logString(_logfile.getTime() + " BEGIN       Update Fluid Objects");

    StopWatch t;
    t.start();
    _updateAddedFluidMeshObjectQueue();
    _updateMeshFluidSources();
    t.stop();

    _timingData.updateFluidObjects += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Update Fluid Objects");
}

/********************************************************************************
    4.  Reconstruct Output Fluid Surface
********************************************************************************/

void FluidSimulation::_getTriangleMeshFileData(TriangleMesh &mesh, std::vector<char> &data) {
    if (_meshOutputFormat == TriangleMeshFormat::ply) {
        mesh.getMeshFileDataPLY(data);
    } else if (_meshOutputFormat == TriangleMeshFormat::bobj) {
        mesh.getMeshFileDataBOBJ(data);
    }
}

void FluidSimulation::_getForceFieldDebugFileData(std::vector<ForceFieldDebugNode> &debugNodes, 
                                                  std::vector<char> &data) {

    std::vector<float> values;
    for (size_t i = 0; i < debugNodes.size(); i++) {
        values.push_back(debugNodes[i].x);
        values.push_back(debugNodes[i].y);
        values.push_back(debugNodes[i].z);
        values.push_back(debugNodes[i].strength);
    }

    int numVertices = (int)debugNodes.size();
    int vertexDataSize = 4 * numVertices * sizeof(float);
    int dataSize = sizeof(int) + vertexDataSize;

    data.clear();
    data.resize(dataSize);
    data.shrink_to_fit();

    int byteOffset = 0;
    std::memcpy(data.data() + byteOffset, &numVertices, sizeof(int));
    byteOffset += sizeof(int);

    std::memcpy(data.data() + byteOffset, (char *)values.data(), vertexDataSize);
    byteOffset += vertexDataSize;
}


void FluidSimulation::_getFluidParticleFileData(std::vector<vmath::vec3> &particles, 
                                                std::vector<int> &binStarts, 
                                                std::vector<float> &binSpeeds, 
                                                std::vector<char> &outdata) {
    int numParticles = (int)particles.size();
    int numBins = (int)binStarts.size();
    int particleDataSize = 3 * numParticles * sizeof(float);
    int binStartsDataSize = numBins * sizeof(float);
    int binSpeedsDataSize = numBins * sizeof(float);
    int binDataSize = binStartsDataSize + binSpeedsDataSize;
    int dataSize = sizeof(int) + particleDataSize + sizeof(int) + binDataSize;

    outdata.clear();
    outdata.resize(dataSize);
    outdata.shrink_to_fit();

    int byteOffset = 0;
    std::memcpy(outdata.data() + byteOffset, &numParticles, sizeof(int));
    byteOffset += sizeof(int);

    std::memcpy(outdata.data() + byteOffset, (char*)particles.data(), particleDataSize);
    byteOffset += particleDataSize;

    std::memcpy(outdata.data() + byteOffset, &numBins, sizeof(int));
    byteOffset += sizeof(int);

    std::memcpy(outdata.data() + byteOffset, (char*)binStarts.data(), binStartsDataSize);
    byteOffset += binStartsDataSize;

    std::memcpy(outdata.data() + byteOffset, (char*)binSpeeds.data(), binSpeedsDataSize);
    byteOffset += binSpeedsDataSize;
}

std::string FluidSimulation::_numberToString(int number) {
    std::ostringstream ss;
    ss << number;
    return ss.str();
}

std::string FluidSimulation::_getFrameString(int number) {
    std::string currentFrame = _numberToString(number);
    currentFrame.insert(currentFrame.begin(), 6 - currentFrame.size(), '0');
    return currentFrame;
}

void FluidSimulation::_smoothSurfaceMesh(TriangleMesh &mesh) {
    mesh.smooth(_surfaceReconstructionSmoothingValue, 
                _surfaceReconstructionSmoothingIterations);
}

void FluidSimulation::_invertContactNormals(TriangleMesh &mesh) {
    if (!_isInvertedContactNormalsEnabled) {
        return;
    }

    float eps = _contactThresholdDistance * _dx;
    std::vector<bool> contactVertices(mesh.vertices.size(), false);
    for (size_t i = 0; i < mesh.vertices.size(); i++) {
        if (_solidSDF.trilinearInterpolate(mesh.vertices[i]) < eps) {
            contactVertices[i] = true;
        }
    }

    for (size_t i = 0; i < mesh.triangles.size(); i++) {
        Triangle t = mesh.triangles[i];
        if (contactVertices[t.tri[0]] || contactVertices[t.tri[1]] || contactVertices[t.tri[2]]) {
            int temp = t.tri[1];
            t.tri[1] = t.tri[2];
            t.tri[2] = temp;
            mesh.triangles[i] = t;
        }
    }
}

void FluidSimulation::_removeMeshNearDomain(TriangleMesh &mesh) {
    if (!_isRemoveSurfaceNearDomainEnabled) {
        return;
    }

    Array3d<bool> validCells(_isize, _jsize, _ksize, false);
    int width = 2 + _removeSurfaceNearDomainDistance;
    for (int k = 0 + width; k < _ksize - width; k++) {
        for (int j = 0 + width; j < _jsize - width; j++) {
            for (int i = 0 + width; i < _isize - width; i++) {
                validCells.set(i, j, k, true);
            }
        }
    }

    std::vector<int> removalTriangles;
    for (size_t tidx = 0; tidx < mesh.triangles.size(); tidx++) {
        Triangle t = mesh.triangles[tidx];
        vmath::vec3 centroid = (mesh.vertices[t.tri[0]] + 
                                mesh.vertices[t.tri[1]] + 
                                mesh.vertices[t.tri[2]]) / 3.0;
        GridIndex g = Grid3d::positionToGridIndex(centroid, _dx);
        if (!validCells.isIndexInRange(g) || !validCells(g)) {
            removalTriangles.push_back(tidx);
        } else {
            if (_isMeshingVolumeSet) {
                float d = _meshingVolumeSDF.trilinearInterpolate(centroid);
                if (d < _dx) {
                    removalTriangles.push_back(tidx);
                }
            }
        }
    }

    mesh.removeTriangles(removalTriangles);
    mesh.removeExtraneousVertices();
}

void FluidSimulation::_computeDomainBoundarySDF(MeshLevelSet *sdf) {
    AABB bbox = _getBoundaryAABB();
    vmath::vec3 minp = bbox.getMinPoint();
    vmath::vec3 maxp = bbox.getMaxPoint();
    GridIndex gmin = Grid3d::positionToGridIndex(minp, _dx);
    GridIndex gmax = Grid3d::positionToGridIndex(maxp, _dx);

    /*
    for (int k = gmin.k + 1; k <= gmax.k; k++) {
        for (int j = gmin.j + 1; j <= gmax.j; j++) {
            for (int i = gmin.i + 1; i <= gmax.i; i++) {
                sdf->set(i, j, k, -sdf->get(i, j, k));
            }
        }
    }
    */

    // -X side
    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = gmin.i; i <= gmin.i + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }

    // +X side
    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = gmax.i; i <= gmax.i + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }

    // -Y side
    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = gmin.j; j <= gmin.j + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }

    // +Y side
    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = gmax.j; j <= gmax.j + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }

    // -Z side
    for (int k = gmin.k; k <= gmin.k + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }

    // +Z side
    for (int k = gmax.k; k <= gmax.k + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                vmath::vec3 p = Grid3d::GridIndexToPosition(i, j, k, _dx);
                float d = std::max(bbox.getSignedDistance(p), 0.0f);
                sdf->set(i, j, k, d);
            }   
        }
    }
}

void FluidSimulation::_generateOutputSurface(TriangleMesh &surface, TriangleMesh &preview,
                                               std::vector<vmath::vec3> *particles,
                                               MeshLevelSet *solidSDF) {
    
    _applyMeshingVolumeToSDF(solidSDF);
    _filterParticlesOutsideMeshingVolume(particles);

    if (_markerParticles.empty()) {
        surface = TriangleMesh();
        preview = TriangleMesh();
        return;
    }

    if (_isObstacleMeshingOffsetEnabled) {
        float eps = 1e-9;
        float offset = (float)(_obstacleMeshingOffset * _dx);
        if (std::abs(offset) > eps) {
            // Values near boundary are unchanged so that the mesh is generated
            // directly against domain boundary
            for (int k = 3; k < _ksize - 2; k++) {
                for (int j = 3; j < _jsize - 2; j++) {
                    for (int i = 3; i < _isize - 2; i++) {
                        solidSDF->set(i, j, k, solidSDF->get(i, j, k) + offset);
                    }
                }
            }
        }
    } else {
        float fillval = 3.0 * _dx;
        for (int k = 0; k < _ksize + 1; k++) {
            for (int j = 0; j < _jsize + 1; j++) {
                for (int i = 0; i < _isize + 1; i++) {
                    solidSDF->set(i, j, k, fillval);
                }
            }
        }
        _computeDomainBoundarySDF(solidSDF);
    }

    ParticleMesherParameters params;
    params.isize = _isize;
    params.jsize = _jsize;
    params.ksize = _ksize;
    params.dx = _dx;
    params.subdivisions = _outputFluidSurfaceSubdivisionLevel;
    params.computechunks = _numSurfaceReconstructionPolygonizerSlices;
    params.radius = _markerParticleRadius*_markerParticleScale;
    params.particles = particles;
    params.solidSDF = solidSDF;
    params.isPreviewMesherEnabled = _isPreviewSurfaceMeshEnabled;
    if (_isPreviewSurfaceMeshEnabled) {
        params.previewdx = _previewdx;
    }

    ParticleMesher mesher;
    surface = mesher.meshParticles(params);
    if (_isPreviewSurfaceMeshEnabled) {
        preview = mesher.getPreviewMesh();
    }

    surface.removeMinimumTriangleCountPolyhedra(_minimumSurfacePolyhedronTriangleCount);
    _removeMeshNearDomain(surface);
    _removeMeshNearDomain(preview);
}

void FluidSimulation::_updateMeshingVolumeSDF() {
    if (!_isMeshingVolumeSet || _currentFrameTimeStepNumber != 0) {
        return;
    }

    MeshObjectStatus s = _meshingVolume->getStatus();
    _meshingVolume->clearObjectStatus();
    if (s.isStateChanged || (s.isEnabled && s.isAnimated && s.isMeshChanged)) {
        _isMeshingVolumeLevelSetUpToDate = false;
    }

    if (_isMeshingVolumeLevelSetUpToDate) {
        return;
    }

    _meshingVolumeSDF.reset();
    _meshingVolumeSDF.disableVelocityData();
    _meshingVolume->getMeshLevelSet(_currentFrameDeltaTime, 0.0, _solidLevelSetExactBand, _meshingVolumeSDF);
    _meshingVolumeSDF.negate();

    _isMeshingVolumeLevelSetUpToDate = true;
}

void FluidSimulation::_applyMeshingVolumeToSDF(MeshLevelSet *sdf) {
    if (!_isMeshingVolumeSet) {
        return;
    }

    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                float d1 = sdf->get(i, j, k);
                float d2 = _meshingVolumeSDF(i, j, k);
                if (d2 < d1) {
                    sdf->set(i, j, k, d2);
                }
            }
        }
    }
}

void FluidSimulation::_filterParticlesOutsideMeshingVolume(std::vector<vmath::vec3> *particles) {
    if (!_isMeshingVolumeSet) {
        return;
    }

    std::vector<bool> isSolid;
    _meshingVolumeSDF.trilinearInterpolateSolidPoints(*particles, isSolid);
    _removeItemsFromVector(*particles, isSolid);
}

void FluidSimulation::_generateSurfaceMotionBlurData(TriangleMesh &surface, MACVelocityField *vfield) {
    if (!_isSurfaceMotionBlurEnabled) {
        return;
    }

    TriangleMesh blurData;
    blurData.vertices.reserve(surface.vertices.size());
    double dt = _currentFrameDeltaTime;
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i];
        vmath::vec3 t = vfield->evaluateVelocityAtPositionLinear(p) * _domainScale * dt;
        blurData.vertices.push_back(t);
    }

    _getTriangleMeshFileData(blurData, _outputData.surfaceBlurData);
    _outputData.frameData.surfaceblur.enabled = 1;
    _outputData.frameData.surfaceblur.vertices = (int)blurData.vertices.size();
    _outputData.frameData.surfaceblur.triangles = (int)blurData.triangles.size();
    _outputData.frameData.surfaceblur.bytes = (unsigned int)_outputData.surfaceBlurData.size();

}

void FluidSimulation::_generateSurfaceVelocityAttributeData(TriangleMesh &surface, MACVelocityField *vfield) {
    if (!_isSurfaceVelocityAttributeEnabled && !_isSurfaceSpeedAttributeEnabled) {
        return;
    }

    TriangleMesh velocityData;
    if (_isSurfaceVelocityAttributeEnabled) {
        velocityData.vertices.reserve(surface.vertices.size());
    }

    std::vector<float> speedData;
    if (_isSurfaceSpeedAttributeEnabled) {
        speedData.reserve(surface.vertices.size());
    }

    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i];
        vmath::vec3 v = vfield->evaluateVelocityAtPositionLinear(p);

        if (_isSurfaceVelocityAttributeEnabled) {
            velocityData.vertices.push_back(v);
        }

        if (_isSurfaceSpeedAttributeEnabled) {
            speedData.push_back(v.length());
        }
    }

    if (_isSurfaceVelocityAttributeEnabled) {
        _getTriangleMeshFileData(velocityData, _outputData.surfaceVelocityAttributeData);
        _outputData.frameData.surfacevelocity.enabled = 1;
        _outputData.frameData.surfacevelocity.vertices = (int)velocityData.vertices.size();
        _outputData.frameData.surfacevelocity.triangles = (int)velocityData.triangles.size();
        _outputData.frameData.surfacevelocity.bytes = (unsigned int)_outputData.surfaceVelocityAttributeData.size();
    }

    if (_isSurfaceSpeedAttributeEnabled) {
        size_t datasize = speedData.size() * sizeof(float);
        _outputData.surfaceSpeedAttributeData = std::vector<char>(datasize);
        std::memcpy(_outputData.surfaceSpeedAttributeData.data(), (char *)speedData.data(), datasize);

        _outputData.frameData.surfacespeed.enabled = 1;
        _outputData.frameData.surfacespeed.vertices = speedData.size();
        _outputData.frameData.surfacespeed.triangles = 0;
        _outputData.frameData.surfacespeed.bytes = (unsigned int)_outputData.surfaceSpeedAttributeData.size();
    }
}

void FluidSimulation::_generateSurfaceVorticityAttributeData(TriangleMesh &surface) {
    if (!_isSurfaceVorticityAttributeEnabled) {
        return;
    }

    TriangleMesh vorticityData;
    vorticityData.vertices.reserve(surface.vertices.size());

    vmath::vec3 offset(0.5 * _dx, 0.5 * _dx, 0.5 * _dx);
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i] - offset;
        vmath::vec3 curl = Interpolation::trilinearInterpolate(p, _dx, _vorticityAttributeGrid);
        vorticityData.vertices.push_back(curl);
    }

    _getTriangleMeshFileData(vorticityData, _outputData.surfaceVorticityAttributeData);
    _outputData.frameData.surfacevorticity.enabled = 1;
    _outputData.frameData.surfacevorticity.vertices = (int)vorticityData.vertices.size();
    _outputData.frameData.surfacevorticity.triangles = (int)vorticityData.triangles.size();
    _outputData.frameData.surfacevorticity.bytes = (unsigned int)_outputData.surfaceVorticityAttributeData.size();
}

void FluidSimulation::_generateSurfaceAgeAttributeData(TriangleMesh &surface) {
    if (!_isSurfaceAgeAttributeEnabled) {
        return;
    }

    vmath::vec3 goffset(0.5f * _dx, 0.5f * _dx, 0.5f * _dx);

    std::vector<float> ageData;
    ageData.reserve(surface.vertices.size());
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i];
        float age = Interpolation::trilinearInterpolate(p - goffset, _dx, _ageAttributeGrid);
        ageData.push_back(age);
    }

    size_t datasize = ageData.size() * sizeof(float);
    _outputData.surfaceAgeAttributeData = std::vector<char>(datasize);
    std::memcpy(_outputData.surfaceAgeAttributeData.data(), (char *)ageData.data(), datasize);

    _outputData.frameData.surfaceage.enabled = 1;
    _outputData.frameData.surfaceage.vertices = ageData.size();
    _outputData.frameData.surfaceage.triangles = 0;
    _outputData.frameData.surfaceage.bytes = (unsigned int)_outputData.surfaceAgeAttributeData.size();
}

void FluidSimulation::_generateSurfaceColorAttributeData(TriangleMesh &surface) {
    if (!_isSurfaceSourceColorAttributeEnabled) {
        return;
    }

    vmath::vec3 goffset(0.5f * _dx, 0.5f * _dx, 0.5f * _dx);

    TriangleMesh colorData;
    colorData.vertices.reserve(surface.vertices.size());
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i];
        float r = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridR);
        float g = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridG);
        float b = Interpolation::trilinearInterpolate(p - goffset, _dx, _colorAttributeGridB);
        vmath::vec3 color(r, g, b);

        color = _RGBToHSV(color);
        color.y = std::min(color.y * _mixboxSaturationFactor, 1.0f);
        color = _HSVToRGB(color);
        color.x = _clamp(color.x, 0.0f, 1.0f);
        color.y = _clamp(color.y, 0.0f, 1.0f);
        color.z = _clamp(color.z, 0.0f, 1.0f);
        
        colorData.vertices.push_back(color);
    }

    _getTriangleMeshFileData(colorData, _outputData.surfaceColorAttributeData);
    _outputData.frameData.surfacecolor.enabled = 1;
    _outputData.frameData.surfacecolor.vertices = (int)colorData.vertices.size();
    _outputData.frameData.surfacecolor.triangles = (int)colorData.triangles.size();
    _outputData.frameData.surfacecolor.bytes = (unsigned int)_outputData.surfaceColorAttributeData.size();
}

void FluidSimulation::_generateSurfaceSourceIDAttributeData(TriangleMesh &surface, 
                                                            std::vector<vmath::vec3> &positions, 
                                                            std::vector<int> *sourceID) {
    if (!_isSurfaceSourceIDAttributeEnabled) {
        return;
    }

    Array3d<bool> validGrid(_isize, _jsize, _ksize, false);
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 v = surface.vertices[i];
        GridIndex g = Grid3d::positionToGridIndex(v, _dx);
        validGrid.set(g, true);
    }

    GridUtils::featherGrid26(&validGrid, ThreadUtils::getMaxThreadCount());

    int maxCellCount = 16;
    Array3d<char> cellCounts(_isize, _jsize, _ksize, (char)0);
    for (size_t i = 0; i < positions.size(); i++) {
        vmath::vec3 p = positions[i];
        GridIndex g = Grid3d::positionToGridIndex(p, _dx);
        int count = (int)cellCounts(g);
        if (!validGrid(g) || count >= maxCellCount) {
            continue;
        }

        count++;
        cellCounts.set(g, count);
    }

    int totalCount = 0;
    Array3d<int> startIndexGrid(_isize, _jsize, _ksize, -1);
    for (int k = 0; k < _ksize; k++) {
        for (int j = 0; j < _jsize; j++) {
            for (int i = 0; i < _isize; i++) {
                int count = (int)cellCounts(i, j, k);
                if (count == 0) {
                    continue;
                }

                startIndexGrid.set(i, j, k, totalCount);
                totalCount += count;
            }
        }
    }

    struct PointData {
        vmath::vec3 position;
        int sourceID = 0;
    };

    std::vector<PointData> data(totalCount);
    Array3d<int> startIndexGridCopy = startIndexGrid;
    Array3d<char> cellCountsCopy = cellCounts;
    for (size_t i = 0; i < positions.size(); i++) {
        vmath::vec3 p = positions[i];
        GridIndex g = Grid3d::positionToGridIndex(p, _dx);
        int count = (int)cellCountsCopy(g);
        if (!validGrid(g) || count == 0) {
            continue;
        }

        PointData pd;
        pd.position = p;
        pd.sourceID = sourceID->at(i);
        data[startIndexGridCopy(g)] = pd;
        startIndexGridCopy.add(g, 1);

        count--;
        cellCountsCopy.set(g, count);
    }

    std::vector<int> sourceIDData;
    sourceIDData.reserve(surface.vertices.size());
    for (size_t vidx = 0; vidx < surface.vertices.size(); vidx++) {
        vmath::vec3 v = surface.vertices[vidx];
        GridIndex g = Grid3d::positionToGridIndex(v, _dx);
        
        int imin = std::max(g.i - 1, 0);
        int jmin = std::max(g.j - 1, 0);
        int kmin = std::max(g.k - 1, 0);
        int imax = std::min(g.i + 1, _isize - 1);
        int jmax = std::min(g.j + 1, _jsize - 1);
        int kmax = std::min(g.k + 1, _ksize - 1);

        float minDistance = std::numeric_limits<float>::infinity();
        int minSourceID = -1;
        for (int k = kmin; k <= kmax; k++) {
            for (int j = jmin; j <= jmax; j++) {
                for (int i = imin; i <= imax; i++) {

                    int count = (int)cellCounts(i, j, k);
                    if (!validGrid(i, j, k) || count == 0) {
                        continue;
                    }
                    int startidx = startIndexGrid(i, j, k);
                    int endidx = startidx + count;

                    for (int pidx = startidx; pidx < endidx; pidx++) {
                        vmath::vec3 p = data[pidx].position;
                        float d = vmath::length(v - p);
                        if (d < minDistance) {
                            minDistance = d;
                            minSourceID = data[pidx].sourceID;
                        }
                    }

                }
            }
        }

        sourceIDData.push_back(minSourceID);
    }

    size_t datasize = sourceIDData.size() * sizeof(int);
    _outputData.surfaceSourceIDAttributeData = std::vector<char>(datasize);
    std::memcpy(_outputData.surfaceSourceIDAttributeData.data(), (char *)sourceIDData.data(), datasize);

    _outputData.frameData.surfacesourceid.enabled = 1;
    _outputData.frameData.surfacesourceid.vertices = sourceIDData.size();
    _outputData.frameData.surfacesourceid.triangles = 0;
    _outputData.frameData.surfacesourceid.bytes = (unsigned int)_outputData.surfaceSourceIDAttributeData.size();
}

void FluidSimulation::_generateSurfaceViscosityAttributeData(TriangleMesh &surface) {
    if (!_isSurfaceSourceViscosityAttributeEnabled) {
        return;
    }

    vmath::vec3 goffset(0.5f * _dx, 0.5f * _dx, 0.5f * _dx);

    std::vector<float> viscosityData;
    viscosityData.reserve(surface.vertices.size());
    for (size_t i = 0; i < surface.vertices.size(); i++) {
        vmath::vec3 p = surface.vertices[i];
        float viscosity = Interpolation::trilinearInterpolate(p - goffset, _dx, _viscosityAttributeGrid);
        viscosityData.push_back(viscosity);
    }

    size_t datasize = viscosityData.size() * sizeof(float);
    _outputData.surfaceViscosityAttributeData = std::vector<char>(datasize);
    std::memcpy(_outputData.surfaceViscosityAttributeData.data(), (char *)viscosityData.data(), datasize);

    _outputData.frameData.surfaceviscosity.enabled = 1;
    _outputData.frameData.surfaceviscosity.vertices = viscosityData.size();
    _outputData.frameData.surfaceviscosity.triangles = 0;
    _outputData.frameData.surfaceviscosity.bytes = (unsigned int)_outputData.surfaceViscosityAttributeData.size();
}

void FluidSimulation::_outputSurfaceMeshThread(std::vector<vmath::vec3> *particles,
                                               MeshLevelSet *solidSDF, 
                                               MACVelocityField *vfield,
                                               std::vector<int> *sourceID) {
    if (!_isSurfaceMeshReconstructionEnabled) { return; }

    _logfile.logString(_logfile.getTime() + " BEGIN       Generate Surface Mesh");

    StopWatch t;
    t.start();

    std::vector<vmath::vec3> particlesCopy;
    if (_isSurfaceSourceIDAttributeEnabled || _isSurfaceSourceColorAttributeEnabled) {
        particlesCopy = *particles;
    }

    TriangleMesh surfacemesh, previewmesh;
    _generateOutputSurface(surfacemesh, previewmesh, particles, solidSDF);
    delete particles;
    delete solidSDF;

    surfacemesh.removeMinimumTriangleCountPolyhedra(_minimumSurfacePolyhedronTriangleCount);

    _removeMeshNearDomain(surfacemesh);
    _smoothSurfaceMesh(surfacemesh);
    _smoothSurfaceMesh(previewmesh);
    _invertContactNormals(surfacemesh);

    AABB bbox = _getBoundaryAABB();
    bbox.expand(-8.0f * _dx);
    vmath::vec3 bboxcenter = bbox.position + 0.5 * vmath::vec3(bbox.width, bbox.height, bbox.depth);

    if (bbox.width >= bbox.height && bbox.width >= bbox.depth) {
        TriangleMesh addmesh;
        for (size_t i = 0; i < _vertx.size(); i += 3) {
            vmath::vec3 v(_vertx[i+0], _vertx[i+1], _vertx[i+2]);
            addmesh.vertices.push_back(v);
        }
        for (size_t i = 0; i < _trix.size(); i += 3) {
            Triangle t;
            t.tri[0] = _trix[i+0];
            t.tri[1] = _trix[i+1];
            t.tri[2] = _trix[i+2];
            addmesh.triangles.push_back(t);
        }

        float width = 0.8 * bbox.width;
        addmesh.scale(vmath::vec3(width, width, width));
        vmath::vec3 translate(bboxcenter);
        addmesh.translate(translate);
        surfacemesh.append(addmesh);
        previewmesh.append(addmesh);
    } else if (bbox.height >= bbox.width || bbox.height >= bbox.depth) {
        TriangleMesh addmesh;
        for (size_t i = 0; i < _vertx.size(); i += 3) {
            vmath::vec3 v(_verty[i+0], _verty[i+1], _verty[i+2]);
            addmesh.vertices.push_back(v);
        }
        for (size_t i = 0; i < _trix.size(); i += 3) {
            Triangle t;
            t.tri[0] = _triy[i+0];
            t.tri[1] = _triy[i+1];
            t.tri[2] = _triy[i+2];
            addmesh.triangles.push_back(t);
        }

        float width = 0.8 * bbox.height;
        addmesh.scale(vmath::vec3(width, width, width));
        vmath::vec3 translate(bboxcenter);
        addmesh.translate(translate);
        surfacemesh.append(addmesh);
        previewmesh.append(addmesh);
    } else {
        TriangleMesh addmesh;
        for (size_t i = 0; i < _vertx.size(); i += 3) {
            vmath::vec3 v(_vertz[i+0], _vertz[i+1], _vertz[i+2]);
            addmesh.vertices.push_back(v);
        }
        for (size_t i = 0; i < _trix.size(); i += 3) {
            Triangle t;
            t.tri[0] = _triz[i+0];
            t.tri[1] = _triz[i+1];
            t.tri[2] = _triz[i+2];
            addmesh.triangles.push_back(t);
        }

        float width = 0.8 * bbox.depth;
        addmesh.scale(vmath::vec3(width, width, width));
        vmath::vec3 translate(bboxcenter);
        addmesh.translate(translate);
        surfacemesh.append(addmesh);
        previewmesh.append(addmesh);
    }

    if (_isSurfaceMotionBlurEnabled) {
        TriangleMesh blurData;
        blurData.vertices.reserve(surfacemesh.vertices.size());
        double dt = _currentFrameDeltaTime;
        for (size_t i = 0; i < surfacemesh.vertices.size(); i++) {
            vmath::vec3 p = surfacemesh.vertices[i];
            vmath::vec3 t = _MACVelocity.evaluateVelocityAtPositionLinear(p) * _domainScale * dt;
            blurData.vertices.push_back(t);
        }
    }

    _generateSurfaceMotionBlurData(surfacemesh, vfield);
    _generateSurfaceVelocityAttributeData(surfacemesh, vfield);
    delete vfield;

    _generateSurfaceVorticityAttributeData(surfacemesh);

    _generateSurfaceSourceIDAttributeData(surfacemesh, particlesCopy, sourceID);
    delete sourceID;

    particlesCopy.clear();
    particlesCopy.shrink_to_fit();

    _generateSurfaceViscosityAttributeData(surfacemesh);
    _generateSurfaceAgeAttributeData(surfacemesh);
    _generateSurfaceColorAttributeData(surfacemesh);

    _invertContactNormals(surfacemesh);

    vmath::vec3 scale(_domainScale, _domainScale, _domainScale);
    surfacemesh.scale(scale);
    surfacemesh.translate(_domainOffset);

    _getTriangleMeshFileData(surfacemesh, _outputData.surfaceData);
    _outputData.frameData.surface.enabled = 1;
    _outputData.frameData.surface.vertices = (int)surfacemesh.vertices.size();
    _outputData.frameData.surface.triangles = (int)surfacemesh.triangles.size();
    _outputData.frameData.surface.bytes = (unsigned int)_outputData.surfaceData.size();

    if (_isPreviewSurfaceMeshEnabled) {
        previewmesh.scale(scale);
        previewmesh.translate(_domainOffset);

        _getTriangleMeshFileData(previewmesh, _outputData.surfacePreviewData);
        _outputData.frameData.preview.enabled = 1;
        _outputData.frameData.preview.vertices = (int)previewmesh.vertices.size();
        _outputData.frameData.preview.triangles = (int)previewmesh.triangles.size();
        _outputData.frameData.preview.bytes = (unsigned int)_outputData.surfacePreviewData.size();
    }

    t.stop();
    _timingData.outputMeshSimulationData += t.getTime();

    _logfile.logString(_logfile.getTime() + " COMPLETE    Generate Surface Mesh");
}

void FluidSimulation::_launchOutputSurfaceMeshThread() {
    std::vector<vmath::vec3> *positions;
    _markerParticles.getAttributeValues("POSITION", positions);

    // Particles will be deleted within the thread after use
    std::vector<vmath::vec3> *particles = new std::vector<vmath::vec3>();
    particles->reserve(positions->size());
    for (size_t i = 0; i < positions->size(); i++) {
        particles->push_back(positions->at(i));
    }

    // solidSDF will be deleted within the thread after use
    MeshLevelSet *tempSolidSDF = new MeshLevelSet();
    tempSolidSDF->constructMinimalSignedDistanceField(_solidSDF);

    // Velocity Field will be deleted within the thread after use
    MACVelocityField *vfield = new MACVelocityField();
    if (_isSurfaceMotionBlurEnabled || _isSurfaceVelocityAttributeEnabled || _isSurfaceSpeedAttributeEnabled) {
        if (_isSurfaceVelocityAttributeAgainstObstaclesEnabled) {
            *vfield = _velocityAttributeGrid;
        } else {
            *vfield = _MACVelocity;
        }
    }

    // SourceID will be deleted within the thread after use
    std::vector<int> *sourceID = new std::vector<int>();
    if (_isSurfaceSourceIDAttributeEnabled) {
        std::vector<int> *ids = nullptr;
        _markerParticles.getAttributeValues("SOURCEID", ids);

        sourceID->reserve(ids->size());
        for (size_t i = 0; i < ids->size(); i++) {
            sourceID->push_back(ids->at(i));
        }
    }

    _mesherThread = std::thread(&FluidSimulation::_outputSurfaceMeshThread, this,
                                particles, tempSolidSDF, vfield, sourceID);

    if (!_isAsynchronousMeshingEnabled) {
        _mesherThread.join();
    }
}

void FluidSimulation::_joinOutputSurfaceMeshThread() {
    _mesherThread.join();
}

void FluidSimulation::_outputDiffuseMaterial() {
    if (!_isDiffuseMaterialOutputEnabled) { return; }

    if (_isDiffuseMaterialFilesSeparated) {
        _diffuseMaterial.getFoamParticleFileDataWWP(_outputData.diffuseFoamData);
        _diffuseMaterial.getBubbleParticleFileDataWWP(_outputData.diffuseBubbleData);
        _diffuseMaterial.getSprayParticleFileDataWWP(_outputData.diffuseSprayData);
        _diffuseMaterial.getDustParticleFileDataWWP(_outputData.diffuseDustData);

        int nspray, nbubble, nfoam, ndust;
        _diffuseMaterial.getDiffuseParticleTypeCounts(&nfoam, 
                                                      &nbubble, 
                                                      &nspray,
                                                      &ndust);

        _outputData.frameData.foam.enabled = 1;
        _outputData.frameData.foam.vertices = nfoam;
        _outputData.frameData.foam.triangles = 0;
        _outputData.frameData.foam.bytes = _outputData.diffuseFoamData.size();

        _outputData.frameData.bubble.enabled = 1;
        _outputData.frameData.bubble.vertices = nbubble;
        _outputData.frameData.bubble.triangles = 0;
        _outputData.frameData.bubble.bytes = _outputData.diffuseBubbleData.size();

        _outputData.frameData.spray.enabled = 1;
        _outputData.frameData.spray.vertices = nspray;
        _outputData.frameData.spray.triangles = 0;
        _outputData.frameData.spray.bytes = _outputData.diffuseSprayData.size();

        _outputData.frameData.dust.enabled = 1;
        _outputData.frameData.dust.vertices = ndust;
        _outputData.frameData.dust.triangles = 0;
        _outputData.frameData.dust.bytes = _outputData.diffuseDustData.size();

        if (_isWhitewaterMotionBlurEnabled) {
            double dt = _currentFrameDeltaTime;
            _diffuseMaterial.getFoamParticleBlurFileDataWWP(_outputData.diffuseFoamBlurData, dt);
            _diffuseMaterial.getBubbleParticleBlurFileDataWWP(_outputData.diffuseBubbleBlurData, dt);
            _diffuseMaterial.getSprayParticleBlurFileDataWWP(_outputData.diffuseSprayBlurData, dt);
            _diffuseMaterial.getDustParticleBlurFileDataWWP(_outputData.diffuseDustBlurData, dt);

            _outputData.frameData.foamblur.enabled = 1;
            _outputData.frameData.foamblur.vertices = nfoam;
            _outputData.frameData.foamblur.triangles = 0;
            _outputData.frameData.foamblur.bytes = (unsigned int)_outputData.diffuseFoamBlurData.size();

            _outputData.frameData.bubbleblur.enabled = 1;
            _outputData.frameData.bubbleblur.vertices = nbubble;
            _outputData.frameData.bubbleblur.triangles = 0;
            _outputData.frameData.bubbleblur.bytes = (unsigned int)_outputData.diffuseBubbleBlurData.size();

            _outputData.frameData.sprayblur.enabled = 1;
            _outputData.frameData.sprayblur.vertices = nspray;
            _outputData.frameData.sprayblur.triangles = 0;
            _outputData.frameData.sprayblur.bytes = (unsigned int)_outputData.diffuseSprayBlurData.size();

            _outputData.frameData.dustblur.enabled = 1;
            _outputData.frameData.dustblur.vertices = ndust;
            _outputData.frameData.dustblur.triangles = 0;
            _outputData.frameData.dustblur.bytes = (unsigned int)_outputData.diffuseDustBlurData.size();
        }

        if (_isWhitewaterVelocityAttributeEnabled) {
            _diffuseMaterial.getFoamParticleVelocityAttributeFileDataWWP(_outputData.whitewaterFoamVelocityAttributeData);
            _diffuseMaterial.getBubbleParticleVelocityAttributeFileDataWWP(_outputData.whitewaterBubbleVelocityAttributeData);
            _diffuseMaterial.getSprayParticleVelocityAttributeFileDataWWP(_outputData.whitewaterSprayVelocityAttributeData);
            _diffuseMaterial.getDustParticleVelocityAttributeFileDataWWP(_outputData.whitewaterDustVelocityAttributeData);

            _outputData.frameData.foamvelocity.enabled = 1;
            _outputData.frameData.foamvelocity.vertices = nfoam;
            _outputData.frameData.foamvelocity.triangles = 0;
            _outputData.frameData.foamvelocity.bytes = (unsigned int)_outputData.whitewaterFoamVelocityAttributeData.size();

            _outputData.frameData.bubblevelocity.enabled = 1;
            _outputData.frameData.bubblevelocity.vertices = nbubble;
            _outputData.frameData.bubblevelocity.triangles = 0;
            _outputData.frameData.bubblevelocity.bytes = (unsigned int)_outputData.whitewaterBubbleVelocityAttributeData.size();

            _outputData.frameData.sprayvelocity.enabled = 1;
            _outputData.frameData.sprayvelocity.vertices = nspray;
            _outputData.frameData.sprayvelocity.triangles = 0;
            _outputData.frameData.sprayvelocity.bytes = (unsigned int)_outputData.whitewaterSprayVelocityAttributeData.size();

            _outputData.frameData.dustvelocity.enabled = 1;
            _outputData.frameData.dustvelocity.vertices = ndust;
            _outputData.frameData.dustvelocity.triangles = 0;
            _outputData.frameData.dustvelocity.bytes = (unsigned int)_outputData.whitewaterDustVelocityAttributeData.size();
        }

        if (_isWhitewaterIDAttributeEnabled) {
            _diffuseMaterial.getFoamParticleIDAttributeFileDataWWI(_outputData.whitewaterFoamIDAttributeData);
            _diffuseMaterial.getBubbleParticleIDAttributeFileDataWWI(_outputData.whitewaterBubbleIDAttributeData);
            _diffuseMaterial.getSprayParticleIDAttributeFileDataWWI(_outputData.whitewaterSprayIDAttributeData);
            _diffuseMaterial.getDustParticleIDAttributeFileDataWWI(_outputData.whitewaterDustIDAttributeData);

            _outputData.frameData.foamid.enabled = 1;
            _outputData.frameData.foamid.vertices = nfoam;
            _outputData.frameData.foamid.triangles = 0;
            _outputData.frameData.foamid.bytes = (unsigned int)_outputData.whitewaterFoamIDAttributeData.size();

            _outputData.frameData.bubbleid.enabled = 1;
            _outputData.frameData.bubbleid.vertices = nbubble;
            _outputData.frameData.bubbleid.triangles = 0;
            _outputData.frameData.bubbleid.bytes = (unsigned int)_outputData.whitewaterBubbleIDAttributeData.size();

            _outputData.frameData.sprayid.enabled = 1;
            _outputData.frameData.sprayid.vertices = nspray;
            _outputData.frameData.sprayid.triangles = 0;
            _outputData.frameData.sprayid.bytes = (unsigned int)_outputData.whitewaterSprayIDAttributeData.size();

            _outputData.frameData.dustid.enabled = 1;
            _outputData.frameData.dustid.vertices = ndust;
            _outputData.frameData.dustid.triangles = 0;
            _outputData.frameData.dustid.bytes = (unsigned int)_outputData.whitewaterDustIDAttributeData.size();
        }

        if (_isWhitewaterLifetimeAttributeEnabled) {
            _diffuseMaterial.getFoamParticleLifetimeAttributeFileDataWWF(_outputData.whitewaterFoamLifetimeAttributeData);
            _diffuseMaterial.getBubbleParticleLifetimeAttributeFileDataWWF(_outputData.whitewaterBubbleLifetimeAttributeData);
            _diffuseMaterial.getSprayParticleLifetimeAttributeFileDataWWF(_outputData.whitewaterSprayLifetimeAttributeData);
            _diffuseMaterial.getDustParticleLifetimeAttributeFileDataWWF(_outputData.whitewaterDustLifetimeAttributeData);

            _outputData.frameData.foamlifetime.enabled = 1;
            _outputData.frameData.foamlifetime.vertices = nfoam;
            _outputData.frameData.foamlifetime.triangles = 0;
            _outputData.frameData.foamlifetime.bytes = (unsigned int)_outputData.whitewaterFoamLifetimeAttributeData.size();

            _outputData.frameData.bubblelifetime.enabled = 1;
            _outputData.frameData.bubblelifetime.vertices = nbubble;
            _outputData.frameData.bubblelifetime.triangles = 0;
            _outputData.frameData.bubblelifetime.bytes = (unsigned int)_outputData.whitewaterBubbleLifetimeAttributeData.size();

            _outputData.frameData.spraylifetime.enabled = 1;
            _outputData.frameData.spraylifetime.vertices = nspray;
            _outputData.frameData.spraylifetime.triangles = 0;
            _outputData.frameData.spraylifetime.bytes = (unsigned int)_outputData.whitewaterSprayLifetimeAttributeData.size();

            _outputData.frameData.dustlifetime.enabled = 1;
            _outputData.frameData.dustlifetime.vertices = ndust;
            _outputData.frameData.dustlifetime.triangles = 0;
            _outputData.frameData.dustlifetime.bytes = (unsigned int)_outputData.whitewaterDustLifetimeAttributeData.size();
        }

    } else {
        _diffuseMaterial.getDiffuseParticleFileDataWWP(_outputData.diffuseData);
    }
}

float FluidSimulation::_calculateParticleSpeedPercentileThreshold(float pct) {
    std::vector<vmath::vec3> *velocities;
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    float eps = 1e-3;
    float maxs = fmax(_getMaximumMarkerParticleSpeed(), eps);
    float invmax = 1.0f / maxs;
    int nbins = 10000;
    std::vector<int> binCounts(nbins, 0);
    for (size_t i = 0; i < _markerParticles.size(); i++) {
        float s = vmath::length(velocities->at(i));
        int binidx = (int)fmin(floor(s * invmax * (nbins - 1)), nbins - 1);
        binCounts[binidx]++;
    }

    float pthresh = 0.995;
    int threshCount = floor(pthresh * _markerParticles.size());
    int currentCount = 0;
    float slimit = maxs;
    for (size_t i = 0; i < binCounts.size(); i++) {
        currentCount += binCounts[i];
        if (currentCount >= threshCount) {
            slimit = ((float)i / (float)(nbins - 1)) * maxs;
            break;
        }
    }

    return fmax(slimit, eps);
}

void FluidSimulation::_outputFluidParticles() {
    if (!_isFluidParticleOutputEnabled) { return; }

    std::vector<vmath::vec3> *positions, *velocities;
    _markerParticles.getAttributeValues("POSITION", positions);
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    float maxSpeed = _calculateParticleSpeedPercentileThreshold(0.995);
    float invmax = 1.0f / maxSpeed;
    int nbins = 1024;
    std::vector<int> binCounts(nbins, 0);
    for (size_t i = 0; i < velocities->size(); i++) {
        float s = vmath::length(velocities->at(i));
        int binidx = (int)fmin(floor(s * invmax * (nbins - 1)), nbins - 1);
        binCounts[binidx]++;
    }

    std::vector<int> binStarts(nbins, 0);
    std::vector<float> binSpeeds(nbins, 0);
    int currentIdx = 0;
    for (size_t i = 0; i < binCounts.size(); i++) {
        binStarts[i] = currentIdx;
        currentIdx += binCounts[i];

        binSpeeds[i] = ((float)i / (float)(nbins - 1)) * maxSpeed;
    }

    std::vector<vmath::vec3> sortedParticles(_markerParticles.size());
    std::vector<int> binStartsCopy = binStarts;
    for (size_t i = 0; i < velocities->size(); i++) {
        float s = vmath::length(velocities->at(i));
        int binidx = (int)fmin(floor(s * invmax * (nbins - 1)), nbins - 1);
        int vidx = binStartsCopy[binidx];
        binStartsCopy[binidx]++;

        vmath::vec3 p = positions->at(i);
        p *= _domainScale;
        p += _domainOffset;
        sortedParticles[vidx] = p;
    }

    _getFluidParticleFileData(sortedParticles, binStarts, binSpeeds, 
                              _outputData.fluidParticleData);

    _outputData.frameData.particles.enabled = 1;
    _outputData.frameData.particles.vertices = (int)sortedParticles.size();
    _outputData.frameData.particles.triangles = 0;
    _outputData.frameData.particles.bytes = _outputData.fluidParticleData.size();
}

void FluidSimulation::_outputInternalObstacleMesh() {
    if (!_isInternalObstacleMeshOutputEnabled) { return; }

    ScalarField field = ScalarField(_isize + 1, _jsize + 1, _ksize + 1, _dx);
    field.setSurfaceThreshold(0.0);
    for (int k = 0; k < _ksize + 1; k++) {
        for (int j = 0; j < _jsize + 1; j++) {
            for (int i = 0; i < _isize + 1; i++) {
                field.setScalarFieldValue(i, j, k, -_solidSDF(i, j, k));
                if (i <= 1 || j <= 1 || k <= 1 || 
                    i >= _isize - 1 || j >= _jsize - 1 || k >= _ksize - 1) {
                    field.setScalarFieldValue(i, j, k, _solidSDF(i, j, k));
                }
            }
        }
    }

    Polygonizer3d polygonizer(&field);
    TriangleMesh sdfmesh = polygonizer.polygonizeSurface();
    
    vmath::vec3 scale(_domainScale, _domainScale, _domainScale);
    sdfmesh.scale(scale);
    sdfmesh.translate(_domainOffset);

    _getTriangleMeshFileData(sdfmesh, _outputData.internalObstacleMeshData);

    _outputData.frameData.obstacle.enabled = 1;
    _outputData.frameData.obstacle.vertices = (int)sdfmesh.vertices.size();
    _outputData.frameData.obstacle.triangles = (int)sdfmesh.triangles.size();
    _outputData.frameData.obstacle.bytes = _outputData.internalObstacleMeshData.size();
}

void FluidSimulation::_outputForceFieldDebugData() {
    if (!_isForceFieldDebugOutputEnabled) { 
        return; 
    }

    std::vector<ForceFieldDebugNode> debugNodes;
    _forceFieldGrid.generateDebugNodes(debugNodes);

    for (size_t i = 0; i < debugNodes.size(); i++) {
        ForceFieldDebugNode n = debugNodes[i];
        n.x = n.x * _domainScale + _domainOffset.x;
        n.y = n.y * _domainScale + _domainOffset.y;
        n.z = n.z * _domainScale + _domainOffset.z;
        debugNodes[i] = n;
    }

    _getForceFieldDebugFileData(debugNodes, _outputData.forceFieldDebugData);

    _outputData.frameData.obstacle.enabled = 1;
    _outputData.frameData.obstacle.vertices = (int)debugNodes.size();
    _outputData.frameData.obstacle.triangles = 0;
    _outputData.frameData.forcefield.bytes = _outputData.forceFieldDebugData.size();
}

void FluidSimulation::_outputSimulationLogFile() {
    _outputData.logfileData = _logfile.flush();
}

void FluidSimulation::_outputSimulationData() {
    if (_currentFrameTimeStepNumber == 0) {
        _logfile.logString(_logfile.getTime() + " BEGIN       Generate Output Data");

        StopWatch t;
        t.start();
        _launchOutputSurfaceMeshThread();
        _outputDiffuseMaterial();
        _outputFluidParticles();
        _outputInternalObstacleMesh();
        _outputForceFieldDebugData();
        t.stop();

        _timingData.outputNonMeshSimulationData += t.getTime();
        _logfile.logString(_logfile.getTime() + " COMPLETE    Generate Output Data");
    }

    if (_isLastFrameTimeStep && _isAsynchronousMeshingEnabled) {
        _joinOutputSurfaceMeshThread();
    }
}

/********************************************************************************
    TIME STEP
********************************************************************************/

void FluidSimulation::_stepFluid(double dt) {
    srand(_currentFrame + _currentFrameTimeStepNumber);
    if (!_isSkippedFrame) {
        _launchUpdateObstacleObjectsThread(dt);
        _joinUpdateObstacleObjectsThread();
        _launchUpdateLiquidLevelSetThread();
        _joinUpdateLiquidLevelSetThread();
        _launchAdvectVelocityFieldThread();
        _joinAdvectVelocityFieldThread();
        _launchCalculateFluidCurvatureGridThread();
        _saveVelocityField();
        _applyBodyForcesToVelocityField(dt);
        _applyViscosityToVelocityField(dt);

        if (_isSurfaceTensionEnabled) {
            _joinCalculateFluidCurvatureGridThread();
        }

        _pressureSolve(dt);
        _constrainVelocityFields();

        if (_isDiffuseMaterialOutputEnabled) {
            _joinCalculateFluidCurvatureGridThread();
        }

        _updateDiffuseMaterial(dt);

        if (_isSheetSeedingEnabled) {
            _joinCalculateFluidCurvatureGridThread();
        }

        _updateSheetSeeding();
        _updateMarkerParticleVelocities();
        _deleteSavedVelocityField();
        _advanceMarkerParticles(dt);
        _updateFluidObjects();
        _updateMarkerParticleAttributes(dt);
        _outputSimulationData();
    }
}

double FluidSimulation::_getMaximumMeshObjectFluidVelocity(MeshObject *object, 
                                                           vmath::vec3 fluidVelocity) {
    double maxu = 0.0;
    if (object->isAppendObjectVelocityEnabled()) {
        RigidBodyVelocity rv = object->getRigidBodyVelocity(_currentFrameDeltaTime);
        TriangleMesh m = object->getMesh();
        for (size_t vidx = 0; vidx < m.vertices.size(); vidx++) {
            vmath::vec3 vert = m.vertices[vidx];
            vmath::vec3 rotv = vmath::cross(rv.angular * rv.axis, vert - rv.centroid);
            vmath::vec3 v = fluidVelocity + rv.linear + rotv;
            maxu = fmax(v.length(), maxu);
        }
    } else {
        maxu = fmax(fluidVelocity.length(), maxu);
    }

    return maxu;
}

double FluidSimulation::_predictMaximumMarkerParticleSpeed(double dt) {
    double maxu = 0.0;
    for (size_t i = 0; i < _addedFluidMeshObjectQueue.size(); i++) {
        MeshObject object = _addedFluidMeshObjectQueue[i].object;
        vmath::vec3 fluidVelocity = _addedFluidMeshObjectQueue[i].velocity;
        maxu = fmax(_getMaximumMeshObjectFluidVelocity(&object, fluidVelocity), maxu);
    }

    for (size_t i = 0; i < _meshFluidSources.size(); i++) {
        MeshFluidSource *source = _meshFluidSources[i];
        if (!source->isEnabled() || !source->isInflow()) {
            continue;
        }

        MeshObject *object = source->getMeshObject();
        vmath::vec3 fluidVelocity = source->getVelocity();
        maxu = fmax(_getMaximumMeshObjectFluidVelocity(object, fluidVelocity), maxu);
    }

    maxu += vmath::length(_getConstantBodyForce()) * dt;

    return maxu;
}

double FluidSimulation::_getMaximumMarkerParticleSpeed() {
    std::vector<vmath::vec3> *velocities;
    _markerParticles.getAttributeValues("VELOCITY", velocities);

    double maxsq = 0.0;
    for (unsigned int i = 0; i < velocities->size(); i++) {
        vmath::vec3 v = velocities->at(i);
        double distsq = vmath::dot(v, v);
        if (distsq > maxsq) {
            maxsq = distsq;
        }
    }

    return sqrt(maxsq);
}

double FluidSimulation::_getMaximumObstacleSpeed(double dt) {
    if (!_isAdaptiveObstacleTimeSteppingEnabled) {
        return 0.0;
    }

    AABB domainBounds(0.0, 0.0, 0.0, _isize * _dx, _jsize * _dx, _ksize * _dx);

    double maxu = 0.0;
    for (size_t i = 0; i < _obstacles.size(); i++) {
        MeshObject *obj = _obstacles[i];
        if (!obj->isEnabled()) {
            continue;
        }

        TriangleMesh m = obj->getMesh();
        std::vector<vmath::vec3> vels = obj->getFrameVertexVelocities(_currentFrame, dt);
        for (size_t vidx = 0; vidx < vels.size(); vidx++) {
            if (domainBounds.isPointInside(m.vertices[vidx])) {
                maxu = fmax(vels[vidx].length(), maxu);
            }
        }
    }

    return maxu;
}

double FluidSimulation::_calculateNextTimeStep(double dt) {
    double maxu = 0.0;
    if (_currentFrame == 0 && _currentFrameTimeStepNumber == 0) {
        // Fluid has not yet been added to the simulation, so estimate the
        // fluid speed
        maxu = _predictMaximumMarkerParticleSpeed(dt);
    } else {
        maxu = _getMaximumMarkerParticleSpeed();
    }
    maxu = fmax(_getMaximumObstacleSpeed(dt), maxu);

    double eps = 1e-6;
    double timeStep = _CFLConditionNumber * _dx / (maxu + eps);

    if (_isSurfaceTensionEnabled) {
        double restriction = sqrt(_dx * _dx * _dx) * sqrt(1.0 / (_surfaceTensionConstant + eps));
        timeStep = fmin(timeStep, _surfaceTensionConditionNumber * restriction);
    }

    return timeStep;
}

double FluidSimulation::_getFrameInterpolation() {
    double frameTime = _currentFrameDeltaTimeRemaining + _currentFrameTimeStep;
    return 1.0 - (frameTime / _currentFrameDeltaTime);
}

void FluidSimulation::_updateTimingData() {
    double diffuseCurvatureTimeFactor = 0.0;
    if (_isSurfaceTensionEnabled && _isDiffuseMaterialOutputEnabled) {
        diffuseCurvatureTimeFactor = 0.5;
    } else if (_isSurfaceTensionEnabled) {
        diffuseCurvatureTimeFactor = 0.0;
    } else if (_isDiffuseMaterialOutputEnabled) {
        diffuseCurvatureTimeFactor = 1.0;
    }

    _timingData.normalizeTimes();
    TimingData tdata = _timingData;
    FluidSimulationTimingStats tstats;
    tstats.total = tdata.frameTime;
    tstats.mesh = tdata.outputNonMeshSimulationData + tdata.outputMeshSimulationData;
    tstats.advection = tdata.advectVelocityField;
    tstats.particles = tdata.updateSheetSeeding + 
                       tdata.updateMarkerParticleVelocities + 
                       tdata.advanceMarkerParticles + 
                       tdata.updateLiquidLevelSet;
    tstats.pressure = tdata.pressureSolve;
    tstats.diffuse = diffuseCurvatureTimeFactor * tdata.calculateFluidCurvatureGrid + tdata.updateDiffuseMaterial;
    tstats.viscosity = tdata.applyViscosityToVelocityField;
    tstats.objects = tdata.updateObstacleObjects + tdata.updateFluidObjects;
    _outputData.frameData.timing = tstats;
}

void FluidSimulation::_logFrameInfo() {
    struct PrintData {
        PrintData() {}
        PrintData(std::string s, double t) : str(s), time(t) {}
        std::string str;
        double time = 0.0;
    };

    TimingData tdata = _timingData;
    std::vector<PrintData> data({
        PrintData("Update Obstacle Objects              ", tdata.updateObstacleObjects),
        PrintData("Update Liquid Level Set              ", tdata.updateLiquidLevelSet),
        PrintData("Advect Velocity Field                ", tdata.advectVelocityField),
        PrintData("Save Velocity Field                  ", tdata.saveVelocityField),
        PrintData("Calculate Surface Curvature          ", tdata.calculateFluidCurvatureGrid),
        PrintData("Apply Force Fields                   ", tdata.applyBodyForcesToVelocityField),
        PrintData("Apply Viscosity                      ", tdata.applyViscosityToVelocityField),
        PrintData("Solve Pressure System                ", tdata.pressureSolve),
        PrintData("Constrain Velocity Fields            ", tdata.constrainVelocityFields),
        PrintData("Simulate Diffuse Material            ", tdata.updateDiffuseMaterial),
        PrintData("Update Sheet Seeding                 ", tdata.updateSheetSeeding),
        PrintData("Update Marker Particle Attributes    ", tdata.updateMarkerParticleVelocities),
        PrintData("Delete Saved Velocity Field          ", tdata.deleteSavedVelocityField),
        PrintData("Advance Marker Particles             ", tdata.advanceMarkerParticles),
        PrintData("Update Fluid Objects                 ", tdata.updateFluidObjects),
        PrintData("Output Simulation Data               ", tdata.outputNonMeshSimulationData),
        PrintData("Generate Surface Mesh                ", tdata.outputMeshSimulationData)
    });

    _logfile.logString("*** Frame Timing Stats ***");
    _logfile.newline();

    for (size_t i = 0; i < data.size(); i++) {
        std::stringstream ss;
        ss << std::fixed << std::setw(8) << std::setprecision(3) << data[i].time;
        std::string timestr = ss.str();

        ss.str("");
        double percentval = (data[i].time / tdata.frameTime) * 100;
        ss << std::fixed << std::setprecision(1) << percentval;
        std::string pctstr = ss.str();
        if (pctstr.size() == 3) {
            pctstr.insert(0, " ");
        }

        float eps = 1e-5;
        int n = 60;
        int progress = 0;
        if (tdata.frameTime > eps) {
            progress = (data[i].time / tdata.frameTime) * n;
        }
        std::string progressBar(progress, '|');

        std::string pstring = data[i].str + timestr + "s    " + pctstr + "%  |" + progressBar;
        _logfile.logString(pstring);
    }

    _logfile.newline();
    _logfile.log("Frame Time:   ", tdata.frameTime, 3);
    _logfile.log("Total Time:   ", _totalSimulationTime, 3);
    _logfile.newline();
}

void FluidSimulation::_logStepInfo() {
    _logfile.newline();
    _logfile.logString("*** Time Step Stats ***");
    _logfile.newline();

    std::stringstream ss;
    ss << "Fluid Particles:   " << _markerParticles.size() << std::endl << 
          "Fluid Cells:       " << _getNumFluidCells();
    _logfile.logString(ss.str());

    if (_isDiffuseMaterialOutputEnabled) {
        int spraycount, bubblecount, foamcount, dustcount;
        _diffuseMaterial.getDiffuseParticleTypeCounts(&foamcount, 
                                                      &bubblecount, 
                                                      &spraycount,
                                                      &dustcount);
        std::stringstream dss;
        dss << "Diffuse Particles: " << getNumDiffuseParticles() << std::endl << 
          "    Foam:          " << foamcount << std::endl << 
          "    Bubble:        " << bubblecount << std::endl << 
          "    Spray:         " << spraycount << std::endl << 
          "    Dust:          " << dustcount;
        _logfile.newline();
        _logfile.logString(dss.str());
    }

    if (_pressureSolverStatus.size() > 0) {
        _logfile.newline();
        _logfile.logString(_pressureSolverStatus);
    }
    if (_isViscosityEnabled && _viscositySolverStatus.size() > 0) {
        _logfile.newline();
        _logfile.logString(_viscositySolverStatus);
    }
    _logfile.newline();
}

void FluidSimulation::_logGreeting() {
    _logfile.separator();
    std::stringstream ss;
    ss << "Fluid Engine Version " << VersionUtils::getLabel();
    _logfile.logString(ss.str());
    _logfile.separator();
}

void FluidSimulation::update(double dt) {
    if (!_isSimulationInitialized) {
        std::string msg = "Error: FluidSimulation must be initialized before update.\n";
        throw std::runtime_error(msg);
    }

    if (dt < 0.0) {
        std::string msg = "Error: delta time must be greater than or equal to 0.\n";
        msg += "delta time: " + _toString(dt) + "\n";
        throw std::domain_error(msg);
    }

    _timingData = TimingData();

    StopWatch frameTimer;
    frameTimer.start();

    double epsdt = 1e-6;
    _isZeroLengthDeltaTime = dt < epsdt;
    dt = std::max(dt, epsdt);

    _isCurrentFrameFinished = false;

    _currentFrameDeltaTime = dt;
    _currentFrameDeltaTimeRemaining = dt;
    _currentFrameTimeStepNumber = 0;
    bool isDebuggingEnabled = _isFluidParticleOutputEnabled || _isInternalObstacleMeshOutputEnabled || _isForceFieldDebugOutputEnabled;
    _isSkippedFrame = _isZeroLengthDeltaTime && _outputData.isInitialized && !isDebuggingEnabled;
    double substepTime = _currentFrameDeltaTime / (double)_minFrameTimeSteps;

    // Initialize status of solvers
    // Status values will be updated after pressure/viscosity solve
    _pressureSolverSuccess = true;
    _pressureSolverIterations = 0;
    _pressureSolverError = 0.0f;
    _viscositySolverSuccess = true;
    _viscositySolverIterations = 0;
    _viscositySolverError = 0.0f;

    double eps = 1e-9;
    do {
        StopWatch stepTimer;
        stepTimer.start();

        _currentFrameTimeStep = fmin(_calculateNextTimeStep(dt), 
                                     _currentFrameDeltaTimeRemaining);

        double timeCompleted = _currentFrameDeltaTime - _currentFrameDeltaTimeRemaining;
        double stepLimit = (_currentFrameTimeStepNumber + 1) * substepTime;
        if (timeCompleted + _currentFrameTimeStep > stepLimit) {
            _currentFrameTimeStep = fmin(substepTime, _currentFrameDeltaTimeRemaining);
        }

        if (_currentFrameTimeStepNumber == _maxFrameTimeSteps - 1) {
            _currentFrameTimeStep = _currentFrameDeltaTimeRemaining;
        }

        _currentFrameDeltaTimeRemaining -= _currentFrameTimeStep;
        _isLastFrameTimeStep = fabs(_currentFrameDeltaTimeRemaining) < eps;

        double frameProgress = 100 * (1.0 - _currentFrameDeltaTimeRemaining/dt);
        int numFrames = _timelineFrameEnd - _timelineFrameStart + 1;
        std::ostringstream ss;
        // ss << "Frame: " << _currentFrame << " (Step " << _currentFrameTimeStepNumber + 1 << ")\n" <<
        //      "Step time: " << _currentFrameTimeStep << " (" << frameProgress << "% of frame)\n";

        ss << "Simulation Frame: " << _currentFrame + 1 << " / " << numFrames << "\n"
              "  Timeline Frame: " << _currentFrame + _timelineFrameStart << " / " << _timelineFrameEnd << "\n"
              "         Substep: " << _currentFrameTimeStepNumber + 1 << "\n"
              "    Substep Time: " << _currentFrameTimeStep << "s (" << frameProgress << "% of frame)\n";

        _logfile.separator();
        _logfile.timestamp();
        _logfile.newline();
        _logfile.log(ss);
        _logfile.newline();

        _stepFluid(_currentFrameTimeStep);
        _logStepInfo();

        stepTimer.stop();
        _logfile.log("Step Update Time:   ", stepTimer.getTime(), 3);
        _logfile.newline();

        _currentFrameTimeStepNumber++;
    } while (_currentFrameDeltaTimeRemaining > eps);

    frameTimer.stop();
    _timingData.frameTime = frameTimer.getTime();
    _totalSimulationTime += frameTimer.getTime();

    _updateTimingData();
    _logFrameInfo();

    _outputData.frameData.frame = _currentFrame;
    _outputData.frameData.substeps = _currentFrameTimeStepNumber;
    _outputData.frameData.deltaTime = dt;
    _outputData.frameData.timing.total = frameTimer.getTime();
    _outputData.frameData.fluidParticles = (int)_markerParticles.size();
    _outputData.frameData.diffuseParticles = (int)(_diffuseMaterial.getDiffuseParticles()->size());
    
    _outputData.frameData.pressureSolverEnabled = 1;
    _outputData.frameData.pressureSolverSuccess = (int)_pressureSolverSuccess;
    _outputData.frameData.pressureSolverError = (double)_pressureSolverError;
    _outputData.frameData.pressureSolverIterations = _pressureSolverIterations;
    _outputData.frameData.pressureSolverMaxIterations = getPressureSolverMaxIterations();

    _outputData.frameData.viscositySolverEnabled = (int)_isViscosityEnabled;
    _outputData.frameData.viscositySolverSuccess = (int)_viscositySolverSuccess;
    _outputData.frameData.viscositySolverError = (double)_viscositySolverError;
    _outputData.frameData.viscositySolverIterations = _viscositySolverIterations;
    _outputData.frameData.viscositySolverMaxIterations = getPressureSolverMaxIterations();

    _outputData.isInitialized = true;

    _outputSimulationLogFile();

    _currentFrame++;

    _isCurrentFrameFinished = true;
}
